Package: EFAtools
Title: Fast and Flexible Implementations of Exploratory Factor Analysis
        Tools
Version: 1.1.0
Authors@R: 
    c(person(given = "Markus",
             family = "Steiner",
             role = c("aut", "cre"),
             email = "markus.d.steiner@gmail.com",
             comment = c(ORCID = "0000-0002-8126-0757")),
      person(given = "Silvia",
             family = "Steiner",
             role = "aut",
             email = "silvia.steiner.grieder@gmail.com",
             comment = c(ORCID = "0000-0002-0118-7722")),
      person(given = "William",
             family = "Revelle",
             role = "ctb"),
      person(given = "Max",
             family = "Auerswald",
             role = "ctb"),
      person(given = "Morten",
             family = "Moshagen",
             role = "ctb"),
      person(given = "John",
             family = "Ruscio",
             role = "ctb"),
      person(given = "Brendan",
             family = "Roche",
             role = "ctb"),
      person(given = "Urbano",
             family = "Lorenzo-Seva",
             role = "ctb"),
      person(given = "David",
             family = "Navarro-Gonzalez",
             role = "ctb"),
      person(given = "Johan",
             family = "Braeken",
             role = "ctb"),
      person(given = "Andreas",
             family = "Soteriades",
             role = "ctb"))
Description: Provides a complete workflow for exploratory factor analysis
    (EFA). It covers data screening and factorability checks, a suite of
    factor retention criteria for choosing the number of factors, and factor
    extraction by principal axis factoring, maximum likelihood, unweighted
    least squares, or diagonally weighted least squares from Pearson,
    Spearman, Kendall, polychoric, tetrachoric, or two-stage
    full-information maximum likelihood correlations. A built-in rotation
    engine offers a range of orthogonal and oblique rotations, and standard
    errors for loadings and related quantities can be obtained by analytic,
    robust, or bootstrap methods. Further tools support model averaging
    across analytic choices, multigroup EFA with factor congruence, EFA on
    multiply imputed data, Schmid-Leiman transformation, reliability
    coefficients (including McDonald's omegas), factor score estimation,
    data simulation, and power analysis. Computationally intensive
    procedures are implemented in 'C++' for speed.
Depends: R (>= 4.1.0)
License: GPL-3
Encoding: UTF-8
LazyData: true
URL: https://github.com/mdsteiner/EFAtools,
        https://mdsteiner.github.io/EFAtools/
BugReports: https://github.com/mdsteiner/EFAtools/issues
Imports: psych, stats, ggplot2 (>= 3.4.0), cli, Rcpp, future.apply,
        checkmate, progressr, rlang, clue, lifecycle
LinkingTo: Rcpp, RcppArmadillo, roptim
Suggests: testthat (>= 3.0.0), future, GPArotation (>= 2022.4-1),
        lavaan, lavaan.mi, MASS, mice, nFactors, knitr, rmarkdown,
        vdiffr, polycor, mnormt, semTools, withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-20 19:53:40 UTC; ma1159938
Author: Markus Steiner [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8126-0757>),
  Silvia Steiner [aut] (ORCID: <https://orcid.org/0000-0002-0118-7722>),
  William Revelle [ctb],
  Max Auerswald [ctb],
  Morten Moshagen [ctb],
  John Ruscio [ctb],
  Brendan Roche [ctb],
  Urbano Lorenzo-Seva [ctb],
  David Navarro-Gonzalez [ctb],
  Johan Braeken [ctb],
  Andreas Soteriades [ctb]
Maintainer: Markus Steiner <markus.d.steiner@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-21 12:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(format,OMEGA)
S3method(format,efa)
S3method(format,efa_average)
S3method(format,efa_bartlett)
S3method(format,efa_compare)
S3method(format,efa_estimate_control)
S3method(format,efa_group)
S3method(format,efa_kmo)
S3method(format,efa_loadings)
S3method(format,efa_mi)
S3method(format,efa_power)
S3method(format,efa_reliability)
S3method(format,efa_retain)
S3method(format,efa_retention)
S3method(format,efa_rotate_control)
S3method(format,efa_schmid_leiman)
S3method(format,efa_scores)
S3method(format,efa_screen)
S3method(format,efa_simulated)
S3method(format,efa_sl_loadings)
S3method(format,summary.efa)
S3method(format,summary.efa_scores)
S3method(plot,efa_average)
S3method(plot,efa_compare)
S3method(plot,efa_group)
S3method(plot,efa_power)
S3method(plot,efa_retain)
S3method(plot,efa_retention)
S3method(print,OMEGA)
S3method(print,efa)
S3method(print,efa_average)
S3method(print,efa_bartlett)
S3method(print,efa_compare)
S3method(print,efa_estimate_control)
S3method(print,efa_group)
S3method(print,efa_kmo)
S3method(print,efa_loadings)
S3method(print,efa_mi)
S3method(print,efa_power)
S3method(print,efa_reliability)
S3method(print,efa_retain)
S3method(print,efa_retention)
S3method(print,efa_rotate_control)
S3method(print,efa_schmid_leiman)
S3method(print,efa_scores)
S3method(print,efa_screen)
S3method(print,efa_simulated)
S3method(print,efa_sl_loadings)
S3method(print,summary.efa)
S3method(print,summary.efa_scores)
S3method(residuals,efa)
S3method(summary,efa)
S3method(summary,efa_mi)
S3method(summary,efa_scores)
export(BARTLETT)
export(CD)
export(COMPARE)
export(EFA)
export(EFA_AVERAGE)
export(EFA_POOLED)
export(EKC)
export(FACTOR_SCORES)
export(HULL)
export(KGC)
export(KMO)
export(MAP)
export(NEST)
export(N_FACTORS)
export(OMEGA)
export(PARALLEL)
export(PROCRUSTES)
export(SCREE)
export(SL)
export(SMT)
export(efa_average)
export(efa_bartlett)
export(efa_cd)
export(efa_compare)
export(efa_ekc)
export(efa_fit)
export(efa_group)
export(efa_hull)
export(efa_kgc)
export(efa_kmo)
export(efa_map)
export(efa_mi)
export(efa_nest)
export(efa_parallel)
export(efa_power)
export(efa_procrustes)
export(efa_reliability)
export(efa_retain)
export(efa_schmid_leiman)
export(efa_scores)
export(efa_scree)
export(efa_screen)
export(efa_simulate)
export(efa_smt)
export(estimate_control)
export(rotate_control)
importFrom(Rcpp,sourceCpp)
importFrom(rlang,.data)
importFrom(utils,modifyList)
useDynLib(EFAtools, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# EFAtools

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/EFAtools)](https://CRAN.R-project.org/package=EFAtools)
[![DOI](https://joss.theoj.org/papers/10.21105/joss.02521/status.svg)](https://doi.org/10.21105/joss.02521)

<!-- badges: end -->

The EFAtools package provides functions to perform exploratory factor
analysis (EFA) procedures and compare their solutions. The goal is to
provide state-of-the-art factor retention methods and a high degree of
flexibility in the EFA procedures. This way, implementations from R
psych and SPSS can be compared. Moreover, functions for Schmid-Leiman
transformation, and computation of omegas are provided. To speed up the
analyses, some of the iterative procedures like principal axis factoring
(PAF) are implemented in C++.

## Installation

You can install the release version from
[CRAN](https://cran.r-project.org/) with:

``` r
install.packages("EFAtools")
```

You can install the development version from
[GitHub](https://github.com/) with:

``` r
install.packages("devtools")
devtools::install_github("mdsteiner/EFAtools")
```

To also build the vignette when installing the development version, use:

``` r
install.packages("devtools")
devtools::install_github("mdsteiner/EFAtools", build_vignettes = TRUE)
```

## Example

Here are a few examples on how to perform the analyses with the
different types and how to compare the results using the `COMPARE`
function. For more details, see the vignette by running
`vignette("EFAtools", package = "EFAtools")`. The vignette provides a
high-level introduction into the functionalities of the package.

``` r
# load the package
library(EFAtools)

# Run multiple factor retention methods
N_FACTORS(test_models$baseline$cormat, N = 500)
#> Warning in N_FACTORS(test_models$baseline$cormat, N = 500): ! 'x' was a correlation matrix but CD needs raw data. Skipping CD.
#> ℹ The default implementation of EKC has changed compared to EFAtools version <= 0.5.0 to reflect the original version by Braeken and van Assen (2017). The previous version (which often yields different results from the original) is available with type = 'AM2019'. See details in the help page.
#> 
#> ── Tests for the suitability of the data for factor analysis ───────────────────
#> 
#> Bartlett's test of sphericity
#> 
#> ✔ The Bartlett's test of sphericity was significant at an alpha level of .05.
#>   These data are probably suitable for factor analysis.
#> 
#>   𝜒²(153) = 2173.28, p < .001
#> 
#> Kaiser-Meyer-Olkin criterion (KMO)
#> 
#> ✔ The overall KMO value for your data is marvellous with 0.916.
#>   These data are probably suitable for factor analysis.
#> 
#> ── Number of factors suggested by the different factor retention criteria ──────
#> 
#> • Comparison data: NA
#> • EKC (original implementation, type 'BvA2017'): 3
#> • Hull method with CAF: 3
#> • Hull method with CFI: 1
#> • Hull method with RMSEA: 1
#> • Parallel analysis with SMC: 3
#> • NEST: 3

# A type SPSS EFA to mimick the SPSS implementation with
# promax rotation
EFA_SPSS <- EFA(test_models$baseline$cormat, n_factors = 3, type = "SPSS",
                  rotation = "promax")

# look at solution
EFA_SPSS
#> 
#> EFA performed with type = 'SPSS', method = 'PAF', and rotation = 'promax'.
#> 
#> ── Rotated Loadings ────────────────────────────────────────────────────────────
#> 
#>       F1      F2      F3      h2      u2  
#> V1   -.048    .035    .613    .367    .633
#> V2   -.001    .067    .482    .277    .723
#> V3    .060    .056    .453    .283    .717
#> V4    .101   -.009    .551    .378    .622
#> V5    .157   -.018    .438    .293    .707
#> V6   -.072   -.049    .704    .399    .601
#> V7    .001    .533    .093    .357    .643
#> V8   -.016    .581    .030    .349    .651
#> V9    .038    .550   -.001    .330    .670
#> V10  -.022    .674   -.071    .383    .617
#> V11   .015    .356    .232    .297    .703
#> V12   .020    .651   -.010    .432    .568
#> V13   .614    .086   -.067    .394    .606
#> V14   .548   -.068    .088    .322    .678
#> V15   .561    .128   -.070    .363    .637
#> V16   .555   -.050    .091    .344    .656
#> V17   .664   -.037   -.027    .390    .610
#> V18   .555    .004    .050    .350    .650
#> 
#> ── Factor Intercorrelations ────────────────────────────────────────────────────
#> 
#>       F1      F2      F3  
#> F1    1.000   0.617   0.648
#> F2    0.617   1.000   0.632
#> F3    0.648   0.632   1.000
#> 
#> ── Variances Accounted for ─────────────────────────────────────────────────────
#> 
#>                       F1      F2      F3  
#> SS loadings           2.198   2.074   2.034
#> Prop Tot Var          0.122   0.115   0.113
#> Cum Prop Tot Var      0.122   0.237   0.350
#> Prop Comm Var         0.349   0.329   0.323
#> Cum Prop Comm Var     0.349   0.677   1.000
#> 
#> ── Model Fit ───────────────────────────────────────────────────────────────────
#> 
#> CAF: .50
#> df: 102

# A type psych EFA to mimick the psych::fa() implementation with
# promax rotation
EFA_psych <- EFA(test_models$baseline$cormat, n_factors = 3, type = "psych",
                  rotation = "promax")

# compare the type psych and type SPSS implementations
COMPARE(EFA_SPSS$rot_loadings, EFA_psych$rot_loadings,
        x_labels = c("SPSS", "psych"))
#> Mean [min, max] absolute difference:  0.0090 [ 0.0001,  0.0245]
#> Median absolute difference:  0.0095
#> Max decimals where all numbers are equal: 0
#> Minimum number of decimals provided: 17
#> 
#>        F1      F2      F3  
#> V1    0.0150  0.0142 -0.0195
#> V2    0.0109  0.0109 -0.0138
#> V3    0.0095  0.0103 -0.0119
#> V4    0.0118  0.0131 -0.0154
#> V5    0.0084  0.0105 -0.0109
#> V6    0.0183  0.0169 -0.0245
#> V7   -0.0026 -0.0017  0.0076
#> V8   -0.0043 -0.0035  0.0102
#> V9   -0.0055 -0.0040  0.0117
#> V10  -0.0075 -0.0066  0.0151
#> V11   0.0021  0.0029  0.0001
#> V12  -0.0064 -0.0050  0.0136
#> V13  -0.0109 -0.0019  0.0163
#> V14  -0.0049  0.0028  0.0070
#> V15  -0.0107 -0.0023  0.0161
#> V16  -0.0051  0.0028  0.0074
#> V17  -0.0096 -0.0001  0.0136
#> V18  -0.0066  0.0014  0.0098
```

<img src="man/figures/README-unnamed-chunk-5-1.png" width="100%" />

``` r

# Run EFA and compute standard errors and confidence intervals based
# on non-parametric bootstrap:
EFA(GRiPS_raw, n_factors = 1, method = "ML", se = "np-boot")
#> ℹ 'x' was not a correlation matrix. Correlations are found from entered raw data.
#> 
#> EFA performed with type = 'EFAtools', method = 'ML', and rotation = 'none'.
#> 
#> ── Unrotated Loadings ──────────────────────────────────────────────────────────
#> 
#>               F1      h2      u2  
#> fun           .796    .634    .366
#> friends       .851    .725    .275
#> enjoy         .872    .760    .240
#> hurt          .767    .588    .412
#> part          .817    .667    .333
#> commonly      .832    .692    .308
#> chances       .788    .621    .379
#> attracted     .845    .715    .285
#> 
#> ── Variances Accounted for ─────────────────────────────────────────────────────
#> 
#>               F1  
#> SS loadings   5.401
#> Prop Tot Var  0.675
#> 
#> ── Model Fit ───────────────────────────────────────────────────────────────────
#> 
#> 𝜒²(20) = 59.92, p < .001
#> CFI =1.00
#> RMSEA [90% CI] = .05 [.04; .06]
#> AIC = 19.92
#> BIC = -74.02
#> CAF = .50


# Average solution across many different EFAs with oblique rotations
EFA_AV <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                      method = c("PAF", "ML", "ULS"), rotation = "oblique",
                      show_progress = FALSE)

# look at solution
EFA_AV
#> 
#> Averaging performed with averaging method mean (trim = 0) across 162 EFAs, varying the following settings: method, init_comm, criterion_type, start_method, rotation, k_promax, P_type, and varimax_type.
#> 
#> The error rate is at 0%. Of the solutions that did not result in an error, 100% converged, 0% contained Heywood cases, and 100% were admissible.
#> 
#> 
#> ══ Indicator-to-Factor Correspondences ═════════════════════════════════════════
#> 
#> For each cell, the proportion of solutions including the respective indicator-to-factor correspondence. A salience threshold of 0.3 was used to determine indicator-to-factor correspondences.
#> 
#>       F1      F2      F3 
#> V1    .11     .00    1.00
#> V2    .11     .00    1.00
#> V3    .11     .00     .94
#> V4    .11     .00    1.00
#> V5    .11     .00     .94
#> V6    .11     .00    1.00
#> V7    .11     .94     .00
#> V8    .11    1.00     .00
#> V9    .11     .94     .00
#> V10   .11    1.00     .00
#> V11   .11     .89     .00
#> V12   .11    1.00     .00
#> V13  1.00     .00     .00
#> V14  1.00     .00     .00
#> V15  1.00     .00     .00
#> V16  1.00     .00     .00
#> V17  1.00     .00     .00
#> V18  1.00     .00     .00
#> 
#> 
#> ══ Loadings ════════════════════════════════════════════════════════════════════
#> 
#> ── Mean ────────────────────────────────────────────────────────────────────────
#> 
#>       F1      F2      F3  
#> V1    .025    .048    .576
#> V2    .060    .077    .451
#> V3    .115    .066    .425
#> V4    .157    .007    .518
#> V5    .198   -.002    .412
#> V6    .002   -.028    .658
#> V7    .074    .497    .102
#> V8    .056    .538    .046
#> V9    .100    .510    .018
#> V10   .048    .625   -.046
#> V11   .082    .336    .228
#> V12   .094    .606    .007
#> V13   .597    .083   -.047
#> V14   .531   -.056    .093
#> V15   .548    .122   -.049
#> V16   .540   -.041    .097
#> V17   .633   -.033   -.009
#> V18   .542    .009    .060
#> 
#> ── Range ───────────────────────────────────────────────────────────────────────
#> 
#>       F1      F2      F3  
#> V1    0.513   0.086   0.239
#> V2    0.431   0.093   0.186
#> V3    0.394   0.108   0.179
#> V4    0.415   0.110   0.214
#> V5    0.315   0.122   0.177
#> V6    0.514   0.104   0.267
#> V7    0.527   0.255   0.089
#> V8    0.520   0.275   0.078
#> V9    0.470   0.276   0.080
#> V10   0.533   0.313   0.097
#> V11   0.482   0.176   0.102
#> V12   0.548   0.324   0.103
#> V13   0.081   0.289   0.114
#> V14   0.063   0.220   0.117
#> V15   0.091   0.280   0.107
#> V16   0.072   0.230   0.122
#> V17   0.108   0.270   0.124
#> V18   0.081   0.246   0.118
#> 
#> 
#> ══ Factor Intercorrelations from Oblique Solutions ═════════════════════════════
#> 
#> ── Mean ────────────────────────────────────────────────────────────────────────
#> 
#>       F1      F2      F3  
#> F1    1.000   0.431   0.518
#> F2    0.431   1.000   0.454
#> F3    0.518   0.454   1.000
#> 
#> ── Range ───────────────────────────────────────────────────────────────────────
#> 
#>       F1      F2      F3  
#> F1    0.000   1.276   0.679
#> F2    1.276   0.000   1.316
#> F3    0.679   1.316   0.000
#> 
#> 
#> ══ Variances Accounted for ═════════════════════════════════════════════════════
#> 
#> ── Mean ────────────────────────────────────────────────────────────────────────
#> 
#>                   F1      F2      F3  
#> SS loadings       2.443   1.929   1.904
#> Prop Tot Var      0.136   0.107   0.106
#> Prop Comm Var     0.389   0.307   0.303
#> 
#> ── Range ───────────────────────────────────────────────────────────────────────
#> 
#>                   F1      F2      F3  
#> SS loadings       2.831   1.356   1.291
#> Prop Tot Var      0.157   0.075   0.072
#> Prop Comm Var     0.419   0.215   0.215
#> 
#> 
#> ══ Model Fit ═══════════════════════════════════════════════════════════════════
#> 
#>        M (SD) [Min; Max]
#> 𝜒²: 101.73 (34.62) [53.23; 125.98]
#> df: 102
#> p: .369 (.450) [.054; 1.000]
#> CFI: 1.00 (.00) [1.00; 1.00]
#> RMSEA: .01 (.01) [.00; .02]
#> AIC: -102.27 (34.62) [-150.77; -78.02]
#> BIC: -532.16 (34.62) [-580.66; -507.91]
#> CAF: .50 (.00) [.50; .50]
```

<img src="man/figures/README-unnamed-chunk-5-2.png" width="100%" />

``` r

# Perform a Schmid-Leiman transformation
SL <- SL(EFA_psych)

# Based on a specific salience threshold for the loadings (here: .20):
factor_corres <- SL$sl[, c("F1", "F2", "F3")] >= .2

# Compute omegas from the Schmid-Leiman solution
OMEGA(SL, factor_corres = factor_corres)
#> Omega total, omega hierarchical, omega subscale, H index, explained common variance (ECV), and percent of uncontaminated correlations (PUC) for the general factor (top row) and omegas and H index for the group factors:
#> 
#>      tot  hier   sub     H   ECV   PUC
#> g  0.883 0.750 0.122 0.845 0.668 0.706
#> F1 0.769 0.498 0.272 0.465            
#> F2 0.764 0.494 0.270 0.473            
#> F3 0.745 0.543 0.202 0.380
```

## Citation

If you use this package in your research, please acknowledge it by
citing:

Steiner, M.D., & Grieder, S.G. (2020). EFAtools: An R package with fast
and flexible implementations of exploratory factor analysis tools.
*Journal of Open Source Software*, *5*(53), 2521.
<https://doi.org/10.21105/joss.02521>

## Contribute or Report Bugs

If you want to contribute or report bugs, please open an issue on GitHub
or email us at <markus.d.steiner@gmail.com> or
<silvia.grieder@gmail.com>.

## ---- include = FALSE---------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.align = "center"
)


if (!requireNamespace("psych", quietly = TRUE)) {
      stop("Package \"psych\" needed for this vignette to work. Please install it.",
      call. = FALSE)
}


## ----setup--------------------------------------------------------------------
library(psych)
library(EFAtools)

## -----------------------------------------------------------------------------
# EFAtools::EFA with type = "psych" without rotation
EFA_psych_paf <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych")
# EFAtools::EFA with type = "SPSS" without rotation
EFA_SPSS_paf <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS")

## -----------------------------------------------------------------------------
# psych::fa without rotation
psych_paf <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "none")

## -----------------------------------------------------------------------------
# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_paf$unrot_loadings, psych_paf$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_paf$unrot_loadings, SPSS_27$DOSPERT$paf_load)


## -----------------------------------------------------------------------------
# Compare loadings from psych::fa and SPSS
COMPARE(psych_paf$loadings, SPSS_27$DOSPERT$paf_load)

## -----------------------------------------------------------------------------
## Fit the models

# EFAtools::EFA with type = "psych" with varimax rotation
EFA_psych_var <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych", rotation = "varimax")
# EFAtools::EFA with type = "SPSS" with varimax rotation
EFA_SPSS_var <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS", rotation = "varimax")
# psych::fa with varimax rotation
psych_var <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "varimax")

## Check replication of results

# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_var$rot_loadings, psych_var$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_var$rot_loadings, SPSS_27$DOSPERT$var_load)

## Compare original results (just to see the difference)

# Compare loadings from psych::fa and SPSS
COMPARE(psych_var$loadings, SPSS_27$DOSPERT$var_load)


## -----------------------------------------------------------------------------
## Fit the models

# EFAtools::EFA with type = "psych" with promax rotation
EFA_psych_pro <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych", rotation = "promax")
# EFAtools::EFA with type = "SPSS" with promax rotation
EFA_SPSS_pro <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS", rotation = "promax")
# psych::fa with promax rotation
psych_pro <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "Promax")

## Check replication of results

# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_pro$rot_loadings, psych_pro$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_pro$rot_loadings, SPSS_27$DOSPERT$pro_load)

## Compare original results (just to see the difference)

# Compare loadings from psych::fa and SPSS
COMPARE(psych_pro$loadings, SPSS_27$DOSPERT$pro_load)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.align = "center"
)


if (!requireNamespace("psych", quietly = TRUE)) {
      stop("Package \"psych\" needed for this vignette to work. Please install it.",
      call. = FALSE)
}


## ----setup--------------------------------------------------------------------
library(psych)
library(EFAtools)

## -----------------------------------------------------------------------------
# EFAtools::EFA with type = "psych" without rotation
EFA_psych_paf <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych")
# EFAtools::EFA with type = "SPSS" without rotation
EFA_SPSS_paf <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS")

## -----------------------------------------------------------------------------
# psych::fa without rotation
psych_paf <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "none")

## -----------------------------------------------------------------------------
# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_paf$unrot_loadings, psych_paf$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_paf$unrot_loadings, SPSS_27$DOSPERT$paf_load)


## -----------------------------------------------------------------------------
# Compare loadings from psych::fa and SPSS
COMPARE(psych_paf$loadings, SPSS_27$DOSPERT$paf_load)

## -----------------------------------------------------------------------------
## Fit the models

# EFAtools::EFA with type = "psych" with varimax rotation
EFA_psych_var <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych", rotation = "varimax")
# EFAtools::EFA with type = "SPSS" with varimax rotation
EFA_SPSS_var <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS", rotation = "varimax")
# psych::fa with varimax rotation
psych_var <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "varimax")

## Check replication of results

# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_var$rot_loadings, psych_var$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_var$rot_loadings, SPSS_27$DOSPERT$var_load)

## Compare original results (just to see the difference)

# Compare loadings from psych::fa and SPSS
COMPARE(psych_var$loadings, SPSS_27$DOSPERT$var_load)


## -----------------------------------------------------------------------------
## Fit the models

# EFAtools::EFA with type = "psych" with promax rotation
EFA_psych_pro <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych", rotation = "promax")
# EFAtools::EFA with type = "SPSS" with promax rotation
EFA_SPSS_pro <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS", rotation = "promax")
# psych::fa with promax rotation
psych_pro <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "Promax")

## Check replication of results

# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_pro$rot_loadings, psych_pro$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_pro$rot_loadings, SPSS_27$DOSPERT$pro_load)

## Compare original results (just to see the difference)

# Compare loadings from psych::fa and SPSS
COMPARE(psych_pro$loadings, SPSS_27$DOSPERT$pro_load)

#' Print function for KGC objects
#'
#' @param x a list of class KGC. Output from \link{KGC} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print KGC
#'
#' @examples
#' KGC_base <- KGC(test_models$baseline$cormat)
#' KGC_base
#'
print.KGC <- function(x, plot = TRUE, ...) {

  nfac_PCA <- x$n_fac_PCA
  nfac_SMC <- x$n_fac_SMC
  nfac_EFA <- x$n_fac_EFA
  eigen_type <-x$settings$eigen_type

  cat("\n")
  cat("Eigenvalues were found using ", .settings_string(eigen_type), ".",
      sep = "")
  cat("\n")
  cat("\n")

  cat(cli::rule("Number of factors suggested by Kaiser-Guttmann criterion",
                col = "blue"))
  cat("\n")
  cat("\n")

  if("PCA" %in% eigen_type){

    cat(crayon::blue(cli::symbol$bullet, "With PCA-determined eigenvalues: "),
                     crayon::bold(nfac_PCA))
    cat("\n")

  }

  if("SMC" %in% eigen_type){

    cat(crayon::blue(cli::symbol$bullet,
                     "With SMC-determined eigenvalues: "),
        crayon::bold(nfac_SMC))
    cat("\n")

  }

  if("EFA" %in% eigen_type){

    cat(crayon::blue(cli::symbol$bullet,
                     "With EFA-determined eigenvalues: "),
        crayon::bold(nfac_EFA))
    cat("\n")

  }

  cat("\n")

  if (isTRUE(plot)) {
    graphics::plot(x)
  }

}

#' Empirical Kaiser Criterion
#'
#' The empirical Kaiser criterion incorporates random sampling variations of the
#' eigenvalues from the Kaiser-Guttman criterion (\code{\link{KGC}}; see Auerswald & Moshagen
#' , 2019; Braeken & van Assen, 2017). The code is based on Braeken & van Assen, (2017) and on Auerswald and Moshagen
#' (2019).
#'
#' @param x data.frame or matrix. data.frame or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. The number of observations. Only needed if x is a correlation
#'  matrix.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#'  data is given as input. Default is  \code{"pairwise.complete.obs"}.
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#'  Default is  \code{"pearson"}.
#' @param type character. The calculation of EKC. type \code{"BvA2017"} is the original implementation; type \code{"AM2019"} differs from the original implementation but was used in simulation studies (Auerswald & Moshagen, 2019; Caron, 2025). See details.
#'  Use \code{type = c("BvA2017", "AM2019")} for both implementations. Make sure
#'  to report which version you used.
#'
#' @details The Kaiser-Guttman criterion was defined with the intend that a factor
#'  should only be extracted if it explains at least as much variance as a single
#'  factor (see \code{\link{KGC}}). However, this only applies to population-level
#'  correlation matrices. Due to sampling variation, the KGC strongly overestimates
#'  the number of factors to retrieve (e.g., Zwick & Velicer, 1986). To account
#'  for this and to introduce a factor retention method that performs well with
#'  small number of indicators and correlated factors (cases where the performance
#'  of parallel analysis, see \code{\link{PARALLEL}}, is known to deteriorate)
#'  Braeken and van Assen (2017) introduced the empirical Kaiser criterion in
#'  which a series of reference eigenvalues is created as a function of the
#'  variables-to-sample-size ratio and the observed eigenvalues.
#'
#'  Braeken and van Assen (2017) showed that "(a) EKC performs about as well as
#'  parallel analysis for data arising from the null, 1-factor, or orthogonal
#'  factors model; and (b) clearly outperforms parallel analysis for the specific
#'  case of oblique factors, particularly whenever factor intercorrelation is
#'  moderate to high and the number of variables per factor is small, which is
#'  characteristic of many applications these days" (p.463-464).
#'
#'  In EFAtools version <= 0.5.0 only the implementation of Auerswald and
#'  Moshagen (2019) was implemented (now available with
#'  \code{type = "AM2019"}). However, this implementation, that was probably also used in Caron (2025), differs from the
#'  original implementation by Braeken and van Assen (2017) in that it corrects by the reference values, i.e., without
#'  using the empirical eigenvalues used in the original implementation.
#'  Thanks to Luis Eduardo Garrido for pointing this out and to Johan Braeken for sharing
#'  sample code, based on which the original version is now implemented and used
#'  by default with \code{type = "BvA2017"}.
#'
#'  While the adapted version performed relatively well
#'  in the simulation studies by Auerswald and Moshagen (2019) and Caron (2025),
#'  the theoretical derivations of the EKC as introduced by Braeken and van Assen (2017)
#'  may no longer hold. Currently we are unaware of studies comparing the two implementations,
#'  but based on our own brief comparisons across multiple datasets, the two implementations
#'  appear to often differ substantially regarding the number of factors suggested.
#'
#'  As both implementations exist in different packages and studies, we provide both
#'  versions here. Be sure to state clearly which version you use when reporting your
#'  results to avoid confusion and ensure reproducibility.
#'
#'  The \code{EKC} function can also be called together with other factor
#'   retention criteria in the \code{\link{N_FACTORS}} function.
#'
#' @return A list of class EKC containing
#'
#' \item{eigenvalues}{A vector containing the eigenvalues found on the correlation matrix of the entered data.}
#' \item{n_factors_BvA2017}{The number of factors to retain according to the original empirical Kaiser criterion by Braeken and van Assen (2017).}
#' \item{n_factors_AM2019}{The number of factors to retain according to the adapted empirical Kaiser criterion by Auerswald and Moshagen (2019).}
#' \item{references}{The reference eigenvalues.}
#' \item{settings}{A list with the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Braeken, J., & van Assen, M. A. (2017). An empirical Kaiser criterion.
#' Psychological Methods, 22, 450 – 466. http://dx.doi.org/10.1037/met0000074
#'
#' @source Caron, P.-O. (2025). A Comparison of the Next Eigenvalue Sufficiency Test to Other Stopping Rules for the Number of Factors in Factor Analysis.
#' Educational and Psychological Measurement, Online-first publication. https://doi.org/10.1177/00131644241308528
#'
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. Psychological Bulletin, 99,
#' 432–442. http://dx.doi.org/10.1037/0033-2909.99.3.432
#'
#' @seealso Other factor retention criteria: \code{\link{CD}},
#'  \code{\link{HULL}}, \code{\link{KGC}}, \code{\link{PARALLEL}},
#'  \code{\link{SMT}}
#'
#'   \code{\link{N_FACTORS}} as a wrapper function for this and all
#'   the above-mentioned factor retention criteria.
#' @export
#'
#' @examples
#' # original implementation
#' EKC(test_models$baseline$cormat, N = 500)
EKC <- function(x, N = NA,
                use = c("pairwise.complete.obs", "all.obs",
                           "complete.obs", "everything",
                           "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall"),
                type = "BvA2017") {

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  checkmate::assert_count(N, na.ok = TRUE)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  type <- match.arg(type, choices = c("BvA2017", "AM2019"), several.ok = TRUE)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

    if (is.na(N)) {

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Argument 'N' was NA but correlation matrix was entered. Please either provide N or raw data.\n"))

    }

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }

  message(cli::col_cyan(cli::symbol$info, " The default implementation of EKC has changed compared to EFAtools version <= 0.5.0 to reflect the original version by Braeken and van Assen (2017). The previous version (which often yields different results from the original) is available with type = 'AM2019' for comparison purposes. See details in the help page.\n"))

  n_factors_BvA2017 <- NA
  refs_BvA2017 <- NA
  n_factors_AM2019 <- NA
  refs_AM2019 <- NA


  if ("BvA2017" %in% type) {



    ### implementation in Braeken & van Assen, 2017. An Empirical Kaiser
    ### Criterion. Psychological Methods, 22(3). pp. 450-466
    ### Calculation based on p. 454 and adapted code by Johan Braeken

    # n variables
    J <- ncol(R)

    # eigenvalues
    lambda <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

    # lup: asymptortic max sample eigen value under null model
    #      used as first reference eigenvalue ()
    lup <- (1 + sqrt(J/N))^2

    # correction factor
    correction_factor <- c(J,(J-cumsum(lambda))[-J])/(J:1)

    # Unrestricted EKC reference values
    l_REF <- lup* correction_factor

    # Restricted EKC reference values
    l_EKC <- l_REF
    l_EKC[l_REF < 1] <- 1

    # number of factors to retain
    temp <- cumsum(lambda > l_EKC) * cumprod(lambda > l_EKC)
    if (sum(temp) == 0) {
      n_factors_BvA2017 <- 0
    } else {
      n_factors_BvA2017 <- which.max(temp)
    }
    refs_BvA2017 <- l_EKC

  }

  if ("AM2019" %in% type) {

    # implementation based on Auerswald and Moshagen 2019:
    # https://osf.io/fnc86?view_only=d03efba1fd0f4c849a87db82e6705668


    p <- ncol(R)

    # eigenvalues
    lambda <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

    # reference values
    refs <- vector("double", p)
    for (i in seq_len(p)) {
      refs[i] <- max( ((1 + sqrt(p / N))^2) * (p - sum(refs))/
                        (p - i + 1), 1)

    }

    n_factors_AM2019 <- which(lambda <= refs)[1] - 1
    refs_AM2019 <- refs

  }


  out <- list(
    eigenvalues = lambda,
    n_factors_BvA2017 = n_factors_BvA2017,
    n_factors_AM2019 = n_factors_AM2019,
    references = data.frame(
      BvA2017 = refs_BvA2017,
      AM2019 = refs_AM2019
    ),
    settings = list(
      use = use,
      cor_method = cor_method,
      N = N,
      type = type
    )
  )

  class(out) <- "EKC"

  return(out)

}

ekc_cor <- EKC(test_models$baseline$cormat, N = 500)
ekc_raw <- EKC(GRiPS_raw)

test_that("output class and dimensions are correct", {
  expect_is(ekc_cor, "EKC")
  expect_output(str(ekc_cor), "List of 4")
  expect_is(ekc_raw, "EKC")
  expect_output(str(ekc_raw), "List of 4")
})


test_that("found eigenvalues are correct", {
  expect_equal(sum(ekc_cor$eigenvalues),
               ncol(test_models$baseline$cormat))
  expect_equal(sum(ekc_raw$eigenvalues), ncol(GRiPS_raw))
  expect_length(ekc_cor$eigenvalues,
                ncol(test_models$baseline$cormat))
  expect_length(ekc_raw$eigenvalues, ncol(GRiPS_raw))
})

test_that("reference eigenvalues are correct", {
  expect_equal(ekc_cor$references$BvA2017[floor(ncol(test_models$baseline$cormat) / 2)], 1)
  expect_equal(ekc_raw$references$BvA2017[floor(ncol(GRiPS_raw) / 2)], 1)
  expect_length(ekc_cor$references$BvA2017,
                length(ekc_cor$eigenvalues))
  expect_length(ekc_raw$references$BvA2017, length(ekc_raw$eigenvalues))
})

test_that("identified number of factors is correct", {
  expect_equal(ekc_cor$n_factors_BvA2017, 3)
  expect_equal(ekc_raw$n_factors_BvA2017, 1)
})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z, rnorm(10), rnorm(10), rnorm(10)), ncol = 6)
cor_sing <- stats::cor(dat_sing)

cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

test_that("errors are thrown correctly", {
  expect_error(EKC(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_error(EKC(test_models$baseline$cormat), " Argument 'N' was NA but correlation matrix was entered. Please either provide N or raw data.\n")
  expect_message(EKC(GRiPS_raw), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_warning(EKC(GRiPS_raw, N = 20), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(EKC(dat_sing), " Correlation matrix is singular, no further analyses are performed\n")
  expect_error(EKC(cor_sing, N = 20), " Correlation matrix is singular, no further analyses are performed\n")
  expect_warning(EKC(cor_nposdef, N = 20), "Matrix was not positive definite, smoothing was done")
})

test_that("settings are returned correctly", {
  expect_named(ekc_cor$settings, c("use", "cor_method", "N", "type"))
  expect_named(ekc_raw$settings, c("use", "cor_method", "N", "type"))

  expect_equal(ekc_cor$settings$N, 500)
  expect_equal(ekc_raw$settings$N, 810)

  expect_equal(ekc_cor$settings$use, "pairwise.complete.obs")
  expect_equal(ekc_raw$settings$use, "pairwise.complete.obs")

  expect_equal(ekc_cor$settings$cor_method, "pearson")
  expect_equal(ekc_raw$settings$cor_method, "pearson")
})

rm(ekc_cor, ekc_raw, x, y, z, dat_sing, cor_sing, cor_nposdef)

#' Print function for CD objects
#'
#' @param x a list of class CD. Output from \link{CD} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print CD
#'
#' @examples
#' \donttest{
#' # determine n factors of the GRiPS
#' CD(GRiPS_raw)
#' }
#'
print.CD <- function(x, plot = TRUE, ...) {

  nfac <- x$n_factors

  cat("\n")

  if(nfac == 1){
    cat("Comparison Data analysis suggests", crayon::bold(nfac), "factor")
    cat("\n")
    cat("\n")
  } else {
    cat("Comparison Data analysis suggests", crayon::bold(nfac), "factors")
    cat("\n")
    cat("\n")
  }

  if (isTRUE(plot)) {
    graphics::plot(x)
  }


}

#' Various Factor Retention Criteria
#'
#' Among the most important decisions for an exploratory factor analysis (EFA) is
#' the choice of the number of factors to retain. Several factor retention
#' criteria have been developed for this. With this function, various factor
#'  retention criteria can be performed simultaneously. Additionally, the data
#'  can be checked for their suitability for factor analysis.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations. If \code{"CD"} is included as a criterion, x must be raw
#'  data.
#' @param criteria character. A vector with the factor retention methods to
#' perform. Possible inputs are: \code{"CD"}, \code{"EKC"}, \code{"HULL"},
#' \code{"KGC"}, \code{"PARALLEL"}, \code{"NEST"}, \code{"SCREE"}, and \code{"SMT"}
#' (see details). By default, a subset of often used, well-performing methods are performed.
#' @param suitability logical. Whether the data should be checked for suitability
#' for factor analysis using the Bartlett's test of sphericity and the
#' Kaiser-Guttmann criterion (see details). Default is \code{TRUE}.
#' @param N  numeric. The number of observations. Only needed if x is a
#' correlation matrix.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#' data is given as input. Default is \code{"pairwise.complete.obs"}.
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}
#' Default is  \code{"pearson"}.
#' @param n_factors_max numeric. Passed to \code{\link{CD}}.The maximum number
#' of factors to test against.
#' Larger numbers will increase the duration the procedure takes, but test more
#' possible solutions. Maximum possible is number of variables / 2. Default is
#' NA. If not specified, number of variables / 2 is used.
#' @param N_pop numeric. Passed to \code{\link{CD}}. Size of finite populations
#' of comparison data. Default is 10000.
#' @param N_samples numeric. Passed to \code{\link{CD}}. Number of samples drawn
#'  from each population. Default is 500.
#' @param alpha numeric. Passed to \code{\link{CD}}. The alpha level used to test
#'  the significance of the improvement added by an additional factor.
#'  Default is .30.
#' @param max_iter_CD numeric. Passed to \code{\link{CD}}. The maximum number of
#'  iterations to perform after which the iterative PAF procedure is halted.
#'   Default is 50.
#' @param n_fac_theor numeric. Passed to \code{\link{HULL}}. Theoretical number
#'  of factors to retain. The maximum of this number and the number of factors
#'  suggested by \link{PARALLEL} plus one will be used in the Hull method.
#' @param method character. Passed to \code{\link{EFA}} in \code{\link{HULL}},
#' \code{\link{KGC}}, \code{\link{SCREE}}, \code{\link{PARALLEL}}, and \code{\link{NEST}}. The
#' estimation method to use. One of  \code{"PAF"}, \code{"ULS"}, or  \code{"ML"},
#' for principal axis factoring, unweighted least squares, and maximum
#' likelihood, respectively.
#' @param gof character. Passed to \code{\link{HULL}}. The goodness of fit index
#' to use. Either \code{"CAF"}, \code{"CFI"}, or \code{"RMSEA"}, or any
#' combination of them. If \code{method = "PAF"} is used, only
#' the CAF can be used as goodness of fit index. For details on the CAF, see
#' Lorenzo-Seva, Timmerman, and Kiers (2011).
#' @param eigen_type_HULL character. Passed to  \code{\link{PARALLEL}} in
#' \code{\link{HULL}}. On what the
#' eigenvalues should be found in the parallel analysis. Can be one of
#' \code{"SMC"}, \code{"PCA"}, or \code{"EFA"}. If using  \code{"SMC"} (default),
#' the diagonal of the correlation matrices is
#' replaced by the squared multiple correlations (SMCs) of the indicators. If
#' using  \code{"PCA"}, the diagonal values of the correlation
#' matrices are left to be 1. If using  \code{"EFA"}, eigenvalues are found on the
#' correlation  matrices with the final communalities of an EFA solution as
#' diagonal.
#' @param eigen_type_other character. Passed to \code{\link{KGC}},
#' \code{\link{SCREE}}, and \code{\link{PARALLEL}}. The same as eigen_type_HULL,
#' but multiple inputs
#' are possible here. Default is to use all inputs, that is, \code{c("PCA",
#' "SMC", "EFA"})
#' @param n_factors numeric. Passed to \code{\link{PARALLEL}} (also within
#' \code{\link{HULL}}), \code{\link{KGC}}, and \code{\link{SCREE}}. Number of
#' factors to extract if \code{"EFA"} is included in \code{eigen_type_HULL} or
#'  \code{eigen_type_other}. Default is 1.
#' @param n_datasets numeric. Passed to \code{\link{PARALLEL}} (also within
#' \code{\link{HULL}}). The number of datasets to simulate. Default is 1000.
#' @param percent numeric. Passed to \code{\link{PARALLEL}} (also within
#' \code{\link{HULL}}). A vector of percentiles to take the simulated eigenvalues
#'  from. Default is 95.
#' @param decision_rule character. Passed to \code{\link{PARALLEL}} (also within
#'  \code{\link{HULL}}). Which rule to use to determine the number of
#'  factors to retain. Default is \code{"means"}, which will use the average
#'  simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
#'  in percent. \code{"crawford"} uses the 95th percentile for the first factor
#'  and the mean afterwards (based on Crawford et al, 2010).
#' @param ekc_type character. Passed to the \code{type} argument of \code{\link{EKC}}.
#'   Either \code{"BvA2017"} for the original implementation by Braeken and van Assen
#'   (2017), or \code{"AM2019"} for the adapted implementation by Auerswald and Moshagen
#'   (2019).
#' @param n_datasets_nest numeric. The number of datasets to simulate in \code{\link{NEST}}. Default is 1000.
#' @param alpha_nest numeric. The alpha level to use in \code{\link{NEST}} (i.e., 1-alpha percentile of eigenvalues is used for reference values).
#' @param show_progress logical. Whether a progress bar should be shown in the
#'   console. Default is FALSE.
#' @param ... Further arguments passed to \code{\link{EFA}} in
#' \code{\link{PARALLEL}} (also within \code{\link{HULL}}) and \code{\link{KGC}}.
#'
#' @details
#' By default, the entered data are checked for suitability for factor analysis
#' using the following methods (see respective documentations for details):
#' \itemize{
#' \item{Bartlett's test of sphericity (see \code{\link{BARTLETT}})}
#' \item{Kaiser-Meyer-Olkin criterion (see \code{\link[EFAtools]{KMO}})}}
#'
#' The available factor retention criteria are the following (see respective
#'  documentations for details):
#'  \itemize{
#' \item{Comparison data (see \code{\link{CD}})}
#' \item{Empirical Kaiser criterion (see \code{\link{EKC}})}
#' \item{Hull method (see \code{\link{HULL}})}
#' \item{Kaiser-Guttman criterion (see \code{\link{KGC}})}
#' \item{Parallel analysis (see \code{\link{PARALLEL}})}
#' \item{Next Eigenvalue Sufficiency Test, NEST (see \code{\link{NEST}})}
#' \item{Scree plot (see \code{\link{SCREE}})}
#' \item{Sequential chi-square model tests, RMSEA lower bound, and AIC
#' (see \code{\link{SMT}})}
#' }
#'
#' @return A list of class N_FACTORS containing
#' \item{outputs}{A list with the outputs from \code{\link{BARTLETT}} and
#'  \code{\link[EFAtools]{KMO}} and the factor retention criteria.}
#' \item{n_factors}{A named vector containing the suggested number of factors
#' from each factor retention criterion.}
#' \item{settings}{A list of the settings used.}
#'
#' @export
#'
#' @examples
#' \donttest{
#' # Default criteria, with correlation matrix and fit method "ML" (where needed)
#' # This will throw a warning for CD, as no raw data were specified
#' nfac_all <- N_FACTORS(test_models$baseline$cormat, N = 500, method = "ML")
#'
#' # The same as above, but without "CD"
#' nfac_wo_CD <- N_FACTORS(test_models$baseline$cormat, criteria = c("EKC",
#'                         "HULL", "PARALLEL", "NEST"), N = 500,
#'                         method = "ML")
#'
#' # Use PAF instead of ML (this will take longer). For this, gof has
#' # to be set to "CAF" for the Hull method.
#' nfac_PAF <- N_FACTORS(test_models$baseline$cormat, criteria = c("EKC",
#'                         "HULL", "PARALLEL", "NEST"), N = 500,
#'                       gof = "CAF")
#'
#' # Do KGC and PARALLEL with only "PCA" type of eigenvalues
#' nfac_PCA <- N_FACTORS(test_models$baseline$cormat, criteria = c("EKC",
#'                       "HULL", "PARALLEL", "NEST"), N = 500,
#'                       method = "ML", eigen_type_other = "PCA")
#'
#' # Use raw data, such that CD can also be performed
#' nfac_raw <- N_FACTORS(GRiPS_raw, method = "ML")
#'}
N_FACTORS <- function(x, criteria = c("CD", "EKC", "HULL", "PARALLEL",
                                      "NEST"),
                      suitability = TRUE, N = NA,
                      use = c("pairwise.complete.obs", "all.obs",
                              "complete.obs", "everything", "na.or.complete"),
                      cor_method = c("pearson", "spearman", "kendall"),
                      n_factors_max = NA, N_pop = 10000, N_samples = 500,
                      alpha = .30, max_iter_CD = 50, n_fac_theor = NA,
                      method = c("ML", "PAF", "ULS"),
                      gof = c("CAF", "CFI", "RMSEA"),
                      eigen_type_HULL = c("SMC", "PCA", "EFA"),
                      eigen_type_other = c("SMC"),
                      n_factors = 1, n_datasets = 1000,
                      percent = 95,
                      decision_rule = c("means", "percentile", "crawford"),
                      ekc_type = c("BvA2017"),
                      n_datasets_nest = 1000, alpha_nest = .05,
                      show_progress = FALSE,
                      ...){

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  ## Perform argument checks and prepare input
  criteria <- match.arg(criteria, several.ok = TRUE,
                        choices = c("CD", "EKC", "HULL", "KGC", "PARALLEL",
                                    "SCREE", "SMT", "NEST"))
  suitability <- checkmate::assert_flag(suitability)
  eigen_type_HULL <- match.arg(eigen_type_HULL)
  eigen_type_other <- match.arg(eigen_type_other, several.ok = TRUE,
                                choices = c("PCA", "SMC", "EFA"))
  gof <- match.arg(gof, several.ok = TRUE, choices = c("CAF", "CFI", "RMSEA"))
  cor_method <- match.arg(cor_method)
  use <- match.arg(use)
  method <- match.arg(method)
  decision_rule <- match.arg(decision_rule)
  ekc_type <- match.arg(ekc_type, c("BvA2017", "AM2019"), several.ok = TRUE)
  checkmate::assert_number(alpha_nest, lower = 0, upper = 1)
  checkmate::assert_count(n_datasets_nest, na.ok = FALSE, positive = TRUE)

  if (isTRUE(show_progress)) {
    criteria <- sort(criteria)
  }

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite
  if(any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)){

    R <- psych::cor.smooth(R)

  }

  # Set all outputs to NA for a start
  bart_out <- NA
  kmo_out <- NA
  cd_out <- NA
  ekc_out <- NA
  hull_out <- NA
  kgc_out <- NA
  parallel_out <- NA
  nest_out <- NA
  scree_out <- NA
  smt_out <- NA

  nfac_CD <- NA
  nfac_EKC_BvA2017 <- NA
  nfac_EKC_AM2019 <- NA
  nfac_HULL_CAF <- NA
  nfac_HULL_CFI <- NA
  nfac_HULL_RMSEA <- NA
  nfac_KGC_PCA <- NA
  nfac_KGC_SMC <- NA
  nfac_KGC_EFA <- NA
  nfac_PA_PCA <- NA
  nfac_PA_SMC <- NA
  nfac_PA_EFA <- NA
  nfac_NEST <- NA
  nfac_SMT_chi <- NA
  nfac_RMSEA <- NA
  nfac_AIC <- NA

  ## Tests for suitability of factor analysis
  if(isTRUE(suitability)){

  # Bartlett's Test of Sphericity
  bart_out <- BARTLETT(R, N = N, use = use, cor_method = cor_method)

  # Kaiser-Meyer_Olkin criterion
  kmo_out <- KMO(R, use = use, cor_method = cor_method)

  }

  ## Factor retention criteria

  # Comparison data
  if("CD" %in% criteria){

    if (!.is_cormat(x)) {

      if (isTRUE(show_progress)) {
        .show_progress(criteria, "CD")
      }

      cd_out <- CD(x, n_factors_max = n_factors_max, N_pop = N_pop,
                   N_samples = N_samples, alpha = alpha, use = use,
                   cor_method = cor_method, max_iter = max_iter_CD)

      nfac_CD <- cd_out$n_factors

    } else {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'x' was a correlation matrix but CD needs raw data. Skipping CD.\n"))
    }

  }

  # Empirical Kaiser Criterion
  if("EKC" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "EKC")
    }

    ekc_out <- EKC(R, N = N, use = use, cor_method = cor_method,
                   type = ekc_type)

    nfac_EKC_BvA2017 <- ekc_out$n_factors_BvA2017
    nfac_EKC_AM2019 <- ekc_out$n_factors_AM2019


  }

  # HULL method
  if("HULL" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "HULL")
    }

    hull_out <- HULL(R, N = N, n_fac_theor = n_fac_theor, method = method,
                     eigen_type = eigen_type_HULL, gof = gof, use = use,
                     cor_method = cor_method, n_datasets = n_datasets,
                     percent = percent, decision_rule = decision_rule,
                     n_factors = n_factors, ...)

    nfac_HULL_CAF <- hull_out$n_fac_CAF
    nfac_HULL_CFI <- hull_out$n_fac_CFI
    nfac_HULL_RMSEA <- hull_out$n_fac_RMSEA

  }

  # Kaiser-Guttman criterion
  if("KGC" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "KGC")
    }

    kgc_out <- KGC(R, eigen_type = eigen_type_other, use = use,
                   cor_method = cor_method, n_factors = n_factors,
                   method = method, ...)

    nfac_KGC_PCA <- kgc_out$n_fac_PCA
    nfac_KGC_SMC <- kgc_out$n_fac_SMC
    nfac_KGC_EFA <- kgc_out$n_fac_EFA

  }

  # Parallel analysis
  if("PARALLEL" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "PARALLEL")
    }

    parallel_out <- try(PARALLEL(R, N = N,
                                 n_datasets = n_datasets, percent = percent,
                                 eigen_type = eigen_type_other, use = use,
                                 cor_method = cor_method,
                                 decision_rule = decision_rule,
                                 n_factors = n_factors, method = method, ...))

    nfac_PA_PCA <- parallel_out$n_fac_PCA
    nfac_PA_SMC <- parallel_out$n_fac_SMC
    nfac_PA_EFA <- parallel_out$n_fac_EFA

  }

  # Next Eigenvalue Sufficiency Test
  if("NEST" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "NEST")
    }

    nest_out <- NEST(R, N = N, use = use, cor_method = cor_method,
                     alpha = alpha_nest, n_datasets = n_datasets_nest,
                     method = method)

    nfac_NEST <- nest_out$n_factors


  }

  # Scree plot
  if("SCREE" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "SCREE")
    }

    scree_out <- SCREE(R, eigen_type = eigen_type_other, use = use,
                     cor_method = cor_method, n_factors = n_factors,
                     method = method, ...)
  }

  # Sequential chi square tests, RMSEA lower bound and AIC
  if("SMT" %in% criteria){

    if (isTRUE(show_progress)) {
      .show_progress(criteria, "SMT")
    }

    smt_out <- SMT(R, N = N, use = use, cor_method = cor_method)

    nfac_SMT_chi <- smt_out$nfac_chi
    nfac_RMSEA <- smt_out$nfac_RMSEA
    nfac_AIC <- smt_out$nfac_AIC

  }

  # Prepare settings here
  settings <- list(criteria = criteria,
                   suitability = suitability,
                   N = N,
                   use = use,
                   n_factors_max = n_factors_max,
                   N_pop = N_pop,
                   N_samples = N_samples,
                   alpha = alpha,
                   cor_method = cor_method,
                   max_iter_CD = max_iter_CD,
                   n_fac_theor = n_fac_theor,
                   method = method,
                   gof = gof,
                   eigen_type_HULL = eigen_type_HULL,
                   eigen_type_other = eigen_type_other,
                   n_factors = n_factors,
                   n_datasets = n_datasets,
                   percent = percent,
                   decision_rule = decision_rule,
                   ekc_type = ekc_type,
                   n_datasets_nest = n_datasets_nest,
                   alpha_nest = alpha_nest)

  # Prepare the output
  n_factors <- c(nfac_CD = nfac_CD,
                 nfac_EKC_BvA2017 = nfac_EKC_BvA2017,
                 nfac_EKC_AM2019 = nfac_EKC_AM2019,
                 nfac_HULL_CAF = nfac_HULL_CAF,
                 nfac_HULL_CFI = nfac_HULL_CFI,
                 nfac_HULL_RMSEA = nfac_HULL_RMSEA,
                 nfac_KGC_PCA = nfac_KGC_PCA,
                 nfac_KGC_SMC = nfac_KGC_SMC,
                 nfac_KGC_EFA = nfac_KGC_EFA,
                 nfac_PA_PCA = nfac_PA_PCA,
                 nfac_PA_SMC = nfac_PA_SMC,
                 nfac_PA_EFA = nfac_PA_EFA,
                 nfac_NEST = nfac_NEST,
                 nfac_SMT_chi = nfac_SMT_chi,
                 nfac_RMSEA = nfac_RMSEA,
                 nfac_AIC = nfac_AIC)

  outputs <- list(bart_out = bart_out,
                  kmo_out = kmo_out,
                  cd_out = cd_out,
                  ekc_out = ekc_out,
                  hull_out = hull_out,
                  kgc_out = kgc_out,
                  parallel_out = parallel_out,
                  nest_out = nest_out,
                  scree_out = scree_out,
                  smt_out = smt_out)

  output <- list(outputs = outputs,
                 n_factors = n_factors,
                 settings = settings)

  class(output) <- "N_FACTORS"

  if (isTRUE(show_progress)) {
    .show_progress(criteria, "done", TRUE)
  }

  return(output)

}

nf_grips <- suppressMessages(suppressWarnings(N_FACTORS(GRiPS_raw)))

test_that("output class and dimensions are correct", {
  expect_is(nf_grips, "N_FACTORS")
  expect_is(nf_grips$outputs, "list")
  expect_is(nf_grips$settings, "list")
  expect_is(nf_grips$n_factors, "numeric")

  expect_named(nf_grips, c("outputs", "n_factors", "settings"))
  expect_named(nf_grips$outputs, c("bart_out", "kmo_out", "cd_out", "ekc_out",
                                   "hull_out", "kgc_out", "parallel_out",
                                  "nest_out", "scree_out", "smt_out"))
  expect_named(nf_grips$n_factors, c("nfac_CD",
                                     "nfac_EKC_BvA2017", "nfac_EKC_AM2019",
                                     "nfac_HULL_CAF",
                                     "nfac_HULL_CFI", "nfac_HULL_RMSEA",
                                     "nfac_KGC_PCA", "nfac_KGC_SMC",
                                     "nfac_KGC_EFA",
                                     "nfac_PA_PCA", "nfac_PA_SMC", "nfac_PA_EFA",
                                     "nfac_NEST",
                                     "nfac_SMT_chi", "nfac_RMSEA", "nfac_AIC"))
  expect_named(nf_grips$settings, c("criteria", "suitability", "N", "use",
                                    "n_factors_max", "N_pop", "N_samples", "alpha",
                                    "cor_method", "max_iter_CD", "n_fac_theor",
                                    "method", "gof", "eigen_type_HULL",
                                    "eigen_type_other", "n_factors", "n_datasets",
                                    "percent", "decision_rule",
                                    "ekc_type", "n_datasets_nest",
                                    "alpha_nest"))
})

x <- rnorm(100)
y <- rnorm(100)
z <- x + y



burt <- matrix(c(1.00,  0.83,  0.81,  0.80,   0.71, 0.70, 0.54, 0.53,  0.59,  0.24, 0.13,
                 0.83,  1.00,  0.87,  0.62,   0.59, 0.44, 0.58, 0.44,  0.23,  0.45,  0.21,
                 0.81,  0.87,  1.00,  0.63,   0.37, 0.31, 0.30, 0.12,  0.33,  0.33,  0.36,
                 0.80,  0.62,  0.63,  1.00,   0.49, 0.54, 0.30, 0.28,  0.42,  0.29, -0.06,
                 0.71,  0.59,  0.37,  0.49,   1.00, 0.54, 0.34, 0.55,  0.40,  0.19, -0.10,
                 0.70,  0.44,  0.31,  0.54,   0.54, 1.00, 0.50, 0.51,  0.31,  0.11,  0.10,
                 0.54,  0.58,  0.30,  0.30,   0.34, 0.50, 1.00, 0.38,  0.29,  0.21,  0.08,
                 0.53,  0.44,  0.12,  0.28,   0.55, 0.51, 0.38, 1.00,  0.53,  0.10, -0.16,
                 0.59,  0.23,  0.33,  0.42,   0.40, 0.31, 0.29, 0.53,  1.00, -0.09, -0.10,
                 0.24,  0.45,  0.33,  0.29,   0.19, 0.11, 0.21, 0.10, -0.09,  1.00,  0.41,
                 0.13,  0.21,  0.36, -0.06,  -0.10, 0.10, 0.08, -0.16, -0.10, 0.41,  1.00),
               nrow = 11, ncol = 11)

test_that("errors etc. are thrown correctly", {
  expect_error(N_FACTORS(1:10), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_warning(N_FACTORS(GRiPS_raw, N = 10), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(N_FACTORS(cbind(x, y, z, z + 1, y + 1, x + 1)), " Correlation matrix is singular, no further analyses are performed\n")
  expect_warning(N_FACTORS(test_models$baseline$cormat, N = 500), " 'x' was a correlation matrix but CD needs raw data. Skipping CD.\n")
  expect_warning(N_FACTORS(burt, N = 170, criteria = c("PARALLEL", "EKC")), "Matrix was not positive definite, smoothing was done")
})

rm(nf_grips, x, y, z, burt)

#' Plot CD object
#'
#' Plot method showing a summarized output of the \link{CD} function
#'
#' @param x a list of class CD. An output from the \link{CD} function.
#' @param ... not used.
#'
#' @export
#' @method plot CD
plot.CD <- function(x, ...) {

  x_len <- x$n_factors

  ys <- colMeans(x$RMSE_eigenvalues)[seq_len(x$n_factors)]

  graphics::plot.new()
  graphics::plot.window(xlim = c(1, x_len),
                        ylim = c(0, max(pretty(ys))))
  graphics::axis(1, seq_len(x_len))
  graphics::axis(2, pretty(c(0, ys)),
                 las = 1)

  graphics::mtext("N Factors", side = 1, line = 3, cex = 1.5, padj =-.5)
  graphics::mtext("RMSE eigenvalues", side = 2, line = 3, cex = 1.5, padj =.25)

  graphics::lines(seq_len(x_len), ys)
  graphics::points(seq_len(x_len), ys, pch = 16)

  if (!is.na(x$n_factors)) {
    graphics::points(x$n_factors, ys[x$n_factors],
                     pch = 1, cex = 2, col = "red")
    graphics::text(x$n_factors, ys[x$n_factors],
                   x$n_factors, pos = 3, cex = 1.5, col = "red",
                   font = 1, offset = .75)
  }

  factors_text <- paste0("N factors suggested by comparison data analysis: ",
                         x$n_factors)

  graphics::title(factors_text, cex.main = 1.3)

}

#' Print COMPARE object
#'
#' Print Method showing a summarized output of the \code{\link{COMPARE}} function.
#'
#' @param x  list. An object of class COMPARE to be printed
#' @param ... Further arguments for print.
#'
#' @export
#' @method print COMPARE
#' @examples
#' # A type SPSS EFA to mimick the SPSS implementation
#' EFA_SPSS_5 <- EFA(IDS2_R, n_factors = 5, type = "SPSS")
#'
#' # A type psych EFA to mimick the psych::fa() implementation
#' EFA_psych_5 <- EFA(IDS2_R, n_factors = 5, type = "psych")
#'
#' # compare the two
#' COMPARE(EFA_SPSS_5$unrot_loadings, EFA_psych_5$unrot_loadings,
#'         x_labels = c("SPSS", "psych"))
print.COMPARE <- function(x, ...) {

  # extract summary statistics
  diff <- x$diff
  mean_abs_diff <- x$mean_abs_diff
  median_abs_diff <- x$median_abs_diff
  min_abs_diff <- x$min_abs_diff
  max_abs_diff <- x$max_abs_diff
  max_dec <- x$max_dec
  are_equal <- x$are_equal

  # extract control settings
  digits <- x$settings$digits
  m_red <- x$settings$m_red
  range_red <- x$settings$range_red
  round_red <- x$settings$round_red
  print_diff <- x$settings$print_diff
  x_labels <- x$settings$x_labels
  plot <- x$settings$plot
  plot_red <- x$settings$plot_red

  # prepare to print statistics

  if (mean_abs_diff <= m_red) {
    mean_out <- crayon::green$bold(.numformat(mean_abs_diff, digits,
                                              TRUE))
  } else {
    mean_out <- crayon::red$bold(.numformat(mean_abs_diff, digits, TRUE))
  }

  if (median_abs_diff <= m_red) {
    median_out <- crayon::green$bold(.numformat(median_abs_diff, digits,
                                                TRUE))
  } else {
    median_out <- crayon::red$bold(.numformat(median_abs_diff, digits,
                                              TRUE))
  }

  if (max_abs_diff <= range_red) {
    max_out <- crayon::green$bold(.numformat(max_abs_diff, digits, TRUE))
    min_out <- crayon::green$bold(.numformat(min_abs_diff, digits, TRUE))
  } else {
    max_out <- crayon::red$bold(.numformat(max_abs_diff, digits, TRUE))
    min_out <- crayon::red$bold(.numformat(min_abs_diff, digits, TRUE))
  }

  if (is.na(are_equal)) {
    equal_out <- crayon::red$bold("none")
  } else if (are_equal < round_red) {
    equal_out <- crayon::red$bold(are_equal)
  } else {
    equal_out <- crayon::green$bold(are_equal)
  }



  cat("Mean [min, max] absolute difference: ")
  cat(paste0(mean_out, " [", min_out, ", ", max_out, "]"))
  cat("\n")
  cat(paste0("Median absolute difference: ", median_out))
  cat("\n")
  cat(paste0("Max decimals where all numbers are equal: ",
             equal_out))
  cat("\n")
  cat(paste0("Minimum number of decimals provided: ", crayon::blue$bold(max_dec)))

  # create the difference object
  if (isTRUE(print_diff)) {
    cat("\n")
    cat("\n")
    if (inherits(diff, "matrix")) {

      out_diff <- .get_compare_matrix(diff, digits = digits, r_red = range_red)

    } else {

      out_diff <- .get_compare_vector(diff, digits = digits, r_red = range_red)

    }

    cat(out_diff)

  }


  if (isTRUE(plot) && length(c(diff)) > 2) {

    if (length(x_labels) < 2) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" Less than two x_labels specified, using 'x' and 'y'."))
      x_labels <- c("x", "y")
    } else if (length(x_labels) > 2) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" More than two x_labels specified, using only the first two."))
      x_labels <- x_labels[1:2]
    }

    # prepare variable for plot
    diff_dat <- tibble::tibble(diffs = as.vector(abs(diff))) %>%
      dplyr::mutate(color = dplyr::case_when(diffs >= plot_red ~ "large difference",
                                             TRUE ~ "acceptable difference"),
                    comp = paste(x_labels, collapse = " vs. "))

    diff_plot <- ggplot2::ggplot(diff_dat, ggplot2::aes_string("comp", "diffs",
                                                               col = "color")) +
      ggplot2::geom_violin(col = "grey20", width = .7, size = .7) +
      ggplot2::geom_hline(yintercept = plot_red, lty = 2, alpha = .5,
                          size = 1.25) +
      ggplot2::geom_jitter(alpha = .5, width = 0.05, height = 0, size = 2) +
      ggplot2::scale_color_manual(values = c("black", "red")) +
      ggplot2::theme_bw() +
      ggplot2::labs(
        subtitle = paste("Threshold for difference coloring:", plot_red),
        x = "Compared Variables",
        y = "Absolute Difference"
      ) +
      ggplot2::theme(
        legend.position = "none",
        strip.text = ggplot2::element_text(size = 11, face = "bold"),
        axis.text = ggplot2::element_text(size = 11),
        axis.title = ggplot2::element_text(size = 13,face = "bold")
      )

    print(diff_plot)

  }


}

#' Print function for NEST objects
#'
#' @param x a list of class NEST Output from \code{\link{NEST}} function.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print NEST
#'
#' @examples
#' NEST(test_models$baseline$cormat, N = 500)
#'
print.NEST <- function(x, ...) {

  nfac <- x$n_factors

  cat("\n")

  if(nfac == 1){
    cat("NEST suggests", crayon::bold(nfac), "factor")
    cat("\n")
    cat("\n")
  } else {
    cat("NEST suggests", crayon::bold(nfac), "factors")
    cat("\n")
    cat("\n")
  }

}

#' Bartlett's test of sphericity
#'
#' This function tests whether a correlation matrix is significantly different
#' from an identity matrix (Bartlett, 1951). If the Bartlett's test is not
#' significant, the correlation matrix is not suitable for factor analysis
#' because the variables show too little covariance.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw data
#' is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#'
#' @details Bartlett (1951) proposed this statistic to determine a correlation
#' matrix' suitability for factor analysis. The statistic is approximately
#' chi square distributed with \eqn{df = \frac{p(p - 1)}{2}} and is given by
#'
#' \deqn{chi^2 = -log(det(R)) (N - 1 - (2 * p + 5)/6)}
#'
#' where \eqn{det(R)} is the determinant of the correlation matrix, \eqn{N} is
#' the sample size, and \eqn{p} is the number of variables.
#'
#' This tests requires multivariate normality. If this condition is not met,
#' the Kaiser-Meyer-Olkin criterion (\code{\link[EFAtools]{KMO}})
#' can still be used.
#'
#' This function was heavily influenced by the \code{\link[psych:cortest.bartlett]{psych::cortest.bartlett}} function from the psych package.
#'
#' The \code{BARTLETT} function can also be called together with the
#'  (\code{\link[EFAtools]{KMO}}) function and with factor retention criteria
#'  in the \code{\link{N_FACTORS}} function.
#'
#' @return A list containing
#' \item{chisq}{The chi square statistic.}
#' \item{p_value}{The p value of the chi square statistic.}
#' \item{df}{The degrees of freedom for the chi square statistic.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Bartlett, M. S. (1951). The effect of standardization on a Chi-square
#' approximation in factor analysis. Biometrika, 38, 337-344.
#'
#' @seealso \code{\link[EFAtools]{KMO}} for another measure to determine
#'  suitability for factor analysis.
#'
#'  \code{\link{N_FACTORS}} as a wrapper function for this function,
#'  \code{\link[EFAtools]{KMO}} and several factor retention criteria.
#'
#' @export
#'
#' @examples
#' BARTLETT(test_models$baseline$cormat, N = 500)
#'
BARTLETT <- function(x, N = NA, use = c("pairwise.complete.obs", "all.obs",
                                        "complete.obs", "everything",
                                        "na.or.complete"),
                     cor_method = c("pearson", "spearman", "kendall")){

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  checkmate::assert_count(N, na.ok = TRUE)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

    if (is.na(N)) {

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Argument 'N' was NA, Bartlett's test could not be executed. Please provide either N or raw data.\n"))

    }

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, Bartlett's test cannot be executed.\n"))
  }

  # Check if correlation matrix is positive definite
  if(any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)){

    R <- psych::cor.smooth(R)

  }

  # Calculate test statistic
  p <- nrow(R)
  detR <- det(R)
  statistic <- -log(detR) * (N - 1 - (2 * p + 5)/6)
  df <- p * (p - 1)/2
  pval <- stats::pchisq(statistic, df, lower.tail = FALSE)

  # prepare the output
  settings <- list(N = N,
                   use = use,
                   cor_method = cor_method)

  output <- list(chisq = statistic,
                 p_value = pval,
                 df = df,
                 settings = settings)

  class(output) <- "BARTLETT"

  return(output)

}

bart_cor <- BARTLETT(test_models$baseline$cormat, N = 500)
bart_raw <- BARTLETT(GRiPS_raw)

set.seed(500)
bart_rand <- BARTLETT(matrix(rnorm(100), ncol = 4))

test_that("output class and dimensions are correct", {
  expect_is(bart_cor, "BARTLETT")
  expect_output(str(bart_cor), "List of 4")
  expect_is(bart_raw, "BARTLETT")
  expect_output(str(bart_raw), "List of 4")
})


test_that("p-values and df are correct", {
  expect_lt(bart_cor$p_value, 0.0001)
  expect_lt(bart_raw$p_value, 0.0001)
  expect_gt(bart_rand$p_value, 0.05)

  expect_equal(bart_cor$df, 153)
  expect_equal(bart_raw$df, 28)
  expect_equal(bart_rand$df, 6)
})

test_that("settings are returned correctly", {
  expect_named(bart_cor$settings, c("N", "use", "cor_method"))
  expect_named(bart_raw$settings, c("N", "use", "cor_method"))
  expect_named(bart_rand$settings, c("N", "use", "cor_method"))

  expect_equal(bart_cor$settings$N, 500)
  expect_equal(bart_raw$settings$N, 810)
  expect_equal(bart_rand$settings$N, 25)

  expect_equal(bart_cor$settings$use, "pairwise.complete.obs")
  expect_equal(bart_raw$settings$use, "pairwise.complete.obs")
  expect_equal(bart_rand$settings$use, "pairwise.complete.obs")

  expect_equal(bart_cor$settings$cor_method, "pearson")
  expect_equal(bart_raw$settings$cor_method, "pearson")
  expect_equal(bart_rand$settings$cor_method, "pearson")
})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y

dat_sing <- matrix(c(x, y, z), ncol = 3)
cor_sing <- stats::cor(dat_sing)

cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

test_that("errors are thrown correctly", {
  expect_error(BARTLETT(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_error(BARTLETT(test_models$baseline$cormat), " Argument 'N' was NA, Bartlett's test could not be executed. Please provide either N or raw data.\n")
  expect_message(BARTLETT(GRiPS_raw), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_warning(BARTLETT(GRiPS_raw, N = 20), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(BARTLETT(dat_sing), " Correlation matrix is singular, Bartlett's test cannot be executed.\n")
  expect_error(BARTLETT(cor_sing, N = 10), " Correlation matrix is singular, Bartlett's test cannot be executed.\n")
  expect_warning(BARTLETT(cor_nposdef, N = 10), "Matrix was not positive definite, smoothing was done")
})

rm(bart_cor, bart_raw, bart_rand, x, y, z, dat_sing, cor_sing, cor_nposdef)

#' Scree Plot
#'
#' The scree plot was originally introduced by Cattell (1966) to perform the
#' scree test. In a scree plot, the eigenvalues of the factors / components are
#' plotted against the index of the factors / components, ordered from 1 to N
#' factors components, hence from largest to smallest eigenvalue. According to
#' the scree test, the number of factors / components to retain is the number of
#' factors / components to the left of the "elbow" (where the curve starts to
#' level off) in the scree plot.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param eigen_type character. On what the eigenvalues should be found. Can be
#'  either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
#'  the diagonal values of the correlation matrices are left to be 1. If using
#'  "SMC", the diagonal of the
#'  correlation matrices is replaced by the squared multiple correlations (SMCs)
#'  of the indicators. If using "EFA", eigenvalues are found on the correlation
#'  matrices with the final communalities of an exploratory factor analysis
#'  solution (default is principal axis factoring extracting 1 factor) as
#'  diagonal.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#'  data is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#' @param n_factors numeric. Number of factors to extract if "EFA" is included in
#' \code{eigen_type}. Default is 1.
#' @param ... Additional arguments passed to \code{\link{EFA}}. For example,
#' to change the extraction method (PAF is default).
#'
#' @details As the scree test requires visual examination, the test has been
#' especially criticized for its subjectivity and with this low inter-rater
#' reliability. Moreover, a scree plot can be ambiguous if there are either no
#' clear "elbow" or multiple "elbows", making it difficult to judge just where
#' the eigenvalues do level off. Finally, the scree test has also been found to
#' be less accurate than other factor retention criteria. For all these reasons,
#' the scree test has been recommended against, at least for exclusive use as a
#' factor retention criterion (Zwick & Velicer, 1986)
#'
#' The \code{SCREE} function can also be called together with other factor
#' retention criteria in the \code{\link{N_FACTORS}} function.
#'
#' @return A list of class SCREE containing
#'
#' \item{eigen_PCA}{ A vector containing the eigenvalues found with PCA.}
#' \item{eigen_SMC}{ A vector containing the eigenvalues found with SMCs.}
#' \item{eigen_EFA}{ A vector containing the eigenvalues found with EFA.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Cattell, R. B. (1966). The scree test for the number of factors.
#' Multivariate Behavioral Research, 1(2), 245–276.
#' https://doi.org/10.1207/s15327906mbr0102_10
#'
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. Psychological Bulletin, 99,
#' 432–442. http://dx.doi.org/10.1037/0033-2909.99.3.432
#'
#' @seealso Other factor retention criteria: \code{\link{CD}}, \code{\link{EKC}},
#' \code{\link{HULL}}, \code{\link{PARALLEL}}, \code{\link{SMT}}
#'
#' \code{\link{N_FACTORS}} as a wrapper function for this and all the
#' above-mentioned factor retention criteria.
#'
#' @export
#'
#' @examples
#' SCREE(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
SCREE <- function(x, eigen_type = c("PCA", "SMC", "EFA"),
                  use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                        "everything", "na.or.complete"),
                  cor_method = c("pearson", "spearman", "kendall"),
                  n_factors = 1, ...){

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  eigen_type <- match.arg(eigen_type, several.ok = TRUE)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  checkmate::assert_count(n_factors)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed.\n"))

  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }


  # Store original R for later
  R_orig <- R

  # Prepare objects
  eigen_R_PCA <- NA
  eigen_R_SMC <- NA
  eigen_R_EFA <- NA

  if("PCA" %in% eigen_type) {

    # Calculate eigenvalues and determine number of factors
    eigen_R_PCA <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  }

  if("SMC" %in% eigen_type) {

    # Calculate SMCs and replace diagonal of correlation matrix with these
    inverse_R <- solve(R)
    SMCs <- 1 - 1 / diag(inverse_R)
    diag(R) <- SMCs

    # Calculate eigenvalues and determine number of factors
    eigen_R_SMC <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  }

  if("EFA" %in% eigen_type) {

    R <- R_orig

    # Do an EFA to get final communality estimates and replace diagonal of
    # correlation matrix with these
    EFA_h2 <- suppressMessages(suppressWarnings(EFA(R, n_factors = n_factors, ...)$h2))
    diag(R) <- EFA_h2

    # Calculate eigenvalues and determine number of factors
    eigen_R_EFA <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  }

  # prepare settings
  settings <- list(eigen_type = eigen_type,
                   use = use,
                   cor_method = cor_method,
                   n_factors = n_factors)

  # Prepare the output
  output <- list(
    eigen_PCA = eigen_R_PCA,
    eigen_SMC = eigen_R_SMC,
    eigen_EFA = eigen_R_EFA,
    settings = settings
  )

  class(output) <- "SCREE"

  return(output)

}

#' Print function for HULL objects
#'
#' @param x a list of class HULL. Output from the \code{\link{HULL}} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print HULL
#'
#' @examples
#' \donttest{
#' HULL(test_models$baseline$cormat, N = 500, method = "ML")
#' }
print.HULL <- function(x, plot = TRUE, ...) {

  method <- x$settings$method
  gof <- x$settings$gof

  cat("Hull Analysis performed testing ", crayon::bold("0"), " to ",
      crayon::bold(x$n_fac_max), " factors.", sep = "")

  if(length(gof) == 1){
  cat("\n")
  cat(crayon::bold(method), " estimation and the ",
      .settings_string(gof), " fit index was used.", sep = "")
  cat("\n")
  } else {
    cat("\n")
    cat(crayon::bold(method), " estimation and the ",
        .settings_string(gof), " fit indices were used.", sep = "")
    cat("\n")
  }

  cat("\n")
  cat(cli::rule(left = "Number of factors suggested by the Hull method",
                col = "blue"))
  cat("\n")

  if("CAF" %in% gof){
    cat("\n")
    cat(crayon::blue(cli::symbol$bullet, "With", crayon::bold("CAF:   ")),
        crayon::bold(x$n_fac_CAF))
    cat("\n")
  }

  if("CFI" %in% gof){
    cat(crayon::blue(cli::symbol$bullet, "With", crayon::bold("CFI:   ")),
        crayon::bold(x$n_fac_CFI))
    cat("\n")
  }

  if("RMSEA" %in% gof){
    cat(crayon::blue(cli::symbol$bullet, "With", crayon::bold("RMSEA: ")),
        crayon::bold(x$n_fac_RMSEA))
    cat("\n")
  }

  if (isTRUE(plot)) {
    graphics::plot(x)
  }

}

#' Parallel analysis
#'
#' Various methods for performing parallel analysis. This function uses
#' \link[future.apply]{future_lapply} for which a parallel processing plan can
#' be selected. To do so, call \code{library(future)} and, for example,
#'  \code{plan(multisession)}; see examples.
#'
#' @param x matrix or data.frame. The real data to compare the simulated eigenvalues
#'  against. Must not contain variables of classes other than numeric. Can be a
#'  correlation matrix or raw data.
#' @param N numeric. The number of cases / observations to simulate. Only has to
#'  be specified if \code{x} is either a correlation matrix or \code{NULL}. If
#'  x contains raw data, \code{N} is found from the dimensions of \code{x}.
#' @param n_vars numeric. The number of variables / indicators to simulate.
#' Only has to be specified if \code{x} is left as \code{NULL} as otherwise the
#' dimensions are taken from \code{x}.
#' @param n_datasets numeric. The number of datasets to simulate. Default is 1000.
#' @param percent numeric. The percentile to take from the simulated eigenvalues.
#'  Default is 95.
#' @param eigen_type character. On what the eigenvalues should be found. Can be
#'  either "SMC", "PCA", or "EFA". If using "SMC", the diagonal of the correlation
#'  matrix is replaced by the squared multiple correlations (SMCs) of the
#'  indicators. If using "PCA", the diagonal values of the correlation matrices
#'  are left to be 1. If using "EFA", eigenvalues are found on the correlation
#'  matrices with the final communalities of an EFA solution as diagonal.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw data
#' is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}
#' Default is "pearson".
#' @param decision_rule character. Which rule to use to determine the number of
#'  factors to retain. Default is \code{"means"}, which will use the average
#'  simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
#'  in percent. \code{"crawford"} uses the 95th percentile for the first factor
#'  and the mean afterwards (based on Crawford et al, 2010).
#' @param n_factors numeric. Number of factors to extract if "EFA" is included in
#' \code{eigen_type}. Default is 1.
#' @param ... Additional arguments passed to \code{\link{EFA}}. For example,
#' the extraction method can be changed here (default is "PAF"). PAF is more
#' robust, but it will take longer compared to the other estimation methods
#' available ("ML" and "ULS").
#'
#' @details Parallel analysis (Horn, 1965) compares the eigenvalues obtained from
#' the sample
#'  correlation matrix against those of null model correlation matrices (i.e.,
#'  with uncorrelated variables) of the same sample size. This way, it accounts
#'  for the variation in eigenvalues introduced by sampling error and thus
#'  eliminates the main problem inherent in the Kaiser-Guttman criterion
#'  (\code{\link{KGC}}).
#'
#'  Three different ways of finding the eigenvalues under the factor model are
#'  implemented, namely "SMC", "PCA", and "EFA". PCA leaves the diagonal elements
#'  of the correlation matrix as they are and is thus equivalent to what is done
#'  in PCA. SMC uses squared multiple correlations as communality estimates with
#'  which the diagonal of the correlation matrix is replaced. Finally, EFA performs
#'  an \code{\link{EFA}} with one factor (can be adapted to more factors) to estimate
#'  the communalities and based on the correlation matrix with these as diagonal
#'  elements, finds the eigenvalues.
#'
#'  Parallel analysis is often argued to be one of the most accurate factor
#'  retention criteria. However, for highly correlated
#'  factor structures it has been shown to underestimate the correct number of
#'  factors. The reason for this is that a null model (uncorrelated variables)
#'  is used as reference. However, when factors are highly correlated, the first
#'  eigenvalue will be much larger compared to the following ones, as
#'  later eigenvalues are conditional on the earlier ones in the sequence and thus
#'  the shared variance is already accounted in the first eigenvalue (e.g.,
#'  Braeken & van Assen, 2017).
#'
#'  The \code{PARALLEL} function can also be called together with other factor
#'  retention criteria in the \code{\link{N_FACTORS}} function.
#'
#' @return A list of class PARALLEL containing the following objects
#' \item{eigenvalues_PCA}{A matrix containing the eigenvalues of the real and the simulated data found with eigen_type = "PCA"}
#' \item{eigenvalues_SMC}{A matrix containing the eigenvalues of the real and the simulated data found with eigen_type = "SMC"}
#' \item{eigenvalues_EFA}{A matrix containing the eigenvalues of the real and the simulated data found with eigen_type = "EFA"}
#' \item{n_fac_PCA}{The number of factors to retain according to the parallel procedure with eigen_type = "PCA".}
#' \item{n_fac_SMC}{The number of factors to retain according to the parallel procedure with eigen_type = "SMC".}
#' \item{n_fac_EFA}{The number of factors to retain according to the parallel procedure with eigen_type = "EFA".}
#' \item{settings}{A list of control settings used in the print function.}
#'
#' @source Braeken, J., & van Assen, M. A. (2017). An empirical Kaiser criterion.
#' Psychological Methods, 22, 450 – 466. http://dx.doi.org/10.1037/ met0000074
#'
#' @source Crawford, A. V., Green, S. B., Levy, R., Lo, W. J., Scott, L.,
#' Svetina, D., & Thompson, M. S. (2010). Evaluation of parallel analysis methods
#' for determining the number of factors. Educational and Psychological
#' Measurement, 70(6), 885-901.
#'
#' @source Horn, J. L. (1965). A rationale and test for the number of factors in
#' factor analysis. Psychometrika, 30(2), 179–185. doi: 10.1007/BF02289447
#'
#' @seealso Other factor retention criteria: \code{\link{CD}}, \code{\link{EKC}},
#' \code{\link{HULL}}, \code{\link{KGC}}, \code{\link{SMT}}
#'
#' \code{\link{N_FACTORS}} as a wrapper function for this and all the
#' above-mentioned factor retention criteria.
#'
#' @export
#'
#' @examples
#' \donttest{
#' # example without real data
#' pa_unreal <- PARALLEL(N = 500, n_vars = 10)
#'
#' # example with correlation matrix with all eigen_types and PAF estimation
#' pa_paf <- PARALLEL(test_models$case_11b$cormat, N = 500)
#'
#' # example with correlation matrix with all eigen_types and ML estimation
#' # this will be faster than the above with PAF)
#' pa_ml <- PARALLEL(test_models$case_11b$cormat, N = 500, method = "ML")
#'}
#'
#'\dontrun{
#' # for parallel computation
#' future::plan(future::multisession)
#' pa_faster <- PARALLEL(test_models$case_11b$cormat, N = 500)
#' }

PARALLEL <- function(x = NULL,
                     N = NA,
                     n_vars = NA,
                     n_datasets = 1000,
                     percent = 95,
                     eigen_type = c("PCA", "SMC", "EFA"),
                     use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                             "everything", "na.or.complete"),
                     cor_method = c("pearson", "spearman", "kendall"),
                     decision_rule = c("means", "percentile", "crawford"),
                     n_factors = 1,
                     ...) {


  if(!is.null(x) && !inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither NULL, nor a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data or leave x at NULL.\n"))

  }
  eigen_type <- match.arg(eigen_type, several.ok = TRUE)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  decision_rule <- match.arg(decision_rule)
  checkmate::assert_count(n_factors)
  checkmate::assert_count(N, na.ok = TRUE)
  checkmate::assert_count(n_vars, na.ok = TRUE)
  checkmate::assert_count(n_datasets)
  checkmate::assert_number(percent, lower = 0, upper = 100)

  n_cores <- future::nbrOfWorkers()
  size_vec <- rep(round(n_datasets / n_cores), n_cores - 1)
  size_vec[n_cores] <- n_datasets - sum(size_vec)

  eigenvalues_PCA <- NA
  eigenvalues_SMC <- NA
  eigenvalues_EFA <- NA
  n_fac_PCA <- NA
  n_fac_SMC <- NA
  n_fac_EFA <- NA
  x_dat <- FALSE

  if (!is.null(x)){

      if (!is.na(n_vars)) {
        warning(crayon::yellow$bold("!"), crayon::yellow(" n_vars was set and data entered. Taking n_vars from data\n"))
      }
      n_vars <- ncol(x)
      x_dat <- TRUE

      # Check if it is a correlation matrix
      if(.is_cormat(x)){

        R <- x

      } else {

        message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

        if (!is.na(N)) {
          warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
        }

        R <- stats::cor(x, use = use, method = cor_method)
        colnames(R) <- colnames(x)
        N <- nrow(x)

      }


      # Check if correlation matrix is invertable, if it is not, stop with message
      R_i <- try(solve(R), silent = TRUE)

      if (inherits(R_i, "try-error")) {
        stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, parallel analysis is not possible.\n"))
      }

      # Check if correlation matrix is positive definite
      if(any(eigen(R)$values <= .Machine$double.eps^.6)){

        R <- psych::cor.smooth(R)

      }

      if ("PCA" %in% eigen_type) {
        eigvals_real_PCA <- matrix(eigen(R, symmetric = TRUE,
                                     only.values = TRUE)$values, ncol = 1)
        colnames(eigvals_real_PCA) <- "Real Eigenvalues"
      }

      if ("SMC" %in% eigen_type) {
        # compute smcs
        R_SMC <- R
        diag(R_SMC) <- 1 - (1 / diag(solve(R_SMC)))
        eigvals_real_SMC <- matrix(eigen(R_SMC, symmetric = TRUE,
                                     only.values = TRUE)$values, ncol = 1)
        colnames(eigvals_real_SMC) <- "Real Eigenvalues"
      }

      if ("EFA" %in% eigen_type) {
        eigvals_real_EFA <- matrix(EFA(R, n_factors = n_factors, N = N,
                                   ...)$final_eigen,  ncol = 1)
        colnames(eigvals_real_EFA) <- "Real Eigenvalues"
      }

  }

  if (is.na(n_vars)) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' "n_vars" was not set and could not be taken from data. Please specify n_vars and try again.\n'))
  }

  if (is.na(N)) {

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' "N" was not set and could not be taken from data. Please specify N and try again.\n'))

  }

  if (N <= n_vars) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' "N" was smaller than or equal to the number of variables but must be larger.\n'))
  }

    if ("PCA" %in% eigen_type) {

      eigvals_PCA <- try(future.apply::future_lapply(size_vec, .parallel_sim, N = N,
                                             n_vars = n_vars, eigen_type = 1,
                                             future.seed = TRUE),
                         silent = TRUE)

      it_i <- 1
      while (inherits(eigvals_PCA, "try-error") && it_i < 25) {
        eigvals_PCA <- try(future.apply::future_lapply(size_vec, .parallel_sim,
                                                       N = N,
                                                       n_vars = n_vars,
                                                       eigen_type = 1,
                                                       future.seed = TRUE),
                           silent = TRUE)
        it_i <- it_i + 1
      }

      if (inherits(eigvals_PCA, "try-error")) {
        stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Eigenvalues based on simulated data and PCA could not be found in 25 tries; likely due to occurrences of singular matrices. Aborting.\n"))
      }

      eigvals_PCA <- do.call(rbind, eigvals_PCA)

      results_PCA <- .parallel_summarise(eigvals_PCA, percent = percent,
                                        n_datasets = n_datasets, n_vars = n_vars)

      colnames(results_PCA) <- c("Means", paste(percent, "Percentile"))

      if (isTRUE(x_dat)) {
        n_fac_PCA <- .determine_factors(decision_rule = decision_rule,
                                        eigvals_real = eigvals_real_PCA,
                                        results = results_PCA,
                                        percent = percent)

        eigenvalues_PCA <- cbind(eigvals_real_PCA, results_PCA)

      } else {

        eigenvalues_PCA <- results_PCA

      }

    }

    if ("SMC" %in% eigen_type) {

      eigvals_SMC <- try(future.apply::future_lapply(size_vec, .parallel_sim,
                                                     N = N, n_vars = n_vars,
                                                     eigen_type = 2,
                                                     maxit = n_datasets * 10,
                                                     future.seed = TRUE),
                         silent = TRUE)
      it_i <- 1
      while (inherits(eigvals_SMC, "try-error") && it_i < 25) {
        eigvals_SMC <- try(future.apply::future_lapply(size_vec, .parallel_sim,
                                                       N = N,
                                                       n_vars = n_vars,
                                                       eigen_type = 2,
                                                       maxit = n_datasets * 10,
                                                       future.seed = TRUE),
                           silent = TRUE)
        it_i <- it_i + 1
      }

      if (inherits(eigvals_SMC, "try-error")) {
        stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Eigenvalues based on simulated data and SMCs could not be found in 25 tries; likely due to occurrences of singular matrices. Aborting.\n"))
      }
      eigvals_SMC <- do.call(rbind, eigvals_SMC)

      results_SMC <- .parallel_summarise(eigvals_SMC, percent = percent,
                                        n_datasets = n_datasets, n_vars = n_vars)

      colnames(results_SMC) <- c("Means", paste(percent, "Percentile"))

      if (isTRUE(x_dat)) {
      n_fac_SMC <- .determine_factors(decision_rule = decision_rule,
                                      eigvals_real = eigvals_real_SMC,
                                      results = results_SMC,
                                      percent = percent)

      eigenvalues_SMC <- cbind(eigvals_real_SMC, results_SMC)

      } else {

        eigenvalues_SMC <- results_SMC

      }

    }

    if ("EFA" %in% eigen_type) {

      eigvals_EFA <- future.apply::future_lapply(size_vec, .parallel_EFA_sim,
                                             n_vars = n_vars, N = N,
                                             n_factors = n_factors, ...,
                                             future.seed = TRUE)
      eigvals_EFA <- do.call(rbind, eigvals_EFA)

      results_EFA <- .parallel_summarise(eigvals_EFA, percent = percent,
                                        n_datasets = n_datasets, n_vars = n_vars)

      colnames(results_EFA) <- c("Means", paste(percent, "Percentile"))

      if (isTRUE(x_dat)) {
      n_fac_EFA <- .determine_factors(decision_rule = decision_rule,
                                      eigvals_real = eigvals_real_EFA,
                                      results = results_EFA,
                                      percent = percent)

      eigenvalues_EFA <- cbind(eigvals_real_EFA, results_EFA)

      } else {

        eigenvalues_EFA <- results_EFA

      }

    }

  settings <- list(
    x_dat = x_dat,
    N = N,
    n_vars = n_vars,
    n_datasets = n_datasets,
    percent = percent,
    eigen_type = eigen_type,
    use = use,
    cor_method = cor_method,
    decision_rule = decision_rule,
    n_factors = n_factors
  )

  out <- list(
    eigenvalues_PCA = eigenvalues_PCA,
    eigenvalues_SMC = eigenvalues_SMC,
    eigenvalues_EFA = eigenvalues_EFA,
    n_fac_PCA = n_fac_PCA,
    n_fac_SMC = n_fac_SMC,
    n_fac_EFA = n_fac_EFA,
    settings = settings
  )

  class(out) <- "PARALLEL"

  return(out)

}


.parallel_EFA_sim <- function(n_datasets, n_vars, N, n_factors, ...){

  eigvals <- matrix(nrow = n_datasets, ncol = n_vars)

  for(i in seq_len(n_datasets)){

    x <- matrix(stats::rnorm(N * n_vars), nrow = N, ncol = n_vars)
    R <- stats::cor(x)
    eigvals_i <- try(suppressWarnings(suppressMessages(EFA(R, n_factors = n_factors, N = N,
                                          ...)$final_eigen)), silent = TRUE)
    it_i <- 1
    while (inherits(eigvals_i, "try-error") && it_i < 25) {
      x <- matrix(stats::rnorm(N * n_vars), nrow = N, ncol = n_vars)
      R <- stats::cor(x)
      eigvals_i <- try(suppressWarnings(suppressMessages(EFA(R, n_factors = n_factors, N = N,
                                            ...)$final_eigen)), silent = TRUE)
      it_i <- it_i + 1
    }

    if (inherits(eigvals_i, "try-error")) {
      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Eigenvalues based on simulated data and EFA could not be found in 25 tries; likely due to occurrences of singular matrices. Aborting.\n"))
    }

    eigvals[i,] <- eigvals_i

  }

  return(eigvals)
}

.determine_factors <- function(decision_rule, eigvals_real, results, percent){

# determine the number of factors to retain
if (decision_rule == "crawford") {
  # n factors from resampling
  if ("95 Percentile" %in% colnames(results)) {
    crawford <- c(results[1, "95 Percentile"],
                  results[-1, "Means"])
    n_fac <- which(!(eigvals_real > crawford))[1] - 1
  } else {
    warning(crayon::yellow$bold("!"), crayon::yellow(" decision_rule == 'crawford' is specified, but 95 percentile was not used. Using means instead. To use 'crawford', make sure to specify percent = 95.\n"))
    n_fac <- which(!(eigvals_real > results[, "Means"]))[1] - 1
    decision_rule <- "means"
  }

} else if (decision_rule == "means") {
  n_fac <- which(!(eigvals_real > results[, "Means"]))[1] - 1

} else if (decision_rule == "percentile") {

  pp <- paste(percent, "Percentile")
  n_fac <- which(!(eigvals_real > results[, pp]))[1] - 1

}

  return(n_fac)

}

.parallel_summarise <- function(eig_vals, percent, n_datasets, n_vars) {

  results <- matrix(NA, nrow = n_vars, ncol = length(percent) + 1)
  results[, 1] <- colMeans(eig_vals)

  for (root in seq_len(n_vars)) {
    for (perc_i in seq_along(percent)) {
      ind <- round((percent[perc_i] * n_datasets) / 100)
      results[root, 1 + perc_i] <- sort(eig_vals[, root])[ind]
    }
  }

  return(results)
}

pa_cor <- PARALLEL(test_models$baseline$cormat, N = 500)
pa_cor_pca <- PARALLEL(test_models$baseline$cormat, N = 500, eigen_type = "PCA")
pa_raw <- PARALLEL(GRiPS_raw)
pa_nodat <- PARALLEL(N = 20, n_vars = 5)
pa_craw <- PARALLEL(test_models$baseline$cormat, N = 500, eigen_type = "PCA",
                    decision_rule = "crawford")
pa_perc <- PARALLEL(test_models$baseline$cormat, N = 500, eigen_type = "PCA",
                    decision_rule = "percentile")

test_that("output class and dimensions are correct", {
  expect_is(pa_cor, "PARALLEL")
  expect_output(str(pa_cor), "List of 7")
  expect_is(pa_raw, "PARALLEL")
  expect_output(str(pa_raw), "List of 7")
  expect_is(pa_cor_pca, "PARALLEL")
  expect_output(str(pa_cor_pca), "List of 7")
  expect_is(pa_nodat, "PARALLEL")
  expect_output(str(pa_nodat), "List of 7")
  expect_is(pa_craw, "PARALLEL")
  expect_output(str(pa_craw), "List of 7")
  expect_is(pa_perc, "PARALLEL")
  expect_output(str(pa_perc), "List of 7")
})


test_that("found eigenvalues are correct", {
  expect_equal(sum(pa_cor$eigenvalues_PCA[, 1]),
               ncol(test_models$baseline$cormat))
  expect_equal(sum(pa_cor$eigenvalues_PCA[, 2]),
               ncol(test_models$baseline$cormat))
  expect_gt(sum(pa_cor$eigenvalues_PCA[, 3]),
               ncol(test_models$baseline$cormat))
  expect_lt(sum(pa_cor$eigenvalues_SMC[, 1]),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(pa_cor$eigenvalues_SMC[, 2]),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(pa_cor$eigenvalues_EFA[, 1]),
               ncol(test_models$baseline$cormat))
  expect_lt(sum(pa_cor$eigenvalues_EFA[, 2]),
               ncol(test_models$baseline$cormat))

  expect_equal(sum(pa_raw$eigenvalues_PCA[, 1]),
               ncol(GRiPS_raw))
  expect_equal(sum(pa_raw$eigenvalues_PCA[, 2]),
               ncol(GRiPS_raw))
  expect_gt(sum(pa_raw$eigenvalues_PCA[, 3]),
            ncol(GRiPS_raw))
  expect_lt(sum(pa_raw$eigenvalues_SMC[, 1]),
            ncol(GRiPS_raw))
  expect_lt(sum(pa_raw$eigenvalues_SMC[, 2]),
            ncol(GRiPS_raw))
  expect_lt(sum(pa_raw$eigenvalues_EFA[, 1]),
            ncol(GRiPS_raw))
  expect_lt(sum(pa_raw$eigenvalues_EFA[, 2]),
            ncol(GRiPS_raw))

  expect_is(pa_cor$eigenvalues_PCA, "matrix")
  expect_is(pa_cor$eigenvalues_SMC, "matrix")
  expect_is(pa_cor$eigenvalues_EFA, "matrix")

  expect_is(pa_raw$eigenvalues_PCA, "matrix")
  expect_is(pa_raw$eigenvalues_SMC, "matrix")
  expect_is(pa_raw$eigenvalues_EFA, "matrix")

  expect_is(pa_nodat$eigenvalues_PCA, "matrix")
  expect_is(pa_nodat$eigenvalues_SMC, "matrix")
  expect_is(pa_nodat$eigenvalues_EFA, "matrix")

  expect_is(pa_cor_pca$eigenvalues_PCA, "matrix")
  expect_equal(c(pa_cor_pca$eigenvalues_SMC, pa_cor_pca$eigenvalues_EFA), c(NA, NA))

  expect_is(pa_craw$eigenvalues_PCA, "matrix")
  expect_equal(c(pa_craw$eigenvalues_SMC, pa_craw$eigenvalues_EFA), c(NA, NA))

  expect_is(pa_perc$eigenvalues_PCA, "matrix")
  expect_equal(c(pa_perc$eigenvalues_SMC, pa_perc$eigenvalues_EFA), c(NA, NA))

})


test_that("identified number of factors is correct", {
  expect_equal(pa_cor$n_fac_PCA, 3)
  expect_equal(pa_cor$n_fac_SMC, 3)
  expect_equal(pa_cor$n_fac_EFA, 7, tolerance = 2)

  expect_equal(pa_raw$n_fac_PCA, 1)
  expect_equal(pa_raw$n_fac_SMC, 1)
  expect_equal(pa_raw$n_fac_EFA, 3, tolerance = 2)

  expect_equal(c(pa_nodat$n_fac_PCA, pa_nodat$n_fac_SMC, pa_nodat$n_fac_EFA),
               rep(NA, 3))

  expect_equal(pa_cor_pca$n_fac_PCA, 3)
  expect_equal(c(pa_cor_pca$n_fac_SMC, pa_cor_pca$n_fac_EFA), c(NA, NA))

  expect_equal(pa_craw$n_fac_PCA, 3)
  expect_equal(c(pa_craw$n_fac_SMC, pa_craw$n_fac_EFA), c(NA, NA))

  expect_equal(pa_perc$n_fac_PCA, 3)
  expect_equal(c(pa_perc$n_fac_SMC, pa_perc$n_fac_EFA), c(NA, NA))
})


test_that("settings are returned correctly", {
  expect_named(pa_cor$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                  "eigen_type", "use", "cor_method", "decision_rule",
                                  "n_factors"))
  expect_named(pa_raw$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                  "eigen_type", "use", "cor_method", "decision_rule",
                                  "n_factors"))
  expect_named(pa_cor_pca$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))
  expect_named(pa_nodat$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))
  expect_named(pa_craw$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))
  expect_named(pa_perc$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))

  expect_true(pa_cor$settings$x_dat)
  expect_true(pa_raw$settings$x_dat)
  expect_true(pa_cor_pca$settings$x_dat)
  expect_false(pa_nodat$settings$x_dat)
  expect_true(pa_craw$settings$x_dat)
  expect_true(pa_perc$settings$x_dat)

  expect_equal(pa_cor$settings$N, 500)
  expect_equal(pa_raw$settings$N, 810)
  expect_equal(pa_cor_pca$settings$N, 500)
  expect_equal(pa_nodat$settings$N, 20)
  expect_equal(pa_craw$settings$N, 500)
  expect_equal(pa_perc$settings$N, 500)

  expect_equal(pa_cor$settings$n_vars, 18)
  expect_equal(pa_raw$settings$n_vars, 8)
  expect_equal(pa_cor_pca$settings$n_vars, 18)
  expect_equal(pa_nodat$settings$n_vars, 5)
  expect_equal(pa_craw$settings$n_vars, 18)
  expect_equal(pa_perc$settings$n_vars, 18)

  expect_equal(pa_cor$settings$n_datasets, 1000)
  expect_equal(pa_raw$settings$n_datasets, 1000)
  expect_equal(pa_cor_pca$settings$n_datasets, 1000)
  expect_equal(pa_nodat$settings$n_datasets, 1000)
  expect_equal(pa_craw$settings$n_datasets, 1000)
  expect_equal(pa_perc$settings$n_datasets, 1000)

  expect_equal(pa_cor$settings$percent, 95)
  expect_equal(pa_raw$settings$percent, 95)
  expect_equal(pa_cor_pca$settings$percent, 95)
  expect_equal(pa_nodat$settings$percent, 95)
  expect_equal(pa_craw$settings$percent, 95)
  expect_equal(pa_perc$settings$percent, 95)

  expect_equal(pa_cor$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(pa_raw$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(pa_cor_pca$settings$eigen_type, "PCA")
  expect_equal(pa_nodat$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(pa_craw$settings$eigen_type, "PCA")
  expect_equal(pa_perc$settings$eigen_type, "PCA")

  expect_equal(pa_cor$settings$use, "pairwise.complete.obs")
  expect_equal(pa_raw$settings$use, "pairwise.complete.obs")
  expect_equal(pa_cor_pca$settings$use, "pairwise.complete.obs")
  expect_equal(pa_nodat$settings$use, "pairwise.complete.obs")
  expect_equal(pa_craw$settings$use, "pairwise.complete.obs")
  expect_equal(pa_perc$settings$use, "pairwise.complete.obs")

  expect_equal(pa_cor$settings$cor_method, "pearson")
  expect_equal(pa_raw$settings$cor_method, "pearson")
  expect_equal(pa_cor_pca$settings$cor_method, "pearson")
  expect_equal(pa_nodat$settings$cor_method, "pearson")
  expect_equal(pa_craw$settings$cor_method, "pearson")
  expect_equal(pa_perc$settings$cor_method, "pearson")

  expect_equal(pa_cor$settings$decision_rule, "means")
  expect_equal(pa_raw$settings$decision_rule, "means")
  expect_equal(pa_cor_pca$settings$decision_rule, "means")
  expect_equal(pa_nodat$settings$decision_rule, "means")
  expect_equal(pa_craw$settings$decision_rule, "crawford")
  expect_equal(pa_perc$settings$decision_rule, "percentile")

  expect_equal(pa_cor$settings$n_factors, 1)
  expect_equal(pa_raw$settings$n_factors, 1)
  expect_equal(pa_cor_pca$settings$n_factors, 1)
  expect_equal(pa_nodat$settings$n_factors, 1)
  expect_equal(pa_craw$settings$n_factors, 1)
  expect_equal(pa_perc$settings$n_factors, 1)

})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z), ncol = 3)
cor_sing <- stats::cor(dat_sing)

burt <- matrix(c(1.00,  0.83,  0.81,  0.80,   0.71, 0.70, 0.54, 0.53,  0.59,  0.24, 0.13,
                 0.83,  1.00,  0.87,  0.62,   0.59, 0.44, 0.58, 0.44,  0.23,  0.45,  0.21,
                 0.81,  0.87,  1.00,  0.63,   0.37, 0.31, 0.30, 0.12,  0.33,  0.33,  0.36,
                 0.80,  0.62,  0.63,  1.00,   0.49, 0.54, 0.30, 0.28,  0.42,  0.29, -0.06,
                 0.71,  0.59,  0.37,  0.49,   1.00, 0.54, 0.34, 0.55,  0.40,  0.19, -0.10,
                 0.70,  0.44,  0.31,  0.54,   0.54, 1.00, 0.50, 0.51,  0.31,  0.11,  0.10,
                 0.54,  0.58,  0.30,  0.30,   0.34, 0.50, 1.00, 0.38,  0.29,  0.21,  0.08,
                 0.53,  0.44,  0.12,  0.28,   0.55, 0.51, 0.38, 1.00,  0.53,  0.10, -0.16,
                 0.59,  0.23,  0.33,  0.42,   0.40, 0.31, 0.29, 0.53,  1.00, -0.09, -0.10,
                 0.24,  0.45,  0.33,  0.29,   0.19, 0.11, 0.21, 0.10, -0.09,  1.00,  0.41,
                 0.13,  0.21,  0.36, -0.06,  -0.10, 0.10, 0.08, -0.16, -0.10, 0.41,  1.00),
               nrow = 11, ncol = 11)


#sim_NA <- data.frame(rnorm(30), rnorm(30), rnorm(30), rep("a", 30))

test_that("errors are thrown correctly", {
  expect_error(PARALLEL(1:5), " 'x' is neither NULL, nor a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data or leave x at NULL.\n")
  expect_warning(suppressMessages(PARALLEL(GRiPS_raw, n_vars = 5)), " n_vars was set and data entered. Taking n_vars from data\n")
  expect_warning(suppressMessages(PARALLEL(GRiPS_raw, N = 20, eigen_type = "PCA")), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(suppressMessages(PARALLEL(N = 500)), ' "n_vars" was not set and could not be taken from data. Please specify n_vars and try again.\n')
  expect_message(PARALLEL(GRiPS_raw, eigen_type = "PCA"), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_error(PARALLEL(test_models$baseline$cormat, eigen_type = "PCA"), '"N" was not set and could not be taken from data. Please specify N and try again.\n')
  expect_warning(PARALLEL(test_models$baseline$cormat, N = 500,
                          eigen_type = "PCA", decision_rule = "crawford",
                          percent = 80), "decision_rule == 'crawford' is specified, but 95 percentile was not used. Using means instead. To use 'crawford', make sure to specify percent = 95.\n")
  expect_error(PARALLEL(dat_sing), " Correlation matrix is singular, parallel analysis is not possible.\n")
  expect_error(PARALLEL(cor_sing, N = 10), " Correlation matrix is singular, parallel analysis is not possible.\n")
  expect_warning(PARALLEL(burt, N = 100), "Matrix was not positive definite, smoothing was done")
  expect_error(PARALLEL(test_models$baseline$cormat, N = 15), ' "N" was smaller than or equal to the number of variables but must be larger.\n')
  expect_error(PARALLEL(test_models$baseline$cormat, N = 18), ' "N" was smaller than or equal to the number of variables but must be larger.\n')
})

rm(pa_cor, pa_cor_pca, pa_raw, pa_nodat, pa_craw, pa_perc, x, y, z, dat_sing,
   cor_sing, burt)

#' Plot SCREE object
#'
#' Plot method showing a summarized output of the \link{SCREE} function
#'
#' @param x a list of class SCREE An output from the \link{SCREE} function.
#' @param ... not used.
#'
#' @export
#' @method plot SCREE
#'
#' @examples
#' SCREE_base <- SCREE(test_models$baseline$cormat)
#' plot(SCREE_base)
#'
plot.SCREE <- function(x, ...) {

  eigen_PCA <- x$eigen_PCA
  eigen_SMC <- x$eigen_SMC
  eigen_EFA <- x$eigen_EFA

  # Create plots
  if(!all(is.na(eigen_PCA))){

    .plot_SCREE_helper(eigen = eigen_PCA, eigen_name = "PCA")

  }

  if (!all(is.na(eigen_SMC))) {

    .plot_SCREE_helper(eigen = eigen_SMC, eigen_name = "SMC")

  }

  if (!all(is.na(eigen_EFA))) {

    .plot_SCREE_helper(eigen = eigen_EFA, eigen_name = "EFA")

  }

}

.plot_SCREE_helper <- function(eigen, eigen_name){
  # eigen = eigenvalues found with specific type (PCA, SMC or EFA)
  # eigen_name = name of type (one of "PCA", "SMC", "EFA")

  x_len <- length(eigen)

  p_eigen <- pretty(c(min(eigen) * .9, eigen, max(eigen) * 1.1))

  graphics::plot.new()
  graphics::plot.window(xlim = c(1, x_len),
                        ylim = c(min(p_eigen), max(p_eigen)))
  graphics::axis(1, seq_len(x_len))
  graphics::axis(2, p_eigen, las = 1)

  graphics::mtext("Factor", side = 1, line = 3, cex = 1)
  graphics::mtext("Eigenvalues", side = 2, line = 3, cex = 1, padj =.5)

  graphics::lines(seq_len(x_len), eigen)
  graphics::points(seq_len(x_len), eigen, pch = 16)

  title <- paste0("Scree plot with ", eigen_name, "-determined eigenvalues",
                  sep = "")

  graphics::title(title, cex.main = 1.2)

}

#' Print EFA_AVERAGE object
#'
#' Print Method showing a summarized output of the \link{EFA_AVERAGE} function
#'
#' @param x list. An object of class EFA_AVERAGE to be printed
#' @param stat character. A vector with the statistics to print. Possible inputs
#' are "average", "sd", "range", "min", and "max". Default is "average" and
#' "range".
#' @param plot logical. Whether a plot of the average and min- max loadings should
#' be created. Default is TRUE. If more than 10 factors are extracted, no plot is
#' created.
#' @param ...  Further arguments for print.
#'
#' @export
#'
#' @method print EFA_AVERAGE
#'
#' @examples
#' \dontrun{
#' EFA_aver <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)
#' EFA_aver
#' }
print.EFA_AVERAGE <- function(x, stat = c("average", "range"),
                              plot = TRUE, ...) {

  checkmate::assert_subset(stat, c("average", "sd", "range", "min", "max"),
                           empty.ok = FALSE)

  # extract settings
  settings <- x$settings
  method <- settings$method
  rotation <- settings$rotation
  N <- settings$N
  grid <- x$implementations_grid
  averaging <- settings$averaging

  # settings that were varied
  varied_settings <- grid
  varied_settings[, c("errors", "error_m", "converged", "heywood", "chisq",
                      "p_chi", "cfi", "caf", "rmsea", "aic", "bic")] <- NULL
  varied_settings <- apply(varied_settings, 2, function(x)unique(x[!is.na(x)]))
  varied_settings <- sapply(varied_settings, length)
  varied_settings <- names(varied_settings[varied_settings > 1])

  # extract other stuff
  no_efas <- nrow(grid)

  cat("\n")
  cat("Averaging performed with averaging method ",
      crayon::bold(ifelse(averaging == "median", "median",
                          paste0("mean (trim = ", settings$trim, ")", sep = ""))),
      " across ", crayon::bold(no_efas), " EFAs, ",
      "varying the following settings: ",
      .settings_string(varied_settings), ".", sep = "")
  cat("\n")

  cat("\n")
  cat("The error rate is at ",
      crayon::bold(round(mean(grid$errors, na.rm = TRUE) * 100), "%", sep = ""),
                   ". Of the solutions that did not result in an error, ",
      crayon::bold(round(mean(grid$converged == 0, na.rm = TRUE) * 100), "%",
                   sep = ""),
      " converged, ",
      crayon::bold(round(mean(grid$heywood, na.rm = TRUE) * 100), "%", sep = ""),
      " contained Heywood cases, and ",
      crayon::bold(round(mean(grid$admissible, na.rm = TRUE) * 100), "%", sep = ""),
      " were admissible.", sep = "")
  cat("\n")
  cat("\n")

  # If no solutions were achieved across which averaging could be performed,
  # stop here with a message. Else, continue printing loadings etc.
  if(all(grid$converged != 0 | grid$errors | grid$heywood)){

    warning(crayon::yellow$bold("!"), crayon::yellow(" No solutions were achieved across which averaging was possible. Best try again with a different number of factors.\n"))

  } else {

    fit <- x$fit_indices
    rownames(fit) <- fit$index

  # Indicator-to-factor correspondences
  cat("\n")
  cat(cli::rule(left = crayon::bold("Indicator-to-Factor Correspondences"),
                col = "blue", line = 2))
  cat("\n")
  cat("\n")
  cat("For each cell, the proportion of solutions including the respective indicator-to-factor correspondence. A salience threshold of",
      crayon::bold(settings$salience_threshold),
      "was used to determine indicator-to-factor correspondences.")
  cat("\n")
  cat("\n")
  cat(print.LOADINGS(x$ind_fac_corres, cutoff = 1e-4, digits = 2))
  cat("\n")

  # Print the loadings
    cat("\n")
    cat(cli::rule(left = crayon::bold("Loadings"), col = "blue", line = 2))
    cat("\n")
    .print_average(x, what = c("loadings"), stat = stat, averaging = averaging)
    cat("\n")

  ## Print Phi for oblique solutions
  if(!all(is.na(x$Phi))){

    cat("\n")
    cat(cli::rule(left = crayon::bold("Factor Intercorrelations from Oblique Solutions"), col = "blue", line = 2))
    cat("\n")
    .print_average(x, what = c("Phi"), stat = stat, averaging = averaging)
    cat("\n")
  }

  # Variances accounted for
    cat("\n")
    cat(cli::rule(left = crayon::bold("Variances Accounted for"), col = "blue",
                  line = 2))
    cat("\n")
    .print_average(x, what = c("vars_accounted"), stat = stat,
                   averaging = averaging)
    cat("\n")


  # Print fit indices
  if (fit["df", "average"] == 0) {
    cat("\n")
    cat(crayon::yellow$bold("!"), crayon::yellow(" The model is just identified (df = 0). Goodness of fit indices may not be interpretable."))
    cat("\n")
  }

    cat("\n")
    cat(cli::rule(left = crayon::bold("Model Fit"), col = "blue", line = 2))
    cat("\n")
    cat("\n")
    cat(crayon::blue("       ", ifelse(averaging == "mean", "M", "Md"),
                     " (SD) [Min; Max]", sep = ""))
    cat("\n")

  if(all(method == "PAF") || is.na(N)){

    .print_gof(fit, ind = "caf", ind_name = "CAF:  ", print_zero = FALSE, digits = 2)
    cat(crayon::blue("df: "),
        .numformat(fit["df", "average"], 0, print_zero = TRUE), "\n", sep = "")

  } else {

    .print_gof(fit, ind = c("chisq"), ind_name = "\U1D712\U00B2: ",
               print_zero = TRUE, digits = 2)
    cat(crayon::blue("df: "),
        .numformat(fit["df", "average"], 0, print_zero = TRUE), "\n", sep = "")
    .print_gof(fit, ind = c("p_chi", "cfi", "rmsea", "aic", "bic", "caf"),
               ind_name = c(crayon::italic("p: "), "CFI: ", "RMSEA: ",
                            "AIC: ", "BIC: ", "CAF: "),
               print_zero = c(FALSE, FALSE, FALSE, TRUE, TRUE, FALSE),
               digits = c(3, 2, 2, 2, 2, 2))

  }

    # Plot loadings
    if(isTRUE(plot)){

    if(ncol(x$loadings$average) <= 10){

      plot(x)

    } else {

      message(cli::col_cyan(cli::symbol$info, " The factor solution contained more than 10 factors, no plot was generated. If you still want to create the plot, use 'plot(", substitute(x) ,")'.\n"))

    }

    }

  }

}

.print_average <- function(x, what, stat, averaging){

  if("average" %in% stat){

    if(averaging == "mean"){

      cat("\n")
      cat(cli::rule(left = crayon::bold("Mean"), col = "blue"))
      cat("\n")
      cat("\n")

    } else {

      cat("\n")
      cat(cli::rule(left = crayon::bold("Median"), col = "blue"))
      cat("\n")
      cat("\n")

    }

    if(what == "loadings"){

      print(x$loadings$average)

      } else if(what == "Phi"){

        cat(.get_compare_matrix(x$Phi$average, r_red = Inf, n_char = 17,
                                var_names = paste0("F",
                                                   seq_len(ncol(x$Phi$average)))))

      } else {

        cat(.get_compare_matrix(x$vars_accounted$average, r_red = Inf,
                                n_char = 17))

      }

    }

  if("sd" %in% stat){

    cat("\n")
      cat(cli::rule(left = crayon::bold("Standard Deviation"), col = "blue"))
      cat("\n")
      cat("\n")

      if(what == "loadings"){

        cat(.get_compare_matrix(x$loadings$sd, r_red = Inf, n_char = 17))

      } else if(what == "Phi"){

        cat(.get_compare_matrix(x$Phi$sd, r_red = Inf, n_char = 17,
                                var_names = paste0("F",
                                                   seq_len(ncol(x$Phi$sd)))))

      } else {

        cat(.get_compare_matrix(x$vars_accounted$sd, r_red = Inf,
                                n_char = 17))

      }

  }

  if("range" %in% stat){

    cat("\n")
    cat(cli::rule(left = crayon::bold("Range"), col = "blue"))
    cat("\n")
    cat("\n")

    if(what == "loadings"){

      cat(.get_compare_matrix(x$loadings$range, r_red = Inf, n_char = 17))

    } else if(what == "Phi"){

      cat(.get_compare_matrix(x$Phi$range, r_red = Inf, n_char = 17,
                              var_names = paste0("F",
                                                 seq_len(ncol(x$Phi$range)))))

    } else {

      cat(.get_compare_matrix(x$vars_accounted$range, r_red = Inf,
                              n_char = 17))

    }

  }

  if("min" %in% stat){

    cat("\n")
    cat(cli::rule(left = crayon::bold("Minimum"), col = "blue"))
    cat("\n")
    cat("\n")

    if(what == "loadings"){

      print(x$loadings$min)

    } else if(what == "Phi"){

      cat(.get_compare_matrix(x$Phi$min, r_red = Inf, n_char = 17,
                              var_names = paste0("F",
                                                 seq_len(ncol(x$Phi$min)))))

    } else {

      cat(.get_compare_matrix(x$vars_accounted$min, r_red = Inf,
                              n_char = 17))

    }

  }

  if("max" %in% stat){

    cat("\n")
    cat(cli::rule(left = crayon::bold("Maximum"), col = "blue"))
    cat("\n")
    cat("\n")

    if(what == "loadings"){

      print(x$loadings$max)

    } else if(what == "Phi"){

      cat(.get_compare_matrix(x$Phi$max, r_red = Inf, n_char = 17,
                              var_names = paste0("F",
                                                 seq_len(ncol(x$Phi$max)))))

    } else {

      cat(.get_compare_matrix(x$vars_accounted$max, r_red = Inf,
                              n_char = 17))

    }

  }
}

.print_gof <- function(fit, ind, ind_name, print_zero, digits){

    for(i in seq_along(ind)){

      if(ind[i] %in% c("p_chi", "cfi", "rmsea", "caf")){

  cat(crayon::blue(ind_name[i], sep = ""),
      ifelse(round(fit[ind[i], "average"], digits[i]) < 1,
             substr(.numformat(fit[ind[i], "average"], digits = digits[i],
                               print_zero = print_zero[i]),
                    2, digits + 2),
             .numformat(fit[ind[i], "average"], digits = digits[i],
                        print_zero = print_zero[i])), " (",
      ifelse(round(fit[ind[i], "sd"], digits[i]) < 1,
             substr(.numformat(fit[ind[i], "sd"], digits = digits[i],
                               print_zero = print_zero[i]),
                    2, digits + 2),
             .numformat(fit[ind[i], "sd"], digits = digits[i],
                        print_zero = print_zero[i])),
      ") [",
      ifelse(round(fit[ind[i], "min"], digits[i]) < 1,
             substr(.numformat(fit[ind[i], "min"], digits = digits[i],
                               print_zero = print_zero[i]),
                    2, digits + 2),
             .numformat(fit[ind[i], "min"], digits = digits[i],
                        print_zero = print_zero[i])), "; ",
      ifelse(round(fit[ind[i], "max"], digits[i]) < 1,
             substr(.numformat(fit[ind[i], "max"], digits = digits[i],
                               print_zero = print_zero[i]),
                    2, digits + 2),
             .numformat(fit[ind[i], "max"], digits = digits[i],
                        print_zero = print_zero[i])),
      "]\n", sep = "")

      } else {

        cat(crayon::blue(ind_name[i], sep = ""),
            .numformat(fit[ind[i], "average"], digits = digits[i],
                                     print_zero = print_zero[i]), " (",
            .numformat(fit[ind[i], "sd"], digits = digits[i],
                       print_zero = print_zero[i]),
            ") [",
            .numformat(fit[ind[i], "min"], digits = digits[i],
                       print_zero = print_zero[i]),
            "; ",
            .numformat(fit[ind[i], "max"], digits = digits[i],
                       print_zero = print_zero[i]),
            "]\n", sep = "")


      }

    }

  }

#' Print function for EKC objects
#'
#' @param x a list of class EKC. Output from \code{\link{EKC}} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print EKC
#'
#' @examples
#' EKC_base <- EKC(test_models$baseline$cormat, N = 500)
#' EKC_base
#'
print.EKC <- function(x, plot = TRUE, ...) {

  n_factors_BvA2017 <- x$n_factors_BvA2017
  n_factors_AM2019 <- x$n_factors_AM2019
  ekc_type <- x$settings$type

  cat(cli::rule("Number of factors suggested by EKC",
                col = "blue"))
  cat("\n")
  cat("\n")

  if("BvA2017" %in% ekc_type){

    cat(crayon::blue(cli::symbol$bullet, "Original implementation (Braeken & van Assen, 2017): "),
        crayon::bold(n_factors_BvA2017))
    cat("\n")

  }

  if("AM2019" %in% ekc_type){

    cat(crayon::blue(cli::symbol$bullet,
                     "Adapted implementation (Auerswald & Moshagen, 2019): "),
        crayon::bold(n_factors_AM2019))
    cat("\n")

  }

  cli::cli_alert_info("Different implementations of EKC exist. Make sure to report which one you relied on (see EKC help page for details)")


  if (isTRUE(plot)) {
    graphics::plot(x)
  }

}

# Orthogonal factor rotations with GPArotation package
.ROTATE_ORTH <- function(x, rotation = c("equamax", "quartimax", "geominT",
                                        "bentlerT", "bifactorT"),
                        type = c("EFAtools", "psych", "SPSS", "none"),
                        normalize = TRUE, precision = 1e-5, order_type = NA,
                        ...){

if (type == "none") {

  if (is.na(order_type)) {

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' "order_type" was NA and no valid "type" was specified. Either use one of "EFAtools", "psych", or "SPSS" for type, or specify the "order_type" argument\n'))
  }

} else if (type == "EFAtools") {

  if (isFALSE(normalize)) {

    warning(crayon::yellow$bold("!"), crayon::yellow(" Type and normalize is specified. normalize is used with value '", normalize, "'. Results may differ from the specified type\n"))
  }

  if (is.na(order_type)) {
    order_type <- "eigen"
  } else {
    warning(crayon::yellow$bold("!"), crayon::yellow(" Type and order_type is specified. order_type is used with value '", order_type, "'. Results may differ from the specified type\n"))
  }

} else if (type == "psych") {

  if (isFALSE(normalize)) {

    warning(crayon::yellow$bold("!"), crayon::yellow(" Type and normalize is specified. normalize is used with value '", normalize, "'. Results may differ from the specified type\n"))
  }

  if (is.na(order_type)) {
    order_type <- "eigen"
  } else {
    warning(crayon::yellow$bold("!"), crayon::yellow(" Type and order_type is specified. order_type is used with value '", order_type, "'. Results may differ from the specified type\n"))
  }

} else if (type == "SPSS") {

  if (isFALSE(normalize)) {

    warning(crayon::yellow$bold("!"), crayon::yellow(" Type and normalize is specified. normalize is used with value '", normalize, "'. Results may differ from the specified type\n"))
  }

  if (is.na(order_type)) {
    order_type <- "ss_factors"
  } else {
    warning(crayon::yellow$bold("!"), crayon::yellow(" Type and order_type is specified. order_type is used with value '", order_type, "'. Results may differ from the specified type\n"))
  }

}

  # extract loadings and dim names
  L <- x$unrot_loadings
  dim_names <- dimnames(L)

  # prepare settings
  settings <- list(normalize = normalize, precision = precision,
                   order_type = order_type)

  if (ncol(L) < 2) {

    # prepare and return output list
    output <- list(rot_loadings = L,
                   rotmat = NA,
                   vars_accounted_rot = NA,
                   settings = settings)

    warning(crayon::yellow$bold("!"), crayon::yellow(" Cannot rotate single factor. Unrotated loadings returned.\n"))
    return(output)
  }

  # perform the requested rotation
  if(rotation == "equamax"){
  AV <- GPArotation::cfT(L, eps = precision, kappa = ncol(L)/(2 * nrow(L)),
                         normalize = normalize, ...)

  } else if (rotation == "bentlerT"){
    AV <- GPArotation::bentlerT(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "quartimax"){
    AV <- GPArotation::bentlerT(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "geominT"){
    AV <- GPArotation::geominT(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "bifactorT"){
    AV <- GPArotation::bifactorT(L, eps = precision, normalize = normalize, ...)

  }

  # Extract loading and rotation matrix
  load_mat <- AV$loadings
  dimnames(load_mat) <- dim_names
  rotmat <- AV$Th

  # reflect factors with negative sums
  signs <- sign(colSums(load_mat))
  signs[signs == 0] <- 1
  load_mat <- load_mat %*% diag(signs)

  if (order_type == "ss_factors") {

    # reorder the factors according to largest sums of squares
    ss <- colSums(load_mat^2)
    ss_order <- order(ss, decreasing = TRUE)

    load_mat <- load_mat[, ss_order]

    rotmat <- rotmat[ss_order, ss_order]

    dim_names[[2]] <- dim_names[[2]][ss_order]

  } else if (order_type == "eigen") {

    # order according to communalities
    eig_rotated <- diag(t(load_mat) %*% load_mat)
    eig_order <- order(eig_rotated, decreasing = TRUE)
    load_mat <- load_mat[, eig_order]
    rotmat <- rotmat[eig_order, eig_order]

    dim_names[[2]] <- dim_names[[2]][eig_order]

  }

  # compute explained variances
  vars_accounted_rot <- .compute_vars(L_unrot = L, L_rot = load_mat)
  colnames(vars_accounted_rot) <- colnames(load_mat)

  # prepare output and return output list
  class(load_mat) <- "LOADINGS"

  output <- list(rot_loadings = load_mat,
                 rotmat = rotmat,
                 vars_accounted_rot = vars_accounted_rot,
                 settings = settings)
  output
}

#' Print function for N_FACTORS objects
#'
#' @param x a list of class N_FACTORS. Output from \link{N_FACTORS} function.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print N_FACTORS
#'
#' @examples
#' \donttest{
#' # All criteria except "CD", with correlation matrix and fit method "ML"
#' # (where needed)
#' N_FACTORS(test_models$baseline$cormat, criteria = c("EKC", "HULL", "KGC",
#'           "PARALLEL", "SCREE", "SMT"), N = 500, method = "ML")
#'}
print.N_FACTORS <- function(x, ...){

  suitability <- x$settings$suitability
  criteria <- x$settings$criteria
  n_fac <- x$n_factors
  gof <- x$settings$gof
  eigen_type_other <- x$settings$eigen_type_other
  kmo_out <- x$outputs$kmo_out
  KMO <- kmo_out$KMO

  if(isTRUE(suitability)){

    cat("\n")
    cat(cli::rule(left = crayon::bold("Tests for the suitability of the data for factor analysis"), col = "blue"))
    cat("\n")

    cat("\n")
    cat(crayon::blue$bold("Bartlett's test of sphericity"))
    cat("\n")

    print(x$output$bart_out)
    cat("\n")

    cat(crayon::blue$bold("Kaiser-Meyer-Olkin criterion (KMO)"))
    cat("\n")

    if(!is.na(KMO) && !is.null(KMO)){

      if(KMO >= .9){
        symb <- crayon::green$bold(cli::symbol$tick)
        label <- crayon::green$bold("marvellous")
      } else if(KMO >= .8){
        symb <- crayon::green$bold(cli::symbol$tick)
        label <- crayon::green$bold("meritorious")
      } else if(KMO >= .7){
        symb <- crayon::green$bold(cli::symbol$tick)
        label <- crayon::green$bold("middling")
      } else if(KMO >= .6){
        symb <- crayon::yellow$bold("!")
        label <- crayon::yellow$bold("mediocre")
      } else if (KMO >= .5){
        symb <- crayon::red$bold(cli::symbol$cross)
        label <- crayon::red$bold("miserable")
      } else {
        symb <- crayon::red$bold(cli::symbol$cross)
        label <- crayon::red$bold("unacceptable")
      }

      cat("\n")
      cat(symb, " The overall KMO value for your data is ", label,
          " with ", crayon::bold(round(KMO, 3)), ".", sep = "")
      cat("\n")


      if(KMO < .5){
        cat(crayon::bold(" "), "These data are not suitable for factor analysis.")
        cat("\n")
      } else if(KMO < .6){
        cat(crayon::bold(" "), "These data are hardly suitable for factor analysis.")
        cat("\n")
      } else {
        cat(crayon::bold(" "), "These data are probably suitable for factor analysis.")
        cat("\n")
      }

    } else {

      cat("\n")
      cat("The overall KMO value for your data is not available.")
      cat("\n")

    }

  }

  cat("\n")
  cat(cli::rule(crayon::bold("Number of factors suggested by the different factor",
  "retention criteria"), col = "blue"))
  cat("\n")
  cat("\n")

  if("CD" %in% criteria){

  cat(crayon::blue(cli::symbol$bullet, "Comparison data: "),
      crayon::bold(n_fac["nfac_CD"]), sep = "")
  cat("\n")

  }

  if("EKC" %in% criteria){

    ekc_type <- x$settings$ekc_type

    if("BvA2017" %in% ekc_type){

      cat(crayon::blue(cli::symbol$bullet, "EKC (original implementation, type 'BvA2017'): "),
          crayon::bold(n_fac["nfac_EKC_BvA2017"]),
          sep = "")
      cat("\n")

    }

    if("AM2019" %in% ekc_type){

      cat(crayon::blue(cli::symbol$bullet, "EKC (adapted implementation, type 'AM2019'): "),
          crayon::bold(n_fac["nfac_EKC_AM2019"]),
          sep = "")
      cat("\n")

    }

  }

  if("HULL" %in% criteria){

    if("CAF" %in% gof){
    cat(crayon::blue(cli::symbol$bullet, "Hull method with CAF: "),
        crayon::bold(n_fac["nfac_HULL_CAF"]),
        sep = "")
    cat("\n")
    }
    if("CFI" %in% gof){
    cat(crayon::blue(cli::symbol$bullet, "Hull method with CFI: "),
        crayon::bold(n_fac["nfac_HULL_CFI"]),
        sep = "")
    cat("\n")
    }
    if("RMSEA" %in% gof){
    cat(crayon::blue(cli::symbol$bullet, "Hull method with RMSEA: "),
        crayon::bold(n_fac["nfac_HULL_RMSEA"]),
        sep = "")
    cat("\n")
    }

  }

  if("KGC" %in% criteria){

    if("PCA" %in% eigen_type_other){
    cat(crayon::blue(cli::symbol$bullet,
                     "Kaiser-Guttman criterion with PCA: "),
        crayon::bold(n_fac["nfac_KGC_PCA"]), sep = "")
    cat("\n")
    }
    if("SMC" %in% eigen_type_other){
    cat(crayon::blue(cli::symbol$bullet,
                     "Kaiser-Guttman criterion with SMC: "),
        crayon::bold(n_fac["nfac_KGC_SMC"]), sep = "")
    cat("\n")
    }
    if("EFA" %in% eigen_type_other){
    cat(crayon::blue(cli::symbol$bullet,
                     "Kaiser-Guttman criterion with EFA: "),
        crayon::bold(n_fac["nfac_KGC_EFA"]), sep = "")
    cat("\n")
    }

  }

  if("PARALLEL" %in% criteria){

    if("PCA" %in% eigen_type_other){
    cat(crayon::blue(cli::symbol$bullet,
                     "Parallel analysis with PCA: "),
        crayon::bold(n_fac["nfac_PA_PCA"]),
        sep = "")
    cat("\n")
    }
    if("SMC" %in% eigen_type_other){
    cat(crayon::blue(cli::symbol$bullet,
                     "Parallel analysis with SMC: "),
        crayon::bold(n_fac["nfac_PA_SMC"]),
        sep = "")
    cat("\n")
    }
    if("EFA" %in% eigen_type_other){
    cat(crayon::blue(cli::symbol$bullet,
                     "Parallel analysis with EFA: "),
        crayon::bold(n_fac["nfac_PA_EFA"]),
        sep = "")
    cat("\n")
    }

  }

  if("NEST" %in% criteria){

      cat(crayon::blue(cli::symbol$bullet,
                       "NEST: "),
          crayon::bold(n_fac["nfac_NEST"]),
          sep = "")
      cat("\n")

  }

  if("SMT" %in% criteria){

    cat(crayon::blue(cli::symbol$bullet,
                     "Sequential \U1D712\U00B2 model tests: "),
        crayon::bold(n_fac["nfac_SMT_chi"]), sep = "")
    cat("\n")
    cat(crayon::blue(cli::symbol$bullet,
                     "Lower bound of RMSEA 90% confidence interval: "),
        crayon::bold(n_fac["nfac_RMSEA"]), sep = "")
    cat("\n")
    cat(crayon::blue(cli::symbol$bullet,
                     "Akaike Information Criterion: "),
        crayon::bold(n_fac["nfac_AIC"]),
        sep = "")
    cat("\n")

  }

  if("SCREE" %in% criteria){

    graphics::plot(x$output$scree_out)

  }

}

#' Print SLLOADINGS object
#'
#' @param x class SLLOADINGS matrix.
#' @param cutoff numeric. The number above which to print loadings in bold
#'  (default is .2).
#' @param digits numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
#'  to round the loadings to (default is 3).
#' @param ... additional arguments passed to print
#'
#' @method print SLLOADINGS
#' @export
#'
#' @examples
#' EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                type = "EFAtools", method = "PAF", rotation = "promax")
#' SL(EFA_mod, type = "EFAtools", method = "PAF")
#'
print.SLLOADINGS <- function(x, cutoff = .2, digits = 3, ...) {

  # create factor names to display
  factor_names <- colnames(x)
  if (is.null(factor_names)) {
    factor_names <- paste0("F", seq_len(ncol(x)))
  }

  # for equal spacing, fill the factor names such that they match the columns
  fn_nchar <- sapply(factor_names, nchar)
  factor_names[which(fn_nchar > digits + 2)] <- substr(
    factor_names[which(fn_nchar > digits + 2)] , 1, digits + 2)
  factor_names <- stringr::str_pad(factor_names, digits + 2, side = "both")

  var_names <- rownames(x)
  if (is.null(var_names)) {
    var_names <- paste0("V", seq_len(nrow(x)))
  }

  max_char <- max(sapply(var_names, nchar))

  if (max_char > 10) {
    vn_nchar <- sapply(var_names, nchar)
    var_names[which(vn_nchar > 10)] <- substr(var_names[which(vn_nchar > 10)] ,
                                                      1, 10)
    max_char <- 10
  }

  var_names <- stringr::str_pad(var_names, max_char, side = "right")

  n_col <- ncol(x)

  # create the string to paste using the crayon package
  temp <- apply(matrix(seq_len(nrow(x)), ncol = 1), 1,
                function(ind, x, cutoff, n_col, vn, digits){
    i <- x[ind,]

    tt <- crayon::blue(vn[ind])
    for (kk in seq_len(n_col)) {
      if (kk <= n_col - 2) {
        if (abs(i[kk]) < cutoff) {
          tt <- c(tt, crayon::silver(.numformat(round(i[kk], digits = digits),
                                                digits = digits)))
        } else if (abs(i[kk]) <= 1) {
          tt <- c(tt, crayon::bold(.numformat(round(i[kk], digits = digits),
                                              digits = digits)))
        } else {
          tt <- c(tt, crayon::red$bold(.numformat(round(i[kk], digits = digits),
                                                  digits = digits)))
        }
      } else {
        if (i[kk] <= 1 & i[kk] >= 0) {
          tt <- c(tt, .numformat(round(i[kk], digits = digits), digits = digits))
        } else {
          tt <- c(tt, crayon::red$bold(.numformat(round(i[kk], digits = digits),
                                                  digits = digits)))
        }
      }
    }
    stringr::str_c(tt, collapse = "\t")
  }, cutoff = cutoff, n_col = n_col, digits = digits, x = x, vn = var_names)

  # add header
  factor_names <- stringr::str_c(factor_names,
                                 collapse = "\t")
  factor_names <- crayon::blue(stringr::str_c( stringr::str_pad(" ", max_char),
                                               "\t", factor_names))

  temp <- stringr::str_c(temp, collapse = "\n")
  temp <- stringr::str_c(factor_names, "\n", temp)

  # warn from Heywood cases
  if (sum(x[, -n_col] > 1) == 1) {
    temp <- paste(temp,
                  crayon::red$bold("\nWarning: Results contain a Heywood case!"),
                  collapse = "\n")
  } else if (sum(x[, -(n_col-1):-n_col] > 1) > 1) {
    temp <- paste(temp, crayon::red$bold("\nWarning: Results contain",
                                         sum(x[, -(n_col-1):-n_col] > 1),
                                         "Heywood cases!"),
                  collapse = "\n")
  }

  temp <- stringr::str_c(temp, "\n")
  # print the results to the console
  cat(temp)
}

#' Sequential Chi Square Model Tests, RMSEA lower bound, and AIC
#'
#' Sequential Chi Square Model Tests (SMT) are a factor retention method where
#' multiple
#' EFAs with increasing numbers of factors are fitted and the number of factors
#' for which the Chi Square value first becomes non-significant is taken as the
#' suggested number of factors.
#' Preacher, Zhang, Kim, & Mels (2013) suggested a similar approach with the
#' lower bound of the 90\% confidence interval of the Root Mean Square Error of
#' Approximation (RMSEA; Browne & Cudeck, 1992; Steiger & Lind, 1980), and with
#' the Akaike Information Criterion (AIC). For the RMSEA, the
#' number of factors for which this lower bound first falls below .05 is the
#' suggested number of factors to retain. For the AIC, it is the number of factors
#' where the AIC is lowest.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#' data is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#'  Default is "pearson".
#'
#' @details
#' As a first step in the procedure, a maximum number of factors to extract is
#' determined for which the model is still over-identified (df > 0).
#'
#' Then, EFAs with increasing numbers of factors from 1 to the maximum number are
#' fitted with maximum likelihood estimation.
#'
#' For the SMT, first the significance of the chi
#' square value for a model with 0 factors is determined. If this value is
#' not significant, 0 factors are suggested to retain. If it is significant,
#' a model with 1 factor is estimated and the significance of its chi square value
#' is determined, and so on, until a non-significant result is obtained. The
#' suggested number of factors is the number of factors for the model where the
#' chi square value first becomes non-significant.
#'
#' Regarding the RMSEA, the suggested number of factors is the number of factors
#' for the model where the lower bound of the 90\% confidence interval of the
#' RMSEA first falls below the .05 threshold.
#'
#' Regarding the AIC, the suggested number of factors is the number of factors
#' for the model with the lowest AIC.
#'
#' In comparison with other prominent factor retention criteria, SMT performed
#' well at determining the number of factors to extract in EFA (Auerswald &
#' Moshagen, 2019). The RMSEA lower bound also performed well at determining the true
#' number of factors, while the AIC performed well at determining the
#' most generalizable model (Preacher, Zhang, Kim, & Mels, 2013).
#'
#' The \code{SMT} function can also be called together with other factor
#' retention criteria in the \code{\link{N_FACTORS}} function.
#'
#' @return A list of class SMT containing
#' \item{nfac_chi}{The number of factors to retain according to the significance
#' of the chi square value.}
#' \item{nfac_RMSEA}{The number of factors to retain according to the RMSEA lower
#' bound}
#' \item{nfac_AIC}{The number of factors to retain according to the AIC}
#' \item{p_null}{The p-value for the null model (zero factors)}
#' \item{ps_chi}{The p-values for EFA models with increasing numbers of factors,
#' starting with 1 factor}
#' \item{RMSEA_LB_null}{The lower bounds of the 90\% confidence interval for the RMSEA
#' for the null model (zero factors).}
#' \item{RMSEA_LBs}{The lower bounds of the 90\% confidence interval for the RMSEA
#' for EFA models with increasing numbers of factors, starting with 1 factor}
#' \item{AIC_null}{The AICs for the null model (zero factors)}
#' \item{AICs}{The AICs for EFA models with increasing numbers of factors,
#' starting with 1 factor}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#' @source Browne, M.W., & Cudeck, R. (1992). Alternative ways of assessing model
#' fit. Sociological Methods and Research, 21, 230–258.
#' @source Preacher, K. J., Zhang G., Kim, C., & Mels, G. (2013). Choosing the
#' Optimal Number of Factors in Exploratory Factor Analysis: A Model Selection
#' Perspective, Multivariate Behavioral Research, 48(1), 28-56,
#' doi:10.108/00273171.2012.710386
#' @source Steiger, J. H., & Lind, J. C. (1980, May). Statistically based tests
#' for the number of common factors. Paper presented at the annual meeting of
#' the Psychometric Society, Iowa City, IA.
#'
#' @seealso Other factor retention criteria: \code{\link{CD}}, \code{\link{EKC}},
#' \code{\link{HULL}}, \code{\link{KGC}}, \code{\link{PARALLEL}}
#'
#' \code{\link{N_FACTORS}} as a wrapper function for this and all the
#' above-mentioned factor retention criteria.
#'
#' @export
#'
#' @examples
#' SMT_base <- SMT(test_models$baseline$cormat, N = 500)
#' SMT_base
#'
SMT <- function(x, N = NA, use = c("pairwise.complete.obs", "all.obs",
                                     "complete.obs", "everything",
                                     "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall")){

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  checkmate::assert_count(N, na.ok = TRUE)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  if (is.na(N)) {

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Argument 'N' was NA. Either provide N or raw data.\n"))

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if(any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }

  # Prepare objects for sequential tests
  max_fac <- .det_max_factors(ncol(R))

  if(max_fac <= 0){
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" The model is either underidentified or just identified with 1 factor already. SMTs cannot be performed. Please provide more indicators.\n"))
  }

  ps <- vector("double", max_fac)
  RMSEA_LB <- vector("double", max_fac)
  AIC <- vector("double", max_fac)
  nfac_chi <- NA
  nfac_RMSEA <- NA
  nfac_AIC <- NA

    # sequentially perform EFAs with 1 to the maximum number of factors
    for (i in seq_len(max_fac)) {

      temp <- suppressWarnings(suppressMessages(EFA(R, n_factors = i,
                                                    method = "ML",
                                                    rotate ="none", N = N)))
      ps[i] <- stats::pchisq(temp$fit_indices$chi, temp$fit_indices$df,
                             lower.tail = FALSE)
      RMSEA_LB[i] <- temp$fit_indices$RMSEA_LB
      AIC[i] <- temp$fit_indices$AIC

    }

  # With which number of factors does the chi square first become
  # non-significant?

  # First check if 0 factors already result in nonsignificant chi square
  p_null <- stats::pchisq(temp$fit_indices$chi_null,
                          temp$fit_indices$df_null, lower.tail = F)

    if(p_null > 0.05){

      nfac_chi <- 0

    } else if(any(ps > 0.05, na.rm = TRUE)) {

      nfac_chi <- which(ps > 0.05)[1]

    } else {

      nfac_chi <- NA

    }

  # Calculate RMSEA (incl. lower bound of 90% CI) and AIC for the null model
  chi_null <- temp$fit_indices$chi_null
  df_null <- temp$fit_indices$df_null

  RMSEA_null <- sqrt(max(0, chi_null - df_null) / (df_null * N - 1))

  p_chi_fun <- function(x, val, df, goal){goal - stats::pchisq(val, df, ncp = x)}

  if (stats::pchisq(chi_null, df = df_null, ncp = 0) >= .95) {
    lambda_l <- stats::uniroot(f = p_chi_fun, interval = c(1e-10, 10000), val = chi_null,
                               df = df_null, goal = .95, extendInt = "upX",
                               maxiter = 100L)$root
  } else {
    lambda_l <- 0
  }

  RMSEA_LB_null <- sqrt(lambda_l / (df_null * N))

  AIC_null <- chi_null - 2 * df_null

  # With which number of factors does the RMSEA lower bound first fall below .05?
  if(RMSEA_LB_null < .05){

    nfac_RMSEA <- 0

  } else {

    if(any(RMSEA_LB < .05, na.rm = TRUE)){

      nfac_RMSEA <- which(RMSEA_LB < .05)[1]

    } else {

      nfac_RMSEA <- NA

    }

  }

  # With which number of factors is the AIC lowest?
  AIC_all <- c(AIC_null, AIC)
  nfac_AIC <- which(AIC_all == min(AIC_all)) - 1

  # Prepare the output
  output <- list(nfac_chi = nfac_chi,
                 nfac_RMSEA = nfac_RMSEA,
                 nfac_AIC = nfac_AIC,
                 p_null = p_null,
                 ps_chi = ps,
                 RMSEA_LB_null = RMSEA_LB_null,
                 RMSEA_LBs = RMSEA_LB,
                 AIC_null = AIC_null,
                 AICs = AIC,
                 settings = list(N = N,
                                 use = use,
                                 cor_method = cor_method))

  class(output) <- "SMT"

  return(output)

}

smt_cor <- SMT(test_models$baseline$cormat, N = 500)
smt_zero <- SMT(diag(nrow = 5, ncol = 5), N = 500)
smt_raw <- SMT(GRiPS_raw)

test_that("output class and dimensions are correct", {
  expect_is(smt_cor, "SMT")
  expect_output(str(smt_cor), "List of 10")
  expect_is(smt_raw, "SMT")
  expect_output(str(smt_raw), "List of 10")
  expect_is(smt_zero, "SMT")
  expect_output(str(smt_zero), "List of 10")
})

test_that("number of factors are correct", {
  expect_equal(smt_cor$nfac_chi, 3)
  expect_equal(smt_cor$nfac_RMSEA, 2)
  expect_equal(smt_cor$nfac_AIC, 3)

  expect_equal(smt_raw$nfac_chi, 3)
  expect_equal(smt_raw$nfac_RMSEA, 1)
  expect_equal(smt_raw$nfac_AIC, 3)

  expect_equal(smt_zero$nfac_chi, 0)
  expect_equal(smt_zero$nfac_RMSEA, 0)
  expect_equal(smt_zero$nfac_AIC, 0)
})

test_that("p-values are correct", {
  expect_lt(smt_cor$p_null, 0.05)
  expect_lt(smt_raw$p_null, 0.05)
  expect_gte(smt_zero$p_null, 0.05)

  expect_lt(smt_cor$ps_chi[1],  0.05)
  expect_lt(smt_cor$ps_chi[2],  0.05)
  expect_gte(smt_cor$ps_chi[3],  0.05)
  expect_gte(smt_cor$ps_chi[4],  0.05)

  expect_lt(smt_raw$ps_chi[1],  0.05)
  expect_lt(smt_raw$ps_chi[2],  0.05)
  expect_gte(smt_raw$ps_chi[3],  0.05)
  expect_gte(smt_raw$ps_chi[4],  0.05)

  expect_gte(smt_zero$ps_chi[1], 0.05)
})

test_that("RMSEA_LB and AIC values are correct", {
  expect_equal(smt_cor$RMSEA_LB_null, 0.264456, tolerance = 1e-4)
  expect_equal(smt_raw$RMSEA_LB_null, 0.662764, tolerance = 1e-4)
  expect_equal(smt_zero$RMSEA_LB_null, 0, tolerance = 1e-4)

  expect_equal(smt_cor$RMSEA_LBs, c(0.05674033, 0.03975791, rep(0, 10)),
               tolerance = 1e-4)
  expect_equal(smt_raw$RMSEA_LBs, c(0.03547387, 0.02637614, rep(0, 2)),
               tolerance = 1e-4)
  expect_equal(smt_zero$RMSEA_LBs, rep(0, 2), tolerance = 1e-4)

  expect_equal(smt_cor$AIC_null, 5441.203, tolerance = 0.1)
  expect_equal(smt_raw$AIC_null, 10264.07, tolerance = 0.1)
  expect_equal(smt_zero$AIC_null, -20, tolerance = 0.1)

  expect_equal(smt_cor$AICs, c(139.36142, 17.28713, -78.02180, -77.23762,
                               -74.12983, -66.95662, -56.34764, -45.65766,
                               -35.95682, -25.891317, -16.952594, -5.103373),
               tolerance = 0.1)
  expect_equal(smt_raw$AICs, c(19.919177, 7.833463, -4.447209, -1.852061),
               tolerance = 0.1)
  expect_equal(smt_zero$AICs, c(-10, -2), tolerance = 0.1)
})

test_that("settings are returned correctly", {
  expect_named(smt_cor$settings, c("N", "use", "cor_method"))
  expect_named(smt_raw$settings, c("N", "use", "cor_method"))
  expect_named(smt_zero$settings, c("N", "use", "cor_method"))

  expect_equal(smt_cor$settings$N, 500)
  expect_equal(smt_raw$settings$N, 810)
  expect_equal(smt_zero$settings$N, 500)

  expect_equal(smt_cor$settings$use, "pairwise.complete.obs")
  expect_equal(smt_raw$settings$use, "pairwise.complete.obs")
  expect_equal(smt_zero$settings$use, "pairwise.complete.obs")

  expect_equal(smt_cor$settings$cor_method, "pearson")
  expect_equal(smt_raw$settings$cor_method, "pearson")
  expect_equal(smt_zero$settings$cor_method, "pearson")

})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z), ncol = 3)
cor_sing <- stats::cor(dat_sing)

burt <- matrix(c(1.00,  0.83,  0.81,  0.80,   0.71, 0.70, 0.54, 0.53,  0.59,  0.24, 0.13,
                 0.83,  1.00,  0.87,  0.62,   0.59, 0.44, 0.58, 0.44,  0.23,  0.45,  0.21,
                 0.81,  0.87,  1.00,  0.63,   0.37, 0.31, 0.30, 0.12,  0.33,  0.33,  0.36,
                 0.80,  0.62,  0.63,  1.00,   0.49, 0.54, 0.30, 0.28,  0.42,  0.29, -0.06,
                 0.71,  0.59,  0.37,  0.49,   1.00, 0.54, 0.34, 0.55,  0.40,  0.19, -0.10,
                 0.70,  0.44,  0.31,  0.54,   0.54, 1.00, 0.50, 0.51,  0.31,  0.11,  0.10,
                 0.54,  0.58,  0.30,  0.30,   0.34, 0.50, 1.00, 0.38,  0.29,  0.21,  0.08,
                 0.53,  0.44,  0.12,  0.28,   0.55, 0.51, 0.38, 1.00,  0.53,  0.10, -0.16,
                 0.59,  0.23,  0.33,  0.42,   0.40, 0.31, 0.29, 0.53,  1.00, -0.09, -0.10,
                 0.24,  0.45,  0.33,  0.29,   0.19, 0.11, 0.21, 0.10, -0.09,  1.00,  0.41,
                 0.13,  0.21,  0.36, -0.06,  -0.10, 0.10, 0.08, -0.16, -0.10, 0.41,  1.00),
               nrow = 11, ncol = 11)

test_that("errors are thrown correctly", {
  expect_error(SMT(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_error(SMT(test_models$baseline$cormat), " Argument 'N' was NA. Either provide N or raw data.\n")
  expect_message(SMT(GRiPS_raw), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_warning(SMT(GRiPS_raw, N = 20), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(SMT(dat_sing), " Correlation matrix is singular, no further analyses are performed\n")
  expect_error(SMT(cor_sing, N = 10), " Correlation matrix is singular, no further analyses are performed\n")
  expect_error(SMT(matrix(rnorm(50), ncol = 2)), " The model is either underidentified or just identified with 1 factor already. SMTs cannot be performed. Please provide more indicators.\n") # underidentified case
  expect_error(SMT(matrix(rnorm(60), ncol = 3)), " The model is either underidentified or just identified with 1 factor already. SMTs cannot be performed. Please provide more indicators.\n") # just identified case
  # expect_warning(SMT(burt, N = 170), "Matrix was not positive definite, smoothing was done")
})

rm(smt_cor, smt_raw, smt_zero, x, y, z, dat_sing, cor_sing, burt)

#' Plot EKC object
#'
#' Plot method showing a summarized output of the \link{EKC} function
#'
#' @param x a list of class EKC. An output from the \link{EKC} function.
#' @param ... not used.
#'
#' @export
#' @method plot EKC
#'
#' @examples
#' EKC_base <- EKC(test_models$baseline$cormat, N = 500)
#' plot(EKC_base)
#'
plot.EKC <- function(x, ...) {

  ekc_type <- x$settings$type

  if("BvA2017" %in% ekc_type){

    .plot_EKC_helper(eigvls = x$eigenvalues,
                     references = x$references$BvA2017,
                     n_factors = x$n_factors_BvA2017,
                     type = "BvA2017")

  }

  if("AM2019" %in% ekc_type){

    .plot_EKC_helper(eigvls = x$eigenvalues,
                     references = x$references$AM2019,
                     n_factors = x$n_factors_AM2019,
                     type = "AM2019")

  }

}



.plot_EKC_helper <- function(eigvls, references, n_factors, type){
  # eigen = eigenvalues found with specific type (PCA, SMC or EFA)
  # references = reference eigenvalues
  # n_factors = number of factors suggested by EKC
  # type = EKC type used

  p_eigvls <- pretty(eigvls)
  x_len <- length(eigvls)

  graphics::plot.new()
  graphics::plot.window(xlim = c(1, x_len),
                        ylim = c(min(p_eigvls),
                                 max(p_eigvls)))
  graphics::axis(1, seq_len(x_len))
  graphics::axis(2, p_eigvls, las = 1)

  graphics::mtext("Factor", side = 1, line = 3, cex = 1.5, padj =-.5)
  graphics::mtext("Eigenvalues", side = 2, line = 3, cex = 1.5, padj =.5)

  graphics::lines(seq_len(x_len), eigvls)
  graphics::points(seq_len(x_len), eigvls, pch = 16)

  graphics::lines(seq_len(x_len), references, lty = 2, lwd = 1.25,
                  col = "darkgray")

  if (!is.na(n_factors)) {
    graphics::points(n_factors, eigvls[n_factors],
                     pch = 1, cex = 2, col = "red")
    graphics::text(n_factors, eigvls[n_factors],
                   n_factors, pos = 3, cex = 1.5, col = "red",
                   font = 1, offset = .75)
  }

  factors_text <- paste0("N factors suggested by EKC (type '", type, "'): ",
                         n_factors)

  graphics::title(factors_text, cex.main = 1.3)

}

#' Exploratory factor analysis on multiple data imputations
#'
#' @author Andreas Soteriades, Markus Steiner
#'
#' @description
#' Fits \code{\link{EFA}} to each of several imputed datasets, aligns the
#' factor solutions to a common factor space, and pools the resulting estimates
#' and selected fit quantities across imputations.
#'
#' @details
#' The function first fits the same \code{\link{EFA}} model to each imputed
#' dataset. Unrotated loading matrices can optionally be put into a common
#' signed/permuted factor order before averaging. Rotated loading matrices are
#' aligned with either a consensus Procrustes target or with the first
#' imputation's rotated solution as a fixed target. For oblique solutions,
#' factor intercorrelations are transformed/aligned together with the loading
#' matrices so that the factor model remains internally consistent.
#'
#' Point estimates are pooled by arithmetic averaging after alignment. For
#' oblique rotations, the returned structure matrix is computed from the pooled
#' aligned pattern matrix and the pooled factor correlation matrix,
#' \eqn{Structure = \Lambda \Phi}. Communalities are always computed as the
#' diagonal of the common-factor reproduced correlation matrix,
#' \eqn{diag(\Lambda \Phi \Lambda')} for oblique rotations and
#' \eqn{diag(\Lambda \Lambda')} otherwise.
#'
#' Residuals are not averaged from the per-imputation residual matrices. Instead,
#' the observed correlation matrices are averaged across imputations and residuals
#' are calculated from the pooled observed correlation matrix minus the
#' model-implied correlation matrix of the pooled solution. Consequently,
#' residual-based fit indices such as RMSR/SRMR are based on pooled residuals.
#'
#' Fit indices based on the model chi-square are not arithmetic means of the
#' per-imputation fit indices. If possible, chi-square-type fit is pooled with
#' the D2 rule. For CFI, the null-model chi-square is D2-pooled as well when
#' complete-data null-model chi-squares are available. The asymptotic
#' chi-square approximation to D2 is then used for RMSEA and CFI. AIC and BIC,
#' if returned, are chi-square-derived descriptive quantities based on this D2
#' approximation and should not be interpreted as likelihood-based MI
#' information criteria. D3/D4 pooling is not implemented here because the
#' current \code{EFA} objects do not expose the likelihood quantities needed
#' for those methods.
#'
#' If each component \code{EFA} call was run with \code{se = "np-boot"} and
#' returned \code{boot.arrays}, pooled bootstrap SEs and Wald-type MI confidence
#' intervals are computed for loadings, communalities, residuals, and, when
#' applicable, factor correlations and structure coefficients. Importantly, the
#' rotated bootstrap loading matrices stored by the component \code{EFA} calls
#' are not reused directly, because they were aligned to imputation-specific
#' targets. Instead, the unrotated bootstrap loading matrices are re-aligned to
#' the final MI target before estimating within-imputation bootstrap covariance
#' matrices. Rubin-type MI pooling is then applied with
#' \eqn{T = Ubar + (1 + 1 / m) B}. Confidence intervals for loadings, Phi,
#' communalities, residuals, and structure coefficients are Wald-type MI
#' intervals. The confidence level of these pooled intervals is controlled by
#' \code{p}; the \code{ci} argument passed through \code{...} to the component
#' \code{EFA} calls is not used for the pooled intervals.
#' \code{boot.CI$fit_indices_pooled_algorithm}, when available, is a
#' percentile-style summary obtained by re-running the pooled-fit algorithm over
#' matched bootstrap replicate indices.
#'
#' @param data_list A list of length \eqn{m}, where \eqn{m} is the number of
#' imputations. Each list element is a data frame or matrix of raw data, or a
#' correlation matrix. See argument \code{x} in \code{\link{EFA}}.
#' @param p Numeric in \eqn{(0, 1)}. One minus the confidence level used for
#' pooled Wald-type bootstrap/MI confidence intervals when bootstrap arrays are
#' available. For example, \code{p = .05} gives 95\% intervals.
#' @param target_method Character. \code{"consensus"} aligns all solutions to an
#' iteratively updated consensus target. \code{"first_target"} aligns all
#' solutions to the first imputation's rotated solution.
#' @param align_unrotated Character. How to align unrotated loadings before
#' pooling. \code{"signed_tucker_congruence"} preserves the unrotated axes up
#' to factor reordering and sign changes using Tucker congruence.
#' \code{"procrustes"} aligns the unrotated matrices to the first imputation by
#' orthogonal Procrustes rotation. \code{"none"} averages unrotated loadings as
#' returned by \code{EFA}.
#' @param fit_pool_method Character. Currently only \code{"D2"} is implemented
#' for chi-square-type fit. If no chi-square is available, only residual-based
#' fit and descriptive quantities are returned.
#' @param consensus_args List of additional arguments passed to
#' \code{CONSENSUS_PROCRUSTES}.
#' @param procrustes_args List of additional arguments passed to \code{PROCRUSTES}
#' for fixed-target alignment.
#' @param rmsea_ci_level Numeric. Confidence level for the RMSEA CI.
#' @param rmsr_upper Logical. If \code{TRUE}, compute RMSR from the unique
#' off-diagonal residual correlations. If \code{FALSE}, use the full off-diagonal
#' matrix.
#' @param ... Additional arguments passed to \code{\link{EFA}}.
#'
#' @return A list of class \code{"EFA_POOLED"} containing pooled estimates,
#' residuals, fit indices, the individual fits, and MI diagnostics.
#'
#' @export
#'
#' @examples
#'
#' # create a list of three datasets, mimicking a list you would obtain from
#' # e.g. mice.
#' dat_list <- lapply(1:3, function(x) GRiPS_raw[sample(1:nrow(GRiPS_raw), replace = TRUE),])
#' mod <- EFA_POOLED(dat_list, n_factors = 1, method = "ML")
#' mod
#'
#' \donttest{
#' # add computation of standard errors and CIs
#' mod <- EFA_POOLED(dat_list, n_factors = 1, method = "ML", se = "np-boot")
#' mod
#' }
EFA_POOLED <- function(data_list,
                       p = 0.05,
                       target_method = c("consensus", "first_target"),
                       align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
                       fit_pool_method = c("D2"),
                       consensus_args = list(),
                       procrustes_args = list(),
                       rmsea_ci_level = .90,
                       rmsr_upper = TRUE,
                       ...) {

  efa_args <- list(...)

  checkmate::assert_list(data_list, min.len = 2, null.ok = FALSE)
  lapply(data_list, checkmate::assert_multi_class, c("matrix", "data.frame"))
  checkmate::assert_number(p, na.ok = FALSE, lower = 0, upper = 1)
  checkmate::assert_list(consensus_args, null.ok = FALSE)
  checkmate::assert_list(procrustes_args, null.ok = FALSE)
  checkmate::assert_number(rmsea_ci_level, na.ok = FALSE, lower = 0, upper = 1)
  checkmate::assert_flag(rmsr_upper)

  if (p <= 0 || p >= 1) {
    stop("`p` must be strictly between 0 and 1.", call. = FALSE)
  }
  if (rmsea_ci_level <= 0 || rmsea_ci_level >= 1) {
    stop("`rmsea_ci_level` must be strictly between 0 and 1.", call. = FALSE)
  }
  if (!is.null(efa_args$ci) && length(efa_args$ci) == 1L &&
      is.finite(efa_args$ci) &&
      abs(as.numeric(efa_args$ci) - (1 - p)) > sqrt(.Machine$double.eps)) {
    warning(
      "`EFA_POOLED()` uses `p`, not the component `EFA()` argument `ci`, ",
      "to set pooled bootstrap/MI confidence intervals.",
      call. = FALSE
    )
  }

  target_method <- match.arg(target_method)
  align_unrotated <- match.arg(align_unrotated)
  fit_pool_method <- match.arg(fit_pool_method)

  m_imp <- length(data_list)

  ## -------------------------------------------------------------------------
  ## Fit EFA to each imputed dataset
  ## -------------------------------------------------------------------------

  fits <- lapply(data_list, function(data_list_subset) {
    do.call(EFA, c(list(x = data_list_subset), efa_args))
  })

  .efa_pooled_check_fits(fits)

  settings <- fits[[1]]$settings
  method <- settings$method
  rotation <- settings$rotation
  var_names <- rownames(fits[[1]]$orig_R)
  if (is.null(var_names)) {
    var_names <- colnames(data_list[[1]])
  }

  if (is.null(rotation) || identical(rotation, "none")) {
    rotation_type <- "none"
  } else if (rotation %in% c("varimax", "quartimax", "equamax", "bentlerT",
                             "geominT", "bifactorT")) {
    rotation_type <- "orthogonal"
  } else if (rotation %in% c("promax", "oblimin", "quartimin", "simplimax",
                             "bentlerQ", "geominQ", "bifactorQ")) {
    rotation_type <- "oblique"
  } else {
    stop("Unknown rotation type: ", rotation, call. = FALSE)
  }

  ## -------------------------------------------------------------------------
  ## Extract and align unrotated loadings
  ## -------------------------------------------------------------------------

  unrot_loadings <- lapply(.extract_list_object(fits, "unrot_loadings"),
                           .change_class, "matrix")

  unrot_loadings_aligned <- .efa_pooled_align_unrotated_list(
    unrot_loadings = unrot_loadings,
    align_unrotated = align_unrotated
  )

  mean_unrot_loadings <- .average_matrices(unrot_loadings_aligned)

  ## -------------------------------------------------------------------------
  ## Align rotated loadings and Phi, if a rotation was requested
  ## -------------------------------------------------------------------------

  phis <- NULL
  structure_loadings <- NULL

  if (rotation_type != "none") {
    rot_loadings_initial <- lapply(.extract_list_object(fits, "rot_loadings"),
                                   .change_class, "matrix")

    if (target_method == "first_target") {
      target_rotations <- vector("list", m_imp)
      rot_loadings <- vector("list", m_imp)
      phis <- vector("list", m_imp)

      rot_loadings[[1]] <- rot_loadings_initial[[1]]
      if (rotation_type == "oblique") {
        phis[[1]] <- fits[[1]]$Phi
      }

      point_rotation_failures <- logical(m_imp)
      for (d in 2:m_imp) {
        target_rotations[[d]] <- do.call(
          PROCRUSTES,
          c(list(A = unrot_loadings[[d]],
                 Target = rot_loadings_initial[[1]],
                 rotation = rotation_type),
            procrustes_args)
        )
        point_rotation_failures[d] <- !is.null(target_rotations[[d]]$convergence) &&
          isFALSE(target_rotations[[d]]$convergence)
        rot_loadings[[d]] <- target_rotations[[d]]$loadings
        if (rotation_type == "oblique") {
          phis[[d]] <- target_rotations[[d]]$Phi
        }
      }

      if (any(point_rotation_failures)) {
        warning(
          "At least one fixed-target Procrustes alignment did not converge. ",
          "The returned pooled point estimates still use the best available ",
          "alignment from PROCRUSTES(). Inspect `alignment$point_rotation_failures`.",
          call. = FALSE
        )
      }

      final_target <- rot_loadings_initial[[1]]
      alignment <- list(method = "first_target",
                        target = final_target,
                        target_rotations = target_rotations,
                        point_rotation_failures = which(point_rotation_failures))

    } else if (target_method == "consensus") {
      consensus <- do.call(
        CONSENSUS_PROCRUSTES,
        c(list(unrotated_list = unrot_loadings,
               init_targets = rot_loadings_initial,
               rotation = rotation_type),
          consensus_args)
      )

      if (!isTRUE(consensus$converged)) {
        warning(
          "Consensus Procrustes alignment did not meet its convergence criterion. ",
          "Inspect `alignment$history` and consider stricter or multi-start ",
          "`consensus_args`.",
          call. = FALSE
        )
      }

      rot_loadings <- consensus$aligned_loadings
      phis <- consensus$aligned_phi
      final_target <- consensus$target
      alignment <- consensus
    }

    mean_rot_loadings <- .average_matrices(rot_loadings)
    rownames(mean_rot_loadings) <- var_names

    if (rotation_type == "oblique") {
      mean_phis <- .average_matrices(phis)
      # Ensure mean_phis is symmetric (needed, as due to floating point imprecision,
      # the averaged matrix can be asymmetric, leading to issues in downstream checks
      # of matrix symmetry)
      mean_phis <- (mean_phis + t(mean_phis)) / 2
      structure_loadings <- Map(function(L, Phi) L %*% Phi, rot_loadings, phis)
      # Keep Structure parallel to the returned pooled pattern matrix and Phi:
      # it is the plug-in structure of the pooled solution, not the arithmetic
      # mean of the imputation-specific structure matrices.
      mean_structure_loadings <- mean_rot_loadings %*% mean_phis
      dimnames(mean_structure_loadings) <- dimnames(mean_rot_loadings)
    } else {
      mean_phis <- NULL
      mean_structure_loadings <- NULL
    }

  } else {
    rot_loadings <- NULL
    mean_rot_loadings <- NULL
    mean_phis <- NULL
    mean_structure_loadings <- NULL
    final_target <- NULL
    alignment <- NULL
  }

  ## -------------------------------------------------------------------------
  ## Pooled observed and model-implied correlation matrices; residuals and RMSR
  ## -------------------------------------------------------------------------

  orig_R_list <- .extract_list_object(fits, "orig_R")

  pooled_orig_R <- .average_matrices(orig_R_list)

  if (rotation_type == "oblique") {
    common_R <- mean_rot_loadings %*% mean_phis %*% t(mean_rot_loadings)
  } else if (rotation_type == "orthogonal") {
    common_R <- mean_rot_loadings %*% t(mean_rot_loadings)
  } else {
    common_R <- mean_unrot_loadings %*% t(mean_unrot_loadings)
  }

  h2 <- diag(common_R)
  names(h2) <- rownames(common_R)

  model_implied_R <- common_R
  diag(model_implied_R) <- 1
  dimnames(model_implied_R) <- list(var_names, var_names)

  residuals <- pooled_orig_R - model_implied_R
  diag(residuals) <- 0
  dimnames(residuals) <- list(var_names, var_names)

  RMSR <- .rmsr(residuals, upper = rmsr_upper)

  ## -------------------------------------------------------------------------
  ## Variance-accounted tables
  ## -------------------------------------------------------------------------

  mean_vars_accounted <- .compute_vars(
    L_unrot = mean_unrot_loadings,
    L_rot = mean_unrot_loadings,
    Phi = NULL
  )

  if (rotation_type == "oblique") {
    mean_vars_accounted_rot <- .compute_vars(
      L_unrot = mean_unrot_loadings,
      L_rot = mean_rot_loadings,
      Phi = mean_phis
    )
  } else if (rotation_type == "orthogonal") {
    mean_vars_accounted_rot <- .compute_vars(
      L_unrot = mean_unrot_loadings,
      L_rot = mean_rot_loadings,
      Phi = NULL
    )
  } else {
    mean_vars_accounted_rot <- NULL
  }

  ## -------------------------------------------------------------------------
  ## Fit indices: residual-based from pooled residuals, chi-square-based via D2
  ## -------------------------------------------------------------------------

  Ns <- .efa_pooled_get_Ns(data_list, fits, efa_args)
  Ns_ok <- Ns[is.finite(Ns)]
  if (length(Ns_ok) == 0L) {
    N_pool <- NA_real_
    if (!identical(method, "PAF")) {
      warning(
        "N could not be recovered for any imputation. ",
        "Chi-square-based fit indices cannot be computed.",
        call. = FALSE
      )
    }
  } else {
    if (length(Ns_ok) < length(Ns)) {
      warning(
        "N could not be recovered for every imputation. ",
        "Fit indices use the mean of the available Ns.",
        call. = FALSE
      )
    }
    if (length(unique(Ns_ok)) > 1L) {
      warning(
        "The imputed datasets appear to have different N. ",
        "Fit indices use the mean N across imputations.",
        call. = FALSE
      )
    }
    N_pool <- mean(Ns_ok)
  }

  fit_indices <- .efa_pooled_fit_indices(
    fits = fits,
    pooled_R = pooled_orig_R,
    residuals = residuals,
    RMSR = RMSR,
    N = N_pool,
    method = method,
    pool_method = fit_pool_method,
    rmsea_ci_level = rmsea_ci_level
  )

  ## -------------------------------------------------------------------------
  ## Bootstrap SEs and CIs for pooled MI estimates, if available
  ## -------------------------------------------------------------------------

  boot_pooled <- .efa_pooled_bootstrap_pool(
    fits = fits,
    orig_R_list = orig_R_list,
    unrot_loadings_aligned = unrot_loadings_aligned,
    mean_unrot_loadings = mean_unrot_loadings,
    rot_loadings = rot_loadings,
    phis = phis,
    structure_loadings = structure_loadings,
    mean_structure_loadings = mean_structure_loadings,
    final_target = final_target,
    rotation_type = rotation_type,
    align_unrotated = align_unrotated,
    procrustes_args = procrustes_args,
    h2 = h2,
    residuals = residuals,
    fit_indices = fit_indices,
    pooled_orig_R = pooled_orig_R,
    N = N_pool,
    method = method,
    pool_method = fit_pool_method,
    rmsea_ci_level = rmsea_ci_level,
    alpha = p,
    rmsr_upper = rmsr_upper
  )

  ## -------------------------------------------------------------------------
  ## Return object
  ## -------------------------------------------------------------------------

  mean_unrot_loadings <- .change_class(mean_unrot_loadings, "LOADINGS")
  if (!is.null(mean_rot_loadings)) {
    mean_rot_loadings <- .change_class(mean_rot_loadings, "LOADINGS")
  }
  if (!is.null(mean_structure_loadings)) {
    mean_structure_loadings <- .change_class(mean_structure_loadings, "LOADINGS")
  }

  settings_pooled <- settings
  settings_pooled$N <- N_pool
  settings_pooled$pooled_N <- N_pool
  settings_pooled$pooled <- TRUE
  settings_pooled$component_se <- settings$se
  if (is.null(boot_pooled)) {
    # Avoid a misleading bootstrap note in print.EFA_POOLED() when component EFA
    # calls requested bootstrap SEs but pooled bootstrap/MI SEs could not be
    # computed (e.g., missing arrays or too few bootstrap replicates).
    settings_pooled$se <- "none"
  }
  settings_pooled$n_imputations <- m_imp
  settings_pooled$target_method <- target_method
  settings_pooled$align_unrotated <- align_unrotated
  settings_pooled$fit_pool_method <- fit_pool_method
  settings_pooled$p <- p
  settings_pooled$ci <- 1 - p
  settings_pooled$rmsea_ci_level <- rmsea_ci_level
  settings_pooled$rmsr_upper <- rmsr_upper
  if (!is.null(boot_pooled) && !is.null(boot_pooled$n_boot)) {
    settings_pooled$b_boot <- boot_pooled$n_boot
  }

  results <- list(
    h2 = h2,
    unrot_loadings = mean_unrot_loadings,
    vars_accounted = mean_vars_accounted,
    fit_indices = fit_indices,
    model_implied_R = model_implied_R,
    residuals = residuals,
    orig_R = pooled_orig_R,
    settings = settings_pooled,
    fits = fits,
    alignment = alignment,
    mi_diagnostics = fit_indices$mi_diagnostics
  )

  if (rotation_type != "none") {
    results$rot_loadings <- mean_rot_loadings
    results$vars_accounted_rot <- mean_vars_accounted_rot
  }

  if (rotation_type == "oblique") {
    results$Phi <- mean_phis
    results$Structure <- mean_structure_loadings
  }

  if (!is.null(boot_pooled)) {
    results$boot.SE <- boot_pooled$SE
    results$boot.CI <- boot_pooled$CI
    results$boot.arrays <- boot_pooled$arrays
    results$boot.MI <- boot_pooled$MI
    results$standardized_residuals <- results$residuals / boot_pooled$SE$residuals
  }

  class(results) <- c("EFA_POOLED", "EFA")
  results
}

## =============================================================================
## Internal helpers for EFA_POOLED
## =============================================================================

.efa_pooled_get_Ns <- function(data_list, fits, efa_args) {
  # Recover the N used in each EFA fit. Correlation-matrix input may not carry N,
  # so return NA unless N was supplied to EFA() or stored in settings.
  vapply(seq_along(data_list), function(d) {
    if (!is.null(fits[[d]]$settings$N) && !is.na(fits[[d]]$settings$N)) {
      return(as.numeric(fits[[d]]$settings$N))
    }
    if (!is.null(efa_args$N) && !is.na(efa_args$N)) {
      return(as.numeric(efa_args$N))
    }
    if (!.is_cormat(data_list[[d]])) {
      return(nrow(data_list[[d]]))
    }
    NA_real_
  }, numeric(1))
}

.efa_pooled_check_fits <- function(fits) {
  # Fail early if the fitted EFA objects are not conformable. Pooling only makes
  # sense when all imputations estimate the same model on the same variables.
  if (length(fits) < 2L) {
    stop("At least two EFA fits are required for MI pooling.", call. = FALSE)
  }

  dims <- vapply(fits, function(x) {
    paste(dim(as.matrix(x$unrot_loadings)), collapse = "x")
  }, character(1))
  if (length(unique(dims)) != 1L) {
    stop("All unrotated loading matrices must have the same dimensions.", call. = FALSE)
  }

  var_names <- lapply(fits, function(x) rownames(as.matrix(x$orig_R)))
  if (!all(vapply(var_names[-1], identical, logical(1), var_names[[1]]))) {
    stop("All imputations must contain the same variables in the same order.", call. = FALSE)
  }

  setting_value <- function(x, name) {
    val <- x$settings[[name]]
    if (is.null(val)) NA_character_ else as.character(val)
  }

  for (nm in c("method", "rotation", "n_factors")) {
    vals <- vapply(fits, setting_value, character(1), name = nm)
    vals <- vals[!is.na(vals)]
    if (length(unique(vals)) > 1L) {
      stop("All imputations must use the same ", nm, ".", call. = FALSE)
    }
  }

  invisible(TRUE)
}

.efa_pooled_align_unrotated_list <- function(unrot_loadings,
                                             align_unrotated = c("signed_tucker_congruence", "none", "procrustes")) {
  # Unrotated factor axes are arbitrary up to ordering and signs. This helper
  # puts them into a common orientation before simple arithmetic averaging.
  # Unlike the rotated solution below, this step should not seek simple
  # structure; it only removes indeterminacy in the unrotated axes.
  align_unrotated <- match.arg(align_unrotated)

  if (align_unrotated == "none") {
    return(unrot_loadings)
  }

  target <- as.matrix(unrot_loadings[[1]])
  out <- vector("list", length(unrot_loadings))
  out[[1]] <- target

  for (d in seq_along(unrot_loadings)[-1]) {
    Ld <- as.matrix(unrot_loadings[[d]])

    if (align_unrotated == "signed_tucker_congruence") {
      # Preserve the axes up to a one-to-one permutation/sign change based on
      # Tucker congruence. .align_solution() implements the assignment and sign
      # correction without applying a continuous rotation.
      out[[d]] <- .align_solution(L_target = target, L = Ld)$loadings
    } else if (align_unrotated == "procrustes") {
      # Optional continuous orthogonal alignment for users who prefer a least-
      # squares Procrustes match for the unrotated matrices.
      out[[d]] <- .change_class(PROCRUSTES(
        A = Ld,
        Target = target,
        rotation = "orthogonal"
      )$loadings, "matrix")
    }
  }

  out
}

.efa_pooled_rmsea_ci <- function(chi, df, N, level = .90) {
  # RMSEA CI from the noncentral chi-square inversion used by EFAtools.
  # The chi-square supplied here should already be the pooled test statistic.
  if (is.na(chi) || is.na(df) || is.na(N) || df <= 0 || N <= 1) {
    return(c(lower = NA_real_, upper = NA_real_))
  }

  alpha <- 1 - level
  lower_goal <- 1 - alpha / 2
  upper_goal <- alpha / 2

  pfun <- function(lambda, goal) {
    goal - stats::pchisq(chi, df = df, ncp = lambda)
  }

  lambda_l <- 0
  lambda_u <- 0

  if (stats::pchisq(chi, df = df, ncp = 0) >= lower_goal) {
    lambda_l <- tryCatch(
      stats::uniroot(pfun, interval = c(1e-10, 10000), goal = lower_goal,
                     extendInt = "upX", maxiter = 100L)$root,
      error = function(e) NA_real_
    )
  }

  if (stats::pchisq(chi, df = df, ncp = 0) >= upper_goal) {
    lambda_u <- tryCatch(
      stats::uniroot(pfun, interval = c(1e-10, 10000), goal = upper_goal,
                     extendInt = "upX", maxiter = 100L)$root,
      error = function(e) NA_real_
    )
  }

  denom <- df * (N - 1)
  if (!is.finite(denom) || denom <= 0) {
    return(c(lower = NA_real_, upper = NA_real_))
  }
  c(lower = sqrt(lambda_l / denom), upper = sqrt(lambda_u / denom))
}

.efa_pooled_D2 <- function(chis, df) {
  # D2 pools complete-data chi-square statistics across imputations. The
  # returned chi value is the asymptotic chi-square approximation df * F_D2,
  # which is used downstream for RMSEA/CFI.
  chis <- chis[is.finite(chis)]
  M <- length(chis)

  if (M < 2L || !is.finite(df) || df <= 0) {
    return(NULL)
  }

  Tbar <- mean(chis)

  if (Tbar <= 0) {
    return(list(
      F = 0,
      chi = 0,
      df1 = df,
      df2 = Inf,
      p = 1,
      ARIV = 0,
      FMI = 0,
      Tbar = Tbar,
      M = M
    ))
  }

  ARIV <- (1 + 1 / M) * sum((sqrt(chis) - sqrt(Tbar))^2) / (M - 1)

  if (ARIV <= .Machine$double.eps) {
    F_D2 <- Tbar / df
    df2 <- Inf
    p_val <- stats::pchisq(Tbar, df = df, lower.tail = FALSE)
  } else {
    F_D2 <- (Tbar / df - ((M + 1) / (M - 1)) * ARIV) / (1 + ARIV)
    df2 <- df^(-3 / M) * (M - 1) * (1 + 1 / ARIV)^2
    p_val <- if (F_D2 < 0) 1 else stats::pf(F_D2, df1 = df, df2 = df2, lower.tail = FALSE)
  }

  chi_D2 <- max(0, df * F_D2)
  FMI <- ARIV / (1 + ARIV)

  list(
    F = F_D2,
    chi = chi_D2,
    df1 = df,
    df2 = df2,
    p = p_val,
    ARIV = ARIV,
    FMI = FMI,
    Tbar = Tbar,
    M = M
  )
}

.efa_pooled_fit_indices <- function(fits,
                                    pooled_R,
                                    residuals,
                                    RMSR,
                                    N,
                                    method,
                                    pool_method = "D2",
                                    rmsea_ci_level = .90) {

  fit_list <- .extract_list_object(fits, "fit_indices")
  fit_list <- fit_list[!vapply(fit_list, is.null, logical(1))]

  p_vars <- nrow(pooled_R)
  df <- NA_real_
  chis <- numeric(0)
  chis_null <- numeric(0)

  if (length(fit_list) > 0L) {
    dfs <- vapply(fit_list, function(x) {
      if (!is.null(x$df)) x$df else NA_real_
    }, numeric(1))
    finite_dfs <- unique(stats::na.omit(dfs))
    if (length(finite_dfs) > 1L) {
      stop("Cannot D2-pool chi-square fit because the imputation-specific dfs differ.",
           call. = FALSE)
    }
    if (length(finite_dfs) == 1L) {
      df <- finite_dfs[[1L]]
    }

    chis <- vapply(fit_list, function(x) {
      if (!is.null(x$chi)) x$chi else NA_real_
    }, numeric(1))
    chis_null <- vapply(fit_list, function(x) {
      if (!is.null(x$chi_null)) x$chi_null else NA_real_
    }, numeric(1))
  }

  D2 <- NULL
  if (identical(pool_method, "D2") && length(chis) > 0L && is.finite(df)) {
    D2 <- .efa_pooled_D2(chis, df)
  }

  if (!is.null(D2)) {
    chi <- D2$chi
    p_chi <- D2$p
    Fm <- if (is.finite(N) && N > 1) chi / (N - 1) else NA_real_
  } else {
    chi <- NA_real_
    p_chi <- NA_real_
    Fm <- NA_real_
  }

  df_null <- p_vars * (p_vars - 1) / 2
  D2_null <- NULL
  if (identical(pool_method, "D2") && length(chis_null) > 0L &&
      is.finite(df_null) && df_null > 0) {
    D2_null <- .efa_pooled_D2(chis_null, df_null)
  }

  if (!is.null(D2_null)) {
    chi_null <- D2_null$chi
    p_null <- D2_null$p
  } else if (is.finite(N) && N > 1) {
    chi_null <- sum(pooled_R[upper.tri(pooled_R)]^2) * (N - 1)
    p_null <- stats::pchisq(chi_null, df_null, lower.tail = FALSE)
  } else {
    chi_null <- NA_real_
    p_null <- NA_real_
  }

  if (is.finite(chi) && is.finite(chi_null) &&
      is.finite(df) && is.finite(df_null) && df >= 0) {
    delta_null <- chi_null - df_null
    delta_model <- chi - df
    if (is.finite(delta_null) && delta_null > 0) {
      CFI <- (delta_null - delta_model) / delta_null
      CFI <- max(0, min(1, CFI))
    } else {
      CFI <- NA_real_
    }
  } else {
    CFI <- NA_real_
  }

  rmsea_denom <- df * (N - 1)
  if (is.finite(chi) && is.finite(df) && is.finite(N) &&
      df > 0 && N > 1 && is.finite(rmsea_denom) && rmsea_denom > 0) {
    RMSEA <- sqrt(max((chi - df) / rmsea_denom, 0))
    rmsea_ci <- .efa_pooled_rmsea_ci(chi, df, N, level = rmsea_ci_level)
  } else {
    RMSEA <- NA_real_
    rmsea_ci <- c(lower = NA_real_, upper = NA_real_)
  }

  # These mirror the chi-square-derived quantities used elsewhere in EFAtools.
  # Under MI/D2 pooling they are descriptive only, not likelihood-based MI AIC/BIC.
  AIC <- if (is.finite(chi) && is.finite(df)) chi - 2 * df else NA_real_
  BIC <- if (is.finite(chi) && is.finite(df) && is.finite(N)) chi - log(N) * df else NA_real_

  ## CAF in EFAtools is 1 - KMO(delta_hat), with diagonal set to 1.
  # Ensure symmetry of residuals by (residuals + t(residuals)) / 2. Asymmetry
  # can arise due to floating point imprecision from averaging the matrices.
  delta_hat <- (residuals + t(residuals)) / 2
  diag(delta_hat) <- 1
  CAF <- tryCatch(1 - KMO(delta_hat)$KMO,
                  error = function(e) NA_real_)

  out <- list(
    chi = chi,
    df = df,
    p_chi = p_chi,
    CAF = CAF,
    CFI = CFI,
    RMSEA = RMSEA,
    RMSEA_LB = rmsea_ci[["lower"]],
    RMSEA_UB = rmsea_ci[["upper"]],
    AIC = AIC,
    BIC = BIC,
    Fm = Fm,
    RMSR = RMSR,
    SRMR = RMSR,
    chi_null = chi_null,
    df_null = df_null,
    p_null = p_null,
    pool_method = pool_method,
    mi_diagnostics = list(
      D2_F = if (!is.null(D2)) D2$F else NA_real_,
      D2_df1 = if (!is.null(D2)) D2$df1 else NA_real_,
      D2_df2 = if (!is.null(D2)) D2$df2 else NA_real_,
      D2_chi_asymptotic = if (!is.null(D2)) D2$chi else NA_real_,
      chi_bar_naive = if (!is.null(D2)) D2$Tbar else if (length(chis) > 0L) mean(chis, na.rm = TRUE) else NA_real_,
      D2_null_F = if (!is.null(D2_null)) D2_null$F else NA_real_,
      D2_null_df1 = if (!is.null(D2_null)) D2_null$df1 else NA_real_,
      D2_null_df2 = if (!is.null(D2_null)) D2_null$df2 else NA_real_,
      D2_null_chi_asymptotic = if (!is.null(D2_null)) D2_null$chi else NA_real_,
      chi_null_bar_naive = if (!is.null(D2_null)) D2_null$Tbar else if (length(chis_null) > 0L) mean(chis_null, na.rm = TRUE) else NA_real_,
      ARIV = if (!is.null(D2)) D2$ARIV else NA_real_,
      FMI = if (!is.null(D2)) D2$FMI else NA_real_,
      ARIV_null = if (!is.null(D2_null)) D2_null$ARIV else NA_real_,
      FMI_null = if (!is.null(D2_null)) D2_null$FMI else NA_real_,
      m = if (!is.null(D2)) D2$M else length(fits)
    )
  )

  out
}

## -----------------------------------------------------------------------------
## Bootstrap pooling helpers
## -----------------------------------------------------------------------------

.efa_pooled_has_boot_arrays <- function(fits) {
  # Bootstrap MI pooling requires the raw unrotated bootstrap loading arrays.
  # Scalar bootstrap SEs/CIs are not enough because the replicates must be
  # re-expressed in the final MI target space.
  vapply(fits, function(x) {
    !is.null(x$boot.arrays) &&
      !is.null(x$boot.arrays$unrot_loadings)
  }, logical(1))
}

.efa_pooled_vec <- function(M) {
  # Column-major vectorization so matrix estimates can be Rubin-pooled.
  as.vector(as.matrix(M))
}

.efa_pooled_vech <- function(M) {
  # Vectorize the lower triangle of a symmetric Phi matrix without duplicates.
  M <- as.matrix(M)
  M[lower.tri(M, diag = TRUE)]
}

.efa_pooled_unvech <- function(v, k) {
  # Reconstruct a symmetric matrix from vech() output.
  M <- matrix(NA_real_, k, k)
  M[lower.tri(M, diag = TRUE)] <- v
  M[upper.tri(M)] <- t(M)[upper.tri(M)]
  M
}

.efa_pooled_communalities <- function(Lambda, Phi = NULL) {
  # Communalities are the diagonal of the common-factor reproduced matrix.
  Lambda <- as.matrix(Lambda)
  common_R <- if (is.null(Phi)) {
    Lambda %*% t(Lambda)
  } else {
    Lambda %*% as.matrix(Phi) %*% t(Lambda)
  }
  diag(common_R)
}

.efa_pooled_model_implied <- function(Lambda, Phi = NULL) {
  # Correlation-model implied matrix: common-factor part plus uniquenesses,
  # implemented by setting the diagonal of Lambda Phi Lambda' to one.
  Lambda <- as.matrix(Lambda)
  common_R <- if (is.null(Phi)) {
    Lambda %*% t(Lambda)
  } else {
    Lambda %*% as.matrix(Phi) %*% t(Lambda)
  }
  implied <- common_R
  diag(implied) <- 1
  implied
}

.efa_pooled_residual_from_solution <- function(R, Lambda, Phi = NULL) {
  # Residuals for one imputation in the same target space as its aligned solution.
  E <- as.matrix(R) - .efa_pooled_model_implied(Lambda, Phi)
  diag(E) <- 0
  E
}

.efa_pooled_make_ci <- function(est, se, df, alpha) {
  # Wald-type confidence intervals using Rubin degrees of freedom when finite.
  crit <- ifelse(
    is.finite(df),
    stats::qt(1 - alpha / 2, df = df),
    stats::qnorm(1 - alpha / 2)
  )
  list(
    lower = est - crit * se,
    upper = est + crit * se
  )
}


.efa_pooled_col_means <- function(x) {
  # Column means that return NA for columns without finite values.
  x <- as.matrix(x)
  if (ncol(x) < 1L) {
    return(numeric(0))
  }
  vapply(seq_len(ncol(x)), function(j) {
    vals <- x[, j]
    vals <- vals[is.finite(vals)]
    if (length(vals) < 1L) {
      NA_real_
    } else {
      mean(vals)
    }
  }, numeric(1))
}

.efa_pooled_col_vars <- function(x) {
  # Column variances that return NA for columns with fewer than two finite values.
  # Rubin pooling only needs diagonal within- and between-imputation variances,
  # so computing full covariance matrices is unnecessary and can be prohibitively
  # memory-intensive for residual matrices.
  x <- as.matrix(x)
  if (ncol(x) < 1L) {
    return(numeric(0))
  }
  if (nrow(x) < 2L) {
    return(rep(NA_real_, ncol(x)))
  }
  vapply(seq_len(ncol(x)), function(j) {
    vals <- x[, j]
    vals <- vals[is.finite(vals)]
    if (length(vals) < 2L) {
      NA_real_
    } else {
      stats::var(vals)
    }
  }, numeric(1))
}

.efa_pooled_rubin_pool <- function(q_list, boot_mat_list, alpha = 0.05,
                                   est_override = NULL) {
  # Rubin-style pooling for one parameter family.
  #
  # q_list contains one aligned point-estimate vector per imputation. boot_mat_list
  # contains aligned bootstrap replicates for estimating the within-imputation
  # variances U_d. Only diagonal variances are required for the returned SEs, CIs,
  # RIV, FMI, and df, so the implementation intentionally avoids building full
  # p x p covariance matrices for each parameter vector.
  m <- length(q_list)
  q_mat <- do.call(rbind, q_list)
  q_bar <- .efa_pooled_col_means(q_mat)

  U_diag_mat <- do.call(rbind, lapply(boot_mat_list, .efa_pooled_col_vars))
  U_bar <- colMeans(U_diag_mat)
  B_mi <- if (m > 1L) {
    .efa_pooled_col_vars(q_mat)
  } else {
    rep(0, ncol(q_mat))
  }

  T_mi <- U_bar + (1 + 1 / m) * B_mi
  T_mi[is.finite(T_mi) & T_mi < 0 & abs(T_mi) < sqrt(.Machine$double.eps)] <- 0
  se <- sqrt(T_mi)

  r <- rep(0, length(se))
  has_U <- is.finite(U_bar) & U_bar > 0
  r[has_U] <- ((1 + 1 / m) * B_mi[has_U]) / U_bar[has_U]
  r[!has_U & is.finite(B_mi) & B_mi > 0] <- Inf
  r[!is.finite(B_mi) & !is.infinite(r)] <- NA_real_

  df <- rep(Inf, length(se))
  finite_pos_r <- is.finite(r) & r > 0
  df[finite_pos_r] <- (m - 1) * (1 + 1 / r[finite_pos_r])^2
  df[is.infinite(r)] <- m - 1
  df[is.na(r)] <- NA_real_

  fmi <- r / (1 + r)
  fmi[is.infinite(r)] <- 1
  fmi[is.nan(fmi)] <- 0

  est <- if (is.null(est_override)) q_bar else est_override
  ci <- .efa_pooled_make_ci(est, se, df, alpha)

  list(
    est = est,
    q_bar = q_bar,
    se = se,
    ci = ci,
    U_bar = U_bar,
    B_mi = B_mi,
    T_mi = T_mi,
    RIV = r,
    FMI = fmi,
    df = df
  )
}


.efa_pooled_align_unrot_boot <- function(L_boot, target, align_unrotated) {
  # Align one bootstrap unrotated loading matrix to the pooled unrotated target.
  if (align_unrotated == "none") {
    return(as.matrix(L_boot))
  }

  if (align_unrotated == "procrustes") {
    return(.change_class(PROCRUSTES(
      A = as.matrix(L_boot),
      Target = as.matrix(target),
      rotation = "orthogonal"
    )$loadings, "matrix"))
  }

  .align_solution(
    L_target = as.matrix(target),
    L = as.matrix(L_boot)
  )$loadings
}

.efa_pooled_align_rot_boot <- function(L_boot_unrot, final_target, rotation_type,
                                       procrustes_args = list()) {
  # Re-rotate/re-align the bootstrap unrotated loadings to the final MI target.
  # This is the key step that makes within-imputation bootstrap variances refer
  # to the same estimand as the pooled rotated solution.
  if (rotation_type == "none") {
    return(NULL)
  }

  pr <- do.call(
    PROCRUSTES,
    c(list(A = as.matrix(L_boot_unrot),
           Target = as.matrix(final_target),
           rotation = rotation_type),
      procrustes_args)
  )

  if (!is.null(pr$convergence) && isFALSE(pr$convergence)) {
    return(NULL)
  }

  list(
    Lambda = .change_class(pr$loadings, "matrix"),
    Phi = if (!is.null(pr$Phi)) as.matrix(pr$Phi) else NULL
  )
}

.efa_pooled_rubin_matrix_result <- function(pool, nrow, ncol, dimnames = NULL) {
  # Convert vectorized Rubin output back into matrix-shaped estimates.
  est <- matrix(pool$est, nrow = nrow, ncol = ncol)
  se <- matrix(pool$se, nrow = nrow, ncol = ncol)
  lower <- matrix(pool$ci$lower, nrow = nrow, ncol = ncol)
  upper <- matrix(pool$ci$upper, nrow = nrow, ncol = ncol)

  if (!is.null(dimnames)) {
    dimnames(est) <- dimnames
    dimnames(se) <- dimnames
    dimnames(lower) <- dimnames
    dimnames(upper) <- dimnames
  }

  list(est = est, se = se, ci = list(lower = lower, upper = upper))
}

.efa_pooled_rubin_symmetric_result <- function(pool, k, dimnames = NULL) {
  # Convert vech-based Rubin output back into symmetric matrix form.
  est <- .efa_pooled_unvech(pool$est, k)
  se <- .efa_pooled_unvech(pool$se, k)
  lower <- .efa_pooled_unvech(pool$ci$lower, k)
  upper <- .efa_pooled_unvech(pool$ci$upper, k)

  if (!is.null(dimnames)) {
    dimnames(est) <- dimnames
    dimnames(se) <- dimnames
    dimnames(lower) <- dimnames
    dimnames(upper) <- dimnames
  }

  list(est = est, se = se, ci = list(lower = lower, upper = upper))
}

.efa_pooled_scalar_boot_summary <- function(x, alpha = 0.05) {
  # Percentile-style summaries for scalar quantities generated by the full
  # pooled bootstrap algorithm. Columns with no finite bootstrap values return
  # NA instead of failing inside quantile().
  x <- as.matrix(x)
  nms <- colnames(x)

  se <- sqrt(.efa_pooled_col_vars(x))
  qfun <- function(vals, prob) {
    vals <- vals[is.finite(vals)]
    if (length(vals) < 1L) {
      NA_real_
    } else {
      stats::quantile(vals, probs = prob, names = FALSE, type = 7)
    }
  }

  ci <- list(
    lower = vapply(seq_len(ncol(x)), function(j) qfun(x[, j], alpha / 2), numeric(1)),
    upper = vapply(seq_len(ncol(x)), function(j) qfun(x[, j], 1 - alpha / 2), numeric(1))
  )

  if (!is.null(nms)) {
    names(se) <- nms
    names(ci$lower) <- nms
    names(ci$upper) <- nms
  }

  list(se = se, ci = ci)
}


.efa_pooled_bootstrap_pool <- function(fits,
                                       orig_R_list,
                                       unrot_loadings_aligned,
                                       mean_unrot_loadings,
                                       rot_loadings = NULL,
                                       phis = NULL,
                                       structure_loadings = NULL,
                                       mean_structure_loadings = NULL,
                                       final_target = NULL,
                                       rotation_type = c("none", "orthogonal", "oblique"),
                                       align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
                                       procrustes_args = list(),
                                       h2,
                                       residuals,
                                       fit_indices,
                                       pooled_orig_R,
                                       N,
                                       method,
                                       pool_method = "D2",
                                       rmsea_ci_level = .90,
                                       alpha = 0.05,
                                       rmsr_upper = TRUE) {

  rotation_type <- match.arg(rotation_type)
  align_unrotated <- match.arg(align_unrotated)

  has_boot <- .efa_pooled_has_boot_arrays(fits)
  if (!all(has_boot)) {
    if (any(has_boot)) {
      warning(
        "Bootstrap arrays were found for only some imputations. Pooled bootstrap SEs/CIs were not computed.",
        call. = FALSE
      )
    }
    return(NULL)
  }

  m <- length(fits)
  p_vars <- nrow(as.matrix(mean_unrot_loadings))
  k <- ncol(as.matrix(mean_unrot_loadings))
  loading_dimnames <- dimnames(as.matrix(mean_unrot_loadings))

  B_vec <- vapply(fits, function(x) dim(x$boot.arrays$unrot_loadings)[3], integer(1))
  if (length(unique(B_vec)) > 1L) {
    B_use <- min(B_vec)
    warning(
      "The number of bootstrap replicates differs across imputations. Using the minimum number available in all imputations.",
      call. = FALSE
    )
  } else {
    B_use <- B_vec[[1]]
  }

  if (!is.finite(B_use) || B_use < 2L) {
    warning(
      "At least two bootstrap replicates per imputation are required for pooled bootstrap SEs/CIs.",
      call. = FALSE
    )
    return(NULL)
  }

  q_unrot <- lapply(unrot_loadings_aligned, .efa_pooled_vec)
  boot_unrot <- vector("list", m)

  q_rot <- boot_rot <- NULL
  q_phi <- boot_phi <- NULL
  q_structure <- boot_structure <- NULL
  q_h2 <- vector("list", m)
  boot_h2 <- vector("list", m)
  q_residuals <- vector("list", m)
  boot_residuals <- vector("list", m)

  if (rotation_type != "none") {
    q_rot <- lapply(rot_loadings, .efa_pooled_vec)
    boot_rot <- vector("list", m)
  }
  if (rotation_type == "oblique") {
    q_phi <- lapply(phis, .efa_pooled_vech)
    boot_phi <- vector("list", m)
    q_structure <- lapply(structure_loadings, .efa_pooled_vec)
    boot_structure <- vector("list", m)
  }

  scalar_fit_names <- names(fit_indices)[vapply(fit_indices, function(x) {
    is.numeric(x) && length(x) == 1L
  }, logical(1))]
  can_boot_fit_algorithm <- length(scalar_fit_names) > 0L &&
    all(vapply(fits, function(x) {
      fit_arr <- x$boot.arrays$fit_indices
      res_arr <- x$boot.arrays$residuals
      !is.null(fit_arr) && !is.null(res_arr) &&
        nrow(as.matrix(fit_arr)) >= B_use &&
        length(dim(res_arr)) == 3L && dim(res_arr)[3L] >= B_use
    }, logical(1)))
  pooled_fit_boot <- NULL
  if (can_boot_fit_algorithm) {
    pooled_fit_boot <- matrix(
      NA_real_, nrow = B_use, ncol = length(scalar_fit_names),
      dimnames = list(NULL, scalar_fit_names)
    )
  }

  nonconv_procrustes <- integer(m)

  for (d in seq_len(m)) {
    arr <- fits[[d]]$boot.arrays
    unrot_arr <- arr$unrot_loadings

    boot_unrot[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * k)

    if (rotation_type != "none") {
      boot_rot[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * k)
    }
    if (rotation_type == "oblique") {
      boot_phi[[d]] <- matrix(NA_real_, nrow = B_use, ncol = k * (k + 1) / 2)
      boot_structure[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * k)
    }

    boot_h2[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars)
    boot_residuals[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * p_vars)

    if (rotation_type == "oblique") {
      q_h2[[d]] <- .efa_pooled_communalities(rot_loadings[[d]], phis[[d]])
      q_residuals[[d]] <- .efa_pooled_vec(.efa_pooled_residual_from_solution(
        orig_R_list[[d]], rot_loadings[[d]], phis[[d]]
      ))
    } else if (rotation_type == "orthogonal") {
      q_h2[[d]] <- .efa_pooled_communalities(rot_loadings[[d]], NULL)
      q_residuals[[d]] <- .efa_pooled_vec(.efa_pooled_residual_from_solution(
        orig_R_list[[d]], rot_loadings[[d]], NULL
      ))
    } else {
      q_h2[[d]] <- .efa_pooled_communalities(unrot_loadings_aligned[[d]], NULL)
      q_residuals[[d]] <- .efa_pooled_vec(.efa_pooled_residual_from_solution(
        orig_R_list[[d]], unrot_loadings_aligned[[d]], NULL
      ))
    }

    for (b in seq_len(B_use)) {
      Lb_unrot0 <- unrot_arr[, , b]
      Lb_unrot <- .efa_pooled_align_unrot_boot(
        L_boot = Lb_unrot0,
        target = mean_unrot_loadings,
        align_unrotated = align_unrotated
      )
      boot_unrot[[d]][b, ] <- .efa_pooled_vec(Lb_unrot)

      if (rotation_type != "none") {
        # Do not use arr$rot_loadings here. Those matrices were aligned to the
        # imputation-specific EFA target. MI pooling requires every bootstrap
        # replicate to be represented in the final MI target space.
        rot_b <- .efa_pooled_align_rot_boot(
          L_boot_unrot = Lb_unrot0,
          final_target = final_target,
          rotation_type = rotation_type,
          procrustes_args = procrustes_args
        )

        if (!is.null(rot_b)) {
          boot_rot[[d]][b, ] <- .efa_pooled_vec(rot_b$Lambda)

          if (rotation_type == "oblique") {
            boot_phi[[d]][b, ] <- .efa_pooled_vech(rot_b$Phi)
            boot_structure[[d]][b, ] <- .efa_pooled_vec(rot_b$Lambda %*% rot_b$Phi)
            boot_h2[[d]][b, ] <- .efa_pooled_communalities(rot_b$Lambda, rot_b$Phi)
          } else {
            boot_h2[[d]][b, ] <- .efa_pooled_communalities(rot_b$Lambda, NULL)
          }
        } else {
          nonconv_procrustes[d] <- nonconv_procrustes[d] + 1L
        }
      } else {
        boot_h2[[d]][b, ] <- .efa_pooled_communalities(Lb_unrot, NULL)
      }

      if (!is.null(arr$residuals)) {
        boot_residuals[[d]][b, ] <- .efa_pooled_vec(arr$residuals[, , b])
      }
    }
  }

  if (any(nonconv_procrustes > 0L)) {
    warning(
      "Some bootstrap Procrustes rotations did not converge. ",
      "Pooled bootstrap SEs/CIs are based on the valid aligned replicates.",
      call. = FALSE
    )
  }

  pool_unrot <- .efa_pooled_rubin_pool(q_unrot, boot_unrot, alpha = alpha)
  pool_h2 <- .efa_pooled_rubin_pool(
    q_h2, boot_h2, alpha = alpha,
    est_override = as.vector(h2)
  )
  pool_residuals <- .efa_pooled_rubin_pool(
    q_residuals, boot_residuals, alpha = alpha,
    est_override = .efa_pooled_vec(residuals)
  )

  unrot_res <- .efa_pooled_rubin_matrix_result(
    pool_unrot, p_vars, k, loading_dimnames
  )
  residual_res <- .efa_pooled_rubin_matrix_result(
    pool_residuals, p_vars, p_vars, dimnames(as.matrix(residuals))
  )

  SE <- list(
    h2 = pool_h2$se,
    unrot_loadings = unrot_res$se,
    residuals = residual_res$se
  )

  CI <- list(
    h2 = list(lower = pool_h2$ci$lower, upper = pool_h2$ci$upper),
    unrot_loadings = unrot_res$ci,
    residuals = residual_res$ci
  )

  arrays <- list(
    unrot_loadings = boot_unrot,
    h2 = boot_h2,
    residuals = boot_residuals
  )

  MI <- list(
    unrot_loadings = list(RIV = pool_unrot$RIV, FMI = pool_unrot$FMI, df = pool_unrot$df),
    h2 = list(RIV = pool_h2$RIV, FMI = pool_h2$FMI, df = pool_h2$df),
    residuals = list(RIV = pool_residuals$RIV, FMI = pool_residuals$FMI, df = pool_residuals$df),
    bootstrap_rotation_failures = nonconv_procrustes,
    bootstrap_rotation_valid = B_use - nonconv_procrustes
  )

  if (rotation_type != "none") {
    pool_rot <- .efa_pooled_rubin_pool(q_rot, boot_rot, alpha = alpha)
    rot_res <- .efa_pooled_rubin_matrix_result(pool_rot, p_vars, k, loading_dimnames)
    SE$rot_loadings <- rot_res$se
    CI$rot_loadings <- rot_res$ci
    arrays$rot_loadings <- boot_rot
    MI$rot_loadings <- list(RIV = pool_rot$RIV, FMI = pool_rot$FMI, df = pool_rot$df)
  }

  if (rotation_type == "oblique") {
    pool_phi <- .efa_pooled_rubin_pool(q_phi, boot_phi, alpha = alpha)
    phi_res <- .efa_pooled_rubin_symmetric_result(
      pool_phi, k, dimnames(as.matrix(phis[[1]]))
    )
    SE$Phi <- phi_res$se
    CI$Phi <- phi_res$ci
    arrays$Phi <- boot_phi
    MI$Phi <- list(RIV = pool_phi$RIV, FMI = pool_phi$FMI, df = pool_phi$df)

    pool_structure <- .efa_pooled_rubin_pool(
      q_structure, boot_structure, alpha = alpha,
      est_override = .efa_pooled_vec(mean_structure_loadings)
    )
    structure_res <- .efa_pooled_rubin_matrix_result(
      pool_structure, p_vars, k, loading_dimnames
    )
    SE$Structure <- structure_res$se
    CI$Structure <- structure_res$ci
    arrays$Structure <- boot_structure
    MI$Structure <- list(
      RIV = pool_structure$RIV,
      FMI = pool_structure$FMI,
      df = pool_structure$df
    )
  }

  ## Fit-index summaries ------------------------------------------------------
  ## fit_indices_descriptive: Rubin-Wald summaries of per-imputation fit indices.
  ## fit_indices_pooled_algorithm: percentile summaries obtained by re-running
  ## the pooled residual/D2 calculations bootstrap-index by bootstrap-index.

  fit_q_names <- Reduce(intersect, lapply(fits, function(x) names(x$fit_indices)))
  fit_q_names <- fit_q_names[vapply(fits[[1]]$fit_indices[fit_q_names], function(x) {
    is.numeric(x) && length(x) == 1L
  }, logical(1))]

  can_describe_fit_boot <- length(fit_q_names) > 0L &&
    all(vapply(fits, function(x) {
      fit_arr <- x$boot.arrays$fit_indices
      !is.null(fit_arr) &&
        nrow(as.matrix(fit_arr)) >= B_use &&
        ncol(as.matrix(fit_arr)) >= length(fit_q_names)
    }, logical(1)))

  if (can_describe_fit_boot) {
    q_fit <- lapply(fits, function(x) unlist(x$fit_indices[fit_q_names]))
    boot_fit <- vector("list", m)

    for (d in seq_len(m)) {
      fit_arr <- as.matrix(fits[[d]]$boot.arrays$fit_indices)
      if (is.null(colnames(fit_arr))) {
        colnames(fit_arr) <- names(fits[[d]]$fit_indices)
      }
      boot_fit[[d]] <- fit_arr[seq_len(B_use), fit_q_names, drop = FALSE]
    }

    pool_fit_desc <- .efa_pooled_rubin_pool(q_fit, boot_fit, alpha = alpha)
    SE$fit_indices_descriptive <- stats::setNames(pool_fit_desc$se, fit_q_names)
    CI$fit_indices_descriptive <- list(
      lower = stats::setNames(pool_fit_desc$ci$lower, fit_q_names),
      upper = stats::setNames(pool_fit_desc$ci$upper, fit_q_names)
    )
    arrays$fit_indices_descriptive <- boot_fit
    MI$fit_indices_descriptive <- list(
      RIV = stats::setNames(pool_fit_desc$RIV, fit_q_names),
      FMI = stats::setNames(pool_fit_desc$FMI, fit_q_names),
      df = stats::setNames(pool_fit_desc$df, fit_q_names)
    )
  }

  if (!is.null(pooled_fit_boot)) {
    # Use bootstrap index b as a Monte Carlo pairing across imputations. The
    # bootstrap samples are independent across imputations, so the pairing has
    # no substantive meaning; it simply creates B replicate runs of the complete
    # pooled-fit algorithm.
    for (b in seq_len(B_use)) {
      Lb_unrot_list <- vector("list", m)
      Lb_rot_list <- if (rotation_type != "none") vector("list", m) else NULL
      Phib_list <- if (rotation_type == "oblique") vector("list", m) else NULL
      Rb_list <- vector("list", m)
      boot_fit_objects <- vector("list", m)

      for (d in seq_len(m)) {
        arr <- fits[[d]]$boot.arrays
        Lb_unrot <- matrix(boot_unrot[[d]][b, ], nrow = p_vars, ncol = k)
        Lb_unrot_list[[d]] <- Lb_unrot

        if (rotation_type != "none") {
          Lrow <- boot_rot[[d]][b, ]
          if (anyNA(Lrow)) {
            next
          }
          Lb_rot <- matrix(Lrow, nrow = p_vars, ncol = k)
          Lb_rot_list[[d]] <- Lb_rot

          if (rotation_type == "oblique") {
            Phirow <- boot_phi[[d]][b, ]
            if (anyNA(Phirow)) {
              next
            }
            Phib <- .efa_pooled_unvech(Phirow, k)
            Phib_list[[d]] <- Phib
            model_b <- .efa_pooled_model_implied(Lb_rot, Phib)
          } else {
            model_b <- .efa_pooled_model_implied(Lb_rot, NULL)
          }
        } else {
          model_b <- .efa_pooled_model_implied(Lb_unrot, NULL)
        }

        # EFA() stores bootstrap residuals. Adding the model-implied matrix for
        # the aligned bootstrap solution reconstructs the bootstrap observed R;
        # because rotations preserve the reproduced correlation matrix, this is
        # consistent with the final target-space representation.
        Rb <- arr$residuals[, , b] + model_b
        diag(Rb) <- 1
        Rb_list[[d]] <- Rb

        fit_arr <- as.matrix(arr$fit_indices)
        if (is.null(colnames(fit_arr))) {
          colnames(fit_arr) <- names(fits[[d]]$fit_indices)
        }
        boot_fit_objects[[d]] <- list(fit_indices = as.list(fit_arr[b, , drop = TRUE]))
      }

      if (any(vapply(Rb_list, is.null, logical(1)))) {
        next
      }

      pooled_R_b <- .average_matrices(Rb_list)

      if (rotation_type == "oblique") {
        if (any(vapply(Lb_rot_list, is.null, logical(1))) ||
            any(vapply(Phib_list, is.null, logical(1)))) {
          next
        }
        L_pool_b <- .average_matrices(Lb_rot_list)
        Phi_pool_b <- .average_matrices(Phib_list)
        model_pool_b <- .efa_pooled_model_implied(L_pool_b, Phi_pool_b)
      } else if (rotation_type == "orthogonal") {
        if (any(vapply(Lb_rot_list, is.null, logical(1)))) {
          next
        }
        L_pool_b <- .average_matrices(Lb_rot_list)
        model_pool_b <- .efa_pooled_model_implied(L_pool_b, NULL)
      } else {
        L_pool_b <- .average_matrices(Lb_unrot_list)
        model_pool_b <- .efa_pooled_model_implied(L_pool_b, NULL)
      }

      residual_b <- pooled_R_b - model_pool_b
      diag(residual_b) <- 0
      RMSR_b <- .rmsr(residual_b, upper = rmsr_upper)

      if (any(vapply(boot_fit_objects, is.null, logical(1)))) {
        next
      }

      fit_b <- .efa_pooled_fit_indices(
        fits = boot_fit_objects,
        pooled_R = pooled_R_b,
        residuals = residual_b,
        RMSR = RMSR_b,
        N = N,
        method = method,
        pool_method = pool_method,
        rmsea_ci_level = rmsea_ci_level
      )

      pooled_fit_boot[b, ] <- unlist(fit_b[scalar_fit_names])
    }

    fit_boot_summary <- .efa_pooled_scalar_boot_summary(pooled_fit_boot, alpha = alpha)
    SE$fit_indices_pooled_algorithm <- fit_boot_summary$se
    CI$fit_indices_pooled_algorithm <- fit_boot_summary$ci
    arrays$fit_indices_pooled_algorithm <- pooled_fit_boot
  }

  list(
    SE = SE,
    CI = CI,
    arrays = arrays,
    MI = MI,
    n_boot = B_use
  )
}

# Property tests for the internal GPA-consensus engine that backs
# EFA_POOLED(target_method = "consensus"): symmetry across the m inputs,
# monotone descent under the standard Gower 1975 GPA update, agreement with
# a hand-coded GPA reference loop, stationary-point convergence, and the
# oblique-rotation guard.

gpa_consensus_target <- EFAtools:::.gpa_consensus_target

# Synthetic m-list around a simple-structure L_true with random per-imputation
# orthogonal unrotations + small noise. Inputs are kept small to minimise
# test runtime.
.make_imp_list <- function(m, p, k, sigma = 0, seed = 1L) {
  set.seed(seed)
  L_true <- matrix(0, p, k)
  per <- p %/% k
  for (j in seq_len(k)) {
    L_true[((j - 1L) * per + 1L):(j * per), j] <- runif(per, 0.6, 0.8)
  }
  unrot <- lapply(seq_len(m), function(d) {
    L_true %*% qr.Q(qr(matrix(rnorm(k * k), k, k))) +
      matrix(rnorm(p * k, sd = sigma), p, k)
  })
  list(L_true = L_true, unrot = unrot)
}


test_that(".gpa_consensus_target is symmetric in its m inputs (orthogonal)", {
  d <- .make_imp_list(m = 3L, p = 8L, k = 2L, sigma = 0.02, seed = 11)
  fit1 <- gpa_consensus_target(
    unrotated_list = d$unrot,
    init_targets = list(d$L_true), start = 1L,
    rotation = "orthogonal",
    tol = 1e-8, loss_tol = 1e-10, max_iter = 100, min_iter = 2L
  )
  perm <- c(3L, 1L, 2L)
  fit2 <- gpa_consensus_target(
    unrotated_list = d$unrot[perm],
    init_targets = list(d$L_true), start = 1L,
    rotation = "orthogonal",
    tol = 1e-8, loss_tol = 1e-10, max_iter = 100, min_iter = 2L
  )
  expect_equal(unname(fit1$target), unname(fit2$target), tolerance = 1e-7)
})


test_that("orthogonal consensus loss is monotone non-increasing (Gower 1975 GPA)", {
  # match_target = FALSE, alpha = 1, rotation = orthogonal: block-coordinate
  # GPA, guaranteed monotone non-increase of the mean squared discrepancy.
  d <- .make_imp_list(m = 4L, p = 9L, k = 3L, sigma = 0.05, seed = 23)
  fit <- gpa_consensus_target(
    unrotated_list = d$unrot,
    rotation = "orthogonal",
    match_target = FALSE, alpha = 1,
    tol = 1e-10, loss_tol = 1e-12, max_iter = 100, min_iter = 2L
  )
  expect_true(all(diff(fit$history$loss) <= 1e-12))
})


test_that("consensus target is a fixed point of one more centroid update", {
  d <- .make_imp_list(m = 3L, p = 8L, k = 2L, sigma = 0.03, seed = 31)
  tol <- 1e-8
  fit <- gpa_consensus_target(
    unrotated_list = d$unrot,
    rotation = "orthogonal",
    tol = tol, loss_tol = 1e-12, max_iter = 200, min_iter = 2L
  )
  next_aligned <- lapply(d$unrot, function(L) {
    efa_procrustes(L, fit$target, rotation = "orthogonal")$loadings
  })
  centroid_next <- Reduce(`+`, next_aligned) / length(next_aligned)
  rel <- norm(unname(centroid_next) - unname(fit$target), type = "F") /
    (norm(unname(fit$target), type = "F") + 1e-12)
  expect_lt(rel, 100 * tol)
})


test_that("orthogonal consensus matches a hand-coded Gower 1975 GPA loop (m = 2)", {
  # Independent reference: alternate orthogonal Procrustes per d, then
  # arithmetic-mean centroid. Same limit at the convergence tolerance.
  d <- .make_imp_list(m = 2L, p = 8L, k = 2L, sigma = 0.05, seed = 37)
  fit <- gpa_consensus_target(
    unrotated_list = d$unrot,
    init_targets = list(d$unrot[[1L]]), start = 1L,
    rotation = "orthogonal",
    match_target = FALSE, alpha = 1,
    tol = 1e-12, loss_tol = 1e-14, max_iter = 200, min_iter = 2L
  )
  target <- d$unrot[[1L]]
  for (it in seq_len(200L)) {
    aligned <- lapply(d$unrot, function(L) {
      s <- svd(crossprod(L, target))
      L %*% (s$u %*% t(s$v))
    })
    new_target <- Reduce(`+`, aligned) / length(aligned)
    rel <- norm(new_target - target, type = "F") /
      (norm(target, type = "F") + 1e-12)
    target <- new_target
    if (rel < 1e-12) break
  }
  expect_equal(unname(fit$target), unname(target), tolerance = 1e-8)
})


test_that("oblique GPA-consensus aborts with a classed condition for k > 1", {
  # The naive iterated-oblique-Procrustes design has degenerate fixed points
  # for k > 1; Lorenzo-Seva & Van Ginkel (2016) instead apply a Promin step
  # on top of the centroid, which this engine does not implement.
  d <- .make_imp_list(m = 2L, p = 6L, k = 2L, sigma = 0.02, seed = 43)
  expect_error(
    gpa_consensus_target(
      unrotated_list = d$unrot,
      init_targets = list(d$L_true), start = 1L,
      rotation = "oblique"
    ),
    class = "efa_consensus_oblique_unsupported"
  )
})

# Procrustes-aligned analytic-SE pooling in EFA_POOLED().
#
# Companion to test-EFA_POOLED-rubin-unrot.R, focused on
# align_unrotated = "procrustes". The Procrustes orthogonal transform Q_d
# mixes loading columns, so the per-imputation full unrotated vcov is
# propagated through the Kronecker identity
#
#     Var(vec(L_d Q_d)) = (Q_d' (x) I_p) V_d (Q_d (x) I_p)
#
# (column-major vec) before per-element Rubin pooling. Q_d is treated as fixed
# (Schoenemann 1966; lavaan / mice convention).
#
# Assertion groups:
#   1. Identity-imputation invariance.
#   2. Signed-permutation orbit: procrustes pool recovers the anchor SE.
#   3. Helper kronecker check: closed-form aligned diagonals match the full
#      Kronecker propagation and a Monte-Carlo simulation.
#   4. Continuous-rotation orbit: procrustes pool recovers the anchor SE.
#   5. Cross-check that the procrustes pool and the signed_tucker_congruence
#      pool agree on a signed-permutation orbit (both reduce to the anchor SE).
#   6. NA-vcov gating: NA-filled vcov on any imputation aborts with
#      `efa_pooled_unreliable_vcov`; per-row NA in the helper still
#      NA-propagates when the helper is called directly.
#   7. Classed `efa_pooled_no_vcov` abort when any component fit is missing
#      `vcov_unrot_loadings` under `align_unrotated = "procrustes"`.

cormat <- test_models$baseline$cormat
p_vars <- ncol(cormat)
N_id   <- 500L
k      <- 3L
m_id   <- 5L
identical_list <- replicate(m_id, cormat, simplify = FALSE)

# -- Synthetic-fit helpers ----------------------------------------------------

# Build the (p k x p k) Kronecker that maps vec(L) -> vec(L S) under
# column-major vec: vec(L S) = (S' (x) I_p) vec(L). Used to transport the
# anchor's vcov_unrot_loadings into a known gauge so the test inputs satisfy
# the same Kronecker identity the production helper relies on.
.k_lhs <- function(S, p) kronecker(t(S), diag(p))

# Marginal SEs of (L S) implied by V_d in the gauged frame.
.row_marginal_se <- function(V, p, k) {
  out <- matrix(NA_real_, p, k)
  for (i in seq_len(p)) {
    idx <- (seq_len(k) - 1L) * p + i
    out[i, ] <- sqrt(pmax(diag(V[idx, idx, drop = FALSE]), 0))
  }
  out
}

# Apply a k x k gauge S to a base fit: replace the unrotated loadings,
# unrot_loading SEs, and full vcov to reflect the (L S, (S' (x) I) V (S (x) I))
# transform. Uniquenesses (rotation-invariant) are kept identical to the base.
.gauge_fit <- function(base_fit, S) {
  L_base <- unclass(base_fit$unrot_loadings)
  p <- nrow(L_base)
  kk <- ncol(L_base)

  L_g <- L_base %*% S

  V_base <- base_fit$vcov_unrot_loadings
  J <- .k_lhs(S, p)
  V_g <- J %*% V_base %*% t(J)
  V_g <- (V_g + t(V_g)) / 2

  SE_g <- .row_marginal_se(V_g, p, kk)
  dimnames(SE_g) <- dimnames(L_base)

  fit <- base_fit
  fit$unrot_loadings <- structure(L_g,
    class    = class(base_fit$unrot_loadings),
    dimnames = dimnames(base_fit$unrot_loadings)
  )
  fit$SE$unrot_loadings <- SE_g
  fit$vcov_unrot_loadings <- V_g
  fit
}

# Rotation matrix in factor space (k x k orthogonal). For k > 2, a single
# Givens rotation in the (1, 2) plane is enough to mix columns; the identity
# elsewhere keeps the test focused on the Kronecker propagation rather than
# k-specific structure.
.givens <- function(theta, k, i = 1L, j = 2L) {
  R <- diag(k)
  c_ <- cos(theta); s_ <- sin(theta)
  R[i, i] <-  c_; R[j, j] <- c_
  R[i, j] <- -s_; R[j, i] <- s_
  R
}

# Signed permutation matrix S with S[perm[j], j] = signs[j] (so L S permutes
# and sign-flips columns of L).
.signed_perm <- function(perm, signs) {
  k <- length(perm)
  S <- matrix(0, k, k)
  for (j in seq_len(k)) S[perm[j], j] <- signs[j]
  S
}

# -- Shared fixtures ----------------------------------------------------------

# The single unrotated analytic-SE fit is the anchor for every gauge orbit below
# and the oracle for Group 1, always with the same arguments, so it is fitted
# once here. `.gauge_fit()` copies before it writes, so no group can mutate the
# shared object, and the ML / rotation = "none" path is deterministic and draws
# no random numbers, so sharing it cannot move the stream the seeded groups use.
base_fit <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                rotation = "none", se = "information")

pooled_pr <- suppressMessages(EFA_POOLED(
  identical_list, n_factors = k, N = N_id, method = "ML",
  rotation = "none", se = "information",
  align_unrotated = "procrustes"
))

# ---- Group 1 ---------------------------------------------------------------

test_that("identity imputations: procrustes pool equals the per-fit SE and FMI is zero", {
  expect_identical(pooled_pr$settings$align_unrotated, "procrustes")
  expect_identical(pooled_pr$settings$se, "information")

  expect_equal(pooled_pr$SE$unrot_loadings, base_fit$SE$unrot_loadings,
               tolerance = 1e-10)
  expect_equal(pooled_pr$SE$uniquenesses,   base_fit$SE$uniquenesses,
               tolerance = 1e-10)
  expect_equal(unclass(pooled_pr$unrot_loadings),
               unclass(base_fit$unrot_loadings), tolerance = 1e-10)

  # Q_d = I_k under identical imputations only up to LAPACK SVD numerical
  # noise. Tolerate that noise; the signed_tucker_congruence pool gives an
  # exact 0 because it permutes integer column indices, not floats.
  fmi <- pooled_pr$MI$unrot_loadings$FMI
  riv <- pooled_pr$MI$unrot_loadings$RIV
  expect_true(all(abs(fmi[is.finite(fmi)]) < 1e-10))
  expect_true(all(abs(riv[is.finite(riv)]) < 1e-10))
})

# ---- Group 2 ---------------------------------------------------------------

test_that("procrustes pool over a signed-permutation orbit returns the anchor SE", {
  SE_base <- base_fit$SE$unrot_loadings
  psi_se  <- base_fit$SE$uniquenesses

  gauges <- list(
    diag(k),
    .signed_perm(c(2L, 1L, 3L), c(-1,  1,  1)),
    .signed_perm(c(3L, 2L, 1L), c( 1, -1,  1)),
    .signed_perm(c(1L, 3L, 2L), c( 1,  1, -1)),
    .signed_perm(c(2L, 3L, 1L), c(-1, -1,  1))
  )
  gauge_fits <- lapply(gauges, .gauge_fit, base_fit = base_fit)

  unrot_list <- lapply(gauge_fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list,
    align_unrotated = "procrustes"
  )

  pool <- EFAtools:::.efa_pooled_analytic_pool(
    fits = gauge_fits,
    unrot_loadings_aligned = aligned$loadings,
    align_meta = aligned$meta,
    ci = 0.95,
    align_unrotated = "procrustes"
  )

  expect_equal(unname(pool$SE$unrot_loadings), unname(SE_base), tolerance = 1e-10)
  expect_equal(unname(pool$SE$uniquenesses),   unname(psi_se),  tolerance = 1e-12)
})

# ---- Group 3 ---------------------------------------------------------------

test_that(".efa_pooled_propagate_procrustes_vcov matches the full kronecker formula and a Monte-Carlo estimate", {
  set.seed(20260621)
  p <- 4L
  kk <- 2L

  # Positive definite V of dimension (p k) x (p k); rescaling keeps numbers in O(1e-2) so
  # the MC variances do not get drowned by row-block magnitude mismatches.
  A <- matrix(stats::rnorm((p * kk)^2), p * kk, p * kk)
  V <- crossprod(A) / (p * kk)

  theta <- pi / 7
  Q <- .givens(theta, kk)

  helper <- EFAtools:::.efa_pooled_propagate_procrustes_vcov(V, Q, p, kk)

  # Brute-force Kronecker reference: V_aligned = (Q' (x) I_p) V (Q (x) I_p);
  # row-marginal variances are the (i + (j - 1) p)-th diagonal element.
  J <- kronecker(t(Q), diag(p))
  V_aligned <- J %*% V %*% t(J)
  ref <- matrix(diag(V_aligned), p, kk)
  expect_equal(helper, sqrt(pmax(ref, 0)), tolerance = 1e-12)

  # Monte-Carlo sanity: draw vec(L) ~ N(0, V), align by Q, take the empirical
  # row-marginal variance. Use a sample of 5e4; the MC standard error of a
  # variance estimate is sqrt(2 / n) ~ 6e-3, so 2e-2 is a safe tolerance.
  n_draws <- 50000L
  draws <- matrix(stats::rnorm(n_draws * p * kk), n_draws) %*% chol(V)
  mc_var <- matrix(0, p, kk)
  for (b in seq_len(n_draws)) {
    L_b <- matrix(draws[b, ], p, kk)
    mc_var <- mc_var + (L_b %*% Q)^2
  }
  mc_var <- mc_var / n_draws
  expect_equal(mc_var, ref, tolerance = 2e-2)
})

# ---- Group 4 ---------------------------------------------------------------

test_that("procrustes pool recovers the anchor SE under a continuous rotation gauge", {
  SE_base <- base_fit$SE$unrot_loadings

  rotations <- list(
    diag(k),
    .givens(pi / 7,  k, 1L, 2L),
    .givens(pi / 5,  k, 2L, 3L),
    .givens(pi / 11, k, 1L, 3L)
  )
  gauge_fits <- lapply(rotations, .gauge_fit, base_fit = base_fit)

  unrot_list <- lapply(gauge_fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list,
    align_unrotated = "procrustes"
  )

  pool <- EFAtools:::.efa_pooled_analytic_pool(
    fits = gauge_fits,
    unrot_loadings_aligned = aligned$loadings,
    align_meta = aligned$meta,
    ci = 0.95,
    align_unrotated = "procrustes"
  )

  # All aligned vcovs collapse to V_base, so B = 0 and the pool returns the
  # anchor's marginal SE matrix up to floating-point noise from the SVD-based
  # Q recovery and Kronecker round-trip.
  expect_equal(unname(pool$SE$unrot_loadings), unname(SE_base), tolerance = 1e-8)
})

# ---- Group 5 ---------------------------------------------------------------

test_that("procrustes pool and signed_tucker_congruence pool agree on the signed-permutation orbit", {
  gauges <- list(
    diag(k),
    .signed_perm(c(2L, 1L, 3L), c(-1,  1,  1)),
    .signed_perm(c(3L, 2L, 1L), c( 1, -1,  1)),
    .signed_perm(c(1L, 3L, 2L), c( 1,  1, -1)),
    .signed_perm(c(2L, 3L, 1L), c(-1, -1,  1))
  )
  gauge_fits <- lapply(gauges, .gauge_fit, base_fit = base_fit)

  unrot_list <- lapply(gauge_fits, function(f) unclass(f$unrot_loadings))

  aligned_stc <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "signed_tucker_congruence"
  )
  pool_stc <- EFAtools:::.efa_pooled_analytic_pool(
    fits = gauge_fits,
    unrot_loadings_aligned = aligned_stc$loadings,
    align_meta = aligned_stc$meta,
    ci = 0.95,
    align_unrotated = "signed_tucker_congruence"
  )

  aligned_pr <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "procrustes"
  )
  pool_pr <- EFAtools:::.efa_pooled_analytic_pool(
    fits = gauge_fits,
    unrot_loadings_aligned = aligned_pr$loadings,
    align_meta = aligned_pr$meta,
    ci = 0.95,
    align_unrotated = "procrustes"
  )

  expect_equal(unname(pool_pr$SE$unrot_loadings),
               unname(pool_stc$SE$unrot_loadings), tolerance = 1e-9)
  expect_equal(unname(pool_pr$SE$uniquenesses),
               unname(pool_stc$SE$uniquenesses),   tolerance = 1e-12)
})

# ---- Group 6 ---------------------------------------------------------------

test_that("NA-filled vcov_unrot_loadings on any imputation aborts with classed condition", {
  gauges <- list(diag(k),
                 .signed_perm(c(2L, 1L, 3L), c(-1, 1, 1)),
                 .signed_perm(c(1L, 3L, 2L), c( 1, 1,-1)))
  gauge_fits <- lapply(gauges, .gauge_fit, base_fit = base_fit)

  # `EFA()` NA-fills the whole `vcov_unrot_loadings` matrix when the analytic
  # covariance is unreliable (Heywood / singular bordered information); mimic
  # that here. The dispatcher must abort rather than silently produce all-NA
  # pooled SEs (the propagation helper would otherwise NA-skip every row and
  # the fail-closed mask would blank every pooled element with no diagnostic).
  gauge_fits[[3]]$vcov_unrot_loadings[] <- NA_real_

  unrot_list <- lapply(gauge_fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "procrustes"
  )

  expect_error(
    EFAtools:::.efa_pooled_analytic_pool(
      fits = gauge_fits,
      unrot_loadings_aligned = aligned$loadings,
      align_meta = aligned$meta,
      ci = 0.95,
      align_unrotated = "procrustes"
    ),
    class = "efa_pooled_unreliable_vcov"
  )
})

test_that(".efa_pooled_propagate_procrustes_vcov NA-fills rows whose vcov block has NAs", {
  set.seed(20260622)
  p <- 4L
  kk <- 2L
  A <- matrix(stats::rnorm((p * kk)^2), p * kk, p * kk)
  V <- crossprod(A) / (p * kk)
  Q <- .givens(pi / 7, kk)

  i_bad <- 2L
  idx_bad <- (seq_len(kk) - 1L) * p + i_bad
  V[idx_bad, idx_bad[1]] <- NA_real_

  out <- EFAtools:::.efa_pooled_propagate_procrustes_vcov(V, Q, p, kk)
  expect_true(all(is.na(out[i_bad, ])))
  other_rows <- setdiff(seq_len(p), i_bad)
  expect_true(all(!is.na(out[other_rows, , drop = FALSE])))
})

# ---- Group 7 ---------------------------------------------------------------

test_that("missing vcov_unrot_loadings under align_unrotated = 'procrustes' aborts with classed condition", {
  gauges <- list(diag(k), .signed_perm(c(2L, 1L, 3L), c(-1, 1, 1)))
  gauge_fits <- lapply(gauges, .gauge_fit, base_fit = base_fit)
  gauge_fits[[2]]$vcov_unrot_loadings <- NULL

  unrot_list <- lapply(gauge_fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "procrustes"
  )

  expect_error(
    EFAtools:::.efa_pooled_analytic_pool(
      fits = gauge_fits,
      unrot_loadings_aligned = aligned$loadings,
      align_meta = aligned$meta,
      ci = 0.95,
      align_unrotated = "procrustes"
    ),
    class = "efa_pooled_no_vcov"
  )
})

test_that("procrustes pool returns correctly shaped slots", {
  loading_dn <- dimnames(unclass(pooled_pr$unrot_loadings))

  expect_identical(dim(pooled_pr$SE$unrot_loadings),         c(p_vars, k))
  expect_identical(dimnames(pooled_pr$SE$unrot_loadings),    loading_dn)
  expect_identical(dim(pooled_pr$CI$unrot_loadings$lower),   c(p_vars, k))
  expect_identical(dim(pooled_pr$CI$unrot_loadings$upper),   c(p_vars, k))
  expect_identical(dimnames(pooled_pr$CI$unrot_loadings$lower), loading_dn)

  expect_setequal(names(pooled_pr$MI$unrot_loadings), c("RIV", "FMI", "df"))
  expect_identical(dim(pooled_pr$MI$unrot_loadings$RIV), c(p_vars, k))
  expect_identical(dim(pooled_pr$MI$unrot_loadings$FMI), c(p_vars, k))
  expect_identical(dim(pooled_pr$MI$unrot_loadings$df),  c(p_vars, k))

  expect_length(pooled_pr$SE$uniquenesses,             p_vars)
  expect_length(pooled_pr$CI$uniquenesses$lower,       p_vars)
  expect_length(pooled_pr$CI$uniquenesses$upper,       p_vars)
  expect_setequal(names(pooled_pr$MI$uniquenesses),    c("RIV", "FMI", "df"))

  expect_null(pooled_pr$replicates)
})

# Rubin's-rules pooling of analytic unrotated-loading and uniqueness SEs in
# EFA_POOLED(). Covers the se = "information" path, where component fits expose
# closed-form expected-information SE matrices and no bootstrap replicate cubes.
#
# Assertion groups:
#   1. Identity-imputation invariance.
#   2. Signed-permutation invariance of marginal SEs after alignment.
#   3. lavaan.mi cross-check (SE only; df conventions differ).
#   4. Rubin (1987) df sanity.
#   5. Uniquenesses pooling shape and invariance.
#   6. Wald CI symmetry and monotonicity.
#   7. FMI / RIV bounds and per-element NA propagation (fail-closed).
#   8. Slot shape, settings preservation, and Procrustes abort.

cormat <- test_models$baseline$cormat
p_vars <- ncol(cormat)
N_id   <- 500L
k      <- 3L
m_id   <- 5L
identical_list <- replicate(m_id, cormat, simplify = FALSE)

# The identity-imputation pool and the corresponding single fit are needed by
# several groups below with byte-identical arguments, so they are built once
# here. Both are read-only at every use (the gauge variants in Group 2 are
# copy-on-modify replicas), and the ML / rotation = "none" / analytic-SE path is
# deterministic and consumes no RNG, so sharing them cannot move the stream that
# the seeded groups draw from.
pooled_info <- suppressMessages(EFA_POOLED(
  identical_list, n_factors = k, N = N_id, method = "ML",
  rotation = "none", se = "information"
))
single_fit <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                  rotation = "none", se = "information")

# ---- Group 1 ---------------------------------------------------------------

test_that("identity imputations: pooled SE equals the per-fit SE and FMI is zero", {
  expect_identical(pooled_info$settings$se, "information")
  expect_identical(pooled_info$settings$component_se, "information")

  expect_equal(pooled_info$SE$unrot_loadings, single_fit$SE$unrot_loadings,
               tolerance = 1e-10)
  expect_equal(pooled_info$SE$uniquenesses, single_fit$SE$uniquenesses,
               tolerance = 1e-10)
  expect_equal(unclass(pooled_info$unrot_loadings),
               unclass(single_fit$unrot_loadings), tolerance = 1e-10)

  fmi <- pooled_info$MI$unrot_loadings$FMI
  riv <- pooled_info$MI$unrot_loadings$RIV
  expect_true(all(fmi[is.finite(fmi)] == 0))
  expect_true(all(riv[is.finite(riv)] == 0))

  # Plain Rubin df (nu_com = Inf): identical imputations leave B = 0, so the
  # relative increase in variance r = 0 and nu_old = (m - 1)(1 + 1/r)^2 -> Inf.
  # Every element's df is therefore Inf (the Wald interval reduces to the normal
  # reference), matching lavaan.mi's df -> Inf for asymptotically-normal estimates.
  df <- pooled_info$MI$unrot_loadings$df
  expect_true(all(is.infinite(df)))
})

# ---- Group 2 ---------------------------------------------------------------

test_that("aligned-pool SEs are invariant to per-imputation sign flips and column permutations", {
  L_base  <- unclass(single_fit$unrot_loadings)
  SE_base <- single_fit$SE$unrot_loadings
  psi_se  <- single_fit$SE$uniquenesses

  # The first entry is the anchor (identity gauge); the rest are deliberate
  # column-permutation + sign-flip variants of the same fit. Marginal SEs are
  # non-negative scalars (sign-flip is a no-op) and merely permute under
  # column reordering, so the analytic pool must recover the anchor SE matrix
  # exactly (B = 0 within the identity-gauge orbit).
  gauges <- list(
    list(perm = c(1L, 2L, 3L), signs = c( 1,  1,  1)),
    list(perm = c(2L, 1L, 3L), signs = c(-1,  1,  1)),
    list(perm = c(3L, 2L, 1L), signs = c( 1, -1,  1)),
    list(perm = c(1L, 3L, 2L), signs = c( 1,  1, -1)),
    list(perm = c(2L, 3L, 1L), signs = c(-1, -1,  1))
  )

  permuted_fits <- lapply(gauges, function(g) {
    fit <- single_fit
    L_g  <- L_base[, g$perm, drop = FALSE]
    L_g  <- sweep(L_g, 2, g$signs, `*`)
    fit$unrot_loadings <- structure(L_g,
      class    = class(single_fit$unrot_loadings),
      dimnames = dimnames(single_fit$unrot_loadings)
    )
    fit$SE$unrot_loadings <- SE_base[, g$perm, drop = FALSE]
    fit$SE$uniquenesses   <- psi_se
    fit
  })

  unrot_list <- lapply(permuted_fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list,
    align_unrotated = "signed_tucker_congruence"
  )

  pool <- EFAtools:::.efa_pooled_analytic_pool(
    fits = permuted_fits,
    unrot_loadings_aligned = aligned$loadings,
    align_meta = aligned$meta,
    ci = 0.95
  )

  expect_equal(unname(pool$SE$unrot_loadings), unname(SE_base), tolerance = 1e-12)
  expect_equal(unname(pool$SE$uniquenesses),   unname(psi_se),  tolerance = 1e-12)
})

# ---- Group 3 ---------------------------------------------------------------

test_that("Rubin pooling of marginal SEs matches lavaan.mi when fed identical per-imp inputs", {
  # This test isolates the pooling math from the fit math. EFAtools' ML
  # extraction and lavaan's ML estimator use different identifications and
  # produce per-imputation loadings that can drift by ~5% in finite samples,
  # so feeding each pipeline its own per-imp fits would mix that drift into
  # the comparison. Instead we fit lavaan per imputation, capture est + se,
  # pool those inputs through .efa_pooled_rubin_core, and compare to
  # lavaan.mi's pooled SE. Pooled SEs are df-agnostic and must agree to ~1e-6.
  skip_on_cran()
  skip_if_not_installed("lavaan")
  skip_if_not_installed("lavaan.mi")

  set.seed(20260621)
  p_lav <- 6L
  N_lav <- 400L
  m_lav <- 10L
  L_true <- matrix(c(0.70, 0.80, 0.65, 0.55, 0.75, 0.60), p_lav, 1)
  Sigma  <- tcrossprod(L_true)
  diag(Sigma) <- 1
  # All uniquenesses are strictly positive, so Sigma is positive definite and
  # chol() is unique (positive-diagonal convention). Drawing through the
  # Cholesky factor therefore reproduces the same sample from the same seed on
  # every LAPACK build, which an eigen-based draw would not: eigenvector signs
  # are not pinned by the decomposition.
  Rt <- chol(Sigma)

  imps <- lapply(seq_len(m_lav), function(i) {
    X <- matrix(stats::rnorm(N_lav * p_lav), N_lav, p_lav) %*% Rt
    colnames(X) <- paste0("V", seq_len(p_lav))
    as.data.frame(X)
  })

  vars  <- colnames(imps[[1]])
  model <- paste0("f =~ ", paste(vars, collapse = " + "))

  # Per-imp single-fit estimates and SEs from lavaan, in canonical loading order.
  est_lav <- matrix(NA_real_, m_lav, p_lav)
  se_lav  <- matrix(NA_real_, m_lav, p_lav)
  for (d in seq_len(m_lav)) {
    fit_d <- suppressWarnings(suppressMessages(
      lavaan::cfa(model, data = imps[[d]], std.lv = TRUE)
    ))
    pe <- lavaan::parameterEstimates(fit_d)
    ld <- pe[pe$op == "=~" & pe$lhs == "f", ]
    ld <- ld[match(vars, ld$rhs), ]
    est_lav[d, ] <- ld$est
    se_lav[d, ]  <- ld$se
  }
  if (anyNA(est_lav) || anyNA(se_lav)) skip("lavaan per-imp fits returned NAs")

  Qbar <- colMeans(est_lav)
  Ubar <- colMeans(se_lav^2)
  B    <- apply(est_lav, 2, stats::var)

  efa_core <- EFAtools:::.efa_pooled_rubin_core(
    Qbar, Ubar, B, m = m_lav, N_pool = Inf, alpha = 0.05
  )

  fit_mi <- tryCatch(
    suppressWarnings(suppressMessages(
      lavaan.mi::cfa.mi(model, data = imps, std.lv = TRUE)
    )),
    error = function(e) e
  )
  if (inherits(fit_mi, "error")) {
    skip(paste0("lavaan.mi::cfa.mi failed: ", conditionMessage(fit_mi)))
  }

  # asymptotic = TRUE selects the textbook Rubin (1987) formula
  # T = W + (1 + 1/m) B with normal-z CIs. lavaan.mi's default
  # (asymptotic = FALSE) instead rescales W by (1 + 1/m), which is a
  # lavaan-specific adjustment and not what EFA_POOLED implements.
  pe_mi <- tryCatch(
    lavaan.mi::parameterEstimates.mi(fit_mi, ci = TRUE, level = 0.95,
                                     asymptotic = TRUE),
    error = function(e) NULL
  )
  if (is.null(pe_mi)) skip("lavaan.mi::parameterEstimates.mi failed")
  ld_mi <- pe_mi[pe_mi$op == "=~" & pe_mi$lhs == "f", ]
  ld_mi <- ld_mi[match(vars, ld_mi$rhs), ]
  if (anyNA(ld_mi$se)) skip("lavaan.mi returned NA standard errors")

  # Same per-imp inputs -> same pooled SE under the textbook Rubin formula.
  expect_equal(efa_core$se, ld_mi$se, tolerance = 1e-6)
})

# ---- Group 4 ---------------------------------------------------------------

test_that(".efa_pooled_rubin_core: BR df inflates the critical, is monotone in N_pool, and limits to Rubin", {
  Qbar <- 0.5
  Ubar <- 0.01
  B    <- 0.005
  m    <- 5L

  core_500 <- EFAtools:::.efa_pooled_rubin_core(Qbar, Ubar, B, m,
                                                N_pool = 500, alpha = 0.05)
  core_50  <- EFAtools:::.efa_pooled_rubin_core(Qbar, Ubar, B, m,
                                                N_pool = 50,  alpha = 0.05)
  core_inf <- EFAtools:::.efa_pooled_rubin_core(Qbar, Ubar, B, m,
                                                N_pool = Inf, alpha = 0.05)

  expect_true(is.finite(core_500$df))
  expect_gt(stats::qt(0.975, core_500$df), stats::qnorm(0.975))

  rubin_df <- (m - 1) * (1 + Ubar / ((1 + 1 / m) * B))^2
  expect_equal(core_inf$df, rubin_df, tolerance = 1e-9)

  # Shrinking N_pool shrinks the BR df.
  expect_lt(core_50$df, core_500$df)
  expect_lt(core_500$df, core_inf$df)

  # SE is df-agnostic; identical across all three calls.
  expect_equal(core_50$se,  core_inf$se, tolerance = 1e-12)
  expect_equal(core_500$se, core_inf$se, tolerance = 1e-12)
})

test_that(".efa_pooled_rubin_core: BR fmi = 1 (Ubar = 0, B > 0) falls back to plain-Rubin df", {
  # Within-imputation analytic SE = 0 with non-zero between-imputation variance
  # makes fmi -> 1 and nu_obs -> 0 in the BR formula, which would otherwise
  # collapse df to 0 and qt(., 0) to NaN. The core falls back to df_old.
  core <- EFAtools:::.efa_pooled_rubin_core(
    Qbar = 0.5, Ubar = 0, B = 0.01, m = 5L, N_pool = 200, alpha = 0.05
  )
  expect_true(is.finite(core$df))
  expect_equal(core$df, 5L - 1L, tolerance = 1e-12)
  expect_true(is.finite(core$ci$lower))
  expect_true(is.finite(core$ci$upper))
  expect_equal(core$FMI, 1, tolerance = 1e-12)
})

# ---- Group 5 ---------------------------------------------------------------

test_that("uniqueness pooling has correct length, is non-negative and identity-invariant", {
  expect_length(pooled_info$SE$uniquenesses, p_vars)
  expect_true(all(
    pooled_info$SE$uniquenesses[!is.na(pooled_info$SE$uniquenesses)] >= 0
  ))
  expect_equal(unname(pooled_info$SE$uniquenesses),
               unname(single_fit$SE$uniquenesses), tolerance = 1e-10)
})

# ---- Group 6 ---------------------------------------------------------------

test_that("pooled Wald CIs are symmetric around the pooled estimate", {
  set.seed(42)
  imps <- lapply(seq_len(5), function(i) {
    GRiPS_raw[sample.int(nrow(GRiPS_raw), replace = TRUE), , drop = FALSE]
  })
  pooled <- suppressWarnings(suppressMessages(EFA_POOLED(
    imps, n_factors = 1, method = "ML", rotation = "none", se = "information"
  )))

  est <- unclass(pooled$unrot_loadings)
  lo  <- pooled$CI$unrot_loadings$lower
  hi  <- pooled$CI$unrot_loadings$upper

  finite <- is.finite(lo) & is.finite(hi) & is.finite(est)
  expect_true(any(finite))
  expect_true(all(lo[finite] < est[finite]))
  expect_true(all(est[finite] < hi[finite]))
  expect_equal(unname((lo[finite] + hi[finite]) / 2),
               unname(est[finite]), tolerance = 1e-10)
})

# ---- Group 7 ---------------------------------------------------------------

test_that("FMI is in [0,1], RIV >= 0; per-element NA in any imputation propagates", {
  set.seed(42)
  imps <- lapply(seq_len(5), function(i) {
    GRiPS_raw[sample.int(nrow(GRiPS_raw), replace = TRUE), , drop = FALSE]
  })
  pooled <- suppressWarnings(suppressMessages(EFA_POOLED(
    imps, n_factors = 1, method = "ML", rotation = "none", se = "information"
  )))

  fmi <- pooled$MI$unrot_loadings$FMI
  riv <- pooled$MI$unrot_loadings$RIV
  expect_true(all(fmi[is.finite(fmi)] >= 0))
  expect_true(all(fmi[is.finite(fmi)] <= 1))
  expect_true(all(riv[is.finite(riv)] >= 0))

  # Inject NA into one imputation's SE for a single element and re-pool via
  # the helper directly. Expect NA at that position in SE, both CI bounds, and
  # all three MI fields; surrounding elements must remain finite.
  fits <- pooled$fits
  fits[[3]]$SE$unrot_loadings[2, 1] <- NA_real_

  unrot_list <- lapply(fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "signed_tucker_congruence"
  )
  na_pool <- EFAtools:::.efa_pooled_analytic_pool(
    fits = fits,
    unrot_loadings_aligned = aligned$loadings,
    align_meta = aligned$meta,
    ci = 0.95
  )

  expect_true(is.na(na_pool$SE$unrot_loadings[2, 1]))
  expect_true(is.na(na_pool$CI$unrot_loadings$lower[2, 1]))
  expect_true(is.na(na_pool$CI$unrot_loadings$upper[2, 1]))
  expect_true(is.na(na_pool$MI$unrot_loadings$FMI[2, 1]))
  expect_true(is.na(na_pool$MI$unrot_loadings$RIV[2, 1]))
  expect_true(is.na(na_pool$MI$unrot_loadings$df[2, 1]))

  # Other positions in the same column unaffected.
  expect_false(is.na(na_pool$SE$unrot_loadings[1, 1]))
  expect_false(is.na(na_pool$SE$unrot_loadings[3, 1]))
})

test_that("an unreliable per-imputation unrotated SE warns and NA-blanks the pooled block", {
  base_fit <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                  rotation = "none", se = "information")
  fits <- replicate(3L, base_fit, simplify = FALSE)

  # Mimic a gauge-degenerate fit: EFA() NA-fills the marginal unrotated SEs while the
  # full covariance stays finite, so the whole-fit reliability gate passes and the
  # fail-closed mask silently blanks the entire pooled unrotated SE block.
  fits[[2]]$SE$unrot_loadings[] <- NA_real_

  unrot_list <- lapply(fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "signed_tucker_congruence"
  )

  expect_warning(
    pooled <- EFAtools:::.efa_pooled_analytic_pool(
      fits = fits,
      unrot_loadings_aligned = aligned$loadings,
      align_meta = aligned$meta,
      ci = 0.95
    ),
    class = "efa_pooled_unrotated_se_unreliable"
  )

  expect_true(all(is.na(pooled$SE$unrot_loadings)))
  # the uniquenesses were untouched, so the warning is specific to the affected family
  expect_false(any(is.na(pooled$SE$uniquenesses)))
})

# ---- Group 8 ---------------------------------------------------------------

test_that("analytic-pool output has correct slot shapes and preserves se = 'information'", {
  loading_dn <- dimnames(unclass(pooled_info$unrot_loadings))

  expect_identical(dim(pooled_info$SE$unrot_loadings),         c(p_vars, k))
  expect_identical(dimnames(pooled_info$SE$unrot_loadings),    loading_dn)
  expect_identical(dim(pooled_info$CI$unrot_loadings$lower),   c(p_vars, k))
  expect_identical(dim(pooled_info$CI$unrot_loadings$upper),   c(p_vars, k))
  expect_identical(dimnames(pooled_info$CI$unrot_loadings$lower), loading_dn)

  expect_setequal(names(pooled_info$MI$unrot_loadings), c("RIV", "FMI", "df"))
  expect_identical(dim(pooled_info$MI$unrot_loadings$RIV), c(p_vars, k))
  expect_identical(dim(pooled_info$MI$unrot_loadings$FMI), c(p_vars, k))
  expect_identical(dim(pooled_info$MI$unrot_loadings$df),  c(p_vars, k))

  expect_length(pooled_info$SE$uniquenesses,             p_vars)
  expect_length(pooled_info$CI$uniquenesses$lower,       p_vars)
  expect_length(pooled_info$CI$uniquenesses$upper,       p_vars)
  expect_setequal(names(pooled_info$MI$uniquenesses),    c("RIV", "FMI", "df"))
  expect_length(pooled_info$MI$uniquenesses$RIV,         p_vars)
  expect_length(pooled_info$MI$uniquenesses$FMI,         p_vars)
  expect_length(pooled_info$MI$uniquenesses$df,          p_vars)

  # Analytic path: replicates slot is present-but-NULL (matches the EFA() schema
  # contract: `expect_true("replicates" %in% names(fit));
  # expect_null(fit$replicates)`), no rotation-failure counters.
  expect_true("replicates" %in% names(pooled_info))
  expect_null(pooled_info$replicates)
  expect_null(pooled_info$MI$unrot_loadings$bootstrap_rotation_failures)
  expect_null(pooled_info$MI$unrot_loadings$bootstrap_rotation_valid)
})

test_that("analytic-pooled summary note advertises Wald-from-information CIs, not bootstrap", {
  # summary() exercises the full view that emits the provenance note; the brief
  # print() view suppresses it (passes ci = "none" to .print_efa_bootstrap_note).
  out <- testthat::capture_output(print(summary(pooled_info)))
  expect_match(out, "Wald CIs from the expected information matrix",
               fixed = TRUE)
  expect_false(grepl("Bootstrap/MI CIs", out, fixed = TRUE))
})

#' Plot EFA_AVERAGE object
#'
#' Plot method showing a summarized output of the \link{EFA_AVERAGE} function
#'
#' @param x list. An output from the \link{EFA_AVERAGE} function.
#' @param ... not used.
#'
#' @importFrom rlang .data
#' @export
#' @method plot EFA_AVERAGE
#'
#' @examples
#' \dontrun{
#' EFA_aver <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)
#' EFA_aver
#' }
#'
plot.EFA_AVERAGE <- function(x, ...) {

  averaging <- x$settings$averaging

  # Prepare data
  dat <- lapply(x$loadings, function(temp){
    temp <- as.data.frame(unclass(temp))

    temp <- temp  %>%
      tibble::rownames_to_column() %>%
      tidyr::pivot_longer(-.data$rowname, names_to = "colname", values_to = "loadings")

    return(temp)
  })

  dat <- do.call(cbind, dat)

  dat <- dat[, c("average.rowname", "average.colname", "average.loadings",
                 "min.loadings", "max.loadings")]
  names(dat) <- c("row_ind", "col_ind", "average", "min", "max")
  dat$row_ind <- factor(dat$row_ind, levels = rownames(x$loadings$average))
  dat$col_ind <- factor(dat$col_ind, levels = colnames(x$loadings$average))

  # Create plot faceted for variables and factors

  plot_load <- ggplot2::ggplot(dat) +
    ggplot2::geom_segment(ggplot2::aes(x = min, xend = max, y = 0, yend = 0)) +
    ggplot2::geom_segment(ggplot2::aes(x = min, xend = min, y = -0.5, yend = 0.5)) +
    ggplot2::geom_segment(ggplot2::aes(x = max, xend = max, y = -0.5, yend = 0.5)) +
    ggplot2::geom_rect(xmin = -x$settings$salience_threshold,
                       xmax = x$settings$salience_threshold,
                       ymin = -2, ymax = 2, fill = ggplot2::alpha("grey", 0.3)) +
    ggplot2::scale_y_continuous(limits = c(-1, 1)) +
    ggplot2::geom_point(ggplot2::aes(.data$average, 0), color = "darkred") +
    ggplot2::facet_grid(rows = ggplot2::vars(.data$row_ind),
                        cols = ggplot2::vars(.data$col_ind),
                        switch = "y") +
    ggplot2::theme_minimal() +
    ggplot2::ggtitle(paste0("Minimum, Maximum, and ",
                                ifelse(averaging == "mean", "Mean", "Median"),
                                " Loadings"), ) +
    ggplot2::theme(axis.title.y = ggplot2::element_blank(),
          axis.title.x = ggplot2::element_blank(),
          axis.text.y = ggplot2::element_blank(),
          axis.line.x = ggplot2::element_line(color = "black", size = 0.2),
          axis.ticks.x = ggplot2::element_line(color = "black", size = 0.2),
          plot.title = ggplot2::element_text(face = "bold", hjust = 0.5),
          panel.grid.minor.y = ggplot2::element_blank(),
          panel.grid.major.y = ggplot2::element_blank(),
          panel.spacing.y = ggplot2::unit(0, "mm"),
          strip.text.y.left = ggplot2::element_text(angle = 0, hjust = 0),
          strip.text.x = ggplot2::element_text(face = "bold"),
          strip.background.x = ggplot2::element_rect(color = "black", size = 0.2),
          panel.border = ggplot2::element_rect(color = "gray", fill = NA,
                                               size = 0.2)
          )

  print(plot_load)

}

#' Print BARTLETT object
#'
#' @param x list of class BARTLETT (output from the \link{BARTLETT} function)
#' @param ... additional arguments passed to print
#'
#' @method print BARTLETT
#'
#' @export
#'
#' @examples
#' BARTLETT(test_models$baseline$cormat, N = 500)
#'
print.BARTLETT <- function(x, ...) {

  pval <- x$p_value

  if(!is.na(pval) && !is.null(pval)){

  if(pval < .05){

    cat("\n")
    cat(crayon::green$bold(cli::symbol$tick), "The",
        crayon::bold("Bartlett's test of sphericity"), "was",
        crayon::green$bold("significant"),
    "at an alpha level of .05.")
    cat("\n")
    cat(crayon::bold(" "), "These data are probably suitable for factor analysis.")

  } else {

    cat("\n")
    cat(crayon::red$bold(cli::symbol$cross),
        "The Bartlett's test of sphericity was", crayon::red$bold("not significant"),
    "at an alpha level of .05.")
    cat("\n")
    cat(crayon::bold(" "), "These data are probably not suitable for factor analysis.")

  }

} else {

    cat("\n")
    cat(crayon::yellow(crayon::bold("!"), "The Bartlett's test of sphericity did not render a result."))
    cat("\n")

}

  cat("\n")
  cat("\n")
  cat(crayon::bold("  "), "\U1D712\U00B2(", x$df, ") = ", round(x$chisq, 2), ", ",
      crayon::italic("p"), ifelse(pval < .001, " < .001",
                                  paste(" = ", round(pval, 3), sep = "")),
      sep = "")
  cat("\n")

}

#' Plot KGC object
#'
#' Plot method showing a summarized output of the \link{KGC} function
#'
#' @param x a list of class KGC. An output from the \link{KGC} function.
#' @param ... not used.
#'
#' @export
#' @method plot KGC
#'
#' @examples
#' KGC_base <- KGC(test_models$baseline$cormat)
#' plot(KGC_base)
#'
plot.KGC <- function(x, ...) {

  eigen_PCA <- x$eigen_PCA
  eigen_SMC <- x$eigen_SMC
  eigen_EFA <- x$eigen_EFA
  nfac_PCA <- x$n_fac_PCA
  nfac_SMC <- x$n_fac_SMC
  nfac_EFA <- x$n_fac_EFA

  # Create plots
  if(!is.na(nfac_PCA)){

    .plot_KGC_helper(eigen = eigen_PCA, n_fac = nfac_PCA, eigen_name = "PCA")

  }

  if (!is.na(nfac_SMC)) {

    .plot_KGC_helper(eigen = eigen_SMC, n_fac = nfac_SMC, eigen_name = "SMC")

  }

  if (!is.na(nfac_EFA)) {

    .plot_KGC_helper(eigen = eigen_EFA, n_fac = nfac_EFA, eigen_name = "EFA")

  }

}

.plot_KGC_helper <- function(eigen, n_fac, eigen_name){
  # eigen = eigenvalues found with specific type (PCA, SMC or EFA)
  # n_fac = number of factors suggested with this type
  # eigen_name = name of type (one of "PCA", "SMC", "EFA")

x_len <- length(eigen)

p_eigen <- pretty(c(min(eigen) * .9, eigen, max(eigen) * 1.1))

graphics::plot.new()
graphics::plot.window(xlim = c(1, x_len),
                      ylim = c(min(p_eigen), max(p_eigen)))
graphics::axis(1, seq_len(x_len))
graphics::axis(2, p_eigen, las = 1)

graphics::mtext("Factor", side = 1, line = 3, cex = 1)
graphics::mtext("Eigenvalues", side = 2, line = 3, cex = 1, padj =.5)

graphics::lines(seq_len(x_len), eigen)
graphics::points(seq_len(x_len), eigen, pch = 16)
graphics::abline(h = 1, lty = 2)

graphics::points(n_fac, eigen[n_fac], pch = 1, cex = 2, col = "red")

graphics::text(n_fac, eigen[n_fac], n_fac,
               pos = 4, cex = 1.2, col = "red",
               font = 1, offset = .75)

title <- paste0("N factors suggested by Kaiser-Guttman criterion with ",
                eigen_name, ": ", n_fac)

graphics::title(title, cex.main = 1.2)

}

#' Velicer's Minimum Average Partial (MAP) Criterion
#'
#' @description
#' Computes Velicer's Minimum Average Partial (MAP) criterion for determining the number of
#' factors/components to retain. The function implements the original MAP criterion
#' (Velicer, 1976), expressed via the \eqn{\mathrm{TR2}} representation, and the revised
#' \eqn{\mathrm{TR4}} variant proposed by Velicer, Eaton, and Fava (2000).
#'
#' @details
#'
#' #' MAP is based on the idea that systematic common variance is increasingly removed from a
#' correlation matrix \eqn{R} as principal components are partialled out. After removing the
#' first \eqn{m} components, a residual (partial) covariance matrix is obtained as
#' \deqn{C_m = R - A_m A_m',}
#' where \eqn{A_m} contains the first \eqn{m} principal component loading vectors (PCA loadings).
#' This residual matrix is then standardized to a partial correlation matrix
#' \deqn{R^*_m = D_m^{-1/2} \, C_m \, D_m^{-1/2},}
#' with \eqn{D_m = \mathrm{diag}(C_m)}. The MAP criteria summarize the off-diagonal association
#' remaining in \eqn{R^*_m}. The recommended number of factors/components is the \eqn{m} that
#' minimizes the chosen criterion.
#'
#' This function returns two MAP criteria:
#' \itemize{
#'   \item \strong{TR2 (original MAP):} \deqn{\mathrm{MAP}_m = \frac{\mathrm{Trace}(R^{*2}_m) - p}{p(p-1)}}
#'   which is algebraically equivalent to the mean squared off-diagonal partial correlations and
#'   corresponds to Velicer's original MAP procedure.
#'   \item \strong{TR4 (revised MAP):} \deqn{\mathrm{MAP4}_m = \frac{\mathrm{Trace}(R^{*4}_m) - p}{p(p-1)}}
#'   a higher-order variant that places more weight on dominant residual association structure.
#' }
#'
#' \strong{Input handling.} \code{x} can be a correlation matrix or raw data. If \code{x} is not a
#' correlation matrix, correlations are computed using \code{\link[stats]{cor}} with the requested
#' missing-data handling (\code{use}) and association measure (\code{cor_method}). If a correlation
#' matrix is supplied, \code{N} must be provided.
#'
#' \strong{Matrix conditioning.} The function stops if the correlation matrix is singular (non-invertible),
#' because subsequent computations rely on stable matrix operations. If the correlation matrix is not
#' positive definite (e.g., due to sampling error), it is smoothed using \code{\link[psych]{cor.smooth}}.
#'
#' \strong{PCA-based partialing.} The PCA loading matrix \eqn{A} is obtained from the eigen-decomposition
#' of \eqn{R} as \eqn{A = V \Lambda^{1/2}}. For each \eqn{m = 0, \dots, p-1}, the first \eqn{m} columns
#' of \eqn{A} are used to compute \eqn{C_m = R - A_m A_m'}. The residual is re-standardized to the
#' partial correlation matrix \eqn{R^*_m} using \eqn{D_m^{-1/2}} (i.e., dividing by the square roots of
#' residual variances).
#'
#' \strong{Termination.} If residual variances (the diagonal of \eqn{C_m}) become non-positive or
#' numerically unstable, the loop terminates early because \eqn{R^*_m} cannot be formed reliably.
#'
#' @param x A numeric \code{matrix} or \code{data.frame}. Can be either (a) a correlation matrix, or
#'   (b) raw data (rows = observations, columns = variables) from which correlations are computed.
#' @param use Character string specifying the treatment of missing values when computing correlations.
#'   Passed to \code{\link[stats]{cor}}. Defaults to \code{"pairwise.complete.obs"}.
#' @param cor_method Character string specifying the correlation coefficient to be computed if raw
#'   data are supplied. Passed to \code{\link[stats]{cor}}. Defaults to \code{"pearson"}.
#'
#' @return An object of class \code{"MAP"} with the following elements:
#' \itemize{
#'   \item \code{eigenvalues}: Eigenvalues of the (possibly smoothed) correlation matrix.
#'   \item \code{n_factors_TR2}: Index \eqn{m} that minimizes the TR2 (original MAP) criterion.
#'   \item \code{n_factors_TR4}: Index \eqn{m} that minimizes the TR4 (revised MAP) criterion.
#'   \item \code{criteria}: A \code{matrix} with columns \code{m}, \code{TR2 (orig. MAP)}, and
#'   \code{TR4 (revised MAP)}.
#'   \item \code{settings}: A list containing \code{use}, \code{cor_method}, and \code{N}.
#' }
#'
#' @references
#' Velicer, W. F. (1976). Determining the number of components from the matrix of partial correlations.
#' \emph{Psychometrika, 41}, 321--327.
#'
#' Velicer, W. F., Eaton, C. A., & Fava, J. L. (2000). Construct explication through factor or component analysis: A review and evaluation of alternative procedures for determining the number of factors or components. In Goffin, R. D. & Helmes, E. (Eds.), \emph{Problems and Solutions in
#' Human Assessment: Honoring Douglas N. Jackson at Seventy} (pp. 41--71). Boston: Kluwer.
#'
#'
#' @examples
#' ## Example with raw data
#' res <- MAP(GRiPS_raw)
#' res
#'
#' ## Example with a correlation matrix
#' res2 <- MAP(test_models$baseline$cormat)
#' res2
#'
#' @export
MAP <- function(x,
                use = c("pairwise.complete.obs", "all.obs",
                        "complete.obs", "everything",
                        "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall")) {

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  use <- match.arg(use)
  cor_method <- match.arg(cor_method)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }

  p <- ncol(R)

  # set up m_grid
  m_max <- p - 1
  ms <- 0:m_max
  criteria <- matrix(NA_real_, nrow = p, ncol = 3)
  criteria[, 1] <- ms
  colnames(criteria) <- c("m", "TR2 (orig. MAP)", "TR4 (revised MAP)")


  # PCA to get the loadings A
  ed <- eigen(R, symmetric = TRUE)
  vals <- ed$values
  vecs <- ed$vectors
  A <- vecs %*% sqrt(diag(vals))

  map_from_trace_power <- function(M) {
    # "Trace of the matrix ... to the k-th power" in the VEF (2000) sense (matrix power),
    # then transformed to the MAP scale:
    #   (Trace(M^k) - p) / (p(p-1))  (matches the TR2-to-MAP equivalence described in the chapter)
    M2 <- M %*% M
    M4 <- M2 %*% M2

    c((sum(diag(M2)) - p) / (p * (p - 1)), # TR2 criterion
      (sum(diag(M4)) - p) / (p * (p - 1))) # TR4 criterion
  }

  # m=0, Rstar = R
  criteria[1, c(2, 3)] <- map_from_trace_power(R)

  # run through ms
  for (m in seq_len(m_max)) {

    Am <- A[, 1:m]

    # Partial covariance: C = R - AA'
    Cm <- R - Am %*% t(Am)

    d <- diag(Cm)
    # Guard against zero/negative residual variances (can happen at very high m or numerical issues)
    if (any(!is.finite(d)) || any(d <= 1e-5)) {
      break
    }
    # D^(-1/2) to standardize to correlation matrix
    Dm <- diag(1 / sqrt(d))

    Rstar <- Dm %*% Cm %*% Dm
    criteria[m + 1, c(2, 3)] <- map_from_trace_power(Rstar)

  }


  out <- list(
    eigenvalues = vals,
    n_factors_TR2 = ms[which.min(criteria[, 2])],
    n_factors_TR4 = ms[which.min(criteria[, 3])],
    criteria = criteria,
    settings = list(
      use = use,
      cor_method = cor_method
    )
  )

  class(out) <- "MAP"

  return(out)

}

#' Print SL object
#'
#' Print Method showing a summarized output of the \link{SL} function.
#'
#' @param x list. An object of class SL to be printed
#' @param ...  Further arguments for print.
#'
#' @export
#' @method print SL
#'
#' @examples
#' EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                type = "EFAtools", method = "PAF", rotation = "promax")
#' SL(EFA_mod, type = "EFAtools", method = "PAF")
#'
print.SL <- function(x, ...) {

  if(!all(is.na(x$settings))){
  # extract the settings for EFA not depending on the method
  method <- x$settings$method
  type <- x$settings$type

  # Settings intro message
  cat("\n")
  cat("EFA for second-order loadings performed with type = '",
      crayon::bold(type), "' and method = '", crayon::bold(method),
      "'", sep = "")
  cat("\n")

  if (!is.null(x$settings$max_iter) && x$iter > x$settings$max_iter) {
    cat("\n")
    cat(crayon::red$bold(cli::symbol$cross,
                         "Maximum number of iterations reached",
                         "without convergence"))
    cat("\n")
  }

  }

  # print the loadings and the variances
  cat("\n")
  cat(cli::rule(left = crayon::bold("Schmid-Leiman Solution"), col = "blue"))
  cat("\n")
  cat("\n")
  print(x$sl)
  cat("\n")
  cat("\n")
  cat(cli::rule(left = crayon::bold("Variances Accounted for"), col = "blue"))
  cat("\n")
  cat("\n")
  cat(.get_compare_matrix(x$vars_accounted, r_red = Inf, n_char = 17))

}

#' Print function for MAP objects
#'
#' @param x a list of class MAP. Output from \code{\link{MAP}} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print MAP
#'
#' @examples
#' MAP_base <- MAP(test_models$baseline$cormat)
#' MAP_base
#'
print.MAP <- function(x, plot = TRUE, ...) {

  n_factors_TR2 <- x$n_factors_TR2
  n_factors_TR4 <- x$n_factors_TR4

  cat(cli::rule("Number of factors suggested by MAP",
                col = "blue"))
  cat("\n")
  cat("\n")


    cat(crayon::blue(cli::symbol$bullet, "Original implementation (TR2): "),
        crayon::bold(n_factors_TR2))
    cat("\n")

    cat(crayon::blue(cli::symbol$bullet,
                     "Revised implementation (TR4): "),
        crayon::bold(n_factors_TR4))
    cat("\n")


}

#' Print SMT object
#'
#' @param x list of class SMT (output from the \link{SMT} function)
#' @param ... additional arguments passed to print
#'
#' @method print SMT
#'
#' @export
#'
#' @examples
#' SMT_base <- SMT(test_models$baseline$cormat, N = 500)
#' SMT_base
#'
print.SMT <- function(x, ...) {

  nfac_chi <- x$nfac_chi
  nfac_RMSEA <- x$nfac_RMSEA
  nfac_AIC <- x$nfac_AIC

  if(!is.na(nfac_chi)){

  cat("\n")
  cat("Sequential \U1D712\U00B2 Model Tests suggest ", crayon::bold(nfac_chi),
      " factor", ifelse(nfac_chi > 1 | nfac_chi == 0 | is.na(nfac_chi), "s.", "."),
      sep = "")
  cat("\n")
  cat("\n")

  }

  if(!is.na(nfac_RMSEA)){

  cat("Lower bound of RMSEA 90% confidence interval suggests ",
      crayon::bold(nfac_RMSEA), " factor",
      ifelse(nfac_RMSEA > 1 | nfac_RMSEA == 0 | is.na(nfac_RMSEA), "s.", "."),
      sep = "")
  cat("\n")
  cat("\n")

  }

  if(!is.na(nfac_AIC)){

  cat("AIC suggests ", crayon::bold(nfac_AIC), " factor",
      ifelse(nfac_AIC > 1 | nfac_AIC == 0 | is.na(nfac_AIC), "s.", "."),
      sep = "")
  cat("\n")
  cat("\n")

  }

}

#' Compare two vectors or matrices (communalities or loadings)
#'
#' The function takes two objects of the same dimensions containing numeric
#' information (loadings or communalities) and returns a list of class COMPARE
#' containing summary information of the differences of the objects.
#'
#' @param x matrix, or vector. Loadings or communalities of a factor
#'  analysis output.
#' @param y matrix, or vector. Loadings or communalities of another
#'  factor analysis output to compare to x.
#' @param reorder character. Whether and how elements / columns should be
#' reordered. If "congruence" (default), reordering is done according to Tuckers
#' correspondence coefficient, if "names", objects according to their names,
#' if "none", no reordering is done.
#' @param corres logical. Whether factor correspondences should be compared if a
#'  matrix is entered.
#' @param thresh numeric. The threshold to classify a pattern coefficient as substantial. Default is .3.
#' @param digits numeric. Number of decimals to print in the output. Default is 4.
#' @param m_red numeric. Number above which the mean and median should be printed
#'  in red (i.e., if .001 is used, the mean will be in red if it is larger than
#'  .001, otherwise it will be displayed in green.) Default is .001.
#' @param range_red numeric. Number above which the min and max should be printed
#'  in red (i.e., if .001 is used, min and max will be in red if the max is larger
#'   than .001, otherwise it will be displayed in green. Default is .001). Note that
#'   the color of min also depends on max, that is min will be displayed in the
#'   same color as max.
#' @param round_red  numeric. Number above which the max decimals to round to where
#' all corresponding elements of x and y are still equal are displayed in red
#' (i.e., if 3 is used, the number will be in red if it is smaller than
#'  3, otherwise it will be displayed in green). Default is 3.
#' @param print_diff logical. Whether the difference vector or matrix should be
#'  printed or not. Default is TRUE.
#' @param na.rm logical. Whether NAs should be removed in the mean, median, min,
#'  and max functions. Default is FALSE.
#' @param x_labels character. A vector of length two containing identifying
#'  labels for the two objects x and y that will be compared. These will be used
#'  as labels on the x-axis of the plot. Default is "x" and "y".
#' @param plot logical. If TRUE (default), a plot illustrating the differences
#'  will be shown.
#' @param plot_red numeric. Threshold above which to plot the absolute differences
#'  in red. Default is .001.
#'
#' @return A list of class COMPARE containing summary statistics on the differences
#'  of x and y.
#'
#' \item{diff}{The vector or matrix containing the differences between x and y.}
#' \item{mean_abs_diff}{The mean absolute difference between x and y.}
#' \item{median_abs_diff}{The median absolute difference between x and y.}
#' \item{min_abs_diff}{The minimum absolute difference between x and y.}
#' \item{max_abs_diff}{The maximum absolute difference between x and y.}
#' \item{max_dec}{The maximum number of decimals to which a comparison makes sense.
#'  For example, if x contains only values up to the third decimals, and y is a
#'  normal double, max_dec will be three.}
#' \item{are_equal}{The maximal number of decimals to which all elements of x and y
#'  are equal.}
#' \item{diff_corres}{The number of differing variable-to-factor correspondences
#'  between x and y, when only the highest loading is considered.}
#' \item{diff_corres_cross}{The number of differing variable-to-factor correspondences
#'  between x and y when all loadings \code{>= thresh} are considered.}
#' \item{g}{The root mean squared distance (RMSE) between x and y.}
#' \item{settings}{List of the settings used.}
#'
#' @export
#'
#' @examples
#' # A type SPSS EFA to mimick the SPSS implementation
#' EFA_SPSS_6 <- EFA(test_models$case_11b$cormat, n_factors = 6, type = "SPSS")
#'
#' # A type psych EFA to mimick the psych::fa() implementation
#' EFA_psych_6 <- EFA(test_models$case_11b$cormat, n_factors = 6, type = "psych")
#'
#' # compare the two
#' COMPARE(EFA_SPSS_6$unrot_loadings, EFA_psych_6$unrot_loadings,
#'         x_labels = c("SPSS", "psych"))
COMPARE <- function(x,
                    y,
                    reorder = c("congruence", "names", "none"),
                    corres = TRUE,
                    thresh = .3,
                    digits = 4,
                    m_red = .001,
                    range_red = .001,
                    round_red = 3,
                    print_diff = TRUE,
                    na.rm = FALSE,
                    x_labels = c("x", "y"),
                    plot = TRUE,
                    plot_red = .01)  {

  reorder <- match.arg(reorder)
  checkmate::assert_flag(corres)
  checkmate::assert_number(thresh)
  checkmate::assert_count(digits)
  checkmate::assert_number(m_red)
  checkmate::assert_number(range_red)
  checkmate::assert_number(round_red)
  checkmate::assert_flag(print_diff)
  checkmate::assert_flag(na.rm)
  checkmate::assert_character(x_labels, len = 2)
  checkmate::assert_flag(plot)
  checkmate::assert_number(plot_red)

  # reclass data.frames and tibbles to matrices so the stats functions afterwards
  # work
  if ((inherits(x, c("loadings", "LOADINGS", "SLLOADINGS", "matrix"))) &&
      (inherits(y, c("loadings", "LOADINGS", "SLLOADINGS", "matrix")))) {

    if (inherits(x, c("loadings", "LOADINGS", "SLLOADINGS"))) {
      x <- unclass(x)
    }

    if (inherits(y, c("loadings", "LOADINGS", "SLLOADINGS"))) {
      y <- unclass(y)
    }

    # check if dimensions match:
    if (any(dim(x) != dim(y))) {

      stop(crayon::red$bold(cli::symbol$circle_cross),
           crayon::red(" 'x' and 'y' have different dimensions. Can only compare matrices with identical dimensions.\n"))

    }

  } else if (inherits(x, c("numeric", "integer")) &&
             inherits(y, c("numeric", "integer"))) {

    if (length(x) != length(y)) {

      stop(crayon::red$bold(cli::symbol$circle_cross),
           crayon::red(" 'x' and 'y' have different lengths Compare only works with identical dimensions.\n"))

    }

  } else {

    stop(crayon::red$bold(cli::symbol$circle_cross),
         crayon::red(" 'x' is of class", class(x), "and 'y' is of class",
                     class(y), "but must be numeric vectors or matrices\n"))

  }

  if (inherits(x, "matrix")) {

    n_factors <- ncol(x)

    if (reorder == "congruence" && n_factors > 1) {
      # get Tucker's congruence coefficients
      congruence <- .factor_congruence(x, y)

      if (any(is.na(congruence))) {
        stop(crayon::red$bold(cli::symbol$circle_cross),
             crayon::red(" Tucker's congruence coefficients contained NAs, cannot reorder columns based on congruence. Try another reordering method.\n"))
      }

      # factor order for y
      factor_order <- apply(abs(congruence), 1, which.max)

      # obtain signs to reflect signs of y if necessary
      factor_sign <- sapply(seq_len(n_factors),
                            function(ll, congruence, factor_order){
                              sign(congruence[ll, factor_order[ll]])
                              }, congruence = congruence,
                            factor_order = factor_order)

      factor_sign <- rep(factor_sign, each = nrow(x))

      # reorder
      y <- y[, factor_order]

      # reflect signs if necessary
      y <- y * factor_sign
    } else if (reorder == "names" && n_factors > 1) {

      if (!is.null(colnames(x)) && !is.null(colnames(y))) {
        x <- x[, order(colnames(x))]
        y <- y[, order(colnames(y))]

        if(!all(colnames(x) == colnames(y))) {
          warning(crayon::yellow$bold("!"), crayon::yellow(" reorder = 'names' was used but colnames of x and y were not identical. Results might be inaccurate.\n"))
        }
      } else if (is.null(colnames(x)) || is.null(colnames(y))) {
        warning(crayon::yellow$bold("!"), crayon::yellow(" reorder was set to 'names' but at least one of 'x' and 'y' was not named. Proceeding without reordering.\n"))
      }

    }

    if (n_factors > 1 && isTRUE(corres)) {
      # factor correspondences
      corres_list <- .factor_corres(x, y, thresh = thresh)
      diff_corres <- corres_list$diff_corres
      diff_corres_cross <- corres_list$diff_corres_cross
    } else {
      diff_corres <- 0
      diff_corres_cross <- 0
    }


  } else if (inherits(x, c("numeric", "integer"))) {

    if (reorder == "congruence" && !is.null(names(x)) && !is.null(names(y))){

      warning(crayon::yellow$bold("!"), crayon::yellow(" reorder was set to 'congruence', but this only works for matrices. To reorder vectors, set reorder = 'names'. Proceeding without reordering.\n"))

    } else if (reorder == "names") {

      if (!is.null(names(x)) && !is.null(names(y))) {

        x <- x[order(names(x))]
        y <- y[order(names(y))]

        if (!all(names(x) == names(y))) {
          warning(crayon::yellow$bold("!"), crayon::yellow(" reorder = 'names' was used but names of x and y were not identical. Results might be inaccurate.\n"))
        }



      } else if (is.null(names(x)) || is.null(names(y))) {
        warning(crayon::yellow$bold("!"), crayon::yellow(" reorder was set to 'names' but at least one of 'x' and 'y' was not named. Proceeding without reordering.\n"))
      }

      }

      diff_corres <- NA
      diff_corres_cross <- NA

      g <- NA

  }

  # compute differences and statistics
  diff <- x - y

  if(inherits(x, "matrix")) {
    g <- sqrt(sum(diag(t(diff) %*% (diff))) / prod(dim(x)))
  } else {
    g <- sqrt((sum(diff ** 2) / length(diff)))
  }


  mean_abs_diff <- mean(abs(diff), na.rm = na.rm)
  median_abs_diff <- stats::median(abs(diff), na.rm = na.rm)

  min_abs_diff <- min(abs(diff), na.rm = na.rm)
  max_abs_diff <- max(abs(diff), na.rm = na.rm)

  are_equal_v <- c()

  max_dec <- min(c(.decimals(x), .decimals(y)))

  class(x) <- "character"
  class(y) <- "character"
  x <- gsub("-|\\.", "", x)
  y <- gsub("-|\\.", "", y)
  for (ii in seq_len((max_dec + 1))) {
    are_equal_v[ii] <- all(substr(x, 1, ii) == substr(y, 1, ii))
  }

  are_equal <- utils::tail(which(are_equal_v), 1)

  if(length(are_equal) == 0){
    are_equal <- 0
  } else {
     are_equal <- are_equal - 1
  }

  settings <- list(
    reorder = reorder,
    corres = corres,
    digits = digits,
    thresh = thresh,
    m_red = m_red,
    range_red = range_red,
    round_red = round_red,
    print_diff = print_diff,
    na.rm = na.rm,
    x_labels = x_labels,
    plot = plot,
    plot_red = plot_red
  )



  # create output list
  out <- list(
    diff = diff,
    mean_abs_diff = mean_abs_diff,
    median_abs_diff = median_abs_diff,
    min_abs_diff = min_abs_diff,
    max_abs_diff = max_abs_diff,
    max_dec = max_dec,
    are_equal = are_equal,
    diff_corres = diff_corres,
    diff_corres_cross = diff_corres_cross,
    g = g,
    settings = settings
  )

  class(out) <- "COMPARE"

  out
}

#' Comparison Data
#'
#' Factor retention method introduced by Ruscio and Roche (2012). The code was
#' adapted from the CD code by Auerswald and Moshagen (2017) available at
#' \url{https://osf.io/x5cz2/?view_only=d03efba1fd0f4c849a87db82e6705668}
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data.
#' @param n_factors_max numeric. The maximum number of factors to test against.
#'  Larger numbers will increase the duration the procedure takes, but test more
#'  possible solutions. If left NA (default) the maximum number of factors for
#'  which the model is still over-identified (df > 0) is used.
#' @param N_pop numeric. Size of finite populations of comparison data. Default
#'  is 10000.
#' @param N_samples numeric. Number of samples drawn from each population.
#'  Default is 500.
#' @param alpha numeric. The alpha level used to test the significance of the
#'  improvement added by an additional factor. Default is .30.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}}. Default
#'  is "pairwise.complete.obs". However, for the comparison data procedure,
#'  \code{NA} values will be excluded using na.omit(). If missing data should
#'  be handled differently (e.g., imputation), do this before passing the data to
#'  \code{CD()}.
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#' @param max_iter numeric. The maximum number of iterations to perform after
#'  which the iterative PAF procedure is halted. Default is 50.
#'
#' @details "Parallel analysis (PA) is an effective stopping rule that compares
#' the eigenvalues of randomly generated data with those for the actual data.
#' PA takes into account sampling error, and at present it is widely considered
#' the best available method. We introduce a variant of PA that goes even further
#' by reproducing the observed correlation matrix rather than generating random
#' data. Comparison data (CD) with known factorial structure are first generated
#' using 1 factor, and then the number of factors is increased until the
#' reproduction of the observed eigenvalues fails to improve significantly"
#' (Ruscio & Roche, 2012, p. 282).
#'
#' The CD implementation here is based on the code by Ruscio and Roche (2012), but
#' is slightly adapted to increase speed by performing the principal axis factoring
#' using a C++ based function.
#'
#' Note that if the data contains missing values, these will be removed for the
#' comparison data procedure using \code{\link[stats:na.fail]{stats::na.omit}}. If
#' missing data should be treated differently, e.g., by imputation, do this outside
#' \code{CD} and then pass the complete data.
#'
#' The \code{CD} function can also be called together with other factor retention
#' criteria in the \code{\link{N_FACTORS}} function.
#'
#' @return A list of class CD containing
#'
#' \item{n_factors}{The number of factors to retain according to comparison data results.}
#' \item{eigenvalues}{A vector containing the eigenvalues of the entered data.}
#' \item{RMSE_eigenvalues}{A matrix containing the RMSEs between the eigenvalues of the generated data and those of the entered data.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Ruscio, J., & Roche, B. (2012). Determining the number of factors to
#' retain in an exploratory factor analysis using comparison data of known
#' factorial structure. Psychological Assessment, 24, 282–292.
#' doi: 10.1037/a0025697
#'
#' @seealso Other factor retention criteria: \code{\link{EKC}},
#'  \code{\link{HULL}}, \code{\link{KGC}}, \code{\link{PARALLEL}}, \code{\link{SMT}}
#'
#'   \code{\link{N_FACTORS}} as a wrapper function for this and all
#'   the above-mentioned factor retention criteria.
#'
#' @export
#'
#'
#' @examples
#' \donttest{
#' # determine n factors of the GRiPS
#' CD(GRiPS_raw)
#'
#' # determine n factors of the DOSPERT risk subscale
#' CD(DOSPERT_raw)
#'}
CD <- function(x, n_factors_max = NA, N_pop = 10000, N_samples = 500, alpha = .30,
               use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                    "everything", "na.or.complete"),
               cor_method = c("pearson", "spearman", "kendall"),
               max_iter = 50) {

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Provide a dataframe or matrix with raw data.\n"))

  }

  if (.is_cormat(x)) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is a correlation matrix, but CD only works with raw data.\n"))
  }

  if (inherits(x, c("tbl_df", "tbl"))) {
    x <- as.data.frame(x)
  }

  use <- match.arg(use)
  cor_method <- match.arg(cor_method)

  checkmate::assert_count(n_factors_max, na.ok = TRUE)
  checkmate::assert_count(N_pop)
  checkmate::assert_count(N_samples)
  checkmate::assert_number(alpha, lower = 0, upper = 1)
  checkmate::assert_count(max_iter)

  # Create correlation matrix
  R <- stats::cor(x, use = use, method = cor_method)
  colnames(R) <- colnames(x)

  if (any(is.na(x))) {
    n_row_complete <- nrow(x)
    x <- stats::na.omit(x)
    n_row_new <- nrow(x)
    n_rows_removed <- n_row_complete - n_row_new

    warning(crayon::yellow$bold("!"),
            crayon::yellow(" The data contained missing values that were removed using stats::na.omit().", n_rows_removed, "row(s) were removed.\n"))
  }
  n_cases <- nrow(x)
  k <- ncol(x)

  m_possible <- .det_max_factors(k)

  if (is.na(n_factors_max) || n_factors_max > m_possible) {

    if (!is.na(n_factors_max) & n_factors_max > m_possible) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" n_factors_max was set to",
              n_factors_max, "but maximum possible",
              "factors to extract is", m_possible, ". Setting n_factors_max to",
              m_possible, ".\n"))
    }

    n_factors_max <- m_possible

  }

  eigvals_real <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  # initialize objects for iterative procedures
  RMSE_eigvals <- matrix(0, nrow = N_samples, ncol = n_factors_max)
  sig <- TRUE
  n_factors <- 1



  while (n_factors <= n_factors_max && isTRUE(sig)) {

    pop <- .gen_data(x, cor_method = cor_method, use, n_factors, N_pop,
                     max_iter = max_iter)

    for (j in 1:N_samples) {

      samp <- pop[sample(1:N_pop, size = n_cases, replace = TRUE),]
      R_samp <- stats::cor(samp, method = cor_method)
      eigvals_samp <- eigen(R_samp, symmetric = TRUE, only.values = TRUE)$values
      RMSE_eigvals[j,n_factors] <- sqrt(sum((eigvals_samp - eigvals_real) *
                                         (eigvals_samp - eigvals_real)) / k)
    }

    if (n_factors > 1) {

      sig <- (stats::wilcox.test(RMSE_eigvals[,n_factors],
                          RMSE_eigvals[,(n_factors - 1)], "less")$p.value < alpha)
    }
    if (isTRUE(sig)) {
      n_factors <- n_factors + 1
    }

  }

  n_factors <- n_factors - 1

  settings <- list(
    n_factors_max = n_factors_max,
    N_pop = N_pop,
    N_samples = N_samples,
    alpha = alpha,
    use = use,
    cor_method = cor_method,
    max_iter = max_iter
  )

  out <- list(
    n_factors = n_factors,
    eigenvalues = eigvals_real,
    RMSE_eigenvalues = RMSE_eigvals,
    settings = settings
  )

  class(out) <- "CD"

  return(out)

}


.gen_data <- function(x, cor_method, use, n_factors, N, max_trials = 5,
                      initial_multiplier = 1, max_iter = 100) {
  # Steps refer to description in the following article:
  # Ruscio, J., & Kaczetow, W. (2008). Simulating multivariate nonnormal data using an iterative algorithm.
  # Multivariate Behavioral Research, 43(3), 355-381.

  # Initialize variables and (if applicable) set random number seed (step 1) -------------------------------------

  k <- ncol(x)
  sim_dat <- matrix(0, nrow = N, ncol = k)         # Matrix to store the simulated data
  dists <- matrix(0, nrow = N, ncol = k)   # Matrix to store each variable's score distribution
  iter <- 0                                   # Iteration counter
  best_RMSE <- 1                                   # Lowest RMSE correlation
  t_no_impr <- 0                  # Trial counter

  # Generate distribution for each variable (step 2) -------------------------------------------------------------

  for (i in 1:k) {
    dists[,i] <- sort(sample(x[,i], size = N, replace = TRUE))
  }


  # Calculate and store a copy of the target correlation matrix (step 3) -----------------------------------------

  R <- stats::cor(x, method = cor_method, use = use)
  R_inter <- R

  # Generate random normal data for shared and unique components, initialize factor loadings (steps 5, 6) --------

  shared_comp <- matrix(stats::rnorm(N * n_factors, 0, 1), nrow = N,
                        ncol = n_factors)
  unique_comp <- matrix(stats::rnorm(N * k, 0, 1), nrow = N, ncol = k)
  shared_load <- matrix(0, nrow = k, ncol = n_factors)
  unique_load <- matrix(0, nrow = k, ncol = 1)

  # Begin loop that ends when specified number of iterations pass without improvement in RMSE correlation --------

  while (t_no_impr < max_trials) {
    iter <- iter + 1

    # Calculate factor loadings and apply to reproduce desired correlations (steps 7, 8) ---------------------------

    L <- suppressWarnings(.paf_iter(rep(1, k), criterion = .001, R = R_inter,
                   n_fac = n_factors, abs_eig = TRUE, crit_type = 2,
                   max_iter = max_iter)$L)

    shared_load[,1:n_factors] <- L

    # get rid of Heywood cases
    shared_load[shared_load > 1] <- 1
    shared_load[shared_load < -1] <- -1

    if (shared_load[1, 1] < 0) {
      shared_load <- shared_load * -1
    }

    for (i in seq_len(k)) {
      if (sum(shared_load[i,] * shared_load[i,]) < 1) {
        unique_load[i, 1] <-
          (1 - sum(shared_load[i,] * shared_load[i,]))
      } else {
        unique_load[i, 1] <- 0
      }
    }

    unique_load <- sqrt(unique_load)

    for (i in seq_len(k)) {
      sim_dat[, i] <- (shared_comp %*% t(shared_load))[, i] +
        unique_comp[, i] * unique_load[i, 1]
    }


      # Replace normal with nonnormal distributions (step 9) ---------------------------------------------------------

      for (i in seq_len(k)) {
        sim_dat <- sim_dat[sort.list(sim_dat[, i]),]
        sim_dat[,i] <- dists[, i]
      }

      # Calculate RMSE correlation, compare to lowest value, take appropriate action (steps 10, 11, 12) --------------

      R_rep <- stats::cor(sim_dat, method = cor_method)
      R_res <- R - R_rep
      # check whether this also works:
      # sqrt(sum(diag(t(R_res) %*% (R_res))) / prod(dim(R)))
      RMSE <- sqrt(sum(R_res[lower.tri(R_res)] * R_res[lower.tri(R_res)]) /
                     (.5 * (k * k - k)))

      if (RMSE < best_RMSE) {
        best_RMSE <- RMSE
        R_best <- R_inter
        res_best <- R_res
        R_inter <- R_inter + initial_multiplier * R_res
        t_no_impr <- 0
      } else {
        t_no_impr <- t_no_impr + 1
        current_multiplier <- initial_multiplier * .5 ^ t_no_impr
        R_inter <- R_best + current_multiplier * res_best
      }
  }

  # Construct the data set with the lowest RMSE correlation (step 13) --------------------------------------------

  L <- suppressWarnings(.paf_iter(rep(1, k), criterion = .001, R = R_best,
                n_fac = n_factors, abs_eig = TRUE, crit_type = 2,
                max_iter = max_iter)$L)
  shared_load[, seq_len(n_factors)] <- L

  shared_load[shared_load > 1] <- 1
  shared_load[shared_load < -1] <- -1
  if (shared_load[1, 1] < 0) {
    shared_load <- shared_load * -1
  }

  for (i in seq_len(k)) {
    if (sum(shared_load[i,] * shared_load[i,]) < 1) {
      unique_load[i, 1] <-
        (1 - sum(shared_load[i,] * shared_load[i,]))
    } else {
      unique_load[i, 1] <- 0
    }
  }

  unique_load <- sqrt(unique_load)

  for (i in seq_len(k)) {
    sim_dat[, i] <- (shared_comp %*% t(shared_load))[, i] +
      unique_comp[, i] * unique_load[i, 1]
  }

  sim_dat <- apply(sim_dat, 2, scale) # standardizes each variable in the matrix
  for (i in seq_len(k)) {
    sim_dat <- sim_dat[sort.list(sim_dat[, i]),]
    sim_dat[,i] <- dists[, i]
  }

  # Return the simulated data set (step 14) ----------------------------------------------------------------------

  return(sim_dat)
}

#' Plot PARALLEL object
#'
#' Plot method showing a summarized output of the \link{PARALLEL} function
#'
#' @param x list of class PARALLEL. An output from the \link{PARALLEL} function.
#' @param ... not used.
#'
#' @export
#' @method plot PARALLEL
#'
#' @examples
#' \donttest{
#' # example with correlation matrix and "ML" estimation
#' x <- PARALLEL(test_models$case_11b$cormat, N = 500, method = "ML")
#' plot(x)
#' }
plot.PARALLEL <- function(x, ...) {

  n_vars <- x$settings$n_vars
  eigen_type <- x$settings$eigen_type
  eigen_PCA <- x$eigenvalues_PCA
  eigen_SMC <- x$eigenvalues_SMC
  eigen_EFA <- x$eigenvalues_EFA
  n_fac_PCA <- x$n_fac_PCA
  n_fac_SMC <- x$n_fac_SMC
  n_fac_EFA <- x$n_fac_EFA
  x_dat <- x$settings$x_dat
  decision_rule <- x$settings$decision_rule
  percent <- x$settings$percent

  # Create plots depending on eigen_type and if real data were entered or not
  if("PCA" %in% eigen_type){

    .plot_PA_helper(eigenvalues = eigen_PCA, n_vars = n_vars, n_fac = n_fac_PCA,
                    x_dat = x_dat, decision_rule = decision_rule,
                    percent = percent, eigen_type = "PCA")

  }

  if("SMC" %in% eigen_type){

    .plot_PA_helper(eigenvalues = eigen_SMC, n_vars = n_vars, n_fac = n_fac_SMC,
                    x_dat = x_dat, decision_rule = decision_rule,
                    percent = percent, eigen_type = "SMC")

  }

  if("EFA" %in% eigen_type){

    .plot_PA_helper(eigenvalues = eigen_EFA, n_vars = n_vars, n_fac = n_fac_EFA,
                    x_dat = x_dat, decision_rule = decision_rule,
                    percent = percent, eigen_type = "EFA")

  }

}

.plot_PA_helper <- function(eigenvalues, n_vars, n_fac, x_dat,
                            decision_rule = decision_rule, percent = percent,
                            eigen_type){

  p_eigen <- pretty(c(min(eigenvalues) * .9, eigenvalues,
                      max(eigenvalues) * 1.15))
  graphics::plot.new()
  graphics::plot.window(xlim = c(1, n_vars),
                        ylim = c(min(p_eigen), max(p_eigen)))
  graphics::axis(1, seq_len(n_vars))
  graphics::axis(2, p_eigen, las = 1)
  graphics::mtext("Factor", side = 1, line = 3, cex = 1.5, padj =-.5)
  graphics::mtext("Eigenvalues", side = 2, line = 3, cex = 1.5, padj =.5)

  if (isTRUE(x_dat)) {
    graphics::lines(seq_len(n_vars), eigenvalues[,"Real Eigenvalues"])
    graphics::points(seq_len(n_vars), eigenvalues[,"Real Eigenvalues"], pch = 16)
    if (!is.na(n_fac)) {
      graphics::points(n_fac, eigenvalues[n_fac,"Real Eigenvalues"],
                       pch = 1, cex = 2, col = "red")
      graphics::text(n_fac, eigenvalues[n_fac,"Real Eigenvalues"],
                     n_fac, pos = 3, cex = 1.5, col = "red",
                     font = 1, offset = .75)
    }

  }

  cols <- viridisLite::viridis(ncol(eigenvalues) - x_dat, end = .8)
  names(cols) <- colnames(eigenvalues)[(as.numeric(x_dat) + 1):ncol(eigenvalues)]

  graphics::lines(seq_len(n_vars), eigenvalues[,"Means"], lty = 2, lwd = 1.25,
                  col = cols[1])

  for (perc_i in percent) {
    graphics::lines(seq_len(n_vars), eigenvalues[,paste(perc_i, "Percentile")],
                    lty = 2, col = cols[paste(perc_i, "Percentile")], lwd = 1.25)
  }

  factors_text <- paste0("N Factors with Decision Rule '", decision_rule,
                         "' and Eigen Type '", eigen_type, "': ", n_fac)

  graphics::title(factors_text)

  graphics::legend("topright",
                   colnames(eigenvalues), lty = c(1, rep(2, length(cols))),
                   col = c("black", cols))


}

#' Kaiser-Meyer-Olkin criterion
#'
#' This function computes the Kaiser-Meyer-Olkin (KMO) criterion overall and for
#' each variable in a correlation matrix. The KMO represents the degree to
#' which each observed variable is predicted by the other variables in the
#' dataset and with this indicates the suitability for factor analysis.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#'  correlations.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#'  data is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#'
#' @details Kaiser (1970) proposed this index, originally called measure of
#' sampling adequacy (MSA), that indicates how near the inverted correlation
#' matrix \eqn{R^{-1}} is to a diagonal matrix \eqn{S} to determine a given
#' correlation matrix's (\eqn{R}) suitability for factor analysis.
#' The index is
#' \deqn{KMO = \frac{\sum\limits_{i<j}\sum r_{ij}^2}{\sum\limits_{i<j}\sum r_{ij}^2 + \sum\limits_{i<j}\sum q_{ij}^2}}
#' with \eqn{Q = SR^{-1}S} and S = \eqn{(diag R^{-1})^{-1/2}} where
#' \eqn{\sum\limits_{i<j}\sum r_{ij}^2} is the sum of squares of the upper
#' off-diagonal elements of \eqn{R} and \eqn{\sum\limits_{i<j}\sum q_{ij}^2} is the
#' sum of squares of the upper off-diagonal elements of \eqn{Q} (see also Cureton & D'Augustino, 1983).
#'
#' So KMO varies between 0 and 1, with larger values indicating higher suitability
#' for factor analysis. Kaiser and Rice (1974) suggest that KMO should at least
#' exceed .50 for a correlation matrix to be suitable for factor analysis.
#'
#' This function was heavily influenced by the \code{\link[psych:KMO]{psych::KMO}}
#' function.
#'
#' See also \code{\link{BARTLETT}} for another test of suitability for factor
#' analysis.
#'
#' The \code{KMO} function can also be called together with the
#' \code{\link{BARTLETT}} function and with factor retention criteria in the
#'  \code{\link{N_FACTORS}} function.
#'
#' @return A list containing
#' \item{KMO}{Overall KMO.}
#' \item{KMO_i}{KMO for each variable.}
#' \item{settings}{A list of the settings used.}
#'
#' @export
#'
#' @source Kaiser, H. F. (1970). A second generation little jiffy. Psychometrika,
#' 35, 401-415.
#' @source Kaiser, H. F. & Rice, J. (1974). Little jiffy, mark IV. Educational
#' and Psychological Measurement, 34, 111-117.
#' @source Cureton, E. E. & D'Augustino, R. B. (1983). Factor analysis: An
#'  applied approach. Hillsdale, N.J.: Lawrence Erlbaum Associates, Inc.
#'
#' @seealso \code{\link{BARTLETT}} for another measure to determine
#' suitability for factor analysis.
#'
#' \code{\link{N_FACTORS}} as a wrapper function for this function,
#' \code{\link{BARTLETT}} and several factor retention criteria.
#'
#' @examples
#' KMO(test_models$baseline$cormat)
KMO <- function(x, use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                           "everything", "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall")) {

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  use <- match.arg(use)
  cor_method <- match.arg(cor_method)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed.\n"))
  }

  # Check if correlation matrix is positive definite
  if(any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)){

    R <- psych::cor.smooth(R)
    R_i <- try(solve(R), silent = TRUE)

  }

  # Start computations
  S <- diag(diag(R_i)^(-1/2))
  Q <- S %*% R_i %*% S
  diag(Q) <- 0
  diag(R) <- 0
  sumQ2 <- sum(Q^2)
  sumR2 <- sum(R^2)
  KMO <- sumR2/(sumR2 + sumQ2)
  KMO_i <- colSums(R^2)/(colSums(R^2) + colSums(Q^2))

  if(!is.null(colnames(R))){

    names(KMO_i) <- colnames(R)

  } else if(!is.null(rownames(R))) {

    names(KMO_i) <- rownames(R)

  } else {

    names(KMO_i) <- paste0("V", seq_len(ncol(R)))

  }

  # Prepare settings
  settings <- list(use = use,
                   cor_method = cor_method)

  output <- list(KMO = KMO,
                 KMO_i = KMO_i,
                 settings = settings)

  class(output) <- "KMO"

  return(output)

}

kmo_cor <- KMO(test_models$baseline$cormat)
kmo_raw <- KMO(GRiPS_raw)
dat_nonames <- test_models$baseline$cormat
colnames(dat_nonames) <- NULL
kmo_nona <- KMO(dat_nonames)

test_that("output class and dimensions are correct", {
  expect_is(kmo_cor, "KMO")
  expect_output(str(kmo_cor), "List of 3")
  expect_is(kmo_raw, "KMO")
  expect_output(str(kmo_raw), "List of 3")
  expect_is(kmo_nona, "KMO")
  expect_output(str(kmo_nona), "List of 3")
})

test_that("KMO values are correct", {
  expect_equal(kmo_cor$KMO, 0.916, tolerance = 1e-3)
  expect_equal(kmo_raw$KMO, 0.955, tolerance = 1e-3)
  expect_equal(kmo_nona$KMO, 0.916, tolerance = 1e-3)

  expect_length(kmo_cor$KMO_i, ncol(test_models$baseline$cormat))
  expect_length(kmo_raw$KMO_i, ncol(GRiPS_raw))
  expect_length(kmo_nona$KMO_i, ncol(dat_nonames))
})

test_that("settings are returned correctly", {
  expect_named(kmo_cor$settings, c("use", "cor_method"))
  expect_named(kmo_raw$settings, c("use", "cor_method"))
  expect_named(kmo_nona$settings, c("use", "cor_method"))

  expect_equal(kmo_cor$settings$use, "pairwise.complete.obs")
  expect_equal(kmo_raw$settings$use, "pairwise.complete.obs")
  expect_equal(kmo_nona$settings$use, "pairwise.complete.obs")

  expect_equal(kmo_cor$settings$cor_method, "pearson")
  expect_equal(kmo_raw$settings$cor_method, "pearson")
  expect_equal(kmo_nona$settings$cor_method, "pearson")
})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z), ncol = 3)
cor_sing <- stats::cor(dat_sing)

cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

test_that("errors are thrown correctly", {
  expect_error(KMO(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_message(KMO(GRiPS_raw), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_error(KMO(dat_sing), " Correlation matrix is singular, no further analyses are performed.\n")
  expect_error(KMO(cor_sing), " Correlation matrix is singular, no further analyses are performed.\n")
  expect_warning(KMO(cor_nposdef), "Matrix was not positive definite, smoothing was done")
})

rm(kmo_cor, kmo_raw, dat_nonames, kmo_nona, x, y, z, dat_sing, cor_sing, cor_nposdef)

kmo_cor <- efa_kmo(test_models$baseline$cormat)
kmo_raw <- efa_kmo(GRiPS_raw)
dat_nonames <- test_models$baseline$cormat
colnames(dat_nonames) <- NULL
kmo_nona <- efa_kmo(dat_nonames)

test_that("output class and dimensions are correct", {
  expect_identical(class(kmo_cor), c("efa_kmo", "KMO"))
  expect_output(str(kmo_cor), "List of 3")
  expect_identical(class(kmo_raw), c("efa_kmo", "KMO"))
  expect_output(str(kmo_raw), "List of 3")
  expect_identical(class(kmo_nona), c("efa_kmo", "KMO"))
  expect_output(str(kmo_nona), "List of 3")
})

test_that("KMO values are correct", {
  expect_equal(kmo_cor$KMO, 0.916, tolerance = 1e-3)
  expect_equal(kmo_raw$KMO, 0.955, tolerance = 1e-3)
  expect_equal(kmo_nona$KMO, 0.916, tolerance = 1e-3)

  expect_length(kmo_cor$KMO_i, ncol(test_models$baseline$cormat))
  expect_length(kmo_raw$KMO_i, ncol(GRiPS_raw))
  expect_length(kmo_nona$KMO_i, ncol(dat_nonames))
})

test_that("per-variable and overall KMO match psych", {
  skip_on_cran()
  skip_if_not_installed("psych")

  psych_cor <- psych::KMO(test_models$baseline$cormat)
  expect_equal(unname(kmo_cor$KMO_i), unname(psych_cor$MSAi), tolerance = 1e-4)
  expect_equal(kmo_cor$KMO, unname(psych_cor$MSA), tolerance = 1e-4)

  expect_equal(unname(kmo_raw$KMO_i),
               unname(psych::KMO(stats::cor(GRiPS_raw))$MSAi),
               tolerance = 1e-4)
})

test_that("settings are returned correctly", {
  expect_named(kmo_cor$settings, c("use", "cor_method"))
  expect_named(kmo_raw$settings, c("use", "cor_method"))
  expect_named(kmo_nona$settings, c("use", "cor_method"))

  expect_equal(kmo_cor$settings$use, "pairwise.complete.obs")
  expect_equal(kmo_raw$settings$use, "pairwise.complete.obs")
  expect_equal(kmo_nona$settings$use, "pairwise.complete.obs")

  expect_equal(kmo_cor$settings$cor_method, "pearson")
  expect_equal(kmo_raw$settings$cor_method, "pearson")
  expect_equal(kmo_nona$settings$cor_method, "pearson")
})

test_that("errors are thrown correctly", {
  expect_error(efa_kmo(1:5), class = "efa_input_not_matrix")
  expect_message(efa_kmo(GRiPS_raw), class = "efa_cor_from_data")
  expect_error(efa_kmo(sing_raw), class = "efa_cor_singular")
  expect_error(efa_kmo(sing_cor), class = "efa_cor_singular")
  expect_warning(efa_kmo(cor_nposdef), class = "efa_cor_smoothed")
})

test_that("print output is stable", {
  local_reproducible_output()

  # high KMO: tick, verdict, and the per-variable values
  expect_snapshot(print(kmo_cor), transform = scrub_num)

  # low KMO: cross and the "not suitable" verdict
  kmo_low <- structure(list(KMO = 0.45, KMO_i = c(V1 = 0.40, V2 = 0.50, V3 = 0.45)),
                       class = c("efa_kmo", "KMO"))
  expect_snapshot(print(kmo_low), transform = scrub_num)

  # KMO value not available
  kmo_na <- structure(list(KMO = NA_real_, KMO_i = NA_real_),
                      class = c("efa_kmo", "KMO"))
  expect_snapshot(print(kmo_na), transform = scrub_num)
})

rm(kmo_cor, kmo_raw, dat_nonames, kmo_nona)

#' Model averaging across different EFA methods and types
#'
#' Not all EFA procedures always arrive at the same solution. This function allows
#' you perform a number of EFAs from different methods (e.g., Maximum Likelihood
#' and Principal Axis Factoring), with different implementations (e.g., the SPSS
#' and psych implementations of Principal Axis Factoring), and across different
#' rotations of the same type (e.g., multiple oblique rotations, like promax and
#' oblimin). EFA_AVERAGE will then run all these EFAs (using the \code{\link{EFA}}
#' function) and provide a summary across the different solutions.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations. If raw data is entered, the correlation matrix is found from the
#' data.
#' @param n_factors numeric. Number of factors to extract.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used. If input is a correlation matrix and \code{N} = NA
#' (default), not all fit indices can be computed.
#' @param method character vector. Any combination of  "PAF", "ML", and "ULS",
#' to use principal axis factoring, maximum likelihood, or unweighted least
#' squares (also called minres), respectively, to fit the EFAs. Default is "PAF".
#' @param rotation character vector. Either perform no rotation ("none"),
#' any combination of orthogonal rotations ("varimax", "equamax", "quartimax", "geominT",
#' "bentlerT", and "bifactorT"; using "orthogonal" runs all of these), or of
#' oblique rotations ("promax", "oblimin", "quartimin", "simplimax", "bentlerQ",
#' "geominQ", and "bifactorQ"; using "oblique" runs all of these). Rotation types
#' (no rotation, orthogonal rotations, and oblique rotations) cannot be mixed.
#' Default is "promax".
#' @param type character vector. Any combination of "none" (default), "EFAtools",
#' "psych", and "SPSS" can be entered. "none" allows the specification of various
#' combinations of the arguments controlling both factor extraction methods and
#' the rotations. The others ("EFAtools", "psych", and "SPSS"), control the execution
#' of the respective factor extraction method and rotation to be in line with how
#' it is executed in this package (i.e., the respective default procedure), in the
#' psych package, and in SPSS. A specific psych implementation exists for PAF, ML, varimax,
#' and promax. The SPSS implementation exists for PAF, varimax, and promax. For
#' details, see \code{\link{EFA}}.
#' @param averaging character. One of "mean" (default), and "median". Controls
#' whether the different results should be averaged using the (trimmed) mean,
#' or the median.
#' @param trim numeric. If averaging is set to "mean", this argument controls
#' the trimming of extremes (for details see \code{\link[base:mean]{base::mean}}).
#' By default no trimming is done (i.e., trim = 0).
#' @param salience_threshold numeric. The threshold to use to classify a pattern
#' coefficient or loading as salient (i.e., substantial enough to assign it to
#' a factor). Default is 0.3. Indicator-to-factor correspondences will be inferred
#' based on this threshold. Note that this may not be meaningful if rotation = "none"
#' and n_factors > 1 are used, as no simple structure is present there.
#' @param max_iter numeric. The maximum number of iterations to perform after which
#' the iterative PAF procedure is halted with a warning. Default is 10,000. Note
#' that non-converged procedures are excluded from the averaging procedure.
#' @param init_comm character vector. Any combination of "smc", "mac", and "unity".
#' Controls the methods to estimate the initial communalities in \code{PAF} if
#' "none" is among the specified types. "smc" will use squared multiple
#' correlations, "mac" will use maximum absolute correlations, "unity" will use
#' 1s (for details see \code{\link{EFA}}). Default is \code{c("smc", "mac", "unity")}.
#' @param criterion numeric vector. The convergence criterion used for PAF if
#' "none" is among the specified types.
#' If the change in communalities from one iteration to the next is smaller than
#' this criterion the solution is accepted and the procedure ends.
#' Default is \code{0.001}.
#' @param criterion_type character vector. Any combination of "max_individual" and
#' "sum". Type of convergence criterion used for PAF if "none" is among the
#' specified types. "max_individual" selects the maximum change in any of the
#' communalities from one iteration to the next and tests it against the
#' specified criterion. "sum" takes the difference of
#' the sum of all communalities in one iteration and the sum of all communalities
#' in the next iteration and tests this against the criterion
#' (for details see \code{\link{EFA}}). Default is \code{c("sum", "max_individual")}.
#' @param abs_eigen logical vector. Any combination of TRUE and FALSE.
#' Which algorithm to use in the PAF iterations if "none" is among the specified
#' types. If FALSE, the loadings are computed from the eigenvalues. This is also
#' used by the \code{\link[psych:fa]{psych::fa}} function. If TRUE the
#' loadings are computed with the absolute eigenvalues as done by SPSS
#' (for details see \code{\link{EFA}}). Default is \code{TRUE}.
#' @param varimax_type character vector. Any combination of "svd" and "kaiser".
#' The type of the varimax rotation performed if "none" is among the specified
#' types and "varimax", "promax", "orthogonal", or "oblique" is among the specified
#' rotations. "svd" uses singular value decomposition, as
#' \link[stats:varimax]{stats::varimax} does, and "kaiser" uses the varimax
#' procedure performed in SPSS. This is the original procedure from Kaiser (1958),
#' but with slight alterations in the varimax criterion (for details, see
#' \code{\link{EFA}} and Grieder & Steiner, 2020).
#' Default is \code{c("svd", "kaiser")}.
#' @param normalize logical vector. Any combination of TRUE and FALSE.
#' \code{TRUE} performs a kaiser normalization before the specified rotation(s).
#' Default is \code{TRUE}.
#' @param k_promax numeric vector. The power used for computing the target matrix
#' P in the promax rotation if "none" is among the specified types and "promax"
#' or "oblique" is among the specified rotations. Default is \code{2:4}.
#' @param k_simplimax numeric. The number of 'close to zero loadings' for the
#' simplimax rotation (see \code{\link[GPArotation:GPA]{GPArotation::GPFoblq}})
#' if "simplimax" or "oblique" is among the specified rotations. Default
#' is \code{ncol(x)}, where x is the entered data.
#' @param P_type character vector. Any combination of "norm" and "unnorm".
#' This specifies how the target matrix P is computed in promax rotation if
#' "none" is among the specified types and "promax" or "oblique" is among the
#' specified rotations. "unnorm" will use the unnormalized target matrix as
#' originally done in Hendrickson and White (1964). "norm" will use a
#' normalized target matrix (for details see \code{\link{EFA}}).
#' Default is \code{c("norm", "unnorm")}.
#' @param precision numeric vector. The tolerance for stopping in the rotation
#' procedure(s). Default is 10^-5.
#' @param start_method character vector. Any combination of "psych" and "factanal".
#' How to specify the starting values for the optimization procedure for ML.
#' "psych" takes the starting values specified in \link[psych:fa]{psych::fa}.
#' "factanal" takes the starting values specified in the
#' \link[stats:factanal]{stats::factanal} function. Default is
#' \code{c("psych", "factanal")}.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw data
#' is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#' @param show_progress logical. Whether a progress bar should be shown in the
#' console. Default is TRUE.
#'
#' @details
#'
#' As a first step in this function, a grid is produced containing the setting
#' combinations for the to-be-performed EFAs. These settings are then entered as
#' arguments to the \code{\link{EFA}} function and the EFAs are run in a second
#' step. After all EFAs are run, the factor solutions are averaged and their
#' variability determined in a third step.
#'
#' The grid containing the setting combinations is produced based on the entries
#' to the respective arguments. To this end, all possible combinations resulting
#' in unique EFA models are considered. That is, if, for example, the \code{type}
#' argument was set to \code{c("none", "SPSS")} and one combination of the specific
#' settings entered was identical to the SPSS combination, this combination
#' would be included in the grid and run only once. We include here a list
#' of arguments that are only evaluated under specific conditions:
#'
#' The arguments \code{init_comm}, \code{criterion}, \code{criterion_type},
#' \code{abs_eigen} are only evaluated if "PAF" is included in \code{method}
#' and "none" is included in \code{type}.
#'
#' The argument \code{varimax_type} is only evaluated if "varimax", "promax",
#' "oblique", or "orthogonal" is included in \code{rotation} and "none" is
#' included in \code{type}.
#'
#' The argument \code{normalize} is only evaluated if \code{rotation} is not
#' set to "none" and "none" is included in \code{type}.
#'
#' The argument \code{k_simplimax} is only evaluated if "simplimax" or "oblique"
#' is included in \code{rotation}.
#'
#' The arguments \code{k_promax} and \code{P_type} are only evaluated if
#' "promax" or "oblique" is included in \code{rotation} and "none" is included
#' in \code{type}.
#'
#' The argument \code{start_method} is only evaluated if "ML" is included in
#' \code{method}.
#'
#' To avoid a bias in the averaged factor solutions from problematic solutions,
#' these are excluded prior to averaging. A solution is deemed problematic if
#' at least one of the following is true: an error occurred, the model did not
#' converge, or there is at least one Heywood case (defined as a loading or communality of >= .998).
#' Information on errors, convergence, and Heywood cases are returned in the
#' implementations_grid and a summary of these is given when printing the output.
#' In addition to these, information on the admissibility of the factor solutions
#' is also included. A solution was deemed admissible if (1) no error occurred,
#' (2) the model converged, (3) no Heywood cases are present, and (4) there are
#' at least two salient loadings (i.e., loadings exceeding the specified
#' \code{salience_threshold}) for each factor. So, solutions failing one of the
#' first three of these criteria of admissibility are also deemed problematic and
#' therefore excluded from averaging. However, solutions failing only
#' the fourth criterion of admissibility are still included for averaging.
#' Finally, if all solutions are problematic (e.g., all solutions contain
#' Heywood cases), no averaging is performed and the respective outputs are NA.
#' In this case, the implementations_grid should be inspected to see if there
#' are any error messages, and the separate EFA solutions that are also included
#' in the output can be inspected as well, for example, to see where Heywood
#' cases occurred.
#'
#' A core output of this function includes the average, minimum, and maximum
#' loadings derived from all non-problematic (see above) factor solutions. Please
#' note that these are not entire solutions, but the matrices include the average,
#' minimum, or maximum value for each cell (i.e., each loading separately). This
#' means that, for example, the matrix with the minimum loadings will contain
#' the minimum value in any of the factor solutions for each specific loading,
#' and therefore most likely contains loadings from different factor solutions.
#' The matrices containing the minimum and maximum factor solutions can
#' therefore not be interpreted as whole factor solutions.
#'
#' The output also includes information on the average, minimum, maximum, and
#' variability of the fit indices across the non-problematic factor solutions.
#' It is important to note that not all fit indices are computed for all fit
#' methods: For ML and ULS, all fit indices can be computed, while for PAF, only
#' the common part accounted for (CAF) index (Lorenzo-Seva, Timmerman, & Kiers, 2011)
#' can be computed. As a consequence, if only "PAF" is included in the
#' \code{method} argument, averaging can only be performed for the CAF, and the
#' other fit indices are NA. If a combination of "PAF" and "ML" and/or "ULS" are
#' included in the \code{method} argument, the CAF is averaged across all non-
#' problematic factor solutions, while all other fit indices are only averaged
#' across the ML and ULS solutions. The user should therefore keep in mind that
#' the number of EFAs across which the fit indices are averaged can diverge for
#' the CAF compared to all other fit indices.
#'
#' @return A list of class EFA_AVERAGE containing
#' \item{orig_R}{Original correlation matrix.}
#' \item{h2}{A list with the average, standard deviation, minimum, maximum, and
#' range of the final communality estimates across the factor solutions.}
#' \item{loadings}{A list with the average, standard deviation, minimum, maximum,
#' and range of the final loadings across the factor solutions. If rotation was
#' "none", the unrotated loadings, otherwise the rotated loadings (pattern
#' coefficients).}
#' \item{Phi}{A list with the average, standard deviation, minimum, maximum, and
#' range of the factor intercorrelations across factor solutions obtained with
#' oblique rotations.}
#' \item{ind_fac_corres}{A matrix with each cell containing the proportion of
#' the factor solutions in which the respective indicator-to-factor correspondence
#' occurred, i.e., in which the loading exceeded the specified salience threshold.
#' Note: Rowsums can exceed 1 due to cross-loadings.}
#' \item{vars_accounted}{A list with the average, standard deviation, minimum,
#' maximum, and range of explained variances and sums of squared loadings across
#' the factor solutions. Based on the unrotated loadings.}
#' \item{fit_indices}{A matrix containing the average, standard deviation,
#' minimum, maximum, and range for all applicable fit indices across the respective
#' factor solutions, and the degrees of freedom (df). If the method argument
#' contains ML or ULS: Fit indices derived
#' from the unrotated factor loadings: Chi Square (chisq), including significance
#' level, Comparative Fit Index (CFI), Root Mean Square Error of Approximation
#' (RMSEA), Akaike Information Criterion (AIC), Bayesian Information Criterion
#' (BIC)and the common part accounted for (CAF) index as proposed by
#' Lorenzo-Seva, Timmerman, & Kiers (2011). For PAF, only the CAF can be
#' calculated (see details).}
#' \item{implementations_grid}{A matrix containing, for each performed EFA,
#' the setting combination, if an error occurred (logical), the error message
#' (character), an integer code for convergence as returned by
#' \code{\link[stats:optim]{stats:optim}} (0 indicates successful completion.),
#' if heywood cases occurred (logical, see details for definition), if the
#' solution was admissible (logical, see details for definition), and the fit
#' indices.}
#' \item{efa_list}{A list containing the outputs of all performed EFAs. The names
#' correspond to the rownames from the implementations_grid.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Grieder, S., & Steiner, M.D. (2020). Algorithmic Jingle Jungle:
#' A Comparison of Implementations of Principal Axis Factoring and Promax Rotation
#'  in R and SPSS. Manuscript in Preparation.
#' @source Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
#' rotation to oblique simple structure. British Journal of Statistical Psychology,
#' 17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x
#' @source Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
#' Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
#' Research, 46, 340-364, doi: 10.1080/00273171.2011.564527
#' @source Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
#' factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Averaging across different implementations of PAF and promax rotation (72 EFAs)
#' Aver_PAF <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)
#'
#' # Use median instead of mean for averaging (72 EFAs)
#' Aver_PAF_md <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                            averaging = "median")
#'
#' # Averaging across different implementations of PAF and promax rotation,
#' # and across ULS and different versions of ML (108 EFAs)
#' Aver_meth_ext <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                              method = c("PAF", "ULS", "ML"))
#'
#' # Averaging across one implementation each of PAF (EFAtools type), ULS, and
#' # ML with one implementation of promax (EFAtools type) (3 EFAs)
#' Aver_meth <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                          method = c("PAF", "ULS", "ML"), type = "EFAtools",
#'                          start_method = "psych")
#'
#' # Averaging across different oblique rotation methods, using one implementation
#' # of ML and one implementation of promax (EFAtools type) (7 EFAs)
#' Aver_rot <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                          method = "ML", rotation = "oblique", type = "EFAtools",
#'                          start_method = "psych")
#'}
#'
EFA_AVERAGE <- function(x, n_factors, N = NA, method = "PAF", rotation = "promax",
                        type = "none", averaging = c("mean", "median"), trim = 0,
                        salience_threshold = .3,
                        max_iter = 1e4,
                        init_comm = c("smc", "mac", "unity"),
                        criterion = c(1e-3),
                        criterion_type = c("sum", "max_individual"),
                        abs_eigen = c(TRUE),
                        varimax_type = c("svd", "kaiser"),
                        normalize = TRUE,
                        k_promax = 2:4, k_simplimax = ncol(x),
                        P_type = c("norm", "unnorm"), precision = 1e-5,
                        start_method = c("psych", "factanal"),
                        use = c("pairwise.complete.obs", "all.obs",
                                "complete.obs", "everything", "na.or.complete"),
                        cor_method = c("pearson", "spearman", "kendall"),
                        show_progress = TRUE) {

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  checkmate::assert_count(n_factors)
  checkmate::assert_count(N, na.ok = TRUE)
  checkmate::assert_subset(method, c("PAF", "ML", "ULS"),
                           empty.ok = FALSE)
  checkmate::assert_subset(rotation, c("none", "orthogonal", "oblique", "varimax",
                                       "equamax", "quartimax", "geominT", "bentlerT",
                                       "bifactorT", "promax", "oblimin", "quartimin",
                                       "simplimax", "bentlerQ", "geominQ", "bifactorQ"),
                           empty.ok = FALSE)
  checkmate::assert_subset(type, c("none", "EFAtools", "psych", "SPSS"),
                           empty.ok = FALSE)
  averaging <- match.arg(averaging)
  checkmate::assert_number(trim, lower = 0, upper = 0.5)
  checkmate::assert_number(salience_threshold, lower = 0, upper = 1)
  checkmate::assert_count(max_iter)
  checkmate::assert_subset(init_comm, c("smc", "mac", "unity"),
                           empty.ok = FALSE)
  checkmate::assert_vector(criterion, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_true(all(criterion > 0 & criterion < 1))
  checkmate::assert_subset(criterion_type, c("max_individual", "sum"),
                           empty.ok = FALSE)
  checkmate::assert_subset(abs_eigen, c(TRUE, FALSE),
                           empty.ok = FALSE)
  checkmate::assert_subset(varimax_type, c("svd", "kaiser"),
                           empty.ok = FALSE)
  checkmate::assert_subset(normalize, c(TRUE, FALSE),
                           empty.ok = FALSE)
  checkmate::assert_vector(k_promax, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_vector(k_simplimax, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_subset(P_type, c("unnorm", "norm"),
                           empty.ok = FALSE)
  checkmate::assert_vector(precision, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_true(all(precision > 0 & precision < 1))
  checkmate::assert_subset(start_method, c("psych", "factanal"),
                           empty.ok = FALSE)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  checkmate::assert_flag(show_progress)



  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }

  # Check if model is identified

  # calculate degrees of freedom
  m <- ncol(R)
  df <- ((m - n_factors)**2 - (m + n_factors)) / 2

  if(df < 0){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" The model is underidentified. Please enter a lower number of factors or use a larger number of indicators and try again.\n"))

  } else if (df == 0){

    warning(crayon::yellow$bold("!"), crayon::yellow(" The model is just identified (df = 0). We suggest to try again with a lower number of factors or a larger number of indicators.\n"))

  }

  if (n_factors == 1 && !all(rotation == "none")) {
    message(cli::col_cyan(cli::symbol$info, " 'n_factors' is 1, but rotation != 'none'. Setting rotation to 'none' to avoid many warnings, as 1-factor solutions cannot be rotated.\n"))
    rotation <- "none"
  }

  ### create the grid with all combinations of the input arguments

  grid_list <- list()


  if ("PAF" %in% method) {

    if ("EFAtools" %in% type) {
      grid_list[["ftls_pf"]] <- .type_grid(method = "PAF", init_comm = "smc",
                                           criterion = 1e-3, criterion_type = "sum",
                                           abs_eigen = TRUE, start_method = NA,
                                           rotation = rotation, k_promax = 4,
                                           normalize = TRUE, P_type = "norm",
                                           precision = 1e-5,
                                           varimax_type = "kaiser",
                                           k_simplimax = k_simplimax)
    }

    if ("psych" %in% type) {
      grid_list[["psch_pf"]] <- .type_grid(method = "PAF", init_comm = "smc",
                                           criterion = 1e-3, criterion_type = "sum",
                                           abs_eigen = FALSE, start_method = NA,
                                           rotation = rotation, k_promax = 4,
                                           normalize = TRUE, P_type = "unnorm",
                                           precision = 1e-5,
                                           varimax_type = "svd",
                                           k_simplimax = k_simplimax)
    }

    if ("SPSS" %in% type) {
      grid_list[["spss_pf"]] <- .type_grid(method = "PAF", init_comm = "smc",
                                           criterion = 1e-3, criterion_type = "max_individual",
                                           abs_eigen = TRUE, start_method = NA,
                                           rotation = rotation, k_promax = 4,
                                           normalize = TRUE, P_type = "norm",
                                           precision = 1e-5,
                                           varimax_type = "kaiser",
                                           k_simplimax = k_simplimax)
    }

    if ("none" %in% type) {

        grid_list[["nn_pf"]] <- .type_grid(method = "PAF", init_comm = init_comm,
                                           criterion = criterion, criterion_type = criterion_type,
                                           abs_eigen = abs_eigen, start_method = NA,
                                           rotation = rotation, k_promax = k_promax,
                                           normalize = normalize, P_type = P_type,
                                           precision = precision,
                                           varimax_type = varimax_type,
                                           k_simplimax = k_simplimax)
    }
  }

    if ("ML" %in% method) {

      if ("EFAtools" %in% type) {
        grid_list[["ftls_ml"]] <- .type_grid(method = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, start_method = "psych",
                                             rotation = rotation, k_promax = 4,
                                             normalize = TRUE, P_type = "norm",
                                             precision = 1e-5,
                                             varimax_type = "kaiser",
                                             k_simplimax = k_simplimax)
      }

      if ("psych" %in% type) {
        grid_list[["psch_ml"]] <- .type_grid(method = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, start_method = "psych",
                                             rotation = rotation, k_promax = 4,
                                             normalize = TRUE, P_type = "unnorm",
                                             precision = 1e-5,
                                             varimax_type = "svd",
                                             k_simplimax = k_simplimax)
      }

      if ("SPSS" %in% type) {
        grid_list[["spss_ml"]] <- .type_grid(method = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, start_method = "psych",
                                             rotation = rotation, k_promax = 4,
                                             normalize = TRUE, P_type = "norm",
                                             precision = 1e-5,
                                             varimax_type = "kaiser",
                                             k_simplimax = k_simplimax)
      }

      if ("none" %in% type) {
          grid_list[["nn_ml"]] <- .type_grid(method = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, start_method = start_method,
                                             rotation = rotation, k_promax = k_promax,
                                             normalize = normalize, P_type = P_type,
                                             precision = precision,
                                             varimax_type = varimax_type,
                                             k_simplimax = k_simplimax)

      }
    }

      if ("ULS" %in% method) {

        if ("EFAtools" %in% type) {
          grid_list[["ftls_ls"]] <- .type_grid(method = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, start_method = NA,
                                               rotation = rotation, k_promax = 4,
                                               normalize = TRUE, P_type = "norm",
                                               precision = 1e-5,
                                               varimax_type = "kaiser",
                                               k_simplimax = k_simplimax)
        }

        if ("psych" %in% type) {
          grid_list[["psch_ls"]] <- .type_grid(method = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, start_method = NA,
                                               rotation = rotation, k_promax = 4,
                                               normalize = TRUE, P_type = "unnorm",
                                               precision = 1e-5,
                                               varimax_type = "svd",
                                               k_simplimax = k_simplimax)
        }

        if ("SPSS" %in% type) {
          grid_list[["spss_ls"]] <- .type_grid(method = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, start_method = NA,
                                               rotation = rotation, k_promax = 4,
                                               normalize = TRUE, P_type = "norm",
                                               precision = 1e-5,
                                               varimax_type = "kaiser",
                                               k_simplimax = k_simplimax)
        }

        if ("none" %in% type) {

            grid_list[["nn_ls"]] <- .type_grid(method = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, start_method = NA,
                                               rotation = rotation, k_promax = k_promax,
                                               normalize = normalize, P_type = P_type,
                                               precision = precision,
                                               varimax_type = varimax_type,
                                               k_simplimax = k_simplimax)

        }
      }

  arg_grid <- unique(do.call(rbind, grid_list))

  ### Run all efas

  if (nrow(arg_grid) == 1) {

    warning(crayon::yellow$bold("!"), crayon::yellow(" There was only one combination of arguments, returning normal EFA output.\n"))

    return(EFA(R, n_factors, N = N, method = arg_grid$method, rotation = arg_grid$rotation,
        type = "none", max_iter = max_iter, init_comm = arg_grid$init_comm,
        criterion = arg_grid$criterion, criterion_type = arg_grid$criterion_type,
        abs_eigen = arg_grid$abs_eigen, varimax_type = arg_grid$varimax_type,
        k = ifelse(arg_grid$rotation == "promax", arg_grid$k_promax, arg_grid$k_simplimax),
        normalize = arg_grid$normalize, P_type = arg_grid$P_type, precision = precision,
        order_type = "eigen", start_method = arg_grid$start_method))

  }

  if (isTRUE(show_progress)) {
    progressr::handlers("progress")
  } else {
    progressr::handlers("void")
  }

  progressr::with_progress({
    n_efa <- nrow(arg_grid)
    if (n_efa <= 10) {
      stepsize <- 1
    } else {
      stepsize <- round(n_efa / 100 * 10)
    }

    efa_progress_bar <- progressr::progressor(steps = n_efa / stepsize)
    efa_progress_bar("Running EFAs:", class = "sticky", amount = 0)
    efa_list <- future.apply::future_lapply(1:n_efa,
                                            function(i, methods, rotations,
                                                     init_comms, criteria,
                                                     criterion_types, abs_eigens,
                                                     varimax_types, k_ps, k_ss,
                                                     normalizes, P_types, start_methods) {
      if (i %% stepsize == 0){
        efa_progress_bar(message = sprintf("Running EFA %g of %g", i, n_efa))
      }

      try(EFA(R, n_factors, N = N, method = methods[i], rotation = rotations[i],
              type = "none", max_iter = max_iter, init_comm = init_comms[i],
              criterion = criteria[i], criterion_type = criterion_types[i],
              abs_eigen = abs_eigens[i], varimax_type = varimax_types[i],
              k = ifelse(rotations[i] == "promax",k_ps[i], k_ss[i]),
              normalize = normalizes[i], P_type = P_types[i], precision = precision,
              order_type = "eigen", start_method = start_methods[i], maxit = 5e4),
          silent = TRUE)
    }, methods = arg_grid$method, rotations = arg_grid$rotation,
    init_comms = arg_grid$init_comm, criteria = arg_grid$criterion,
    criterion_types = arg_grid$criterion_type, abs_eigens = arg_grid$abs_eigen,
    varimax_types = arg_grid$varimax_type, k_ps = arg_grid$k_promax,
    k_ss = arg_grid$k_simplimax, normalizes = arg_grid$normalize,
    P_types = arg_grid$P_type, start_methods = arg_grid$start_method,
    future.seed = TRUE)

    if (n_efa %% 10 != 0){
      efa_progress_bar(message = "Done Running EFAs ")
    }
  })

  names(efa_list) <- rownames(arg_grid)

  ### Extract relevant information from EFA outputs
  if (isTRUE(show_progress)) {
    .show_av_progress("\U0001f3c3", "Extracting data...")
  }
  ext_list <- .extract_data(efa_list, R, n_factors, n_efa, rotation, salience_threshold)

  if (all(isTRUE(ext_list$for_grid$errors) | ext_list$for_grid$converged != 0 |
          isTRUE(ext_list$for_grid$heywood))) {

    av_list <- list(
      h2 = NA,
      loadings = NA,
      phi = NA,
      ind_fac_corres = NA,
      vars_accounted = NA,
      fit_indices = NA)

  } else {

    if (n_factors > 1) {
      if (isTRUE(show_progress)) {
        .show_av_progress("\U0001f6b6", "Reordering factors...")
      }

      re_list <- .array_reorder(ext_list$vars_accounted, ext_list$L, ext_list$L_corres,
                                ext_list$phi, ext_list$extract_phi, n_factors)
    } else {
      re_list <- ext_list
    }


    if (isTRUE(show_progress)) {
      .show_av_progress("\U0001f3c3", "Averaging data...")
    }
    av_list <- suppressWarnings(
      .average_values(re_list$vars_accounted, re_list$L, re_list$L_corres,
                      ext_list$h2, re_list$phi, ext_list$extract_phi, averaging,
                      trim, ext_list$for_grid[, c("chisq", "p_chi", "caf", "cfi",
                                                  "rmsea", "aic", "bic")], df,
                      colnames(R)))

  }


  arg_grid <- cbind(arg_grid, ext_list$for_grid)


  settings <- list(
    method = method,
    rotation = rotation,
    type = type,
    n_factors = n_factors,
    N = N,
    init_comm = init_comm,
    criterion = criterion,
    criterion_type = criterion_type,
    abs_eigen = abs_eigen,
    varimax_type = varimax_type,
    normalize = normalize,
    k_promax = k_promax,
    k_simplimax = k_simplimax,
    P_type = P_type,
    precision = precision,
    start_method = start_method,
    use = use,
    cor_method = cor_method,
    max_iter = max_iter,
    averaging = averaging,
    trim = trim,
    salience_threshold = salience_threshold
  )

  # Create output
  output <- list(
    orig_R = R,
    h2 = av_list$h2,
    loadings = av_list$loadings,
    Phi = av_list$phi,
    ind_fac_corres = av_list$ind_fac_corres,
    vars_accounted = av_list$vars_accounted,
    fit_indices = av_list$fit_indices,
    implementations_grid = arg_grid,
    efa_list = efa_list,
    settings = settings
  )

  class(output) <- "EFA_AVERAGE"

  if (isTRUE(show_progress)) {
    .show_av_progress("", "", done = TRUE)
  }

  return(output)

}

efa_def <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)
efa_ml <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                      method = "ML")
efa_uls <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                      method = "ULS")

efa_all <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                       method = c("PAF", "ML", "ULS"),
                       type = c("none", "EFAtools", "psych", "SPSS"),
                       salience_threshold = .2)
efa_all_np <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                       method = c("PAF", "ML", "ULS"),
                       type = c("none", "EFAtools", "psych", "SPSS"),
                       salience_threshold = .2, show_progress = FALSE)

efa_all_oblq <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                            method = c("PAF", "ML", "ULS"),
                            type = c("none", "EFAtools", "psych", "SPSS"),
                            rotation = "oblique")
efa_all_orth <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                            method = c("PAF", "ML", "ULS"),
                            type = c("none", "EFAtools", "psych", "SPSS"),
                            rotation = "orthogonal")
efa_all_none <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                            method = c("PAF", "ML", "ULS"),
                            type = c("none", "EFAtools", "psych", "SPSS"),
                            rotation = "none")

efa_all_md <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                            method = c("PAF", "ML", "ULS"),
                            type = c("none", "EFAtools", "psych", "SPSS"),
                            rotation = "oblique", averaging = "median")
efa_all_tm <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                          method = c("PAF", "ML", "ULS"),
                          type = c("none", "EFAtools", "psych", "SPSS"),
                          rotation = "oblique", averaging = "mean",
                          trim = .2)
efa_raw <- EFA_AVERAGE(GRiPS_raw, n_factors = 1, rotation = "none")
efa_raw_p <- EFA_AVERAGE(GRiPS_raw, n_factors = 2, rotation = "promax")

test_that("output class and dimensions are correct", {
  expect_is(efa_def, "EFA_AVERAGE")
  expect_is(efa_ml, "EFA_AVERAGE")
  expect_is(efa_uls, "EFA_AVERAGE")
  expect_is(efa_all, "EFA_AVERAGE")
  expect_is(efa_all_oblq, "EFA_AVERAGE")
  expect_is(efa_all_orth, "EFA_AVERAGE")
  expect_is(efa_all_none, "EFA_AVERAGE")
  expect_is(efa_all_md, "EFA_AVERAGE")
  expect_is(efa_all_tm, "EFA_AVERAGE")
  expect_is(efa_raw, "EFA_AVERAGE")
  expect_is(efa_raw_p, "EFA_AVERAGE")

  expect_is(efa_def$loadings$average, "LOADINGS")
  expect_is(efa_ml$loadings$average, "LOADINGS")
  expect_is(efa_uls$loadings$average, "LOADINGS")
  expect_is(efa_all$loadings$average, "LOADINGS")
  expect_is(efa_all_oblq$loadings$average, "LOADINGS")
  expect_is(efa_all_orth$loadings$average, "LOADINGS")
  expect_is(efa_all_none$loadings$average, "LOADINGS")
  expect_is(efa_all_md$loadings$average, "LOADINGS")
  expect_is(efa_all_tm$loadings$average, "LOADINGS")
  expect_is(efa_raw$loadings$average, "LOADINGS")
  expect_is(efa_raw_p$loadings$average, "LOADINGS")

  expect_named(efa_def, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_ml, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_uls, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_oblq, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_orth, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_none, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_md, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_tm, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_raw, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_raw_p, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
})

test_that("settings are returned correctly", {
  expect_named(efa_def$settings, c("method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold"))
  expect_named(efa_ml$settings, c("method", "rotation", "type", "n_factors", "N",
                                  "init_comm", "criterion", "criterion_type",
                                  "abs_eigen", "varimax_type", "normalize",
                                  "k_promax", "k_simplimax", "P_type",
                                  "precision", "start_method", "use",
                                  "cor_method", "max_iter", "averaging",
                                  "trim", "salience_threshold"))
  expect_named(efa_uls$settings, c("method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold"))
  expect_named(efa_all$settings, c("method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold"))
  expect_named(efa_all_oblq$settings, c("method", "rotation", "type", "n_factors", "N",
                                        "init_comm", "criterion", "criterion_type",
                                        "abs_eigen", "varimax_type", "normalize",
                                        "k_promax", "k_simplimax", "P_type",
                                        "precision", "start_method", "use",
                                        "cor_method", "max_iter", "averaging",
                                        "trim", "salience_threshold"))
  expect_named(efa_all_orth$settings, c("method", "rotation", "type", "n_factors", "N",
                                        "init_comm", "criterion", "criterion_type",
                                        "abs_eigen", "varimax_type", "normalize",
                                        "k_promax", "k_simplimax", "P_type",
                                        "precision", "start_method", "use",
                                        "cor_method", "max_iter", "averaging",
                                        "trim", "salience_threshold"))
  expect_named(efa_all_none$settings, c("method", "rotation", "type", "n_factors", "N",
                                        "init_comm", "criterion", "criterion_type",
                                        "abs_eigen", "varimax_type", "normalize",
                                        "k_promax", "k_simplimax", "P_type",
                                        "precision", "start_method", "use",
                                        "cor_method", "max_iter", "averaging",
                                        "trim", "salience_threshold"))
  expect_named(efa_all_md$settings, c("method", "rotation", "type", "n_factors", "N",
                                      "init_comm", "criterion", "criterion_type",
                                      "abs_eigen", "varimax_type", "normalize",
                                      "k_promax", "k_simplimax", "P_type",
                                      "precision", "start_method", "use",
                                      "cor_method", "max_iter", "averaging",
                                      "trim", "salience_threshold"))
  expect_named(efa_all_tm$settings, c("method", "rotation", "type", "n_factors", "N",
                                      "init_comm", "criterion", "criterion_type",
                                      "abs_eigen", "varimax_type", "normalize",
                                      "k_promax", "k_simplimax", "P_type",
                                      "precision", "start_method", "use",
                                      "cor_method", "max_iter", "averaging",
                                      "trim", "salience_threshold"))
  expect_named(efa_raw$settings, c("method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold"))
  expect_named(efa_raw_p$settings, c("method", "rotation", "type", "n_factors", "N",
                                     "init_comm", "criterion", "criterion_type",
                                     "abs_eigen", "varimax_type", "normalize",
                                     "k_promax", "k_simplimax", "P_type",
                                     "precision", "start_method", "use",
                                     "cor_method", "max_iter", "averaging",
                                     "trim", "salience_threshold"))


  expect_equal(efa_def$settings$method, "PAF")
  expect_equal(efa_ml$settings$method, "ML")
  expect_equal(efa_uls$settings$method, "ULS")
  expect_equal(efa_all$settings$method, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_oblq$settings$method, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_orth$settings$method, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_none$settings$method, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_md$settings$method, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_tm$settings$method, c("PAF", "ML", "ULS"))
  expect_equal(efa_raw$settings$method, "PAF")
  expect_equal(efa_raw_p$settings$method, "PAF")

  expect_equal(efa_def$settings$rotation, "promax")
  expect_equal(efa_ml$settings$rotation, "promax")
  expect_equal(efa_uls$settings$rotation, "promax")
  expect_equal(efa_all$settings$rotation, "promax")
  expect_equal(efa_all_oblq$settings$rotation, "oblique")
  expect_equal(efa_all_orth$settings$rotation, "orthogonal")
  expect_equal(efa_all_none$settings$rotation, "none")
  expect_equal(efa_all_md$settings$rotation, "oblique")
  expect_equal(efa_all_tm$settings$rotation, "oblique")
  expect_equal(efa_raw$settings$rotation, "none")
  expect_equal(efa_raw_p$settings$rotation, "promax")

  expect_equal(efa_def$settings$type, "none")
  expect_equal(efa_ml$settings$type, "none")
  expect_equal(efa_uls$settings$type, "none")
  expect_equal(efa_all$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_oblq$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_orth$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_none$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_md$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_tm$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_raw$settings$type, "none")
  expect_equal(efa_raw_p$settings$type, "none")

  expect_equal(efa_def$settings$n_factors, 3)
  expect_equal(efa_ml$settings$n_factors, 3)
  expect_equal(efa_uls$settings$n_factors, 3)
  expect_equal(efa_all$settings$n_factors, 3)
  expect_equal(efa_all_oblq$settings$n_factors, 3)
  expect_equal(efa_all_orth$settings$n_factors, 3)
  expect_equal(efa_all_none$settings$n_factors, 3)
  expect_equal(efa_all_md$settings$n_factors, 3)
  expect_equal(efa_all_tm$settings$n_factors, 3)
  expect_equal(efa_raw$settings$n_factors, 1)
  expect_equal(efa_raw_p$settings$n_factors, 2)

  expect_equal(efa_def$settings$N, 500)
  expect_equal(efa_ml$settings$N, 500)
  expect_equal(efa_uls$settings$N, 500)
  expect_equal(efa_all$settings$N, 500)
  expect_equal(efa_all_oblq$settings$N, 500)
  expect_equal(efa_all_orth$settings$N, 500)
  expect_equal(efa_all_none$settings$N, 500)
  expect_equal(efa_all_md$settings$N, 500)
  expect_equal(efa_all_tm$settings$N, 500)
  expect_equal(efa_raw$settings$N, 810)
  expect_equal(efa_raw_p$settings$N, 810)

  expect_equal(efa_def$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_ml$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_uls$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_oblq$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_orth$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_none$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_md$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_tm$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_raw$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_raw_p$settings$init_comm, c("smc", "mac", "unity"))

  expect_equal(efa_def$settings$criterion, 0.001)
  expect_equal(efa_ml$settings$criterion, 0.001)
  expect_equal(efa_uls$settings$criterion, 0.001)
  expect_equal(efa_all$settings$criterion, 0.001)
  expect_equal(efa_all_oblq$settings$criterion, 0.001)
  expect_equal(efa_all_orth$settings$criterion, 0.001)
  expect_equal(efa_all_none$settings$criterion, 0.001)
  expect_equal(efa_all_md$settings$criterion, 0.001)
  expect_equal(efa_all_tm$settings$criterion, 0.001)
  expect_equal(efa_raw$settings$criterion, 0.001)
  expect_equal(efa_raw_p$settings$criterion, 0.001)

  expect_equal(efa_def$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_ml$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_uls$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_oblq$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_orth$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_none$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_md$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_tm$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_raw$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_raw_p$settings$criterion_type, c("sum", "max_individual"))

  expect_equal(efa_def$settings$abs_eigen, TRUE)
  expect_equal(efa_ml$settings$abs_eigen, TRUE)
  expect_equal(efa_uls$settings$abs_eigen, TRUE)
  expect_equal(efa_all$settings$abs_eigen, TRUE)
  expect_equal(efa_all_oblq$settings$abs_eigen, TRUE)
  expect_equal(efa_all_orth$settings$abs_eigen, TRUE)
  expect_equal(efa_all_none$settings$abs_eigen, TRUE)
  expect_equal(efa_all_md$settings$abs_eigen, TRUE)
  expect_equal(efa_all_tm$settings$abs_eigen, TRUE)
  expect_equal(efa_raw$settings$abs_eigen, TRUE)
  expect_equal(efa_raw_p$settings$abs_eigen, TRUE)

  expect_equal(efa_def$settings$abs_eigen, TRUE)
  expect_equal(efa_ml$settings$abs_eigen, TRUE)
  expect_equal(efa_uls$settings$abs_eigen, TRUE)
  expect_equal(efa_all$settings$abs_eigen, TRUE)
  expect_equal(efa_all_oblq$settings$abs_eigen, TRUE)
  expect_equal(efa_all_orth$settings$abs_eigen, TRUE)
  expect_equal(efa_all_none$settings$abs_eigen, TRUE)
  expect_equal(efa_all_md$settings$abs_eigen, TRUE)
  expect_equal(efa_all_tm$settings$abs_eigen, TRUE)
  expect_equal(efa_raw$settings$abs_eigen, TRUE)
  expect_equal(efa_raw_p$settings$abs_eigen, TRUE)

  expect_equal(efa_def$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_ml$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_uls$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_oblq$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_orth$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_none$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_md$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_tm$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_raw$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_raw_p$settings$varimax_type, c("svd", "kaiser"))

  expect_equal(efa_def$settings$normalize, TRUE)
  expect_equal(efa_ml$settings$normalize, TRUE)
  expect_equal(efa_uls$settings$normalize, TRUE)
  expect_equal(efa_all$settings$normalize, TRUE)
  expect_equal(efa_all_oblq$settings$normalize, TRUE)
  expect_equal(efa_all_orth$settings$normalize, TRUE)
  expect_equal(efa_all_none$settings$normalize, TRUE)
  expect_equal(efa_all_md$settings$normalize, TRUE)
  expect_equal(efa_all_tm$settings$normalize, TRUE)
  expect_equal(efa_raw$settings$normalize, TRUE)
  expect_equal(efa_raw_p$settings$normalize, TRUE)

  expect_equal(efa_def$settings$k_promax, 2:4)
  expect_equal(efa_ml$settings$k_promax, 2:4)
  expect_equal(efa_uls$settings$k_promax, 2:4)
  expect_equal(efa_all$settings$k_promax, 2:4)
  expect_equal(efa_all_oblq$settings$k_promax, 2:4)
  expect_equal(efa_all_orth$settings$k_promax, 2:4)
  expect_equal(efa_all_none$settings$k_promax, 2:4)
  expect_equal(efa_all_md$settings$k_promax, 2:4)
  expect_equal(efa_all_tm$settings$k_promax, 2:4)
  expect_equal(efa_raw$settings$k_promax, 2:4)
  expect_equal(efa_raw_p$settings$k_promax, 2:4)

  expect_equal(efa_def$settings$k_simplimax, 18)
  expect_equal(efa_ml$settings$k_simplimax, 18)
  expect_equal(efa_uls$settings$k_simplimax, 18)
  expect_equal(efa_all$settings$k_simplimax, 18)
  expect_equal(efa_all_oblq$settings$k_simplimax, 18)
  expect_equal(efa_all_orth$settings$k_simplimax, 18)
  expect_equal(efa_all_none$settings$k_simplimax, 18)
  expect_equal(efa_all_md$settings$k_simplimax, 18)
  expect_equal(efa_all_tm$settings$k_simplimax, 18)
  expect_equal(efa_raw$settings$k_simplimax, 8)
  expect_equal(efa_raw_p$settings$k_simplimax, 8)

  expect_equal(efa_def$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_ml$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_uls$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_oblq$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_orth$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_none$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_md$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_tm$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_raw$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_raw_p$settings$P_type, c("norm", "unnorm"))

  expect_equal(efa_def$settings$precision, 1e-5)
  expect_equal(efa_ml$settings$precision, 1e-5)
  expect_equal(efa_uls$settings$precision, 1e-5)
  expect_equal(efa_all$settings$precision, 1e-5)
  expect_equal(efa_all_oblq$settings$precision, 1e-5)
  expect_equal(efa_all_orth$settings$precision, 1e-5)
  expect_equal(efa_all_none$settings$precision, 1e-5)
  expect_equal(efa_all_md$settings$precision, 1e-5)
  expect_equal(efa_all_tm$settings$precision, 1e-5)
  expect_equal(efa_raw$settings$precision, 1e-5)
  expect_equal(efa_raw_p$settings$precision, 1e-5)

  expect_equal(efa_def$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_ml$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_uls$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_oblq$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_orth$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_none$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_md$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_tm$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_raw$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_raw_p$settings$start_method, c("psych", "factanal"))

  expect_equal(efa_def$settings$use, "pairwise.complete.obs")
  expect_equal(efa_ml$settings$use, "pairwise.complete.obs")
  expect_equal(efa_uls$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_oblq$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_orth$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_none$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_md$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_tm$settings$use, "pairwise.complete.obs")
  expect_equal(efa_raw$settings$use, "pairwise.complete.obs")
  expect_equal(efa_raw_p$settings$use, "pairwise.complete.obs")

  expect_equal(efa_def$settings$cor_method, "pearson")
  expect_equal(efa_ml$settings$cor_method, "pearson")
  expect_equal(efa_uls$settings$cor_method, "pearson")
  expect_equal(efa_all$settings$cor_method, "pearson")
  expect_equal(efa_all_oblq$settings$cor_method, "pearson")
  expect_equal(efa_all_orth$settings$cor_method, "pearson")
  expect_equal(efa_all_none$settings$cor_method, "pearson")
  expect_equal(efa_all_md$settings$cor_method, "pearson")
  expect_equal(efa_all_tm$settings$cor_method, "pearson")
  expect_equal(efa_raw$settings$cor_method, "pearson")
  expect_equal(efa_raw_p$settings$cor_method, "pearson")

  expect_equal(efa_def$settings$max_iter, 10000)
  expect_equal(efa_ml$settings$max_iter, 10000)
  expect_equal(efa_uls$settings$max_iter, 10000)
  expect_equal(efa_all$settings$max_iter, 10000)
  expect_equal(efa_all_oblq$settings$max_iter, 10000)
  expect_equal(efa_all_orth$settings$max_iter, 10000)
  expect_equal(efa_all_none$settings$max_iter, 10000)
  expect_equal(efa_all_md$settings$max_iter, 10000)
  expect_equal(efa_all_tm$settings$max_iter, 10000)
  expect_equal(efa_raw$settings$max_iter, 10000)
  expect_equal(efa_raw_p$settings$max_iter, 10000)

  expect_equal(efa_def$settings$averaging, "mean")
  expect_equal(efa_ml$settings$averaging, "mean")
  expect_equal(efa_uls$settings$averaging, "mean")
  expect_equal(efa_all$settings$averaging, "mean")
  expect_equal(efa_all_oblq$settings$averaging, "mean")
  expect_equal(efa_all_orth$settings$averaging, "mean")
  expect_equal(efa_all_none$settings$averaging, "mean")
  expect_equal(efa_all_md$settings$averaging, "median")
  expect_equal(efa_all_tm$settings$averaging, "mean")
  expect_equal(efa_raw$settings$averaging, "mean")
  expect_equal(efa_raw_p$settings$averaging, "mean")

  expect_equal(efa_def$settings$trim, 0)
  expect_equal(efa_ml$settings$trim, 0)
  expect_equal(efa_uls$settings$trim, 0)
  expect_equal(efa_all$settings$trim, 0)
  expect_equal(efa_all_oblq$settings$trim, 0)
  expect_equal(efa_all_orth$settings$trim, 0)
  expect_equal(efa_all_none$settings$trim, 0)
  expect_equal(efa_all_md$settings$trim, 0)
  expect_equal(efa_all_tm$settings$trim, 0.2)
  expect_equal(efa_raw$settings$trim, 0)
  expect_equal(efa_raw_p$settings$trim, 0)

  expect_equal(efa_def$settings$salience_threshold, 0.3)
  expect_equal(efa_ml$settings$salience_threshold, 0.3)
  expect_equal(efa_uls$settings$salience_threshold, 0.3)
  expect_equal(efa_all$settings$salience_threshold, 0.2)
  expect_equal(efa_all_oblq$settings$salience_threshold, 0.3)
  expect_equal(efa_all_orth$settings$salience_threshold, 0.3)
  expect_equal(efa_all_none$settings$salience_threshold, 0.3)
  expect_equal(efa_all_md$settings$salience_threshold, 0.3)
  expect_equal(efa_all_tm$settings$salience_threshold, 0.3)
  expect_equal(efa_raw$settings$salience_threshold, 0.3)
  expect_equal(efa_raw_p$settings$salience_threshold, 0.3)


})


# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z), ncol = 3)
cor_sing <- stats::cor(dat_sing)

cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

test_that("errors are thrown correctly", {
  expect_error(EFA_AVERAGE(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_message(EFA_AVERAGE(GRiPS_raw, n_factors = 2, method = "PAF", type = c("EFAtools", "psych")),
                 " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_warning(EFA_AVERAGE(GRiPS_raw, n_factors = 2, method = "PAF", type = c("EFAtools", "psych"),
                             N = 20),
                 " 'N' was set and data entered. Taking N from data.\n")
  expect_error(EFA_AVERAGE(dat_sing, n_factors = 1),
               " Correlation matrix is singular, no further analyses are performed\n")
  expect_error(EFA_AVERAGE(cor_sing, N = 10, n_factors = 1),
               " Correlation matrix is singular, no further analyses are performed\n")
  expect_error(EFA_AVERAGE(matrix(rnorm(30), ncol = 3), n_factors = 2),
               " The model is underidentified. Please enter a lower number of factors or use a larger number of indicators and try again.\n")
  expect_warning(EFA_AVERAGE(matrix(rnorm(30), ncol = 3), n_factors = 1,
                             method = "PAF", type = c("EFAtools", "psych")),
                 " The model is just identified (df = 0). We suggest to try again with a lower number of factors or a larger number of indicators.\n", fixed = TRUE)
  expect_warning(EFA_AVERAGE(cor_nposdef, n_factors = 1, N = 10, method = "PAF",
                     type = c("EFAtools", "psych")), "Matrix was not positive definite, smoothing was done")
  expect_message(EFA_AVERAGE(GRiPS_raw, n_factors = 1, method = "PAF", type = c("EFAtools", "psych")),
                 " 'n_factors' is 1, but rotation != 'none'. Setting rotation to 'none' to avoid many warnings, as 1-factor solutions cannot be rotated.\n")
  expect_warning(EFA_AVERAGE(GRiPS_raw, n_factors = 1, method = "PAF", type = c("EFAtools"),
                             rotation = "none"),
                 " There was only one combination of arguments, returning normal EFA output.\n")
})

rm(efa_def, efa_ml, efa_uls, efa_all, efa_all_oblq, efa_all_orth, efa_all_none,
   efa_all_md, efa_all_tm, efa_raw, efa_raw_p)

#' Rotate a loading matrix to a target using Procrustes alignment
#'
#' `PROCRUSTES()` aligns one loading matrix to a target loading matrix with the
#' same dimensions. It is used internally by `EFA_POOLED()`, but can also be used
#' directly when factor columns must be brought into a common orientation before
#' averaging or comparing solutions.
#'
#' For `rotation = "orthogonal"`, the function solves the closed-form orthogonal
#' Procrustes problem
#'
#' \deqn{\min_T \frac{1}{2}\|A T - B\|_F^2 \quad \textrm{subject to}\quad T'T = I,}
#'
#' where `A` is the loading matrix and `B` is `Target`.
#'
#' For `rotation = "oblique"`, the function calls the compiled
#' `.oblique_procrustes()` optimizer. The oblique convention is the same as in
#' `GPArotation::targetQ()`:
#'
#' \deqn{L = A T^{-T}, \qquad \Phi = T'T, \qquad diag(\Phi) = 1.}
#'
#' Random starts are only used for oblique alignment. For one-factor models,
#' oblique and orthogonal alignment are equivalent, so the function uses the
#' stable one-factor orthogonal solution instead of calling the oblique optimizer.
#'
#' @param A Numeric loading matrix to be aligned.
#' @param Target Numeric target matrix with the same dimensions as `A`.
#' @param rotation Character string, either `"orthogonal"` or `"oblique"`.
#' @param S Optional `k x k` cross-product matrix `crossprod(A)`. Supplying this
#'   is useful when the same `A` is rotated repeatedly. `S` is used only when
#'   `oblique_normalize = FALSE`; if Kaiser normalization is requested, the
#'   cross-product must be recomputed on the normalized matrix.
#' @param T_init Optional `k x k` nonsingular starting transformation matrix for
#'   the oblique solver. Its columns are normalized internally.
#' @param oblique_eps Positive convergence tolerance for the projected-gradient
#'   norm in the oblique solver.
#' @param oblique_maxit Non-negative integer. Maximum number of projected-gradient
#'   updates in the full oblique solver.
#' @param oblique_max_line_search Non-negative integer. Maximum number of
#'   step-halving attempts after the initial line-search step.
#' @param oblique_step0 Positive initial step size for the oblique solver.
#' @param oblique_normalize Logical; if `TRUE`, apply Kaiser row normalization in
#'   the oblique solver and back-transform the aligned loadings afterwards.
#' @param oblique_random_starts Non-negative integer. Number of additional random
#'   starts used by the oblique solver.
#' @param oblique_screen_keep Non-negative integer. Number of random starts
#'   retained after cheap objective screening and sent to triage optimization.
#' @param oblique_triage_maxit Non-negative integer. Number of short optimization
#'   iterations used in the triage stage.
#' @param oblique_triage_improve_tol Non-negative scalar. Relative improvement
#'   required for a triaged start to be promoted to full optimization.
#'
#' @returns A list containing aligned `loadings`, transformation matrix `T`,
#'   factor intercorrelation matrix `Phi`, target criterion `value`, convergence
#'   diagnostics, line-search diagnostics, and multi-start summaries. Row and
#'   column names are preserved where possible.
#'
#' @export
#'
PROCRUSTES <- function(A,
                       Target,
                       rotation = c("orthogonal", "oblique"),
                       S = NULL,
                       T_init = NULL,
                       oblique_eps = 1e-5,
                       oblique_maxit = 1000,
                       oblique_max_line_search = 10,
                       oblique_step0 = 1,
                       oblique_normalize = FALSE,
                       oblique_random_starts = 0,
                       oblique_screen_keep = 2,
                       oblique_triage_maxit = 25,
                       oblique_triage_improve_tol = 0) {
  rotation <- match.arg(rotation)

  mats <- .procrustes_validate_matrix_pair(A, Target)
  A <- mats$A
  Target <- mats$B
  k <- ncol(A)

  if (rotation == "orthogonal" || k == 1L) {
    out <- .orthogonal_procrustes(A, Target)
    if (rotation == "oblique" && k == 1L) {
      out$method <- "single_factor_procrustes"
    }
    return(out)
  }

  controls <- .procrustes_validate_oblique_controls(
    oblique_eps = oblique_eps,
    oblique_maxit = oblique_maxit,
    oblique_max_line_search = oblique_max_line_search,
    oblique_step0 = oblique_step0,
    oblique_normalize = oblique_normalize,
    oblique_random_starts = oblique_random_starts,
    oblique_screen_keep = oblique_screen_keep,
    oblique_triage_maxit = oblique_triage_maxit,
    oblique_triage_improve_tol = oblique_triage_improve_tol
  )

  S_arg <- if (is.null(S)) NULL else .procrustes_validate_crossprod(S, k)
  T_arg <- if (is.null(T_init)) NULL else .procrustes_validate_t_init(T_init, k)

  out <- .oblique_procrustes(
    A = A,
    B = Target,
    S_r = S_arg,
    T_init_r = T_arg,
    eps = controls$oblique_eps,
    maxit = controls$oblique_maxit,
    max_line_search = controls$oblique_max_line_search,
    step0 = controls$oblique_step0,
    normalize = controls$oblique_normalize,
    random_starts = controls$oblique_random_starts,
    screen_keep = controls$oblique_screen_keep,
    triage_maxit = controls$oblique_triage_maxit,
    triage_improve_tol = controls$oblique_triage_improve_tol
  )

  out$method <- "oblique_procrustes"
  .procrustes_apply_dimnames(out, A, Target)
}


#' Consensus Procrustes alignment across multiple loading matrices
#'
#' Align several loading matrices to a common Procrustes consensus target.
#'
#' The function iterates between two steps:
#'
#' 1. each loading matrix is independently aligned to the current target with
#'    `PROCRUSTES()`; and
#' 2. the target is updated to the centroid, i.e., the elementwise average of the
#'    aligned matrices.
#'
#' This makes the target symmetric across imputations or samples: no single
#' solution is permanently privileged as the reference. The outer loop can stop
#' when the target stabilizes, when the consensus loss stabilizes, or when both
#' criteria are satisfied.
#'
#' If `multi_start = FALSE`, one consensus run is performed. If
#' `multi_start = TRUE`, the same single-start engine is repeated for the
#' selected starting targets, and the run with the smallest final mean loss is
#' returned as the main result. All runs and a between-run congruence summary are
#' retained in the `multi_start` component.
#'
#' @param unrotated_list List of unrotated loading matrices to be aligned. All
#'   matrices must be numeric, finite, and have identical dimensions.
#' @param init_targets Optional list of starting target matrices. These are
#'   typically rotated loading matrices from the corresponding analyses. If
#'   `NULL`, `unrotated_list` is used.
#' @param rotation Character string, either `"orthogonal"` or `"oblique"`.
#' @param start Either a single integer selecting an element of `init_targets`,
#'   or an explicit target matrix. Used when `multi_start = FALSE`.
#' @param multi_start Logical. If `FALSE`, perform one consensus-target run. If
#'   `TRUE`, repeat the single-start algorithm for each element of `starts`.
#' @param starts Integer vector selecting elements of `init_targets` used as
#'   starting targets when `multi_start = TRUE`. If `NULL`, all elements of
#'   `init_targets` are used. Duplicate entries are removed.
#' @param tol Positive relative Frobenius-norm convergence tolerance for the
#'   outer target update.
#' @param loss_tol Positive tolerance for the relative change in the outer
#'   consensus loss. If `NULL`, loss-based convergence is disabled. It cannot be
#'   `NULL` when `convergence` is `"loss"` or `"both"`.
#' @param loss_patience Positive integer. Number of consecutive iterations with
#'   relative loss change below `loss_tol` required for loss-based convergence.
#' @param convergence Character string controlling the stopping rule. `"either"`
#'   stops when either target or loss convergence is satisfied; `"target"` uses
#'   only target change; `"loss"` uses only loss change; `"both"` requires both.
#' @param min_iter Non-negative integer. Minimum number of outer iterations
#'   before convergence can be declared.
#' @param max_iter Positive integer. Maximum number of outer consensus
#'   iterations.
#' @param alpha Damping factor for the target update. `alpha = 1` uses the full
#'   centroid update. Smaller values, such as `0.5`, can reduce oscillation.
#' @param match_target Logical. If `TRUE`, the updated centroid is signed and
#'   column-matched to the previous target before convergence is evaluated.
#' @param hyper_cutoff Non-negative cutoff used by `.hyperplane_count()` for
#'   summary output.
#' @param oblique_maxit,oblique_eps,oblique_max_line_search,oblique_step0,oblique_normalize Parameters
#'   passed to `PROCRUSTES()` when `rotation = "oblique"`.
#' @param oblique_random_starts Number of random starts used by the inner oblique
#'   solver. See `oblique_random_starts_stage` for when these are used.
#' @param oblique_random_starts_stage Character string controlling whether random
#'   starts are used during the outer consensus loop, the final alignment pass,
#'   both, or neither. The default `"final"` keeps the outer loop smooth by using
#'   warm starts during iteration and applies random-start protection only in the
#'   final pass.
#' @param oblique_screen_keep,oblique_triage_maxit,oblique_triage_improve_tol
#'   Screening and triage parameters passed to the compiled oblique solver when
#'   random starts are used.
#' @param verbose Logical; if `TRUE`, print convergence messages for the outer
#'   loop.
#'
#' @returns A list with the converged target, aligned matrices, pooled loadings,
#'   pooled `Phi`, convergence history, inner-alignment diagnostics, and
#'   hyperplane summaries. If `multi_start = TRUE`, the `multi_start` element also
#'   contains the per-start losses, convergence indicators, run summaries, all
#'   run objects, and between-run Tucker congruence matrices.
#'
#' @export
#'
#' @references
#' Gower, J. C. (1975). Generalized Procrustes analysis. *Psychometrika*, 40,
#' 33-51.
CONSENSUS_PROCRUSTES <- function(unrotated_list,
                                 init_targets = NULL,
                                 rotation = c("orthogonal", "oblique"),
                                 start = 1,
                                 multi_start = FALSE,
                                 starts = NULL,
                                 tol = 1e-3,
                                 loss_tol = 1e-6,
                                 loss_patience = 5,
                                 convergence = c("either", "target", "loss", "both"),
                                 min_iter = 2,
                                 max_iter = 200,
                                 alpha = 1,
                                 match_target = TRUE,
                                 hyper_cutoff = 0.15,
                                 oblique_maxit = 500,
                                 oblique_eps = 1e-5,
                                 oblique_max_line_search = 10,
                                 oblique_step0 = 1,
                                 oblique_normalize = FALSE,
                                 oblique_random_starts = 0,
                                 oblique_random_starts_stage = c("final", "none", "outer", "both"),
                                 oblique_screen_keep = 2,
                                 oblique_triage_maxit = 25,
                                 oblique_triage_improve_tol = 0,
                                 verbose = FALSE) {
  rotation <- match.arg(rotation)
  convergence <- match.arg(convergence)
  oblique_random_starts_stage <- match.arg(oblique_random_starts_stage)
  multi_start <- .procrustes_check_flag(multi_start, "multi_start")

  single_args <- list(
    unrotated_list = unrotated_list,
    init_targets = init_targets,
    rotation = rotation,
    tol = tol,
    loss_tol = loss_tol,
    loss_patience = loss_patience,
    convergence = convergence,
    min_iter = min_iter,
    max_iter = max_iter,
    alpha = alpha,
    match_target = match_target,
    hyper_cutoff = hyper_cutoff,
    oblique_maxit = oblique_maxit,
    oblique_eps = oblique_eps,
    oblique_max_line_search = oblique_max_line_search,
    oblique_step0 = oblique_step0,
    oblique_normalize = oblique_normalize,
    oblique_random_starts = oblique_random_starts,
    oblique_random_starts_stage = oblique_random_starts_stage,
    oblique_screen_keep = oblique_screen_keep,
    oblique_triage_maxit = oblique_triage_maxit,
    oblique_triage_improve_tol = oblique_triage_improve_tol,
    verbose = verbose
  )

  if (!multi_start) {
    out <- do.call(
      .consensus_target_procrustes_single,
      c(single_args, list(start = start))
    )

    out$multi_start <- list(
      enabled = FALSE,
      starts = out$start,
      best_start = out$start,
      best_index = 1L,
      losses = out$mean_loss,
      converged = out$converged,
      iterations = out$iterations,
      summary = data.frame(
        start = out$start,
        loss = out$mean_loss,
        converged = out$converged,
        iterations = out$iterations
      ),
      runs = NULL,
      congruence = NULL
    )

    return(out)
  }

  if (is.null(init_targets)) {
    single_args$init_targets <- unrotated_list
    n_targets <- length(unrotated_list)
  } else {
    n_targets <- length(init_targets)
  }
  starts <- .procrustes_validate_starts(starts, n_targets)

  runs <- lapply(starts, function(s) {
    do.call(
      .consensus_target_procrustes_single,
      c(single_args, list(start = s))
    )
  })

  losses <- vapply(runs, function(x) x$mean_loss, numeric(1L))
  names(losses) <- paste0("start_", starts)

  finite_losses <- is.finite(losses)
  if (!any(finite_losses)) {
    stop("No multi-start consensus run produced a finite final loss.", call. = FALSE)
  }
  best_idx <- which.min(ifelse(finite_losses, losses, Inf))
  best <- runs[[best_idx]]

  n_runs <- length(runs)
  run_names <- paste0("start_", starts)
  congruence <- vector("list", n_runs)
  names(congruence) <- run_names

  for (i in seq_len(n_runs)) {
    congruence[[i]] <- vector("list", n_runs)
    names(congruence[[i]]) <- run_names
  }

  for (i in seq_len(n_runs)) {
    for (j in i:n_runs) {
      cij <- .tucker_congruence(
        runs[[i]]$pooled_loadings,
        runs[[j]]$pooled_loadings
      )
      congruence[[i]][[j]] <- cij
      congruence[[j]][[i]] <- if (i == j) cij else t(cij)
    }
  }

  run_summary <- data.frame(
    start = starts,
    loss = losses,
    converged = vapply(runs, function(x) isTRUE(x$converged), logical(1L)),
    iterations = vapply(runs, function(x) x$iterations, integer(1L)),
    final_failures = vapply(runs, function(x) length(x$final_failures), integer(1L)),
    row.names = run_names
  )

  best$multi_start <- list(
    enabled = TRUE,
    starts = starts,
    best_start = starts[best_idx],
    best_index = best_idx,
    losses = losses,
    converged = run_summary$converged,
    iterations = run_summary$iterations,
    summary = run_summary,
    runs = runs,
    congruence = congruence
  )

  best
}

#' Hull method for determining the number of factors to retain
#'
#' Implementation of the Hull method suggested by Lorenzo-Seva, Timmerman,
#' and Kiers (2011), with an extension to principal axis factoring. See details for
#' parallelization.
#'
#' @param x matrix or data.frame. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. Number of cases in the data. This is passed to \link{PARALLEL}.
#'  Only has to be specified if x is a correlation matrix, otherwise it is determined
#'  based on the dimensions of x.
#' @param n_fac_theor numeric. Theoretical number of factors to retain. The maximum
#'   of this number and the number of factors suggested by \link{PARALLEL} plus
#'   one will be used in the Hull method.
#' @param method character. The estimation method to use. One of  \code{"PAF"},
#'    \code{"ULS"}, or  \code{"ML"}, for principal axis factoring, unweighted
#'    least squares, and maximum likelihood, respectively.
#' @param gof character. The goodness of fit index to use. Either \code{"CAF"},
#'   \code{"CFI"}, or \code{"RMSEA"}, or any combination of them.
#'   If \code{method = "PAF"} is used, only
#'   the CAF can be used as goodness of fit index. For details on the CAF, see
#'   Lorenzo-Seva, Timmerman, and Kiers (2011).
#' @param eigen_type character. On what the eigenvalues should be found in the
#'  parallel analysis. Can be one of \code{"SMC"}, \code{"PCA"}, or \code{"EFA"}.
#'   If using  \code{"SMC"} (default), the diagonal of the correlation matrices is
#'    replaced by the squared multiple correlations (SMCs) of the indicators. If
#'     using  \code{"PCA"}, the diagonal values of the correlation
#'  matrices are left to be 1. If using  \code{"EFA"}, eigenvalues are found on the
#'  correlation  matrices with the final communalities of an EFA solution as
#'  diagonal. This is passed to  \code{\link{PARALLEL}}.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw data
#' is given as input. Default is \code{"pairwise.complete.obs"}.
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#'  Default is  \code{"pearson"}.
#' @param n_datasets numeric. The number of datasets to simulate. Default is 1000.
#'   This is passed to \code{\link{PARALLEL}}.
#' @param percent numeric. A vector of percentiles to take the simulated eigenvalues from.
#'  Default is 95. This is passed to \code{\link{PARALLEL}}.
#' @param decision_rule character. Which rule to use to determine the number of
#' factors to retain. Default is \code{"means"}, which will use the average
#' simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
#' in percent. \code{"crawford"} uses the 95th percentile for the first factor
#' and the mean afterwards (based on Crawford et al, 2010). This is passed to \code{\link{PARALLEL}}.
#' @param n_factors numeric. Number of factors to extract if  \code{"EFA"} is
#' included in \code{eigen_type}. Default is 1. This is passed to
#' \code{\link{PARALLEL}}.
#' @param ... Further arguments passed to \code{\link{EFA}}, also in
#' \code{\link{PARALLEL}}.
#'
#' @details The Hull method aims to find a model with an optimal balance between
#'  model fit and number of parameters. That is, it aims to retrieve only major
#'  factors (Lorenzo-Seva, Timmerman, & Kiers, 2011). To this end, it performs
#'  the following steps (Lorenzo-Seva, Timmerman, & Kiers, 2011, p.351):
#'  \enumerate{
#'    \item It performs parallel analysis and adds one to the identified number of factors (this number is denoted \emph{J}). \emph{J} is taken as an upper bound of the number of factors to retain in the hull method. Alternatively, a theoretical number of factors can be entered. In this case \emph{J} will be set to whichever of these two numbers (from parallel analysis or based on theory) is higher.
#'    \item For all 0 to \emph{J} factors, the goodness-of-fit (one of \emph{CAF}, \emph{RMSEA}, or \emph{CFI}) and the degrees of freedom (\emph{df}) are computed.
#'    \item The solutions are ordered according to their \emph{df}.
#'    \item Solutions that are not on the boundary of the convex hull are eliminated (see Lorenzo-Seva, Timmerman, & Kiers, 2011, for details).
#'    \item All the triplets of adjacent solutions are considered consecutively. The middle solution is excluded if its point is below or on the line connecting its neighbors in a plot of the goodness-of-fit versus the degrees of freedom.
#'    \item Step 5 is repeated until no solution can be excluded.
#'    \item The \emph{st} values of the “hull” solutions are determined.
#'    \item The solution with the highest \emph{st} value is selected.
#'  }
#'
#' The \link{PARALLEL} function and the principal axis factoring of the
#'   different number of factors can be parallelized using the future framework,
#'   by calling the \link[future:plan]{future::plan} function. The examples
#'    provide example code on how to enable parallel processing.
#'
#'   Note that if \code{gof = "RMSEA"} is used, 1 - RMSEA is actually used to
#'   compare the different solutions. Thus, the threshold of .05 is then .95. This
#'   is necessary due to how the heuristic to locate the elbow of the hull works.
#'
#'   The ML estimation method uses the \link[stats:factanal]{stats::factanal}
#'    starting values. See also the \link{EFA} documentation.
#'
#'    The \code{HULL} function can also be called together with other factor
#'    retention criteria in the \code{\link{N_FACTORS}} function.
#' @return A list of class HULL containing the following objects
#' \item{n_fac_CAF}{The number of factors to retain according to the Hull method
#' with the CAF.}
#' \item{n_fac_CFI}{The number of factors to retain according to the Hull method
#' with the CFI.}
#' \item{n_fac_RMSEA}{The number of factors to retain according to the Hull method
#' with the RMSEA.}
#' \item{solutions_CAF}{A matrix containing the CAFs, degrees of freedom, and for the factors lying on the hull, the st values of the hull solution (see Lorenzo-Seva, Timmerman, and Kiers 2011 for details).}
#' \item{solutions_CFI}{A matrix containing the CFIs, degrees of freedom, and for the factors lying on the hull, the st values of the hull solution (see Lorenzo-Seva, Timmerman, and Kiers 2011 for details).}
#' \item{solutions_RMSEA}{A matrix containing the RMSEAs, degrees of freedom, and for the factors lying on the hull, the st values of the hull solution (see Lorenzo-Seva, Timmerman, and Kiers 2011 for details).}
#' \item{n_fac_max}{The upper bound \emph{J} of the number of factors to extract (see details).}
#' \item{settings}{A list of the settings used.}
#'
#' @source Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. (2011).
#' The Hull method for selecting the number of common factors. Multivariate
#' Behavioral Research, 46(2), 340-364.
#'
#' @seealso Other factor retention criteria: \code{\link{CD}}, \code{\link{EKC}},
#' \code{\link{KGC}}, \code{\link{PARALLEL}}, \code{\link{SMT}}
#'
#' \code{\link{N_FACTORS}} as a wrapper function for this and all the
#' above-mentioned factor retention criteria.
#'
#' @export
#'
#' @examples
#' \donttest{
#' # using PAF (this will throw a warning if gof is not specified manually
#' # and CAF will be used automatically)
#' HULL(test_models$baseline$cormat, N = 500, gof = "CAF")
#'
#' # using ML with all available fit indices (CAF, CFI, and RMSEA)
#' HULL(test_models$baseline$cormat, N = 500, method = "ML")
#'
#' # using ULS with only RMSEA
#' HULL(test_models$baseline$cormat, N = 500, method = "ULS", gof = "RMSEA")
#'}
#'
#'\dontrun{
#' # using parallel processing (Note: plans can be adapted, see the future
#' # package for details)
#' future::plan(future::multisession)
#' HULL(test_models$baseline$cormat, N = 500, gof = "CAF")
#' }
HULL <- function(x, N = NA, n_fac_theor = NA,
                 method = c("PAF", "ULS", "ML"), gof = c("CAF", "CFI", "RMSEA"),
                 eigen_type = c("SMC", "PCA", "EFA"),
                 use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                         "everything", "na.or.complete"),
                 cor_method = c("pearson", "spearman", "kendall"),
                 n_datasets = 1000, percent = 95,
                 decision_rule = c("means", "percentile", "crawford"),
                 n_factors = 1, ...) {
  # Perform hull method following Lorenzo-Seva, Timmerman, and Kiers (2011)

  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  method <- match.arg(method)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  gof <- match.arg(gof, several.ok = TRUE)
  eigen_type <- match.arg(eigen_type)
  checkmate::assert_count(n_fac_theor, na.ok = TRUE)
  checkmate::assert_count(N, na.ok = TRUE)
  decision_rule <- match.arg(decision_rule)
  checkmate::assert_count(n_factors)
  checkmate::assert_count(n_datasets)
  checkmate::assert_number(percent, lower = 0, upper = 100)

  if (ncol(x) < 6) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red( "Data has fewer than 6 indicators. Hull method needs at least 6.\n"))
  }

  if (method == "PAF" && !all(gof == "CAF")) {
    cli::cli_alert_info(cli::col_cyan('Only CAF can be used as gof if method "PAF" is used. Setting gof to "CAF"\n'))
    gof <- "CAF"
  }

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  if (is.na(N)) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' "N" is not specified but is needed for computation of some of the fit indices.\n'))
  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' Correlation matrix is singular, the HULL method cannot be exectued.\n'))
  }

  # Check if correlation matrix is positive definite
  if(any(eigen(R)$values <= .Machine$double.eps^.6)){

    R <- psych::cor.smooth(R)

  }

  m <- ncol(R)

  # 1) perform parallel analysis to find J as n_fac_theor + 1
  par_res <- PARALLEL(R, N = N, eigen_type = eigen_type, method = method,
                      n_datasets = n_datasets, percent = percent,
                      decision_rule = decision_rule, n_factors = n_factors,
                      ...)

  if(eigen_type == "SMC"){
    n_fac_PA <- par_res$n_fac_SMC
  } else if(eigen_type == "PCA"){
    n_fac_PA <- par_res$n_fac_PCA
  } else if(eigen_type == "EFA"){
    n_fac_PA <- par_res$n_fac_EFA
  }

  if (is.na(n_fac_PA)) {

    if (!is.na(n_fac_theor)) {
      J <- n_fac_theor + 1
    } else {
      J <- .det_max_factors(ncol(R))
    }

  } else {

    J <- max(c(n_fac_PA, n_fac_theor), na.rm = TRUE) + 1

  }

  if (J > .det_max_factors(ncol(R))) {
    J <- .det_max_factors(ncol(R))
    warning(crayon::yellow$bold("!"), crayon::yellow(' Setting maximum number of factors to',
                                                     J, 'to ensure overidentified models.\n'))
  }

  if (J < 3) {
    warning(crayon::yellow$bold("!"),
            crayon::yellow(" Suggested maximum number of factors was", J,
                           "but must be at least 3 for hull method to work.",
                           "Setting it to 3.\n"))
    J <- 3

  }

  # 2) perform factor analysis for the range of dimensions 1:J and compute f and
  #    df for every solution
  s <- matrix(0, ncol = 4, nrow = J + 1)
  s[, 1] <- 0:J

  # first for 0 factors
  if ("CAF" %in% gof) {
    s_CAF <- s
    colnames(s_CAF) <- c("nfactors", "CAF", "df", "st")
    s_CAF[1, 2] <- 1 - KMO(R)$KMO
    s_CAF[1, 3] <- (m**2 - m) / 2
  }

  if ("CFI" %in% gof) {
    s_CFI <- s
    colnames(s_CFI) <- c("nfactors", "CFI", "df", "st")
    s_CFI[1, 2] <- 0
    s_CFI[1, 3] <- (m**2 - m) / 2

    # for later use in loop
    chi_null <- sum(R[upper.tri(R)] ^ 2) * (N - 1)
    df_null <- (m**2 - m) / 2
    delta_hat_null <- max(0, chi_null - df_null)

  }

  if ("RMSEA" %in% gof) {
    s_RMSEA <- s
    colnames(s_RMSEA) <- c("nfactors", "RMSEA", "df", "st")
    Fm <- sum(R[upper.tri(R)] ^ 2)
    chi <- Fm * (N - 1)
    df <- (m**2 - m) / 2
    # compute 1 - RMSEA
    s_RMSEA[1, 2] <- 1 - sqrt(max(0, chi - df) / (df * N - 1))
    s_RMSEA[1, 3] <- (m**2 - m) / 2

  }

  # Calculate loadings with EFA function
  loadings <- suppressWarnings(future.apply::future_lapply(seq_len(J), EFA,
                                                           x = R,
                                                           method = method,
                                                           N = N, ...,
                                                           future.seed = FALSE))

  # then for 1 to J factors
  for (i in seq_len(J)) {
    if (method == "PAF") {
      # compute goodness of fit "f" as CAF (common part accounted for; Eq 3)
      # compute CAF
      s_CAF[i + 1, 2] <- loadings[[i]]$fit_indices$CAF
      # compute dfs (Eq 4 provides the number of free parameters; using dfs yields
      # th same numbers, as the difference in df equals the difference in free
      # parameters)
      s_CAF[i + 1, 3] <- loadings[[i]]$fit_indices$df
    } else {
      if ("CAF" %in% gof) {
        # compute goodness of fit "f" as CAF (common part accounted for; Eq 3)
        # compute CAF
        s_CAF[i + 1, 2] <- loadings[[i]]$fit_indices$CAF
        # compute dfs (Eq 4 provides the number of free parameters; using dfs yields
        # th same numbers, as the difference in df equals the difference in free
        # parameters)
        s_CAF[i + 1, 3] <- loadings[[i]]$fit_indices$df
      }

      if ("CFI" %in% gof) {
        # compute CFI
        s_CFI[i + 1, 2] <- loadings[[i]]$fit_indices$CFI
        # compute dfs (Eq 4 provides the number of free parameters; using dfs yields
        # th same numbers, as the difference in df equals the difference in free
        # parameters)
        s_CFI[i + 1, 3] <- loadings[[i]]$fit_indices$df

      }

      if ("RMSEA" %in% gof) {
        # compute 1 - RMSEA
        s_RMSEA[i + 1, 2] <- 1 - loadings[[i]]$fit_indices$RMSEA
        # compute dfs (Eq 4 provides the number of free parameters; using dfs yields
        # th same numbers, as the difference in df equals the difference in free
        # parameters)
        s_RMSEA[i + 1, 3] <- loadings[[i]]$fit_indices$df

      }

    }

  }

  out_CAF <- list(s_complete = NA, retain = NA)
  out_CFI <- list(s_complete = NA, retain = NA)
  out_RMSEA <- list(s_complete = NA, retain = NA)

  if("CAF" %in% gof) {
    out_CAF <- .hull_calc(s = s_CAF, J = J, gof_t = "CAF")
  }
  if("CFI" %in% gof) {
    out_CFI <- .hull_calc(s = s_CFI, J = J, gof_t = "CFI")
  }
  if("RMSEA" %in% gof) {
    out_RMSEA <- .hull_calc(s = s_RMSEA, J = J, gof_t = "RMSEA")
  }

  out <- list(
    n_fac_CAF = out_CAF$retain,
    n_fac_CFI = out_CFI$retain,
    n_fac_RMSEA = out_RMSEA$retain,
    solutions_CAF = out_CAF$s_complete,
    solutions_CFI = out_CFI$s_complete,
    solutions_RMSEA = out_RMSEA$s_complete,
    n_fac_max = J,
    settings = list(N = N,
                    method = method,
                    gof = gof,
                    n_fac_theor = n_fac_theor,
                    eigen_type = eigen_type,
                    use = use,
                    cor_method = cor_method)
  )

  class(out) <- "HULL"

  return(out)

}


.hull_calc <- function(s, J, gof_t){

  # 3) sort n solutions by their df values and denoted by s (already done)

  # 4) all solutions s are excluded for which a solution sj (j<i) exists such
  #    that fj > fi (eliminate solutions not on the boundary of the convex hull)

  s_complete <- s
  d_s <- diff(s[, 2])
  while (any(d_s < 0)) {
    s <- s[c(1, d_s) > 0, , drop = FALSE]
    if(nrow(s) == 1){
      break
    }
    d_s <- diff(s[, 2])
  }

  if (nrow(s) < 3) {
    warning(crayon::yellow$bold("!"),
            crayon::yellow(" Less than three solutions located on the hull have",
            "been identified when using", gof_t, "as goodness of fit index.",
                           "Proceeding by taking the value with the maximum",
                           gof_t, "as heuristic. You may want to consider",
                           "additional indices or methods as robustness check.\n"))

    # combine values
    for (row_i in 0:J) {

      if (row_i %in% s[,1]) {
        s_complete[row_i + 1, 4] <- s[s[,1] == row_i, 4]
      } else {
        s_complete[row_i + 1, 4] <- NA
      }

    }
    s_complete[, 4] <- rep(NA, nrow(s_complete))

    # 8) select solution with highest gof value
    retain <- s[which.max(s[, 2]), 1]


  } else {


    # 5) all triplets of adjacent solutions are considered consecutively.
    #    middle solution is excluded if its point is below or on the line
    #    connecting its neighbors in GOF vs df

    # 6) repeat 5) until no solution can be excluded

    nr_s <- nrow(s)
    i <- 2

    while(i < nr_s - 1) {

      f1 <- s[i - 1, 2]
      f2 <- s[i, 2]
      f3 <- s[i + 1, 2]
      df1 <- s[i - 1, 3]
      df2 <- s[i, 3]
      df3 <- s[i + 1, 3]

      # compute f2 if it were on the line between f1 and f3
      p_f2 <- f1 + (f3 - f1) / (df3 - df1) * (df2 - df1)

      # check if f2 is below or on the predicted line and if so, remove it
      if (f2 <= p_f2) {
        s <- s[-i, ]
        nr_s <- nr_s -1
        i <- 1
      }
      i <- i + 1
    }


    # 7) the st values of the hull solutions are determined (Eq 5)
    for (i in 2:(nrow(s) - 1)) {

      f_i <- s[i, 2]
      f_p <- s[i - 1, 2]
      f_n <- s[i + 1, 2]
      df_i <- s[i, 3]
      df_p <- s[i - 1, 3]
      df_n <- s[i + 1, 3]

      s[i, 4] <- ((f_i - f_p) / (df_i - df_p)) / ((f_n - f_i) / (df_n - df_i))

    }

    # combine values
    for (row_i in 0:J) {

      if (row_i %in% s[,1]) {
        s_complete[row_i + 1, 4] <- s[s[,1] == row_i, 4]
      } else {
        s_complete[row_i + 1, 4] <- NA
      }

    }

    # 8) select solution with highest st value
    retain <- s[which.max(s[, 4]), 1]

  }


  out <- list(s_complete = s_complete,
              retain = unname(retain))

  return(out)

}

hull_cor_paf <- suppressMessages(HULL(test_models$baseline$cormat, N = 500))
hull_cor_ml <- HULL(test_models$baseline$cormat, N = 500, method = "ML")
hull_cor_uls <- HULL(test_models$baseline$cormat, N = 500, method = "ULS")
hull_cor_uls_CFI <- HULL(test_models$baseline$cormat, N = 500, method = "ULS",
                         gof = "CFI")
hull_cor_ml_nf <- suppressWarnings(HULL(test_models$baseline$cormat, N = 500,
                                        method = "ML", n_fac_theor = 12))

hull_PCA <- HULL(test_models$baseline$cormat, N = 500, method = "ML",
                 eigen_type = "PCA")
hull_EFA <- HULL(test_models$baseline$cormat, N = 500, method = "ML",
                 eigen_type = "EFA")

hull_raw_paf <- suppressMessages(suppressWarnings(HULL(GRiPS_raw)))
hull_raw_ml <- suppressMessages(suppressWarnings(HULL(GRiPS_raw, method = "ML")))
hull_raw_uls <- suppressMessages(suppressWarnings(HULL(GRiPS_raw, method = "ULS")))
hull_raw_uls_CFI <- suppressMessages(suppressWarnings(HULL(GRiPS_raw,
                                                           method = "ULS",
                                                           gof = "CFI")))
hull_raw_ml_nf <- suppressMessages(suppressWarnings(HULL(GRiPS_raw, N = 500,
                                                         method = "ML",
                                                         n_fac_theor = 7)))


test_that("output class and dimensions are correct", {
  expect_is(hull_cor_paf, "HULL")
  expect_output(str(hull_cor_paf), "List of 8")
  expect_is(hull_cor_ml, "HULL")
  expect_output(str(hull_cor_ml), "List of 8")
  expect_is(hull_cor_uls, "HULL")
  expect_output(str(hull_cor_uls), "List of 8")
  expect_is(hull_cor_uls_CFI, "HULL")
  expect_output(str(hull_cor_uls_CFI), "List of 8")
  expect_is(hull_PCA, "HULL")
  expect_output(str(hull_PCA), "List of 8")
  expect_is(hull_EFA, "HULL")
  expect_output(str(hull_EFA), "List of 8")

  expect_is(hull_raw_paf, "HULL")
  expect_output(str(hull_raw_paf), "List of 8")
  expect_is(hull_raw_ml, "HULL")
  expect_output(str(hull_raw_ml), "List of 8")
  expect_is(hull_raw_uls, "HULL")
  expect_output(str(hull_raw_uls), "List of 8")
  expect_is(hull_raw_uls_CFI, "HULL")
  expect_output(str(hull_raw_uls_CFI), "List of 8")
  expect_is(hull_PCA, "HULL")
  expect_output(str(hull_PCA), "List of 8")
  expect_is(hull_EFA, "HULL")
  expect_output(str(hull_EFA), "List of 8")

  expect_output(str(hull_cor_paf$settings), "List of 7")
  expect_output(str(hull_cor_ml$settings), "List of 7")
  expect_output(str(hull_cor_uls$settings), "List of 7")
  expect_output(str(hull_cor_uls_CFI$settings), "List of 7")
  expect_output(str(hull_PCA$settings), "List of 7")
  expect_output(str(hull_EFA$settings), "List of 7")

  expect_output(str(hull_raw_paf$settings), "List of 7")
  expect_output(str(hull_raw_ml$settings), "List of 7")
  expect_output(str(hull_raw_uls$settings), "List of 7")
  expect_output(str(hull_raw_uls_CFI$settings), "List of 7")
  expect_output(str(hull_PCA$settings), "List of 7")
  expect_output(str(hull_EFA$settings), "List of 7")
})

test_that("n_fac_max is correctly specified", {
  expect_lte(hull_cor_paf$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_cor_ml$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_cor_uls$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_cor_uls_CFI$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_PCA$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_EFA$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))

  expect_lte(hull_raw_paf$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))
  expect_lte(hull_raw_ml$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))
  expect_lte(hull_raw_uls$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))
  expect_lte(hull_raw_uls_CFI$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))

  expect_equal(hull_cor_ml_nf$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_equal(hull_raw_ml_nf$n_fac_max,
               .det_max_factors(ncol(GRiPS_raw)))
})

test_that("solution matrices are correctly returned", {
  expect_is(hull_cor_paf$solutions_CAF, "matrix")
  expect_equal(hull_cor_paf$solutions_CFI, NA)
  expect_equal(hull_cor_paf$solutions_RMSEA, NA)

  expect_is(hull_cor_uls$solutions_CAF, "matrix")
  expect_is(hull_cor_uls$solutions_CFI, "matrix")
  expect_is(hull_cor_uls$solutions_RMSEA, "matrix")

  expect_is(hull_cor_uls_CFI$solutions_CFI, "matrix")
  expect_equal(hull_cor_uls_CFI$solutions_CAF, NA)
  expect_equal(hull_cor_uls_CFI$solutions_RMSEA, NA)

  expect_is(hull_raw_paf$solutions_CAF, "matrix")
  expect_equal(hull_raw_paf$solutions_CFI, NA)
  expect_equal(hull_raw_paf$solutions_RMSEA, NA)

  expect_is(hull_raw_uls$solutions_CAF, "matrix")
  expect_is(hull_raw_uls$solutions_CFI, "matrix")
  expect_is(hull_raw_uls$solutions_RMSEA, "matrix")

  expect_is(hull_raw_uls_CFI$solutions_CFI, "matrix")
  expect_equal(hull_raw_uls_CFI$solutions_CAF, NA)
  expect_equal(hull_raw_uls_CFI$solutions_RMSEA, NA)
})

test_that("n_factors are correctly returned", {
  expect_equal(hull_cor_paf$n_fac_CAF, 3)
  expect_equal(hull_cor_paf$n_fac_CFI, NA)
  expect_equal(hull_cor_paf$n_fac_RMSEA, NA)

  expect_equal(hull_cor_uls$n_fac_CAF, 3)
  expect_equal(hull_cor_uls$n_fac_CFI, 3)
  expect_equal(hull_cor_uls$n_fac_RMSEA, 3)

  expect_equal(hull_cor_uls_CFI$n_fac_CAF, NA)
  expect_equal(hull_cor_uls_CFI$n_fac_CFI, 3)
  expect_equal(hull_cor_uls_CFI$n_fac_RMSEA, NA)

  expect_equal(hull_raw_paf$n_fac_CAF, 1)
  expect_equal(hull_raw_paf$n_fac_CFI, NA)
  expect_equal(hull_raw_paf$n_fac_RMSEA, NA)

  expect_equal(hull_raw_uls$n_fac_CAF, 1)
  expect_equal(hull_raw_uls$n_fac_CFI, 1)
  expect_equal(hull_raw_uls$n_fac_RMSEA, 1)

  expect_equal(hull_raw_uls_CFI$n_fac_CAF, NA)
  expect_equal(hull_raw_uls_CFI$n_fac_CFI, 1)
  expect_equal(hull_raw_uls_CFI$n_fac_RMSEA, NA)
})


# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z, rnorm(10), rnorm(10), rnorm(10)), ncol = 6)
cor_sing <- stats::cor(dat_sing)

burt <- matrix(c(1.00,  0.83,  0.81,  0.80,   0.71, 0.70, 0.54, 0.53,  0.59,  0.24, 0.13,
                 0.83,  1.00,  0.87,  0.62,   0.59, 0.44, 0.58, 0.44,  0.23,  0.45,  0.21,
                 0.81,  0.87,  1.00,  0.63,   0.37, 0.31, 0.30, 0.12,  0.33,  0.33,  0.36,
                 0.80,  0.62,  0.63,  1.00,   0.49, 0.54, 0.30, 0.28,  0.42,  0.29, -0.06,
                 0.71,  0.59,  0.37,  0.49,   1.00, 0.54, 0.34, 0.55,  0.40,  0.19, -0.10,
                 0.70,  0.44,  0.31,  0.54,   0.54, 1.00, 0.50, 0.51,  0.31,  0.11,  0.10,
                 0.54,  0.58,  0.30,  0.30,   0.34, 0.50, 1.00, 0.38,  0.29,  0.21,  0.08,
                 0.53,  0.44,  0.12,  0.28,   0.55, 0.51, 0.38, 1.00,  0.53,  0.10, -0.16,
                 0.59,  0.23,  0.33,  0.42,   0.40, 0.31, 0.29, 0.53,  1.00, -0.09, -0.10,
                 0.24,  0.45,  0.33,  0.29,   0.19, 0.11, 0.21, 0.10, -0.09,  1.00,  0.41,
                 0.13,  0.21,  0.36, -0.06,  -0.10, 0.10, 0.08, -0.16, -0.10, 0.41,  1.00),
               nrow = 11, ncol = 11)

test_that("errors etc are thrown correctly", {
  expect_error(HULL(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_message(suppressWarnings(HULL(GRiPS_raw)), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_warning(suppressMessages(HULL(GRiPS_raw)), " Suggested maximum number of factors was 2 but must be at least 3 for hull method to work. Setting it to 3.\n")
  expect_warning(suppressMessages(HULL(IDS2_R, N = 20)), " Less than three solutions located on the hull have been identified when using CAF as goodness of fit index. Proceeding by taking the value with the maximum CAF as heuristic. You may want to consider additional indices or methods as robustness check.\n")
  expect_warning(suppressMessages(HULL(GRiPS_raw, N = 20)), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(HULL(test_models$baseline$cormat), ' "N" is not specified but is needed for computation of some of the fit indices.\n')
  expect_error(HULL(test_models$baseline$cormat, method = "ML"), ' "N" is not specified but is needed for computation of some of the fit indices.\n')
  expect_error(HULL(test_models$baseline$cormat, method = "ULS"), ' "N" is not specified but is needed for computation of some of the fit indices.\n')

  expect_error(HULL(dat_sing, method = "ML"), ' Correlation matrix is singular, the HULL method cannot be exectued.\n')
  expect_error(HULL(cor_sing, N = 20, method = "ML"), ' Correlation matrix is singular, the HULL method cannot be exectued.\n')

  expect_error(HULL(matrix(rnorm(50), ncol = 5)), "Data has fewer than 6 indicators. Hull method needs at least 6.\n")

  expect_message(suppressWarnings(HULL(GRiPS_raw)), 'Only CAF can be used as gof if method "PAF" is used. Setting gof to "CAF"\n')

  expect_warning(HULL(test_models$baseline$cormat, n_fac_theor = 13, N = 500), ' Setting maximum number of factors to 12 to ensure overidentified models.\n')
  expect_warning(HULL(burt, N = 20, method = "ML"), "Matrix was not positive definite, smoothing was done")
  expect_warning(HULL(burt, N = 20, method = "ML", n_fac_theor = 1), "Matrix was not positive definite, smoothing was done")
})

rm(hull_cor_paf, hull_cor_ml, hull_cor_uls, hull_cor_uls_CFI, hull_raw_paf,
   hull_raw_ml, hull_raw_uls, hull_raw_uls_CFI, hull_raw_ml_nf, hull_cor_ml_nf,
   hull_PCA, hull_EFA, x, y, z, dat_sing, cor_sing, burt)

#' Schmid-Leiman Transformation
#'
#' This function implements the Schmid-Leiman (SL) transformation
#' (Schmid & Leiman, 1957). It takes the pattern coefficients and factor
#' intercorrelations from an oblique factor solution as
#' input and can reproduce the results from \code{\link[psych:schmid]{psych::schmid}}
#' and from the SPSS implementation from Wolff & Preising (2005). Other arguments
#' from \code{\link{EFA}} can be used to control the procedure to find the
#' second-order loadings more flexibly. The function can also be used on a
#' second-order confirmatory factor analysis (CFA) solution from lavaan.
#'
#' @param x object of class \code{\link{EFA}}, class \code{\link[psych:fa]{psych::fa}},
#' class \code{\link[lavaan]{lavaan}} or matrix. If class \code{\link{EFA}} or
#' class \code{\link[psych:fa]{psych::fa}}, pattern coefficients and factor
#' intercorrelations are taken from this object. If class \code{\link[lavaan]{lavaan}},
#' it must be a second-order CFA solution. In this case first-order and second-order
#'  factor loadings are taken from this object and the \code{g_name} argument has
#'  to be specified.
#' x can also be a pattern matrix from an oblique factor solution (see \code{Phi})
#' or a matrix of first-order factor loadings from a higher-order confirmatory factor
#' analysis (see \code{L2}).
#' @param Phi matrix. A matrix of factor intercorrelations from an oblique factor
#' solution. Only needs to be specified if a pattern matrix is entered directly
#' into \code{x}.
#' @param type character. One of "EFAtools" (default), "psych", "SPSS", or "none".
#' This is used to control the procedure of the second-order factor analysis. See
#' \code{\link{EFA}} for details.
#' @param method character. One of "PAF", "ML", or "ULS" to use
#' principal axis factoring, maximum likelihood, or unweighted least squares
#' (also called minres), respectively, used in \code{\link{EFA}} to find the second-order
#' loadings.
#' @param g_name character. The name of the general factor. This needs only be
#' specified if \code{x} is a \code{lavaan} second-order solution. Default is "g".
#' @param ... Arguments to be passed to \code{\link{EFA}}.
#'
#' @details
#' The SL transformation (also called SL orthogonalization) is a procedure with
#' which an oblique factor solution is transformed into a hierarchical,
#' orthogonalized solution. As a first step, the factor intercorrelations are
#' again factor analyzed to find second-order factor loadings. If there is only
#' one higher-order factor, this step of the procedure stops there, resulting in
#' a second-order factor structure. The first-order factor and the second-order
#' factor are then orthogonalized, resulting in an orthogonalized factor solution
#' with proportionality constraints. The procedure thus makes a suggested
#' hierarchical data structure based on factor intercorrelations explicit. One
#' major advantage of SL transformation is that it enables variance
#' partitioning between higher-order and first-order factors, including the
#' calculation of McDonald's omegas (see \code{\link{OMEGA}}).
#'
#' @return A list of class SL containing the following
#' \item{orig_R}{Original correlation matrix.}
#' \item{sl}{A matrix with general factor loadings, group factor loadings, communalities,
#' and uniquenesses.}
#' \item{L2}{Second-order factor loadings.}
#' \item{vars_accounted}{A matrix of explained variances and sums of squared loadings.}
#' \item{iter}{The number of iterations needed for convergence in EFA.}
#' \item{settings}{list. The settings (arguments) used in EFA to get the
#' second-order loadings.}
#'
#' @source Schmid, J. & Leiman, J. M. (1957). The development of hierarchical
#' factor solutions. Psychometrika, 22(1), 53–61. doi:10.1007/BF02289209
#' @source Wolff, H.-G., & Preising, K. (2005). Exploring item and higher order
#' factor structure with the Schmid-Leiman solution: Syntax codes for SPSS and
#' SAS. Behavior Research Methods, 37 , 48–58. doi:10.3758/BF03206397
#'
#' @export
#'
#' @examples
#' ## Use with an output from the EFAtools::EFA function, both with type EFAtools
#' EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                type = "EFAtools", method = "PAF", rotation = "promax")
#' SL_EFAtools <- SL(EFA_mod, type = "EFAtools", method = "PAF")
#'
#' \donttest{
#' ## Use with an output from the psych::fa function with type psych in SL
#' fa_mod <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
#'                     fm = "pa", rotate = "Promax")
#' SL_psych <- SL(fa_mod, type = "psych", method = "PAF")
#' }
#'
#' ## Use more flexibly by entering a pattern matrix and phi directly (useful if
#' ## a factor solution found with another program should be subjected to SL
#' ## transformation)
#'
#' ## For demonstration, take pattern matrix and phi from an EFA output
#' ## This gives the same solution as the first example
#' EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                type = "EFAtools", method = "PAF", rotation = "promax")
#' SL_flex <- SL(EFA_mod$rot_loadings, Phi = EFA_mod$Phi, type = "EFAtools",
#'               method = "PAF")
#'
#' \donttest{
#' ## Use with a lavaan second-order CFA output
#'
#' # Create and fit model in lavaan (assume all variables have SDs of 1)
#' mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
#'         F2 =~ V7 + V8 + V9 + V10 + V11 + V12
#'         F3 =~ V13 + V14 + V15 + V16 + V17 + V18
#'         g =~ F1 + F2 + F3'
#' fit <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
#'                    sample.nobs = 500, estimator = "ml")
#'
#' SL_lav <- SL(fit, g_name = "g")
#'
#' }
SL <- function(x, Phi = NULL, type = c("EFAtools", "psych", "SPSS", "none"),
               method = c("PAF", "ML", "ULS"), g_name = "g", ...) {

  # Perform argument checks
  checkmate::assert_matrix(Phi, null.ok = TRUE)
  type <- match.arg(type)
  method <- match.arg(method)
  checkmate::assert_string(g_name)

  if(!inherits(x, c("EFA", "fa", "lavaan", "matrix", "LOADINGS", "loadings"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither an object of class EFA, fa, or lavaan, nor a matrix, nor of class LOADINGS or loadings.\n"))

  }

  if(inherits(x, "EFA")) {

    if("Phi" %in% names(x)){

      L1 <- x$rot_loadings
      n_first_fac <- ncol(x$rot_loadings)
      orig_R <- x$orig_R

      if(!is.null(Phi)){
        warning(crayon::yellow$bold("!"), crayon::yellow(" Phi argument is specified. Specified factor intercorrelations are taken. To take factor intercorrelations from the EFA output, leave Phi = NULL\n"))

      } else {

        Phi <- x$Phi

      }

    } else {

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is either a non-rotated or orthogonal factor solution. SL needs an oblique factor solution\n"))

    }

    n_order <- order(as.numeric(gsub("F", "", colnames(L1))))
    L1 <- L1[, n_order]
    Phi <- Phi[n_order, n_order]

  } else if(inherits(x, "fa")) {

    if("Phi" %in% names(x)){

      L1 <- unclass(x$loadings)
      n_first_fac <- ncol(x$loadings)
      orig_R <- unclass(x$r)

      if(!is.null(Phi)){
        warning(crayon::yellow$bold("!"), crayon::yellow(" Phi argument is specified. Specified factor intercorrelations are taken. To take factor intercorrelations from the psych fa output, leave Phi = NULL\n"))

      } else {

        Phi <- x$Phi

      }

    } else {

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is either a non-rotated or orthogonal factor solution. SL needs an oblique factor solution\n"))

    }

    n_order <- order(colnames(L1))
    L1 <- L1[, n_order]
    Phi <- Phi[n_order, n_order]

  } else if(inherits(x, "lavaan")){

    if(lavaan::lavInspect(x, what = "converged") == FALSE){
      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Model did not converge. No omegas are computed.\n"))
    }

    std_sol <- suppressWarnings(lavaan::lavInspect(x, what = "std"))

    if(any(is.na(std_sol$lambda))){
      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Some loadings are NA or NaN. No omegas are computed.\n"))
    }

    if(any(std_sol$lambda >= 1)){
      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" A Heywood case was detected (loading equal to or larger than 1). No omegas are computed.\n"))
    }

    # Create list with factor and corresponding subtest names
    col_names <- colnames(std_sol$lambda)

    if(!any(col_names %in% g_name)){
      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Could not find the specified name of the general factor in the entered lavaan solution. Please check the spelling.\n"))
    }

    if(!all(std_sol$lambda[, g_name] == 0)){

      warning(crayon::yellow$bold("!"), crayon::yellow(" The second-order factor you specified contains first-order loadings. Did you really enter a second-order CFA solution? Or did you enter the wrong factor name in g_name?\n"))

    }

    col_names <- col_names[!col_names %in% g_name]
    fac_names <- c(g_name, col_names)

    n_first_fac <- length(col_names)

  } else {

    if(is.null(Phi)){

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Phi not provided. Either enter an oblique factor solution from EFAtools::EFA or from psych::fa, or a second-order CFA solution from lavaan, or provide Phi\n"))

    }

    if (!is.null(colnames(x))) {
      n_order <- order(colnames(x))
      x <- x[, n_order]

      if(!is.null(Phi)){
      Phi <- Phi[n_order, n_order]
      }

    }

    L1 <- x
    n_first_fac <- ncol(x)
    orig_R <- NA

  }

  if(inherits(x, "lavaan")){

    # Calculate direct g loadings
    L1 <- std_sol$lambda[, col_names]
    L2 <- std_sol$beta[col_names, g_name]

    L_sls_2 <- L1 %*% L2

    # Calculate direct group factor loadings
    L_sls_1 <- L1 %*% sqrt(std_sol$psi[col_names, col_names])

    orig_R <- NA
    iter <- NA
    settings <- NA

  } else {

    # perform a factor analysis on the intercorrelation matrix of the first order
    # factors (N is only specified to avoid a warning)
    EFA_phi <- suppressWarnings(EFA(Phi, n_factors = 1, N = 100, type = type,
                                    method = method, rotation = "none", ...))

    if (ncol(Phi) <= 2) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" The second-order EFA is underidentified.\n"))
    }

    iter <- EFA_phi$iter
    settings <- EFA_phi$settings

    # extract second order loadings
    L2 <- EFA_phi$unrot_loadings

    # Schmid-Leiman solution, direct loadings of second order factor
    L_sls_2 <- L1 %*% L2

    # compute uniqueness of higher order factor
    u2_h <- sqrt(1 - diag(L2 %*% t(L2)))

    # Schmid-Leiman solution, residualized first order factor loadings
    L_sls_1 <- L1 %*% diag(u2_h)

  }

  # Combine the Schmid-Leiman loadings in a data frame
  sl_load <- cbind(L_sls_2, L_sls_1)

  # Compute communalities and uniquenesses of the Schmid-Leiman solution
  h2_sl <- rowSums(sl_load^2)
  u2_sl <- 1 - h2_sl

  vars_accounted <- .compute_vars(L_unrot = sl_load, L_rot = sl_load)

  colnames(vars_accounted) <-c("g", paste0("F", seq_len(n_first_fac)))

  # Finalize output object
  sl <- cbind(sl_load, h2_sl, u2_sl)
  colnames(sl) <- c("g", paste0("F", seq_len(n_first_fac)), "h2", "u2")
  class(sl) <- "SLLOADINGS"

  output <- list(
    orig_R = orig_R,
    sl = sl,
    L2 = L2,
    vars_accounted = vars_accounted,
    iter = iter,
    settings = settings
    )

  class(output) <- "SL"

  output

}

#' Woodcock Johnson IV: ages 40 to 90 plus
#'
#' A list containing the bivariate correlations (N = 1,146) of the 47
#' cognitive and achievement subtests from the WJ IV for 40- to 90+-year-olds
#' from the standardization sample obtained from
#' the WJ-IV technical manual (McGrew, LaForte, & Schrank, 2014).
#' Tables are reproduced with permission from the publisher.
#'
#'
#' @format A list of 2 with elements "cormat" (47 x 47 matrix of bivariate correlations)
#' and "N". The correlation matrix contains the following variables:
#' \describe{
#'   \item{ORLVOC}{(numeric) - Oral Vocabulary.}
#'   \item{NUMSER}{(numeric) - Number Series.}
#'   \item{VRBATN}{(numeric) - Verbal Attention.}
#'   \item{LETPAT}{(numeric) - Letter-Pattern Matching.}
#'   \item{PHNPRO}{(numeric) - Phonological Processing.}
#'   \item{STYREC}{(numeric) - Story Recall.}
#'   \item{VISUAL}{(numeric) - Visualization.}
#'   \item{GENINF}{(numeric) - General Information.}
#'   \item{CONFRM}{(numeric) - Concept Formation.}
#'   \item{NUMREV}{(numeric) - Numbers Reversed.}
#'   \item{NUMPAT}{(numeric) - Number-Pattern Matching.}
#'   \item{NWDREP}{(numeric) - Nonword Repetition.}
#'   \item{VAL}{(numeric) - Visual-Auditory Learning.}
#'   \item{PICREC}{(numeric) - Picture Recognition.}
#'   \item{ANLSYN}{(numeric) - Analysis-Synthesis.}
#'   \item{OBJNUM}{(numeric) - Object-Number Sequencing.}
#'   \item{PAIRCN}{(numeric) - Pair Cancellation.}
#'   \item{MEMWRD}{(numeric) - Memory for Words.}
#'   \item{PICVOC}{(numeric) - Picture Vocabulary.}
#'   \item{ORLCMP}{(numeric) - Oral Comprehension.}
#'   \item{SEGMNT}{(numeric) - Segmentation.}
#'   \item{RPCNAM}{(numeric) - Rapid Picture Naming.}
#'   \item{SENREP}{(numeric) - Sentence Repetition.}
#'   \item{UNDDIR}{(numeric) - Understanding Directions.}
#'   \item{SNDBLN}{(numeric) - Sound Blending.}
#'   \item{RETFLU}{(numeric) - Retrieval Fluency.}
#'   \item{SNDAWR}{(numeric) - Sound Awareness.}
#'   \item{LWIDNT}{(numeric) - Letter-Word Identification.}
#'   \item{APPROB}{(numeric) - Applied Problems.}
#'   \item{SPELL}{(numeric) - Spelling.}
#'   \item{PSGCMP}{(numeric) - Passage Comprehension.}
#'   \item{CALC}{(numeric) - Calculation.}
#'   \item{WRTSMP}{(numeric) - Writing Samples.}
#'   \item{WRDATK}{(numeric) - Word Attack.}
#'   \item{ORLRDG}{(numeric) - Oral Reading.}
#'   \item{SNRDFL}{(numeric) - Sentence Reading Fluency.}
#'   \item{MTHFLU}{(numeric) - Math Facts Fluency.}
#'   \item{SNWRFL}{(numeric) - Sentence Writing Fluency.}
#'   \item{RDGREC}{(numeric) - Reading Recall.}
#'   \item{NUMMAT}{(numeric) - Number Matrices.}
#'   \item{EDIT}{(numeric) - Editing.}
#'   \item{WRDFLU}{(numeric) - Word Reading Fluency.}
#'   \item{SPLSND}{(numeric) - Spelling of Sounds.}
#'   \item{RDGVOC}{(numeric) - Reading Vocabulary.}
#'   \item{SCI}{(numeric) - Science.}
#'   \item{SOC}{(numeric) - Social Studies.}
#'   \item{HUM}{(numeric) - Humanities.}
#'  }
#' @source McGrew, K. S., LaForte, E. M., & Schrank, F. A. (2014). Technical
#'  Manual. Woodcock-Johnson IV. Rolling Meadows, IL: Riverside.
#'
#' @source Schrank, F. A., McGrew, K. S., & Mather, N. (2014). Woodcock-Johnson IV.
#' Rolling Meadows, IL: Riverside.
"WJIV_ages_40_90"

# Declarative control objects for the estimation and rotation tuning knobs.
#
# `estimate_control()` and `rotate_control()` bundle the tuning arguments that are
# otherwise passed one by one to a fit. Each returns a validated, classed list that
# is a 1:1 declarative surface over the settings resolved internally by
# `.estimate_model()` (via `.PAF()`) and `.rotate_model()`. The constructors only
# validate and store: they carry the chosen `type` preset and the declared knobs
# (with `NA` marking "unset, resolve from the preset") and leave the actual preset
# resolution -- and any override warning -- to the fitting cores, because which
# preset applies depends on the estimator and rotation, which a control object does
# not know.

# TRUE when a knob is left unset: a single genuine NA, the "resolve from the type
# preset" sentinel. NaN is deliberately not treated as unset (is.na(NaN) is TRUE) so a
# stray NaN falls through to the type check and is rejected rather than silently stored.
.is_control_unset <- function(x) length(x) == 1L && is.na(x) && !is.nan(x)

# Index of a single knob value among its choices, case-folded and allowing unambiguous
# abbreviations. NA marks "no usable match": a non-string input, no match at all, or -- from
# charmatch()'s 0 -- an abbreviation that prefixes several choices.
.control_choice_index <- function(x, choices) {
  if (!checkmate::test_string(x)) return(NA_integer_)
  idx <- charmatch(.fold_upper(x), .fold_upper(choices))
  if (!is.na(idx) && idx == 0L) NA_integer_ else idx
}

# Resolve a choice-valued knob (or NA, when `na_ok`) against its choices, case-insensitively
# and allowing unambiguous abbreviations as everywhere else on the interface, and return the
# canonical spelling so the stored knob and the printed control always carry it. Classed abort
# otherwise.
.assert_control_choice <- function(x, choices, arg, na_ok = TRUE) {
  if (na_ok && .is_control_unset(x)) return(x)
  idx <- .control_choice_index(x, choices)
  if (is.na(idx)) {
    na_txt <- if (na_ok) " (or NA to resolve from the type preset)" else ""
    cli::cli_abort(
      "{.arg {arg}} must be one of {.val {choices}}{na_txt}.",
      class = "efa_control_input"
    )
  }
  choices[idx]
}

# Validate a numeric knob against its admissible range (optionally a whole number, optionally
# NA); classed abort otherwise. `positive` demands x > 0; `lower` bounds it inclusively and
# `upper` inclusively unless `upper_strict`, which demands x < upper. The bounds mirror the ones
# the fitting core enforces, so a control object can never carry a value the fit would go on to
# reject.
.assert_control_number <- function(x, arg, positive = TRUE, lower = -Inf, upper = Inf,
                                   na_ok = TRUE, int = FALSE, upper_strict = FALSE) {
  if (na_ok && .is_control_unset(x)) return(invisible(NULL))
  ok <- if (int) checkmate::test_int(x) else checkmate::test_number(x, finite = TRUE)
  if (isTRUE(ok) && positive) ok <- x > 0
  if (isTRUE(ok)) ok <- x >= lower && if (upper_strict) x < upper else x <= upper
  if (!isTRUE(ok)) {
    kind <- if (int) "whole number" else "number"
    bounds <- character()
    if (positive) bounds <- c(bounds, "greater than 0")
    if (!positive && is.finite(lower)) bounds <- c(bounds, paste("of at least", lower))
    if (is.finite(upper)) {
      bounds <- c(bounds, paste(if (upper_strict) "smaller than" else "of at most", upper))
    }
    bound <- if (length(bounds)) paste0(" ", paste(bounds, collapse = " and ")) else ""
    na_txt <- if (na_ok) " (or NA to resolve from the type preset)" else ""
    cli::cli_abort(
      "{.arg {arg}} must be a single {kind}{bound}{na_txt}.",
      class = "efa_control_input"
    )
  }
  invisible(NULL)
}

# Resolve a choice-valued knob that the flat interface partial-matched, keeping that
# convenience: an abbreviation resolves to the full value and NA stays unset (it means "not
# needed by this estimator"). Returns the resolved value; classed abort otherwise.
.match_control_choice <- function(x, choices, arg) {
  if (.is_control_unset(x)) return(x)
  idx <- .control_choice_index(x, choices)
  if (is.na(idx)) {
    cli::cli_abort(
      "{.arg {arg}} must be one of {.val {choices}} (or NA to leave it unset).",
      class = "efa_control_input"
    )
  }
  choices[idx]
}

# ASCII-only upper-casing for case-folded matching. toupper() is locale-dependent -- under a
# Turkish/Azeri locale toupper("i") is the dotted capital "İ", so an input like "minres"
# or "oblimin" would fold to a string the (already-uppercase) choice can never equal. The
# choice values are all ASCII, so folding both sides with a fixed a-z -> A-Z map keeps the
# matching identical on every platform.
.fold_upper <- function(x) chartr("abcdefghijklmnopqrstuvwxyz",
                                  "ABCDEFGHIJKLMNOPQRSTUVWXYZ", x)

# Case-insensitive match.arg(): resolve a choice-valued argument against its choices and
# return the canonical spelling(s) from `choices`, so downstream code, stored settings, and
# printed output always carry the canonical value. Values are compared case-folded, and an
# exact (case-folded) match beats a prefix match, so unambiguous abbreviations still resolve
# as they do under match.arg(). It is deliberately stricter than match.arg() in `several.ok`
# mode: an unmatched element aborts here, whereas match.arg() silently drops it. Mirroring
# match.arg(), an `arg` identical to `choices` selects the default -- the first choice, or
# all of them when `several.ok` -- and a missing `choices` is taken from the caller's formal
# default; a NULL `arg` selects the documented default too, see the comment on that branch.
# Ambiguous or unmatched values are a classed abort listing the choices; `class` lets a
# caller raise its own condition class (the control constructors reuse their
# `efa_control_input`).
.match_arg_ci <- function(arg, choices, several.ok = FALSE,
                          arg_name = deparse1(substitute(arg)),
                          class = "efa_bad_choice") {
  caller <- sys.parent()
  from_formals <- missing(choices)
  if (from_formals) {
    choices <- eval(formals(sys.function(caller))[[arg_name]],
                    envir = sys.frame(caller))
  }
  # NULL means "use the default", as in match.arg(), but the default is the caller's formal
  # default, NOT choices[1L]: the two part company under `several.ok` whenever the formal
  # default is a whole vector (efa_hull()'s `gof`, efa_kgc()'s `eigen_type`), and wherever an
  # explicit `choices` is a superset of the formal default (efa_retain()'s `criteria`) or
  # orders it differently (`eigen_type_other`). Taking choices[1L] there silently runs
  # something narrower than the help page documents, with no condition raised. A formal
  # without a default, or one that is itself NULL (efa_simulate()'s `match`, documented as
  # resolving to "thresholds"), falls back to choices[1L].
  if (is.null(arg)) {
    if (from_formals) {
      # `choices` already is that default -- no need to look it up a second time
      arg <- choices
    } else {
      # Keep the default inside a length-one list while testing it: a formal without a
      # default is the empty symbol, and binding that to a variable (or passing it on)
      # forces it into R's "argument is missing" error. Comparing the enclosing list never
      # evaluates it.
      default <- formals(sys.function(caller))[arg_name]
      arg <- if (identical(unname(as.list(default)), list(quote(expr = )))) {
        NULL
      } else {
        eval(default[[1L]], envir = sys.frame(caller))
      }
    }
    if (is.null(arg)) return(choices[1L])
  }
  if (identical(arg, choices)) return(if (several.ok) choices else choices[1L])

  # charmatch(): NA = no match, 0 = ambiguous (a prefix of several choices)
  idx <- if (is.character(arg)) charmatch(.fold_upper(arg), .fold_upper(choices)) else NA_integer_
  len_ok <- if (several.ok) length(arg) >= 1L else length(arg) == 1L
  if (!len_ok || anyNA(idx) || any(idx == 0L)) {
    no_match <- if (is.character(arg)) unique(arg[is.na(idx)]) else character()
    ambiguous <- if (is.character(arg)) unique(arg[!is.na(idx) & idx == 0L]) else character()
    what <- if (several.ok) "one or more" else "one"
    cli::cli_abort(
      c("{.arg {arg_name}} must be {what} of {.val {choices}}.",
        if (!is.character(arg)) c("x" = "You supplied {.obj_type_friendly {arg}}."),
        if (!several.ok && is.character(arg) && length(arg) > 1L) {
          c("x" = "You supplied {length(arg)} values.")
        },
        if (length(no_match) > 0L) {
          c("x" = "{.val {no_match}} {?does/do} not match any of the choices.")
        },
        if (length(ambiguous) > 0L) {
          c("x" = "{.val {ambiguous}} {?matches/match} more than one choice.")
        }),
      class = class
    )
  }
  choices[idx]
}

# Case-insensitively map the elements of a vector-valued choice argument onto the canonical
# spelling of `choices` (exact case-folded matches only -- no abbreviations, matching the
# subset semantics of the callers). An element matching no choice is returned unchanged, so
# the caller's subset assertion still rejects it with its usual message.
.map_subset_ci <- function(x, choices) {
  if (!is.character(x)) return(x)
  idx <- match(.fold_upper(x), .fold_upper(choices))
  found <- !is.na(idx)
  x[found] <- choices[idx[found]]
  x
}

# Validate a logical knob (or NA, when `na_ok`); classed abort otherwise.
.assert_control_flag <- function(x, arg, na_ok = TRUE) {
  if (na_ok && .is_control_unset(x)) return(invisible(NULL))
  if (!checkmate::test_flag(x)) {
    na_txt <- if (na_ok) " (or NA to resolve from the type preset)" else ""
    cli::cli_abort(
      "{.arg {arg}} must be a single logical value{na_txt}.",
      class = "efa_control_input"
    )
  }
  invisible(NULL)
}

#' Control objects for estimation and rotation settings
#'
#' `estimate_control()` and `rotate_control()` collect the estimation and rotation
#' *tuning* arguments of a factor analysis into two small, validated objects. They
#' are a declarative surface over the same settings resolved internally by the
#' package's estimation and rotation engines, so that a fit's many tuning knobs can
#' be prepared, inspected, and reused as a single value instead of being passed one
#' by one.
#'
#' Each argument that governs a `type` preset defaults to `NA`, meaning "leave this
#' knob to the preset". Setting `type` to one of `"EFAtools"`, `"psych"`, or
#' `"SPSS"` fills those knobs from the corresponding preset when the fit is run;
#' setting `type = "none"` requires the relevant knobs to be supplied explicitly.
#' The control object only records the chosen `type` and the knobs you set: the
#' preset is resolved (and any "argument set alongside `type`" warning issued) when
#' the object is used to fit a model, exactly as it is today, because which preset
#' applies depends on the estimator and rotation.
#'
#' @param type character. One of `"EFAtools"` (default), `"psych"`, `"SPSS"`, or
#'   `"none"`. Selects the preset that fills the `NA`-defaulted knobs below when the
#'   control is used to fit a model.
#' @param init_comm character. Method for the initial communalities in principal
#'   axis factoring: `"smc"` (squared multiple correlations), `"mac"` (maximum
#'   absolute correlations), or `"unity"`. `NA` (default) resolves from `type`.
#' @param criterion numeric. The convergence criterion for principal axis
#'   factoring: iteration stops once the change in communalities falls below it. A
#'   single number greater than 0 and smaller than 1; `NA` (default) resolves from `type`.
#' @param criterion_type character. The convergence criterion type for principal
#'   axis factoring: `"max_individual"` (the largest change in any communality, as
#'   in SPSS) or `"sum"` (the change in the summed communalities, as in
#'   [psych::fa()]). `NA` (default) resolves from `type`.
#' @param max_iter numeric. The maximum number of principal-axis-factoring
#'   iterations before the procedure is halted with a warning. A single whole
#'   number of at least 1; `NA` (default) resolves from `type`.
#' @param abs_eigen logical. Which algorithm the principal-axis-factoring
#'   iterations use: `FALSE` computes the loadings from the eigenvalues (as in
#'   [psych::fa()]); `TRUE` uses the absolute eigenvalues (as in SPSS). `NA`
#'   (default) resolves from `type`.
#' @param start_method character. Starting values for the maximum-likelihood
#'   optimiser: `"psych"` (default, the [psych::fa()] starts) or `"factanal"` (the
#'   [stats::factanal()] starts); abbreviations are matched. Not governed by `type`.
#'   Only maximum likelihood uses it, so `NA` leaves it unset and is rejected only by
#'   a fit that is actually run with `estimator = "ML"`.
#' @param fiml_max_iter numeric. The maximum number of EM iterations used to estimate the
#'   two-stage full-information maximum-likelihood moments from raw data with missing
#'   values (`cor_method = "fiml"`); the last iterate is returned, with a warning, if the
#'   cap is reached. A single whole number of at least 1; default `500`. Not governed by
#'   `type`, and unused by every other correlation method. The EM converges linearly and
#'   needs more iterations the larger the fraction of missing information, so raise it when
#'   a fit reports that the moments did not converge.
#' @param fiml_tol numeric. The convergence tolerance of that EM: iteration stops once the
#'   largest change in the standardized moments (the standardized means, log-variances, and
#'   correlations) falls below it, so it does not depend on the variables' measurement
#'   scale. A single number greater than 0 and smaller than 1 (at or above 1 the criterion is
#'   met immediately and the starting moments would be returned as converged); default `1e-5`.
#'   Not governed by `type`.
#'
#' @returns `estimate_control()` returns a list of class `efa_estimate_control` with
#'   the components `type`, `init_comm`, `criterion`, `criterion_type`, `max_iter`,
#'   `abs_eigen`, `start_method`, `fiml_max_iter`, and `fiml_tol`.
#'   `rotate_control()` returns a list of class
#'   `efa_rotate_control` with the components `type`, `normalize`, `precision`,
#'   `order_type`, `varimax_type`, `p_type`, `k`, `random_starts`, and `extra_args`
#'   (a named list of any additional arguments forwarded to the rotation engine).
#'
#' @seealso [efa_fit()], which takes both controls; [efa_retain()], the retention
#'   criteria, and [efa_schmid_leiman()], which take an `estimate_control` for the
#'   fits they run.
#'
#' @family Control functions
#'
#' @examples
#' # Estimation knobs taken entirely from a preset:
#' estimate_control(type = "SPSS")
#'
#' # A preset with one knob pinned to a non-preset value:
#' estimate_control(type = "EFAtools", max_iter = 500)
#'
#' # Every knob supplied explicitly (type = "none"):
#' estimate_control(type = "none", init_comm = "smc", criterion = 1e-3,
#'                  criterion_type = "sum", max_iter = 300, abs_eigen = TRUE)
#'
#' @export
estimate_control <- function(type = c("EFAtools", "psych", "SPSS", "none"),
                             init_comm = NA, criterion = NA, criterion_type = NA,
                             max_iter = NA, abs_eigen = NA, start_method = "psych",
                             fiml_max_iter = 500, fiml_tol = 1e-5) {

  # Match `type` case-insensitively, but keep the constructor's own condition class so a bad
  # `type` is caught alongside the other knobs (all `efa_control_input`).
  type <- .match_arg_ci(type, class = "efa_control_input")

  init_comm <- .assert_control_choice(init_comm, c("smc", "mac", "unity"), "init_comm")
  # The convergence tolerance is compared against a change in communalities, so a value at or
  # above 1 stops the iterations immediately; the fit rejects it (see .PAF()), and the control
  # must not be able to carry a value the fit would go on to reject.
  .assert_control_number(criterion, "criterion", upper = 1, upper_strict = TRUE)
  criterion_type <- .assert_control_choice(criterion_type, c("max_individual", "sum"),
                                           "criterion_type")
  .assert_control_number(max_iter, "max_iter", int = TRUE)
  .assert_control_flag(abs_eigen, "abs_eigen")
  # `start_method` only governs the ML optimiser, so NA ("not needed here") is admissible and
  # is rejected by the fit itself, and only when `estimator = "ML"`.
  start_method <- .match_control_choice(start_method, c("psych", "factanal"), "start_method")
  # The two FIML knobs are not preset-driven, so they carry real defaults rather than the NA
  # sentinel and NA is not admissible. The tolerance is bounded above at 1 for the same reason
  # `criterion` is: it is compared against a change in standardized moments -- standardized means,
  # log-variances, and correlations -- so a value at or above 1 is met on the first iteration and
  # the EM would return its starting moments while reporting itself converged. Unlike `criterion`
  # the fit does not reject such a value downstream (the EM only demands tol > 0), so the bound has
  # to be enforced here or not at all.
  .assert_control_number(fiml_max_iter, "fiml_max_iter", int = TRUE, na_ok = FALSE)
  .assert_control_number(fiml_tol, "fiml_tol", upper = 1, upper_strict = TRUE, na_ok = FALSE)

  structure(
    list(type = type, init_comm = init_comm, criterion = criterion,
         criterion_type = criterion_type, max_iter = max_iter,
         abs_eigen = abs_eigen, start_method = start_method,
         fiml_max_iter = fiml_max_iter, fiml_tol = fiml_tol),
    class = "efa_estimate_control"
  )
}

#' @rdname estimate_control
#'
#' @param normalize logical. If `TRUE` (default), a Kaiser normalization is
#'   performed before the rotation. The one knob that is always on unless you turn
#'   it off with `FALSE`.
#' @param precision numeric. The convergence tolerance of the rotation procedure. A
#'   single number greater than 0 and at most 1; default `1e-5`. Each rotation stage
#'   monitors its own quantity, so the same number is not the same tolerance everywhere:
#'   `varimax_type = "kaiser"` stops on the *absolute* change in the varimax simplicity
#'   criterion, which is an average over variables (and so does not scale with how many
#'   there are) but rises with the number of factors, roughly toward
#'   `1 - 1 / n_factors`, so a fixed value is a relatively weaker tolerance the more
#'   factors are extracted; `varimax_type = "svd"` stops on the *relative* change in the
#'   singular values (as in [stats::varimax()]); and the criterion rotations fitted by
#'   gradient projection stop when the *projected-gradient norm* falls below it. Promax
#'   inherits whichever of the two varimax tests its `varimax_type` selects, because it
#'   rotates a varimax base.
#' @param order_type character. How the factors are ordered: `"eigen"` (by
#'   descending sum of squared loadings, as in [psych::fa()]) or `"ss_factors"` (by
#'   descending unweighted sum of squared loadings). `NA` (default) resolves from
#'   `type`.
#' @param varimax_type character. The varimax variant used (for the varimax and
#'   promax rotations): `"svd"` (as in [stats::varimax()]) or `"kaiser"` (the SPSS
#'   / Kaiser (1958) procedure). `NA` (default) resolves from `type`.
#' @param p_type character. How the promax target matrix is computed: `"unnorm"`
#'   (the unnormalized target of Hendrickson & White (1964), also used by psych and
#'   stats) or `"norm"` (the normalized target used by SPSS). `NA` (default)
#'   resolves from `type`.
#' @param k numeric. The promax power (for the target matrix) or the number of
#'   near-zero loadings for simplimax. A single number greater than 0; `NA`
#'   (default) leaves it to the fit (the `type`-dependent promax value, or
#'   `nrow(loadings)` for simplimax). Simplimax counts loadings, so a fit using it
#'   additionally requires a whole number no larger than the number of loadings in
#'   the solution; promax's power has no such restriction.
#' @param random_starts numeric. The number of random starts used by the
#'   criterion-based rotations to guard against local minima. A single whole number
#'   of at least 0, where `0` runs the rotation from its warm start only; default
#'   `100`. The default suffices for the smooth criteria; `simplimax` remains
#'   materially start-dependent at it, so raise it there (see the *Rotations* section
#'   of [efa_fit()]).
#' @param ... Additional arguments forwarded to the rotation engine. Only the names
#'   a rotation engine can consume are accepted: `maxit` (a whole number of at least 0
#'   bounding a *single* gradient-projection optimization -- the multi-start search runs
#'   several of them and each is bounded separately, so it is not a budget for the run as a
#'   whole; varimax and promax have no such stage and take only `precision`), and the
#'   criterion parameters `gam` (oblimin; `gam = 0` is the
#'   recommended default, and larger values increasingly reward correlated factors and
#'   can drive the solution toward factor collapse, so inspect `Phi` before interpreting
#'   a fit with `gam > 0`) and `delta` (geomin; a positive number, default `0.01`);
#'   anything else is rejected as a
#'   misspelling. They are stored in `extra_args` and passed on to the rotation engine
#'   when the control is used to fit a model; an extra a given fit's rotation does not
#'   consume is ignored by that fit, so one control can serve fits with different
#'   rotations. An estimation knob (which belongs in [estimate_control()]) or one of the
#'   former spellings `P_type` and `randomStarts` is likewise rejected here, because the
#'   fit would silently drop it.
#'
#' @examples
#'
#' # Rotation knobs taken from a preset:
#' rotate_control(type = "psych")
#'
#' # A criterion-specific extra argument, forwarded to the rotation engine:
#' rotate_control(type = "EFAtools", k = 3, gam = 0.5)
#'
#' @export
rotate_control <- function(type = c("EFAtools", "psych", "SPSS", "none"),
                           normalize = TRUE, precision = 1e-5, order_type = NA,
                           varimax_type = NA, p_type = NA, k = NA,
                           random_starts = 100, ...) {

  type <- .match_arg_ci(type, class = "efa_control_input")

  .assert_control_flag(normalize, "normalize", na_ok = FALSE)
  .assert_control_number(precision, "precision", upper = 1, na_ok = FALSE)
  order_type <- .assert_control_choice(order_type, c("eigen", "ss_factors"), "order_type")
  varimax_type <- .assert_control_choice(varimax_type, c("svd", "kaiser"), "varimax_type")
  p_type <- .assert_control_choice(p_type, c("norm", "unnorm"), "p_type")
  .assert_control_number(k, "k")
  # 0 is a meaningful setting, not a missing one: it runs the rotation from its warm start
  # only, with no random restarts.
  .assert_control_number(random_starts, "random_starts", positive = FALSE, lower = 0,
                         int = TRUE, na_ok = FALSE)

  extra_args <- list(...)
  # A misplaced tuning knob is not a rotation engine extra: its name collides with a formal
  # of the fit, so the extras splice in efa_fit() would drop it and the fit would quietly run
  # the preset value instead. Refuse to carry it rather than let it be silently ignored.
  bad_est <- intersect(names(extra_args), .flat_estimate_knobs)
  bad_flat <- intersect(names(extra_args), c("P_type", "randomStarts"))
  if (length(bad_est) > 0L || length(bad_flat) > 0L) {
    cli::cli_abort(
      c("{.arg {c(bad_est, bad_flat)}} cannot be passed to {.fn rotate_control} as a rotation
         engine extra.",
        if (length(bad_est) > 0L) c("i" = "The estimation knobs live in {.fn estimate_control}."),
        if (length(bad_flat) > 0L) c("i" = "The former {.arg P_type} and {.arg randomStarts} are
                                            the formals {.arg p_type} and {.arg random_starts}.")),
      class = "efa_control_input"
    )
  }
  # Any other extra must be a name some rotation's engine can consume (the engines read their
  # extras by exact name); an unknown one is a misspelling that the fit would silently drop.
  # Deliberately validated against the across-rotation union, not a single rotation: a control
  # is reusable across fits, so it may legitimately carry an extra that some of its fits (e.g.
  # the unrotated or promax rows of an efa_average() grid) never consume.
  extra_nms <- names(extra_args)
  if (is.null(extra_nms)) extra_nms <- rep("", length(extra_args))
  bad_extra <- unique(setdiff(extra_nms[nzchar(extra_nms)], .rotation_extra_union))
  if (length(bad_extra) > 0L || any(!nzchar(extra_nms))) {
    msg <- if (length(bad_extra) > 0L) {
      "{.arg {bad_extra}} {?is/are} not {?a name/names} a rotation engine can consume."
    } else {
      "The extra engine arguments in {.arg ...} must be named."
    }
    cli::cli_abort(
      c(msg,
        "i" = "The accepted engine extras are {.arg {(.rotation_extra_union)}}.",
        "i" = "Check for a misspelled engine argument (for example {.arg gam}, not
               {.arg gamma}); the rotation tuning knobs are {.fn rotate_control}'s named
               arguments."),
      class = "efa_control_input"
    )
  }

  structure(
    list(type = type, normalize = normalize, precision = precision,
         order_type = order_type, varimax_type = varimax_type, p_type = p_type,
         k = k, random_starts = random_starts, extra_args = extra_args),
    class = "efa_rotate_control"
  )
}

# The flat estimation and rotation tuning knobs the pre-control interface exposed, split by the
# control object each one now lives in.
.flat_estimate_knobs <- c("init_comm", "criterion", "criterion_type", "max_iter",
                          "abs_eigen", "start_method")
.flat_rotate_knobs <- c("normalize", "precision", "order_type", "varimax_type", "p_type",
                        "k", "random_starts")

# The efa_fit() arguments that govern its inference machinery (standard errors and the
# random-number stream feeding them). They are efa_fit() formals, so a caller forwarding
# `...` into a fit accepts them by construction -- which is wrong wherever the fit is only
# an internal step whose standard errors are discarded (see .reject_inference_dots()).
.fit_inference_args <- c("se", "b_boot", "ci", "seed")

# Reject a flat tuning knob that reached a fit as a bare dot. No fitting function has a formal
# for any of these any more -- they live in the control objects -- so a bare copy would be taken
# for a rotation extra and dropped, and the fit would quietly run the default preset instead
# of the requested settings. Takes the argument NAMES (`...names()`), so the dots are never
# forced; `fn` names the function the caller actually reached, so the message points at the call
# that has to change.
.reject_flat_knobs <- function(nms, fn = "efa_fit") {
  # The renamed `method` argument is checked first: it is the more specific mistake, and
  # every function that must reject the flat knobs must reject the former estimator
  # spelling too, so one call site covers both.
  .reject_renamed_method(nms, fn = fn)
  bad <- intersect(nms, c("type", .flat_estimate_knobs, .flat_rotate_knobs,
                          "P_type", "randomStarts"))
  if (length(bad) == 0L) return(invisible(NULL))
  # point the example at the constructor the offending knob actually lives in
  example <- if (all(bad %in% c(.flat_rotate_knobs, "P_type", "randomStarts"))) {
    paste0(fn, "(x, ..., rotate_control = rotate_control(k = 3))")
  } else {
    paste0(fn, "(x, ..., estimate_control = estimate_control(max_iter = 500))")
  }
  cli::cli_abort(
    c("{.arg {bad}} cannot be passed to {.fn {fn}} directly.",
      "i" = "The estimation and rotation tuning knobs live in {.fn estimate_control} and
             {.fn rotate_control}.",
      "i" = "For example: {.code {example}}."),
    class = "efa_flat_knob_in_dots"
  )
}

# Reject a `...` name that no fit could ever consume. The retention criteria and their
# orchestrator forward their `...` into efa_fit(), but only when a fit actually runs (e.g.
# eigen_type includes "EFA", or a criterion that fits a model is selected). A misspelled name
# is otherwise silently dropped and the criterion runs with the default it looks like it was
# told to change. A name that is neither an efa_fit() formal nor a rotation-engine extra is
# such a misspelling and is refused here, independently of whether a fit runs. Takes the
# argument NAMES (`...names()`) so the dots are never forced; `fn` names the function reached.
#
# It must stay lenient for the superseded wrappers, which repack their flat tuning knobs into
# `estimate_control` / `rotate_control` objects and splice those into these same dots: both
# control names are efa_fit() formals, so they are in the accepted set by construction.
.reject_unknown_fit_dots <- function(nms, fn, unrotated = FALSE) {
  nms <- nms[nzchar(nms)]
  if (length(nms) == 0L) return(invisible(NULL))
  known <- c(setdiff(names(formals(efa_fit)), "..."), .rotation_extra_union)
  # A caller whose fits are always unrotated runs no rotation engine, so an engine extra
  # can never be consumed there. Refusing it here, rather than letting it reach the fit,
  # is what makes the error name the function the user called and arrive before the
  # criterion does its work (`rotation` itself is refused by .reject_rotation_dots()).
  # efa_fit()'s inference arguments are refused for the same reason, but with a message of
  # their own: they ARE efa_fit() arguments, so "check for a misspelled name" would be wrong.
  if (unrotated) {
    known <- setdiff(known, .rotation_extra_union)
    .reject_inference_dots(nms, fn = fn)
  }
  # setdiff() already returns unique values, so the offenders are named once each
  bad <- setdiff(nms, known)
  if (length(bad) == 0L) return(invisible(NULL))
  cli::cli_abort(
    c("{.arg {bad}} {?is/are} not {?an argument/arguments} of {.fn {fn}} or {.fn efa_fit}.",
      "i" = "Check for a misspelled argument name."),
    class = "efa_unused_dots"
  )
}

# Reject efa_fit()'s inference arguments in the `...` of a caller whose fits are internal
# steps. Two distinct hazards, both silent before this guard:
#   * `se`, `b_boot`, `ci`: the fit does compute the standard errors -- they are efa_fit()
#     formals, so they pass any whitelist built from them -- but the caller keeps only a few
#     quantities from the fit and never reports its standard errors, so the result is
#     identical and the work is thrown away (measurably: an information-based SE roughly
#     doubles an efa_kgc() run).
#   * `seed`: efa_fit() seeds itself with .set_local_seed(), which restores the caller's
#     stream when the fit returns, so the seed is spent on a fit that draws no random
#     numbers. Where the caller draws at all, that draw still comes from the caller's stream.
#     The name therefore looks like it pins the run and pins nothing at all; set.seed() is
#     what governs it.
# The hints must hold for every caller, so neither claims that this particular call fits, or
# simulates, or is a retention criterion: efa_retain() with a criterion such as MAP fits no
# model at all, efa_kgc() and efa_scree() fit one but simulate nothing, and
# efa_schmid_leiman() is not a criterion and keeps its internal fit's loadings rather than
# eigenvalues.
# Takes the argument NAMES (`...names()`), so the dots are never forced.
.reject_inference_dots <- function(nms, fn) {
  bad <- intersect(nms, .fit_inference_args)
  if (length(bad) == 0L) return(invisible(NULL))
  msg <- c("{.arg {bad}} cannot be passed to {.fn {fn}}.")
  if (any(c("se", "b_boot", "ci") %in% bad)) {
    msg <- c(msg, "i" = "Any fit run here is an internal step whose standard errors are not
                         reported, so they would be computed and then discarded.")
  }
  if ("seed" %in% bad) {
    msg <- c(msg, "i" = "The fits run here draw no random numbers, so {.arg seed} has
                         nothing to govern. Whatever else the call draws comes from the
                         caller's stream: call {.fn set.seed} before {.fn {fn}}.")
  }
  cli::cli_abort(msg, class = "efa_unused_dots")
}

# The retention criteria's fits are always unrotated, so a rotation setting passed through
# their `...` is meaningless: a user's `rotation = ...` makes a criterion fit spin up a
# rotation engine for nothing -- and a criterion-based rotation draws its random starts from
# the caller's stream, so it silently moves a seeded result -- while a `rotate_control` is
# simply ignored by the unrotated fit. Refuse
# both. The superseded N_FACTORS() wrapper still repacks a frozen `type` (or a flat rotation
# knob) into a rotate_control() object that legitimately rides through these dots, so a
# `rotate_control` that IS such a control object is exempt, as is a `NULL` (efa_fit()'s own
# "not supplied" default, which the repack passes through untouched) -- only some other value
# under that name, or any `rotation`, is refused. Takes the dots by value for that check.
.reject_rotation_dots <- function(dots, fn) {
  nms <- names(dots)
  if (is.null(nms)) return(invisible(NULL))
  bad <- character(0)
  if ("rotation" %in% nms) bad <- c(bad, "rotation")
  rc <- dots[nms == "rotate_control"]
  ok_rc <- vapply(rc, function(v) is.null(v) || inherits(v, "efa_rotate_control"), logical(1))
  if (!all(ok_rc)) bad <- c(bad, "rotate_control")
  if (length(bad) == 0L) return(invisible(NULL))
  cli::cli_abort(
    c("{.arg {bad}} cannot be passed to {.fn {fn}}.",
      "i" = "The fits run by the retention criteria are always unrotated."),
    class = "efa_unused_dots"
  )
}

# Reject the former `method` argument (the estimator is selected with `estimator`) when it
# lands in a fit function's `...`: it would otherwise be forwarded into the rotation extras
# or a criterion fit -- surfacing as an opaque downstream error or being silently ignored --
# instead of selecting the estimator.
.reject_renamed_method <- function(nms, fn = "efa_fit") {
  if (!"method" %in% nms) return(invisible(NULL))
  cli::cli_abort(
    c("{.arg method} is not an argument of {.fn {fn}}.",
      "i" = "The estimator is selected with {.arg estimator}, e.g.
             {.code {fn}(x, ..., estimator = \"ML\")}."),
    class = "efa_renamed_arg"
  )
}

# Validate an estimation control (NULL means "the efa_fit() default"). Shared by efa_fit() and by
# the functions that only pass the control on to a fit, so a bogus object is rejected even when
# the call ends up running no fit at all (e.g. efa_kgc() with eigen_type = "PCA").
.assert_estimate_control <- function(x) {
  if (is.null(x) || inherits(x, "efa_estimate_control")) return(invisible(NULL))
  cli::cli_abort(
    "{.arg estimate_control} must be a control object from {.fn estimate_control}.",
    class = "efa_control_input"
  )
}

# The legacy alias spellings .repack_flat_dots() translates rather than forwards. Kept as a
# named constant so the drop filter below and the repack agree by construction.
.flat_alias_names <- c("method", "type", "P_type", "randomStarts")

# The frozen wrappers that forward their `...` (N_FACTORS(), PARALLEL(), KGC(), SCREE(),
# HULL(), NEST(), EFA_POOLED()) keep the flat interface's silent-ignore contract: a name
# that nothing on the old surface could ever consume was dropped without comment, so it is
# dropped here too -- before .repack_flat_dots() translates the flat knobs and before the
# successors' guards (.reject_unknown_fit_dots(), .reject_rotation_dots(), efa_fit()'s
# per-rotation extras whitelist) would refuse it. The kept universe is everything the old
# surface consumed: the efa_fit() formals, the flat tuning knobs, and the alias spellings.
# Keeping the flat knobs matters when a control object rides alongside one (the repack then
# translates nothing), so such a mixed call still fails loudly in the successor's flat-knob
# guard instead of losing the knob here. The retention wrappers' criterion fits are always
# unrotated (the default), so a `rotation` setting and the rotation-engine extras
# (`maxit`/`gam`/`delta`) were never consumable there -- 0.8.0 ignored them without effect on
# the result -- and are dropped with the junk; EFA_POOLED() instead passes `unrotated =
# FALSE` plus the extras its selected rotation's engine reads, mirroring EFA()'s own
# rotation-aware filter. Keeping a name here is not the same as accepting it: efa_fit()'s
# inference arguments (`se`/`b_boot`/`ci`/`seed`) pass this filter as efa_fit() formals and
# are then refused by .reject_inference_dots() in the retention successors. That is a
# DELIBERATE exception to the silent-ignore rule above, not an instance of it: unlike the
# rotation extras, the flat EFA() did take all four and a pre-rename call could name them.
# It changes no legacy result, because on this path they were inert -- these wrappers hand
# EFA() a correlation matrix, which never reached its seeding branch at all, and a standard
# error the fit computed was thrown away by the criterion (`se` even aborted outright through
# some of them: "sandwich" for want of an acov, "information" through KGC()/SCREE() for want
# of an N). See .reject_inference_dots() for the two hazards themselves.
# A successor-only name carrying a malformed value (e.g.
# `rotate_control = "SPSS"`) is deliberately KEPT: no pre-rename code could have used that
# name, so the successor's loud classed validation is the right outcome, exactly as EFA()
# rejects the successor-only names outright. Unnamed dots pass through untouched, as on the
# reject side.
.drop_unknown_frozen_dots <- function(dots, extras = character(), unrotated = TRUE) {
  nms <- names(dots)
  if (is.null(nms)) return(dots)
  known <- c(setdiff(names(formals(efa_fit)), "..."),
             .flat_estimate_knobs, .flat_rotate_knobs, .flat_alias_names, extras)
  if (unrotated) known <- setdiff(known, "rotation")
  # charmatch(), not %in%: the flat interface reached its target through do.call(), where R
  # partial-matched an abbreviated name against the formals ahead of `...`, so an
  # abbreviation was a name the old surface could consume. Dropping it here would turn such
  # a call into a silently different fit; keeping it (an ambiguous abbreviation too, which
  # base R would also have refused) lets it travel on and fail loudly instead.
  dots[!nzchar(nms) | !is.na(charmatch(nms, known))]
}

# Repack the flat tuning knobs a call may still carry in its `...` into the control objects the
# fitting interface now takes. `efa_fit()` has no flat tuning formals left, so a bare knob
# forwarded through `...` would otherwise be dropped and the fit would quietly run the default
# preset instead of the requested one. Returns the arguments to splice into the call: the two
# controls -- built only when the caller actually named `type` or a knob -- followed by every
# remaining dot (the genuine rotation extras, e.g. `maxit`/`gam`) unchanged. One `type` governs
# both presets, exactly as the single flat `type` always did.
.repack_flat_dots <- function(dots, type = NULL) {

  # The flat interface selected the estimator with `method`; the fitting functions take
  # `estimator`. Translate the name first -- unlike the knob repacking below it must happen
  # even when a control object is passed alongside, because `method` never lives in a control.
  # Both names at once is a half-migrated call: silently letting either win would fit with
  # an estimator the caller did not (knowingly) choose, so it is rejected instead.
  if ("method" %in% names(dots)) {
    if ("estimator" %in% names(dots)) {
      cli::cli_abort(
        c("{.arg method} and {.arg estimator} cannot both be supplied.",
          "i" = "{.arg method} is the former name of {.arg estimator}; pass only one."),
        class = "efa_renamed_arg"
      )
    }
    dots$estimator <- dots$method
    dots$method <- NULL
  }

  # A control object passed explicitly is the current interface; never second-guess it.
  if (any(c("estimate_control", "rotate_control") %in% names(dots))) return(dots)

  # the former argument spellings reach the controls under their current names
  if ("P_type" %in% names(dots)) {
    dots$p_type <- dots$P_type
    dots$P_type <- NULL
  }
  if ("randomStarts" %in% names(dots)) {
    dots$random_starts <- dots$randomStarts
    dots$randomStarts <- NULL
  }

  if (is.null(type) && "type" %in% names(dots)) type <- dots$type
  dots$type <- NULL

  est_named <- intersect(names(dots), .flat_estimate_knobs)
  rot_named <- intersect(names(dots), .flat_rotate_knobs)

  # nothing to translate: hand the dots on exactly as they came in
  if (is.null(type) && length(est_named) == 0L && length(rot_named) == 0L) return(dots)

  # the preset the flat interface applied when a call tuned a knob without naming a `type`
  if (is.null(type)) type <- "EFAtools"

  # keep the leftovers by logical indexing: indexing by name would corrupt an unnamed dot
  # (its "" matches nothing, so it would come back as a NULL named NA)
  nms <- names(dots)
  if (is.null(nms)) nms <- character(length(dots))
  c(list(estimate_control = do.call(estimate_control, c(list(type = type), dots[est_named])),
         rotate_control = do.call(rotate_control, c(list(type = type), dots[rot_named]))),
    dots[!(nms %in% c(est_named, rot_named))])
}

# Render one "<label>: <value>" line inside a cli container. An unset (NA)
# preset-driven knob shows as a dim marker; other values are shown verbatim.
.control_knob_line <- function(label, value) {
  shown <- if (.is_control_unset(value)) {
    cli::col_grey("<from type preset>")
  } else {
    paste(format(value), collapse = ", ")
  }
  cli::cli_text("{.field {label}}: {shown}")
}

#' Print and format a control object
#'
#' `print()` shows the chosen `type` and each tuning knob, with an unset (`NA`)
#' preset-driven knob marked as resolved from the `type` preset. `format()`
#' assembles the same report and returns it as a character vector; `print()` is
#' `cat(format(x), sep = "\n")`. The lines follow the active console theme, so they
#' are plain when colours are disabled.
#'
#' @param x A control object from [estimate_control()] or [rotate_control()].
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @seealso [estimate_control()], [rotate_control()]
#'
#' @family Control functions
#'
#' @name print.efa_control
#'
#' @examples
#' est <- estimate_control(type = "SPSS")
#' est
#' writeLines(format(est))
#'
NULL

#' @rdname print.efa_control
#' @export
#' @method print efa_estimate_control
print.efa_estimate_control <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_control
#' @export
#' @method format efa_estimate_control
format.efa_estimate_control <- function(x, ...) {
  cli::cli_format_method({
    cli::cli_text("{.strong Estimation control} ({.field type}: {.val {x$type}})")
    cli::cli_text("")
    .control_knob_line("init_comm", x$init_comm)
    .control_knob_line("criterion", x$criterion)
    .control_knob_line("criterion_type", x$criterion_type)
    .control_knob_line("max_iter", x$max_iter)
    .control_knob_line("abs_eigen", x$abs_eigen)
    .control_knob_line("start_method", x$start_method)
    .control_knob_line("fiml_max_iter", x$fiml_max_iter)
    .control_knob_line("fiml_tol", x$fiml_tol)
  })
}

#' @rdname print.efa_control
#' @export
#' @method print efa_rotate_control
print.efa_rotate_control <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_control
#' @export
#' @method format efa_rotate_control
format.efa_rotate_control <- function(x, ...) {
  cli::cli_format_method({
    cli::cli_text("{.strong Rotation control} ({.field type}: {.val {x$type}})")
    cli::cli_text("")
    .control_knob_line("normalize", x$normalize)
    .control_knob_line("precision", x$precision)
    .control_knob_line("order_type", x$order_type)
    .control_knob_line("varimax_type", x$varimax_type)
    .control_knob_line("p_type", x$p_type)
    .control_knob_line("k", x$k)
    .control_knob_line("random_starts", x$random_starts)

    if (length(x$extra_args) > 0) {
      nms <- names(x$extra_args)
      if (is.null(nms)) nms <- rep("", length(x$extra_args))
      vals <- vapply(x$extra_args,
                     function(v) paste(format(v), collapse = ", "), character(1))
      # `pairs` is interpolated as a variable so any user-supplied value cannot be
      # parsed as cli markup.
      pairs <- paste0(nms, " = ", vals)
      cli::cli_text("{.field extra_args}: {pairs}")
    }
  })
}

## Use with an output from the EFAtools::EFA function, both with type EFAtools
EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
SL_EFAtools <- efa_schmid_leiman(EFA_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF")

# with type SPSS and method ULS. The second-order fit notes that only PAF is validated
# against SPSS; that note is asserted where it belongs (test-superseded.R) and is muffled
# here so the fixture does not raise a warning outside any test.
SL_SPSS <- suppressWarnings(
  efa_schmid_leiman(EFA_mod, estimate_control = estimate_control(type = "SPSS"),
                    estimator = "ULS"))

## Use with an output from the psych::fa function with type psych
fa_mod <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                    fm = "pa", rotate = "Promax", n.rotations = 1)
SL_psych <- efa_schmid_leiman(fa_mod, estimate_control = estimate_control(type = "psych"), estimator = "PAF")

## Use more flexibly by entering a pattern matrix and phi directly, with method
## ML
SL_flex <- efa_schmid_leiman(EFA_mod$rot_loadings, Phi = EFA_mod$Phi, estimator = "ML",
                             estimate_control = estimate_control(type = "EFAtools"))

## Use with a second-order lavaan solution
if (requireNamespace("lavaan", quietly = TRUE)) {
lav_mod_ho <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18
               g =~ F1 + F2 + F3'
lav_fit_ho <- suppressWarnings(lavaan::cfa(lav_mod_ho,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))
SL_lav <- efa_schmid_leiman(lav_fit_ho, g_name = "g")
}

test_that("output class and dimensions are correct", {
  skip_if_not_installed("lavaan")
  expect_identical(class(SL_EFAtools), c("efa_schmid_leiman", "SL"))
  expect_identical(class(SL_EFAtools$sl), c("efa_sl_loadings", "SLLOADINGS"))
  expect_s3_class(SL_SPSS, "efa_schmid_leiman")
  expect_s3_class(SL_psych, "efa_schmid_leiman")
  expect_s3_class(SL_flex, "efa_schmid_leiman")
  expect_s3_class(SL_lav, "efa_schmid_leiman")

  expect_output(str(SL_EFAtools), "List of 7")
  expect_output(str(SL_SPSS), "List of 7")
  expect_output(str(SL_psych), "List of 7")
  expect_output(str(SL_flex), "List of 7")
  expect_output(str(SL_lav), "List of 7")
})

test_that("non-convergence of the second-order EFA is recorded and surfaced", {
  testthat::local_reproducible_output()

  # The convergence code of the second-order EFA is carried on the SL object.
  expect_equal(SL_SPSS$convergence, 0)

  # A non-zero code shows the banner: the optimiser estimators use the generic
  # message, PAF the iteration-specific one. Set the code explicitly so the test
  # does not depend on the second-order fit's actual convergence.
  sl_uls <- SL_SPSS
  sl_uls$convergence <- 1L
  expect_true(any(grepl("The optimiser did not converge", format(sl_uls), fixed = TRUE)))

  sl_paf <- SL_EFAtools
  sl_paf$convergence <- 1L
  expect_true(any(grepl("Maximum number of iterations reached without convergence",
                        format(sl_paf), fixed = TRUE)))

  # A converged fit shows no banner.
  expect_false(any(grepl("did not converge", format(SL_SPSS), fixed = TRUE)))
})

test_that("original correlation is correct", {
  skip_if_not_installed("lavaan")
  expect_equal(SL_EFAtools$orig_R, test_models$baseline$cormat)
  expect_equal(SL_SPSS$orig_R, test_models$baseline$cormat)
  expect_equal(SL_psych$orig_R, test_models$baseline$cormat)
  expect_equal(SL_flex$orig_R, NA)
  expect_equal(SL_lav$orig_R, NA)
})

test_that("sl solution is correct", {
  skip_if_not_installed("lavaan")
  expect_equal(unname(SL_EFAtools$sl[, "h2"]) + unname(SL_EFAtools$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_SPSS$sl[, "h2"]) + unname(SL_SPSS$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_psych$sl[, "h2"]) + unname(SL_psych$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_flex$sl[, "h2"]) + unname(SL_flex$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_lav$sl[, "h2"]) + unname(SL_lav$sl[, "u2"]),
               rep(1, 18))

  expect_equal(unname(SL_EFAtools$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_SPSS$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_psych$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_flex$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_lav$sl[, "g"]) >= .20, rep(TRUE, 18))

  expect_equal(unname(SL_EFAtools$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_SPSS$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_psych$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_flex$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_lav$sl[1:6, "F1"]) >= .20, rep(TRUE, 6))

  expect_equal(unname(SL_EFAtools$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_SPSS$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_psych$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_flex$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_lav$sl[7:18, "F1"]) < .20, rep(TRUE, 12))

  expect_equal(unname(SL_EFAtools$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_SPSS$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_psych$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_flex$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_lav$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))

  expect_equal(unname(SL_EFAtools$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_SPSS$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_psych$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_flex$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_lav$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))

  expect_equal(unname(SL_EFAtools$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_SPSS$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_psych$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_flex$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_lav$sl[13:18, "F3"]) >= .20, rep(TRUE, 6))

  expect_equal(unname(SL_EFAtools$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_SPSS$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_psych$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_flex$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_lav$sl[1:12, "F3"]) < .20, rep(TRUE, 12))
})

test_that("sl loadings reproduce psych::schmid", {
  skip_on_cran()
  skip_if_not_installed("psych")

  # The Schmid-Leiman transformation of the same psych Promax solution must match
  # psych::schmid()'s g and group loadings (its columns are g, F1*, F2*, F3*).
  sch <- suppressMessages(suppressWarnings(
    psych::schmid(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                  fm = "pa", rotate = "Promax")))

  expect_equal(unname(SL_psych$sl[, c("g", "F1", "F2", "F3")]),
               unname(sch$sl[, c("g", "F1*", "F2*", "F3*")]),
               tolerance = 1e-4)
})

test_that("settings are returned correctly", {
  skip_if_not_installed("lavaan")
  expect_named(SL_EFAtools$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen"))
  expect_named(SL_SPSS$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed"))
  expect_named(SL_psych$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                       "N", "use", "cor_method", "input_type",
                                       "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                       "init_comm", "criterion", "criterion_type",
                                       "abs_eigen"))
  expect_named(SL_flex$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                  "N", "use", "cor_method", "input_type",
                                  "cor_method_used", "se", "b_boot", "ci", "seed", "start_method"))
  expect_equal(SL_lav$settings, NA)

  expect_equal(SL_EFAtools$settings$estimator, "PAF")
  expect_equal(SL_SPSS$settings$estimator, "ULS")
  expect_equal(SL_psych$settings$estimator, "PAF")
  expect_equal(SL_flex$settings$estimator, "ML")

  expect_equal(SL_EFAtools$settings$rotation, "none")
  expect_equal(SL_SPSS$settings$rotation, "none")
  expect_equal(SL_psych$settings$rotation, "none")
  expect_equal(SL_flex$settings$rotation, "none")

  expect_equal(SL_EFAtools$settings$type, "EFAtools")
  expect_equal(SL_SPSS$settings$type, "SPSS")
  expect_equal(SL_psych$settings$type, "psych")
  expect_equal(SL_flex$settings$type, "EFAtools")

  expect_equal(SL_EFAtools$settings$n_factors, 1)
  expect_equal(SL_SPSS$settings$n_factors, 1)
  expect_equal(SL_psych$settings$n_factors, 1)
  expect_equal(SL_flex$settings$n_factors, 1)

  expect_equal(SL_EFAtools$settings$N, 100)
  expect_equal(SL_SPSS$settings$N, 100)
  expect_equal(SL_psych$settings$N, 100)
  expect_equal(SL_flex$settings$N, 100)

  expect_equal(SL_EFAtools$settings$use, "pairwise.complete.obs")
  expect_equal(SL_SPSS$settings$use, "pairwise.complete.obs")
  expect_equal(SL_psych$settings$use, "pairwise.complete.obs")
  expect_equal(SL_flex$settings$use, "pairwise.complete.obs")

  expect_equal(SL_EFAtools$settings$cor_method, "pearson")
  expect_equal(SL_SPSS$settings$cor_method, "pearson")
  expect_equal(SL_psych$settings$cor_method, "pearson")
  expect_equal(SL_flex$settings$cor_method, "pearson")

  expect_equal(SL_EFAtools$settings$max_iter, 300)
  expect_equal(SL_psych$settings$max_iter, 50)

  expect_equal(SL_EFAtools$settings$init_comm, "smc")
  expect_equal(SL_psych$settings$init_comm, "smc")

  expect_equal(SL_EFAtools$settings$criterion, 0.001)
  expect_equal(SL_psych$settings$criterion,  0.001)

  expect_equal(SL_EFAtools$settings$criterion_type, "sum")
  expect_equal(SL_psych$settings$criterion_type, "sum")

  expect_equal(SL_EFAtools$settings$abs_eigen, TRUE)
  expect_equal(SL_psych$settings$abs_eigen, FALSE)

  expect_equal(SL_flex$settings$start_method, "psych")
})


EFA_mod_unrot <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                     type = "EFAtools", method = "PAF", rotation = "none")
EFA_mod_orth <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                     type = "EFAtools", method = "PAF", rotation = "varimax")
fa_mod_unrot <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                          fm = "pa", rotate = "none")
fa_mod_orth <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                         fm = "pa", rotate = "Varimax", n.rotations = 1)

if (requireNamespace("lavaan", quietly = TRUE)) {
lav_mod_NA <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6 + V17
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12 + V2
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18 + V10
               g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                    V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_NA <- suppressWarnings(lavaan::cfa(lav_mod_NA,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

lav_mod_ho_inv <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
                   F2 =~ V7 + V8 + V9 + V10 + V11 + V12
                   g =~ F1 + F2 + V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_ho_inv <- suppressWarnings(lavaan::cfa(lav_mod_ho_inv,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

# Improper second-order solution: a first-order factor residual variance turns
# negative while all loadings stay below 1 (a variance Heywood case).
lav_mod_var_hey <- 'F1 =~ V1 + V2
                    F2 =~ V3 + V4
                    F3 =~ V5 + V6
                    g =~ F1 + F2 + F3'
lav_fit_var_hey <- suppressWarnings(lavaan::cfa(lav_mod_var_hey,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

# Bifactor solution (general factor loads on the items, not the first-order
# factors): SL needs a second-order model, so this must abort cleanly.
lav_mod_bifactor <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
                     F2 =~ V7 + V8 + V9 + V10 + V11 + V12
                     F3 =~ V13 + V14 + V15 + V16 + V17 + V18
                     g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 +
                          V12 + V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_bifactor <- suppressWarnings(lavaan::cfa(lav_mod_bifactor,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml",
                                           orthogonal = TRUE))
}

test_that("errors are thrown correctly", {
  skip_if_not_installed("lavaan")
  expect_error(efa_schmid_leiman(1:5), class = "efa_sl_bad_input")
  expect_warning(efa_schmid_leiman(EFA_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF", Phi = EFA_mod$Phi),
                 class = "efa_sl_phi_specified")
  expect_error(efa_schmid_leiman(EFA_mod_unrot, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF"), class = "efa_sl_not_oblique")
  expect_error(efa_schmid_leiman(EFA_mod_orth, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF"), class = "efa_sl_not_oblique")
  expect_warning(efa_schmid_leiman(fa_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF", Phi = fa_mod$Phi), class = "efa_sl_phi_specified")
  expect_error(efa_schmid_leiman(fa_mod_unrot, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF"), class = "efa_sl_not_oblique")
  expect_error(efa_schmid_leiman(fa_mod_orth, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF"), class = "efa_sl_not_oblique")
  expect_error(efa_schmid_leiman(lav_fit_NA, g_name = "g"), class = "efa_sl_na_loadings")
  expect_error(efa_schmid_leiman(lav_fit_var_hey, g_name = "g"), class = "efa_sl_heywood")
  expect_error(efa_schmid_leiman(lav_fit_ho, g_name = "fu"), class = "efa_sl_g_name")
  expect_warning(efa_schmid_leiman(lav_fit_ho_inv, g_name = "g"), class = "efa_sl_second_order_loadings")
  expect_error(efa_schmid_leiman(EFA_mod$rot_loadings, estimate_control = estimate_control(type = "EFAtools"), estimator = "ML"), class = "efa_sl_phi_missing")
  expect_error(efa_schmid_leiman(lav_fit_bifactor, g_name = "g"), class = "efa_sl_not_second_order")
})

test_that("the second-order fit's substantive warnings reach the user", {
  # A factor space with an analytic Heywood case: for one factor on three variables
  # lambda1^2 = r12 * r13 / r23, so these intercorrelations imply a communality of 1.44.
  # PAF and ULS abort on it; ML clamps the communality at the boundary and warns, and
  # that warning must not be swallowed -- the residualized loadings it leads to are
  # essentially zero and everything computed from them rests on it.
  Phi_hey <- matrix(c(1, .9, .8, .9, 1, .5, .8, .5, 1), 3, 3)
  L1 <- matrix(0, 9, 3)
  L1[1:3, 1] <- L1[4:6, 2] <- L1[7:9, 3] <- .7
  colnames(L1) <- c("F1", "F2", "F3")

  expect_warning(efa_schmid_leiman(L1, Phi = Phi_hey, estimator = "ML"),
                 class = "efa_heywood")
  # PAF aborts on the same input, after the fit has named the improper first-order
  # factor -- the abort says that there is a Heywood case, the warning says where.
  expect_warning(
    expect_error(efa_schmid_leiman(L1, Phi = Phi_hey, estimator = "PAF"),
                 class = "efa_sl_heywood"),
    class = "efa_heywood")

  # The identification warnings of the internal fit are structural -- one factor on a
  # k by k matrix of factor intercorrelations is just identified at k = 3 for every
  # input -- so a sound three-factor solution stays silent.
  expect_no_warning(efa_schmid_leiman(EFA_mod, estimator = "PAF"))
  expect_no_warning(efa_schmid_leiman(EFA_mod, estimator = "ML"))
})

test_that("too few first-order factors are reported on the user's solution", {
  # One first-order factor leaves nothing to orthogonalize. The diagnostic must name the
  # supplied solution rather than the internal one-factor fit on the 1 x 1 matrix of
  # factor intercorrelations, which would otherwise abort about a single "variable".
  expect_error(
    efa_schmid_leiman(matrix(c(.7, .6, .5, .4), 4, 1), Phi = matrix(1, 1, 1),
                      estimator = "PAF"),
    class = "efa_sl_too_few_factors")

  # Two first-order factors do not identify the second-order fit either, which is a
  # warning and is raised before that fit runs.
  L2 <- matrix(0, 4, 2)
  L2[1:2, 1] <- 0.7
  L2[3:4, 2] <- 0.7
  colnames(L2) <- c("F1", "F2")
  expect_warning(
    efa_schmid_leiman(L2, Phi = matrix(c(1, .5, .5, 1), 2), estimator = "PAF"),
    class = "efa_sl_underidentified")
})

test_that("efa_fit()'s inference arguments are refused", {
  # They are efa_fit() formals, so they used to ride through the dots into the second-order
  # fit -- which runs against a placeholder N and reports none of them, so a standard error
  # was computed from a sample size that is not the data's and then thrown away, and `seed`
  # governed a fit that draws no random numbers. Only `settings` ever recorded them.
  args <- list(EFA_mod, estimate_control = estimate_control(type = "EFAtools"),
               estimator = "ML")
  expect_error(do.call(efa_schmid_leiman, c(args, list(se = "information"))),
               class = "efa_unused_dots")
  expect_error(do.call(efa_schmid_leiman, c(args, list(seed = 42))),
               class = "efa_unused_dots")
  expect_error(do.call(efa_schmid_leiman, c(args, list(b_boot = 10, ci = .90))),
               class = "efa_unused_dots")

  # the frozen wrapper forwards its dots into the same fit, so the refusal reaches it too
  expect_error(SL(EFA_mod, type = "EFAtools", method = "ML", se = "information"),
               class = "efa_unused_dots")

  # An abbreviation must be refused as well: the dots are spliced into the second-order fit
  # with do.call(), where R partial-matches them against efa_fit()'s formals, so `b_b` would
  # arrive there as `b_boot` without ever matching the refused name exactly. Only the
  # whitelist, which spells every accepted name in full, closes that route.
  expect_error(do.call(efa_schmid_leiman, c(args, list(b_b = 10))),
               class = "efa_unused_dots")
  expect_error(do.call(efa_schmid_leiman, c(args, list(see = 42))),
               class = "efa_unused_dots")
  expect_error(SL(EFA_mod, type = "EFAtools", method = "ML", b_b = 10),
               class = "efa_unused_dots")

  # a genuine efa_fit() argument still travels through the dots untouched -- `use` is an
  # efa_fit() formal and not one of efa_schmid_leiman()'s, so it really does go through `...`
  expect_no_error(suppressWarnings(suppressMessages(
    do.call(efa_schmid_leiman, c(args, list(use = "complete.obs"))))))
})

test_that("a second-order Heywood case raises a classed error", {
  # A pattern matrix with a highly (and unevenly) intercorrelated factor space
  # makes the single second-order factor improper: a second-order communality
  # exceeds 1, so the residualized first-order loadings would be undefined.
  L1 <- matrix(0, 6, 3)
  L1[1:2, 1] <- 0.7
  L1[3:4, 2] <- 0.7
  L1[5:6, 3] <- 0.7
  colnames(L1) <- c("F1", "F2", "F3")
  Phi_heywood <- matrix(c(1, 0.8, 0.95,
                          0.8, 1, 0.7,
                          0.95, 0.7, 1), nrow = 3)

  # The second-order fit names the improper first-order factor before the abort.
  expect_warning(
    expect_error(efa_schmid_leiman(L1, Phi = Phi_heywood, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF"),
                 class = "efa_sl_heywood"),
    class = "efa_heywood")
})

test_that("oblique solutions with unlabelled factor columns keep all factors", {

  # A loading matrix whose columns carry no labels leaves the first-order factors to be
  # taken in the order they come in. Reordering them by a factor number parsed from
  # absent labels used to subset the loadings and Phi down to zero columns, which made
  # the second-order EFA see no variables. efa_fit() labels every rotation's columns, so
  # strip the labels to reach the path an externally built or bare-matrix input takes.
  cormat <- test_models$baseline$cormat
  cormat_list <- list(cormat, cormat, cormat)

  for (k in c(3L, 4L)) {

    fitted <- suppressMessages(efa_fit(cormat, N = 500, n_factors = k,
                                       estimator = "PAF", rotation = "oblimin"))
    pooled <- suppressMessages(efa_mi(cormat_list, N = 500, n_factors = k,
                                      estimator = "PAF", rotation = "oblimin"))

    # the fixtures only test what they are meant to if the columns are unlabelled
    colnames(fitted$rot_loadings) <- NULL
    colnames(pooled$rot_loadings) <- NULL
    dimnames(fitted$Phi) <- NULL
    dimnames(pooled$Phi) <- NULL

    for (obj in list(fitted, pooled)) {

      sl <- efa_schmid_leiman(obj, estimator = "PAF")

      expect_s3_class(sl, "efa_schmid_leiman")
      expect_identical(colnames(sl$sl),
                       c("g", paste0("F", seq_len(k)), "h2", "u2"))
      expect_identical(nrow(sl$sl), ncol(cormat))
      expect_false(anyNA(sl$sl))

      # the second-order EFA saw all k first-order factors: this is what the
      # zero-column subset destroyed, so it is the assertion that pins the bug
      expect_identical(dim(sl$L2), c(k, 1L))

      # loadings are real rather than the all-zero block a collapsed L1 produces
      expect_true(all(abs(sl$sl[, "g"]) > 0))
      expect_true(all(sl$sl[, "h2"] > 0 & sl$sl[, "h2"] < 1))

    }

  }

})

test_that("labelled factor columns are still ordered by factor number", {

  # The fallback for unlabelled columns must not disable the reordering when the
  # labels are present but out of order.
  L1 <- EFA_mod$rot_loadings[, c(2, 3, 1)]
  Phi_1 <- EFA_mod$Phi[c(2, 3, 1), c(2, 3, 1)]
  colnames(L1) <- c("F2", "F3", "F1")

  sl_scrambled <- efa_schmid_leiman(L1, Phi = Phi_1, estimator = "PAF",
                                    estimate_control = estimate_control(type = "EFAtools"))

  expect_equal(sl_scrambled$sl, SL_EFAtools$sl)

})

test_that("a named Phi is aligned to the loading factors before reordering", {
  L1 <- unclass(EFA_mod$rot_loadings)
  Phi <- EFA_mod$Phi
  base <- efa_schmid_leiman(
    L1, Phi = Phi, estimator = "PAF",
    estimate_control = estimate_control(type = "EFAtools")
  )

  perm <- c(3L, 1L, 2L)
  aligned <- efa_schmid_leiman(
    L1, Phi = Phi[perm, perm], estimator = "PAF",
    estimate_control = estimate_control(type = "EFAtools")
  )

  expect_equal(aligned$sl, base$sl)
  expect_equal(aligned$L2, base$L2)
})

test_that("efa_schmid_leiman runs on an oblique fit whose column labels are permuted", {
  skip_if_not_installed("GPArotation")

  # The first-order columns are reordered by the number in their labels, which the
  # oblimin rotation returns permuted (F2, F1, F3). An unguarded reorder empties Phi,
  # and the second-order fit then aborts on a model with no variables.
  EFA_obl <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                     estimator = "PAF", rotation = "oblimin")
  SL_obl <- efa_schmid_leiman(EFA_obl, estimator = "PAF",
                              estimate_control = estimate_control(type = "EFAtools"))

  expect_s3_class(SL_obl, "efa_schmid_leiman")
  expect_identical(class(SL_obl$sl), c("efa_sl_loadings", "SLLOADINGS"))
  expect_identical(dim(unclass(SL_obl$sl)), dim(unclass(SL_EFAtools$sl)))
  expect_true(all(is.finite(SL_obl$L2)))
  expect_length(SL_obl$L2, 3)
})

test_that("print output is stable", {
  local_reproducible_output()

  expect_snapshot(print(SL_EFAtools), transform = scrub_num)
})

test_that("the second-order header follows the console width", {
  # Emitted verbatim so the "estimator = 'value'" token is never split, and packed to the
  # console around it -- the same treatment the efa_fit() header gets.
  header <- function(w) {
    out <- withr::with_options(list(cli.width = w, cli.num_colors = 1), format(SL_EFAtools))
    wrapped_item(out, "^EFA for second-order")
  }

  for (w in c(120L, 80L, 60L)) {
    lines <- header(w)
    expect_true(all(cli::ansi_nchar(lines, type = "width") <= w))
    expect_identical(paste(trimws(lines), collapse = " "),
                     "EFA for second-order loadings performed with estimator = 'PAF'")
  }
  expect_length(header(60L), 2L)
})

rm(EFA_mod, SL_EFAtools, SL_SPSS, fa_mod, SL_psych, SL_flex, EFA_mod_unrot,
   EFA_mod_orth, fa_mod_unrot, fa_mod_orth)
if (requireNamespace("lavaan", quietly = TRUE)) {
  rm(lav_mod_ho, lav_fit_ho, SL_lav, lav_mod_NA, lav_fit_NA, lav_mod_ho_inv,
     lav_fit_ho_inv, lav_mod_var_hey, lav_fit_var_hey, lav_mod_bifactor,
     lav_fit_bifactor)
}


test_that("the input error names the averaged matrix an efa_average object carries", {
  # An efa_average object is rejected while "a matrix" is listed as acceptable, and its
  # averaged loadings are exactly that. The message now says where they are, together with
  # the averaged factor correlations, which this transformation always needs.
  avg <- suppressWarnings(suppressMessages(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = c("PAF", "ULS"), rotation = "promax", type = "EFAtools",
                start_method = "psych", show_progress = FALSE)))

  e <- tryCatch(efa_schmid_leiman(avg), efa_sl_bad_input = function(e) e)
  expect_s3_class(e, "efa_sl_bad_input")
  expect_snapshot(cat(conditionMessage(e)))

  # The components the message names are accepted.
  expect_s3_class(
    suppressWarnings(suppressMessages(
      efa_schmid_leiman(avg$loadings$average, Phi = avg$Phi$average))),
    "efa_schmid_leiman")
})

# Unit tests for the estimate_control() / rotate_control() control constructors.

test_that("constructors return classed lists with the documented fields", {
  ec <- estimate_control()
  expect_s3_class(ec, "efa_estimate_control")
  expect_equal(ec$type, "EFAtools")
  expect_true(is.na(ec$init_comm))
  expect_true(is.na(ec$criterion))
  expect_true(is.na(ec$abs_eigen))
  expect_equal(ec$start_method, "psych")
  # The two FIML knobs are not preset-driven, so they carry the EM's own defaults rather
  # than the NA "resolve from the type preset" sentinel.
  expect_equal(ec$fiml_max_iter, 500)
  expect_equal(ec$fiml_tol, 1e-5)
  expect_named(ec, c("type", "init_comm", "criterion", "criterion_type",
                     "max_iter", "abs_eigen", "start_method",
                     "fiml_max_iter", "fiml_tol"))

  rc <- rotate_control()
  expect_s3_class(rc, "efa_rotate_control")
  expect_equal(rc$type, "EFAtools")
  expect_true(rc$normalize)
  expect_equal(rc$precision, 1e-5)
  expect_true(is.na(rc$p_type))
  expect_equal(rc$random_starts, 100)
  expect_type(rc$extra_args, "list")
  expect_length(rc$extra_args, 0)
  expect_named(rc, c("type", "normalize", "precision", "order_type",
                     "varimax_type", "p_type", "k", "random_starts", "extra_args"))
})

test_that("rotate_control captures criterion-specific extras in extra_args", {
  rc <- rotate_control(type = "EFAtools", k = 3, gam = 0.5, delta = 0.02)
  expect_equal(rc$k, 3)
  expect_equal(rc$extra_args$gam, 0.5)
  expect_equal(rc$extra_args$delta, 0.02)
})

test_that("rotate_control rejects a misplaced tuning knob among its extras", {
  # the fit drops an extra whose name collides with one of its formals, so a knob
  # smuggled in here would be silently ignored -- refuse it at construction instead
  expect_error(rotate_control(max_iter = 2), class = "efa_control_input")
  expect_error(rotate_control(criterion = 1e-4), class = "efa_control_input")
  expect_error(rotate_control(P_type = "norm"), class = "efa_control_input")
  expect_error(rotate_control(randomStarts = 7), class = "efa_control_input")
})

test_that("bad estimate_control inputs abort with a classed condition", {
  expect_error(estimate_control(init_comm = "xyz"), class = "efa_control_input")
  expect_error(estimate_control(criterion = -1), class = "efa_control_input")
  expect_error(estimate_control(criterion = 0), class = "efa_control_input")
  expect_error(estimate_control(criterion_type = "foo"), class = "efa_control_input")
  expect_error(estimate_control(max_iter = 0), class = "efa_control_input")
  expect_error(estimate_control(max_iter = 2.5), class = "efa_control_input")
  expect_error(estimate_control(abs_eigen = "yes"), class = "efa_control_input")
  expect_error(estimate_control(start_method = "bad"), class = "efa_control_input")
  # NaN is not the unset sentinel; it must be rejected, not stored.
  expect_error(estimate_control(criterion = NaN), class = "efa_control_input")
  expect_error(estimate_control(init_comm = NaN), class = "efa_control_input")
  expect_error(estimate_control(abs_eigen = NaN), class = "efa_control_input")
  # a criterion at or above 1 is not a convergence tolerance; the bound matches the one the
  # fit enforces, so a control can never carry a value the fit would go on to reject.
  expect_error(estimate_control(criterion = 5), class = "efa_control_input")
  expect_error(estimate_control(criterion = 1), class = "efa_control_input")
  # the bound is exclusive at 1 only: anything below it is a legitimate tolerance
  expect_s3_class(estimate_control(criterion = 0.999), "efa_estimate_control")
  # an invalid `type` is rejected with the same class as the other knobs
  expect_error(estimate_control(type = "bogus"), class = "efa_control_input")
  expect_error(rotate_control(type = "bogus"), class = "efa_control_input")
  # the FIML EM knobs mirror the bounds the EM itself enforces, and -- not being
  # preset-driven -- do not accept the NA sentinel either
  expect_error(estimate_control(fiml_max_iter = 0), class = "efa_control_input")
  expect_error(estimate_control(fiml_max_iter = 2.5), class = "efa_control_input")
  expect_error(estimate_control(fiml_max_iter = NA), class = "efa_control_input")
  expect_error(estimate_control(fiml_tol = 0), class = "efa_control_input")
  expect_error(estimate_control(fiml_tol = -1e-4), class = "efa_control_input")
  expect_error(estimate_control(fiml_tol = NA), class = "efa_control_input")
  # A tolerance at or above 1 is met on the first EM iteration, which would return the
  # starting moments while reporting convergence; the EM itself only demands tol > 0, so the
  # control is the only place that can refuse it (as it does for `criterion`).
  expect_error(estimate_control(fiml_tol = 1), class = "efa_control_input")
  expect_error(estimate_control(fiml_tol = 5), class = "efa_control_input")
  expect_s3_class(estimate_control(fiml_max_iter = 2000, fiml_tol = 0.999),
                  "efa_estimate_control")
})

test_that("efa_fit rejects unused dots when no rotation runs", {
  # with rotation = "none" no rotation engine runs, so nothing can consume the dots;
  # this also catches a misspelled control argument instead of silently ignoring it
  expect_error(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, rotation = "none",
            estimate_controls = estimate_control(max_iter = 500)),
    class = "efa_unused_dots")
  expect_error(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, gam = 0.5),
    class = "efa_unused_dots")
})

test_that("efa_fit rejects a dot argument the selected rotation does not consume", {
  cm <- test_models$baseline$cormat

  # a misspelled criterion parameter would silently run the engine default (gam = 0)
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, rotation = "oblimin", gamma = 0.5),
    class = "efa_unused_dots")
  # another criterion's parameter is just as unconsumable as a misspelling
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, rotation = "geominT", gam = 0.5),
    class = "efa_unused_dots")
  # varimax and promax run outside the native engines and take no extras at all
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, rotation = "varimax", bogus_arg = 1),
    class = "efa_unused_dots")
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, rotation = "promax", maxit = 100),
    class = "efa_unused_dots")

  # the consumable extras still reach the engine: oblimin's gam changes the criterion.
  # gam = 0.5 rewards correlated factors strongly enough to trip the extreme-Phi check on
  # this fixture, which is beside the point here -- only that the extra reaches the engine.
  fit_default <- efa_fit(cm, n_factors = 3, N = 500, rotation = "oblimin")
  fit_gam <- suppressWarnings(
    efa_fit(cm, n_factors = 3, N = 500, rotation = "oblimin", gam = 0.5))
  expect_false(identical(fit_default$rot_loadings, fit_gam$rot_loadings))
  expect_no_error(efa_fit(cm, n_factors = 3, N = 500, rotation = "geominQ", delta = 0.05))

  # a control-carried extra stays lenient per rotation: one control may serve fits whose
  # rotation never consumes it (efa_average() carries maxit across its whole grid)
  expect_no_error(efa_fit(cm, n_factors = 3, N = 500, rotation = "varimax",
                          rotate_control = rotate_control(maxit = 5e4)))
  expect_no_error(efa_fit(cm, n_factors = 3, N = 500, rotation = "none",
                          rotate_control = rotate_control(maxit = 5e4)))
})

test_that("the fit stays the backstop for a control edited after construction", {
  # The constructor's bounds mirror the fit's, so a control it builds can never carry a value the
  # fit rejects. A control is a plain list, though, so it can be edited afterwards -- the fit
  # keeps its own guard for that, and it must not be mistaken for dead code.
  ec <- estimate_control(criterion = 1e-3)
  ec$criterion <- 1

  expect_error(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            estimate_control = ec),
    class = "efa_criterion_too_large")
})

test_that("efa_fit's argument assertions abort with a classed condition", {
  cm <- test_models$baseline$cormat

  expect_error(efa_fit(cm, n_factors = 3, N = -5), class = "efa_invalid_argument")
  expect_error(efa_fit(cm, n_factors = -1, N = 500), class = "efa_invalid_argument")

  # a value routed through a control reaches the same guard
  ec <- estimate_control()
  ec$max_iter <- -1
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", estimate_control = ec),
    class = "efa_invalid_argument")
})

test_that("estimate_control accepts the start_method spellings the flat interface took", {
  # start_method only governs the ML optimiser, so NA ("not needed here") is admissible; the
  # fit rejects it, and only when the estimator really is ML.
  expect_true(is.na(estimate_control(start_method = NA)$start_method))
  # abbreviations resolve, as checkmate::matchArg() did for the flat interface
  expect_identical(estimate_control(start_method = "fact")$start_method, "factanal")
  expect_identical(estimate_control(start_method = "psy")$start_method, "psych")
  expect_error(estimate_control(start_method = "bad"), class = "efa_control_input")
})

test_that("bad rotate_control inputs abort with a classed condition", {
  # an extra no rotation engine can consume is a misspelling; refuse to store it
  expect_error(rotate_control(gamma = 0.5), class = "efa_control_input")
  expect_error(rotate_control(eps = 1e-6), class = "efa_control_input")
  expect_error(rotate_control(order_type = "x"), class = "efa_control_input")
  expect_error(rotate_control(varimax_type = "x"), class = "efa_control_input")
  expect_error(rotate_control(p_type = "x"), class = "efa_control_input")
  expect_error(rotate_control(k = -1), class = "efa_control_input")
  expect_error(rotate_control(precision = 0), class = "efa_control_input")
  expect_error(rotate_control(random_starts = -1), class = "efa_control_input")
  expect_error(rotate_control(random_starts = 2.5), class = "efa_control_input")
  expect_error(rotate_control(k = NaN), class = "efa_control_input")
  expect_error(rotate_control(p_type = NaN), class = "efa_control_input")
  # a rotation tolerance above 1 is not a tolerance; the bound matches the fit's.
  expect_error(rotate_control(precision = 2), class = "efa_control_input")
  # normalize, precision and random_starts have no NA state.
  expect_error(rotate_control(normalize = NA), class = "efa_control_input")
  expect_error(rotate_control(precision = NA), class = "efa_control_input")
})

test_that("rotate_control accepts random_starts = 0 (warm start only)", {
  # 0 is a meaningful setting, not a missing one: it runs the rotation from its warm start
  # with no random restarts, which the flat interface always allowed.
  expect_equal(rotate_control(random_starts = 0)$random_starts, 0)
})

test_that("NA-defaulted knobs are accepted (they resolve from the preset later)", {
  expect_s3_class(estimate_control(init_comm = NA, criterion = NA,
                                   criterion_type = NA, max_iter = NA,
                                   abs_eigen = NA),
                  "efa_estimate_control")
  expect_s3_class(rotate_control(order_type = NA, varimax_type = NA,
                                 p_type = NA, k = NA),
                  "efa_rotate_control")
})

# The core requirement: a control object built from a given type resolves (through
# the same resolver the fitting cores use) to the same settings EFA() would use.

test_that("estimate_control(type) resolves to the settings EFA() uses for PAF", {
  fields <- c("init_comm", "criterion", "criterion_type", "max_iter", "abs_eigen")
  for (ty in c("EFAtools", "psych", "SPSS")) {
    ec <- estimate_control(type = ty)
    resolved <- .resolve_settings(
      type = ty,
      user = list(init_comm = ec$init_comm, criterion = ec$criterion,
                  criterion_type = ec$criterion_type, max_iter = ec$max_iter,
                  abs_eigen = ec$abs_eigen),
      preset = .efa_presets$PAF
    )
    efa <- suppressWarnings(
      EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
          method = "PAF", type = ty))
    for (nm in fields) expect_equal(resolved[[nm]], efa$settings[[nm]])
  }
})

test_that("rotate_control(type) resolves to the settings EFA() uses per rotation", {
  cormat <- test_models$baseline$cormat

  for (ty in c("EFAtools", "psych", "SPSS")) {
    rc <- rotate_control(type = ty)

    # promax: the richest preset (normalize, P_type, order_type, varimax_type, k).
    resolved_p <- .resolve_settings(
      type = ty,
      user = list(normalize = rc$normalize, P_type = rc$p_type,
                  order_type = rc$order_type, varimax_type = rc$varimax_type,
                  k = rc$k),
      preset = .efa_presets$PROMAX
    )
    efa_p <- suppressWarnings(
      EFA(cormat, n_factors = 3, N = 500, rotation = "promax", type = ty))
    for (nm in c("normalize", "P_type", "order_type", "varimax_type", "k")) {
      expect_equal(resolved_p[[nm]], efa_p$settings[[nm]])
    }

    # varimax: normalize, order_type, varimax_type.
    resolved_v <- .resolve_settings(
      type = ty,
      user = list(normalize = rc$normalize, order_type = rc$order_type,
                  varimax_type = rc$varimax_type),
      preset = .efa_presets$VARIMAX
    )
    efa_v <- suppressWarnings(
      EFA(cormat, n_factors = 3, N = 500, rotation = "varimax", type = ty))
    for (nm in c("normalize", "order_type", "varimax_type")) {
      expect_equal(resolved_v[[nm]], efa_v$settings[[nm]])
    }

    # oblimin: only normalize and order_type are preset-driven.
    resolved_o <- .resolve_settings(
      type = ty,
      user = list(normalize = rc$normalize, order_type = rc$order_type),
      preset = .efa_presets$ROTATE_OBLQ
    )
    efa_o <- suppressWarnings(
      EFA(cormat, n_factors = 3, N = 500, rotation = "oblimin", type = ty))
    for (nm in c("normalize", "order_type")) {
      expect_equal(resolved_o[[nm]], efa_o$settings[[nm]])
    }
  }
})

test_that("a knob pinned alongside a type is carried through and wins over the preset", {
  ec <- estimate_control(type = "SPSS", max_iter = 10)
  expect_equal(ec$max_iter, 10)

  resolved <- suppressWarnings(.resolve_settings(
    type = "SPSS",
    user = list(init_comm = ec$init_comm, criterion = ec$criterion,
                criterion_type = ec$criterion_type, max_iter = ec$max_iter,
                abs_eigen = ec$abs_eigen),
    preset = .efa_presets$PAF
  ))
  expect_equal(resolved$max_iter, 10)
  # the SPSS preset default (25) is overridden by the pin.
  expect_equal(resolved$criterion_type, "max_individual")
})

test_that("control objects print", {
  testthat::local_reproducible_output()

  expect_snapshot(print(estimate_control(type = "SPSS")))
  expect_snapshot(print(estimate_control(type = "none", init_comm = "smc",
                                         criterion = 1e-3, criterion_type = "sum",
                                         max_iter = 300, abs_eigen = TRUE)))
  expect_snapshot(print(rotate_control(type = "psych")))
  expect_snapshot(print(rotate_control(type = "none", normalize = FALSE,
                                       order_type = "eigen", varimax_type = "kaiser",
                                       p_type = "norm", k = 4, random_starts = 50,
                                       gam = 0.5)))
})

# Regression net against the bundled SPSS FACTOR reference output (`SPSS_23`, `SPSS_27`).
# Reproducing SPSS's principal axis factoring, varimax, and promax is what the `"SPSS"`
# presets exist for, and these reference objects are the only oracles in the suite that are
# SPSS output rather than another R implementation, so the presets are otherwise pinned only
# against their own literals.
#
# Two things are checked, because they can fail independently:
#
#  (a) the ROTATIONS on their own, by rotating SPSS's own `paf_load` under `type = "SPSS"`.
#      The extraction is then out of the comparison, so any deviation belongs to the Kaiser
#      varimax, the "norm" promax target, or the ss_factors ordering.
#  (b) the FULL preset path, by extracting from the correlation matrix with
#      `estimate_control(type = "SPSS")` and rotating with `rotate_control(type = "SPSS")`,
#      which adds the PAF preset itself to the comparison.
#
# Columns and signs are matched with the package's own congruence alignment before
# comparing, and agreement is reported as the largest absolute deviation of any single
# coefficient, which is the quantity the tolerances below were set from; a relative
# difference averaged over the matrix would hide a single badly placed loading.

spss_refs <- list(SPSS_23 = SPSS_23, SPSS_27 = SPSS_27)

# The simulated cases whose PAF iteration needs more than SPSS's own default of 25
# iterations to reach the stored reference (see `?SPSS_23`).
spss_slow_cases <- c("case_1a", "case_11b")

max_abs_diff <- function(x, y) max(abs(unclass(x) - unclass(y)))

# largest absolute deviation after matching columns and signs to the reference
aligned_diff <- function(reference, loadings) {
  max_abs_diff(.align_solution(L_target = reference, L = unclass(loadings))$loadings,
               reference)
}

test_that("the SPSS rotation presets reproduce the bundled SPSS FACTOR rotations", {
  skip_on_cran()

  # Tolerance: the largest deviation observed over all 17 reference solutions is 1.2e-2, on
  # the 29-variable seven-factor WJIV_3_5 promax, and it is dominated by convergence slack
  # rather than by an algorithmic difference -- tightening `precision` to 1e-8 takes it to
  # 1.1e-3. Three of the solutions reproduce to machine precision. 2e-2 therefore sits above
  # the worst honest residual and far below any genuine regression, which moves loadings in
  # the second decimal.
  for (ver in names(spss_refs)) {
    for (nm in names(spss_refs[[ver]])) {

      ent <- spss_refs[[ver]][[nm]]
      lab <- paste(ver, nm)
      unrot <- list(unrot_loadings = ent$paf_load)

      vm <- .rotate_model(unrot, rotation = "varimax", type = "SPSS")
      expect_lt(aligned_diff(ent$var_load, vm$rot_loadings), 2e-2,
                label = paste("varimax deviation,", lab))

      pm <- .rotate_model(unrot, rotation = "promax", type = "SPSS")
      al <- .align_solution(L_target = ent$pro_load, L = unclass(pm$rot_loadings),
                            Phi = unclass(pm$Phi))
      expect_lt(max_abs_diff(al$loadings, ent$pro_load), 2e-2,
                label = paste("promax deviation,", lab))
      expect_lt(max_abs_diff(al$Phi, ent$pro_phi), 2e-2,
                label = paste("promax Phi deviation,", lab))
    }
  }
})

test_that("the SPSS presets reproduce SPSS FACTOR from the correlation matrix", {
  skip_on_cran()

  # The four simulated cases are the entries whose correlation matrices ship as well
  # (`test_models`), so they are the only ones where the extraction can be checked rather
  # than taken from the reference. The stored solutions were produced with the iteration
  # limit raised above SPSS's default of 25 (see `?SPSS_23`), so the extraction is run the
  # same way; at that budget the loadings agree to about 1e-14 and the pin is set at 1e-8,
  # six orders below the 2e-2 to 3e-2 disagreement the capped iteration produces.
  #
  # The fits depend only on the case, so each is run once and then compared against both
  # stored reference versions.
  for (cs in names(test_models)) {

    tm <- test_models[[cs]]
    fit <- function(rotation, max_iter) {
      suppressWarnings(efa_fit(
        tm$cormat, n_factors = tm$n_factors, N = tm$N, estimator = "PAF",
        rotation = rotation,
        estimate_control = estimate_control(type = "SPSS", max_iter = max_iter),
        rotate_control = rotate_control(type = "SPSS")))
    }

    unrot <- fit("none", 500)
    vm <- fit("varimax", 500)
    pm <- fit("promax", 500)
    # Only the slow cases need the capped run: where the extraction converges within
    # SPSS's own limit, `unrot$iter <= 25` already says the cap never bit and the capped
    # fit would be the identical solution.
    slow <- cs %in% spss_slow_cases
    capped <- if (slow) fit("none", 25) else NULL

    for (ver in names(spss_refs)) {

      ent <- spss_refs[[ver]][[cs]]
      lab <- paste(ver, cs)

      expect_lt(aligned_diff(ent$paf_load, unrot$unrot_loadings), 1e-8,
                label = paste("PAF deviation,", lab))

      # Rotating the preset's own extraction rather than SPSS's costs nothing beyond the
      # deviation already measured in the rotation-only block above, so the same class of
      # tolerance applies; the worst case here is 3.4e-4.
      expect_lt(aligned_diff(ent$var_load, vm$rot_loadings), 1e-3,
                label = paste("varimax deviation,", lab))

      al <- .align_solution(L_target = ent$pro_load, L = unclass(pm$rot_loadings),
                            Phi = unclass(pm$Phi))
      expect_lt(max_abs_diff(al$loadings, ent$pro_load), 1e-3,
                label = paste("promax deviation,", lab))
      expect_lt(max_abs_diff(al$Phi, ent$pro_phi), 1e-3,
                label = paste("promax Phi deviation,", lab))

      # At SPSS's own iteration limit the two slow cases miss the stored reference by two
      # orders of magnitude, which is why the documentation records that the references
      # need a raised limit.
      if (slow) {
        expect_gt(aligned_diff(ent$paf_load, capped$unrot_loadings), 1e-3,
                  label = paste("capped PAF deviation,", lab))
      }
    }

    # Convergence is a property of the fit, not of the reference version it is compared
    # against, so it is asserted once per case.
    expect_equal(unrot$convergence, 0, label = paste("PAF convergence,", cs))
    if (slow) {
      expect_equal(capped$convergence, 1, label = paste("capped convergence,", cs))
      expect_gt(unrot$iter, 25, label = paste("iterations used,", cs))
    } else {
      expect_lte(unrot$iter, 25, label = paste("iterations used,", cs))
    }
  }
})

rm(spss_refs, spss_slow_cases, max_abs_diff, aligned_diff)

# Case-insensitive choice matching for the efa_* interface: .match_arg_ci() resolves
# match.arg()-style choice arguments accepting any capitalization and unambiguous
# abbreviations, .map_subset_ci() maps vector-valued subset arguments. Both always
# return the canonical spelling from the choices, so stored settings, registry keys,
# and printed output are unaffected by how a value was capitalized.

test_that(".match_arg_ci returns the default when the argument is left at the formals", {
  f <- function(estimator = c("PAF", "ML", "ULS")) .match_arg_ci(estimator)
  expect_identical(f(), "PAF")

  # with several.ok, the full default resolves to all choices, as in match.arg()
  g <- function(gof = c("CAF", "CFI", "RMSEA")) .match_arg_ci(gof, several.ok = TRUE)
  expect_identical(g(), c("CAF", "CFI", "RMSEA"))
})

test_that(".match_arg_ci matches case-insensitively and returns the canonical spelling", {
  expect_identical(.match_arg_ci("ml", c("PAF", "ML", "ULS")), "ML")
  expect_identical(.match_arg_ci("Paf", c("PAF", "ML", "ULS")), "PAF")
  expect_identical(.match_arg_ci("GEOMINT", c("geominT", "geominQ")), "geominT")
  # exact spellings keep working unchanged
  expect_identical(.match_arg_ci("ULS", c("PAF", "ML", "ULS")), "ULS")
})

test_that(".match_arg_ci resolves abbreviations, with an exact match beating a prefix", {
  expect_identical(.match_arg_ci("pa", c("PAF", "ML", "ULS")), "PAF")
  expect_identical(.match_arg_ci("minr", c("PAF", "ML", "ULS", "MINRES")), "MINRES")
  # "efa" is both a full choice and a prefix of another one: the exact match wins
  expect_identical(.match_arg_ci("efa", c("EFAtools", "EFA")), "EFA")
})

test_that(".match_arg_ci with several.ok matches every element", {
  crits <- c("CD", "EKC", "HULL", "KGC", "PARALLEL", "SCREE", "SMT", "NEST", "MAP")
  expect_identical(.match_arg_ci(c("hull", "parallel"), crits, several.ok = TRUE),
                   c("HULL", "PARALLEL"))
})

test_that(".match_arg_ci rejects ambiguous and unmatched values with a classed error", {
  expect_error(.match_arg_ci("geomin", c("geominT", "geominQ")),
               class = "efa_bad_choice")
  expect_error(.match_arg_ci("bogus", c("PAF", "ML", "ULS")),
               class = "efa_bad_choice")
  # a bad element is an error even when the others match (stricter than match.arg(),
  # which silently drops it)
  expect_error(.match_arg_ci(c("HULL", "bogus"), c("HULL", "EKC"), several.ok = TRUE),
               class = "efa_bad_choice")
  # several values without several.ok, an empty vector, and a non-character input
  expect_error(.match_arg_ci(c("PAF", "ML"), c("PAF", "ML", "ULS")),
               class = "efa_bad_choice")
  expect_error(.match_arg_ci(character(), c("PAF", "ML"), several.ok = TRUE),
               class = "efa_bad_choice")
  expect_error(.match_arg_ci(1L, c("PAF", "ML")), class = "efa_bad_choice")
})

test_that(".match_arg_ci case-folds locale-independently and lets the caller set the class", {
  # ASCII folding, not toupper(): a value with an "i" must resolve regardless of the
  # locale's upper-casing rules (toupper("i") is the dotted capital under a Turkish locale).
  expect_identical(.match_arg_ci("minres", c("PAF", "ML", "ULS", "MINRES")), "MINRES")
  expect_identical(.match_arg_ci("oblimin", c("oblimin", "quartimin")), "oblimin")
  # the caller can raise its own condition class (the control constructors do)
  expect_error(.match_arg_ci("bogus", c("PAF", "ML"), class = "efa_control_input"),
               class = "efa_control_input")
})

test_that(".map_subset_ci maps onto the canonical spellings and leaves non-matches alone", {
  expect_identical(.map_subset_ci(c("paf", "MINRES"), c("PAF", "ML", "ULS", "MINRES")),
                   c("PAF", "MINRES"))
  # exact case-folded matches only: no abbreviations, and an unknown value passes
  # through unchanged for the caller's subset assertion to reject
  expect_identical(.map_subset_ci(c("pa", "bogus"), c("PAF", "ML")), c("pa", "bogus"))
  expect_identical(.map_subset_ci(TRUE, c("PAF", "ML")), TRUE)
})

test_that("efa_fit stores the canonical estimator and rotation from lowercase input", {
  fit <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                 estimator = "ml", rotation = "geomint")
  expect_identical(fit$settings$estimator, "ML")
  # the settings keep a `method` entry duplicating `estimator`; it must stay in sync
  expect_identical(fit$settings$method, "ML")
  expect_identical(fit$settings$rotation, "geominT")
})

test_that("the retention criteria canonicalize case-insensitive choices (fast path)", {
  cmat <- test_models$baseline$cormat

  # efa_kgc's eigen_type (several.ok) and efa_ekc's use (single choice) resolve without an
  # EFA fit, so these exercise the .match_arg_ci swap on a default run
  kgc <- efa_kgc(cmat, eigen_type = c("pca", "smc"))
  expect_identical(kgc$settings$eigen_type, c("PCA", "SMC"))

  ekc <- efa_ekc(cmat, N = 500, use = "PAIRWISE.COMPLETE.OBS")
  expect_identical(ekc$settings$use, "pairwise.complete.obs")
})

test_that("efa_retain matches criteria case-insensitively and keeps canonical names", {
  skip_if_not_slow()

  set.seed(500)
  out <- efa_retain(test_models$baseline$cormat, N = 500,
                    criteria = c("hull", "parallel"), gof = "caf",
                    eigen_type_other = "smc", suitability = FALSE,
                    show_progress = FALSE)
  expect_identical(out$settings$criteria, c("HULL", "PARALLEL"))
  expect_identical(out$settings$gof, "CAF")
  expect_identical(out$settings$eigen_type_other, "SMC")
  # the per-criterion results are keyed by the canonical uppercase names
  expect_true(any(grepl("^HULL", names(out$n_factors))))
  expect_true(any(grepl("^PARALLEL", names(out$n_factors))))
})

test_that("the control constructors match the type preset case-insensitively", {
  expect_identical(estimate_control(type = "spss")$type, "SPSS")
  expect_identical(rotate_control(type = "efatools")$type, "EFAtools")
})

# The simulation-side choice arguments. `return_pop = TRUE` stops after the population is
# built, which is all these need: the value is canonicalized before any data are drawn.
sim_Lambda <- population_models$loadings$baseline
sim_Phi <- population_models$phis_3$moderate

test_that("efa_simulate matches marginals case-insensitively", {
  sim <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi,
                      marginals = "vm", return_pop = TRUE)
  expect_identical(sim$settings$marginals, "VM")
  expect_error(efa_simulate(Lambda = sim_Lambda, marginals = "bogus", return_pop = TRUE),
               class = "efa_bad_choice")
})

test_that("efa_simulate matches the model-error method case-insensitively", {
  sim <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi,
                      model_error = "tkl", target_rmsea = 0.05, return_pop = TRUE)
  expect_identical(sim$model_error$method, "TKL")
  expect_error(efa_simulate(Lambda = sim_Lambda, model_error = "bogus",
                            target_rmsea = 0.05, return_pop = TRUE),
               class = "efa_bad_choice")
})

test_that("efa_simulate matches the missing-data mechanism case-insensitively", {
  sim <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi,
                      missing = "mcar", missing_prop = 0.1, return_pop = TRUE)
  expect_identical(sim$settings$missing, "MCAR")
  expect_error(efa_simulate(Lambda = sim_Lambda, missing = "bogus",
                            missing_prop = 0.1, return_pop = TRUE),
               class = "efa_bad_choice")
})

test_that("efa_simulate matches the threshold-matching rule case-insensitively", {
  sim <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi, categories = 4,
                      match = "Polychoric", return_pop = TRUE)
  expect_identical(sim$settings$match, "polychoric")
  expect_error(efa_simulate(Lambda = sim_Lambda, categories = 4, match = "bogus",
                            return_pop = TRUE),
               class = "efa_bad_choice")
})

test_that("efa_power matches mode and type case-insensitively", {
  # the analytic path resolves both without any simulation
  pw <- efa_power(mode = "RMSEA", type = "NotClose", p = 18, k = 3, N = 200)
  expect_identical(pw$settings$mode, "rmsea")
  expect_identical(pw$settings$type, "notclose")
  expect_error(efa_power(mode = "bogus"), class = "efa_bad_choice")
  expect_error(efa_power(type = "bogus", p = 18, k = 3, N = 200),
               class = "efa_bad_choice")
})

test_that("efa_power matches the model-error method case-insensitively", {
  pw <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                  n_datasets = 3, criteria = "EKC", model_error = "tkl",
                  target_rmsea = 0.05, seed = 42)
  expect_identical(pw$settings$model_error, "TKL")
  expect_error(efa_power("simulation", Lambda = sim_Lambda, N = 150, n_datasets = 3,
                         criteria = "EKC", model_error = "bogus", target_rmsea = 0.05),
               class = "efa_bad_choice")
})

test_that("efa_mi matches the fit-pooling method case-insensitively", {
  cmats <- list(test_models$baseline$cormat, test_models$baseline$cormat)
  pooled <- efa_mi(cmats, n_factors = 3, N = 500, estimator = "PAF",
                   rotation = "promax", fit_pool_method = "d2")
  expect_identical(pooled$settings$fit_pool_method, "D2")
  expect_error(efa_mi(cmats, n_factors = 3, N = 500, fit_pool_method = "bogus"),
               class = "efa_bad_choice")
})

test_that("efa_scores matches the factor-score method case-insensitively", {
  fit <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500)
  expect_message(
    fs <- efa_scores(test_models$baseline$cormat, f = fit, method = "bartlett"),
    class = "efa_scores_needs_raw"
  )
  expect_identical(fs$settings$method, "Bartlett")
})

# --------------------------------------------------------------------------------------
# The rule, enumerated
# --------------------------------------------------------------------------------------
# One table listing every choice-valued argument a user can supply on the efa_* interface.
# It is the rule in executable form: a choice argument that is missing from it, or that is
# present and does not reject an unmatched value with the interface's condition class, is
# what this block exists to catch. Rejection happens in each function's argument-check
# block, before anything is computed, so the whole table runs in well under a second.

.choice_arg_cases <- function() {
  cmat <- test_models$baseline$cormat
  fit <- efa_fit(cmat, n_factors = 3, N = 500, rotation = "promax")
  loads <- unclass(fit$rot_loadings)
  case <- function(fn, args, ..., class = "efa_bad_choice") {
    lapply(c(...), function(a) list(fn = fn, args = args, arg = a, class = class))
  }
  c(
    case("efa_fit", list(cmat, n_factors = 3, N = 500), "se", "use", "cor_method"),
    case("efa_bartlett", list(cmat, N = 500), "use", "cor_method"),
    case("efa_kmo", list(cmat), "use", "cor_method"),
    case("efa_map", list(cmat), "use", "cor_method"),
    case("efa_smt", list(cmat, N = 500), "use", "cor_method"),
    case("efa_ekc", list(cmat, N = 500), "use", "cor_method"),
    case("efa_kgc", list(cmat), "use", "cor_method", "eigen_type"),
    case("efa_scree", list(cmat), "use", "cor_method", "eigen_type"),
    case("efa_hull", list(cmat, N = 500),
         "use", "cor_method", "decision_rule", "estimator", "gof", "eigen_type"),
    case("efa_parallel", list(cmat, N = 500, n_datasets = 5),
         "use", "cor_method", "decision_rule", "eigen_type"),
    case("efa_nest", list(cmat, N = 500, n_datasets = 5), "use", "cor_method"),
    case("efa_cd", list(GRiPS_raw[1:50, ]), "cor_method"),
    case("efa_screen", list(cmat, N = 500), "use", "cor_method"),
    case("efa_retain", list(cmat, N = 500, suitability = FALSE),
         "use", "cor_method", "decision_rule", "criteria", "estimator", "gof",
         "eigen_type_HULL", "eigen_type_other"),
    case("efa_average", list(cmat, n_factors = 3, N = 500, show_progress = FALSE),
         "use", "cor_method", "averaging"),
    case("efa_compare", list(loads, loads), "reorder"),
    case("efa_reliability", list(fit), "variance"),
    case("efa_procrustes", list(loads, loads), "rotation"),
    case("efa_mi", list(list(cmat, cmat), n_factors = 3, N = 500),
         "target_method", "align_unrotated", "fit_pool_method"),
    case("efa_scores", list(cmat, f = fit), "method"),
    case("efa_power", list(p = 18, k = 3, N = 200), "mode", "type"),
    # the simulation branch validates its own choices before it draws anything
    case("efa_power", list("simulation"), "criteria", "estimator", "model_error"),
    case("estimate_control", list(),
         "type", "init_comm", "criterion_type", "start_method",
         class = "efa_control_input"),
    case("rotate_control", list(),
         "type", "order_type", "varimax_type", "p_type",
         class = "efa_control_input"),
    # the print and plot layer, which takes its own choice arguments
    case("format", list(fit), "sort_loadings"),
    case("summary", list(fit), "ci", "ci_filter", "sort_loadings"),
    # `view` is not listed here: it is set internally ("brief" by format.efa(), "full" by
    # format.summary.efa()) and never reaches the renderer from a user's call
    case("print", list(summary(fit)), "ci", "ci_filter", "sort_loadings"),
    case("format", list(fit$rot_loadings), "name_style", "sort_loadings"),
    case("residuals", list(fit), "type")
  )
}

test_that("every choice-valued argument rejects an unmatched value with a classed error", {
  for (case in .choice_arg_cases()) {
    expect_error(
      do.call(case$fn, c(case$args, stats::setNames(list("bogus"), case$arg))),
      class = case$class,
      info = paste0(case$fn, "(", case$arg, " = \"bogus\")")
    )
  }
})

test_that("efa_average's grid arguments are case-insensitive but keep the subset check", {
  # These take a whole vector of values to build the averaging grid from, so they are
  # validated by checkmate::assert_subset() rather than by .match_arg_ci(): the value is
  # mapped onto the canonical spellings first, and an element matching none of them falls
  # through to the subset assertion, whose message already names the argument and lists the
  # admissible values. The condition is therefore deliberately unclassed here.
  cmat <- test_models$baseline$cormat
  avg <- efa_average(cmat, n_factors = 3, N = 500, estimator = c("paf", "PAF"),
                     rotation = c("Promax", "Oblimin"), type = "psych", init_comm = "SMC",
                     criterion_type = "SUM", varimax_type = "SVD", p_type = "NORM",
                     start_method = "Psych", show_progress = FALSE)
  # "paf" and "PAF" fold onto one grid entry, so the grid runs each requested value once
  expect_identical(avg$settings$estimator, "PAF")
  expect_identical(avg$settings$rotation, c("promax", "oblimin"))

  for (arg in c("estimator", "rotation", "type", "init_comm", "criterion_type",
                "varimax_type", "p_type", "start_method")) {
    expect_error(
      do.call(efa_average, c(list(cmat, n_factors = 3, N = 500, show_progress = FALSE),
                             stats::setNames(list("bogus"), arg))),
      info = arg
    )
  }
})

test_that("plot.efa_group rejects an unmatched type with a classed error", {
  mg <- efa_group(GRiPS_raw[1:100, ],
                  groups = rep(c("g1", "g2"), length.out = 100), n_factors = 1)
  expect_error(plot(mg, type = "bogus"), class = "efa_bad_choice")
  expect_s3_class(plot(mg, type = "Congruence"), "ggplot")
})

test_that("the correlation and inference arguments canonicalize a flipped spelling", {
  cmat <- test_models$baseline$cormat

  fit <- efa_fit(cmat, n_factors = 3, N = 500, estimator = "ML",
                 cor_method = "Pearson", use = "Complete.Obs", se = "Information")
  expect_identical(fit$settings$cor_method, "pearson")
  expect_identical(fit$settings$use, "complete.obs")
  expect_identical(fit$settings$se, "information")

  expect_identical(efa_kmo(cmat, cor_method = "Kendall")$settings$cor_method, "kendall")
  # efa_hull() does not record `decision_rule`, so the flipped spelling is pinned by the
  # result it produces being the one the canonical spelling produces
  expect_identical(efa_parallel(cmat, N = 500, n_datasets = 20, decision_rule = "Crawford",
                                eigen_type = "PCA")$settings$decision_rule, "crawford")
  expect_identical(efa_average(cmat, n_factors = 3, N = 500, rotation = "none",
                               type = "none", averaging = "Median",
                               show_progress = FALSE)$settings$averaging, "median")

  # the tuning knobs, which reach the constructors from the flat interface as well
  expect_identical(estimate_control(init_comm = "SMC")$init_comm, "smc")
  expect_identical(estimate_control(criterion_type = "SUM")$criterion_type, "sum")
  expect_identical(estimate_control(start_method = "Factanal")$start_method, "factanal")
  expect_identical(rotate_control(order_type = "Eigen")$order_type, "eigen")
  expect_identical(rotate_control(varimax_type = "SVD")$varimax_type, "svd")
  expect_identical(rotate_control(p_type = "NORM")$p_type, "norm")

  # the print layer returns text, so the flipped value is asserted by it not erroring
  expect_identical(format(fit, sort_loadings = "Primary"),
                   format(fit, sort_loadings = "primary"))
  expect_identical(residuals(fit, type = "Raw"), residuals(fit, type = "raw"))
})

test_that("the superseded uppercase functions keep matching their choices exactly", {
  # These four are frozen on the pre-rename behaviour (see R/EFAtools-superseded.R), so a
  # flipped spelling must still be refused there even though the successors accept it.
  cmat <- test_models$baseline$cormat
  efa <- EFA(cmat, n_factors = 3, N = 500, rotation = "promax")

  # A bare expect_error() would also be satisfied by the successors' own condition, which is
  # exactly what must not appear here, so assert the class the frozen path may never raise.
  refuses_exactly <- function(expr) {
    cnd <- tryCatch({expr; NULL}, error = function(e) e)
    !is.null(cnd) && !inherits(cnd, "efa_bad_choice")
  }

  expect_true(refuses_exactly(EFA(cmat, n_factors = 3, N = 500, type = "efatools")))
  expect_true(refuses_exactly(EFA(cmat, n_factors = 3, N = 500, rotation = "Promax")))
  expect_true(refuses_exactly(SL(efa, n_factors = 3, type = "efatools")))
  expect_true(refuses_exactly(OMEGA(efa, type = "efatools")))
  expect_true(refuses_exactly(FACTOR_SCORES(cmat, f = efa, method = "Regression")))
})

# --------------------------------------------------------------------------------------
# NULL selects the documented default
# --------------------------------------------------------------------------------------
# .match_arg_ci() resolves a NULL argument to the caller's formal default, not to the first
# admissible value. The two differ wherever an argument takes several values at once, or
# wherever its choices are listed in another order than its default, and taking the first
# choice there silently ran something narrower than the help page documents.

test_that(".match_arg_ci resolves NULL to the caller's formal default", {
  f <- function(estimator = c("ML", "PAF", "ULS")) .match_arg_ci(estimator)
  expect_identical(f(NULL), "ML")

  # several.ok: the whole default vector, not its first element
  g <- function(gof = c("CAF", "CFI", "RMSEA")) .match_arg_ci(gof, several.ok = TRUE)
  expect_identical(g(NULL), c("CAF", "CFI", "RMSEA"))

  # an explicit `choices` ordered differently from the default
  h <- function(eigen_type = "SMC") {
    .match_arg_ci(eigen_type, c("PCA", "SMC", "EFA"), several.ok = TRUE)
  }
  expect_identical(h(NULL), "SMC")

  # a formal that is itself NULL, or has no default at all, falls back to the first choice
  k <- function(match = NULL) .match_arg_ci(match, c("thresholds", "polychoric"))
  expect_identical(k(NULL), "thresholds")
  m <- function(match) .match_arg_ci(match, c("thresholds", "polychoric"))
  expect_identical(m(NULL), "thresholds")
})

test_that("NULL selects the documented default at every affected call site", {
  cmat <- test_models$baseline$cormat

  expect_identical(efa_kgc(cmat, eigen_type = NULL)$settings$eigen_type,
                   efa_kgc(cmat)$settings$eigen_type)
  expect_identical(efa_scree(cmat, eigen_type = NULL)$settings$eigen_type,
                   efa_scree(cmat)$settings$eigen_type)
  expect_identical(efa_parallel(cmat, N = 500, n_datasets = 5,
                                eigen_type = NULL)$settings$eigen_type,
                   c("PCA", "SMC", "EFA"))
  expect_identical(efa_hull(cmat, N = 500, estimator = "ML", gof = NULL)$settings$gof,
                   c("CAF", "CFI", "RMSEA"))
  expect_identical(
    efa_retain(cmat, N = 500, criteria = "KGC", suitability = FALSE,
               eigen_type_other = NULL)$settings$eigen_type_other,
    "SMC")
  expect_identical(
    efa_retain(cmat, N = 500, criteria = "HULL", suitability = FALSE, estimator = "ML",
               gof = NULL)$settings$gof,
    c("CAF", "CFI", "RMSEA"))

  # efa_simulate()'s `match` documents NULL as "thresholds"; that must keep working
  sim <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi, categories = 4,
                      match = NULL, return_pop = TRUE)
  expect_identical(sim$settings$match, "thresholds")
})

test_that("efa_retain(criteria = NULL) runs the six documented default criteria", {
  skip_if_not_slow()

  set.seed(500)
  # CD is one of the six defaults and needs raw data, so a correlation matrix skips it with
  # a warning -- which is itself evidence that all six were requested
  expect_warning(
    out <- efa_retain(test_models$baseline$cormat, N = 500, criteria = NULL,
                      suitability = FALSE, n_datasets = 50, n_datasets_nest = 50,
                      show_progress = FALSE),
    class = "efa_criterion_skipped"
  )
  expect_identical(out$settings$criteria,
                   eval(formals(efa_retain)$criteria))
})

test_that("efa_power(criteria = NULL) evaluates the documented default criteria", {
  skip_if_not_slow()

  pw <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                  n_datasets = 3, criteria = NULL, seed = 42)
  expect_identical(pw$settings$criteria, eval(formals(efa_power)$criteria))
})

#' Kaiser-Meyer-Olkin criterion
#'
#' This function computes the Kaiser-Meyer-Olkin (KMO) criterion overall and for
#' each variable in a correlation matrix. The KMO represents the degree to
#' which each observed variable is predicted by the other variables in the
#' dataset and with this indicates the suitability for factor analysis.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#'  correlations.
#' @param use character. The missing-data policy for raw data. Passed to
#'  [stats::cor()] for `"pearson"`, `"spearman"`, and `"kendall"`; for `"poly"` /
#'  `"tetra"` the same policies are applied to the raw data before the polychoric
#'  estimation, where `"all.obs"` and `"everything"` abort on a missing value instead
#'  of returning `NA` correlations. Default is "pairwise.complete.obs".
#' @param cor_method character. Correlation computed from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (passed to [stats::cor()]), or `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations of ordinal / binary data
#'   (a two-step estimator).
#' Default is "pearson".
#'
#' @details Kaiser (1970) proposed this index, originally called measure of
#' sampling adequacy (MSA), that indicates how near the inverted correlation
#' matrix \eqn{R^{-1}} is to a diagonal matrix to determine a given
#' correlation matrix's (\eqn{R}) suitability for factor analysis.
#' The index is
#' \deqn{KMO = \frac{\sum_{i \neq j} r_{ij}^2}{\sum_{i \neq j} r_{ij}^2 + \sum_{i \neq j} q_{ij}^2}}
#' with \eqn{Q = SR^{-1}S} and S = \eqn{(diag R^{-1})^{-1/2}} where
#' \eqn{\sum_{i \neq j} r_{ij}^2} is the sum of squares of the off-diagonal
#' elements of \eqn{R} and \eqn{\sum_{i \neq j} q_{ij}^2} is the sum of squares of
#' the off-diagonal elements of \eqn{Q} (see also Cureton & D'Agostino, 1983).
#'
#' So KMO varies between 0 and 1, with larger values indicating higher suitability
#' for factor analysis. Kaiser and Rice (1974) suggest that KMO should at least
#' exceed .50 for a correlation matrix to be suitable for factor analysis.
#'
#' This function was heavily influenced by the [psych::KMO()]
#' function.
#'
#' See also [efa_bartlett()] for another test of suitability for factor
#' analysis.
#'
#' The `efa_kmo` function can also be called together with the
#' [efa_bartlett()] function and with factor retention criteria in the
#'  [efa_retain()] function.
#'
#' @return A list containing
#' \item{KMO}{Overall KMO.}
#' \item{KMO_i}{KMO for each variable.}
#' \item{settings}{A list of the settings used.}
#'
#' @export
#'
#' @source Kaiser, H. F. (1970). A second generation little jiffy. Psychometrika,
#' 35, 401-415.
#' @source Kaiser, H. F. & Rice, J. (1974). Little jiffy, mark IV. Educational
#' and Psychological Measurement, 34, 111-117.
#' @source Cureton, E. E. & D'Agostino, R. B. (1983). Factor analysis: An
#'  applied approach. Hillsdale, N.J.: Lawrence Erlbaum Associates, Inc.
#'
#' @family factor analysis suitability
#'
#' @seealso [efa_bartlett()] for another measure to determine
#' suitability for factor analysis.
#'
#' [efa_retain()] as a wrapper function for this function,
#' [efa_bartlett()] and several factor retention criteria.
#'
#' @examples
#' efa_kmo(test_models$baseline$cormat)
efa_kmo <- function(x, use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                               "everything", "na.or.complete"),
                    cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")) {

  # Perform argument checks
  .assert_cor_input(x)

  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, use = use, cor_method = cor_method,
                             N_policy = "none")
  R <- prep$R

  KMO_list <- .compute_kmo(R)
  KMO <- KMO_list$KMO
  KMO_i <- KMO_list$KMO_i

  if(!is.null(colnames(R))){

    names(KMO_i) <- colnames(R)

  } else if(!is.null(rownames(R))) {

    names(KMO_i) <- rownames(R)

  } else {

    names(KMO_i) <- paste0("V", seq_len(ncol(R)))

  }

  # Prepare settings
  settings <- list(use = use,
                   cor_method = cor_method)

  output <- list(KMO = KMO,
                 KMO_i = KMO_i,
                 settings = settings)

  # the trailing legacy class is load-bearing: it keeps `inherits(x, "KMO")`
  # working in code written against the superseded name. Do not drop it.
  class(output) <- c("efa_kmo", "KMO")

  return(output)

}

# outsourced computation of KMO to
# use it when computing CAF which needs
# different input checks
#
# `R_inv` lets a caller that has already formed R^-1 hand it over instead of paying for a
# second p x p inversion (efa_screen() builds one inverse for several measures at once).
.compute_kmo <- function(R, R_inv = NULL) {

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- if (is.null(R_inv)) try(solve(R), silent = TRUE) else R_inv

  if (inherits(R_i, "try-error")) {
    cli::cli_abort("The matrix is singular; KMO cannot be computed.",
                   class = "efa_cor_singular")
  }

  # Start computations
  S <- diag(diag(R_i)^(-0.5))
  Q <- S %*% R_i %*% S
  diag(Q) <- 0
  diag(R) <- 0
  sumQ2 <- sum(Q^2)
  sumR2 <- sum(R^2)
  KMO <- sumR2/(sumR2 + sumQ2)
  KMO_i <- colSums(R^2)/(colSums(R^2) + colSums(Q^2))
  list(
    KMO = KMO,
    KMO_i = KMO_i
  )
}

#' Print function for PARALLEL objects
#'
#' @param x a list of class PARALLEL. Output from \link{PARALLEL} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print PARALLEL
#'
#' @examples
#' \donttest{
#' # example without real data
#' PARALLEL(N = 500, n_vars = 10)
#'
#' # example with correlation matrix and "ML" estimation
#' PARALLEL(test_models$case_11b$cormat, N = 500, method = "ML")
#' }
print.PARALLEL <- function(x, plot = TRUE, ...) {

  eigen_type <- x$settings$eigen_type

  cat("Parallel Analysis performed using ", crayon::bold(x$settings$n_datasets),
      " simulated random data sets", sep = "")
  cat("\n")

  cat("Eigenvalues were found using ", .settings_string(x$settings$eigen_type),
      "", sep = "")
  cat("\n")

  if (isTRUE(x$settings$x_dat)) {

    cat("\n")
    cat("Decision rule used:", crayon::bold(x$settings$decision_rule))
    cat("\n")
    cat("\n")
    cat(cli::rule(left = crayon::bold("Number of factors to retain according to"),
                  col = "blue"))
    cat("\n")
    cat("\n")

    if("PCA" %in% eigen_type){
      cat(crayon::blue(cli::symbol$bullet, "PCA-determined eigenvalues: "),
          crayon::bold(x$n_fac_PCA))
      cat("\n")
    }

    if("SMC" %in% eigen_type){
      cat(crayon::blue(cli::symbol$bullet, "SMC-determined eigenvalues: "),
          crayon::bold(x$n_fac_SMC))
      cat("\n")
    }

    if("EFA" %in% eigen_type){
      cat(crayon::blue(cli::symbol$bullet, "EFA-determined eigenvalues: "),
          crayon::bold(x$n_fac_EFA))
      cat("\n")
    }

  } else {

    cat("\n")
    cat(cli::col_cyan(cli::symbol$info, " No data were entered to base number of factors on. Plotting simulated eigenvalues."))
    cat("\n")

  }

  if (isTRUE(plot)) {
    graphics::plot(x)
  }

}

#' Residuals function for EFA objects
#'
#' @param object a list of class EFA. Output from \code{\link{EFA}} or
#'   \code{\link{EFA_POOLED}}.
#' @param print logical. Whether to print residuals. If FALSE, they are just returned.
#' @param digits numeric. The number of digits to round printed residuals.
#' @param ... Further arguments.
#'
#' @export
#' @method residuals EFA
residuals.EFA <- function(object, print = TRUE, digits = 3, ...) {

  out <- list(
    residuals = object$residuals
  )

  if ("standardized_residuals" %in% names(object)) {
    out$standardized_residuals <- object$standardized_residuals
  }

  if (isFALSE(print)) {

    return(out)
  }

  cli::cli_h2("Raw residuals, based on {.code orig_R - model_implied_R}")

  print(round(object$residuals, digits = digits))



  if ("standardized_residuals" %in% names(object)) {

    cli::cli_h2("Standardized residuals, based on {.code residuals / SE}")

    print(round(object$standardized_residuals, digits = digits))

  }

  invisible(out)
}

#' Print and format an OMEGA object
#'
#' `print()` shows the omega coefficients computed by [OMEGA()]: omega total (and,
#' for multi-factor solutions, omega hierarchical, omega subscale, the H index,
#' the explained common variance, and the percent of uncontaminated correlations)
#' for the general factor and the group factors, for a single group or for each
#' group. `format()` assembles the same report and returns it as a character
#' vector; `print()` is `cat(format(x), sep = "\n")`. The lines follow the active
#' console theme, so they are plain when colours are disabled (for example when
#' captured into a file or stripped with [cli::ansi_strip()]).
#'
#' @param x An object of class `OMEGA` (output from [OMEGA()]).
#' @param digits Integer. The number of decimal places the coefficients are
#'   rounded to (passed to [base::round()]). Default is 3.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @method print OMEGA
#'
#' @export
#'
#' @examples
#' efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' sl_mod <- efa_schmid_leiman(efa_mod, estimator = "PAF")
#'
#' om <- OMEGA(sl_mod, type = "EFAtools",
#'             factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2)
#' om
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(om))
#'
print.OMEGA <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.OMEGA
#' @export
#' @method format OMEGA
format.OMEGA <- function(x, digits = 3, ...) {

  cli::cli_format_method({

    # Render an omega-coefficient matrix as before (base R's matrix layout with blank NA
    # cells), emitted verbatim so cli does not reflow the aligned columns.
    emit_matrix <- function(mat) {
      cli::cli_verbatim(utils::capture.output(
        print(round(unclass(mat), digits = digits), na.print = "")))
    }

    if (is.list(x)) {

      # In case of multiple groups
      group_names <- names(x)

      if (length(x[[1]]) == 1) {

        # A single factor, omega total only
        cli::cli_text("{.strong Omega total for the single factor for each group:}")
        for (i in seq_along(group_names)) {
          grp <- group_names[i]
          value <- round(x[[i]], digits = digits)
          cli::cli_text("")
          cli::cli_text("Group {.emph {grp}}: {value}")
        }

      } else if (length(x[[1]]) == 2) {

        # A single factor, omega total and H index
        cli::cli_text("{.strong Omega total and H index for the single factor for each group:}")
        for (i in seq_along(group_names)) {
          grp <- group_names[i]
          omega <- round(x[[i]][1], digits = digits)
          h_index <- round(x[[i]][2], digits = digits)
          cli::cli_text("")
          cli::cli_text("Group {.emph {grp}}:")
          cli::cli_text("Omega: {omega}")
          cli::cli_text("H index: {h_index}")
        }

      } else {

        # The full coefficient matrix
        desc <- if (ncol(x[[1]]) == 3) {
          paste0("Omega total, omega hierarchical, and omega subscale for the ",
                 "general factor (top row) and the group factors for each group:")
        } else {
          paste0("Omega total, omega hierarchical, omega subscale, H index, ",
                 "explained common variance (ECV), and percent of uncontaminated ",
                 "correlations (PUC) for the general factor (top row) and omegas and ",
                 "H index for the group factors for each group:")
        }
        cli::cli_text("{.strong {desc}}")

        for (i in seq_along(group_names)) {
          grp <- group_names[i]
          cli::cli_text("")
          cli::cli_text("Group {.emph {grp}}:")
          emit_matrix(x[[i]])
        }

      }

    } else {

      # In case of a single group
      if (length(x) == 1) {

        value <- round(unclass(x), digits = digits)
        cli::cli_text("{.strong Omega total for the single factor:} {value}")

      } else if (length(x) == 2) {

        omega <- round(unclass(x)[1], digits = digits)
        h_index <- round(unclass(x)[2], digits = digits)
        cli::cli_text("{.strong Omega total and H index for the single factor:}")
        cli::cli_text("")
        cli::cli_text("Omega: {omega}")
        cli::cli_text("H index: {h_index}")

      } else {

        desc <- if (ncol(x) == 3) {
          paste0("Omega total, omega hierarchical, and omega subscale for the ",
                 "general factor (top row) and the group factors:")
        } else {
          paste0("Omega total, omega hierarchical, omega subscale, H index, ",
                 "explained common variance (ECV), and percent of uncontaminated ",
                 "correlations (PUC) for the general factor (top row) and omegas and ",
                 "H index for the group factors:")
        }
        cli::cli_text("{.strong {desc}}")
        cli::cli_text("")
        emit_matrix(x)

      }

    }
  })
}

#' population_models
#'
#' Population factor models, some of which (baseline to case_11e) used for the
#' simulation analyses reported in Grieder and Steiner (2022). All combinations
#' of the pattern matrices and the factor
#' intercorrelations were used in the simulations. Many models are based on cases
#' used in de Winter and Dodou (2012).
#'
#'
#' @format A list of 3 lists "loadings", "phis_3", and "phis_6".
#' \describe{
#' `loadings` contains the following matrices of pattern coefficients:
#'   \item{baseline}{(matrix) - The pattern coefficients of the baseline model. Three factors with six indicators each, all with pattern coefficients of .6. Same baseline model as used in de Winter and Dodou (2012).}
#'   \item{case_1a}{(matrix) - Three factors with 2 indicators per factor.}
#'   \item{case_1b}{(matrix) - Three factors with 3 indicators per factor. Case 5 in de Winter and Dodou (2012).}
#'   \item{case_1c}{(matrix) - Three factors with 4 indicators per factor.}
#'   \item{case_1d}{(matrix) - Three factors with 5 indicators per factor.}
#'   \item{case_2}{(matrix) - Same as baseline model but with low pattern coefficients of .3.}
#'   \item{case_3}{(matrix) - Same as baseline model but with high pattern coefficients of .9.}
#'   \item{case_4}{(matrix) - Three factors with different pattern coefficients *between* factors (one factor with .9, one with .6, and one with .3, respectively). Case 7 in de Winter and Dodou (2012).}
#'   \item{case_5}{(matrix) - Three factors with different pattern coefficients *within* factors (each factor has two pattern coefficients of each .9, .6, and .3). Similar to cases 8/ 9 in de Winter and Dodou (2012).}
#'   \item{case_6a}{(matrix) - Same as baseline model but with one cross loading of .4. Similar to case 10 in de Winter and Dodou (2012).}
#'   \item{case_6b}{(matrix) - Same as baseline model but with three cross loading of .4 (One factor with 2 and one with 1 crossloading). Similar to case 10 in de Winter and Dodou (2012).}
#'   \item{case_7}{(matrix) - Three factors with different number of indicators per factor (2, 4, and 6 respectively). Similar to cases 11/ 12 in de Winter and Dodou (2012).}
#'   \item{case_8}{(matrix) - Three factors with random variation in pattern coefficients added, drawn from a uniform distribution between \[-.2, .2\]. Case 13 in de Winter and Dodou (2012).}
#'   \item{case_9a}{(matrix) - Three factors with 2 indicators per factor, with different pattern coefficients within one of the factors.}
#'   \item{case_9b}{(matrix) - Three factors with 3 indicators per factor, with different pattern coefficients.}
#'   \item{case_9c}{(matrix) - Three factors with 4 indicators per factor, with different pattern coefficients.}
#'   \item{case_9d}{(matrix) - Three factors with 5 indicators per factor, with different pattern coefficients.}
#'   \item{case_10a}{(matrix) - Six factors with 2 indicators per factor, all with pattern coefficients of .6.}
#'   \item{case_10b}{(matrix) - Six factors with 3 indicators per factor, all with pattern coefficients of .6.}
#'   \item{case_10c}{(matrix) - Six factors with 4 indicators per factor, all with pattern coefficients of .6.}
#'   \item{case_10d}{(matrix) - Six factors with 5 indicators per factor, all with pattern coefficients of .6.}
#'   \item{case_10e}{(matrix) - Six factors with 6 indicators per factor, all with pattern coefficients of .6.}
#'   \item{case_11a}{(matrix) - Six factors with 2 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
#'   \item{case_11b}{(matrix) - Six factors with 3 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
#'   \item{case_11c}{(matrix) - Six factors with 4 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
#'   \item{case_11d}{(matrix) - Six factors with 5 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
#'   \item{case_11e}{(matrix) - Six factors with 6 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
#'   \item{case_12a}{(matrix) - One factor, with 2 equal pattern coefficients (.6).}
#'   \item{case_12b}{(matrix) - One factor, with 3 equal pattern coefficients (.6).}
#'   \item{case_12c}{(matrix) - One factor, with 6 equal pattern coefficients (.6).}
#'   \item{case_12d}{(matrix) - One factor, with 10 equal pattern coefficients (.6).}
#'   \item{case_12e}{(matrix) - One factor, with 15 equal pattern coefficients (.6).}
#'   \item{case_13a}{(matrix) - One factor, with 2 different pattern coefficients (.3, and .6).}
#'   \item{case_13b}{(matrix) - One factor, with 3 different pattern coefficients (.3, .6, and .9).}
#'   \item{case_13c}{(matrix) - One factor, with 6 different pattern coefficients (.3, .6, and .9).}
#'   \item{case_13d}{(matrix) - One factor, with 10 different pattern coefficients (.3, .6, and .9).}
#'   \item{case_13e}{(matrix) - One factor, with 15 different pattern coefficients (.3, .6, and .9).}
#'   \item{case_14a}{(matrix) - No factor, 2 variables (0).}
#'   \item{case_14b}{(matrix) - No factor, 3 variables (0).}
#'   \item{case_14c}{(matrix) - No factor, 6 variables (0).}
#'   \item{case_14d}{(matrix) - No factor, 10 variables (0).}
#'   \item{case_14e}{(matrix) - No factor, 15 variables (0).}
#'   `phis_3` contains the following 3x3 matrices:
#'   \item{zero}{(matrix) - Matrix of factor intercorrelations of 0. Same intercorrelations as used in de Winter and Dodou (2012).}
#'   \item{moderate}{(matrix) - Matrix of moderate factor intercorrelations of .3.}
#'   \item{mixed}{(matrix) - Matrix of mixed (.3, .5, and .7) factor intercorrelations.}
#'   \item{strong}{(matrix) - Matrix of strong factor intercorrelations of .7. Same intercorrelations as used in de Winter and Dodou (2012).}
#'   `phis_6` contains the following 6x6 matrices:
#'   \item{zero}{(matrix) - Matrix of factor intercorrelations of 0. Same intercorrelations as used in de Winter and Dodou (2012).}
#'   \item{moderate}{(matrix) - Matrix of moderate factor intercorrelations of .3.}
#'   \item{mixed}{(matrix) - Matrix of mixed (around .3, .5, and .7; smoothing was necessary for the matrix to be positive definite) factor intercorrelations.}
#'   \item{strong}{(matrix) - Matrix of strong factor intercorrelations of .7. Same intercorrelations as used in de Winter and Dodou (2012).}
#'  }
#' @source Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
#' of implementations of principal axis factoring and promax rotation in R and SPSS.
#' Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x
#' @source de Winter, J.C.F., & Dodou, D. (2012). Factor recovery by principal axis factoring and maximum likelihood factor analysis as a function of factor pattern and sample size. Journal of Applied Statistics. 39.
"population_models"

#' Print KMO object
#'
#' @param x list of class KMO (output from the \link{KMO} function)
#' @param ... additional arguments passed to print
#'
#' @method print KMO
#'
#' @export
#'
#' @examples
#' KMO_base <- KMO(test_models$baseline$cormat)
#' KMO_base
#'
print.KMO <- function(x, ...) {

  KMO <- x$KMO

  cat("\n")
  cat(cli::rule(left = "Kaiser-Meyer-Olkin criterion (KMO)", col = "blue"))
  cat("\n")

  if(!is.na(KMO) && !is.null(KMO)){

    if(KMO >= .9){
      symb <- crayon::green$bold(cli::symbol$tick)
      label <- crayon::green$bold("marvellous")
    } else if(KMO >= .8){
      symb <- crayon::green$bold(cli::symbol$tick)
      label <- crayon::green$bold("meritorious")
    } else if(KMO >= .7){
      symb <- crayon::green$bold(cli::symbol$tick)
      label <- crayon::green$bold("middling")
    } else if(KMO >= .6){
      symb <- crayon::yellow$bold("!")
      label <- crayon::yellow$bold("mediocre")
    } else if (KMO >= .5){
      symb <- crayon::red$bold(cli::symbol$cross)
      label <- crayon::red$bold("miserable")
    } else {
      symb <- crayon::red$bold(cli::symbol$cross)
      label <- crayon::red$bold("unacceptable")
    }

    cat("\n")
    cat(symb, " The overall KMO value for your data is ", label, ".", sep = "")
    cat("\n")

    if(KMO < .5){
      cat(crayon::bold(" "), "These data are not suitable for factor analysis.")
      cat("\n")
    } else if(KMO < .6){
      cat(crayon::bold(" "), "These data are hardly suitable for factor analysis.")
      cat("\n")
    } else {
      cat(crayon::bold(" "), "These data are probably suitable for factor analysis.")
      cat("\n")
    }

    cat("\n")
    cat(crayon::bold(" "), crayon::blue("Overall:"), crayon::bold(round(KMO, 3)))
    cat("\n")
    cat("\n")

    cat(crayon::bold(" "), crayon::blue("For each variable:"))
    cat("\n")
    print(round(x$KMO_i, 3))

  } else {

    cat("\n")
    cat(crayon::yellow(crayon::bold("!"), "Sorry, the KMO value for your data is not available."))
    cat("\n")

  }

}

#' Estimate factor scores and score-quality diagnostics for an EFA model
#'
#' Computes factor-score weights, and (from raw data) the factor scores
#' themselves, for an [efa_fit()] solution or a directly supplied loading matrix.
#' It also returns score-quality diagnostics: the score intercorrelations, the
#' determinacy (validity) and univocality of each score, and Guttman's
#' indeterminacy index. Factor scores are returned only when raw data are
#' supplied; a correlation matrix yields the weights and diagnostics alone.
#'
#' @details
#' Each method combines the loadings with some or all of the factor correlations and the
#' scoring correlation matrix into weights in a different way:
#' \describe{
#'   \item{`"regression"`}{Thurstone's (1935) regression scores.}
#'   \item{`"Bartlett"`}{Bartlett's (1937) scores.}
#'   \item{`"Anderson"`}{Anderson & Rubin's (1956) scores.}
#'   \item{`"tenBerge"`}{ten Berge, Krijnen, Wansbeek & Shapiro's (1999) scores.}
#'   \item{`"Harman"`}{Harman's (1976) scores, based on an idealized variable (a
#'     hypothetical variable that would correlate perfectly with the factor).}
#'   \item{`"components"`}{Component scores. These are formed from the raw,
#'     uncentered data (`X %*% W`) rather than the standardized data, so unlike
#'     the other methods they are on the scale of the input variables. The
#'     diagnostics below describe the standardized combination `scale(X) %*% W`,
#'     and therefore differ from the realized correlations of the returned scores
#'     whenever the variables have unequal variances.}
#' }
#' The determinacy (validity) of a score is its correlation with the factor it
#' estimates, computed from the returned weights; for regression scores it is the
#' multiple correlation between the factor and the observed variables (Guttman,
#' 1955; Grice, 2001). The off-diagonal score-factor correlations give the
#' univocality (the correlation of a score with the *other* factors). Guttman's
#' (1955) indeterminacy index, `2 rho^2 - 1`, is the minimum correlation between
#' two equally valid sets of scores. For a `method` other than `"regression"`
#' both quantities are specific to those scores: the determinacy is the method's
#' own score-factor correlation (never larger than the regression value), and
#' the reported `guttman` follows from it.
#'
#' Determinacies close to 1 mean the scores stand in for the factor with little
#' loss; Grice (2001) regards values of about .90 and above as the level required
#' before scores are interpreted for individual cases, and treats lower values as
#' usable only for group-level research. The Guttman index makes the same point
#' more sharply, because a factor score is never the factor: at `rho = .90` two
#' equally valid sets of scores can still correlate as low as .62, and at
#' `rho = .80` as low as .28, so the rank order of cases is not unique.
#'
#' Which `method` to prefer follows from what the scores are for. Regression
#' scores correlate most highly with the factor, but they are biased towards it
#' and correlate across factors even when the model is orthogonal. Bartlett
#' scores are conditionally unbiased, which makes them the choice when the scores
#' stand in for the factor in a later model. `"tenBerge"` reproduces the factor
#' intercorrelations `Phi`, so it is the choice when the scores will be correlated
#' with each other or with external variables. `"Anderson"` forces the scores to
#' be uncorrelated with unit variance and is appropriate only when the factors
#' themselves are orthogonal. `"components"` is a weighted sum of the observed
#' variables rather than an estimate of a common factor.
#'
#' @param x data.frame or matrix. Raw data (needed to obtain factor scores) or a
#'   correlation matrix (yields weights and diagnostics only). When `f` is a
#'   directly supplied loading matrix, a correlation-matrix `x` also supplies the
#'   correlations the weights are derived from; when `f` is an [efa_fit()] object,
#'   its own fitted correlations are used instead (supply `rho` to derive the
#'   weights from another matrix). `x` describes the model variables either way.
#'   When raw data carry column names, they are matched to the model variables by
#'   name (any extra columns are ignored, and a model variable missing from `x` is
#'   an error). A named correlation matrix is likewise matched to the loading rows
#'   by name; its row and column names must use the same order, and it must carry
#'   one row and column for each model variable. Unnamed input is matched by
#'   position.
#'
#'   Raw data are scored as supplied: no imputation is performed, so a case with a
#'   missing value on any model variable receives `NA` scores (and is reported as
#'   not scored). A model variable that carries no usable spread in `x` -- constant,
#'   infinite, or observed fewer than twice -- is an error.
#' @param f object of class [efa_fit()], an `efa_loadings` object, or a matrix of factor
#'   loadings.
#' @param Phi matrix. Factor intercorrelations. Only used when a loading matrix is
#'   supplied directly in `f`; taken from the `efa` object otherwise, in which
#'   case a supplied `Phi` is ignored with a warning. Named rows and columns are
#'   matched to the loading columns and must use the same order.
#'   Default is `NULL`, in which case the factors are assumed uncorrelated.
#' @param rho matrix. Correlation matrix used to derive the scoring weights.
#'   Defaults to `NULL`, in which case `f$orig_R` is used for an `efa` object; for
#'   a directly supplied loading matrix, `x` itself when it is a correlation
#'   matrix, and `cor(x, use = "pairwise")` otherwise. Pass a matrix here to score
#'   against a correlation other than the one implied by `f`/`x`. Named rows and
#'   columns are matched to the loading rows; row and column names must use the
#'   same order.
#' @param method character. The factor-score method: one of `"regression"`
#'   (default), `"Bartlett"`, `"Anderson"`, `"tenBerge"`, `"Harman"`, or
#'   `"components"`.
#'
#' @return An object of class `efa_scores`, a list containing:
#'
#' \item{weights}{The `p` by `m` factor-score weight matrix.}
#' \item{scores}{The factor scores (`n` by `m`), or `NULL` when a correlation
#'   matrix was supplied. A case with a missing value on any model variable is not
#'   scored and keeps `NA` in every column.}
#' \item{r.scores}{The `m` by `m` correlations of the factor-score estimates (see
#'   Details for the `"components"`-method scale caveat).}
#' \item{score_cor}{The `m` by `m` score-factor correlation matrix; its diagonal
#'   is the determinacy (validity) of each score and its off-diagonals the
#'   univocality.}
#' \item{determinacy}{A data frame with, per factor, the determinacy `rho`, the
#'   squared determinacy `rho2`, and Guttman's indeterminacy index `guttman`.}
#' \item{settings}{A list of the settings used, including the number of supplied
#'   observations `n_obs` and the number of them that could be scored `n_scored`.}
#'
#' @source Thurstone, L. L. (1935). *The vectors of mind*. University of Chicago
#'   Press.
#'
#'   Bartlett, M. S. (1937). The statistical conception of mental factors.
#'   *British Journal of Psychology, 28*, 97-104.
#'
#'   Anderson, T. W., & Rubin, H. (1956). Statistical inference in factor
#'   analysis. In *Proceedings of the Third Berkeley Symposium on Mathematical
#'   Statistics and Probability* (Vol. 5, pp. 111-150). University of California
#'   Press.
#'
#'   Guttman, L. (1955). The determinacy of factor score matrices with
#'   implications for five other basic problems of common-factor theory.
#'   *British Journal of Statistical Psychology, 8*, 65-81.
#'
#'   ten Berge, J. M. F., Krijnen, W. P., Wansbeek, T., & Shapiro, A. (1999).
#'   Some new results on correlation-preserving factor scores prediction methods.
#'   *Linear Algebra and its Applications, 289*, 311-318.
#'
#'   Grice, J. W. (2001). Computing and evaluating factor scores. *Psychological
#'   Methods, 6*, 430-450.
#'
#' @seealso [efa_fit()] for the solution these are computed from.
#'
#' @family factor scoring
#'
#' @export
#'
#' @examples
#' # Weights and score diagnostics from an EFA on a correlation matrix
#' efa <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                estimator = "PAF", rotation = "oblimin")
#' fs <- efa_scores(test_models$baseline$cormat, f = efa)
#' fs
#' summary(fs)
#'
#' # Factor scores from raw data (Bartlett method)
#' \donttest{
#' efa_raw <- efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF")
#' efa_scores(GRiPS_raw, f = efa_raw, method = "Bartlett")
#' }
#'
#' # Loadings supplied directly, with the factor intercorrelations
#' efa_scores(test_models$baseline$cormat, f = efa$rot_loadings, Phi = efa$Phi)
#'
efa_scores <- function(x, f, Phi = NULL, rho = NULL,
                       method = c("regression", "Bartlett", "Anderson",
                                  "tenBerge", "Harman", "components")) {

  .assert_cor_input(x)

  # A correlation matrix carries no cases, so scores cannot be formed; only the
  # weights and diagnostics are returned.
  is_cmat <- .is_cormat(x)
  if (is_cmat) {
    cli::cli_inform(
      c("i" = "{.arg x} is a correlation matrix; factor scores cannot be computed. Only factor weights and score diagnostics are returned. Enter raw data to get factor scores."),
      class = "efa_scores_needs_raw"
    )
  }

  method <- .match_arg_ci(method)
  checkmate::assert_matrix(Phi, null.ok = TRUE)

  if (!inherits(f, c("EFA", "matrix", "LOADINGS"))) {
    cli::cli_abort(
      c("{.arg f} must be an {.cls efa} object (from {.fn efa_fit}), a matrix, or an {.cls efa_loadings} object.",
        "i" = "From an {.cls efa_average} object, use the averaged loading matrix in
               {.code $loadings$average}, with the averaged factor correlations in
               {.code $Phi$average} for an oblique solution."),
      class = "efa_scores_bad_f"
    )
  }

  cor_method <- NULL
  rotation <- NULL
  orig_R <- NULL

  if (inherits(f, "EFA")) {

    # The factor intercorrelations come from the fitted solution, so an explicitly
    # supplied `Phi` cannot be honoured here; say so instead of discarding it silently.
    if (!is.null(Phi)) {
      cli::cli_warn(
        c("{.arg Phi} is ignored when {.arg f} is an {.cls efa} object.",
          "i" = "The factor intercorrelations of the fitted solution are used instead.",
          "i" = "Pass the loading matrix (for example {.code f$rot_loadings}) in {.arg f} to score against a different {.arg Phi}."),
        class = "efa_scores_phi_ignored"
      )
    }

    Phi <- f$Phi
    cor_method <- f$settings$cor_method
    rotation <- f$settings$rotation
    # The loadings and communalities that the weights must stay consistent with:
    # the rotated pattern for a rotated solution, the unrotated loadings otherwise.
    # h2 is the model communality diag(Lambda Phi Lambda') (rotation-invariant), the
    # correct uniqueness source for the Bartlett/Anderson weighting.
    if (!identical(rotation, "none")) {
      Lambda <- unclass(f$rot_loadings)
    } else {
      Lambda <- unclass(f$unrot_loadings)
    }
    h2 <- f$h2
    orig_R <- unclass(f$orig_R)

  } else {

    Lambda <- unclass(f)

    if (is.null(Phi)) {
      cli::cli_inform(
        c("i" = "{.arg Phi} was left {.code NULL} and loadings were entered directly in {.arg f}; assuming uncorrelated factors."),
        class = "efa_scores_phi_null"
      )
    }

  }

  if (!is.matrix(Lambda) || !is.numeric(Lambda) ||
      nrow(Lambda) < 1L || ncol(Lambda) < 1L || any(!is.finite(Lambda))) {
    cli::cli_abort(
      "{.arg f} must provide a non-empty numeric loading matrix with only finite values.",
      class = "efa_scores_bad_loadings"
    )
  }

  # Label the factor dimension (F1..Fm) when the loadings carry no usable factor
  # names, so the weights, diagnostics, scores, and determinacy table are
  # self-describing (the same factor naming the EFA print methods use). Duplicate
  # names are treated as unusable too: they would otherwise abort the determinacy
  # table below with an opaque base-R "duplicate row.names" error.
  factor_names <- colnames(Lambda)
  if (is.null(factor_names) || anyNA(factor_names) ||
      !all(nzchar(factor_names)) || anyDuplicated(factor_names) > 0L) {
    colnames(Lambda) <- paste0("F", seq_len(ncol(Lambda)))
  }

  if (!is_cmat) {
    raw_columns_ok <- if (is.data.frame(x)) {
      all(vapply(x, function(z) is.numeric(z) || is.logical(z), logical(1)))
    } else {
      is.numeric(x) || is.logical(x)
    }
    if (!raw_columns_ok) {
      cli::cli_abort(
        "Raw {.arg x} must contain only numeric or logical variables.",
        class = "efa_scores_bad_x"
      )
    }
    # Logical variables represent binary 0/1 observations and are valid inputs,
    # but matrix algebra and stats::cor() should see numeric storage explicitly.
    if (is.logical(x) || (is.data.frame(x) && any(vapply(x, is.logical, logical(1))))) {
      x <- as.matrix(x)
      storage.mode(x) <- "double"
    }
  }

  # Uncorrelated factors when no Phi is available (EFA with an orthogonal/unrotated
  # solution, or a directly supplied loading matrix without Phi).
  if (is.null(Phi)) {
    Phi <- diag(ncol(Lambda))
  }

  Phi <- .align_correlation_axis(
    Phi, n = ncol(Lambda), target_names = colnames(Lambda), arg = "Phi"
  )

  # A directly supplied loading matrix carries no stored communalities, so derive
  # them from the (now resolved) model: h2 = diag(Lambda Phi Lambda').
  if (!inherits(f, "EFA")) {
    h2 <- diag(Lambda %*% Phi %*% t(Lambda))
  }

  # Anderson-Rubin orthogonalisation is undefined with a single factor (there is
  # nothing to make uncorrelated), so guard it before any computation.
  if (method == "Anderson" && ncol(Lambda) == 1L) {
    cli::cli_abort(
      c("Anderson-Rubin scores are not defined for a single factor.",
        "i" = "There is nothing to orthogonalise with one factor; use {.code method = \"regression\"} or {.code method = \"Bartlett\"}."),
      class = "efa_scores_anderson_single"
    )
  }

  # Anderson-Rubin scores are defined for orthogonal factors: the weights force the
  # score intercorrelations to the identity, so for a model whose factors are
  # correlated they report uncorrelated scores for factors that are not. Warn when Phi
  # is not diagonal up to rounding -- an orthogonal or unrotated solution carries no
  # Phi and resolves to an exact identity above, so this fires only on a genuinely
  # oblique solution (or a hand-supplied Phi).
  if (method == "Anderson") {
    off <- abs(Phi[upper.tri(Phi)])
    if (any(off > sqrt(.Machine$double.eps), na.rm = TRUE)) {
      cli::cli_warn(
        c("Anderson-Rubin scores are defined for orthogonal factors, but the factors of {.arg f} are correlated (largest {.code |r|} = {round(max(off, na.rm = TRUE), 3)}).",
          "i" = "The scores are orthogonalised regardless, so they are uncorrelated even though the factors of the model are not.",
          "i" = "Use {.code method = \"tenBerge\"} to preserve the factor intercorrelations."),
        class = "efa_scores_anderson_oblique"
      )
    }
  }

  # Align raw `x` to the model variables. The weight matrix `W` (and the structure
  # products it is built from) combines variables by column position, but `W`, like
  # `Lambda`, carries the fitted model's variable names in the model's order. A raw
  # `x` whose columns are in a different order would otherwise be scored against the
  # wrong variables -- and in the loadings-matrix path also corrupt the `cor(x)` the
  # weights are derived from. When both sides carry usable, unique names, reorder `x`
  # to the model order by name (dropping any extra, non-model columns); a model
  # variable absent from `x` cannot be scored, so abort. Fall back to positional
  # matching when either side is unnamed or carries duplicate names.
  if (!is_cmat) {
    model_vars <- rownames(Lambda)
    x_names <- colnames(x)
    usable_model_vars <- !is.null(model_vars) && !anyNA(model_vars) &&
      all(nzchar(model_vars)) && anyDuplicated(model_vars) == 0L
    if (usable_model_vars && !is.null(x_names)) {
      if (anyNA(x_names) || !all(nzchar(x_names)) || anyDuplicated(x_names) > 0L) {
        cli::cli_abort(
          "The column names of {.arg x} must be complete, non-empty, and unique for name-based alignment.",
          class = "efa_scores_x_names"
        )
      }
      missing_vars <- setdiff(model_vars, x_names)
      if (length(missing_vars) > 0L) {
        cli::cli_abort(
          c("{.arg x} is missing {cli::qty(missing_vars)} model variable{?s}: {.val {missing_vars}}.",
            "i" = "Factor scores need every variable in the loading matrix of {.arg f}."),
          class = "efa_scores_missing_vars"
        )
      }
      x <- x[, model_vars, drop = FALSE]
    }
  } else {
    # A correlation matrix `x` describes the same model axis, and takes the checks the
    # scoring matrix takes, whether or not it ends up being that matrix: a fitted `f`
    # carries the correlations the model was estimated from, so `x` is not read for the
    # weights there, and left unchecked a matrix of the wrong size or of entirely
    # different variables returned the fitted solution's weights as though it had produced
    # them. Aligned as well as checked, so the branch below that does read it gets the
    # model's variable order, as the raw branch above does. Coerced first because `x` also
    # takes a data frame, which a correlation matrix is as readily given as.
    x <- .align_correlation_axis(as.matrix(x), n = nrow(Lambda),
                                 target_names = rownames(Lambda), arg = "x")
  }

  # A model variable that carries no usable spread cannot be scored: a constant column
  # divides by a zero standard deviation and becomes `NaN`, and because a score sums
  # over all variables, one such column destroys the scores of every case. The rule is
  # the same for every method -- a model variable that does not vary means the fitted
  # correlation structure does not apply to these data. Checked after the alignment
  # above, so only the variables actually used for scoring are examined, and before the
  # scoring correlation is resolved, where the same column would otherwise surface as an
  # opaque non-finite correlation.
  #
  # sd() is exactly 0 for a constant column, NaN for one carrying an infinity, and NA
  # for one with fewer than two observed values, so the causes are separated before they
  # are reported -- an infinite value is not an absent one -- the same split the
  # uncomputable-correlation diagnosis makes. A small but positive standard deviation is
  # ill-conditioned rather than undefined and is left to standardise as usual. Each test
  # yields FALSE rather than NA on a column it cannot judge, so `any()` cannot abort with
  # an unclassed base error in place of the classed condition.
  if (!is_cmat) {
    sds <- if (is.data.frame(x)) {
      vapply(x, stats::sd, numeric(1), na.rm = TRUE)
    } else {
      apply(x, 2L, stats::sd, na.rm = TRUE)
    }
    nonfinite <- is.nan(sds)
    unobserved <- is.na(sds) & !nonfinite
    constant <- !is.na(sds) & sds == 0
    unusable <- nonfinite | unobserved | constant
    if (any(unusable)) {
      # Name the offenders, falling back to their column positions as character labels
      # so that cli::qty() reads the number of them rather than a position value. Long
      # lists go through the shared cap, as in every other condition that enumerates
      # affected variables.
      nms <- colnames(x)
      if (is.null(nms)) nms <- as.character(seq_along(sds))
      n_bad <- sum(unusable)
      cap <- .cap_label_list(nms[unusable])
      why <- if (all(constant[unusable])) {
        "{cli::qty(n_bad)}{?It has/They have} no variance, so {?it carries/they carry} no information about any factor."
      } else if (all(nonfinite[unusable])) {
        "{cli::qty(n_bad)}{?It contains/They contain} infinite values, which combine with no factor."
      } else if (all(unobserved[unusable])) {
        "{cli::qty(n_bad)}{?It has/They have} fewer than two observed values, so {?its/their} spread is undefined."
      } else {
        "{cli::qty(n_bad)}{?It has/They have} no variance, infinite values, or fewer than two observed values."
      }
      cli::cli_abort(
        c("{cli::qty(n_bad)}Model variable{?s} {.val {cap$shown}}{cap$rest}{cli::qty(n_bad)} cannot be scored from {.arg x}.",
          "x" = why,
          "i" = "Correct the affected values, or drop {cli::qty(n_bad)}{?it/them} from the factor model."),
        class = "efa_scores_constant_var"
      )
    }
  }

  # Scoring correlation matrix R. A user-supplied `rho` wins; otherwise use the
  # matrix the EFA was fit on (f$orig_R), so the weights are consistent with the
  # loadings even for a non-Pearson fit; otherwise fall back to the correlation
  # of `x` (the matrix itself when `x` is already a correlation matrix).
  if (!is.null(rho)) {
    R <- unclass(rho)
  } else if (!is.null(orig_R)) {
    R <- orig_R
  } else if (is_cmat) {
    R <- unclass(as.matrix(x))
  } else {
    R <- stats::cor(x, use = "pairwise")
  }

  R_source <- if (!is.null(rho)) "rho" else if (!is.null(orig_R)) "f$orig_R" else "x"
  R <- .align_correlation_axis(
    R, n = nrow(Lambda), target_names = rownames(Lambda), arg = R_source
  )

  # A non-Pearson EFA scored on raw data with rho = NULL: the weights and
  # diagnostics correctly use the fitted correlation (f$orig_R), but the scores are
  # formed from the Pearson-standardised raw data, so realised score properties can
  # differ from the reported diagnostics. Warn (and point to `rho`) unless the fit
  # is Pearson, a matrix was supplied via `rho`, or `x` is a correlation matrix.
  if (!is.null(cor_method) && !identical(cor_method, "pearson") &&
      is.null(rho) && !is_cmat) {
    cli::cli_warn(
      c("{.arg f} was fit with {.code cor_method = {.val {cor_method}}}; the weights and diagnostics use that fitted correlation ({.code f$orig_R}).",
        "i" = "The factor scores are formed from the Pearson-standardised raw data, so their realised properties may differ from the reported diagnostics.",
        "i" = "Pass the {.val {cor_method}} correlation matrix (or another) via {.arg rho} to control the scoring correlation."),
      class = "efa_scores_cor_method"
    )
  }

  # psych's components path multiplies `x %*% W` without coercing; coerce a raw
  # data.frame up front (a no-op for the scale()-based methods).
  if (is.data.frame(x)) {
    x <- as.matrix(x)
  }

  W <- .factor_score_weights(Lambda, Phi, R, h2, method = method)
  S <- Lambda %*% Phi
  diag_out <- .factor_score_diagnostics(W, R, S)

  # Factor scores only from raw data. Component scores are the raw (uncentered)
  # data times the loadings; the other methods use the standardized data. A case with
  # a missing value on any model variable cannot be scored (no imputation is
  # performed), so its score row is NA -- count those cases and say so, rather than
  # returning NA rows the caller only discovers downstream.
  scores <- NULL
  n_scored <- NA_integer_
  if (!is_cmat) {
    scores <- if (method == "components") x %*% W else scale(x) %*% W
    n_na <- sum(!stats::complete.cases(scores))
    n_scored <- nrow(scores) - n_na
    if (n_na > 0L) {
      cli::cli_warn(
        c("{n_na} case{?s} {?has/have} a missing value on at least one model variable, so {?its/their} factor score{?s} {?is/are} {.code NA}.",
          "i" = "No imputation is performed: handle the missing values before scoring, or drop the affected cases."),
        class = "efa_scores_incomplete_cases"
      )
    }
  }

  determinacy <- data.frame(
    rho = diag_out$determinacy,
    rho2 = diag_out$determinacy^2,
    guttman = diag_out$guttman,
    row.names = colnames(Lambda)
  )

  output <- list(
    weights = W,
    scores = scores,
    r.scores = diag_out$r.scores,
    score_cor = diag_out$score_cor,
    determinacy = determinacy,
    settings = list(
      method = method,
      cor_method = cor_method,
      n_factors = ncol(Lambda),
      n_obs = if (is.null(scores)) NA_integer_ else nrow(scores),
      n_scored = n_scored
    )
  )

  class(output) <- "efa_scores"

  output

}

# Validate a factor- or variable-correlation matrix and, when both dimensions and
# the corresponding model axis carry unique names, align it to the model by name.
# A wholly unnamed matrix keeps the documented positional semantics; partially
# named or duplicate-named input cannot be aligned safely and is rejected.
.align_correlation_axis <- function(x, n, target_names, arg) {
  if (!is.matrix(x) || !is.numeric(x) || !identical(dim(x), c(n, n))) {
    cli::cli_abort(
      "{.arg {arg}} must be a numeric {n} by {n} correlation matrix.",
      class = "efa_scores_matrix_dim"
    )
  }
  if (any(!is.finite(x))) {
    cli::cli_abort(
      "{.arg {arg}} must contain only finite values.",
      class = "efa_scores_matrix_nonfinite"
    )
  }

  usable_target <- !is.null(target_names) && !anyNA(target_names) &&
    all(nzchar(target_names)) && anyDuplicated(target_names) == 0L
  if (usable_target) {
    rn <- rownames(x)
    cn <- colnames(x)
    wholly_unnamed <- is.null(rn) && is.null(cn)

    if (!wholly_unnamed) {
      usable_names <- !is.null(rn) && !is.null(cn) && identical(rn, cn) &&
        !anyNA(rn) && !anyNA(cn) && all(nzchar(rn)) && all(nzchar(cn)) &&
        anyDuplicated(rn) == 0L && anyDuplicated(cn) == 0L
      same_names <- usable_names && setequal(rn, target_names) &&
        setequal(cn, target_names)
      if (!same_names) {
        cli::cli_abort(
          c("The row and column names of {.arg {arg}} do not identify the model axis unambiguously.",
            "i" = "Use the same unique names as the corresponding loading matrix dimension, or remove both sets of names to match by position."),
          class = "efa_scores_matrix_names"
        )
      }
      x <- x[target_names, target_names, drop = FALSE]
    }
  }

  # Validate after name alignment. Requiring the two labelled axes to use the same
  # order above also prevents a numerically symmetric matrix with one mislabelled
  # axis from becoming invalid only after subsetting.
  tol <- sqrt(.Machine$double.eps)
  if (max(abs(x - t(x))) > tol * max(1, max(abs(x)))) {
    cli::cli_abort(
      "{.arg {arg}} must be symmetric.",
      class = "efa_scores_matrix_symmetric"
    )
  }
  # Canonicalize accepted round-off asymmetry so symmetric eigensolvers and general
  # matrix operations consume the same matrix on every BLAS/LAPACK implementation.
  x <- x / 2 + t(x) / 2
  range_tol <- 100 * .Machine$double.eps
  if (max(abs(diag(x) - 1)) > range_tol || any(abs(x) > 1 + range_tol)) {
    cli::cli_abort(
      "{.arg {arg}} must have a unit diagonal and entries between -1 and 1.",
      class = "efa_scores_matrix_correlation"
    )
  }
  diag(x) <- 1
  x[x > 1] <- 1
  x[x < -1] <- -1

  x
}


#' Print and format an efa_scores object
#'
#' `print()` shows a concise overview of an [efa_scores()] result: a header naming
#' the method and whether factor scores were computed, and the per-factor
#' determinacy table (determinacy, squared determinacy, and Guttman index).
#' `summary()` returns a `summary.efa_scores` object whose print method adds the
#' full factor-weight matrix, the score validity/univocality matrix, and the score
#' intercorrelations. `format()` assembles the same report and returns it as a
#' character vector; `print()` is `cat(format(x), sep = "\n")`. The lines follow
#' the active console theme, so they are plain when colours are disabled (for
#' example when captured into a file or stripped with [cli::ansi_strip()]).
#'
#' @param x,object An object of class `efa_scores`; for the `summary.efa_scores`
#'   methods, the object returned by `summary()`.
#' @param digits numeric. Number of decimal places for the printed tables.
#'   Default is 3.
#' @param ... Not used; for consistency with the generics.
#'
#' @returns `print()` and the print method for `summary.efa_scores` objects return
#'   their argument invisibly. `format()` returns a character vector with the
#'   report lines. `summary()` returns an object of class `summary.efa_scores`.
#'
#' @family factor scoring
#'
#' @export
#'
#' @method print efa_scores
#'
#' @examples
#' efa <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                estimator = "PAF", rotation = "oblimin")
#' fs <- efa_scores(test_models$baseline$cormat, f = efa)
#' fs
#' summary(fs)
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(fs))
#'
print.efa_scores <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_scores
#' @export
#' @method format efa_scores
format.efa_scores <- function(x, digits = 3, ...) {
  .render_efa_scores(x, view = "brief", digits = digits)
}

#' @rdname print.efa_scores
#' @export
#' @method summary efa_scores
summary.efa_scores <- function(object, digits = 3, ...) {
  structure(list(scores = object, opts = list(digits = digits)),
            class = "summary.efa_scores")
}

#' @rdname print.efa_scores
#' @export
#' @method print summary.efa_scores
print.summary.efa_scores <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_scores
#' @export
#' @method format summary.efa_scores
format.summary.efa_scores <- function(x, digits = x$opts$digits, ...) {
  .render_efa_scores(x$scores, view = "full", digits = digits)
}

# Render the factor-score report at the requested depth: "brief" (print) shows the
# header and the per-factor determinacy table; "full" (summary) additionally shows
# the factor-weight matrix, the score validity/univocality matrix, and the score
# intercorrelations. Assembled inside a single cli container so prose wraps to the
# console width; the aligned tables are emitted verbatim (no reflow).
.render_efa_scores <- function(x, view = c("brief", "full"), digits = 3) {
  view <- .match_arg_ci(view)
  full <- identical(view, "full")

  method <- x$settings$method
  title <- paste0("Factor scores (", method, ")")

  # Every table here holds bounded coefficients or correlations, so all four go through
  # the shared renderer and print in the package's number convention (leading zero
  # dropped), matching the coefficient tables of efa_reliability() and the loading
  # tables. The renderer also splits a table wider than the console into stacked column
  # blocks, which the weight matrix reaches on a wide solution.
  cli::cli_format_method({
    cli::cli_text("")
    cli::cli_rule(left = "{.strong {title}}")
    cli::cli_text("")

    if (is.null(x$scores)) {
      cli::cli_text("Weights and diagnostics only (correlation-matrix input; no scores).")
    } else {
      n_scored <- x$settings$n_scored
      n_obs <- x$settings$n_obs
      n_fac <- x$settings$n_factors
      # The pointer names a field of the printed object, which is what `print()` was
      # handed; `summary()` wraps that object in a field of the same name, so the
      # pointer would resolve to the wrapper there and is left off the fuller report.
      # Written out twice rather than interpolated, because a `{.code ...}` carried in
      # through a substitution is emitted as literal text instead of being styled.
      if (full) {
        cli::cli_text("Scored {n_scored} of {n_obs} observation{?s} on {n_fac} factor{?s}.")
      } else {
        cli::cli_text("Scored {n_scored} of {n_obs} observation{?s} on {n_fac} factor{?s} (see {.code $scores}).")
      }
    }

    cli::cli_text("")
    cli::cli_rule(left = "{.strong Score determinacy}")
    cli::cli_text("")
    .efa_emit_lines(.efa_corr_lines(as.matrix(x$determinacy), digits = digits))

    if (full) {
      cli::cli_text("")
      cli::cli_rule(left = "{.strong Factor weights}")
      cli::cli_text("")
      .efa_emit_lines(.efa_corr_lines(x$weights, digits = digits))

      cli::cli_text("")
      cli::cli_rule(left = "{.strong Score validity and univocality}")
      cli::cli_text("")
      cli::cli_text("Diagonal: validity (score-factor correlation). Off-diagonal: univocality.")
      cli::cli_text("")
      .efa_emit_lines(.efa_corr_lines(x$score_cor, digits = digits))

      cli::cli_text("")
      cli::cli_rule(left = "{.strong Score intercorrelations}")
      cli::cli_text("")
      .efa_emit_lines(.efa_corr_lines(x$r.scores, digits = digits))
    }
  })
}


# Native factor-score weight engine ------------------------------------------
#
# Computes the p x m factor-score weight matrix W (standardized scores are
# scale(X) %*% W) natively, without psych::factor.scores(). With the structure
# matrix S = Lambda %*% Phi and the model uniquenesses Psi = diag(1 - h2), the
# supported methods are:
#   Thurstone / regression : W = R^-1 S                            (Thurstone, 1935)
#   Bartlett               : W = Psi^-1 Lambda (Lambda' Psi^-1 Lambda)^-1
#                                                                   (Bartlett, 1937)
#   Anderson (orthogonal)  : W = Psi^-1 Lambda
#                                (Lambda' Psi^-1 R Psi^-1 Lambda)^-1/2
#                                                          (Anderson & Rubin, 1956)
#   tenBerge               : L = Lambda Phi^1/2;
#                            W = R^-1/2 [R^-1/2 L (L' R^-1 L)^-1/2] Phi^1/2
#                     (ten Berge, Krijnen, Wansbeek & Shapiro, 1999; Grice, 2001)
#   Harman                 : M = Lambda (Lambda Phi)', diag(M) = 1; W = M^-1 Lambda
#                                                                    (Harman, 1976)
#   components             : W = Lambda
# For orthogonal regression the diagonal of Lambda' R^-1 Lambda gives the
# factor-score validity coefficients rho^2, i.e. each factor's squared multiple
# correlation with the observed variables (Grice, 2001).

# Symmetric matrix square root A^1/2 = V diag(sqrt(lambda)) V' via the eigen-
# decomposition; negative eigenvalues are clamped to .Machine$double.eps (as in
# psych's matSqrt). `nrow=` keeps diag() a matrix for a single eigenvalue.
.mat_sqrt <- function(x) {
  e <- eigen(x, symmetric = TRUE)
  ev <- e$values
  ev[ev < 0] <- .Machine$double.eps
  e$vectors %*% diag(sqrt(ev), nrow = length(ev)) %*% t(e$vectors)
}

# Symmetric inverse matrix square root A^-1/2; eigenvalues below
# .Machine$double.eps are clamped to 100 * .Machine$double.eps (as in psych's
# invMatSqrt). `nrow=` keeps diag() a matrix for a single eigenvalue.
.inv_mat_sqrt <- function(x) {
  e <- eigen(x, symmetric = TRUE)
  ev <- e$values
  ev[ev < .Machine$double.eps] <- 100 * .Machine$double.eps
  e$vectors %*% diag(1 / sqrt(ev), nrow = length(ev)) %*% t(e$vectors)
}

# Moore-Penrose pseudo-inverse via the SVD, dropping singular values at or below
# tol * (largest singular value); fallback for a singular correlation or Gram
# matrix (mirrors psych's Pinv).
.pinv <- function(X, tol = sqrt(.Machine$double.eps)) {
  s <- svd(X)
  pos <- s$d > max(tol * s$d[1], 0)
  if (all(pos)) {
    s$v %*% (t(s$u) / s$d)
  } else {
    s$v[, pos, drop = FALSE] %*% (t(s$u[, pos, drop = FALSE]) / s$d[pos])
  }
}

# Factor-score weight matrix W (p x m) for the requested method; see the block
# comment above for the formulas and references. `h2` must be the model-
# reproduced communalities diag(Lambda Phi Lambda') (not rowSums(Lambda^2),
# which differ for oblique solutions) so the Bartlett/Anderson uniquenesses are
# correct. Anderson-Rubin scores are defined for orthogonal factors only.
.factor_score_weights <- function(Lambda, Phi, R, h2,
                                  method = c("Thurstone", "regression", "tenBerge",
                                             "Anderson", "Bartlett", "Harman",
                                             "components")) {

  method <- .match_arg_ci(method)
  if (method == "regression") method <- "Thurstone"

  S <- Lambda %*% Phi

  # solve(A, B) with a Moore-Penrose fallback when A is singular.
  solve_or_pinv <- function(A, B) {
    tryCatch(solve(A, B), error = function(e) .pinv(A) %*% B)
  }

  # The model uniquenesses u2 = 1 - h2 that the Bartlett and Anderson methods weight
  # by. Both need Psi^-1 Lambda = diag(1 / u2) %*% Lambda, which is the row scaling
  # `Lambda / u2`, so the dense p x p diagonal matrix is never formed. A non-positive
  # uniqueness makes the weighting undefined; abort instead of dividing by zero or
  # returning finite-looking output from an invalid negative weight. A very small
  # positive uniqueness is defined but ill-conditioned, so retain the existing warning
  # for that distinct case.
  uniquenesses <- function() {
    u2 <- 1 - h2
    nonpositive <- u2 <= 0
    if (any(nonpositive)) {
      bad <- rownames(Lambda)[nonpositive]
      if (is.null(bad)) bad <- which(nonpositive)
      cli::cli_abort(
        c("Non-positive uniqueness values make the {method} factor-score weights undefined (affected variables: {.val {bad}}).",
          "i" = "Use a scoring method that does not invert uniquenesses, or revise the factor model."),
        class = "efa_scores_nonpositive_uniqueness"
      )
    }
    near_zero <- u2 < .Machine$double.eps
    if (any(near_zero)) {
      bad <- rownames(Lambda)[near_zero]
      if (is.null(bad)) bad <- which(near_zero)
      cli::cli_warn(
        c("Near-zero uniqueness values were detected for {.val {bad}}.",
          "i" = "The affected {method} factor-score weights are unstable."),
        class = "efa_scores_heywood"
      )
    }
    u2
  }

  W <- switch(
    method,
    Thurstone = solve_or_pinv(R, S),
    Bartlett = {
      Lu <- Lambda / uniquenesses()                 # Psi^-1 Lambda
      Lu %*% solve_or_pinv(crossprod(Lu, Lambda), diag(ncol(Lambda)))
    },
    Anderson = {
      Lu <- Lambda / uniquenesses()                 # Psi^-1 Lambda
      Lu %*% .inv_mat_sqrt(crossprod(Lu, R %*% Lu))
    },
    tenBerge = {
      phi_sqrt <- .mat_sqrt(Phi)
      L <- Lambda %*% phi_sqrt
      r_inv_sqrt <- .inv_mat_sqrt(R)
      C <- r_inv_sqrt %*% L %*% .inv_mat_sqrt(crossprod(L, solve_or_pinv(R, L)))
      r_inv_sqrt %*% C %*% phi_sqrt
    },
    Harman = {
      M <- Lambda %*% t(S)
      diag(M) <- 1
      solve_or_pinv(M, Lambda)
    },
    components = Lambda
  )

  dimnames(W) <- dimnames(Lambda)
  W

}


# Score-quality diagnostics for a factor-score weight matrix W (from
# .factor_score_weights). With the standardized scores Fhat = Z %*% W (Z the standardized
# data, cov(Z) = R) and the structure matrix S = Lambda %*% Phi, C = W' R W is the
# covariance of the score estimates. Reports:
#   r.scores    : cov2cor(C), the score intercorrelations; for a well-behaved solution
#                 these reproduce the factor correlations Phi (exactly for ten Berge, and
#                 the identity for Anderson-Rubin).
#   score_cor   : the full score-factor correlation matrix diag(sd.in) W' S,
#                 sd.in = 1/sqrt(diag(C)); its diagonal is each score's determinacy /
#                 validity rho, its off-diagonals the univocality (the correlation of a
#                 score with the *other* factors).
#   determinacy : the diagonal of score_cor, the correlation between each factor and its
#                 estimated score for the *returned* (method-specific) weights, capped at
#                 1 with non-positive values set to NA (psych's convention).
#   guttman     : Guttman's (1955) indeterminacy index 2 rho^2 - 1, the minimum
#                 correlation between two equally valid sets of factor scores.
# The factors carry unit variance (diag(Phi) == 1), so Phi does not enter the score-
# factor correlations and is not needed here.
# References: Guttman (1955); ten Berge, Krijnen, Wansbeek & Shapiro (1999); Grice (2001).
.factor_score_diagnostics <- function(W, R, S) {

  C <- crossprod(W, R %*% W)                 # t(W) %*% R %*% W  (m x m)
  r.scores <- stats::cov2cor(C)

  sd.in <- 1 / sqrt(diag(C))
  # corr(Fhat_i, F_j): the factors are standardised (unit variance), so only the score
  # standard deviations enter the score-factor correlations.
  score_cor <- diag(sd.in, nrow = length(sd.in)) %*% crossprod(W, S)
  dimnames(score_cor) <- list(colnames(W), colnames(W))
  dimnames(r.scores)  <- dimnames(score_cor)

  # Determinacy / validity: the diagonal. Following psych's convention, cap at 1 and set
  # any non-positive value to NA (an ill-conditioned score whose "correlation" leaves
  # [0, 1]); a non-finite entry (a NaN reachable only from a non-PD R) is treated as NA
  # too, so the reported determinacy and Guttman index never leak a NaN.
  determinacy <- diag(score_cor)
  determinacy[determinacy > 1] <- 1
  determinacy[!is.finite(determinacy) | determinacy <= 0] <- NA
  names(determinacy) <- colnames(W)

  guttman <- 2 * determinacy^2 - 1

  list(r.scores = r.scores, score_cor = score_cor,
       determinacy = determinacy, guttman = guttman)

}

# Fixtures: oblique and orthogonal EFA fits on a correlation matrix, plus a
# raw-data single-factor fit for the scores path.
efa_ob <- suppressMessages(suppressWarnings(
  EFA(test_models$baseline$cormat, n_factors = 3, N = 500, type = "EFAtools",
      method = "PAF", rotation = "oblimin")))
efa_or <- suppressMessages(suppressWarnings(
  EFA(test_models$baseline$cormat, n_factors = 3, N = 500, type = "EFAtools",
      method = "PAF", rotation = "none")))
efa_grips <- suppressMessages(EFA(GRiPS_raw, n_factors = 1, type = "EFAtools",
                                  method = "PAF"))

fs_ob_cor  <- suppressMessages(efa_scores(test_models$baseline$cormat, f = efa_ob))
fs_grips   <- suppressMessages(efa_scores(GRiPS_raw, f = efa_grips, method = "Bartlett"))


test_that("output has the expected class, names, and shapes", {
  expect_s3_class(fs_ob_cor, "efa_scores")
  expect_s3_class(fs_grips, "efa_scores")

  expect_named(fs_ob_cor, c("weights", "scores", "r.scores", "score_cor",
                            "determinacy", "settings"))
  expect_named(fs_grips, c("weights", "scores", "r.scores", "score_cor",
                           "determinacy", "settings"))

  # Weights keep the variable rownames and carry factor labels; correlation input
  # returns no scores.
  L <- unclass(efa_ob$rot_loadings)
  expect_equal(dim(fs_ob_cor$weights), dim(L))
  expect_identical(rownames(fs_ob_cor$weights), rownames(L))
  expect_identical(colnames(fs_ob_cor$weights), c("F1", "F2", "F3"))
  expect_null(fs_ob_cor$scores)

  # Raw input returns one score row per observation, one column per factor.
  expect_equal(dim(fs_grips$scores), c(nrow(GRiPS_raw), 1L))
  expect_false(anyNA(fs_grips$scores))

  # Diagnostics matrices are m x m and share the factor names.
  m <- ncol(L)
  expect_equal(dim(fs_ob_cor$r.scores), c(m, m))
  expect_equal(dim(fs_ob_cor$score_cor), c(m, m))
  expect_identical(dimnames(fs_ob_cor$r.scores), dimnames(fs_ob_cor$score_cor))
})

test_that("the determinacy table is consistent with score_cor", {
  expect_s3_class(fs_ob_cor$determinacy, "data.frame")
  expect_named(fs_ob_cor$determinacy, c("rho", "rho2", "guttman"))
  expect_identical(rownames(fs_ob_cor$determinacy), colnames(fs_ob_cor$weights))

  # The determinacy is the score_cor diagonal; rho2 and guttman are its transforms.
  expect_equal(fs_ob_cor$determinacy$rho, diag(fs_ob_cor$score_cor),
               ignore_attr = TRUE)
  expect_equal(fs_ob_cor$determinacy$rho2, fs_ob_cor$determinacy$rho^2)
  expect_equal(fs_ob_cor$determinacy$guttman, 2 * fs_ob_cor$determinacy$rho^2 - 1)
})

test_that("settings record the resolved method and dimensions", {
  expect_equal(fs_grips$settings$method, "Bartlett")
  expect_equal(fs_ob_cor$settings$method, "regression")   # the default
  expect_equal(fs_ob_cor$settings$n_factors, 3L)
  expect_true(is.na(fs_ob_cor$settings$n_obs))            # correlation input
  expect_true(is.na(fs_ob_cor$settings$n_scored))
  # n_obs counts the supplied rows, n_scored the rows that could be scored; with
  # complete data the two agree.
  expect_equal(fs_grips$settings$n_obs, nrow(GRiPS_raw))
  expect_equal(fs_grips$settings$n_scored, nrow(GRiPS_raw))
})

test_that("a constant model variable in the scoring data is an error", {
  x_const <- as.matrix(GRiPS_raw)
  x_const[, 2] <- 3
  expect_error(
    suppressMessages(efa_scores(x_const, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_constant_var")

  # A data.frame takes the same route, and two constant variables render the
  # plural branch of the message.
  df_const <- as.data.frame(GRiPS_raw)
  df_const[[1]] <- 2
  df_const[[3]] <- 5
  expect_error(
    suppressMessages(efa_scores(df_const, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_constant_var")

  # A model variable that is entirely missing, or that carries an infinite value, has
  # no usable spread either; each renders its own cause in the same condition.
  x_all_na <- as.matrix(GRiPS_raw)
  x_all_na[, 3] <- NA_real_
  expect_error(
    suppressMessages(efa_scores(x_all_na, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_constant_var")

  x_inf <- as.matrix(GRiPS_raw)
  x_inf[1, 4] <- Inf
  expect_error(
    suppressMessages(efa_scores(x_inf, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_constant_var")

  # Unnamed input reports the column position instead of a name.
  x_unnamed <- unname(as.matrix(GRiPS_raw))
  x_unnamed[, 2] <- 3
  expect_error(
    suppressMessages(efa_scores(x_unnamed, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_constant_var")

  # A constant column that is not a model variable is dropped by name alignment
  # and must not trip the guard.
  x_extra <- cbind(as.matrix(GRiPS_raw), constant_extra = 1)
  expect_no_error(
    suppressMessages(efa_scores(x_extra, f = efa_grips, method = "Bartlett")))
})

test_that("cases with a missing model variable are counted and reported", {
  x_na <- as.matrix(GRiPS_raw)
  x_na[1, 1] <- NA

  expect_warning(
    fs_one <- suppressMessages(efa_scores(x_na, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_incomplete_cases")

  # n_obs stays the number of supplied rows; only n_scored drops.
  expect_equal(fs_one$settings$n_obs, nrow(GRiPS_raw))
  expect_equal(fs_one$settings$n_scored, nrow(GRiPS_raw) - 1L)
  expect_true(is.na(fs_one$scores[1, 1]))
  expect_false(anyNA(fs_one$scores[-1, , drop = FALSE]))

  # The plural branch of the message is a separate rendering.
  x_na[5, 2] <- NA
  x_na[9, 3] <- NA
  expect_warning(
    fs_many <- suppressMessages(efa_scores(x_na, f = efa_grips, method = "Bartlett")),
    class = "efa_scores_incomplete_cases")
  expect_equal(fs_many$settings$n_scored, nrow(GRiPS_raw) - 3L)
})

test_that("Phi supplied alongside an efa object is ignored with a warning", {
  expect_warning(
    fs <- suppressMessages(efa_scores(test_models$baseline$cormat, f = efa_ob,
                                      Phi = diag(3))),
    class = "efa_scores_phi_ignored")
  expect_equal(fs$weights, fs_ob_cor$weights)
})

test_that("classed conditions fire as expected", {
  expect_error(efa_scores(1:5), class = "efa_input_not_matrix")
  expect_error(efa_scores(GRiPS_raw, f = 1:5), class = "efa_scores_bad_f")
  expect_message(efa_scores(test_models$baseline$cormat, f = efa_ob),
                 class = "efa_scores_needs_raw")
  # Raw input isolates the phi_null message (a correlation matrix would also emit
  # needs_raw).
  expect_message(efa_scores(GRiPS_raw, f = unclass(efa_grips$unrot_loadings)),
                 class = "efa_scores_phi_null")

  # Anderson-Rubin is undefined for a single factor.
  expect_error(efa_scores(GRiPS_raw, f = efa_grips, method = "Anderson"),
               class = "efa_scores_anderson_single")

  # Anderson-Rubin forces uncorrelated scores, so it is inappropriate for an oblique
  # solution: warn there, but not for an orthogonal one (whose Phi is the identity).
  expect_warning(
    suppressMessages(efa_scores(test_models$baseline$cormat, f = efa_ob,
                                method = "Anderson")),
    class = "efa_scores_anderson_oblique")
  expect_no_warning(
    suppressMessages(efa_scores(test_models$baseline$cormat, f = efa_or,
                                method = "Anderson")))
})

test_that("a correlation matrix in `x` is checked against the model it is scored with", {
  cm <- test_models$baseline$cormat

  # A fitted solution carries the correlations it was estimated from, so `x` is not read
  # for the weights. Unchecked, a matrix of any size and of entirely different variables
  # returned the fitted solution's weights as though it had produced them.
  five <- matrix(.2, 5, 5)
  diag(five) <- 1
  dimnames(five) <- list(paste0("X", 1:5), paste0("X", 1:5))
  expect_error(suppressMessages(efa_scores(five, f = efa_ob)),
               class = "efa_scores_matrix_dim")

  # The right size, but the variables of some other model.
  foreign <- cm
  dimnames(foreign) <- list(paste0("Q", 1:18), paste0("Q", 1:18))
  expect_error(suppressMessages(efa_scores(foreign, f = efa_ob)),
               class = "efa_scores_matrix_names")

  # Named on one axis only, which identifies neither by name nor by position: refused
  # here as it is for `rho`, rather than passed over as it was when `x` went unchecked.
  half <- cm
  colnames(half) <- NULL
  expect_error(suppressMessages(efa_scores(half, f = efa_ob)),
               class = "efa_scores_matrix_names")

  # Where `f` is a loading matrix the solution carries no correlations, so `x` is the
  # scoring matrix. Named for the model and aligned to it there, so a permuted matrix is
  # reordered rather than scored against the wrong variables. (With a fitted `f` the same
  # pair agrees whatever the order, `x` not being read, so the model axis is what this
  # can be asserted on.)
  L <- unclass(efa_ob$rot_loadings)
  p <- c(4:18, 1:3)
  expect_equal(suppressMessages(efa_scores(cm[p, p], f = L, Phi = efa_ob$Phi)),
               suppressMessages(efa_scores(cm, f = L, Phi = efa_ob$Phi)))

  # The inputs the check must not turn away: a data frame, which `x` is documented to
  # take, and a wholly unnamed matrix, which is matched by position.
  expect_equal(suppressMessages(efa_scores(as.data.frame(cm), f = efa_ob)),
               suppressMessages(efa_scores(cm, f = efa_ob)))
  unnamed <- cm
  dimnames(unnamed) <- NULL
  expect_s3_class(suppressMessages(efa_scores(unnamed, f = efa_ob)), "efa_scores")

  # The wrapper forwards to the same check, so it is refused there too.
  expect_error(suppressMessages(FACTOR_SCORES(five, f = efa_ob)),
               class = "efa_scores_matrix_dim")
})

test_that("loadings with missing or duplicate factor names are relabelled F1..Fm", {
  L <- unclass(efa_or$unrot_loadings)          # NULL colnames
  colnames(L) <- c("F1", "F1", "F2")           # duplicate factor labels

  # Must not abort with the opaque base-R "duplicate row.names" error.
  fs <- suppressMessages(efa_scores(test_models$baseline$cormat, f = L, Phi = diag(3)))
  expect_identical(colnames(fs$weights), c("F1", "F2", "F3"))
  expect_identical(rownames(fs$determinacy), c("F1", "F2", "F3"))
})

test_that("a non-Pearson EFA with rho = NULL on raw data warns, rho silences it", {
  skip_on_cran()

  efa_sp <- suppressMessages(EFA(GRiPS_raw, n_factors = 1, type = "EFAtools",
                                 method = "PAF", cor_method = "spearman"))
  expect_warning(efa_scores(GRiPS_raw, f = efa_sp, method = "regression"),
                 class = "efa_scores_cor_method")

  rho_sp <- stats::cor(GRiPS_raw, method = "spearman")
  expect_no_warning(efa_scores(GRiPS_raw, f = efa_sp, rho = rho_sp,
                               method = "regression"))
})

test_that("raw-data columns are aligned to the model by name, not position", {
  # Reversing the column order must not change the scores or weights: efa_scores
  # aligns x to the variable names carried by the model, not by column position.
  rev_idx <- rev(seq_len(ncol(GRiPS_raw)))

  fs_ordered  <- suppressMessages(
    efa_scores(GRiPS_raw, f = efa_grips, method = "Bartlett"))
  fs_shuffled <- suppressMessages(
    efa_scores(GRiPS_raw[, rev_idx], f = efa_grips, method = "Bartlett"))

  expect_equal(fs_shuffled$scores, fs_ordered$scores)
  expect_equal(fs_shuffled$weights, fs_ordered$weights)

  # Guard against a positional regression: scoring the shuffled columns by position
  # (the pre-alignment behaviour) gives materially different scores.
  positional <- scale(as.matrix(GRiPS_raw[, rev_idx])) %*% fs_ordered$weights
  expect_gt(max(abs(positional - fs_shuffled$scores)), 0.1)
})

test_that("column name alignment handles missing, extra, and unnamed columns", {
  # A model variable missing from x cannot be scored.
  expect_error(
    suppressMessages(efa_scores(GRiPS_raw[, -1], f = efa_grips, method = "Bartlett")),
    class = "efa_scores_missing_vars")

  fs_plain <- suppressMessages(
    efa_scores(GRiPS_raw, f = efa_grips, method = "Bartlett"))

  # Extra, non-model columns are ignored (dropped by name alignment).
  x_extra <- cbind(as.matrix(GRiPS_raw), extra_col = 0)
  fs_extra <- suppressMessages(
    efa_scores(x_extra, f = efa_grips, method = "Bartlett"))
  expect_equal(fs_extra$scores, fs_plain$scores)
  expect_identical(rownames(fs_extra$weights), rownames(fs_plain$weights))

  # Unnamed data fall back to positional matching (order already matches the model).
  x_unnamed <- unname(as.matrix(GRiPS_raw))
  fs_unnamed <- suppressMessages(
    efa_scores(x_unnamed, f = efa_grips, method = "Bartlett"))
  expect_equal(fs_unnamed$scores, fs_plain$scores, ignore_attr = TRUE)
})

test_that("alignment holds for the loadings-matrix path and multiple factors", {
  # A bare loading matrix with more than one factor exercises the path where the
  # scoring correlation is cor(x), recomputed from x: the reorder must precede it
  # so the weights stay consistent. Two factors also check that the score columns
  # stay matched under a column reorder. Both the standardized (Bartlett) and the
  # uncentered (components) score multiplications are covered.
  efa2 <- suppressMessages(suppressWarnings(
    EFA(GRiPS_raw, n_factors = 2, type = "EFAtools", method = "PAF",
        rotation = "oblimin")))
  L   <- unclass(efa2$rot_loadings)
  Phi <- efa2$Phi
  rev_idx <- rev(seq_len(ncol(GRiPS_raw)))

  for (m in c("Bartlett", "components")) {
    fs_ordered <- suppressMessages(suppressWarnings(
      efa_scores(GRiPS_raw, f = L, Phi = Phi, method = m)))
    fs_shuffled <- suppressMessages(suppressWarnings(
      efa_scores(GRiPS_raw[, rev_idx], f = L, Phi = Phi, method = m)))

    expect_equal(fs_shuffled$scores, fs_ordered$scores, info = m)
    expect_equal(fs_shuffled$weights, fs_ordered$weights, info = m)
    expect_equal(dim(fs_ordered$scores), c(nrow(GRiPS_raw), 2L), info = m)
  }
})

test_that("named scoring and factor correlations are aligned to the loadings", {
  L <- unclass(efa_ob$rot_loadings)
  R <- test_models$baseline$cormat
  Phi <- efa_ob$Phi

  base <- suppressMessages(efa_scores(R, f = L, Phi = Phi))

  var_perm <- rev(seq_len(nrow(R)))
  R_perm <- R[var_perm, var_perm]
  from_permuted_R <- suppressMessages(efa_scores(R_perm, f = L, Phi = Phi))
  expect_equal(from_permuted_R$weights, base$weights)
  expect_equal(from_permuted_R$score_cor, base$score_cor)

  # A correlation matrix must use the same variable order on its two labelled axes.
  col_perm <- c(seq.int(2L, ncol(R)), 1L)
  R_axis_perm <- R[var_perm, col_perm]
  expect_error(
    suppressMessages(efa_scores(R, f = L, Phi = Phi, rho = R_axis_perm)),
    class = "efa_scores_matrix_names"
  )

  fac_perm <- c(3L, 1L, 2L)
  Phi_perm <- Phi[fac_perm, fac_perm]
  from_permuted_Phi <- suppressMessages(efa_scores(R, f = L, Phi = Phi_perm))
  expect_equal(from_permuted_Phi$weights, base$weights)
  expect_equal(from_permuted_Phi$score_cor, base$score_cor)

  # rho follows the same name alignment when raw observations are scored.
  R_grips <- stats::cor(GRiPS_raw)
  rho_base <- suppressMessages(
    efa_scores(GRiPS_raw, f = efa_grips, rho = R_grips, method = "regression"))
  grips_perm <- rev(seq_len(nrow(R_grips)))
  rho_perm <- suppressMessages(
    efa_scores(GRiPS_raw, f = efa_grips,
               rho = R_grips[grips_perm, grips_perm],
               method = "regression"))
  expect_equal(rho_perm$weights, rho_base$weights)
  expect_equal(rho_perm$scores, rho_base$scores)
})

test_that("invalid scoring-correlation matrices fail before matrix algebra", {
  L <- unclass(efa_ob$rot_loadings)
  R <- test_models$baseline$cormat

  expect_error(
    suppressMessages(efa_scores(R, f = L, Phi = efa_ob$Phi,
                                rho = diag(nrow(R) - 1L))),
    class = "efa_scores_matrix_dim")

  R_asym <- R
  R_asym[1, 2] <- R_asym[1, 2] + 0.1
  expect_error(
    suppressMessages(efa_scores(R, f = L, Phi = efa_ob$Phi, rho = R_asym)),
    class = "efa_scores_matrix_symmetric")

  R_bad_names <- R
  dimnames(R_bad_names) <- list(paste0("bad", seq_len(nrow(R))),
                                paste0("bad", seq_len(ncol(R))))
  expect_error(
    suppressMessages(efa_scores(R_bad_names, f = L, Phi = efa_ob$Phi)),
    class = "efa_scores_matrix_names")

  # Matching name sets do not excuse inconsistent labels on the two axes.
  R_bad_axes <- R
  colnames(R_bad_axes) <- rev(colnames(R_bad_axes))
  expect_error(
    suppressMessages(efa_scores(R_bad_axes, f = L, Phi = efa_ob$Phi)),
    class = "efa_scores_matrix_names")
})

test_that("score labels and uniqueness inputs fail or fall back cleanly", {
  R2 <- diag(2)
  dimnames(R2) <- list(c("v1", "v2"), c("v1", "v2"))

  # Missing factor labels are not valid data-frame row names; replace the whole
  # factor axis with deterministic labels before diagnostics are constructed.
  L_bad_factor_names <- matrix(c(.6, .2, .1, .5), 2, 2,
                               dimnames = list(c("v1", "v2"), c("F1", NA)))
  relabelled <- suppressMessages(efa_scores(R2, f = L_bad_factor_names))
  expect_identical(colnames(relabelled$weights), c("F1", "F2"))

  # Named raw data can only be aligned when its labels identify columns uniquely.
  x_bad_names <- GRiPS_raw
  colnames(x_bad_names)[2] <- colnames(x_bad_names)[1]
  expect_error(
    suppressMessages(efa_scores(x_bad_names, f = efa_grips)),
    class = "efa_scores_x_names"
  )

  # Bartlett/Anderson uniqueness weighting is undefined at exactly zero; do not
  # continue into 1 / 0 and return non-finite weights or an eigensolver error.
  L_zero_u2 <- matrix(c(1, .5), 2, 1,
                      dimnames = list(c("v1", "v2"), "F1"))
  expect_error(
    suppressMessages(efa_scores(R2, f = L_zero_u2, method = "Bartlett")),
    class = "efa_scores_nonpositive_uniqueness"
  )

  expect_error(
    suppressMessages(efa_scores(R2, f = matrix("x", 2, 1))),
    class = "efa_scores_bad_loadings"
  )
  L_nonfinite <- matrix(c(.5, Inf), 2, 1)
  expect_error(
    suppressMessages(efa_scores(R2, f = L_nonfinite)),
    class = "efa_scores_bad_loadings"
  )
  expect_error(
    suppressMessages(efa_scores(matrix("x", 4, 2), f = L_zero_u2)),
    class = "efa_scores_bad_x"
  )
})

test_that("accepted correlation round-off is canonicalized", {
  R_near <- diag(3)
  R_near[1, 2] <- .2 + 1e-10
  R_near[2, 1] <- .2
  dimnames(R_near) <- list(letters[1:3], letters[1:3])
  out <- .align_correlation_axis(R_near, 3L, letters[1:3], "rho")
  expect_identical(out, t(out))
  expect_identical(unname(diag(out)), rep(1, 3))

  R_outside <- R_near
  R_outside[1, 2] <- R_outside[2, 1] <- 1 + 1e-12
  expect_error(
    .align_correlation_axis(R_outside, 3L, letters[1:3], "rho"),
    class = "efa_scores_matrix_correlation"
  )
})

test_that("component scores stay raw-scale while their diagnostics are standardized", {
  # method = "components" multiplies the raw, uncentered data by the loadings, but
  # r.scores / score_cor / determinacy are computed from W' R W, i.e. they describe
  # scale(x) %*% W. With unequal item variances the two diverge, which is what the
  # help page states; every other method returns standardized scores, where they agree.
  withr::local_seed(42)
  R <- test_models$baseline$cormat
  L <- unclass(efa_ob$rot_loadings)
  p <- ncol(R)
  x <- matrix(stats::rnorm(300 * p), 300, p) %*% chol(R)
  x <- sweep(x, 2L, seq(1, 10, length.out = p), "*")  # unequal variances
  x <- x + 10                                         # and non-zero means
  colnames(x) <- rownames(L)

  fs <- suppressMessages(suppressWarnings(
    efa_scores(x, f = L, Phi = efa_ob$Phi, method = "components")))

  expect_equal(fs$scores, x %*% fs$weights, ignore_attr = TRUE)
  # Uncentered. Taken on the absolute mean because a rotated factor carries no fixed
  # sign, so a column of the solution may be negated on another platform.
  expect_gt(min(abs(colMeans(fs$scores))), 1)
  expect_equal(unname(fs$r.scores),
               unname(stats::cor(scale(x) %*% fs$weights)), tolerance = 1e-8)
  expect_gt(max(abs(fs$r.scores - stats::cor(fs$scores))), 0.01)

  fs_b <- suppressMessages(suppressWarnings(
    efa_scores(x, f = L, Phi = efa_ob$Phi, method = "Bartlett")))
  expect_equal(unname(fs_b$r.scores), unname(stats::cor(fs_b$scores)),
               tolerance = 1e-8)
})

test_that("print and summary output are stable", {
  skip_on_cran()
  testthat::local_reproducible_output()

  # The printed values are eigen- and inverse-derived, so their last digits can drift
  # across BLAS implementations; scrub_num pins the layout, headers, and wording instead.

  # Brief print, raw-data scores.
  expect_snapshot(print(fs_grips), transform = scrub_num)

  # Brief print, correlation input (weights only).
  expect_snapshot(print(fs_ob_cor), transform = scrub_num)

  # Full summary: weight matrix, validity/univocality, and score intercorrelations.
  expect_snapshot(print(summary(fs_ob_cor)), transform = scrub_num)
})


# --- FACTOR_SCORES delegation: byte-identical regression against psych --------
#
# The rewritten FACTOR_SCORES() delegates to efa_scores(). Its weights/scores/
# r.scores must reproduce the original psych::factor.scores() pass-through
# bit-for-bit on the non-Heywood fixtures; only the R2 slot changes, now the
# corrected native determinacy^2 (which for oblique fits diverges from psych's
# rowSums(pattern^2) approximation but equals the textbook diag(S' R^-1 S)).
test_that("FACTOR_SCORES reproduces the original psych engine bit-for-bit", {
  skip_on_cran()
  skip_if_not_installed("psych")

  efa_raw <- suppressMessages(suppressWarnings(
    EFA(DOSPERT_raw, n_factors = 10, type = "EFAtools", method = "PAF",
        rotation = "oblimin")))

  fixtures <- list(
    list(x = DOSPERT_raw, f = efa_raw, method = "Bartlett"),
    list(x = test_models$baseline$cormat, f = efa_ob, method = "Thurstone"),
    list(x = test_models$baseline$cormat, f = efa_or, method = "Thurstone")
  )

  for (fx in fixtures) {
    lbl <- paste0(fx$method, "/", if (.is_cormat(fx$x)) "cor" else "raw")
    new <- suppressMessages(suppressWarnings(
      FACTOR_SCORES(fx$x, f = fx$f, method = fx$method)))

    rot <- fx$f$settings$rotation
    L <- unclass(if (!identical(rot, "none")) fx$f$rot_loadings else fx$f$unrot_loadings)
    ps <- psych::factor.scores(x = as.matrix(fx$x), f = L, Phi = fx$f$Phi,
                               method = fx$method)

    expect_equal(new$weights, ps$weights, tolerance = 1e-12,
                 ignore_attr = TRUE, info = lbl)
    expect_equal(new$r.scores, ps$r.scores, tolerance = 1e-12,
                 ignore_attr = TRUE, info = lbl)
    if (is.null(ps$scores)) {
      expect_null(new$scores, info = lbl)
    } else {
      expect_equal(new$scores, ps$scores, tolerance = 1e-12,
                   ignore_attr = TRUE, info = lbl)
    }

    # R2 is the corrected native determinacy^2; for regression it is diag(S' R^-1 S).
    es <- suppressMessages(suppressWarnings(
      efa_scores(fx$x, f = fx$f,
                 method = if (fx$method == "Thurstone") "regression" else fx$method)))
    expect_equal(unname(new$R2), unname(es$determinacy$rho2),
                 tolerance = 1e-12, info = lbl)
    if (fx$method == "Thurstone") {
      Phi <- if (is.null(fx$f$Phi)) diag(ncol(L)) else unclass(fx$f$Phi)
      S <- L %*% Phi
      Rr <- unclass(fx$f$orig_R)
      expect_equal(unname(new$R2), unname(diag(t(S) %*% solve(Rr, S))),
                   tolerance = 1e-10, info = lbl)
    }
  }
})

test_that("FACTOR_SCORES keeps its legacy output shape and Thurstone alias", {
  # Legacy names: raw carries `missing`, correlation input does not.
  fs_raw <- suppressMessages(suppressWarnings(
    FACTOR_SCORES(GRiPS_raw, f = efa_grips, method = "Bartlett")))
  fs_cor <- suppressMessages(FACTOR_SCORES(test_models$baseline$cormat, f = efa_ob))

  expect_named(fs_raw, c("scores", "weights", "r.scores", "missing", "R2", "settings"))
  expect_named(fs_cor, c("scores", "weights", "r.scores", "R2", "settings"))
  expect_s3_class(fs_raw, "FACTOR_SCORES")

  # "Thurstone" maps to efa_scores' "regression": identical weights.
  fs_thu <- suppressMessages(FACTOR_SCORES(test_models$baseline$cormat, f = efa_ob,
                                           method = "Thurstone"))
  es_reg <- suppressMessages(efa_scores(test_models$baseline$cormat, f = efa_ob,
                                        method = "regression"))
  expect_equal(fs_thu$weights, es_reg$weights, ignore_attr = TRUE)
})

test_that("the f error names the averaged matrix an efa_average object carries", {
  # An efa_average object is rejected while "a matrix" is listed as acceptable, and its
  # averaged loadings are exactly that. The message now says where they are.
  avg <- suppressWarnings(suppressMessages(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = c("PAF", "ULS"), rotation = "promax", type = "EFAtools",
                start_method = "psych", show_progress = FALSE)))

  e <- tryCatch(efa_scores(GRiPS_raw, f = avg), efa_scores_bad_f = function(e) e)
  expect_s3_class(e, "efa_scores_bad_f")
  expect_snapshot(cat(conditionMessage(e)))

  # The component the message names is accepted. Supplied on its own it carries no factor
  # correlations, which efa_scores() reports itself (efa_scores_phi_null), so the message
  # does not have to.
  expect_s3_class(
    suppressMessages(efa_scores(test_models$baseline$cormat, f = avg$loadings$average)),
    "efa_scores")
})

rm(efa_ob, efa_or, efa_grips, fs_ob_cor, fs_grips)

# Synthetic fixtures: a two-factor model with a known pattern and factor correlation.
# The model-implied correlation matrix R = Lambda Phi Lambda' + diag(1 - h2) is used so
# the diagnostics line up with the theoretical determinacy of the regression scores (the
# quantity the psych oracle and the closed forms below report). The r.scores identities
# hold for any PD R, so the same fixtures serve every test.
Lambda <- matrix(c(0.8, 0.1,
                   0.7, 0.0,
                   0.6, 0.1,
                   0.1, 0.8,
                   0.0, 0.7,
                   0.2, 0.6), nrow = 6, byrow = TRUE,
                 dimnames = list(paste0("V", 1:6), c("F1", "F2")))
Phi <- matrix(c(1, 0.3, 0.3, 1), 2, dimnames = list(c("F1", "F2"), c("F1", "F2")))
h2  <- diag(Lambda %*% Phi %*% t(Lambda))
S   <- Lambda %*% Phi
R   <- Lambda %*% Phi %*% t(Lambda) + diag(1 - h2)

# Orthogonal counterpart (Phi = I) for the Anderson-Rubin identity.
Phi_or <- diag(2)
h2_or  <- diag(Lambda %*% t(Lambda))
S_or   <- Lambda
R_or   <- Lambda %*% t(Lambda) + diag(1 - h2_or)


test_that("r.scores reproduces the target factor correlations", {
  # ten Berge scores preserve the factor correlations (W' R W = Phi), so their
  # intercorrelation matrix equals Phi (ten Berge et al., 1999).
  W_tb <- EFAtools:::.factor_score_weights(Lambda, Phi, R, h2, method = "tenBerge")
  d_tb <- EFAtools:::.factor_score_diagnostics(W_tb, R, S)
  expect_equal(d_tb$r.scores, Phi, tolerance = 1e-8, ignore_attr = TRUE)

  # Anderson-Rubin scores are uncorrelated with unit variance (W' R W = I), so their
  # intercorrelation matrix is the identity (Anderson & Rubin, 1956).
  W_an <- EFAtools:::.factor_score_weights(Lambda, Phi_or, R_or, h2_or, method = "Anderson")
  d_an <- EFAtools:::.factor_score_diagnostics(W_an, R_or, S_or)
  expect_equal(d_an$r.scores, diag(2), tolerance = 1e-8, ignore_attr = TRUE)
})

test_that("score_cor has the right shape, carries the factor names, and holds determinacy", {
  W <- EFAtools:::.factor_score_weights(Lambda, Phi, R, h2, method = "Thurstone")
  d <- EFAtools:::.factor_score_diagnostics(W, R, S)

  expect_equal(dim(d$score_cor), c(ncol(Lambda), ncol(Lambda)))
  expect_identical(dimnames(d$score_cor), list(colnames(Lambda), colnames(Lambda)))
  expect_identical(dimnames(d$r.scores), dimnames(d$score_cor))
  # For a proper solution the reported determinacy is exactly the score_cor diagonal.
  expect_equal(d$determinacy, diag(d$score_cor), tolerance = 1e-12, ignore_attr = TRUE)
})

test_that("determinacy follows psych's cap-at-1 / non-positive-to-NA convention", {
  # A single-factor toy where the weight is scaled so the score-factor "correlation"
  # exceeds 1: with R = 1, W'S / sqrt(W'RW) = S / |W| for W, S scalars.
  over <- EFAtools:::.factor_score_diagnostics(W = matrix(1), R = matrix(1),
                                               S = matrix(2))
  expect_equal(unname(over$determinacy), 1)            # capped from 2
  expect_equal(unname(over$guttman), 1)                # 2 * 1^2 - 1

  # A negative score-factor "correlation" is set to NA (and propagates to guttman).
  neg <- EFAtools:::.factor_score_diagnostics(W = matrix(1), R = matrix(1),
                                              S = matrix(-0.5))
  expect_true(is.na(neg$determinacy))
  expect_true(is.na(neg$guttman))

  # A non-PD R (here a 1x1 "correlation" of -1) makes diag(C) < 0, so the score sd and
  # the validity are NaN; the guard must report NA (not a leaked NaN). Warnings from the
  # degenerate input (sqrt/cov2cor) are expected and suppressed.
  nan <- suppressWarnings(
    EFAtools:::.factor_score_diagnostics(W = matrix(1), R = matrix(-1), S = matrix(1)))
  expect_true(is.na(nan$determinacy))
  expect_true(is.na(nan$guttman))
})

test_that("regression determinacy matches psych's factor-score validity", {
  skip_on_cran()
  skip_if_not_installed("psych")

  W  <- EFAtools:::.factor_score_weights(Lambda, Phi, R, h2, method = "Thurstone")
  d  <- EFAtools:::.factor_score_diagnostics(W, R, S)

  # psych's Grice = TRUE value is gricef(), diag(S' R^-1 S) -- the same regression
  # determinacy we report, so the two must agree to machine precision. psych's default
  # (Grice = FALSE) builds the augmented error block from rowSums(f^2) rather than the
  # true communality diag(Lambda Phi Lambda'), and diverges for oblique factors; it is
  # a different quantity, not a looser version of this one.
  fs <- psych::factor.scores(x = R, f = unclass(Lambda), Phi = unname(Phi),
                             method = "regression", Grice = TRUE)

  # psych reports R2 (squared validity); compare against our determinacy (the validity
  # rho). The measured agreement is 1.1e-16; 1e-10 keeps headroom for BLAS variation.
  expect_equal(unname(d$determinacy), sqrt(fs$R2), tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("the Guttman index matches its closed form", {
  skip_on_cran()

  W <- EFAtools:::.factor_score_weights(Lambda, Phi, R, h2, method = "Thurstone")
  d <- EFAtools:::.factor_score_diagnostics(W, R, S)

  # Guttman's indeterminacy index is the minimum correlation between two sets of factor
  # scores equally consistent with the model: G = 2 rho^2 - 1, where rho is the multiple
  # correlation between a factor and its regression scores, rho_j^2 = s_j' R^-1 s_j taken
  # on the structure matrix S = Lambda Phi (Guttman, 1955, Psychometrika, 20(3), 173-197,
  # doi:10.1007/BF02289047). The check evaluates that definition directly, independently of
  # the weight matrix the implementation routes through. Cross-validated once against
  # fungible::fsIndeterminacy on this fixture, which agrees to 4.4e-16.
  rho <- sqrt(diag(t(S) %*% solve(R) %*% S))
  expect_equal(unname(d$guttman), unname(2 * rho^2 - 1), tolerance = 1e-6)
})

rm(Lambda, Phi, h2, S, R, Phi_or, h2_or, S_or, R_or)

# Fixtures: an oblique and an orthogonal EFA fit supply Lambda / Phi / R that are
# passed identically to the native engine and to psych::factor.scores(), so the
# two weight matrices must agree regardless of the fit quality.
R_mat <- test_models$baseline$cormat

# The alias check below compares two separate invocations, so it uses expect_equal() with an
# explicit tolerance rather than expect_identical(): a threaded BLAS (Apple's Accelerate, for
# one) is free to vary its GEMM reduction order between calls, so one function on one input
# can differ in the last ulp. waldo still compares S3 classes, names, attributes and structure
# exactly, so only numeric values are given slack -- an alias resolved to a different weight
# formula moves the weights in the second or third decimal and still fails loudly. (A set
# tolerance does relax integer against double, which these weights never are.)
fp_tol <- 1e-8

efa_ob <- suppressMessages(suppressWarnings(
  EFA(R_mat, n_factors = 3, N = 500, type = "EFAtools", method = "PAF",
      rotation = "oblimin")))
Lambda_ob <- unclass(efa_ob$rot_loadings)
Phi_ob <- unclass(efa_ob$Phi)
h2_ob <- diag(Lambda_ob %*% Phi_ob %*% t(Lambda_ob))

efa_or <- suppressMessages(suppressWarnings(
  EFA(R_mat, n_factors = 3, N = 500, type = "EFAtools", method = "PAF",
      rotation = "varimax")))
Lambda_or <- unclass(efa_or$rot_loadings)
Phi_or <- diag(ncol(Lambda_or))
h2_or <- diag(Lambda_or %*% Phi_or %*% t(Lambda_or))


test_that("native weights match psych::factor.scores for every method", {
  skip_on_cran()
  skip_if_not_installed("psych")

  # Oblique fixture: every method except Anderson (which is orthogonal only).
  for (m in c("Thurstone", "Bartlett", "tenBerge", "Harman", "components")) {
    W <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                          method = m)
    W_psych <- psych::factor.scores(x = R_mat, f = Lambda_ob, Phi = Phi_ob,
                                    method = m)$weights
    expect_equal(W, W_psych, tolerance = 1e-6, ignore_attr = TRUE,
                 info = paste("method:", m))
  }

  # Anderson-Rubin is defined for orthogonal factors only.
  W_and <- EFAtools:::.factor_score_weights(Lambda_or, Phi_or, R_mat, h2_or,
                                            method = "Anderson")
  W_and_psych <- psych::factor.scores(x = R_mat, f = Lambda_or, Phi = Phi_or,
                                      method = "Anderson")$weights
  expect_equal(W_and, W_and_psych, tolerance = 1e-6, ignore_attr = TRUE)
})

test_that("regression is an alias of Thurstone", {
  W_thu <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                            method = "Thurstone")
  W_reg <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                            method = "regression")
  expect_equal(W_thu, W_reg, tolerance = fp_tol)
})

test_that("orthogonal regression weights reproduce the SMC validity coefficients", {
  # For Phi = I the regression weights are W = R^-1 Lambda, so diag(Lambda' W)
  # equals diag(Lambda' R^-1 Lambda): each factor's squared multiple correlation
  # with the observed variables (Grice, 2001).
  W <- EFAtools:::.factor_score_weights(Lambda_or, Phi_or, R_mat, h2_or,
                                        method = "regression")
  smc <- diag(t(Lambda_or) %*% solve(R_mat, Lambda_or))
  expect_equal(diag(crossprod(Lambda_or, W)), smc, tolerance = 1e-8)
})

test_that("weights satisfy their estimator-defining properties (psych-free)", {
  # Closed-form identities that pin the weight *values* without psych, so the
  # numeric correctness is still guarded on CRAN (where the psych cross-check is
  # skipped).

  # Bartlett is conditionally unbiased: Lambda' W = I (Bartlett, 1937).
  W_bart <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                             method = "Bartlett")
  expect_equal(crossprod(Lambda_ob, W_bart), diag(ncol(Lambda_ob)),
               tolerance = 1e-8, ignore_attr = TRUE)

  # Anderson-Rubin scores are uncorrelated with unit variance: W' R W = I
  # (Anderson & Rubin, 1956).
  W_and <- EFAtools:::.factor_score_weights(Lambda_or, Phi_or, R_mat, h2_or,
                                            method = "Anderson")
  expect_equal(crossprod(W_and, R_mat %*% W_and), diag(ncol(Lambda_or)),
               tolerance = 1e-8, ignore_attr = TRUE)

  # ten Berge scores preserve the factor correlations: W' R W = Phi
  # (ten Berge et al., 1999).
  W_tb <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                           method = "tenBerge")
  expect_equal(crossprod(W_tb, R_mat %*% W_tb), Phi_ob,
               tolerance = 1e-8, ignore_attr = TRUE)
})

test_that("Bartlett and Anderson weights lie in the column space of Psi^-1 Lambda", {
  # The identities above pin the right-hand normalizer but not the uniqueness
  # weighting: Lambda' W = I holds for W = D Lambda (Lambda' D Lambda)^-1 with
  # *any* symmetric D, and W' R W = I holds for the analogous Anderson-Rubin form,
  # so both stay at machine precision even if Psi^-1 is replaced by Psi, by
  # diag(1 / h2), or by the identity. What identifies the weighting is the column
  # space: both estimators are Psi^-1 Lambda times an invertible m x m factor, so
  # W must leave no residual when regressed on Psi^-1 Lambda. The residuals are
  # judged against the scale of W itself, so the bound does not silently loosen if
  # the fixture is exchanged for one with larger weights.

  # Bartlett: W = Psi^-1 Lambda (Lambda' Psi^-1 Lambda)^-1 (Bartlett, 1937).
  Lu_ob <- Lambda_ob / (1 - h2_ob)
  W_bart <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                             method = "Bartlett")
  expect_lt(max(abs(qr.resid(qr(Lu_ob), W_bart))), 1e-8 * max(abs(W_bart)))

  # Anderson-Rubin: W = Psi^-1 Lambda (Lambda' Psi^-1 R Psi^-1 Lambda)^-1/2
  # (Anderson & Rubin, 1956).
  Lu_or <- Lambda_or / (1 - h2_or)
  W_and <- EFAtools:::.factor_score_weights(Lambda_or, Phi_or, R_mat, h2_or,
                                            method = "Anderson")
  expect_lt(max(abs(qr.resid(qr(Lu_or), W_and))), 1e-8 * max(abs(W_and)))

  # W' R W = I fixes the Anderson-Rubin right factor only up to an orthogonal
  # rotation. The estimator takes the *symmetric* inverse root, which leaves
  # Lu' R W = (Lu' R Lu)^1/2 symmetric; a mis-specified Psi makes the right factor
  # non-symmetric and shows up here even though it cannot disturb W' R W = I.
  M_and <- crossprod(Lu_or, R_mat %*% W_and)
  expect_lt(max(abs(M_and - t(M_and))), 1e-8 * max(abs(M_and)))
})

test_that("the singular-matrix pseudo-inverse fallback works", {
  # .pinv is the Moore-Penrose inverse of a rank-deficient matrix: M Mp M = M
  # and Mp M Mp = Mp.
  set.seed(7)
  B <- matrix(rnorm(12), 4, 3)
  M <- tcrossprod(cbind(B, B[, 1]))  # 4x4, rank 3 (singular)
  Mp <- EFAtools:::.pinv(M)
  expect_equal(M %*% Mp %*% M, M, tolerance = 1e-8)
  expect_equal(Mp %*% M %*% Mp, Mp, tolerance = 1e-8)

  # An exactly singular correlation matrix routes solve() through .pinv instead
  # of erroring, so the engine still returns the pseudo-inverse weights.
  R_sing <- matrix(c(1, 0.5, 0.5,
                     0.5, 1,   1,
                     0.5, 1,   1), nrow = 3)  # rows 2 and 3 identical -> singular
  Lam <- matrix(c(0.7, 0.2,
                  0.5, 0.3,
                  0.4, 0.4), nrow = 3, byrow = TRUE)
  h2s <- diag(Lam %*% t(Lam))
  W_sing <- EFAtools:::.factor_score_weights(Lam, diag(2), R_sing, h2s,
                                             method = "Thurstone")
  expect_true(all(is.finite(W_sing)))
  expect_equal(W_sing, EFAtools:::.pinv(R_sing) %*% Lam,
               tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("non-positive uniquenesses abort for Bartlett and Anderson", {
  Lambda <- matrix(c(0.9, 0.4,
                     0.5, 0.3,
                     0.4, 0.8), nrow = 3, byrow = TRUE)
  Phi <- diag(2)
  R <- matrix(c(1,    0.30, 0.20,
                0.30, 1,    0.25,
                0.20, 0.25, 1),    nrow = 3)
  h2 <- c(1.02, 0.4, 0.5)  # first communality > 1 -> non-positive uniqueness

  expect_error(
    EFAtools:::.factor_score_weights(Lambda, Phi, R, h2, method = "Bartlett"),
    class = "efa_scores_nonpositive_uniqueness")
  expect_error(
    EFAtools:::.factor_score_weights(Lambda, Phi, R, h2, method = "Anderson"),
    class = "efa_scores_nonpositive_uniqueness")

  # A proper solution (all uniquenesses positive) does not warn.
  Lambda_ok <- matrix(c(0.7, 0.1,
                        0.6, 0.2,
                        0.2, 0.7), nrow = 3, byrow = TRUE)
  h2_ok <- diag(Lambda_ok %*% Phi %*% t(Lambda_ok))
  expect_no_warning(
    EFAtools:::.factor_score_weights(Lambda_ok, Phi, R, h2_ok, method = "Bartlett"))
})

test_that("matrix (inverse) square-root helpers are correct, including 1x1", {
  set.seed(1)
  A <- crossprod(matrix(rnorm(4), 2, 2)) + diag(2)  # 2x2 symmetric positive definite

  expect_equal(EFAtools:::.mat_sqrt(A) %*% EFAtools:::.mat_sqrt(A), A,
               tolerance = 1e-10)
  expect_equal(EFAtools:::.inv_mat_sqrt(A), solve(EFAtools:::.mat_sqrt(A)),
               tolerance = 1e-10)
  expect_equal(EFAtools:::.mat_sqrt(A) %*% EFAtools:::.inv_mat_sqrt(A), diag(2),
               tolerance = 1e-10)

  # 1x1 inputs must return the scalar (inverse) root, not a mis-dimensioned
  # identity from diag(scalar).
  expect_equal(EFAtools:::.mat_sqrt(matrix(4, 1, 1)), matrix(2, 1, 1),
               tolerance = 1e-12)
  expect_equal(EFAtools:::.inv_mat_sqrt(matrix(4, 1, 1)), matrix(0.5, 1, 1),
               tolerance = 1e-12)

  # Indefinite / singular inputs exercise the eigenvalue clamps: without them
  # .mat_sqrt would take the root of a negative eigenvalue (NaN) and
  # .inv_mat_sqrt would divide by a zero eigenvalue (Inf).
  expect_true(all(is.finite(EFAtools:::.mat_sqrt(diag(c(4, 1, -0.25))))))
  expect_true(all(is.finite(EFAtools:::.inv_mat_sqrt(diag(c(4, 1, 0))))))
})

test_that("W has the right shape and carries the loading dimnames for every method", {
  # The psych cross-check uses ignore_attr = TRUE, so dimnames are pinned here
  # (across methods, since the engine sets them in one shared step).
  for (m in c("Thurstone", "Bartlett", "tenBerge", "Harman", "components")) {
    W <- EFAtools:::.factor_score_weights(Lambda_ob, Phi_ob, R_mat, h2_ob,
                                          method = m)
    expect_equal(dim(W), dim(Lambda_ob), info = m)
    expect_identical(dimnames(W), dimnames(Lambda_ob), info = m)
  }
})

rm(R_mat, efa_ob, Lambda_ob, Phi_ob, h2_ob, efa_or, Lambda_or, Phi_or, h2_or)

#' Kaiser-Guttman Criterion
#'
#' Probably the most popular factor retention criterion. Kaiser and Guttman suggested
#' to retain as many factors as there are sample eigenvalues greater than 1.
#' This is why the criterion is also known as eigenvalues-greater-than-one rule.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param eigen_type character. On what the eigenvalues should be found. Can be
#'  either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
#'  the diagonal values of the correlation matrices are left to be 1. If using
#'  "SMC", the diagonal of the
#'  correlation matrices is replaced by the squared multiple correlations (SMCs)
#'  of the indicators. If using "EFA", eigenvalues are found on the correlation
#'  matrices with the final communalities of an exploratory factor analysis
#'  solution (default is principal axis factoring extracting 1 factor) as
#'  diagonal.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#'  data is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#' @param n_factors numeric. Number of factors to extract if "EFA" is included in
#' \code{eigen_type}. Default is 1.
#' @param ... Additional arguments passed to \code{\link{EFA}}. For example,
#' to change the extraction method (PAF is default).
#'
#' @details Originally, the Kaiser-Guttman criterion was intended for the use
#' with prinicpal components, hence with eigenvalues derived from the original
#' correlation matrix. This can be done here by setting \code{eigen_type} to
#' "PCA". However, it is well-known that this criterion is often inaccurate and
#' that it tends to overestimate the number of factors, especially for unidimensional
#' or orthogonal factor structures (e.g., Zwick & Velicer, 1986).
#'
#' The criterion's inaccuracy in these cases is somewhat addressed if it is
#' applied on the correlation matrix with communalities in the diagonal, either
#' initial communalities estimated from SMCs (done setting \code{eigen_type} to
#' "SMC") or final communality estimates from an EFA (done setting \code{eigen_type}
#' to "EFA"; see Auerswald & Moshagen, 2019). However, although this variant
#' of the KGC is more accurate in some cases compared to the traditional KGC, it
#' is at the same time less accurate than the PCA-variant in other cases, and it
#' is still often less accurate than other factor retention methods, for
#' example parallel analysis (\code{\link{PARALLEL}}), the Hull method
#' \code{\link{HULL}}, or sequential \eqn{chi^2} model tests (\code{\link{SMT}};
#' see Auerswald & Moshagen, 2019).
#'
#' The \code{KGC} function can also be called together with other factor
#' retention criteria in the \code{\link{N_FACTORS}} function.
#'
#' @return A list of class KGC containing
#'
#' \item{eigen_PCA}{ A vector containing the eigenvalues found with PCA.}
#' \item{eigen_SMC}{ A vector containing the eigenvalues found with SMCs.}
#' \item{eigen_EFA}{ A vector containing the eigenvalues found with EFA.}
#' \item{n_fac_PCA}{ The number of factors to retain according to the Kaiser-
#' Guttmann criterion with PCA eigenvalues type.}
#' \item{n_fac_SMC}{ The number of factors to retain according to the Kaiser-
#' Guttmann criterion with SMC eigenvalues type.}
#' \item{n_fac_EFA}{ The number of factors to retain according to the Kaiser-
#' Guttmann criterion with EFA eigenvalues type.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Guttman, L. (1954). Some necessary conditions for common-factor analysis.
#' Psychometrika, 19, 149 –161. http://dx.doi.org/10.1007/BF02289162
#'
#' @source Kaiser, H. F. (1960). The application of electronic computers to factor
#' analysis. Educational and Psychological Measurement, 20, 141–151.
#' http://dx.doi.org/10.1177/001316446002000116
#'
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. Psychological Bulletin, 99,
#' 432–442. http://dx.doi.org/10.1037/0033-2909.99.3.432
#'
#' @seealso Other factor retention criteria: \code{\link{CD}}, \code{\link{EKC}},
#' \code{\link{HULL}}, \code{\link{PARALLEL}}, \code{\link{SMT}}
#'
#' \code{\link{N_FACTORS}} as a wrapper function for this and all the
#' above-mentioned factor retention criteria.
#'
#' @export
#'
#' @examples
#' KGC(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
KGC <- function(x, eigen_type = c("PCA", "SMC", "EFA"),
                use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                        "everything", "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall"), n_factors = 1,
                ...){

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  eigen_type <- match.arg(eigen_type, several.ok = TRUE)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  checkmate::assert_count(n_factors)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed.\n"))

  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }


  # Store original R for later
  R_orig <- R

  # Prepare objects
  n_fac_PCA <- NA
  n_fac_SMC <- NA
  n_fac_EFA <- NA
  eigen_R_PCA <- NA
  eigen_R_SMC <- NA
  eigen_R_EFA <- NA

  if("PCA" %in% eigen_type) {

    # Calculate eigenvalues and determine number of factors
    eigen_R_PCA <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
    n_fac_PCA <- sum(eigen_R_PCA >= 1)

  }

  if("SMC" %in% eigen_type) {

    # Calculate SMCs and replace diagonal of correlation matrix with these
    inverse_R <- solve(R)
    SMCs <- 1 - 1 / diag(inverse_R)
    diag(R) <- SMCs

    # Calculate eigenvalues and determine number of factors
    eigen_R_SMC <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
    n_fac_SMC <- sum(eigen_R_SMC >= 1)

  }

  if("EFA" %in% eigen_type) {

    R <- R_orig

    # Do an EFA to get final communality estimates and replace diagonal of
    # correlation matrix with these
    EFA_h2 <- suppressMessages(suppressWarnings(EFA(R, n_factors = n_factors, ...)$h2))
    diag(R) <- EFA_h2

    # Calculate eigenvalues and determine number of factors
    eigen_R_EFA <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
    n_fac_EFA <- sum(eigen_R_EFA >= 1)

  }

  # prepare settings
  settings <- list(eigen_type = eigen_type,
                   use = use,
                   cor_method = cor_method,
                   n_factors = n_factors)

  # Prepare the output
  output <- list(
    eigen_PCA = eigen_R_PCA,
    eigen_SMC = eigen_R_SMC,
    eigen_EFA = eigen_R_EFA,
    n_fac_PCA = n_fac_PCA,
    n_fac_SMC = n_fac_SMC,
    n_fac_EFA = n_fac_EFA,
    settings = settings
  )

  class(output) <- "KGC"

  return(output)

}

kgc_cor <- KGC(test_models$baseline$cormat)
kgc_cor_smc <- KGC(test_models$baseline$cormat, eigen_type = "SMC")
kgc_raw <- KGC(GRiPS_raw)
# Check with an argument passed to "EFA"
kgc_efa_ml<- KGC(test_models$baseline$cormat, eigen_type = "EFA", method = "ML")

test_that("output class and dimensions are correct", {
  expect_is(kgc_cor, "KGC")
  expect_output(str(kgc_cor), "List of 7")
  expect_is(kgc_cor_smc, "KGC")
  expect_output(str(kgc_cor_smc), "List of 7")
  expect_is(kgc_raw, "KGC")
  expect_output(str(kgc_raw), "List of 7")
  expect_is(kgc_efa_ml, "KGC")
  expect_output(str(kgc_efa_ml), "List of 7")
})

test_that("found eigenvalues are correct", {
  expect_equal(sum(kgc_cor$eigen_PCA), ncol(test_models$baseline$cormat))
  expect_lt(sum(kgc_cor$eigen_SMC), ncol(test_models$baseline$cormat))
  expect_lt(sum(kgc_cor$eigen_EFA), ncol(test_models$baseline$cormat))

  expect_equal(sum(kgc_raw$eigen_PCA), ncol(GRiPS_raw))
  expect_lt(sum(kgc_raw$eigen_SMC), ncol(GRiPS_raw))
  expect_lt(sum(kgc_raw$eigen_EFA), ncol(GRiPS_raw))

  expect_lt(sum(kgc_raw$eigen_SMC), ncol(test_models$baseline$cormat))
  expect_equal(c(kgc_cor_smc$eigen_PCA, kgc_cor_smc$eigen_EFA), c(NA, NA))

  expect_lt(sum(kgc_raw$eigen_EFA), ncol(test_models$baseline$cormat))
  expect_equal(c(kgc_efa_ml$eigen_PCA, kgc_efa_ml$eigen_SMC), c(NA, NA))
})

test_that("identified number of factors is correct", {
  expect_equal(kgc_cor$n_fac_PCA, 3)
  expect_equal(kgc_cor$n_fac_SMC, 1)
  expect_equal(kgc_cor$n_fac_EFA, 1)

  expect_equal(kgc_raw$n_fac_PCA, 1)
  expect_equal(kgc_raw$n_fac_SMC, 1)
  expect_equal(kgc_raw$n_fac_EFA, 1)

  expect_equal(kgc_cor_smc$n_fac_SMC, 1)
  expect_equal(c(kgc_cor_smc$n_fac_PCA, kgc_cor_smc$n_fac_EFA), c(NA, NA))

  expect_equal(kgc_efa_ml$n_fac_EFA, 1)
  expect_equal(c(kgc_efa_ml$n_fac_PCA, kgc_efa_ml$n_fac_SMC), c(NA, NA))
})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z), ncol = 3)
cor_sing <- stats::cor(dat_sing)

cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

test_that("errors are thrown correctly", {
  expect_error(KGC(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_message(KGC(GRiPS_raw, eigen_type = "PCA"), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_error(KGC(dat_sing), " Correlation matrix is singular, no further analyses are performed.\n")
  expect_error(KGC(cor_sing, N = 10), " Correlation matrix is singular, no further analyses are performed.\n")
  expect_warning(KGC(cor_nposdef, N = 10), "Matrix was not positive definite, smoothing was done")
})

test_that("settings are returned correctly", {
  expect_named(kgc_cor$settings, c("eigen_type", "use", "cor_method", "n_factors"))
  expect_named(kgc_raw$settings, c("eigen_type", "use", "cor_method", "n_factors"))
  expect_named(kgc_cor_smc$settings, c("eigen_type", "use", "cor_method", "n_factors"))
  expect_named(kgc_efa_ml$settings, c("eigen_type", "use", "cor_method", "n_factors"))

  expect_equal(kgc_cor$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(kgc_raw$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(kgc_cor_smc$settings$eigen_type, "SMC")
  expect_equal(kgc_efa_ml$settings$eigen_type, "EFA")

  expect_equal(kgc_cor$settings$use, "pairwise.complete.obs")
  expect_equal(kgc_raw$settings$use, "pairwise.complete.obs")
  expect_equal(kgc_cor_smc$settings$use, "pairwise.complete.obs")
  expect_equal(kgc_efa_ml$settings$use, "pairwise.complete.obs")

  expect_equal(kgc_cor$settings$cor_method, "pearson")
  expect_equal(kgc_raw$settings$cor_method, "pearson")
  expect_equal(kgc_cor_smc$settings$cor_method, "pearson")
  expect_equal(kgc_efa_ml$settings$cor_method, "pearson")

  expect_equal(kgc_cor$settings$n_factors, 1)
  expect_equal(kgc_raw$settings$n_factors, 1)
  expect_equal(kgc_cor_smc$settings$n_factors, 1)
  expect_equal(kgc_efa_ml$settings$n_factors, 1)

})

rm(kgc_cor, kgc_cor_smc, kgc_raw, kgc_efa_ml, x, y, z, dat_sing, cor_sing,
   cor_nposdef)

#' Next eigenvalue sufficiency test (NEST)
#'
#' NEST uses many synthetic datasets to generate reference eigenvalues against
#' which to compare the empirical eigenvalues. This is similar to parallel
#' analysis, but other than parallel analysis, NEST does not just rely on
#' synthetic eigenvalues based on an identity matrix as null model.
#' It was introduced by Achim (2017), see also Brandenburg and Papenberg (2024) and
#' Caron (2025) for further simulation studies including NEST.
#'
#' @param x data.frame or matrix. data.frame or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. The number of observations. Only needed if x is a correlation
#'  matrix.
#' @param alpha numeric. The alpha level to use (i.e., 1-alpha percentile of eigenvalues is used for reference values).
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
#'  data is given as input. Default is  \code{"pairwise.complete.obs"}.
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#'  Default is  \code{"pearson"}.
#' @param n_datasets numeric. The number of datasets to simulate. Default is 1000.
#' @param ... Additional arguments passed to \code{\link{EFA}}. For example,
#' the extraction method can be changed here (default is "PAF"). PAF is more
#' robust, but it will take longer compared to the other estimation methods
#' available ("ML" and "ULS").
#'
#' @details NEST compares the first empirical eigenvalue against the first eigenvalues
#' of \code{n_dataset} synthetic datasets based on a null model  (i.e.,
#'  with uncorrelated variables; same as in parallel analysis, see \code{\link{PARALLEL}}).
#'  The following eigenvalues are compared against synthetic datasets based on an EFA-model with one fewer factors
#'  than the position of the respective empirical eigenvalue. E.g, the second
#'  empirical eigenvalue is compared against synthetic data based on a one-factor
#'  model. The \code{alpha}-level defines against which percentile of the synthetic
#'  eigenvalue distribution to compare the empirical eigenvalues against, i.e., an
#'  alpha of .05 (the default) uses the 95th percentile as reference value.
#'
#'  For details on the method, including simulation studies, see Achim (2017),
#'  Brandenburg and Papenberg (2024), and Caron (2025).
#'
#'  The \code{NEST} function can also be called together with other factor
#'   retention criteria in the \code{\link{N_FACTORS}} function.
#'
#'
#' @return A list of class NEST containing the following objects
#' \item{eigenvalues}{A vector containing the empirical eigenvalues of the entered data.}
#' \item{n_factors}{The number of factors to retain according to the NEST procedure.}
#' \item{references}{A vector containing the reference eigenvalues.}
#' \item{prob}{For the first n_factors + 1 empirical eigenvalues, the proportion <= the set of n_datasets  reference eigenvalues.}
#' \item{settings}{A list of control settings used in the print function.}
#'
#' @source Achim, A. (2017). Testing the number of required dimensions in exploratory factor analysis. The Quantitative Methods for Psychology, 13(1), 64–74. https://doi.org/10.20982/tqmp.13.1.p064
#' @source Brandenburg, N., & Papenberg, M. (2024). Reassessment of innovative methods to determine the number of factors: A simulation-based comparison of exploratory graph analysis and Next Eigenvalue Sufficiency Test. Psychological Methods, 29(1), 21–47. https://doi.org/10.1037/met0000527
#' @source Caron, P.-O. (2025). A Comparison of the Next Eigenvalue Sufficiency Test to Other Stopping Rules for the Number of Factors in Factor Analysis.
#' Educational and Psychological Measurement, Online-first publication. https://doi.org/10.1177/00131644241308528
#'
#' @export
#'
#' @examples
#'
#' # with correlation matrix
#' NEST(test_models$baseline$cormat, N = 500)
#'
#' # with raw data
#' NEST(GRiPS_raw)
NEST <- function(x, N = NA,
                 alpha = .05,
                 use = c("pairwise.complete.obs", "all.obs",
                         "complete.obs", "everything",
                         "na.or.complete"),
                 cor_method = c("pearson", "spearman", "kendall"),
                 n_datasets = 1000,
                 ...) {


  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  checkmate::assert_count(N, na.ok = TRUE)
  checkmate::assert_number(alpha, lower = 0, upper = 1)
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  checkmate::assert_count(n_datasets, na.ok = FALSE,
                          positive = TRUE)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

    R <- x

    if (is.na(N)) {

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Argument 'N' was NA but correlation matrix was entered. Please either provide N or raw data.\n"))

    }

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

  }

  # Check if correlation matrix is invertable, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error")) {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {

    R <- psych::cor.smooth(R)

  }

  emp_eigen <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
  nvar <- ncol(R)

  # rule by Achim (2017)
  max_fac <- floor(.8 * nvar)

  references <- vector(mode = "double", length = max_fac)
  prob <- vector(mode = "double", length = max_fac)

  for (nf in 1:max_fac) {

    if (nf == 1) {

      # for the first factor, use identity matrix as reference data
      M <- diag(nvar)

    } else {


      # For further factors, use model with nf - 1 as reference data
      mm <- EFA(R, N = N, n_factors = nf - 1, ...)
      L <- mm$unrot_loadings
      u2 <- diag(sqrt(1-mm$h2))

      # bind loadings and uniquenesses together to create data according to
      # factor score rule X = F*L + e, which is faster than using the model
      # implied R to draw from a multivariate normal distribution.
      M <- cbind(L, u2)


    }

    # use C++ to simulate datasets
    ref_values <- .nest_sym(nf, N, t(M), n_datasets)

    references[nf] <- stats::quantile(ref_values, probs = 1 - alpha)
    prob[nf] <- mean(emp_eigen[nf] < ref_values)
    if (emp_eigen[nf] <= references[nf]) {
      break
    }


  }

  n_factors <- nf - 1
  references <- references[1:nf]
  prob <- prob[1:nf]

  out <- list(
    eigenvalues = emp_eigen,
    n_factors = n_factors,
    references = references,
    prob = prob,
    settings = list(
      alpha = alpha,
      N = N,
      n_datasets = n_datasets,
      use = use,
      cor_method = cor_method
    )
  )

  class(out) <- "NEST"

  out

}

nest_cor <- NEST(test_models$baseline$cormat, N = 500)
nest_raw <- NEST(GRiPS_raw)

test_that("output class and dimensions are correct", {
  expect_is(nest_cor, "NEST")
  expect_output(str(nest_cor), "List of 5")
  expect_is(nest_raw, "NEST")
  expect_output(str(nest_raw), "List of 5")
})


test_that("found eigenvalues are correct", {
  expect_equal(sum(nest_cor$eigenvalues),
               ncol(test_models$baseline$cormat))
  expect_equal(sum(nest_raw$eigenvalues), ncol(GRiPS_raw))
  expect_length(nest_cor$eigenvalues,
                ncol(test_models$baseline$cormat))
  expect_length(nest_raw$eigenvalues, ncol(GRiPS_raw))
})

test_that("reference eigenvalues are correct", {
  expect_length(nest_cor$references, nest_cor$n_factors + 1)
  expect_length(nest_raw$references, nest_raw$n_factors + 1)
  expect_lte(nest_cor$eigenvalues[nest_cor$n_factors + 1],
             nest_cor$references[nest_cor$n_factors + 1])
  expect_true(all(nest_cor$eigenvalues[1:nest_cor$n_factors] >
            nest_cor$references[1:nest_cor$n_factors]))
  expect_lte(nest_raw$eigenvalues[nest_raw$n_factors + 1],
             nest_raw$references[nest_raw$n_factors + 1])
  expect_true(all(nest_raw$eigenvalues[1:nest_raw$n_factors] >
            nest_raw$references[1:nest_raw$n_factors]))
})

test_that("identified number of factors is correct", {
  expect_equal(nest_cor$n_factors, 3)
  expect_equal(nest_raw$n_factors, 1)
})

# Create singular correlation matrix for tests
x <- rnorm(10)
y <- rnorm(10)
z <- x + y
dat_sing <- matrix(c(x, y, z, rnorm(10), rnorm(10), rnorm(10)), ncol = 6)
cor_sing <- stats::cor(dat_sing)

cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

test_that("errors are thrown correctly", {
  expect_error(NEST(1:5), " 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n")
  expect_error(NEST(test_models$baseline$cormat), " Argument 'N' was NA but correlation matrix was entered. Please either provide N or raw data.\n")
  expect_message(NEST(GRiPS_raw), " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n")
  expect_warning(NEST(GRiPS_raw, N = 20), " 'N' was set and data entered. Taking N from data.\n")
  expect_error(NEST(dat_sing), " Correlation matrix is singular, no further analyses are performed\n")
  expect_error(NEST(cor_sing, N = 20), " Correlation matrix is singular, no further analyses are performed\n")
})

test_that("settings are returned correctly", {

  expect_equal(nest_cor$settings$N, 500)
  expect_equal(nest_raw$settings$N, 810)

  expect_equal(nest_cor$settings$use, "pairwise.complete.obs")
  expect_equal(nest_raw$settings$use, "pairwise.complete.obs")

  expect_equal(nest_cor$settings$cor_method, "pearson")
  expect_equal(nest_raw$settings$cor_method, "pearson")
})

rm(nest_cor, nest_raw, x, y, z, dat_sing, cor_sing, cor_nposdef)

.efa_pooled_medoid_anchor <- function(unrot_loadings) {
  # Index of the imputation that is closest, in aligned squared Frobenius
  # distance, to all the others. Anchoring the alignment there instead of on
  # whichever imputation happens to be first makes the pooled solution invariant
  # to the order of `data_list`: the medoid is a property of the set. Cost is
  # O(D^2) cheap column matchings.
  #
  # Aligning to a single arbitrarily chosen imputation is the usual approach --
  # it is what the reference-target route below does, and what psych's
  # fa.pooled() does -- but it makes the result depend on which imputation that
  # is. The alternative in the literature is to drop the reference and iterate a
  # Generalized Procrustes consensus instead (Gower 1975; van Ginkel &
  # Kroonenberg 2014); choosing a central reference keeps the pooled estimate a
  # plain average of matched solutions, which the consensus centroid is not.
  D <- length(unrot_loadings)
  if (D < 2L) return(1L)
  Ls <- lapply(unrot_loadings, as.matrix)
  # The distance is symmetric -- the column matching of (b, a) is the inverse of
  # the matching of (a, b), and an orthogonal factor leaves the norm alone -- so
  # only the upper triangle needs matching, halving the O(D^2) work.
  dm <- matrix(0, D, D)
  for (a in seq_len(D - 1L)) {
    A <- Ls[[a]]
    for (b in seq.int(a + 1L, D)) {
      v <- sum((unclass(.align_solution(L_target = A, L = Ls[[b]])$loadings) - A)^2)
      dm[a, b] <- v
      dm[b, a] <- v
    }
  }
  tot <- rowSums(dm)

  # Ties are not a numerical edge case here: with two imputations every
  # candidate is tied by construction, and identical imputations tie for any D.
  # Breaking them on list position would put back exactly the order dependence
  # the medoid removes, so break them on the candidates' own content -- the
  # squared Frobenius norm, which no alignment can change and which does not
  # depend on where the imputation sits in the list. A candidate set that is
  # still tied after that falls back to the lowest position, which is order
  # dependent -- it needs two imputations with the same total distance *and* the
  # same norm, which outside of exactly duplicated solutions does not happen.
  scale <- max(1, max(abs(tot)))
  cand <- which(tot <= min(tot) + 1e-10 * scale)
  if (length(cand) == 1L) return(cand)
  ss <- vapply(cand, function(i) sum(Ls[[i]]^2), numeric(1))
  cand[which.max(round(ss, 10))]
}

.efa_pooled_canonical_gauge <- function(M, anchor) {
  # The extraction identifies the unrotated solution by making L' Psi^-1 L
  # diagonal with descending diagonal (Lawley & Maxwell 1971), so a single fit's
  # unrotated loadings are fixed up to column sign. Averaging aligned
  # per-imputation solutions does not preserve that constraint, which leaves the
  # pooled matrix in a gauge no component fit uses and makes it incomparable
  # element-by-element with any single-fit unrotated solution. Returns the k x k
  # orthogonal C that puts M C back in the canonical gauge. Psi is taken from the
  # pooled matrix itself, so the gauge is consistent with the loadings actually
  # reported.
  #
  # C is common to every imputation and orthogonal, so it leaves all
  # gauge-invariant quantities (communalities, total variance accounted for, the
  # model-implied correlation matrix and hence the residuals and RMSR) exactly
  # unchanged; only the split of variance across factors moves.
  k <- ncol(M)
  # Which constraint the extraction uses depends on the estimator: principal-axis
  # extractions diagonalise L'L, maximum likelihood diagonalises L' Psi^-1 L.
  # Rather than assume one, read it off `anchor`, which is an untouched
  # single-fit solution and therefore already satisfies whichever holds.
  gram  <- function(L) crossprod(L)
  # Psi is reconstructed from the loadings, and it has to be floored at the same
  # value the fitting routines use. A maximum likelihood solution with a
  # uniqueness pinned at the boundary diagonalises L' Psi^-1 L against the
  # *floored* psi, whereas 1 - rowSums(L^2) comes back a whisker below it; the
  # inverse multiplies that tiny gap by ~1/floor and turns it into a defect large
  # enough to fail the tolerance below, which would silently switch the
  # re-gauging off exactly for the boundary solutions. Flooring here reproduces
  # the psi the extraction actually used.
  wgram <- function(L) crossprod(L, L / pmax(1 - rowSums(L^2), .uniqueness_floor))
  defect <- function(A) {
    dg <- sum(abs(diag(A)))
    if (!is.finite(dg) || dg <= 0) return(Inf)
    sum(abs(A[upper.tri(A)])) / dg
  }
  d_g <- defect(gram(anchor))
  d_w <- defect(wgram(anchor))
  # Loose enough to clear the extraction's own convergence noise, tight enough
  # that the constraint the estimator does *not* use does not pass. The margin is
  # narrower than it looks: over a sweep of 492 fits (PAF/ML/ULS/MINRES/DWLS x
  # pearson/poly/tetra) the smallest defect for a *used* constraint was 1.7e-3
  # away on the ML side, while the *unused* constraint came as low as 2.5e-4
  # (DWLS with polychoric correlations). Do not widen this to 1e-3 on the
  # assumption that the two are an order of magnitude apart -- they are not.
  # Ties are settled by taking the smaller defect, not by the tolerance alone.
  tol <- 1e-4
  A <- if (is.finite(d_g) && d_g <= tol && d_g <= d_w) {
    gram(M)
  } else if (is.finite(d_w) && d_w <= tol) {
    # A row at or past the Heywood boundary carries no usable uniqueness, so leave an
    # improper average in the gauge the alignment produced rather than letting a
    # degenerate row steer the weighting. `wgram()` floors psi at `.uniqueness_floor`,
    # which already bounds the weight (at 1 / 0.005 = 200, not the unbounded blow-up an
    # unfloored Psi^-1 would give), so this guard is the second line of defence and not
    # the thing that keeps the weight finite -- do not remove the floor on the strength
    # of it.
    if (any(1 - rowSums(M^2) <= 1e-6)) return(diag(k))
    wgram(M)
  } else {
    # The component fits sit in no gauge this recognises (an improper solution,
    # say). Imposing a foreign one would be worse than leaving the average alone.
    return(diag(k))
  }
  # If the average never left the gauge -- identical or near-identical
  # imputations -- there is nothing to correct, and rotating anyway would move
  # the result by the extraction's convergence tolerance for no gain.
  if (defect(A) <= tol) return(diag(k))
  e <- eigen((A + t(A)) / 2, symmetric = TRUE)
  # Two (near-)equal eigenvalues leave their eigenvectors undetermined: every
  # rotation inside that eigenspace diagonalises A equally well, so the gauge
  # would be settled by rounding noise and could come out differently under a
  # different BLAS. Leave such a solution as the alignment produced it.
  # The sensitivity of an eigenvector goes as (machine epsilon)/gap, so the
  # threshold has to be well above epsilon to bound it: at a relative gap of
  # 1e-6 the worst-case movement is around 1e-10, comfortably below anything a
  # user would see, whereas a threshold at 1e-8 would still admit bases that
  # move in the sixth decimal between one BLAS and another.
  if (k > 1L) {
    ev <- e$values
    if (min(abs(diff(ev))) <= 1e-6 * max(abs(ev))) return(diag(k))
  }
  C <- e$vectors
  # Orient each column to agree with the anchor rather than by an internal rule,
  # so the pooled solution keeps the sign convention its component fits use. The
  # anchor is the medoid, so this stays independent of the imputation order.
  MC <- M %*% C
  s <- sign(colSums(MC * anchor))
  s[s == 0] <- 1
  C %*% diag(s, nrow = k)
}

.efa_pooled_align_unrotated_list <- function(unrot_loadings,
                                             align_unrotated = c("signed_tucker_congruence", "none", "procrustes")) {
  # Unrotated factor axes are arbitrary up to ordering and signs. This helper
  # puts them into a common orientation before simple arithmetic averaging.
  # Unlike the rotated solution below, this step should not seek simple
  # structure; it only removes indeterminacy in the unrotated axes.
  #
  # Returns the aligned loadings together with a parallel list of per-imputation
  # alignment metadata, so that downstream consumers can apply the same column
  # permutation (and, when signed gauges matter, sign vector) to other
  # per-imputation matrices that share the loading gauge, e.g. marginal-SE
  # matrices used by Rubin's-rules pooling of analytic SEs.
  align_unrotated <- match.arg(align_unrotated)

  D <- length(unrot_loadings)
  if (align_unrotated == "none") {
    k_first <- if (D > 0L) ncol(as.matrix(unrot_loadings[[1]])) else 0L
    identity_meta <- list(type         = "none",
                          factor_order = seq_len(k_first),
                          factor_sign  = rep(1, k_first))
    meta <- rep(list(identity_meta), D)
    return(list(loadings = unrot_loadings, meta = meta))
  }

  # The signed-permutation default anchors on the medoid, so the pooled solution
  # does not depend on the order of the imputation list. `procrustes` keeps its
  # documented first-imputation anchor.
  anchor <- if (align_unrotated == "signed_tucker_congruence") {
    .efa_pooled_medoid_anchor(unrot_loadings)
  } else {
    1L
  }
  target <- as.matrix(unrot_loadings[[anchor]])
  k <- ncol(target)

  out  <- vector("list", D)
  meta <- vector("list", D)
  out[[anchor]] <- target
  meta[[anchor]] <- if (align_unrotated == "signed_tucker_congruence") {
    list(type         = "signed_tucker_congruence",
         factor_order = seq_len(k),
         factor_sign  = rep(1, k))
  } else {
    # Procrustes anchor gauge: Q_1 = I_k, so the aligned analytic-SE propagation
    # reduces to the identity on the first imputation.
    list(type = "procrustes", Q = diag(k))
  }

  for (d in setdiff(seq_along(unrot_loadings), anchor)) {
    Ld <- as.matrix(unrot_loadings[[d]])

    if (align_unrotated == "signed_tucker_congruence") {
      # Preserve the axes up to a one-to-one permutation/sign change based on
      # Tucker congruence. .align_solution() implements the assignment and sign
      # correction without applying a continuous rotation.
      al <- .align_solution(L_target = target, L = Ld)
      out[[d]]  <- al$loadings
      meta[[d]] <- list(type         = "signed_tucker_congruence",
                        factor_order = as.integer(al$factor_order),
                        factor_sign  = as.numeric(al$factor_sign))
    } else if (align_unrotated == "procrustes") {
      # Continuous orthogonal alignment: Q_d minimises ||L_d Q_d - L_1||_F over
      # orthogonal Q. The transform is retained alongside the aligned loadings
      # so the analytic-SE pool can propagate the per-imputation full unrotated
      # covariance via the Kronecker identity Var(vec(L_d Q_d)) = (Q_d' (x) I_p)
      # V_d (Q_d (x) I_p).
      pr <- efa_procrustes(A = Ld, Target = target, rotation = "orthogonal")
      out[[d]]  <- .change_class(pr$loadings, "matrix")
      meta[[d]] <- list(type = "procrustes", Q = pr$T)
    }
  }

  if (align_unrotated == "signed_tucker_congruence") {
    # Put the pooled solution back into the gauge every component fit uses. The
    # same orthogonal C is applied to each imputation, so the arithmetic mean of
    # the returned list is itself canonical and the per-imputation estimates stay
    # in the gauge of the pooled estimate they are pooled against.
    C <- .efa_pooled_canonical_gauge(.average_matrices(out), target)
    # Record the rotation only when it does something. Leaving `C` absent keeps
    # the cheaper marginal-SE alignment in `.efa_pooled_analytic_pool()`, which
    # needs no covariance block, for every case that does not need re-gauging.
    if (max(abs(C - diag(k))) > 1e-8) {
      dn <- dimnames(target)
      for (d in seq_len(D)) {
        out[[d]] <- structure(out[[d]] %*% C, dimnames = dn)
        meta[[d]]$C <- C
      }
    }
  }

  list(loadings = out, meta = meta)
}

.efa_pooled_rmsea_ci <- function(chi, df, N, point, level = .90) {
  # RMSEA confidence bounds for a pooled test statistic: the same noncentral
  # chi-square inversion .chi_fit_indices() uses for a single fit (.rmsea_lambda();
  # Browne & Cudeck, 1992), at an arbitrary confidence `level` and on the pooled
  # (N - 1) scale. The chi-square supplied here should already be the pooled
  # statistic. An undefined bound is reported as NA rather than aborting.
  # `point` is the RMSEA the caller reports alongside these bounds. It is required
  # rather than re-derived here: the two callers report point estimates built on
  # different statistics and scales, so a locally recomputed one would not be the
  # value the interval has to contain.
  if (is.na(chi) || is.na(df) || is.na(N) || df <= 0 || N <= 1) {
    return(c(lower = NA_real_, upper = NA_real_))
  }

  denom <- df * (N - 1)
  if (!is.finite(denom) || denom <= 0) {
    return(c(lower = NA_real_, upper = NA_real_))
  }

  alpha <- 1 - level
  # Cap the bounds at 1 like the non-pooled .chi_fit_indices() RMSEA CI, so all
  # pooled routes (information, bootstrap, MI2S) report RMSEA in [0, 1].
  lower <- min(sqrt(.rmsea_lambda(chi, df, 1 - alpha / 2) / denom), 1)
  upper <- min(sqrt(.rmsea_lambda(chi, df, alpha / 2) / denom), 1)

  # Withhold an interval that does not contain the point estimate reported with it, by the
  # same rule .chi_fit_indices() applies to a single fit.
  if (!.rmsea_ci_contains(lower, upper, point)) {
    return(c(lower = NA_real_, upper = NA_real_))
  }

  c(lower = lower, upper = upper)
}

.efa_pooled_D2 <- function(chis, df) {
  # D2 pools complete-data chi-square statistics across imputations. The
  # returned chi value is the asymptotic chi-square approximation df * F_D2,
  # which is used downstream for RMSEA/CFI.
  chis <- chis[is.finite(chis)]
  M <- length(chis)

  if (M < 2L || !is.finite(df) || df <= 0) {
    return(NULL)
  }

  Tbar <- mean(chis)

  if (Tbar <= 0) {
    return(list(
      F = 0,
      chi = 0,
      df1 = df,
      df2 = Inf,
      p = 1,
      ARIV = 0,
      FMI = 0,
      Tbar = Tbar,
      M = M
    ))
  }

  # Average relative increase in variance (Li, Meng, Raghunathan & Rubin, 1991):
  # the between-imputation variance of the sqrt-transformed statistics is centred
  # on mean(sqrt(chi^2)) = stats::var(sqrt(chis)), not on sqrt(mean(chi^2)).
  ARIV <- (1 + 1 / M) * stats::var(sqrt(chis))

  if (ARIV <= .Machine$double.eps) {
    F_D2 <- Tbar / df
    df2 <- Inf
    p_val <- stats::pchisq(Tbar, df = df, lower.tail = FALSE)
  } else {
    F_D2 <- (Tbar / df - ((M + 1) / (M - 1)) * ARIV) / (1 + ARIV)
    df2 <- df^(-3 / M) * (M - 1) * (1 + 1 / ARIV)^2
    p_val <- if (F_D2 < 0) 1 else stats::pf(F_D2, df1 = df, df2 = df2, lower.tail = FALSE)
  }

  chi_D2 <- max(0, df * F_D2)
  FMI <- ARIV / (1 + ARIV)

  list(
    F = F_D2,
    chi = chi_D2,
    df1 = df,
    df2 = df2,
    p = p_val,
    ARIV = ARIV,
    FMI = FMI,
    Tbar = Tbar,
    M = M
  )
}

# The fit indices every pooled route reports, in the order they are reported in. The
# routes reach them differently -- the Rubin routes pool the component statistics, the
# two-stage route reads them off its single fit on the pooled inputs -- so the order is
# fixed here rather than at each assembly point, and route-specific extras (a two-stage
# fit's Fm and scaled-statistic descriptors) follow the common block.
.efa_pooled_fit_index_order <- c("chi", "df", "p_chi", "CAF", "CFI", "TLI", "RMSEA",
                                 "RMSEA_LB", "RMSEA_UB", "AIC", "BIC", "ECVI", "RMSR",
                                 "SRMR", "chi_null", "df_null", "p_null", "pool_method")

.efa_pooled_order_fit_indices <- function(fit_indices) {
  # Put a pooled fit-index list into the common order, leaving any route-specific extra
  # after it in the order it arrived in. Reordering is by name, so an unnamed list is
  # returned untouched rather than subset away to nothing.
  nms <- names(fit_indices)
  if (is.null(nms)) {
    return(fit_indices)
  }
  common <- intersect(.efa_pooled_fit_index_order, nms)
  fit_indices[c(common, setdiff(nms, common))]
}

.efa_pooled_fit_indices <- function(fits,
                                    pooled_R,
                                    residuals,
                                    RMSR,
                                    N,
                                    Ns,
                                    pool_method = "D2",
                                    rmsea_ci_level = .90) {

  fit_list <- .extract_list_object(fits, "fit_indices")
  keep <- !vapply(fit_list, is.null, logical(1))
  fit_list <- fit_list[keep]

  # Per-imputation N aligned to the kept fits. Each imputation's chi-square was
  # computed in .gof() with that imputation's own N, so the rescaling to the
  # common (N - 1) scale used by the CFI/TLI diagnostics below must use the same
  # per-imputation N.
  Ns_kept <- as.numeric(Ns)[keep]

  p_vars <- nrow(pooled_R)
  df <- NA_real_
  chis <- numeric(0)
  chis_null <- numeric(0)

  if (length(fit_list) > 0L) {
    dfs <- vapply(fit_list, function(x) {
      if (!is.null(x$df)) x$df else NA_real_
    }, numeric(1))
    finite_dfs <- unique(stats::na.omit(dfs))
    if (length(finite_dfs) > 1L) {
      cli::cli_abort("Cannot D2-pool chi-square fit because the imputation-specific dfs differ.",
                     class = "efa_pooled_chisq_df")
    }
    if (length(finite_dfs) == 1L) {
      df <- finite_dfs[[1L]]
    }

    chis <- vapply(fit_list, function(x) {
      if (!is.null(x$chi)) x$chi else NA_real_
    }, numeric(1))
    chis_null <- vapply(fit_list, function(x) {
      if (!is.null(x$chi_null)) x$chi_null else NA_real_
    }, numeric(1))
  }

  # D2-pool a vector of per-imputation chi-squares at `dfree` degrees of freedom,
  # or NULL when pooling does not apply (non-D2 request, no statistics, or a
  # degenerate df). Shared by the reported model/baseline statistics and their
  # (N - 1)-scaled diagnostic counterparts (chi_cfi / chi_null_cfi).
  .pool_chi <- function(chi_stats, dfree) {
    if (identical(pool_method, "D2") && length(chi_stats) > 0L &&
        is.finite(dfree) && dfree > 0) {
      .efa_pooled_D2(chi_stats, dfree)
    } else {
      NULL
    }
  }

  D2 <- .pool_chi(chis, df)
  if (!is.null(D2)) {
    chi <- D2$chi
    p_chi <- D2$p
  } else {
    chi <- NA_real_
    p_chi <- NA_real_
  }

  df_null <- p_vars * (p_vars - 1) / 2
  D2_null <- .pool_chi(chis_null, df_null)

  if (!is.null(D2_null)) {
    chi_null <- D2_null$chi
    p_null <- D2_null$p
  } else if (is.finite(N) && N > 1) {
    chi_null <- .null_chisq(pooled_R, N)
    p_null <- stats::pchisq(chi_null, df_null, lower.tail = FALSE)
  } else {
    chi_null <- NA_real_
    p_null <- NA_real_
  }

  # The incremental indices CFI and TLI are reported as the average of the
  # per-imputation indices rather than formed from separately pooled model/baseline
  # chi-squares; see the rationale in the function documentation (@details). Average
  # via .efa_pooled_col_means(), the same finite-mean primitive that centres the
  # bootstrap fit_indices_descriptive CI, so the reported point estimate and that CI
  # share one definition.
  cfis <- vapply(fit_list, function(x) {
    if (!is.null(x$CFI)) x$CFI else NA_real_
  }, numeric(1))
  tlis <- vapply(fit_list, function(x) {
    if (!is.null(x$TLI)) x$TLI else NA_real_
  }, numeric(1))
  # Columns are positional (col_means drops dimnames): column 1 is CFI, column 2 TLI.
  inc_means <- .efa_pooled_col_means(cbind(cfis, tlis))
  CFI <- inc_means[[1L]]
  TLI <- inc_means[[2L]]

  # The model and baseline chi-squares D2-pooled on the common (N - 1) noncentrality
  # scale, the basis on which lavaan.mi/semTools form pooled incremental indices.
  # chi_cfi is the (N - 1)-scaled pooled model statistic that drives the reported RMSEA
  # below; chi_null_cfi is exposed via mi_diagnostics so the pooled fit can be reconciled
  # against that reference (the reported CFI/TLI are averaged per imputation, not formed
  # from these). Each imputation's Bartlett-corrected chi is mapped back to the
  # (N - 1) scale -- chi = F * mult implies F * (N - 1) = chi * (N - 1) / mult, with
  # .bartlett_mult() shared with .gof()/.null_chisq() so the multiplier cannot drift
  # -- before pooling, since D2 is non-linear in a constant rescaling. The model and
  # baseline use different multipliers (the factor-count term (2q)/3 enters only the
  # model), so a small-N imputation can pass one finiteness mask (ok_m / ok_n) and
  # not the other; the model mask ok_m gates the RMSEA statistic, ok_n the chi_null_cfi
  # diagnostic. q is read from the first fit, valid because .efa_pooled_check_fits()
  # pins an equal factor count across imputations.
  chi_cfi <- NA_real_
  chi_null_cfi <- NA_real_
  if (identical(pool_method, "D2") && length(chis) > 0L && is.finite(df) && df > 0) {
    q <- ncol(as.matrix(fits[[1L]]$unrot_loadings))
    mult_model <- .bartlett_mult(Ns_kept, p_vars, q)
    mult_null <- .bartlett_mult(Ns_kept, p_vars)
    chis_cfi <- rep(NA_real_, length(chis))
    ok_m <- is.finite(Ns_kept) & is.finite(mult_model) & mult_model > 0
    chis_cfi[ok_m] <- chis[ok_m] * (Ns_kept[ok_m] - 1) / mult_model[ok_m]
    chis_null_cfi <- rep(NA_real_, length(chis_null))
    ok_n <- is.finite(Ns_kept) & is.finite(mult_null) & mult_null > 0
    chis_null_cfi[ok_n] <- chis_null[ok_n] * (Ns_kept[ok_n] - 1) / mult_null[ok_n]

    D2_cfi <- .pool_chi(chis_cfi, df)
    if (!is.null(D2_cfi)) chi_cfi <- D2_cfi$chi
    D2_null_cfi <- .pool_chi(chis_null_cfi, df_null)
    if (!is.null(D2_null_cfi)) chi_null_cfi <- D2_null_cfi$chi
  }

  # AIC, BIC and ECVI are chi-square-derived information criteria with no standard
  # interpretation once the component statistic is a scaled one (the Satorra-Bentler
  # corrected two-stage statistic of Yuan, Marshall & Bentler, 2002, and its
  # scaled-shifted variants), so .gof() withholds them from every fit that reports
  # such a statistic. A D2 pool cannot rehabilitate them, so take the components'
  # decision rather than recomputing from the pooled chi-square: a component that
  # produced a chi-square but no AIC withheld it deliberately. Reading it off the
  # returned indices covers every route at once -- the corrected `chi_scaled_type =
  # "scaled.shifted"` path, the `"uncorrected.lrt"` two-stage FIML fallback, and the
  # just-identified FIML fit, which carries no tag at all -- because all three withhold
  # the three criteria whether or not a correction was applied. A fit with no chi-square
  # at all (PAF, DWLS) is not a withholding: its pooled chi is NA, which zeroes these anyway.
  withheld_components <- any(vapply(fit_list, function(x) {
    !is.null(x$chi) && is.finite(x$chi) && (is.null(x$AIC) || !is.finite(x$AIC))
  }, logical(1)))

  if (!withheld_components && is.finite(chi) && is.finite(df) && is.finite(N) && N > 1) {
    n_params <- p_vars * (p_vars + 1) / 2 - df
    ECVI <- (chi + 2 * n_params) / (N - 1)
  } else {
    ECVI <- NA_real_
  }

  # RMSEA is built on the uncorrected (N - 1) discrepancy scale (as in
  # .chi_fit_indices()), so it uses the (N - 1)-scaled pooled model statistic chi_cfi,
  # not the Bartlett-corrected chi (which drives the reported model chi-square test).
  rmsea_denom <- df * (N - 1)
  if (is.finite(chi_cfi) && is.finite(df) && is.finite(N) &&
      df > 0 && N > 1 && is.finite(rmsea_denom) && rmsea_denom > 0) {
    # Capped at 1 as .chi_fit_indices() caps the single-fit RMSEA, so the point estimate and
    # the bounds below (also capped) stay on one scale.
    RMSEA <- min(.rmsea_point(chi_cfi, df, N), 1)
    rmsea_ci <- .efa_pooled_rmsea_ci(chi_cfi, df, N, point = RMSEA,
                                     level = rmsea_ci_level)
  } else {
    RMSEA <- NA_real_
    rmsea_ci <- c(lower = NA_real_, upper = NA_real_)
  }

  # These mirror the chi-square-derived quantities used elsewhere in EFAtools.
  # Under MI/D2 pooling they are descriptive only, not likelihood-based MI AIC/BIC,
  # and they are withheld entirely when the component fits withheld them (see
  # `withheld_components` above).
  AIC <- if (!withheld_components && is.finite(chi) && is.finite(df)) {
    chi - 2 * df
  } else {
    NA_real_
  }
  BIC <- if (!withheld_components && is.finite(chi) && is.finite(df) && is.finite(N)) {
    chi - log(N) * df
  } else {
    NA_real_
  }

  ## CAF in EFAtools is 1 - KMO(delta_hat), with diagonal set to 1.
  # Ensure symmetry of residuals by (residuals + t(residuals)) / 2. Asymmetry
  # can arise due to floating point imprecision from averaging the matrices.
  delta_hat <- (residuals + t(residuals)) / 2
  diag(delta_hat) <- 1
  CAF <- .compute_caf(delta_hat)

  ## SRMR (standardized RMR; Bentler, 1995) over the pooled residuals
  SRMR <- .srmr(residuals)

  # The fit indices and the pooling diagnostics are returned separately: the pooled
  # object carries the diagnostics in its own top-level `mi_diagnostics` slot, so the
  # `fit_indices` slot keeps the same all-scalar shape a single efa_fit() returns.
  list(
    fit_indices = .efa_pooled_order_fit_indices(list(
      chi = chi,
      df = df,
      p_chi = p_chi,
      CAF = CAF,
      CFI = CFI,
      TLI = TLI,
      RMSEA = RMSEA,
      RMSEA_LB = rmsea_ci[["lower"]],
      RMSEA_UB = rmsea_ci[["upper"]],
      AIC = AIC,
      BIC = BIC,
      ECVI = ECVI,
      RMSR = RMSR,
      SRMR = SRMR,
      chi_null = chi_null,
      df_null = df_null,
      p_null = p_null,
      pool_method = pool_method
    )),
    mi_diagnostics = list(
      D2_F = if (!is.null(D2)) D2$F else NA_real_,
      D2_df1 = if (!is.null(D2)) D2$df1 else NA_real_,
      D2_df2 = if (!is.null(D2)) D2$df2 else NA_real_,
      D2_chi_asymptotic = if (!is.null(D2)) D2$chi else NA_real_,
      chi_bar_naive = if (!is.null(D2)) D2$Tbar else if (length(chis) > 0L) mean(chis, na.rm = TRUE) else NA_real_,
      D2_null_F = if (!is.null(D2_null)) D2_null$F else NA_real_,
      D2_null_df1 = if (!is.null(D2_null)) D2_null$df1 else NA_real_,
      D2_null_df2 = if (!is.null(D2_null)) D2_null$df2 else NA_real_,
      D2_null_chi_asymptotic = if (!is.null(D2_null)) D2_null$chi else NA_real_,
      chi_null_bar_naive = if (!is.null(D2_null)) D2_null$Tbar else if (length(chis_null) > 0L) mean(chis_null, na.rm = TRUE) else NA_real_,
      ARIV = if (!is.null(D2)) D2$ARIV else NA_real_,
      FMI = if (!is.null(D2)) D2$FMI else NA_real_,
      ARIV_null = if (!is.null(D2_null)) D2_null$ARIV else NA_real_,
      FMI_null = if (!is.null(D2_null)) D2_null$FMI else NA_real_,
      # Pooled model and baseline chi-squares on the common (N - 1) noncentrality
      # scale -- the lavaan.mi/semTools basis for pooled incremental indices, kept
      # for reconciliation against that reference. The reported CFI/TLI average the
      # per-imputation indices instead; these are diagnostic and distinct from the
      # reported chi / chi_null, which keep their Bartlett multipliers.
      chi_cfi = chi_cfi,
      chi_null_cfi = chi_null_cfi,
      m = if (!is.null(D2)) D2$M else length(fits)
    )
  )
}

## -----------------------------------------------------------------------------
## Rubin pooling helpers, shared by the analytic and bootstrap routes
## -----------------------------------------------------------------------------

.efa_pooled_has_replicates <- function(fits) {
  # Bootstrap MI pooling requires the raw unrotated bootstrap loading replicates.
  # Scalar bootstrap SEs/CIs are not enough because the replicates must be
  # re-expressed in the final MI target space.
  vapply(fits, function(x) {
    !is.null(x$replicates) &&
      !is.null(x$replicates$unrot_loadings)
  }, logical(1))
}

.efa_pooled_vec <- function(M) {
  # Column-major vectorization so matrix estimates can be Rubin-pooled.
  as.vector(as.matrix(M))
}

.efa_pooled_communalities <- function(Lambda, Phi = NULL) {
  # Communalities are the diagonal of the common-factor reproduced matrix; read it
  # directly instead of forming the full p x p product:
  # diag(L L') = rowSums(L^2) and diag(L Phi L') = rowSums((L Phi) * L).
  Lambda <- as.matrix(Lambda)
  if (is.null(Phi)) {
    unname(rowSums(Lambda^2))
  } else {
    unname(rowSums((Lambda %*% as.matrix(Phi)) * Lambda))
  }
}

.efa_pooled_model_implied <- function(Lambda, Phi = NULL) {
  # Correlation-model implied matrix: common-factor part plus uniquenesses,
  # implemented by setting the diagonal of Lambda Phi Lambda' to one.
  Lambda <- as.matrix(Lambda)
  common_R <- if (is.null(Phi)) {
    Lambda %*% t(Lambda)
  } else {
    Lambda %*% as.matrix(Phi) %*% t(Lambda)
  }
  implied <- common_R
  diag(implied) <- 1
  implied
}

.efa_pooled_residual_from_solution <- function(R, Lambda, Phi = NULL) {
  # Residuals for one imputation in the same target space as its aligned solution.
  E <- as.matrix(R) - .efa_pooled_model_implied(Lambda, Phi)
  diag(E) <- 0
  E
}

.efa_pooled_make_ci <- function(est, se, df, alpha) {
  # Wald-type confidence intervals using Rubin degrees of freedom when finite.
  # `est` may be a plug-in estimate (communalities, Structure, symmetrised Phi)
  # supplied via the callers' `est_override`, in which case the interval is
  # centred on the plug-in while its half-width (se, df) comes from the pooled
  # column-mean's between/within variance; the two coincide when the plug-in
  # equals that column mean and differ only by the (typically small) plug-in gap.
  crit <- ifelse(
    is.finite(df),
    stats::qt(1 - alpha / 2, df = df),
    stats::qnorm(1 - alpha / 2)
  )
  list(
    lower = est - crit * se,
    upper = est + crit * se
  )
}


.efa_pooled_col_means <- function(x) {
  # Column means that return NA for columns without finite values.
  x <- as.matrix(x)
  if (ncol(x) < 1L) {
    return(numeric(0))
  }
  vapply(seq_len(ncol(x)), function(j) {
    vals <- x[, j]
    vals <- vals[is.finite(vals)]
    if (length(vals) < 1L) {
      NA_real_
    } else {
      mean(vals)
    }
  }, numeric(1))
}

.efa_pooled_col_vars <- function(x) {
  # Column variances that return NA for columns with fewer than two finite values.
  # Rubin pooling only needs diagonal within- and between-imputation variances,
  # so computing full covariance matrices is unnecessary and can be prohibitively
  # memory-intensive for residual matrices.
  x <- as.matrix(x)
  if (ncol(x) < 1L) {
    return(numeric(0))
  }
  if (nrow(x) < 2L) {
    return(rep(NA_real_, ncol(x)))
  }
  vapply(seq_len(ncol(x)), function(j) {
    vals <- x[, j]
    vals <- vals[is.finite(vals)]
    if (length(vals) < 2L) {
      NA_real_
    } else {
      stats::var(vals)
    }
  }, numeric(1))
}

.efa_pooled_rubin_core <- function(Qbar, Ubar, B, m, N_pool = Inf,
                                   alpha = 0.05) {
  # Element-wise Rubin-style pooling core, shared by the bootstrap (analytic SEs
  # absent) and analytic (per-imputation marginal SEs available) paths.
  #
  # Qbar  pooled point estimate per parameter (vector or matrix-as-vector).
  # Ubar  mean within-imputation variance (vector).
  # B     between-imputation variance (sample, m-1 denominator) (vector).
  # m     number of imputations.
  # N_pool complete-data sample size for Barnard-Rubin small-sample df. When
  #        non-finite, the plain Rubin (1987) df ν_old = (m-1)(1 + 1/r)^2 is
  #        used (lavaan.mi convention). When finite, the Barnard & Rubin
  #        (1999, Biometrika 86) adjustment is applied:
  #           ν_obs = ((ν_com + 1)/(ν_com + 3)) ν_com (1 - λ),  ν_com = N_pool - 1
  #           ν_BR  = 1 / (1/ν_old + 1/ν_obs)
  #        Every pooling route in the package reports the asymptotically normal
  #        reference and so passes N_pool = Inf; the finite-N_pool branch is a
  #        tested extension point, kept because it is the only place the
  #        small-sample reference is implemented. It reproduces
  #        mice::pool.scalar(n = N_pool) up to the documented lambda-vs-gamma FMI
  #        convention below, and is exercised by the test suite.
  # alpha 1 - confidence level for the Wald interval.
  #
  # Returns a list(se, ci, RIV, FMI, df). Bootstrap callers receive the same
  # df as before (N_pool defaults to Inf for byte-identical behavior).
  Tot <- Ubar + (1 + 1 / m) * B
  Tot[is.finite(Tot) & Tot < 0 & abs(Tot) < sqrt(.Machine$double.eps)] <- 0
  se <- sqrt(Tot)

  r <- rep(0, length(se))
  has_U <- is.finite(Ubar) & Ubar > 0
  r[has_U] <- ((1 + 1 / m) * B[has_U]) / Ubar[has_U]
  r[!has_U & is.finite(B) & B > 0] <- Inf
  r[!is.finite(B) & !is.infinite(r)] <- NA_real_

  df_old <- rep(Inf, length(se))
  finite_pos_r <- is.finite(r) & r > 0
  df_old[finite_pos_r] <- (m - 1) * (1 + 1 / r[finite_pos_r])^2
  df_old[is.infinite(r)] <- m - 1
  df_old[is.na(r)] <- NA_real_

  # Reported FMI is Rubin's asymptotic lambda = r / (1 + r) (the limiting fraction
  # of missing information; equals lavaan.mi's `fmi`), not mice's finite-m gamma
  # = (r + 2/(df_old + 3)) / (r + 1), which adds a small-m correction. The two
  # agree as m grows; lambda is the conventional EFA/SEM-MI reporting choice.
  fmi <- r / (1 + r)
  fmi[is.infinite(r)] <- 1
  fmi[is.nan(fmi)] <- 0

  if (is.finite(N_pool) && N_pool > 1) {
    nu_com <- N_pool - 1
    nu_obs <- ((nu_com + 1) / (nu_com + 3)) * nu_com * (1 - fmi)
    # ν_BR = 1/(1/ν_old + 1/ν_obs). When ν_old is Inf (no between-variance),
    # 1/ν_old = 0 and ν_BR collapses to ν_obs (the complete-data t reference).
    df <- 1 / (1 / df_old + 1 / nu_obs)
    df[is.infinite(df_old)] <- nu_obs[is.infinite(df_old)]
    # When fmi -> 1 (Ubar = 0, B > 0), ν_obs -> 0 and the BR formula collapses
    # to 0, sending qt(., 0) to NaN. Fall back to the plain-Rubin ν_old in that
    # regime so the Wald interval is well-defined (wide, but finite).
    no_obs <- is.finite(nu_obs) & nu_obs == 0
    df[no_obs] <- df_old[no_obs]
    df[is.na(df_old)] <- NA_real_
  } else {
    df <- df_old
  }

  ci <- .efa_pooled_make_ci(Qbar, se, df, alpha)

  list(
    se  = se,
    Tot = Tot,
    ci  = ci,
    RIV = r,
    FMI = fmi,
    df  = df
  )
}

.efa_pooled_rubin_pool <- function(q_list, boot_mat_list, alpha = 0.05,
                                   est_override = NULL) {
  # Rubin-style pooling for one parameter family.
  #
  # q_list contains one aligned point-estimate vector per imputation. boot_mat_list
  # contains aligned bootstrap replicates for estimating the within-imputation
  # variances U_d. Only diagonal variances are required for the returned SEs, CIs,
  # RIV, FMI, and df, so the implementation intentionally avoids building full
  # p x p covariance matrices for each parameter vector.
  m <- length(q_list)
  q_mat <- do.call(rbind, q_list)
  q_bar <- .efa_pooled_col_means(q_mat)

  U_diag_mat <- do.call(rbind, lapply(boot_mat_list, .efa_pooled_col_vars))
  U_bar <- colMeans(U_diag_mat)
  B_mi <- .efa_pooled_col_vars(q_mat)

  est <- if (is.null(est_override)) q_bar else est_override
  core <- .efa_pooled_rubin_core(est, U_bar, B_mi, m, N_pool = Inf, alpha = alpha)

  list(
    est = est,
    se  = core$se,
    ci  = core$ci,
    RIV = core$RIV,
    FMI = core$FMI,
    df  = core$df
  )
}

.efa_pooled_propagate_procrustes_vcov <- function(V, Q, p, k) {
  # Marginal SEs of L_d Q_d implied by the per-imputation full unrotated
  # covariance V_d (p k x p k over column-major vec(L_d)) and a fixed orthogonal
  # Q_d (k x k). Under the column-major identity vec(L_d Q_d) = (Q_d' (x) I_p)
  # vec(L_d), V_d^aligned = (Q_d' (x) I_p) V_d (Q_d (x) I_p). For element (i, j)
  # only the variance of L_d Q_d at row i is required, so the implementation
  # extracts the k x k row sub-block V_d^(i) at indices (0:(k-1)) p + i and
  # returns sqrt(diag(Q_d' V_d^(i) Q_d)) per row. Cost is O(p k^3).
  #
  # Q_d is treated as fixed (Schoenemann 1966; mirrors the lavaan / mice
  # convention of not differentiating the per-imputation orthogonal Procrustes
  # solution back through the SVD on vec(L_d)).
  #
  # NA propagation is fail-closed at the row level: if any element of the row
  # block of V or of Q is non-finite, the entire row of the returned p x k
  # marginal-SE matrix is NA so the per-element NA mask in the analytic pool
  # can blank out the affected pooled outputs.
  q_finite <- all(is.finite(Q))
  out <- matrix(NA_real_, p, k)
  if (!q_finite) return(out)
  for (i in seq_len(p)) {
    idx <- (seq_len(k) - 1L) * p + i
    block <- V[idx, idx, drop = FALSE]
    if (!all(is.finite(block))) next
    # diag(Q' (block Q)) = colSums(Q * (block Q)); avoids forming the full k x k
    # product only to keep its diagonal.
    BQ <- block %*% Q
    out[i, ] <- sqrt(pmax(colSums(Q * BQ), 0))
  }
  out
}

.efa_pooled_analytic_marginal <- function(q_mat, se_mat, m, alpha,
                                          est_override = NULL) {
  # Rubin pool of one parameter family from aligned per-imputation point estimates
  # (rows of `q_mat`) and aligned per-imputation marginal SEs (rows of `se_mat`, so
  # the within-imputation variance U_d is `se_mat^2`). Only diagonal variances are
  # needed. Fail-closed: any per-imputation NA in q or se for an element blanks the
  # pooled outputs (SE, CI, RIV, FMI, df) at that element. `est_override` supplies
  # the pooled point estimate for plug-in quantities (communalities, structure)
  # whose returned estimate is computed from the pooled solution, not the mean of
  # the per-imputation values.
  Qbar <- .efa_pooled_col_means(q_mat)
  Ubar <- .efa_pooled_col_means(se_mat^2)
  B    <- .efa_pooled_col_vars(q_mat)

  na <- apply(is.na(q_mat) | is.na(se_mat), 2, any)
  Qbar[na] <- NA_real_; Ubar[na] <- NA_real_; B[na] <- NA_real_

  est <- if (is.null(est_override)) Qbar else est_override
  # Plain Rubin (1987) df (N_pool = Inf): analytic information-matrix loadings are
  # asymptotically normal (Wald-z), so the complete-data reference is normal
  # (nu_com = Inf) and the Barnard-Rubin small-sample adjustment does not apply -- it
  # reduces to plain Rubin. This matches lavaan.mi, which reports df -> Inf for
  # asymptotically-normal SEM parameters.
  .efa_pooled_rubin_core(est, Ubar, B, m, N_pool = Inf, alpha = alpha)
}

.efa_pooled_assemble_family <- function(pool, reshape, method = NULL) {
  # Reshape one Rubin-pooled family's SE, Wald CI bounds, and MI diagnostics with a
  # single reshape function (loading matrix, uniqueness/communality vector, or
  # vech-to-symmetric). `method` tags the gauge-alignment provenance when supplied.
  mi <- list(RIV = reshape(pool$RIV), FMI = reshape(pool$FMI), df = reshape(pool$df))
  if (!is.null(method)) mi$method <- method
  list(
    SE = reshape(pool$se),
    CI = list(lower = reshape(pool$ci$lower), upper = reshape(pool$ci$upper)),
    MI = mi
  )
}

.efa_pooled_rubin_result <- function(pool, reshape, dimnames = NULL) {
  # Convert vectorized Rubin output (estimate, SE, and Wald CI bounds) back into one
  # parameter family's own shape with a single `reshape` function -- a rectangular
  # matrix or a vech-to-symmetric expansion -- the way .efa_pooled_assemble_family()
  # does for the analytic route's SE/CI/MI triple.
  shape <- function(v) {
    out <- reshape(v)
    if (!is.null(dimnames)) dimnames(out) <- dimnames
    out
  }

  list(est = shape(pool$est), se = shape(pool$se),
       ci = list(lower = shape(pool$ci$lower), upper = shape(pool$ci$upper)))
}

# Tests for efa_mi(): structure of the returned object, pooling math,
# classed conditions, bootstrap/MI pooling, and the print/format methods.

cormat <- test_models$baseline$cormat
p_vars <- ncol(cormat)
cormat_list <- list(cormat, cormat, cormat)

pooled_obl <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                         rotation = "promax")
pooled_orth <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                          rotation = "varimax")

# The single-solution counterpart of `pooled_obl`: pooling identical imputations must
# reproduce it, so several blocks below compare against the same fit. PAF extraction and
# promax rotation are deterministic, so one fit serves all of them.
single_obl <- EFA(cormat, n_factors = 3, N = 500, method = "PAF", rotation = "promax")

set.seed(42)
grips_list <- lapply(1:3, function(i) {
  GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), replace = TRUE), ]
})
pooled_none <- suppressMessages(
  efa_mi(grips_list, n_factors = 1, estimator = "ML")
)

test_that("efa_mi returns a well-formed pooled object", {
  expect_s3_class(pooled_obl, c("efa_mi", "EFA_POOLED", "efa", "EFA"), exact = TRUE)

  core <- c("h2", "unrot_loadings", "vars_accounted", "fit_indices",
            "model_implied_R", "residuals", "orig_R", "settings", "fits",
            "alignment", "mi_diagnostics", "rot_loadings",
            "vars_accounted_rot", "Phi", "Structure")
  expect_true(all(core %in% names(pooled_obl)))

  expect_type(pooled_obl$h2, "double")
  expect_length(pooled_obl$h2, p_vars)
  expect_named(pooled_obl$h2)

  expect_s3_class(pooled_obl$unrot_loadings, "LOADINGS")
  expect_s3_class(pooled_obl$rot_loadings, "LOADINGS")
  expect_s3_class(pooled_obl$Structure, "LOADINGS")
  expect_identical(dim(unclass(pooled_obl$rot_loadings)), c(p_vars, 3L))

  expect_identical(dim(pooled_obl$Phi), c(3L, 3L))
  expect_equal(pooled_obl$Phi, t(pooled_obl$Phi))

  expect_equal(unname(diag(pooled_obl$model_implied_R)), rep(1, p_vars))
  expect_equal(unname(diag(pooled_obl$residuals)), rep(0, p_vars))

  expect_length(pooled_obl$fits, 3)
  expect_true(all(vapply(pooled_obl$fits, inherits, logical(1), "EFA")))
  expect_true(isTRUE(pooled_obl$alignment$converged))

  expect_type(pooled_obl$fit_indices, "list")
  expect_true(is.finite(pooled_obl$fit_indices$CAF))
  expect_true(is.finite(pooled_obl$fit_indices$RMSR))
})

test_that("pooled CAF reproduces the single-solution CAF on identical imputations", {
  # Pooling identical correlation matrices reproduces the single EFA solution, so
  # the pooled CAF (computed on the residual matrix with a unit diagonal) must
  # equal the CAF that EFA() reports for that solution.
  expect_equal(pooled_obl$fit_indices$CAF, single_obl$fit_indices$CAF, tolerance = 1e-6)
  expect_gt(pooled_obl$fit_indices$CAF, 0)
  expect_lt(pooled_obl$fit_indices$CAF, 1)
})

test_that("efa_mi records the pooling settings", {
  s <- pooled_obl$settings
  expect_true(s$pooled)
  expect_identical(s$n_imputations, 3L)
  expect_identical(s$target_method, "first_target")
  expect_identical(s$align_unrotated, "signed_tucker_congruence")
  expect_identical(s$fit_pool_method, "D2")
  expect_equal(s$p, 0.05)
  expect_equal(s$ci, 0.95)
  # no bootstrap arrays were available, so the pooled object must not claim SEs;
  # neither the current names nor the historic flattened ones may appear.
  expect_identical(s$se, "none")
  expect_false(any(c("SE", "CI", "replicates", "MI",
                     "vcov_unrot_loadings", "Gamma",
                     "boot.SE", "boot.CI", "boot.arrays", "boot.MI") %in%
                     names(pooled_obl)))
})

test_that("efa_mi records the component fits' admissibility", {
  adm <- pooled_obl$mi_admissibility
  expect_identical(adm$m, 3L)
  expect_identical(adm$n_heywood_items, c(0L, 0L, 0L))
  expect_identical(adm$heywood_imputations, integer(0))
  expect_identical(adm$nonconverged, integer(0))
  expect_true(all(is.finite(adm$iter)))

  # The record is read off the component fits, so an inadmissible or non-converged
  # component is captured whatever the pooled matrix looks like.
  affected <- pooled_obl$fits
  affected[[1]]$heywood <- c(V1 = 1L, V5 = 5L)
  affected[[3]]$heywood <- c(V3 = 3L)
  affected[[3]]$convergence <- 1L
  adm_affected <- .efa_pooled_admissibility(affected)
  expect_identical(adm_affected$n_heywood_items, c(2L, 0L, 1L))
  expect_identical(adm_affected$heywood_imputations, c(1L, 3L))
  expect_identical(adm_affected$nonconverged, 3L)
})

test_that("summary() reports component Heywood cases behind a clean pooled count", {
  # Averaging aligned solutions pulls boundary communalities back inside the admissible
  # range, so a pool of improper solutions can have a proper pooled matrix. The pooled
  # count alone then reads as an all-clear for the analysis.
  local_reproducible_output()

  affected <- pooled_obl
  affected$fits[[1]]$heywood <- c(V1 = 1L, V5 = 5L)
  affected$fits[[3]]$heywood <- c(V3 = 3L)
  affected$mi_admissibility <- .efa_pooled_admissibility(affected$fits)

  heywood_line <- function(x) {
    grep("^Heywood cases:", cli::ansi_strip(format(summary(x))), value = TRUE)
  }
  # Three flags, not three distinct variables: a variable flagged in two imputations
  # counts once per imputation.
  expect_identical(heywood_line(affected),
                   "Heywood cases: 0 pooled (3 flags across 2 of 3 imputations)")

  # A pool whose component fits were all proper keeps the unqualified count.
  expect_identical(heywood_line(pooled_obl), "Heywood cases: 0")
})

test_that("a pool of improper component fits does not report zero Heywood cases", {
  skip_on_cran()
  local_reproducible_output()

  # One small imputation among larger ones: several component fits hit the uniqueness
  # boundary while their average stays well inside the admissible range.
  withr::local_seed(3)
  imps <- lapply(1:6, function(i) {
    n <- if (i == 1L) 45L else 300L
    UPPS_raw[sample.int(nrow(UPPS_raw), n, replace = TRUE), 1:12]
  })
  pooled <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 4, estimator = "ML", rotation = "varimax")
  ))

  # precondition: the components really are improper and the pooled matrix is not
  expect_gt(sum(pooled$mi_admissibility$n_heywood_items), 0)
  expect_lt(max(pooled$h2), 1)

  expect_match(
    grep("^Heywood cases:", cli::ansi_strip(format(summary(pooled))), value = TRUE),
    "imputations)", fixed = TRUE
  )
})

test_that("rotation variants include exactly the applicable components", {
  # orthogonal: rotated loadings but no factor intercorrelations
  expect_s3_class(pooled_orth$rot_loadings, "LOADINGS")
  expect_true(!is.null(pooled_orth$vars_accounted_rot))
  expect_null(pooled_orth$Phi)
  expect_null(pooled_orth$Structure)

  # unrotated: no rotated components at all
  expect_null(pooled_none$rot_loadings)
  expect_null(pooled_none$vars_accounted_rot)
  expect_null(pooled_none$Phi)
  expect_null(pooled_none$Structure)
  expect_null(pooled_none$alignment)
})

test_that("alignment variants produce well-formed pooled objects", {
  # first_target: align every imputation to the first rotated solution rather
  # than to an iteratively updated consensus target
  ft <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                   rotation = "promax", target_method = "first_target")
  expect_s3_class(ft, c("efa_mi", "EFA_POOLED", "efa", "EFA"), exact = TRUE)
  expect_identical(ft$settings$target_method, "first_target")
  expect_identical(ft$alignment$method, "first_target")
  expect_s3_class(ft$rot_loadings, "LOADINGS")
  expect_identical(dim(ft$Phi), c(3L, 3L))

  # align_unrotated = "procrustes": orthogonal Procrustes of the unrotated axes
  pr <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                   rotation = "none", align_unrotated = "procrustes")
  expect_identical(pr$settings$align_unrotated, "procrustes")
  expect_s3_class(pr$unrot_loadings, "LOADINGS")

  # align_unrotated = "none": average the unrotated loadings as returned. With
  # identical imputations the pooled result must equal the single fit.
  nn <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                   rotation = "none", align_unrotated = "none")
  expect_identical(nn$settings$align_unrotated, "none")
  single <- EFA(cormat, n_factors = 3, N = 500, method = "PAF",
                rotation = "none")
  expect_equal(unclass(nn$unrot_loadings), unclass(single$unrot_loadings),
               tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("target_method = 'consensus' pools a solution that does not depend on the imputation order", {
  skip_on_cran()

  # One atypical imputation among larger ones. The GPA iteration keeps the gauge of the
  # solution it starts from, so this is exactly the setting in which the anchor decides
  # the pooled orientation.
  withr::local_seed(11)
  imps <- lapply(1:4, function(i) {
    n <- if (i == 1L) 60L else 250L
    GRiPS_raw[sample.int(nrow(GRiPS_raw), n, replace = TRUE), ]
  })
  fit_mi <- function(dl, tm, ...) {
    suppressWarnings(suppressMessages(
      efa_mi(dl, n_factors = 2, estimator = "ML", rotation = "varimax",
             target_method = tm, ...)
    ))
  }
  perm <- c(3L, 1L, 4L, 2L)

  consensus <- fit_mi(imps, "consensus")
  consensus_perm <- fit_mi(imps[perm], "consensus")
  expect_equal(unclass(consensus_perm$rot_loadings), unclass(consensus$rot_loadings),
               tolerance = 1e-8)
  expect_equal(as.numeric(consensus_perm$h2), as.numeric(consensus$h2),
               tolerance = 1e-8)

  # The GPA record is returned and the iteration is started at the medoid rotated
  # solution, which is a property of the set rather than of the list order.
  expect_true(isTRUE(consensus$alignment$converged))
  expect_length(consensus$alignment$aligned_loadings, length(imps))
  expect_equal(
    consensus$alignment$start,
    .efa_pooled_medoid_anchor(lapply(consensus$fits,
                                     function(f) unclass(f$rot_loadings)))
  )

  # The fixture is discriminating: the first-imputation anchor does move under the same
  # permutation, and by more than a relabelling of the columns.
  first_target <- fit_mi(imps, "first_target")
  first_target_perm <- fit_mi(imps[perm], "first_target")
  matched <- .align_solution(L_target = unclass(first_target$rot_loadings),
                             L = unclass(first_target_perm$rot_loadings))$loadings
  expect_gt(max(abs(unclass(matched) - unclass(first_target$rot_loadings))), 1e-3)

  # ... because "first_target" is anchored on the first imputation by definition.
  expect_equal(unclass(first_target$alignment$target),
               unclass(first_target$fits[[1]]$rot_loadings), tolerance = 1e-12)

  # An explicit start overrides the medoid anchor, and is honoured.
  from_first <- fit_mi(imps, "consensus", consensus_args = list(start = 1))
  expect_equal(from_first$alignment$start, 1)
})

test_that("procrustes_args forwards algorithm controls and nothing else", {
  target <- matrix(0, p_vars, 3)
  reserved <- list(list(A = target), list(Target = target),
                   list(rotation = "orthogonal"), list(S = diag(3)),
                   # partially spelled names are matched the way do.call() matches them
                   list(Tar = target))
  for (args in reserved) {
    expect_error(
      efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
             rotation = "promax", procrustes_args = args),
      class = "efa_pooled_bad_procrustes_args"
    )
  }

  # An unknown name is rejected here rather than surfacing as an unused-argument error
  # from inside the alignment, and an unnamed element cannot be matched at all.
  expect_error(
    efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
           rotation = "promax", procrustes_args = list(oblique_maxitt = 5)),
    class = "efa_pooled_bad_procrustes_args"
  )
  expect_error(
    efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
           rotation = "promax", procrustes_args = list(50)),
    class = "efa_pooled_bad_procrustes_args"
  )

  # A genuine control reaches efa_procrustes(): capping the oblique solver at one
  # iteration stops the alignment exactly there.
  tuned <- suppressWarnings(efa_mi(
    cormat_list, n_factors = 3, N = 500, estimator = "PAF", rotation = "promax",
    procrustes_args = list(oblique_maxit = 1L)
  ))
  expect_identical(
    vapply(tuned$alignment$target_rotations[-1L],
           function(x) as.numeric(x$iterations), numeric(1)),
    rep(1, length(cormat_list) - 1L)
  )
  expect_false(tuned$alignment$converged)
})

test_that("pooling identical imputations reproduces the single fit", {
  # unrotated alignment is a pure sign/permutation step, so it must be exact
  expect_equal(unclass(pooled_obl$unrot_loadings),
               unclass(single_obl$unrot_loadings),
               tolerance = 1e-10, ignore_attr = TRUE)
  # rotated solutions are re-derived by oblique Procrustes alignment, which
  # recovers the promax solution up to the solver tolerance
  expect_equal(unclass(pooled_obl$rot_loadings),
               unclass(single_obl$rot_loadings),
               tolerance = 1e-4, ignore_attr = TRUE)
  expect_equal(pooled_obl$Phi, single_obl$Phi, tolerance = 1e-4,
               ignore_attr = TRUE)
  expect_equal(pooled_obl$orig_R, cormat, ignore_attr = TRUE)
})

test_that("pooled components are internally consistent", {
  L <- unclass(pooled_obl$rot_loadings)
  Phi <- pooled_obl$Phi

  expect_equal(unclass(pooled_obl$Structure), L %*% Phi, ignore_attr = TRUE)
  expect_equal(unname(pooled_obl$h2), unname(diag(L %*% Phi %*% t(L))))

  res <- pooled_obl$orig_R - pooled_obl$model_implied_R
  diag(res) <- 0
  expect_equal(pooled_obl$residuals, res, ignore_attr = TRUE)
})

test_that("RMSR is the root mean square of the unique off-diagonal residuals", {
  for (pooled in list(pooled_obl, pooled_orth, pooled_none)) {
    res <- pooled$residuals
    expect_equal(pooled$fit_indices$RMSR,
                 sqrt(mean(res[upper.tri(res)]^2)))
    # The pooled residual matrix is symmetric, which is why counting each residual
    # pair twice instead of once cannot change that mean square.
    expect_equal(res, t(res), ignore_attr = TRUE)
    expect_equal(pooled$fit_indices$RMSR,
                 sqrt(mean(res[row(res) != col(res)]^2)))
  }
})

test_that("the deprecated rmsr_upper is accepted, warned about, and ignored", {
  expect_warning(
    deprecated <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                         rotation = "promax", rmsr_upper = TRUE),
    class = "lifecycle_warning_deprecated"
  )
  expect_equal(deprecated$fit_indices$RMSR, pooled_obl$fit_indices$RMSR)

  expect_warning(
    flipped <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "PAF",
                      rotation = "promax", rmsr_upper = FALSE),
    class = "lifecycle_warning_deprecated"
  )
  expect_equal(flipped$fit_indices$RMSR, pooled_obl$fit_indices$RMSR)

  # It no longer records a setting, and the frozen wrapper still accepts it without
  # passing it on, so legacy calls stay silent.
  expect_null(pooled_obl$settings$rmsr_upper)
  expect_no_warning(
    legacy <- EFA_POOLED(cormat_list, n_factors = 3, N = 500, method = "PAF",
                         rotation = "promax", rmsr_upper = TRUE),
    class = "lifecycle_warning_deprecated"
  )
  expect_equal(legacy$fit_indices$RMSR, pooled_obl$fit_indices$RMSR)
})

test_that("pooled fit indices D2-pool the imputation chi-squares", {
  fi <- pooled_none$fit_indices
  md <- pooled_none$mi_diagnostics
  expect_identical(fi$pool_method, "D2")
  # The reported names and their order are pinned by one shared definition, so a new
  # index cannot be added to the assembled list without also being placed there.
  expect_identical(names(fi), EFAtools:::.efa_pooled_fit_index_order)
  expect_equal(fi$df, pooled_none$fits[[1]]$fit_indices$df)
  expect_true(is.finite(fi$chi) && fi$chi >= 0)
  expect_true(is.finite(fi$p_chi))
  expect_true(is.finite(fi$RMSR))

  # The pooled set reports CFI/TLI as the average of the per-imputation indices
  # (kept in range and consistent with the component fits; verified in
  # test-EFA_POOLED-cfi-scale.R). The separately pooled (N - 1)-scale model and
  # baseline noncentralities remain exposed via mi_diagnostics chi_cfi /
  # chi_null_cfi for reconciliation against lavaan.mi, while ECVI keeps the
  # reported pooled chi-square. No mislabeled Fm.
  expect_true(is.finite(fi$TLI))
  expect_true(is.finite(fi$ECVI))
  expect_null(fi$Fm)
  expect_true(is.finite(md$chi_cfi))
  expect_true(is.finite(md$chi_null_cfi))
  N_used <- pooled_none$settings$N
  n_vars <- ncol(pooled_none$orig_R)
  n_params <- n_vars * (n_vars + 1) / 2 - fi$df
  expect_equal(fi$ECVI, (fi$chi + 2 * n_params) / (N_used - 1))

  # The diagnostics live in the top-level slot only; `fit_indices` holds scalars, the
  # same shape a single efa_fit() returns.
  expect_null(fi$mi_diagnostics)
  expect_true(all(vapply(fi, function(v) {
    (is.numeric(v) || is.character(v)) && length(v) == 1L
  }, logical(1))))

  expect_identical(md$m, 3L)
  expect_gte(md$ARIV, 0)
  expect_gte(md$FMI, 0)
  expect_lte(md$FMI, 1)

  # the pooled observed correlation matrix is the mean across imputations
  R_mean <- Reduce(`+`, lapply(pooled_none$fits, function(f) f$orig_R)) / 3
  expect_equal(pooled_none$orig_R, R_mean, ignore_attr = TRUE)
})

test_that(".efa_pooled_D2 ARIV matches the Li et al. (1991) formula", {
  # The average relative increase in variance is the between-imputation variance
  # of the sqrt-transformed statistics, i.e. (1 + 1/M) * var(sqrt(chi^2)).
  chis <- c(30, 35, 50)
  d2 <- .efa_pooled_D2(chis, df = 20)
  expect_equal(d2$ARIV, (1 + 1 / length(chis)) * stats::var(sqrt(chis)))
  expect_equal(d2$FMI, d2$ARIV / (1 + d2$ARIV))
})

test_that("a negative D2 statistic is reported as such, and the pooled fit stays in range", {
  # The pooling statistic is returned unfloored, so it can go negative when the
  # between-imputation variability of the component statistics exceeds the pooled
  # discrepancy; that is a diagnostic of the pool. The fit it feeds must not inherit
  # it: the pooled chi-square is floored at zero and its p-value is 1.
  d2 <- .efa_pooled_D2(c(10, 20, 40, 60, 80), df = 33)
  expect_lt(d2$F, 0)
  expect_identical(d2$chi, 0)
  expect_identical(d2$p, 1)
  expect_gt(d2$ARIV, 0)

  # An ordinary pool keeps a positive statistic and a chi-square derived from it.
  ok <- .efa_pooled_D2(c(30, 35, 50), df = 20)
  expect_gt(ok$F, 0)
  expect_equal(ok$chi, ok$df1 * ok$F)
})

test_that("seed and b_boot in the dots govern the whole pooled call", {
  skip_on_cran()

  withr::local_seed(20260812)
  boot_list <- lapply(1:2, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), 200, replace = TRUE), 1:5]
  })
  run <- function(seed) {
    suppressWarnings(suppressMessages(
      efa_mi(boot_list, n_factors = 1, estimator = "ML", se = "np-boot",
             b_boot = 4, seed = seed)
    ))
  }

  a <- run(99)
  b <- run(99)
  expect_identical(a$SE$unrot_loadings, b$SE$unrot_loadings)
  # b_boot is the per-imputation replicate count and is recorded as such
  expect_equal(a$settings$b_boot, 4)
  expect_identical(dim(a$replicates$unrot_loadings[[1]])[1], 4L)

  # A different seed moves the bootstrap SEs, so the reproducibility above is not
  # vacuous ...
  expect_false(isTRUE(all.equal(a$SE$unrot_loadings, run(100)$SE$unrot_loadings)))

  # ... and the caller's own random stream is left where it was.
  set.seed(7)
  expected <- stats::runif(1)
  set.seed(7)
  invisible(run(99))
  expect_identical(stats::runif(1), expected)
})

test_that("efa_mi validates its arguments with classed conditions", {
  expect_error(
    efa_mi(cormat_list, p = 0, n_factors = 3, N = 500, estimator = "PAF",
               rotation = "none"),
    class = "efa_pooled_bad_p"
  )
  expect_error(
    efa_mi(cormat_list, rmsea_ci_level = 1, n_factors = 3, N = 500,
               estimator = "PAF", rotation = "none"),
    class = "efa_pooled_bad_ci_level"
  )
  expect_warning(
    efa_mi(cormat_list, ci = .8, n_factors = 3, N = 500, estimator = "PAF",
               rotation = "none"),
    class = "efa_pooled_ci_ignored"
  )
  # Too few imputations signals the documented condition class, not a bare
  # assertion error, and does so before any component fit is run.
  expect_error(
    efa_mi(list(cormat), n_factors = 3, N = 500, estimator = "PAF",
               rotation = "none"),
    class = "efa_pooled_min_fits"
  )
  expect_error(
    efa_mi(list(), n_factors = 3, N = 500, estimator = "PAF", rotation = "none"),
    class = "efa_pooled_min_fits"
  )
  # A mids object is itself a list, so without the dedicated check it would fail deep
  # inside the per-dataset assertion instead of naming the conversion.
  expect_error(
    efa_mi(structure(list(data = cormat, m = 2L), class = "mids"),
               n_factors = 3, N = 500, estimator = "PAF", rotation = "none"),
    class = "efa_pooled_mids_input"
  )
})

test_that("the raw-data correlation note is given once, not once per imputation", {
  # The imputations are versions of one data set, so the note the per-imputation efa_fit()
  # calls raise states one fact about the input. It stays a catchable classed condition.
  n_notes <- 0L
  withCallingHandlers(
    efa_mi(grips_list, n_factors = 1, estimator = "ML"),
    efa_cor_from_data = function(cnd) {
      n_notes <<- n_notes + 1L
      invokeRestart("muffleMessage")
    },
    message = function(cnd) invokeRestart("muffleMessage")
  )
  expect_identical(n_notes, 1L)
})

test_that("a failing component fit names its imputation and keeps the original condition", {
  local_reproducible_output()

  mk <- function(seed) {
    set.seed(seed)
    f <- stats::rnorm(60)
    X <- outer(f, rep(0.7, 4)) + matrix(stats::rnorm(240, sd = 0.7), 60, 4)
    colnames(X) <- paste0("V", seq_len(4))
    X
  }
  raw <- lapply(1:3, mk)
  # A constant column in one imputation only: the pool is all-or-nothing, so the whole
  # call ends, and without the index the user has to bisect `data_list` by hand.
  raw[[2]][, 3] <- 1

  err <- expect_error(
    efa_mi(raw, n_factors = 1, estimator = "PAF", rotation = "none"),
    class = "efa_pooled_fit_failed"
  )
  # the component condition is chained, not replaced, so its own diagnosis survives
  expect_s3_class(err$parent, "efa_cor_uncomputable")

  expect_snapshot(
    error = TRUE,
    efa_mi(raw, n_factors = 1, estimator = "PAF", rotation = "none")
  )
})

test_that("efa_mi rejects non-conformable imputations", {
  expect_error(
    efa_mi(list(cormat, cormat[1:(p_vars - 1), 1:(p_vars - 1)]),
               n_factors = 3, N = 500, estimator = "PAF", rotation = "none"),
    class = "efa_pooled_dim_mismatch"
  )

  renamed <- cormat
  dimnames(renamed) <- list(paste0("X", seq_len(p_vars)),
                            paste0("X", seq_len(p_vars)))
  expect_error(
    efa_mi(list(cormat, renamed), n_factors = 3, N = 500, estimator = "PAF",
               rotation = "none"),
    class = "efa_pooled_var_mismatch"
  )
})

test_that("efa_mi warns when imputations have different N", {
  expect_warning(
    suppressMessages(
      efa_mi(list(GRiPS_raw[1:300, ], GRiPS_raw[1:400, ]), n_factors = 1,
                 estimator = "PAF", rotation = "none")
    ),
    class = "efa_pooled_unequal_n"
  )
})

test_that("efa_mi warns when N cannot be recovered", {
  # correlation-matrix input carries no N; a method that needs N for chi-square
  # fit cannot compute chi-square-based indices. The component ML fits also warn
  # about the missing N; muffle those so only the pooled class is asserted.
  suppressWarnings(
    expect_warning(
      suppressMessages(
        efa_mi(cormat_list, n_factors = 3, estimator = "ML", rotation = "none")
      ),
      class = "efa_pooled_no_n"
    )
  )

  # N recoverable for the raw-data imputation but not the correlation matrix
  expect_warning(
    suppressMessages(
      efa_mi(list(GRiPS_raw, stats::cor(GRiPS_raw)), n_factors = 1,
                 estimator = "PAF", rotation = "none")
    ),
    class = "efa_pooled_partial_n"
  )
})

test_that("bootstrap arrays are pooled into MI SEs and CIs", {
  skip_on_cran()
  local_reproducible_output()

  set.seed(1)
  boot_list <- lapply(1:2, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), 250, replace = TRUE), ]
  })
  pooled_boot <- suppressMessages(
    efa_mi(boot_list, n_factors = 1, estimator = "ML",
               se = "np-boot", b_boot = 6)
  )

  expect_true(all(c("SE", "CI", "replicates", "MI",
                    "standardized_residuals") %in% names(pooled_boot)))
  # the historic flattened slot names must not leak in alongside the new ones
  expect_false(any(c("boot.SE", "boot.CI", "boot.arrays", "boot.MI") %in%
                     names(pooled_boot)))

  L <- unclass(pooled_boot$unrot_loadings)
  se <- pooled_boot$SE$unrot_loadings
  expect_identical(dim(se), dim(L))
  # the SEs must actually be computed (finite), not silently all-NA, and the
  # bootstrap must have produced real variation (positive SE somewhere)
  expect_true(all(is.finite(se)))
  expect_true(all(se >= 0))
  expect_true(any(se > 0))

  # Wald-type MI intervals: finite, correctly ordered, and centred on the
  # pooled point estimate. The centring check is the real test (it fails if the
  # interval is built around the wrong quantity or the bounds are swapped);
  # asserting only lower <= L <= upper would hold for any symmetric interval.
  ci <- pooled_boot$CI$unrot_loadings
  expect_true(all(is.finite(ci$lower) & is.finite(ci$upper)))
  expect_true(all(ci$upper >= ci$lower))
  expect_equal((ci$lower + ci$upper) / 2, L, ignore_attr = TRUE)

  expect_equal(pooled_boot$settings$b_boot, 6)
  expect_identical(pooled_boot$settings$se, "np-boot")

  # Communality SEs/CIs use efa_fit()'s name on this route too, with `h2` kept as an
  # alias of it; the MI diagnostics carry the canonical name alone, so each family
  # enters the printed FMI/RIV summary once.
  expect_false(is.null(pooled_boot$SE$communalities))
  expect_identical(pooled_boot$SE$h2, pooled_boot$SE$communalities)
  expect_identical(pooled_boot$CI$h2, pooled_boot$CI$communalities)
  expect_false(is.null(pooled_boot$MI$communalities))
  expect_null(pooled_boot$MI$h2)

  # FMIs must be computed (at least some finite) and in [0, 1]; an all-NA vector
  # would pass a bare all(..., na.rm = TRUE) range check vacuously.
  fmi <- pooled_boot$MI$unrot_loadings$FMI
  expect_true(any(is.finite(fmi)))
  expect_true(all(fmi[is.finite(fmi)] >= 0 & fmi[is.finite(fmi)] <= 1))

  expect_identical(dim(pooled_boot$standardized_residuals),
                   dim(pooled_boot$residuals))

  fit_ci <- pooled_boot$CI$fit_indices_descriptive
  expect_true(all(is.finite(fit_ci$lower[c("SRMR", "TLI", "ECVI")])))
  expect_true(all(is.finite(fit_ci$upper[c("SRMR", "TLI", "ECVI")])))
  expect_true("RMSR" %in% names(pooled_boot$fit_indices))

  summary_lines <- cli::ansi_strip(format(summary(pooled_boot)))
  expect_false(any(grepl("^RMSR\\b", summary_lines)))
  expect_true(any(grepl("^SRMR \\[95% bootstrap/MI-CI\\]:", summary_lines)))
  # The incremental indices carry the averaged-over-imputations label ahead of the CI tag
  expect_true(any(grepl("^TLI \\(avg\\. over imputations\\) \\[95% bootstrap/MI-CI\\]:",
                        summary_lines)))
  expect_true(any(grepl("^ECVI \\[95% bootstrap/MI-CI\\]:", summary_lines)))

  # summary() additionally shows the MI uncertainty summary
  expect_snapshot(print(summary(pooled_boot)), transform = scrub_num)
})

test_that("oblique bootstrap pooling produces rotated SEs, CIs, and Phi", {
  skip_on_cran()

  set.seed(2)
  boot_list <- lapply(1:2, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), 250, replace = TRUE), ]
  })
  pooled_boot <- suppressWarnings(suppressMessages(
    efa_mi(boot_list, n_factors = 2, estimator = "PAF", rotation = "promax",
               se = "np-boot", b_boot = 6)
  ))

  # the oblique branch pools rotated loadings, factor correlations, and
  # structure coefficients in addition to the unrotated quantities
  for (comp in c("rot_loadings", "Phi", "Structure")) {
    se <- pooled_boot$SE[[comp]]
    ci <- pooled_boot$CI[[comp]]
    expect_false(is.null(se))
    expect_true(all(is.finite(se)))
    expect_true(all(se >= 0))
    expect_true(all(is.finite(ci$lower) & is.finite(ci$upper)))
    expect_true(all(ci$upper >= ci$lower))
  }

  expect_identical(dim(pooled_boot$SE$Phi), dim(as.matrix(pooled_boot$Phi)))
  expect_identical(dim(pooled_boot$SE$Structure),
                   dim(unclass(pooled_boot$Structure)))

  fmi <- pooled_boot$MI$Phi$FMI
  expect_true(any(is.finite(fmi)))
  expect_true(all(fmi[is.finite(fmi)] >= 0 & fmi[is.finite(fmi)] <= 1))
})

test_that("a failed bootstrap replicate is skipped, not fatal", {
  # A component EFA NA-fills a bootstrap replicate it could not fit; the pooled
  # bootstrap must skip that replicate (classed warning) and still produce finite
  # SEs from the valid ones, rather than aborting on the NA in the alignment step.
  withr::local_seed(1)
  p <- 3L; k <- 1L; B <- 4L; m <- 2L
  mk_arr <- function(na = NULL) {
    a <- array(stats::rnorm(p * k * B, 0.6, 0.05), dim = c(p, k, B))
    if (!is.null(na)) a[, , na] <- NA_real_
    a
  }
  mk_res <- function(na = NULL) {
    a <- array(stats::rnorm(p * p * B, 0, 0.02), dim = c(p, p, B))
    if (!is.null(na)) a[, , na] <- NA_real_
    a
  }
  # bootstrap fit-index arrays so the Rubin-Wald descriptive fit path also runs
  mk_fit <- function(na = NULL) {
    f <- matrix(stats::rnorm(B * 2, 1, 0.1), nrow = B,
                dimnames = list(NULL, c("chi", "CFI")))
    if (!is.null(na)) f[na, ] <- NA_real_
    f
  }
  L <- matrix(0.6, p, k)
  fits <- list(
    list(fit_indices = list(chi = 1, CFI = 1),
         replicates = list(unrot_loadings = mk_arr(na = 2), residuals = mk_res(na = 2),
                           fit_indices = mk_fit(na = 2))),
    list(fit_indices = list(chi = 1, CFI = 1),
         replicates = list(unrot_loadings = mk_arr(),       residuals = mk_res(),
                           fit_indices = mk_fit()))
  )
  orig_R <- replicate(m, { R <- matrix(0.4, p, p); diag(R) <- 1; R }, simplify = FALSE)
  args <- list(
    fits = fits, orig_R_list = orig_R,
    unrot_loadings_aligned = replicate(m, L, simplify = FALSE),
    mean_unrot_loadings = L, rotation_type = "none",
    align_unrotated = "signed_tucker_congruence",
    h2 = rep(0.36, p), residuals = matrix(0, p, p)
  )

  expect_warning(
    pooled <- do.call(.efa_pooled_bootstrap_pool, args),
    class = "efa_pooled_boot_failed"
  )
  expect_true(all(is.finite(pooled$SE$unrot_loadings)))
  # the Rubin-Wald descriptive fit path also ran (its failed replicate skipped,
  # not fatal) and produced finite fit-index SEs
  expect_false(is.null(pooled$SE$fit_indices_descriptive))
  expect_true(all(is.finite(pooled$SE$fit_indices_descriptive)))

  # The skipped (NA-filled) replicate is tallied as a source failure, and the
  # valid-rotation count subtracts it (not only rotation failures), so it never
  # overstates the replicates that entered the pool.
  expect_identical(pooled$MI$bootstrap_source_failures, c(1L, 0L))
  expect_identical(pooled$MI$bootstrap_rotation_failures, c(0L, 0L))
  expect_identical(pooled$MI$bootstrap_rotation_valid, c(B - 1L, B))

  # if an imputation is left with fewer than two valid replicates, no SEs can be
  # computed and the pooled bootstrap returns NULL (the existing "no SEs" path)
  fits_fail <- fits
  fits_fail[[1]]$replicates$unrot_loadings <- mk_arr(na = seq_len(B))
  args_fail <- args
  args_fail$fits <- fits_fail
  expect_warning(
    pooled_fail <- do.call(.efa_pooled_bootstrap_pool, args_fail),
    class = "efa_pooled_boot_insufficient"
  )
  expect_null(pooled_fail)
})

test_that("the pooled bootstrap sample count reports the usable replicates when they differ", {
  # A pooled fit records per-imputation FAILURE counts rather than a survivor count, so without a
  # fallback the printed sample count states the requested b_boot whatever the survival rate -- the
  # same claim of an unearned precision that a single fit used to make.
  spec <- list(b_boot = 4L, is_pooled = TRUE, np_boot = TRUE)
  x <- list(MI = list(bootstrap_source_failures = c(1L, 0L)),
            settings = list(b_boot = 4L))

  expect_identical(EFAtools:::.efa_valid_replicate_counts(x), c(3L, 4L))
  expect_identical(EFAtools:::.efa_usable_replicate_text(x, spec),
                   " (between 3 and 4 usable per imputation)")

  # A run in which every replicate survived, in either shape, says nothing extra.
  x_clean <- list(MI = list(bootstrap_source_failures = c(0L, 0L)),
                  settings = list(b_boot = 4L))
  expect_identical(EFAtools:::.efa_usable_replicate_text(x_clean, spec), "")
  expect_identical(
    EFAtools:::.efa_usable_replicate_text(list(SE = list(valid_replicates = 4L)), spec), "")

  # A single fit keeps the unqualified wording.
  expect_identical(
    EFAtools:::.efa_usable_replicate_text(list(SE = list(valid_replicates = 3L)),
                                          list(b_boot = 4L)),
    " (3 usable)")
})

test_that("print.efa_mi output is stable (PAF, promax)", {
  local_reproducible_output()

  expect_snapshot(print(pooled_obl), transform = scrub_num)
})

test_that("print.efa_mi output is stable (ML, unrotated)", {
  local_reproducible_output()

  expect_snapshot(print(pooled_none), transform = scrub_num)
})

test_that("summary.efa_mi output is stable (PAF, promax)", {
  local_reproducible_output()

  expect_snapshot(print(summary(pooled_obl)), transform = scrub_num)
})

test_that("summary.efa_mi output is stable (ML, unrotated)", {
  local_reproducible_output()

  expect_snapshot(print(summary(pooled_none)), transform = scrub_num)
})

test_that("the pooled header and pooling settings follow the console width", {
  # The pooled header names the analysis, and its settings line is the longest line of the
  # report; both are emitted verbatim so a "setting = 'value'" token is never split, and both
  # therefore have to be packed to the console at the separators between those tokens.

  # The settings an item reports, independent of where its lines happen to break.
  tokens <- function(lines) {
    txt <- sub("^Pooling( settings)?: ", "", paste(trimws(lines), collapse = " "))
    strsplit(sub("\\.$", "", txt), ", ", fixed = TRUE)[[1L]]
  }
  at_width <- function(x, w) withr::with_options(list(cli.width = w), cli::ansi_strip(format(x)))

  reference <- tokens(wrapped_item(at_width(pooled_obl, 120L), "^Pooling settings:"))
  expect_length(reference, 3L)

  for (w in c(120L, 80L, 60L)) {
    out <- at_width(pooled_obl, w)
    lines <- c(wrapped_item(out, "^Pooled EFA"), wrapped_item(out, "^Pooling settings:"))
    expect_true(all(cli::ansi_nchar(lines, type = "width") <= w))
    # wrapping only moves the line breaks: the same settings, in the same order ...
    expect_identical(tokens(wrapped_item(out, "^Pooling settings:")), reference)
    # ... and no line ends inside a "setting = 'value'" token
    expect_false(any(grepl("= '[^']*$", lines)))
    expect_true(any(grepl("estimator = 'PAF'", lines, fixed = TRUE)))
    expect_true(any(grepl("rotation = 'promax'", lines, fixed = TRUE)))

    # the summary's diagnostics entry reports the same settings and is packed the same way
    pooling <- wrapped_item(at_width(summary(pooled_obl), w), "^Pooling:")
    expect_true(all(cli::ansi_nchar(pooling, type = "width") <= w))
    expect_identical(tokens(pooling), reference)
  }
})

test_that("the wrapped pooled header keeps its styling when colours are on", {
  withr::local_options(cli.num_colors = 256, cli.width = 60)

  out <- format(pooled_obl)
  # the estimator/rotation values stay emphasised after the line is packed ...
  expect_true(cli::ansi_has_any(out[2L]) || cli::ansi_has_any(out[3L]))
  # ... and with colours off the same lines are plain text, wrapped identically
  withr::local_options(cli.num_colors = 1)
  plain <- format(pooled_obl)
  expect_false(cli::ansi_has_any(paste(plain[1:4], collapse = "")))
  expect_identical(cli::ansi_strip(out[1:4]), plain[1:4])
})

test_that("format.efa_mi matches the printed output", {
  local_reproducible_output()

  expect_identical(format(pooled_obl),
                   utils::capture.output(print(pooled_obl)))
})

test_that("the averaged-index flag names only the indices that were reported", {
  local_reproducible_output()

  inc_lines <- function(x) {
    lines <- cli::ansi_strip(format(x))
    lines[grepl("^CFI|^TLI|averaged over the imputations", lines)]
  }

  # Both incremental indices defined: both are labelled and the note is plural.
  both <- inc_lines(pooled_none)
  expect_true(any(grepl("^CFI \\(avg\\. over imputations\\):", both)))
  expect_true(any(grepl("^TLI \\(avg\\. over imputations\\):", both)))
  expect_true(any(grepl("CFI and TLI are averaged over the imputations", both)))

  # A degenerate baseline leaves TLI undefined: its line is dropped, so the note must
  # name CFI alone and in the singular rather than claiming a TLI that was not printed.
  one <- pooled_none
  one$fit_indices$TLI <- NA_real_
  one_lines <- inc_lines(one)
  expect_true(any(grepl("^CFI \\(avg\\. over imputations\\):", one_lines)))
  expect_false(any(grepl("^TLI", one_lines)))
  expect_true(any(grepl("CFI is averaged over the imputations", one_lines)))

  # Neither index defined: no label on the bare CFI line and no note at all.
  none <- pooled_none
  none$fit_indices$CFI <- NA_real_
  none$fit_indices$TLI <- NA_real_
  none_lines <- cli::ansi_strip(format(none))
  expect_true(any(grepl("^CFI: ", none_lines)))
  expect_false(any(grepl("avg\\. over imputations", none_lines)))
  expect_false(any(grepl("averaged over the imputations", none_lines)))

  # The D2 chi-square note is independent of the incremental-index note.
  expect_true(any(grepl("the pooled .* is the D2 statistic", none_lines)))
})

# ---- Default unrotated alignment: medoid anchor and canonical gauge ---------

# These use resampled imputations on purpose. A `list(cormat, cormat, cormat)`
# fixture has no between-imputation variance, so the aligned solutions are
# identical, their average is still in the extraction's gauge, and neither the
# anchor choice nor the re-gauging can be detected.

make_resamples <- function(dat, n = 5, seed = 42) {
  set.seed(seed)
  lapply(seq_len(n), function(i) {
    dat[sample.int(nrow(dat), replace = TRUE), , drop = FALSE]
  })
}

test_that("pooled unrotated loadings do not depend on the order of data_list", {
  # A weakly determined second factor makes the column matching ambiguous, which
  # is precisely when the anchor choice matters: with a first-imputation anchor
  # the pooled loadings move by up to 0.84 across orderings of this fixture.
  # Where the factors are well separated the matching is unambiguous and every
  # anchor already agrees, so such a fixture could not detect the difference.
  imps <- make_resamples(GRiPS_raw)
  base <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 2, estimator = "ML", rotation = "none")))

  set.seed(7)
  for (i in seq_len(3)) {
    permuted <- suppressWarnings(suppressMessages(
      efa_mi(imps[sample(length(imps))], n_factors = 2, estimator = "ML",
             rotation = "none")))
    expect_equal(unclass(permuted$unrot_loadings),
                 unclass(base$unrot_loadings), tolerance = 1e-8)
    expect_equal(as.numeric(permuted$h2), as.numeric(base$h2), tolerance = 1e-8)
    expect_equal(permuted$fit_indices$RMSR, base$fit_indices$RMSR,
                 tolerance = 1e-8)
  }
})

test_that("pooled unrotated loadings are returned in the extraction's gauge", {
  imps <- make_resamples(DOSPERT_raw)

  # ML identifies the unrotated solution by diagonal L' Psi^-1 L ...
  ml <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 3, estimator = "ML", rotation = "none")))
  L <- unclass(ml$unrot_loadings)
  A <- crossprod(L, L / pmax(1 - rowSums(L^2), 1e-6))
  expect_lt(sum(abs(A[upper.tri(A)])) / sum(abs(diag(A))), 1e-8)

  # ... a principal-axis extraction by diagonal L'L, and the pooled solution
  # must follow whichever its component fits use.
  paf <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 3, estimator = "PAF", rotation = "none")))
  Lp <- unclass(paf$unrot_loadings)
  G <- crossprod(Lp)
  expect_lt(sum(abs(G[upper.tri(G)])) / sum(abs(diag(G))), 1e-8)
})

test_that("the canonical re-gauging leaves gauge-invariant quantities alone", {
  imps <- make_resamples(DOSPERT_raw)
  fits <- lapply(imps, function(d) suppressWarnings(suppressMessages(
    efa_fit(d, n_factors = 3, estimator = "ML", rotation = NULL))))
  Ls <- lapply(fits, function(f) unclass(f$unrot_loadings))

  aligned <- .efa_pooled_align_unrotated_list(
    Ls, align_unrotated = "signed_tucker_congruence")
  C <- aligned$meta[[1]]$C
  expect_false(is.null(C))
  expect_equal(crossprod(C), diag(ncol(C)), tolerance = 1e-10)

  # Undo the common rotation to recover the pre-canonical average, then check
  # that the quantities a rotation cannot change are indeed unchanged.
  M <- .average_matrices(aligned$loadings)
  raw <- M %*% t(C)
  expect_equal(rowSums(M^2), rowSums(raw^2), tolerance = 1e-10)
  expect_equal(M %*% t(M), raw %*% t(raw), tolerance = 1e-10)
})

test_that("analytic SEs survive the column-mixing gauge rotation", {
  imps <- make_resamples(DOSPERT_raw)
  pooled <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 3, estimator = "ML", rotation = "none",
           se = "information")))

  se <- pooled$SE$unrot_loadings
  expect_false(is.null(se))
  expect_true(all(is.finite(as.matrix(se))))
  expect_true(all(as.matrix(se) > 0))

  # Wald intervals stay centred on the pooled estimate they belong to.
  est <- unclass(pooled$unrot_loadings)
  lo <- pooled$CI$unrot_loadings$lower
  hi <- pooled$CI$unrot_loadings$upper
  expect_equal(unname((lo + hi) / 2), unname(est), tolerance = 1e-8)
})

test_that("an unreliable per-imputation SE still blanks the pooled element", {
  # The re-gauging reads the full covariance block rather than the marginal SEs,
  # so the marginal NA mask has to be applied explicitly. Because the rotation
  # mixes columns, one NA contaminates its whole row.
  imps <- make_resamples(DOSPERT_raw)
  pooled <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 3, estimator = "ML", rotation = "none",
           se = "information")))

  fits <- pooled$fits
  fits[[3]]$SE$unrot_loadings[2, 1] <- NA_real_
  Ls <- lapply(fits, function(f) unclass(f$unrot_loadings))
  aligned <- .efa_pooled_align_unrotated_list(
    Ls, align_unrotated = "signed_tucker_congruence")
  na_pool <- .efa_pooled_analytic_pool(
    fits = fits, unrot_loadings_aligned = aligned$loadings,
    align_meta = aligned$meta, ci = 0.95)

  expect_true(all(is.na(na_pool$SE$unrot_loadings[2, ])))
  expect_true(all(is.finite(na_pool$SE$unrot_loadings[3, ])))
})

test_that("a missing gauge covariance is reported once, not by both unrotated warnings", {
  # Dropping the covariance block leaves that imputation's row all-NA, which is also
  # the shape the NA-filled-SE warning keys on. The two must not both fire for one
  # downgrade: the missing-covariance branch is the specific diagnosis.
  imps <- make_resamples(DOSPERT_raw)
  pooled <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 3, estimator = "ML", rotation = "none",
           se = "information")))

  fits <- pooled$fits
  fits[[2]]$vcov_unrot_loadings <- NULL
  Ls <- lapply(fits, function(f) unclass(f$unrot_loadings))
  aligned <- .efa_pooled_align_unrotated_list(
    Ls, align_unrotated = "signed_tucker_congruence")
  # precondition: the gauge rotation is active, so the covariance block is required
  expect_false(is.null(aligned$meta[[1]]$C))

  expect_warning(
    gauge_pool <- withCallingHandlers(
      .efa_pooled_analytic_pool(fits = fits,
                                unrot_loadings_aligned = aligned$loadings,
                                align_meta = aligned$meta, ci = 0.95),
      efa_pooled_unrotated_se_unreliable = function(w) {
        testthat::fail("the missing-covariance downgrade was reported twice")
      }
    ),
    class = "efa_pooled_gauge_vcov_missing"
  )
  expect_true(all(is.na(gauge_pool$SE$unrot_loadings)))
})

test_that("the medoid anchor is order-invariant with only two imputations", {
  # Two imputations are tied by construction -- each is exactly as far from the
  # other -- so the anchor has to be settled on the candidates' own content
  # rather than on list position, or the tie puts the order dependence back.
  set.seed(42)
  imps <- lapply(seq_len(2), function(i) {
    GRiPS_raw[sample.int(nrow(GRiPS_raw), replace = TRUE), , drop = FALSE]
  })
  a <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 2, estimator = "ML", rotation = "none")))
  b <- suppressWarnings(suppressMessages(
    efa_mi(rev(imps), n_factors = 2, estimator = "ML", rotation = "none")))

  expect_equal(unclass(a$unrot_loadings), unclass(b$unrot_loadings),
               tolerance = 1e-10)
  expect_equal(as.numeric(a$h2), as.numeric(b$h2), tolerance = 1e-10)

  Ls <- lapply(a$fits, function(x) unclass(x$unrot_loadings))
  expect_identical(.efa_pooled_medoid_anchor(Ls), 1L)
  expect_identical(.efa_pooled_medoid_anchor(rev(Ls)), 2L)
})

test_that("re-gauging stands down when the gauge is not identified", {
  Ls <- lapply(1:3, function(i) unclass(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
            estimator = "ML", rotation = NULL)$unrot_loadings))

  # An improper average would enter Psi^-1 at the clamp and let one variable
  # decide the gauge, so it is left alone.
  heywood <- Ls[[1]]
  heywood[1, ] <- heywood[1, ] / sqrt(sum(heywood[1, ]^2)) * 1.05
  expect_identical(.efa_pooled_canonical_gauge(heywood, Ls[[1]]),
                   diag(ncol(heywood)))

  # So is one whose canonical values are tied, where any rotation inside the
  # tied subspace diagonalises equally well.
  tied <- matrix(0, 6, 2)
  tied[1:3, 1] <- 0.6
  tied[4:6, 2] <- 0.6
  expect_identical(.efa_pooled_canonical_gauge(tied, tied), diag(2))
})

test_that("re-gauging still fires when a uniqueness sits on the estimation floor", {
  # Psi is reconstructed from the pooled loadings, and maximum likelihood pins a
  # uniqueness at .uniqueness_floor. There 1 - rowSums(L^2) lands a whisker below
  # the value that actually diagonalised the fit, and Psi^-1 multiplies that gap
  # by ~1/floor. Reconstructing with a smaller floor pushes the measured defect
  # past the detection tolerance, which would silently switch the re-gauging off
  # for exactly the boundary solutions.
  set.seed(5)
  imps <- lapply(seq_len(5), function(i) {
    UPPS_raw[sample.int(nrow(UPPS_raw), 60, replace = TRUE), 1:12, drop = FALSE]
  })
  fits <- lapply(imps, function(d) suppressWarnings(suppressMessages(
    efa_fit(d, n_factors = 4, estimator = "ML", rotation = NULL))))

  # precondition: at least one fit really is on the floor, or the test is vacuous
  min_psi <- min(vapply(fits, function(f) min(1 - as.numeric(f$h2)), numeric(1)))
  expect_lte(min_psi, .uniqueness_floor + 1e-9)

  Ls <- lapply(fits, function(f) unclass(f$unrot_loadings))
  anchor <- Ls[[.efa_pooled_medoid_anchor(Ls)]]
  defect <- function(A) sum(abs(A[upper.tri(A)])) / sum(abs(diag(A)))
  wgram <- function(L, floor) crossprod(L, L / pmax(1 - rowSums(L^2), floor))

  # too small a floor makes the anchor look like it is in no recognised gauge ...
  expect_gt(defect(wgram(anchor, 1e-6)), 1e-4)
  # ... while the estimation floor recovers the constraint the fit satisfies
  expect_lt(defect(wgram(anchor, .uniqueness_floor)), 1e-4)

  # so detection must fire and the pooled matrix must come back canonical
  aligned <- .efa_pooled_align_unrotated_list(
    Ls, align_unrotated = "signed_tucker_congruence")
  expect_false(is.null(aligned$meta[[1]]$C))

  pooled <- suppressWarnings(suppressMessages(
    efa_mi(imps, n_factors = 4, estimator = "ML", rotation = "none")))
  L <- unclass(pooled$unrot_loadings)
  expect_lt(defect(wgram(L, .uniqueness_floor)), 1e-8)
})

test_that("pooled AIC/BIC/ECVI are withheld when the component fits are scaled", {
  # A missing-at-random fixture: column 1 is fully observed and drives the
  # missingness in the others, so cor_method = "fiml" fits a genuine two-stage
  # model and each component carries the corrected (scaled-shifted) statistic.
  imps <- lapply(1:3, function(i) {
    set.seed(100 + i)
    L <- matrix(0, 6, 2)
    L[1:3, 1] <- 0.7
    L[4:6, 2] <- 0.7
    S <- tcrossprod(L)
    diag(S) <- 1
    # Colour the deviates with the Cholesky factor, not an eigendecomposition: this population
    # covariance has repeated eigenvalues, whose eigenvector basis is undetermined and settled
    # by rounding, whereas chol() is unique for a positive definite matrix, so the seed above
    # reproduces the same imputation on every LAPACK build.
    X <- matrix(stats::rnorm(400 * 6), 400) %*% chol(S)
    colnames(X) <- paste0("V", seq_len(6))
    X[X[, 1] > 0.8, 2] <- NA
    X[X[, 1] < -0.8, 3] <- NA
    X[X[, 1] > 1.2, 4] <- NA
    X
  })

  pooled <- suppressMessages(suppressWarnings(
    efa_mi(imps, n_factors = 2, estimator = "ML", rotation = "none",
           cor_method = "fiml")
  ))

  # precondition: the components really are scaled and withhold the three indices
  expect_true(all(vapply(pooled$fits,
                         function(f) !is.null(f$fit_indices$chi_scaled_type),
                         logical(1))))
  expect_true(all(vapply(pooled$fits,
                         function(f) is.na(f$fit_indices$AIC), logical(1))))

  # so the D2 pool of those statistics must withhold them as well ...
  expect_true(is.na(pooled$fit_indices$AIC))
  expect_true(is.na(pooled$fit_indices$BIC))
  expect_true(is.na(pooled$fit_indices$ECVI))

  # ... while the pooled chi-square and the indices that remain interpretable stay
  expect_true(is.finite(pooled$fit_indices$chi))
  expect_true(is.finite(pooled$fit_indices$CFI))
  expect_true(is.finite(pooled$fit_indices$RMSEA))
})

test_that("pooled AIC/BIC/ECVI are withheld when the components fell back to the plain LRT", {
  # Same fixture as above, but with the Stage-1 saturated covariance forced to fail, so no
  # component can form its correction and each reports the tagged plain two-stage
  # likelihood-ratio statistic instead. The pooler takes the components' withholding
  # decision, so the three criteria must stay NA on this route too -- the tag tells a reader
  # which statistic was pooled, and it must never say "corrected" here.
  imps <- lapply(1:3, function(i) {
    set.seed(100 + i)
    L <- matrix(0, 6, 2)
    L[1:3, 1] <- 0.7
    L[4:6, 2] <- 0.7
    S <- tcrossprod(L)
    diag(S) <- 1
    X <- matrix(stats::rnorm(400 * 6), 400) %*% chol(S)
    colnames(X) <- paste0("V", seq_len(6))
    X[X[, 1] > 0.8, 2] <- NA
    X[X[, 1] < -0.8, 3] <- NA
    X[X[, 1] > 1.2, 4] <- NA
    X
  })

  testthat::local_mocked_bindings(
    .fiml_saturated_acov = function(...) {
      cli::cli_abort("forced degenerate saturated covariance",
                     class = "efa_fiml_singular_information")
    }
  )

  pooled <- suppressMessages(suppressWarnings(
    efa_mi(imps, n_factors = 2, estimator = "ML", rotation = "none",
           cor_method = "fiml")
  ))

  expect_true(all(vapply(
    pooled$fits,
    function(f) identical(f$fit_indices$chi_scaled_type, "uncorrected.lrt"),
    logical(1))))
  expect_true(all(vapply(pooled$fits,
                         function(f) is.na(f$fit_indices$AIC), logical(1))))

  expect_true(is.na(pooled$fit_indices$AIC))
  expect_true(is.na(pooled$fit_indices$BIC))
  expect_true(is.na(pooled$fit_indices$ECVI))
  expect_true(is.finite(pooled$fit_indices$chi))
})

test_that("pooled AIC/BIC/ECVI are reported on an unscaled ML pool", {
  # The guard is conditional, not a blanket suppression: with plain (unscaled)
  # component statistics the descriptive criteria are still returned.
  pooled_ml <- efa_mi(cormat_list, n_factors = 3, N = 500, estimator = "ML",
                      rotation = "none")

  expect_true(all(vapply(pooled_ml$fits,
                         function(f) is.null(f$fit_indices$chi_scaled_type),
                         logical(1))))
  expect_true(is.finite(pooled_ml$fit_indices$AIC))
  expect_true(is.finite(pooled_ml$fit_indices$BIC))
  expect_true(is.finite(pooled_ml$fit_indices$ECVI))
})

test_that("a pooled RMSEA interval that misses its point estimate is reported as undefined", {
  # The same non-convergent noncentrality solve the single-fit path guards: at a chi in the
  # millions stats::pchisq() stops converging and both bounds collapse below the point
  # estimate. An interval that does not contain the RMSEA reported beside it is not an
  # interval, so it is withheld rather than shipped as a usable range.
  point_big <- .rmsea_point(4101811, 135, 5e6)
  ci_big <- .efa_pooled_rmsea_ci(4101811, df = 135, N = 5e6, point = point_big)
  expect_true(is.na(ci_big[["lower"]]))
  expect_true(is.na(ci_big[["upper"]]))

  # An ordinary pooled statistic keeps its interval, and that interval brackets the point
  # estimate the caller reports.
  point_ok <- .rmsea_point(200, 102, 500)
  ci_ok <- .efa_pooled_rmsea_ci(200, df = 102, N = 500, point = point_ok)
  expect_true(is.finite(ci_ok[["lower"]]))
  expect_true(is.finite(ci_ok[["upper"]]))
  expect_lte(ci_ok[["lower"]], point_ok)
  expect_lte(point_ok, ci_ok[["upper"]])

  # Containment is judged against the caller's point estimate, not one re-derived here from
  # chi and N: the two callers report estimates built on different statistics and scales.
  outside <- .efa_pooled_rmsea_ci(200, df = 102, N = 500,
                                  point = ci_ok[["upper"]] + .05)
  expect_true(is.na(outside[["lower"]]))
  expect_true(is.na(outside[["upper"]]))
})

# Rubin's-rules pooling of analytic ROTATED-quantity SEs in EFA_POOLED() on the
# se = "information" path: rotated loadings, communalities (h2), and (oblique)
# factor correlations (Phi) and structure coefficients.
#
# A rotated-loading SE is conditional on the rotation criterion (Archer &
# Jennrich 1973; Jennrich 1973, 1974; Zhang, Preacher, & Jennrich 2012; Zhang &
# Preacher 2015), so for ORTHOGONAL and OBLIQUE
# rotations alike the within-imputation variance is each fit's own criterion-aware
# delta-method rotated SE (the quantity EFA() returns), reused after a
# signed-permutation alignment to the MI target; the between-imputation variance
# is the sample variance of the aligned per-imputation rotated loadings. Both
# rotation families are flagged MI$<param>$method = "signed_permutation_approx".
# A fixed-rotation linear map of the unrotated covariance is deliberately NOT used
# for rotated-loading SEs (it would inflate them by redistributing the rotational
# indeterminacy variance the criterion pins down). Communalities are
# rotation-invariant and pool with no alignment.
#
# Assertion groups:
#   1. Identity-imputation invariance (orthogonal): pooled rotated SE equals the
#      per-fit criterion-aware delta SE; B = 0; FMI = 0.
#   2. Point-estimate / orthogonal-Procrustes sanity.
#   3. Regression guard: orthogonal rotated SE is the delta SE, not the inflated
#      fixed-Q propagation.
#   4. Communalities pooling (gauge-invariant, no alignment).
#   5. np-boot cross-check on an orthogonal solution (informational).
#   6. Oblique approximation: signed-permutation path, flagged, does not abort.
#   7. Phi and Structure (oblique): shapes, diag(Phi SE) = 0, symmetry, invariance.
#   8. Slot shape under se = "information" + rotation.

cormat <- test_models$baseline$cormat
p_vars <- ncol(cormat)
N_id   <- 500L
k      <- 2L
m_id   <- 5L
identical_list <- replicate(m_id, cormat, simplify = FALSE)

# The orthogonal (varimax) identity-imputation pool and its single-fit oracle
# are needed by several groups below with byte-identical arguments, so they are
# built once here and only ever read. Varimax is the closed-form SVD path with
# no random start, so neither object consumes RNG and sharing them cannot move
# the stream the oblique groups below draw from. The oblique fixtures are
# deliberately NOT shared: an oblique engine draws random starts, so each of
# those groups seeds and fits its own solution.
pooled_varimax <- suppressMessages(EFA_POOLED(
  identical_list, n_factors = k, N = N_id, method = "ML",
  rotation = "varimax", se = "information"
))
oracle_varimax <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                      rotation = "varimax", se = "information")

# ---- Group 1 ---------------------------------------------------------------

test_that("identity imputations (orthogonal): pooled rotated SE equals the per-fit criterion-aware delta SE, B = 0", {
  expect_identical(pooled_varimax$settings$se, "information")

  # Pooled point estimate is the MI target (= the single fit's varimax solution).
  expect_equal(unclass(pooled_varimax$rot_loadings),
               unclass(oracle_varimax$rot_loadings), tolerance = 1e-8)

  # The within-imputation variance is each fit's criterion-aware delta-method
  # rotated SE (the quantity EFA() returns); with identical imputations the
  # signed-permutation alignment is the identity and B = 0, so the pooled rotated
  # SE equals the single-fit rotated SE exactly.
  expect_equal(unname(pooled_varimax$SE$rot_loadings),
               unname(oracle_varimax$SE$rot_loadings), tolerance = 1e-6)

  fmi <- pooled_varimax$MI$rot_loadings$FMI
  riv <- pooled_varimax$MI$rot_loadings$RIV
  expect_true(all(abs(fmi[is.finite(fmi)]) < 1e-6))
  expect_true(all(abs(riv[is.finite(riv)]) < 1e-6))

  # Provenance flag: the criterion-aware signed-permutation reuse (orthogonal and
  # oblique alike).
  expect_identical(pooled_varimax$MI$rot_loadings$method,
                   "signed_permutation_approx")
})

# ---- Group 2 ---------------------------------------------------------------

test_that("orthogonal point estimate: L_d^unrot Q_d reproduces the aligned rotated loadings", {
  A      <- unclass(oracle_varimax$unrot_loadings)
  target <- unclass(oracle_varimax$rot_loadings)
  Q      <- EFAtools:::.procrustes_orthogonal_T(A, target)

  # Q recovers the orthogonal rotation, so A Q reproduces the rotated loadings.
  expect_equal(unname(A %*% Q), unname(target), tolerance = 1e-8)
  # Orthogonality of the recovered transform.
  expect_equal(crossprod(Q), diag(k), tolerance = 1e-10)
})

# ---- Group 3 ---------------------------------------------------------------

test_that("orthogonal rotated SE is the criterion-aware delta SE, not the inflated fixed-Q propagation", {
  # The pooled orthogonal rotated SE is the per-fit criterion-aware delta SE.
  expect_equal(unname(pooled_varimax$SE$rot_loadings),
               unname(oracle_varimax$SE$rot_loadings), tolerance = 1e-6)

  # Regression guard: the former fixed-Q Kronecker propagation of the
  # unrotated covariance is a DISTINCT matrix from the reported (criterion-aware)
  # SE -- an orthogonal Q preserves each row's total variance, so it redistributes
  # the rotational-indeterminacy variance the criterion otherwise pins down (and is
  # never smaller). Confirm the reported SE is not that propagated quantity; the
  # gap grows from modest on this well-separated structure to several-fold when
  # factors are weakly separated.
  A      <- unclass(oracle_varimax$unrot_loadings)
  V      <- oracle_varimax$vcov_unrot_loadings
  target <- unclass(oracle_varimax$rot_loadings)
  Q      <- EFAtools:::.procrustes_orthogonal_T(A, target)
  kron_se <- EFAtools:::.efa_pooled_propagate_procrustes_vcov(V, Q, p_vars, k)

  reported <- unclass(pooled_varimax$SE$rot_loadings)
  expect_gt(max(abs(kron_se - reported)), 1e-3)
})

# ---- Group 4 ---------------------------------------------------------------

test_that("communalities pool is gauge-invariant: Rubin's rules on per-fit communality SEs, no alignment", {
  # Identical imputations -> B = 0 -> pooled communality SE equals the per-fit
  # communality SE exactly (no alignment is applied; communalities are
  # rotation-invariant).
  expect_length(pooled_varimax$SE$communalities, p_vars)
  expect_equal(unname(pooled_varimax$SE$communalities),
               unname(oracle_varimax$SE$communalities), tolerance = 1e-10)
  se_comm <- pooled_varimax$SE$communalities
  expect_true(all(se_comm[!is.na(se_comm)] >= 0))

  fmi <- pooled_varimax$MI$communalities$FMI
  expect_true(all(abs(fmi[is.finite(fmi)]) < 1e-12))
  expect_identical(pooled_varimax$MI$communalities$method, "gauge_invariant")
})

test_that("`h2` stays available as an alias of the canonical communality SE/CI", {
  # `communalities` is what EFA() and the two-stage route call it, so it is the
  # canonical name on every route; `h2` is what this route has always returned.
  expect_identical(pooled_varimax$SE$h2, pooled_varimax$SE$communalities)
  expect_identical(pooled_varimax$CI$h2, pooled_varimax$CI$communalities)

  # The MI diagnostics carry the canonical name only: the printed FMI/RIV summary
  # walks the whole MI tree, so an aliased family would be counted twice there.
  expect_null(pooled_varimax$MI$h2)
})

# ---- Group 4b --------------------------------------------------------------

test_that("distinct imputations (B > 0): pooled rotated-loading and h2 SEs match an independent Rubin computation", {
  # Identical imputations leave B = 0, which cannot exercise the
  # between-imputation variance, the (1 + 1/m) B scaling, or the column
  # alignment under genuine disagreement. Drive B > 0 with deterministically
  # distinct datasets (base-R draws, no MASS) from a 2-factor population and
  # check the pooled SEs against a hand-computed Rubin reference.
  set.seed(20260621)
  p_b <- 6L
  k_b <- 2L
  N_b <- 250L
  m_b <- 4L
  Lp <- matrix(0, p_b, k_b)
  Lp[1:3, 1] <- c(0.72, 0.65, 0.60)
  Lp[4:6, 2] <- c(0.72, 0.65, 0.60)
  Sig <- tcrossprod(Lp)
  diag(Sig) <- 1
  Rt <- chol(Sig)
  imps_b <- lapply(seq_len(m_b), function(i) {
    X <- matrix(stats::rnorm(N_b * p_b), N_b, p_b) %*% Rt
    colnames(X) <- paste0("V", seq_len(p_b))
    X
  })

  pooled_b <- tryCatch(suppressWarnings(suppressMessages(EFA_POOLED(
    imps_b, n_factors = k_b, method = "ML", rotation = "varimax",
    se = "information"
  ))), error = function(e) e)
  if (inherits(pooled_b, "error")) skip("EFA_POOLED failed on synthetic data")

  m <- length(pooled_b$fits)

  # --- Communalities (gauge-invariant): independent Rubin from per-fit pieces ---
  comm    <- t(vapply(pooled_b$fits, function(f) as.numeric(f$h2), numeric(p_b)))
  se_comm <- t(vapply(pooled_b$fits, function(f) as.numeric(f$SE$communalities),
                      numeric(p_b)))
  if (anyNA(se_comm)) skip("a component fit had non-finite communality SEs")
  Ubar_h2 <- colMeans(se_comm^2)
  B_h2    <- apply(comm, 2, stats::var)
  ref_se_h2 <- sqrt(Ubar_h2 + (1 + 1 / m) * B_h2)
  expect_equal(unname(pooled_b$SE$communalities), unname(ref_se_h2), tolerance = 1e-8)
  expect_true(any(B_h2 > 0))                 # the regime is genuinely B > 0
  expect_true(any(pooled_b$MI$communalities$FMI > 0))  # and it flows through to FMI

  # --- Rotated loadings: independent Rubin from the per-fit criterion-aware delta
  # SE. Mirror the production gauge exactly: the MI target is the first fit's
  # rotated solution; imputation 1 uses it directly and 2..m are Procrustes-aligned
  # to it. The within-imputation variance is each fit's $SE$rot_loadings, reordered
  # by the signed-permutation alignment of its own rotated solution to Lr_d; B is
  # the sample variance of the aligned rotated loadings. ---
  target  <- as.matrix(pooled_b$alignment$target)
  est_mat <- matrix(NA_real_, m, p_b * k_b)
  se_mat  <- matrix(NA_real_, m, p_b * k_b)
  for (d in seq_len(m)) {
    if (d == 1L) {
      Lr_d <- target
    } else {
      A    <- unclass(pooled_b$fits[[d]]$unrot_loadings)
      Lr_d <- unclass(EFAtools::efa_procrustes(A, target, rotation = "orthogonal")$loadings)
    }
    se_d <- pooled_b$fits[[d]]$SE$rot_loadings
    if (is.null(se_d) || anyNA(se_d)) skip("a component fit had non-finite rotated SEs")
    fo <- EFAtools:::.align_solution(
      L_target = Lr_d, L = unclass(pooled_b$fits[[d]]$rot_loadings)
    )$factor_order
    est_mat[d, ] <- as.vector(Lr_d)
    se_mat[d, ]  <- as.vector(unclass(se_d)[, fo, drop = FALSE])
  }
  Ubar_r <- colMeans(se_mat^2)
  B_r    <- apply(est_mat, 2, stats::var)
  ref_se_rot <- matrix(sqrt(Ubar_r + (1 + 1 / m) * B_r), p_b, k_b)
  expect_equal(unname(pooled_b$SE$rot_loadings), unname(ref_se_rot),
               tolerance = 1e-6)
  expect_true(any(pooled_b$MI$rot_loadings$FMI > 0))

  # Factor-column labels are consistent across the loading-shaped SE families.
  expect_identical(colnames(pooled_b$SE$rot_loadings),
                   colnames(pooled_b$SE$unrot_loadings))
})

# ---- Group 5 ---------------------------------------------------------------

test_that("np-boot cross-check: analytic and bootstrap pooled orthogonal rotated SEs agree to within ~order of magnitude (informational)", {
  skip_on_cran()
  skip_if_not_slow()

  set.seed(20260621)
  p_syn <- 6L
  N_syn <- 300L
  m_syn <- 3L
  # Independent-cluster 2-factor population (orthogonal), strong primary loadings.
  L_true <- matrix(0, p_syn, 2)
  L_true[1:3, 1] <- c(0.75, 0.70, 0.65)
  L_true[4:6, 2] <- c(0.75, 0.70, 0.65)
  Sigma <- tcrossprod(L_true)
  diag(Sigma) <- 1
  # As in the block above, draw through the Cholesky factor: every uniqueness is
  # positive so Sigma is positive definite, chol() is then unique, and the seeded
  # sample is reproduced on every LAPACK build.
  Rt <- chol(Sigma)

  imps <- lapply(seq_len(m_syn), function(i) {
    X <- matrix(stats::rnorm(N_syn * p_syn), N_syn, p_syn) %*% Rt
    colnames(X) <- paste0("V", seq_len(p_syn))
    as.data.frame(X)
  })

  pooled_info <- tryCatch(suppressWarnings(suppressMessages(EFA_POOLED(
    imps, n_factors = 2, method = "ML", rotation = "varimax", se = "information"
  ))), error = function(e) e)
  if (inherits(pooled_info, "error")) skip("information EFA_POOLED failed on synthetic data")

  pooled_boot <- tryCatch(suppressWarnings(suppressMessages(EFA_POOLED(
    imps, n_factors = 2, method = "ML", rotation = "varimax",
    se = "np-boot", b_boot = 200
  ))), error = function(e) e)
  if (inherits(pooled_boot, "error")) skip("np-boot EFA_POOLED failed on synthetic data")

  se_info <- pooled_info$SE$rot_loadings
  se_boot <- pooled_boot$SE$rot_loadings
  est     <- unclass(pooled_info$rot_loadings)

  # Compare on the primary (marker) loadings where the SE is well determined.
  primary <- abs(est) >= 0.3 & is.finite(se_info) & is.finite(se_boot) &
    se_boot > 0
  expect_true(any(primary))
  ratio <- se_info[primary] / se_boot[primary]

  # Informational: both target the same MI-gauge estimand (Procrustes to the MI
  # target), so the analytic fixed-Q SE and the bootstrap SE should agree to
  # within roughly 10-20%; the band is deliberately loose (not a strict oracle).
  expect_gt(stats::median(ratio), 0.5)
  expect_lt(stats::median(ratio), 1.6)
})

# ---- Group 6 ---------------------------------------------------------------

test_that("oblique rotated SE pooling uses the signed-permutation fallback, is flagged, and does not abort", {
  set.seed(42)
  pooled <- suppressMessages(EFA_POOLED(
    identical_list, n_factors = k, N = N_id, method = "ML",
    rotation = "oblimin", se = "information"
  ))
  oracle <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                rotation = "oblimin", se = "information")

  # Results are produced (no abort) and carry the documentation flag.
  expect_false(is.null(pooled$SE$rot_loadings))
  expect_identical(pooled$MI$rot_loadings$method, "signed_permutation_approx")

  # Signed-permutation reuse of the per-fit rotated SE: with identical imputations
  # the alignment is the identity, so the within-imputation variance (Ubar) is the
  # single-fit SE exactly. The pooled SE differs only by the tiny between-imputation
  # variance the iterative oblique Procrustes alignment injects into the per-fit
  # rotated point estimates (the closed-form orthogonal path has none); its
  # magnitude depends on the oblimin flat-optimum geometry, so allow headroom.
  expect_equal(unname(pooled$SE$rot_loadings), unname(oracle$SE$rot_loadings),
               tolerance = 1e-3)

  fmi <- pooled$MI$rot_loadings$FMI
  expect_true(all(abs(fmi[is.finite(fmi)]) < 1e-2))
})

# ---- Group 7 ---------------------------------------------------------------

test_that("oblique Phi and Structure SE pooling: shapes, diag(Phi SE) = 0, symmetry, invariance, flags", {
  set.seed(42)
  pooled <- suppressMessages(EFA_POOLED(
    identical_list, n_factors = k, N = N_id, method = "ML",
    rotation = "oblimin", se = "information"
  ))
  oracle <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                rotation = "oblimin", se = "information")

  # Phi SE: k x k, fixed unit diagonal -> zero SE, symmetric. (1e-3 headroom for
  # the iterative-oblique-Procrustes B injection, as in the previous group.)
  expect_identical(dim(pooled$SE$Phi), c(k, k))
  expect_equal(diag(pooled$SE$Phi), rep(0, k), tolerance = 1e-12, ignore_attr = TRUE)
  expect_equal(pooled$SE$Phi, t(pooled$SE$Phi), tolerance = 1e-12)
  expect_equal(unname(pooled$SE$Phi), unname(oracle$SE$Phi), tolerance = 1e-3)
  expect_identical(pooled$MI$Phi$method, "signed_permutation_approx")

  # The fixed unit diagonal carries no MI diagnostics (NA), unlike the off-diagonal.
  expect_true(all(is.na(diag(pooled$MI$Phi$FMI))))
  expect_true(all(is.na(diag(pooled$MI$Phi$df))))

  # Structure SE: p x k, invariance under identical imputations.
  expect_identical(dim(pooled$SE$Structure), c(p_vars, k))
  expect_equal(unname(pooled$SE$Structure), unname(oracle$SE$Structure),
               tolerance = 1e-3)
  expect_identical(pooled$MI$Structure$method, "signed_permutation_approx")
})

# ---- Group 8 ---------------------------------------------------------------

test_that("se = 'information' + rotation fills the rotated SE/CI/MI slots with correct shapes", {
  # Orthogonal: rot_loadings + communalities; no Phi/Structure.
  rot_dn <- dimnames(unclass(pooled_varimax$rot_loadings))

  expect_identical(dim(pooled_varimax$SE$rot_loadings), c(p_vars, k))
  expect_identical(dimnames(pooled_varimax$SE$rot_loadings), rot_dn)
  expect_identical(dim(pooled_varimax$CI$rot_loadings$lower), c(p_vars, k))
  expect_identical(dim(pooled_varimax$CI$rot_loadings$upper), c(p_vars, k))
  expect_setequal(names(pooled_varimax$MI$rot_loadings),
                  c("RIV", "FMI", "df", "method"))
  expect_identical(dim(pooled_varimax$MI$rot_loadings$df), c(p_vars, k))

  expect_length(pooled_varimax$SE$communalities, p_vars)
  expect_length(pooled_varimax$CI$communalities$lower, p_vars)
  expect_setequal(names(pooled_varimax$MI$communalities),
                  c("RIV", "FMI", "df", "method"))

  expect_null(pooled_varimax$SE$Phi)
  expect_null(pooled_varimax$SE$Structure)
  expect_null(pooled_varimax$replicates)

  # Oblique: adds Phi and Structure.
  pooled_oblq <- suppressMessages(EFA_POOLED(
    identical_list, n_factors = k, N = N_id, method = "ML",
    rotation = "oblimin", se = "information"
  ))
  expect_identical(dim(pooled_oblq$SE$Phi), c(k, k))
  expect_identical(dim(pooled_oblq$CI$Phi$lower), c(k, k))
  expect_identical(dim(pooled_oblq$SE$Structure), c(p_vars, k))
  expect_setequal(names(pooled_oblq$MI$Structure),
                  c("RIV", "FMI", "df", "method"))

  # Unrotated families keep their RIV/FMI/df-only MI shape (no method tag).
  expect_setequal(names(pooled_oblq$MI$unrot_loadings), c("RIV", "FMI", "df"))
  expect_setequal(names(pooled_oblq$MI$uniquenesses), c("RIV", "FMI", "df"))
})

# ---- Group 9 ---------------------------------------------------------------

test_that("an unreliable per-imputation rotated SE warns and NA-blanks the pooled rotated SE family", {
  base_fit <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                  rotation = "oblimin", se = "information")
  fits <- replicate(3L, base_fit, simplify = FALSE)

  # Mimic a fit whose rotated-SE Jacobian could not be reproduced: EFA() leaves the
  # rotated SE matrices NA-filled while the unrotated covariance stays reliable, so
  # the whole-fit vcov reliability gate does not catch it.
  fits[[2]]$SE$rot_loadings[] <- NA_real_
  fits[[2]]$SE$Phi[]          <- NA_real_
  fits[[2]]$SE$Structure[]    <- NA_real_

  unrot_list <- lapply(fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "signed_tucker_congruence"
  )
  rot_list <- lapply(fits, function(f) unclass(f$rot_loadings))
  phi_list <- lapply(fits, function(f) f$Phi)
  str_list <- Map(function(L, Phi) L %*% Phi, rot_list, phi_list)
  h2 <- EFAtools:::.efa_pooled_communalities(rot_list[[1]], phi_list[[1]])

  expect_warning(
    pooled <- EFAtools:::.efa_pooled_analytic_pool(
      fits = fits,
      unrot_loadings_aligned = aligned$loadings,
      align_meta = aligned$meta,
      ci = 0.95,
      align_unrotated = "signed_tucker_congruence",
      rotation_type = "oblique",
      rot_loadings = rot_list,
      phis = phi_list,
      structure_loadings = str_list,
      mean_structure_loadings = str_list[[1]],
      mean_phis = phi_list[[1]],
      h2 = h2
    ),
    class = "efa_pooled_rotated_se_unreliable"
  )

  # The affected rotated families are fully NA-blanked by the fail-closed mask,
  # while the gauge-invariant communalities and the unrotated families stay finite.
  expect_true(all(is.na(pooled$SE$rot_loadings)))
  expect_true(all(is.na(pooled$SE$Structure)))
  expect_true(all(is.na(pooled$SE$Phi[upper.tri(pooled$SE$Phi)])))
  expect_false(all(is.na(pooled$SE$unrot_loadings)))
  expect_false(all(is.na(pooled$SE$communalities)))
})

#' Exploratory factor analysis (EFA)
#'
#' This function does an EFA with either `PAF`, `ML`, `ULS`/`MINRES`,
#' or `DWLS` with or without subsequent rotation.
#' Estimation and rotation are controlled through the control objects built by
#' [estimate_control()] and [rotate_control()]; each accepts a `type`
#' ("EFAtools", "SPSS", "psych", or "none") that fills in its remaining settings.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations. If raw data is entered, the correlation matrix is found from the
#' data.
#' @param n_factors numeric. Number of factors to extract. Must be at least 1 and
#' smaller than the number of variables (the common factor model is not identified
#' otherwise). Use [efa_retain()] to decide on a value.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used; with raw data, `N` is found from the data instead.
#'  - With `N = NA`, not all fit indices can be computed; a positive `N` that is very
#'    small relative to the number of variables leaves the chi-square-derived indices
#'    unavailable as well, with a warning.
#'  - With raw data, `N` is the number of cases the correlation matrix was actually
#'    computed from -- see `use` for the general rule and how missing values change it.
#'    Under `cor_method = "fiml"`, `use` is ignored and `N` is instead the number of
#'    cases carrying at least one observed value.
#' @param estimator character. The estimator used to fit the EFA: "PAF" (principal axis
#' factoring), "ML" (maximum likelihood), "ULS" (unweighted least squares; "MINRES" is an
#' accepted alias returning identical results), or "DWLS" (diagonally weighted least
#' squares, for ordinal data). See the *Estimators* section in Details for their
#' properties and data requirements. Lower-case versions (e.g., "paf") are also accepted.
#' @param rotation character. Either perform no rotation ("none"; default),
#' an orthogonal rotation ("varimax", "equamax", "quartimax", "geominT",
#' "bentlerT", or "bifactorT"), or an oblique rotation ("promax", "oblimin",
#' "quartimin", "simplimax", "bentlerQ", "geominQ", or "bifactorQ"). See the
#' *Rotations* section in Details for their properties and known issues.
#' @param se character. Whether and how to compute standard errors (and matching
#'  confidence intervals): "none" (default), "information" (analytic
#'  standard errors from the expected Fisher information of the ML solution), "sandwich"
#'  (robust "sandwich" standard errors from raw data, which stay reliable under
#'  non-normality or a misspecified estimator weight), or "np-boot" (non-parametric
#'  bootstrap). The
#'  methods differ in their assumptions, their data requirements, and which estimator,
#'  rotation, and `cor_method` combinations they support; see the *Standard errors*
#'  section in Details.
#' @param use character. Passed to [stats::cor()] if raw data
#' is given as input. Default is "pairwise.complete.obs". It is ignored when
#' `cor_method = "fiml"` (which handles the missingness itself, so every case
#' contributes), and it is overridden to listwise deletion whenever an asymptotic
#' covariance is required (the "DWLS" estimator, or `se = "sandwich"`), because the
#' covariance must describe the same cases as the correlation matrix.
#' @param cor_method character. How the correlation is computed from raw data:
#'   `"pearson"`, `"spearman"`, or `"kendall"` (passed to [stats::cor()]); `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations of ordinal / binary data; or
#'   `"fiml"` for a two-stage full-information maximum-likelihood correlation from raw data
#'   with missing values. See the *Correlation methods* section in Details for their
#'   properties and the combinations they support. Default is "pearson".
#' @param estimate_control a control object from [estimate_control()] bundling the
#'  estimation control arguments: the `type` preset; the principal-axis-factoring iteration
#'  settings `init_comm`, `criterion`, `criterion_type`, `max_iter`, and `abs_eigen`; and
#'  the maximum-likelihood `start_method`. Defaults to `estimate_control()`, which uses
#'  the `"EFAtools"` type. See [estimate_control()].
#' @param rotate_control a control object from [rotate_control()] bundling the rotation
#'  control arguments: the `type` preset; Kaiser `normalize`; the convergence `precision`; the
#'  factor `order_type`; the varimax/promax settings `varimax_type` and `p_type`; the
#'  simplimax/promax `k`; and `random_starts`. Defaults to `rotate_control()`, which uses
#'  the `"EFAtools"` type. The estimation and rotation presets
#'  are independent. See [rotate_control()] for details.
#' @param b_boot numeric. The number of bootstrap samples to draw. Default is 1000.
#'  Must be at least 2, the smallest number from which a standard error is defined.
#'  Under `cor_method = "fiml"` each bootstrap sample re-runs the EM moment
#'  estimation, so a smaller value may be advisable.
#' @param ci numeric. The level of the confidence intervals: the percentile intervals from
#'  the bootstrap samples under `se = "np-boot"`, and the analytic Wald intervals under
#'  `se = "information"` and `se = "sandwich"`, the corrected two-stage intervals of
#'  `cor_method = "fiml"` included. Must be greater than 0 and smaller than 1. Default is
#'  .95 for 95% CIs.
#' @param seed numeric. An optional seed for the random-number generator.
#' @param ... Additional arguments forwarded to the rotation engine (usually not needed);
#'  an unrecognized one is an error, not a setting that is silently ignored.
#'
#' @details
#'
#' ## Estimators
#'
#' The estimator is chosen with `estimator`.
#'
#' - **PAF** (principal axis factoring) iteratively estimates the communalities and makes
#'   no distributional assumptions, which makes it robust and a good general-purpose
#'   default. Because it minimises no likelihood or weighted discrepancy it provides no
#'   model chi-square, and hence no chi-square-based fit indices (see *Fit indices*).
#' - **ML** (maximum likelihood) maximises the normal-theory likelihood. It yields the
#'   full set of fit indices and is the only estimator with analytic expected-information
#'   standard errors (`se = "information"`), but it assumes multivariate normality and is
#'   the most prone to Heywood (improper) cases. Its starting values are set by
#'   `start_method`.
#' - **ULS** (unweighted least squares) minimises the sum of squared correlation
#'   residuals. "MINRES" (minimum residual) is the same estimator under a different name
#'   and returns identical results. It makes no normality assumption, is robust to mild
#'   non-normality, and yields the full set of fit indices.
#' - **DWLS** (diagonally weighted least squares) is the recommended estimator for ordinal
#'   data. It weights each off-diagonal correlation residual by the inverse asymptotic
#'   variance of the corresponding polychoric correlation (Muthén, du Toit, & Spisic, 1997),
#'   reproducing the loadings of a diagonally weighted least squares fit. It therefore
#'   requires raw ordinal data with `cor_method = "poly"` or `"tetra"`. Because the
#'   weighting follows the polychoric
#'   asymptotic covariance, the matrix and the weights are estimated on the
#'   listwise-complete cases. Its fit-index behaviour is described under *Fit indices*.
#'
#' ## Correlation methods
#'
#' When raw data are supplied, `cor_method` selects how the correlation matrix is computed
#' (it is ignored when a correlation matrix is entered directly).
#'
#' - **"pearson"** (default), **"spearman"**, and **"kendall"** are passed to
#'   [stats::cor()]. The factor model assumes a Pearson correlation, but a rank
#'   correlation is analysed on its own scale, not converted to a Pearson-equivalent
#'   value; Kendall's tau in particular gives the most attenuated loadings of the three.
#'   For ordinal items prefer `"poly"` / `"tetra"` below, which estimate the correlation
#'   of the underlying continuous variables.
#' - **"poly"** / **"tetra"** compute polychoric / tetrachoric correlations for ordinal /
#'   binary data, assuming an underlying bivariate-normal latent variable. They use a
#'   two-step estimator. The polychoric asymptotic covariance that
#'   underlies both the DWLS weights and the scaled (sandwich) statistic relies on
#'   large-sample theory that degrades for empty or near-empty response-category
#'   combinations; with very sparse cells the resulting weights and standard errors can be
#'   unreliable (a warning is issued when empty cells are present), so interpret them with
#'   caution and consider collapsing rare categories.
#'   Each of the `p(p - 1)/2` variable pairs is a separate numerical optimisation, so a
#'   polychoric matrix takes much longer to compute than a Pearson one, and the difference grows
#'   quadratically in the number of variables; with `se = "np-boot"` the whole matrix is
#'   re-estimated for every bootstrap replicate.
#' - **"fiml"** estimates a two-stage full-information maximum-likelihood correlation. The
#'   saturated multivariate-normal mean and covariance are estimated from raw data with
#'   missing values by an EM algorithm assuming the data are missing at random (Yuan,
#'   Marshall, & Bentler, 2002; Little & Rubin, 2002), and the standardized covariance is
#'   then analysed. The model fit indices are corrected
#'   two-stage statistics wherever the correction can be formed (see *Fit indices*).
#'   `"fiml"` uses every case and handles the
#'   missingness itself, so `use` is ignored; it supplies a continuous (Pearson-type)
#'   correlation only and is therefore not compatible with `estimator = "DWLS"`. Standard
#'   errors are available analytically for `estimator = "ML"` or `"ULS"` and, for any estimator,
#'   by the non-parametric bootstrap (see *Standard errors*). For multiply imputed data,
#'   [efa_mi()] is the alternative route to handling missingness. Both routes assume
#'   the values are missing at random (MAR), and which one to prefer is largely
#'   practical: FIML is a single, efficient fit and is the simpler default when the
#'   analysis model is the whole story, whereas multiple imputation is more flexible when
#'   the imputation model should draw on auxiliary variables not in the factor model, or
#'   when the same imputations feed several downstream analyses.
#'
#' ## Rotations
#'
#' A rotation transforms the unrotated loadings toward a simpler, more interpretable
#' pattern; all rotations are performed by rotation engines built into the package.
#' Orthogonal rotations keep the factors uncorrelated, whereas oblique rotations let them
#' correlate (returning a pattern matrix, a structure matrix, and the factor
#' intercorrelations `Phi`) and are usually more realistic for psychological constructs.
#'
#' For an oblique solution the **pattern** matrix (`rot_loadings`) holds each variable's
#' unique contribution from each factor, with the other factors partialled out; it is what
#' is normally interpreted and reported. The **structure** matrix
#' (`Structure = rot_loadings %*% Phi`) holds the plain variable-factor correlations, which
#' are inflated by the factor intercorrelations. The two coincide only when `Phi` is the
#' identity, which is why an orthogonal rotation returns `rot_loadings` alone.
#'
#' Orthogonal rotations:
#' - **varimax** maximises the variance of the squared loadings within each factor (column
#'   simplicity). It is the most widely used orthogonal rotation and spreads variance
#'   across factors rather than concentrating it in a general factor.
#' - **quartimax** simplifies the variables (rows) so that each loads mainly on one factor;
#'   it tends to produce a strong general factor.
#' - **equamax** is a Crawford-Ferguson compromise between varimax (column) and quartimax
#'   (row) simplicity.
#' - **geominT** uses a geometric-mean criterion that rewards a sparse pattern and
#'   tolerates variables with cross-loadings; a smaller offset `delta` gives a sparser
#'   solution but sharper local minima.
#' - **bentlerT** uses Bentler's invariant pattern simplicity criterion.
#' - **bifactorT** is the Jennrich-Bentler orthogonal bifactor criterion: a general factor
#'   plus group factors (bifactor simple structure). It needs at least two group factors
#'   (`n_factors >= 3`): with two factors the criterion is identically zero, so no rotation
#'   is performed and the unrotated loadings are returned with a warning.
#'
#' Oblique rotations:
#' - **promax** is a fast two-step rotation: a varimax solution is raised to a power
#'   (controlled by `k` and `p_type`) to form a target that is then fitted obliquely.
#' - **quartimin** simplifies the variables (rows), like quartimax, so each loads mainly
#'   on one factor, but factors are allowed to correlate.
#' - **oblimin** is quartimin with a tunable argument, `gam` (default 0, i.e. quartimin
#'   itself); turning it up trades some of that row simplicity for more strongly
#'   correlated factors, and can drive the solution toward factor collapse, so inspect
#'   `Phi` before interpreting a fit with `gam > 0`.
#' - **simplimax** drives the `k` smallest loadings toward zero. Its criterion is only
#'   piecewise smooth and strongly multimodal, which makes it by far the most
#'   start-dependent rotation offered here: different random seeds can reach noticeably
#'   different solutions. Because every start is fully optimised (there is no screening
#'   stage), raising `random_starts` to several hundred costs proportionally more time but
#'   buys a better optimum.
#' - **bentlerQ** is the oblique Bentler invariant pattern simplicity criterion.
#' - **geominQ** is the oblique geomin criterion; it handles complex (cross-loading)
#'   structure well but is multimodal, so it benefits from more `random_starts` (and uses a
#'   more thorough multi-start search internally).
#' - **bifactorQ** is the oblique (correlated) Jennrich-Bentler bifactor criterion, with the
#'   same two-group-factor requirement as **bifactorT**.
#'
#' Prefer an oblique rotation unless there is a substantive reason to force the factors to
#' be uncorrelated: if `Phi` comes back near zero the oblique solution is essentially the
#' orthogonal one anyway, whereas imposing orthogonality on genuinely correlated factors
#' distorts the pattern.
#'
#' The criterion-based rotations (all except varimax and promax) are fitted by gradient
#' projection with `random_starts` random starts to guard against local minima; the
#' complexity criteria (simplimax and geominQ in particular) are the most multimodal. The
#' starts are drawn from the random-number generator, so different starts can reach
#' genuinely different optima and such a fit is reproducible only when the generator is
#' controlled: pass `seed`, or call [base::set.seed()] beforehand.
#'
#' ## Standard errors
#'
#' `se` selects whether and how standard errors (and matching confidence intervals) are
#' computed. Which quantities they cover depends on the method. The analytic methods
#' (`"information"` and `"sandwich"`) cover the unrotated loadings, the uniquenesses and
#' the communalities and, when a rotation is applied, the rotated loadings and -- for
#' oblique rotations -- the factor correlations and the structure coefficients. The
#' bootstrap (`"np-boot"`) covers the unrotated loadings, the residuals, and the fit
#' indices and, when a rotation is applied, the rotated loadings and -- for oblique
#' rotations -- the factor correlations and the structure coefficients; it reports no
#' uniqueness or communality standard errors.
#'
#' - **"none"** (default) computes no standard errors.
#' - **"information"** returns analytic standard errors from the expected (Fisher)
#'   information matrix of the maximum-likelihood solution, and therefore requires
#'   `estimator = "ML"` and `cor_method = "pearson"` (or `"fiml"`, see below). The rotated
#'   standard errors are obtained by propagating the unrotated-loading covariance through
#'   the rotation (Jennrich, 1973); because rotated quantities do not depend on how the
#'   unrotated solution happens to be oriented, they are directly comparable across
#'   programs. Unlike the bootstrap it also works from a correlation matrix as long as `N`
#'   is supplied. These standard errors assume multivariate normality and a correctly
#'   specified model; under heavy-tailed data or model misfit they can understate the
#'   sampling variability, where `"sandwich"` or `"np-boot"` are more robust.
#'
#'   The rotated loadings, `Phi`, the structure coefficients, the uniquenesses, and the
#'   communalities do not depend on how the unrotated solution happens to be oriented, and
#'   so are comparable across programs. The **unrotated** loading standard errors are not:
#'   a program using a different orientation will report different unrotated loading
#'   standard errors for the same fit. That orientation can also fail to be well defined --
#'   for example when two factors are of near-equal strength -- in which case the
#'   unrotated loadings have no well-defined sampling distribution and their standard
#'   errors diverge. `efa_fit()` detects this and returns `NA` for the unrotated loading
#'   standard errors with a warning, rather than reporting a number that looks like a
#'   standard error but is an artefact of the orientation; the rotated loadings, `Phi`, the
#'   structure coefficients, the uniquenesses, and the communalities are unaffected and
#'   still reported. Use those, or `se = "np-boot"`, when the unrotated loadings
#'   themselves are the quantity of interest. This detection applies to every analytic
#'   route (the Pearson and polychoric paths and the two-stage `cor_method = "fiml"`
#'   sandwich alike). A Heywood case (a uniqueness at its lower boundary) is
#'   separate and more severe: there
#'   the whole `SE`/`CI` block comes back `NA` with a warning, because the standard-error
#'   approximation fails for every parameter. The scaled chi-square (from `"sandwich"`, or
#'   from `cor_method = "fiml"`) does not rely on that approximation and is still
#'   reported, so the fit indices are unaffected.
#' - **"sandwich"** returns robust ("sandwich") standard errors estimated from raw data,
#'   combining the estimator weight with a distribution-free covariance of the
#'   correlations, so it stays valid under non-normality and weight misspecification
#'   (Browne, 1984; Satorra & Bentler, 1994). It is available either for ordinal data with
#'   `cor_method = "poly"` or `"tetra"` and `estimator` one of `"ML"`, `"ULS"`, or
#'   `"DWLS"`, or for continuous data with `cor_method = "pearson"` and
#'   `estimator = "ML"` or `"ULS"`. It reports the same coverage as `"information"` and
#'   additionally fills the model fit's chi-square block with a scaled chi-square (see
#'   *Fit indices*); the statistic reported as `chi` is always the scaled-and-shifted one
#'   (flagged by `chi_scaled_type`), and a mean-adjusted alternative is returned alongside
#'   it as `chi_mean_adjusted`.
#'   Because the covariance must describe the same cases as the correlation matrix, the
#'   sandwich (like `estimator = "DWLS"`) is computed on the listwise-complete cases; on
#'   data with missing values the reported `N`, the correlation matrix, and the point
#'   estimate therefore reflect the complete cases regardless of `use`.
#' - **"np-boot"** draws a non-parametric (case-resampling) bootstrap and needs raw data.
#'   A correlation matrix carries no cases to resample; alone among the unsupported
#'   combinations this one does not error but warns and downgrades `se` to `"none"`, so
#'   the fit is returned without an `SE` slot.
#'   It is the most general method -- available for any `estimator`, `rotation`, and
#'   `cor_method` -- and the most robust to non-normality and misfit, at the cost of speed;
#'   its intervals are bootstrap percentile intervals. The replicate fits run across
#'   replicates with the `future` framework; by default they run sequentially, but
#'   registering a plan with [future::plan()] runs them in parallel instead, as the
#'   examples show. With a fixed `seed` the bootstrap is reproducible and yields the same
#'   result regardless of the number of workers. Under `cor_method = "fiml"` each resample
#'   also re-runs the EM moment estimation and is therefore slow, so a smaller `b_boot`
#'   may be advisable.
#'
#'   The percentile intervals are centred on the point estimate for the loadings, the
#'   factor correlations, the structure coefficients, and the residuals, but **not** for
#'   the indices derived from the chi-square (`RMSEA`, `AIC`, `BIC`, and `ECVI`): a
#'   resample already carries the model's own misfit, so those intervals ride upward and
#'   can even place the point estimate below their own lower bound. Read them as a spread
#'   rather than a range for the point estimate; correcting that shift needs resampling
#'   from a population transformed to fit the model (Bollen & Stine, 1992), which is not
#'   what this bootstrap does. `CFI` and `TLI` are not affected, being ratios in which the
#'   baseline chi-square shifts along with the model one.
#'
#' The analytic methods (`"information"` and `"sandwich"`) are not available with the
#' `"promax"` or `"simplimax"` rotations, which have no supported analytic route for the
#' rotated standard errors; use `"np-boot"` there. Under `cor_method = "fiml"`,
#' `"information"` and `"sandwich"` instead return, for `estimator = "ML"` or `"ULS"`,
#' corrected two-stage sandwich standard errors (Yuan & Bentler, 2000; Savalei & Bentler,
#' 2009). `estimator = "PAF"` carries no Stage-2 weight to build the sandwich from, so use
#' `se = "np-boot"` there instead.
#'
#' ## Fit indices
#'
#' For ML and ULS, [efa_fit()] returns the model chi-square (with its p-value and degrees of
#' freedom), the Comparative Fit Index (CFI; Bentler, 1990), the Tucker-Lewis Index (TLI,
#' also called the non-normed fit index; Tucker & Lewis, 1973), the Root Mean Square Error
#' of Approximation (RMSEA) with its 90% confidence interval (Browne & Cudeck, 1992), the
#' Akaike and Bayesian Information Criteria (AIC, BIC), the Expected Cross-Validation Index
#' (ECVI; Browne & Cudeck, 1989), the Root Mean Squared Residual (RMSR), the Standardized
#' Root Mean Squared Residual (SRMR; Bentler, 1995), and the common-part-accounted-for
#' (CAF) index (Lorenzo-Seva, Timmerman, & Kiers, 2011). They come with the
#' independence-baseline statistics `chi_null`, `df_null`, and `p_null`. On the unscaled ML
#' and ULS paths `chi_null` is Bartlett's test of sphericity; a scaled chi-square, and the
#' two-stage statistic of a `cor_method = "fiml"` fit, each carry their own baseline
#' instead.
#'
#' The degrees of freedom depend on the number of variables and factors; the baseline
#' degrees of freedom `df_null` are \eqn{p(p - 1)/2} for \eqn{p} variables.
#'
#' RMSR is the root mean square of the off-diagonal residuals; SRMR rescales it by a
#' fixed factor that depends only on the number of variables. The print and summary
#' methods show SRMR, not RMSR; RMSR remains in the returned object for backward
#' compatibility. Both are computed over the same residuals, so an unavailable residual
#' leaves both `NA`. (\pkg{psych}'s `rms` uses a different divisor and equals
#' RMSR\eqn{/\sqrt{2}}; the two are not directly comparable.)
#'
#' The model chi-square is the Bartlett-corrected discrepancy (matching
#' [stats::factanal()] for ML). For ULS it is the same maximum-likelihood discrepancy,
#' evaluated at the ULS-fitted solution, as in [psych::fa()] -- not the least-squares
#' criterion \pkg{lavaan} reports as its standard ULS test statistic, so the two are not
#' comparable. AIC and BIC are built on this chi-square and can therefore be negative;
#' ECVI (built on the same chi-square plus a non-negative penalty) cannot. Because of the
#' Bartlett correction, ECVI differs slightly from the uncorrected Browne-Cudeck form
#' reported by \pkg{lavaan} and Mplus (their AIC/BIC use an unrelated log-likelihood-based
#' formula, so they are not comparable). On the unscaled ML/ULS path, CFI and
#' TLI are computed on a slightly different discrepancy scale than the reported
#' chi-square, so you cannot recompute one from the other by hand there; on the scaled
#' (sandwich) and FIML paths, the reported chi and chi_null are exactly the pair the
#' indices use.
#'
#' Which indices are reported depends on the estimator:
#' - **ML and ULS** compute the full set above.
#' - **PAF** returns only the descriptive residual indices (RMSR, SRMR, CAF) and df; the
#'   printed model-fit block shows CAF and SRMR. The chi-square-based indices are `NA`,
#'   because PAF minimises no discrepancy.
#' - **DWLS** by default returns only RMSR, SRMR, CAF, and df, because the ordinary
#'   maximum-likelihood discrepancy is not its fit function. When `se = "sandwich"`, a
#'   scaled chi-square and the CFI, TLI, and RMSEA derived from it are reported (AIC and
#'   BIC remain `NA`); that statistic is a two-stage correction applied to the polychoric
#'   correlation residuals (Browne, 1984), not identical to the full WLSMV test of
#'   \pkg{lavaan} or Mplus, which also corrects for the response thresholds.
#' - **`cor_method = "fiml"`** (with ML or ULS) reports two-stage-corrected statistics
#'   (Yuan, Marshall, & Bentler, 2002); AIC, BIC, and ECVI are left `NA`, as for any
#'   scaled chi-square. The correction itself can be degenerate -- typically with a small
#'   sample, a high proportion of missing values, or near-collinear variables -- in which
#'   case an uncorrected likelihood-ratio statistic is reported in its place with a
#'   warning, and the print methods label that line `uncorrected` rather than `scaled`;
#'   read its p-value and the CFI, TLI, and RMSEA derived from it as indicative only.
#'
#' Beyond the estimator, the chi-square and everything derived from it are `NA` whenever
#' the statistic itself is undefined: when `N` is not supplied, when the model is
#' underidentified (a negative df), and when a positive `N` is too small relative to the
#' number of variables and factors for the small-sample correction to remain valid. Each
#' case raises its own warning. The residual summaries (RMSR, SRMR, CAF) and the degrees
#' of freedom are still returned there, but residual size does not establish that a model
#' is identified: below zero degrees of freedom a near-zero residual is an artefact of
#' over-parameterisation, not close fit.
#'
#' Whenever the chi-square is a scaled one (`se = "sandwich"`, or a `cor_method = "fiml"`
#' fit whose correction could be formed), AIC, BIC, and ECVI are `NA`; see the
#' `fit_indices` entry in Value for the additional components then returned. AIC, BIC,
#' and ECVI are `NA` on every `cor_method = "fiml"` fit, including the uncorrected
#' fallback above. Lorenzo-Seva, Timmerman, and Kiers (2011) describe CAF as ranging from
#' 0 to 1, with values near 1 indicating close fit; that does not hold here, where a
#' well-fitting model produces a CAF near 0.5, not near 1. Read it as a relative rather
#' than an absolute measure.
#'
#' ## Available combinations
#'
#' Not every estimator, rotation, standard-error, and correlation method can be combined:
#'
#' - **Estimator and correlation method.** `estimator = "DWLS"` requires ordinal data with
#'   `cor_method = "poly"` or `"tetra"`. `cor_method = "fiml"` works with PAF, ML, and ULS
#'   (not DWLS) and needs raw data with missing values.
#' - **Standard errors.** `se = "information"` requires `estimator = "ML"` and
#'   `cor_method = "pearson"` or `"fiml"`, and can be computed from a correlation matrix
#'   when `N` is supplied. `se = "sandwich"` requires raw data,
#'   with either a polychoric/tetrachoric `cor_method` (ML, ULS, or DWLS) or a Pearson
#'   `cor_method` (ML or ULS); it is not available for PAF. Under `cor_method = "fiml"`,
#'   `"information"` and `"sandwich"` are available for ML and ULS only and both return the
#'   corrected two-stage sandwich. `se = "np-boot"` requires raw data and works with any
#'   estimator, rotation, and correlation method. Neither `"information"` nor `"sandwich"`
#'   is available with the `"promax"` or `"simplimax"` rotations.
#' - **Fit indices.** The chi-square-based indices are available for ML and ULS (and, as
#'   scaled statistics, for `cor_method = "fiml"` and for DWLS with `se = "sandwich"`); PAF
#'   and DWLS otherwise report only the descriptive residual indices.
#'
#' @return A list of class `c("efa", "EFA")` containing (a subset of) the following:
#'
#' \item{orig_R}{Original correlation matrix.}
#' \item{h2_init}{Initial communality estimates from PAF.}
#' \item{h2}{Final communality estimates from the unrotated solution.}
#' \item{orig_eigen}{Eigen values of the original correlation matrix.}
#' \item{init_eigen}{Initial eigenvalues, obtained from the correlation matrix
#'  with the initial communality estimates as diagonal in PAF.}
#' \item{final_eigen}{Eigenvalues obtained from the correlation matrix
#'  with the final communality estimates as diagonal.}
#' \item{iter}{For PAF, the number of iterations until convergence. For ML, ULS, and
#'  DWLS, the number of objective-function evaluations used by the optimiser (not the
#'  number of optimiser iterations).}
#' \item{convergence}{Integer convergence code (0 = converged), using the codes of
#'  [`stats::optim()`][stats::optim]. For ML and ULS it is the code from the bounded
#'  optimiser; for DWLS the fit runs a bounded warm start followed by an
#'  unconstrained polish, and the reported code is from the final polish. For PAF
#'  it is 1 if the maximum number of iterations was reached
#'  without meeting the convergence criterion and 0 otherwise. A non-zero code is
#'  also reported with a warning.}
#' \item{heywood}{A named integer vector indicating which variables have a
#'  Heywood (improper) case in the unrotated solution; empty if there are none.}
#' \item{unrot_loadings}{Loading matrix containing the final unrotated loadings.}
#' \item{vars_accounted}{Matrix of explained variances and sums of squared loadings. Based
#' on the unrotated loadings. Its rows are `"SS loadings"` and `"Prop Tot Var"`, followed by
#' `"Cum Prop Tot Var"`, `"Prop Comm Var"`, and `"Cum Prop Comm Var"`; those last three are
#' omitted for a single-factor solution, where they would only repeat the two above them. So
#' code that indexes a row by name (for example `["Prop Comm Var", ]`) must allow for the
#' two-row form at `n_factors = 1`.}
#' \item{fit_indices}{A named list of fit indices computed from the unrotated
#' loadings. ML and ULS report the full set: the model Chi Square (with its
#' p-value and df), CFI, TLI, RMSEA with its 90% confidence interval, AIC, BIC,
#' ECVI, RMSR, SRMR, and CAF. PAF and DWLS report only RMSR, SRMR, CAF, and df;
#' the Chi-Square-based indices are `NA` there, unless DWLS uses
#' `se = "sandwich"`, which fills the full block from a scaled Chi Square
#' instead. RMSR and SRMR are both included, but the print and summary methods
#' display only SRMR (see *Fit indices* in Details). Whenever the Chi Square is
#' scaled (`se = "sandwich"`, or a `cor_method = "fiml"` fit whose correction
#' could be formed), AIC, BIC, and ECVI are `NA`, and a few extra
#' scaled-statistic fields are appended for advanced diagnostics (see
#' *Fit indices* in Details). The list also carries `Fm`, the estimator's own
#' objective value at the solution -- on its own scale, not the discrepancy
#' the Chi Square is built from -- and the independence-model baseline
#' `chi_null`, `df_null`, and `p_null` that CFI and TLI are computed from.}
#' \item{model_implied_R}{The model implied correlation
#' matrix.}
#' \item{residuals}{Residual correlations, i.e., orig_R - model_implied_R}
#' \item{standardized_residuals}{Residual correlations standardized by their
#'  bootstrap standard errors. Only returned, if `se = "np-boot"`.}
#' \item{rot_loadings}{Loading matrix containing the final rotated loadings. For an oblique
#' rotation this is the pattern matrix -- each variable's unique contribution from each
#' factor, with the other factors partialled out -- and is the matrix normally interpreted
#' (see *Rotations*).}
#' \item{Phi}{The factor intercorrelations (only for oblique rotations).}
#' \item{Structure}{The structure matrix `rot_loadings %*% Phi`, holding the plain
#' variable-factor correlations, which are inflated by the factor intercorrelations (only
#' for oblique rotations).}
#' \item{rotmat}{The rotation matrix. The rotated loadings are recovered from the
#' unrotated loadings as `unrot_loadings %*% rotmat` for orthogonal rotations and
#' for promax, and as `unrot_loadings %*% t(solve(rotmat))` for the other oblique
#' rotations.}
#' \item{vars_accounted_rot}{Matrix of explained variances and sums of squared
#' loadings. Based on rotated loadings and, for oblique rotations, the factor
#' intercorrelations. Same rows as `vars_accounted`; it is returned only when a rotation
#' was actually applied, which never happens for a single-factor solution, so it always
#' has the five-row form.}
#' \item{settings}{A list of the settings used, including `seed` (`NULL` when none was
#' supplied), `input_type` (`"raw"` or `"correlation"`, what `x` was), and
#' `cor_method_used` (the correlation method that actually ran, `NA_character_` for a
#' correlation-matrix input, which consumes none). `cor_method` keeps the requested value
#' whether or not it was used. For the criterion rotations fitted by gradient projection
#' it additionally carries `rotation_diagnostics`, a list summarising the multi-start run:
#'  - `converged`: whether the start whose solution is returned reached the convergence
#'    tolerance (reported separately from the counts below because the returned solution
#'    is the one with the lowest criterion value, which need not be a converged one).
#'  - `n_starts_total`: the `random_starts` random starts plus the rational start.
#'  - `n_optimized`: how many of those starts were actually optimized -- fewer than
#'    `n_starts_total` whenever the solver screens the random starts and optimizes only
#'    the most promising ones.
#'  - `n_converged`: how many optimized starts reached the convergence tolerance.
#'  - `n_distinct_minima`: how many distinct local optima those converged starts found;
#'    more than one means the criterion is multimodal on these data.
#'  - `criterion_spread`: the range of the criterion values they attained.
#'  - `criterion_best`: the criterion value of the returned solution.
#'
#'  When `normalize = TRUE`, `criterion_best` and `criterion_spread` are evaluated on
#'  the Kaiser-normalized loadings the criterion is optimized on, not on the returned
#'  `rot_loadings`, so they are not directly comparable to a criterion recomputed from
#'  the returned loadings.}
#' \item{fiml}{Diagnostics of the FIML correlation's EM estimation, present only for
#' `cor_method = "fiml"`: whether it `converged` before `fiml_max_iter`, how many
#' iterations it took (`iter`), the number of distinct missingness patterns
#' (`n_patterns`), and the number of cases used (`n`, the reported `N`). A `converged`
#' of `FALSE` means the analysed correlation is the EM's last iterate, not the true
#' FIML estimate; this is flagged in the printed output. Set
#' `fiml_max_iter` and `fiml_tol` through [estimate_control()].}
#' \item{SE}{A named list of standard error matrices, returned only when `se` is not
#' `"none"` (see *Standard errors* in Details for what each method assumes and when a
#' component comes back `NA`). For `se = "np-boot"`: bootstrap SDs of the loadings
#' (rotated too, if a rotation was applied), the residuals, and the fit indices, plus
#' -- for oblique rotations -- `Phi` and the structure coefficients; `valid_replicates`
#' (and, when rotated, `valid_target_rotations`) records how many bootstrap replicates
#' each of those is based on. For `se = "information"` and `se = "sandwich"`: Wald-type
#' SEs for the loadings, the uniquenesses, and the communalities -- identical to each
#' other, since communality is just `1 - uniqueness` -- plus `Phi` and the structure
#' coefficients for oblique rotations. `"sandwich"` stays valid under non-normality;
#' `"information"` assumes the model is correctly specified.}
#' \item{CI}{A named list of confidence intervals of width `ci`, matching the
#' components of `SE`: percentile intervals for `se = "np-boot"`, Wald intervals for
#' `se = "information"` and `se = "sandwich"`. Only returned if `se` is not `"none"`.}
#' \item{replicates}{A named list of raw bootstrap replicate arrays -- the aligned
#' loadings, `Phi`, structure coefficients, residuals, and fit indices behind `SE` and
#' `CI` -- with one replicate per array's last dimension (first dimension for
#' `fit_indices`). Failed replicates are left `NA`. Populated only for
#' `se = "np-boot"`; `NULL` otherwise.}
#' \item{vcov_unrot_loadings}{The full unrotated-loading covariance matrix behind
#' `SE$unrot_loadings`: a `p * n_factors` by `p * n_factors` matrix in column-major
#' `vec(Lambda)` order, with rows and columns labelled `"<variable>_<factor>"`.
#' Populated for `se = "information"` and `se = "sandwich"` -- always the unrotated
#' block, even when a rotation is applied -- and `NA`-filled if the analytic covariance
#' is unreliable (a Heywood case or a singular information matrix); `NULL` for
#' `se = "np-boot"` and `se = "none"`. It can be populated even when
#' `SE$unrot_loadings` is `NA`: a weakly determined rotational orientation invalidates
#' only the marginal SEs, not the underlying covariance.}
#' \item{Gamma}{The asymptotic covariance of the off-diagonal sample correlations --
#' the meat of the robust sandwich SEs. A `p (p - 1) / 2` by `p (p - 1) / 2` matrix,
#' rows and columns ordered by [utils::combn()] over the column pairs and labelled
#' `"<var_i>-<var_j>"`. Populated for `se = "sandwich"` on the polychoric/tetrachoric
#' and Pearson paths; `NULL` otherwise (including under `cor_method = "fiml"`, whose
#' meat is not returned). It typically dominates a sandwich fit's size (about 11 MB at
#' `p = 50`) and is kept because [efa_mi()] needs it from each per-imputation fit to
#' build the pooled covariance.}
#'
#' @source Bollen, K. A., & Stine, R. A. (1992). Bootstrapping goodness-of-fit measures
#' in structural equation models. Sociological Methods & Research, 21, 205–229.
#' doi: 10.1177/0049124192021002004
#' @source Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
#' of implementations of principal axis factoring and promax rotation in R and SPSS.
#' Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x
#' @source Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
#' rotation to oblique simple structure. British Journal of Statistical Psychology,
#' 17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x
#' @source Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
#' Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
#' Research, 46, 340-364, doi: 10.1080/00273171.2011.564527
#' @source Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
#' factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233
#' @source Lawley, D. N., & Maxwell, A. E. (1971). Factor analysis as a statistical
#' method (2nd ed.). Butterworths.
#' @source Cudeck, R. (1989). Analysis of correlation matrices using covariance
#' structure models. Psychological Bulletin, 105, 317–327.
#' doi: 10.1037/0033-2909.105.2.317
#' @source Olkin, I., & Siotani, M. (1976). Asymptotic distribution of functions of a
#' correlation matrix. In S. Ikeda (Ed.), Essays in probability and statistics
#' (pp. 235–251). Shinko Tsusho.
#' @source Jennrich, R. I. (1973). Standard errors for obliquely rotated factor
#' loadings. Psychometrika, 38, 593–604. doi: 10.1007/BF02291497
#' @source Zhang, G., & Preacher, K. J. (2015). Factor rotation and standard errors
#' in exploratory factor analysis. Journal of Educational and Behavioral Statistics,
#' 40, 579–603. doi: 10.3102/1076998615606098
#' @source Browne, M. W. (1984). Asymptotically distribution-free methods for the
#' analysis of covariance structures. British Journal of Mathematical and Statistical
#' Psychology, 37, 62–83. doi: 10.1111/j.2044-8317.1984.tb00789.x
#' @source Satorra, A., & Bentler, P. M. (1994). Corrections to test statistics and
#' standard errors in covariance structure analysis. In A. von Eye & C. C. Clogg (Eds.),
#' Latent variables analysis: Applications for developmental research (pp. 399–419). Sage.
#' @source Asparouhov, T., & Muthén, B. (2010). Simple second order chi-square
#' correction. Mplus Technical Appendix.
#' @source Muthén, B., du Toit, S. H. C., & Spisic, D. (1997). Robust inference using
#' weighted least squares and quadratic estimating equations in latent variable modeling
#' with categorical and continuous outcomes. Unpublished manuscript.
#' @source Yuan, K.-H., & Bentler, P. M. (2000). Three likelihood-based methods for mean
#' and covariance structure analysis with nonnormal missing data. Sociological Methodology,
#' 30, 165–200. doi: 10.1111/0081-1750.00078
#' @source Yuan, K.-H., Marshall, L. L., & Bentler, P. M. (2002). A unified approach to
#' exploratory factor analysis with missing data, nonnormal data, and in the presence of
#' outliers. Psychometrika, 67, 95–121. doi: 10.1007/BF02294711
#' @source Savalei, V., & Bentler, P. M. (2009). A two-stage approach to missing data:
#' Theory and application to auxiliary variables. Structural Equation Modeling, 16, 477–497.
#' doi: 10.1080/10705510903008238
#' @source Little, R. J. A., & Rubin, D. B. (2002). Statistical analysis with missing data
#' (2nd ed.). Wiley.
#' @source Bartlett, M. S. (1951). The effect of standardization on a Chi-square
#' approximation in factor analysis. Biometrika, 38, 337–344.
#' @source Bentler, P. M. (1990). Comparative fit indexes in structural models.
#' Psychological Bulletin, 107, 238–246. doi: 10.1037/0033-2909.107.2.238
#' @source Tucker, L. R., & Lewis, C. (1973). A reliability coefficient for maximum
#' likelihood factor analysis. Psychometrika, 38, 1–10. doi: 10.1007/BF02291170
#' @source Browne, M. W., & Cudeck, R. (1989). Single sample cross-validation indices
#' for covariance structures. Multivariate Behavioral Research, 24, 445–455.
#' doi: 10.1207/s15327906mbr2404_4
#' @source Browne, M. W., & Cudeck, R. (1992). Alternative ways of assessing model fit.
#' Sociological Methods & Research, 21, 230–258. doi: 10.1177/0049124192021002005
#' @source Bentler, P. M. (1995). EQS structural equations program manual. Multivariate
#' Software.
#'
#' @family factor analysis
#'
#' @seealso [estimate_control()] and [rotate_control()] for the estimation and rotation
#'  tuning knobs. [efa_retain()] for choosing `n_factors`, and [efa_scores()],
#'  [efa_reliability()], [efa_schmid_leiman()], and [efa_compare()] for working with the
#'  fitted solution.
#'
#' @export
#'
#' @examples
#'
#' # Principal axis factoring with oblimin rotation
#' mod_oblimin <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                        rotation = "oblimin")
#' mod_oblimin
#' summary(mod_oblimin)
#'
#' # ML estimation with oblimin rotation
#' mod_oblimin <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                        estimator = "ML", rotation = "oblimin")
#' mod_oblimin
#' summary(mod_oblimin)
#'
#' # Tuning knobs are supplied through the control objects. Here the SPSS preset is
#' # used for the estimation and rotation, with the maximum PAF iterations raised.
#' mod_spss <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                     rotation = "promax",
#'                     estimate_control = estimate_control(type = "SPSS", max_iter = 500),
#'                     rotate_control = rotate_control(type = "SPSS"))
#' mod_spss
#'
#' # Analytic (expected-information) standard errors for the above
#' ML_info <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                    estimator = "ML", rotation = "oblimin", se = "information")
#' ML_info
#' summary(ML_info)
#'
#' \donttest{
#' # Robust (sandwich) standard errors and a scaled chi-square for ordinal raw data.
#' # These need a polychoric/tetrachoric correlation method and estimator ML, ULS, or DWLS.
#' DWLS_rob <- efa_fit(DOSPERT_raw, n_factors = 6, cor_method = "poly",
#'                     estimator = "DWLS", rotation = "oblimin", se = "sandwich")
#' DWLS_rob
#' summary(DWLS_rob)
#'
#' # The same robust SEs and scaled chi-square for continuous data: a Pearson
#' # correlation with estimator ML or ULS (the fourth-moment ADF covariance).
#' ML_rob <- efa_fit(GRiPS_raw, n_factors = 1, cor_method = "pearson",
#'                   estimator = "ML", rotation = "none", se = "sandwich")
#' ML_rob
#' summary(ML_rob)
#' }
#'
#' \donttest{
#' # Two-stage FIML correlations from raw data with missing values: the saturated
#' # multivariate-normal moments are EM-estimated (assuming the data are missing at
#' # random) and the standardized covariance is analysed.
#' x_miss <- GRiPS_raw
#' x_miss[cbind(1:20, 1)] <- NA
#' efa_fiml <- efa_fit(x_miss, n_factors = 1, estimator = "ML", cor_method = "fiml")
#' efa_fiml
#' }
#'
#' \dontrun{
#' # Bootstrap standard errors from raw data, reproducible via a fixed seed and run
#' # in parallel across replicates. future::plan() returns the plan it replaces, so
#' # on.exit() puts the session back as it was -- also if the fit fails.
#' efa_boot <- local({
#'   old_plan <- future::plan(future::multisession, workers = 2)
#'   on.exit(future::plan(old_plan), add = TRUE)
#'   efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF", rotation = "none",
#'           se = "np-boot", b_boot = 1000, seed = 42)
#' })
#' }
#'
efa_fit <- function(x, n_factors, N = NA,
                    estimator = c("PAF", "ML", "ULS", "MINRES", "DWLS"),
                    rotation = c("none", "varimax", "equamax", "quartimax", "geominT",
                                 "bentlerT", "bifactorT", "promax", "oblimin",
                                 "quartimin", "simplimax", "bentlerQ", "geominQ",
                                 "bifactorQ"),
                    se = c("none", "information", "sandwich", "np-boot"),
                    cor_method = c("pearson", "spearman", "kendall", "poly", "tetra",
                                   "fiml"),
                    use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                            "everything", "na.or.complete"),
                    estimate_control = NULL,
                    rotate_control = NULL,
                    b_boot = 1000, ci = .95, seed = NULL, ...) {

  # The control arguments share their names with the constructor functions, so a literal
  # `estimate_control = estimate_control()` default would recurse into the promise. Default
  # them to NULL and build the preset controls here instead; in call position R resolves the
  # namesake to the (function) constructor, skipping the NULL argument binding.
  if (is.null(estimate_control)) estimate_control <- estimate_control()
  if (is.null(rotate_control)) rotate_control <- rotate_control()

  # Perform argument checks
  .reject_flat_knobs(...names())
  .assert_cor_input(x)

  # n_factors has no default: catch its omission here so the most common first-call
  # mistake gets a classed condition pointing at the function that answers it,
  # rather than R's bare "argument is missing" error.
  if (missing(n_factors)) {
    cli::cli_abort(
      c("{.arg n_factors} is required.",
        "i" = "Use {.fn efa_retain} to decide how many factors to extract."),
      class = "efa_missing_n_factors"
    )
  }

  # The companion mistake to omitting n_factors: handing over the whole retention object. The
  # shared assertion below names the class the value is not, never the component that holds
  # the counts, so name it here. Both retention classes carry that component: `efa_retain()`
  # one count per criterion, a single criterion one per variant it reports (two for MAP, three
  # for the eigenvalue-based ones). Only a one-variant criterion therefore yields a value the
  # assertion would take, so the number is reported and the reader is asked to choose.
  if (inherits(n_factors, c("efa_retain", "efa_retention"))) {
    n_suggested <- length(n_factors$n_factors)
    cli::cli_abort(
      c("{.arg n_factors} must be a single count, not an {.cls {class(n_factors)[[1]]}} object.",
        "i" = "Its {.code $n_factors} holds {n_suggested} suggested count{?s}; supply
               {?it/one of them}."),
      class = "efa_n_factors_object"
    )
  }

  estimator <- .match_arg_ci(estimator)
  # "MINRES" is a synonym for "ULS" (same estimator); resolve it once here so the
  # rest of efa_fit() and the reported settings use the single canonical name.
  if (estimator == "MINRES") estimator <- "ULS"
  rotation <- .match_arg_ci(rotation)
  se <- .match_arg_ci(se)
  np_boot <- se == "np-boot"
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)

  # The dots only carry extras for the rotation engine, which reads them by exact name:
  # `maxit` for the GPArotation-style engines, plus the selected criterion's parameter
  # (`gam` for oblimin, `delta` for geomin). Any other dot -- a misspelled name (e.g.
  # `gamma` for `gam`), another criterion's parameter, or any extra with varimax/promax,
  # which take none -- would be silently dropped and the fit would quietly run the engine
  # defaults instead, so it is rejected here. With no rotation there is no engine, so
  # nothing can consume the dots at all.
  if (...length() > 0L) {
    allowed <- .rotation_dot_extras(rotation)
    nms <- ...names()
    if (is.null(nms)) nms <- rep("", ...length())
    bad <- unique(setdiff(nms[nzchar(nms)], allowed))
    if (length(bad) > 0L || any(!nzchar(nms))) {
      if (rotation == "none") {
        msg <- if (length(bad) > 0L) {
          "{.arg {bad}} {?is/are} not used: with {.code rotation = \"none\"} no rotation engine
           runs, so {.arg ...} must be empty."
        } else {
          "{.arg ...} must be empty with {.code rotation = \"none\"}: no rotation engine runs,
           so the extra arguments are not used."
        }
        info <- c("i" = "Check for a misspelled argument name (for example {.arg estimate_control}
                         or {.arg rotate_control}).")
      } else {
        msg <- if (length(bad) > 0L) {
          "{.arg {bad}} {?is/are} not consumed by the {.val {rotation}} rotation."
        } else {
          "The arguments in {.arg ...} must be named; the {.val {rotation}} rotation engine
           reads its extras by exact name."
        }
        info <- c(
          if (length(allowed) > 0L) {
            c("i" = "The {.val {rotation}} rotation accepts {.arg {allowed}} through {.arg ...}.")
          } else {
            c("i" = "The {.val {rotation}} rotation takes no extra arguments through {.arg ...}.")
          },
          "i" = "Check for a misspelled engine argument; the rotation tuning knobs live in
                 {.fn rotate_control}."
        )
      }
      cli::cli_abort(c(msg, info), class = "efa_unused_dots")
    }
  }

  .assert_estimate_control(estimate_control)
  if (!inherits(rotate_control, "efa_rotate_control")) {
    cli::cli_abort(
      "{.arg rotate_control} must be a control object from {.fn rotate_control}.",
      class = "efa_control_input"
    )
  }

  # Unbundle the estimation and rotation tuning knobs from the two control objects (already
  # validated by their constructors) into the local variables the fitting pipeline expects.
  # Each control carries its own `type` preset, so estimation and rotation may resolve from
  # different presets. The extra rotation arguments the rotate control stored are merged with
  # `...` (with `...` winning on a name clash) and forwarded to the rotation engine.
  est_type <- estimate_control$type
  init_comm <- estimate_control$init_comm
  criterion <- estimate_control$criterion
  criterion_type <- estimate_control$criterion_type
  max_iter <- estimate_control$max_iter
  abs_eigen <- estimate_control$abs_eigen
  start_method <- estimate_control$start_method
  fiml_max_iter <- estimate_control$fiml_max_iter
  fiml_tol <- estimate_control$fiml_tol

  rot_type <- rotate_control$type
  normalize <- rotate_control$normalize
  precision <- rotate_control$precision
  order_type <- rotate_control$order_type
  varimax_type <- rotate_control$varimax_type
  p_type <- rotate_control$p_type
  k <- rotate_control$k
  random_starts <- rotate_control$random_starts
  rot_extra_args <- modifyList(rotate_control$extra_args, list(...))

  if (is.na(start_method) && estimator == "ML") {
    cli::cli_abort(
      c("{.arg start_method} must be set when {.code estimator = \"ML\"}.",
        "i" = "Set {.arg start_method} to {.val psych} or {.val factanal}."),
      class = "efa_ml_start_missing"
    )
  }

  # Detect a correlation-matrix input once, up front: the FIML guard below and the
  # bootstrap/DWLS guards further down all branch on it, and resolving it before R is
  # validated/smoothed keeps those warnings from being pre-empted by a singular or
  # non-positive-definite matrix.
  is_cormat <- .is_cormat(x)

  # Two-stage / full-information ML correlations are EM-estimated from raw data with missing
  # values (Yuan, Marshall, & Bentler, 2002). Reject the input and option combinations that path
  # cannot honour here, before any computation, so they fail with a dedicated message rather than
  # a downstream one. The analytic standard errors are handled by the corrected two-stage sandwich
  # (.se_fiml()) for ML and ULS; DWLS, a correlation-matrix input, and PAF with analytic standard
  # errors are unsupported.
  if (cor_method == "fiml") {
    if (is_cormat) {
      cli::cli_abort(
        c("{.code cor_method = \"fiml\"} needs raw data, not a correlation matrix.",
          "x" = "A correlation matrix carries no cases to estimate the FIML moments from.",
          "i" = "Supply raw data (with missing values), or choose another {.arg cor_method}."),
        class = "efa_fiml_needs_raw"
      )
    }
    if (estimator == "DWLS") {
      cli::cli_abort(
        c("{.code estimator = \"DWLS\"} is not compatible with {.code cor_method = \"fiml\"}.",
          "x" = "DWLS needs a polychoric asymptotic covariance, which the continuous FIML correlation does not provide.",
          "i" = "Use {.code estimator = \"ML\"}, {.val ULS}, or {.val PAF}, or {.code cor_method = \"poly\"}/{.val tetra} for DWLS."),
        class = "efa_fiml_unsupported_method"
      )
    }
    # The corrected two-stage sandwich (.se_fiml()) reuses the Stage-2 ML/ULS weight; PAF minimises
    # no discrepancy and so carries no weight for it, exactly as the polychoric/continuous sandwich
    # rejects PAF. Reject the analytic-SE request here (the bootstrap stays available for PAF).
    if (se %in% c("information", "sandwich") && !(estimator %in% c("ML", "ULS"))) {
      cli::cli_abort(
        c("Analytic standard errors under {.code cor_method = \"fiml\"} require {.code estimator = \"ML\"} or {.val ULS}.",
          "x" = "You requested {.code estimator = {.val {estimator}}}.",
          "i" = "{.val PAF} minimises no discrepancy, so it has no weight for the two-stage sandwich; use {.code se = \"np-boot\"} for {.val {estimator}}."),
        class = "efa_se_unsupported"
      )
    }
    if (use != "pairwise.complete.obs") {
      cli::cli_warn(
        c("{.arg use} is ignored when {.code cor_method = \"fiml\"}.",
          "i" = "FIML uses every case and handles the missingness itself, so {.code use = {.val {use}}} has no effect."),
        class = "efa_fiml_use_ignored"
      )
    }
  }

  # Analytic standard errors cover only a subset of estimators and rotations. Reject the
  # unsupported combinations here, before any computation, with a clear pointer to the
  # bootstrap. Information-matrix SEs are derived from the ML discrepancy and so require
  # estimator = "ML"; promax (a two-step target rotation) and simplimax (a non-smooth, piecewise
  # criterion) have no usable analytic rotation Jacobian; PAF has no discrepancy-based information
  # from which a sandwich could be built.
  if (se == "information" && estimator != "ML" && cor_method != "fiml") {
    cli::cli_abort(
      c("{.code se = \"information\"} is only available for {.code estimator = \"ML\"}.",
        "x" = "You requested {.code estimator = {.val {estimator}}}.",
        "i" = "Use {.code se = \"np-boot\"} for {.val {estimator}}."),
      class = "efa_se_unsupported"
    )
  }
  # The expected information is built from the normal-theory asymptotic covariance of PEARSON
  # correlations, so it presumes normal-theory sampling behaviour of the analysed matrix. A
  # polychoric/tetrachoric correlation additionally carries first-stage threshold estimation error
  # and a rank correlation is not a Pearson moment at all; neither is accounted for here. The
  # sandwich covers both, so point at it rather than at the bootstrap.
  if (se == "information" && !(cor_method %in% c("pearson", "fiml"))) {
    cli::cli_abort(
      c("{.code se = \"information\"} requires {.code cor_method = \"pearson\"} or {.val fiml}.",
        "x" = "You requested {.code cor_method = {.val {cor_method}}}.",
        "i" = "The expected information assumes normal-theory sampling behaviour of the analysed correlations.",
        # From a bare correlation matrix the sandwich is unavailable too (it needs the raw data its
        # meat is estimated from), so pointing there would be a dead end.
        "i" = if (is_cormat) {
          "No analytic standard error is available for a {.val {cor_method}} correlation matrix; supply the raw data and use {.code se = \"sandwich\"}."
        } else {
          "Use {.code se = \"sandwich\"} for {.val {cor_method}}."
        }),
      class = "efa_se_unsupported"
    )
  }
  if (se == "sandwich" && estimator == "PAF" && cor_method != "fiml") {
    cli::cli_abort(
      c("{.code se = \"sandwich\"} is not available for {.code estimator = \"PAF\"}.",
        "i" = "Use {.code se = \"np-boot\"} for {.val PAF}."),
      class = "efa_se_unsupported"
    )
  }
  if (se %in% c("information", "sandwich") && rotation %in% c("promax", "simplimax")) {
    cli::cli_abort(
      c("{.code se = {.val {se}}} is not available with {.code rotation = {.val {rotation}}}.",
        "i" = "Use {.code se = \"np-boot\"} for {rotation}-rotated solutions."),
      class = "efa_se_unsupported"
    )
  }
  # Sandwich (robust) SEs combine the estimator weight with an asymptotic-distribution-free
  # covariance of the correlations (the robust meat), so they require raw data (enforced once N
  # is resolved below) to estimate that covariance. Two paths supply it: the polychoric/
  # tetrachoric asymptotic covariance for ordinal data (any estimator), or the fourth-moment
  # (Browne, 1984) covariance for continuous data with cor_method = "pearson" and estimator ML
  # or ULS. Spearman/Kendall correlations and continuous DWLS have no such covariance.
  if (se == "sandwich" && cor_method != "fiml" && !.is_poly_cor(cor_method) &&
      !(cor_method == "pearson" && estimator %in% c("ML", "ULS"))) {
    cli::cli_abort(
      c("{.code se = \"sandwich\"} is not available for this correlation/estimator combination.",
        "x" = "You requested {.code cor_method = {.val {cor_method}}} with {.code estimator = {.val {estimator}}}.",
        "i" = "Use {.code cor_method = \"poly\"}/{.val tetra} (any estimator), or {.code cor_method = \"pearson\"} with {.code estimator = \"ML\"} or {.val ULS}; otherwise use {.code se = \"np-boot\"}."),
      class = "efa_se_unsupported"
    )
  }

  .assert_args({
    checkmate::assert_count(n_factors)
    # N = 0 is not an unknown sample size but an impossible one: it would silently NA the
    # chi-square block through the same route a valid-but-small N does. NA remains the
    # supported way of saying "sample size unknown".
    checkmate::assert_count(N, positive = TRUE, na.ok = TRUE)
    checkmate::assert_count(max_iter, na.ok = TRUE)
    checkmate::assert_choice(init_comm, c("smc", "mac", "unity", NA))
    checkmate::assert_number(criterion, lower = 0, upper = 1, na.ok = TRUE)
    if (!is.na(criterion) && criterion >= 1) {
      cli::cli_abort(
        c("{.arg criterion} must be smaller than 1.",
          "x" = "You supplied {.arg criterion} = {criterion}.",
          "i" = "Use a small positive convergence tolerance such as {.val 0.001}."),
        class = "efa_criterion_too_large"
      )
    }
    checkmate::assert_choice(criterion_type, c("max_individual", "sum", NA))
    checkmate::assert_flag(abs_eigen, na.ok = TRUE)
    checkmate::assert_number(k, na.ok = TRUE)
    checkmate::assert_choice(varimax_type, c("svd", "kaiser", NA))
    checkmate::assert_flag(normalize, na.ok = TRUE)
    checkmate::assert_choice(p_type, c("unnorm", "norm", NA))
    checkmate::assert_number(precision, lower = 0, upper = 1)
    checkmate::assert_choice(order_type, c("eigen", "ss_factors", NA))
    checkmate::assert_integerish(b_boot, len = 1, any.missing = FALSE)
    # A bootstrap standard error is the dispersion across replicates, so two is the smallest number
    # from which one is defined at all: at b_boot = 1 every SE is the sd() of a single value and comes
    # back NA, and the percentile interval collapses onto that replicate.
    if (b_boot < 2) {
      cli::cli_abort(
        c("{.arg b_boot} must be at least 2.",
          "x" = "You supplied {.arg b_boot} = {b_boot}.",
          "i" = "A bootstrap standard error is the spread across replicates and is undefined below
                 two of them. The default is {.val {1000}}."),
        class = "efa_b_boot_too_small"
      )
    }
    checkmate::assert_number(ci)
    # Both ends of the level are open. Every path turns `ci` into a quantile of the sampling
    # distribution: the analytic paths into z = qnorm(1 - (1 - ci) / 2), the bootstrap into the
    # matching percentile pair. At ci = 1 that quantile is infinite and each analytic Wald bound
    # comes back -Inf/Inf; at ci = 0 it is zero and each interval collapses onto the point
    # estimate. Neither value is a confidence level, and both used to pass without a condition.
    if (ci <= 0 || ci >= 1) {
      cli::cli_abort(
        c("{.arg ci} must be greater than 0 and smaller than 1.",
          "x" = "You supplied {.arg ci} = {ci}.",
          "i" = "At 1 every analytic bound is infinite, and at 0 every interval has zero width.
                 The default is {.val {0.95}}."),
        class = "efa_ci_out_of_bounds"
      )
    }
    checkmate::assert_int(seed, null.ok = TRUE)
  })

  # The common-factor model needs at least one factor and fewer factors than
  # variables: with n_factors >= n_variables it is not identified and the
  # eigenvalue-based extraction in the ML, ULS, and PAF fitters reads past the
  # available eigenvalues (undefined behaviour in an unchecked build), and with
  # n_factors = 0 there is no model to fit. Both bounds are checked here so the
  # C++ kernels' own bound checks are never the user-facing error.
  n_vars <- ncol(x)
  # With a single variable no number of factors is admissible, so only state the
  # range when one exists (it would otherwise read "1 to 0").
  range_hint <- if (n_vars > 1) " The admissible range is 1 to {n_vars - 1}." else ""
  if (n_factors < 1) {
    cli::cli_abort(
      c("{.arg n_factors} must be at least 1.",
        "x" = paste0("You requested {n_factors} factor{?s}.", range_hint),
        "i" = "A factor retention criterion that returns 0 means no factor is worth
               extracting, so there is no exploratory factor model to fit."),
      class = "efa_too_few_factors"
    )
  }
  if (n_factors >= n_vars) {
    cli::cli_abort(
      c("{.arg n_factors} must be smaller than the number of variables.",
        "x" = paste0("You requested {n_factors} factor{?s} for {n_vars} variable{?s}.",
                     range_hint),
        "i" = "Extract fewer factors."),
      class = "efa_too_many_factors"
    )
  }

  # DWLS weights each polychoric correlation residual by the inverse of its asymptotic
  # variance, so it needs a polychoric/tetrachoric asymptotic covariance. That is only
  # available from raw ordinal data with cor_method = "poly" or "tetra"; there is no
  # fallback to unit weights. Resolve this before any computation so an unsupported
  # request fails with a single clear error rather than downstream.
  if (estimator == "DWLS" && !(.is_poly_cor(cor_method) && !is_cormat)) {
    cli::cli_abort(
      c("{.code estimator = \"DWLS\"} requires a polychoric asymptotic covariance.",
        "x" = if (is_cormat) {
          "You supplied a correlation matrix, so no asymptotic covariance can be estimated."
        } else {
          "{.code cor_method = {.val {cor_method}}} does not produce one."
        },
        "i" = "Supply raw ordinal data with {.code cor_method = \"poly\"} or {.code \"tetra\"}."),
      class = "efa_dwls_no_acov"
    )
  }

  # A correlation matrix carries no cases to resample, so the bootstrap is impossible.
  # Analytic SEs (information/sandwich) need only R, the loadings, the uniquenesses and N,
  # so they remain available from a correlation matrix as long as N is supplied (checked
  # below, once N is resolved).
  if (is_cormat && isTRUE(np_boot)) {

    cli::cli_warn(
      c("Cannot compute bootstrap standard errors from correlation matrix.",
      "x" = "You've supplied {.var se} = {.val {se}}, but {.var x} is a correlation matrix.",
      "i" = "Setting {.var se} to {.val none}. Rerun with raw data to calculate bootstrap SEs."),
      class = "efa_boot_cormat"
    )

    np_boot <- FALSE
    se <- "none"

  }

  # A fixed seed makes the whole fit reproducible: every stochastic step downstream of
  # this point draws from the state it sets. That covers the rotation's random starts on
  # the point estimate -- the criterion rotations draw `random_starts` random orthogonal
  # starts from the ambient stream, so without a seed a simplimax or geominQ fit is not
  # reproducible run to run -- and, when `se = "np-boot"`, the case resampling, the
  # replicate rotations, and the Procrustes random starts as well. The bootstrap is also
  # independent of the number of parallel workers: the case resampling advances the
  # global RNG by a b_boot-dependent amount, and the parallel replicate fit then adds a
  # fixed, worker-count-independent step when future.seed = TRUE derives a per-replicate
  # L'Ecuyer stream. Both advances are deterministic given the seed, so the downstream
  # draws -- and the result -- are identical at any number of workers. The caller's RNG
  # stream is saved and restored afterwards -- or, if none existed yet, the state
  # set.seed() creates is removed again -- so efa_fit() leaves no side effect on it.
  .set_local_seed(seed)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                             N_policy = "optional",
                             # Sandwich SEs need the full asymptotic covariance of the
                             # correlations (the robust meat); for DWLS its diagonal also
                             # supplies the per-element weights, so "full" subsumes the DWLS
                             # "diag" request. dwls = TRUE builds those weights; the ML / ULS
                             # sandwich path uses the meat alone. A correlation matrix carries no
                             # raw data to estimate the covariance from, so do not request one
                             # there (the sandwich is rejected just below); requesting it would
                             # only draw a spurious "acov ignored" warning before that abort.
                             acov = if (se == "sandwich" && !is_cormat && cor_method != "fiml") "full"
                                    else if (estimator == "DWLS") "diag" else "none",
                             dwls = estimator == "DWLS",
                             # EM budget for the two-stage FIML moments; ignored by every other
                             # cor_method.
                             fiml_max_iter = fiml_max_iter, fiml_tol = fiml_tol,
                             # The psych preset tolerates a singular correlation matrix only
                             # for PAF, whose starting communalities fall back to a
                             # pseudo-inverse. No other estimator has that fallback -- ML and
                             # ULS solve R for their starting values and the ML discrepancy
                             # needs log|R|, and DWLS would silently drop to a flat start --
                             # so the check stays on for them under every preset.
                             check_singular = !(est_type == "psych" && estimator == "PAF"),
                             posdef_abort = est_type == "SPSS")
  R <- prep$R
  N <- prep$N
  # DWLS weight matrix (1 / diag(Gamma)); NULL for the other estimators.
  weights <- prep$weights
  # Full asymptotic covariance of the off-diagonal correlations (the sandwich meat); NULL unless
  # se = "sandwich" on a raw-data path (polychoric/tetrachoric, or continuous Pearson).
  Gamma <- prep$Gamma
  # Two-stage / full-information ML moments for the fit-index likelihood-ratio chi-square; the
  # saturated mean/covariance and log-likelihood EM-estimated in .prepare_cor_input() are
  # carried alongside the raw data so .gof() can evaluate the model and baseline FIML log-
  # likelihoods. Reuse prep$fiml (no second EM run); NULL for every other cor_method.
  # The Stage-1 EM diagnostics ride along so `.efa_core()` can report them on the fitted
  # object: whether the EM met its tolerance, how many iterations and missingness patterns it
  # used, and how many rows carried information.
  fiml_pt <- if (cor_method == "fiml") {
    list(data = x, mu = prep$fiml$mu, sigma = prep$fiml$sigma, logl = prep$fiml$logl,
         converged = prep$fiml$converged, iter = prep$fiml$iter,
         n_patterns = prep$fiml$n_patterns, n = prep$fiml$n)
  }
  # On an analytic-SE fit the saturated FIML correlation covariance is needed twice from the same
  # moments -- by the scaled chi-square (.fiml_scaled_test, via .gof()) and by the SE sandwich
  # (.se_fiml_core) -- so build it once here and cache it on the fiml list; both read it and skip the
  # rebuild. Only the analytic-SE path duplicates it: with se = "none"/"np-boot" only .gof() forms it
  # (once), and the bootstrap replicate lists omit the field and recompute per replicate. NULL on a
  # degenerate covariance, where each consumer falls back to its own guarded recompute.
  if (!is.null(fiml_pt) && se %in% c("information", "sandwich")) {
    fiml_pt$acov_cor <- tryCatch(
      .fiml_saturated_acov(fiml_pt$data, fiml_pt$mu, fiml_pt$sigma)$cor,
      error = function(e) NULL)
  }

  # Analytic SEs scale the inverse information by 1 / (N - 1), so they need the sample
  # size. Raw data always supplies it (N = number of rows); a correlation matrix does not,
  # so require it explicitly there.
  if (se %in% c("information", "sandwich") && is.na(N)) {
    cli::cli_abort(
      c("{.code se = {.val {se}}} requires the sample size {.arg N}.",
        "i" = "Supply {.arg N} when {.arg x} is a correlation matrix."),
      class = "efa_se_no_n"
    )
  }

  # The sandwich meat (the asymptotic covariance of the correlations) can only be estimated from
  # raw data; a correlation matrix carries no such covariance, so reject that combination
  # explicitly (rather than failing later when the meat is NULL).
  if (se == "sandwich" && is.null(Gamma) && cor_method != "fiml") {
    cli::cli_abort(
      c("{.code se = \"sandwich\"} requires raw data to estimate the asymptotic covariance of the correlations.",
        "x" = "You supplied a correlation matrix.",
        "i" = "Provide raw data (ordinal with {.code cor_method = \"poly\"}/{.val tetra}, or continuous with {.code cor_method = \"pearson\"}), or use {.code se = \"np-boot\"}."),
      class = "efa_se_unsupported"
    )
  }

  if (!is_cormat && isTRUE(np_boot)) {

    m <- ncol(R)
    # Resample the cases the correlation matrix was actually built from. Under
    # listwise deletion that is the complete cases (N of them), not the first N
    # row positions, so the case bootstrap stays a faithful resample of the
    # estimator that produced R (Efron & Tibshirani, 1993). FIML uses every row
    # that carries at least one observed value (fully-missing rows are dropped, so
    # N counts only these) and `use` does not apply; resampling exactly those rows
    # -- not all nrow(x) positions -- keeps each replicate's sample size equal to N.
    rows <- if (cor_method == "fiml") {
      which(rowSums(!is.na(x)) > 0L)
    } else if (.is_listwise_use(use) || estimator == "DWLS") {
      which(stats::complete.cases(x))
    } else {
      seq_len(nrow(x))
    }

    # create bootstrap samples and from these, correlation matrices
    R_boot_array <- array(NA_real_, c(m, m, b_boot), dimnames = list(colnames(x),
                                                               colnames(x),
                                                               NULL))

    poly_cor <- .is_poly_cor(cor_method)
    tetra_cor <- cor_method == "tetra"
    fiml_cor <- cor_method == "fiml"
    dwls <- estimator == "DWLS"

    # DWLS reweights each replicate by the inverse of its own polychoric asymptotic
    # variances, so the per-element weights are recomputed alongside the matrix and
    # carried into the lean fit; NULL for the other estimators, which need no weights.
    # The weights are positional, so no dimnames are needed.
    W_boot_array <- if (dwls) array(NA_real_, c(m, m, b_boot)) else NULL

    # FIML carries each replicate's resample and its own EM moments (saturated mean, covariance,
    # and log-likelihood) into the lean fit, so the per-replicate fit indices use the same
    # likelihood-ratio chi-square as the point estimate. The resample is stored as the row
    # INDICES, not as the resampled matrix: the list is a closure global of the replicate fits, so
    # a matrix per replicate would be held here and serialised to every parallel worker on top of
    # it, growing as B * N * p where the indices grow as B * N (at N = 3123, p = 30, B = 1000 that
    # is the difference between roughly 715 MB and 3 MB). `.boot_fun()` receives the raw data once
    # and rebuilds each slice. One list element per replicate (NULL for a dropped resample); NULL
    # for the other cor_methods.
    fiml_boot <- if (fiml_cor) vector("list", b_boot) else NULL

    for (boot_i in seq_len(b_boot)) {
      ind <- sample(rows, size = N, replace = TRUE)
      if (poly_cor) {
        # A resample can be degenerate -- a constant column, a pair with no
        # overlapping cases, a numerically uncomputable matrix, or (for DWLS) a pair
        # whose asymptotic variance is non-positive -- and make .polychoric() or the
        # weight construction fail. Any genuine bug would already have surfaced on the
        # point-estimate fit over the full data above, so a failure here is necessarily
        # resample-specific; fall back to an all-NA matrix so the replicate is dropped at
        # the fit stage, mirroring how stats::cor() returns NA for a degenerate Pearson
        # resample and how .boot_fun() drops unfittable replicates. The bootstrap is
        # parallelised at the fit (across replicates, via future). DWLS requests the diagonal
        # ACOV and builds the weights inside the same try so the matrix and weights share one
        # resample and a degenerate weight drops the replicate too.
        rep_i <- tryCatch(
          suppressWarnings({
            poly <- .polychoric(x[ind, , drop = FALSE], nearest_pd = FALSE,
                                binary_only = tetra_cor,
                                acov = if (dwls) "diag" else "none",
                                label_acov = FALSE)
            list(R = poly$R,
                 W = if (dwls) .poly_weight_matrix(poly$acov, m) else NULL)
          }),
          error = function(e) NULL)
        if (is.null(rep_i)) {
          R_boot_array[,, boot_i] <- matrix(NA_real_, m, m)
        } else {
          R_boot_array[,, boot_i] <- rep_i$R
          if (dwls) W_boot_array[,, boot_i] <- rep_i$W
        }
      } else if (fiml_cor) {
        # Two-stage FIML: re-estimate the EM moments on the resample and standardise to a
        # correlation, mirroring the point-estimate path. The moments, the saturated
        # log-likelihood, and the resample's row indices are retained so the lean fit can
        # form this replicate's FIML likelihood-ratio chi-square. A degenerate resample (a
        # constant or collinear column, an EM breakdown) makes .fiml_em_moments() abort; fall
        # back to an all-NA matrix so the replicate is dropped at the fit stage, as on the poly
        # and Pearson paths. The EM recompute is serial here; the replicate fits parallelise
        # downstream.
        em_i <- tryCatch(suppressWarnings(.fiml_em_moments(x[ind, , drop = FALSE],
                                                           max_iter = fiml_max_iter,
                                                           tol = fiml_tol)),
                         error = function(e) NULL)
        if (is.null(em_i)) {
          R_boot_array[,, boot_i] <- matrix(NA_real_, m, m)
        } else {
          R_boot_array[,, boot_i] <- stats::cov2cor(em_i$sigma)
          fiml_boot[[boot_i]] <- list(rows = ind, mu = em_i$mu,
                                      sigma = em_i$sigma, logl = em_i$logl)
        }
      } else {
        R_boot_array[,, boot_i] <- stats::cor(x[ind, , drop = FALSE], use = use,
                                              method = cor_method)
      }
    }

  }

  core_args <- list(
    R = R, N = N, weights = weights, Gamma = Gamma,
    R_boot_array = if (isTRUE(np_boot)) R_boot_array else NULL,
    W_boot_array = if (isTRUE(np_boot)) W_boot_array else NULL,
    fiml = fiml_pt, fiml_boot = if (isTRUE(np_boot)) fiml_boot else NULL,
    np_boot = np_boot, b_boot = b_boot, estimator = estimator, rotation = rotation,
    type = est_type, rot_type = rot_type, n_factors = n_factors, se = se, ci = ci,
    use = use, cor_method = cor_method, max_iter = max_iter, init_comm = init_comm,
    criterion = criterion, criterion_type = criterion_type,
    abs_eigen = abs_eigen, start_method = start_method, normalize = normalize,
    precision = precision, order_type = order_type, varimax_type = varimax_type,
    P_type = p_type, k = k, randomStarts = random_starts)

  # Splice the criterion-specific rotation extras (from `rotate_control()` or `...`) into
  # `.efa_core()`'s dots, exactly where the old flat `...` used to land. `.efa_core()` still
  # declares the former flat tuning knobs (`type`, `max_iter`, `k`, ...) as named formals, so
  # a caller that forwards one of them through `...` -- e.g. the retention criteria, efa_group,
  # or EFA_POOLED passing a former flat argument -- would otherwise supply that formal twice
  # and abort with "matched by multiple actual arguments". Drop any extra whose name is already
  # an explicit `core_args` entry (those knobs now live in the control objects, so a bare copy
  # is silently ignored); genuine rotation extras such as `maxit` or `gam` are not core_args and
  # pass through untouched.
  extra <- rot_extra_args[setdiff(names(rot_extra_args), names(core_args))]
  out <- do.call(.efa_core, c(core_args, extra))

  # Provenance .efa_core() cannot record: it starts from an already-prepared correlation
  # matrix, so it can see neither what `x` was nor how the run was seeded. `cor_method`
  # keeps the requested value (a correlation-matrix input ignores it, and the settings
  # echo has always reported it either way); `cor_method_used` says what actually ran.
  out$settings <- append(
    out$settings,
    list(input_type = if (is_cormat) "correlation" else "raw",
         cor_method_used = if (is_cormat) NA_character_ else cor_method),
    after = match("cor_method", names(out$settings)))
  out$settings <- append(out$settings, list(seed = seed),
                         after = match("ci", names(out$settings)))
  out
}

# Explanation for a non-zero optimiser convergence code, branched on what the code means:
# 1 is the fixed iteration limit, 51 and 52 are L-BFGS-B's line-search warning and error
# (which are not iteration limits and are not fixed by running longer).
.nonconvergence_cause <- function(convergence) {
  if (isTRUE(convergence == 1)) {
    paste("It reached the optimiser's fixed iteration limit before meeting the",
          "convergence tolerance; that limit is not user-adjustable.")
  } else if (isTRUE(convergence %in% c(51, 52))) {
    paste("The bounded optimiser's line search failed, which typically means the",
          "objective is flat or ill-conditioned near the point it stopped at.")
  } else {
    "It stopped before meeting the convergence tolerance."
  }
}

# Fit the common-factor model from already-prepared inputs: a correlation matrix R, the
# sample size N, optional DWLS weights, the optional sandwich meat Gamma, and -- for the
# bootstrap -- pre-resampled correlation/weight arrays. Split out from efa_fit() so multiple-
# imputation pooling can drive the same estimate -> rotate -> standard-error pipeline on
# pooled inputs (the MI2S route in EFA_POOLED()) without re-entering efa_fit()'s raw-data
# preparation and input guards. `fiml` (point estimate) and `fiml_boot` (per-replicate)
# carry the two-stage EM moments + raw data for the FIML likelihood-ratio fit indices and
# default to NULL; EFA_POOLED()'s MI2S route never supplies them (it is gated to poly/tetra/
# pearson), so its fit indices stay on the standard discrepancy path.
.efa_core <- function(R, N, weights = NULL, Gamma = NULL,
                      R_boot_array = NULL, W_boot_array = NULL,
                      fiml = NULL, fiml_boot = NULL,
                      np_boot = FALSE, b_boot = 1000, estimator, rotation, type,
                      rot_type = type,
                      n_factors, se = "none", ci = .95,
                      use = "pairwise.complete.obs", cor_method = "pearson",
                      max_iter = NA, init_comm = NA, criterion = NA,
                      criterion_type = NA, abs_eigen = NA, start_method = "psych",
                      normalize = TRUE, precision = 1e-5, order_type = NA,
                      varimax_type = NA, P_type = NA, k = NA, randomStarts = 100,
                      ...) {

  # Check if model is identified

  # calculate degrees of freedom
  m <- ncol(R)
  df <- .efa_df(m, n_factors)

  if(df < 0){

    chisq_block <- .fit_unavailable_text("chisq_block")
    residuals_kept <- .fit_unavailable_text("residuals_kept")
    not_identification <- .fit_unavailable_text("residuals_not_identification")
    cli::cli_warn(
      c("The model is underidentified ({.code df = {df}}); {chisq_block}.",
        "!" = "{residuals_kept}",
        "!" = "{not_identification}",
        "i" = "Use fewer factors or more indicators."),
      class = "efa_underidentified"
    )

  } else if (df == 0){

    cli::cli_warn(
      c("The model is just identified ({.code df = 0}).",
        "i" = "Consider fewer factors or more indicators."),
      class = "efa_just_identified"
    )

  }

  # run factor analysis with the respective estimator

  if (estimator %in% c("ML", "ULS", "DWLS")) {

    if (type == "SPSS") {

      cli::cli_warn(
        c("Only {.val PAF} is validated against the SPSS implementation.",
          "i" = "{.val {estimator}} results may differ from those returned by SPSS."),
        class = "efa_spss_method_untested"
      )

    }

    if (is.na(N)) {

      cli::cli_warn(
        c("{.arg N} is {.val NA}; not all fit indices could be computed.",
          "i" = "Provide {.arg N} or raw data to compute all fit indices."),
        class = "efa_fit_na_n"
      )

    } else if (estimator %in% c("ML", "ULS") && is.null(fiml) && se != "sandwich" &&
               df >= 0 && .bartlett_mult(N, m, n_factors) <= 0) {

      # A known but very small N relative to the number of variables turns Bartlett's
      # small-sample multiplier non-positive, so .gof() leaves the whole chi-square block
      # undefined -- the same output PAF produces, and previously with no explanation. The
      # warning belongs here rather than in .gof(), which the bootstrap calls once per
      # replicate. The FIML path (is.null(fiml)) uses a likelihood-ratio statistic that
      # carries no Bartlett correction, DWLS reports no ML discrepancy at all, and
      # se = "sandwich" refills the whole block from the scaled statistic afterwards, which
      # the multiplier never enters -- so in none of those cases is the block missing.
      chisq_block <- .fit_unavailable_text("chisq_block")
      mult_note <- .fit_unavailable_text("bartlett_mult", N = N, p = m, q = n_factors)
      residuals_kept <- .fit_unavailable_text("residuals_kept")
      cli::cli_warn(
        c("{.arg N} is too small for the small-sample correction; {chisq_block}.",
          "i" = "{mult_note}",
          "i" = "{residuals_kept}"),
        class = "efa_fit_indices_undefined"
      )

    }

  }

  fit_out <- .estimate_model(R, method = estimator, n_factors = n_factors, N = N,
                             type = type, max_iter = max_iter,
                             init_comm = init_comm, criterion = criterion,
                             criterion_type = criterion_type,
                             abs_eigen = abs_eigen, start_method = start_method,
                             weights = weights, fiml = fiml)

  # Surface Heywood cases from the point-estimate solution (the detector runs in
  # .finalize_fit for every fit). This fires once per efa_fit() call; efa_average(),
  # which fits one EFA per grid cell, suppresses these per-model warnings and
  # reports a single summary instead.
  if (length(fit_out$heywood) > 0) {
    # Fall back to the indices if the correlation matrix carried no variable names.
    heywood_vars <- names(fit_out$heywood) %||% as.character(fit_out$heywood)
    cli::cli_warn(
      c(paste("{cli::qty(heywood_vars)}Heywood case{?s} detected for {.val {heywood_vars}}:",
              "the solution is improper (a communality at or above 1, or a uniqueness",
              "fixed at the estimation boundary)."),
        "i" = "Interpret the affected loadings and uniquenesses with caution."),
      class = "efa_heywood"
    )
  }

  # Surface optimiser non-convergence from the point-estimate solution for the
  # iterative estimators (ML, ULS, DWLS), whose fitters return a non-zero
  # convergence code when the optimiser stops before meeting its tolerance. PAF
  # raises its own non-convergence warning from inside .PAF(). This fires once per
  # efa_fit() call; the bootstrap replicates suppress their per-fit warnings and are
  # tallied separately in .boot_fun().
  if (estimator %in% c("ML", "ULS", "DWLS") && isTRUE(fit_out$convergence != 0)) {
    cli::cli_warn(
      c("The {.val {estimator}} optimiser did not converge (convergence code {fit_out$convergence}).",
        "i" = paste(.nonconvergence_cause(fit_out$convergence),
                    "The results may not be interpretable."),
        "i" = paste("Try extracting fewer factors, or check the correlation matrix for",
                    "near-collinear variables.")),
      class = "efa_nonconvergence"
    )
  }

  # Only the bootstrap produces replicate fits; the analytic SE methods leave this NULL.
  boot_fits <- NULL
  if (isTRUE(np_boot)) {

    boot_fits <- .boot_fun(R_boot_array, b_boot, .estimate_model,
                           # .estimate_model arguments:
                           method = estimator, n_factors = n_factors, N = N,
                           type = type, max_iter = max_iter,
                           init_comm = init_comm, criterion = criterion,
                           criterion_type = criterion_type,
                           abs_eigen = abs_eigen, start_method = start_method,
                           # Each replicate fits only the quantities the bootstrap
                           # aggregation consumes (see .finalize_fit()).
                           lean = TRUE,
                           # DWLS carries the per-replicate weight matrices; NULL otherwise.
                           weights_array = W_boot_array,
                           # FIML carries the per-replicate EM moments and resample indices for
                           # the likelihood-ratio fit indices, plus one copy of the raw data the
                           # indices point into; NULL for the other cor_methods.
                           fiml_list = fiml_boot,
                           fiml_data = fiml$data)

  }

  # rotate factor analysis results
  if (rotation == "none") {

    output <- fit_out
    boot_rot <- "none"

  } else {

    # Only promax and varimax are validated against the SPSS implementation.
    if (rot_type == "SPSS" && !rotation %in% c("promax", "varimax")) {

      cli::cli_warn(
        c("Only the {.val promax} and {.val varimax} rotations are validated against the SPSS implementation.",
          "i" = "{.val {rotation}} results may differ from those returned by SPSS."),
        class = "efa_spss_rotation_untested"
      )

    }

    rot_out <- .rotate_model(fit_out, rotation = rotation, type = rot_type,
                             normalize = normalize, precision = precision,
                             order_type = order_type, varimax_type = varimax_type,
                             P_type = P_type, k = k, randomStarts = randomStarts,
                             ...)

    boot_rot <- .rotation_family(rotation)

  }

  if (rotation != "none"){

    if(estimator %in% c("ULS", "DWLS")){

      settings <- rot_out$settings
      output <- c(fit_out, within(rot_out, rm(settings)),
                  settings = list(settings))

    } else {

      settings <- c(fit_out$settings, rot_out$settings)
      output <- c(within(fit_out, rm(settings)), within(rot_out, rm(settings)),
                  settings = list(settings))

    }

  }

  # Add settings used to output
  settings_EFA <- list(
    estimator = estimator,
    # back-compat alias: the output settings keep the former field name alongside the
    # current one, as they do for the frozen P_type/randomStarts rotate keys
    method = estimator,
    rotation = rotation,
    type = type,
    n_factors = n_factors,
    N = N,
    use = use,
    cor_method = cor_method,
    se = se,
    b_boot = b_boot,
    ci = ci
  )

  if(estimator %in% c("ULS", "DWLS") & rotation == "none"){

    output <- c(output, settings = list(settings_EFA))

  } else {

    settings <- c(settings_EFA, output$settings)

    output <- c(within(output, rm(settings)),
                settings = list(settings))

  }

  # Stage-1 diagnostics of the two-stage FIML correlation. The EM's non-convergence warning
  # fires once, inside the correlation preparation, and callers that suppress per-fit warnings
  # (efa_average(), the bootstrap replicates) swallow it, so record it on the object as well:
  # a `converged` of FALSE says the analysed matrix is the last iterate, not the FIML estimate.
  # Present only under cor_method = "fiml", which is the only path that estimates them.
  if (!is.null(fiml)) {
    output$fiml <- list(converged = fiml$converged, iter = fiml$iter,
                        n_patterns = fiml$n_patterns, n = fiml$n)
  }

  # Persist the full unrotated loading covariance (populated below for the analytic SE
  # methods) and the asymptotic covariance of the off-diagonal correlations (populated
  # for `se = "sandwich"`). Both stay present-but-NULL elsewhere so downstream consumers
  # (multiple-imputation pooling) can probe them by name without an `is.null(names(...))`
  # dance.
  output["vcov_unrot_loadings"] <- list(NULL)
  output["Gamma"] <- list(Gamma)

  if (se != "none") {
    if (rotation == "none") {
      L_rot <- NULL
      rot_info <- NULL
    } else {
      L_rot <- rot_out$rot_loadings
      # Analytic rotated SEs re-solve the rotation when finite-differencing its Jacobian, so they
      # need the converged transformation, the factor correlations, the resolved Kaiser-
      # normalization flag, and the criterion's tuning arguments (the same `.gpf_crit` defaults the
      # rotation itself used). NULL for the bootstrap, which carries its own replicate rotations.
      rot_info <- if (se %in% c("information", "sandwich")) {
        list(rotation = rotation,
             rotmat = rot_out$rotmat,
             rot_loadings = rot_out$rot_loadings,
             Phi = rot_out$Phi,
             normalize = rot_out$settings$normalize,
             crit_args = list(gam = .gpf_crit(list(...), "gam", 0),
                              delta = .gpf_crit(list(...), "delta", 0.01)))
      } else {
        NULL
      }
    }
    boot_out <- .compute_se_ci(fit_out, L_rot, se_method = se,
                               boot_fits = boot_fits, boot_rot = boot_rot,
                               ci = ci, b = b_boot, N = N, rot_info = rot_info,
                               gamma = Gamma, method = estimator, fiml = fiml)
    # The sandwich also returns the robust scaled chi-square block; it is patched into the
    # fit indices below rather than carried in the SE schema, so strip it before merging.
    scaled_test <- boot_out$scaled_test
    boot_out$scaled_test <- NULL
    output$SE <- boot_out$SE
    output$CI <- boot_out$CI
    # Single-bracket list assignment preserves a present-but-NULL slot for the
    # analytic SE methods (information, sandwich) where there are no replicate
    # arrays; `output$replicates <- NULL` would remove the slot entirely.
    output["replicates"] <- list(boot_out$replicates)
    # The analytic SE paths (information, sandwich) carry through the full unrotated
    # loading covariance the marginal SEs were derived from; the bootstrap path leaves it
    # NULL. Same single-bracket pattern as `replicates` so the slot is present-but-NULL
    # on the bootstrap rather than absent.
    output["vcov_unrot_loadings"] <- list(boot_out$vcov_unrot_loadings)
    # Only the bootstrap yields a sampling SD for every residual; the analytic methods do
    # not, so standardise the residuals only when those SEs are available.
    if (!is.null(boot_out$SE$residuals)) {
      # The residual diagonal is fixed at 0 with SE 0, so 0/0 would yield NaN on the
      # diagonal; set it to 0 so the off-diagonal standardised residuals are usable.
      std_resid <- output$residuals / boot_out$SE$residuals
      diag(std_resid) <- 0
      output$standardized_residuals <- std_resid
    }
    # Fill the chi-square block of the fit indices with the robust scaled chi-square
    # (Satorra-Bentler / scaled-shifted): .gof() leaves it undefined for DWLS, and for
    # ML/ULS the unscaled discrepancy chi-square is not robust to the ordinal weighting.
    if (se == "sandwich" && !is.null(scaled_test)) {
      output$fit_indices <- .apply_scaled_test(output$fit_indices, scaled_test, N)
    }
  }

  class(output) <- c("efa", "EFA")

  return(output)

}



.boot_fun <- function(x, b, call_fun, ..., weights_array = NULL, fiml_list = NULL,
                      fiml_data = NULL) {

  # The per-replicate fits are independent and estimation is RNG-free, so they are
  # run in parallel across replicates at the R/process level with future.apply. A
  # parallel processing plan can be selected with future::plan(); the default plan
  # runs sequentially. future.seed = TRUE assigns each replicate its own reproducible
  # L'Ecuyer-CMRG stream, so the bootstrap is reproducible and independent of the
  # number of workers (any RNG a fitter might draw is bound to the replicate index,
  # never to the worker).
  #
  # A replicate whose (possibly degenerate) resampled correlation matrix cannot be
  # fit returns NULL and is skipped later, rather than aborting the whole call.
  #
  # Per-replicate warnings are suppressed here: they repeat identical information b
  # times. The type/preset-override notice depends only on the (type, pinned
  # arguments) combination, which is the same for every replicate and already
  # surfaced once by the point-estimate fit; the iterative fitter's max-iteration
  # warning would otherwise fire once per non-converged replicate. Non-convergence
  # is instead tallied and reported once, after all replicates have been fitted.
  boot_list <- future.apply::future_lapply(seq_len(b), function(boot_i) {
    # DWLS passes the replicate's own weight matrix; the other estimators ignore the
    # NULL (.estimate_model()'s weights default).
    w_i <- if (is.null(weights_array)) NULL else weights_array[,, boot_i]
    # FIML passes the replicate's own EM moments plus its resample for the likelihood-ratio fit
    # indices; the other paths ignore the NULL (.estimate_model()'s fiml default). The resample is
    # rebuilt here from the stored row indices and the single copy of the raw data, so what the
    # list -- and hence what future.apply serialises to each worker -- carries per replicate is
    # N integers rather than an N x p matrix. A dropped resample is a NULL element and stays NULL.
    f_i <- if (is.null(fiml_list) || is.null(fiml_list[[boot_i]])) {
      NULL
    } else {
      fl <- fiml_list[[boot_i]]
      list(data = fiml_data[fl$rows, , drop = FALSE], mu = fl$mu,
           sigma = fl$sigma, logl = fl$logl)
    }
    tryCatch(suppressWarnings(call_fun(x[,, boot_i], ..., weights = w_i, fiml = f_i)),
             error = function(e) NULL)
  }, future.seed = TRUE)

  n_nonconverged <- sum(vapply(boot_list,
                               function(fit_i) isTRUE(fit_i$convergence != 0),
                               logical(1)))

  if (n_nonconverged > 0L) {
    cli::cli_warn(
      c("{n_nonconverged} of {b} bootstrap replicate{?s} did not converge.",
        "i" = "Their bootstrap standard errors and confidence intervals may be unreliable."),
      class = "efa_boot_nonconvergence"
    )
  }

  boot_list

}

# Unit tests for .extract_data() (R/averaging.R), which stacks the per-run efa_fit()
# solutions of efa_average() into the loading/Phi/fit-index arrays the averaging step
# consumes. Covered across three fixtures: a clean run, one carrying a non-converged
# fit, and a rotated run.

efa_list <- list(EFA(test_models$baseline$cormat, n_factors = 3, N = 500),
                 EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     method = "ML"),
                 EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     method = "ULS"),
                 suppressWarnings(EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     max_iter = 2)))


ext_a <- .extract_data(efa_list, test_models$baseline$cormat, 3, 4, "none", .3)

efa_list_er <- list(EFA(test_models$baseline$cormat, n_factors = 3, N = 500),
                 EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     method = "ML"),
                 EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     method = "ULS"),
                 try(suppressWarnings(EFA(test_models$baseline$cormat, n_factors = 15, N = 500,
                         type = "psych")),
                     silent = TRUE))

ext_er <- .extract_data(efa_list_er, test_models$baseline$cormat, 3, 4, "none", .3)

efa_list_rot <- list(EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                         rotation = "promax"),
                 EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     method = "ML", rotation = "promax"),
                 EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                     method = "ULS", rotation = "promax"))

ext_rot <- .extract_data(efa_list_rot, test_models$baseline$cormat, 3, 3, "promax",
                         .3)

test_that(".extract_data works", {
  ### tests for ext_a with one non-convergence; no rotation; no error
  expect_type(ext_a, "list")
  expect_named(ext_a, c("L", "L_corres", "phi", "extract_phi", "h2",
                        "vars_accounted", "for_grid"))
  expect_named(ext_a$for_grid, c("errors", "error_m", "converged", "heywood",
                                 "admissible", "chisq", "p_chi", "caf", "cfi",
                                 "rmsea", "aic", "bic", "srmr", "tli", "ecvi",
                                 "rmsr"))
  checkmate::expect_array(ext_a$L)
  expect_equal(dim(ext_a$L), c(ncol(test_models$baseline$cormat), 3, 3))
  checkmate::expect_array(ext_a$L_corres)
  expect_equal(dim(ext_a$L_corres), c(ncol(test_models$baseline$cormat), 3, 3))
  expect_equal(ext_a$phi, NA)
  expect_equal(ext_a$extract_phi, FALSE)
  checkmate::expect_matrix(ext_a$h2)
  expect_equal(dim(ext_a$h2), c(4, ncol(test_models$baseline$cormat)))
  expect_equal(ext_a$for_grid$errors, rep(FALSE, 4))
  expect_true(all(is.na(ext_a$for_grid$error_m)))
  expect_equal(ext_a$for_grid$converged, c(0, 0, 0, 1))
  expect_equal(ext_a$for_grid$heywood, c(FALSE, FALSE, FALSE, NA))
  expect_equal(sum(is.na(ext_a$for_grid$chisq)), 2)
  expect_type(ext_a$for_grid$chisq, "double")
  expect_equal(sum(is.na(ext_a$for_grid$p_chi)), 2)
  expect_type(ext_a$for_grid$p_chi, "double")
  expect_equal(round(ext_a$for_grid$caf, 2), c(0.5, 0.5, 0.5, NA))
  expect_equal(ext_a$for_grid$cfi > .95, c(NA, TRUE, TRUE, NA))
  expect_equal(ext_a$for_grid$rmsea < .05, c(NA, TRUE, TRUE, NA))
  expect_equal(sign(ext_a$for_grid$aic), c(NA, -1, -1, NA))
  expect_equal(sign(ext_a$for_grid$bic), c(NA, -1, -1, NA))
  # SRMR/RMSR are residual-based (also computed for PAF); TLI/ECVI need a
  # chi-square and are NA for PAF. The non-converged run contributes NA to all.
  expect_type(ext_a$for_grid$srmr, "double")
  expect_equal(sum(is.na(ext_a$for_grid$srmr)), 1)
  expect_type(ext_a$for_grid$rmsr, "double")
  expect_equal(sum(is.na(ext_a$for_grid$rmsr)), 1)
  expect_type(ext_a$for_grid$tli, "double")
  expect_equal(sum(is.na(ext_a$for_grid$tli)), 2)
  expect_type(ext_a$for_grid$ecvi, "double")
  expect_equal(sum(is.na(ext_a$for_grid$ecvi)), 2)
  checkmate::expect_array(ext_a$vars_accounted)
  expect_equal(dim(ext_a$vars_accounted), c(3, 3, 3))

  ### tests for ext_er with one error; no rotation
  expect_type(ext_er, "list")
  expect_named(ext_er, c("L", "L_corres", "phi", "extract_phi", "h2",
                         "vars_accounted", "for_grid"))
  expect_named(ext_er$for_grid, c("errors", "error_m", "converged", "heywood",
                                  "admissible", "chisq", "p_chi", "caf", "cfi",
                                  "rmsea", "aic", "bic", "srmr", "tli", "ecvi",
                                  "rmsr"))
  checkmate::expect_array(ext_er$L)
  expect_equal(dim(ext_er$L), c(ncol(test_models$baseline$cormat), 3, 3))
  checkmate::expect_array(ext_er$L_corres)
  expect_equal(dim(ext_er$L_corres), c(ncol(test_models$baseline$cormat), 3, 3))
  expect_equal(ext_er$phi, NA)
  expect_equal(ext_er$extract_phi, FALSE)
  checkmate::expect_matrix(ext_er$h2)
  expect_equal(dim(ext_er$h2), c(4, ncol(test_models$baseline$cormat)))
  expect_equal(ext_er$for_grid$errors, c(rep(FALSE, 3), TRUE))
  expect_equal(sum(is.na(ext_er$for_grid$error_m)), 3)
  expect_equal(ext_er$for_grid$converged, c(0, 0, 0, NA))
  expect_equal(ext_er$for_grid$heywood, c(FALSE, FALSE, FALSE, NA))
  expect_equal(sum(is.na(ext_er$for_grid$chisq)), 2)
  expect_type(ext_er$for_grid$chisq, "double")
  expect_equal(sum(is.na(ext_er$for_grid$p_chi)), 2)
  expect_type(ext_er$for_grid$p_chi, "double")
  expect_equal(round(ext_er$for_grid$caf, 2), c(0.5, 0.5, 0.5, NA))
  expect_equal(ext_er$for_grid$cfi > .95, c(NA, TRUE, TRUE, NA))
  expect_equal(ext_er$for_grid$rmsea < .05, c(NA, TRUE, TRUE, NA))
  expect_equal(sign(ext_er$for_grid$aic), c(NA, -1, -1, NA))
  expect_equal(sign(ext_er$for_grid$bic), c(NA, -1, -1, NA))
  expect_type(ext_er$for_grid$srmr, "double")
  expect_equal(sum(is.na(ext_er$for_grid$srmr)), 1)
  expect_type(ext_er$for_grid$rmsr, "double")
  expect_equal(sum(is.na(ext_er$for_grid$rmsr)), 1)
  expect_type(ext_er$for_grid$tli, "double")
  expect_equal(sum(is.na(ext_er$for_grid$tli)), 2)
  expect_type(ext_er$for_grid$ecvi, "double")
  expect_equal(sum(is.na(ext_er$for_grid$ecvi)), 2)
  checkmate::expect_array(ext_er$vars_accounted)
  expect_equal(dim(ext_er$vars_accounted), c(3, 3, 3))


  ### tests for ext_rot with no errors; promax rotation
  expect_type(ext_rot, "list")
  expect_named(ext_rot, c("L", "L_corres", "phi", "extract_phi", "h2",
                          "vars_accounted", "for_grid"))
  expect_named(ext_rot$for_grid, c("errors", "error_m", "converged", "heywood",
                                   "admissible", "chisq", "p_chi", "caf", "cfi",
                                   "rmsea", "aic", "bic", "srmr", "tli", "ecvi",
                                   "rmsr"))
  checkmate::expect_array(ext_rot$L)
  expect_equal(dim(ext_rot$L), c(ncol(test_models$baseline$cormat), 3, 3))
  checkmate::expect_array(ext_rot$L_corres)
  expect_equal(dim(ext_rot$L_corres), c(ncol(test_models$baseline$cormat), 3, 3))
  checkmate::expect_array(ext_rot$phi)
  expect_equal(dim(ext_rot$phi), c(3, 3, 3))
  checkmate::expect_matrix(ext_rot$h2)
  expect_equal(dim(ext_rot$h2), c(3, ncol(test_models$baseline$cormat)))
  expect_equal(ext_rot$for_grid$errors, rep(FALSE, 3))
  expect_equal(ext_rot$for_grid$converged, c(0, 0, 0))
  expect_equal(ext_rot$for_grid$heywood, c(FALSE, FALSE, FALSE))
  expect_equal(sum(is.na(ext_rot$for_grid$chisq)), 1)
  expect_type(ext_rot$for_grid$chisq, "double")
  expect_equal(sum(is.na(ext_rot$for_grid$p_chi)), 1)
  expect_type(ext_rot$for_grid$p_chi, "double")
  expect_equal(round(ext_rot$for_grid$caf, 2), c(0.5, 0.5, 0.5))
  expect_equal(ext_rot$for_grid$cfi > .95, c(NA, TRUE, TRUE))
  expect_equal(ext_rot$for_grid$rmsea < .05, c(NA, TRUE, TRUE))
  expect_equal(sign(ext_rot$for_grid$aic), c(NA, -1, -1))
  expect_equal(sign(ext_rot$for_grid$bic), c(NA, -1, -1))
  expect_type(ext_rot$for_grid$srmr, "double")
  expect_equal(sum(is.na(ext_rot$for_grid$srmr)), 0)
  expect_type(ext_rot$for_grid$rmsr, "double")
  expect_equal(sum(is.na(ext_rot$for_grid$rmsr)), 0)
  expect_type(ext_rot$for_grid$tli, "double")
  expect_equal(sum(is.na(ext_rot$for_grid$tli)), 1)
  expect_type(ext_rot$for_grid$ecvi, "double")
  expect_equal(sum(is.na(ext_rot$for_grid$ecvi)), 1)
  checkmate::expect_array(ext_rot$vars_accounted)
  expect_equal(dim(ext_rot$vars_accounted), c(3, 3, 3))
})


rm(efa_list, ext_a, efa_list_er, ext_er, efa_list_rot, ext_rot)

efa_cor <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500)
efa_raw <- EFA(GRiPS_raw, n_factors = 1)

# different types
efa_psych <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                 type = "psych", rotation = "promax")
efa_spss <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                type = "SPSS", rotation = "promax")

# different methods
efa_ml <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
              method = "ML")
efa_uls <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
               method = "ULS")

# different rotation methods from GPA rotation package (orthogonal and oblique)
efa_equa <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                rotation = "equamax")
efa_quart <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                 rotation = "quartimin")

# PAF with promax rotation without a specified type
efa_none <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                type = "none", method = "PAF", rotation = "promax",
                max_iter = 500, init_comm = "unity", criterion = 1e-4,
                criterion_type = "sum", abs_eigen = FALSE, k = 3,
                P_type = "unnorm", precision= 1e-5, order_type = "eigen",
                varimax_type = "svd")

# create correlation matrices from population models
cormat_zero <- population_models$loadings$baseline %*% population_models$phis_3$zero %*% t(population_models$loadings$baseline)
diag(cormat_zero) <- 1

cormat_moderate <- population_models$loadings$baseline %*% population_models$phis_3$moderate %*% t(population_models$loadings$baseline)
diag(cormat_moderate) <- 1

efa_paf_zero <- EFA(cormat_zero, 3, 500, rotation = "varimax")
efa_ml_zero <- EFA(cormat_zero, 3, 500, method = "ML", rotation = "varimax")
efa_uls_zero <- EFA(cormat_zero, 3, 500, method = "ULS", rotation = "varimax")

efa_paf_moderate <- EFA(cormat_moderate, 3, 500, rotation = "promax")
efa_ml_moderate <- EFA(cormat_moderate, 3, 500, method = "ML",
                       rotation = "promax")
efa_uls_moderate <- EFA(cormat_moderate, 3, 500, method = "ULS",
                        rotation = "promax")


test_that("output class and dimensions are correct", {
  # Objects carry the new `efa` class first and keep the legacy `EFA` string for
  # inherits()/back-compat; the loading matrices are dual-classed the same way.
  expect_identical(class(efa_cor), c("efa", "EFA"))
  expect_s3_class(efa_raw, "EFA")
  expect_s3_class(efa_psych, "EFA")
  expect_s3_class(efa_spss, "EFA")
  expect_s3_class(efa_ml, "EFA")
  expect_s3_class(efa_uls, "EFA")
  expect_s3_class(efa_equa, "EFA")
  expect_s3_class(efa_quart, "EFA")
  expect_s3_class(efa_none, "EFA")

  expect_identical(class(efa_cor$unrot_loadings), c("efa_loadings", "LOADINGS"))
  expect_s3_class(efa_raw$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_psych$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_spss$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_ml$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_uls$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_equa$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_quart$unrot_loadings, "LOADINGS")
  expect_s3_class(efa_none$unrot_loadings, "LOADINGS")

  expect_identical(class(efa_psych$rot_loadings), c("efa_loadings", "LOADINGS"))
  expect_s3_class(efa_spss$rot_loadings, "LOADINGS")
  expect_s3_class(efa_equa$rot_loadings, "LOADINGS")
  expect_s3_class(efa_quart$rot_loadings, "LOADINGS")
  expect_s3_class(efa_none$rot_loadings, "LOADINGS")

  expect_named(efa_cor, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                          "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                          "vars_accounted", "fit_indices", "model_implied_R",
                          "residuals", "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_raw, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                          "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                          "vars_accounted", "fit_indices", "model_implied_R",
                          "residuals", "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_psych, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                            "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                            "vars_accounted", "fit_indices", "model_implied_R",
                            "residuals", "rot_loadings",
                            "Phi", "Structure", "rotmat", "vars_accounted_rot",
                            "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_spss, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                           "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                           "vars_accounted", "fit_indices", "model_implied_R",
                           "residuals", "rot_loadings",
                           "Phi", "Structure", "rotmat", "vars_accounted_rot",
                           "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_ml, c("orig_R", "h2", "orig_eigen", "final_eigen", "iter",
                         "convergence", "heywood", "unrot_loadings", "vars_accounted",
                         "fit_indices", "model_implied_R",
                         "residuals", "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_uls, c("orig_R", "h2", "orig_eigen", "final_eigen", "iter",
                         "convergence", "heywood", "unrot_loadings", "vars_accounted",
                         "fit_indices", "model_implied_R",
                         "residuals", "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_equa, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                           "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                           "vars_accounted", "fit_indices", "model_implied_R",
                           "residuals", "rot_loadings",
                           "rotmat", "vars_accounted_rot",
                           "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_quart, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                            "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                            "vars_accounted", "fit_indices", "model_implied_R",
                            "residuals", "rot_loadings",
                            "Phi", "Structure", "rotmat", "vars_accounted_rot",
                            "settings", "vcov_unrot_loadings", "Gamma"))
  expect_named(efa_none, c("orig_R", "h2_init", "h2", "orig_eigen", "init_eigen",
                           "final_eigen", "iter", "convergence", "heywood", "unrot_loadings",
                           "vars_accounted", "fit_indices", "model_implied_R",
                           "residuals", "rot_loadings",
                           "Phi", "Structure", "rotmat", "vars_accounted_rot",
                           "settings", "vcov_unrot_loadings", "Gamma"))
})

test_that("settings are returned correctly", {
  expect_named(efa_cor$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen"))
  expect_named(efa_raw$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen"))
  expect_named(efa_psych$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                     "N", "use", "cor_method", "input_type",
                                     "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                     "init_comm", "criterion", "criterion_type",
                                     "abs_eigen", "normalize", "P_type", "precision",
                                     "order_type", "varimax_type", "k"))
  expect_named(efa_spss$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                    "N", "use", "cor_method", "input_type",
                                    "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                    "init_comm", "criterion", "criterion_type",
                                    "abs_eigen", "normalize", "P_type", "precision",
                                    "order_type", "varimax_type", "k"))
  expect_named(efa_ml$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                    "N", "use", "cor_method", "input_type",
                                    "cor_method_used", "se", "b_boot", "ci", "seed", "start_method"))
  expect_named(efa_uls$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed"))
  expect_named(efa_equa$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "normalize", "precision",
                                   "order_type", "randomStarts", "rotation_diagnostics"))
  expect_named(efa_quart$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "normalize", "precision",
                                   "order_type", "k", "randomStarts", "rotation_diagnostics"))
  expect_named(efa_none$settings, c("estimator", "method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "input_type",
                                   "cor_method_used", "se", "b_boot", "ci", "seed", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "normalize", "P_type", "precision",
                                   "order_type", "varimax_type", "k"))

  expect_equal(efa_cor$settings$estimator, "PAF")
  expect_equal(efa_raw$settings$estimator, "PAF")
  expect_equal(efa_psych$settings$estimator, "PAF")
  expect_equal(efa_spss$settings$estimator, "PAF")
  expect_equal(efa_ml$settings$estimator, "ML")
  expect_equal(efa_uls$settings$estimator, "ULS")
  expect_equal(efa_equa$settings$estimator, "PAF")
  expect_equal(efa_quart$settings$estimator, "PAF")
  expect_equal(efa_none$settings$estimator, "PAF")

  expect_equal(efa_cor$settings$rotation, "none")
  expect_equal(efa_raw$settings$rotation, "none")
  expect_equal(efa_psych$settings$rotation, "promax")
  expect_equal(efa_spss$settings$rotation, "promax")
  expect_equal(efa_ml$settings$rotation, "none")
  expect_equal(efa_uls$settings$rotation, "none")
  expect_equal(efa_equa$settings$rotation, "equamax")
  expect_equal(efa_quart$settings$rotation, "quartimin")
  expect_equal(efa_none$settings$rotation, "promax")

  expect_equal(efa_cor$settings$type, "EFAtools")
  expect_equal(efa_raw$settings$type, "EFAtools")
  expect_equal(efa_psych$settings$type, "psych")
  expect_equal(efa_spss$settings$type, "SPSS")
  expect_equal(efa_ml$settings$type, "EFAtools")
  expect_equal(efa_uls$settings$type, "EFAtools")
  expect_equal(efa_equa$settings$type, "EFAtools")
  expect_equal(efa_quart$settings$type, "EFAtools")
  expect_equal(efa_none$settings$type, "none")

  expect_equal(efa_cor$settings$n_factors, 3)
  expect_equal(efa_raw$settings$n_factors, 1)
  expect_equal(efa_psych$settings$n_factors, 3)
  expect_equal(efa_spss$settings$n_factors, 3)
  expect_equal(efa_ml$settings$n_factors, 3)
  expect_equal(efa_uls$settings$n_factors, 3)
  expect_equal(efa_equa$settings$n_factors, 3)
  expect_equal(efa_quart$settings$n_factors, 3)
  expect_equal(efa_none$settings$n_factors, 3)

  expect_equal(efa_cor$settings$N, 500)
  expect_equal(efa_raw$settings$N, 810)
  expect_equal(efa_psych$settings$N, 500)
  expect_equal(efa_spss$settings$N, 500)
  expect_equal(efa_ml$settings$N, 500)
  expect_equal(efa_uls$settings$N, 500)
  expect_equal(efa_equa$settings$N, 500)
  expect_equal(efa_quart$settings$N, 500)
  expect_equal(efa_none$settings$N, 500)

  expect_equal(efa_cor$settings$use, "pairwise.complete.obs")
  expect_equal(efa_raw$settings$use, "pairwise.complete.obs")
  expect_equal(efa_psych$settings$use, "pairwise.complete.obs")
  expect_equal(efa_spss$settings$use, "pairwise.complete.obs")
  expect_equal(efa_ml$settings$use, "pairwise.complete.obs")
  expect_equal(efa_uls$settings$use, "pairwise.complete.obs")
  expect_equal(efa_equa$settings$use, "pairwise.complete.obs")
  expect_equal(efa_quart$settings$use, "pairwise.complete.obs")
  expect_equal(efa_none$settings$use, "pairwise.complete.obs")

  expect_equal(efa_cor$settings$cor_method, "pearson")
  expect_equal(efa_raw$settings$cor_method, "pearson")
  expect_equal(efa_psych$settings$cor_method, "pearson")
  expect_equal(efa_spss$settings$cor_method, "pearson")
  expect_equal(efa_ml$settings$cor_method, "pearson")
  expect_equal(efa_uls$settings$cor_method, "pearson")
  expect_equal(efa_equa$settings$cor_method, "pearson")
  expect_equal(efa_quart$settings$cor_method, "pearson")
  expect_equal(efa_none$settings$cor_method, "pearson")

  expect_equal(efa_cor$settings$max_iter, 300)
  expect_equal(efa_raw$settings$max_iter, 300)
  expect_equal(efa_psych$settings$max_iter, 50)
  expect_equal(efa_spss$settings$max_iter, 25)
  expect_equal(efa_equa$settings$max_iter, 300)
  expect_equal(efa_quart$settings$max_iter, 300)
  expect_equal(efa_none$settings$max_iter, 500)

  expect_equal(efa_cor$settings$init_comm, "smc")
  expect_equal(efa_raw$settings$init_comm, "smc")
  expect_equal(efa_psych$settings$init_comm, "smc")
  expect_equal(efa_spss$settings$init_comm, "smc")
  expect_equal(efa_equa$settings$init_comm, "smc")
  expect_equal(efa_quart$settings$init_comm, "smc")
  expect_equal(efa_none$settings$init_comm, "unity")

  expect_equal(efa_cor$settings$criterion, 0.001)
  expect_equal(efa_raw$settings$criterion,  0.001)
  expect_equal(efa_psych$settings$criterion,  0.001)
  expect_equal(efa_spss$settings$criterion,  0.001)
  expect_equal(efa_equa$settings$criterion,  0.001)
  expect_equal(efa_quart$settings$criterion,  0.001)
  expect_equal(efa_none$settings$criterion,  1e-4)

  expect_equal(efa_cor$settings$criterion_type, "sum")
  expect_equal(efa_raw$settings$criterion_type, "sum")
  expect_equal(efa_psych$settings$criterion_type, "sum")
  expect_equal(efa_spss$settings$criterion_type, "max_individual")
  expect_equal(efa_equa$settings$criterion_type, "sum")
  expect_equal(efa_quart$settings$criterion_type, "sum")
  expect_equal(efa_none$settings$criterion_type, "sum")

  expect_equal(efa_cor$settings$abs_eigen, TRUE)
  expect_equal(efa_raw$settings$abs_eigen,  TRUE)
  expect_equal(efa_psych$settings$abs_eigen, FALSE)
  expect_equal(efa_spss$settings$abs_eigen,  TRUE)
  expect_equal(efa_equa$settings$abs_eigen, TRUE)
  expect_equal(efa_quart$settings$abs_eigen,  TRUE)
  expect_equal(efa_none$settings$abs_eigen, FALSE)

  expect_equal(efa_psych$settings$normalize, TRUE)
  expect_equal(efa_spss$settings$normalize, TRUE)
  expect_equal(efa_equa$settings$normalize, TRUE)
  expect_equal(efa_quart$settings$normalize, TRUE)
  expect_equal(efa_none$settings$normalize, TRUE)

  expect_equal(efa_psych$settings$P_type, "unnorm")
  expect_equal(efa_spss$settings$P_type, "norm")
  expect_equal(efa_none$settings$P_type, "unnorm")

  expect_equal(efa_psych$settings$precision, 1e-05)
  expect_equal(efa_spss$settings$precision, 1e-05)
  expect_equal(efa_equa$settings$precision, 1e-05)
  expect_equal(efa_quart$settings$precision, 1e-05)
  expect_equal(efa_none$settings$precision, 1e-05)

  expect_equal(efa_psych$settings$order_type, "eigen")
  expect_equal(efa_spss$settings$order_type, "ss_factors")
  expect_equal(efa_equa$settings$order_type, "eigen")
  expect_equal(efa_quart$settings$order_type, "eigen")
  expect_equal(efa_none$settings$order_type, "eigen")

  expect_equal(efa_psych$settings$varimax_type, "svd")
  expect_equal(efa_spss$settings$varimax_type, "kaiser")
  expect_equal(efa_none$settings$varimax_type, "svd")

  expect_equal(efa_psych$settings$k, 4)
  expect_equal(efa_spss$settings$k, 4)
  expect_equal(efa_none$settings$k, 3)

  expect_equal(efa_ml$settings$start_method, "psych")
})

test_that("settings record the seed and how the correlations were obtained", {
  cormat <- test_models$baseline$cormat

  # a correlation matrix consumes no correlation method, so the resolved record is NA
  # while the requested value is kept unchanged
  s_cor <- efa_fit(cormat, n_factors = 3, N = 500)$settings
  expect_identical(s_cor$input_type, "correlation")
  expect_identical(s_cor$cor_method_used, NA_character_)
  expect_identical(s_cor$cor_method, "pearson")
  expect_null(s_cor$seed)

  # a method requested but never run is still echoed as requested, and still not
  # reported as the one that ran
  s_tetra <- efa_fit(cormat, n_factors = 3, N = 500, cor_method = "tetra")$settings
  expect_identical(s_tetra$cor_method, "tetra")
  expect_identical(s_tetra$cor_method_used, NA_character_)

  # raw data records the method that actually produced the correlations, and the seed
  s_raw <- suppressMessages(
    efa_fit(GRiPS_raw, n_factors = 1, cor_method = "spearman", seed = 42))$settings
  expect_identical(s_raw$input_type, "raw")
  expect_identical(s_raw$cor_method_used, "spearman")
  expect_equal(s_raw$seed, 42)
})

test_that("factor analyses are performed correctly", {
  expect_equal(mean(abs(efa_compare(efa_paf_zero$rot_loadings,
                                   population_models$loadings$baseline,
                                   plot = FALSE)$diff)), 0, tolerance = .01)
  expect_equal(mean(abs(efa_compare(efa_ml_zero$rot_loadings,
                                   population_models$loadings$baseline,
                                   plot = FALSE)$diff)), 0, tolerance = .01)
  expect_equal(mean(abs(efa_compare(efa_uls_zero$rot_loadings,
                                   population_models$loadings$baseline,
                                   plot = FALSE)$diff)), 0, tolerance = .01)
  # expect_equal(mean(abs(efa_compare(efa_paf_moderate$rot_loadings,
  #                                  population_models$loadings$baseline,
  #                                  plot = FALSE)$diff)), 0, tolerance = .01)
  expect_equal(mean(abs(efa_compare(efa_ml_moderate$rot_loadings,
                                   population_models$loadings$baseline,
                                   plot = FALSE)$diff)), 0, tolerance = .01)
  expect_equal(mean(abs(efa_compare(efa_uls_moderate$rot_loadings,
                                   population_models$loadings$baseline,
                                   plot = FALSE)$diff)), 0, tolerance = .01)
})

test_that("errors are thrown correctly", {
  expect_error(EFA(1:5), class = "efa_input_not_matrix")
  expect_error(EFA(cor_nposdef, n_factors = 1, N = 10, type = "SPSS"), class = "efa_cor_not_posdef")
  expect_message(EFA(GRiPS_raw, n_factors = 1), class = "efa_cor_from_data")
  expect_warning(EFA(GRiPS_raw, N = 20, n_factors = 1), class = "efa_n_from_data")
  expect_error(EFA(sing_raw, n_factors = 1), class = "efa_cor_singular")
  expect_error(EFA(sing_cor, N = sing_N, n_factors = 1), class = "efa_cor_singular")
  # a criterion of 1 is not a convergence tolerance. EFA() repacks it into the estimation
  # control, which rejects it up front -- as it already did for any larger value -- so the
  # invalid call is caught at the argument rather than deep inside the fit.
  expect_error(EFA(test_models$baseline$cormat, n_factors = 3, N = 500, method = "PAF", criterion = 1),
               class = "efa_control_input")
  # Identification is a property of the dimensions, not of the values, but fitting a
  # 10 x 3 sample does depend on the draw: whether the extraction Heywoods or runs out of
  # iterations varies from one realization to the next. Seed the fixture and let those two
  # incidental conditions through, so the assertion stays about identification alone.
  set.seed(1)
  dat_ident <- matrix(rnorm(30), ncol = 3)
  expect_warning(EFA(dat_ident, n_factors = 2), class = "efa_underidentified")
  expect_warning(
    suppressWarnings(
      EFA(dat_ident, n_factors = 1),
      classes = c("efa_heywood", "efa_paf_nonconvergence")
    ),
    class = "efa_just_identified"
  )
  expect_warning(EFA(test_models$baseline$cormat, n_factors = 3, method = "ML"), class = "efa_fit_na_n")
  expect_warning(EFA(test_models$baseline$cormat, n_factors = 3, method = "ULS"), class = "efa_fit_na_n")
  expect_warning(
    suppressWarnings(
      EFA(cor_nposdef, n_factors = 1, N = 10),
      classes = c("efa_just_identified", "efa_heywood")
    ),
    class = "efa_cor_smoothed"
  )
  expect_warning(EFA(test_models$baseline$cormat, n_factors = 3, method = "ML", N = 500, type = "SPSS"), class = "efa_spss_method_untested")
  expect_warning(EFA(test_models$baseline$cormat, n_factors = 3, method = "ULS", N = 500, type = "SPSS"), class = "efa_spss_method_untested")
  expect_warning(EFA(test_models$baseline$cormat, n_factors = 3, rotation = "oblimin", N = 500, type = "SPSS"), class = "efa_spss_rotation_untested")
  expect_warning(EFA(test_models$baseline$cormat, n_factors = 3, rotation = "quartimax", N = 500, type = "SPSS"), class = "efa_spss_rotation_untested")
})

test_that("an unusable N is rejected or reported, and never left silent", {
  R <- test_models$baseline$cormat

  # N = 0 is an impossible sample size, not an unknown one: it goes through the classed
  # argument-validation path. NA remains the supported way of saying "sample size unknown",
  # and still warns.
  expect_error(efa_fit(R, n_factors = 3, N = 0, estimator = "ML"),
               class = "efa_invalid_argument")
  expect_warning(efa_fit(R, n_factors = 3, N = NA, estimator = "ML"),
                 class = "efa_fit_na_n")

  # A valid but very small N relative to p turns Bartlett's small-sample multiplier
  # non-positive, so the chi-square block is undefined. That was previously silent and
  # indistinguishable from a PAF fit; it now warns exactly once, and the residual summaries
  # and df it leaves behind are still populated.
  expect_warning(fit_small <- efa_fit(R, n_factors = 3, N = 8, estimator = "ML"),
                 class = "efa_fit_indices_undefined")
  expect_true(is.na(fit_small$fit_indices$chi))
  expect_true(is.na(fit_small$fit_indices$CFI))
  expect_true(is.na(fit_small$fit_indices$RMSEA))
  expect_true(is.finite(fit_small$fit_indices$SRMR))
  expect_equal(fit_small$fit_indices$df, .efa_df(ncol(R), 3))

  # PAF reports no chi-square at any N, so the small-sample warning would be noise there.
  expect_no_warning(efa_fit(R, n_factors = 3, N = 8, estimator = "PAF"),
                    class = "efa_fit_indices_undefined")

  # An underidentified model leaves the same block undefined, and its warning must not claim
  # that no fit indices at all were computed: the residual summaries are returned, and are
  # exactly the numbers a reader would otherwise misread as a flawless solution.
  expect_warning(
    fit_ui <- suppressWarnings(efa_fit(R, n_factors = 17, N = 500, estimator = "ML"),
                               classes = c("efa_heywood", "efa_nonconvergence")),
    class = "efa_underidentified"
  )
  expect_lt(fit_ui$fit_indices$df, 0)
  expect_true(is.na(fit_ui$fit_indices$chi))
  expect_true(is.finite(fit_ui$fit_indices$SRMR))
  expect_true(is.finite(fit_ui$fit_indices$CAF))
})

test_that("a singular correlation matrix is rejected for every type except psych + PAF", {
  # The psych preset relaxes the singularity check because psych::smc() falls back
  # to a pseudo-inverse for the PAF starting communalities. ML and ULS have no such
  # fallback -- they solve R for their starting values, and the ML discrepancy needs
  # log|R| -- so the check must stay on for them under every preset, and the abort
  # must be the package's classed one rather than a raw LAPACK error.
  for (ty in c("EFAtools", "psych", "SPSS")) {
    for (est in c("PAF", "ML", "ULS")) {
      if (ty == "psych" && est == "PAF") next
      expect_error(
        efa_fit(sing_cor, n_factors = 1, N = sing_N, estimator = est,
                estimate_control = estimate_control(type = ty)),
        class = "efa_cor_singular"
      )
    }
  }

  # psych + PAF keeps the documented pseudo-inverse behaviour and still fits.
  expect_no_error(
    suppressWarnings(
      efa_fit(sing_cor, n_factors = 1, N = sing_N, estimator = "PAF",
              estimate_control = estimate_control(type = "psych"))
    )
  )
})

test_that("EFA rejects n_factors >= number of variables", {
  dat <- matrix(rnorm(30), ncol = 3)  # 3 variables

  # n_factors == n_variables: out of bounds for ML, degenerate for ULS/PAF
  expect_error(EFA(dat, n_factors = 3, method = "ML"),  class = "efa_too_many_factors")
  expect_error(EFA(dat, n_factors = 3, method = "ULS"), class = "efa_too_many_factors")
  expect_error(EFA(dat, n_factors = 3, method = "PAF"), class = "efa_too_many_factors")

  # n_factors > n_variables
  expect_error(EFA(dat, n_factors = 4, method = "ML"),  class = "efa_too_many_factors")

  # the bootstrap path reuses n_factors and is guarded before any resampling runs
  expect_error(EFA(dat, n_factors = 3, method = "ML", se = "np-boot"),
               class = "efa_too_many_factors")
})

test_that("EFA rejects n_factors below 1 and a missing n_factors", {
  cm <- test_models$baseline$cormat

  # Several factor retention criteria legitimately return 0 ("no factor worth
  # extracting"), so the retain-then-fit idiom can hand efa_fit() a zero. It must
  # be rejected in R with a classed condition, not by the C++ bound check.
  for (est in c("PAF", "ML", "ULS")) {
    expect_error(efa_fit(cm, n_factors = 0, N = 500, estimator = est),
                 class = "efa_too_few_factors")
  }
  expect_error(EFA(cm, n_factors = 0, N = 500), class = "efa_too_few_factors")

  # Omitting n_factors points at the function whose job is to answer it.
  expect_error(efa_fit(cm, N = 500), class = "efa_missing_n_factors")
  expect_error(EFA(cm, N = 500), class = "efa_missing_n_factors")
})

test_that("a whole retention object as n_factors names the component that holds the counts", {
  # The companion mistake to omitting n_factors. The shared argument assertion answered with
  # the class the value is not ("Must be of type 'count', not 'efa_retain/N_FACTORS'"), never
  # with the component to take the count from. Both retention classes reach the guard.
  cm <- test_models$baseline$cormat
  bad_n_factors <- function(nf) {
    e <- tryCatch(efa_fit(cm, n_factors = nf, N = 500),
                  efa_n_factors_object = function(e) e)
    expect_s3_class(e, "efa_n_factors_object")
    conditionMessage(e)
  }

  ret <- suppressWarnings(suppressMessages(
    efa_retain(cm, N = 500, criteria = c("EKC", "MAP"), suitability = FALSE)))
  expect_snapshot(cat(bad_n_factors(ret)))
  # A single criterion reporting a single variant is the singular branch of the count.
  expect_snapshot(cat(bad_n_factors(suppressMessages(efa_ekc(cm, N = 500)))))

  # `$n_factors` carries one count per criterion (or per variant), so a single element of it
  # is the value the message asks for, and it is accepted.
  expect_s3_class(efa_fit(cm, n_factors = ret$n_factors[["MAP_TR4"]], N = 500), "efa")
})

test_that("print.efa output is stable (PAF, promax)", {
  local_reproducible_output()

  expect_snapshot(print(efa_psych), transform = scrub_num)
})

test_that("print.efa output is stable (ML, promax)", {
  local_reproducible_output()

  expect_snapshot(print(efa_ml_moderate), transform = scrub_num)
})

test_that("format.efa is the source of truth and honours the colour state", {
  # print() is exactly cat(format(x), sep = "\n"), so the two agree line for line.
  expect_identical(utils::capture.output(print(efa_psych)), format(efa_psych))

  old <- options(cli.num_colors = 256)
  on.exit(options(old), add = TRUE)

  # With colours on the report embeds ANSI ...
  expect_true(any(grepl("\033", format(efa_psych), fixed = TRUE)))

  # ... and with colours off it is plain.
  options(cli.num_colors = 1)
  expect_false(any(grepl("\033", format(efa_psych), fixed = TRUE)))
})

test_that("the model header follows the console width", {
  # The header names the analysis and is emitted verbatim, so a "setting = 'value'" token is
  # never split; it is packed to the console around those tokens instead.
  header <- function(w, colors = 1) {
    out <- withr::with_options(list(cli.width = w, cli.num_colors = colors),
                               cli::ansi_strip(format(efa_psych)))
    wrapped_item(out, "^EFA performed")
  }

  for (w in c(120L, 80L, 60L)) {
    lines <- header(w)
    expect_true(all(cli::ansi_nchar(lines, type = "width") <= w))
    # wrapping moves the line break only: the same sentence, with both tokens intact
    expect_identical(paste(trimws(lines), collapse = " "),
                     "EFA performed with estimator = 'PAF' and rotation = 'promax'.")
    expect_false(any(grepl("= '[^']*$", lines)))
  }

  # a narrow console does split the line, and does so identically with colours on
  expect_length(header(60L), 2L)
  expect_identical(header(60L), header(60L, colors = 256))
})

test_that("summary.efa output is stable (PAF, promax)", {
  local_reproducible_output()

  expect_snapshot(print(summary(efa_psych)), transform = scrub_num)
})

test_that("summary.efa output is stable (ML, promax)", {
  local_reproducible_output()

  expect_snapshot(print(summary(efa_ml_moderate)), transform = scrub_num)
})

test_that("format.summary.efa is the source of truth and honours the colour state", {
  s <- summary(efa_psych)

  # print() is exactly cat(format(x), sep = "\n"), so the two agree line for line.
  expect_identical(utils::capture.output(print(s)), format(s))

  old <- options(cli.num_colors = 256)
  on.exit(options(old), add = TRUE)

  # With colours on the report embeds ANSI ...
  expect_true(any(grepl("\033", format(s), fixed = TRUE)))

  # ... and with colours off it is plain.
  options(cli.num_colors = 1)
  expect_false(any(grepl("\033", format(s), fixed = TRUE)))
})
test_that("SRMR is the printed residual fit index while RMSR remains returned", {
  residuals <- matrix(c(0, .10, -.20,
                        .10, 0, .30,
                        -.20, .30, 0), nrow = 3, byrow = TRUE)
  p <- nrow(residuals)
  rmsr <- .rmsr(residuals)
  srmr <- sqrt(sum(residuals[upper.tri(residuals)]^2) / (p * (p + 1) / 2))

  expect_equal(srmr, rmsr * sqrt((p - 1) / (p + 1)))
  expect_lt(srmr, rmsr)

  expect_true(all(c("RMSR", "SRMR") %in% names(efa_ml$fit_indices)))
  out <- cli::ansi_strip(format(efa_ml))
  summary_out <- cli::ansi_strip(format(summary(efa_ml)))
  expect_false(any(grepl("^RMSR\\b", out)))
  expect_false(any(grepl("^RMSR\\b", summary_out)))
  expect_true(any(grepl("^SRMR:", out)))
  expect_true(any(grepl("^SRMR:", summary_out)))
})

test_that("print/summary.efa omit the inapplicable tables for a rotated single factor", {
  local_reproducible_output()

  efa_1fac <- suppressWarnings(
    EFA(test_models$baseline$cormat, n_factors = 1, N = 500, method = "PAF",
        rotation = "promax")
  )

  # A single factor cannot be rotated, so the rotation-only outputs are absent and
  # their print sections are skipped (rather than rendering a stray NA).
  expect_null(efa_1fac$Phi)
  expect_null(efa_1fac$vars_accounted_rot)

  # The variance table also drops the three cumulative/common-variance rows here: with one
  # factor they would only repeat the two above them. This shape is part of the documented
  # return, because user code indexing a row by name has to allow for it.
  expect_identical(rownames(efa_1fac$vars_accounted), c("SS loadings", "Prop Tot Var"))

  expect_snapshot(print(efa_1fac), transform = scrub_num)
  expect_snapshot(print(summary(efa_1fac)), transform = scrub_num)
})

test_that("EFA records rotation diagnostics for multistart rotations", {
  # A native gradient-projection rotation records the per-start diagnostics.
  efa_geo <- suppressWarnings(
    EFA(test_models$baseline$cormat, n_factors = 3, N = 500, rotation = "geominQ")
  )
  rd <- efa_geo$settings$rotation_diagnostics
  expect_type(rd, "list")
  expect_named(rd, c("converged", "n_starts_total", "n_optimized", "n_converged",
                     "n_distinct_minima", "criterion_spread", "criterion_best"))
  # the convergence status of the solution that is actually returned, so a saved fit can be
  # audited for it after the fact rather than only through the transient warning
  expect_true(rd$converged)
  # 100 random starts plus the rational (identity) start; only a few of them are ever
  # optimized under the screen-and-triage strategy.
  expect_equal(rd$n_starts_total, 101L)
  expect_lte(rd$n_optimized, rd$n_starts_total)
  # The distinct-optima count is taken over the converged starts only, so it is at least
  # one (the rational start converges here) and never exceeds the converged count, which
  # in turn can never exceed the number of optimized starts.
  expect_gte(rd$n_converged, 1L)
  expect_lte(rd$n_converged, rd$n_optimized)
  expect_gte(rd$n_distinct_minima, 1L)
  expect_lte(rd$n_distinct_minima, rd$n_converged)
  expect_true(is.finite(rd$criterion_best))
  expect_gte(rd$criterion_spread, 0)

  # simplimax also records diagnostics. It runs full multistart, so every requested random
  # start is optimized and the two counts coincide.
  efa_simp <- suppressWarnings(
    EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
        rotation = "simplimax", randomStarts = 20)
  )
  rd_simp <- efa_simp$settings$rotation_diagnostics
  expect_equal(rd_simp$n_starts_total, 21L)
  expect_equal(rd_simp$n_optimized, 21L)
  expect_true(is.finite(rd_simp$criterion_best))

  # Rotations that do not use random starts carry no diagnostics.
  expect_null(efa_psych$settings$rotation_diagnostics)   # promax
  expect_null(efa_cor$settings$rotation_diagnostics)     # rotation = "none"

  # summary() surfaces the diagnostic only for the multistart rotations.
  local_reproducible_output()
  geo_summary <- utils::capture.output(print(summary(efa_geo)))
  expect_true(any(grepl("Rotation local optima", geo_summary, fixed = TRUE)))
  # The line reports the starts that were actually optimized alongside the total, so it
  # cannot overstate the evidence by counting screened-and-discarded starts as optima.
  expect_true(any(grepl(
    paste0(rd$n_distinct_minima, " distinct from ", rd$n_optimized,
           " of ", rd$n_starts_total, " starts"),
    geo_summary, fixed = TRUE)))
  promax_summary <- utils::capture.output(print(summary(efa_psych)))
  expect_false(any(grepl("Rotation local optima", promax_summary, fixed = TRUE)))
})

test_that("the rotation diagnostics record the returned solution's own convergence", {
  # The returned solution is the start with the lowest criterion value, which need not be a
  # converged one: a start that stopped short of the tolerance at a lower value beats a
  # converged start at a higher one. `converged` therefore has to be carried through from
  # the winning start rather than read off the aggregate count, or such a solution would be
  # presented as converged because its rivals were. Exercised on the helper so the
  # disagreement is deterministic instead of depending on which start a seed happens to win.
  rd <- .rotation_diagnostics(all_values = c(0.5, 0.1), all_converged = c(1L, 0L),
                              randomStarts = 1, converged = FALSE)
  expect_false(rd$converged)
  expect_equal(rd$n_converged, 1L)
  expect_equal(rd$criterion_best, 0.1)

  # end to end: with a one-iteration budget no start can converge, and the object records
  # that instead of leaving it in the transient warning only
  capped <- suppressWarnings(
    EFA(test_models$baseline$cormat, n_factors = 3, N = 500, rotation = "geominQ",
        maxit = 1)
  )
  expect_false(capped$settings$rotation_diagnostics$converged)
})

test_that("residuals.efa is a pure extractor", {
  # returns the residual matrix with no printing side effect
  expect_output(residuals(efa_psych), NA)
  expect_identical(residuals(efa_psych), efa_psych$residuals)
  expect_identical(residuals(efa_psych, "raw"), efa_psych$residuals)

  # standardized residuals need bootstrap SEs, which efa_psych does not have
  expect_error(residuals(efa_psych, "standardized"),
               class = "efa_no_standardized_residuals")

  # an unknown type is rejected
  expect_error(residuals(efa_psych, "bogus"))
})

# A near-unit first loading pins item1's uniqueness at the ML/ULS estimation
# boundary, giving a genuine improper solution with every communality still below 1.
boundary_heywood_cormat <- function() {
  L <- c(.999, .7, .6, .5)
  R <- L %*% t(L)
  diag(R) <- 1
  dimnames(R) <- list(paste0("item", seq_along(L)), paste0("item", seq_along(L)))
  R
}

test_that("Heywood cases are detected, warned, and recorded", {
  # Over-extracting the baseline model with PAF yields a Heywood case (V14).
  expect_warning(
    suppressWarnings(
      EFA(test_models$baseline$cormat, 6, N = 500, method = "PAF"),
      classes = "efa_paf_nonconvergence"
    ),
    class = "efa_heywood"
  )

  efa_hey <- suppressWarnings(EFA(test_models$baseline$cormat, 6, N = 500,
                                  method = "PAF"))
  expect_gt(length(efa_hey$heywood), 0)
  expect_true(all(efa_hey$h2[efa_hey$heywood] >= 1))

  # A proper solution records no Heywood cases and emits no Heywood warning.
  efa_clean <- suppressWarnings(EFA(test_models$baseline$cormat, 3, N = 500,
                                    method = "ML"))
  expect_length(efa_clean$heywood, 0)
  # ... and it is still a named integer vector when empty, as documented
  expect_named(efa_clean$heywood)

  # ML/ULS constrain the uniquenesses to a lower bound, so an improper solution
  # keeps the communality just below 1 and instead pins a uniqueness at that
  # bound. A near-unit first loading forces this boundary case; it must be flagged
  # for both estimators even though no communality reaches 1.
  R_bnd <- boundary_heywood_cormat()

  for (m in c("ML", "ULS")) {
    expect_warning(EFA(R_bnd, 1, N = 200, method = m), class = "efa_heywood")
    efa_bnd <- suppressWarnings(EFA(R_bnd, 1, N = 200, method = m))
    expect_gt(length(efa_bnd$heywood), 0)
    expect_true(all(efa_bnd$h2[efa_bnd$heywood] < 1))
    # heywood is documented as a *named* integer vector: the optimiser-based
    # fitters return the uniquenesses as a one-column matrix, whose dim attribute
    # would otherwise displace the variable names when the two criteria are
    # combined, leaving the warning below with nothing to name.
    expect_named(efa_bnd$heywood, "item1")
  }

  # The same matrix with a clearly proper structure is not flagged.
  L_ok <- c(.9, .7, .6, .5)
  R_ok <- L_ok %*% t(L_ok)
  diag(R_ok) <- 1
  expect_length(suppressWarnings(EFA(R_ok, 1, N = 200, method = "ML"))$heywood, 0)
})

test_that("the Heywood warning names the affected variables", {
  local_reproducible_output()

  R_bnd <- boundary_heywood_cormat()

  # ML and ULS reach the Heywood case through a uniqueness pinned at the
  # estimation boundary; the message must name the variable rather than report a
  # bare count.
  expect_snapshot({
    for (est in c("ML", "ULS")) {
      cat("--", est, "--\n")
      withCallingHandlers(
        efa_fit(R_bnd, 1, N = 200, estimator = est),
        efa_heywood = function(w) cat(conditionMessage(w), "\n"),
        warning = function(w) invokeRestart("muffleWarning")
      )
    }
  })
})

test_that("point-estimate non-convergence is surfaced for the iterative estimators", {
  R <- test_models$baseline$cormat

  # Force the convergence code via a mock so the assertions do not depend on the
  # well-behaved baseline fits actually converging. EFA() must warn on a non-zero
  # code and stay silent on a zero code, for ML, ULS, and DWLS. heywood is cleared
  # so the only candidate warning is the convergence one. Each mock is scoped to its
  # own local() frame.
  for (m in c("ML", "ULS")) {
    local({
      fit <- suppressWarnings(.estimate_model(R, method = m, n_factors = 3, N = 500,
                                              start_method = "psych"))
      fit$heywood <- integer(0)
      testthat::local_mocked_bindings(.estimate_model = function(...) fit)

      fit$convergence <- 1L
      expect_warning(EFA(R, 3, N = 500, method = m), class = "efa_nonconvergence")

      # A zero code must not warn (the mock makes this deterministic).
      fit$convergence <- 0L
      conv_classes <- character(0)
      withCallingHandlers(
        EFA(R, 3, N = 500, method = m),
        warning = function(w) {
          conv_classes <<- c(conv_classes, class(w))
          invokeRestart("muffleWarning")
        }
      )
      expect_false("efa_nonconvergence" %in% conv_classes)
    })
  }

  # DWLS needs raw ordinal data with cor_method = "poly". The mocked fitter makes
  # the real estimation irrelevant, so a minimal fit carrying only the fields EFA()
  # reads after the fit suffices (the polychoric matrix is still built inside EFA()).
  local({
    xo <- DOSPERT_raw[, 1:6]
    testthat::local_mocked_bindings(
      .estimate_model = function(...) list(convergence = 1L, heywood = integer(0))
    )
    expect_warning(EFA(xo, 2, method = "DWLS", cor_method = "poly"),
                   class = "efa_nonconvergence")
  })

  # PAF surfaces non-convergence through its own warning class, never the
  # optimiser-estimator one.
  paf_classes <- character(0)
  withCallingHandlers(
    EFA(R, 3, N = 500, method = "PAF", type = "none", init_comm = "smc",
        criterion = 1e-3, criterion_type = "sum", abs_eigen = TRUE, max_iter = 1),
    warning = function(w) {
      paf_classes <<- c(paf_classes, class(w))
      invokeRestart("muffleWarning")
    }
  )
  expect_true("efa_paf_nonconvergence" %in% paf_classes)
  expect_false("efa_nonconvergence" %in% paf_classes)
})

test_that("the non-convergence explanation is branched on the convergence code", {
  # Code 1 is the optimiser's iteration limit; 51 and 52 come from L-BFGS-B's line
  # search and mean something different, so they must not be explained as an iteration
  # limit. Anything else falls back to the generic explanation.
  cap <- .nonconvergence_cause(1L)
  line_search <- .nonconvergence_cause(52L)
  generic <- .nonconvergence_cause(10L)

  for (msg in list(cap, line_search, generic)) {
    expect_type(msg, "character")
    expect_length(msg, 1L)
    expect_true(nzchar(msg))
  }

  # the two L-BFGS-B line-search codes share one explanation, distinct from the others
  expect_identical(.nonconvergence_cause(51L), line_search)
  expect_false(identical(cap, line_search))
  expect_false(identical(cap, generic))
  expect_false(identical(line_search, generic))

  # the branch is what the warning carries: a mocked code-52 fit must not be explained
  # with the iteration-limit text. The message is compared with its line breaks
  # collapsed, so console wrapping cannot decide the assertion.
  testthat::local_reproducible_output()
  squash <- function(x) gsub("\\s+", " ", paste(x, collapse = " "))

  R <- test_models$baseline$cormat
  fit <- suppressWarnings(.estimate_model(R, method = "ULS", n_factors = 3, N = 500))
  fit$heywood <- integer(0)
  fit$convergence <- 52L
  testthat::local_mocked_bindings(.estimate_model = function(...) fit)

  w <- expect_warning(EFA(R, 3, N = 500, method = "ULS"),
                      class = "efa_nonconvergence")
  msg <- squash(conditionMessage(w))
  expect_true(grepl(squash(line_search), msg, fixed = TRUE))
  expect_false(grepl(squash(cap), msg, fixed = TRUE))
})

test_that("the print banner reports optimiser non-convergence", {
  testthat::local_reproducible_output()
  R <- test_models$baseline$cormat
  efa <- suppressWarnings(EFA(R, 3, N = 500, method = "ML"))

  # Set the code explicitly so the assertions do not depend on the fixture's actual
  # convergence. A zero code shows no banner; a non-zero code shows the generic
  # optimiser banner for ML/ULS/DWLS.
  efa$convergence <- 0L
  expect_false(any(grepl("The optimiser did not converge", format(efa), fixed = TRUE)))
  efa$convergence <- 1L
  expect_true(any(grepl("The optimiser did not converge", format(efa), fixed = TRUE)))

  # PAF keeps its iteration-specific banner.
  paf <- suppressWarnings(
    EFA(R, 3, N = 500, method = "PAF", type = "none", init_comm = "smc",
        criterion = 1e-3, criterion_type = "sum", abs_eigen = TRUE, max_iter = 1)
  )
  expect_true(any(grepl("Maximum number of iterations reached without convergence",
                        format(paf), fixed = TRUE)))

  # An object carrying no convergence code (e.g. one saved by an older version) still
  # shows the banner via the iteration-count fallback (iter >= max_iter).
  paf$convergence <- NULL
  expect_true(any(grepl("Maximum number of iterations reached without convergence",
                        format(paf), fixed = TRUE)))
})

rm(efa_cor, efa_raw, efa_psych, efa_spss, efa_ml, efa_uls, efa_equa, efa_quart,
   efa_none, cormat_zero, cormat_moderate, efa_paf_zero, efa_ml_zero, efa_uls_zero,
   efa_paf_moderate, efa_ml_moderate, efa_uls_moderate)

test_that("print.efa argument validators raise classed conditions", {
  efa <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500)

  expect_error(print(efa, cutoff = -1), class = "efa_print_invalid_cutoff")
  expect_error(print(efa, digits = 1.5), class = "efa_print_invalid_digits")
  expect_error(print(efa, max_name_length = 0),
               class = "efa_print_invalid_max_name_length")
  expect_error(print(efa, diagnostics_top_n = -1),
               class = "efa_print_invalid_diagnostics_top_n")
  expect_error(print(efa, residual_cutoff = -1),
               class = "efa_print_invalid_residual_cutoff")
  expect_error(print(efa, residual_top_n = 0),
               class = "efa_print_invalid_residual_top_n")
  expect_error(print(efa, show_structure = NA),
               class = "efa_print_invalid_show_structure")
  expect_error(print(efa, show_loading_legend = NA),
               class = "efa_print_invalid_show_loading_legend")
  expect_error(print(efa, cross_loading_cutoff = -1),
               class = "efa_print_invalid_cross_loading_cutoff")
  expect_error(print(efa, min_primary_gap = -1),
               class = "efa_print_invalid_min_primary_gap")
  expect_error(print(efa, min_salient_per_factor = 0),
               class = "efa_print_invalid_min_salient_per_factor")
  expect_error(print(efa, max_factors_per_block = 0),
               class = "efa_print_invalid_max_factors_per_block")
  expect_error(print(efa, show_mi_diagnostics = NA),
               class = "efa_print_invalid_show_mi_diagnostics")
})

# Regression values are the RMSEA power-analysis figures tabulated by MacCallum,
# Browne & Sugawara (1996) for a 100-df model, reproduced with base R's noncentral
# chi-square functions (no external dependency).

test_that("RMSEA close-fit power reproduces MacCallum, Browne & Sugawara (1996)", {
  cf <- efa_power(df = 100, N = 200, type = "close", eps0 = 0.05, eps1 = 0.08)
  expect_equal(cf$power, 0.9549, tolerance = 1e-4)
  expect_equal(cf$solve_for, "power")
  expect_equal(cf$N, 200L)
  # ncp = (N - 1) * df * eps^2: 199 * 100 * .0025 and 199 * 100 * .0064
  expect_equal(unname(cf$ncp), c(49.75, 127.36))
  expect_named(cf$ncp, c("H0", "H1"))
})

test_that("RMSEA not-close-fit power reproduces the tabled value", {
  nc <- efa_power(df = 100, N = 200, type = "notclose", eps0 = 0.05, eps1 = 0.01)
  expect_equal(nc$power, 0.8701, tolerance = 1e-4)
  expect_equal(nc$solve_for, "power")
})

test_that("required sample size is the smallest N reaching the target power", {
  rn <- efa_power(df = 100, power = 0.80, type = "close")
  expect_equal(rn$N, 132L)
  expect_equal(rn$solve_for, "N")

  # 132 is minimal: one fewer falls short of .80, 132 clears it.
  expect_lt(efa_power(df = 100, N = 131)$power, 0.80)
  expect_gte(efa_power(df = 100, N = 132)$power, 0.80)

  # With neither N nor power supplied the target defaults to .80.
  expect_equal(efa_power(df = 100)$N, 132L)
  expect_equal(efa_power(df = 100)$settings$power, 0.80)
})

test_that("df is taken from p and k via the EFA df formula", {
  # p = 20, k = 5 -> ((20 - 5)^2 - (20 + 5)) / 2 = 100
  from_pk <- efa_power(p = 20, k = 5, N = 200)
  from_df <- efa_power(df = 100, N = 200)
  expect_equal(from_pk$settings$df, 100)
  expect_equal(from_pk$power, from_df$power)
})

test_that("the number of groups scales the noncentrality", {
  g1 <- efa_power(df = 100, N = 200, group = 1)
  g2 <- efa_power(df = 100, N = 200, group = 2)
  expect_equal(unname(g2$ncp), unname(g1$ncp) / 2)
})

test_that("the solved N is the total across groups, split by N_per_group", {
  # The 1 / group factor in the noncentrality makes the solved N a total, so it
  # grows linearly in `group` while the per-group requirement stays near-constant.
  g1 <- efa_power(df = 102, power = 0.80, group = 1)
  g2 <- efa_power(df = 102, power = 0.80, group = 2)
  g3 <- efa_power(df = 102, power = 0.80, group = 3)
  expect_equal(g1$N, 130L)
  expect_equal(g2$N, 260L)
  expect_equal(g3$N, 390L)
  expect_equal(g1$N_per_group, 130)
  expect_equal(g2$N_per_group, 130)
  expect_equal(g3$N_per_group, 130)
})

test_that("a solved N is a total that a study can collect", {
  # All groups have the same size, so a required total must divide by `group`; a total
  # of 259 over two groups would ask for 129.5 persons in each. Rounding the total up
  # keeps the reported power the power at the total that is reported. `N_per_group` is
  # `N / group`, so the modulus is the whole property -- no separate whole-number check.
  for (g in 1:4) {
    x <- efa_power(df = 102, power = 0.80, group = g)
    expect_equal(x$N %% g, 0)
    expect_gte(x$power, 0.80)
    # Still the smallest such total: one person less in each group misses the target.
    expect_lt(efa_power(df = 102, N = x$N - g, group = g)$power, 0.80)
  }

  # A target below `alpha` is reached at any size, which returns one person per group
  # rather than a total of one shared out over `group` groups.
  floor_g2 <- efa_power(df = 102, power = 0.01, group = 2)
  expect_equal(floor_g2$N, 2L)
  expect_equal(floor_g2$N_per_group, 1)
  expect_equal(efa_power(df = 102, power = 0.01, group = 1)$N, 1L)

  # A supplied N is the caller's own total and is reported unchanged, fraction and all.
  odd <- efa_power(df = 102, N = 259L, group = 2)
  expect_equal(odd$N, 259L)
  expect_equal(odd$N_per_group, 129.5)
})

test_that("the solved total matches semTools::findRMSEAsamplesize per group", {
  skip_if_not_installed("semTools")

  # findRMSEAsamplesize() searches the same noncentrality but reports the per-group
  # figure (its last line divides the solved total by `group`), so its value is the
  # N_per_group counterpart of the total efa_power() returns.
  #
  # It divides without rounding, so for a total that does not divide by `group` it
  # returns a fraction of a person (129.5 at df = 102 with two groups). efa_power()
  # rounds the total up to the next multiple of `group` instead, because all `group`
  # groups have the same size. The two therefore agree up to that one step, and the
  # oracle is compared as ceiling(): a definitional offset, not a numerical one.
  # Compared as named vectors so a failure names the group count that broke.
  groups <- c(1, 2, 3)
  ref <- vapply(groups, function(g) {
    semTools::findRMSEAsamplesize(rmsea0 = 0.05, rmseaA = 0.08, df = 102,
                                  power = 0.80, group = g)
  }, numeric(1))
  pw <- vapply(groups, function(g) {
    efa_power(df = 102, power = 0.80, group = g)$N_per_group
  }, numeric(1))
  tot <- vapply(groups, function(g) {
    as.numeric(efa_power(df = 102, power = 0.80, group = g)$N)
  }, numeric(1))
  names(ref) <- names(pw) <- names(tot) <- paste0("group", groups)

  expect_equal(pw, ceiling(ref), tolerance = 1e-8)
  expect_equal(tot, ceiling(ref) * groups, tolerance = 1e-8, ignore_attr = "names")
})

test_that("output has the expected class and structure", {
  x <- efa_power(df = 100, N = 200)
  expect_s3_class(x, "efa_power")
  expect_named(x, c("power", "N", "N_per_group", "crit", "ncp", "solve_for",
                    "settings"))
  expect_named(x$settings, c("mode", "type", "eps0", "eps1", "df", "p", "k",
                             "alpha", "group", "power"))
})

test_that("classed conditions fire as expected", {
  # Equal null and alternative RMSEA: nothing to detect.
  expect_error(efa_power(df = 100, N = 200, eps0 = 0.05, eps1 = 0.05),
               class = "efa_power_unreachable")
  # Non-positive degrees of freedom, given directly or from an unidentified model.
  expect_error(efa_power(df = 0, N = 200), class = "efa_power_bad_df")
  expect_error(efa_power(p = 3, k = 3, N = 200), class = "efa_power_bad_df")
  # No way to obtain df.
  expect_error(efa_power(N = 200), class = "efa_power_missing_df")
  # Both N and power supplied: nothing to solve for.
  expect_error(efa_power(df = 100, N = 200, power = 0.80),
               class = "efa_power_overdetermined")
  # Out-of-range alpha and target power.
  expect_error(efa_power(df = 100, N = 200, alpha = 0),
               class = "efa_power_bad_alpha")
  expect_error(efa_power(df = 100, power = 1.2), class = "efa_power_bad_power")

  # eps0/eps1 on the wrong side for the chosen test: a message, and the value is
  # still returned (the branch follows `type`, not the ordering of the epsilons).
  expect_message(
    val <- efa_power(df = 100, N = 200, type = "close", eps0 = 0.08, eps1 = 0.05),
    class = "efa_power_wrong_side")
  expect_s3_class(val, "efa_power")
})

test_that("an unreachable target aborts cleanly without warning noise", {
  # Wrong-side epsilons make close-fit power fall short at every N, so the search
  # runs up to the cap and aborts (the wrong-side message is muffled here).
  expect_error(
    suppressMessages(
      efa_power(df = 100, power = 0.80, type = "close", eps0 = 0.08, eps1 = 0.05)),
    class = "efa_power_unreached")

  # The search evaluates the noncentral chi-square routines at extreme
  # noncentrality; those base-R precision/convergence notes must not leak.
  expect_no_warning(suppressMessages(tryCatch(
    efa_power(df = 100, power = 0.80, type = "close", eps0 = 0.08, eps1 = 0.05),
    error = function(e) NULL)))
})

test_that("printed output is stable", {
  local_reproducible_output()

  # Solve for power (test of close fit).
  expect_snapshot(print(efa_power(df = 100, N = 200)))
  # Solve for the required sample size.
  expect_snapshot(print(efa_power(df = 100, power = 0.80)))
  # Test of not-close fit with more than one group.
  expect_snapshot(print(efa_power(df = 100, N = 200, type = "notclose", group = 2)))
})

test_that("plot() returns a ggplot for the various curves", {
  # Base curve (solved for power), and one solved for the required N.
  expect_s3_class(plot(efa_power(df = 100, N = 200)), "ggplot")
  expect_s3_class(plot(efa_power(df = 100, power = 0.80)), "ggplot")
  # Sweep df, sweep eps1, and a caller-supplied sample-size sequence.
  expect_s3_class(plot(efa_power(df = 100, N = 200), df = c(50, 100, 200)), "ggplot")
  expect_s3_class(plot(efa_power(df = 100, N = 200), eps1 = c(0.06, 0.08, 0.10)),
                  "ggplot")
  expect_s3_class(plot(efa_power(df = 100, N = 200), n = seq(50, 400, by = 50)),
                  "ggplot")
  # Not-close fit with more than one group still plots.
  expect_s3_class(plot(efa_power(df = 100, N = 200, type = "notclose", group = 2)),
                  "ggplot")

  # Wrong-side epsilons: the automatic N axis cannot solve for 0.99 power, so the
  # curve span falls back to a multiple of the object's N rather than erroring out.
  wrong_side <- suppressMessages(
    efa_power(df = 100, N = 200, type = "close", eps0 = 0.08, eps1 = 0.05))
  expect_s3_class(suppressMessages(plot(wrong_side)), "ggplot")
})

test_that("the object's point is marked only for its own curve", {
  pw <- efa_power(df = 100, N = 200)
  # Own curve: reference lines and the point are drawn (hline, vline, line, point).
  expect_length(plot(pw)$layers, 4L)
  # A scalar df/eps1 override redraws the curve, so the object's point no longer lies
  # on it and the marks are dropped -- only the line remains.
  expect_length(plot(pw, df = 50)$layers, 1L)
  expect_length(plot(pw, eps1 = 0.10)$layers, 1L)
  # A supplied N range that excludes the object's N would place the point off-axis, so
  # the marks are dropped there too.
  expect_length(plot(pw, n = seq(10, 50, by = 5))$layers, 1L)
})

test_that("the subtitle comparator follows the test direction", {
  # Close fit bounds the null from above, not-close fit from below.
  expect_match(plot(efa_power(df = 100, N = 200))$labels$subtitle, "H0 RMSEA \u2264")
  expect_match(plot(efa_power(df = 100, N = 200, type = "notclose"))$labels$subtitle,
               "H0 RMSEA \u2265")
})

test_that("overrides uphold the constructor's invariants", {
  pw <- efa_power(df = 100, N = 200)
  # A non-positive df has no identified model, exactly as efa_power() rejects it.
  expect_error(plot(pw, df = 0), class = "efa_power_bad_df")
  expect_error(plot(pw, df = c(0, 100)), class = "efa_power_bad_df")
  # An alternative RMSEA equal to the null leaves nothing to detect.
  expect_error(plot(pw, eps1 = 0.05), class = "efa_power_unreachable")
})

test_that("sweeping both df and eps1 is refused", {
  expect_error(
    plot(efa_power(df = 100, N = 200), df = c(50, 100), eps1 = c(0.06, 0.08)),
    class = "efa_power_plot_grid")
})

test_that("the power curve is visually stable", {
  skip_if_not_installed("vdiffr")

  vdiffr::expect_doppelganger("efa_power RMSEA power curve",
                              plot(efa_power(df = 100, N = 200)))
  vdiffr::expect_doppelganger("efa_power power curve swept df",
                              plot(efa_power(df = 100, N = 200), df = c(50, 100, 200)))
})


# --- Simulation mode -----------------------------------------------------------
# A clean, strong three-factor population fixture reused across the simulation-mode
# tests: the retention criteria and the loadings recovery should both succeed on it.
sim_Lambda <- population_models$loadings$baseline
sim_Phi <- population_models$phis_3$moderate

test_that("simulation mode is reproducible and leaves the RNG stream unchanged", {
  a <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                 n_datasets = 8, criteria = "EKC", seed = 42)
  b <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                 n_datasets = 8, criteria = "EKC", seed = 42)
  expect_equal(a, b)

  # A fixed seed restores the caller's random-number stream (no global side effect).
  set.seed(99)
  before <- get(".Random.seed", envir = globalenv(), inherits = FALSE)
  efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
            n_datasets = 4, criteria = "EKC", seed = 7)
  after <- get(".Random.seed", envir = globalenv(), inherits = FALSE)
  expect_identical(before, after)
})

test_that("simulation mode is reproducible at 1 vs 2 workers", {
  skip_on_cran()
  skip_if_not_slow()
  # Each replicate is analysed in its own future, and future.seed = TRUE binds its RNG
  # stream to its index, so the result must be identical regardless of the number of
  # workers. The multisession workers are fresh R processes that load the installed
  # package, so run this under devtools::check() / after devtools::install().
  # (multicore is unavailable on Windows.)
  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  run <- function() {
    efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
              n_datasets = 6, criteria = "EKC", seed = 2024)
  }

  future::plan(future::sequential)
  one <- run()
  future::plan(future::multisession, workers = 2)
  two <- run()

  expect_equal(one$hit_rate, two$hit_rate)
  expect_equal(one$recovery, two$recovery)
  expect_equal(one$convergence, two$convergence)
  expect_equal(one$replicates, two$replicates)
})

test_that("simulation mode recovers a clean population", {
  sim <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 300,
                   n_datasets = 20, criteria = c("EKC", "MAP"), seed = 42)
  expect_s3_class(sim, "efa_power")
  expect_identical(sim$settings$mode, "simulation")
  expect_equal(sim$k_true, 3L)
  expect_named(sim, c("hit_rate", "hits", "recovery", "convergence", "replicates",
                      "k_true", "model_error", "settings"))

  # EKC and the loadings recovery should both be near-perfect on a clean 3-factor model.
  expect_gte(sim$hit_rate[["EKC_BvA2017"]], 0.9)
  expect_gte(sim$recovery$min_rate, 0.9)
  expect_equal(sim$recovery$threshold, 0.95)
  expect_equal(sim$convergence$convergence_rate, 1)
  expect_equal(sim$convergence$heywood_rate, 0)

  # Every fit completed, so no replicate carries a failure reason.
  expect_true(all(sim$replicates$fit_ok))
  expect_true(all(is.na(sim$replicates$fit_error)))

  # hits data frame is consistent with the hit_rate vector.
  expect_equal(sim$hits$hit_rate, unname(sim$hit_rate))
  expect_true(all(sim$hits$hits <= sim$hits$n_valid))
})

test_that("simulation mode muffles the per-replicate fit messages", {
  expect_no_message(
    efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 120,
              n_datasets = 3, criteria = "EKC", seed = 1))
})

test_that("a criterion that fails on every replicate is reported, not dropped", {
  # N < p makes each replicate's correlation matrix singular, so MAP errors on every
  # replicate. The criterion must still appear in the results -- with no valid replicates
  # and an NA hit-rate -- and be named in a classed warning, rather than vanishing while
  # the settings still record the request.
  # The singular matrix also stops the recovery fit, which warns in its own right.
  expect_warning(
    expect_warning(
      sim <- efa_power("simulation", Lambda = sim_Lambda, N = 6, n_datasets = 3,
                       criteria = "MAP", seed = 1),
      class = "efa_power_criterion_failed"),
    class = "efa_power_fit_failed")
  expect_s3_class(sim, "efa_power")
  expect_named(sim$hit_rate, "MAP")
  expect_true(is.na(sim$hit_rate[["MAP"]]))
  expect_equal(sim$hits$n_valid, 0L, ignore_attr = TRUE)
  expect_equal(sim$hits$hits, 0, ignore_attr = TRUE)
  expect_identical(sim$settings$criteria, "MAP")
  # The formatter reports the missing rate rather than dropping the line or erroring.
  local_reproducible_output()
  expect_snapshot(print(sim))
})

test_that("a criterion that decides on every replicate raises no failure warning", {
  expect_no_warning(
    efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
              n_datasets = 4, criteria = "EKC", seed = 3))
})

test_that("a recovery fit that fails everywhere is reported, not thrown away", {
  # The recovery fit fails on every replicate while the retention criteria are
  # unaffected: the completed retention results must still be returned, with a warning,
  # rather than the whole (expensive) run being discarded.
  testthat::local_mocked_bindings(
    efa_fit = function(x, ...) {
      cli::cli_abort("Mocked recovery failure.", class = "efa_mock_fit_failure")
    })

  expect_warning(
    sim <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                     n_datasets = 3, criteria = "EKC", seed = 1),
    class = "efa_power_fit_failed")
  expect_s3_class(sim, "efa_power")
  expect_equal(sim$hits$n_valid, 3L, ignore_attr = TRUE)
  expect_true(is.finite(sim$hit_rate[["EKC_BvA2017"]]))

  # The fit half is missing, and every replicate keeps the reason it failed.
  expect_equal(sim$convergence$fit_rate, 0)
  expect_true(is.na(sim$convergence$convergence_rate))
  expect_true(is.na(sim$recovery$min_rate))
  expect_equal(sim$recovery$n_valid, 0L)
  expect_false(any(sim$replicates$fit_ok))
  expect_length(sim$replicates$fit_error, 3L)
  expect_true(all(grepl("Mocked recovery failure", sim$replicates$fit_error,
                        fixed = TRUE)))
})

test_that("a partly failing recovery fit keeps the replicates that did fit", {
  # A deterministic per-replicate property of the drawn data decides which fits fail,
  # so the run has both kinds of replicate under a fixed seed.
  real_fit <- efa_fit
  testthat::local_mocked_bindings(
    efa_fit = function(x, ...) {
      if (x[1, 1] > 0) {
        cli::cli_abort("Mocked recovery failure.", class = "efa_mock_fit_failure")
      }
      real_fit(x, ...)
    })

  expect_warning(
    sim <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                     n_datasets = 6, criteria = "EKC", seed = 11),
    class = "efa_power_fit_failed")
  failed <- !sim$replicates$fit_ok
  expect_true(any(failed))
  expect_false(all(failed))

  # The rates are over the completed fits, and each failure keeps its parent cause at
  # its own replicate index.
  expect_equal(sim$convergence$n_fit_ok, sum(!failed))
  expect_equal(sim$convergence$fit_rate, mean(!failed))
  expect_equal(sim$recovery$n_valid, sum(!failed))
  expect_true(all(grepl("Mocked recovery failure",
                        sim$replicates$fit_error[failed], fixed = TRUE)))
  expect_true(all(is.na(sim$replicates$fit_error[!failed])))
})

test_that("an unusable recovery rotation aborts before any data is drawn", {
  # `rotation` is resolved against efa_fit()'s own choices, so a value it would reject
  # costs no simulation time and is reported instead of silently emptying the report.
  expect_error(
    efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 200,
              n_datasets = 500, criteria = "EKC", rotation = "nonesuch"),
    class = "efa_bad_choice")
  expect_error(
    efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 200,
              n_datasets = 500, criteria = "EKC",
              rotation = c("varimax", "promax")),
    class = "efa_bad_choice")
  # A valid rotation is matched case-insensitively and stored canonically, as elsewhere.
  vm <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 120,
                  n_datasets = 2, criteria = "EKC", rotation = "Varimax", seed = 1)
  expect_identical(vm$settings$rotation, "varimax")
  expect_true(all(is.na(vm$replicates$fit_error)))
})

test_that("simulation mode: R-only population turns recovery off but keeps the hit-rate", {
  R_pop <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi, return_pop = TRUE)$population
  rr <- efa_power("simulation", R = R_pop, k = 3, N = 200, n_datasets = 6,
                  criteria = "EKC", seed = 42)
  expect_null(rr$recovery)
  expect_false(rr$settings$has_recovery)
  expect_equal(rr$k_true, 3L)
  expect_true(is.finite(rr$hit_rate[["EKC_BvA2017"]]))
})

test_that("model error follows each function's own documented default", {
  # The two defaults deliberately differ: efa_power() perturbs with TKL, whose minor
  # common factors degrade structure recovery as well as fit, while efa_simulate()
  # defaults to CB. The same target passed to both therefore builds different
  # populations unless `model_error` is set explicitly.
  pw <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 150,
                  n_datasets = 2, criteria = "EKC", target_rmsea = 0.05, seed = 1)
  expect_identical(pw$settings$model_error, "TKL")
  expect_identical(pw$model_error$method, "TKL")
  expect_equal(pw$model_error$target_rmsea, 0.05)
  expect_lt(abs(pw$model_error$rmsea - 0.05), 0.02)

  sim <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi, target_rmsea = 0.05,
                      return_pop = TRUE, seed = 1)
  expect_identical(sim$model_error$method, "CB")

  # Without a target neither function perturbs the population, whatever the method.
  exact <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 120,
                     n_datasets = 2, criteria = "EKC", seed = 1)
  expect_null(exact$model_error)
})

test_that("simulation-mode classed conditions fire as expected", {
  R_pop <- efa_simulate(Lambda = sim_Lambda, Phi = sim_Phi, return_pop = TRUE)$population

  # The population must be given exactly one way.
  expect_error(efa_power("simulation", N = 200, n_datasets = 2),
               class = "efa_power_input")
  expect_error(efa_power("simulation", Lambda = sim_Lambda, R = R_pop, N = 200,
                         n_datasets = 2), class = "efa_power_input")
  # k is required with a bare R, and must agree with Lambda otherwise.
  expect_error(efa_power("simulation", R = R_pop, N = 200, n_datasets = 2),
               class = "efa_power_missing_k")
  expect_error(efa_power("simulation", Lambda = sim_Lambda, k = 2, N = 200,
                         n_datasets = 2), class = "efa_power_bad_k")
  # p is read off the population, like k: a conflicting value is an error on both
  # population routes rather than being silently replaced.
  expect_error(efa_power("simulation", Lambda = sim_Lambda, p = 99, N = 200,
                         n_datasets = 2), class = "efa_power_bad_p")
  expect_error(efa_power("simulation", R = R_pop, k = 3, p = 99, N = 200,
                         n_datasets = 2), class = "efa_power_bad_p")
  # A matching p is accepted and recorded.
  expect_identical(
    efa_power("simulation", Lambda = sim_Lambda, p = nrow(sim_Lambda), N = 120,
              n_datasets = 2, criteria = "EKC", seed = 1)$settings$p,
    nrow(sim_Lambda))
  # N is required in simulation mode (it is optional in RMSEA mode, where it is solved
  # for), and both N and n_datasets must be positive whole numbers.
  expect_error(efa_power("simulation", Lambda = sim_Lambda), class = "efa_power_input")
  expect_error(efa_power("simulation", Lambda = sim_Lambda, N = -5, n_datasets = 2),
               class = "efa_power_input")
  expect_error(efa_power("simulation", Lambda = sim_Lambda, N = 200, n_datasets = 0),
               class = "efa_power_input")
  # Out-of-range recovery threshold.
  expect_error(efa_power("simulation", Lambda = sim_Lambda, N = 200, n_datasets = 2,
                         recovery_threshold = 0), class = "efa_power_bad_threshold")
  # A visual criterion makes no numeric suggestion to score.
  expect_error(efa_power("simulation", Lambda = sim_Lambda, N = 200, n_datasets = 2,
                         criteria = "SCREE"))
  # A simulation object has no analytic power curve to plot.
  sim <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 120,
                   n_datasets = 3, criteria = "EKC", seed = 42)
  expect_error(plot(sim), class = "efa_power_no_plot")
})

test_that("simulation mode runs the internally-simulating criteria", {
  skip_if_not_slow()
  sim <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 250,
                   n_datasets = 10, criteria = c("EKC", "NEST", "PARALLEL"),
                   seed = 42)
  expect_true("EKC_BvA2017" %in% names(sim$hit_rate))
  expect_true(any(grepl("NEST", names(sim$hit_rate))))
  expect_true(any(grepl("PARALLEL", names(sim$hit_rate))))
  expect_gte(sim$hit_rate[["EKC_BvA2017"]], 0.8)
})

test_that("simulation-mode printed output is stable", {
  local_reproducible_output()
  sim <- efa_power("simulation", Lambda = sim_Lambda, Phi = sim_Phi, N = 200,
                   n_datasets = 10, criteria = c("EKC", "MAP"), seed = 42)
  # Every decimal here -- the retention hit rates and the median minimum congruence alike --
  # is averaged over ten simulated fits and moves with the BLAS, so all of them are masked.
  # What this snapshot holds fixed is the layout, the section order, the criterion labels and
  # the integer replicate counts. The rates themselves are pinned numerically instead, by the
  # expect_gte() bounds at lines 268 and 361.
  expect_snapshot(print(sim), transform = scrub_num)
})

#' Exploratory factor analysis (EFA)
#'
#' This function does an EFA with either \code{PAF}, \code{ML},
#' or \code{ULS} with or without subsequent rotation.
#' All arguments with default value \code{NA} can be left to default if \code{type}
#' is set to one of "EFAtools", "SPSS", or "psych". The respective specifications are
#' then handled according to the specified type (see details). For all rotations
#' except varimax and promax, the \code{GPArotation} package is needed.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations. If raw data is entered, the correlation matrix is found from the
#' data.
#' @param n_factors numeric. Number of factors to extract.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used. If input is a correlation matrix and \code{N} = NA
#' (default), not all fit indices can be computed.
#' @param method character. One of "PAF", "ML", or "ULS" to use principal axis
#' factoring, maximum likelihood, or unweighted least squares (also called minres),
#' respectively, to fit the EFA.
#' @param rotation character. Either perform no rotation ("none"; default),
#' an orthogonal rotation ("varimax", "equamax", "quartimax", "geominT",
#' "bentlerT", or "bifactorT"), or an oblique rotation ("promax", "oblimin",
#' "quartimin", "simplimax", "bentlerQ", "geominQ", or "bifactorQ").
#' @param se character. Whether and how standard errors should be computed.
#'  Currently, only "np-boot" for non-parametric bootstrap is available
#'  and needs raw data to work. Default is "none".
#' @param type character. If one of "EFAtools" (default), "psych", or "SPSS" is
#'  used, and the following arguments with default NA are left with
#'  NA, these implementations are executed according to the respective program
#'  ("psych" and "SPSS") or according to the best solution found in Grieder &
#'  Steiner (2020; "EFAtools"). Individual properties can be adapted using one of
#'  the three types and specifying some of the following arguments. If set to
#'  "none" additional arguments must be specified depending on the \code{method}
#'  and \code{rotation} used (see details).
#' @param max_iter numeric. The maximum number of iterations to perform after which
#' the iterative PAF procedure is halted with a warning. If \code{type} is one of
#' "EFAtools", "SPSS", or "psych", this is automatically specified if \code{max_iter} is
#' left to be \code{NA}, but can be overridden by entering a number. Default is
#' \code{NA}.
#' @param init_comm character. The method to estimate the initial communalities
#' in \code{PAF}. "smc" will use squared multiple correlations, "mac" will use
#' maximum absolute correlations, "unity" will use 1s (see details).
#' Default is \code{NA}.
#' @param criterion numeric. The convergence criterion used for PAF.
#' If the change in communalities from one iteration to the next is smaller than
#' this criterion the solution is accepted and the procedure ends.
#' Default is \code{NA}.
#' @param criterion_type character. Type of convergence criterion used for
#' PAF. "max_individual" selects the maximum change in any of the
#' communalities from one iteration to the next and tests it against the
#' specified criterion. This is also used by SPSS. "sum" takes the difference of
#' the sum of all communalities in one iteration and the sum of all communalities
#' in the next iteration and tests this against the criterion. This procedure is
#' used by the \code{\link[psych:fa]{psych::fa}} function. Default is \code{NA}.
#' @param abs_eigen logical. Which algorithm to use in the PAF
#' iterations. If FALSE, the loadings are computed from the eigenvalues. This is
#' also used by the \code{\link[psych:fa]{psych::fa}} function. If TRUE the
#' loadings are computed with the absolute eigenvalues as done by SPSS.
#' Default is \code{NA}.
#' @param use character. Passed to \code{\link[stats:cor]{stats::cor}} if raw data
#' is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Passed to \code{\link[stats:cor]{stats::cor}}.
#' Default is "pearson".
#' @param k numeric. Either the power used for computing the target matrix P in
#' the promax rotation or the number of 'close to zero loadings' for the simplimax
#' rotation (see \code{\link[GPArotation:GPA]{GPArotation::GPFoblq}}). If left to
#' \code{NA} (default), the value for promax depends on the specified type.
#' For simplimax, \code{nrow(L)}, where L is the matrix of unrotated loadings,
#' is used by default.
#' @param normalize logical. If \code{TRUE}, a kaiser normalization is
#' performed before the specified rotation. Default is \code{TRUE}.
#' @param P_type character. This specifies how the target
#' matrix P is computed in promax rotation. If "unnorm" it will use the
#' unnormalized target matrix as originally done in Hendrickson and White (1964).
#' This is also used in the psych and stats packages. If "norm" it will use the
#' normalized target matrix as used in SPSS. Default is \code{NA}.
#' @param precision numeric. The tolerance for stopping in the rotation
#' procedure. Default is 10^-5 for all rotation methods.
#' @param varimax_type character. The type of the varimax rotation performed.
#' If "svd", singular value decomposition is used, as \link[stats:varimax]{stats::varimax} does. If "kaiser", the varimax procedure performed in SPSS is used.
#' This is the original procedure from Kaiser (1958), but with slight alterations
#' in the varimax criterion (see details, and Grieder & Steiner, 2020). Default is \code{NA}.
#' @param order_type character. How to order the factors. "eigen" will reorder
#' the factors according to the largest to lowest eigenvalues of the matrix of
#' rotated loadings. "ss_factors" will reorder the factors according to descending
#' sum of squared factor loadings per factor. Default is \code{NA}.
#' @param start_method character. How to specify the starting values for the
#' optimization procedure for ML. Default is "psych" which takes the
#' starting values specified in \link[psych:fa]{psych::fa}. "factanal" takes the
#' starting values specified in the \link[stats:factanal]{stats::factanal} function.
#' Solutions are very similar.
#' @param b_boot numeric. The number of bootstrap samples to draw. Default is 1000.
#' @param ci numeric. The confidence interval to create from the bootstrap samples.
#'  Must be between 0 and 1. Default ist .95 for 95\% CIs.
#' @param randomStarts numeric. The number of random starts to use in rotations
#'  that use the \code{GPArotation} package. Some rotations are prone to produce
#'  local minima and sometimes many random starts are needed (see the GPArotation
#'  package documentation for details).
#' @param ... Additional arguments passed to rotation functions from the \code{GPArotation} package (e.g., \code{maxit} for maximum number of iterations).
#'
#' @details There are two main ways to use this function. The easiest way is to
#' use it with a specified \code{type} (see above), which sets most of the other
#' arguments accordingly. Another way is to use it more flexibly by explicitly
#' specifying all arguments used and set \code{type} to "none" (see examples).
#' A mix of the two can also be done by specifying a \code{type} as well as
#' additional arguments. However, this will throw warnings to avoid unintentional
#' deviations from the implementations according to the specified \code{type}.
#'
#' The \code{type} argument is evaluated for PAF and for all rotations (mainly
#' important for the varimax and promax rotations). The type-specific settings
#' for these functions are detailed below.
#'
#' For PAF, the values of \code{init_comm}, \code{criterion}, \code{criterion_type},
#' and \code{abs_eigen} depend on the \code{type} argument.
#'
#' \code{type = "EFAtools"} will use the following argument specification:
#' \code{init_comm = "smc", criterion = .001, criterion_type = "sum",
#' abs_eigen = TRUE}.
#'
#' \code{type = "psych"} will use the following argument specification:
#' \code{init_comm = "smc", criterion = .001, criterion_type = "sum",
#' abs_eigen = FALSE}.
#'
#' \code{type = "SPSS"} will use the following argument specification:
#' \code{init_comm = "smc", criterion = .001, criterion_type = "max_individual",
#' abs_eigen = TRUE}.
#'
#' If SMCs fail, SPSS takes "mac". However, as SPSS takes absolute eigenvalues,
#' this is hardly ever the case. Psych, on the other hand, takes "unity" if SMCs
#' fail, but uses the Moore-Penrose Psudo Inverse of a matrix, thus, taking "unity"
#' is only necessary if negative eigenvalues occur afterwards in the iterative
#' PAF procedure. The EFAtools type setting combination was the best in terms of accuracy
#' and number of Heywood cases compared to all the
#' other setting combinations tested in simulation studies in Grieder & Steiner
#' (2020), which is why this type is used as a default here.
#'
#' For varimax, the values of \code{varimax_type} and \code{order_type} depend on
#' the \code{type} argument.
#'
#' \code{type = "EFAtools"} will use the following argument specification:
#' \code{varimax_type = "kaiser", order_type = "eigen"}.
#'
#' \code{type = "psych"} will use the following argument specification:
#' \code{varimax_type = "svd", order_type = "eigen"}.
#'
#' \code{type = "SPSS"} will use the following argument specification:
#' \code{varimax_type = "kaiser", order_type = "ss_factors"}.
#'
#' For promax, the values of \code{P_type},
#' \code{order_type}, and \code{k} depend on the \code{type} argument.
#'
#' \code{type = "EFAtools"} will use the following argument specification:
#' \code{P_type = "norm", order_type = "eigen", k = 4}.
#'
#' \code{type = "psych"} will use the following argument specification:
#' \code{P_type = "unnorm", order_type = "eigen", k = 4}.
#'
#' \code{type = "SPSS"} will use the following argument specification:
#' \code{P_type = "norm", order_type = "ss_factors", k = 4}.
#'
#' The \code{P_type} argument can take two values, "unnorm" and "norm". It controls
#' which formula is used to compute the target matrix P in the promax rotation.
#' "unnorm" uses the formula from Hendrickson and White (1964), specifically:
#' \code{P = abs(A^(k + 1)) / A},
#' where A is the unnormalized matrix containing varimax rotated loadings.
#' "SPSS" uses the normalized varimax rotated loadings. Specifically it used the
#' following formula, which can be found in the SPSS 23 and SPSS 27 Algorithms manuals:
#' \code{P = abs(A / sqrt(rowSums(A^2))) ^(k + 1) * (sqrt(rowSums(A^2)) / A)}.
#' As for PAF, the EFAtools type setting combination for promax was the best
#' compared to the other setting combinations tested in simulation studies in
#' Grieder & Steiner (2020).
#'
#' The \code{varimax_type} argument can take two values, "svd", and "kaiser". "svd" uses
#' singular value decomposition, by calling \link[stats:varimax]{stats::varimax}. "kaiser"
#' performs the varimax procedure as described in the SPSS 23 Algorithms manual and as described
#' by Kaiser (1958). However, there is a slight alteration in computing the varimax criterion, which
#' we found to better align with the results obtain from SPSS. Specifically, the original varimax
#' criterion as described in the SPSS 23 Algorithms manual is
#' \code{sum(n*colSums(lambda ^ 4) - colSums(lambda ^ 2) ^ 2) / n ^ 2}, where n is the
#' number of indicators, and lambda is the rotated loadings matrix. However, we found the following
#' to produce results more similar to those of SPSS:
#' \code{sum(n*colSums(abs(lambda)) - colSums(lambda ^ 4) ^ 2) / n^2}.
#'
#' For all other rotations except varimax and promax, the \code{type} argument
#' only controls the \code{order_type} argument with the same values as stated
#' above for the varimax and promax rotations. For these other rotations, the
#' \code{GPArotation} package is needed. Additional arguments can also be
#' specified and will be passed to the respective \code{GPArotation} function
#' (e.g., maxit to change the maximum number of iterations for the rotation procedure).
#'
#' The \code{type} argument has no effect on ULS and ML. For ULS, no additional
#' arguments are needed. For ML, an additional argument
#' \code{start_method} is needed to determine the starting values for the
#' optimization procedure. Default for this argument is "factanal" which takes
#' the starting values specified in the \link[stats:factanal]{stats::factanal} function.
#'
#'
#' @return A list of class EFA containing (a subset of) the following:
#'
#' \item{orig_R}{Original correlation matrix.}
#' \item{h2_init}{Initial communality estimates from PAF.}
#' \item{h2}{Final communality estimates from the unrotated solution.}
#' \item{orig_eigen}{Eigen values of the original correlation matrix.}
#' \item{init_eigen}{Initial eigenvalues, obtained from the correlation matrix
#'  with the initial communality estimates as diagonal in PAF.}
#' \item{final_eigen}{Eigenvalues obtained from the correlation matrix
#'  with the final communality estimates as diagonal.}
#' \item{iter}{The number of iterations needed for convergence.}
#' \item{convergence}{Integer code for convergence as returned by
#' \code{\link[stats:optim]{stats:optim}} (only for ML and ULS).
#' 0 indicates successful completion.}
#' \item{unrot_loadings}{Loading matrix containing the final unrotated loadings.}
#' \item{vars_accounted}{Matrix of explained variances and sums of squared loadings. Based on the unrotated loadings.}
#' \item{fit_indices}{For ML and ULS: Fit indices derived from the unrotated
#' factor loadings: Chi Square, including significance level, degrees of freedom
#' (df), Comparative Fit Index (CFI), Root Mean Square Error of Approximation
#' (RMSEA), including its 90\% confidence interval, Akaike Information Criterion
#' (AIC), Bayesian Information Criterion (BIC), and the common part accounted
#' for (CAF) index as proposed by Lorenzo-Seva, Timmerman, & Kiers (2011).
#' For PAF, only the CAF and dfs are returned.}
#' \item{model_implied_R}{The model implied correlation
#' matrix.}
#' \item{residuals}{Residual correlations, i.e., orig_R - model_implied_R}
#' \item{rot_loadings}{Loading matrix containing the final rotated loadings
#' (pattern matrix).}
#' \item{Phi}{The factor intercorrelations (only for oblique rotations).}
#' \item{Structure}{The structure matrix (only for oblique rotations).}
#' \item{rotmat}{The rotation matrix.}
#' \item{vars_accounted_rot}{Matrix of explained variances and sums of squared
#' loadings. Based on rotated loadings and, for oblique rotations, the factor
#' intercorrelations.}
#' \item{settings}{A list of the settings used.}
#' \item{boot.SE}{A list bootstrap standard errors for loadings (rotated and unrotated), structure coefficients (if rotated obliquely), factor correlations (Phi, only if rotated), and fit indices. Only returned, if \code{se = "np-boot"}.}
#' \item{boot.CI}{A list bootstrap confidence intervals of width \code{ci} for loadings (rotated and unrotated), structure coefficients (if rotated obliquely), factor correlations (Phi, if obliquely rotated), and fit indices. Only returned, if \code{se = "np-boot"}.}
#' \item{boot.arrays}{A list of arrays with the bootstrapped loadings (aligned rotated and unrotated), aligned structure coefficients (if rotated obliquely), aligned factor correlations (Phi, if obliquely rotated), and fit indices. Only returned, if \code{se = "np-boot"}.}
#'
#' @source Grieder, S., & Steiner, M.D. (2020). Algorithmic Jingle Jungle:
#' A Comparison of Implementations of Principal Axis Factoring and Promax Rotation
#'  in R and SPSS. Manuscript in Preparation.
#' @source Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
#' rotation to oblique simple structure. British Journal of Statistical Psychology,
#' 17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x
#' @source Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
#' Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
#' Research, 46, 340-364, doi: 10.1080/00273171.2011.564527
#' @source Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
#' factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233
#'
#' @export
#'
#' @examples
#' # A type EFAtools (as presented in Steiner and Grieder, 2020) EFA
#' EFAtools_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                     type = "EFAtools", method = "PAF", rotation = "none")
#'
#' # A type SPSS EFA to mimick the SPSS implementation (this will throw a warning,
#' # see below)
#' SPSS_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                 type = "SPSS", method = "PAF", rotation = "none")
#'
#' # A type psych EFA to mimick the psych::fa() implementation
#' psych_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                  type = "psych", method = "PAF", rotation = "none")
#'
#' # Use ML instead of PAF with type EFAtools
#' EFAtools_ML <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                    type = "EFAtools", method = "ML", rotation = "none")
#'
#' # Use oblimin rotation instead of no rotation with type EFAtools
#' EFAtools_oblim <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                       type = "EFAtools", method = "PAF", rotation = "oblimin")
#'
#' # Do a PAF without rotation without specifying a type, so the arguments
#' # can be flexibly specified (this is only recommended if you know what you're
#' # doing)
#' PAF_none <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                 type = "none", method = "PAF", rotation = "none",
#'                 max_iter = 500, init_comm = "mac", criterion = 1e-4,
#'                 criterion_type = "sum", abs_eigen = FALSE)
#'
#' # Add a promax rotation
#' PAF_pro <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                type = "none", method = "PAF", rotation = "promax",
#'                max_iter = 500, init_comm = "mac", criterion = 1e-4,
#'                criterion_type = "sum", abs_eigen = FALSE, k = 3,
#'                P_type = "unnorm", precision= 1e-5, order_type = "eigen",
#'                varimax_type = "svd")
#'
EFA <- function(x, n_factors, N = NA, method = c("PAF", "ML", "ULS"),
                rotation = c("none", "varimax", "equamax", "quartimax", "geominT",
                             "bentlerT", "bifactorT", "promax", "oblimin",
                             "quartimin", "simplimax", "bentlerQ", "geominQ",
                             "bifactorQ"),
                se = c("none", "np-boot"),
                type = c("EFAtools", "psych", "SPSS", "none"), max_iter = NA,
                init_comm = NA, criterion = NA, criterion_type = NA,
                abs_eigen = NA, use = c("pairwise.complete.obs", "all.obs",
                                          "complete.obs", "everything",
                                          "na.or.complete"),
                varimax_type = NA,
                k = NA, normalize = TRUE, P_type = NA, precision = 1e-5,
                order_type = NA, start_method = "psych",
                cor_method = c("pearson", "spearman", "kendall"),
                b_boot = 1000, ci = .95,
                randomStarts = 10,
                ...) {

  # Perform argument checks
  if(!inherits(x, c("matrix", "data.frame"))){

    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" 'x' is neither a matrix nor a dataframe. Either provide a correlation matrix or a dataframe or matrix with raw data.\n"))

  }

  method <- match.arg(method)
  rotation <- match.arg(rotation)
  se <- match.arg(se)
  np_boot <- se == "np-boot"
  use <- match.arg(use)
  cor_method <- match.arg(cor_method)
  type <- match.arg(type)
  start_method <- checkmate::matchArg(start_method, c("psych", "factanal", NA))

  if (is.na(start_method) && method == "ML") {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" method is 'ML' but no start method is defined. Please set start_method to 'psych' or 'factanal'.\n"))
  }

  checkmate::assert_count(n_factors)
  checkmate::assert_count(N, na.ok = TRUE)
  checkmate::assert_count(max_iter, na.ok = TRUE)
  checkmate::assert_choice(init_comm, c("smc", "mac", "unity", NA))
  checkmate::assert_number(criterion, lower = 0, upper = 1, na.ok = TRUE)
  checkmate::assert_choice(criterion_type, c("max_individual", "sum", NA))
  checkmate::assert_flag(abs_eigen, na.ok = TRUE)
  checkmate::assert_number(k, na.ok = TRUE)
  checkmate::assert_choice(varimax_type, c("svd", "kaiser", NA))
  checkmate::assert_flag(normalize, na.ok = TRUE)
  checkmate::assert_choice(P_type, c("unnorm", "norm", NA))
  checkmate::assert_number(precision, lower = 0, upper = 1)
  checkmate::assert_choice(order_type, c("eigen", "ss_factors", NA))
  checkmate::assert_integerish(b_boot, len = 1, any.missing = FALSE)
  checkmate::assert_number(ci, lower = 0, upper = 1)

  # Check if it is a correlation matrix
  if(.is_cormat(x)){

      R <- x

      if (isTRUE(np_boot)) {
        cli::cli_warn(
          c("Cannot compute bootstrap standard errors from correlation matrix.",
          "x" = "You've supplied {.var se} = {.val {se}}, but {.var x} is a correlation matrix.",
          "i" = "Setting {.var se} to {.val none}. Rerun with raw data to calculate bootstrap SEs.")
        )
      }

      np_boot <- FALSE
      se <- "none"

  } else {

    message(cli::col_cyan(cli::symbol$info, " 'x' was not a correlation matrix. Correlations are found from entered raw data.\n"))

    if (!is.na(N)) {
      warning(crayon::yellow$bold("!"), crayon::yellow(" 'N' was set and data entered. Taking N from data.\n"))
    }

    R <- stats::cor(x, use = use, method = cor_method)
    colnames(R) <- colnames(x)
    N <- nrow(x)

    if (isTRUE(np_boot)) {
      m <- ncol(R)
      rows <- 1:N

      # create bootstrap samples and from these, correlation matrices
      R_boot_array <- array(NA_real_, c(m, m, b_boot), dimnames = list(colnames(x),
                                                                 colnames(x),
                                                                 NULL))

      for (boot_i in seq_len(b_boot)) {
        ind <- sample(rows, size = N, replace = TRUE)
        R_boot_array[,, boot_i] <- stats::cor(x[ind,], use = use, method = cor_method)
      }

    }

  }

  # Check if correlation matrix is invertible, if it is not, stop with message
  R_i <- try(solve(R), silent = TRUE)

  if (inherits(R_i, "try-error") && type != "psych") {
    stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is singular, no further analyses are performed\n"))
  }

  # Check if correlation matrix is positive definite, if it is not,
  # smooth the matrix (cor.smooth throws a warning)
  if (any(eigen(R, symmetric = TRUE, only.values = TRUE)$values <= .Machine$double.eps^.6)) {
    if (type == "SPSS") {
      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(" Correlation matrix is not positive definite, no further analyses are performed.\n"))
    }

    R <- psych::cor.smooth(R)

  }

  # Check if model is identified

  # calculate degrees of freedom
  m <- ncol(R)
  df <- ((m - n_factors)**2 - (m + n_factors)) / 2

  if(df < 0){

    warning(crayon::yellow$bold("!"), crayon::yellow(" The model is underidentified. No fit indices were calculated. Please enter a lower number of factors or use a larger number of indicators.\n"))

  } else if (df == 0){

    warning(crayon::yellow$bold("!"), crayon::yellow(" The model is just identified (df = 0). We suggest to try again with a lower number of factors or a larger number of indicators.\n"))

  }

  # run factor analysis with respective fit method

  if (method == "PAF") {

  fit_out <- .PAF(R, n_factors = n_factors, N = N, type = type,
                 max_iter = max_iter, init_comm = init_comm,
                 criterion = criterion, criterion_type = criterion_type,
                 abs_eigen = abs_eigen)


  if (isTRUE(np_boot)) {

    boot_fits <- .boot_fun(R_boot_array, b_boot, .PAF,
                           # .PAF arguments:
                           n_factors = n_factors, N = N, type = type,
                           max_iter = max_iter, init_comm = init_comm,
                           criterion = criterion, criterion_type = criterion_type,
                           abs_eigen = abs_eigen)

  }

  } else if (method == "ML") {

    if (type == "SPSS") {

      warning(crayon::yellow$bold("!"), crayon::yellow(" 'ML' is used as method with type 'SPSS'. Note that only the 'PAF' method is tested to provide the SPSS implementation and results may therefore differ from those returned by SPSS.\n"))

    }

    if (is.na(N)) {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Argument 'N' was NA, not all fit indices could be computed. To get all fit indices, either provide N or raw data.\n"))

    }

    fit_out <- .ML(R, n_factors = n_factors, N = N, start_method = start_method)

    if (isTRUE(np_boot)) {

      boot_fits <- .boot_fun(R_boot_array, b_boot, .ML,
                             # .ML arguments:
                             n_factors = n_factors, N = N,
                             start_method = start_method)

    }

  } else if (method == "ULS") {

    if (type == "SPSS") {

      warning(crayon::yellow$bold("!"), crayon::yellow(" 'ULS' is used as method with type 'SPSS'. Note that only the 'PAF' method is tested to provide the SPSS implementation and results may therefore differ from those returned by SPSS.\n"))

    }

    if (is.na(N)) {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Argument 'N' was NA, not all fit indices could be computed. To get all fit indices, either provide N or raw data.\n"))

    }

    fit_out <- .ULS(R, n_factors = n_factors, N = N)

    if (isTRUE(np_boot)) {

      boot_fits <- .boot_fun(R_boot_array, b_boot, .ULS,
                             # .ULS arguments:
                             n_factors = n_factors, N = N)

    }
  }

  # rotate factor analysis results
  if (rotation == "promax") {

    rot_out <- .PROMAX(fit_out, type = type, normalize = normalize, P_type = P_type,
                      precision = precision, order_type = order_type,
                      varimax_type = varimax_type, k = k)

    boot_rot <- "oblique"


  } else if (rotation == "varimax") {

    rot_out <- .VARIMAX(fit_out, type = type, normalize = normalize,
                       precision = precision, varimax_type = varimax_type,
                       order_type = order_type)

    boot_rot <- "orthogonal"

  } else if (rotation == "quartimax" || rotation == "equamax" ||
             rotation == "bentlerT" || rotation == "geominT" ||
             rotation == "bifactorT") {

    if (type == "SPSS") {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Note that only the 'promax' and 'varimax' rotations are tested to match the SPSS implementation and results may therefore differ from those returned by SPSS.\n"))

    }

    rot_out <- .ROTATE_ORTH(fit_out, type = type, rotation = rotation,
                           normalize = normalize, precision = precision,
                           order_type = order_type,
                           randomStarts = randomStarts, ...)

    boot_rot <- "orthogonal"

  } else if (rotation == "oblimin" || rotation == "quartimin" ||
             rotation == "simplimax" || rotation == "bentlerQ" ||
             rotation == "geominQ" || rotation == "bifactorQ") {

    if (type == "SPSS") {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Note that only the 'promax' and 'varimax' rotations are tested to match the SPSS implementation and results may therefore differ from those returned by SPSS.\n"))

    }

    rot_out <- .ROTATE_OBLQ(fit_out, type = type, rotation = rotation,
                           normalize = normalize, precision = precision,
                           order_type = order_type, k = k,
                           randomStarts = randomStarts, ...)

    boot_rot <- "oblique"

  } else {

    output <- fit_out

    boot_rot <- "none"

  }

  if (rotation != "none"){

    if(method == "ULS"){

      settings <- rot_out$settings
      output <- c(fit_out, within(rot_out, rm(settings)),
                  settings = list(settings))

    } else {

      settings <- c(fit_out$settings, rot_out$settings)
      output <- c(within(fit_out, rm(settings)), within(rot_out, rm(settings)),
                  settings = list(settings))

    }

  }

  # Add settings used to output
  settings_EFA <- list(
    method = method,
    rotation = rotation,
    type = type,
    n_factors = n_factors,
    N = N,
    use = use,
    cor_method = cor_method,
    se = se,
    b_boot = b_boot,
    ci = ci
  )

  if(method == "ULS" & rotation == "none"){

    output <- c(output, settings = list(settings_EFA))

  } else {

    settings <- c(settings_EFA, output$settings)

    output <- c(within(output, rm(settings)),
                settings = list(settings))

  }

  if (isTRUE(np_boot)) {
    if (rotation == "none") {
      L_rot <- NULL
    } else {
      L_rot <- rot_out$rot_loadings
    }
    boot_out <- .boot_se_ci(fit_out, L_rot,
                            boot_fits, boot_rot, ci, b_boot)
    output <- c(output, boot = boot_out)
    output$standardized_residuals <- output$residuals / boot_out$SE$residuals
  }

  class(output) <- "EFA"

  return(output)

}



.boot_fun <- function(x, b, call_fun, ...) {

  boot_list <- list()

  for (boot_i in seq_len(b)) {

    # save complete output, as this has to be passed to rotation method
    boot_list[[boot_i]] <- call_fun(x[,, boot_i], ...)
  }

  boot_list

}

.boot_se_ci <- function(fit_target, L_rot, boot_fit, boot_rot, ci, b) {

  l_ci <- (1 - ci) / 2
  ps <- c(l_ci, ci + l_ci)

  ### calculate stats for unrot loadings and gof measures
  L_unrot <- fit_target$unrot_loadings

  ncol_L <- ncol(L_unrot)
  nrow_L <- nrow(L_unrot)
  colnam_L <- colnames(L_unrot)
  rownam_L <- rownames(L_unrot)

  L_unrot_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))
  gof_boot <- matrix(NA_real_, ncol = length(fit_target$fit_indices), nrow = b)
  residuals_boot <- array(NA_real_, c(nrow_L, nrow_L, b),
                          dimnames = list(rownam_L, rownam_L,
                                          NULL))

  for (boot_i in seq_len(b)) {

    # save aligned loading matrix
    L_unrot_boot[,, boot_i] <- .align_solution(L_unrot, boot_fit[[boot_i]]$unrot_loadings)$loadings
    gof_boot[boot_i, ] <- unlist(boot_fit[[boot_i]]$fit_indices)
    residuals_boot[,, boot_i] <- boot_fit[[boot_i]]$residuals

  }

  # se = sd of bootstrap replications (Zhang, 2014, Estimating Standard Errors
  # in Exploratory Factor Analysis)
  L_unrot_se_ci <- .array_se_ci(L_unrot_boot, ps)
  gof_se_ci <- .array_se_ci(gof_boot, ps, M = 2)
  residuals_se_ci <- .array_se_ci(residuals_boot, ps)
  names(gof_se_ci$se) <- names(fit_target$fit_indices)
  names(gof_se_ci$ci$lower) <- names(fit_target$fit_indices)
  names(gof_se_ci$ci$upper) <- names(fit_target$fit_indices)

  if(boot_rot == "oblique") {

    colnam_L <- colnames(L_rot)

    L_rot_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))
    Phi_rot_boot <- array(NA_real_, c(ncol_L, ncol_L, b),
                          dimnames = list(colnam_L, colnam_L,
                                          NULL))
    Structure_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))

    nonconv_counter <- 0
    for (boot_i in seq_len(b)) {

      rot_i <- boot_fit[[boot_i]]
      # save target-rotated loading matrix
      aligned_i <- PROCRUSTES(boot_fit[[boot_i]]$unrot_loadings,
                              Target = L_rot, rotation = "oblique",
                              oblique_random_starts = 5)
      if (isFALSE(aligned_i$convergence)) {
        nonconv_counter <- nonconv_counter + 1
        next
      }
      L_rot_boot[,, boot_i] <- aligned_i$loadings
      Phi_rot_boot[,, boot_i] <- aligned_i$Phi
      Structure_boot[,, boot_i] <- aligned_i$loadings %*% aligned_i$Phi
    }

    if (nonconv_counter > 0) {
      cli::cli_warn(c("{nonconv_counter} target rotations in bootstrap procedure did not converge.",
                    "i" = "Bootstrap SE and CI of rotated loadings, factor correlations and structure coefficients are based on {b - nonconv_counter} bootstrap samples."))
    }

    L_rot_se_ci <- .array_se_ci(L_rot_boot, ps)
    Phi_rot_se_ci <- .array_se_ci(Phi_rot_boot, ps)
    Structure_se_ci <- .array_se_ci(Structure_boot, ps)

    out <- list(
      SE = list(
        unrot_loadings = L_unrot_se_ci$se,
        rot_loadings = L_rot_se_ci$se,
        Phi = Phi_rot_se_ci$se,
        Structure = Structure_se_ci$se,
        fit_indices = gof_se_ci$se,
        residuals = residuals_se_ci$se,
        valid_target_rotations = b - nonconv_counter
      ),
      CI = list(
        unrot_loadings = L_unrot_se_ci$ci,
        rot_loadings = L_rot_se_ci$ci,
        Phi = Phi_rot_se_ci$ci,
        Structure = Structure_se_ci$ci,
        fit_indices = gof_se_ci$ci,
        residuals = residuals_se_ci$ci
      ),
      arrays = list(
        unrot_loadings = L_unrot_boot,
        rot_loadings = L_rot_boot,
        Phi = Phi_rot_boot,
        Structure = Structure_boot,
        fit_indices = gof_boot,
        residuals = residuals_boot
      )
    )

  } else if (boot_rot == "orthogonal") {
    colnam_L <- colnames(L_rot)

    L_rot_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))

    nonconv_counter <- 0

    for (boot_i in seq_len(b)) {

      rot_i <- boot_fit[[boot_i]]
      # save target-rotated loading matrix
      aligned_i <- PROCRUSTES(boot_fit[[boot_i]]$unrot_loadings,
                              Target = L_rot, rotation = "orthogonal")
      L_rot_boot[,, boot_i] <- aligned_i$loadings
    }

    if (nonconv_counter > 0) {
      cli::cli_warn(c("{nonconv_counter} target rotations in bootstrap procedure did not converge.",
                    "i" = "Bootstrap SE and CI of rotated loadings are based on {b - nonconv_counter} bootstrap samples."))
    }

    L_rot_se_ci <- .array_se_ci(L_rot_boot, ps)

    out <- list(
      SE = list(
        unrot_loadings = L_unrot_se_ci$se,
        rot_loadings = L_rot_se_ci$se,
        fit_indices = gof_se_ci$se,
        residuals = residuals_se_ci$se,
        valid_target_rotations = b - nonconv_counter
      ),
      CI = list(
        unrot_loadings = L_unrot_se_ci$ci,
        rot_loadings = L_rot_se_ci$ci,
        fit_indices = gof_se_ci$ci,
        residuals = residuals_se_ci$ci
      ),
      arrays = list(
        unrot_loadings = L_unrot_boot,
        rot_loadings = L_rot_boot,
        fit_indices = gof_boot,
        residuals = residuals_boot
      )
    )


  } else {

    out <- list(
      SE = list(
        unrot_loadings = L_unrot_se_ci$se,
        fit_indices = gof_se_ci$se,
        residuals = residuals_se_ci$se
      ),
      CI = list(
        unrot_loadings = L_unrot_se_ci$ci,
        fit_indices = gof_se_ci$ci,
        residuals = residuals_se_ci$ci
      ),
      arrays = list(
        unrot_loadings = L_unrot_boot,
        fit_indices = gof_boot,
        residuals = residuals_boot
      )
    )

  }


  out

}

.array_se_ci <- function(x, probs, M = c(1, 2)) {
  se <- apply(x, M, stats::sd, na.rm = TRUE)
  ci <- lapply(probs, function(p) {
    apply(x, M, stats::quantile, probs = p, na.rm = TRUE)
  })
  names(ci) <- c("lower", "upper")

  list(
    se = se,
    ci = ci
  )
}

## Use with an output from the EFAtools::EFA function, both with type EFAtools
EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
SL_EFAtools <- SL(EFA_mod, type = "EFAtools", method = "PAF")

# with type SPSS and method ULS
SL_SPSS <- SL(EFA_mod, type = "SPSS", method = "ULS")

## Use with an output from the psych::fa function with type psych in SL
fa_mod <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                    fm = "pa", rotate = "Promax")
SL_psych <- SL(fa_mod, type = "psych", method = "PAF")

## Use more flexibly by entering a pattern matrix and phi directly, with method
## ML
SL_flex <- SL(EFA_mod$rot_loadings, Phi = EFA_mod$Phi, type = "EFAtools",
              method = "ML")

## Use with a second-order lavaan solution
lav_mod_ho <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18
               g =~ F1 + F2 + F3'
lav_fit_ho <- suppressWarnings(lavaan::cfa(lav_mod_ho,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))
SL_lav <- SL(lav_fit_ho, g_name = "g")

test_that("output class and dimensions are correct", {
  expect_is(SL_EFAtools, "SL")
  expect_is(SL_SPSS, "SL")
  expect_is(SL_psych, "SL")
  expect_is(SL_flex, "SL")
  expect_is(SL_lav, "SL")

  expect_output(str(SL_EFAtools), "List of 6")
  expect_output(str(SL_SPSS), "List of 6")
  expect_output(str(SL_psych), "List of 6")
  expect_output(str(SL_flex), "List of 6")
  expect_output(str(SL_lav), "List of 6")
})

test_that("original correlation is correct", {
  expect_equal(SL_EFAtools$orig_R, test_models$baseline$cormat)
  expect_equal(SL_SPSS$orig_R, test_models$baseline$cormat)
  expect_equal(SL_psych$orig_R, test_models$baseline$cormat)
  expect_equal(SL_flex$orig_R, NA)
  expect_equal(SL_lav$orig_R, NA)
})

test_that("sl solution is correct", {
  expect_equal(unname(SL_EFAtools$sl[, "h2"]) + unname(SL_EFAtools$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_SPSS$sl[, "h2"]) + unname(SL_SPSS$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_psych$sl[, "h2"]) + unname(SL_psych$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_flex$sl[, "h2"]) + unname(SL_flex$sl[, "u2"]),
               rep(1, 18))
  expect_equal(unname(SL_lav$sl[, "h2"]) + unname(SL_lav$sl[, "u2"]),
               rep(1, 18))

  expect_equal(unname(SL_EFAtools$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_SPSS$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_psych$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_flex$sl[, "g"]) >= .20, rep(TRUE, 18))
  expect_equal(unname(SL_lav$sl[, "g"]) >= .20, rep(TRUE, 18))

  expect_equal(unname(SL_EFAtools$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_SPSS$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_psych$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_flex$sl[13:18, "F1"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_lav$sl[1:6, "F1"]) >= .20, rep(TRUE, 6))

  expect_equal(unname(SL_EFAtools$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_SPSS$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_psych$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_flex$sl[1:12, "F1"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_lav$sl[7:18, "F1"]) < .20, rep(TRUE, 12))

  expect_equal(unname(SL_EFAtools$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_SPSS$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_psych$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_flex$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_lav$sl[7:12, "F2"]) >= .20, rep(TRUE, 6))

  expect_equal(unname(SL_EFAtools$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_SPSS$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_psych$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_flex$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_lav$sl[c(1:6, 13:18), "F2"]) < .20, rep(TRUE, 12))

  expect_equal(unname(SL_EFAtools$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_SPSS$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_psych$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_flex$sl[1:6, "F3"]) >= .20, rep(TRUE, 6))
  expect_equal(unname(SL_lav$sl[13:18, "F3"]) >= .20, rep(TRUE, 6))

  expect_equal(unname(SL_EFAtools$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_SPSS$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_psych$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_flex$sl[7:18, "F3"]) < .20, rep(TRUE, 12))
  expect_equal(unname(SL_lav$sl[1:12, "F3"]) < .20, rep(TRUE, 12))
})

test_that("settings are returned correctly", {
  expect_named(SL_EFAtools$settings, c("method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method", "max_iter",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen"))
  expect_named(SL_SPSS$settings, c("method", "rotation", "type", "n_factors",
                                   "N", "use", "cor_method"))
  expect_named(SL_psych$settings, c("method", "rotation", "type", "n_factors",
                                       "N", "use", "cor_method", "max_iter",
                                       "init_comm", "criterion", "criterion_type",
                                       "abs_eigen"))
  expect_named(SL_flex$settings, c("method", "rotation", "type", "n_factors",
                                  "N", "use", "cor_method", "start_method"))
  expect_equal(SL_lav$settings, NA)

  expect_equal(SL_EFAtools$settings$method, "PAF")
  expect_equal(SL_SPSS$settings$method, "ULS")
  expect_equal(SL_psych$settings$method, "PAF")
  expect_equal(SL_flex$settings$method, "ML")

  expect_equal(SL_EFAtools$settings$rotation, "none")
  expect_equal(SL_SPSS$settings$rotation, "none")
  expect_equal(SL_psych$settings$rotation, "none")
  expect_equal(SL_flex$settings$rotation, "none")

  expect_equal(SL_EFAtools$settings$type, "EFAtools")
  expect_equal(SL_SPSS$settings$type, "SPSS")
  expect_equal(SL_psych$settings$type, "psych")
  expect_equal(SL_flex$settings$type, "EFAtools")

  expect_equal(SL_EFAtools$settings$n_factors, 1)
  expect_equal(SL_SPSS$settings$n_factors, 1)
  expect_equal(SL_psych$settings$n_factors, 1)
  expect_equal(SL_flex$settings$n_factors, 1)

  expect_equal(SL_EFAtools$settings$N, 100)
  expect_equal(SL_SPSS$settings$N, 100)
  expect_equal(SL_psych$settings$N, 100)
  expect_equal(SL_flex$settings$N, 100)

  expect_equal(SL_EFAtools$settings$use, "pairwise.complete.obs")
  expect_equal(SL_SPSS$settings$use, "pairwise.complete.obs")
  expect_equal(SL_psych$settings$use, "pairwise.complete.obs")
  expect_equal(SL_flex$settings$use, "pairwise.complete.obs")

  expect_equal(SL_EFAtools$settings$cor_method, "pearson")
  expect_equal(SL_SPSS$settings$cor_method, "pearson")
  expect_equal(SL_psych$settings$cor_method, "pearson")
  expect_equal(SL_flex$settings$cor_method, "pearson")

  expect_equal(SL_EFAtools$settings$max_iter, 300)
  expect_equal(SL_psych$settings$max_iter, 50)

  expect_equal(SL_EFAtools$settings$init_comm, "smc")
  expect_equal(SL_psych$settings$init_comm, "smc")

  expect_equal(SL_EFAtools$settings$criterion, 0.001)
  expect_equal(SL_psych$settings$criterion,  0.001)

  expect_equal(SL_EFAtools$settings$criterion_type, "sum")
  expect_equal(SL_psych$settings$criterion_type, "sum")

  expect_equal(SL_EFAtools$settings$abs_eigen, TRUE)
  expect_equal(SL_psych$settings$abs_eigen, FALSE)

  expect_equal(SL_flex$settings$start_method, "psych")
})


EFA_mod_unrot <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                     type = "EFAtools", method = "PAF", rotation = "none")
EFA_mod_orth <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                     type = "EFAtools", method = "PAF", rotation = "varimax")
fa_mod_unrot <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                          fm = "pa", rotate = "none")
fa_mod_orth <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                         fm = "pa", rotate = "varimax")

lav_mod_NA <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6 + V17
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12 + V2
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18 + V10
               g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                    V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_NA <- suppressWarnings(lavaan::cfa(lav_mod_NA,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

lav_mod_ho_inv <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
                   F2 =~ V7 + V8 + V9 + V10 + V11 + V12
                   g =~ F1 + F2 + V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_ho_inv <- suppressWarnings(lavaan::cfa(lav_mod_ho_inv,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

test_that("errors are thrown correctly", {
  expect_error(SL(1:5), " 'x' is neither an object of class EFA, fa, or lavaan, nor a matrix, nor of class LOADINGS or loadings.\n")
  expect_warning(SL(EFA_mod, type = "EFAtools", method = "PAF", Phi = EFA_mod$Phi),
                 " Phi argument is specified. Specified factor intercorrelations are taken. To take factor intercorrelations from the EFA output, leave Phi = NULL\n")
  expect_error(SL(EFA_mod_unrot, type = "EFAtools", method = "PAF"), " 'x' is either a non-rotated or orthogonal factor solution. SL needs an oblique factor solution\n")
  expect_error(SL(EFA_mod_orth, type = "EFAtools", method = "PAF"), " 'x' is either a non-rotated or orthogonal factor solution. SL needs an oblique factor solution\n")
  expect_warning(SL(fa_mod, type = "EFAtools", method = "PAF", Phi = fa_mod$Phi), " Phi argument is specified. Specified factor intercorrelations are taken. To take factor intercorrelations from the psych fa output, leave Phi = NULL\n")
  expect_error(SL(fa_mod_unrot, type = "EFAtools", method = "PAF"), " 'x' is either a non-rotated or orthogonal factor solution. SL needs an oblique factor solution\n")
  expect_error(SL(fa_mod_orth, type = "EFAtools", method = "PAF"), " 'x' is either a non-rotated or orthogonal factor solution. SL needs an oblique factor solution\n")
  expect_error(SL(lav_fit_NA, g_name = "g"), " Some loadings are NA or NaN. No omegas are computed.\n")
  expect_error(SL(lav_fit_ho, g_name = "fu"), " Could not find the specified name of the general factor in the entered lavaan solution. Please check the spelling.\n")
  expect_warning(SL(lav_fit_ho_inv, g_name = "g"), " The second-order factor you specified contains first-order loadings. Did you really enter a second-order CFA solution? Or did you enter the wrong factor name in g_name?\n", fixed = TRUE)
  expect_error(SL(EFA_mod$rot_loadings, type = "EFAtools", method = "ML"), " Phi not provided. Either enter an oblique factor solution from EFAtools::EFA or from psych::fa, or a second-order CFA solution from lavaan, or provide Phi\n")
})

rm(EFA_mod, SL_EFAtools, SL_SPSS, fa_mod, SL_psych, SL_flex, lav_mod_ho,
   lav_fit_ho, SL_lav, EFA_mod_unrot, EFA_mod_orth, fa_mod_unrot, fa_mod_orth,
   lav_mod_NA, lav_fit_NA, lav_mod_ho_inv, lav_fit_ho_inv)

bart_cor <- efa_bartlett(test_models$baseline$cormat, N = 500)
bart_raw <- efa_bartlett(GRiPS_raw)

set.seed(500)
bart_rand <- efa_bartlett(matrix(rnorm(100), ncol = 4))

test_that("output class and dimensions are correct", {
  expect_identical(class(bart_cor), c("efa_bartlett", "BARTLETT"))
  expect_output(str(bart_cor), "List of 4")
  expect_identical(class(bart_raw), c("efa_bartlett", "BARTLETT"))
  expect_output(str(bart_raw), "List of 4")
})


test_that("p-values and df are correct", {
  expect_lt(bart_cor$p_value, 0.0001)
  expect_lt(bart_raw$p_value, 0.0001)
  expect_gt(bart_rand$p_value, 0.05)

  expect_equal(bart_cor$df, 153)
  expect_equal(bart_raw$df, 28)
  expect_equal(bart_rand$df, 6)
})

test_that("chi-square statistic matches psych", {
  skip_on_cran()
  skip_if_not_installed("psych")

  # correlation-matrix branch (also pinned to the known value so a wrong Bartlett
  # correction is caught even without psych)
  expect_equal(bart_cor$chisq,
               psych::cortest.bartlett(test_models$baseline$cormat, n = 500)$chisq,
               tolerance = 1e-4)
  expect_equal(bart_cor$chisq, 2173.283, tolerance = 1e-3)

  # raw-data branch (N recovered from the data)
  expect_equal(bart_raw$chisq,
               psych::cortest.bartlett(stats::cor(GRiPS_raw),
                                       n = nrow(GRiPS_raw))$chisq,
               tolerance = 1e-4)

  # non-significant random branch
  set.seed(500)
  rand_mat <- matrix(rnorm(100), ncol = 4)
  expect_equal(bart_rand$chisq,
               psych::cortest.bartlett(stats::cor(rand_mat),
                                       n = nrow(rand_mat))$chisq,
               tolerance = 1e-4)
})

test_that("settings are returned correctly", {
  expect_named(bart_cor$settings, c("N", "use", "cor_method"))
  expect_named(bart_raw$settings, c("N", "use", "cor_method"))
  expect_named(bart_rand$settings, c("N", "use", "cor_method"))

  expect_equal(bart_cor$settings$N, 500)
  expect_equal(bart_raw$settings$N, 810)
  expect_equal(bart_rand$settings$N, 25)

  expect_equal(bart_cor$settings$use, "pairwise.complete.obs")
  expect_equal(bart_raw$settings$use, "pairwise.complete.obs")
  expect_equal(bart_rand$settings$use, "pairwise.complete.obs")

  expect_equal(bart_cor$settings$cor_method, "pearson")
  expect_equal(bart_raw$settings$cor_method, "pearson")
  expect_equal(bart_rand$settings$cor_method, "pearson")
})

test_that("the chi-square is NA when N is too small for the Bartlett correction", {
  # the documented guard: the multiplier N - 1 - (2p + 5)/6 turns non-positive at
  # N = 5 for these p = 18 variables, so no statistic is reported. N relative to p is the
  # only actionable fact behind that NA, so it is reported rather than left silent.
  expect_warning(bart_small <- efa_bartlett(test_models$baseline$cormat, N = 5),
                 class = "efa_bartlett_n_too_small")
  expect_true(is.na(bart_small$chisq))
  expect_true(is.na(bart_small$p_value))

  # An N that keeps the multiplier positive must not warn.
  expect_no_warning(efa_bartlett(test_models$baseline$cormat, N = 500),
                    class = "efa_bartlett_n_too_small")
})

test_that("errors are thrown correctly", {
  expect_error(efa_bartlett(1:5), class = "efa_input_not_matrix")
  expect_error(efa_bartlett(test_models$baseline$cormat), class = "efa_n_required")
  expect_message(efa_bartlett(GRiPS_raw), class = "efa_cor_from_data")
  expect_warning(efa_bartlett(GRiPS_raw, N = 20), class = "efa_n_from_data")
  expect_error(efa_bartlett(sing_raw), class = "efa_cor_singular")
  expect_error(efa_bartlett(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_warning(efa_bartlett(cor_nposdef, N = 10), class = "efa_cor_smoothed")
})

test_that("a near-singular but positive-definite matrix still yields a statistic", {
  # det() underflows to 0 for a large, highly correlated (but positive-definite)
  # matrix; the null-model chi-square must be recovered from the log-determinant
  # rather than turning into NA and silently NA-ing Bartlett's test or crashing
  # the CFI/TLI block of the model fit indices.
  R <- matrix(0.999, 110, 110)
  diag(R) <- 1
  expect_gt(min(eigen(R, symmetric = TRUE, only.values = TRUE)$values), 0)
  expect_equal(det(R), 0)  # documents why a naive determinant fails here

  bart <- efa_bartlett(R, N = 500)
  expect_true(is.finite(bart$chisq))
  expect_true(is.finite(bart$p_value))

  # the same matrix must not crash the model fit indices (CFI/TLI use the null)
  expect_no_error(
    suppressWarnings(suppressMessages(EFA(R, n_factors = 1, N = 500, method = "ML"))))
})

test_that("print output is stable", {
  local_reproducible_output()

  # significant
  expect_snapshot(print(bart_cor), transform = scrub_num)

  # not significant
  expect_snapshot(print(bart_rand), transform = scrub_num)

  # test did not render a result
  bart_na <- structure(list(chisq = NA_real_, df = NA_integer_, p_value = NA_real_,
                            settings = list()), class = c("efa_bartlett", "BARTLETT"))
  expect_snapshot(print(bart_na), transform = scrub_num)

  # missing components (NULL) must not error and must still print a stable line
  bart_null <- structure(list(chisq = NULL, df = NULL, p_value = NULL,
                              settings = list()), class = c("efa_bartlett", "BARTLETT"))
  expect_snapshot(print(bart_null), transform = scrub_num)
})

rm(bart_cor, bart_raw, bart_rand)

# Regression net: pin the current estimator output (PAF / ML / ULS) against the
# reference implementations it is designed to reproduce, so an accidental change is
# caught. Comparisons use quantities that are invariant to the sign, order, and
# orthogonal rotation of the unrotated factors (per-variable uniquenesses and the
# off-diagonal reproduced correlations L %*% t(L)); both are robust yet still pin the
# identified solution. Tolerance is set generously relative to the observed agreement
# (noted per block) so the contract is portable across BLAS implementations.

# off-diagonal of the reproduced correlation matrix from a loading matrix
repro_offdiag <- function(loadings) {
  L <- unclass(loadings)
  M <- L %*% t(L)
  M[upper.tri(M)]
}

# baseline correlation matrix (3 factors) and a single-factor matrix of a different
# size; GRiPS is passed as a correlation matrix so the estimator, not the correlation
# construction, is what is compared.
reg_fixtures <- list(
  list(R = test_models$baseline$cormat, N = 500, k = 3),
  list(R = stats::cor(GRiPS_raw, use = "pairwise.complete.obs"), N = 810, k = 1)
)

# Several blocks below read different quantities -- uniquenesses, reproduced correlations,
# chi-square, SRMR, the closed-form fit indices -- off what is argument for argument the
# same fit, so each (fixture, estimator) pair is estimated once and cached. The cache is
# filled on first request rather than eagerly, so the skip gates in each block still
# short-circuit before any estimation is paid for. Extraction is deterministic and draws no
# random numbers, so caching cannot perturb the RNG stream.
reg_fit <- local({
  cache <- list()
  function(i, method) {
    key <- paste(i, method)
    if (!key %in% names(cache)) {
      fx <- reg_fixtures[[i]]
      cache[[key]] <<- suppressWarnings(EFA(fx$R, n_factors = fx$k, N = fx$N,
                                            method = method))
    }
    cache[[key]]
  }
})

# The same holds for the lavaan reference: the three lavaan blocks below fit an identical
# model per fixture and inspect different measures on it, so it too is fitted once and
# cached. A fit that errors is cached as NULL, which each consumer turns into its own skip
# exactly as an inline failure would.
lav_efa_fit <- local({
  cache <- list()
  function(i) {
    key <- as.character(i)
    if (!key %in% names(cache)) {
      fx <- reg_fixtures[[i]]
      cache[key] <<- list(tryCatch(
        lavaan::efa(sample.cov = fx$R, sample.nobs = fx$N, nfactors = fx$k,
                    rotation = "none", estimator = "ML", se = "none")[[1]],
        error = function(e) NULL))
    }
    cache[[key]]
  }
})

test_that("PAF reproduces psych::fa(fm = 'pa')", {
  skip_on_cran()
  skip_if_not_installed("psych")

  for (fx in reg_fixtures) {
    efa <- suppressWarnings(EFA(fx$R, n_factors = fx$k, N = fx$N,
                                type = "psych", method = "PAF"))
    ref <- suppressMessages(suppressWarnings(
      psych::fa(fx$R, nfactors = fx$k, n.obs = fx$N, fm = "pa",
                rotate = "none", SMC = TRUE, max.iter = 50)))

    # observed agreement on the baseline model: ~1e-15
    expect_equal(unname(1 - efa$h2), unname(1 - ref$communality), tolerance = 1e-4)
    expect_equal(repro_offdiag(efa$unrot_loadings),
                 repro_offdiag(ref$loadings), tolerance = 1e-4)
  }
})

test_that("ML reproduces stats::factanal", {
  skip_on_cran()

  for (i in seq_along(reg_fixtures)) {
    fx <- reg_fixtures[[i]]
    efa <- reg_fit(i, "ML")
    ref <- stats::factanal(covmat = fx$R, factors = fx$k, n.obs = fx$N,
                           rotation = "none")

    # observed agreement on the baseline model: ~2e-6
    expect_equal(unname(1 - efa$h2), unname(ref$uniquenesses), tolerance = 1e-4)
    expect_equal(repro_offdiag(efa$unrot_loadings),
                 repro_offdiag(ref$loadings), tolerance = 1e-4)
  }
})

# The reference for ULS is psych::fa(fm = "uls"), which minimises the same criterion:
# the squared Frobenius norm of the reduced residual, unhalved (psych sums the full
# residual matrix where `Fm` halves it). psych's fm = "minres" is the same optimiser on
# the same gradient but sums only the strictly lower triangle, a function that gradient
# does not differentiate, so it terminates short of its own optimum -- on the baseline
# model its solution sits 5.7e-6 from fm = "uls" and 6.3e-6 from fm = "ols", which is
# psych's internal spread rather than a property of this package. Tolerance stays at 1e-4:
# the observed agreement of ~1e-15 reflects a shared optimiser path, not an algebraic
# identity, so it is not a bound that would hold across BLAS implementations or a future
# psych release.
test_that("ULS reproduces psych::fa(fm = 'uls')", {
  skip_on_cran()
  skip_if_not_installed("psych")

  for (i in seq_along(reg_fixtures)) {
    fx <- reg_fixtures[[i]]
    efa <- reg_fit(i, "ULS")
    ref <- suppressMessages(suppressWarnings(
      psych::fa(fx$R, nfactors = fx$k, n.obs = fx$N, fm = "uls",
                rotate = "none")))

    # observed agreement on the baseline model: ~8e-15
    expect_equal(unname(1 - efa$h2), unname(1 - ref$communality), tolerance = 1e-4)
    expect_equal(repro_offdiag(efa$unrot_loadings),
                 repro_offdiag(ref$loadings), tolerance = 1e-4)

    # `Fm` is the ULS objective at the solution, half the squared Frobenius norm of the
    # reduced residual. Its diagonal vanishes at an interior optimum, so it reduces to
    # the squared residual correlations summed over the unique variable pairs -- half
    # the criterion psych minimises, which counts each pair twice. Observed agreement on
    # the baseline model: ~1e-11 relative -- looser than the loadings above because `Fm`
    # also carries the residual diagonal, which is only zero in the limit.
    res_pairs <- fx$R[upper.tri(fx$R)] - repro_offdiag(ref$loadings)
    expect_equal(efa$fit_indices$Fm, sum(res_pairs^2), tolerance = 1e-4)
  }
})

# Fit-index parity: the model chi-square is the Bartlett-corrected ML discrepancy at the
# model-implied matrix. For ML this reproduces stats::factanal / psych::fa(fm = "ml"); for
# ULS it reproduces psych::fa(fm = "uls"). lavaan uses the multiplier N (no Bartlett
# correction) but the same fit function, so its statistic differs by exactly bart / N.
test_that("ML chi-square reproduces stats::factanal and psych::fa(fm = 'ml')", {
  skip_on_cran()
  skip_if_not_installed("psych")

  for (i in seq_along(reg_fixtures)) {
    fx <- reg_fixtures[[i]]
    efa <- reg_fit(i, "ML")
    ref_fa <- stats::factanal(covmat = fx$R, factors = fx$k, n.obs = fx$N,
                              rotation = "none")
    ref_ps <- suppressMessages(suppressWarnings(
      psych::fa(fx$R, nfactors = fx$k, n.obs = fx$N, fm = "ml", rotate = "none")))

    # observed agreement on the baseline model: ~1e-9
    expect_equal(efa$fit_indices$chi, unname(ref_fa$STATISTIC), tolerance = 1e-4)
    expect_equal(efa$fit_indices$chi, unname(ref_ps$STATISTIC), tolerance = 1e-4)
    expect_equal(efa$fit_indices$df, ref_fa$dof)
  }
})

test_that("ULS chi-square reproduces psych::fa(fm = 'uls')", {
  skip_on_cran()
  skip_if_not_installed("psych")

  for (i in seq_along(reg_fixtures)) {
    fx <- reg_fixtures[[i]]
    efa <- reg_fit(i, "ULS")
    ref <- suppressMessages(suppressWarnings(
      psych::fa(fx$R, nfactors = fx$k, n.obs = fx$N, fm = "uls", rotate = "none")))

    # observed agreement on the baseline model: ~1e-14 relative
    expect_equal(efa$fit_indices$chi, unname(ref$STATISTIC), tolerance = 1e-3)
  }
})

test_that("ML chi-square relates to lavaan's via the documented Bartlett factor", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  for (i in seq_along(reg_fixtures)) {
    fx <- reg_fixtures[[i]]
    p <- ncol(fx$R)
    bart <- fx$N - 1 - (2 * p + 5) / 6 - (2 * fx$k) / 3
    efa <- reg_fit(i, "ML")

    # lavaan::efa()'s baseline / robust paths can fail on a covariance-matrix input;
    # take the standard test statistic from the fitted model and skip if unavailable.
    lav_stat <- tryCatch({
      lavaan::lavInspect(lav_efa_fit(i), "test")$standard$stat
    }, error = function(e) NA_real_)
    skip_if(is.na(lav_stat), "lavaan::efa() did not return a standard test statistic")

    expect_equal(efa$fit_indices$chi, lav_stat * bart / fx$N, tolerance = 1e-3)
  }
})

test_that("SRMR reproduces lavaan", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  for (i in seq_along(reg_fixtures)) {
    efa <- reg_fit(i, "ML")

    lav_srmr <- tryCatch({
      unname(lavaan::fitMeasures(lav_efa_fit(i), "srmr"))
    }, error = function(e) NA_real_)
    skip_if(is.na(lav_srmr), "lavaan::efa() did not return SRMR")

    # SRMR is convention-free, so it matches lavaan exactly
    expect_equal(efa$fit_indices$SRMR, lav_srmr, tolerance = 1e-5)
  }
})

test_that("TLI equals its (N - 1)-scale closed form; ECVI uses the Bartlett-corrected chi", {
  for (i in seq_along(reg_fixtures)) {
    fx <- reg_fixtures[[i]]
    p <- ncol(fx$R)
    for (mth in c("ML", "ULS")) {
      fi <- reg_fit(i, mth)$fit_indices

      # CFI/TLI compare the model and baseline noncentralities on a common (N - 1) scale,
      # so the reported (Bartlett-corrected) model and baseline chi-squares are rescaled
      # from their own multipliers to (N - 1).
      mult_m    <- fx$N - 1 - (2 * p + 5) / 6 - (2 * fx$k) / 3
      mult_null <- fx$N - 1 - (2 * p + 5) / 6
      chi_cfi      <- fi$chi      * (fx$N - 1) / mult_m
      chi_null_cfi <- fi$chi_null * (fx$N - 1) / mult_null
      expect_equal(fi$TLI,
                   ((chi_null_cfi / fi$df_null) - (chi_cfi / fi$df)) /
                     ((chi_null_cfi / fi$df_null) - 1),
                   tolerance = 1e-8)

      # ECVI is still built on the Bartlett-corrected model chi-square.
      n_params <- p * (p + 1) / 2 - fi$df
      expect_equal(fi$ECVI, (fi$chi + 2 * n_params) / (fx$N - 1), tolerance = 1e-8)
    }
  }
})

# The PAF variant code paths reproduce the same psych::fa(fm = "pa") solution as the
# already-tested "psych" type. The EFAtools/SPSS variants (absolute eigenvalues, the
# max-individual convergence criterion) hit the fixed point to ~1e-15; the alternate
# initial-communality paths (mac/unity) reach it within the PAF stopping tolerance.
test_that("PAF variant paths reproduce psych::fa(fm = 'pa')", {
  skip_on_cran()
  skip_if_not_installed("psych")

  R <- test_models$baseline$cormat
  N <- 500
  k <- 3
  ref <- suppressMessages(suppressWarnings(
    psych::fa(R, nfactors = k, n.obs = N, fm = "pa",
              rotate = "none", SMC = TRUE, max.iter = 50)))

  variants <- list(
    list(label = "EFAtools", args = list(type = "EFAtools"),                tol = 1e-4),
    list(label = "SPSS",     args = list(type = "SPSS"),                     tol = 1e-4),
    list(label = "mac",      args = list(type = "psych", init_comm = "mac"),   tol = 5e-3),
    list(label = "unity",    args = list(type = "psych", init_comm = "unity"), tol = 5e-3)
  )

  for (v in variants) {
    efa <- suppressWarnings(do.call(
      EFA, c(list(R, n_factors = k, N = N, method = "PAF"), v$args)))
    expect_equal(unname(1 - efa$h2), unname(1 - ref$communality),
                 tolerance = v$tol, info = v$label)
    expect_equal(repro_offdiag(efa$unrot_loadings),
                 repro_offdiag(ref$loadings), tolerance = v$tol, info = v$label)
  }
})

# AIC and BIC follow the documented chi-square forms, not lavaan's -2logL-based versions.
# CFI and RMSEA place the model (and, for CFI, the baseline) noncentrality on a common
# (N - 1) scale, matching lavaan once its statistics are rescaled from the multiplier N to
# N - 1. The Bartlett correction enters only the reported chi-square test (and the AIC/BIC
# derived from it), not these approximation indices.
test_that("CFI, RMSEA, AIC, and BIC match their references on the baseline", {
  skip_on_cran()

  # the baseline fixture, whose ML fit and lavaan reference are shared with the blocks above
  N <- reg_fixtures[[1]]$N
  fi <- reg_fit(1L, "ML")$fit_indices

  expect_equal(fi$AIC, fi$chi - 2 * fi$df, tolerance = 1e-8)
  expect_equal(fi$BIC, fi$chi - log(N) * fi$df, tolerance = 1e-8)

  skip_if_not_installed("lavaan")

  lav <- tryCatch({
    lavaan::fitMeasures(lav_efa_fit(1L),
                        c("chisq", "df", "baseline.chisq", "baseline.df"))
  }, error = function(e) NULL)
  skip_if(is.null(lav), "lavaan::efa() did not return the baseline fit measures")

  df_m <- unname(lav["df"])
  df_0 <- unname(lav["baseline.df"])
  # lavaan's statistics use the multiplier N; CFI and RMSEA need them on the common (N - 1)
  # noncentrality scale.
  chi_m_n1 <- unname(lav["chisq"])          * (N - 1) / N
  chi_0_n1 <- unname(lav["baseline.chisq"]) * (N - 1) / N

  d_m  <- max(chi_m_n1 - df_m, 0)
  d_0  <- max(chi_0_n1 - df_0, 0)
  cfi_ref   <- 1 - d_m / max(d_m, d_0)
  rmsea_ref <- sqrt(max(0, chi_m_n1 - df_m) / (df_m * (N - 1)))

  # observed agreement on the baseline model: ~1e-13
  expect_equal(fi$CFI, cfi_ref, tolerance = 1e-3)
  expect_equal(fi$RMSEA, rmsea_ref, tolerance = 1e-3)
})

# DWLS reproduces lavaan's diagonally weighted least squares loadings on ordinal data.
# The comparison uses the rotation-invariant reproduced off-diagonal correlations and the
# communalities. Agreement is limited by how closely the two polychoric matrices agree
# (~2e-5), so the observed loading agreement is ~3e-5; the tolerance is set generously at
# 1e-3. The reference is estimator = "DWLS" specifically (not ULS/ULSMV).
test_that("DWLS reproduces lavaan::efa(estimator = 'DWLS') on DOSPERT_raw", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), , drop = FALSE]
  k <- 3
  efa <- suppressWarnings(EFA(x, n_factors = k, method = "DWLS",
                              cor_method = "poly", rotation = "none"))
  fit <- suppressWarnings(lavaan::efa(data = as.data.frame(x), nfactors = k,
                                      ordered = colnames(x), estimator = "DWLS",
                                      rotation = "none")[[1]])
  L_ref <- lavaan::lavInspect(fit, "est")$lambda

  expect_equal(repro_offdiag(efa$unrot_loadings), repro_offdiag(L_ref), tolerance = 1e-3)
  expect_equal(unname(efa$h2), unname(diag(tcrossprod(L_ref))), tolerance = 1e-3)
})

test_that("DWLS reproduces lavaan::efa(estimator = 'DWLS') on UPPS_raw", {
  skip_on_cran()
  skip_if_not_slow()
  skip_if_not_installed("lavaan")

  x <- UPPS_raw[stats::complete.cases(UPPS_raw), , drop = FALSE]
  k <- 5
  efa <- suppressWarnings(EFA(x, n_factors = k, method = "DWLS",
                              cor_method = "poly", rotation = "none"))
  fit <- suppressWarnings(lavaan::efa(data = as.data.frame(x), nfactors = k,
                                      ordered = colnames(x), estimator = "DWLS",
                                      rotation = "none")[[1]])
  L_ref <- lavaan::lavInspect(fit, "est")$lambda

  expect_equal(repro_offdiag(efa$unrot_loadings), repro_offdiag(L_ref), tolerance = 1e-3)
  expect_equal(unname(efa$h2), unname(diag(tcrossprod(L_ref))), tolerance = 1e-3)
})

rm(repro_offdiag, reg_fixtures, reg_fit, lav_efa_fit)

EFA_raw <- suppressMessages(EFA(DOSPERT_raw, n_factors = 10, type = "EFAtools",
                                method = "PAF", rotation = "oblimin"))
fac_scores_raw <- FACTOR_SCORES(DOSPERT_raw, f = EFA_raw, method = "Bartlett")

EFA_cor <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
               type = "EFAtools", method = "PAF", rotation = "oblimin")
fac_scores_cor <- suppressMessages(FACTOR_SCORES(test_models$baseline$cormat,
                                                 f = EFA_cor))

EFA_unrot <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                 type = "EFAtools", method = "PAF", rotation = "none")
fac_scores_unrot <- suppressMessages(FACTOR_SCORES(test_models$baseline$cormat,
                                                   f = EFA_unrot))

fac_scores_man <- suppressMessages(FACTOR_SCORES(test_models$baseline$cormat,
                                                 f = unclass(EFA_unrot$unrot_loadings)))


test_that("output is correct", {
  expect_s3_class(fac_scores_raw, "FACTOR_SCORES")
  expect_s3_class(fac_scores_cor, "FACTOR_SCORES")
  expect_s3_class(fac_scores_unrot, "FACTOR_SCORES")
  expect_s3_class(fac_scores_man, "FACTOR_SCORES")

  expect_named(fac_scores_raw, c("scores", "weights", "r.scores", "missing",
                                 "R2", "settings"))
  expect_named(fac_scores_cor, c("scores", "weights", "r.scores",
                                 "R2", "settings"))
  expect_named(fac_scores_unrot, c("scores", "weights", "r.scores", "R2",
                                   "settings"))
  expect_named(fac_scores_man, c("scores", "weights", "r.scores",
                                 "R2", "settings"))
})

test_that("factor scores have the expected shape and recover a known factor", {
  skip_on_cran()

  # raw-data scores: one row per observation, one column per factor (guards against
  # a transposed or weight-matrix mix-up in the returned object)
  expect_equal(dim(fac_scores_raw$scores), c(nrow(DOSPERT_raw), 10L))
  expect_false(anyNA(fac_scores_raw$scores))

  # On a strongly unidimensional scale the single estimated factor score must track
  # the factor it represents; the mean item response is a faithful proxy for that
  # latent factor, so the two correlate almost perfectly.
  efa_1f <- suppressMessages(EFA(GRiPS_raw, n_factors = 1, type = "EFAtools",
                                 method = "PAF"))
  fs_1f <- suppressMessages(FACTOR_SCORES(GRiPS_raw, f = efa_1f, method = "Bartlett"))
  expect_equal(dim(fs_1f$scores), c(nrow(GRiPS_raw), 1L))
  expect_gt(abs(stats::cor(fs_1f$scores[, 1], rowMeans(GRiPS_raw))), 0.95)
})

test_that("settings are returned correctly", {
  expect_named(fac_scores_raw$settings, "method")
  expect_named(fac_scores_cor$settings, "method")
  expect_named(fac_scores_unrot$settings, "method")
  expect_named(fac_scores_man$settings, "method")

  expect_equal(fac_scores_raw$settings$method, "Bartlett")
  expect_equal(fac_scores_cor$settings$method, "Thurstone")
  expect_equal(fac_scores_unrot$settings$method, "Thurstone")
  expect_equal(fac_scores_man$settings$method, "Thurstone")
})


test_that("warnings and errors are thrown correctly", {
  expect_error(FACTOR_SCORES(1:5), class = "efa_input_not_matrix")
  expect_error(FACTOR_SCORES(DOSPERT_raw, f = 1:5), class = "efa_scores_bad_f")
  expect_message(FACTOR_SCORES(test_models$baseline$cormat, f = EFA_cor), class = "efa_scores_needs_raw")
  expect_message(FACTOR_SCORES(test_models$baseline$cormat,
                               f = unclass(EFA_unrot$unrot_loadings)), class = "efa_scores_phi_null")

})


test_that("a non-Pearson EFA with rho = NULL on raw data warns", {
  skip_on_cran()

  # When rho is NULL, psych derives the score machinery from the Pearson cor(x):
  # for regression-based methods the weights/scores, and for every method the
  # score intercorrelations (r.scores), become inconsistent with a non-Pearson
  # fit; warn unless the matching matrix is supplied via rho.
  efa_sp <- suppressMessages(EFA(GRiPS_raw, n_factors = 1, type = "EFAtools",
                                 method = "PAF", cor_method = "spearman"))
  expect_warning(FACTOR_SCORES(GRiPS_raw, f = efa_sp, method = "Thurstone"),
                 class = "efa_scores_cor_method")

  # Supplying the matching correlation via rho silences the warning.
  rho_sp <- stats::cor(GRiPS_raw, method = "spearman")
  expect_no_warning(FACTOR_SCORES(GRiPS_raw, f = efa_sp, rho = rho_sp,
                                  method = "Thurstone"),
                    class = "efa_scores_cor_method")

  # Bartlett weights/scores are correlation-free, but psych still computes
  # r.scores from the Pearson cor(x), so the warning still fires for a
  # non-Pearson fit with rho = NULL.
  expect_warning(FACTOR_SCORES(GRiPS_raw, f = efa_sp, method = "Bartlett"),
                 class = "efa_scores_cor_method")
})

test_that("components scores work on a raw data.frame and Anderson guards single factor", {
  skip_on_cran()

  # psych's components path does `x %*% w` without coercing, so a data.frame raw
  # input must be coerced to a matrix by the wrapper; this previously errored.
  # EFA_raw (file top) is the matching 10-factor DOSPERT fit and is still in scope.
  fs_comp <- FACTOR_SCORES(DOSPERT_raw, f = EFA_raw, method = "components")
  expect_s3_class(fs_comp, "FACTOR_SCORES")
  expect_equal(dim(fs_comp$scores), c(nrow(DOSPERT_raw), 10L))
  expect_false(anyNA(fs_comp$scores))

  # Anderson-Rubin orthogonalisation is undefined for a single factor (and psych
  # aborts opaquely); the wrapper raises a classed error instead.
  efa_1f <- suppressMessages(EFA(GRiPS_raw, n_factors = 1, type = "EFAtools",
                                 method = "PAF"))
  expect_error(FACTOR_SCORES(GRiPS_raw, f = efa_1f, method = "Anderson"),
               class = "efa_scores_anderson_single")
})


rm(EFA_raw, fac_scores_raw, EFA_cor, fac_scores_cor, EFA_unrot, fac_scores_unrot,
   fac_scores_man)

## Use with a lavaan output
if (requireNamespace("lavaan", quietly = TRUE)) {
lav_mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
        F2 =~ V7 + V8 + V9 + V10 + V11 + V12
        F3 =~ V13 + V14 + V15 + V16 + V17 + V18
        g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
             V13 + V14 + V15 + V16 + V17 + V18'
lav_fit <- lavaan::cfa(lav_mod, sample.cov = test_models$baseline$cormat,
                   sample.nobs = 500, estimator = "ml", orthogonal = TRUE)
om_lav <- OMEGA(lav_fit, g_name = "g")

# Second-order model (g loads on the first-order factors): exercises the
# Schmid-Leiman transform of a higher-order solution.
lav_mod_ho <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18
               g =~ F1 + F2 + F3'
lav_fit_ho <- suppressWarnings(lavaan::cfa(lav_mod_ho,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))
om_lav_ho <- suppressMessages(OMEGA(lav_fit_ho, g_name = "g"))
}

## Use with an output from the SL function, with type EFAtools
efa_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
sl_mod <- SL(efa_mod, type = "EFAtools", method = "PAF")
om_sl <- OMEGA(sl_mod, type = "EFAtools",
               factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2)

## Use with an output from the psych::schmid function, with type psych
schmid_mod <- psych::schmid(test_models$baseline$cormat, nfactors = 3,
                            n.obs = 500, fm = "pa", rotate = "Promax")
# Find correlation matrix from phi and pattern matrix from psych::schmid outpu
om_schmid <- OMEGA(schmid_mod, type = "psych")

## Manually specify components
om_man <- OMEGA(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
                factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2)

test_that("output class and dimensions are correct", {
  skip_if_not_installed("lavaan")
  expect_s3_class(om_lav, "OMEGA")
  expect_s3_class(om_sl, "OMEGA")
  expect_s3_class(om_schmid, "OMEGA")
  expect_s3_class(om_man, "OMEGA")

  expect_output(str(om_lav), "List of 2")
  expect_output(str(om_sl), "List of 2")
  expect_output(str(om_schmid), "List of 2")
  expect_output(str(om_man), "List of 2")
})

test_that("a second-order lavaan model is Schmid-Leiman transformed", {
  skip_if_not_installed("lavaan")

  # The second-order general factor triggers the SL transform of the higher-order
  # solution; OMEGA informs the user and returns the full coefficient matrix.
  expect_message(OMEGA(lav_fit_ho, g_name = "g"),
                 class = "efa_omega_g_second_order")
  expect_s3_class(om_lav_ho, "OMEGA")
  expect_identical(dim(om_lav_ho), c(4L, 6L))
  expect_identical(rownames(om_lav_ho), c("g", "F1", "F2", "F3"))
  expect_identical(colnames(om_lav_ho),
                   c("tot", "hier", "sub", "H", "ECV", "PUC"))

  # The transformed group-factor loadings yield finite omega coefficients within
  # [0, 1], and omega hierarchical never exceeds omega total.
  coefs <- unclass(om_lav_ho)[, c("tot", "hier", "sub")]
  expect_true(all(is.finite(coefs)))
  expect_true(all(coefs >= 0 & coefs <= 1))
  expect_true(all(om_lav_ho[, "tot"] >= om_lav_ho[, "hier"]))
})

test_that("OMEGA lavaan numbers are unchanged (regression)", {
  skip_if_not_installed("lavaan")

  # Reference coefficient outputs captured from OMEGA()'s lavaan path, which scores
  # the fitted solution through the shared reliability core. These pin the coefficients
  # (non-Heywood fixtures) to a tight tolerance so any change to that path is caught;
  # the comparison is not bit-exact because the last bits differ across BLAS/platforms.
  ref_bi <- structure(c(0.88301174174790287, 0.74751625677111733, 0.766678791034203,
    0.7685498467792552, 0.76548235222658689, 0.5703940426328894,
    0.50146589411449449, 0.49415766808792488, 0.11752938952131604,
    0.17712221413822785, 0.26521289691970851, 0.2743921786913302,
    0.84884383208407344, 0.37901193583533627, 0.48206335211458812,
    0.47292561296446434, 0.67182974097583426, NA, NA, NA, 0.70588235294117641,
    NA, NA, NA), dim = c(4L, 6L), dimnames = list(c("g", "F1", "F2", "F3"),
    c("tot", "hier", "sub", "H", "ECV", "PUC")), class = "OMEGA")
  expect_equal(om_lav, ref_bi, tolerance = 1e-6)

  ref_ho <- structure(c(0.88158341868254952, 0.74400564775764777, 0.76359153637480304,
    0.76786220097481273, 0.76381778912616816, 0.54868574924161106,
    0.50423868000554961, 0.50487682924077948, 0.11776562955638133,
    0.19531989851603676, 0.25935285636925343, 0.26298537173403319,
    0.84773834907453449, 0.36084418350415226, 0.44810509596935005,
    0.45435071956970075, 0.68379181434818159, NA, NA, NA, 0.70588235294117641,
    NA, NA, NA), dim = c(4L, 6L), dimnames = list(c("g", "F1", "F2", "F3"),
    c("tot", "hier", "sub", "H", "ECV", "PUC")), class = "OMEGA")
  expect_equal(om_lav_ho, ref_ho, tolerance = 1e-6)

  # A single factor returns a named c(Omega, H) vector.
  ref_sf <- structure(c(Omega = 0.86802682659256869, H = 0.86873279471693676),
                      class = "OMEGA")
  fit_sf <- suppressWarnings(lavaan::cfa(
    'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
          V13 + V14 + V15 + V16 + V17 + V18',
    sample.cov = test_models$baseline$cormat, sample.nobs = 500,
    estimator = "ml", orthogonal = TRUE))
  expect_equal(suppressMessages(OMEGA(fit_sf, g_name = "g")), ref_sf, tolerance = 1e-6)

  # Multiple groups return a named list of (unclassed) coefficient matrices.
  ref_mg <- structure(list(A = structure(c(0.88301175860369074, 0.74751650041121109,
    0.7666785723706202, 0.76854962402215088, 0.7654826788236242, 0.57038939781173303,
    0.50147399385562386, 0.49415517812176823, 0.11752907978006645, 0.17712710259947803,
    0.26520457851499635, 0.27439444590038276, 0.84884398691526919, 0.37901393509779763,
    0.48206053742597404, 0.47292785665908216, 0.67183010049739167, NA, NA, NA,
    0.70588235294117641, NA, NA, NA), dim = c(4L, 6L), dimnames = list(
    c("g", "F1", "F2", "F3"), c("tot", "hier", "sub", "H", "ECV", "PUC"))),
    B = structure(c(0.88301177654016794, 0.74751702091877537, 0.76667849734450322,
    0.76854951437076557, 0.76548249348611708, 0.57038915952811098, 0.50147394437097303,
    0.49415464215133981, 0.1175292830540509, 0.17712786139066439, 0.26520455297353013,
    0.27439487221942582, 0.84884395119146006, 0.37901520165768471, 0.48206079699355442,
    0.47292821672600055, 0.6718296371036121, NA, NA, NA, 0.70588235294117641,
    NA, NA, NA), dim = c(4L, 6L), dimnames = list(c("g", "F1", "F2", "F3"),
    c("tot", "hier", "sub", "H", "ECV", "PUC")))), class = "OMEGA")
  fit_mg <- suppressWarnings(lavaan::cfa(lav_mod,
    sample.cov = list(test_models$baseline$cormat, test_models$baseline$cormat),
    sample.nobs = c(500, 500), estimator = "ml", orthogonal = TRUE))
  # Looser than the single-group pins above: these coefficients come out of lavaan's
  # multigroup ML optimizer, which settles on a slightly different point under a
  # different BLAS and moves them by about 1e-6. Drift in the omega arithmetic itself
  # would move them by orders of magnitude more.
  expect_equal(OMEGA(fit_mg, g_name = "g", group_names = c("A", "B")), ref_mg,
               tolerance = 1e-4)
})

test_that("the general-factor row is labelled with g_name, not a hardcoded 'g'", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          gen =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                 V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))
  expect_identical(rownames(OMEGA(fit, g_name = "gen")),
                   c("gen", "F1", "F2", "F3"))
})

test_that("errors are thrown correctly", {
  expect_error(OMEGA(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                     g_load = sl_mod$sl[, "g"], s_load = 1:7,
                     u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
                     factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2), class = "efa_omega_bad_s_load")
  expect_error(OMEGA(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                     g_load = sl_mod$sl[, "g"],
                     s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                     u2 = sl_mod$sl[, "u2"],
                     factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                     pattern = 1:5), class = "efa_omega_bad_pattern")
  expect_warning(OMEGA(sl_mod, type = "EFAtools", g_load = sl_mod$sl[, "g"],
                       s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                       u2 = sl_mod$sl[, "u2"],
                       factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2), class = "efa_omega_model_args_ignored")
  expect_error(OMEGA(model = 1:4), class = "efa_omega_invalid_model")
  expect_error(OMEGA(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                     s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                     u2 = sl_mod$sl[, "u2"],
                     factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2), class = "efa_omega_missing_args")
})

test_that("a user cormat is honoured for a flexible-input SL object", {
  # A flexible-input SL object stores orig_R = NA; OMEGA must use the supplied
  # cormat rather than overwriting it (and must not crash).
  sl_flex <- SL(efa_mod$rot_loadings, Phi = efa_mod$Phi, type = "EFAtools",
                method = "PAF")
  expect_true(identical(sl_flex$orig_R, NA))

  fc <- sl_flex$sl[, c("F1", "F2", "F3")] >= .2
  om_flex <- OMEGA(sl_flex, type = "EFAtools",
                   cormat = test_models$baseline$cormat, factor_corres = fc)
  expect_s3_class(om_flex, "OMEGA")

  # The coefficients match the equivalent model = NULL manual path with the same
  # loadings, uniquenesses, and correlation matrix (the SL object carries factor
  # names, so only the row labels differ).
  om_flex_ref <- OMEGA(model = NULL, type = "EFAtools",
                       var_names = rownames(sl_flex$sl),
                       g_load = sl_flex$sl[, "g"],
                       s_load = sl_flex$sl[, c("F1", "F2", "F3")],
                       u2 = sl_flex$sl[, "u2"],
                       cormat = test_models$baseline$cormat, factor_corres = fc)
  expect_equal(unname(unclass(om_flex)), unname(unclass(om_flex_ref)))
})

test_that("a solution whose variables are keyed against each other is flagged", {
  # The coefficients describe the unit-weighted sum of the variables as supplied, so a
  # scale whose reverse-worded items were never reverse-coded is signalled rather than
  # returned as a merely poor one. Shared with efa_reliability() through the reliability
  # engine, so it reaches this surface too.
  flip <- rep(1, 18)
  flip[7:12] <- -1
  cormat_flipped <- diag(flip) %*% test_models$baseline$cormat %*% diag(flip)
  dimnames(cormat_flipped) <- dimnames(test_models$baseline$cormat)
  efa_flipped <- suppressWarnings(EFA(cormat_flipped, N = 500, n_factors = 3,
                                      type = "EFAtools", method = "PAF",
                                      rotation = "promax"))
  sl_flipped <- suppressWarnings(SL(efa_flipped, type = "EFAtools", method = "PAF"))
  fc <- sl_flipped$sl[, c("F1", "F2", "F3")] >= .2

  expect_warning(OMEGA(sl_flipped, type = "EFAtools", factor_corres = fc),
                 class = "efa_reliability_mixed_keying")
  expect_no_warning(OMEGA(sl_mod, type = "EFAtools",
                          factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2),
                    class = "efa_reliability_mixed_keying")
})

test_that("print output is stable", {
  local_reproducible_output()

  # single group, full coefficient matrix (general factor plus group factors)
  expect_snapshot(print(om_sl), transform = scrub_num)

  # single group, single factor
  expect_snapshot(print(structure(0.85, class = "OMEGA")), transform = scrub_num)
  expect_snapshot(print(structure(c(0.85, 2.10), class = "OMEGA")), transform = scrub_num)

  # multiple groups, full coefficient matrix (long header with ECV and PUC)
  om_mg6 <- structure(list(GroupA = unclass(om_sl), GroupB = unclass(om_sl)),
                      class = "OMEGA")
  expect_snapshot(print(om_mg6), transform = scrub_num)

  # multiple groups, three-column matrix (omega total, hierarchical, subscale)
  om_mg3 <- structure(list(GroupA = unclass(om_sl)[, 1:3],
                           GroupB = unclass(om_sl)[, 1:3]), class = "OMEGA")
  expect_snapshot(print(om_mg3), transform = scrub_num)

  # multiple groups, single factor
  expect_snapshot(print(structure(list(GroupA = 0.85, GroupB = 0.80), class = "OMEGA")),
                  transform = scrub_num)
  expect_snapshot(print(structure(list(GroupA = c(0.85, 2.10), GroupB = c(0.80, 1.90)),
                                  class = "OMEGA")), transform = scrub_num)
})

rm(efa_mod, sl_mod, om_sl, schmid_mod, om_schmid, om_man)
if (requireNamespace("lavaan", quietly = TRUE)) {
  rm(lav_mod, lav_fit, om_lav, lav_mod_ho, lav_fit_ho, om_lav_ho)
}

# Schema contract for the top-level inference slots on EFA: SE, CI, replicates.
# Pins which keys live in each slot under each (se method x rotation) combination,
# and the absence of accidental dual storage under the old `boot.` names.

dat <- GRiPS_raw[, 1:6]
R   <- stats::cor(dat)
N   <- nrow(dat)

# Assert the slot contract for one fit. Old `boot.` names must be absent
# everywhere, the expected SE/CI keys must be present, and `replicates` is a
# present-but-NULL slot under analytic SE methods (a list under "np-boot").
expect_field_contract <- function(fit, expected_SE, expected_CI,
                                  expected_replicates, info = "") {
  expect_false(any(c("boot.SE", "boot.CI", "boot.arrays", "boot.MI")
                   %in% names(fit)), info = info)
  expect_setequal(names(fit$SE), expected_SE)
  expect_setequal(names(fit$CI), expected_CI)
  if (is.null(expected_replicates)) {
    expect_true("replicates" %in% names(fit), info = info)
    expect_null(fit$replicates, info = info)
  } else {
    expect_true(is.list(fit$replicates), info = info)
    expect_setequal(names(fit$replicates), expected_replicates)
  }
}


test_that("se = 'none' leaves no SE/CI/replicates/MI slot on the fit", {
  fit <- EFA(R, n_factors = 2, N = N, method = "ML", rotation = "none",
             se = "none")
  expect_false(any(c("SE", "CI", "replicates", "MI",
                     "boot.SE", "boot.CI", "boot.arrays", "boot.MI")
                   %in% names(fit)))
})


test_that("information SEs fill the schema across all rotation families", {
  # Unrotated: unrot_loadings + uniquenesses; no rotated-loading components.
  fit_none <- EFA(R, n_factors = 2, N = N, method = "ML", rotation = "none",
                  se = "information")
  expect_field_contract(
    fit_none,
    expected_SE = c("unrot_loadings", "uniquenesses", "communalities"),
    expected_CI = c("unrot_loadings", "uniquenesses", "communalities"),
    expected_replicates = NULL,
    info = "information / none"
  )

  # Orthogonal: adds rot_loadings + communalities; no Phi/Structure.
  fit_orth <- EFA(R, n_factors = 2, N = N, method = "ML", rotation = "varimax",
                  se = "information")
  expect_field_contract(
    fit_orth,
    expected_SE = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities"),
    expected_CI = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities"),
    expected_replicates = NULL,
    info = "information / varimax"
  )

  # Oblique: adds Phi and Structure on top of the orthogonal schema.
  fit_oblq <- EFA(R, n_factors = 2, N = N, method = "ML", rotation = "oblimin",
                  se = "information")
  expect_field_contract(
    fit_oblq,
    expected_SE = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities", "Phi", "Structure"),
    expected_CI = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities", "Phi", "Structure"),
    expected_replicates = NULL,
    info = "information / oblimin"
  )
})


test_that("sandwich SEs fill the same schema as information across rotations", {
  skip_on_cran()
  # The Browne (1984) ADF Gamma for Pearson correlations needs the raw data, so
  # these fits use `dat` instead of the precomputed cormat.
  fit_none <- EFA(dat, n_factors = 2, method = "ML", rotation = "none",
                  se = "sandwich")
  expect_field_contract(
    fit_none,
    expected_SE = c("unrot_loadings", "uniquenesses", "communalities"),
    expected_CI = c("unrot_loadings", "uniquenesses", "communalities"),
    expected_replicates = NULL,
    info = "sandwich / none"
  )

  fit_orth <- EFA(dat, n_factors = 2, method = "ML", rotation = "varimax",
                  se = "sandwich")
  expect_field_contract(
    fit_orth,
    expected_SE = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities"),
    expected_CI = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities"),
    expected_replicates = NULL,
    info = "sandwich / varimax"
  )

  fit_oblq <- EFA(dat, n_factors = 2, method = "ML", rotation = "oblimin",
                  se = "sandwich")
  expect_field_contract(
    fit_oblq,
    expected_SE = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities", "Phi", "Structure"),
    expected_CI = c("unrot_loadings", "uniquenesses",
                    "rot_loadings", "communalities", "Phi", "Structure"),
    expected_replicates = NULL,
    info = "sandwich / oblimin"
  )
})


test_that("np-boot fills `replicates` and adds fit_indices/residuals SEs", {
  skip_on_cran()
  # Small `b` -- this test pins schema/slot names; the replicate count enters only
  # via dim(...)[3L] == b, not via any numeric agreement.
  b <- 4L

  # Unrotated: SE/CI cover the unrotated loadings, fit indices, and residuals;
  # `replicates` is a list of three 3D cubes whose last dimension is b.
  set.seed(42)
  fit_none <- suppressWarnings(suppressMessages(
    EFA(dat, n_factors = 2, method = "ML", rotation = "none",
        se = "np-boot", b_boot = b)
  ))
  expect_field_contract(
    fit_none,
    expected_SE = c("unrot_loadings", "fit_indices", "residuals",
                    "valid_replicates"),
    expected_CI = c("unrot_loadings", "fit_indices", "residuals"),
    expected_replicates = c("unrot_loadings", "fit_indices", "residuals"),
    info = "np-boot / none"
  )
  expect_identical(dim(fit_none$replicates$unrot_loadings)[3L], b)

  # Orthogonal: adds rot_loadings; SE also reports valid_replicates and
  # valid_target_rotations (the CI list does not). `replicates` mirrors SE minus
  # those two integer counters.
  set.seed(42)
  fit_orth <- suppressWarnings(suppressMessages(
    EFA(dat, n_factors = 2, method = "ML", rotation = "varimax",
        se = "np-boot", b_boot = b)
  ))
  expect_field_contract(
    fit_orth,
    expected_SE = c("unrot_loadings", "rot_loadings", "fit_indices", "residuals",
                    "valid_replicates", "valid_target_rotations"),
    expected_CI = c("unrot_loadings", "rot_loadings",
                    "fit_indices", "residuals"),
    expected_replicates = c("unrot_loadings", "rot_loadings",
                            "fit_indices", "residuals"),
    info = "np-boot / varimax"
  )
  expect_identical(dim(fit_orth$replicates$rot_loadings)[3L], b)

  # Oblique: adds Phi and Structure on top of the orthogonal schema.
  set.seed(42)
  fit_oblq <- suppressWarnings(suppressMessages(
    EFA(dat, n_factors = 2, method = "ML", rotation = "oblimin",
        se = "np-boot", b_boot = b)
  ))
  expect_field_contract(
    fit_oblq,
    expected_SE = c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                    "fit_indices", "residuals", "valid_replicates",
                    "valid_target_rotations"),
    expected_CI = c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                    "fit_indices", "residuals"),
    expected_replicates = c("unrot_loadings", "rot_loadings", "Phi",
                            "Structure", "fit_indices", "residuals"),
    info = "np-boot / oblimin"
  )
  expect_identical(dim(fit_oblq$replicates$Phi)[3L], b)
})

unrot <- EFA(test_models$baseline$cormat, 3, N = 500)
prom <- .rotate_model(unrot, rotation = "promax", type = "EFAtools")
prom_psych <- .rotate_model(unrot, rotation = "promax", type = "psych")
prom_spss <- .rotate_model(unrot, rotation = "promax", type = "SPSS")

unrot_1 <- EFA(test_models$baseline$cormat, 1, N = 500)
prom_1 <- suppressWarnings(.rotate_model(unrot_1, rotation = "promax", type = "EFAtools"))

test_that("output class and dimensions are correct", {
  expect_type(prom, "list")
  expect_type(prom_1, "list")
  expect_named(prom, c("rot_loadings", "Phi", "Structure", "rotmat",
                       "vars_accounted_rot", "settings"))
  expect_named(prom_1, c("rot_loadings", "Phi", "Structure", "rotmat",
                       "vars_accounted_rot", "settings"))

  expect_s3_class(prom$rot_loadings, "LOADINGS")
  checkmate::expect_matrix(prom$Phi)
  expect_s3_class(prom$Structure, "LOADINGS")
  checkmate::expect_matrix(prom$rotmat)
  checkmate::expect_matrix(prom$vars_accounted_rot)
  expect_type(prom$settings, "list")

  expect_s3_class(prom_1$rot_loadings, "LOADINGS")
  expect_null(prom_1$Phi)
  expect_null(prom_1$Structure)
  expect_equal(prom_1$rotmat, NA)
  expect_null(prom_1$vars_accounted_rot)
  expect_type(prom_1$settings, "list")
})

test_that("settings are returned correctly", {
  expect_named(prom$settings, c("normalize", "P_type", "precision",
                                "order_type", "varimax_type", "k"))
  expect_named(prom_psych$settings, c("normalize", "P_type", "precision",
                                      "order_type", "varimax_type", "k"))
  expect_named(prom_spss$settings, c("normalize", "P_type", "precision",
                                     "order_type", "varimax_type", "k"))
  expect_named(prom_1$settings, c("normalize", "P_type", "precision",
                                  "order_type", "varimax_type", "k"))

  expect_equal(prom$settings$normalize, TRUE)
  expect_equal(prom_psych$settings$normalize, TRUE)
  expect_equal(prom_spss$settings$normalize, TRUE)
  expect_equal(prom_1$settings$normalize, TRUE)

  expect_equal(prom$settings$P_type, "norm")
  expect_equal(prom_psych$settings$P_type, "unnorm")
  expect_equal(prom_spss$settings$P_type, "norm")
  expect_equal(prom_1$settings$P_type, "norm")

  expect_equal(prom$settings$precision, 1e-05)
  expect_equal(prom_psych$settings$precision, 1e-05)
  expect_equal(prom_spss$settings$precision, 1e-05)
  expect_equal(prom_1$settings$precision, 1e-05)

  expect_equal(prom$settings$order_type, "eigen")
  expect_equal(prom_psych$settings$order_type, "eigen")
  expect_equal(prom_spss$settings$order_type, "ss_factors")
  expect_equal(prom_1$settings$order_type, "eigen")

  expect_equal(prom$settings$varimax_type, "kaiser")
  expect_equal(prom_psych$settings$varimax_type, "svd")
  expect_equal(prom_spss$settings$varimax_type, "kaiser")
  expect_equal(prom_1$settings$varimax_type, "kaiser")

  expect_equal(prom$settings$k, 4)
  expect_equal(prom_psych$settings$k, 4)
  expect_equal(prom_spss$settings$k, 4)
  expect_equal(prom_1$settings$k, 4)

})

test_that("errors etc. are thrown correctly", {

  expect_error(.rotate_model(unrot, rotation = "promax", type = "none"), class = "efa_type_none")

  expect_warning(.rotate_model(unrot, rotation = "promax", type = "EFAtools", normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "EFAtools", P_type = "norm"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "EFAtools", order_type = "ss_factors"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "EFAtools", k = 2), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "promax", type = "psych", normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "psych", P_type = "norm"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "psych", order_type = "ss_factors"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "psych", k = 2), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "promax", type = "SPSS", normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "SPSS", P_type = "unnorm"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "SPSS", order_type = "eigen"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "SPSS", k = 2), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "promax", type = "SPSS", varimax_type = "svd"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot_1, rotation = "promax", type = "EFAtools"), class = "efa_single_factor")
})

test_that("promax rotation matrix reproduces the rotated loadings", {
  # the rotation matrix must carry the sign reflection and factor reordering so that
  # L_unrot %*% rotmat == rot_loadings, for both ordering branches (eigen reorders the
  # pattern after the fit; ss_factors reorders the varimax base before it)
  L <- unclass(unrot$unrot_loadings)
  expect_equal(L %*% prom$rotmat, unclass(prom$rot_loadings),       # order_type "eigen"
               ignore_attr = TRUE, tolerance = 1e-6)
  expect_equal(L %*% prom_spss$rotmat, unclass(prom_spss$rot_loadings),  # "ss_factors"
               ignore_attr = TRUE, tolerance = 1e-6)
})

rm(unrot, prom, unrot_1, prom_1, prom_psych, prom_spss)

# Two-step polychoric / tetrachoric correlation matrix from raw ordinal data. Recodes
# each variable to consecutive 0-based categories, then defers the numerics to the C++
# backend (.polychoric_cpp); this wrapper owns the input validation and the classed
# conditions, which can only be raised at the R level.
.polychoric <- function(x, nearest_pd = FALSE,
                        binary_only = FALSE, acov = c("none", "diag", "full"),
                        label_acov = TRUE, error_call = rlang::caller_env()) {

  acov <- match.arg(acov)

  # data.matrix() maps a factor to its LEVEL INDEX, so the level order silently becomes the
  # response order. That is correct for an ordered factor (and for numeric codes, which pass
  # through untouched), but an unordered factor built over character labels is ordered
  # alphabetically -- "always" < "never" < "often" < "rarely" -- which scrambles the responses
  # and attenuates the correlations badly (a true rho of 0.69 estimates as 0.33) without
  # producing anything a user could notice: the result is still a clean, unit-diagonal matrix.
  # Character columns take the same alphabetical route. Ordering is load-bearing here, so
  # refuse to guess it rather than accept the input on a silently arbitrary order.
  if (is.data.frame(x)) {
    bad <- vapply(x, function(col) is.character(col) ||
                    (is.factor(col) && !is.ordered(col)), logical(1))
    if (any(bad)) {
      nms_bad <- names(x)[bad]
      cli::cli_abort(
        c("Polychoric correlations need an explicit response order, but {cli::qty(nms_bad)} column{?s} {.val {nms_bad}} {?is/are} unordered.",
          "x" = "An unordered factor or character column would be ranked by its level order, which is alphabetical unless you set it.",
          "i" = "Supply numeric response codes, or convert with {.fun ordered} using the correct level order."),
        class = "efa_cor_unordered_factor", call = error_call)
    }
  } else if (is.character(x)) {
    cli::cli_abort(
      c("Polychoric correlations need numeric response codes, but {.arg x} is a character matrix.",
        "i" = "Supply numeric response codes, or convert the columns with {.fun ordered} using the correct level order."),
      class = "efa_cor_unordered_factor", call = error_call)
  }

  # Ordered-factor columns of a data frame become their level indices; numeric data passes
  # through. The recoding below maps each column onto its own 0-based categories.
  x <- data.matrix(x)
  if (is.null(colnames(x))) {
    colnames(x) <- paste0("V", seq_len(ncol(x)))
  }
  nms <- colnames(x)
  p <- ncol(x)

  # A single variable has no correlation to estimate. Caught here so the request fails with a
  # classed condition; the C++ backend keeps the same check as unreachable defence.
  if (p < 2L) {
    cli::cli_abort(
      c("Polychoric correlations need at least two variables.",
        "x" = "{.arg x} has {p} column{?s}."),
      class = "efa_cor_too_few_vars", call = error_call)
  }

  # When an asymptotic covariance is requested, the point estimate, thresholds, and ACOV must
  # all come from the SAME cases: a sandwich covariance is only valid for the estimator that
  # produced the estimates, so a pairwise matrix paired with a listwise covariance is not the
  # covariance of any estimator (lavaan likewise computes categorical standard errors on a
  # single case set). Reduce to the listwise-complete rows up front, so the recode below also
  # collapses any category that survives only in incomplete rows. Without an ACOV the matrix
  # stays pairwise-complete for data efficiency (exploratory factor analysis needs no standard
  # errors from the matrix itself).
  if (acov != "none") {
    complete <- stats::complete.cases(x)
    if (sum(complete) < 2L) {
      cli::cli_abort(
        c("A polychoric asymptotic covariance needs at least two listwise-complete observations.",
          "x" = "{sum(complete)} row{?s} {?is/are} complete across all variables.",
          "i" = "Supply complete data, or impute the missing values, before requesting an {.arg acov}."),
        class = "efa_cor_no_complete_cases", call = error_call)
    }
    x <- x[complete, , drop = FALSE]
  }

  codes <- matrix(NA_integer_, nrow(x), p)
  n_cat <- integer(p)
  for (j in seq_len(p)) {
    col <- x[, j]
    lv <- sort(unique(col[!is.na(col)]))
    n_cat[j] <- length(lv)
    codes[, j] <- match(col, lv) - 1L
  }
  colnames(codes) <- nms

  # A constant (or empty) column has no thresholds and no information about any
  # correlation; abort rather than returning a degenerate matrix.
  if (any(n_cat < 2L)) {
    bad <- nms[n_cat < 2L]
    cli::cli_abort(
      c("Polychoric correlations need at least two response categories per variable.",
        "x" = "{cli::qty(bad)}Variable{?s} {.val {bad}} {?is/are} constant."),
      class = "efa_cor_constant_col", call = error_call)
  }

  # Tetrachoric correlations are the binary special case; reject any variable with
  # more than two categories so the request is not silently widened to polychoric.
  if (isTRUE(binary_only) && any(n_cat > 2L)) {
    bad <- nms[n_cat > 2L]
    cli::cli_abort(
      c("Tetrachoric correlations require binary variables (at most two categories).",
        "x" = "{cli::qty(bad)}Variable{?s} {.val {bad}} {?has/have} more than two categories."),
      class = "efa_cor_not_binary", call = error_call)
  }

  # Many categories usually means the variable is continuous, not ordinal; warn (rather than
  # abort, so a genuinely ordinal many-category variable still works) but proceed.
  if (any(n_cat >= 10L)) {
    many <- nms[n_cat >= 10L]
    cli::cli_warn(
      c("{cli::qty(many)} Variable{?s} {.val {many}} {?has/have} 10 or more response categories.",
        "i" = "Polychoric correlations assume a few ordered categories; check that {cli::qty(many)} {?this variable is/these variables are} not continuous.",
        "i" = "Each pair is estimated on a contingency table with one cell per pair of
               observed values, so the work per pair grows with the square of the number of
               categories: a continuous variable, whose every distinct value is a category,
               is far slower than an ordinal one."),
      class = "efa_cor_many_categories")
  }

  res <- .polychoric_cpp(codes, acov, isTRUE(nearest_pd))

  R <- res$R
  dimnames(R) <- list(nms, nms)

  # A pair with no overlapping complete observations is uncomputable and comes back as
  # NA; abort with the same classed condition .prepare_cor_input() uses for an NA
  # correlation rather than returning a matrix EFA cannot use.
  if (anyNA(R)) {
    bad <- which(is.na(R) & upper.tri(R), arr.ind = TRUE)
    pairs <- paste0(nms[bad[, 1L]], "-", nms[bad[, 2L]])
    cli::cli_abort(
      c("The polychoric correlation could not be computed for {cli::qty(pairs)} variable pair{?s} {.val {pairs}}.",
        "i" = "{cli::qty(pairs)}{?This pair has/These pairs have} no overlapping complete observations."),
      class = "efa_cor_na", call = error_call)
  }

  # Surface the nearest-PD projection as the same classed warning .prepare_cor_input()
  # uses for smoothing, so downstream handling is consistent.
  if (isTRUE(res$pd_adjusted)) {
    cli::cli_warn(
      c("The polychoric correlation matrix was not positive definite; it has been projected to the nearest positive definite matrix.",
        "i" = "Inspect the results carefully."),
      class = "efa_cor_smoothed")
  }

  # Pair labels are needed only by the diagnostics and by the labelled covariance, both of which
  # are skipped for the per-replicate bootstrap recompute, so they are not built for it either.
  labels <- if (label_acov) .pair_labels(nms) else NULL

  # A binary pair whose table has a single empty cell is at a Frechet bound of its own marginals,
  # but it is repaired rather than reported: the empty cell is nudged by a continuity correction so
  # the pair can be estimated in the interior (see poly_zero_correct_2x2 in src/polychoric.cpp).
  # Name those pairs anyway -- the estimate is not a plain function of the observed counts, and the
  # correction is the reason the pair has no asymptotic variance -- but keep it a separate, milder
  # condition from the boundary report below, because these pairs DO get a usable correlation. The
  # 0.5 below states the POLY_ZERO_ADD constant the backend actually applies; the two must be
  # changed together. `label_acov = FALSE` keeps the per-replicate bootstrap recompute silent, as
  # for the boundary and sparse-cell diagnostics.
  if (label_acov && any(res$zero_corrected)) {
    n_zc <- sum(res$zero_corrected)
    cap <- .cap_label_list(labels[res$zero_corrected])
    cli::cli_warn(
      c("{n_zc} binary variable pair{?s} {?has/have} a response combination that never occurs, so a continuity correction of {.val {0.5}} was applied before estimating {cli::qty(n_zc)}{?its/their} correlation.",
        "x" = "Affected {cli::qty(n_zc)}pair{?s}: {.val {cap$shown}}{cap$rest}.",
        "i" = "Every two-by-two table with an empty cell is reproduced exactly by a correlation of 1 (or -1), so without the correction the estimate would be the boundary value whatever the underlying correlation. The correction adds {.val {0.5}} to the empty cell while preserving the table margins, as {.pkg lavaan} and {.pkg psych} do by default.",
        "i" = "The corrected correlation is a point estimate only: no asymptotic variance or standard error is available for {cli::qty(n_zc)} {?this pair/these pairs}."),
      class = "efa_cor_zero_cell")
  }

  # A pair whose response table is one of the two Frechet bounds of its own marginals (perfectly
  # ordered, or perfectly reversed, with no discordant observations) and that the correction above
  # does not cover is reproduced exactly by the model at rho = +/-1: the estimate is the boundary of
  # the parameter space, reported as the largest correlation the estimator admits. The likelihood is
  # numerically flat over the whole approach to that boundary, so this is the only value that is
  # reproducible -- an optimiser left to search there stops at an arbitrary point. Report it,
  # because the pair's estimate is a bound rather than an interior maximum, and because it usually
  # means two items are redundant.
  if (label_acov && any(res$at_bound)) {
    n_bnd <- sum(res$at_bound)
    cap <- .cap_label_list(labels[res$at_bound])
    # A table at the LOWER bound is estimated at the negative endpoint, so the message quotes the
    # values actually reported (both, when the two directions occur together) rather than naming the
    # positive one for every pair. They come from the backend's own boundary estimates: the returned
    # matrix cannot be used because `nearest_pd` may have projected the entry away from the
    # endpoint. Sorted decreasing so a mixed report reads "0.9999 and -0.9999".
    bnd_vals <- sort(unique(res$bound_rho[res$at_bound]), decreasing = TRUE)
    cli::cli_warn(
      c("{n_bnd} variable pair{?s} {?has/have} a perfectly monotone response table, so {?its/their} correlation is estimated at the boundary {cli::qty(length(bnd_vals))}value{?s} {.val {bnd_vals}}.",
        "x" = "Affected {cli::qty(n_bnd)}pair{?s}: {.val {cap$shown}}{cap$rest}.",
        "i" = "The two variables are never observed out of order -- either they always vary together, or (for a negative estimate) they always vary in opposite directions -- which the model can reproduce only with a correlation of 1 or -1, so the data put no limit on how close to that boundary the correlation lies and no asymptotic variance or standard error is available.",
        "i" = "Consider collapsing rare response categories in the variables involved, or dropping one item of a redundant pair."),
      class = "efa_cor_boundary")
  }

  out <- list(R = R)

  # The asymptotic covariance of the off-diagonal correlations (Muthen, 1984; Joreskog,
  # 1994), at the requested level: a variance per element ("diag") or the full cross-pair
  # matrix ("full"), on the variance scale (Var(rho-hat)). It is the covariance of the
  # (un-projected) two-step estimates computed on the listwise-complete rows; `nearest_pd`
  # only smooths the returned R, it does not change the covariance. Pairs follow column order
  # of the upper triangle (i < j), so the labels match utils::combn(nms, 2). A pair at a Frechet
  # bound, and a binary pair whose empty cell was continuity-corrected, both have no asymptotic
  # variance and come back as NA (their whole row and column at the "full" level); `at_bound` and
  # `zero_corrected` name those pairs.
  if (acov != "none") {
    av <- res$acov
    # The point estimate names the asymptotic covariance by variable pair and runs the
    # structural-sparsity diagnostic below; the per-replicate bootstrap recompute consumes the
    # covariance positionally and skips both (label_acov = FALSE), so neither is repeated per
    # replicate.
    if (label_acov) {
      # The two-step asymptotic covariance (and the DWLS weights / robust standard errors derived
      # from it) rely on large-sample theory that degrades for structurally empty contingency
      # cells, where the analytic variances can be far too small. Warn only when a cell is empty
      # despite a non-negligible expected count under independence (a structural rather than
      # incidental zero, by the usual >= 5 contingency-table rule); a rare corner combination in a
      # many-category table is not flagged. The covariance is still returned.
      n_obs <- nrow(codes)
      pairs_ij <- utils::combn(p, 2L)
      # Every pair is tabulated (rather than stopping at the first sparse one) so the warning
      # can name the variables the user has to act on. Each table is a single tabulate() and
      # this runs once per fit, not per bootstrap replicate (label_acov = FALSE skips it).
      # Only the pair indices are collected here; .pair_labels() below supplies the names, so
      # the warning and the covariance's dimnames cannot drift apart.
      is_sparse <- logical(ncol(pairs_ij))
      for (k in seq_len(ncol(pairs_ij))) {
        i <- pairs_ij[1L, k]; j <- pairs_ij[2L, k]
        tab <- matrix(tabulate(codes[, i] * n_cat[j] + codes[, j] + 1L,
                               nbins = n_cat[i] * n_cat[j]),
                      n_cat[i], n_cat[j], byrow = TRUE)
        expected <- outer(rowSums(tab), colSums(tab)) / n_obs
        is_sparse[k] <- any(tab == 0L & expected >= 5)
      }
      # A boundary or continuity-corrected pair is reported by the warnings above, which say
      # strictly more: its covariance is not "unreliable" but absent. Reporting the same pair
      # twice, the second time with a weaker diagnosis, would only obscure that -- so drop those
      # pairs here. A pair that is sparse without being either is unaffected and still warns
      # exactly as before.
      is_sparse <- is_sparse & !res$at_bound & !res$zero_corrected

      if (any(is_sparse)) {
        n_sparse <- sum(is_sparse)
        cap <- .cap_label_list(labels[is_sparse])
        cli::cli_warn(
          c("{n_sparse} variable pair{?s} {?has/have} an empty response-category combination despite a non-negligible expected count.",
            "x" = "Affected {cli::qty(n_sparse)}pair{?s}: {.val {cap$shown}}{cap$rest}.",
            "i" = "The polychoric asymptotic covariance (and any DWLS weights or robust standard errors derived from it) can be unreliable for such structurally sparse cells; interpret them with caution and consider collapsing rare response categories in these variables."),
          class = "efa_cor_sparse_cells")
      }

      if (acov == "diag") {
        names(av) <- labels
      } else {
        dimnames(av) <- list(labels, labels)
      }
    }
    out$acov <- av
  }

  # Which pairs are at a bound, and which had a binary empty cell continuity-corrected, in the same
  # pair order, so a consumer can name them without re-deriving the structural test. Returned for
  # every call, not only when an asymptotic covariance was requested: the flags explain the estimate
  # as much as the missing variance, and the warnings above name those pairs on the plain matrix
  # path too.
  out$at_bound <- if (label_acov) stats::setNames(res$at_bound, labels) else res$at_bound
  out$zero_corrected <- if (label_acov) {
    stats::setNames(res$zero_corrected, labels)
  } else {
    res$zero_corrected
  }

  out
}

#' Scree plot
#'
#' The scree plot was originally introduced by Cattell (1966) to perform the
#' scree test. In a scree plot, the eigenvalues of the factors / components are
#' plotted against the index of the factors / components, ordered from 1 to N
#' factors components, hence from largest to smallest eigenvalue. According to
#' the scree test, the number of factors / components to retain is the number of
#' factors / components to the left of the "elbow" (where the curve starts to
#' level off) in the scree plot.
#'
#' @inheritParams efa_kgc
#'
#' @details As the scree test requires visual examination, the test has been
#' especially criticized for its subjectivity and with this low inter-rater
#' reliability. Moreover, a scree plot can be ambiguous if there are either no
#' clear "elbow" or multiple "elbows", making it difficult to judge just where
#' the eigenvalues do level off. Finally, the scree test has also been found to
#' be less accurate than other factor retention criteria. For all these reasons,
#' the scree test has been recommended against, at least for exclusive use as a
#' factor retention criterion (Zwick & Velicer, 1986)
#'
#' The `efa_scree` function can also be called together with other factor
#' retention criteria in the [efa_retain()] function.
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). The scree plot is a
#'   visual criterion, so it returns no numeric suggestion. Its main fields are:
#' \item{results}{A list with one record per requested eigenvalue type, each
#'   holding the eigenvalues used for the scree plot.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Cattell, R. B. (1966). The scree test for the number of factors.
#' Multivariate Behavioral Research, 1(2), 245–276.
#' https://doi.org/10.1207/s15327906mbr0102_10
#'
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. Psychological Bulletin, 99,
#' 432–442. https://doi.org/10.1037/0033-2909.99.3.432
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#' @examples
#' efa_scree(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
efa_scree <- function(x, eigen_type = c("PCA", "SMC", "EFA"),
                  use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                        "everything", "na.or.complete"),
                  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                  n_factors = 1, estimate_control = NULL, ...){

  # Perform argument checks
  .reject_flat_knobs(...names(), fn = "efa_scree")
  .reject_unknown_fit_dots(...names(), fn = "efa_scree", unrotated = TRUE)
  .reject_rotation_dots(list(...), fn = "efa_scree")
  .assert_cor_input(x)

  eigen_type <- .match_arg_ci(eigen_type, several.ok = TRUE)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  checkmate::assert_count(n_factors)
  .assert_estimate_control(estimate_control)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, use = use, cor_method = cor_method,
                             N_policy = "none")
  R <- prep$R


  # Calculate the PCA / SMC / EFA eigenvalues for the requested types
  eigen_list <- .three_eigen(R, eigen_type, n_factors = n_factors,
                             estimate_control = estimate_control, ...)

  # one eigenvalue record per requested type; the scree plot is purely visual, so
  # there is no numeric suggestion (n_factors is NA)
  results <- list()
  for (et in c("PCA", "SMC", "EFA")) {
    if (!(et %in% eigen_type)) next
    eig <- eigen_list[[et]]
    results[[et]] <- list(
      name = et,
      label = et,
      n_factors = NA_real_,
      plot_type = "eigen",
      x = seq_along(eig),
      y = eig
    )
  }

  output <- .new_efa_retention(
    "SCREE",
    results = unname(results),
    settings = list(eigen_type = eigen_type, use = use,
                    cor_method = cor_method, n_factors = n_factors),
    subtitle = .eigen_subtitle(eigen_type),
    note = "Scree plot is a visual criterion; call plot(x) to identify the elbow."
  )

  return(output)

}

scree_cor <- efa_scree(test_models$baseline$cormat)
scree_cor_smc <- efa_scree(test_models$baseline$cormat, eigen_type = "SMC")
scree_raw <- efa_scree(GRiPS_raw)
# Check with an argument passed to "EFA"
scree_efa_ml<- efa_scree(test_models$baseline$cormat, eigen_type = "EFA", estimator = "ML")

test_that("output class and dimensions are correct", {
  expect_s3_class(scree_cor, "efa_retention")
  expect_length(scree_cor, 6)
  expect_s3_class(scree_cor_smc, "efa_retention")
  expect_length(scree_cor_smc, 6)
  expect_s3_class(scree_raw, "efa_retention")
  expect_s3_class(scree_efa_ml, "efa_retention")

  expect_named(scree_cor$n_factors, c("PCA", "SMC", "EFA"))
  expect_named(scree_cor_smc$n_factors, "SMC")
  expect_named(scree_efa_ml$n_factors, "EFA")
  # the scree plot is purely visual: no numeric suggestion
  expect_true(all(is.na(scree_cor$n_factors)))
  expect_equal(.retention_record(scree_cor, "PCA")$plot_type, "eigen")
})

test_that("found eigenvalues are correct", {
  expect_equal(sum(.retention_record(scree_cor, "PCA")$y),
               ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(scree_cor, "SMC")$y),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(scree_cor, "EFA")$y),
            ncol(test_models$baseline$cormat))

  expect_equal(sum(.retention_record(scree_raw, "PCA")$y), ncol(GRiPS_raw))
  expect_lt(sum(.retention_record(scree_raw, "SMC")$y), ncol(GRiPS_raw))
  expect_lt(sum(.retention_record(scree_raw, "EFA")$y), ncol(GRiPS_raw))

  # Only the requested eigenvalue type produces a record
  expect_null(.retention_record(scree_cor_smc, "PCA"))
  expect_null(.retention_record(scree_cor_smc, "EFA"))
  expect_null(.retention_record(scree_efa_ml, "PCA"))
  expect_null(.retention_record(scree_efa_ml, "SMC"))
})


test_that("errors are thrown correctly", {
  expect_error(efa_scree(1:5), class = "efa_input_not_matrix")
  expect_message(efa_scree(GRiPS_raw, eigen_type = "PCA"), class = "efa_cor_from_data")
  expect_error(efa_scree(sing_raw), class = "efa_cor_singular")
  expect_error(efa_scree(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_warning(efa_scree(cor_nposdef, N = 10), class = "efa_cor_smoothed")
})

test_that("settings are returned correctly", {
  expect_named(scree_cor$settings, c("eigen_type", "use", "cor_method",
                                     "n_factors"))
  expect_named(scree_raw$settings, c("eigen_type", "use", "cor_method",
                                     "n_factors"))
  expect_named(scree_cor_smc$settings, c("eigen_type", "use", "cor_method",
                                         "n_factors"))
  expect_named(scree_efa_ml$settings, c("eigen_type", "use", "cor_method",
                                        "n_factors"))

  expect_equal(scree_cor$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(scree_raw$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(scree_cor_smc$settings$eigen_type, "SMC")
  expect_equal(scree_efa_ml$settings$eigen_type, "EFA")

  expect_equal(scree_cor$settings$use, "pairwise.complete.obs")
  expect_equal(scree_raw$settings$use, "pairwise.complete.obs")
  expect_equal(scree_cor_smc$settings$use, "pairwise.complete.obs")
  expect_equal(scree_efa_ml$settings$use, "pairwise.complete.obs")

  expect_equal(scree_cor$settings$cor_method, "pearson")
  expect_equal(scree_raw$settings$cor_method, "pearson")
  expect_equal(scree_cor_smc$settings$cor_method, "pearson")
  expect_equal(scree_efa_ml$settings$cor_method, "pearson")

  expect_equal(scree_cor$settings$n_factors, 1)
  expect_equal(scree_raw$settings$n_factors, 1)
  expect_equal(scree_cor_smc$settings$n_factors, 1)
  expect_equal(scree_efa_ml$settings$n_factors, 1)

})

rm(scree_cor, scree_cor_smc, scree_raw, scree_efa_ml)

# Oblique factor rotations with GPArotation package
.ROTATE_OBLQ <- function(x, rotation = c("oblimin", "quartimin", "simplimax",
                                        "bentlerQ", "geominQ", "bifactorQ"),
                        type = c("EFAtools", "psych", "SPSS", "none"),
                        normalize = TRUE, precision = 1e-5, order_type = NA,
                        k = NA, ...){

  if (type == "none") {

    if (is.na(order_type)) {

      stop(crayon::red$bold(cli::symbol$circle_cross), crayon::red(' "order_type" was NA and no valid "type" was specified. Either use one of "EFAtools", "psych", or "SPSS" for type, or specify the "order_type" argument\n'))

    }

  } else if (type == "EFAtools") {

    if (isFALSE(normalize)) {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Type and normalize is specified. normalize is used with value '", normalize, "'. Results may differ from the specified type\n"))
    }

    if (is.na(order_type)) {
      order_type <- "eigen"
    } else {
      warning(crayon::yellow$bold("!"), crayon::yellow(" Type and order_type is specified. order_type is used with value '", order_type, "'. Results may differ from the specified type\n"))
    }

  } else if (type == "psych") {

    # if not specified, set PAF properties. If specified, throw warning that
    # results may not exactly match the specified type

    if (isFALSE(normalize)) {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Type and normalize is specified. normalize is used with value '", normalize, "'. Results may differ from the specified type\n"))
    }

    if (is.na(order_type)) {
      order_type <- "eigen"
    } else {
      warning(crayon::yellow$bold("!"), crayon::yellow(" Type and order_type is specified. order_type is used with value '", order_type, "'. Results may differ from the specified type\n"))
    }

  } else if (type == "SPSS") {

    if (isFALSE(normalize)) {

      warning(crayon::yellow$bold("!"), crayon::yellow(" Type and normalize is specified. normalize is used with value '", normalize, "'. Results may differ from the specified type\n"))
    }

    if (is.na(order_type)) {
      order_type <- "ss_factors"
    } else {
      warning(crayon::yellow$bold("!"), crayon::yellow(" Type and order_type is specified. order_type is used with value '", order_type, "'. Results may differ from the specified type\n"))
    }

  }

  # extract loadings and dim names
  L <- x$unrot_loadings
  dim_names <- dimnames(L)

  # prepare settings
  settings <- list(normalize = normalize, precision = precision,
                   order_type = order_type, k = k)

  if (ncol(L) < 2) {

    # prepare and return output list
    output <- list(rot_loadings = L,
                   rotmat = NA,
                   Phi = NA,
                   Structure = NA,
                   vars_accounted_rot = NA,
                   settings = settings)

    warning(crayon::yellow$bold("!"), crayon::yellow(" Cannot rotate single factor. Unrotated loadings returned.\n"))
    return(output)
  }

  # perform the requested rotation
  if(rotation == "bentlerQ"){
    AV <- GPArotation::bentlerQ(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "oblimin"){
    AV <- GPArotation::oblimin(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "quartimin"){
    AV <- GPArotation::quartimin(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "geominQ"){
    AV <- GPArotation::geominQ(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "bifactorQ"){
    AV <- GPArotation::bifactorQ(L, eps = precision, normalize = normalize, ...)

  } else if (rotation == "simplimax"){

    if(is.na(k)){

      k <- nrow(L)
      # update settings
      settings$k <- k

    }

    AV <- GPArotation::simplimax(L, eps = precision, normalize = normalize,
                                 k = k, ...)

  }

  # get pattern and rotation matrix and factor intercorrelations
  patt_mat <- AV$loadings
  dimnames(patt_mat) <- dim_names
  rotmat <- AV$Th
  Phi <- AV$Phi

  # reflect factors with negative sums
  signs <- sign(colSums(patt_mat))
  signs[signs == 0] <- 1
  patt_mat <- patt_mat %*% diag(signs)

  if (order_type == "ss_factors") {

    # reorder the factors according to largest sums of squares
    ss <- colSums(patt_mat^2)
    ss_order <- order(ss, decreasing = TRUE)

    patt_mat <- patt_mat[, ss_order]

    rotmat <- rotmat[ss_order, ss_order]

    dim_names[[2]] <- dim_names[[2]][ss_order]

  }

    if (order_type == "eigen") {
      # reflect factors with negative sums
      signs <- sign(colSums(patt_mat))
      signs[signs == 0] <- 1
      patt_mat <- patt_mat %*% diag(signs)

      # order according to communalities
      eig_rotated <- diag(t(patt_mat) %*% patt_mat)
      eig_order <- order(eig_rotated, decreasing = TRUE)
      patt_mat <- patt_mat[, eig_order]

      Phi <- diag(signs) %*% Phi %*% diag(signs)
      Phi <- Phi[eig_order, eig_order]

      dim_names[[2]] <- dim_names[[2]][eig_order]

    }

  # get structure matrix
  Structure <- patt_mat %*% Phi

  # compute explained variances
  vars_accounted_rot <- .compute_vars(L_unrot = L, L_rot = patt_mat, Phi = Phi)
  colnames(vars_accounted_rot) <- colnames(patt_mat)

  # prepare and return output list
  class(patt_mat) <- "LOADINGS"

  output <- list(rot_loadings = patt_mat,
                 Phi = Phi,
                 Structure = Structure,
                 rotmat = rotmat,
                 vars_accounted_rot = vars_accounted_rot,
                 settings = settings)
  output
}

#' Velicer's minimum average partial (MAP) criterion
#'
#' @description
#' Computes Velicer's Minimum Average Partial (MAP) criterion for determining the number of
#' factors/components to retain. The function implements the original MAP criterion
#' (Velicer, 1976), expressed via the \eqn{\mathrm{TR2}} representation, and the revised
#' \eqn{\mathrm{TR4}} variant proposed by Velicer, Eaton, and Fava (2000).
#'
#' @details
#'
#' MAP partials successive principal components out of the correlation matrix and,
#' after removing \eqn{m} components, summarizes the off-diagonal partial
#' correlations \eqn{r^*_{ij}} that remain in the \eqn{m}-th partial correlation
#' matrix \eqn{M} (which has a unit diagonal); the suggested number of factors is
#' the \eqn{m} that minimizes the criterion. Two criteria are returned, each
#' rescaling the trace of a matrix power of \eqn{M} by the number of off-diagonal
#' cells \eqn{p(p-1)}:
#' \itemize{
#'   \item **TR2 (original MAP; Velicer, 1976):** the average squared off-diagonal
#'   partial correlation,
#'   \deqn{\mathrm{TR2}_m = \frac{\mathrm{tr}(M^2) - p}{p(p-1)} = \frac{\sum_{i \neq j} (r^*_{ij})^2}{p(p-1)},}
#'   where subtracting \eqn{p} removes the \eqn{p} unit diagonal entries.
#'   \item **TR4 (revised MAP; Velicer, Eaton, & Fava, 2000):** the analogous
#'   fourth-power summary, formed from the trace of the fourth matrix power,
#'   \deqn{\mathrm{TR4}_m = \frac{\mathrm{tr}(M^4) - p}{p(p-1)}.}
#'   Moving from the squared to the fourth power downweights the small partial
#'   correlations relative to the large ones, which can sharpen the minimum.
#'   Unlike TR2, \eqn{\mathrm{tr}(M^4)} is *not* the sum of the fourth powers of the
#'   individual partial correlations; the matrix power is intended and is what
#'   Velicer, Eaton, and Fava (2000) describe.
#' }
#'
#' Both criteria are returned for every call and they can suggest different numbers
#' of factors on the same correlation matrix. Both are in use in the literature and
#' neither is treated as the default here, so be sure to state which of the two you
#' report, as you would for any other analysis choice.
#'
#' MAP is most dependable when the components are well determined, that is with many
#' indicators per factor and substantial loadings. It has a well-documented tendency
#' to under-extract, particularly with few indicators per factor or weak loadings
#' (Zwick & Velicer, 1986; Auerswald & Moshagen, 2019), so it is best read as a lower
#' bound and paired with a criterion that errs in the other direction, such as the
#' Kaiser-Guttman criterion ([efa_kgc()]).
#'
#' The criterion is evaluated over \eqn{m = 0, \ldots, p - 1}. Each step standardizes
#' the partial covariance matrix by its residual standard deviations, which requires
#' every residual variance to stay positive. Partialling out all but one component
#' leaves a rank-one residual, so the final point \eqn{m = p - 1} is undefined for
#' most correlation matrices and is routinely returned as `NA`. A residual variance
#' can also reach zero earlier, most often on a near-singular matrix; the search then
#' stops there, the criterion values that could be computed are kept, the remaining
#' values stay `NA`, and a warning (class `efa_map_truncated`) reports how far the
#' grid was searched. In that case the suggested \eqn{m} is the minimum over the
#' evaluated range only, so it should be read together with the returned series.
#'
#' A non-positive-definite input correlation matrix (e.g. from sampling error) is
#' smoothed with [psych::cor.smooth()].
#'
#' @param x A numeric `matrix` or `data.frame`. Can be either (a) a correlation matrix, or
#'   (b) raw data (rows = observations, columns = variables) from which correlations are computed.
#' @param use Character string specifying the treatment of missing values when computing correlations.
#'   Passed to [stats::cor()]. Defaults to `"pairwise.complete.obs"`.
#' @param cor_method Character string specifying the correlation coefficient to be computed if raw
#'   data are supplied. One of `"pearson"`, `"spearman"`, or `"kendall"` (passed to
#'   [stats::cor()]), or `"poly"` / `"tetra"` for polychoric / tetrachoric correlations
#'   of ordinal / binary data (a two-step estimator). Defaults to `"pearson"`.
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] for the
#'   print method). MAP has no plot; [plot.efa_retention()] returns `NULL` with a
#'   message for it. Its main elements are:
#' \itemize{
#'   \item `n_factors`: A named numeric vector (`"TR2"`, `"TR4"`) with the index
#'   \eqn{m} that minimizes the original (TR2) and revised (TR4) MAP criterion.
#'   \item `results`: A list with one record per criterion, each holding the
#'   criterion values over \eqn{m} and, in `m_last`, the largest \eqn{m} at which
#'   the criterion could be evaluated (see details).
#'   \item `settings`: A list containing `use` and `cor_method`.
#' }
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. *Psychological Methods, 24*(4), 468--491.
#' https://doi.org/10.1037/met0000200
#' @source Velicer, W. F. (1976). Determining the number of components from the matrix of partial correlations.
#' *Psychometrika, 41*, 321--327.
#' @source Velicer, W. F., Eaton, C. A., & Fava, J. L. (2000). Construct explication through factor or component analysis: A review and evaluation of alternative procedures for determining the number of factors or components. In Goffin, R. D. & Helmes, E. (Eds.), *Problems and Solutions in
#' Human Assessment: Honoring Douglas N. Jackson at Seventy* (pp. 41--71). Boston: Kluwer.
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. *Psychological Bulletin, 99*,
#' 432--442. https://doi.org/10.1037/0033-2909.99.3.432
#'
#'
#' @examples
#' ## Example with raw data
#' res <- efa_map(GRiPS_raw)
#' res
#'
#' ## Example with a correlation matrix
#' res2 <- efa_map(test_models$baseline$cormat)
#' res2
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
efa_map <- function(x,
                use = c("pairwise.complete.obs", "all.obs",
                        "complete.obs", "everything",
                        "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")) {

  # Perform argument checks
  .assert_cor_input(x)

  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, use = use, cor_method = cor_method,
                             N_policy = "none")
  R <- prep$R

  p <- ncol(R)

  # set up m_grid
  m_max <- p - 1L
  ms <- 0:m_max
  criteria <- matrix(NA_real_, nrow = p, ncol = 2)
  colnames(criteria) <- c("TR2 (orig. MAP)", "TR4 (revised MAP)")


  # PCA to get the loadings A
  ed <- eigen(R, symmetric = TRUE)
  vals <- ed$values
  vecs <- ed$vectors
  # A = V Lambda^(1/2): scale each eigenvector column by the square root of its
  # eigenvalue (equivalent to vecs %*% sqrt(diag(vals)) but without forming the
  # p x p diagonal matrix)
  A <- sweep(vecs, 2, sqrt(vals), "*")

  map_from_partials <- function(M) {
    # Velicer's (1976) MAP and the revised criterion of Velicer, Eaton & Fava (2000)
    M2 <- M %*% M
    M4 <- M2 %*% M2
    c((sum(diag(M2)) - p) / (p * (p - 1)), # TR2 criterion (original MAP)
      (sum(diag(M4)) - p) / (p * (p - 1))) # TR4 criterion (revised MAP)
  }

  # m=0, Rstar = R
  criteria[1, ] <- map_from_partials(R)

  # Largest m at which the criterion could be evaluated, and whether that stop cut
  # the grid short. Partialling out all but one component leaves a rank-one residual,
  # so the final point m = p - 1 is undefined for most matrices; only a stop before
  # that shortens the searched range in a way the user needs to know about.
  m_last <- m_max
  truncated <- FALSE

  # run through ms
  for (m in seq_len(m_max)) {

    Am <- A[, 1:m]

    # Partial covariance: C = R - AA'
    Cm <- R - Am %*% t(Am)

    d <- diag(Cm)
    # Guard against zero/negative residual variances (can happen at very high m or numerical issues)
    if (any(!is.finite(d)) || any(d <= 1e-5)) {
      m_last <- m - 1L
      truncated <- m < m_max
      break
    }
    # D^(-1/2) to standardize to correlation matrix
    Dm <- diag(1 / sqrt(d))

    Rstar <- Dm %*% Cm %*% Dm
    criteria[m + 1, ] <- map_from_partials(Rstar)

  }

  if (isTRUE(truncated)) {
    cli::cli_warn(
      c("The MAP criterion could only be evaluated up to {m_last} partialled component{?s} of {m_max}.",
        "i" = "A residual variance reached zero there, so the criterion is {.code NA} beyond that point and the suggestion is the minimum over the range that could be computed.",
        "i" = "A stop well before the end of the grid usually indicates a near-singular correlation matrix."),
      class = "efa_map_truncated"
    )
  }

  n_factors_TR2 <- ms[which.min(criteria[, "TR2 (orig. MAP)"])]
  n_factors_TR4 <- ms[which.min(criteria[, "TR4 (revised MAP)"])]

  # one record per MAP criterion (criterion values over the number of partialled
  # components m)
  results <- list(
    list(name = "TR2", label = "Original implementation (TR2)",
         n_factors = n_factors_TR2, plot_type = "none",
         x = ms, y = criteria[, "TR2 (orig. MAP)"], m_last = m_last),
    list(name = "TR4", label = "Revised implementation (TR4)",
         n_factors = n_factors_TR4, plot_type = "none",
         x = ms, y = criteria[, "TR4 (revised MAP)"], m_last = m_last)
  )

  out <- .new_efa_retention(
    "MAP",
    results = results,
    settings = list(use = use, cor_method = cor_method),
    note = if (isTRUE(truncated)) {
      paste0("The criterion could only be evaluated up to ", m_last, " of ",
             m_max, " partialled components; the suggestion is the minimum over ",
             "that range, not over the full grid.")
    }
  )

  return(out)

}

map_cor <- efa_map(test_models$baseline$cormat)
map_raw <- efa_map(GRiPS_raw)

# Three uncorrelated triplets whose within-triplet correlations are near-unit but
# *different* from each other. The leading eigenvalue (2.99998) is then simple and
# well separated from the next two (2.80, 2.60), so its eigenvector is determined up
# to sign on every LAPACK driver and concentrates on the first triplet. That
# triplet's residual variance is 6.7e-06 after one component, below the 1e-5 guard,
# so the partialling degenerates at m = 1 on any platform. Do not equalise the three
# correlations: that makes the leading eigenvalues numerically tied, and any
# orthonormal basis of the resulting three-dimensional eigenspace is then a valid
# result -- a balanced eigenvector leaves every residual variance at 0.667 and the
# search does not truncate at all.
map_trunc_cor <- local({
  R <- diag(9)
  for (b in 0:2) {
    idx <- (b * 3 + 1):(b * 3 + 3)
    R[idx, idx] <- c(.99999, .90, .80)[b + 1]
  }
  diag(R) <- 1
  R
})

test_that("output class and dimensions are correct", {
  expect_s3_class(map_cor, "efa_retention")
  expect_length(map_cor, 6)
  expect_s3_class(map_raw, "efa_retention")
  expect_length(map_raw, 6)

  expect_named(map_cor$n_factors, c("TR2", "TR4"))
  expect_equal(.retention_record(map_cor, "TR2")$plot_type, "none")
})

test_that("identified number of factors is correct", {
  expect_equal(map_cor$n_factors[["TR2"]], 1)
  expect_equal(map_cor$n_factors[["TR4"]], 3)
})

test_that("criterion series are returned", {
  tr2 <- .retention_record(map_cor, "TR2")
  expect_named(tr2, c("name", "label", "n_factors", "plot_type", "x", "y",
                      "m_last"))
  expect_length(tr2$y, ncol(test_models$baseline$cormat))
  expect_length(tr2$x, ncol(test_models$baseline$cormat))
  expect_equal(tr2$x[1], 0)
  # the suggested m minimizes the criterion
  expect_equal(tr2$x[which.min(tr2$y)], map_cor$n_factors[["TR2"]])
})

test_that("the TR2 criterion series matches psych::vss", {
  skip_if_not_installed("psych")
  # Velicer's original MAP is TR2, which psych::vss returns as $map. EFAtools
  # additionally reports m = 0 (part of Velicer's procedure, omitted by psych), and
  # returns NA once the partialling degenerates at the largest m, where psych
  # substitutes 1; compare the m >= 1 values both report. `n.obs` feeds psych's
  # other statistics only -- MAP is a function of the correlation matrix alone --
  # so one nominal value covers both matrices.
  for (cmat in list(test_models$baseline$cormat, stats::cor(GRiPS_raw))) {
    ours <- .retention_record(efa_map(cmat), "TR2")$y[-1]
    theirs <- psych::vss(cmat, n = ncol(cmat) - 1, n.obs = 500,
                         plot = FALSE)$map

    compared <- !is.na(ours)
    # guard against a vacuously passing comparison
    expect_gt(sum(compared), 0.5 * length(ours))
    # The two sides compute the same quantity through different operation sequences --
    # ours partials with A = V sqrt(Lambda) and takes the trace of a matrix product,
    # psych's sums the squared residuals elementwise off its own principal() fit -- so
    # their rounding does not cancel. The partial matrix is standardised by
    # 1 / sqrt(d_i d_j), which at the largest m amplifies those last-digit differences
    # before they are squared and summed. A definitional divergence (a wrong partialling
    # order, a missing offset) moves the criterion by ~1e-2, so 1e-8 still catches one.
    expect_equal(ours[compared], theirs[compared], tolerance = 1e-8)
  }
})

test_that("TR4 is the trace of the fourth matrix power", {
  # For a three-variable equicorrelation matrix with r = .2, the eigenvalues
  # are 1.4, .8, .8. Hence tr(R^4) = 1.4^4 + 2 * .8^4 and the m = 0
  # revised-MAP value has the following closed form. It is far from the
  # element-wise fourth-power average (.2^4), so this guards that distinction.
  R <- matrix(.2, 3, 3)
  diag(R) <- 1
  tr4_expected <- (1.4^4 + 2 * .8^4 - 3) / (3 * 2)

  tr4 <- .retention_record(efa_map(R), "TR4")$y[1]
  expect_equal(tr4, tr4_expected, tolerance = 1e-14)
  expect_gt(abs(tr4 - .2^4), .1)
})

test_that("an ordinary run searches the whole grid and stays silent", {
  # Partialling out all but one component leaves a rank-one residual, so the final
  # point m = p - 1 is undefined on this matrix; that is the expected end of the
  # grid and must not raise the truncation warning.
  expect_no_warning(res <- efa_map(test_models$baseline$cormat),
                    class = "efa_map_truncated")
  expect_equal(.retention_record(res, "TR2")$m_last,
               ncol(test_models$baseline$cormat) - 2)
  expect_null(res$note)

  # GRiPS reaches the last point, so nothing is NA there
  expect_equal(.retention_record(map_raw, "TR2")$m_last, ncol(GRiPS_raw) - 1)
  expect_false(any(is.na(.retention_record(map_raw, "TR2")$y)))
})

test_that("a truncated grid warns, keeps the computed prefix, and records how far it got", {
  expect_warning(res <- efa_map(map_trunc_cor), class = "efa_map_truncated")

  for (nm in c("TR2", "TR4")) {
    rec <- .retention_record(res, nm)
    # the search stopped well short of the grid, which is what the warning reports
    expect_equal(rec$m_last, 0)
    expect_lt(rec$m_last, ncol(map_trunc_cor) - 2)
    # the prefix that could be computed is kept, the rest stays NA -- asserted
    # against m_last itself, so the contract holds wherever the stop lands
    kept <- seq_len(rec$m_last + 1)
    expect_false(any(is.na(rec$y[kept])))
    expect_true(all(is.na(rec$y[-kept])))
    # the suggestion is the minimum over the evaluated range
    expect_equal(res$n_factors[[nm]], rec$x[which.min(rec$y)])
  }

  # the print says so too, rather than presenting a fully searched grid
  expect_length(res$note, 1)
  expect_match(res$note, "not over the full grid", fixed = TRUE)
})

test_that("a truncated MAP does not take the criterion out of efa_retain()", {
  expect_warning(
    nf <- efa_retain(map_trunc_cor, N = 500, suitability = FALSE,
                     criteria = "MAP"),
    class = "efa_map_truncated"
  )
  expect_named(nf$outputs, "MAP")
  expect_null(nf$not_run)
})

test_that("settings are returned correctly", {
  expect_named(map_cor$settings, c("use", "cor_method"))
  expect_equal(map_cor$settings$use, "pairwise.complete.obs")
  expect_equal(map_cor$settings$cor_method, "pearson")
})

test_that("errors are thrown correctly", {
  expect_error(efa_map(1:5), class = "efa_input_not_matrix")
  expect_message(efa_map(GRiPS_raw), class = "efa_cor_from_data")
})

rm(map_cor, map_raw, map_trunc_cor)

#' Print and format an efa_bartlett object
#'
#' `print()` reports the outcome of [efa_bartlett()]'s test of sphericity: a verdict
#' on whether the test was significant (and what that implies for the suitability
#' of the data for factor analysis), followed by the chi-square statistic, its
#' degrees of freedom, and the p-value. `format()` assembles the same report and
#' returns it as a character vector; `print()` is `cat(format(x), sep = "\n")`.
#' The lines follow the active console theme, so they are plain when colours are
#' disabled (for example when captured into a file or stripped with
#' [cli::ansi_strip()]).
#'
#' @param x An object of class `efa_bartlett` (output from [efa_bartlett()]).
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @export
#'
#' @method print efa_bartlett
#'
#' @examples
#' bart <- efa_bartlett(test_models$baseline$cormat, N = 500)
#' bart
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(bart))
#'
print.efa_bartlett <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

# The undefined-statistic verdict, shared by format.efa_bartlett() and format.efa_screen() so
# the two entry points cannot drift. Names the reason as well as the outcome: a non-positive
# small-sample multiplier is what leaves the statistic undefined, and the object carries no
# other clue. Wording shared with the warning raised when that NA is computed.
.print_bartlett_no_result <- function() {
  mult_note <- .fit_unavailable_text("bartlett_mult")
  cli::cli_alert_warning("The Bartlett's test of sphericity did not render a result.",
                         wrap = TRUE)
  cli::cli_text("{mult_note}")
  invisible(NULL)
}

#' @rdname print.efa_bartlett
#' @export
#' @method format efa_bartlett
format.efa_bartlett <- function(x, ...) {
  pval <- x$p_value

  cli::cli_format_method({
    cli::cli_text("")

    if (!is.null(pval) && !is.na(pval)) {
      if (pval < .05) {
        significant <- cli::col_green(cli::style_bold("significant"))
        cli::cli_alert_success(
          "The {.strong Bartlett's test of sphericity} was {significant} at an alpha level of .05.",
          wrap = TRUE
        )
        cli::cli_text("These data are probably suitable for factor analysis.")
      } else {
        not_significant <- cli::col_red(cli::style_bold("not significant"))
        cli::cli_alert_danger(
          "The Bartlett's test of sphericity was {not_significant} at an alpha level of .05.",
          wrap = TRUE
        )
        cli::cli_text("These data are probably not suitable for factor analysis.")
      }
    } else {
      .print_bartlett_no_result()
    }

    # The chi-square line is shown for every object, including the "no result" case above, so
    # render the statistic and p-value defensively when they are NA/NULL (.screen_p_str()
    # handles the p-value tail, shared with the screening report). It is a fixed-format
    # statistic line, emitted verbatim so cli never reflows it mid-token.
    chisq_str <- if (is.null(x$chisq) || is.na(x$chisq)) "NA" else round(x$chisq, 2)
    p_str <- .screen_p_str(pval)

    cli::cli_text("")
    cli::cli_verbatim(paste0("\U03C7\U00B2(", x$df, ") = ", chisq_str, ", ",
                             cli::style_italic("p"), p_str))
  })
}

.efa_pooled_analytic_pool <- function(fits,
                                      unrot_loadings_aligned,
                                      align_meta,
                                      ci = 0.95,
                                      align_unrotated = c("signed_tucker_congruence",
                                                          "none", "procrustes"),
                                      rotation_type = c("none", "orthogonal",
                                                        "oblique"),
                                      rot_loadings = NULL,
                                      phis = NULL,
                                      structure_loadings = NULL,
                                      mean_structure_loadings = NULL,
                                      mean_phis = NULL,
                                      h2 = NULL) {
  # Rubin's-rules pool of analytic per-imputation marginal SEs for the unrotated
  # loadings and uniquenesses. Serves component fits run with se = "information":
  # each fit carries $SE$unrot_loadings (p x k) and $SE$uniquenesses (length p),
  # and there are no bootstrap replicate cubes to estimate within-imputation
  # variance from. (The sandwich route is handled separately by
  # .efa_pooled_mi2s() and never reaches this helper.)
  #
  # Under `align_unrotated = "signed_tucker_congruence"` or `"none"`, the
  # per-imputation gauge transform is a signed permutation that preserves
  # marginal SEs up to a column permutation, so $SE$unrot_loadings is reused
  # directly after applying the alignment's `factor_order`.
  #
  # Under `align_unrotated = "procrustes"`, the per-imputation orthogonal Q_d
  # MIXES loading columns, so marginal SEs alone no longer determine the aligned
  # marginal SE matrix. The full unrotated covariance persisted on each fit
  # (`vcov_unrot_loadings`, populated by `se = "information"` and
  # `se = "sandwich"`) is propagated by
  # `.efa_pooled_propagate_procrustes_vcov()`, which implements
  # diag(Q' V_d^(i) Q) per row i. Q_d is treated as fixed.
  #
  # Per-element NA propagation is fail-closed: if any imputation has NA for an
  # element (in either Q_d or SE_d), every pooled output at that position (SE,
  # CI bounds, RIV, FMI, df) is NA.
  #
  # The returned shape mirrors .efa_pooled_bootstrap_pool() so the efa_mi()
  # assembly block can route either pool through the same downstream code:
  # list(SE, CI, replicates = NULL, MI, n_boot = NULL). Empty inputs or fits
  # missing the analytic SE slot return NULL so the caller falls through to the
  # existing silent-downgrade branch. Fits missing the full vcov slot under the
  # procrustes branch abort with a classed condition (fail-closed; the marginal
  # SE alone cannot represent the aligned gauge).
  align_unrotated <- match.arg(align_unrotated)
  rotation_type <- match.arg(rotation_type)
  if (length(fits) == 0L) return(NULL)
  has_se <- vapply(fits, function(x) {
    !is.null(x$SE) &&
      !is.null(x$SE$unrot_loadings) &&
      !is.null(x$SE$uniquenesses)
  }, logical(1L))
  if (!all(has_se)) {
    if (any(has_se)) {
      cli::cli_warn(
        c("Analytic SE slots were found for only some imputations; pooled analytic SEs/CIs were not computed.",
          "i" = "Inspect each fit's {.code $SE} slot to identify the imputations with missing analytic SEs."),
        class = "efa_pooled_analytic_partial_se"
      )
    }
    return(NULL)
  }

  # Whole-fit reliability gate (every alignment method). `EFA()` NA-fills the
  # entire `vcov_unrot_loadings` when the analytic covariance is unreliable (a
  # Heywood case or a singular bordered information matrix), NA-filling the
  # marginal SE matrices in lockstep and firing `efa_se_unreliable` at fit time.
  # Pooling those would yield an all-NA SE bundle with no diagnostic, so abort
  # here; `efa_mi()` converts this into a classed `efa_pooled_se_unavailable`
  # warning and falls back to point-estimate-only pooling. The gate keys on the
  # whole-matrix vcov NA fill, not on individual `$SE` elements, so a genuine
  # per-element NA (one imputation NA at a single cell) still fail-closes per
  # element via the mask below rather than dropping the entire pool.
  vcov_reliable <- vapply(fits, function(x) {
    is.null(x$vcov_unrot_loadings) || !anyNA(x$vcov_unrot_loadings)
  }, logical(1L))
  if (!all(vcov_reliable)) {
    bad <- which(!vcov_reliable)
    n_bad <- length(bad)
    cli::cli_abort(
      c("Pooled analytic standard errors require a reliable unrotated loading covariance on every imputation.",
        "i" = "{cli::qty(n_bad)}Imputation{?s} {.val {bad}} {cli::qty(n_bad)}carr{?ies/y} an NA-filled {.code $vcov_unrot_loadings} (Heywood case or singular information). Drop or re-fit {cli::qty(n_bad)}{?it/them}."),
      class = "efa_pooled_unreliable_vcov"
    )
  }

  if (align_unrotated == "procrustes") {
    # The Procrustes branch propagates the full covariance through Q_d, so the
    # slot must be present (not merely reliable) on every imputation.
    has_vcov <- vapply(fits, function(x) {
      !is.null(x$vcov_unrot_loadings)
    }, logical(1L))
    if (!all(has_vcov)) {
      bad <- which(!has_vcov)
      n_bad <- length(bad)
      cli::cli_abort(
        c("{.code align_unrotated = \"procrustes\"} with {.code se = \"information\"} requires the full unrotated loading covariance on every imputation.",
          "i" = "{cli::qty(n_bad)}Imputation{?s} {.val {bad}} {cli::qty(n_bad)}{?is/are} missing {.code $vcov_unrot_loadings}; re-fit those imputations or use {.code align_unrotated = \"signed_tucker_congruence\"}."),
        class = "efa_pooled_no_vcov"
      )
    }
  }

  m <- length(fits)
  L_target <- as.matrix(unrot_loadings_aligned[[1]])
  p_vars   <- nrow(L_target)
  k_fac    <- ncol(L_target)
  loading_dn <- dimnames(L_target)
  var_names  <- rownames(L_target)

  # Per-imputation aligned loadings and aligned SEs.
  # - signed_tucker_congruence / none: column-permute the marginal SE matrix by
  #   the alignment's factor_order.
  # - procrustes: propagate the full vcov through Q_d row-by-row to obtain the
  #   aligned marginal variances, then take sqrt to land back on the same
  #   variance-of-vec(L_d Q_d) scale as Ubar_L (which squares SE_mat).
  Q_mat   <- matrix(NA_real_, nrow = m, ncol = p_vars * k_fac)
  SE_mat  <- matrix(NA_real_, nrow = m, ncol = p_vars * k_fac)
  Q_psi   <- matrix(NA_real_, nrow = m, ncol = p_vars)
  SE_psi_mat <- matrix(NA_real_, nrow = m, ncol = p_vars)
  # Imputations whose covariance block the canonical-gauge branch needed but did
  # not find; collected so the silent all-NA block below can be reported.
  gauge_no_vcov <- integer(0)
  for (d in seq_len(m)) {
    Q_mat[d, ] <- as.vector(as.matrix(unrot_loadings_aligned[[d]]))

    if (identical(align_meta[[d]]$type, "procrustes")) {
      V_d  <- fits[[d]]$vcov_unrot_loadings
      Q_d  <- align_meta[[d]]$Q
      se_d <- .efa_pooled_propagate_procrustes_vcov(V_d, Q_d, p_vars, k_fac)
      # Reading the covariance block bypasses the marginal-SE NA mask, so apply
      # it here too: an SE the component fit could not certify must not come
      # back finite. Q_d mixes columns, so one NA contaminates its whole row.
      se_marg <- fits[[d]]$SE$unrot_loadings
      if (!is.null(se_marg)) {
        se_d[apply(is.na(as.matrix(se_marg)), 1, any), ] <- NA_real_
      }
      SE_mat[d, ] <- as.vector(se_d)
    } else if (!is.null(align_meta[[d]]$C)) {
      # The canonical-gauge rotation mixes loading columns, so the marginal SEs
      # cannot merely be permuted: the full unrotated covariance is propagated
      # through the composed orthogonal transform E_d C, where E_d is the signed
      # permutation that aligned imputation d. Fail-closed at the imputation
      # level -- a missing or non-finite covariance block leaves this row NA and
      # the per-element NA mask blanks the affected pooled outputs -- rather than
      # aborting, so a component fit without the block still pools everything else.
      fo <- align_meta[[d]]$factor_order
      sg <- align_meta[[d]]$factor_sign
      if (is.null(fo) || is.null(sg)) {
        cli::cli_abort(
          "Per-imputation alignment metadata is missing for imputation {d}; analytic-SE pooling requires a column-permutation/sign alignment.",
          class = "efa_pooled_analytic_align_meta_missing"
        )
      }
      E <- matrix(0, k_fac, k_fac)
      E[cbind(fo, seq_len(k_fac))] <- sg
      V_d <- fits[[d]]$vcov_unrot_loadings
      if (is.null(V_d)) {
        gauge_no_vcov <- c(gauge_no_vcov, d)
      } else {
        se_aligned <- .efa_pooled_propagate_procrustes_vcov(
          as.matrix(V_d), E %*% align_meta[[d]]$C, p_vars, k_fac)
        # Keep the marginal-SE NA mask meaningful. The propagation reads the
        # covariance block, so an unreliable marginal SE would otherwise be
        # silently replaced by a finite value. Because the gauge rotation mixes
        # columns, one NA contaminates its whole row.
        se_marg <- fits[[d]]$SE$unrot_loadings
        if (!is.null(se_marg)) {
          bad_row <- apply(is.na(as.matrix(se_marg)), 1, any)
          se_aligned[bad_row, ] <- NA_real_
        }
        SE_mat[d, ] <- as.vector(se_aligned)
      }
    } else {
      se_d <- as.matrix(fits[[d]]$SE$unrot_loadings)
      fo <- align_meta[[d]]$factor_order
      if (is.null(fo)) {
        # All currently-supported producers of `align_meta` populate
        # `factor_order` (the `none` and `signed_tucker_congruence` branches of
        # `.efa_pooled_align_unrotated_list`); procrustes meta is routed into
        # the V_d / Q_d branch above by the `type` test and never reaches
        # here. Catch any future bypass at the helper boundary rather than
        # silently falling back to an identity permutation that would pool
        # SEs against unaligned loadings.
        cli::cli_abort(
          "Per-imputation alignment metadata is missing for imputation {d}; analytic-SE pooling requires a column-permutation/sign alignment.",
          class = "efa_pooled_analytic_align_meta_missing"
        )
      }
      SE_mat[d, ] <- as.vector(se_d[, fo, drop = FALSE])
    }

    psi_d <- .efa_pooled_communalities(unrot_loadings_aligned[[d]], NULL)
    Q_psi[d, ] <- 1 - as.numeric(psi_d)
    SE_psi_mat[d, ] <- as.numeric(fits[[d]]$SE$uniquenesses)
  }

  if (length(gauge_no_vcov)) {
    # Without the covariance block the gauge rotation cannot be propagated, so
    # those rows stay NA and the per-element mask blanks the whole unrotated
    # loading family. Say so: an all-NA block returned in silence is not
    # diagnosable, and every sibling branch here signals its downgrade.
    cli::cli_warn(
      c("The unrotated loading covariance is missing for {cli::qty(length(gauge_no_vcov))}imputation{?s} {.val {gauge_no_vcov}}; pooled unrotated loading SEs/CIs were not computed.",
        "i" = "The pooled solution is returned in the extraction's canonical gauge, which mixes loading columns, so the full {.code $vcov_unrot_loadings} is required rather than the marginal SEs alone."),
      class = "efa_pooled_gauge_vcov_missing"
    )
  }

  # A component fit whose analytic covariance was unreliable at the loading level carries
  # an NA-filled $SE$unrot_loadings while keeping a finite $vcov_unrot_loadings, so the
  # whole-fit gate above passes and the fail-closed mask then blanks the entire pooled
  # unrotated SE block. Name the affected imputations, as the rotated family does: an
  # all-NA block returned in silence reads as a computation that succeeded. The
  # imputations already reported by the missing-covariance warning are excluded so a
  # single downgrade is not announced twice.
  unrot_se_unreliable <- setdiff(
    which(apply(SE_mat, 1L, function(r) all(is.na(r)))),
    gauge_no_vcov
  )
  if (length(unrot_se_unreliable) > 0L) {
    n_bad <- length(unrot_se_unreliable)
    cli::cli_warn(
      c("Pooled unrotated standard errors could not be produced for {n_bad} imputation{?s}.",
        "i" = "Affected: {.val {unrot_se_unreliable}} (each left an NA-filled unrotated standard error, typically the upstream {.code efa_se_unreliable} signal).",
        "i" = "The pooled unrotated loading standard errors are returned as {.code NA}."),
      class = "efa_pooled_unrotated_se_unreliable"
    )
  }

  # Fail-closed NA mask and Rubin pooling are shared with the rotated families
  # below via .efa_pooled_analytic_marginal().
  alpha <- 1 - ci
  pool_L   <- .efa_pooled_analytic_marginal(Q_mat, SE_mat,     m, alpha)
  pool_psi <- .efa_pooled_analytic_marginal(Q_psi, SE_psi_mat, m, alpha)

  reshape_loading <- function(v) {
    M <- matrix(v, p_vars, k_fac)
    dimnames(M) <- loading_dn
    M
  }
  reshape_psi <- function(v) {
    out <- as.numeric(v)
    if (!is.null(var_names)) names(out) <- var_names
    out
  }

  unrot_asm <- .efa_pooled_assemble_family(pool_L,   reshape_loading)
  psi_asm   <- .efa_pooled_assemble_family(pool_psi, reshape_psi)

  SE <- list(unrot_loadings = unrot_asm$SE, uniquenesses = psi_asm$SE)
  CI <- list(unrot_loadings = unrot_asm$CI, uniquenesses = psi_asm$CI)
  MI <- list(unrot_loadings = unrot_asm$MI, uniquenesses = psi_asm$MI)

  ## Rotated quantities -------------------------------------------------------
  ## Pool the rotated loadings, communalities, and (oblique) factor correlations
  ## and structure coefficients, each in a common MI rotational gauge. A
  ## rotated-loading standard error is conditional on the rotation criterion
  ## (Archer & Jennrich 1973; Jennrich 1973, 1974; Zhang, Preacher, & Jennrich
  ## 2012; Zhang & Preacher 2015), so for orthogonal
  ## and oblique rotations alike the within-imputation variance is each fit's own
  ## criterion-aware delta-method rotated SE (the quantity EFA() returns), reused
  ## after a signed-permutation alignment to the MI target. The between-imputation
  ## spread of the aligned point estimates supplies B, so the pooled SE shares the
  ## estimand of the np-boot pool, which likewise aligns each replicate to the
  ## target. This is a deliberate approximation -- the per-fit SE is conditional on
  ## that fit's rotation optimum rather than the common gauge -- and is flagged in
  ## MI$<param>$method = "signed_permutation_approx". Communalities are
  ## rotation-invariant and pool without any alignment.
  if (rotation_type != "none") {
    # Phi and Structure exist only for a genuine oblique rotation (a single-factor
    # solution carries no factor correlations); gate on k_fac so the oblique branch
    # never dereferences the NULL Phi/Structure slots of a one-factor fit.
    oblique <- rotation_type == "oblique" && k_fac >= 2L

    q_rot  <- matrix(NA_real_, m, p_vars * k_fac)
    se_rot <- matrix(NA_real_, m, p_vars * k_fac)
    q_h2   <- matrix(NA_real_, m, p_vars)
    se_h2  <- matrix(NA_real_, m, p_vars)
    if (oblique) {
      vlen   <- k_fac * (k_fac + 1L) / 2L
      q_phi  <- matrix(NA_real_, m, vlen)
      se_phi <- matrix(NA_real_, m, vlen)
      q_str  <- matrix(NA_real_, m, p_vars * k_fac)
      se_str <- matrix(NA_real_, m, p_vars * k_fac)
    }

    for (d in seq_len(m)) {
      Lr_d  <- as.matrix(rot_loadings[[d]])
      Phi_d <- if (oblique) as.matrix(phis[[d]]) else NULL
      q_rot[d, ] <- .efa_pooled_vec(Lr_d)
      q_h2[d, ]  <- .efa_pooled_communalities(Lr_d, Phi_d)
      se_h2[d, ] <- as.numeric(fits[[d]]$SE$communalities)

      # Within-imputation variance for the rotated loadings = this fit's
      # criterion-aware delta-method rotated SE, reused after a signed-permutation
      # alignment to THIS imputation's pooled rotated loading Lr_d. Aligning to
      # Lr_d (rather than the global target) keeps the SE column order matched to
      # the order q_rot/q_phi/q_str use; aligning to the global target can pick a
      # different order when factors are weakly separated, pairing a cell's pooled
      # estimate with another factor's within-imputation variance. Only the column
      # order is used (SE magnitudes are sign-invariant).
      fo <- .align_solution(
        L_target = Lr_d,
        L        = as.matrix(fits[[d]]$rot_loadings)
      )$factor_order
      se_rot[d, ] <- .efa_pooled_vec(
        as.matrix(fits[[d]]$SE$rot_loadings)[, fo, drop = FALSE]
      )

      if (oblique) {
        # Apply the same column alignment to the per-fit Phi and Structure SEs.
        se_phi[d, ] <- .vech(
          as.matrix(fits[[d]]$SE$Phi)[fo, fo, drop = FALSE]
        )
        q_phi[d, ]  <- .vech(Phi_d)
        se_str[d, ] <- .efa_pooled_vec(
          as.matrix(fits[[d]]$SE$Structure)[, fo, drop = FALSE]
        )
        q_str[d, ]  <- .efa_pooled_vec(as.matrix(structure_loadings[[d]]))
      }
    }

    # A component fit whose rotated-SE Jacobian could not be reproduced carries an
    # all-NA $SE$rot_loadings (the upstream efa_se_unreliable signal) even when its
    # unrotated covariance is fine, so the whole-fit reliability gate above does not
    # catch it. Such a fit leaves an all-NA row in se_rot, which the fail-closed mask
    # in .efa_pooled_analytic_marginal() then propagates to every cell of the pooled
    # rotated-loading (and, for oblique fits, Phi/Structure) SE family. Name the
    # affected imputations so the resulting all-NA rotated SE block is diagnosable
    # rather than silent.
    rot_se_unreliable <- which(apply(se_rot, 1L, function(r) all(is.na(r))))
    if (length(rot_se_unreliable) > 0L) {
      n_bad <- length(rot_se_unreliable)
      affected <- if (oblique) {
        "rotated-loading, factor-correlation, and structure-coefficient"
      } else {
        "rotated-loading"
      }
      cli::cli_warn(
        c("Pooled rotated standard errors could not be produced for {n_bad} imputation{?s}.",
          "i" = "Affected: {.val {rot_se_unreliable}} (each carried an NA-filled rotated standard error, the upstream {.code efa_se_unreliable} signal).",
          "i" = "The pooled {affected} standard errors are returned as {.code NA}."),
        class = "efa_pooled_rotated_se_unreliable"
      )
    }

    pool_rot <- .efa_pooled_analytic_marginal(q_rot, se_rot, m, alpha)
    pool_h2  <- .efa_pooled_analytic_marginal(q_h2, se_h2, m, alpha,
                                              est_override = as.numeric(h2))

    # Rotated loadings and structure are p x k like the unrotated loadings, so they
    # share the same factor-column labels (reshape_loading carries loading_dn).
    rot_asm <- .efa_pooled_assemble_family(
      pool_rot, reshape_loading,
      method = "signed_permutation_approx"
    )
    SE$rot_loadings <- rot_asm$SE; CI$rot_loadings <- rot_asm$CI
    MI$rot_loadings <- rot_asm$MI

    h2_asm <- .efa_pooled_assemble_family(pool_h2, reshape_psi,
                                          method = "gauge_invariant")
    SE$communalities <- h2_asm$SE; CI$communalities <- h2_asm$CI
    MI$communalities <- h2_asm$MI
    # `communalities` is the name EFA() uses, so it is the canonical one here too;
    # `h2` is what this route has always returned and is kept as an alias of it. The
    # MI diagnostics carry no alias: the printed FMI/RIV summary walks the whole MI
    # tree, so a duplicated family would enter those statistics twice.
    SE$h2 <- SE$communalities; CI$h2 <- CI$communalities

    if (oblique) {
      phi_dn <- dimnames(as.matrix(phis[[1]]))
      unvech_phi <- function(v) {
        M <- .unvech(v, k_fac)
        dimnames(M) <- phi_dn
        M
      }
      # Center the Phi CI on the same symmetrised pooled estimate reported as $Phi
      # (mean_phis, always present on this oblique branch), matching the est_override
      # the h2 and Structure families use; without it the off-diagonal CI would
      # center on the unsymmetrised column mean of vech(Phi), which can differ from
      # the reported (symmetrised) estimate.
      pool_phi <- .efa_pooled_analytic_marginal(
        q_phi, se_phi, m, alpha,
        est_override = .vech(mean_phis)
      )
      phi_asm <- .efa_pooled_assemble_family(pool_phi, unvech_phi,
                                             method = "signed_permutation_approx")
      # The fixed unit diagonal of Phi is not an estimated parameter: zero its SE
      # and NA its MI diagnostics so the fixed cells stay out of the printed
      # FMI/RIV summaries.
      diag(phi_asm$SE) <- 0
      diag(phi_asm$MI$RIV) <- NA_real_
      diag(phi_asm$MI$FMI) <- NA_real_
      diag(phi_asm$MI$df)  <- NA_real_
      SE$Phi <- phi_asm$SE; CI$Phi <- phi_asm$CI; MI$Phi <- phi_asm$MI

      pool_str <- .efa_pooled_analytic_marginal(
        q_str, se_str, m, alpha,
        est_override = .efa_pooled_vec(mean_structure_loadings)
      )
      str_asm <- .efa_pooled_assemble_family(pool_str, reshape_loading,
                                             method = "signed_permutation_approx")
      SE$Structure <- str_asm$SE; CI$Structure <- str_asm$CI
      MI$Structure <- str_asm$MI
    }
  }

  list(SE = SE, CI = CI, replicates = NULL, MI = MI, n_boot = NULL)
}


.efa_pooled_align_unrot_boot <- function(L_boot, target, align_unrotated) {
  # Align one bootstrap unrotated loading matrix to the pooled unrotated target.
  if (align_unrotated == "none") {
    return(as.matrix(L_boot))
  }

  if (align_unrotated == "procrustes") {
    return(.change_class(efa_procrustes(
      A = as.matrix(L_boot),
      Target = as.matrix(target),
      rotation = "orthogonal"
    )$loadings, "matrix"))
  }

  .align_solution(
    L_target = as.matrix(target),
    L = as.matrix(L_boot)
  )$loadings
}

.efa_pooled_align_rot_boot <- function(L_boot_unrot, final_target, rotation_type,
                                       procrustes_args = list()) {
  # Re-rotate/re-align the bootstrap unrotated loadings to the final MI target.
  # This is the key step that makes within-imputation bootstrap variances refer
  # to the same estimand as the pooled rotated solution.
  if (rotation_type == "none") {
    return(NULL)
  }

  pr <- do.call(
    efa_procrustes,
    c(list(A = as.matrix(L_boot_unrot),
           Target = as.matrix(final_target),
           rotation = rotation_type),
      procrustes_args)
  )

  # Drop the replicate only when no valid alignment was produced. A best
  # multi-start fit that did not formally converge is still the lowest-objective
  # alignment available, with well-defined loadings, and is retained.
  if (isFALSE(pr$valid)) {
    return(NULL)
  }

  list(
    Lambda = .change_class(pr$loadings, "matrix"),
    Phi = if (!is.null(pr$Phi)) as.matrix(pr$Phi) else NULL
  )
}

.efa_pooled_bootstrap_pool <- function(fits,
                                       orig_R_list,
                                       unrot_loadings_aligned,
                                       mean_unrot_loadings,
                                       rot_loadings = NULL,
                                       phis = NULL,
                                       structure_loadings = NULL,
                                       mean_structure_loadings = NULL,
                                       final_target = NULL,
                                       rotation_type = c("none", "orthogonal", "oblique"),
                                       align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
                                       procrustes_args = list(),
                                       h2,
                                       residuals,
                                       alpha = 0.05) {

  rotation_type <- match.arg(rotation_type)
  align_unrotated <- match.arg(align_unrotated)

  has_boot <- .efa_pooled_has_replicates(fits)
  if (!all(has_boot)) {
    if (any(has_boot)) {
      cli::cli_warn("Bootstrap replicates were found for only some imputations; pooled bootstrap SEs/CIs were not computed.",
                    class = "efa_pooled_partial_boot")
    }
    return(NULL)
  }

  m <- length(fits)
  # Coerce the loop-invariant alignment targets once; mean_unrot_loadings is a
  # LOADINGS object whose as.matrix() coercion is otherwise repeated on every
  # bootstrap replicate inside the d/b loops.
  mean_unrot_target <- as.matrix(mean_unrot_loadings)
  final_target_m <- if (is.null(final_target)) NULL else as.matrix(final_target)
  p_vars <- nrow(mean_unrot_target)
  k <- ncol(mean_unrot_target)
  loading_dimnames <- dimnames(mean_unrot_target)

  # A single factor cannot be rotated and has no factor correlations, so an
  # oblique request on one factor degenerates to the no-Phi case (matching a
  # single-fit EFA()); align such a solution by sign only.
  oblique <- rotation_type == "oblique" && k >= 2L
  boot_proc_rotation <- if (oblique) "oblique" else "orthogonal"

  B_vec <- vapply(fits, function(x) dim(x$replicates$unrot_loadings)[3], integer(1))
  if (length(unique(B_vec)) > 1L) {
    B_use <- min(B_vec)
    cli::cli_warn("The number of bootstrap replicates differs across imputations; using the minimum number available in all imputations.",
                  class = "efa_pooled_unequal_boot")
  } else {
    B_use <- B_vec[[1]]
  }

  if (!is.finite(B_use) || B_use < 2L) {
    cli::cli_warn("At least two bootstrap replicates per imputation are required for pooled bootstrap SEs/CIs.",
                  class = "efa_pooled_min_boot")
    return(NULL)
  }

  q_unrot <- lapply(unrot_loadings_aligned, .efa_pooled_vec)
  boot_unrot <- vector("list", m)

  q_rot <- boot_rot <- NULL
  q_phi <- boot_phi <- NULL
  q_structure <- boot_structure <- NULL
  q_h2 <- vector("list", m)
  boot_h2 <- vector("list", m)
  q_residuals <- vector("list", m)
  boot_residuals <- vector("list", m)

  if (rotation_type != "none") {
    q_rot <- lapply(rot_loadings, .efa_pooled_vec)
    boot_rot <- vector("list", m)
  }
  if (oblique) {
    q_phi <- lapply(phis, .vech)
    boot_phi <- vector("list", m)
    q_structure <- lapply(structure_loadings, .efa_pooled_vec)
    boot_structure <- vector("list", m)
  }

  nonconv_procrustes <- integer(m)
  boot_failures <- integer(m)

  for (d in seq_len(m)) {
    arr <- fits[[d]]$replicates
    unrot_arr <- arr$unrot_loadings

    boot_unrot[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * k)

    if (rotation_type != "none") {
      boot_rot[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * k)
    }
    if (oblique) {
      boot_phi[[d]] <- matrix(NA_real_, nrow = B_use, ncol = k * (k + 1) / 2)
      boot_structure[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * k)
    }

    boot_h2[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars)
    boot_residuals[[d]] <- matrix(NA_real_, nrow = B_use, ncol = p_vars * p_vars)

    if (oblique) {
      q_h2[[d]] <- .efa_pooled_communalities(rot_loadings[[d]], phis[[d]])
      q_residuals[[d]] <- .efa_pooled_vec(.efa_pooled_residual_from_solution(
        orig_R_list[[d]], rot_loadings[[d]], phis[[d]]
      ))
    } else if (rotation_type != "none") {
      q_h2[[d]] <- .efa_pooled_communalities(rot_loadings[[d]], NULL)
      q_residuals[[d]] <- .efa_pooled_vec(.efa_pooled_residual_from_solution(
        orig_R_list[[d]], rot_loadings[[d]], NULL
      ))
    } else {
      q_h2[[d]] <- .efa_pooled_communalities(unrot_loadings_aligned[[d]], NULL)
      q_residuals[[d]] <- .efa_pooled_vec(.efa_pooled_residual_from_solution(
        orig_R_list[[d]], unrot_loadings_aligned[[d]], NULL
      ))
    }

    for (b in seq_len(B_use)) {
      Lb_unrot0 <- unrot_arr[, , b]

      # A component EFA NA-fills a bootstrap replicate it could not fit; skip it
      # (its pre-initialized rows stay NA and are dropped by the NA-tolerant
      # pooling) and tally it, rather than letting the NA reach the alignment /
      # Procrustes step and abort the whole pooled bootstrap.
      if (!all(is.finite(Lb_unrot0))) {
        boot_failures[d] <- boot_failures[d] + 1L
        next
      }

      Lb_unrot <- .efa_pooled_align_unrot_boot(
        L_boot = Lb_unrot0,
        target = mean_unrot_target,
        align_unrotated = align_unrotated
      )
      boot_unrot[[d]][b, ] <- .efa_pooled_vec(Lb_unrot)

      if (rotation_type != "none") {
        # Do not use arr$rot_loadings here. Those matrices were aligned to the
        # imputation-specific EFA target. MI pooling requires every bootstrap
        # replicate to be represented in the final MI target space.
        rot_b <- .efa_pooled_align_rot_boot(
          L_boot_unrot = Lb_unrot0,
          final_target = final_target_m,
          rotation_type = boot_proc_rotation,
          procrustes_args = procrustes_args
        )

        if (!is.null(rot_b)) {
          boot_rot[[d]][b, ] <- .efa_pooled_vec(rot_b$Lambda)

          if (oblique) {
            boot_phi[[d]][b, ] <- .vech(rot_b$Phi)
            boot_structure[[d]][b, ] <- .efa_pooled_vec(rot_b$Lambda %*% rot_b$Phi)
            boot_h2[[d]][b, ] <- .efa_pooled_communalities(rot_b$Lambda, rot_b$Phi)
          } else {
            boot_h2[[d]][b, ] <- .efa_pooled_communalities(rot_b$Lambda, NULL)
          }
        } else {
          nonconv_procrustes[d] <- nonconv_procrustes[d] + 1L
        }
      } else {
        boot_h2[[d]][b, ] <- .efa_pooled_communalities(Lb_unrot, NULL)
      }

      if (!is.null(arr$residuals)) {
        boot_residuals[[d]][b, ] <- .efa_pooled_vec(arr$residuals[, , b])
      }
    }
  }

  # Insufficient implies some failed, so test it first and emit a single warning.
  if (any(B_use - boot_failures < 2L)) {
    cli::cli_warn("Too few valid bootstrap replicates remain after dropping failures; pooled bootstrap SEs/CIs could not be computed.",
                  class = "efa_pooled_boot_insufficient")
    return(NULL)
  }

  if (any(boot_failures > 0L)) {
    cli::cli_warn("Some bootstrap replicates failed in the component EFAs; pooled bootstrap SEs/CIs are based on the valid replicates.",
                  class = "efa_pooled_boot_failed")
  }

  if (any(nonconv_procrustes > 0L)) {
    cli::cli_warn("Some bootstrap Procrustes rotations did not converge; pooled bootstrap SEs/CIs are based on the valid aligned replicates.",
                  class = "efa_pooled_boot_nonconv")
  }

  pool_unrot <- .efa_pooled_rubin_pool(q_unrot, boot_unrot, alpha = alpha)
  pool_h2 <- .efa_pooled_rubin_pool(
    q_h2, boot_h2, alpha = alpha,
    est_override = as.vector(h2)
  )
  pool_residuals <- .efa_pooled_rubin_pool(
    q_residuals, boot_residuals, alpha = alpha,
    est_override = .efa_pooled_vec(residuals)
  )

  # Reshapers for the two rectangular families pooled below (loading-shaped and
  # correlation-matrix-shaped); Phi is expanded from its vech instead.
  as_loading <- \(v) matrix(v, nrow = p_vars, ncol = k)

  unrot_res <- .efa_pooled_rubin_result(pool_unrot, as_loading, loading_dimnames)
  residual_res <- .efa_pooled_rubin_result(
    pool_residuals, \(v) matrix(v, nrow = p_vars, ncol = p_vars),
    dimnames(as.matrix(residuals))
  )

  SE <- list(
    communalities = pool_h2$se,
    unrot_loadings = unrot_res$se,
    residuals = residual_res$se
  )

  CI <- list(
    communalities = list(lower = pool_h2$ci$lower, upper = pool_h2$ci$upper),
    unrot_loadings = unrot_res$ci,
    residuals = residual_res$ci
  )

  arrays <- list(
    unrot_loadings = boot_unrot,
    h2 = boot_h2,
    residuals = boot_residuals
  )

  # Same canonical-name/alias split as the analytic route above.
  SE$h2 <- SE$communalities; CI$h2 <- CI$communalities

  MI <- list(
    unrot_loadings = list(RIV = pool_unrot$RIV, FMI = pool_unrot$FMI, df = pool_unrot$df),
    communalities = list(RIV = pool_h2$RIV, FMI = pool_h2$FMI, df = pool_h2$df),
    residuals = list(RIV = pool_residuals$RIV, FMI = pool_residuals$FMI, df = pool_residuals$df),
    # Per-imputation alignment outcomes. `boot_failures` are replicates the
    # component EFA could not fit (NA-filled, dropped before rotation);
    # `bootstrap_rotation_failures` are replicates whose Procrustes alignment to
    # the MI target did not produce a valid rotation. The valid count subtracts
    # both, so it never overstates the replicates that actually entered the pool.
    bootstrap_source_failures = boot_failures,
    bootstrap_rotation_failures = nonconv_procrustes,
    bootstrap_rotation_valid = B_use - boot_failures - nonconv_procrustes
  )

  if (rotation_type != "none") {
    pool_rot <- .efa_pooled_rubin_pool(q_rot, boot_rot, alpha = alpha)
    rot_res <- .efa_pooled_rubin_result(pool_rot, as_loading, loading_dimnames)
    SE$rot_loadings <- rot_res$se
    CI$rot_loadings <- rot_res$ci
    arrays$rot_loadings <- boot_rot
    MI$rot_loadings <- list(RIV = pool_rot$RIV, FMI = pool_rot$FMI, df = pool_rot$df)
  }

  if (oblique) {
    pool_phi <- .efa_pooled_rubin_pool(q_phi, boot_phi, alpha = alpha)
    phi_res <- .efa_pooled_rubin_result(
      pool_phi, \(v) .unvech(v, k), dimnames(as.matrix(phis[[1]]))
    )
    SE$Phi <- phi_res$se
    CI$Phi <- phi_res$ci
    arrays$Phi <- boot_phi
    MI$Phi <- list(RIV = pool_phi$RIV, FMI = pool_phi$FMI, df = pool_phi$df)

    pool_structure <- .efa_pooled_rubin_pool(
      q_structure, boot_structure, alpha = alpha,
      est_override = .efa_pooled_vec(mean_structure_loadings)
    )
    structure_res <- .efa_pooled_rubin_result(
      pool_structure, as_loading, loading_dimnames
    )
    SE$Structure <- structure_res$se
    CI$Structure <- structure_res$ci
    arrays$Structure <- boot_structure
    MI$Structure <- list(
      RIV = pool_structure$RIV,
      FMI = pool_structure$FMI,
      df = pool_structure$df
    )
  }

  ## Fit-index summaries ------------------------------------------------------
  ## fit_indices_descriptive: Rubin-Wald MI summaries of the per-imputation
  ## bootstrap fit indices. Chi-square-type fit is pooled with the D2 rule and its
  ## reference distribution supplies the point-estimate uncertainty, so no
  ## bootstrap-percentile interval is constructed for the pooled fit statistic.

  fit_q_names <- Reduce(intersect, lapply(fits, function(x) names(x$fit_indices)))
  fit_q_names <- fit_q_names[vapply(fits[[1]]$fit_indices[fit_q_names], function(x) {
    is.numeric(x) && length(x) == 1L
  }, logical(1))]

  # The bootstrap does not aggregate every fit index -- the RMSEA confidence bounds have no
  # per-replicate value, because the replicate fits run with ci = FALSE -- so the replicate
  # matrix is narrower than the point-estimate fit-index list. Take the pooled quantities from
  # the columns each fit's replicate matrix actually carries, rather than assuming the two line
  # up.
  boot_fit_arrs <- lapply(fits, function(x) {
    fit_arr <- x$replicates$fit_indices
    if (is.null(fit_arr)) return(NULL)
    fit_arr <- as.matrix(fit_arr)
    if (is.null(colnames(fit_arr))) {
      # An object saved before the replicate matrix was labelled carries no dimnames; there the
      # fit-index names are the only available labels, and only if they match by position.
      nms <- names(x$fit_indices)
      if (length(nms) != ncol(fit_arr)) return(NULL)
      colnames(fit_arr) <- nms
    }
    fit_arr
  })
  fit_q_names <- intersect(
    fit_q_names,
    Reduce(intersect, lapply(boot_fit_arrs, function(a) {
      if (is.null(a)) character(0) else colnames(a)
    }))
  )

  can_describe_fit_boot <- length(fit_q_names) > 0L &&
    all(vapply(boot_fit_arrs, function(a) {
      !is.null(a) && nrow(a) >= B_use
    }, logical(1)))

  if (can_describe_fit_boot) {
    q_fit <- lapply(fits, function(x) unlist(x$fit_indices[fit_q_names]))
    boot_fit <- lapply(boot_fit_arrs, function(a) {
      a[seq_len(B_use), fit_q_names, drop = FALSE]
    })

    pool_fit_desc <- .efa_pooled_rubin_pool(q_fit, boot_fit, alpha = alpha)
    SE$fit_indices_descriptive <- stats::setNames(pool_fit_desc$se, fit_q_names)
    CI$fit_indices_descriptive <- list(
      lower = stats::setNames(pool_fit_desc$ci$lower, fit_q_names),
      upper = stats::setNames(pool_fit_desc$ci$upper, fit_q_names)
    )
    arrays$fit_indices_descriptive <- boot_fit
    MI$fit_indices_descriptive <- list(
      RIV = stats::setNames(pool_fit_desc$RIV, fit_q_names),
      FMI = stats::setNames(pool_fit_desc$FMI, fit_q_names),
      df = stats::setNames(pool_fit_desc$df, fit_q_names)
    )
  }

  list(
    SE = SE,
    CI = CI,
    replicates = arrays,
    MI = MI,
    n_boot = B_use
  )
}

#' Plot HULL object
#'
#' Plot method showing a summarized output of the \code{\link{HULL}} function
#'
#' @param x list of class HULL. An output from the \code{\link{HULL}} function.
#' @param ... not used.
#'
#' @export
#' @method plot HULL
#'
#' @examples
#' \donttest{
#' x <- HULL(test_models$baseline$cormat, N = 500, method = "ML")
#' plot(x)
#'}
plot.HULL <- function(x, ...) {

  gof <- x$settings$gof
  method <- x$settings$method
  sol_CAF <- as.data.frame(x$solutions_CAF)
  sol_CFI <- as.data.frame(x$solutions_CFI)
  sol_RMSEA <- as.data.frame(x$solutions_RMSEA)
  n_fac_CAF <- x$n_fac_CAF
  n_fac_CFI <- x$n_fac_CFI
  n_fac_RMSEA <- x$n_fac_RMSEA

  if(!all(is.na(sol_CAF)) && !all(is.na(sol_CAF[, 4]))){

    .plot_HULL_helper(sol_CAF, method = method, gof = "CAF")

  }

  if(!all(is.na(sol_CFI)) && !all(is.na(sol_CFI[, 4]))){

    .plot_HULL_helper(sol_CFI, method = method, gof = "CFI")

  }

  if(!all(is.na(sol_RMSEA)) && !all(is.na(sol_RMSEA[, 4]))){

    .plot_HULL_helper(sol_RMSEA, method = method, gof = "RMSEA")

  }


}

.plot_HULL_helper <- function(sol, method, gof){
  # Sol = solution output for given fit index
  # Method = estimation method of EFA (PAF, ULS, ML)
  # gof = the fit index
  # nfac_suggested = number of factors suggested by HULL with given fit index

  df <- sol$df
  p_df <- pretty(df)

  fit_ind <- sol[[gof]]
  p_fit_ind <- pretty(c(0, fit_ind, max(fit_ind) * 1.25))
  st <- sol$st
  nfac_tested <- sol$nfactors

  graphics::plot.new()
  graphics::plot.window(xlim = c(min(p_df), max(p_df)),
                        ylim = c(min(p_fit_ind), max(p_fit_ind)),
                        xaxs = "i")
  graphics::axis(1, p_df)
  graphics::axis(2, p_fit_ind, las = 1)
  graphics::mtext(expression(italic(df)), side = 1, line = 3, cex = 1.5, padj =-.5)
  graphics::mtext(expression(italic(f)), side = 2, line = 3, cex = 1.5, padj =.5)
  graphics::title(paste("Hull Method with", method, "estimation and", gof))

  graphics::points(df[is.na(st)], fit_ind[is.na(st)], pch = 16,
                   col = "darkgrey")
  graphics::points(df[!is.na(st)], fit_ind[!is.na(st)], pch = 16,
                   cex = 1.5, type = "b")
  graphics::points(df[which.max(st)], fit_ind[which.max(st)], pch = 16,
                   cex = 1.5)
  graphics::points(df[which.max(st)], fit_ind[which.max(st)], pch = 1,
                   cex = 2.5, col = "red")
  if (any(is.na(st))) {
    graphics::text(df[is.na(st)], fit_ind[is.na(st)],
                   nfac_tested[is.na(st)], pos = 1, col = "darkgrey")
  }

  graphics::text(df[!is.na(st) & nfac_tested != which.max(st) - 1],
                 fit_ind[!is.na(st) & nfac_tested != which.max(st) - 1],
                 nfac_tested[!is.na(st) & nfac_tested != which.max(st) - 1],
                 pos = 3, cex = 1.25)
  graphics::text(df[which.max(st)], fit_ind[which.max(st)],
                 nfac_tested[which.max(st)],
                 pos = 3, cex = 1.5, col = "red", font = 1, offset = .75)


}

#' Reliability and common-variance coefficients for a factor solution
#'
#' Computes model-based reliability coefficients for a factor solution:
#' McDonald's omega (total, hierarchical, and subscale), standardized Cronbach's
#' alpha, and the H index. For a bifactor solution, it also computes two
#' common-variance indices, ECV and PUC, for the general factor. The result is a
#' tidy, long-format table with one row per coefficient.
#'
#' The function reads many kinds of input: a Schmid-Leiman solution
#' ([efa_schmid_leiman()] or [psych::schmid()]), an oblique [efa_fit()]
#' (correlated-factors) solution, a `lavaan` fit (single-factor,
#' correlated-factors, second-order, or bifactor), a raw bifactor loading matrix,
#' an oblique pattern matrix given with its factor intercorrelations, or manually
#' supplied components.
#'
#' @details ## Coefficients
#'
#' The reliability coefficients are McDonald's omegas (McDonald, 1978, 1985,
#' 1999; Zinbarg et al., 2005, 2006, for omega hierarchical specifically),
#' standardized Cronbach's alpha (Cronbach, 1951), and the H index (construct
#' replicability; Hancock & Mueller, 2001).
#'
#' The omegas give the share of true score variance in a unit-weighted
#' composite. Omega total is the share due to all factors together. Omega
#' hierarchical is the share due to the general factor only. Omega subscale is
#' the share due to the group factors: for the whole scale, or for one specific
#' group factor in a subscale composite.
#'
#' Alpha is the standardized coefficient, computed from the correlation matrix
#' of the items. Where no such matrix is available -- for `lavaan` input, and
#' for components supplied without one -- alpha is computed from the
#' model-implied correlation matrix instead, so it then reflects the fitted
#' model rather than the raw data.
#'
#' The H index is the reliability of an optimally weighted composite. A low
#' value means the factor is not well defined by its indicators.
#'
#' All of these coefficients describe the raw sum of the variables as fitted,
#' without reverse-coding. If some items are keyed in the opposite direction --
#' for example, a reverse-worded item that was not reverse-scored -- they lower
#' that sum's true-score variance. The coefficients then look poor even though
#' the model fits well. The function warns when it detects this. Reverse-code
#' such items before fitting the solution (Flora, 2020).
#'
#' The sum these coefficients describe is not always the sum of the raw answers.
#' Polychoric and tetrachoric correlations describe the continuous latent
#' responses assumed to underlie ordinal answers; a `lavaan` fit that declares
#' its indicators `ordered` does the same. In that case the loadings, the
#' uniquenesses, the correlation matrix, and the coefficients are all on that
#' latent-response metric. They give the reliability of the unit-weighted sum of
#' the latent responses, which is not observed -- not the reliability of the
#' ordinal sum score the user actually computes from the answers. The two can
#' differ substantially, especially where the answers use few categories or are
#' strongly skewed. Green and Yang (2009) give an omega for the ordinal sum
#' score itself, computed from the fitted model and its thresholds; this
#' package does not compute it. Pearson correlations raise no such distinction,
#' because their metric is the answers as scored.
#'
#' Omega total is lower when a solution reproduces the observed correlations
#' poorly, because residual covariance does not count as true score.
#' [psych::omega()] computes the whole-scale omega total differently:
#' residually, from the observed total-score variance. The two agree when the
#' model reproduces the correlations exactly, and diverge otherwise.
#'
#' The three coefficients are not generally additive. Omega total need not
#' equal omega hierarchical plus omega subscale, except on the whole-scale row
#' under `variance = "sums_load"`.
#'
#' A single-factor solution is scored as such on every input route, and
#' returns omega total, alpha, and the H index. Alpha assumes essentially
#' tau-equivalent items, an assumption nested within a one-factor model -- so a
#' single factor is the case where reporting alpha is defensible, not merely
#' possible.
#'
#' The other coefficients are omitted because a single factor does not define
#' them: omega subscale is the variance due to the group factors, and there
#' are none; omega hierarchical would equal omega total, since the one factor
#' accounts for all common variance; and ECV and PUC would both be 1 by
#' construction, which reflects the number of factors in the model rather than
#' evidence of unidimensionality.
#'
#' Each coefficient answers a different question:
#'
#' - Omega hierarchical: can the total score be read as a measure of one
#'   construct?
#' - Omega subscale: does a subscale score add anything beyond the general
#'   factor?
#' - The H index: is a factor well defined by its indicators?
#' - ECV together with PUC: is a unidimensional model defensible?
#'
#' Alpha assumes essentially tau-equivalent items. Factor analysis instead
#' yields congeneric solutions, for which alpha is only a lower bound. For a
#' multidimensional scale, alpha is rarely the coefficient to report (Gignac,
#' 2014).
#'
#' Composite reliability and average variance extracted (AVE) are not among
#' the coefficients this function computes. Use `semTools::compRelSEM()` and
#' `semTools::AVE()` to compute them from a `lavaan` fit.
#'
#' The common-variance indices ECV and PUC (Bonifay et al., 2015; Reise et al.,
#' 2013; Rodriguez et al., 2016a, 2016b) describe the general factor, so they
#' are reported for the general factor only. ECV is the share of the common
#' variance explained by the general factor. PUC is the proportion of
#' correlations that reflect general-factor variance alone -- correlations
#' between indicators of different group factors. The higher the PUC, the more
#' the general factor resembles the single factor of a unidimensional model.
#'
#' ## Input
#'
#' `efa_reliability()` reads several kinds of input, illustrated in the
#' examples below.
#'
#' - **An oblique [efa_fit()] solution** is scored as the correlated-factors
#'   model it is. It has no general factor, so it omits the bifactor indices
#'   (omega hierarchical, ECV, and PUC). Its whole-scale row is labelled
#'   `"total"` rather than `"g"`.
#' - **An [efa_schmid_leiman()] solution** -- or a `schmid` object from
#'   [psych::schmid()], or a raw bifactor loading matrix with the general
#'   factor in its first column -- is scored as a bifactor solution, with the
#'   whole-scale row labelled `"g"`. Indicator-to-factor correspondences come
#'   from `factor_map` if supplied (see `factor_map` below). Without one, an
#'   [efa_schmid_leiman()] or [psych::schmid()] solution is mapped by each
#'   variable's strongest group loading. A raw bifactor matrix instead
#'   defaults to its exact zero pattern -- supply `factor_map` explicitly if
#'   that matrix has no exact zeros (e.g. an estimated rather than a target
#'   matrix).
#' - **A `lavaan` fit** -- single-factor, correlated-factors, second-order, or
#'   bifactor -- is scored per its structure. The structure is detected
#'   automatically, not taken from `g_name`. The general-factor coefficients
#'   need the latent factors to be uncorrelated: fit a bifactor model with
#'   `orthogonal = TRUE` (`lavaan`'s default leaves the factors correlated),
#'   and leave a second-order model's first-order factor covariances at zero.
#'   A fit whose factors correlate is rejected rather than scored as though
#'   they did not. `variance` is not used for `lavaan` input: its composite
#'   variances are always model-implied, computed separately per group for a
#'   multiple-group fit.
#' - **A pattern matrix with its `Phi`** -- for example, the loadings and
#'   `Phi` of an oblique [efa_fit()] solution -- is scored as the
#'   correlated-factors solution it represents. Uniquenesses are derived from
#'   the loadings under `Phi`, unless `u2` is given. Supply the pattern
#'   matrix, not the structure matrix -- the structure matrix would be read
#'   as a pattern too, giving the wrong result. Unlike a fitted solution, this
#'   pair carries no correlation matrix. `variance = "correlation"` then
#'   needs one, supplied in `cormat`, and errors without it. `variance =
#'   "sums_load"` needs none.
#'
#' A one-factor solution -- a one-column loading matrix, a single-factor
#' `lavaan` fit, a single-factor [efa_fit()] solution, or single-factor
#' components -- is scored the same way regardless of route (see Coefficients
#' above for what it returns). Its row is labelled with the factor's own name,
#' with `fac_names` if supplied, or with `"F1"` if neither names it; it is
#' never labelled `"g"`.
#'
#' Manually supplied components (`g_load`, `s_load`, `u2`, `var_names`, `Phi`)
#' follow the same reading rules as the matrix routes above. Do not pass a
#' correlation matrix as `s_load` (or as `model`, for the matrix routes) --
#' supply it as `cormat` instead.
#'
#' @param model an [efa_schmid_leiman()], `schmid` ([psych::schmid()]), [efa_fit()]
#'   (oblique), or `lavaan` object; a raw bifactor loading matrix (general factor
#'   first), the loading table of an [efa_schmid_leiman()] solution, or the pattern
#'   matrix of an oblique solution together with its `Phi`; or `NULL` to
#'   supply the components manually via `g_load`, `s_load`, `u2`, and `var_names`.
#' @param coefficients character. An optional subset of the coefficients to
#'   return, any of `"omega_total"`, `"omega_hierarchical"`, `"omega_subscale"`,
#'   `"alpha"`, `"H"`, `"ECV"`, and `"PUC"`. Default `NULL` returns all of them.
#' @param g_name character. The name of the general factor in the `lavaan`
#'   solution. Only needed for a `lavaan` second-order or bifactor fit. A fit in
#'   which every variable loads on a single factor has no general factor, and
#'   reads none. No other input is affected by it. Default is `"g"`.
#' @param group_names character. An optional vector of group names for a `lavaan`
#'   multiple-group fit. Its length must match the number of groups. Not used for
#'   any other input -- including a single-group `lavaan` fit, since every one of
#'   those is scored as a single ungrouped solution with no group label.
#' @param factor_map matrix. A logical or 0/1 matrix indicating which variable
#'   belongs to which group factor, with the same dimensions as the group
#'   loading matrix (cross-loadings are allowed). Match its columns to the
#'   group factors by position -- a map given in a different factor order than
#'   the solution still runs, but produces meaningless subscale coefficients.
#'   The function warns if a mapped item loads weakly on its assigned factor
#'   and more strongly on another one. If `NULL` (default), each variable is
#'   assigned to the group factor on which it loads most strongly. Not used
#'   for `lavaan` input.
#' @param variance character. The total-variance denominator for the
#'   coefficients. `"correlation"` (default) takes each composite's variance
#'   from the correlation matrix, giving the observed-score omega. `"sums_load"`
#'   uses the model-implied composite variance from the loadings and the
#'   uniquenesses instead; it needs no correlation matrix, so it is the way to
#'   score a bare loading matrix or manual components given without one.
#'   `lavaan` input fixes the convention: its composite variances are always
#'   model-implied, and include any freed residual covariance as well as the
#'   residual variances. Neither convention changes the metric the coefficients
#'   are on -- with polychoric or tetrachoric correlations, or an `ordered`
#'   `lavaan` fit, both describe the latent-response composite rather than the
#'   ordinal sum score (see Details).
#' @param var_names character. Subtest names in the row order of the loadings.
#'   Only needed when `model` is `NULL`.
#' @param fac_names character. An optional vector of group-factor names in the
#'   column order of the loadings. Taken from the input if `NULL`. A
#'   single-factor solution has no group factors; `fac_names` then labels its
#'   one factor instead. If neither `fac_names` nor the solution names that
#'   factor, it is labelled `"F1"`. Not used for `lavaan` input, whose factor
#'   labels come from the model syntax.
#' @param g_load numeric. General-factor loadings. Only needed when `model` is
#'   `NULL`.
#' @param s_load matrix. Group-factor loadings. Only needed when `model` is
#'   `NULL`.
#' @param u2 numeric. Uniquenesses. Only needed when `model` is `NULL`, or to
#'   override the communality-based default for a loading matrix. Under
#'   `variance = "correlation"`, the coefficients follow from the loadings and
#'   the correlation matrix alone, so `u2` only enters the check for an improper
#'   solution. Under `"sums_load"`, it is part of every composite's variance.
#' @param cormat matrix or data.frame. A correlation matrix used when
#'   `variance = "correlation"`. It must hold the same variables as the
#'   solution: named variables in a different order are reordered to match; a
#'   different set of variables, or a different number of them, is an error.
#'   The matrix must use the solution's own variable names -- for manually
#'   supplied components, that means the row names of `s_load`, not
#'   `var_names`, which only labels the output. Without row names, the
#'   variables are matched by position, so supply the matrix in the row order
#'   of the loadings. If `NULL`, it is taken from the input object, or
#'   reconstructed from `pattern` and `Phi` where possible. Supply the matrix
#'   on the same metric as the loadings: a polychoric or tetrachoric matrix
#'   keeps the coefficients on the latent-response metric, while a Pearson
#'   matrix given with loadings fitted to a polychoric one mixes the two (see
#'   Details).
#' @param pattern matrix. Pattern coefficients from a separate oblique solution, used
#'   with `Phi` to reconstruct a correlation matrix when `model` is `NULL`. Supply it for
#'   a Schmid-Leiman input, whose `s_load` holds the orthogonalized group loadings rather
#'   than the oblique ones. It is an alternative to `cormat`, not a supplement: giving
#'   both is an error.
#' @param Phi matrix. Factor intercorrelations. `NULL` (default) means
#'   uncorrelated group factors, as a Schmid-Leiman or bifactor solution has.
#'   It cannot be combined with a general factor: supplied together with a
#'   non-zero `g_load`, or with the loading table of an
#'   [efa_schmid_leiman()] solution, it is an error.
#'
#'   Supply it together with `s_load` (manually supplied components, with
#'   `g_load` zero throughout), or with a loading matrix of two or more
#'   factors in `model`, to score the input as a correlated-factors solution
#'   instead.
#'
#'   Without `Phi`, a loading matrix in `model` is read as a bifactor
#'   solution (general factor first). The one exception is a matrix that
#'   still carries the loading class [efa_fit()] gives its output -- a class
#'   that subsetting a matrix or reordering its rows drops. Such a matrix is
#'   rejected instead; supply `Phi` to score it as a correlated-factors
#'   solution.
#'
#'   With `Phi`, a matrix in `model` that carries no loading class is read as
#'   a correlated-factors solution, and a warning names this reading (drop
#'   `Phi` to read it as a bifactor matrix instead).
#'
#'   Given together with `pattern` instead, see `pattern` above. With a
#'   fitted solution already in `model`, `Phi` is ignored, with a warning,
#'   since that solution carries its own factor intercorrelations.
#'
#' @returns An object of class `efa_reliability`: a long-format data frame with
#'   one row per computed coefficient, with columns
#'   \item{coefficient}{the coefficient name (e.g. `"omega_total"`).}
#'   \item{level}{`"general"` for the general-factor row, `"total"` for the
#'     whole-scale row of a correlated-factors solution, and `"group"` for
#'     every other row. A single-factor solution has only the general-factor
#'     row.}
#'   \item{factor}{the factor label: `"g"` for the general factor of a
#'     solution with group factors. A single-factor solution's one factor
#'     takes its own name, or `"F1"` if neither the input nor `fac_names`
#'     names it. A correlated-factors solution has no general factor, so its
#'     first row is labelled `"total"` instead -- it describes the composite
#'     of every variable, not a factor of the model.}
#'   \item{group}{the group label, or `NA` for a single ungrouped solution.}
#'   \item{value}{the coefficient value.}
#'   Structurally undefined cells (for example, ECV and PUC on a group factor)
#'   are omitted. The object also carries a `settings` attribute (the
#'   total-variance convention used, and whether the solution has a general
#'   factor) and a `kind` attribute tagging each coefficient as a reliability
#'   coefficient or a common-variance index. It has a
#'   [print.efa_reliability()] method.
#'
#' @source McDonald, R. P. (1978). Generalizability in factorable domains: Domain
#'   validity and generalizability. Educational and Psychological Measurement, 38,
#'   75-79.
#' @source McDonald, R. P. (1985). Factor analysis and related methods. Hillsdale,
#'   NJ: Erlbaum.
#' @source McDonald, R. P. (1999). Test theory: A unified treatment. Mahwah,
#'   NJ: Erlbaum.
#' @source Cronbach, L. J. (1951). Coefficient alpha and the internal structure of
#'   tests. Psychometrika, 16, 297-334.
#' @source Gignac, G. E. (2014). On the inappropriateness of using items to
#'   calculate total scale score reliability via coefficient alpha for
#'   multidimensional scales. European Journal of Psychological Assessment, 30,
#'   130-139.
#' @source Flora, D. B. (2020). Your coefficient alpha is probably wrong, but which
#'   coefficient omega is right? A tutorial on using R to obtain better reliability
#'   estimates. Advances in Methods and Practices in Psychological Science, 3,
#'   484-501.
#' @source Green, S. B., & Yang, Y. (2009). Reliability of summed item scores using
#'   structural equation modeling: An alternative to coefficient alpha. Psychometrika,
#'   74, 155-167.
#' @source Zinbarg, R. E., Revelle, W., Yovel, I., & Li, W. (2005). Cronbach's
#'   alpha, Revelle's beta, and McDonald's omega H: Their relations with each
#'   other and two alternative conceptualizations of reliability. Psychometrika,
#'   70, 123-133.
#' @source Zinbarg, R. E., Yovel, I., Revelle, W., & McDonald, R. P. (2006).
#'   Estimating generalizability to a latent variable common to all of a scale's
#'   indicators: A comparison of estimators for omega H. Applied Psychological
#'   Measurement, 30, 121-144.
#' @source Hancock, G. R., & Mueller, R. O. (2001). Rethinking construct
#'   reliability within latent variable systems. In R. Cudeck, S. du Toit, & D.
#'   Sörbom (Eds.), Structural equation modeling: Present and future - A
#'   Festschrift in honor of Karl Jöreskog (pp. 195-216). Lincolnwood, IL:
#'   Scientific Software International.
#' @source Bonifay, W. E., Reise, S. P., Scheines, R., & Meijer, R. R. (2015). When
#'   are multidimensional data unidimensional enough for structural equation
#'   modeling? An evaluation of the DETECT multidimensionality index. Structural
#'   Equation Modeling, 22, 504-516.
#' @source Reise, S. P., Scheines, R., Widaman, K. F., & Haviland, M. G. (2013).
#'   Multidimensionality and structural coefficient bias in structural equation
#'   modeling: A bifactor perspective. Educational and Psychological Measurement,
#'   73, 5-26.
#' @source Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016a). Applying bifactor
#'   statistical indices in the evaluation of psychological measures. Journal of
#'   Personality Assessment, 98, 223-237.
#' @source Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016b). Evaluating
#'   bifactor models: Calculating and interpreting statistical indices.
#'   Psychological Methods, 21, 137-150.
#'
#' @family reliability coefficients
#' @seealso [efa_fit()] for the solution these are computed from, and [OMEGA()], the
#'   superseded function that returns these same coefficients in a wide, per-factor layout.
#'
#' @export
#'
#' @examples
#' ## From an oblique EFA (correlated-factors) solution. With no factor_map, each
#' ## item is auto-assigned to its highest-loading factor.
#' efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' efa_reliability(efa_mod)
#'
#' ## From a Schmid-Leiman solution, with an explicit indicator-to-factor map.
#' sl_mod <- efa_schmid_leiman(efa_mod, estimator = "PAF")
#' fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2
#' efa_reliability(sl_mod, factor_map = fc)
#'
#' ## Request a subset of the coefficients only.
#' efa_reliability(sl_mod, factor_map = fc,
#'                 coefficients = c("omega_total", "alpha"))
#'
#' ## From an oblique pattern matrix and its factor intercorrelations. This is
#' ## the same correlated-factors solution, and gives the same coefficients.
#' efa_reliability(efa_mod$rot_loadings, Phi = efa_mod$Phi,
#'                 cormat = test_models$baseline$cormat)
#'
#' ## From lavaan fits: a bifactor solution, and a correlated-factors one.
#' \donttest{
#' if (requireNamespace("lavaan", quietly = TRUE)) {
#' mod_cf <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
#'            F2 =~ V7 + V8 + V9 + V10 + V11 + V12
#'            F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
#' mod <- paste(mod_cf, 'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 +
#'                            V11 + V12 + V13 + V14 + V15 + V16 + V17 + V18',
#'              sep = "\n")
#' fit <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
#'                    sample.nobs = 500, estimator = "ml", orthogonal = TRUE)
#' print(efa_reliability(fit, g_name = "g"))
#'
#' ## No general factor: omega hierarchical, ECV, and PUC are omitted.
#' fit_cf <- lavaan::cfa(mod_cf, sample.cov = test_models$baseline$cormat,
#'                       sample.nobs = 500, estimator = "ml")
#' efa_reliability(fit_cf)
#' }
#' }
#'
efa_reliability <- function(model = NULL,
                            coefficients = NULL,
                            g_name = "g",
                            group_names = NULL,
                            factor_map = NULL,
                            variance = c("correlation", "sums_load"),
                            var_names = NULL,
                            fac_names = NULL,
                            g_load = NULL,
                            s_load = NULL,
                            u2 = NULL,
                            cormat = NULL,
                            pattern = NULL,
                            Phi = NULL) {

  variance <- .match_arg_ci(variance)
  checkmate::assert_string(g_name)
  checkmate::assert_character(group_names, null.ok = TRUE)

  # Validate the coefficient selector against the surfaced menu.
  menu <- .reliability_registry()$coefficient
  if (!is.null(coefficients)) {
    checkmate::assert_character(coefficients, any.missing = FALSE, min.len = 1L)
    bad <- setdiff(coefficients, menu)
    if (length(bad) > 0) {
      cli::cli_abort(
        c("Unknown {.arg coefficients}: {.val {bad}}.",
          "i" = "Choose from {.val {menu}}."),
        class = "efa_reliability_bad_coefficients"
      )
    }
  }

  # Read by both argument reports below, the second of which covers a `NULL` `model` too.
  lavaan_fit <- inherits(model, "lavaan")

  # These arguments describe manually supplied components; a solution passed in `model`
  # carries its own and the adapters never read them. Report them rather than drop them: a
  # user who expects a `Phi` to define the group factors' correlations, or a `u2` to replace
  # the solution's own, would otherwise get an orthogonal or unchanged solution's
  # coefficients with nothing to say why.
  if (!is.null(model)) {
    # Two of them are read by a matrix `model` and so are not reported for one: `u2`, as an
    # override of the uniquenesses the adapter would derive, and `Phi`, which a matrix never
    # merely ignores -- it makes the columns of an oblique pattern matrix correlated factors,
    # and with a matrix read as a hierarchy it is refused rather than dropped, where the
    # dispatch can say which of the two readings was taken. A `lavaan` fit reads neither the
    # correspondences nor a correlation matrix -- it is scored per group from its own -- so
    # those two are discarded there as well, and pointing at them would send the user
    # nowhere.
    unused <- c("g_load", "s_load", "var_names", "pattern")
    if (!is.matrix(model)) unused <- c(unused, "Phi", "u2")
    if (lavaan_fit) unused <- c(unused, "cormat", "factor_map")
    given <- unused[!vapply(mget(unused), is.null, logical(1))]
    if (length(given) > 0) {
      cli::cli_warn(
        c("{.arg {given}} {?is/are} ignored for a fitted solution.",
          "i" = "{cli::qty(given)}{?It/They} describe{?s/} a solution supplied through the components; {.arg model} carries its own.",
          "i" = if (lavaan_fit) {
            "A {.cls lavaan} fit is scored per group from its own loadings, uniquenesses, and model-implied composite variances."
          } else {
            "Supply {.arg cormat} to score the solution against an observed correlation matrix, and {.arg factor_map} to set the indicator-to-factor correspondences."
          }),
        class = "efa_reliability_args_ignored"
      )
    }
  }

  # `fac_names` and `group_names` label the result rather than describe a solution, so the
  # reason above -- that `model` carries its own components -- is not theirs: each is read on
  # one kind of input only. `fac_names` names the factors of every solution but a `lavaan`
  # fit, whose factor labels come from the model syntax; `group_names` names the blocks of a
  # `lavaan` multiple-group fit, the only input scored as more than one solution. Reported for
  # the same reason the components are: a caller who labelled their factors or their groups
  # would otherwise get the input's own labels back with nothing to say why -- and a
  # `fac_names` of the wrong length is not even refused for a `lavaan` fit, the length check
  # below belonging to the routes that read it. Both are checked outside the block above,
  # which a `NULL` `model` does not enter although it reads no `group_names` either.
  if (lavaan_fit && !is.null(fac_names)) {
    cli::cli_warn(
      c("{.arg fac_names} is not used for a {.cls lavaan} fit.",
        "i" = "Its factor labels come from the model syntax; rename the latent variables there to change them."),
      class = "efa_reliability_arg_not_applicable"
    )
  }
  if (!is.null(group_names)) {
    # A single-group `lavaan` fit reads the labels no more than a non-`lavaan` input does:
    # .rel_adapt_lavaan() checks their length and the result builder then drops them, one
    # block needing no key, so its `group` column stays NA like every ungrouped solution's.
    # Keyed on the group count rather than on the class for exactly that reason. lavaan is
    # asked for the count only when there are labels to place.
    if (lavaan_fit) .require_lavaan()
    multigroup <- lavaan_fit && lavaan::lavInspect(model, "ngroups") > 1L
    if (!multigroup) {
      cli::cli_warn(
        c("{.arg group_names} is only used for a {.cls lavaan} multiple-group fit.",
          "i" = "Every other input is scored as a single ungrouped solution, whose {.field group} column is {.code NA}."),
        class = "efa_reliability_arg_not_applicable"
      )
    }
  }

  # Dispatch to the right adapter, then score each factor solution through the
  # shared reliability core (add_rel = TRUE so standardized alpha is available;
  # the registry drops the core's CR/AVE columns). lavaan input is scored per
  # group with the model-implied composite variance.
  no_general <- FALSE

  if (inherits(model, "lavaan")) {

    .require_lavaan()
    adapt <- .rel_adapt_lavaan(model, g_name = g_name, group_names = group_names)
    used_variance <- adapt$variance
    no_general <- isTRUE(adapt$correlated)

    # Distinct group labels: the group column keys each block of the tidy result,
    # so duplicate labels would silently merge groups.
    if (length(adapt$groups) > 1L && anyDuplicated(adapt$group_names)) {
      cli::cli_abort(
        c("{.arg group_names} must be unique.",
          "i" = "Duplicate labels would merge groups in the result."),
        class = "efa_reliability_group_names"
      )
    }

    if (isTRUE(adapt$higher_order)) {
      cli::cli_inform(
        c("i" = "The specified general factor is a second-order factor; coefficients are computed on the Schmid-Leiman transformed second-order solution."),
        class = "efa_reliability_g_second_order"
      )
    }

    if (isTRUE(adapt$few_loadings)) {
      cli::cli_inform(
        c("i" = "Some variables have fewer than two loadings; did you enter a bifactor model? Provide a bifactor model, a second-order model, or a single-factor model."),
        class = "efa_reliability_few_loadings"
      )
    }

    # No variable loads on more than one factor, so the fit carries no general factor over
    # and above its factors. Said rather than left to be read off a missing column: a user
    # who fitted what they think of as a bifactor model, or who named `g_name`, would
    # otherwise only see a table without the general-factor coefficients.
    if (no_general) {
      cli::cli_inform(
        c("i" = "Each variable loads on one factor only, so the fit is scored as a correlated-factors solution; omega hierarchical, ECV, and PUC are not defined for it and are omitted."),
        class = "efa_reliability_no_general_factor"
      )
    }

    mats <- vector("list", length(adapt$groups))
    informed_single <- FALSE
    for (i in seq_along(adapt$groups)) {
      grp <- adapt$groups[[i]]
      if (isTRUE(grp$single)) {
        if (!informed_single) {
          .rel_inform_single_factor()
          informed_single <- TRUE
        }
        # A single factor is scored through the core on the spec the adapter normalized it
        # to, as every other group is; the coefficients it does not define are then dropped,
        # exactly as on the other input routes.
        mats[[i]] <- .rel_drop_single_factor(
          .reliability_core(grp, used_variance, add_ind = TRUE, add_rel = TRUE,
                            arg = "factor_map"),
          fac_names = grp$fac_label)
      } else {
        mats[[i]] <- .reliability_core(grp, used_variance, add_ind = TRUE,
                                       add_rel = TRUE, arg = "factor_map")
        # A correlated-factors fit defines only some of the coefficients the core
        # computes; the rest are dropped, per group, as on the non-lavaan routes.
        if (no_general) mats[[i]] <- .rel_drop_general(mats[[i]])
      }
    }

    # One matrix (single ungrouped fit) or a named list (multiple groups).
    x <- if (length(mats) == 1L) mats[[1]] else
      stats::setNames(mats, adapt$group_names)

  } else {

    spec <- .rel_dispatch_spec(model, factor_map = factor_map,
                               var_names = var_names, fac_names = fac_names,
                               g_load = g_load, s_load = s_load, u2 = u2,
                               cormat = cormat, pattern = pattern, Phi = Phi)

    # A solution with exactly one factor is scored as the single factor it is, whichever slot
    # that factor arrived in -- as the general factor of a one-column bifactor matrix, or as
    # the only group factor of a one-factor `efa_fit()` solution or of manual components with
    # a zero `g_load`. Normalizing them to one spec is what makes every route report the same
    # coefficients for the same solution (see .rel_single_factor_spec).
    single_spec <- .rel_single_factor_spec(spec)
    single <- !is.null(single_spec)

    # One label per factor, checked once the solution is in hand and its factors counted.
    # `fac_names` labels the group-factor rows, of which a single-factor solution has one --
    # its own factor, whichever slot it arrived in -- and any other solution has one per
    # group loading column; the general-factor row is labelled "g" and takes none. Nothing
    # downstream can say so: too few or too many reach `rownames()` on the coefficient
    # matrix and fail there in base R's terms, naming neither the argument nor the count it
    # needed, and a single-factor solution does not fail at all -- .rel_single_factor_label()
    # reads a vector that is not of length one as no name and falls back to "F1", so the
    # labels the caller asked for are dropped without a word.
    n_fac <- if (single) 1L else ncol(as.matrix(spec$s_load))
    if (!is.null(fac_names) && length(fac_names) != n_fac) {
      cli::cli_abort(
        c("{.arg fac_names} must have one name per factor.",
          "x" = "It has {length(fac_names)} name{?s}, but the solution has {n_fac} factor{?s}.",
          "i" = if (single) {
            "A single-factor solution has no group factors; {.arg fac_names} labels its one factor."
          } else {
            "Name the group factors in the column order of the loadings; the general factor is labelled {.val g}."
          }),
        class = "efa_reliability_fac_names_length"
      )
    }

    single_fac_names <- NULL
    if (single) {
      # Read off the spec as it arrived, before it is replaced: the normalized one has no
      # group factor left to carry the name the input gave this one.
      single_fac_names <- spec$fac_names
      spec <- single_spec
    }

    # An all-zero general factor marks a solution with no general factor -- an oblique EFA,
    # or a bifactor/manual input whose general column is zero. It says nothing about
    # whether the group factors are correlated: only a spec that carries the factor
    # intercorrelations describes correlated ones, and the coefficients account for them
    # in either variance convention. A single factor is not such a solution: it is itself
    # the general factor, and the normalization above has put it there.
    no_general <- !single && isTRUE(all(spec$g_load == 0))

    # Only correlation mode needs a correlation matrix, whether or not the solution has a
    # general factor. The model-implied "sums_load" variance needs none, and is correct for
    # every solution the adapters produce -- it reads the factor intercorrelations where the
    # spec carries them. Abort with an actionable message when correlation mode is asked for
    # and no matrix is available (the adapters supply one when they can; an SL object from a
    # bare loading matrix, or manual components without cormat/pattern/Phi, leave it NULL)
    # rather than divide the omega denominators by sum(NULL) = 0 and return non-finite values.
    if (variance == "correlation" && is.null(spec$cormat)) {
      cli::cli_abort(
        c("A correlation matrix is required for {.code variance = \"correlation\"} but none is available.",
          "i" = "Supply {.arg cormat} (or {.arg pattern} and {.arg Phi}), or use {.code variance = \"sums_load\"}."),
        class = "efa_reliability_need_cormat"
      )
    }

    if (single) .rel_inform_single_factor()

    used_variance <- variance
    x <- .reliability_core(spec, variance, add_ind = TRUE, add_rel = TRUE,
                           arg = "factor_map")

    # A correlated-factors solution (no general factor) does not identify the
    # general-factor decomposition, and a single-factor one leaves the general-factor
    # decomposition with nothing to decompose, so each drops the coefficients it does not
    # define (see .rel_drop_general() and .rel_drop_single_factor() for which, and why).
    # NA-ing the cells lets the result builder omit them as any other undefined coefficient.
    if (single) {
      x <- .rel_drop_single_factor(x, fac_names = single_fac_names)
    } else if (no_general) {
      x <- .rel_drop_general(x)
    }

  }

  res <- .reliability_result(x, settings = list(variance = used_variance,
                                                no_general = no_general))

  if (!is.null(coefficients)) {
    # Note any requested coefficient this solution does not define (e.g. omega
    # hierarchical for a correlated-factors EFA, or omega subscale for a single-factor
    # solution), so a reduced or empty selection is not silent.
    undefined <- setdiff(coefficients, unique(res$coefficient))
    if (length(undefined) > 0) {
      cli::cli_inform(
        c("i" = "{cli::qty(undefined)}The coefficient{?s} {.val {undefined}} {?is/are} not defined for this solution and {?is/are} omitted."),
        class = "efa_reliability_coef_undefined"
      )
    }
    res <- .reliability_select(res, coefficients)
  }

  res

}


# Read a loading matrix and its factor intercorrelations as one correlated-factors
# solution, returning the normalized reliability spec. Shared by the two matrix routes
# that reach this reading, so a pattern matrix gives the same coefficients whether or not
# it still carries the loading class.
.rel_read_pattern <- function(model, Phi, u2, factor_map, cormat, fac_names) {
  # The loadings and any supplied uniquenesses get the checks the bifactor-matrix route
  # applies, and for the same reasons.
  .assert_args({
    checkmate::assert_numeric(model, any.missing = FALSE, finite = TRUE)
    if (!is.null(u2)) {
      checkmate::assert_numeric(u2, any.missing = FALSE, finite = TRUE,
                                len = nrow(model))
    }
  })
  L <- as.matrix(unclass(model))
  # Checked here because the uniquenesses are derived from it just below, which needs
  # it conformable with the loadings; .rel_adapt_manual() checks it again on its own.
  .rel_check_phi(Phi, ncol(L))
  # 1 - diag(L Phi L'), the communalities the oblique pattern implies under these
  # factor correlations, unless the caller supplied uniquenesses of their own. The
  # same expression the oblique `efa_fit()` route uses, so both reach the same ones.
  if (is.null(u2)) u2 <- 1 - diag(L %*% Phi %*% t(L))
  var_names <- rownames(L)
  if (is.null(var_names)) var_names <- paste0("V", seq_len(nrow(L)))
  if (is.null(fac_names)) {
    fac_names <- colnames(L)
    if (is.null(fac_names)) fac_names <- paste0("F", seq_len(ncol(L)))
  }
  # `pattern` is reported as ignored for any `model`, so it is not passed on: here
  # `Phi` describes the loading matrix that was given, not a separate solution.
  .rel_adapt_manual(g_load = rep(0, nrow(L)), s_load = L, u2 = u2,
                    var_names = var_names, factor_corres = factor_map,
                    type = "psych", cormat = cormat, Phi = Phi,
                    fac_names = fac_names)
}


# Dispatch a non-lavaan input to its front-end adapter, returning the normalized
# reliability spec. Mirrors OMEGA()'s model dispatch, adding the oblique EFA and
# raw bifactor-matrix paths that OMEGA does not expose. The internal adapters take a
# `type` that gates how the item-to-factor map is derived when `factor_map` is
# absent; efa_reliability always auto-maps to the highest-loading factor (adapter
# type "psych"), so a supplied map is used and an omitted one is derived.
.rel_dispatch_spec <- function(model, factor_map, var_names, fac_names,
                               g_load, s_load, u2, cormat, pattern, Phi) {

  if (inherits(model, "SL")) {
    .rel_adapt_SL(model, factor_corres = factor_map, type = "psych",
                  cormat = cormat, fac_names = fac_names)
  } else if (inherits(model, "schmid")) {
    .rel_adapt_schmid(model, factor_corres = factor_map, type = "psych",
                      cormat = cormat, fac_names = fac_names)
  } else if (inherits(model, "EFA")) {
    .rel_adapt_efa(model, factor_corres = factor_map, type = "psych",
                   cormat = cormat, fac_names = fac_names)
  } else if (is.matrix(model)) {
    if (!is.numeric(model)) {
      cli::cli_abort(
        c("A matrix {.arg model} must be a numeric bifactor loading matrix (general factor in the first column).",
          "i" = "Enter numeric loadings, or pass a fitted model object instead."),
        class = "efa_reliability_bad_matrix"
      )
    }
    # Recorded before the Schmid-Leiman branch below replaces `model` with its factor
    # columns, which drops the class that says so. A Schmid-Leiman table states that it is
    # a hierarchy; a matrix that carries no loading class merely fails to state anything,
    # which is not the same evidence, and the two are told apart where `Phi` arrives.
    declared_hierarchy <- inherits(model, c("efa_sl_loadings", "SLLOADINGS"))
    if (declared_hierarchy) {
      # A Schmid-Leiman loading table is a bifactor matrix plus the communality and
      # uniqueness columns it prints. Score it as such, dropping those two -- they would
      # otherwise be read as two more group factors -- and taking the uniquenesses from
      # the table rather than recomputing them from the loadings.
      sl <- unclass(model)
      nms <- colnames(sl)
      # The table is always built as the general factor, then the group factors, then h2
      # and u2, so an unlabelled one is read off that layout rather than left with its
      # two display columns scored as group factors.
      if (is.null(nms) && ncol(sl) > 3L) {
        display <- seq_len(ncol(sl)) > ncol(sl) - 2L
        u2_col <- ncol(sl)
      } else {
        # A table too short to carry that layout keeps every column, as before.
        if (is.null(nms)) nms <- character(ncol(sl))
        display <- nms %in% c("h2", "u2")
        u2_col <- match("u2", nms)
      }
      if (is.null(u2) && !is.na(u2_col)) u2 <- sl[, u2_col]
      model <- sl[, !display, drop = FALSE]
      # An estimated Schmid-Leiman table carries no structural zeros, so the raw-matrix
      # default -- every non-zero group loading -- would assign every variable to every
      # group factor. Derive the correspondences the same way the Schmid-Leiman object
      # does instead, so both routes to one solution report the same coefficients.
      if (is.null(factor_map)) {
        factor_map <- .rel_map(model[, -1, drop = FALSE], type = "psych")
      }
    } else if (inherits(model, c("efa_loadings", "LOADINGS"))) {
      # An ordinary loading table is a pattern or structure matrix: its first column is
      # simply the first factor, not a general factor. Read as a bifactor matrix it
      # yields plausible but meaningless coefficients.
      #
      # With `Phi` it is a complete correlated-factors solution -- the oblique pattern and
      # the correlations of the factors it belongs to -- and is scored as one, through the
      # same components the manual route takes for such a solution. Without it nothing says
      # the columns are correlated factors rather than a hierarchy, so a matrix of two or
      # more factors is refused and the hierarchy asked for instead. The class states that
      # the matrix is such a solution's pattern, so reading it as one overrides nothing and
      # is not reported, where the same reading of a matrix that states nothing is.
      if (!is.null(Phi)) {
        return(.rel_read_pattern(model, Phi = Phi, u2 = u2, factor_map = factor_map,
                                 cormat = cormat, fac_names = fac_names))
      }
      # A single column is the one case in which there is nothing to confuse: one factor is
      # no hierarchy, and a solution with one factor has no factor intercorrelations to
      # supply, so neither remedy the refusal offers is one its caller could take. It falls
      # through to the bare-matrix route below, which reads a one-column matrix as a general
      # factor with no group factors; the front end then rewrites that spec into the
      # single-factor one (.rel_single_factor_spec), as it does for every other route that
      # can carry one factor, so the classed matrix, the unclassed one, and the one-factor
      # `efa_fit()` object it came from all reach the same coefficients.
      if (ncol(model) > 1L) {
        cli::cli_abort(
          c("{.arg model} is a loading matrix of an ordinary factor solution, not a bifactor loading matrix.",
            "i" = "Pass the {.fn efa_fit} object itself for a correlated-factors solution, or, if this is its pattern matrix, supply the matching factor intercorrelations in {.arg Phi} to score it as one.",
            "i" = "To read it as a hierarchy instead, transform it with {.fn efa_schmid_leiman} first.",
            "i" = "If it really is a bifactor solution, pass {.code unclass(x)} to read its first column as the general factor."),
          class = "efa_reliability_pattern_matrix"
        )
      }
    } else if (!anyNA(model) && .is_cormat(model)) {
      # `.is_cormat()` aborts on a missing value in a matrix that is otherwise shaped like a
      # correlation matrix, and does so in terms of its own argument, hence the `anyNA()`
      # guard: an incomplete matrix is named by the loading check below rather than by a
      # misdirected error.
      #
      # A correlation matrix is square, symmetric, and has a unit diagonal at once; a
      # bifactor loading matrix is essentially never all three, since a unit diagonal would
      # mean every variable loads exactly 1 on one factor. Read as loadings it takes the
      # first variable's correlations for general-factor loadings and returns plausible but
      # meaningless coefficients, so it is refused where the mistake can still be named.
      cli::cli_abort(
        c("{.arg model} is a correlation matrix, not a bifactor loading matrix.",
          "i" = "Fit a solution first, e.g. with {.fn efa_fit} or {.fn efa_schmid_leiman}, and pass that.",
          "i" = "To score a solution against these correlations, pass the matrix as {.arg cormat}."),
        class = "efa_reliability_cormat_as_model"
      )
    }
    # Everything still here is read as a hierarchy unless `Phi` says otherwise: a bare
    # bifactor loading matrix, or the loading table of a Schmid-Leiman solution reduced to
    # its factor columns above. Read as one, its general factor is the first column and its
    # group factors are orthogonal by construction, so there are no correlated group factors
    # for `Phi` to be the correlation matrix of, and the general and group parts would no
    # longer partition the composite.
    # A Schmid-Leiman table states that it is a hierarchy, so the combination contradicts
    # itself and is refused, as it is through the components. A matrix that carries no
    # loading class states nothing: the class is not a reliable signal here, since `[`,
    # `rbind()` and a round-trip through a data frame all drop it while the argument stays,
    # so a caller who subsets an oblique pattern and keeps its `Phi` holds a complete
    # correlated-factors solution in two objects. `Phi` settles the reading for that matrix
    # -- the group factors of a hierarchy are orthogonal and have no factor correlations to
    # supply -- so it is read as the pattern it can only be. The reading is reported rather
    # than taken quietly, because the two give different coefficients: a hierarchy also
    # defines omega hierarchical, the ECV and the PUC, which a correlated-factors solution
    # does not, and a caller who meant the bifactor reading would otherwise see them absent
    # with nothing to say why.
    #
    # One column is the exception this route makes throughout: one factor is not a hierarchy,
    # and the only correlation matrix its factor can have is the 1 by 1 identity, which says
    # nothing either way. It is checked and then changes nothing, as on the classed route.
    if (!is.null(Phi)) {
      # Before the conformability check, so a Schmid-Leiman table is told that its own
      # reading refuses `Phi` rather than that this `Phi` is the wrong shape: no `Phi` of
      # any shape belongs with one, and naming the shape would send the caller to fix it.
      if (ncol(model) > 1L && declared_hierarchy) {
        cli::cli_abort(
          c("{.arg Phi} describes correlated group factors, which the loading table of a Schmid-Leiman solution does not have.",
            "i" = "To score the pattern matrix of an oblique solution, pass the {.fn efa_fit} object, or supply the components: the pattern in {.arg s_load}, a {.arg g_load} that is zero throughout, and {.arg u2} and {.arg var_names}, with this {.arg Phi}.",
            "i" = "To reconstruct a correlation matrix for a hierarchy instead, supply its components together with the {.arg pattern} of the separate oblique solution {.arg Phi} belongs to.",
            "i" = "Otherwise drop {.arg Phi}: the group factors of a hierarchy are uncorrelated."),
          class = "efa_reliability_phi_with_general"
        )
      }
      if (ncol(model) > 1L) {
        # Read first, report after. Everything the reading can refuse -- a `Phi` that
        # cannot be these columns' correlation matrix, a missing or infinite loading or
        # uniqueness -- is named on its own that way, rather than under a warning about a
        # reading that is then abandoned. The spec is built before the warning and returned
        # after it, so the two cannot come apart.
        spec <- .rel_read_pattern(model, Phi = Phi, u2 = u2, factor_map = factor_map,
                                  cormat = cormat, fac_names = fac_names)
        cli::cli_warn(
          c("{.arg model} is read as the pattern matrix of a correlated-factors solution, because {.arg Phi} gives its factor correlations.",
            "i" = "Every column is a group factor, so the coefficients a general factor defines -- omega hierarchical, the ECV, and the PUC -- are not reported.",
            "i" = "Drop {.arg Phi} to read the first column as a general factor and score the matrix as a bifactor one."),
          class = "efa_reliability_phi_as_pattern"
        )
        return(spec)
      }
      # One column is not a hierarchy and has no correlated group factors, so it falls
      # through to the bifactor route below; the check is all that applies to it.
      .rel_check_phi(Phi, ncol(model))
    }
    # Uniquenesses reach this route either from the caller, overriding the ones the adapter
    # would derive from the loadings, or off a Schmid-Leiman table's own `u2` column. Both
    # get the check the manual route's do, and for the same reason: a missing or infinite
    # one is scored rather than refused, and the table it produces looks ordinary -- an
    # infinite uniqueness drives its composite's omega total to a flat 0, a missing one
    # drops the coefficients of every composite the variable belongs to -- while nothing
    # downstream can name it.
    # The loadings get the same check, and not only because a missing or infinite one is no
    # more a loading than a missing uniqueness is a uniqueness: when `u2` is left out the
    # adapter derives it as 1 - rowSums(loadings^2), so an unchecked loading carries a
    # missing or infinite uniqueness straight past the check above and into the coefficients.
    .assert_args({
      checkmate::assert_numeric(model, any.missing = FALSE, finite = TRUE)
      if (!is.null(u2)) {
        checkmate::assert_numeric(u2, any.missing = FALSE, finite = TRUE,
                                  len = nrow(model))
      }
    })

    .rel_adapt_bifactor(model, factor_corres = factor_map, u2 = u2,
                        cormat = cormat, fac_names = fac_names)
  } else if (is.null(model)) {
    if (is.null(var_names) || is.null(g_load) || is.null(s_load) || is.null(u2)) {
      cli::cli_abort(
        "Specify all of {.arg var_names}, {.arg g_load}, {.arg s_load}, and {.arg u2}.",
        class = "efa_reliability_missing_args"
      )
    }
    # Validate the manually supplied components, as OMEGA() does: a non-matrix
    # s_load would otherwise be coerced (e.g. a vector to a one-column matrix) and
    # silently yield wrong coefficients instead of an error.
    if (!inherits(s_load, c("matrix", "SLLOADINGS"))) {
      cli::cli_abort(
        c("Invalid {.arg s_load}.",
          "i" = "Supply a Schmid-Leiman or bifactor group-loading matrix of class {.cls matrix} or {.cls efa_sl_loadings}."),
        class = "efa_reliability_bad_s_load"
      )
    }
    # A loading or a uniqueness that is missing or infinite is not one, and nothing
    # downstream can say so: every check that might is blind to exactly these values. The
    # Heywood report tests `u2 < eps`, which skips NA; the keying check and the
    # standardization check both compare totals that are themselves non-finite and so
    # return early. What the caller gets instead is not an empty table but a plausible one
    # -- an infinite uniqueness drives its composite's omega total to a flat 0, and an
    # infinite general loading leaves an ordinary-looking coefficient -- so these have to
    # be refused where they enter. Checking the group loadings as a numeric vector also
    # catches a matrix that is not numeric at all, which the class test above admits.
    .assert_args({
      checkmate::assert_numeric(s_load, any.missing = FALSE, finite = TRUE)
      checkmate::assert_numeric(g_load, any.missing = FALSE, finite = TRUE)
      checkmate::assert_numeric(u2, any.missing = FALSE, finite = TRUE)
    })
    # Square, symmetric, and a unit diagonal at once identifies a correlation matrix: a
    # group-loading matrix is essentially never all three, since a unit diagonal would mean
    # every variable loads exactly 1 on one group factor. Nothing else here would say so --
    # the class test above admits it, the numeric checks pass, and its p rows match p
    # variables, so it is scored as loadings and yields plausible but meaningless
    # coefficients. Named where the mistake is made, as for a `model` matrix. After the
    # numeric checks, so a `s_load` carrying missing values is reported as such rather than
    # by `.is_cormat()`, which aborts on one in terms of its own argument.
    if (.is_cormat(s_load)) {
      cli::cli_abort(
        c("{.arg s_load} is a correlation matrix, not a group-factor loading matrix.",
          "i" = "Supply the group-factor loadings of a Schmid-Leiman or bifactor solution.",
          "i" = "To score the components against these correlations, pass the matrix as {.arg cormat}."),
        class = "efa_reliability_cormat_as_s_load"
      )
    }
    # The whole printed Schmid-Leiman table, rather than its group-factor columns: its `h2`
    # and `u2` columns would be scored as two more group factors, and its general column as
    # a third, on top of the `g_load` given separately. The matrix route reads that table by
    # its layout because it is given as the entire solution; here the components are named
    # one by one, so the columns to pass are the ones the caller means.
    display_cols <- intersect(c("h2", "u2"), colnames(s_load))
    if (length(display_cols) > 0) {
      cli::cli_abort(
        c("{.arg s_load} carries the {.val {display_cols}} display column{?s} of a Schmid-Leiman loading table, which {?is/are} not group factors.",
          "i" = "Pass the group-factor columns only, e.g. {.code s_load = x$sl[, c(\"F1\", \"F2\")]}.",
          "i" = "To score the whole table as one solution, pass it as {.arg model} instead."),
        class = "efa_reliability_s_load_display_cols"
      )
    }
    # The components must agree in length so they combine into one item-by-factor spec;
    # otherwise they only fail deep in the algebra with an opaque base-R error.
    n_items <- nrow(as.matrix(s_load))
    if (length(g_load) != n_items || length(u2) != n_items ||
        length(var_names) != n_items) {
      cli::cli_abort(
        c("{.arg g_load}, {.arg u2}, and {.arg var_names} must each have one entry per variable.",
          "i" = "{.arg s_load} has {n_items} row{?s}, but {.arg g_load} has {length(g_load)}, {.arg u2} has {length(u2)}, and {.arg var_names} has {length(var_names)}."),
        class = "efa_reliability_length_mismatch"
      )
    }
    .rel_adapt_manual(g_load = g_load, s_load = s_load, u2 = u2,
                      var_names = var_names, factor_corres = factor_map,
                      type = "psych", cormat = cormat, pattern = pattern, Phi = Phi,
                      fac_names = fac_names)
  } else {
    cli::cli_abort(
      c("Invalid {.arg model}.",
        "i" = "Enter a {.cls lavaan}, {.fn psych::schmid}, {.cls SL}, or {.cls EFA} object, a bifactor loading matrix, or specify {.arg var_names}, {.arg g_load}, {.arg s_load}, and {.arg u2}.",
        "i" = "From an {.cls efa_average} object, use the averaged loading matrix in
               {.code $loadings$average}, with {.code $Phi$average} as {.arg Phi} and
               {.code $orig_R} as {.arg cormat}."),
      class = "efa_reliability_invalid_model"
    )
  }

}


# Keep only the requested coefficients in an efa_reliability result, preserving the
# object's schema, settings, and kind tagging (base data.frame subsetting drops the
# custom attributes).
.reliability_select <- function(res, coefficients) {
  settings <- attr(res, "settings")
  kind <- attr(res, "kind")
  out <- res[res$coefficient %in% coefficients, , drop = FALSE]
  rownames(out) <- NULL
  attr(out, "settings") <- settings
  attr(out, "kind") <- kind[names(kind) %in% coefficients]
  class(out) <- c("efa_reliability", "data.frame")
  out
}

#' Plot a multigroup factor analysis
#'
#' Two views of an [efa_group()] result, selected by `type`:
#'
#' - `"congruence"` (the default) plots the matched Tucker congruence of each factor
#'   between every group pair, with a percentile bootstrap confidence interval when one
#'   was computed (`b_boot > 0`). The Lorenzo-Seva and ten Berge (2006) reference bands
#'   (`.95` "equal", `.85` "fair") are drawn so a factor's cross-group similarity can be
#'   read against them at a glance.
#' - `"differences"` draws a heatmap of the signed cross-group loading differences (item
#'   by factor, one panel per group pair). Cells whose absolute difference reaches the
#'   salience threshold `delta` are outlined.
#'
#' @param x An object of class `efa_group` (output from [efa_group()]).
#' @param type character. Which plot to draw: `"congruence"` (per-factor congruence with
#'   confidence intervals) or `"differences"` (a per-item loading-difference heatmap).
#' @param ... Not used; for consistency with the generic.
#'
#' @returns A [ggplot2::ggplot] object.
#'
#' @references
#' Lorenzo-Seva, U., and ten Berge, J. M. F. (2006). Tucker's congruence coefficient as a
#' meaningful index of factor similarity. *Methodology*, 2, 57-64.
#' doi: 10.1027/1614-2241.2.2.57
#'
#' @family factor analysis
#'
#' @importFrom rlang .data
#' @export
#' @method plot efa_group
#'
#' @examples
#' g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
#' mg <- efa_group(GRiPS_raw, groups = g, n_factors = 1)
#'
#' # Per-factor congruence against the Lorenzo-Seva & ten Berge bands
#' plot(mg)
#'
#' # Per-item cross-group loading-difference heatmap
#' plot(mg, type = "differences")
plot.efa_group <- function(x, type = c("congruence", "differences"), ...) {
  type <- .match_arg_ci(type)
  if (type == "congruence") .gg_group_congruence(x) else .gg_group_differences(x)
}

# Per-factor matched Tucker congruence between every group pair, as points (or point
# ranges when a bootstrap supplied confidence intervals), against the Lorenzo-Seva & ten
# Berge (2006) reference bands. Returns a ggplot.
#' @importFrom rlang .data
.gg_group_congruence <- function(x) {
  M <- x$congruence$matched
  groups <- dimnames(M)[[1L]]
  fac_names <- dimnames(M)[[3L]]
  lower <- x$congruence$matched_ci$lower
  upper <- x$congruence$matched_ci$upper
  m <- length(groups)

  rows <- list()
  r <- 0L
  for (g in seq_len(m - 1L)) {
    for (h in (g + 1L):m) {
      for (f in seq_along(fac_names)) {
        r <- r + 1L
        rows[[r]] <- data.frame(
          pair = paste(groups[g], "vs", groups[h]),
          factor = fac_names[f],
          phi = M[g, h, f],
          lower = if (is.null(lower)) NA_real_ else lower[g, h, f],
          upper = if (is.null(upper)) NA_real_ else upper[g, h, f],
          stringsAsFactors = FALSE)
      }
    }
  }
  dat <- do.call(rbind, rows)
  dat$factor <- factor(dat$factor, levels = fac_names)
  dat$pair <- factor(dat$pair, levels = unique(dat$pair))
  has_ci <- any(!is.na(dat$lower))

  y_min <- suppressWarnings(min(c(dat$phi, dat$lower), na.rm = TRUE))
  if (!is.finite(y_min)) y_min <- 0.8

  # The .85 ("fair") and .95 ("equal") band edges mirror the cut-offs in `.invariance_band()`;
  # keep them in step if those are ever retuned.
  p <- ggplot2::ggplot(dat, ggplot2::aes(.data$factor, .data$phi, colour = .data$pair)) +
    ggplot2::annotate("rect", xmin = -Inf, xmax = Inf, ymin = 0.85, ymax = 0.95,
                      fill = "grey80", alpha = 0.35) +
    ggplot2::geom_hline(yintercept = c(0.85, 0.95), linetype = 2, colour = "grey50")

  p <- if (has_ci) {
    p + ggplot2::geom_pointrange(
      ggplot2::aes(ymin = .data$lower, ymax = .data$upper),
      position = ggplot2::position_dodge(width = 0.5), size = 0.5, na.rm = TRUE)
  } else {
    p + ggplot2::geom_point(position = ggplot2::position_dodge(width = 0.5),
                            size = 2.5, na.rm = TRUE)
  }

  p +
    ggplot2::coord_cartesian(ylim = c(min(0.8, y_min), 1)) +
    ggplot2::scale_colour_viridis_d(end = 0.8) +
    ggplot2::labs(x = "Factor", y = "Tucker's congruence",
                  title = "Factor congruence across groups", colour = "Group pair") +
    .gg_theme()
}

# Heatmap of the signed cross-group loading differences (item by factor), one panel per
# group pair; cells reaching the salience threshold `delta` are outlined. Returns a ggplot.
#' @importFrom rlang .data
.gg_group_differences <- function(x) {
  d <- x$flags
  pair_labels <- paste(d$group_1, "vs", d$group_2)
  d$pair <- factor(pair_labels, levels = unique(pair_labels))
  d$indicator <- factor(d$indicator, levels = rev(unique(d$indicator)))
  d$factor <- factor(d$factor, levels = unique(d$factor))
  # which() (not the logical vector) so an NA `flagged`, should one ever arise, drops out
  # rather than injecting an all-NA row that geom_tile would try to outline.
  flagged <- d[which(d$flagged), , drop = FALSE]

  lim <- max(abs(d$diff), na.rm = TRUE)
  if (!is.finite(lim) || lim == 0) lim <- 1

  p <- ggplot2::ggplot(d, ggplot2::aes(.data$factor, .data$indicator)) +
    ggplot2::geom_tile(ggplot2::aes(fill = .data$diff), colour = "grey92")
  if (nrow(flagged) > 0L) {
    p <- p + ggplot2::geom_tile(data = flagged, fill = NA, colour = "black",
                                linewidth = 0.7)
  }

  p +
    ggplot2::scale_fill_gradient2(low = "#2166AC", mid = "white", high = "#B2182B",
                                  midpoint = 0, limits = c(-lim, lim)) +
    ggplot2::facet_wrap(ggplot2::vars(.data$pair)) +
    ggplot2::labs(x = "Factor", y = NULL, fill = "Loading\ndifference",
                  title = "Cross-group loading differences",
                  subtitle = paste0("Outlined cells: |difference| >= ",
                                    .efa_group_delta_str(x$settings$delta))) +
    .gg_theme()
}

#' Multigroup exploratory factor analysis
#'
#' Fit an exploratory factor analysis in each of several groups at a common
#' number of factors and bring the per-group solutions into one shared
#' orientation so their loadings can be compared. Each group is fitted with
#' [efa_fit()]; the solutions are then aligned either to a symmetric consensus target
#' or to a chosen reference group (see *Alignment*).
#'
#' @details
#' ## Input
#' Groups can be supplied in two ways: raw data together with a grouping vector
#' (`x` a data frame or matrix, `groups` one value per row), or a named list of
#' per-group data sets in `x` (with `groups` left `NULL`). The list may hold raw
#' data frames or correlation matrices (supply `N`), but not a mix of the two.
#' All groups must contain the same items in the same order; a different item set
#' or order is an error rather than being silently reordered.
#'
#' Every group is fitted at the same `n_factors`. Extra arguments in `...` (for example
#' `estimator`, `rotation`, `cor_method`, or an [estimate_control()] / [rotate_control()]
#' carrying the tuning knobs) are forwarded unchanged to each [efa_fit()] call, so the
#' estimator and rotation are common to all groups.
#'
#' The requested number of factors must be small enough, relative to the number of items,
#' for the `n_factors`-factor model to be identified for the shared item set. Unlike a single
#' [efa_fit()] fit -- which only warns on an under-identified model -- a multigroup fit aborts
#' when this fails, because a shared alignment target across an under-identified group is not
#' interpretable.
#'
#' ## Alignment
#' A factor solution is identified only up to a rotation of its factors, so the
#' per-group solutions must be brought into a common orientation before their
#' loadings can be compared. Two strategies are available and are chosen
#' automatically:
#'
#' - **Consensus** (the default for orthogonal rotations and for unrotated
#'   solutions): a symmetric target is built across all groups using Generalized Procrustes
#'   Analysis (Gower, 1975), and every group's loadings are rotated to it. Because this
#'   target's own orientation is arbitrary, it is then rotated once more into a fixed
#'   convention (called the gauge), and the same transform is applied to every group. The
#'   gauge uses the same simple-structure criterion as the requested rotation, applied to the
#'   target itself, so the shared loadings are in the same kind of frame as the per-group
#'   solutions they summarise. Where no rotation criterion identifies a unique frame (an
#'   unrotated solution, or a two-factor `bifactorT` request), the target's principal-axes
#'   orientation is used instead. Either way the columns are ordered by decreasing sum of
#'   squares and signed by their column sums, and the shared orientation -- and hence every
#'   reported congruence, difference, and flag -- does not depend on the order the groups are
#'   supplied, to well beyond the precision loadings are reported at.
#' - **Reference**: every group's loadings are aligned by Procrustes rotation to
#'   one reference group's loadings, which are kept fixed. This path is used when
#'   `reference_group` is given, and is used automatically for oblique rotations
#'   because the consensus iteration is not defined for oblique transforms with
#'   more than one factor. When an oblique rotation triggers the reference path
#'   without an explicit `reference_group`, the first group is used and a message
#'   reports this; the requested rotation is never silently changed.
#'
#' In both cases the returned per-group loadings share the column order and sign
#' of the returned `target`.
#'
#' ## Comparing the aligned loadings
#' Because the per-group loadings share one orientation, they can be compared cell by cell.
#' `efa_group()` reports a per-pair summary of their differences (`diffs`) and a per-item,
#' per-factor flag table (`flags`) marking cells whose absolute difference reaches `delta`; a
#' bootstrap (`b_boot > 0`) additionally reports, for every cell, whether its difference's
#' confidence interval excludes zero. With `invariance = TRUE`, each factor and group pair also gets an
#' approximate-invariance verdict based on the matched Tucker congruence (see *Value* for the
#' similarity bands and how a bootstrap is used).
#'
#' @param x A data frame or matrix of raw data (with `groups`), or a named list
#'   of per-group data sets -- either raw data frames/matrices or correlation
#'   matrices (all of one kind).
#' @param groups A vector with one value per row of `x`, giving each row's group.
#'   Only used when `x` is a single raw data set; leave `NULL` when `x` is a list.
#'   Rows with a missing group value are dropped with a warning.
#' @param n_factors numeric. The common number of factors extracted in every
#'   group.
#' @param N numeric. The number of observations per group, used only for
#'   correlation-matrix input: either a single value applied to all groups or one
#'   value per group. Ignored for raw data, where `N` is taken from each group's
#'   data. Default is `NA`.
#' @param reference_group The group to align the others to (a group name or an
#'   integer index). If `NULL` (default), orthogonal and unrotated solutions use
#'   the symmetric consensus target; oblique solutions with more than one factor
#'   fall back to the first group as reference. Supplying a value forces the
#'   reference alignment.
#' @param b_boot numeric. The number of non-parametric bootstrap replicates used to
#'   form percentile confidence intervals for the between-group Tucker congruences.
#'   `0` (the default) skips the bootstrap and returns the congruence point estimates
#'   only. Bootstrapping requires raw data; it is skipped with a warning for
#'   correlation-matrix input.
#' @param ci numeric. The confidence level for the bootstrap congruence intervals, a
#'   single value in `(0, 1)`. Default is `0.95`.
#' @param seed numeric or `NULL`. An optional seed making the analysis reproducible. It
#'   covers the per-group fits, some of whose rotations draw random starts, whether or not a
#'   bootstrap is run. With a bootstrap, it also makes the result independent of how many
#'   parallel workers are used (bootstrap replicates run with
#'   [future_lapply()][future.apply::future_lapply], configurable via [future::plan()]). The
#'   caller's random-number stream is restored afterwards, leaving no side effect. Default is
#'   `NULL`.
#' @param delta numeric. The salience threshold for the per-item loading-difference flag
#'   table: an item's loading on a factor is flagged for a group pair when the groups'
#'   aligned loadings differ by at least `delta` in absolute value. This is a descriptive
#'   salience heuristic, not a significance test; common alternatives are `0.15` and `0.20`.
#'   The threshold applies to whatever loading metric the chosen rotation produces (pattern
#'   coefficients for an oblique rotation). `0` flags every cell. Default is `0.1`.
#'
#'   The geomin rotations take a criterion parameter of the same name. A `delta` given
#'   directly is always this salience threshold and never reaches the rotation; give the
#'   geomin parameter as `rotate_control(delta = ...)`. With `rotation = "geominT"` or
#'   `"geominQ"`, a supplied `delta` gives a warning that says which of the two applies.
#' @param invariance logical. Whether to add an approximate-invariance verdict per factor and
#'   group pair from the Lorenzo-Seva and ten Berge (2006) congruence bands (see *Value*).
#'   Default is `FALSE`.
#' @param se Not used. `efa_group()` itself sets the standard-error method of the
#'   per-group [efa_fit()] calls, so a supplied value is dropped with a warning. Ask for
#'   bootstrap confidence intervals of the between-group congruences with `b_boot`.
#'   Default is `NULL`.
#' @param ... Additional arguments passed to [efa_fit()] for every group (for
#'   example `estimator`, `rotation`, or `cor_method`). The [estimate_control()] and
#'   [rotate_control()] objects are accepted through `...` as well, although they are not
#'   declared formals: pass them as `estimate_control =` / `rotate_control =` exactly
#'   as you would to [efa_fit()]. A name that is neither an [efa_fit()] argument nor a
#'   rotation-engine extra is rejected.
#'
#'   A rotation-engine extra that shares a name with an `efa_group()` argument cannot reach
#'   the rotation through `...`, because the argument takes the name first (for example
#'   geomin's `delta`; see `delta` above).
#'
#' @returns An object of class `efa_group`, a list containing:
#' \item{loadings}{A named list of the aligned per-group loading matrices. Their
#'   columns match the columns of `target` in order and sign.}
#' \item{target}{The alignment target: the symmetric consensus target, or the
#'   reference group's own loadings.}
#' \item{Phi}{A named list of the aligned per-group factor intercorrelations for
#'   an oblique rotation; `NULL` otherwise.}
#' \item{congruence}{Tucker congruence between the aligned group loadings, a list
#'   with: `matrices`, a nested list whose `[[g]][[h]]` element is the
#'   factor-by-factor congruence matrix between the aligned loadings of groups
#'   `g` and `h`; `matched`, a groups-by-groups-by-factors array of the
#'   matched-factor congruences (the diagonal of each pairwise matrix); and
#'   `degenerate`, a groups-by-groups logical matrix flagging pairs whose
#'   congruence is undefined (for example, a near-zero factor), for which the
#'   corresponding entries are `NA`. When `b_boot > 0` (raw data), three further
#'   elements are added: `matched_se`, the bootstrap standard error of each
#'   matched congruence; `matched_ci`, a list of `lower` and `upper` percentile
#'   confidence limits (each a groups-by-groups-by-factors array); and `n_boot`,
#'   the number of bootstrap replicates that aligned in every group and so
#'   contributed to the intervals (a replicate whose fit did not converge is
#'   retained, as in [efa_fit()]).}
#' \item{diffs}{A data frame with one row per group pair summarising the differences
#'   between their aligned loadings: the mean, median, minimum, and maximum absolute
#'   difference, the root-mean-square difference (`rmse`), and `n_flagged`, the number of
#'   loading cells whose absolute difference reaches `delta`.}
#' \item{flags}{A data frame with one row per group pair, item, and factor giving the signed
#'   loading difference (`diff`), its absolute value (`abs_diff`), and `flagged`, whether it
#'   reaches `delta`. When a bootstrap was run (`b_boot > 0`, raw data), `ci_lower`,
#'   `ci_upper`, and `ci_excludes_0` add the percentile confidence interval for the
#'   difference and whether it excludes zero; otherwise these are `NA`.}
#' \item{invariance}{When `invariance = TRUE`, a data frame with one row per group pair and
#'   factor giving the matched Tucker congruence (`phi`), its bootstrap CI lower bound
#'   (`phi_lower`, `NA` without a bootstrap), and an approximate-invariance `verdict` based on
#'   the Lorenzo-Seva and ten Berge (2006) similarity bands: `phi >= 0.95` is "equal" and
#'   `[0.85, 0.95)` is "fair"; congruences `< 0.85`, below their bands, are labelled
#'   "incongruent". The verdict is read from `phi_lower` when a bootstrap is available
#'   (conservative) and from `phi` otherwise. A wide interval therefore lowers the verdict:
#'   `phi` = 0.989 with `phi_lower` = 0.726 is labelled "incongruent", because the band is
#'   applied to the lower bound. Tucker's congruence is invariant to a
#'   proportional rescaling of a factor's loadings, so a factor can be graded "equal"
#'   even when one group's loadings on it are uniformly stronger; read the verdict
#'   alongside `diffs`. `NULL` when `invariance = FALSE`.}
#' \item{efa}{The named list of per-group [efa_fit()] objects (each retains its own
#'   diagnostics, e.g. `heywood`).}
#' \item{alignment}{The alignment result: the consensus object (see
#'   [efa_procrustes()]), or a list with the reference group, the target, and the
#'   per-group Procrustes results. On the consensus path, this is the raw Procrustes
#'   iteration output: its `target`/`aligned_loadings` are in a different (pre-gauge)
#'   orientation than the gauged `target`/`loadings` returned above. Use `target`/`loadings`
#'   above for comparisons.}
#' \item{settings}{A list of the settings used, including the per-group `N`, the
#'   alignment method, the group that seeded the consensus frame (`alignment_start`,
#'   `NULL` on the reference path), the orientation the shared frame was put in
#'   (`gauge`: the rotation's own name, `"principal_axes"`, or `"identity"` for a
#'   single factor; `NULL` on the reference path), the rotation, the estimator, the
#'   input type, whether a bootstrap is available (`can_bootstrap`, `FALSE` for
#'   correlation-matrix input), and `seed` (`NULL` when none was supplied).}
#'
#' @references
#' Efron, B., & Tibshirani, R. J. (1993). *An Introduction to the Bootstrap*.
#' Chapman & Hall.
#'
#' Gower, J. C. (1975). Generalized Procrustes analysis. *Psychometrika*, 40,
#' 33-51. doi: 10.1007/BF02291478
#'
#' Lorenzo-Seva, U., and ten Berge, J. M. F. (2006). Tucker's congruence coefficient as a
#' meaningful index of factor similarity. *Methodology*, 2, 57-64.
#' doi: 10.1027/1614-2241.2.2.57
#'
#' @family factor analysis
#'
#' @export
#'
#' @examples
#' # Raw data split by a grouping vector (unrotated, consensus alignment)
#' g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
#' mg <- efa_group(GRiPS_raw, groups = g, n_factors = 1)
#' mg$loadings
#'
#' # Per-pair difference summary and the per-item salience-flag table
#' mg$diffs
#' mg$flags
#'
#' \donttest{
#' # Percentile bootstrap confidence intervals for the between-group congruences, with an
#' # approximate-invariance verdict read conservatively off the congruence CI lower bound
#' mg_ci <- efa_group(GRiPS_raw, groups = g, n_factors = 1, b_boot = 100, seed = 42,
#'                    invariance = TRUE)
#' mg_ci$congruence$matched_ci
#' mg_ci$invariance
#'
#' # A named list of correlation matrices sharing the same items, common
#' # three-factor model, orthogonal rotation -> symmetric consensus target
#' bands <- list(age_6_8 = WJIV_ages_6_8$cormat, age_14_19 = WJIV_ages_14_19$cormat)
#' Ns <- c(WJIV_ages_6_8$N, WJIV_ages_14_19$N)
#' efa_group(bands, n_factors = 3, N = Ns, rotation = "varimax")
#'
#' # An oblique rotation aligns to a reference group (reported via a message)
#' efa_group(bands, n_factors = 3, N = Ns, rotation = "promax")
#' }
#'
efa_group <- function(x, groups = NULL, n_factors, N = NA,
                      reference_group = NULL, b_boot = 0L, ci = 0.95,
                      seed = NULL, delta = 0.1, invariance = FALSE, se = NULL, ...) {

  # Both guards read `...names()` and so run before the dots are forced: a rejected name
  # is reported as such even when its value would itself have errored.
  # A flat tuning knob or the former `method` spelling in the dots would only surface from
  # inside the first per-group fit, re-wrapped as an efa_group_fit_failed "fit failed for
  # group X" error; reject it here so the message names the actual mistake.
  .reject_flat_knobs(...names(), fn = "efa_group")
  # Same reason for an unknown name: the per-group fits run inside a tryCatch that
  # re-labels their errors, so a misspelled argument would be reported as a statistical
  # failure of whichever group was fitted first. The fits are rotated, so a rotation
  # engine extra is a legitimate name here.
  .reject_unknown_fit_dots(...names(), fn = "efa_group")
  efa_args <- list(...)
  checkmate::assert_count(n_factors, positive = TRUE)
  checkmate::assert_count(b_boot)
  b_boot <- as.integer(b_boot)
  # 0 switches the bootstrap off; from 1 upward it is a request for congruence intervals, and those
  # need at least two replicates to have a spread. Rejected here rather than by the per-group
  # efa_fit(), whose abort would be re-raised as a failure of whichever group happened to be fitted
  # first and so would name a group instead of the argument at fault.
  if (b_boot == 1L) {
    cli::cli_abort(
      c("{.arg b_boot} must be 0 (no bootstrap) or at least 2.",
        "x" = "You supplied {.arg b_boot} = {b_boot}.",
        "i" = "A bootstrap confidence interval is built from the spread across replicates and is
               undefined below two of them."),
      class = "efa_group_bad_b_boot"
    )
  }
  if (!checkmate::test_number(ci) || ci <= 0 || ci >= 1) {
    cli::cli_abort("{.arg ci} must be a single number strictly between 0 and 1.",
                   class = "efa_group_bad_ci")
  }
  if (!is.null(seed)) checkmate::assert_number(seed, finite = TRUE)
  checkmate::assert_number(delta, lower = 0, finite = TRUE)
  checkmate::assert_flag(invariance)
  # `delta` names two different things: this salience threshold and the geomin criterion
  # parameter. The formal takes the name, so the caller's intention cannot be read off the
  # call; recorded here and reported below, once the rotation is resolved.
  delta_supplied <- !missing(delta)

  # Resolve the input into a named list of per-group data sets, tagging whether
  # they are raw data or correlation matrices (which cannot be bootstrapped).
  resolved <- .efa_group_resolve_input(x, groups)
  group_data <- resolved$group_data
  group_names <- names(group_data)
  input_type <- resolved$input_type
  m <- length(group_data)

  # Bootstrap confidence intervals for the congruences need raw data to resample;
  # correlation-matrix input carries no cases, so a requested bootstrap is skipped
  # with a note and only the point estimates are returned.
  do_boot <- b_boot > 0L && input_type == "raw"
  if (b_boot > 0L && input_type != "raw") {
    cli::cli_warn(
      c("Bootstrap confidence intervals for the congruences need raw data.",
        "i" = "{.arg x} holds correlation matrices, which carry no cases to resample; the point estimates are returned without intervals."),
      class = "efa_group_boot_unavailable"
    )
  }

  if (m < 2L) {
    cli::cli_abort(
      c("{.fn efa_group} needs at least two groups.",
        "x" = "Found {m} group{?s}."),
      class = "efa_group_too_few_groups"
    )
  }

  # Enforce an identical item set and order across groups.
  .efa_group_check_items(group_data)
  p <- ncol(group_data[[1L]])

  # Validate an explicit reference group now, before the per-group fits, so a bad
  # name/index fails fast. The oblique fall-back reference is resolved later, once
  # the (common) rotation is known.
  ref_requested <- !is.null(reference_group)
  ref_idx_requested <- if (ref_requested) {
    .efa_group_resolve_reference(reference_group, group_names)
  }

  # The common-k model must be identified for the shared items: non-negative EFA
  # degrees of freedom (the Ledermann bound). This also rules out k >= p.
  df <- .efa_df(p, n_factors)
  if (df < 0) {
    cli::cli_abort(
      c("The {n_factors}-factor model is under-identified for {p} variable{?s}.",
        "x" = "It has {df} degree{?s} of freedom; a non-negative value is required.",
        "i" = "Extract fewer factors or include more variables."),
      class = "efa_group_under_identified"
    )
  }

  # N to pass to EFA: derived from the data for raw groups, supplied for
  # correlation matrices (scalar recycled or one value per group).
  Ns_in <- .efa_group_resolve_N(N, m, input_type)

  # A supplied `seed` makes the whole analysis reproducible and leaves the caller's RNG
  # stream untouched: it is saved and restored on exit (or, if none existed, the state
  # set.seed() creates is removed again). It is set regardless of `do_boot`, because the
  # per-group fits are stochastic on their own whenever the rotation draws random starts;
  # with a bootstrap it additionally makes the congruence intervals independent of the
  # number of parallel workers. The per-group efa_fit() fits are called with seed = NULL
  # so they advance this one seeded stream in sequence rather than each resetting it;
  # efa_fit()'s own future.seed = TRUE keeps the replicate fits worker-count-independent.
  # Mirrors the seed handling in efa_fit().
  .set_local_seed(seed)

  # Fit every group at the common number of factors. A degenerate group (too few
  # cases, a constant item, a non-computable matrix) makes efa_fit abort; re-raise it
  # with the group's name so the failure is attributable. efa_fit warnings (e.g. a
  # Heywood case) are left to surface; the flagged variables stay in the fit's
  # `heywood` element.
  #
  # When a bootstrap is requested, each group is additionally fitted with
  # se = "np-boot" so its replicate unrotated-loading cube is produced; the cube is
  # captured before the fit is stripped back to its point-estimate form (out$efa stays
  # lean and identical in shape whether or not a bootstrap was run). The loop runs
  # strictly serially in a fixed group order: each group advances the shared (seeded)
  # RNG stream, so group g+1's resampling depends on group g -- do not parallelise it.
  #
  # efa_group controls the SE method itself, so a supplied `se` is dropped with a warning:
  # otherwise it would trigger an unrequested, unseeded per-group bootstrap or leave
  # SE/replicate payload in out$efa. Congruence intervals are requested with `b_boot`, not
  # `se`. `se` is a declared formal only so that this path is reachable: every formal
  # precedes `...`, so R partially matches `se` to `seed` and a dots-only `se` would abort
  # in the `seed` check with a message naming an argument the caller never wrote.
  if (!is.null(se)) {
    cli::cli_warn(
      "{.arg se} is ignored by {.fn efa_group}; use {.arg b_boot} for congruence confidence intervals.",
      class = "efa_group_se_ignored"
    )
  }
  boot_cubes <- if (do_boot) stats::setNames(vector("list", m), group_names)

  fits <- vector("list", m)
  names(fits) <- group_names
  # All groups come from one `x`, so the per-fit "computing correlations from the raw data"
  # note states one fact about one input; report it once instead of once per group.
  .cor_note_once(
    for (g in seq_len(m)) {
      fit_args <- c(list(x = group_data[[g]], n_factors = n_factors, N = Ns_in[[g]]),
                    efa_args)
      if (do_boot) {
        fit_args <- c(fit_args,
                      list(se = "np-boot", b_boot = b_boot, ci = ci, seed = NULL))
      }
      fit_g <- tryCatch(
        do.call(efa_fit, fit_args),
        error = function(e) {
          # The cause is carried by `parent`, which rlang prints in full under "Caused by
          # error:" with its own bullets intact. Repeating conditionMessage() here as a
          # single bullet would both duplicate it and flatten it: a parent that raises more
          # than one bullet -- most of efa_fit()'s input guards do -- has its own markers
          # rendered inside this one, mid-sentence.
          cli::cli_abort(
            "The {.fn efa_fit} fit failed for group {.val {group_names[[g]]}}.",
            class = "efa_group_fit_failed", parent = e
          )
        }
      )
      if (do_boot) {
        boot_cubes[[g]] <- fit_g$replicates$unrot_loadings
        fit_g <- .efa_strip_boot(fit_g)
      }
      fits[[g]] <- fit_g
    }
  )

  # The estimator and rotation are common (one `...` applied to every group), so
  # read them from the first fit. `settings$rotation` is always populated.
  settings1 <- fits[[1L]]$settings
  rotation <- settings1$rotation
  rotation_type <- if (is.null(rotation)) "none" else .rotation_family(rotation)
  oblique <- rotation_type == "oblique" && n_factors >= 2L

  # The geomin rotations take a criterion parameter that is also called `delta`. It cannot
  # arrive through the dots, because the salience-threshold formal takes the name first, so
  # the geomin criterion silently stayed at its default. Report which of the two readings
  # was applied and how to get the other one.
  if (delta_supplied && !is.null(rotation) && rotation %in% c("geominT", "geominQ")) {
    cli::cli_warn(
      c("{.arg delta} is the salience threshold of the flag table, not the
         {.val {rotation}} criterion parameter.",
        "i" = "The flag threshold is {delta}. The rotation criterion parameter is not
               changed by it.",
        "i" = "To set the criterion parameter, pass {.code rotate_control(delta = ...)}."),
      class = "efa_group_delta_ambiguous"
    )
  }

  # Loadings as plain matrices (efa_fit returns them classed as efa_loadings/LOADINGS).
  unrot_loadings <- lapply(fits, function(f) .change_class(f$unrot_loadings, "matrix"))
  rot_loadings <- if (rotation_type != "none") {
    lapply(fits, function(f) .change_class(f$rot_loadings, "matrix"))
  }

  # The reference path is used when a reference group is named, and always for an
  # oblique rotation (the consensus iteration is undefined for oblique k > 1).
  use_reference <- ref_requested || oblique

  if (use_reference) {
    ref_idx <- if (ref_requested) ref_idx_requested else 1L

    # Surface the automatic routing when an oblique rotation forced it without an
    # explicit choice; the rotation itself is kept, not switched to orthogonal.
    if (oblique && !ref_requested) {
      cli::cli_inform(
        c("Oblique rotations are aligned to a reference group, not a symmetric consensus target.",
          "i" = "The consensus target is undefined for oblique rotations with more than one factor.",
          "i" = "Group {.val {group_names[[ref_idx]]}} is used as the reference; set {.arg reference_group} to choose another."),
        class = "efa_group_oblique_reference"
      )
    }

    ref_loadings <- if (rotation_type == "none") {
      unrot_loadings[[ref_idx]]
    } else {
      rot_loadings[[ref_idx]]
    }
    ref_phi <- if (oblique) .change_class(fits[[ref_idx]]$Phi, "matrix")

    aligned <- .efa_group_align_reference(
      unrot_loadings = unrot_loadings, ref_loadings = ref_loadings,
      ref_idx = ref_idx, ref_phi = ref_phi, oblique = oblique,
      group_names = group_names
    )
    alignment_method <- "reference"
  } else {
    # The gauge runs the requested criterion on the shared frame, so a criterion parameter
    # the fits were tuned with has to reach it too -- `delta` changes what geomin optimizes,
    # and gauging at the default while the groups were rotated at another value would put
    # the shared frame in a different simple structure than the solutions it summarises. It
    # is read off the rotate control, the only place it can be: `delta` is efa_group()'s own
    # salience-flag argument, so a geomin delta can never arrive through the dots, and the
    # fit settings do not record it. (`gam` is oblimin-only and so never reaches this path,
    # which oblique rotations do not take.)
    rc <- efa_args$rotate_control
    gauge_args <- if (inherits(rc, "efa_rotate_control")) {
      # The criterion parameters are the rotation's engine extras less `maxit`, which is a
      # solver budget, not part of the criterion -- the gauge pins its own. Derived from
      # .rotation_dot_extras() so a criterion parameter added there reaches the gauge too.
      crit_extras <- setdiff(.rotation_dot_extras(rotation), "maxit")
      rc$extra_args[intersect(names(rc$extra_args), crit_extras)]
    } else {
      list()
    }
    aligned <- .efa_group_align_consensus(
      unrot_loadings = unrot_loadings, rot_loadings = rot_loadings,
      rotation_type = rotation_type, group_names = group_names,
      rotation = rotation,
      # Kaiser normalization changes what the criterion optimizes just as `delta` does, so
      # the gauge takes the setting the fits ran under. It is already resolved in the fit
      # settings (through the type preset), unlike `delta`, which is not recorded there.
      normalize = settings1$normalize %||% TRUE,
      rotation_args = gauge_args
    )
    alignment_method <- "consensus"
  }

  # Tucker congruence between the aligned group loadings (full pairwise matrices,
  # the matched-factor diagonals, and a flag for any degenerate pair).
  congruence <- .efa_group_congruence(aligned$loadings)

  # Percentile-bootstrap confidence intervals for the matched congruences (raw data
  # and b_boot > 0). Each group's replicate unrotated loadings are re-aligned to the
  # frozen point-estimate target with the same Procrustes mode as the point estimate,
  # and the per-factor congruence is recomputed per replicate. This runs in the main
  # process (serial) under the seed umbrella above, so it is reproducible and
  # worker-count-independent.
  if (do_boot) {
    proc_rotation <- if (oblique) "oblique" else "orthogonal"
    boot <- .efa_group_boot_congruence(boot_cubes, aligned$target, proc_rotation, ci)
    congruence$matched_se <- boot$matched_se
    congruence$matched_ci <- boot$matched_ci
    congruence$n_boot <- boot$n_boot
  }

  # Cross-group loading differences on the aligned solutions: a per-pair magnitude summary
  # and a per-item, per-factor salience-flag table, paired with the bootstrap difference
  # intervals when a bootstrap was run.
  diffs <- .efa_group_diffs(aligned$loadings, delta,
                            diff_ci = if (do_boot) boot$diff_ci else NULL)

  # Optional approximate-invariance verdict per factor and pair (Lorenzo-Seva & ten Berge,
  # 2006), read conservatively off the congruence CI lower bound when bootstrapped.
  invariance_tbl <- if (isTRUE(invariance)) .efa_group_invariance(congruence)

  # The N actually used per group (derived for raw data, supplied otherwise).
  Ns_used <- vapply(fits, function(f) {
    n <- f$settings$N
    if (is.null(n)) NA_real_ else as.numeric(n)
  }, numeric(1L))
  names(Ns_used) <- group_names

  settings <- list(
    n_factors = n_factors,
    N = Ns_used,
    reference_group = if (use_reference) group_names[[ref_idx]] else NULL,
    alignment = alignment_method,
    # The consensus iteration records the seeding target as a position in the supplied
    # list; report it as the group name, so it identifies a group the way its sibling
    # `reference_group` does rather than being a bare index whose meaning changes with
    # the order the groups were passed in.
    alignment_start = if (use_reference) NULL else group_names[[aligned$alignment$start]],
    # Which orientation the shared frame was put in. It is not always derivable from
    # `rotation`: a criterion that degenerates at the requested number of factors, and a
    # rotation engine that fails outright, both fall back to the principal axes silently.
    gauge = aligned$gauge,
    rotation = rotation,
    rotation_family = rotation_type,
    estimator = settings1$estimator,
    cor_method = settings1$cor_method,
    input_type = input_type,
    can_bootstrap = input_type == "raw",
    b_boot = if (do_boot) b_boot else 0L,
    ci = ci,
    delta = delta,
    invariance = invariance,
    seed = seed,
    groups = group_names,
    efa_args = efa_args
  )

  .new_efa_group(
    loadings = aligned$loadings,
    target = aligned$target,
    Phi = aligned$Phi,
    congruence = congruence,
    diffs = diffs$diffs,
    flags = diffs$flags,
    invariance = invariance_tbl,
    efa = fits,
    alignment = aligned$alignment,
    settings = settings
  )
}


# Assemble an efa_group object with its stable field set and class. The single place the
# object's shape is defined, so every field is listed here in a fixed order.
.new_efa_group <- function(loadings, target, Phi, congruence, diffs, flags,
                           invariance, efa, alignment, settings) {
  structure(
    list(
      loadings = loadings,
      target = target,
      Phi = Phi,
      congruence = congruence,
      diffs = diffs,
      flags = flags,
      invariance = invariance,
      efa = efa,
      alignment = alignment,
      settings = settings
    ),
    class = "efa_group"
  )
}


# Resolve `x`/`groups` into a named list of per-group data plus the input type
# ("raw" or "cormat"). Single raw data are split by `groups`; a list is taken as
# already-split groups (all raw or all correlation matrices, never a mix).
.efa_group_resolve_input <- function(x, groups) {

  if (is.data.frame(x) || is.matrix(x)) {
    if (.is_cormat(x)) {
      cli::cli_abort(
        c("A correlation matrix cannot be split into groups.",
          "x" = "{.arg x} looks like a correlation matrix, which carries no cases to split.",
          "i" = "Supply raw data with a {.arg groups} vector, or a named list of per-group correlation matrices."),
        class = "efa_group_cormat_needs_list"
      )
    }
    if (is.null(groups)) {
      cli::cli_abort(
        c("A single data set needs a {.arg groups} vector.",
          "i" = "Supply {.arg groups} (one value per row of {.arg x}), or pass a named list of per-group data sets."),
        class = "efa_group_needs_groups"
      )
    }
    if (length(groups) != nrow(x)) {
      cli::cli_abort(
        c("{.arg groups} must have one value per row of {.arg x}.",
          "x" = "{.arg x} has {nrow(x)} row{?s} but {.arg groups} has length {length(groups)}."),
        class = "efa_group_groups_length"
      )
    }

    g <- as.factor(groups)
    na_g <- is.na(g)
    if (any(na_g)) {
      cli::cli_warn(
        "{sum(na_g)} row{?s} with a missing group value {?was/were} dropped.",
        class = "efa_group_na_group"
      )
      x <- x[!na_g, , drop = FALSE]
      g <- g[!na_g]
    }
    # Drop empty levels -- both those left by the NA removal above and any unused
    # levels a factor `groups` already carried -- so they do not become phantom
    # zero-row groups.
    g <- droplevels(g)

    lv <- levels(g)
    group_data <- lapply(lv, function(l) x[g == l, , drop = FALSE])
    names(group_data) <- lv
    return(list(group_data = group_data, input_type = "raw"))
  }

  if (is.list(x)) {
    if (!is.null(groups)) {
      cli::cli_abort(
        c("{.arg groups} is only used with a single data set.",
          "i" = "When {.arg x} is a list of groups, drop {.arg groups}."),
        class = "efa_group_groups_with_list"
      )
    }

    for (i in seq_along(x)) {
      if (!inherits(x[[i]], c("matrix", "data.frame"))) {
        cli::cli_abort(
          c("Every element of the group list must be a matrix or data frame.",
            "x" = "Element {i} is {.obj_type_friendly {x[[i]]}}."),
          class = "efa_group_input"
        )
      }
    }

    is_cm <- vapply(x, .is_cormat, logical(1L))
    if (length(unique(is_cm)) > 1L) {
      cli::cli_abort(
        c("The group list mixes correlation matrices and raw data.",
          "i" = "Supply either all raw data sets or all correlation matrices."),
        class = "efa_group_mixed_input"
      )
    }

    nm <- names(x)
    gen <- paste0("group", seq_along(x))
    if (is.null(nm)) {
      nm <- gen
    } else {
      nm[!nzchar(nm) | is.na(nm)] <- gen[!nzchar(nm) | is.na(nm)]
    }

    # Duplicated names would leave the second group unaddressable by name and make
    # a name-based `reference_group` ambiguous, so require them to be unique.
    if (anyDuplicated(nm)) {
      dup <- unique(nm[duplicated(nm)])
      cli::cli_abort(
        c("The groups must have unique names.",
          "x" = "Duplicated group name{?s}: {.val {dup}}."),
        class = "efa_group_duplicate_groups"
      )
    }

    group_data <- x
    names(group_data) <- nm
    return(list(group_data = group_data,
                input_type = if (isTRUE(is_cm[[1L]])) "cormat" else "raw"))
  }

  cli::cli_abort(
    c("{.arg x} must be raw data (with {.arg groups}) or a named list of per-group data sets.",
      "x" = "You supplied {.obj_type_friendly {x}}."),
    class = "efa_group_input"
  )
}


# Abort unless every group has the same items in the same order. Item identity is
# compared by column name when names are present, otherwise only the count.
.efa_group_check_items <- function(group_data) {
  ref_names <- colnames(group_data[[1L]])
  ref_p <- ncol(group_data[[1L]])

  for (i in seq_along(group_data)[-1L]) {
    gi <- group_data[[i]]
    if (ncol(gi) != ref_p) {
      cli::cli_abort(
        c("All groups must have the same items.",
          "x" = "Group {i} has {ncol(gi)} item{?s} but group 1 has {ref_p}."),
        class = "efa_group_unequal_items"
      )
    }
    ni <- colnames(gi)
    if (!identical(ni, ref_names)) {
      msg <- if (!is.null(ni) && !is.null(ref_names) && setequal(ni, ref_names)) {
        c("All groups must have their items in the same order.",
          "x" = "Group {i} has the same items as group 1 but in a different order.",
          "i" = "Reorder the columns so every group matches.")
      } else {
        c("All groups must have the same items.",
          "x" = "The item names in group {i} do not match group 1.")
      }
      cli::cli_abort(msg, class = "efa_group_unequal_items")
    }
  }
  invisible(TRUE)
}


# The N to pass to each EFA call: NA for raw data (EFA derives it from the data),
# a single value recycled or one value per group for correlation matrices.
.efa_group_resolve_N <- function(N, m, input_type) {
  if (input_type == "raw") {
    return(rep(list(NA), m))
  }
  if (length(N) == 1L) {
    return(rep(list(N), m))
  }
  if (length(N) != m) {
    cli::cli_abort(
      c("{.arg N} must be a single value or one value per group.",
        "x" = "You supplied {length(N)} value{?s} for {m} group{?s}."),
      class = "efa_group_bad_n"
    )
  }
  as.list(N)
}


# Resolve `reference_group` (a group name or an integer index) to a group index.
.efa_group_resolve_reference <- function(reference_group, group_names) {
  m <- length(group_names)
  if (length(reference_group) != 1L) {
    cli::cli_abort("{.arg reference_group} must be a single group name or index.",
                   class = "efa_group_bad_reference")
  }
  if (is.character(reference_group)) {
    idx <- match(reference_group, group_names)
    if (is.na(idx)) {
      cli::cli_abort(
        c("{.arg reference_group} must name one of the groups.",
          "x" = "{.val {reference_group}} is not one of {.val {group_names}}."),
        class = "efa_group_bad_reference"
      )
    }
    return(idx)
  }
  if (is.numeric(reference_group) && is.finite(reference_group) &&
      reference_group == round(reference_group) &&
      reference_group >= 1L && reference_group <= m) {
    return(as.integer(reference_group))
  }
  cli::cli_abort(
    c("{.arg reference_group} must be a group name or an integer between 1 and {m}.",
      "x" = "You supplied {reference_group}."),
    class = "efa_group_bad_reference"
  )
}


# Consensus alignment: rotate every group's unrotated loadings to a symmetric
# GPA-consensus target. Used for orthogonal and unrotated solutions only, so the
# factor correlations are always the identity and are reported as NULL.
.efa_group_align_consensus <- function(unrot_loadings, rot_loadings,
                                       rotation_type, group_names,
                                       rotation = "none", normalize = TRUE,
                                       rotation_args = list()) {
  consensus <- .gpa_consensus_target(
    unrotated_list = unrot_loadings,
    init_targets = if (rotation_type == "none") NULL else rot_loadings,
    rotation = "orthogonal",
    # The shared frame is order-invariant only to the accuracy the centroid iteration
    # reaches, and the iteration is seeded by the first group supplied. At the engine's
    # own default (1e-3) it stops after two outer iterations and leaves a residual of
    # about 3e-5 between group orders, which is large enough to move a borderline
    # salience flag. It is therefore run well past the point where the frame is
    # visually settled: the extra outer iterations are closed-form orthogonal
    # Procrustes steps (one k x k SVD and one p x k product per group), which is cheap
    # against the per-group fits that produced these loadings.
    tol = 1e-8
  )

  if (!isTRUE(consensus$converged)) {
    cli::cli_warn(
      c("The consensus alignment did not meet its convergence criterion.",
        "i" = "Inspect {.code alignment$history}; consider a reference group."),
      class = "efa_group_align_failed"
    )
  }

  # The GPA-consensus objective mean_g ||A_g T_g - M||^2 is invariant under a global
  # orthogonal rotation (M, T_g) -> (M Q, T_g Q), so the converged frame is oriented
  # only up to Q -- fixed in practice by whichever group seeds the iteration. Rotate
  # that frame into a canonical, order-invariant gauge and carry the identical
  # transform through every aligned matrix, so the reported congruences, differences,
  # and flags do not depend on the order the groups are supplied in. The column order
  # is then fixed by descending sum of squares and the column signs by
  # `.reflect_signs()`, as elsewhere.
  M <- unname(consensus$target)
  gauged <- .consensus_gauge(M, rotation, rotation_type, normalize, rotation_args)
  Q <- gauged$Q
  MQ <- M %*% Q
  ord <- order(colSums(MQ^2), decreasing = TRUE)
  signs <- .reflect_signs(MQ[, ord, drop = FALSE])
  D <- diag(signs, nrow = length(signs))
  gauge <- function(L) (unname(L) %*% Q)[, ord, drop = FALSE] %*% D

  # gauge(consensus$target) would recompute M %*% Q; MQ already holds it
  target <- MQ[, ord, drop = FALSE] %*% D
  dimnames(target) <- dimnames(consensus$target)

  loadings <- lapply(consensus$aligned_loadings, function(L) {
    Lg <- gauge(L)
    dimnames(Lg) <- dimnames(consensus$target)
    Lg
  })
  names(loadings) <- group_names

  list(loadings = loadings, target = target, Phi = NULL,
       alignment = consensus, gauge = gauged$gauge)
}


# Canonical orientation for the converged consensus frame M. Returns the transform `Q` that
# puts M into the simple structure of the criterion the caller asked for -- so the shared
# target looks like the per-group solutions it summarises instead of being reported in a
# frame borrowed from a different criterion -- together with a `gauge` label naming what was
# actually applied, which is not always derivable from the rotation (see the fallbacks below).
#
# Order-invariance comes from equivariance, not from luck: every orthogonal simple-structure
# criterion depends on the loadings alone, so if R minimizes f(M R) then Q0' R minimizes
# f(M Q0 R) and yields the SAME rotated matrix M R. A different group order hands the frame
# over as M Q0, and the gauge undoes exactly that Q0. What can break the argument is an
# optimizer that returns an inferior local optimum for one of the two orientations, which is
# why the gauge is run harder than the fits are: at a tolerance of 1e-10 rather than the fit's
# `precision`, and with a wider screen-and-triage budget than any criterion's own preset. The
# geomin criterion needs it -- at the shipped preset a minority of orientations land in a basin
# ~0.02 above the global optimum, which moves the reported frame by ~0.6 -- and the cost is
# under a millisecond, so the strong setting is applied to every criterion rather than
# special-cased. `random_starts` is pinned for the same reason: a user's `random_starts = 0`
# would leave the gauge on a single rational start and break the invariance outright.
#
# Kaiser normalization, by contrast, IS taken from the fit: it changes what the criterion
# optimizes exactly as `delta` does, so gauging with it on while the groups rotated with it
# off would put the shared frame in a different simple structure than the solutions it
# summarises. It does not put the invariance at risk, because the row weights are communalities
# and those are unchanged by a global orthogonal rotation of M.
#
# The principal-axes orientation is used whenever no criterion identifies a frame. Its columns
# are the eigenvectors of crossprod(M), which leaves t(M) %*% M diagonal with a decreasing
# diagonal -- the gauge an unrotated extraction reports anyway. Three cases reach it: an
# unrotated (or oblique) request, which has no criterion to borrow; a two-factor bifactor
# request, where the Jennrich-Bentler criterion sums lambda_ij^2 lambda_il^2 over j != l with
# j, l >= 2 and so is identically zero with a single group factor, making every rotation a
# global optimum and the gauge a no-op; and any engine failure.
.consensus_gauge <- function(M, rotation, rotation_type, normalize = TRUE,
                             rotation_args = list()) {

  # The engines require at least two factors, and SO(1) is trivial anyway: only the sign
  # gauge applies to a single column.
  if (ncol(M) < 2L) return(list(Q = diag(1), gauge = "identity"))

  # One decomposition serves every path: the principal-axes fallback and the equivariant
  # pre-rotation below read the same eigenvectors.
  V <- eigen(crossprod(M), symmetric = TRUE)$vectors
  principal_axes <- function(why) list(Q = V, gauge = why)

  if (rotation_type != "orthogonal") return(principal_axes("principal_axes"))
  if (rotation == "bifactorT" && ncol(M) < 3L) return(principal_axes("principal_axes"))

  # Hand the criterion a frame that does not depend on the group order, rather than trusting
  # it to reach the same optimum from two different orientations of the same loadings. The
  # equivariance above is a statement about the criterion's global optimum; the solver only
  # approximates it, and its random starts are drawn in the frame it is handed, so two
  # orientations explore genuinely different points. On a five-factor fixture that was enough
  # to leave geomin in an optimum 0.1% above the global one for one group order and not the
  # other, moving the reported target by 0.15 and changing three salience flags. Pre-rotating
  # to the principal axes -- themselves exactly equivariant, with the column order and signs
  # already pinned -- makes both orders hand the solver the same matrix, so they agree whether
  # or not it finds the global optimum. The returned gauge is the composition.
  P <- V
  MP <- M %*% P
  P <- P %*% diag(.reflect_signs(MP), nrow = ncol(MP))
  Mc <- M %*% P

  # varimax runs through .varimax_svd() rather than a native engine, so a degenerate frame
  # (a row with no loading at all) is gauged rather than lost: the floored Kaiser weights
  # keep the zero row from dividing by zero and aborting the SVD, as they do on every other
  # rotation path. The shared fallback below still catches a genuinely unusable frame.
  # `varimax_type` is deliberately not consulted: .VARIMAX_SPSS and stats::varimax optimize
  # the identical normal-varimax criterion (the SPSS pairwise angle is its closed-form plane
  # maximizer), so the gauged frame agrees to convergence tolerance under either variant.
  engine <- if (rotation == "varimax") {
    function(L, ...) list(Th = .varimax_svd(L, normalize = normalize, precision = 1e-10)$rotmat)
  } else {
    .orth_engines[[rotation]]
  }
  if (is.null(engine)) return(principal_axes("principal_axes"))

  Q <- tryCatch(
    withCallingHandlers(
      do.call(engine,
              c(list(Mc, eps = 1e-10, normalize = normalize, randomStarts = 100L,
                     screen_keep = 20L, triage_maxit = 100L,
                     triage_improve_tol = 0),
                rotation_args))$Th,
      # A gauge that stalled short of the tolerance is still an orthogonal frame, and the
      # fallback below catches a genuinely unusable one; the fits have already reported
      # their own convergence, so do not warn a second time about this internal rotation.
      efa_rotation_no_convergence = function(w) invokeRestart("muffleWarning")
    ),
    error = function(e) NULL
  )

  if (is.null(Q) || !is.matrix(Q) || !all(is.finite(Q))) {
    # The fallback is sound but must not be invisible: the printed output does not show
    # settings$gauge, so without a signal the user would read a principal-axes frame as the
    # requested criterion's. The alignment itself has succeeded at this point.
    cli::cli_warn(
      c("The {.val {rotation}} gauge rotation of the shared consensus frame failed;
         its principal-axes orientation is reported instead.",
        "i" = "The group alignment itself succeeded; see {.code settings$gauge}."),
      class = "efa_group_gauge_failed"
    )
    principal_axes("principal_axes")
  } else {
    list(Q = P %*% Q, gauge = rotation)
  }
}


# Reference alignment: keep the reference group's loadings fixed and Procrustes-
# rotate every other group's unrotated loadings onto them (orthogonally, or
# obliquely for an oblique rotation with more than one factor).
.efa_group_align_reference <- function(unrot_loadings, ref_loadings, ref_idx,
                                       ref_phi, oblique, group_names) {
  m <- length(unrot_loadings)
  proc_rotation <- if (oblique) "oblique" else "orthogonal"

  loadings <- vector("list", m)
  names(loadings) <- group_names
  phis <- if (oblique) stats::setNames(vector("list", m), group_names)
  procrustes <- stats::setNames(vector("list", m), group_names)
  valid <- rep(TRUE, m)

  loadings[[ref_idx]] <- ref_loadings
  if (oblique) phis[[ref_idx]] <- ref_phi

  for (g in seq_len(m)[-ref_idx]) {
    pr <- efa_procrustes(A = unrot_loadings[[g]], Target = ref_loadings,
                         rotation = proc_rotation)
    loadings[[g]] <- pr$loadings
    if (oblique) phis[[g]] <- pr$Phi
    procrustes[[g]] <- pr
    valid[g] <- isTRUE(pr$valid)
  }

  if (any(!valid)) {
    cli::cli_warn(
      c("At least one group could not be aligned to a valid rotation.",
        "i" = "Inspect {.code alignment$procrustes}; the best available alignment is used."),
      class = "efa_group_align_failed"
    )
  }

  alignment <- list(
    method = "reference",
    reference = group_names[[ref_idx]],
    reference_index = ref_idx,
    target = ref_loadings,
    procrustes = procrustes,
    valid = stats::setNames(valid, group_names)
  )

  list(loadings = loadings, target = ref_loadings,
       Phi = if (oblique) phis, alignment = alignment)
}


# Pairwise Tucker congruence between the aligned group loadings: the full
# factor-by-factor matrix for every group pair, the matched-factor congruence
# (each matrix's diagonal) as a groups-by-groups-by-factors array, and a flag for
# pairs whose congruence is undefined. A near-zero (or otherwise non-finite) factor
# makes `.tucker_congruence()` abort; that condition is caught and turned into an
# NA-flagged pair rather than failing the run, while a structural error (a dimension
# or finiteness violation of the loadings) is left to surface.
.efa_group_congruence <- function(loadings) {
  m <- length(loadings)
  group_names <- names(loadings)
  k <- ncol(loadings[[1L]])

  fac_names <- colnames(loadings[[1L]])
  if (is.null(fac_names)) fac_names <- paste0("F", seq_len(k))

  na_mat <- matrix(NA_real_, k, k, dimnames = list(fac_names, fac_names))

  matrices <- stats::setNames(vector("list", m), group_names)
  for (g in seq_len(m)) {
    matrices[[g]] <- stats::setNames(vector("list", m), group_names)
  }

  matched <- array(
    NA_real_, dim = c(m, m, k),
    dimnames = list(group_names, group_names, fac_names)
  )
  degenerate <- matrix(FALSE, m, m, dimnames = list(group_names, group_names))

  # Upper triangle and diagonal; the reverse pair is the transpose. A near-zero
  # factor makes the congruence undefined -- `.tucker_congruence()` aborts, which is
  # caught (only the undefined-congruence conditions, not a structural error) and
  # turned into an NA-filled, flagged pair rather than propagating the abort.
  for (i in seq_len(m)) {
    for (j in i:m) {
      cij <- tryCatch(
        .tucker_congruence(loadings[[i]], loadings[[j]]),
        efa_zero_column = function(e) NULL,
        efa_undefined_congruence = function(e) NULL
      )

      if (is.null(cij)) {
        matrices[[i]][[j]] <- na_mat
        degenerate[i, j] <- TRUE
        if (i != j) {
          matrices[[j]][[i]] <- na_mat
          degenerate[j, i] <- TRUE
        }
        next  # matched stays NA (the array is initialised to NA).
      }

      dimnames(cij) <- list(fac_names, fac_names)
      d <- diag(cij)
      matrices[[i]][[j]] <- cij
      matched[i, j, ] <- d
      if (i != j) {
        matrices[[j]][[i]] <- t(cij)
        matched[j, i, ] <- d
      }
    }
  }

  list(matrices = matrices, matched = matched, degenerate = degenerate)
}


# Drop the bootstrap SE/CI/replicate payload from a group fit so the stored EFA
# object reads as a point-estimate fit (out$efa is identical in shape whether or not
# a bootstrap was run; the replicate cube is captured separately before stripping).
.efa_strip_boot <- function(fit) {
  fit$SE <- NULL
  fit$CI <- NULL
  fit$replicates <- NULL
  fit$standardized_residuals <- NULL
  fit$settings$se <- "none"
  fit
}


# Percentile-bootstrap confidence intervals for the matched Tucker congruences and the
# per-item, per-factor loading differences between groups.
# `boot_cubes` holds one p x k x b cube of replicate unrotated loadings per group
# (from EFA(se = "np-boot")). Each replicate is re-aligned to the frozen point-estimate
# `target` with the same Procrustes mode as the point estimate ("orthogonal" for
# consensus / reference-orthogonal, "oblique" for reference-oblique), and the
# per-factor congruence and the pairwise loading differences are recomputed from the same
# re-aligned replicates for every group pair. A replicate contributes only if every group
# aligned in it; incomplete replicates are dropped with a single classed warning. Aggregation
# matches .boot_se_ci()/.array_se_ci(): the percentile interval at level `ci`, and the
# replicate SD as the standard error.
.efa_group_boot_congruence <- function(boot_cubes, target, proc_rotation, ci) {
  m <- length(boot_cubes)
  group_names <- names(boot_cubes)
  b <- dim(boot_cubes[[1L]])[3L]

  target <- as.matrix(target)
  p <- nrow(target)
  k <- ncol(target)
  fac_names <- colnames(target)
  if (is.null(fac_names)) fac_names <- paste0("F", seq_len(k))
  item_names <- rownames(target)
  if (is.null(item_names)) item_names <- paste0("V", seq_len(p))

  # Re-align every replicate of every group to the frozen target. The oblique solver
  # is left at its default random_starts = 0: warm-started from the closed-form
  # orthogonal Procrustes solution it is deterministic (draws no RNG), which keeps the
  # bootstrap worker-count-independent and makes the alignment invariant to the sign /
  # column permutation the replicate cubes already carry from their own alignment.
  aligned <- vector("list", m)
  valid <- matrix(FALSE, nrow = m, ncol = b)
  for (g in seq_len(m)) {
    cube <- boot_cubes[[g]]
    ag <- array(NA_real_, dim = c(p, k, b))
    for (i in seq_len(b)) {
      Li <- cube[, , i]
      # A replicate EFA() could not fit is NA-filled by EFA; skip it.
      if (!all(is.finite(Li))) next
      pr <- tryCatch(
        efa_procrustes(A = Li, Target = target, rotation = proc_rotation),
        error = function(e) NULL
      )
      if (is.null(pr) || !isTRUE(pr$valid)) next
      ag[, , i] <- pr$loadings
      valid[g, i] <- TRUE
    }
    aligned[[g]] <- ag
  }

  # A replicate is usable only when every group aligned in it.
  complete <- apply(valid, 2L, all)
  n_valid <- sum(complete)

  if (n_valid == 0L) {
    cli::cli_abort(
      c("All {b} bootstrap replicate{?s} failed; no congruence confidence intervals could be computed.",
        "i" = "The resampled correlation matrices may be degenerate; try more observations or fewer factors."),
      class = "efa_group_boot_all_failed"
    )
  }
  if (n_valid < b) {
    n_failed <- b - n_valid
    cli::cli_warn(
      c("{n_failed} bootstrap replicate{?s} failed and {?was/were} excluded.",
        "i" = "The congruence confidence intervals are based on {n_valid} replicate{?s}."),
      class = "efa_group_boot_failed"
    )
  }

  # Matched (per-factor) congruence between every group pair, per usable replicate.
  # Both members of a pair are aligned to the same target, so factor f of one matches
  # factor f of the other and the matched congruence is the diagonal of the pairwise
  # matrix. A near-zero factor makes `.tucker_congruence()` abort; only the degeneracy
  # conditions are caught (as NA), so a structural error still surfaces.
  matched_boot <- array(
    NA_real_, dim = c(m, m, k, b),
    dimnames = list(group_names, group_names, fac_names, NULL)
  )
  for (i in which(complete)) {
    for (a in seq_len(m)) {
      for (h in a:m) {
        d <- tryCatch(
          diag(.tucker_congruence(aligned[[a]][, , i], aligned[[h]][, , i])),
          efa_zero_column = function(e) rep(NA_real_, k),
          efa_undefined_congruence = function(e) rep(NA_real_, k)
        )
        matched_boot[a, h, , i] <- d
        if (a != h) matched_boot[h, a, , i] <- d
      }
    }
  }

  # se = SD of the replicates; ci = percentile interval at level `ci` (matches
  # .boot_se_ci()). Aggregation is over the replicate margin; a pair-factor cell that
  # is all-NA (every usable replicate degenerate there) yields a silent NA.
  l_ci <- (1 - ci) / 2
  probs <- c(l_ci, ci + l_ci)
  agg <- .array_se_ci(matched_boot, probs, M = c(1L, 2L, 3L))

  # Percentile CIs for the per-item, per-factor loading differences L_g - L_h between every
  # group pair, from the same re-aligned replicates. Both members of a pair share the frozen
  # target, so element (v, f) of one lines up with element (v, f) of the other and the
  # cellwise difference is well defined. Only the unordered pairs (g < h) are used downstream,
  # so the diagonal (a zero self-difference) and the lower triangle are left NA. Every complete
  # replicate has finite aligned loadings, so no all-NA quantile arises here. Percentile limits
  # only are needed (no SE), so aggregate the replicate margin directly rather than via
  # .array_se_ci(), which would also build a discarded SD array per pair.
  comp <- which(complete)
  diff_lower <- array(
    NA_real_, dim = c(m, m, p, k),
    dimnames = list(group_names, group_names, item_names, fac_names)
  )
  diff_upper <- diff_lower
  for (a in seq_len(m - 1L)) {
    for (h in (a + 1L):m) {
      dab <- aligned[[a]][, , comp, drop = FALSE] - aligned[[h]][, , comp, drop = FALSE]
      qs <- apply(dab, c(1L, 2L), stats::quantile, probs = probs, na.rm = TRUE)
      diff_lower[a, h, , ] <- qs[1L, , ]
      diff_upper[a, h, , ] <- qs[2L, , ]
    }
  }

  list(matched_se = agg$se, matched_ci = agg$ci, n_boot = n_valid,
       diff_ci = list(lower = diff_lower, upper = diff_upper))
}


# Cross-group loading differences on the already-aligned per-group loadings: a per-pair
# magnitude summary and a per-item, per-factor flag table. `loadings` is the named list of
# aligned matrices (all p x k, sharing the target's orientation), so the pairwise comparison
# is a pure difference -- the printless `.compare_loadings()` core does no reordering, and
# `decimals = FALSE` skips its decimal-agreement scan, which only the printed comparison
# reports. A cell is flagged when its absolute loading difference reaches `delta` (a
# descriptive salience heuristic, not a test). When `diff_ci` (bootstrap percentile limits from
# .efa_group_boot_congruence()) is supplied, each flag is paired with whether that cell's
# difference interval excludes zero.
.efa_group_diffs <- function(loadings, delta, diff_ci = NULL) {
  m <- length(loadings)
  group_names <- names(loadings)
  p <- nrow(loadings[[1L]])
  k <- ncol(loadings[[1L]])

  item_names <- rownames(loadings[[1L]])
  if (is.null(item_names)) item_names <- paste0("V", seq_len(p))
  fac_names <- colnames(loadings[[1L]])
  if (is.null(fac_names)) fac_names <- paste0("F", seq_len(k))

  # Column-major labels (item within factor), matching as.vector() of a p x k matrix.
  indicator <- rep(item_names, times = k)
  factor <- rep(fac_names, each = p)

  n_pairs <- m * (m - 1L) / 2L
  diffs_rows <- vector("list", n_pairs)
  flags_rows <- vector("list", n_pairs)
  r <- 0L
  for (g in seq_len(m - 1L)) {
    for (h in (g + 1L):m) {
      r <- r + 1L
      cmp <- .compare_loadings(loadings[[g]], loadings[[h]], corres = FALSE,
                               decimals = FALSE)
      d <- cmp$diff
      ad <- abs(d)
      flagged <- ad >= delta

      if (is.null(diff_ci)) {
        lo <- rep(NA_real_, p * k)
        up <- rep(NA_real_, p * k)
        exc <- rep(NA, p * k)
      } else {
        lo <- as.vector(diff_ci$lower[g, h, , ])
        up <- as.vector(diff_ci$upper[g, h, , ])
        exc <- (lo > 0) | (up < 0)
      }

      flags_rows[[r]] <- data.frame(
        group_1 = group_names[g],
        group_2 = group_names[h],
        indicator = indicator,
        factor = factor,
        diff = as.vector(d),
        abs_diff = as.vector(ad),
        flagged = as.vector(flagged),
        ci_lower = lo,
        ci_upper = up,
        ci_excludes_0 = exc,
        stringsAsFactors = FALSE
      )

      diffs_rows[[r]] <- data.frame(
        group_1 = group_names[g],
        group_2 = group_names[h],
        mean_abs_diff = cmp$mean_abs_diff,
        median_abs_diff = cmp$median_abs_diff,
        min_abs_diff = cmp$min_abs_diff,
        max_abs_diff = cmp$max_abs_diff,
        rmse = cmp$g,
        n_flagged = sum(flagged),
        stringsAsFactors = FALSE
      )
    }
  }

  list(
    diffs = do.call(rbind, diffs_rows),
    flags = do.call(rbind, flags_rows)
  )
}


# Approximate-invariance bands for a Tucker congruence, following Lorenzo-Seva & ten Berge
# (2006): phi >= .95 the factors are "equal", [.85, .95) "fair" similarity; congruences below
# .85 (below their named bands) are labelled "incongruent" here. Vectorised and NA-safe -- an
# undefined (e.g. degenerate) congruence stays NA rather than being classified.
.invariance_band <- function(phi) {
  out <- rep(NA_character_, length(phi))
  ok <- !is.na(phi)
  out[ok & phi >= 0.95] <- "equal"
  out[ok & phi >= 0.85 & phi < 0.95] <- "fair"
  out[ok & phi < 0.85] <- "incongruent"
  out
}


# Approximate-invariance verdict per group pair and factor from the matched Tucker
# congruences. Reads the point congruences (`congruence$matched`) and, when a bootstrap was
# run, their percentile CI lower bounds (`congruence$matched_ci$lower`). The verdict is read
# conservatively off the CI lower bound when available -- a factor is judged "equal" only if
# even the lower bound clears the .95 band -- otherwise off the point estimate, with the
# `phi_lower` column left NA to signal a non-conservative, point-based reading. The basis is
# chosen once by whether a bootstrap ran, so the whole table is uniformly conservative or
# uniformly point-based. A degenerate pair (NA congruence) yields an NA verdict.
.efa_group_invariance <- function(congruence) {
  matched <- congruence$matched
  dn <- dimnames(matched)
  group_names <- dn[[1L]]
  fac_names <- dn[[3L]]
  m <- dim(matched)[1L]
  k <- dim(matched)[3L]

  has_ci <- !is.null(congruence$matched_ci)
  lower <- if (has_ci) congruence$matched_ci$lower

  rows <- vector("list", m * (m - 1L) / 2L * k)
  r <- 0L
  for (g in seq_len(m - 1L)) {
    for (h in (g + 1L):m) {
      for (f in seq_len(k)) {
        r <- r + 1L
        phi <- matched[g, h, f]
        phi_lower <- if (has_ci) lower[g, h, f] else NA_real_
        basis <- if (has_ci) phi_lower else phi
        rows[[r]] <- data.frame(
          group_1 = group_names[g],
          group_2 = group_names[h],
          factor = fac_names[f],
          phi = phi,
          phi_lower = phi_lower,
          verdict = .invariance_band(basis),
          stringsAsFactors = FALSE
        )
      }
    }
  }
  do.call(rbind, rows)
}

# Tests for efa_group(): per-group fitting at a common number of factors and the
# shared consensus/reference alignment. Deterministic rotations (varimax, promax)
# and identical-group invariants are used so no random-start behaviour is relied
# on. Conditions are asserted by class, not message text.

cmat <- test_models$baseline$cormat
p <- ncol(cmat)

# Three copies of one correlation matrix: the identical-group case, where every aligned
# solution must collapse onto the shared target and every congruence must be one. PAF and
# varimax are both deterministic and neither draws a random start, so one fit serves every
# block that reads it -- and a second one with the invariance verdict switched on, since
# that adds tables without changing anything that is fitted.
ident_bands <- list(a = cmat, b = cmat, c = cmat)
ident_fit <- efa_group(ident_bands, n_factors = 3, N = 500, rotation = "varimax")
ident_fit_inv <- efa_group(ident_bands, n_factors = 3, N = 500, rotation = "varimax",
                           invariance = TRUE)

# The three WJIV age bands cut to their twelve leading variables: genuinely different
# correlation matrices over one item set, which is what an order- or gauge-invariance claim
# has to be demonstrated on (identical groups would satisfy it trivially).
wj_trio <- list(a = WJIV_ages_6_8$cormat[1:12, 1:12],
                b = WJIV_ages_14_19$cormat[1:12, 1:12],
                c = WJIV_ages_20_39$cormat[1:12, 1:12])
wj_trio_N <- c(WJIV_ages_6_8$N, WJIV_ages_14_19$N, WJIV_ages_20_39$N)


test_that("efa_group fits every group at a common k and returns an efa_group object", {
  mg <- ident_fit

  expect_s3_class(mg, "efa_group")
  expect_length(mg$loadings, 3)
  expect_named(mg$loadings, c("a", "b", "c"))
  expect_true(all(vapply(mg$efa, inherits, logical(1), "EFA")))
  expect_true(all(vapply(mg$efa, function(f) as.numeric(f$settings$n_factors),
                         numeric(1)) == 3))
  expect_identical(dim(mg$loadings[[1]]), c(p, 3L))
  expect_identical(dim(mg$target), c(p, 3L))
  expect_identical(mg$settings$alignment, "consensus")
  expect_null(mg$Phi)
})


test_that("identical groups align to a shared consensus target", {
  mg <- ident_fit

  # With identical inputs every aligned solution collapses onto the target.
  expect_equal(mg$loadings$a, mg$loadings$b, tolerance = 1e-6, ignore_attr = TRUE)
  expect_equal(mg$loadings$a, mg$loadings$c, tolerance = 1e-6, ignore_attr = TRUE)
  expect_equal(mg$loadings$a, mg$target, tolerance = 1e-6, ignore_attr = TRUE)
})


test_that("consensus alignment is invariant to the order of the groups", {
  skip_on_cran()
  # The consensus frame is only identified up to a global rotation of its factors, so
  # without a fixed gauge every reported statistic would depend on which group seeds the
  # iteration -- i.e. on the order the groups are supplied in. The gauge is the simple
  # structure of the requested rotation (varimax here) evaluated on the consensus target,
  # and it removes that dependence: three genuinely different WJIV age bands are aligned
  # in every group order and must yield the same shared frame and statistics.
  fit <- function(o) {
    efa_group(wj_trio[o], n_factors = 2, N = wj_trio_N[o],
              rotation = "varimax", invariance = TRUE)
  }
  # All six orders, not just the reversal: the centroid iteration is seeded by the
  # first group supplied, so each order hands the gauge a different starting frame.
  perms <- list(c(1, 2, 3), c(1, 3, 2), c(2, 1, 3),
                c(2, 3, 1), c(3, 1, 2), c(3, 2, 1))
  fits <- lapply(perms, fit)
  mg1 <- fits[[1L]]

  # Salience flags compared per unordered pair, item, and factor: the absolute
  # difference and the flag are orientation-independent (only the signed diff flips
  # when a pair is listed the other way round).
  flag_key <- function(fl) {
    pair <- vapply(seq_len(nrow(fl)), function(i) {
      paste(sort(c(fl$group_1[i], fl$group_2[i])), collapse = "|")
    }, character(1L))
    o <- order(paste(pair, fl$indicator, fl$factor, sep = "|"))
    list(abs_diff = fl$abs_diff[o], flagged = fl$flagged[o])
  }
  k1 <- flag_key(mg1$flags)

  # The seeding group is surfaced by name so a frame can be reproduced; it is the
  # first group supplied, so it changes with the order. The reference path has none.
  expect_identical(
    vapply(fits, function(f) f$settings$alignment_start, character(1L)),
    vapply(perms, function(o) names(wj_trio)[o[1L]], character(1L))
  )

  # The residual is the consensus iteration's own convergence floor and nothing else,
  # so the tolerances are set from it rather than from the reporting precision: the
  # worst deviation observed over these six orders is ~3e-12 on the loadings. They are
  # deliberately far tighter than the ~3e-5 a loosely converged centroid leaves behind,
  # which is the scale at which a borderline salience flag could still flip.
  for (i in seq_along(fits)[-1L]) {
    mg2 <- fits[[i]]
    info <- paste(names(wj_trio)[perms[[i]]], collapse = "")

    # The gauge fixes the shared frame, so each group's aligned loadings and the
    # shared target match however the groups are ordered.
    for (g in c("a", "b", "c")) {
      expect_equal(mg1$loadings[[g]], mg2$loadings[[g]], tolerance = 1e-8,
                   ignore_attr = TRUE, info = paste(info, g))
    }
    expect_equal(mg1$target, mg2$target, tolerance = 1e-8, ignore_attr = TRUE,
                 info = info)

    # ... and hence every reported statistic. Matched congruences are indexed by group
    # name, so they are directly comparable across orderings.
    for (pair in list(c("a", "b"), c("a", "c"), c("b", "c"))) {
      expect_equal(mg1$congruence$matched[pair[1L], pair[2L], ],
                   mg2$congruence$matched[pair[1L], pair[2L], ], tolerance = 1e-8,
                   info = paste(info, paste(pair, collapse = "-")))
    }

    k2 <- flag_key(mg2$flags)
    expect_equal(k1$abs_diff, k2$abs_diff, tolerance = 1e-8, info = info)
    expect_identical(k1$flagged, k2$flagged, info = info)
  }
  expect_null(efa_group(wj_trio[1:2], n_factors = 2, N = wj_trio_N[1:2],
                        rotation = "varimax",
                        reference_group = "a")$settings$alignment_start)
})


test_that("an unrotated consensus solution uses the principal-axes gauge", {
  skip_on_cran()
  # With no requested rotation there is no simple structure to borrow, so the shared
  # frame is fixed by the principal axes of the target: t(M) %*% M diagonal with a
  # decreasing diagonal and non-negative column sums. That gauge is order-invariant too.
  fit <- function(o) {
    efa_group(wj_trio[o], n_factors = 2, N = wj_trio_N[o], rotation = "none")
  }
  mg1 <- fit(c(1, 2, 3))
  mg2 <- fit(c(3, 2, 1))

  expect_identical(mg1$settings$alignment, "consensus")

  G <- crossprod(unclass(mg1$target))
  expect_lt(max(abs(G[upper.tri(G)])), 1e-8)        # t(M) %*% M is diagonal
  expect_false(is.unsorted(rev(diag(G))))           # ... with a decreasing diagonal
  expect_true(all(colSums(unclass(mg1$target)) >= 0))

  # ... and, as on the varimax path, the frame does not depend on the group order.
  expect_equal(mg1$target, mg2$target, tolerance = 1e-8, ignore_attr = TRUE)
  expect_equal(mg1$congruence$matched["a", "b", ],
               mg2$congruence$matched["a", "b", ], tolerance = 1e-8)
})


# Best-|phi| column matching between two loading matrices. Needed because a group's OWN
# rotated solution and the shared target need not be in the same column order, so the
# question "is the shared frame the same kind of structure" is about the columns as a set.
# This is not what efa_group reports: `congruence` is positional (`diag(cij)`), since the
# groups have already been rotated into the shared frame by then.
matched_phi <- function(A, B) {
  P <- abs(.tucker_congruence(unclass(A), unclass(B)))
  used <- integer(0)
  vapply(seq_len(ncol(P)), function(j) {
    cand <- setdiff(seq_len(ncol(P)), used)
    i <- cand[which.max(P[j, cand])]
    used <<- c(used, i)
    P[j, i]
  }, numeric(1))
}


test_that("the consensus gauge follows the requested rotation criterion", {
  skip_on_cran()
  # The shared frame is put in the simple structure of the rotation that was asked for, not
  # in a borrowed one. bifactorT is the sharpest case: its per-group solutions carry a
  # general factor loading on every variable, which a varimax gauge destroys -- that gauge
  # returned a 9/4/6 salience split with a leading column whose smallest |loading| was .14,
  # and matched congruences against the per-group solutions of only .74-.81.
  mg <- suppressWarnings(
    efa_group(wj_trio, n_factors = 3, N = wj_trio_N,
              rotation = "bifactorT", seed = 11))
  tgt <- unclass(mg$target)
  expect_identical(mg$settings$gauge, "bifactorT")

  # a general column loading saliently on every variable, and two sparse group columns
  expect_identical(sum(abs(tgt[, 1L]) >= .3), nrow(tgt))
  expect_gt(min(abs(tgt[, 1L])), .3)
  expect_true(all(colSums(abs(tgt[, -1L, drop = FALSE]) >= .3) < nrow(tgt) / 2))

  # and a frame the per-group solutions actually agree with
  for (g in names(mg$efa)) {
    expect_gt(mean(matched_phi(tgt, mg$efa[[g]]$rot_loadings)), .90)
  }
})


test_that("the criterion gauge is invariant to the order of the groups", {
  skip_on_cran()
  # Order-invariance is not specific to varimax: every orthogonal criterion depends on the
  # loadings alone, so rotating M and rotating M %*% Q0 reach the same rotated matrix and
  # the gauge undoes whichever Q0 the group order produced. The residual is the consensus
  # iteration's own convergence floor, not the gauge's, so the tolerance is set from that.
  fit <- function(o, rot) {
    suppressWarnings(
      efa_group(wj_trio[o], n_factors = 3, N = wj_trio_N[o],
                rotation = rot, seed = 11))
  }

  for (rot in c("quartimax", "geominT", "bifactorT")) {
    mg1 <- fit(c(1, 2, 3), rot)
    mg2 <- fit(c(3, 2, 1), rot)
    expect_equal(mg1$target, mg2$target, tolerance = 1e-6, ignore_attr = TRUE,
                 info = rot)
    for (g in c("a", "b", "c")) {
      expect_equal(mg1$loadings[[g]], mg2$loadings[[g]], tolerance = 1e-6,
                   ignore_attr = TRUE, info = paste(rot, g))
    }
    # These criteria are optimised iteratively rather than in closed form, so the gauge
    # adds its own stopping slack on top of the consensus floor: the worst deviation
    # observed here is 4e-9 (bifactorT), against 3e-12 on the varimax path above.
    expect_equal(mg1$congruence$matched["a", "b", ],
                 mg2$congruence$matched["a", "b", ], tolerance = 1e-6, info = rot)
  }
})


test_that("the gauge is equivariant: it undoes an arbitrary reorientation of the frame", {
  skip_on_cran()
  # The property the whole gauge rests on, tested directly on .consensus_gauge() rather than
  # through efa_group(): gauging M and gauging a reoriented copy M %*% Q0 must land on the
  # same frame, since that is exactly what a different group order hands over. It is not
  # enough that the criterion's global optimum has this property -- the solver only
  # approximates it, and it draws its random starts in whatever frame it is given, so two
  # orientations explore different points. A five-factor frame with cross-loadings is where
  # that bites: without the principal-axes pre-rotation the geomin gauge settled 0.1% above
  # the global optimum for some orientations and not others, moving the frame by 0.15.
  set.seed(4004)
  p <- 20L; k <- 5L
  M <- matrix(stats::rnorm(p * k, 0, 0.15), p, k)
  M[cbind(seq_len(p), rep(seq_len(k), length.out = p))] <- stats::runif(p, 0.4, 0.85)
  M <- M / sqrt(max(rowSums(M^2)) / 0.8)

  canon <- function(L) {
    ord <- order(colSums(L^2), decreasing = TRUE)
    L <- L[, ord, drop = FALSE]
    L %*% diag(.reflect_signs(L), nrow = ncol(L))
  }

  for (rot in c("varimax", "quartimax", "equamax", "geominT", "bentlerT", "bifactorT")) {
    devs <- vapply(1:5, function(i) {
      set.seed(700 + i)
      Q0 <- qr.Q(qr(matrix(stats::rnorm(k * k), k, k)))
      if (i %% 2L == 0L) Q0[, 1L] <- -Q0[, 1L]          # cover det(Q0) = -1 as well
      set.seed(11)
      A <- canon(M %*% .consensus_gauge(M, rot, "orthogonal")$Q)
      set.seed(11)
      B <- canon((M %*% Q0) %*% .consensus_gauge(M %*% Q0, rot, "orthogonal")$Q)
      max(abs(A - B))
    }, numeric(1))
    expect_lt(max(devs), 1e-6, label = paste0("max gauge deviation for ", rot))
  }
})


test_that("a criterion parameter set on the fits also reaches the gauge", {
  skip_on_cran()
  # `delta` changes what geomin optimizes, so gauging at the default while the groups were
  # rotated at another value would leave the shared frame in a different simple structure
  # than the solutions it summarises. It can only travel by the rotate control: `delta` is
  # efa_group()'s own salience-flag argument, so it can never arrive through the dots, and
  # the fit settings do not record it. Kaiser normalization is the same kind of setting and
  # is taken from the fit settings, where it is recorded.
  fit <- function(...) {
    suppressWarnings(efa_group(wj_trio, n_factors = 3, N = wj_trio_N,
                               rotation = "geominT", seed = 11, ...))
  }
  base <- fit()
  wide <- fit(rotate_control = rotate_control(delta = 0.5))
  expect_false(isTRUE(all.equal(unclass(base$target), unclass(wide$target),
                                tolerance = 1e-3)))

  # directly on the gauge, so the assertion does not depend on the fits also moving
  M <- unname(base$alignment$target)
  expect_false(isTRUE(all.equal(
    .consensus_gauge(M, "geominT", "orthogonal")$Q,
    .consensus_gauge(M, "geominT", "orthogonal",
                     rotation_args = list(delta = 0.5))$Q,
    tolerance = 1e-3)))

  # normalization likewise reaches it, and from the fit settings rather than the control
  expect_false(isTRUE(all.equal(
    .consensus_gauge(M, "geominT", "orthogonal", normalize = TRUE)$Q,
    .consensus_gauge(M, "geominT", "orthogonal", normalize = FALSE)$Q,
    tolerance = 1e-3)))
})


test_that("a two-factor bifactor request falls back to the principal-axes gauge", {
  skip_on_cran()
  # The Jennrich-Bentler orthogonal bifactor criterion sums lambda_ij^2 lambda_il^2 over
  # j != l with j, l >= 2, so with a single group factor the sum is empty and the criterion
  # is identically zero: every rotation is a global optimum, the engine returns the identity,
  # and the gauge would be a no-op leaving the frame wherever the iteration happened to stop
  # (loadings moved by more than 1.0 between group orders). The principal-axes gauge takes
  # over, which is both order-invariant and what the per-group solutions look like anyway.
  fit <- function(o) {
    suppressWarnings(
      efa_group(wj_trio[o], n_factors = 2, N = wj_trio_N[o],
                rotation = "bifactorT", seed = 11))
  }
  mg1 <- fit(c(1, 2, 3))
  mg2 <- fit(c(3, 2, 1))

  # the fallback is reported, since `rotation` alone does not reveal it
  expect_identical(mg1$settings$gauge, "principal_axes")

  G <- crossprod(unclass(mg1$target))
  expect_lt(max(abs(G[upper.tri(G)])), 1e-8)        # the principal-axes signature
  expect_false(is.unsorted(rev(diag(G))))
  expect_true(all(colSums(unclass(mg1$target)) >= 0))

  expect_equal(mg1$target, mg2$target, tolerance = 1e-4, ignore_attr = TRUE)
})


test_that(".consensus_gauge falls back where no criterion identifies a frame", {
  M <- matrix(c(0.8, 0.7, 0.6, 0.2, 0.1, -0.3, 0.5, 0.4), nrow = 4)
  pa <- eigen(crossprod(M), symmetric = TRUE)$vectors

  # a single factor: the engines require two, and SO(1) leaves only the sign gauge
  one <- .consensus_gauge(M[, 1L, drop = FALSE], "quartimax", "orthogonal")
  expect_identical(one$Q, diag(1))
  expect_identical(one$gauge, "identity")

  # nothing to borrow from an unrotated or an oblique request, and nothing to run for a
  # rotation with no orthogonal engine of its own
  for (args in list(list("none", "none"), list("promax", "oblique"),
                    list("not_a_rotation", "orthogonal"))) {
    g <- .consensus_gauge(M, args[[1L]], args[[2L]])
    expect_equal(g$Q, pa, info = args[[1L]])
    expect_identical(g$gauge, "principal_axes", info = args[[1L]])
  }

  # ... whereas a criterion rotation really runs its engine. Orthogonality alone would not
  # show that -- the principal-axes fallback is orthogonal too -- so check that the frame
  # differs from the fallback and that it lowers the criterion the caller asked for.
  g <- .consensus_gauge(M, "quartimax", "orthogonal")
  expect_identical(g$gauge, "quartimax")
  expect_equal(crossprod(g$Q), diag(ncol(M)), tolerance = 1e-10, ignore_attr = TRUE)
  expect_false(isTRUE(all.equal(unname(g$Q), unname(pa))))
  quartimax_crit <- function(L) -sum(L^4)
  expect_lt(quartimax_crit(M %*% g$Q), quartimax_crit(M %*% pa))
})


test_that("the alignment method follows the rotation and reference_group", {
  bands <- list(a = cmat, b = cmat)
  expect_identical(
    efa_group(bands, n_factors = 3, N = 500, rotation = "varimax")$settings$alignment,
    "consensus")
  expect_identical(
    efa_group(bands, n_factors = 3, N = 500, rotation = "varimax",
              reference_group = 1)$settings$alignment,
    "reference")
})


test_that("reference alignment keeps the reference group fixed", {
  bands <- list(a = cmat, b = cmat)
  mg <- efa_group(bands, n_factors = 3, N = 500, rotation = "varimax",
                  reference_group = "b")

  expect_identical(mg$settings$alignment, "reference")
  expect_identical(mg$settings$reference_group, "b")
  # The reference group's aligned loadings are its own rotated loadings, and are
  # the shared target.
  expect_equal(mg$loadings$b, mg$efa$b$rot_loadings, ignore_attr = TRUE)
  expect_equal(mg$target, mg$efa$b$rot_loadings, ignore_attr = TRUE)
  # Identical data -> the other group aligns exactly onto the reference.
  expect_equal(mg$loadings$a, mg$loadings$b, tolerance = 1e-6, ignore_attr = TRUE)
})


test_that("an oblique rotation routes to reference-Procrustes with a note", {
  bands <- list(a = cmat, b = cmat)

  # The note fires and the requested rotation is not silently changed. The fit it announces
  # is captured from that same call, so the routing is read off the run that was asserted on.
  expect_message(
    mg <- efa_group(bands, n_factors = 3, N = 500, rotation = "promax"),
    class = "efa_group_oblique_reference"
  )

  expect_identical(mg$settings$alignment, "reference")
  expect_identical(mg$settings$rotation, "promax")
  expect_false(is.null(mg$Phi))
  expect_length(mg$Phi, 2)
  expect_identical(dim(mg$Phi[[1]]), c(3L, 3L))
})


test_that("a list of correlation matrices disables the bootstrap and threads N", {
  b1 <- WJIV_ages_14_19$cormat[1:12, 1:12]
  b2 <- WJIV_ages_20_39$cormat[1:12, 1:12]
  mg <- suppressWarnings(
    efa_group(list(young = b1, adult = b2), n_factors = 2,
              N = c(WJIV_ages_14_19$N, WJIV_ages_20_39$N), rotation = "varimax"))

  expect_identical(mg$settings$input_type, "cormat")
  expect_false(mg$settings$can_bootstrap)
  expect_equal(unname(mg$settings$N),
               c(WJIV_ages_14_19$N, WJIV_ages_20_39$N))
  expect_length(mg$loadings, 2)
  expect_identical(dim(mg$loadings$young), c(12L, 2L))
})


test_that("a single factor aligns by sign only", {
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  mg <- suppressMessages(efa_group(GRiPS_raw, groups = g, n_factors = 1))

  expect_identical(mg$settings$input_type, "raw")
  expect_true(mg$settings$can_bootstrap)
  expect_identical(ncol(mg$target), 1L)
  # k = 1: the aligned loadings differ from the per-group unrotated loadings only
  # by a global sign.
  expect_equal(abs(mg$loadings$g1), abs(unclass(mg$efa$g1$unrot_loadings)),
               tolerance = 1e-8, ignore_attr = TRUE)
})


test_that("unused factor levels in groups do not create phantom groups", {
  g <- factor(rep(c("a", "b"), length.out = nrow(GRiPS_raw)),
              levels = c("a", "b", "c"))
  mg <- suppressMessages(efa_group(GRiPS_raw, groups = g, n_factors = 1))

  expect_length(mg$loadings, 2)
  expect_named(mg$loadings, c("a", "b"))
  expect_named(mg$efa, c("a", "b"))
})


test_that("Tucker congruence matches psych::factor.congruence and is symmetric", {
  skip_on_cran()
  skip_if_not_installed("psych")

  b1 <- WJIV_ages_14_19$cormat[1:12, 1:12]
  b2 <- WJIV_ages_20_39$cormat[1:12, 1:12]
  mg <- efa_group(list(young = b1, adult = b2), n_factors = 2,
                  N = c(WJIV_ages_14_19$N, WJIV_ages_20_39$N),
                  rotation = "varimax")

  cong <- mg$congruence

  # Full pairwise matrices, a groups x groups x factors matched diagonal, and no
  # degeneracy on a regular fit.
  expect_named(cong, c("matrices", "matched", "degenerate"))
  expect_identical(dim(cong$matched), c(2L, 2L, 2L))
  expect_false(any(cong$degenerate))

  # The full factor-by-factor matrix equals psych's Tucker congruence to ~1e-12.
  # psych rounds its return value to `digits`, so ask for full precision.
  ref <- psych::factor.congruence(mg$loadings$young, mg$loadings$adult,
                                  digits = 15)
  expect_equal(cong$matrices$young$adult, ref, tolerance = 1e-12,
               ignore_attr = TRUE)
  # The matched diagonal is that matrix's diagonal.
  expect_equal(cong$matched["young", "adult", ], diag(ref), tolerance = 1e-12,
               ignore_attr = TRUE)

  # The reverse pair is checked against an independent psych call with the groups
  # swapped (not merely the transpose of the forward pair), so a dropped or wrong
  # transpose would be caught.
  ref_rev <- psych::factor.congruence(mg$loadings$adult, mg$loadings$young,
                                      digits = 15)
  expect_equal(cong$matrices$adult$young, ref_rev, tolerance = 1e-12,
               ignore_attr = TRUE)
  expect_equal(cong$matched["adult", "young", ], diag(ref_rev), tolerance = 1e-12,
               ignore_attr = TRUE)
})


test_that("identical groups have matched congruence of one", {
  mg <- ident_fit

  # Matched-factor congruence between identical groups is 1 across every pair.
  expect_equal(as.numeric(mg$congruence$matched), rep(1, 3 * 3 * 3),
               tolerance = 1e-6)
  # A group with itself is exactly congruent, factor by factor.
  expect_equal(diag(mg$congruence$matrices$a$a), rep(1, 3), tolerance = 1e-12,
               ignore_attr = TRUE)
  expect_false(any(mg$congruence$degenerate))
})


test_that("a degenerate factor yields NA congruence and a flag, not an error", {
  set.seed(123)
  clean <- matrix(stats::rnorm(12), nrow = 6, ncol = 2)
  degen <- clean
  degen[, 2] <- 1e-14  # an all-near-zero factor

  # The near-zero factor must not abort the run.
  cong <- .efa_group_congruence(list(g1 = clean, g2 = degen))

  # Every pair touching the degenerate group is NA and flagged in both directions.
  expect_true(all(is.na(cong$matched["g2", , ])))
  expect_true(all(is.na(cong$matched[, "g2", ])))
  expect_true(all(is.na(cong$matrices$g1$g2)))
  expect_true(all(is.na(cong$matrices$g2$g1)))
  expect_true(cong$degenerate["g1", "g2"])
  expect_true(cong$degenerate["g2", "g1"])
  expect_true(cong$degenerate["g2", "g2"])

  # The clean group's self-pair is unaffected and finite.
  expect_false(cong$degenerate["g1", "g1"])
  expect_equal(diag(cong$matrices$g1$g1), rep(1, 2), tolerance = 1e-12,
               ignore_attr = TRUE)
  expect_true(all(is.finite(cong$matched["g1", "g1", ])))
})


test_that("efa_group guards its inputs with classed conditions", {
  # Same item count, different names -> unequal items.
  m1 <- cmat[1:5, 1:5]
  m2 <- m1
  dimnames(m2) <- list(paste0("z", 1:5), paste0("z", 1:5))
  expect_error(efa_group(list(m1, m2), n_factors = 1, N = 500),
               class = "efa_group_unequal_items")

  # Fewer than two groups.
  expect_error(efa_group(list(only = cmat), n_factors = 3, N = 500),
               class = "efa_group_too_few_groups")

  # Under-identified model (p = 4, k = 2 -> df = -1).
  m4 <- cmat[1:4, 1:4]
  expect_error(efa_group(list(m4, m4), n_factors = 2, N = 500),
               class = "efa_group_under_identified")

  # A correlation matrix cannot be split into groups.
  expect_error(efa_group(cmat, groups = rep(1:2, length.out = ncol(cmat)),
                         n_factors = 2, N = 500),
               class = "efa_group_cormat_needs_list")

  # Single raw data set without a grouping vector.
  expect_error(efa_group(GRiPS_raw, n_factors = 1),
               class = "efa_group_needs_groups")

  # A grouping vector supplied together with a list.
  expect_error(efa_group(list(cmat, cmat), groups = c(1, 2), n_factors = 3, N = 500),
               class = "efa_group_groups_with_list")

  # A list mixing raw data and a correlation matrix.
  raw <- as.data.frame(GRiPS_raw[1:50, ])
  cor_raw <- stats::cor(GRiPS_raw, use = "pairwise")
  expect_error(efa_group(list(raw, cor_raw), n_factors = 1),
               class = "efa_group_mixed_input")

  # An unknown reference group.
  expect_error(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500,
                         reference_group = "nope"),
               class = "efa_group_bad_reference")

  # Duplicated group names in a list.
  expect_error(efa_group(list(a = cmat, a = cmat), n_factors = 3, N = 500),
               class = "efa_group_duplicate_groups")

  # A confidence level outside (0, 1).
  expect_error(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500, ci = 1.5),
               class = "efa_group_bad_ci")

  # N length neither 1 nor one-per-group (correlation-matrix input).
  expect_error(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = c(100, 200, 300)),
               class = "efa_group_bad_n")

  # A single bootstrap replicate has no spread to build an interval from.
  expect_error(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500, b_boot = 1),
               class = "efa_group_bad_b_boot")

  # A grouping vector shorter than the data it splits.
  expect_error(efa_group(GRiPS_raw, groups = rep(1:2, length.out = 50), n_factors = 1),
               class = "efa_group_groups_length")

  # A list element that is neither a matrix nor a data frame.
  expect_error(efa_group(list(a = cmat, b = 1:5), n_factors = 3, N = 500),
               class = "efa_group_input")

  # Neither raw data with `groups` nor a list of per-group data sets.
  expect_error(efa_group(1:5, n_factors = 3, N = 500),
               class = "efa_group_input")

  # A misspelled argument in the dots is refused before any group is fitted, so it is
  # not reported as a statistical failure of whichever group happened to be fitted first.
  expect_error(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500, bogus = 1),
               class = "efa_unused_dots")

  # the guard reads the names only, so the rejection wins over evaluating the value
  expect_error(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500,
                         bogus = stop("must not be evaluated")),
               class = "efa_unused_dots")
})

test_that("a failing group names the group and hands the cause on unaltered", {
  # The re-label carries the cause as the condition's parent, which rlang renders in full
  # under "Caused by error:" with its own bullets intact. Repeating the parent's message as a
  # bullet of this one used to flatten it: every marker of a multi-bullet cause -- most of
  # efa_fit()'s input guards raise one -- was rendered inline, mid-sentence.
  e <- tryCatch(efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500,
                          estimator = "DWLS"),
                efa_group_fit_failed = function(e) e)
  expect_s3_class(e, "efa_group_fit_failed")
  # The cause keeps its own class, so a caller can still dispatch on it.
  expect_s3_class(e$parent, "efa_dwls_no_acov")
  # ... and it is still readable from the message, which rlang builds from the whole chain.
  expect_snapshot(cat(conditionMessage(e)))
})

test_that("efa_group forwards genuine efa_fit arguments and rotation extras", {
  # the guard above must not refuse what the per-group fits legitimately consume
  expect_s3_class(
    efa_group(list(a = cmat, b = cmat), n_factors = 3, N = 500,
              estimator = "ML", rotation = "oblimin", maxit = 750),
    "efa_group"
  )
})


test_that("rows with a missing group value are dropped with a classed warning", {
  x <- GRiPS_raw[1:200, ]
  g <- rep(c("a", "b"), each = 100)
  g[1:2] <- NA

  expect_warning(mg <- efa_group(x, groups = g, n_factors = 1),
                 class = "efa_group_na_group")
  # the two rows are dropped, not assigned to a phantom third group
  expect_length(mg$loadings, 2)
  expect_named(mg$loadings, c("a", "b"))
})


# Bootstrap percentile CIs for the between-group Tucker congruences. GRiPS_raw is well
# conditioned, so every replicate should fit; a genuine (not identical-group) split
# keeps the congruences comfortably below the 1.0 ceiling so the point estimate sits
# inside the percentile interval.

test_that("a bootstrap adds percentile CIs to the matched congruences", {
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  mg <- suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "varimax",
              estimator = "PAF", b_boot = 60, seed = 2024)))

  cong <- mg$congruence
  expect_named(cong, c("matrices", "matched", "degenerate",
                       "matched_se", "matched_ci", "n_boot"))
  expect_identical(dim(cong$matched_ci$lower), c(2L, 2L, 2L))
  expect_identical(dim(cong$matched_ci$upper), c(2L, 2L, 2L))
  expect_identical(dim(cong$matched_se), c(2L, 2L, 2L))
  expect_identical(dimnames(cong$matched_ci$lower),
                   list(c("g1", "g2"), c("g1", "g2"), c("F1", "F2")))

  # Every replicate fits on well-conditioned data.
  expect_identical(cong$n_boot, 60L)
  expect_identical(mg$settings$b_boot, 60L)
  expect_identical(mg$settings$ci, 0.95)

  # A valid interval everywhere; a self-pair is exactly congruent with itself.
  expect_true(all(cong$matched_ci$lower <= cong$matched_ci$upper))
  expect_equal(diag(cong$matched_ci$lower[, , 1]), c(1, 1), ignore_attr = TRUE)
  expect_equal(diag(cong$matched_ci$upper[, , 1]), c(1, 1), ignore_attr = TRUE)
})


test_that("the bootstrap CIs bracket the point congruence", {
  skip_on_cran()
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  mg <- suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "varimax",
              estimator = "PAF", b_boot = 300, seed = 2024)))

  pt <- mg$congruence$matched
  lo <- mg$congruence$matched_ci$lower
  up <- mg$congruence$matched_ci$upper

  # The point estimate lies inside the percentile interval (a small tolerance guards
  # the Monte Carlo edge; the congruences here are ~0.97, comfortably below 1).
  expect_true(all(lo - 0.01 <= pt & pt <= up + 0.01, na.rm = TRUE))
  # Off-diagonal intervals are non-degenerate; the self-pairs are the [1, 1] ceiling.
  expect_true(up["g1", "g2", 1] > lo["g1", "g2", 1])
  expect_equal(as.numeric(pt[cbind(1:2, 1:2, 1)]), c(1, 1), tolerance = 1e-8)
})


test_that("np-boot congruence CIs are reproducible at 1 vs 2 workers", {
  skip_on_cran()
  skip_if_not_slow()
  # The replicate fits are parallelised across workers inside each per-group EFA() with
  # future.apply (future.seed = TRUE), while efa_group's re-alignment is deterministic,
  # so a fixed `seed` must return the same intervals regardless of the number of
  # workers. The multisession workers are fresh R processes that load the installed
  # package, so run this under devtools::check() / after devtools::install() for the
  # worker fits to match the main process. (multicore is unavailable on Windows.)
  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  run <- function() suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "varimax",
              estimator = "PAF", b_boot = 12, seed = 2024)))

  future::plan(future::sequential)
  one <- run()

  future::plan(future::multisession, workers = 2)
  two <- run()

  expect_equal(one$congruence$matched_ci, two$congruence$matched_ci, tolerance = 1e-10)
  expect_equal(one$congruence$matched_se, two$congruence$matched_se, tolerance = 1e-10)
  expect_identical(one$congruence$n_boot, two$congruence$n_boot)
})


test_that("a supplied seed makes the CIs reproducible and restores the RNG", {
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  boot <- function() suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "varimax",
              estimator = "PAF", b_boot = 30, seed = 99)))

  set.seed(1)
  state_before <- get(".Random.seed", envir = globalenv(), inherits = FALSE)
  one <- boot()
  # The seeded bootstrap leaves the caller's stream untouched.
  expect_identical(state_before,
                   get(".Random.seed", envir = globalenv(), inherits = FALSE))

  two <- boot()
  expect_equal(one$congruence$matched_ci, two$congruence$matched_ci)

  # the seed is part of the record, so a saved object says how it was seeded
  expect_equal(one$settings$seed, 99)
  expect_null(suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "varimax",
              estimator = "PAF")))$settings$seed)
})


test_that("a supplied seed covers the per-group fits without a bootstrap", {
  # `seed` applies whether or not congruence intervals are requested: the per-group fits
  # are stochastic on their own when the rotation draws random starts. simplimax is the
  # most multimodal criterion, so two seeded runs agreeing is meaningful.
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  run <- function() suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "simplimax",
              estimator = "PAF", seed = 7)))

  set.seed(1)
  state_before <- get(".Random.seed", envir = globalenv(), inherits = FALSE)
  one <- run()
  # The restore-on-exit contract holds on the non-bootstrap path too.
  expect_identical(state_before,
                   get(".Random.seed", envir = globalenv(), inherits = FALSE))

  two <- run()
  expect_equal(one$loadings, two$loadings)
  expect_equal(one$congruence$matched, two$congruence$matched)
})


test_that("bootstrap replicate failures are dropped with a classed warning", {
  set.seed(1)
  p <- 6L; k <- 2L; b <- 10L
  target <- matrix(stats::rnorm(p * k), p, k)
  jitter_cube <- function() {
    a <- array(NA_real_, c(p, k, b))
    for (i in seq_len(b)) a[, , i] <- target + matrix(stats::rnorm(p * k, sd = 0.05), p, k)
    a
  }
  c1 <- jitter_cube()
  c2 <- jitter_cube()
  c2[, , c(2, 5, 8)] <- NA_real_  # three unfittable replicates in group 2
  cubes <- list(g1 = c1, g2 = c2)

  expect_warning(
    res <- .efa_group_boot_congruence(cubes, target, "orthogonal", 0.95),
    class = "efa_group_boot_failed"
  )
  # Only the replicates complete in both groups are used.
  expect_identical(res$n_boot, 7L)
  expect_identical(dim(res$matched_ci$lower), c(2L, 2L, 2L))

  # Every replicate failing is a hard error, not an empty interval.
  all_na <- list(g1 = array(NA_real_, c(p, k, b)), g2 = array(NA_real_, c(p, k, b)))
  expect_error(
    .efa_group_boot_congruence(all_na, target, "orthogonal", 0.95),
    class = "efa_group_boot_all_failed"
  )
})


test_that("a correlation-matrix input skips the bootstrap with a warning", {
  b1 <- WJIV_ages_14_19$cormat[1:12, 1:12]
  b2 <- WJIV_ages_20_39$cormat[1:12, 1:12]

  expect_warning(
    mg <- efa_group(list(young = b1, adult = b2), n_factors = 2,
                    N = c(WJIV_ages_14_19$N, WJIV_ages_20_39$N),
                    rotation = "varimax", b_boot = 50),
    class = "efa_group_boot_unavailable"
  )
  # No interval fields are added on the point-estimate-only path.
  expect_named(mg$congruence, c("matrices", "matched", "degenerate"))
  expect_identical(mg$settings$b_boot, 0L)
})


test_that("a supplied se is dropped with a classed warning", {
  # efa_group manages the SE method itself (bootstrap via b_boot); a stray se must not
  # trigger an unrequested, unseeded per-group bootstrap or leave payload in out$efa.
  # `se` is a declared formal, so it reaches this path whether or not `seed` is given:
  # every formal precedes `...`, so a dots-only `se` is partially matched to `seed` and
  # aborts in the `seed` check instead.
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  expect_warning(
    mg <- suppressMessages(
      efa_group(GRiPS_raw, groups = g, n_factors = 1, se = "np-boot")),
    class = "efa_group_se_ignored"
  )
  expect_null(mg$efa$g1$SE)
  expect_null(mg$efa$g1$replicates)
  expect_named(mg$congruence, c("matrices", "matched", "degenerate"))

  # An explicit `seed` forces exact matching, and a requested bootstrap makes efa_group
  # set `se` for its own replicate fits; both still drop the supplied value.
  expect_warning(
    suppressMessages(
      efa_group(GRiPS_raw, groups = g, n_factors = 1, se = "np-boot", seed = 1)),
    class = "efa_group_se_ignored"
  )
  expect_warning(
    mg_boot <- suppressMessages(
      efa_group(GRiPS_raw, groups = g, n_factors = 1, se = "np-boot", b_boot = 2,
                seed = 1)),
    class = "efa_group_se_ignored"
  )
  expect_null(mg_boot$efa$g1$SE)
  expect_null(mg_boot$efa$g1$replicates)
  expect_identical(mg_boot$settings$b_boot, 2L)
})


test_that("a delta given with a geomin rotation reports which reading was applied", {
  # `delta` names both the salience threshold and the geomin criterion parameter, and the
  # formal takes the name. The criterion parameter can therefore only be set through
  # rotate_control(); the warning gives the reading applied and the route to the other one.
  expect_warning(
    mg <- suppressMessages(
      efa_group(wj_trio, n_factors = 3, N = wj_trio_N, rotation = "geominT",
                delta = 0.001, seed = 1)),
    class = "efa_group_delta_ambiguous"
  )
  expect_identical(mg$settings$delta, 0.001)

  # The reading the warning states: the flag threshold moves, the rotation does not.
  base <- suppressMessages(
    efa_group(wj_trio, n_factors = 3, N = wj_trio_N, rotation = "geominT", seed = 1))
  expect_equal(mg$loadings, base$loadings)
  expect_true(all(mg$diffs$n_flagged > base$diffs$n_flagged))

  # The other reading, which only rotate_control() can express, does move the rotation.
  rc <- suppressMessages(
    efa_group(wj_trio, n_factors = 3, N = wj_trio_N, rotation = "geominT", seed = 1,
              rotate_control = rotate_control(delta = 0.001)))
  expect_identical(rc$settings$delta, 0.1)
  # Movement, not its size: the criterion optimum at this delta is ill-conditioned, so the
  # threshold only has to clear numerical noise (~1e-8), not the 0.024 measured here.
  expect_gt(max(abs(unlist(rc$loadings) - unlist(base$loadings))), 1e-3)

  # Silent where the two readings cannot be confused: a rotation without a `delta`
  # criterion parameter, and a `delta` left at its default.
  expect_no_warning(
    suppressMessages(
      efa_group(wj_trio, n_factors = 3, N = wj_trio_N, rotation = "varimax",
                delta = 0.001, seed = 1)),
    class = "efa_group_delta_ambiguous"
  )
  expect_no_warning(
    suppressMessages(
      efa_group(wj_trio, n_factors = 3, N = wj_trio_N, rotation = "geominT", seed = 1)),
    class = "efa_group_delta_ambiguous"
  )
})


test_that("the raw-data correlation note is given once, not once per group", {
  # Every group is cut from one `x`, so the note the per-group efa_fit() calls raise states
  # one fact about one input. It stays a catchable classed condition.
  g <- rep(c("g1", "g2", "g3"), length.out = nrow(GRiPS_raw))
  n_notes <- 0L
  withCallingHandlers(
    efa_group(GRiPS_raw, groups = g, n_factors = 1),
    efa_cor_from_data = function(cnd) {
      n_notes <<- n_notes + 1L
      invokeRestart("muffleMessage")
    },
    message = function(cnd) invokeRestart("muffleMessage")
  )
  expect_identical(n_notes, 1L)
})


# Cross-group loading differences, per-item salience flags, and the approximate-invariance
# verdict. The table builders are exercised directly on hand-built aligned loadings (so an
# injected non-invariant item is fully controlled), and the classifier against known phi.

test_that(".invariance_band maps congruences to the Lorenzo-Seva & ten Berge bands", {
  expect_identical(
    .invariance_band(c(0.98, 0.95, 0.90, 0.85, 0.8499, 0.50, NA)),
    c("equal", "equal", "fair", "fair", "incongruent", "incongruent", NA)
  )
})


test_that(".efa_group_diffs flags an injected non-invariant loading and summarises the pair", {
  L1 <- matrix(c(0.70, 0.65, 0.10, 0.15,
                 0.12, 0.10, 0.72, 0.68), nrow = 4, ncol = 2,
               dimnames = list(paste0("i", 1:4), c("F1", "F2")))
  L2 <- L1
  L2["i2", "F1"] <- L1["i2", "F1"] + 0.25  # one salient non-invariant cell
  loadings <- list(g1 = L1, g2 = L2)

  dd <- .efa_group_diffs(loadings, delta = 0.10)

  # Flag table: one row per item x factor, exactly the injected cell flagged.
  expect_named(dd$flags, c("group_1", "group_2", "indicator", "factor", "diff",
                           "abs_diff", "flagged", "ci_lower", "ci_upper", "ci_excludes_0"))
  expect_identical(nrow(dd$flags), 8L)
  expect_equal(sum(dd$flags$flagged), 1)
  row <- dd$flags[dd$flags$flagged, ]
  expect_identical(row$indicator, "i2")
  expect_identical(row$factor, "F1")
  expect_equal(row$abs_diff, 0.25)
  expect_equal(row$diff, -0.25)  # g1 - g2
  # No bootstrap supplied -> the interval columns are NA.
  expect_true(all(is.na(dd$flags$ci_lower)))
  expect_true(all(is.na(dd$flags$ci_upper)))
  expect_true(all(is.na(dd$flags$ci_excludes_0)))

  # Diffs summary: one row for the single pair.
  expect_named(dd$diffs, c("group_1", "group_2", "mean_abs_diff", "median_abs_diff",
                           "min_abs_diff", "max_abs_diff", "rmse", "n_flagged"))
  expect_identical(nrow(dd$diffs), 1L)
  expect_identical(dd$diffs$group_1, "g1")
  expect_identical(dd$diffs$group_2, "g2")
  expect_equal(dd$diffs$n_flagged, 1)
  expect_equal(dd$diffs$max_abs_diff, 0.25)
  expect_equal(dd$diffs$mean_abs_diff, 0.25 / 8)

  # delta is a movable salience threshold: a larger one clears the flag, a smaller one keeps it.
  expect_equal(sum(.efa_group_diffs(loadings, delta = 0.30)$flags$flagged), 0)
  expect_equal(sum(.efa_group_diffs(loadings, delta = 0.20)$flags$flagged), 1)
})


test_that(".efa_group_diffs pairs each flag with whether the bootstrap CI excludes zero", {
  L1 <- matrix(c(0.70, 0.65, 0.10, 0.15,
                 0.12, 0.10, 0.72, 0.68), nrow = 4, ncol = 2,
               dimnames = list(paste0("i", 1:4), c("F1", "F2")))
  loadings <- list(g1 = L1, g2 = L1)

  # A synthetic difference-CI cube (groups x groups x items x factors) for the g1-g2 pair:
  # most cells bracket zero; (i2, F1) is a strictly positive interval and (i3, F2) a strictly
  # negative one, so both exclude zero.
  gn <- c("g1", "g2"); it <- paste0("i", 1:4); fn <- c("F1", "F2")
  lower <- array(NA_real_, dim = c(2, 2, 4, 2), dimnames = list(gn, gn, it, fn))
  upper <- lower
  lo_slice <- matrix(-0.05, 4, 2); up_slice <- matrix(0.05, 4, 2)
  lo_slice[2, 1] <- 0.10; up_slice[2, 1] <- 0.30    # (i2, F1): excludes 0 (positive)
  lo_slice[3, 2] <- -0.30; up_slice[3, 2] <- -0.10  # (i3, F2): excludes 0 (negative)
  lower[1, 2, , ] <- lo_slice
  upper[1, 2, , ] <- up_slice
  diff_ci <- list(lower = lower, upper = upper)

  fl <- .efa_group_diffs(loadings, delta = 0.10, diff_ci = diff_ci)$flags

  expect_equal(sum(fl$ci_excludes_0), 2)
  i2F1 <- fl$indicator == "i2" & fl$factor == "F1"
  i3F2 <- fl$indicator == "i3" & fl$factor == "F2"
  expect_true(fl$ci_excludes_0[i2F1])
  expect_true(fl$ci_excludes_0[i3F2])
  expect_equal(fl$ci_lower[i2F1], 0.10)
  expect_equal(fl$ci_upper[i2F1], 0.30)
  # A cell whose interval brackets zero is not marked.
  expect_false(fl$ci_excludes_0[fl$indicator == "i1" & fl$factor == "F1"])
})


test_that(".efa_group_invariance reads the verdict conservatively off the CI lower bound", {
  gn <- c("g1", "g2"); fn <- c("F1", "F2")
  matched <- array(NA_real_, dim = c(2, 2, 2), dimnames = list(gn, gn, fn))
  matched[1, 2, ] <- c(0.99, 0.97); matched[2, 1, ] <- c(0.99, 0.97)
  matched[1, 1, ] <- 1; matched[2, 2, ] <- 1
  lower <- array(NA_real_, dim = c(2, 2, 2), dimnames = list(gn, gn, fn))
  # F1's lower bound stays in the "equal" band; F2's drops into "fair" though its point is "equal".
  lower[1, 2, ] <- c(0.96, 0.88); lower[2, 1, ] <- c(0.96, 0.88)
  lower[1, 1, ] <- 1; lower[2, 2, ] <- 1

  inv <- .efa_group_invariance(list(matched = matched,
                                    matched_ci = list(lower = lower)))

  expect_named(inv, c("group_1", "group_2", "factor", "phi", "phi_lower", "verdict"))
  expect_identical(nrow(inv), 2L)  # one pair x two factors
  expect_identical(inv$verdict[inv$factor == "F1"], "equal")
  expect_identical(inv$verdict[inv$factor == "F2"], "fair")  # conservative: point 0.97 would be "equal"
  expect_equal(inv$phi[inv$factor == "F2"], 0.97)
  expect_equal(inv$phi_lower[inv$factor == "F2"], 0.88)

  # Without a bootstrap the verdict is point-based and phi_lower is left NA.
  inv_pt <- .efa_group_invariance(list(matched = matched))
  expect_identical(inv_pt$verdict, c("equal", "equal"))
  expect_true(all(is.na(inv_pt$phi_lower)))

  # A degenerate (NA) congruence yields an NA verdict rather than "incongruent".
  matched_deg <- matched
  matched_deg[1, 2, 2] <- NA; matched_deg[2, 1, 2] <- NA
  inv_deg <- .efa_group_invariance(list(matched = matched_deg))
  expect_true(is.na(inv_deg$verdict[inv_deg$factor == "F2"]))
})


test_that("efa_group always returns diffs/flags and gates the invariance verdict", {
  mg <- ident_fit_inv

  expect_s3_class(mg$diffs, "data.frame")
  expect_s3_class(mg$flags, "data.frame")
  expect_s3_class(mg$invariance, "data.frame")
  # choose(3, 2) = 3 pairs; the invariance table adds a row per factor.
  expect_identical(nrow(mg$diffs), 3L)
  expect_identical(nrow(mg$invariance), 9L)
  expect_identical(nrow(mg$flags), 3L * p * 3L)

  # Identical groups: nothing salient, congruence 1, every factor "equal".
  expect_equal(sum(mg$flags$flagged), 0)
  expect_true(all(mg$diffs$n_flagged == 0))
  expect_true(all(mg$invariance$verdict == "equal"))
  expect_equal(mg$invariance$phi, rep(1, 9), tolerance = 1e-6)
  # No bootstrap -> the interval-based columns are NA.
  expect_true(all(is.na(mg$invariance$phi_lower)))
  expect_true(all(is.na(mg$flags$ci_excludes_0)))

  expect_equal(mg$settings$delta, 0.1)
  expect_true(mg$settings$invariance)

  # The verdict is gated: off by default, and diffs/flags are still present.
  mg0 <- ident_fit
  expect_null(mg0$invariance)
  expect_false(mg0$settings$invariance)
  expect_s3_class(mg0$diffs, "data.frame")
  expect_s3_class(mg0$flags, "data.frame")
})


test_that("a bootstrap fills the flag CIs and the point difference sits inside them", {
  skip_on_cran()
  # An oblique rotation routes to the reference-Procrustes path. The displayed diff (from the
  # aligned point loadings) and the bootstrap difference CI (from the re-aligned replicate
  # cubes) must share one coordinate frame, so every point diff must lie inside its interval.
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  mg <- suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "promax",
              estimator = "PAF", b_boot = 60, seed = 2024)))

  expect_identical(mg$settings$alignment, "reference")

  fl <- mg$flags
  # The bootstrap populates the interval columns.
  expect_false(all(is.na(fl$ci_lower)))
  expect_false(all(is.na(fl$ci_excludes_0)))
  # Every point-estimate difference lies within its own percentile interval (a small
  # tolerance guards the Monte Carlo edge; a genuine frame offset would push it far outside).
  expect_true(all(fl$ci_lower - 0.02 <= fl$diff & fl$diff <= fl$ci_upper + 0.02,
                  na.rm = TRUE))
})


# ---- print / format / plot -------------------------------------------------

# scrub_num, plus a mask for the salience-flag count `n_flagged / (p*k)`: n_flagged is a hard
# threshold count over BLAS-sensitive aligned loadings, so it can drift by one across
# platforms, and scrub_num leaves it verbatim (it carries no decimal point). Pin the layout
# and wording, not the fragile count.
#
# The mask has to swallow the whitespace ahead of the count as well, and replace it with a
# single space: the count is printed right-aligned in its column, so a drift that changes its
# NUMBER OF DIGITS (6/12 to 12/12) moves the padding rather than the digits and would otherwise
# still change the scrubbed line. scrub_num cannot absorb that padding on its own -- it only
# eats leading whitespace for tokens that carry a decimal point.
scrub_group <- function(lines) sub("\\s*[0-9]+/[0-9]+", " <flagged>", scrub_num(lines))

# Two full WJIV age bands are the standard report fixture: two genuinely different groups
# over the whole item set, so every section of the printout has something to say. varimax
# draws no random start and the consensus gauge is closed form, so the fit is deterministic
# and the blocks below share one -- with the invariance verdict on and off, which changes
# which tables are printed but nothing that is fitted.
wj_pair <- list(age_6_8 = WJIV_ages_6_8$cormat, age_14_19 = WJIV_ages_14_19$cormat)
wj_pair_N <- c(WJIV_ages_6_8$N, WJIV_ages_14_19$N)
wj_pair_fit <- efa_group(wj_pair, n_factors = 3, N = wj_pair_N, rotation = "varimax")
wj_pair_fit_inv <- efa_group(wj_pair, n_factors = 3, N = wj_pair_N,
                             rotation = "varimax", invariance = TRUE)

test_that("print and format render the efa_group report", {
  local_reproducible_output()

  # consensus alignment, orthogonal rotation, invariance verdicts, no bootstrap
  mg <- wj_pair_fit_inv
  expect_snapshot(print(mg), transform = scrub_group)

  # print() is exactly cat(format(x), sep = "\n"), so the two agree line for line
  expect_identical(utils::capture.output(print(mg)), format(mg))

  # three groups -> one row per pair; identical groups -> perfect congruence
  cmat <- test_models$baseline$cormat
  mg_ident <- efa_group(list(a = cmat, b = cmat, c = cmat), n_factors = 2, N = 500,
                        rotation = "varimax", invariance = TRUE)
  expect_snapshot(print(mg_ident), transform = scrub_group)

  # oblique rotation -> reference-alignment header, no invariance section
  mg_ob <- suppressMessages(suppressWarnings(
    efa_group(wj_pair, n_factors = 3, N = wj_pair_N, rotation = "promax")))
  expect_snapshot(print(mg_ob), transform = scrub_group)
})


test_that("print reports the bootstrap congruence intervals and verdicts", {
  skip_on_cran()
  local_reproducible_output()

  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  mg <- suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 1, b_boot = 50, seed = 42,
              invariance = TRUE)))
  # the congruence section notes the CIs, the differences section notes the
  # CI-excluding-0 count, and the invariance heading reads off the CI lower bound
  expect_snapshot(print(mg), transform = scrub_group)
})


test_that("print points out an 'equal' verdict that rests on scale-invariance", {
  local_reproducible_output()

  # Tucker's congruence is invariant to a proportional rescaling of a factor's loadings,
  # so a group whose loadings are uniformly 0.55x another's is graded "equal" on every
  # factor while every cell differs. Build that case exactly: two correlation matrices
  # implied by proportional loading patterns.
  L <- matrix(c(.8, .8, .8, 0, 0, 0,
                0, 0, 0, .8, .8, .8), nrow = 6, ncol = 2,
              dimnames = list(paste0("V", 1:6), c("F1", "F2")))
  implied <- function(L) {
    R <- tcrossprod(L)
    diag(R) <- 1
    dimnames(R) <- list(rownames(L), rownames(L))
    R
  }
  scaled <- list(g1 = implied(L), g2 = implied(0.55 * L))

  mg <- efa_group(scaled, n_factors = 2, N = c(500, 500), rotation = "varimax",
                  invariance = TRUE, seed = 42)
  # the premise: perfect congruence, but substantial cell-wise differences
  expect_equal(unname(mg$congruence$matched["g1", "g2", ]), c(1, 1), tolerance = 1e-6)
  expect_true(all(mg$invariance$verdict == "equal"))
  expect_gt(mg$diffs$mean_abs_diff, 0.1)

  expect_snapshot(print(mg), transform = scrub_group)

  # The pointer is keyed off the salience threshold, so it stays silent when `delta = 0`
  # (which flags every cell by construction) and when the groups genuinely agree.
  shows_pointer <- function(x) {
    any(grepl("proportional rescaling", format(x), fixed = TRUE))
  }
  expect_true(shows_pointer(mg))

  mg_d0 <- efa_group(scaled, n_factors = 2, N = c(500, 500), rotation = "varimax",
                     invariance = TRUE, delta = 0, seed = 42)
  expect_false(shows_pointer(mg_d0))

  mg_bands <- wj_pair_fit_inv
  expect_true(all(mg_bands$invariance$verdict == "equal"))
  expect_lt(max(mg_bands$diffs$mean_abs_diff), 0.1)
  expect_false(shows_pointer(mg_bands))
})


test_that("the report wraps its header and splits its tables to the console width", {
  # A narrow console with four groups: the header no longer fits on one line and the
  # six-column difference table no longer fits in one block.
  local_reproducible_output(width = 60)
  withr::local_options(cli.width = 60)

  bands <- list(age_6_8 = WJIV_ages_6_8$cormat, age_9_13 = WJIV_ages_9_13$cormat,
                age_14_19 = WJIV_ages_14_19$cormat, age_20_39 = WJIV_ages_20_39$cormat)
  Ns <- c(WJIV_ages_6_8$N, WJIV_ages_9_13$N, WJIV_ages_14_19$N, WJIV_ages_20_39$N)
  mg <- efa_group(bands, n_factors = 3, N = Ns, rotation = "varimax",
                  invariance = TRUE, seed = 42)

  lines <- format(mg)
  # nothing overruns the console (the rules are drawn to exactly that width)
  expect_lte(max(cli::ansi_nchar(lines, type = "width")), 60L)
  expect_snapshot(print(mg), transform = scrub_group)

  # the same holds for the report that carries the scale-invariance pointer, which is
  # long enough to overrun any console unless it is wrapped
  L <- matrix(c(.8, .8, .8, 0, 0, 0,
                0, 0, 0, .8, .8, .8), nrow = 6, ncol = 2,
              dimnames = list(paste0("V", 1:6), c("F1", "F2")))
  R <- tcrossprod(0.55 * L)
  diag(R) <- 1
  R2 <- tcrossprod(L)
  diag(R2) <- 1
  dimnames(R) <- dimnames(R2) <- list(rownames(L), rownames(L))
  mg_scaled <- efa_group(list(g1 = R2, g2 = R), n_factors = 2, N = c(500, 500),
                         rotation = "varimax", invariance = TRUE, seed = 42)
  expect_lte(max(cli::ansi_nchar(format(mg_scaled), type = "width")), 60L)
})


test_that(".compare_loadings can skip the decimal-agreement scan", {
  x <- matrix(c(0.57, 0.20, 0.31, 0.44), 2, 2)
  y <- matrix(c(0.57, 0.21, 0.31, 0.44), 2, 2)

  full <- .compare_loadings(x, y)
  quick <- .compare_loadings(x, y, decimals = FALSE)

  # the difference summaries are untouched ...
  expect_equal(full[c("diff", "mean_abs_diff", "median_abs_diff",
                      "min_abs_diff", "max_abs_diff", "g")],
               quick[c("diff", "mean_abs_diff", "median_abs_diff",
                       "min_abs_diff", "max_abs_diff", "g")])
  # ... and only the two decimal-place statistics are dropped
  expect_equal(full$max_dec, 2)
  expect_equal(full$are_equal, 1)
  expect_true(is.na(quick$max_dec))
  expect_true(is.na(quick$are_equal))
})


test_that("plot methods return ggplot objects", {
  mg <- wj_pair_fit

  expect_s3_class(plot(mg, type = "congruence"), "ggplot")
  expect_s3_class(plot(mg, type = "differences"), "ggplot")
  expect_s3_class(plot(mg), "ggplot")            # default is "congruence"
  expect_error(plot(mg, type = "nope"), class = "efa_bad_choice")
})


test_that("the bootstrap congruence plot builds", {
  skip_on_cran()
  g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
  mg <- suppressMessages(suppressWarnings(
    efa_group(GRiPS_raw, groups = g, n_factors = 2, rotation = "promax",
              estimator = "PAF", b_boot = 30, seed = 7)))
  # with a bootstrap the congruence plot carries a point-range (CI) layer; smoke only
  expect_s3_class(plot(mg, type = "congruence"), "ggplot")
  expect_s3_class(plot(mg, type = "differences"), "ggplot")
})


test_that("the deterministic plots match their vdiffr baselines", {
  skip_if_not_installed("vdiffr")
  mg <- wj_pair_fit

  vdiffr::expect_doppelganger("efa_group congruence plot",
                              plot(mg, type = "congruence"))
  vdiffr::expect_doppelganger("efa_group differences plot",
                              plot(mg, type = "differences"))
})

#' Print function for SCREE objects
#'
#' @param x a list of class SCREE Output from \link{SCREE} function.
#' @param plot logical. Whether to plot the results.
#' @param ... Further arguments for print.
#'
#' @export
#' @method print SCREE
#'
#' @examples
#' SCREE_base <- SCREE(test_models$baseline$cormat)
#' SCREE_base
#'
print.SCREE <- function(x, plot = TRUE, ...) {

  eigen_type <-x$settings$eigen_type

  cat("\n")
  cat("Eigenvalues were found using ", .settings_string(eigen_type), ".",
      sep = "")
  cat("\n")
  cat("\n")

  cat("\n")

  if (isTRUE(plot)) {
    graphics::plot(x)
  }

}

# SPSS varimax implementation as described in the SPSS manual: ftp://public.dhe.ibm.com/software/analytics/spss/documentation/statistics/23.0/en/client/Manuals/IBM_SPSS_Statistics_Algorithms.pdf

.VARIMAX_SPSS <- function(x, normalize = TRUE, precision = 1e-5) {

  # get dimensions
  n_cols <- ncol(x)
  n_rows <- nrow(x)

  if (isTRUE(normalize)) {
    # Kaiser row normalisation. Floor the weights so a zero-communality row
    # (h2 = 0) does not divide by zero and inject NaN into the rotation; such a
    # row stays zero throughout and is restored unchanged below. The length-nrow
    # w recycles column-major, scaling row i by 1/w[i] (mirrors efa_procrustes.R).
    h2 <- diag(x %*% t(x))
    w <- sqrt(h2)
    w[!is.finite(w) | w < 1e-15] <- 1
    x <- x / w
  }


  # initialize rotation matrix as identity matrix
  rotmat <- diag(1, n_cols)

  SV_now <- .SV(x)


  for (it in 1:1000) {

    for (col_j in 1:(n_cols - 1)) {
      for (col_k in (col_j + 1):n_cols) {

        u_p <- x[,col_j] ** 2 - x[,col_k] ** 2
        v_p <- 2 * x[,col_j] * x[,col_k]

        A <- sum(u_p)
        B <- sum(v_p)
        C <- sum(u_p ** 2 - v_p ** 2)
        D <- sum(2 * u_p * v_p)

        X <- D - 2 * A * B / n_rows
        Y <- C - (A ** 2 - B ** 2) / n_rows

        P <- 1/4 * atan2(X, Y)

        if (abs(sin(P)) > 1e-15) {
          sub_rot <- matrix(c(cos(P), -sin(P),
                              sin(P), cos(P)),
                            ncol = 2, byrow = TRUE)

          x[, c(col_j, col_k)] <- x[, c(col_j, col_k)] %*% sub_rot
          rotmat[, c(col_j, col_k)] <- rotmat[, c(col_j, col_k)] %*% sub_rot
        }

      }
    }

    SV_old <- SV_now
    SV_now <- .SV(x)

    if (abs(SV_now - SV_old) <= precision) {
      break
    }

  }

  .warn_rotation_no_convergence(abs(SV_now - SV_old) <= precision, 1000L,
                                tunable_maxit = FALSE)

  if (isTRUE(normalize)) {
    x <- x * w
  }

  # Create output list
  list(rotmat = rotmat, loadings = x, iter = it)

}

# SVD-based varimax (stats::varimax) with the Kaiser-weight floor the other rotation
# paths use. stats::varimax() normalises by dividing each row by sqrt(h2) without a
# floor, so a zero-communality row (h2 = 0) injects NaN and the rotation dies inside
# La.svd with an uninformative, unclassed error. Normalising here with the floored
# weights instead -- and handing stats::varimax() an already-normalised matrix -- puts
# the guard where .VARIMAX_SPSS() and .rotate_promax() have it. stats::varimax()
# returns loadings = x %*% rotmat, so scaling the rows by 1/w before and by w after
# reproduces its own normalise-rotate-denormalise sequence exactly, leaving loadings
# with no zero-communality row unchanged.
.varimax_svd <- function(x, normalize = TRUE, precision = 1e-5) {

  if (!isTRUE(normalize)) {
    return(stats::varimax(x, normalize = FALSE, eps = precision))
  }

  w <- sqrt(rowSums(x^2))
  w[!is.finite(w) | w < 1e-15] <- 1

  AV <- stats::varimax(x / w, normalize = FALSE, eps = precision)
  AV$loadings <- AV$loadings * w
  AV
}

unrot <- EFA(test_models$baseline$cormat, 3, N = 500)
vari <- .rotate_model(unrot, rotation = "varimax", type = "EFAtools")
vari_psych <- .rotate_model(unrot, rotation = "varimax", type = "psych")
vari_spss <- .rotate_model(unrot, rotation = "varimax", type = "SPSS")

unrot_1 <- EFA(test_models$baseline$cormat, 1, N = 500)
vari_1 <- suppressWarnings(.rotate_model(unrot_1, rotation = "varimax", type = "EFAtools"))

test_that("output class and dimensions are correct", {
  expect_type(vari, "list")
  expect_type(vari_1, "list")
  expect_named(vari, c("rot_loadings", "rotmat", "vars_accounted_rot", "settings"))
  expect_named(vari_1, c("rot_loadings", "rotmat", "vars_accounted_rot", "settings"))

  expect_s3_class(vari$rot_loadings, "LOADINGS")
  checkmate::expect_matrix(vari$rotmat)
  checkmate::expect_matrix(vari$vars_accounted_rot)
  expect_type(vari$settings, "list")

  expect_s3_class(vari_1$rot_loadings, "LOADINGS")
  expect_equal(vari_1$rotmat, NA)
  expect_null(vari_1$vars_accounted_rot)
  expect_type(vari_1$settings, "list")
})

test_that("settings are returned correctly", {
  expect_named(vari$settings, c("normalize", "precision", "order_type",
                                "varimax_type"))
  expect_named(vari_psych$settings, c("normalize", "precision", "order_type",
                                      "varimax_type"))
  expect_named(vari_spss$settings, c("normalize", "precision", "order_type",
                                     "varimax_type"))
  expect_named(vari_1$settings, c("normalize", "precision", "order_type",
                                  "varimax_type"))

  expect_equal(vari$settings$normalize, TRUE)
  expect_equal(vari_psych$settings$normalize, TRUE)
  expect_equal(vari_spss$settings$normalize, TRUE)
  expect_equal(vari_1$settings$normalize, TRUE)

  expect_equal(vari$settings$precision, 1e-05)
  expect_equal(vari_psych$settings$precision, 1e-05)
  expect_equal(vari_spss$settings$precision, 1e-5)
  expect_equal(vari_1$settings$precision, 1e-05)

  expect_equal(vari$settings$order_type, "eigen")
  expect_equal(vari_psych$settings$order_type, "eigen")
  expect_equal(vari_spss$settings$order_type, "ss_factors")
  expect_equal(vari_1$settings$order_type, "eigen")

  expect_equal(vari$settings$varimax_type, "kaiser")
  expect_equal(vari_psych$settings$varimax_type, "svd")
  expect_equal(vari_spss$settings$varimax_type, "kaiser")
  expect_equal(vari_1$settings$varimax_type, "kaiser")

})

test_that("errors etc. are thrown correctly", {

  expect_error(.rotate_model(unrot, rotation = "varimax", type = "none"), class = "efa_type_none")

  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "EFAtools", normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "EFAtools", order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "psych", normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "psych", order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "SPSS", normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "SPSS", order_type = "eigen"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "varimax", type = "SPSS", varimax_type = "svd"), class = "efa_type_override")
  expect_warning(.rotate_model(unrot_1, rotation = "varimax", type = "EFAtools"), class = "efa_single_factor")
})

test_that(".VARIMAX_SPSS tolerates a zero-communality row (Kaiser normalisation)", {
  # A zero-communality row makes 1/sqrt(h2) = Inf; without floored weights Kaiser
  # normalisation injects NaN that corrupts the rotation or triggers an
  # uninformative if(NA) error. The floored row must stay zero throughout.
  L <- matrix(c(0.8, 0.1,
                0.7, 0.2,
                0.0, 0.0,
                0.1, 0.9,
                0.2, 0.8), ncol = 2, byrow = TRUE)
  res <- .VARIMAX_SPSS(L, normalize = TRUE)
  expect_false(anyNA(res$loadings))
  expect_equal(res$loadings[3, ], c(0, 0))
})

test_that("the svd varimax path tolerates a zero-communality row (Kaiser normalisation)", {
  # stats::varimax() Kaiser-normalises by dividing each row by sqrt(h2) without a floor,
  # so the same zero-communality row that .VARIMAX_SPSS() handles above makes the
  # rotation die inside La.svd with an uninformative, unclassed error. .varimax_svd()
  # normalises with the floored weights instead, so both varimax_type values accept the
  # same data rather than the choice of `type` deciding whether it errors or succeeds.
  L <- matrix(c(0.8, 0.1,
                0.7, 0.2,
                0.0, 0.0,
                0.1, 0.9,
                0.2, 0.8), ncol = 2, byrow = TRUE)

  res <- .varimax_svd(L, normalize = TRUE)
  expect_false(anyNA(res$loadings))
  expect_equal(unclass(res$loadings)[3, ], c(0, 0))
  # the reproduction identity loadings == L %*% rotmat survives the pre/post scaling
  expect_equal(unclass(res$loadings), L %*% res$rotmat, ignore_attr = TRUE)

  # the floor is arithmetically neutral: with no zero-communality row the floored path
  # reproduces stats::varimax() exactly
  L2 <- L
  L2[3, ] <- c(0.6, 0.3)
  expect_equal(unclass(.varimax_svd(L2, normalize = TRUE)$loadings),
               unclass(stats::varimax(L2, normalize = TRUE, eps = 1e-5)$loadings))

  # normalize = FALSE skips Kaiser normalisation entirely, so it must reach
  # stats::varimax() unnormalized rather than picking up the floored weights
  expect_equal(unclass(.varimax_svd(L2, normalize = FALSE)$loadings),
               unclass(stats::varimax(L2, normalize = FALSE, eps = 1e-5)$loadings))
  expect_false(isTRUE(all.equal(unclass(.varimax_svd(L2, normalize = FALSE)$loadings),
                                unclass(.varimax_svd(L2, normalize = TRUE)$loadings))))

  # the guard reaches both rotations that build on the svd varimax
  for (rn in c("varimax", "promax")) {
    out <- .rotate_model(list(unrot_loadings = L), rotation = rn, type = "psych")
    expect_false(anyNA(out$rot_loadings))
    expect_equal(unclass(out$rot_loadings)[3, ], c(0, 0), ignore_attr = TRUE)
  }
})

test_that("the SPSS varimax non-convergence warning names the tolerance alone", {
  # The Kaiser sweep runs a budget of 1000 iterations that is fixed in the algorithm, so
  # `maxit` is no remedy here and only `precision` is named. The branch is a backstop -- the
  # sweep drives the change in the simplicity criterion to exactly zero on every input tried,
  # so no correlation matrix reaches it -- and the warning is raised directly to make sure the
  # text it would show is the text recorded here.
  w <- tryCatch(.warn_rotation_no_convergence(FALSE, 1000L, tunable_maxit = FALSE),
                efa_rotation_no_convergence = function(w) w)
  expect_s3_class(w, "efa_rotation_no_convergence")
  expect_snapshot(cat(conditionMessage(w)))
})

rm(unrot, vari, unrot_1, vari_1, vari_psych, vari_spss)

#' Print LOADINGS object
#'
#' @details
#' The method prints a loading matrix in a compact, console-oriented table.
#' Loadings with absolute value greater than or equal to `cutoff` are emphasized,
#' smaller loadings are de-emphasized, and Heywood-relevant communality/
#' uniqueness values are marked when `h2` is supplied. Long variable names can
#' be truncated, abbreviated, or printed in full. If the matrix has many factor
#' columns, the table is split into column blocks so that the output remains
#' readable in narrower consoles.
#'
#' If `h2` is named and `x` has row names, `h2` is matched to the row names of
#' `x` before any optional row sorting is applied. If `x` has no row names, a
#' named `h2` vector is used in the supplied order.
#'
#' @param x class LOADINGS matrix.
#' @param cutoff numeric. The number above which to print loadings in bold
#'  default is .3.
#' @param digits numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
#'  to round the loadings to (default is 3).
#' @param max_name_length numeric. The maximum length of the variable names to
#'  display. Everything beyond this will be cut from the right unless
#'  \code{name_style = "abbreviate"} or \code{name_style = "full"} is used.
#' @param h2 numeric. Vector of communalities to print. If named and \code{x}
#'  has row names, names are used to align communalities to rows.
#' @param color logical. Whether to apply console styling using \pkg{crayon}.
#'  Default is \code{TRUE}.
#' @param name_style character. How to shorten variable names longer than
#'  \code{max_name_length}. \code{"truncate"} cuts names from the right,
#'  \code{"abbreviate"} uses \code{\link[base:abbreviate]{abbreviate}}, and
#'  \code{"full"} prints full names.
#' @param max_factor_name_length numeric or \code{NULL}. Optional maximum length
#'  of factor names. If \code{NULL}, factor names are not shortened.
#' @param max_factors_per_block numeric or \code{NULL}. Maximum number of factor
#'  columns to print per block. If \code{NULL}, the number is chosen from the
#'  console width.
#' @param sort_loadings character. Optional row sorting. \code{"none"} preserves
#'  the input order, \code{"primary"} groups rows by the factor with the largest
#'  absolute loading, and \code{"clustered"} additionally sorts within each
#'  factor by the size of the primary loading.
#' @param legend logical. Whether to append a short explanation of the styling.
#'  Default is \code{FALSE} for standalone loading matrices.
#' @param ... additional arguments passed to print or format
#'
#' @method print LOADINGS
#' @export
#'
#' @examples
#' EFAtools_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                     type = "EFAtools", method = "PAF", rotation = "promax")
#' EFAtools_PAF
#'
print.LOADINGS <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                           h2 = NULL, color = TRUE,
                           name_style = c("truncate", "abbreviate", "full"),
                           max_factor_name_length = NULL,
                           max_factors_per_block = NULL,
                           sort_loadings = c("none", "primary", "clustered"),
                           legend = FALSE, ...) {

  out <- format.LOADINGS(x,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = h2,
    color = color,
    name_style = name_style,
    max_factor_name_length = max_factor_name_length,
    max_factors_per_block = max_factors_per_block,
    sort_loadings = sort_loadings,
    legend = legend,
    ...
  )

  cat(out, sep = "\n")
  cat("\n")

  invisible(x)
}

#' @rdname print.LOADINGS
#' @method format LOADINGS
#' @export
format.LOADINGS <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                            h2 = NULL, color = TRUE,
                            name_style = c("truncate", "abbreviate", "full"),
                            max_factor_name_length = NULL,
                            max_factors_per_block = NULL,
                            sort_loadings = c("none", "primary", "clustered"),
                            legend = FALSE, ...) {

  name_style <- match.arg(name_style)
  sort_loadings <- match.arg(sort_loadings)

  .validate_loadings_print_args(
    x = x,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = h2,
    color = color,
    max_factor_name_length = max_factor_name_length,
    max_factors_per_block = max_factors_per_block,
    legend = legend
  )

  spec <- .loadings_print_spec(
    x = x,
    h2 = h2,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    name_style = name_style,
    max_factor_name_length = max_factor_name_length,
    sort_loadings = sort_loadings
  )

  .format_loadings_table(
    spec,
    color = color,
    max_factors_per_block = max_factors_per_block,
    legend = legend
  )
}

.validate_loadings_print_args <- function(x, cutoff, digits, max_name_length,
                                          h2, color, max_factor_name_length,
                                          max_factors_per_block, legend) {
  if (!is.matrix(x) || !is.numeric(x)) {
    stop("`x` must be a numeric matrix.", call. = FALSE)
  }

  if (nrow(x) < 1L || ncol(x) < 1L) {
    stop("`x` must have at least one row and one column.", call. = FALSE)
  }

  if (!is.numeric(cutoff) || length(cutoff) != 1L ||
      !is.finite(cutoff) || cutoff < 0) {
    stop("`cutoff` must be a single finite non-negative number.", call. = FALSE)
  }

  if (!is.numeric(digits) || length(digits) != 1L || !is.finite(digits) ||
      digits < 0 || digits != as.integer(digits)) {
    stop("`digits` must be a single finite non-negative integer.", call. = FALSE)
  }

  if (!is.numeric(max_name_length) || length(max_name_length) != 1L ||
      !is.finite(max_name_length) || max_name_length < 1 ||
      max_name_length != as.integer(max_name_length)) {
    stop("`max_name_length` must be a single finite positive integer.", call. = FALSE)
  }

  if (!is.null(max_factor_name_length) &&
      (!is.numeric(max_factor_name_length) || length(max_factor_name_length) != 1L ||
       !is.finite(max_factor_name_length) || max_factor_name_length < 1 ||
       max_factor_name_length != as.integer(max_factor_name_length))) {
    stop("`max_factor_name_length` must be NULL or a single finite positive integer.", call. = FALSE)
  }

  if (!is.null(max_factors_per_block) &&
      (!is.numeric(max_factors_per_block) || length(max_factors_per_block) != 1L ||
       !is.finite(max_factors_per_block) || max_factors_per_block < 1 ||
       max_factors_per_block != as.integer(max_factors_per_block))) {
    stop("`max_factors_per_block` must be NULL or a single finite positive integer.", call. = FALSE)
  }

  if (!is.logical(color) || length(color) != 1L || is.na(color)) {
    stop("`color` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(legend) || length(legend) != 1L || is.na(legend)) {
    stop("`legend` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.null(h2)) {
    if (!is.numeric(h2) || length(h2) != nrow(x)) {
      stop("`h2` must be a numeric vector with one value per row of `x`.", call. = FALSE)
    }
  }

  invisible(TRUE)
}

.loadings_print_spec <- function(x, h2, cutoff, digits, max_name_length,
                                 name_style, max_factor_name_length,
                                 sort_loadings) {
  x <- as.matrix(x)
  n_factors <- ncol(x)

  factor_names <- colnames(x)
  if (is.null(factor_names)) {
    factor_names <- paste0("F", seq_len(n_factors))
  }

  original_var_names <- rownames(x)
  has_row_names <- !is.null(original_var_names)
  var_names <- original_var_names
  if (!has_row_names) {
    var_names <- paste0("V", seq_len(nrow(x)))
  }

  has_h2 <- !is.null(h2)
  if (has_h2) {
    h2 <- .align_loadings_h2(
      h2 = h2,
      var_names = var_names,
      has_row_names = has_row_names
    )
  }

  row_order <- .loadings_row_order(x, sort_loadings)
  x <- x[row_order, , drop = FALSE]
  var_names <- var_names[row_order]
  if (has_h2) {
    h2 <- h2[row_order]
  }

  col_type <- rep("loading", n_factors)
  values <- x

  if (has_h2) {
    values <- cbind(values, "h2" = h2, "u2" = 1 - h2)
    factor_names <- c(factor_names, "h2", "u2")
    col_type <- c(col_type, "h2", "u2")
  }

  display_var_names <- .shorten_loadings_names(
    var_names,
    max_length = max_name_length,
    name_style = name_style
  )

  display_factor_names <- .shorten_loadings_factor_names(
    factor_names,
    max_length = max_factor_name_length
  )

  value_strings <- .format_loadings_values(values, digits = digits)

  list(
    values = values,
    value_strings = value_strings,
    var_names = display_var_names,
    factor_names = display_factor_names,
    col_type = col_type,
    cutoff = cutoff,
    digits = digits,
    n_factors = n_factors,
    has_h2 = has_h2
  )
}

.loadings_row_order <- function(x, sort_loadings = c("none", "primary", "clustered")) {
  sort_loadings <- match.arg(sort_loadings)

  if (identical(sort_loadings, "none") || nrow(x) < 2L || ncol(x) < 1L) {
    return(seq_len(nrow(x)))
  }

  abs_x <- abs(x)
  abs_x[is.na(abs_x)] <- -Inf
  primary_factor <- max.col(abs_x, ties.method = "first")
  primary_loading <- apply(abs_x, 1L, max, na.rm = TRUE)
  primary_loading[!is.finite(primary_loading)] <- -Inf

  if (identical(sort_loadings, "primary")) {
    return(order(primary_factor, seq_along(primary_factor)))
  }

  order(primary_factor, -primary_loading, seq_along(primary_factor))
}

.align_loadings_h2 <- function(h2, var_names, has_row_names = TRUE) {
  # Only align by name when the loading matrix itself has row names. Generated
  # fallback names (V1, V2, ...) should not be used to reorder a named h2 vector.
  h2_names <- names(h2)
  has_complete_h2_names <- !is.null(h2_names) && all(nzchar(h2_names))

  if (isTRUE(has_row_names) && has_complete_h2_names) {
    if (!all(var_names %in% h2_names)) {
      stop(
        "If `h2` is named and `x` has row names, names(h2) must include all row names of `x`.",
        call. = FALSE
      )
    }
    h2 <- h2[var_names]
  }

  as.numeric(h2)
}

.shorten_loadings_names <- function(x, max_length, name_style) {
  x <- as.character(x)

  if (identical(name_style, "full")) {
    return(x)
  }

  if (all(nchar(x) <= max_length)) {
    return(x)
  }

  if (identical(name_style, "abbreviate")) {
    return(as.character(abbreviate(
      x,
      minlength = max_length,
      strict = TRUE,
      method = "both.sides"
    )))
  }

  substr(x, 1L, max_length)
}

.shorten_loadings_factor_names <- function(x, max_length = NULL) {
  x <- as.character(x)

  if (is.null(max_length) || all(nchar(x) <= max_length)) {
    return(x)
  }

  substr(x, 1L, max_length)
}

.format_loadings_values <- function(x, digits) {
  values <- as.numeric(x)
  out <- .numformat(round(values, digits = digits), digits = digits)
  out[is.na(values)] <- "NA"

  matrix(out, nrow = nrow(x), ncol = ncol(x), dimnames = dimnames(x))
}

.format_loadings_number <- function(x, digits) {
  if (is.na(x)) {
    return("NA")
  }

  .numformat(round(x, digits = digits), digits = digits)
}

.format_loadings_table <- function(spec, color = TRUE,
                                   max_factors_per_block = NULL,
                                   legend = FALSE) {
  row_width <- max(nchar(spec$var_names), 1L)
  col_widths <- pmax(
    nchar(spec$factor_names),
    apply(spec$value_strings, 2L, function(z) max(nchar(z)))
  )

  blocks <- .loadings_column_blocks(
    spec = spec,
    row_width = row_width,
    col_widths = col_widths,
    max_factors_per_block = max_factors_per_block
  )

  out <- character(0)
  multi_block <- length(blocks) > 1L

  for (bb in seq_along(blocks)) {
    cols <- blocks[[bb]]

    if (multi_block) {
      if (bb > 1L) {
        out <- c(out, "")
      }
      out <- c(out, .loadings_block_label(spec, cols, bb, length(blocks), color))
    }

    out <- c(out, .format_loadings_block(
      spec = spec,
      cols = cols,
      row_width = row_width,
      col_widths = col_widths,
      color = color
    ))
  }

  if (isTRUE(legend)) {
    out <- c(out, "", .loadings_legend(spec, color = color))
  }

  out
}

.loadings_column_blocks <- function(spec, row_width, col_widths,
                                    max_factors_per_block = NULL) {
  factor_cols <- seq_len(spec$n_factors)
  aux_cols <- if (isTRUE(spec$has_h2)) {
    (spec$n_factors + 1L):length(spec$col_type)
  } else {
    integer(0)
  }

  if (length(factor_cols) <= 1L) {
    return(list(c(factor_cols, aux_cols)))
  }

  if (!is.null(max_factors_per_block)) {
    split_points <- ceiling(seq_along(factor_cols) / max_factors_per_block)
    return(lapply(split(factor_cols, split_points), function(cols) c(cols, aux_cols)))
  }

  console_width <- .loadings_console_width()
  full_width <- .loadings_table_width(c(factor_cols, aux_cols), row_width, col_widths)
  if (full_width <= console_width) {
    return(list(c(factor_cols, aux_cols)))
  }

  blocks <- list()
  current <- integer(0)

  for (col in factor_cols) {
    candidate <- c(current, col, aux_cols)
    if (length(current) > 0L &&
        .loadings_table_width(candidate, row_width, col_widths) > console_width) {
      blocks[[length(blocks) + 1L]] <- c(current, aux_cols)
      current <- col
    } else {
      current <- c(current, col)
    }
  }

  if (length(current) > 0L) {
    blocks[[length(blocks) + 1L]] <- c(current, aux_cols)
  }

  blocks
}

.loadings_console_width <- function() {
  width <- tryCatch(cli::console_width(), error = function(e) NA_integer_)
  if (!is.finite(width) || width < 40L) {
    return(80L)
  }
  as.integer(width)
}

.loadings_table_width <- function(cols, row_width, col_widths) {
  # Two spaces between the row-label column and every value column.
  row_width + 2L + sum(col_widths[cols]) + 2L * (length(cols) - 1L)
}

.loadings_block_label <- function(spec, cols, block, n_blocks, color) {
  factor_cols <- cols[cols <= spec$n_factors]
  first <- spec$factor_names[min(factor_cols)]
  last <- spec$factor_names[max(factor_cols)]

  label <- if (identical(first, last)) {
    paste0("Factors ", first, " (block ", block, "/", n_blocks, ")")
  } else {
    paste0("Factors ", first, "-", last, " (block ", block, "/", n_blocks, ")")
  }

  .loadings_style(label, style = "label", color = color)
}

.format_loadings_block <- function(spec, cols, row_width, col_widths, color = TRUE) {
  header <- .loadings_style(
    stringr::str_c(
      stringr::str_pad("", row_width),
      "  ",
      stringr::str_c(
        stringr::str_pad(spec$factor_names[cols], col_widths[cols], side = "both"),
        collapse = "  "
      )
    ),
    style = "label",
    color = color
  )

  rows <- vapply(seq_len(nrow(spec$values)), function(ii) {
    cells <- vapply(cols, function(jj) {
      cell <- stringr::str_pad(spec$value_strings[ii, jj], col_widths[jj], side = "left")
      .style_loadings_cell(
        cell = cell,
        value = spec$values[ii, jj],
        col_type = spec$col_type[jj],
        cutoff = spec$cutoff,
        color = color
      )
    }, character(1L))

    stringr::str_c(
      .loadings_style(
        stringr::str_pad(spec$var_names[ii], row_width, side = "right"),
        style = "label",
        color = color
      ),
      "  ",
      stringr::str_c(cells, collapse = "  ")
    )
  }, character(1L))

  c(header, rows)
}

.style_loadings_cell <- function(cell, value, col_type, cutoff, color) {
  if (identical(col_type, "loading")) {
    if (is.na(value)) {
      return(cell)
    }

    if (abs(value) < cutoff) {
      return(.loadings_style(cell, style = "weak", color = color))
    }

    return(.loadings_style(cell, style = "salient", color = color))
  }

  if (identical(col_type, "h2") && is.finite(value) && value > 1) {
    return(.loadings_style(cell, style = "heywood", color = color))
  }

  if (identical(col_type, "u2") && is.finite(value) && value < 0) {
    return(.loadings_style(cell, style = "heywood", color = color))
  }

  cell
}

.loadings_legend <- function(spec, color = TRUE) {
  cutoff <- .format_loadings_number(spec$cutoff, digits = spec$digits)

  if (isTRUE(color)) {
    text <- paste0(
      crayon::italic("Legend: \n"),
      crayon::bold("  bold"), " = |loading| >= ", cutoff,
      "\n", crayon::silver("  grey"), " = below cutoff"
    )
    if (isTRUE(spec$has_h2)) {
      text <- paste0(text, "\n", crayon::red$bold("  red h2/u2"), " = Heywood-relevant value")
    }
    return(text)
  }

  text <- paste0("Legend: salient loadings use cutoff |loading| >= ", cutoff)
  if (isTRUE(spec$has_h2)) {
    text <- paste0(text, "; h2 > 1 or u2 < 0 indicates a Heywood-relevant value")
  }
  text
}

.loadings_style <- function(x, style = c("label", "weak", "salient", "heywood"),
                            color = TRUE) {
  style <- match.arg(style)

  if (!isTRUE(color)) {
    return(x)
  }

  switch(style,
    label = crayon::blue(x),
    weak = crayon::silver(x),
    salient = crayon::bold(x),
    heywood = crayon::red$bold(x)
  )
}

# One emitted report item: the line matching `start` plus the continuation lines the
# console-width packer indented under it. Used by the width tests of every printer that packs
# a long line, so they all agree on what counts as one item (the indent is .efa_wrap_chunks()'s
# `exdent`, and a blank line or any unindented line ends the item).
wrapped_item <- function(lines, start) {
  i <- grep(start, lines)[1L]
  j <- i
  while (j < length(lines) && startsWith(lines[j + 1L], "  ")) j <- j + 1L
  lines[i:j]
}

# Shared snapshot scrub used by the print/format snapshot tests.
#
# Masks every decimal number, together with the sign and any padding in front of
# it (so a sign flip of a near-zero value cannot change the masked line), and
# collapses the value-driven inter-column spacing of factor-label matrix headers
# (lines of the form "F1    F2    F3"). A matrix header carries no decimal for the
# number mask to absorb, so its column alignment would otherwise be pinned verbatim
# and shift whenever a body value changes width. Print snapshots recorded under
# local_reproducible_output() therefore pin down layout, section order, headers
# (labels and their order), and wording, but not computed decimals or the column
# widths they drive, both of which can drift across BLAS implementations, platforms,
# and rotation-package versions. Integer output (factor counts, df, N, residual
# counts) is kept verbatim.
scrub_num <- function(lines) {
  lines <- gsub("\\s*-?(\\d+)?\\.\\d+", " <num>", lines)
  hdr <- grepl("^\\s*F[0-9]+( +F[0-9]+)+ *$", lines)
  lines[hdr] <- gsub("(F[0-9]+) +", "\\1 ", lines[hdr])
  lines
}

# Like scrub_num, but also masks integer percentages and the effective number of solutions
# each fit index is averaged over. The efa_average() print reports
# error/convergence/Heywood/admissibility rates as `round(mean(...) * 100)` over many inner
# solutions, and the Model Fit block counts the solutions that reached a usable fit index;
# both are driven by the same per-solution convergence and Heywood outcomes, which can flip
# across BLAS implementations. The snapshot therefore pins the wording but not those counts.
# (The deterministic grid size - the "N EFAs" sentence and the "of N" denominators - is left
# verbatim.)
scrub_num_pct <- function(lines) {
  lines <- gsub("[0-9]+%", "<pct>", lines)
  lines <- gsub("averaged over [0-9]+ of", "averaged over <n> of", lines)
  scrub_num(lines)
}

#' Plot the RMSEA power curve
#'
#' @description
#' Draws the analytic RMSEA power (MacCallum, Browne, & Sugawara, 1996) of an
#' [efa_power()] result as a function of the total sample size, mirroring
#' `semTools::plotRMSEApower()` but returning a [ggplot2::ggplot] object rather than
#' drawing to the active device. The test, its null and alternative RMSEA, the
#' significance level, and the number of groups are taken from the object; only the
#' sample-size axis is swept, with an optional sweep of the degrees of freedom or the
#' alternative RMSEA to overlay several curves.
#'
#' @details
#' When the plotted curve is the object's own -- a single curve with neither `df` nor
#' `eps1` overridden -- it is annotated with the object's result: a dashed vertical line
#' at its sample size `x$N`, a dashed horizontal line at the reference power (the target
#' power when a sample size was solved for, otherwise the power achieved at `x$N`), and a
#' point at `x$N` and the achieved power. Overriding `df` or `eps1`, sweeping either as a
#' vector, or supplying an `n` that does not span `x$N` moves that point off the drawn
#' curve, so the marks are then omitted.
#'
#' @param x An object of class `efa_power` (output from [efa_power()]).
#' @param n numeric. The total sample sizes to evaluate. If `NULL` (the default) a
#'   sequence bracketing the object's sample size is chosen automatically.
#' @param df numeric. The model degrees of freedom (must be positive). Defaults to the
#'   object's `df`; a vector of length greater than one draws one curve per value.
#' @param eps1 numeric. The alternative-hypothesis RMSEA (must differ from the null
#'   `eps0`). Defaults to the object's `eps1`; a vector of length greater than one draws
#'   one curve per value. At most one of `df` and `eps1` may be a vector.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns A [ggplot2::ggplot] object.
#'
#' @references
#' MacCallum, R. C., Browne, M. W., & Sugawara, H. M. (1996). Power analysis and
#' determination of sample size for covariance structure modeling. *Psychological
#' Methods, 1*(2), 130-149. \doi{10.1037/1082-989X.1.2.130}
#'
#' @family power analysis
#'
#' @importFrom rlang .data
#' @export
#' @method plot efa_power
#'
#' @examples
#' pw <- efa_power(df = 100, N = 200)
#'
#' # Power curve for the test of close fit, marking the object's own N
#' plot(pw)
#'
#' # Overlay several models by sweeping the degrees of freedom
#' plot(pw, df = c(50, 100, 200))
#'
#' # Sweep the alternative RMSEA instead
#' plot(pw, eps1 = c(0.06, 0.08, 0.10))
plot.efa_power <- function(x, n = NULL, df = NULL, eps1 = NULL, ...) {
  # Simulation-mode results are Monte-Carlo rates, not an analytic power curve, so
  # there is no sample-size axis to sweep.
  if (identical(x$settings$mode, "simulation")) {
    cli::cli_abort(
      c("There is no power curve to plot for a simulation-mode {.cls efa_power} object.",
        "i" = "Inspect the hit-rate, recovery, and convergence rates in the printed summary instead."),
      class = "efa_power_no_plot")
  }
  s <- x$settings
  if (is.null(df)) df <- s$df
  if (is.null(eps1)) eps1 <- s$eps1
  checkmate::assert_numeric(df, finite = TRUE, any.missing = FALSE, min.len = 1)
  checkmate::assert_numeric(eps1, lower = 0, finite = TRUE, any.missing = FALSE,
                            min.len = 1)
  # An override must uphold the same invariants efa_power() enforces on the object: a
  # positive df, and an alternative RMSEA distinct from the null (else there is nothing
  # to detect). Reuse the constructor's condition classes so callers can catch either.
  if (any(df <= 0)) {
    cli::cli_abort("{.arg df} must be positive.", class = "efa_power_bad_df")
  }
  if (any(eps1 == s$eps0)) {
    cli::cli_abort(
      c("{.arg eps1} must differ from the null RMSEA {.arg eps0} ({s$eps0}).",
        "i" = "Equal null and alternative RMSEA leave no difference to detect."),
      class = "efa_power_unreachable")
  }

  # "Sweep df OR eps1": at most one dimension may vary, so a single colour aesthetic
  # carries the overlay. A 2-D grid of both would need faceting and is out of scope.
  df_swept <- length(df) > 1L
  eps1_swept <- length(eps1) > 1L
  if (df_swept && eps1_swept) {
    cli::cli_abort(
      c("Only one of {.arg df} and {.arg eps1} can be swept at a time.",
        "i" = "Give a vector to one of them and leave the other a single value."),
      class = "efa_power_plot_grid")
  }
  swept <- df_swept || eps1_swept

  # The swept argument supplies the colour levels; the other is held fixed. curve_df and
  # curve_eps1 pair element-wise with the levels for the per-curve power evaluation.
  if (eps1_swept) {
    levels_val <- eps1
    level_lab <- "RMSEA (H1)"
    curve_df <- rep(df[1L], length(levels_val))
    curve_eps1 <- eps1
  } else {
    levels_val <- df
    level_lab <- "df"
    curve_df <- df
    curve_eps1 <- rep(eps1[1L], length(levels_val))
  }

  # Sample-size axis: honour a supplied `n`, otherwise span from 1 up to where the
  # slowest curve saturates (its N for 0.99 power), so every curve reaches its plateau
  # and the object's N sits inside. `.efa_power_solve_N()` aborts (efa_power_unreached)
  # when a target is unreachable (e.g. wrong-side epsilons); fall back to twice the
  # object's N there, but let any other error surface.
  if (is.null(n)) {
    hi <- vapply(seq_along(levels_val), function(i) {
      tryCatch(
        .efa_power_solve_N(curve_df[i], s$eps0, curve_eps1[i], s$alpha, s$group,
                           s$type, target = 0.99),
        efa_power_unreached = function(e) 2 * x$N)
    }, numeric(1))
    n_hi <- max(hi, x$N)
    n <- unique(round(seq(1, n_hi, length.out = 100L)))
  } else {
    checkmate::assert_numeric(n, lower = 1, finite = TRUE, any.missing = FALSE,
                              min.len = 1)
  }

  # Power on the (N x curve) grid, one long-format row per point.
  rows <- lapply(seq_along(levels_val), function(i) {
    pw <- vapply(n, function(nn) {
      .efa_power_rmsea(nn, curve_df[i], s$eps0, curve_eps1[i], s$alpha, s$group,
                       s$type)$power
    }, numeric(1))
    data.frame(N = n, power = pw, level = levels_val[i])
  })
  dat <- do.call(rbind, rows)
  dat$level <- factor(dat$level, levels = unique(levels_val))

  fit_lbl <- if (s$type == "close") "close fit" else "not-close fit"
  n_lab <- if (s$group > 1) "Total sample size (N)" else "Sample size (N)"
  # The null bounds RMSEA from above for close fit and from below for not-close fit,
  # matching the comparator in format.efa_power().
  cmp <- if (s$type == "close") "\u2264" else "\u2265"
  sub <- paste0("alpha = ", .efa_num(s$alpha, pad = FALSE),
                " \u00b7 df = ", if (df_swept) "swept" else df[1L],
                " \u00b7 H0 RMSEA ", cmp, " ", .efa_num(s$eps0, pad = FALSE))
  if (!eps1_swept) {
    sub <- paste0(sub, " \u00b7 H1 RMSEA = ", .efa_num(eps1[1L], pad = FALSE))
  }

  # Annotate the object's own point only when the drawn curve is the object's -- a single
  # unmodified curve whose sample-size axis actually spans x$N. Any override or an `n`
  # that excludes x$N would place (x$N, x$power) off the curve.
  mark_object <- !swept && df[1L] == s$df && eps1[1L] == s$eps1 &&
    x$N >= min(n) && x$N <= max(n)

  if (swept) {
    p <- ggplot2::ggplot(dat, ggplot2::aes(.data$N, .data$power, colour = .data$level)) +
      ggplot2::geom_line(linewidth = 0.8) +
      ggplot2::scale_colour_viridis_d(end = 0.85) +
      ggplot2::labs(colour = level_lab)
  } else {
    p <- ggplot2::ggplot(dat, ggplot2::aes(.data$N, .data$power))
    if (mark_object) {
      # Reference power: the target when N was solved for, else the power at N.
      ref_power <- if (x$solve_for == "N") s$power else x$power
      p <- p +
        ggplot2::geom_hline(yintercept = ref_power, linetype = 2, colour = "grey50") +
        ggplot2::geom_vline(xintercept = x$N, linetype = 2, colour = "grey50")
    }
    p <- p + ggplot2::geom_line(linewidth = 0.8, colour = "#3B528B")
    if (mark_object) {
      p <- p + ggplot2::geom_point(data = data.frame(N = x$N, power = x$power),
                                   size = 2.5, colour = "#3B528B")
    }
  }

  p +
    ggplot2::coord_cartesian(ylim = c(0, 1)) +
    ggplot2::labs(x = n_lab, y = "Power",
                  title = paste0("Power for the RMSEA test of ", fit_lbl),
                  subtitle = sub) +
    .gg_theme()
}

# Factor congruence metrics and solution alignment: Tucker congruence between two loading
# matrices, and reordering/sign-matching a solution to a target.

# Pairwise Tucker congruence between the columns of two loading matrices: entry
# (i, j) is the cosine <L1_i, L2_j> / (||L1_i|| ||L2_j||). No validation and
# NA-tolerant, so the wrappers below add their own checks (`.tucker_congruence`)
# or missing-value handling (`.factor_congruence`) around this shared core. A
# caller that already has the column norms can pass them in to avoid recomputing.
.congruence_matrix <- function(L1, L2,
                               n1 = sqrt(colSums(L1 * L1)),
                               n2 = sqrt(colSums(L2 * L2))) {
  crossprod(L1, L2) / tcrossprod(n1, n2)
}

.factor_congruence <- function(x, y, na.rm = TRUE, skip_checks = FALSE) {

  if (isFALSE(skip_checks)) {

    if (anyNA(x) || anyNA(y)) {
      if (isTRUE(na.rm)) {
        cli::cli_warn(
          c("The input contained missing values.",
            "i" = "The analysis is performed on complete cases."),
          class = "efa_missing_complete"
        )
        if (any(is.na(x))) {
          xc <- x[stats::complete.cases(x), , drop = FALSE]
          y <- y[stats::complete.cases(x), , drop = FALSE]
          x <- xc
        }
        if (any(is.na(y))) {
          yc <- y[stats::complete.cases(y), , drop = FALSE]
          x <- x[stats::complete.cases(y), , drop = FALSE]
          y <- yc
        }
        if (nrow(x) < 2L) {
          cli::cli_abort(
            c("Fewer than two complete cases remain after removing missing values.",
              "i" = "Cannot compute factor congruence; check the missing-value pattern."),
            class = "efa_too_few_complete"
          )
        }
      } else {
        cli::cli_warn(
          c("The input contained missing values.",
            "i" = "Check your data or rerun with {.code na.rm = TRUE}."),
          class = "efa_missing_check"
        )
      }
    }

  }


  .congruence_matrix(x, y)

}


### Align solution according to factor congruence

.align_solution <- function(L_target, L, Phi = NULL) {

  L_target <- as.matrix(L_target)
  L <- as.matrix(L)
  if (!identical(dim(L_target), dim(L))) {
    cli::cli_abort("{.arg L_target} and {.arg L} must have identical dimensions.",
                   class = "efa_dim_mismatch")
  }

  m <- ncol(L_target)
  target_colnames <- colnames(L_target)
  target_rownames <- rownames(L_target)

  congruence <- .tucker_congruence(L_target, L)

  if (m > 1) {

    cost <- max(abs(congruence), na.rm = TRUE) - abs(congruence)

    factor_order <- as.integer(clue::solve_LSAP(cost))

    # Reorder columns of L so column j is matched to target column j.
    L <- L[, factor_order, drop = FALSE]

    if (!is.null(Phi)) {
      Phi <- Phi[factor_order, factor_order, drop = FALSE]
    }

    # Get signs after reordering.
    factor_sign <- sign(diag(crossprod(L, L_target)))

    # Avoid accidental zero columns if an exact dot product is zero.
    factor_sign[factor_sign == 0] <- 1

    factor_sign_matrix <- diag(factor_sign, nrow = m, ncol = m)

    # Apply signs.
    L <- L %*% factor_sign_matrix

    if (!is.null(Phi)) {
      Phi <- factor_sign_matrix %*% Phi %*% factor_sign_matrix
    }

  } else {

    factor_order <- 1L
    factor_sign <- as.numeric(sign(congruence[1L, 1L]))
    if (factor_sign == 0) factor_sign <- 1

    L <- L * factor_sign

    # Phi unchanged for one factor, because sign^2 = 1.
  }

  # Only attach dimnames when the target actually carries some: assigning
  # list(NULL, NULL) would leave an empty-but-present attribute that callers
  # then have to scrub before it surfaces in downstream results (e.g. a diff).
  if (is.null(target_rownames) && is.null(target_colnames)) {
    dimnames(L) <- NULL
  } else {
    dimnames(L) <- list(target_rownames, target_colnames)
  }
  if (!is.null(Phi)) {
    dimnames(Phi) <- if (is.null(target_colnames)) {
      NULL
    } else {
      list(target_colnames, target_colnames)
    }
  }

  list(
    loadings = L,
    Phi = Phi,
    factor_order = factor_order,
    factor_sign = factor_sign
  )
}

#' Tucker congruence between factors
#'
#' Compute the Tucker congruence matrix between the columns of two loading
#' matrices.
#'
#' @param L1 Numeric matrix.
#' @param L2 Numeric matrix with the same dimensions as `L1`.
#'
#' @returns A square matrix whose `(i, j)` entry is the Tucker congruence
#'   between column `i` of `L1` and column `j` of `L2`.
#'
#' @references
#' Lorenzo-Seva, U., and ten Berge, J. M. F. (2006). Tucker's congruence
#' coefficient as a meaningful index of factor similarity. *Methodology*, 2,
#' 57-64.
#'
#' @keywords internal
.tucker_congruence <- function(L1, L2) {
  L1 <- .procrustes_as_matrix(L1, "L1")
  L2 <- .procrustes_as_matrix(L2, "L2")

  if (!identical(dim(L1), dim(L2))) {
    cli::cli_abort("{.arg L1} and {.arg L2} must have identical dimensions.",
                   class = "efa_dim_mismatch")
  }

  # Column Euclidean norms.
  n1 <- sqrt(colSums(L1 * L1))
  n2 <- sqrt(colSums(L2 * L2))

  if (any(n1 <= sqrt(.Machine$double.eps))) {
    cli::cli_abort("{.arg L1} contains at least one zero or near-zero column.",
                   class = "efa_zero_column")
  }
  if (any(n2 <= sqrt(.Machine$double.eps))) {
    cli::cli_abort("{.arg L2} contains at least one zero or near-zero column.",
                   class = "efa_zero_column")
  }

  # Full pairwise congruence matrix (shared core; reuse the validated norms).
  cong <- .congruence_matrix(L1, L2, n1, n2)

  # A non-finite congruence (e.g. column magnitudes that overflow when squared)
  # is as undefined as a zero column; flag it so callers get a clean classed
  # abort instead of a downstream solver or assignment failure.
  if (any(!is.finite(cong))) {
    cli::cli_abort(
      "{.arg L1} and {.arg L2} produced a non-finite Tucker congruence.",
      class = "efa_undefined_congruence"
    )
  }

  rownames(cong) <- colnames(L1)
  colnames(cong) <- colnames(L2)
  cong
}

# Unit tests for .array_reorder() (R/averaging.R), which aligns each efa_average()
# replicate's factors to a common column order and sign before they are averaged.
# Inputs are literal arrays with a known permutation, so both the reordered output
# and the optimality of the chosen permutation can be asserted exactly.

arr_re_NA <- .array_reorder(L = array(c(rep(.6, 6), rep(0, 12),
                                     rep(0, 6), rep(.6, 6), rep(0, 6),
                                     rep(0, 6), rep(0, 6), rep(.6, 6),
                                     rep(.6, 6), rep(0, 12),
                                     rep(0, 6), rep(.6, 6), rep(0, 6),
                                     rep(0, 6), rep(0, 6), rep(.6, 6),
                                     rep(0, 6), rep(.6, 6), rep(0, 6),
                                     rep(-.6, 6), rep(0, 12),
                                     rep(0, 6), rep(0, 6), rep(.6, 6)),
                                   c(18, 3, 3)),
                            vars_accounted = array(c(rep(.2, 3),
                                                     rep(.3, 3),
                                                     rep(.4, 3),
                                                     rep(.2, 3),
                                                     rep(.3, 3),
                                                     rep(.4, 3),
                                                     rep(.3, 3),
                                                     rep(.2, 3),
                                                     rep(.4, 3)),
                                                   c(3, 3, 3)),
                         L_corres = array(as.numeric(abs(c(rep(.6, 6), rep(0, 12),
                                            rep(0, 6), rep(.6, 6), rep(0, 6),
                                            rep(0, 6), rep(0, 6), rep(.6, 6),
                                            rep(.6, 6), rep(0, 12),
                                            rep(0, 6), rep(.6, 6), rep(0, 6),
                                            rep(0, 6), rep(0, 6), rep(.6, 6),
                                            rep(0, 6), rep(.6, 6), rep(0, 6),
                                            rep(.6, 6), rep(0, 12),
                                            rep(0, 6), rep(0, 6), rep(.6, 6)))> .3),
                                          c(18, 3, 3)),
                         phi = NA, extract_phi = FALSE, n_factors = 3)

arr_re <- .array_reorder(L = array(c(rep(.6, 6), rep(0, 12),
                                        rep(0, 6), rep(.6, 6), rep(0, 6),
                                        rep(0, 6), rep(0, 6), rep(.6, 6),
                                        rep(.6, 6), rep(0, 12),
                                        rep(0, 6), rep(.6, 6), rep(0, 6),
                                        rep(0, 6), rep(0, 6), rep(.6, 6),
                                        rep(0, 6), rep(.6, 6), rep(0, 6),
                                        rep(-.6, 6), rep(0, 12),
                                        rep(0, 6), rep(0, 6), rep(.6, 6)),
                                      c(18, 3, 3)),
                         vars_accounted = array(c(rep(.2, 3),
                                                  rep(.3, 3),
                                                  rep(.4, 3),
                                                  rep(.2, 3),
                                                  rep(.3, 3),
                                                  rep(.4, 3),
                                                  rep(.3, 3),
                                                  rep(.2, 3),
                                                  rep(.4, 3)),
                                                c(3, 3, 3)),
                            L_corres = array(as.numeric(abs(c(rep(.6, 6), rep(0, 12),
                                                              rep(0, 6), rep(.6, 6), rep(0, 6),
                                                              rep(0, 6), rep(0, 6), rep(.6, 6),
                                                              rep(.6, 6), rep(0, 12),
                                                              rep(0, 6), rep(.6, 6), rep(0, 6),
                                                              rep(0, 6), rep(0, 6), rep(.6, 6),
                                                              rep(0, 6), rep(.6, 6), rep(0, 6),
                                                              rep(.6, 6), rep(0, 12),
                                                              rep(0, 6), rep(0, 6), rep(.6, 6)))> .3),
                                             c(18, 3, 3)),
                            phi = array(rep(c(1, .3, .4, .3, 1, .2, .4, .2, 1), 3), c(3, 3, 3)),
                         extract_phi = TRUE, n_factors = 3)
test_that(".array_reorder works", {
  ### tests for arr_re_NA with phi = NA and extract_phi = FALSE
  expect_type(arr_re_NA, "list")
  expect_named(arr_re_NA, c("L", "L_corres", "phi", "vars_accounted"))
  checkmate::expect_array(arr_re_NA$L)
  expect_equal(dim(arr_re_NA$L), c(18, 3, 3))
  expect_equal(arr_re_NA$L,
               array(c(rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6),
                       rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6),
                       rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6)),
                      c(18, 3, 3)))
  checkmate::expect_array(arr_re_NA$L_corres)
  expect_equal(dim(arr_re_NA$L_corres), c(18, 3, 3))
  expect_equal(arr_re_NA$L_corres,
               array(as.numeric(c(rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6),
                       rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6),
                       rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6)) > .3),
                     c(18, 3, 3)))
  expect_equal(arr_re_NA$phi, NA)
  checkmate::expect_array(arr_re_NA$vars_accounted)
  expect_equal(arr_re_NA$vars_accounted,
               array(c(rep(.2, 3),
                       rep(.3, 3),
                       rep(.4, 3),
                       rep(.2, 3),
                       rep(.3, 3),
                       rep(.4, 3),
                       rep(.2, 3),
                       rep(.3, 3),
                       rep(.4, 3)),
                     c(3, 3, 3)))

  ### tests for arr_re with phi = array() and extract_phi = TRUE
  expect_type(arr_re, "list")
  expect_named(arr_re, c("L", "L_corres", "phi", "vars_accounted"))
  checkmate::expect_array(arr_re$L)
  expect_equal(dim(arr_re$L), c(18, 3, 3))
  expect_equal(arr_re$L,
               array(c(rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6),
                       rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6),
                       rep(.6, 6), rep(0, 12),
                       rep(0, 6), rep(.6, 6), rep(0, 6),
                       rep(0, 6), rep(0, 6), rep(.6, 6)),
                     c(18, 3, 3)))
  checkmate::expect_array(arr_re$L_corres)
  expect_equal(dim(arr_re$L_corres), c(18, 3, 3))
  expect_equal(arr_re$L_corres,
               array(as.numeric(c(rep(.6, 6), rep(0, 12),
                                  rep(0, 6), rep(.6, 6), rep(0, 6),
                                  rep(0, 6), rep(0, 6), rep(.6, 6),
                                  rep(.6, 6), rep(0, 12),
                                  rep(0, 6), rep(.6, 6), rep(0, 6),
                                  rep(0, 6), rep(0, 6), rep(.6, 6),
                                  rep(.6, 6), rep(0, 12),
                                  rep(0, 6), rep(.6, 6), rep(0, 6),
                                  rep(0, 6), rep(0, 6), rep(.6, 6)) > .3),
                     c(18, 3, 3)))
  checkmate::expect_array(arr_re$phi)
  expect_equal(dim(arr_re$phi), c(3, 3, 3))
  expect_equal(arr_re$phi,
               array(c(1, .3, .4, .3, 1, .2, .4, .2, 1,
                       1, .3, .4, .3, 1, .2, .4, .2, 1,
                       1, -.3, -.2, -.3, 1, .4, -.2, .4, 1), c(3, 3, 3)))
  checkmate::expect_array(arr_re$vars_accounted)
  expect_equal(arr_re$vars_accounted,
               array(c(rep(.2, 3),
                       rep(.3, 3),
                       rep(.4, 3),
                       rep(.2, 3),
                       rep(.3, 3),
                       rep(.4, 3),
                       rep(.2, 3),
                       rep(.3, 3),
                       rep(.4, 3)),
                     c(3, 3, 3)))

})

test_that(".array_reorder uses an optimal permutation alignment", {
  # Two solutions engineered so that a greedy per-row congruence match maps both
  # target factors 1 and 3 onto source column 1 (order 1, 2, 1): column 1 would
  # be duplicated and column 3 dropped, corrupting the average. The optimal
  # (linear-sum-assignment) alignment instead recovers the true permutation.
  L1 <- cbind(c(.9, .8, 0, 0, 0, 0),
              c(0, 0, .9, .8, 0, 0),
              c(0, 0, 0, 0, .9, .8))
  Ln <- cbind(c(.2, .1, 0, 0, .9, .8),
              c(0, 0, .9, .8, 0, 0),
              c(0, 0, 0, 0, .2, .1))

  # The greedy which.max alignment used previously collapses to a non-permutation.
  greedy_order <- apply(abs(.factor_congruence(L1, Ln, skip_checks = TRUE)), 1,
                        which.max)
  expect_equal(greedy_order, c(1, 2, 1))

  L_arr <- array(c(L1, Ln), dim = c(6, 3, 2))
  corres_arr <- array(as.numeric(abs(c(L1, Ln)) > .3), dim = c(6, 3, 2))
  va_arr <- array(1, dim = c(3, 3, 2))

  res <- .array_reorder(vars_accounted = va_arr, L = L_arr, L_corres = corres_arr,
                        phi = NA, extract_phi = FALSE, n_factors = 3)

  aligned <- res$L[, , 2]
  # All three source factors are retained (no duplicated/dropped column) ...
  expect_equal(nrow(unique(round(t(aligned), 8))), 3)
  # ... and the recovered order is the true permutation, leaving Ln in place.
  expect_equal(aligned, Ln)
})


rm(arr_re_NA, arr_re)

#' Print and format an efa_average object
#'
#' `print()` shows a summarised output of the [efa_average()] function: the
#' averaging settings, the error/convergence/Heywood/admissibility rates, the
#' indicator-to-factor correspondences, the averaged loadings (and, for oblique
#' solutions, the factor intercorrelations), the variances accounted for, and the
#' model fit. `format()` assembles the same report and returns it as a character
#' vector; by default (`plot = FALSE`) `print()` is `cat(format(x), sep = "\n")`.
#' With `plot = TRUE` it additionally draws the loading plot, which is the one
#' thing `format()` cannot return: the printed lines are the same, but the call
#' has a side effect beyond them. The lines follow the active console theme, so
#' they are plain when colours are disabled (for example when captured into a file
#' or stripped with [cli::ansi_strip()]).
#'
#' @param x An object of class `efa_average` (output from [efa_average()]).
#' @param stat character. A vector with the statistics to print. Possible inputs
#' are "average", "sd", "range", "min", and "max". Default is "average" and
#' "range".
#' @param plot logical. Whether a plot of the average and min- max loadings should
#' be created. Default is FALSE. If more than 10 factors are extracted, no plot is
#' created. Only used by `print()`; [plot.efa_average()] draws the same plot and
#' returns it, so it is the route to take when the plot object itself is wanted.
#' @param ...  Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @export
#'
#' @method print efa_average
#'
#' @examples
#' \donttest{
#' EFA_aver <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500)
#' EFA_aver
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(EFA_aver))
#' }
print.efa_average <- function(x, stat = c("average", "range"),
                              plot = FALSE, ...) {
  cat(format(x, stat = stat, ...), sep = "\n")

  # Plotting is a genuine side-effect, so it stays in print() (format() is text-only). Skip it
  # when no solution survived: the loadings are all NA and format() has already flagged this.
  if (isTRUE(plot) && !.efa_average_no_solutions(x$implementations_grid)) {

    if (ncol(x$loadings$average) <= 10) {

      print(plot(x))

    } else {

      obj_name <- deparse1(substitute(x))
      cli::cli_inform(c(
        "i" = "The factor solution contained more than 10 factors, no plot was generated.",
        "i" = "If you still want to create the plot, use {.code plot({obj_name})}."
      ))

    }

  }

  invisible(x)
}

#' @rdname print.efa_average
#' @export
#' @method format efa_average
format.efa_average <- function(x, stat = c("average", "range"), ...) {

  checkmate::assert_subset(stat, c("average", "sd", "range", "min", "max"),
                           empty.ok = FALSE)

  # extract settings
  settings <- x$settings
  estimator <- settings$estimator
  N <- settings$N
  grid <- x$implementations_grid
  averaging <- settings$averaging

  # settings that were varied: keep only the grid columns that correspond to EFA
  # settings. The per-model outcome columns appended to the grid (errors,
  # convergence, Heywood/admissibility flags, fit indices) are never part of the
  # settings list, so this excludes them without a column to maintain.
  varied_settings <- grid[, intersect(names(grid), names(settings)), drop = FALSE]
  # vapply keeps one count per column: apply() would simplify to a matrix when every
  # column has the same number of distinct values, and the count would then run over
  # the cells instead of the columns.
  n_unique <- vapply(varied_settings, function(z) length(unique(z[!is.na(z)])),
                     integer(1))
  varied_settings <- names(n_unique[n_unique > 1])
  # The grid and settings keep the historical key `P_type`; the argument is `p_type`, so
  # display it under that name.
  varied_settings[varied_settings == "P_type"] <- "p_type"

  # quantities embedded in the averaging sentence (the rates are built by
  # .efa_emit_average_rates() from the grid alone)
  no_efas <- nrow(grid)
  averaging_method <- if (averaging == "median") {
    "median"
  } else {
    paste0("mean (trim = ", settings$trim, ")")
  }

  cli::cli_format_method({

    # Prose sentences wrap to the console width; the varied settings collapse into a
    # comma-and-"and" list automatically (each value emphasised individually).
    cli::cli_text("")
    cli::cli_text("Averaging performed with averaging method {.strong {averaging_method}} across {.strong {no_efas}} EFAs, varying the following settings: {.strong {varied_settings}}.")

    cli::cli_text("")
    .efa_emit_average_rates(grid)

    # If no solutions were achieved across which averaging could be performed, flag it and stop
    # after the summary (there are no loadings or fit indices to show). Otherwise show the full
    # report.
    if (.efa_average_no_solutions(grid)) {

      cli::cli_text("")
      cli::cli_alert_warning("No solutions were achieved across which averaging was possible. Best try again with a different number of factors.",
                             wrap = TRUE)

    } else {

      fit <- x$fit_indices
      rownames(fit) <- fit$index

      # Indicator-to-factor correspondences
      .efa_section_rule("Indicator-to-Factor Correspondences")
      salience_threshold <- settings$salience_threshold
      cli::cli_text("")
      cli::cli_text("For each cell, the proportion of solutions including the respective indicator-to-factor correspondence. A salience threshold of {.strong {salience_threshold}} was used to determine indicator-to-factor correspondences.")
      cli::cli_text("")
      .efa_emit_lines(format(
        structure(x$ind_fac_corres, class = c("efa_loadings", "LOADINGS")),
        cutoff = 1e-4, digits = 2))

      # Loadings
      .efa_section_rule("Loadings")
      .print_average(x, what = "loadings", stat = stat, averaging = averaging)

      # Factor intercorrelations for oblique solutions
      if (!all(is.na(x$Phi))) {
        .efa_section_rule("Factor Intercorrelations from Oblique Solutions")
        .print_average(x, what = "Phi", stat = stat, averaging = averaging)
      }

      # Variances accounted for
      .efa_section_rule("Variances Accounted for")
      .print_average(x, what = "vars_accounted", stat = stat, averaging = averaging)

      # Model fit
      if (fit["df", "average"] == 0) {
        cli::cli_text("")
        cli::cli_alert_warning("The model is just identified (df = 0). Goodness of fit indices may not be interpretable.",
                               wrap = TRUE)
      }

      .efa_section_rule("Model Fit")
      cli::cli_text("")
      # `averaging` is exactly "mean" or "median", so it names the aggregation the block
      # actually reports (the header below abbreviates the same choice as M or Md).
      cli::cli_text("The fit indices are the {averaging} of the per-solution fit indices, not the fit of the averaged loadings printed above, which are a cell-wise summary rather than a fitted solution.")

      # Effective n per index class. .extract_data() records a solution's fit indices only
      # once it converged without a Heywood case, and leaves the chi-square-based ones NA
      # for PAF throughout, so the two classes can be averaged over different numbers of
      # solutions. Each class is counted on the column that defines it (`chisq` for the
      # chi-square-based indices, `caf` for the residual-based ones), which is exactly the
      # non-missing count the averaging itself used.
      n_resid <- sum(!is.na(grid$caf))
      chisq_based <- !(all(estimator == "PAF") || is.na(N))
      cli::cli_text("")
      if (chisq_based) {
        n_chisq <- sum(!is.na(grid$chisq))
        cli::cli_text("Chi-square-based indices averaged over {n_chisq} of {no_efas} solution{?s}.")
      }
      cli::cli_text("CAF, RMSR, and SRMR averaged over {n_resid} of {no_efas} solution{?s}.")
      cli::cli_text("")
      cli::cli_verbatim(paste0("       ", if (averaging == "mean") "M" else "Md",
                               " (SD) [Min; Max]"))

      if (!chisq_based) {

        lines <- .gof_lines(fit, ind = c("caf", "rmsr", "srmr"),
                            ind_name = c("CAF:  ", "RMSR: ", "SRMR: "),
                            print_zero = c(FALSE, FALSE, FALSE),
                            digits = c(2, 2, 2))
        lines <- c(lines, paste0("df: ",
          .efa_num(fit["df", "average"], 0, print_zero = TRUE)))
        cli::cli_verbatim(lines)

      } else {

        chisq <- .gof_lines(fit, ind = "chisq", ind_name = "\U03C7\U00B2: ",
                            print_zero = TRUE, digits = 2)
        df_line <- paste0("df: ",
                          .efa_num(fit["df", "average"], 0, print_zero = TRUE))
        rest <- .gof_lines(fit,
          ind = c("p_chi", "cfi", "tli", "rmsea", "aic", "bic",
                  "ecvi", "caf", "rmsr", "srmr"),
          ind_name = c(.efa_style("p: ", "italic"), "CFI: ", "TLI: ",
                       "RMSEA: ", "AIC: ", "BIC: ", "ECVI: ", "CAF: ",
                       "RMSR: ", "SRMR: "),
          print_zero = c(FALSE, FALSE, FALSE, FALSE, TRUE, TRUE,
                         TRUE, FALSE, FALSE, FALSE),
          digits = c(3, 2, 2, 2, 2, 2, 2, 2, 2, 2))
        cli::cli_verbatim(c(chisq, df_line, rest))

      }

    }
  })
}

# TRUE when no fitted solution survived to be averaged (every run errored, did not converge,
# or was a Heywood case), so there are no loadings or fit indices to report. Consumed by both
# format() (to flag it and stop after the summary) and print() (to suppress plotting).
.efa_average_no_solutions <- function(grid) {
  all((grid$converged != 0 | grid$errors | grid$heywood) %in% TRUE)
}

# Emit the error/convergence/Heywood/admissibility summary for an implementations grid.
#
# Each conditional rate has its own denominator, because .extract_data() records an outcome
# only once the preceding stage succeeded: `converged` is set for every solution that did not
# error, while `heywood` and `admissible` are set only for those that then converged (the
# unassessed entries stay NA and drop out of the rate). Name the denominator each rate is
# conditioned on, and drop a clause when its denominator is empty: mean(<all NA>, na.rm =
# TRUE) is NaN, and a rate over no solutions is undefined rather than 0%. Each rate is
# therefore computed on the branch that reports it, which leaves an empty denominator no
# variable to sit in.
.efa_emit_average_rates <- function(grid) {

  pct <- function(x) paste0(round(mean(x, na.rm = TRUE) * 100), "%")

  error_pct <- pct(grid$errors)
  clauses <- "The error rate is at {.strong {error_pct}}."

  # `%in% TRUE` keeps an NA flag from propagating into the condition, as in
  # .efa_average_no_solutions() above.
  if (all(grid$errors %in% TRUE)) {

    clauses <- c(clauses, "No solution could be fitted, so convergence, Heywood cases, and admissibility could not be assessed.")

  } else {

    converged_pct <- pct(grid$converged == 0)
    clauses <- c(clauses, "Of the solutions that did not result in an error, {.strong {converged_pct}} converged.")

    if (!any(grid$converged == 0, na.rm = TRUE)) {

      clauses <- c(clauses, "No solution converged, so Heywood cases and admissibility could not be assessed.")

    } else {

      heywood_pct <- pct(grid$heywood)
      admissible_pct <- pct(grid$admissible)
      clauses <- c(clauses, "Of the solutions that converged, {.strong {heywood_pct}} contained Heywood cases and {.strong {admissible_pct}} were admissible.")

    }

  }

  # A single cli_text() so the clauses wrap as one paragraph rather than one block each.
  cli::cli_text(paste(clauses, collapse = " "))

  invisible(NULL)
}

# Emit a double-ruled (==) header for a major efa_average section: a blank line followed by
# the rule, mirroring the original `cat(rule(line = 2))` spacing (the following blank, if any,
# is supplied by the next element). The double rule keeps the major/minor (==/--) visual
# hierarchy, with the minor sub-rules emitted by `.print_efa_rule()`.
.efa_section_rule <- function(title) {
  cli::cli_text("")
  cli::cli_verbatim(cli::rule(left = cli::style_bold(title), line = 2))
  invisible(NULL)
}

.print_average <- function(x, what, stat, averaging) {

  # Emit one statistic's table into the cli report: average/min/max loadings are
  # LOADINGS-classed (styled loading table), captured verbatim with their colour; everything
  # else (sd/range loadings, Phi, variances) is a plain "corr" matrix. Phi is symmetric, so
  # only its lower triangle shows.
  emit <- function(stat_key) {
    if (what == "loadings" && stat_key %in% c("average", "min", "max")) {
      .efa_emit_lines(format(x$loadings[[stat_key]]))
    } else if (what == "Phi") {
      .efa_emit_lines(.efa_corr_lines(x$Phi[[stat_key]], lower_only = TRUE))
    } else if (what == "loadings") {
      .efa_emit_lines(.efa_corr_lines(x$loadings[[stat_key]]))
    } else {
      .efa_emit_lines(.efa_corr_lines(x$vars_accounted[[stat_key]]))
    }
  }

  # Section title per statistic, in the order they are printed (which is fixed here, not
  # taken from the order of `stat`). The range is labelled with its definition, since the
  # cells are widths, not the [Min; Max] intervals the Model Fit block reports.
  titles <- c(average = if (averaging == "mean") "Mean" else "Median",
              sd = "Standard Deviation",
              range = "Range (max \u2212 min)",
              min = "Minimum",
              max = "Maximum")

  for (stat_key in intersect(names(titles), stat)) {
    .print_efa_rule(titles[[stat_key]])
    emit(stat_key)
  }

  invisible(NULL)
}

# Build the "M (SD) [Min; Max]" report line for each averaged fit index. `.efa_num` drops the
# leading zero of values in (-1, 1) when print_zero is FALSE (and keeps it otherwise) and
# preserves the sign, so it already produces the stripped/unstripped form each index needs.
# The integer-valued indices (print_zero = TRUE, e.g. AIC/BIC) are left-padded to align the
# bracketed range; the decimal-only ones are not.
.gof_lines <- function(fit, ind, ind_name, print_zero, digits) {
  fmt <- function(stat, i) {
    .efa_num(fit[ind[i], stat], digits = digits[i],
             print_zero = print_zero[i], pad = print_zero[i])
  }

  vapply(seq_along(ind), function(i) {
    paste0(ind_name[i], fmt("average", i), " (", fmt("sd", i), ") [",
           fmt("min", i), "; ", fmt("max", i), "]")
  }, character(1L))
}

# Detect or build the analysis correlation matrix, then validate it once: assert the
# input type, smooth a non-positive-definite matrix, and resolve N. Shared by EFA and
# the suitability/retention functions so these checks live in one place.

# Unit diagonal on one shared tolerance: the correlation-vs-covariance classification
# hinges on .is_cormat() and .is_covmat() applying the SAME test, so any matrix falls on
# exactly one side of it.
.has_unit_diag <- function(x, na.rm = FALSE) {
  all(abs(diag(x) - 1) <= .Machine$double.eps * 100, na.rm = na.rm)
}

# Everything that identifies a correlation matrix except symmetry: at least two columns,
# square, all-numeric, a unit diagonal, and every entry in [-1, 1]. Two callers need exactly
# this test -- .is_cormat(), which adds the symmetry requirement, and .assert_cor_input(),
# which rejects an input that meets it but is asymmetric -- so it lives in one place and the
# two cannot drift apart. Returns the matrix form of `x` when the shape matches and NULL
# otherwise, so the caller neither re-tests nor re-coerces.
.cormat_shape <- function(x) {

  # Cheapest test first, and before any coercion, for the reason .is_covmat() documents: raw
  # data is almost never square, and as.matrix() on a data frame copies the whole thing. The
  # two-column minimum matches .is_covmat()'s own guard and keeps a 1x1 input -- not an
  # analysable correlation matrix -- on the raw-data route, where it is rejected.
  if (ncol(x) < 2L || nrow(x) != ncol(x)) return(NULL)

  # A data frame is coerced here rather than tested in place: diag() on a data frame is a hard
  # base error, and isSymmetric() needs a real matrix too. The all-numeric test guards both the
  # coercion and the range comparisons that follow -- on a factor column the comparisons would
  # dispatch to Ops.factor and fail with an unclassed base error. Anything non-numeric is not a
  # correlation matrix, so it falls through to the raw-data path and is rejected there with a
  # classed condition.
  if (is.data.frame(x)) {
    if (!all(vapply(x, is.numeric, logical(1)))) return(NULL)
    x <- as.matrix(x)
  }
  if (!is.numeric(x)) return(NULL)

  if (!all(x >= (-1 - .Machine$double.eps * 100), na.rm = TRUE) ||
      !all(x <= (1 + .Machine$double.eps * 100), na.rm = TRUE)) return(NULL)

  if (!.has_unit_diag(x, na.rm = TRUE)) return(NULL)

  x
}

# Heuristically decide whether x is already a correlation matrix rather than raw data:
# .cormat_shape() plus symmetry. This cannot distinguish a correlation matrix from a (rare)
# raw-data matrix that happens to meet all of these conditions, and it treats a covariance
# matrix as raw data (its diagonal is not all ones); callers that accept either input assume
# raw data unless these conditions hold.
.is_cormat <- function(x){

  xm <- .cormat_shape(x)

  if (is.null(xm) || !isSymmetric(unclass(unname(xm)))) return(FALSE)

  if (any(is.na(xm))) {

    cli::cli_abort(
      c("{.arg x} looks like a correlation matrix but contains missing values.",
        "i" = "Please check the entered data."),
      class = "efa_cormat_has_na"
    )

  }

  TRUE

}

# Abort unless `x` is a matrix or data frame. `raw_only = TRUE` tailors the
# message for functions that accept only raw data (e.g. CD), which reject a
# correlation matrix downstream.
.assert_cor_input <- function(x, raw_only = FALSE, error_call = rlang::caller_env()) {
  if (!inherits(x, c("matrix", "data.frame"))) {
    lead <- if (raw_only) {
      "{.arg x} must be a data frame/matrix of raw data."
    } else {
      "{.arg x} must be a correlation matrix or a data frame/matrix of raw data."
    }
    cli::cli_abort(
      c(lead, "x" = "You supplied {.obj_type_friendly {x}}."),
      class = "efa_input_not_matrix",
      call = error_call
    )
  }
  if (.is_covmat(x)) {
    # cov2cor() is the fix only for a caller that accepts a correlation matrix; one that
    # needs raw data (raw_only) would just hit the correlation-matrix rejection next.
    cov_lead <- if (raw_only) {
      "{.arg x} looks like a covariance matrix, not a data frame/matrix of raw data."
    } else {
      "{.arg x} looks like a covariance matrix, not a correlation matrix or raw data."
    }
    cov_remedy <- if (raw_only) {
      "Supply the raw observations the covariance matrix was computed from."
    } else {
      "Standardise it first: {.code cov2cor(x)}."
    }
    cli::cli_abort(
      c(cov_lead,
        "x" = "It is square and symmetric, with variances rather than ones on the diagonal.",
        "i" = cov_remedy),
      class = "efa_input_is_covmat", call = error_call
    )
  }
  # A square, unit-diagonal, in-range matrix that is not symmetric is a correlation matrix
  # transcribed one triangle at a time, or typed at too few decimals to mirror to machine
  # precision (isSymmetric()'s tolerance is 100 * .Machine$double.eps). Neither classifier
  # accepts it, so passed on it would be fed to stats::cor() as if its p columns were p cases
  # -- singular by construction, and reported as such, which sends the user looking for
  # collinearity that does not exist. It is deliberately not symmetrised here: with one
  # triangle left empty the other is authoritative, but an asymmetry from a transcription slip
  # leaves neither triangle authoritative, and repairing it either way would analyse a matrix
  # the user never entered.
  xm <- .cormat_shape(x)
  if (!is.null(xm) && !isSymmetric(unclass(unname(xm)))) {
    sym_lead <- if (raw_only) {
      "{.arg x} looks like a (non-symmetric) correlation matrix, not a data frame/matrix of raw data."
    } else {
      "{.arg x} looks like a correlation matrix but is not symmetric."
    }
    # The unit-diagonal and range tests run with na.rm = TRUE, so a matrix whose diagonal or
    # whose unused triangle was left empty reaches this branch; do not claim of missing cells
    # that they are in range.
    sym_detail <- if (anyNA(xm)) {
      "Every entry that is present lies in [-1, 1] and no diagonal entry departs from one, but part of the matrix is missing and {.code x[i, j]} and {.code x[j, i]} differ."
    } else {
      "Its diagonal is all ones and every entry lies in [-1, 1], but {.code x[i, j]} and {.code x[j, i]} differ."
    }
    # Which triangle to mirror has to be read off the data: the hint for a transcribed lower
    # triangle overwrites the entries of a transcribed upper one, which would leave a matrix
    # that passes every later check and analyses as the identity. An unused triangle is left
    # either empty (NA) or at zero. Mirroring is offered only when exactly one triangle was
    # filled in: when both carry entries neither is authoritative (the same reasoning that
    # keeps this branch from symmetrising on its own), and when neither does there is nothing
    # to mirror -- copying an empty triangle would also produce the identity.
    tri_l <- xm[lower.tri(xm)]
    tri_u <- xm[upper.tri(xm)]
    empty <- function(tri) all(is.na(tri) | tri == 0)
    sym_remedy <- if (raw_only) {
      "Supply the raw observations the correlation matrix was computed from."
    } else if (empty(tri_u) && !empty(tri_l)) {
      "Only the lower triangle carries entries; mirror it: {.code x[upper.tri(x)] <- t(x)[upper.tri(x)]}."
    } else if (empty(tri_l) && !empty(tri_u)) {
      "Only the upper triangle carries entries; mirror it: {.code x[lower.tri(x)] <- t(x)[lower.tri(x)]}."
    } else if (empty(tri_l) && empty(tri_u)) {
      "Neither triangle carries a correlation, so there is nothing to mirror; enter the off-diagonal correlations."
    } else {
      "Both triangles carry entries but they disagree, so neither can be mirrored onto the other; check the entries of the pairs that differ."
    }
    cli::cli_abort(
      c(sym_lead,
        "x" = sym_detail,
        "i" = sym_remedy),
      class = "efa_input_not_symmetric", call = error_call
    )
  }
  invisible(x)
}

# Recognise a covariance matrix, which is neither a correlation matrix (its diagonal is not
# all ones, so .is_cormat() returns FALSE) nor usable as raw data: passed on it would be fed
# to stats::cor(), i.e. the correlation of the covariance matrix's own p columns over p
# "cases", which is singular by construction and misdiagnoses the user's input. A square,
# symmetric, all-finite numeric matrix with positive non-unit diagonal entries is never raw
# data in practice -- symmetry alone is already decisive. Positive semi-definiteness is
# deliberately NOT required: the covariance matrices this guard exists for -- built with
# pairwise deletion, or transcribed from published tables at few decimals -- are routinely
# indefinite, and an indefinite correlation matrix is likewise accepted (and smoothed
# downstream), so demanding PSD of the covariance side only would send exactly the
# problematic inputs down the raw-data misdiagnosis. The positive-diagonal test keeps
# zero-diagonal distance/dissimilarity matrices out.
.is_covmat <- function(x) {

  # Cheapest test first, and before any coercion: raw data is almost never square, and
  # as.matrix() on a data frame copies the whole thing -- which would otherwise be paid on
  # every call for every raw-data input.
  if (ncol(x) < 2L || nrow(x) != ncol(x)) return(FALSE)

  if (is.data.frame(x)) {
    if (!all(vapply(x, is.numeric, logical(1)))) return(FALSE)
    x <- as.matrix(x)
  }
  if (!is.numeric(x) || !all(is.finite(x))) return(FALSE)

  if (any(diag(x) <= 0)) return(FALSE)

  # A unit diagonal (on the shared tolerance, so the two classifiers are exactly
  # complementary) means this is a correlation matrix, handled by .is_cormat().
  if (.has_unit_diag(x)) return(FALSE)

  isSymmetric(unclass(unname(x)))
}

# Single source of truth for the correlation methods that are estimated from the
# raw data by .polychoric() rather than stats::cor(). Used both to route the
# computation and to reject criteria whose reference data are continuous.
.is_poly_cor <- function(cor_method) cor_method %in% c("poly", "tetra")

# Labels for the off-diagonal variable pairs in utils::combn(p, 2) column order
# ("name_i-name_j"), shared by every routine that names a pair-indexed vector or matrix
# (the polychoric and ADF asymptotic covariances and DWLS weights).
.pair_labels <- function(nms) {
  idx <- utils::combn(length(nms), 2L)
  paste(nms[idx[1L, ]], nms[idx[2L, ]], sep = "-")
}

# Cap a list of affected labels for a condition message at the first five plus a count of the
# rest: a heterogeneous ordinal set can have dozens of bad variable pairs, and a data set read
# in as character has one bad column per variable, so a message that prints them all is
# unreadable. Returns the pieces rather than a finished string so each caller keeps its own cli
# template. When the list is truncated, `shown` is a cli_vec whose "and" is dropped so the
# trailing count closes the enumeration instead of adding a second conjunction. Shared by every
# condition that enumerates affected variables -- the boundary and sparse-cell diagnostics, the
# asymptotic covariance screen, the DWLS weight abort, and the uncomputable-correlation column
# diagnosis -- so their wording cannot drift apart.
.cap_label_list <- function(labels) {
  shown <- utils::head(labels, 5L)
  more <- length(labels) - length(shown)
  if (more == 0L) return(list(shown = shown, rest = ""))
  list(shown = cli::cli_vec(shown, style = list("vec-last" = ", ")),
       rest = paste0(", and ", more, " more"))
}

# Build the symmetric per-element DWLS weight matrix W (W_ij = 1 / Var(rho_hat_ij), zero
# diagonal) from the diagonal polychoric asymptotic covariance returned by .polychoric():
# a length p(p - 1)/2 vector of off-diagonal variances ordered by the upper triangle
# (i < j) in utils::combn() column order. The combn order is row-major, which differs
# from R's column-major upper.tri(), so the pairs are placed via explicit (i, j) indexing
# rather than `W[upper.tri(W)] <- ...`. Shared by the DWLS point estimate and the per-
# replicate bootstrap recompute.
.poly_weight_matrix <- function(acov_diag, p) {
  # A non-positive or non-finite asymptotic variance means a variable pair whose inverse-variance
  # weight is undefined; refuse rather than emit Inf/NaN weights that would silently corrupt the
  # fit. NA arrives for a pair estimated at the boundary of the parameter space, which has no
  # asymptotic variance at all (see .polychoric()), so this refusal is deterministic rather than
  # dependent on which side of zero a floating-point artefact happened to land. (The bootstrap
  # recompute catches this and drops the replicate.)
  bad <- !is.finite(acov_diag) | acov_diag <= 0
  if (any(bad)) {
    n_bad <- sum(bad)
    pairs <- if (is.null(names(acov_diag))) paste0("pair ", which(bad)) else names(acov_diag)[bad]
    cap <- .cap_label_list(pairs)
    cli::cli_abort(
      c("DWLS needs an inverse-variance weight for every variable pair, but {n_bad} pair{?s} {?has/have} no usable asymptotic variance.",
        "x" = "Affected {cli::qty(n_bad)}pair{?s}: {.val {cap$shown}}{cap$rest}.",
        "i" = "The most common cause with binary variables is a response combination that never occurs: the correlation of such a pair is estimated with a continuity correction, which repairs the estimate but leaves it without a variance.",
        "i" = "Otherwise the pair carries too little information about its correlation to have one: its responses are perfectly ordered (never observed in reversed order), so the correlation sits on the boundary of the parameter space, or a response category is (near-)empty.",
        "i" = "Fit with {.code estimator = \"ULS\"} instead, which needs no such weights. Collapsing rare response categories in the variables involved, or dropping one item of a redundant pair, also resolves the cases that are genuinely under-identified."),
      class = "efa_dwls_degenerate_weight"
    )
  }
  W <- matrix(0, p, p)
  idx <- utils::combn(p, 2L)
  W[t(idx)] <- 1 / acov_diag
  W + t(W)
}

# One shared degeneracy check on the diagonal of the asymptotic covariance, applied where it
# is received so that every consumer -- the DWLS weights and the sandwich standard errors --
# takes the same branch on the same data. A variable pair whose asymptotic variance is unusable
# leaves the DWLS weight meaningless and the sandwich standard errors of that pair meaningless
# rather than merely large, and neither outcome should be silent, so warn once and name the pairs.
#
# The main case is a pair estimated at the boundary of the parameter space (a contingency table at
# a Frechet bound, i.e. perfectly ordered), which has no asymptotic variance at all and reaches
# this screen as NA by construction -- see .polychoric(). That is caught by the non-finite clause,
# so the branch taken here does not depend on the platform. The `> 1` clause is a residual net for
# any remaining pair whose variance is finite but useless: a variance above 1 already implies a
# +/- 1 SE interval as wide as the whole [-1, 1] range of a correlation. (Searches over
# strongly-correlated tables found no table off a Frechet bound that reaches it -- the largest was
# ~0.008 -- so it is expected to be inert, but it costs nothing to keep.) Non-positive values are
# degenerate a fortiori.
.warn_acov_degenerate <- function(acov_diag, labels = names(acov_diag)) {
  bad <- !is.finite(acov_diag) | acov_diag <= 0 | acov_diag > 1
  if (!any(bad)) return(invisible(NULL))

  pairs <- if (is.null(labels)) paste0("pair ", which(bad)) else labels[bad]
  cap <- .cap_label_list(pairs)

  cli::cli_warn(
    c("The polychoric asymptotic covariance is unavailable for {cli::qty(pairs)} variable pair{?s} {.val {cap$shown}}{cap$rest}.",
      "x" = "{cli::qty(pairs)}{?This pair was/These pairs were} either estimated with a continuity correction for a response combination that never occurs -- which repairs the correlation but leaves it without a variance -- or {?is/are} not identified well enough to have a usable one (e.g. a perfectly ordered or near-empty response table).",
      "i" = "Any robust/sandwich standard error involving {cli::qty(pairs)} {?it/them} is withheld. A DWLS fit refuses the data outright when the variance is missing or non-positive, and down-weights {cli::qty(pairs)} {?it/them} out of the solution when it is merely far too large.",
      "i" = "Fitting with {.code estimator = \"ULS\"} avoids the weights entirely; collapsing rare response categories in the variables involved resolves the under-identified cases."),
    class = "efa_acov_degenerate"
  )
  invisible(NULL)
}

# Asymptotic-distribution-free (ADF; Browne, 1984) covariance of the off-diagonal sample
# correlations from raw continuous data, on the variance scale Var(rho-hat) and in
# utils::combn(p, 2) column order -- the continuous analogue of the polychoric asymptotic
# covariance, and the meat of the robust/sandwich standard errors. For case n and pair (i < j)
# the per-case influence is IF[n, (ij)] = (w_ni w_nj - 1/2 r_ij (w_ni^2 + w_nj^2)) / N, with w
# the columns standardised to unit variance; Gamma = crossprod(IF). The -1/2 r (w_ni^2 + w_nj^2)
# terms are the delta-method correction for estimating the marginal SDs (the standardisation
# nuisance, the continuous counterpart of the polychoric threshold influence). N * Gamma equals
# lavaan's correlation-structure NACOV (the basis of the MLM/MLR robust statistics). Kept in R:
# the cost is the BLAS crossprod (about 0.4 s at p = 40, N = 800); the influence build is cheap.
.adf_gamma <- function(x) {
  x <- as.matrix(x)
  N <- nrow(x)
  p <- ncol(x)

  # Centre each column and scale by its population SD, so colMeans(W^2) == 1 and crossprod(W)/N
  # is exactly the Pearson correlation matrix (the 1/N vs 1/(N-1) factors cancel).
  W <- sweep(x, 2L, colMeans(x), "-")
  W <- sweep(W, 2L, sqrt(colSums(W^2) / N), "/")
  Wsq <- W^2

  pairs <- utils::combn(p, 2L)
  pi <- pairs[1, ]
  pj <- pairs[2, ]
  r <- (crossprod(W) / N)[cbind(pi, pj)]

  # One influence column per off-diagonal pair (combn order), with the 1/N folded in so the
  # crossprod lands on the variance scale.
  IF <- (W[, pi, drop = FALSE] * W[, pj, drop = FALSE] -
           0.5 * (Wsq[, pi, drop = FALSE] + Wsq[, pj, drop = FALSE]) * rep(r, each = N)) / N
  Gamma <- crossprod(IF)
  Gamma <- (Gamma + t(Gamma)) / 2          # symmetrise away round-off

  if (!is.null(colnames(x))) {
    labels <- .pair_labels(colnames(x))
    dimnames(Gamma) <- list(labels, labels)
  }
  Gamma
}

# The `use` policies that listwise-delete incomplete rows (so N is the number of
# complete cases and resampling/correlation run on complete data).
.is_listwise_use <- function(use) use %in% c("complete.obs", "na.or.complete")

# Project a symmetric matrix onto the positive definite correlation matrices: floor the
# eigenvalues, rescale the spectrum to keep the trace at p, rebuild, and standardise back to a
# unit diagonal. Same construction and same floor as psych::cor.smooth(), so the two agree
# wherever both act; used where cor.smooth() declines to act on a matrix that is still not
# positive definite (see .prepare_cor_input()).
.project_cor_pd <- function(R) {
  eig_tol <- 1e-12                       # cor.smooth()'s own `eig.tol`, and its 100x floor
  e <- eigen(R, symmetric = TRUE)
  ev <- e$values
  ev[ev < eig_tol] <- 100 * eig_tol
  ev <- ev * ncol(R) / sum(ev)
  # ev * t(V) scales row i of t(V) by ev[i], i.e. diag(ev) %*% t(V) without forming diag(ev).
  out <- stats::cov2cor(e$vectors %*% (ev * t(e$vectors)))
  dimnames(out) <- dimnames(R)
  out
}

# Detect or compute the correlation matrix, check invertibility, and smooth a
# non-positive-definite matrix. Assumes `x` has already been validated as a
# matrix or data frame (see .assert_cor_input()). Shared by EFA and the
# suitability/retention functions so these checks live in one place. The flags
# reproduce each function's specifics (whether N is needed, the wording of the
# singular message, the SPSS positive-definite abort, etc.).
.prepare_cor_input <- function(x,
                               N = NA_real_,
                               use = "pairwise.complete.obs",
                               cor_method = "pearson",
                               N_policy = c("optional", "none", "required"),
                               acov = c("none", "diag", "full"),
                               dwls = FALSE,
                               # EM budget for the two-stage FIML moments; ignored by every other
                               # cor_method. Taken FROM the control constructor rather than
                               # restated, so the entry points that thread an `estimate_control()`
                               # (efa_fit(), and what forwards to it) and those that cannot
                               # (efa_average()) can never estimate different correlations from the
                               # same data because one default was changed and the other was not.
                               # Both are evaluated lazily, inside the FIML branch only.
                               fiml_max_iter = estimate_control()$fiml_max_iter,
                               fiml_tol = estimate_control()$fiml_tol,
                               inform_from_data = TRUE,
                               check_singular = TRUE,
                               posdef_abort = FALSE,
                               singular_tail = "no further analyses are performed",
                               N_required_msg = c(
                                 "{.arg N} is {.val NA} but a correlation matrix was entered.",
                                 "i" = "Provide {.arg N} or raw data."),
                               error_call = rlang::caller_env()) {

  N_policy <- match.arg(N_policy)
  acov <- match.arg(acov)

  # DWLS weights (1 / diag(Gamma)); populated only on the polychoric raw-data path when
  # an asymptotic covariance is requested, NULL otherwise.
  weights <- NULL
  # Full p(p-1)/2 x p(p-1)/2 asymptotic covariance of the off-diagonal correlations
  # (Var(rho-hat) scale); populated only when acov = "full" (the sandwich-SE path), NULL
  # otherwise. It is the meat of the robust/sandwich standard errors.
  Gamma <- NULL
  # The EM-estimated saturated mean/covariance and saturated log-likelihood; populated only
  # on the two-stage FIML raw-data path (cor_method = "fiml"), NULL otherwise.
  fiml <- NULL
  # TRUE once `x` has been reduced to its listwise-complete rows below, so that a failure
  # diagnosed afterwards describes the rows the correlation was actually attempted on.
  listwise_reduced <- FALSE

  # TRUE exactly when an asymptotic covariance forces listwise deletion of incomplete rows:
  # the polychoric path deletes for any acov (inside .polychoric()), the Pearson/rank path only
  # for the full ADF covariance. Used both to report the override and to resolve N, so neither
  # claims a deletion that did not happen.
  acov_listwise <- acov == "full" || (.is_poly_cor(cor_method) && acov != "none")

  is_cormat <- .is_cormat(x)

  if (is_cormat) {

    # A data frame that passed .is_cormat() is a correlation matrix; carry it forward as a
    # matrix so every downstream consumer gets the type it expects (determinant(), the
    # eigendecompositions and the compiled estimators all reject a data frame).
    R <- as.matrix(x)

    if (N_policy == "required" && is.na(N)) {
      cli::cli_abort(N_required_msg, class = "efa_n_required", call = error_call)
    }

    # A correlation matrix carries no raw data, so an asymptotic covariance cannot be estimated
    # from it; report that the request is ignored rather than silently returning no weights.
    if (acov != "none") {
      cli::cli_warn(
        c("An asymptotic covariance was requested but {.arg x} is already a correlation matrix.",
          "i" = "{.arg acov} needs raw data and is ignored here."),
        class = "efa_acov_ignored"
      )
    }

  } else {

    if (inform_from_data) {
      cli::cli_inform(
        c("i" = "{.arg x} is not a correlation matrix; computing correlations from the raw data."),
        class = "efa_cor_from_data"
      )
    }

    if (N_policy != "none" && !is.na(N)) {
      cli::cli_warn(
        c("Both {.arg N} and raw data were supplied.",
          "i" = "Taking {.arg N} from the data."),
        class = "efa_n_from_data"
      )
    }

    # The asymptotic covariance is available from continuous (Pearson) data only as the full ADF
    # meat: a diagonal (DWLS-weight) covariance and any rank-correlation covariance are ordinal
    # constructs supplied on the polychoric path, not here. Reject the unsupported combinations
    # before reporting any listwise deletion below, so an override is not announced for a
    # computation that is then refused. EFA() gates these earlier; the guard keeps the helper's
    # contract local for any other caller.
    if (!.is_poly_cor(cor_method) && acov != "none" &&
        !(acov == "full" && cor_method == "pearson")) {
      cli::cli_abort(
        c("That asymptotic covariance is not available for {.code cor_method = {.val {cor_method}}}.",
          "x" = "Only {.code acov = \"full\"} with {.code cor_method = \"pearson\"} is supported for continuous data; {.code acov = \"diag\"} (DWLS weights) needs {.code cor_method = \"poly\"} or {.code \"tetra\"}."),
        class = "efa_acov_unsupported", call = error_call)
    }

    # An asymptotic covariance (the DWLS weights or the sandwich meat) must describe a single
    # set of complete cases, so incomplete rows are listwise-deleted below even though `use`
    # asks for pairwise-complete estimation. Surface that override rather than letting it change
    # the estimates silently.
    if (acov_listwise && use == "pairwise.complete.obs" && anyNA(x)) {
      cli::cli_inform(
        c("i" = "An asymptotic covariance requires complete cases; incomplete rows were dropped (listwise), overriding {.code use = \"pairwise.complete.obs\"}."),
        class = "efa_acov_listwise"
      )
    }

    if (cor_method == "fiml") {

      # Two-stage / full-information ML: EM-estimate the saturated multivariate-normal mean
      # and covariance from the raw data with missing values (assuming the data are missing
      # at random; Yuan, Marshall, & Bentler, 2002; Little & Rubin, 2002), then analyse the
      # standardised covariance. Every case contributes, so `use` does not apply and N is the
      # EM case count (resolved below). Routed apart from stats::cor(), which would instead
      # delete cases. The EM moments are carried forward in the returned list for downstream use.
      em <- .fiml_em_moments(x, max_iter = fiml_max_iter, tol = fiml_tol)
      R <- stats::cov2cor(em$sigma)
      # Mirror the other raw-data paths: label R by the input's column names (NULL for an
      # unnamed matrix) rather than the V1..Vp fallback the EM synthesises, so the analysis
      # matrix is named consistently across cor_method.
      dimnames(R) <- list(colnames(x), colnames(x))
      fiml <- list(mu = em$mu, sigma = em$sigma, logl = em$logl,
                   converged = em$converged, iter = em$iter,
                   n_patterns = em$n_patterns, n = em$n)
      # The saturated mean/covariance and saturated log-likelihood are carried forward so the
      # downstream fit indices can form the FIML likelihood-ratio chi-square (model and
      # independence-baseline log-likelihoods over the missingness patterns) rather than a naive
      # complete-case discrepancy on R; see .gof_fiml_chisq(). The Stage-1 diagnostics travel
      # with them so the caller can report that the analysed matrix is the EM's last iterate
      # rather than the converged FIML estimate -- the warning alone leaves no record on the
      # fitted object, and callers that suppress per-fit warnings would surface nothing at all.

    } else if (.is_poly_cor(cor_method)) {

      # Polychoric/tetrachoric correlations come from the raw ordinal data.
      # .polychoric() handles missing data per pair, so reproduce the other `use`
      # policies on the data first to keep `use` meaning the same as for
      # stats::cor(): missing values abort under "all.obs"/"everything" (which do
      # not delete and so would yield an uncomputable result, matching how
      # stats::cor() errors or returns NAs there), listwise deletion applies under
      # "complete.obs"/"na.or.complete", and "pairwise.complete.obs" maps to the
      # per-pair handling. The wrapper owns the ordinal validation and classed
      # conditions (including the NA-pair abort); nearest_pd = FALSE so a
      # non-positive-definite result is smoothed once below by the shared
      # psych::cor.smooth() step rather than projected here as well. The returned
      # matrix already carries the variable names, so the colnames step that the
      # Pearson branch needs is not repeated.
      if (use %in% c("all.obs", "everything") && anyNA(x)) {
        cli::cli_abort(
          c("The correlation matrix could not be computed from the raw data because of missing values.",
            "i" = "Adjust {.arg use} (e.g. {.val pairwise.complete.obs}) or supply data with fewer missing values."),
          class = "efa_cor_na", call = error_call)
      }
      if (.is_listwise_use(use)) {
        x <- x[stats::complete.cases(x), , drop = FALSE]
        # Listwise deletion can remove every row; report that as missing data
        # rather than letting .polychoric() raise a misleading "constant variable".
        if (nrow(x) == 0L) {
          cli::cli_abort(
            c("The correlation matrix could not be computed from the raw data because of missing values.",
              "i" = "No complete cases remain under {.code use = {.val {use}}}; try {.val pairwise.complete.obs}."),
            class = "efa_cor_na", call = error_call)
        }
      }
      poly <- .polychoric(x, nearest_pd = FALSE,
                          binary_only = cor_method == "tetra",
                          acov = acov, error_call = error_call)
      R <- poly$R
      # When an asymptotic covariance is requested (the DWLS path), .polychoric()
      # estimates the matrix and the covariance on the listwise-complete rows, so the
      # per-element variances pair with the returned matrix; turn them into the symmetric
      # DWLS weight matrix. The weights are the asymptotic variances of the observed
      # (un-projected) polychoric correlations, as in lavaan, and are kept on that scale even
      # if R is subsequently projected to positive definiteness below.
      if (acov != "none") {
        # Screen the asymptotic variances once, before either consumer uses them, so a
        # degenerate pair is reported identically whether it reaches DWLS or the sandwich.
        # The diagonal is the same quantity at both levels (diag(Gamma) == the "diag" vector),
        # and .polychoric() has already labelled it by variable pair (diag() keeps those names).
        .warn_acov_degenerate(if (acov == "diag") poly$acov else diag(poly$acov))
      }
      if (acov == "diag") {
        weights <- .poly_weight_matrix(poly$acov, ncol(R))
      } else if (acov == "full") {
        # poly$acov is the full p(p-1)/2 x p(p-1)/2 asymptotic covariance: the sandwich meat.
        # Only DWLS needs the per-element inverse-variance weights, and its diagonal supplies
        # them (diag(Gamma) == the acov = "diag" vector); the ML / ULS sandwich path uses Gamma
        # alone, so the weights are built only when the estimator is DWLS.
        Gamma <- poly$acov
        if (dwls) weights <- .poly_weight_matrix(diag(Gamma), ncol(R))
      }

    } else {

      # A full ADF asymptotic covariance (the continuous sandwich meat) must describe the same
      # cases the correlation matrix was computed from -- a sandwich covariance is only valid for
      # the estimator that produced the estimates -- so reduce to the listwise-complete rows
      # first, mirroring the polychoric acov path. Without an ACOV the matrix stays pairwise for
      # data efficiency. (Only the full level is requested on the Pearson path; the diagonal DWLS
      # weights are an ordinal construct, so "diag" is left to the polychoric branch above.)
      if (acov == "full") {
        x <- x[stats::complete.cases(x), , drop = FALSE]
        # Recorded so the diagnosis below can say that it judged the rows actually used: a
        # column can be constant among the listwise-complete rows and vary in the full data.
        listwise_reduced <- TRUE
        if (nrow(x) < 2L) {
          cli::cli_abort(
            c("An asymptotic-distribution-free covariance needs at least two listwise-complete observations.",
              "x" = "{nrow(x)} row{?s} {?is/are} complete across all variables.",
              "i" = "Supply complete data, or impute the missing values, before requesting an {.arg acov}."),
            class = "efa_cor_no_complete_cases", call = error_call)
        }
      }

      # Missing values can make stats::cor() either throw a hard base error (e.g.
      # use = "all.obs", or "complete.obs" with no complete cases) or return NAs
      # (e.g. a column with no complete pairs under the chosen `use`). Catch both
      # instead of failing with an opaque base error here or later in
      # solve()/eigen(). A try-error without any NAs in the data has another cause
      # (e.g. a non-numeric or zero-variance column), so report that separately
      # rather than blaming missing values.
      #
      # stats::cor() raises exactly one warning of its own, "the standard deviation is zero"
      # for a constant column, and it always leaves NAs in the result -- so it is always
      # followed by the classed abort below, which names that column. Muffle it rather than
      # letting an untranslated base condition print ahead of, and in competition with, the
      # package's own diagnosis.
      R <- withCallingHandlers(
        try(stats::cor(x, use = use, method = cor_method), silent = TRUE),
        warning = function(w) invokeRestart("muffleWarning"))
      if (inherits(R, "try-error") || anyNA(R)) {
        if (anyNA(x)) {
          cli::cli_abort(
            c("The correlation matrix could not be computed from the raw data because of missing values.",
              "i" = "Adjust {.arg use} (e.g. {.val pairwise.complete.obs}) or supply data with fewer missing values."),
            class = "efa_cor_na", call = error_call)
        }
        # Name the columns rather than leaving the user to find them in a wide data set. The
        # three causes have different remedies, so they are reported separately, and any of
        # them can apply to several columns at once (a whole data set read in as character,
        # say). Usability is the test stats::cor() itself applies, is.numeric() OR
        # is.logical(), so dichotomous items stored as TRUE/FALSE -- which cor() correlates
        # without complaint -- are not misreported as non-numeric.
        #
        # Both remaining tests are written to yield FALSE rather than NA on a column they
        # cannot judge: sd() is NA for fewer than two values and NaN for a column carrying an
        # infinity, and a single NA in `const` would make `any(const)` NA and abort with an
        # unclassed base error in place of the classed condition below. Such a column falls
        # through to the generic bullet instead, which is what this branch reported before.
        nms <- colnames(x)
        if (is.null(nms)) nms <- paste0("V", seq_len(ncol(x)))   # unnamed input: by position
        usable <- if (is.data.frame(x)) {
          vapply(x, function(v) is.numeric(v) || is.logical(v), logical(1))
        } else {
          rep(is.numeric(x) || is.logical(x), ncol(x))
        }
        nonfin <- logical(ncol(x))
        const <- logical(ncol(x))
        for (j in which(usable)) {
          col_j <- if (is.data.frame(x)) x[[j]] else x[, j]
          if (!all(is.finite(col_j))) {
            nonfin[j] <- TRUE
          } else {
            # The exact test stats::cor() applies internally: sd == 0, not a tolerance.
            const[j] <- isTRUE(stats::sd(col_j) == 0)
          }
        }

        # The name lists are capped through the same helper as every other condition in this
        # file that enumerates affected variables, so the two cannot render differently. The
        # cli::qty() is repeated after the capped list because the empty `rest` string is
        # itself a substitution and would otherwise set the pluralisation quantity to one.
        bullets <- "The correlation matrix could not be computed from the raw data."
        if (any(!usable)) {
          n_type <- sum(!usable)
          cap_type <- .cap_label_list(nms[!usable])
          bullets <- c(
            bullets,
            "x" = "{cli::qty(n_type)}Column{?s} {.val {cap_type$shown}}{cap_type$rest}{cli::qty(n_type)} {?is/are} not numeric.",
            "i" = "Ordinal items stored as factors or character strings are correlated with {.code cor_method = \"poly\"} ({.val tetra} for binary items); anything else has to be converted or dropped."
          )
        }
        if (any(nonfin)) {
          n_inf <- sum(nonfin)
          cap_inf <- .cap_label_list(nms[nonfin])
          bullets <- c(
            bullets,
            "x" = "{cli::qty(n_inf)}Column{?s} {.val {cap_inf$shown}}{cap_inf$rest}{cli::qty(n_inf)} {?contains/contain} infinite values.",
            "i" = "An infinite value has no correlation with anything; check for a division by zero or an out-of-range missing-value code."
          )
        }
        if (any(const)) {
          n_const <- sum(const)
          cap_const <- .cap_label_list(nms[const])
          # Under a listwise-reducing `acov` the variance was judged on the complete cases
          # only, so a named column can still vary in the data the user supplied; say which
          # rows the verdict describes rather than advising a drop that may be wrong.
          const_remedy <- if (listwise_reduced) {
            "The variance was computed on the listwise-complete rows an {.arg acov} needs, which can be far fewer than the data supplied; either supply more complete cases or drop the constant {cli::qty(n_const)}column{?s}."
          } else {
            "A constant variable correlates with nothing; drop {cli::qty(n_const)}{?it/them} before the analysis."
          }
          bullets <- c(
            bullets,
            "x" = "{cli::qty(n_const)}Column{?s} {.val {cap_const$shown}}{cap_const$rest}{cli::qty(n_const)} {?has/have} zero variance.",
            "i" = const_remedy
          )
        }
        # Derived from what was actually reported, so adding a fourth cause above cannot leave
        # the fallback firing alongside a named one.
        if (length(bullets) == 1L) {
          too_few_rows <- if (nrow(x) < 2L) {
            " Correlations also need at least two observations."
          } else {
            ""
          }
          bullets <- c(
            bullets,
            "i" = paste0("Check that all columns are numeric and have non-zero variance.",
                         too_few_rows)
          )
        }
        cli::cli_abort(bullets, class = "efa_cor_uncomputable", call = error_call)
      }
      colnames(R) <- colnames(x)

      # Full ADF covariance of the off-diagonal correlations (Browne, 1984): the meat of the
      # continuous robust/sandwich SEs, on the variance scale Var(rho-hat) and in utils::combn()
      # order, computed on the same listwise-complete rows as R.
      if (acov == "full") {
        Gamma <- .adf_gamma(x)
      }

    }

    if (N_policy != "none") {
      # Under listwise deletion stats::cor() drops incomplete rows, so N must be
      # the number of complete cases rather than the raw row count. Requesting a
      # polychoric ACOV (the DWLS path) likewise reduces to the listwise-complete rows
      # inside .polychoric(), regardless of `use`, so N follows the complete cases there
      # too. Under pairwise-complete estimation with missing data, N is the nominal row
      # count: each correlation is estimated from its own pairwise-complete subset, which
      # may be smaller, so fit statistics and analytic standard errors that scale with N
      # treat the data as if complete.
      N <- if (cor_method == "fiml") {
        # FIML analyses every case that carries at least one observed value (the EM case
        # count), independent of `use`.
        em$n
      } else if (.is_listwise_use(use) || acov_listwise) {
        sum(stats::complete.cases(x))
      } else {
        nrow(x)
      }
    }

  }

  # One symmetric eigendecomposition answers both checks below, so no separate p x p inverse
  # is built and discarded as a singularity test. For a symmetric matrix
  # min(|lambda|) / max(|lambda|) is exactly the reciprocal 2-norm condition number, and
  # min(lambda) < .Machine$double.eps is the positive-definiteness test. max(|lambda|) >= 1
  # here because R always carries a unit diagonal, so the ratio is well defined.
  #
  # The singularity threshold is p * .Machine$double.eps, the standard numerical-rank
  # tolerance (as in MASS::ginv and LAPACK's own rank estimates), NOT a bare epsilon. A
  # backward-stable symmetric eigensolver returns eigenvalues with absolute error of order
  # eps * ||R||_2 (Golub & Van Loan, 2013, Matrix Computations, 4th ed., sec. 8.1; Higham,
  # 2002, Accuracy and Stability of Numerical Algorithms, 2nd ed., ch. 5), so for an exactly
  # rank-deficient matrix the computed min(|lambda|) IS that rounding error: testing it
  # against eps alone compares a quantity against the noise floor that produced it, and the
  # verdict becomes a coin flip. Measured on 495 rank-deficient correlation matrices
  # (p = 4 to 40; a duplicated item, an item equal to the sum of two others, two redundant
  # items), a bare epsilon missed 124 of them while p * eps missed none.
  #
  # This threshold refuses everything solve() refused, which is what the check did before:
  # kappa_1 <= p * kappa_2 for a symmetric matrix (the norm-equivalence bounds in Golub & Van
  # Loan, sec. 2.3), so solve()'s gate (reciprocal 1-norm
  # condition number below eps) implies min(|lambda|) / max(|lambda|) < p * eps. It also
  # refuses a narrow band just above it -- 47 of the 600 matrices in that sweep -- which is
  # the intended direction: a correlation matrix with a condition number above 1 / (p * eps)
  # carries no usable inverse. Well-conditioned data is nowhere near the threshold; every
  # correlation matrix shipped with the package sits at 1e13 to 1e15 times eps.
  #
  # The two checks are NOT interchangeable and both are kept, in this order: an indefinite
  # matrix can be perfectly well conditioned (it is smoothed, not refused -- only the
  # eigenvalue test sees it, which is why the singularity test takes the absolute value and
  # the smoothing test does not), and a positive definite matrix can be numerically singular
  # (it is refused before any smoothing is attempted). The smoothing threshold stays at a bare
  # eps because it is pinned to psych::cor.smooth()'s own trigger, not to a rank tolerance.
  ev <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
  # The number of variables sets the rank tolerance below and is reported by the abort that
  # tolerance triggers, so both read the one value.
  p_R <- ncol(R)

  # Check if the correlation matrix is invertible, if it is not, stop with message
  if (check_singular &&
      min(abs(ev)) / max(abs(ev)) < p_R * .Machine$double.eps) {
    # The offending matrix travels with the condition, so a caller that wants to name
    # the culprits (efa_screen() reports the perfectly correlated pairs) does not have
    # to recompute a correlation matrix this function has already built.
    #
    # Two different causes reach this point, and they have different remedies. A sample no
    # larger than the number of variables makes a full-rank correlation matrix arithmetically
    # impossible -- the matrix is taken about the sample means, so its rank is at most N - 1 --
    # and no variable is at fault; naming the two numbers is the whole of the diagnosis. The
    # bound is N <= p rather than N < p because the centering costs that one degree of freedom.
    # For any larger N the deficiency is in the variables themselves, which is what
    # efa_screen() reads off this same matrix: it names the perfectly correlated pairs, and
    # says so when the redundancy is a combination of several variables instead.
    # isTRUE() rather than a test on N: this abort is reached from every entry point, and
    # several of them do not constrain N at all (N_policy = "none" leaves it NA). An absent,
    # missing, or multi-valued N must fall through to the general hint rather than turn a
    # classed condition into an unclassed comparison error.
    hint <- if (isTRUE(N <= p_R)) {
      "With N = {N} and {p_R} variables the correlation matrix cannot have full rank.
       N must be larger than the number of variables."
    } else {
      "Use {.fn efa_screen}, which reports the variables responsible."
    }
    cli::cli_abort(c("The correlation matrix is singular; {singular_tail}.", "i" = hint),
                   class = "efa_cor_singular", R = R, call = error_call)
  }

  # Check if correlation matrix is positive definite, if it is not, either stop
  # (SPSS type) or smooth the matrix and surface a single classed warning.
  # The threshold matches psych::cor.smooth()'s own trigger (smallest eigenvalue
  # below .Machine$double.eps), so a matrix that has already been smoothed - whose
  # eigenvalue floor sits well above this - is not re-flagged on each downstream
  # call (e.g. inside HULL -> PARALLEL -> EFA).
  if (min(ev) < .Machine$double.eps) {

    if (posdef_abort) {
      cli::cli_abort(
        "The correlation matrix is not positive definite; no further analyses are performed.",
        class = "efa_cor_not_posdef", call = error_call)
    }

    # Muffle only cor.smooth's routine "smoothing was done" note (re-surfaced
    # below as a classed warning); let its serious "eigen values are NA" failure
    # warning propagate so an unrepairable matrix is not silently accepted.
    R <- withCallingHandlers(
      psych::cor.smooth(R),
      warning = function(w) {
        if (grepl("smoothing was done", conditionMessage(w), fixed = TRUE)) {
          invokeRestart("muffleWarning")
        }
      }
    )

    # cor.smooth() decides for itself whether to act, and it decides from an eigendecomposition
    # that asks for the eigenvectors, whereas the eigenvalues above are values-only. LAPACK runs
    # a different driver for each, so their smallest eigenvalues differ by the rounding of the
    # decomposition itself (of order eps * ||R||). For a matrix that is singular to working
    # precision that rounding IS the eigenvalue, so the two verdicts disagree on a substantial
    # share of such matrices -- measured at two in five over a sweep of exactly rank-deficient
    # correlation matrices of order 5 to 40 -- and cor.smooth() then returns the matrix
    # untouched, leaving a matrix reported as smoothed still indefinite. Which side of the
    # disagreement a build lands on is a property of its LAPACK, so verify the result here and
    # project it on the same eigenvalue floor when it was left alone.
    if (min(eigen(R, symmetric = TRUE, only.values = TRUE)$values) <
        .Machine$double.eps) {
      R <- .project_cor_pd(R)
    }

    # The remedy is named as the eigenvalue floor rather than as cor.smooth(), because on the
    # branch above it is .project_cor_pd() that applies it.
    cli::cli_warn(
      c("The correlation matrix was not positive definite; it has been smoothed.",
        "i" = "Its smallest eigenvalues were floored, as in {.fun psych::cor.smooth}; inspect the results carefully."),
      class = "efa_cor_smoothed"
    )

  }

  list(R = R, N = N, is_cormat = is_cormat, weights = weights, Gamma = Gamma,
       fiml = fiml)
}

# Evaluate `expr` with the "computing correlations from the raw data" note limited to its
# first occurrence. An entry point that calls efa_fit() once per group or per imputation
# gets one note from every inner fit, so one fact about one input is stated as many times
# as there are fits. The note keeps the classed condition of the first fit, so a caller can
# still catch it. The pre-fit diagnostics do not report the computation at all and suppress
# it with `inform_from_data = FALSE` instead.
.cor_note_once <- function(expr) {
  seen <- FALSE
  withCallingHandlers(
    expr,
    efa_cor_from_data = function(cnd) {
      if (seen) invokeRestart("muffleMessage")
      seen <<- TRUE
    }
  )
}

# Polychoric/tetrachoric correlations describe the observed data only. Criteria that cannot
# honour them reject the request with one shared classed condition rather than returning a
# silently invalid answer. `why` states which property fails: by default that the criterion
# compares the data against a separately generated continuous reference (efa_cd(),
# efa_parallel(), efa_nest(), efa_hull()), and for efa_smt() that its sequential tests rest on
# a normal-theory ML chi-square (and the RMSEA and AIC derived from it), which is not valid for
# such correlations. It is a cli template evaluated here, so it can interpolate `fn`.
#
# `fn` is the name the message shows, and every caller passes its own. Only a direct call
# reaches this message -- efa_retain() screens the same combination off the registry's
# `poly_ok` flag and reports it under the criterion id instead -- so naming the function the
# user actually called is what identifies the problem, including through a superseded wrapper,
# which forwards to the successor named here.
.reject_poly_reference <- function(cor_method, fn,
                                   why = "{.fn {fn}} compares the data against a reference computed with a continuous (Pearson) correlation.",
                                   error_call = rlang::caller_env()) {
  if (.is_poly_cor(cor_method)) {
    cli::cli_abort(
      c("{.val {cor_method}} correlations are not supported by {.fn {fn}}.",
        "x" = why,
        # Kendall's tau is deliberately absent: it is not on the Pearson scale the factor
        # model parameterises (for bivariate-normal data tau = (2/pi) asin(rho)), so
        # offering it as the remedy to a user who correctly reached for a polychoric
        # correlation would steer an ordinal analysis onto the most attenuated metric of
        # the three. Spearman is kept: it stays a correlation coefficient.
        "i" = "Use {.val pearson} or {.val spearman}."),
      class = "efa_cor_method_unsupported", call = error_call)
  }
  invisible(NULL)
}

# Unit tests for the shared correlation-input preparer: correlation-matrix
# detection, the N-handling policy, and the singular / non-positive-definite
# branches.

set.seed(7)                                            # keep the drawn columns deterministic
x <- rnorm(10)                                         # two varying columns for the
y <- rnorm(10)                                         # constant-column fixture below
cormat <- test_models$baseline$cormat                  # well-behaved correlation matrix

test_that("a correlation matrix is detected and returned unchanged", {
  prep <- .prepare_cor_input(cormat, N = 500)

  expect_named(prep, c("R", "N", "is_cormat", "weights", "Gamma", "fiml"))
  expect_true(prep$is_cormat)
  expect_equal(prep$R, cormat)
  expect_equal(prep$N, 500)
  expect_null(prep$fiml)                                # only the FIML path populates it

  # a clean correlation matrix triggers no message, warning, or error
  expect_silent(.prepare_cor_input(cormat, N = 500))
})

test_that("an exact -1 correlation is within the accepted range", {
  R <- matrix(c(1, -1, -1, 1), 2, 2)
  expect_true(.is_cormat(R))
  expect_silent(.assert_cor_input(R))
})

test_that("a correlation matrix supplied as a data frame is detected and coerced", {
  # read.csv(file, row.names = 1) is how a published correlation table usually reaches R, and
  # it yields a data frame. It has to be classified as a correlation matrix -- diag() on a data
  # frame is a hard base error, so the classifier coerces first -- and the preparer has to hand
  # a real matrix downstream, since determinant(), the eigendecompositions and the compiled
  # estimators all reject a data frame.
  cormat_df <- as.data.frame(cormat)

  expect_true(.is_cormat(cormat_df))
  expect_silent(.assert_cor_input(cormat_df))

  prep_df <- .prepare_cor_input(cormat_df, N = 500)
  expect_true(prep_df$is_cormat)
  expect_true(is.matrix(prep_df$R))
  expect_equal(prep_df$R, cormat)
  expect_equal(prep_df$N, 500)
  expect_silent(.prepare_cor_input(cormat_df, N = 500))       # not mistaken for raw data

  # every entry point that accepts a correlation matrix returns what the matrix route returns
  expect_equal(efa_kmo(cormat_df), efa_kmo(cormat))
  expect_equal(efa_bartlett(cormat_df, N = 500), efa_bartlett(cormat, N = 500))
  expect_equal(efa_map(cormat_df), efa_map(cormat))
  expect_equal(efa_kgc(cormat_df), efa_kgc(cormat))
  expect_equal(efa_scree(cormat_df), efa_scree(cormat))
  expect_equal(efa_ekc(cormat_df, N = 500), efa_ekc(cormat, N = 500))
  expect_equal(efa_fit(cormat_df, n_factors = 3, N = 500),
               efa_fit(cormat, n_factors = 3, N = 500))
  expect_equal(efa_screen(cormat_df, N = 500), efa_screen(cormat, N = 500))
  expect_equal(efa_retain(cormat_df, N = 500, criteria = c("KGC", "MAP")),
               efa_retain(cormat, N = 500, criteria = c("KGC", "MAP")))
  # including the superseded wrappers, which forward to the same preparer
  expect_equal(EFA(cormat_df, n_factors = 3, N = 500), EFA(cormat, n_factors = 3, N = 500))
  expect_equal(KMO(cormat_df), KMO(cormat))

  # a non-numeric square data frame is still not a correlation matrix, and a covariance data
  # frame is still rejected as one
  expect_false(.is_cormat(data.frame(a = factor(c("x", "y")), b = c(1, 2))))
  expect_error(.assert_cor_input(as.data.frame(diag(c(2, 3, 4)))),
               class = "efa_input_is_covmat")
})

test_that("a 1x1 input is not classified as a correlation matrix", {
  # .is_covmat() has always required at least two columns; .is_cormat() applies the same guard,
  # so a 1x1 input takes the raw-data route and is rejected there rather than being accepted
  # and failing somewhere downstream.
  expect_false(.is_cormat(matrix(1)))
  expect_false(.is_covmat(matrix(1)))
})

test_that("a non-symmetric correlation matrix is named rather than reported as singular", {
  # Neither classifier accepts an asymmetric matrix, so without this branch it reaches
  # stats::cor() as if its p columns were p cases and is reported as singular -- sending the
  # user after collinearity that does not exist.
  Rl <- cormat
  Rl[upper.tri(Rl)] <- 0                       # a lower triangle transcribed on its own
  expect_error(.assert_cor_input(Rl), class = "efa_input_not_symmetric")

  Ra <- cormat
  Ra[1, 2] <- Ra[1, 2] + 1e-7                  # values that do not mirror to machine precision
  expect_error(.assert_cor_input(Ra), class = "efa_input_not_symmetric")

  # an empty (NA) upper triangle is the same transcription error, and must take the same route
  # rather than the missing-values one
  Rna <- cormat
  Rna[upper.tri(Rna)] <- NA
  expect_error(.assert_cor_input(Rna), class = "efa_input_not_symmetric")

  # a data frame is classified identically
  expect_error(.assert_cor_input(as.data.frame(Rl)), class = "efa_input_not_symmetric")

  # the rejection reaches every entry point, including the raw-data-only route, which points
  # at the observations instead of at mirroring
  expect_error(suppressMessages(efa_kmo(Rl)), class = "efa_input_not_symmetric")
  expect_error(suppressMessages(efa_fit(Rl, n_factors = 3, N = 500)),
               class = "efa_input_not_symmetric")
  expect_error(suppressMessages(efa_screen(Rl)), class = "efa_input_not_symmetric")
  expect_error(suppressMessages(efa_map(Rl)), class = "efa_input_not_symmetric")
  expect_error(suppressMessages(efa_cd(Rl)), class = "efa_input_not_symmetric")

  # the suggested repair is accepted and recovers the original matrix
  Rl[upper.tri(Rl)] <- t(Rl)[upper.tri(Rl)]
  expect_silent(.assert_cor_input(Rl))
  expect_equal(Rl, cormat)

  # The hint has to name the triangle that is actually empty. Mirroring the lower triangle
  # onto the upper one is the repair for a lower-triangle transcription and DESTROYS an
  # upper-triangle one -- it overwrites every entered correlation with the empty triangle and
  # leaves the identity matrix, which passes every later check and analyses silently.
  Ru <- cormat
  Ru[lower.tri(Ru)] <- 0                       # only the UPPER triangle was entered
  expect_error(.assert_cor_input(Ru), class = "efa_input_not_symmetric")
  Ru[lower.tri(Ru)] <- t(Ru)[lower.tri(Ru)]    # the repair this input needs
  expect_silent(.assert_cor_input(Ru))
  expect_equal(Ru, cormat)
  # the other direction really would have wiped it
  Rwipe <- cormat
  Rwipe[lower.tri(Rwipe)] <- 0
  Rwipe[upper.tri(Rwipe)] <- t(Rwipe)[upper.tri(Rwipe)]
  expect_equal(unname(Rwipe), diag(ncol(cormat)))

  # square raw data is untouched: it is asymmetric but not unit-diagonal
  set.seed(3)
  expect_silent(.assert_cor_input(matrix(stats::runif(9, -1, 1), 3)))
  expect_silent(.assert_cor_input(GRiPS_raw))
})

test_that("raw data is converted to a correlation matrix", {
  expect_message(prep <- .prepare_cor_input(GRiPS_raw), class = "efa_cor_from_data")

  expect_false(prep$is_cormat)
  expect_equal(prep$R,
               stats::cor(GRiPS_raw, use = "pairwise.complete.obs",
                          method = "pearson"))
  expect_equal(colnames(prep$R), colnames(GRiPS_raw))
  expect_equal(prep$N, nrow(GRiPS_raw))

  # inform_from_data = FALSE silences the "computed from raw data" message
  expect_no_message(.prepare_cor_input(GRiPS_raw, inform_from_data = FALSE))
})

test_that("N_policy governs how N is handled", {
  # required: a correlation matrix without N aborts, but a supplied N is kept
  expect_error(.prepare_cor_input(cormat, N_policy = "required"),
               class = "efa_n_required")
  prep_req <- .prepare_cor_input(cormat, N = 200, N_policy = "required")
  expect_equal(prep_req$N, 200)

  # none: N is passed through untouched and the both-N-and-raw warning is muted
  expect_no_warning(
    prep_none <- suppressMessages(
      .prepare_cor_input(GRiPS_raw, N = 999, N_policy = "none")))
  expect_equal(prep_none$N, 999)

  # optional (default): supplying N with raw data warns and N is taken from data
  expect_warning(
    prep_opt <- suppressMessages(.prepare_cor_input(GRiPS_raw, N = 999)),
    class = "efa_n_from_data")
  expect_equal(prep_opt$N, nrow(GRiPS_raw))
})

test_that("listwise-deletion modes set N to the number of complete cases", {
  # raw data with two incomplete rows: under "complete.obs"/"na.or.complete"
  # stats::cor() drops them listwise, so N must be the complete-case count.
  set.seed(11)
  dat_miss <- matrix(rnorm(30 * 3), ncol = 3)
  dat_miss[2, 1] <- NA
  dat_miss[7, 3] <- NA
  n_complete <- sum(stats::complete.cases(dat_miss))

  for (u in c("complete.obs", "na.or.complete")) {
    prep <- suppressMessages(.prepare_cor_input(dat_miss, use = u))
    expect_equal(prep$N, n_complete)
    expect_lt(prep$N, nrow(dat_miss))
  }

  # other modes keep the raw row count
  prep_pw <- suppressMessages(
    .prepare_cor_input(dat_miss, use = "pairwise.complete.obs"))
  expect_equal(prep_pw$N, nrow(dat_miss))
})

test_that("a non-positive-definite matrix is smoothed, or aborts under posdef_abort", {
  # default: the matrix is smoothed to positive-definiteness with a classed warning
  expect_warning(prep_sm <- .prepare_cor_input(cor_nposdef, N = 10),
                 class = "efa_cor_smoothed")
  expect_true(prep_sm$is_cormat)
  # Positive definite with a margin, not merely positive: the smallest eigenvalue of a smoothed
  # matrix sits on the projection's own floor (about 1e-10), so asserting a floor of 1e-12
  # states the guarantee the warning makes. Asserting only "> 0" would instead be satisfied by
  # an unsmoothed matrix whose smallest eigenvalue happens to be a positive rounding artefact.
  expect_gt(min(eigen(prep_sm$R, symmetric = TRUE, only.values = TRUE)$values), 1e-12)

  # posdef_abort = TRUE turns the same input into an error instead
  expect_error(.prepare_cor_input(cor_nposdef, N = 10, posdef_abort = TRUE),
               class = "efa_cor_not_posdef")

  # an already-smoothed matrix is not re-flagged on a subsequent pass (the
  # non-positive-definite check uses psych::cor.smooth()'s own trigger), and it
  # is returned unchanged
  smoothed <- prep_sm$R
  expect_no_warning(prep_again <- .prepare_cor_input(smoothed, N = 10))
  expect_equal(prep_again$R, smoothed)
})

test_that("the positive-definite projection floors the spectrum and keeps the correlation form", {
  # The fallback for a matrix psych::cor.smooth() declines to act on. It has to return a
  # correlation matrix -- unit diagonal, symmetric, dimnames intact -- whose smallest
  # eigenvalue is clear of the boundary rather than sitting at the rounding floor.
  R_pd <- .project_cor_pd(cor_nposdef)
  expect_gt(min(eigen(R_pd, symmetric = TRUE, only.values = TRUE)$values), 1e-12)
  expect_equal(diag(R_pd), rep(1, ncol(cor_nposdef)))
  expect_true(isSymmetric(R_pd))

  # the eigendecomposition drops dimnames, so they are restored explicitly
  named <- cor_nposdef
  dimnames(named) <- list(letters[1:3], letters[1:3])
  expect_equal(dimnames(.project_cor_pd(named)), dimnames(named))

  # an already positive definite matrix passes through: no eigenvalue is floored, so the
  # spectrum is unchanged and the rebuild returns the input up to rounding
  expect_equal(.project_cor_pd(cormat), cormat, tolerance = 1e-8)
})

test_that("a matrix psych::cor.smooth() leaves untouched is projected anyway", {
  # cor.smooth() re-decides whether to act on its own eigendecomposition, which asks for the
  # eigenvectors where the gate that calls it reads a values-only one; LAPACK runs a different
  # driver for each, so on a matrix that is singular to working precision the two verdicts
  # disagree and cor.smooth() returns the matrix untouched. Which side of that a build lands
  # on cannot be arranged from the data, so the branch is driven by standing in for
  # cor.smooth() with the no-op it becomes when it declines. The matrix must still come back
  # positive definite, under the same classed warning.
  local_mocked_bindings(cor.smooth = function(x, ...) x, .package = "psych")

  expect_warning(prep <- .prepare_cor_input(cor_nposdef, N = 10),
                 class = "efa_cor_smoothed")
  expect_gt(min(eigen(prep$R, symmetric = TRUE, only.values = TRUE)$values), 1e-12)
  expect_equal(diag(prep$R), rep(1, ncol(cor_nposdef)))
})

test_that("NA in the raw-data correlation matrix aborts with a classed error", {
  # under use = "everything" a missing value leaves NAs in the computed
  # correlations; this must be a clear classed error, not an opaque base crash
  dat_na <- matrix(c(1, 2, 3, 4, 5, 6, 7, 8,
                     8, 6, 7, 5, 3, 4, 1, 2,
                     2, 5, 1, 8, 3, 7, 4, NA), ncol = 3)
  expect_error(suppressMessages(.prepare_cor_input(dat_na, use = "everything")),
               class = "efa_cor_na")
  # under use = "all.obs" stats::cor() throws a hard base error before any NA
  # check; this must route to the same classed abort, not an unclassed crash
  expect_error(suppressMessages(.prepare_cor_input(dat_na, use = "all.obs")),
               class = "efa_cor_na")
})

test_that("non-numeric raw data aborts with a distinct classed error", {
  # A character column makes stats::cor() fail for a reason other than NAs (the
  # data here has none); report that as efa_cor_uncomputable rather than blaming
  # missing values.
  dat_chr <- data.frame(a = c("x", "y", "z", "w"), b = c(1, 2, 3, 4),
                        d = c(4, 3, 2, 1))
  expect_error(suppressMessages(.prepare_cor_input(dat_chr)),
               class = "efa_cor_uncomputable")

  # A SQUARE data frame reaches .is_cormat(), where a factor column would otherwise
  # dispatch to Ops.factor and fail with an unclassed base error; it must take the same
  # classed route as any other non-numeric raw data.
  dat_fct <- data.frame(a = factor(c("x", "y", "z")), b = c(1, 2, 3), d = c(3, 1, 2))
  expect_false(.is_cormat(dat_fct))
  expect_error(suppressMessages(.prepare_cor_input(dat_fct)),
               class = "efa_cor_uncomputable")

  # A constant column takes the same classed route. stats::cor() warns about it in base R's
  # own words, which are neither classed nor translated; that warning must not escape ahead
  # of the abort that diagnoses the same column.
  dat_const <- cbind(x, y, const = 1)
  expect_no_warning(
    expect_error(suppressMessages(.prepare_cor_input(dat_const)),
                 class = "efa_cor_uncomputable"))
  # every correlation method routes through the same abort
  for (m in c("pearson", "spearman", "kendall")) {
    expect_no_warning(
      expect_error(suppressMessages(.prepare_cor_input(dat_const, cor_method = m)),
                   class = "efa_cor_uncomputable"))
  }

  # A column the constancy test cannot judge must not turn the classed abort into an unclassed
  # base error: sd() is NaN for a column carrying an infinity and NA for fewer than two
  # values, and a single NA in that test would make the branch selection itself fail.
  expect_error(
    suppressMessages(.prepare_cor_input(cbind(a = c(1, 2, 3, 4), b = c(Inf, 1, 2, 3),
                                              d = c(4, 3, 2, 1)))),
    class = "efa_cor_uncomputable")
  expect_error(suppressMessages(.prepare_cor_input(matrix(c(1, 2, 3), nrow = 1))),
               class = "efa_cor_uncomputable")
  expect_error(suppressMessages(.prepare_cor_input(matrix(numeric(0), ncol = 3))),
               class = "efa_cor_uncomputable")
  # the 1x1 input the ncol < 2 guard sends down this route is rejected here, as claimed
  expect_error(suppressMessages(.prepare_cor_input(matrix(1))),
               class = "efa_cor_uncomputable")

  # Logical columns are usable: stats::cor() accepts is.numeric() OR is.logical(), so
  # dichotomous TRUE/FALSE items must not be blamed as non-numeric while the column that
  # actually broke the correlation goes unnamed.
  # (the wording is pinned by the snapshots below, which show only the constant column named)
  dat_lgl <- data.frame(a = c(TRUE, FALSE, TRUE, FALSE), b = rep(1, 4), d = c(4, 1, 3, 2))
  expect_error(suppressMessages(.prepare_cor_input(dat_lgl)),
               class = "efa_cor_uncomputable")
  # and an all-logical data set correlates without reaching this branch at all
  dat_lgl2 <- data.frame(a = c(TRUE, FALSE, TRUE, FALSE), b = c(TRUE, TRUE, FALSE, FALSE))
  expect_no_error(suppressMessages(.prepare_cor_input(dat_lgl2)))
})

test_that("the uncomputable-correlation abort names the offending columns", {
  local_reproducible_output()

  # Which column is at fault, and what to do about it, is the whole content of this message:
  # a factor or character column is what an ordinal item read from a file looks like, and
  # cor_method = "poly" is the fix rather than a reason to drop it.
  dat_fct <- data.frame(item_1 = factor(c("x", "y", "z")), item_2 = c(1, 2, 3),
                        item_3 = c(3, 1, 2))
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_fct, inform_from_data = FALSE))

  # a constant column has a different remedy, and both causes can be present at once
  dat_const <- data.frame(item_1 = c(1, 2, 3, 4), item_2 = rep(2, 4), item_3 = c(4, 1, 3, 2))
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_const, inform_from_data = FALSE))

  dat_both <- data.frame(item_1 = letters[1:4], item_2 = rep(2, 4), item_3 = c(4, 1, 3, 2))
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_both, inform_from_data = FALSE))

  # an unnamed matrix falls back to positional labels
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(cbind(c(1, 2, 3, 4), rep(2, 4), c(4, 1, 3, 2)),
                                     inform_from_data = FALSE))

  # a logical column is usable, so only the constant column is named and the polychoric hint
  # is not offered for a column that needs nothing done to it
  dat_lgl <- data.frame(a = c(TRUE, FALSE, TRUE, FALSE), b = rep(1, 4), d = c(4, 1, 3, 2))
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_lgl, inform_from_data = FALSE))

  # an infinite value is its own cause, with its own remedy
  dat_inf <- data.frame(item_1 = c(1, 2, 3, 4), item_2 = c(Inf, 1, 2, 3),
                        item_3 = c(4, 3, 2, 1))
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_inf, inform_from_data = FALSE))

  # fewer than two observations leaves no column judgeable, so the generic bullet stands
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(matrix(c(1, 2, 3), nrow = 1), inform_from_data = FALSE))

  # An asymptotic covariance reduces the data to its listwise-complete rows before the
  # correlation is attempted, so a column can be constant there and vary in the data the user
  # supplied. The verdict has to say which rows it describes rather than advising a drop.
  set.seed(9)
  dat_lw <- cbind(a = stats::rnorm(60), b = stats::rnorm(60), c = stats::rnorm(60))
  keep <- c(3, 11, 25, 40)
  dat_lw[-keep, "a"] <- NA                    # only these rows are complete
  dat_lw[keep, "c"] <- 7                      # constant among the complete rows only
  expect_gt(stats::sd(dat_lw[, "c"]), 0)      # but not in the data as supplied
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_lw, acov = "full", inform_from_data = FALSE))

  # the plural forms of both bullets, and the five-name cap on the list
  dat_many <- data.frame(a = letters[1:4], b = LETTERS[1:4],
                         c1 = rep(2, 4), c2 = rep(3, 4), ok = 1:4)
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_many, inform_from_data = FALSE))

  dat_wide <- as.data.frame(matrix(rep(1, 35), nrow = 5))
  dat_wide$ok <- 1:5
  expect_snapshot(error = TRUE,
                  .prepare_cor_input(dat_wide, inform_from_data = FALSE))
})

test_that("the non-symmetric abort names the triangle that is actually empty", {
  local_reproducible_output()

  # Which triangle to mirror is read off the data: printing the lower-triangle repair for an
  # upper-triangle transcription would tell the user to overwrite everything they entered.
  R_lower <- cormat[1:5, 1:5]
  R_lower[upper.tri(R_lower)] <- 0
  expect_snapshot(error = TRUE, .assert_cor_input(R_lower))

  R_upper <- cormat[1:5, 1:5]
  R_upper[lower.tri(R_upper)] <- 0
  expect_snapshot(error = TRUE, .assert_cor_input(R_upper))

  # an empty triangle entered as NA rather than 0 is the same transcription, and the message
  # must not claim the missing cells were checked and found to be in range
  R_na <- cormat[1:5, 1:5]
  R_na[upper.tri(R_na)] <- NA
  expect_snapshot(error = TRUE, .assert_cor_input(R_na))

  # both triangles filled but disagreeing: neither is authoritative, so no mirroring is offered
  R_mismatch <- cormat[1:5, 1:5]
  R_mismatch[1, 2] <- R_mismatch[1, 2] + 1e-7
  expect_snapshot(error = TRUE, .assert_cor_input(R_mismatch))

  # neither triangle filled in (a table pre-allocated but never transcribed): mirroring an
  # empty triangle onto the other would silently produce the identity matrix, so it must not
  # be suggested
  R_none <- cormat[1:5, 1:5]
  R_none[upper.tri(R_none)] <- NA
  R_none[lower.tri(R_none)] <- 0
  expect_snapshot(error = TRUE, .assert_cor_input(R_none))

  # the raw-data-only route points at the observations instead
  expect_snapshot(error = TRUE, .assert_cor_input(R_lower, raw_only = TRUE))
})

test_that("a singular matrix aborts unless the check is disabled", {
  # LAPACK returns the eigenvalues of a rank-deficient matrix with absolute error of order
  # eps * ||R||, so the smallest computed eigenvalue IS that rounding error and a rank
  # tolerance of a bare eps decides by coin flip. The shared fixture is sized so the ratio
  # sits well below ncol(R) * eps; it is the archetypal EFA singularity -- a composite
  # entered alongside its components -- and must be refused rather than inverted downstream.
  expect_equal(qr(sing_cor, tol = 1e-10)$rank, 9L)         # rank deficient by one
  expect_error(.prepare_cor_input(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_error(suppressMessages(.prepare_cor_input(sing_raw)),
               class = "efa_cor_singular")
  expect_error(suppressMessages(efa_fit(sing_cor, n_factors = 2, N = sing_N)),
               class = "efa_cor_singular")

  # The two causes take different bullets. Here the rank deficiency is in the variables, so
  # the message hands the reader on to the function that names them.
  expect_snapshot(error = TRUE, .prepare_cor_input(sing_cor, N = sing_N))

  # A sample no larger than the number of variables is the other cause, and no variable is at
  # fault: the matrix is taken about the sample means, so its rank is at most N - 1 and the
  # numbers are the whole diagnosis. The boundary is N <= p, so the equality case takes this
  # branch too -- with N = p, efa_screen() would report an exact linear combination that is an
  # artefact of the sample size rather than a property of the variables.
  expect_snapshot(error = TRUE,
                  suppressMessages(.prepare_cor_input(GRiPS_raw[1:4, ])))
  expect_snapshot(error = TRUE,
                  suppressMessages(.prepare_cor_input(GRiPS_raw[1:8, ])))

  # check_singular = FALSE skips the abort. Whether the smoothing branch is entered at all is a
  # property of the build rather than of the data: for an exactly rank-deficient matrix the
  # smallest computed eigenvalue IS the eigensolver's rounding, so its sign -- and with it the
  # verdict of a gate at .Machine$double.eps -- differs between LAPACK implementations. What
  # holds on every build is that nothing is refused and the matrix that comes back is a
  # correlation matrix that is not materially indefinite; the positive definiteness that
  # smoothing guarantees is pinned above on `cor_nposdef`, whose smallest eigenvalue is -0.41
  # and which therefore takes the branch everywhere.
  # The two outcomes are asserted as the disjunction they are, rather than as a bound loose
  # enough to admit both: either the branch was not entered and the matrix comes back exactly
  # as it went in, or it was and the result is positive definite with the projection's margin.
  prep_ns <- suppressWarnings(
    .prepare_cor_input(sing_cor, N = sing_N, check_singular = FALSE))
  expect_true(
    identical(prep_ns$R, sing_cor) ||
      min(eigen(prep_ns$R, symmetric = TRUE, only.values = TRUE)$values) > 1e-12)

  # and a well-conditioned matrix is nowhere near the threshold
  ev_ok <- eigen(cormat, symmetric = TRUE, only.values = TRUE)$values
  expect_gt(min(abs(ev_ok)) / max(abs(ev_ok)),
            1e6 * ncol(cormat) * .Machine$double.eps)
  expect_silent(.prepare_cor_input(cormat, N = 500))
})

test_that("a covariance matrix is rejected with a pointer to cov2cor()", {
  # A covariance matrix is square and symmetric but carries variances on its diagonal, so
  # it is neither a correlation matrix nor raw data. Passed on, it would be fed to
  # stats::cor() as if its columns were cases -- singular by construction and a wrong
  # diagnosis -- so the shared input assertion rejects it up front.
  sds <- seq(1, 5, length.out = ncol(cormat))
  S <- diag(sds) %*% cormat %*% diag(sds)               # a valid covariance matrix

  expect_error(.assert_cor_input(S), class = "efa_input_is_covmat")
  # a diagonal (uncorrelated) covariance is caught too: variances on the diagonal
  expect_error(.assert_cor_input(diag(c(2, 3, 4))), class = "efa_input_is_covmat")

  # an INDEFINITE covariance matrix -- as arises from pairwise deletion or a table
  # transcribed at few decimals -- is a covariance matrix too: PSD is deliberately not
  # required (an indefinite correlation matrix is likewise accepted and smoothed), so
  # exactly the problematic inputs do not fall through to the raw-data misdiagnosis
  S_ind <- S
  S_ind[1, 2] <- S_ind[2, 1] <- sqrt(S[1, 1] * S[2, 2]) * 1.05
  expect_lt(min(eigen(S_ind, symmetric = TRUE, only.values = TRUE)$values), 0)
  expect_error(.assert_cor_input(S_ind), class = "efa_input_is_covmat")

  # a non-finite entry must not abort inside the classifier (eigen() is no longer called);
  # the matrix is simply not classified as a covariance matrix
  S_inf <- S
  S_inf[1, 2] <- S_inf[2, 1] <- Inf
  expect_false(.is_covmat(S_inf))
  expect_silent(.assert_cor_input(S_inf))

  # a zero diagonal (a distance-like matrix) is not a covariance matrix
  expect_false(.is_covmat(as.matrix(stats::dist(t(cormat[1:4, 1:4])))))

  # the rejection reaches every entry point that accepts a correlation matrix or raw data
  expect_error(suppressMessages(efa_fit(S, n_factors = 2, N = 200)),
               class = "efa_input_is_covmat")
  expect_error(suppressMessages(efa_kmo(S)), class = "efa_input_is_covmat")
  expect_error(suppressMessages(efa_screen(S)), class = "efa_input_is_covmat")
  expect_error(suppressMessages(efa_bartlett(S, N = 200)), class = "efa_input_is_covmat")
  # including the retention criteria, which reach the shared preparer by their own routes
  expect_error(suppressMessages(efa_parallel(S, N = 200)), class = "efa_input_is_covmat")
  expect_error(suppressMessages(efa_map(S)), class = "efa_input_is_covmat")

  # a raw-data-only function points at the observations, not at cov2cor() (which would only
  # produce a correlation matrix it rejects in turn)
  expect_error(suppressMessages(efa_cd(S)), class = "efa_input_is_covmat")

  # standardising first is accepted; a genuine correlation matrix and raw data still pass
  expect_silent(.assert_cor_input(stats::cov2cor(S)))
  expect_silent(.assert_cor_input(cormat))
  expect_silent(.assert_cor_input(GRiPS_raw))
})

## Diagonally-Weighted Least Squares Estimation of Factor Loadings
## Thin fitter: returns the raw DWLS results; shared post-processing happens in
## .finalize_fit() / .estimate_model(). Unlike ULS, DWLS optimises the loadings
## directly (weighted off-diagonal least squares) so the per-element weights shape the
## solution; see .fit_dwls() and the DwlsFunctor in src/estimate.cpp.
.DWLS <- function(x, n_factors, weights) {

  if (is.null(weights)) {
    cli::cli_abort(
      c("DWLS estimation requires a per-element weight matrix.",
        "i" = "Weights are produced by {.fn .prepare_cor_input} with {.code acov = \"diag\"} on raw ordinal data."),
      class = "efa_dwls_no_weights"
    )
  }

  # Get correlation matrix entered or created in EFA
  R <- x

  dwls <- .fit_dwls(R, n_factors, weights)

  L <- dwls$loadings
  orig_R <- R
  h2 <- rowSums(L^2)            # diag(L L'), without forming the full p x p product
  diag(R) <- h2

  # raw fit, finalized by .estimate_model(). `Fm` is the weighted off-diagonal objective the
  # C++ backend already minimised. DWLS optimises the loadings with no lower bound on the
  # uniquenesses, so the boundary-uniqueness Heywood heuristic in .finalize_fit() (which only
  # applies to the box-constrained ML/ULS optimisers) does not apply here; psi is left NULL
  # so a Heywood case is flagged only by a communality at or above 1, which the unconstrained
  # optimiser can still produce.
  list(
    L = L,
    h2 = h2,
    psi = NULL,
    Fm = dwls$Fm,
    iter = dwls$iter,
    convergence = dwls$convergence,
    orig_R = orig_R,
    R_final = R
  )
}

# Obtain the DWLS fit. The C++ backend warm-starts with the weighted-ULS solution (the
# eigen extraction that minimises the weighted objective over the uniquenesses) and then
# polishes the free loadings to the diagonally weighted least squares optimum.
.fit_dwls <- function(R, n_fac, weights) {

  dwls <- .fit_dwls_cpp(R, n_fac, weights)

  list(loadings = dwls$loadings, Fm = dwls$Fm, iter = dwls$iter,
       convergence = dwls$convergence)
}

#' Print a loading matrix
#'
#' @details
#' The method prints a loading matrix in a compact, console-oriented table.
#' Loadings with absolute value greater than or equal to `cutoff` are emphasized,
#' smaller loadings are de-emphasized, and Heywood-relevant communality/
#' uniqueness values are marked when `h2` is supplied. Long variable names can
#' be truncated, abbreviated, or printed in full. If the matrix has many factor
#' columns, the table is split into column blocks so that the output remains
#' readable in narrower consoles.
#'
#' If `h2` is named and `x` has row names, `h2` is matched to the row names of
#' `x` before any optional row sorting is applied. If `x` has no row names, a
#' named `h2` vector is used in the supplied order.
#'
#' @param x a loading matrix of class `efa_loadings` (or the legacy `LOADINGS`).
#' @param cutoff numeric. The value at or above which loadings are emphasized;
#'  default is .3.
#' @param digits numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
#'  to round the loadings to (default is 3).
#' @param max_name_length numeric. The maximum length of the variable names to
#'  display. Everything beyond this will be cut from the right unless
#'  `name_style = "abbreviate"` or `name_style = "full"` is used. Cutting never
#'  leaves two variables sharing a row label: if it would (as for items with a
#'  long common prefix), the names are abbreviated instead, and numbered if
#'  needed.
#' @param h2 numeric. Vector of communalities to print. If named and `x`
#'  has row names, names are used to align communalities to rows.
#' @param color logical. Whether to apply console styling using \pkg{cli}.
#'  Default is `TRUE`.
#' @param name_style character. How to shorten variable names longer than
#'  `max_name_length`. `"truncate"` cuts names from the right,
#'  `"abbreviate"` uses [base::abbreviate()], and
#'  `"full"` prints full names.
#' @param max_factor_name_length numeric or `NULL`. Optional maximum length
#'  of factor names. If `NULL`, factor names are not shortened.
#' @param max_factors_per_block numeric or `NULL`. Maximum number of factor
#'  columns to print per block. If `NULL`, the number is chosen from the
#'  console width.
#' @param sort_loadings character. Optional row sorting. `"none"` preserves
#'  the input order, `"primary"` groups rows by the factor with the largest
#'  absolute loading, and `"clustered"` additionally sorts within each
#'  factor by the size of the primary loading.
#' @param legend logical. Whether to append a short explanation of the styling.
#'  Default is `FALSE` for standalone loading matrices. The legend is only printed
#'  when the styling it describes is actually rendered (`color = TRUE` and a
#'  colour-capable console); in plain output it is omitted and this argument has
#'  no effect.
#' @param ... additional arguments passed to print or format
#'
#' @returns `print()` returns its argument `x` invisibly; it is
#'   `cat(format(x, ...), sep = "\n")` followed by a blank line for console
#'   spacing. `format()` returns a character vector with the table lines (styled
#'   to the active console theme; plain when colours are disabled).
#'
#' @method print efa_loadings
#' @export
#'
#' @examples
#' EFAtools_PAF <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                         estimator = "PAF", rotation = "promax")
#' EFAtools_PAF
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(EFAtools_PAF$rot_loadings))
#'
print.efa_loadings <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  # a trailing blank line, so a loading table printed at the console is set off from
  # whatever follows it
  cat("\n")

  invisible(x)
}

#' @rdname print.efa_loadings
#' @method format efa_loadings
#' @export
format.efa_loadings <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                           h2 = NULL, color = TRUE,
                           name_style = c("truncate", "abbreviate", "full"),
                           max_factor_name_length = NULL,
                           max_factors_per_block = NULL,
                           sort_loadings = c("none", "primary", "clustered"),
                           legend = FALSE, ...) {

  name_style <- .match_arg_ci(name_style)
  sort_loadings <- .match_arg_ci(sort_loadings)

  .validate_loadings_print_args(
    x = x,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = h2,
    color = color,
    max_factor_name_length = max_factor_name_length,
    max_factors_per_block = max_factors_per_block,
    legend = legend
  )

  spec <- .loadings_print_spec(
    x = x,
    h2 = h2,
    max_name_length = max_name_length,
    name_style = name_style,
    max_factor_name_length = max_factor_name_length,
    sort_loadings = sort_loadings
  )

  cli::cli_format_method({
    # Through .efa_emit_lines(), not cli_verbatim() directly: a table too wide for the
    # console is rendered as stacked column blocks separated by a blank line, and
    # cli_verbatim() drops interior empty strings, which would run the blocks together.
    .efa_emit_lines(.efa_format_matrix(
      values = spec$values,
      row_labels = spec$var_names,
      col_labels = spec$factor_names,
      col_roles = spec$col_type,
      cutoff = cutoff,
      digits = digits,
      color = color,
      max_factors_per_block = max_factors_per_block
    ))

    # The legend only describes console styling, so it is emitted only when that styling
    # is actually rendered. In plain output the loadings themselves carry the information,
    # and a legend naming marks the reader cannot see would be misleading.
    if (isTRUE(legend) && .efa_styling_visible(color)) {
      cli::cli_verbatim("")
      cli::cli_verbatim(.efa_loadings_legend(cutoff, spec$has_h2, digits, color))
    }
  })
}

.validate_loadings_print_args <- function(x, cutoff, digits, max_name_length,
                                          h2, color, max_factor_name_length,
                                          max_factors_per_block, legend) {
  if (!is.matrix(x) || !is.numeric(x)) {
    cli::cli_abort("{.arg x} must be a numeric matrix.",
                   class = "efa_print_invalid_x")
  }

  if (nrow(x) < 1L || ncol(x) < 1L) {
    cli::cli_abort("{.arg x} must have at least one row and one column.",
                   class = "efa_print_invalid_x")
  }

  .efa_check_nonneg_number(cutoff, "cutoff", "efa_print_invalid_cutoff")
  .efa_check_nonneg_int(digits, "digits", "efa_print_invalid_digits")
  .efa_check_pos_int(max_name_length, "max_name_length",
                     "efa_print_invalid_max_name_length")
  .efa_check_opt_pos_int(max_factor_name_length, "max_factor_name_length",
                         "efa_print_invalid_max_factor_name_length")
  .efa_check_opt_pos_int(max_factors_per_block, "max_factors_per_block",
                         "efa_print_invalid_max_factors_per_block")
  .efa_check_flag(color, "color", "efa_print_invalid_color")
  .efa_check_flag(legend, "legend", "efa_print_invalid_legend")

  if (!is.null(h2)) {
    if (!is.numeric(h2) || length(h2) != nrow(x)) {
      cli::cli_abort("{.arg h2} must be a numeric vector with one value per row of {.arg x}.",
                     class = "efa_print_invalid_h2")
    }
  }

  invisible(TRUE)
}

.loadings_print_spec <- function(x, h2, max_name_length,
                                 name_style, max_factor_name_length,
                                 sort_loadings) {
  x <- as.matrix(x)
  n_factors <- ncol(x)

  factor_names <- colnames(x)
  if (is.null(factor_names)) {
    factor_names <- paste0("F", seq_len(n_factors))
  }

  original_var_names <- rownames(x)
  has_row_names <- !is.null(original_var_names)
  var_names <- original_var_names
  if (!has_row_names) {
    var_names <- paste0("V", seq_len(nrow(x)))
  }

  has_h2 <- !is.null(h2)
  if (has_h2) {
    h2 <- .align_loadings_h2(
      h2 = h2,
      var_names = var_names,
      has_row_names = has_row_names
    )
  }

  # Shorten before reordering, so a row label is a property of the variable rather than of
  # where it happens to sit: the numbering that resolves a collision is assigned in vector
  # order, and deriving it from the sorted order would label the same item differently under
  # different `sort_loadings`.
  display_var_names <- .shorten_loadings_names(
    var_names,
    max_length = max_name_length,
    name_style = name_style
  )

  row_order <- .efa_loading_row_order(x, sort_loadings)
  x <- x[row_order, , drop = FALSE]
  display_var_names <- display_var_names[row_order]
  if (has_h2) {
    h2 <- h2[row_order]
  }

  col_type <- rep("loading", n_factors)
  values <- x

  if (has_h2) {
    values <- cbind(values, "h2" = h2, "u2" = 1 - h2)
    factor_names <- c(factor_names, "h2", "u2")
    col_type <- c(col_type, "h2", "u2")
  }

  display_factor_names <- .shorten_loadings_factor_names(
    factor_names,
    max_length = max_factor_name_length
  )

  list(
    values = values,
    var_names = display_var_names,
    factor_names = display_factor_names,
    col_type = col_type,
    has_h2 = has_h2
  )
}

.align_loadings_h2 <- function(h2, var_names, has_row_names = TRUE) {
  # Only align by name when the loading matrix itself has row names. Generated
  # fallback names (V1, V2, ...) should not be used to reorder a named h2 vector.
  h2_names <- names(h2)
  has_complete_h2_names <- !is.null(h2_names) && all(nzchar(h2_names))

  if (isTRUE(has_row_names) && has_complete_h2_names) {
    if (!all(var_names %in% h2_names)) {
      cli::cli_abort(
        "If {.arg h2} is named and {.arg x} has row names, {.code names(h2)} must include all row names of {.arg x}.",
        class = "efa_print_invalid_h2"
      )
    }
    h2 <- h2[var_names]
  }

  as.numeric(h2)
}

# Shorten variable names to `max_length` for a loading table. Cutting from the right can map
# distinct items onto one row label -- "premeditation_1" through "premeditation_11" all become
# "premeditat" -- leaving rows the reader cannot tell apart, with nothing on screen to signal
# it. So a truncation that collides falls back to abbreviation and, if that still collides, to
# numbered labels: a row label is never shared by two different variables. Names the caller
# supplied as duplicates are left as they are; those collisions are not the shortening's doing.
.shorten_loadings_names <- function(x, max_length, name_style) {
  x <- as.character(x)

  if (identical(name_style, "full")) {
    return(x)
  }

  if (all(nchar(x) <= max_length)) {
    return(x)
  }

  if (identical(name_style, "abbreviate")) {
    return(.abbreviate_loadings_names(x, max_length))
  }

  out <- substr(x, 1L, max_length)
  if (anyDuplicated(out) < 1L || anyDuplicated(x) > 0L) {
    return(out)
  }

  # `abbreviate()` warns on non-ASCII input, and a table repairing its own row labels must
  # not raise a warning for doing so; such names go straight to the numbering below. The
  # test is `abbreviate()`'s own, so the two agree on what counts as non-ASCII.
  if (all(nchar(x, type = "bytes") == nchar(x, type = "chars"))) {
    abbreviated <- .abbreviate_loadings_names(x, max_length)
    if (anyDuplicated(abbreviated) < 1L) {
      return(abbreviated)
    }
    out <- abbreviated
  }

  .disambiguate_loadings_names(out, max_length)
}

# `abbreviate()` with `strict = TRUE` never exceeds `minlength`, so the result still respects
# the requested width; `method = "both.sides"` also drops characters from a shared prefix,
# which is what separates names that differ only near their end.
.abbreviate_loadings_names <- function(x, max_length) {
  as.character(abbreviate(
    x,
    minlength = max_length,
    strict = TRUE,
    method = "both.sides"
  ))
}

# Last-resort disambiguation of labels that are still duplicated: the numeric suffix
# `make.unique()` appends is budgeted out of `max_length` by trimming the label it is appended
# to, so a disambiguated label keeps the requested width instead of growing past it. Trimming
# can in principle re-collide, and a width too narrow to hold a suffix cannot be budgeted at
# all; both fall back to the untrimmed form, which `make.unique()` guarantees is unique. An
# over-wide label is still readable, an ambiguous one is not.
.disambiguate_loadings_names <- function(x, max_length) {
  suffixed <- make.unique(x, sep = "~")
  suffix <- substr(suffixed, nchar(x) + 1L, nchar(suffixed))

  keep <- max_length - nchar(suffix)
  if (any(keep < 1L)) {
    return(suffixed)
  }

  out <- paste0(substr(x, 1L, keep), suffix)
  if (anyDuplicated(out) > 0L) suffixed else out
}

.shorten_loadings_factor_names <- function(x, max_length = NULL) {
  x <- as.character(x)

  if (is.null(max_length) || all(nchar(x) <= max_length)) {
    return(x)
  }

  substr(x, 1L, max_length)
}

# Build the cli legend lines describing the loading-matrix styling. Styling is dropped when
# `color = FALSE` or colours are off, so the lines embed cleanly in plain output.
.efa_loadings_legend <- function(cutoff, has_h2, digits, color = TRUE) {
  cutoff_str <- .efa_num(cutoff, digits = digits, pad = FALSE)

  bold_word <- if (isTRUE(color)) cli::style_bold("bold") else "bold"
  grey_word <- if (isTRUE(color)) cli::col_grey("grey") else "grey"

  lines <- c(
    "Legend:",
    paste0("  ", bold_word, " = |loading| >= ", cutoff_str),
    paste0("  ", grey_word, " = below cutoff")
  )

  if (isTRUE(has_h2)) {
    red_word <- if (isTRUE(color)) {
      cli::style_bold(cli::col_red("red h2/u2"))
    } else {
      "red h2/u2"
    }
    lines <- c(lines, paste0("  ", red_word, " = Heywood-relevant value"))
  }

  lines
}

#' Four test models used in Grieder and Steiner (2022)
#'
#' Correlation matrices created from simulated data from four of the
#' [`population_models`] cases, each with strong factor intercorrelations.
#' These are used in Grieder & Steiner (2022) to compare the psych and SPSS
#' implementations in this package with the actual implementations of the programs.
#' For details on the cases, see [`population_models`].
#'
#' @format A list of 4 lists "baseline", "case_1a", "case_6b", and "case_11b", each with the following elements.
#' \describe{
#'   \item{cormat}{(matrix) - The correlation matrix of the simulated data.}
#'   \item{n_factors}{(numeric) - The true number of factors.}
#'   \item{N}{(numeric) - The sample size of the generated data.}
#'  }
#' @source Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
#' of implementations of principal axis factoring and promax rotation in R and SPSS.
#' Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x
"test_models"

#' Print EFA object
#'
#' Print Method showing a summarized output of the \link{EFA} function.
#'
#' @details
#' The method is shared by single-imputation `EFA` objects and pooled
#' `EFA_POOLED` objects. It prints, in order, a compact model header, optional
#' diagnostics, the loading table, optional confidence-interval tables, variance
#' accounted for, model fit, bootstrap notes, optional multiple-imputation
#' uncertainty diagnostics, and residual diagnostics.
#'
#' For `EFA_POOLED` objects, the header and diagnostics report the number of
#' imputations and the alignment/pooling settings stored in the object. Loading
#' and factor-correlation confidence intervals are printed only when `boot.CI`
#' is present. Pooled intervals created by `EFA_POOLED()` are labelled as
#' bootstrap/MI intervals.
#'
#' `ci_filter` controls which loading intervals are shown. `salient` reports
#' intervals for loadings whose absolute point estimate is at least `cutoff`;
#' `nonzero` reports intervals excluding zero; and `all` reports every finite
#' interval.
#'
#' @param x list. An object of class EFA to be printed
#' @param cutoff numeric. Passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.
#' The number above which to print loadings in bold. Default is .3.
#' @param digits numeric. Passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}
#' Number of digits to round the loadings to (default is 3).
#' @param max_name_length numeric. Passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.
#' The maximum length of the variable names to display. Everything beyond this
#' will be cut from the right.
#' @param ci character. Whether to print confidence intervals for loadings and
#' factor intercorrelations, if available. \code{"auto"} and \code{"separate"}
#' print separate CI sections when CIs were computed; \code{"none"} suppresses
#' these sections. Default is \code{"auto"}.
#' @param ci_filter character. Which loading CIs to print. \code{"salient"}
#' prints CIs for loadings with absolute values greater than or equal to
#' \code{cutoff}; \code{"all"} prints all loading CIs; \code{"nonzero"} prints
#' loading CIs that exclude zero. Default is \code{"salient"}.
#' @param details character. Amount of information to print. \code{"standard"}
#' prints the regular output, \code{"compact"} suppresses longer diagnostic
#' sections, and \code{"full"} additionally prints available optional sections.
#' @param diagnostics logical. Whether to print model and simple-structure
#' diagnostics. Default is \code{TRUE}.
#' @param diagnostics_top_n numeric. Maximum number of item-level diagnostic
#' entries to print per diagnostic type.
#' @param residual_cutoff numeric. Absolute residual cutoff used in the residual
#' diagnostics section. Default is .1.
#' @param residual_top_n numeric. Maximum number of residuals to print. Use
#' \code{Inf} to print all residuals above \code{residual_cutoff}.
#' @param show_structure logical. Whether to print the structure matrix for
#' oblique solutions when available. Default is \code{FALSE}.
#' @param sort_loadings character. Optional row sorting for loading tables.
#' See \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.
#' @param show_loading_legend logical. Whether to print a compact legend for
#' loading-table styling. Default is \code{TRUE}.
#' @param cross_loading_cutoff numeric. Cutoff for counting cross-loadings in
#' the diagnostics. Defaults to \code{cutoff}.
#' @param min_primary_gap numeric. Minimum desired absolute difference between
#' the largest and second-largest absolute loading of an item. Used only for
#' descriptive diagnostics.
#' @param min_salient_per_factor numeric. Minimum number of salient indicators
#' per factor used in the diagnostics. Default is 3.
#' @param max_factors_per_block numeric or \code{NULL}. Maximum number of factor
#' columns per loading-table block. If \code{NULL}, this is chosen from the
#' console width.
#' @param show_mi_diagnostics logical or \code{NULL}. Whether to print a compact
#' MI uncertainty summary for pooled EFAs when available. \code{NULL} prints it
#' only for \code{details = "full"}.
#' @param ... Further arguments passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.
#'
#' @export
#'
#' @method print EFA
#'
#' @examples
#' mod <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
#'            method = "PAF", rotation = "promax")
#' mod
#'
#' # With SEs and CIs, needs raw-data
#' mod <- EFA(GRiPS_raw, n_factors = 1, method = "ML", se = "np-boot")
#' mod
#' # with additional infos
#' print(mod, details = "full")
#' # omit some diagnostic parts
#' print(mod, details = "compact")
#'
print.EFA <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                      ci = c("auto", "none", "separate"),
                      ci_filter = c("salient", "all", "nonzero"),
                      details = c("standard", "compact", "full"),
                      diagnostics = TRUE,
                      diagnostics_top_n = 10,
                      residual_cutoff = .1,
                      residual_top_n = 10,
                      show_structure = FALSE,
                      sort_loadings = c("none", "primary", "clustered"),
                      show_loading_legend = TRUE,
                      cross_loading_cutoff = cutoff,
                      min_primary_gap = .20,
                      min_salient_per_factor = 3,
                      max_factors_per_block = NULL,
                      show_mi_diagnostics = NULL, ...) {
  ci <- match.arg(ci)
  ci_filter <- match.arg(ci_filter)
  details <- match.arg(details)
  sort_loadings <- match.arg(sort_loadings)

  .print_EFA(x,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    details = details,
    diagnostics = diagnostics,
    diagnostics_top_n = diagnostics_top_n,
    residual_cutoff = residual_cutoff,
    residual_top_n = residual_top_n,
    show_structure = show_structure,
    sort_loadings = sort_loadings,
    show_loading_legend = show_loading_legend,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    max_factors_per_block = max_factors_per_block,
    show_mi_diagnostics = show_mi_diagnostics,
    ...
  )

  invisible(x)
}

#' @rdname print.EFA
#' @export
#' @method print EFA_POOLED
print.EFA_POOLED <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                             ci = c("auto", "none", "separate"),
                             ci_filter = c("salient", "all", "nonzero"),
                             details = c("standard", "compact", "full"),
                             diagnostics = TRUE,
                             diagnostics_top_n = 10,
                             residual_cutoff = .1,
                             residual_top_n = 10,
                             show_structure = FALSE,
                             sort_loadings = c("none", "primary", "clustered"),
                             show_loading_legend = TRUE,
                             cross_loading_cutoff = cutoff,
                             min_primary_gap = .20,
                             min_salient_per_factor = 3,
                             max_factors_per_block = NULL,
                             show_mi_diagnostics = NULL, ...) {
  ci <- match.arg(ci)
  ci_filter <- match.arg(ci_filter)
  details <- match.arg(details)
  sort_loadings <- match.arg(sort_loadings)

  .print_EFA(x,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    details = details,
    diagnostics = diagnostics,
    diagnostics_top_n = diagnostics_top_n,
    residual_cutoff = residual_cutoff,
    residual_top_n = residual_top_n,
    show_structure = show_structure,
    sort_loadings = sort_loadings,
    show_loading_legend = show_loading_legend,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    max_factors_per_block = max_factors_per_block,
    show_mi_diagnostics = show_mi_diagnostics,
    ...
  )

  invisible(x)
}

#' @rdname print.EFA
#' @export
#' @method format EFA
format.EFA <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                       ci = c("auto", "none", "separate"),
                       ci_filter = c("salient", "all", "nonzero"),
                       details = c("standard", "compact", "full"),
                       diagnostics = TRUE,
                       diagnostics_top_n = 10,
                       residual_cutoff = .1,
                       residual_top_n = 10,
                       show_structure = FALSE,
                       sort_loadings = c("none", "primary", "clustered"),
                       show_loading_legend = TRUE,
                       cross_loading_cutoff = cutoff,
                       min_primary_gap = .20,
                       min_salient_per_factor = 3,
                       max_factors_per_block = NULL,
                       show_mi_diagnostics = NULL, ...) {
  ci <- match.arg(ci)
  ci_filter <- match.arg(ci_filter)
  details <- match.arg(details)
  sort_loadings <- match.arg(sort_loadings)

  utils::capture.output(
    .print_EFA(x,
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      ci = ci,
      ci_filter = ci_filter,
      details = details,
      diagnostics = diagnostics,
      diagnostics_top_n = diagnostics_top_n,
      residual_cutoff = residual_cutoff,
      residual_top_n = residual_top_n,
      show_structure = show_structure,
      sort_loadings = sort_loadings,
      show_loading_legend = show_loading_legend,
      cross_loading_cutoff = cross_loading_cutoff,
      min_primary_gap = min_primary_gap,
      min_salient_per_factor = min_salient_per_factor,
      max_factors_per_block = max_factors_per_block,
      show_mi_diagnostics = show_mi_diagnostics,
      ...
    )
  )
}

#' @rdname print.EFA
#' @export
#' @method format EFA_POOLED
format.EFA_POOLED <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                              ci = c("auto", "none", "separate"),
                              ci_filter = c("salient", "all", "nonzero"),
                              details = c("standard", "compact", "full"),
                              diagnostics = TRUE,
                              diagnostics_top_n = 10,
                              residual_cutoff = .1,
                              residual_top_n = 10,
                              show_structure = FALSE,
                              sort_loadings = c("none", "primary", "clustered"),
                              show_loading_legend = TRUE,
                              cross_loading_cutoff = cutoff,
                              min_primary_gap = .20,
                              min_salient_per_factor = 3,
                              max_factors_per_block = NULL,
                              show_mi_diagnostics = NULL, ...) {
  ci <- match.arg(ci)
  ci_filter <- match.arg(ci_filter)
  details <- match.arg(details)
  sort_loadings <- match.arg(sort_loadings)

  utils::capture.output(
    .print_EFA(x,
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      ci = ci,
      ci_filter = ci_filter,
      details = details,
      diagnostics = diagnostics,
      diagnostics_top_n = diagnostics_top_n,
      residual_cutoff = residual_cutoff,
      residual_top_n = residual_top_n,
      show_structure = show_structure,
      sort_loadings = sort_loadings,
      show_loading_legend = show_loading_legend,
      cross_loading_cutoff = cross_loading_cutoff,
      min_primary_gap = min_primary_gap,
      min_salient_per_factor = min_salient_per_factor,
      max_factors_per_block = max_factors_per_block,
      show_mi_diagnostics = show_mi_diagnostics,
      ...
    )
  )
}

.print_EFA <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                       ci = c("auto", "none", "separate"),
                       ci_filter = c("salient", "all", "nonzero"),
                       details = c("standard", "compact", "full"),
                       diagnostics = TRUE,
                       diagnostics_top_n = 10,
                       residual_cutoff = .1,
                       residual_top_n = 10,
                       show_structure = FALSE,
                       sort_loadings = c("none", "primary", "clustered"),
                       show_loading_legend = TRUE,
                       cross_loading_cutoff = cutoff,
                       min_primary_gap = .20,
                       min_salient_per_factor = 3,
                       max_factors_per_block = NULL,
                       show_mi_diagnostics = NULL, ...) {
  ci <- match.arg(ci)
  ci_filter <- match.arg(ci_filter)
  details <- match.arg(details)
  sort_loadings <- match.arg(sort_loadings)

  .efa_validate_print_options(
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    diagnostics = diagnostics,
    diagnostics_top_n = diagnostics_top_n,
    residual_cutoff = residual_cutoff,
    residual_top_n = residual_top_n,
    show_structure = show_structure,
    show_loading_legend = show_loading_legend,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    max_factors_per_block = max_factors_per_block,
    show_mi_diagnostics = show_mi_diagnostics
  )

  spec <- .efa_print_spec(x)
  show_structure <- .efa_resolve_show_structure(show_structure, details)
  show_mi_diagnostics <- .efa_resolve_show_mi_diagnostics(
    show_mi_diagnostics,
    details = details,
    spec = spec
  )

  .print_efa_header(spec)

  if (isTRUE(diagnostics) && !identical(details, "compact")) {
    .print_efa_diagnostics_section(x, spec,
      cutoff = cutoff,
      digits = digits,
      cross_loading_cutoff = cross_loading_cutoff,
      min_primary_gap = min_primary_gap,
      min_salient_per_factor = min_salient_per_factor
    )
  }

  .print_efa_loadings_section(x, spec,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    show_structure = show_structure,
    sort_loadings = sort_loadings,
    show_loading_legend = show_loading_legend,
    max_factors_per_block = max_factors_per_block,
    ...
  )

  if (isTRUE(diagnostics) && !identical(details, "compact")) {
    .print_efa_simple_structure_section(x, spec,
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      cross_loading_cutoff = cross_loading_cutoff,
      min_primary_gap = min_primary_gap,
      min_salient_per_factor = min_salient_per_factor,
      diagnostics_top_n = diagnostics_top_n,
      sort_loadings = sort_loadings
    )
  }

  if (!identical(details, "compact")) {
    .print_efa_variances_section(x, spec)
  }

  if (!.print_efa_identification_warning(spec)) {
    return(invisible(x))
  }

  .print_efa_model_fit_section(x, spec)
  .print_efa_bootstrap_note(x, spec)

  if (isTRUE(show_mi_diagnostics)) {
    .print_efa_mi_diagnostics_section(x, spec, digits = digits)
  }

  if (!identical(details, "compact")) {
    .print_efa_residuals_section(x,
      residual_cutoff = residual_cutoff,
      residual_top_n = residual_top_n,
      digits = digits
    )
  }

  invisible(x)
}

.efa_print_spec <- function(x) {
  settings <- x$settings
  if (is.null(settings)) {
    settings <- list()
  }

  se <- settings$se
  is_pooled <- inherits(x, "EFA_POOLED") || isTRUE(settings$pooled)

  list(
    method = settings$method,
    rotation = settings$rotation,
    type = settings$type,
    N = settings$N,
    fit = x$fit_indices,
    h2 = x$h2,
    np_boot = identical(.efa_setting_text(se), "np-boot"),
    ci = .efa_ci_level(x),
    rmsea_ci_level = .efa_rmsea_ci_level(x),
    b_boot = settings$b_boot,
    max_iter = settings$max_iter,
    iter = x$iter,
    is_pooled = is_pooled,
    n_imputations = .efa_n_imputations(x),
    target_method = settings$target_method,
    align_unrotated = settings$align_unrotated,
    fit_pool_method = settings$fit_pool_method,
    component_se = settings$component_se,
    has_boot_ci = !is.null(x$boot.CI)
  )
}

.print_efa_header <- function(spec) {
  cat("\n")

  if (isTRUE(spec$is_pooled)) {
    .print_efa_pooled_header(spec)
  } else {
    cat("EFA performed with type = '", crayon::bold(spec$type),
      "', method = '", crayon::bold(spec$method),
      "', and rotation = '", crayon::bold(spec$rotation), "'.",
      sep = ""
    )
    cat("\n")
  }

  if (.efa_iteration_nonconvergence(spec)) {
    cat("\n")
    cat(crayon::red$bold(
      cli::symbol$cross,
      "Maximum number of iterations reached",
      "without convergence"
    ))
    cat("\n")
  }
}


.print_efa_pooled_header <- function(spec) {
  cat("Pooled EFA")

  n_imputations <- .efa_setting_text(spec$n_imputations)
  if (nzchar(n_imputations)) {
    cat(" across ", crayon::bold(n_imputations), " imputations", sep = "")
  }

  cat(" performed with type = '", crayon::bold(spec$type),
    "', method = '", crayon::bold(spec$method),
    "', and rotation = '", crayon::bold(spec$rotation), "'.",
    sep = ""
  )
  cat("\n")

  pooling_settings <- .efa_pooled_settings_text(spec)
  if (length(pooling_settings) > 0L) {
    cat("Pooling settings: ", paste(pooling_settings, collapse = ", "), ".", sep = "")
    cat("\n")
  }

  invisible(NULL)
}

.efa_pooled_settings_text <- function(spec) {
  out <- character(0)

  target_method <- .efa_setting_text(spec$target_method)
  if (nzchar(target_method) && !identical(spec$rotation, "none")) {
    out <- c(out, paste0("target_method = '", target_method, "'"))
  }

  align_unrotated <- .efa_setting_text(spec$align_unrotated)
  if (nzchar(align_unrotated)) {
    out <- c(out, paste0("align_unrotated = '", align_unrotated, "'"))
  }

  fit_pool_method <- .efa_setting_text(spec$fit_pool_method)
  if (nzchar(fit_pool_method)) {
    out <- c(out, paste0("fit_pool_method = '", fit_pool_method, "'"))
  }

  if (is.finite(spec$rmsea_ci_level) &&
      abs(spec$rmsea_ci_level - .90) > sqrt(.Machine$double.eps)) {
    out <- c(out, paste0("rmsea_ci_level = ", .efa_ci_level_text(spec$rmsea_ci_level)))
  }

  out
}

.print_efa_loadings_section <- function(x, spec, cutoff, digits, max_name_length,
                                        ci, ci_filter, show_structure,
                                        sort_loadings, show_loading_legend,
                                        max_factors_per_block, ...) {
  if (identical(spec$rotation, "none")) {
    .print_efa_rule("Unrotated Loadings")

    if (is.null(x$unrot_loadings)) {
      cat("No unrotated loading matrix available.\n")
      return(invisible(NULL))
    }

    print(x$unrot_loadings,
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      h2 = x$h2,
      sort_loadings = sort_loadings,
      legend = show_loading_legend,
      max_factors_per_block = max_factors_per_block,
      ...
    )
    .print_efa_heywood_warning(spec$h2)
    .print_efa_loading_ci_section(x, spec,
      component = "unrot_loadings",
      loadings = x$unrot_loadings,
      loading_label = "unrotated",
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      ci = ci,
      ci_filter = ci_filter,
      sort_loadings = sort_loadings
    )
    return(invisible(NULL))
  }

  .print_efa_rule("Rotated Loadings")

  if (is.null(x$rot_loadings)) {
    cat("No rotated loading matrix available.\n")
    return(invisible(NULL))
  }

  print(x$rot_loadings,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = x$h2,
    sort_loadings = sort_loadings,
    legend = show_loading_legend,
    max_factors_per_block = max_factors_per_block,
    ...
  )
  .print_efa_heywood_warning(spec$h2)
  .print_efa_loading_ci_section(x, spec,
    component = "rot_loadings",
    loadings = x$rot_loadings,
    loading_label = "rotated",
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    sort_loadings = sort_loadings
  )

  if (!is.null(x$Phi)) {
    phi <- as.matrix(x$Phi)
    factor_names <- .efa_factor_names(phi)

    .print_efa_rule("Factor Intercorrelations")
    cat(.get_compare_matrix(phi,
      r_red = Inf,
      n_char = 17,
      var_names = factor_names,
      factor_names = factor_names
    ))
    .print_efa_phi_ci_section(x, spec, digits, ci)
  }

  if (!is.null(x$Structure)) {
    if (isTRUE(show_structure)) {
      .print_efa_structure_section(x,
        cutoff = cutoff,
        digits = digits,
        max_name_length = max_name_length,
        sort_loadings = sort_loadings,
        max_factors_per_block = max_factors_per_block,
        ...
      )
    }
    # else {
    #   .print_efa_structure_note()
    # }
  }

  invisible(NULL)
}

.print_efa_variances_section <- function(x, spec) {
  .print_efa_rule("Variances Accounted for")

  if (identical(spec$rotation, "none")) {
    vars_accounted <- x$vars_accounted
  } else {
    vars_accounted <- x$vars_accounted_rot
  }

  if (is.null(vars_accounted)) {
    cat("No variance-accounted table available.\n")
    return(invisible(NULL))
  }

  cat(.get_compare_matrix(vars_accounted, r_red = Inf, n_char = 17))

  invisible(NULL)
}

.print_efa_identification_warning <- function(spec) {
  df <- .efa_fit_scalar(spec$fit, "df")

  if (!is.finite(df)) {
    return(TRUE)
  }

  if (df == 0) {
    cat("\n")
    cat(
      crayon::yellow$bold("!"),
      crayon::yellow(
        " The model is just identified (df = 0). Goodness of fit indices may not be interpretable."
      )
    )
    cat("\n")
    return(TRUE)
  }

  if (df < 0) {
    cat("\n")
    cat(
      crayon::yellow$bold("!"),
      crayon::yellow(
        " The model is underidentified (df < 0). No goodness of fit indices were calculated."
      )
    )
    cat("\n")
    return(FALSE)
  }

  TRUE
}

.print_efa_model_fit_section <- function(x, spec) {
  fit <- spec$fit
  fit_ci <- .efa_fit_ci_labels(x, spec)

  .print_efa_rule("Model Fit")

  if (is.null(fit) || length(fit) < 1L) {
    cat("No model-fit indices available.\n")
    return(invisible(NULL))
  }

  method <- .efa_setting_text(spec$method)
  df <- .efa_fit_scalar(fit, "df")
  chi <- .efa_fit_scalar(fit, "chi")
  p_chi <- .efa_fit_scalar(fit, "p_chi")

  if (identical(method, "PAF") || .efa_is_missing_number(spec$N) ||
      !is.finite(chi) || !is.finite(df)) {
    cat(crayon::blue("CAF", fit_ci$label, ":"),
      .efa_format_fit_value(fit, "CAF"), fit_ci$CAF, "\n",
      sep = ""
    )
    cat(crayon::blue("RMSR", fit_ci$label, ":"),
      .efa_format_fit_value(fit, "RMSR"), fit_ci$RMSR, "\n",
      sep = ""
    )
    cat(crayon::blue("df: "),
      .efa_format_fit_value(fit, "df", digits = 0, print_zero = TRUE, pad = FALSE), "\n",
      sep = ""
    )
    return(invisible(NULL))
  }

  p_text <- if (!is.finite(p_chi)) {
    " = NA"
  } else if (p_chi < .001) {
    " < .001"
  } else {
    paste0(" = ", .efa_format_number(p_chi, digits = 3, pad = FALSE))
  }

  cat(crayon::blue("\u03c7\u00b2(", sep = ""),
    .efa_format_fit_value(fit, "df", digits = 0, print_zero = TRUE, pad = FALSE),
    crayon::blue(") = ", sep = ""),
    .efa_format_fit_value(fit, "chi", digits = 2, print_zero = TRUE), ", ",
    crayon::blue(crayon::italic("p")),
    p_text,
    "\n",
    sep = ""
  )

  cat(crayon::blue("CFI", fit_ci$label, ": "),
    .efa_format_fit_value(fit, "CFI", pad = FALSE), fit_ci$CFI, "\n",
    sep = ""
  )

  rmsea_label <- paste0("RMSEA [", .efa_ci_level_text(spec$rmsea_ci_level), " CI]")
  cat(crayon::blue(rmsea_label, fit_ci$label, ": "),
    paste0(
      .efa_format_fit_value(fit, "RMSEA", pad = FALSE), " [",
      .efa_format_fit_value(fit, "RMSEA_LB", pad = FALSE), "; ",
      .efa_format_fit_value(fit, "RMSEA_UB", pad = FALSE), "]"
    ),
    fit_ci$RMSEA, "\n",
    sep = ""
  )

  cat(crayon::blue("AIC", fit_ci$label, ": "),
    .efa_format_fit_value(fit, "AIC", print_zero = TRUE), fit_ci$AIC, "\n",
    sep = ""
  )
  cat(crayon::blue("BIC", fit_ci$label, ": "),
    .efa_format_fit_value(fit, "BIC", print_zero = TRUE), fit_ci$BIC, "\n",
    sep = ""
  )
  cat(crayon::blue("CAF", fit_ci$label, ":"),
    .efa_format_fit_value(fit, "CAF"), fit_ci$CAF, "\n",
    sep = ""
  )
  cat(crayon::blue("RMSR", fit_ci$label, ":"),
    .efa_format_fit_value(fit, "RMSR"), fit_ci$RMSR, "\n",
    sep = ""
  )

  invisible(NULL)
}

.efa_fit_ci_labels <- function(x, spec) {
  out <- list(
    label = "",
    CAF = "",
    RMSR = "",
    CFI = "",
    RMSEA = "",
    AIC = "",
    BIC = ""
  )

  if (!isTRUE(spec$has_boot_ci)) {
    return(out)
  }

  fitind_ci <- .efa_get_fit_ci(x, spec)
  if (is.null(fitind_ci)) {
    return(out)
  }

  out$label <- paste0(" [", .efa_ci_header_label(spec), "-CI]")
  out$CAF <- .efa_paste_gof_ci(fitind_ci, "CAF")
  out$RMSR <- .efa_paste_gof_ci(fitind_ci, "RMSR")
  out$CFI <- .efa_paste_gof_ci(fitind_ci, "CFI")
  out$RMSEA <- .efa_paste_gof_ci(fitind_ci, "RMSEA")
  out$AIC <- .efa_paste_gof_ci(fitind_ci, "AIC")
  out$BIC <- .efa_paste_gof_ci(fitind_ci, "BIC")

  out
}

.print_efa_bootstrap_note <- function(x, spec) {
  if (!isTRUE(spec$has_boot_ci)) {
    return(invisible(NULL))
  }

  note_label <- if (isTRUE(spec$is_pooled)) {
    "Bootstrap/MI CIs"
  } else {
    "Bootstrap CIs"
  }

  cat("\n", crayon::italic("Note: "), note_label, sep = "")

  b_text <- .efa_bootstrap_sample_text(spec)
  if (nzchar(b_text)) {
    cat(" based on ", b_text, sep = "")
  }

  target_rotation_note <- .efa_target_rotation_note(x, spec)
  if (nzchar(target_rotation_note)) {
    cat("; ", target_rotation_note, sep = "")
  }

  cat(".\n")

  invisible(NULL)
}

.print_efa_residuals_section <- function(x, residual_cutoff = .1,
                                     residual_top_n = 10, digits = 3) {
  .print_efa_rule("Residual Diagnostics")

  if (is.null(x$residuals)) {
    cat("No residual matrix available.\n")
    return(invisible(NULL))
  }

  residuals <- as.matrix(x$residuals)
  idx <- which(upper.tri(residuals) & is.finite(residuals), arr.ind = TRUE)

  if (nrow(idx) < 1L) {
    cat("No finite off-diagonal residuals available.\n")
    return(invisible(NULL))
  }

  values <- residuals[idx]
  keep <- abs(values) > residual_cutoff
  n_large <- sum(keep, na.rm = TRUE)
  largest_abs <- max(abs(values), na.rm = TRUE)

  cat(crayon::blue("Residual cutoff: "), "|r| > ",
    .efa_format_plain_number(residual_cutoff, digits), "\n",
    sep = ""
  )
  cat(crayon::blue("Number of large residuals: "), n_large, "\n", sep = "")
  cat(crayon::blue("Largest absolute residual: "),
    .efa_format_plain_number(largest_abs, digits), "\n",
    sep = ""
  )

  if (n_large < 1L) {
    cat("\nNo absolute residuals > ",
      .efa_format_plain_number(residual_cutoff, digits),
      " occurred.\n",
      sep = ""
    )
    cat("\nInspect the residual matrix for details (e.g., with residuals()).\n")
    return(invisible(NULL))
  }

  large_idx <- idx[keep, , drop = FALSE]
  large_values <- values[keep]
  ord <- order(abs(large_values), decreasing = TRUE)
  large_idx <- large_idx[ord, , drop = FALSE]
  large_values <- large_values[ord]

  if (is.finite(residual_top_n)) {
    n_print <- min(length(large_values), as.integer(residual_top_n))
  } else {
    n_print <- length(large_values)
  }

  large_idx <- large_idx[seq_len(n_print), , drop = FALSE]
  large_values <- large_values[seq_len(n_print)]

  var_names <- colnames(residuals)
  if (is.null(var_names)) {
    var_names <- rownames(residuals)
  }
  if (is.null(var_names)) {
    var_names <- paste0("V", seq_len(ncol(residuals)))
  }

  cat("\n")
  heading <- if (n_print < n_large) {
    paste0("Largest residuals (top ", n_print, " of ", n_large, "):")
  } else {
    "Largest residuals:"
  }
  cat(crayon::blue(heading), "\n", sep = "")

  for (i in seq_along(large_values)) {
    cat(cli::symbol$bullet, " ",
      var_names[large_idx[i, 1]], " ~~ ", var_names[large_idx[i, 2]],
      ": ", .efa_format_plain_number(large_values[i], digits), "\n",
      sep = ""
    )
  }

  cat("\nInspect the residual matrix for details (e.g., with residuals()).\n")

  invisible(NULL)
}

.print_efa_loading_ci_section <- function(x, spec, component, loadings, loading_label,
                                          cutoff, digits, max_name_length,
                                          ci, ci_filter,
                                          sort_loadings = c("none", "primary", "clustered")) {
  sort_loadings <- match.arg(sort_loadings)

  if (!.efa_should_print_ci(x, ci)) {
    return(invisible(NULL))
  }

  loading_ci <- .efa_get_component_ci(x, component)
  if (is.null(loading_ci)) {
    return(invisible(NULL))
  }

  loadings <- as.matrix(loadings)
  lower <- as.matrix(loading_ci$lower)
  upper <- as.matrix(loading_ci$upper)

  if (!all(dim(loadings) == dim(lower)) || !all(dim(loadings) == dim(upper))) {
    return(invisible(NULL))
  }

  row_order <- .efa_loading_row_order(loadings, sort_loadings)
  loadings <- loadings[row_order, , drop = FALSE]
  lower <- lower[row_order, , drop = FALSE]
  upper <- upper[row_order, , drop = FALSE]

  keep <- .efa_select_loading_cis(loadings, lower, upper, cutoff, ci_filter)
  title <- .efa_loading_ci_title(spec, loading_label, ci_filter)
  .print_efa_rule(title)

  if (!any(keep, na.rm = TRUE)) {
    cat("No loading CIs matched ci_filter = '", ci_filter, "'.\n", sep = "")
    return(invisible(NULL))
  }

  idx <- which(keep, arr.ind = TRUE)
  row_names <- rownames(loadings)
  if (is.null(row_names)) {
    row_names <- paste0("V", seq_len(nrow(loadings)))
  }
  row_names <- .efa_truncate_names(row_names, max_name_length)

  factor_names <- colnames(loadings)
  if (is.null(factor_names)) {
    factor_names <- paste0("F", seq_len(ncol(loadings)))
  }

  .print_efa_ci_table(
    key = row_names[idx[, 1]],
    second_key = factor_names[idx[, 2]],
    estimate = loadings[idx],
    lower = lower[idx],
    upper = upper[idx],
    digits = digits,
    key_header = "Variable",
    second_key_header = "Factor",
    highlight = abs(loadings[idx]) >= cutoff
  )

  invisible(NULL)
}

.print_efa_phi_ci_section <- function(x, spec, digits, ci) {
  if (!.efa_should_print_ci(x, ci)) {
    return(invisible(NULL))
  }

  phi_ci <- .efa_get_component_ci(x, "Phi")
  if (is.null(phi_ci) || is.null(x$Phi)) {
    return(invisible(NULL))
  }

  phi <- as.matrix(x$Phi)
  lower <- as.matrix(phi_ci$lower)
  upper <- as.matrix(phi_ci$upper)

  if (!all(dim(phi) == dim(lower)) || !all(dim(phi) == dim(upper))) {
    return(invisible(NULL))
  }

  keep <- lower.tri(phi) & is.finite(phi) & is.finite(lower) & is.finite(upper)
  if (!any(keep, na.rm = TRUE)) {
    return(invisible(NULL))
  }

  idx <- which(keep, arr.ind = TRUE)
  factor_names <- colnames(phi)
  if (is.null(factor_names)) {
    factor_names <- paste0("F", seq_len(ncol(phi)))
  }

  .print_efa_rule(.efa_phi_ci_title(spec))
  .print_efa_ci_table(
    key = paste0(factor_names[idx[, 2]], " ~~ ", factor_names[idx[, 1]]),
    second_key = NULL,
    estimate = phi[idx],
    lower = lower[idx],
    upper = upper[idx],
    digits = digits,
    key_header = "Factors",
    second_key_header = NULL,
    highlight = NULL
  )

  invisible(NULL)
}

.print_efa_heywood_warning <- function(h2) {
  if (is.null(h2)) {
    return(invisible(NULL))
  }

  n_heywood <- sum(h2 >= 1 + .Machine$double.eps, na.rm = TRUE)

  if (n_heywood == 1) {
    cat(crayon::red$bold("\nWarning: A Heywood case was detected!"))
    cat("\n")
  } else if (n_heywood > 1) {
    cat(crayon::red$bold("\nWarning:", n_heywood, "Heywood cases were detected!"))
    cat("\n")
  }

  invisible(NULL)
}

.print_efa_rule <- function(title) {
  cat("\n")
  cat(cli::rule(left = crayon::bold(title), col = "blue"))
  cat("\n")
  cat("\n")

  invisible(NULL)
}

.efa_should_print_ci <- function(x, ci) {
  !identical(ci, "none") && !is.null(x$boot.CI)
}

.efa_get_component_ci <- function(x, component) {
  if (is.null(x$boot.CI)) {
    return(NULL)
  }

  ci <- x$boot.CI[[component]]
  if (.efa_is_ci_pair(ci)) {
    return(ci)
  }

  NULL
}

.efa_get_fit_ci <- function(x, spec = .efa_print_spec(x)) {
  if (is.null(x$boot.CI)) {
    return(NULL)
  }

  components <- if (isTRUE(spec$is_pooled)) {
    c("fit_indices_pooled_algorithm", "fit_indices_descriptive", "fit_indices")
  } else {
    c("fit_indices", "fit_indices_pooled_algorithm", "fit_indices_descriptive")
  }

  for (component in components) {
    ci <- x$boot.CI[[component]]
    if (.efa_is_ci_pair(ci)) {
      return(ci)
    }
  }

  NULL
}

.efa_is_ci_pair <- function(x) {
  is.list(x) && !is.null(x$lower) && !is.null(x$upper)
}

.efa_select_loading_cis <- function(loadings, lower, upper, cutoff, ci_filter) {
  finite <- is.finite(loadings) & is.finite(lower) & is.finite(upper)

  if (identical(ci_filter, "all")) {
    return(finite)
  }

  if (identical(ci_filter, "nonzero")) {
    return(finite & (lower > 0 | upper < 0))
  }

  finite & abs(loadings) >= cutoff
}

.efa_n_imputations <- function(x) {
  n_imputations <- x$settings$n_imputations

  if ((is.null(n_imputations) || length(n_imputations) < 1L || is.na(n_imputations[1])) &&
      !is.null(x$fits)) {
    n_imputations <- length(x$fits)
  }

  if (is.null(n_imputations) || length(n_imputations) < 1L || is.na(n_imputations[1])) {
    return(NA_integer_)
  }

  as.integer(n_imputations[1])
}

.efa_setting_text <- function(x) {
  if (is.null(x) || length(x) < 1L || is.na(x[1])) {
    return("")
  }

  as.character(x[1])
}

.efa_ci_level <- function(x) {
  val <- x$settings$ci
  if (is.null(val) || length(val) < 1L || is.na(val[1])) {
    return(NA_real_)
  }

  out <- suppressWarnings(as.numeric(val[1]))
  if (length(out) < 1L || is.na(out)) {
    return(NA_real_)
  }

  out
}

.efa_rmsea_ci_level <- function(x) {
  val <- x$settings$rmsea_ci_level
  if (is.null(val) || length(val) < 1L || is.na(val[1])) {
    return(.90)
  }

  out <- suppressWarnings(as.numeric(val[1]))
  if (length(out) < 1L || !is.finite(out) || out <= 0 || out >= 1) {
    return(.90)
  }

  out
}

.efa_ci_level_text <- function(level) {
  if (!is.finite(level)) {
    return("")
  }

  paste0(formatC(level * 100, format = "fg", digits = 4, width = -1), "%")
}

.efa_ci_source_text <- function(spec) {
  if (isTRUE(spec$is_pooled)) {
    return("bootstrap/MI")
  }

  "bootstrap"
}

.efa_target_rotation_note <- function(x, spec) {
  counts <- .efa_valid_target_rotations(x)
  if (is.null(counts) || length(counts) < 1L || all(is.na(counts))) {
    return("")
  }

  counts <- as.integer(stats::na.omit(counts))
  if (length(counts) < 1L) {
    return("")
  }

  total <- .efa_total_bootstrap_samples(x, spec)
  if (!is.na(total) && !any(counts < total)) {
    return("")
  }

  component_text <- .efa_target_rotation_component_text(x)
  count_text <- .efa_valid_target_rotation_count_text(counts, spec)
  total_text <- if (!is.na(total)) {
    paste0(" out of ", total)
  } else {
    ""
  }
  scope_text <- if (isTRUE(spec$is_pooled)) {
    " per imputation"
  } else {
    ""
  }

  paste0(
    component_text,
    " based on ", count_text,
    " valid target-rotated bootstrap samples",
    total_text,
    scope_text
  )
}

.efa_valid_target_rotations <- function(x) {
  if (!is.null(x$boot.SE$valid_target_rotations)) {
    return(x$boot.SE$valid_target_rotations)
  }

  if (!is.null(x$boot.MI$bootstrap_rotation_valid)) {
    return(x$boot.MI$bootstrap_rotation_valid)
  }

  NULL
}

.efa_total_bootstrap_samples <- function(x, spec) {
  if (!is.null(spec$b_boot) && length(spec$b_boot) >= 1L &&
      is.finite(spec$b_boot[1])) {
    return(as.integer(spec$b_boot[1]))
  }

  if (!is.null(x$boot.arrays$unrot_loadings)) {
    arr <- x$boot.arrays$unrot_loadings
    if (length(dim(arr)) >= 3L) {
      return(dim(arr)[3L])
    }
  }

  NA_integer_
}

.efa_target_rotation_component_text <- function(x) {
  if (!is.null(x$Phi)) {
    return("rotated-loading, factor-intercorrelation, and structure-coefficient CIs")
  }

  "rotated-loading CIs"
}

.efa_valid_target_rotation_count_text <- function(counts, spec) {
  counts <- as.integer(counts)

  if (!isTRUE(spec$is_pooled) || length(unique(counts)) == 1L) {
    return(as.character(counts[1L]))
  }

  paste0("between ", min(counts), " and ", max(counts))
}

.efa_ci_header_label <- function(spec) {
  level <- .efa_ci_level_text(spec$ci)
  source <- .efa_ci_source_text(spec)

  if (nzchar(level)) {
    paste(level, source)
  } else {
    source
  }
}

.efa_loading_ci_title <- function(spec, loading_label, ci_filter) {
  filter_text <- switch(ci_filter,
    salient = "salient",
    all = "all",
    nonzero = "nonzero",
    ci_filter
  )

  paste0(
    .efa_ci_header_label(spec),
    " CIs for ", filter_text, " ", loading_label, " loadings"
  )
}

.efa_phi_ci_title <- function(spec) {
  paste0(.efa_ci_header_label(spec), " CIs for factor intercorrelations")
}

.efa_truncate_names <- function(x, max_name_length) {
  x <- as.character(x)
  too_long <- nchar(x) > max_name_length
  x[too_long] <- substr(x[too_long], 1, max_name_length)
  x
}

.efa_format_ci_num <- function(x, digits) {
  vapply(x, function(z) {
    .numformat(round(z, digits = digits), digits = digits)
  }, character(1))
}

.print_efa_ci_table <- function(key, second_key = NULL, estimate, lower, upper,
                                digits, key_header, second_key_header = NULL,
                                highlight = NULL) {
  has_second_key <- !is.null(second_key)

  estimate_chr <- .efa_format_ci_num(estimate, digits)
  lower_chr <- .efa_format_ci_num(lower, digits)
  upper_chr <- .efa_format_ci_num(upper, digits)

  if (has_second_key) {
    plain_rows <- data.frame(
      key = as.character(key),
      second_key = as.character(second_key),
      estimate = estimate_chr,
      lower = lower_chr,
      upper = upper_chr,
      stringsAsFactors = FALSE
    )
    headers <- c(key_header, second_key_header, "est", "lower", "upper")
  } else {
    plain_rows <- data.frame(
      key = as.character(key),
      estimate = estimate_chr,
      lower = lower_chr,
      upper = upper_chr,
      stringsAsFactors = FALSE
    )
    headers <- c(key_header, "est", "lower", "upper")
  }

  widths <- vapply(seq_along(headers), function(j) {
    max(nchar(c(headers[j], plain_rows[[j]])), na.rm = TRUE)
  }, numeric(1))

  header <- paste(
    stringr::str_pad(headers, widths, side = "right"),
    collapse = "  "
  )
  cat(crayon::blue(header), "\n", sep = "")

  for (i in seq_len(nrow(plain_rows))) {
    cells <- as.character(unlist(plain_rows[i, , drop = FALSE], use.names = FALSE))
    padded <- stringr::str_pad(cells, widths, side = "right")

    if (has_second_key) {
      padded[1] <- crayon::blue(padded[1])
      padded[2] <- crayon::blue(padded[2])
      if (!is.null(highlight) && isTRUE(highlight[i])) {
        padded[3] <- crayon::bold(padded[3])
      }
    } else {
      padded[1] <- crayon::blue(padded[1])
    }

    cat(paste(padded, collapse = "  "), "\n", sep = "")
  }

  invisible(NULL)
}


.efa_validate_print_options <- function(cutoff, digits, max_name_length,
                                        diagnostics, diagnostics_top_n,
                                        residual_cutoff, residual_top_n,
                                        show_structure, show_loading_legend,
                                        cross_loading_cutoff, min_primary_gap,
                                        min_salient_per_factor,
                                        max_factors_per_block,
                                        show_mi_diagnostics) {
  if (!is.numeric(cutoff) || length(cutoff) != 1L ||
      !is.finite(cutoff) || cutoff < 0) {
    stop("`cutoff` must be a single finite non-negative number.", call. = FALSE)
  }

  if (!is.numeric(digits) || length(digits) != 1L || !is.finite(digits) ||
      digits < 0 || digits != as.integer(digits)) {
    stop("`digits` must be a single finite non-negative integer.", call. = FALSE)
  }

  if (!is.numeric(max_name_length) || length(max_name_length) != 1L ||
      !is.finite(max_name_length) || max_name_length < 1 ||
      max_name_length != as.integer(max_name_length)) {
    stop("`max_name_length` must be a single finite positive integer.", call. = FALSE)
  }

  if (!is.logical(diagnostics) || length(diagnostics) != 1L || is.na(diagnostics)) {
    stop("`diagnostics` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!.efa_is_top_n(diagnostics_top_n)) {
    stop("`diagnostics_top_n` must be a positive integer or Inf.", call. = FALSE)
  }

  if (!is.numeric(residual_cutoff) || length(residual_cutoff) != 1L ||
      !is.finite(residual_cutoff) || residual_cutoff < 0) {
    stop("`residual_cutoff` must be a single finite non-negative number.", call. = FALSE)
  }

  if (!.efa_is_top_n(residual_top_n)) {
    stop("`residual_top_n` must be a positive integer or Inf.", call. = FALSE)
  }

  if (!is.logical(show_structure) || length(show_structure) != 1L ||
      is.na(show_structure)) {
    stop("`show_structure` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.logical(show_loading_legend) || length(show_loading_legend) != 1L ||
      is.na(show_loading_legend)) {
    stop("`show_loading_legend` must be TRUE or FALSE.", call. = FALSE)
  }

  if (!is.numeric(cross_loading_cutoff) || length(cross_loading_cutoff) != 1L ||
      !is.finite(cross_loading_cutoff) || cross_loading_cutoff < 0) {
    stop("`cross_loading_cutoff` must be a single finite non-negative number.", call. = FALSE)
  }

  if (!is.numeric(min_primary_gap) || length(min_primary_gap) != 1L ||
      !is.finite(min_primary_gap) || min_primary_gap < 0) {
    stop("`min_primary_gap` must be a single finite non-negative number.", call. = FALSE)
  }

  if (!is.numeric(min_salient_per_factor) || length(min_salient_per_factor) != 1L ||
      !is.finite(min_salient_per_factor) || min_salient_per_factor < 1 ||
      min_salient_per_factor != as.integer(min_salient_per_factor)) {
    stop("`min_salient_per_factor` must be a single finite positive integer.", call. = FALSE)
  }

  if (!is.null(max_factors_per_block) &&
      (!is.numeric(max_factors_per_block) || length(max_factors_per_block) != 1L ||
       !is.finite(max_factors_per_block) || max_factors_per_block < 1 ||
       max_factors_per_block != as.integer(max_factors_per_block))) {
    stop("`max_factors_per_block` must be NULL or a single finite positive integer.", call. = FALSE)
  }

  if (!is.null(show_mi_diagnostics) &&
      (!is.logical(show_mi_diagnostics) || length(show_mi_diagnostics) != 1L ||
       is.na(show_mi_diagnostics))) {
    stop("`show_mi_diagnostics` must be TRUE, FALSE, or NULL.", call. = FALSE)
  }

  invisible(TRUE)
}

.efa_iteration_nonconvergence <- function(spec) {
  iter <- .efa_as_scalar_number(spec$iter)
  max_iter <- .efa_as_scalar_number(spec$max_iter)

  is.finite(iter) && is.finite(max_iter) && iter > max_iter
}

.efa_as_scalar_number <- function(x) {
  if (is.null(x) || length(x) < 1L) {
    return(NA_real_)
  }

  out <- tryCatch(
    suppressWarnings(as.numeric(x[1])),
    error = function(e) NA_real_
  )
  if (length(out) < 1L) {
    return(NA_real_)
  }

  out
}

.efa_is_missing_number <- function(x) {
  val <- .efa_as_scalar_number(x)
  !is.finite(val)
}

.efa_fit_scalar <- function(fit, name) {
  if (is.null(fit) || is.null(fit[[name]]) || length(fit[[name]]) < 1L) {
    return(NA_real_)
  }

  out <- tryCatch(
    suppressWarnings(as.numeric(fit[[name]][1])),
    error = function(e) NA_real_
  )
  if (length(out) < 1L) {
    return(NA_real_)
  }

  out
}

.efa_format_number <- function(x, digits = 2, print_zero = FALSE, pad = TRUE) {
  if (length(x) != 1L || !is.finite(x)) {
    return("NA")
  }

  .numformat(round(x, digits = digits), digits = digits,
             print_zero = print_zero, pad = pad)
}

.efa_format_fit_value <- function(fit, name, digits = 2,
                                  print_zero = FALSE, pad = TRUE) {
  .efa_format_number(
    .efa_fit_scalar(fit, name),
    digits = digits,
    print_zero = print_zero,
    pad = pad
  )
}

.efa_bootstrap_sample_text <- function(spec) {
  b_boot <- .efa_as_scalar_number(spec$b_boot)
  if (!is.finite(b_boot)) {
    return("")
  }

  label <- if (isTRUE(spec$is_pooled)) {
    "bootstrap samples per imputation"
  } else {
    "bootstrap samples"
  }

  paste(as.integer(b_boot), label)
}

.efa_alignment_summary <- function(x) {
  alignment <- x$alignment
  if (is.null(alignment)) {
    return("")
  }

  out <- character(0)

  method <- .efa_setting_text(alignment$method)
  if (!nzchar(method)) {
    method <- .efa_setting_text(x$settings$target_method)
  }
  if (nzchar(method)) {
    out <- c(out, paste0("method = '", method, "'"))
  }

  if (!is.null(alignment$converged)) {
    converged <- isTRUE(alignment$converged)
    out <- c(out, if (converged) "converged" else "not converged")
  }

  failures <- alignment$point_rotation_failures
  if (!is.null(failures) && length(failures) > 0L) {
    failures <- failures[!is.na(failures)]
    if (length(failures) > 0L) {
      out <- c(out, paste0("point-alignment failures = ", length(failures)))
    }
  }

  paste(out, collapse = ", ")
}

.efa_is_top_n <- function(x) {
  is.numeric(x) && length(x) == 1L && !is.na(x) &&
    (is.infinite(x) || (x >= 1 && x == as.integer(x)))
}

.efa_resolve_show_structure <- function(show_structure, details) {
  isTRUE(show_structure) || identical(details, "full")
}

.efa_resolve_show_mi_diagnostics <- function(show_mi_diagnostics, details, spec) {
  if (!is.null(show_mi_diagnostics)) {
    return(isTRUE(show_mi_diagnostics))
  }

  isTRUE(spec$is_pooled) && identical(details, "full")
}

.efa_main_loadings <- function(x, spec) {
  if (identical(spec$rotation, "none")) {
    if (is.null(x$unrot_loadings)) {
      return(NULL)
    }
    return(as.matrix(x$unrot_loadings))
  }

  if (is.null(x$rot_loadings)) {
    return(NULL)
  }

  as.matrix(x$rot_loadings)
}

.efa_variable_names <- function(x) {
  rn <- rownames(x)
  if (is.null(rn)) {
    rn <- paste0("V", seq_len(nrow(x)))
  }
  rn
}

.efa_factor_names <- function(x) {
  cn <- colnames(x)
  if (is.null(cn)) {
    cn <- paste0("F", seq_len(ncol(x)))
  }
  cn
}

.print_efa_diagnostics_section <- function(x, spec, cutoff, digits,
                                           cross_loading_cutoff,
                                           min_primary_gap,
                                           min_salient_per_factor) {
  loadings <- .efa_main_loadings(x, spec)

  title <- if (isTRUE(spec$is_pooled)) {
    "Pooled Model Diagnostics"
  } else {
    "Model Diagnostics"
  }
  .print_efa_rule(title)

  if (is.null(loadings)) {
    cat("No loading matrix available for diagnostics.\n")
    return(invisible(NULL))
  }

  .efa_print_key_value("Factors", ncol(loadings))
  .efa_print_key_value("Variables", nrow(loadings))

  n_text <- .efa_setting_text(spec$N)
  if (nzchar(n_text) && !identical(n_text, "NA")) {
    .efa_print_key_value("N", n_text)
  }

  if (isTRUE(spec$is_pooled)) {
    n_imp <- .efa_setting_text(spec$n_imputations)
    if (nzchar(n_imp)) {
      .efa_print_key_value("Imputations", n_imp)
    }
    pooling <- .efa_pooled_settings_text(spec)
    if (length(pooling) > 0L) {
      .efa_print_key_value("Pooling", paste(pooling, collapse = ", "))
    }

    alignment_text <- .efa_alignment_summary(x)
    if (nzchar(alignment_text)) {
      .efa_print_key_value("Alignment", alignment_text)
    }
  }

  if (isTRUE(spec$np_boot) && !is.null(spec$b_boot)) {
    sample_label <- if (isTRUE(spec$is_pooled)) {
      "Bootstrap samples per imputation"
    } else {
      "Bootstrap samples"
    }
    .efa_print_key_value(sample_label, spec$b_boot)

    valid_text <- .efa_valid_target_rotation_summary(x, spec)
    if (nzchar(valid_text)) {
      .efa_print_key_value("Valid target-rotated samples", valid_text)
    }
  }

  h2 <- spec$h2
  heywood <- if (is.null(h2)) {
    0L
  } else {
    sum(h2 >= 1 + .Machine$double.eps, na.rm = TRUE)
  }
  .efa_print_key_value("Heywood cases", heywood)

  cross_salient <- abs(loadings) >= cross_loading_cutoff
  cross_salient[is.na(cross_salient)] <- FALSE
  cross_loading_items <- sum(rowSums(cross_salient) > 1L)

  salient <- abs(loadings) >= cutoff
  salient[is.na(salient)] <- FALSE
  no_salient_items <- sum(rowSums(salient) < 1L)
  .efa_print_key_value(
    paste0("Cross-loading items (|loading| >= ",
           .efa_format_plain_number(cross_loading_cutoff, digits), ")"),
    cross_loading_items
  )
  .efa_print_key_value(
    paste0("Items without salient loading (|loading| >= ",
           .efa_format_plain_number(cutoff, digits), ")"),
    no_salient_items
  )

  weak_factors <- .efa_factors_below_indicator_count(
    loadings,
    cutoff = cutoff,
    min_salient_per_factor = min_salient_per_factor
  )
  .efa_print_key_value(
    paste0("Factors with fewer than ", min_salient_per_factor,
           " salient indicators"),
    length(weak_factors)
  )

  gap_count <- .efa_small_primary_gap_count(loadings,
    cutoff = cutoff,
    min_primary_gap = min_primary_gap
  )
  .efa_print_key_value(
    paste0("Items with primary-loading gap < ",
           .efa_format_plain_number(min_primary_gap, digits)),
    gap_count
  )

  if (!is.null(x$residuals)) {
    largest_resid <- .efa_largest_abs_residual(x$residuals)
    if (is.finite(largest_resid)) {
      .efa_print_key_value("Largest |residual|",
        .efa_format_plain_number(largest_resid, digits)
      )
    }
  }

  if (!is.null(x$Phi)) {
    high_phi <- .efa_high_factor_correlation_count(x$Phi, cutoff = .85)
    value <- if (high_phi > 0L) {
      as.character(high_phi)
    } else {
      "none"
    }
    .efa_print_key_value("Factor intercorrelations > .85", value)
  }

  invisible(NULL)
}

.efa_print_key_value <- function(key, value) {
  cat(crayon::blue(paste0(key, ": ")), value, "\n", sep = "")
  invisible(NULL)
}

.efa_valid_target_rotation_summary <- function(x, spec) {
  counts <- .efa_valid_target_rotations(x)
  if (is.null(counts) || length(counts) < 1L || all(is.na(counts))) {
    return("")
  }

  counts <- as.integer(stats::na.omit(counts))
  if (length(counts) < 1L) {
    return("")
  }

  total <- .efa_total_bootstrap_samples(x, spec)
  count_text <- .efa_valid_target_rotation_count_text(counts, spec)

  if (!is.na(total)) {
    count_text <- paste0(count_text, " out of ", total)
  }

  if (isTRUE(spec$is_pooled)) {
    count_text <- paste0(count_text, " per imputation")
  }

  count_text
}

.efa_factors_below_indicator_count <- function(loadings, cutoff,
                                               min_salient_per_factor) {
  salient <- abs(loadings) >= cutoff
  salient[is.na(salient)] <- FALSE
  counts <- colSums(salient)
  which(counts < min_salient_per_factor)
}

.efa_small_primary_gap_count <- function(loadings, cutoff, min_primary_gap) {
  if (ncol(loadings) < 2L) {
    return(0L)
  }

  abs_loadings <- abs(loadings)
  abs_loadings[is.na(abs_loadings)] <- -Inf
  sorted <- t(apply(abs_loadings, 1L, sort, decreasing = TRUE))
  primary <- sorted[, 1L]
  secondary <- sorted[, 2L]
  sum(is.finite(primary) & primary >= cutoff &
        is.finite(secondary) & (primary - secondary) < min_primary_gap)
}

.efa_largest_abs_residual <- function(residuals) {
  residuals <- as.matrix(residuals)
  values <- residuals[upper.tri(residuals)]
  values <- values[is.finite(values)]
  if (length(values) < 1L) {
    return(NA_real_)
  }
  max(abs(values), na.rm = TRUE)
}

.efa_high_factor_correlation_count <- function(phi, cutoff = .85) {
  phi <- as.matrix(phi)
  values <- phi[lower.tri(phi)]
  values <- values[is.finite(values)]
  sum(abs(values) > cutoff, na.rm = TRUE)
}

.print_efa_simple_structure_section <- function(x, spec, cutoff, digits,
                                                max_name_length,
                                                cross_loading_cutoff,
                                                min_primary_gap,
                                                min_salient_per_factor,
                                                diagnostics_top_n,
                                                sort_loadings) {
  loadings <- .efa_main_loadings(x, spec)
  row_order <- .efa_loading_row_order(loadings, sort_loadings)
  loadings <- loadings[row_order, , drop = FALSE]

  summary <- .efa_simple_structure_summary(
    loadings = loadings,
    cutoff = cutoff,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    digits = digits,
    max_name_length = max_name_length
  )

  if (!summary$has_output) {
    return(invisible(NULL))
  }

  .print_efa_rule("Simple Structure Diagnostics")

  if (length(summary$no_salient) > 0L) {
    .efa_print_limited_bullets(
      heading = paste0("Items with no salient loading, |loading| >= ",
                       .efa_format_plain_number(cutoff, digits), ":"),
      values = summary$no_salient,
      top_n = diagnostics_top_n
    )
  }

  if (length(summary$cross_loadings) > 0L) {
    .efa_print_limited_bullets(
      heading = paste0("Items with cross-loadings, |loading| >= ",
                       .efa_format_plain_number(cross_loading_cutoff, digits),
                       " on multiple factors:"),
      values = summary$cross_loadings,
      top_n = diagnostics_top_n
    )
  }

  if (length(summary$small_gaps) > 0L) {
    .efa_print_limited_bullets(
      heading = paste0("Items with primary-loading gap < ",
                       .efa_format_plain_number(min_primary_gap, digits), ":"),
      values = summary$small_gaps,
      top_n = diagnostics_top_n
    )
  }

  if (length(summary$weak_factors) > 0L) {
    .efa_print_limited_bullets(
      heading = paste0("Factors with fewer than ", min_salient_per_factor,
                       " salient indicators:"),
      values = summary$weak_factors,
      top_n = Inf
    )
  }

  invisible(NULL)
}

.efa_simple_structure_summary <- function(loadings, cutoff, cross_loading_cutoff,
                                          min_primary_gap,
                                          min_salient_per_factor,
                                          digits, max_name_length) {
  var_names <- .efa_truncate_names(.efa_variable_names(loadings), max_name_length)
  factor_names <- .efa_factor_names(loadings)

  abs_loadings <- abs(loadings)
  salient_for_items <- abs_loadings >= cutoff
  salient_for_items[is.na(salient_for_items)] <- FALSE

  cross_salient <- abs_loadings >= cross_loading_cutoff
  cross_salient[is.na(cross_salient)] <- FALSE

  no_salient <- var_names[rowSums(salient_for_items) < 1L]

  cross_rows <- which(rowSums(cross_salient) > 1L)
  cross_loadings <- vapply(cross_rows, function(i) {
    idx <- which(cross_salient[i, ])
    paste0(var_names[i], ": ",
      .efa_format_loading_pairs(loadings[i, idx], factor_names[idx], digits)
    )
  }, character(1L))

  small_gap_rows <- .efa_small_primary_gap_rows(loadings,
    cutoff = cutoff,
    min_primary_gap = min_primary_gap
  )
  small_gaps <- vapply(small_gap_rows, function(i) {
    ord <- order(abs(loadings[i, ]), decreasing = TRUE)
    idx <- ord[seq_len(min(2L, length(ord)))]
    paste0(var_names[i], ": ",
      .efa_format_loading_pairs(loadings[i, idx], factor_names[idx], digits)
    )
  }, character(1L))

  factor_counts <- colSums(salient_for_items)
  weak_factor_idx <- which(factor_counts < min_salient_per_factor)
  if (length(weak_factor_idx) > 0L) {
    weak_factors <- paste0(
      factor_names[weak_factor_idx],
      ": ",
      factor_counts[weak_factor_idx],
      " salient indicator",
      ifelse(factor_counts[weak_factor_idx] == 1L, "", "s")
    )
  } else {
    weak_factors <- NULL
  }


  list(
    no_salient = no_salient,
    cross_loadings = cross_loadings,
    small_gaps = small_gaps,
    weak_factors = weak_factors,
    has_output = length(no_salient) > 0L ||
      length(cross_loadings) > 0L ||
      length(small_gaps) > 0L ||
      length(weak_factors) > 0L
  )
}

.efa_small_primary_gap_rows <- function(loadings, cutoff, min_primary_gap) {
  if (ncol(loadings) < 2L) {
    return(integer(0))
  }

  abs_loadings <- abs(loadings)
  abs_loadings[is.na(abs_loadings)] <- -Inf
  sorted <- t(apply(abs_loadings, 1L, sort, decreasing = TRUE))
  primary <- sorted[, 1L]
  secondary <- sorted[, 2L]

  which(is.finite(primary) & primary >= cutoff &
          is.finite(secondary) & (primary - secondary) < min_primary_gap)
}

.efa_format_loading_pairs <- function(values, factor_names, digits) {
  ord <- order(abs(values), decreasing = TRUE)
  values <- values[ord]
  factor_names <- factor_names[ord]

  paste0(
    factor_names,
    " = ",
    vapply(values, .efa_format_plain_number, character(1L), digits = digits),
    collapse = ", "
  )
}

.efa_print_limited_bullets <- function(heading, values, top_n) {
  cat(crayon::blue(heading), "\n", sep = "")

  if (length(values) < 1L) {
    return(invisible(NULL))
  }

  n_print <- if (is.finite(top_n)) {
    min(length(values), as.integer(top_n))
  } else {
    length(values)
  }

  for (i in seq_len(n_print)) {
    cat(cli::symbol$bullet, " ", values[i], "\n", sep = "")
  }

  if (n_print < length(values)) {
    cat(cli::symbol$bullet, " ... ", length(values) - n_print,
      " more not shown\n",
      sep = ""
    )
  }

  cat("\n")
  invisible(NULL)
}

.print_efa_structure_section <- function(x, cutoff, digits, max_name_length,
                                         sort_loadings, max_factors_per_block,
                                         ...) {
  .print_efa_rule("Structure Matrix")
  print(x$Structure,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = NULL,
    sort_loadings = sort_loadings,
    legend = FALSE,
    max_factors_per_block = max_factors_per_block,
    ...
  )
  invisible(NULL)
}

.print_efa_structure_note <- function() {
  cat("\n")
  cat(crayon::italic("Note: "),
    "Structure coefficients are available; use show_structure = TRUE to print them.",
    "\n",
    sep = ""
  )
  invisible(NULL)
}

.print_efa_mi_diagnostics_section <- function(x, spec, digits = 3) {
  if (!isTRUE(spec$is_pooled) || is.null(x$boot.MI)) {
    return(invisible(NULL))
  }

  fmi_values <- .efa_collect_mi_values(x$boot.MI, pattern = "fmi")
  riv_values <- .efa_collect_mi_values(x$boot.MI, pattern = "riv")

  if (length(fmi_values) < 1L && length(riv_values) < 1L) {
    return(invisible(NULL))
  }

  .print_efa_rule("MI Uncertainty Summary")

  if (length(fmi_values) > 0L) {
    .efa_print_key_value("Largest available FMI",
      .efa_format_plain_number(max(fmi_values, na.rm = TRUE), digits)
    )
    .efa_print_key_value("Median available FMI",
      .efa_format_plain_number(stats::median(fmi_values, na.rm = TRUE), digits)
    )
  }

  if (length(riv_values) > 0L) {
    .efa_print_key_value("Largest available RIV",
      .efa_format_plain_number(max(riv_values, na.rm = TRUE), digits)
    )
    .efa_print_key_value("Median available RIV",
      .efa_format_plain_number(stats::median(riv_values, na.rm = TRUE), digits)
    )
  }

  invisible(NULL)
}

.efa_collect_mi_values <- function(x, pattern, path = "") {
  pieces <- .efa_collect_mi_values_impl(x, pattern = pattern, path = path)
  if (length(pieces) < 1L) {
    return(numeric(0))
  }

  values <- unlist(pieces, use.names = FALSE)
  values[is.finite(values)]
}

.efa_collect_mi_values_impl <- function(x, pattern, path = "") {
  if (is.list(x)) {
    nm <- names(x)
    if (is.null(nm)) {
      nm <- rep("", length(x))
    }

    pieces <- vector("list", length(x))
    for (i in seq_along(x)) {
      child_path <- paste(c(path, nm[i]), collapse = "/")
      pieces[[i]] <- .efa_collect_mi_values_impl(x[[i]], pattern, child_path)
    }
    return(unlist(pieces, recursive = FALSE, use.names = FALSE))
  }

  if (!is.numeric(x) || !grepl(pattern, path, ignore.case = TRUE)) {
    return(list())
  }

  list(as.numeric(x))
}

.efa_loading_row_order <- function(x, sort_loadings = c("none", "primary", "clustered")) {
  sort_loadings <- match.arg(sort_loadings)

  if (identical(sort_loadings, "none") || nrow(x) < 2L || ncol(x) < 1L) {
    return(seq_len(nrow(x)))
  }

  abs_x <- abs(x)
  abs_x[is.na(abs_x)] <- -Inf
  primary_factor <- max.col(abs_x, ties.method = "first")
  primary_loading <- apply(abs_x, 1L, max, na.rm = TRUE)
  primary_loading[!is.finite(primary_loading)] <- -Inf

  if (identical(sort_loadings, "primary")) {
    return(order(primary_factor, seq_along(primary_factor)))
  }

  order(primary_factor, -primary_loading, seq_along(primary_factor))
}

.efa_format_plain_number <- function(x, digits = 3, pad = FALSE) {
  if (length(x) != 1L || is.na(x)) {
    return("NA")
  }

  .numformat(round(x, digits = digits), digits = digits, pad = pad)
}


.efa_paste_gof_ci <- function(fitind_ci, name) {
  if (!.efa_is_ci_pair(fitind_ci)) {
    return("")
  }

  lower_names <- names(fitind_ci$lower)
  upper_names <- names(fitind_ci$upper)
  if (is.null(lower_names) || is.null(upper_names) ||
      !(name %in% lower_names) || !(name %in% upper_names)) {
    return("")
  }

  tryCatch(.paste_gof_ci(fitind_ci, name), error = function(e) "")
}

#' Various factor retention criteria
#'
#' Choosing the number of factors to retain is one of the most important
#' decisions in an exploratory factor analysis (EFA). Many criteria exist to
#' help with this choice. This function runs several of them together, and can
#' also check whether the data are suitable for factor analysis.
#'
#' @param x data.frame or matrix. Raw data, or a correlation matrix. If `"CD"` is
#' included as a criterion, x must be raw data.
#' @param criteria character. Which factor retention methods to run: one or
#' more of `"CD"`, `"EKC"`, `"HULL"`, `"KGC"`, `"MAP"`, `"NEST"`, `"PARALLEL"`,
#' `"SCREE"`, and `"SMT"` (see details). The default runs a subset of commonly
#' used, well-performing methods, listed in the details.
#' @param suitability logical. Whether the data should be checked for suitability
#' for factor analysis using Bartlett's test of sphericity and the
#' Kaiser-Meyer-Olkin criterion (see details). Default is `TRUE`.
#' @param N  numeric. The number of observations. Only needed if x is a
#' correlation matrix.
#' @param use character. Passed to [stats::cor()] if raw
#' data is given as input. Default is `"pairwise.complete.obs"`.
#' @param cor_method character. Correlation computed from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (passed to [stats::cor()]), or `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations (a two-step estimator).
#'   `CD`, `PARALLEL`, `NEST`, `HULL`, and `SMT` do not support `"poly"` / `"tetra"`
#'   and are skipped automatically if you request them together.
#' Default is `"pearson"`.
#' @param n_factors_max numeric. Passed to [efa_cd()]. The maximum number
#' of factors to test against.
#' Larger numbers will increase the duration the procedure takes, but test more
#' possible solutions. If left NA (default), the maximum number of factors for
#' which the model is still over-identified (df > 0) is used.
#' @param N_pop numeric. Passed to [efa_cd()]. Size of finite populations
#' of comparison data. Default is 10000.
#' @param N_samples numeric. Passed to [efa_cd()]. Number of samples drawn
#'  from each population. Default is 500.
#' @param alpha numeric. Passed to [efa_cd()]. The alpha level used to test
#'  the significance of the improvement added by an additional factor.
#'  Default is .30.
#' @param max_iter_CD numeric. Passed to [efa_cd()]. The maximum number of
#'  iterations to perform after which the iterative PAF procedure is halted.
#'   Default is 50.
#' @param n_fac_theor numeric. Passed to [efa_hull()]. Theoretical number
#'  of factors to retain. The Hull method uses one plus the larger of this
#'  number and the number of factors suggested by [efa_parallel()] as its
#'  upper bound.
#' @param estimator character. Passed to [efa_fit()] in [efa_hull()],
#' [efa_kgc()], [efa_scree()], [efa_parallel()], and [efa_nest()]. The
#' estimator to use. One of  `"PAF"`, `"ULS"`, or  `"ML"`,
#' for principal axis factoring, unweighted least squares, and maximum
#' likelihood, respectively. The default here is `"ML"`. Some criteria default to
#' something else when called on their own (for example, [efa_hull()] defaults to
#' `"PAF"`), so results from `efa_retain()` can differ from calling that criterion
#' directly unless you set `estimator` to match. In [efa_kgc()], [efa_scree()], and
#' [efa_parallel()] it only takes effect when the respective `eigen_type` includes `"EFA"`.
#' @param gof character. Passed to [efa_hull()]. The goodness of fit index
#' to use. Either `"CAF"`, `"CFI"`, or `"RMSEA"`, or any
#' combination of them. With the `"PAF"` estimator, only
#' the CAF can be used as goodness of fit index. For details on the CAF, see
#' Lorenzo-Seva, Timmerman, and Kiers (2011).
#' @param eigen_type_HULL character. Passed to [efa_parallel()] in
#' [efa_hull()]. What the eigenvalues in the parallel analysis are based on.
#' One of `"SMC"`, `"PCA"`, or `"EFA"` -- different ways of estimating how much
#' variance each indicator shares with the others before the eigenvalues are
#' computed. `"SMC"` (default) uses each indicator's squared multiple
#' correlation with the others (its diagonal value in the correlation matrix).
#' `"PCA"` leaves the diagonal at 1, so each indicator's total variance -- not
#' just the shared part -- feeds into the eigenvalues. `"EFA"` uses the
#' communalities from a fitted EFA solution instead.
#' @param eigen_type_other character. Passed to [efa_kgc()],
#' [efa_scree()], and [efa_parallel()]. The same as eigen_type_HULL,
#' but multiple inputs are possible here (any combination of `"PCA"`, `"SMC"`,
#' and `"EFA"`). Default is `"SMC"`.
#' @param n_factors numeric. Passed to [efa_parallel()] (also within
#' [efa_hull()]), [efa_kgc()], and [efa_scree()]. Number of
#' factors to extract if `"EFA"` is included in `eigen_type_HULL` or
#'  `eigen_type_other`. Default is 1.
#' @param n_datasets numeric. Passed to [efa_parallel()] (also within
#' [efa_hull()]). The number of datasets to simulate. Default is 1000.
#' @param percent numeric. Passed to [efa_parallel()] (also within
#' [efa_hull()]). The percentile to take from the simulated eigenvalues.
#'  Default is 95.
#' @param decision_rule character. Passed to [efa_parallel()] (also within
#'  [efa_hull()]). Which rule to use to determine the number of
#'  factors to retain. Default is `"means"`, which uses the average simulated
#'  eigenvalues. `"percentile"` uses the percentiles specified in `percent`.
#'  `"crawford"` uses the 95th percentile for the first factor and the mean
#'  afterwards (based on Crawford et al., 2010).
#' @param ekc_type `r lifecycle::badge("deprecated")` Accepted and ignored. It
#'   used to select between two ways to compute the [efa_ekc()] reference values.
#'   The `"AM2019"` reference values do not depend on the observed eigenvalues.
#'   They therefore skip the empirical correction that defines the criterion, so
#'   they are no longer computed.
#' @param n_datasets_nest numeric. Passed to [efa_nest()]. The number of
#'   datasets to simulate. Default is 1000.
#' @param alpha_nest numeric. Passed to [efa_nest()]. The alpha level to use.
#'   The reference values are the eigenvalues at the (1 - alpha_nest)
#'   percentile. Default is .05.
#' @param show_progress logical. Whether a progress bar should be shown in the
#'   console. Default is FALSE.
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  [efa_fit()] fits run by the criteria that fit a model ([efa_hull()], [efa_kgc()],
#'  [efa_scree()], [efa_parallel()], [efa_nest()], and [efa_smt()]). `NULL` (default) uses the
#'  [efa_fit()] defaults. It only applies to criteria that fit a model, and only to the parts of
#'  that fit each criterion actually runs. [efa_cd()], [efa_ekc()], and [efa_map()] fit no model.
#'  [efa_kgc()], [efa_scree()], and [efa_parallel()] only fit one when their `eigen_type`
#'  includes `"EFA"`. [efa_smt()] fits with maximum likelihood by definition, so only
#'  `start_method` takes effect there. All fits are unrotated, so no rotation settings apply.
#' @param ... Further arguments passed to [efa_fit()] in
#' [efa_parallel()], [efa_kgc()], [efa_scree()], [efa_nest()], and (through its parallel
#' analysis and its own candidate fits) [efa_hull()]. An argument that [efa_fit()] does not
#' recognize causes an error. The estimation tuning knobs are not passed here; they live in
#' `estimate_control`. The standard-error arguments (`se`, `b_boot`, `ci`, `seed`) are not
#' accepted, because the criterion fits are internal steps whose standard errors are not
#' reported. Arguments listed after `...` must be given by their full name (R matches an
#' abbreviated name only against the arguments before `...`), so a tuning knob such as
#' `max_iter` cannot be mistaken for `max_iter_CD`.
#'
#' @details
#' By default, the entered data are checked for suitability for factor analysis
#' using the following methods (see the respective documentation for details):
#' \itemize{
#'   \item Bartlett's test of sphericity (see [efa_bartlett()])
#'   \item Kaiser-Meyer-Olkin criterion (see [efa_kmo()])
#' }
#'
#' The available factor retention criteria are the following (see the
#'  respective documentation for details):
#'  \itemize{
#'   \item Comparison data (see [efa_cd()])
#'   \item Empirical Kaiser criterion (see [efa_ekc()])
#'   \item Hull method (see [efa_hull()])
#'   \item Kaiser-Guttman criterion (see [efa_kgc()])
#'   \item Velicer's minimum average partial, MAP (see [efa_map()])
#'   \item Next Eigenvalue Sufficiency Test, NEST (see [efa_nest()])
#'   \item Parallel analysis (see [efa_parallel()])
#'   \item Scree plot (see [efa_scree()])
#'   \item Sequential chi-square model tests, RMSEA lower bound, and AIC
#'     (see [efa_smt()])
#' }
#'
#' The default `criteria` are comparison data, the empirical Kaiser criterion, the
#' Hull method, MAP, NEST, and parallel analysis. No single criterion is the most
#' accurate in all conditions. `efa_retain()` therefore runs several criteria
#' together, and the printed summary gives the range of their suggestions and the
#' most common one. Auerswald and Moshagen (2019) compare the criteria and give
#' guidance on the selection.
#'
#' The comparison data, parallel analysis, and NEST criteria compare the data
#' against simulated reference data, so their suggested numbers of factors vary
#' slightly from run to run. The Hull method also varies, because it calls
#' [efa_parallel()] to set its upper bound. Call [base::set.seed()] before
#' `efa_retain()` to make the results reproducible.
#'
#' @returns A list of class `c("efa_retain", "N_FACTORS")`, the trailing class
#'   keeping `inherits(x, "N_FACTORS")` working for code written against the
#'   superseded name. It contains
#' \item{suitability}{A list with the results from [efa_bartlett()] and
#'   [efa_kmo()] (`bartlett` and `kmo`), or `NULL` if
#'   `suitability = FALSE`.}
#' \item{outputs}{A named list with one `efa_retention` object per factor
#'   retention criterion that was run (see, e.g., [efa_ekc()]).}
#' \item{n_factors}{A named numeric vector with the suggested number of factors
#'   per criterion and, where a criterion has several variants, per variant
#'   (e.g. `EKC_BvA2017` or `PARALLEL_SMC`). Criteria without a numeric
#'   suggestion (the scree plot) are not included. The printed summary's "most common"
#'   value is based on each criterion's own most frequent (modal) suggestion among its
#'   variants, not a plain tally of this vector, so counting entries here by hand can
#'   give a different answer.}
#' \item{not_run}{A named character vector with the criteria that were skipped
#'   or failed and the reason, or `NULL` if all requested criteria ran.}
#' \item{settings}{A list of the settings used. Its `criteria` element records the
#'   requested criteria, in the order they were given, while `outputs` and `n_factors`
#'   are in the order in which the criteria were run. `gof` records the requested Hull
#'   goodness-of-fit indices, and `gof_used` records the ones the Hull method actually
#'   computed (it reduces them to `"CAF"` for the PAF estimator). `gof_used` is `NA`
#'   when HULL was not requested, was skipped, or failed.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @seealso [efa_screen()] for data screening before retention, and [efa_fit()] to extract
#'  the chosen number of factors.
#'
#' @family factor retention criteria
#'
#' @export
#'
#' @examples
#' \donttest{
#' # Default criteria, with correlation matrix and estimator "ML" (where needed)
#' # This will throw a warning for CD, as no raw data were specified
#' # The simulation-based criteria are seeded to make the run reproducible
#' set.seed(42)
#' nfac_all <- efa_retain(test_models$baseline$cormat, N = 500, estimator = "ML",
#'                        n_datasets = 100, n_datasets_nest = 100)
#'
#' # The same as above, but without "CD"
#' nfac_wo_CD <- efa_retain(test_models$baseline$cormat, criteria = c("EKC",
#'                          "HULL", "PARALLEL", "NEST"), N = 500,
#'                          estimator = "ML", n_datasets = 100,
#'                          n_datasets_nest = 100)
#'
#' # Use PAF instead of ML (this will take longer). PAF only supports "CAF" as
#' # gof for the Hull method, so set it explicitly to avoid the automatic message.
#' nfac_PAF <- efa_retain(test_models$baseline$cormat, criteria = c("EKC",
#'                        "HULL", "PARALLEL", "NEST"), N = 500,
#'                        estimator = "PAF", gof = "CAF", n_datasets = 100,
#'                        n_datasets_nest = 100)
#'
#' # Back to the default ML estimator (unlike above), with only "PCA" type eigenvalues
#' nfac_PCA <- efa_retain(test_models$baseline$cormat, criteria = c("EKC",
#'                        "HULL", "PARALLEL", "NEST"), N = 500,
#'                        estimator = "ML", eigen_type_other = "PCA",
#'                        n_datasets = 100, n_datasets_nest = 100)
#'
#' # Use raw data, such that CD can also be performed
#' nfac_raw <- efa_retain(GRiPS_raw, estimator = "ML", N_pop = 500,
#'                         N_samples = 20, n_datasets = 100,
#'                         n_datasets_nest = 100)
#'}
efa_retain <- function(x, criteria = c("CD", "EKC", "HULL", "MAP", "NEST", "PARALLEL"),
                       suitability = TRUE, N = NA,
                       use = c("pairwise.complete.obs", "all.obs",
                               "complete.obs", "everything", "na.or.complete"),
                       cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                       n_factors_max = NA, N_pop = 10000, N_samples = 500,
                       alpha = .30,
                       # `...` sits here deliberately, and must stay ahead of `max_iter_CD`: R
                       # matches an argument to a formal by unique prefix, but only for the
                       # formals BEFORE `...`. With `max_iter_CD` behind it, a bare `max_iter`
                       # can no longer be taken for it -- silently capping the comparison-data
                       # iterations instead of tuning the fits -- and lands in the dots, where
                       # the flat-knob guard below names the control object it belongs to. Every
                       # argument below must therefore be passed by its full name. Keep any new
                       # formal here too unless it can never be confused with a tuning knob.
                       ...,
                       max_iter_CD = 50, n_fac_theor = NA,
                       estimator = c("ML", "PAF", "ULS"),
                       gof = c("CAF", "CFI", "RMSEA"),
                       eigen_type_HULL = c("SMC", "PCA", "EFA"),
                       eigen_type_other = c("SMC"),
                       n_factors = 1, n_datasets = 1000,
                       percent = 95,
                       decision_rule = c("means", "percentile", "crawford"),
                       ekc_type = lifecycle::deprecated(),
                       n_datasets_nest = 1000, alpha_nest = .05,
                       show_progress = FALSE,
                       estimate_control = NULL){

  # Perform argument checks
  # Reject a flat tuning knob here rather than letting it travel into the criteria that fit a
  # model: each criterion is run under a tryCatch, so the knob would surface as an opaque
  # "criterion could not be run" instead of naming the control object it belongs to.
  .reject_flat_knobs(...names(), fn = "efa_retain")
  # A dot the criteria cannot forward to a fit is a misspelling; reject it here rather than
  # letting it be silently dropped (when the selected criteria run no fit) or surface as an
  # opaque "criterion could not be run". The criterion fits are always unrotated, so a
  # rotation setting is refused too (the N_FACTORS() wrapper's repacked rotate_control() object
  # is exempt -- see .reject_rotation_dots()), as are efa_fit()'s inference arguments
  # (`se`/`b_boot`/`ci`/`seed`), which a criterion fit computes and then discards -- see
  # .reject_inference_dots().
  .reject_unknown_fit_dots(...names(), fn = "efa_retain", unrotated = TRUE)
  .reject_rotation_dots(list(...), fn = "efa_retain")
  .assert_cor_input(x)
  .assert_estimate_control(estimate_control)

  ## Perform argument checks and prepare input
  criteria <- .match_arg_ci(criteria, several.ok = TRUE,
                            choices = c("CD", "EKC", "HULL", "KGC", "PARALLEL",
                                        "SCREE", "SMT", "NEST", "MAP"))
  suitability <- checkmate::assert_flag(suitability)
  eigen_type_HULL <- .match_arg_ci(eigen_type_HULL)
  eigen_type_other <- .match_arg_ci(eigen_type_other, several.ok = TRUE,
                                    choices = c("PCA", "SMC", "EFA"))
  gof <- .match_arg_ci(gof, several.ok = TRUE, choices = c("CAF", "CFI", "RMSEA"))
  cor_method <- .match_arg_ci(cor_method)
  use <- .match_arg_ci(use)
  estimator <- .match_arg_ci(estimator)
  decision_rule <- .match_arg_ci(decision_rule)
  if (lifecycle::is_present(ekc_type)) .deprecate_ekc_type("efa_retain(ekc_type)")
  checkmate::assert_number(alpha_nest, lower = 0, upper = 1)
  checkmate::assert_count(n_datasets_nest, na.ok = FALSE, positive = TRUE)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                             N_policy = "optional", inform_from_data = FALSE)
  R <- prep$R
  N <- prep$N

  ## Tests for suitability of factor analysis
  suitability_out <- NULL
  if (isTRUE(suitability)) {
    # Bartlett's test of sphericity needs N; the KMO criterion does not. When N
    # is unavailable (a correlation matrix without N), skip Bartlett's test with a
    # note and still report KMO, instead of aborting and thereby blocking the
    # N-free retention criteria too.
    bartlett_out <- if (is.na(N)) {
      cli::cli_warn(
        c("{.arg N} is {.val NA}; Bartlett's test of sphericity was skipped.",
          "i" = "Provide {.arg N} or raw data to include it."),
        class = "efa_suitability_no_n"
      )
      NULL
    } else {
      efa_bartlett(R, N = N, use = use, cor_method = cor_method)
    }
    suitability_out <- list(
      bartlett = bartlett_out,
      kmo = efa_kmo(R, use = use, cor_method = cor_method)
    )
  }

  ## Factor retention criteria, driven by the registry (run in registry order)
  ctl <- .n_factors_ctl(N = N, use = use, cor_method = cor_method,
                        n_factors_max = n_factors_max, N_pop = N_pop,
                        N_samples = N_samples, alpha = alpha,
                        max_iter_CD = max_iter_CD, n_fac_theor = n_fac_theor,
                        estimator = estimator, gof = gof,
                        eigen_type_HULL = eigen_type_HULL,
                        eigen_type_other = eigen_type_other,
                        n_factors = n_factors, n_datasets = n_datasets,
                        percent = percent, decision_rule = decision_rule,
                        n_datasets_nest = n_datasets_nest,
                        alpha_nest = alpha_nest,
                        estimate_control = estimate_control, dots = list(...))

  run <- intersect(names(.retention_registry), criteria)
  outputs <- list()
  not_run <- character(0)   # id -> reason it was skipped or failed
  # the first criterion error, chained into the all-failed abort below so the underlying
  # cause travels with it (a criterion that was skipped raises no condition to chain)
  first_cnd <- NULL

  for (id in run) {

    if (isTRUE(show_progress)) {
      # the message is a glue template evaluated when the bar is terminated -- by which time
      # `id` has advanced to the next criterion -- so build the string eagerly and freeze the
      # criterion name in it, or every completed step is labelled with its successor's name
      cli::cli_progress_step(paste0("Running ", id))
    }

    entry <- .retention_registry[[id]]

    if (isTRUE(entry$needs_raw) && .is_cormat(x)) {
      cli::cli_warn(
        c("{.arg x} is a correlation matrix, but {.val {id}} needs raw data.",
          "i" = "Skipping {.val {id}}."),
        class = "efa_criterion_skipped"
      )
      not_run[[id]] <- "needs raw data, but a correlation matrix was supplied"
      next
    }

    # Some criteria do not support polychoric/tetrachoric correlations (they
    # compare against continuous reference data, or their normal-theory test is
    # invalid for such correlations); skip them with a clear note rather than
    # letting them run on an inappropriate matrix.
    if (.is_poly_cor(cor_method) && isFALSE(entry$poly_ok)) {
      cli::cli_warn(
        c("{.val {id}} does not support {.code cor_method = {.val {cor_method}}}.",
          "i" = "Skipping {.val {id}}."),
        class = "efa_criterion_skipped"
      )
      not_run[[id]] <- paste0("does not support cor_method = \"", cor_method, "\"")
      next
    }

    # a failing criterion is excluded with a warning; the others still run
    out_id <- try(entry$fun(if (isTRUE(entry$needs_raw)) x else R, ctl),
                  silent = TRUE)

    if (inherits(out_id, "try-error")) {
      # `conditionMessage()` returns the message already wrapped to the console width, so its
      # first line is only the first *wrapped* line -- truncating the reason mid-sentence at
      # narrow widths. A cli condition keeps the unwrapped headline in `$message` (the bullets
      # live in `$body`), so take that and squish the newlines the source string carries.
      # A condition whose headline is absent or empty carries its text in `$body` instead, so
      # fall back to the formatted message there rather than reporting an empty reason.
      cnd <- attr(out_id, "condition")
      reason <- trimws(paste(cnd$message, collapse = " "))
      if (!nzchar(reason)) reason <- conditionMessage(cnd)
      reason <- gsub("\\s+", " ", trimws(reason))
      if (is.null(first_cnd)) first_cnd <- cnd
      cli::cli_warn(
        c("{.val {id}} could not be run and is excluded from the results.",
          "i" = "Error: {reason}"),
        class = "efa_criterion_failed"
      )
      not_run[[id]] <- reason
      next
    }

    outputs[[id]] <- out_id

  }

  if (isTRUE(show_progress)) {
    cli::cli_progress_done()
  }

  # nothing ran: a result with no criteria is not meaningful
  if (length(outputs) == 0) {
    cli::cli_abort(
      c("None of the requested factor retention criteria could be run.",
        "x" = "Could not run: {.val {names(not_run)}}.",
        "i" = "See the warnings above for the reason in each case."),
      class = "efa_no_criteria",
      # chains the first criterion error onto the abort, so the cause is reported with it
      # rather than only in the warning that has already scrolled past
      parent = first_cnd
    )
  }

  # Prepare settings here
  settings <- list(criteria = criteria,
                   suitability = suitability,
                   N = N,
                   use = use,
                   n_factors_max = n_factors_max,
                   N_pop = N_pop,
                   N_samples = N_samples,
                   alpha = alpha,
                   cor_method = cor_method,
                   max_iter_CD = max_iter_CD,
                   n_fac_theor = n_fac_theor,
                   estimator = estimator,
                   # back-compat alias, as for the frozen P_type key
                   method = estimator,
                   gof = gof,
                   # `gof` above is what was requested; the Hull method reduces it to CAF
                   # whenever the PAF estimator is used, so record what actually ran
                   # alongside it. NA when HULL was not among the criteria, was skipped, or
                   # failed -- no Hull goodness-of-fit index was computed in those cases.
                   gof_used = if (is.null(outputs[["HULL"]])) NA_character_
                              else outputs[["HULL"]]$settings$gof,
                   eigen_type_HULL = eigen_type_HULL,
                   eigen_type_other = eigen_type_other,
                   n_factors = n_factors,
                   n_datasets = n_datasets,
                   percent = percent,
                   decision_rule = decision_rule,
                   n_datasets_nest = n_datasets_nest,
                   alpha_nest = alpha_nest,
                   estimate_control = estimate_control)

  # Aggregate the suggested numbers of factors as "<id>" or "<id>_<variant>".
  # Visual criteria (the scree plot) make no numeric suggestion and are omitted;
  # for the others NA suggestions are kept (named), so a criterion that ran but
  # could not determine a number stays visible.
  n_factors_out <- unlist(lapply(names(outputs), function(id) {
    if (isTRUE(.retention_registry[[id]]$visual)) return(NULL)
    .retention_key(id, outputs[[id]]$n_factors)
  }))
  if (is.null(n_factors_out)) n_factors_out <- numeric(0)

  output <- list(suitability = suitability_out,
                 outputs = outputs,
                 n_factors = n_factors_out,
                 not_run = if (length(not_run) > 0) not_run else NULL,
                 settings = settings)

  # the trailing legacy class is load-bearing: it keeps `inherits(x, "N_FACTORS")`
  # working in code written against the superseded name. Do not drop it.
  class(output) <- c("efa_retain", "N_FACTORS")

  return(output)

}

#' Print method for efa_retain objects
#'
#' @param x an object of class efa_retain, returned by [efa_retain()].
#' @param ... not used.
#'
#' @returns `print()` returns its argument `x` invisibly; it is
#'   `cat(format(x), sep = "\n")`.
#'
#' @export
#' @method print efa_retain
#'
#' @examples
#' \donttest{
#' efa_retain(test_models$baseline$cormat, criteria = c("EKC", "SMT"), N = 500)
#' }
print.efa_retain <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' Format method for efa_retain objects
#'
#' @param x an object of class efa_retain, returned by [efa_retain()].
#' @param ... not used.
#'
#' @returns A character vector with the report lines (styled to the active
#'   console theme; plain when colours are disabled).
#'
#' @export
#' @method format efa_retain
#'
#' @examples
#' \donttest{
#' nf <- efa_retain(test_models$baseline$cormat, criteria = c("EKC", "SMT"),
#'                  N = 500)
#' writeLines(format(nf))
#' }
format.efa_retain <- function(x, ...) {
  cli::cli_format_method({

    if (!is.null(x$suitability)) {

      cli::cli_rule(left = "Tests for the suitability of the data for factor analysis")
      cli::cli_text("")

      bart <- x$suitability$bartlett
      pval <- bart$p_value
      if (!is.null(pval) && !is.na(pval)) {
        stats_text <- paste0("\u03c7\u00b2(", bart$df, ") = ",
                             round(bart$chisq, 2), ", p", .screen_p_str(pval), ".")
        if (.screen_is_sig(pval)) {
          cli::cli_bullets(c("v" = paste0(
            "The Bartlett's test of sphericity was significant at an alpha level of .05: ",
            stats_text,
            " These data are probably suitable for factor analysis.")))
        } else {
          cli::cli_bullets(c("x" = paste0(
            "The Bartlett's test of sphericity was not significant at an alpha level of .05: ",
            stats_text,
            " These data are probably not suitable for factor analysis.")))
        }
      } else {
        cli::cli_bullets(c("!" =
          "The Bartlett's test of sphericity did not render a result."))
      }

      kmo <- x$suitability$kmo$KMO
      if (!is.null(kmo) && !is.na(kmo)) {
        # Kaiser's verbal bands, shared with efa_kmo() and efa_screen() so that the same
        # value is never reported at two different severities
        band <- .kmo_band(kmo)
        kmo_text <- paste0("The Kaiser-Meyer-Olkin criterion is ", band$label,
                           " (KMO = ", round(kmo, 3), "). These data are ",
                           band$suitability, " suitable for factor analysis.")
        cli::cli_bullets(stats::setNames(kmo_text, band$symbol))
      } else {
        cli::cli_bullets(c("!" =
          "The overall KMO value for your data is not available."))
      }

      cli::cli_text("")
    }

    is_visual <- vapply(names(x$outputs),
                        function(id) isTRUE(.retention_registry[[id]]$visual),
                        logical(1))

    cli::cli_rule(left = "Suggested number of factors")

    # one line answering "how many factors, and how much do the criteria disagree?"
    # before the per-criterion detail; interpolated as a variable so that cli
    # substitutes the value instead of re-parsing it as a template
    summary_line <- .retention_summary(x$outputs[!is_visual])
    if (!is.null(summary_line)) {
      cli::cli_text("")
      cli::cli_text("{summary_line}")
    }

    # criteria with a numeric suggestion: one group with one bullet per variant
    # (a variant the criterion could not determine shows as "not applicable"),
    # separated by a blank line for readability
    for (out in x$outputs[!is_visual]) {
      cli::cli_text("")
      cli::cli_text("{out$criterion[['label']]}")
      # the criterion's own context line, where it has one (which eigenvalues, which
      # estimator): the bullets below are keyed by eigenvalue type or fit index, so
      # without it the aggregate report would not say what those keys refer to
      if (!is.null(out$subtitle)) {
        cli::cli_text("{out$subtitle}")
      }
      cli::cli_ul(.retention_bullets(out$results))
    }

    # visual criteria (e.g. the scree plot) are pointed to the plot
    for (out in x$outputs[is_visual]) {
      cli::cli_text("")
      label <- out$criterion[["label"]]
      cli::cli_bullets(c("i" =
        "{label} provides no numeric suggestion; inspect the plot."))
    }

    # criteria that were skipped or failed
    if (!is.null(x$not_run)) {
      cli::cli_text("")
      cli::cli_rule(left = "Criteria that could not be run")
      cli::cli_text("")
      for (id in names(x$not_run)) {
        reason <- x$not_run[[id]]
        cli::cli_bullets(c("!" = "{id}: {reason}"))
      }
    }

  })
}

# One-line summary of the criteria's suggestions for format.efa_retain(): how many
# suggestions how many criteria made, the range they span, and, when they disagree, the
# most common suggestion. The most common value is counted with ONE VOTE PER CRITERION --
# each criterion's own modal variant -- and never per variant, so that a criterion with
# several variants (e.g. the Hull method with three goodness-of-fit indices) does not
# outvote a single-variant one. A criterion whose variants have no single modal value is
# undecided and casts no vote. `outputs` are the criterion results with a numeric
# suggestion (the visual ones are excluded by the caller); returns NULL when none of them
# determined a number, or when a single suggestion would only restate the bullet below.
.retention_summary <- function(outputs) {

  # per criterion, the numbers it determined (a variant that came out NA determined none)
  per_crit <- lapply(outputs, function(o) o$n_factors[!is.na(o$n_factors)])
  per_crit <- per_crit[lengths(per_crit) > 0]
  if (length(per_crit) == 0) return(NULL)

  nf <- unlist(per_crit, use.names = FALSE)
  n_sug <- length(nf)
  n_crit <- length(per_crit)
  # there is nothing to summarise about a lone suggestion; the bullet below already
  # carries the number
  if (n_sug == 1L) return(NULL)

  counts <- paste0(n_sug, " ", .screen_plural(n_sug, "suggestion", "suggestions"),
                   " from ", n_crit, " ",
                   .screen_plural(n_crit, "criterion", "criteria"))

  if (min(nf) == max(nf)) {
    return(paste0(counts, ", all suggesting ", .retention_count(nf[1]), " ",
                  .screen_plural(nf[1], "factor", "factors"), "."))
  }

  # the most frequent value(s) of a vector, with how often the winner occurred; table()
  # names its cells in ascending numeric order, so tied values come out sorted
  modal <- function(v) {
    tab <- table(v)
    list(values = as.numeric(names(tab))[tab == max(tab)], n = max(tab))
  }
  votes <- unlist(lapply(per_crit, function(v) {
    m <- modal(v)
    if (length(m$values) == 1L) m$values else NULL
  }))

  spread <- paste0(counts, ", ranging from ", .retention_count(min(nf)), " to ",
                   .retention_count(max(nf)), " factors")
  if (is.null(votes)) return(paste0(spread, "."))

  # a most common value only means something once some value was chosen by more than one
  # criterion. Where every deciding criterion picked a different number there is none, and
  # listing them all under "most common" would read as agreement where there is the
  # maximum possible disagreement, so the range is left to speak for itself.
  best <- modal(votes)
  if (best$n == 1L) return(paste0(spread, "."))

  paste0(spread, " (most common: ",
         .screen_and_list(.retention_count(best$values)), ").")

}

#' Plot method for efa_retain objects
#'
#' Plots every factor-retention criterion in the [efa_retain()] result that has
#' a plottable outcome (see [plot.efa_retention()]). Criteria without a plot
#' (e.g. [efa_map()] or [efa_smt()]) are skipped.
#'
#' @param x an object of class efa_retain, returned by [efa_retain()].
#' @param ... not used.
#'
#' @returns A named list of [ggplot2::ggplot] objects, one per criterion with a
#'   plottable result, or invisibly `NULL` if there is none.
#'
#' @export
#' @method plot efa_retain
#'
#' @examples
#' \donttest{
#' nf <- efa_retain(test_models$baseline$cormat, criteria = c("EKC", "SMT"),
#'                  N = 500)
#' plot(nf)
#' }
plot.efa_retain <- function(x, ...) {

  # plot.efa_retention returns NULL (with a message) for criteria with no plot
  # (e.g. MAP/SMT) or a degenerate one (e.g. CD suggesting 0 factors); drop those
  plots <- lapply(x$outputs, function(o) suppressMessages(plot(o)))
  plots <- Filter(Negate(is.null), plots)

  if (length(plots) == 0) {
    cli::cli_inform("No plot is available for the criteria that were run.",
                    class = "efa_no_plot")
    return(invisible(NULL))
  }

  plots

}

# efa_retain aggregates the heavy retention criteria (CD's bootstrap simulation,
# PARALLEL's eigen Monte Carlo, HULL's per-factor refits), so the file-top
# `nf_grips` fixture dominates this file (~28s). Skipped by default; opt in with
# `Sys.setenv(EFATOOLS_TEST_SLOW = "true")`. See helper-slow.R.
if (is_slow_test()) {
set.seed(42)
nf_grips <- suppressMessages(suppressWarnings(efa_retain(GRiPS_raw)))
}  # is_slow_test()

test_that("output class and dimensions are correct", {
  skip_if_not_slow()
  expect_s3_class(nf_grips, c("efa_retain", "N_FACTORS"), exact = TRUE)
  expect_named(nf_grips, c("suitability", "outputs", "n_factors", "not_run",
                           "settings"))
  expect_type(nf_grips$outputs, "list")
  expect_type(nf_grips$settings, "list")
  expect_type(nf_grips$n_factors, "double")

  # default criteria, in registry order, each a unified retention object
  expect_named(nf_grips$outputs, c("CD", "EKC", "HULL", "MAP", "NEST",
                                   "PARALLEL"))
  for (out in nf_grips$outputs) {
    expect_s3_class(out, "efa_retention")
  }

  # all requested criteria ran (GRiPS is raw data, so CD is not skipped)
  expect_null(nf_grips$not_run)

  expect_s3_class(nf_grips$suitability$bartlett, "efa_bartlett")
  expect_s3_class(nf_grips$suitability$kmo, "efa_kmo")

  expect_named(nf_grips$n_factors,
               c("CD", "EKC_BvA2017", "HULL_CAF", "HULL_CFI", "HULL_RMSEA",
                 "MAP_TR2", "MAP_TR4", "NEST", "PARALLEL_SMC"))

  expect_named(nf_grips$settings, c("criteria", "suitability", "N", "use",
                                    "n_factors_max", "N_pop", "N_samples", "alpha",
                                    "cor_method", "max_iter_CD", "n_fac_theor",
                                    "estimator", "method", "gof", "gof_used",
                                    "eigen_type_HULL",
                                    "eigen_type_other", "n_factors", "n_datasets",
                                    "percent", "decision_rule", "n_datasets_nest",
                                    "alpha_nest", "estimate_control"))
})

test_that("the summary line counts one vote per criterion", {
  # `outputs` elements are efa_retention objects; only their $n_factors is read
  crit <- function(...) list(n_factors = c(...))

  # nothing to summarise: no criterion determined a number, or a lone suggestion is
  # already carried by the bullet printed below the line
  expect_null(.retention_summary(list(A = crit(A = NA_real_))))
  expect_null(.retention_summary(list(A = crit(A = 3))))

  expect_equal(.retention_summary(list(A = crit(A = 3), B = crit(B = 3))),
               "2 suggestions from 2 criteria, all suggesting 3 factors.")

  # a three-variant criterion casts one vote, so it cannot outvote two single-variant
  # criteria that agree
  expect_equal(
    .retention_summary(list(HULL = crit(HULL_CAF = 4, HULL_CFI = 4, HULL_RMSEA = 4),
                            EKC = crit(EKC = 2), MAP = crit(MAP = 2))),
    "5 suggestions from 3 criteria, ranging from 2 to 4 factors (most common: 2).")

  # that one vote is the criterion's own modal variant
  expect_equal(
    .retention_summary(list(HULL = crit(HULL_CAF = 2, HULL_CFI = 2, HULL_RMSEA = 5),
                            EKC = crit(EKC = 2))),
    "4 suggestions from 2 criteria, ranging from 2 to 5 factors (most common: 2).")

  # a criterion whose variants tie has no modal value and abstains: here 3 is the mode
  # of the per-variant vector, but only SMT decides, so no value wins more than one vote
  expect_equal(
    .retention_summary(list(MAP = crit(MAP_TR2 = 3, MAP_TR4 = 2),
                            KGC = crit(KGC_PCA = 3, KGC_SMC = 1),
                            SMT = crit(SMT_chi = 3, SMT_RMSEA = 3))),
    "6 suggestions from 3 criteria, ranging from 1 to 3 factors.")

  # every deciding criterion picked a different number: the range speaks for itself
  expect_equal(
    .retention_summary(list(A = crit(A = 1), B = crit(B = 2), C = crit(C = 3))),
    "3 suggestions from 3 criteria, ranging from 1 to 3 factors.")

  # tied winners among the voters are both reported
  expect_equal(
    .retention_summary(list(A = crit(A = 2), B = crit(B = 2),
                            C = crit(C = 4), D = crit(D = 4))),
    "4 suggestions from 4 criteria, ranging from 2 to 4 factors (most common: 2 and 4).")

  # a variant that could not determine a number contributes nothing to either count
  expect_equal(
    .retention_summary(list(A = crit(A_x = 2, A_y = NA_real_), B = crit(B = 3))),
    "2 suggestions from 2 criteria, ranging from 2 to 3 factors.")
})

test_that("the summary line renders its counts whatever options(scipen) is", {
  # An efa_retention object holds its counts as doubles, and a double is rendered in
  # scientific notation under a negative options(scipen) ("3e+00" instead of "3"). Both
  # branches of the line carry counts, so both are pinned.
  crit <- function(...) list(n_factors = c(...))
  withr::local_options(scipen = -5)

  expect_equal(.retention_summary(list(A = crit(A = 3), B = crit(B = 3))),
               "2 suggestions from 2 criteria, all suggesting 3 factors.")
  expect_equal(
    .retention_summary(list(A = crit(A = 2), B = crit(B = 2), C = crit(C = 4))),
    "3 suggestions from 3 criteria, ranging from 2 to 4 factors (most common: 2).")
})

test_that("settings record the Hull goodness-of-fit indices that actually ran", {
  skip_if_not_slow()
  cormat <- test_models$baseline$cormat

  # PAF supports only the CAF, so the Hull method reduces the requested set to it
  r_paf <- suppressMessages(
    efa_retain(cormat, N = 500, suitability = FALSE, criteria = "HULL", estimator = "PAF"))
  expect_equal(r_paf$settings$gof, c("CAF", "CFI", "RMSEA"))
  expect_equal(r_paf$settings$gof_used, "CAF")
  expect_named(r_paf$n_factors, "HULL_CAF")

  # with an estimator that supports all three, requested and used coincide
  r_ml <- efa_retain(cormat, N = 500, suitability = FALSE, criteria = "HULL",
                     estimator = "ML")
  expect_equal(r_ml$settings$gof_used, r_ml$settings$gof)

  # no HULL result to read it off
  r_no_hull <- efa_retain(cormat, N = 500, suitability = FALSE, criteria = "EKC")
  expect_identical(r_no_hull$settings$gof_used, NA_character_)
})

test_that("print and plot work on the aggregate object", {
  skip_if_not_slow()
  # the suitability block is rendered, including the Bartlett verdict and the
  # KMO label ladder (GRiPS has a marvellous KMO and a significant Bartlett test)
  txt <- paste(format(nf_grips), collapse = "\n")
  expect_match(txt, "Bartlett", fixed = TRUE)
  expect_match(txt, "significant", fixed = TRUE)
  expect_match(txt, "Kaiser-Meyer-Olkin", fixed = TRUE)
  expect_match(txt, "marvellous", fixed = TRUE)

  # one ggplot per plottable criterion that was run (MAP has no plot)
  p <- plot(nf_grips)
  expect_named(p, c("CD", "EKC", "HULL", "NEST", "PARALLEL"))
  for (p_i in p) {
    expect_s3_class(p_i, "ggplot")
  }
})

test_that("the aggregate report carries each criterion's own context line", {
  # The lines are compared verbatim, so pin the console width: the subtitle is longer
  # than a narrow console and would otherwise be wrapped across two of them.
  local_reproducible_output()

  # The bullets are keyed by eigenvalue type / fit index, so the group has to repeat
  # the criterion's subtitle -- otherwise the aggregate report shows a bare "PCA: 3"
  # with nothing saying what that key is. Criteria without a subtitle add no line.
  nf <- efa_retain(test_models$baseline$cormat, N = 500, suitability = FALSE,
                   criteria = c("EKC", "KGC"), eigen_type_other = c("PCA", "SMC"))
  txt <- format(nf)

  expect_true(any(txt == nf$outputs$KGC$subtitle))
  # and it sits between the criterion heading and that criterion's bullets
  expect_equal(which(txt == nf$outputs$KGC$subtitle),
               which(txt == "Kaiser-Guttman criterion") + 1)
  # EKC has no subtitle, so its heading is followed straight by its bullet
  expect_null(nf$outputs$EKC$subtitle)
})

test_that("format.efa_retain is the source of truth and honours the colour state", {
  skip_if_not_slow()
  # print() is exactly cat(format(x), sep = "\n"), so the two agree line for line.
  expect_identical(utils::capture.output(print(nf_grips)), format(nf_grips))

  old <- options(cli.num_colors = 256)
  on.exit(options(old), add = TRUE)

  # With colours on the report embeds ANSI ...
  expect_true(any(grepl("\033", format(nf_grips), fixed = TRUE)))

  # ... and with colours off it is plain.
  options(cli.num_colors = 1)
  expect_false(any(grepl("\033", format(nf_grips), fixed = TRUE)))
})

test_that("visual criteria and suitability = FALSE are handled", {
  nf_scree <- efa_retain(test_models$baseline$cormat, suitability = FALSE,
                        criteria = c("MAP", "SCREE"))

  expect_null(nf_scree$suitability)
  expect_named(nf_scree$outputs, c("MAP", "SCREE"))
  # the visual scree plot contributes no numeric suggestion
  expect_named(nf_scree$n_factors, c("MAP_TR2", "MAP_TR4"))

  txt <- paste(format(nf_scree), collapse = "\n")
  expect_match(txt, "no numeric suggestion", fixed = TRUE)
  expect_false(grepl("suitability of the data", txt, fixed = TRUE))
})

test_that("missing N skips Bartlett's test but still runs N-free criteria", {
  # A correlation matrix without N: Bartlett's test needs N and is skipped with a
  # classed warning, while KMO and the N-free retention criteria still run -
  # instead of the suitability check aborting the whole call.
  expect_warning(
    nf_noN <- efa_retain(test_models$baseline$cormat,
                        criteria = c("KGC", "MAP", "SCREE")),
    class = "efa_suitability_no_n"
  )
  expect_null(nf_noN$suitability$bartlett)
  expect_s3_class(nf_noN$suitability$kmo, "efa_kmo")
  expect_named(nf_noN$outputs, c("KGC", "MAP", "SCREE"), ignore.order = TRUE)
})

test_that("a failing criterion warns, is reported, and the others still run", {
  # EKC needs N, which is missing for a correlation matrix; MAP still runs
  expect_warning(
    nf_fail <- efa_retain(test_models$baseline$cormat, suitability = FALSE,
                         criteria = c("EKC", "MAP")),
    class = "efa_criterion_failed"
  )
  expect_named(nf_fail$outputs, "MAP")
  expect_named(nf_fail$n_factors, c("MAP_TR2", "MAP_TR4"))
  expect_named(nf_fail$not_run, "EKC")
  expect_match(paste(format(nf_fail), collapse = "\n"), "could not be run",
               fixed = TRUE)
})

test_that("a skipped criterion is reported in not_run", {
  # CD needs raw data and is skipped on a correlation matrix; pairing it with the cheap MAP
  # exercises the skip path without the simulating default criteria, so this runs on CRAN.
  expect_warning(
    nf_skip <- efa_retain(test_models$baseline$cormat, N = 500,
                          criteria = c("CD", "MAP")),
    class = "efa_criterion_skipped"
  )
  # CD needs raw data and is skipped on a correlation matrix
  expect_named(nf_skip$not_run, "CD")
  expect_false("CD" %in% names(nf_skip$outputs))
})

test_that("an all-failed run is a hard error", {
  # CD is the only requested criterion and is skipped on a correlation matrix
  expect_error(
    suppressWarnings(efa_retain(test_models$baseline$cormat, suitability = FALSE,
                               criteria = "CD")),
    class = "efa_no_criteria"
  )

  # when a criterion failed with an error rather than being skipped, its condition is
  # chained onto the abort, so the cause is reported with it
  err <- tryCatch(
    suppressWarnings(efa_retain(test_models$baseline$cormat, suitability = FALSE,
                                criteria = "EKC")),
    error = function(e) e
  )
  expect_s3_class(err, "efa_no_criteria")
  expect_s3_class(err$parent, "efa_n_required")
})

burt <- matrix(c(1.00,  0.83,  0.81,  0.80,   0.71, 0.70, 0.54, 0.53,  0.59,  0.24, 0.13,
                 0.83,  1.00,  0.87,  0.62,   0.59, 0.44, 0.58, 0.44,  0.23,  0.45,  0.21,
                 0.81,  0.87,  1.00,  0.63,   0.37, 0.31, 0.30, 0.12,  0.33,  0.33,  0.36,
                 0.80,  0.62,  0.63,  1.00,   0.49, 0.54, 0.30, 0.28,  0.42,  0.29, -0.06,
                 0.71,  0.59,  0.37,  0.49,   1.00, 0.54, 0.34, 0.55,  0.40,  0.19, -0.10,
                 0.70,  0.44,  0.31,  0.54,   0.54, 1.00, 0.50, 0.51,  0.31,  0.11,  0.10,
                 0.54,  0.58,  0.30,  0.30,   0.34, 0.50, 1.00, 0.38,  0.29,  0.21,  0.08,
                 0.53,  0.44,  0.12,  0.28,   0.55, 0.51, 0.38, 1.00,  0.53,  0.10, -0.16,
                 0.59,  0.23,  0.33,  0.42,   0.40, 0.31, 0.29, 0.53,  1.00, -0.09, -0.10,
                 0.24,  0.45,  0.33,  0.29,   0.19, 0.11, 0.21, 0.10, -0.09,  1.00,  0.41,
                 0.13,  0.21,  0.36, -0.06,  -0.10, 0.10, 0.08, -0.16, -0.10, 0.41,  1.00),
               nrow = 11, ncol = 11)

test_that("errors etc. are thrown correctly", {
  expect_error(efa_retain(1:10), class = "efa_input_not_matrix")
  expect_warning(efa_retain(GRiPS_raw, N = 10, criteria = "MAP",
                           suitability = FALSE), class = "efa_n_from_data")
  expect_error(efa_retain(sing_raw), class = "efa_cor_singular")
  # CD is skipped on a correlation matrix; a small criteria set keeps the skip-warning check
  # off the simulating default criteria so it runs on CRAN.
  expect_warning(efa_retain(test_models$baseline$cormat, N = 500,
                            criteria = c("CD", "MAP")), class = "efa_criterion_skipped")
  # burt is near-singular: it is smoothed, and PARALLEL finds every real SMC
  # eigenvalue above its reference (no crossing), so both warnings are expected.
  # PARALLEL's verdict comes out of a simulation, so seed it rather than inheriting
  # whatever state the preceding tests left (which differs between the default and the
  # slow gate, because the gated bodies return before drawing).
  set.seed(42)
  expect_warning(
    expect_warning(efa_retain(burt, N = 170, criteria = c("PARALLEL", "EKC")),
                   class = "efa_parallel_no_crossing"),
    class = "efa_cor_smoothed")
})

if (is_slow_test()) rm(nf_grips)
rm(burt)

## Promax rotation engine: build a varimax base, fit the oblique promax target,
## and reflect/order the solution. Dispatched from .rotate_model(), which resolves
## the `type` preset and guards the single-factor case. Promax reflects and orders
## its varimax base before the fit (ss_factors) or its pattern after the fit
## (eigen), so it finalizes its own solution rather than using .reflect_and_order().
.rotate_promax <- function(L, normalize, P_type, precision, order_type,
                           varimax_type, k) {

  dim_names <- dimnames(L)

  # perform the varimax rotation
  if (varimax_type == "svd") {
    AV <- .varimax_svd(L, normalize = normalize, precision = precision)
  } else if (varimax_type == "kaiser") {
    AV <- .VARIMAX_SPSS(L, normalize = normalize, precision = precision)
  }


  if (order_type == "ss_factors") {

    # reflect factors with negative sums
    signs <- .reflect_signs(AV$loadings)
    AV$loadings <- AV$loadings %*% diag(signs, nrow = length(signs))

    # reorder the factors according to largest sums of squares
    ss <- colSums(AV$loadings ^2)
    ss_order <- order(ss, decreasing = TRUE)

    AV$loadings <- AV$loadings[, ss_order]

    AV$rotmat <- (AV$rotmat %*% diag(signs))[, ss_order, drop = FALSE]

  }

  # store the loading matrix in a separate object for use
  A <- AV$loadings

  if (P_type == "unnorm") {

    # Promax target of Hendrickson and White (1964): P = sign(A) |A|^k. Written
    # without dividing by A -- the algebraically identical abs(A^(k + 1)) / A form
    # yields 0/0 = NaN for an exact-zero loading -- as stats::promax also does.
    P <- sign(A) * abs(A)^k

  } else if (P_type == "norm") {

    # SPSS-normalized target: row-normalize A by its row norm before raising to the
    # power, P = sign(A) |A / rowNorm|^k. The row norm is floored so a zero-
    # communality row (norm 0) stays 0 instead of producing 0/0 = NaN, mirroring the
    # Kaiser-weight floor in .VARIMAX_SPSS; an isolated zero loading in a non-zero
    # row already maps to 0 without the floor.
    rn <- sqrt(rowSums(A^2))
    rn[!is.finite(rn) | rn < 1e-15] <- 1
    P <- sign(A) * abs(A / rn)^k
  }

  # run the least squares fit
  fit <- stats::lm.fit(A, P)
  if (fit$rank < ncol(A)) {
    cli::cli_abort(
      c("The varimax base passed to promax is rank-deficient, so the oblique target fit is not identified.",
        "i" = "Inspect the unrotated loadings or extract fewer factors."),
      class = "efa_promax_rank_deficient"
    )
  }
  U <- fit$coefficients

  # rescale the transformation matrix
  D <- diag(solve(t(U) %*% U))
  U <- U %*% diag(sqrt(D))
  dimnames(U) <- NULL

  # compute the factor pattern matrix
  AP <- A %*% U

  # get rotation matrix and prepare output
  U <- AV$rotmat %*% U
  Ui <- solve(U)
  Phi <- Ui %*% t(Ui)

  if (order_type == "eigen") {
    # reflect factors with negative sums
    signs <- .reflect_signs(AP)
    AP <- AP %*% diag(signs, nrow = length(signs))
    Phi <- diag(signs) %*% Phi %*% diag(signs)

    # order the factors by their explained variance, largest first: the
    # Phi-weighted sum of squares diag(Phi L'L), which is the quantity reported as
    # the "SS loadings", so the reported variances decrease monotonically (as in
    # psych)
    ss_rotated <- diag(Phi %*% crossprod(AP))
    eig_order <- order(ss_rotated, decreasing = TRUE)
    AP <- AP[, eig_order, drop = FALSE]
    Phi <- Phi[eig_order, eig_order]

    # the rotation matrix follows the same sign reflection and factor ordering as
    # the pattern, so L_unrot %*% rotmat reproduces the rotated loadings (the
    # ss_factors branch reorders its varimax base before the fit, so U is already
    # in order there)
    U <- (U %*% diag(signs))[, eig_order, drop = FALSE]

  }

  # the factor columns are labelled "F1".."Fk" by their position in the ordered
  # pattern: promax mixes all k factors, so a rotated column corresponds to no single
  # unrotated factor, and permuting the unrotated labels along with the reordering
  # above would only attach an arbitrary permutation to a stable solution
  dimnames(AP) <- dim_names
  # label the factor intercorrelations with the same factor names as the pattern, so
  # every rotation returns labelled Phi/Structure columns
  dimnames(Phi) <- list(colnames(AP), colnames(AP))

  vars_accounted_rot <- .compute_vars(L_unrot = L, L_rot = AP, Phi = Phi)

  colnames(vars_accounted_rot) <- colnames(AP)

  # get structure matrix
  Structure <- AP %*% Phi
  dimnames(Structure) <- list(rownames(AP), colnames(AP))

  # prepare and return output list
  class(AP) <- c("efa_loadings", "LOADINGS")
  class(Structure) <- c("efa_loadings", "LOADINGS")

  list(rot_loadings = AP,
       Phi = Phi,
       Structure = Structure,
       rotmat = U,
       vars_accounted_rot = vars_accounted_rot)
}

#' Print and format a multigroup factor analysis
#'
#' `print()` turns an [efa_group()] result into a sectioned report: a header recapping
#' the groups, the common number of factors, the estimator, the rotation, and the
#' alignment; a group-pair table of the matched Tucker congruences between the aligned
#' loadings; a per-pair summary of the cross-group loading differences (with the salient
#' and, when a bootstrap was run, the confidence-interval flags); and, when
#' `invariance = TRUE`, a group-pair by factor grid of the approximate-invariance verdicts.
#' `format()` assembles the same report and returns it as a character vector; `print()` is
#' `cat(format(x), sep = "\n")`. The lines follow the active console theme, so they are
#' plain when colours are disabled (for example when captured into a file or stripped with
#' [cli::ansi_strip()]). `print()` does not draw a plot; use [plot.efa_group()].
#'
#' @param x An object of class `efa_group` (output from [efa_group()]).
#' @param digits Integer. The number of decimal places the reported values are rounded to.
#'   Default is 3.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a character
#'   vector with the report lines.
#'
#' @family factor analysis
#'
#' @export
#' @method print efa_group
#'
#' @examples
#' g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
#' mg <- efa_group(GRiPS_raw, groups = g, n_factors = 1)
#' mg
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(mg))
#'
print.efa_group <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_group
#' @export
#' @method format efa_group
format.efa_group <- function(x, digits = 3, ...) {

  settings <- x$settings
  groups <- settings$groups
  m <- length(groups)
  k <- settings$n_factors
  # One row per group pair, in the same order as $diffs / $flags / $invariance.
  pairs <- paste(x$diffs$group_1, "vs", x$diffs$group_2)
  fac_names <- dimnames(x$congruence$matched)[[3L]]
  n_boot <- x$congruence$n_boot
  boot <- !is.null(n_boot)

  cli::cli_format_method({

    # -- Header ----------------------------------------------------------------
    cli::cli_rule(left = "{.strong Multigroup exploratory factor analysis}")
    cli::cli_text("")
    rot_txt <- if (is.null(settings$rotation) || identical(settings$rotation, "none")) {
      "unrotated"
    } else {
      paste0(settings$rotation, " rotation")
    }
    # Emitted verbatim so arbitrary group names are never interpreted as cli/glue markup.
    .efa_group_verbatim(paste0(
      m, " groups (", paste(groups, collapse = ", "), ") \u00b7 ",
      k, if (k == 1L) " factor" else " factors", " \u00b7 ",
      settings$estimator, " extraction \u00b7 ", rot_txt))
    .efa_group_verbatim(if (identical(settings$alignment, "consensus")) {
      "Aligned to a symmetric consensus target."
    } else {
      paste0("Aligned to reference group ", settings$reference_group, ".")
    })
    .efa_group_verbatim(paste0(
      "N: ", paste0(names(settings$N), " = ",
                    format(settings$N, trim = TRUE), collapse = ", ")))

    # -- Factor congruence -----------------------------------------------------
    .print_efa_rule("Factor congruence")
    cli::cli_text("Tucker's congruence between the aligned group loadings (matched factors):")
    cli::cli_text("")

    phi_mat <- matrix(NA_real_, nrow = length(pairs), ncol = k,
                      dimnames = list(pairs, fac_names))
    r <- 0L
    for (g in seq_len(m - 1L)) {
      for (h in (g + 1L):m) {
        r <- r + 1L
        phi_mat[r, ] <- x$congruence$matched[g, h, ]
      }
    }
    .efa_emit_lines(.efa_format_matrix(
      values = phi_mat, row_labels = pairs, col_labels = fac_names,
      col_roles = rep("corr", k), digits = digits))

    if (any(x$congruence$degenerate)) {
      cli::cli_text("")
      cli::cli_alert_info(
        "{.val NA}: a near-zero factor left the congruence undefined for that pair (see {.code $congruence$degenerate}).",
        wrap = TRUE)
    }
    if (boot) {
      cli::cli_text("")
      cli::cli_alert_info(
        "{round(settings$ci * 100)}% percentile bootstrap CIs ({n_boot} replicate{?s}) are in {.code $congruence$matched_ci}.",
        wrap = TRUE)
    }

    # -- Loading differences ---------------------------------------------------
    .print_efa_rule("Loading differences")
    dstr <- .efa_group_delta_str(settings$delta)
    cli::cli_text("Absolute loading differences by group pair (salience threshold |\u0394| \u2265 {dstr}):")
    cli::cli_text("")

    pk <- nrow(x$loadings[[1L]]) * k
    d <- x$diffs
    diff_cells <- cbind(
      .efa_num(d$mean_abs_diff, digits = digits, pad = FALSE),
      .efa_num(d$median_abs_diff, digits = digits, pad = FALSE),
      .efa_num(d$min_abs_diff, digits = digits, pad = FALSE),
      .efa_num(d$max_abs_diff, digits = digits, pad = FALSE),
      .efa_num(d$rmse, digits = digits, pad = FALSE),
      paste0(d$n_flagged, "/", pk))
    .efa_emit_lines(.efa_group_table(
      pairs, c("mean", "median", "min", "max", "rmse", "flagged"), diff_cells))

    if (boot) {
      n_exc <- sum(x$flags$ci_excludes_0, na.rm = TRUE)
      if (n_exc > 0L) {
        cli::cli_text("")
        cli::cli_alert_info(
          "{n_exc} loading{?s} {?has/have} a bootstrap CI excluding 0 (see {.code $flags}).",
          wrap = TRUE)
      }
    }

    # -- Approximate invariance (only when requested) --------------------------
    if (!is.null(x$invariance)) {
      .print_efa_rule("Approximate invariance")
      if (boot) {
        cli::cli_text("Congruence bands (Lorenzo-Seva & ten Berge, 2006), read off the CI lower bound:")
      } else {
        cli::cli_text("Congruence bands (Lorenzo-Seva & ten Berge, 2006: \u2265 .95 equal, \u2265 .85 fair):")
      }
      cli::cli_text("")

      # $invariance is enumerated in the same pair-major, factor-minor order as
      # `pairs` x `fac_names` (all follow the shared g < h enumeration over the groups),
      # so its verdicts fill the grid positionally -- as the congruence table above does --
      # rather than by re-joining and matching group-name strings.
      verdicts <- matrix(x$invariance$verdict, nrow = length(pairs), ncol = k,
                         byrow = TRUE, dimnames = list(pairs, fac_names))
      verdicts[is.na(verdicts)] <- "NA"
      .efa_emit_lines(.efa_group_table(
        pairs, fac_names, verdicts, style = .efa_group_verdict_style))

      # Tucker's congruence is invariant to a proportional rescaling of a factor's
      # loadings, so uniformly stronger loadings in one group are still graded "equal"
      # -- a verdict grid that then sits directly above a difference table contradicting
      # it. Point at the differences when that gap is actually present: every factor
      # graded equal, yet some pair's *average* absolute difference already reaches the
      # threshold set for a single salient cell. `delta = 0` flags every cell by
      # construction and so says nothing about the size of the gap.
      max_mad <- max(d$mean_abs_diff)
      if (settings$delta > 0 && !anyNA(x$invariance$verdict) &&
          all(x$invariance$verdict == "equal") && max_mad >= settings$delta) {
        # At least two decimals whatever `digits` is: a difference that rounds to "0"
        # would contradict the sentence it is quoted in.
        mad_str <- .efa_num(max_mad, digits = max(digits, 2L), pad = FALSE)
        cli::cli_text("")
        cli::cli_alert_info(
          "Every factor is graded {.val equal} while the aligned loadings differ by up to {mad_str} on average: Tucker's congruence is invariant to a proportional rescaling of a factor's loadings, so read the verdicts alongside {.code $diffs}.",
          wrap = TRUE)
      }
    }

  })
}

# Render a small labelled table (left-aligned row labels, centred bold headers,
# right-aligned cells) into console lines, following the column maths of
# .efa_format_matrix() without its role machinery. `cells` is an already-formatted
# character matrix; `style(padded, raw)` optionally restyles a cell (the raw, unpadded
# value drives the styling; the padded cell keeps the column aligned). A table wider
# than the console is split into vertically stacked column blocks by the same helper
# .efa_format_matrix() uses; there are no auxiliary columns here, so every block repeats
# only the row labels and its own header row.
.efa_group_table <- function(row_labels, headers, cells, style = NULL) {
  cells <- as.matrix(cells)
  n_col <- ncol(cells)

  row_width <- max(cli::ansi_nchar(row_labels, type = "width"), 1L)
  col_widths <- vapply(seq_len(n_col), function(j) {
    max(cli::ansi_nchar(headers[j], type = "width"),
        max(cli::ansi_nchar(cells[, j], type = "width")))
  }, integer(1L))

  blocks <- .efa_matrix_blocks(seq_len(n_col), integer(0), row_width, col_widths)

  out <- character(0)
  for (bb in seq_along(blocks)) {
    cols <- blocks[[bb]]

    header_cells <- vapply(cols, function(j) {
      as.character(cli::ansi_align(cli::style_bold(headers[j]),
                                   width = col_widths[j], align = "center"))
    }, character(1L))
    header <- paste0(strrep(" ", row_width), "  ",
                     paste(header_cells, collapse = "  "))

    body <- vapply(seq_len(nrow(cells)), function(i) {
      row_cells <- vapply(cols, function(j) {
        padded <- as.character(cli::ansi_align(cells[i, j], width = col_widths[j],
                                               align = "right"))
        if (is.null(style)) padded else style(padded, cells[i, j])
      }, character(1L))
      label <- as.character(cli::ansi_align(row_labels[i], width = row_width,
                                            align = "left"))
      paste0(label, "  ", paste(row_cells, collapse = "  "))
    }, character(1L))

    # A blank line separates stacked blocks; the repeated header row names their columns.
    out <- c(out, if (bb > 1L) "", header, body)
  }

  sub("[ ]+$", "", out)
}

# Emit a header line verbatim, wrapped to the console width with continuation lines
# indented: cli_verbatim() keeps arbitrary group names out of cli/glue markup, but it also
# does no wrapping of its own. strwrap() breaks *below* its `width`, so it is given one
# column more than the console has in order to fill it exactly. It wraps on display width
# (so wide characters are counted correctly) but it also normalises runs of whitespace and
# breaks only at spaces, so a group name carrying doubled spaces prints with single ones
# and a name longer than the console still overruns it.
.efa_group_verbatim <- function(txt) {
  cli::cli_verbatim(strwrap(txt, width = .efa_console_width() + 1L, exdent = 2L))
}

# Colour an approximate-invariance verdict cell by its band (a no-op when colour is off,
# since .efa_style's cli primitives are then no-ops).
.efa_group_verdict_style <- function(padded, raw) {
  colour <- switch(trimws(raw), equal = "green", fair = "yellow",
                   incongruent = "red", NULL)
  if (is.null(colour)) padded else .efa_style(padded, colour)
}

# Format the salience threshold for display: locale-independent (sprintf always uses ".",
# unlike format() under a comma OutDec) with the leading zero dropped, matching the
# correlation-scale numbers in the tables (which flow through .efa_num / sprintf). Shared by
# the print header and the differences-plot subtitle so the two show the threshold alike.
.efa_group_delta_str <- function(delta) sub("^(-?)0\\.", "\\1.", sprintf("%g", delta))

#' Rotate a loading matrix to a target using Procrustes alignment
#'
#' `efa_procrustes()` aligns one loading matrix to a target loading matrix with the
#' same dimensions. It is used internally by `efa_mi()`, but can also be used
#' directly when factor columns must be brought into a common orientation before
#' averaging or comparing solutions.
#'
#' For `rotation = "orthogonal"`, the function solves the closed-form orthogonal
#' Procrustes problem
#'
#' \deqn{\min_T \frac{1}{2}\|A T - B\|_F^2 \quad \textrm{subject to}\quad T'T = I,}
#'
#' where `A` is the loading matrix and `B` is `Target`.
#'
#' For `rotation = "oblique"`, the function calls the compiled
#' `.oblique_procrustes()` optimizer. The oblique convention is the same as in
#' `GPArotation::targetQ()`:
#'
#' \deqn{L = A T^{-T}, \qquad \Phi = T'T, \qquad diag(\Phi) = 1.}
#'
#' By default the oblique solver is warm-started from the closed-form orthogonal
#' Procrustes solution, which resolves the factor permutation and sign
#' indeterminacy and avoids the poor local minima an identity start can fall
#' into. Supply `T_init` to override this start. Random starts are only used for
#' oblique alignment. For one-factor models, oblique and orthogonal alignment are
#' equivalent, so the function uses the stable one-factor orthogonal solution
#' instead of calling the oblique optimizer.
#'
#' @param A Numeric loading matrix to be aligned.
#' @param Target Numeric target matrix with the same dimensions as `A`.
#' @param rotation Character string, either `"orthogonal"` or `"oblique"`.
#' @param S Optional `k x k` cross-product matrix `crossprod(A)`, kept for
#'   compatibility. It enters both the oblique criterion and its gradient, so any
#'   other matrix would minimize a different criterion: where `S` is used it is
#'   checked against `crossprod(A)` and must agree with it up to a relative
#'   tolerance of `1e-8`. That check forms `crossprod(A)` itself, so passing `S`
#'   no longer avoids any work: omitting it gives the same result for slightly
#'   less. `S` is used, and therefore checked, only on the oblique path with more
#'   than one factor and `oblique_normalize = FALSE`; if Kaiser normalization is
#'   requested, the cross-product must be recomputed on the normalized matrix and
#'   `S` is ignored.
#' @param T_init Optional `k x k` starting transformation matrix for the oblique
#'   solver. Its columns are normalized internally, and the normalized matrix must
#'   be well enough conditioned to define a proper factor correlation matrix: its
#'   smallest singular value must be at least `1e-4`, the same floor the solver
#'   applies to every candidate it evaluates. If `NULL` (the default), the oblique
#'   solver is warm-started from the closed-form orthogonal Procrustes solution.
#' @param oblique_eps Positive convergence tolerance for the projected-gradient
#'   norm in the oblique solver.
#' @param oblique_maxit Non-negative integer. Maximum number of projected-gradient
#'   updates in the full oblique solver.
#' @param oblique_max_line_search Non-negative integer. Maximum number of
#'   step-halving attempts after the initial line-search step.
#' @param oblique_step0 Positive initial step size for the oblique solver.
#' @param oblique_normalize Logical; if `TRUE`, apply Kaiser row normalization to
#'   the loadings (only) in the oblique solver and back-transform the aligned
#'   loadings afterwards, leaving `Target` unnormalized (as in
#'   `GPArotation::targetQ(normalize = TRUE)`).
#' @param oblique_random_starts Non-negative integer. Number of additional random
#'   starts used by the oblique solver.
#' @param oblique_screen_keep Non-negative integer. Number of random starts
#'   retained after cheap objective screening and sent to triage optimization.
#' @param oblique_triage_maxit Non-negative integer. Number of short optimization
#'   iterations used in the triage stage.
#' @param oblique_triage_improve_tol Non-negative scalar. Relative improvement
#'   required for a triaged start to be promoted to full optimization.
#'
#' @returns A list. Every path returns the following components:
#'
#' \item{loadings}{Aligned loading matrix.}
#' \item{T}{Transformation matrix.}
#' \item{Phi}{Factor intercorrelation matrix; the identity for orthogonal and
#'   one-factor alignment.}
#' \item{value}{Target criterion at the returned solution.}
#' \item{convergence}{Logical; `TRUE` for the closed-form orthogonal solution.}
#' \item{valid}{Logical; whether the transformation defines an admissible `Phi`.}
#' \item{iterations}{Number of solver iterations; `0` for the closed-form
#'   orthogonal solution.}
#' \item{kappa_T}{Condition number of `T`; a constant `1` on the orthogonal path.}
#' \item{Table}{Iteration history with columns `iter`, `f`, `log10_s`, and `step`;
#'   a single placeholder row on the orthogonal path.}
#' \item{method}{`"orthogonal_procrustes"`, `"oblique_procrustes"`, or
#'   `"single_factor_procrustes"` for a one-factor oblique request.}
#' \item{line_search_failed}{Logical line-search diagnostic.}
#' \item{best_start_index, all_start_indices, all_values, all_converged,
#'   all_iterations}{Multi-start summary of the starts that were fully
#'   optimized; each has a single entry when no random starts were used.}
#'
#' The oblique solver additionally returns `screen_start_indices` and
#' `screen_values` (the starts kept by cheap objective screening and their
#' criterion values) together with the counts `n_random_starts`, `n_screened`,
#' `n_triaged`, and `n_fully_optimized`. These six components are absent for
#' `rotation = "orthogonal"` and for one-factor models, which are aligned with the
#' orthogonal solution.
#'
#' Row and column names are preserved where possible. When
#' `oblique_normalize = TRUE` the returned `loadings` are back-transformed to the
#' original scale, but `value` is the criterion on the Kaiser-normalized loadings,
#' so it is not `0.5 * sum((loadings - Target)^2)`.
#'
#' @family factor rotation
#'
#' @export
#'
#' @examples
#' ## Align an estimated loading matrix to a known target pattern: fit an
#' ## unrotated three-factor model, then rotate its loadings toward the true
#' ## population pattern.
#' efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "none")
#' target <- population_models$loadings$baseline
#'
#' ## Orthogonal target rotation (rigid rotation/reflection):
#' efa_procrustes(efa_mod$unrot_loadings, target, rotation = "orthogonal")
#'
#' ## Oblique target rotation (lets the aligned factors correlate):
#' efa_procrustes(efa_mod$unrot_loadings, target, rotation = "oblique")
#'
efa_procrustes <- function(A,
                           Target,
                           rotation = c("orthogonal", "oblique"),
                           S = NULL,
                           T_init = NULL,
                           oblique_eps = 1e-5,
                           oblique_maxit = 1000,
                           oblique_max_line_search = 10,
                           oblique_step0 = 1,
                           oblique_normalize = FALSE,
                           oblique_random_starts = 0,
                           oblique_screen_keep = 2,
                           oblique_triage_maxit = 25,
                           oblique_triage_improve_tol = 0) {
  rotation <- .match_arg_ci(rotation)

  mats <- .procrustes_validate_matrix_pair(A, Target)
  A <- mats$A
  Target <- mats$B
  k <- ncol(A)

  if (rotation == "orthogonal" || k == 1L) {
    out <- .orthogonal_procrustes(A, Target)
    if (rotation == "oblique" && k == 1L) {
      out$method <- "single_factor_procrustes"
    }
    return(out)
  }

  controls <- .procrustes_validate_oblique_controls(
    oblique_eps = oblique_eps,
    oblique_maxit = oblique_maxit,
    oblique_max_line_search = oblique_max_line_search,
    oblique_step0 = oblique_step0,
    oblique_normalize = oblique_normalize,
    oblique_random_starts = oblique_random_starts,
    oblique_screen_keep = oblique_screen_keep,
    oblique_triage_maxit = oblique_triage_maxit,
    oblique_triage_improve_tol = oblique_triage_improve_tol
  )

  # Under Kaiser normalization the solver works on the row-weighted loadings and
  # recomputes the cross-product itself, so a supplied `S` is neither consumed nor
  # checked there; on the orthogonal and one-factor paths the function has already
  # returned above.
  S_arg <- if (is.null(S) || controls$oblique_normalize) {
    NULL
  } else {
    .procrustes_validate_crossprod(S, k, A)
  }
  T_arg <- if (is.null(T_init)) NULL else .procrustes_validate_t_init(T_init, k)

  # Warm-start the oblique solver from the closed-form orthogonal Procrustes
  # solution when the user supplies no start. The identity start can be trapped
  # in poor local minima on well-conditioned problems (and trapped solutions
  # still report convergence), whereas the orthogonal solution resolves the
  # factor permutation and sign structure and lands in the basin of the global
  # oblique optimum. The consensus engine warm-starts its inner alignments the
  # same way. An explicit T_init overrides this.
  if (is.null(T_arg)) {
    if (controls$oblique_normalize) {
      # Compute the start in the same Kaiser row-normalized space the compiled
      # solver optimizes in: the loadings are row-weighted, the target is left raw
      # (mirrors the row weighting in .oblique_procrustes()).
      w <- sqrt(rowSums(A^2))
      w[!is.finite(w) | w < 1e-15] <- 1
      T_arg <- .procrustes_orthogonal_T(A / w, Target)
    } else {
      T_arg <- .procrustes_orthogonal_T(A, Target)
    }
  }

  out <- .oblique_procrustes(
    A = A,
    B = Target,
    S_r = S_arg,
    T_init_r = T_arg,
    eps = controls$oblique_eps,
    maxit = controls$oblique_maxit,
    max_line_search = controls$oblique_max_line_search,
    step0 = controls$oblique_step0,
    normalize = controls$oblique_normalize,
    random_starts = controls$oblique_random_starts,
    screen_keep = controls$oblique_screen_keep,
    triage_maxit = controls$oblique_triage_maxit,
    triage_improve_tol = controls$oblique_triage_improve_tol
  )

  out$method <- "oblique_procrustes"
  .procrustes_apply_dimnames(out, A, Target)
}


#' Generalized Procrustes Analysis consensus target across loading matrices
#'
#' Internal helper that constructs a Generalized Procrustes Analysis (GPA)
#' consensus target across a list of loading matrices and returns the aligned
#' loadings, the centroid target, and convergence diagnostics. Used by
#' [efa_mi()] under `target_method = "consensus"` to build a common
#' rotation target across imputations. Oblique rotations are not supported
#' here: the iteration is degenerate for oblique transforms with more than
#' one factor (cf. Lorenzo-Seva & Van Ginkel 2016, who use a Promin step on
#' top of the centroid rather than iterated oblique Procrustes); callers
#' should pass the unrotated solutions of an orthogonal rotation, or use
#' `target_method = "first_target"`.
#'
#' The iteration alternates two steps:
#'
#' 1. each loading matrix is aligned to the current target with `efa_procrustes()`;
#' 2. the target is updated to the elementwise centroid of the aligned matrices.
#'
#' The outer loop stops when the target stabilises, when the consensus loss
#' stabilises, or when both criteria are satisfied.
#'
#' If `multi_start = FALSE`, one consensus run is performed. If
#' `multi_start = TRUE`, the same engine is repeated for the selected starting
#' targets and the run with the smallest final mean loss is returned as the
#' main result; all runs and a between-run congruence summary are retained in
#' the `multi_start` component.
#'
#' @param unrotated_list List of unrotated loading matrices to be aligned. All
#'   matrices must be numeric, finite, and have identical dimensions.
#' @param init_targets Optional list of starting target matrices. These are
#'   typically rotated loading matrices from the corresponding analyses. If
#'   `NULL`, `unrotated_list` is used.
#' @param rotation Character string, either `"orthogonal"` or `"oblique"`.
#' @param start Either a single integer selecting an element of `init_targets`,
#'   or an explicit target matrix. Used when `multi_start = FALSE`.
#' @param multi_start Logical. If `FALSE`, perform one consensus-target run. If
#'   `TRUE`, repeat the single-start algorithm for each element of `starts`.
#' @param starts Integer vector selecting elements of `init_targets` used as
#'   starting targets when `multi_start = TRUE`. If `NULL`, all elements of
#'   `init_targets` are used. Duplicate entries are removed.
#' @param tol Positive relative Frobenius-norm convergence tolerance for the
#'   outer target update.
#' @param loss_tol Positive tolerance for the relative change in the outer
#'   consensus loss. If `NULL`, loss-based convergence is disabled. It cannot be
#'   `NULL` when `convergence` is `"loss"` or `"both"`.
#' @param loss_patience Positive integer. Number of consecutive iterations with
#'   relative loss change below `loss_tol` required for loss-based convergence.
#' @param convergence Character string controlling the stopping rule. `"either"`
#'   stops when either target or loss convergence is satisfied; `"target"` uses
#'   only target change; `"loss"` uses only loss change; `"both"` requires both.
#' @param min_iter Non-negative integer. Minimum number of outer iterations
#'   before convergence can be declared.
#' @param max_iter Positive integer. Maximum number of outer consensus
#'   iterations.
#' @param alpha Damping factor for the target update. `alpha = 1` uses the full
#'   centroid update. Smaller values, such as `0.5`, can reduce oscillation.
#' @param match_target Logical. If `TRUE`, the updated centroid is signed and
#'   column-matched to the previous target before convergence is evaluated.
#' @param hyper_cutoff Non-negative cutoff used by `.hyperplane_count()` for
#'   summary output.
#' @param verbose Logical; if `TRUE`, print convergence messages for the outer
#'   loop.
#'
#' @returns A list with the converged target, aligned matrices, pooled loadings,
#'   pooled `Phi`, convergence history, inner-alignment diagnostics, and
#'   hyperplane summaries. If `multi_start = TRUE`, the `multi_start` element also
#'   contains the per-start losses, convergence indicators, run summaries, all
#'   run objects, and between-run Tucker congruence matrices.
#'
#' @references
#' Gower, J. C. (1975). Generalized Procrustes analysis. *Psychometrika*, 40,
#' 33-51.
#'
#' Van Ginkel, J. R., & Kroonenberg, P. M. (2014). Using Generalized
#' Procrustes Analysis for Multiple Imputation in Principal Component
#' Analysis. *Journal of Classification*, 31, 242-269.
#'
#' Lorenzo-Seva, U., & Van Ginkel, J. R. (2016). Multiple Imputation of
#' missing values in exploratory factor analysis of multidimensional scales:
#' estimating latent trait scores. *Anales de Psicologia*, 32, 596-608.
#'
#' @keywords internal
.gpa_consensus_target <- function(unrotated_list,
                                 init_targets = NULL,
                                 rotation = c("orthogonal", "oblique"),
                                 start = 1,
                                 multi_start = FALSE,
                                 starts = NULL,
                                 tol = 1e-3,
                                 loss_tol = 1e-6,
                                 loss_patience = 5,
                                 convergence = c("either", "target", "loss", "both"),
                                 min_iter = 2,
                                 max_iter = 200,
                                 alpha = 1,
                                 match_target = TRUE,
                                 hyper_cutoff = 0.15,
                                 verbose = FALSE) {
  rotation <- match.arg(rotation)
  convergence <- match.arg(convergence)
  multi_start <- .procrustes_check_flag(multi_start, "multi_start")

  # The iteration is well-behaved for orthogonal Procrustes (closed-form per
  # step; classical Gower 1975 GPA) but the extra column-skew freedom of an
  # oblique transform lets the centroid loss collapse to degenerate targets
  # for k > 1; Lorenzo-Seva & Van Ginkel (2016) sidestep this by running a
  # Promin target rotation on top of the centroid, which this engine does
  # not implement. NCOL() treats NULL/vectors/1-D arrays as a single column
  # so the guard fires only on genuinely multi-column inputs.
  if (rotation == "oblique" && is.list(unrotated_list) &&
      length(unrotated_list) >= 1L &&
      isTRUE(NCOL(unrotated_list[[1L]]) >= 2L)) {
    cli::cli_abort(
      c("GPA-consensus alignment does not support oblique rotations with more than one factor.",
        "i" = "Use {.code target_method = \"first_target\"} in {.fn efa_mi}, or pass orthogonal unrotated loadings."),
      class = "efa_consensus_oblique_unsupported"
    )
  }

  single_args <- list(
    unrotated_list = unrotated_list,
    init_targets = init_targets,
    rotation = rotation,
    tol = tol,
    loss_tol = loss_tol,
    loss_patience = loss_patience,
    convergence = convergence,
    min_iter = min_iter,
    max_iter = max_iter,
    alpha = alpha,
    match_target = match_target,
    hyper_cutoff = hyper_cutoff,
    verbose = verbose
  )

  if (!multi_start) {
    out <- do.call(
      .consensus_target_procrustes_single,
      c(single_args, list(start = start))
    )

    out$multi_start <- list(
      enabled = FALSE,
      starts = out$start,
      best_start = out$start,
      best_index = 1L,
      losses = out$mean_loss,
      converged = out$converged,
      iterations = out$iterations,
      summary = data.frame(
        start = out$start,
        loss = out$mean_loss,
        converged = out$converged,
        iterations = out$iterations
      ),
      runs = NULL,
      congruence = NULL
    )

    return(out)
  }

  if (is.null(init_targets)) {
    single_args$init_targets <- unrotated_list
    n_targets <- length(unrotated_list)
  } else {
    n_targets <- length(init_targets)
  }
  starts <- .procrustes_validate_starts(starts, n_targets)

  runs <- lapply(starts, function(s) {
    do.call(
      .consensus_target_procrustes_single,
      c(single_args, list(start = s))
    )
  })

  losses <- vapply(runs, function(x) x$mean_loss, numeric(1L))
  names(losses) <- paste0("start_", starts)

  finite_losses <- is.finite(losses)
  if (!any(finite_losses)) {
    cli::cli_abort("No multi-start consensus run produced a finite final loss.",
                   class = "efa_consensus_no_finite_loss")
  }
  best_idx <- which.min(ifelse(finite_losses, losses, Inf))
  best <- runs[[best_idx]]

  n_runs <- length(runs)
  run_names <- paste0("start_", starts)
  congruence <- vector("list", n_runs)
  names(congruence) <- run_names

  for (i in seq_len(n_runs)) {
    congruence[[i]] <- vector("list", n_runs)
    names(congruence[[i]]) <- run_names
  }

  for (i in seq_len(n_runs)) {
    for (j in i:n_runs) {
      cij <- .tucker_congruence(
        runs[[i]]$pooled_loadings,
        runs[[j]]$pooled_loadings
      )
      congruence[[i]][[j]] <- cij
      congruence[[j]][[i]] <- if (i == j) cij else t(cij)
    }
  }

  run_summary <- data.frame(
    start = starts,
    loss = losses,
    converged = vapply(runs, function(x) isTRUE(x$converged), logical(1L)),
    iterations = vapply(runs, function(x) x$iterations, integer(1L)),
    final_failures = vapply(runs, function(x) length(x$final_failures), integer(1L)),
    row.names = run_names
  )

  best$multi_start <- list(
    enabled = TRUE,
    starts = starts,
    best_start = starts[best_idx],
    best_index = best_idx,
    losses = losses,
    converged = run_summary$converged,
    iterations = run_summary$iterations,
    summary = run_summary,
    runs = runs,
    congruence = congruence
  )

  best
}

#' Next eigenvalue sufficiency test (NEST)
#'
#' NEST uses many synthetic datasets to generate reference eigenvalues against
#' which to compare the empirical eigenvalues. This is similar to parallel
#' analysis, but other than parallel analysis, NEST does not just rely on
#' synthetic eigenvalues based on an identity matrix as null model.
#' It was introduced by Achim (2017), see also Brandenburg and Papenberg (2024) and
#' Caron (2025) for further simulation studies including NEST.
#'
#' @inheritParams efa_kgc
#' @param N numeric. The number of observations. Only needed if x is a correlation
#'  matrix. Must be larger than the number of variables.
#' @param alpha numeric. The alpha level to use (i.e., 1-alpha percentile of eigenvalues is used for reference values).
#' @param cor_method character. One of `"pearson"`, `"spearman"`, or `"kendall"`,
#'   passed to [stats::cor()]. `"poly"` and `"tetra"` are not supported because
#'   `NEST` compares the data against simulated continuous reference data.
#'  Default is  `"pearson"`.
#' @param n_datasets numeric. The number of datasets to simulate. Default is 1000.
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  [efa_fit()] reference-model fits. `NULL` (default) uses the [efa_fit()] defaults. The
#'  reference models are unrotated, so no rotation settings apply.
#' @param ... Additional arguments passed to [efa_fit()]. For example,
#' `estimator`, to change the estimator (default is "PAF"). PAF is more
#' robust, but it will take longer compared to the other estimators
#' available ("ML" and "ULS"). The estimation tuning knobs are not passed here; they
#' live in `estimate_control`, and the standard-error arguments (`se`, `b_boot`, `ci`,
#' `seed`) are not accepted because the reference-model fits are internal steps.
#'
#' @details NEST compares the first empirical eigenvalue against the first eigenvalues
#' of `n_dataset` synthetic datasets based on a null model  (i.e.,
#'  with uncorrelated variables; same as in parallel analysis, see [efa_parallel()]).
#'  The following eigenvalues are compared against synthetic datasets based on an EFA-model with one fewer factors
#'  than the position of the respective empirical eigenvalue. E.g, the second
#'  empirical eigenvalue is compared against synthetic data based on a one-factor
#'  model. In each comparison the \eqn{k}-th empirical eigenvalue is tested against
#'  the \eqn{k}-th largest eigenvalue of the synthetic datasets. The `alpha`-level
#'  defines against which percentile of the synthetic
#'  eigenvalue distribution to compare the empirical eigenvalues against, i.e., an
#'  alpha of .05 (the default) uses the 95th percentile as reference value.
#'
#'  The number of factors tested is capped at \eqn{\lfloor 0.8 \times p \rfloor}
#'  (with \eqn{p} the number of variables; Achim, 2017) and additionally limited so
#'  that the \eqn{(k - 1)}-factor reference model used at each step stays
#'  over-identified. If no empirical eigenvalue falls at or below its reference
#'  within this range, every tested factor is accepted and this capped number is
#'  returned.
#'
#'  Because each reference model carries the factors already retained, NEST does not
#'  lose accuracy for the strongly correlated factor structures where parallel
#'  analysis tends to under-extract, and it was among the more accurate criteria in
#'  the simulation studies of Brandenburg and Papenberg (2024) and Caron (2025). The
#'  price is runtime: a fresh set of `n_datasets` reference datasets is drawn and
#'  eigen-decomposed at every candidate factor count, which makes NEST one of the
#'  slowest criteria available here.
#'
#'  The reference models are fitted without inequality constraints. A Heywood case in
#'  one of them leaves no unique variance to simulate the reference data from, so
#'  NEST aborts rather than continuing from an inadmissible reference.
#'
#'  The reference eigenvalues are obtained from simulated data, so the suggested number
#'  of factors varies slightly from run to run. Call [base::set.seed()] beforehand to make a
#'  run reproducible.
#'
#'  For details on the method, including simulation studies, see Achim (2017),
#'  Brandenburg and Papenberg (2024), and Caron (2025).
#'
#'  The `efa_nest` function can also be called together with other factor
#'   retention criteria in the [efa_retain()] function.
#'
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). Its main fields are:
#' \item{n_factors}{A named numeric vector (`"NEST"`) with the suggested number of
#'   factors according to the NEST procedure.}
#' \item{results}{A list with a single record holding the empirical eigenvalues
#'   and the reference eigenvalues. Only the positions the search actually tested
#'   carry a reference value; beyond the position at which it stopped the series is
#'   `NA`.}
#' \item{settings}{A list of control settings used.}
#'
#' @source Achim, A. (2017). Testing the number of required dimensions in exploratory factor analysis. The Quantitative Methods for Psychology, 13(1), 64–74. https://doi.org/10.20982/tqmp.13.1.p064
#' @source Brandenburg, N., & Papenberg, M. (2024). Reassessment of innovative methods to determine the number of factors: A simulation-based comparison of exploratory graph analysis and Next Eigenvalue Sufficiency Test. Psychological Methods, 29(1), 21–47. https://doi.org/10.1037/met0000527
#' @source Caron, P.-O. (2025). A Comparison of the Next Eigenvalue Sufficiency Test to Other Stopping Rules for the Number of Factors in Factor Analysis.
#' Educational and Psychological Measurement, Online-first publication. https://doi.org/10.1177/00131644241308528
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#' @examples
#' \donttest{
#' # with correlation matrix
#' efa_nest(test_models$baseline$cormat, N = 500)
#'
#' # with raw data
#' efa_nest(GRiPS_raw)
#' }
efa_nest <- function(x, N = NA,
                 alpha = .05,
                 use = c("pairwise.complete.obs", "all.obs",
                         "complete.obs", "everything",
                         "na.or.complete"),
                 cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                 n_datasets = 1000,
                 estimate_control = NULL,
                 ...) {


  # Perform argument checks
  .reject_flat_knobs(...names(), fn = "efa_nest")
  .reject_unknown_fit_dots(...names(), fn = "efa_nest", unrotated = TRUE)
  .reject_rotation_dots(list(...), fn = "efa_nest")
  .assert_cor_input(x)

  checkmate::assert_count(N, na.ok = TRUE, positive = TRUE)
  checkmate::assert_number(alpha, lower = 0, upper = 1)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  .assert_estimate_control(estimate_control)
  .reject_poly_reference(cor_method, "efa_nest")
  checkmate::assert_count(n_datasets, na.ok = FALSE,
                          positive = TRUE)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                             N_policy = "required")
  R <- prep$R
  N <- prep$N

  .assert_n_gt_vars(N, ncol(R))

  emp_eigen <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
  nvar <- ncol(R)

  # Number of factors to test. Achim's (2017) rule caps this at floor(.8 * nvar);
  # it is additionally bounded so the (nf - 1)-factor reference model fit at each
  # step stays over-identified (df > 0). The reference data for the nf-th empirical
  # eigenvalue come from an (nf - 1)-factor EFA, whose loadings are only
  # trustworthy when that model has positive degrees of freedom
  # (df = ((nvar - k)^2 - (nvar + k)) / 2 for k factors); the largest reference
  # factor count with df > 0 caps the search at one factor beyond it.
  ref_k <- seq_len(nvar) - 1L
  ref_df <- .efa_df(nvar, ref_k)
  max_fac <- min(floor(.8 * nvar), max(ref_k[ref_df > 0]) + 1L)

  references <- vector(mode = "double", length = max_fac)

  # `stopped` records whether the search ended because an empirical eigenvalue
  # fell at/below its reference (i.e. a factor was rejected). If none does so
  # within the tested range, every tested factor was accepted and the last tested
  # model is retained (see the boundary handling after the loop).
  stopped <- FALSE
  for (nf in 1:max_fac) {

    if (nf == 1) {

      # For the first factor the reference is the null (identity) model: no loadings
      # and unit uniquenesses.
      Lambda <- matrix(numeric(0), nvar, 0)
      Psi <- rep(1, nvar)

    } else {


      # For further factors, use a model with nf - 1 factors as reference. The fit is an
      # internal step run once per tested factor; suppress its warnings so a forwarded
      # estimator (e.g. estimator = "ML") cannot turn per-model Heywood or non-convergence
      # warnings into one warning per loop iteration. A Heywood case is handled
      # explicitly below.
      mm <- suppressWarnings(efa_fit(R, N = N, n_factors = nf - 1,
                                     estimate_control = estimate_control, ...))
      Lambda <- mm$unrot_loadings

      # A Heywood case in the reference model (communality at or above 1) makes
      # the uniqueness negative, so the reference data cannot be simulated; abort
      # with an actionable message instead of passing a negative variance to the
      # C++ simulator.
      Psi <- 1 - mm$h2
      if (any(!is.finite(Psi) | Psi < 0)) {
        cli::cli_abort(
          c("A Heywood case occurred in the {nf - 1}-factor reference model used by NEST.",
            "x" = "A communality at or above 1 leaves no unique variance to simulate the reference data from.",
            "i" = "Try a different {.arg estimator} (e.g. {.val ML} or {.val ULS}) or fewer indicators."),
          class = "efa_nest_heywood"
        )
      }

    }

    # Simulate the reference datasets and return the nf-th largest eigenvalue of each.
    # The shared kernel draws via the factor-score rule X = F*L' + e from the reference
    # model (loadings Lambda, uniquenesses Psi), which is faster than drawing from the
    # model-implied correlation matrix.
    ref_values <- .simulate_cfm_eigen(nf, N, Lambda, Psi, n_datasets)

    references[nf] <- stats::quantile(ref_values, probs = 1 - alpha)
    if (emp_eigen[nf] <= references[nf]) {
      stopped <- TRUE
      break
    }


  }

  # On a rejection the retained model is the one with one fewer factor than the
  # rejected position. At the no-stop boundary (no rejection within the tested
  # range) every tested factor was accepted, so the last tested model
  # (`nf == max_fac`) is retained rather than the just-past index.
  n_factors <- if (stopped) nf - 1L else nf
  # Only the tested positions have a reference; pad the rest with NA so the series
  # aligns with the empirical eigenvalues (the shared eigenvalue plot binds the two
  # into one data frame and drops the NAs).
  references <- c(references[seq_len(nf)],
                  rep(NA_real_, length(emp_eigen) - nf))

  results <- list(list(
    name = "NEST",
    label = "Suggested number of factors",
    n_factors = n_factors,
    plot_type = "eigen",
    x = seq_along(emp_eigen),
    y = emp_eigen,
    reference = references,
    # mark the retained solution in the plot, as the other eigenvalue-based
    # criteria do; there is nothing to mark when no factor is retained
    highlight = if (n_factors >= 1) n_factors else NULL
  ))

  out <- .new_efa_retention(
    "NEST",
    results = results,
    settings = list(alpha = alpha, N = N, n_datasets = n_datasets,
                    use = use, cor_method = cor_method)
  )

  out

}

#' Power analysis for exploratory factor analysis
#'
#' @description
#' Analyses power for exploratory factor analysis, in one of two modes chosen with
#' `mode`.
#'
#' `mode = "rmsea"` (the default) gives the analytic power of the root mean square
#' error of approximation (RMSEA) tests of close and not-close fit (MacCallum,
#' Browne, & Sugawara, 1996). Give a sample size to get the power of the test, or
#' give a target power to get the sample size needed to reach it.
#'
#' `mode = "simulation"` runs a Monte-Carlo study: it draws `n_datasets` samples from
#' a known population (via [efa_simulate()]), analyses each one, and reports how well
#' the analysis recovers that population. See *Details* for what is reported.
#'
#' Here the number of variables is `p` and the number of factors is `k` (elsewhere in
#' the package: `n_vars` and `n_factors`).
#'
#' @details
#' # RMSEA mode
#'
#' Power rises with a larger sample, a larger model (more degrees of freedom), and a
#' bigger gap between the null and alternative RMSEA (MacCallum, Browne, & Sugawara,
#' 1996).
#'
#' Two tests are supported, chosen with `type` (never by the order of `eps0` and
#' `eps1`):
#' \describe{
#'   \item{`"close"`}{Tests *close fit* (MacCallum et al., 1996). The null
#'     hypothesis is that the fit is close (RMSEA \eqn{\le} `eps0`; conventionally
#'     0.05). Power is the chance of detecting a worse alternative (`eps1`;
#'     conventionally 0.08, so `eps0 < eps1`), in the upper tail.}
#'   \item{`"notclose"`}{Tests *not-close fit*. The null hypothesis is that the fit
#'     is not close (RMSEA \eqn{\ge} `eps0`). Power is the chance of detecting a
#'     better alternative (`eps1`; conventionally 0.01, so `eps0 > eps1`), in the
#'     lower tail.}
#' }
#' When `eps0` and `eps1` are in the wrong order for the chosen `type`, a message is
#' shown but the requested test still runs. Equal `eps0` and `eps1` leave nothing to
#' detect and are an error.
#'
#' Power always increases with `N`, so the required sample size (the smallest `N`
#' reaching `power`) is found by bisection. `N` is the **total** sample size across
#' groups: with `group > 1` the power calculation divides by `group` (the
#' `1 / group` factor), so spreading a fixed total over more groups gives less
#' power. The matching per-group sample size, `N / group`, is returned as
#' `N_per_group`.
#'
#' The `1 / group` factor makes all `group` groups the same size, so a required
#' total is rounded up to the next multiple of `group`. A solved `N_per_group` is thus
#' a whole number of persons, and the reported `power` is the power at a total that a
#' study can collect. With `group = 2` and `df = 102`, for example, the required total
#' is 260, or 130 per group. Bisection on the total alone gives 259, which asks for
#' 129.5 persons in each group.
#'
#' # Simulation mode
#'
#' The population is passed to [efa_simulate()], which draws `n_datasets` samples of
#' size `N` from it. The population's true number of factors, `k_true`, is
#' `ncol(Lambda)` for a factor-model population, or `k` for a bare `R`. By default
#' the population fits the factor model exactly, which overstates how well the
#' criteria and the fit recover its structure; setting a misfit target makes the
#' population more realistic (MacCallum, 2003).
#'
#' Each replicate is analysed three ways:
#' \describe{
#'   \item{**Hit-rate**}{The share of replicates where a criterion's suggested factor
#'     count (from `criteria`) matches `k_true`. A replicate where the criterion
#'     errored or gave no answer is left out of this count -- it does not count as a
#'     miss.}
#'   \item{**Structure recovery** (factor-model populations only)}{The `k_true`-factor
#'     model is fitted with [efa_fit()], its loadings are matched to the population
#'     loadings, and the matched-factor Tucker congruences (Lorenzo-Seva & ten Berge,
#'     2006) are compared with `recovery_threshold`. A replicate succeeds when its
#'     smallest (`min`) or average (`mean`) matched congruence reaches the
#'     threshold.}
#'   \item{**Convergence**}{Among the replicates whose fit completed, the share that
#'     converged and the share that produced a Heywood case.}
#' }
#' A replicate whose fit fails completely is not counted in any of the three
#' measures above. If any fit fails, a warning reports how many failed and the cause
#' of the first failure.
#'
#' Replicates are analysed in parallel with \pkg{future.apply}; choose a parallel
#' plan with [future::plan()]. Each replicate uses its own reproducible
#' random-number stream, so with a fixed `seed` the result does not depend on the
#' number of workers, and the caller's random-number state is left unchanged.
#'
#' @param mode character. The kind of power analysis: `"rmsea"` (the default; analytic
#'   RMSEA power) or `"simulation"` (Monte-Carlo hit-rate and structure recovery).
#'   `type`, `eps0`, `eps1`, `df`, `alpha`, `power`, and `group` apply to RMSEA mode only;
#'   the arguments marked *Simulation mode* below apply to the other; `N`, `p`, and `k` are
#'   used in both.
#' @param type character. The RMSEA test: `"close"` (test of close fit) or
#'   `"notclose"` (test of not-close fit). See *Details*.
#' @param eps0 numeric. The null-hypothesis RMSEA. Default is `0.05`.
#' @param eps1 numeric. The alternative-hypothesis RMSEA (the true RMSEA power is
#'   evaluated at). Default is `0.08` for `type = "close"` and `0.01` for
#'   `type = "notclose"`.
#' @param N numeric. In `"rmsea"` mode, the total sample size across groups (the
#'   plain sample size when `group` is `1`): give `N` to compute power, or leave it
#'   `NULL` to solve for the required `N` at a target `power`. In `"simulation"`
#'   mode `N` is required and is the size of each drawn sample, with no sample size
#'   solved for and no `group` division.
#' @param p numeric. The number of observed variables. In `"rmsea"` mode, used with
#'   `k` to derive `df` when `df` is not given directly. In `"simulation"` mode it is
#'   read off the population, so leave it unset (or matching `nrow(Lambda)` / `nrow(R)`).
#' @param k numeric. The number of factors. In `"rmsea"` mode, used with `p` to
#'   derive `df` when `df` is not given directly. In `"simulation"` mode it is the
#'   true number of factors: it is required with an `R` population and must be left
#'   unset (or match `ncol(Lambda)`) with a factor-model population.
#' @param df numeric. The model degrees of freedom. Either supply `df` directly or
#'   supply both `p` and `k`, from which `df = ((p - k)^2 - (p + k)) / 2`. Must be
#'   positive.
#' @param alpha numeric. The significance level. Default is `0.05`.
#' @param power numeric. The target power. Give `power` (or leave both `power` and
#'   `N` `NULL`, defaulting to `0.80`) to solve for the required `N`; leave it
#'   `NULL` while giving `N` to compute power. Exactly one of `N` and `power` is
#'   solved for.
#' @param group numeric. The number of groups. Default is `1`. `N` is the total
#'   across all `group` groups, not the size of each one, and a solved `N` is a
#'   multiple of `group`. See *Details*.
#' @param Lambda matrix. Simulation mode. A `p` by `k_true` population loading matrix.
#'   Supply this (optionally with `Phi`/`Psi`) to build a factor-model population;
#'   structure recovery is available only with this form. Passed to [efa_simulate()].
#' @param Phi matrix. Simulation mode. The `k_true` by `k_true` population factor
#'   intercorrelations. Only used with `Lambda`; defaults to orthogonal factors. When
#'   `rotation` is unset, an oblique `Phi` selects a `"promax"` recovery fit and an
#'   orthogonal one a `"varimax"` fit.
#' @param Psi numeric or matrix. Simulation mode. The population unique variances (a
#'   length-`p` vector or a `p` by `p` matrix). Only used with `Lambda`. Passed to
#'   [efa_simulate()].
#' @param R matrix. Simulation mode. A `p` by `p` population correlation matrix to draw
#'   from directly, instead of a factor model. Structure recovery is not available for
#'   this form (there are no population loadings to recover), and `k` is required.
#' @param n_datasets numeric. Simulation mode. The number of samples to draw and
#'   analyse. Default is `500`.
#' @param criteria character. Simulation mode. The factor-retention criteria to
#'   evaluate the hit-rate for, any of `"CD"`, `"EKC"`, `"HULL"`, `"KGC"`, `"MAP"`,
#'   `"NEST"`, `"PARALLEL"`, and `"SMT"` (see [efa_retain()]). Default is
#'   `c("EKC", "MAP")`. Criteria that simulate internally (`"CD"`, `"HULL"`,
#'   `"NEST"`, `"PARALLEL"`) make each run substantially slower.
#' @param estimator character. Simulation mode. The estimator (`"PAF"`, `"ML"`,
#'   or `"ULS"`) used for the recovery fit and the retention criteria. Default
#'   is `"PAF"`.
#' @param rotation character. Simulation mode. The rotation for the recovery fit,
#'   passed to [efa_fit()]. Default is `NULL`, which matches the population: `"varimax"`
#'   for orthogonal factors and `"promax"` for oblique ones (a single factor is left
#'   unrotated). Recovery aligns the fitted loadings to the population pattern by
#'   permutation and sign only. A rotation that does not seek that structure -- for
#'   example `"none"` with more than one factor -- will understate recovery, so keep
#'   the default (or another structure-seeking rotation) for a meaningful recovery
#'   rate.
#' @param recovery_threshold numeric. Simulation mode. The matched-factor Tucker
#'   congruence a replicate must reach to count as recovered. Default is `0.95`:
#'   Lorenzo-Seva and ten Berge (2006) treat congruence at or above this level as
#'   indicating the same factor.
#' @param model_error character. Simulation mode. The [efa_simulate()] method that
#'   perturbs the population with model error: `"TKL"` (Tucker-Koopman-Linn, the
#'   default here), `"CB"` (Cudeck-Browne), `"WB"` (Wu-Browne), or `"none"` for an
#'   exact population. It only takes effect when a target is supplied (`target_rmsea`
#'   and/or `target_cfi`), and only for a factor-model population; without a target
#'   the population stays exact whatever the method. `"TKL"` adds minor common
#'   factors, giving a realistically imperfect population but lowering both the
#'   hit-rate and structure recovery; `"CB"` and `"WB"` target the RMSEA only. `"CB"`
#'   keeps the population loadings as the exact minimizer of the perturbed
#'   population, so recovery stays close to perfect; `"WB"`'s loadings are not the
#'   minimizer, so they carry no such guarantee. Note that [efa_simulate()] itself
#'   defaults to `"CB"`: the same `target_rmsea` passed to both functions gives an
#'   easier population there, unless `model_error` is also set explicitly here.
#' @param target_rmsea numeric. Simulation mode. The population RMSEA the model should
#'   have relative to the perturbed population, activating model error. Default is
#'   `NULL`. Passed to [efa_simulate()].
#' @param target_cfi numeric. Simulation mode. The population CFI target (only with
#'   `model_error = "TKL"`). Default is `NULL`. Passed to [efa_simulate()].
#' @param seed numeric. Simulation mode. Optional seed making the draws and analysis
#'   reproducible and worker-count independent; the caller's random-number stream is
#'   restored afterwards. Default is `NULL`.
#'
#' @returns An object of class `efa_power`. For `mode = "rmsea"`, a list containing:
#' \item{power}{The power of the test at `N` (the achieved power, which for a
#'   solved sample size is at least the target).}
#' \item{N}{The total sample size across groups: the supplied `N`, or the solved
#'   required sample size (a multiple of `group`).}
#' \item{N_per_group}{The per-group sample size `N / group`, equal to `N` when
#'   `group` is `1`. A whole number for a solved `N`; for a supplied `N` that is not
#'   a multiple of `group` it is the fraction that the noncentrality uses.}
#' \item{crit}{The critical chi-square value the fit statistic is compared against.}
#' \item{ncp}{The noncentrality parameters under the null (`H0`, from `eps0`) and
#'   the alternative (`H1`, from `eps1`).}
#' \item{solve_for}{`"power"` or `"N"`, recording which quantity was solved for.}
#' \item{settings}{A list of the inputs: `mode`, `type`, `eps0`, `eps1`, `df`,
#'   `p`, `k`, `alpha`, `group`, and the target `power` (the value solved to when
#'   `solve_for` is `"N"`, otherwise `NULL`).}
#'
#' For `mode = "simulation"`, a list containing:
#' \item{hit_rate}{A named numeric vector of the retention hit-rate per criterion (and,
#'   where a criterion has several variants, per variant); `NA` for a criterion that
#'   returned no suggestion on any replicate.}
#' \item{hits}{A data frame with one row per criterion (`criterion`) giving the
#'   number of replicates it returned a definite suggestion on (`n_valid`), the
#'   number of those that matched `k_true` (`hits`), and the `hit_rate`
#'   (`hits / n_valid`).}
#' \item{recovery}{For a factor-model population, a list with the structure-recovery
#'   rates (`min_rate`, `mean_rate`), the `threshold`, and the number of usable fits
#'   (`n_valid`); `NULL` for an `R` population. Rates are over every replicate whose fit
#'   returned loadings, including non-converged or Heywood solutions (their rates are
#'   reported separately in `convergence`).}
#' \item{convergence}{A list with the number of datasets (`n_datasets`), the number
#'   of fits that completed (`n_fit_ok`), how many of those converged
#'   (`n_converged`) and how many produced a Heywood case (`n_heywood`), and the
#'   corresponding rates: `fit_rate` (fits completed, over all datasets) and
#'   `convergence_rate` / `heywood_rate` (converged / Heywood, over the completed
#'   fits).}
#' \item{replicates}{The raw per-replicate values: the suggested factor counts
#'   (`n_hat`), the matched congruences (`rec_min`, `rec_mean`), the `converged`,
#'   `heywood`, and `fit_ok` flags, and `fit_error`, the message of the fit that did
#'   not complete (`NA` where it did).}
#' \item{k_true}{The true number of factors.}
#' \item{model_error}{The [efa_simulate()] model-error record, or `NULL`.}
#' \item{settings}{A list of the simulation inputs.}
#'
#' @references
#' MacCallum, R. C., Browne, M. W., & Sugawara, H. M. (1996). Power analysis and
#' determination of sample size for covariance structure modeling. *Psychological
#' Methods, 1*(2), 130-149. \doi{10.1037/1082-989X.1.2.130}
#'
#' MacCallum, R. C. (2003). 2001 Presidential Address: Working with imperfect models.
#' *Multivariate Behavioral Research, 38*(1), 113-139. \doi{10.1207/S15327906MBR3801_5}
#'
#' Lorenzo-Seva, U., & ten Berge, J. M. F. (2006). Tucker's congruence coefficient as a
#' meaningful index of factor similarity. *Methodology, 2*(2), 57-64.
#' \doi{10.1027/1614-2241.2.2.57}
#'
#' @seealso [efa_simulate()] draws the replicate datasets used in simulation mode.
#'   [efa_retain()] implements the retention criteria whose hit-rates simulation
#'   mode reports.
#'
#' @family power analysis
#'
#' @export
#'
#' @examples
#' # Power of the test of close fit at N = 200 for a 100-df model
#' efa_power(df = 100, N = 200)
#'
#' # Deriving df from the model dimensions instead of giving it directly
#' efa_power(p = 20, k = 3, N = 200)
#'
#' # Required total sample size for 80% power
#' efa_power(df = 100, power = 0.80)
#'
#' # Test of not-close fit
#' efa_power(df = 100, N = 200, type = "notclose")
#'
#' \donttest{
#' # Simulation mode: retention hit-rate and structure recovery for a known
#' # three-factor population at N = 300
#' efa_power("simulation", Lambda = population_models$loadings$baseline,
#'           Phi = population_models$phis_3$moderate, N = 300,
#'           n_datasets = 50, criteria = c("EKC", "MAP"), seed = 42)
#' }
#'
efa_power <- function(mode = c("rmsea", "simulation"),
                      type = c("close", "notclose"),
                      eps0 = NULL, eps1 = NULL, N = NULL, p = NULL, k = NULL,
                      df = NULL, alpha = 0.05, power = NULL, group = 1,
                      Lambda = NULL, Phi = NULL, Psi = NULL, R = NULL,
                      n_datasets = 500, criteria = c("EKC", "MAP"),
                      estimator = "PAF", rotation = NULL, recovery_threshold = 0.95,
                      model_error = c("TKL", "CB", "WB", "none"),
                      target_rmsea = NULL, target_cfi = NULL, seed = NULL) {

  mode <- .match_arg_ci(mode)

  # Simulation mode is a self-contained Monte-Carlo path with its own inputs (the
  # population, the retention criteria, the recovery fit); the analytic RMSEA
  # arguments do not apply to it, so it branches off before they are validated.
  if (mode == "simulation") {
    return(.efa_power_simulation(
      N = N, Lambda = Lambda, Phi = Phi, Psi = Psi, R = R, p = p, k = k,
      n_datasets = n_datasets, criteria = criteria, estimator = estimator,
      rotation = rotation, recovery_threshold = recovery_threshold,
      model_error = model_error, target_rmsea = target_rmsea,
      target_cfi = target_cfi, seed = seed))
  }

  type <- .match_arg_ci(type)

  # RMSEA defaults: a common null of .05, tested against a worse (.08) alternative
  # for close fit or a better (.01) alternative for not-close fit.
  if (is.null(eps0)) eps0 <- 0.05
  if (is.null(eps1)) eps1 <- if (type == "close") 0.08 else 0.01
  checkmate::assert_number(eps0, lower = 0, finite = TRUE)
  checkmate::assert_number(eps1, lower = 0, finite = TRUE)
  checkmate::assert_count(group, positive = TRUE)
  if (!checkmate::test_number(alpha) || alpha <= 0 || alpha >= 1) {
    cli::cli_abort("{.arg alpha} must be a single number strictly between 0 and 1.",
                   class = "efa_power_bad_alpha")
  }

  # Degrees of freedom: taken from `df` directly, or derived from the model
  # dimensions via the shared EFA df formula. A non-positive df has no test.
  # `p` / `k` are validated whenever supplied -- even alongside an explicit `df` --
  # so a malformed value never lands in the returned settings.
  if (!is.null(p)) checkmate::assert_count(p, positive = TRUE)
  if (!is.null(k)) checkmate::assert_count(k, positive = TRUE)
  if (is.null(df)) {
    if (is.null(p) || is.null(k)) {
      cli::cli_abort(
        c("The model degrees of freedom are missing.",
          "i" = "Supply {.arg df} directly, or both {.arg p} (variables) and {.arg k} (factors)."),
        class = "efa_power_missing_df")
    }
    df <- .efa_df(p, k)
  } else {
    checkmate::assert_number(df, finite = TRUE)
  }
  if (df <= 0) {
    cli::cli_abort(
      c("The model has {df} degree{?s} of freedom; a positive value is required.",
        "i" = "Use more variables or fewer factors."),
      class = "efa_power_bad_df")
  }

  # Equal null and alternative RMSEA cannot be told apart, so no sample size (and
  # no power beyond alpha) exists.
  if (eps0 == eps1) {
    cli::cli_abort(
      c("{.arg eps0} and {.arg eps1} are both {eps0}, so the test has no power to detect a difference.",
        "i" = "Give them different RMSEA values (for example {.val 0.05} vs {.val 0.08} for a close-fit test)."),
      class = "efa_power_unreachable")
  }

  # The branch is chosen by `type`, not by the ordering of eps0/eps1; a reversed
  # pair is reported but still computed as requested.
  wrong_side <- (type == "close" && eps0 > eps1) ||
    (type == "notclose" && eps0 < eps1)
  if (wrong_side) {
    expected <- if (type == "close") {
      "A close-fit test expects the null RMSEA below the alternative ({.arg eps0} < {.arg eps1})."
    } else {
      "A not-close-fit test expects the null RMSEA above the alternative ({.arg eps0} > {.arg eps1})."
    }
    cli::cli_inform(
      c("!" = "For a {.val {type}}-fit test, {.arg eps0} ({eps0}) and {.arg eps1} ({eps1}) look reversed.",
        "i" = expected),
      class = "efa_power_wrong_side")
  }

  # Solve for whichever of N / power is left NULL; supplying both is over-determined.
  if (!is.null(N) && !is.null(power)) {
    cli::cli_abort(
      c("Both {.arg N} and {.arg power} were supplied, so there is nothing to solve for.",
        "i" = "Leave {.arg N} as {.code NULL} to find the required sample size, or {.arg power} as {.code NULL} to find the power at {.arg N}."),
      class = "efa_power_overdetermined")
  }

  if (is.null(N)) {
    solve_for <- "N"
    target <- if (is.null(power)) 0.80 else power
    if (!checkmate::test_number(target) || target <= 0 || target >= 1) {
      cli::cli_abort("{.arg power} must be a single number strictly between 0 and 1.",
                     class = "efa_power_bad_power")
    }
    N <- .efa_power_solve_N(df, eps0, eps1, alpha, group, type, target)
    settings_power <- target
  } else {
    solve_for <- "power"
    checkmate::assert_count(N, positive = TRUE)
    N <- as.integer(N)
    settings_power <- NULL
  }

  res <- .efa_power_rmsea(N, df, eps0, eps1, alpha, group, type)

  structure(
    list(
      power = res$power,
      N = N,
      N_per_group = N / group,
      crit = res$crit,
      ncp = c(H0 = res$ncp0, H1 = res$ncp1),
      solve_for = solve_for,
      settings = list(mode = mode, type = type, eps0 = eps0, eps1 = eps1, df = df,
                      p = p, k = k, alpha = alpha, group = group,
                      power = settings_power)
    ),
    class = "efa_power"
  )
}

# Power of the RMSEA test of (not-)close fit at a single sample size, from the
# noncentral chi-square framework of MacCallum, Browne & Sugawara (1996). The
# noncentrality parameter is (N - 1) * df * eps^2 / group. For close fit the
# rejection region is the upper tail (critical value = upper-alpha point of the
# null); for not-close fit it is the lower tail. Returns the power, the critical
# value, and both noncentrality parameters.
.efa_power_rmsea <- function(N, df, eps0, eps1, alpha, group, type) {
  ncp0 <- (N - 1) * df * eps0^2 / group
  ncp1 <- (N - 1) * df * eps1^2 / group
  # At very large noncentrality (huge N) qchisq()/pchisq() emit precision and
  # non-convergence notes while still returning valid quantiles/probabilities. Those
  # notes carry no information for the caller, so suppress them and keep power
  # reporting to the function's own (classed) conditions.
  suppressWarnings(
    if (type == "close") {
      crit <- stats::qchisq(alpha, df, ncp = ncp0, lower.tail = FALSE)
      pow <- stats::pchisq(crit, df, ncp = ncp1, lower.tail = FALSE)
    } else {
      crit <- stats::qchisq(1 - alpha, df, ncp = ncp0, lower.tail = FALSE)
      pow <- 1 - stats::pchisq(crit, df, ncp = ncp1, lower.tail = FALSE)
    }
  )
  list(power = pow, crit = crit, ncp0 = ncp0, ncp1 = ncp1)
}

# Smallest total sample size reaching `target` power that a design with `group` equal
# groups can have. Power rises monotonically with N (from about `alpha` toward 1), so
# double an upper bound until it clears the target, then bisect. `cap` bounds the search:
# an alternative RMSEA too close to the null (or a target power too high) can otherwise
# never be reached.
.efa_power_solve_N <- function(df, eps0, eps1, alpha, group, type, target,
                               cap = 1e7) {
  powfun <- function(N) .efa_power_rmsea(N, df, eps0, eps1, alpha, group, type)$power

  # The noncentrality divides the total by `group`, so the design has `group` groups of
  # equal size and only a total that is a multiple of `group` is possible. Power increases
  # with the total, so the smallest possible total is the bisected one rounded up to the
  # next multiple. Without this step a total of, say, 259 over two groups asks for 129.5
  # persons in each group, which is not a sample size.
  design_N <- function(N) as.integer(ceiling(N / group) * group)

  # Rounding up happens after the search, so the search itself must stop at the largest
  # collectable total within `cap`; otherwise a returned total could pass the bound that
  # the abort message below states (10000002 for `group` = 3 at the default cap).
  cap <- floor(cap / group) * group

  lo <- 1
  if (powfun(lo) >= target) return(design_N(1))
  # Double the upper bound, clamped to `cap`, until it clears the target. Probing
  # `cap` itself before giving up keeps the reachable ceiling equal to the documented
  # bound (a plain `hi * 2` would abort once past the last power of two below `cap`).
  hi <- 2
  while (powfun(hi) < target) {
    if (hi >= cap) {
      cli::cli_abort(
        c("No sample size up to {.val {cap}} reaches a power of {target}.",
          "i" = "The alternative RMSEA may be too close to the null, or the target power too high."),
        class = "efa_power_unreached")
    }
    hi <- min(hi * 2, cap)
  }
  while (hi - lo > 1) {
    mid <- floor((lo + hi) / 2)
    if (powfun(mid) >= target) hi <- mid else lo <- mid
  }
  design_N(hi)
}


# Simulation-mode power: draw `n_datasets` samples of size `N` from a known
# population, and over the replicates report how often each factor-retention
# criterion recovers the true factor count (hit-rate), how often the fitted
# loadings recover the population structure (Tucker congruence >= threshold), and
# the convergence/Heywood rate of the k_true-factor fit. Parallelised over
# replicates with a per-replicate reproducible RNG stream.
.efa_power_simulation <- function(N, Lambda, Phi, Psi, R, p, k, n_datasets,
                                  criteria, estimator, rotation, recovery_threshold,
                                  model_error, target_rmsea, target_cfi, seed) {

  model_error <- .match_arg_ci(model_error, c("TKL", "CB", "WB", "none"))
  estimator <- .match_arg_ci(estimator, c("PAF", "ML", "ULS"))
  # Only criteria that make a numeric suggestion can score a hit-rate; the visual
  # scree plot is excluded.
  valid_ids <- names(.retention_registry)[
    !vapply(.retention_registry, function(e) isTRUE(e$visual), logical(1))]
  # `criteria` is always passed on from efa_power(), so this function's own formal carries no
  # default for .match_arg_ci() to fall back on when the caller supplied NULL. Read it off
  # efa_power() instead of restating it, so the documented default stays in one place.
  if (is.null(criteria)) criteria <- eval(formals(efa_power)$criteria)
  criteria <- .match_arg_ci(criteria, valid_ids, several.ok = TRUE)

  # `N` is optional in RMSEA mode (it is solved for at a target power), which makes
  # omitting it the most likely simulation-mode mistake; there is nothing to solve for
  # here, so report it as a requirement rather than as a bare type assertion.
  if (is.null(N)) {
    cli::cli_abort(
      c("{.arg N} is required in simulation mode.",
        "x" = "It is the size of each drawn sample; no sample size is solved for here.",
        "i" = "Set {.arg N}, e.g. {.code N = 300}, or use {.code mode = \"rmsea\"} to solve for the required sample size."),
      class = "efa_power_input")
  }
  if (!checkmate::test_count(N, positive = TRUE)) {
    cli::cli_abort(
      c("{.arg N} must be a single positive whole number.",
        "x" = "It is the number of cases in each drawn sample."),
      class = "efa_power_input")
  }
  if (!checkmate::test_count(n_datasets, positive = TRUE)) {
    cli::cli_abort(
      c("{.arg n_datasets} must be a single positive whole number.",
        "x" = "It is the number of samples drawn and analysed."),
      class = "efa_power_input")
  }
  if (!checkmate::test_number(recovery_threshold) ||
      recovery_threshold <= 0 || recovery_threshold > 1) {
    cli::cli_abort("{.arg recovery_threshold} must be a single number in (0, 1].",
                   class = "efa_power_bad_threshold")
  }

  # The population is given exactly one way; efa_simulate() owns the deep checks
  # (symmetry, positive-semidefiniteness, Heywood, dimensions), raised as classed
  # efa_simulate_* conditions.
  have_lambda <- !is.null(Lambda)
  have_R <- !is.null(R)
  if (have_lambda == have_R) {
    cli::cli_abort(
      c("Specify the simulation population in exactly one way.",
        "x" = "Provide either {.arg R}, or {.arg Lambda} (with optional {.arg Phi}/{.arg Psi})."),
      class = "efa_power_input")
  }

  # k_true (the true factor count) comes from Lambda for a factor-model population --
  # a conflicting `k` is an error -- and must be given for a bare correlation matrix.
  # `p` is likewise read off the population; a conflicting one is an error rather than
  # silently replaced, so the two dimension arguments are policed the same way.
  if (!is.null(p)) checkmate::assert_count(p, positive = TRUE)
  if (have_lambda) {
    Lambda <- as.matrix(Lambda)
    k_true <- ncol(Lambda)
    if (!is.null(k) && k != k_true) {
      cli::cli_abort(
        c("{.arg k} ({k}) does not match the number of columns of {.arg Lambda} ({k_true}).",
          "i" = "With a factor-model population the true factor count is {.code ncol(Lambda)}; leave {.arg k} unset."),
        class = "efa_power_bad_k")
    }
    pop_arg <- "Lambda"
    p_true <- nrow(Lambda)
  } else {
    if (is.null(k)) {
      cli::cli_abort(
        c("{.arg k} (the true number of factors) is required with an {.arg R} population.",
          "i" = "A bare correlation matrix carries no factor count."),
        class = "efa_power_missing_k")
    }
    checkmate::assert_count(k, positive = TRUE)
    k_true <- as.integer(k)
    pop_arg <- "R"
    p_true <- nrow(as.matrix(R))
  }
  if (!is.null(p) && p != p_true) {
    cli::cli_abort(
      c("{.arg p} ({p}) does not match the number of rows of {.arg {pop_arg}} ({p_true}).",
        "i" = "In simulation mode the number of variables comes from the population; leave {.arg p} unset."),
      class = "efa_power_bad_p")
  }
  p <- p_true

  # Structure recovery aligns the fitted loadings against the population loadings, so
  # it needs a factor-model population; a bare R carries no loadings to recover.
  has_recovery <- have_lambda

  # Recovery-fit rotation: resolved against efa_fit()'s own choices before any data is
  # drawn, so a value it would reject costs milliseconds rather than a full run whose
  # recovery, convergence, and Heywood columns then come back empty. The choices are
  # read off efa_fit()'s formal rather than restated, so the two cannot drift apart.
  # (A test that replaces efa_fit() with a stub replaces its formals too, so such a
  # stub must either carry a `rotation` formal or leave `rotation` unset here.)
  if (!is.null(rotation)) {
    rotation <- .match_arg_ci(rotation, eval(formals(efa_fit)$rotation))
  }

  # Match the population when unset -- varimax for orthogonal factors, promax for
  # oblique ones. Both are deterministic, so the run stays reproducible; a single
  # factor has no rotation.
  if (is.null(rotation)) {
    oblique_pop <- have_lambda && !is.null(Phi) &&
      !isTRUE(all.equal(unname(as.matrix(Phi)), diag(k_true)))
    rotation <- if (oblique_pop) "promax" else "varimax"
  }
  if (k_true == 1L) rotation <- "none"

  # One seed makes the whole call reproducible and independent of the worker count,
  # and the caller's random-number stream is left untouched (mirrors efa_simulate()).
  if (!is.null(seed)) {
    checkmate::assert_int(seed)
    .set_local_seed(seed)
  }

  # Draw the replicate datasets via efa_simulate(). seed = NULL inherits (and advances)
  # the umbrella stream above, decorrelating the data draw from the analysis draw while
  # staying deterministic; efa_simulate()'s own future.seed = TRUE keeps the datasets
  # worker-count independent.
  sim <- efa_simulate(N = N, Lambda = Lambda, Phi = Phi, Psi = Psi, R = R,
                      model_error = model_error, target_rmsea = target_rmsea,
                      target_cfi = target_cfi, n_datasets = n_datasets, seed = NULL)
  datasets <- sim$data
  if (!is.list(datasets)) datasets <- list(datasets)

  # Shared retention control list: the N_FACTORS() defaults with the sample size `N`
  # and `estimator` threaded in (`gof` follows `estimator` inside
  # .n_factors_ctl(): PAF supports only the CAF), both fixed across replicates.
  ctl <- .n_factors_ctl(N = N, estimator = estimator)

  # Analyse every replicate; future.seed = TRUE binds each to its own reproducible
  # L'Ecuyer stream, so criteria that simulate internally (PARALLEL, NEST, CD) are
  # reproducible and independent of the number of workers.
  # Recovery aligns against the population loadings, so pass `Lambda` (already NULL for
  # an R population, which disables recovery inside .efa_power_analyze_one).
  per_rep <- future.apply::future_lapply(
    datasets, .efa_power_analyze_one,
    Lambda = Lambda, k_true = k_true, criteria = criteria, ctl = ctl,
    estimator = estimator, rotation = rotation, future.seed = TRUE)

  # --- Aggregate over replicates ---
  # Hit-rate: union the criterion/variant keys (a criterion can fail on a replicate,
  # or emit several variants), then score each against k_true over the replicates
  # where it produced a suggestion.
  # unlist() drops the NULLs a failed criterion contributes, and returns NULL when every
  # criterion failed on every replicate, so coerce to a character vector before matching.
  observed <- as.character(unique(unlist(lapply(per_rep, function(r) names(r$n_hat)))))
  # A criterion that failed on *every* replicate contributes no key at all, so without
  # this it would silently vanish from the results while `settings$criteria` still records
  # the request -- worst for the internally-simulating criteria on a hard population,
  # where a long run would quietly drop the criterion the user was paying for. Seed the
  # missing ids so each gets a row with n_valid = 0 (and, via the guard below, an NA
  # hit-rate), and say so. A key is a criterion's own id or `id_variant`
  # (see .retention_key()).
  failed <- criteria[!vapply(criteria, function(id) {
    any(observed == id | startsWith(observed, paste0(id, "_")))
  }, logical(1))]
  if (length(failed)) {
    cli::cli_warn(
      c("{cli::qty(failed)}Criteri{?on/a} {.val {failed}} produced no suggestion on any of the {n_datasets} replicate{?s}.",
        "x" = "{cli::qty(failed)}{?It is/They are} reported with a missing hit-rate over zero valid replicates.",
        "i" = "Such a criterion errored or was undecided throughout; a larger {.arg N}, a different {.arg estimator}, or a less demanding population may let it decide."),
      class = "efa_power_criterion_failed")
  }
  # `criteria` always holds at least one id, so seeding the failed ones leaves `keys`
  # non-empty even when no criterion decided anywhere: that case now runs the ordinary
  # path and comes out as n_valid = 0 with an NA hit-rate.
  keys <- c(observed, failed)
  # Extract this replicate's suggestion for every key; a replicate where a criterion
  # failed (its key absent) contributes NA there, and one where all of them failed has a
  # NULL `n_hat` whose indexing gives the wrong length.
  pull <- function(r) {
    v <- r$n_hat[keys]
    if (length(v) != length(keys)) v <- rep(NA_real_, length(keys))
    v
  }
  hit_mat <- matrix(
    vapply(per_rep, pull, numeric(length(keys))),
    nrow = length(keys), dimnames = list(keys, NULL))
  hit_n_valid <- rowSums(!is.na(hit_mat))
  hit_hits <- rowSums(hit_mat == k_true, na.rm = TRUE)
  hit_rate <- ifelse(hit_n_valid > 0, hit_hits / hit_n_valid, NA_real_)
  names(hit_rate) <- keys
  hits <- data.frame(criterion = keys, n_valid = hit_n_valid, hits = hit_hits,
                     hit_rate = hit_rate, row.names = NULL)

  rec_min <- vapply(per_rep, function(r) r$rec_min, numeric(1))
  rec_mean <- vapply(per_rep, function(r) r$rec_mean, numeric(1))
  converged <- vapply(per_rep, function(r) r$converged, logical(1))
  heywood <- vapply(per_rep, function(r) r$heywood, logical(1))
  fit_ok <- vapply(per_rep, function(r) r$fit_ok, logical(1))
  fit_error <- vapply(per_rep, function(r) r$fit_error, character(1))

  recovery <- if (has_recovery) {
    n_rec <- sum(!is.na(rec_min))
    list(
      min_rate = if (n_rec > 0) mean(rec_min >= recovery_threshold, na.rm = TRUE) else NA_real_,
      mean_rate = if (n_rec > 0) mean(rec_mean >= recovery_threshold, na.rm = TRUE) else NA_real_,
      threshold = recovery_threshold, n_valid = n_rec)
  }

  # `fit_rate` is over all replicates; the convergence and Heywood rates are conditional
  # on a fit having completed (denominator `n_fit_ok`), so a replicate whose fit errored
  # is not silently counted as a converged, non-Heywood solution.
  n_fit_ok <- sum(fit_ok)
  convergence <- list(
    n_datasets = as.integer(n_datasets), n_fit_ok = n_fit_ok,
    n_converged = sum(converged), n_heywood = sum(heywood),
    fit_rate = mean(fit_ok),
    convergence_rate = if (n_fit_ok > 0) sum(converged) / n_fit_ok else NA_real_,
    heywood_rate = if (n_fit_ok > 0) sum(heywood) / n_fit_ok else NA_real_)

  # Mirrors the criterion-failure warning above. A systematic failure -- a rotation or
  # estimator that cannot fit this population, a population the fit cannot handle -- is
  # otherwise indistinguishable from per-replicate numerical trouble, and when every fit
  # fails the whole recovery/convergence half of the report is NA with no stated cause.
  # The replicate's own error message is carried through so the user does not have to
  # reproduce the failure to see it. The completed retention results are still returned.
  n_fit_failed <- as.integer(n_datasets) - n_fit_ok
  if (n_fit_failed > 0L) {
    first_failed <- which(!fit_ok)[1L]
    first_cause <- fit_error[first_failed]
    # An R population has no loadings to recover, so the same fit is only the source of
    # the convergence and Heywood rates there; name it for what it did on this run.
    what_fit <- if (has_recovery) "recovery fit" else "replicate fit"
    affected <- if (has_recovery) {
      "Structure recovery, the convergence rate, and the Heywood rate"
    } else {
      "The convergence rate and the Heywood rate"
    }
    cli::cli_warn(
      c("The {what_fit} failed on {n_fit_failed} of the {n_datasets} replicate{?s}.",
        "x" = if (n_fit_ok == 0L) {
          "{affected} are missing."
        } else {
          "{affected} are over the {n_fit_ok} fit{?s} that completed."
        },
        "i" = "Replicate {first_failed} failed with: {first_cause}",
        "i" = "Check {.arg rotation} ({.val {rotation}}) and {.arg estimator} ({.val {estimator}}) against what {.fn efa_fit} accepts."),
      class = "efa_power_fit_failed")
  }

  settings <- list(mode = "simulation", N = as.integer(N),
                   n_datasets = as.integer(n_datasets), p = p, k = k_true,
                   criteria = criteria, estimator = estimator, rotation = rotation,
                   recovery_threshold = recovery_threshold,
                   has_recovery = has_recovery, model_error = model_error,
                   target_rmsea = target_rmsea, target_cfi = target_cfi, seed = seed)

  structure(
    list(
      hit_rate = hit_rate, hits = hits, recovery = recovery,
      convergence = convergence,
      replicates = list(n_hat = t(hit_mat), rec_min = rec_min, rec_mean = rec_mean,
                        converged = converged, heywood = heywood, fit_ok = fit_ok,
                        fit_error = fit_error),
      k_true = k_true, model_error = sim$model_error, settings = settings),
    class = "efa_power")
}

# Analyse a single simulated dataset: (i) run each requested retention criterion
# straight off the registry and record its suggested factor count, (ii) fit the
# k_true-factor EFA once (its convergence code and Heywood flag), and (iii) align the
# fitted loadings to the population loadings and read the matched-factor Tucker
# congruences (recovery). Returns a per-replicate record for aggregation.
.efa_power_analyze_one <- function(dat, Lambda, k_true, criteria, ctl, estimator,
                                   rotation) {

  # Retention: the registry funs take raw data (needs_raw criteria) or a correlation
  # matrix; a criterion that errors on this replicate contributes NA.
  Rmat <- stats::cor(dat)
  n_hat <- unlist(lapply(criteria, function(id) {
    entry <- .retention_registry[[id]]
    xarg <- if (isTRUE(entry$needs_raw)) dat else Rmat
    # Muffle the criteria's routine progress/informational output (e.g. "computing
    # correlations from the raw data"). A criterion that errors contributes nothing on
    # this replicate (NULL is dropped by unlist below), lowering its `n_valid` -- rather
    # than a bare-id NA that would not match its variant-named success key.
    out <- suppressMessages(suppressWarnings(
      tryCatch(entry$fun(xarg, ctl), error = function(e) NULL)))
    if (is.null(out)) return(NULL)
    # Name the suggestion(s) exactly as N_FACTORS() does.
    .retention_key(id, out$n_factors)
  }))

  # k_true-factor fit (once): the source of the convergence/Heywood rate and the
  # loadings recovery aligns against. EFA's own Heywood/non-convergence warnings are
  # muffled here; the flags are read off the returned object instead. A failure keeps
  # its cause, so the aggregation can report why the fit did not complete instead of
  # leaving the user with a column of NAs.
  fit <- suppressMessages(suppressWarnings(tryCatch(
    efa_fit(dat, n_factors = k_true, estimator = estimator, rotation = rotation),
    error = function(e) e)))
  fit_error <- NA_character_
  if (inherits(fit, "error")) {
    fit_error <- conditionMessage(fit)
    fit <- NULL
  }
  fit_ok <- !is.null(fit)
  converged <- isTRUE(fit_ok && fit$convergence == 0)
  heywood <- isTRUE(fit_ok && length(fit$heywood) > 0)

  # Structure recovery: permute/sign-match the fitted loadings to the population
  # loadings, then take the matched-factor congruences (the diagonal). A degenerate
  # (near-zero) factor makes the congruence undefined -- caught as NA.
  rec_min <- NA_real_
  rec_mean <- NA_real_
  if (!is.null(Lambda) && fit_ok) {
    loadings <- if (rotation == "none") {
      .change_class(fit$unrot_loadings, "matrix")
    } else {
      .change_class(fit$rot_loadings, "matrix")
    }
    matched <- tryCatch({
      aligned <- .align_solution(Lambda, loadings)
      diag(.tucker_congruence(Lambda, aligned$loadings))
    },
    efa_zero_column = function(e) NULL,
    efa_undefined_congruence = function(e) NULL,
    error = function(e) NULL)
    if (!is.null(matched)) {
      rec_min <- min(matched)
      rec_mean <- mean(matched)
    }
  }

  list(n_hat = n_hat, converged = converged, heywood = heywood, fit_ok = fit_ok,
       fit_error = fit_error, rec_min = rec_min, rec_mean = rec_mean)
}

#' Print and format an efa_power object
#'
#' `print()` turns an [efa_power()] result into a short report, and `format()`
#' builds the same report as a character vector (`print()` is
#' `cat(format(x), sep = "\n")`).
#'
#' For an RMSEA-mode object, the report has a header naming the test, the null and
#' alternative hypotheses with the significance level and degrees of freedom, the
#' headline result (the power at the sample size, or the required sample size for
#' the target power), and the critical value and noncentrality parameters.
#'
#' For a simulation-mode object, the report instead has the population and design,
#' the retention hit-rate per criterion, the structure-recovery rate, and the
#' convergence and Heywood-case rate.
#'
#' The lines follow the active console theme, so they print as plain text when
#' colours are disabled -- for example when captured into a file, or stripped with
#' [cli::ansi_strip()].
#'
#' @param x An object of class `efa_power` (output from [efa_power()]).
#' @param digits Integer. The number of decimal places the reported values are
#'   rounded to. Default is 3.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @family power analysis
#'
#' @export
#'
#' @method print efa_power
#'
#' @examples
#' pw <- efa_power(df = 100, N = 200)
#' pw
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(pw))
#'
print.efa_power <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_power
#' @export
#' @method format efa_power
format.efa_power <- function(x, digits = 3, ...) {

  # Simulation-mode results carry a different payload (hit-rate, recovery,
  # convergence) and are rendered by their own formatter.
  if (identical(x$settings$mode, "simulation")) {
    return(.format_efa_power_simulation(x, digits = digits))
  }

  s <- x$settings
  test_lbl <- if (s$type == "close") "Test of close fit" else "Test of not-close fit"
  # <= bounds the null for close fit, >= for not-close fit.
  cmp <- if (s$type == "close") "\u2264" else "\u2265"
  e0 <- .efa_num(s$eps0, digits = digits, pad = FALSE)
  e1 <- .efa_num(s$eps1, digits = digits, pad = FALSE)
  a <- .efa_num(s$alpha, digits = digits, pad = FALSE)
  pw <- .efa_num(x$power, digits = digits, pad = FALSE)
  crit <- .efa_num(x$crit, digits = digits, pad = FALSE)
  n0 <- .efa_num(x$ncp[["H0"]], digits = digits, pad = FALSE)
  n1 <- .efa_num(x$ncp[["H1"]], digits = digits, pad = FALSE)

  cli::cli_format_method({
    cli::cli_text("")
    cli::cli_rule(left = "{.strong RMSEA power analysis}")
    cli::cli_text("")

    cli::cli_text("{test_lbl}: H0 RMSEA {cmp} {e0} vs. H1 RMSEA = {e1}.")
    if (s$group > 1) {
      cli::cli_text("alpha = {a} \u00b7 df = {s$df} \u00b7 groups = {s$group}")
    } else {
      cli::cli_text("alpha = {a} \u00b7 df = {s$df}")
    }
    cli::cli_text("")

    # With more than one group `N` is the total across them, which is the quantity a
    # multiple-group study is most likely to misread as per-group; name it and give the
    # per-group figure outright.
    # A sample size, not a coefficient: format it as a count rather than through
    # .efa_num(), whose fixed `digits` would render 100 as "100.000".
    # A solved `N` is a multiple of `group`, so the per-group figure is a whole number.
    # A supplied `N` is the caller's own total and is never changed: when it does not
    # divide by `group`, the fraction the noncentrality uses is shown as it is.
    # An object serialized before `N_per_group` existed carries `group` but not the field;
    # its `N` was per-group back then, so no note can be derived from it -- omit the note
    # rather than erroring in round(NULL) or dividing a per-group N by `group` again.
    n_note <- if (s$group > 1 && !is.null(x$N_per_group)) {
      paste0(" (total; ", format(round(x$N_per_group, 1)), " per group)")
    } else {
      ""
    }
    if (x$solve_for == "power") {
      cli::cli_text("{.strong Power = {pw}} at N = {x$N}{n_note}.")
    } else {
      tgt <- .efa_num(s$power, digits = digits, pad = FALSE)
      cli::cli_text("{.strong Required N = {x$N}}{n_note} for a power of {tgt} (achieved {pw}).")
    }
    cli::cli_text(
      "Critical value \u03c7\u00b2({s$df}) = {crit} \u00b7 noncentrality H0 = {n0}, H1 = {n1}.")
  })
}

# Report lines for a simulation-mode efa_power object: the population and design, the
# retention hit-rate per criterion, the structure-recovery rate, and the
# convergence/Heywood rate. Numbers are pre-formatted into plain strings so cli's
# `{...}` interpolation does not mistake `.efa_num(...)` for an inline style.
.format_efa_power_simulation <- function(x, digits = 3) {

  s <- x$settings
  conv <- x$convergence

  hit_lines <- vapply(seq_len(nrow(x$hits)), function(i) {
    h <- x$hits[i, ]
    paste0(h$criterion, ": ", .efa_num(h$hit_rate, digits = digits, pad = FALSE),
           " (n = ", h$n_valid, ")")
  }, character(1))

  conv_rate <- .efa_num(conv$convergence_rate, digits = digits, pad = FALSE)
  heywood_rate <- .efa_num(conv$heywood_rate, digits = digits, pad = FALSE)
  fit_rate <- .efa_num(conv$fit_rate, digits = digits, pad = FALSE)

  cli::cli_format_method({
    cli::cli_text("")
    cli::cli_rule(left = "{.strong EFA power simulation}")
    cli::cli_text("")

    cli::cli_text(
      "{s$p} variable{?s} \u00b7 {x$k_true} factor{?s} \u00b7 N = {s$N} \u00b7 {s$n_datasets} dataset{?s}")
    cli::cli_text("Estimation: {s$estimator} \u00b7 rotation: {s$rotation}")
    if (!is.null(x$model_error)) {
      me <- x$model_error
      me_rmsea <- .efa_num(me$rmsea, digits = digits, pad = FALSE)
      me_cfi <- .efa_num(me$cfi, digits = digits, pad = FALSE)
      cli::cli_text(
        "Model error ({me$method}): RMSEA = {me_rmsea} \u00b7 CFI = {me_cfi}")
    } else {
      cli::cli_text(
        "Model error: none. The population is exact, so the hit-rate and recovery are optimistic; set {.arg target_rmsea} for realism.")
    }
    cli::cli_text("")

    cli::cli_text("{.strong Retention hit-rate} P(k-hat = {x$k_true})")
    # Every requested criterion now gets a row, so a fresh object always has lines here;
    # an object serialized before that carries no row for a criterion that failed on every
    # replicate, and prints this instead of an empty section (as for `N_per_group` above).
    if (length(hit_lines) > 0) {
      cli::cli_ul(hit_lines)
    } else {
      cli::cli_text("No criterion produced a suggestion on any replicate.")
    }
    cli::cli_text("")

    if (!is.null(x$recovery)) {
      r <- x$recovery
      thr <- .efa_num(r$threshold, digits = digits, pad = FALSE)
      min_rate <- .efa_num(r$min_rate, digits = digits, pad = FALSE)
      mean_rate <- .efa_num(r$mean_rate, digits = digits, pad = FALSE)
      # min_rate/mean_rate are the proportions of replicates clearing the threshold,
      # not congruences. The median raw congruence is added alongside them so a rate
      # of .000 still says how far off the recovered structure was.
      med_min <- .efa_num(stats::median(x$replicates$rec_min, na.rm = TRUE),
                          digits = digits, pad = FALSE)
      cli::cli_text("{.strong Structure recovery} (Tucker congruence \u2265 {thr})")
      cli::cli_ul(c(
        paste0("recovery rate (min congruence): ", min_rate, " (n = ", r$n_valid, ")"),
        paste0("recovery rate (mean congruence): ", mean_rate, " (n = ", r$n_valid, ")"),
        paste0("median min congruence: ", med_min)))
    } else {
      cli::cli_text(
        "{.strong Structure recovery}: not available (needs a factor-model population).")
    }
    cli::cli_text("")

    cli::cli_text("{.strong Convergence}")
    cli::cli_ul(c(
      paste0("fits completed: ", fit_rate, " (", conv$n_fit_ok, "/", conv$n_datasets, ")"),
      paste0("converged (of completed): ", conv_rate),
      paste0("Heywood cases (of completed): ", heywood_rate)))
  })
}

# Declarative `type` presets and the single resolver that applies them.
#
# Every estimator/rotation helper exposes a `type` preset ("EFAtools", "psych",
# "SPSS") that fills a block of tuning arguments, plus "none" which requires the
# user to set them explicitly. `.resolve_settings()` is the one place that turns
# a chosen `type` and the user-supplied (possibly NA) arguments into the final
# settings, warning once about any argument pinned alongside a preset.

# Per-function argument defaults for each preset. One source of truth, consumed
# by .PAF() (via .estimate_model()) and by .rotate_model() for every rotation.
.efa_presets <- list(
  PAF = list(
    EFAtools = list(init_comm = "smc", criterion = 1e-3,
                    criterion_type = "sum", max_iter = 300, abs_eigen = TRUE),
    psych    = list(init_comm = "smc", criterion = 1e-3,
                    criterion_type = "sum", max_iter = 50, abs_eigen = FALSE),
    SPSS     = list(init_comm = "smc", criterion = 1e-3,
                    criterion_type = "max_individual", max_iter = 25,
                    abs_eigen = TRUE)
  ),
  PROMAX = list(
    EFAtools = list(normalize = TRUE, P_type = "norm", order_type = "eigen",
                    varimax_type = "kaiser", k = 4),
    psych    = list(normalize = TRUE, P_type = "unnorm", order_type = "eigen",
                    varimax_type = "svd", k = 4),
    SPSS     = list(normalize = TRUE, P_type = "norm", order_type = "ss_factors",
                    varimax_type = "kaiser", k = 4)
  ),
  VARIMAX = list(
    EFAtools = list(normalize = TRUE, order_type = "eigen",
                    varimax_type = "kaiser"),
    psych    = list(normalize = TRUE, order_type = "eigen",
                    varimax_type = "svd"),
    SPSS     = list(normalize = TRUE, order_type = "ss_factors",
                    varimax_type = "kaiser")
  ),
  ROTATE_OBLQ = list(
    EFAtools = list(normalize = TRUE, order_type = "eigen"),
    psych    = list(normalize = TRUE, order_type = "eigen"),
    SPSS     = list(normalize = TRUE, order_type = "ss_factors")
  ),
  ROTATE_ORTH = list(
    EFAtools = list(normalize = TRUE, order_type = "eigen"),
    psych    = list(normalize = TRUE, order_type = "eigen"),
    SPSS     = list(normalize = TRUE, order_type = "ss_factors")
  )
)

# The settings lists and the rotation engine keep the historical internal spellings of two
# knobs; messages must show the names the current interface takes (rotate_control()).
.knob_display_names <- function(nms) {
  renames <- c(P_type = "p_type", randomStarts = "random_starts")
  hit <- nms %in% names(renames)
  nms[hit] <- renames[nms[hit]]
  nms
}

# Resolve a preset block into the final settings list. `type` is the chosen
# preset (or "none"); `user` holds the user-supplied values (NA where unset);
# `preset` is the `.efa_presets` slice for the calling function. Every setting
# uses NA to signal "unset" except `normalize`, the one on/off toggle whose
# preset value is always TRUE, so the user only overrides it by passing FALSE.
.resolve_settings <- function(type, user, preset) {

  setting_names <- names(preset[[1]])
  required <- setdiff(setting_names, "normalize")

  if (type == "none") {

    # Without a preset every NA-defaulted argument must be supplied by the user.
    missing <- required[vapply(user[required], is.na, logical(1))]

    if (length(missing) > 0) {
      missing <- .knob_display_names(missing)
      cli::cli_abort(
        c(
          paste("{cli::qty(missing)} {.arg type} is {.val none} but",
                "{?a required argument is/required arguments are} unset:",
                "{.arg {missing}}."),
          "i" = paste("Either set {.arg type} to {.val EFAtools},",
                      "{.val psych}, or {.val SPSS}, or supply",
                      "{cli::qty(missing)} {?it/them} explicitly.")
        ),
        class = "efa_type_none"
      )
    }

    return(user[setting_names])
  }

  # A `type` that names no block of this preset slice would make `defaults` NULL, and
  # `resolved[[nm]] <- defaults[[nm]]` then DELETES the element instead of filling it --
  # handing the engine a settings list with holes rather than failing. Every public caller
  # matches `type` against the four supported names first, so reaching this is an internal
  # fault: a preset slice added without one of the type names.
  if (!type %in% names(preset)) {
    cli::cli_abort(
      c("No {.val {type}} preset is defined for this function.",
        "i" = "Available presets: {.val {names(preset)}}."),
      class = "efa_preset_undefined",
      .internal = TRUE
    )
  }

  defaults <- preset[[type]]
  resolved <- user[setting_names]
  pinned <- character(0)

  for (nm in setting_names) {

    if (identical(nm, "normalize")) {
      if (isFALSE(user[[nm]])) pinned <- c(pinned, nm)
    } else if (is.na(user[[nm]])) {
      resolved[[nm]] <- defaults[[nm]]
    } else {
      pinned <- c(pinned, nm)
    }

  }

  if (length(pinned) > 0) {

    # `pairs` is interpolated as a variable (never baked into the glue template)
    # so user-supplied values cannot be parsed as cli markup.
    pairs <- paste0(.knob_display_names(pinned), " = ",
                    vapply(pinned, function(nm) format(resolved[[nm]]),
                           character(1)))

    cli::cli_warn(
      c(
        paste("{cli::qty(pinned)} {?An argument was/Arguments were} set",
              "together with {.arg type} = {.val {type}}; the supplied",
              "value{?s} {?is/are} used and may differ from the",
              "{.val {type}} preset:"),
        "*" = "{pairs}"
      ),
      class = "efa_type_override"
    )
  }

  resolved
}

# Unit tests for the type-preset resolver shared by the estimation and rotation
# helpers.

test_that("preset defaults are filled when arguments are NA", {
  res <- .resolve_settings(
    type = "SPSS",
    user = list(init_comm = NA, criterion = NA, criterion_type = NA,
                max_iter = NA, abs_eigen = NA),
    preset = .efa_presets$PAF
  )
  expect_equal(res$init_comm, "smc")
  expect_equal(res$criterion_type, "max_individual")
  expect_equal(res$max_iter, 25)
  expect_equal(res$abs_eigen, TRUE)

  res_psych <- .resolve_settings(
    type = "psych",
    user = list(normalize = TRUE, order_type = NA, varimax_type = NA),
    preset = .efa_presets$VARIMAX
  )
  expect_equal(res_psych$order_type, "eigen")
  expect_equal(res_psych$varimax_type, "svd")
})

test_that("specified arguments are kept and untouched ones take the preset default", {
  res <- suppressWarnings(.resolve_settings(
    type = "EFAtools",
    user = list(init_comm = NA, criterion = NA, criterion_type = "max_individual",
                max_iter = NA, abs_eigen = NA),
    preset = .efa_presets$PAF
  ))
  expect_equal(res$criterion_type, "max_individual")
  expect_equal(res$max_iter, 300)
})

test_that("one consolidated warning is emitted regardless of how many arguments are pinned", {
  expect_warning(
    .resolve_settings(
      type = "EFAtools",
      user = list(init_comm = "smc", criterion = NA, criterion_type = NA,
                  max_iter = NA, abs_eigen = NA),
      preset = .efa_presets$PAF
    ),
    class = "efa_type_override"
  )

  warns <- capture_warnings(.resolve_settings(
    type = "EFAtools",
    user = list(init_comm = "smc", criterion = 0.01, criterion_type = "sum",
                max_iter = 100, abs_eigen = FALSE),
    preset = .efa_presets$PAF
  ))
  expect_length(warns, 1)
})

test_that("normalize is treated as pinned only when FALSE", {
  expect_warning(
    .resolve_settings(
      type = "EFAtools",
      user = list(normalize = FALSE, order_type = NA, varimax_type = NA),
      preset = .efa_presets$VARIMAX
    ),
    class = "efa_type_override"
  )

  expect_no_warning(.resolve_settings(
    type = "EFAtools",
    user = list(normalize = TRUE, order_type = NA, varimax_type = NA),
    preset = .efa_presets$VARIMAX
  ))
})

test_that("type = 'none' passes the user values through without a condition", {
  expect_no_warning(res <- .resolve_settings(
    type = "none",
    user = list(normalize = TRUE, order_type = "eigen", varimax_type = "svd"),
    preset = .efa_presets$VARIMAX
  ))
  expect_equal(res$order_type, "eigen")
  expect_equal(res$varimax_type, "svd")
})

test_that("type = 'none' errors when a required argument is missing", {
  expect_error(
    .resolve_settings(
      type = "none",
      user = list(normalize = TRUE, order_type = NA, varimax_type = "svd"),
      preset = .efa_presets$VARIMAX
    ),
    class = "efa_type_none"
  )

  # normalize has no NA state, so leaving it at its default must not error.
  expect_silent(.resolve_settings(
    type = "none",
    user = list(normalize = TRUE, order_type = "eigen"),
    preset = .efa_presets$ROTATE_ORTH
  ))
})

test_that("a type that names no preset block errors instead of dropping settings", {
  # Assigning a NULL default deletes the element rather than filling it, so an
  # unmatched type would silently hand the engine a settings list with holes.
  expect_error(
    .resolve_settings(
      type = "SPPS",
      user = list(normalize = TRUE, order_type = NA, varimax_type = NA),
      preset = .efa_presets$VARIMAX
    ),
    class = "efa_preset_undefined"
  )

  # every supported type still resolves to a complete settings list
  for (tp in names(.efa_presets$VARIMAX)) {
    res <- .resolve_settings(
      type = tp,
      user = list(normalize = TRUE, order_type = NA, varimax_type = NA),
      preset = .efa_presets$VARIMAX
    )
    expect_named(res, c("normalize", "order_type", "varimax_type"))
    expect_false(anyNA(unlist(res)))
  }
})

test_that("preset tables hold the documented defaults", {
  expect_equal(.efa_presets$PAF$SPSS$max_iter, 25)
  expect_equal(.efa_presets$PAF$psych$abs_eigen, FALSE)
  expect_equal(.efa_presets$PROMAX$psych$P_type, "unnorm")
  expect_equal(.efa_presets$ROTATE_ORTH$SPSS$order_type, "ss_factors")
})

#' Plot efa_average object
#'
#' Plot method showing a summarized output of the [efa_average()] function
#'
#' @param x list. An output from the [efa_average()] function.
#' @param ... not used.
#'
#' @returns A ggplot object showing, for each indicator and factor, the minimum,
#'  maximum, and average (mean or median) loading across the averaged solutions.
#'  Each panel carries a point at the average, a bar spanning the minimum to the
#'  maximum with a tick at each endpoint, and a grey band marking the loadings
#'  that fall below the salience threshold; the caption names the four marks.
#'
#' @importFrom rlang .data
#' @export
#' @method plot efa_average
#'
#' @examples
#' \donttest{
#' EFA_aver <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500)
#' plot(EFA_aver)
#' }
#'
plot.efa_average <- function(x, ...) {

  averaging <- x$settings$averaging

  # Prepare data
  dat <- lapply(x$loadings, function(temp){
    df <- as.data.frame(unclass(temp))

    data.frame(
      rowname  = rep(rownames(df), each = ncol(df)),
      colname  = rep(colnames(df), times = nrow(df)),
      loadings = as.vector(t(as.matrix(df))),
      stringsAsFactors = FALSE
    )
  })

  dat <- do.call(cbind, dat)

  dat <- dat[, c("average.rowname", "average.colname", "average.loadings",
                 "min.loadings", "max.loadings")]
  names(dat) <- c("row_ind", "col_ind", "average", "min", "max")
  dat$row_ind <- factor(dat$row_ind, levels = rownames(x$loadings$average))
  dat$col_ind <- factor(dat$col_ind, levels = colnames(x$loadings$average))

  # Create plot faceted for variables and factors

  # No aesthetic is mapped to a scale (the four marks are drawn by four separate layers
  # with fixed appearance), so ggplot2 renders no legend; name the marks in a caption
  # instead. Without it the panel is uninterpretable: the grey band in particular is the
  # decision-relevant element and nothing else in the figure identifies it.
  centre <- if (averaging == "mean") "mean" else "median"
  caption <- paste0("Point = ", centre, "; bar = min-max range; ticks = endpoints; ",
                    "grey band = |loading| < ", format(x$settings$salience_threshold))

  plot_load <- ggplot2::ggplot(dat) +
    ggplot2::geom_segment(ggplot2::aes(x = min, xend = max, y = 0, yend = 0)) +
    ggplot2::geom_segment(ggplot2::aes(x = min, xend = min, y = -0.5, yend = 0.5)) +
    ggplot2::geom_segment(ggplot2::aes(x = max, xend = max, y = -0.5, yend = 0.5)) +
    ggplot2::geom_rect(xmin = -x$settings$salience_threshold,
                       xmax = x$settings$salience_threshold,
                       ymin = -2, ymax = 2, fill = ggplot2::alpha("grey", 0.3)) +
    ggplot2::scale_y_continuous(limits = c(-1, 1)) +
    ggplot2::geom_point(ggplot2::aes(.data$average, 0), color = "darkred") +
    ggplot2::facet_grid(rows = ggplot2::vars(.data$row_ind),
                        cols = ggplot2::vars(.data$col_ind),
                        switch = "y") +
    .gg_theme() +
    ggplot2::ggtitle(paste0("Minimum, Maximum, and ",
                                ifelse(averaging == "mean", "Mean", "Median"),
                                " Loadings")) +
    # The x axis is the loading scale; the y axis is a structural dummy (every mark sits
    # at y = 0, and the vertical extent only gives the ticks and the band their height),
    # so it keeps neither a title nor labels.
    ggplot2::labs(x = "Loading", caption = caption) +
    ggplot2::theme(axis.title.y = ggplot2::element_blank(),
          axis.text.y = ggplot2::element_blank(),
          axis.line.x = ggplot2::element_line(color = "black", linewidth = 0.2),
          axis.ticks.x = ggplot2::element_line(color = "black", linewidth = 0.2),
          panel.grid.minor.y = ggplot2::element_blank(),
          panel.grid.major.y = ggplot2::element_blank(),
          panel.spacing.y = ggplot2::unit(0, "mm"),
          strip.text.y.left = ggplot2::element_text(angle = 0, hjust = 0),
          strip.text.x = ggplot2::element_text(face = "bold"),
          strip.background.x = ggplot2::element_rect(color = "black", linewidth = 0.2),
          panel.border = ggplot2::element_rect(color = "gray", fill = NA,
                                               linewidth = 0.2)
          )

  plot_load

}

#' Print an efa_sl_loadings object
#'
#' @details
#' Prints a Schmid-Leiman loading matrix (general factor, group factors, and the
#' communality/uniqueness columns) as a styled, decimal-aligned table. Loadings with
#' absolute value greater than or equal to `cutoff` are emphasised, smaller loadings are
#' de-emphasised, and Heywood-relevant cells (a loading or communality above 1, or a
#' negative uniqueness) are highlighted. If the matrix has many columns or the console is
#' narrow, the table is split into stacked column blocks so the output stays readable.
#'
#' @param x class efa_sl_loadings matrix.
#' @param cutoff numeric. The value at or above which loadings are emphasized
#'  (default is .2). The default is lower than the .3 of an ordinary loading
#'  table ([print.efa_loadings()]): the group-factor loadings are residualized,
#'  that is, they carry only the variance left once the general factor has been
#'  partialled out, and are therefore smaller than the corresponding first-order
#'  loadings.
#' @param digits numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
#'  to round the loadings to (default is 3).
#' @param max_name_length numeric. The maximum length of the variable names to
#'  display; see [print.efa_loadings()].
#' @param color logical. Whether to apply console styling using \pkg{cli}.
#'  Default is `TRUE`.
#' @param name_style character. How to shorten variable names longer than
#'  `max_name_length`; see [print.efa_loadings()].
#' @param max_factors_per_block numeric or `NULL`. Maximum number of factor
#'  columns to print per block. If `NULL`, the number is chosen from the
#'  console width.
#' @param sort_loadings character. Optional row sorting; see
#'  [print.efa_loadings()]. The default `"none"` keeps the input order. When
#'  sorting is requested, rows are grouped by their largest *group*-factor
#'  loading: the general factor is left out of the comparison, since it is the
#'  largest loading of almost every item and sorting on it would leave the order
#'  untouched.
#' @param ... additional arguments passed to print or format.
#'
#' @returns `print()` returns its argument `x` invisibly; it is
#'   `cat(format(x, ...), sep = "\n")` followed by a blank line for console
#'   spacing. `format()` returns a character vector with the table lines (styled
#'   to the active console theme; plain when colours are disabled).
#'
#' @method print efa_sl_loadings
#' @export
#'
#' @examples
#' EFA_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' efa_schmid_leiman(EFA_mod, estimator = "PAF")
#'
print.efa_sl_loadings <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  # a trailing blank line, so a loading table printed at the console is set off from
  # whatever follows it
  cat("\n")

  invisible(x)
}

#' @rdname print.efa_sl_loadings
#' @method format efa_sl_loadings
#' @export
format.efa_sl_loadings <- function(x, cutoff = .2, digits = 3, max_name_length = 10,
                                   color = TRUE,
                                   name_style = c("truncate", "abbreviate", "full"),
                                   max_factors_per_block = NULL,
                                   sort_loadings = c("none", "primary", "clustered"),
                                   ...) {

  name_style <- .match_arg_ci(name_style)
  sort_loadings <- .match_arg_ci(sort_loadings)

  mat <- unclass(x)

  # The same validator format.efa_loadings() runs, so the shared display API rejects the
  # same values with the same wording and condition classes in both places. The arguments
  # this class does not offer are passed at their no-op values.
  .validate_loadings_print_args(
    x = mat,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = NULL,
    color = color,
    max_factor_name_length = NULL,
    max_factors_per_block = max_factors_per_block,
    legend = FALSE
  )

  n_col <- ncol(mat)

  factor_names <- colnames(mat)
  if (is.null(factor_names)) {
    factor_names <- paste0("F", seq_len(n_col))
  }

  var_names <- rownames(mat)
  if (is.null(var_names)) {
    var_names <- paste0("V", seq_len(nrow(mat)))
  }

  # The columns are [g, F1..Fn, h2, u2]; the last two are communality/uniqueness.
  col_roles <- rep("loading", n_col)
  col_roles[n_col - 1L] <- "h2"
  col_roles[n_col] <- "u2"

  loading_cols <- seq_len(n_col - 2L)

  # Shortened before any reordering, for the same reason as in .loadings_print_spec(): a row
  # label has to be a property of the variable, not of where the sorting puts it.
  var_names <- .shorten_loadings_names(var_names, max_length = max_name_length,
                                       name_style = name_style)

  # Row sorting compares the group factors only. The general factor carries the largest
  # loading of almost every item in a Schmid-Leiman solution, so including it would assign
  # every row to the same primary factor and leave the order as it was.
  group_cols <- setdiff(loading_cols, 1L)
  if (length(group_cols) > 0L) {
    row_order <- .efa_loading_row_order(mat[, group_cols, drop = FALSE], sort_loadings)
    mat <- mat[row_order, , drop = FALSE]
    var_names <- var_names[row_order]
  }

  # Count items (rows) flagged as Heywood cases: a loading with absolute value
  # above 1, a communality (h2) above 1, or a negative uniqueness. These are
  # deterministically coupled within an item, so each affected item counts once.
  n_heywood <- sum(
    rowSums(abs(mat[, loading_cols, drop = FALSE]) > 1, na.rm = TRUE) > 0 |
      (!is.na(mat[, n_col - 1L]) & mat[, n_col - 1L] > 1) |
      (!is.na(mat[, n_col]) & mat[, n_col] < 0))

  cli::cli_format_method({
    # Through .efa_emit_lines(), for the same reason as in format.efa_loadings(): the blank
    # line separating stacked column blocks would not survive cli_verbatim().
    .efa_emit_lines(.efa_format_matrix(
      values = mat,
      row_labels = var_names,
      col_labels = factor_names,
      col_roles = col_roles,
      cutoff = cutoff,
      digits = digits,
      color = color,
      max_factors_per_block = max_factors_per_block
    ))

    if (n_heywood >= 1L) {
      cli::cli_verbatim("")
      if (n_heywood == 1L) {
        cli::cli_alert_warning("Results contain a Heywood case!")
      } else {
        cli::cli_alert_warning("Results contain {n_heywood} Heywood cases!")
      }
    }
  })
}

# Stage-1 of two-stage / full-information maximum likelihood (FIML) correlation estimation:
# from raw data with missing values, EM-estimate the saturated multivariate-normal mean and
# covariance assuming multivariate normality and ignorable (MAR) missingness, plus the
# saturated observed-data log-likelihood that the fit-index and standard-error paths reuse.
# Two-stage EFA with missing data: Yuan, Marshall & Bentler (2002, Psychometrika 67:95-121).
# MAR/ignorability: Little & Rubin (2002). EM reference implementation: lavaan's
# lav_mvn_mi_h1_est_moments() / lav_mvnorm_missing_*().
#
# The asymptotic covariance of the saturated estimates (the meat for corrected two-stage
# standard errors) is the inverse observed information of the saturated log-likelihood. Under
# FIML the observed -- not the expected -- information is used: the expected information would
# have to be averaged over the random missingness distribution and is inconsistent under MAR
# (Yuan & Bentler, 2000; Savalei & Bentler, 2009, SEM). lavaan/Mplus make the same choice.

# Group row indices by missingness pattern (unique sets of observed columns); returns, per
# pattern, the observed/missing column indices and the rows that share it, so the EM and the
# log-likelihood can accumulate per pattern rather than per case.
.fiml_patterns <- function(obs) {
  # One key per row via a vectorised column-wise paste of the 0/1 mask (avoids a per-row
  # apply() and stays exact for any number of variables).
  keys <- do.call(paste0, asplit(matrix(as.integer(obs), nrow(obs)), 2L))
  lapply(split(seq_len(nrow(obs)), keys), function(rows) {
    pat <- obs[rows[1L], ]
    list(rows = rows, o = which(pat), m = which(!pat), freq = length(rows))
  })
}

# EM estimate of the saturated MVN mean and covariance from raw data with missing values
# (the FIML / two-stage Stage-1 engine; Yuan, Marshall & Bentler, 2002).
.fiml_em_moments <- function(data, max_iter = 500L, tol = 1e-5) {

  if (length(max_iter) != 1L || !is.finite(max_iter) || max_iter < 1 ||
      max_iter != as.integer(max_iter)) {
    cli::cli_abort("{.arg max_iter} must be a single positive integer.",
                   class = "efa_fiml_bad_max_iter")
  }

  if (length(tol) != 1L || !is.finite(tol) || tol <= 0) {
    cli::cli_abort("{.arg tol} must be a single positive number.",
                   class = "efa_fiml_bad_tol")
  }

  Y <- as.matrix(data)
  if (!is.numeric(Y)) {
    cli::cli_abort(
      c("{.arg data} must be numeric to estimate FIML moments.",
        "x" = "You supplied {.obj_type_friendly {data}}."),
      class = "efa_fiml_not_numeric"
    )
  }

  p <- ncol(Y)
  nms <- colnames(Y)
  if (is.null(nms)) nms <- paste0("V", seq_len(p))

  obs <- !is.na(Y)

  # Fully-missing rows carry no information for the saturated moments; drop them so n counts
  # the cases actually used.
  keep <- rowSums(obs) > 0L
  Y <- Y[keep, , drop = FALSE]
  obs <- obs[keep, , drop = FALSE]
  n <- nrow(Y)

  # Coverage: diagonal = observed cases per variable, off-diagonal = jointly-observed cases
  # per pair. A variable or a pair with zero coverage cannot be estimated.
  coverage <- crossprod(obs)
  if (any(diag(coverage) == 0)) {
    empty <- nms[diag(coverage) == 0]
    n_empty <- length(empty)
    cli::cli_abort(
      c("Every variable must have at least one observed value.",
        "x" = "{n_empty} variable{?s} ({.val {empty}}) {?has/have} no observed values."),
      class = "efa_fiml_no_observed"
    )
  }
  zero_pairs <- which(coverage == 0 & upper.tri(coverage), arr.ind = TRUE)
  if (nrow(zero_pairs) > 0L) {
    pair <- paste(nms[zero_pairs[1L, 1L]], nms[zero_pairs[1L, 2L]], sep = " & ")
    cli::cli_abort(
      c("Every variable pair must have at least one jointly observed case.",
        "x" = "Variables {.val {pair}} are never observed together (zero coverage)."),
      class = "efa_fiml_zero_coverage"
    )
  }

  patterns <- .fiml_patterns(obs)
  n_patterns <- length(patterns)

  # Init: mu from the available data; sigma from the complete cases when there are enough of
  # them and their covariance is well-conditioned, else a diagonal of column variances
  # (floored to stay positive definite). A near-singular complete-case covariance would seed
  # the EM with an ill-conditioned start, so it is rejected by an rcond threshold rather than
  # by chol() alone (which accepts a tiny positive pivot).
  mu <- colMeans(Y, na.rm = TRUE)
  cc <- stats::complete.cases(Y)
  sigma <- NULL
  if (sum(cc) > p) {
    s_cc <- stats::cov(Y[cc, , drop = FALSE])
    if (!inherits(try(chol(s_cc), silent = TRUE), "try-error") &&
        rcond(s_cc) > .Machine$double.eps) {
      sigma <- s_cc
    }
  }
  if (is.null(sigma)) {
    v <- apply(Y, 2L, stats::var, na.rm = TRUE)
    v[!is.finite(v) | v <= 0] <- 1
    sigma <- diag(v, nrow = p)
  }

  # Convergence is judged on the scale-invariant standardised moments (the standardised means
  # mu/sd, the log-variances, and the off-diagonal correlations), so the criterion does not
  # depend on the variables' measurement scale; an absolute tolerance on the raw moments would
  # stop early on small-variance data and over-iterate on large-variance data. lavaan's
  # lav_mvnorm_missing_* uses a likelihood-based criterion for the same scale-invariance.
  lt_off <- lower.tri(sigma)
  std_par <- function(mu, sigma) {
    v <- diag(sigma)
    s <- sqrt(v)
    c(mu / s, log(v), (sigma / tcrossprod(s))[lt_off])
  }
  par_old <- std_par(mu, sigma)

  converged <- FALSE
  for (iter in seq_len(max_iter)) {

    # E-step accumulators: T1 = sum of completed rows, T2 = sum of their outer products.
    T1 <- numeric(p)
    T2 <- matrix(0, p, p)

    for (pat in patterns) {
      o <- pat$o
      m <- pat$m
      Yo <- Y[pat$rows, o, drop = FALSE]

      if (length(m) == 0L) {
        # Complete pattern: the observed block contributes directly.
        T1[o] <- T1[o] + colSums(Yo)
        T2[o, o] <- T2[o, o] + crossprod(Yo)
        next
      }

      # Solve each pattern's Sigma_oo directly (lavaan instead inverts the full Sigma once and
      # downdates each pattern from it).
      Soo_inv <- chol2inv(chol(sigma[o, o, drop = FALSE]))
      B <- sigma[m, o, drop = FALSE] %*% Soo_inv          # conditional-regression block

      # Complete each row: observed entries as-is, missing entries -> conditional means
      # yhat_m = mu_m + B (y_o - mu_o).
      Yc <- matrix(0, length(pat$rows), p)
      Yc[, o] <- Yo
      Yc[, m] <- sweep(sweep(Yo, 2L, mu[o], "-") %*% t(B), 2L, mu[m], "+")

      T1 <- T1 + colSums(Yc)
      T2 <- T2 + crossprod(Yc)

      # Conditional-covariance correction for the missing x missing block, added once per case
      # (constant within the pattern). Omitting it biases sigma downward.
      corr <- sigma[m, m, drop = FALSE] - B %*% sigma[o, m, drop = FALSE]
      T2[m, m] <- T2[m, m] + pat$freq * corr
    }

    # M-step.
    mu <- T1 / n
    sigma <- T2 / n - tcrossprod(mu)
    sigma <- (sigma + t(sigma)) / 2                       # symmetrise away round-off

    # A loss of positive-definiteness means the completed-data covariance has lost rank. That
    # has two quite different causes and the message must name both: a (near-)constant or
    # collinear variable, and simply too few informative cases for p variables -- the latter is
    # what n <= p produces, where a user sent to inspect the variables would inspect the wrong
    # thing. The concrete counts are added only when the sample is that small, so the extra
    # bullet never appears where it would be a red herring.
    if (inherits(try(chol(sigma), silent = TRUE), "try-error")) {
      cli::cli_abort(
        c("The FIML covariance estimate stopped being positive definite during the EM iteration.",
          "i" = "The completed-data covariance lost rank: a variable may be (near-)constant or
                 collinear, or the observed cases may carry too little information to estimate
                 {p} variance{?s} and their covariances.",
          if (n <= p) {
            c("x" = "Only {n} case{?s} carr{?ies/y} an observed value, for {p} variable{?s}.")
          }),
        class = "efa_fiml_not_posdef"
      )
    }

    par_new <- std_par(mu, sigma)
    if (max(abs(par_new - par_old)) < tol) {
      converged <- TRUE
      break
    }
    par_old <- par_new
  }

  if (!converged) {
    # Every remedy named here has to be reachable from every entry point that estimates FIML
    # correlations, not only from the ones that take an `estimate_control()`: this warning is
    # raised during correlation preparation, which `efa_average()` also runs and which exposes no
    # control object. So the remedy that always applies is stated first and unconditionally, and
    # the knobs are named together with where they can be supplied.
    cli::cli_warn(
      c("FIML EM did not converge in {max_iter} iteration{?s}; the last iterate is returned and
         analysed in place of the FIML estimate.",
        "i" = "The EM converges linearly and needs more iterations the larger the fraction of
               missing information, so analysing fewer variables, or ones with less missingness,
               helps wherever the FIML correlation is estimated.",
        "i" = "Where an {.fn estimate_control} can be supplied -- {.fn efa_fit}, and the functions
               that forward one to it -- raise {.arg fiml_max_iter} or relax {.arg fiml_tol}
               there."),
      class = "efa_fiml_em_nonconvergence"
    )
  }

  names(mu) <- nms
  dimnames(sigma) <- list(nms, nms)

  # The saturated log-likelihood; this call also re-validates positive-definiteness, so a
  # near-singular fixed point that slipped past the per-iteration chol() guard (chol accepts a
  # sub-eps eigenvalue) aborts here with efa_fiml_not_posdef rather than being returned.
  logl <- .fiml_loglik(Y, mu, sigma, patterns = patterns)

  list(mu = mu, sigma = sigma, logl = logl, n = n,
       n_patterns = n_patterns, iter = iter, converged = converged)
}

# Saturated (h1) observed-data multivariate-normal log-likelihood for data with missing
# values: the sum over cases of the observed-variable MVN log-density, grouped by missingness
# pattern. Reused by the FIML fit-index and standard-error paths.
.fiml_loglik <- function(data, mu, sigma, patterns = NULL) {

  # A non-positive-definite sigma has no valid MVN density; refuse with a classed error rather
  # than letting chol() raise a bare base error (this helper is reused with externally supplied
  # moments). A principal submatrix of a positive-definite matrix is itself positive definite,
  # so this single check covers Sigma_oo for every pattern. The threshold matches the
  # non-positive-definite guard in .prepare_cor_input().
  if (any(eigen(sigma, symmetric = TRUE, only.values = TRUE)$values < .Machine$double.eps)) {
    cli::cli_abort(
      c("The covariance matrix is not positive definite, so the FIML log-likelihood is undefined.",
        "i" = "A variable may be (near-)constant or collinear."),
      class = "efa_fiml_not_posdef"
    )
  }

  Y <- as.matrix(data)

  # The EM engine already grouped the rows it passes (over its fully-missing-row-filtered data),
  # so accept those patterns directly; otherwise group here so the helper stays self-contained
  # for external callers that pass raw, unfiltered data (fully-missing rows carry no density).
  if (is.null(patterns)) {
    obs <- !is.na(Y)
    keep <- rowSums(obs) > 0L
    Y <- Y[keep, , drop = FALSE]
    patterns <- .fiml_patterns(obs[keep, , drop = FALSE])
  }

  log2pi <- log(2 * pi)
  ll <- 0
  for (pat in patterns) {
    o <- pat$o
    Yo <- Y[pat$rows, o, drop = FALSE]
    ch <- chol(sigma[o, o, drop = FALSE])
    logdet <- 2 * sum(log(diag(ch)))                     # log|Sigma_oo|
    C <- sweep(Yo, 2L, mu[o], "-")
    quad <- rowSums((C %*% chol2inv(ch)) * C)            # (y_o - mu_o)' Sigma_oo^{-1} (.)
    ll <- ll + sum(-0.5 * (length(o) * log2pi + logdet + quad))
  }
  ll
}

# Analytic gradient (score) of the saturated observed-data log-likelihood with respect to
# theta = (mu, vech(sigma)), accumulated per missingness pattern; vech is the column-major
# lower triangle including the diagonal (the .vech() convention). The covariance
# block is returned with respect to the distinct free parameters: an off-diagonal sigma_ij
# appears in both Sigma[i, j] and Sigma[j, i], so its derivative carries the factor 2 from the
# duplication-matrix chain rule. Differentiating this score gives the observed information.
.fiml_saturated_score <- function(data, mu, sigma, patterns = NULL) {

  # No valid score without a positive-definite covariance (mirrors .fiml_loglik() /
  # .fiml_saturated_acov(); a principal submatrix of a positive-definite matrix is positive
  # definite, so every pattern's Sigma_oo is then safe to factor below).
  if (any(eigen(sigma, symmetric = TRUE, only.values = TRUE)$values < .Machine$double.eps)) {
    cli::cli_abort(
      c("The covariance matrix is not positive definite, so the FIML score is undefined.",
        "i" = "A variable may be (near-)constant or collinear."),
      class = "efa_fiml_not_posdef"
    )
  }

  Y <- as.matrix(data)
  p <- ncol(Y)

  # The EM passes its (fully-missing-row-filtered) patterns; otherwise group here, matching
  # .fiml_loglik() so an external caller can pass raw, unfiltered data.
  if (is.null(patterns)) {
    obs <- !is.na(Y)
    keep <- rowSums(obs) > 0L
    Y <- Y[keep, , drop = FALSE]
    patterns <- .fiml_patterns(obs[keep, , drop = FALSE])
  }

  mu_grad <- numeric(p)
  Gfull <- matrix(0, p, p)                                # symmetric-matrix gradient w.r.t. Sigma

  for (pat in patterns) {
    o <- pat$o
    C <- sweep(Y[pat$rows, o, drop = FALSE], 2L, mu[o], "-")
    Soo_inv <- chol2inv(chol(sigma[o, o, drop = FALSE]))
    S1 <- colSums(C)                                      # sum of (y_o - mu_o)
    S2 <- crossprod(C)                                    # sum of (y_o - mu_o)(y_o - mu_o)'

    mu_grad[o] <- mu_grad[o] + as.numeric(Soo_inv %*% S1)
    Gfull[o, o] <- Gfull[o, o] -
      0.5 * (pat$freq * Soo_inv - Soo_inv %*% S2 %*% Soo_inv)
  }

  # Map the symmetric matrix gradient to the distinct vech parameters: double the off-diagonals
  # (chain rule), leave the diagonal as is.
  Gpar <- 2 * Gfull
  diag(Gpar) <- diag(Gfull)
  c(mu_grad, .vech(Gpar))
}

# Observed information of the saturated observed-data log-likelihood at theta =
# (mu, vech(sigma)): the negative Hessian of .fiml_loglik(), whose gradient is
# .fiml_saturated_score(), in closed form and accumulated per missingness pattern.
#
# Write M for the pattern's inverse Sigma_oo^-1 padded back to p x p by the selection matrix
# that picks its observed columns, u = M sum_i (y_i - mu) and W = M (sum_i (y_i - mu)(y_i -
# mu)') M. Differentiating the score of a pattern with n_g cases gives the three blocks
#
#   mu    x mu    :  n_g M
#   sigma x mu    :  D' (u kron M)
#   sigma x sigma :  (1/2) D' [ W kron M + M kron W - n_g (M kron M) ] D
#
# with D the duplication matrix that maps vech(Sigma) to vec(Sigma) -- the same closed form
# lavaan accumulates in lav_mvnorm_missing_logl_hessian_samplestats(). Its derivation is the
# product rule applied to d(Sigma_oo^-1) = -Sigma_oo^-1 (d Sigma_oo) Sigma_oo^-1 (Magnus &
# Neudecker, 2019, Matrix Differential Calculus, 3rd ed., ch. 10). No Kronecker product is
# formed: every entry of the covariance block is a product of two entries of M and W, which
# the index vectors below read off directly, and only the pattern's own observed parameters
# are touched, since M and W vanish outside its observed block.
#
# Cost is one factorisation per missingness pattern, against the 1 + 2 * (p + p(p+1)/2) score
# evaluations -- and as many factorisations per pattern -- that a numerical Jacobian of the
# score needs; the difference is what makes a FIML bootstrap, which rebuilds this for every
# replicate, affordable at realistic p.
#
# `sigma` must be positive definite and `Y` and `patterns` must be the matrix and grouping
# .fiml_saturated_acov() has already prepared: it raises the classed efa_fiml_not_posdef
# before calling, so every pattern's Sigma_oo below is safe to factor and the check is not
# repeated per replicate.
.fiml_saturated_information <- function(Y, mu, sigma, patterns) {

  p <- ncol(Y)
  pstar <- p * (p + 1L) / 2L

  # Column-major lower-triangle position of each (row >= col) entry of vech(sigma).
  ij <- which(lower.tri(diag(p), diag = TRUE), arr.ind = TRUE)
  vech_pos <- matrix(0L, p, p)
  vech_pos[ij] <- seq_len(pstar)

  info <- matrix(0, p + pstar, p + pstar)

  for (pat in patterns) {
    o <- pat$o
    k <- length(o)
    ng <- pat$freq

    A <- chol2inv(chol(sigma[o, o, drop = FALSE]))        # M on the observed block
    C <- sweep(Y[pat$rows, o, drop = FALSE], 2L, mu[o], "-")
    u <- as.numeric(A %*% colSums(C))
    W <- A %*% crossprod(C) %*% A
    P <- ng * A - W

    # The pattern's own vech pairs (local indices, row >= col) and where they sit in the
    # global vech(sigma).
    lij <- which(lower.tri(diag(k), diag = TRUE), arr.ind = TRUE)
    b <- lij[, 1L]
    cl <- lij[, 2L]
    off <- b != cl                                        # the duplicated parameters
    gi <- p + vech_pos[cbind(o[b], o[cl])]

    # Mean block.
    info[o, o] <- info[o, o] + ng * A

    # Mean-by-covariance block.
    X <- A[, b, drop = FALSE] * rep(u[cl], each = k)
    X[, off] <- X[, off] +
      A[, cl[off], drop = FALSE] * rep(u[b[off]], each = k)
    info[o, gi] <- info[o, gi] + X
    info[gi, o] <- info[gi, o] + t(X)

    # Covariance block. Each entry pairs the row parameter sigma_bc with the column
    # parameter sigma_de. A column for an off-diagonal parameter gains the term with d and e
    # exchanged, because that parameter sits in both Sigma[d, e] and Sigma[e, d]; a row for
    # one carries the matching factor two from the same chain rule.
    H <- 0.5 * (A[b, b, drop = FALSE] * P[cl, cl, drop = FALSE] -
                  W[b, b, drop = FALSE] * A[cl, cl, drop = FALSE])
    H[, off] <- H[, off] +
      0.5 * (A[b, cl[off], drop = FALSE] * P[cl, b[off], drop = FALSE] -
               W[b, cl[off], drop = FALSE] * A[cl, b[off], drop = FALSE])
    H[off, ] <- 2 * H[off, ]
    info[gi, gi] <- info[gi, gi] - H
  }

  # Symmetric by construction; the average removes the accumulated round-off.
  (info + t(info)) / 2
}

# Asymptotic covariance of the Stage-1 saturated FIML estimates: the inverse observed
# information (negative Hessian of the saturated log-likelihood) of theta = (mu, vech(sigma)),
# and its correlation-scale variant for the off-diagonal correlations of cov2cor(sigma). The
# information comes from .fiml_saturated_information(); the correlation variant standardises
# the covariance block by the delta method. Returns a list
# with `theta` (the full ACOV on the (mu, vech(sigma)) scale) and `cor` (the off-diagonal
# correlation ACOV in utils::combn(p, 2) order, the meat the corrected two-stage standard errors
# reuse, on the same scale and layout as the continuous ADF covariance .adf_gamma()).
.fiml_saturated_acov <- function(data, mu, sigma, patterns = NULL) {

  # No valid information without a positive-definite covariance (mirrors .fiml_loglik(); a
  # principal submatrix of a positive-definite matrix is positive definite, so every pattern's
  # Sigma_oo is then safe).
  if (any(eigen(sigma, symmetric = TRUE, only.values = TRUE)$values < .Machine$double.eps)) {
    cli::cli_abort(
      c("The covariance matrix is not positive definite, so the FIML information is undefined.",
        "i" = "A variable may be (near-)constant or collinear."),
      class = "efa_fiml_not_posdef"
    )
  }

  Y <- as.matrix(data)
  p <- ncol(Y)
  nms <- colnames(Y)
  if (is.null(nms)) nms <- names(mu)
  if (is.null(nms)) nms <- paste0("V", seq_len(p))

  # Group the rows once so the information accumulates over the same missingness patterns
  # (and the same fully-missing-row filtering) as the log-likelihood and the score.
  if (is.null(patterns)) {
    obs <- !is.na(Y)
    keep <- rowSums(obs) > 0L
    Y <- Y[keep, , drop = FALSE]
    patterns <- .fiml_patterns(obs[keep, , drop = FALSE])
  }

  pstar <- p * (p + 1L) / 2L

  # Index pairs (row >= col) of vech(sigma) in column-major order, shared below by the theta
  # labels and the standardisation Jacobian.
  sds <- sqrt(diag(sigma))
  ij <- which(lower.tri(diag(p), diag = TRUE), arr.ind = TRUE)

  info <- .fiml_saturated_information(Y, mu, sigma, patterns = patterns)

  # A singular information signals genuine near-degeneracy (coverage is validated upstream in
  # .fiml_em_moments()); refuse with its own classed condition -- distinct from the non-positive-
  # definite-covariance abort above, which is a different failure -- rather than a bare solve()
  # error.
  ch <- try(chol(info), silent = TRUE)
  if (inherits(ch, "try-error") || rcond(info) < .Machine$double.eps) {
    cli::cli_abort(
      c("The FIML information matrix is singular, so the saturated asymptotic covariance is undefined.",
        "i" = "A variable or variable pair may be (near-)degenerate or have very low coverage."),
      class = "efa_fiml_singular_information"
    )
  }
  Omega <- chol2inv(ch)
  Omega <- (Omega + t(Omega)) / 2

  # Label theta: means by variable name, vech(sigma) by "name_i-name_j" in the column-major
  # lower-triangle order (ij computed above).
  theta_lab <- c(nms, paste(nms[ij[, 1L]], nms[ij[, 2L]], sep = "-"))
  dimnames(Omega) <- list(theta_lab, theta_lab)

  # Correlation-scale variant. Take the vech(sigma) block of Omega (the MARGINAL covariance of
  # the covariance estimates -- not the inverse of the information's sigma-block -- so the
  # missingness-induced mean/covariance dependence is carried through), then standardise it by
  # the delta method to the off-diagonal correlations r_ij = sigma_ij / sqrt(sigma_ii sigma_jj).
  sig_idx <- p + seq_len(pstar)
  Omega_sig <- Omega[sig_idx, sig_idx, drop = FALSE]

  # Standardisation Jacobian J = d r_ij / d vech(sigma): rows are the off-diagonal pairs in
  # utils::combn() order, columns the vech(sigma) positions. vech_pos maps an element (row, col),
  # row >= col, to its column-major vech index (ij computed above). The two variance terms are the
  # delta-method correction (the continuous analogue of .adf_gamma()'s -1/2 r (w_i^2 + w_j^2)).
  vech_pos <- matrix(0L, p, p)
  vech_pos[ij] <- seq_len(pstar)
  pairs <- utils::combn(p, 2L)
  J <- matrix(0, ncol(pairs), pstar)
  for (t in seq_len(ncol(pairs))) {
    i <- pairs[1L, t]; j <- pairs[2L, t]                 # i < j
    rij <- sigma[i, j] / (sds[i] * sds[j])
    J[t, vech_pos[j, i]] <- 1 / (sds[i] * sds[j])        # d r / d sigma_ij  (lower-tri (j, i))
    J[t, vech_pos[i, i]] <- -0.5 * rij / sigma[i, i]
    J[t, vech_pos[j, j]] <- -0.5 * rij / sigma[j, j]
  }
  Omega_R <- J %*% Omega_sig %*% t(J)
  Omega_R <- (Omega_R + t(Omega_R)) / 2
  pair_lab <- .pair_labels(nms)
  dimnames(Omega_R) <- list(pair_lab, pair_lab)

  list(theta = Omega, cor = Omega_R)
}

# Stage-1 FIML engine: the EM estimate of the saturated MVN mean/covariance from raw data
# with missing values, and the saturated observed-data log-likelihood. Numerics are
# cross-checked against lavaan's two-stage / FIML moments (missing = "ml").

# A missing-at-random fixture: column 1 is fully observed and drives the missingness in the
# others, so the mechanism depends only on observed data. The normal deviates are coloured by
# the Cholesky factor rather than by an eigendecomposition, because chol() is unique for a
# positive definite matrix (the positive-diagonal convention pins it) whereas eigenvector signs
# are not, so the same seed reproduces the same sample on every LAPACK build.
make_mar_data <- function(n = 600, seed = 456) {
  set.seed(seed)
  p <- 5
  Sig <- 0.5 ^ abs(outer(seq_len(p), seq_len(p), "-"))     # AR(1)-type positive-definite cov
  X <- matrix(stats::rnorm(n * p), n) %*% chol(Sig)
  colnames(X) <- paste0("V", seq_len(p))
  X[X[, 1] >  0.7, 2] <- NA
  X[X[, 1] < -0.7, 3] <- NA
  X[X[, 1] >  1.0, 4] <- NA
  X[X[, 1] < -1.0, 5] <- NA
  X
}

test_that("complete data reproduces the MLE covariance and the Pearson correlation", {
  skip_on_cran()

  set.seed(123)
  p <- 5
  Sig <- 0.5 ^ abs(outer(seq_len(p), seq_len(p), "-"))
  X <- matrix(stats::rnorm(400 * p), 400) %*% chol(Sig)
  colnames(X) <- paste0("V", seq_len(p))

  em <- .fiml_em_moments(X)
  n <- nrow(X)

  expect_true(em$converged)
  expect_equal(em$n, n)
  expect_equal(em$n_patterns, 1L)
  # With no missing data the EM fixed point is the ML covariance (1/n scaling) and its
  # standardisation is exactly the Pearson correlation.
  expect_equal(em$sigma, (n - 1) / n * stats::cov(X), tolerance = 1e-8)
  expect_equal(stats::cov2cor(em$sigma), stats::cor(X), tolerance = 1e-8)
})

test_that("FIML moments match lavaan two-stage (missing = 'ml') under MAR", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  X <- make_mar_data()
  df <- as.data.frame(X)
  em <- .fiml_em_moments(X)
  expect_true(em$converged)
  expect_gt(em$n_patterns, 1L)

  fit <- lavaan::lavCor(df, missing = "ml", output = "fit")
  ss <- lavaan::lavInspect(fit, "sampstat")
  expect_equal(em$sigma, ss$cov, tolerance = 1e-4, ignore_attr = TRUE)
  # testthat's numeric tolerance is relative, which is unstable for means close to zero.
  # The two independent optimisers agree to a few 1e-5 across BLAS implementations, so
  # assert the intended absolute tolerance explicitly.
  expect_lt(max(abs(unname(em$mu) - unname(ss$mean))), 1e-4)

  Rml <- lavaan::lavCor(df, missing = "ml", output = "cor")
  expect_equal(stats::cov2cor(em$sigma), as.matrix(Rml),
               tolerance = 1e-4, ignore_attr = TRUE)
})

test_that("the saturated log-likelihood matches a direct per-case computation", {
  skip_on_cran()

  X <- make_mar_data()
  em <- .fiml_em_moments(X)

  # Transparent reference: sum the observed-subvector MVN log-density case by case (the
  # pattern-grouped helper must reproduce it).
  ref <- 0
  for (i in seq_len(nrow(X))) {
    o <- which(!is.na(X[i, ]))
    So <- em$sigma[o, o, drop = FALSE]
    d <- X[i, o] - em$mu[o]
    ref <- ref - 0.5 * (length(o) * log(2 * pi) +
                          as.numeric(determinant(So, logarithm = TRUE)$modulus) +
                          drop(crossprod(d, solve(So, d))))
  }
  expect_equal(.fiml_loglik(X, em$mu, em$sigma), ref, tolerance = 1e-8)
})

test_that("structural guards raise classed errors", {
  # A variable pair that is never jointly observed (zero coverage).
  zero_cov <- matrix(c(1, 2, NA, NA,
                       NA, NA, 3, 4), ncol = 2)
  expect_error(.fiml_em_moments(zero_cov), class = "efa_fiml_zero_coverage")

  # A non-numeric column.
  chr <- data.frame(a = c("x", "y", "z"), b = c(1, 2, 3))
  expect_error(.fiml_em_moments(chr), class = "efa_fiml_not_numeric")

  # A variable with no observed values at all.
  no_obs <- cbind(c(1, 2, 3, 4), c(NA, NA, NA, NA))
  expect_error(.fiml_em_moments(no_obs), class = "efa_fiml_no_observed")
})

test_that("too few cases for the number of variables aborts with the classed error", {
  # A second, independent route to efa_fiml_not_posdef: the sample size alone. With fewer
  # informative cases than variables the completed-data covariance cannot have full rank, so the
  # EM loses positive definiteness on its first M-step -- no constant or collinear variable
  # involved, which is what the constant-column test below covers. The gap is wide (rank 2 of 8)
  # rather than marginal, so the Cholesky verdict does not depend on rounding; the abort also
  # exercises the sample-size clause the message adds when n <= p.
  set.seed(11)
  small <- matrix(stats::rnorm(3 * 8), nrow = 3, ncol = 8)
  expect_error(.fiml_em_moments(small), class = "efa_fiml_not_posdef")
})

test_that("hitting max_iter warns and reports non-convergence", {
  skip_on_cran()

  X <- make_mar_data()
  expect_warning(em <- .fiml_em_moments(X, max_iter = 1L),
                 class = "efa_fiml_em_nonconvergence")
  expect_false(em$converged)
  expect_equal(em$iter, 1L)
})

test_that("the missing-data EM fixed point is stationary under an independent E-M step", {
  skip_on_cran()

  # The complete-data and log-likelihood tests do not pin the missing-data E-step (the
  # conditional-mean completion and the missing-by-missing covariance correction) to an
  # independent reference without lavaan. Here an independently coded per-case E-M step
  # (using solve(), not the engine's pattern-grouped chol2inv) must reproduce the converged
  # moments: a sign error or omission in the correction would move the engine's fixed point
  # away from this correct operator's, failing the check.
  X <- make_mar_data()
  em <- .fiml_em_moments(X, tol = 1e-10)
  mu <- unname(em$mu)
  sigma <- unname(em$sigma)
  n <- em$n
  p <- ncol(X)

  T1 <- numeric(p)
  T2 <- matrix(0, p, p)
  for (i in seq_len(nrow(X))) {
    o <- which(!is.na(X[i, ]))
    m <- which(is.na(X[i, ]))
    yc <- numeric(p)
    yc[o] <- X[i, o]
    extra <- matrix(0, p, p)
    if (length(m)) {
      B <- sigma[m, o, drop = FALSE] %*% solve(sigma[o, o, drop = FALSE])
      yc[m] <- mu[m] + B %*% (X[i, o] - mu[o])
      extra[m, m] <- sigma[m, m, drop = FALSE] - B %*% sigma[o, m, drop = FALSE]
    }
    T1 <- T1 + yc
    T2 <- T2 + tcrossprod(yc) + extra
  }
  mu_ref <- T1 / n
  sigma_ref <- T2 / n - tcrossprod(mu_ref)

  expect_equal(mu, mu_ref, tolerance = 1e-7)
  expect_equal(sigma, sigma_ref, tolerance = 1e-7)
})

test_that("the FIML correlation is invariant to per-column rescaling", {
  skip_on_cran()

  # FIML correlations must not depend on the variables' measurement scale; an absolute
  # convergence tolerance would stop early on small-variance columns and return a wrong
  # correlation. Rescaling each column must leave cov2cor(sigma) unchanged.
  X <- make_mar_data()
  em1 <- .fiml_em_moments(X)
  Xs <- sweep(X, 2L, c(1e-3, 10, 1, 1e2, 0.5), "*")
  em2 <- .fiml_em_moments(Xs)

  expect_true(em1$converged && em2$converged)
  expect_equal(stats::cov2cor(em1$sigma), stats::cov2cor(em2$sigma),
               tolerance = 1e-5, ignore_attr = TRUE)
})

test_that("a non-positive-definite covariance aborts with a classed error", {
  set.seed(11)
  degenerate <- cbind(stats::rnorm(60), rep(3, 60), stats::rnorm(60))   # column 2 is constant
  expect_error(.fiml_em_moments(degenerate), class = "efa_fiml_not_posdef")
})

test_that("the diagonal-init fallback (too few complete cases) still converges", {
  skip_on_cran()

  set.seed(202)
  p <- 4
  n <- 300
  Z <- matrix(stats::rnorm(n * p), n) %*%
    chol(0.4 ^ abs(outer(seq_len(p), seq_len(p), "-")))
  colnames(Z) <- paste0("V", seq_len(p))
  # One NA per row (rotating column) leaves only two complete rows (< p + 1), forcing the
  # diagonal-init fallback, while every pair keeps ample joint coverage.
  for (i in seq_len(n - 2L)) Z[i, (i %% p) + 1L] <- NA

  expect_lt(sum(stats::complete.cases(Z)), p + 1L)       # the fallback path is exercised
  em <- .fiml_em_moments(Z)
  expect_true(em$converged)
  expect_true(all(eigen(em$sigma, symmetric = TRUE, only.values = TRUE)$values > 0))
})

test_that("fully-missing rows are dropped from n and the log-likelihood", {
  skip_on_cran()

  X <- make_mar_data()
  Xd <- rbind(X, NA)                                      # append an all-NA row
  em <- .fiml_em_moments(Xd)

  expect_equal(em$n, nrow(X))
  expect_lt(em$n, nrow(Xd))
  # .fiml_loglik ignores the all-NA row too, so it returns the same value with or without it.
  expect_equal(.fiml_loglik(Xd, em$mu, em$sigma),
               .fiml_loglik(X, em$mu, em$sigma), tolerance = 1e-10)
})

test_that("a non-positive or non-integer max_iter is rejected", {
  X <- matrix(c(1, 2, 3, 4, 5, 6), ncol = 2)
  expect_error(.fiml_em_moments(X, max_iter = 0L), class = "efa_fiml_bad_max_iter")
  expect_error(.fiml_em_moments(X, max_iter = 2.5), class = "efa_fiml_bad_max_iter")
})

test_that("a non-positive tol is rejected", {
  X <- matrix(c(1, 2, 3, 4, 5, 6), ncol = 2)
  expect_error(.fiml_em_moments(X, tol = 0), class = "efa_fiml_bad_tol")
  expect_error(.fiml_em_moments(X, tol = -1), class = "efa_fiml_bad_tol")
})

# cor_method = "fiml": two-stage / full-information ML correlations wired into EFA().
# The EM moments engine itself is covered in test-fiml-moments.R; these tests cover the
# routing through .prepare_cor_input() and EFA(), the point estimates (which match
# psych::corFiml() -> psych::fa() and lavaan(missing = "two.stage"), not lavaan::efa()),
# and the classed conditions guarding the unsupported input/option combinations.

# A clean two-factor population covariance: items 1..p/2 load on factor 1 at `loadings[1]`, the
# rest on factor 2 at `loadings[2]`, with unit diagonal.
fiml_pop_cov <- function(p = 6, loadings = c(0.7, 0.7)) {
  L <- matrix(0, p, 2)
  half <- p / 2
  L[seq_len(half), 1] <- loadings[1]
  L[(half + 1):p, 2] <- loadings[2]
  S <- tcrossprod(L)
  diag(S) <- 1
  S
}

# A missing-at-random fixture: column 1 is fully observed and drives the missingness in the
# others, so the mechanism depends only on observed data. The multivariate normal draw goes
# through the Cholesky factor rather than an eigendecomposition: at the default (equal) loadings
# the population covariance has repeated eigenvalues, whose eigenvector basis is mathematically
# undetermined and therefore settled by rounding, while chol() is unique for a positive definite
# matrix, so the same seed yields the same sample on every LAPACK build.
fiml_mar_data <- function(n = 800, seed = 456, loadings = c(0.7, 0.7)) {
  set.seed(seed)
  X <- matrix(stats::rnorm(n * 6), n) %*% chol(fiml_pop_cov(6, loadings))
  colnames(X) <- paste0("V", seq_len(6))
  X[X[, 1] >  0.8, 2] <- NA
  X[X[, 1] < -0.8, 3] <- NA
  X[X[, 1] >  1.2, 4] <- NA
  X[X[, 1] < -1.2, 5] <- NA
  X
}

# The same fixture with the two factors at different strengths. At the default equal loadings the
# canonical variances diag(Lambda' Psi^-1 Lambda) that identify the unrotated ML solution coincide
# exactly in the population: the loadings can then be rotated within that plane without leaving the
# identifying constraint, so the unrotated orientation -- and with it the unrotated loading standard
# errors -- has no well-defined sampling distribution, and a fitted solution sits right at the
# detector's floor. Every test of an unrotated analytic standard error needs a solution where that
# orientation is determined, which separating the two strengths provides (canonical variances 4.83
# against 0.90, a transversal an order of magnitude clear of the floor). Rotated quantities are
# gauge-invariant and need no such care.
fiml_mar_data_2f <- function(n = 800, seed = 456) {
  fiml_mar_data(n = n, seed = seed, loadings = c(0.8, 0.5))
}

# A single-factor MAR fixture: a clean SE / chi-square oracle, with no rotational indeterminacy.
fiml_mar_data_1f <- function(n = 1000, p = 6, load = 0.65, seed = 11) {
  set.seed(seed)
  L <- matrix(load, p, 1)
  S <- tcrossprod(L); diag(S) <- 1
  X <- matrix(stats::rnorm(n * p), n) %*% chol(S)
  colnames(X) <- paste0("V", seq_len(p))
  X[X[, 1] >  0.8, 2] <- NA
  X[X[, 1] < -0.8, 3] <- NA
  X[X[, 1] >  1.0, 4] <- NA
  X[X[, 1] < -1.0, 5] <- NA
  X
}

test_that("complete data: FIML loadings equal the Pearson-correlation loadings", {
  skip_on_cran()

  set.seed(123)
  X <- matrix(stats::rnorm(500 * 6), 500) %*% chol(fiml_pop_cov(6))
  colnames(X) <- paste0("V", seq_len(6))

  fiml <- EFA(X, n_factors = 2, method = "ML", cor_method = "fiml")
  pear <- EFA(X, n_factors = 2, method = "ML", cor_method = "pearson")

  # With no missing data the FIML moments reduce to the sample correlation, so the
  # extracted loadings coincide.
  expect_equal(unclass(fiml$unrot_loadings), unclass(pear$unrot_loadings),
               tolerance = 1e-6, ignore_attr = TRUE)
})

test_that("MAR-missing data: FIML loadings match psych::fa(corFiml)", {
  skip_on_cran()
  skip_if_not_installed("psych")

  X <- fiml_mar_data()
  k <- 2L

  efa <- EFA(X, n_factors = k, method = "ML", cor_method = "fiml")

  # Reference: psych's FIML correlation analysed by psych's ML factor analysis, unrotated.
  R_ref  <- suppressWarnings(psych::corFiml(X))
  fa_ref <- suppressWarnings(psych::fa(R_ref, nfactors = k, fm = "ml",
                                       rotate = "none", n.obs = nrow(X)))
  L_ref <- unclass(fa_ref$loadings)
  L     <- unclass(efa$unrot_loadings)

  # Sign- and column-order-invariant comparison via the reproduced common-factor
  # correlations (L %*% t(L)); the FIML correlation and the ML extraction match the
  # established corFiml -> fa workflow. Both come from the same EM -> ML workflow, so the
  # reproduced common parts agree far tighter than the eyeball 0.01: a tolerance two orders
  # of magnitude looser than they differ would let a real drift pass.
  expect_lt(max(abs(tcrossprod(L) - tcrossprod(L_ref))), 1e-3)
})

test_that("FIML resolves N to the EM case count (rows with any observed value)", {
  skip_on_cran()

  X <- fiml_mar_data()
  Xd <- rbind(X, NA)                                     # an all-missing row carries no info

  efa <- EFA(Xd, n_factors = 2, method = "ML", cor_method = "fiml")

  # N is the EM case count: every row with at least one observed value, independent of `use`.
  # This pins the FIML branch of the N resolution, distinguishing em$n from both the raw row
  # count and the (smaller) complete-case count.
  expect_equal(efa$settings$N, nrow(X))                  # the all-NA row is dropped
  expect_lt(efa$settings$N, nrow(Xd))                    # ... so N is not the raw row count
  expect_gt(efa$settings$N, sum(stats::complete.cases(Xd)))  # ... nor the complete-case count
})

test_that("the FIML correlation matches lavaan two-stage (missing = 'ml')", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  X <- fiml_mar_data()
  efa <- EFA(X, n_factors = 2, method = "ML", cor_method = "fiml")

  # EFA() analyses the FIML two-stage correlation, not a pairwise/Pearson one.
  Rml <- lavaan::lavCor(as.data.frame(X), missing = "ml", output = "cor")
  expect_equal(unclass(efa$orig_R), as.matrix(Rml),
               tolerance = 1e-3, ignore_attr = TRUE)
})

test_that("FIML fit indices are the corrected (Satorra-Bentler) two-stage statistic", {
  skip_on_cran()

  X <- fiml_mar_data()
  k <- 2L
  efa <- EFA(X, n_factors = k, method = "ML", cor_method = "fiml")
  fi <- efa$fit_indices

  # The reported chi-square is the scaled-shifted Satorra-Bentler statistic a * T + b, not the
  # plain two-stage likelihood-ratio statistic (which is not asymptotically chi^2(df) under the
  # two-stage estimator). Pin the self-consistency of the scaled-statistic block.
  expect_identical(fi$chi_scaled_type, "scaled.shifted")
  expect_equal(fi$chi, fi$chi_scaling * fi$chi_unscaled + fi$chi_shift, tolerance = 1e-8)
  expect_equal(fi$df, ((6 - k)^2 - (6 + k)) / 2)        # = 4, the EFA correlation-structure df
  expect_equal(fi$df_null, 6 * 5 / 2)                   # = 15
  expect_true(all(is.finite(c(fi$chi, fi$chi_unscaled, fi$chi_mean_adjusted, fi$chi_mean_var,
                              fi$df_mean_var, fi$chi_null, fi$CFI, fi$TLI, fi$RMSEA))))
  expect_gt(fi$chi, 0)

  # AIC/BIC/ECVI are likelihood-ratio chi-square criteria with no meaning for the moment-scaled
  # statistic, so they are NA (as on the sandwich path), not the chi-square-based forms.
  expect_true(all(is.na(c(fi$AIC, fi$BIC, fi$ECVI))))
})

# Force the Stage-1 saturated covariance to fail, so the corrected two-stage statistic cannot be
# formed. It is the covariance that is degenerate, not the model, so the plain two-stage
# likelihood-ratio statistic is still available -- which is exactly the situation the fallback
# contract covers. Used by the tests below to reach it deterministically.
.mock_fiml_acov_failure <- function(env = parent.frame()) {
  testthat::local_mocked_bindings(
    .fiml_saturated_acov = function(...) {
      cli::cli_abort("forced degenerate saturated covariance",
                     class = "efa_fiml_singular_information")
    },
    .env = env
  )
}

test_that("a formable correction is tagged as the corrected statistic, with no fallback warning", {
  skip_on_cran()

  expect_no_warning(
    fi <- efa_fit(fiml_mar_data(), n_factors = 2, estimator = "ML",
                  cor_method = "fiml")$fit_indices,
    class = "efa_fiml_uncorrected_chisq"
  )

  # The success side of the tag: `chi` is the corrected statistic, and the components that
  # describe the correction are present alongside it.
  expect_identical(fi$chi_scaled_type, "scaled.shifted")
  expect_true(all(is.finite(c(fi$chi_scaling, fi$chi_shift, fi$chi_unscaled))))
})

test_that("a degenerate saturated covariance falls back to the tagged plain two-stage LRT", {
  skip_on_cran()

  X <- fiml_mar_data()
  .mock_fiml_acov_failure()

  expect_warning(
    fit <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml"),
    class = "efa_fiml_uncorrected_chisq"
  )
  fi <- fit$fit_indices

  # Tagged, so no consumer can read the substituted statistic as the Satorra-Bentler one ...
  expect_identical(fi$chi_scaled_type, "uncorrected.lrt")
  # ... and none of the scaling components ships, because nothing was rescaled.
  expect_null(fi$chi_scaling)
  expect_null(fi$chi_shift)
  expect_null(fi$chi_unscaled)
  expect_null(fi$chi_mean_adjusted)

  # The reported statistic is the plain two-stage likelihood ratio, referenced to the
  # chi-square(df) tail, and the block built on it is complete rather than NA-ed away: the
  # statistic is approximate under the two-stage estimator, not meaningless.
  em <- .fiml_em_moments(X)
  ref <- .gof_fiml_chisq(tcrossprod(unclass(fit$unrot_loadings)), N = fit$settings$N,
                         method = "ML", df = fi$df, m = ncol(X),
                         fiml = list(data = X, mu = em$mu, sigma = em$sigma, logl = em$logl))
  expect_equal(fi$chi, ref$chi, tolerance = 1e-8)
  expect_equal(fi$p_chi, stats::pchisq(fi$chi, fi$df, lower.tail = FALSE), tolerance = 1e-10)
  expect_true(all(is.finite(c(fi$CFI, fi$TLI, fi$RMSEA))))

  # AIC/BIC/ECVI stay withheld: they are likelihood-ratio criteria the two-stage estimator
  # cannot support, whether or not the correction was formed.
  expect_true(all(is.na(c(fi$AIC, fi$BIC, fi$ECVI))))
})

test_that("the fallback chi-square prints as uncorrected, never as scaled", {
  skip_on_cran()
  local_reproducible_output()

  .mock_fiml_acov_failure()
  expect_warning(
    fit <- efa_fit(fiml_mar_data(), n_factors = 2, estimator = "ML", cor_method = "fiml"),
    class = "efa_fiml_uncorrected_chisq"
  )

  printed <- cli::ansi_strip(capture.output(print(fit)))
  expect_true(any(grepl("uncorrected χ²(", printed, fixed = TRUE)))
  expect_false(any(grepl("scaled χ²(", printed, fixed = TRUE)))
})

test_that("a just-identified FIML fit is left untagged rather than called uncorrected", {
  skip_on_cran()

  # With df = 0 there is no chi-square test to mislabel, so the fallback tag and its warning
  # must not fire: the tag exists to distinguish a corrected statistic from an uncorrected
  # one, and this fit has neither.
  set.seed(77)
  L <- matrix(0.7, 3, 1)
  S <- tcrossprod(L); diag(S) <- 1
  X <- matrix(stats::rnorm(500 * 3), 500) %*% chol(S)
  colnames(X) <- paste0("V", seq_len(3))
  X[X[, 1] >  0.8, 2] <- NA
  X[X[, 1] < -0.8, 3] <- NA

  # Collect the condition classes rather than suppressing wholesale: the fit legitimately
  # raises efa_just_identified, and a blanket suppressWarnings() would also hide the very
  # condition this test asserts is absent.
  seen <- character()
  fi <- withCallingHandlers(
    efa_fit(X, n_factors = 1, estimator = "ML", cor_method = "fiml")$fit_indices,
    warning = function(w) {
      seen <<- c(seen, class(w))
      invokeRestart("muffleWarning")
    }
  )
  expect_true("efa_just_identified" %in% seen)
  expect_false("efa_fiml_uncorrected_chisq" %in% seen)
  expect_equal(fi$df, 0)
  expect_null(fi$chi_scaled_type)
})

test_that("FIML fit indices match lavaan two-stage (Satorra-Bentler)", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  # A single-factor model: a clean oracle where EFAtools' quadratic-form statistic is close to
  # lavaan's exact two-stage likelihood ratio (the gap widens with misfit, so a saturated-ish
  # cross-loading model is a looser comparison).
  X <- fiml_mar_data_1f()
  efa <- EFA(X, n_factors = 1, method = "ML", cor_method = "fiml")
  fi  <- efa$fit_indices

  vn <- colnames(X)
  mod <- paste0("f =~ ", paste(vn, collapse = " + "))
  lfit <- lavaan::cfa(mod, data = as.data.frame(X), std.lv = TRUE, missing = "two.stage")
  sb <- lavaan::lavInspect(lfit, "test")[["satorra.bentler"]]

  # lavaan's two-stage estimator reports the Satorra-Bentler (mean-adjusted) corrected statistic;
  # EFAtools keeps it as `chi_mean_adjusted` (the reported `chi` is the scaled-shifted variant of
  # the same correction). The mean-adjusted statistic is the quadratic-form approximation of
  # lavaan's exact two-stage likelihood-ratio statistic, matching within a few percent.
  expect_equal(fi$df, sb$df)                            # EFA correlation-structure df match exactly
  expect_equal(fi$chi_mean_adjusted, sb$stat, tolerance = 0.05)

  # The robust CFI/TLI/RMSEA come from the scaled model and baseline statistics and sit near the
  # perfect-fit boundary for this correctly specified model, so they match lavaan's scaled measures.
  fm <- lavaan::fitMeasures(lfit, c("cfi.scaled", "tli.scaled", "rmsea.scaled"))
  expect_lt(abs(fi$CFI   - unname(fm[["cfi.scaled"]])),   0.02)
  expect_lt(abs(fi$TLI   - unname(fm[["tli.scaled"]])),   0.02)
  expect_lt(abs(fi$RMSEA - unname(fm[["rmsea.scaled"]])), 0.02)
})

test_that("FIML leaves the chi-square NA for PAF but keeps the residual indices", {
  skip_on_cran()

  X <- fiml_mar_data()
  efa <- EFA(X, n_factors = 2, method = "PAF", cor_method = "fiml")
  fi  <- efa$fit_indices

  # PAF has no discrepancy-based chi-square (the package convention), so the chi-square-derived
  # indices are NA under FIML as on every other cor_method; the residual indices are still
  # computed against the EM correlation.
  expect_true(is.na(fi$chi))
  expect_true(is.na(fi$CFI))
  expect_true(is.na(fi$RMSEA))
  expect_true(is.finite(fi$CAF))
  expect_true(is.finite(fi$SRMR))
})

test_that("FIML leaves AIC/BIC/ECVI NA for a just-identified (df = 0) fit", {
  skip_on_cran()

  # A one-factor model on three indicators is just-identified (df = 0), so the corrected two-stage
  # statistic is not formed and .gof() falls back to the plain likelihood-ratio chi-square. AIC,
  # BIC, and ECVI are likelihood-ratio criteria with no meaning under the two-stage estimator, so
  # they stay NA on this fallback too -- not the finite chi-square-derived values the plain LRT
  # would otherwise produce.
  set.seed(77)
  L <- matrix(0.7, 3, 1)
  S <- tcrossprod(L); diag(S) <- 1
  X <- matrix(stats::rnorm(500 * 3), 500) %*% chol(S)
  colnames(X) <- paste0("V", seq_len(3))
  X[X[, 1] >  0.8, 2] <- NA
  X[X[, 1] < -0.8, 3] <- NA

  # The just-identified df = 0 model warns (efa_just_identified); assert it so the fit-index
  # check below runs on a clean condition stack.
  expect_warning(
    fi <- EFA(X, n_factors = 1, method = "ML", cor_method = "fiml")$fit_indices,
    class = "efa_just_identified"
  )
  expect_equal(fi$df, 0)
  expect_true(all(is.na(c(fi$AIC, fi$BIC, fi$ECVI))))
})

test_that("FIML NA's the whole chi-square block for a non-PD model-implied matrix", {
  skip_on_cran()

  X  <- fiml_mar_data()
  em <- .fiml_em_moments(X)

  # A Heywood loading (here > 1) makes the model-implied correlation LL' (unit diagonal) and hence
  # Sigma_model non-positive-definite, so the model deviance is undefined. The model and baseline
  # statistics must NA out together (as on every other cor_method), not report a finite baseline
  # against an undefined model.
  LLt <- tcrossprod(matrix(c(1.1, 1.0, 0.2, 0.1, 0.1, 0.1), ncol = 1))
  fc <- .gof_fiml_chisq(LLt, N = em$n, method = "ML", df = 9, m = 6,
                        fiml = list(data = X, mu = em$mu, sigma = em$sigma, logl = em$logl))

  expect_true(is.na(fc$chi))
  expect_true(is.na(fc$chi_null))
  expect_true(is.na(fc$df_null))
  expect_true(is.na(fc$chi_cfi))
  expect_true(is.na(fc$chi_null_cfi))
})

test_that("FIML fit indices ignore a fully-missing row (point-estimate filter)", {
  skip_on_cran()

  X  <- fiml_mar_data()
  Xd <- rbind(X, NA)                                    # an all-missing row carries no information

  fi  <- EFA(X,  n_factors = 2, method = "ML", cor_method = "fiml")$fit_indices
  fid <- EFA(Xd, n_factors = 2, method = "ML", cor_method = "fiml")$fit_indices

  # .gof_fiml_chisq() re-applies the EM's fully-missing-row filter to fiml$data, so the all-NA row
  # drops out of the model and baseline log-likelihoods exactly as it does from the saturated one:
  # the likelihood-ratio chi-squares are identical with or without it. A regression dropping that
  # filter would compute logl_model/logl_null over a row set wider than logl_sat and diverge here.
  expect_equal(fid$chi, fi$chi, tolerance = 1e-8)
  expect_equal(fid$chi_null, fi$chi_null, tolerance = 1e-8)
})

test_that("FIML np-boot fit-index SEs come from the corrected two-stage chi-square", {
  skip_on_cran()

  X <- fiml_mar_data()
  efa <- suppressWarnings(suppressMessages(
    EFA(X, n_factors = 2, method = "ML", cor_method = "fiml",
        se = "np-boot", b_boot = 20, seed = 42)
  ))

  # Each resample re-estimates the EM moments and forms the same corrected (Satorra-Bentler)
  # two-stage statistic as the point estimate, so the chi-square-derived fit-index SEs are
  # populated and finite (not the NA they would be if the bootstrap path silently dropped the FIML
  # moments). AIC/BIC/ECVI are NA for the scaled statistic, so their bootstrap SE is NA too.
  fi_se <- efa$SE$fit_indices
  expect_true(all(is.finite(fi_se[c("chi", "CFI", "TLI", "RMSEA")])))
  expect_gt(fi_se[["chi"]], 0)
})

test_that("a converged FIML fit reports its Stage-1 EM diagnostics", {
  skip_on_cran()

  X <- fiml_mar_data()
  fit <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml")

  expect_named(fit$fiml, c("converged", "iter", "n_patterns", "n"))
  expect_true(fit$fiml$converged)
  expect_true(fit$fiml$iter >= 1L && fit$fiml$iter < 500L)
  # The pattern count and the informative-row count are the EM's own, so they must match a
  # direct read of the missingness structure rather than the raw data dimensions.
  obs <- !is.na(X)
  keep <- rowSums(obs) > 0L
  expect_equal(fit$fiml$n_patterns, length(.fiml_patterns(obs[keep, , drop = FALSE])))
  expect_equal(fit$fiml$n, sum(keep))
  expect_equal(fit$fiml$n, fit$settings$N)

  # The slot belongs to the FIML route alone; no other correlation method estimates them.
  pear <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "pearson")
  expect_null(pear$fiml)
})

test_that("estimate_control's FIML EM knobs reach the moment estimation", {
  skip_on_cran()

  X <- fiml_mar_data()

  # A cap of two iterations is reached on this fixture, so the warning fires and the fit
  # records that the analysed matrix is the last iterate rather than the FIML estimate.
  expect_warning(
    capped <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml",
                      estimate_control = estimate_control(fiml_max_iter = 2)),
    class = "efa_fiml_em_nonconvergence"
  )
  expect_false(capped$fiml$converged)
  expect_equal(capped$fiml$iter, 2L)

  # The knobs are not inert: a capped EM stops at a different correlation matrix from the
  # converged one, and a looser tolerance converges in fewer iterations.
  full <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml")
  expect_false(isTRUE(all.equal(unclass(capped$orig_R), unclass(full$orig_R))))

  loose <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml",
                   estimate_control = estimate_control(fiml_tol = 1e-2))
  expect_true(loose$fiml$converged)
  expect_lt(loose$fiml$iter, full$fiml$iter)
})

test_that("a non-converged FIML EM is visible in the printed output", {
  skip_on_cran()
  local_reproducible_output()

  expect_warning(
    capped <- efa_fit(fiml_mar_data(), n_factors = 2, estimator = "ML", cor_method = "fiml",
                      estimate_control = estimate_control(fiml_max_iter = 2)),
    class = "efa_fiml_em_nonconvergence"
  )

  printed <- cli::ansi_strip(capture.output(print(capped)))
  expect_true(any(grepl("FIML EM stopped after 2 iterations without converging",
                        printed, fixed = TRUE)))
  # summary() renders the same header, so the diagnostic is surfaced there as well.
  expect_match(cli::ansi_strip(format(summary(capped))),
               "without converging", fixed = TRUE, all = FALSE)

  # A converged fit says nothing, so the line marks the exception rather than the rule.
  ok <- cli::ansi_strip(capture.output(print(
    efa_fit(fiml_mar_data(), n_factors = 2, estimator = "ML", cor_method = "fiml"))))
  expect_false(any(grepl("without converging", ok, fixed = TRUE)))
})

test_that("FIML aborts on unsupported input/option combinations", {
  raw <- matrix(c(1, 2, 3, NA, 5, 6,
                  2, 1, 4,  3, 2, 5,
                  3, 5, 2,  4, 1, 6), ncol = 3)
  colnames(raw) <- c("a", "b", "c")
  cmat <- stats::cor(raw[stats::complete.cases(raw), ])

  # A correlation matrix carries no cases to estimate the FIML moments from.
  expect_error(EFA(cmat, n_factors = 1, N = 50, cor_method = "fiml"),
               class = "efa_fiml_needs_raw")
  # DWLS needs a polychoric asymptotic covariance the FIML correlation cannot supply.
  expect_error(EFA(raw, n_factors = 1, method = "DWLS", cor_method = "fiml"),
               class = "efa_fiml_unsupported_method")
  # PAF carries no Stage-2 discrepancy weight for the corrected two-stage sandwich, so analytic
  # standard errors are rejected for it under FIML (the bootstrap stays available for PAF).
  expect_error(EFA(raw, n_factors = 1, method = "PAF", cor_method = "fiml", se = "sandwich"),
               class = "efa_se_unsupported")
  expect_error(EFA(raw, n_factors = 1, method = "PAF", cor_method = "fiml", se = "information"),
               class = "efa_se_unsupported")
})


# Corrected two-stage (Yuan & Bentler, 2000; Savalei & Bentler, 2009) analytic standard errors for
# cor_method = "fiml" (se = "information"/"sandwich" both route to the same two-stage sandwich).

# Independent reference: the two-stage sandwich with the gauge-constraint Jacobian taken by CENTRAL
# FINITE DIFFERENCES of the orientation function, sharing no code with .se_fiml_core()'s analytic
# constraint (.se_sandwich_constraint). H is the normal-theory ML weight; the meat is the FIML
# saturated covariance Omega (already on the variance scale Var(r-hat)), put on the unit
# asymptotic-variance scale by N and the covariance divided by N - 1, the package-wide small-sample
# convention shared with the expected-information and polychoric/ADF sandwich paths.
.ref_fiml_loading_se <- function(L, Omega, gauge, N, weight = "ML") {
  p <- nrow(L); k <- ncol(L); pk <- p * k
  pr <- utils::combn(p, 2L); pii <- pr[1, ]; pjj <- pr[2, ]; n <- ncol(pr)
  Delta <- matrix(0, n, pk)
  for (f in seq_len(k)) {
    Delta[cbind(seq_len(n), (f - 1L) * p + pii)] <- L[pjj, f]
    Delta[cbind(seq_len(n), (f - 1L) * p + pjj)] <- L[pii, f]
  }
  # Stage-2 weight matched to the estimator: ULS the identity, ML the normal-theory weight.
  if (weight == "ULS") {
    H <- diag(n)
  } else {
    Sig <- tcrossprod(L); diag(Sig) <- 1; P <- solve(Sig)
    H <- matrix(0, n, n)
    for (s in seq_len(n)) { a <- pii[s]; b <- pjj[s]; H[, s] <- 0.5 * (P[pii, a] * P[pjj, b] + P[pii, b] * P[pjj, a]) }
  }
  A <- crossprod(Delta, H %*% Delta)
  gfun <- function(par) {
    Lm <- matrix(par, p, k)
    M <- if (gauge == "LtL") crossprod(Lm) else { ps <- 1 - rowSums(Lm^2); crossprod(Lm, Lm / ps) }
    M[upper.tri(M)]
  }
  nc <- k * (k - 1L) / 2L; Cmat <- matrix(0, nc, pk); h <- 1e-6; par0 <- as.vector(L)
  for (j in seq_len(pk)) { pp <- par0; pp[j] <- pp[j] + h; pm <- par0; pm[j] <- pm[j] - h; Cmat[, j] <- (gfun(pp) - gfun(pm)) / (2 * h) }
  Aug <- if (nc > 0) rbind(cbind(A, t(Cmat)), cbind(Cmat, matrix(0, nc, nc))) else A
  Ab <- solve(Aug)[seq_len(pk), seq_len(pk)]
  HD <- H %*% Delta
  V_AA <- Ab %*% crossprod(HD, (N * Omega) %*% HD) %*% Ab / (N - 1)
  matrix(sqrt(diag(V_AA)), p, k)
}

test_that("FIML unrotated loading and uniqueness SEs match lavaan two-stage", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  X <- fiml_mar_data_1f()
  efa <- EFA(X, n_factors = 1, method = "ML", cor_method = "fiml", se = "sandwich")

  vn <- colnames(X)
  mod <- paste0("f =~ ", paste(vn, collapse = " + "))
  lfit <- lavaan::cfa(mod, data = as.data.frame(X), std.lv = TRUE, missing = "two.stage")
  lam <- lavaan::standardizedSolution(lfit)
  lam <- lam[lam$op == "=~", ]
  L_lav <- lam$est.std; SE_lav <- lam$se
  sgn <- sign(sum(as.numeric(efa$unrot_loadings) * L_lav))

  # One factor: no rotational indeterminacy, so the standardized loadings (up to sign) and their
  # corrected two-stage SEs are directly comparable. The residual gap is the EM-vs-optimiser and
  # N-vs-(N-1) difference.
  expect_equal(as.numeric(efa$unrot_loadings) * sgn, L_lav, tolerance = 0.01)
  expect_equal(as.numeric(efa$SE$unrot_loadings), SE_lav, tolerance = 0.03)
  expect_true(all(is.finite(efa$SE$uniquenesses)) && all(efa$SE$uniquenesses > 0))
})

test_that("FIML two-stage loading SEs match an independent reference sandwich", {
  skip_on_cran()

  # The unrotated loading SEs are gauge-dependent, so this needs the fixture whose rotational
  # orientation is determined.
  X <- fiml_mar_data_2f()
  efa <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml", se = "sandwich")
  L <- unclass(efa$unrot_loadings)

  em <- .fiml_em_moments(X)
  Omega <- .fiml_saturated_acov(X, em$mu, em$sigma)$cor
  # ML reports loadings in the Lambda' Psi^-1 Lambda orientation; the reference takes its gauge
  # constraint by central differences, sharing no code with .se_fiml_core().
  ref <- .ref_fiml_loading_se(L, Omega, "LtPiL", N = nrow(X))
  expect_equal(unclass(efa$SE$unrot_loadings), ref, tolerance = 1e-5, ignore_attr = TRUE)
})

test_that("FIML rotated loading SEs match lavaan two-stage under a supported rotation", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  X <- fiml_mar_data_2f()
  efa <- efa_fit(X, n_factors = 2, estimator = "ML", rotation = "geominQ",
                 cor_method = "fiml", se = "sandwich")

  # Match EFAtools' geominQ epsilon (default delta = 0.01) so the two rotations coincide; the
  # two-stage estimator carries the corrected rotated-loading SEs through standardizedSolution().
  if (utils::packageVersion("lavaan") >= "0.7.0") {
    lf <- lavaan::efa(as.data.frame(X), nfactors = 2,
                      rotation = list("geomin", geomin_epsilon = 0.01),
                      missing = "two.stage")
  } else {
    # lavaan < 0.7 used a separate rotation.args argument.
    lf <- lavaan::efa(as.data.frame(X), nfactors = 2, rotation = "geomin",
                      missing = "two.stage", rotation.args = list(geomin.epsilon = 0.01))
  }
  lfit <- if (is.list(lf) && !inherits(lf, "lavaan")) lf[[1]] else lf
  lam <- lavaan::standardizedSolution(lfit)
  lam <- lam[lam$op == "=~", ]
  SE_lav <- matrix(lam$se, nrow = 6, ncol = 2)
  EST_lav <- matrix(lam$est.std, nrow = 6, ncol = 2)

  SE_efa <- unclass(efa$SE$rot_loadings)
  EST_efa <- unclass(efa$rot_loadings)
  # The rotated column order is arbitrary; match EFAtools columns to lavaan by the salient pattern.
  cmap <- apply(abs(crossprod(EST_efa, EST_lav)), 1L, which.max)
  expect_setequal(cmap, 1:2)
  expect_equal(SE_efa[, cmap], SE_lav, tolerance = 0.05, ignore_attr = TRUE)
})

test_that("FIML sandwich SEs fill the analytic SE/CI schema (oblique)", {
  skip_on_cran()

  X <- fiml_mar_data_2f()
  efa <- efa_fit(X, n_factors = 2, estimator = "ML", rotation = "oblimin",
                 cor_method = "fiml", se = "sandwich")

  expect_named(efa$SE, c("unrot_loadings", "uniquenesses", "rot_loadings", "communalities",
                         "Phi", "Structure"))
  for (nm in c("unrot_loadings", "rot_loadings", "Structure")) {
    expect_true(all(is.finite(efa$SE[[nm]])), info = nm)
  }
  expect_true(all(efa$SE$unrot_loadings > 0))
  expect_length(efa$SE$uniquenesses, 6L)
  expect_true(all(is.finite(efa$SE$uniquenesses)))
  # Factor correlations: symmetric SE matrix with a fixed (zero-variance) unit diagonal. It carries
  # the factor names, so compare the diagonal on its values alone.
  expect_equal(unname(diag(efa$SE$Phi)), c(0, 0))
  expect_equal(efa$SE$Phi, t(efa$SE$Phi))
  expect_true(is.finite(efa$SE$Phi[1, 2]))
  # Communalities are rotation-invariant, so their SEs equal the unrotated uniqueness SEs.
  expect_equal(unname(efa$SE$communalities), unname(efa$SE$uniquenesses), tolerance = 1e-8)
  # Wald intervals bracket the point estimates.
  expect_true(all(efa$CI$unrot_loadings$lower <= unclass(efa$unrot_loadings)))
  expect_true(all(efa$CI$unrot_loadings$upper >= unclass(efa$unrot_loadings)))
})

test_that("FIML 'information' and 'sandwich' give the same corrected two-stage SEs", {
  skip_on_cran()

  X <- fiml_mar_data_2f()
  fit_i <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml", se = "information")
  fit_s <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml", se = "sandwich")
  # Under FIML both analytic settings return the corrected two-stage sandwich, so the SEs and the
  # persisted loading covariance are identical.
  expect_equal(fit_i$SE$unrot_loadings, fit_s$SE$unrot_loadings)
  expect_equal(fit_i$SE$uniquenesses, fit_s$SE$uniquenesses)
  expect_equal(fit_i$vcov_unrot_loadings, fit_s$vcov_unrot_loadings)
})

test_that("FIML ULS uses the estimator's own (identity) Stage-2 weight, not the ML weight", {
  skip_on_cran()

  X <- fiml_mar_data_2f()
  efa_u <- efa_fit(X, n_factors = 2, estimator = "ULS", cor_method = "fiml", se = "sandwich")
  L <- unclass(efa_u$unrot_loadings)
  expect_true(all(is.finite(efa_u$SE$unrot_loadings)) && all(efa_u$SE$unrot_loadings > 0))
  expect_true(all(is.finite(efa_u$SE$uniquenesses)))

  em <- .fiml_em_moments(X)
  Omega <- .fiml_saturated_acov(X, em$mu, em$sigma)$cor
  # ULS reports loadings in the Lambda' Lambda orientation and weights the discrepancy by the
  # identity, so the reference sandwich uses the identity weight in that same gauge.
  ref_uls <- .ref_fiml_loading_se(L, Omega, "LtL", N = nrow(X), weight = "ULS")
  expect_equal(unclass(efa_u$SE$unrot_loadings), ref_uls, tolerance = 1e-5, ignore_attr = TRUE)
  # In the same gauge, the SE must match the estimator-specific ULS reference more closely than
  # the ML-weighted sandwich reference.
  ref_ml <- .ref_fiml_loading_se(L, Omega, "LtL", N = nrow(X), weight = "ML")
  err_uls <- max(abs(unclass(efa_u$SE$unrot_loadings) - ref_uls))
  err_ml <- max(abs(unclass(efa_u$SE$unrot_loadings) - ref_ml))
  expect_lt(err_uls, err_ml)

  # 'information' and 'sandwich' coincide for ULS too (both route to the corrected sandwich).
  efa_ui <- efa_fit(X, n_factors = 2, estimator = "ULS", cor_method = "fiml", se = "information")
  expect_equal(efa_ui$SE$unrot_loadings, efa_u$SE$unrot_loadings)
})

test_that("FIML analytic SEs NA-fill with a classed warning when the covariance is unusable", {
  skip_on_cran()

  X <- fiml_mar_data()
  em <- .fiml_em_moments(X)
  fiml <- list(data = X, mu = em$mu, sigma = em$sigma)

  # A non-finite (NA) loading makes the model derivative and the parameter covariance undefined;
  # the unrotated SE schema must NA-fill with the classed efa_se_unreliable warning rather than
  # ship a finite SE next to an unusable covariance. This is the unusable-covariance route, which
  # is distinct from the boundary route tested above: there the covariance is fine and it is the
  # Wald approximation that is not.
  fo <- list(unrot_loadings = matrix(c(0.7, NA, 0.6, 0.5, 0.4, 0.3,
                                       0.1, 0.2, 0.15, 0.7, 0.6, 0.5), 6, 2),
             orig_R = stats::cov2cor(em$sigma))
  expect_warning(
    res <- .se_fiml(fo, rot_info = NULL, N = em$n, ci = 0.95, fiml = fiml, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(res$SE$unrot_loadings)))
  expect_true(all(is.na(res$vcov_unrot_loadings)))
})

test_that("the FIML sandwich picks its gauge from the loadings, not from the estimator label", {
  skip_on_cran()

  X <- fiml_mar_data_2f()
  em <- .fiml_em_moments(X)
  fiml <- list(data = X, mu = em$mu, sigma = em$sigma)
  Omega <- .fiml_saturated_acov(X, em$mu, em$sigma)$cor

  # ML-fitted loadings are reported in the Lambda' Psi^-1 Lambda orientation. Ask for the ULS
  # Stage-2 weight on those same loadings: the gauge must still be the one the loadings are
  # actually in, because it is detected from the fitted uniquenesses rather than assumed from the
  # estimator name -- the unrotated loading SEs are scaled in whichever gauge is used, so picking
  # the other one would report them in an orientation the solution is not in.
  L <- unclass(efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml")$unrot_loadings)
  fo <- list(unrot_loadings = L, orig_R = stats::cov2cor(em$sigma))
  core <- .se_fiml_core(fo, fiml, N = em$n, method = "ULS")

  ref_ltpil <- .ref_fiml_loading_se(L, Omega, "LtPiL", N = em$n, weight = "ULS")
  ref_ltl <- .ref_fiml_loading_se(L, Omega, "LtL", N = em$n, weight = "ULS")
  expect_equal(core$loadings_se, ref_ltpil, tolerance = 1e-5, ignore_attr = TRUE)
  # The two gauges really do differ here, so the check above is not satisfied by both.
  expect_gt(max(abs(ref_ltpil - ref_ltl)), 1e-3)
})

test_that("the FIML sandwich reports a degenerate rotational gauge", {
  skip_on_cran()

  X <- fiml_mar_data_2f()

  # A weakly determined orientation inflates the unrotated loading SEs without bound while leaving
  # every gauge-invariant quantity intact. This fixture's orientation is well determined, so the
  # transversal is forced below its floor: the mock is then the only cause, and what is under test
  # is that the two-stage path acts on the diagnostic at all, which it previously could not.
  expect_true(all(is.finite(
    efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml",
            se = "sandwich")$SE$unrot_loadings)))

  testthat::local_mocked_bindings(.gauge_transversal = function(...) 0)

  expect_warning(
    fit <- efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml", se = "sandwich"),
    class = "efa_se_unreliable"
  )

  # The unrotated loadings alone are withheld ...
  expect_true(all(is.na(fit$SE$unrot_loadings)))
  # ... while the gauge-invariant uniquenesses and communalities, and the covariance itself, stay:
  # the divergence lives purely in the gauge, so it must not take them down with it.
  expect_true(all(is.finite(fit$SE$uniquenesses)))
  expect_true(all(is.finite(fit$SE$communalities)))
  expect_true(all(is.finite(fit$vcov_unrot_loadings)))
})

test_that("the FIML sandwich withholds standard errors at a uniqueness pinned at the floor", {
  skip_on_cran()

  X <- fiml_mar_data()
  em <- .fiml_em_moments(X)
  fiml <- list(data = X, mu = em$mu, sigma = em$sigma)
  R <- stats::cov2cor(em$sigma)

  # The two-stage sandwich reports the same Wald quantities as the other analytic paths, from a
  # different covariance, so it withholds on the same boundary test: on the parameter-space
  # boundary the Wald approximation fails for every parameter regardless of how the covariance was
  # estimated. The boundary an ML/ULS fitter can actually reach is the uniqueness floor, not zero,
  # so this solution is strictly interior to the gauge detection's own `psi <= 0` test.
  L <- matrix(0, 6, 2)
  L[, 1] <- c(sqrt(1 - .uniqueness_floor), 0.6, 0.5, 0.1, 0.15, 0.2)
  L[, 2] <- c(0.00, 0.05, 0.20, 0.70, 0.60, 0.55)
  psi <- 1 - rowSums(L^2)
  expect_lte(min(psi), .uniqueness_floor + sqrt(.Machine$double.eps))
  expect_gt(min(psi), 0)

  fo <- list(unrot_loadings = L, orig_R = R)

  # Without the gate this covariance is perfectly usable, so what withholds the SEs below is the
  # boundary test and not a degenerate saturated covariance.
  core <- .se_fiml_core(fo, fiml, N = em$n, method = "ML")
  expect_true(core$reliable)
  expect_true(all(is.finite(core$loadings_se)))

  expect_warning(
    out <- .se_fiml(fo, rot_info = NULL, N = em$n, ci = 0.95, fiml = fiml, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(out$SE$unrot_loadings)))
  expect_true(all(is.na(out$SE$uniquenesses)))
  expect_true(all(is.na(out$SE$communalities)))
  expect_true(all(is.na(out$vcov_unrot_loadings)))
  expect_true(all(is.na(unlist(out$CI))))

  # Under a rotation the withholding reaches the rotated quantities too: they are propagated from
  # the same loading covariance, which the gate leaves unusable.
  rot_info <- list(rotation = "oblimin", rotmat = diag(2), rot_loadings = L,
                   Phi = diag(2), normalize = FALSE, crit_args = list(gam = 0, delta = 0.01))
  expect_warning(
    rot <- .se_fiml(fo, rot_info, N = em$n, ci = 0.95, fiml = fiml, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(rot$SE$rot_loadings)))
  expect_true(all(is.na(rot$SE$Phi)))
  expect_true(all(is.na(rot$SE$Structure)))

  # The corrected two-stage chi-square is a discrepancy-function quantity, not a Wald one, so the
  # boundary does not invalidate it: it is built outside the standard-error route and is kept.
  st <- .fiml_scaled_test(L, R, N = em$n, method = "ML", df = 4, m = 6, fiml = fiml)
  expect_false(is.null(st))
  expect_true(is.finite(st$chi))
})

test_that("the FIML fit indices honour ci = FALSE", {
  skip_on_cran()

  X <- fiml_mar_data()
  em <- .fiml_em_moments(X)
  fiml <- list(data = X, mu = em$mu, sigma = em$sigma, logl = em$logl)
  R <- stats::cov2cor(em$sigma)
  L <- unclass(efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml")$unrot_loadings)

  with_ci <- .gof(L, R, em$n, "ML", Fm = NA_real_, ci = TRUE, fiml = fiml)
  no_ci <- .gof(L, R, em$n, "ML", Fm = NA_real_, ci = FALSE, fiml = fiml)

  # The scaled block replaces the whole chi-square-derived block, so it has to honour the request
  # not to solve the RMSEA noncentrality bounds -- which is exactly the per-replicate bootstrap
  # path, where the bounds are dropped again by the aggregation.
  expect_true(any(is.finite(c(with_ci$RMSEA_LB, with_ci$RMSEA_UB))))
  expect_true(is.na(no_ci$RMSEA_LB))
  expect_true(is.na(no_ci$RMSEA_UB))

  # Nothing else changes with it: the flag governs the bounds alone.
  expect_identical(with_ci$chi_scaled_type, "scaled.shifted")
  expect_identical(no_ci$chi_scaled_type, "scaled.shifted")
  expect_equal(no_ci$chi, with_ci$chi)
  expect_equal(no_ci$RMSEA, with_ci$RMSEA)
  expect_equal(no_ci$CFI, with_ci$CFI)
})

test_that("the FIML CI-provenance note names the corrected two-stage sandwich", {
  skip_on_cran()
  local_reproducible_output()

  efa <- EFA(fiml_mar_data_1f(), n_factors = 1, method = "ML",
             cor_method = "fiml", se = "sandwich")
  body <- cli::ansi_strip(format(summary(efa)))
  note <- body[grepl("Note:", body)]
  # For a non-pooled analytic fit both "information" and "sandwich" yield the corrected two-stage
  # sandwich, so the note must describe that covariance, not the expected information matrix.
  expect_true(any(grepl("two-stage (FIML) sandwich", note, fixed = TRUE)))
  expect_false(any(grepl("expected information matrix", note)))
})

test_that("print/summary.efa label FIML correlations in the header", {
  skip_on_cran()
  local_reproducible_output()

  efa <- EFA(fiml_mar_data(), n_factors = 2, method = "ML", cor_method = "fiml")

  # The header flags the two-stage FIML correlation so the analysed matrix is not read as an
  # ordinary complete-case one. Assert the labels without pinning platform-dependent spacing.
  printed <- cli::ansi_strip(capture.output(print(efa)))
  expect_true(any(grepl(
    "Correlations: FIML (two-stage, missing data)", printed, fixed = TRUE
  )))
  expect_true(any(grepl("Unrotated Loadings", printed, fixed = TRUE)))

  # summary() renders the same header, so the label is surfaced there as well.
  expect_match(format(summary(efa)), "FIML (two-stage, missing data)",
               fixed = TRUE, all = FALSE)
})

test_that("use is ignored under cor_method = 'fiml' (classed warning)", {
  set.seed(1)
  n <- 200
  f <- stats::rnorm(n)
  load <- 0.7
  X <- vapply(seq_len(4),
              function(j) load * f + stats::rnorm(n, sd = sqrt(1 - load^2)),
              numeric(n))
  colnames(X) <- paste0("V", seq_len(4))

  expect_warning(
    EFA(X, n_factors = 1, method = "ML", cor_method = "fiml", use = "complete.obs"),
    class = "efa_fiml_use_ignored"
  )
})

test_that("FIML np-boot returns the full SE/CI schema with finite SEs (oblique)", {
  skip_on_cran()

  X <- fiml_mar_data()
  b <- 20L

  efa <- suppressWarnings(suppressMessages(
    EFA(X, n_factors = 2, method = "ML", rotation = "oblimin",
        cor_method = "fiml", se = "np-boot", b_boot = b, seed = 42)
  ))

  expect_s3_class(efa, "EFA")

  # The bootstrap SE/CI schema is the same one the pearson/poly paths produce: an oblique
  # rotation adds the factor-correlation (Phi) and structure-coefficient blocks. The
  # replicate cube's last dimension indexes the bootstrap samples.
  expect_named(efa$SE, c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                         "fit_indices", "residuals", "valid_replicates",
                         "valid_target_rotations"))
  expect_named(efa$CI, c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                         "fit_indices", "residuals"))
  expect_named(efa$replicates, c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                                 "fit_indices", "residuals"))
  expect_identical(dim(efa$replicates$unrot_loadings)[3], b)

  # SEs are finite and non-negative throughout, the loadings strictly positive (Phi's fixed
  # unit diagonal carries SE 0, so it is only checked >= 0); confidence intervals are ordered.
  for (nm in c("unrot_loadings", "rot_loadings", "Phi", "Structure")) {
    expect_true(all(is.finite(efa$SE[[nm]])), info = nm)
    expect_true(all(efa$SE[[nm]] >= 0), info = nm)
    expect_true(all(efa$CI[[nm]]$lower <= efa$CI[[nm]]$upper), info = nm)
  }
  expect_true(all(efa$SE$unrot_loadings > 0))
})

test_that("the FIML bootstrap carries resample indices, not a matrix per replicate", {
  skip_on_cran()

  X <- fiml_mar_data()
  b <- 8L

  # Intercept what efa_fit() hands the replicate fits, then run them for real: the per-replicate
  # list is a closure global of the parallel map, so whatever it carries is held once per worker
  # on top of the parent.
  real_boot <- .boot_fun
  captured <- NULL
  testthat::local_mocked_bindings(
    .boot_fun = function(..., fiml_list = NULL, fiml_data = NULL) {
      captured <<- list(fiml_list = fiml_list, fiml_data = fiml_data)
      real_boot(..., fiml_list = fiml_list, fiml_data = fiml_data)
    }
  )

  fit <- suppressWarnings(suppressMessages(
    efa_fit(X, n_factors = 2, estimator = "ML", cor_method = "fiml",
            se = "np-boot", b_boot = b, seed = 42)
  ))
  expect_s3_class(fit, "efa")

  reps <- Filter(Negate(is.null), captured$fiml_list)
  expect_length(reps, b)
  for (r in reps) {
    expect_named(r, c("rows", "mu", "sigma", "logl"))
    expect_null(r$data)
    expect_length(r$rows, fit$settings$N)
  }
  # The raw data travels once, alongside the indices that point into it.
  expect_equal(captured$fiml_data, X)

  # The indices identify exactly the resample the stored moments were estimated from, so the
  # replicate fit sees the same data the old per-replicate matrix carried.
  em_1 <- suppressWarnings(.fiml_em_moments(captured$fiml_data[reps[[1]]$rows, , drop = FALSE]))
  expect_equal(em_1$sigma, reps[[1]]$sigma)
  expect_equal(em_1$mu, reps[[1]]$mu)
  expect_equal(em_1$logl, reps[[1]]$logl)

  # ... and it is materially smaller than storing the slices themselves.
  with_data <- lapply(reps, function(r) {
    list(data = X[r$rows, , drop = FALSE], mu = r$mu, sigma = r$sigma, logl = r$logl)
  })
  expect_lt(as.numeric(utils::object.size(captured$fiml_list)),
            as.numeric(utils::object.size(with_data)) / 3)
})

test_that("the FIML bootstrap replicate fit is unchanged by the index representation", {
  skip_on_cran()

  X <- fiml_mar_data()
  rows <- which(rowSums(!is.na(X)) > 0L)
  set.seed(99)
  ind <- sample(rows, size = length(rows), replace = TRUE)
  em_i <- suppressWarnings(.fiml_em_moments(X[ind, , drop = FALSE]))
  R_i <- stats::cov2cor(em_i$sigma)

  # Drive one replicate through .boot_fun() with the index representation ...
  boot <- .boot_fun(array(R_i, c(ncol(X), ncol(X), 1L)), 1L, .estimate_model,
                    method = "ML", n_factors = 2, N = length(rows), type = "EFAtools",
                    start_method = "psych", lean = TRUE,
                    fiml_list = list(list(rows = ind, mu = em_i$mu, sigma = em_i$sigma,
                                          logl = em_i$logl)),
                    fiml_data = X)

  # ... and independently with the resampled matrix supplied directly, as the caller used to
  # store it. The fit -- loadings and the FIML likelihood-ratio fit indices alike -- must be
  # identical: the representation is a storage change, not a numerical one.
  direct <- suppressWarnings(.estimate_model(
    R_i, method = "ML", n_factors = 2, N = length(rows), type = "EFAtools",
    start_method = "psych", lean = TRUE,
    fiml = list(data = X[ind, , drop = FALSE], mu = em_i$mu, sigma = em_i$sigma,
                logl = em_i$logl)))

  expect_equal(boot[[1]]$unrot_loadings, direct$unrot_loadings)
  expect_equal(boot[[1]]$fit_indices, direct$fit_indices)
})

test_that("FIML np-boot is reproducible given a fixed seed", {
  skip_on_cran()

  X <- fiml_mar_data()
  run <- function() suppressWarnings(suppressMessages(
    EFA(X, n_factors = 2, method = "ML", rotation = "oblimin",
        cor_method = "fiml", se = "np-boot", b_boot = 8, seed = 7)
  ))

  # The `seed` argument reseeds inside EFA(), so two calls reproduce the case resampling
  # (and hence the EM recompute and the replicate fits) independently of the caller's RNG.
  a <- run()
  b <- run()

  expect_equal(a$SE$unrot_loadings, b$SE$unrot_loadings)
  expect_equal(a$SE$Phi, b$SE$Phi)
})

test_that("a degenerate FIML resample is dropped rather than aborting np-boot", {
  skip_on_cran()

  X <- fiml_mar_data()
  b <- 10L

  # Capture the real EM engine, then mock it: succeed on the point-estimate call (the first
  # one, inside .prepare_cor_input()) and force two later bootstrap resamples to abort,
  # standing in for a degenerate resample. The loop must store an all-NA matrix for those so
  # the replicate is dropped at the fit stage rather than aborting the whole bootstrap.
  real_em <- .fiml_em_moments
  n_call <- 0L
  testthat::local_mocked_bindings(
    .fiml_em_moments = function(...) {
      n_call <<- n_call + 1L
      if (n_call %in% c(3L, 6L)) {
        cli::cli_abort("forced degenerate resample", class = "efa_fiml_not_posdef")
      }
      real_em(...)
    }
  )

  efa <- suppressWarnings(suppressMessages(
    EFA(X, n_factors = 2, method = "ML", cor_method = "fiml",
        se = "np-boot", b_boot = b, seed = 1)
  ))

  expect_s3_class(efa, "EFA")
  expect_true(all(is.finite(efa$SE$unrot_loadings)))
  # the two dropped replicates leave NA slices; the surviving ones drive the SEs.
  expect_true(any(is.na(efa$replicates$unrot_loadings)))
})

test_that("FIML np-boot resamples the informative rows when a row is fully missing", {
  skip_on_cran()

  # A fully-missing row carries no information and is excluded from the EM case count
  # (N = em$n < nrow), so the bootstrap must resample the rows with at least one
  # observed value -- not all row positions -- to keep each replicate's sample size
  # equal to N. The np-boot path runs cleanly and N stays the informative-row count.
  X <- fiml_mar_data()
  Xd <- rbind(X, NA)

  efa <- suppressWarnings(suppressMessages(
    EFA(Xd, n_factors = 2, method = "ML", cor_method = "fiml",
        se = "np-boot", b_boot = 12, seed = 3)
  ))

  expect_s3_class(efa, "EFA")
  expect_equal(efa$settings$N, nrow(X))                 # the all-NA row is excluded
  expect_lt(efa$settings$N, nrow(Xd))
  expect_true(all(is.finite(efa$SE$unrot_loadings)))
})

# Asymptotic covariance of the Stage-1 saturated FIML estimates: the inverse observed
# information (negative Hessian) of theta = (mu, vech(sigma)) and its correlation-scale variant.
# The analytic score is cross-checked against a numerical gradient of the saturated
# log-likelihood, the theta-scale covariance against lavaan's saturated-model vcov, and the
# correlation-scale variant against a finite-difference standardisation and a Monte-Carlo
# sampling covariance.

# MAR fixture: column 1 is fully observed and drives the missingness in the others, so the
# mechanism depends only on observed data (mirrors the Stage-1 engine tests). The deviates are
# coloured by the Cholesky factor rather than by an eigendecomposition, because chol() is unique
# for a positive definite matrix while eigenvector signs are settled by rounding, so the same
# seed reproduces the same sample on every LAPACK build.
.acov_mar_data <- function(n = 500, seed = 321) {
  set.seed(seed)
  p <- 4
  Sig <- 0.5 ^ abs(outer(seq_len(p), seq_len(p), "-"))     # AR(1)-type positive-definite cov
  X <- matrix(stats::rnorm(n * p), n) %*% chol(Sig)
  colnames(X) <- paste0("V", seq_len(p))
  X[X[, 1] >  0.8, 2] <- NA
  X[X[, 1] < -0.8, 3] <- NA
  X[X[, 1] >  1.1, 4] <- NA
  X
}

test_that("the analytic score matches a numerical gradient of the log-likelihood", {
  skip_on_cran()

  X <- .acov_mar_data()
  em <- .fiml_em_moments(X)
  p <- ncol(X)
  pstar <- p * (p + 1L) / 2L

  loglik <- function(theta) {
    sig <- .unvech(theta[p + seq_len(pstar)], p)
    .fiml_loglik(X, theta[seq_len(p)], sig)
  }

  # Evaluate away from the EM fixed point (where the score is ~0), keeping sigma posdef.
  set.seed(9)
  theta <- c(em$mu, .vech(em$sigma)) + 0.02 * stats::rnorm(p + pstar)
  sig <- .unvech(theta[p + seq_len(pstar)], p)
  skip_if(any(eigen(sig, symmetric = TRUE, only.values = TRUE)$values < 1e-6))

  ana <- .fiml_saturated_score(X, theta[seq_len(p)], sig)

  h <- 1e-5
  num <- vapply(seq_along(theta), function(k) {
    tp <- theta; tp[k] <- tp[k] + h
    tm <- theta; tm[k] <- tm[k] - h
    (loglik(tp) - loglik(tm)) / (2 * h)
  }, numeric(1))

  expect_equal(ana, num, tolerance = 1e-5, ignore_attr = TRUE)
})

test_that("the score vanishes at the EM fixed point", {
  skip_on_cran()

  X <- .acov_mar_data()
  em <- .fiml_em_moments(X, tol = 1e-10, max_iter = 1000L)
  sc <- .fiml_saturated_score(X, em$mu, em$sigma)
  # The score is O(n); a near-zero value relative to n confirms the fixed point is the MLE.
  expect_lt(max(abs(sc)) / em$n, 1e-5)
})

test_that("the analytic information matches a numerical Jacobian of the score", {
  X <- .acov_mar_data()
  em <- .fiml_em_moments(X)
  p <- ncol(X)
  pstar <- p * (p + 1L) / 2L
  patterns <- .fiml_patterns(!is.na(X))

  ana <- .fiml_saturated_information(X, em$mu, em$sigma, patterns)

  # Central differences of the analytic score, at the eps^(1/3) optimum step scaled by each
  # parameter's own magnitude (a mean is on the variable's scale, sigma_ij on sd_i sd_j).
  theta0 <- c(em$mu, .vech(em$sigma))
  sds <- sqrt(diag(em$sigma))
  ij <- which(lower.tri(diag(p), diag = TRUE), arr.ind = TRUE)
  h <- .Machine$double.eps^(1 / 3) *
    pmax(abs(theta0), c(sds, sds[ij[, 1L]] * sds[ij[, 2L]]))
  score_at <- function(theta) {
    .fiml_saturated_score(X, theta[seq_len(p)], .unvech(theta[p + seq_len(pstar)], p),
                          patterns = patterns)
  }
  num <- vapply(seq_along(theta0), function(k) {
    tp <- theta0; tp[k] <- tp[k] + h[k]
    tm <- theta0; tm[k] <- tm[k] - h[k]
    -(score_at(tp) - score_at(tm)) / (2 * h[k])
  }, numeric(length(theta0)))

  expect_equal(ana, (num + t(num)) / 2, tolerance = 1e-6, ignore_attr = TRUE)
  expect_true(isSymmetric(ana, tol = 0))
})

test_that("the analytic information equals the complete-data closed form", {
  # With no missing values the observed information of the saturated multivariate normal has
  # the textbook block-diagonal form n Sigma^-1 for the means and n/2 D'(Sigma^-1 kron
  # Sigma^-1) D for vech(Sigma), with D the duplication matrix (Magnus & Neudecker, 2019).
  set.seed(11)
  p <- 5L
  n <- 200L
  Sig <- 0.5 ^ abs(outer(seq_len(p), seq_len(p), "-"))
  X <- matrix(stats::rnorm(n * p), n) %*% chol(Sig)
  colnames(X) <- paste0("V", seq_len(p))
  mu_hat <- colMeans(X)
  S_hat <- crossprod(sweep(X, 2L, mu_hat, "-")) / n

  ana <- .fiml_saturated_information(X, mu_hat, S_hat, .fiml_patterns(!is.na(X)))

  pstar <- p * (p + 1L) / 2L
  ij <- which(lower.tri(diag(p), diag = TRUE), arr.ind = TRUE)
  D <- matrix(0, p * p, pstar)
  D[cbind((ij[, 2L] - 1L) * p + ij[, 1L], seq_len(pstar))] <- 1
  D[cbind((ij[, 1L] - 1L) * p + ij[, 2L], seq_len(pstar))] <- 1
  S_inv <- solve(S_hat)
  closed <- matrix(0, p + pstar, p + pstar)
  closed[seq_len(p), seq_len(p)] <- n * S_inv
  closed[p + seq_len(pstar), p + seq_len(pstar)] <-
    (n / 2) * crossprod(D, kronecker(S_inv, S_inv) %*% D)

  expect_equal(ana, closed, tolerance = 1e-10, ignore_attr = TRUE)
})

test_that("the information and asymptotic covariances are well-formed", {
  skip_on_cran()

  X <- .acov_mar_data()
  em <- .fiml_em_moments(X)
  p <- ncol(X)
  pstar <- p * (p + 1L) / 2L
  npair <- p * (p - 1L) / 2L

  ac <- .fiml_saturated_acov(X, em$mu, em$sigma)

  # theta-scale: square (p + p*), symmetric, positive definite, labelled.
  expect_equal(dim(ac$theta), c(p + pstar, p + pstar))
  expect_equal(rownames(ac$theta)[seq_len(p)], colnames(X))
  expect_true(isSymmetric(unname(ac$theta), tol = 1e-8))
  expect_gt(min(eigen(ac$theta, symmetric = TRUE, only.values = TRUE)$values), 0)

  # cor-scale: off-diagonal pairs, combn order/labels (same layout as .adf_gamma / .pair_labels).
  expect_equal(dim(ac$cor), c(npair, npair))
  expect_equal(rownames(ac$cor), .pair_labels(colnames(X)))
  expect_true(isSymmetric(unname(ac$cor), tol = 1e-8))
  expect_gt(min(eigen(ac$cor, symmetric = TRUE, only.values = TRUE)$values), -1e-8)
  # each correlation carries strictly positive sampling variance (guards against a vacuously
  # PSD all-zero block from a mis-indexed Jacobian or sigma-block extraction).
  expect_true(all(diag(ac$cor) > 0))
})

test_that("a non-positive-definite covariance aborts with a classed error", {
  # A deterministic indefinite covariance (one negative eigenvalue); both the score and the ACOV
  # must refuse with the shared classed condition rather than leaking a bare chol() error.
  X <- matrix(c(1, 2, 3, 2, 1, 4, 3, 5, 2, 4, 1, 6), ncol = 3)
  colnames(X) <- paste0("V", seq_len(3))
  mu <- colMeans(X)
  bad <- matrix(c(1, 0.99, 0.99, 0.99, 1, -0.99, 0.99, -0.99, 1), 3)
  expect_lt(min(eigen(bad, symmetric = TRUE, only.values = TRUE)$values), 0)

  expect_error(.fiml_saturated_score(X, mu, bad), class = "efa_fiml_not_posdef")
  expect_error(.fiml_saturated_acov(X, mu, bad), class = "efa_fiml_not_posdef")
})

test_that("the theta-scale asymptotic covariance matches lavaan's saturated vcov", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  X <- .acov_mar_data()
  df <- as.data.frame(X)
  p <- ncol(X)
  vn <- colnames(X)
  em <- .fiml_em_moments(X)
  ours <- .fiml_saturated_acov(X, em$mu, em$sigma)$theta

  # Explicit saturated model: free every variance/covariance and mean, fit by FIML with the
  # observed information so the parameter vcov is the inverse observed information we compute.
  ij <- which(lower.tri(diag(p), diag = TRUE), arr.ind = TRUE)
  cov_lines <- sprintf("%s ~~ %s", vn[ij[, 2L]], vn[ij[, 1L]])
  mean_lines <- sprintf("%s ~ 1", vn)
  mod <- paste(c(cov_lines, mean_lines), collapse = "\n")
  fit <- suppressWarnings(
    lavaan::lavaan(mod, data = df, missing = "ml",
                   information = "observed", se = "standard"))
  V <- lavaan::vcov(fit)

  # Map lavaan's parameters to our theta order by canonicalised labels (sorted endpoints for a
  # covariance; the orientation lavaan stores a "~~" pair in is not guaranteed).
  canon <- function(lab) {
    lab <- gsub("\\s", "", lab)
    if (grepl("~~", lab, fixed = TRUE)) {
      vv <- sort(strsplit(lab, "~~", fixed = TRUE)[[1]])
      paste0("cov:", vv[1L], ":", vv[2L])
    } else {
      paste0("mean:", sub("~1$", "", lab))
    }
  }
  lav_keys <- vapply(rownames(V), canon, character(1))
  cov_keys <- vapply(seq_len(nrow(ij)), function(t) {
    vv <- sort(c(vn[ij[t, 1L]], vn[ij[t, 2L]]))
    paste0("cov:", vv[1L], ":", vv[2L])
  }, character(1))
  our_keys <- c(paste0("mean:", vn), cov_keys)

  ord <- match(our_keys, lav_keys)
  expect_false(anyNA(ord))
  lav <- V[ord, ord, drop = FALSE]

  # Split tolerance as the polychoric ACOV tests do: variances relatively, the tiny off-diagonal
  # covariances absolutely (scaled to the variance magnitude). The residual is our EM vs
  # lavaan's optimiser plus the finite-difference Jacobian; measured agreement is far tighter.
  expect_lt(max(abs(diag(ours) - diag(lav)) / diag(lav)), 0.05)
  off <- upper.tri(ours)
  expect_lt(max(abs(ours[off] - lav[off])), 0.05 * max(diag(lav)))
})

test_that("the correlation-scale ACOV equals a finite-difference standardisation", {
  skip_on_cran()

  X <- .acov_mar_data()
  em <- .fiml_em_moments(X)
  p <- ncol(X)
  pstar <- p * (p + 1L) / 2L
  ac <- .fiml_saturated_acov(X, em$mu, em$sigma)

  # Off-diagonal correlations of cov2cor(sigma), in combn order, as a function of vech(sigma).
  pairs <- utils::combn(p, 2L)
  cor_off <- function(vs) {
    R <- stats::cov2cor(.unvech(vs, p))
    R[cbind(pairs[1L, ], pairs[2L, ])]
  }
  vs0 <- .vech(em$sigma)
  h <- 1e-6
  Jn <- vapply(seq_len(pstar), function(k) {
    vp <- vs0; vp[k] <- vp[k] + h
    vm <- vs0; vm[k] <- vm[k] - h
    (cor_off(vp) - cor_off(vm)) / (2 * h)
  }, numeric(ncol(pairs)))

  sig_idx <- p + seq_len(pstar)
  Omega_sig <- ac$theta[sig_idx, sig_idx]
  Omega_R_fd <- Jn %*% Omega_sig %*% t(Jn)

  expect_equal(unname(ac$cor), unname(Omega_R_fd), tolerance = 1e-5)
})

test_that("the correlation-scale ACOV tracks the Monte-Carlo sampling covariance", {
  skip_on_cran()
  skip_if_not_slow()

  # The analytic asymptotic variance of each off-diagonal FIML correlation must track the
  # empirical sampling variance over many fresh MAR samples from the same population. The
  # analytic variance is averaged over the samples to remove single-sample noise (it is itself
  # estimated from data), so the check isolates the method's calibration; the bounds then
  # cover the Monte-Carlo error in the empirical variances.
  p <- 4
  pairs <- utils::combn(p, 2L)
  R <- 1000L
  rmat <- matrix(NA_real_, R, ncol(pairs))
  amat <- matrix(NA_real_, R, ncol(pairs))
  for (r in seq_len(R)) {
    Xr <- .acov_mar_data(seed = 1000L + r)
    emr <- .fiml_em_moments(Xr)
    rmat[r, ] <- stats::cov2cor(emr$sigma)[cbind(pairs[1L, ], pairs[2L, ])]
    amat[r, ] <- diag(.fiml_saturated_acov(Xr, emr$mu, emr$sigma)$cor)
  }
  empirical <- apply(rmat, 2L, stats::var)
  analytic <- colMeans(amat)

  # Each empirical variance is itself a variance over R replicates, so it carries a
  # Monte-Carlo standard deviation of sqrt(2 / (R - 1)) = 4.5%. Calibration is therefore
  # asserted on the median ratio, where that noise largely averages out, and the worst of the
  # six ratios only has to stay inside a band several Monte-Carlo standard deviations wide --
  # a systematic mis-scaling of the Jacobian or of the saturated block moves every ratio well
  # beyond both.
  ratio <- analytic / empirical
  expect_lt(abs(stats::median(ratio) - 1), 0.10)
  expect_lt(max(abs(ratio - 1)), 0.25)
})

#' McDonald's omega
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `OMEGA()` has been superseded by [efa_reliability()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' This function finds omega total, hierarchical, and subscale, as well as additional
#' model-based indices of interpretive relevance (H index, ECV, PUC)
#' from a Schmid-Leiman (SL) solution or lavaan single factor, second-order (see below),
#' or bifactor solution. The SL-based omegas can either be found from a
#' [psych::schmid()], [efa_schmid_leiman()], or,
#' in a more flexible way, by leaving
#' `model = NULL` and specifying additional arguments. The `type` argument selects
#' how variables are assigned to group factors, and can reproduce the assignment
#' [psych::omega()] makes.
#'
#' @param model class [efa_schmid_leiman()], class `schmid`, or class
#' `lavaan` object. That is, an output object from [efa_schmid_leiman()] or
#' [psych::schmid()], or a `lavaan` fit object with a
#' single factor, second-order, or bifactor solution. If of class `lavaan`,
#' only `g_name` needs to be specified additionally. If of class
#' [efa_schmid_leiman()] or `schmid`, only the arguments `factor_corres`
#' and `cormat` need to be specified additionally.
#' @param type character. Either `"EFAtools"` (default) or `"psych"`
#' (see details)
#' @param g_name character. The name of the general factor from the lavaan solution.
#' This needs only be specified if `model` is a `lavaan` second-order
#' or bifactor solution. Default is "g".
#' @param group_names character. An optional vector of group names. The length
#' must correspond to the number of groups for which the `lavaan` model
#' was fitted.
#' @param add_ind logical. Whether additional indices (H index, ECV, PUC) should
#' be calculated or not (see details for these indices). If FALSE, only omegas
#' are returned. Default is `TRUE`.
#' @param factor_corres matrix. A logical matrix or a numeric matrix containing
#' 0's and 1's that indicates which variable corresponds to which group factor.
#' Must have the same dimensions as the matrix of group factor loadings from the
#' SL solution. Cross-loadings are allowed here. See examples for use.
#' @param var_names character. A vector with subtest names in the order
#' of the rows from the SL solution. This needs only be specified if `model`
#' is left `NULL`.
#' @param fac_names character. An optional vector of group factor names in the
#' order of the columns of the SL solution. If left `NULL`, names of the
#' group factors from the entered solution are taken.
#' @param g_load numeric. A vector of general factor loadings from an SL solution.
#' This needs only be specified if `model` is left `NULL`.
#' @param s_load matrix. A matrix of group factor loadings from an SL solution.
#' This needs only be specified if `model` is left `NULL`.
#' @param u2 numeric. A vector of uniquenesses from an SL solution. This needs
#' only be specified if `model` is left `NULL`.
#' @param cormat matrix. A correlation matrix to be used when
#' `variance = "correlation"`. If left `NULL` and an [efa_schmid_leiman()]
#' output is entered in `model`, the correlation matrix is taken from the
#' output. If left `NULL` and a [psych::schmid()]
#' output is entered, the correlation matrix will be found based on the pattern
#' matrix and Phi from the [psych::schmid()] output
#' using [psych::factor.model()].
#' If left `NULL` and model is also left `NULL`, the correlation matrix
#' is found based on the pattern matrix and Phi entered. However, if the
#' correlation matrix is available, `cormat` should be specified instead
#' of `Phi` and `pattern`.
#' @param pattern matrix. Pattern coefficients from an oblique factor solution.
#' This needs only be specified if `model` is left `NULL`,
#' `variance = "correlation"` and `cormat` is also left `NULL`.
#' @param Phi matrix. Factor intercorrelations from an oblique factor solution.
#' This needs only be specified if `model` is left `NULL`,
#' `variance = "correlation"` and `cormat` is also left `NULL`.
#' @param variance character. If `"correlation"` (default), then total
#' variances for the whole scale as well as for the subscale composites are
#' calculated based on the correlation
#' matrix. If `"sums_load"`, then total variances are calculated using the
#' squared sums of general factor loadings and group factor loadings and
#' the sum of uniquenesses (see details).
#'
#' @details ## What this function does
#'
#' All types of McDonald's omegas (total, hierarchical, and subscale; McDonald,
#' 1978, 1985, 1999) are calculated for
#' the general factor as well as for the subscales / group factors (see, e.g.,
#' Gignac, 2014; Rodriguez et al., 2016a, 2016b). Omegas refer to the correlation
#' between a factor and a unit-weighted composite score and thus the
#' true score variance in a unit-weighted composite based on the respective
#' indicators. Omega total is the total true score variance in a composite.
#' Omega hierarchical is the true score variance in a composite that is attributable
#' to the general factor, and omega subscale is the true score variance in a
#' composite attributable to all subscales / group factors (for the whole scale)
#' or to the specific subscale / group factor (for subscale composites).
#'
#' Accordingly, on a subscale row the `hier` column reports the share of that
#' subscale's composite variance due to the general factor and the `sub` column
#' the share due to the subscale-specific factor; the latter corresponds to the
#' omega hierarchical subscale of Rodriguez et al. (2016a, 2016b).
#'
#' The H index (also construct reliability or replicability index) is the
#' correlation between an optimally-weighted composite score
#' and a factor (Hancock & Mueller, 2001; Rodriguez et al., 2016a, 2016b). It, too,
#' can be calculated for the whole scale / general factor as well as for the
#' subscales / group factors. Low values indicate that a latent variable is not well
#' defined by its indicators.
#'
#' The ECV (Rodriguez et al., 2016a, 2016b) is the ratio of the
#' variance explained by the general factor and the variance explained by the
#' general factor and the group factors.
#'
#' The PUC (Bonifay et al., 2015; Reise et al., 2013, Rodriguez et al., 2016a,
#' 2016b) refers to the proportion
#' of correlations in the underlying correlation matrix that is not contaminated
#' by variance of both the general factor and the group factors (i.e., correlations
#' between indicators from different group factors, which reflect only general
#' factor variance). The higher the PUC, the more similar a general factor from
#' a multidimensional model will be to the single factor from a unidimensional
#' model.
#'
#' ## How to use this function
#'
#' If `model` is a `lavaan` second-order or bifactor solution,
#' only the name of the general factor from the lavaan model needs to be specified
#' additionally with the `g_name` argument. It is then determined whether this
#' general factor is a second-order factor (second-order model with one second-order
#' factor assumed) or a breadth factor (bifactor model assumed). Please note that
#' this function only works for second-order models if they contain no more than
#' one second-order factor. In case of a second-order solution, a
#' Schmid-Leiman transformation is performed on the first- and second-order loadings
#' and omega coefficients are obtained from the transformed (orthogonalized) solution
#' (see [efa_schmid_leiman()] for more information on Schmid-Leiman transformation).
#' There is also the possibility to enter a `lavaan` single factor solution.
#' In this case, `g_name` is not needed. Finally, if a solution from a
#' `lavaan` multiple group analysis is entered, the indices are computed for
#' each group. For `lavaan` input the composite variances entering the omegas are
#' model-implied: they are computed from the fitted loadings and the fitted residual
#' covariance matrix, and count any freed residual covariance as well as the residual
#' variances. The coefficients thus coincide with the observed-score versions when the
#' model fits perfectly. The omegas split a composite's variance into a general part and one part
#' per group factor, which needs uncorrelated latent variables: fit a bifactor model
#' with `orthogonal = TRUE` (not `lavaan`'s default) and leave the covariances between
#' a second-order model's first-order factors at zero. A fit whose factors correlate is
#' rejected rather than scored as though they did not.
#' The type argument is not evaluated if `model` is of class
#' `lavaan`.
#'
#' If `model` is of class [efa_schmid_leiman()] or
#' [psych::schmid()] only the
#' `type` and, depending on the type (see below), the `factor_corres`
#' arguments need to be specified additionally. If model is of class
#' [psych::schmid()] and `variance = "correlation"`
#' (default), it is
#' recommended to also provide the original correlation matrix in `cormat`
#' to get more accurate results. Otherwise, the correlation matrix will be found
#' based on the pattern matrix and Phi from the
#' [psych::schmid()] output
#' using the [psych::factor.model()] function.
#'
#' If `model = NULL`, the arguments `type`, `factor_corres`
#' (depending on the type, see below), `var_names`, `g_load`, `s_load`,
#' and `u2` and either `cormat` (recommended) or `Phi` and
#' `pattern` need to be specified. If `Phi` and `pattern` are
#' specified instead of `cormat`, the correlation matrix is found using
#' the [psych::factor.model()] function.
#'
#' The only difference between `type = "EFAtools"` and `type = "psych"`
#' is the determination of variable-to-factor correspondences. `type = "psych"`
#' derives them as [psych::omega()] does, by
#' taking the highest
#' group factor loading for each variable as the relevant group factor loading.
#' To do this, `factor_corres` must be left `NULL`.
#'
#' Both settings score a composite by the true score variance the model attributes
#' to it, counting every factor its variables load on; they differ only in the
#' variance that is divided into. `variance = "correlation"` uses the composite's
#' observed variance, giving the observed-score form of omega; `"sums_load"` uses its
#' model-implied variance, which partitions exactly into omega hierarchical plus omega
#' subscale on the whole-scale row. The two settings agree up to model misfit, and
#' differ mainly in the whole-scale omega subscale, which counts all group-factor
#' variance under `"sums_load"` but only the assigned subscale composites under
#' `"correlation"`.
#'
#'
#' @return If found for an SL or `lavaan` second-order or bifactor solution
#' without multiple groups:
#' A matrix with omegas for the whole scale and for the subscales and (only if
#' `add_ind = TRUE`) with the H index, ECV, and PUC.
#' \item{tot}{Omega total.}
#' \item{hier}{Omega hierarchical.}
#' \item{sub}{Omega subscale.}
#' \item{H}{H index.}
#' \item{ECV}{Explained common variance.}
#' \item{PUC}{Percent of uncontaminated correlations.}
#'
#' If found for a `lavaan` single factor solution without multiple groups:
#' A (named) vector with omega total and (if `add_ind = TRUE`) the H index
#' for the single factor.
#'
#' If found for a `lavaan` output from a multiple group analysis: A list
#' containing the output described above for each group.
#'
#' @source McDonald, R. P. (1978). Generalizability in factorable domains: ‘‘Domain
#' validity and generalizability’’. Educational and Psychological Measurement,
#' 38, 75–79.
#' @source McDonald, R. P. (1985). Factor analysis and related methods. Hillsdale,
#' NJ: Erlbaum.
#' @source McDonald, R. P. (1999). Test theory: A unified treatment. Mahwah,
#' NJ: Erlbaum.
#' @source Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016a). Applying bifactor
#' statistical indices in the evaluation of psychological measures. Journal of
#' Personality Assessment, 98, 223-237.
#' @source Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016b). Evaluating
#' bifactor models: Calculating and interpreting statistical indices.
#' Psychological Methods, 21, 137-150.
#' @source Hancock, G. R., & Mueller, R. O. (2001). Rethinking construct reliability
#' within latent variable systems. In R. Cudeck, S. du Toit, & D. Sörbom (Eds.),
#' Structural equation modeling: Present and future—A Festschrift in honor of Karl
#' Jöreskog (pp. 195–216). Lincolnwood, IL: Scientific Software International.
#' @source Reise, S. P., Scheines, R., Widaman, K. F., & Haviland, M. G. (2013).
#' Multidimensionality and structural coefficient bias in structural equation
#' modeling: A bifactor perspective. Educational and Psychological Measurement,
#' 73, 5–26.
#' @source Bonifay, W. E., Reise, S. P., Scheines, R., & Meijer, R. R. (2015).
#' When are multidimensional data unidimensional enough for structural equation
#' modeling?: An evaluation of the DETECT multidimensionality index. Structural
#' Equation Modeling, 22, 504—516.
#' @source Gignac, G. E. (2014). On the Inappropriateness of Using Items to
#' Calculate Total Scale Score Reliability via Coefficient Alpha for Multidimensional
#' Scales. European Journal of Psychological Assessment, 30, 130-139.
#'
#' @seealso [efa_reliability()] for the same coefficients in a tidy, long-format
#'   result.
#'
#' @export
#'
#' @examples
#' \donttest{
#' ## Use with lavaan outputs
#' if (requireNamespace("lavaan", quietly = TRUE)) {
#'
#' # Create and fit bifactor model in lavaan (assume all variables have SDs of 1)
#' mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
#'         F2 =~ V7 + V8 + V9 + V10 + V11 + V12
#'         F3 =~ V13 + V14 + V15 + V16 + V17 + V18
#'         g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
#'              V13 + V14 + V15 + V16 + V17 + V18'
#' fit_bi <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
#'                       sample.nobs = 500, estimator = "ml", orthogonal = TRUE)
#'
#' # Compute omegas and additional indices for bifactor solution
#' OMEGA(fit_bi, g_name = "g")
#'
#' # Compute only omegas
#' OMEGA(fit_bi, g_name = "g", add_ind = FALSE)
#'
#' # Create and fit second-order model in lavaan (assume all variables have SDs of 1)
#' mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
#'         F2 =~ V7 + V8 + V9 + V10 + V11 + V12
#'         F3 =~ V13 + V14 + V15 + V16 + V17 + V18
#'         g =~ F1 + F2 + F3'
#' fit_ho <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
#'                       sample.nobs = 500, estimator = "ml")
#'
#' # Compute omegas and additional indices for second-order solution
#' OMEGA(fit_ho, g_name = "g")
#' }
#' }
#'
#' ## Use with an output from the SL function, with type EFAtools
#' efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' sl_mod <- efa_schmid_leiman(efa_mod, estimator = "PAF")
#'
#' # Indicator-to-factor correspondences from a salience threshold (here: .20):
#' factor_corres_1 <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2
#'
#' OMEGA(sl_mod, type = "EFAtools", factor_corres = factor_corres_1)
#'
#' ## Use with an output from the psych::schmid function, with type psych for
#' ## OMEGA
#' schmid_mod <- psych::schmid(test_models$baseline$cormat, nfactors = 3,
#'                             n.obs = 500, fm = "pa", rotate = "Promax")
#' # Find correlation matrix from phi and pattern matrix from psych::schmid output
#' OMEGA(schmid_mod, type = "psych")
#' # Use specified correlation matrix
#' OMEGA(schmid_mod, type = "psych", cormat = test_models$baseline$cormat)
#'
#' ## Manually specify components (useful if omegas should be computed for a SL
#' ## or bifactor solution found with another program)
#' ## As an example, we extract the elements from an SL output here. This gives
#' ## the same results as in the second example above.
#'
#' factor_corres <- matrix(c(rep(0, 12), rep(1, 6), rep(0, 6), rep(1, 6),
#'                         rep(0, 6), rep(1, 6), rep(0, 12)), ncol = 3,
#'                         byrow = FALSE)
#'
#' OMEGA(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
#'       g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
#'       u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
#'       factor_corres = factor_corres)
#'
OMEGA <- function(model = NULL, type = c("EFAtools", "psych"), g_name = "g",
                  group_names = NULL, add_ind = TRUE, factor_corres = NULL,
                  var_names = NULL, fac_names = NULL,
                  g_load = NULL, s_load = NULL, u2 = NULL, cormat = NULL,
                  pattern = NULL, Phi = NULL, variance = c("correlation",
                                                           "sums_load")){

  # Perform argument checks
  type <- match.arg(type)
  checkmate::assert_string(g_name)
  checkmate::assert_logical(add_ind)
  checkmate::assert_character(group_names, null.ok = TRUE)
  # Check for factor_corres in OMEGA_helper
  checkmate::assert_character(var_names, null.ok = TRUE)
  checkmate::assert_character(fac_names, null.ok = TRUE)
  checkmate::assert_numeric(g_load, null.ok = TRUE)
  if(!is.null(s_load) && !inherits(s_load, c("matrix", "SLLOADINGS"))){

    cli::cli_abort(
      c("Invalid {.arg s_load}.",
        "i" = "Leave it {.code NULL} for a model input, or supply a Schmid-Leiman loading matrix of class {.cls matrix} or {.cls SLLOADINGS}."),
      class = "efa_omega_bad_s_load"
    )

  }
  checkmate::assert_numeric(u2, null.ok = TRUE)
  checkmate::assert_matrix(cormat, null.ok = TRUE)
  if(!is.null(pattern) && !inherits(pattern, c("matrix", "loadings", "LOADINGS"))){

    cli::cli_abort(
      c("Invalid {.arg pattern}.",
        "i" = "Leave it {.code NULL}, or supply a pattern matrix from an oblique solution of class {.cls matrix}, {.cls loadings}, or {.cls LOADINGS}."),
      class = "efa_omega_bad_pattern"
    )

  }
  checkmate::assert_matrix(Phi, null.ok = TRUE)
  variance <- match.arg(variance)

  # Determine which function to use
  if(!is.null(model) & (!is.null(var_names) || !is.null(g_load) || !is.null(s_load)
     || !is.null(u2))){

    cli::cli_warn(
      c("You entered a {.arg model} and also specified {.arg var_names}, {.arg g_load}, {.arg s_load}, or {.arg u2}; these are ignored.",
        "i" = "To use specific values, leave {.code model = NULL} and specify all arguments separately."),
      class = "efa_omega_model_args_ignored"
    )

  }

  if(inherits(model, "lavaan")){

    .require_lavaan()

    .OMEGA_LAVAAN(model = model, g_name = g_name, group_names = group_names,
                  add_ind = add_ind)

  } else {

    # Validate the non-lavaan inputs, then hand both the schmid/SL and the
    # validated manual (model = NULL) cases to the single flexible-omega worker.
    if(!inherits(model, c("schmid", "SL"))){

      if(!is.null(model)){

        cli::cli_abort(
          c("Invalid {.arg model}.",
            "i" = "Enter a lavaan, {.fn psych::schmid}, or SL object, or specify {.arg var_names}, {.arg g_load}, and {.arg s_load}."),
          class = "efa_omega_invalid_model"
        )

      } else if(is.null(var_names) || is.null(g_load) || is.null(s_load) || is.null(u2)){

        cli::cli_abort("Specify all of {.arg var_names}, {.arg g_load}, {.arg s_load}, and {.arg u2}.",
                       class = "efa_omega_missing_args")

      }

    }

    .OMEGA_FLEX(model = model, type = type, factor_corres = factor_corres,
                var_names = var_names, fac_names = fac_names, g_load = g_load,
                s_load = s_load, u2 = u2, cormat = cormat, pattern = pattern,
                Phi = Phi, variance = variance, add_ind = add_ind)

  }
}

#' Estimate factor scores for an EFA model
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `FACTOR_SCORES()` has been superseded by [efa_scores()], which is the
#' recommended interface going forward. It remains available so existing code
#' keeps working. Note that `R2` is now the squared factor-score determinacy of
#' the *requested* `method`: the squared correlation between a factor and the
#' scores that method produces. For `method = "Thurstone"` this is each factor's
#' squared multiple correlation with the observed variables (the value
#' [psych::factor.scores()] returns with `Grice = TRUE`); for every other method
#' it is smaller, because no estimator correlates more highly with the factor
#' than the regression estimator does. Earlier versions returned psych's default
#' `Grice = FALSE` validity coefficient, so the slot is not comparable across
#' versions.
#'
#' A convenience wrapper around [efa_scores()] that returns factor scores and
#' weights in a compact list. Factor scores are calculated according to the
#' specified method if raw data are provided, and only factor weights if a
#' correlation matrix is provided.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data (needed to get
#' factor scores) or matrix with correlations.
#' @param f object of class [efa_fit()] or matrix.
#' @param Phi matrix. A matrix of factor intercorrelations. Only needs to be
#' specified if a factor loadings matrix is entered directly into `f`; for an
#' [efa_fit()] object the intercorrelations are taken from the object, and a
#' supplied `Phi` is ignored with a warning. Default is `NULL`, in which case the
#' intercorrelations of a directly supplied loading matrix are assumed to be zero.
#' @param rho matrix. Correlation matrix used to derive the scoring weights.
#' Defaults to `NULL`, in which case the matrix the EFA in `f` was fit on
#' (`f$orig_R`) is used, so the weights stay consistent with the loadings even for
#' a non-Pearson correlation (e.g. polychoric); for a directly supplied loading
#' matrix, `x` itself is used when it is a correlation matrix, otherwise the
#' Pearson correlation of `x`. Pass a matrix here to score against a different
#' correlation.
#' @param method character. The method used to calculate factor scores. One of
#' "Thurstone" (regression-based; default), "tenBerge", "Anderson", "Bartlett",
#' "Harman", or "components".
#'
#' @return A list of class FACTOR_SCORES containing the following:
#'
#' \item{scores}{The factor scores (only if raw data are provided.)}
#' \item{weights}{The factor weights.}
#' \item{r.scores}{The correlations of the factor score estimates.}
#' \item{missing}{Whether the raw data contained missing values (only if raw data
#' are provided).}
#' \item{R2}{The squared factor-score determinacy for each factor: the squared
#' correlation between a factor and the score the requested `method` produces.
#' For `method = "Thurstone"` this equals the squared multiple correlation
#' between the factor and the observed variables; for every other method it is
#' specific to those scores and smaller. See [efa_scores()] for the underlying
#' score-quality diagnostics.}
#' \item{settings}{A list of the settings used.}
#'
#' @seealso [efa_scores()] for the factor-score weights together with the full
#' set of score-quality diagnostics (determinacy, univocality, and Guttman
#' indeterminacy index) and a print/summary method.
#'
#' @export
#'
#' @examples
#' # Example with raw data with method "Bartlett"
#' EFA_raw <- efa_fit(DOSPERT_raw, n_factors = 10, estimator = "PAF",
#'                    rotation = "oblimin",
#'                    rotate_control = rotate_control(random_starts = 1))
#' fac_scores_raw <- FACTOR_SCORES(DOSPERT_raw, f = EFA_raw, method = "Bartlett")
#'
#' # Same as above, but with raw data AND a correlation matrix
#' cor_pearson <- cor(DOSPERT_raw)
#' EFA_cor_pearson <- efa_fit(cor_pearson, n_factors = 10, N = nrow(DOSPERT_raw),
#'                            estimator = "PAF", rotation = "oblimin",
#'                            rotate_control = rotate_control(random_starts = 1))
#' fac_scores_cor_pearson <- FACTOR_SCORES(DOSPERT_raw, f = EFA_cor_pearson,
#'                                         rho = cor_pearson,
#'                                         method = "Bartlett")
#'
#' # Scores between two alternatives above are identical
#' isTRUE(all.equal(fac_scores_raw$scores, fac_scores_cor_pearson$scores,
#'                  check.attributes = FALSE))
#'
#' # Example with a correlation matrix only (does not return factor scores)
#' EFA_cor <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                    estimator = "PAF", rotation = "oblimin")
#' fac_scores_cor <- FACTOR_SCORES(test_models$baseline$cormat, f = EFA_cor)
#'
FACTOR_SCORES <- function(x, f, Phi = NULL, rho = NULL,
                          method = c("Thurstone", "tenBerge", "Anderson",
                                     "Bartlett", "Harman", "components")){

  method <- match.arg(method)

  # "Thurstone" is the regression method under efa_scores' naming; every other
  # label is shared. Delegate the whole computation (and the classed conditions).
  es <- efa_scores(x, f = f, Phi = Phi, rho = rho,
                   method = if (method == "Thurstone") "regression" else method)

  # Reassemble the legacy FACTOR_SCORES output shape.
  output <- list(
    scores = es$scores,
    weights = es$weights,
    r.scores = es$r.scores
  )

  # Only raw data (with returned scores) carries the `missing` field.
  if (!is.null(es$scores)) {
    output$missing <- anyNA(x)
  }

  # Squared factor-score determinacy, named by factor.
  output$R2 <- stats::setNames(es$determinacy$rho2, rownames(es$determinacy))

  # Preserve the original (user-facing) method label in the settings.
  output$settings <- list(method = method)

  class(output) <- "FACTOR_SCORES"

  output

}

#' Simulate data from a common-factor population model
#'
#' Draws data from a population correlation matrix, given either directly or
#' built from a factor model. The population correlation is either supplied in
#' `R`, or assembled from a loading matrix `Lambda`, the factor intercorrelations
#' `Phi`, and the unique variances `Psi` as
#' \eqn{R = Lambda\, Phi\, Lambda' + Psi}, standardized to a correlation matrix.
#' `marginals` chooses the marginal distribution of the drawn cases, `categories`
#' discretizes them into ordered categories, `missing` imposes a missing-data mechanism,
#' and a misfit target perturbs the population with model error; see *Details*.
#'
#' @details
#' Provide the population either as a ready correlation matrix in `R`, or through
#' the model components `Lambda`, `Phi`, and `Psi`; the two ways are mutually
#' exclusive. When the model components are used, `Phi` defaults to the identity
#' matrix (orthogonal factors) and `Psi` defaults to the unique variances that
#' make the population a correlation matrix (\eqn{1 - \mathrm{diag}(Lambda\, Phi\,
#' Lambda')}); the assembled covariance is standardized with
#' [stats::cov2cor()] so a non-standardized `Psi` still yields a correlation
#' matrix. With the default `Psi`, a factor model whose implied communalities
#' exceed 1 (a Heywood case) leaves no unique variance and is rejected; a `Psi`
#' you supply is instead only required to give positive variances and a
#' positive-semidefinite (a mathematically valid, internally consistent
#' correlation/covariance structure) population. Cases with normal marginals (the
#' default) are drawn through a matrix square root of the population correlation --
#' a Cholesky factor, or a symmetric eigen square root when it is singular (e.g. a
#' communality of exactly 1).
#'
#' With `marginals = "empirical"`, the iterative rank-matching algorithm of Ruscio
#' and Kaczetow (2008) reproduces the population correlation while each variable
#' takes the empirical marginal distribution of the matching column of
#' `marginal_data` (resampled with replacement). Only the marginals of
#' `marginal_data` are used; its own correlations are ignored, and the drawn columns
#' follow the population's variables, not those of `marginal_data`.
#'
#' With `marginals = "VM"` (Vale-Maurelli, 1983) or `"IG"` (the independent-generator
#' method; Foldnes & Olsson, 2016), the cases reproduce the population correlation
#' while carrying non-normal marginals with the target `skewness` and (excess)
#' `kurtosis`. The Vale-Maurelli family does not span every valid non-normal
#' distribution (Foldnes & Grønneberg, 2015); `"IG"` covers distributions `"VM"`
#' cannot. Not every (`skewness`, `kurtosis`) pair is attainable -- every distribution
#' needs excess kurtosis of at least `skewness^2 - 2`, and either method covers a smaller
#' region still -- so an unreachable request is rejected, as is a `"VM"` intermediate
#' correlation matrix that is not positive definite unless `force_pd` allows it.
#'
#' With `categories`, the drawn data are discretized into ordered categories (an integer
#' code `1` to `K`) at the thresholds that reproduce the requested category proportions
#' (Olsson, 1979): the standard-normal quantiles for `marginals = "normal"`, and for
#' `marginals = "VM"` those quantiles mapped through the same Fleishman (1978) cubic the
#' draw uses, so the requested proportions are reproduced on the non-normal scale too. Under
#' `marginals = "IG"` the thresholds stay on the standard-normal scale while the data do
#' not, so the achieved proportions depart from the request systematically rather than by
#' sampling noise, and only the *number* of categories is guaranteed; the same holds for a
#' `"VM"` variable whose Fleishman cubic is not increasing over its own thresholds and the
#' tails beyond them, which keeps the normal-scale thresholds and is reported with a
#' warning. This is more likely with strong skewness/kurtosis or very unequal category
#' proportions; the warning names the affected variable.
#' Because categorization attenuates product-moment correlations, the categorized
#' data's Pearson correlation is smaller in magnitude than the population correlation;
#' under non-normal marginals its polychoric correlation departs from the population as
#' well. Ordinal output is not available with `marginals = "empirical"`. Empty categories
#' left by a draw are reported with a warning, as they destabilize the polychoric
#' correlation and the factor analysis.
#'
#' With `missing`, missing values are introduced into the drawn data under a chosen
#' mechanism (Rubin, 1976), each variable holed at a target expected rate
#' `missing_prop`. `"MCAR"` draws an independent mask, so missingness is unrelated to
#' the data. `"MAR"` and `"MNAR"` set each case's missing probability by a logistic
#' model of a standardized predictor: another variable for `"MAR"` (chosen by
#' `missing_predictor`) or the variable's own value for `"MNAR"`, with slope
#' `missing_strength`. The mechanism acts on the drawn (latent) values, so when
#' `categories` also discretizes the data the missingness is keyed on the underlying
#' value, not the category code. For `"MAR"` the predictor is evaluated on the complete
#' drawn values, so whether the mechanism is *ignorably* MAR depends on which variables
#' carry missing values. By default every variable is holed, which leaves a variable's MAR
#' predictor itself missing for roughly a `missing_prop` fraction of the cases whose
#' missingness it drove. This breaks ignorability: the mechanism is then MAR conditional
#' on the *complete* data but **not** ignorable for an analyst who sees only the observed
#' data. As a result, estimators that are consistent under ignorable MAR -- `cor_method =
#' "fiml"` in [efa_fit()], or the multiple imputation behind [efa_mi()] -- keep a residual
#' bias that grows with `missing_prop` and `missing_strength`. Restricting the holed
#' variables with `missing_vars` and pointing
#' `missing_predictor` at variables outside that set makes every predictor fully observed,
#' which is ignorably MAR and recovers the unbiasedness those estimators are advertised
#' with. The returned matrix carries the `NA`s, which the correlation estimators handle
#' downstream.
#'
#' With `model_error`, the population is perturbed away from the exact factor
#' structure so the `q`-factor model (`q = ncol(Lambda)`) fits it only approximately,
#' at a prescribed misfit; exact factor structures are unrealistic (see
#' `target_rmsea`). The perturbation is applied once to
#' the population, and the achieved misfit of the specified generating model is computed
#' with the same fit-index formulas [efa_fit()] uses and returned in the `model_error`
#' element. It needs a factor-model population with residual degrees of freedom and an
#' exact factor structure (a diagonal `Psi`), and is orthogonal to the marginal, ordinal,
#' and missing-data options. Three methods are available. `"CB"` (Cudeck & Browne,
#' 1992) matches the target RMSEA to numerical precision and keeps the `q`-factor
#' model the exact minimizer (the CFI follows as a derived quantity). `"TKL"` (Tucker,
#' Koopman & Linn, 1969) adds minor common factors tuned so the achieved RMSEA -- and,
#' optionally, CFI -- match the target(s); with a single target the match is close, with
#' both it is a compromise, reported with a warning when the two cannot be reconciled.
#' `"WB"` (Wu & Browne, 2015) draws the population from
#' an inverse-Wishart distribution around the model-implied correlation; its calibration
#' applies to the *best-fitting* model, so the reported misfit of the *generating* model
#' is systematically larger than the target -- about 1.4 times for a typical 12-variable,
#' 3-factor model. Use `"CB"` when the reported RMSEA must equal the target.
#'
#' Replicated draws (`n_datasets > 1`) are generated in parallel across
#' replicates with \pkg{future.apply}; a parallel plan can be selected with
#' [future::plan()] (the default plan runs sequentially). Each replicate is
#' assigned its own reproducible random-number stream, so with a fixed `seed` the
#' output is identical regardless of the number of workers.
#'
#' @param N numeric. Number of cases (rows) to draw per dataset. Required unless
#'   `return_pop = TRUE`.
#' @param Lambda matrix. A `p` by `m` matrix of factor loadings. Supply this
#'   (optionally with `Phi` and `Psi`) instead of `R` to build the population
#'   from a factor model.
#' @param Phi matrix. The `m` by `m` factor intercorrelation matrix. Only used
#'   with `Lambda`. Default is `NULL`, in which case the factors are orthogonal
#'   (an identity matrix).
#' @param Psi numeric vector or matrix. The unique variances: either a length-`p`
#'   vector or a `p` by `p` matrix (added as the residual covariance). Only used
#'   with `Lambda`. Default is `NULL`, in which case the unique variances that
#'   standardize the population to a correlation matrix are used.
#' @param R matrix. A `p` by `p` population correlation matrix to draw from
#'   directly. Supply this instead of `Lambda`/`Phi`/`Psi`. It must have a unit
#'   diagonal; standardize a covariance matrix with [stats::cov2cor()] first.
#' @param model_error character. The method used to perturb the population so the factor model
#'   fits it imperfectly ("model error"): one of `"CB"` (Cudeck-Browne, the default), `"TKL"`
#'   (Tucker-Koopman-Linn), `"WB"` (Wu-Browne), or `"none"`. Model error is only applied when a
#'   target is supplied in `target_rmsea` or `target_cfi`; without one the population is exact,
#'   whatever `model_error`. Only used with a factor-model population (`Lambda`). Note that
#'   [efa_power()] defaults to `"TKL"` instead, because its minor common factors degrade
#'   factor recovery as well as the fit, so passing the same target to the two functions
#'   perturbs the population differently unless `model_error` is set explicitly.
#' @param target_rmsea numeric. The population RMSEA the factor model should have relative to
#'   the perturbed population, a single number strictly in `(0, 1)`. Supplying it activates model
#'   error. Simulating from an exact population overstates recovery, so a realistic value (around
#'   `0.05`) is recommended for simulation studies (MacCallum, 2003). Default is `NULL` (an
#'   exact population; do not pass `0`). Required for `"CB"` and `"WB"`; optional for `"TKL"`.
#' @param target_cfi numeric. Only used with `model_error = "TKL"`: the population CFI to
#'   target, a single number strictly in `(0, 1)`, on its own or together with `target_rmsea`
#'   (TKL then trades the two off). Default is `NULL` (do not pass `1`). `"CB"` and `"WB"` target
#'   the RMSEA only.
#' @param marginals character. The marginal distribution of the drawn data: one
#'   of `"normal"` (the default), which draws normal marginals; `"empirical"`,
#'   which reproduces the population correlation while preserving the empirical
#'   marginals supplied in `marginal_data`; or `"VM"` (Vale-Maurelli) and `"IG"`
#'   (independent generator), which draw non-normal marginals with the target
#'   `skewness` and `kurtosis`.
#' @param marginal_data matrix or data frame. Only used with
#'   `marginals = "empirical"`, where it is required: a data set with one numeric
#'   column per variable (`p` columns), each with at least two distinct values,
#'   whose per-column distributions are resampled as the marginals of the drawn
#'   data. Its correlations are ignored. Default is `NULL`.
#' @param n_factors numeric. Only used with `marginals = "empirical"`: the number
#'   of factors the rank-matching reproduction fits. Default is `NULL`, in which
#'   case it is the number of columns of `Lambda` when the population is built
#'   from a factor model; it must be given when the population is supplied via `R`.
#' @param skewness numeric. Only used with `marginals = "VM"` or `"IG"`: the
#'   target marginal skewness, as a single value applied to every variable or a
#'   length-`p` vector. Default is `NULL` (0, a symmetric marginal). At least one
#'   of `skewness` or `kurtosis` must be given for these marginals.
#' @param kurtosis numeric. Only used with `marginals = "VM"` or `"IG"`: the
#'   target marginal excess kurtosis (0 for a normal marginal), as a single value
#'   applied to every variable or a length-`p` vector. Default is `NULL` (0).
#' @param force_pd logical. Used with `marginals = "VM"` and with Cudeck-Browne model error
#'   (`model_error = "CB"`), where `FALSE` (the default) rejects a population or intermediate
#'   matrix that is not positive definite. What `TRUE` accepts instead differs by path, in
#'   both cases with a warning: for `marginals = "VM"` the intermediate correlation matrix is
#'   projected to the nearest correlation matrix (via [psych::cor.smooth()]), and the returned
#'   `population` becomes the one the projected draw attains; for `model_error = "CB"` there is
#'   no such projection -- the closest positive-definite perturbation is kept instead, whose
#'   achieved RMSEA is **below** the target. Has no effect for the `"TKL"` or `"WB"` methods.
#' @param categories numeric or list. Requests ordinal output by discretizing each
#'   variable into ordered categories. Either a count of equally probable categories
#'   (a single value applied to every variable or a length-`p` vector), or a
#'   length-`p` list of numeric vectors giving the marginal category proportions per
#'   variable (each strictly positive and summing to 1). Default is `NULL`, which
#'   returns the continuous data.
#' @param match character. Only used with `categories`: an assertion about how the
#'   categorization relates to the population correlation. With a normal latent, cutting
#'   at the normal-scale thresholds already leaves the population polychoric correlation
#'   of the categorized data equal to the target correlation, so both values compute the
#'   same thresholds and produce identical data whenever both are legal. `"thresholds"`
#'   (the default) also cuts the `"VM"` and `"IG"` draws, whose ordinal Pearson and
#'   polychoric correlations then both depart from the population; `"polychoric"` states
#'   that the polychoric match is required and therefore rejects non-normal marginals.
#'   Not available with `marginals = "empirical"`. Default is `NULL`
#'   (`"thresholds"` when `categories` is set).
#' @param missing character. An optional missing-data mechanism to impose on the drawn
#'   data: one of `"none"` (the default, complete data), `"MCAR"` (missing completely at
#'   random), `"MAR"` (missing at random, depending on another variable), or `"MNAR"`
#'   (missing not at random, depending on the variable's own value). Introduced values
#'   become `NA`. By default every variable is holed, so under `"MAR"` each variable's
#'   predictor is itself subject to missingness and the mechanism is MAR given the
#'   *complete* data rather than ignorably MAR; `missing_vars` and `missing_predictor`
#'   together give an ignorable design (see Details).
#' @param missing_prop numeric. Only used when `missing` is not `"none"`, where it is
#'   required: the target marginal proportion of missing values per variable, a single
#'   number strictly between 0 and 1. This is the expected rate; the realized rate of a
#'   given draw varies around it.
#' @param missing_strength numeric. Only used with `missing = "MAR"` or `"MNAR"`: the
#'   slope of the logistic missingness model, setting how strongly the missing
#'   probability depends on the predictor. Default is `NULL` (1, a moderate dependence);
#'   `0` removes the dependence (equivalent to MCAR at the same rate) and large
#'   magnitudes make missingness nearly deterministic.
#' @param missing_predictor integer or character. Only used with `missing = "MAR"`:
#'   which variable drives each holed variable's missingness, as one column index or
#'   variable name per variable in `missing_vars` (so one per variable by default), in
#'   that order. A variable cannot be its own predictor; if the same predictor is reused
#'   for several `missing_vars`, it must not itself be one of them. Default is
#'   `NULL`, in which case each variable's missingness is driven by the next variable
#'   cyclically (so variable order matters; supply this explicitly when the order is
#'   arbitrary). Predictors outside `missing_vars` are fully observed (see Details for
#'   why this matters).
#' @param missing_vars integer or character. Only used when `missing` is not `"none"`:
#'   which variables carry missing values, as column indices or variable names. Default is
#'   `NULL`, in which case every variable is holed. Restricting it, together with
#'   pointing `missing_predictor` at variables outside that set, is how an ignorably
#'   MAR design is obtained (see Details).
#' @param n_datasets numeric. The number of datasets to draw. Default is 1. With
#'   more than one, a list of datasets is returned.
#' @param seed numeric. Optional seed for reproducible draws. When supplied, the
#'   caller's random-number stream is saved and restored, so the call leaves the
#'   global RNG state unchanged. Default is `NULL` (no seeding).
#' @param return_pop logical. If `TRUE`, return only the population correlation
#'   matrix and draw no data. Default is `FALSE`. Because no data are drawn, the
#'   Vale-Maurelli intermediate correlation matrix is never formed, so its
#'   positive-definiteness (and any `force_pd` projection of it) is neither checked nor
#'   reflected in the returned population.
#'
#' @returns An object of class `efa_simulated`, a list containing:
#' \item{data}{The simulated data: an `N` by `p` numeric matrix, an integer matrix of
#'   category codes when `categories` is set, or a length-`n_datasets` list of these when
#'   `n_datasets > 1`. `NULL` when `return_pop = TRUE`.}
#' \item{population}{The `p` by `p` population correlation matrix drawn from,
#'   model-error-perturbed when requested. When `force_pd = TRUE` causes `marginals =
#'   "VM"`'s intermediate matrix to be projected, `population` is the population the
#'   projected draw actually reaches, differing from the target by the drift the warning
#'   reports.}
#' \item{model_error}{`NULL`, or, when model error was applied, a list of the method, the
#'   target and achieved RMSEA/CFI, the model degrees of freedom `df`, and the method's
#'   tuning parameter: `kappa` for `"CB"`, `v` and `eps` for `"TKL"`, `m` for `"WB"`.}
#' \item{settings}{The call's key arguments (`N`, `n_datasets`, `marginals`, `categories`,
#'   `match`, `missing`, `seed`), kept for reference and used when printing.}
#'
#' Printing the object shows a compact summary.
#'
#' @references Cudeck, R., & Browne, M. W. (1992). Constructing a covariance matrix
#'   that yields a specified minimizer and a specified minimum discrepancy function
#'   value. Psychometrika, 57(3), 357-369. \doi{10.1007/BF02295424}
#'
#'   Fleishman, A. I. (1978). A method for simulating non-normal
#'   distributions. Psychometrika, 43(4), 521-532. \doi{10.1007/BF02293811}
#'
#'   Foldnes, N., & Grønneberg, S. (2015). How general is the Vale-Maurelli
#'   simulation approach? Psychometrika, 80(4), 1066-1083.
#'   \doi{10.1007/s11336-014-9414-0}
#'
#'   Foldnes, N., & Olsson, U. H. (2016). A simple simulation technique for
#'   nonnormal data with prespecified skewness, kurtosis, and covariance matrix.
#'   Multivariate Behavioral Research, 51(2-3), 207-219.
#'   \doi{10.1080/00273171.2015.1133274}
#'
#'   MacCallum, R. C. (2003). 2001 Presidential Address: Working with imperfect
#'   models. Multivariate Behavioral Research, 38(1), 113-139.
#'   \doi{10.1207/S15327906MBR3801_5}
#'
#'   Olsson, U. (1979). Maximum likelihood estimation of the polychoric correlation
#'   coefficient. Psychometrika, 44(4), 443-460. \doi{10.1007/BF02296207}
#'
#'   Rubin, D. B. (1976). Inference and missing data. Biometrika, 63(3), 581-592.
#'   \doi{10.1093/biomet/63.3.581}
#'
#'   Ruscio, J., & Kaczetow, W. (2008). Simulating multivariate nonnormal data
#'   using an iterative algorithm. Multivariate Behavioral Research, 43(3),
#'   355-381. \doi{10.1080/00273170802285693}
#'
#'   Tucker, L. R., Koopman, R. F., & Linn, R. L. (1969). Evaluation of factor
#'   analytic research procedures by means of simulated correlation matrices.
#'   Psychometrika, 34(4), 421-459. \doi{10.1007/BF02290601}
#'
#'   Vale, C. D., & Maurelli, V. A. (1983). Simulating multivariate nonnormal
#'   distributions. Psychometrika, 48(3), 465-471. \doi{10.1007/BF02293687}
#'
#'   Wu, H., & Browne, M. W. (2015). Quantifying adventitious error in a covariance
#'   structure as a random effect. Psychometrika, 80(3), 571-600.
#'   \doi{10.1007/s11336-015-9451-3}
#'
#' @seealso [efa_power()], whose simulation mode draws its replicate data sets with this
#'   function, and [efa_fit()] for analysing the simulated data.
#'
#' @family data simulation
#'
#' @export
#'
#' @examples
#' # Build a population from a shipped loading pattern and factor correlations
#' Lambda <- population_models$loadings$baseline
#' Phi <- population_models$phis_3$moderate
#'
#' # Draw one normal dataset of 500 cases (the data live in $data)
#' sim <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, seed = 42)
#' dim(sim$data)
#'
#' # Return only the population correlation matrix
#' R_pop <- efa_simulate(Lambda = Lambda, Phi = Phi, return_pop = TRUE)$population
#'
#' # Draw several datasets at once from a supplied correlation matrix
#' sims <- efa_simulate(N = 500, R = R_pop, n_datasets = 3, seed = 42)
#' length(sims$data)
#'
#' # Reproduce the population correlation but with skewed, empirical marginals
#' # (here from a chi-squared source with one column per variable)
#' src <- matrix(rchisq(200 * nrow(Lambda), df = 3), ncol = nrow(Lambda))
#' dat_emp <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi,
#'                         marginals = "empirical", marginal_data = src, seed = 42)
#'
#' # Draw skewed, leptokurtic data with the Vale-Maurelli method
#' dat_vm <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, marginals = "VM",
#'                        skewness = 1.5, kurtosis = 4, seed = 42)
#'
#' # Draw five-category ordinal data whose polychoric correlation matches R
#' dat_ord <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi,
#'                         categories = 5, match = "polychoric", seed = 42)
#'
#' # Draw data with 15% missing at random, driven by a neighbouring item
#' dat_mar <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, missing = "MAR",
#'                         missing_prop = 0.15, seed = 42)
#' colMeans(is.na(dat_mar$data))
#'
#' # An ignorably MAR design: only the first nine items are holed, each driven by one of
#' # the last nine, which stay complete
#' dat_ign <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, missing = "MAR",
#'                         missing_prop = 0.15, missing_vars = 1:9,
#'                         missing_predictor = 10:18, seed = 42)
#' colMeans(is.na(dat_ign$data))
#'
#' # Add realistic model error: a population the model fits with RMSEA of about .05
#' # (Cudeck-Browne, the default method; the achieved fit is reported)
#' sim_me <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi,
#'                        target_rmsea = 0.05, seed = 42)
#' sim_me$model_error$rmsea
efa_simulate <- function(N = NULL, Lambda = NULL, Phi = NULL, Psi = NULL,
                         R = NULL,
                         model_error = c("CB", "TKL", "WB", "none"),
                         target_rmsea = NULL, target_cfi = NULL,
                         marginals = c("normal", "empirical", "VM", "IG"),
                         marginal_data = NULL, n_factors = NULL,
                         skewness = NULL, kurtosis = NULL, force_pd = FALSE,
                         categories = NULL, match = NULL,
                         missing = c("none", "MCAR", "MAR", "MNAR"),
                         missing_prop = NULL, missing_strength = NULL,
                         missing_predictor = NULL, missing_vars = NULL,
                         n_datasets = 1L, seed = NULL, return_pop = FALSE) {

  # Eigenvalues below -tol mark the population as indefinite (not a valid
  # covariance); kept equal to the compiled core's default so both agree.
  tol <- 1e-8

  checkmate::assert_flag(return_pop)
  checkmate::assert_flag(force_pd)
  checkmate::assert_count(n_datasets, positive = TRUE)
  checkmate::assert_int(seed, null.ok = TRUE)

  marginals <- .match_arg_ci(marginals)

  # `model_error` selects the method used to perturb the population so a factor model fits it
  # imperfectly; `target_rmsea`/`target_cfi` set the amount and activate it. A target is what
  # turns model error on -- without one the population is exact, whatever `model_error`. CB and
  # WB target the RMSEA only; TKL can target the RMSEA and/or the CFI. Compatibility is checked
  # here; the perturbation itself runs once the population is built (below).
  model_error <- .match_arg_ci(model_error)
  # The targets are strictly inside (0, 1): a target RMSEA of 0 (or CFI of 1) means no misfit,
  # which is the exact population -- omit the target instead of asking a model-error method to
  # inject zero error (which would only degenerate the solvers).
  if (!is.null(target_rmsea) &&
      (!checkmate::test_number(target_rmsea) || target_rmsea <= 0 || target_rmsea >= 1)) {
    cli::cli_abort(
      c("{.arg target_rmsea} must be a single number in (0, 1).",
        "i" = "Omit {.arg target_rmsea} for an exact (zero-misfit) population."),
      class = "efa_simulate_input")
  }
  if (!is.null(target_cfi) &&
      (!checkmate::test_number(target_cfi) || target_cfi <= 0 || target_cfi >= 1)) {
    cli::cli_abort(
      c("{.arg target_cfi} must be a single number in (0, 1).",
        "i" = "Omit {.arg target_cfi} for an exact (perfect-fit) population."),
      class = "efa_simulate_input")
  }
  have_target <- !is.null(target_rmsea) || !is.null(target_cfi)
  do_model_error <- model_error != "none" && have_target
  if (model_error == "none" && have_target) {
    cli::cli_abort(
      c("{.arg target_rmsea} and {.arg target_cfi} require a model-error method.",
        "x" = "You supplied a target but left {.code model_error = \"none\"}.",
        "i" = "Choose {.code model_error = \"CB\"}, {.val TKL}, or {.val WB}."),
      class = "efa_simulate_input"
    )
  }
  # CB and WB target the RMSEA only.
  if (do_model_error && model_error %in% c("CB", "WB")) {
    if (is.null(target_rmsea)) {
      cli::cli_abort(
        c("{.code model_error = \"{model_error}\"} requires {.arg target_rmsea}.",
          "i" = "{.val CB} and {.val WB} target the RMSEA."),
        class = "efa_simulate_input"
      )
    }
    if (!is.null(target_cfi)) {
      cli::cli_abort(
        c("{.code model_error = \"{model_error}\"} targets the RMSEA only.",
          "x" = "You also supplied {.arg target_cfi}.",
          "i" = "Use {.code model_error = \"TKL\"} to target the CFI (with or without the RMSEA)."),
        class = "efa_simulate_input"
      )
    }
  }

  # marginal_data and n_factors configure the empirical rank-matching path only;
  # skewness and kurtosis configure the non-normal VM/IG paths, and force_pd the
  # Vale-Maurelli path only.
  if (marginals != "empirical" && (!is.null(marginal_data) || !is.null(n_factors))) {
    supplied <- c("marginal_data", "n_factors")[c(!is.null(marginal_data),
                                                  !is.null(n_factors))]
    cli::cli_abort(
      c("{.arg marginal_data} and {.arg n_factors} only apply when {.code marginals = \"empirical\"}.",
        "x" = "You set {.code marginals = \"{marginals}\"} but supplied {.arg {supplied}}."),
      class = "efa_simulate_input"
    )
  }
  if (!(marginals %in% c("VM", "IG")) &&
      (!is.null(skewness) || !is.null(kurtosis))) {
    supplied <- c("skewness", "kurtosis")[c(!is.null(skewness), !is.null(kurtosis))]
    cli::cli_abort(
      c("{.arg skewness} and {.arg kurtosis} only apply when {.code marginals} is {.val VM} or {.val IG}.",
        "x" = "You set {.code marginals = \"{marginals}\"} but supplied {.arg {supplied}}."),
      class = "efa_simulate_input"
    )
  }
  # force_pd rescues the Vale-Maurelli intermediate correlation matrix and, for Cudeck-Browne
  # model error, a population that cannot reach the target RMSEA while staying positive definite;
  # it has no role otherwise (the IG path needs a positive definite population outright).
  if (marginals != "VM" && !do_model_error && isTRUE(force_pd)) {
    cli::cli_abort(
      c("{.arg force_pd} only applies when {.code marginals = \"VM\"} or model error is requested.",
        "x" = "You set {.code force_pd = TRUE} but {.code marginals = \"{marginals}\"} with no model-error target."),
      class = "efa_simulate_input"
    )
  }

  # `categories` requests ordered-category (ordinal) output; `match` selects how
  # that categorization relates to the population correlation and is meaningless
  # without it.
  if (is.null(categories)) {
    if (!is.null(match)) {
      cli::cli_abort(
        c("{.arg match} only applies when {.arg categories} requests ordinal output.",
          "x" = "You set {.arg match} but left {.arg categories} unset, so continuous data are returned."),
        class = "efa_simulate_input"
      )
    }
  } else {
    match <- .match_arg_ci(match, c("thresholds", "polychoric"))
    # The polychoric correlation is defined by a bivariate-normal latent, so it can
    # only equal the target correlation when the latent is normal. Pairing it with
    # non-normal marginals is a contradiction rather than an approximation.
    if (match == "polychoric" && marginals != "normal") {
      cli::cli_abort(
        c("{.code match = \"polychoric\"} requires normal latents, but {.code marginals = \"{marginals}\"} draws non-normal marginals.",
          "x" = "The polychoric correlation assumes a bivariate-normal latent, so it cannot equal the target correlation under non-normal marginals.",
          "i" = "Use {.code match = \"thresholds\"} to cut the non-normal data (its ordinal Pearson correlation is then attenuated), or keep {.code marginals = \"normal\"}."),
        class = "efa_simulate_match_conflict"
      )
    }
    # Thresholds sit on the standard-normal scale, which categorizes the standardized
    # draws (normal, and the unit-variance VM/IG marginals) as intended. The empirical
    # draw instead carries the arbitrary location and scale of `marginal_data`, so
    # those thresholds would not reproduce the requested categories.
    if (marginals == "empirical") {
      cli::cli_abort(
        c("{.arg categories} is not supported with {.code marginals = \"empirical\"}.",
          "x" = "Ordinal thresholds sit on the standard-normal scale, but the empirical draw keeps the arbitrary scale of {.arg marginal_data}, so the requested categories are not reproduced.",
          "i" = "Use {.code marginals = \"normal\"}, {.val VM}, or {.val IG} to draw ordinal data with {.arg categories}."),
        class = "efa_simulate_input"
      )
    }
  }

  # `missing` selects an optional missing-data mechanism; `missing_prop` sets the
  # target marginal missing rate, `missing_strength` the strength of the value
  # dependence in MAR/MNAR, `missing_predictor` the MAR predictor, and `missing_vars`
  # which variables are holed at all. Each companion is meaningful only under the
  # mechanisms that use it (compatibility is checked here; value validation runs later,
  # once the data dimension is known).
  missing <- .match_arg_ci(missing)
  if (missing == "none") {
    supplied <- c("missing_prop", "missing_strength", "missing_predictor",
                  "missing_vars")[
      c(!is.null(missing_prop), !is.null(missing_strength),
        !is.null(missing_predictor), !is.null(missing_vars))]
    if (length(supplied)) {
      cli::cli_abort(
        c("Missing-data settings only apply when {.arg missing} requests a mechanism.",
          "x" = "You supplied {.arg {supplied}} but left {.code missing = \"none\"}."),
        class = "efa_simulate_input"
      )
    }
  } else {
    if (is.null(missing_prop)) {
      cli::cli_abort(
        c("{.arg missing_prop} is required when {.arg missing} is not {.val none}.",
          "i" = "Set the target marginal proportion of missing values, e.g. {.code missing_prop = 0.1}."),
        class = "efa_simulate_input"
      )
    }
    # MCAR removes values independently of the data, so it has no strength or predictor;
    # MNAR's predictor is fixed to each variable's own value, so none may be supplied.
    if (missing == "MCAR" &&
        (!is.null(missing_strength) || !is.null(missing_predictor))) {
      supplied <- c("missing_strength", "missing_predictor")[
        c(!is.null(missing_strength), !is.null(missing_predictor))]
      cli::cli_abort(
        c("{.code missing = \"MCAR\"} removes values completely at random.",
          "x" = "You supplied {.arg {supplied}}, but MCAR has no predictor to depend on."),
        class = "efa_simulate_input"
      )
    }
    if (missing == "MNAR" && !is.null(missing_predictor)) {
      cli::cli_abort(
        c("{.arg missing_predictor} does not apply when {.code missing = \"MNAR\"}.",
          "x" = "MNAR makes each variable's missingness depend on its own value, not another predictor."),
        class = "efa_simulate_input"
      )
    }
  }

  # The population is given either as a ready correlation matrix or as model
  # components, never both and never neither.
  have_lambda <- !is.null(Lambda)
  have_R <- !is.null(R)
  if (have_lambda == have_R) {
    cli::cli_abort(
      c("Specify the population in exactly one way.",
        "x" = "Provide either {.arg R}, or {.arg Lambda} (with optional {.arg Phi}/{.arg Psi})."),
      class = "efa_simulate_input"
    )
  }
  if (have_R && (!is.null(Phi) || !is.null(Psi))) {
    cli::cli_abort(
      c("{.arg Phi} and {.arg Psi} only apply when the population is built from {.arg Lambda}.",
        "x" = "You supplied {.arg R} together with {.arg Phi}/{.arg Psi}."),
      class = "efa_simulate_input"
    )
  }

  if (have_lambda) {

    Lambda <- as.matrix(Lambda)
    checkmate::assert_matrix(Lambda, mode = "numeric", any.missing = FALSE,
                             min.rows = 1, min.cols = 1)
    # assert_matrix's any.missing rejects NA/NaN but not Inf; a non-finite loading
    # would propagate NaN into the communalities and turn the Heywood guard below
    # into an uninformative `if (NA)` error, so reject it up front.
    if (!all(is.finite(Lambda))) {
      cli::cli_abort("{.arg Lambda} must contain only finite values.",
                     class = "efa_simulate_input")
    }
    p <- nrow(Lambda)
    m <- ncol(Lambda)

    if (is.null(Phi)) {
      Phi <- diag(m)
    } else {
      Phi <- as.matrix(Phi)
      checkmate::assert_matrix(Phi, mode = "numeric", nrows = m, ncols = m,
                               any.missing = FALSE)
      if (!isSymmetric(unname(Phi))) {
        cli::cli_abort(c("{.arg Phi} must be a symmetric {m} by {m} matrix."),
                       class = "efa_simulate_input")
      }
    }

    common <- Lambda %*% Phi %*% t(Lambda)

    if (is.null(Psi)) {
      # Unique variances that standardize the population to a correlation matrix.
      # An implied communality above 1 (a Heywood case) leaves no unique variance
      # to simulate from; a communality of exactly 1 gives a drawable, if singular,
      # population and is allowed.
      psi_vec <- 1 - diag(common)
      if (any(psi_vec < 0)) {
        cli::cli_abort(
          c("The factor model implies a communality above 1 (a Heywood case).",
            "x" = "There is no unique variance left to simulate from.",
            "i" = "Provide unique variances in {.arg Psi}, or reduce the loadings."),
          class = "efa_simulate_heywood"
        )
      }
      Psi_mat <- diag(psi_vec, p)
    } else if (is.matrix(Psi)) {
      checkmate::assert_matrix(Psi, mode = "numeric", nrows = p, ncols = p,
                               any.missing = FALSE)
      # A residual covariance must be finite and symmetric; a non-finite or
      # non-symmetric matrix would otherwise be silently averaged into shape by the
      # symmetrization below (as Phi and R are also checked to be symmetric).
      if (!all(is.finite(Psi)) || !isSymmetric(unname(Psi))) {
        cli::cli_abort("{.arg Psi} must be a finite, symmetric {p} by {p} matrix.",
                       class = "efa_simulate_input")
      }
      Psi_mat <- Psi
    } else {
      checkmate::assert_numeric(Psi, len = p, any.missing = FALSE, finite = TRUE)
      Psi_mat <- diag(Psi, p)
    }

    Sigma <- common + Psi_mat
    if (any(diag(Sigma) <= 0)) {
      cli::cli_abort(
        c("The assembled population has a non-positive variance and cannot be standardized.",
          "x" = "Check {.arg Lambda}, {.arg Phi}, and {.arg Psi}."),
        class = "efa_simulate_heywood"
      )
    }
    R_pop <- stats::cov2cor(Sigma)
    var_names <- rownames(Lambda)

  } else {

    R <- as.matrix(R)
    checkmate::assert_matrix(R, mode = "numeric", any.missing = FALSE,
                             min.rows = 1)
    # `any.missing` does not reject Inf, so check finiteness explicitly; otherwise a
    # non-finite R reaches eigen() below and aborts with an unclassed base error.
    if (nrow(R) != ncol(R) || !all(is.finite(R)) || !isSymmetric(unname(R))) {
      cli::cli_abort(
        c("{.arg R} must be a square, symmetric correlation matrix with finite values."),
        class = "efa_simulate_input")
    }
    p <- nrow(R)
    # Ordinal thresholds, the Vale-Maurelli cubic, and the `match = "polychoric"`
    # assertion all live on the standard-normal scale, so a non-unit diagonal is not a
    # rescaling of the requested population but a different one: the drawn variables
    # would keep the supplied variances while the categorization assumes unit variance.
    # Aborting keeps that visible; silently standardizing would hide the mistake.
    diag_dev <- max(abs(diag(R) - 1))
    if (diag_dev > tol) {
      cli::cli_abort(
        c("{.arg R} must be a correlation matrix, with a unit diagonal.",
          "x" = "Its largest deviation of a diagonal element from 1 is {.val {diag_dev}}.",
          "i" = "Standardize a covariance matrix with {.fun stats::cov2cor} first."),
        class = "efa_simulate_input")
    }
    R_pop <- R
    # Variable names live on either dimension of a correlation matrix; fall back to
    # the column names when only those are present.
    var_names <- rownames(R)
    if (is.null(var_names)) var_names <- colnames(R)

  }

  # Remove any floating-point asymmetry so the matrix square root sees an exactly
  # symmetric matrix, and attach variable names.
  R_pop <- (R_pop + t(R_pop)) / 2
  if (is.null(var_names)) var_names <- paste0("V", seq_len(p))
  dimnames(R_pop) <- list(var_names, var_names)

  # Reject an indefinite population: a matrix with a markedly negative eigenvalue
  # is not a valid covariance to simulate from. This runs before the `return_pop`
  # exit too, so `return_pop = TRUE` only ever returns a valid correlation matrix.
  # A positive-semidefinite but singular population (smallest eigenvalue in
  # [-tol, 0]) is allowed and handled by the compiled core's eigen square-root
  # fallback.
  min_ev <- min(eigen(R_pop, symmetric = TRUE, only.values = TRUE)$values)
  if (min_ev < -tol) {
    cli::cli_abort(
      c("The population correlation matrix is not positive semi-definite.",
        "x" = "Its smallest eigenvalue is {.val {min_ev}}.",
        "i" = "A valid population model must have a positive semi-definite correlation matrix."),
      class = "efa_simulate_not_pd"
    )
  }

  # A fixed seed makes the whole call reproducible and independent of the number of parallel
  # workers. The caller's RNG stream is saved and restored afterwards -- or, if none existed yet,
  # the stream set.seed() creates is removed again -- so efa_simulate() leaves no side effect on
  # it (mirrors EFA()'s bootstrap seeding). Seeding here (before the optional model-error draw and
  # before the replicate loop) puts both under one umbrella; the model-error draw runs once in the
  # main process, and nothing between it and the replicate loop consumes random numbers, so a call
  # without model error draws the identical stream it did before this argument existed.
  .set_local_seed(seed)

  # Inject model error: perturb the population so the q-factor model fits it with the target
  # RMSEA (CB/WB) or RMSEA and/or CFI (TKL), and record the achieved fit. Requires the factor-model
  # population (there is no specified model to misfit against a bare `R`, and CB needs the model
  # Jacobian) and residual degrees of freedom. Runs once here so the perturbed population feeds the
  # `return_pop` exit, the VM/IG setup, and every replicate.
  model_error_info <- NULL
  if (do_model_error) {
    if (!have_lambda) {
      cli::cli_abort(
        c("Model error requires a factor-model population.",
          "x" = "You supplied {.arg R}, so there is no factor model to misfit against.",
          "i" = "Provide {.arg Lambda} (with optional {.arg Phi}/{.arg Psi}) to use {.arg model_error}."),
        class = "efa_simulate_input"
      )
    }
    q_me <- ncol(Lambda)
    if (.efa_df(p, q_me) <= 0) {
      cli::cli_abort(
        c("The {q_me}-factor model has no residual degrees of freedom, so no model error can be injected.",
          "i" = "Model error needs {.code ((p - q)^2 - (p + q)) / 2 > 0}."),
        class = "efa_simulate_input"
      )
    }
    # All three methods invert the population (solve()/chol()); a positive-semidefinite but
    # singular population (e.g. a communality of exactly 1) has no inverse, so require it to be
    # positive definite here (min_ev is the smallest eigenvalue computed above) rather than let a
    # base-R linear-algebra error escape unclassed.
    if (min_ev < tol) {
      cli::cli_abort(
        c("Model error requires a positive-definite population correlation matrix.",
          "x" = "Its smallest eigenvalue is {.val {min_ev}} (the population is singular or near-singular).",
          "i" = "Reduce the communalities (a communality of 1 makes the population singular), or omit the model-error target."),
        class = "efa_simulate_model_error"
      )
    }
    me <- .apply_model_error(R_pop, common, Sigma, q_me, model_error,
                            target_rmsea, target_cfi, force_pd, tol)
    R_pop <- (me$population + t(me$population)) / 2
    dimnames(R_pop) <- list(var_names, var_names)
    model_error_info <- me$model_error
    # The perturbed population must still be a valid (positive semi-definite) correlation matrix.
    min_ev <- min(eigen(R_pop, symmetric = TRUE, only.values = TRUE)$values)
    if (min_ev < -tol) {
      cli::cli_abort(
        c("The model-error population correlation matrix is not positive semi-definite.",
          "x" = "Its smallest eigenvalue is {.val {min_ev}}.",
          "i" = "Lower the target RMSEA/CFI, or use a different {.arg model_error} method."),
        class = "efa_simulate_not_pd"
      )
    }
  }

  if (return_pop) {
    return(.new_efa_simulated(
      data = NULL, population = R_pop, model_error = model_error_info,
      N = NA_integer_, n_datasets = NA_integer_, marginals = marginals,
      categories = categories, match = match, missing = missing, seed = seed))
  }

  checkmate::assert_count(N, positive = TRUE)

  # The empirical path needs a marginal source and a factor count for the
  # rank-matching reproduction; validate both against the population dimension.
  if (marginals == "empirical") {
    # The reproduction sorts and correlates the drawn cases, which needs at least
    # two of them: a single case has no correlation and collapses the working
    # matrix to a vector.
    if (N < 2L) {
      cli::cli_abort(
        c("{.arg N} must be at least 2 when {.code marginals = \"empirical\"}.",
          "x" = "You requested {N} case{?s}."),
        class = "efa_simulate_input"
      )
    }
    if (is.null(marginal_data)) {
      cli::cli_abort(
        c("{.arg marginal_data} is required when {.code marginals = \"empirical\"}.",
          "i" = "Supply a matrix or data frame with {p} column{?s} giving the empirical marginals."),
        class = "efa_simulate_input"
      )
    }
    marginal_data <- as.matrix(marginal_data)
    checkmate::assert_matrix(marginal_data, mode = "numeric", any.missing = FALSE,
                             min.rows = 1)
    # any.missing = FALSE rejects NA/NaN but not Inf, so check finiteness too.
    if (!all(is.finite(marginal_data))) {
      cli::cli_abort("{.arg marginal_data} must contain only finite values.",
                     class = "efa_simulate_input")
    }
    if (ncol(marginal_data) != p) {
      cli::cli_abort(
        c("{.arg marginal_data} must have one column per variable ({p}).",
          "x" = "It has {ncol(marginal_data)} column{?s}."),
        class = "efa_simulate_input"
      )
    }
    # Each column must offer at least two distinct values: a constant column is
    # not a distribution to resample and cannot reproduce a correlation (its
    # sample correlation is undefined), and a single-row source would resample a
    # length-one vector -- both of which yield silently wrong or degenerate draws.
    n_distinct <- apply(marginal_data, 2L, function(col) length(unique(col)))
    if (any(n_distinct < 2L)) {
      bad <- as.character(which(n_distinct < 2L))
      cli::cli_abort(
        c("Every column of {.arg marginal_data} must have at least two distinct values.",
          "x" = "{cli::qty(bad)}Constant column{?s}: {bad}."),
        class = "efa_simulate_input"
      )
    }

    # Default the reproduction's factor count to the model's when the population
    # was built from Lambda; a supplied R carries no such count, so require it.
    if (is.null(n_factors)) {
      if (have_lambda) {
        n_factors <- ncol(Lambda)
      } else {
        cli::cli_abort(
          c("{.arg n_factors} is required when the population is supplied via {.arg R}.",
            "i" = "It sets the factor count of the rank-matching reproduction."),
          class = "efa_simulate_input"
        )
      }
    }
    checkmate::assert_count(n_factors, positive = TRUE)
    # The iterative principal-axis step extracts n_factors factors from a p by p
    # matrix, which requires n_factors < p.
    if (n_factors >= p) {
      cli::cli_abort(
        c("{.arg n_factors} must be smaller than the number of variables ({p}).",
          "x" = "The reproduction uses {n_factors} factor{?s} for {p} variable{?s}.",
          "i" = "When not supplied, {.arg n_factors} defaults to the number of columns of {.arg Lambda}."),
        class = "efa_simulate_input"
      )
    }
  }

  # The VM/IG non-normal paths need target marginal moments; validate them,
  # recycle to length p, and precompute the Fleishman transform once. None of the
  # solves below draw random numbers, so they live outside the replicate loop.
  vm_ftable <- vm_Rinter <- ig_ftable <- ig_L <- ig_Lt <- NULL
  kurt_default <- FALSE
  if (marginals %in% c("VM", "IG")) {
    if (is.null(skewness) && is.null(kurtosis)) {
      cli::cli_abort(
        c("{.arg skewness} and/or {.arg kurtosis} are required when {.code marginals} is {.val VM} or {.val IG}.",
          "i" = "The unset one defaults to 0 (the normal-marginal value)."),
        class = "efa_simulate_input"
      )
    }
    # Recorded before the recycling below turns an unset moment into zeros, so an
    # infeasible-moment abort can name the default as the likely cause.
    kurt_default <- is.null(kurtosis)
    skewness <- .moment_vec(skewness, p, "skewness")
    kurtosis <- .moment_vec(kurtosis, p, "kurtosis")
    # Universal moment bound (Pearson, 1916): every distribution has excess
    # kurtosis at least skewness^2 - 2. Fleishman's cubic covers a smaller region
    # still, which the per-variable coefficient solve screens.
    infeasible <- kurtosis < skewness^2 - 2
    if (any(infeasible)) {
      bad <- as.character(which(infeasible))
      cli::cli_abort(
        c("The requested marginal moments are not attainable by any distribution.",
          "x" = "{cli::qty(bad)}Variable{?s} {bad}: excess {.arg kurtosis} is below {.arg skewness}^2 - 2.",
          "i" = "Every distribution must have excess kurtosis >= skewness^2 - 2 (Pearson, 1916)."),
        class = "efa_simulate_infeasible_moments"
      )
    }
  }

  if (marginals == "VM") {
    # Vale-Maurelli (1983): draw normal deviates at an intermediate correlation
    # chosen so the per-variable Fleishman cubic reproduces the target correlation.
    vm_ftable <- .fleishman_table(skewness, kurtosis, kurt_default)
    vm_Rinter <- .vm_intermediate_cor(R_pop, vm_ftable)
    # The pairwise cubic solves ignore joint positive-definiteness, so the
    # intermediate matrix can be indefinite even when R_pop is not. Reject it, or
    # project it to the nearest correlation matrix on request.
    min_ev_i <- min(eigen(vm_Rinter, symmetric = TRUE, only.values = TRUE)$values)
    if (min_ev_i < -tol) {
      if (!force_pd) {
        cli::cli_abort(
          c("The Vale-Maurelli intermediate correlation matrix is not positive definite.",
            "x" = "Its smallest eigenvalue is {.val {min_ev_i}}.",
            "i" = "Set {.code force_pd = TRUE} to project it to the nearest correlation matrix, or lower the target correlations or non-normality."),
          class = "efa_simulate_intermediate_not_pd"
        )
      }
      # Muffle cor.smooth's routine note; re-surface a single classed warning.
      vm_Rinter <- withCallingHandlers(
        psych::cor.smooth(vm_Rinter),
        warning = function(w) {
          if (grepl("smoothing was done", conditionMessage(w), fixed = TRUE)) {
            invokeRestart("muffleWarning")
          }
        }
      )
      vm_Rinter <- (vm_Rinter + t(vm_Rinter)) / 2
      # The projected intermediate matrix no longer solves the Vale-Maurelli cubic, so the
      # data are drawn from a population other than the target. Push the projected
      # intermediate correlations back through that same cubic to obtain the population
      # actually drawn from in closed form (no simulation), and report it in `population`
      # so a recovery or bias study measures against the right target.
      R_ach <- .vm_achieved_cor(vm_Rinter, vm_ftable)
      drift <- max(abs(R_ach - R_pop)[lower.tri(R_pop)])
      R_pop <- R_ach
      dimnames(R_pop) <- list(var_names, var_names)
      cli::cli_warn(
        c("The Vale-Maurelli intermediate correlation matrix was not positive definite; it has been projected to the nearest correlation matrix.",
          "i" = "Projection was applied via {.fun psych::cor.smooth}; the population drawn from therefore drifts from the target by up to {.val {round(drift, 4)}} (largest absolute correlation difference).",
          "i" = "The returned {.field population} is the population actually drawn from, not the target."),
        class = "efa_simulate_pd_forced"
      )
    }

  } else if (marginals == "IG") {
    # Independent generator (Foldnes & Olsson, 2016): X = L Z with L the lower
    # Cholesky factor of R_pop and Z independent non-normal generators.
    ig_L <- tryCatch(t(chol(R_pop)), error = function(e) NULL)
    if (is.null(ig_L)) {
      cli::cli_abort(
        c("{.code marginals = \"IG\"} needs a positive definite population correlation matrix.",
          "x" = "The population is singular, so it has no Cholesky factor.",
          "i" = "Use {.code marginals = \"VM\"} or {.code marginals = \"empirical\"} for a singular population."),
        class = "efa_simulate_input"
      )
    }
    # Cumulant additivity turns the target marginal moments into the independent
    # generators' moments via lower-triangular solves; these depend on the
    # population, so a feasible request can still imply an infeasible generator.
    gen <- .ig_generator_moments(ig_L, skewness, kurtosis)
    # Report generator infeasibility in terms of the request the user made -- the
    # implied generator moments below are internal and not directly controllable.
    # Fleishman's region is tighter than the universal bound, so also translate a
    # solve failure (raised inside .fleishman_table) into this same message.
    ig_infeasible <- function() {
      cli::cli_abort(
        c("No independent generator can produce the requested marginals from this population.",
          "x" = "The moments the generators would need are not attainable.",
          "i" = "Lower the target correlations or the non-normality, or use {.code marginals = \"VM\"}."),
        class = "efa_simulate_infeasible_moments"
      )
    }
    if (any(gen$kurtosis < gen$skewness^2 - 2)) ig_infeasible()
    ig_ftable <- tryCatch(
      .fleishman_table(gen$skewness, gen$kurtosis),
      efa_simulate_infeasible_moments = function(e) ig_infeasible()
    )
    # Transpose of the Cholesky factor for the X = Z L' mixing; loop-invariant, so
    # precompute it once here rather than re-forming it in every replicate.
    ig_Lt <- t(ig_L)
  }

  # Resolve and validate the missing-data settings once (they draw no random numbers).
  # The mask itself is applied per replicate inside the draw so it shares each
  # replicate's reproducible stream. `miss_b` is the logistic slope (MAR/MNAR),
  # `miss_vars` the columns to hole, and `pred_idx` their MAR predictor columns; all
  # stay NULL when unused.
  miss_b <- NULL
  pred_idx <- NULL
  miss_vars <- NULL
  if (missing != "none") {
    # A proportion strictly inside (0, 1): 0 or 1 is not a mechanism to calibrate, and
    # the boundary would send qlogis() to +-Inf and break the intercept root-find.
    if (!checkmate::test_number(missing_prop) ||
        missing_prop <= 0 || missing_prop >= 1) {
      cli::cli_abort(
        c("{.arg missing_prop} must be a single number strictly between 0 and 1.",
          "x" = "It is the target marginal proportion of missing values."),
        class = "efa_simulate_input"
      )
    }
    if (missing %in% c("MAR", "MNAR")) {
      # The slope defaults to 1 (a moderate dependence); 0 removes the dependence, which
      # is MCAR at the same rate; large magnitudes make missingness near-deterministic.
      miss_b <- if (is.null(missing_strength)) 1 else missing_strength
      if (!checkmate::test_number(miss_b, finite = TRUE)) {
        cli::cli_abort(
          c("{.arg missing_strength} must be a single finite number.",
            "i" = "It is the slope of the logistic missingness model; larger magnitudes make missingness depend more strongly on the predictor."),
          class = "efa_simulate_input"
        )
      }
    }
    miss_vars <- .resolve_missing_vars(missing_vars, p, var_names)
    if (missing == "MAR") {
      pred_idx <- .resolve_missing_predictor(missing_predictor, p, var_names,
                                            miss_vars)
    }
  }

  # With `categories`, the drawn data are cut into ordered categories at thresholds
  # on the standard-normal scale. The thresholds draw no random numbers and are the
  # same for every replicate, so they are built once here; the cutting itself runs
  # after the draw (below), keeping the replicate RNG streams untouched.
  ord_thresholds <- if (!is.null(categories)) {
    .ordinal_thresholds(categories, p)
  }
  # The Vale-Maurelli draw is a cubic map of a standard normal, so the normal-scale
  # thresholds would cut it at the wrong quantiles; push them through the same cubic
  # so the requested proportions are reproduced on the non-normal scale. A variable
  # whose cubic does not increase across its own thresholds, or folds tail mass back
  # below an outermost cut, cannot be mapped faithfully and keeps the normal ones.
  if (!is.null(ord_thresholds) && marginals == "VM") {
    mapped <- .vm_thresholds(ord_thresholds, vm_ftable)
    ord_thresholds <- mapped$thresholds
    if (length(mapped$non_monotone)) {
      kept <- var_names[mapped$non_monotone]
      cli::cli_warn(
        c("{cli::qty(kept)}Variable{?s} {.val {kept}} kept standard-normal category thresholds.",
          "x" = "{cli::qty(kept)}{?Its/Their} Fleishman polynomial{?s} {?is/are} not increasing over the requested thresholds and the tails beyond them, so mapping the thresholds would misassign category proportions.",
          "i" = "The achieved category proportions of {cli::qty(kept)}{?this/these} variable{?s} depart from the request; less extreme category proportions or a milder {.arg skewness}/{.arg kurtosis} restore the mapping."),
        class = "efa_simulate_threshold_fallback"
      )
    }
  }

  # Draw each dataset in its own future; future.seed = TRUE assigns every
  # replicate a reproducible L'Ecuyer-CMRG stream, so the result is identical at
  # any number of workers (mirrors the np-boot fit in .boot_fun()).
  datasets <- future.apply::future_lapply(
    seq_len(n_datasets),
    function(i) {
      dat <- if (marginals == "empirical") {
        .simulate_cfm_empirical(R_pop, marginal_data, n_factors, N)
      } else if (marginals == "VM") {
        # normal deviates at the intermediate correlation, then the cubic map.
        .fleishman_transform(.simulate_cfm_mvn(vm_Rinter, N), vm_ftable)
      } else if (marginals == "IG") {
        # independent Fleishman generators mixed by the Cholesky factor: X = Z L'.
        z <- .fleishman_transform(matrix(stats::rnorm(N * p), N, p), ig_ftable)
        z %*% ig_Lt
      } else {
        .simulate_cfm_mvn(R_pop, N)
      }
      colnames(dat) <- var_names
      # Introduce missing values on the drawn (continuous) data; runs inside the future
      # so the mask draws from this replicate's stream (reproducible across worker
      # counts). When `categories` is set, the NAs propagate into the codes below.
      if (missing != "none") {
        dat <- .apply_missing(dat, missing, missing_prop, miss_b, pred_idx,
                              miss_vars)
      }
      dat
    },
    future.seed = TRUE
  )

  # Discretize to ordered categories when requested. Cutting is deterministic given
  # the continuous draw, so it runs in the main process (leaving the draw and its
  # per-replicate streams unchanged) with the same thresholds for every replicate. A
  # draw can leave a requested category empty -- more likely at small `N` or, under
  # `match = "thresholds"`, with strongly non-normal marginals -- which destabilizes
  # the polychoric correlation; report any such variable once.
  if (!is.null(ord_thresholds)) {
    k_req <- vapply(ord_thresholds, function(tau) length(tau) + 1L, integer(1L))
    datasets <- lapply(datasets, .ordinalize, thresholds = ord_thresholds)
    empty <- character(0L)
    for (d in datasets) {
      # Count non-empty categories per variable with an O(N) tabulate over the 1..K
      # codes; tabulate ignores any NA codes (from masked values), so no missing-value
      # filtering is needed.
      filled <- vapply(seq_len(ncol(d)),
                       function(j) sum(tabulate(d[, j], nbins = k_req[j]) > 0L),
                       integer(1L))
      empty <- union(empty, colnames(d)[filled < k_req])
    }
    if (length(empty)) {
      cli::cli_warn(
        c("{cli::qty(empty)}Variable{?s} {.val {empty}} did not fill every requested category.",
          "i" = "Empty or near-empty categories destabilize the polychoric correlation and the factor analysis; increase {.arg N} or request fewer categories."),
        class = "efa_simulate_empty_category"
      )
    }
  }

  data <- if (n_datasets == 1L) datasets[[1L]] else datasets
  .new_efa_simulated(
    data = data, population = R_pop, model_error = model_error_info,
    N = N, n_datasets = n_datasets, marginals = marginals,
    categories = categories, match = match, missing = missing, seed = seed)

}


# Construct the `efa_simulated` result object: the drawn `data` (an N x p matrix, a length-
# n_datasets list of them, or NULL for return_pop = TRUE), the `population` correlation matrix
# actually drawn from (model-error-perturbed when requested), the `model_error` record (NULL when
# the population is exact), and a `settings` list for provenance and printing.
.new_efa_simulated <- function(data, population, model_error, N, n_datasets,
                               marginals, categories, match, missing, seed) {
  structure(
    list(
      data = data,
      population = population,
      model_error = model_error,
      settings = list(N = N, n_datasets = n_datasets, marginals = marginals,
                      categories = categories, match = match, missing = missing,
                      seed = seed)
    ),
    class = "efa_simulated"
  )
}


# Ruscio-Kaczetow iterative rank-matching: draw `N` cases that reproduce the
# target correlation `R` while carrying the empirical column marginals of `x`
# (each column is resampled with replacement from the matching column of `x`).
# The target correlation and the marginal source are separate arguments so an
# arbitrary population `R` can be paired with any supplied marginals; the shared
# rank-matching loop otherwise follows Ruscio & Kaczetow (2008). `stats::cor()`,
# `sort()`, `sort.list()`, and `.paf_iter()` draw no random numbers, so with
# `R = stats::cor(x, method = cor_method)` the stream, and hence the output, is
# identical to the observed-data special case that `CD()` uses to draw its
# comparison populations. `x` may be a data frame or a matrix (`CD()` passes its
# raw input, `efa_simulate()` a matrix); `x[, i]` and `sample()` treat both
# identically.
# Ruscio, J., & Kaczetow, W. (2008). Simulating multivariate nonnormal data
# using an iterative algorithm. Multivariate Behavioral Research, 43(3),
# 355-381. doi: 10.1080/00273170802285693
.simulate_cfm_empirical <- function(R, x, n_factors, N, cor_method = "pearson",
                                    max_trials = 5, initial_multiplier = 1,
                                    max_iter = 100) {

  k <- ncol(x)
  # The reproduction correlates the drawn cases, which needs at least two of them; with
  # fewer there is no correlation to match and the row indexing below has nothing to read.
  # efa_simulate() gates this at its own boundary, so this catches the CD path.
  if (N < 2L) {
    cli::cli_abort(
      c("At least 2 cases must be drawn to reproduce the population correlation.",
        "x" = "You asked for {N}."),
      class = "efa_simulate_degenerate_marginal"
    )
  }
  sim_dat <- matrix(0, nrow = N, ncol = k)         # Simulated data
  dists <- matrix(0, nrow = N, ncol = k)           # Each variable's marginal
  best_RMSE <- 1                                    # Lowest RMSE correlation
  t_no_impr <- 0                                    # Trial counter

  # Resample the supplied empirical marginals (sorted) for each variable. A resample can
  # come out constant even from a non-constant source -- a category rare enough that N
  # draws miss all but one of its values -- and a constant column has no correlation, so
  # the RMSE below would be NA and the acceptance test would fail on a missing value.
  # Report it here instead (the sorted column makes the check its two extremes).
  for (i in 1:k) {
    dists[, i] <- sort(sample(x[, i], size = N, replace = TRUE))
  }
  constant <- dists[1L, ] == dists[N, ]
  if (any(constant)) {
    bad <- as.character(which(constant))
    cli::cli_abort(
      c("{cli::qty(bad)}Resampling the empirical marginals produced {?a constant column/constant columns} ({bad}).",
        "x" = "A constant column has no correlation, so the population cannot be reproduced.",
        "i" = "Draw more cases, or use a marginal whose rarest category is less rare."),
      class = "efa_simulate_degenerate_marginal"
    )
  }

  # The target correlation is supplied; keep a working and a best-so-far copy so
  # the no-improvement branch below is safe even if the first iteration does not
  # lower the initial RMSE (best_RMSE = 1).
  R_inter <- R
  R_best <- R_inter
  res_best <- matrix(0, nrow = k, ncol = k)

  # Random normal shared and unique components; factor loadings applied below.
  shared_comp <- matrix(stats::rnorm(N * n_factors, 0, 1), nrow = N,
                        ncol = n_factors)
  unique_comp <- matrix(stats::rnorm(N * k, 0, 1), nrow = N, ncol = k)
  shared_load <- matrix(0, nrow = k, ncol = n_factors)
  unique_load <- matrix(0, nrow = k, ncol = 1)

  # Iterate until a set number of trials pass without improving the RMSE between
  # the reproduced and target correlations.
  while (t_no_impr < max_trials) {

    # Factor loadings for the current intermediate correlation.
    L <- suppressWarnings(.paf_iter(rep(1, k), criterion = .001, R = R_inter,
                   n_fac = n_factors, abs_eig = TRUE, crit_type = 1,
                   max_iter = max_iter)$L)

    shared_load[, 1:n_factors] <- L

    # get rid of Heywood cases
    shared_load[shared_load > 1] <- 1
    shared_load[shared_load < -1] <- -1

    if (shared_load[1, 1] < 0) {
      shared_load <- shared_load * -1
    }

    for (i in seq_len(k)) {
      if (sum(shared_load[i, ] * shared_load[i, ]) < 1) {
        unique_load[i, 1] <-
          (1 - sum(shared_load[i, ] * shared_load[i, ]))
      } else {
        unique_load[i, 1] <- 0
      }
    }

    unique_load <- sqrt(unique_load)

    for (i in seq_len(k)) {
      sim_dat[, i] <- (shared_comp %*% t(shared_load))[, i] +
        unique_comp[, i] * unique_load[i, 1]
    }

    # Replace the normal scores with the empirical marginals via rank matching.
    for (i in seq_len(k)) {
      sim_dat <- sim_dat[sort.list(sim_dat[, i]), ]
      sim_dat[, i] <- dists[, i]
    }

    # RMSE between reproduced and target correlations; accept or shrink the step.
    R_rep <- stats::cor(sim_dat, method = cor_method)
    R_res <- R - R_rep
    RMSE <- sqrt(sum(R_res[lower.tri(R_res)] * R_res[lower.tri(R_res)]) /
                   (.5 * (k * k - k)))

    if (RMSE < best_RMSE) {
      best_RMSE <- RMSE
      R_best <- R_inter
      res_best <- R_res
      R_inter <- R_inter + initial_multiplier * R_res
      t_no_impr <- 0
    } else {
      t_no_impr <- t_no_impr + 1
      current_multiplier <- initial_multiplier * .5 ^ t_no_impr
      R_inter <- R_best + current_multiplier * res_best
    }
  }

  # Reconstruct the data set from the best intermediate correlation.
  L <- suppressWarnings(.paf_iter(rep(1, k), criterion = .001, R = R_best,
                n_fac = n_factors, abs_eig = TRUE, crit_type = 1,
                max_iter = max_iter)$L)
  shared_load[, seq_len(n_factors)] <- L

  shared_load[shared_load > 1] <- 1
  shared_load[shared_load < -1] <- -1
  if (shared_load[1, 1] < 0) {
    shared_load <- shared_load * -1
  }

  for (i in seq_len(k)) {
    if (sum(shared_load[i, ] * shared_load[i, ]) < 1) {
      unique_load[i, 1] <-
        (1 - sum(shared_load[i, ] * shared_load[i, ]))
    } else {
      unique_load[i, 1] <- 0
    }
  }

  unique_load <- sqrt(unique_load)

  for (i in seq_len(k)) {
    sim_dat[, i] <- (shared_comp %*% t(shared_load))[, i] +
      unique_comp[, i] * unique_load[i, 1]
  }

  for (i in seq_len(k)) {
    sim_dat <- sim_dat[sort.list(sim_dat[, i]), ]
    sim_dat[, i] <- dists[, i]
  }

  sim_dat
}


# Recycle a length-1 or length-p marginal-moment argument to length p (NULL gives
# zeros, the normal-marginal value), validating type and finiteness.
.moment_vec <- function(x, p, arg) {
  if (is.null(x)) return(rep(0, p))
  # A matrix passes is.numeric() and, if it has p entries, the length check below,
  # so reject any dimensioned object here rather than flatten it column-major.
  if (!is.numeric(x) || !is.null(dim(x)) || !all(is.finite(x))) {
    cli::cli_abort("{.arg {arg}} must be a finite numeric vector.",
                   class = "efa_simulate_input")
  }
  if (length(x) == 1L) return(rep(as.numeric(x), p))
  if (length(x) != p) {
    cli::cli_abort(
      c("{.arg {arg}} must have length 1 or {p} (one value per variable).",
        "x" = "It has length {length(x)}."),
      class = "efa_simulate_input"
    )
  }
  as.numeric(x)
}

# Fleishman (1978) cubic coefficients (a, b, c, d) with a = -c that transform a
# standard normal Z into a unit-variance variable with the target skewness and
# excess kurtosis via a + bZ + cZ^2 + dZ^3. The three power-method equations are
# minimised as a sum of squares (as in lavaan/semTools); a non-zero minimum means
# the target lies outside Fleishman's feasible region and is rejected. Fleishman's region
# leaves the zero-kurtosis line at a skewness of about 0.855, so `kurtosis_default` marks
# a request that only became infeasible through the unset moment defaulting to 0 -- the
# most likely cause of a rejection, and one the moments alone do not reveal.
# Fleishman, A. I. (1978). A method for simulating non-normal distributions.
# Psychometrika, 43(4), 521-532.
.fleishman_coef <- function(skewness, kurtosis, kurtosis_default = FALSE) {
  sys <- function(x) {
    b. <- x[1L]; c. <- x[2L]; d. <- x[3L]
    eq1 <- b.^2 + 6 * b. * d. + 2 * c.^2 + 15 * d.^2 - 1
    eq2 <- 2 * c. * (b.^2 + 24 * b. * d. + 105 * d.^2 + 2) - skewness
    eq3 <- 24 * (b. * d. + c.^2 * (1 + b.^2 + 28 * b. * d.) +
                 d.^2 * (12 + 48 * b. * d. + 141 * c.^2 + 225 * d.^2)) - kurtosis
    eq1^2 + eq2^2 + eq3^2
  }
  out <- stats::nlminb(start = c(1, 0, 0), objective = sys, scale = 10,
                       control = list(trace = 0L))
  if (out$objective > 1e-6) {
    cli::cli_abort(
      c("The marginal moments (skewness {.val {skewness}}, excess kurtosis {.val {kurtosis}}) are not representable by a Fleishman polynomial.",
        if (isTRUE(kurtosis_default)) {
          c("x" = "{.arg kurtosis} was not supplied and defaulted to 0, which a skewness of {.val {skewness}} needs more excess kurtosis than.")
        },
        "i" = "Fleishman's cubic covers a smaller region than excess kurtosis >= skewness^2 - 2; choose moments further inside it."),
      class = "efa_simulate_infeasible_moments"
    )
  }
  b. <- out$par[1L]; c. <- out$par[2L]; d. <- out$par[3L]
  c(-c., b., c., d.)
}

# Fleishman coefficients for every variable as a p x 4 matrix of (a, b, c, d).
.fleishman_table <- function(skewness, kurtosis, kurtosis_default = FALSE) {
  t(vapply(seq_along(skewness),
           function(i) .fleishman_coef(skewness[i], kurtosis[i], kurtosis_default),
           numeric(4L)))
}

# Apply each variable's Fleishman cubic to the columns of Z (a p x 4 coefficient
# table): X_i = a_i + b_i Z_i + c_i Z_i^2 + d_i Z_i^3.
.fleishman_transform <- function(Z, ftable) {
  for (i in seq_len(ncol(Z))) {
    z <- Z[, i]
    Z[, i] <- ftable[i, 1L] + ftable[i, 2L] * z + ftable[i, 3L] * z^2 +
      ftable[i, 4L] * z^3
  }
  Z
}

# Pairwise coefficients of the Vale-Maurelli (1983, equation 11) cubic relating the
# intermediate normal correlation rho to the attained correlation of the two Fleishman
# cubics: cor = k1 rho + k2 rho^2 + k3 rho^3. Returned as three p x p matrices, so
# .vm_intermediate_cor() (which solves the cubic for rho) and .vm_achieved_cor() (which
# evaluates it at a given rho) cannot drift apart.
# Vale, C. D., & Maurelli, V. A. (1983). Simulating multivariate nonnormal
# distributions. Psychometrika, 48(3), 465-471.
.vm_cubic_coefs <- function(ftable) {
  b <- ftable[, 2L]; cc <- ftable[, 3L]; d <- ftable[, 4L]
  list(k1 = outer(b, b) + 3 * outer(b, d) + 3 * outer(d, b) + 9 * outer(d, d),
       k2 = 2 * outer(cc, cc),
       k3 = 6 * outer(d, d))
}

# Vale-Maurelli (1983) intermediate normal correlation for each variable pair:
# solve the cubic in rho relating the intermediate normal correlation to the
# target correlation R_ij given the two variables' Fleishman coefficients. The
# cubic can admit several real roots in [-1, 1]; take the one nearest the target
# (Olvera Astivia & Zumbo, 2019). Base polyroot() is used rather than a numeric
# minimiser so all roots are found explicitly.
.vm_intermediate_cor <- function(R, ftable) {
  p <- ncol(R)
  kk <- .vm_cubic_coefs(ftable)
  ICOR <- diag(p)
  for (j in seq_len(p - 1L)) {
    for (i in (j + 1L):p) {
      target <- R[i, j]
      if (target == 0) next                # an uncorrelated pair stays at rho = 0
      # coefficients of c0 + c1 rho + c2 rho^2 + c3 rho^3
      coefs <- c(-target, kk$k1[i, j], kk$k2[i, j], kk$k3[i, j])
      # drop trailing zero coefficients so polyroot sees the true polynomial degree
      while (length(coefs) > 1L && coefs[length(coefs)] == 0) {
        coefs <- coefs[-length(coefs)]
      }
      roots <- polyroot(coefs)
      real <- Re(roots)[abs(Im(roots)) < 1e-8]
      if (!length(real)) real <- Re(roots)[which.min(abs(Im(roots)))]
      in_range <- real[real >= -1 & real <= 1]
      cand <- if (length(in_range)) in_range else real
      # A degenerate cubic (all higher coefficients zero) leaves no root; fall back
      # to the target so the assignment below never sees a zero-length value.
      if (!length(cand)) cand <- target
      rho <- cand[which.min(abs(cand - target))]
      ICOR[i, j] <- ICOR[j, i] <- rho
    }
  }
  ICOR
}

# The population correlation a Vale-Maurelli draw at intermediate correlation `Rinter`
# actually attains: the forward direction of the cubic .vm_intermediate_cor() inverts,
# evaluated at `Rinter` with the shared coefficients. Needed when the intermediate matrix
# has been projected to positive definiteness and therefore no longer solves the cubic for
# the target correlation.
.vm_achieved_cor <- function(Rinter, ftable) {
  kk <- .vm_cubic_coefs(ftable)
  R <- Rinter * kk$k1 + Rinter^2 * kk$k2 + Rinter^3 * kk$k3
  diag(R) <- 1
  (R + t(R)) / 2
}

# Independent-generator moments (Foldnes & Olsson, 2016). With X = L Z, L the
# lower Cholesky factor of the target correlation and Z independent with mean 0
# and variance 1, the additivity and homogeneity of cumulants give
# skewness(X) = (L^3) skewness(Z) and excess-kurtosis(X) = (L^4) kurtosis(Z);
# solve the lower-triangular systems for the generators' moments.
# Foldnes, N., & Olsson, U. H. (2016). A simple simulation technique for
# nonnormal data with prespecified skewness, kurtosis, and covariance matrix.
# Multivariate Behavioral Research, 51(2-3), 207-219.
.ig_generator_moments <- function(L, skewness, kurtosis) {
  list(skewness = forwardsolve(L^3, skewness),
       kurtosis = forwardsolve(L^4, kurtosis))
}


# Interior category thresholds on the standard-normal scale for each variable, from
# a `categories` specification: either a scalar or length-p count of equally probable
# categories, or a length-p list of per-variable category proportions. Cutting a
# standard normal at tau = qnorm(cumsum(prop)), dropping the trailing +Inf, yields
# categories with population proportions `prop` (Olsson, 1979). Returns a length-p
# list of numeric threshold vectors (a variable with K categories has K - 1 of them).
# Olsson, U. (1979). Maximum likelihood estimation of the polychoric correlation
# coefficient. Psychometrika, 44(4), 443-460.
.ordinal_thresholds <- function(categories, p) {
  if (is.list(categories)) {
    if (length(categories) != p) {
      cli::cli_abort(
        c("{.arg categories}, as a list of category proportions, must have one element per variable ({p}).",
          "x" = "It has {length(categories)} element{?s}."),
        class = "efa_simulate_input"
      )
    }
    lapply(seq_along(categories), function(j) {
      prop <- categories[[j]]
      # Reject a dimensioned object (matrix/array): it passes is.numeric() and would
      # be flattened column-major by cumsum() instead of read as a proportion vector
      # (mirrors the guard in .moment_vec()).
      if (!is.numeric(prop) || !is.null(dim(prop)) || length(prop) < 2L ||
          !all(is.finite(prop))) {
        cli::cli_abort(
          c("Each element of {.arg categories} must be a finite numeric vector of at least two category proportions.",
            "x" = "Element {j} is not."),
          class = "efa_simulate_input"
        )
      }
      if (any(prop <= 0)) {
        cli::cli_abort(
          c("Category proportions in {.arg categories} must be strictly positive.",
            "x" = "Element {j} has a non-positive proportion (a zero proportion is an empty category with no threshold)."),
          class = "efa_simulate_input"
        )
      }
      if (abs(sum(prop) - 1) > 1e-6) {
        cli::cli_abort(
          c("Category proportions in {.arg categories} must sum to 1.",
            "x" = "Element {j} sums to {.val {sum(prop)}}."),
          class = "efa_simulate_input"
        )
      }
      # Renormalize onto the exact simplex: floating-point slack in `prop` (within the
      # tolerance above) could otherwise push an interior cumulative proportion past 1,
      # which qnorm would turn into a NaN threshold. The outer gate has already rejected
      # any genuine mismatch. An interior cumulative of exactly 1 survives this and gives
      # a +Inf threshold, which is legal -- findInterval() handles it and the category
      # above is simply always empty, which the empty-category warning reports.
      prop <- prop / sum(prop)
      stats::qnorm(cumsum(prop)[-length(prop)])
    })
  } else {
    # Reject a dimensioned object, as for the proportion-vector branch above.
    if (!is.numeric(categories) || !is.null(dim(categories)) ||
        !all(is.finite(categories)) || any(categories != round(categories))) {
      cli::cli_abort(
        "{.arg categories} must be a whole number of categories per variable, or a list of category proportions.",
        class = "efa_simulate_input"
      )
    }
    if (length(categories) != 1L && length(categories) != p) {
      cli::cli_abort(
        c("{.arg categories}, as a count of categories, must have length 1 or {p} (one per variable).",
          "x" = "It has length {length(categories)}."),
        class = "efa_simulate_input"
      )
    }
    if (any(categories < 2L)) {
      cli::cli_abort(
        c("{.arg categories} must request at least two categories per variable.",
          "x" = "The smallest requested count is {.val {min(categories)}}."),
        class = "efa_simulate_input"
      )
    }
    counts <- if (length(categories) == 1L) rep(categories, p) else categories
    lapply(counts, function(k) stats::qnorm(seq_len(k - 1L) / k))
  }
}


# Exact probability mass a Fleishman cubic f assigns at or below the cut f(t0): decompose
# the real line at the real roots of f(z) = f(t0) and add the standard-normal mass of every
# interval on which f - f(t0) <= 0 (sign tested at an interior point). Used to detect tail
# fold-back: for a non-monotone cubic, P(f(Z) <= f(t0)) can differ from pnorm(t0) even when
# f increases across the thresholds themselves, because a tail segment beyond t0 dips back
# below the cut. Returns NA for a degenerate (constant) map.
.vm_cut_mass <- function(t0, b, cc, d) {
  # f(z) - f(t0) in increasing degree; the constant Fleishman term cancels
  co <- c(-(b * t0 + cc * t0^2 + d * t0^3), b, cc, d)
  while (length(co) > 1L && co[length(co)] == 0) co <- co[-length(co)]
  if (length(co) <= 1L) return(NA_real_)
  r <- polyroot(co)
  re <- sort(unique(Re(r[abs(Im(r)) < 1e-8])))
  cuts <- c(-Inf, re, Inf)
  mass <- 0
  for (i in seq_len(length(cuts) - 1L)) {
    lo <- cuts[i]; hi <- cuts[i + 1L]
    mid <- if (is.infinite(lo) && is.infinite(hi)) 0
    else if (is.infinite(lo)) hi - 1
    else if (is.infinite(hi)) lo + 1
    else (lo + hi) / 2
    if (b * mid + cc * mid^2 + d * mid^3 -
        (b * t0 + cc * t0^2 + d * t0^3) <= 0) {
      mass <- mass + (stats::pnorm(hi) - stats::pnorm(lo))
    }
  }
  mass
}

# Map standard-normal ordinal thresholds onto the Vale-Maurelli scale. The VM draw is
# X = f(Z) with f the per-variable Fleishman cubic a + bZ + cZ^2 + dZ^3, so where f is
# order-preserving over (essentially) all the normal mass, P(X <= f(tau)) = P(Z <= tau) and
# cutting X at f(tau) reproduces exactly the proportions tau was built for. Two screens
# check this. First, f must increase across the requested thresholds themselves, or the cut
# points come out of order: the derivative b + 2cz + 3dz^2 is a parabola, so its minimum
# over the closed threshold range is attained at an endpoint or, when the parabola opens
# upward (d > 0), at the vertex -c/(3d) if that falls inside. Second, the tails beyond the
# outermost thresholds must not fold back below the cuts: a non-monotone cubic (every
# platykurtic marginal has d < 0; a skewed one can dip on one side) re-crosses an outer cut
# value at some more extreme z, and the mass beyond that re-crossing is silently assigned to
# the wrong side -- 75% of a 1.4% top category at excess kurtosis -1, 84% of a 5% bottom
# category at skewness 1.5 / excess kurtosis 3. The exact misassigned mass at the two
# outermost thresholds is compared against the requested outer-category proportions
# (checking the outermost cuts suffices: an interior cut's re-crossing lies further out and
# folds strictly less mass) and a variable fails when it is off by more than 5% of the
# proportion -- departures below that are negligible against the error the normal-scale
# fallback itself incurs. The indices of the failing variables are returned; those keep
# their normal-scale thresholds. Infinite thresholds are passed through: an increasing map
# sends +-Inf to itself, and evaluating the polynomial there could produce a NaN cut point.
# Fleishman, A. I. (1978). A method for simulating non-normal distributions.
# Psychometrika, 43(4), 521-532.
.vm_thresholds <- function(thresholds, ftable) {
  non_monotone <- integer(0L)
  for (j in seq_along(thresholds)) {
    tau <- thresholds[[j]]
    fin <- is.finite(tau)
    if (!any(fin)) next                 # nothing to map
    b <- ftable[j, 2L]; cc <- ftable[j, 3L]; d <- ftable[j, 4L]
    rng <- range(tau[fin])
    z <- rng
    if (d > 0) {
      zv <- -cc / (3 * d)
      if (zv > z[1L] && zv < z[2L]) z <- c(z, zv)
    }
    if (any(b + 2 * cc * z + 3 * d * z^2 <= 0)) {
      non_monotone <- c(non_monotone, j)
      next
    }
    p_lo <- stats::pnorm(rng[1L])
    p_hi <- 1 - stats::pnorm(rng[2L])
    m_lo <- .vm_cut_mass(rng[1L], b, cc, d)
    m_hi <- .vm_cut_mass(rng[2L], b, cc, d)
    if (is.na(m_lo) || is.na(m_hi) ||
        abs(m_lo - p_lo) > 0.05 * p_lo ||
        abs((1 - m_hi) - p_hi) > 0.05 * p_hi) {
      non_monotone <- c(non_monotone, j)
      next
    }
    tau[fin] <- .fleishman_transform(matrix(tau[fin], ncol = 1L),
                                     ftable[j, , drop = FALSE])[, 1L]
    thresholds[[j]] <- tau
  }
  list(thresholds = thresholds, non_monotone = non_monotone)
}


# Discretize the columns of a continuous draw into ordered category codes 1..K,
# cutting each column at its interior thresholds (a length-p list). findInterval
# returns the count of thresholds at or below each value (0..K-1); +1 shifts to
# 1-based codes.
.ordinalize <- function(dat, thresholds) {
  out <- matrix(NA_integer_, nrow(dat), ncol(dat), dimnames = dimnames(dat))
  for (j in seq_len(ncol(dat))) {
    out[, j] <- findInterval(dat[, j], thresholds[[j]]) + 1L
  }
  out
}


# Resolve a column specification (variable names, or column indices) to a vector of
# column indices in 1..p, shared by `missing_vars` and `missing_predictor` so both report
# the same failures under their own argument name. The range is tested on the supplied
# values rather than on as.integer() of them: coercing first turns an index beyond
# .Machine$integer.max into NA, which would leave the range test an `if (NA)` and abort
# with an unclassed base error. A dimensioned or fractional spec is rejected rather than
# flattened column-major or rounded silently (mirrors the guard in .moment_vec()).
.resolve_column_spec <- function(spec, p, var_names, arg) {
  if (is.character(spec)) {
    idx <- match(spec, var_names)
    if (anyNA(idx)) {
      bad <- spec[is.na(idx)]
      cli::cli_abort(
        c("{.arg {arg}} names a variable not in the data.",
          "x" = "{cli::qty(bad)}Unknown name{?s}: {.val {bad}}."),
        class = "efa_simulate_input"
      )
    }
    return(idx)
  }
  if (!is.numeric(spec) || !is.null(dim(spec)) || !all(is.finite(spec)) ||
      any(spec != round(spec))) {
    cli::cli_abort(
      "{.arg {arg}} must be whole-number column indices or variable names.",
      class = "efa_simulate_input"
    )
  }
  out_of_range <- spec < 1 | spec > p
  if (any(out_of_range)) {
    # Formatted to character before interpolation, and given an explicit `cli::qty()`:
    # a numeric vector supplies no usable quantity for the plural marker, so leaving it
    # numeric aborts inside cli once more than one value is out of range.
    bad <- as.character(unique(spec[out_of_range]))
    cli::cli_abort(
      c("{.arg {arg}} indexes a column outside 1 to {p}.",
        "x" = "{cli::qty(bad)}It has value{?s} {.val {bad}}."),
      class = "efa_simulate_input"
    )
  }
  as.integer(spec)
}


# Resolve the `missing_vars` specification to the vector of column indices to hole.
# NULL means every column (the default: the mechanism applies to the whole data set). A
# supplied spec must reference existing, distinct columns; the order is kept, as
# `missing_predictor` is aligned to it element by element. Restricting the holed columns
# is what makes a strictly ignorable MAR design possible: a predictor outside
# `missing_vars` is fully observed, so the missingness depends only on data the analyst
# sees.
.resolve_missing_vars <- function(spec, p, var_names) {
  if (is.null(spec)) return(seq_len(p))
  idx <- .resolve_column_spec(spec, p, var_names, "missing_vars")
  if (length(idx) == 0L || anyDuplicated(idx) > 0L) {
    cli::cli_abort(
      c("{.arg missing_vars} must name at least one variable, each at most once.",
        "i" = "Leave it {.code NULL} to hole every variable."),
      class = "efa_simulate_input"
    )
  }
  idx
}


# Resolve the MAR predictor specification to one column index per holed variable (the
# columns in `vars`, in that order): which other variable's value drives that variable's
# missingness. NULL defaults to the next variable cyclically (variable p is driven by
# variable 1), so every variable has a distinct predictor. A supplied spec (one index or
# name per holed variable) must reference existing columns other than the variable itself
# -- a variable predicting its own missingness would be MNAR, not MAR (which also rules
# out a single recycled predictor, as it would predict itself). The cyclic default is
# order-dependent, so an explicit spec is preferable when the column order is arbitrary;
# it is also the way to keep the predictors outside `vars`, which is what makes the
# mechanism ignorably MAR.
.resolve_missing_predictor <- function(spec, p, var_names, vars = seq_len(p)) {
  if (p < 2L) {
    cli::cli_abort(
      c("{.code missing = \"MAR\"} needs at least two variables.",
        "x" = "With a single variable there is no other predictor for the missingness.",
        "i" = "Use {.code missing = \"MNAR\"} or {.val MCAR} for a single variable."),
      class = "efa_simulate_input"
    )
  }
  # Cyclic next variable: 2, 3, ..., p, 1 (never a variable's own position), taken for
  # the holed columns only.
  if (is.null(spec)) return(c(seq_len(p)[-1L], 1L)[vars])

  idx <- .resolve_column_spec(spec, p, var_names, "missing_predictor")

  if (length(idx) != length(vars)) {
    cli::cli_abort(
      c("{.arg missing_predictor} must give one predictor per variable it applies to ({length(vars)}).",
        "x" = "It has length {length(idx)}.",
        "i" = "A single predictor for every variable is not allowed, as that variable would predict its own missingness (MNAR)."),
      class = "efa_simulate_input"
    )
  }
  if (any(idx == vars)) {
    self <- as.character(vars[idx == vars])
    cli::cli_abort(
      c("{.arg missing_predictor} must reference another variable, not the variable itself.",
        "x" = "{cli::qty(self)}Variable{?s} {self} {?is/are} set to predict {?its/their} own missingness, which is MNAR rather than MAR.",
        "i" = "Use {.code missing = \"MNAR\"} for self-dependent missingness."),
      class = "efa_simulate_input"
    )
  }
  idx
}


# Solve for the logistic intercept a so the mean missing probability
# mean(plogis(a + b * z)) equals the target `prop`. The mean is strictly increasing in a
# from 0 to 1, so a unique root exists; the bracket qlogis(prop) +- (|b| max|z| + 10)
# forces a + b*z <= qlogis(prop) - 10 at the low end (mean below prop) and
# >= qlogis(prop) + 10 at the high end (mean above prop), so the ends are opposite-signed
# for any prop in (0, 1) and any finite b. uniroot draws no random numbers.
.calibrate_missing_intercept <- function(z, b, prop) {
  f <- function(a) mean(stats::plogis(a + b * z)) - prop
  # max|z| lives at an extremum, so range() avoids materializing a length-n abs(z).
  reach <- abs(b) * max(abs(range(z))) + 10
  centre <- stats::qlogis(prop)
  stats::uniroot(f, interval = c(centre - reach, centre + reach),
                 tol = .Machine$double.eps^0.5)$root
}


# Introduce missing values into a drawn data matrix under a chosen mechanism (Rubin,
# 1976). Each variable in `vars` (every column by default) is holed at (in expectation)
# `prop` of its cases; columns outside `vars` stay complete. MCAR draws a Bernoulli(prop)
# mask independent of the data. MAR and MNAR set each case's missing probability by a
# logistic model, logit(p_i) = a + b * z_i, where z standardizes a predictor: another
# variable (MAR, given by `pred_idx`, one entry per element of `vars`) or the variable's
# own value (MNAR). The intercept a is calibrated per variable so the mean missing
# probability is `prop`; the slope b (`strength`) sets the dependence. Predictors are read
# from a single pre-loop copy of the complete draw so an earlier column's holes cannot
# corrupt a later column's predictor (R copy-on-modify leaves `complete` untouched once
# `dat` is written). A constant predictor (sd 0) contributes z = 0, degrading that
# variable to MCAR. Exactly one runif(n) is drawn per holed column in fixed order, never
# short-circuited, so the RNG stream -- and hence the reproducibility across worker
# counts -- does not depend on the mechanism or the achieved probabilities.
# Rubin, D. B. (1976). Inference and missing data. Biometrika, 63(3), 581-592.
.apply_missing <- function(dat, mechanism, prop, strength, pred_idx,
                           vars = seq_len(ncol(dat))) {
  n <- nrow(dat)
  complete <- dat
  for (i in seq_along(vars)) {
    j <- vars[i]
    prob <- if (mechanism == "MCAR") {
      prop                                 # recycled against runif(n) below
    } else {
      xk <- complete[, if (mechanism == "MAR") pred_idx[i] else j]
      s <- stats::sd(xk)
      z <- if (s > 0) (xk - mean(xk)) / s else rep(0, n)
      a <- .calibrate_missing_intercept(z, strength, prop)
      stats::plogis(a + strength * z)
    }
    dat[stats::runif(n) < prob, j] <- NA
  }
  dat
}


# Perturb a population correlation matrix so a `q`-factor model fits it with a prescribed
# discrepancy, injecting "model error" (all models are approximations, so an exact-fit
# population overstates recovery; MacCallum, 2003). Recovers the model's unrotated orthogonal
# loadings from the common part (`common` = Lambda Phi Lambda', `Sigma_cov` the assembled
# covariance) -- the standardized common matrix `C_std` has the communalities on its diagonal,
# and its top-`q` eigenpairs give loadings `L` with `L L' + diag(psi) == R_pop` -- then hands
# them to the chosen method. Returns the perturbed correlation and the achieved fit -- the misfit
# of the specified q-factor (generating) model to the perturbed population, measured with the
# shared population-limit wrapper (.efa_population_fit) using EFA()'s fit-index formulas. For CB
# the generating model is the minimizer, so this equals the target; for TKL/WB it is the
# generating model's misfit (as in fungible/noisemaker), which a re-fitted EFA of the data may
# recover as somewhat smaller since it re-optimizes the loadings.
.apply_model_error <- function(R_pop, common, Sigma_cov, q, method,
                               target_rmsea, target_cfi, force_pd, tol = 1e-8) {
  p <- nrow(R_pop)
  h2 <- diag(common) / diag(Sigma_cov)     # standardized communalities
  u <- 1 - h2                              # uniquenesses
  C_std <- R_pop
  diag(C_std) <- h2
  ev <- eigen(C_std, symmetric = TRUE)
  L <- ev$vectors[, seq_len(q), drop = FALSE] %*%
    diag(sqrt(pmax(ev$values[seq_len(q)], 0)), q)
  psi <- 1 - rowSums(L^2)

  # The recovered L reproduces R_pop only when the population is exactly a rank-q common factor
  # structure with diagonal, non-negative uniquenesses. A correlated-residual (off-diagonal) Psi
  # or a communality above 1 breaks that: the eigen-recovered model would misrepresent the
  # population, so CB would centre on the wrong model (missing its exact-target guarantee), TKL
  # would take the square root of a negative uniqueness, and the reported achieved fit would be
  # measured against the wrong model. Reject such populations rather than silently mislead.
  if (any(psi < -tol) ||
      max(abs(tcrossprod(L) + diag(psi, p) - R_pop)) > 1e-8) {
    cli::cli_abort(
      c("Model error requires a population that is exactly a {q}-factor structure with non-negative uniquenesses.",
        "x" = "The population has correlated residuals (an off-diagonal {.arg Psi}) or a communality above 1, which the {q}-factor model cannot represent.",
        "i" = "Use a factor model with a diagonal {.arg Psi} and communalities below 1, or omit the model-error target."),
      class = "efa_simulate_model_error"
    )
  }

  me <- switch(
    method,
    CB  = .model_error_cb(R_pop, L, q, target_rmsea, force_pd),
    TKL = .model_error_tkl(R_pop, u, q, target_rmsea, target_cfi),
    WB  = .model_error_wb(R_pop, target_rmsea)
  )

  # Achieved fit of the major q-factor model on the perturbed population (see the header
  # above for what each method's value means).
  fit <- .efa_population_fit(L, me$population)
  model_error <- c(
    list(method = method, target_rmsea = target_rmsea, target_cfi = target_cfi,
         rmsea = fit$rmsea, cfi = fit$cfi, df = fit$df),
    me$params
  )
  list(population = me$population, model_error = model_error)
}


# Cudeck & Browne (1992) model error: construct Sigma* = Omega + kappa E such that the q-factor
# model (unrotated loadings `L`; its uniquenesses follow from `R_pop`) remains the minimizer of the
# ML discrepancy and the minimum equals delta = df * target_rmsea^2, so the achieved RMSEA equals
# the target.
# Follows MBESS::Sigma.2.SigmaStar at the correlation scale: differentiate the model-implied
# Sigma(theta) = Lt Lt' + diag(psi) with respect to theta = (vec(L), psi), form the ML-metric
# Jacobian columns B[,i] = -d * vech(W^-1 dSigma_i W^-1) (W = Omega, d the vech weights
# that make an ordinary residual orthogonal in the ML metric), and take E from the part of a
# random symmetric matrix orthogonal to B. G = W^-1 E has the same eigenvalues s as the symmetric
# W^{-1/2} E W^{-1/2}, so the discrepancy is kappa*sum(s) - sum(log(1 + kappa*s)); it is monotone
# in kappa up to the positive-definite boundary 1 + kappa*min(s) > 0. kappa is the exact root
# (uniroot on the monotone equation, unlike MBESS's nlm on |.|, which is less precise near the
# kink and would miss the target here). A direction that cannot reach delta while staying positive
# definite is redrawn; if none succeeds, force_pd = TRUE returns the closest positive-definite
# approximation (a smaller RMSEA, with a warning), otherwise the call aborts.
# Cudeck, R., & Browne, M. W. (1992). Constructing a covariance matrix that yields a specified
# minimizer and a specified minimum discrepancy function value. Psychometrika, 57(3), 357-369.
.model_error_cb <- function(R_pop, L, q, target_rmsea, force_pd,
                            max_tries = 50L) {
  p <- nrow(R_pop)
  df <- .efa_df(p, q)
  delta <- target_rmsea^2 * df
  pstar <- p * (p + 1) / 2

  W <- R_pop
  Wi <- solve(W)
  # vech inner-product weights (1 on the diagonal, 2 off it): with these, an ordinary residual
  # against B is orthogonal in the ML metric, which is what keeps theta the minimizer.
  wmat <- matrix(2, p, p)
  diag(wmat) <- 1
  dvec <- .vech(wmat)

  # The Jacobian columns in closed form. Elementwise Sigma_ij(theta) = sum_k L_ik L_jk +
  # [i == j] psi_i, so
  #   dSigma / dL[a, b] = e_a t(L[, b]) + L[, b] t(e_a)     -> column (b - 1) * p + a
  #   dSigma / dpsi[a]  = e_a t(e_a)                        -> column p * q + a,
  # ordered to match theta = (vec(L), psi), which R fills column-major. Each ML sandwich is then
  # a sum of outer products,
  #   Wi (e_a t(y) + y t(e_a)) Wi = Wi[, a] t(t(Wi) y) + (Wi y) t(Wi[a, ]),
  # whose vech picks x[ii] * y[jj] over the lower-triangle index pairs, so no p x p product is
  # formed per column. Wi[, a] and Wi[a, ] are kept apart because solve() need not return an
  # exactly symmetric inverse.
  #
  # Differentiating in closed form rather than by a finite difference is what keeps the draw
  # below reproducible: differencing divides a cancelling subtraction by the step size, which
  # lifts the q(q - 1)/2 rotational null directions of the model off zero to roughly eps/step.
  # Their orientation is then set purely by rounding, qr() counts them as real columns, and
  # qr.resid() projects them out -- so E, and with it the perturbed population, would swing by
  # tens of percent between BLAS implementations for one and the same seed.
  idx <- which(!upper.tri(matrix(0, p, p)), arr.ind = TRUE)   # the vech pairs, in .vech order
  ii <- idx[, 1L]
  jj <- idx[, 2L]
  WL <- Wi %*% L            # columns: Wi %*% L[, b]
  WtL <- crossprod(Wi, L)   # columns: t(Wi) %*% L[, b]
  npar <- p * q + p
  B <- matrix(0, pstar, npar)
  for (a in seq_len(p)) {
    wi_i <- Wi[ii, a]       # (Wi e_a)[ii]
    wi_j <- Wi[a, jj]       # (t(e_a) Wi)[jj]
    for (b in seq_len(q)) {
      B[, (b - 1L) * p + a] <- -dvec * (wi_i * WtL[jj, b] + WL[ii, b] * wi_j)
    }
    B[, p * q + a] <- -dvec * (wi_i * wi_j)
  }
  Bqr <- qr(B)

  # symmetric whitening W^{-1/2} for the eigenvalues s that drive both the discrepancy and the
  # positive-definite bound on kappa.
  Weig <- eigen(W, symmetric = TRUE)
  Wih <- Weig$vectors %*% diag(1 / sqrt(Weig$values), p) %*% t(Weig$vectors)

  best_pop <- NULL       # closest positive-definite approximation, for the force_pd escape
  best_gap <- -Inf       # its (negative) shortfall in discrepancy relative to delta
  for (tr in seq_len(max_tries)) {
    e <- qr.resid(Bqr, stats::rnorm(pstar))
    E <- matrix(0, p, p)
    E[lower.tri(E, diag = TRUE)] <- e
    E <- E + t(E) - diag(diag(E))
    s <- eigen(Wih %*% E %*% Wih, symmetric = TRUE, only.values = TRUE)$values
    trG <- sum(s)
    fdisc <- function(k) k * trG - sum(log1p(k * s)) - delta
    hi <- if (min(s) < 0) -1 / min(s) * (1 - 1e-9) else 1e6
    gap <- fdisc(hi)                          # discrepancy reachable at the PD boundary, minus delta
    if (gap >= 0) {
      kappa <- stats::uniroot(fdisc, c(1e-12, hi),
                              tol = .Machine$double.eps^0.5)$root
      return(list(population = stats::cov2cor(R_pop + kappa * E),
                  params = list(kappa = kappa)))
    }
    if (gap > best_gap) {                      # track the direction that comes closest to delta
      best_gap <- gap
      best_pop <- stats::cov2cor(R_pop + hi * E)
    }
  }

  # No direction reached delta while staying positive definite: the target is too large for this
  # population. Either accept the closest positive-definite approximation or abort.
  if (isTRUE(force_pd) && !is.null(best_pop)) {
    cli::cli_warn(
      c("The requested {.arg target_rmsea} could not be reached while keeping the population positive definite; the closest positive-definite approximation was used.",
        "i" = "The achieved RMSEA is below the target; lower {.arg target_rmsea} or use {.code model_error = \"TKL\"} for a larger misfit."),
      class = "efa_simulate_pd_forced"
    )
    return(list(population = best_pop, params = list(kappa = NA_real_)))
  }
  cli::cli_abort(
    c("The Cudeck-Browne model error cannot reach {.arg target_rmsea} = {target_rmsea} while keeping the population positive definite.",
      "i" = "Lower {.arg target_rmsea}, use {.code model_error = \"TKL\"} or {.val WB}, or set {.code force_pd = TRUE} to accept the closest positive-definite approximation."),
    class = "efa_simulate_model_error"
  )
}


# Tucker, Koopman & Linn (1969) model error: add a set of `n_minor` minor common factors to the
# population so the q-factor major model misfits it by a target RMSEA (and, optionally, CFI). The
# minor loadings `W0` are standard normal; a geometric column decay (1 - eps)^(0:(n_minor-1))
# concentrates variance in the first minor factors, and each row is rescaled so the minor common
# variance of variable i equals v * u_i (v is the proportion of its uniqueness u_i reallocated to
# minor factors). The perturbed population is R_pop + W W' (diagonal reset to 1). The two knobs
# (v, eps) are chosen by bounded optimization (L-BFGS-B with random restarts) to minimise the
# weighted relative squared deviation from the target(s); with a single target the fit is near
# exact, with both targets it is a compromise. Ft (major-model discrepancy) and Fb (independence
# baseline) use the same ML discrepancy as .gof()/.efa_population_fit().
# Tucker, L. R., Koopman, R. F., & Linn, R. L. (1969). Evaluation of factor analytic research
# procedures by means of simulated correlation matrices. Psychometrika, 34(4), 421-459.
.model_error_tkl <- function(R_pop, u, q, target_rmsea, target_cfi,
                             n_minor = 50L, max_tries = 25L) {
  p <- nrow(R_pop)
  df <- .efa_df(p, q)
  W0 <- matrix(stats::rnorm(p * n_minor), p, n_minor)
  Rinv <- solve(R_pop)
  ldRpop <- as.numeric(determinant(R_pop, logarithm = TRUE)$modulus)
  has_r <- !is.null(target_rmsea)
  has_c <- !is.null(target_cfi)

  obj <- function(par, return_vals = FALSE) {
    v <- par[1L]
    eps <- par[2L]
    # Column j of the minor loadings decays by (1 - eps)^(j-1); scale columns directly rather
    # than multiplying by a dense n_minor x n_minor diagonal on every objective evaluation.
    Wc <- sweep(W0, 2L, (1 - eps)^(0:(n_minor - 1L)), "*")
    wsq <- rowSums(Wc^2)
    Wc <- sqrt(v * u / wsq) * Wc          # scale row i so rowSums(Wc^2) = v * u_i
    RpopME <- R_pop + tcrossprod(Wc)
    diag(RpopME) <- 1
    ldME <- as.numeric(determinant(RpopME, logarithm = TRUE)$modulus)
    Ft <- ldRpop - ldME + sum(Rinv * RpopME) - p     # ML discrepancy of the major model
    Fb <- -ldME                                      # independence baseline
    rmsea <- sqrt(max(0, Ft) / df)
    cfi <- 1 - Ft / Fb
    if (return_vals) {
      return(list(population = RpopME, v = v, eps = eps, rmsea = rmsea, cfi = cfi))
    }
    val <- 0
    if (has_r) val <- val + (rmsea - target_rmsea)^2 / target_rmsea^2
    if (has_c) val <- val + ((1 - cfi) - (1 - target_cfi))^2 / (1 - target_cfi)^2
    if (!is.finite(val)) 1e6 else val
  }

  best_par <- NULL
  best_val <- Inf
  for (tr in seq_len(max_tries)) {
    start <- c(stats::runif(1, 0.02, 0.9), stats::runif(1, 0, 0.8))
    opt <- tryCatch(
      stats::optim(start, obj, method = "L-BFGS-B",
                   lower = c(0.001, 0), upper = c(1, 1)),
      error = function(e) NULL)
    if (is.null(opt)) next
    if (opt$value < best_val) {
      best_val <- opt$value
      best_par <- opt$par
    }
    if (opt$convergence == 0 && opt$value < 1e-8) break   # target(s) essentially hit
  }
  if (is.null(best_par)) {
    cli::cli_abort(
      c("The Tucker-Koopman-Linn optimization failed to find minor-factor weights for the requested target(s).",
        "i" = "Try a different {.arg target_rmsea}/{.arg target_cfi}, or {.code model_error = \"CB\"}."),
      class = "efa_simulate_model_error")
  }
  res <- obj(best_par, return_vals = TRUE)
  # `best_val` is the sum of the squared relative deviations from the requested target(s),
  # so its square root is the relative miss. Two targets over two knobs need not both be
  # reachable -- a high target CFI dominates the weighted objective and can halve the
  # achieved RMSEA -- and the compromise is otherwise only visible by reading the achieved
  # values back off the result. Flag a miss of more than 20% of the target: an ordinary
  # two-target compromise (say .05/.95, which lands within a tenth of both) is close
  # enough to be the intended answer, while a jointly infeasible pair misses by 50% or
  # more, so the threshold sits between the two rather than at the edge of the former.
  if (sqrt(best_val) > 0.2) {
    # Name the target(s) actually requested, so a CFI-only run is not told about an RMSEA
    # target it never set; both achieved values are reported either way.
    requested <- c(if (has_r) "RMSEA" else NULL, if (has_c) "CFI" else NULL)
    targets <- c(if (has_r) target_rmsea else NULL, if (has_c) target_cfi else NULL)
    cli::cli_warn(
      c("The Tucker-Koopman-Linn minor factors could not reach the requested {requested} target{?s} of {.val {targets}}.",
        "x" = "The closest compromise achieves an RMSEA of {.val {round(res$rmsea, 4)}} and a CFI of {.val {round(res$cfi, 4)}}.",
        "i" = "The targets are not jointly attainable by the two minor-factor knobs for this population; relax one of them, supply a single target, or use {.code model_error = \"CB\"} to match the RMSEA exactly."),
      class = "efa_simulate_model_error_compromise")
  }
  list(population = res$population, params = list(v = res$v, eps = res$eps))
}


# Wu & Browne (2015) model error: draw the population from an inverse-Wishart distribution
# centered on the model-implied correlation Omega, Sigma ~ IW(m, m Omega), where the precision
# m = 1 / target_rmsea^2 sets the expected RMSEA to approximately the target (via the Wu-Browne
# small-error relation v ~= epsilon^2; the realized RMSEA is biased slightly upward, more so for
# larger targets). The drawn covariance is
# standardized to a correlation matrix; the achieved RMSEA is random around the target (this is a
# single random draw, not an exact construction), so the realized value is reported. Requires
# m >= p for a proper draw (a very large target RMSEA is rejected).
# Wu, H., & Browne, M. W. (2015). Quantifying adventitious error in a covariance structure as a
# random effect. Psychometrika, 80(3), 571-600.
.model_error_wb <- function(R_pop, target_rmsea) {
  p <- nrow(R_pop)
  m <- 1 / target_rmsea^2
  if (m < p) {
    cli::cli_abort(
      c("{.arg target_rmsea} = {target_rmsea} is too large for the Wu-Browne method with {p} variable{?s}.",
        "i" = "The precision m = 1 / target_rmsea^2 = {round(m, 1)} must be at least the number of variables ({p}); use a smaller {.arg target_rmsea}."),
      class = "efa_simulate_model_error")
  }
  Sigma <- .riwish(m, m * R_pop)
  list(population = stats::cov2cor(Sigma), params = list(m = m))
}


# One inverse-Wishart draw Sigma ~ IW(nu, S) via the Bartlett decomposition (no external
# dependency): S^-1 = C C', and A A' ~ Wishart(nu, S^-1) with A = C T, T lower-triangular whose
# diagonal is sqrt(chi-square(nu - i + 1)) and whose strictly-lower entries are standard normal;
# Sigma = (A A')^-1. Draws its randomness (rchisq, rnorm) from the caller's stream, so it is
# reproducible under a fixed seed.
.riwish <- function(nu, S) {
  p <- nrow(S)
  C <- t(chol(solve(S)))
  A <- matrix(0, p, p)
  diag(A) <- sqrt(stats::rchisq(p, df = nu - seq_len(p) + 1))
  A[lower.tri(A)] <- stats::rnorm(p * (p - 1) / 2)
  CA <- C %*% A
  Sig <- solve(tcrossprod(CA))
  (Sig + t(Sig)) / 2
}

# A three-factor population built from a shipped loading pattern and moderate
# factor intercorrelations; reused across the recovery and reproducibility tests.
Lambda_pop <- population_models$loadings$baseline
Phi_pop <- population_models$phis_3$moderate
R_pop <- efa_simulate(Lambda = Lambda_pop, Phi = Phi_pop, return_pop = TRUE)$population

# The same-seed reproducibility assertions below compare two separate invocations, so they use
# expect_equal() with an explicit tolerance rather than expect_identical(): the draw's core is
# a Cholesky (or eigen) factor times a normal deviate matrix, and a threaded BLAS (Apple's
# Accelerate, for one) is free to vary its GEMM reduction order between calls, so one function
# on one input can differ in the last ulp. waldo still compares S3 classes, names, attributes
# and structure exactly -- including where the NAs sit -- so only numeric values are given
# slack; a seed that was ignored draws wholly different data and still fails loudly. A set
# tolerance does relax integer against double, so the ordinal paths, whose category codes are
# integer by contract, assert their storage mode separately.
fp_tol <- 1e-8

test_that("return_pop yields the population correlation matrix", {
  expect_true(is.matrix(R_pop))
  expect_equal(dim(R_pop), c(nrow(Lambda_pop), nrow(Lambda_pop)))
  expect_equal(diag(R_pop), rep(1, nrow(Lambda_pop)), ignore_attr = TRUE)
  expect_true(isSymmetric(R_pop))
  expect_equal(rownames(R_pop), rownames(Lambda_pop))
  # Assembled from Lambda * Phi * Lambda' + Psi and standardized: the within-factor
  # entries are .6 * 1 * .6 = .36, the between-factor entries .6 * .3 * .6 = .108.
  expect_equal(R_pop[1, 2], 0.36)
  expect_equal(R_pop[1, 7], 0.108)
  # Building from Lambda/Phi and passing the resulting matrix as R agree.
  expect_equal(efa_simulate(R = R_pop, return_pop = TRUE)$population, R_pop)
})

test_that("simulated data recover the population correlation matrix", {
  # At a large N the sample correlations of normal data converge to the population
  # correlation; check both input paths (model components and a supplied R).
  dat_model <- efa_simulate(N = 1e5, Lambda = Lambda_pop, Phi = Phi_pop, seed = 42)$data
  dat_R <- efa_simulate(N = 1e5, R = R_pop, seed = 42)$data

  expect_true(is.matrix(dat_model))
  expect_false(is.data.frame(dat_model))
  expect_null(attr(dat_model, "class"))
  expect_equal(dim(dat_model), c(1e5, nrow(Lambda_pop)))
  expect_equal(colnames(dat_model), rownames(Lambda_pop))

  expect_lt(max(abs(stats::cor(dat_model) - R_pop)), 0.03)
  expect_lt(max(abs(stats::cor(dat_R) - R_pop)), 0.03)
})

test_that("n_datasets > 1 returns a list of datasets", {
  sims <- efa_simulate(N = 200, R = R_pop, n_datasets = 3, seed = 1)$data
  expect_type(sims, "list")
  expect_length(sims, 3)
  expect_true(all(vapply(sims, is.matrix, logical(1))))
  expect_equal(dim(sims[[1]]), c(200, nrow(R_pop)))
  # future.seed = TRUE binds each replicate to its own stream, so the datasets differ.
  expect_false(isTRUE(all.equal(sims[[1]], sims[[2]])))
})

test_that("a positive-semidefinite but singular population uses the eigen fallback", {
  # Two perfectly correlated variables: the correlation matrix is singular (a zero
  # eigenvalue), so the Cholesky fails and the symmetric eigen square root is used.
  R_sing <- matrix(1, 2, 2)
  dat <- efa_simulate(N = 500, R = R_sing, seed = 7)$data
  expect_equal(dim(dat), c(500, 2))
  expect_equal(stats::cor(dat)[1, 2], 1, tolerance = 1e-8)
})

test_that("a fixed seed is reproducible and leaves the RNG stream unchanged", {
  a <- efa_simulate(N = 100, R = R_pop, seed = 123)
  b <- efa_simulate(N = 100, R = R_pop, seed = 123)
  expect_equal(a, b, tolerance = fp_tol)

  # A supplied seed must not have a lasting side effect on the caller's stream.
  set.seed(1)
  state_before <- get(".Random.seed", envir = globalenv(), inherits = FALSE)
  invisible(efa_simulate(N = 100, R = R_pop, seed = 99))
  expect_identical(get(".Random.seed", envir = globalenv(), inherits = FALSE),
                   state_before)

  # When no stream existed yet, the one set.seed() creates is removed again.
  if (exists(".Random.seed", envir = globalenv(), inherits = FALSE)) {
    rm(".Random.seed", envir = globalenv())
  }
  invisible(efa_simulate(N = 100, R = R_pop, seed = 5))
  expect_false(exists(".Random.seed", envir = globalenv(), inherits = FALSE))
})

test_that("a fixed seed is reproducible at 1 vs 2 workers", {
  skip_on_cran()
  skip_if_not_slow()
  # Each dataset is drawn in its own future, and future.seed = TRUE binds every
  # replicate's RNG stream to its index. With a fixed `seed` the draws must
  # therefore be identical regardless of the number of workers. The multisession
  # workers are fresh R processes that load the installed package, so run this
  # under devtools::check() / after devtools::install() for the worker code to
  # match the main process. (multicore is unavailable on Windows.)
  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  run <- function() efa_simulate(N = 200, R = R_pop, n_datasets = 4, seed = 2024)

  future::plan(future::sequential)
  one <- run()

  future::plan(future::multisession, workers = 2)
  two <- run()

  expect_equal(one, two, tolerance = 1e-10)
})

test_that("invalid population specifications raise a classed error", {
  # Neither Lambda nor R.
  expect_error(efa_simulate(N = 10), class = "efa_simulate_input")
  # Both Lambda and R.
  expect_error(efa_simulate(N = 10, Lambda = Lambda_pop, R = R_pop),
               class = "efa_simulate_input")
  # Phi/Psi supplied alongside a ready R.
  expect_error(efa_simulate(N = 10, R = R_pop, Phi = Phi_pop),
               class = "efa_simulate_input")
  # A non-symmetric R.
  R_asym <- R_pop
  R_asym[1, 2] <- R_asym[1, 2] + 0.1
  expect_error(efa_simulate(N = 10, R = R_asym), class = "efa_simulate_input")
})

test_that("a covariance matrix supplied as R is rejected", {
  # Ordinal thresholds, the Vale-Maurelli cubic, and the polychoric match all assume
  # unit variances, so a non-unit diagonal would silently produce a population other
  # than the one those options describe.
  sdv <- c(2, rep(1, nrow(R_pop) - 1))
  S_cov <- diag(sdv) %*% R_pop %*% diag(sdv)
  dimnames(S_cov) <- dimnames(R_pop)
  expect_error(efa_simulate(N = 10, R = S_cov), class = "efa_simulate_input")
  expect_error(efa_simulate(R = S_cov, return_pop = TRUE), class = "efa_simulate_input")
  # The same matrix standardized is accepted, so it is the scale that is refused and
  # not the correlation structure.
  expect_s3_class(efa_simulate(N = 10, R = stats::cov2cor(S_cov), seed = 1),
                  "efa_simulated")
})

test_that("an indefinite population is rejected", {
  # A symmetric matrix with a negative eigenvalue is not a valid covariance.
  R_indef <- matrix(c(1, 1, 0,
                      1, 1, 1,
                      0, 1, 1), 3, 3)
  expect_lt(min(eigen(R_indef, symmetric = TRUE, only.values = TRUE)$values), 0)
  expect_error(efa_simulate(N = 10, R = R_indef), class = "efa_simulate_not_pd")
})

test_that("a Heywood case in the factor model is rejected", {
  # A single-factor loading above 1 implies a communality above 1, leaving no
  # unique variance to simulate from.
  Lambda_hey <- matrix(1.2, 3, 1)
  expect_error(efa_simulate(N = 10, Lambda = Lambda_hey),
               class = "efa_simulate_heywood")
})

test_that("non-finite or non-symmetric inputs raise a classed error", {
  # any.missing = FALSE rejects NA/NaN but not Inf, so finiteness is checked
  # explicitly on every population input.
  L_inf <- Lambda_pop
  L_inf[1, 1] <- Inf
  expect_error(efa_simulate(N = 10, Lambda = L_inf), class = "efa_simulate_input")

  R_inf <- R_pop
  R_inf[1, 1] <- Inf
  expect_error(efa_simulate(N = 10, R = R_inf), class = "efa_simulate_input")

  # A matrix-valued Psi must be finite and symmetric, like Phi and R.
  p <- nrow(Lambda_pop)
  Psi_asym <- diag(0.5, p)
  Psi_asym[1, 2] <- 0.3
  expect_error(efa_simulate(N = 10, Lambda = Lambda_pop, Psi = Psi_asym),
               class = "efa_simulate_input")

  Psi_inf <- diag(0.5, p)
  Psi_inf[1, 1] <- Inf
  expect_error(efa_simulate(N = 10, Lambda = Lambda_pop, Psi = Psi_inf),
               class = "efa_simulate_input")
})

test_that("a valid matrix-valued Psi matches the default unique variances", {
  # A diagonal Psi equal to the standardizing unique variances reproduces the
  # default population exactly.
  Psi <- diag(1 - diag(Lambda_pop %*% Phi_pop %*% t(Lambda_pop)))
  R_from_mat <- efa_simulate(Lambda = Lambda_pop, Phi = Phi_pop, Psi = Psi,
                             return_pop = TRUE)$population
  expect_equal(R_from_mat, R_pop)
})

test_that("return_pop rejects an indefinite population", {
  # The positive-semidefinite screen runs before the return_pop exit, so an
  # invalid population is never returned as 'the population correlation matrix'.
  R_indef <- matrix(c(1, 1, 0,
                      1, 1, 1,
                      0, 1, 1), 3, 3)
  expect_error(efa_simulate(R = R_indef, return_pop = TRUE),
               class = "efa_simulate_not_pd")
})

test_that("a communality of exactly 1 gives a drawable singular population", {
  # psi = 0 (communality exactly 1) is allowed: the population is positive
  # semi-definite but singular and drawn via the eigen fallback, not rejected.
  L1 <- matrix(1, 3, 1)
  R1 <- efa_simulate(Lambda = L1, return_pop = TRUE)$population
  expect_equal(R1, matrix(1, 3, 3), ignore_attr = TRUE)
  dat <- efa_simulate(N = 200, Lambda = L1, seed = 3)$data
  expect_equal(dim(dat), c(200, 3))
})

test_that("variable names fall back to the columns of R when rows are unnamed", {
  R2 <- matrix(c(1, 0.5, 0.5, 1), 2, 2)
  colnames(R2) <- c("item1", "item2")
  Rp <- efa_simulate(R = R2, return_pop = TRUE)$population
  expect_equal(colnames(Rp), c("item1", "item2"))
  dat <- efa_simulate(N = 50, R = R2, seed = 1)$data
  expect_equal(colnames(dat), c("item1", "item2"))
})

test_that("empirical marginals reproduce the target R and are preserved", {
  # Ruscio-Kaczetow reproduces the population correlation whatever the supplied
  # marginals are; check recovery at a large N and that each drawn column's
  # values come only from the matching (skewed) empirical source column.
  p <- nrow(Lambda_pop)
  set.seed(2024)
  x_skew <- matrix(stats::rexp(400 * p), nrow = 400, ncol = p)

  dat <- efa_simulate(N = 1e4, Lambda = Lambda_pop, Phi = Phi_pop,
                      marginals = "empirical", marginal_data = x_skew,
                      n_factors = 3, seed = 42)$data

  expect_true(is.matrix(dat))
  expect_equal(dim(dat), c(1e4, p))
  expect_equal(colnames(dat), rownames(Lambda_pop))
  expect_lt(max(abs(stats::cor(dat) - R_pop)), 0.05)
  # Each drawn column carries only values from its empirical marginal.
  expect_true(all(vapply(seq_len(p),
                         function(i) all(dat[, i] %in% x_skew[, i]),
                         logical(1))))
})

test_that("empirical n_factors defaults to the model's factor count", {
  p <- nrow(Lambda_pop)
  set.seed(3)
  x_emp <- matrix(stats::rnorm(200 * p), nrow = 200, ncol = p)

  # Built from Lambda: n_factors defaults to ncol(Lambda) and the call runs.
  dat <- efa_simulate(N = 500, Lambda = Lambda_pop, Phi = Phi_pop,
                      marginals = "empirical", marginal_data = x_emp, seed = 1)$data
  expect_equal(dim(dat), c(500, p))

  # Supplied via R: no factor count is implied, so n_factors is required.
  expect_error(
    efa_simulate(N = 500, R = R_pop, marginals = "empirical",
                 marginal_data = x_emp, seed = 1),
    class = "efa_simulate_input")
})

test_that("empirical-path inputs are validated", {
  p <- nrow(Lambda_pop)
  set.seed(5)
  x_emp <- matrix(stats::rnorm(100 * p), nrow = 100, ncol = p)

  # marginal_data is required for the empirical path.
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical"),
    class = "efa_simulate_input")
  # Wrong number of columns.
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_emp[, 1:5]),
    class = "efa_simulate_input")
  # n_factors must be smaller than the number of variables.
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_emp, n_factors = p),
    class = "efa_simulate_input")
  # Non-finite marginal_data.
  x_bad <- x_emp
  x_bad[1, 1] <- Inf
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_bad),
    class = "efa_simulate_input")
  # A single-row marginal source would resample a length-one vector (tripping
  # sample()'s length-1 behaviour), so it is rejected.
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_emp[1, , drop = FALSE]),
    class = "efa_simulate_input")
  # A constant (zero-variance) column cannot reproduce a correlation.
  x_const <- x_emp
  x_const[, 2] <- 4
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_const),
    class = "efa_simulate_input")
  # The reproduction needs at least two cases to draw.
  expect_error(
    efa_simulate(N = 1, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_emp),
    class = "efa_simulate_input")
  # marginal_data / n_factors only apply to the empirical path.
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, marginal_data = x_emp),
    class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 100, Lambda = Lambda_pop, Phi = Phi_pop, n_factors = 2),
    class = "efa_simulate_input")
})

test_that("a resample that comes out constant is reported with a classed error", {
  # The source column is non-normal but legal (two distinct values), so it passes the
  # boundary check; its rare category is rare enough that a small draw almost never
  # includes it, leaving a constant resample that has no correlation to match.
  p <- nrow(Lambda_pop)
  set.seed(9)
  x_rare <- matrix(stats::rnorm(2000 * p), nrow = 2000, ncol = p)
  x_rare[, 1] <- c(1, rep(0, 1999))
  # The draw runs in a future, which re-raises the error but also emits its own
  # cancellation warning; only the classed error is of interest here.
  expect_error(
    suppressWarnings(
      efa_simulate(N = 10, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                   marginal_data = x_rare, seed = 1)),
    class = "efa_simulate_degenerate_marginal")
})

test_that("empirical n_datasets > 1 returns a reproducible list", {
  p <- nrow(Lambda_pop)
  set.seed(7)
  x_emp <- matrix(stats::rexp(150 * p), nrow = 150, ncol = p)

  run <- function() {
    efa_simulate(N = 300, Lambda = Lambda_pop, Phi = Phi_pop, marginals = "empirical",
                 marginal_data = x_emp, n_factors = 3, n_datasets = 2, seed = 11)
  }
  sims1 <- run()
  sims2 <- run()

  expect_type(sims1$data, "list")
  expect_length(sims1$data, 2)
  expect_true(all(vapply(sims1$data, is.matrix, logical(1))))
  expect_equal(dim(sims1$data[[1]]), c(300, p))
  # A fixed seed is reproducible; each replicate has its own stream, so the two
  # datasets differ.
  expect_equal(sims1, sims2, tolerance = fp_tol)
  expect_false(isTRUE(all.equal(sims1$data[[1]], sims1$data[[2]])))
})

# Population and empirical-moment estimators (divisor n, matching the theoretical
# targets) shared by the Vale-Maurelli / independent-generator moment tests.
.pop_skew <- function(x) {
  x <- x - mean(x)
  mean(x^3) / mean(x^2)^1.5
}
.pop_exkurt <- function(x) {
  x <- x - mean(x)
  mean(x^4) / mean(x^2)^2 - 3
}
# A compact two-factor population used for moment recovery (six variables keeps
# the large-N draws affordable while the fourth-moment estimator settles).
Lambda_vm <- matrix(0, 6, 2)
Lambda_vm[1:3, 1] <- 0.7
Lambda_vm[4:6, 2] <- 0.7
Phi_vm <- matrix(c(1, 0.3, 0.3, 1), 2, 2)
R_vm <- efa_simulate(Lambda = Lambda_vm, Phi = Phi_vm, return_pop = TRUE)$population

test_that("Vale-Maurelli data carry the target moments and correlation", {
  skip_on_cran()
  sk <- 0.8
  ku <- 1.5
  dat <- efa_simulate(N = 2e5, Lambda = Lambda_vm, Phi = Phi_vm, marginals = "VM",
                      skewness = sk, kurtosis = ku, seed = 42)$data

  expect_true(is.matrix(dat))
  expect_equal(dim(dat), c(2e5, 6))
  # Fleishman marginals are centred with unit variance by construction.
  expect_lt(max(abs(colMeans(dat))), 0.02)
  expect_lt(max(abs(apply(dat, 2, stats::var) - 1)), 0.02)
  # Achieved skewness/(excess) kurtosis match the targets; the fourth moment is a
  # high-variance estimator, so it carries a looser tolerance than skewness.
  expect_lt(max(abs(apply(dat, 2, .pop_skew) - sk)), 0.05)
  expect_lt(max(abs(apply(dat, 2, .pop_exkurt) - ku)), 0.3)
  expect_lt(max(abs(stats::cor(dat) - R_vm)), 0.02)
})

test_that("independent-generator data carry the target moments and correlation", {
  skip_on_cran()
  sk <- 0.8
  ku <- 1.5
  dat <- efa_simulate(N = 2e5, R = R_vm, marginals = "IG",
                      skewness = sk, kurtosis = ku, seed = 42)$data

  expect_equal(dim(dat), c(2e5, 6))
  expect_lt(max(abs(colMeans(dat))), 0.02)
  expect_lt(max(abs(apply(dat, 2, stats::var) - 1)), 0.02)
  expect_lt(max(abs(apply(dat, 2, .pop_skew) - sk)), 0.05)
  expect_lt(max(abs(apply(dat, 2, .pop_exkurt) - ku)), 0.3)
  expect_lt(max(abs(stats::cor(dat) - R_vm)), 0.02)
})

test_that("per-variable skewness and kurtosis are honoured", {
  skip_on_cran()
  # The first three variables are near-normal, the last three markedly skewed.
  sk <- c(rep(0, 3), rep(1, 3))
  ku <- c(rep(0, 3), rep(2, 3))
  dat <- efa_simulate(N = 2e5, R = R_vm, marginals = "VM",
                      skewness = sk, kurtosis = ku, seed = 7)$data
  expect_lt(max(abs(apply(dat, 2, .pop_skew) - sk)), 0.05)
  expect_lt(max(abs(apply(dat, 2, .pop_exkurt) - ku)), 0.3)
})

test_that("unattainable target moments raise a classed error", {
  # Below the universal bound (excess kurtosis >= skewness^2 - 2): skew 3 needs
  # excess kurtosis of at least 7.
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "VM", skewness = 3, kurtosis = 0),
    class = "efa_simulate_infeasible_moments")
  # Inside the universal bound but outside Fleishman's tighter region (its lower
  # kurtosis boundary at skewness 0 is about -1.15): the coefficient solve fails.
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "VM", skewness = 0, kurtosis = -1.4),
    class = "efa_simulate_infeasible_moments")
  # Skewness alone: Fleishman's region leaves the zero-kurtosis line at skewness ~0.855,
  # so the most obvious "give me skewed data" call is infeasible via the defaulted
  # kurtosis rather than via the requested skewness.
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "VM", skewness = 1),
    class = "efa_simulate_infeasible_moments")
  # The independent-generator moments can be infeasible even when the requested
  # marginal moments are attainable, because they depend on the population.
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "IG", skewness = 3, kurtosis = 0),
    class = "efa_simulate_infeasible_moments")
  # A modest, attainable marginal request can still be infeasible via the
  # population: strong correlations blow up the implied generator moments past
  # Fleishman's region. The abort is reported for the request, not the internals.
  R_hi <- matrix(0.98, 4, 4)
  diag(R_hi) <- 1
  expect_error(
    efa_simulate(N = 10, R = R_hi, marginals = "IG", skewness = 1, kurtosis = 2),
    class = "efa_simulate_infeasible_moments")
})

test_that("a non-positive-definite Vale-Maurelli intermediate matrix is guarded", {
  # This population is (barely) positive definite, but the strong non-normality
  # inflates the required intermediate correlations enough to make the pairwise
  # intermediate matrix indefinite.
  R_bad <- matrix(c(1, 0.857, 0.918,
                    0.857, 1, 0.585,
                    0.918, 0.585, 1), 3, 3)
  expect_gt(min(eigen(R_bad, symmetric = TRUE, only.values = TRUE)$values), 0)
  expect_error(
    efa_simulate(N = 10, R = R_bad, marginals = "VM", skewness = 2.5, kurtosis = 10),
    class = "efa_simulate_intermediate_not_pd")
  # force_pd projects the intermediate matrix to the nearest correlation matrix
  # and warns, rather than aborting.
  expect_warning(
    dat <- efa_simulate(N = 500, R = R_bad, marginals = "VM", skewness = 2.5,
                        kurtosis = 10, force_pd = TRUE, seed = 1),
    class = "efa_simulate_pd_forced")
  expect_equal(dim(dat$data), c(500, 3))
  # The projected intermediate matrix no longer solves the Vale-Maurelli cubic for the
  # target, so `population` reports the population the draw actually attains -- which
  # differs from the target and keeps a unit diagonal and the variable names.
  expect_false(isTRUE(all.equal(dat$population, R_bad, check.attributes = FALSE)))
  expect_equal(diag(dat$population), rep(1, 3), ignore_attr = TRUE)
  expect_equal(dimnames(dat$population), list(paste0("V", 1:3), paste0("V", 1:3)))
})

test_that(".vm_achieved_cor inverts the intermediate-correlation cubic", {
  # Where the intermediate matrix is not projected, the achieved population must be the
  # target: the two helpers are the two directions of the same Vale-Maurelli equation.
  ftab <- .fleishman_table(rep(1, ncol(R_vm)), rep(2, ncol(R_vm)))
  expect_equal(.vm_achieved_cor(.vm_intermediate_cor(R_vm, ftab), ftab),
               R_vm, ignore_attr = TRUE, tolerance = 1e-10)
})

test_that("the forced-PD Vale-Maurelli population is the one the draw converges to", {
  skip_on_cran()
  skip_if_not_slow()
  # Strongly correlated variables with heavy, mixed-sign non-normality: the intermediate
  # matrix is far outside the positive-definite cone, so the projection moves the
  # population it implies a long way from the target (here by about 0.21).
  R_bad <- diag(5)
  R_bad[lower.tri(R_bad)] <- c(0.8606, -0.5258, 0.8610, 0.5717, -0.8367, 0.9447,
                               0.8524, -0.7903, -0.9634, 0.8094)
  R_bad <- R_bad + t(R_bad) - diag(5)
  sim <- suppressWarnings(
    efa_simulate(N = 5e4, R = R_bad, marginals = "VM",
                 skewness = c(2.20, 2.39, -1.91, -0.13, 0.30),
                 kurtosis = c(27.6, 8.5, 29.7, 28.7, 7.1),
                 force_pd = TRUE, seed = 3))
  emp <- stats::cor(sim$data)
  off <- function(x) x[lower.tri(x)]
  # The realized correlations converge on the reported population, not on the target the
  # projection abandoned -- the drift is a population property, not sampling noise.
  expect_lt(max(abs(off(emp) - off(sim$population))), 0.02)
  expect_gt(max(abs(off(sim$population) - off(R_bad))), 0.1)
})

test_that("the independent generator requires a positive-definite population", {
  # Two perfectly correlated variables give a singular population, which has no
  # Cholesky factor for the independent-generator mixture.
  R_sing <- matrix(1, 3, 3)
  expect_error(
    efa_simulate(N = 10, R = R_sing, marginals = "IG", skewness = 1, kurtosis = 2),
    class = "efa_simulate_input")
})

test_that("VM/IG moment arguments are validated and gated", {
  # skewness/kurtosis only apply to the VM/IG paths.
  expect_error(efa_simulate(N = 10, R = R_vm, skewness = 1), class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, R = R_vm, kurtosis = 1), class = "efa_simulate_input")
  # force_pd only applies to the Vale-Maurelli path, not normal or IG.
  expect_error(efa_simulate(N = 10, R = R_vm, force_pd = TRUE), class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "IG", skewness = 1, kurtosis = 2,
                 force_pd = TRUE),
    class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "empirical", marginal_data = R_vm,
                 skewness = 1),
    class = "efa_simulate_input")
  # VM/IG need at least one target moment.
  expect_error(efa_simulate(N = 10, R = R_vm, marginals = "VM"),
               class = "efa_simulate_input")
  # A moment vector must have length 1 or p, and must be a vector (not a matrix).
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "IG", skewness = c(1, 2)),
    class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "VM", skewness = matrix(0.5, 6, 1),
                 kurtosis = 1),
    class = "efa_simulate_input")
})

test_that("VM/IG draws are reproducible and support multiple datasets", {
  a <- efa_simulate(N = 200, R = R_vm, marginals = "VM", skewness = 1, kurtosis = 2,
                    seed = 11)
  b <- efa_simulate(N = 200, R = R_vm, marginals = "VM", skewness = 1, kurtosis = 2,
                    seed = 11)
  expect_equal(a, b, tolerance = fp_tol)
  expect_equal(colnames(a$data), colnames(R_vm))

  ig_a <- efa_simulate(N = 200, R = R_vm, marginals = "IG", skewness = 1, kurtosis = 2,
                       seed = 11)
  ig_b <- efa_simulate(N = 200, R = R_vm, marginals = "IG", skewness = 1, kurtosis = 2,
                       seed = 11)
  expect_equal(ig_a, ig_b, tolerance = fp_tol)

  sims <- efa_simulate(N = 200, R = R_vm, marginals = "VM", skewness = 1,
                       kurtosis = 2, n_datasets = 3, seed = 11)$data
  expect_type(sims, "list")
  expect_length(sims, 3)
  # Each replicate has its own stream, so the datasets differ.
  expect_false(isTRUE(all.equal(sims[[1]], sims[[2]])))
})

test_that("categories discretize the draw into ordered category codes", {
  # A five-category cut of normal marginals: the output is an integer matrix of
  # codes 1..5, every category is filled at a moderate N, and categorization
  # attenuates the product-moment correlation below the latent value.
  dat <- efa_simulate(N = 2000, R = R_vm, categories = 5, seed = 1)$data
  expect_true(is.matrix(dat))
  expect_type(dat, "integer")
  expect_equal(dim(dat), c(2000, ncol(R_vm)))
  expect_equal(colnames(dat), colnames(R_vm))
  expect_true(all(dat %in% 1:5))
  expect_true(all(vapply(seq_len(ncol(dat)),
                         function(j) length(unique(dat[, j])) == 5L, logical(1))))
  # Within-factor pair (both load on the first factor): the ordinal Pearson
  # correlation is smaller in magnitude than the latent .49.
  expect_lt(stats::cor(dat)[1, 2], R_vm[1, 2])
})

test_that("match = 'polychoric' round-trips through .polychoric() to the target R", {
  skip_on_cran()
  # Normal latents thresholded into ordered categories: the two-step polychoric
  # estimator is consistent for the latent correlation (Olsson, 1979), so at a
  # large N its estimate returns the target correlation. Both equally probable
  # categories and asymmetric category proportions must round-trip.
  dat_eq <- efa_simulate(N = 1e5, R = R_vm, categories = 5, match = "polychoric",
                         seed = 42)$data
  R_eq <- .polychoric(dat_eq)$R
  expect_lt(max(abs(R_eq - R_vm)), 0.02)

  props <- rep(list(c(0.1, 0.2, 0.4, 0.3)), ncol(R_vm))
  dat_sk <- efa_simulate(N = 1e5, R = R_vm, categories = props, match = "polychoric",
                         seed = 42)$data
  R_sk <- .polychoric(dat_sk)$R
  expect_lt(max(abs(R_sk - R_vm)), 0.02)
})

test_that("polychoric matching recovers the requested category proportions", {
  skip_on_cran()
  # With normal latents, cutting at qnorm(cumsum(p)) gives population category
  # proportions equal to `p`; at a large N the realized proportions match.
  props <- c(0.1, 0.2, 0.4, 0.3)
  dat <- efa_simulate(N = 1e5, R = R_vm, categories = rep(list(props), ncol(R_vm)),
                      match = "polychoric", seed = 7)$data
  realized <- vapply(seq_len(ncol(dat)),
                     function(j) tabulate(dat[, j], nbins = 4L) / nrow(dat),
                     numeric(4L))
  expect_lt(max(abs(realized - props)), 0.01)
})

test_that("Vale-Maurelli thresholds recover the requested category proportions", {
  skip_on_cran()
  # A VM draw is X = f(Z) with f the monotone Fleishman cubic, so P(X <= f(tau)) =
  # P(Z <= tau): cutting at the mapped thresholds f(tau) reproduces the proportions
  # tau was built for, while the raw normal-scale tau would cut the wrong quantiles.
  sk <- 1.5
  ku <- 4
  dat <- efa_simulate(N = 1e5, R = R_vm, marginals = "VM", skewness = sk, kurtosis = ku,
                      categories = 5, seed = 7)$data
  realized <- vapply(seq_len(ncol(dat)),
                     function(j) tabulate(dat[, j], nbins = 5L) / nrow(dat),
                     numeric(5L))
  expect_lt(max(abs(realized - 0.2)), 0.01)

  # Asymmetric requested proportions likewise.
  props <- c(0.1, 0.2, 0.4, 0.3)
  dat_p <- efa_simulate(N = 1e5, R = R_vm, marginals = "VM", skewness = sk, kurtosis = ku,
                        categories = rep(list(props), ncol(R_vm)), seed = 7)$data
  realized_p <- vapply(seq_len(ncol(dat_p)),
                       function(j) tabulate(dat_p[, j], nbins = 4L) / nrow(dat_p),
                       numeric(4L))
  expect_lt(max(abs(realized_p - props)), 0.01)

  # Platykurtic marginals (excess kurtosis below 0) are the case a global-monotonicity
  # test would wrongly reject: d < 0 there, yet the cubic increases across the
  # thresholds, so the proportions must come back just as exactly.
  dat_f <- efa_simulate(N = 1e5, R = R_vm, marginals = "VM", skewness = 0, kurtosis = -1,
                        categories = 5, seed = 7)$data
  realized_f <- vapply(seq_len(ncol(dat_f)),
                       function(j) tabulate(dat_f[, j], nbins = 5L) / nrow(dat_f),
                       numeric(5L))
  expect_lt(max(abs(realized_f - 0.2)), 0.01)

  # The mapping is what buys this: cutting the same draw at the unmapped normal
  # thresholds misses the request by far more than the Monte Carlo error above.
  cont <- efa_simulate(N = 1e5, R = R_vm, marginals = "VM", skewness = sk,
                       kurtosis = ku, seed = 7)$data
  naive <- vapply(seq_len(ncol(cont)),
                  function(j) tabulate(findInterval(cont[, j], stats::qnorm((1:4) / 5)) + 1L,
                                       nbins = 5L) / nrow(cont),
                  numeric(5L))
  expect_gt(max(abs(naive - 0.2)), 0.05)
})

test_that(".vm_thresholds maps a cubic that increases across the thresholds", {
  tau <- stats::qnorm((1:4) / 5)
  # The normal marginal has coefficients (0, 1, 0, 0): an identity map that leaves the
  # thresholds where they are, with no fallback reported.
  norm_map <- .vm_thresholds(list(tau), matrix(c(0, 1, 0, 0), 1, 4))
  expect_equal(norm_map$thresholds[[1]], tau)
  expect_length(norm_map$non_monotone, 0L)

  ft <- .fleishman_table(1.5, 4)
  expect_equal(.vm_thresholds(list(tau), ft)$thresholds[[1]],
               ft[1, 1L] + ft[1, 2L] * tau + ft[1, 3L] * tau^2 + ft[1, 4L] * tau^3)

  # A platykurtic marginal has d < 0, so f' has real roots and the cubic is not monotone
  # on the whole line -- but its turning points are far out in the tails, so it does
  # increase across ordinary thresholds and must still be mapped.
  ft_flat <- .fleishman_table(0, -1)
  expect_lt(ft_flat[1, 4L], 0)
  flat_map <- .vm_thresholds(list(tau), ft_flat)
  expect_length(flat_map$non_monotone, 0L)
  expect_false(is.unsorted(flat_map$thresholds[[1]]))

  # Pushing a threshold past the turning point (|z| = 2.25 here) is what forces the
  # fallback: the mapping would no longer preserve the category order.
  far <- c(-2.6, tau)
  fallback <- .vm_thresholds(list(far), ft_flat)
  expect_equal(fallback$thresholds[[1]], far)
  expect_identical(fallback$non_monotone, 1L)
})

test_that(".vm_thresholds falls back when the tail folds mass back below an outer cut", {
  # A threshold just INSIDE the turning point passes the derivative screen, but the cubic
  # re-crosses the mapped cut a little further out and the tail mass beyond that point
  # lands in the wrong category: at excess kurtosis -1 the turning point is |z| = 2.25 and
  # a top category of .014 (threshold 2.197) would come out near .0035 -- a factor of four
  # off, silently, under the derivative screen alone. The mass screen catches it.
  ft_flat <- .fleishman_table(0, -1)
  tau_tail <- stats::qnorm(0.986)
  expect_gt(ft_flat[1, 2L] + 2 * ft_flat[1, 3L] * tau_tail +
              3 * ft_flat[1, 4L] * tau_tail^2, 0)   # derivative screen alone passes
  fb <- .vm_thresholds(list(tau_tail), ft_flat)
  expect_identical(fb$non_monotone, 1L)
  expect_equal(fb$thresholds[[1]], tau_tail)

  # Not a platykurtosis-only phenomenon: a skewed leptokurtic marginal dips on one side,
  # and a 5% bottom category (threshold -1.645, turning point -1.69) folds back 84% of
  # the requested mass. It must fall back too.
  ft_skew <- .fleishman_table(1.5, 3)
  fb_skew <- .vm_thresholds(list(stats::qnorm(0.05)), ft_skew)
  expect_identical(fb_skew$non_monotone, 1L)

  # The screen is a tolerance, not a monotonicity dogma: ordinary interior thresholds on
  # the same platykurtic marginal still map (their re-crossings carry negligible mass).
  ok <- .vm_thresholds(list(stats::qnorm((1:4) / 5)), ft_flat)
  expect_length(ok$non_monotone, 0L)
})

test_that("thresholds that outrun the Fleishman cubic warn and keep the normal scale", {
  # Same construction end to end: a strongly platykurtic marginal with a category
  # boundary beyond the cubic's turning point falls back with a classed warning.
  props <- rep(list(c(0.005, 0.7, 0.28, 0.015)), ncol(R_vm))
  # The fallback empties that category too, and necessarily so: the untransformed
  # threshold (-2.58) lies outside the range of the platykurtic draw, which is precisely
  # why the normal-scale cut points do not reproduce the requested proportions.
  expect_warning(
    expect_warning(
      efa_simulate(N = 5000, R = R_vm, marginals = "VM", skewness = 0, kurtosis = -1,
                   categories = props, seed = 11),
      class = "efa_simulate_threshold_fallback"),
    class = "efa_simulate_empty_category")
})

test_that("match = 'thresholds' cuts non-normal marginals without error", {
  # The naive mode works with the standardized non-normal marginals: the
  # Vale-Maurelli draw is cut as is, yielding a valid integer ordinal matrix (its
  # polychoric correlation departs from the target, the documented cost of non-normal
  # marginals).
  dat <- efa_simulate(N = 2000, R = R_vm, marginals = "VM", skewness = 1,
                      kurtosis = 2, categories = 4, match = "thresholds", seed = 3)$data
  expect_type(dat, "integer")
  expect_equal(dim(dat), c(2000, ncol(R_vm)))
  expect_true(all(dat %in% 1:4))
  dat2 <- efa_simulate(N = 2000, R = R_vm, marginals = "VM", skewness = 1,
                       kurtosis = 2, categories = 4, match = "thresholds", seed = 3)$data
  # Category codes, not floating-point values: expect_identical() is the right comparison
  # and a tolerance would buy nothing.
  expect_identical(dat, dat2)
})

test_that("match = 'polychoric' rejects non-normal marginals", {
  # The polychoric model assumes a normal latent, so pairing it with non-normal
  # marginals is a contradiction rather than an approximation.
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "VM", skewness = 1, kurtosis = 2,
                 categories = 5, match = "polychoric"),
    class = "efa_simulate_match_conflict")
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "IG", skewness = 1, kurtosis = 2,
                 categories = 5, match = "polychoric"),
    class = "efa_simulate_match_conflict")
  expect_error(
    efa_simulate(N = 10, R = R_vm, marginals = "empirical", categories = 5,
                 match = "polychoric"),
    class = "efa_simulate_match_conflict")
})

test_that("categories are not supported with empirical marginals", {
  # Ordinal thresholds are on the standard-normal scale, but the empirical draw
  # carries the arbitrary scale of marginal_data, so the combination is rejected
  # (rather than silently returning mis-proportioned or degenerate categories).
  set.seed(1)
  md <- matrix(stats::rexp(200 * ncol(R_vm)), ncol = ncol(R_vm))
  expect_error(
    efa_simulate(N = 100, R = R_vm, marginals = "empirical", marginal_data = md,
                 n_factors = 2, categories = 5),
    class = "efa_simulate_input")
})

test_that("match only applies when categories requests ordinal output", {
  expect_error(efa_simulate(N = 10, R = R_vm, match = "polychoric"),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, R = R_vm, match = "thresholds"),
               class = "efa_simulate_input")
})

test_that("invalid category specifications are rejected", {
  p <- ncol(R_vm)
  # Counts must be whole numbers of at least two categories, of length 1 or p.
  expect_error(efa_simulate(N = 10, R = R_vm, categories = 1),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, R = R_vm, categories = 2.5),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, R = R_vm, categories = rep(3L, p + 1L)),
               class = "efa_simulate_input")
  # A list of proportions must have one element per variable, each a valid
  # probability vector of at least two strictly-positive proportions summing to 1.
  expect_error(efa_simulate(N = 10, R = R_vm, categories = list(c(0.5, 0.5))),
               class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 10, R = R_vm, categories = rep(list(c(0.2, 0.2, 0.2)), p)),
    class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 10, R = R_vm, categories = rep(list(c(0, 0.5, 0.5)), p)),
    class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, R = R_vm, categories = rep(list(1), p)),
               class = "efa_simulate_input")
  # A dimensioned object (matrix/array) is rejected rather than flattened
  # column-major into per-variable counts (a p-entry matrix passes the length check).
  expect_error(efa_simulate(N = 10, R = R_vm, categories = matrix(2:7, 2, 3)),
               class = "efa_simulate_input")
  # ... and likewise for a matrix supplied as a proportion list element.
  expect_error(
    efa_simulate(N = 10, R = R_vm,
                 categories = rep(list(matrix(c(0.25, 0.25, 0.25, 0.25), 2, 2)), p)),
    class = "efa_simulate_input")
  # Proportions that pass the sum tolerance but whose interior cumulative sum would
  # cross 1 are renormalized rather than driving qnorm to NaN (which would crash
  # findInterval with an unclassed error); the tiny last category is left empty. The
  # total sits just inside the sum tolerance (1 + 9e-7 < 1 + 1e-6) rather than exactly
  # on it, so the check is not decided by the last-bit rounding of the sum (which
  # differs across BLAS/platforms).
  props_edge <- rep(list(c(0.5, 0.5000005, 4e-7)), p)
  expect_warning(
    dat_edge <- efa_simulate(N = 500, R = R_vm, categories = props_edge, seed = 1),
    class = "efa_simulate_empty_category")
  expect_type(dat_edge$data, "integer")
})

test_that("empty categories are reported with a warning", {
  # Three cases cannot fill five categories, so some are left empty.
  expect_warning(
    efa_simulate(N = 3, R = R_vm, categories = 5, seed = 1),
    class = "efa_simulate_empty_category")
})

test_that("ordinal draws are reproducible and support multiple datasets", {
  a <- efa_simulate(N = 500, R = R_vm, categories = 4, seed = 11)
  b <- efa_simulate(N = 500, R = R_vm, categories = 4, seed = 11)
  expect_equal(a, b, tolerance = fp_tol)
  expect_type(a$data, "integer")

  sims <- efa_simulate(N = 500, R = R_vm, categories = 4, n_datasets = 3, seed = 11)$data
  expect_type(sims, "list")
  expect_length(sims, 3)
  expect_true(all(vapply(sims, function(d) is.matrix(d) && is.integer(d), logical(1))))
  expect_equal(dim(sims[[1]]), c(500, ncol(R_vm)))
  expect_false(isTRUE(all.equal(sims[[1]], sims[[2]])))
})

test_that("MCAR holes each variable at the target expected rate", {
  # MCAR removes values independently of the data, so the achieved marginal rate per
  # variable is the target up to sampling error (expected rate, not an exact count).
  dat <- efa_simulate(N = 1e4, R = R_vm, missing = "MCAR", missing_prop = 0.2, seed = 1)$data
  expect_true(anyNA(dat))
  expect_lt(max(abs(colMeans(is.na(dat)) - 0.2)), 0.02)
})

test_that("MAR and MNAR achieve the target expected missing rate", {
  d_mar <- efa_simulate(N = 1e4, R = R_vm, missing = "MAR", missing_prop = 0.25,
                        seed = 2)$data
  d_mnar <- efa_simulate(N = 1e4, R = R_vm, missing = "MNAR", missing_prop = 0.25,
                         seed = 2)$data
  # Calibrating the logistic intercept per variable targets the mean missing
  # probability, so the achieved marginal rate matches the target.
  expect_lt(max(abs(colMeans(is.na(d_mar)) - 0.25)), 0.02)
  expect_lt(max(abs(colMeans(is.na(d_mnar)) - 0.25)), 0.02)
})

test_that("missingness is orthogonal to the marginal distribution", {
  # Missingness applies to any marginal type (only ordinal cutting is gated to the
  # standard-normal-scale marginals); empirical and VM draws accept it unchanged.
  p <- nrow(Lambda_pop)
  set.seed(30)
  x_emp <- matrix(stats::rexp(300 * p), nrow = 300, ncol = p)
  d_emp <- efa_simulate(N = 5000, Lambda = Lambda_pop, Phi = Phi_pop,
                        marginals = "empirical", marginal_data = x_emp, n_factors = 3,
                        missing = "MAR", missing_prop = 0.2, seed = 31)$data
  expect_lt(max(abs(colMeans(is.na(d_emp)) - 0.2)), 0.03)

  d_vm <- efa_simulate(N = 5000, R = R_vm, marginals = "VM", skewness = 1, kurtosis = 2,
                       missing = "MNAR", missing_prop = 0.2, seed = 32)$data
  expect_lt(max(abs(colMeans(is.na(d_vm)) - 0.2)), 0.03)
})

test_that(".calibrate_missing_intercept hits the target mean probability", {
  set.seed(14)
  z <- as.vector(scale(stats::rnorm(1000)))
  for (b in c(0, 0.5, 1.5, 3)) {
    for (target in c(0.05, 0.2, 0.5, 0.8)) {
      a <- .calibrate_missing_intercept(z, b, target)
      expect_equal(mean(stats::plogis(a + b * z)), target, tolerance = 1e-6)
    }
  }
  # A constant predictor (z = 0) reduces to plogis(a) = target.
  a0 <- .calibrate_missing_intercept(rep(0, 100), 2, 0.3)
  expect_equal(stats::plogis(a0), 0.3, tolerance = 1e-6)
})

test_that("the MAR mask depends on the predictor, not the variable itself", {
  # Independent columns, so a dependence on the predictor cannot leak through
  # correlation with the variable being holed.
  set.seed(11)
  n <- 2e4
  x <- matrix(stats::rnorm(n * 3), n, 3)
  # pred_idx = c(2, 3, 1): variable 1's missingness is driven by variable 2.
  out <- .apply_missing(x, "MAR", prop = 0.3, strength = 3, pred_idx = c(2L, 3L, 1L))
  miss1 <- as.numeric(is.na(out[, 1]))
  expect_gt(stats::cor(miss1, x[, 2]), 0.15)      # depends on the predictor (var 2)
  expect_lt(abs(stats::cor(miss1, x[, 1])), 0.05) # not on variable 1 itself
  expect_equal(mean(miss1), 0.3, tolerance = 0.02)
  # The non-holed values are left byte-identical to the input.
  keep <- !is.na(out[, 1])
  expect_identical(out[keep, 1], x[keep, 1])
})

test_that("the MNAR mask depends on the variable's own value", {
  set.seed(12)
  n <- 2e4
  x <- matrix(stats::rnorm(n * 3), n, 3)
  out <- .apply_missing(x, "MNAR", prop = 0.3, strength = 3, pred_idx = NULL)
  miss1 <- as.numeric(is.na(out[, 1]))
  expect_gt(stats::cor(miss1, x[, 1]), 0.15)
  # A positive slope preferentially removes high values, so the observed mean shifts low.
  expect_lt(mean(out[!is.na(out[, 1]), 1]), -0.1)
})

test_that(".apply_missing reads predictors from the complete draw", {
  # Column 2's predictor is column 1, which is holed first. The mask must use the
  # complete values of column 1, not the NA-holed ones (else the probabilities would be
  # NA and the intended dependence would break).
  set.seed(13)
  n <- 8000
  x <- matrix(stats::rnorm(n * 2), n, 2)
  out <- .apply_missing(x, "MAR", prop = 0.3, strength = 3, pred_idx = c(2L, 1L))
  miss2 <- as.numeric(is.na(out[, 2]))
  expect_false(anyNA(miss2))
  expect_gt(stats::cor(miss2, x[, 1]), 0.15)
  expect_equal(mean(miss2), 0.3, tolerance = 0.03)
})

test_that("missing_vars restricts the holed columns and leaves the rest complete", {
  p <- ncol(R_vm)
  dat <- efa_simulate(N = 4000, R = R_vm, missing = "MAR", missing_prop = 0.2,
                      missing_vars = 1:3, missing_predictor = 4:6, seed = 41)$data
  # Only the named columns carry NAs, and they carry them at the target rate.
  expect_true(all(colMeans(is.na(dat))[1:3] > 0))
  expect_equal(unname(colSums(is.na(dat))[4:6]), rep(0L, 3L))
  expect_lt(max(abs(colMeans(is.na(dat))[1:3] - 0.2)), 0.03)

  # Names select the same columns as indices.
  R_named <- R_vm
  nm <- paste0("v", seq_len(p))
  dimnames(R_named) <- list(nm, nm)
  by_name <- efa_simulate(N = 400, R = R_named, missing = "MCAR", missing_prop = 0.2,
                          missing_vars = nm[c(1, 3)], seed = 42)$data
  by_idx <- efa_simulate(N = 400, R = R_named, missing = "MCAR", missing_prop = 0.2,
                         missing_vars = c(1, 3), seed = 42)$data
  expect_identical(is.na(by_name), is.na(by_idx))

  # The default (NULL) is the all-columns behaviour.
  all_cols <- efa_simulate(N = 400, R = R_vm, missing = "MCAR", missing_prop = 0.2,
                           seed = 43)
  explicit <- efa_simulate(N = 400, R = R_vm, missing = "MCAR", missing_prop = 0.2,
                           missing_vars = seq_len(p), seed = 43)
  expect_equal(all_cols$data, explicit$data, tolerance = fp_tol)
})

test_that("missing_vars is validated and gated", {
  p <- ncol(R_vm)
  # Only meaningful with a mechanism.
  expect_error(efa_simulate(N = 100, R = R_vm, missing_vars = 1:3),
               class = "efa_simulate_input")
  # Out-of-range, duplicated, empty, non-integer, and unknown-name specs.
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                            missing_vars = c(1, 99)),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                            missing_vars = c(1, 1)),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                            missing_vars = integer(0)),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                            missing_vars = 1.5),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                            missing_vars = "nope"),
               class = "efa_simulate_input")
  # An index beyond the integer range is caught on the supplied value: coercing it first
  # would give NA and leave the range test an `if (NA)` base error. Same for the predictor.
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                            missing_vars = 3e9),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MAR", missing_prop = 0.1,
                            missing_predictor = rep(3e9, p)),
               class = "efa_simulate_input")
  # missing_vars applies to every mechanism, not just MAR.
  d_mnar <- efa_simulate(N = 2000, R = R_vm, missing = "MNAR", missing_prop = 0.2,
                         missing_vars = c(2L, 5L), seed = 44)$data
  expect_true(all(colMeans(is.na(d_mnar))[c(2, 5)] > 0))
  expect_equal(unname(colSums(is.na(d_mnar))[c(1, 3, 4, 6)]), rep(0L, 4L))
  # missing_predictor is now one entry per holed variable, not per variable.
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MAR", missing_prop = 0.1,
                            missing_vars = 1:3, missing_predictor = seq_len(p)),
               class = "efa_simulate_input")
  # ... and still may not point a holed variable at itself.
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MAR", missing_prop = 0.1,
                            missing_vars = 1:3, missing_predictor = c(4L, 2L, 6L)),
               class = "efa_simulate_input")
})

test_that("a strictly MAR design leaves FIML moments unbiased", {
  skip_on_cran()
  skip_if_not_slow()
  # With the holed variables disjoint from their predictors, every predictor is fully
  # observed, so the mechanism is ignorably MAR and FIML -- which is consistent under
  # ignorable MAR -- recovers the population correlations. Holing every column instead
  # leaves the predictors partly missing, which is only MAR given the complete data and
  # biases FIML noticeably at this rate and strength. The tolerances are loose: this
  # pins the direction and the order of magnitude, not a Monte-Carlo estimate.
  ign <- efa_simulate(N = 2e4, Lambda = Lambda_pop, Phi = Phi_pop, missing = "MAR",
                      missing_prop = 0.3, missing_strength = 3,
                      missing_vars = 1:9, missing_predictor = 10:18, seed = 71)$data
  all_holed <- efa_simulate(N = 2e4, Lambda = Lambda_pop, Phi = Phi_pop,
                            missing = "MAR", missing_prop = 0.3,
                            missing_strength = 3, seed = 71)$data

  # Mask placement: the ignorable design holes exactly the first nine columns.
  expect_true(all(colMeans(is.na(ign))[1:9] > 0))
  expect_equal(unname(colSums(is.na(ign))[10:18]), rep(0L, 9L))
  expect_true(all(colMeans(is.na(all_holed)) > 0))

  off <- function(x) x[lower.tri(x)]
  R_ign <- stats::cov2cor(.fiml_em_moments(ign)$sigma)
  R_all <- stats::cov2cor(.fiml_em_moments(all_holed)$sigma)
  bias_ign <- mean(off(R_ign) - off(R_pop))
  bias_all <- mean(off(R_all) - off(R_pop))

  # The ignorable design is unbiased on the mean off-diagonal correlation; holing every
  # column attenuates it by an order of magnitude more.
  expect_lt(abs(bias_ign), 0.005)
  expect_lt(bias_all, -0.01)
  expect_lt(abs(bias_ign), abs(bias_all) / 3)
})

test_that(".resolve_missing_predictor resolves and validates specs", {
  p <- 5L
  vn <- paste0("v", seq_len(p))
  # NULL gives the cyclic next neighbour.
  expect_identical(.resolve_missing_predictor(NULL, p, vn), c(2L, 3L, 4L, 5L, 1L))
  # Restricted to a subset of holed variables, the cyclic default and an explicit spec
  # both give one predictor per holed variable.
  expect_identical(.resolve_missing_predictor(NULL, p, vn, c(1L, 3L)), c(2L, 4L))
  expect_identical(.resolve_missing_predictor(c(4L, 5L), p, vn, c(1L, 3L)), c(4L, 5L))
  # A spec of the wrong length for the holed subset, or one that self-predicts, errors.
  expect_error(.resolve_missing_predictor(c(2L, 3L, 4L, 5L, 1L), p, vn, c(1L, 3L)),
               class = "efa_simulate_input")
  expect_error(.resolve_missing_predictor(c(2L, 3L), p, vn, c(1L, 3L)),
               class = "efa_simulate_input")
  # Names resolve to indices.
  expect_identical(.resolve_missing_predictor(vn[c(2:5, 1)], p, vn),
                   c(2L, 3L, 4L, 5L, 1L))
  # A self-predicting index errors (that would be MNAR).
  expect_error(.resolve_missing_predictor(seq_len(p), p, vn),
               class = "efa_simulate_input")
  # A single shared predictor is not allowed (wrong length).
  expect_error(.resolve_missing_predictor(2L, p, vn), class = "efa_simulate_input")
  # Out-of-range index, unknown name, and p < 2 error.
  expect_error(.resolve_missing_predictor(c(2L, 3L, 4L, 5L, 99L), p, vn),
               class = "efa_simulate_input")
  expect_error(.resolve_missing_predictor("nope", p, vn), class = "efa_simulate_input")
  expect_error(.resolve_missing_predictor(NULL, 1L, "v1"), class = "efa_simulate_input")
})

test_that("missingness is reproducible and supports multiple datasets", {
  a <- efa_simulate(N = 300, R = R_vm, missing = "MAR", missing_prop = 0.2, seed = 5)
  b <- efa_simulate(N = 300, R = R_vm, missing = "MAR", missing_prop = 0.2, seed = 5)
  expect_equal(a, b, tolerance = fp_tol)

  sims <- efa_simulate(N = 300, R = R_vm, missing = "MCAR", missing_prop = 0.2,
                       n_datasets = 3, seed = 5)$data
  expect_type(sims, "list")
  expect_length(sims, 3)
  expect_true(all(vapply(sims, anyNA, logical(1))))
  # Each replicate has its own stream, so the NA patterns differ.
  expect_false(identical(is.na(sims[[1]]), is.na(sims[[2]])))
})

test_that("missingness combines with ordinal categories", {
  # The mask acts on the latent draw and the NAs propagate into the codes, so the NA
  # positions match a continuous draw at the same seed and the codes stay valid.
  dat <- efa_simulate(N = 3000, R = R_vm, categories = 4, missing = "MCAR",
                      missing_prop = 0.15, seed = 6)$data
  cont <- efa_simulate(N = 3000, R = R_vm, missing = "MCAR", missing_prop = 0.15,
                       seed = 6)$data
  expect_type(dat, "integer")
  expect_true(anyNA(dat))
  expect_true(all(dat[!is.na(dat)] %in% 1:4))
  expect_lt(max(abs(colMeans(is.na(dat)) - 0.15)), 0.03)
  # Discretizing does not move the holes.
  expect_identical(is.na(dat), is.na(cont))
})

test_that("MCAR and MNAR work with a single variable, MAR does not", {
  R1 <- matrix(1, 1, 1)
  expect_true(anyNA(efa_simulate(N = 500, R = R1, missing = "MCAR",
                                 missing_prop = 0.2, seed = 1)$data))
  expect_true(anyNA(efa_simulate(N = 500, R = R1, missing = "MNAR",
                                 missing_prop = 0.2, seed = 1)$data))
  # MAR needs another variable to predict from.
  expect_error(
    efa_simulate(N = 500, R = R1, missing = "MAR", missing_prop = 0.2),
    class = "efa_simulate_input")
})

test_that("MCAR draws its runif per column, leaving the pre-mask draw intact", {
  # MCAR consumes one runif(n) per column just like MAR/MNAR, so the non-missing values
  # equal the corresponding complete draw (the mask only overwrites, never reorders).
  full <- efa_simulate(N = 1000, R = R_vm, seed = 21)$data
  holed <- efa_simulate(N = 1000, R = R_vm, missing = "MCAR", missing_prop = 0.3,
                        seed = 21)$data
  keep <- !is.na(holed)
  expect_equal(holed[keep], full[keep])
})

test_that("MAR predictor can be given by name or index", {
  R_named <- R_vm
  nm <- paste0("v", seq_len(ncol(R_vm)))
  dimnames(R_named) <- list(nm, nm)
  a <- efa_simulate(N = 400, R = R_named, missing = "MAR", missing_prop = 0.2,
                    missing_predictor = nm[c(2:6, 1)], seed = 8)$data
  b <- efa_simulate(N = 400, R = R_named, missing = "MAR", missing_prop = 0.2,
                    missing_predictor = c(2:6, 1), seed = 8)$data
  expect_identical(is.na(a), is.na(b))
})

test_that("missing-data arguments are validated and gated", {
  p <- ncol(R_vm)
  # Companion settings require a mechanism.
  expect_error(efa_simulate(N = 100, R = R_vm, missing_prop = 0.1),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing_strength = 1),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing_predictor = 2),
               class = "efa_simulate_input")
  # missing_prop is required and must be a single proportion in (0, 1).
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR"),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 1),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 1.2),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 100, R = R_vm, missing = "MCAR",
                            missing_prop = c(0.1, 0.2)),
               class = "efa_simulate_input")
  # strength/predictor do not apply to MCAR.
  expect_error(
    efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                 missing_strength = 1),
    class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 100, R = R_vm, missing = "MCAR", missing_prop = 0.1,
                 missing_predictor = 2),
    class = "efa_simulate_input")
  # A non-finite strength is rejected on the value-dependent paths.
  expect_error(
    efa_simulate(N = 100, R = R_vm, missing = "MNAR", missing_prop = 0.1,
                 missing_strength = Inf),
    class = "efa_simulate_input")
  # MNAR has no separate predictor.
  expect_error(
    efa_simulate(N = 100, R = R_vm, missing = "MNAR", missing_prop = 0.1,
                 missing_predictor = 2),
    class = "efa_simulate_input")
  # A MAR predictor cannot be the variable itself, out of range, or unknown.
  expect_error(
    efa_simulate(N = 100, R = R_vm, missing = "MAR", missing_prop = 0.1,
                 missing_predictor = seq_len(p)),
    class = "efa_simulate_input")
  expect_error(
    efa_simulate(N = 100, R = R_vm, missing = "MAR", missing_prop = 0.1,
                 missing_predictor = c(2:6, 99)),
    class = "efa_simulate_input")
})

# ---- model error (none / CB / TKL / WB) --------------------------------------

# The 18-variable, three-factor baseline population gives ample residual degrees of
# freedom (df = 102) for injecting model error.
Lambda_me <- population_models$loadings$baseline
Phi_me <- population_models$phis_3$moderate
p_me <- nrow(Lambda_me)
q_me <- ncol(Lambda_me)
df_me <- ((p_me - q_me)^2 - (p_me + q_me)) / 2

test_that("model error is off by default and when no target is given", {
  # A bare Lambda call returns the exact population (backward-compatible: the default
  # method is CB, but nothing is perturbed without a target).
  sim <- efa_simulate(N = 200, Lambda = Lambda_me, Phi = Phi_me, seed = 1)
  expect_s3_class(sim, "efa_simulated")
  expect_null(sim$model_error)
  rp <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, return_pop = TRUE)
  expect_equal(rp$population[1, 2], 0.36)
  expect_null(rp$model_error)
})

test_that("CB model error achieves the target RMSEA to numerical precision", {
  skip_on_cran()
  for (tr in c(0.03, 0.05, 0.08)) {
    sim <- efa_simulate(N = 100, Lambda = Lambda_me, Phi = Phi_me,
                        target_rmsea = tr, seed = 1)
    expect_equal(sim$model_error$method, "CB")
    # Self-consistent: the achieved RMSEA equals the target (the model stays the minimizer).
    expect_equal(sim$model_error$rmsea, tr, tolerance = 1e-6)
    # The perturbed population is a proper, positive-definite correlation matrix.
    expect_equal(diag(sim$population), rep(1, p_me), ignore_attr = TRUE)
    expect_gt(min(eigen(sim$population, symmetric = TRUE, only.values = TRUE)$values), 0)
    expect_true(is.finite(sim$model_error$kappa))
  }
})

test_that("CB achieved RMSEA matches an independent factanal fit", {
  skip_on_cran()
  # factanal fits the best-fitting q-factor model; CB makes the specified model that
  # minimizer, so the minimized discrepancy reproduces the target RMSEA.
  rp <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 0.05,
                     return_pop = TRUE, seed = 2)
  fit <- stats::factanal(covmat = rp$population, factors = q_me, n.obs = 1e6)
  rmsea_min <- sqrt(fit$criteria[["objective"]] / df_me)
  expect_equal(rmsea_min, 0.05, tolerance = 1e-4)
  expect_equal(rp$model_error$rmsea, 0.05, tolerance = 1e-6)
})

test_that("TKL matches a single RMSEA target closely and both targets approximately", {
  skip_on_cran()
  s1 <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "TKL",
                     target_rmsea = 0.05, return_pop = TRUE, seed = 3)
  expect_equal(s1$model_error$method, "TKL")
  expect_equal(s1$model_error$rmsea, 0.05, tolerance = 1e-3)
  # Two targets, two knobs: a close compromise rather than an exact match (absolute band).
  s2 <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "TKL",
                     target_rmsea = 0.05, target_cfi = 0.95, return_pop = TRUE, seed = 4)
  expect_lt(abs(s2$model_error$rmsea - 0.05), 0.02)
  expect_lt(abs(s2$model_error$cfi - 0.95), 0.02)
})

test_that("TKL warns when the two targets are jointly unattainable", {
  skip_on_cran()
  # A high target CFI dominates the weighted objective, so the two knobs cannot honour a
  # target RMSEA of .05 alongside it: the achieved RMSEA comes out about half the target.
  # The compromise is otherwise only visible by reading the achieved values back off the
  # result, so it is flagged.
  expect_warning(
    s <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "TKL",
                      target_rmsea = 0.05, target_cfi = 0.99, return_pop = TRUE,
                      seed = 6),
    class = "efa_simulate_model_error_compromise")
  expect_lt(s$model_error$rmsea, 0.04)
  # A single target is matched closely, so nothing is flagged there.
  expect_no_warning(
    efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "TKL",
                 target_rmsea = 0.05, return_pop = TRUE, seed = 6))
})

test_that("TKL can target the CFI alone", {
  skip_on_cran()
  s <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "TKL",
                    target_cfi = 0.95, return_pop = TRUE, seed = 5)
  expect_lt(abs(s$model_error$cfi - 0.95), 0.02)
})

test_that("WB produces a realized RMSEA in a relaxed band around the target", {
  skip_on_cran()
  # A single inverse-Wishart draw: the realized RMSEA varies around (and tends to exceed)
  # the target, so check a broad band across a few seeds rather than a tight tolerance.
  ach <- vapply(1:8, function(s) {
    efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "WB",
                 target_rmsea = 0.05, return_pop = TRUE, seed = s)$model_error$rmsea
  }, numeric(1))
  expect_true(all(ach > 0.02 & ach < 0.12))
  # The precision parameter is m = 1 / target_rmsea^2.
  m <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "WB",
                    target_rmsea = 0.05, return_pop = TRUE, seed = 1)$model_error$m
  expect_equal(m, 1 / 0.05^2)
})

test_that("WB rejects a target RMSEA too large for the number of variables", {
  # m = 1 / rmsea^2 must be at least p; rmsea = 0.3 -> m ~ 11 < 18.
  expect_error(
    efa_simulate(Lambda = Lambda_me, Phi = Phi_me, model_error = "WB",
                 target_rmsea = 0.3, return_pop = TRUE),
    class = "efa_simulate_model_error")
})

test_that("model error is reproducible and shared across replicates", {
  a <- efa_simulate(N = 150, Lambda = Lambda_me, Phi = Phi_me,
                    target_rmsea = 0.05, n_datasets = 2, seed = 7)
  b <- efa_simulate(N = 150, Lambda = Lambda_me, Phi = Phi_me,
                    target_rmsea = 0.05, n_datasets = 2, seed = 7)
  # Compared numerically rather than bit for bit: every step from the population
  # perturbation to the draw runs through the BLAS, and a threaded BLAS need not return
  # identical bits for two identical calls within one session.
  expect_equal(a, b)
  # One population perturbation is shared by all replicates, which still differ.
  expect_equal(a$model_error$rmsea, 0.05, tolerance = 1e-6)
  expect_false(isTRUE(all.equal(a$data[[1]], a$data[[2]])))
})

test_that("model error requires a factor model and residual degrees of freedom", {
  R <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, return_pop = TRUE)$population
  # A bare R has no specified factor model to misfit against.
  expect_error(efa_simulate(N = 10, R = R, target_rmsea = 0.05),
               class = "efa_simulate_input")
  # A model with no residual df (q too large) cannot carry model error: p = 3, q = 1 -> df = 0.
  L_sat <- matrix(0.5, 3, 1)
  expect_error(efa_simulate(N = 10, Lambda = L_sat, target_rmsea = 0.05),
               class = "efa_simulate_input")
})

test_that("model-error arguments are validated and gated", {
  # A target with model_error = "none".
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me,
                            model_error = "none", target_rmsea = 0.05),
               class = "efa_simulate_input")
  # CB / WB target the RMSEA only, not the CFI.
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me, target_cfi = 0.95),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me,
                            model_error = "WB", target_rmsea = 0.05, target_cfi = 0.95),
               class = "efa_simulate_input")
  # Out-of-range targets (the range is the open interval (0, 1); a target of 0 misfit or a CFI
  # of exactly 1 means an exact population and is rejected -- omit the target instead).
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 1.5),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 0),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me,
                            model_error = "TKL", target_cfi = 1.5),
               class = "efa_simulate_input")
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me,
                            model_error = "TKL", target_cfi = 1),
               class = "efa_simulate_input")
  # force_pd needs VM or a model-error target.
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me, force_pd = TRUE),
               class = "efa_simulate_input")
})

test_that("model error rejects singular and non-factor-structure populations", {
  # A positive-semidefinite but SINGULAR population (two items with a communality of 1 are
  # perfectly correlated) is drawable but has no inverse, so the model-error solvers reject it
  # with a classed error rather than an opaque base-R linear-algebra crash.
  L_sing <- matrix(0, 6, 2)
  L_sing[1:3, 1] <- c(1, 1, 0.5)      # items 1 and 2 both fully explained -> R singular
  L_sing[4:6, 2] <- c(0.7, 0.6, 0.5)
  expect_error(efa_simulate(N = 10, Lambda = L_sing, target_rmsea = 0.05),
               class = "efa_simulate_model_error")

  # A correlated-residual (off-diagonal) Psi makes the population not an exact q-factor
  # structure, which the eigen-recovered loadings cannot represent, so it is rejected.
  Psi_corr <- diag(1 - diag(Lambda_me %*% Phi_me %*% t(Lambda_me)))
  Psi_corr[1, 2] <- Psi_corr[2, 1] <- 0.1
  expect_error(efa_simulate(N = 10, Lambda = Lambda_me, Phi = Phi_me, Psi = Psi_corr,
                            target_rmsea = 0.05),
               class = "efa_simulate_model_error")
  # ... and likewise for TKL (which would otherwise take the square root of a negative
  # uniqueness when a negative Psi diagonal pushes a communality above 1).
  Psi_neg <- diag(c(-0.02, rep(0.75, nrow(Lambda_me) - 1L)))
  expect_error(efa_simulate(N = 10, Lambda = matrix(0.5, nrow(Lambda_me), 1),
                            Psi = Psi_neg, model_error = "TKL", target_rmsea = 0.05),
               class = "efa_simulate_model_error")
})

test_that("model error works with an equivalent diagonal-matrix Psi", {
  skip_on_cran()
  # A diagonal matrix Psi is a proper factor structure, so model error still hits the target.
  Psi_diag <- diag(1 - diag(Lambda_me %*% Phi_me %*% t(Lambda_me)))
  sim <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, Psi = Psi_diag,
                      target_rmsea = 0.05, return_pop = TRUE, seed = 1)
  expect_equal(sim$model_error$rmsea, 0.05, tolerance = 1e-6)
})

test_that("model error combines with non-normal marginals and missingness", {
  skip_on_cran()
  # Model error is orthogonal to the marginal, ordinal, and missing-data options.
  sim <- efa_simulate(N = 2000, Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 0.05,
                      marginals = "VM", skewness = 1, kurtosis = 2,
                      missing = "MCAR", missing_prop = 0.1, seed = 8)
  expect_equal(sim$model_error$rmsea, 0.05, tolerance = 1e-6)
  expect_true(anyNA(sim$data))
  expect_lt(max(abs(colMeans(is.na(sim$data)) - 0.1)), 0.03)
})

test_that(".efa_population_fit reports the population-limit fit reused from EFA()", {
  # An exact population fits its own model perfectly (RMSEA 0, CFI 1).
  R <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, return_pop = TRUE)$population
  h2 <- diag(Lambda_me %*% Phi_me %*% t(Lambda_me))
  Cstd <- R
  diag(Cstd) <- h2
  ev <- eigen(Cstd, symmetric = TRUE)
  L <- ev$vectors[, seq_len(q_me)] %*% diag(sqrt(ev$values[seq_len(q_me)]))
  pf <- .efa_population_fit(L, R)
  # The discrepancy is analytically zero here, so what the ML criterion actually returns is
  # the rounding of an 18-term trace against two log-determinants -- of order eps * cond(R).
  # RMSEA takes the square root of that dust, which lifts it to ~1e-8, and the max(0, .)
  # clamp on the criterion hands the assertion a free pass whenever the dust happens to
  # come out negative. Comparing against exactly 0 therefore tests the sign of the last
  # bit. Bound it instead at a level no genuine misfit could reach: 1e-6 still proves a
  # discrepancy below 1e-10, i.e. a perfect population fit, with a hundredfold margin over
  # the dust. CFI is safe as it stands -- its target of 1 makes waldo compare relatively.
  expect_lt(pf$rmsea, 1e-6)
  expect_equal(pf$cfi, 1, tolerance = 1e-8)
  expect_equal(pf$df, df_me)
})

test_that("the efa_simulated object has the documented shape", {
  sim <- efa_simulate(N = 50, Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 0.05, seed = 1)
  expect_s3_class(sim, "efa_simulated")
  expect_named(sim, c("data", "population", "model_error", "settings"))
  expect_true(is.matrix(sim$data))
  expect_equal(dim(sim$population), c(p_me, p_me))
  expect_named(sim$model_error,
               c("method", "target_rmsea", "target_cfi", "rmsea", "cfi", "df", "kappa"))
  # return_pop drops the data but keeps the population and model-error record.
  rp <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 0.05, return_pop = TRUE)
  expect_null(rp$data)
  expect_equal(dim(rp$population), c(p_me, p_me))
  expect_equal(rp$model_error$rmsea, 0.05, tolerance = 1e-6)
})

test_that("print.efa_simulated summarises the object", {
  skip_on_cran()
  testthat::local_reproducible_output()
  # Scrub decimals: the achieved CFI is a BLAS-sensitive derived value, so pin the
  # wording and layout (and the integer dimensions) but not the computed decimals.
  sim <- efa_simulate(N = 100, Lambda = Lambda_me, Phi = Phi_me, target_rmsea = 0.05, seed = 1)
  expect_snapshot(print(sim), transform = scrub_num)
  rp <- efa_simulate(Lambda = Lambda_me, Phi = Phi_me, return_pop = TRUE)
  expect_snapshot(print(rp), transform = scrub_num)
})

rm(Lambda_pop, Phi_pop, R_pop, Lambda_vm, Phi_vm, R_vm,
   Lambda_me, Phi_me, p_me, q_me, df_me)

#' Print and format an efa_compare object
#'
#' `print()` shows a summarised output of the [efa_compare()] function: the mean
#' (with its range), median, and root mean squared distance (RMSE) of the
#' differences, the number of decimals to which all numbers agree, the minimum
#' number of decimals provided, and (for matrices) the number of differing
#' indicator-to-factor correspondences, followed (optionally) by the table of
#' elementwise differences. `format()` assembles the same report and returns it
#' as a character vector; `print()` is `cat(format(x), sep = "\n")`.
#' The lines follow the active console theme, so they are plain when colours are
#' disabled (for example when captured into a file or stripped with
#' [cli::ansi_strip()]).
#'
#' The line reporting the minimum number of decimals provided is shown only when it
#' carries information: two ordinary double matrices carry the full double precision,
#' for which the count is uninformative and the line is omitted.
#'
#' The summary statistics are absolute differences, so they carry no direction. The
#' elementwise differences are signed, and the table is headed by the direction of the
#' subtraction, named with the `x_labels` recorded by [efa_compare()] (`"x"` and `"y"` by
#' default): a negative cell means the first solution is the lower of the two there.
#'
#' @param x An object of class `efa_compare` (output from [efa_compare()]).
#' @param digits,m_red,range_red,round_red,print_diff Display controls, documented
#'   in [efa_compare()]. Each defaults to `NULL`, meaning the value `efa_compare()`
#'   recorded in `x$settings` is used; supplying one overrides it for this call
#'   only, so the comparison need not be recomputed to change the printed report.
#' @param ... Passed from `print()` to `format()`; not otherwise used.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @export
#' @method print efa_compare
#' @examples
#' # A type SPSS EFA to mimick the SPSS implementation
#' EFA_SPSS_5 <- efa_fit(IDS2_R, n_factors = 5,
#'                       estimate_control = estimate_control(type = "SPSS"),
#'                       rotate_control = rotate_control(type = "SPSS"))
#'
#' # A type psych EFA to mimick the psych::fa() implementation
#' EFA_psych_5 <- efa_fit(IDS2_R, n_factors = 5,
#'                        estimate_control = estimate_control(type = "psych"),
#'                        rotate_control = rotate_control(type = "psych"))
#'
#' # compare the two
#' comp <- efa_compare(EFA_SPSS_5$unrot_loadings, EFA_psych_5$unrot_loadings,
#'                     x_labels = c("SPSS", "psych"))
#' comp
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(comp))
#'
#' # the display settings can be changed without recomputing the comparison:
#' print(comp, digits = 2, print_diff = FALSE)
#'
print.efa_compare <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_compare
#' @export
#' @method format efa_compare
format.efa_compare <- function(x, digits = NULL, m_red = NULL, range_red = NULL,
                               round_red = NULL, print_diff = NULL, ...) {

  # extract summary statistics
  diff <- x$diff
  mean_abs_diff <- x$mean_abs_diff
  median_abs_diff <- x$median_abs_diff
  min_abs_diff <- x$min_abs_diff
  max_abs_diff <- x$max_abs_diff
  max_dec <- x$max_dec
  are_equal <- x$are_equal
  g <- x$g
  diff_corres <- x$diff_corres
  diff_corres_cross <- x$diff_corres_cross

  # extract control settings
  corres <- x$settings$corres
  thresh <- x$settings$thresh
  # The labels naming the two compared solutions. `efa_compare()` always records a pair, so
  # the fallback only covers a hand-built or deserialized object; the argument names are the
  # right stand-in, since they are what the sign convention `x - y` refers to.
  x_labels <- x$settings$x_labels
  if (length(x_labels) != 2L) {
    x_labels <- c("x", "y")
  }

  # The display controls are recorded at construction but govern only how the report
  # reads, so an argument supplied here overrides the recorded value for this call and
  # anything left NULL falls back to it. Supplied values are validated exactly as
  # efa_compare() validates its own.
  if (is.null(digits)) digits <- x$settings$digits else checkmate::assert_count(digits)
  if (is.null(m_red)) m_red <- x$settings$m_red else checkmate::assert_number(m_red)
  if (is.null(range_red)) range_red <- x$settings$range_red else checkmate::assert_number(range_red)
  if (is.null(round_red)) round_red <- x$settings$round_red else checkmate::assert_number(round_red)
  if (is.null(print_diff)) print_diff <- x$settings$print_diff else checkmate::assert_flag(print_diff)

  # The report colours each figure the same way: green when it clears its threshold and
  # red when it does not, so a smaller difference reads as a closer match. `isTRUE()`
  # keeps a figure red when the comparison gives NA.
  ok_style <- function(v, ok) {
    .efa_style(v, if (isTRUE(ok)) c("green", "bold") else c("red", "bold"))
  }

  # The statistics keep their decimal padding, so the lines are emitted verbatim below.
  # They are differences between loadings, i.e. bounded coefficients, so they follow the
  # package's number convention and drop the leading zero; the two decimal *counts* below
  # are integers and do not.
  mean_out <- ok_style(.efa_num(mean_abs_diff, digits), mean_abs_diff <= m_red)
  median_out <- ok_style(.efa_num(median_abs_diff, digits), median_abs_diff <= m_red)

  # The minimum and the maximum are printed as one bracketed range and are coloured as
  # one: both follow the test on the maximum, the wider end. The minimum is never larger,
  # so a green maximum always makes a green minimum correct as well.
  min_out <- ok_style(.efa_num(min_abs_diff, digits), max_abs_diff <= range_red)
  max_out <- ok_style(.efa_num(max_abs_diff, digits), max_abs_diff <= range_red)

  # No agreeing decimal place at all is reported as "none" rather than as a count.
  equal_out <- ok_style(if (is.na(are_equal)) "none" else are_equal,
                        !is.na(are_equal) && are_equal >= round_red)

  cli::cli_format_method({
    .print_efa_rule("Summary statistics")

    # Fixed-format statistic lines: the values carry decimal padding and conditional
    # green/red styling, so emit them verbatim (cli never reflows them mid-token).
    cli::cli_verbatim(paste0("Mean [min, max] absolute difference: ",
                             mean_out, " [", min_out, ", ", max_out, "]"))
    cli::cli_verbatim(paste0("Median absolute difference: ", median_out))
    cli::cli_verbatim(paste0("Root mean squared distance (RMSE): ",
                             .efa_style(.efa_num(g, digits), "bold")))
    cli::cli_verbatim(paste0("Max decimals where all numbers agree in absolute value: ",
                             equal_out))

    # The decimals actually carried by the inputs only bound the comparison when one of
    # them was rounded. `.decimals()` renders with `digits = 15`, so an unrounded double
    # always reaches that many decimal places (more for magnitudes below 1), and the count
    # then says nothing about the two solutions; only a smaller count is informative. The
    # threshold therefore tracks the `digits =` argument of `.decimals()` in
    # R/format-helpers.R and must be revisited with it.
    if (max_dec < 15) {
      cli::cli_verbatim(paste0("Minimum number of decimals provided: ",
                               .efa_style(max_dec, "bold")))
    }

    # Differing indicator-to-factor correspondences. Only reported when they were
    # actually compared, which the NA sentinel records for every case that skips the
    # comparison -- vector input, a single-factor matrix, `corres = FALSE`, and a
    # missing loading under `na.rm = FALSE`. The remaining conditions restate the rule
    # `.compare_loadings()` applies, so a hand-built or deserialized object carrying a
    # count it never computed is still not printed as if its solutions had been found
    # to agree. Green when they do agree (0 differing), red otherwise -- mirroring the
    # "smaller difference reads as a closer match" colouring of the statistics above.
    if (isTRUE(corres) && !is.na(diff_corres) && is.matrix(diff) && ncol(diff) > 1L) {
      # The longest line of the report. Emitted verbatim so a styled count is never split
      # from the rule it belongs to, but packed to the console width between the two counts.
      cli::cli_verbatim(.efa_wrap_chunks(c(
        "Differing", "indicator-to-factor", "correspondences:",
        paste0(ok_style(diff_corres, diff_corres == 0), " (highest loading),"),
        paste0(ok_style(diff_corres_cross, diff_corres_cross == 0),
               " (all |loadings| >= ", format(thresh), ")"))))
    }

    if (isTRUE(print_diff)) {
      .print_efa_rule("Elementwise differences")
      # The table below is signed (`x - y`), and its cells are the only place the report
      # shows a direction; name it with the labels the comparison recorded, so a reader can
      # tell which solution a negative difference belongs to without opening the help page.
      cli::cli_verbatim(.efa_wrap_chunks(
        c("Differences:", paste0(x_labels[1L], " - ", x_labels[2L], "."))))
      cli::cli_text("")
      .efa_emit_lines(.compare_diff_lines(diff, digits = digits, r_red = range_red))
    }
  })
}

# Render an efa_compare difference object through the shared matrix renderer into verbatim
# lines, colouring cells whose absolute difference exceeds `r_red` (the efa_compare
# `range_red` threshold) red. A matrix is shown with its variable rows and factor columns; a vector
# becomes a single unlabelled column (one value per row), so its column-header row is dropped.
.compare_diff_lines <- function(diff, digits, r_red) {
  is_vector <- !inherits(diff, "matrix")
  if (is_vector) {
    diff <- matrix(diff, ncol = 1, dimnames = list(names(diff), NULL))
  }

  lines <- .efa_format_matrix(
    values = diff,
    row_labels = .efa_variable_names(diff),
    col_labels = .efa_factor_names(diff),
    col_roles = rep("compare", ncol(diff)),
    cutoff = r_red,
    digits = digits
  )

  if (is_vector) lines[-1L] else lines
}

#' Compare two vectors or matrices (communalities or loadings)
#'
#' The function takes two objects of the same dimensions containing numeric
#' information (loadings or communalities) and returns a list of class
#' `efa_compare` containing summary information of the differences of the objects.
#'
#' @details `digits`, `m_red`, `range_red`, `round_red`, `print_diff`, and `plot_red`
#'  only control how the result is displayed; each is stored in the returned object's
#'  `settings` and can be overridden later without recomputing the comparison --
#'  `digits`, `m_red`, `range_red`, `round_red`, and `print_diff` in a call to
#'  [print.efa_compare()], and `plot_red` in a call to [plot.efa_compare()].
#'
#' @param x matrix, or vector. Loadings or communalities of a factor
#'  analysis output.
#' @param y matrix, or vector. Loadings or communalities of another
#'  factor analysis output to compare to x.
#' @param reorder character. Whether and how elements / columns should be
#' reordered. If "congruence" (default), the columns of `y` are matched to those of
#' `x` by a joint one-to-one assignment that maximizes the total Tucker's congruence
#' coefficient (a standard measure of similarity between two loading vectors) across
#' all columns at once, and each matched column's sign is flipped if needed. This way,
#' mismatched factor order or sign between two solutions does not distort the
#' comparison. It applies to matrices only, and warns when `x`
#' and `y` are vectors. If "names", the columns of a matrix -- or the elements of
#' a vector -- are put in alphabetical order of their names; the rows of a matrix
#' are assumed to be aligned already and are left untouched. If "none", no
#' reordering is done.
#' @param corres logical. Whether factor correspondences should be compared if a
#'  matrix is entered. Default is TRUE.
#' @param thresh numeric. The threshold at or above which a loading is classified as substantial. Default is .3.
#' @param digits numeric. Number of decimals to print in the output. Default is 4.
#' @param m_red numeric. Number above which the mean and median should be printed
#'  in red (i.e., if .001 is used, the mean will be in red if it is larger than
#'  .001, otherwise it will be displayed in green.) Default is .001.
#' @param range_red numeric. Number above which the min and max should be printed
#'  in red (i.e., if .001 is used, min and max will be in red if the max is larger
#'   than .001, otherwise it will be displayed in green). Default is .001. Note that
#'   the color of min also depends on max, that is min will be displayed in the
#'   same color as max.
#' @param round_red numeric. The number of agreeing decimals below which the
#'  report highlights the agreement in red (i.e., if 3 is used, the value is
#'  shown in red when the compared numbers agree to fewer than 3 decimals,
#'  otherwise in green). Default is 3.
#' @param print_diff logical. Whether the difference vector or matrix should be
#'  printed or not. Default is TRUE.
#' @param na.rm logical. Whether NAs should be removed from the difference
#'  summaries and factor-correspondence classifications. With `FALSE`, a missing
#'  loading makes the correspondence counts undefined (`NA`). Default is FALSE.
#' @param x_labels character. A vector of length two containing identifying
#'  labels for the two objects x and y that will be compared. These will be used
#'  as labels on the x-axis of the plot, and to name the direction of the signed
#'  elementwise differences in the printed report (see [print.efa_compare()]).
#'  Default is "x" and "y".
#' @param plot `r lifecycle::badge("superseded")` Accepted and validated, but
#'  without effect; retained for backwards compatibility. The difference plot is
#'  drawn by [plot.efa_compare()]. Default is TRUE.
#' @param plot_red numeric. Threshold above which to plot the absolute differences
#'  in red. Default is .01.
#'
#' @return A list of class `efa_compare` with the following components:
#'
#' \item{diff}{The vector or matrix containing the differences between x and y.}
#' \item{mean_abs_diff}{The mean absolute difference between x and y.}
#' \item{median_abs_diff}{The median absolute difference between x and y.}
#' \item{min_abs_diff}{The minimum absolute difference between x and y.}
#' \item{max_abs_diff}{The maximum absolute difference between x and y.}
#' \item{max_dec}{The maximum number of decimals to which a comparison makes sense.
#'  For example, if x contains only values up to the third decimals, and y is a
#'  normal double, max_dec will be three.}
#' \item{are_equal}{The maximal number of decimals to which all elements of x and y
#'  agree in absolute value. The comparison is on magnitudes, so two elements that
#'  are equal in size but opposite in sign count as agreeing; signed disagreements
#'  are reflected in `diff` and the mean / median / min / max absolute differences.
#'  `0` means the two agree in their integer parts but in no decimal place. `NA`
#'  means there is no agreement at all: either they already differ in their integer
#'  parts, or `na.rm = FALSE` and an element is missing.}
#' \item{diff_corres}{The number of differing variable-to-factor correspondences
#'  between x and y, when only the highest loading is considered. `NA` whenever the
#'  correspondences were not compared: for vector input, for a matrix with a single
#'  column, with `corres = FALSE`, and when a loading is missing under
#'  `na.rm = FALSE`.}
#' \item{diff_corres_cross}{The number of differing variable-to-factor correspondences
#'  between x and y when all loadings `>= thresh` are considered. `NA` under the same
#'  conditions as `diff_corres`.}
#' \item{g}{The root mean squared distance (RMSE) between x and y.}
#' \item{settings}{List of the settings used.}
#'
#' @seealso [efa_fit()] for the solutions being compared, and [efa_procrustes()] to rotate
#'  one solution onto another before comparing.
#'
#' @family factor comparison
#'
#' @export
#'
#' @examples
#' # A type SPSS EFA to mimick the SPSS implementation
#' EFA_SPSS_6 <- efa_fit(test_models$case_11b$cormat, n_factors = 6,
#'                       estimate_control = estimate_control(type = "SPSS"),
#'                       rotate_control = rotate_control(type = "SPSS"))
#'
#' # A type psych EFA to mimick the psych::fa() implementation
#' EFA_psych_6 <- efa_fit(test_models$case_11b$cormat, n_factors = 6,
#'                        estimate_control = estimate_control(type = "psych"),
#'                        rotate_control = rotate_control(type = "psych"))
#'
#' # compare the two
#' efa_compare(EFA_SPSS_6$unrot_loadings, EFA_psych_6$unrot_loadings,
#'             x_labels = c("SPSS", "psych"))
efa_compare <- function(x,
                        y,
                        reorder = c("congruence", "names", "none"),
                        corres = TRUE,
                        thresh = .3,
                        digits = 4,
                        m_red = .001,
                        range_red = .001,
                        round_red = 3,
                        print_diff = TRUE,
                        na.rm = FALSE,
                        x_labels = c("x", "y"),
                        plot = TRUE,
                        plot_red = .01)  {

  reorder <- .match_arg_ci(reorder)
  checkmate::assert_flag(corres)
  checkmate::assert_number(thresh, lower = 0)
  checkmate::assert_count(digits)
  checkmate::assert_number(m_red)
  checkmate::assert_number(range_red)
  checkmate::assert_number(round_red)
  checkmate::assert_flag(print_diff)
  checkmate::assert_flag(na.rm)
  checkmate::assert_character(x_labels, len = 2)
  checkmate::assert_flag(plot)
  checkmate::assert_number(plot_red)

  # reclass data.frames and tibbles to matrices so the stats functions afterwards
  # work
  if ((inherits(x, c("loadings", "LOADINGS", "SLLOADINGS", "matrix"))) &&
      (inherits(y, c("loadings", "LOADINGS", "SLLOADINGS", "matrix")))) {

    if (inherits(x, c("loadings", "LOADINGS", "SLLOADINGS"))) {
      x <- unclass(x)
    }

    if (inherits(y, c("loadings", "LOADINGS", "SLLOADINGS"))) {
      y <- unclass(y)
    }

    if (!is.numeric(x) || !is.numeric(y)) {
      cli::cli_abort(
        "{.arg x} and {.arg y} must be numeric vectors or matrices.",
        class = "efa_compare_bad_input"
      )
    }

    # check if dimensions match:
    if (any(dim(x) != dim(y))) {

      cli::cli_abort("{.arg x} and {.arg y} have different dimensions; {.fun efa_compare} only works with identical dimensions.",
                     class = "efa_compare_dim_mismatch")

    }

  } else if (inherits(x, c("numeric", "integer")) &&
             inherits(y, c("numeric", "integer"))) {

    if (length(x) != length(y)) {

      cli::cli_abort("{.arg x} and {.arg y} have different lengths; {.fun efa_compare} only works with identical dimensions.",
                     class = "efa_compare_dim_mismatch")

    }

  } else {

    # A fitted solution is the commonest thing to hand over here, so name the component that
    # holds its loadings rather than leaving the reader to find it.
    cli::cli_abort(
      c("{.arg x} ({.cls {class(x)}}) and {.arg y} ({.cls {class(y)}}) must be numeric vectors or matrices.",
        "i" = "From an {.cls efa} solution, use {.code $rot_loadings}, or {.code $unrot_loadings}
               if it was fitted without a rotation."),
      class = "efa_compare_bad_input")

  }

  if (length(x) == 0L) {
    cli::cli_abort(
      "{.arg x} and {.arg y} must contain at least one value.",
      class = "efa_compare_empty"
    )
  }

  # Matrices headed for congruence reordering are exempt: that branch reports a
  # non-finite input as a reordering failure, which names the step that actually
  # cannot proceed. The exemption has to cover every matrix the branch handles,
  # single-column ones included, or the same input reports differently depending on
  # how many columns it has.
  if ((any(is.infinite(x)) || any(is.infinite(y))) &&
      !(is.matrix(x) && identical(reorder, "congruence"))) {
    cli::cli_abort(
      "{.arg x} and {.arg y} must not contain infinite values.",
      class = "efa_compare_nonfinite"
    )
  }

  if (inherits(x, "matrix")) {

    n_factors <- ncol(x)

    if (reorder == "congruence") {
      # Match y's columns to those of x with the shared congruence alignment: an
      # optimal one-to-one assignment that maximises the matched absolute Tucker
      # congruences and reflects signs accordingly. The linear assignment
      # guarantees a permutation (a greedy row-wise which.max could send two
      # x-factors to the same y column, duplicating one column and dropping
      # another), and carries x's dimnames onto the realigned y. A single column
      # has nothing to permute but the same arbitrary sign as any other, so it
      # goes through the alignment too: two one-factor solutions that differ only
      # in the sign of their column are the same solution.
      #
      # Tucker's congruence is undefined for missing or non-finite inputs and for
      # zero/near-zero or non-finite congruences. Reject the former up front; map
      # the alignment's classed aborts for the latter to the same reorder error.
      if (!all(is.finite(x)) || !all(is.finite(y))) {
        cli::cli_abort(
          c("{.arg x} or {.arg y} contains missing or non-finite values; cannot reorder columns by congruence.",
            "i" = "Try another reordering method."),
          class = "efa_compare_congruence_na"
        )
      }

      congruence_undefined <- function(e) {
        cli::cli_abort(
          c("Tucker's congruence is undefined for a zero, near-zero, or non-finite column; cannot reorder columns by congruence.",
            "i" = "Try another reordering method."),
          class = "efa_compare_congruence_na"
        )
      }
      y <- tryCatch(
        .align_solution(L_target = x, L = y)$loadings,
        efa_zero_column = congruence_undefined,
        efa_undefined_congruence = congruence_undefined
      )
    } else if (reorder == "names" && n_factors > 1) {

      if (!is.null(colnames(x)) && !is.null(colnames(y))) {
        x <- x[, order(colnames(x))]
        y <- y[, order(colnames(y))]

        if(!all(colnames(x) == colnames(y))) {
          cli::cli_warn("{.code reorder = \"names\"} was used but the colnames of {.arg x} and {.arg y} differ; results might be inaccurate.",
                        class = "efa_compare_reorder_mismatch")
        }
      } else if (is.null(colnames(x)) || is.null(colnames(y))) {
        cli::cli_warn("{.arg reorder} was {.val names} but at least one of {.arg x} and {.arg y} is unnamed; proceeding without reordering.",
                      class = "efa_compare_reorder_unnamed")
      }

    }

  } else if (inherits(x, c("numeric", "integer"))) {

    if (reorder == "congruence"){

      # Congruence reordering needs columns to permute, so it does nothing here.
      # That is true of any vector, named or not, and the default is
      # `"congruence"` -- so the warning does not depend on the names. Only the
      # pointer to the alternative does: `reorder = "names"` is no use to someone
      # whose vectors carry no names.
      msg <- "{.arg reorder} was {.val congruence}, which only works for matrices; proceeding without reordering."
      if (!is.null(names(x)) && !is.null(names(y))) {
        msg <- c(msg, "i" = "To reorder vectors, set {.code reorder = \"names\"}.")
      }
      cli::cli_warn(msg, class = "efa_compare_reorder_vectors")

    } else if (reorder == "names") {

      if (!is.null(names(x)) && !is.null(names(y))) {

        x <- x[order(names(x))]
        y <- y[order(names(y))]

        if (!all(names(x) == names(y))) {
          cli::cli_warn("{.code reorder = \"names\"} was used but the names of {.arg x} and {.arg y} differ; results might be inaccurate.",
                        class = "efa_compare_reorder_mismatch")
        }



      } else if (is.null(names(x)) || is.null(names(y))) {
        cli::cli_warn("{.arg reorder} was {.val names} but at least one of {.arg x} and {.arg y} is unnamed; proceeding without reordering.",
                      class = "efa_compare_reorder_unnamed")
      }

      }

  }

  # delegate the difference statistics and factor correspondences to the
  # printless core; x and y are already coerced and aligned at this point, so the
  # core does not (and must not) reorder them.
  core <- .compare_loadings(x, y, thresh = thresh, na.rm = na.rm, corres = corres)

  settings <- list(
    reorder = reorder,
    corres = corres,
    digits = digits,
    thresh = thresh,
    m_red = m_red,
    range_red = range_red,
    round_red = round_red,
    print_diff = print_diff,
    na.rm = na.rm,
    x_labels = x_labels,
    plot_red = plot_red
  )



  # create output list: the core statistics followed by the recorded settings
  out <- c(core, list(settings = settings))

  # the superseded COMPARE() name keeps its old class string alongside the new one,
  # so `inherits(x, "COMPARE")` in existing code still resolves
  class(out) <- c("efa_compare", "COMPARE")

  out
}


# Difference statistics and factor correspondences for two already coerced and
# aligned loading/communality objects. Kept free of coercion, reordering, and
# printing so a caller that has aligned x and y to a shared target can reuse it
# without triggering a second alignment.
.compare_loadings <- function(x, y, thresh = 0.3, na.rm = FALSE, corres = TRUE,
                              decimals = TRUE) {

  # Factor correspondences need a matrix with more than one column, an enabled
  # comparison, and no missing loading to work around. Wherever one of those is
  # absent the correspondences were not compared, and the result must say so with
  # NA: a count of 0 would read as "they were compared and agreed everywhere".
  if (inherits(x, "matrix")) {
    if (ncol(x) > 1 && isTRUE(corres)) {
      if (!na.rm && (anyNA(x) || anyNA(y))) {
        diff_corres <- NA_integer_
        diff_corres_cross <- NA_integer_
      } else {
        corres_list <- .factor_corres(x, y, thresh = thresh)
        diff_corres <- corres_list$diff_corres
        diff_corres_cross <- corres_list$diff_corres_cross
      }
    } else {
      # a single factor leaves nothing to disagree on, and `corres = FALSE` asked
      # for no comparison at all
      diff_corres <- NA_integer_
      diff_corres_cross <- NA_integer_
    }
  } else {
    # a vector has no factors to correspond to
    diff_corres <- NA_integer_
    diff_corres_cross <- NA_integer_
  }

  # compute differences and statistics
  diff <- x - y

  # RMSE of the elementwise differences. na.rm drops missing entries from both
  # the numerator and the denominator so the statistic stays a proper mean
  # square; the unified form also covers the matrix case (length(diff) is the
  # full element count) without forming t(diff) %*% diff.
  sq <- diff ^ 2
  n_ok <- if (na.rm) sum(!is.na(diff)) else length(diff)
  if (na.rm && n_ok == 0L) {
    g <- mean_abs_diff <- median_abs_diff <- min_abs_diff <- max_abs_diff <- NA_real_
  } else {
    g <- sqrt(sum(sq, na.rm = na.rm) / n_ok)
    mean_abs_diff <- mean(abs(diff), na.rm = na.rm)
    median_abs_diff <- stats::median(abs(diff), na.rm = na.rm)
    min_abs_diff <- min(abs(diff), na.rm = na.rm)
    max_abs_diff <- max(abs(diff), na.rm = na.rm)
  }

  # max_dec: the most decimal places at which a comparison is meaningful (the
  # fewest decimals carried by either input). are_equal: the most decimal places
  # to which every corresponding pair of x and y still agrees. Comparing the
  # truncated values place by place is correct whatever the magnitude of the
  # integer part; concatenating the digit strings would let extra integer digits
  # inflate the count. signif() strips the floating-point representation noise
  # before truncating, so a value such as 0.57 (held as 0.5699999999999999) is
  # not mis-truncated to 0.56 at the second decimal.
  #
  # The two are the only non-trivial work in this function -- a decimal-place scan
  # of every element, then a truncation loop over the decimal places -- and only
  # the printed comparison reports them, so `decimals = FALSE` skips both (leaving
  # them NA) for a caller that needs the difference summaries alone.
  max_dec <- NA_real_
  are_equal <- NA_real_
  if (isTRUE(decimals)) {
    max_dec <- min(c(.decimals(x), .decimals(y)))

    ax <- abs(x)
    ay <- abs(y)
    if (anyNA(diff) && (!na.rm || all(is.na(diff)))) {
      # are_equal is undefined when there is nothing to compare: under na.rm = FALSE
      # any missing element poisons the comparison, and under na.rm = TRUE a fully
      # missing overlap leaves no comparable pair (the loop's all(logical(0)) would
      # otherwise spuriously report agreement to every decimal place). Report NA (the
      # printed "none"), mirroring how mean_abs_diff and g propagate NA, and matching
      # max_dec, which ignores missings.
      are_equal <- NA_real_
    } else {
      # Walk the decimal places from the integer part outwards, stopping at the
      # first one that differs: once a place disagrees no deeper place can count, so
      # there is nothing to gain from testing the rest. The counter starts at NA and
      # is only ever set to a place that was reached, so a comparison that already
      # fails at d = 0 (the integer parts differ) stays NA and is distinguishable
      # from one that succeeds at d = 0 but at no decimal place.
      for (d in 0:max_dec) {
        if (!isTRUE(all(trunc(signif(ax * 10^d, 13)) == trunc(signif(ay * 10^d, 13)),
                        na.rm = na.rm))) {
          break
        }
        are_equal <- as.double(d)
      }
    }
  }

  list(
    diff = diff,
    mean_abs_diff = mean_abs_diff,
    median_abs_diff = median_abs_diff,
    min_abs_diff = min_abs_diff,
    max_abs_diff = max_abs_diff,
    max_dec = max_dec,
    are_equal = are_equal,
    diff_corres = diff_corres,
    diff_corres_cross = diff_corres_cross,
    g = g
  )
}

vec_s <- c("a" = 1, "b" = 2, "c" = 4)
vec_L <- c("A" = 1, "B" = 2, "C" = 4)

mat_s <- matrix(c(0, 0, 0, 1), ncol = 2)
colnames(mat_s) <- c("a", "b")
mat_L <- matrix(c(0, 0, 0, 1), ncol = 2)
colnames(mat_L) <- c("A", "B")

# Vector fixtures ask for no reordering explicitly: the default "congruence" does not
# apply to vectors and warns, which is asserted where it belongs rather than here.
int <- efa_compare(1:10, 1:10, reorder = "none")
dec <- efa_compare(c(.1, .2), c(.1, .1), reorder = "none")
matr <- efa_compare(matrix(c(1,1,1,2), ncol = 2), matrix(c(1,1,1,1), ncol = 2))

SPSS_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                type = "SPSS", method = "PAF", rotation = "none")
psych_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                 type = "psych", method = "PAF", rotation = "none")
load <- efa_compare(SPSS_PAF$unrot_loadings, SPSS_PAF$unrot_loadings)
load_ro1 <- efa_compare(SPSS_PAF$unrot_loadings, SPSS_PAF$unrot_loadings,
                    reorder = "names")
load_ro2 <- efa_compare(SPSS_PAF$unrot_loadings, SPSS_PAF$unrot_loadings,
                    reorder = "none")

SPSS_PAF_1 <- EFA(test_models$baseline$cormat, n_factors = 1, N = 500,
                type = "SPSS", method = "PAF", rotation = "none")
psych_PAF_1 <- EFA(test_models$baseline$cormat, n_factors = 1, N = 500,
                 type = "psych", method = "PAF", rotation = "none")
load_F1 <- efa_compare(SPSS_PAF_1$unrot_loadings, psych_PAF_1$unrot_loadings)

test_that("output class and dimensions are correct", {
  # the old class string is kept alongside the new one, so `inherits(x, "COMPARE")`
  # in code written against the superseded name still resolves
  expect_identical(class(int), c("efa_compare", "COMPARE"))
  expect_identical(class(dec), c("efa_compare", "COMPARE"))
  expect_identical(class(matr), c("efa_compare", "COMPARE"))
  expect_s3_class(load, "efa_compare")
  expect_s3_class(load_ro1, "efa_compare")
  expect_s3_class(load_ro2, "efa_compare")
  expect_s3_class(load_F1, "efa_compare")

  expect_named(int, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                      "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                      "diff_corres_cross", "g", "settings"))
  expect_named(dec, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                      "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                      "diff_corres_cross", "g", "settings"))
  expect_named(matr, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                      "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                      "diff_corres_cross", "g", "settings"))
  expect_named(load, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                       "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                       "diff_corres_cross", "g", "settings"))
  expect_named(load_ro1, c("diff", "mean_abs_diff", "median_abs_diff",
                           "min_abs_diff", "max_abs_diff", "max_dec", "are_equal",
                           "diff_corres", "diff_corres_cross", "g", "settings"))
  expect_named(load_ro2, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                       "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                       "diff_corres_cross", "g", "settings"))
  expect_named(load_F1, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                       "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                       "diff_corres_cross", "g", "settings"))

  expect_type(int$diff, "integer")
  expect_type(int$mean_abs_diff, "double")
  expect_type(int$median_abs_diff, "double")
  expect_type(int$min_abs_diff, "integer")
  expect_type(int$max_abs_diff, "integer")
  expect_type(int$max_dec, "double")
  expect_type(int$are_equal, "double")
  expect_identical(int$diff_corres, NA_integer_)
  expect_identical(int$diff_corres_cross, NA_integer_)
  expect_type(int$g, "double")
  expect_type(int$settings, "list")

  expect_type(dec$diff, "double")
  expect_type(dec$mean_abs_diff, "double")
  expect_type(dec$median_abs_diff, "double")
  expect_type(dec$min_abs_diff, "double")
  expect_type(dec$max_abs_diff, "double")
  expect_type(dec$max_dec, "integer")
  expect_type(dec$are_equal, "double")
  expect_identical(dec$diff_corres, NA_integer_)
  expect_identical(dec$diff_corres_cross, NA_integer_)
  expect_type(dec$g, "double")
  expect_type(dec$settings, "list")

  checkmate::expect_matrix(matr$diff)
  expect_type(matr$mean_abs_diff, "double")
  expect_type(matr$median_abs_diff, "double")
  expect_type(matr$min_abs_diff, "double")
  expect_type(matr$max_abs_diff, "double")
  expect_type(matr$max_dec, "double")
  expect_type(matr$are_equal, "double")
  expect_type(matr$diff_corres, "integer")
  expect_type(matr$diff_corres_cross, "integer")
  expect_type(matr$g, "double")
  expect_type(matr$settings, "list")

  checkmate::expect_matrix(load$diff)
  expect_type(load$mean_abs_diff, "double")
  expect_type(load$median_abs_diff, "double")
  expect_type(load$min_abs_diff, "double")
  expect_type(load$max_abs_diff, "double")
  expect_type(load$max_dec, "integer")
  expect_type(load$are_equal, "double")
  expect_type(load$diff_corres, "integer")
  expect_type(load$diff_corres_cross, "integer")
  expect_type(load$g, "double")
  expect_type(load$settings, "list")

  checkmate::expect_matrix(load_ro1$diff)
  expect_type(load_ro1$mean_abs_diff, "double")
  expect_type(load_ro1$median_abs_diff, "double")
  expect_type(load_ro1$min_abs_diff, "double")
  expect_type(load_ro1$max_abs_diff, "double")
  expect_type(load_ro1$max_dec, "integer")
  expect_type(load_ro1$are_equal, "double")
  expect_type(load_ro1$diff_corres, "integer")
  expect_type(load_ro1$diff_corres_cross, "integer")
  expect_type(load_ro1$g, "double")
  expect_type(load_ro1$settings, "list")

  checkmate::expect_matrix(load_ro2$diff)
  expect_type(load_ro2$mean_abs_diff, "double")
  expect_type(load_ro2$median_abs_diff, "double")
  expect_type(load_ro2$min_abs_diff, "double")
  expect_type(load_ro2$max_abs_diff, "double")
  expect_type(load_ro2$max_dec, "integer")
  expect_type(load_ro2$are_equal, "double")
  expect_type(load_ro2$diff_corres, "integer")
  expect_type(load_ro2$diff_corres_cross, "integer")
  expect_type(load_ro2$g, "double")
  expect_type(load_ro2$settings, "list")

  checkmate::expect_matrix(load_F1$diff)
  expect_type(load_F1$mean_abs_diff, "double")
  expect_type(load_F1$median_abs_diff, "double")
  expect_type(load_F1$min_abs_diff, "double")
  expect_type(load_F1$max_abs_diff, "double")
  expect_type(load_F1$max_dec, "integer")
  expect_type(load_F1$are_equal, "double")
  expect_type(load_F1$diff_corres, "integer")
  expect_type(load_F1$diff_corres_cross, "integer")
  expect_type(load_F1$g, "double")
  expect_type(load_F1$settings, "list")

})

test_that("efa_compare returns the correct values", {
  expect_equal(int$diff, rep(0, 10))
  expect_equal(int$mean_abs_diff, 0)
  expect_equal(int$median_abs_diff, 0)
  expect_equal(int$min_abs_diff, 0)
  expect_equal(int$max_abs_diff, 0)
  expect_equal(int$max_dec, 0)
  expect_equal(int$are_equal, 0)
  expect_equal(int$g, 0)

  expect_equal(dec$diff, c(0, 0.1))
  expect_equal(dec$mean_abs_diff, 0.05)
  expect_equal(dec$median_abs_diff, 0.05)
  expect_equal(dec$min_abs_diff, 0)
  expect_equal(dec$max_abs_diff, .1)
  expect_equal(dec$max_dec, 1)
  expect_equal(dec$are_equal, 0)
  expect_equal(dec$g, 0.0707, tolerance = .01)

  expect_equal(matr$diff, matrix(c(0, 0, 0, 1), ncol = 2))
  expect_equal(matr$mean_abs_diff, 0.25)
  expect_equal(matr$median_abs_diff, 0)
  expect_equal(matr$min_abs_diff, 0)
  expect_equal(matr$max_abs_diff, 1)
  expect_equal(matr$max_dec, 0)
  # 1 vs 2 in one cell: the two differ already in the integer part, so there is no
  # decimal place they agree to -- NA, not the 0 that means "agree to zero decimals".
  expect_true(is.na(matr$are_equal))
  expect_equal(matr$g, 0.5, tolerance = .01)
  expect_equal(matr$diff_corres, 1)
  expect_equal(matr$diff_corres_cross, 0)

  expect_equal(load$diff, matrix(rep(0, 54), ncol = 3,
                                 dimnames = list(c(paste0("V", seq_len(18))),
                                                 c(paste0("F", seq_len(3))))))
  expect_equal(load$mean_abs_diff, 0)
  expect_equal(load$median_abs_diff, 0)
  expect_equal(load$min_abs_diff, 0)
  expect_equal(load$max_abs_diff, 0)
  expect_equal(load$max_dec, 17, tolerance = 1)
  expect_equal(load$are_equal, 17, tolerance = 1)
  expect_equal(load$g, 0)
  expect_equal(load$diff_corres, 0)
  expect_equal(load$diff_corres_cross, 0)

  expect_equal(load_ro1$mean_abs_diff, 0)
  expect_equal(load_ro1$median_abs_diff, 0)
  expect_equal(load_ro1$min_abs_diff, 0)
  expect_equal(load_ro1$max_abs_diff, 0)
  expect_equal(load_ro1$max_dec, 17, tolerance = 1)
  expect_equal(load_ro1$are_equal, 17, tolerance = 1)
  expect_equal(load_ro1$g, 0)
  expect_equal(load_ro1$diff_corres, 0)
  expect_equal(load_ro1$diff_corres_cross, 0)

  expect_equal(load_ro2$mean_abs_diff, 0)
  expect_equal(load_ro2$median_abs_diff, 0)
  expect_equal(load_ro2$min_abs_diff, 0)
  expect_equal(load_ro2$max_abs_diff, 0)
  expect_equal(load_ro2$max_dec, 17, tolerance = 1)
  expect_equal(load_ro2$are_equal, 17, tolerance = 1)
  expect_equal(load_ro2$g, 0)
  expect_equal(load_ro2$diff_corres, 0)
  expect_equal(load_ro2$diff_corres_cross, 0)

  expect_equal(round(load_F1$diff, 4), matrix(rep(0, 18), ncol = 1,
                                              dimnames = list(c(paste0("V",
                                                                       seq_len(18))),
                                                              "F1")))
  expect_equal(round(load_F1$mean_abs_diff, 4), 0)
  expect_equal(round(load_F1$median_abs_diff, 4), 0)
  expect_equal(round(load_F1$min_abs_diff, 4), 0)
  expect_equal(round(load_F1$max_abs_diff, 4), 0)
  expect_equal(load_F1$max_dec, 15)
  expect_equal(load_F1$are_equal, 2)
  expect_equal(round(load_F1$g, 4), 0)
  # a single factor is not compared for correspondences, and says so
  expect_identical(load_F1$diff_corres, NA_integer_)
  expect_identical(load_F1$diff_corres_cross, NA_integer_)

  expect_equal(efa_compare(vec_s, vec_s[c(3, 1, 2)],
                       reorder = "names")$mean_abs_diff, 0)
  expect_equal(efa_compare(mat_s, mat_s[, c(2, 1)],
                       reorder = "names")$mean_abs_diff, 0)
  expect_equal(efa_compare(psych_PAF$unrot_loadings,
                       psych_PAF$unrot_loadings[, c(3, 1, 2)])$mean_abs_diff, 0)
})


test_that("congruence reordering also covers a single column", {
  # A one-factor solution is identified only up to the sign of its column, so two
  # solutions differing only in that sign are the same solution and the congruence
  # alignment -- which handles a single column -- must be applied to them too.
  v <- matrix(c(.8, .7, .6), ncol = 1)

  flipped <- efa_compare(v, -v, reorder = "congruence")
  expect_equal(flipped$max_abs_diff, 0)
  expect_equal(flipped$mean_abs_diff, 0)
  # an already-aligned pair is left as it is
  expect_equal(efa_compare(v, v, reorder = "congruence")$max_abs_diff, 0)
  # and the sign difference is only visible where no reordering was asked for
  expect_equal(efa_compare(v, -v, reorder = "none")$max_abs_diff, 1.6)
})


test_that("factor correspondences report NA whenever they were not compared", {
  m <- matrix(c(.8, .1, .1, .7), ncol = 2)

  # Every route that skips the comparison uses the same sentinel: a count of 0 there
  # would read as "compared, and found to agree everywhere".
  no_corres <- efa_compare(m, m, corres = FALSE)
  expect_identical(no_corres$diff_corres, NA_integer_)
  expect_identical(no_corres$diff_corres_cross, NA_integer_)

  one <- efa_compare(matrix(c(.8, .7, .6), ncol = 1), matrix(c(.8, .7, .6), ncol = 1))
  expect_identical(one$diff_corres, NA_integer_)
  expect_identical(one$diff_corres_cross, NA_integer_)

  vec <- efa_compare(c(.8, .7), c(.7, .6), reorder = "none")
  expect_identical(vec$diff_corres, NA_integer_)
  expect_identical(vec$diff_corres_cross, NA_integer_)

  # ... while a comparison that did run and found no difference still reports 0
  done <- efa_compare(m, m, reorder = "none")
  expect_identical(done$diff_corres, 0L)
  expect_identical(done$diff_corres_cross, 0L)

  # and the report only carries the line where the comparison actually happened
  expect_no_match(paste(cli::ansi_strip(format(no_corres)), collapse = " "),
                  "correspondences", fixed = TRUE)
  expect_match(paste(cli::ansi_strip(format(done)), collapse = " "),
               "correspondences", fixed = TRUE)
})


test_that("vectors are told that congruence reordering does not apply to them", {
  # The lengths compared below are of rendered cli output, so the width, colour, and
  # unicode state have to be fixed for the comparison to mean anything.
  local_reproducible_output()

  # Vectors have no columns to permute, so the default reordering never runs. That
  # is true whether or not they are named; only the pointer to reorder = "names"
  # depends on names being available, since it is no use without them.
  expect_warning(efa_compare(1:3, 1:3), class = "efa_compare_reorder_vectors")
  expect_warning(efa_compare(vec_s, vec_s), class = "efa_compare_reorder_vectors")

  unnamed <- tryCatch(efa_compare(1:3, 1:3), warning = identity)
  named <- tryCatch(efa_compare(vec_s, vec_s), warning = identity)
  expect_lt(nchar(unnamed$message), nchar(named$message))
})


test_that("reorder = 'names' orders matrix columns and leaves the rows alone", {
  # Documented contract: names order the columns of a matrix (or the elements of a
  # vector); matrix rows are assumed to be aligned already.
  A <- matrix(1:6, 3, 2, dimnames = list(c("r1", "r2", "r3"), c("F2", "F1")))
  B <- matrix(1:6, 3, 2, dimnames = list(c("r3", "r2", "r1"), c("F1", "F2")))

  expect_equal(
    efa_compare(A, B, reorder = "names")$diff,
    matrix(c(3, 3, 3, -3, -3, -3), 3, 2,
           dimnames = list(c("r1", "r2", "r3"), c("F1", "F2")))
  )

  # for a vector the elements themselves are reordered, so the same labels line up
  expect_equal(efa_compare(vec_s, vec_s[c(3, 1, 2)], reorder = "names")$max_abs_diff, 0)
})


test_that("are_equal counts decimal agreement without floating-point miscounts", {
  # 0.57 is held as 0.5699999999999999; a plain trunc(x * 100) reads its second
  # decimal as 6 and would wrongly report agreement with 0.5699999 to two
  # decimals. They agree only to the first decimal.
  expect_equal(efa_compare(c(0.57, 0.1), c(0.5699999, 0.1),
                       reorder = "none")$are_equal, 1)
  # 0.6285 * 1e4 underflows to 6284.999...; a plain trunc drops a place and would
  # under-report. 0.6285 and 0.62851 agree to four decimals.
  expect_equal(efa_compare(c(0.6285, 0.1), c(0.62851, 0.1),
                       reorder = "none")$are_equal, 4)
})


test_that("are_equal separates disagreeing integer parts from zero-decimal agreement", {
  # Both pairs disagree from the first decimal on, but only the first pair already
  # disagrees in its integer part; are_equal must not report the same value for the two.
  expect_true(is.na(efa_compare(c(1.5, 2.5, 3.5), c(9.5, 8.5, 7.5),
                                reorder = "none")$are_equal))
  expect_equal(efa_compare(c(0.5, 0.5, 0.5), c(0.9, 0.9, 0.9),
                           reorder = "none")$are_equal, 0)
})


test_that("congruence reordering yields a true permutation when columns collide", {
  # x has orthonormal factors; y is built so that the greedy row-wise which.max
  # match assigns y-column 1 to BOTH x-factor 1 and x-factor 2 (a collision that
  # would duplicate one y column and drop another). The optimal one-to-one
  # assignment is the permutation c(1, 3, 2).
  x <- diag(3)
  y <- matrix(c(0.7, 0.7, 0.10,    # aligned with factors 1 and 2
                0.2, 0.1, 0.97,    # aligned with factor 3
                0.1, 0.2, 0.95),   # aligned with factor 3
              ncol = 3)

  cmp <- efa_compare(x, y, reorder = "congruence", corres = FALSE)

  # y as reordered (and sign-matched) inside efa_compare
  y_reordered <- x - cmp$diff

  # index in y that each reordered column came from
  recovered <- vapply(seq_len(ncol(y_reordered)), function(j) {
    which(vapply(seq_len(ncol(y)), function(k)
      isTRUE(all.equal(y_reordered[, j], y[, k])), logical(1)))
  }, integer(1))

  # a genuine permutation: every factor used exactly once, never duplicated or
  # dropped (the greedy which.max produced c(1, 1, 2))
  expect_setequal(recovered, seq_len(ncol(y)))
  expect_equal(recovered, c(1L, 3L, 2L))
})


test_that("errors etc. are thrown correctly", {
  expect_error(efa_compare(c(1, 2), 1), class = "efa_compare_dim_mismatch")
  expect_error(efa_compare(numeric(), numeric(), reorder = "none"),
               class = "efa_compare_empty")
  expect_error(efa_compare(c(1, 2), c("1", "2")), class = "efa_compare_bad_input")
  expect_error(efa_compare(c(1, 2), data.frame(x = "1", y = "2")), class = "efa_compare_bad_input")
  expect_error(efa_compare(matrix("x"), matrix("x"), reorder = "none"),
               class = "efa_compare_bad_input")
  expect_error(efa_compare(c(1, Inf), c(1, 2), reorder = "none"),
               class = "efa_compare_nonfinite")
  # a matrix bound for congruence reordering reports the non-finite value as a
  # reordering failure instead (pinned below); every other route reports it here
  expect_error(efa_compare(matrix(Inf), matrix(Inf), reorder = "none"),
               class = "efa_compare_nonfinite")
  expect_error(efa_compare(c(1, 2), c(1, 2), thresh = -0.1), "not >= 0")

  expect_error(efa_compare(matrix(c(0, 0, 0, 1), ncol = 2),
                       matrix(c(0, 0, 0, 1), ncol = 1)), class = "efa_compare_dim_mismatch")

  expect_warning(efa_compare(vec_s, vec_s), class = "efa_compare_reorder_vectors")
  expect_warning(efa_compare(vec_s, vec_L, reorder = "names"), class = "efa_compare_reorder_mismatch")
  expect_warning(efa_compare(vec_s, 1:3, reorder = "names"), class = "efa_compare_reorder_unnamed")
  expect_warning(efa_compare(1:3, 1:3, reorder = "names"), class = "efa_compare_reorder_unnamed")

  expect_warning(efa_compare(matrix(c(0, 0, 0, 1), ncol = 2),
                         matrix(c(0, 0, 0, 1), ncol = 2),
                         reorder = "names"), class = "efa_compare_reorder_unnamed")
  expect_warning(efa_compare(mat_s, mat_L, reorder = "names"), class = "efa_compare_reorder_mismatch")

  expect_error(efa_compare(mat_s, mat_s), class = "efa_compare_congruence_na")

  # A column whose norm is below sqrt(.Machine$double.eps) but non-zero makes
  # Tucker's congruence undefined just as an exactly-zero column does; it must
  # surface the reorder error, not a low-level zero-column abort.
  nz <- matrix(c(1e-9, 1e-9, 0.8, 0.6), ncol = 2)
  expect_error(efa_compare(nz, nz, reorder = "congruence", corres = FALSE),
               class = "efa_compare_congruence_na")

  # Non-finite inputs, and finite values whose squared magnitudes overflow,
  # both leave Tucker's congruence undefined; each must surface the same reorder
  # error rather than a non-finite/solver abort from the alignment internals.
  m_inf <- matrix(c(Inf, 0.5, 0.5, 0.8, 0.6, 0.7), ncol = 2)
  expect_error(efa_compare(m_inf, m_inf, reorder = "congruence", corres = FALSE),
               class = "efa_compare_congruence_na")
  # ... and a single column takes the same branch, so it must report the same way
  # rather than falling through to the generic non-finite abort
  expect_error(efa_compare(m_inf[, 1, drop = FALSE], m_inf[, 1, drop = FALSE],
                           reorder = "congruence", corres = FALSE),
               class = "efa_compare_congruence_na")
  m_big <- matrix(c(1e200, 1e200, 1e200, 0.8, 0.6, 0.7), ncol = 2)
  expect_error(efa_compare(m_big, m_big, reorder = "congruence", corres = FALSE),
               class = "efa_compare_congruence_na")

})

test_that("the input error names where a fitted solution keeps its loadings", {
  # A whole efa_fit() object is the commonest thing to hand over here. The error listed the
  # classes it accepts and left the reader to find the loading table inside the object.
  fit <- suppressWarnings(suppressMessages(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "promax")))

  e <- tryCatch(efa_compare(fit, fit), efa_compare_bad_input = function(e) e)
  expect_s3_class(e, "efa_compare_bad_input")
  expect_snapshot(cat(conditionMessage(e)))

  # The component the message names is accepted.
  expect_s3_class(efa_compare(fit$rot_loadings, fit$rot_loadings, plot = FALSE),
                  "efa_compare")
})

test_that("efa_compare is NA-safe and honours na.rm (vector and matrix)", {
  # NA-containing input used to crash in .decimals ("missing value where
  # TRUE/FALSE needed") before efa_compare could return; the documented na.rm
  # argument must work for both vector and matrix input (reorder = "none"
  # avoids the congruence path, which aborts earlier on NAs).
  v_na <- c(1, 2, NA, 4)
  v_ok <- c(1, 2, 3, 4)
  expect_s3_class(efa_compare(v_na, v_ok, reorder = "none", na.rm = TRUE), "efa_compare")
  expect_s3_class(efa_compare(v_na, v_ok, reorder = "none", na.rm = FALSE), "efa_compare")

  m_na <- matrix(c(1, NA, 1, 2), ncol = 2)
  m_ok <- matrix(c(1, 1, 1, 1), ncol = 2)
  expect_s3_class(efa_compare(m_na, m_ok, reorder = "none", na.rm = TRUE), "efa_compare")
  expect_s3_class(efa_compare(m_na, m_ok, reorder = "none", na.rm = FALSE), "efa_compare")

  # na.rm = TRUE drops the missing entry from the RMSE statistic rather than
  # propagating NA into it.
  cmp_rm <- efa_compare(v_na, v_ok, reorder = "none", na.rm = TRUE)
  expect_false(is.na(cmp_rm$g))
  expect_equal(cmp_rm$g, 0)

  # are_equal mirrors the other statistics under missingness: NA (printed "none")
  # when na.rm = FALSE and an element is missing, and a number when na.rm = TRUE
  # (consistent with max_dec, which always drops missings).
  expect_true(is.na(efa_compare(c(0.5, NA), c(0.5, 0.5),
                            reorder = "none", na.rm = FALSE)$are_equal))
  expect_equal(efa_compare(c(0.5, NA), c(0.5, 0.5),
                       reorder = "none", na.rm = TRUE)$are_equal, 1)

  # Even with na.rm = TRUE, if every pair has a missing value there is nothing to
  # compare, so are_equal is NA rather than spuriously climbing to max_dec (the
  # empty all() would otherwise report agreement to every decimal place). The
  # all-missing overlap also makes min()/max() warn ("no non-missing arguments"),
  # which is unrelated to are_equal, so it is suppressed here.
  expect_true(is.na(suppressWarnings(
    efa_compare(c(1.234, NA), c(NA, 5.678), reorder = "none", na.rm = TRUE))$are_equal))

  # With no comparable pair all numerical summaries are undefined, not NaN/Inf,
  # and the print method still produces a report.
  all_missing <- efa_compare(c(1.234, NA), c(NA, 5.678),
                             reorder = "none", na.rm = TRUE)
  expect_true(all(is.na(unlist(all_missing[c("mean_abs_diff", "median_abs_diff",
                                              "min_abs_diff", "max_abs_diff", "g")]))))
  expect_silent(capture.output(print(all_missing)))

  # Missing cells are ignored for correspondences only when requested; retaining
  # them leaves the correspondence comparison undefined.
  mx <- rbind(c(.8, NA), c(.1, .7))
  my <- rbind(c(.8, .1), c(.1, .7))
  cmp_keep <- efa_compare(mx, my, reorder = "none", na.rm = FALSE)
  cmp_drop <- efa_compare(mx, my, reorder = "none", na.rm = TRUE)
  expect_true(is.na(cmp_keep$diff_corres))
  expect_identical(cmp_drop$diff_corres, 0L)
  expect_identical(cmp_drop$diff_corres_cross, 0L)

  # Near-zero values render in scientific notation, which used to crash the
  # decimal-count helper that drives max_dec; efa_compare must still return.
  expect_s3_class(efa_compare(c(0.5, 0.00003), c(0.5, 0.4), reorder = "none"),
                  "efa_compare")
})

test_that("print output is stable", {
  local_reproducible_output()

  # matrix difference (variables x factors). matr/int hold exact 0/1 differences, so the
  # snapshots are recorded literally (no scrub) to pin the column alignment.
  expect_snapshot(print(matr))

  # vector difference (rendered as a single column, one value per row)
  expect_snapshot(print(int))
})

test_that("the signed difference table names its direction", {
  local_reproducible_output()

  # Unlabelled: the argument names are the stand-in, and they already say which way the
  # subtraction runs.
  expect_true(any(grepl("Differences: x - y.", cli::ansi_strip(format(matr)), fixed = TRUE)))

  # Labelled: the labels efa_compare() recorded reach the report, not just the plot.
  labelled <- efa_compare(matrix(c(1, 1, 1, 2), ncol = 2),
                          matrix(c(1, 1, 1, 1), ncol = 2),
                          x_labels = c("SPSS", "psych"))
  out <- cli::ansi_strip(format(labelled))
  expect_true(any(grepl("Differences: SPSS - psych.", out, fixed = TRUE)))
  # it heads the signed table, and only that: the absolute summary block is untouched
  expect_lt(grep("Elementwise differences", out)[1L],
            grep("Differences: SPSS", out, fixed = TRUE)[1L])
  expect_gt(grep("Differences: SPSS", out, fixed = TRUE)[1L],
            grep("Median absolute difference", out)[1L])

  # ... and it is not emitted when the table it heads is suppressed
  expect_false(any(grepl("Differences: SPSS", cli::ansi_strip(format(labelled, print_diff = FALSE)),
                         fixed = TRUE)))
})

test_that("the correspondence line follows the console width", {
  cmp <- efa_compare(matrix(c(1, 1, 1, 2), ncol = 2), matrix(c(1, 1, 1, 1), ncol = 2))

  counts <- function(lines) wrapped_item(lines, "^Differing")

  for (w in c(120L, 80L, 60L)) {
    out <- withr::with_options(list(cli.width = w, cli.num_colors = 1),
                               cli::ansi_strip(format(cmp)))
    lines <- counts(out)
    expect_true(all(cli::ansi_nchar(lines, type = "width") <= w))
    # wrapping moves the line break only; both counts and their qualifiers survive intact
    expect_identical(
      paste(trimws(lines), collapse = " "),
      paste0("Differing indicator-to-factor correspondences: ", cmp$diff_corres,
             " (highest loading), ", cmp$diff_corres_cross, " (all |loadings| >= 0.3)"))
  }

  # the counts keep their conditional colouring once the line is packed
  withr::local_options(cli.width = 60, cli.num_colors = 256)
  expect_true(cli::ansi_has_any(paste(counts(format(cmp)), collapse = "")))
})

test_that("display settings can be overridden at print time", {
  local_reproducible_output()

  # The recorded settings are unchanged; only this call's report differs.
  plain <- cli::ansi_strip(format(matr))
  two <- cli::ansi_strip(format(matr, digits = 2))
  # ".2500" is a substring of "0.2500", so the leading zero has to be excluded
  # explicitly: the statistics follow the package's number convention and drop it.
  expect_match(paste(plain, collapse = " "), ".2500", fixed = TRUE)
  expect_no_match(paste(plain, collapse = " "), "0.2500", fixed = TRUE)
  expect_match(paste(two, collapse = " "), ".25", fixed = TRUE)
  expect_no_match(paste(two, collapse = " "), ".2500", fixed = TRUE)
  expect_equal(matr$settings$digits, 4)

  # print() forwards its dots to format(), so the same override works there.
  expect_identical(cli::ansi_strip(utils::capture.output(print(matr, digits = 2))), two)

  # print_diff drops the difference table.
  expect_true(any(grepl("Elementwise differences", plain, fixed = TRUE)))
  expect_false(any(grepl("Elementwise differences",
                         cli::ansi_strip(format(matr, print_diff = FALSE)), fixed = TRUE)))

  # The three colouring thresholds change only the green/red styling, so they are
  # observable solely with colours on: each must flip the ANSI of its own line while
  # leaving the reported numbers untouched. matr's mean and max are .25 and 1, so a
  # threshold on either side of those is decisive; round_red is read against
  # are_equal, which is NA for matr (always red) and 0 for int.
  old <- options(cli.num_colors = 256)
  on.exit(options(old), add = TRUE)
  expect_false(identical(format(matr, m_red = 1), format(matr, m_red = 0)))
  expect_false(identical(format(matr, range_red = 2), format(matr, range_red = 0)))
  expect_false(identical(format(int, round_red = 0), format(int, round_red = 9)))
  # ... and none of them alters the numbers themselves.
  strip_matr <- function(...) cli::ansi_strip(format(matr, ...))
  expect_identical(strip_matr(m_red = 1), strip_matr(m_red = 0))
  expect_identical(strip_matr(range_red = 2), strip_matr(range_red = 0))
  options(cli.num_colors = 1)

  # An override is validated exactly as efa_compare() validates the argument.
  expect_error(format(matr, digits = -1))

  # omitting an argument leaves the recorded value in force
  expect_identical(format(matr), format(matr, digits = matr$settings$digits))
})

test_that("the minimum-decimals line is dropped when it carries no information", {
  local_reproducible_output()

  # Two ordinary doubles carry full double precision, so the count says nothing.
  full <- efa_compare(c(1/3, 2/7, 1/6), c(1/3, 2/7, 1/6 + 1e-4), reorder = "none")
  expect_gte(full$max_dec, 15)
  expect_no_match(paste(cli::ansi_strip(format(full)), collapse = " "),
                  "Minimum number of decimals provided", fixed = TRUE)

  # A rounded input bounds the comparison, so the line is informative and shown.
  expect_match(paste(cli::ansi_strip(format(dec)), collapse = " "),
               "Minimum number of decimals provided", fixed = TRUE)
})

test_that("plot returns a ggplot and guards too-few differences", {
  # Smoke-test only (no vdiffr baseline) because geom_jitter draws random positions,
  # so the rendered plot is not reproducible across runs.
  p <- plot(matr)
  expect_s3_class(p, "ggplot")

  # plot_red is a drawing setting, overridable per call without recomputing: it
  # decides which differences are flagged and is reported in the subtitle.
  expect_true("large difference" %in% p$data$color)
  p2 <- plot(matr, plot_red = 2)
  expect_setequal(p2$data$color, "acceptable difference")
  expect_match(p2$labels$subtitle, "difference coloring: 2", fixed = TRUE)
  expect_equal(matr$settings$plot_red, 0.01)

  # too few differences to plot
  expect_error(plot(dec), class = "efa_compare_too_few_to_plot")
})


# Reference implementation of the statistics and factor correspondences that
# .compare_loadings() returns. The test below pins the extracted core against
# accidental drift: it mirrors the helper's computation (a regression guard) rather
# than deriving it independently, so the statistic *values* are anchored by the
# value-based efa_compare tests above. Only the vector NA path is written differently
# here (an explicit numeric/integer test, not a catch-all else); the two agree for
# the matrix/numeric/integer inputs exercised.
ref_compare_loadings <- function(x, y, thresh = 0.3, na.rm = FALSE, corres = TRUE) {

  if (inherits(x, "matrix")) {
    if (ncol(x) > 1 && isTRUE(corres)) {
      if (!na.rm && (anyNA(x) || anyNA(y))) {
        diff_corres <- NA_integer_
        diff_corres_cross <- NA_integer_
      } else {
        corres_list <- .factor_corres(x, y, thresh = thresh)
        diff_corres <- corres_list$diff_corres
        diff_corres_cross <- corres_list$diff_corres_cross
      }
    } else {
      diff_corres <- NA_integer_
      diff_corres_cross <- NA_integer_
    }
  } else if (inherits(x, c("numeric", "integer"))) {
    diff_corres <- NA_integer_
    diff_corres_cross <- NA_integer_
  }

  diff <- x - y

  sq <- diff ^ 2
  n_ok <- if (na.rm) sum(!is.na(diff)) else length(diff)
  if (na.rm && n_ok == 0L) {
    g <- mean_abs_diff <- median_abs_diff <- min_abs_diff <- max_abs_diff <- NA_real_
  } else {
    g <- sqrt(sum(sq, na.rm = na.rm) / n_ok)
    mean_abs_diff <- mean(abs(diff), na.rm = na.rm)
    median_abs_diff <- stats::median(abs(diff), na.rm = na.rm)
    min_abs_diff <- min(abs(diff), na.rm = na.rm)
    max_abs_diff <- max(abs(diff), na.rm = na.rm)
  }

  max_dec <- min(c(.decimals(x), .decimals(y)))

  ax <- abs(x)
  ay <- abs(y)
  if (anyNA(diff) && (!na.rm || all(is.na(diff)))) {
    are_equal <- NA_real_
  } else {
    are_equal <- NA_real_
    for (d in 0:max_dec) {
      if (!isTRUE(all(trunc(signif(ax * 10^d, 13)) == trunc(signif(ay * 10^d, 13)),
                      na.rm = na.rm))) {
        break
      }
      are_equal <- as.double(d)
    }
  }

  list(
    diff = diff,
    mean_abs_diff = mean_abs_diff,
    median_abs_diff = median_abs_diff,
    min_abs_diff = min_abs_diff,
    max_abs_diff = max_abs_diff,
    max_dec = max_dec,
    are_equal = are_equal,
    diff_corres = diff_corres,
    diff_corres_cross = diff_corres_cross,
    g = g
  )
}

test_that(".compare_loadings reproduces efa_compare's pre-extraction computation", {
  set_dn <- function(m) {
    colnames(m) <- paste0("F", seq_len(ncol(m)))
    rownames(m) <- paste0("V", seq_len(nrow(m)))
    m
  }

  m2x <- set_dn(matrix(c(0.8, 0.1, 0.2, 0.7, 0.15, 0.75, 0.1, 0.05), ncol = 2))
  m2y <- set_dn(matrix(c(0.75, 0.12, 0.25, 0.68, 0.2, 0.7, 0.08, 0.02), ncol = 2))
  m1x <- set_dn(matrix(c(0.5, 0.6, 0.7, 0.8), ncol = 1))
  m1y <- set_dn(matrix(c(0.52, 0.58, 0.72, 0.79), ncol = 1))
  # signed disagreement (0.57 vs -0.5699999) and float-noise magnitudes exercise
  # are_equal / max_dec.
  vx  <- c(a = 0.57, b = 0.6285, c = 0.1)
  vy  <- c(a = -0.5699999, b = 0.62851, c = 0.1)
  vix <- 1:6
  viy <- c(1L, 2L, 3L, 4L, 5L, 7L)
  # NA-containing matrix, corres kept on to mirror efa_compare's default path.
  mnax <- set_dn(matrix(c(0.8, NA, 0.2, 0.7, 0.15, 0.75, NA, 0.05), ncol = 2))
  mnay <- set_dn(matrix(c(0.75, 0.12, 0.25, 0.68, 0.2, 0.7, 0.08, 0.02), ncol = 2))
  vnax <- c(0.5, NA, 0.30003, 0.4)
  vnay <- c(0.5, 0.5, 0.3, 0.4)
  # near-zero value renders in scientific notation (drives the .decimals path).
  scix <- c(0.5, 0.00003, 0.4)
  sciy <- c(0.5, 0.4, 0.4)
  # every pair missing under na.rm = TRUE -> all summaries are NA.
  allmiss_x <- c(1.234, NA)
  allmiss_y <- c(NA, 5.678)

  cases <- list(
    list(x = m2x, y = m2y, corres = TRUE),
    list(x = m2x, y = m2y, corres = FALSE),
    list(x = m2x, y = m2y, corres = TRUE, thresh = 0.5),
    list(x = m1x, y = m1y),
    list(x = vx, y = vy),
    list(x = vix, y = viy),
    list(x = mnax, y = mnay, na.rm = FALSE),
    list(x = mnax, y = mnay, na.rm = TRUE),
    list(x = vnax, y = vnay, na.rm = FALSE),
    list(x = vnax, y = vnay, na.rm = TRUE),
    list(x = scix, y = sciy),
    list(x = allmiss_x, y = allmiss_y, na.rm = TRUE)
  )

  for (cs in cases) {
    got <- do.call(.compare_loadings, cs)
    ref <- do.call(ref_compare_loadings, cs)
    expect_identical(got, ref)
  }
})

rm(int, dec, matr, SPSS_PAF, psych_PAF, load, load_ro1, load_ro2, SPSS_PAF_1,
   psych_PAF_1, load_F1, vec_s, vec_L, mat_s, mat_L)

# Number of factors a sequential SMT rule suggests. `values` runs over the 0, 1, ...,
# max_fac factor models and `rule` maps that series to a logical "this model satisfies
# the rule" (taken as a function rather than a second vector so the series cannot be
# passed twice inconsistently). The search stops at the first model that satisfies the
# rule, or at the first undefined value: an NA breaks the strictly sequential test, so
# the search stops there rather than skipping ahead to a later model that satisfies it,
# and no number can be suggested. Shared by the chi-square and the RMSEA rule so the
# two cannot drift apart.
.smt_sequential_stop <- function(values, rule) {
  stop_at <- which(rule(values) | is.na(values))[1]
  if (is.na(stop_at) || is.na(values[stop_at])) NA_integer_ else stop_at - 1L
}

#' Sequential chi square model tests, RMSEA lower bound, and AIC
#'
#' Sequential chi square model tests (SMT) are a factor retention method where
#' multiple
#' EFAs with increasing numbers of factors are fitted and the number of factors
#' for which the Chi Square value first becomes non-significant is taken as the
#' suggested number of factors.
#' Preacher, Zhang, Kim, & Mels (2013) suggested a similar approach with the
#' lower bound of the 90% confidence interval of the Root Mean Square Error of
#' Approximation (RMSEA; Browne & Cudeck, 1992; Steiger & Lind, 1980), and with
#' the Akaike Information Criterion (AIC). For the RMSEA, the
#' number of factors for which this lower bound first falls below .05 is the
#' suggested number of factors to retain. For the AIC, it is the number of factors
#' where the AIC is lowest.
#'
#' @inheritParams efa_kgc
#' @param cor_method character. One of `"pearson"`, `"spearman"`, or `"kendall"`,
#'   passed to [stats::cor()]. `"poly"` and `"tetra"` are not supported because
#'   `SMT` rests on a normal-theory chi-square test that is not valid for
#'   polychoric / tetrachoric correlations.
#'  Default is `"pearson"`.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used. Must be larger than the number of variables.
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  sequential [efa_fit()] fits. `NULL` (default) uses the [efa_fit()] defaults. The sequential
#'  models are fitted with maximum likelihood (the chi-square, RMSEA, and AIC the SMT is built on
#'  are defined for it), so of the estimation knobs only `start_method` takes effect; the ones
#'  governing principal axis factoring do not apply. The models are unrotated, so no rotation
#'  settings apply either.
#'
#' @details
#' As a first step in the procedure, a maximum number of factors to extract is
#' determined for which the model is still over-identified (df > 0).
#'
#' Then, EFAs with increasing numbers of factors from 1 to the maximum number are
#' fitted with maximum likelihood estimation.
#'
#' For the SMT, first the significance of the chi
#' square value for a model with 0 factors is determined. If this value is
#' not significant, 0 factors are suggested to retain. If it is significant,
#' a model with 1 factor is estimated and the significance of its chi square value
#' is determined, and so on, until a non-significant result is obtained. The
#' suggested number of factors is the number of factors for the model where the
#' chi square value first becomes non-significant.
#'
#' Regarding the RMSEA, the suggested number of factors is the number of factors
#' for the model where the lower bound of the 90% confidence interval of the
#' RMSEA first falls below the .05 threshold.
#'
#' Regarding the AIC, the suggested number of factors is the number of factors
#' for the model with the lowest AIC.
#'
#' The sequential models are fitted without inequality constraints, so a solution
#' can be inadmissible (a Heywood case, or a fit that did not converge). Only the
#' models the three rules actually select are checked for this; if one of them is
#' inadmissible a warning is raised and the corresponding suggestion should be
#' interpreted with caution.
#'
#' In comparison with other prominent factor retention criteria, SMT performed
#' well at determining the number of factors to extract in EFA (Auerswald &
#' Moshagen, 2019). The RMSEA lower bound also performed well at determining the true
#' number of factors, while the AIC performed well at determining the
#' most generalizable model (Preacher, Zhang, Kim, & Mels, 2013).
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] for
#'   the print method). SMT has no plot; [plot.efa_retention()] returns `NULL`
#'   with a message for it. Its main fields are:
#' \item{n_factors}{A named numeric vector (`"chi"`, `"RMSEA"`, `"AIC"`) with the
#'   suggested number of factors from the sequential chi-square model tests, the
#'   RMSEA lower bound, and the AIC.}
#' \item{results}{A list with one record per criterion, each holding the criterion
#'   values for the null model (zero factors) through the maximum number of
#'   factors.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#' @source Browne, M.W., & Cudeck, R. (1992). Alternative ways of assessing model
#' fit. Sociological Methods and Research, 21, 230–258.
#' @source Preacher, K. J., Zhang G., Kim, C., & Mels, G. (2013). Choosing the
#' Optimal Number of Factors in Exploratory Factor Analysis: A Model Selection
#' Perspective, Multivariate Behavioral Research, 48(1), 28-56,
#' doi:10.1080/00273171.2012.710386
#' @source Steiger, J. H., & Lind, J. C. (1980, May). Statistically based tests
#' for the number of common factors. Paper presented at the annual meeting of
#' the Psychometric Society, Iowa City, IA.
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#' @examples
#' SMT_base <- efa_smt(test_models$baseline$cormat, N = 500)
#' SMT_base
#'
efa_smt <- function(x, N = NA, use = c("pairwise.complete.obs", "all.obs",
                                     "complete.obs", "everything",
                                     "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                estimate_control = NULL){

  # Perform argument checks
  .assert_cor_input(x)

  checkmate::assert_count(N, na.ok = TRUE)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  .reject_poly_reference(
    cor_method, "efa_smt",
    why = "{.fn {fn}} rests on a normal-theory chi-square test that is not valid for such correlations.")
  .assert_estimate_control(estimate_control)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(
    x, N = N, use = use, cor_method = cor_method, N_policy = "required",
    N_required_msg = c("{.arg N} is {.val NA}.",
                       "i" = "Provide {.arg N} or raw data."))
  R <- prep$R
  N <- prep$N
  .assert_n_gt_vars(N, ncol(R))

  # Prepare objects for sequential tests
  max_fac <- .det_max_factors(ncol(R))

  if(max_fac <= 0){
    cli::cli_abort(
      c("The model is underidentified, or just identified with a single factor; SMTs cannot be performed.",
        "i" = "Provide more indicators."),
      class = "efa_smt_underidentified"
    )
  }

  ps <- vector("double", max_fac)
  RMSEA_LB <- vector("double", max_fac)
  AIC <- vector("double", max_fac)
  # admissibility of each fitted model; only the selected ones are reported on below
  n_heywood <- vector("integer", max_fac)
  converged <- vector("logical", max_fac)

    # sequentially perform EFAs with 1 to the maximum number of factors
    for (i in seq_len(max_fac)) {

      temp <- suppressWarnings(suppressMessages(
        efa_fit(R, n_factors = i, estimator = "ML", rotation = "none", N = N,
                estimate_control = estimate_control)))
      ps[i] <- temp$fit_indices$p_chi
      RMSEA_LB[i] <- temp$fit_indices$RMSEA_LB
      AIC[i] <- temp$fit_indices$AIC
      n_heywood[i] <- length(temp$heywood)
      converged[i] <- !isTRUE(temp$convergence != 0)

    }

  # With which number of factors does the chi square first become
  # non-significant?

  # Null-model (zero-factor) statistics. These depend only on R and N (the
  # model-implied matrix is the identity), so compute them directly rather than
  # reading them off the last fitted EFA, whose fit_indices come back NA when
  # that model is degenerate (e.g. a Heywood / non-positive-definite case).
  # Mirrors the null-model block in .gof().
  m <- ncol(R)
  # Reuse one log-determinant of R for both null chi-squares (the Bartlett-corrected statistic
  # below and the uncorrected RMSEA baseline further down) instead of factorising R twice.
  ld_R <- determinant(R, logarithm = TRUE)
  chi_null <- .null_chisq(R, N, ld = ld_R)
  df_null <- (m^2 - m) / 2
  p_null <- stats::pchisq(chi_null, df_null, lower.tail = FALSE)

  # Retain the number of factors of the first model whose chi square is
  # non-significant. An NA p (e.g. a non-converged / Heywood model) stops the search
  # with no suggestion, as does a sequence in which every model is significant.
  p_seq <- c(p_null, ps)
  nfac_chi <- .smt_sequential_stop(p_seq, \(p) p > 0.05)

  # Calculate RMSEA (incl. lower bound of 90% CI) and AIC for the null model. The
  # RMSEA noncentrality is built on the uncorrected (N - 1) discrepancy scale (as in
  # .chi_fit_indices()), so the null model uses the uncorrected baseline chi-square;
  # p_null and AIC_null keep the Bartlett-corrected statistic computed above. When N is
  # too small for the Bartlett correction (chi_null is NA) the null-model asymptotics
  # break down, so leave the RMSEA bound NA too, matching how .gof() drops the
  # fitted-model fit indices at such N.
  chi_null_rmsea <- .null_chisq(R, N, ld = ld_R, corrected = FALSE)
  RMSEA_LB_null <- if (is.na(chi_null)) {
    NA_real_
  } else {
    sqrt(.rmsea_lambda(chi_null_rmsea, df_null, .95) / (df_null * (N - 1)))
  }

  AIC_null <- chi_null - 2 * df_null

  # Retain the number of factors of the first model whose RMSEA lower bound falls
  # below .05, under the same sequential convention as the chi-square rule above.
  RMSEA_seq <- c(RMSEA_LB_null, RMSEA_LB)
  nfac_RMSEA <- .smt_sequential_stop(RMSEA_seq, \(r) r < .05)

  # With which number of factors is the AIC lowest? (which.min returns the first
  # minimum, so ties yield a single, well-defined suggestion). which.min()
  # returns integer(0) when every AIC is NA (a degenerate tiny-N case where the
  # null and all factor models are undefined); fall back to NA so the record
  # keeps a length-1 suggestion.
  AIC_all <- c(AIC_null, AIC)
  nfac_AIC <- if (all(is.na(AIC_all))) NA_integer_ else which.min(AIC_all) - 1

  # The models at the upper end of the sequence over-extract and may be
  # inadmissible, so flag only the solutions the three rules actually select (as
  # efa_hull() does). The null model is not fitted and cannot be inadmissible.
  suggested <- c(chi = nfac_chi, RMSEA = nfac_RMSEA, AIC = nfac_AIC)
  inadmissible <- character(0)
  for (rule in names(suggested)) {
    k <- suggested[[rule]]
    if (is.na(k) || k < 1) next
    issues <- c(if (n_heywood[k] > 0) "Heywood case",
                if (!converged[k]) "non-convergence")
    if (length(issues) > 0) {
      inadmissible <- c(
        inadmissible,
        paste0(rule, ": ", .retention_count(k), " factor",
               if (k != 1) "s" else "",
               " (", paste(issues, collapse = ", "), ")")
      )
    }
  }

  if (length(inadmissible) > 0) {
    cli::cli_warn(
      c("The sequential model tests selected an inadmissible solution: {inadmissible}.",
        "i" = "The selected solution has a Heywood case or did not converge, so the suggested number of factors may be unreliable; interpret it with caution and cross-check with other criteria."),
      class = "efa_smt_inadmissible"
    )
  }

  # one record per criterion (values for the null model through max_fac factors)
  results <- list(
    list(name = "chi", label = "Sequential chi-square model tests",
         n_factors = nfac_chi, plot_type = "none",
         x = 0:max_fac, y = c(p_null, ps)),
    list(name = "RMSEA", label = "Lower bound of RMSEA 90% CI",
         n_factors = nfac_RMSEA, plot_type = "none",
         x = 0:max_fac, y = c(RMSEA_LB_null, RMSEA_LB)),
    list(name = "AIC", label = "Akaike Information Criterion",
         n_factors = nfac_AIC, plot_type = "none",
         x = 0:max_fac, y = AIC_all)
  )

  output <- .new_efa_retention(
    "SMT",
    results = results,
    settings = list(N = N, use = use, cor_method = cor_method)
  )

  return(output)

}

smt_cor <- efa_smt(test_models$baseline$cormat, N = 500)
smt_zero <- efa_smt(diag(nrow = 5, ncol = 5), N = 500)
smt_raw <- efa_smt(GRiPS_raw)

test_that("output class and dimensions are correct", {
  expect_s3_class(smt_cor, "efa_retention")
  expect_length(smt_cor, 6)
  expect_s3_class(smt_raw, "efa_retention")
  expect_length(smt_raw, 6)
  expect_s3_class(smt_zero, "efa_retention")
  expect_length(smt_zero, 6)

  expect_named(smt_cor$n_factors, c("chi", "RMSEA", "AIC"))
  expect_equal(.retention_record(smt_cor, "chi")$plot_type, "none")
})

test_that("number of factors are correct", {
  expect_equal(smt_cor$n_factors[["chi"]], 3)
  expect_equal(smt_cor$n_factors[["RMSEA"]], 2)
  expect_equal(smt_cor$n_factors[["AIC"]], 3)

  expect_equal(smt_raw$n_factors[["chi"]], 3)
  expect_equal(smt_raw$n_factors[["RMSEA"]], 1)
  expect_equal(smt_raw$n_factors[["AIC"]], 3)

  expect_equal(smt_zero$n_factors[["chi"]], 0)
  expect_equal(smt_zero$n_factors[["RMSEA"]], 0)
  expect_equal(smt_zero$n_factors[["AIC"]], 0)
})

test_that("p-values are correct", {
  # the chi record's y is c(p_null, ps_chi)
  chi_cor <- .retention_record(smt_cor, "chi")$y
  expect_lt(chi_cor[1], 0.05)
  expect_lt(chi_cor[2], 0.05)
  expect_lt(chi_cor[3], 0.05)
  expect_gte(chi_cor[4], 0.05)
  expect_gte(chi_cor[5], 0.05)

  chi_raw <- .retention_record(smt_raw, "chi")$y
  expect_lt(chi_raw[1], 0.05)
  expect_lt(chi_raw[2], 0.05)
  expect_lt(chi_raw[3], 0.05)
  expect_gte(chi_raw[4], 0.05)
  expect_gte(chi_raw[5], 0.05)

  chi_zero <- .retention_record(smt_zero, "chi")$y
  expect_gte(chi_zero[1], 0.05)
  expect_gte(chi_zero[2], 0.05)
})

test_that("RMSEA_LB and AIC values are correct", {
  rmsea_cor <- .retention_record(smt_cor, "RMSEA")$y
  rmsea_raw <- .retention_record(smt_raw, "RMSEA")$y
  rmsea_zero <- .retention_record(smt_zero, "RMSEA")$y

  expect_equal(rmsea_cor[1], 0.157856, tolerance = 1e-2)
  expect_equal(rmsea_raw[1], 0.461174, tolerance = 1e-2)
  expect_equal(rmsea_zero[1], 0, tolerance = 1e-2)

  expect_equal(rmsea_cor[-1], c(0.0567972, 0.0397977, rep(0, 10)),
               tolerance = 1e-2)
  expect_equal(rmsea_raw[-1], c(0.0354958, 0.0263924, rep(0, 2)),
               tolerance = 1e-2)
  expect_equal(rmsea_zero[-1], rep(0, 2), tolerance = 1e-2)

  aic_cor <- .retention_record(smt_cor, "AIC")$y
  aic_raw <- .retention_record(smt_raw, "AIC")$y
  aic_zero <- .retention_record(smt_zero, "AIC")$y

  expect_equal(aic_cor[1], 1867.28, tolerance = 0.1)
  expect_equal(aic_raw[1], 4998.06, tolerance = 0.1)
  expect_equal(aic_zero[1], -20, tolerance = 0.1)

  expect_equal(aic_cor[-1], c(133.20917, 13.14179, -80.25186, -79.07978,
                              -75.59415, -68.10822, -57.26154, -46.34868,
                              -36.42090, -26.76733, -17.03907, -5.13003),
               tolerance = 0.1)
  expect_equal(aic_raw[-1], c(19.610600, 7.631330, -4.512150, -1.868430),
               tolerance = 0.1)
  expect_equal(aic_zero[-1], c(-10, -2), tolerance = 0.1)
})

test_that("null-model statistics are computed from R and N, not the fitted model", {
  # The zero-factor (null) statistics depend only on the correlation matrix and
  # N. SMT must derive them directly so they stay finite even when the
  # max-factor model is degenerate (a Heywood / non-positive-definite case
  # leaves that model's fit_indices NA, which previously crashed the p_null and
  # RMSEA_LB_null comparisons).
  R <- test_models$baseline$cormat
  N <- 500

  # SMT smooths a non-positive-definite input before use (the trigger is a
  # smallest eigenvalue below .Machine$double.eps; see .prepare_cor_input). The
  # baseline matrix clears that threshold, so smoothing is a no-op and the R used
  # inside SMT equals the input here; enforce that so the recomputation is valid.
  expect_true(all(eigen(R, symmetric = TRUE, only.values = TRUE)$values >=
                    .Machine$double.eps))

  m <- ncol(R)
  chi_null <- .null_chisq(R, N)
  df_null <- (m^2 - m) / 2
  p_null <- stats::pchisq(chi_null, df_null, lower.tail = FALSE)
  # RMSEA is built on the uncorrected (N - 1) discrepancy scale, so the null-model
  # bound uses the uncorrected baseline chi-square (p_null and AIC_null keep the
  # Bartlett-corrected statistic).
  chi_null_rmsea <- .null_chisq(R, N, corrected = FALSE)
  RMSEA_LB_null <- sqrt(.rmsea_lambda(chi_null_rmsea, df_null, .95) / (df_null * (N - 1)))
  AIC_null <- chi_null - 2 * df_null

  # the null model is the first element of each criterion record's y vector
  null_out <- c(chi   = .retention_record(smt_cor, "chi")$y[1],
                RMSEA = .retention_record(smt_cor, "RMSEA")$y[1],
                AIC   = .retention_record(smt_cor, "AIC")$y[1])
  expect_equal(null_out, c(chi = p_null, RMSEA = RMSEA_LB_null, AIC = AIC_null))

  # finite null statistics in the SMT output are what keep the comparison guards
  # from crashing
  expect_false(anyNA(null_out))
})

test_that("SMT requires more observations than variables", {
  # The sequential tests rest on normal-theory chi-squares whose Bartlett
  # multiplier N - 1 - (2p + 5)/6 is only positive above that boundary.
  R <- stats::cor(GRiPS_raw)  # 8 variables
  expect_error(efa_smt(R, N = 8), class = "efa_n_too_small")
  expect_s3_class(suppressWarnings(efa_smt(R, N = 9)), "efa_retention")
})

test_that("an undefined value stops both sequential rules with no suggestion", {
  # Both sequential rules share one convention: an undefined value breaks the
  # strictly sequential test, so the search stops there and no number is
  # suggested rather than skipping ahead to a later model that meets the rule.
  chi_rule <- \(p) p > 0.05
  expect_equal(.smt_sequential_stop(c(NA, 0.9, 0.9), chi_rule), NA_integer_)
  expect_equal(.smt_sequential_stop(c(0.01, NA, 0.9), chi_rule), NA_integer_)
  expect_equal(.smt_sequential_stop(c(0.01, 0.9, 0.9), chi_rule), 1L)
  # every model significant: no number can be suggested either
  expect_equal(.smt_sequential_stop(c(0.01, 0.01), chi_rule), NA_integer_)

  rmsea_rule <- \(r) r < .05
  expect_equal(.smt_sequential_stop(c(NA, .01, .01), rmsea_rule), NA_integer_)
  expect_equal(.smt_sequential_stop(c(.2, .01, .01), rmsea_rule), 1L)
})

test_that("the chi-square sequence matches stats::factanal", {
  # Two independent ML optimisers on the same matrix: the agreement is tight but
  # not exact, and how tight depends on the BLAS/LAPACK the two share. Pin it
  # where the numerical environment is known rather than across every check
  # flavour (as the comparison-data pin in test-efa_cd.R does).
  skip_on_cran()
  R <- test_models$baseline$cormat
  N <- 500
  # the chi record's y is c(p_null, p_1, ..., p_max_fac)
  ps <- .retention_record(smt_cor, "chi")$y

  compared <- 0L
  for (k in seq_len(.det_max_factors(ncol(R)))) {
    fa <- stats::factanal(covmat = R, factors = k, n.obs = N)
    # Solutions with a uniqueness at the lower bound are boundary solutions: the
    # two optimisers can then attain different local optima (EFAtools reaches the
    # lower discrepancy on this matrix), so only interior solutions are comparable.
    if (min(fa$uniquenesses) <= .uniqueness_floor + 1e-8) next
    expect_equal(ps[k + 1], fa$PVAL[[1]], tolerance = 1e-6)
    compared <- compared + 1L
  }
  # guard against a vacuously passing loop
  expect_gt(compared, 0)
})

test_that("an inadmissible selected solution raises a classed warning", {
  # The warning names the number of factors each rule selected, so it renders counts the
  # way the report does. A negative options(scipen) is set here rather than in a test of
  # its own, so that the one fit covers both the condition and its text.
  withr::local_options(scipen = -5)

  # On this matrix the chi-square and AIC rules both select a 6-factor model that
  # has Heywood cases, while the RMSEA rule selects an admissible 4-factor model.
  w <- expect_warning(smt_bad <- efa_smt(test_models$case_11b$cormat, N = 500),
                      class = "efa_smt_inadmissible")
  expect_equal(smt_bad$n_factors[["chi"]], 6)
  expect_equal(smt_bad$n_factors[["AIC"]], 6)

  # the AIC rule stores its count as a double, so it is the one that would render as
  # "6e+00" were the message built with a bare paste0(); the pattern takes any exponent,
  # because a count of ten or more carries a different one
  expect_false(grepl("e[+-][0-9]", conditionMessage(w)))

  # a clean sequence stays silent
  expect_no_warning(efa_smt(test_models$baseline$cormat, N = 500),
                    class = "efa_smt_inadmissible")
})

test_that("settings are returned correctly", {
  expect_named(smt_cor$settings, c("N", "use", "cor_method"))
  expect_named(smt_raw$settings, c("N", "use", "cor_method"))
  expect_named(smt_zero$settings, c("N", "use", "cor_method"))

  expect_equal(smt_cor$settings$N, 500)
  expect_equal(smt_raw$settings$N, 810)
  expect_equal(smt_zero$settings$N, 500)

  expect_equal(smt_cor$settings$use, "pairwise.complete.obs")
  expect_equal(smt_raw$settings$use, "pairwise.complete.obs")
  expect_equal(smt_zero$settings$use, "pairwise.complete.obs")

  expect_equal(smt_cor$settings$cor_method, "pearson")
  expect_equal(smt_raw$settings$cor_method, "pearson")
  expect_equal(smt_zero$settings$cor_method, "pearson")

})

test_that("errors are thrown correctly", {
  expect_error(efa_smt(1:5), class = "efa_input_not_matrix")
  expect_error(efa_smt(test_models$baseline$cormat), class = "efa_n_required")
  expect_message(efa_smt(GRiPS_raw), class = "efa_cor_from_data")
  expect_warning(efa_smt(GRiPS_raw, N = 20), class = "efa_n_from_data")
  expect_error(efa_smt(sing_raw), class = "efa_cor_singular")
  expect_error(efa_smt(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_error(efa_smt(matrix(rnorm(50), ncol = 2)), class = "efa_smt_underidentified") # underidentified case
  expect_error(efa_smt(matrix(rnorm(60), ncol = 3)), class = "efa_smt_underidentified") # just identified case

  # SMT's sequential tests rest on a normal-theory chi-square that is not valid for
  # polychoric / tetrachoric correlations, so they are rejected rather than run.
  expect_error(efa_smt(GRiPS_raw, cor_method = "poly"),
               class = "efa_cor_method_unsupported")
  expect_error(efa_smt(GRiPS_raw, cor_method = "tetra"),
               class = "efa_cor_method_unsupported")
})

rm(smt_cor, smt_raw, smt_zero)

#' Extract a list object by its name
#'
#' Consider a list of named sub-lists. This function extracts, for each sub-list,
#' the sub-list element that is specified by the user. This function is useful
#' for extracting results from [efa_fit()] for each imputation in
#' [efa_mi()].
#'
#' @param alist A list of sub-lists, typically a list of \eqn{m} objects of class
#' `"efa"`, where \eqn{m} is the number of imputations passed to
#' [efa_mi()].
#' @param object String of length 1. The name of the object to extract e.g.
#' `"h2"` or `"vars_accounted"`.
#'
#' @return A list of length \eqn{m}, with each element containing the extracted
#' `object` for the \eqn{k}th element (\eqn{k = 1,..., m}).
#'
#' @author Andreas Soteriades
#'
#' @keywords internal
.extract_list_object <- function(alist, object) {
  lapply(
    alist,
    function(x) {
      x[[object]]
    }
  )
}

# Set `seed` for the duration of the *calling* function and restore the caller's
# global RNG state when that function returns, so `.Random.seed` carries no lasting
# side effect: the caller's stream is saved and reinstated on exit or -- if none
# existed yet -- the state set.seed() creates is removed again. The restoring
# on.exit() is registered in the caller's frame (`envir`, via do.call) so it fires
# when the public function returns, not when this helper does; the saved state is
# inlined into the handler with bquote() so it resolves in that frame. `add = TRUE`
# composes with any other on.exit() handlers the caller has registered. A NULL
# `seed` is a no-op. Used to make seeded, worker-count-independent bootstraps and
# simulations reproducible without disturbing the caller's RNG stream.
.set_local_seed <- function(seed, envir = parent.frame()) {
  if (is.null(seed)) {
    return(invisible(NULL))
  }
  seed_existed <- exists(".Random.seed", envir = globalenv(), inherits = FALSE)
  saved_seed <- if (seed_existed) {
    get(".Random.seed", envir = globalenv(), inherits = FALSE)
  }
  do.call(
    on.exit,
    list(
      bquote({
        if (.(seed_existed)) {
          assign(".Random.seed", .(saved_seed), envir = globalenv())
        } else if (exists(".Random.seed", envir = globalenv(), inherits = FALSE)) {
          rm(".Random.seed", envir = globalenv())
        }
      }),
      add = TRUE
    ),
    envir = envir
  )
  set.seed(seed)
  invisible(NULL)
}

# Abort (classed) when the suggested lavaan package is unavailable; guards the lavaan
# input paths of OMEGA() and efa_schmid_leiman(). `call` is forwarded so the error points
# at the caller.
.require_lavaan <- function(call = rlang::caller_env()) {
  if (!requireNamespace("lavaan", quietly = TRUE)) {
    cli::cli_abort(
      c("The {.pkg lavaan} package must be installed to use a {.cls lavaan} object as input.",
        "i" = 'Install it with {.code install.packages("lavaan")}.'),
      class = "efa_lavaan_not_installed",
      call = call
    )
  }
}

# One deprecation warning for the retired option of the empirical Kaiser criterion reference
# values. The two arguments that selected it (`efa_ekc(type)` and `efa_retain(ekc_type)`) share
# it, so their wording cannot drift apart.
#
# `always = TRUE`: the fallback changes the number of factors the criterion suggests. A warning
# that goes quiet after the first call of a session is too easy to miss.
#
# `user_env = globalenv()` tells lifecycle that the caller is a user, and it is load-bearing for
# EVERY call, not only the ones that arrive through EKC() or N_FACTORS(). lifecycle reads
# `user_env` to find out whether a user named the deprecated argument, or whether a package named
# it in its own internals; only the second case is rate limited, and it also adds a footer that
# tells the reader to report a bug against the package. The default resolves two frames out from
# lifecycle's own, which is this helper's caller -- so it is efa_ekc() or efa_retain(), this
# package, whichever route the call took. Without the override lifecycle would therefore ignore
# `always` and misattribute the warning to EFAtools even for a direct call.
#
# The assertion is safe because nothing in the package passes either argument on, and the
# superseded wrappers forward the lifecycle sentinel unless the caller named the argument. A call
# that gets here is therefore always a user's.
.deprecate_ekc_type <- function(what) {
  lifecycle::deprecate_warn(
    when = "1.1.0",
    what = what,
    always = TRUE,
    details = c(
      "i" = paste("The reference values of \"AM2019\" do not depend on the observed",
                  "eigenvalues, so they do not apply the empirical correction that",
                  "defines the criterion."),
      "i" = paste("The criterion is now always computed as in Braeken and van Assen",
                  "(2017), so the number of factors can differ from earlier results.")
    ),
    user_env = globalenv()
  )
}

# Evaluate a block of argument checks and re-raise checkmate's assertion failures under
# `efa_invalid_argument`. checkmate's assert_*() functions signal a plain `simpleError`,
# which leaves an invalid argument as the one condition in the package that cannot be
# caught or tested by class; this keeps checkmate's wording but gives it the class the
# rest of the package guarantees. cli_abort() raises an rlang condition rather than a
# `simpleError`, so classed checks inside the block pass through untouched. Only messages
# carrying checkmate's `mstop()` prefix are relabelled: an "object not found" or any other
# unclassed error from the block is an internal fault, not a user input problem, and is
# rethrown unchanged rather than disguised as one. `call` is forwarded so the error points
# at the user's call, not at this helper.
.assert_args <- function(expr, call = rlang::caller_env()) {
  tryCatch(
    expr,
    simpleError = function(e) {
      if (!startsWith(conditionMessage(e), "Assertion on ")) stop(e)
      cli::cli_abort("{conditionMessage(e)}", class = "efa_invalid_argument", call = call)
    }
  )
}

# Second-order Schmid-Leiman group-factor loadings: scale the first-order factor
# loadings by the residual standard deviations of the first-order factors. Only
# the psi diagonal is used (off-diagonal first-order disturbance covariances are
# not residual standard deviations); `nrow` keeps diag() a matrix when there is a
# single first-order factor. Shared by efa_schmid_leiman() and OMEGA() for lavaan
# second-order input. Schmid & Leiman (1957, Psychometrika).
.sl_group_loadings <- function(loadings, psi, col_names) {
  loadings %*% diag(sqrt(diag(psi)[col_names]), nrow = length(col_names))
}

#' Convert an `"efa_loadings"` table to matrix or a matrix to `"efa_loadings"`
#'
#' The loadings tables returned by [efa_fit()] are of class
#' `c("efa_loadings", "LOADINGS")`, which prevents applying functions on them.
#' This function changes their class to `"matrix"`, and changes it back to
#' `"efa_loadings"` when done.
#'
#' @param x A table of class `"matrix"` or `"efa_loadings"`.
#' @param cl A character vector with the class to change the table to. Should be
#' `c("efa_loadings", "LOADINGS")` or `"matrix"`.
#'
#' @return A table with the loadings, of class either `"efa_loadings"` or
#' `"matrix"`.
#'
#' @author Andreas Soteriades
#'
#' @keywords internal
.change_class <- function(x, cl = 'matrix') {
  class(x) <- cl
  return(x)
}

# Vectorize the lower triangle of a symmetric matrix (diagonal included) in column-major
# order, dropping the duplicated upper triangle. This is the standard vech() ordering (the
# one MBESS::vech() returns), which callers rely on when they pair the vectorized entries
# with duplication-matrix weights.
.vech <- function(M) {
  M <- as.matrix(M)
  M[lower.tri(M, diag = TRUE)]
}

# Reconstruct a symmetric matrix from .vech() output.
.unvech <- function(v, k) {
  M <- matrix(NA_real_, k, k)
  M[lower.tri(M, diag = TRUE)] <- v
  M[upper.tri(M)] <- t(M)[upper.tri(M)]
  M
}

# Varimax simplicity criterion monitored for convergence by .VARIMAX_SPSS(); the
# rotation stops once it stabilises. This is the criterion SV given in the SPSS
# Statistics Algorithms manual (FACTOR, "Orthogonal Rotations (Harman, 1976)"),
# evaluated on the Kaiser-normalized loadings: for each factor, n times the sum of
# its fourth-power loadings minus the squared sum of its squared loadings, summed
# over factors and divided by n^2 (Kaiser, 1958).
.SV <- function(lambda) {

  n <- nrow(lambda)

  sum(n * colSums(lambda ** 4) - colSums(lambda ** 2) ** 2) / n ** 2

}


.onUnload <- function (libpath) {
  library.dynam.unload("EFAtools", libpath)
}

.det_max_factors <- function(m) {
  q <- floor((2*m + 1 - sqrt(8 * m + 9)) / 2)
  if(q < 0) q <- 0
  return(q)
}

# Eigenvalues of the correlation matrix under each requested diagonal convention used by the
# eigenvalue-based retention criteria: "PCA" keeps the unit diagonal, "SMC" substitutes the
# squared multiple correlations, and "EFA" substitutes the final communalities of an EFA
# solution (default principal axis factoring extracting `n_factors`). Returns a named list
# (PCA/SMC/EFA); an entry is NA for any convention not requested. `...` is forwarded to efa_fit()
# for the "EFA" convention.
.three_eigen <- function(R, eigen_type, n_factors = 1, ...) {

  eigen_R_PCA <- NA
  eigen_R_SMC <- NA
  eigen_R_EFA <- NA

  if ("PCA" %in% eigen_type) {
    eigen_R_PCA <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
  }

  if ("SMC" %in% eigen_type) {
    R_SMC <- R
    diag(R_SMC) <- .smc_start(R)
    eigen_R_SMC <- eigen(R_SMC, symmetric = TRUE, only.values = TRUE)$values
  }

  if ("EFA" %in% eigen_type) {
    R_EFA <- R
    EFA_h2 <- suppressMessages(suppressWarnings(efa_fit(R, n_factors = n_factors, ...)$h2))
    diag(R_EFA) <- EFA_h2
    eigen_R_EFA <- eigen(R_EFA, symmetric = TRUE, only.values = TRUE)$values
  }

  list(PCA = eigen_R_PCA, SMC = eigen_R_SMC, EFA = eigen_R_EFA)
}

# The common look of every plot the package returns: `theme_minimal()` with a bold,
# centred title. Defined once so that panels from different functions agree when a user
# puts them together in one report. Callers add only the settings that are specific to
# their plot (for example, a hidden legend) after this.
.gg_theme <- function() {
  ggplot2::theme_minimal() +
    ggplot2::theme(plot.title = ggplot2::element_text(face = "bold", hjust = 0.5))
}

# Fall back to `y` when `x` is NULL. Defined here because the base R version was
# only added in R 4.4.0, below the package's minimum.
`%||%` <- function(x, y) if (is.null(x)) y else x

test_that(".efa_num formats numbers, normalises negative zero, returns plain character", {
  # strips the leading zero of values in (-1, 1) and left-pads to a common width
  expect_equal(.efa_num(0.23, digits = 2), " .23")
  expect_equal(.efa_num(0.2345, digits = 3), " .234")
  expect_equal(.efa_num(0.2345, digits = 3, print_zero = TRUE), " 0.234")

  # a value rounding to zero from below is rendered without a sign, like its
  # positive counterpart (no negative zero)
  expect_equal(.efa_num(-0.0001, digits = 3), .efa_num(0.0001, digits = 3))
  expect_equal(.efa_num(-0.0001, digits = 3, print_zero = TRUE),
               .efa_num(0.0001, digits = 3, print_zero = TRUE))
  expect_false(any(grepl("-", .efa_num(-0.0001, digits = 3), fixed = TRUE)))

  # the padded result is always plain character, never a cli ansi_string
  padded <- .efa_num(0.5, digits = 3, pad = TRUE)
  expect_type(padded, "character")
  expect_false(inherits(padded, "cli_ansi_string"))
})


x_base <- population_models$loadings$baseline
x_NA <- population_models$loadings$baseline
x_NA[1, 3] <- NA
y_base <- x_base[, c(3,2,1)]
y_NA <- y_base
y_NA[2, 2] <- NA

test_that(".factor_congruence works", {
  checkmate::expect_matrix(.factor_congruence(x_base, y_base))
  expect_equal(sum(.factor_congruence(x_base, y_base)), 3)
  expect_warning(.factor_congruence(x_NA, y_NA, na.rm = FALSE), class = "efa_missing_check")
  expect_warning(.factor_congruence(x_NA, y_NA), class = "efa_missing_complete")
})

test_that(".factor_congruence has (i, j) = cos(x_i, y_j) orientation", {
  # Asymmetric inputs so the congruence matrix differs from its transpose; this
  # pins the orientation the efa_compare / averaging reordering depends on.
  A <- matrix(c(2, 0, 1,
                0, 1, 0,
                0, 0, 3,
                1, 2, 0), nrow = 4, byrow = TRUE)
  B <- matrix(c(0, 1, 0,
                3, 0, 1,
                0, 0, 2,
                1, 1, 0), nrow = 4, byrow = TRUE)
  # Reference computed entrywise from column dot products, independent of the
  # matrix-algebra implementation.
  ref <- sapply(seq_len(ncol(B)), function(j)
    sapply(seq_len(ncol(A)), function(i)
      sum(A[, i] * B[, j]) / sqrt(sum(A[, i]^2) * sum(B[, j]^2))))
  expect_false(isTRUE(all.equal(ref, t(ref))))
  expect_equal(unname(.factor_congruence(A, B)), ref)
})

test_that(".factor_congruence aborts when too few complete cases remain", {
  x_few <- population_models$loadings$baseline
  x_few[-1, 1] <- NA   # only the first row has complete data
  expect_error(
    suppressWarnings(.factor_congruence(x_few, x_base)),
    class = "efa_too_few_complete"
  )
})

test_that(".is_cormat works", {
  expect_equal(.is_cormat(cor(cbind(rnorm(100), rnorm(100)))), TRUE)
  expect_equal(.is_cormat(cbind(rnorm(100), rnorm(100))), FALSE)
  expect_equal(.is_cormat(cbind(rnorm(2), rnorm(2))), FALSE)
  expect_equal(.is_cormat(cbind(c(1, NA, .57, .85))), FALSE)
  expect_equal(.is_cormat(matrix(c(1, .1, .3, 1), ncol = 2)), FALSE)
  expect_error(.is_cormat(matrix(c(1, NA, NA, 1), ncol = 2)),
               class = "efa_cormat_has_na")
  # A symmetric matrix with entries in [-1, 1] but a non-unit diagonal (here trace 2.6, which
  # rounds to the row count) is not a correlation matrix and must not be detected as one.
  expect_equal(
    .is_cormat(matrix(c(0.6, 0.2, 0.2, 0.2, 1, 0.2, 0.2, 0.2, 1), ncol = 3)),
    FALSE
  )
})

q_p <- .det_max_factors(8) + 1
test_that(".det_max_factors works", {
  expect_type(.det_max_factors(8), "double")
  expect_lte(((8 - q_p)**2 - (8 + q_p)) / 2, 0)
  expect_equal(.det_max_factors(0), 0)
  expect_equal(.det_max_factors(1), 0)
  expect_equal(.det_max_factors(2), 0)
  expect_equal(.det_max_factors(3), 0)
  expect_gt(.det_max_factors(4), 0)
})


test_that(".decimals works", {
  expect_type(.decimals(8), "double")
  expect_equal(.decimals(8), 0)
  expect_type(.decimals(8), "double")
  expect_error(.decimals("a"), class = "efa_not_numeric")

  # NA elements must not trigger "missing value where TRUE/FALSE needed"; they
  # contribute 0 decimals so the rest of the input still drives the result.
  expect_equal(.decimals(NA_real_), 0)
  expect_equal(.decimals(c(1.23, NA, 4.5)), 2)
  expect_equal(.decimals(matrix(c(1.23, NA, 4.5, 6), ncol = 2)), 2)

  # Small magnitudes that R renders in scientific notation (e.g. "3e-05") used to
  # abort with "subscript out of bounds"; the decimal count must be taken from the
  # fixed-notation representation instead.
  expect_equal(.decimals(3e-5), 5)
  expect_equal(.decimals(1e-4), 4)
  expect_equal(.decimals(1e-6), 6)
  expect_equal(.decimals(1.23e-5), 7)
  expect_equal(.decimals(c(0.5, 3e-5)), 5)
  expect_equal(.decimals(matrix(c(0.5, 3e-5, 0.4, 0.2), ncol = 2)), 5)

  # A comma-decimal OutDec must not collapse the count to 0: the separator is
  # pinned to "." so the split still finds the fractional part.
  withr::with_options(list(OutDec = ","), {
    expect_equal(.decimals(0.123), 3)
    expect_equal(.decimals(3e-5), 5)
  })
})

### test .rotation_family

test_that(".rotation_family works", {
  expect_identical(.rotation_family("none"), "none")

  # representative names hardcoded so the test catches a rotation placed in the
  # wrong family vector (including the special-cased varimax/promax and the
  # geomin/bentler/bifactor T-vs-Q pairs)
  expect_identical(.rotation_family("varimax"), "orthogonal")
  expect_identical(.rotation_family("geominT"), "orthogonal")
  expect_identical(.rotation_family("bifactorT"), "orthogonal")
  expect_identical(.rotation_family("promax"), "oblique")
  expect_identical(.rotation_family("geominQ"), "oblique")
  expect_identical(.rotation_family("bifactorQ"), "oblique")

  # and every canonical name dispatches to its family
  for (rot in .orth_rotations) {
    expect_identical(.rotation_family(rot), "orthogonal")
  }

  for (rot in .oblq_rotations) {
    expect_identical(.rotation_family(rot), "oblique")
  }

  expect_error(.rotation_family("bogus"), class = "efa_unknown_rotation")
})



rm(x_base, y_base, q_p)


test_that(".smc_start returns the squared multiple correlations", {
  R <- test_models$baseline$cormat
  # unnamed, so the return shape does not depend on which inverse route ran
  expect_equal(.smc_start(R), unname(1 - 1 / diag(solve(R))))
  expect_null(names(.smc_start(R)))
})

test_that(".smc_start clamps squared multiple correlations to [0, 1]", {
  # A squared multiple correlation cannot leave [0, 1], but 1 - 1/diag(R^-1) does as soon
  # as R is not positive definite -- and the value is written straight onto a correlation
  # diagonal (.three_eigen) or handed to the optimiser as a starting uniqueness.

  # indefinite, invertible: the unclamped expression returns 1.8 for every variable
  R_ind <- matrix(c(1, .6, .6, .6, 1, -.6, .6, -.6, 1), 3)
  raw_ind <- 1 - 1 / diag(solve(R_ind))
  expect_true(any(raw_ind > 1))
  expect_equal(.smc_start(R_ind), pmin(pmax(raw_ind, 0), 1))

  # the shared non-positive-definite fixture escapes both bounds (-Inf and 2)
  raw_np <- 1 - 1 / diag(solve(cor_nposdef))
  expect_true(any(raw_np < 0) && any(raw_np > 1))
  expect_equal(.smc_start(cor_nposdef), pmin(pmax(raw_np, 0), 1))

  # what .three_eigen() puts on the diagonal is the clamped value
  R_smc <- R_ind
  diag(R_smc) <- .smc_start(R_ind)
  expect_true(all(diag(R_smc) >= 0 & diag(R_smc) <= 1))
  expect_equal(.three_eigen(R_ind, "SMC")$SMC,
               eigen(R_smc, symmetric = TRUE, only.values = TRUE)$values)
})

test_that(".smc_start errors on a singular matrix and only on a singular matrix", {
  # .parallel_sim_eig() rejects a simulated draw precisely when .smc_start() errors, so
  # the positive-definite route must neither turn a singular matrix into a silent result
  # nor start rejecting matrices that are merely not positive definite. Both halves of
  # that boundary are asserted here. The error comes from base solve() and carries no
  # condition class.
  expect_error(.smc_start(sing_cor))

  expect_no_error(.smc_start(cor_nposdef))                    # invertible, not PD
  expect_no_error(.smc_start(test_models$baseline$cormat))    # positive definite
})

test_that(".three_eigen returns eigenvalues per requested diagonal convention", {
  R <- test_models$baseline$cormat

  res <- .three_eigen(R, c("PCA", "SMC"))

  # PCA: eigenvalues on the unit-diagonal correlation matrix
  expect_equal(res$PCA, eigen(R, symmetric = TRUE, only.values = TRUE)$values)

  # SMC: eigenvalues with the squared multiple correlations on the diagonal
  R_smc <- R
  diag(R_smc) <- .smc_start(R)
  expect_equal(res$SMC, eigen(R_smc, symmetric = TRUE, only.values = TRUE)$values)

  # a convention that was not requested is NA
  expect_true(is.na(res$EFA))

  # EFA: eigenvalues with the communalities of an n-factor EFA on the diagonal
  res_efa <- .three_eigen(R, "EFA", n_factors = 2)
  R_efa <- R
  diag(R_efa) <- suppressMessages(suppressWarnings(EFA(R, n_factors = 2)$h2))
  expect_equal(res_efa$EFA, eigen(R_efa, symmetric = TRUE, only.values = TRUE)$values)
  expect_true(is.na(res_efa$PCA))
})

# Goodness-of-fit indices for a fitted factor solution: explained-variance accounting,
# RMSR, the KMO-based CAF, and the chi-square / CFI / RMSEA / AIC / BIC block from `.gof()`.

#' Compute explained variances from loadings
#'
#' From unrotated loadings compute the communalities and uniquenesses for total
#' variance. Compute explained variances per factor from rotated loadings (and
#' factor intercorrelations Phi if oblique rotation was used).
#'
#' @param L_unrot matrix. Unrotated factor loadings.
#' @param L_rot matrix. Rotated factor loadings.
#' @param Phi matrix. Factor intercorrelations. Provide only if oblique rotation
#'  is used.
#'
#' @return A matrix with sum of squared loadings, proportion explained variance
#'  from total variance per factor, same as previous but cumulative, Proportion
#'  of explained variance from total explained variance, and same as previous but
#'  cumulative. The three cumulative and common-variance rows are omitted when
#'  `L_rot` has a single column, where they would only repeat the two above them,
#'  so the result has two rows there and five otherwise.
#'
#' @keywords internal
.compute_vars <- function(L_unrot, L_rot, Phi = NULL) {

  if (is.null(Phi)) {
    # sum of squared loadings per factor; colSums() already returns the scalar for
    # a single-factor solution, so no special case is needed.
    vars <- colSums(L_rot^2)
  } else {
    # compute variance proportions
    vars <- diag(Phi %*% t(L_rot) %*% L_rot)
  }

  # Compute the explained variances. The code is based on the psych::fac() function
  # total variance (sum of communalities and uniquenesses)
  h2 <- rowSums(L_unrot^2)            # diag(L L'), without forming the full p x p product
  var_total <- sum(h2 + (1 - h2))
  vars_explained <- rbind(`SS loadings` = vars)
  vars_explained <- rbind(vars_explained, `Prop Tot Var` = vars / var_total)

  if (ncol(L_rot) > 1) {
    vars_explained <- rbind(vars_explained,
                            `Cum Prop Tot Var` = cumsum(vars / var_total))
    vars_explained <- rbind(vars_explained,
                            `Prop Comm Var` = vars / sum(vars))
    vars_explained <- rbind(vars_explained,
                            `Cum Prop Comm Var` = cumsum(vars / sum(vars)))
  }

  vars_explained
}

# RMSR (root mean square residual; Harman, 1976) from a residual matrix: the root mean square
# of the p(p - 1)/2 unique off-diagonal residuals. The diagonal carries no residual information
# for a correlation-structure model and is excluded. Note that psych's `rms` (psych:::fa.stats)
# divides that same one-per-pair sum by p(p - 1), the count of *both* triangles, and so reports
# this value divided by sqrt(2); the two are not directly comparable. Shared by .gof() and the
# multiple-imputation pooler.
# An unavailable residual propagates rather than being dropped: RMSR and SRMR are the same sum
# of squares over the same pairs and differ only by the fixed factor sqrt((p - 1)/(p + 1)),
# which holds only while both are computed over the full set. Averaging RMSR over the surviving
# pairs while .srmr() returns NA would break that relation, and would report a residual summary
# describing fewer variable pairs than it appears to.
.rmsr <- function(residuals) {
  E <- as.matrix(residuals)
  sqrt(mean(E[upper.tri(E, diag = FALSE)]^2))
}

# SRMR (standardized root mean square residual; Bentler, 1995) from a residual matrix whose
# model-implied diagonal is 1, so only the off-diagonal residuals contribute; the denominator
# is the count of non-redundant elements p(p + 1)/2. Distinct from .rmsr() (off-diagonal mean,
# p(p - 1)/2 denominator). Shared by .gof() and the multiple-imputation pooler. An unavailable
# residual propagates, as it does in .rmsr(); see the note there.
.srmr <- function(residuals) {
  E <- as.matrix(residuals)
  # The denominator counts the non-redundant elements of a p x p matrix, so a non-square
  # block would not match the elements upper.tri() selects.
  if (nrow(E) != ncol(E)) {
    cli::cli_abort("{.arg residuals} must be a square matrix.",
                   class = "efa_srmr_not_square")
  }
  sqrt(sum(E[upper.tri(E)]^2) / (nrow(E) * (nrow(E) + 1) / 2))
}

.compute_caf <- function(delta_hat) {

  delta_hat_KMO <- try(.compute_kmo(delta_hat)$KMO, silent = TRUE)

  if (inherits(delta_hat_KMO, "try-error") || is.na(delta_hat_KMO)) {
    CAF <- 0
    cli::cli_warn(
      c("CAF could not be computed; it was set to {.val 0} (the worst value).",
        "i" = "Inspect the results carefully."),
      class = "efa_caf_failed"
    )
  } else {
    CAF <- 1 - delta_hat_KMO
  }

  CAF

}

# Bartlett's (1951) small-sample multiplier for the ML/ULS discrepancy chi-square,
# N - 1 - (2p + 5)/6 - (2q)/3 with p variables and q factors. q = 0 gives the
# independence-model (sphericity) multiplier. Shared by the model chi-square in
# .gof(), the baseline chi-square in .null_chisq(), and EFA_POOLED()'s pooled
# (N - 1) rescaling so all three use one definition.
.bartlett_mult <- function(N, p, q = 0) N - 1 - (2 * p + 5) / 6 - (2 * q) / 3

# Shared wording for the situations in which the chi-square-derived quantities are
# unavailable. efa_fit(), efa_bartlett(), and efa_screen() raise the warnings, and their print
# methods repeat the same fact, so the sentences live here once instead of being restated (and
# drifting) at six call sites. `what` selects the sentence; `N`, `p`, and `q` name the concrete
# numbers where the caller has them (the warnings) and are omitted where it does not (the print
# methods, which see only the resulting NA).
.fit_unavailable_text <- function(what = c("chisq_block", "residuals_kept",
                                           "residuals_not_identification", "bartlett_mult"),
                                  N = NULL, p = NULL, q = NULL) {
  what <- match.arg(what)
  switch(
    what,
    # What becomes unavailable: the chi-square test itself and every index built on it.
    chisq_block = paste("the chi-square test and the fit indices derived from it (CFI, TLI,",
                        "RMSEA, AIC, BIC, ECVI) are not available"),
    # What is still there, so a reader does not conclude that nothing was computed. Says
    # "in the returned object" rather than "reported", because the print methods suppress
    # the fit and residual sections in exactly the cases that use this sentence.
    residuals_kept = paste("The residual summaries (CAF, RMSR, SRMR) and the degrees of freedom",
                           "are still present in the returned object."),
    # ... and what those residuals cannot settle.
    residuals_not_identification = paste("Residual size does not establish that a model is",
                                         "identified; with negative degrees of freedom a small",
                                         "residual is an artefact of over-parameterisation."),
    # Bartlett's (1951) small-sample multiplier, the reason a valid but small N leaves the
    # statistic undefined. `q` is omitted for the sphericity (0-factor) multiplier.
    bartlett_mult = {
      form <- if (is.null(q)) "N - 1 - (2p + 5)/6" else "N - 1 - (2p + 5)/6 - 2q/3"
      if (is.null(N) || is.null(p)) {
        paste0("The Bartlett multiplier ", form, " is not positive for a sample this small ",
               "relative to the number of variables.")
      } else {
        nums <- if (is.null(q)) {
          paste0("N = ", N, " and p = ", p)
        } else {
          paste0("N = ", N, ", p = ", p, ", and q = ", q)
        }
        paste0("The Bartlett multiplier ", form, " is not positive for ", nums, ".")
      }
    }
  )
}

# Report why .null_chisq() left Bartlett's test of sphericity undefined, for the callers that
# expose that statistic to the user (efa_bartlett(), efa_screen()). N relative to p is the only
# actionable fact behind the NA, and it is otherwise visible only as a missing value in the
# returned list. Raises nothing when the multiplier is positive: the NA then came from the other
# route (a non-positive-definite log-determinant) and would be misdescribed by this reason.
.warn_bartlett_n_too_small <- function(N, p) {
  if (!isTRUE(.bartlett_mult(N, p) <= 0)) return(invisible(NULL))
  mult_note <- .fit_unavailable_text("bartlett_mult", N = N, p = p)
  cli::cli_warn(
    c("Bartlett's test of sphericity could not be computed; {.arg N} is too small relative
       to the number of variables.",
      "i" = "{mult_note}"),
    class = "efa_bartlett_n_too_small"
  )
  invisible(NULL)
}

# Independence-model (baseline) chi-square of the null model (model-implied matrix =
# identity), -log|R| * mult. With `corrected = TRUE` (default) mult is the Bartlett
# multiplier N - 1 - (2p + 5)/6: this is Bartlett's test of sphericity, the reported
# baseline statistic shared with BARTLETT(). It is not the CFI/TLI baseline -- .gof() builds
# that one at (N - 1) directly (see chi_null_cfi there). With
# `corrected = FALSE` mult is (N - 1), the uncorrected discrepancy scale on which the
# RMSEA noncentrality is built (the 0-factor reference in HULL() and the null model in SMT()).
.null_chisq <- function(R, N, ld = determinant(R, logarithm = TRUE),
                        corrected = TRUE) {
  p <- ncol(R)
  # Use the log-determinant directly: det(R) underflows to 0 for large,
  # near-singular (but still positive-definite) matrices, which would otherwise
  # turn the statistic into NA and propagate into .gof(), SMT(), and BARTLETT().
  # `ld` defaults to that decomposition; callers that already have it (e.g. .gof()
  # for the model chi-square) pass it in to avoid recomputing determinant(R).
  if (ld$sign <= 0 || !is.finite(ld$modulus)) return(NA_real_)
  # Guard the multiplier: for small N relative to p the Bartlett multiplier turns
  # non-positive, which would flip the baseline chi-square sign and propagate a
  # spurious statistic into .gof(), SMT(), and BARTLETT().
  mult <- if (corrected) .bartlett_mult(N, p) else N - 1
  if (is.na(mult) || mult <= 0) return(NA_real_)
  -as.numeric(ld$modulus) * mult
}

# Noncentrality parameter for an RMSEA confidence bound: solve pchisq(chi, df, ncp) = goal
# for ncp via stats::uniroot (the 90% CI lower bound uses goal = .95, the upper bound goal =
# .05; Browne & Cudeck, 1992). Returns 0 when the model chi-square already lies below the
# target quantile, so the bound collapses to 0. Shared by .gof(), SMT(), and the pooled
# .efa_pooled_rmsea_ci(), so the inversion lives in one place.
.rmsea_lambda <- function(chi, df, goal) {
  # An undefined chi-square or df (e.g. the null model for a tiny N, where .null_chisq()
  # returns NA) has no noncentrality bound; propagate NA rather than failing the
  # `if (pchisq(NA) >= goal)` test with "missing value where TRUE/FALSE needed".
  if (!is.finite(chi) || !is.finite(df)) return(NA_real_)
  if (stats::pchisq(chi, df = df, ncp = 0) >= goal) {
    p_chi_fun <- function(x, val, df, goal) goal - stats::pchisq(val, df, ncp = x)
    # Defensive: for the finite inputs the guard above admits, a sign change always
    # exists -- the gate forces f(0) <= 0 and pchisq(chi, df, ncp) -> 0 as ncp -> Inf --
    # and extendInt = "upX" locates it (verified up to chi = 1e8). Should a root
    # nonetheless prove unbracketable, the bound is undefined rather than an estimation
    # failure, so report NA instead of aborting the whole fit. Callers must therefore keep
    # NA out of `if` conditions (see the min() caps in .chi_fit_indices()).
    # stats::pchisq() stops converging for very large noncentralities (chi in the millions)
    # and emits one base warning per evaluation, so a single fit could leak dozens of them.
    # Muffle them here: the resulting bound is unreliable rather than absent, and the callers
    # detect that from the bound itself (the bracketing checks in .chi_fit_indices() and
    # .efa_pooled_rmsea_ci()). The handler is unconditional, so uniroot()'s own
    # "_NOT_ converged" warning is muffled with them -- selecting on the message is not an
    # option, since both texts are translated -- and those bracketing checks are therefore
    # the only signal that a bound is untrustworthy.
    withCallingHandlers(
      tryCatch(
        stats::uniroot(f = p_chi_fun, interval = c(1e-10, 10000), val = chi, df = df,
                       goal = goal, extendInt = "upX", maxiter = 100L)$root,
        error = function(e) NA_real_
      ),
      warning = function(w) invokeRestart("muffleWarning")
    )
  } else {
    0
  }
}

# RMSEA point estimate from a model chi-square on the (N - 1) noncentrality scale: the
# self-consistent Browne & Cudeck (1992) quantity sqrt(max(0, chi_cfi - df) / (df * (N - 1))).
# Callers guard df > 0 and N > 1 (a positive divisor); the cap at 1, where wanted, is applied by
# the caller. Shared by .chi_fit_indices(), HULL(), and the EFA_POOLED() D2 pooler so the RMSEA
# convention lives in one place.
.rmsea_point <- function(chi_cfi, df, N) {
  sqrt(max(0, chi_cfi - df) / (df * (N - 1)))
}

# TRUE when an RMSEA interval contains the point estimate reported beside it. A noncentrality
# solve that did not converge (a chi in the millions, where stats::pchisq() stops converging)
# can return bounds that collapse below the point estimate; callers withhold such an interval
# rather than report it as a range. isTRUE() makes the test NA-safe, so a bound that is already
# undefined is left alone instead of failing the comparison. Shared by .chi_fit_indices() and
# .efa_pooled_rmsea_ci() so the single-fit and pooled paths apply one rule.
.rmsea_ci_contains <- function(lower, upper, point) {
  !isTRUE(lower > point || point > upper)
}

# CFI (Bentler, 1990) and TLI (Tucker & Lewis, 1973) from model and baseline
# noncentralities on a common scaling constant: chi_cfi and chi_null_cfi, each with
# its df subtracted and floored at 0 so an over-fitting model cannot deflate the
# index and it stays in [0, 1] (matches lavaan::fitMeasures()). Returns NA for both
# when either statistic is undefined. Called from .chi_fit_indices() (the per-model
# path); EFA_POOLED() reports the average of the per-imputation indices this produces
# rather than calling it on pooled statistics.
.incremental_fit <- function(chi_cfi, df, chi_null_cfi, df_null) {
  if (is.na(chi_cfi) || is.na(chi_null_cfi)) {
    return(list(CFI = NA_real_, TLI = NA_real_))
  }
  d_m <- max(chi_cfi - df, 0)
  d_null <- max(chi_null_cfi - df_null, 0)
  CFI <- if (df == 0 || max(d_m, d_null) == 0) 1 else 1 - d_m / max(d_m, d_null)
  if (df == 0) {
    TLI <- 1
  } else {
    ratio_null <- chi_null_cfi / df_null
    # ratio_null == 1 leaves the index undefined (0/0); report a perfect 1.
    TLI <- if (ratio_null == 1) 1 else (ratio_null - chi_cfi / df) / (ratio_null - 1)
  }
  list(CFI = CFI, TLI = TLI)
}

# Chi-square-derived fit indices: the p-values, CFI (Bentler, 1990), TLI/NNFI (Tucker &
# Lewis, 1973), RMSEA with its analytic 90% bounds (Browne & Cudeck, 1992), AIC, BIC, and
# ECVI (Browne & Cudeck, 1989). Computed from a model chi-square and an independence-
# baseline chi-square with their degrees of freedom; `m` is the number of variables, `ci`
# toggles the (uniroot-solved) RMSEA bounds. Returns NA throughout when the model chi-
# square is undefined. Shared by .gof() (the ML/ULS discrepancy chi-square) and the
# scaled-chi-square sandwich-SE path, which passes its scaled model and baseline statistics
# in, so both report the block by exactly the same formulas. `chi_cfi` / `chi_null_cfi` are the
# model and baseline statistics on the common (N - 1) noncentrality scale used for CFI/TLI/RMSEA;
# they are required (not defaulted to `chi`/`chi_null`) so a caller cannot silently report those
# approximation indices off the Bartlett-corrected test scale.
.chi_fit_indices <- function(chi, df, chi_null, df_null, N, m, ci,
                             chi_cfi, chi_null_cfi) {

  if (is.na(chi)) {
    return(list(p_chi = NA_real_, CFI = NA_real_, TLI = NA_real_, RMSEA = NA_real_,
                RMSEA_LB = NA_real_, RMSEA_UB = NA_real_, AIC = NA_real_,
                BIC = NA_real_, ECVI = NA_real_, p_null = NA_real_))
  }

  p_null <- stats::pchisq(chi_null, df_null, lower.tail = F)
  # A just-identified model (df = 0) has no chi-square test: pchisq(chi, 0) returns 0 for
  # any chi > 0, i.e. a spurious "p < .001". Report the p-value as undefined instead,
  # matching the df == 0 handling of CFI/TLI/RMSEA below.
  p_chi <- if (df == 0) NA_real_ else stats::pchisq(chi, df, lower.tail = F)

  # CFI and TLI compare the model and baseline noncentralities, which their definitions
  # (Bentler, 1990; Tucker & Lewis, 1973) require on a single scaling constant. `chi_cfi`
  # and `chi_null_cfi` carry the model and baseline statistics on that common scale (.gof()
  # passes the ML/ULS discrepancy on the (N - 1) noncentrality scale; the scaled sandwich
  # path passes its scaled model and baseline statistics, which are already comparable).
  # When the baseline is undefined (e.g. a degenerate scaled baseline)
  # CFI/TLI are NA while the model-only RMSEA/AIC/BIC/p remain computable.
  inc <- .incremental_fit(chi_cfi, df, chi_null_cfi, df_null)
  CFI <- inc$CFI
  TLI <- inc$TLI

  if (df != 0) {

    # RMSEA point estimate and bounds (Steiger & Lind, 1980; Browne & Cudeck, 1992).
    # The noncentrality estimate chi_cfi - df and its divisor df * (N - 1) are both on the
    # uncorrected (N - 1) discrepancy scale, so the RMSEA is the self-consistent
    # Browne-Cudeck quantity and shares one noncentrality scale with CFI/TLI. The Bartlett
    # small-sample multiplier is applied only to the reported model chi-square test (matching
    # stats::factanal), not to this approximation index, for which it has no role.
    # Capped at 1 with min(), so an undefined statistic propagates as NA instead of
    # failing the comparison (see the bounds below).
    RMSEA <- min(.rmsea_point(chi_cfi, df, N), 1)

    if (isTRUE(ci)) {

      # Analytic 90% RMSEA confidence bounds via the noncentrality solver (Browne &
      # Cudeck, 1992); skipped when ci = FALSE (e.g. per-replicate bootstrap fits).
      lambda_l <- .rmsea_lambda(chi_cfi, df, .95)
      lambda_u <- .rmsea_lambda(chi_cfi, df, .05)

      # Capped at 1 with min(), which passes an undefined bound through: .rmsea_lambda()
      # returns NA when the noncentrality root cannot be located.
      RMSEA_LB <- min(sqrt(lambda_l / (df * (N - 1))), 1)
      RMSEA_UB <- min(sqrt(lambda_u / (df * (N - 1))), 1)

      # At chi around 4e6 both bounds collapse onto the same value below the point estimate.
      # That is not an interval, so report it as undefined rather than clip or reorder it
      # into something that looks usable.
      if (!.rmsea_ci_contains(RMSEA_LB, RMSEA_UB, RMSEA)) {
        RMSEA_LB <- NA_real_
        RMSEA_UB <- NA_real_
      }

    } else {

      RMSEA_LB <- NA_real_
      RMSEA_UB <- NA_real_

    }

  } else {

    RMSEA <- 0
    RMSEA_LB <- 0
    RMSEA_UB <- 0

  }

  AIC <- chi - 2 * df
  BIC <- chi - log(N) * df
  n_params <- m * (m + 1) / 2 - df
  ECVI <- (chi + 2 * n_params) / (N - 1)

  list(p_chi = p_chi, CFI = CFI, TLI = TLI, RMSEA = RMSEA, RMSEA_LB = RMSEA_LB,
       RMSEA_UB = RMSEA_UB, AIC = AIC, BIC = BIC, ECVI = ECVI, p_null = p_null)
}

# Degrees of freedom of an EFA solution with `m` variables and `q` factors, ((m - q)^2 -
# (m + q)) / 2 (equivalently factanal's 0.5 * ((m - q)^2 - m - q)). Shared so the model
# fit and the identification check use one formula.
.efa_df <- function(m, q) ((m - q)^2 - (m + q)) / 2

# Maximum-likelihood (Wishart) discrepancy F(Sigma, R) = tr(Sigma^-1 R) - log|Sigma^-1 R| - m
# between a model-implied matrix `Sigma` (unit diagonal) and an observed/population matrix `R`
# with `m` variables. Computed via a determinant split (log|Sigma^-1 R| = log|R| - log|Sigma|)
# to avoid forming the explicit inverse and to stay stable for ill-conditioned Sigma; the
# caller may pass a precomputed `ldR = determinant(R, logarithm = TRUE)` when it also needs it
# elsewhere. Returns the discrepancy clamped at 0 (a (near-)perfect fit is floating-point dust),
# or NA when Sigma or R is not positive definite or the trace is non-finite. Shared by .gof()
# (the model chi-square) and .efa_population_fit() so simulation and fitting use one formula.
.ml_discrepancy <- function(Sigma, R, m = ncol(R),
                            ldR = determinant(R, logarithm = TRUE)) {
  tryCatch({
    ldSigma <- determinant(Sigma, logarithm = TRUE)
    val <- sum(diag(solve(Sigma, R))) +
      as.numeric(ldSigma$modulus) - as.numeric(ldR$modulus) - m
    if (!is.finite(val) || ldSigma$sign <= 0 || ldR$sign <= 0) {
      NA_real_
    } else {
      max(0, val)
    }
  }, error = function(e) NA_real_)
}

# Population-limit fit of a fixed `q`-factor model (loadings `L`, p by q) to a population
# correlation matrix `R_pop`: the RMSEA, CFI, and TLI of THIS model, in the N -> infinity,
# no-Bartlett-multiplier limit, reusing the same formulas as the fitting side (.ml_discrepancy,
# .efa_df, .chi_fit_indices) rather than reimplementing them. It reports the misfit of the given
# `L` (not a re-optimized best fit), which is what efa_simulate() wants for the generating model;
# a re-fitted EFA that re-optimizes the loadings can attain a smaller discrepancy.
# In that limit the noncentrality parameter equals the discrepancy F per unit (N - 1), so passing
# chi = df + F with N = 2 (N - 1 = 1) makes .chi_fit_indices() return RMSEA = sqrt(F / df),
# CFI = 1 - F / max(F, F_null) (the floor-at-0 CFI from .incremental_fit(); equals 1 - F / F_null
# in the intended regime F <= F_null, which always holds here since L is the generating model), and
# TLI on the exact population scale (the sampling correction and the Bartlett multiplier both drop
# out). The model-implied matrix is LL' with unit diagonal (as in
# .gof()); the independence baseline discrepancy is F_null = -log|R_pop| (as in .null_chisq()).
.efa_population_fit <- function(L, R_pop) {
  m <- nrow(L)
  q <- ncol(L)
  df <- .efa_df(m, q)
  Sigma_model <- tcrossprod(L)
  diag(Sigma_model) <- 1
  ldR <- determinant(R_pop, logarithm = TRUE)
  Fval <- .ml_discrepancy(Sigma_model, R_pop, m, ldR)
  F_null <- -as.numeric(ldR$modulus)
  df_null <- m * (m - 1) / 2
  idx <- .chi_fit_indices(chi = df + Fval, df = df,
                          chi_null = df_null + F_null, df_null = df_null,
                          N = 2, m = m, ci = FALSE,
                          chi_cfi = df + Fval, chi_null_cfi = df_null + F_null)
  list(rmsea = idx$RMSEA, cfi = idx$CFI, tli = idx$TLI, df = df, Fval = Fval)
}

# Two-stage / full-information ML likelihood-ratio chi-square for cor_method = "fiml". The EM
# moments `fiml` (the saturated mean `mu`, covariance `sigma`, and saturated observed-data
# log-likelihood `logl`, plus the raw `data`) come from .prepare_cor_input(); the analysed
# correlation R is cov2cor(sigma). The model-implied correlation R_model = LL' + diag(1 - h2)
# has a unit diagonal (diag(LL') = rowSums(L^2) = h2), so it is the model-implied LL' (the `LLt`
# .gof() already formed) with the diagonal set to 1; put it back on the covariance scale with the
# Stage-1 variances d = diag(sigma) as
# Sigma_model = D^(1/2) R_model D^(1/2). The model leaves the mean unrestricted (mu_model = mu),
# so chi = 2 (logl_sat - logl_model) and df = ((p - q)^2 - (p + q))/2 are unchanged (the mean
# and variance parameters are shared with the saturated model and cancel). The independence
# baseline (free means + free variances, zero covariances) factorises under MAR, so its MLE is
# each variable's available-case mean and ML (/n_j) variance; chi_null = 2 (logl_sat - logl_null)
# with df_null = p(p - 1)/2. No Bartlett correction applies to a likelihood-ratio statistic, so
# the CFI/TLI/RMSEA noncentrality scale uses the same statistics (chi_cfi = chi, chi_null_cfi =
# chi_null). This helper returns the plain two-stage likelihood-ratio statistic; `.gof()` reports
# the Satorra-Bentler-corrected two-stage statistic (.fiml_scaled_test()) in its place, and the
# plain LRT here stands only as the fallback when that correction cannot be formed -- referenced to
# chi^2(df) it is only approximate under the two-stage estimator (Yuan, Marshall, & Bentler, 2002,
# Psychometrika 67:95-121). Returns NA throughout for PAF (no discrepancy), missing N,
# underidentified df, or a non-positive-definite Sigma_model (e.g. a Heywood case), matching how
# the ML/ULS discrepancy block leaves those undefined.
.gof_fiml_chisq <- function(LLt, N, method, df, m, fiml) {

  if (!(method %in% c("ML", "ULS")) || is.na(N) || df < 0) {
    return(list(chi = NA_real_, chi_null = NA_real_, df_null = NA_real_,
                chi_cfi = NA_real_, chi_null_cfi = NA_real_))
  }

  # The saturated log-likelihood was accumulated by the EM over rows with at least one observed
  # value; re-apply the same filter and share one missingness-pattern grouping across the model
  # and baseline log-likelihoods so all three use the identical row set. The filter matters on
  # the point-estimate path, where fiml$data is the user's raw data and may keep a fully-missing
  # row the EM dropped when forming logl_sat (the bootstrap resample pool already excludes such
  # rows).
  data <- as.matrix(fiml$data)         # EFA() accepts a data frame; match .fiml_loglik()
  obs <- !is.na(data)
  keep <- rowSums(obs) > 0L
  data <- data[keep, , drop = FALSE]
  patterns <- .fiml_patterns(obs[keep, , drop = FALSE])
  logl_sat <- fiml$logl

  d <- diag(fiml$sigma)
  R_model <- LLt                       # the p x p product .gof() already formed
  diag(R_model) <- 1
  Sigma_model <- R_model * tcrossprod(sqrt(d))
  logl_model <- tryCatch(.fiml_loglik(data, fiml$mu, Sigma_model, patterns = patterns),
                         error = function(e) NA_real_)
  # A non-positive-definite Sigma_model (e.g. a Heywood case) leaves the model deviance
  # undefined; NA the whole chi-square block together, matching how the ML/ULS discrepancy path
  # NAs the model and baseline in lockstep when the discrepancy is undefined.
  if (is.na(logl_model)) {
    return(list(chi = NA_real_, chi_null = NA_real_, df_null = NA_real_,
                chi_cfi = NA_real_, chi_null_cfi = NA_real_))
  }
  chi <- max(0, 2 * (logl_sat - logl_model))

  # Independence baseline: per-variable available-case mean and ML (/n_j) variance. A diagonal
  # Sigma makes the joint observed-data log-likelihood equal the sum of the univariate marginals,
  # so this is the exact independence-model FIML log-likelihood.
  mu_null <- colMeans(data, na.rm = TRUE)
  var_null <- colMeans(sweep(data, 2L, mu_null, "-")^2, na.rm = TRUE)
  logl_null <- tryCatch(
    .fiml_loglik(data, mu_null, diag(var_null, nrow = m), patterns = patterns),
    error = function(e) NA_real_)
  chi_null <- if (is.na(logl_null)) NA_real_ else max(0, 2 * (logl_sat - logl_null))
  df_null <- m * (m - 1) / 2

  list(chi = chi, chi_null = chi_null, df_null = df_null,
       chi_cfi = chi, chi_null_cfi = chi_null)
}

.gof <- function(L, # The loading/ pattern matrix
                 R, # The correlation matrix
                 N, # The number of cases
                 method, # The estimation method
                 Fm, # Minimized error
                 ci = TRUE, # Compute the analytic RMSEA confidence bounds
                 fiml = NULL) { # EM moments + raw data for the FIML LRT (cor_method = "fiml")
  m <- nrow(L)
  q <- ncol(L)

  # dfs
  df <- .efa_df(m, q)

  # Model-implied (no-uniqueness) correlations LL'; reused for the residual indices
  # below and the chi-square's Sigma, so form the p x p product once.
  LLt <- L %*% t(L)

  ### compute CAF
  delta_hat <- R - LLt
  diag(delta_hat) <- 1
  CAF <- .compute_caf(delta_hat)

  ### compute RMSR
  RMSR <- .rmsr(delta_hat)

  ### compute SRMR (standardized root mean square residual; Bentler, 1995)
  SRMR <- .srmr(delta_hat)

  # Model and baseline chi-square. For cor_method = "fiml" both are likelihood-ratio statistics
  # computed from the EM moments and the raw data (.gof_fiml_chisq); otherwise the Bartlett-
  # corrected ML/ULS discrepancy below. Either way the residual indices above and the shared
  # .chi_fit_indices() call below stay the single source of the rest of the block.
  if (is.null(fiml)) {

    # Model chi-square: the Bartlett-corrected (Bartlett, 1951) ML discrepancy
    # F = tr(Sigma^-1 R) - log|Sigma^-1 R| - p, evaluated at the model-implied correlation
    # matrix Sigma = LL' (unit diagonal), times (N - 1 - (2p + 5)/6 - (2q)/3). For ML this
    # equals the ML objective times the Bartlett multiplier (matching stats::factanal); for
    # ULS the same ML/Wishart discrepancy is evaluated at the ULS-fitted Sigma (matching
    # psych::fa(fm = "uls")), rather than treating the raw least-squares residual sum of
    # squares as the statistic. Its chi-square reference distribution is asymptotically exact
    # under ML and is used here as the conventional approximation for ULS. NA
    # for PAF, missing N, underidentified df, or a non-PD model-implied matrix (e.g. Heywood
    # cases), where the discrepancy is undefined. DWLS is also excluded: the ML discrepancy is
    # not its fit function -- the appropriate categorical statistic is the mean-and-variance-
    # adjusted chi-square from the full asymptotic covariance -- so the chi-square-derived block
    # is left undefined here rather than reported on an inapplicable scale.
    # The Bartlett multiplier goes non-positive for small N relative to the number of
    # variables, which would turn the discrepancy into a negative (meaningless) chi-square;
    # guard it here so the statistic falls through to the chi NA branch below.
    mult <- .bartlett_mult(N, m, q)
    if (!(method %in% c("PAF", "DWLS")) && !is.na(N) && df >= 0 && mult > 0) {
      Sigma <- LLt
      diag(Sigma) <- 1
      # log|R|, reused for the discrepancy below and the CFI/TLI/baseline scale further down.
      ldR <- determinant(R, logarithm = TRUE)
      # discrepancy F (>= 0, clamped); the reported chi-square is F * mult.
      Fchi <- .ml_discrepancy(Sigma, R, m, ldR)
      chi <- Fchi * mult
    } else {
      chi <- NA_real_
      Fchi <- NA_real_
    }

    # null model: reuse the log-determinant already computed for the model chi-square above
    # instead of letting .null_chisq() recompute determinant(R). Undefined (NA) chi-square
    # leaves the whole chi-derived block NA.
    if (is.na(chi)) {
      # chi is already NA_real_ here; NA out the baseline and common-scale quantities to
      # match, so the whole undefined chi-square block is reported as a numeric (NA_real_) NA.
      chi_null <- NA_real_
      df_null <- NA_real_
      chi_cfi <- NA_real_
      chi_null_cfi <- NA_real_
    } else {
      chi_null <- .null_chisq(R, N, ld = ldR)
      df_null <- (m**2 - m) / 2
      # CFI/TLI compare the model and baseline noncentralities and so need both on one
      # scaling constant (Bentler, 1990; Tucker & Lewis, 1973). The reported model and
      # baseline chi-squares keep their own Bartlett corrections (the model matching
      # factanal, the baseline being Bartlett's test of sphericity), but the factor-count
      # term (2q)/3 sits only in the model multiplier and would bias the ratio. Put both on
      # the common (N - 1) scale -- the noncentrality scale RMSEA also uses -- for CFI/TLI.
      chi_cfi <- Fchi * (N - 1)
      chi_null_cfi <- -as.numeric(ldR$modulus) * (N - 1)
    }

  } else {

    fc <- .gof_fiml_chisq(LLt, N, method, df, m, fiml)
    chi <- fc$chi
    chi_null <- fc$chi_null
    df_null <- fc$df_null
    chi_cfi <- fc$chi_cfi
    chi_null_cfi <- fc$chi_null_cfi

  }

  # CFI/TLI/RMSEA/AIC/BIC/ECVI and the p-values. Shared with the scaled-chi-square
  # (sandwich-SE) path, which supplies its own scaled model and baseline statistics
  # (already comparable, so it uses the CFI/TLI scale defaults).
  idx <- .chi_fit_indices(chi, df, chi_null, df_null, N, m, ci,
                          chi_cfi = chi_cfi, chi_null_cfi = chi_null_cfi)

  out <- list(
    chi = chi,
    df = df,
    p_chi = idx$p_chi,
    CAF = CAF,
    RMSR = RMSR,
    SRMR = SRMR,
    CFI = idx$CFI,
    TLI = idx$TLI,
    RMSEA = idx$RMSEA,
    RMSEA_LB = idx$RMSEA_LB,
    RMSEA_UB = idx$RMSEA_UB,
    AIC = idx$AIC,
    BIC = idx$BIC,
    ECVI = idx$ECVI,
    Fm = Fm,
    chi_null = chi_null,
    df_null = df_null,
    p_null = idx$p_null
  )

  # cor_method = "fiml": replace the plain two-stage likelihood-ratio chi-square (computed above)
  # with the asymptotically-correct Satorra-Bentler-corrected two-stage statistic (Yuan, Marshall,
  # & Bentler, 2002), built on the saturated FIML asymptotic covariance. The plain LRT is referenced
  # to chi^2(df), which is not the two-stage estimator's reference distribution, so the p-value and
  # the noncentrality-based CFI/TLI/RMSEA are biased. Applied to every FIML fit (point estimate and
  # bootstrap replicates), independent of `se`; .fiml_scaled_test() returns NULL (PAF, a just/under-
  # identified model, or a degenerate/non-PD saturated covariance), where the NA / likelihood-ratio
  # block above stands.
  if (!is.null(fiml)) {
    # AIC/BIC/ECVI are likelihood-ratio information criteria with no standard interpretation under
    # the corrected two-stage statistic; leave them NA for every FIML fit. .apply_scaled_test() also
    # NAs them on the scaled path, but it does not run on the plain-LRT and just-identified (df == 0)
    # fallbacks, so NA them here too rather than ship a chi-square-derived value the two-stage
    # estimator cannot support.
    out$AIC <- NA_real_
    out$BIC <- NA_real_
    out$ECVI <- NA_real_
    st <- .fiml_scaled_test(L, R, N, method, df, m, fiml)
    if (!is.null(st)) {
      # Thread `ci` on: the scaled block replaces the whole chi-square-derived block, so it must
      # honour the caller's request not to solve the RMSEA bounds -- which is exactly the
      # per-replicate bootstrap path, where the bounds are computed and then discarded.
      out <- .apply_scaled_test(out, st, N, ci = ci)
    } else if (method %in% c("ML", "ULS") && !is.na(df) && df > 0 && !is.na(out$chi)) {
      # The correction could not be formed although the model has a testable chi-square: the
      # saturated FIML covariance is degenerate, which happens on small, heavily incomplete or
      # near-collinear samples. Keep the plain two-stage likelihood-ratio statistic -- it is
      # informative, and NA-ing the whole block would discard a usable (if approximate) test --
      # but tag it, so the reported statistic is never read as the corrected one. Every consumer
      # of the chi-square block branches on this tag: the print method labels the line from it,
      # and the multiple-imputation pooler reads the withheld AIC alongside it. `chi_scaling`,
      # `chi_shift` and the alternative scaled statistics are absent here, because no scaling was
      # applied. The just-identified (df == 0) case is deliberately excluded: it has no
      # chi-square test to mislabel.
      out$chi_scaled_type <- "uncorrected.lrt"
      cli::cli_warn(
        c("The corrected two-stage chi-square could not be formed; the plain two-stage
           likelihood-ratio statistic is reported instead.",
          "i" = "Referenced to a chi-square({df}) distribution it is only approximate under the
                 two-stage estimator, so read the p-value, CFI, TLI, and RMSEA as indicative.",
          "i" = "This follows a degenerate saturated covariance, usually from a small sample with
                 a high proportion of missing values, or from near-collinear variables.",
          "i" = "The reported statistic is identified by {.code fit_indices$chi_scaled_type ==
                 \"uncorrected.lrt\"}."),
        class = "efa_fiml_uncorrected_chisq"
      )
    }
  }

  out

}

# Unit tests for the fit-index helpers in R/fit-indices.R: the variance-accounted
# table (.compute_vars), the goodness-of-fit block (.gof, including the CFI
# noncentrality truncation, the small-N Bartlett guard, and an unlocatable RMSEA
# bound), and the common-part accounted-for index (.compute_caf). The ML/ULS/PAF
# fixtures are fitted once at the top from a seeded random matrix so every .gof
# branch is exercised on the same solutions.

efa_temp <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500)
efa_pro <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                rotation = "promax")
test_that(".compute_vars works", {
  checkmate::expect_matrix(
    .compute_vars(efa_temp$unrot_loadings, efa_temp$unrot_loadings)
  )
  checkmate::expect_matrix(
    .compute_vars(efa_pro$rot_loadings, efa_pro$unrot_loadings, efa_pro$Phi)
  )
})

test_that(".compute_vars drops the cumulative rows for a single factor", {
  # The documented row set: the three cumulative/common-variance rows are defined only for
  # more than one factor, so the table has two rows at k = 1 and five otherwise. User code
  # that indexes a row by name depends on this, so it is part of the contract.
  five <- .compute_vars(efa_temp$unrot_loadings, efa_temp$unrot_loadings)
  expect_identical(rownames(five),
                   c("SS loadings", "Prop Tot Var", "Cum Prop Tot Var", "Prop Comm Var",
                     "Cum Prop Comm Var"))

  one <- .compute_vars(efa_temp$unrot_loadings[, 1, drop = FALSE],
                       efa_temp$unrot_loadings[, 1, drop = FALSE])
  expect_identical(rownames(one), c("SS loadings", "Prop Tot Var"))
})

set.seed(42)
efa_ml <- suppressWarnings(EFA(cbind(rnorm(100), rnorm(100), rnorm(100), rnorm(100),
                                     rnorm(100), rnorm(100)), 3, N = 500,
                               method = "ML"))
efa_uls <- suppressWarnings(EFA(cbind(rnorm(100), rnorm(100), rnorm(100), rnorm(100),
                    rnorm(100), rnorm(100)), 3, method = "ULS"))
efa_paf <- suppressWarnings(EFA(cbind(rnorm(100), rnorm(100), rnorm(100), rnorm(100),
                    rnorm(100), rnorm(100)), 3, method = "PAF"))

# .gof() derives CAF from a KMO on the residual matrix and signals the classed
# efa_caf_failed warning when that solve is unusable, in which case CAF falls back to 0
# instead of the ~.5 these near-null fixtures otherwise give. Which of the two branches ran
# is therefore part of what the expectations below need to know, so record it alongside the
# value. withCallingHandlers reads the condition without unwinding, so one evaluation yields
# both; tryCatch(warning = ) aborts the call and would force .gof() to be run a second time.
# Keying the handler on the class rather than on "warning" means an unrelated warning cannot
# silently divert the CAF expectation to the wrong branch.
.gof_caf <- function(...) {
  caf_failed <- FALSE
  gof <- withCallingHandlers(
    .gof(...),
    efa_caf_failed = function(w) {
      caf_failed <<- TRUE
      invokeRestart("muffleWarning")
    }
  )
  list(gof = gof, caf_failed = caf_failed)
}

res_ml <- .gof_caf(efa_ml$unrot_loadings, efa_ml$orig_R, efa_ml$settings$N,
                   "ML", efa_ml$fit_indices$Fm)
gof_ml <- res_ml$gof
res_uls <- .gof_caf(efa_uls$unrot_loadings, efa_uls$orig_R, efa_uls$settings$N,
                    "ULS", efa_uls$fit_indices$Fm)
gof_uls <- res_uls$gof
res_paf <- .gof_caf(efa_paf$unrot_loadings, efa_paf$orig_R, efa_paf$settings$N,
                    "PAF", NA)
gof_paf <- res_paf$gof
m <- 6 # n variables
q <- 3 # n factors
test_that(".gof works", {
  expect_type(gof_ml, "list")
  expect_named(gof_ml,
               c("chi", "df", "p_chi", "CAF", "RMSR", "SRMR", "CFI", "TLI",
                 "RMSEA", "RMSEA_LB", "RMSEA_UB", "AIC", "BIC", "ECVI", "Fm",
                 "chi_null", "df_null", "p_null"))
  expect_true(is.na(gof_ml$p_chi))    # df == 0: the chi-square test is undefined
  expect_equal(gof_ml$CFI, 1)
  expect_equal(gof_ml$RMSEA, 0)
  if (res_ml$caf_failed) {
    expect_equal(gof_ml$CAF, 0, tolerance = .01)
  } else {
    expect_equal(gof_ml$CAF, .5, tolerance = .1)
  }
  expect_equal(gof_ml$df, ((m - q)**2 - (m + q)) / 2)

  expect_type(gof_uls, "list")
  expect_named(gof_uls,
               c("chi", "df", "p_chi", "CAF", "RMSR", "SRMR", "CFI", "TLI",
                 "RMSEA", "RMSEA_LB", "RMSEA_UB", "AIC", "BIC", "ECVI", "Fm",
                 "chi_null", "df_null", "p_null"))
  expect_true(is.na(gof_uls$p_chi))   # df == 0: the chi-square test is undefined
  expect_equal(gof_uls$CFI, 1)
  expect_equal(gof_uls$RMSEA, 0)
  if (res_uls$caf_failed) {
    expect_equal(gof_uls$CAF, 0, tolerance = .01)
  } else {
    expect_equal(gof_uls$CAF, .5, tolerance = .1)
  }

  expect_equal(gof_uls$df, ((m - q)**2 - (m + q)) / 2)

  expect_type(gof_paf, "list")
  expect_named(gof_paf,
               c("chi", "df", "p_chi", "CAF", "RMSR", "SRMR", "CFI", "TLI",
                 "RMSEA", "RMSEA_LB", "RMSEA_UB", "AIC", "BIC", "ECVI", "Fm",
                 "chi_null", "df_null", "p_null"))
  expect_equal(gof_paf$chi, NA_real_)
  expect_equal(gof_paf$p_chi, NA_real_)
  expect_equal(gof_paf$CFI, NA_real_)
  expect_equal(gof_paf$RMSEA, NA_real_)
  if (res_paf$caf_failed) {
    expect_equal(gof_paf$CAF, 0, tolerance = .01)
  } else {
    expect_equal(gof_paf$CAF, .5, tolerance = .1)
  }
  expect_equal(gof_paf$df, ((m - q)**2 - (m + q)) / 2)
  expect_equal(gof_paf$chi_null, NA_real_)
  expect_equal(gof_paf$df_null, NA_real_)
  expect_equal(gof_paf$p_null, NA_real_)
})


test_that("the documented fit-index definitions are the ones computed", {
  # The help page states df, df_null, the ECVI formula, and the RMSR/SRMR scaling relation
  # explicitly; pin each so the documentation and the code cannot drift apart.
  R <- test_models$baseline$cormat
  p <- ncol(R)
  N <- 500
  fit <- suppressWarnings(suppressMessages(
    EFA(R, n_factors = 3, N = N, method = "ML")))
  fi <- fit$fit_indices

  # df = ((p - k)^2 - (p + k))/2, and the independence baseline df_null = p(p - 1)/2
  expect_equal(fi$df, ((p - 3)^2 - (p + 3)) / 2)
  expect_equal(fi$df_null, p * (p - 1) / 2)
  # chi_null is documented as Bartlett's test of sphericity, so it must be that statistic and
  # not merely some baseline that p_null happens to be a chi-square tail of.
  expect_equal(fi$chi_null, efa_bartlett(R, N = N)$chisq)

  # ECVI = (chi^2 + 2q)/(N - 1) with q = p(p + 1)/2 - df free parameters, built on the
  # reported (Bartlett-corrected) chi-square rather than the uncorrected Browne-Cudeck one.
  n_params <- p * (p + 1) / 2 - fi$df
  expect_equal(fi$ECVI, (fi$chi + 2 * n_params) / (N - 1))

  # AIC and BIC are the chi-square-based forms, on the same corrected statistic
  expect_equal(fi$AIC, fi$chi - 2 * fi$df)
  expect_equal(fi$BIC, fi$chi - log(N) * fi$df)

  # SRMR = RMSR * sqrt((p - 1)/(p + 1)); psych's rms is RMSR/sqrt(2)
  expect_equal(fi$SRMR, fi$RMSR * sqrt((p - 1) / (p + 1)))
})


test_that(".gof CFI uses the Bentler noncentrality truncation (matches lavaan)", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  # Random near-uncorrelated data: a one-factor model barely improves on the
  # baseline, so the model noncentrality (chi - df) is negative. The Bentler
  # convention floors it at 0 (CFI = 1); the previous unfloored ratio deflated
  # CFI toward 0 (here it returned about 0.21).
  set.seed(1)
  X <- matrix(stats::rnorm(100 * 7), 100, 7)
  R <- stats::cor(X)
  colnames(R) <- rownames(R) <- paste0("V", seq_len(7))

  cfi_efa <- suppressWarnings(
    EFA(R, n_factors = 1, N = 100, method = "ML")
  )$fit_indices$CFI

  mod <- paste0("f =~ ", paste(colnames(R), collapse = " + "))
  fit_l <- lavaan::cfa(mod, sample.cov = R, sample.nobs = 100, std.lv = TRUE)
  cfi_lav <- unname(lavaan::fitMeasures(fit_l, "cfi"))

  expect_equal(cfi_efa, cfi_lav, tolerance = 1e-3)
  expect_equal(cfi_efa, 1)
})


test_that(".gof guards the Bartlett multiplier against tiny N (no negative chi)", {
  # For small N relative to the number of variables the model multiplier
  # N - 1 - (2m + 5)/6 - (2q)/3 turns non-positive (here N = 5, m = 6, q = 3).
  # The chi-square must then be NA, not a negative statistic masquerading as
  # perfect fit (p_chi -> 1, RMSEA floored to 0, CFI from a meaningless number).
  gof_tiny <- .gof(efa_ml$unrot_loadings, efa_ml$orig_R, N = 5, "ML",
                   efa_ml$fit_indices$Fm)
  expect_true(is.na(gof_tiny$chi))
  expect_true(is.na(gof_tiny$p_chi))
  expect_true(is.na(gof_tiny$CFI))
  expect_true(is.na(gof_tiny$RMSEA))

  # The null (baseline) multiplier N - 1 - (2p + 5)/6 is guarded the same way.
  expect_true(is.na(.null_chisq(efa_ml$orig_R, N = 3)))

  # A missing N must propagate NA, not crash the `if (mult <= 0)` guard with
  # if(NA), and an undefined chi-square has no RMSEA noncentrality bound.
  expect_true(is.na(.null_chisq(efa_ml$orig_R, N = NA_real_)))
  expect_true(is.na(.rmsea_lambda(NA_real_, df = 10, goal = .95)))
  # An undefined df has no bound either, and must not reach the `if (pchisq(NA) >= goal)`
  # comparison.
  expect_true(is.na(.rmsea_lambda(20, df = NA_real_, goal = .95)))
})


test_that("an unlocatable RMSEA noncentrality bound is NA, not an error", {
  # A statistic whose tail probability never crosses the target quantile leaves the
  # noncentrality root unbracketed even after extendInt: the bound is undefined, so it is
  # reported as NA rather than aborting the fit. chi = Inf makes goal - pchisq(Inf, df, ncp)
  # negative for every ncp, which is exactly that case.
  expect_true(is.na(.rmsea_lambda(Inf, df = 10, goal = .95)))

  # An undefined common-scale statistic must survive the caps in .chi_fit_indices() as NA
  # rather than failing an `if (NA > 1)` comparison, for the point estimate and both
  # bounds. The reported chi-square stays finite here, so only the RMSEA block is undefined.
  idx <- .chi_fit_indices(chi = 30, df = 10, chi_null = 300, df_null = 45, N = 200,
                          m = 10, ci = TRUE, chi_cfi = NA_real_, chi_null_cfi = 300)
  expect_true(is.na(idx$RMSEA))
  expect_true(is.na(idx$RMSEA_LB))
  expect_true(is.na(idx$RMSEA_UB))
  expect_true(is.finite(idx$p_chi))
})


test_that("a non-convergent RMSEA noncentrality solve is contained, not leaked or dressed up", {
  # stats::pchisq() stops converging for noncentralities in the millions and emits one base
  # warning per evaluation, so a single interval solve used to leak dozens of them out of an
  # efa_fit() call. The solver contains them.
  expect_no_warning(lambda <- .rmsea_lambda(4101811, df = 135, goal = .95))
  expect_true(is.finite(lambda))

  # The bounds that same solve produces collapse onto a single value below the point
  # estimate. An interval that does not contain its own point estimate is not an interval:
  # both bounds are reported as undefined rather than clipped or reordered into something
  # that reads as usable. The point estimate itself is unaffected.
  idx <- .chi_fit_indices(chi = 4101811, df = 135, chi_null = 1e8, df_null = 153,
                          N = 5e6, m = 18, ci = TRUE,
                          chi_cfi = 4101811, chi_null_cfi = 1e8)
  expect_true(is.na(idx$RMSEA_LB))
  expect_true(is.na(idx$RMSEA_UB))
  expect_true(is.finite(idx$RMSEA))

  # ... and the whole fit stays quiet and consistent at that N.
  expect_no_warning(
    fit_huge <- suppressMessages(
      efa_fit(test_models$baseline$cormat, n_factors = 1, N = 5e6, estimator = "ML"))
  )
  expect_true(is.finite(fit_huge$fit_indices$RMSEA))
  expect_true(is.na(fit_huge$fit_indices$RMSEA_LB))
  expect_true(is.na(fit_huge$fit_indices$RMSEA_UB))
})


test_that("an ordinary RMSEA interval is preserved and brackets its point estimate", {
  idx <- .chi_fit_indices(chi = 200, df = 102, chi_null = 2000, df_null = 153, N = 500,
                          m = 18, ci = TRUE, chi_cfi = 200, chi_null_cfi = 2000)
  expect_true(is.finite(idx$RMSEA_LB))
  expect_true(is.finite(idx$RMSEA_UB))
  expect_lte(idx$RMSEA_LB, idx$RMSEA)
  expect_lte(idx$RMSEA, idx$RMSEA_UB)
  expect_gt(idx$RMSEA_UB, idx$RMSEA_LB)

  # ci = FALSE still returns no bounds at all, and a just-identified model still reports the
  # degenerate zero interval rather than being caught by the bracketing check.
  no_ci <- .chi_fit_indices(chi = 200, df = 102, chi_null = 2000, df_null = 153, N = 500,
                            m = 18, ci = FALSE, chi_cfi = 200, chi_null_cfi = 2000)
  expect_true(is.na(no_ci$RMSEA_LB))
  expect_true(is.na(no_ci$RMSEA_UB))

  just_id <- .chi_fit_indices(chi = 5, df = 0, chi_null = 2000, df_null = 153, N = 500,
                              m = 18, ci = TRUE, chi_cfi = 5, chi_null_cfi = 2000)
  expect_equal(just_id$RMSEA_LB, 0)
  expect_equal(just_id$RMSEA_UB, 0)
})


test_that("the shared unavailability wording names the block, the residuals, and the multiplier", {
  # One source for the sentences efa_fit(), efa_bartlett(), efa_screen() and their print
  # methods use, so a reader is told the same thing wherever the chi-square block goes away.
  expect_match(.fit_unavailable_text("chisq_block"), "CFI, TLI, RMSEA, AIC, BIC, ECVI",
               fixed = TRUE)
  expect_match(.fit_unavailable_text("residuals_kept"), "CAF, RMSR, SRMR", fixed = TRUE)
  expect_match(.fit_unavailable_text("residuals_not_identification"), "identified",
               fixed = TRUE)

  # The multiplier is named with the factor-count term only where there are factors, and
  # with the caller's numbers only where the caller has them.
  expect_match(.fit_unavailable_text("bartlett_mult", N = 8, p = 18, q = 3),
               "N - 1 - (2p + 5)/6 - 2q/3 is not positive for N = 8, p = 18, and q = 3",
               fixed = TRUE)
  expect_match(.fit_unavailable_text("bartlett_mult", N = 5, p = 18),
               "N - 1 - (2p + 5)/6 is not positive for N = 5 and p = 18", fixed = TRUE)
  expect_match(.fit_unavailable_text("bartlett_mult"), "N - 1 - (2p + 5)/6", fixed = TRUE)

  # A mistyped key is a caller bug, not a silently empty message.
  expect_error(.fit_unavailable_text("not_a_key"))
})


test_that("RMSR and SRMR keep their fixed relation, and both drop out together on an NA", {
  # Both summarise the same p(p - 1)/2 off-diagonal residuals and differ only in their
  # denominator, so SRMR = RMSR * sqrt((p - 1)/(p + 1)) exactly. That is the relation the
  # print methods rely on when they show only one of the two.
  set.seed(11)
  p <- 7
  E <- matrix(0, p, p)
  E[upper.tri(E)] <- stats::rnorm(p * (p - 1) / 2, sd = .05)
  E <- E + t(E)

  expect_equal(.srmr(E), .rmsr(E) * sqrt((p - 1) / (p + 1)))
  expect_equal(.rmsr(E), sqrt(sum(E[upper.tri(E)]^2) / (p * (p - 1) / 2)))

  # A residual that is not available makes the sum, and therefore both summaries, undefined.
  # Averaging RMSR over the surviving pairs would report a number describing fewer variable
  # pairs than it appears to, and would break the relation above.
  E_na <- E
  E_na[1, 3] <- E_na[3, 1] <- NA_real_
  expect_true(is.na(.rmsr(E_na)))
  expect_true(is.na(.srmr(E_na)))

  # An NA on the diagonal is not a residual and must not reach either summary.
  E_diag <- E
  diag(E_diag) <- NA_real_
  expect_equal(.rmsr(E_diag), .rmsr(E))
  expect_equal(.srmr(E_diag), .srmr(E))
})


test_that(".compute_caf returns 0 (with warning) when KMO is not computable", {
  # Hollow residual matrix: solve() succeeds (not a try-error) but the inverse
  # has a negative diagonal, so KMO is NaN. CAF must fall back to 0, not NaN.
  delta_hat <- matrix(c(0, .5, .5, .5, 0, .5, .5, .5, 0), 3)
  expect_warning(caf <- .compute_caf(delta_hat), class = "efa_caf_failed")
  expect_equal(caf, 0)
})


rm(efa_pro, efa_temp, efa_ml, efa_uls, efa_paf, gof_ml, gof_uls, gof_paf,
   res_ml, res_uls, res_paf, .gof_caf, m, q)

#' DOSPERT_raw
#'
#' A data.frame containing responses to the risk subscale of the Domain Specific
#' Risk Taking Scale (DOSPERT; Weber, Blais, & Betz, 2002) based on the publicly
#' available dataset (at <https://osf.io/pjt57/>) by Frey, Duncan, and Weber (2023).
#' The items measure risk-taking propensity on five different domains: social,
#' recreational, financial, health/ safety, and ethical.
#'
#' @format A data.frame with 3,123 rows (participants) and 30 columns, named by a
#' domain prefix and item number, with six items in each of five risk domains:
#' \describe{
#'   \item{ethR_1 to ethR_6}{(numeric) - Ethical-domain risk-taking items.}
#'   \item{finR_1 to finR_6}{(numeric) - Financial-domain risk-taking items.}
#'   \item{heaR_1 to heaR_6}{(numeric) - Health/safety-domain risk-taking items.}
#'   \item{recR_1 to recR_6}{(numeric) - Recreational-domain risk-taking items.}
#'   \item{socR_1 to socR_6}{(numeric) - Social-domain risk-taking items.}
#'  }
#' @source Blais, A.-R., & Weber, E. U. (2006). A domain-specific risk-taking (DOSPERT) scale for adult populations. Judgment and Decision Making, 1(1), 33–47. doi: 10.1017/S1930297500000334
#' @source Frey, R., Duncan, S. M., & Weber, E. U. (2023). Towards a typology of risk preference: Four risk profiles describe two-thirds of individuals in a large sample of the U.S. population. Journal of Risk and Uncertainty, 66(1), 1–17. doi:10.1007/s11166-022-09398-5
#'
"DOSPERT_raw"

# Dispatch standard-error/confidence-interval computation on the requested method. The
# nonparametric bootstrap aggregates resampled fits; the analytic methods derive SEs from
# the fitted model itself. Every branch returns the same list(SE, CI, replicates) schema, so
# the print and summary methods are agnostic to how the SEs were produced. The sandwich branch
# additionally returns a `scaled_test` element (the robust scaled chi-square), which efa_fit()
# strips off and folds into the fit indices. `gamma` (the polychoric ACOV meat) and `method`
# are used only by the sandwich. `fiml` (the two-stage EM moments + raw data) is non-NULL only
# for cor_method = "fiml", where both analytic settings route to the corrected two-stage sandwich.
.compute_se_ci <- function(fit_out, L_rot, se_method, boot_fits = NULL,
                           boot_rot = "none", ci = .95, b = NULL, N = NULL,
                           rot_info = NULL, gamma = NULL, method = NULL,
                           fiml = NULL) {

  # cor_method = "fiml": the EM correlation is a two-stage estimate, so both analytic settings
  # ("information"/"sandwich") return the corrected two-stage sandwich SE built on the saturated
  # FIML covariance. The naive Stage-2 information SE (treating the EM correlation as complete data)
  # is inconsistent under missingness and is never shipped. The bootstrap path is unchanged.
  if (!is.null(fiml) && se_method %in% c("information", "sandwich")) {
    return(.se_fiml(fit_out, rot_info, N, ci, fiml, method))
  }

  switch(se_method,
    "np-boot" = .boot_se_ci(fit_out, L_rot, boot_fits, boot_rot, ci, b),
    # rot_info is non-NULL exactly for an analytic-SE fit under a real rotation; the unrotated
    # path (rotation = "none") leaves it NULL. Both are handled inside .se_information().
    "information" = .se_information(fit_out, rot_info, N, ci, method),
    "sandwich" = .se_sandwich_dispatch(fit_out, rot_info, N, ci, gamma, method),
    NULL
  )
}


# Leading `keep`-block of the inverse of the bordered matrix [A C'; C 0] -- the constrained
# (reflexive generalised) inverse of a gauge-singular information matrix A under the identification
# constraints whose Jacobian is `Cmat` (k(k-1)/2 rows, zero rows for a single factor). Returns NULL
# if the augmented system is singular. Shared by the expected-information and sandwich SEs.
.bordered_inverse_block <- function(A, Cmat, keep) {
  nc <- nrow(Cmat)
  Aug <- if (nc > 0L) {
    rbind(cbind(A, t(Cmat)), cbind(Cmat, matrix(0, nc, nc)))
  } else {
    A
  }
  inv <- tryCatch(solve(Aug), error = function(e) NULL)
  if (is.null(inv)) return(NULL)
  inv[seq_len(keep), seq_len(keep), drop = FALSE]
}

# Delta-method SEs of the communalities h2_i = rowSums(Lambda^2)_i from a loading covariance V
# (p*k x p*k over column-major vec(Lambda)). The gradient of h2_i is 2 Lambda[i, ], nonzero only in
# variable i's loading columns. Shared by every analytic path: psi_i = 1 - h2_i is a function of
# Lambda, not a free parameter, so this single route supplies the coinciding uniqueness and
# communality SEs from whichever loading covariance the path built.
.communality_se <- function(L, V) {
  p <- nrow(L)
  k <- ncol(L)
  G_h <- matrix(0, p, p * k)
  for (i in seq_len(p)) G_h[i, (seq_len(k) - 1L) * p + i] <- 2 * L[i, ]
  sqrt(pmax(rowSums((G_h %*% V) * G_h), 0))
}

# Skeleton gradient of the rotational identification constraint for the factor pair (u, v): a
# p x k matrix with column u set to X[, v] and column v to X[, u]. The off-diagonal (u, v) entry
# of X' Lambda has exactly this gradient in vec(Lambda) -- with X = Lambda it is the gradient of
# off-diag(Lambda' Lambda), and with X = Psi^-1 Lambda the direct part of off-diag(Lambda' Psi^-1
# Lambda); the caller adds the psi(Lambda) chain-rule term. Used by `.se_sandwich_constraint()`,
# which builds the gauge constraint for every analytic path.
.gauge_grad <- function(X, u, v) {
  g <- matrix(0, nrow(X), ncol(X))
  g[, u] <- X[, v]
  g[, v] <- X[, u]
  g
}

# TRUE if M is a (numerically) positive-semidefinite covariance: finite, with a non-negative
# diagonal and no eigenvalue below a small tolerance scaled by its largest diagonal entry. Used to
# gate the analytic parameter covariances before their square roots are reported: checking only the
# diagonal would pass a covariance that is non-PSD off the diagonal (where `pmax(., 0)` would then
# silently floor a negative rotated variance to an understated SE rather than flag it).
.is_psd <- function(M) {
  if (!all(is.finite(M))) return(FALSE)
  d <- diag(M)
  if (any(d < 0)) return(FALSE)
  sym <- (M + t(M)) / 2
  ev <- tryCatch(min(eigen(sym, symmetric = TRUE, only.values = TRUE)$values),
                 error = function(e) NA_real_)
  is.finite(ev) && ev >= -1e-8 * max(d, 1)
}

# TRUE when the solution sits on the parameter-space boundary: a uniqueness at (or below) its lower
# boundary, i.e. the Heywood case `.finalize_fit()` flags. The Wald approximation every analytic
# standard error rests on is not valid there, so this is the single gate on which the analytic paths
# withhold. The ML and ULS fitters constrain the uniquenesses to [.uniqueness_floor, 1] (DWLS does
# not -- see R/DWLS.R -- and can land below zero), so an improper solution from those fitters is
# pinned AT that floor and never reaches zero: a gate keyed on `psi <= 0` would fire only for a
# hand-built covariance and never for a fitted one. Test the boundary the fitters can actually
# reach.
.at_uniqueness_boundary <- function(L) {
  psi <- 1 - rowSums(unclass(L)^2)
  anyNA(psi) || any(psi <= .uniqueness_floor + sqrt(.Machine$double.eps))
}

# The hint the analytic paths attach to `efa_se_unreliable` when the boundary above is what withheld
# the standard errors. Held next to the predicate, and shared by the rotated and unrotated message
# sites, so the wording cannot drift between them. It states a total withholding, so a caller must
# only reach for it once it has established that nothing was reported.
.se_boundary_hint <- paste0(
  "The solution is a Heywood case -- a uniqueness sits at its lower boundary -- where the Wald ",
  "approximation these standard errors rest on is not valid, so none is reported. Use ",
  "{.code se = \"np-boot\"}, or extract fewer factors."
)

# Whether that hint may be attached to a withholding -- a narrower question than the gate above,
# and the reason the two are separate predicates rather than one.
#
# The gate deliberately also fires on a non-finite solution (`anyNA(psi)`): a Wald interval is no
# more valid there, so the standard errors must go either way. But nothing is at any boundary in
# that case -- the parameter covariance is simply undefined -- so the hint's claim that the
# solution is a Heywood case would be false, and "extract fewer factors" would send the user after
# a problem the fit does not have. Those withholdings are described by the callers' general clause
# (a singular information matrix or an unusable covariance) instead, which is what they are.
# Held beside the hint so the wording and the condition it asserts cannot drift apart.
.se_boundary_hint_applies <- function(L) {
  all(is.finite(L)) && .at_uniqueness_boundary(L)
}

# Apply the boundary policy to a core's standard errors: at a solution on the parameter-space
# boundary the Wald approximation fails for every parameter however the covariance was estimated,
# so every analytic path withholds on the same test and the policy lives here rather than being
# restated at each of them. The marginals go, `reliable` goes with them (which makes the consumers
# NA-fill the persisted covariance to match), and `gauge_reliable` is RESET to TRUE because the
# gauge is not what withheld them -- the caller must not attach its gauge-specific message. What is
# deliberately NOT touched is the scaled chi-square: it is a discrepancy-function quantity rather
# than a Wald one, the boundary does not invalidate it, and for DWLS it is the only chi-square block
# the fit has. Returns the core unchanged at an interior solution.
.withhold_at_boundary <- function(core, L) {
  if (!.at_uniqueness_boundary(L)) return(core)
  core$loadings_se[] <- NA_real_
  core$uniquenesses_se[] <- NA_real_
  core$reliable <- FALSE
  core$gauge_reliable <- TRUE
  core
}

# Row/column labels for a vec(Lambda)-ordered loading covariance: "<variable>_<factor>" in the same
# column-major order the block is assembled in, so the documented ordering can be read off the
# object. NULL when the loadings carry no dimnames, in which case the block ships unlabelled rather
# than mislabelled.
.vec_loading_labels <- function(L) {
  rn <- rownames(L)
  cn <- colnames(L)
  if (is.null(rn) || is.null(cn)) return(NULL)
  as.vector(outer(rn, cn, paste, sep = "_"))
}


# Analytic expected-information SEs and Wald CIs for the ML solution, rotated or unrotated.
#
# EFAtools fits a CORRELATION structure: the analysed matrix has a diagonal fixed at exactly 1, and
# the uniquenesses psi = 1 - rowSums(Lambda^2) are a function of Lambda rather than free parameters.
# The Fisher information is therefore the one for theta = vec(Lambda) over the off-diagonal
# correlations alone,
#
#   I(theta) = Delta' Gamma^{-1} Delta,
#
# with Delta = d sigma_offdiag / d vec(Lambda) and Gamma the normal-theory asymptotic covariance of
# the sample correlations evaluated at the model-implied Sigma (Cudeck, 1989; Ogasawara, 1998).
# Assembling the covariance-structure information over (vec(Lambda), psi) instead would attribute
# Wishart sampling variability to a diagonal that carries none, and leak it into the loading block.
#
# I(theta) is exactly the Godambe sandwich of `.se_sandwich_core()` at the optimal (efficient)
# weight V = Gamma^{-1}, where the meat Delta' V Gamma V Delta collapses back onto the bread
# Delta' V Delta. Routing through the sandwich therefore reuses the model Jacobian, the gauge
# constraint, the bordered inverse, the PSD gate and the rotation propagation rather than
# duplicating them. No scaled chi-square is requested: the normal-theory weight is the one the ML
# discrepancy already minimises, so the ordinary chi-square stands.
#
# Cost: Gamma is n x n over the n = p(p-1)/2 pairs and is inverted once, so this is O(n^3) rather
# than the O((p k + p)^3) of a covariance-structure assembly -- a one-shot 0.1 s at p = 18 and under
# a second at p = 40, far below a bootstrap.
.se_information <- function(fit_out, rot_info, N, ci, method) {

  L <- unclass(fit_out$unrot_loadings)

  # At a boundary solution (`.at_uniqueness_boundary()`) the Wald intervals this path reports are
  # not valid, so withhold the covariance: the core then NA-fills through its usual unusable-Gamma
  # branch and the shared efa_se_unreliable warning fires, rather than reporting a boundary standard
  # error. Short-circuiting here also skips assembling and inverting an n x n Gamma the fit would
  # never use; `.se_sandwich_dispatch()` applies the same gate for the paths that still need the
  # core (it builds the scaled chi-square, which stays reportable at a boundary fit).
  Gamma <- if (.at_uniqueness_boundary(L)) {
    NULL
  } else {
    # Model-implied correlation matrix (unit diagonal by construction).
    Sigma <- tcrossprod(L)
    diag(Sigma) <- 1
    # `.se_sandwich_core()` takes Gamma on the variance scale Var(rho-hat) and restores the unit
    # scale itself, so undo the unit scaling `.normal_theory_gamma()` returns.
    .normal_theory_gamma(Sigma, utils::combn(nrow(L), 2L)) / N
  }

  .se_sandwich_dispatch(fit_out, rot_info, N, ci, Gamma, method,
                        optimal_weight = TRUE, scaled = FALSE)
}


# Map a rotation name to the warm-start criterion family and tuning argument used by the compiled
# `.rotation_se_jacobian()`, mirroring the engine selection in `.orth_engines`/`.oblq_engines`
# (R/rotate_model.R). varimax shares the Crawford-Ferguson criterion at kappa = 1 / p, quartimax at
# kappa = 0, equamax at kappa = k / (2 p); quartimin is oblimin at gam = 0. The geomin offset
# `delta` and oblimin `gam` are taken from the resolved criterion arguments (the same `.gpf_crit`
# defaults the rotation itself used). promax and simplimax have no usable analytic Jacobian and are
# rejected before reaching here, so an unrecognised rotation returns NULL.
.rotation_se_method <- function(rotation, p, k, crit_args = list()) {
  gam <- if (is.null(crit_args$gam)) 0 else crit_args$gam
  delta <- if (is.null(crit_args$delta)) 0.01 else crit_args$delta
  switch(rotation,
    varimax   = list(method = "cf",       param = 1 / p,       oblique = FALSE),
    quartimax = list(method = "cf",       param = 0,           oblique = FALSE),
    equamax   = list(method = "cf",       param = k / (2 * p), oblique = FALSE),
    bentlerT  = list(method = "bentler",  param = 0,           oblique = FALSE),
    geominT   = list(method = "geomin",   param = delta,       oblique = FALSE),
    bifactorT = list(method = "bifactor", param = 0,           oblique = FALSE),
    oblimin   = list(method = "oblimin",  param = gam,         oblique = TRUE),
    quartimin = list(method = "oblimin",  param = 0,           oblique = TRUE),
    bentlerQ  = list(method = "bentler",  param = 0,           oblique = TRUE),
    geominQ   = list(method = "geomin",   param = delta,       oblique = TRUE),
    bifactorQ = list(method = "bifactor", param = 0,           oblique = TRUE),
    NULL)
}

# Wald lower/upper interval around an estimate from its standard error. `est` is unclassed so the
# returned bounds are plain numeric matrices/vectors (the rotated estimates carry the "LOADINGS"
# class).
.wald_ci <- function(est, se, z) {
  est <- unclass(est)
  list(lower = est - z * se, upper = est + z * se)
}


# Analytic standard errors and Wald CIs for an obliquely or orthogonally ROTATED solution. The
# caller's covariance of the unrotated loadings (`se0`) is propagated through the rotation by the
# delta method; every analytic path -- expected information, robust sandwich, and the corrected
# two-stage FIML sandwich -- supplies its own, so the propagation is written once. The rotation
# maps the unrotated loadings A
# to the rotated pattern L = A T(A)^{-1 T} (oblique) / A T(A) (orthogonal) and, for oblique
# rotations, the factor correlations Phi = T(A)' T(A); both depend on A alone (not the
# uniquenesses), so only the p*k loading block V of the parameter covariance is needed. The
# rotation Jacobians d vec(L) / d vec(A) and d vec(Phi) / d vec(A) are obtained by
# finite-differencing a warm-started re-rotation of A (`.rotation_se_jacobian()`), which re-solves
# the rotation from the converged transformation so the optimum tracks each perturbation smoothly.
# This is the delta-method standard error of Jennrich (1973), matching lavaan's rotation.se =
# "delta".
# Because rotated quantities are identification-invariant they are comparable across packages
# (unlike the unrotated loadings, whose orientation is identification-dependent).
#
#   Var(vec L)   = J_L V J_L'                                   -> rotated-loading SEs
#   Var(vec Phi) = J_Phi V J_Phi'                              -> factor-correlation SEs (oblique)
#   S = L Phi:  J_S = (Phi' (x) I_p) J_L + (I_k (x) L) J_Phi;  Var(vec S) = J_S V J_S'  (oblique)
#   h2_i = rowSums(A^2)_i = 1 - psi_i (rotation-invariant); SE(h2_i) by the delta method on V
#
# Fills the same SE/CI schema as the bootstrap (rot_loadings/Phi/Structure plus the unrotated
# loadings and uniquenesses); there are no replicate arrays. Falls back to NA rotated SEs (with a
# classed warning) at a Heywood case, a singular information matrix, or a rotation the warm start
# cannot reproduce (e.g. a non-converged transformation). References: Jennrich (1973); Zhang &
# Preacher (2015).
.se_information_rotated <- function(fit_out, rot_info, N, ci, se0) {

  A <- unclass(fit_out$unrot_loadings)
  p <- nrow(A)
  k <- ncol(A)
  pk <- p * k
  psi <- 1 - rowSums(A^2)
  z <- stats::qnorm(1 - (1 - ci) / 2)

  # Unrotated pieces (always reported, mirroring the bootstrap's unrot_loadings) and the loading
  # covariance block propagated through the rotation. Each analytic path supplies its own covariance
  # in `se0` (uniform $vcov/$loadings_se/$uniquenesses_se schema), so the propagation below is
  # identical for all of them and is purely a function of the loading covariance V.
  V <- se0$vcov[seq_len(pk), seq_len(pk), drop = FALSE]
  SE_unrot <- se0$loadings_se
  SE_psi <- se0$uniquenesses_se
  dimnames(SE_unrot) <- dimnames(A)
  names(SE_psi) <- rownames(A)
  # The caller NA's the unrotated SEs whenever the parameter covariance is non-finite OR carries a
  # negative variance on its diagonal (a near-degenerate orientation `solve()` still inverts). Reuse
  # that as the single reliability signal for the rotated quantities too: they are propagated from
  # the same covariance, so if it was too ill-conditioned for the unrotated SEs it must not yield
  # finite rotated SEs either. NA the covariance here, before anything reads it: on the non-PSD path
  # the caller ships a FINITE covariance next to NA marginal SEs, and every quantity derived from it
  # below (the communality delta method, the persisted slot) must fail closed with them rather than
  # report a standard error the caller already judged untrustworthy.
  info_reliable <- !anyNA(SE_unrot)
  if (!info_reliable) V[] <- NA_real_

  # A degenerate rotational gauge is a strictly weaker signal: it diverges only along the gauge
  # directions, which the rotation Jacobian annihilates, so the rotated loadings, Phi, the structure
  # coefficients and the communalities all stay valid and V must NOT be NA'd here. Withhold the
  # unrotated loadings alone -- and only after `info_reliable` has been read, so a degenerate gauge
  # cannot masquerade as an unusable covariance and take the rotated quantities down with it.
  # Every analytic path supplies the flag (they share the gauge constraint it describes); absent
  # rather than FALSE would mean no diagnostic was available, which is then not treated as one.
  gauge_degenerate <- isFALSE(se0$gauge_reliable)
  if (gauge_degenerate) SE_unrot[] <- NA_real_

  spec <- .rotation_se_method(rot_info$rotation, p, k, rot_info$crit_args)
  rotmat <- rot_info$rotmat
  L_rot <- unclass(rot_info$rot_loadings)
  Phi_pt <- rot_info$Phi
  # Orthogonal rotations and the single-factor fall-back carry no factor correlations, so they
  # report rotated loadings and communalities only.
  oblique <- !is.null(Phi_pt)

  # The communalities are the exact complement of the uniquenesses (h2_i = rowSums(A^2)_i =
  # 1 - psi_i = (L Phi L')_ii), so h2_i and psi_i are one estimand up to sign and must share a
  # standard error. Both are the ordinary delta method on the loading covariance V through the
  # gradient 2 A[i, ] of h2_i (`.communality_se()`); because every path's `se0$uniquenesses_se`
  # comes from that same call on that same V, the two agree exactly.
  h2 <- rowSums(A^2)
  names(h2) <- rownames(A)
  SE_h2 <- .communality_se(A, V)
  names(SE_h2) <- rownames(A)

  na_mat <- function() {
    m <- matrix(NA_real_, p, k)
    dimnames(m) <- dimnames(L_rot)
    m
  }
  SE_rot <- na_mat()
  SE_Phi <- if (oblique) {
    matrix(NA_real_, k, k, dimnames = dimnames(Phi_pt))
  } else {
    NULL
  }
  SE_S <- if (oblique) na_mat() else NULL
  S_pt <- if (oblique) `dimnames<-`(L_rot %*% Phi_pt, dimnames(L_rot)) else NULL

  can_rotate <- !is.null(spec) && is.matrix(rotmat) && !anyNA(rotmat) &&
    k >= 2L && !is.null(L_rot) && info_reliable

  if (can_rotate) {
    normalize <- isTRUE(rot_info$normalize)
    # The bifactor criterion exempts a fixed column as the general factor, but `.reflect_and_order`
    # may have moved the general factor out of the first column. It is the one column that loads on
    # every variable -- so the column whose smallest absolute loading is largest, the others being
    # group factors with near-zero out-group loadings. Identify it so the re-rotation optimises the
    # same criterion the point estimate did; the other criteria ignore general_col.
    general_col <- if (identical(spec$method, "bifactor")) {
      which.max(apply(abs(L_rot), 2, min)) - 1L
    } else {
      0L
    }
    # Forward difference of the warm-started re-rotation (the stencil lavaan's delta method uses);
    # the whole p*k finite-difference loop runs in compiled code (.rotation_se_jacobian()) to avoid
    # p*k round trips to R. This is a one-shot cost per fit: a few hundredths of a second for a
    # typical problem (p = 18, k = 3) and about 1 s for a large one (p = 40, k = 8) -- far below a
    # bootstrap, and an order of magnitude faster than the equivalent pure-R numeric differentiation.
    eps <- 1e-4
    jac <- .rotation_se_jacobian(A, rotmat, spec$method, spec$param, normalize,
                                 spec$oblique, eps, general_col)
    # The re-rotation at the unperturbed A must reproduce the reported rotated loadings; a gross
    # mismatch flags a criterion/parameter mismatch or a transformation outside the criterion's
    # basin, for which the Jacobian is not trustworthy. The tolerance absorbs the small drift
    # between the reported optimum and a re-solve to a tighter tolerance -- and, for varimax, between
    # the SVD/SPSS varimax algorithm and the Crawford-Ferguson re-rotation at a flat optimum -- while
    # still catching a genuine basin/criterion mismatch (which moves loadings an order of magnitude
    # more).
    if (isTRUE(jac$valid) && max(abs(jac$base_loadings - L_rot)) < 5e-2) {
      J_L <- jac$J_L
      SE_rot <- matrix(sqrt(pmax(rowSums((J_L %*% V) * J_L), 0)), p, k)
      dimnames(SE_rot) <- dimnames(L_rot)
      if (spec$oblique) {
        J_Phi <- jac$J_Phi
        SE_Phi <- matrix(sqrt(pmax(rowSums((J_Phi %*% V) * J_Phi), 0)), k, k)
        diag(SE_Phi) <- 0       # the unit diagonal of Phi is fixed, so it has no variance
        SE_Phi <- (SE_Phi + t(SE_Phi)) / 2
        dimnames(SE_Phi) <- dimnames(Phi_pt)
        # Structure S = L Phi: dvec(S) = (Phi' (x) I_p) dvec(L) + (I_k (x) L) dvec(Phi); the
        # rotation matrices are the re-rotated point estimates, consistent with the Jacobians.
        L0 <- jac$base_loadings
        Phi0 <- jac$base_Phi
        J_S <- kronecker(t(Phi0), diag(p)) %*% J_L + kronecker(diag(k), L0) %*% J_Phi
        SE_S <- matrix(sqrt(pmax(rowSums((J_S %*% V) * J_S), 0)), p, k)
        dimnames(SE_S) <- dimnames(L_rot)
      }
    }
  } else if (info_reliable && k < 2L) {
    # No rotation was actually performed (single factor): the rotated loadings equal the unrotated
    # ones, so their SEs are the unrotated loading SEs; there is no Phi or structure matrix.
    SE_rot <- SE_unrot
    dimnames(SE_rot) <- dimnames(L_rot)
  }

  SE <- list(unrot_loadings = SE_unrot, uniquenesses = SE_psi,
             rot_loadings = SE_rot, communalities = SE_h2)
  CI <- list(unrot_loadings = .wald_ci(A, SE_unrot, z),
             uniquenesses = .wald_ci(psi, SE_psi, z),
             rot_loadings = .wald_ci(L_rot, SE_rot, z),
             communalities = .wald_ci(h2, SE_h2, z))
  if (oblique) {
    SE$Phi <- SE_Phi
    SE$Structure <- SE_S
    CI$Phi <- .wald_ci(Phi_pt, SE_Phi, z)
    CI$Structure <- .wald_ci(S_pt, SE_S, z)
  }

  if (anyNA(SE_unrot) || anyNA(SE_psi) || anyNA(SE_rot) || anyNA(SE_h2) ||
      (oblique && (anyNA(SE_Phi) || anyNA(SE_S)))) {
    cli::cli_warn(
      c("Analytic standard errors could not be computed for all parameters.",
        "i" = if (gauge_degenerate && info_reliable) {
          "The factor solution's rotational orientation is only weakly determined (two canonical variances nearly coincide), so the unrotated loadings have no well-defined standard error. The rotated loadings and communalities are gauge-invariant and unaffected."
        } else if (!info_reliable && .se_boundary_hint_applies(A)) {
          # Named rather than listed among the possible causes: every analytic path
          # (`.se_sandwich_dispatch()` and `.se_fiml()`) withholds through `.withhold_at_boundary()`,
          # so where the hint applies it can be reported as a fact about the solution. Two things
          # make that attribution safe. `!info_reliable`: a fit that reached the boundary but still
          # produced finite unrotated SEs would contradict the hint's claim that nothing was
          # reported. And `.se_boundary_hint_applies()` rather than the gate itself, which also
          # fires on a non-finite solution that is at no boundary at all. Both fall to the last
          # clause instead.
          .se_boundary_hint
        } else {
          "This occurs when the parameter covariance is singular, or when the rotation could not be reproduced for the standard-error Jacobian."
        }),
      class = "efa_se_unreliable"
    )
  }

  # Surface the pk x pk unrotated loading vcov (whichever analytic covariance `se0` carried) so it
  # can be persisted on the EFA object alongside the rotated SEs. It was already NA-filled above
  # when the unrotated SEs were unreliable, so the slot is consistent with the marginal SEs and
  # downstream consumers can fail closed on `anyNA()`. A degenerate gauge deliberately does NOT
  # NA-fill it: the covariance is finite, PSD and correct, and only its gauge-dependent marginals
  # are unusable (see the matching note in `.se_sandwich_unrotated()`).
  lab <- .vec_loading_labels(A)
  if (!is.null(lab)) dimnames(V) <- list(lab, lab)

  list(SE = SE, CI = CI, replicates = NULL, vcov_unrot_loadings = V)
}


# Robust (sandwich) standard errors and a scaled chi-square for the ordinal/polychoric path.
# The factor model is fitted by minimising a weighted off-diagonal discrepancy
# F(theta) = (s - sigma(theta))' V (s - sigma(theta)) over the free loadings theta = vec(Lambda),
# where s = vech of the off-diagonal sample (polychoric/tetrachoric) correlations,
# sigma_ij(theta) = (Lambda Lambda')_ij, and V is the estimator's weight matrix (DWLS: diagonal
# inverse asymptotic variances; ULS: identity; ML: the normal-theory GLS weight). When V is not
# the inverse of the true asymptotic covariance of s (which it never is for ordinal data), the
# naive inverse-information SEs are biased, and the Godambe sandwich
#   Var(theta) = (1/(N-1)) A^- (Delta' V Gamma V Delta) A^-,   A = Delta' V Delta,
# with Gamma the (threshold-adjusted) asymptotic covariance of s, gives consistent SEs (Browne,
# 1984; Muthen, 1984; Satorra & Bentler, 1994). This mirrors lavaan's se = "robust.sem". The
# rotational over-parameterisation of A is handled by bordering it with the identification-
# constraint Jacobian, exactly as the expected-information path does. The bordered loading
# covariance is propagated through a rotation by the same Jacobian machinery as the information
# SEs (.se_information_rotated), so promax/simplimax stay bootstrap-only.
.se_sandwich_dispatch <- function(fit_out, rot_info, N, ci, Gamma, method,
                                  optimal_weight = FALSE, scaled = TRUE) {

  core <- .se_sandwich_core(fit_out, N, Gamma, method, optimal_weight, scaled)

  # The gate sits here rather than inside the core because the core's own `psi <= 0` test asks a
  # different question -- whether Psi^-1 exists for the gauge constraint -- and must not be moved to
  # the floor (see the note there); and because the core's scaled chi-square is deliberately kept
  # (see `.withhold_at_boundary()`, which owns the policy for every analytic path).
  core <- .withhold_at_boundary(core, fit_out$unrot_loadings)

  res <- if (is.null(rot_info)) {
    .se_sandwich_unrotated(fit_out, core, ci)
  } else {
    # Reuse the information-SE rotation propagation verbatim by supplying the robust loading
    # covariance in place of the expected-information one (identical $vcov schema). Because the
    # rotation Jacobian annihilates the rotational gauge directions, the rotated SEs are
    # identification-invariant and do not depend on which bordering constraint produced V_AA.
    se0 <- list(vcov = core$V_AA,
                loadings_se = core$loadings_se,
                uniquenesses_se = core$uniquenesses_se,
                gauge_reliable = core$gauge_reliable)
    .se_information_rotated(fit_out, rot_info, N, ci, se0 = se0)
  }

  res$scaled_test <- core$scaled_test
  res
}

# Identification-constraint Jacobian for the rotational gauge: the gradient, with respect to the
# column-major vec(Lambda), of the off-diagonal entries the estimator's canonical orientation sets
# to zero. nc = k(k-1)/2 rows, p*k columns; for k = 1 there is no gauge freedom and the matrix has
# zero rows. The constraint must match the orientation the loadings are reported in so the unrotated
# loading SEs are scaled in that gauge (the rotated SEs are gauge-invariant and the same for any
# transversal choice). With `psi = NULL` (eigen-based ULS/DWLS) it fixes off-diag(Lambda'Lambda);
# with `psi` supplied (ML, where psi = 1 - rowSums(Lambda^2)) it fixes off-diag(Lambda' Psi^-1
# Lambda) (Lawley & Maxwell), whose gradient carries a chain-rule term through psi(Lambda).
.se_sandwich_constraint <- function(L, psi = NULL) {
  p <- nrow(L)
  k <- ncol(L)
  nc <- k * (k - 1L) / 2L
  Cmat <- matrix(0, nc, p * k)
  if (k > 1L) {
    Astar <- if (is.null(psi)) NULL else L / psi   # Psi^-1 Lambda for the ML gauge
    uv <- 0L
    for (v in 2:k) {
      for (u in seq_len(v - 1L)) {
        uv <- uv + 1L
        grad <- if (is.null(psi)) {
          .gauge_grad(L, u, v)
        } else {
          # d off-diag(Lambda' Psi^-1 Lambda)_uv / d vec(Lambda): the direct part plus the term
          # from psi_a = 1 - rowSums(Lambda^2)_a depending on Lambda (d psi_a / d lambda_ac = -2
          # lambda_ac), which contributes 2 (Astar[, u] Astar[, v]) Lambda across all columns.
          .gauge_grad(Astar, u, v) + 2 * (Astar[, u] * Astar[, v]) * L
        }
        Cmat[uv, ] <- as.vector(grad)
      }
    }
  }
  Cmat
}

# Smallest gauge-transversal conditioning a solution can have before its rotational orientation is
# treated as degenerate. The quantity `.gauge_transversal()` returns is a cosine in [0, 1] -- the
# smallest principal angle between the gauge-fixing constraint's row space and the gauge directions
# it has to pin down -- so this is a dimensionless angle, not a parameter scale. Blow-ups in the
# reported standard errors set in below about 0.05 and well-determined solutions sit above 0.2, so
# the floor flags only a clearly degenerate orientation.
.gauge_transversal_floor <- 0.05

# Conditioning of the gauge-fixing constraint as a transversal to the rotational gauge orbit.
#
# The bordered inverse of `A` under the constraint `Cmat` is the constrained (reflexive generalised)
# inverse (Rao & Mitra, 1971, ch. 3; Silvey, 1959)
#   A^- = (I - Z (C Z)^-1 C) A^+ (I - C' (C Z)^-1' Z'),
# where Z spans null(A). The whole conditioning of the augmented system therefore sits in the
# k(k-1)/2 x k(k-1)/2 matrix C Z, and the reported unrotated loading covariance is amplified by
# ||(C Z)^-1||. The null space is available in closed form: the gauge directions are dLambda =
# Lambda S for antisymmetric S, since d(Lambda Lambda') = Lambda (S + S') Lambda' = 0, so no
# eigendecomposition of A is needed.
#
# C Z degenerates when the canonical orientation is not pinned down. Dropping the chain-rule term,
# C Z is diagonal with entries (d_u - d_v) for d = diag(Lambda' Psi^-1 Lambda): when two of the
# canonical variances that identify the ML solution collide (Lawley & Maxwell, 1971, sec. 2.3), the
# orientation within that two-plane is arbitrary and the unrotated loadings have no well-defined
# limiting distribution -- the classic rotational indeterminacy of the unrotated factor solution.
# The divergence is genuine, not a numerical artefact -- but it is confined to the gauge, so it must
# be reported rather than reported around: the rotated loadings, Phi and the structure coefficients
# are gauge-invariant (the rotation Jacobian annihilates these directions) and the communalities
# rowSums(Lambda^2) are likewise invariant, so all of them stay valid and are left untouched by the
# caller.
#
# Returns the smallest singular value of the row-normalised constraint restricted to the gauge
# directions -- dimensionless and in [0, 1] -- or Inf when there is no gauge freedom to fix (k = 1).
.gauge_transversal <- function(L, Cmat) {

  k <- ncol(L)
  if (nrow(Cmat) == 0L || k < 2L) return(Inf)
  if (!all(is.finite(L)) || !all(is.finite(Cmat))) return(0)

  # Z = [vec(L S_uv)] over the antisymmetric basis, in the same (u, v) order as the constraint rows.
  # L %*% S_uv is just two sign-flipped column copies (column u becomes -L[, v], column v becomes
  # L[, u]), so fill it directly rather than forming S and multiplying.
  Z <- matrix(0, length(L), nrow(Cmat))
  col <- 0L
  for (v in 2:k) {
    for (u in seq_len(v - 1L)) {
      col <- col + 1L
      Zc <- matrix(0, nrow(L), k)
      Zc[, u] <- -L[, v]
      Zc[, v] <- L[, u]
      Z[, col] <- as.vector(Zc)
    }
  }
  # A rank-deficient Z means the loadings themselves are collinear across factors, so the gauge is
  # not even a well-defined orbit; treat it as maximally degenerate.
  qrZ <- qr(Z)
  if (qrZ$rank < ncol(Z)) return(0)
  Z <- qr.Q(qrZ)

  # Normalise the constraint rows so the singular values measure orientation only: the constraint is
  # a set of equations fixed at zero, so its overall row scaling is arbitrary and must not enter.
  rn <- sqrt(rowSums(Cmat^2))
  if (any(rn <= 0)) return(0)

  min(svd((Cmat / rn) %*% Z)$d)
}

# Normal-theory asymptotic covariance of the off-diagonal Pearson correlations, on the unit scale
# (N * Cov(r-hat)) and in the `pairs` (utils::combn(p, 2)) order the sandwich machinery uses
# throughout. `Sigma` is the correlation matrix it is evaluated at -- the model-implied one for the
# expected-information SEs.
#
# For pairs (i, j) and (k, l) (Pearson & Filon, 1898; Olkin & Siotani, 1976; Steiger, 1980):
#   N Cov(r_ij, r_kl) = 1/2 rho_ij rho_kl (rho_ik^2 + rho_il^2 + rho_jk^2 + rho_jl^2)
#                       + rho_ik rho_jl + rho_il rho_jk
#                       - rho_ij (rho_ik rho_il + rho_jk rho_jl)
#                       - rho_kl (rho_ik rho_jk + rho_il rho_jl),
# which reduces to the familiar (1 - rho^2)^2 on the diagonal. Assembled by whole-matrix indexing
# (`Sigma[ii, ii]` and friends are already n x n in pair order) rather than looping over the n^2
# pair combinations; the result is symmetric analytically and is symmetrised against rounding.
.normal_theory_gamma <- function(Sigma, pairs) {

  ii <- pairs[1, ]
  jj <- pairs[2, ]
  n <- ncol(pairs)

  # Drop the variable dimnames while sub-setting: `Sigma[ii, ii]` would otherwise label the pair
  # axis with repeated variable names ("V1", "V1", ...), which mislabels a pair-indexed matrix and
  # would propagate into the covariance built from it.
  Sigma <- unname(Sigma)
  r_ij <- Sigma[cbind(ii, jj)]
  Rik <- Sigma[ii, ii, drop = FALSE]
  Ril <- Sigma[ii, jj, drop = FALSE]
  Rjk <- Sigma[jj, ii, drop = FALSE]
  Rjl <- Sigma[jj, jj, drop = FALSE]

  # rho_ij indexes the row pair, rho_kl the column pair: `r_ij * M` scales rows (column-major
  # recycling) and `rep(r_ij, each = n) * M` scales columns.
  G <- 0.5 * outer(r_ij, r_ij) * (Rik^2 + Ril^2 + Rjk^2 + Rjl^2) +
    Rik * Rjl + Ril * Rjk -
    r_ij * (Rik * Ril + Rjk * Rjl) -
    rep(r_ij, each = n) * (Rik * Rjk + Ril * Rjl)

  (G + t(G)) / 2
}

# Shared geometry of every off-diagonal sandwich: the estimator's Stage-2 weight V (held as a
# diagonal `vdiag` or a full `Vmat`, the other NULL), the weighted model Jacobian VD, the bordered
# (constrained generalised) inverse of the bread A = Delta' V Delta, and the conditioning of the
# gauge constraint that bordering used. The polychoric/ADF sandwich, the expected-information path
# (`.se_sandwich_core()`) and the corrected two-stage FIML sandwich (`.se_fiml_core()`,
# `.fiml_scaled_test()`) differ only in the meat they sandwich between these pieces, so they share
# them rather than each carrying a copy -- which is what let the gauge choice and the gauge
# diagnostic drift apart between the paths before.
#
# `Gamma` is the asymptotic covariance on the UNIT scale (N * Var(rho-hat)) and is read only by the
# two weights that need it: DWLS's diagonal inverse variances, and the efficient weight
# `optimal_weight = TRUE` (V = Gamma^-1, at which the sandwich collapses onto its bread and returns
# the Fisher information of the correlation structure). The ML/ULS paths -- including both FIML
# ones -- pass `Gamma = NULL`. Returns NULL when the model-implied correlation, the efficient
# weight, or the bordered system is singular, or when a weight that needs `Gamma` was asked for
# without one; every caller NA-fills on that.
#
# `gauge = FALSE` skips the transversal diagnostic for a caller that does not read it (the
# corrected two-stage chi-square, which runs once per bootstrap replicate and is gauge-invariant),
# so that caller does not pay for a QR and an SVD it discards.
.offdiag_sandwich_pieces <- function(L, method, Gamma = NULL, optimal_weight = FALSE,
                                     gauge = TRUE) {

  # The DWLS and efficient weights are built FROM `Gamma`, so neither is defined without one.
  # `.se_sandwich_core()` screens that out before calling, but the FIML callers pass `Gamma = NULL`
  # with a `method` their own callers guarantee is ML or ULS: fail closed here rather than let an
  # unguarded combination surface as a non-conformable-arguments error further down.
  if ((optimal_weight || identical(method, "DWLS")) && is.null(Gamma)) return(NULL)

  L <- unclass(L)
  p <- nrow(L)
  k <- ncol(L)
  pk <- p * k

  pairs <- utils::combn(p, 2L)
  pi <- pairs[1, ]
  pj <- pairs[2, ]
  n <- ncol(pairs)

  # Model Jacobian Delta = d sigma_offdiag / d vec(Lambda) (n x pk). For pair (i, j),
  # d sigma_ij / d Lambda[a, f] = (a == i) Lambda[j, f] + (a == j) Lambda[i, f]; rows in
  # utils::combn(p, 2) order, columns in column-major vec(Lambda) order.
  Delta <- matrix(0, n, pk)
  for (f in seq_len(k)) {
    Delta[cbind(seq_len(n), (f - 1L) * p + pi)] <- L[pj, f]
    Delta[cbind(seq_len(n), (f - 1L) * p + pj)] <- L[pi, f]
  }

  # Estimator weight V (unit scale). DWLS: diagonal inverse variances; ULS: identity; ML: the
  # normal-theory GLS weight 1/2 (Sigma^-1 (x) Sigma^-1) restricted to the off-diagonal pairs, at
  # the model-implied correlation matrix Sigma = Lambda Lambda' (unit diagonal).
  if (optimal_weight) {
    # Efficient weight V = Gamma^-1. The meat Delta' V Gamma V Delta then equals the bread
    # Delta' V Delta = Delta' Gamma^-1 Delta, so V_AA reduces to the (bordered) inverse of the
    # correlation-structure Fisher information.
    Vmat <- tryCatch(solve(Gamma), error = function(e) NULL)
    if (is.null(Vmat)) return(NULL)
    Vmat <- (Vmat + t(Vmat)) / 2
    vdiag <- NULL
  } else if (method == "DWLS") {
    vdiag <- 1 / diag(Gamma)
    Vmat <- NULL
  } else if (method == "ULS") {
    vdiag <- rep(1, n)
    Vmat <- NULL
  } else {
    Sigma <- tcrossprod(L)
    diag(Sigma) <- 1
    P <- tryCatch(solve(Sigma), error = function(e) NULL)
    if (is.null(P)) return(NULL)
    vdiag <- NULL
    Vmat <- matrix(0, n, n)
    for (s in seq_len(n)) {
      a <- pi[s]
      b <- pj[s]
      Vmat[, s] <- 0.5 * (P[pi, a] * P[pj, b] + P[pi, b] * P[pj, a])
    }
  }
  is_diag <- is.null(Vmat)

  # VD = V Delta; bread A = Delta' V Delta (singular by the k(k-1)/2 rotational gauge freedoms).
  VD <- if (is_diag) vdiag * Delta else Vmat %*% Delta
  A <- crossprod(Delta, VD)

  # Border A with the gauge-fixing constraint so the augmented system is invertible; the leading
  # pk-block of its inverse is the reflexive generalised inverse of A. The constraint fixes the
  # rotational orientation, so it must match the one the solution's loadings are reported in: the
  # unrotated loading SEs are scaled in that gauge (the rotated SEs and scaled chi-square are
  # gauge-invariant, so the choice does not affect them). Rather than assume the orientation from the
  # estimator label, detect it from the loadings: an eigen-based ULS/DWLS solution is
  # Lambda'Lambda-diagonal, an ML solution is Lambda'Psi^-1 Lambda-diagonal (Lawley & Maxwell). The
  # orientation the solution is in leaves a (scale-free) relative off-diagonal at the optimiser floor
  # while the other is O(0.01-1), so the smaller ratio identifies the gauge -- and detecting keeps it
  # tied to the actual solution, so a future estimator with either identification is handled without
  # special-casing. Two cases the loadings cannot resolve fall back to a fixed choice: the
  # Lambda'Psi^-1 Lambda gauge needs Psi^-1, so it is unavailable where that is undefined (psi <= 0)
  # or at a non-finite solution -- routed to Lambda'Lambda; and homogeneous uniquenesses (Psi
  # proportional to I) make BOTH orientations diagonal, so the gauge is undetermined by the loadings
  # (yet the two still give different SEs through the chain-rule term), broken by the estimator's
  # identification. A single factor has no rotational freedom (the constraint is empty either way).
  #
  # The test here is psi <= 0 and deliberately NOT the `.uniqueness_floor` boundary that the callers
  # treat as a Heywood case. The two ask different questions: that one asks whether the solution sits
  # on the parameter-space boundary (where a Wald interval is invalid), this one only whether Psi^-1
  # exists. At a uniqueness pinned at the floor Psi^-1 is large but perfectly well defined, and the
  # solution is still oriented Lambda'Psi^-1 Lambda-diagonal -- so forcing the Lambda'Lambda
  # constraint there would mismatch the orientation the loadings are reported in and hand
  # `.gauge_transversal()` a near-tangent transversal (0.002 against 0.16 for the matching constraint
  # on a boundary ML fit), withholding standard errors that are fine.
  psi <- 1 - rowSums(L^2)
  use_ltpil_gauge <- if (k < 2L || anyNA(psi) || any(psi <= 0)) {
    FALSE
  } else {
    rel_off <- function(M) max(abs(M[upper.tri(M)])) / max(abs(diag(M)), .Machine$double.eps)
    r_ltl <- rel_off(crossprod(L))
    r_ltpil <- rel_off(crossprod(L, L / psi))
    # Both orientations near-diagonal (Psi proportional to I): the loadings cannot distinguish the
    # gauge, so use the estimator's known identification; otherwise the smaller ratio wins.
    if (max(r_ltl, r_ltpil) < 1e-4) method == "ML" else r_ltpil < r_ltl
  }
  Cmat <- if (use_ltpil_gauge) {
    .se_sandwich_constraint(L, psi = psi)
  } else {
    .se_sandwich_constraint(L)
  }
  Abread <- .bordered_inverse_block(A, Cmat, pk)
  if (is.null(Abread)) return(NULL)

  # Conditioning of that constraint as a transversal to the gauge orbit. A near-degenerate
  # transversal amplifies the unrotated loading covariance by ||(C Z)^-1|| without making it
  # non-finite or non-PSD, so it passes the callers' `.is_psd()` gate and would otherwise ship a
  # silently meaningless standard error. Reported separately from a caller's `reliable` flag because
  # the divergence lives purely in the gauge: everything gauge-invariant derived from the covariance
  # stays valid. Computed after the bordered inverse so an exactly singular augmented system -- which
  # returns NULL above and discards this flag -- does not pay for the decomposition, and skipped
  # entirely for a caller that does not read it.
  gauge_reliable <- if (gauge) {
    .gauge_transversal(L, Cmat) >= .gauge_transversal_floor
  } else {
    NULL
  }

  list(vdiag = vdiag, Vmat = Vmat, VD = VD, Abread = Abread, pairs = pairs,
       gauge_reliable = gauge_reliable)
}

# Build the robust loading covariance V_AA (p*k x p*k), the unrotated loading/uniqueness SEs, and
# the scaled chi-square from the fitted loadings and the polychoric ACOV `Gamma`. `Gamma` enters on
# the variance scale Var(rho-hat) and is converted here to the unit asymptotic-variance scale
# (N * Var) used by the WLS/sandwich formulas (= lavaan's NACOV). Returns NA SEs (reliable = FALSE)
# at a singular bordered information matrix or an unusable covariance.
#
# The linear algebra is O(n^2 q) for the meat and O(n^3) for the chi-square trace (n = p(p-1)/2
# pairs, q = p*k), and is kept in R: it is a one-shot cost of a few hundredths of a second for a
# typical problem and about 0.3 s (p = 30) to 1.5 s (p = 40, n = 780 pairs) for a large one --
# negligible next to the polychoric estimate it builds on, and far below a bootstrap.
#
# `optimal_weight = TRUE` replaces the estimator weight with V = Gamma^-1, at which the sandwich
# collapses onto its bread and returns the Fisher information of the correlation structure (the
# se = "information" path; see `.se_information()`). `scaled = FALSE` skips the scaled chi-square,
# which only the robust paths report.
.se_sandwich_core <- function(fit_out, N, Gamma, method,
                              optimal_weight = FALSE, scaled = TRUE) {

  L <- unclass(fit_out$unrot_loadings)
  p <- nrow(L)
  k <- ncol(L)
  pk <- p * k

  # `gauge_reliable = TRUE` on the NA path: everything is already withheld, and the gauge is not
  # what withheld it, so the caller's gauge-specific message must not be attached to this branch.
  na_core <- list(
    V_AA = matrix(NA_real_, pk, pk),
    loadings_se = matrix(NA_real_, p, k),
    uniquenesses_se = rep(NA_real_, p),
    scaled_test = NULL,
    reliable = FALSE,
    gauge_reliable = TRUE
  )

  if (is.null(Gamma) || anyNA(Gamma) || nrow(Gamma) != p * (p - 1L) / 2L) {
    return(na_core)
  }

  # Variance scale (Var(rho-hat)) -> unit asymptotic-variance scale (N * Var); the 1/(N-1) on the
  # final covariance and the N on the chi-square then follow lavaan's robust.sem conventions.
  Gamma <- N * Gamma

  # Model Jacobian, estimator weight, bordered bread, and gauge conditioning; shared verbatim with
  # the corrected two-stage FIML sandwich, which supplies a different meat below.
  pieces <- .offdiag_sandwich_pieces(L, method, Gamma = Gamma,
                                     optimal_weight = optimal_weight)
  if (is.null(pieces)) return(na_core)
  VD <- pieces$VD
  Abread <- pieces$Abread

  # Meat Delta' V Gamma V Delta = (V Delta)' Gamma (V Delta); robust covariance A^- meat A^- /(N-1).
  Gamma_theta <- crossprod(VD, Gamma %*% VD)
  V_AA <- (Abread %*% Gamma_theta %*% Abread) / (N - 1)

  # The covariance must be positive semidefinite: a non-finite entry or a negative eigenvalue (even
  # with a still-positive diagonal) would otherwise corrupt the marginal, uniqueness, and rotated
  # SEs, which read the full V_AA.
  reliable <- .is_psd(V_AA)
  loadings_se <- if (reliable) sqrt(diag(V_AA)) else rep(NA_real_, pk)

  # Uniqueness SE = communality SE (psi_i = 1 - rowSums(Lambda^2)_i), via the shared gradient.
  uniq_se <- if (!reliable) rep(NA_real_, p) else .communality_se(L, V_AA)

  scaled_test <- if (!reliable || !scaled) NULL else {
    .scaled_chisq(fit_out, Gamma, pieces$pairs, VD, pieces$vdiag, pieces$Vmat, Abread, N)
  }

  list(
    V_AA = V_AA,
    loadings_se = matrix(loadings_se, p, k),
    uniquenesses_se = uniq_se,
    scaled_test = scaled_test,
    reliable = reliable,
    gauge_reliable = pieces$gauge_reliable
  )
}

# Scaled chi-square test statistics (Satorra & Bentler, 1994; Asparouhov & Muthen, 2010) for the
# weighted off-diagonal fit. T = N * (s - sigma)' V (s - sigma) is the (unscaled) fit statistic;
# U = V - V Delta A^- Delta' V is the residual projector; c1 = tr(U Gamma), c2 = tr((U Gamma)^2)
# its trace coefficients (Gamma on the unit scale). The projector spans only the off-diagonal
# correlation residuals, so this is the two-stage correlation-structure correction (Browne, 1984)
# and is not identical to the full WLSMV statistic of lavaan/Mplus, which also projects the
# thresholds. Returns the scaled-shifted, mean-adjusted, and mean-and-variance-adjusted statistics
# plus the scaled baseline statistic for the robust CFI/TLI/RMSEA. NULL when the model is
# just-identified (df <= 0) or the traces degenerate.
.scaled_chisq <- function(fit_out, Gamma, pairs, VD, vdiag, Vmat, Abread, N) {

  L <- unclass(fit_out$unrot_loadings)
  df <- fit_out$fit_indices$df
  if (is.null(df) || is.na(df) || df <= 0) return(NULL)

  n <- ncol(pairs)
  is_diag <- is.null(Vmat)
  apply_V <- function(x) if (is_diag) vdiag * x else as.vector(Vmat %*% x)

  # The sample correlations s and residuals e are read from the (possibly smoothed) analysis matrix,
  # while Gamma is the asymptotic covariance of the observed un-projected polychoric correlations --
  # the same convention the DWLS point estimate and lavaan use (the weights/ACOV describe the
  # observed correlations even when the matrix is projected to positive definiteness).
  R <- unclass(fit_out$orig_R)
  s <- R[t(pairs)]                       # off-diagonal sample correlations
  e <- (R - tcrossprod(L))[t(pairs)]     # residuals r_ij - (Lambda Lambda')_ij

  Tstat <- N * sum(e * apply_V(e))

  # Residual projector U = V - V Delta A^- Delta' V. The dense V Delta A^- Delta' V term makes U
  # dense anyway, so add the (diagonal or full) weight onto it in place instead of materialising a
  # dense V.
  U <- -(VD %*% Abread %*% t(VD))
  if (is_diag) diag(U) <- diag(U) + vdiag else U <- U + Vmat
  c1 <- sum(U * Gamma)
  UG <- U %*% Gamma
  c2 <- sum(UG * t(UG))
  if (!is.finite(c1) || !is.finite(c2) || c1 <= 0 || c2 <= 0) return(NULL)

  # Independence baseline (all off-diagonal correlations fixed to 0): residual is s, projector the
  # baseline weight V0. V0 is diagonal for every estimator -- it equals the (model-independent)
  # weight for DWLS/ULS, and 1/2 I for ML (the normal-theory weight at Sigma = I) -- so the baseline
  # traces use only its diagonal.
  v0 <- if (is_diag) vdiag else rep(0.5, n)
  Tbase <- N * sum(s * (v0 * s))
  df_base <- n
  c1_base <- sum(v0 * diag(Gamma))       # tr(diag(v0) Gamma)
  V0G <- v0 * Gamma                      # diag(v0) %*% Gamma (row scaling)
  c2_base <- sum(V0G * t(V0G))           # tr((diag(v0) Gamma)^2)

  scaled <- .scaled_variants(Tstat, df, c1, c2)
  chi_null <- if (is.finite(c1_base) && is.finite(c2_base) && c2_base > 0) {
    .scaled_variants(Tbase, df_base, c1_base, c2_base)$T_ss
  } else {
    NA_real_
  }

  list(
    m = nrow(L),
    df = df,
    chi = scaled$T_ss,                   # scaled-shifted (WLSMV default) -> the reported chi-square
    chi_scaling = scaled$a,
    chi_shift = scaled$b,
    chi_unscaled = Tstat,
    chi_mean_adjusted = scaled$T_mean,
    chi_mean_var = scaled$T_mv,
    df_mean_var = scaled$df_mv,
    chi_null = chi_null,
    df_null = df_base
  )
}

# The three scalings of a fit statistic T (Satorra & Bentler, 1994; Asparouhov & Muthen, 2010)
# from its trace coefficients c1 = tr(U Gamma), c2 = tr((U Gamma)^2) and degrees of freedom df:
# mean-adjusted (T * df/c1, reference df); scaled-shifted a*T + b with a = sqrt(df/c2),
# b = df - a*c1 (reference df, the WLSMV default); mean-and-variance-adjusted T * c1/c2 with the
# Satterthwaite df* = c1^2/c2.
.scaled_variants <- function(Tstat, df, c1, c2) {
  a <- sqrt(df / c2)
  list(
    T_mean = Tstat * df / c1,
    a = a,
    b = df - a * c1,
    T_ss = a * Tstat + (df - a * c1),
    df_mv = c1^2 / c2,
    T_mv = Tstat * c1 / c2
  )
}

# Unrotated analytic SE/CI wrapper: fills the uniform list(SE, CI, replicates) schema from the
# core's loading/uniqueness SEs. Shared by every analytic path -- expected information, robust
# sandwich, and the corrected two-stage FIML sandwich -- which differ only in the covariance the
# core built (`.se_information_rotated()` plays the same role under a rotation).
.se_sandwich_unrotated <- function(fit_out, core, ci) {

  L <- unclass(fit_out$unrot_loadings)
  SE_L <- core$loadings_se
  SE_psi <- core$uniquenesses_se
  dimnames(SE_L) <- dimnames(L)
  names(SE_psi) <- rownames(L)

  # A degenerate rotational gauge inflates the unrotated loading SEs without bound while leaving
  # every gauge-invariant quantity intact, so withhold the unrotated loadings alone and keep the
  # communalities and uniquenesses, which are invariant under Lambda -> Lambda T and unaffected.
  gauge_degenerate <- isFALSE(core$gauge_reliable)
  if (gauge_degenerate) SE_L[] <- NA_real_

  if (anyNA(SE_L) || anyNA(SE_psi)) {
    cli::cli_warn(
      c("Analytic standard errors could not be computed for all parameters.",
        "i" = if (gauge_degenerate) {
          "The factor solution's rotational orientation is only weakly determined (two canonical variances nearly coincide), so the unrotated loadings have no well-defined standard error. The communalities and uniquenesses are unaffected; apply a rotation, or use {.code se = \"np-boot\"}, for loading-level uncertainty."
        } else if (anyNA(SE_L) && .se_boundary_hint_applies(L)) {
          # As in `.se_information_rotated()`: the hint claims a total withholding, so it is only
          # attached once the loading SEs are actually gone. They and the uniqueness SEs are
          # all-or-nothing together today, but a path that withheld only one of them would
          # otherwise be described by a hint that its own output contradicts. And it claims a
          # Heywood case, so a non-finite solution -- which the gate withholds on too, at no
          # boundary -- takes the general clause below rather than this one.
          .se_boundary_hint
        } else {
          "This occurs when the bordered information matrix is singular, or when the asymptotic covariance is not usable."
        }),
      class = "efa_se_unreliable"
    )
  }

  # h2_i = 1 - psi_i exactly, so the communality is the uniqueness up to sign and shares its
  # standard error verbatim (the delta method the core already ran through the gradient
  # 2 L[i, ]). Report it here too, so an unrotated fit does not force the user to know the
  # identity to obtain communality intervals.
  h2 <- rowSums(L^2)
  names(h2) <- rownames(L)
  psi <- 1 - h2
  SE_h2 <- SE_psi
  z <- stats::qnorm(1 - (1 - ci) / 2)

  # `.se_sandwich_core()` always returns a numeric V_AA, even when `reliable = FALSE` (it is
  # only the marginal `loadings_se` that gets NA-filled in the unreliable branch). NA-fill the
  # persisted covariance to match, so a finite-but-not-PSD V_AA does not silently ship next to
  # NA SEs and propagate sqrt(NaN) into downstream pooling.
  # A degenerate gauge does NOT NA-fill the persisted block. The documented contract for that slot
  # is "NA-filled if the analytic covariance is unreliable (a Heywood case or a singular bordered
  # information matrix)", and a weakly determined orientation is neither: the covariance is finite,
  # PSD and -- verified against the closed-form reflexive inverse -- correct. What is unusable is
  # reading its diagonal as the standard error of a parameter the gauge does not pin down. Keeping
  # it finite also leaves consumers that pool in a COMMON gauge across fits (`efa_mi()`) able to
  # recover the gauge-invariant quantities, which a wholesale NA fill would take down with it.
  V_AA <- core$V_AA
  if (!isTRUE(core$reliable)) V_AA[] <- NA_real_
  lab <- .vec_loading_labels(L)
  if (!is.null(lab)) dimnames(V_AA) <- list(lab, lab)

  list(
    SE = list(unrot_loadings = SE_L, uniquenesses = SE_psi,
              communalities = SE_h2),
    CI = list(unrot_loadings = .wald_ci(L, SE_L, z),
              uniquenesses = .wald_ci(psi, SE_psi, z),
              communalities = .wald_ci(h2, SE_h2, z)),
    replicates = NULL,
    vcov_unrot_loadings = V_AA
  )
}

# Fold the robust scaled chi-square into the fit indices, overwriting the chi-square-derived block
# (.gof() leaves it undefined for DWLS and reports the non-robust discrepancy for ML/ULS). The
# reported chi-square is the scaled-shifted (WLSMV-default) statistic; the mean-adjusted and
# mean-and-variance-adjusted statistics and the scaling/shift are added as extra fields. CFI/TLI/
# RMSEA come from the scaled model and baseline statistics via the shared .chi_fit_indices().
# `ci` is the caller's own "solve the analytic RMSEA bounds" flag, threaded through rather than
# fixed: `.gof()` runs this on every cor_method = "fiml" fit, including the bootstrap replicates it
# calls with ci = FALSE, whose bounds are discarded by `.boot_se_ci()` anyway.
.apply_scaled_test <- function(fit_indices, st, N, ci = TRUE) {

  if (is.null(st)) return(fit_indices)

  # The scaled (sandwich) model and baseline statistics are already on a comparable scale, so they
  # serve directly as the CFI/TLI/RMSEA noncentrality inputs (chi_cfi / chi_null_cfi).
  idx <- .chi_fit_indices(st$chi, st$df, st$chi_null, st$df_null, N, st$m, ci = ci,
                          chi_cfi = st$chi, chi_null_cfi = st$chi_null)

  fit_indices$chi <- st$chi
  fit_indices$df <- st$df
  fit_indices$p_chi <- idx$p_chi
  fit_indices$CFI <- idx$CFI
  fit_indices$TLI <- idx$TLI
  fit_indices$RMSEA <- idx$RMSEA
  fit_indices$RMSEA_LB <- idx$RMSEA_LB
  fit_indices$RMSEA_UB <- idx$RMSEA_UB
  # AIC/BIC/ECVI are likelihood-ratio chi-square information criteria; they have no standard
  # interpretation when built on the moment-scaled (Satorra-Bentler) statistic, so leave them NA
  # rather than report a misleading model-comparison number.
  fit_indices$AIC <- NA_real_
  fit_indices$BIC <- NA_real_
  fit_indices$ECVI <- NA_real_
  fit_indices$chi_null <- st$chi_null
  fit_indices$df_null <- st$df_null
  fit_indices$p_null <- idx$p_null

  fit_indices$chi_scaled_type <- "scaled.shifted"
  fit_indices$chi_scaling <- st$chi_scaling
  fit_indices$chi_shift <- st$chi_shift
  fit_indices$chi_unscaled <- st$chi_unscaled
  fit_indices$chi_mean_adjusted <- st$chi_mean_adjusted
  fit_indices$chi_mean_var <- st$chi_mean_var
  fit_indices$df_mean_var <- st$df_mean_var

  fit_indices
}


.boot_se_ci <- function(fit_target, L_rot, boot_fit, boot_rot, ci, b) {

  l_ci <- (1 - ci) / 2
  ps <- c(l_ci, ci + l_ci)

  ### calculate stats for unrot loadings and gof measures
  L_unrot <- fit_target$unrot_loadings

  ncol_L <- ncol(L_unrot)
  nrow_L <- nrow(L_unrot)
  colnam_L <- colnames(L_unrot)
  rownam_L <- rownames(L_unrot)

  L_unrot_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))
  # The bootstrap aggregates only the numeric fit indices: a scaled chi-square fit (the
  # cor_method = "fiml" two-stage statistic, or any se = "sandwich" fit) adds the character
  # `chi_scaled_type` tag and the extra scaled-statistic components, which the quantile/SD
  # aggregation below cannot consume. Aligning each replicate to these names also lets a
  # replicate whose scaled statistic degenerated (and so lacks a component) contribute NA
  # there rather than shifting every column.
  gof_names <- names(fit_target$fit_indices)[
    vapply(fit_target$fit_indices, is.numeric, logical(1))]
  # The RMSEA confidence bounds are dropped: the replicate fits are run with ci = FALSE, so
  # they are NA in every replicate, and a bootstrap standard error of an interval bound is
  # not a meaningful quantity in the first place.
  gof_names <- setdiff(gof_names, c("RMSEA_LB", "RMSEA_UB"))
  gof_boot <- matrix(NA_real_, ncol = length(gof_names), nrow = b,
                     dimnames = list(NULL, gof_names))
  residuals_boot <- array(NA_real_, c(nrow_L, nrow_L, b),
                          dimnames = list(rownam_L, rownam_L,
                                          NULL))

  # Track replicates that could not be fit (NULL) or aligned, so they are
  # excluded from the bootstrap statistics rather than aborting the whole call.
  failed <- vapply(boot_fit, is.null, logical(1))

  for (boot_i in seq_len(b)) {

    if (failed[boot_i]) next

    # save aligned loading matrix
    aligned <- tryCatch(
      .align_solution(L_unrot, boot_fit[[boot_i]]$unrot_loadings),
      error = function(e) NULL
    )
    if (is.null(aligned)) {
      failed[boot_i] <- TRUE
      next
    }

    L_unrot_boot[,, boot_i] <- aligned$loadings
    fi_i <- boot_fit[[boot_i]]$fit_indices
    gof_boot[boot_i, ] <- vapply(gof_names, function(nm) {
      v <- fi_i[[nm]]
      if (is.null(v) || !is.numeric(v)) NA_real_ else as.numeric(v[1L])
    }, numeric(1))
    residuals_boot[,, boot_i] <- boot_fit[[boot_i]]$residuals

  }

  n_failed <- sum(failed)
  # Reported on the object so the effective B behind the standard errors stays recoverable
  # from a saved fit, not only from the warning below.
  valid_reps <- b - n_failed
  if (n_failed == b) {
    cli::cli_abort(
      c("All {b} bootstrap replicates failed; no bootstrap standard errors could be computed.",
        "i" = "The resampled correlation matrices may be degenerate; try more observations or fewer factors."),
      class = "efa_boot_all_failed"
    )
  }
  if (n_failed > 0) {
    cli::cli_warn(
      c("{n_failed} bootstrap replicate{?s} failed and {?was/were} excluded.",
        "i" = "Bootstrap standard errors and confidence intervals are based on {b - n_failed} replicate{?s}."),
      class = "efa_boot_replicate_failed"
    )
  }
  # A bootstrap standard error is the dispersion across replicates, so fewer than two leaves it
  # undefined (`sd()` of a single value is NA) and collapses the percentile interval onto that one
  # replicate. `b_boot` is bounded below at 2, so this is reachable only when replicates fail at
  # run time -- including the extreme case where all but one of a large `b_boot` do.
  if (valid_reps < 2L) {
    cli::cli_warn(
      c("Bootstrap standard errors are not defined from {valid_reps} usable replicate{?s}.",
        "i" = "The standard errors are {.val {NA}} and the confidence bounds collapse onto the single replicate; raise {.arg b_boot}, or check why the replicate fits failed."),
      class = "efa_se_unreliable"
    )
  }

  # se = sd of bootstrap replications (Zhang, 2014, Estimating Standard Errors
  # in Exploratory Factor Analysis)
  L_unrot_se_ci <- .array_se_ci(L_unrot_boot, ps)
  gof_se_ci <- .array_se_ci(gof_boot, ps, M = 2)
  residuals_se_ci <- .array_se_ci(residuals_boot, ps)
  names(gof_se_ci$se) <- gof_names
  names(gof_se_ci$ci$lower) <- gof_names
  names(gof_se_ci$ci$upper) <- gof_names

  if(boot_rot == "oblique") {

    colnam_L <- colnames(L_rot)

    L_rot_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))
    Phi_rot_boot <- array(NA_real_, c(ncol_L, ncol_L, b),
                          dimnames = list(colnam_L, colnam_L,
                                          NULL))
    Structure_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))

    failed_rot <- 0

    # Align every successfully fit replicate to the target in a single compiled
    # call over the loading cube, rather than one efa_procrustes() round trip per
    # replicate. Replicates that failed to fit or could not be aligned to the
    # point estimate are excluded from the cube and stay NA in the output arrays.
    keep <- which(!failed)
    if (length(keep) > 0) {
      A_cube <- array(NA_real_, c(nrow_L, ncol_L, length(keep)))
      for (j in seq_along(keep)) {
        A_cube[, , j] <- boot_fit[[keep[j]]]$unrot_loadings
      }

      aligned <- .oblique_procrustes_batch(A_cube, L_rot, random_starts = 5)

      for (j in seq_along(keep)) {
        boot_i <- keep[j]
        # Exclude a replicate only when no valid alignment could be produced. The
        # best multi-start fit is kept even if it did not formally converge: it is
        # the lowest-objective alignment available and its loadings are well-defined.
        if (!isTRUE(aligned$valid[j])) {
          failed_rot <- failed_rot + 1
          next
        }
        L_j <- matrix(aligned$loadings[, , j], nrow_L, ncol_L)
        Phi_j <- matrix(aligned$Phi[, , j], ncol_L, ncol_L)
        L_rot_boot[, , boot_i] <- L_j
        Phi_rot_boot[, , boot_i] <- Phi_j
        Structure_boot[, , boot_i] <- L_j %*% Phi_j
      }
    }

    valid_rot <- b - n_failed - failed_rot
    if (failed_rot > 0) {
      cli::cli_warn(c("{failed_rot} target rotation{?s} in the bootstrap procedure could not be aligned.",
                    "i" = "Bootstrap SE and CI of rotated loadings, factor correlations and structure coefficients are based on {valid_rot} bootstrap sample{?s}."),
                    class = "efa_boot_rotation_failed")
    }

    L_rot_se_ci <- .array_se_ci(L_rot_boot, ps)
    Phi_rot_se_ci <- .array_se_ci(Phi_rot_boot, ps)
    Structure_se_ci <- .array_se_ci(Structure_boot, ps)

    out <- list(
      SE = list(
        unrot_loadings = L_unrot_se_ci$se,
        rot_loadings = L_rot_se_ci$se,
        Phi = Phi_rot_se_ci$se,
        Structure = Structure_se_ci$se,
        fit_indices = gof_se_ci$se,
        residuals = residuals_se_ci$se,
        valid_replicates = valid_reps,
        valid_target_rotations = valid_rot
      ),
      CI = list(
        unrot_loadings = L_unrot_se_ci$ci,
        rot_loadings = L_rot_se_ci$ci,
        Phi = Phi_rot_se_ci$ci,
        Structure = Structure_se_ci$ci,
        fit_indices = gof_se_ci$ci,
        residuals = residuals_se_ci$ci
      ),
      replicates = list(
        unrot_loadings = L_unrot_boot,
        rot_loadings = L_rot_boot,
        Phi = Phi_rot_boot,
        Structure = Structure_boot,
        fit_indices = gof_boot,
        residuals = residuals_boot
      )
    )

  } else if (boot_rot == "orthogonal") {
    colnam_L <- colnames(L_rot)

    L_rot_boot <- array(NA_real_, c(nrow_L, ncol_L, b),
                        dimnames = list(rownam_L, colnam_L,
                                        NULL))

    failed_rot <- 0

    for (boot_i in seq_len(b)) {

      if (failed[boot_i]) next

      # save target-rotated loading matrix
      aligned_i <- tryCatch(
        efa_procrustes(boot_fit[[boot_i]]$unrot_loadings,
                       Target = L_rot, rotation = "orthogonal"),
        error = function(e) NULL
      )
      if (is.null(aligned_i)) {
        failed_rot <- failed_rot + 1
        next
      }
      L_rot_boot[,, boot_i] <- aligned_i$loadings
    }

    valid_rot <- b - n_failed - failed_rot
    if (failed_rot > 0) {
      cli::cli_warn(c("{failed_rot} target rotation{?s} in the bootstrap procedure could not be aligned.",
                    "i" = "Bootstrap SE and CI of rotated loadings are based on {valid_rot} bootstrap sample{?s}."),
                    class = "efa_boot_rotation_failed")
    }

    L_rot_se_ci <- .array_se_ci(L_rot_boot, ps)

    out <- list(
      SE = list(
        unrot_loadings = L_unrot_se_ci$se,
        rot_loadings = L_rot_se_ci$se,
        fit_indices = gof_se_ci$se,
        residuals = residuals_se_ci$se,
        valid_replicates = valid_reps,
        valid_target_rotations = valid_rot
      ),
      CI = list(
        unrot_loadings = L_unrot_se_ci$ci,
        rot_loadings = L_rot_se_ci$ci,
        fit_indices = gof_se_ci$ci,
        residuals = residuals_se_ci$ci
      ),
      replicates = list(
        unrot_loadings = L_unrot_boot,
        rot_loadings = L_rot_boot,
        fit_indices = gof_boot,
        residuals = residuals_boot
      )
    )


  } else {

    out <- list(
      SE = list(
        unrot_loadings = L_unrot_se_ci$se,
        fit_indices = gof_se_ci$se,
        residuals = residuals_se_ci$se,
        valid_replicates = valid_reps
      ),
      CI = list(
        unrot_loadings = L_unrot_se_ci$ci,
        fit_indices = gof_se_ci$ci,
        residuals = residuals_se_ci$ci
      ),
      replicates = list(
        unrot_loadings = L_unrot_boot,
        fit_indices = gof_boot,
        residuals = residuals_boot
      )
    )

  }


  out

}

# Bootstrap standard deviation and percentile interval of a replicate array, over the margins `M`
# (the replicate index is whichever dimension is left over).
#
# Both interval bounds come from ONE sweep. `stats::quantile()` sorts each cell's length-B
# replicate vector, so requesting the two probabilities in separate sweeps sorts every cell twice;
# asking for both at once halves the sorting and the traversals. `apply()` puts the two results on
# a new LEADING dimension, so `q` is length(probs) x dim(x)[M] whatever `M` is; flattening it to a
# 2 x length(se) matrix indexes that dimension without having to know the rank. `se` carries the
# target shape and dimnames and the bounds are reshaped onto it, reproducing what a per-probability
# `apply()` returned for every rank this is called with (a vector for a matrix input, a matrix for
# a 3-D cube, a 3-D array for the group bootstrap's 4-D cube). Quantile type (the default 7) and
# `na.rm` are unchanged, so the bounds are identical to sweeping twice.
.array_se_ci <- function(x, probs, M = c(1, 2)) {
  se <- apply(x, M, stats::sd, na.rm = TRUE)
  q <- apply(x, M, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE)
  dim(q) <- c(length(probs), length(se))

  bound <- function(i) {
    b <- q[i, ]
    if (is.null(dim(se))) {
      names(b) <- names(se)
    } else {
      dim(b) <- dim(se)
      dimnames(b) <- dimnames(se)
    }
    b
  }

  list(
    se = se,
    ci = list(lower = bound(1L), upper = bound(2L))
  )
}

# Analytic expected-information standard errors for the ML solution.

# Independent reference for the correlation-structure information. theta = vec(Lambda) alone (the
# analysed diagonal is fixed at 1, so the uniquenesses are a function of Lambda, not free
# parameters): Gamma is assembled by an elementwise double loop over the pair index using the
# Olkin-Siotani formula, the model Jacobian Delta and the constraint Jacobian by central finite
# differences, and the information Delta' Gamma^-1 Delta is bordered, inverted and scaled by
# 1 / (N - 1). This shares no code with the package's vectorised assembly.
.ref_se_information <- function(L, psi, N) {
  p <- nrow(L); k <- ncol(L); pk <- p * k
  Sigma <- tcrossprod(L); diag(Sigma) <- 1
  prs <- utils::combn(p, 2L); n <- ncol(prs)

  # N Cov(r_ij, r_kl) under normality, one pair combination at a time.
  G <- matrix(0, n, n)
  for (s in seq_len(n)) for (t in seq_len(n)) {
    i <- prs[1, s]; j <- prs[2, s]; kk <- prs[1, t]; l <- prs[2, t]
    G[s, t] <- 0.5 * Sigma[i, j] * Sigma[kk, l] *
      (Sigma[i, kk]^2 + Sigma[i, l]^2 + Sigma[j, kk]^2 + Sigma[j, l]^2) +
      Sigma[i, kk] * Sigma[j, l] + Sigma[i, l] * Sigma[j, kk] -
      Sigma[i, j] * (Sigma[i, kk] * Sigma[i, l] + Sigma[j, kk] * Sigma[j, l]) -
      Sigma[kk, l] * (Sigma[i, kk] * Sigma[j, kk] + Sigma[i, l] * Sigma[j, l])
  }

  # sigma_offdiag(vec Lambda) and the gauge constraint off-diag(Lambda' Psi^-1 Lambda) = 0, both
  # differentiated by central finite differences.
  sig <- function(th) { Lm <- matrix(th, p, k); tcrossprod(Lm)[t(prs)] }
  g <- function(th) { Lm <- matrix(th, p, k); ps <- 1 - rowSums(Lm^2)
                      M <- crossprod(Lm, Lm / ps); M[upper.tri(M)] }
  th0 <- as.vector(L); nc <- k * (k - 1L) / 2L
  Delta <- matrix(0, n, pk); Cmat <- matrix(0, nc, pk); h <- 1e-6
  for (jj in seq_len(pk)) {
    tp <- th0; tp[jj] <- tp[jj] + h; tm <- th0; tm[jj] <- tm[jj] - h
    Delta[, jj] <- (sig(tp) - sig(tm)) / (2 * h)
    if (nc > 0) Cmat[, jj] <- (g(tp) - g(tm)) / (2 * h)
  }

  A <- crossprod(Delta, solve(G, Delta))
  Aug <- if (nc > 0) rbind(cbind(A, t(Cmat)), cbind(Cmat, matrix(0, nc, nc))) else A
  V <- solve(Aug)[seq_len(pk), seq_len(pk), drop = FALSE] / (N - 1)
  se <- sqrt(diag(V))
  # h2_i = rowSums(Lambda^2)_i has gradient 2 Lambda[i, ]; psi_i = 1 - h2_i shares its SE.
  G_h <- matrix(0, p, pk)
  for (i in seq_len(p)) G_h[i, (seq_len(k) - 1L) * p + i] <- 2 * L[i, ]
  list(loadings_se = matrix(se, p, k),
       uniquenesses_se = sqrt(rowSums((G_h %*% V) * G_h)))
}


# Match the columns of B to the columns of A by maximum absolute loading-column correlation,
# returning the permutation and signs that align B to A. Rotated solutions order and sign factors
# arbitrarily across packages and bootstrap replicates, so the rotated SEs must be aligned before
# they are compared.
.align_factor_cols <- function(A, B) {
  k <- ncol(A); perm <- integer(k); sgn <- numeric(k); used <- logical(k)
  for (j in seq_len(k)) {
    cors <- vapply(seq_len(k),
                   function(i) if (used[i]) NA_real_ else stats::cor(A[, j], B[, i]), 0)
    i <- which.max(abs(cors)); perm[j] <- i; sgn[j] <- sign(cors[i]); used[i] <- TRUE
  }
  list(perm = perm, sgn = sgn)
}


# The three-factor information-SE fits on the baseline correlation matrix -- one unrotated, one
# oblimin -- are each needed by more than one block below, and each assembles and inverts a
# 54 x 54 bordered information matrix, so each is computed once and shared. The memoisation is
# lazy so that a block skipped before it asks never pays for the fit, and so that the oblimin fit's
# random rotation starts are drawn at the point in the RNG stream where the first consumer runs.
.info_fit_unrot <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      cache <<- EFA(test_models$baseline$cormat, n_factors = 3, N = 500, method = "ML",
                    rotation = "none", se = "information")
    }
    cache
  }
})

.info_fit_oblq <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      cache <<- EFA(test_models$baseline$cormat, n_factors = 3, N = 500, method = "ML",
                    rotation = "oblimin", se = "information")
    }
    cache
  }
})


test_that("information SEs match an independent correlation-structure reference", {
  # The two factors carry deliberately different strengths, so the ML canonical variances
  # diag(Lambda' Psi^-1 Lambda) are well separated (4.23 vs 1.16) and the rotational gauge is
  # firmly identified. Equal strengths leave the orientation unpinned, and the unrotated loading
  # SEs then diverge along the gauge -- a regime the path deliberately withholds (see the gauge
  # tests below), which would leave nothing here to compare against the reference.
  L <- matrix(c(0.80, 0.75, 0.70, 0.10, 0.00, 0.20,
                0.10, 0.05, 0.20, 0.55, 0.50, 0.45), 6, 2)
  psi <- 1 - rowSums(L^2)
  fit_out <- list(unrot_loadings = L)

  ours <- EFAtools:::.se_information(fit_out, rot_info = NULL, N = 250, ci = 0.95,
                                     method = "ML")
  ref <- .ref_se_information(L, psi, N = 250)

  expect_equal(unname(ours$SE$unrot_loadings), ref$loadings_se, tolerance = 1e-5)
  expect_equal(unname(ours$SE$uniquenesses), ref$uniquenesses_se, tolerance = 1e-5)
})


test_that("the normal-theory Gamma reproduces the known variance of a correlation", {
  # The diagonal of the Olkin-Siotani covariance is the classical asymptotic variance of a
  # Pearson correlation, N Var(r_ij) = (1 - rho_ij^2)^2, which pins the formula's scale and its
  # pair ordering independently of the off-diagonal terms.
  R <- test_models$baseline$cormat
  prs <- utils::combn(ncol(R), 2L)
  G <- EFAtools:::.normal_theory_gamma(R, prs)

  expect_equal(unname(diag(G)), unname((1 - R[t(prs)]^2)^2), tolerance = 1e-12)
  expect_true(isSymmetric(G, tol = 1e-12))
  # It is the covariance of the sample correlations, so it must be positive definite at a
  # positive-definite Sigma -- the information Delta' Gamma^-1 Delta is built by inverting it.
  expect_gt(min(eigen(G, symmetric = TRUE, only.values = TRUE)$values), 0)
})


test_that("information uniqueness and loading SEs agree with the robust sandwich", {
  skip_on_cran()

  # `information` and `sandwich` build the same correlation-structure covariance from different
  # meats -- the normal-theory Gamma at the model-implied Sigma versus the fourth-moment (ADF)
  # Gamma of the observed correlations -- so on multivariate normal data they estimate the same
  # quantity and agree up to finite-sample meat noise. That makes the sandwich an independent
  # check on the uniqueness SEs within the correlation metric. lavaan's default fit is NOT the
  # comparator here: it analyses a covariance structure, so its uniqueness SEs carry the sampling
  # variability of a diagonal that is fixed at 1 in the model EFAtools fits (and under
  # `correlation = TRUE` lavaan derives the uniquenesses and reports their SE as 0).
  set.seed(11)
  R0 <- test_models$baseline$cormat
  X <- matrix(stats::rnorm(3000 * ncol(R0)), 3000) %*% chol(R0)
  colnames(X) <- colnames(R0)

  fi <- EFA(X, n_factors = 3, method = "ML", rotation = "none", se = "information")
  fs <- EFA(X, n_factors = 3, method = "ML", rotation = "none", se = "sandwich")

  expect_lt(abs(stats::median(fi$SE$uniquenesses / fs$SE$uniquenesses) - 1), 0.1)
  expect_lt(abs(stats::median(fi$SE$unrot_loadings / fs$SE$unrot_loadings) - 1), 0.1)
  expect_gt(stats::cor(c(fi$SE$unrot_loadings), c(fs$SE$unrot_loadings)), 0.9)
})


test_that("single-factor information loading SEs match lavaan's correlation structure", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  # With one factor there is no rotational indeterminacy, so the loadings are identified (up to
  # sign) the same way in both packages and their SEs are directly comparable. lavaan must be
  # asked for a CORRELATION structure (`correlation = TRUE`): its default fits a covariance
  # structure, which treats the analysed diagonal as carrying sampling variability and returns
  # loading SEs roughly twice as large. The residual gap is the N vs N - 1 scaling plus lavaan
  # evaluating its weight at the sample rather than the model-implied matrix.
  R <- stats::cor(GRiPS_raw)
  N <- nrow(GRiPS_raw)
  fit <- EFA(GRiPS_raw, n_factors = 1, method = "ML", rotation = "none", se = "information")

  mod <- paste0("f =~ ", paste(colnames(R), collapse = " + "))
  lav_fit <- lavaan::cfa(mod, data = as.data.frame(GRiPS_raw), std.lv = TRUE,
                         correlation = TRUE)
  pe <- lavaan::parameterEstimates(lav_fit)
  ld <- pe[pe$op == "=~", ]; ld <- ld[match(colnames(R), ld$rhs), ]

  # The residual few percent is the evaluation point: EFAtools uses the EXPECTED information, so
  # Gamma is taken at the model-implied Sigma, whereas lavaan evaluates it at the sample matrix.
  expect_equal(as.vector(fit$SE$unrot_loadings), ld$se, tolerance = 0.08)

  # ... and NOT lavaan's covariance-structure SEs, which are materially larger. Pinned as a
  # direction so a regression back to the covariance-structure information cannot pass silently.
  lav_cov <- lavaan::cfa(mod, data = as.data.frame(GRiPS_raw), std.lv = TRUE)
  se_cov <- lavaan::parameterEstimates(lav_cov)
  se_cov <- se_cov[se_cov$op == "=~", ]
  se_cov <- se_cov[match(colnames(R), se_cov$rhs), "se"]
  expect_gt(stats::median(se_cov / as.vector(fit$SE$unrot_loadings)), 1.2)
})


test_that("information loading SEs track the bootstrap on an identifiable solution", {
  skip_on_cran()
  skip_if_not_slow()

  # Well-separated three-factor structure at a large N, so the unrotated solution is
  # cleanly identified and the bootstrap alignment is stable. The data are multivariate
  # normal and the model is correctly specified, the regime in which the expected-information
  # and the bootstrap SE are both consistent and coincide (Yuan & Hayashi, 2006); they are
  # different finite-sample estimators, so they agree in structure (strong correlation)
  # rather than to a tight per-element tolerance.
  set.seed(123)
  R0 <- test_models$baseline$cormat
  X <- matrix(rnorm(2500 * ncol(R0)), 2500) %*% chol(R0)
  colnames(X) <- colnames(R0)

  fa <- EFA(X, n_factors = 3, method = "ML", rotation = "none", se = "information")
  fb <- EFA(X, n_factors = 3, method = "ML", rotation = "none",
            se = "np-boot", b_boot = 250, seed = 1)

  expect_gt(cor(c(fa$SE$unrot_loadings), c(fb$SE$unrot_loadings)), 0.9)
  expect_lt(abs(median(fa$SE$unrot_loadings / fb$SE$unrot_loadings) - 1), 0.35)
})


test_that("information SEs populate the bootstrap SE/CI schema", {
  R <- test_models$baseline$cormat
  fit <- .info_fit_unrot()

  expect_setequal(names(fit$SE), c("unrot_loadings", "uniquenesses", "communalities"))
  expect_setequal(names(fit$CI), c("unrot_loadings", "uniquenesses", "communalities"))
  expect_null(fit$replicates)

  expect_equal(dim(fit$SE$unrot_loadings), dim(fit$unrot_loadings))
  expect_length(fit$SE$uniquenesses, ncol(R))
  expect_false(anyNA(fit$SE$unrot_loadings))

  # h2 = 1 - psi exactly, so the unrotated path reports the communality SEs it already holds,
  # with the mirrored Wald interval.
  expect_equal(unname(fit$SE$communalities), unname(fit$SE$uniquenesses))
  expect_equal(unname(fit$CI$communalities$upper), unname(1 - fit$CI$uniquenesses$lower))
  expect_equal(unname(fit$CI$communalities$lower), unname(1 - fit$CI$uniquenesses$upper))

  # Wald intervals bracket the point estimates.
  expect_true(all(fit$CI$unrot_loadings$lower <= fit$unrot_loadings))
  expect_true(all(fit$unrot_loadings <= fit$CI$unrot_loadings$upper))
  expect_named(fit$CI$uniquenesses, c("lower", "upper"))
})


test_that("information SEs populate the rotated SE/CI schema under an oblique rotation", {
  R <- test_models$baseline$cormat
  fit <- .info_fit_oblq()

  expect_setequal(names(fit$SE),
                  c("unrot_loadings", "uniquenesses", "rot_loadings",
                    "communalities", "Phi", "Structure"))
  expect_setequal(names(fit$CI),
                  c("unrot_loadings", "uniquenesses", "rot_loadings",
                    "communalities", "Phi", "Structure"))
  expect_null(fit$replicates)

  expect_equal(dim(fit$SE$rot_loadings), dim(fit$rot_loadings))
  expect_equal(dim(fit$SE$Structure), dim(fit$rot_loadings))
  expect_equal(dim(fit$SE$Phi), dim(fit$Phi))
  expect_length(fit$SE$communalities, ncol(R))
  expect_false(anyNA(fit$SE$rot_loadings))

  # The unit diagonal of Phi is fixed, so it carries no sampling variance.
  expect_true(all(diag(fit$SE$Phi) == 0))

  # Wald intervals bracket every rotated point estimate.
  expect_true(all(fit$CI$rot_loadings$lower <= fit$rot_loadings &
                    fit$rot_loadings <= fit$CI$rot_loadings$upper))
  expect_true(all(fit$CI$Structure$lower <= fit$Structure &
                    fit$Structure <= fit$CI$Structure$upper))
  expect_named(fit$CI$Phi, c("lower", "upper"))

  # h2_i = 1 - psi_i exactly, so the communality and uniqueness are a single estimand up to sign
  # and must carry the same standard error (and a mirrored Wald interval). Both are the ordinary
  # delta method on the loading covariance, so they agree by construction -- under a
  # covariance-structure information the two routes disagreed by up to 18%. The unrotated
  # assembly path is a different one, and pins the same contract in the block above.
  expect_equal(unname(fit$SE$communalities), unname(fit$SE$uniquenesses))
  expect_equal(unname(fit$CI$communalities$upper), unname(1 - fit$CI$uniquenesses$lower))
  expect_equal(unname(fit$CI$communalities$lower), unname(1 - fit$CI$uniquenesses$upper))
})


test_that("an unusable covariance NAs the communality SEs along with the uniqueness SEs", {
  # The post-solve PSD gate leaves a FINITE covariance next to NA marginal SEs (only the marginals
  # are NA-filled). The communalities are a delta method on that covariance, so they must fail
  # closed with it -- otherwise the object ships a finite `SE$communalities` beside an all-NA
  # `SE$uniquenesses` and an all-NA `vcov_unrot_loadings`, contradicting both the one-estimand
  # contract and the "no finite SE next to an NA covariance" invariant.
  R <- test_models$baseline$cormat
  fit <- EFA(R, n_factors = 3, N = 500, method = "ML", rotation = "oblimin", se = "none")
  fit_out <- list(unrot_loadings = fit$unrot_loadings)
  rot_info <- list(rotation = "oblimin", rotmat = fit$rotmat, rot_loadings = fit$rot_loadings,
                   Phi = fit$Phi, normalize = fit$settings$normalize,
                   crit_args = list(gam = 0, delta = 0.01))

  out <- testthat::with_mocked_bindings(
    suppressWarnings(
      EFAtools:::.se_information(fit_out, rot_info, N = 500, ci = 0.95, method = "ML")),
    .is_psd = function(M) FALSE,
    .package = "EFAtools"
  )

  expect_true(all(is.na(out$SE$uniquenesses)))
  expect_true(all(is.na(out$SE$communalities)))
  expect_true(all(is.na(out$SE$rot_loadings)))
  expect_true(all(is.na(out$vcov_unrot_loadings)))
})


test_that("information SEs under an orthogonal rotation omit Phi and the structure matrix", {
  R <- test_models$baseline$cormat
  fit <- EFA(R, n_factors = 3, N = 500, method = "ML", rotation = "varimax", se = "information")

  expect_setequal(names(fit$SE),
                  c("unrot_loadings", "uniquenesses", "rot_loadings", "communalities"))
  expect_null(fit$SE$Phi)
  expect_null(fit$SE$Structure)
  expect_equal(dim(fit$SE$rot_loadings), dim(fit$rot_loadings))
  expect_false(anyNA(fit$SE$rot_loadings))
  expect_false(anyNA(fit$SE$communalities))
})


test_that("information SEs are produced for every supported native rotation", {
  # The lavaan oracle below validates the SE magnitude for a representative rotation; this guards
  # every native criterion against a wrong criterion mapping in `.rotation_se_method` or a
  # warm-start reproduction failure that would silently degrade its rotated SEs to NA. All eleven
  # mapped rotations are covered: equamax (param = k / (2 p)) and geominT/geominQ (param = delta)
  # are the only entries whose Jacobian parameter is COMPUTED rather than fixed, so a transposed
  # k / p or a drifted delta default would hide there. The bifactor REORDER path is exercised by
  # the next block; this block covers the non-reorder case.
  R <- test_models$baseline$cormat
  orth <- c("varimax", "quartimax", "equamax", "bentlerT", "geominT", "bifactorT")
  oblq <- c("oblimin", "quartimin", "bentlerQ", "geominQ", "bifactorQ")
  for (rot in c(orth, oblq)) {
    # The native criterion rotations run `random_starts` random starts, so without a seed inside
    # the loop which local optimum is tested depends on the RNG state left by the preceding
    # iteration (and by preceding test blocks).
    set.seed(51)
    fit <- EFA(R, n_factors = 3, N = 500, method = "ML", rotation = rot, se = "information")
    expect_false(anyNA(fit$SE$rot_loadings), info = rot)
    expect_equal(dim(fit$SE$rot_loadings), dim(fit$rot_loadings), info = rot)
    expect_false(anyNA(fit$SE$communalities), info = rot)
    if (rot %in% oblq) {
      expect_false(anyNA(fit$SE$Phi), info = rot)
      expect_false(anyNA(fit$SE$Structure), info = rot)
      expect_true(all(diag(fit$SE$Phi) == 0), info = rot)
    } else {
      expect_null(fit$SE$Phi, info = rot)
    }
  }
})


test_that("bifactor rotated SEs survive a general-factor reorder", {
  # When a group factor has a larger sum of squared loadings than the general factor,
  # `.reflect_and_order` moves the general factor out of the first column. The analytic SE
  # re-rotation must exempt the general factor wherever it landed (not blindly column 1), or it
  # cannot reproduce the reported loadings and the SEs collapse to NA. A weak general factor (0.30
  # on every variable) with two strong group factors (0.75) forces that reorder.
  set.seed(1)
  p <- 12
  Lp <- matrix(0, p, 3); Lp[, 1] <- 0.30; Lp[1:6, 2] <- 0.75; Lp[7:12, 3] <- 0.75
  Sig <- Lp %*% t(Lp); diag(Sig) <- 1
  X <- matrix(stats::rnorm(800 * p), 800) %*% chol(Sig)
  colnames(X) <- paste0("v", seq_len(p))

  for (rot in c("bifactorT", "bifactorQ")) {
    # The two group factors are equally strong by construction, so their canonical variances
    # nearly coincide and the unrotated orientation is weakly determined. The fit therefore
    # warns (`efa_se_unreliable`) and withholds the unrotated loading SEs -- that is the
    # documented gauge behaviour, not a failure, and the point of this test is that it must
    # not take the rotated SEs down with it. The warning is suppressed rather than asserted
    # by class because it sits on a numeric threshold and need not fire identically on every
    # BLAS; the substantive assertion below is what this test pins.
    fit <- suppressWarnings(
      EFA(X, n_factors = 3, method = "ML", rotation = rot, se = "information"))
    # the general factor (the column loading on every variable) is not in column 1 here
    expect_gt(which.max(apply(abs(unclass(fit$rot_loadings)), 2, min)), 1L)
    expect_false(anyNA(fit$SE$rot_loadings), info = rot)
  }
})


test_that("rotated information loading and Phi SEs match lavaan's delta method", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  set.seed(42)
  R0 <- test_models$baseline$cormat
  X <- matrix(stats::rnorm(1500 * ncol(R0)), 1500) %*% chol(R0)
  colnames(X) <- colnames(R0)
  N <- 1500

  # normalize = FALSE matches lavaan's row.weights = "none"; oblimin (gam = 0) is quartimin, the
  # criterion lavaan rotates to under rotation = "oblimin".
  fit <- suppressWarnings(
    EFA(X, n_factors = 3, method = "ML", rotation = "oblimin", normalize = FALSE,
        se = "information"))
  if (utils::packageVersion("lavaan") >= "0.7.0") {
    ef <- lavaan::efa(sample.cov = stats::cor(X), sample.nobs = N, nfactors = 3,
                      rotation = list("oblimin", row_weights = "none"),
                      rotation.se = "delta")
  } else {
    # lavaan < 0.7 used a separate rotation.args argument.
    ef <- lavaan::efa(sample.cov = stats::cor(X), sample.nobs = N, nfactors = 3,
                      rotation = "oblimin", rotation.se = "delta",
                      rotation.args = list(row.weights = "none"))
  }
  lf <- Filter(function(e) inherits(e, "lavaan"), ef)[[1]]
  Ll <- lavaan::lavInspect(lf, "est")$lambda
  SEl <- lavaan::lavInspect(lf, "se")$lambda

  Lt <- unclass(fit$rot_loadings)
  al <- .align_factor_cols(Ll, Lt)
  Lt_a <- sweep(Lt[, al$perm, drop = FALSE], 2, al$sgn, "*")
  SEt_a <- fit$SE$rot_loadings[, al$perm, drop = FALSE]

  # Both packages reach the same (identification-invariant) rotated solution.
  expect_equal(max(abs(Lt_a - Ll)), 0, tolerance = 0.01)
  # Their delta-method loading SEs agree closely overall -- the same rotation Jacobian applied to
  # the same solution -- but not exactly, because `lavaan::efa()` propagates a COVARIANCE-structure
  # unrotated covariance (it has no correlation-structure EFA path) while EFAtools propagates the
  # correlation-structure one.
  expect_gt(stats::cor(c(SEt_a), c(SEl)), 0.93)
  expect_lt(mean(abs(SEt_a - SEl)), 0.004)

  # That difference has a signature, and it is the one that would disappear if the unrotated
  # covariance regressed to a covariance structure: it is confined to the SALIENT loadings (where
  # the spurious standardisation variance accumulates) and leaves the near-zero loadings alone.
  salient <- abs(Ll) > 0.4
  expect_lt(max((SEt_a / SEl)[salient]), 0.98)
  expect_lt(abs(stats::median((SEt_a / SEl)[!salient]) - 1), 0.05)

  # Factor-correlation SEs, matched by correlation value (the two packages order factors differently
  # so the off-diagonal positions do not line up).
  pe <- lavaan::parameterEstimates(lf)
  fc <- pe[pe$op == "~~" & pe$lhs != pe$rhs, ]
  ours <- data.frame(corr = fit$Phi[lower.tri(fit$Phi)],
                     se = fit$SE$Phi[lower.tri(fit$SE$Phi)])
  ours <- ours[order(ours$corr), ]
  fc <- fc[order(fc$est), ]
  expect_lt(max(abs(ours$se - fc$se)), 0.01)

  # TODO: `SE$Structure` has no external oracle here. `lavaan::efa()` reports no standard error
  # for the structure matrix, so validating it against lavaan would mean re-deriving
  # J_S = (Phi' (x) I_p) J_L + (I_k (x) L) J_Phi from lavaan's own parameter covariance inside the
  # test -- i.e. re-implementing the code under test. It is currently checked for magnitude
  # against a 300-replicate bootstrap in the block below (slow-gated) only.
})


test_that("rotated information SEs track the bootstrap", {
  skip_on_cran()
  skip_if_not_slow()

  set.seed(7)
  R0 <- test_models$baseline$cormat
  X <- matrix(stats::rnorm(2000 * ncol(R0)), 2000) %*% chol(R0)
  colnames(X) <- colnames(R0)

  fa <- EFA(X, n_factors = 3, method = "ML", rotation = "oblimin", se = "information")
  fb <- EFA(X, n_factors = 3, method = "ML", rotation = "oblimin",
            se = "np-boot", b_boot = 300, seed = 1)

  al <- .align_factor_cols(unclass(fa$rot_loadings), unclass(fb$rot_loadings))
  SEb <- fb$SE$rot_loadings[, al$perm, drop = FALSE]

  # Bootstrap rotated-loading SEs carry target-rotation (Procrustes) alignment noise that the
  # analytic delta method does not, so the two are different finite-sample estimators (Yuan &
  # Hayashi, 2006) that agree in overall magnitude (median ratio near 1) rather than cell by cell.
  expect_lt(abs(stats::median(fa$SE$rot_loadings / SEb) - 1), 0.3)

  SEs_b <- fb$SE$Structure[, al$perm, drop = FALSE]
  expect_lt(abs(stats::median(fa$SE$Structure / SEs_b) - 1), 0.3)

  # Communalities are rotation-invariant (no alignment), so they are compared cell by cell. The
  # analytic communality SE is the delta-method value it shares with the uniqueness
  # (h2 = 1 - psi); the bootstrap has no communality slot, so the reference is recomputed from the
  # replicate loadings. The two agree in magnitude (median ratio near 1); the normal-theory SE and
  # the bootstrap are different finite-sample estimators, so the per-variable pattern correlation is
  # positive but looser than the magnitude agreement.
  comm_b <- apply(fb$replicates$unrot_loadings, 3, function(L) rowSums(L^2))
  comm_b_se <- apply(comm_b, 1, stats::sd, na.rm = TRUE)
  expect_gt(stats::cor(fa$SE$communalities, comm_b_se), 0.4)
  expect_lt(abs(stats::median(fa$SE$communalities / comm_b_se) - 1), 0.4)
})


test_that("a Heywood case under a rotation yields NA rotated SEs with a classed warning", {
  # A communality above one drives a uniqueness below its zero boundary, where the solution sits
  # on the parameter-space boundary and the Wald intervals this path reports are not valid, so no
  # analytic SE -- rotated or unrotated -- is offered.
  L <- matrix(c(sqrt(1.11), 0.6, 0.5, 0.4, 0.2, 0.7,
                0.1, 0.05, 0.2, 0.3, 0.15, 0.25), 6, 2)   # h2[1] > 1, so psi[1] < 0
  fit_out <- list(unrot_loadings = L)
  rot_info <- list(rotation = "oblimin", rotmat = diag(2), rot_loadings = L,
                   Phi = diag(2), normalize = FALSE, crit_args = list(gam = 0, delta = 0.01))

  expect_warning(
    out <- EFAtools:::.se_information(fit_out, rot_info, N = 200, ci = 0.95, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(anyNA(out$SE$rot_loadings))
  expect_true(anyNA(out$SE$Phi))
  expect_true(anyNA(out$SE$communalities))
})


test_that("information SEs are available from a correlation matrix and match the raw fit", {
  set.seed(5)
  R0 <- test_models$baseline$cormat
  X <- matrix(rnorm(800 * ncol(R0)), 800) %*% chol(R0)
  colnames(X) <- colnames(R0)

  raw <- EFA(X, n_factors = 3, method = "ML", rotation = "none", se = "information")
  cmat <- EFA(stats::cor(X), n_factors = 3, N = 800, method = "ML",
              rotation = "none", se = "information")

  expect_equal(raw$SE$unrot_loadings, cmat$SE$unrot_loadings, tolerance = 1e-5)
  expect_equal(raw$SE$uniquenesses, cmat$SE$uniquenesses, tolerance = 1e-5)
})


test_that("a Heywood case yields NA SEs with a classed warning", {
  # A communality above one drives a uniqueness below its zero boundary. The correlation-structure
  # information is still assembleable there, but the solution sits on the parameter-space boundary,
  # where a Wald interval is not valid, so the path declines to report one.
  L <- matrix(c(sqrt(1.11), 0.6, 0.5, 0.4), 4, 1)   # h2[1] = 1.11 > 1, so psi[1] < 0
  fit_out <- list(unrot_loadings = L)

  expect_warning(
    out <- EFAtools:::.se_information(fit_out, rot_info = NULL, N = 200, ci = 0.95,
                                      method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(anyNA(out$SE$unrot_loadings))
})


test_that("a uniqueness pinned at the fitter's floor is treated as a Heywood case", {
  # The ML/ULS/DWLS fitters constrain the uniquenesses to [.uniqueness_floor, 1], so an improper
  # solution is pinned AT that floor and never reaches zero. A gate keyed on `psi <= 0` therefore
  # fires only for a hand-built covariance and never for a fitted one, letting a boundary solution
  # report a hugely inflated Wald standard error. Pin the boundary the fitters can actually reach.
  L <- matrix(c(sqrt(1 - EFAtools:::.uniqueness_floor), 0.6, 0.5, 0.4,
                0.00, 0.05, 0.20, 0.30), 4, 2)
  psi <- 1 - rowSums(L^2)
  expect_lte(min(psi), EFAtools:::.uniqueness_floor + sqrt(.Machine$double.eps))
  expect_gt(min(psi), 0)                       # strictly interior to the old psi <= 0 gate

  expect_warning(
    out <- EFAtools:::.se_information(list(unrot_loadings = L), rot_info = NULL,
                                      N = 250, ci = 0.95, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(out$SE$unrot_loadings)))
  expect_true(all(is.na(out$SE$uniquenesses)))
})


test_that("a Heywood case from an actual fit withholds the analytic SEs", {
  # The companion test above drives `.se_information()` with a hand-built psi, which bypasses the
  # optimiser -- and the optimiser is precisely what makes the boundary reachable only AT the floor
  # rather than below zero. Drive the public path on data that fits an improper solution, so the
  # gate is exercised against a uniqueness the fitter actually produced.
  set.seed(2)
  Lh <- matrix(0, 6, 2)
  Lh[1:3, 1] <- c(.95, .55, .50)
  Lh[4:6, 2] <- c(.90, .50, .45)
  Rh <- tcrossprod(Lh)
  diag(Rh) <- 1
  x <- matrix(stats::rnorm(150 * 6), 150, 6) %*% chol(Rh)
  colnames(x) <- paste0("V", seq_len(6))

  fit <- suppressMessages(suppressWarnings(
    efa_fit(x, n_factors = 2, estimator = "ML", rotation = "none", se = "information")
  ))
  skip_if(length(fit$heywood) == 0L, "this platform's optimiser did not land on the boundary")

  # The uniqueness is pinned AT the floor, strictly above the zero the old gate tested for.
  psi <- 1 - rowSums(unclass(fit$unrot_loadings)^2)
  expect_lte(min(psi), EFAtools:::.uniqueness_floor + sqrt(.Machine$double.eps))
  expect_gt(min(psi), 0)

  expect_true(all(is.na(fit$SE$unrot_loadings)))
  expect_true(all(is.na(fit$SE$uniquenesses)))
  expect_true(anyNA(fit$vcov_unrot_loadings))
})


test_that("a weakly determined rotational gauge withholds only the unrotated loading SEs", {
  # When two ML canonical variances diag(Lambda' Psi^-1 Lambda) coincide, the orientation within
  # that two-plane is arbitrary: the gauge-fixing constraint stops being a transversal to the gauge
  # orbit, the reflexive inverse is amplified by ||(C Z)^-1||, and the unrotated loadings have no
  # well-defined limiting distribution. The covariance stays finite and PSD, so it passes the
  # PSD gate -- without the transversal check the path would report a standard error of order 10 or
  # 100 for a parameter bounded in [-1, 1].
  L <- matrix(c(0.70, 0.60, 0.50, 0.10, 0.00, 0.20,
                0.10, 0.05, 0.20, 0.70, 0.60, 0.50), 6, 2)
  psi <- 1 - rowSums(L^2)
  canon <- diag(crossprod(L, L / psi))
  expect_lt(abs(diff(canon)) / max(canon), 1e-2)       # the two-plane really is degenerate

  Cmat <- EFAtools:::.se_sandwich_constraint(L, psi = psi)
  expect_lt(EFAtools:::.gauge_transversal(L, Cmat), EFAtools:::.gauge_transversal_floor)

  expect_warning(
    out <- EFAtools:::.se_information(list(unrot_loadings = L), rot_info = NULL,
                                      N = 250, ci = 0.95, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(out$SE$unrot_loadings)))
  # The communalities rowSums(Lambda^2) are invariant under Lambda -> Lambda T, so they survive.
  expect_false(anyNA(out$SE$communalities))
  expect_equal(unname(out$SE$uniquenesses), unname(out$SE$communalities))
  # The covariance itself is correct; only its gauge-dependent marginals are unusable.
  expect_false(anyNA(out$vcov_unrot_loadings))
})


test_that("a weakly determined gauge leaves the rotated SEs intact", {
  # The rotation Jacobian annihilates the gauge directions, so the rotated loadings, Phi and the
  # structure coefficients are gauge-invariant and must NOT be withheld along with the unrotated
  # ones -- otherwise the gate would destroy the very quantities that remain trustworthy.
  # Drive the whole public path so the rotation, its Jacobian and the gauge diagnostic all come
  # from the same fitted solution rather than a hand-built rot_info. Two disjoint blocks make
  # Lambda' Psi^-1 Lambda exactly diagonal, so the population solution IS the canonical one; the
  # 0.98 scaling puts the two canonical variances close together (2.371 / 2.203) without making
  # them equal, which is the near-degenerate regime the gate targets. At exactly equal strengths
  # the bordered matrix is exactly singular and the pre-existing singular-information path takes
  # over instead.
  Lp <- matrix(0, 6, 2)
  Lp[1:3, 1] <- c(0.72, 0.65, 0.60)
  Lp[4:6, 2] <- c(0.72, 0.65, 0.60) * 0.98
  R <- tcrossprod(Lp)
  diag(R) <- 1
  colnames(R) <- rownames(R) <- paste0("V", seq_len(6))

  expect_warning(
    fit <- efa_fit(R, n_factors = 2, N = 250, estimator = "ML", rotation = "oblimin",
                   se = "information"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(fit$SE$unrot_loadings)))
  expect_false(anyNA(fit$SE$rot_loadings))
  expect_false(anyNA(fit$SE$communalities))
  expect_false(anyNA(fit$SE$Phi))
  # ... and the surviving rotated SEs are of a plausible magnitude, not gauge-inflated.
  expect_lt(max(fit$SE$rot_loadings), 0.5)
  # The covariance the rotated SEs were propagated from stays intact, so consumers that pool in a
  # common gauge across fits (efa_mi) keep the gauge-invariant quantities.
  expect_false(anyNA(fit$vcov_unrot_loadings))
})


test_that("a well-separated gauge is not flagged and reports finite SEs", {
  # Guard the other direction: the transversal check must not fire on an ordinary solution.
  L <- matrix(c(0.80, 0.75, 0.70, 0.10, 0.00, 0.20,
                0.10, 0.05, 0.20, 0.55, 0.50, 0.45), 6, 2)
  psi <- 1 - rowSums(L^2)
  Cmat <- EFAtools:::.se_sandwich_constraint(L, psi = psi)
  expect_gt(EFAtools:::.gauge_transversal(L, Cmat), EFAtools:::.gauge_transversal_floor)

  expect_no_warning(
    out <- EFAtools:::.se_information(list(unrot_loadings = L), rot_info = NULL,
                                      N = 250, ci = 0.95, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_false(anyNA(out$SE$unrot_loadings))
  expect_lt(max(out$SE$unrot_loadings), 0.5)
})


test_that("the gauge transversal is the exact conditioning of the bordered inverse", {
  # `.gauge_transversal()` restricts the constraint to null(A) = {vec(Lambda S) : S antisymmetric},
  # using that basis in closed form rather than decomposing A. The claim rests on the gauge
  # direction leaving the modelled off-diagonal of Sigma = Lambda Lambda' unchanged, which is
  # checked here directly: the change along Lambda S must be second order in the step, so the
  # information A = Delta' Gamma^-1 Delta annihilates it whatever Gamma is.
  L <- matrix(c(0.70, 0.60, 0.50, 0.10, 0.00, 0.20,
                0.10, 0.05, 0.20, 0.66, 0.58, 0.47), 6, 2)
  psi <- 1 - rowSums(L^2)
  k <- ncol(L)
  nc <- k * (k - 1L) / 2L

  S <- matrix(c(0, -1, 1, 0), 2, 2)
  off <- function(M) { diag(M) <- 0; max(abs(M)) }
  step <- vapply(c(1e-2, 1e-3), function(e) off(tcrossprod(L + e * (L %*% S)) - tcrossprod(L)),
                 numeric(1))
  # Halving the step ten-fold cuts the deviation a hundred-fold: second order, no first-order part.
  expect_equal(step[1] / step[2], 100, tolerance = 1e-3)

  # The value agrees with a basis obtained by decomposing the information matrix instead. Build A
  # through the package's own assembly via the sandwich core at the efficient weight, then read its
  # null space off the smallest `nc` eigenvectors.
  Sigma <- tcrossprod(L); diag(Sigma) <- 1
  prs <- utils::combn(nrow(L), 2L)
  Gamma <- EFAtools:::.normal_theory_gamma(Sigma, prs)
  Delta <- t(vapply(seq_len(ncol(prs)), function(s) {
    G <- matrix(0, nrow(L), k)
    G[prs[1, s], ] <- L[prs[2, s], ]
    G[prs[2, s], ] <- G[prs[2, s], ] + L[prs[1, s], ]
    as.vector(G)
  }, numeric(length(L))))
  A <- crossprod(Delta, solve(Gamma) %*% Delta)
  ev <- eigen((A + t(A)) / 2, symmetric = TRUE)
  Z_eig <- ev$vectors[, seq(ncol(A) - nc + 1L, ncol(A)), drop = FALSE]

  Cmat <- EFAtools:::.se_sandwich_constraint(L, psi = psi)
  rn <- sqrt(rowSums(Cmat^2))
  expect_equal(EFAtools:::.gauge_transversal(L, Cmat),
               min(svd((Cmat / rn) %*% Z_eig)$d), tolerance = 1e-8)

  # A single factor has no rotational freedom, so there is no transversal to condition.
  expect_identical(
    EFAtools:::.gauge_transversal(L[, 1, drop = FALSE],
                                  EFAtools:::.se_sandwich_constraint(L[, 1, drop = FALSE])),
    Inf
  )
})


test_that("unsupported se combinations abort early with a clear class", {
  R <- test_models$baseline$cormat

  expect_error(
    EFA(R, n_factors = 3, N = 500, method = "PAF", rotation = "none", se = "information"),
    class = "efa_se_unsupported"
  )
  expect_error(
    EFA(R, n_factors = 3, N = 500, method = "ML", rotation = "promax", se = "information"),
    class = "efa_se_unsupported"
  )
  # simplimax has a non-smooth (piecewise) criterion, so it has no usable analytic rotation Jacobian.
  expect_error(
    EFA(R, n_factors = 3, N = 500, method = "ML", rotation = "simplimax", se = "information"),
    class = "efa_se_unsupported"
  )
  expect_error(
    EFA(R, n_factors = 3, N = 500, method = "PAF", rotation = "none", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # Sandwich SEs require an ordinal correlation method; a continuous (Pearson) matrix is rejected.
  expect_error(
    EFA(R, n_factors = 3, N = 500, method = "ML", rotation = "none", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  expect_error(
    EFA(R, n_factors = 3, N = 500, method = "ULS", rotation = "none", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # A correlation matrix carries no sample size, which the analytic methods require.
  expect_error(
    EFA(R, n_factors = 3, method = "ML", rotation = "none", se = "information"),
    class = "efa_se_no_n"
  )
})


test_that("information SEs are rejected for non-Pearson correlation methods", {
  # The expected information is built from the normal-theory covariance of Pearson correlations.
  # A polychoric correlation additionally carries first-stage threshold estimation error and a rank
  # correlation is not a Pearson moment at all, so neither may reach this path silently -- the
  # sandwich is the supported route for both.
  x <- DOSPERT_raw[1:300, 1:8]

  for (cm in c("poly", "tetra", "spearman", "kendall")) {
    expect_error(
      EFA(x, n_factors = 2, method = "ML", rotation = "none", se = "information",
          cor_method = cm),
      class = "efa_se_unsupported",
      info = cm
    )
  }

  # Pearson and fiml stay available.
  expect_s3_class(
    EFA(x, n_factors = 2, method = "ML", rotation = "none", se = "information",
        cor_method = "pearson"),
    "EFA"
  )
})


test_that("an analytic-SE fit prints and summarises without error", {
  fit <- .info_fit_unrot()

  expect_no_error(testthat::capture_output(print(fit)))
  expect_no_error(testthat::capture_output(print(summary(fit))))

  # The Wald provenance note (with the CI table) appears in the full summary, but not in the
  # brief print() view nor when intervals are suppressed with ci = "none".
  expect_true(any(grepl("Wald", testthat::capture_output(print(summary(fit))), fixed = TRUE)))
  expect_false(any(grepl("Wald", testthat::capture_output(print(fit)), fixed = TRUE)))
  expect_false(any(grepl("Wald", testthat::capture_output(print(summary(fit, ci = "none"))), fixed = TRUE)))
})


test_that("every component of the analytic SE schema carries the point estimate's dimnames", {
  # The SEs are subset by name as routinely as the estimates they belong to, so an unlabelled
  # component is an inconsistency a reader trips over -- and at k = 6+ leaves them counting columns.
  fit <- .info_fit_oblq()

  expect_identical(dimnames(fit$SE$unrot_loadings), dimnames(unclass(fit$unrot_loadings)))
  expect_identical(dimnames(fit$SE$rot_loadings), dimnames(unclass(fit$rot_loadings)))
  expect_identical(dimnames(fit$SE$Structure), dimnames(unclass(fit$rot_loadings)))
  expect_identical(dimnames(fit$SE$Phi), dimnames(fit$Phi))
  expect_identical(names(fit$SE$uniquenesses), rownames(unclass(fit$unrot_loadings)))
  expect_identical(names(fit$SE$communalities), rownames(unclass(fit$unrot_loadings)))

  # A withheld (all-NA) SE must be labelled the same way, so it stays subsettable by name.
  L <- matrix(c(sqrt(1.11), 0.6, 0.5, 0.4, 0.2, 0.7,
                0.1, 0.05, 0.2, 0.3, 0.15, 0.25), 6, 2,
              dimnames = list(paste0("V", 1:6), c("F1", "F2")))
  Phi <- matrix(c(1, 0.3, 0.3, 1), 2, dimnames = list(c("F1", "F2"), c("F1", "F2")))
  rot_info <- list(rotation = "oblimin", rotmat = diag(2), rot_loadings = L, Phi = Phi,
                   normalize = FALSE, crit_args = list(gam = 0, delta = 0.01))
  expect_warning(
    na_out <- EFAtools:::.se_information(list(unrot_loadings = L), rot_info, N = 200,
                                         ci = 0.95, method = "ML"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(na_out$SE$Phi)))
  expect_identical(dimnames(na_out$SE$Phi), dimnames(Phi))
})


test_that("a boundary solution can still report unrotated SEs when only the rotation failed", {
  # The two-stage FIML sandwich builds its own `se0` and reaches the rotation propagation without
  # passing through the boundary gate, so a Heywood solution can arrive here with a perfectly usable
  # loading covariance and lose only the rotated quantities. The withholding hint must not then be
  # attributed to the boundary: the unrotated loadings and uniquenesses ARE reported.
  L <- matrix(c(sqrt(1 - EFAtools:::.uniqueness_floor), 0.6, 0.5, 0.4,
                0.00, 0.05, 0.20, 0.30), 4, 2,
              dimnames = list(paste0("V", 1:4), c("F1", "F2")))
  expect_true(EFAtools:::.at_uniqueness_boundary(L))

  pk <- length(L)
  se0 <- list(vcov = diag(pk) * 1e-3,
              loadings_se = matrix(0.05, 4, 2),
              uniquenesses_se = rep(0.05, 4))
  # An unusable rotation matrix is what strands the rotated block while the unrotated one stands.
  rot_info <- list(rotation = "oblimin", rotmat = matrix(NA_real_, 2, 2), rot_loadings = L,
                   Phi = diag(2), normalize = FALSE, crit_args = list(gam = 0, delta = 0.01))

  expect_warning(
    out <- EFAtools:::.se_information_rotated(list(unrot_loadings = L), rot_info,
                                              N = 250, ci = 0.95, se0 = se0),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.finite(out$SE$unrot_loadings)))
  expect_true(all(is.finite(out$SE$uniquenesses)))
  expect_true(all(is.finite(out$SE$communalities)))
  expect_true(all(is.na(out$SE$rot_loadings)))
})

# Robust (Godambe sandwich) standard errors and the scaled (Satorra-Bentler) chi-square for the
# ordinal/polychoric path (method ML/ULS/DWLS, cor_method poly/tetra).

# The unrotated two-factor polychoric DWLS fit on the first eight DOSPERT items is the reference
# sandwich fit for the SE/CI schema, the CI-provenance note and the unscaled statistic alike. Its
# cost is the polychoric correlations and their full 28 x 28 ACOV over 3,123 rows, so it is built
# once and shared. The memoisation is lazy because one of the three consumers is skipped on CRAN,
# where building it would be wasted work; the fit takes no random draws, so sharing it cannot
# perturb any RNG stream.
.dwls_sandwich_fit <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      cache <<- EFA(DOSPERT_raw[, 1:8], n_factors = 2, cor_method = "poly", method = "DWLS",
                    rotation = "none", se = "sandwich")
    }
    cache
  }
})

test_that("sandwich SEs fill the unrotated SE/CI schema and a scaled chi-square", {
  fit <- .dwls_sandwich_fit()

  # Unrotated loading and uniqueness SEs are present, finite, and positive.
  expect_equal(dim(fit$SE$unrot_loadings), c(8L, 2L))
  expect_true(all(is.finite(fit$SE$unrot_loadings)))
  expect_true(all(fit$SE$unrot_loadings > 0))
  expect_length(fit$SE$uniquenesses, 8L)
  expect_true(all(is.finite(fit$SE$uniquenesses)))

  # Wald intervals bracket the point estimates.
  ci <- fit$CI$unrot_loadings
  expect_true(all(ci$lower <= unclass(fit$unrot_loadings)))
  expect_true(all(ci$upper >= unclass(fit$unrot_loadings)))

  # The scaled chi-square block is populated (it is NA from .gof() for DWLS).
  fi <- fit$fit_indices
  expect_true(is.finite(fi$chi))
  expect_true(is.finite(fi$p_chi))
  expect_true(all(is.finite(c(fi$CFI, fi$TLI, fi$RMSEA))))
  expect_identical(fi$chi_scaled_type, "scaled.shifted")
  expect_true(all(is.finite(c(fi$chi_scaling, fi$chi_shift, fi$chi_unscaled,
                              fi$chi_mean_adjusted, fi$chi_mean_var, fi$df_mean_var))))
  # AIC/BIC/ECVI are likelihood-ratio chi-square criteria with no meaning for the moment-scaled
  # statistic, so they are deliberately left NA on the sandwich path.
  expect_true(all(is.na(c(fi$AIC, fi$BIC, fi$ECVI))))
})


test_that("the unrotated sandwich CI-provenance note names the robust sandwich, not the information matrix", {
  skip_on_cran()

  fit <- .dwls_sandwich_fit()

  # The full summary carries the CI-provenance note; for a robust (Godambe sandwich) fit it must
  # describe the sandwich covariance, not the expected information matrix (which is correct only
  # for se = "information").
  body <- cli::ansi_strip(format(summary(fit)))
  note <- body[grepl("Note:", body)]
  expect_true(any(grepl("sandwich", note)))
  expect_false(any(grepl("expected information matrix", note)))
})


test_that(".chi_fit_indices is robust to an undefined (NA) baseline chi-square", {
  # A degenerate scaled baseline yields chi_null = NA while the model chi stays finite; the shared
  # fit-index helper must return NA CFI/TLI (not error) and still compute the model-only RMSEA. The
  # scaled path passes its statistics as the noncentrality inputs (chi_cfi / chi_null_cfi).
  idx <- EFAtools:::.chi_fit_indices(chi = 50, df = 10, chi_null = NA_real_,
                                     df_null = 28, N = 500, m = 8, ci = TRUE,
                                     chi_cfi = 50, chi_null_cfi = NA_real_)
  expect_true(is.na(idx$CFI))
  expect_true(is.na(idx$TLI))
  expect_true(is.na(idx$p_null))
  expect_true(is.finite(idx$RMSEA))
  expect_true(is.finite(idx$p_chi))
})


test_that("the unscaled sandwich statistic equals the DWLS objective Fm", {
  fit <- .dwls_sandwich_fit()
  # T = N (s - sigma)' V (s - sigma) with V = 1/diag(Gamma) equals the weighted off-diagonal
  # objective the DWLS backend minimised, because the variance-scale weights absorb the N.
  expect_equal(fit$fit_indices$chi_unscaled, fit$fit_indices$Fm, tolerance = 1e-8)
})


test_that("requesting the sandwich does not change the DWLS point estimate", {
  dat <- DOSPERT_raw[, 1:6]
  base <- EFA(dat, n_factors = 1, cor_method = "poly", method = "DWLS",
              rotation = "none", se = "none")
  rob <- EFA(dat, n_factors = 1, cor_method = "poly", method = "DWLS",
             rotation = "none", se = "sandwich")
  # Both fit on the listwise-complete rows, and the full ACOV's diagonal reproduces the DWLS
  # "diag" weights, so the converged loadings are identical (up to the optimiser tolerance).
  expect_equal(unclass(rob$unrot_loadings), unclass(base$unrot_loadings), tolerance = 1e-6)
})


test_that("sandwich SEs propagate through an oblique rotation", {
  dat <- DOSPERT_raw[, 1:8]
  fit <- EFA(dat, n_factors = 2, cor_method = "poly", method = "DWLS",
             rotation = "oblimin", se = "sandwich")

  expect_equal(dim(fit$SE$rot_loadings), c(8L, 2L))
  expect_true(all(is.finite(fit$SE$rot_loadings)))
  # Factor correlations: a symmetric SE matrix with a fixed (zero-variance) unit diagonal. It
  # carries the factor names, so compare the diagonal on its values alone.
  expect_equal(dim(fit$SE$Phi), c(2L, 2L))
  expect_equal(unname(diag(fit$SE$Phi)), c(0, 0))
  expect_equal(fit$SE$Phi, t(fit$SE$Phi))
  expect_true(is.finite(fit$SE$Phi[1, 2]))
  # Structure coefficients and communalities are reported for the oblique solution.
  expect_equal(dim(fit$SE$Structure), c(8L, 2L))
  expect_length(fit$SE$communalities, 8L)
  # Communalities are rotation-invariant, so their SEs equal the unrotated uniqueness SEs.
  expect_equal(unname(fit$SE$communalities), unname(fit$SE$uniquenesses),
               tolerance = 1e-8)

  # TODO: the ROTATED sandwich SEs are asserted for shape and finiteness only -- no external
  # oracle. `lavaan::efa()` offers no robust rotated standard errors for a polychoric EFA, so the
  # nearest reference would be a bootstrap, which is what the (slow-gated) comparison in
  # "continuous sandwich SEs are in the same ballpark as the bootstrap" does on the unrotated
  # continuous path. The rotation Jacobian these SEs pass through is shared verbatim with
  # `se = "information"` (`.se_information_rotated()`), which IS validated against lavaan's delta
  # method in test-se-information.R, so what stays unvalidated is the covariance handed to it.
})


test_that("sandwich SEs are available for ULS and ML on the ordinal path", {
  dat <- DOSPERT_raw[, 1:8]
  for (m in c("ULS", "ML")) {
    fit <- EFA(dat, n_factors = 2, cor_method = "poly", method = m,
               rotation = "none", se = "sandwich")
    expect_true(all(is.finite(fit$SE$unrot_loadings)),
                info = paste("method", m))
    expect_true(is.finite(fit$fit_indices$chi), info = paste("method", m))
  }
})


test_that("sandwich SEs reject the unsupported method/rotation/correlation combinations", {
  dat <- DOSPERT_raw[, 1:6]
  cormat <- test_models$baseline$cormat

  # Continuous correlations have no polychoric ACOV for the robust meat.
  expect_error(
    EFA(dat, n_factors = 2, cor_method = "pearson", method = "DWLS", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # PAF has no discrepancy-based information for a sandwich.
  expect_error(
    EFA(dat, n_factors = 2, cor_method = "poly", method = "PAF", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # promax and simplimax have no usable analytic rotation Jacobian.
  expect_error(
    EFA(dat, n_factors = 2, cor_method = "poly", method = "DWLS",
        rotation = "promax", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  expect_error(
    EFA(dat, n_factors = 2, cor_method = "poly", method = "DWLS",
        rotation = "simplimax", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # A correlation matrix carries no raw data to estimate the polychoric ACOV from. (DWLS rejects a
  # correlation matrix even earlier, via its own weight precondition, so use ML here to exercise
  # the sandwich-specific raw-data guard.)
  expect_error(
    EFA(cormat, n_factors = 3, N = 500, cor_method = "poly", method = "ML", se = "sandwich"),
    class = "efa_se_unsupported"
  )
})


test_that("single-factor sandwich loading SEs match lavaan robust.sem (DWLS and ULS)", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  dat <- as.data.frame(DOSPERT_raw[, 1:6])
  vn <- colnames(dat)
  mod <- paste0("f =~ ", paste(vn, collapse = " + "))

  for (pair in list(c("DWLS", "DWLS"), c("ULS", "ULS"))) {
    fit <- EFA(dat, n_factors = 1, cor_method = "poly", method = pair[1],
               rotation = "none", se = "sandwich")

    lfit <- lavaan::cfa(mod, data = dat, ordered = vn, estimator = pair[2],
                        se = "robust.sem", test = "scaled.shifted", std.lv = TRUE)
    pe <- lavaan::parameterEstimates(lfit, standardized = TRUE)
    lam <- pe[pe$op == "=~", ]
    L_lav <- lam$std.all; SE_lav <- lam$se

    L_efa <- as.numeric(fit$unrot_loadings)
    sgn <- sign(sum(L_efa * L_lav))

    # One factor: no rotational indeterminacy, so the loadings (up to sign) and their robust SEs
    # are directly comparable. The residual gap is the small polychoric-correlation difference.
    expect_equal(as.numeric(fit$unrot_loadings) * sgn, L_lav, tolerance = 0.01,
                 info = pair[1])
    expect_equal(as.numeric(fit$SE$unrot_loadings), SE_lav, tolerance = 0.01,
                 info = pair[1])
  }
})


test_that("the scaled chi-square matches lavaan WLSMV", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  dat <- as.data.frame(DOSPERT_raw[, 1:8])
  vn <- colnames(dat)

  fit <- EFA(dat, n_factors = 2, cor_method = "poly", method = "DWLS",
             rotation = "none", se = "sandwich")

  lf <- lavaan::efa(dat, nfactors = 2, ordered = vn, estimator = "DWLS",
                    rotation = "geomin", test = "scaled.shifted", se = "robust.sem")
  lfit <- lf[[1]]
  ss <- lavaan::lavInspect(lfit, "test")[["scaled.shifted"]]

  # The scaled-shifted statistic, its shift, and the degrees of freedom match lavaan WLSMV; the
  # chi-square gap is the small polychoric-correlation difference (the shift, a function of the
  # scale-invariant trace coefficients, matches more tightly).
  expect_equal(fit$fit_indices$df, ss$df)
  expect_equal(fit$fit_indices$chi, ss$stat, tolerance = 0.02)
  expect_equal(fit$fit_indices$chi_shift, ss$shift.parameter, tolerance = 1e-3)
  # lavaan stores the scaling factor as the reciprocal of the multiplier a.
  expect_equal(1 / fit$fit_indices$chi_scaling, ss$scaling.factor, tolerance = 1e-3)
})


test_that("the EFAtools polychoric Gamma matches lavaan's NACOV up to the N scale", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  dat <- as.data.frame(DOSPERT_raw[, 1:6])
  vn <- colnames(dat)

  poly <- EFAtools:::.polychoric(as.matrix(dat), nearest_pd = FALSE,
                                 binary_only = FALSE, acov = "full")
  N <- sum(stats::complete.cases(dat))

  lf <- lavaan::efa(dat, nfactors = 1, ordered = vn, estimator = "DWLS", se = "robust.sem")
  lfit <- lf[[1]]
  Glav <- lavaan::lavInspect(lfit, "gamma")
  if (is.list(Glav)) Glav <- Glav[[1]]
  cor_rows <- grep("~~", rownames(Glav))
  Glav_corr <- unname(Glav[cor_rows, cor_rows])

  # EFAtools' Gamma is on the variance scale Var(rho-hat); lavaan's NACOV is N times that.
  rel_F <- norm(N * poly$acov - Glav_corr, "F") / norm(Glav_corr, "F")
  expect_lt(rel_F, 1e-3)
})


# Continuous (Pearson) path: the fourth-moment (ADF; Browne, 1984) covariance of the sample
# correlations is the robust meat for method ML/ULS, giving MLM/MLR-style robust SEs and a scaled
# (Satorra-Bentler) chi-square without an ordinal correlation method.

# Deterministic non-normal continuous fixture: a k-factor model whose common factors and unique
# errors are standardised chi-square(1) draws (mean 0, variance 1, skew ~ 2.8, excess kurtosis
# ~ 12), so the columns are markedly non-normal and the ADF covariance departs clearly from
# normal theory. Each column has approximately unit variance and the population correlations are
# the simple-structure loading products.
.adf_fixture <- function(N, loadings, seed) {
  set.seed(seed)
  p <- nrow(loadings)
  k <- ncol(loadings)
  skew <- function(n) (stats::rchisq(n, df = 1) - 1) / sqrt(2)
  Fmat <- matrix(skew(N * k), N, k)
  E <- matrix(skew(N * p), N, p)
  uniq <- sqrt(pmax(1 - rowSums(loadings^2), 0))
  X <- Fmat %*% t(loadings) + sweep(E, 2L, uniq, "*")
  colnames(X) <- paste0("V", seq_len(p))
  as.data.frame(X)
}


test_that("continuous Pearson sandwich SEs fill the SE/CI schema and a scaled chi-square", {
  dat <- .adf_fixture(N = 600, loadings = matrix(c(.7, .65, .6, .75, .55, .5, .45, .6), ncol = 1L),
                      seed = 1)

  for (m in c("ML", "ULS")) {
    fit <- EFA(dat, n_factors = 1, cor_method = "pearson", method = m,
               rotation = "none", se = "sandwich")

    # Unrotated loading and uniqueness SEs are present, finite, and positive.
    expect_true(all(is.finite(fit$SE$unrot_loadings)), info = m)
    expect_true(all(fit$SE$unrot_loadings > 0), info = m)
    expect_true(all(is.finite(fit$SE$uniquenesses)), info = m)

    # Wald intervals bracket the point estimates.
    ci <- fit$CI$unrot_loadings
    expect_true(all(ci$lower <= unclass(fit$unrot_loadings)), info = m)
    expect_true(all(ci$upper >= unclass(fit$unrot_loadings)), info = m)

    # The scaled chi-square block is populated (the unscaled ML/ULS discrepancy is not robust to
    # non-normality).
    fi <- fit$fit_indices
    expect_true(is.finite(fi$chi), info = m)
    expect_true(is.finite(fi$p_chi), info = m)
    expect_true(all(is.finite(c(fi$CFI, fi$TLI, fi$RMSEA))), info = m)
    expect_identical(fi$chi_scaled_type, "scaled.shifted")
    expect_true(all(is.na(c(fi$AIC, fi$BIC, fi$ECVI))), info = m)
  }
})


test_that("continuous Pearson sandwich SEs propagate through an oblique rotation", {
  L <- matrix(0, 8L, 2L)
  L[1:4, 1] <- c(.7, .65, .6, .55)
  L[5:8, 2] <- c(.7, .6, .65, .5)
  dat <- .adf_fixture(N = 700, loadings = L, seed = 2)

  fit <- EFA(dat, n_factors = 2, cor_method = "pearson", method = "ML",
             rotation = "oblimin", se = "sandwich")

  expect_equal(dim(fit$SE$rot_loadings), c(8L, 2L))
  expect_true(all(is.finite(fit$SE$rot_loadings)))
  expect_equal(dim(fit$SE$Phi), c(2L, 2L))
  expect_equal(unname(diag(fit$SE$Phi)), c(0, 0))
  expect_equal(fit$SE$Phi, t(fit$SE$Phi))
  expect_true(is.finite(fit$SE$Phi[1, 2]))
  expect_equal(dim(fit$SE$Structure), c(8L, 2L))
  expect_length(fit$SE$communalities, 8L)
  # Communalities are rotation-invariant, so their SEs equal the unrotated uniqueness SEs.
  expect_equal(unname(fit$SE$communalities), unname(fit$SE$uniquenesses),
               tolerance = 1e-8)
})


test_that("continuous Pearson sandwich rejects the unsupported method/rotation/data combinations", {
  dat <- .adf_fixture(N = 400, loadings = matrix(c(.7, .6, .65, .55, .5, .6), ncol = 1L), seed = 3)
  cormat <- test_models$baseline$cormat

  # DWLS has no continuous asymptotic covariance (it is an ordinal estimator).
  expect_error(
    EFA(dat, n_factors = 1, cor_method = "pearson", method = "DWLS", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # promax and simplimax have no usable analytic rotation Jacobian.
  expect_error(
    EFA(dat, n_factors = 2, cor_method = "pearson", method = "ML",
        rotation = "promax", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  expect_error(
    EFA(dat, n_factors = 2, cor_method = "pearson", method = "ML",
        rotation = "simplimax", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # Spearman/Kendall correlations have no fourth-moment covariance in this implementation.
  expect_error(
    EFA(dat, n_factors = 1, cor_method = "spearman", method = "ML", se = "sandwich"),
    class = "efa_se_unsupported"
  )
  # A correlation matrix carries no raw data to estimate the ADF covariance from.
  expect_error(
    EFA(cormat, n_factors = 3, N = 500, cor_method = "pearson", method = "ML", se = "sandwich"),
    class = "efa_se_unsupported"
  )
})


test_that(".prepare_cor_input rejects a full ADF covariance for non-Pearson correlations", {
  dat <- .adf_fixture(N = 200, loadings = matrix(c(.7, .6, .5, .55), ncol = 1L), seed = 7)
  # The fourth-moment ADF covariance is Pearson-specific; a rank correlation with acov = "full"
  # would mismatch R and Gamma. EFA() gates this earlier, but the helper guards its own contract.
  expect_error(
    EFAtools:::.prepare_cor_input(as.matrix(dat), cor_method = "spearman", acov = "full"),
    class = "efa_acov_unsupported"
  )
})


test_that("the continuous ADF Gamma matches lavaan's correlation NACOV up to the N scale", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  skip_if(utils::packageVersion("lavaan") < "0.6.9")

  L <- matrix(c(.7, .65, .6, .75, .55, .5), ncol = 1L)
  dat <- .adf_fixture(N = 600, loadings = L, seed = 4)
  vn <- colnames(dat)
  N <- nrow(dat)

  Gamma <- EFAtools:::.adf_gamma(as.matrix(dat))

  # correlation = TRUE reparameterises lavaan onto the correlation structure: its NACOV omits the
  # variance (Vi~~Vi) rows and orders the off-diagonal ~~ rows exactly as utils::combn(p, 2).
  mod <- paste0("f =~ ", paste(vn, collapse = " + "))
  lfit <- lavaan::cfa(mod, data = dat, estimator = "MLM", std.lv = TRUE,
                      correlation = TRUE, meanstructure = TRUE)
  Glav <- lavaan::lavInspect(lfit, "gamma")
  if (is.list(Glav)) Glav <- Glav[[1]]
  # correlation = TRUE drops the variance (Vi~~Vi) rows, leaving exactly the p(p-1)/2 off-diagonal
  # correlation rows in utils::combn(p, 2) order. `correlation` is consumed via ... (not a formal
  # cfa() argument), so a lavaan build that silently ignored it keeps the variance rows (a
  # covariance-structure NACOV that is not comparable) -- detect that and skip rather than fail on
  # a feature-version guard that does not track when the option landed.
  sides <- strsplit(rownames(Glav), "~~")
  is_var <- vapply(sides, function(s) length(s) == 2L && s[1] == s[2], logical(1))
  skip_if(any(is_var), "this lavaan build did not apply correlation = TRUE")
  cor_rows <- which(vapply(sides, function(s) length(s) == 2L && s[1] != s[2], logical(1)))
  Glav_corr <- unname(Glav[cor_rows, cor_rows])

  # EFAtools' Gamma is on the variance scale Var(rho-hat); lavaan's NACOV is N times that.
  rel_F <- norm(N * unname(Gamma) - Glav_corr, "F") / norm(Glav_corr, "F")
  expect_lt(rel_F, 1e-4)
})


test_that("single-factor continuous sandwich loading SEs and scaled chi match lavaan robust", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  skip_if(utils::packageVersion("lavaan") < "0.6.9")

  L <- matrix(c(.7, .65, .6, .75, .55, .5), ncol = 1L)
  dat <- .adf_fixture(N = 700, loadings = L, seed = 5)
  vn <- colnames(dat)
  mod <- paste0("f =~ ", paste(vn, collapse = " + "))

  # Compared against ML, the unambiguous oracle: the MLMV shortcut (= ML + scaled.shifted test +
  # robust.sem SEs) routes through lavaan's working correlation-structure path, whereas the manual
  # se = "robust.sem" override errors there. ULS has no such lavaan oracle (its continuous robust
  # correlation path is unsupported), so it is cross-checked against the bootstrap below instead.
  fit <- EFA(dat, n_factors = 1, cor_method = "pearson", method = "ML",
             rotation = "none", se = "sandwich")

  lfit <- lavaan::cfa(mod, data = dat, estimator = "MLMV", std.lv = TRUE,
                      correlation = TRUE, meanstructure = TRUE)
  pe <- lavaan::parameterEstimates(lfit)
  lam <- pe[pe$op == "=~", ]
  L_lav <- lam$est; SE_lav <- lam$se

  L_efa <- as.numeric(fit$unrot_loadings)
  sgn <- sign(sum(L_efa * L_lav))

  # One factor: no rotational indeterminacy, and the Pearson correlations are identical to
  # lavaan's, so the loadings (up to sign) and their robust SEs match closely. The SEs differ only
  # by the N vs N-1 convention (a uniform ~sqrt(N/(N-1)) ratio), hence the tight 2e-3 band.
  expect_equal(L_efa * sgn, L_lav, tolerance = 0.01)
  expect_equal(as.numeric(fit$SE$unrot_loadings), SE_lav, tolerance = 2e-3)

  ss <- lavaan::lavInspect(lfit, "test")[["scaled.shifted"]]
  expect_equal(fit$fit_indices$df, ss$df)
  # The chi-square carries a small (~1%) gap: EFAtools' statistic is the quadratic-form
  # approximation N (s - sigma)' V (s - sigma), lavaan's is the exact ML likelihood ratio.
  expect_equal(fit$fit_indices$chi, ss$stat, tolerance = 0.02)
  # The shift is invariant to the (arbitrary) scale of the ML weight matrix, so it matches lavaan
  # tightly; the scaling factor a = sqrt(df / c2) is not (EFAtools uses the 1/2 normal-theory ML
  # weight, lavaan a different normalisation), but it cancels in the statistic.
  expect_equal(fit$fit_indices$chi_shift, ss$shift.parameter, tolerance = 1e-3)
})


test_that("continuous sandwich SEs are in the same ballpark as the bootstrap (ML and ULS)", {
  skip_on_cran()

  L <- matrix(c(.7, .65, .6, .75, .55, .5), ncol = 1L)
  dat <- .adf_fixture(N = 500, loadings = L, seed = 6)

  for (m in c("ML", "ULS")) {
    rob <- EFA(dat, n_factors = 1, cor_method = "pearson", method = m,
               rotation = "none", se = "sandwich")
    boot <- EFA(dat, n_factors = 1, cor_method = "pearson", method = m,
                rotation = "none", se = "np-boot", b_boot = 50, seed = 123)

    se_rob <- as.numeric(rob$SE$unrot_loadings)
    se_boot <- as.numeric(boot$SE$unrot_loadings)

    # Robust and bootstrap SEs estimate the same sampling variability (observed median ratio ~1.02
    # for both ML and ULS). This is a coarse cross-check robust to the b_boot = 50 Monte-Carlo
    # noise -- the precise validation is the lavaan oracle above (ML); the band still catches a
    # gross (>~40%) systematic miscalibration of the robust SE, which is the only numeric guard for
    # ULS (lavaan's continuous ULS robust correlation path is unsupported).
    expect_true(all(se_rob > 0 & se_boot > 0), info = m)
    expect_gt(stats::median(se_rob / se_boot), 0.7)
    expect_lt(stats::median(se_rob / se_boot), 1.4)
  }
})


# Unrotated robust loading SEs are gauge-dependent: they must be scaled in the orientation the
# estimator reports its loadings in (Lambda'Lambda diagonal for the eigen-based ULS/DWLS, Lambda'
# Psi^-1 Lambda diagonal for ML; Lawley & Maxwell). Independent reference: the Godambe sandwich with
# the gauge-constraint Jacobian taken by CENTRAL FINITE DIFFERENCES of the orientation function,
# sharing no code with the package's analytic constraint (.se_sandwich_constraint).
.ref_sandwich_loading_se <- function(L, Gamma, N, method, gauge) {
  p <- nrow(L); k <- ncol(L); pk <- p * k; G <- N * Gamma
  pr <- utils::combn(p, 2L); pii <- pr[1, ]; pjj <- pr[2, ]; n <- ncol(pr)
  Delta <- matrix(0, n, pk)
  for (f in seq_len(k)) {
    Delta[cbind(seq_len(n), (f - 1L) * p + pii)] <- L[pjj, f]
    Delta[cbind(seq_len(n), (f - 1L) * p + pjj)] <- L[pii, f]
  }
  if (method == "ULS") {
    VD <- Delta
  } else {
    Sig <- tcrossprod(L); diag(Sig) <- 1; P <- solve(Sig)
    Vmat <- matrix(0, n, n)
    for (s in seq_len(n)) {
      a <- pii[s]; b <- pjj[s]
      Vmat[, s] <- 0.5 * (P[pii, a] * P[pjj, b] + P[pii, b] * P[pjj, a])
    }
    VD <- Vmat %*% Delta
  }
  A <- crossprod(Delta, VD)
  gfun <- function(par) {
    Lm <- matrix(par, p, k)
    M <- if (gauge == "LtL") crossprod(Lm) else {
      ps <- 1 - rowSums(Lm^2); crossprod(Lm, Lm / ps)
    }
    M[upper.tri(M)]
  }
  nc <- k * (k - 1L) / 2L; Cmat <- matrix(0, nc, pk); h <- 1e-6; par0 <- as.vector(L)
  for (j in seq_len(pk)) {
    pp <- par0; pp[j] <- pp[j] + h; pm <- par0; pm[j] <- pm[j] - h
    Cmat[, j] <- (gfun(pp) - gfun(pm)) / (2 * h)
  }
  Aug <- rbind(cbind(A, t(Cmat)), cbind(Cmat, matrix(0, nc, nc)))
  Ab <- solve(Aug)[seq_len(pk), seq_len(pk)]
  V_AA <- (Ab %*% crossprod(VD, G %*% VD) %*% Ab) / (N - 1)
  matrix(sqrt(diag(V_AA)), p, k)
}


test_that("the unrotated sandwich loading SEs match the estimator's reporting gauge", {
  # A two-factor structure with cross-loadings, so the Lambda'Lambda and Lambda'Psi^-1 Lambda
  # orientations differ noticeably (and a wrong gauge would be caught).
  set.seed(2024)
  Lp <- matrix(c(.75, .70, .65, .20, .10, .15,
                 .15, .10, .20, .75, .70, .65), 6, 2)
  Sig <- Lp %*% t(Lp); diag(Sig) <- 1
  X <- matrix(stats::rnorm(800 * 6), 800) %*% chol(Sig)
  colnames(X) <- paste0("v", seq_len(6))
  N <- nrow(X)

  # ML reports loadings in the Lambda'Psi^-1 Lambda orientation, so its robust unrotated loading
  # SEs must be scaled in that gauge -- and are materially different from the Lambda'Lambda gauge.
  fit_ml <- EFA(X, n_factors = 2, cor_method = "pearson", method = "ML",
                rotation = "none", se = "sandwich")
  L_ml <- unclass(fit_ml$unrot_loadings)
  ref_ml_correct <- .ref_sandwich_loading_se(L_ml, fit_ml$Gamma, N, "ML", "LtPiL")
  ref_ml_wrong   <- .ref_sandwich_loading_se(L_ml, fit_ml$Gamma, N, "ML", "LtL")
  expect_equal(unclass(fit_ml$SE$unrot_loadings), ref_ml_correct,
               tolerance = 1e-5, ignore_attr = TRUE)
  expect_gt(max(abs(ref_ml_correct - ref_ml_wrong)) / mean(ref_ml_correct), 0.05)

  # ULS reports loadings in the (eigen-based) Lambda'Lambda orientation, so its robust unrotated
  # loading SEs stay in that gauge.
  fit_uls <- EFA(X, n_factors = 2, cor_method = "pearson", method = "ULS",
                 rotation = "none", se = "sandwich")
  L_uls <- unclass(fit_uls$unrot_loadings)
  ref_uls_correct <- .ref_sandwich_loading_se(L_uls, fit_uls$Gamma, N, "ULS", "LtL")
  expect_equal(unclass(fit_uls$SE$unrot_loadings), ref_uls_correct,
               tolerance = 1e-5, ignore_attr = TRUE)
})


test_that(".is_psd accepts PSD covariances and rejects non-PSD ones with a non-negative diagonal", {
  expect_true(EFAtools:::.is_psd(diag(3)))
  expect_true(EFAtools:::.is_psd(matrix(c(2, 1, 1, 2), 2)))
  # Positive diagonal with a round-off-level negative eigenvalue (eigenvalues ~ 2 and -1e-9): inside
  # the -1e-8 tolerance, so accepted. Exercises the eigenvalue gate, not just the diagonal.
  expect_true(EFAtools:::.is_psd(matrix(c(1, 1 + 1e-9, 1 + 1e-9, 1), 2)))
  # All diagonal entries are positive, but the matrix is indefinite (eigenvalues -1 and 3): the
  # diagonal-only gate would wrongly accept it.
  expect_false(EFAtools:::.is_psd(matrix(c(1, 2, 2, 1), 2)))
  expect_false(EFAtools:::.is_psd(matrix(c(1, NA, NA, 1), 2)))
  expect_false(EFAtools:::.is_psd(matrix(c(-1, 0, 0, 1), 2)))
})


test_that("the sandwich core degrades gracefully when the gauge is undefined or undetermined", {
  # Regression: an NA loading or a Heywood uniqueness (psi <= 0) makes the Lambda'Psi^-1 Lambda
  # orientation undefined. Gauge detection must rule it out and fall back to the Lambda'Lambda gauge
  # without erroring on the comparison, and must not abort the (gauge-invariant) rest of the
  # computation.
  G <- diag(3L)   # p = 3 -> p (p - 1) / 2 = 3 off-diagonal pairs

  # An NA loading propagates to psi and to the off-diagonal comparison; detection must coerce the
  # undefined gauge to the Lambda'Lambda fallback rather than raise "missing value where TRUE/FALSE
  # needed".
  fo_na <- list(unrot_loadings = matrix(c(0.7, NA, 0.6, 0.5, 0.4, 0.3), 3, 2),
                orig_R = diag(3), fit_indices = list(df = 1))
  expect_no_error(out_na <- EFAtools:::.se_sandwich_core(fo_na, N = 200, Gamma = G, method = "ML"))
  expect_false(out_na$reliable)
  expect_true(all(is.na(out_na$loadings_se)))

  # A communality above one drives psi[1] < 0; the core must not error.
  fo_hey <- list(unrot_loadings = matrix(c(0.95, 0.6, 0.5, 0.4, 0.3, 0.2), 3, 2),
                 orig_R = diag(3), fit_indices = list(df = 1))
  expect_no_error(EFAtools:::.se_sandwich_core(fo_hey, N = 200, Gamma = G, method = "ML"))

  # Homogeneous uniquenesses (Psi proportional to I) make BOTH orientations diagonal, so the gauge is
  # undetermined by the loadings and the detection takes the method tie-break branch. Build such a
  # solution -- an equal-norm simple structure rotated to introduce cross-loadings -- and confirm the
  # tie-break runs for either estimator without error (this construction is numerically degenerate at
  # the bordering step, so both return reliable = FALSE rather than a value).
  L0 <- matrix(0, 6L, 2L); L0[1:3, 1] <- 0.6; L0[4:6, 2] <- 0.6
  Q <- matrix(c(cos(pi / 6), -sin(pi / 6), sin(pi / 6), cos(pi / 6)), 2L)
  Sig_tie <- tcrossprod(L0 %*% Q); diag(Sig_tie) <- 1
  fo_tie <- list(unrot_loadings = L0 %*% Q, orig_R = Sig_tie, fit_indices = list(df = 4))
  G_tie <- diag(15L)   # p = 6 -> 15 off-diagonal pairs
  expect_no_error(EFAtools:::.se_sandwich_core(fo_tie, N = 300, Gamma = G_tie, method = "ML"))
  expect_no_error(EFAtools:::.se_sandwich_core(fo_tie, N = 300, Gamma = G_tie, method = "ULS"))
})


# A two-factor solution with variable 1's uniqueness pinned at the fitter's floor: the boundary an
# ML/ULS optimiser can actually reach, and strictly interior to the sandwich core's own `psi <= 0`
# gauge test. Shared by the two tests below so both exercise the same solution.
.boundary_loadings <- function() {
  L <- matrix(0, 6, 2)
  L[, 1] <- c(sqrt(1 - EFAtools:::.uniqueness_floor), 0.6, 0.5, 0.1, 0.15, 0.2)
  L[, 2] <- c(0.00, 0.05, 0.20, 0.70, 0.60, 0.55)
  L
}

test_that("a non-finite solution is withheld but is not described as a Heywood case", {
  # The gate the analytic paths withhold on fires on two quite different situations: a uniqueness
  # pinned at the fitter's floor, and a solution that is not finite at all. Both must withhold --
  # a Wald interval is invalid either way -- but only the first is a Heywood case, so only the
  # first may be described by the boundary hint. Asserting the two predicates directly pins that
  # split without pinning any wording.
  L_boundary <- .boundary_loadings()
  L_na <- L_boundary
  L_na[1, ] <- c(0.7, NA)
  L_interior <- L_boundary
  L_interior[1, 1] <- 0.7

  # withholding: the boundary and the non-finite solution alike
  expect_true(EFAtools:::.at_uniqueness_boundary(L_boundary))
  expect_true(EFAtools:::.at_uniqueness_boundary(L_na))
  expect_false(EFAtools:::.at_uniqueness_boundary(L_interior))

  # describing it as a Heywood case: the boundary only
  expect_true(EFAtools:::.se_boundary_hint_applies(L_boundary))
  expect_false(EFAtools:::.se_boundary_hint_applies(L_na))
  expect_false(EFAtools:::.se_boundary_hint_applies(L_interior))
})

test_that("the sandwich withholds standard errors at a uniqueness pinned at the fitter's floor", {
  # The robust path reports the same Wald quantities as the expected-information path, from a
  # different covariance, so it must withhold on the same boundary test: on the parameter-space
  # boundary the Wald approximation fails for every parameter regardless of how the covariance was
  # estimated.
  L <- .boundary_loadings()
  psi <- 1 - rowSums(L^2)
  expect_lte(min(psi), EFAtools:::.uniqueness_floor + sqrt(.Machine$double.eps))
  expect_gt(min(psi), 0)

  Sigma <- tcrossprod(L)
  diag(Sigma) <- 1
  Gamma <- EFAtools:::.normal_theory_gamma(Sigma, utils::combn(6L, 2L)) / 300
  fit_out <- list(unrot_loadings = L, orig_R = Sigma, fit_indices = list(df = 4))

  # Without the gate this covariance is perfectly usable, so what withholds the SEs below is the
  # boundary test and not a degenerate covariance.
  core <- EFAtools:::.se_sandwich_core(fit_out, N = 300, Gamma = Gamma, method = "ULS")
  expect_true(core$reliable)
  expect_true(all(is.finite(core$loadings_se)))

  expect_warning(
    out <- EFAtools:::.se_sandwich_dispatch(fit_out, rot_info = NULL, N = 300, ci = 0.95,
                                            Gamma = Gamma, method = "ULS"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(out$SE$unrot_loadings)))
  expect_true(all(is.na(out$SE$uniquenesses)))
  expect_true(all(is.na(out$SE$communalities)))
  expect_true(all(is.na(out$vcov_unrot_loadings)))
  expect_true(all(is.na(unlist(out$CI))))

  # The scaled chi-square is a discrepancy-function quantity, not a Wald one, so the boundary does
  # not invalidate it and it is deliberately kept when the standard errors are withheld.
  expect_false(is.null(out$scaled_test))
  expect_true(is.finite(out$scaled_test$chi))

  # Under a rotation the withholding has to reach the rotated quantities as well: they are
  # propagated from the same loading covariance, which the gate leaves unusable.
  rot_info <- list(rotation = "oblimin", rotmat = diag(2), rot_loadings = L,
                   Phi = diag(2), normalize = FALSE, crit_args = list(gam = 0, delta = 0.01))
  expect_warning(
    rot <- EFAtools:::.se_sandwich_dispatch(fit_out, rot_info, N = 300, ci = 0.95,
                                            Gamma = Gamma, method = "ULS"),
    class = "efa_se_unreliable"
  )
  expect_true(all(is.na(rot$SE$rot_loadings)))
  expect_true(all(is.na(rot$SE$Phi)))
  expect_true(all(is.na(rot$SE$Structure)))
  expect_true(all(is.na(rot$SE$communalities)))
  expect_true(all(is.na(rot$vcov_unrot_loadings)))
  expect_false(is.null(rot$scaled_test))
})


test_that("a Heywood case from an actual fit withholds the sandwich SEs but keeps the fit indices", {
  # The companion test above drives the dispatcher with a hand-built psi. Drive the public path on
  # data that fits an improper solution, so the gate is exercised against a uniqueness the fitter
  # actually produced -- the case a user meets, where the sandwich previously reported a standard
  # error of 0.42 for a uniqueness of 0.005.
  set.seed(2)
  Lh <- matrix(0, 6, 2)
  Lh[1:3, 1] <- c(.95, .55, .50)
  Lh[4:6, 2] <- c(.90, .50, .45)
  Rh <- tcrossprod(Lh)
  diag(Rh) <- 1
  x <- matrix(stats::rnorm(150 * 6), 150, 6) %*% chol(Rh)
  colnames(x) <- paste0("V", seq_len(6))

  fit <- suppressMessages(suppressWarnings(
    efa_fit(x, n_factors = 2, estimator = "ML", rotation = "none", se = "sandwich",
            cor_method = "pearson")
  ))
  skip_if(length(fit$heywood) == 0L, "this platform's optimiser did not land on the boundary")

  psi <- 1 - rowSums(unclass(fit$unrot_loadings)^2)
  expect_lte(min(psi), EFAtools:::.uniqueness_floor + sqrt(.Machine$double.eps))

  expect_true(all(is.na(fit$SE$unrot_loadings)))
  expect_true(all(is.na(fit$SE$uniquenesses)))
  expect_true(all(is.na(fit$SE$communalities)))
  expect_true(anyNA(fit$vcov_unrot_loadings))

  # The scaled statistic and everything derived from it survive the boundary: for a DWLS fit it is
  # the only chi-square block there is, so withholding it would take the fit indices down with the
  # standard errors.
  expect_identical(fit$fit_indices$chi_scaled_type, "scaled.shifted")
  expect_true(all(is.finite(c(fit$fit_indices$chi, fit$fit_indices$CFI,
                              fit$fit_indices$TLI, fit$fit_indices$RMSEA))))
})

# Tests for the persisted unrotated loading vcov and correlation ACOV slots on EFA() results.
# Contracts pinned here:
#   * `vcov_unrot_loadings`: pk x pk numeric matrix; the FULL unrotated loading covariance the
#     marginal `SE$unrot_loadings` were derived from. Populated for se = "information" and
#     se = "sandwich" (in both unrotated and rotated paths), NA-filled at a Heywood/unreliable
#     fit, NULL for se = "none" and se = "np-boot".
#   * `Gamma`: n x n asymptotic covariance of the off-diagonal correlations on the
#     variance scale (n = p(p-1)/2, ordered by utils::combn(p, 2)). Populated for
#     se = "sandwich" only, NULL otherwise.
# Downstream MI pooling depends on these slots; the assertions below are the storage contract.

# Self-contained continuous fixture (skewed simple-structure factor model, ~unit variance per
# column). Independent of test-se-sandwich.R so the file can run in isolation.
make_adf_fixture <- function(N, loadings, seed) {
  set.seed(seed)
  p <- nrow(loadings)
  k <- ncol(loadings)
  skew <- function(n) (stats::rchisq(n, df = 1) - 1) / sqrt(2)
  Fmat <- matrix(skew(N * k), N, k)
  E <- matrix(skew(N * p), N, p)
  uniq <- sqrt(pmax(1 - rowSums(loadings^2), 0))
  X <- Fmat %*% t(loadings) + sweep(E, 2L, uniq, "*")
  colnames(X) <- paste0("V", seq_len(p))
  as.data.frame(X)
}

# The single-factor polychoric ULS sandwich fit on the first eight DOSPERT items backs both the
# vcov and the Gamma storage contract, and the polychoric correlations plus their full ACOV over
# 3,123 rows are the expensive part, so it is computed once and shared. Both consumers are skipped
# on CRAN, hence the lazy memoisation: nothing is built there. The fit takes no random draws, so
# sharing it cannot perturb any RNG stream.
# The off-diagonal correlation rows of a lavaan ADF/NACOV matrix, in utils::combn() order,
# which is what both Gamma comparisons below need. lavaan indexes that matrix by parameter
# name, and a bare grep("~~") would also take the variance rows Vi~~Vi, so the rows are
# selected by splitting the name rather than by regex. Variance rows being present at all
# means the fit was parameterised on the covariance scale -- for the continuous comparison
# that happens when a lavaan build ignores `correlation = TRUE`, which reaches cfa() through
# ... rather than as a formal argument. Such a NACOV is not comparable, so it is reported as
# a skip rather than as a failure, and without leaning on a version guard that does not track
# when the option landed.
.lav_cor_rows <- function(Glav) {
  kind <- vapply(strsplit(rownames(Glav), "~~"), function(s) {
    if (length(s) != 2L) "other" else if (identical(s[1], s[2])) "var" else "cov"
  }, character(1))
  testthat::skip_if(any(kind == "var"),
                    "this lavaan build did not apply the correlation parameterization")
  which(kind == "cov")
}

.poly_sandwich_fit <- local({
  cache <- NULL
  function() {
    if (is.null(cache)) {
      cache <<- EFA(DOSPERT_raw[, 1:8], n_factors = 1, cor_method = "poly", method = "ULS",
                    rotation = "none", se = "sandwich")
    }
    cache
  }
})


test_that("se = 'information' persists the full unrotated loading vcov (unrotated path)", {
  cormat <- test_models$baseline$cormat
  p <- nrow(cormat); k <- 3L; pk <- p * k

  fit <- EFA(cormat, n_factors = k, N = 500, method = "ML", rotation = "none",
             se = "information")

  expect_true("vcov_unrot_loadings" %in% names(fit))
  V <- fit$vcov_unrot_loadings
  expect_true(is.matrix(V) && is.numeric(V))
  expect_equal(dim(V), c(pk, pk))
  expect_true(isSymmetric(V, tol = 1e-10))

  # The persisted vcov must be the EXACT matrix that produced the marginal SEs:
  # sqrt(diag(V)) reshaped column-major == SE$unrot_loadings element-wise.
  diag_se <- sqrt(pmax(diag(V), 0))
  expect_equal(matrix(diag_se, p, k), unname(fit$SE$unrot_loadings), tolerance = 1e-12)
})


test_that("se = 'information' persists the unrotated loading vcov under rotation", {
  cormat <- test_models$baseline$cormat
  p <- nrow(cormat); k <- 3L; pk <- p * k

  fit <- EFA(cormat, n_factors = k, N = 500, method = "ML", rotation = "oblimin",
             se = "information")

  expect_true("vcov_unrot_loadings" %in% names(fit))
  V <- fit$vcov_unrot_loadings
  expect_equal(dim(V), c(pk, pk))
  expect_true(isSymmetric(V, tol = 1e-10))

  # Still the UNROTATED block: its diagonal must match the unrotated loading SEs, NOT the
  # rotated ones. This is what U2/M2 will pool across imputations.
  diag_se <- sqrt(pmax(diag(V), 0))
  expect_equal(matrix(diag_se, p, k), unname(fit$SE$unrot_loadings), tolerance = 1e-12)
})


test_that("se = 'sandwich' persists the unrotated robust loading vcov on the polychoric path", {
  skip_on_cran()

  dat <- DOSPERT_raw[, 1:8]
  p <- ncol(dat); k <- 1L; pk <- p * k

  fit <- .poly_sandwich_fit()

  expect_true("vcov_unrot_loadings" %in% names(fit))
  V <- fit$vcov_unrot_loadings
  expect_equal(dim(V), c(pk, pk))
  expect_true(isSymmetric(V, tol = 1e-10))

  diag_se <- sqrt(pmax(diag(V), 0))
  expect_equal(matrix(diag_se, p, k), unname(fit$SE$unrot_loadings), tolerance = 1e-12)
})


test_that("se = 'sandwich' persists the unrotated robust loading vcov on the continuous path", {
  dat <- make_adf_fixture(N = 500,
                          loadings = matrix(c(.7, .65, .6, .75, .55, .5), ncol = 1L),
                          seed = 11)
  p <- ncol(dat); k <- 1L; pk <- p * k

  fit <- EFA(dat, n_factors = k, cor_method = "pearson", method = "ML", rotation = "none",
             se = "sandwich")

  V <- fit$vcov_unrot_loadings
  expect_equal(dim(V), c(pk, pk))
  expect_true(isSymmetric(V, tol = 1e-10))

  diag_se <- sqrt(pmax(diag(V), 0))
  expect_equal(matrix(diag_se, p, k), unname(fit$SE$unrot_loadings), tolerance = 1e-12)
})


test_that("se = 'sandwich' under rotation also persists the unrotated block", {
  L <- matrix(0, 8, 2)
  L[1:4, 1] <- c(.7, .65, .6, .55)
  L[5:8, 2] <- c(.7, .6, .65, .5)
  dat <- make_adf_fixture(N = 700, loadings = L, seed = 12)
  p <- ncol(dat); k <- 2L; pk <- p * k

  fit <- EFA(dat, n_factors = k, cor_method = "pearson", method = "ML",
             rotation = "oblimin", se = "sandwich")

  V <- fit$vcov_unrot_loadings
  expect_equal(dim(V), c(pk, pk))
  expect_true(isSymmetric(V, tol = 1e-10))
  diag_se <- sqrt(pmax(diag(V), 0))
  expect_equal(matrix(diag_se, p, k), unname(fit$SE$unrot_loadings), tolerance = 1e-12)
})


test_that("an unreliable information fit yields an NA-filled vcov of the correct shape, not NULL", {
  # Drive `.se_information()` straight into its Heywood branch with a synthetic loading whose
  # first communality exceeds one (psi[1] < 0), where the solution sits on the parameter-space
  # boundary and the Wald intervals this path reports are not valid. Going through EFA() is not a
  # reliable route here because the optimiser pulls the boundary off zero (e.g. psi[1] ~ 0.005
  # at the .999-loading fixture), so the SE path completes with finite numbers; the storage
  # contract under genuine unreliability is the one we need to pin.
  L <- matrix(c(sqrt(1.11), 0.6, 0.5, 0.4), 4, 1)   # h2[1] = 1.11 > 1, so psi[1] < 0
  fit_out <- list(unrot_loadings = L)
  p <- nrow(L); k <- ncol(L); pk <- p * k

  out <- suppressWarnings(
    EFAtools:::.se_information(fit_out, rot_info = NULL, N = 200, ci = 0.95, method = "ML"))

  expect_true("vcov_unrot_loadings" %in% names(out))
  V <- out$vcov_unrot_loadings
  # The unreliability gate must surface as a present, correctly shaped NA matrix so
  # pooling code can fail closed on anyNA() rather than on a missing slot.
  expect_false(is.null(V))
  expect_true(is.matrix(V))
  expect_equal(dim(V), c(pk, pk))
  expect_true(all(is.na(V)))
})


test_that("an unreliable sandwich fit yields an NA-filled V_AA of the correct shape, not NULL", {
  # `.se_sandwich_core()` returns `na_core` (NA-filled V_AA, reliable = FALSE) whenever
  # Gamma is missing, the wrong shape, or carries an NA. Force that path with an
  # invalid Gamma; the wrapper must still attach a correctly shaped NA matrix on the
  # persisted slot.
  L <- matrix(c(.7, .65, .6, .55), 4, 1)
  fit_out <- list(unrot_loadings = L)
  p <- nrow(L); k <- ncol(L); pk <- p * k

  core <- EFAtools:::.se_sandwich_core(fit_out, N = 200, Gamma = NULL, method = "ULS")
  out <- suppressWarnings(EFAtools:::.se_sandwich_unrotated(fit_out, core, ci = 0.95))

  expect_true("vcov_unrot_loadings" %in% names(out))
  V <- out$vcov_unrot_loadings
  expect_false(is.null(V))
  expect_true(is.matrix(V))
  expect_equal(dim(V), c(pk, pk))
  expect_true(all(is.na(V)))
})


test_that("a finite-but-not-PSD information vcov is NA-filled to match the marginal SEs", {
  # The information path has TWO unreliability routes: (i) the Heywood early return already
  # covered above; (ii) the post-solve PSD gate, where the bordered inverse succeeds but yields a
  # covariance carrying a negative eigenvalue, so the marginal SEs are NA while the covariance
  # itself stays finite. The persisted slot must NA-fill on the second route too, so a non-PSD
  # covariance never ships next to NA marginal SEs and silently propagates `sqrt(NaN)` downstream.
  # Forcing `.is_psd()` to fail isolates that gate on a fit that is otherwise perfectly healthy.
  L <- matrix(c(.7, .65, .6, .55, .5, .6), 6, 1)
  fit_out <- list(unrot_loadings = L)
  p <- nrow(L); k <- ncol(L); pk <- p * k

  out <- testthat::with_mocked_bindings(
    suppressWarnings(
      EFAtools:::.se_information(fit_out, rot_info = NULL, N = 200, ci = 0.95, method = "ML")),
    .is_psd = function(M) FALSE,
    .package = "EFAtools"
  )

  expect_true(anyNA(out$SE$unrot_loadings))

  V <- out$vcov_unrot_loadings
  expect_false(is.null(V))
  expect_equal(dim(V), c(pk, pk))
  expect_true(all(is.na(V)))
})


test_that("a finite-but-not-PSD sandwich V_AA is NA-filled to match the marginal SEs", {
  # `.se_sandwich_core()` returns `V_AA = V_AA` (unchanged) even when `reliable = FALSE`,
  # only the marginal `loadings_se` is NA-filled in that branch. The persistence wrapper
  # must mirror the marginal NA into the persisted vcov.
  L <- matrix(c(.7, .65, .6, .55), 4, 1)
  fit_out <- list(unrot_loadings = L)
  p <- nrow(L); k <- ncol(L); pk <- p * k

  fake_core <- list(
    V_AA = matrix(1.0, pk, pk),         # finite numeric
    loadings_se = matrix(NA_real_, p, k),
    uniquenesses_se = rep(NA_real_, p),
    scaled_test = NULL,
    reliable = FALSE
  )
  out <- suppressWarnings(EFAtools:::.se_sandwich_unrotated(fit_out, fake_core, ci = 0.95))

  V <- out$vcov_unrot_loadings
  expect_false(is.null(V))
  expect_equal(dim(V), c(pk, pk))
  expect_true(all(is.na(V)))
})


test_that("the top-level name vector under analytic SE is pinned", {
  # The 9 `expect_named()` blocks in test-EFA.R only fire on se = "none" fits, where the
  # analytic-SE branch is skipped, so the order `..., settings, SE, CI, replicates,
  # vcov_unrot_loadings` is never asserted there. Pin it here so a future refactor that
  # reorders the slot assignments around `.compute_se_ci()` cannot silently drift.
  fit <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500, method = "ML",
             rotation = "none", se = "information")
  expect_named(fit, c("orig_R", "h2", "orig_eigen", "final_eigen", "iter",
                      "convergence", "heywood", "unrot_loadings", "vars_accounted",
                      "fit_indices", "model_implied_R", "residuals", "settings",
                      "vcov_unrot_loadings", "Gamma", "SE", "CI", "replicates"))
})


test_that("vcov_unrot_loadings labels its column-major vec(Lambda) ordering", {
  # The documented ordering is otherwise only stated in prose, leaving the user of a 54 x 54 matrix
  # to take it on trust. The labels make it readable off the object -- and pin that the block really
  # is column-major (all variables for factor 1, then factor 2, ...).
  cormat <- test_models$baseline$cormat
  p <- nrow(cormat); k <- 3L

  fit <- EFA(cormat, n_factors = k, N = 500, method = "ML", rotation = "oblimin",
             se = "information")
  vars <- rownames(unclass(fit$unrot_loadings))
  facs <- colnames(unclass(fit$unrot_loadings))
  expected <- as.vector(outer(vars, facs, paste, sep = "_"))

  expect_identical(rownames(fit$vcov_unrot_loadings), expected)
  expect_identical(colnames(fit$vcov_unrot_loadings), expected)
  expect_identical(expected[1:2], paste0(vars[1:2], "_", facs[1]))
  expect_identical(expected[p + 1L], paste0(vars[1], "_", facs[2]))

  # The unrotated path assembles the block in a different function; it must label it identically.
  unrot <- EFA(cormat, n_factors = k, N = 500, method = "ML", rotation = "none",
               se = "information")
  expect_identical(rownames(unrot$vcov_unrot_loadings), expected)
  expect_identical(colnames(unrot$vcov_unrot_loadings), expected)
})


test_that("se = 'none' leaves vcov_unrot_loadings present but NULL", {
  fit <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500, method = "ML",
             rotation = "none", se = "none")

  expect_true("vcov_unrot_loadings" %in% names(fit))
  expect_null(fit$vcov_unrot_loadings)
})


test_that("se = 'np-boot' leaves vcov_unrot_loadings present but NULL", {
  # Tiny bootstrap; this test only asserts the storage convention.
  dat <- make_adf_fixture(N = 200,
                          loadings = matrix(c(.7, .6, .55, .65, .5, .5), ncol = 1L),
                          seed = 13)
  fit <- suppressWarnings(EFA(dat, n_factors = 1, cor_method = "pearson", method = "ML",
                              rotation = "none", se = "np-boot", b_boot = 20, seed = 1))

  expect_true("vcov_unrot_loadings" %in% names(fit))
  expect_null(fit$vcov_unrot_loadings)
})


test_that("se = 'sandwich' persists Gamma on the polychoric path with the right schema", {
  skip_on_cran()

  dat <- DOSPERT_raw[, 1:8]
  p <- ncol(dat); n_pairs <- p * (p - 1L) / 2L

  fit <- .poly_sandwich_fit()

  expect_true("Gamma" %in% names(fit))
  G <- fit$Gamma
  expect_true(is.matrix(G) && is.numeric(G))
  expect_equal(dim(G), c(n_pairs, n_pairs))
  expect_true(isSymmetric(G, tol = 1e-10))
  expect_true(all(diag(G) > 0))

  # Pair ordering matches utils::combn(p, 2) -- the dimnames are set by .polychoric() in
  # that order. Reconstruct the expected labels and confirm.
  vn <- colnames(dat)
  labels <- apply(utils::combn(vn, 2L), 2L, paste, collapse = "-")
  expect_equal(rownames(G), labels)
  expect_equal(colnames(G), labels)
})


test_that("se = 'sandwich' persists Gamma on the continuous Pearson path with the right schema", {
  dat <- make_adf_fixture(N = 500,
                          loadings = matrix(c(.7, .65, .6, .75, .55, .5), ncol = 1L),
                          seed = 14)
  p <- ncol(dat); n_pairs <- p * (p - 1L) / 2L

  fit <- EFA(dat, n_factors = 1, cor_method = "pearson", method = "ML", rotation = "none",
             se = "sandwich")

  G <- fit$Gamma
  expect_true(is.matrix(G) && is.numeric(G))
  expect_equal(dim(G), c(n_pairs, n_pairs))
  expect_true(isSymmetric(G, tol = 1e-10))
  expect_true(all(diag(G) > 0))

  vn <- colnames(dat)
  labels <- apply(utils::combn(vn, 2L), 2L, paste, collapse = "-")
  expect_equal(rownames(G), labels)
  expect_equal(colnames(G), labels)
})


test_that("Gamma is present but NULL outside the sandwich path", {
  cormat <- test_models$baseline$cormat

  fit_none <- EFA(cormat, n_factors = 3, N = 500, method = "ML", rotation = "none",
                  se = "none")
  fit_info <- EFA(cormat, n_factors = 3, N = 500, method = "ML", rotation = "none",
                  se = "information")
  fit_boot <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, method = "ML", rotation = "none",
        se = "np-boot", b_boot = 10, seed = 1)
  )

  for (fit in list(fit_none, fit_info, fit_boot)) {
    expect_true("Gamma" %in% names(fit))
    expect_null(fit$Gamma)
  }
})


test_that("Gamma matches lavaan's correlation NACOV up to the N scale (continuous Pearson)", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  skip_if(utils::packageVersion("lavaan") < "0.6.9")

  L <- matrix(c(.7, .65, .6, .75, .55, .5), ncol = 1L)
  dat <- make_adf_fixture(N = 600, loadings = L, seed = 15)
  vn <- colnames(dat)
  N <- nrow(dat)

  fit <- EFA(dat, n_factors = 1, cor_method = "pearson", method = "ML", rotation = "none",
             se = "sandwich")

  # The persisted Gamma is the variance-scale ADF covariance; lavaan's NACOV is N * Gamma.
  mod <- paste0("f =~ ", paste(vn, collapse = " + "))
  lfit <- lavaan::cfa(mod, data = dat, estimator = "MLM", std.lv = TRUE,
                      correlation = TRUE, meanstructure = TRUE)
  Glav <- lavaan::lavInspect(lfit, "gamma")
  if (is.list(Glav)) Glav <- Glav[[1]]
  # With meanstructure = TRUE lavaan returns the means PLUS the off-diagonal correlation rows
  # (in utils::combn() order); subset to the off-diagonal correlation rows for a clean
  # comparison.
  cor_rows <- .lav_cor_rows(Glav)
  Glav_corr <- unname(Glav[cor_rows, cor_rows])
  expect_equal(nrow(Glav_corr), choose(length(vn), 2L))

  rel_F <- norm(N * unname(fit$Gamma) - Glav_corr, "F") / norm(Glav_corr, "F")
  expect_lt(rel_F, 1e-4)
})


test_that("Gamma matches lavaan's polychoric NACOV up to the N scale (ordinal)", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  # lavaan::efa() wants a data frame with an `ordered` column list; the EFAtools fit routes
  # its input through data.matrix() on the polychoric path, so the shared matrix-form fixture
  # is the same object (verified: identical Gamma, loadings and vcov) and is reused here
  # rather than paying for a third polychoric ACOV over 3,123 rows.
  dat <- as.data.frame(DOSPERT_raw[, 1:8])
  vn <- colnames(dat)
  N <- sum(stats::complete.cases(dat))

  fit <- .poly_sandwich_fit()

  # lavaan::efa() returns an `efaList`; lavInspect expects a single lavaan fit, so take [[1]].
  lf <- lavaan::efa(dat, nfactors = 1, ordered = vn, estimator = "DWLS", se = "robust.sem")
  lfit <- lf[[1]]
  Glav <- lavaan::lavInspect(lfit, "gamma")
  if (is.list(Glav)) Glav <- Glav[[1]]
  cor_rows <- .lav_cor_rows(Glav)
  Glav_corr <- unname(Glav[cor_rows, cor_rows])
  expect_equal(nrow(Glav_corr), choose(length(vn), 2L))

  rel_F <- norm(N * unname(fit$Gamma) - Glav_corr, "F") / norm(Glav_corr, "F")
  expect_lt(rel_F, 1e-3)
})

# Polychoric / tetrachoric correlation matrix (.polychoric + the C++ backend).
#
# The two-step estimator is validated against the conditional-ML two-step references
# polycor::polychor(ML = FALSE) and psych::polychoric(correct = FALSE, global = FALSE);
# the thresholds match those functions' per-pair thresholds (the full-column marginals on
# complete data, the pairwise-complete marginals under missing data), so the estimates agree.
# Near-collinear pairs, where those fixed-quadrature references themselves lose accuracy, are
# instead checked against an mnormt two-step gold. The bivariate-normal rectangle that drives
# the likelihood is cross-checked against mnormt::sadmvn. Heavy reference loops over the large
# item sets run only under EFATOOLS_TEST_SLOW.

# Reference two-step polychoric matrix via polycor, pairwise.
.ref_polychor <- function(x) {
  p <- ncol(x)
  m <- diag(p)
  for (i in seq_len(p - 1L)) for (j in (i + 1L):p) {
    m[i, j] <- m[j, i] <- suppressWarnings(polycor::polychor(x[, i], x[, j], ML = FALSE))
  }
  m
}

# Exact two-step polychoric negative log-likelihood for one pair, integrating each cell with
# mnormt's adaptive bivariate-normal rule. Returned as a closure over the fixed thresholds and
# the contingency table, so a candidate estimate can be scored on the same objective
# .gold_polychor maximises.
.gold_nll <- function(a, b) {
  ai <- match(a, sort(unique(a))) - 1L; bi <- match(b, sort(unique(b))) - 1L
  Ki <- max(ai) + 1L; Kj <- max(bi) + 1L; N <- length(ai)
  ti <- stats::qnorm(cumsum(tabulate(ai + 1L, Ki)) / N)[-Ki]
  tj <- stats::qnorm(cumsum(tabulate(bi + 1L, Kj)) / N)[-Kj]
  rc <- c(-Inf, ti, Inf); cc <- c(-Inf, tj, Inf)
  nab <- matrix(0, Ki, Kj)
  for (r in seq_len(N)) nab[ai[r] + 1L, bi[r] + 1L] <- nab[ai[r] + 1L, bi[r] + 1L] + 1
  function(rho) {
    V <- matrix(c(1, rho, rho, 1), 2L); s <- 0
    for (i in seq_len(Ki)) for (j in seq_len(Kj)) if (nab[i, j] > 0)
      s <- s + nab[i, j] * log(max(as.numeric(mnormt::sadmvn(
        c(rc[i], cc[j]), c(rc[i + 1L], cc[j + 1L]), c(0, 0), V)), 1e-300))
    -s
  }
}

# Exact two-step polychoric for one pair: the maximiser of .gold_nll. This is the oracle in the
# near-collinear regime, where the fixed-quadrature reference estimators (polycor / psych)
# themselves lose accuracy.
.gold_polychor <- function(a, b) {
  stats::optimize(.gold_nll(a, b), c(-0.9999, 0.9999), tol = 1e-9)$minimum
}

# Expand a contingency table of counts into raw two-column ordinal data (row, column codes).
.expand_table <- function(tab) {
  do.call(rbind, lapply(which(tab > 0), function(k) {
    a <- ((k - 1L) %% nrow(tab)) + 1L
    b <- ((k - 1L) %/% nrow(tab)) + 1L
    cbind(rep(a, tab[a, b]), rep(b, tab[a, b]))
  }))
}

# Deterministic raw ordinal data whose two-step polychoric matrix is not positive
# definite: four items are noisy monotone functions of the common score z1 + z2, so the
# columns are near-collinear and disattenuation pushes the smallest eigenvalue clearly
# negative (~ -0.07). Built without quantile cut points so it cannot tie / collapse a
# category across RNG streams.
.nonpd_ordinal <- function() {
  set.seed(1)
  n <- 120L
  z1 <- sample(1:5, n, TRUE)
  z2 <- sample(1:5, n, TRUE)
  core <- z1 + z2
  derive <- function() {
    v <- core + sample(c(-1L, 0L, 1L), n, TRUE)
    as.integer(pmin(6L, pmax(1L, v - 1L)))
  }
  cbind(z1, z2, derive(), derive(), derive(), derive())
}

# Listwise-complete GRiPS and its default polychoric matrix, computed once and
# reused by the structural blocks that would otherwise rebuild them. Tests that
# need non-default arguments (`nearest_pd`, `acov`) or that mutate the data still
# build their own.
g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), ]
poly_g <- .polychoric(g)

test_that(".bvn_rect_cpp matches mnormt::sadmvn rectangle probabilities", {
  skip_on_cran()
  skip_if_not_installed("mnormt")

  cuts <- c(-Inf, -2.5, -1, 0, 1, 2.5, Inf)
  worst <- 0
  for (rho in c(0, 0.3, 0.5, 0.75, 0.9, 0.95, 0.99, -0.8)) {
    V <- matrix(c(1, rho, rho, 1), 2L)
    for (ia in seq_len(length(cuts) - 1L)) for (ib in seq_len(length(cuts) - 1L)) {
      ours <- .bvn_rect_cpp(cuts[ia], cuts[ia + 1L], cuts[ib], cuts[ib + 1L], rho)
      ref <- as.numeric(mnormt::sadmvn(c(cuts[ia], cuts[ib]),
                                       c(cuts[ia + 1L], cuts[ib + 1L]),
                                       c(0, 0), V))
      worst <- max(worst, abs(ours - ref))
    }
  }
  # The quadrature reproduces the reference bivariate-normal integrator to better than 1e-6 over
  # this grid: the 12-node rule for |rho| up to 0.95, and the finer rule (used once |rho| exceeds
  # the refinement threshold) at rho = 0.99, where the conditional transition is narrow. This is
  # far tighter than the ~1e-4 agreement the polychoric matrix itself needs.
  expect_lt(worst, 1e-6)
})

test_that(".bvn_rect_cpp keeps its relative accuracy in the far tails", {
  # The absolute check above cannot see a 1e-45 rectangle collapse to zero, and that is exactly
  # what the opposite-corner cells of a strongly correlated pair are. Such a cell can be
  # OBSERVED -- a few careless or extreme responses put it there -- and then its probability
  # enters the log-likelihood and its score the asymptotic covariance, so what matters is
  # relative accuracy at magnitudes far below any absolute tolerance. The upper-tail rectangle
  # is the demanding one: its conditional band has to be taken from the complementary tail,
  # since as a plain difference of normal CDFs it is 1 - 1 = 0 and the rectangle vanishes
  # however large it really is.
  #
  # The oracle is the (X, Y) -> (-X, -Y) symmetry of the bivariate normal, which maps an
  # upper-tail rectangle onto its mirror image in the LOWER tail -- where a CDF difference
  # cannot cancel, because both terms are small. The identity is exact, needs no external
  # integrator (mnormt itself returns 0 for some of these rectangles), and holds to machine
  # precision here because the two sides are the same quadrature rule on mirrored bands.
  #
  # Compared as a RATIO, not as two values: expect_equal() on quantities this small would
  # compare them absolutely -- any two numbers below 1e-45 are within any usable tolerance of
  # each other -- and the assertion would hold whatever the upper tail returned. On the ratio the
  # tolerance means what it says, and the pre-fix values fail it by 6e-11 at rho = 0.9 and by
  # 7.7e-06 at rho = 0.95, where the cancelled band is inaccurate rather than absent.
  cut7 <- stats::qnorm(seq_len(6) / 7)
  # rho and the corner offset k; deeper than (0.999, 2) the rectangle genuinely underflows,
  # and at |rho| > 0.95 the finer quadrature rule is used, so both rules are covered.
  grid <- list(c(0.9, 1), c(0.9, 2), c(0.95, 1), c(0.95, 2),
               c(0.99, 1), c(0.99, 2), c(0.999, 2))
  for (gr in grid) {
    rho <- gr[1]; k <- gr[2]
    a <- cut7[k]; b <- cut7[7L - k]
    up <- .bvn_rect_cpp(-Inf, a, b, Inf, rho)          # X <= a, Y > b   (upper Y tail)
    lo <- .bvn_rect_cpp(-a, Inf, -Inf, -b, rho)        # X > -a, Y <= -b (its mirror image)
    expect_gt(up, 0)                                   # the cancelled band collapses to exactly 0
    expect_equal(up / lo, 1, tolerance = 1e-12)
  }
})

test_that("near-collinear pairs match the exact two-step estimate", {
  skip_on_cran()
  skip_if_not_installed("mnormt")

  # Near |rho| = 1 the contingency table has near-impossible off-diagonal cells whose
  # probability underflows; a naive expected-information step false-converges at the warm start,
  # and a fixed-order quadrature under-resolves the narrow conditional band. The estimator must
  # still recover the exact two-step value (and the reference estimators polycor / psych are
  # themselves biased here, so the oracle is the mnormt two-step).
  cut7 <- stats::qnorm(seq_len(6) / 7)
  set.seed(101)
  N <- 4000
  for (rt in c(0.97, 0.99, 0.995, 0.999)) {
    L <- chol(matrix(c(1, rt, rt, 1), 2L))
    Z <- matrix(stats::rnorm(2 * N), N, 2L) %*% L
    a <- findInterval(Z[, 1], cut7); b <- findInterval(Z[, 2], cut7)
    expect_equal(.polychoric(cbind(a, b))$R[1, 2], .gold_polychor(a, b), tolerance = 1e-4)
  }
})

test_that("binary pairs below the refinement gate hold the documented 1e-4 accuracy", {
  skip_on_cran()
  skip_if_not_installed("mnormt")

  # The band just below the gate at which the finer quadrature rule takes over is where the
  # fixed 12-node rule is least accurate, and a two-by-two table is its worst case: with only
  # one threshold per variable the whole likelihood rests on four wide bands whose conditional
  # transition the base rule under-resolves. The help page promises about 1e-4 there (against
  # about 1e-5 elsewhere), so that is what is pinned. The bound is set a few times the
  # measured discrepancy -- about 3e-5 at rho = .85 and 1.5e-4 at rho = .92 across seeds --
  # because it must hold on any BLAS, while still catching a loss of the refinement rule or a
  # coarser base rule, which cost an order of magnitude.
  gate_tol <- 5e-4
  for (rt in c(0.85, 0.92, -0.85, -0.92)) {
    for (seed in c(11, 12)) {
      set.seed(seed)
      L <- chol(matrix(c(1, rt, rt, 1), 2L))
      Z <- matrix(stats::rnorm(2 * 1500), 1500, 2L) %*% L
      a <- as.integer(Z[, 1] > 0); b <- as.integer(Z[, 2] > 0)
      ours <- .polychoric(cbind(a, b))$R[1, 2]
      expect_lt(abs(ours - .gold_polychor(a, b)), gate_tol)
      # Reversing one variable's coding reflects the latent variable about its threshold,
      # so the estimate must come back negated -- the accuracy above is a quadrature
      # residual, not an asymmetry between the two signs. Measured at exactly zero
      # difference here, but pinned loosely enough that a last-ulp difference in the
      # complementary threshold on another platform's qnorm cannot fail it, and still
      # six orders tighter than the accuracy band it has to distinguish.
      expect_equal(.polychoric(cbind(a, 1L - b))$R[1, 2], -ours, tolerance = 1e-10)
    }
  }

  # Unequal marginals put the single threshold off centre, which moves the bands the rule has
  # to resolve; the same accuracy has to hold there.
  set.seed(31)
  L <- chol(matrix(c(1, 0.92, 0.92, 1), 2L))
  Z <- matrix(stats::rnorm(2 * 1500), 1500, 2L) %*% L
  a <- as.integer(Z[, 1] > 0.6); b <- as.integer(Z[, 2] > 0.6)
  expect_lt(abs(.polychoric(cbind(a, b))$R[1, 2] - .gold_polychor(a, b)), gate_tol)
})

test_that("an empty pairwise-complete marginal category does not corrupt the estimate", {
  skip_on_cran()
  skip_if_not_installed("polycor")

  set.seed(303)
  N <- 1500
  L <- chol(matrix(c(1, .5, .5, 1), 2L))
  Z <- matrix(stats::rnorm(2 * N), N, 2L) %*% L
  a <- findInterval(Z[, 1], stats::qnorm(c(.2, .5, .8)))   # categories 0..3
  b <- findInterval(Z[, 2], stats::qnorm(c(.3, .7)))       # categories 0..2

  # Drop b wherever a is in its top (then bottom) category, so that category is absent from the
  # pair's pairwise-complete table while still present in a's full column. The per-pair
  # cumulative proportion then lands exactly on 1 (or 0), which must map to an infinite cut and
  # a zero-width empty band -- not a NaN that silently corrupts the estimate.
  for (empty in c(3L, 0L)) {
    bm <- b; bm[a == empty] <- NA
    ours <- .polychoric(cbind(a, bm))$R[1, 2]
    expect_false(is.na(ours))
    expect_equal(ours, suppressWarnings(polycor::polychor(a, bm, ML = FALSE)),
                 tolerance = 1e-3)
  }
})

test_that("an observed near-impossible cell does not bias a strongly-correlated pair", {
  skip_on_cran()
  skip_if_not_installed("mnormt")

  cut7 <- stats::qnorm(seq_len(6) / 7)
  set.seed(404)
  N <- 5000
  L <- chol(matrix(c(1, 0.995, 0.995, 1), 2L))
  Z <- matrix(stats::rnorm(2 * N), N, 2L) %*% L
  a <- findInterval(Z[, 1], cut7); b <- findInterval(Z[, 2], cut7)
  # A few careless responses fall in the two opposite corners, whose model probabilities at the
  # optimum are about 3e-45 and 1e-44. Both are far above the smallest representable double, but
  # the upper-right one is reached only from the complementary tail: taken as a plain difference
  # of normal CDFs its band is 1 - 1 = 0 exactly, which annihilates the cell and biases the pair
  # by ~4e-3 (the two mirror-image corners then contribute quite differently despite carrying the
  # same three cases each). Computed from the nearer tail, both corners contribute and the pair
  # tracks the exact two-step value to ~1e-5.
  a[1:3] <- 0L; b[1:3] <- 6L
  a[4:6] <- 6L; b[4:6] <- 0L

  # The asymptotic variance of this pair is available too: it is estimated in the interior, so it
  # is neither a boundary nor a continuity-corrected pair, and nothing in its influence function
  # overflows once the corner probabilities are accurate. Before that it came out NaN, which made
  # the pair unusable for DWLS and withheld every robust standard error involving it.
  # (The empty interior cells of this table trip the separate sparse-cell diagnostic, which is
  # computed from the observed counts and is not what is under test here.)
  d <- suppressWarnings(.polychoric(cbind(a, b), acov = "diag"),
                        classes = "efa_cor_sparse_cells")
  f <- suppressWarnings(.polychoric(cbind(a, b), acov = "full"),
                        classes = "efa_cor_sparse_cells")
  expect_equal(d$R[1, 2], .gold_polychor(a, b), tolerance = 1e-4)
  expect_false(d$at_bound[[1]])
  expect_false(d$zero_corrected[[1]])
  expect_true(is.finite(d$acov[[1]]) && d$acov[[1]] > 0)
  expect_equal(unname(d$acov), unname(diag(f$acov)), tolerance = 1e-10)
})

test_that("polychoric matrix matches polycor and psych on GRiPS", {
  skip_on_cran()
  skip_if_not_installed("polycor")

  ours <- poly_g$R

  expect_lt(max(abs(ours - .ref_polychor(g))), 1e-4)

  psych_rho <- suppressWarnings(
    psych::polychoric(g, correct = FALSE, global = FALSE)$rho)
  expect_lt(max(abs(ours - unname(psych_rho))), 1e-4)
})

test_that("tetrachoric (2 categories) matches polycor and psych", {
  skip_on_cran()
  skip_if_not_installed("polycor")

  # Pinned against the UNcorrected references even though binary pairs are continuity-corrected:
  # no pair of this median split has an empty cell, so the correction never fires and the two
  # conventions coincide here (psych gives the identical matrix for correct = 0.5 and FALSE).
  gb <- apply(g, 2L, function(col) as.integer(col > stats::median(col)))
  poly_gb <- .polychoric(gb)
  expect_false(any(poly_gb$zero_corrected))
  ours <- poly_gb$R

  expect_lt(max(abs(ours - .ref_polychor(gb))), 1e-4)

  tet_rho <- suppressWarnings(psych::tetrachoric(gb, correct = FALSE)$rho)
  expect_lt(max(abs(ours - unname(tet_rho))), 1e-4)
})

test_that("the matrix is a valid, named correlation matrix", {
  res <- poly_g

  expect_equal(dim(res$R), c(ncol(g), ncol(g)))
  expect_equal(dimnames(res$R), list(colnames(g), colnames(g)))
  expect_equal(diag(res$R), rep(1, ncol(g)), ignore_attr = TRUE)
  expect_true(isSymmetric(res$R))
  expect_true(all(res$R >= -1 & res$R <= 1))
})

test_that("the result is deterministic", {
  # Compared against the file-top `poly_g`, an independent earlier call with the same
  # arguments on the same object: two invocations separated by other work are a stronger
  # determinism probe than two adjacent ones, and one matrix fewer is built.
  expect_identical(poly_g$R, .polychoric(g)$R)
})

test_that("a constant column is rejected with a classed condition", {
  # local copy: this block mutates the column to force the constant-column path
  gx <- g
  gx[, 1L] <- 3L
  expect_error(.polychoric(gx), class = "efa_cor_constant_col")
})

test_that("the category guards read correctly for one and for several variables", {
  # Both bullets open with cli::qty() and then name the offending variables, so one variable
  # must read "Variable ... is" and several "Variables ... are". A space after the qty() call
  # renders as a second space behind the bullet marker, which is why neither branch has one.
  base <- cbind(a = rep(0:2, length.out = 60L), b = rep(0:1, length.out = 60L),
                c = rep(0:2, length.out = 60L))

  constant_msg <- function(x) {
    e <- tryCatch(.polychoric(x), efa_cor_constant_col = function(e) e)
    expect_s3_class(e, "efa_cor_constant_col")
    conditionMessage(e)
  }
  one_const <- base
  one_const[, "a"] <- 1L
  two_const <- one_const
  two_const[, "b"] <- 2L
  expect_snapshot(cat(constant_msg(one_const)))
  expect_snapshot(cat(constant_msg(two_const)))

  binary_msg <- function(x) {
    e <- tryCatch(.polychoric(x, binary_only = TRUE), efa_cor_not_binary = function(e) e)
    expect_s3_class(e, "efa_cor_not_binary")
    conditionMessage(e)
  }
  one_wide <- base
  one_wide[, "c"] <- rep(0:1, length.out = 60L)
  expect_snapshot(cat(binary_msg(one_wide)))
  expect_snapshot(cat(binary_msg(base)))
})

test_that("a single variable is rejected at the R level with a classed condition", {
  # The C++ backend keeps the same check, but it raises an unclassed Rcpp error, so the
  # abort must happen in the wrapper.
  expect_error(.polychoric(g[, 1L, drop = FALSE]), class = "efa_cor_too_few_vars")
})

test_that("a non-positive-definite matrix is left alone unless nearest_pd is requested", {
  x <- .nonpd_ordinal()

  raw <- .polychoric(x)$R
  expect_lt(min(eigen(raw, symmetric = TRUE, only.values = TRUE)$values), 0)

  expect_warning(adj <- .polychoric(x, nearest_pd = TRUE)$R, class = "efa_cor_smoothed")
  expect_gt(min(eigen(adj, symmetric = TRUE, only.values = TRUE)$values), 0)
  expect_equal(diag(adj), rep(1, ncol(x)), ignore_attr = TRUE)
})

test_that("missing data is handled pairwise-complete", {
  gm <- g
  gm[1:20, 1] <- NA          # scattered missingness; every pair still overlaps
  gm[21:40, 2] <- NA
  res <- .polychoric(gm)
  expect_false(anyNA(res$R))
  expect_true(isSymmetric(res$R))
  expect_equal(diag(res$R), rep(1, ncol(gm)), ignore_attr = TRUE)
})

test_that("a pair with no overlapping complete cases is rejected with a classed condition", {
  v6 <- c(1L, 2L, 3L, 1L, 2L, 3L)
  na6 <- rep(NA_integer_, 6L)
  # columns 1 and 2 are never observed on the same row -> that pair is uncomputable;
  # column 3 overlaps both so only the 1-2 entry is NA.
  m <- cbind(c(v6, na6), c(na6, v6), rep(1:3, 4L))
  expect_error(.polychoric(m), class = "efa_cor_na")
})

test_that("the uncomputable-pair error reads correctly for one and for several pairs", {
  # The bullet opens with cli::qty(); a space after that call renders as a second space
  # behind the bullet marker, so it has none, and both plural branches are recorded here.
  na_msg <- function(x) {
    e <- tryCatch(.polychoric(x), efa_cor_na = function(e) e)
    expect_s3_class(e, "efa_cor_na")
    conditionMessage(e)
  }
  n <- 200L
  half <- rep(0:1, length.out = n)
  # `a` and `b` are never observed on the same row, so that pair is uncomputable; `c` overlaps
  # both. `d` follows `a`'s missingness pattern, which makes a second uncomputable pair.
  a <- half; a[1:100] <- NA
  b <- half; b[101:200] <- NA
  d <- half; d[1:100] <- NA
  expect_snapshot(cat(na_msg(data.frame(a = a, b = b, c = half))))
  expect_snapshot(cat(na_msg(data.frame(a = a, b = b, d = d, c = half))))
})

test_that("the boundary estimate of a Frechet-bound pair attains the maximum likelihood", {
  skip_if_not_installed("mnormt")
  # This table is the comonotone coupling of its margins -- the Frechet upper bound (Frechet,
  # 1951). On the cumulative scale that bound is the copula M(u, v) = min(u, v), which is exactly
  # the bivariate normal copula at rho = 1, and the two-step thresholds fix the model's margins to
  # this table's margins. So at rho = 1 the model reproduces the table exactly, the log-likelihood
  # attains its saturated value, and the maximiser is the boundary of the parameter space. The
  # estimator reports the boundary of the domain it maximises over (+/- 0.9999, as in polycor);
  # the likelihood is numerically flat over the whole approach to it, so an optimiser left to
  # search there stops at an arbitrary, platform-dependent point instead.
  #
  # What is checked here is the statistical claim behind that: the reported value fits at least as
  # well as the mnormt two-step gold, i.e. it really is a maximiser and not merely a bound. The
  # 1e-6 slack is measured, not inherited from sadmvn's abseps: at a perfect fit every cell's
  # n_ab / P_ab equals N, so an abseps-sized (1e-6) probability error would move this nll by as
  # much as sum(n_ab / P_ab) * abseps = 1.5e-3. sadmvn is far more accurate than it promises on
  # these smooth rectangles -- the measured spread over the whole plateau is ~6e-14 -- leaving
  # 1e-6 seven orders above the noise and three below the ~1.4e-3 a mislocated estimate pays.
  x <- .expand_table(rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47)))
  ours <- suppressWarnings(.polychoric(x))$R[1, 2]
  expect_identical(ours, 0.9999)
  nll <- .gold_nll(x[, 1], x[, 2])
  gold <- .gold_polychor(x[, 1], x[, 2])
  expect_lte(nll(ours), nll(gold) + 1e-6)
})

test_that("a Frechet-bound pair is detected exactly and reported at the boundary", {
  # The detector compares the table cell-by-cell against both bound couplings of its own
  # marginals, built from the cumulative margins by inclusion-exclusion. Every quantity is an
  # integer count, so the verdict involves no floating point and is identical on every platform --
  # which is the whole point: the estimate and the asymptotic variance of such a pair are boundary
  # quantities that an optimiser cannot pin down (the variance of this fixture has been observed
  # anywhere between 1e-13 and 1e+26 depending on the platform).
  upper <- rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47))

  # Upper bound (perfectly ordered) and its mirror, the lower bound (perfectly reversed): the
  # sign of the estimate follows the bound, and neither has an asymptotic variance.
  for (case in list(list(tab = upper,          rho = 0.9999),
                    list(tab = upper[, 2:1],   rho = -0.9999))) {
    res <- suppressWarnings(.polychoric(.expand_table(case$tab), acov = "diag"))
    expect_identical(res$R[1, 2], case$rho)
    expect_true(res$at_bound[[1L]])
    expect_identical(unname(res$acov), NA_real_)
  }

  # A 2x2 table with a zero cell is at a bound too, but it is repaired by the continuity
  # correction rather than reported (see the tetrachoric tests below). Two empty cells leave
  # nothing to repair -- the correction is defined for a single empty cell -- so those tables
  # stay on the boundary path, in both directions.
  for (case in list(list(tab = rbind(c(40, 0), c(0, 200)), rho = 0.9999),
                    list(tab = rbind(c(0, 40), c(200, 0)), rho = -0.9999))) {
    res <- suppressWarnings(.polychoric(.expand_table(case$tab), acov = "diag"))
    expect_identical(res$R[1, 2], case$rho)
    expect_true(res$at_bound[[1L]])
    expect_false(res$zero_corrected[[1L]])
    expect_identical(unname(res$acov), NA_real_)
  }

  # One count off the bound and the pair is ordinary again: not flagged, estimated in the
  # interior, and with a finite positive variance. There is no grey zone between the two.
  off <- suppressWarnings(.polychoric(.expand_table(rbind(c(40, 2), c(60, 198))), acov = "diag"))
  expect_false(off$at_bound[[1L]])
  expect_false(off$zero_corrected[[1L]])
  expect_lt(abs(off$R[1, 2]), 0.95)
  expect_true(is.finite(off$acov[[1L]]) && off$acov[[1L]] > 0)
})

test_that("a Frechet-bound pair is reproducible under a one-count perturbation", {
  # Moving a single observation between populated cells leaves the table at the bound (the
  # marginals change, but the monotone rearrangement of the new marginals is again this table), so
  # every member of this family is the same boundary problem. It is precisely where the old
  # arbitrary-stopping-point estimate made the asymptotic variance vary by 39 orders of magnitude
  # across the family, non-monotonically, and differ between platforms. All of them must now give
  # the identical answer. The probe is only meaningful among perturbations that stay on the same
  # side of the bound -- one that moved a count INTO a structurally empty cell would leave the
  # family, and is expected to be estimated in the interior instead (covered above).
  tab <- rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47))
  for (k in which(tab > 0)) {
    for (step in c(-1L, 1L)) {
      t2 <- tab
      t2[k] <- t2[k] + step
      res <- suppressWarnings(.polychoric(.expand_table(t2), acov = "diag"))
      expect_identical(res$R[1, 2], 0.9999)
      expect_identical(unname(res$acov), NA_real_)
    }
  }
})

test_that("a binary pair with one empty cell is continuity-corrected, not reported at the bound", {
  # Every 2x2 table with an empty cell is the extreme coupling of its own margins, so without a
  # correction its estimate would be +/-1 whatever the underlying correlation -- which for binary
  # data is the common case, not a corner case. A continuity correction of 0.5 is added to the
  # empty cell instead, so the pair is estimated in the interior. The sign still follows the
  # corner the empty cell sits in.
  for (case in list(list(tab = rbind(c(40, 0), c(60, 200)),  sign = 1),
                    list(tab = rbind(c(0, 40), c(200, 60)),  sign = -1),
                    list(tab = rbind(c(150, 50), c(100, 0)), sign = -1))) {
    res <- suppressWarnings(.polychoric(.expand_table(case$tab), acov = "diag"))
    expect_false(res$at_bound[[1L]])
    expect_true(res$zero_corrected[[1L]])
    expect_identical(sign(res$R[1, 2]), case$sign)
    expect_lt(abs(res$R[1, 2]), 0.9999)
    # The correction repairs the point estimate only: the sandwich would treat the nudged counts
    # as data, and its coverage in this regime is not trustworthy, so no variance is reported.
    expect_identical(unname(res$acov), NA_real_)
  }

  # The mirror image must give the mirrored estimate, so the correction cannot depend on which
  # corner is empty. Compared numerically rather than bit-wise: the two tables put the correction
  # in different cells and so take slightly different routes through the iteration, which leaves
  # them a single ulp apart.
  up <- suppressWarnings(.polychoric(.expand_table(rbind(c(40, 0), c(60, 200)))))$R[1, 2]
  dn <- suppressWarnings(.polychoric(.expand_table(rbind(c(0, 40), c(200, 60)))))$R[1, 2]
  expect_equal(up, -dn, tolerance = 1e-12)
})

test_that("the continuity correction preserves the table margins", {
  # The correction adds 0.5 to the empty cell AND to its diagonal opposite, and takes 0.5 out of
  # the other two, so every row and column sum is unchanged. That is what keeps the shared
  # per-variable thresholds valid, and it is checked here independently: with the margins fixed a
  # 2x2 has one degree of freedom, so the two-step estimate is the rho solving
  # Phi2(t1, t2, rho) = (a - 0.5) / N with t1, t2 taken from the ORIGINAL margins. Agreement to
  # quadrature accuracy shows the margins -- hence the thresholds -- really were left alone.
  tab <- rbind(c(40, 0), c(60, 200))
  N <- sum(tab)
  t1 <- stats::qnorm(sum(tab[1, ]) / N)
  t2 <- stats::qnorm(sum(tab[, 1]) / N)
  target <- (tab[1, 1] - 0.5) / N
  gold <- stats::uniroot(function(rho) .bvn_rect_cpp(-Inf, t1, -Inf, t2, rho) - target,
                         c(-0.9999, 0.9999), tol = 1e-10)$root
  expect_equal(suppressWarnings(.polychoric(.expand_table(tab)))$R[1, 2], gold, tolerance = 1e-4)
})

test_that("the continuity correction preserves the margins under missing data too", {
  # With missing data each pair uses the thresholds of its OWN pairwise-complete cases, derived
  # from the contingency table -- the corrected one. That is the case the margin-preserving design
  # exists for: a correction that shifted the margins would move the thresholds with them. Build a
  # pair whose complete cases form a 2x2 with one empty cell, add missing values in a third
  # variable so the local-threshold branch is taken, and check the pair still matches the estimate
  # implied by its ORIGINAL margins.
  tab <- rbind(c(40, 0), c(60, 200))
  xy <- .expand_table(tab)
  x <- cbind(xy, rep(c(1L, 2L, NA), length.out = nrow(xy)))
  colnames(x) <- c("a", "b", "z")

  res <- suppressWarnings(.polychoric(x))
  expect_true(res$zero_corrected[["a-b"]])

  N <- sum(tab)
  t1 <- stats::qnorm(sum(tab[1, ]) / N)
  t2 <- stats::qnorm(sum(tab[, 1]) / N)
  gold <- stats::uniroot(
    function(rho) .bvn_rect_cpp(-Inf, t1, -Inf, t2, rho) - (tab[1, 1] - 0.5) / N,
    c(-0.9999, 0.9999), tol = 1e-10)$root
  expect_equal(res$R[1, 2], gold, tolerance = 1e-4)
})

test_that("the corrected tetrachoric matches lavaan, which applies the same correction", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  # lavaan corrects a 2x2 zero cell by default (`zero.add = c(0.5, 0)`) and keeps the margins
  # (`zero.keep.margins`), which is exactly the convention implemented here -- so the estimates
  # must agree. With `zero.add = 0` lavaan instead returns its own `maxcor` of 0.999, confirming
  # that the two estimators differ only by the correction and not in the estimator itself.
  tab <- rbind(c(40, 0), c(60, 200))
  d <- as.data.frame(.expand_table(tab))
  names(d) <- c("a", "b")
  d[] <- lapply(d, ordered)

  ours <- suppressWarnings(.polychoric(.expand_table(tab)))$R[1, 2]
  ref <- lavaan::lavCor(d, ordered = names(d), output = "cor")[1, 2]
  expect_equal(ours, ref, tolerance = 1e-4)
  # Regression pin, so a change in the correction is caught even without lavaan installed.
  expect_equal(ours, 0.9276913, tolerance = 1e-6)
})

# Collect the class strings of every warning raised while evaluating `expr`, so one expected class
# can be asserted (and another denied) when several conditions fire on the same data.
.poly_warn_classes <- function(expr) {
  cls <- character()
  withCallingHandlers(
    expr,
    warning = function(w) { cls <<- c(cls, class(w)); invokeRestart("muffleWarning") }
  )
  cls
}

# Variables with structurally empty cells whose expected counts are large (32), estimated well
# inside the interior (rho = 0.82) rather than at a Frechet bound: a common ordinal score nudged up
# by each variable's own 0/1 pattern and capped at the top category. Reaching the lowest category
# needs an unnudged zero score and the highest a nudged or already-top score, so cells (0, 2) and
# (2, 0) are both empty in every pair's table -- a pattern no extreme coupling has, which is what
# keeps the pairs off the bound.
.sparse_interior <- function(n_var = 2L) {
  r <- 0:383L
  u <- (r %/% 64L) %% 3L
  out <- vapply(seq_len(n_var),
                function(k) pmin(2L, u + bitwAnd(bitwShiftR(r, k - 1L), 1L)),
                integer(length(r)))
  colnames(out) <- paste0("v", seq_len(n_var))
  out
}

test_that("a structurally empty cell warns when an asymptotic covariance is requested", {
  # An empty cell with a non-negligible expected count is flagged; an otherwise dense ordinal set
  # whose only empty cells are rare corners is not. The covariance is still returned.
  sparse <- .sparse_interior()
  expect_warning(.polychoric(sparse, acov = "diag"), class = "efa_cor_sparse_cells")
  expect_no_warning(
    .polychoric(DOSPERT_raw[, 1:8], acov = "diag"),
    class = "efa_cor_sparse_cells"
  )
  # The per-replicate bootstrap recompute (label_acov = FALSE) skips the diagnostic entirely.
  expect_no_warning(
    .polychoric(sparse, acov = "diag", label_acov = FALSE),
    class = "efa_cor_sparse_cells"
  )
  # A pair at a Frechet bound has empty cells too, but it is reported as a boundary solution
  # instead: that warning says strictly more (there is no asymptotic covariance at all, rather
  # than an unreliable one), so repeating the weaker sparse-cell diagnosis for it would only
  # obscure the diagnosis.
  bound <- .expand_table(rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47)))
  cls <- .poly_warn_classes(.polychoric(bound, acov = "diag"))
  expect_true("efa_cor_boundary" %in% cls)
  expect_false("efa_cor_sparse_cells" %in% cls)
  # The bootstrap recompute is silent about the boundary too.
  expect_no_warning(.polychoric(bound, acov = "diag", label_acov = FALSE))
})

test_that("the sparse-cell warning names the offending pairs and caps the list", {
  testthat::local_reproducible_output()

  # The warning itself is the subject here, so catch it rather than snapshotting whatever
  # else the fixture emits.
  sparse_msg <- function(x) {
    w <- tryCatch(.polychoric(x, acov = "diag"),
                  efa_cor_sparse_cells = function(w) w)
    expect_s3_class(w, "efa_cor_sparse_cells")
    conditionMessage(w)
  }

  one <- .sparse_interior()
  colnames(one) <- c("risk", "gamble")
  expect_snapshot(cat(sparse_msg(one)))

  # The same construction over six items: cells (0, 2) and (2, 0) are structurally empty in all
  # 15 tables, each with an expected count of 32, and the nudges keep every pair well short of
  # perfect dependence (rho = 0.82) so none of them lands on a Frechet bound -- which the naive
  # monotone construction would. The reported list is capped at five plus a count.
  nudged <- .sparse_interior(6L)
  expect_snapshot(cat(sparse_msg(nudged)))
})

test_that("the boundary warning names the offending pairs and caps the list", {
  testthat::local_reproducible_output()

  boundary_msg <- function(x) {
    w <- tryCatch(.polychoric(x, acov = "diag"), efa_cor_boundary = function(w) w)
    expect_s3_class(w, "efa_cor_boundary")
    conditionMessage(w)
  }

  upper <- rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47))
  one <- .expand_table(upper)
  colnames(one) <- c("risk", "gamble")
  expect_snapshot(cat(boundary_msg(one)))

  # A perfectly REVERSED table is estimated at the negative endpoint, so the snapshot must show
  # -0.9999. Reporting the positive value here, or describing the pair as perfectly ordered, would
  # contradict the correlation the same call returns.
  rev <- .expand_table(upper[, 2:1])
  colnames(rev) <- c("risk", "gamble")
  expect_snapshot(cat(boundary_msg(rev)))

  # Both directions in one matrix: the report names both values rather than picking one. `gamble`
  # is monotone increasing in `risk` and `mirror` is its reverse, so the two pairs involving
  # `mirror` sit on the opposite bound from the `risk`-`gamble` pair (the `gamble`-`mirror` table
  # has two empty cells, so it is not continuity-corrected and stays a boundary pair).
  both <- cbind(one, mirror = max(one[, 2L]) - one[, 2L])
  expect_snapshot(cat(boundary_msg(both)))
})

test_that("the boundary warning quotes the estimate, not the nearest-PD projection of it", {
  # The reported value comes from the backend's own boundary estimate rather than from the returned
  # matrix, because `nearest_pd` projects the matrix afterwards and moves the entry off the
  # endpoint. Without this the message would quote the projected value (measured here: 0.9954538
  # instead of 0.9999), so the check is what keeps the estimate and the message from drifting apart.
  x <- cbind(.expand_table(rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47))),
             .sparse_interior(2L)[seq_len(300L), ])
  colnames(x) <- paste0("v", 1:4)

  projected <- suppressWarnings(.polychoric(x, nearest_pd = TRUE))$R[1, 2]
  expect_lt(projected, 0.9999)                      # the projection really did move it

  # The projection also warns that it happened; that is asserted elsewhere, so muffle it here and
  # let only the boundary report through.
  w <- tryCatch(
    suppressWarnings(.polychoric(x, nearest_pd = TRUE), classes = "efa_cor_smoothed"),
    efa_cor_boundary = function(w) w)
  expect_s3_class(w, "efa_cor_boundary")
  expect_snapshot(cat(conditionMessage(w)))
})

test_that("the continuity-correction warning names the offending pairs and caps the list", {
  testthat::local_reproducible_output()

  zero_cell_msg <- function(x) {
    w <- tryCatch(.polychoric(x, acov = "diag"), efa_cor_zero_cell = function(w) w)
    expect_s3_class(w, "efa_cor_zero_cell")
    conditionMessage(w)
  }

  # Seven binary items that are all monotone in one another (an item is answered 1 only if every
  # easier item is), so every one of the 21 tables has a single empty cell. Being 2x2 they take
  # the continuity correction rather than the boundary report, which is what separates this case
  # from the larger table above. The reported list is capped at five plus a count.
  guttman <- vapply(1:7, function(k) as.integer((0:279L) %% 8L >= k), integer(280L))
  colnames(guttman) <- paste0("i", 1:7)
  expect_snapshot(cat(zero_cell_msg(guttman)))

  # Two monotone binary items give exactly one corrected pair, which is the singular branch of
  # every plural in the message. The correction constant is interpolated as a number, and a
  # number sets the pluralisation quantity for everything after it, so the last plural of the
  # opening line used to read "their correlation" over a single pair.
  one <- cbind(a = as.integer((0:199L) %% 4L >= 1L),
               b = as.integer((0:199L) %% 4L >= 2L))
  expect_snapshot(cat(zero_cell_msg(one)))

  # The pairs are corrected, not bounded, and none of them carries a variance.
  res <- suppressWarnings(.polychoric(guttman, acov = "diag"))
  expect_true(all(res$zero_corrected))
  expect_false(any(res$at_bound))
  expect_true(all(is.na(res$acov)))
})

test_that("a likely-continuous (many-category) variable warns", {
  x <- cbind(rep(1:12, length.out = 100L), rep(1:3, length.out = 100L))
  expect_warning(.polychoric(x), class = "efa_cor_many_categories")
})

test_that("polychoric matches polycor on DOSPERT and UPPS", {
  skip_on_cran()
  skip_if_not_slow()
  skip_if_not_installed("polycor")

  for (data in list(DOSPERT_raw, UPPS_raw)) {
    x <- data[stats::complete.cases(data), ]
    ours <- .polychoric(x)$R
    expect_lt(max(abs(ours - .ref_polychor(x))), 1e-4)
    psych_rho <- suppressWarnings(
      psych::polychoric(x, correct = FALSE, global = FALSE)$rho)
    expect_lt(max(abs(ours - unname(psych_rho))), 1e-4)
  }
})

# --- Asymptotic covariance of the polychoric correlations (acov = "diag" / "full") -------
#
# The two-step ACOV accounts for the estimated thresholds (Muthen, 1984; Joreskog, 1994);
# it is validated against lavaan's NACOV and an independent Monte-Carlo simulation. It is
# NOT checked against polycor::polychor(std.err = TRUE): that returns the threshold-FIXED
# conditional variance 1/(N * I_rho), which omits threshold-estimation uncertainty and runs
# ~7-23% below the true variance, so it is the wrong reference for the threshold-corrected
# quantity DWLS and robust SEs require (and that diag = diag(full) mandates).

# lavaan's variance-scale NACOV for the off-diagonal correlations, labelled "Vi-Vj" to match
# .polychoric()'s pair order. lavaan reports the per-observation Gamma, so divide by N.
.lav_acov <- function(x) {
  df <- as.data.frame(lapply(as.data.frame(x), ordered))
  fit <- lavaan::lavCor(df, ordered = names(df), output = "fit", se = "robust.sem")
  G <- lavaan::lavInspect(fit, "gamma")
  ci <- grep("~~", rownames(G))
  out <- G[ci, ci, drop = FALSE] / nrow(x)
  lab <- gsub("~~", "-", rownames(out))
  dimnames(out) <- list(lab, lab)
  out
}

# Independent reference for one rho's asymptotic variance: the same two-step (Muthen/Joreskog)
# outer-product sandwich, but the bread/meat assembly (A11/A21/A22/IF) is rebuilt from scratch
# in plain R, and the rho-derivative uses the bivariate-normal density at the four corners
# (Plackett, 1954) -- a different formula than the backend's conditioning integral. This
# catches an error in the C++ influence-function assembly that diag == diag(full) cannot. The
# cell probabilities (via .bvn_rect_cpp) and the thresholds (qnorm of the cumulative
# proportions) are deliberately the SAME as the backend, so this guards the assembly while the
# lavaan cross-checks independently guard the cell-probability quadrature. Inputs are 0-based
# category codes for one complete-data pair.
.if_diag_pair <- function(a, b) {
  N <- length(a)
  Ki <- max(a) + 1L; Kj <- max(b) + 1L
  ti <- stats::qnorm(cumsum(tabulate(a + 1L, Ki)) / N)[-Ki]
  tj <- stats::qnorm(cumsum(tabulate(b + 1L, Kj)) / N)[-Kj]
  rcut <- c(-Inf, ti, Inf); ccut <- c(-Inf, tj, Inf)
  rho <- .polychoric(cbind(a, b))$R[1, 2]; s <- sqrt(1 - rho^2)
  dbn <- function(xv, yv) if (is.finite(xv) && is.finite(yv)) {
    exp(-(xv^2 - 2 * rho * xv * yv + yv^2) / (2 * (1 - rho^2))) / (2 * pi * s)
  } else 0
  P <- dP <- matrix(0, Ki, Kj)
  for (aa in seq_len(Ki)) for (bb in seq_len(Kj)) {
    P[aa, bb]  <- .bvn_rect_cpp(rcut[aa], rcut[aa + 1L], ccut[bb], ccut[bb + 1L], rho)
    dP[aa, bb] <- dbn(rcut[aa + 1L], ccut[bb + 1L]) - dbn(rcut[aa], ccut[bb + 1L]) -
                  dbn(rcut[aa + 1L], ccut[bb]) + dbn(rcut[aa], ccut[bb])
  }
  Pf <- pmax(P, .Machine$double.xmin); dxr <- dP / Pf
  n_ab <- matrix(0, Ki, Kj)
  for (r in seq_len(N)) n_ab[a[r] + 1L, b[r] + 1L] <- n_ab[a[r] + 1L, b[r] + 1L] + 1
  A22 <- sum(n_ab * dxr^2)
  A21i <- vapply(seq_len(Ki - 1L), function(k) {
    Bc <- stats::pnorm((ccut[-1L] - rho * ti[k]) / s) -
          stats::pnorm((ccut[-(Kj + 1L)] - rho * ti[k]) / s)
    stats::dnorm(ti[k]) * sum(Bc * (n_ab[k, ] * dxr[k, ] / Pf[k, ] -
                                    n_ab[k + 1L, ] * dxr[k + 1L, ] / Pf[k + 1L, ]))
  }, numeric(1))
  A21j <- vapply(seq_len(Kj - 1L), function(k) {
    Bc <- stats::pnorm((rcut[-1L] - rho * tj[k]) / s) -
          stats::pnorm((rcut[-(Ki + 1L)] - rho * tj[k]) / s)
    stats::dnorm(tj[k]) * sum(Bc * (n_ab[, k] * dxr[, k] / Pf[, k] -
                                    n_ab[, k + 1L] * dxr[, k + 1L] / Pf[, k + 1L]))
  }, numeric(1))
  bread <- function(K, tau, m) {                     # IFth = S (A11)^{-1}, A11 = S' diag(m) S
    S <- matrix(0, K, K - 1L)
    for (aa in 0:(K - 1L)) {
      ipm <- if (m[aa + 1L] > 0) N / m[aa + 1L] else 0
      if (aa <= K - 2L) S[aa + 1L, aa + 1L] <- S[aa + 1L, aa + 1L] + stats::dnorm(tau[aa + 1L]) * ipm
      if (aa >= 1L)     S[aa + 1L, aa]       <- S[aa + 1L, aa]       - stats::dnorm(tau[aa]) * ipm
    }
    S %*% solve(t(S) %*% diag(m) %*% S)
  }
  Ti <- as.numeric(bread(Ki, ti, tabulate(a + 1L, Ki)) %*% A21i)
  Tj <- as.numeric(bread(Kj, tj, tabulate(b + 1L, Kj)) %*% A21j)
  IF <- (dxr - outer(Ti, rep(1, Kj)) - outer(rep(1, Ki), Tj)) / A22
  sum(n_ab * IF^2)
}

test_that("acov diag/full are well-formed and mutually consistent", {
  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), 1:6]
  p <- ncol(x)
  pstar <- p * (p - 1L) / 2L
  labels <- apply(utils::combn(colnames(x), 2L), 2L, paste, collapse = "-")

  d <- .polychoric(x, acov = "diag")
  f <- .polychoric(x, acov = "full")

  # shapes + labelling
  expect_length(d$acov, pstar)
  expect_named(d$acov, labels)
  expect_equal(dim(f$acov), c(pstar, pstar))
  expect_equal(dimnames(f$acov), list(labels, labels))

  # the diagonal is a valid set of variances; the cheap diag path equals diag(full) (a
  # structural check that the per-cell influence is scattered/cross-multiplied correctly)
  expect_true(all(is.finite(d$acov)))
  expect_true(all(d$acov > 0))
  expect_equal(unname(d$acov), unname(diag(f$acov)), tolerance = 1e-10)

  # full Gamma is symmetric and positive-semidefinite (it is a Gram matrix by construction)
  expect_true(isSymmetric(f$acov))
  expect_gt(min(eigen(f$acov, symmetric = TRUE, only.values = TRUE)$values), -1e-8)

  # default is matrix-only; a bad level is rejected (base match.arg, as elsewhere in the pkg)
  expect_null(.polychoric(x)$acov)
  expect_error(.polychoric(x, acov = "nope"))
})

test_that("a Frechet-bound pair has no asymptotic covariance and contaminates no other pair", {
  # A boundary pair has no asymptotic variance to report: the rho-information vanishes identically
  # at the bound, so the influence function (score - threshold correction) / A22 is 0/0 -- both
  # parts underflow to zero well before the boundary is reached, and the limit of the ratio is
  # infinite. Recomputed from scratch at high accuracy, A22 falls from 1.6e-03 at rho = 0.90 to
  # 3.8e-16 at 0.97 and to exactly 0 beyond, so any finite value here is an artefact of the
  # probability floor rather than a variance. It is fail-closed to NA at both levels, which is what
  # makes the branch every consumer takes independent of the platform.
  #
  # The other five pairs are ordinary and must come through untouched: the boundary pair's
  # influence column is zeroed before the crossprod precisely so its NAs cannot spread into them.
  x <- cbind(.expand_table(rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47))),
             .sparse_interior(2L)[seq_len(300L), ])
  colnames(x) <- paste0("v", 1:4)
  d <- suppressWarnings(.polychoric(x, acov = "diag"))
  f <- suppressWarnings(.polychoric(x, acov = "full"))$acov

  bnd <- d$at_bound
  expect_identical(unname(which(bnd)), 1L)                 # the v1-v2 pair, in combn order
  expect_true(all(is.na(d$acov[bnd])))
  expect_true(all(is.finite(d$acov[!bnd]) & d$acov[!bnd] > 0))

  # The full level withholds the boundary pair's whole row and column -- its covariance with every
  # other pair rests on the same vanishing information -- and leaves the rest estimable.
  expect_true(all(is.na(f[bnd, ])) && all(is.na(f[, bnd])))
  expect_true(all(is.finite(f[!bnd, !bnd])))

  # The cheap diagonal still equals diag(full), NA entries included: the two are computed by
  # different routes (a direct cell sum versus scatter-then-crossprod).
  expect_equal(unname(d$acov), unname(diag(f)), tolerance = 1e-6)
})

test_that("a continuity-corrected pair has no asymptotic covariance at either level", {
  # A corrected pair is withheld for a different reason than a boundary pair -- its estimate is
  # interior, but it comes from a nudged table the sandwich would treat as data -- and it is
  # withheld at BOTH levels, not just the cheap diagonal. The other pairs must still come through
  # estimable, which is what the zeroed influence column before the crossprod is for.
  x <- cbind(.expand_table(rbind(c(40, 0), c(60, 200))),
             .sparse_interior(2L)[seq_len(300L), ])
  colnames(x) <- paste0("v", 1:4)
  d <- suppressWarnings(.polychoric(x, acov = "diag"))
  f <- suppressWarnings(.polychoric(x, acov = "full"))$acov

  zc <- d$zero_corrected
  expect_identical(unname(which(zc)), 1L)                  # the v1-v2 pair, in combn order
  expect_false(any(d$at_bound))
  expect_true(all(is.na(d$acov[zc])))
  expect_true(all(is.finite(d$acov[!zc]) & d$acov[!zc] > 0))

  expect_true(all(is.na(f[zc, ])) && all(is.na(f[, zc])))
  expect_true(all(is.finite(f[!zc, !zc])))
  expect_equal(unname(d$acov), unname(diag(f)), tolerance = 1e-6)
})

test_that("a near-collinear pair with structurally empty tail cells still has a variance", {
  skip_on_cran()

  # A pair correlated this strongly has empty cells whose model probability is astronomically
  # small -- at rho = 0.999 some are below the smallest representable double outright. Those
  # cells carry no cases, so they must contribute exactly nothing to the influence function: the
  # asymptotic variance has to come out finite and positive, and the cheap diagonal has to agree
  # with the crossprod route. The pair is estimated in the interior (its table is off both
  # Frechet bounds and it is not binary), so nothing upstream withholds the variance and the
  # whole assembly is exercised. The estimates themselves are pinned against the exact two-step
  # value elsewhere; this covers the ACOV those same tables never requested.
  #
  # Same seed and same four rt values as that estimate test, so these are the same four tables:
  # each draw depends on the ones before it, and a shorter loop would silently substitute
  # different data.
  cut7 <- stats::qnorm(seq_len(6) / 7)
  set.seed(101)
  N <- 4000
  for (rt in c(0.97, 0.99, 0.995, 0.999)) {
    L <- chol(matrix(c(1, rt, rt, 1), 2L))
    Z <- matrix(stats::rnorm(2 * N), N, 2L) %*% L
    a <- findInterval(Z[, 1], cut7); b <- findInterval(Z[, 2], cut7)
    d <- suppressWarnings(.polychoric(cbind(a, b), acov = "diag"),
                          classes = "efa_cor_sparse_cells")
    f <- suppressWarnings(.polychoric(cbind(a, b), acov = "full"),
                          classes = "efa_cor_sparse_cells")
    expect_false(d$at_bound[[1]])
    expect_false(d$zero_corrected[[1]])
    expect_true(is.finite(d$acov[[1]]) && d$acov[[1]] > 0)
    expect_equal(unname(d$acov), unname(diag(f$acov)), tolerance = 1e-10)
  }
})

# Deterministic ordinal data whose alphabetical label order differs from the response order:
# a bivariate normal (rho = 0.7) cut into four categories 1..4, then labelled in the TRUE
# order never < rarely < often < always. Alphabetical sorting scrambles that to
# always, never, often, rarely, so an unordered factor / character column would attenuate rho.
.labelled_ordinal <- function() {
  set.seed(123)
  N <- 600L
  z1 <- stats::rnorm(N)
  z2 <- 0.7 * z1 + sqrt(1 - 0.7^2) * stats::rnorm(N)
  br <- stats::qnorm(c(.25, .5, .75))
  a <- findInterval(z1, br) + 1L
  b <- findInterval(z2, br) + 1L
  lab <- c("never", "rarely", "often", "always")        # the true response order
  list(
    num = data.frame(a = a, b = b),
    ord = data.frame(a = ordered(lab[a], levels = lab), b = ordered(lab[b], levels = lab)),
    uno = data.frame(a = factor(lab[a]), b = factor(lab[b])),
    chr = data.frame(a = lab[a], b = lab[b], stringsAsFactors = FALSE)
  )
}

test_that("ordered factors match numeric codes but unordered / character columns are rejected", {
  d <- .labelled_ordinal()

  # An ordered factor carries the response order, so it gives exactly the numeric-code answer.
  expect_equal(.polychoric(d$ord)$R, .polychoric(d$num)$R)

  # An unordered factor or character column has no response order; data.matrix() would rank it
  # alphabetically and silently attenuate the correlation, so it is refused up front.
  expect_error(.polychoric(d$uno), class = "efa_cor_unordered_factor")
  expect_error(.polychoric(d$chr), class = "efa_cor_unordered_factor")
  expect_error(.polychoric(as.matrix(d$chr)), class = "efa_cor_unordered_factor")
})

test_that("acov diag matches an independent influence-function computation", {
  # Simulated equicorrelated ordinal items with well-populated categories (so the threshold
  # bread is non-singular); compares the backend's diagonal to the from-scratch R reference.
  set.seed(3)
  N <- 1500L; pp <- 4L
  Lc <- chol(matrix(0.4, pp, pp) + diag(0.6, pp))
  Z <- matrix(stats::rnorm(N * pp), N, pp) %*% Lc
  thr <- list(stats::qnorm(c(.3, .65)), stats::qnorm(c(.25, .5, .75)),
              stats::qnorm(c(.2, .5, .8)), stats::qnorm(c(.4, .7)))
  x <- vapply(seq_len(pp), function(j) findInterval(Z[, j], thr[[j]]), integer(N))

  d <- .polychoric(x, acov = "diag")$acov
  codes <- apply(x, 2L, function(col) match(col, sort(unique(col))) - 1L)
  ref <- numeric(length(d)); t <- 1L
  for (i in seq_len(pp - 1L)) for (j in (i + 1L):pp) {
    ref[t] <- .if_diag_pair(codes[, i], codes[, j]); t <- t + 1L
  }
  expect_equal(unname(d), ref, tolerance = 1e-5)
})

test_that("acov uses the listwise-complete rows consistently under missingness", {
  g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), 1:6]
  gm <- g
  gm[1:60, 1] <- NA
  gm[61:110, 2] <- NA

  res <- suppressWarnings(
    .polychoric(gm, acov = "full"),
    classes = "efa_cor_sparse_cells"
  )
  lw <- .polychoric(gm[stats::complete.cases(gm), ])

  # the returned matrix and the ACOV come from the SAME (listwise) case set
  expect_equal(res$R, lw$R)
  expect_true(all(is.finite(res$acov)))
  expect_equal(
    diag(res$acov),
    suppressWarnings(
      .polychoric(gm, acov = "diag"),
      classes = "efa_cor_sparse_cells"
    )$acov,
    ignore_attr = TRUE
  )
})

test_that("acov aborts when there are too few listwise-complete cases", {
  m <- 60L
  # the two columns are never observed on the same row: no listwise-complete case exists
  X <- cbind(c(rep(1:3, length.out = m), rep(NA, m)),
             c(rep(NA, m), rep(1:3, length.out = m)))
  expect_error(.polychoric(X, acov = "diag"), class = "efa_cor_no_complete_cases")
  expect_error(.polychoric(X, acov = "full"), class = "efa_cor_no_complete_cases")
  # without an ACOV the same data is handled pairwise, failing as an uncomputable pair instead
  expect_error(.polychoric(X), class = "efa_cor_na")
})

test_that("acov full and diag match lavaan's NACOV on DOSPERT and UPPS", {
  skip_on_cran()
  skip_if_not_installed("lavaan")

  for (data in list(DOSPERT_raw, UPPS_raw)) {
    x <- data[stats::complete.cases(data), 1:6]
    f <- suppressWarnings(
      .polychoric(x, acov = "full"),
      classes = "efa_cor_sparse_cells"
    )$acov
    lav <- suppressWarnings(.lav_acov(x))
    ord <- match(rownames(lav), rownames(f))
    ours <- f[ord, ord]

    # The residual is our two-step rho vs lavaan's optimiser plus the cancellation-free vs
    # corner cell-probability derivatives. Check the variances relatively and the (tiny,
    # ~1e-4) off-diagonal covariances absolutely -- mixing the two scales in one absolute
    # cap would let the variance-scale diagonal dominate. Measured agreement is ~3e-5 (diag,
    # relative) and ~1e-8 (off-diagonal, absolute), so these still catch a real sign/scale
    # error while leaving headroom against lavaan/BLAS version drift.
    expect_lt(max(abs(diag(ours) - diag(lav)) / diag(lav)), 1e-3)
    expect_lt(max(abs(ours[upper.tri(ours)] - lav[upper.tri(lav)])), 1e-7)
  }
})

test_that("acov full matches lavaan's NACOV on the full item sets", {
  skip_on_cran()
  skip_if_not_slow()
  skip_if_not_installed("lavaan")

  for (data in list(DOSPERT_raw, UPPS_raw)) {
    x <- data[stats::complete.cases(data), ]
    f <- suppressWarnings(
      .polychoric(x, acov = "full"),
      classes = "efa_cor_sparse_cells"
    )$acov
    lav <- suppressWarnings(.lav_acov(x))
    ord <- match(rownames(lav), rownames(f))
    ours <- f[ord, ord]
    expect_lt(max(abs(diag(ours) - diag(lav)) / diag(lav)), 2e-3)
    expect_lt(max(abs(ours[upper.tri(ours)] - lav[upper.tri(lav)])), 1e-6)
  }
})

test_that("acov diag recovers the Monte-Carlo sampling variance of rho", {
  skip_on_cran()
  skip_if_not_slow()

  # Discretise a known bivariate normal into well-populated 4-category items; the analytic
  # diagonal variance must track the empirical Var(rho-hat) over many fresh samples. The
  # tolerance covers the Monte-Carlo error (~2% at R = 4000) plus the small outer-product
  # bias, with margin against RNG/platform drift.
  rho_true <- 0.5
  thr <- stats::qnorm(c(.25, .5, .75))
  disc <- function(z) findInterval(z, thr)
  N <- 2000L
  Lc <- chol(matrix(c(1, rho_true, rho_true, 1), 2L))

  set.seed(7)
  Z <- matrix(stats::rnorm(2 * N), N, 2L) %*% Lc
  analytic <- .polychoric(cbind(disc(Z[, 1]), disc(Z[, 2])), acov = "diag")$acov[[1]]

  set.seed(99)
  R <- 4000L
  rho_hat <- numeric(R)
  for (r in seq_len(R)) {
    Zb <- matrix(stats::rnorm(2 * N), N, 2L) %*% Lc
    rho_hat[r] <- .polychoric(cbind(disc(Zb[, 1]), disc(Zb[, 2])))$R[1, 2]
  }
  expect_lt(abs(analytic / stats::var(rho_hat) - 1), 0.15)
})

#' Schmid-Leiman transformation
#'
#' This function implements the Schmid-Leiman (SL) transformation
#' (Schmid & Leiman, 1957). It takes the pattern coefficients and factor
#' intercorrelations from an oblique factor solution as
#' input and can reproduce the results from [psych::schmid()]
#' and from the SPSS implementation from Wolff & Preising (2005). Other arguments
#' from [efa_fit()] can be used to control the procedure to find the
#' second-order loadings more flexibly. The function can also be used on a
#' second-order confirmatory factor analysis (CFA) solution from lavaan.
#' The group factors of the returned solution are sorted and relabelled, so their
#' column order can differ from the input solution's (see Details).
#'
#' @param x object of class [efa_fit()], class [psych::fa()],
#' class `lavaan::lavaan()`, a matrix, or an `efa_loadings`/`loadings` object. If class [efa_fit()] or
#' class [psych::fa()], pattern coefficients and factor
#' intercorrelations are taken from this object. If class `lavaan::lavaan()`,
#' it must be a second-order CFA solution. In this case first-order and second-order
#'  factor loadings are taken from this object and the `g_name` argument has
#'  to be specified.
#' x can also be a pattern matrix from an oblique factor solution (see `Phi`).
#' @param Phi matrix. A matrix of factor intercorrelations from an oblique factor
#' solution. Only needs to be specified if a pattern matrix is entered directly
#' into `x`.
#' @param estimator character. One of "PAF", "ML", or "ULS" to use
#' principal axis factoring, maximum likelihood, or unweighted least squares,
#' respectively, used in [efa_fit()] to find the second-order loadings. "MINRES" is
#' accepted as a synonym for "ULS" (the same estimator).
#' @param g_name character. The name of the general factor. This needs only be
#' specified if `x` is a `lavaan` second-order solution. Default is "g".
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  second-order [efa_fit()] fit, including the `type` preset. `NULL` (default) uses the
#'  [efa_fit()] defaults. The second-order fit is unrotated, so no rotation settings apply.
#' @param ... Arguments to be passed to [efa_fit()]. The estimation tuning knobs are not passed
#' here; they live in `estimate_control`, and the standard-error arguments (`se`, `b_boot`,
#' `ci`, `seed`) are not accepted because the second-order fit is an internal step run against
#' a placeholder sample size and only its loadings are kept.
#'
#' @details
#' The SL transformation (also called SL orthogonalization) is a procedure with
#' which an oblique factor solution is transformed into a hierarchical,
#' orthogonalized solution. As a first step, the factor intercorrelations are
#' factor analyzed to extract a single second-order (general) factor, yielding a
#' two-level hierarchical structure. The first-order factor and the second-order
#' factor are then orthogonalized, resulting in an orthogonalized factor solution
#' with proportionality constraints. The procedure thus makes a suggested
#' hierarchical data structure based on factor intercorrelations explicit. One
#' major advantage of SL transformation is that it enables variance
#' partitioning between higher-order and first-order factors, including the
#' calculation of McDonald's omegas (see [efa_reliability()]).
#'
#' Where the first-order factors come from a loading matrix -- an [efa_fit()] or a
#' [psych::fa()] solution, or a pattern matrix supplied with `Phi` -- they are sorted
#' by the number in their column labels, so that `"F10"` follows `"F2"` rather than
#' `"F1"`. The sort needs a number in every column label; columns that carry no
#' labels, or a label without a number, keep the order they arrive in. A second-order
#' `lavaan` solution is not sorted at all: its first-order factors keep the order the
#' model declares them in.
#'
#' The columns are then labelled `"F1"` to `"Fk"` by position, on every route, and the
#' input solution's own factor names are not carried over. A factor a `lavaan` model
#' calls `"F3"` can therefore come back as `"F1"`. Where the sort applies it is
#' independent of how the input orders its factors, so the group
#' factors of the returned `sl` matrix can also be in a different order from the columns
#' they came from. A [psych::fa()] solution shows this most readily: it orders its
#' columns by their sums of squared loadings, but keeps each factor's own number in its
#' label, so those numbers arrive out of order. A
#' solution whose columns are `"PA2"`, `"PA3"`, `"PA1"` comes back with those same
#' three factors sorted as `PA1`, `PA2`, `PA3` and labelled `"F1"`, `"F2"`, `"F3"`.
#' The first group factor of the result is then the third column of the input. The
#' same holds against [psych::schmid()], whose columns keep the input order: the two
#' solutions agree column for column only after one of them is permuted to the
#' other's order. An [efa_fit()] solution already labels its factors `"F1"` to `"Fk"`
#' in that order, so nothing moves for one; the reordering shows itself for a
#' [psych::fa()] solution, and for a pattern matrix supplied with labels of its own.
#'
#' Read the group factors from the returned matrix, therefore, rather than from the
#' input. An indicator-to-factor map is matched to the group factors by position, so
#' one built in the input solution's column order lines up only where the columns did
#' not move; where they did, [efa_reliability()] or [OMEGA()] scores each composite
#' against the wrong factor. Build such a map from the `"F1"` to `"Fk"` columns of the
#' returned `sl` matrix instead, which is right on every route. The `fac_names` of
#' [efa_reliability()] are matched by position in the same way, so names given in the
#' input solution's order label the wrong subscales, and do so without any sign.
#'
#' @return A list of class `c("efa_schmid_leiman", "SL")` containing the following
#' \item{orig_R}{Original correlation matrix.}
#' \item{sl}{A matrix with general factor loadings, group factor loadings, communalities,
#' and uniquenesses.}
#' \item{L2}{Second-order factor loadings.}
#' \item{vars_accounted}{A matrix of explained variances and sums of squared loadings.}
#' \item{iter}{The number of iterations needed for convergence in EFA.}
#' \item{convergence}{Integer convergence code of the second-order EFA (0 =
#' converged); `NA` for a lavaan input. See [efa_fit()].}
#' \item{settings}{list. The settings (arguments) used in EFA to get the
#' second-order loadings.}
#'
#' @source Schmid, J. & Leiman, J. M. (1957). The development of hierarchical
#' factor solutions. Psychometrika, 22(1), 53–61. doi:10.1007/BF02289209
#' @source Wolff, H.-G., & Preising, K. (2005). Exploring item and higher order
#' factor structure with the Schmid-Leiman solution: Syntax codes for SPSS and
#' SAS. Behavior Research Methods, 37 , 48–58. doi:10.3758/BF03206397
#'
#' @family factor rotation
#' @family reliability coefficients
#'
#' @export
#'
#' @examples
#' ## Use with an output from the EFAtools::efa_fit function, both with type EFAtools
#' EFA_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' SL_EFAtools <- efa_schmid_leiman(EFA_mod, estimator = "PAF",
#'                                  estimate_control = estimate_control(type = "EFAtools"))
#'
#' \donttest{
#' ## Use with an output from the psych::fa function with type psych
#' fa_mod <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
#'                     fm = "pa", rotate = "Promax")
#' SL_psych <- efa_schmid_leiman(fa_mod, estimator = "PAF",
#'                               estimate_control = estimate_control(type = "psych"))
#' }
#'
#' ## Use more flexibly by entering a pattern matrix and phi directly (useful if
#' ## a factor solution found with another program should be subjected to SL
#' ## transformation)
#'
#' ## For demonstration, take pattern matrix and phi from an EFA output
#' ## This gives the same solution as the first example
#' SL_flex <- efa_schmid_leiman(EFA_mod$rot_loadings, Phi = EFA_mod$Phi, estimator = "PAF",
#'                              estimate_control = estimate_control(type = "EFAtools"))
#'
#' \donttest{
#' ## Use with a lavaan second-order CFA output
#' if (requireNamespace("lavaan", quietly = TRUE)) {
#'
#' # Create and fit model in lavaan (assume all variables have SDs of 1)
#' mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
#'         F2 =~ V7 + V8 + V9 + V10 + V11 + V12
#'         F3 =~ V13 + V14 + V15 + V16 + V17 + V18
#'         g =~ F1 + F2 + F3'
#' fit <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
#'                    sample.nobs = 500, estimator = "ml")
#'
#' SL_lav <- efa_schmid_leiman(fit, g_name = "g")
#'
#' }
#' }
efa_schmid_leiman <- function(x, Phi = NULL,
                              estimator = c("PAF", "ML", "ULS", "MINRES"),
                              g_name = "g", estimate_control = NULL, ...) {

  # Perform argument checks
  .reject_flat_knobs(...names(), fn = "efa_schmid_leiman")
  # The second-order fit is an internal step: of its output only the loadings, the iteration
  # count, the convergence flag and the recorded settings are kept, and it runs against a
  # placeholder N, so a standard error asked for through these dots would be computed from a
  # sample size that is not the data's and then discarded -- reaching nothing but that
  # settings record. It is also unrotated with no bootstrap, so `seed` governs nothing there.
  # The guard must be the same one the retention criteria use rather than a bare
  # .reject_inference_dots(): the dots are spliced into the second-order fit with do.call(),
  # where R partial-matches them against efa_fit()'s formals, so an abbreviation such as
  # `b_b` would arrive as `b_boot` without ever matching the refused names exactly. It is the
  # whitelist here -- every accepted name spelled in full -- that closes that route.
  .reject_unknown_fit_dots(...names(), fn = "efa_schmid_leiman", unrotated = TRUE)
  checkmate::assert_matrix(Phi, null.ok = TRUE)
  .assert_estimate_control(estimate_control)
  estimator <- .match_arg_ci(estimator)
  # "MINRES" is a synonym for "ULS" (same estimator); resolve to the canonical name.
  if (estimator == "MINRES") estimator <- "ULS"
  checkmate::assert_string(g_name)

  if(!inherits(x, c("EFA", "fa", "lavaan", "matrix", "LOADINGS", "loadings"))){

    cli::cli_abort(
      c("{.arg x} must be an {.cls EFA}, {.cls fa}, or {.cls lavaan} object, a matrix, or a {.cls LOADINGS}/{.cls loadings} object.",
        "i" = "From an {.cls efa_average} object, use the averaged loading matrix in
               {.code $loadings$average}, with the averaged factor correlations in
               {.code $Phi$average}."),
      class = "efa_sl_bad_input")

  }

  if(inherits(x, "EFA")) {

    if("Phi" %in% names(x)){

      L1 <- x$rot_loadings
      n_first_fac <- ncol(x$rot_loadings)
      orig_R <- x$orig_R

      if(!is.null(Phi)){
        cli::cli_warn(
          c("{.arg Phi} is specified; the supplied factor intercorrelations are used.",
            "i" = "To use the intercorrelations from the EFA output, leave {.code Phi = NULL}."),
          class = "efa_sl_phi_specified"
        )

      } else {

        Phi <- x$Phi

      }

    } else {

      cli::cli_abort("{.arg x} is a non-rotated or orthogonal factor solution, but SL needs an oblique solution.",
                     class = "efa_sl_not_oblique")

    }

    Phi <- .align_correlation_axis(
      Phi, n = n_first_fac, target_names = colnames(L1), arg = "Phi"
    )
    n_order <- .sl_factor_order(colnames(L1), n_first_fac)
    L1 <- L1[, n_order, drop = FALSE]
    Phi <- Phi[n_order, n_order, drop = FALSE]

  } else if(inherits(x, "fa")) {

    if("Phi" %in% names(x)){

      L1 <- unclass(x$loadings)
      n_first_fac <- ncol(x$loadings)
      orig_R <- unclass(x$r)

      if(!is.null(Phi)){
        cli::cli_warn(
          c("{.arg Phi} is specified; the supplied factor intercorrelations are used.",
            "i" = "To use the intercorrelations from the {.fn psych::fa} output, leave {.code Phi = NULL}."),
          class = "efa_sl_phi_specified"
        )

      } else {

        Phi <- x$Phi

      }

    } else {

      cli::cli_abort("{.arg x} is a non-rotated or orthogonal factor solution, but SL needs an oblique solution.",
                     class = "efa_sl_not_oblique")

    }

    Phi <- .align_correlation_axis(
      Phi, n = n_first_fac, target_names = colnames(L1), arg = "Phi"
    )
    n_order <- .sl_factor_order(colnames(L1), n_first_fac)
    L1 <- L1[, n_order, drop = FALSE]
    Phi <- Phi[n_order, n_order, drop = FALSE]

  } else if(inherits(x, "lavaan")){

    .require_lavaan()

    if(lavaan::lavInspect(x, what = "converged") == FALSE){
      cli::cli_abort("The model did not converge; the Schmid-Leiman transformation is not performed.",
                     class = "efa_sl_no_converge")
    }

    std_sol <- suppressWarnings(lavaan::lavInspect(x, what = "std"))

    if(any(is.na(std_sol$lambda))){
      cli::cli_abort("Some loadings are {.val NA} or {.val NaN}; the Schmid-Leiman transformation is not performed.",
                     class = "efa_sl_na_loadings")
    }

    if(any(std_sol$lambda >= 1)){
      cli::cli_abort("A Heywood case was detected (a loading of 1 or larger); the Schmid-Leiman transformation is not performed.",
                     class = "efa_sl_heywood")
    }

    if(any(diag(std_sol$theta) <= 0) || any(diag(std_sol$psi) <= 0)){
      cli::cli_abort("A Heywood case was detected (a variance of 0 or negative); the Schmid-Leiman transformation is not performed.",
                     class = "efa_sl_heywood")
    }

    # Create list with factor and corresponding subtest names
    col_names <- colnames(std_sol$lambda)

    if(!any(col_names %in% g_name)){
      cli::cli_abort(
        c("Could not find the specified general-factor name in the lavaan solution.",
          "i" = "Please check the spelling."),
        class = "efa_sl_g_name"
      )
    }

    # SL needs a second-order CFA: the general factor must load on the first-order
    # factors, which lavaan stores in the `beta` (latent regression) matrix. A
    # bifactor solution has no such structure (`beta` is empty), so direct the
    # user to efa_reliability() rather than failing later in the loadings algebra.
    if(is.null(std_sol$beta) || !(g_name %in% colnames(std_sol$beta)) ||
       all(std_sol$beta[, g_name] == 0)){
      cli::cli_abort(
        c("{.arg x} does not appear to be a second-order CFA solution.",
          "i" = "{.fn efa_schmid_leiman} needs a second-order model; pass a bifactor solution directly to {.fn efa_reliability}."),
        class = "efa_sl_not_second_order"
      )
    }

    if(!all(std_sol$lambda[, g_name] == 0)){

      cli::cli_warn(
        c("The specified second-order factor contains first-order loadings.",
          "i" = "Did you enter a second-order CFA solution, or the wrong factor name in {.arg g_name}?"),
        class = "efa_sl_second_order_loadings"
      )

    }

    col_names <- col_names[!col_names %in% g_name]
    fac_names <- c(g_name, col_names)

    n_first_fac <- length(col_names)

  } else {

    if(is.null(Phi)){

      cli::cli_abort(
        c("{.arg Phi} was not provided.",
          "i" = "Enter an oblique solution from {.fn efa_fit} or {.fn psych::fa}, a second-order CFA from lavaan, or provide {.arg Phi}."),
        class = "efa_sl_phi_missing"
      )

    }

    Phi <- .align_correlation_axis(
      Phi, n = ncol(x), target_names = colnames(x), arg = "Phi"
    )
    n_order <- .sl_factor_order(colnames(x), ncol(x))
    x <- x[, n_order, drop = FALSE]
    # Phi is guaranteed non-NULL here (the abort above fires otherwise).
    Phi <- Phi[n_order, n_order, drop = FALSE]

    L1 <- x
    n_first_fac <- ncol(x)
    orig_R <- NA

  }

  if(inherits(x, "lavaan")){

    # Calculate direct g loadings
    L1 <- std_sol$lambda[, col_names]
    L2 <- std_sol$beta[col_names, g_name]

    L_sls_2 <- L1 %*% L2

    # Calculate direct group factor loadings (see .sl_group_loadings).
    L_sls_1 <- .sl_group_loadings(L1, std_sol$psi, col_names)

    orig_R <- NA
    iter <- NA
    convergence <- NA
    settings <- NA

  } else {

    # The transformation regresses the first-order factors on a single second-order
    # factor, so it needs at least two of them; with one there is nothing to
    # orthogonalize. Both conditions are reported before the second-order fit runs, so
    # the diagnostic describes the solution the user supplied rather than the internal
    # one-factor fit on a 1 x 1 matrix of intercorrelations.
    if (n_first_fac < 2) {
      cli::cli_abort(
        c("A Schmid-Leiman transformation needs at least two first-order factors.",
          "x" = "{.arg x} has {n_first_fac} first-order factor{?s}."),
        class = "efa_sl_too_few_factors"
      )
    }

    if (n_first_fac == 2) {
      cli::cli_warn("The second-order EFA is underidentified.",
                    class = "efa_sl_underidentified")
    }

    # perform a factor analysis on the intercorrelation matrix of the first order
    # factors (N is only specified to avoid a warning).
    #
    # What that fit reports about the user's solution stays visible: a Heywood case or
    # a failure to converge there makes the residualized first-order loadings, and
    # every coefficient computed from them, questionable. Only the two identification
    # warnings are muffled, because they say nothing about the data: fitting one factor
    # to the k by k matrix of factor intercorrelations is just identified at k = 3 and
    # underidentified at k = 2 by construction, for every input, and the k = 2 case is
    # already reported above in this function's own terms.
    EFA_phi <- withCallingHandlers(
      do.call(efa_fit, c(
        list(Phi, n_factors = 1, N = 100, estimator = estimator, rotation = "none",
             estimate_control = estimate_control),
        list(...))),
      warning = function(w) {
        if (inherits(w, c("efa_just_identified", "efa_underidentified"))) {
          invokeRestart("muffleWarning")
        }
      })

    iter <- EFA_phi$iter
    convergence <- EFA_phi$convergence
    settings <- EFA_phi$settings

    # extract second order loadings
    L2 <- EFA_phi$unrot_loadings

    # Schmid-Leiman solution, direct loadings of second order factor
    L_sls_2 <- L1 %*% L2

    # Communalities of the second-order factor. A value at or above 1 is a
    # Heywood case: the residualized first-order loadings would be undefined
    # (the square root below would be taken of a negative number).
    comm_h <- rowSums(L2^2)

    if(any(comm_h >= 1 + .Machine$double.eps)){
      cli::cli_abort(
        c("A Heywood case was detected in the second-order factor analysis; no Schmid-Leiman solution is computed.",
          "i" = "A second-order communality is 1 or larger, so the residualized first-order loadings are undefined."),
        class = "efa_sl_heywood"
      )
    }

    # compute uniqueness of higher order factor. A communality a hair above 1
    # (within floating-point noise, below the Heywood abort threshold above) is
    # clamped to 1 so the square root never sees a tiny negative value.
    u2_h <- sqrt(pmax(0, 1 - comm_h))

    # Schmid-Leiman solution, residualized first order factor loadings
    L_sls_1 <- L1 %*% diag(u2_h)

  }

  # Combine the Schmid-Leiman loadings in a data frame
  sl_load <- cbind(L_sls_2, L_sls_1)

  # Compute communalities and uniquenesses of the Schmid-Leiman solution
  h2_sl <- rowSums(sl_load^2)
  u2_sl <- 1 - h2_sl

  vars_accounted <- .compute_vars(L_unrot = sl_load, L_rot = sl_load)

  colnames(vars_accounted) <-c("g", paste0("F", seq_len(n_first_fac)))

  # Finalize output object
  sl <- cbind(sl_load, h2_sl, u2_sl)
  colnames(sl) <- c("g", paste0("F", seq_len(n_first_fac)), "h2", "u2")
  class(sl) <- c("efa_sl_loadings", "SLLOADINGS")

  output <- list(
    orig_R = orig_R,
    sl = sl,
    L2 = L2,
    vars_accounted = vars_accounted,
    iter = iter,
    convergence = convergence,
    settings = settings
    )

  class(output) <- c("efa_schmid_leiman", "SL")

  output

}

# Column order of the first-order factors. The loadings are sorted by the factor
# number in the column labels, so that, e.g., "F10" follows "F2" and not "F1".
# The sort needs one label for each factor, and a digit in each label. If one of
# these two conditions is not true, the columns keep the order they have. The
# loadings from `efa_fit()` and `psych::fa()` always have labels; a loading
# matrix that comes directly from the user can have no labels.
.sl_factor_order <- function(nms, n) {

  # covers unlabelled columns too: length(NULL) is 0, so absent labels take the
  # fallback rather than reaching the parse below
  if (length(nms) != n) return(seq_len(n))

  num <- suppressWarnings(as.numeric(gsub("[^0-9]", "", nms)))

  if (anyNA(num)) return(seq_len(n))

  order(num)

}

#' Extract residuals from an efa object
#'
#' Returns the residual correlation matrix of an [efa_fit()] or [efa_mi()]
#' solution. Residuals are a pure extractor here; their diagnostics and a
#' formatted display are part of [summary.efa()].
#'
#' @param object a list of class `efa`. Output from [efa_fit()] or [efa_mi()].
#' @param type character. Which residuals to return. `"raw"` (default) returns
#'   `orig_R - model_implied_R`; `"standardized"` returns the standardized
#'   residuals (residuals divided by their standard errors), available only when
#'   the object was fitted with bootstrap standard errors.
#' @param ... Further arguments (currently unused).
#'
#' @returns A numeric matrix of residual correlations.
#'
#' @export
#' @method residuals efa
#'
#' @examples
#' efa <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500)
#' residuals(efa)
residuals.efa <- function(object, type = c("raw", "standardized"), ...) {
  type <- .match_arg_ci(type)

  if (identical(type, "standardized")) {
    if (is.null(object$standardized_residuals)) {
      cli::cli_abort(
        c(
          "Standardized residuals are not available for this object.",
          "i" = "They require bootstrap standard errors ({.code se = \"np-boot\"})."
        ),
        class = "efa_no_standardized_residuals"
      )
    }
    return(object$standardized_residuals)
  }

  object$residuals
}

## Maximum Likelihood Estimation of Factor Loadings
## Thin fitter: returns the raw ML results; shared post-processing happens in
## .finalize_fit() / .estimate_model().
.ML <- function(x, n_factors, start_method = c("psych", "factanal")) {

  # Get correlation matrix entered or created in EFA
  R <- x

  ml <- .fit_ml(R, n_factors, start_method)

  L <- ml$loadings
  orig_R <- R
  h2 <- rowSums(L^2)
  diag(R) <- h2

  # raw fit, finalized by .estimate_model()
  list(
    L = L,
    h2 = h2,
    psi = ml$res$par,
    Fm = ml$res$value,
    iter = ml$res$counts[1],
    convergence = ml$res$convergence,
    orig_R = orig_R,
    R_final = R,
    settings = list(start_method = start_method)
  )
}

# function to obtain the ML fit; adapted from the psych package
.fit_ml <- function(R, n_fac, start_method) {

  if (start_method == "psych") {
    R.smc <- .smc_start(R)
    # psych::fa()-style start: the uniqueness start diag(R) - smc, or a flat 0.5
    # fallback sized to the variables (the optimiser parameter is the length-p
    # uniqueness vector).
    if((sum(R.smc) == n_fac) && (n_fac > 1)) {
      start <- rep(.5, ncol(R))
    }  else {
      start <- diag(R)- R.smc
    }
  } else if (start_method == "factanal") {
    start <- (1 - 0.5 * n_fac / ncol(R)) / diag(solve(R))
  }

  # Bounded L-BFGS-B over the uniquenesses, run entirely in C++ (roptim drives
  # R's lbfgsb routine, so the box constraints and control match stats::optim).
  ml <- .fit_ml_cpp(R, n_fac, start, .uniqueness_floor)

  res <- list(par = ml$psi, value = ml$Fm, counts = c(ml$iter, NA_integer_),
              convergence = ml$convergence)

  list(loadings = ml$loadings, res = res, R = R)
}

#' Model averaging across different EFA estimators and types
#'
#' Not all EFA procedures always arrive at the same solution. This function allows
#' you perform a number of EFAs from different estimators (e.g., Maximum Likelihood
#' and Principal Axis Factoring), with different implementations (e.g., the SPSS
#' and psych implementations of Principal Axis Factoring), and across different
#' rotations of the same type (e.g., multiple oblique rotations, like promax and
#' oblimin). `efa_average()` will then run all these EFAs (using the [efa_fit()]
#' function) and provide a summary across the different solutions.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations. If raw data is entered, the correlation matrix is found from the
#' data.
#' @param n_factors numeric. Number of factors to extract.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used. If input is a correlation matrix and `N` = NA
#' (default), not all fit indices can be computed.
#' @param estimator character vector. Any combination of  "PAF", "ML", and "ULS",
#' to use principal axis factoring, maximum likelihood, or unweighted least
#' squares, respectively, to fit the EFAs. "MINRES" is accepted as a synonym for
#' "ULS" (the same estimator). Default is "PAF".
#' "DWLS", which [efa_fit()] does accept, is deliberately not
#' offered here: it weights each residual correlation by the inverse of its
#' asymptotic variance, which is only available from raw ordinal data analysed
#' with `cor_method = "poly"` or `"tetra"`, whereas every EFA in the grid is fitted
#' to the single correlation matrix computed once from `x`. Fit a DWLS solution
#' with [efa_fit()] directly.
#' @param rotation character vector. Either perform no rotation ("none"),
#' any combination of orthogonal rotations ("varimax", "equamax", "quartimax", "geominT",
#' "bentlerT", and "bifactorT"; using "orthogonal" runs all of these), or of
#' oblique rotations ("promax", "oblimin", "quartimin", "simplimax", "bentlerQ",
#' "geominQ", and "bifactorQ"; using "oblique" runs all of these). Rotation types
#' (no rotation, orthogonal rotations, and oblique rotations) cannot be mixed.
#' Default is "promax".
#' @param type character vector. Any combination of "none" (default), "EFAtools",
#' "psych", and "SPSS" can be entered. "none" allows the specification of various
#' combinations of the arguments controlling both factor extraction methods and
#' the rotations. The others ("EFAtools", "psych", and "SPSS") take the extraction
#' and rotation tuning of the respective implementation: this package's default
#' procedure, the psych package's, and SPSS's. A specific psych implementation
#' exists for PAF, ML, varimax, and promax. The SPSS implementation exists for
#' PAF, varimax, and promax. For details, see [efa_fit()]. The factor ordering is
#' the one setting a named `type` does not bring here: every solution in the grid
#' is fitted with the eigenvalue-based ordering, so that the solutions can be
#' realigned to a common target before averaging.
#' @param averaging character. One of "mean" (default), and "median". Controls
#' whether the different results should be averaged using the (trimmed) mean,
#' or the median.
#' @param trim numeric. If averaging is set to "mean", this argument controls
#' the trimming of extremes (for details see [base::mean()]).
#' By default no trimming is done (i.e., trim = 0).
#' @param salience_threshold numeric. The threshold to use to classify a pattern
#' coefficient or loading as salient (i.e., substantial enough to assign it to
#' a factor). Default is 0.3. Indicator-to-factor correspondences will be inferred
#' based on this threshold. Note that this may not be meaningful if rotation = "none"
#' and n_factors > 1 are used, as no simple structure is present there.
#' @param max_iter numeric. The maximum number of iterations to perform after which
#' the iterative PAF procedure is halted with a warning. Default is 10,000. It is
#' only evaluated for the "PAF" solutions run under `type` "none": a named `type`
#' brings the iteration cap that defines it ("SPSS" 25, "psych" 50, and
#' "EFAtools" 300), and "ML" and "ULS" do not iterate this way. Note
#' that non-converged procedures are excluded from the averaging procedure.
#' @param init_comm character vector. Any combination of "smc", "mac", and "unity".
#' Controls the methods to estimate the initial communalities in `PAF` if
#' "none" is among the specified types. "smc" will use squared multiple
#' correlations, "mac" will use maximum absolute correlations, "unity" will use
#' 1s (for details see [efa_fit()]). Default is `c("smc", "mac", "unity")`.
#' @param criterion numeric vector. The convergence criterion used for PAF if
#' "none" is among the specified types.
#' If the change in communalities from one iteration to the next is smaller than
#' this criterion the solution is accepted and the procedure ends.
#' Default is `0.001`.
#' @param criterion_type character vector. Any combination of "max_individual" and
#' "sum". Type of convergence criterion used for PAF if "none" is among the
#' specified types. "max_individual" selects the maximum change in any of the
#' communalities from one iteration to the next and tests it against the
#' specified criterion. "sum" takes the difference of
#' the sum of all communalities in one iteration and the sum of all communalities
#' in the next iteration and tests this against the criterion
#' (for details see [efa_fit()]). Default is `c("sum", "max_individual")`.
#' @param abs_eigen logical vector. Any combination of TRUE and FALSE.
#' Which algorithm to use in the PAF iterations if "none" is among the specified
#' types. If FALSE, the loadings are computed from the eigenvalues. This is also
#' used by the [psych::fa()] function. If TRUE the
#' loadings are computed with the absolute eigenvalues as done by SPSS
#' (for details see [efa_fit()]). Default is `TRUE`.
#' @param varimax_type character vector. Any combination of "svd" and "kaiser".
#' The type of the varimax rotation performed if "none" is among the specified
#' types and "varimax", "promax", "orthogonal", or "oblique" is among the specified
#' rotations. "svd" uses singular value decomposition, as
#' [stats::varimax()] does, and "kaiser" uses the varimax
#' procedure performed in SPSS. This is the original procedure from Kaiser (1958),
#' but with slight alterations in the varimax criterion (for details, see
#' [efa_fit()] and Grieder & Steiner, 2022).
#' Default is `c("svd", "kaiser")`.
#' @param normalize logical vector. Any combination of TRUE and FALSE.
#' `TRUE` performs a kaiser normalization before the specified rotation(s).
#' Default is `TRUE`.
#' @param k_promax numeric vector. The power used for computing the target matrix
#' P in the promax rotation if "none" is among the specified types and "promax"
#' or "oblique" is among the specified rotations. Default is `2:4`.
#' @param k_simplimax numeric. The number of 'close to zero loadings' for the
#' simplimax rotation if "simplimax" or "oblique" is among the specified rotations. Default
#' is `ncol(x)`, where x is the entered data. It counts loadings, so each value must be a
#' whole number no larger than the number of loadings in the solution; a simplimax fit
#' given anything else fails and is reported as an errored solution in the grid.
#' @param p_type character vector. Any combination of "norm" and "unnorm".
#' This specifies how the target matrix P is computed in promax rotation if
#' "none" is among the specified types and "promax" or "oblique" is among the
#' specified rotations. "unnorm" will use the unnormalized target matrix as
#' originally done in Hendrickson and White (1964). "norm" will use a
#' normalized target matrix (for details see [efa_fit()]).
#' Default is `c("norm", "unnorm")`.
#' @param precision numeric vector. The tolerance for stopping in the rotation
#' procedure(s). Default is 10^-5.
#' @param start_method character vector. Any combination of "psych" and "factanal".
#' How to specify the starting values for the optimization procedure for ML.
#' "psych" takes the starting values specified in [psych::fa()].
#' "factanal" takes the starting values specified in the
#' [stats::factanal()] function. Default is
#' `c("psych", "factanal")`.
#' @param use character. Passed to [stats::cor()] if raw data
#' is given as input. Default is "pairwise.complete.obs". It is ignored when
#' `cor_method = "fiml"`, which handles the missingness itself, so every case
#' contributes.
#' @param cor_method character. Correlation computed from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (passed to [stats::cor()]), `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations of ordinal / binary data
#'   (a two-step estimator), or `"fiml"`
#'   for a two-stage full-information maximum-likelihood correlation from raw data
#'   with missing values. With `"fiml"` the saturated multivariate-normal mean and
#'   covariance are estimated by an EM algorithm assuming the data are missing at
#'   random and the standardized covariance is analysed, reproducing
#'   `psych::corFiml()` followed by `psych::fa()` and
#'   `lavaan(missing = "two.stage")`, *not* `lavaan::efa(missing = "ml")` (see
#'   [efa_fit()] and the details). Default is "pearson".
#' @param show_progress logical. Whether a progress bar should be shown in the
#' console. Default is TRUE.
#' @param seed numeric or `NULL`. An optional seed making the averaging run
#' reproducible and independent of the number of parallel workers (the grid runs with
#' [future_lapply()][future.apply::future_lapply], for which a parallel plan can be set
#' via [future::plan()]). It matters whenever a criterion-based rotation is included,
#' since those draw random starts. When supplied, the caller's random-number stream is
#' restored afterwards, leaving no side effect. Default is `NULL`.
#' @param P_type `r lifecycle::badge("superseded")` Former name of `p_type`. Still
#' accepted (silently) for backwards compatibility; please use `p_type`.
#'
#' @details
#'
#' As a first step in this function, a grid is produced containing the setting
#' combinations for the to-be-performed EFAs. These settings are then entered as
#' arguments to the [efa_fit()] function and the EFAs are run in a second
#' step. After all EFAs are run, the factor solutions are averaged and their
#' variability determined in a third step.
#'
#' When raw data are supplied, the correlation matrix is computed once before the
#' grid is run and reused for every EFA in it. Under `cor_method = "fiml"` this
#' means the saturated multivariate-normal moments are EM-estimated a single time
#' (from the raw data with missing values, assuming the data are missing at
#' random) and the resulting two-stage correlation is analysed by every solution
#' in the grid; the EM is not re-run per solution. Under `cor_method = "fiml"`,
#' `use` does not select cases (every case contributes to the EM). The averaged
#' loadings and communalities are then the two-stage FIML estimates, but the
#' averaged Chi-Square and the indices derived from it (CFI, TLI, RMSEA, AIC, BIC,
#' ECVI) are the ordinary ML/ULS discrepancy statistics on the EM correlation, not
#' the corrected two-stage (Satorra-Bentler) statistics that a standalone
#' [efa_fit()] with `cor_method = "fiml"` reports; in particular the averaged AIC and
#' BIC are finite here rather than `NA`.
#'
#' The grid containing the setting combinations is produced based on the entries
#' to the respective arguments. To this end, all possible combinations resulting
#' in unique EFA models are considered: combinations that resolve to the same
#' model are run only once. Two combinations are the same model only if every
#' setting the fit consumes agrees, and for "PAF" that includes the iteration cap
#' `max_iter`. Since a named `type` brings its own cap, a `type` of
#' `c("none", "SPSS")` whose specific settings match the SPSS combination in
#' every other respect still gives two "PAF" models, unless `max_iter` is also
#' set to the cap of the SPSS implementation. We include here a list
#' of arguments that are only evaluated under specific conditions:
#'
#' The arguments `init_comm`, `criterion`, `criterion_type`,
#' `abs_eigen`, and `max_iter` are only evaluated if "PAF" is included in
#' `estimator` and "none" is included in `type`.
#'
#' The argument `varimax_type` is only evaluated if "varimax", "promax",
#' "oblique", or "orthogonal" is included in `rotation` and "none" is
#' included in `type`.
#'
#' The argument `normalize` is only evaluated if `rotation` is not
#' set to "none" and "none" is included in `type`.
#'
#' The argument `k_simplimax` is only evaluated if "simplimax" or "oblique"
#' is included in `rotation`.
#'
#' The arguments `k_promax` and `p_type` are only evaluated if
#' "promax" or "oblique" is included in `rotation` and "none" is included
#' in `type`.
#'
#' The argument `start_method` is only evaluated if "ML" is included in
#' `estimator`.
#'
#' Every solution in the grid is fitted with the eigenvalue-based factor ordering,
#' including under a named `type`: the solutions are realigned to a common target
#' before averaging, so a per-fit ordering (SPSS orders by the sum of squared
#' loadings) would not survive into the averaged result. That target is the first
#' solution the grid retains, in grid order; solutions dropped for an error,
#' non-convergence, or a Heywood case cannot become it. The averaged loadings are
#' therefore in the factor order and sign of that solution. This is the one setting a
#' named `type` does not carry, and it is visible in the two places the individual
#' fits are: the solutions returned in `efa_list` are eigenvalue-ordered, and so is
#' the single [efa_fit()] object returned when the grid collapses to one row. Their
#' loadings can therefore appear in a different factor order than the same fit run
#' through [efa_fit()] under that `type`, even though the solutions are the same.
#'
#' To avoid a bias in the averaged factor solutions from problematic solutions,
#' these are excluded prior to averaging. A solution is deemed problematic if
#' at least one of the following is true: an error occurred, the model did not
#' converge, or there is at least one Heywood (improper) case (a communality at
#' or above 1, or, for `ML`/`ULS`, a uniqueness pinned at the estimator's lower
#' bound).
#' Information on errors, convergence, and Heywood cases are returned in the
#' implementations_grid and a summary of these is given when printing the output.
#' In addition to these, information on the admissibility of the factor solutions
#' is also included. A solution was deemed admissible if (1) no error occurred,
#' (2) the model converged, (3) no Heywood cases are present, and (4) there are
#' at least two salient loadings (i.e., loadings exceeding the specified
#' `salience_threshold`) for each factor. So, solutions failing one of the
#' first three of these criteria of admissibility are also deemed problematic and
#' therefore excluded from averaging. However, solutions failing only
#' the fourth criterion of admissibility are still included for averaging.
#' Finally, if all solutions are problematic (e.g., all solutions contain
#' Heywood cases), no averaging is performed and the respective outputs are NA.
#' In this case, the implementations_grid should be inspected to see if there
#' are any error messages, and the separate EFA solutions that are also included
#' in the output can be inspected as well, for example, to see where Heywood
#' cases occurred.
#'
#' A core output of this function includes the average, minimum, and maximum
#' loadings derived from all non-problematic (see above) factor solutions. Please
#' note that these are not entire solutions, but the matrices include the average,
#' minimum, or maximum value for each cell (i.e., each loading separately). This
#' means that, for example, the matrix with the minimum loadings will contain
#' the minimum value in any of the factor solutions for each specific loading,
#' and therefore most likely contains loadings from different factor solutions.
#' The matrices containing the minimum and maximum factor solutions can
#' therefore not be interpreted as whole factor solutions.
#'
#' The averaged loading matrix is likewise a cell-wise summary rather than a fitted
#' solution: it is not itself the solution of any EFA, does not in general reproduce
#' the correlation matrix, and need not reproduce the averaged communalities. The
#' fit indices described below are correspondingly the mean (or, under
#' `averaging = "median"`, the median) of the per-solution fit indices, not the fit
#' of the averaged loadings, so the averaged loadings and the reported fit do not
#' describe one and the same model.
#'
#' The output also includes information on the average, minimum, maximum, and
#' variability of the fit indices across the non-problematic factor solutions.
#' It is important to note that not all fit indices are computed for all fit
#' methods: For ML and ULS, all fit indices can be computed, while for PAF the
#' chi-square-based indices (the chi-square statistic and its significance, CFI,
#' TLI, RMSEA, AIC, BIC, and ECVI) are NA. The common part accounted for (CAF)
#' index (Lorenzo-Seva, Timmerman, & Kiers, 2011) and the residual-based SRMR and
#' RMSR are still computed for PAF. As a consequence, if only "PAF" is included in
#' the `estimator` argument, averaging is performed for the CAF, SRMR, and RMSR, while
#' the chi-square-based indices are NA. If a combination of "PAF" and "ML" and/or
#' "ULS" are included in the `estimator` argument, the CAF, SRMR, and RMSR are
#' averaged across all non-problematic factor solutions, while the chi-square-based
#' indices are only averaged across the ML and ULS solutions. The user should
#' therefore keep in mind that the number of EFAs across which the fit indices are
#' averaged can diverge for the CAF, SRMR, and RMSR compared to the chi-square-based
#' indices.
#'
#' Each reported fit index is summarised across the (non-problematic) solutions in
#' the same descriptive way: the average, standard deviation, minimum, and maximum
#' of the per-solution values. This includes the chi-square significance level
#' (`p_chi`), which is therefore the mean (or median) of the per-solution p-values
#' and is purely descriptive; it is not the p-value of any pooled chi-square test.
#'
#' @return A list of class `c("efa_average", "EFA_AVERAGE")` containing the
#' components below. Throughout, `range` is the width `maximum - minimum` of each
#' cell across the factor solutions, not the interval `[minimum, maximum]`, and
#' `average` is the (trimmed) mean or the median, following `averaging`.
#' \item{orig_R}{Original correlation matrix.}
#' \item{h2}{A list with the average, standard deviation, minimum, maximum, and
#' range of the final communality estimates across the factor solutions.}
#' \item{loadings}{A list with the average, standard deviation, minimum, maximum,
#' and range of the final loadings across the factor solutions. If rotation was
#' "none", the unrotated loadings, otherwise the rotated loadings (pattern
#' coefficients). `average` is a cell-wise summary of the solutions and not itself
#' a fitted solution (see Details).}
#' \item{Phi}{A list with the average, standard deviation, minimum, maximum, and
#' range of the factor intercorrelations across factor solutions obtained with
#' oblique rotations.}
#' \item{ind_fac_corres}{A matrix with each cell containing the proportion of
#' the factor solutions in which the respective indicator-to-factor correspondence
#' occurred, i.e., in which the loading exceeded the specified salience threshold.
#' Note: Rowsums can exceed 1 due to cross-loadings.}
#' \item{vars_accounted}{A list with the average, standard deviation, minimum,
#' maximum, and range of explained variances and sums of squared loadings across
#' the factor solutions. Based on the unrotated loadings if rotation was "none"
#' or only one factor was extracted, otherwise on the rotated loadings. Each entry
#' is a matrix with rows `"SS loadings"`, `"Prop Tot Var"`, and `"Prop Comm Var"`;
#' the last is omitted for a single-factor solution, where it is identically 1, so
#' those matrices have two rows there and three otherwise.}
#' \item{fit_indices}{A matrix containing the average, standard deviation,
#' minimum, maximum, and range for all applicable fit indices across the respective
#' factor solutions, and the degrees of freedom (df). If the estimator argument
#' contains ML or ULS: Fit indices derived
#' from the unrotated factor loadings: Chi Square (chisq), including significance
#' level, Comparative Fit Index (CFI), Tucker-Lewis Index (TLI), Root Mean Square
#' Error of Approximation (RMSEA), Akaike Information Criterion (AIC), Bayesian
#' Information Criterion (BIC), Expected Cross-Validation Index (ECVI), and the
#' common part accounted for (CAF) index as proposed by Lorenzo-Seva, Timmerman,
#' & Kiers (2011). The residual-based Standardized Root Mean Square Residual
#' (SRMR) and Root Mean Square Residual (RMSR) and the CAF are also computed for
#' PAF; for PAF the remaining (chi-square-based) indices are not available (see
#' details).}
#' \item{implementations_grid}{A matrix containing, for each performed EFA,
#' the setting combination, if an error occurred (logical), the error message
#' (character), an integer convergence code (0 = converged; for ML and ULS the
#' same codes as [`stats::optim()`][stats::optim]'s "L-BFGS-B", for PAF 1 if the
#' maximum number of iterations was reached without meeting the convergence
#' criterion and 0 otherwise),
#' if heywood cases occurred (logical, see details for definition), if the
#' solution was admissible (logical, see details for definition), and the fit
#' indices.}
#' \item{efa_list}{A list containing the outputs of all performed EFAs. The names
#' correspond to the rownames from the implementations_grid.}
#' \item{settings}{A list of the settings used, including `seed` (`NULL` when none was
#'   supplied).}
#'
#' If the supplied arguments admit only a single EFA, there is nothing to average
#' across: that one [efa_fit()] object is returned instead, with a warning. Its
#' `settings` are that fit's, with `seed` recording the seed the run was governed by.
#' They therefore describe the concrete arguments the fit ran under rather than the
#' `type` that supplied them: a row taken from a named preset records `type = "none"`
#' together with the preset's resolved values (for example `max_iter = 25` for
#' "SPSS"), and `order_type = "eigen"` as for every other row in the grid.
#'
#' @source Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
#' of implementations of principal axis factoring and promax rotation in R and SPSS.
#' Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x
#' @source Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
#' rotation to oblique simple structure. British Journal of Statistical Psychology,
#' 17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x
#' @source Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
#' Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
#' Research, 46, 340-364, doi: 10.1080/00273171.2011.564527
#' @source Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
#' factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233
#'
#' @family factor analysis
#'
#' @export
#'
#' @examples
#' # Averaging across one implementation each of PAF (EFAtools type), ULS, and
#' # ML with one implementation of promax (EFAtools type) (3 EFAs)
#' Aver_meth <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                          estimator = c("PAF", "ULS", "ML"), type = "EFAtools",
#'                          start_method = "psych")
#'
#' \donttest{
#' # Averaging across different implementations of PAF and promax rotation (72 EFAs)
#' Aver_PAF <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500)
#'
#' # Use median instead of mean for averaging (72 EFAs)
#' Aver_PAF_md <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                            averaging = "median")
#'
#' # Averaging across different implementations of PAF and promax rotation,
#' # and across ULS and different versions of ML (108 EFAs)
#' Aver_meth_ext <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                              estimator = c("PAF", "ULS", "ML"))
#'
#' # Averaging across different oblique rotation methods, using one implementation
#' # of ML and one implementation of promax (EFAtools type) (7 EFAs)
#' Aver_rot <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                          estimator = "ML", rotation = "oblique", type = "EFAtools",
#'                          start_method = "psych")
#'}
#'
#' \donttest{
#' # Two-stage FIML correlations from raw data with missing values: the EM
#' # saturated moments are estimated once and the resulting correlation is
#' # averaged across the grid of EFAs.
#' x_miss <- GRiPS_raw
#' x_miss[cbind(1:20, 1)] <- NA
#' Aver_fiml <- efa_average(x_miss, n_factors = 1, estimator = c("PAF", "ML"),
#'                          cor_method = "fiml")
#' }
#'
efa_average <- function(x, n_factors, N = NA, estimator = "PAF", rotation = "promax",
                        type = "none", averaging = c("mean", "median"), trim = 0,
                        salience_threshold = .3,
                        max_iter = 1e4,
                        init_comm = c("smc", "mac", "unity"),
                        criterion = c(1e-3),
                        criterion_type = c("sum", "max_individual"),
                        abs_eigen = c(TRUE),
                        varimax_type = c("svd", "kaiser"),
                        normalize = TRUE,
                        k_promax = 2:4, k_simplimax = ncol(x),
                        p_type = c("norm", "unnorm"), precision = 1e-5,
                        start_method = c("psych", "factanal"),
                        use = c("pairwise.complete.obs", "all.obs",
                                "complete.obs", "everything", "na.or.complete"),
                        cor_method = c("pearson", "spearman", "kendall", "poly",
                                       "tetra", "fiml"),
                        show_progress = TRUE, seed = NULL,
                        P_type = lifecycle::deprecated()) {

  # Perform argument checks
  .assert_cor_input(x)

  # `P_type` was renamed to the snake_case `p_type`; the old spelling is still
  # accepted (silently) for backwards compatibility and mapped onto `p_type` here.
  if (lifecycle::is_present(P_type)) p_type <- P_type

  checkmate::assert_count(n_factors)
  # Matches efa_fit()'s contract: NA means "sample size unknown", 0 is an impossible one.
  # Checked here as well as there, because the grid runs each fit inside try(), which would
  # otherwise turn the argument error into an unexplained all-solutions-failed result.
  # Raised through .assert_args() so the rejection carries the same condition class here as
  # it does from efa_fit().
  .assert_args(checkmate::assert_count(N, positive = TRUE, na.ok = TRUE))
  # The vector-valued choice arguments are case-insensitive: map them onto the
  # canonical spellings first, so the subset checks and the type grid below only
  # ever see the canonical values.
  estimator <- .map_subset_ci(estimator, c("PAF", "ML", "ULS", "MINRES"))
  checkmate::assert_subset(estimator, c("PAF", "ML", "ULS", "MINRES"),
                           empty.ok = FALSE)
  # "MINRES" is a synonym for "ULS" (same estimator); resolve to the canonical
  # name so the type grid below is keyed once per requested estimator.
  estimator[estimator == "MINRES"] <- "ULS"
  estimator <- unique(estimator)
  rotation_choices <- c("none", "orthogonal", "oblique", "varimax",
                        "equamax", "quartimax", "geominT", "bentlerT",
                        "bifactorT", "promax", "oblimin", "quartimin",
                        "simplimax", "bentlerQ", "geominQ", "bifactorQ")
  rotation <- .map_subset_ci(rotation, rotation_choices)
  checkmate::assert_subset(rotation, rotation_choices, empty.ok = FALSE)
  # Drop duplicates the case-folding may have collapsed (e.g. "promax" and "Promax"), as for
  # the estimators above, so the averaging grid runs each requested rotation and type once.
  rotation <- unique(rotation)
  type <- .map_subset_ci(type, c("none", "EFAtools", "psych", "SPSS"))
  checkmate::assert_subset(type, c("none", "EFAtools", "psych", "SPSS"),
                           empty.ok = FALSE)
  type <- unique(type)
  averaging <- .match_arg_ci(averaging)
  checkmate::assert_number(trim, lower = 0, upper = 0.5)
  # `trim` only reaches mean(); the median has no trimming to do. Say so rather than
  # silently recording a trim in the settings that never affected a single value.
  if (averaging == "median" && trim > 0) {
    cli::cli_inform(
      c("i" = "{.arg trim} is only used when {.code averaging = \"mean\"}; it is ignored for the median."),
      class = "efa_avg_trim_ignored"
    )
  }
  checkmate::assert_number(salience_threshold, lower = 0, upper = 1)
  checkmate::assert_count(max_iter)
  init_comm <- .map_subset_ci(init_comm, c("smc", "mac", "unity"))
  checkmate::assert_subset(init_comm, c("smc", "mac", "unity"),
                           empty.ok = FALSE)
  init_comm <- unique(init_comm)
  checkmate::assert_vector(criterion, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_true(all(criterion > 0 & criterion < 1))
  criterion_type <- .map_subset_ci(criterion_type, c("max_individual", "sum"))
  checkmate::assert_subset(criterion_type, c("max_individual", "sum"),
                           empty.ok = FALSE)
  criterion_type <- unique(criterion_type)
  checkmate::assert_subset(abs_eigen, c(TRUE, FALSE),
                           empty.ok = FALSE)
  varimax_type <- .map_subset_ci(varimax_type, c("svd", "kaiser"))
  checkmate::assert_subset(varimax_type, c("svd", "kaiser"),
                           empty.ok = FALSE)
  varimax_type <- unique(varimax_type)
  checkmate::assert_subset(normalize, c(TRUE, FALSE),
                           empty.ok = FALSE)
  checkmate::assert_vector(k_promax, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_vector(k_simplimax, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  p_type <- .map_subset_ci(p_type, c("unnorm", "norm"))
  checkmate::assert_subset(p_type, c("unnorm", "norm"),
                           empty.ok = FALSE)
  p_type <- unique(p_type)
  checkmate::assert_vector(precision, strict = TRUE, any.missing = FALSE,
                           min.len = 1)
  checkmate::assert_true(all(precision > 0 & precision < 1))
  start_method <- .map_subset_ci(start_method, c("psych", "factanal"))
  checkmate::assert_subset(start_method, c("psych", "factanal"),
                           empty.ok = FALSE)
  start_method <- unique(start_method)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  checkmate::assert_flag(show_progress)



  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                             N_policy = "optional")
  R <- prep$R
  N <- prep$N

  # Check if model is identified

  # calculate degrees of freedom
  m <- ncol(R)
  df <- .efa_df(m, n_factors)

  if(df < 0){

    cli::cli_abort(
      c("The model is underidentified.",
        "i" = "Use fewer factors or more indicators, then try again."),
      class = "efa_underidentified"
    )

  } else if (df == 0){

    cli::cli_warn(
      c("The model is just identified ({.code df = 0}).",
        "i" = "Consider fewer factors or more indicators."),
      class = "efa_just_identified"
    )

  }

  if (n_factors == 1 && !all(rotation == "none")) {
    cli::cli_inform(
      c("i" = "{.arg n_factors} is 1 but {.arg rotation} is not {.val none}; setting {.arg rotation} to {.val none}, as single-factor solutions cannot be rotated."),
      class = "efa_avg_single_factor_rotation"
    )
    rotation <- "none"
  }

  ### create the grid with all combinations of the input arguments

  arg_grid <- .build_avg_grid(
    estimator = estimator, type = type, rotation = rotation, init_comm = init_comm,
    criterion = criterion, criterion_type = criterion_type, abs_eigen = abs_eigen,
    max_iter = max_iter, start_method = start_method, k_promax = k_promax,
    normalize = normalize, P_type = p_type, precision = precision,
    varimax_type = varimax_type, k_simplimax = k_simplimax)

  # The grid leaves precision NA for unrotated rows, where it is not applicable;
  # those rows still need a valid tolerance to pass to EFA(), so fall back to its
  # default. Kept as one constant so both the single- and multi-EFA paths agree.
  default_precision <- 1e-5

  # Iteration budget handed to the rotation engine, well above its own default so a
  # criterion-based rotation is not stopped short of the requested precision. Kept as one
  # constant for the same reason as `default_precision`: a rotation that converges inside
  # the grid must not fail to converge when the same settings collapse to a single row.
  rotation_maxit <- 5e4

  ### Run all efas

  # A supplied `seed` makes the averaging run reproducible and leaves the caller's RNG
  # stream untouched: it is saved and restored on exit (or, if none existed, the state
  # set.seed() creates is removed again). The grid's fits are stochastic whenever a
  # criterion-based rotation draws random starts, and future_lapply(future.seed = TRUE)
  # derives its per-fit L'Ecuyer streams from the state set here, which also makes the
  # grid independent of the number of parallel workers. Set before the single-combination
  # branch below so that shortcut is covered too. Mirrors the seed handling in efa_fit().
  .set_local_seed(seed)

  if (nrow(arg_grid) == 1) {

    cli::cli_warn("There was only one combination of arguments; returning a normal EFA output.",
                  class = "efa_avg_single_combination")

    single <- efa_fit(R, n_factors, N = N, estimator = arg_grid$estimator,
        rotation = arg_grid$rotation,
        estimate_control = estimate_control(type = "none",
            init_comm = arg_grid$init_comm, criterion = arg_grid$criterion,
            # The grid carries the row's iteration cap: the type's own on a named-type
            # PAF row (SPSS 25, psych 50, EFAtools 300), the max_iter argument on a PAF
            # row of type "none", and NA on an ML/ULS row that never uses it.
            criterion_type = arg_grid$criterion_type, max_iter = arg_grid$max_iter,
            abs_eigen = arg_grid$abs_eigen,
            # The grid leaves start_method NA for the non-ML rows that ignore it; the
            # estimate control requires a concrete value, so fall back to the default.
            start_method = if (is.na(arg_grid$start_method)) "psych" else arg_grid$start_method),
        rotate_control = rotate_control(type = "none",
            # normalize and precision are NA on the unrotated rows that ignore them; the
            # rotate control requires concrete values, so fall back to the defaults there.
            normalize = if (is.na(arg_grid$normalize)) TRUE else arg_grid$normalize,
            precision = if (is.na(arg_grid$precision)) default_precision else arg_grid$precision,
            # order_type is fixed to "eigen" (the ordering only affects the factor
            # order of this single returned fit, not any averaged quantity).
            order_type = "eigen", varimax_type = arg_grid$varimax_type,
            p_type = arg_grid$P_type,
            k = ifelse(arg_grid$rotation == "promax", arg_grid$k_promax, arg_grid$k_simplimax),
            # engine extra carried in the control, not efa_fit()'s dots: the
            # unrotated rows never consume it, and bare dots are rejected there
            maxit = rotation_maxit))
    # The fit drew from the stream .set_local_seed() just seeded, so this shortcut
    # result was produced under `seed` even though efa_fit() was not handed it; record
    # it, or the one documented return path that is not an efa_average object would be
    # the only one that cannot say how it was seeded.
    # (single-bracket assignment: `$<-` with a NULL value would DELETE the field
    # rather than set it, dropping it from the schema whenever no seed was supplied)
    single$settings["seed"] <- list(seed)
    return(single)

  }

  # Set the progress handler only for the duration of this call, restoring the
  # user's original setting on exit (capturing the raw option restores an unset
  # state to NULL rather than to an explicit empty handler list).
  old_handlers <- getOption("progressr.handlers")
  progressr::handlers(if (isTRUE(show_progress)) "cli" else "void")
  on.exit(options(progressr.handlers = old_handlers), add = TRUE)

  progressr::with_progress({
    n_efa <- nrow(arg_grid)
    if (n_efa <= 10) {
      stepsize <- 1
    } else {
      stepsize <- round(n_efa / 100 * 10)
    }

    # One declared step per update actually sent: the loop below reports on every
    # stepsize-th EFA and on the last one, which is `ceiling(n_efa / stepsize)` updates,
    # plus the closing message. Declaring fewer completes the bar while EFAs are still
    # being fitted, and an update sent to an already-completed progressor is dropped
    # (the closing message would never be shown) or makes with_progress() warn that it
    # is no longer listening to this progressor.
    efa_progress_bar <- progressr::progressor(steps = ceiling(n_efa / stepsize) + 1)
    efa_progress_bar("Running EFAs:", class = "sticky", amount = 0)
    efa_list <- future.apply::future_lapply(1:n_efa,
                                            function(i, estimators, rotations,
                                                     init_comms, criteria,
                                                     criterion_types, abs_eigens,
                                                     max_iters, varimax_types, k_ps,
                                                     k_ss, normalizes, P_types,
                                                     precisions, start_methods) {
      # Report every stepsize-th EFA, and the last one whatever its position, so the
      # bar tracks the grid to its end instead of completing on the last whole group.
      if (i %% stepsize == 0 || i == n_efa){
        efa_progress_bar(message = sprintf("Running EFA %g of %g", i, n_efa))
      }

      # precision only affects rotated solutions; the grid leaves it NA for the
      # unrotated rows, so fall back to the default tolerance there.
      precision_i <- if (is.na(precisions[i])) default_precision else precisions[i]

      # Per-EFA warnings (e.g. Heywood cases) are summarised once after the grid
      # is run, so they are suppressed here to avoid one warning per model.
      try(suppressWarnings(
          efa_fit(R, n_factors, N = N, estimator = estimators[i], rotation = rotations[i],
              estimate_control = estimate_control(type = "none", init_comm = init_comms[i],
                  criterion = criteria[i], criterion_type = criterion_types[i],
                  # Per-row iteration cap: the type's own on a named-type PAF row
                  # (SPSS 25, psych 50, EFAtools 300), the max_iter argument on a PAF
                  # row of type "none", and NA on the ML/ULS rows that never use it.
                  max_iter = max_iters[i], abs_eigen = abs_eigens[i],
                  # The grid leaves start_method NA for the non-ML rows that ignore it; the
                  # estimate control requires a concrete value, so fall back to the default.
                  start_method = if (is.na(start_methods[i])) "psych" else start_methods[i]),
              rotate_control = rotate_control(type = "none",
                  # normalize is NA on the unrotated rows that ignore it; the rotate control
                  # requires a concrete value, so fall back to the default there.
                  normalize = if (is.na(normalizes[i])) TRUE else normalizes[i],
                  # order_type is fixed to "eigen": the averaging re-aligns every solution
                  # to the first by congruence, so the per-fit factor order is irrelevant.
                  precision = precision_i, order_type = "eigen",
                  varimax_type = varimax_types[i], p_type = P_types[i],
                  k = ifelse(rotations[i] == "promax", k_ps[i], k_ss[i]),
                  # engine extra carried in the control, not efa_fit()'s dots: the
                  # unrotated rows never consume it, and bare dots are rejected there
                  maxit = rotation_maxit))),
          silent = TRUE)
    }, estimators = arg_grid$estimator, rotations = arg_grid$rotation,
    init_comms = arg_grid$init_comm, criteria = arg_grid$criterion,
    criterion_types = arg_grid$criterion_type, abs_eigens = arg_grid$abs_eigen,
    max_iters = arg_grid$max_iter, varimax_types = arg_grid$varimax_type,
    k_ps = arg_grid$k_promax, k_ss = arg_grid$k_simplimax,
    normalizes = arg_grid$normalize, P_types = arg_grid$P_type,
    precisions = arg_grid$precision, start_methods = arg_grid$start_method,
    future.seed = TRUE)

    # The last of the declared steps, so the bar completes on this message rather than
    # on the final EFA. It has to consume a step: progressr drops an update sent to a
    # progressor that has already reached its last one.
    efa_progress_bar(message = "Done Running EFAs")
  })

  names(efa_list) <- rownames(arg_grid)

  ### Extract relevant information from EFA outputs
  if (isTRUE(show_progress)) {
    # The post-grid stages run sequentially in this process, so they are reported with
    # cli's own progress steps (each call ticks off the previous one). The grid above
    # keeps progressr, which is what relays progress out of the future workers.
    cli::cli_progress_step("Extracting data")
  }
  ext_list <- .extract_data(efa_list, R, n_factors, n_efa, rotation, salience_threshold)

  bad <- ext_list$for_grid$converged != 0 | ext_list$for_grid$errors |
         ext_list$for_grid$heywood

  # Summarise the per-model problems once (the individual EFA warnings were
  # suppressed during the grid run to avoid one warning per model).
  n_excluded <- sum(bad %in% TRUE)
  if (n_excluded > 0L) {
    cli::cli_warn(
      c(paste("Excluded {n_excluded} of {length(bad)} EFA{?s} from averaging due to",
              "errors, non-convergence, or Heywood cases."),
        "i" = "Inspect {.field implementations_grid} in the returned object for the per-model details."),
      class = "efa_avg_excluded_solutions"
    )
  }

  if (all(bad %in% TRUE)) {

    av_list <- list(
      h2 = NA,
      loadings = NA,
      phi = NA,
      ind_fac_corres = NA,
      vars_accounted = NA,
      fit_indices = NA)

  } else {

    if (n_factors > 1) {
      if (isTRUE(show_progress)) {
        cli::cli_progress_step("Reordering factors")
      }

      re_list <- .array_reorder(ext_list$vars_accounted, ext_list$L, ext_list$L_corres,
                                ext_list$phi, ext_list$extract_phi, n_factors)
    } else {
      re_list <- ext_list
    }


    if (isTRUE(show_progress)) {
      cli::cli_progress_step("Averaging data")
    }
    av_list <- .average_values(re_list$vars_accounted, re_list$L, re_list$L_corres,
                               ext_list$h2, re_list$phi, ext_list$extract_phi,
                               averaging, trim,
                               ext_list$for_grid[, c("chisq", "p_chi", "caf", "cfi",
                                                     "rmsea", "aic", "bic", "srmr",
                                                     "tli", "ecvi", "rmsr")], df,
                               colnames(R))

  }


  arg_grid <- cbind(arg_grid, ext_list$for_grid)


  settings <- list(
    estimator = estimator,
    # back-compat alias, as for the frozen P_type key
    method = estimator,
    rotation = rotation,
    type = type,
    n_factors = n_factors,
    N = N,
    init_comm = init_comm,
    criterion = criterion,
    criterion_type = criterion_type,
    abs_eigen = abs_eigen,
    varimax_type = varimax_type,
    normalize = normalize,
    k_promax = k_promax,
    k_simplimax = k_simplimax,
    P_type = p_type,
    precision = precision,
    start_method = start_method,
    use = use,
    cor_method = cor_method,
    max_iter = max_iter,
    averaging = averaging,
    trim = trim,
    salience_threshold = salience_threshold,
    seed = seed
  )

  # Create output
  output <- list(
    orig_R = R,
    h2 = av_list$h2,
    loadings = av_list$loadings,
    Phi = av_list$phi,
    ind_fac_corres = av_list$ind_fac_corres,
    vars_accounted = av_list$vars_accounted,
    fit_indices = av_list$fit_indices,
    implementations_grid = arg_grid,
    efa_list = efa_list,
    settings = settings
  )

  class(output) <- c("efa_average", "EFA_AVERAGE")

  if (isTRUE(show_progress)) {
    # Ticks off the last step that was started; which one that is depends on how far
    # the branches above got (no solution to average leaves it at "Extracting data").
    cli::cli_progress_done()
  }

  return(output)

}

# Unit tests for .average_values() (R/averaging.R), which reduces the stacked
# efa_average() arrays to their average, range, and dispersion summaries. Inputs are
# literal arrays with hand-chosen values, so the expected means, medians, trimmed
# means, and NA propagation can be written out exactly.

# Five solutions per cell, all carrying the same five values. They are chosen so the
# three averaging modes are mutually distinguishable: the untrimmed mean is 6, trim = .2
# drops the most extreme value at each end and averages the remaining three to 4, and the
# median is 3. Three values per cell could not tell them apart -- with n = 3 nothing is
# trimmed until trim reaches .5, where the trimmed mean is exactly the median.
av_values <- c(1, 2, 3, 7, 17)

# Indicator-to-factor correspondence: four solutions assign item 2 to F2 and item 3 to
# F3, the fifth swaps them, so the averaged correspondence is .8 / .2. This summary is a
# plain proportion and is never trimmed.
av_corres_plain <- c(1, 0, 0, 0, 1, 0, 0, 0, 1)
av_corres_swap <- c(1, 0, 0, 0, 0, 1, 0, 1, 0)

av_fit_names <- c("chisq", "p_chi", "caf", "cfi", "rmsea", "aic", "bic")
av_for_grid <- as.data.frame(
  matrix(rep(av_values, times = length(av_fit_names)), nrow = length(av_values),
         dimnames = list(NULL, av_fit_names))
)
av_h2 <- matrix(rep(av_values, times = 3), nrow = length(av_values), ncol = 3)

# One .average_values() call. Every argument is fixed except the three the fixtures
# actually vary, so what each fixture tests is visible on its own line below.
av_fixture <- function(extract_phi = FALSE, averaging = "mean", trim = 0,
                       for_grid = av_for_grid, h2 = av_h2) {
  stacked <- array(rep(av_values, each = 9), c(3, 3, 5))
  .average_values(
    L = stacked,
    L_corres = array(c(av_corres_plain, av_corres_swap,
                       rep(av_corres_plain, times = 3)), c(3, 3, 5)),
    vars_accounted = stacked,
    h2 = h2,
    phi = if (isTRUE(extract_phi)) stacked else NA,
    extract_phi = extract_phi,
    averaging = averaging,
    trim = trim,
    for_grid = for_grid,
    df = 5,
    ind_names = paste0("Ind", 1:3)
  )
}

av_mean_NA <- av_fixture(extract_phi = FALSE, averaging = "mean", trim = 0)
av_mean_NA_t20 <- av_fixture(extract_phi = FALSE, averaging = "mean", trim = .2)
av_mean <- av_fixture(extract_phi = TRUE, averaging = "mean", trim = 0)
av_median <- av_fixture(extract_phi = TRUE, averaging = "median", trim = 0)
av_median_NA <- av_fixture(extract_phi = FALSE, averaging = "median", trim = .2)

# Dimnames and the closed-form dispersion of av_values, written once.
av_ind_dn <- list(paste0("Ind", 1:3), paste0("F", 1:3))
av_var_dn <- list(c("SS loadings", "Prop Tot Var", "Prop Comm Var"), paste0("F", 1:3))
av_fac_dn <- list(paste0("F", 1:3), paste0("F", 1:3))
# sum((av_values - 6)^2) = 172, over n - 1 = 4
av_sd <- sqrt(43)


test_that(".average_values works", {
  ### tests for av_mean_NA with extract_phi = FALSE and trim = 0
  expect_type(av_mean_NA, "list")
  expect_named(av_mean_NA, c("h2", "loadings", "phi", "vars_accounted",
                              "ind_fac_corres", "fit_indices"))
  expect_type(av_mean_NA$h2, "list")
  expect_named(av_mean_NA$h2, c("average", "sd", "min", "max", "range"))
  expect_type(av_mean_NA$h2$average, "double")
  expect_equal(unname(av_mean_NA$h2$average), rep(6, 3))
  expect_named(av_mean_NA$h2$average, paste0("Ind", 1:3))
  expect_equal(unname(av_mean_NA$h2$sd), rep(av_sd, 3))
  expect_equal(unname(av_mean_NA$h2$min), rep(1, 3))
  expect_equal(unname(av_mean_NA$h2$max), rep(17, 3))
  expect_type(av_mean_NA$loadings, "list")
  expect_named(av_mean_NA$loadings, c("average", "sd", "min", "max", "range"))
  expect_s3_class(av_mean_NA$loadings$average, "LOADINGS")
  expect_equal(unclass(av_mean_NA$loadings$average), matrix(rep(6, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(av_mean_NA$loadings$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(unclass(av_mean_NA$loadings$min), matrix(rep(1, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(unclass(av_mean_NA$loadings$max), matrix(rep(17, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(av_mean_NA$loadings$range, matrix(rep(16, 9), ncol = 3, dimnames = av_ind_dn))
  expect_type(av_mean_NA$vars_accounted, "list")
  expect_named(av_mean_NA$vars_accounted, c("average", "sd", "min", "max", "range"))
  checkmate::expect_matrix(av_mean_NA$vars_accounted$average)
  expect_equal(av_mean_NA$vars_accounted$average, matrix(rep(6, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA$vars_accounted$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA$vars_accounted$min, matrix(rep(1, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA$vars_accounted$max, matrix(rep(17, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA$vars_accounted$range, matrix(rep(16, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA$phi, NA)
  checkmate::expect_matrix(av_mean_NA$ind_fac_corres)
  expect_equal(av_mean_NA$ind_fac_corres,
               matrix(c(1, 0, 0, 0, .8, .2, 0, .2, .8), ncol = 3, dimnames = av_ind_dn))
  expect_s3_class(av_mean_NA$fit_indices, "data.frame")
  expect_named(av_mean_NA$fit_indices, c("index", "average", "sd", "range",
                                          "min", "max"))
  expect_type(av_mean_NA$fit_indices$index, "character")
  expect_equal(av_mean_NA$fit_indices$index, c(av_fit_names, "df"))
  expect_type(av_mean_NA$fit_indices$average, "double")
  expect_equal(av_mean_NA$fit_indices$average, c(rep(6, 7), 5))
  expect_type(av_mean_NA$fit_indices$sd, "double")
  expect_equal(av_mean_NA$fit_indices$sd, c(rep(av_sd, 7), 0))
  expect_type(av_mean_NA$fit_indices$range, "double")
  expect_equal(av_mean_NA$fit_indices$range, c(rep(16, 7), 0))
  expect_type(av_mean_NA$fit_indices$min, "double")
  expect_equal(av_mean_NA$fit_indices$min, c(rep(1, 7), 5))
  expect_type(av_mean_NA$fit_indices$max, "double")
  expect_equal(av_mean_NA$fit_indices$max, c(rep(17, 7), 5))


  ### tests for av_mean_NA_t20 with extract_phi = FALSE and trim = .20
  expect_type(av_mean_NA_t20, "list")
  expect_named(av_mean_NA_t20, c("h2", "loadings", "phi", "vars_accounted",
                                  "ind_fac_corres", "fit_indices"))
  expect_type(av_mean_NA_t20$h2, "list")
  expect_named(av_mean_NA_t20$h2, c("average", "sd", "min", "max", "range"))
  expect_type(av_mean_NA_t20$h2$average, "double")
  expect_equal(unname(av_mean_NA_t20$h2$average), rep(4, 3))
  expect_named(av_mean_NA_t20$h2$average, paste0("Ind", 1:3))
  # only the average is trimmed; the dispersion summaries still describe all five values
  expect_equal(unname(av_mean_NA_t20$h2$sd), rep(av_sd, 3))
  expect_equal(unname(av_mean_NA_t20$h2$min), rep(1, 3))
  expect_equal(unname(av_mean_NA_t20$h2$max), rep(17, 3))
  expect_type(av_mean_NA_t20$loadings, "list")
  expect_named(av_mean_NA_t20$loadings, c("average", "sd", "min", "max", "range"))
  expect_s3_class(av_mean_NA_t20$loadings$average, "LOADINGS")
  expect_equal(unclass(av_mean_NA_t20$loadings$average), matrix(rep(4, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(av_mean_NA_t20$loadings$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(unclass(av_mean_NA_t20$loadings$min), matrix(rep(1, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(unclass(av_mean_NA_t20$loadings$max), matrix(rep(17, 9), ncol = 3, dimnames = av_ind_dn))
  expect_type(av_mean_NA_t20$vars_accounted, "list")
  expect_named(av_mean_NA_t20$vars_accounted, c("average", "sd", "min", "max", "range"))
  checkmate::expect_matrix(av_mean_NA_t20$vars_accounted$average)
  expect_equal(av_mean_NA_t20$vars_accounted$average, matrix(rep(4, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA_t20$vars_accounted$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA_t20$vars_accounted$min, matrix(rep(1, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA_t20$vars_accounted$max, matrix(rep(17, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA_t20$vars_accounted$range, matrix(rep(16, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_mean_NA_t20$phi, NA)
  checkmate::expect_matrix(av_mean_NA_t20$ind_fac_corres)
  expect_equal(av_mean_NA_t20$ind_fac_corres,
               matrix(c(1, 0, 0, 0, .8, .2, 0, .2, .8), ncol = 3, dimnames = av_ind_dn))
  expect_s3_class(av_mean_NA_t20$fit_indices, "data.frame")
  expect_named(av_mean_NA_t20$fit_indices, c("index", "average", "sd", "range",
                                          "min", "max"))
  expect_type(av_mean_NA_t20$fit_indices$index, "character")
  expect_equal(av_mean_NA_t20$fit_indices$index, c(av_fit_names, "df"))
  expect_type(av_mean_NA_t20$fit_indices$average, "double")
  expect_equal(av_mean_NA_t20$fit_indices$average, c(rep(4, 7), 5))
  expect_type(av_mean_NA_t20$fit_indices$sd, "double")
  expect_equal(av_mean_NA_t20$fit_indices$sd, c(rep(av_sd, 7), 0))
  expect_type(av_mean_NA_t20$fit_indices$range, "double")
  expect_equal(av_mean_NA_t20$fit_indices$range, c(rep(16, 7), 0))
  expect_type(av_mean_NA_t20$fit_indices$min, "double")
  expect_equal(av_mean_NA_t20$fit_indices$min, c(rep(1, 7), 5))
  expect_type(av_mean_NA_t20$fit_indices$max, "double")
  expect_equal(av_mean_NA_t20$fit_indices$max, c(rep(17, 7), 5))


  ### tests for av_mean with extract_phi = TRUE (only affected output tested)
  expect_type(av_mean$phi, "list")
  checkmate::expect_matrix(av_mean$phi$average)
  expect_equal(av_mean$phi$average, matrix(rep(6, 9), ncol = 3, dimnames = av_fac_dn))
  checkmate::expect_matrix(av_mean$phi$sd)
  expect_equal(av_mean$phi$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_fac_dn))
  checkmate::expect_matrix(av_mean$phi$min)
  expect_equal(av_mean$phi$min, matrix(rep(1, 9), ncol = 3, dimnames = av_fac_dn))
  checkmate::expect_matrix(av_mean$phi$max)
  expect_equal(av_mean$phi$max, matrix(rep(17, 9), ncol = 3, dimnames = av_fac_dn))


  ### tests for av_median_NA with extract_phi = FALSE
  expect_type(av_median_NA, "list")
  expect_named(av_median_NA, c("h2", "loadings", "phi", "vars_accounted",
                                "ind_fac_corres", "fit_indices"))
  expect_type(av_median_NA$h2, "list")
  expect_named(av_median_NA$h2, c("average", "sd", "min", "max", "range"))
  expect_type(av_median_NA$h2$average, "double")
  expect_equal(unname(av_median_NA$h2$average), rep(3, 3))
  expect_named(av_median_NA$h2$average, paste0("Ind", 1:3))
  expect_equal(unname(av_median_NA$h2$sd), rep(av_sd, 3))
  expect_equal(unname(av_median_NA$h2$min), rep(1, 3))
  expect_equal(unname(av_median_NA$h2$max), rep(17, 3))
  expect_type(av_median_NA$loadings, "list")
  expect_named(av_median_NA$loadings, c("average", "sd", "min", "max", "range"))
  expect_s3_class(av_median_NA$loadings$average, "LOADINGS")
  expect_equal(unclass(av_median_NA$loadings$average), matrix(rep(3, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(av_median_NA$loadings$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(unclass(av_median_NA$loadings$min), matrix(rep(1, 9), ncol = 3, dimnames = av_ind_dn))
  expect_equal(unclass(av_median_NA$loadings$max), matrix(rep(17, 9), ncol = 3, dimnames = av_ind_dn))
  expect_type(av_median_NA$vars_accounted, "list")
  expect_named(av_median_NA$vars_accounted, c("average", "sd", "min", "max", "range"))
  checkmate::expect_matrix(av_median_NA$vars_accounted$average)
  expect_equal(av_median_NA$vars_accounted$average, matrix(rep(3, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_median_NA$vars_accounted$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_median_NA$vars_accounted$min, matrix(rep(1, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_median_NA$vars_accounted$max, matrix(rep(17, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_median_NA$vars_accounted$range, matrix(rep(16, 9), ncol = 3, dimnames = av_var_dn))
  expect_equal(av_median_NA$phi, NA)
  checkmate::expect_matrix(av_median_NA$ind_fac_corres)
  expect_equal(av_median_NA$ind_fac_corres,
               matrix(c(1, 0, 0, 0, .8, .2, 0, .2, .8), ncol = 3, dimnames = av_ind_dn))
  expect_s3_class(av_median_NA$fit_indices, "data.frame")
  expect_named(av_median_NA$fit_indices, c("index", "average", "sd", "range",
                                              "min", "max"))
  expect_type(av_median_NA$fit_indices$index, "character")
  expect_equal(av_median_NA$fit_indices$index, c(av_fit_names, "df"))
  expect_type(av_median_NA$fit_indices$average, "double")
  expect_equal(av_median_NA$fit_indices$average, c(rep(3, 7), 5))
  expect_type(av_median_NA$fit_indices$sd, "double")
  expect_equal(av_median_NA$fit_indices$sd, c(rep(av_sd, 7), 0))
  expect_type(av_median_NA$fit_indices$range, "double")
  expect_equal(av_median_NA$fit_indices$range, c(rep(16, 7), 0))
  expect_type(av_median_NA$fit_indices$min, "double")
  expect_equal(av_median_NA$fit_indices$min, c(rep(1, 7), 5))
  expect_type(av_median_NA$fit_indices$max, "double")
  expect_equal(av_median_NA$fit_indices$max, c(rep(17, 7), 5))


  ### tests for av_median with extract_phi = TRUE (only affected output tested)
  expect_type(av_median$phi, "list")
  checkmate::expect_matrix(av_median$phi$average)
  expect_equal(av_median$phi$average, matrix(rep(3, 9), ncol = 3, dimnames = av_fac_dn))
  checkmate::expect_matrix(av_median$phi$sd)
  expect_equal(av_median$phi$sd, matrix(rep(av_sd, 9), ncol = 3, dimnames = av_fac_dn))
  checkmate::expect_matrix(av_median$phi$min)
  expect_equal(av_median$phi$min, matrix(rep(1, 9), ncol = 3, dimnames = av_fac_dn))
  checkmate::expect_matrix(av_median$phi$max)
  expect_equal(av_median$phi$max, matrix(rep(17, 9), ncol = 3, dimnames = av_fac_dn))


})


test_that("trim moves the mean and is inert under the median", {
  # The point of the five-value fixture: a trimmed mean that is neither the untrimmed
  # mean nor the median. Dropping trim from the mean() call, or mis-scaling it, would
  # collapse one of these differences.
  expect_false(isTRUE(all.equal(av_mean_NA_t20$h2$average, av_mean_NA$h2$average)))
  expect_false(isTRUE(all.equal(av_mean_NA_t20$h2$average, av_median_NA$h2$average)))
  expect_false(isTRUE(all.equal(av_mean_NA$h2$average, av_median_NA$h2$average)))

  # trim is a parameter of mean() alone, so the median route ignores it entirely
  expect_equal(av_median_NA$h2$average, av_median$h2$average)
  expect_equal(unclass(av_median_NA$loadings$average),
               unclass(av_median$loadings$average))
  expect_equal(av_median_NA$fit_indices$average, av_median$fit_indices$average)

  # trimming changes only the average, never the dispersion summaries
  expect_equal(av_mean_NA_t20$fit_indices$sd, av_mean_NA$fit_indices$sd)
  expect_equal(av_mean_NA_t20$fit_indices$min, av_mean_NA$fit_indices$min)
  expect_equal(av_mean_NA_t20$fit_indices$max, av_mean_NA$fit_indices$max)
})


test_that("a summary with nothing to summarise is NA and raises no warning", {
  # A PAF-only grid carries no chi-square-based fit index, so those columns are missing
  # for every solution -- the routine case. base min()/max() would warn there and return
  # -Inf/Inf, which must never reach the reported range.
  no_chisq <- av_for_grid
  no_chisq$chisq <- NA_real_
  no_h2 <- av_h2
  no_h2[, 2] <- NA_real_

  expect_no_warning(empty <- av_fixture(for_grid = no_chisq, h2 = no_h2))

  chisq_row <- empty$fit_indices$index == "chisq"
  expect_true(is.na(empty$fit_indices$average[chisq_row]))
  expect_true(is.na(empty$fit_indices$min[chisq_row]))
  expect_true(is.na(empty$fit_indices$max[chisq_row]))
  expect_true(is.na(empty$fit_indices$range[chisq_row]))

  expect_true(is.na(empty$h2$min[[2]]))
  expect_true(is.na(empty$h2$max[[2]]))
  expect_true(is.na(empty$h2$range[[2]]))

  # the indices that do have values are unaffected
  expect_equal(empty$fit_indices$min[empty$fit_indices$index == "caf"], 1)
  expect_equal(empty$fit_indices$max[empty$fit_indices$index == "caf"], 17)
  expect_equal(unname(empty$h2$min[c(1, 3)]), c(1, 1))
})


rm(av_mean_NA, av_mean_NA_t20, av_mean, av_median, av_median_NA, av_fixture,
   av_values, av_corres_plain, av_corres_swap, av_fit_names, av_for_grid, av_h2,
   av_ind_dn, av_var_dn, av_fac_dn, av_sd)

# The shared rank-deficient input fixture, used wherever a test pins the
# `efa_cor_singular` branch.
#
# A composite variable entered alongside its own components is the archetypal way an EFA
# correlation matrix comes out rank deficient, so that is what this builds: nine
# independent variables plus their tenth column, the sum of the first two.
#
# The dimension is not incidental. The singularity check compares
# min(abs(ev)) / max(abs(ev)) against ncol(R) * .Machine$double.eps, and for an exactly
# rank-deficient matrix the smallest computed eigenvalue IS LAPACK's rounding, of order
# eps * ||R||. A 3x3 fixture therefore leaves the ratio within a factor of about 1.5 of
# the threshold -- close enough that a LAPACK build with a slightly larger noise floor
# decides the other way and no error is raised. For this seeded p = 10 fixture the ratio
# is 2.7e-16 against a threshold of 2.2e-15, a factor of 8, so the verdict is the same
# whatever the build's noise floor. The seed keeps the realization fixed on top of that.
#
# set.seed() here is deliberate and its effect is global: helper files are sourced once,
# before any test file, so the ambient stream every unseeded fixture in the suite draws
# from starts from this point. Changing the draw below therefore re-rolls those fixtures.
set.seed(1)
sing_raw <- matrix(stats::rnorm(400 * 9), 400)
sing_raw <- cbind(sing_raw, sing_raw[, 1] + sing_raw[, 2])
sing_cor <- stats::cor(sing_raw)
sing_N <- nrow(sing_raw)

# Invertible but not positive definite: a valid correlation matrix cannot have a
# negative eigenvalue, and this one does, so it exercises the separate non-PD branch.
cor_nposdef <- matrix(c(1, 1, 0, 1, 1, 1, 0, 1, 1), ncol = 3)

#' @keywords internal
#' @useDynLib EFAtools, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom utils modifyList
## `lifecycle` is a run-time dependency, not a documentation-only one: `EFA()`,
## `efa_average()` and `efa_mi()` give their legacy arguments
## `lifecycle::deprecated()` defaults and resolve them with
## `lifecycle::is_present()` / `lifecycle::deprecate_warn()`. It is additionally
## used at documentation build time by the `lifecycle::badge()` calls that mark the
## superseded functions and the deprecated argument. Every call is qualified, so no
## import is declared.
"_PACKAGE"

#' Comparison data
#'
#' Factor retention method introduced by Ruscio and Roche (2012). The code was
#' adapted from the CD code published by Auerswald and Moshagen (2019), available
#' at <https://osf.io/x5cz2/>.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data.
#' @param n_factors_max numeric. The maximum number of factors to test against.
#'  Larger numbers will increase the duration the procedure takes, but test more
#'  possible solutions. If left NA (default) the maximum number of factors for
#'  which the model is still over-identified (df > 0) is used.
#' @param N_pop numeric. Size of finite populations of comparison data. Default
#'  is 10000.
#' @param N_samples numeric. Number of samples drawn from each population.
#'  Default is 500.
#' @param alpha numeric. The alpha level used to test the significance of the
#'  improvement added by an additional factor. Default is .30.
#' @param cor_method character. One of `"pearson"`, `"spearman"`, or `"kendall"`,
#'   passed to [stats::cor()]. `"poly"` and `"tetra"` are not supported because
#'   `CD` compares the data against simulated continuous reference data.
#' Default is "pearson".
#' @param max_iter numeric. The maximum number of iterations after which the
#'  iterative PAF procedure inside the comparison-data generation is halted; it
#'  does not cap an EFA of `x`. Default is 50.
#'
#' @details Comparison data (CD) extends parallel analysis by reproducing the
#' observed correlation matrix rather than generating random data: datasets with a
#' known factor structure are generated with an increasing number of factors, and
#' the smallest number for which adding a further factor no longer significantly
#' improves the reproduction of the observed eigenvalues is retained (Ruscio &
#' Roche, 2012).
#'
#' Because it reproduces the observed correlation structure instead of a null model,
#' CD was among the more accurate criteria across a broad range of conditions in
#' Ruscio and Roche (2012). It is, however, the only criterion in this family that
#' requires raw data, and by some margin the most computationally intensive one: a
#' finite population of `N_pop` cases is generated and `N_samples` samples are drawn
#' from it at every candidate factor count. It is therefore a good choice when the
#' raw data are at hand and the runtime is acceptable, and a poor one for a quick
#' look at a correlation matrix.
#'
#' The comparison data are obtained by simulation, so the suggested number of factors
#' varies slightly from run to run. Call [base::set.seed()] beforehand to make a run
#' reproducible.
#'
#' Note that if the data contains missing values, these will be removed for the
#' comparison data procedure using [`stats::na.omit()`][stats::na.fail]. If
#' missing data should be treated differently, e.g., by imputation, do this outside
#' `efa_cd()` and then pass the complete data.
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). Its main fields are:
#' \item{n_factors}{A named numeric vector (`"CD"`) with the suggested number of
#'   factors according to comparison data results.}
#' \item{results}{A list with a single record holding the mean RMSE between the
#'   eigenvalues of the generated and the entered data per number of factors
#'   (used for the plot) and, in `rmse_eigenvalues`, the per-sample RMSE matrix
#'   (rows are samples, columns are factor counts; columns beyond the last tested
#'   factor count are left as zero).}
#' \item{settings}{A list of the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Ruscio, J., & Roche, B. (2012). Determining the number of factors to
#' retain in an exploratory factor analysis using comparison data of known
#' factorial structure. Psychological Assessment, 24, 282–292.
#' doi: 10.1037/a0025697
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#'
#' @examples
#' \donttest{
#' # determine n factors of the GRiPS
#' efa_cd(GRiPS_raw, N_pop = 500, N_samples = 20)
#'
#' # determine n factors of the DOSPERT risk subscale
#' efa_cd(DOSPERT_raw, N_pop = 500, N_samples = 20)
#'}
efa_cd <- function(x, n_factors_max = NA, N_pop = 10000, N_samples = 500, alpha = .30,
               cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
               max_iter = 50) {

  # Perform argument checks
  .assert_cor_input(x, raw_only = TRUE)

  if (.is_cormat(x)) {
    cli::cli_abort("{.arg x} is a correlation matrix, but CD only works with raw data.",
                   class = "efa_cd_needs_raw")
  }

  if (inherits(x, c("tbl_df", "tbl"))) {
    x <- as.data.frame(x)
  }

  cor_method <- .match_arg_ci(cor_method)
  .reject_poly_reference(cor_method, "efa_cd")

  checkmate::assert_count(n_factors_max, na.ok = TRUE, positive = TRUE)
  checkmate::assert_count(N_pop)
  checkmate::assert_count(N_samples, positive = TRUE)
  checkmate::assert_number(alpha, lower = 0, upper = 1)
  checkmate::assert_count(max_iter, positive = TRUE)

  if (N_pop < 2L) {
    cli::cli_abort(
      c("The comparison-data population needs at least 2 cases.",
        "x" = "{.arg N_pop} is {N_pop}."),
      class = "efa_cd_degenerate_population"
    )
  }

  if (any(is.na(x))) {
    n_row_complete <- nrow(x)
    x <- stats::na.omit(x)
    n_row_new <- nrow(x)
    n_rows_removed <- n_row_complete - n_row_new

    cli::cli_warn(
      c("The data contained missing values, removed with {.fn stats::na.omit}.",
        "i" = "{n_rows_removed} row{?s} removed."),
      class = "efa_cd_missing_removed"
    )
  }
  n_cases <- nrow(x)
  k <- ncol(x)

  # A constant variable has no correlation, so its row and column of R are NA and the
  # eigenvalue decomposition below would abort with a base-R error. It also cannot be
  # reproduced by the comparison-data generation. Reject it here, naming the columns.
  # Checked after the listwise deletion above, which can leave a variable constant.
  n_distinct <- apply(x, 2L, function(col) length(unique(col)))
  if (any(n_distinct < 2L)) {
    bad <- colnames(x)[n_distinct < 2L]
    if (is.null(bad)) bad <- as.character(which(n_distinct < 2L))
    cli::cli_abort(
      c("Every variable must have at least two distinct values.",
        "x" = "{cli::qty(bad)}Constant variable{?s}: {.val {bad}}.",
        "i" = "{cli::qty(bad)}Remove {?it/them}; comparison data cannot reproduce a variable with no variance."),
      class = "efa_cd_constant_variable"
    )
  }

  m_possible <- .det_max_factors(k)

  # Comparison data needs at least one over-identified model (df > 0) to compare
  # against; with too few indicators no such model exists and the search range is
  # empty, so abort rather than silently returning zero factors.
  if (m_possible < 1) {
    cli::cli_abort(
      c("Comparison data cannot be run because no factor model with {k} indicator{?s} is over-identified.",
        "i" = "Provide more indicators."),
      class = "efa_cd_min_indicators"
    )
  }

  if (is.na(n_factors_max) || n_factors_max > m_possible) {

    if (!is.na(n_factors_max) & n_factors_max > m_possible) {
      cli::cli_warn(
        c("{.arg n_factors_max} was set to {n_factors_max}, but at most {m_possible} factor{?s} can be extracted.",
          "i" = "Setting {.arg n_factors_max} to {m_possible}."),
        class = "efa_cd_max_factors"
      )
    }

    n_factors_max <- m_possible

  }

  # Create correlation matrix (x has no missing values at this point: incomplete
  # rows were removed above)
  R <- stats::cor(x, method = cor_method)

  eigvals_real <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  # initialize objects for iterative procedures
  RMSE_eigvals <- matrix(0, nrow = N_samples, ncol = n_factors_max)
  sig <- TRUE
  n_factors <- 1



  while (n_factors <= n_factors_max && isTRUE(sig)) {

    # The comparison-data generator is the shared Ruscio-Kaczetow kernel, which reports a
    # degenerate resample under its own name; re-raise it as an efa_cd condition, naming
    # the argument that controls the population size here.
    pop <- tryCatch(
      .simulate_cfm_empirical(R = R, x = x, n_factors = n_factors, N = N_pop,
                              cor_method = cor_method, max_iter = max_iter),
      efa_simulate_degenerate_marginal = function(e) {
        cli::cli_abort(
          c("The comparison-data population could not be generated from these data.",
            "i" = "Increase {.arg N_pop} (currently {N_pop}), or collapse response categories too rare to be drawn."),
          class = "efa_cd_degenerate_population", parent = e
        )
      })

    for (j in seq_len(N_samples)) {

      samp <- pop[sample.int(N_pop, size = n_cases, replace = TRUE),]
      R_samp <- stats::cor(samp, method = cor_method)
      eigvals_samp <- eigen(R_samp, symmetric = TRUE, only.values = TRUE)$values
      RMSE_eigvals[j,n_factors] <- sqrt(sum((eigvals_samp - eigvals_real) *
                                         (eigvals_samp - eigvals_real)) / k)
    }

    if (n_factors > 1) {

      sig <- (stats::wilcox.test(RMSE_eigvals[,n_factors],
                          RMSE_eigvals[,(n_factors - 1)], "less")$p.value < alpha)
    }
    if (isTRUE(sig)) {
      n_factors <- n_factors + 1
    }

  }

  n_factors <- n_factors - 1

  settings <- list(
    n_factors_max = n_factors_max,
    N_pop = N_pop,
    N_samples = N_samples,
    alpha = alpha,
    cor_method = cor_method,
    max_iter = max_iter
  )

  # The RMSE curve is plotted over every tested factor count, including the first
  # count whose lack of significant improvement stopped the search (one beyond the
  # retained number), so the flattening that drives the decision stays visible.
  # When the search instead ran to n_factors_max, all columns are populated.
  n_tested <- min(n_factors + 1, n_factors_max)

  # single record: the mean RMSE curve over candidate factor counts (the full
  # per-sample RMSE matrix is kept in rmse_eigenvalues)
  results <- list(list(
    name = "CD",
    label = "Suggested number of factors",
    n_factors = n_factors,
    plot_type = "eigen",
    x = seq_len(n_tested),
    y = colMeans(RMSE_eigvals)[seq_len(n_tested)],
    highlight = if (n_factors >= 1) n_factors else NULL,
    y_label = "RMSE eigenvalues",
    rmse_eigenvalues = RMSE_eigvals
  ))

  out <- .new_efa_retention(
    "CD",
    results = results,
    settings = settings
  )

  return(out)

}

# CD is stochastic; seed and use a small simulation for a stable, fast test
set.seed(123)
cd_grips <- efa_cd(GRiPS_raw, N_pop = 1000, N_samples = 100)

test_that("output class and dimensions are correct", {
  expect_s3_class(cd_grips, "efa_retention")
  expect_length(cd_grips, 6)
  expect_named(cd_grips$n_factors, "CD")

  rec <- .retention_record(cd_grips, "CD")
  expect_equal(rec$plot_type, "eigen")
  expect_equal(rec$y_label, "RMSE eigenvalues")
  checkmate::expect_matrix(rec$rmse_eigenvalues)
})

test_that("CD returns the correct values", {
  expect_equal(cd_grips$n_factors[["CD"]], 1)

  rec <- .retention_record(cd_grips, "CD")
  # the RMSE curve runs one factor beyond the retained number (the first count
  # whose lack of significant improvement stopped the search)
  expect_length(rec$y, cd_grips$n_factors[["CD"]] + 1)
  expect_length(rec$x, cd_grips$n_factors[["CD"]] + 1)
})

test_that("the comparison-data RMSE curve is unchanged (regression)", {
  # Regression pin on the comparison-data draw: CD generates its populations
  # through the shared Ruscio-Kaczetow kernel, so pin the mean RMSE-eigenvalue
  # curve of the seeded run to catch any drift.
  #
  # The kernel is a stochastic search whose accept-or-shrink step compares two
  # floating-point RMSEs. Once the search plateaus that comparison is decided by
  # rounding, so the trajectory -- and with it this curve -- is reproducible only
  # within one BLAS; the retained factor count the method exists to report is
  # stable regardless and is asserted above. The pin therefore runs where the
  # BLAS is known rather than across every check flavour.
  skip_on_cran()
  rec <- .retention_record(cd_grips, "CD")
  # Only the tested factor counts (1 and 2) are populated; later columns stay
  # zero because the search stops after the first factor.
  expect_equal(
    colMeans(rec$rmse_eigenvalues),
    c(0.033821350175717364, 0.038807011125946637, 0, 0),
    tolerance = 1e-6
  )
})

grips_na <- GRiPS_raw
grips_na[c(1,3,5), c(2, 4)] <- NA
test_that("errors etc. are thrown correctly", {
  expect_error(efa_cd(1:10), class = "efa_input_not_matrix")
  expect_error(efa_cd(test_models$baseline$cormat), class = "efa_cd_needs_raw")

  expect_warning(efa_cd(GRiPS_raw, n_factors_max = 5, N_pop = 500, N_samples = 20), class = "efa_cd_max_factors")
  expect_warning(efa_cd(grips_na, n_factors_max = 3, N_pop = 500, N_samples = 20), class = "efa_cd_missing_removed")

  # With three indicators the one-factor model is just identified (df = 0), so no
  # over-identified model exists to compare against and CD must abort rather than
  # silently return zero factors. The check runs before any data are generated.
  expect_error(efa_cd(GRiPS_raw[, 1:3]), class = "efa_cd_min_indicators")

  # A constant variable has no correlation, which would abort the eigenvalue
  # decomposition with a base-R error; it is reported as a classed condition instead.
  grips_const <- GRiPS_raw
  grips_const[, 2] <- 3
  expect_error(efa_cd(grips_const), class = "efa_cd_constant_variable")
  # Listwise deletion can be what leaves a variable constant, so the check runs on the
  # complete cases: here only the rows kept below are non-constant in column 2.
  grips_late <- GRiPS_raw
  grips_late[, 2] <- c(1, rep(3, nrow(GRiPS_raw) - 1L))
  grips_late[1, 3] <- NA
  expect_error(suppressWarnings(efa_cd(grips_late)),
               class = "efa_cd_constant_variable")

  # A comparison-data population too small to carry a correlation is reported under
  # efa_cd's own class, naming N_pop, rather than as a condition from the shared kernel.
  expect_error(efa_cd(GRiPS_raw[1:60, ], N_pop = 1, N_samples = 2),
               class = "efa_cd_degenerate_population")
  expect_error(efa_cd(GRiPS_raw[1:60, ], N_pop = 0, N_samples = 2),
               class = "efa_cd_degenerate_population")

  # Zero-length loops used to fall through to invalid indexing/replacement errors.
  expect_error(efa_cd(GRiPS_raw, n_factors_max = 0), "Must be >= 1")
  expect_error(efa_cd(GRiPS_raw, N_samples = 0), "Must be >= 1")
  expect_error(efa_cd(GRiPS_raw, max_iter = 0), "Must be >= 1")
})

rm(cd_grips, grips_na)

#' Bartlett's test of sphericity
#'
#' This function tests whether a correlation matrix is significantly different
#' from an identity matrix (Bartlett, 1951). If the Bartlett's test is not
#' significant, the correlation matrix is not suitable for factor analysis
#' because the variables show too little covariance.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. The number of observations. Needs only be specified if a
#' correlation matrix is used.
#' @param use character. The missing-data policy for raw data. Passed to
#' [stats::cor()] for `"pearson"`, `"spearman"`, and `"kendall"`; for `"poly"` /
#' `"tetra"` the same policies are applied to the raw data before the polychoric
#' estimation, where `"all.obs"` and `"everything"` abort on a missing value instead
#' of returning `NA` correlations. Default is "pairwise.complete.obs".
#' @param cor_method character. Correlation computed from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (passed to [stats::cor()]), or `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations of ordinal / binary data
#'   (a two-step estimator).
#' Default is "pearson".
#'
#' @details Bartlett (1951) proposed this statistic to determine a correlation
#' matrix' suitability for factor analysis. The statistic is approximately
#' chi square distributed with \eqn{df = \frac{p(p - 1)}{2}} and is given by
#'
#' \deqn{chi^2 = -log(det(R)) (N - 1 - (2 * p + 5)/6)}
#'
#' where \eqn{det(R)} is the determinant of the correlation matrix, \eqn{N} is
#' the sample size, and \eqn{p} is the number of variables.
#'
#' This test requires multivariate normality. If this condition is not met,
#' the Kaiser-Meyer-Olkin criterion ([efa_kmo()])
#' can still be used.
#'
#' This function was heavily influenced by the [psych::cortest.bartlett()] function from the psych package.
#'
#' The `efa_bartlett` function can also be called together with the
#'  ([efa_kmo()]) function and with factor retention criteria
#'  in the [efa_retain()] function.
#'
#' @return A list containing
#' \item{chisq}{The chi square statistic, or `NA`, with a warning, if `N` is too
#'   small for the Bartlett correction (i.e. \eqn{N - 1 - (2p + 5)/6 \le 0}).}
#' \item{p_value}{The p value of the chi square statistic, or `NA` when `chisq`
#'   is `NA`.}
#' \item{df}{The degrees of freedom for the chi square statistic.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Bartlett, M. S. (1951). The effect of standardization on a Chi-square
#' approximation in factor analysis. Biometrika, 38, 337-344.
#'
#' @family factor analysis suitability
#'
#' @seealso [efa_kmo()] for another measure to determine
#'  suitability for factor analysis.
#'
#'  [efa_retain()] as a wrapper function for this function,
#'  [efa_kmo()] and several factor retention criteria.
#'
#' @export
#'
#' @examples
#' efa_bartlett(test_models$baseline$cormat, N = 500)
#'
efa_bartlett <- function(x, N = NA, use = c("pairwise.complete.obs", "all.obs",
                                            "complete.obs", "everything",
                                            "na.or.complete"),
                         cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")){

  # Perform argument checks
  .assert_cor_input(x)

  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  checkmate::assert_count(N, na.ok = TRUE)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(
    x, N = N, use = use, cor_method = cor_method, N_policy = "required",
    singular_tail = "Bartlett's test cannot be executed",
    N_required_msg = c("{.arg N} is {.val NA}; Bartlett's test could not be executed.",
                       "i" = "Provide {.arg N} or raw data."))
  R <- prep$R
  N <- prep$N

  # Calculate test statistic. Bartlett's test of sphericity is the likelihood-ratio
  # test of the independence model, i.e. the null-model chi-square shared with the
  # CFI baseline in .gof().
  p <- nrow(R)
  statistic <- .null_chisq(R, N)
  df <- p * (p - 1)/2

  # An undefined statistic needs a reason at the point of use; the helper owns both the
  # condition and the wording, shared with efa_screen().
  if (is.na(statistic)) .warn_bartlett_n_too_small(N, p)

  pval <- stats::pchisq(statistic, df, lower.tail = FALSE)

  # prepare the output
  settings <- list(N = N,
                   use = use,
                   cor_method = cor_method)

  output <- list(chisq = statistic,
                 p_value = pval,
                 df = df,
                 settings = settings)

  # the trailing legacy class is load-bearing: it keeps `inherits(x, "BARTLETT")`
  # working in code written against the superseded name. Do not drop it.
  class(output) <- c("efa_bartlett", "BARTLETT")

  return(output)

}

# Procrustes rotation and the consensus-target engine: input validation, the closed-form
# orthogonal Procrustes solution, congruence/hyperplane diagnostics, and the single-start
# GPA-consensus iterator behind .gpa_consensus_target().

#' Average a list of matrices elementwise
#'
#' @param x List of conformable matrices.
#'
#' @returns A matrix with the same dimensions as the inputs.
#'
#' @noRd
.average_matrices <- function(x) {
  if (!is.list(x) || length(x) < 1L) {
    cli::cli_abort("{.arg x} must be a non-empty list of conformable matrices.",
                   class = "efa_not_matrix_list")
  }
  out <- Reduce(`+`, x) / length(x)
  dimnames(out) <- dimnames(x[[1L]])
  out
}


# Internal scalar checks for Procrustes routines
#
# These helpers intentionally use base R rather than checkmate. They are used
# by both the public wrappers and the internal consensus engine, so keeping the
# checks local gives clearer error messages and avoids partially entering the
# compiled optimizer with invalid control values.
.procrustes_is_scalar <- function(x) {
  is.numeric(x) && length(x) == 1L && !is.na(x) && is.finite(x)
}

.procrustes_check_numeric_scalar <- function(x, name, lower = -Inf, upper = Inf,
                                             lower_closed = TRUE,
                                             upper_closed = TRUE) {
  if (!.procrustes_is_scalar(x)) {
    cli::cli_abort("{.arg {name}} must be a finite numeric scalar.",
                   class = "efa_not_scalar")
  }

  lower_ok <- if (lower_closed) x >= lower else x > lower
  upper_ok <- if (upper_closed) x <= upper else x < upper

  if (!lower_ok || !upper_ok) {
    left <- if (lower_closed) "[" else "("
    right <- if (upper_closed) "]" else ")"
    cli::cli_abort("{.arg {name}} must be in {left}{lower}, {upper}{right}.",
                   class = "efa_out_of_range")
  }

  x
}

.procrustes_check_integer_scalar <- function(x, name, lower = -Inf, upper = Inf) {
  if (!.procrustes_is_scalar(x) || abs(x - round(x)) > sqrt(.Machine$double.eps)) {
    cli::cli_abort("{.arg {name}} must be a finite integer-like scalar.",
                   class = "efa_not_integer")
  }
  x <- as.integer(round(x))
  if (x < lower || x > upper) {
    cli::cli_abort("{.arg {name}} must be an integer in [{lower}, {upper}].",
                   class = "efa_out_of_range")
  }
  x
}

.procrustes_check_flag <- function(x, name) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    cli::cli_abort("{.arg {name}} must be {.code TRUE} or {.code FALSE}.",
                   class = "efa_not_flag")
  }
  x
}

.procrustes_as_matrix <- function(x, name) {
  if (inherits(x, "LOADINGS")) {
    x <- unclass(x)
  }

  x <- as.matrix(x)
  if (!is.numeric(x)) {
    cli::cli_abort("{.arg {name}} must be a numeric matrix.", class = "efa_not_matrix")
  }
  storage.mode(x) <- "double"

  if (length(dim(x)) != 2L || any(dim(x) < 1L)) {
    cli::cli_abort("{.arg {name}} must have at least one row and one column.",
                   class = "efa_empty_matrix")
  }
  if (any(!is.finite(x))) {
    cli::cli_abort("{.arg {name}} must contain only finite values.",
                   class = "efa_nonfinite")
  }

  x
}

.procrustes_validate_matrix_pair <- function(A, B) {
  A <- .procrustes_as_matrix(A, "A")
  B <- .procrustes_as_matrix(B, "Target")

  if (!identical(dim(A), dim(B))) {
    cli::cli_abort("{.arg A} and {.arg Target} must have identical dimensions.",
                   class = "efa_dim_mismatch")
  }

  list(A = A, B = B)
}

.procrustes_validate_matrix_list <- function(x, name, min_length = 1L,
                                             expected_dim = NULL) {
  if (!is.list(x) || length(x) < min_length) {
    cli::cli_abort("{.arg {name}} must be a list with at least {min_length} matri{?x/ces}.",
                   class = "efa_not_matrix_list")
  }

  out <- lapply(seq_along(x), function(i) {
    .procrustes_as_matrix(x[[i]], paste0(name, "[[", i, "]]"))
  })

  ref_dim <- if (is.null(expected_dim)) dim(out[[1L]]) else expected_dim
  ok_dims <- vapply(out, function(z) identical(dim(z), ref_dim), logical(1L))
  if (!all(ok_dims)) {
    bad <- paste(which(!ok_dims), collapse = ", ")
    cli::cli_abort(
      c("All matrices in {.arg {name}} must have dimensions {paste(ref_dim, collapse = ' x ')}.",
        "x" = "Offending element(s): {bad}."),
      class = "efa_dim_mismatch"
    )
  }

  out
}

.procrustes_default_names <- function(n, prefix) {
  paste0(prefix, seq_len(n))
}

.procrustes_names <- function(A, B) {
  p <- nrow(A)
  k <- ncol(A)

  item_names <- rownames(B)
  if (is.null(item_names)) item_names <- rownames(A)
  if (is.null(item_names)) item_names <- .procrustes_default_names(p, "V")

  target_factor_names <- colnames(B)
  if (is.null(target_factor_names)) target_factor_names <- colnames(A)
  if (is.null(target_factor_names)) target_factor_names <- .procrustes_default_names(k, "F")

  source_factor_names <- colnames(A)
  if (is.null(source_factor_names)) source_factor_names <- target_factor_names

  list(
    items = item_names,
    source_factors = source_factor_names,
    target_factors = target_factor_names
  )
}

.procrustes_apply_dimnames <- function(out, A, B) {
  dn <- .procrustes_names(A, B)

  if (!is.null(out$loadings)) {
    dimnames(out$loadings) <- list(dn$items, dn$target_factors)
  }
  if (!is.null(out$T)) {
    dimnames(out$T) <- list(dn$source_factors, dn$target_factors)
  }
  if (!is.null(out$Phi)) {
    dimnames(out$Phi) <- list(dn$target_factors, dn$target_factors)
  }

  if (!is.null(out$all_values) && is.null(names(out$all_values))) {
    start_names <- if (!is.null(out$all_start_indices)) {
      paste0("start_", out$all_start_indices)
    } else {
      paste0("start_", seq_along(out$all_values))
    }
    names(out$all_values) <- start_names
  }
  if (!is.null(out$all_converged) && is.null(names(out$all_converged)) &&
      !is.null(names(out$all_values))) {
    names(out$all_converged) <- names(out$all_values)
  }
  if (!is.null(out$all_iterations) && is.null(names(out$all_iterations)) &&
      !is.null(names(out$all_values))) {
    names(out$all_iterations) <- names(out$all_values)
  }

  out
}

# Validate the optional cross-product shortcut against the matrix it stands for.
# `S` replaces `crossprod(A)` inside the compiled oblique objective and its analytic
# gradient, so anything else makes the solver minimize a different criterion while
# still reporting `convergence` and `valid`: the returned `value` then does not
# measure the returned loadings and is not even bounded below by zero. Checking
# costs the very O(p k^2) product the argument was meant to save, which is the
# right trade against an iterative solver that performs hundreds of k x k
# factorizations -- `S` survives as a compatibility argument, not an optimization.
.procrustes_validate_crossprod <- function(S, k, A) {
  S <- .procrustes_as_matrix(S, "S")
  if (!identical(dim(S), c(k, k))) {
    cli::cli_abort("{.arg S} must be a {k} x {k} matrix.", class = "efa_dim_mismatch")
  }
  SA <- crossprod(A)
  dev <- max(abs(S - SA))
  if (dev > 1e-8 * max(1, max(abs(SA)))) {
    cli::cli_abort(
      c("{.arg S} must be {.code crossprod(A)}.",
        "x" = "Largest absolute deviation from {.code crossprod(A)}: {signif(dev, 3)}.",
        "i" = "Omit {.arg S} to have it computed from {.arg A}."),
      class = "efa_bad_crossprod"
    )
  }
  # The quadratic form depends only on the symmetric part of S, and its analytic
  # gradient assumes that same matrix. `crossprod(A)` is symmetric and the check
  # above admits only deviations far below the solver's tolerance, so symmetrizing
  # cannot move the solution; divide before adding to avoid overflow near double.xmax.
  S / 2 + t(S) / 2
}

.procrustes_validate_t_init <- function(T_init, k) {
  T_init <- .procrustes_as_matrix(T_init, "T_init")
  if (!identical(dim(T_init), c(k, k))) {
    cli::cli_abort("{.arg T_init} must be a {k} x {k} matrix.", class = "efa_dim_mismatch")
  }
  if (any(sqrt(colSums(T_init * T_init)) <= sqrt(.Machine$double.eps))) {
    cli::cli_abort("{.arg T_init} must not contain zero or near-zero columns.",
                   class = "efa_zero_column")
  }
  if (qr(T_init)$rank < k) {
    cli::cli_abort("{.arg T_init} must be nonsingular.", class = "efa_singular")
  }
  T_init
}

.procrustes_validate_oblique_controls <- function(oblique_eps,
                                                  oblique_maxit,
                                                  oblique_max_line_search,
                                                  oblique_step0,
                                                  oblique_normalize,
                                                  oblique_random_starts,
                                                  oblique_screen_keep,
                                                  oblique_triage_maxit,
                                                  oblique_triage_improve_tol) {
  list(
    oblique_eps = .procrustes_check_numeric_scalar(
      oblique_eps, "oblique_eps", lower = 0, lower_closed = FALSE
    ),
    oblique_maxit = .procrustes_check_integer_scalar(
      oblique_maxit, "oblique_maxit", lower = 0L
    ),
    oblique_max_line_search = .procrustes_check_integer_scalar(
      oblique_max_line_search, "oblique_max_line_search", lower = 0L
    ),
    oblique_step0 = .procrustes_check_numeric_scalar(
      oblique_step0, "oblique_step0", lower = 0, lower_closed = FALSE
    ),
    oblique_normalize = .procrustes_check_flag(oblique_normalize, "oblique_normalize"),
    oblique_random_starts = .procrustes_check_integer_scalar(
      oblique_random_starts, "oblique_random_starts", lower = 0L
    ),
    oblique_screen_keep = .procrustes_check_integer_scalar(
      oblique_screen_keep, "oblique_screen_keep", lower = 0L
    ),
    oblique_triage_maxit = .procrustes_check_integer_scalar(
      oblique_triage_maxit, "oblique_triage_maxit", lower = 0L
    ),
    oblique_triage_improve_tol = .procrustes_check_numeric_scalar(
      oblique_triage_improve_tol, "oblique_triage_improve_tol", lower = 0
    )
  )
}

.procrustes_validate_start_index <- function(start, n_targets) {
  if (!.procrustes_is_scalar(start) || abs(start - round(start)) > sqrt(.Machine$double.eps)) {
    cli::cli_abort("{.arg start} must be either an integer index or a target matrix.",
                   class = "efa_bad_start")
  }
  start <- as.integer(round(start))
  if (start < 1L || start > n_targets) {
    cli::cli_abort("{.arg start} must be between 1 and {n_targets}.",
                   class = "efa_out_of_range")
  }
  start
}

.procrustes_validate_starts <- function(starts, n_targets) {
  if (is.null(starts)) {
    return(seq_len(n_targets))
  }
  if (!is.numeric(starts) || length(starts) < 1L || anyNA(starts) ||
      any(!is.finite(starts)) || any(abs(starts - round(starts)) > sqrt(.Machine$double.eps))) {
    cli::cli_abort("{.arg starts} must be a non-empty integer vector.",
                   class = "efa_bad_start")
  }
  starts <- as.integer(round(starts))
  if (any(starts < 1L | starts > n_targets)) {
    cli::cli_abort("All entries in {.arg starts} must be between 1 and {n_targets}.",
                   class = "efa_out_of_range")
  }
  if (anyDuplicated(starts)) {
    cli::cli_warn("Duplicate entries in {.arg starts} were removed.",
                  class = "efa_duplicate_starts")
    starts <- unique(starts)
  }
  starts
}

#' Count near-zero loadings
#'
#' Hyperplane count is the number of loadings with absolute value smaller than a
#' user-specified cutoff.
#'
#' @param L Numeric loading matrix.
#' @param cutoff Numeric scalar. Loadings with `abs(L) < cutoff` are counted as
#'   being in the hyperplane.
#'
#' @returns A list with the total hyperplane count and counts by factor and item.
#'
#' @noRd
.hyperplane_count <- function(L, cutoff = 0.15) {
  L <- .procrustes_as_matrix(L, "L")
  cutoff <- .procrustes_check_numeric_scalar(cutoff, "cutoff", lower = 0)

  in_hyperplane <- abs(L) < cutoff
  list(
    total = sum(in_hyperplane),
    by_factor = colSums(in_hyperplane),
    by_item = rowSums(in_hyperplane),
    cutoff = cutoff
  )
}

#' Mean squared discrepancy to a consensus target
#'
#' @param aligned_loadings List of aligned loading matrices.
#' @param target Consensus target matrix.
#'
#' @returns Mean sum of squared deviations from the target across matrices.
#'
#' @noRd
.consensus_loss <- function(aligned_loadings, target) {
  # Inputs are validated by the public/internal Procrustes entry points. Keeping
  # this inner-loop helper allocation-light matters for consensus and bootstrap
  # workflows where the loss is evaluated many times.
  mean(vapply(aligned_loadings, function(L) {
    sum((L - target)^2)
  }, numeric(1L)))
}


# Bare closed-form orthogonal Procrustes transform: the orthogonal `T` minimizing
# `||A %*% T - B||_F`. For warm starts where `A` and `B` are already validated;
# `.orthogonal_procrustes()` wraps this with input validation and full output.
.procrustes_orthogonal_T <- function(A, B) {
  s <- svd(crossprod(A, B))
  s$u %*% t(s$v)
}

#' Closed-form orthogonal Procrustes rotation
#'
#' Rotate `A` to the orthogonal target `B` by minimizing
#' `||A %*% T - B||_F^2` subject to `t(T) %*% T = I`.
#'
#' @param A Numeric matrix to be rotated.
#' @param B Numeric target matrix with the same dimensions as `A`.
#'
#' @returns A list with the rotated loadings, orthogonal transformation matrix,
#'   target criterion value, and basic diagnostics.
#'
#' @references
#' Schoenemann, P. H. (1966). A generalized solution of the orthogonal
#' Procrustes problem. *Psychometrika*, 31, 1-10.
#'
#' @keywords internal
.orthogonal_procrustes <- function(A, B) {
  mats <- .procrustes_validate_matrix_pair(A, B)
  A <- mats$A
  B <- mats$B

  Tmat <- .procrustes_orthogonal_T(A, B)

  Lrot <- A %*% Tmat
  value <- 0.5 * sum((Lrot - B)^2)

  out <- list(
    loadings = Lrot,
    T = Tmat,
    Phi = diag(ncol(A)),
    value = value,
    convergence = TRUE,
    # the closed-form orthogonal solution always exists; report the same `valid`
    # field the oblique path returns so callers can test the contract uniformly
    valid = TRUE,
    iterations = 0L,
    kappa_T = 1,
    Table = cbind(iter = 0, f = value, log10_s = NA_real_, step = NA_real_),
    method = "orthogonal_procrustes",
    best_start_index = 1L,
    all_start_indices = 1L,
    all_values = value,
    all_converged = TRUE,
    all_iterations = 0L,
    line_search_failed = FALSE
  )

  .procrustes_apply_dimnames(out, A, B)
}


#' Internal single-start GPA-consensus engine
#'
#' Performs a single GPA-consensus run from one starting target. The
#' multi-start wrapper [.gpa_consensus_target()] dispatches here.
#'
#' @inheritParams .gpa_consensus_target
#'
#' @keywords internal
.consensus_target_procrustes_single <- function(unrotated_list,
                                                init_targets = NULL,
                                                rotation = c("orthogonal", "oblique"),
                                                start = 1,
                                                tol = 1e-3,
                                                loss_tol = 1e-6,
                                                loss_patience = 5,
                                                convergence = c("either", "target", "loss", "both"),
                                                min_iter = 2,
                                                max_iter = 200,
                                                alpha = 1,
                                                match_target = TRUE,
                                                hyper_cutoff = 0.15,
                                                verbose = FALSE) {
  rotation <- match.arg(rotation)
  convergence <- match.arg(convergence)

  alpha <- .procrustes_check_numeric_scalar(alpha, "alpha", lower = 0, upper = 1,
                                            lower_closed = FALSE)
  tol <- .procrustes_check_numeric_scalar(tol, "tol", lower = 0, lower_closed = FALSE)
  if (!is.null(loss_tol)) {
    loss_tol <- .procrustes_check_numeric_scalar(loss_tol, "loss_tol", lower = 0,
                                                 lower_closed = FALSE)
  }
  loss_patience <- .procrustes_check_integer_scalar(loss_patience, "loss_patience", lower = 1L)
  min_iter <- .procrustes_check_integer_scalar(min_iter, "min_iter", lower = 0L)
  max_iter <- .procrustes_check_integer_scalar(max_iter, "max_iter", lower = 1L)
  if (min_iter > max_iter) {
    cli::cli_abort("{.arg min_iter} must not exceed {.arg max_iter}.",
                   class = "efa_out_of_range")
  }
  if (is.null(loss_tol) && convergence %in% c("loss", "both")) {
    cli::cli_abort("{.arg loss_tol} must not be {.code NULL} when {.arg convergence} is {.val loss} or {.val both}.",
                   class = "efa_bad_loss_tol")
  }

  match_target <- .procrustes_check_flag(match_target, "match_target")
  verbose <- .procrustes_check_flag(verbose, "verbose")
  hyper_cutoff <- .procrustes_check_numeric_scalar(hyper_cutoff, "hyper_cutoff", lower = 0)

  unrotated_list <- .procrustes_validate_matrix_list(
    unrotated_list, "unrotated_list", min_length = 2L
  )

  m <- length(unrotated_list)
  p <- nrow(unrotated_list[[1L]])
  k <- ncol(unrotated_list[[1L]])

  if (is.null(init_targets)) {
    init_targets <- unrotated_list
  } else {
    init_targets <- .procrustes_validate_matrix_list(
      init_targets, "init_targets", min_length = 1L, expected_dim = c(p, k)
    )
  }

  if (is.numeric(start) && length(start) == 1L) {
    start_index <- .procrustes_validate_start_index(start, length(init_targets))
    target <- init_targets[[start_index]]
    start_label <- start_index
  } else {
    target <- .procrustes_as_matrix(start, "start")
    if (!identical(dim(target), c(p, k))) {
      cli::cli_abort("If {.arg start} is a matrix, it must have the same dimensions as the loadings.",
                     class = "efa_dim_mismatch")
    }
    start_index <- NA_integer_
    start_label <- "matrix"
  }

  if (!is.na(start_index)) {
    dimnames(target) <- dimnames(init_targets[[start_index]])
  }
  if (is.null(rownames(target)) || is.null(colnames(target))) {
    dn <- .procrustes_names(unrotated_list[[1L]], target)
    dimnames(target) <- list(dn$items, dn$target_factors)
  }

  hist_iter <- integer(max_iter)
  hist_rel_change <- numeric(max_iter)
  hist_loss <- numeric(max_iter)
  hist_rel_loss_change <- rep(NA_real_, max_iter)
  hist_inner_mean_value <- numeric(max_iter)
  hist_inner_failures <- integer(max_iter)
  hist_inner_max_kappa <- numeric(max_iter)
  hist_loss_stable_count <- integer(max_iter)
  hist_target_ok <- logical(max_iter)
  hist_loss_ok <- logical(max_iter)
  hist_stop_rule_met <- logical(max_iter)

  converged <- FALSE
  convergence_reason <- NA_character_
  prev_loss <- NA_real_
  loss_stable_count <- 0L
  n_history <- 0L

  for (iter in seq_len(max_iter)) {
    aligned <- vector("list", m)

    for (d in seq_len(m)) {
      aligned[[d]] <- efa_procrustes(
        A = unrotated_list[[d]],
        Target = target,
        rotation = rotation
      )
    }

    aligned_loadings <- lapply(aligned, `[[`, "loadings")
    inner_values <- vapply(aligned, `[[`, numeric(1L), "value")
    inner_converged <- vapply(aligned, function(x) isTRUE(x$convergence), logical(1L))
    inner_kappa <- vapply(aligned, function(x) {
      if (is.null(x$kappa_T)) NA_real_ else x$kappa_T
    }, numeric(1L))

    centroid <- .average_matrices(aligned_loadings)
    if (isTRUE(match_target) && k > 1L) {
      centroid <- .align_solution(L = centroid, L_target = target)$loadings
    }

    new_target <- (1 - alpha) * target + alpha * centroid
    dimnames(new_target) <- dimnames(target)

    rel_change <- norm(new_target - target, type = "F") /
      (norm(target, type = "F") + 1e-12)

    # Per-iteration loss proxy: the slices were aligned to the previous target, so
    # this scores them against the freshly updated target rather than re-aligning
    # to it. It drives the loss-stabilisation stopping rule and the history; once
    # the target has stabilised the proxy and the exact GPA loss coincide.
    current_loss <- .consensus_loss(aligned_loadings, new_target)

    rel_loss_change <- if (is.na(prev_loss)) {
      NA_real_
    } else {
      abs(prev_loss - current_loss) / (abs(prev_loss) + 1e-12)
    }

    if (!is.null(loss_tol) && !is.na(rel_loss_change) && rel_loss_change < loss_tol) {
      loss_stable_count <- loss_stable_count + 1L
    } else {
      loss_stable_count <- 0L
    }

    target_ok <- rel_change < tol
    loss_ok <- !is.null(loss_tol) && loss_stable_count >= loss_patience

    stop_rule_met <- switch(
      convergence,
      target = target_ok,
      loss = loss_ok,
      either = target_ok || loss_ok,
      both = target_ok && loss_ok
    )

    n_history <- iter
    hist_iter[iter] <- iter
    hist_rel_change[iter] <- rel_change
    hist_loss[iter] <- current_loss
    hist_rel_loss_change[iter] <- rel_loss_change
    hist_inner_mean_value[iter] <- mean(inner_values)
    hist_inner_failures[iter] <- sum(!inner_converged)
    hist_inner_max_kappa[iter] <- if (all(is.na(inner_kappa))) NA_real_ else max(inner_kappa, na.rm = TRUE)
    hist_loss_stable_count[iter] <- loss_stable_count
    hist_target_ok[iter] <- target_ok
    hist_loss_ok[iter] <- loss_ok
    hist_stop_rule_met[iter] <- stop_rule_met

    if (verbose) {
      cli::cli_inform(sprintf(
        paste0(
          "iter %d | rel_change = %.3e | loss = %.8f | ",
          "rel_loss_change = %s | stable = %d | inner_failures = %d"
        ),
        iter,
        rel_change,
        current_loss,
        ifelse(is.na(rel_loss_change), "NA", sprintf("%.3e", rel_loss_change)),
        loss_stable_count,
        sum(!inner_converged)
      ))
    }

    target <- new_target
    prev_loss <- current_loss

    if (iter >= min_iter && isTRUE(stop_rule_met)) {
      converged <- TRUE
      convergence_reason <- if (target_ok && loss_ok) {
        "target_and_loss"
      } else if (target_ok) {
        "target"
      } else if (loss_ok) {
        "loss"
      } else {
        convergence
      }
      break
    }
  }

  keep <- seq_len(n_history)
  history <- data.frame(
    iter = hist_iter[keep],
    rel_change = hist_rel_change[keep],
    loss = hist_loss[keep],
    rel_loss_change = hist_rel_loss_change[keep],
    inner_mean_value = hist_inner_mean_value[keep],
    inner_failures = hist_inner_failures[keep],
    inner_max_kappa = hist_inner_max_kappa[keep],
    loss_stable_count = hist_loss_stable_count[keep],
    target_ok = hist_target_ok[keep],
    loss_ok = hist_loss_ok[keep],
    stop_rule_met = hist_stop_rule_met[keep]
  )

  final_aligned <- vector("list", m)
  for (d in seq_len(m)) {
    final_aligned[[d]] <- efa_procrustes(
      A = unrotated_list[[d]],
      Target = target,
      rotation = rotation
    )
  }

  final_aligned_loadings <- lapply(final_aligned, `[[`, "loadings")
  final_T <- lapply(final_aligned, `[[`, "T")
  final_Phi <- lapply(final_aligned, `[[`, "Phi")
  final_values <- vapply(final_aligned, `[[`, numeric(1L), "value")
  final_converged <- vapply(final_aligned, function(x) isTRUE(x$convergence), logical(1L))
  final_kappa <- vapply(final_aligned, function(x) {
    if (is.null(x$kappa_T)) NA_real_ else x$kappa_T
  }, numeric(1L))

  pooled_loadings <- .average_matrices(final_aligned_loadings)
  dimnames(pooled_loadings) <- dimnames(target)

  pooled_phi <- if (rotation == "orthogonal" || k == 1L) {
    diag(k)
  } else {
    .average_matrices(final_Phi)
  }
  pooled_phi <- (pooled_phi + t(pooled_phi)) / 2
  diag(pooled_phi) <- 1
  dimnames(pooled_phi) <- list(colnames(target), colnames(target))

  final_loss <- .consensus_loss(final_aligned_loadings, target)

  outer_converged <- converged
  final_inner_converged <- all(final_converged)
  overall_converged <- outer_converged && final_inner_converged
  reported_convergence_reason <- if (overall_converged) {
    convergence_reason
  } else if (!outer_converged) {
    convergence_reason
  } else {
    "final_inner_alignment"
  }

  list(
    converged = overall_converged,
    outer_converged = outer_converged,
    final_inner_converged = final_inner_converged,
    convergence_reason = reported_convergence_reason,
    outer_convergence_reason = convergence_reason,
    iterations = nrow(history),
    tol = tol,
    loss_tol = loss_tol,
    loss_patience = loss_patience,
    convergence = convergence,
    min_iter = min_iter,
    alpha = alpha,
    match_target = match_target,
    rotation = rotation,
    start = start_label,
    start_index = start_index,
    target = target,
    history = history,
    aligned_loadings = final_aligned_loadings,
    aligned_phi = final_Phi,
    transformations = final_T,
    pooled_loadings = pooled_loadings,
    pooled_phi = pooled_phi,
    mean_loss = final_loss,
    final_values = final_values,
    final_converged = final_converged,
    final_failures = which(!final_converged),
    final_kappa_T = final_kappa,
    hyperplane_target = .hyperplane_count(target, cutoff = hyper_cutoff),
    hyperplane_pooled = .hyperplane_count(pooled_loadings, cutoff = hyper_cutoff)
  )
}

# Number formatting and styled-table builders shared by the print methods: padded decimal
# formatting, the LOADINGS/efa_compare cell renderers, and small settings/CI string helpers.

# Apply named logical styles to a pre-assembled string, mapping each to its cli primitive
# (in order). A no-op when `style` is empty or when colours are off (the cli primitives are
# themselves no-ops then). One home for all print colouring. Blue is
# intentionally absent: rules and labels use cli's default colour.
.efa_style <- function(x, style = NULL) {
  styles <- list(red = cli::col_red, green = cli::col_green, yellow = cli::col_yellow,
                 bold = cli::style_bold, italic = cli::style_italic)
  for (s in style) {
    x <- styles[[s]](x)
  }
  x
}

# Canonical number formatter for printed output: rounds to `digits`, drops the leading 0 of
# values in (-1, 1) unless `print_zero`, optionally left-pads to a common width, and renders
# non-finite values as `na` (default "NA"; pass "" to blank structurally-absent cells, as
# base R's na.print does). Accepts a scalar, vector, or matrix and returns the same shape
# (dimnames preserved).
.efa_num <- function(x, digits = 3, print_zero = FALSE, pad = TRUE, na = "NA") {

  dims <- dim(x)
  dn <- dimnames(x)
  v <- as.numeric(x)
  finite <- is.finite(v)

  out <- sprintf(paste0("%.", digits, "f"), v)
  if (isFALSE(print_zero)) {
    out <- sub("^(-?)0\\.", "\\1.", out)
  }
  # Normalise negative zero (e.g. "-.000" or "-0.000") to its unsigned form, as
  # base R and psych do; a value rounding to zero from below should not print a sign.
  out <- sub("^-([0.]+)$", "\\1", out)
  out[!finite] <- na

  if (isTRUE(pad)) {
    width <- if (isFALSE(print_zero)) digits + 2L else digits + 3L
    # `ansi_align()` returns a cli ansi_string; coerce so the result is always plain
    # character, even when unstyled.
    out <- as.character(cli::ansi_align(out, width = width, align = "right"))
  }

  if (!is.null(dims)) {
    out <- matrix(out, nrow = dims[1L], ncol = dims[2L], dimnames = dn)
  }
  out
}

# Render a numeric matrix as decimal-aligned console lines. `col_roles` tags each column. The
# data columns -- "loading", and the non-loading "corr" (factor-correlation and variance
# tables) and "compare" (efa_compare difference tables) -- are split into vertically stacked
# blocks when the table is wider than the console (or, for loadings, `max_factors_per_block`
# is set); the h2/u2 columns are auxiliary and repeat in every block, as does the row-label
# column. Individual rows are never wrapped. "corr" cells are plain; "compare" cells whose
# |x| exceeds `cutoff` are red. Salient loadings (|x| >= cutoff) are bold, weaker ones grey,
# and Heywood-relevant cells (|loading| > 1, h2 > 1, u2 < 0) red; styling is dropped when
# `color = FALSE` or colours are off. With `lower_only = TRUE` the strictly-upper triangle is
# left blank (for symmetric matrices such as factor intercorrelations); later column blocks
# omit any rows that fall entirely within that blank triangle.
.efa_format_matrix <- function(values, row_labels, col_labels, col_roles,
                               cutoff = 0, digits = 3, color = TRUE,
                               max_factors_per_block = NULL, lower_only = FALSE,
                               na = "NA") {

  values <- as.matrix(values)
  n_col <- ncol(values)
  # Auxiliary columns (h2/u2) repeat in every block; every other column is data to be split
  # across console-fitting blocks. For loading tables this matches loading vs h2/u2 exactly;
  # for non-loading tables (all "corr"/"compare") there are no aux columns, so the data
  # columns themselves are split with only the row-label column repeated.
  aux_cols <- which(col_roles %in% c("h2", "u2"))
  split_cols <- setdiff(seq_len(n_col), aux_cols)

  cell_str <- .efa_num(values, digits = digits, print_zero = FALSE, pad = FALSE, na = na)
  if (isTRUE(lower_only)) {
    cell_str[upper.tri(cell_str)] <- ""
  }

  # Display width (not code-point count) so the column maths matches cli::ansi_align below.
  row_width <- max(cli::ansi_nchar(row_labels, type = "width"), 1L)
  col_widths <- vapply(seq_len(n_col), function(j) {
    max(cli::ansi_nchar(col_labels[j], type = "width"),
        max(cli::ansi_nchar(cell_str[, j], type = "width")))
  }, integer(1L))

  blocks <- .efa_matrix_blocks(split_cols, aux_cols, row_width, col_widths,
                               max_factors_per_block)

  out <- character(0)
  multi_block <- length(blocks) > 1L

  for (bb in seq_along(blocks)) {
    cols <- blocks[[bb]]

    if (multi_block) {
      if (bb > 1L) {
        out <- c(out, "")
      }
      out <- c(out, .efa_matrix_block_label(col_labels, split_cols, cols,
                                            bb, length(blocks)))
    }

    # With `lower_only` the strict-upper triangle is blank, so a later column block's leading
    # rows can be entirely empty; render only the rows with a cell in this block's columns.
    rows <- if (isTRUE(lower_only)) {
      which(rowSums(cell_str[, cols, drop = FALSE] != "") > 0L)
    } else {
      seq_len(nrow(cell_str))
    }

    out <- c(out, .efa_matrix_block(
      cell_str = cell_str,
      values = values,
      col_roles = col_roles,
      col_labels = col_labels,
      row_labels = row_labels,
      cols = cols,
      rows = rows,
      row_width = row_width,
      col_widths = col_widths,
      cutoff = cutoff,
      color = color
    ))
  }

  # Drop trailing blank padding (e.g. the upper triangle under `lower_only`, or a centred
  # last header cell) so the lines carry no trailing whitespace. Styled cells end in an ANSI
  # reset, so this only ever trims plain padding.
  sub("[ ]+$", "", out)
}

# Assemble the header and body lines for one column block of `.efa_format_matrix()`.
.efa_matrix_block <- function(cell_str, values, col_roles, col_labels, row_labels,
                              cols, rows, row_width, col_widths, cutoff, color) {

  header_cells <- vapply(cols, function(j) {
    # Style the label first, then centre it, so the alignment padding stays outside the
    # ANSI wrap and the trailing-whitespace trim in `.efa_format_matrix()` can reach it.
    label <- if (isTRUE(color)) cli::style_bold(col_labels[j]) else col_labels[j]
    cli::ansi_align(label, width = col_widths[j], align = "center")
  }, character(1L))
  header <- paste0(strrep(" ", row_width), "  ",
                   paste(header_cells, collapse = "  "))

  body <- vapply(rows, function(ii) {
    cells <- vapply(cols, function(jj) {
      plain <- cli::ansi_align(cell_str[ii, jj], width = col_widths[jj],
                               align = "right")
      .efa_style_cell(plain, values[ii, jj], col_roles[jj], cutoff, color)
    }, character(1L))
    label <- cli::ansi_align(row_labels[ii], width = row_width, align = "left")
    paste0(label, "  ", paste(cells, collapse = "  "))
  }, character(1L))

  c(header, body)
}

# Style a single padded cell by column role (no-op when colour is off).
.efa_style_cell <- function(cell, value, role, cutoff, color) {
  if (!isTRUE(color)) {
    return(cell)
  }

  if (identical(role, "loading")) {
    if (is.na(value)) {
      return(cell)
    }
    if (abs(value) > 1) {
      return(cli::style_bold(cli::col_red(cell)))
    }
    if (abs(value) < cutoff) {
      return(cli::col_grey(cell))
    }
    return(cli::style_bold(cell))
  }

  if (identical(role, "h2") && is.finite(value) && value > 1) {
    return(cli::col_red(cell))
  }
  if (identical(role, "u2") && is.finite(value) && value < 0) {
    return(cli::col_red(cell))
  }
  # "compare" columns (efa_compare difference tables) flag salient differences: cells whose
  # absolute value exceeds `cutoff` (the efa_compare `range_red` threshold) are coloured red.
  if (identical(role, "compare") && is.finite(value) && abs(value) > cutoff) {
    return(cli::col_red(cell))
  }

  cell
}

# Split the data columns (`split_cols`: loading/corr/compare) into console-fitting blocks;
# the h2/u2 (aux) columns repeat in every block. With `max_factors_per_block` the split is
# fixed; otherwise it follows the console width. Non-loading tables have no aux columns, so
# each block holds only its own data columns (the row-label column is repeated separately).
.efa_matrix_blocks <- function(split_cols, aux_cols, row_width, col_widths,
                               max_factors_per_block = NULL) {
  if (length(split_cols) <= 1L) {
    return(list(c(split_cols, aux_cols)))
  }

  if (!is.null(max_factors_per_block)) {
    split_points <- ceiling(seq_along(split_cols) / max_factors_per_block)
    return(lapply(split(split_cols, split_points),
                  function(cols) c(cols, aux_cols)))
  }

  console_width <- .efa_console_width()
  full_width <- .efa_table_width(c(split_cols, aux_cols), row_width, col_widths)
  if (full_width <= console_width) {
    return(list(c(split_cols, aux_cols)))
  }

  blocks <- list()
  current <- integer(0)

  for (col in split_cols) {
    candidate <- c(current, col, aux_cols)
    if (length(current) > 0L &&
        .efa_table_width(candidate, row_width, col_widths) > console_width) {
      blocks[[length(blocks) + 1L]] <- c(current, aux_cols)
      current <- col
    } else {
      current <- c(current, col)
    }
  }

  if (length(current) > 0L) {
    blocks[[length(blocks) + 1L]] <- c(current, aux_cols)
  }

  blocks
}

# Greedily pack pre-assembled chunks into console-width lines, joined by single spaces. A
# chunk is never split, so a report line that has to be emitted verbatim -- because a
# "setting = 'value'" token must not be broken across a line break -- can still be packed to
# the console at the separators between its tokens. Widths are measured with
# `cli::ansi_nchar(type = "width")`, so a styled chunk is packed by what it displays rather
# than by the length of its escape sequences, and continuation lines are indented by `exdent`
# so a wrapped line still reads as one item. A single chunk wider than the console is emitted
# as it is: there is no split point inside it that would keep the token intact.
.efa_wrap_chunks <- function(chunks, width = .efa_console_width(), exdent = 2L) {
  chunks <- chunks[nzchar(chunks)]
  if (length(chunks) < 1L) {
    return(character(0))
  }

  chunk_width <- cli::ansi_nchar(chunks, type = "width")
  indent <- strrep(" ", exdent)

  out <- character(0)
  line <- chunks[1L]
  line_width <- chunk_width[1L]

  for (i in seq_along(chunks)[-1L]) {
    joined <- line_width + 1L + chunk_width[i]
    if (joined > width) {
      out <- c(out, line)
      line <- paste0(indent, chunks[i])
      line_width <- exdent + chunk_width[i]
    } else {
      line <- paste(line, chunks[i])
      line_width <- joined
    }
  }

  c(out, line)
}

# Turn "setting = 'value'" tokens into wrap chunks for `.efa_wrap_chunks()`: the comma
# separator travels with the token it follows, so a wrapped line never starts with one, and
# `end` closes the last token (e.g. with the sentence's full stop).
.efa_setting_chunks <- function(tokens, end = "") {
  if (length(tokens) < 1L) {
    return(character(0))
  }

  paste0(tokens, c(rep(",", length(tokens) - 1L), end))
}

.efa_console_width <- function() {
  width <- tryCatch(cli::console_width(), error = function(e) NA_integer_)
  if (!is.finite(width) || width < 40L) {
    return(80L)
  }
  as.integer(width)
}

# Whether console styling will actually be rendered: the caller has to ask for it AND the
# output target has to support it. cli emits no escapes at one colour (knitr, an active
# sink, NO_COLOR, a non-tty target), so `color = TRUE` alone does not mean styling is visible.
.efa_styling_visible <- function(color = TRUE) {
  isTRUE(color) && cli::num_ansi_colors() > 1L
}

# Printed width of a block: row label + two spaces + value columns + inter-column gaps.
.efa_table_width <- function(cols, row_width, col_widths) {
  row_width + 2L + sum(col_widths[cols]) + 2L * (length(cols) - 1L)
}

.efa_matrix_block_label <- function(col_labels, split_cols, cols, block, n_blocks) {
  factor_cols <- intersect(cols, split_cols)
  first <- col_labels[min(factor_cols)]
  last <- col_labels[max(factor_cols)]

  if (identical(first, last)) {
    paste0("Factors ", first, " (block ", block, "/", n_blocks, ")")
  } else {
    paste0("Factors ", first, "-", last, " (block ", block, "/", n_blocks, ")")
  }
}

.decimals <- function(x) {

  if ((is.null(dim(x)) && !(inherits(x, c("numeric", "integer")))) ||
      (!is.null(dim(x)) && !(inherits(x, c("matrix", "loadings", "LOADINGS",
                                          "SLLOADINGS"))))) {
    cli::cli_abort("{.arg x} must be a numeric vector or matrix, not {.cls {class(x)}}.",
                   class = "efa_not_numeric")
  }

  # Decimal places carried by a single value. format(scientific = FALSE) is used
  # rather than as.character() so small magnitudes (e.g. 3e-5) render as "0.00003"
  # instead of "3e-05": the scientific form has no "." to split on (or splits off a
  # meaningless exponent), which otherwise aborts with "subscript out of bounds" or
  # mis-counts. digits = 15 matches the precision as.character() uses for in-range
  # values, so the count is unchanged for ordinary loading magnitudes. decimal.mark =
  # "." pins the separator so the "." split stays correct under a comma-decimal
  # `OutDec` option, which both format() and as.character() would otherwise honour.
  n_dec <- function(ll) {
    if (is.na(ll) || abs(ll - round(ll)) <= .Machine$double.eps^0.5) {
      return(0)
    }
    frac <- strsplit(format(ll, scientific = FALSE, trim = TRUE, digits = 15,
                            decimal.mark = "."), ".", fixed = TRUE)[[1]]
    if (length(frac) < 2L) 0L else nchar(sub("0+$", "", frac[[2]]))
  }

  if (!is.null(dim(x))) {
    max(apply(x, 1:2, n_dec), na.rm = TRUE)
  } else if (length(x) > 1) {
    max(sapply(x, n_dec), na.rm = TRUE)
  } else {
    n_dec(x)
  }

}

.paste_gof_ci <- function(indices, index) {
  paste0(" [", .efa_num(indices$lower[index], digits = 2, pad = FALSE), ", ",
         .efa_num(indices$upper[index], digits = 2, pad = FALSE),
         "]")
}

#' Parallel analysis
#'
#' Various methods for performing parallel analysis. This function uses
#' [future_lapply()][future.apply::future_lapply] for which a parallel processing plan can
#' be selected. To do so, register a plan with [future::plan()], for example
#'  `future::plan(future::multisession, workers = 2)`; see examples.
#'
#' @param x matrix or data.frame. The real data to compare the simulated eigenvalues
#'  against. Must not contain variables of classes other than numeric. Can be a
#'  correlation matrix or raw data.
#' @param N numeric. The number of cases / observations to simulate. Only has to
#'  be specified if `x` is either a correlation matrix or `NULL`. If
#'  x contains raw data, `N` is found from the dimensions of `x`. Must be larger
#'  than the number of variables.
#' @param n_vars numeric. The number of variables / indicators to simulate.
#' Only has to be specified if `x` is left as `NULL` as otherwise the
#' dimensions are taken from `x`.
#' @param n_datasets numeric. The number of datasets to simulate. Must be at
#'  least 1. Default is 1000.
#' @param percent numeric. The percentile to take from the simulated eigenvalues.
#'  Default is 95.
#' @param eigen_type character. On what the eigenvalues should be found. Can be
#'  either "SMC", "PCA", or "EFA". If using "SMC", the diagonal of the correlation
#'  matrix is replaced by the squared multiple correlations (SMCs) of the
#'  indicators. If using "PCA", the diagonal values of the correlation matrices
#'  are left to be 1. If using "EFA", eigenvalues are found on the correlation
#'  matrices with the final communalities of an EFA solution as diagonal. Default
#'  is `c("PCA", "SMC", "EFA")`, i.e. all three, which costs roughly six times a
#'  single non-EFA type: `"EFA"` fits an EFA to every simulated dataset and
#'  dominates that total. Pass a single type if the run is time-critical.
#' @param use character. Passed to [stats::cor()] if raw data
#' is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. One of `"pearson"`, `"spearman"`, or `"kendall"`,
#'   passed to [stats::cor()]. `"poly"` and `"tetra"` are not supported because
#'   `PARALLEL` compares the data against simulated continuous reference data.
#' Default is "pearson".
#' @param decision_rule character. Which rule to use to determine the number of
#'  factors to retain. Default is `"means"`, which will use the average
#'  simulated eigenvalues. `"percentile"`, uses the percentiles specified
#'  in percent. `"crawford"` uses the 95th percentile for the first factor
#'  and the mean afterwards (based on Crawford et al, 2010). All three rules retain
#'  the factors up to the first observed eigenvalue that fails to exceed its
#'  reference value; an eigenvalue further down the series that rises above its own
#'  reference again therefore adds no factor. Because the average simulated
#'  eigenvalue is a lower reference than the percentile, `"means"` tends to retain
#'  more factors than the more conservative `"percentile"` rule (Glorfeld, 1995).
#' @param n_factors numeric. Number of factors to extract if "EFA" is included in
#' `eigen_type`. Default is 1.
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  [efa_fit()] fits (of both the real and the simulated data) when `"EFA"` is included in
#'  `eigen_type`. `NULL` (default) uses the [efa_fit()] defaults. The fits are unrotated, so no
#'  rotation settings apply.
#' @param ... Additional arguments passed to [efa_fit()]. For example,
#' `estimator`, to change the estimator (default is "PAF"). PAF is more
#' robust, but it will take longer compared to the other estimators
#' available ("ML" and "ULS"). The estimation tuning knobs are not passed here; they live in
#' `estimate_control`, and the standard-error arguments (`se`, `b_boot`, `ci`, `seed`) are
#' not accepted because the fits are internal steps that keep only their eigenvalues.
#'
#' @details Parallel analysis (Horn, 1965) compares the eigenvalues obtained from
#' the sample
#'  correlation matrix against those of null model correlation matrices (i.e.,
#'  with uncorrelated variables) of the same sample size. This way, it accounts
#'  for the variation in eigenvalues introduced by sampling error and thus
#'  eliminates the main problem inherent in the Kaiser-Guttman criterion
#'  ([efa_kgc()]).
#'
#'  Parallel analysis is often argued to be one of the most accurate factor
#'  retention criteria. However, for highly correlated
#'  factor structures it has been shown to underestimate the correct number of
#'  factors. The reason for this is that a null model (uncorrelated variables)
#'  is used as reference. However, when factors are highly correlated, the first
#'  eigenvalue will be much larger compared to the following ones, as
#'  later eigenvalues are conditional on the earlier ones in the sequence and thus
#'  the shared variance is already accounted in the first eigenvalue (e.g.,
#'  Braeken & van Assen, 2017).
#'
#'  The reference eigenvalues are obtained from simulated data, so the suggested number
#'  of factors varies slightly from run to run. Call [base::set.seed()] beforehand to make a
#'  run reproducible; the result is then also independent of the parallel plan set via
#'  [future::plan()], so it can be reproduced on a machine with a different number of
#'  cores. For `"PCA"` and `"SMC"` the simulation is drawn in independently seeded blocks;
#'  a block that fails -- which happens when a simulated correlation matrix is singular, so
#'  that no eigenvalues can be taken from it -- is redrawn on its own, leaving the blocks
#'  that succeeded with the draws they already made. The `"EFA"` series instead redraws the
#'  single dataset that could not be fitted; if that dataset still cannot be fitted, the
#'  call stops with an error.
#'
#'  When both `"PCA"` and `"SMC"` are requested, the two are read off the *same* simulated
#'  datasets rather than from two independent simulations: they differ only in the diagonal
#'  substituted into the simulated correlation matrix, so one set of draws serves both and
#'  the two reference series are paired dataset by dataset. A draw that cannot be used for
#'  the SMC series -- a simulated matrix with no inverse, and hence no squared multiple
#'  correlations -- is discarded for the `"PCA"` series as well, so that the pairing stays
#'  exact. `"EFA"` fits a model to each simulated dataset and draws its own.
#'
#'  The `efa_parallel` function can also be called together with other factor
#'  retention criteria in the [efa_retain()] function.
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). Its main fields are:
#' \item{n_factors}{A named numeric vector with the suggested number of factors for
#'   each requested eigenvalue type (`"PCA"`, `"SMC"`, and/or `"EFA"`). These are
#'   `NA` when no real data are supplied (i.e. only `N` and `n_vars` are given). When
#'   every observed eigenvalue exceeds its reference value (no crossing is found), all
#'   `n_vars` components are retained and a warning is issued.}
#' \item{results}{A list with one record per eigenvalue type, each holding the
#'   observed eigenvalues (when real data were supplied) and the simulated reference
#'   values (means and percentiles) used for printing and plotting.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Braeken, J., & van Assen, M. A. (2017). An empirical Kaiser criterion.
#' Psychological Methods, 22, 450--466. https://doi.org/10.1037/met0000074
#'
#' @source Crawford, A. V., Green, S. B., Levy, R., Lo, W. J., Scott, L.,
#' Svetina, D., & Thompson, M. S. (2010). Evaluation of parallel analysis methods
#' for determining the number of factors. Educational and Psychological
#' Measurement, 70(6), 885-901.
#'
#' @source Glorfeld, L. W. (1995). An improvement on Horn's parallel analysis
#' methodology for selecting the correct number of factors to retain. Educational
#' and Psychological Measurement, 55(3), 377-393.
#'
#' @source Horn, J. L. (1965). A rationale and test for the number of factors in
#' factor analysis. Psychometrika, 30(2), 179--185. https://doi.org/10.1007/BF02289447
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#' @examples
#' \donttest{
#' # example without real data
#' pa_unreal <- efa_parallel(N = 500, n_vars = 10, n_datasets = 100)
#'
#' # example with correlation matrix with all eigen_types and PAF estimation
#' pa_paf <- efa_parallel(test_models$case_11b$cormat, N = 500, n_datasets = 100)
#'
#' # example with correlation matrix with all eigen_types and ML estimation
#' # this will be faster than the above with PAF)
#' pa_ml <- efa_parallel(test_models$case_11b$cormat, N = 500, estimator = "ML",
#'                       n_datasets = 100)
#'}
#'
#'\dontrun{
#' # for parallel computation. future::plan() returns the plan it replaces, so
#' # on.exit() puts the session back as it was -- also if the call fails.
#' pa_faster <- local({
#'   old_plan <- future::plan(future::multisession, workers = 2)
#'   on.exit(future::plan(old_plan), add = TRUE)
#'   efa_parallel(test_models$case_11b$cormat, N = 500)
#' })
#' }

efa_parallel <- function(x = NULL,
                     N = NA,
                     n_vars = NA,
                     n_datasets = 1000,
                     percent = 95,
                     eigen_type = c("PCA", "SMC", "EFA"),
                     use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                             "everything", "na.or.complete"),
                     cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                     decision_rule = c("means", "percentile", "crawford"),
                     n_factors = 1,
                     estimate_control = NULL,
                     ...) {

  .reject_flat_knobs(...names(), fn = "efa_parallel")
  .reject_unknown_fit_dots(...names(), fn = "efa_parallel", unrotated = TRUE)
  .reject_rotation_dots(list(...), fn = "efa_parallel")

  if(!is.null(x) && !inherits(x, c("matrix", "data.frame"))){

    cli::cli_abort(
      c("{.arg x} must be {.code NULL}, a correlation matrix, or a data frame/matrix of raw data.",
        "x" = "You supplied {.obj_type_friendly {x}}."),
      class = "efa_input_not_matrix"
    )

  }
  eigen_type <- .match_arg_ci(eigen_type, several.ok = TRUE)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  .reject_poly_reference(cor_method, "efa_parallel")
  decision_rule <- .match_arg_ci(decision_rule)
  .assert_estimate_control(estimate_control)
  .assert_args({
    checkmate::assert_count(n_factors, positive = TRUE)
    checkmate::assert_count(N, na.ok = TRUE, positive = TRUE)
    checkmate::assert_count(n_vars, na.ok = TRUE, positive = TRUE)
    checkmate::assert_count(n_datasets, positive = TRUE)
    checkmate::assert_number(percent, lower = 0, upper = 100)
  })

  # The simulated datasets are drawn in chunks, one future per chunk, under
  # future.seed = TRUE -- which assigns one L'Ecuyer stream per element of the chunk
  # vector. The number of chunks therefore has to be independent of the number of
  # workers: deriving it from nbrOfWorkers() would make the per-chunk streams, and hence
  # the reference eigenvalues, differ between a sequential and a multisession plan for the
  # same set.seed(). A fixed chunk count keeps a seeded run reproducible on any plan. The
  # trade-off is that the granularity no longer adapts to the pool: a plan with more than
  # 20 workers leaves the surplus idle, so a very wide pool is slower than it would be with
  # worker-matched chunking. 20 is the compromise -- enough chunks to keep a typical pool
  # busy, few enough that the per-chunk dispatch stays negligible. The chunk count never
  # exceeds n_datasets, so no chunk is empty; the lower bound of one is a backstop only,
  # since n_datasets is refused above unless it is at least one.
  size_vec <- .parallel_chunks(n_datasets, max(1L, min(n_datasets, 20L)))

  # Prepare objects
  results_PCA <- NA
  results_SMC <- NA
  results_EFA <- NA
  eigvals_real_PCA <- NA
  eigvals_real_SMC <- NA
  eigvals_real_EFA <- NA
  n_fac_PCA <- NA
  n_fac_SMC <- NA
  n_fac_EFA <- NA
  x_dat <- FALSE

  if (!is.null(x)){

      .assert_cor_input(x)

      if (!is.na(n_vars)) {
        cli::cli_warn(
          c("Both {.arg n_vars} and {.arg x} were supplied.",
            "i" = "Taking {.arg n_vars} from {.arg x}."),
          class = "efa_nvars_from_data"
        )
      }
      n_vars <- ncol(x)
      x_dat <- TRUE

      # Detect or compute the correlation matrix, check it, and smooth it if needed
      prep <- .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                                 N_policy = "optional",
                                 singular_tail = "parallel analysis is not possible")
      R <- prep$R
      N <- prep$N
      eigvals_R <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

      if ("PCA" %in% eigen_type) {
        eigvals_real_PCA <- matrix(eigvals_R, ncol = 1)
        colnames(eigvals_real_PCA) <- "Real Eigenvalues"
      }

      if ("SMC" %in% eigen_type) {
        # compute smcs
        R_SMC <- R
        diag(R_SMC) <- .smc_start(R)
        eigvals_real_SMC <- matrix(eigen(R_SMC, symmetric = TRUE,
                                     only.values = TRUE)$values, ncol = 1)
        colnames(eigvals_real_SMC) <- "Real Eigenvalues"
      }

      if ("EFA" %in% eigen_type) {
        # Internal fit used only for its eigenvalues; suppress its warnings so a
        # forwarded estimator that does not converge does not raise a warning from
        # inside efa_parallel().
        eigvals_real_EFA <- matrix(suppressWarnings(
          efa_fit(R, n_factors = n_factors, N = N,
                  estimate_control = estimate_control, ...)$final_eigen),  ncol = 1)
        colnames(eigvals_real_EFA) <- "Real Eigenvalues"
      }

  }

  if (is.na(n_vars)) {
    cli::cli_abort(
      c("{.arg n_vars} was not set and could not be taken from the data.",
        "i" = "Specify {.arg n_vars} and try again."),
      class = "efa_nvars_required"
    )
  }

  if (is.na(N)) {

    cli::cli_abort(
      c("{.arg N} was not set and could not be taken from the data.",
        "i" = "Specify {.arg N} and try again."),
      class = "efa_n_required"
    )

  }

  .assert_n_gt_vars(N, n_vars)

    # PCA and SMC differ only in the diagonal substituted into the same simulated
    # correlation matrix, so when both are requested they are read off one set of
    # simulated datasets instead of two independent ones -- which halves the simulation
    # and pairs the two reference series dataset by dataset.
    if (all(c("PCA", "SMC") %in% eigen_type)) {

      eigvals_both <- .parallel_sim_chunks(size_vec, label = "PCA and SMCs", N = N,
                                           n_vars = n_vars, eigen_type = 3,
                                           cor_method = cor_method,
                                           maxit = n_datasets * 10)
      eigvals_PCA <- eigvals_both[, seq_len(n_vars), drop = FALSE]
      eigvals_SMC <- eigvals_both[, n_vars + seq_len(n_vars), drop = FALSE]

    } else if ("PCA" %in% eigen_type) {

      eigvals_PCA <- .parallel_sim_chunks(size_vec, label = "PCA", N = N,
                                          n_vars = n_vars, eigen_type = 1,
                                          cor_method = cor_method)

    } else if ("SMC" %in% eigen_type) {

      eigvals_SMC <- .parallel_sim_chunks(size_vec, label = "SMCs", N = N,
                                          n_vars = n_vars, eigen_type = 2,
                                          cor_method = cor_method,
                                          maxit = n_datasets * 10)

    }

    if ("PCA" %in% eigen_type) {

      results_PCA <- .parallel_summarise(eigvals_PCA, percent = percent,
                                        n_vars = n_vars)

      colnames(results_PCA) <- c("Means", paste(percent, "Percentile"))

      if (isTRUE(x_dat)) {
        n_fac_PCA <- .determine_factors(decision_rule = decision_rule,
                                        eigvals_real = eigvals_real_PCA,
                                        results = results_PCA,
                                        percent = percent)
      }

    }

    if ("SMC" %in% eigen_type) {

      results_SMC <- .parallel_summarise(eigvals_SMC, percent = percent,
                                        n_vars = n_vars)

      colnames(results_SMC) <- c("Means", paste(percent, "Percentile"))

      if (isTRUE(x_dat)) {
      n_fac_SMC <- .determine_factors(decision_rule = decision_rule,
                                      eigvals_real = eigvals_real_SMC,
                                      results = results_SMC,
                                      percent = percent)
      }

    }

    if ("EFA" %in% eigen_type) {

      eigvals_EFA <- future.apply::future_lapply(size_vec, .parallel_EFA_sim,
                                             n_vars = n_vars, N = N,
                                             n_factors = n_factors,
                                             cor_method = cor_method,
                                             estimate_control = estimate_control, ...,
                                             future.seed = TRUE)
      eigvals_EFA <- do.call(rbind, eigvals_EFA)

      results_EFA <- .parallel_summarise(eigvals_EFA, percent = percent,
                                        n_vars = n_vars)

      colnames(results_EFA) <- c("Means", paste(percent, "Percentile"))

      if (isTRUE(x_dat)) {
      n_fac_EFA <- .determine_factors(decision_rule = decision_rule,
                                      eigvals_real = eigvals_real_EFA,
                                      results = results_EFA,
                                      percent = percent)
      }

    }

  settings <- list(
    x_dat = x_dat,
    N = N,
    n_vars = n_vars,
    n_datasets = n_datasets,
    percent = percent,
    eigen_type = eigen_type,
    use = use,
    cor_method = cor_method,
    decision_rule = decision_rule,
    n_factors = n_factors
  )

  # one record per requested eigenvalue type: the real eigenvalues (the solid
  # line, absent when no real data are given) plus the simulated reference series
  # (means and percentile) drawn as dashed lines
  sim_list <- list(PCA = results_PCA, SMC = results_SMC, EFA = results_EFA)
  real_list <- list(PCA = eigvals_real_PCA, SMC = eigvals_real_SMC,
                    EFA = eigvals_real_EFA)
  nfac_list <- list(PCA = n_fac_PCA, SMC = n_fac_SMC, EFA = n_fac_EFA)

  results <- list()
  for (et in c("PCA", "SMC", "EFA")) {
    if (!(et %in% eigen_type)) next
    sim <- sim_list[[et]]
    refs <- stats::setNames(lapply(seq_len(ncol(sim)), function(j) sim[, j]),
                            colnames(sim))
    if (isTRUE(x_dat)) {
      n_fac <- nfac_list[[et]]
      y <- as.numeric(real_list[[et]])
      highlight <- if (!is.na(n_fac) && n_fac >= 1) n_fac else NULL
    } else {
      # no real data: no real-eigenvalue series and no suggestion
      n_fac <- NA_real_
      y <- NULL
      highlight <- NULL
    }
    results[[et]] <- list(
      name = et,
      label = et,
      n_factors = n_fac,
      plot_type = "eigen",
      x = seq_len(n_vars),
      y = y,
      references = refs,
      highlight = highlight
    )
  }

  out <- .new_efa_retention(
    "PARALLEL",
    results = unname(results),
    settings = settings,
    subtitle = .eigen_subtitle(
      eigen_type,
      paste0(.retention_count(n_datasets), " simulated datasets")),
    note = if (isTRUE(x_dat)) {
      paste0("Number of factors retained using the \"", decision_rule,
             "\" decision rule.")
    } else {
      "No data were entered; showing the simulated eigenvalues only. No number of factors is suggested."
    }
  )

  return(out)

}


.parallel_EFA_sim <- function(n_datasets, n_vars, N, n_factors, cor_method,
                              estimate_control = NULL, ...){

  eigvals <- matrix(nrow = n_datasets, ncol = n_vars)

  # The null-model reference data are uncorrelated variables, drawn with the shared
  # multivariate-normal kernel from the identity correlation (the no-factor case).
  R_null <- diag(n_vars)

  for(i in seq_len(n_datasets)){

    x <- .simulate_cfm_mvn(R_null, N)
    R <- stats::cor(x, method = cor_method)
    eigvals_i <- try(suppressWarnings(suppressMessages(
      efa_fit(R, n_factors = n_factors, N = N,
              estimate_control = estimate_control, ...)$final_eigen)), silent = TRUE)
    it_i <- 1
    while (inherits(eigvals_i, "try-error") && it_i < 25) {
      x <- .simulate_cfm_mvn(R_null, N)
      R <- stats::cor(x, method = cor_method)
      eigvals_i <- try(suppressWarnings(suppressMessages(
        efa_fit(R, n_factors = n_factors, N = N,
                estimate_control = estimate_control, ...)$final_eigen)), silent = TRUE)
      it_i <- it_i + 1
    }

    if (inherits(eigvals_i, "try-error")) {
      cli::cli_abort(
        c("Eigenvalues from simulated data via {.val EFA} could not be found in 25 tries.",
          "i" = "This is likely due to singular matrices."),
        class = "efa_parallel_sim_failed"
      )
    }

    eigvals[i,] <- eigvals_i

  }

  return(eigvals)
}

# One simulation chunk, reporting a failure instead of raising it, so that one bad draw
# costs its own chunk rather than the whole batch (see .parallel_sim_chunks()).
.parallel_sim_eig_try <- function(n_datasets, ...) {
  try(.parallel_sim_eig(n_datasets, ...), silent = TRUE)
}

# Reference eigenvalues for the whole simulation, one future per chunk of `size_vec`, the
# chunks stacked into one matrix of n_datasets rows.
#
# A chunk can fail: the simulation refuses a draw whose correlation matrix is singular and
# gives up once it has exhausted its own per-draw budget. Such a chunk is redrawn on its own,
# up to `max_tries` times, while the chunks that already succeeded keep the draws they made.
# Repeating the whole batch instead would discard every chunk that had nothing wrong with it
# for one bad draw and -- because future.seed = TRUE spawns fresh random-number streams for
# every call -- would replace their draws as well.
#
# The try() around the batch covers the parallel backend rather than the simulation: a chunk
# reports its own failure through .parallel_sim_eig_try(), so what is caught here is a failure
# of future.apply itself (a lost worker, say), which is retried the same way.
#
# `label` names the eigenvalue type in the abort, which carries the failure that defeated the
# last attempt as its parent, so the reason a chunk kept failing is not lost.
.parallel_sim_chunks <- function(size_vec, label, ..., max_tries = 25L) {

  out <- vector("list", length(size_vec))
  todo <- seq_along(size_vec)
  cause <- NULL

  for (i in seq_len(max_tries)) {

    res <- try(future.apply::future_lapply(size_vec[todo], .parallel_sim_eig_try, ...,
                                           future.seed = TRUE),
               silent = TRUE)

    if (inherits(res, "try-error")) {
      cause <- attr(res, "condition")
      next
    }

    out[todo] <- res
    failed <- vapply(res, inherits, logical(1L), what = "try-error")
    if (!any(failed)) return(do.call(rbind, out))

    todo <- todo[failed]
    cause <- attr(out[[todo[1L]]], "condition")
  }

  cli::cli_abort(
    c("Eigenvalues from simulated data via {.val {label}} could not be found in
       {max_tries} tries.",
      "i" = "This is likely due to singular matrices."),
    class = "efa_parallel_sim_failed",
    parent = cause
  )
}

# Reference eigenvalues for one simulation chunk. Horn's (1965) parallel analysis
# requires the reference matrices to be built with the same correlation estimator
# as the observed data. The default Pearson case uses the compiled .parallel_sim()
# for its allocation-light, eigenvalue-only loop; rank-based estimators ("spearman",
# "kendall") draw the null-model data with the shared multivariate-normal kernel
# (.simulate_cfm_mvn on the identity correlation -- uncorrelated variables) and then
# correlate them with stats::cor(), so the reference matches the observed correlation
# estimator rather than always being Pearson.
.parallel_sim_eig <- function(n_datasets, n_vars, N, eigen_type, cor_method,
                              maxit = 10000) {

  if (cor_method == "pearson") {
    return(.parallel_sim(n_datasets, n_vars, N, eigen_type, maxit))
  }

  want_pca <- eigen_type %in% c(1, 3)
  want_smc <- eigen_type %in% c(2, 3)

  eig_vals <- matrix(NA_real_, nrow = n_datasets,
                     ncol = (want_pca + want_smc) * n_vars)
  smc_col <- if (want_pca) n_vars else 0L
  success <- 0L
  iter <- 0L

  # The null-model reference data are uncorrelated variables, drawn with the shared
  # kernel from the identity correlation (the no-factor case).
  R_null <- diag(n_vars)

  # One loop over the draws for both series, matching the compiled .parallel_sim(): PCA
  # and SMC differ only in the diagonal substituted into the same simulated correlation
  # matrix, so one draw serves both and a draw the SMC series cannot use is discarded
  # for the PCA series too. Only the SMC series can reject a draw and so needs the maxit
  # retry bound; the PCA series runs all n_datasets draws regardless of maxit.
  while (success < n_datasets && (!want_smc || iter < maxit)) {
    iter <- iter + 1L
    x <- .simulate_cfm_mvn(R_null, N)
    R <- stats::cor(x, method = cor_method)

    if (want_smc) { # SMC: replace the diagonal with squared multiple correlations;
      # skip the draw if the simulated matrix is singular.
      smc <- try(.smc_start(R), silent = TRUE)
      if (inherits(smc, "try-error")) next
    }

    success <- success + 1L

    if (want_pca) {
      eig_vals[success, seq_len(n_vars)] <-
        eigen(R, symmetric = TRUE, only.values = TRUE)$values
    }

    if (want_smc) {
      diag(R) <- smc
      eig_vals[success, smc_col + seq_len(n_vars)] <-
        eigen(R, symmetric = TRUE, only.values = TRUE)$values
    }
  }

  if (success < n_datasets) {
    cli::cli_abort("Could not generate enough non-singular matrices.",
                   class = "efa_parallel_sim_failed")
  }

  eig_vals
}

.determine_factors <- function(decision_rule, eigvals_real, results, percent){

# determine the number of factors to retain
if (decision_rule == "crawford") {
  # n factors from resampling
  if ("95 Percentile" %in% colnames(results)) {
    crawford <- c(results[1, "95 Percentile"],
                  results[-1, "Means"])
    n_fac <- which(!(eigvals_real > crawford))[1] - 1
  } else {
    cli::cli_warn(
      c("{.code decision_rule = \"crawford\"} was specified, but the 95th percentile was not used; using means instead.",
        "i" = "To use {.val crawford}, set {.code percent = 95}."),
      class = "efa_parallel_crawford"
    )
    n_fac <- which(!(eigvals_real > results[, "Means"]))[1] - 1
  }

} else if (decision_rule == "means") {
  n_fac <- which(!(eigvals_real > results[, "Means"]))[1] - 1

} else if (decision_rule == "percentile") {

  pp <- paste(percent, "Percentile")
  n_fac <- which(!(eigvals_real > results[, pp]))[1] - 1

}

  # When every observed eigenvalue exceeds its reference the rule finds no crossing
  # (`which()` is empty, so the index is NA). Every dimension then sits above the
  # noise reference, so retain all tested components (the same "all-exceed"
  # convention as the empirical Kaiser criterion in [efa_ekc()]) and flag the boundary
  # with a classed warning rather than returning a silent NA.
  if (is.na(n_fac)) {
    n_fac <- length(eigvals_real)
    cli::cli_warn(
      c("All observed eigenvalues exceeded their parallel analysis reference value; no crossing was found.",
        "i" = "Retaining all {n_fac} component{?s}. This often indicates a near-singular or highly collinear correlation matrix; interpret the suggestion with caution."),
      class = "efa_parallel_no_crossing"
    )
  }

  return(n_fac)

}

.parallel_summarise <- function(eig_vals, percent, n_vars) {

  results <- matrix(NA, nrow = n_vars, ncol = length(percent) + 1)
  results[, 1] <- colMeans(eig_vals)

  # percentile reference series via stats::quantile (type 7, matching
  # psych::fa.parallel) rather than a manual order statistic
  for (root in seq_len(n_vars)) {
    results[root, -1] <- stats::quantile(eig_vals[, root], probs = percent / 100,
                                         names = FALSE, na.rm = TRUE)
  }

  return(results)
}

# Split n_datasets into n_chunks non-negative integer chunks that sum to n_datasets,
# distributing the remainder one per chunk. Avoids a negative final chunk when
# n_datasets is not much larger than the number of chunks.
.parallel_chunks <- function(n_datasets, n_chunks) {
  size_vec <- rep(n_datasets %/% n_chunks, n_chunks)
  rem <- n_datasets %% n_chunks
  if (rem > 0) size_vec[seq_len(rem)] <- size_vec[seq_len(rem)] + 1
  size_vec
}

# Each efa_parallel() call simulates ~100 datasets and eigen-decomposes them, so this
# fixture block dominates the file (~50s). Skipped by default; opt in with
# `Sys.setenv(EFATOOLS_TEST_SLOW = "true")` to run. See helper-slow.R.

# The kernel-routing check below compares two separate eigen-decomposition runs, so it uses
# expect_equal() with an explicit tolerance rather than expect_identical(): a threaded BLAS
# (Apple's Accelerate, for one) is free to vary its GEMM reduction order between calls, so
# one function on one input can differ in the last ulp. waldo still compares S3 classes,
# names, attributes and structure exactly, so only numeric values are given slack -- a draw
# that did not route through the shared kernel consumes a different random-number stream and
# produces wholly different eigenvalues, so it still fails loudly. (A set tolerance does relax
# integer against double, which these eigenvalues never are.)
fp_tol <- 1e-8

if (is_slow_test()) {
# seed the parallel-analysis simulation so the retained-factor counts are
# reproducible (future_lapply uses future.seed = TRUE under the sequential plan)
set.seed(42)
pa_cor <- efa_parallel(test_models$baseline$cormat, N = 500)
pa_cor_pca <- efa_parallel(test_models$baseline$cormat, N = 500, eigen_type = "PCA")
pa_raw <- efa_parallel(GRiPS_raw)
pa_nodat <- efa_parallel(N = 20, n_vars = 5)
pa_craw <- efa_parallel(test_models$baseline$cormat, N = 500, eigen_type = "PCA",
                    decision_rule = "crawford")
pa_perc <- efa_parallel(test_models$baseline$cormat, N = 500, eigen_type = "PCA",
                    decision_rule = "percentile")
}  # is_slow_test()

test_that("output class and dimensions are correct", {
  skip_if_not_slow()
  expect_s3_class(pa_cor, "efa_retention")
  expect_length(pa_cor, 6)
  expect_s3_class(pa_raw, "efa_retention")
  expect_length(pa_raw, 6)
  expect_s3_class(pa_cor_pca, "efa_retention")
  expect_s3_class(pa_nodat, "efa_retention")
  expect_s3_class(pa_craw, "efa_retention")
  expect_s3_class(pa_perc, "efa_retention")

  expect_named(pa_cor$n_factors, c("PCA", "SMC", "EFA"))
  expect_named(pa_cor_pca$n_factors, "PCA")
  expect_named(pa_nodat$n_factors, c("PCA", "SMC", "EFA"))
  expect_equal(.retention_record(pa_cor, "PCA")$plot_type, "eigen")
})


test_that("found eigenvalues are correct", {
  skip_if_not_slow()
  # real eigenvalues form the solid line (record $y)
  expect_equal(sum(.retention_record(pa_cor, "PCA")$y),
               ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(pa_cor, "SMC")$y),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(pa_cor, "EFA")$y),
            ncol(test_models$baseline$cormat))

  # simulated eigenvalues form the dashed reference series (record $references)
  expect_equal(sum(.retention_record(pa_cor, "PCA")$references$Means),
               ncol(test_models$baseline$cormat))
  expect_gt(sum(.retention_record(pa_cor, "PCA")$references[["95 Percentile"]]),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(pa_cor, "SMC")$references$Means),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(pa_cor, "EFA")$references$Means),
            ncol(test_models$baseline$cormat))

  expect_equal(sum(.retention_record(pa_raw, "PCA")$y), ncol(GRiPS_raw))
  expect_named(.retention_record(pa_raw, "PCA")$references,
               c("Means", "95 Percentile"))
  expect_equal(sum(.retention_record(pa_raw, "PCA")$references$Means),
               ncol(GRiPS_raw))
  expect_gt(sum(.retention_record(pa_raw, "PCA")$references[["95 Percentile"]]),
            ncol(GRiPS_raw))
  expect_lt(sum(.retention_record(pa_raw, "SMC")$y), ncol(GRiPS_raw))
  expect_lt(sum(.retention_record(pa_raw, "EFA")$y), ncol(GRiPS_raw))

  # references is a named two-series list (means + percentile)
  refs <- .retention_record(pa_cor, "PCA")$references
  checkmate::expect_list(refs, len = 2)
  expect_named(refs, c("Means", "95 Percentile"))

  # only the requested eigenvalue type produces a record
  expect_null(.retention_record(pa_cor_pca, "SMC"))
  expect_null(.retention_record(pa_cor_pca, "EFA"))

  # no real data: no real-eigenvalue series, but the simulated references remain
  expect_null(.retention_record(pa_nodat, "PCA")$y)
  expect_named(.retention_record(pa_nodat, "PCA")$references,
               c("Means", "95 Percentile"))
})


test_that("identified number of factors is correct", {
  skip_if_not_slow()
  expect_equal(pa_cor$n_factors[["PCA"]], 3)
  expect_equal(pa_cor$n_factors[["SMC"]], 3)
  expect_equal(pa_cor$n_factors[["EFA"]], 7, tolerance = 2)

  expect_equal(pa_raw$n_factors[["PCA"]], 1)
  expect_equal(pa_raw$n_factors[["SMC"]], 1)
  expect_equal(pa_raw$n_factors[["EFA"]], 3, tolerance = 2)

  expect_true(all(is.na(pa_nodat$n_factors)))

  expect_equal(pa_cor_pca$n_factors[["PCA"]], 3)
  expect_equal(pa_craw$n_factors[["PCA"]], 3)
  expect_equal(pa_perc$n_factors[["PCA"]], 3)
})


test_that("settings are returned correctly", {
  skip_if_not_slow()
  expect_named(pa_cor$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                  "eigen_type", "use", "cor_method", "decision_rule",
                                  "n_factors"))
  expect_named(pa_raw$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                  "eigen_type", "use", "cor_method", "decision_rule",
                                  "n_factors"))
  expect_named(pa_cor_pca$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))
  expect_named(pa_nodat$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))
  expect_named(pa_craw$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))
  expect_named(pa_perc$settings, c("x_dat", "N", "n_vars", "n_datasets", "percent",
                                      "eigen_type", "use", "cor_method",
                                      "decision_rule", "n_factors"))

  expect_true(pa_cor$settings$x_dat)
  expect_true(pa_raw$settings$x_dat)
  expect_true(pa_cor_pca$settings$x_dat)
  expect_false(pa_nodat$settings$x_dat)
  expect_true(pa_craw$settings$x_dat)
  expect_true(pa_perc$settings$x_dat)

  expect_equal(pa_cor$settings$N, 500)
  expect_equal(pa_raw$settings$N, 810)
  expect_equal(pa_cor_pca$settings$N, 500)
  expect_equal(pa_nodat$settings$N, 20)
  expect_equal(pa_craw$settings$N, 500)
  expect_equal(pa_perc$settings$N, 500)

  expect_equal(pa_cor$settings$n_vars, 18)
  expect_equal(pa_raw$settings$n_vars, 8)
  expect_equal(pa_cor_pca$settings$n_vars, 18)
  expect_equal(pa_nodat$settings$n_vars, 5)
  expect_equal(pa_craw$settings$n_vars, 18)
  expect_equal(pa_perc$settings$n_vars, 18)

  expect_equal(pa_cor$settings$n_datasets, 1000)
  expect_equal(pa_raw$settings$n_datasets, 1000)
  expect_equal(pa_cor_pca$settings$n_datasets, 1000)
  expect_equal(pa_nodat$settings$n_datasets, 1000)
  expect_equal(pa_craw$settings$n_datasets, 1000)
  expect_equal(pa_perc$settings$n_datasets, 1000)

  expect_equal(pa_cor$settings$percent, 95)
  expect_equal(pa_raw$settings$percent, 95)
  expect_equal(pa_cor_pca$settings$percent, 95)
  expect_equal(pa_nodat$settings$percent, 95)
  expect_equal(pa_craw$settings$percent, 95)
  expect_equal(pa_perc$settings$percent, 95)

  expect_equal(pa_cor$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(pa_raw$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(pa_cor_pca$settings$eigen_type, "PCA")
  expect_equal(pa_nodat$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(pa_craw$settings$eigen_type, "PCA")
  expect_equal(pa_perc$settings$eigen_type, "PCA")

  expect_equal(pa_cor$settings$use, "pairwise.complete.obs")
  expect_equal(pa_raw$settings$use, "pairwise.complete.obs")
  expect_equal(pa_cor_pca$settings$use, "pairwise.complete.obs")
  expect_equal(pa_nodat$settings$use, "pairwise.complete.obs")
  expect_equal(pa_craw$settings$use, "pairwise.complete.obs")
  expect_equal(pa_perc$settings$use, "pairwise.complete.obs")

  expect_equal(pa_cor$settings$cor_method, "pearson")
  expect_equal(pa_raw$settings$cor_method, "pearson")
  expect_equal(pa_cor_pca$settings$cor_method, "pearson")
  expect_equal(pa_nodat$settings$cor_method, "pearson")
  expect_equal(pa_craw$settings$cor_method, "pearson")
  expect_equal(pa_perc$settings$cor_method, "pearson")

  expect_equal(pa_cor$settings$decision_rule, "means")
  expect_equal(pa_raw$settings$decision_rule, "means")
  expect_equal(pa_cor_pca$settings$decision_rule, "means")
  expect_equal(pa_nodat$settings$decision_rule, "means")
  expect_equal(pa_craw$settings$decision_rule, "crawford")
  expect_equal(pa_perc$settings$decision_rule, "percentile")

  expect_equal(pa_cor$settings$n_factors, 1)
  expect_equal(pa_raw$settings$n_factors, 1)
  expect_equal(pa_cor_pca$settings$n_factors, 1)
  expect_equal(pa_nodat$settings$n_factors, 1)
  expect_equal(pa_craw$settings$n_factors, 1)
  expect_equal(pa_perc$settings$n_factors, 1)

})

burt <- .burt_cormat()


#sim_NA <- data.frame(rnorm(30), rnorm(30), rnorm(30), rep("a", 30))

test_that("errors are thrown correctly", {
  skip_if_not_slow()
  expect_error(efa_parallel(1:5), class = "efa_input_not_matrix")
  expect_warning(suppressMessages(efa_parallel(GRiPS_raw, n_vars = 5)), class = "efa_nvars_from_data")
  expect_warning(suppressMessages(efa_parallel(GRiPS_raw, N = 20, eigen_type = "PCA")), class = "efa_n_from_data")
  expect_error(suppressMessages(efa_parallel(N = 500)), class = "efa_nvars_required")
  expect_message(efa_parallel(GRiPS_raw, eigen_type = "PCA"), class = "efa_cor_from_data")
  expect_error(efa_parallel(test_models$baseline$cormat, eigen_type = "PCA"), class = "efa_n_required")
  expect_warning(efa_parallel(test_models$baseline$cormat, N = 500,
                          eigen_type = "PCA", decision_rule = "crawford",
                          percent = 80), class = "efa_parallel_crawford")
  expect_error(efa_parallel(sing_raw), class = "efa_cor_singular")
  expect_error(efa_parallel(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_warning(efa_parallel(burt, N = 100, eigen_type = "PCA"), class = "efa_cor_smoothed")
  expect_error(efa_parallel(test_models$baseline$cormat, N = 15), class = "efa_n_too_small")
  expect_error(efa_parallel(test_models$baseline$cormat, N = 18), class = "efa_n_too_small")
})

test_that("the n_vars override names the input instead of assuming raw data", {
  # The override runs for any non-NULL `x`, so a correlation matrix reaches it as well. The
  # message used to call that input "raw data", which is the one thing a correlation matrix
  # is not. The handler exits at the warning, so no reference datasets are simulated.
  w <- tryCatch(
    efa_parallel(test_models$baseline$cormat, N = 500, n_vars = 18, n_datasets = 20),
    efa_nvars_from_data = function(w) w)
  expect_s3_class(w, "efa_nvars_from_data")
  expect_snapshot(cat(conditionMessage(w)))
})

test_that("parallel-analysis simulation counts must be positive", {
  # all four are refused at the R boundary with the package's argument class,
  # rather than reaching the simulation and answering from an empty reference
  expect_error(efa_parallel(N = 0, n_vars = 4, n_datasets = 1),
               class = "efa_invalid_argument")
  expect_error(efa_parallel(N = 20, n_vars = 0, n_datasets = 1),
               class = "efa_invalid_argument")
  expect_error(efa_parallel(N = 20, n_vars = 4, n_datasets = 0),
               class = "efa_invalid_argument")
  expect_error(efa_parallel(N = 20, n_vars = 4, n_factors = 0),
               class = "efa_invalid_argument")
})

test_that("the simulated PCA reference matches psych::fa.parallel", {
  skip_if_not_slow()
  skip_if_not_installed("psych")
  # Both draw their reference eigenvalues from independent random normal data, so
  # they agree only up to Monte-Carlo error; 2000 datasets bring that well inside
  # the tolerance (observed max absolute difference ~0.0013).
  R <- test_models$baseline$cormat

  set.seed(42)
  ours <- efa_parallel(R, N = 500, n_datasets = 2000, eigen_type = "PCA")

  # fa.parallel reports its suggestion with cat(), so capture that output
  invisible(utils::capture.output(
    theirs <- suppressWarnings(psych::fa.parallel(R, n.obs = 500, fa = "pc",
                                                  n.iter = 2000, plot = FALSE))
  ))

  expect_equal(.retention_record(ours, "PCA")$references$Means, theirs$pc.sim,
               tolerance = 0.01)
  expect_equal(ours$n_factors[["PCA"]], theirs$ncomp)
})

test_that(".parallel_chunks splits exactly into non-negative integer chunks", {
  # the documented degenerate case: 11 datasets across 7 workers must not yield a
  # negative chunk (the old round-and-backfill produced c(2, 2, 2, 2, 2, 2, -1))
  chunks_11_7 <- .parallel_chunks(11, 7)
  expect_length(chunks_11_7, 7)
  expect_equal(sum(chunks_11_7), 11)
  expect_true(all(chunks_11_7 >= 0))

  # a range of awkward (n_datasets, n_cores) pairs: always exact, never negative
  for (n_cores in 1:8) {
    for (n_datasets in c(1, 3, 11, 100, 1000)) {
      chunks <- .parallel_chunks(n_datasets, n_cores)
      expect_length(chunks, n_cores)
      expect_equal(sum(chunks), n_datasets)
      expect_true(all(chunks >= 0))
    }
  }

  # a single chunk takes all datasets
  expect_equal(.parallel_chunks(1000, 1), 1000)
})


test_that("efa_parallel does not size its chunks from the worker count", {
  # The simulated datasets are split into chunks, one future per chunk, and
  # future.seed = TRUE assigns one random-number stream per chunk. If the chunk count
  # tracked the worker count, the streams -- and hence the reference eigenvalues -- would
  # differ between parallel plans for the same set.seed().
  #
  # Assert that through efa_parallel() itself rather than by re-computing the chunk
  # formula here: a test that only restates the arithmetic would still pass if the call
  # site were reverted to nbrOfWorkers(). Faking the reported worker count exercises the
  # call site while keeping the run sequential, so no workers are started and the check
  # runs on CRAN -- unlike the end-to-end test below.
  skip_if_not_installed("testthat", "3.2.0")

  run_reporting <- function(n_workers) {
    testthat::local_mocked_bindings(nbrOfWorkers = function(...) n_workers,
                                    .package = "future")
    set.seed(2024)
    suppressWarnings(suppressMessages(
      efa_parallel(N = 250, n_vars = 6, n_datasets = 24, eigen_type = "PCA")))
  }

  expect_equal(run_reporting(1L)$results[[1]]$references,
               run_reporting(4L)$results[[1]]$references)
})


test_that(".parallel_chunks is exact at the sizes efa_parallel asks for", {
  # fewer datasets than the fixed chunk count: one chunk each, never an empty chunk
  expect_equal(.parallel_chunks(5, max(1L, min(5, 20L))), rep(1, 5))

  # the degenerate n_datasets = 0 must not divide by a zero chunk count
  expect_equal(.parallel_chunks(0, max(1L, min(0, 20L))), 0)
})


test_that("only the simulation chunk that failed is redrawn", {
  # Deterministic chunk failure: the mocked simulation fails one chunk evaluation and only
  # that one, so the redraw meets a working simulation again. Counting the evaluations shows
  # what a failure costs -- redrawing the whole batch would repeat every chunk, which for a
  # failure in the last of twenty chunks is nineteen sound chunks thrown away.
  skip_if_not_installed("testthat", "3.2.0")

  # The mock replaces the simulation in this process only, so the chunks have to resolve
  # here; an ambient parallel plan would run the real one in a worker and count nothing.
  old_plan <- future::plan(future::sequential)
  on.exit(future::plan(old_plan), add = TRUE)

  real <- .parallel_sim_eig
  n_chunks <- length(.parallel_chunks(40, max(1L, min(40, 20L))))

  count <- function(fail_on) {
    calls <- 0L
    local_mocked_bindings(
      .parallel_sim_eig = function(n_datasets, ...) {
        calls <<- calls + 1L
        if (isTRUE(calls == fail_on)) {
          cli::cli_abort("simulated chunk failure", class = "efa_test_chunk_failure")
        }
        real(n_datasets, ...)
      }
    )
    set.seed(11)
    suppressWarnings(suppressMessages(
      efa_parallel(N = 200, n_vars = 5, n_datasets = 40, eigen_type = "PCA")))
    calls
  }

  expect_identical(count(NA_integer_), n_chunks)
  expect_identical(count(3L), n_chunks + 1L)
  expect_identical(count(n_chunks), n_chunks + 1L)
})


test_that("a redrawn chunk leaves the other chunks' draws untouched", {
  # The retry runs on the failing chunk alone, so the chunks that succeeded keep the exact
  # draws of the first attempt; only the redrawn one gets a fresh random-number stream.
  skip_if_not_installed("testthat", "3.2.0")

  old_plan <- future::plan(future::sequential)
  on.exit(future::plan(old_plan), add = TRUE)

  real <- .parallel_sim_eig
  size_vec <- .parallel_chunks(40, max(1L, min(40, 20L)))
  failed_rows <- (sum(size_vec[1:2]) + 1L):sum(size_vec[1:3])

  run <- function(fail_on) {
    calls <- 0L
    local_mocked_bindings(
      .parallel_sim_eig = function(n_datasets, ...) {
        calls <<- calls + 1L
        if (isTRUE(calls == fail_on)) {
          cli::cli_abort("simulated chunk failure", class = "efa_test_chunk_failure")
        }
        real(n_datasets, ...)
      }
    )
    set.seed(11)
    .parallel_sim_chunks(size_vec, label = "PCA", N = 200, n_vars = 5,
                         eigen_type = 1, cor_method = "pearson")
  }

  clean <- run(NA_integer_)
  broken <- run(3L)

  expect_identical(broken[-failed_rows, ], clean[-failed_rows, ])
  expect_false(identical(broken[failed_rows, ], clean[failed_rows, ]))
})


test_that("the parallel-analysis abort carries the failure that caused it", {
  skip_if_not_installed("testthat", "3.2.0")

  old_plan <- future::plan(future::sequential)
  on.exit(future::plan(old_plan), add = TRUE)

  local_mocked_bindings(
    .parallel_sim_eig = function(n_datasets, ...) {
      cli::cli_abort("no usable draw", class = "efa_test_sim_failure")
    }
  )

  cnd <- expect_error(
    .parallel_sim_chunks(.parallel_chunks(4, 2), label = "PCA", N = 200, n_vars = 5,
                         eigen_type = 1, cor_method = "pearson", max_tries = 3L),
    class = "efa_parallel_sim_failed")
  expect_s3_class(cnd$parent, "efa_test_sim_failure")
})


test_that("a failing parallel backend is retried and reported the same way", {
  # A chunk reports its own failure, so what the batch-level retry covers is the backend
  # itself -- a lost worker, say. It must reach the same classed abort, not leak out raw.
  skip_if_not_installed("testthat", "3.2.0")

  local_mocked_bindings(
    future_lapply = function(...) {
      cli::cli_abort("the workers are gone", class = "efa_test_backend_failure")
    },
    .package = "future.apply")

  cnd <- expect_error(
    .parallel_sim_chunks(.parallel_chunks(4, 2), label = "SMCs", N = 200, n_vars = 5,
                         eigen_type = 2, cor_method = "pearson", max_tries = 2L),
    class = "efa_parallel_sim_failed")
  expect_s3_class(cnd$parent, "efa_test_backend_failure")
})


test_that("a seeded efa_parallel run is invariant to the number of workers", {
  skip_on_cran()
  skip_if_not_slow()
  # End-to-end counterpart to the chunk-vector test above: the same set.seed() must give
  # the same reference eigenvalues sequentially and on a two-worker multisession plan.
  # The multisession workers are fresh R processes that load the *installed* package, so
  # run this under devtools::check() / after devtools::install() for the worker code to
  # match the main process (multicore is unavailable on Windows).
  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  run <- function() {
    set.seed(2024)
    suppressWarnings(suppressMessages(
      efa_parallel(N = 300, n_vars = 8, n_datasets = 60, eigen_type = "PCA")))
  }

  future::plan(future::sequential)
  one <- run()

  future::plan(future::multisession, workers = 2)
  two <- run()

  # `results[[1]]$references` holds the simulated reference eigenvalues (means and
  # percentiles), which is what the chunk streams determine. Bit-for-bit equality is not
  # asserted: the workers are separate processes whose BLAS/LAPACK may sum in a different
  # order.
  expect_equal(one$results[[1]]$references, two$results[[1]]$references,
               tolerance = 1e-10)
  expect_equal(one$n_factors, two$n_factors)
})


test_that(".parallel_summarise uses stats::quantile for the percentile series", {
  set.seed(42)
  eig_vals <- matrix(stats::rnorm(1000 * 4), nrow = 1000, ncol = 4)

  res <- .parallel_summarise(eig_vals, percent = 95, n_vars = 4)
  expect_equal(dim(res), c(4, 2))
  expect_equal(res[, 1], colMeans(eig_vals))
  for (root in 1:4) {
    expect_equal(res[root, 2],
                 stats::quantile(eig_vals[, root], probs = 0.95, names = FALSE))
  }

  # multiple percentiles produce one column each
  res2 <- .parallel_summarise(eig_vals, percent = c(50, 95), n_vars = 4)
  expect_equal(dim(res2), c(4, 3))
  for (root in 1:4) {
    expect_equal(res2[root, 2:3],
                 stats::quantile(eig_vals[, root], probs = c(0.5, 0.95),
                                 names = FALSE))
  }

  # a NA/NaN simulated eigenvalue is tolerated, not turned into a hard error
  eig_na <- eig_vals
  eig_na[1, 1] <- NA
  expect_no_error(res_na <- .parallel_summarise(eig_na, percent = 95, n_vars = 4))
  expect_equal(res_na[1, 2],
               stats::quantile(eig_na[, 1], probs = 0.95, names = FALSE,
                               na.rm = TRUE))
})

test_that(".determine_factors retains all components and warns when no crossing", {
  # every real eigenvalue exceeds its reference: the rule finds no crossing, so
  # instead of a silent NA it retains all tested components and flags the boundary
  eigvals_real <- matrix(c(3, 2, 1.5), ncol = 1)
  results <- cbind(c(2, 1, 0.5), c(2.5, 1.2, 0.6))
  colnames(results) <- c("Means", "95 Percentile")

  for (rule in c("means", "percentile", "crawford")) {
    expect_warning(
      n_fac <- .determine_factors(rule, eigvals_real, results, percent = 95),
      class = "efa_parallel_no_crossing"
    )
    expect_equal(n_fac, nrow(eigvals_real), info = rule)
  }
})

test_that(".determine_factors returns a normal crossing and 0 without the warning", {
  results <- cbind(c(1, 1, 1), c(1.2, 1.2, 1.2))
  colnames(results) <- c("Means", "95 Percentile")

  # first eigenvalue exceeds the reference, the second does not -> retain 1
  expect_no_warning(
    n1 <- .determine_factors("means", matrix(c(3, 0.5, 0.2), ncol = 1),
                             results, percent = 95),
    class = "efa_parallel_no_crossing"
  )
  expect_equal(n1, 1)

  # the first eigenvalue is already at/below the reference -> retain 0
  expect_no_warning(
    n0 <- .determine_factors("means", matrix(c(0.5, 0.4, 0.3), ncol = 1),
                             results, percent = 95),
    class = "efa_parallel_no_crossing"
  )
  expect_equal(n0, 0)
})

test_that("the null-model reference draw matches the compiled fast path", {
  # The rank-method and EFA paths draw the null model with the shared MVN kernel on the
  # identity correlation, while the Pearson PCA/SMC path uses the compiled .parallel_sim().
  # This pins that the two draw the same underlying stream: the kernel's eigenvalues agree
  # with the compiled fast path up to the compiled cor() vs stats::cor() rounding (~1e-15)
  # under a fixed seed. (That PARALLEL's rank/EFA paths actually route through the kernel is
  # guarded by the two tests below, which drive .parallel_sim_eig() and .parallel_EFA_sim().)
  N <- 60; p <- 5; nd <- 3

  set.seed(101)
  e_fast <- .parallel_sim(nd, p, N, 1L)          # compiled PCA fast path

  set.seed(101)
  e_kernel <- t(vapply(seq_len(nd), function(i) {
    eigen(stats::cor(.simulate_cfm_mvn(diag(p), N)), symmetric = TRUE,
          only.values = TRUE)$values
  }, numeric(p)))

  expect_equal(e_kernel, e_fast, tolerance = 1e-10)
})

test_that(".parallel_sim_eig draws its rank-method reference from the shared kernel", {
  # The rank-method PCA path never rejects a draw, so its reference eigenvalues must match a
  # direct per-draw recompute through the shared kernel under the same random-number stream
  # -- a guard that the draw routes through .simulate_cfm_mvn().
  N <- 60; p <- 5; nd <- 4

  set.seed(202)
  got <- .parallel_sim_eig(nd, n_vars = p, N = N, eigen_type = 1L,
                           cor_method = "spearman")

  set.seed(202)
  want <- t(vapply(seq_len(nd), function(i) {
    eigen(stats::cor(.simulate_cfm_mvn(diag(p), N), method = "spearman"),
          symmetric = TRUE, only.values = TRUE)$values
  }, numeric(p)))

  expect_equal(got, want, tolerance = fp_tol)
})

test_that("the PCA and SMC series are taken from one set of simulated datasets", {
  # PCA and SMC differ only in the diagonal put into the same simulated correlation matrix,
  # so asking for both must simulate n_datasets datasets, not 2 * n_datasets. Counted at the
  # kernel rather than by restating the branch, so a call site that reverted to two
  # independent simulations would fail here.
  skip_if_not_installed("testthat", "3.2.0")

  # The mock counts draws in this process, so the chunks have to resolve here.
  old_plan <- future::plan(future::sequential)
  on.exit(future::plan(old_plan), add = TRUE)

  real <- .parallel_sim
  n_datasets <- 20
  n_chunks <- length(.parallel_chunks(n_datasets, max(1L, min(n_datasets, 20L))))

  traced <- function(eigen_type) {
    drawn <- 0L
    types <- numeric(0)
    local_mocked_bindings(
      .parallel_sim = function(n_datasets, n_vars, N, eigen_type, maxit) {
        drawn <<- drawn + n_datasets
        types <<- c(types, eigen_type)
        real(n_datasets, n_vars, N, eigen_type, maxit)
      })
    set.seed(3)
    suppressWarnings(suppressMessages(
      efa_parallel(N = 200, n_vars = 5, n_datasets = n_datasets,
                   eigen_type = eigen_type)))
    list(drawn = drawn, types = types)
  }

  pca <- traced("PCA")
  expect_identical(pca$drawn, n_datasets)
  expect_equal(pca$types, rep(1, n_chunks))

  smc <- traced("SMC")
  expect_identical(smc$drawn, n_datasets)
  expect_equal(smc$types, rep(2, n_chunks))

  both <- traced(c("PCA", "SMC"))
  expect_identical(both$drawn, n_datasets)
  expect_equal(both$types, rep(3, n_chunks))

  # the default set adds "EFA", which fits a model per dataset and draws its own
  all_three <- traced(c("PCA", "SMC", "EFA"))
  expect_identical(all_three$drawn, n_datasets)
  expect_equal(all_three$types, rep(3, n_chunks))
})

test_that("the shared draw feeds both series with the eigenvalues each would get alone", {
  # The shared kernel run must return exactly the two series the single-type runs produce
  # from the same random-number stream: the PCA eigenvalues in the first n_vars columns and
  # the SMC eigenvalues in the next, replicate by replicate.
  N <- 200L; p <- 6L; nd <- 25L

  set.seed(4)
  shared <- .parallel_sim(nd, p, N, 3L, nd * 10L)
  set.seed(4)
  pca_only <- .parallel_sim(nd, p, N, 1L, nd * 10L)
  set.seed(4)
  smc_only <- .parallel_sim(nd, p, N, 2L, nd * 10L)

  expect_identical(dim(shared), c(nd, 2L * p))
  expect_identical(shared[, seq_len(p)], pca_only)
  expect_identical(shared[, p + seq_len(p)], smc_only)
})

test_that("a draw the SMC series cannot use is discarded for the PCA series too", {
  # A simulated correlation matrix with no inverse yields no squared multiple correlations,
  # and the draw is then dropped from both series so that they stay paired. Admissible input
  # essentially never produces such a matrix, so the rejection is forced here, on the
  # rank-based route where it is taken in R.
  skip_if_not_installed("testthat", "3.2.0")

  N <- 60; p <- 5
  real <- .smc_start

  set.seed(606)
  clean <- .parallel_sim_eig(4, n_vars = p, N = N, eigen_type = 3L,
                             cor_method = "spearman")

  calls <- 0L
  local_mocked_bindings(
    .smc_start = function(R, ...) {
      calls <<- calls + 1L
      if (calls == 2L) stop("singular simulated matrix")
      real(R, ...)
    })
  set.seed(606)
  rejected <- .parallel_sim_eig(3, n_vars = p, N = N, eigen_type = 3L,
                                cor_method = "spearman")

  # the second draw is gone from both halves, and the third took its place in both
  expect_identical(rejected[1L, ], clean[1L, ])
  expect_identical(rejected[2L, ], clean[3L, ])
  expect_identical(rejected[3L, ], clean[4L, ])
})

test_that(".parallel_EFA_sim draws its reference from the shared kernel", {
  # The EFA path likewise draws the null-model data with the shared kernel; under a
  # fixed seed its eigenvalues equal a direct recompute that draws the same way.
  N <- 60; p <- 5; nd <- 2

  set.seed(303)
  got <- .parallel_EFA_sim(nd, n_vars = p, N = N, n_factors = 1,
                           cor_method = "pearson")

  set.seed(303)
  want <- t(vapply(seq_len(nd), function(i) {
    R <- stats::cor(.simulate_cfm_mvn(diag(p), N))
    suppressWarnings(suppressMessages(EFA(R, n_factors = 1, N = N)$final_eigen))
  }, numeric(p)))

  expect_equal(got, want)
})

if (is_slow_test()) {
  rm(pa_cor, pa_cor_pca, pa_raw, pa_nodat, pa_craw, pa_perc)
}
rm(burt)

#' Various outputs from SPSS (version 27) FACTOR
#'
#' Various outputs from SPSS (version 27) FACTOR for the IDS-2 (Grob & Hagmann-von Arx, 2018), the WJIV (3 to 5 and 20 to 39 years; McGrew, LaForte, & Schrank, 2014), the DOSPERT (Frey et al., 2017; Weber,
#' Blais, & Betz, 2002), and four simulated datasets (baseline, case_1a, case_6b, and case_11b, see [test_models] and [population_models]) used in Grieder and Steiner (2022).
#'
#'
#' @format A list of 8 containing EFA results for each of the data sets mentioned above. Each of these eight entries is a list of 4, of the following structure:
#' \describe{
#'   \item{paf_load}{(matrix) - F1 to FN = unrotated factor loadings obtained with the FACTOR algorithm with PAF. Rownames are the abbreviated subtest or item names, and generic variable labels (V1 to VN) for the simulated datasets.}
#'   \item{var_load}{(matrix) - F1 to FN = varimax rotated factor loadings obtained with the FACTOR algorithm with PAF. Rownames are the abbreviated subtest or item names, and generic variable labels (V1 to VN) for the simulated datasets.}
#'   \item{pro_load}{(matrix) - F1 to FN = promax rotated factor loadings obtained with the FACTOR algorithm with PAF. Rownames are the abbreviated subtest or item names, and generic variable labels (V1 to VN) for the simulated datasets.}
#'   \item{pro_phi}{(matrix) - F1 to FN = intercorrelations of the promax rotated loadings.}
#'  }
#'
#' @details The principal axis factoring was run with the iteration limit raised above
#'   SPSS's own default of 25, so reproducing these solutions requires the same: `case_1a`
#'   needs 60 iterations and `case_11b` needs 33, and at `max_iter = 25` both stop short of
#'   convergence and differ from the stored loadings in the second decimal. Use
#'   `estimate_control(type = "SPSS", max_iter = 500)` when checking a preset against these
#'   references; the other two simulated cases converge in six iterations and are
#'   unaffected.
#'
#' @source Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
#' of implementations of principal axis factoring and promax rotation in R and SPSS.
#' Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x
#' @source Grieder, S., & Grob, A. (2019). Exploratory factor analyses of the intelligence and development scales–2: Implications for theory and practice. Assessment. Advance online publication. doi:10.1177/1073191119845051
#' @source Grob, A., & Hagmann-von Arx, P. (2018). Intelligence and Development Scales--2 (IDS-2). Intelligenz- und Entwicklungsskalen für Kinder und Jugendliche.
#' \[Intelligence and Development Scales for Children and Adolescents.\]. Bern, Switzerland: Hogrefe.
#' @source Frey, R., Pedroni, A., Mata, R., Rieskamp, J., & Hertwig, R. (2017). Risk preference shares the psychometric structure of major psychological traits. Science Advances, 3, e1701381.
#' @source McGrew, K. S., LaForte, E. M., & Schrank, F. A. (2014). Technical
#'  Manual. Woodcock-Johnson IV. Rolling Meadows, IL: Riverside.
#' @source Schrank, F. A., McGrew, K. S., & Mather, N. (2014). Woodcock-Johnson IV.
#' Rolling Meadows, IL: Riverside.
"SPSS_27"

#' Exploratory factor analysis on multiple data imputations
#'
#' @author Andreas Soteriades, Markus Steiner
#'
#' @description
#' Fits [efa_fit()] to each of several imputed datasets, aligns the
#' factor solutions to a common factor space, and pools the resulting estimates
#' and selected fit quantities across imputations.
#'
#' @details
#' `efa_mi()` is the multiple-imputation route to handling missing data:
#' several imputed datasets are each fitted with [efa_fit()] and the solutions pooled.
#' A single-fit alternative is full-information maximum likelihood, available
#' directly in [efa_fit()] as `cor_method = "fiml"`, which EM-estimates a two-stage
#' correlation from one raw dataset with missing values. Both feed the same
#' correlation-scale EFA core and differ only in how the missingness is handled;
#' FIML is intentionally not routed through `efa_mi()`, which is a multi-fit
#' pooler by construction.
#'
#' Both routes assume the values are missing at random (MAR). Which one to prefer
#' is largely practical: FIML is a single, efficient fit and is the simpler
#' default when the analysis model is the whole story, whereas multiple imputation
#' is more flexible when the imputation model should draw on auxiliary variables
#' not in the factor model, or when the same imputations feed several downstream
#' analyses.
#'
#' ## Standard-error pooling routes
#'
#' The pooling pathway is selected automatically from the `se` method recorded on
#' the component [efa_fit()] fits, which must be identical across imputations:
#'
#' - `se = "none"`: no standard errors are pooled.
#' - `se = "information"`: the per-imputation expected-information standard errors
#'   are pooled with Rubin's (1987) rules (Wald intervals).
#' - `se = "sandwich"`: the two-stage pooled-inputs (MI2S) approach fits a single
#'   model on the Rubin-pooled correlation matrix and asymptotic covariance.
#' - `se = "np-boot"`: the non-parametric bootstrap replicates are re-aligned to
#'   the multiple-imputation target and Rubin-pooled.
#'
#' On the information and np-boot routes, if pooled standard errors cannot be
#' produced (for example an unreliable analytic covariance or too few bootstrap
#' replicates) the pool falls back to point-estimate-only pooling and downgrades
#' `settings$se` to `"none"`. The MI2S route is the exception: its single fit
#' fuses the point estimates and standard errors through the pooled asymptotic
#' covariance, so a structural failure aborts directly rather than falling back.
#'
#' ## Aligning solutions across imputations
#'
#' The same [efa_fit()] model is fitted to each imputed dataset and the solutions are
#' put into a common factor space before averaging. For oblique solutions the
#' factor intercorrelations are aligned together with the loadings so the model
#' stays internally consistent.
#'
#' `target_method` controls how rotated solutions are aligned. `"first_target"`
#' (the default) aligns every imputation to the first imputation's rotated
#' solution by one Procrustes rotation each. `"consensus"` instead refines a
#' centroid target by Generalized Procrustes Analysis (Gower 1975; van Ginkel &
#' Kroonenberg 2014; Lorenzo-Seva & Van Ginkel 2016), starting from the *medoid*
#' imputation's rotated solution -- the one closest in aligned squared distance to
#' all the others. `"consensus"` is supported for orthogonal rotations only.
#'
#' Factor loadings are only unique up to rotation; "gauge" here means which
#' particular rotation, or orientation, a solution is expressed in. The two
#' methods differ in the rotational gauge the pooled solution ends up in, and so
#' in how it responds to the order of `data_list`. The GPA iteration moves its
#' target toward the centroid but keeps the gauge of the solution it started from,
#' so starting it at the medoid -- a property of the set, not of the list order --
#' makes the pooled rotated solution invariant to that order, as the pooled
#' unrotated solution already is. `"first_target"` anchors on the first imputation
#' by construction, so an atypical first imputation fixes the orientation for
#' every other one. Permuting `data_list` therefore moves the pooled pattern --
#' by a few hundredths of a loading unit when imputations are similar, more when
#' they disagree. For oblique rotations, the factor correlations move with it.
#' Where the two anchors coincide -- and, more generally,
#' where the imputations agree -- the two methods give effectively the same pooled
#' estimate and `"consensus"` is simply the more expensive; where they do not, the
#' pooled patterns differ by the rotation between the two gauges. Passing `start`
#' through `consensus_args` overrides the medoid anchor and makes the consensus
#' order dependent again.
#'
#' `align_unrotated` controls how unrotated loadings are aligned before pooling:
#' `"signed_tucker_congruence"` (the default) matches them up to factor reordering
#' and sign changes, `"procrustes"` aligns them to the first imputation by
#' orthogonal Procrustes rotation, and `"none"` averages them as returned by
#' [efa_fit()].
#'
#' The default anchors the matching on the *medoid* imputation (defined above)
#' rather than on whichever imputation happens to come first, so the pooled
#' *unrotated* solution does not depend on the order of `data_list`. The rotated
#' solution is aligned separately, against a reference chosen by `target_method`,
#' and still depends on that reference.
#'
#' The pooled unrotated matrix is then returned in the same gauge as a single
#' [efa_fit()] fit. Its identifying constraint differs by extraction method -- a
#' principal-axis extraction and a maximum-likelihood extraction fix the
#' rotation differently (Anderson & Rubin 1956; Lawley & Maxwell 1971) -- and is
#' read off each component fit automatically, so the pooled matrix can be
#' compared element-by-element with an [efa_fit()] solution. A solution that
#' meets neither constraint -- an improper one, say -- is left as aligned. The
#' correction is a common orthogonal rotation, so communalities, the total
#' variance accounted for, the model-implied correlation matrix, the residuals,
#' and RMSR are unchanged; only the split of variance across factors moves.
#' `"procrustes"` and `"none"` keep their first-imputation anchor and are
#' returned as aligned.
#'
#' ## Pooling point estimates
#'
#' Point estimates are pooled by arithmetic averaging after alignment. For oblique
#' rotations the structure matrix is recomputed from the pooled pattern matrix and
#' pooled factor correlations, \eqn{Structure = \Lambda \Phi}, and communalities
#' are the diagonal of the reproduced correlation matrix,
#' \eqn{diag(\Lambda \Phi \Lambda')} for oblique rotations and
#' \eqn{diag(\Lambda \Lambda')} otherwise. Residuals are not averaged across
#' imputations; they are the pooled observed correlation matrix minus the
#' model-implied correlation of the pooled solution, so RMSR/SRMR are based on
#' these pooled residuals. Both are returned, though the print and summary methods
#' show SRMR only.
#'
#' ## Pooling the model chi-square and fit indices
#'
#' The model chi-square and the indices derived from it (ECVI and the descriptive
#' AIC/BIC) are pooled with the D2 rule (Li, Meng, Raghunathan &
#' Rubin, 1991), not arithmetically averaged. RMSEA is pooled by the same rule but
#' from a second D2 pool of the per-imputation discrepancies taken on the
#' uncorrected \eqn{N - 1} scale. The printed RMSEA therefore does not reconcile
#' by hand with the printed chi-square; the statistic it is formed from is
#' `chi_cfi` in `mi_diagnostics`. Because D2 shrinks the pooled chi-square in
#' proportion to the between-imputation variability, the pooled RMSEA can fall
#' below the mean of the per-imputation RMSEAs; read it together with the
#' per-imputation fit. The incremental indices CFI (Bentler, 1990) and TLI
#' (Tucker & Lewis, 1973) are instead the average of the per-imputation indices,
#' which keeps them consistent with the component fits and avoids out-of-range
#' values; the separately pooled model and baseline chi-squares those indices
#' would be formed from remain available in `mi_diagnostics`.
#' AIC and BIC, if returned, are chi-square-derived descriptive quantities and are
#' not likelihood-based MI information criteria. They are reported only where the
#' component fits report them: whenever a component withholds them -- any
#' `cor_method = "fiml"` fit, and any fit whose chi-square is a scaled statistic,
#' such as `se = "sandwich"` -- the pooled AIC, BIC, and ECVI are `NA` too, matching
#' what [efa_fit()] returns for a single such fit. On the sandwich/MI2S route the
#' chi-square is the single fit's scaled statistic rather than a D2 pool.
#'
#' ## Bootstrap pooling (np-boot)
#'
#' If each component [efa_fit()] call was run with `se = "np-boot"`, pooled bootstrap SEs
#' and Wald-type MI confidence intervals are computed for loadings, communalities,
#' residuals, and, when applicable, factor correlations and structure
#' coefficients. The unrotated bootstrap replicates are re-aligned to the final MI
#' target before the within-imputation covariance is estimated, and Rubin pooling
#' is applied with \eqn{T = Ubar + (1 + 1/m) B}. The confidence level of the pooled
#' intervals is set by `p`, not by the component [efa_fit()] calls' `ci`.
#'
#' ## Analytic pooling (information)
#'
#' With `se = "information"`, the analytic unrotated-loading and uniqueness SEs
#' returned by each fit are pooled element-wise with Rubin's rules
#' (\eqn{T = Ubar + (1 + 1/m) B}), with Wald intervals on the plain Rubin (1987)
#' degrees of freedom (the analytic loadings are asymptotically normal, so the
#' Barnard-Rubin (1999) adjustment reduces to this form). NA
#' propagation is fail-closed: if any imputation is NA at an element, all pooled
#' outputs for that element are NA. When a rotation was requested, the rotated
#' loadings, communalities, and (for oblique rotations) factor correlations and
#' structure coefficients are pooled as well; residual SE pooling is available only
#' on the bootstrap path. Under `align_unrotated = "procrustes"` the full unrotated
#' covariance `vcov_unrot_loadings` (populated by `se = "information"`) is
#' propagated through the alignment, so it must be present and reliable on every
#' fit. The default alignment also mixes loading columns, through the common
#' canonical-gauge rotation, and so propagates the same covariance; where a fit
#' does not carry it, the unrotated standard errors are returned as `NA` rather
#' than aborting, and the remaining families still pool.
#'
#' A rotated-loading standard error is conditional on the rotation criterion
#' used (see References). For both orthogonal and oblique rotations the
#' within-imputation variance is therefore each fit's own
#' criterion-aware delta-method rotated SE (the quantity `efa_fit()` returns), reused
#' after a signed-permutation alignment to the MI target, and the
#' between-imputation variance is the sample variance of the aligned rotated
#' loadings. This is a deliberate approximation -- each SE is conditional on its
#' own fit's rotation optimum rather than on a common gauge -- and is flagged by
#' `MI$<param>$method = "signed_permutation_approx"`. Communalities are
#' rotation-invariant and pool element-wise. For a fully gauge-consistent rotated
#' uncertainty, cross-check with `se = "np-boot"`.
#'
#' ## Two-stage pooling (sandwich / MI2S)
#'
#' With `se = "sandwich"` (robust SEs from a polychoric/tetrachoric or
#' continuous-Pearson asymptotic covariance), pooling follows the two-stage,
#' pooled-inputs approach (Chung & Cai 2019; Sriutaisuk, Liu, Chung, Kim & Gu
#' 2025): the correlation matrix and the asymptotic covariance of its off-diagonal
#' entries are Rubin-pooled across imputations, and a single `EFA` model is
#' fitted to the pooled correlation with the pooled covariance, \eqn{\tilde\Gamma},
#' as the robust meat (its diagonal as the weights for
#' `estimator = "DWLS"`). Because there is only one fit and one rotational gauge, this
#' route bypasses the per-imputation alignment: `target_method` and
#' `align_unrotated` do not apply. The fitted object carries native scaled-shifted
#' chi-square statistics and sandwich SEs that already reflect the
#' multiple-imputation uncertainty, so the chi-square is not D2-pooled and the
#' likelihood-ratio-based AIC/BIC/ECVI are `NA`; it is returned in the `mi_fit`
#' slot, with the per-imputation `fits` retained for diagnostics. The pooled fit
#' uses the same [estimate_control()] and [rotate_control()] tuning (including
#' any rotation-engine extras) as the per-imputation fits. At least 20
#' imputations are recommended for the scaled-shifted statistic, and more (around
#' 100) at higher rates of missingness (Sriutaisuk et al. 2025). The
#' polychoric/tetrachoric (ordinal) case is the primary, best-evaluated target; the
#' continuous-Pearson case uses the same recipe but is less benchmarked.
#'
#' @section Conditions:
#' Errors and warnings raised by `efa_mi()` are classed, with an `efa_pooled_`
#' prefix (`efa_consensus_` for the consensus target) -- except the dots validation
#' shared with [efa_fit()], which signals `efa_flat_knob_in_dots` or `efa_renamed_arg`
#' -- so they can be caught programmatically. The message shown explains what went
#' wrong and, where relevant, how to fix it.
#'
#' @param data_list A list of length \eqn{m}, where \eqn{m} is the number of
#' imputations. Each list element is a data frame or matrix of raw data, or a
#' correlation matrix. See argument `x` in [efa_fit()]. A `mids` object from
#' \pkg{mice} must be converted first, with `mice::complete(x, "all")`.
#' @param p Numeric in \eqn{(0, 1)}. One minus the confidence level for the
#' pooled confidence intervals, whichever `se` method produced them
#' (`"information"`, `"np-boot"`, or `"sandwich"`). For example, `p = .05`
#' gives 95% intervals.
#' @param target_method Character. How rotated solutions are aligned across imputations
#' before pooling: `"first_target"` (the default) aligns every imputation to the first
#' imputation's rotated solution, while `"consensus"` refines a centroid target by
#' Generalized Procrustes Analysis, started from the medoid imputation so that the
#' pooled rotated solution does not depend on the order of `data_list` (orthogonal
#' rotations only). See *Aligning solutions across imputations* in Details.
#' @param align_unrotated Character. How unrotated loadings are aligned before pooling:
#' `"signed_tucker_congruence"` (the default; sign/permutation via Tucker congruence,
#' anchored on the medoid imputation and returned in the extraction's canonical
#' gauge), `"procrustes"` (orthogonal Procrustes to the first imputation), or
#' `"none"`. See *Aligning solutions across imputations* in Details.
#' @param fit_pool_method Character. Only `"D2"` is implemented for pooling
#' chi-square-type fit. If no chi-square is available, only residual-based
#' fit and descriptive quantities are returned. See *Pooling the model chi-square and
#' fit indices* in Details.
#' @param consensus_args List of additional arguments controlling the
#' GPA-consensus iteration when `target_method = "consensus"`. Recognised tuning
#' parameters include the convergence tolerances `tol` and `loss_tol`, the
#' iteration bounds `min_iter` and `max_iter`, the target-update damping `alpha`,
#' the multi-start controls `multi_start` and `starts`, and `start`, which
#' overrides the medoid imputation the iteration is otherwise started from.
#' @param procrustes_args List of [efa_procrustes()] algorithm controls for
#' fixed-target alignment, for example `oblique_maxit` or `oblique_random_starts`.
#' The loadings `A`, the alignment `Target`, the `rotation` family, and the
#' cross-product `S` are derived from the imputations and cannot be set here.
#' @param rmsea_ci_level Numeric. Confidence level for the RMSEA CI.
#' @param rmsr_upper `r lifecycle::badge("deprecated")` Deprecated and ignored.
#' `efa_mi()` now always computes RMSR the same way, from the unique
#' off-diagonal residuals; SRMR is reported alongside it. Supplying it to
#' [efa_mi()] signals a deprecation warning; the superseded [EFA_POOLED()]
#' accepts it silently.
#' @param ... Additional arguments passed to [efa_fit()] (e.g. `estimator`, `rotation`, `se`,
#' `n_factors`, `N`). These select the estimator, rotation, standard-error method, and
#' fit indices used for every imputation; see [efa_fit()] for the available options, their
#' properties, and which combinations are valid. Two of them shape the pooled object
#' rather than a single fit: `seed` sets the random state once for the whole
#' `efa_mi()` call -- every component bootstrap and every random-start rotation draws
#' from it, so a seeded call is reproducible as a whole, and the caller's random stream
#' is restored afterwards -- and `b_boot` sets the number of bootstrap replicates drawn
#' per imputation under `se = "np-boot"`, which is what the pooled within-imputation
#' variances are estimated from and is recorded in `settings$b_boot`. The
#' [estimate_control()] and
#' [rotate_control()] objects are accepted through `...` as well, although they are not
#' declared formals: pass them as `estimate_control =` / `rotate_control =` exactly as you
#' would to [efa_fit()].
#'
#' @return A list of class `c("efa_mi", "EFA_POOLED", "efa", "EFA")` containing
#' pooled estimates, residuals, fit indices, the individual fits, and MI
#' diagnostics. The trailing legacy classes keep `inherits(x, "EFA_POOLED")` and
#' the single-fit EFA accessors and S3 dispatch working. In
#' addition to the slots inherited from [efa_fit()] (including `SE`, `CI`, and,
#' on the bootstrap path, `replicates`), the object carries:
#' \describe{
#' \item{SE, CI}{Pooled standard errors and confidence intervals, named as
#' [efa_fit()] names them: where a pooled communality standard error and interval
#' are produced, they are `SE$communalities` and `CI$communalities` on every route.
#' The Rubin routes (`se = "information"`, `se = "np-boot"`) additionally return
#' them under the compatibility alias `h2`, which holds the same values. The
#' analytic route builds the communality family only when a rotation was requested;
#' an unrotated analytic pool reports `uniquenesses` instead.}
#' \item{fit_indices}{The pooled fit indices. Every route reports `chi`, `df`,
#' `p_chi`, `CAF`, `CFI`, `TLI`, `RMSEA`, `RMSEA_LB`, `RMSEA_UB`, `AIC`, `BIC`,
#' `ECVI`, `RMSR`, `SRMR`, `chi_null`, `df_null`, `p_null`, and `pool_method`
#' under those names and in that order. `pool_method` records the rule the model
#' chi-square was pooled with (`"D2"`); it is `NA` on the `se = "sandwich"` (MI2S)
#' path, which fits once on the pooled inputs and reports that fit's own scaled
#' statistic rather than pooling several. On that path a few extra scaled-statistic
#' fields are appended after the common block, for advanced diagnostics.}
#' \item{standardized_residuals}{The pooled residuals divided by their pooled
#' bootstrap standard errors, with a zero diagonal. Returned on the
#' `se = "np-boot"` path only, the one route that pools a residual standard error.}
#' \item{MI}{Multiple-imputation diagnostics for each pooled parameter family.
#' On the bootstrap path: `unrot_loadings`, `communalities`, `residuals`, optionally
#' `rot_loadings`, `Phi`, `Structure`, and `fit_indices_descriptive`, plus
#' integer vectors `bootstrap_source_failures` (replicates the component [efa_fit()]
#' could not fit), `bootstrap_rotation_failures` (replicates whose Procrustes
#' alignment to the target was invalid), and `bootstrap_rotation_valid` (those
#' that entered the pool, `B - source - rotation` failures). Both paths use the
#' plain Rubin (1987) df. On the analytic path (`se = "information"`):
#' `unrot_loadings` and `uniquenesses`, plus, when a rotation was requested,
#' `rot_loadings`, `communalities`, and (oblique) `Phi` and `Structure`. The
#' communality family is keyed by its canonical name here, without the `SE`/`CI`
#' alias, so each family is counted once in the printed FMI/RIV summary. Each per-family
#' entry is a list with `RIV` (relative increase in variance), `FMI` (the
#' fraction of missing information, reported as Rubin's asymptotic
#' \eqn{\lambda = RIV / (1 + RIV)}), and `df`; the
#' rotated families on the analytic path additionally carry a `method` string
#' recording the gauge alignment used (`"gauge_invariant"` for communalities and
#' `"signed_permutation_approx"` for rotated loadings and, for oblique rotations,
#' factor correlations and structure coefficients). `fit_indices_descriptive`, on
#' the bootstrap path, pools every fit index the bootstrap replicates carry, so the
#' structural constants among them (`df`, `df_null`) appear with a standard error
#' of 0. The RMSEA confidence bounds are not among them: the replicate fits run
#' without confidence intervals, so no per-replicate value exists to pool.}
#' \item{mi_fit}{On the `se = "sandwich"` (MI2S) path only: the single [efa_fit()]
#' fit on the pooled correlation matrix \eqn{\bar r} and pooled asymptotic
#' covariance \eqn{\tilde\Gamma}. Its `orig_R` is \eqn{\bar r} and its `Gamma`
#' is \eqn{\tilde\Gamma}; the pooled `SE` and `CI` are taken from it, as are the
#' pooled `fit_indices` (put into the common order above and extended with
#' `pool_method`, while `mi_fit` keeps [efa_fit()]'s own layout). `MI` is `NULL`
#' on this path because the imputation uncertainty is carried by
#' \eqn{\tilde\Gamma} rather than by per-parameter Rubin pooling.}
#' \item{mi_diagnostics}{Diagnostics for the pooled model fit, `NULL` on the
#' `se = "sandwich"` (MI2S) path, where there is one fit and no D2 pool. `m` is the
#' number of imputations that entered the pool. `D2_F`, `D2_df1`, `D2_df2`,
#' `D2_chi_asymptotic`, `ARIV` and `FMI` describe the D2 pool of the model
#' chi-square (the average relative increase in variance and the fraction of
#' missing information it implies), and `chi_bar_naive` is the plain mean of the
#' per-imputation statistics for comparison; the `*_null` entries are the same
#' quantities for the independence baseline. `D2_F` is the rule's raw statistic and
#' is reported unfloored, so it is negative whenever the between-imputation
#' variability of the component statistics exceeds the pooled discrepancy -- a
#' diagnostic of the pool rather than a fit statistic. The reported fit is not
#' affected: the pooled chi-square is floored at zero and its p-value is 1 in that
#' case. `chi_cfi` and `chi_null_cfi` are the
#' pooled model and baseline **chi-squares** on the common \eqn{N - 1}
#' noncentrality scale. `chi_cfi` is the statistic the reported RMSEA is formed
#' from; the pair also gives a reference CFI formed the conventional way,
#' `1 - (chi_cfi - df) / (chi_null_cfi - df_null)` (and analogously for TLI) --
#' a different quantity from the reported CFI/TLI, which average the
#' per-imputation indices.}
#' \item{mi_admissibility}{Admissibility and convergence of the component fits, kept on
#' the pooled object so a saved solution carries the record independently of `fits`: `m`
#' (the number of imputations that entered the pool), `heywood_imputations` (the indices
#' of the fits with at least one Heywood case -- an improper solution where a variable's
#' communality is at or above 1, or its uniqueness is fixed at the estimation
#' boundary), `n_heywood_items` (the number of flagged variables per imputation),
#' `nonconverged` (the indices whose extraction
#' reported a non-zero convergence code), and `iter` (the iterations each extraction
#' used). Averaging aligned solutions pulls boundary communalities back inside the
#' admissible range, so a pooled matrix with no Heywood case can still rest on component
#' fits that had them; `summary()` reports the pooled count together with these.}
#' \item{fits}{The list of \eqn{m} component [efa_fit()] fits, in the order of
#' `data_list`, kept for per-imputation diagnostics. On the MI2S path these are the
#' per-imputation fits whose inputs were pooled, not the pooled fit itself (which
#' is `mi_fit`).}
#' \item{alignment}{Metadata from aligning the rotated solutions, `NULL` when no
#' rotation was requested or on the MI2S path (one fit, one gauge). Under
#' `target_method = "first_target"`: the `method` used, the `target` it aligned to,
#' the per-imputation `target_rotations`, the indices of any
#' `point_rotation_failures`, and whether every inner alignment `converged`. Under
#' `target_method = "consensus"` it is the full [efa_procrustes()]-based GPA record:
#' the converged `target`, the `aligned_loadings` and `aligned_phi`, the iteration
#' `history`, convergence flags, and the multi-start summary.}
#' \item{settings}{The component fits' [efa_fit()] settings with the pooling
#' settings added: `pooled` (always `TRUE`), `pooled_N` and `N` (the mean N across
#' imputations), `n_imputations`, `component_se` (the `se` the component fits
#' used), `target_method`, `align_unrotated`, `fit_pool_method`, `p`, `ci` and
#' `rmsea_ci_level`. `se` records what was actually pooled, so it
#' is `"none"` when pooled standard errors could not be produced although the
#' component fits computed them (`component_se` keeps the request).}
#' }
#'
#'
#' @references
#' Anderson, T. W., & Rubin, H. (1956). Statistical inference in factor analysis.
#' In *Proceedings of the Third Berkeley Symposium on Mathematical Statistics and
#' Probability* (Vol. 5, pp. 111-150). University of California Press.
#'
#' Archer, C. O., & Jennrich, R. I. (1973). Standard errors for rotated factor
#' loadings. *Psychometrika*, 38(4), 581-592.
#'
#' Barnard, J., & Rubin, D. B. (1999). Small-sample degrees of freedom with
#' multiple imputation. *Biometrika*, 86(4), 948-955.
#'
#' Bentler, P. M. (1990). Comparative fit indexes in structural models.
#' *Psychological Bulletin*, 107(2), 238-246.
#'
#' Chung, S., & Cai, L. (2019). Alternative multiple imputation inference for
#' categorical structural equation modeling. *Multivariate Behavioral
#' Research*, 54(3), 323-337.
#'
#' Gower, J. C. (1975). Generalized Procrustes analysis. *Psychometrika*,
#' 40(1), 33-51.
#'
#' Jennrich, R. I. (1973). Standard errors for obliquely rotated factor
#' loadings. *Psychometrika*, 38(4), 593-604.
#'
#' Jennrich, R. I. (1974). Simplified formulae for standard errors in
#' maximum-likelihood factor analysis. *British Journal of Mathematical and
#' Statistical Psychology*, 27(1), 122-131.
#'
#' Lawley, D. N., & Maxwell, A. E. (1971). *Factor analysis as a statistical
#' method* (2nd ed.). Butterworths.
#'
#' Li, K. H., Meng, X.-L., Raghunathan, T. E., & Rubin, D. B. (1991).
#' Significance levels from repeated p-values with multiply-imputed data.
#' *Statistica Sinica*, 1(1), 65-92.
#'
#' Lorenzo-Seva, U., & Van Ginkel, J. R. (2016). Multiple imputation of missing
#' values in exploratory factor analysis of multidimensional scales. *Anales de
#' Psicologia*, 32(2), 596-608.
#'
#' Rubin, D. B. (1987). *Multiple imputation for nonresponse in surveys*. Wiley.
#'
#' Schoenemann, P. H. (1966). A generalized solution of the orthogonal
#' Procrustes problem. *Psychometrika*, 31(1), 1-10.
#'
#' Sriutaisuk, S., Liu, Y., Chung, S., Kim, H., & Gu, F. (2025). Evaluating
#' imputation-based fit statistics in structural equation modeling with ordinal
#' data: The MI2S approach. *Educational and Psychological Measurement*, 85(1),
#' 82-113.
#'
#' Tucker, L. R., & Lewis, C. (1973). A reliability coefficient for maximum
#' likelihood factor analysis. *Psychometrika*, 38(1), 1-10.
#'
#' van Ginkel, J. R., & Kroonenberg, P. M. (2014). Using generalized Procrustes
#' analysis for multiple imputation in principal component analysis. *Journal of
#' Classification*, 31(2), 242-269.
#'
#' Zhang, G., & Preacher, K. J. (2015). Factor rotation and standard errors in
#' exploratory factor analysis. *Journal of Educational and Behavioral
#' Statistics*, 40(6), 579-603.
#'
#' Zhang, G., Preacher, K. J., & Jennrich, R. I. (2012). The infinitesimal
#' jackknife with exploratory factor analysis. *Psychometrika*, 77(4), 634-648.
#'
#' @family factor analysis
#'
#' @export
#'
#' @examples
#'
#' # create a list of three datasets, mimicking a list you would obtain from
#' # e.g. mice.
#' dat_list <- lapply(1:3, function(x) GRiPS_raw[sample(1:nrow(GRiPS_raw), replace = TRUE),])
#' mod <- efa_mi(dat_list, n_factors = 1, estimator = "ML")
#' mod
#'
#' \donttest{
#' # add computation of standard errors and CIs
#' mod <- efa_mi(dat_list, n_factors = 1, estimator = "ML", se = "np-boot")
#' mod
#' }
efa_mi <- function(data_list,
                   p = 0.05,
                   target_method = c("first_target", "consensus"),
                   align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
                   fit_pool_method = c("D2"),
                   consensus_args = list(),
                   procrustes_args = list(),
                   rmsea_ci_level = .90,
                   rmsr_upper = lifecycle::deprecated(),
                   ...) {

  efa_args <- list(...)

  # A flat tuning knob or the former `method` spelling in the dots would only surface from
  # inside the first per-imputation fit, with the error attributed to efa_fit(); reject it
  # here so the message names the function the user called.
  .reject_flat_knobs(...names(), fn = "efa_mi")

  # The per-imputation do.call(efa_fit, ...) binds the dots by R's argument matching,
  # which accepts unique abbreviations of efa_fit()'s formals (e.g. `rotate_c =` for
  # `rotate_control`), but the pooling routes below read `efa_args` by exact name.
  # Canonicalize abbreviated names up front with the same exact-first, unique-prefix
  # matching do.call() applies; ambiguous or unknown names are left untouched for
  # efa_fit() to reject.
  arg_names <- names(efa_args)
  if (length(arg_names)) {
    efa_fit_formals <- setdiff(names(formals(efa_fit)), "...")
    canonical <- pmatch(arg_names, efa_fit_formals)
    names(efa_args)[!is.na(canonical)] <- efa_fit_formals[canonical[!is.na(canonical)]]
  }

  # A mids object is itself a list, so it passes the checks below and only fails deep in
  # the per-dataset assertion with a message naming the lapply variable. It is the most
  # likely wrong input here, so name it and give the conversion.
  if (inherits(data_list, "mids")) {
    cli::cli_abort(
      c("{.arg data_list} must be a list of completed datasets, not a {.cls mids} object.",
        "i" = "Convert it first with {.code mice::complete(x, \"all\")}."),
      class = "efa_pooled_mids_input"
    )
  }

  checkmate::assert_list(data_list, null.ok = FALSE)
  # Arity is asserted here rather than through checkmate's `min.len`, so that too few
  # imputations signal the documented `efa_pooled_min_fits` condition instead of a
  # plain checkmate error.
  if (length(data_list) < 2L) {
    cli::cli_abort(
      c("At least two imputations are required for MI pooling.",
        "i" = "{.arg data_list} has {length(data_list)} element{?s}."),
      class = "efa_pooled_min_fits"
    )
  }
  lapply(data_list, checkmate::assert_multi_class, c("matrix", "data.frame"))
  checkmate::assert_number(p, na.ok = FALSE, lower = 0, upper = 1)
  checkmate::assert_list(consensus_args, null.ok = FALSE)
  checkmate::assert_list(procrustes_args, null.ok = FALSE)
  .efa_pooled_check_procrustes_args(procrustes_args)
  checkmate::assert_number(rmsea_ci_level, na.ok = FALSE, lower = 0, upper = 1)

  # `rmsr_upper` never changed a returned value; the argument's documentation gives the
  # argument, and the warning below repeats it for the caller.
  if (lifecycle::is_present(rmsr_upper)) {
    lifecycle::deprecate_warn(
      when = "1.1.0",
      what = "efa_mi(rmsr_upper)",
      details = c(
        "i" = paste("It selected between two RMSR conventions that coincide for a symmetric",
                    "residual matrix, so it never affected the reported RMSR."),
        "i" = paste("RMSR is the root mean square of the unique off-diagonal residuals; SRMR,",
                    "which divides the same sum by the non-redundant elements, is reported",
                    "alongside it.")
      )
    )
  }

  if (p <= 0 || p >= 1) {
    cli::cli_abort("{.arg p} must be strictly between 0 and 1.", class = "efa_pooled_bad_p")
  }
  if (rmsea_ci_level <= 0 || rmsea_ci_level >= 1) {
    cli::cli_abort("{.arg rmsea_ci_level} must be strictly between 0 and 1.", class = "efa_pooled_bad_ci_level")
  }
  if (!is.null(efa_args$ci) && length(efa_args$ci) == 1L &&
      is.finite(efa_args$ci) &&
      abs(as.numeric(efa_args$ci) - (1 - p)) > sqrt(.Machine$double.eps)) {
    cli::cli_warn("{.fn efa_mi} uses {.arg p}, not the component {.fn efa_fit} argument {.arg ci}, to set pooled bootstrap/MI confidence intervals.",
                  class = "efa_pooled_ci_ignored")
  }

  target_method <- .match_arg_ci(target_method)
  align_unrotated <- .match_arg_ci(align_unrotated)
  fit_pool_method <- .match_arg_ci(fit_pool_method)

  m_imp <- length(data_list)

  ## -------------------------------------------------------------------------
  ## Fit EFA to each imputed dataset
  ## -------------------------------------------------------------------------

  # Pooling is all-or-nothing, so one unfittable imputation ends the call. Name it:
  # the component condition is raised by efa_fit() and identifies no imputation, which
  # leaves a long `data_list` to be bisected by hand. The original condition is kept as
  # the parent so its own diagnosis stays visible.
  fits <- vector("list", m_imp)
  # The imputations are versions of one data set, so the per-fit "computing correlations
  # from the raw data" note states one fact about the input; report it once instead of once
  # per imputation.
  .cor_note_once(
    for (d in seq_len(m_imp)) {
      fits[[d]] <- tryCatch(
        do.call(efa_fit, c(list(x = data_list[[d]]), efa_args)),
        error = function(e) {
          cli::cli_abort(
            c("Imputation {d} could not be fitted.",
              "i" = "Inspect {.code data_list[[{d}]]}, then re-fit or replace it."),
            class = "efa_pooled_fit_failed", parent = e
          )
        }
      )
    }
  )
  # `fits` is returned to the user, so it keeps whatever names `data_list` carried.
  names(fits) <- names(data_list)

  .efa_pooled_check_fits(fits)

  # Select the SE-pooling pathway from the component fits' shared se method. A
  # heterogeneous se cannot be pooled into a single MI estimate, so abort rather
  # than silently producing an uninterpretable mixture.
  route <- .efa_pooled_route(fits)
  if (identical(route, "mixed")) {
    .efa_pooled_mixed_se_abort(fits, data_list)
  }

  settings <- fits[[1]]$settings
  estimator <- settings$estimator
  rotation <- settings$rotation
  var_names <- rownames(fits[[1]]$orig_R)
  if (is.null(var_names)) {
    var_names <- colnames(data_list[[1]])
  }

  rotation_type <- if (is.null(rotation)) "none" else .rotation_family(rotation)

  ## -------------------------------------------------------------------------
  ## MI2S route (Chung & Cai 2019; Sriutaisuk et al. 2025): when the component
  ## fits carry sandwich (robust) standard errors, pool the correlation matrix
  ## and the asymptotic covariance of its off-diagonal entries across imputations
  ## and fit a single model on the pooled inputs. The single fit yields native
  ## scaled-shifted chi-square and sandwich SEs that reflect the multiple-
  ## imputation uncertainty, so this path bypasses the per-imputation rotation
  ## alignment and Rubin pooling used by the bootstrap/information routes below.
  ## -------------------------------------------------------------------------
  if (identical(route, "sandwich")) {
    return(.efa_pooled_mi2s(
      fits = fits, data_list = data_list, efa_args = efa_args,
      settings = settings, estimator = estimator, rotation = rotation,
      rotation_type = rotation_type, target_method = target_method,
      align_unrotated = align_unrotated, fit_pool_method = fit_pool_method,
      p = p, rmsea_ci_level = rmsea_ci_level
    ))
  }

  ## -------------------------------------------------------------------------
  ## Extract and align unrotated loadings
  ## -------------------------------------------------------------------------

  unrot_loadings <- lapply(.extract_list_object(fits, "unrot_loadings"),
                           .change_class, "matrix")

  aligned_unrot <- .efa_pooled_align_unrotated_list(
    unrot_loadings = unrot_loadings,
    align_unrotated = align_unrotated
  )
  unrot_loadings_aligned <- aligned_unrot$loadings
  unrot_align_meta       <- aligned_unrot$meta

  mean_unrot_loadings <- .average_matrices(unrot_loadings_aligned)

  ## -------------------------------------------------------------------------
  ## Align rotated loadings and Phi, if a rotation was requested
  ## -------------------------------------------------------------------------

  # A single-factor solution cannot be rotated and has no factor correlations, so
  # an oblique request on one factor is treated as the no-Phi case throughout,
  # matching a single-fit EFA() (which returns Phi = NULL for one factor).
  oblique <- rotation_type == "oblique" && ncol(unrot_loadings_aligned[[1]]) >= 2L

  phis <- NULL
  structure_loadings <- NULL

  if (rotation_type != "none") {
    rot_loadings_initial <- lapply(.extract_list_object(fits, "rot_loadings"),
                                   .change_class, "matrix")

    if (target_method == "first_target") {
      target_rotations <- vector("list", m_imp)
      rot_loadings <- vector("list", m_imp)
      phis <- vector("list", m_imp)
      # One factor is aligned by sign only; an oblique Procrustes is neither
      # needed nor well defined for a single column.
      proc_rotation <- if (oblique) "oblique" else "orthogonal"

      rot_loadings[[1]] <- rot_loadings_initial[[1]]
      if (oblique) {
        phis[[1]] <- fits[[1]]$Phi
      }

      point_rotation_failures <- logical(m_imp)
      for (d in 2:m_imp) {
        target_rotations[[d]] <- do.call(
          efa_procrustes,
          c(list(A = unrot_loadings[[d]],
                 Target = rot_loadings_initial[[1]],
                 rotation = proc_rotation),
            procrustes_args)
        )
        point_rotation_failures[d] <- isFALSE(target_rotations[[d]]$valid)
        rot_loadings[[d]] <- target_rotations[[d]]$loadings
        if (oblique) {
          phis[[d]] <- target_rotations[[d]]$Phi
        }
      }

      if (any(point_rotation_failures)) {
        cli::cli_warn(
          c("At least one fixed-target Procrustes alignment could not be aligned to a valid rotation.",
            "i" = "The pooled point estimates still use the best available {.fn efa_procrustes} alignment; inspect {.code alignment$point_rotation_failures}."),
          class = "efa_pooled_align_failed"
        )
      }

      final_target <- rot_loadings_initial[[1]]
      inner_converged <- vapply(target_rotations[-1L], function(x) {
        isTRUE(x$valid) && (is.null(x$convergence) || isTRUE(x$convergence))
      }, logical(1L))
      alignment <- list(method = "first_target",
                        target = final_target,
                        target_rotations = target_rotations,
                        point_rotation_failures = which(point_rotation_failures),
                        converged = all(inner_converged))

    } else if (target_method == "consensus") {
      # The GPA iteration starts from one imputation's rotated solution and moves it
      # toward the centroid, inheriting the rotational gauge of wherever it started, so
      # the start decides the gauge of the pooled rotated solution. Start it at the
      # medoid rather than at the first imputation: the medoid is a property of the set,
      # so the converged target -- and with it the pooled rotated pattern -- no longer
      # depends on the order of `data_list`. `"first_target"` keeps its documented
      # first-imputation anchor, which is what its name promises. An explicit `start` in
      # `consensus_args` wins; it is matched the way do.call() would match it, so a
      # partially spelled one is not silently duplicated.
      consensus_call <- consensus_args
      start_supplied <- any(!is.na(pmatch(names(consensus_args), "start")))
      if (!start_supplied) {
        consensus_call$start <- .efa_pooled_medoid_anchor(rot_loadings_initial)
      }
      consensus <- do.call(
        .gpa_consensus_target,
        c(list(unrotated_list = unrot_loadings,
               init_targets = rot_loadings_initial,
               rotation = rotation_type),
          consensus_call)
      )

      if (!isTRUE(consensus$converged)) {
        cli::cli_warn(
          c("Consensus Procrustes alignment did not meet its convergence criterion.",
            "i" = "Inspect {.code alignment$history} and consider stricter or multi-start {.arg consensus_args}."),
          class = "efa_pooled_align_failed"
        )
      }

      rot_loadings <- consensus$aligned_loadings
      phis <- consensus$aligned_phi
      final_target <- consensus$target
      alignment <- consensus
    }

    mean_rot_loadings <- .average_matrices(rot_loadings)
    rownames(mean_rot_loadings) <- var_names

    if (oblique) {
      mean_phis <- .average_matrices(phis)
      # Ensure mean_phis is symmetric (needed, as due to floating point imprecision,
      # the averaged matrix can be asymmetric, leading to issues in downstream checks
      # of matrix symmetry)
      mean_phis <- (mean_phis + t(mean_phis)) / 2
      structure_loadings <- Map(function(L, Phi) L %*% Phi, rot_loadings, phis)
      # Keep Structure parallel to the returned pooled pattern matrix and Phi:
      # it is the plug-in structure of the pooled solution, not the arithmetic
      # mean of the imputation-specific structure matrices.
      mean_structure_loadings <- mean_rot_loadings %*% mean_phis
      dimnames(mean_structure_loadings) <- dimnames(mean_rot_loadings)
    } else {
      mean_phis <- NULL
      mean_structure_loadings <- NULL
    }

  } else {
    rot_loadings <- NULL
    mean_rot_loadings <- NULL
    mean_phis <- NULL
    mean_structure_loadings <- NULL
    final_target <- NULL
    alignment <- NULL
  }

  ## -------------------------------------------------------------------------
  ## Pooled observed and model-implied correlation matrices; residuals and RMSR
  ## -------------------------------------------------------------------------

  orig_R_list <- .extract_list_object(fits, "orig_R")

  pooled_orig_R <- .average_matrices(orig_R_list)

  if (oblique) {
    common_R <- mean_rot_loadings %*% mean_phis %*% t(mean_rot_loadings)
  } else if (rotation_type != "none") {
    common_R <- mean_rot_loadings %*% t(mean_rot_loadings)
  } else {
    common_R <- mean_unrot_loadings %*% t(mean_unrot_loadings)
  }

  h2 <- diag(common_R)
  names(h2) <- rownames(common_R)

  model_implied_R <- common_R
  diag(model_implied_R) <- 1
  dimnames(model_implied_R) <- list(var_names, var_names)

  residuals <- pooled_orig_R - model_implied_R
  diag(residuals) <- 0
  dimnames(residuals) <- list(var_names, var_names)

  RMSR <- .rmsr(residuals)

  ## -------------------------------------------------------------------------
  ## Variance-accounted tables
  ## -------------------------------------------------------------------------

  mean_vars_accounted <- .compute_vars(
    L_unrot = mean_unrot_loadings,
    L_rot = mean_unrot_loadings,
    Phi = NULL
  )

  if (oblique) {
    mean_vars_accounted_rot <- .compute_vars(
      L_unrot = mean_unrot_loadings,
      L_rot = mean_rot_loadings,
      Phi = mean_phis
    )
  } else if (rotation_type != "none") {
    mean_vars_accounted_rot <- .compute_vars(
      L_unrot = mean_unrot_loadings,
      L_rot = mean_rot_loadings,
      Phi = NULL
    )
  } else {
    mean_vars_accounted_rot <- NULL
  }

  ## -------------------------------------------------------------------------
  ## Fit indices: residual-based from pooled residuals, chi-square-based via D2
  ## -------------------------------------------------------------------------

  Ns <- .efa_pooled_get_Ns(data_list, fits, efa_args)
  Ns_ok <- Ns[is.finite(Ns)]
  if (length(Ns_ok) == 0L) {
    N_pool <- NA_real_
    if (!identical(estimator, "PAF")) {
      cli::cli_warn(
        c("N could not be recovered for any imputation.",
          "i" = "Chi-square-based fit indices cannot be computed."),
        class = "efa_pooled_no_n"
      )
    }
  } else {
    if (length(Ns_ok) < length(Ns)) {
      cli::cli_warn(
        c("N could not be recovered for every imputation.",
          "i" = "Fit indices use the mean of the available Ns."),
        class = "efa_pooled_partial_n"
      )
    }
    if (length(unique(Ns_ok)) > 1L) {
      cli::cli_warn(
        c("The imputed datasets appear to have different N.",
          "i" = "Fit indices use the mean N across imputations."),
        class = "efa_pooled_unequal_n"
      )
    }
    N_pool <- mean(Ns_ok)
  }

  pooled_fit <- .efa_pooled_fit_indices(
    fits = fits,
    pooled_R = pooled_orig_R,
    residuals = residuals,
    RMSR = RMSR,
    N = N_pool,
    Ns = Ns,
    pool_method = fit_pool_method,
    rmsea_ci_level = rmsea_ci_level
  )
  fit_indices <- pooled_fit$fit_indices
  mi_diagnostics <- pooled_fit$mi_diagnostics

  ## -------------------------------------------------------------------------
  ## SEs and CIs for pooled MI estimates. Two routes: a bootstrap pool that
  ## requires component-level replicate cubes, and an analytic pool that
  ## consumes the per-imputation closed-form SE matrices populated by
  ## se = "information" via Rubin's (1987) rules.
  ## -------------------------------------------------------------------------

  # The route fixes which pool runs: "information" consumes the per-imputation
  # closed-form SE matrices, "np-boot" the replicate cubes. A pooling failure on
  # either route is non-fatal and falls back to point-estimate-only pooling,
  # signalling the umbrella `efa_pooled_se_unavailable` condition. The "sandwich"
  # route returned via .efa_pooled_mi2s() above, and "none" requests no SEs at all.
  boot_pooled <- NULL
  se_failed <- FALSE
  # Shared fallback: a classed analytic-pool abort becomes a point-estimate-only
  # solution flagged with the umbrella condition. `route` (captured here) resolves
  # to the active route inside each branch, so it labels the warning correctly.
  se_unavailable_handler <- function(e) {
    se_failed <<- TRUE
    .efa_pooled_se_unavailable(route, parent = e)
    NULL
  }
  if (identical(route, "information")) {
    # Catch only the analytic pool's own classed aborts (an unreliable/absent
    # unrotated covariance, or missing alignment metadata) and fall back to
    # point-estimate-only pooling; an unexpected error is a real bug and must
    # propagate rather than be silently downgraded to "se unavailable".
    boot_pooled <- tryCatch(
      .efa_pooled_analytic_pool(
        fits = fits,
        unrot_loadings_aligned = unrot_loadings_aligned,
        align_meta = unrot_align_meta,
        ci = 1 - p,
        align_unrotated = align_unrotated,
        rotation_type = rotation_type,
        rot_loadings = rot_loadings,
        phis = phis,
        structure_loadings = structure_loadings,
        mean_structure_loadings = mean_structure_loadings,
        mean_phis = mean_phis,
        h2 = h2
      ),
      efa_pooled_unreliable_vcov = se_unavailable_handler,
      efa_pooled_no_vcov = se_unavailable_handler,
      efa_pooled_analytic_align_meta_missing = se_unavailable_handler
    )
  } else if (identical(route, "np-boot")) {
    # The bootstrap pool soft-returns NULL (with its own classed warning) when it
    # cannot pool; that NULL is handled by the fallback check below. It has no
    # classed aborts of its own, so call it directly and let any unexpected error
    # propagate rather than masking it as "se unavailable".
    boot_pooled <- .efa_pooled_bootstrap_pool(
      fits = fits,
      orig_R_list = orig_R_list,
      unrot_loadings_aligned = unrot_loadings_aligned,
      mean_unrot_loadings = mean_unrot_loadings,
      rot_loadings = rot_loadings,
      phis = phis,
      structure_loadings = structure_loadings,
      mean_structure_loadings = mean_structure_loadings,
      final_target = final_target,
      rotation_type = rotation_type,
      align_unrotated = align_unrotated,
      procrustes_args = procrustes_args,
      h2 = h2,
      residuals = residuals,
      alpha = p
    )
  }

  # A soft failure (the helper returned NULL after emitting its own specific
  # condition, e.g. too few bootstrap replicates) still downgrades to no SE
  # pooling; surface the umbrella condition unless the tryCatch handler already
  # did so for a classed abort.
  if ((identical(route, "information") || identical(route, "np-boot")) &&
      is.null(boot_pooled) && !se_failed) {
    .efa_pooled_se_unavailable(route)
  }

  ## -------------------------------------------------------------------------
  ## Return object
  ## -------------------------------------------------------------------------

  mean_unrot_loadings <- .change_class(mean_unrot_loadings,
                                       c("efa_loadings", "LOADINGS"))
  if (!is.null(mean_rot_loadings)) {
    mean_rot_loadings <- .change_class(mean_rot_loadings,
                                       c("efa_loadings", "LOADINGS"))
  }
  if (!is.null(mean_structure_loadings)) {
    mean_structure_loadings <- .change_class(mean_structure_loadings,
                                             c("efa_loadings", "LOADINGS"))
  }

  settings_pooled <- settings
  settings_pooled$N <- N_pool
  settings_pooled$pooled_N <- N_pool
  settings_pooled$pooled <- TRUE
  settings_pooled$component_se <- settings$se
  # Downgrade se only on the fallback path: a requested se = "information" /
  # "np-boot" whose pooled SEs could not be produced (an `efa_pooled_se_unavailable`
  # warning was emitted above). The se = "none" route never requested SEs, so it
  # keeps "none" without a warning or downgrade. `component_se` retains the
  # requested method so print.efa_mi() can still describe what the component
  # fits used.
  if (is.null(boot_pooled) && !identical(route, "none")) {
    settings_pooled$se <- "none"
  }
  settings_pooled$n_imputations <- m_imp
  settings_pooled$target_method <- target_method
  settings_pooled$align_unrotated <- align_unrotated
  settings_pooled$fit_pool_method <- fit_pool_method
  settings_pooled$p <- p
  settings_pooled$ci <- 1 - p
  settings_pooled$rmsea_ci_level <- rmsea_ci_level
  if (!is.null(boot_pooled) && !is.null(boot_pooled$n_boot)) {
    settings_pooled$b_boot <- boot_pooled$n_boot
  }

  results <- list(
    h2 = h2,
    unrot_loadings = mean_unrot_loadings,
    vars_accounted = mean_vars_accounted,
    fit_indices = fit_indices,
    model_implied_R = model_implied_R,
    residuals = residuals,
    orig_R = pooled_orig_R,
    settings = settings_pooled,
    fits = fits,
    alignment = alignment,
    mi_diagnostics = mi_diagnostics,
    mi_admissibility = .efa_pooled_admissibility(fits)
  )

  if (rotation_type != "none") {
    results$rot_loadings <- mean_rot_loadings
    results$vars_accounted_rot <- mean_vars_accounted_rot
  }

  if (oblique) {
    results$Phi <- mean_phis
    results$Structure <- mean_structure_loadings
  }

  if (!is.null(boot_pooled)) {
    results$SE <- boot_pooled$SE
    results$CI <- boot_pooled$CI
    results$MI <- boot_pooled$MI
    # Single-bracket list assignment preserves a present-but-NULL `replicates`
    # slot on the analytic path, matching the EFA() schema contract pinned in
    # test-EFA-fields.R; `results$replicates <- NULL` would remove the slot.
    results["replicates"] <- list(boot_pooled$replicates)
    if (!is.null(boot_pooled$SE$residuals)) {
      # The residual diagonal is fixed at 0 with SE 0, so 0/0 would yield NaN on
      # the diagonal; the off-diagonal standardised residuals are the meaningful
      # quantities, so set the diagonal to 0 (matches single-fit EFA()).
      std_resid <- results$residuals / boot_pooled$SE$residuals
      diag(std_resid) <- 0
      results$standardized_residuals <- std_resid
    }
  }

  # The trailing legacy classes are load-bearing: "EFA_POOLED" keeps
  # `inherits(x, "EFA_POOLED")` resolving, and "efa"/"EFA" keep the shared print,
  # format, summary, and residuals methods (and the single-fit accessors) working.
  class(results) <- c("efa_mi", "EFA_POOLED", "efa", "EFA")
  results
}

## =============================================================================
## Internal helpers for efa_mi
## =============================================================================

.efa_pooled_setting_chr <- function(f, name) {
  # A component fit's settings value as a length-1 character (NA when absent), for
  # comparing settings across imputations.
  val <- f$settings[[name]]
  if (is.null(val)) NA_character_ else as.character(val)
}

.efa_pooled_route <- function(fits) {
  # Decide which multiple-imputation SE-pooling pathway a set of component fits
  # selects, from the standard-error method recorded on each fit. Pooling is only
  # defined when every imputation used the same se, so a heterogeneous (or
  # absent) se returns "mixed" and the caller aborts. Returns one of "none",
  # "np-boot", "information", "sandwich", or "mixed".
  ses <- vapply(fits, .efa_pooled_setting_chr, character(1), name = "se")
  if (anyNA(ses) || length(unique(ses)) != 1L) {
    return("mixed")
  }
  ses[[1L]]
}

.efa_pooled_check_procrustes_args <- function(procrustes_args) {
  # Restrict `procrustes_args` to efa_procrustes() algorithm controls. The matrices and
  # the rotation family are derived from the imputations, and `S` is the cross-product
  # of the loadings being rotated: supplying any of them here would either duplicate a
  # formal do.call() already supplies or replace an internally derived quantity with one
  # that does not belong to the alignment being run. efa_procrustes() has no dots, so R
  # would match a partially spelled name onto a formal; canonicalize the names the same
  # way before deciding, or `Tar =` would slip through as an override of `Target`.
  # `charmatch()` rather than `pmatch()`: it separates an abbreviation that matches
  # several formals (0) from one that matches none (NA), so an ambiguous name is
  # diagnosed as ambiguous instead of as a typo.
  if (length(procrustes_args) == 0L) {
    return(invisible(TRUE))
  }
  nms <- names(procrustes_args)
  if (is.null(nms) || any(!nzchar(nms))) {
    cli::cli_abort(
      c("Every element of {.arg procrustes_args} must be named.",
        "i" = "Positional arguments cannot be matched to {.fn efa_procrustes} formals."),
      class = "efa_pooled_bad_procrustes_args"
    )
  }

  procrustes_formals <- names(formals(efa_procrustes))
  matched <- charmatch(nms, procrustes_formals)
  resolved <- !is.na(matched) & matched > 0L
  canonical <- nms
  canonical[resolved] <- procrustes_formals[matched[resolved]]
  reserved <- c("A", "Target", "rotation", "S")

  bad_reserved <- nms[canonical %in% reserved]
  if (length(bad_reserved) > 0L) {
    cli::cli_abort(
      c("{.arg procrustes_args} must not set {.arg {bad_reserved}}.",
        "i" = "{.fn efa_mi} derives the loadings, the alignment target, the rotation family, and the cross-product {.arg S} from the imputations; pass only {.fn efa_procrustes} algorithm controls (for example {.arg oblique_maxit})."),
      class = "efa_pooled_bad_procrustes_args"
    )
  }

  ambiguous <- nms[!is.na(matched) & matched == 0L]
  if (length(ambiguous) > 0L) {
    cli::cli_abort(
      c("{cli::qty(length(ambiguous))}{?An/Some} abbreviated {.arg procrustes_args} name{?s} {cli::qty(length(ambiguous))}match{?es/} more than one {.fn efa_procrustes} argument.",
        "x" = "Ambiguous: {.val {ambiguous}}.",
        "i" = "Spell {cli::qty(length(ambiguous))}{?it/them} out."),
      class = "efa_pooled_bad_procrustes_args"
    )
  }

  bad_unknown <- nms[is.na(matched)]
  if (length(bad_unknown) > 0L) {
    cli::cli_abort(
      c("{.arg procrustes_args} may only contain {.fn efa_procrustes} arguments.",
        "x" = "{cli::qty(length(bad_unknown))}Unknown name{?s}: {.val {bad_unknown}}.",
        "i" = "Available controls: {.arg {setdiff(procrustes_formals, reserved)}}."),
      class = "efa_pooled_bad_procrustes_args"
    )
  }

  invisible(TRUE)
}

.efa_pooled_mixed_se_abort <- function(fits, data_list) {
  # Abort a pool whose component fits recorded different `se` methods, naming the
  # imputations that broke the shared method. A caller passes one `se` to efa_mi(), so a
  # mixture is never something they set directly: efa_fit() records `se = "none"` for a
  # fit whose requested standard errors it could not compute -- a correlation matrix
  # under `se = "np-boot"` being the reachable case -- and "re-fit with the same se" is
  # then advice the caller already followed. Point at the imputations instead.
  ses <- vapply(fits, .efa_pooled_setting_chr, character(1), name = "se")
  known <- ses[!is.na(ses)]
  majority <- if (length(known) > 0L) {
    names(sort(table(known), decreasing = TRUE))[[1L]]
  } else {
    NA_character_
  }
  odd <- which(is.na(ses) | ses != majority)

  bullets <- "The component {.fn efa_fit} fits use different {.arg se} methods, so their standard errors cannot be pooled."
  if (!is.na(majority) && length(odd) > 0L) {
    odd_ses <- unique(ses[odd])
    n_rest <- length(ses) - length(odd)
    bullets <- c(
      bullets,
      "x" = "{cli::qty(length(odd))}Imputation{?s} {.val {odd}} recorded {.code se} {.val {odd_ses}}; the remaining {n_rest} recorded {.val {majority}}."
    )
  }

  cormat_odd <- odd[vapply(data_list[odd], .is_cormat, logical(1))]
  bullets <- if (length(cormat_odd) > 0L) {
    c(bullets,
      "i" = "{cli::qty(length(cormat_odd))}Imputation{?s} {.val {cormat_odd}} {cli::qty(length(cormat_odd))}{?is a correlation matrix/are correlation matrices}, which cannot be bootstrapped, so {.fn efa_fit} fitted {cli::qty(length(cormat_odd))}{?it/them} with {.code se = \"none\"}.")
  } else {
    c(bullets,
      "i" = "Re-fit every imputation with the same {.arg se} (all {.val none}, {.val information}, {.val sandwich}, or {.val np-boot}).")
  }

  cli::cli_abort(bullets, class = "efa_pooled_mixed_se")
}

.efa_pooled_admissibility <- function(fits) {
  # Admissibility and convergence of the component fits, aggregated onto the pooled
  # object. Averaging aligned solutions pulls boundary communalities back inside the
  # admissible range, so a pooled matrix can be proper while several component fits were
  # not; without this record the pooled object holds no evidence of that outside `$fits`,
  # and the per-fit conditions raised at fitting time are gone once it is saved.
  n_heywood <- vapply(fits, function(f) length(f[["heywood"]]), integer(1))
  # Read non-convergence with the same predicate the print layer uses, so the recorded
  # indices cannot disagree with the banner a component fit would print on its own.
  nonconverged <- vapply(fits, function(f) {
    .efa_iteration_nonconvergence(list(convergence = f[["convergence"]],
                                       iter = f[["iter"]],
                                       max_iter = f$settings$max_iter))
  }, logical(1))
  iter <- vapply(fits, function(f) {
    it <- f[["iter"]]
    if (is.numeric(it) && length(it) == 1L) as.integer(it) else NA_integer_
  }, integer(1))

  list(
    m = length(fits),
    heywood_imputations = which(n_heywood > 0L),
    n_heywood_items = n_heywood,
    nonconverged = which(nonconverged),
    iter = iter
  )
}

.efa_pooled_se_unavailable <- function(route, parent = NULL) {
  # Umbrella condition for "the requested standard-error method could not be
  # pooled". Emitted whenever efa_mi() falls back to point-estimate-only
  # pooling on the information / np-boot routes, layered over any specific
  # condition the pooling helper raised (carried as `parent` for a classed
  # abort, NULL when the helper soft-returned NULL after its own warning).
  cli::cli_warn(
    c("Pooled standard errors could not be produced for the {.val {route}} standard-error method; the pooled point estimates are returned without standard errors.",
      "i" = "Inspect the component {.fn efa_fit} fits and any preceding condition for the cause."),
    parent = parent, class = "efa_pooled_se_unavailable"
  )
}

## -----------------------------------------------------------------------------
## MI2S (multiple imputation, two-stage) pooled-inputs route for sandwich SEs
## -----------------------------------------------------------------------------

.efa_pooled_mi2s_inputs <- function(fits) {
  # Pool the per-imputation correlation matrices and their asymptotic covariances
  # into a single correlation matrix r_bar and a single ACOV Gamma_tilde via the
  # two-stage pooled-input ACOV of Chung & Cai (2019) and Sriutaisuk, Liu, Chung,
  # Kim & Gu (2025, Educ. Psychol. Meas.). This is the ACOV-pooling step of MI2S,
  # distinct from the parameter-pooling Rubin's-rules routes (which carry FMI/df):
  #   r_bar       = (1/m) sum_d R_d                         (element-wise mean)
  #   Gamma_W     = (1/m) sum_d Gamma_d                     (within-imputation ACOV)
  #   Gamma_B     = (1/(m-1)) sum_d (v_d - v_bar)(v_d - v_bar)'   (between, = cov())
  #   Gamma_tilde = Gamma_W + (1 + 1/m) Gamma_B             (total pooled ACOV)
  # All quantities stay on the per-fit variance scale (Var(rho-hat)) and in the
  # utils::combn(p, 2) off-diagonal order that EFA()'s $Gamma uses, so the pooled
  # ACOV feeds the sandwich core unchanged.
  m <- length(fits)
  R_list <- lapply(fits, function(f) as.matrix(f$orig_R))
  G_list <- lapply(fits, function(f) as.matrix(f$Gamma))
  p <- ncol(R_list[[1]])

  idx <- utils::combn(p, 2L)
  pair_ij <- cbind(idx[1L, ], idx[2L, ])
  q <- ncol(idx)

  r_bar <- Reduce(`+`, R_list) / m
  r_bar <- (r_bar + t(r_bar)) / 2          # symmetrise away round-off
  diag(r_bar) <- 1

  # Stack the off-diagonal correlations (combn order) for the between-imputation
  # covariance; stats::cov() uses the m - 1 divisor required by Rubin's rule.
  V <- t(vapply(R_list, function(R) R[pair_ij], numeric(q)))

  Gamma_W <- Reduce(`+`, G_list) / m
  Gamma_B <- stats::cov(V)
  Gamma_tilde <- Gamma_W + (1 + 1 / m) * Gamma_B
  Gamma_tilde <- (Gamma_tilde + t(Gamma_tilde)) / 2

  # Gamma_tilde = Gamma_W + (1 + 1/m) Gamma_B is in exact arithmetic a sum of two
  # positive-semidefinite matrices (Gamma_W averages PSD per-fit ACOVs; Gamma_B is
  # a sample covariance), so it cannot become indefinite from a small m alone. A
  # negative eigenvalue therefore signals floating-point round-off or a degenerate
  # / corrupted per-fit Gamma rather than imputation noise. Abort rather than
  # silently project to the nearest PD matrix (which would distort the sandwich
  # meat and the test statistic without warning).
  ev <- eigen(Gamma_tilde, symmetric = TRUE, only.values = TRUE)$values
  tol <- sqrt(.Machine$double.eps) * max(abs(ev))
  if (min(ev) < -tol) {
    cli::cli_abort(
      c("The pooled asymptotic covariance is not positive semidefinite.",
        "x" = "Its smallest eigenvalue is {.val {min(ev)}}, beyond round-off; a per-imputation asymptotic covariance is likely degenerate or corrupted.",
        "i" = "Check the per-imputation {.code $Gamma} matrices (e.g. a near-singular polychoric correlation); re-fitting the affected imputations usually resolves it."),
      class = "efa_pooled_mi2s_acov_not_psd"
    )
  }

  # Preserve the off-diagonal labelling/order from the per-fit Gamma so the
  # sandwich core and any downstream consumer see the same schema.
  dimnames(Gamma_tilde) <- dimnames(G_list[[1]])
  dimnames(r_bar) <- dimnames(R_list[[1]])

  list(r_bar = r_bar, Gamma_tilde = Gamma_tilde, m = m)
}

.efa_pooled_mi2s <- function(fits, data_list, efa_args, settings, estimator,
                             rotation, rotation_type, target_method,
                             align_unrotated, fit_pool_method, p,
                             rmsea_ci_level) {
  # Two-stage (MI2S) pooled-inputs route for component fits carrying sandwich
  # (robust) SEs. Pools the correlation matrix and its asymptotic covariance
  # across imputations (.efa_pooled_mi2s_inputs), then fits the model once on the
  # pooled inputs via the shared .efa_core(), so the resulting object carries
  # native scaled-shifted chi-square and sandwich SEs that already reflect the
  # multiple-imputation uncertainty. No per-imputation rotation alignment or
  # Rubin pooling of estimates is performed on this path.
  m_imp <- length(fits)

  ## ---- Fail closed on inconsistent inputs --------------------------------
  ses <- vapply(fits, .efa_pooled_setting_chr, character(1), name = "se")
  cor_methods <- vapply(fits, .efa_pooled_setting_chr, character(1),
                        name = "cor_method")
  gamma_ok <- vapply(fits, function(f) {
    !is.null(f$Gamma) && !anyNA(f$Gamma)
  }, logical(1))
  if (anyNA(ses) || !all(ses == "sandwich") ||
      length(unique(cor_methods)) != 1L ||
      !cor_methods[[1L]] %in% c("poly", "tetra", "pearson") ||
      !all(gamma_ok)) {
    # The settings can be perfectly consistent and still unsupported here (e.g.
    # cor_method = "fiml", whose sandwich carries no poolable correlation ACOV);
    # advising the user to "re-fit with the same settings" would then be wrong.
    unsupported_cor <- !anyNA(ses) && all(ses == "sandwich") &&
      length(unique(cor_methods)) == 1L &&
      !cor_methods[[1L]] %in% c("poly", "tetra", "pearson")
    hint <- if (unsupported_cor) {
      c("i" = "The imputations were fitted with {.code cor_method = \"{cor_methods[[1L]]}\"}, which carries no poolable asymptotic correlation covariance; use {.code se = \"np-boot\"} to pool such fits.")
    } else {
      c("i" = "Re-fit every imputation with the same {.code se = \"sandwich\"} setting and {.arg cor_method}.")
    }
    cli::cli_abort(
      c("MI2S pooling requires every imputation to be fitted with {.code se = \"sandwich\"} and the same {.arg cor_method} ({.val poly}, {.val tetra}, or {.val pearson}), each carrying a valid asymptotic covariance.",
        hint),
      class = "efa_pooled_mi2s_inputs_inconsistent"
    )
  }
  cor_method <- cor_methods[[1L]]
  n_factors <- settings$n_factors
  use_setting <- settings$use
  if (is.null(use_setting)) use_setting <- "pairwise.complete.obs"

  ## ---- Sample size -------------------------------------------------------
  Ns <- .efa_pooled_get_Ns(data_list, fits, efa_args)
  Ns_ok <- Ns[is.finite(Ns)]
  if (length(Ns_ok) == 0L) {
    cli::cli_abort(
      c("MI2S sandwich pooling requires the sample size {.arg N}.",
        "i" = "Supply raw data or {.arg N} for the imputations."),
      class = "efa_pooled_mi2s_no_n"
    )
  }
  if (length(unique(Ns_ok)) > 1L) {
    cli::cli_warn(
      c("The imputed datasets appear to have different N.",
        "i" = "MI2S pooling uses the mean N across imputations."),
      class = "efa_pooled_unequal_n"
    )
  }
  N_pool <- mean(Ns_ok)

  ## ---- Imputation-count guidance -----------------------------------------
  if (m_imp < 20L) {
    cli::cli_warn(
      c("MI2S pooling was run with only {m_imp} imputation{?s}.",
        "i" = "The scaled-shifted statistic is calibrated for 20 or more imputations (more at higher rates of missingness; Sriutaisuk et al. 2025); interpret the pooled fit and SEs with caution."),
      class = "efa_pooled_mi2s_n_too_small"
    )
  }

  ## ---- Alignment settings do not apply to a single pooled fit ------------
  if (!identical(target_method, "first_target") ||
      !identical(align_unrotated, "signed_tucker_congruence")) {
    cli::cli_warn(
      c("{.arg target_method} and {.arg align_unrotated} are ignored on the MI2S sandwich path.",
        "i" = "MI2S pools the inputs and fits a single solution, so no per-imputation rotation alignment is performed."),
      class = "efa_pooled_mi2s_alignment_ignored"
    )
  }

  ## ---- Pool inputs and fit once ------------------------------------------
  pooled <- .efa_pooled_mi2s_inputs(fits)
  r_bar <- pooled$r_bar
  Gamma_tilde <- pooled$Gamma_tilde

  weights <- if (estimator == "DWLS") {
    # A non-positive pooled asymptotic variance makes .poly_weight_matrix() abort
    # with the low-level efa_dwls_degenerate_weight; relabel it to the documented
    # MI2S condition with the "increase imputations" remediation.
    tryCatch(
      .poly_weight_matrix(diag(Gamma_tilde), ncol(r_bar)),
      efa_dwls_degenerate_weight = function(e) {
        cli::cli_abort(
          c("The pooled asymptotic covariance has a non-positive diagonal entry, so DWLS weights cannot be formed.",
            "i" = "Increase the number of imputations (Sriutaisuk et al. 2025 recommend 20 or more, and more at higher rates of missingness)."),
          class = "efa_pooled_mi2s_acov_not_psd", parent = e
        )
      }
    )
  } else {
    NULL
  }

  # Reuse the same estimate -> rotate -> SE pipeline efa_fit() runs, with the same
  # estimation/rotation tuning the component fits used: unbundle the two control
  # objects from the component call (defaulting to the constructors' presets,
  # exactly as efa_fit() does) and pass the knobs to .efa_core() under its
  # historical formal names. The arguments resolved explicitly here are dropped
  # from the dots to avoid passing them twice through do.call(); the remaining
  # dots are merged over the rotation-engine extras stored in rotate_control(),
  # with the dots winning on a name clash (mirroring efa_fit()).
  ec <- efa_args$estimate_control
  if (is.null(ec)) ec <- estimate_control()
  rc <- efa_args$rotate_control
  if (is.null(rc)) rc <- rotate_control()
  drop_args <- c("n_factors", "N", "estimator", "rotation", "se", "use",
                 "cor_method", "ci", "seed",
                 "estimate_control", "rotate_control")
  extra_args <- modifyList(rc$extra_args,
                           efa_args[setdiff(names(efa_args), drop_args)])
  core_args <- list(R = r_bar, N = N_pool, weights = weights, Gamma = Gamma_tilde,
                    np_boot = FALSE, estimator = estimator, rotation = rotation,
                    type = ec$type, rot_type = rc$type, n_factors = n_factors,
                    se = "sandwich", ci = 1 - p, use = use_setting,
                    cor_method = cor_method, max_iter = ec$max_iter,
                    init_comm = ec$init_comm, criterion = ec$criterion,
                    criterion_type = ec$criterion_type, abs_eigen = ec$abs_eigen,
                    start_method = ec$start_method, normalize = rc$normalize,
                    precision = rc$precision, order_type = rc$order_type,
                    varimax_type = rc$varimax_type, P_type = rc$p_type, k = rc$k,
                    randomStarts = rc$random_starts)
  # Mirror efa_fit()'s extras splice: an extra whose name is already an explicit
  # core argument would abort do.call() with "matched by multiple actual
  # arguments", so drop it (the resolved value wins).
  extra_args <- extra_args[setdiff(names(extra_args), names(core_args))]
  mi_fit <- do.call(.efa_core, c(core_args, extra_args))

  # .efa_core()/EFA() compute RMSEA confidence bounds at a fixed 90% level; honor
  # the requested rmsea_ci_level on the pooled object by recomputing the bounds
  # from the single fit's pooled test statistic, matching how the information and
  # bootstrap routes apply rmsea_ci_level (a no-op when rmsea_ci_level == 0.90).
  fi <- mi_fit$fit_indices
  # A finite point estimate is required as well as a finite statistic: only the bounds are
  # re-solved here, so without one there is nothing for the new interval to be an interval
  # around, and replacing the bounds would advertise a range for an undefined RMSEA.
  if (!is.null(fi) && is.finite(fi$chi) && is.finite(fi$df) && fi$df > 0 &&
      is.finite(fi$RMSEA)) {
    # `fi$RMSEA` stays as .chi_fit_indices() computed it, so pass it in as the point estimate
    # the new interval has to contain rather than letting the helper derive a second one from
    # `fi$chi` and `N_pool`.
    rmsea_ci <- .efa_pooled_rmsea_ci(fi$chi, fi$df, N_pool, point = fi$RMSEA,
                                     level = rmsea_ci_level)
    fi$RMSEA_LB <- unname(rmsea_ci[["lower"]])
    fi$RMSEA_UB <- unname(rmsea_ci[["upper"]])
    mi_fit$fit_indices <- fi
  }

  # The pooled fit indices are the single fit's, reported under the same names and in
  # the same order as on the other routes so that code reading `names(fit_indices)`
  # does not have to branch on the route. `pool_method` is present but `NA`: this path
  # applies no pooling rule to the chi-square -- the statistic is the pooled-inputs
  # fit's own scaled one -- and an absent field would read as "not recorded" exactly
  # where a caller asks whether the chi-square was D2-pooled. The extra quantities the
  # scaled statistic carries follow the common block. `mi_fit` itself is left as
  # .efa_core() returned it, so it stays comparable with a single efa_fit() solution.
  fit_indices_pooled <- mi_fit$fit_indices
  if (!is.null(fit_indices_pooled)) {
    fit_indices_pooled$pool_method <- NA_character_
    fit_indices_pooled <- .efa_pooled_order_fit_indices(fit_indices_pooled)
  }

  ## ---- Assemble the pooled object ----------------------------------------
  settings_pooled <- settings
  settings_pooled$N <- N_pool
  settings_pooled$pooled_N <- N_pool
  settings_pooled$pooled <- TRUE
  settings_pooled$component_se <- "sandwich"
  settings_pooled$se <- "sandwich"
  settings_pooled$n_imputations <- m_imp
  settings_pooled$target_method <- target_method
  settings_pooled$align_unrotated <- align_unrotated
  settings_pooled$fit_pool_method <- fit_pool_method
  settings_pooled$p <- p
  settings_pooled$ci <- 1 - p
  settings_pooled$rmsea_ci_level <- rmsea_ci_level

  results <- list(
    h2 = mi_fit$h2,
    unrot_loadings = mi_fit$unrot_loadings,
    vars_accounted = mi_fit$vars_accounted,
    fit_indices = fit_indices_pooled,
    model_implied_R = mi_fit$model_implied_R,
    residuals = mi_fit$residuals,
    orig_R = mi_fit$orig_R,
    settings = settings_pooled,
    fits = fits,
    alignment = NULL,
    mi_diagnostics = NULL,
    mi_admissibility = .efa_pooled_admissibility(fits),
    mi_fit = mi_fit
  )

  if (rotation_type != "none") {
    results$rot_loadings <- mi_fit$rot_loadings
    results$vars_accounted_rot <- mi_fit$vars_accounted_rot
  }
  if (rotation_type == "oblique") {
    results$Phi <- mi_fit$Phi
    results$Structure <- mi_fit$Structure
  }

  results$SE <- mi_fit$SE
  results$CI <- mi_fit$CI
  # No per-parameter Rubin pooling on the MI2S path: the imputation uncertainty
  # is carried by Gamma_tilde, not by element-wise between-imputation variance.
  # The slots stay present-but-NULL so the schema matches the other pooled paths.
  results["MI"] <- list(NULL)
  results["replicates"] <- list(NULL)

  # Same load-bearing legacy classes as the Rubin/bootstrap return above.
  class(results) <- c("efa_mi", "EFA_POOLED", "efa", "EFA")
  results
}

.efa_pooled_get_Ns <- function(data_list, fits, efa_args) {
  # Recover the N used in each EFA fit. Correlation-matrix input may not carry N,
  # so return NA unless N was supplied to EFA() or stored in settings.
  vapply(seq_along(data_list), function(d) {
    if (!is.null(fits[[d]]$settings$N) && !is.na(fits[[d]]$settings$N)) {
      return(as.numeric(fits[[d]]$settings$N))
    }
    if (!is.null(efa_args$N) && !is.na(efa_args$N)) {
      return(as.numeric(efa_args$N))
    }
    if (!.is_cormat(data_list[[d]])) {
      return(nrow(data_list[[d]]))
    }
    NA_real_
  }, numeric(1))
}

.efa_pooled_check_fits <- function(fits) {
  # Fail early if the fitted EFA objects are not conformable. Pooling only makes
  # sense when all imputations estimate the same model on the same variables. The
  # arity of the list is already asserted on `data_list` by the caller.
  dims <- vapply(fits, function(x) {
    paste(dim(as.matrix(x$unrot_loadings)), collapse = "x")
  }, character(1))
  if (length(unique(dims)) != 1L) {
    cli::cli_abort("All unrotated loading matrices must have the same dimensions.", class = "efa_pooled_dim_mismatch")
  }

  var_names <- lapply(fits, function(x) rownames(as.matrix(x$orig_R)))
  if (!all(vapply(var_names[-1], identical, logical(1), var_names[[1]]))) {
    cli::cli_abort("All imputations must contain the same variables in the same order.", class = "efa_pooled_var_mismatch")
  }

  for (nm in c("estimator", "rotation", "n_factors")) {
    vals <- vapply(fits, .efa_pooled_setting_chr, character(1), name = nm)
    vals <- vals[!is.na(vals)]
    if (length(unique(vals)) > 1L) {
      cli::cli_abort("All imputations must use the same {.arg {nm}}.", class = "efa_pooled_setting_mismatch")
    }
  }

  invisible(TRUE)
}

# MI2S (multiple imputation, two-stage) pooled-inputs route for se = "sandwich"
# in EFA_POOLED(). Pools the correlation matrix and the asymptotic covariance of
# its off-diagonal entries across imputations (Chung & Cai 2019; Sriutaisuk et
# al. 2025), then fits one model on the pooled inputs.

# ---- deterministic fixtures -------------------------------------------------

# m ordinal imputations from a one-factor model, discretised at the quartiles.
.mi2s_ord_imps <- function(m, n = 300L, p = 6L, base_seed = 100L,
                           identical = FALSE) {
  mk <- function(seed) {
    set.seed(seed)
    f <- stats::rnorm(n)
    X <- outer(f, rep(0.6, p)) + matrix(stats::rnorm(n * p, sd = 0.85), n, p)
    d <- apply(X, 2L, function(col) {
      as.integer(cut(col, c(-Inf, stats::quantile(col, c(.25, .5, .75)), Inf)))
    })
    d <- as.data.frame(d)
    names(d) <- paste0("x", seq_len(p))
    d
  }
  if (identical) {
    rep(list(mk(base_seed)), m)
  } else {
    lapply(seq_len(m), function(d) mk(base_seed + d))
  }
}

# m continuous imputations from a one-factor model.
.mi2s_cont_imps <- function(m, n = 400L, p = 6L, base_seed = 200L) {
  lapply(seq_len(m), function(d) {
    set.seed(base_seed + d)
    f <- stats::rnorm(n)
    X <- outer(f, rep(0.6, p)) + matrix(stats::rnorm(n * p, sd = 0.8), n, p)
    out <- as.data.frame(X)
    names(out) <- paste0("v", seq_len(p))
    out
  })
}

# Minimal fit-shaped mocks for the pure pooling helper: a valid correlation
# matrix and a PSD asymptotic covariance over the off-diagonal pairs.
.mi2s_mock_fit <- function(seed, p = 4L, se = "sandwich", cor_method = "poly") {
  set.seed(seed)
  q <- p * (p - 1L) / 2L
  R <- stats::cor(matrix(stats::rnorm(200L * p), 200L, p))
  G <- crossprod(matrix(stats::rnorm(60L * q), 60L, q)) / 60L
  list(orig_R = R, Gamma = G, unrot_loadings = matrix(0.5, p, 1L),
       settings = list(se = se, cor_method = cor_method, n_factors = 1L,
                       estimator = "ULS",
                       use = "pairwise.complete.obs", type = "EFAtools"))
}

# Run the MI2S route quietly (the m guidance / "not a correlation matrix"
# messages are not under test here).
.mi2s_fit <- function(imps, ...) {
  suppressMessages(suppressWarnings(
    EFA_POOLED(imps, se = "sandwich", ...)
  ))
}


test_that("identity imputations reproduce the single-fit sandwich solution", {
  imps <- .mi2s_ord_imps(m = 4L, identical = TRUE)

  pooled <- .mi2s_fit(imps, n_factors = 1L, method = "ULS", rotation = "none",
                      cor_method = "poly")
  single <- suppressMessages(
    EFA(imps[[1L]], n_factors = 1L, method = "ULS", rotation = "none",
        se = "sandwich", cor_method = "poly")
  )

  # With identical imputations Gamma_B = 0 and r_bar = r_1, so the pooled fit is
  # the single-imputation fit element-wise.
  expect_equal(unclass(pooled$unrot_loadings), unclass(single$unrot_loadings),
               tolerance = 1e-10)
  expect_equal(pooled$SE$unrot_loadings, single$SE$unrot_loadings,
               tolerance = 1e-10)
  expect_equal(pooled$SE$uniquenesses, single$SE$uniquenesses,
               tolerance = 1e-10)
  expect_equal(pooled$fit_indices$chi, single$fit_indices$chi,
               tolerance = 1e-10)
  expect_equal(pooled$fit_indices$chi_shift, single$fit_indices$chi_shift,
               tolerance = 1e-10)
})


test_that("Gamma_tilde equals the Rubin total ACOV to machine precision", {
  fits <- lapply(seq_len(6L), .mi2s_mock_fit, p = 4L)
  out <- .efa_pooled_mi2s_inputs(fits)

  m <- length(fits)
  p <- ncol(fits[[1L]]$orig_R)
  idx <- utils::combn(p, 2L)
  pair <- cbind(idx[1L, ], idx[2L, ])
  R_list <- lapply(fits, `[[`, "orig_R")
  G_list <- lapply(fits, `[[`, "Gamma")

  V <- t(vapply(R_list, function(R) R[pair], numeric(ncol(idx))))
  Gamma_W <- Reduce(`+`, G_list) / m
  # Recompute the between-imputation ACOV from first principles (centred
  # cross-product with the m - 1 divisor) rather than via stats::cov(), so the
  # m - 1 divisor and the (1 + 1/m) factor are checked independently of the
  # implementation.
  Vc <- sweep(V, 2L, colMeans(V))
  Gamma_B <- crossprod(Vc) / (m - 1L)
  Gamma_ref <- Gamma_W + (1 + 1 / m) * Gamma_B
  Gamma_ref <- (Gamma_ref + t(Gamma_ref)) / 2

  expect_equal(unname(out$Gamma_tilde), unname(Gamma_ref), tolerance = 1e-12)

  r_ref <- Reduce(`+`, R_list) / m
  expect_equal(out$r_bar[lower.tri(out$r_bar)], r_ref[lower.tri(r_ref)],
               tolerance = 1e-12)
  # The off-diagonal stacking follows combn(p, 2): pair k is (idx[1,k], idx[2,k]).
  expect_identical(dim(out$Gamma_tilde), c(ncol(idx), ncol(idx)))
})


test_that("a non-PSD pooled ACOV aborts instead of being projected", {
  fits <- lapply(seq_len(4L), .mi2s_mock_fit, p = 4L)
  q <- ncol(fits[[1L]]$Gamma)

  # PSD inputs pool to a PSD Gamma_tilde -> no error.
  expect_silent(invisible(.efa_pooled_mi2s_inputs(fits)))

  # Inject a clearly indefinite asymptotic covariance into one imputation.
  bad <- diag(q)
  bad[1L, 1L] <- -5
  fits[[1L]]$Gamma <- bad
  expect_error(.efa_pooled_mi2s_inputs(fits),
               class = "efa_pooled_mi2s_acov_not_psd")
})


test_that("MI2S matches the lavaan.mi poolSat input-pooling oracle", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  skip_if_not_installed("lavaan.mi")

  m <- 10L
  imps_num <- .mi2s_ord_imps(m = m, n = 400L, p = 6L, base_seed = 1000L)
  vn <- paste0("x", seq_len(6L))
  imps_ord <- lapply(imps_num, function(d) {
    for (j in names(d)) d[[j]] <- ordered(d[[j]])
    d
  })

  pooled <- .mi2s_fit(imps_num, n_factors = 1L, method = "DWLS",
                      rotation = "none", cor_method = "poly")

  # poolSat() pools the inputs the same way (scale.W = FALSE gives exactly the
  # Rubin total ACOV), so a single cfa() on its lavMoments is the matched oracle.
  # lavaan.mi advises attaching itself; that note is irrelevant to the oracle.
  model <- paste("F =~", paste(vn, collapse = " + "))
  lfit <- suppressWarnings(suppressMessages({
    pre <- lavaan.mi::poolSat(imps_ord, ordered = vn, scale.W = FALSE)
    lavaan::cfa(model, data = pre, std.lv = TRUE)
  }))
  pe <- lavaan::parameterEstimates(lfit)
  pe <- pe[pe$op == "=~", ]
  ss <- lavaan::lavInspect(lfit, "test")[["scaled.shifted"]]

  sc <- pre$sample.cov
  if (is.list(sc)) sc <- sc[[1L]]
  rbar <- pooled$orig_R
  expect_equal(rbar[lower.tri(rbar)], sc[lower.tri(sc)], tolerance = 1e-4)

  L_mine <- as.numeric(pooled$unrot_loadings)
  sgn <- sign(sum(L_mine)) * sign(sum(pe$est))
  expect_equal(L_mine, pe$est * sgn, tolerance = 5e-3)
  expect_equal(as.numeric(pooled$SE$unrot_loadings), pe$se, tolerance = 0.02)

  # Native scaled-shifted statistic: df exact, statistic/shift within the
  # polychoric-optimiser gap; lavaan stores the reciprocal of the multiplier.
  expect_equal(pooled$fit_indices$df, ss$df)
  expect_equal(pooled$fit_indices$chi, ss$stat, tolerance = 0.02)
  expect_equal(pooled$fit_indices$chi_shift, ss$shift.parameter, tolerance = 0.02)
  expect_equal(1 / pooled$fit_indices$chi_scaling, ss$scaling.factor,
               tolerance = 0.02)
})


test_that("the pooled fit carries a single native scaled-shifted chi-square", {
  imps <- .mi2s_ord_imps(m = 8L)
  pooled <- .mi2s_fit(imps, n_factors = 1L, method = "DWLS", rotation = "none",
                      cor_method = "poly")
  fi <- pooled$fit_indices

  expect_identical(fi$chi_scaled_type, "scaled.shifted")
  # Scalars from the single fit, not D2 pools of m statistics.
  expect_length(fi$chi_scaling, 1L)
  expect_length(fi$chi_shift, 1L)
  expect_true(is.finite(fi$chi_scaling) && is.finite(fi$chi_shift))
  # Likelihood-ratio information criteria have no meaning on the scaled statistic.
  expect_true(all(is.na(c(fi$AIC, fi$BIC, fi$ECVI))))
})


test_that("the MI2S path invokes no rotation-alignment machinery", {
  imps <- .mi2s_ord_imps(m = 6L)

  called <- new.env(parent = emptyenv())
  called$any <- FALSE
  spy <- function(...) {
    called$any <- TRUE
    NULL
  }
  testthat::local_mocked_bindings(
    efa_procrustes = spy,
    .gpa_consensus_target = spy,
    .align_solution = spy,
    .efa_pooled_rubin_core = spy
  )

  pooled <- .mi2s_fit(imps, n_factors = 1L, method = "ULS", rotation = "oblimin",
                      cor_method = "poly")

  expect_false(called$any)
  expect_s3_class(pooled, "EFA_POOLED")
})


test_that("the pooled object exposes mi_fit alongside the component fits", {
  imps <- .mi2s_ord_imps(m = 5L)
  pooled <- .mi2s_fit(imps, n_factors = 2L, method = "DWLS", rotation = "oblimin",
                      cor_method = "poly")

  expect_s3_class(pooled, "EFA_POOLED")
  expect_s3_class(pooled$mi_fit, "EFA")
  expect_length(pooled$fits, 5L)

  # Pooled point estimates and fit indices come straight from the single fit; the
  # indices are its own values, put into the order every pooled route reports and
  # extended with pool_method (see the cross-route schema test below).
  expect_equal(pooled$orig_R, pooled$mi_fit$orig_R)
  expect_identical(pooled$fit_indices[names(pooled$mi_fit$fit_indices)],
                   pooled$mi_fit$fit_indices)
  expect_identical(pooled$SE, pooled$mi_fit$SE)
  expect_identical(pooled$CI, pooled$mi_fit$CI)

  # Oblique sandwich SE schema flows through to the pooled object.
  expect_setequal(names(pooled$SE),
                  c("unrot_loadings", "uniquenesses", "rot_loadings",
                    "communalities", "Phi", "Structure"))
  # The rotated sandwich SE values must be real numbers, not an all-NA regression
  # (a broken oblique Jacobian or unreliable bordered information would NA them
  # while leaving the schema names intact).
  expect_true(all(is.finite(pooled$SE$rot_loadings)))
  expect_true(all(pooled$SE$rot_loadings > 0))
  expect_true(all(is.finite(pooled$SE$Phi)))
  expect_true(all(is.finite(pooled$SE$Structure)))
  expect_true(all(is.finite(pooled$SE$communalities)))
  # No per-parameter Rubin pooling on this path.
  expect_null(pooled$MI)
  expect_null(pooled$replicates)
  expect_true(isTRUE(pooled$settings$pooled))
  expect_identical(pooled$settings$n_imputations, 5L)
})


test_that("the two-stage pool reports its own fit's Heywood cases", {
  # MI2S returns one fit on the pooled inputs, so that fit's Heywood detector is the
  # pooled solution's. Without it the count falls back to h2 >= 1, which cannot see the
  # boundary uniqueness an ML/ULS solution expresses an improper solution as.
  local_reproducible_output()

  imps <- .mi2s_ord_imps(m = 5L)
  pooled <- .mi2s_fit(imps, n_factors = 1L, method = "ULS", rotation = "none",
                      cor_method = "poly")
  pooled$mi_fit$heywood <- c(x3 = 3L)

  # precondition: the h2 fallback cannot see this case
  expect_lt(max(pooled$h2), 1)
  expect_identical(
    grep("^Heywood cases:", cli::ansi_strip(format(summary(pooled))), value = TRUE),
    "Heywood cases: 1"
  )
})


test_that("the pooled fit indices share one schema and order across routes", {
  # Code that reads names(fit_indices) must not have to branch on the pooling route:
  # the common indices are reported under the same names, in the same order, with the
  # two-stage extras after them rather than interleaved.
  imps <- .mi2s_ord_imps(m = 5L)
  mi2s <- .mi2s_fit(imps, n_factors = 2L, method = "ULS", rotation = "none",
                    cor_method = "poly")
  # The Rubin route assembles the same fit-index list whatever the correlations are, so
  # the comparison pool uses the default Pearson correlations rather than paying for a
  # second round of polychorics.
  rubin <- suppressMessages(suppressWarnings(
    efa_mi(imps, n_factors = 2L, estimator = "ULS", rotation = "none", se = "none")
  ))

  common <- names(rubin$fit_indices)
  expect_identical(names(mi2s$fit_indices)[seq_along(common)], common)

  # pool_method is reported on both routes: on MI2S no rule was applied to the
  # chi-square (the inputs were pooled, not the statistic), which is NA rather than
  # an absent field that would read as "not recorded".
  expect_identical(rubin$fit_indices$pool_method, "D2")
  expect_true(is.na(mi2s$fit_indices$pool_method))

  # The scaled-statistic quantities are genuinely route-specific and stay outside the
  # common block.
  extras <- setdiff(names(mi2s$fit_indices), common)
  expect_true(all(c("Fm", "chi_scaled_type") %in% extras))
})


test_that("estimate_control and rotate_control reach the pooled MI2S fit", {
  # Regression test: the pooled .efa_core() fit must run with the same tuning as
  # the component fits; the control objects used to land unread in .efa_core()'s
  # dots, so the pooled solution was silently computed with the preset defaults.
  quiet <- function(expr) suppressMessages(suppressWarnings(expr))
  # Two IDENTICAL imputations: (x + x) / 2 is exact in IEEE arithmetic and the
  # between-imputation covariance is exactly zero, so the pooled inputs are
  # bit-identical to the single fit's and the single fit is an exact oracle
  # (m = 3+ would round the mean and could flip an optimiser iteration).
  imps <- .mi2s_ord_imps(m = 2L, identical = TRUE)
  # The SPSS rotation preset resolves order_type = "ss_factors" (EFAtools and
  # psych both resolve "eigen" for the oblique engines), so it also pins the
  # rot_type forwarding, which the .efa_core() default would collapse onto the
  # estimation preset.
  rc <- rotate_control(type = "SPSS", normalize = FALSE, random_starts = 0L,
                       gam = 0.5)

  set.seed(1)
  pooled_default <- quiet(
    efa_mi(imps, n_factors = 2L, estimator = "ULS", rotation = "oblimin",
           se = "sandwich", cor_method = "poly")
  )
  set.seed(1)
  pooled_ctl <- quiet(
    efa_mi(imps, n_factors = 2L, estimator = "ULS", rotation = "oblimin",
           se = "sandwich", cor_method = "poly", rotate_control = rc)
  )
  set.seed(1)
  single_ctl <- quiet(
    efa_fit(imps[[1L]], n_factors = 2L, estimator = "ULS", rotation = "oblimin",
            se = "sandwich", cor_method = "poly", rotate_control = rc)
  )

  # The single pooled fit records the requested rotation tuning, not the preset
  # defaults (normalize = TRUE, randomStarts = 100, order_type = "eigen").
  expect_identical(pooled_ctl$mi_fit$settings$normalize, FALSE)
  expect_equal(pooled_ctl$mi_fit$settings$randomStarts, 0)
  expect_identical(pooled_ctl$mi_fit$settings$order_type, "ss_factors")
  expect_identical(pooled_default$mi_fit$settings$normalize, TRUE)
  expect_identical(pooled_default$mi_fit$settings$order_type, "eigen")

  # The compared solutions are proper, and the override visibly changes the
  # pooled solution.
  expect_true(all(is.finite(unclass(pooled_ctl$rot_loadings))))
  expect_false(isTRUE(all.equal(unclass(pooled_ctl$rot_loadings),
                                unclass(pooled_default$rot_loadings),
                                tolerance = 1e-6)))

  # With identical imputations the pooled fit reproduces the single fit run with
  # the same rotate_control, including the oblimin gam engine extra.
  expect_equal(unclass(pooled_ctl$rot_loadings), unclass(single_ctl$rot_loadings),
               tolerance = 1e-10)
  expect_equal(pooled_ctl$SE$rot_loadings, single_ctl$SE$rot_loadings,
               tolerance = 1e-10)

  # A bare engine extra in the dots beats the one stored in rotate_control
  # (gam = 0.7 over rc's 0.5); if the control won, the two runs would be equal.
  set.seed(1)
  pooled_dots <- quiet(
    efa_mi(imps, n_factors = 2L, estimator = "ULS", rotation = "oblimin",
           se = "sandwich", cor_method = "poly", rotate_control = rc, gam = 0.7)
  )
  expect_false(isTRUE(all.equal(unclass(pooled_dots$rot_loadings),
                                unclass(pooled_ctl$rot_loadings),
                                tolerance = 1e-8)))

  # estimate_control side: the ML start method must reach the pooled fit (it is
  # recorded in the single fit's settings; .efa_core()'s own default is "psych").
  imps_cont <- .mi2s_cont_imps(m = 2L)
  pooled_ml <- quiet(
    efa_mi(imps_cont, n_factors = 1L, estimator = "ML", rotation = "none",
           se = "sandwich", cor_method = "pearson",
           estimate_control = estimate_control(start_method = "factanal"))
  )
  expect_identical(pooled_ml$mi_fit$settings$start_method, "factanal")
})


test_that("abbreviated control-argument names reach the pooled MI2S fit", {
  # do.call(efa_fit, ...) accepts unique abbreviations of efa_fit()'s formals
  # through R's argument matching, so the pooled fit must resolve them too
  # instead of silently falling back to the preset controls.
  quiet <- function(expr) suppressMessages(suppressWarnings(expr))
  imps <- .mi2s_ord_imps(m = 2L, identical = TRUE)

  set.seed(1)
  pooled_abbrev <- quiet(
    efa_mi(imps, n_factors = 2L, estimator = "ULS", rotation = "oblimin",
           se = "sandwich", cor_method = "poly",
           rotate_c = rotate_control(normalize = FALSE, random_starts = 0L))
  )

  expect_identical(pooled_abbrev$mi_fit$settings$normalize, FALSE)
  expect_equal(pooled_abbrev$mi_fit$settings$randomStarts, 0)
})


test_that("continuous Pearson imputations run through MI2S (ML and ULS)", {
  imps <- .mi2s_cont_imps(m = 6L)

  for (meth in c("ML", "ULS")) {
    pooled <- .mi2s_fit(imps, n_factors = 1L, method = meth, rotation = "none",
                        cor_method = "pearson")
    expect_s3_class(pooled, "EFA_POOLED")
    expect_true(all(is.finite(pooled$SE$unrot_loadings)), info = meth)
    expect_true(all(pooled$SE$unrot_loadings > 0), info = meth)
    expect_identical(pooled$fit_indices$chi_scaled_type, "scaled.shifted")
    expect_s3_class(pooled$mi_fit, "EFA")
  }
})


# Drive .efa_pooled_mi2s() directly with hand-built fit mocks, supplying a
# matching data_list of correlation matrices (so .efa_pooled_get_Ns falls back to
# NA rather than nrow()) unless raw N is provided some other way.
.mi2s_call <- function(fits, efa_args = list()) {
  .efa_pooled_mi2s(
    fits = fits, data_list = lapply(fits, `[[`, "orig_R"), efa_args = efa_args,
    settings = fits[[1L]]$settings, estimator = "ULS", rotation = "none",
    rotation_type = "none", target_method = "first_target",
    align_unrotated = "signed_tucker_congruence", fit_pool_method = "D2",
    p = 0.05, rmsea_ci_level = 0.90
  )
}

test_that("inconsistent component inputs fail closed", {
  # Mixed se: one sandwich, one information.
  expect_error(
    .mi2s_call(list(.mi2s_mock_fit(1L, se = "sandwich"),
                    .mi2s_mock_fit(2L, se = "information"))),
    class = "efa_pooled_mi2s_inputs_inconsistent"
  )

  # Mixed cor_method across imputations.
  expect_error(
    .mi2s_call(list(.mi2s_mock_fit(3L, cor_method = "poly"),
                    .mi2s_mock_fit(4L, cor_method = "pearson"))),
    class = "efa_pooled_mi2s_inputs_inconsistent"
  )

  # A consistent but unsupported cor_method (no sandwich asymptotic covariance).
  expect_error(
    .mi2s_call(list(.mi2s_mock_fit(5L, cor_method = "spearman"),
                    .mi2s_mock_fit(6L, cor_method = "spearman"))),
    class = "efa_pooled_mi2s_inputs_inconsistent"
  )

  # A missing per-fit asymptotic covariance.
  no_gamma <- list(.mi2s_mock_fit(7L), .mi2s_mock_fit(8L))
  no_gamma[[2L]]$Gamma <- NULL
  expect_error(.mi2s_call(no_gamma),
               class = "efa_pooled_mi2s_inputs_inconsistent")
})


test_that("component fits with mismatched settings fail closed", {
  fits <- list(.mi2s_mock_fit(1L), .mi2s_mock_fit(2L))
  fits[[2L]]$settings$estimator <- "ML"
  expect_error(.efa_pooled_check_fits(fits),
               class = "efa_pooled_setting_mismatch")
})


test_that("MI2S aborts when no imputation supplies N", {
  # Consistent sandwich/poly mocks with valid Gamma but no recoverable N: the
  # data_list holds correlation matrices and efa_args carries no N.
  fits <- lapply(seq_len(3L), .mi2s_mock_fit)
  expect_error(.mi2s_call(fits, efa_args = list()),
               class = "efa_pooled_mi2s_no_n")
})


test_that("MI2S warns on too few imputations and on ignored alignment settings", {
  imps <- .mi2s_ord_imps(m = 5L)

  # Fewer than 20 imputations: the scaled-shifted statistic is under-calibrated.
  expect_warning(
    suppressMessages(EFA_POOLED(imps, n_factors = 1L, method = "ULS",
                                rotation = "none", se = "sandwich",
                                cor_method = "poly")),
    class = "efa_pooled_mi2s_n_too_small"
  )

  # A non-default alignment is inert on the single-fit MI2S path. The m guidance
  # warning also fires here; muffle it so only the alignment note is asserted.
  expect_warning(
    withCallingHandlers(
      suppressMessages(EFA_POOLED(imps, n_factors = 1L, method = "ULS",
                                  rotation = "none", se = "sandwich",
                                  cor_method = "poly",
                                  align_unrotated = "procrustes")),
      efa_pooled_mi2s_n_too_small = function(w) invokeRestart("muffleWarning")
    ),
    class = "efa_pooled_mi2s_alignment_ignored"
  )
})


test_that("tetrachoric (binary) imputations run through MI2S end to end", {
  skip_on_cran()
  # Dichotomise the ordinal generator at the median for a binary fixture.
  imps <- lapply(.mi2s_ord_imps(m = 6L, p = 6L, base_seed = 300L), function(d) {
    as.data.frame(lapply(d, function(col) as.integer(col > stats::median(col))))
  })

  pooled <- .mi2s_fit(imps, n_factors = 1L, method = "DWLS", rotation = "none",
                      cor_method = "tetra")

  expect_s3_class(pooled, "EFA_POOLED")
  expect_true(all(is.finite(pooled$SE$unrot_loadings)))
  expect_true(all(pooled$SE$unrot_loadings > 0))
  expect_identical(pooled$fit_indices$chi_scaled_type, "scaled.shifted")
  expect_identical(pooled$settings$cor_method, "tetra")
})


test_that("MI2S print omits inert pooling settings and labels the CI source", {
  imps <- .mi2s_ord_imps(m = 6L)
  pooled <- .mi2s_fit(imps, n_factors = 1L, method = "DWLS", rotation = "none",
                      cor_method = "poly")

  header <- cli::ansi_strip(format(pooled))
  # The single-fit MI2S path has no per-imputation alignment / chi pooling, so
  # those settings must not appear in the pooled header.
  expect_false(any(grepl("align_unrotated|target_method|fit_pool_method", header)))

  # The full summary carries the CI-provenance note; it must describe the robust
  # sandwich covariance, not the expected information matrix.
  body <- cli::ansi_strip(format(summary(pooled)))
  note <- body[grepl("Note:", body)]
  expect_true(any(grepl("sandwich", note)))
  expect_false(any(grepl("expected information matrix", note)))
})

# Automatic routing of EFA_POOLED()'s standard-error pooling by the SE method of
# its component fits, and the classed conditions that replace the former silent
# `settings$se <- "none"` downgrade.
#
#   - .efa_pooled_route() maps the (identical) component se onto one of
#     "none" / "information" / "sandwich" / "np-boot", or "mixed".
#   - Mixed se aborts (efa_pooled_mixed_se).
#   - A route that cannot produce pooled SEs warns (efa_pooled_se_unavailable)
#     and falls back to point-estimate-only pooling, downgrading settings$se.
#   - The sandwich/MI2S route aborts directly on a structural failure
#     (efa_pooled_mi2s_acov_not_psd), with no se-unavailable fallback.

cormat <- test_models$baseline$cormat
p_vars <- ncol(cormat)
k      <- 3L
N_id   <- 500L

# Fail if the umbrella se-unavailable warning fires; return the expression value.
expect_no_se_unavailable <- function(expr) {
  withCallingHandlers(
    expr,
    efa_pooled_se_unavailable = function(w) {
      testthat::fail("unexpected efa_pooled_se_unavailable warning")
    }
  )
}

# Minimal EFA-shaped fit mock: enough for .efa_pooled_check_fits() and
# .efa_pooled_route() to run (loadings/orig_R dims + names, matching estimator /
# rotation / n_factors settings, and the se under test).
.route_mock_fit <- function(se, p = 4L, kk = 1L) {
  R <- diag(p)
  dimnames(R) <- list(paste0("v", seq_len(p)), paste0("v", seq_len(p)))
  L <- matrix(0.5, p, kk)
  list(unrot_loadings = L, orig_R = R,
       settings = list(se = se, estimator = "ML", rotation = "none",
                       n_factors = kk))
}

# m ordinal imputations from a one-factor model, discretised at the quartiles.
.route_ord_imps <- function(m, n = 300L, p = 6L, base_seed = 700L) {
  lapply(seq_len(m), function(d) {
    set.seed(base_seed + d)
    f <- stats::rnorm(n)
    X <- outer(f, rep(0.6, p)) + matrix(stats::rnorm(n * p, sd = 0.85), n, p)
    out <- as.data.frame(apply(X, 2L, function(col) {
      as.integer(cut(col, c(-Inf, stats::quantile(col, c(.25, .5, .75)), Inf)))
    }))
    names(out) <- paste0("x", seq_len(p))
    out
  })
}

# ---- 1. .efa_pooled_route() -------------------------------------------------

test_that(".efa_pooled_route maps identical component se and flags mixed se", {
  for (s in c("none", "information", "sandwich", "np-boot")) {
    fits <- replicate(3L, .route_mock_fit(s), simplify = FALSE)
    expect_identical(EFAtools:::.efa_pooled_route(fits), s)
  }

  mixed <- list(.route_mock_fit("information"), .route_mock_fit("sandwich"))
  expect_identical(EFAtools:::.efa_pooled_route(mixed), "mixed")

  # A NULL/absent se is treated as non-conformable (mixed), not silently coerced.
  missing_se <- list(.route_mock_fit("information"),
                     list(settings = list(estimator = "ML")))
  expect_identical(EFAtools:::.efa_pooled_route(missing_se), "mixed")
})

# ---- 2. Information happy path ----------------------------------------------

test_that("information route pools SEs without an se-unavailable warning", {
  imps <- replicate(5L, cormat, simplify = FALSE)

  pooled <- expect_no_se_unavailable(suppressMessages(EFA_POOLED(
    imps, n_factors = k, N = N_id, method = "ML", rotation = "none",
    se = "information"
  )))

  expect_identical(pooled$settings$se, "information")
  expect_identical(pooled$settings$component_se, "information")
  expect_false(is.null(pooled$SE$unrot_loadings))
  expect_true(all(is.finite(pooled$SE$unrot_loadings)))
})

# ---- 3. Sandwich (MI2S) happy path ------------------------------------------

test_that("sandwich route fits a single MI2S model exposed as mi_fit", {
  skip_on_cran()
  imps <- .route_ord_imps(m = 5L)

  pooled <- expect_no_se_unavailable(suppressMessages(suppressWarnings(
    EFA_POOLED(imps, n_factors = 1L, method = "DWLS", rotation = "none",
               se = "sandwich", cor_method = "poly")
  )))

  expect_identical(pooled$settings$se, "sandwich")
  expect_identical(pooled$settings$component_se, "sandwich")
  expect_s3_class(pooled$mi_fit, "EFA")
  expect_true(all(is.finite(pooled$SE$unrot_loadings)))

  # The print labels the SE source as the robust sandwich covariance.
  note <- cli::ansi_strip(format(summary(pooled)))
  note <- note[grepl("Note:", note)]
  expect_true(any(grepl("sandwich", note)))
})

# ---- 4. np-boot happy path --------------------------------------------------

test_that("np-boot route pools bootstrap replicates without downgrade", {
  skip_on_cran()
  set.seed(42)
  imps <- lapply(1:3, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), replace = TRUE), ]
  })

  pooled <- expect_no_se_unavailable(suppressMessages(suppressWarnings(
    EFA_POOLED(imps, n_factors = 1L, method = "ML", rotation = "none",
               se = "np-boot", b_boot = 6L)
  )))

  expect_identical(pooled$settings$se, "np-boot")
  expect_false(is.null(pooled$SE$unrot_loadings))
  expect_false(is.null(pooled$replicates))

  # The print labels the SE source as the bootstrap/MI intervals.
  body_np <- cli::ansi_strip(format(summary(pooled)))
  expect_true(any(grepl("Bootstrap", body_np)))
})

# ---- 5. None route ----------------------------------------------------------

test_that("none route pools point estimates with no SE machinery or warning", {
  imps <- replicate(3L, cormat, simplify = FALSE)

  pooled <- expect_no_se_unavailable(
    EFA_POOLED(imps, n_factors = k, N = N_id, method = "PAF", rotation = "none")
  )

  expect_identical(pooled$settings$se, "none")
  expect_false(any(c("SE", "CI", "MI") %in% names(pooled)))
})

# ---- 6. Mixed se aborts -----------------------------------------------------

test_that("mixed component se aborts with efa_pooled_mixed_se", {
  i <- 0L
  fake_efa <- function(x, ...) {
    i <<- i + 1L
    .route_mock_fit(if (i == 1L) "information" else "sandwich",
                    p = p_vars, kk = k)
  }
  testthat::local_mocked_bindings(efa_fit = fake_efa)

  expect_error(
    EFA_POOLED(list(cormat, cormat), n_factors = k, N = N_id, method = "ML",
               rotation = "none"),
    class = "efa_pooled_mixed_se"
  )
})

test_that("the mixed-se abort names the imputations that broke the shared se", {
  # A caller passes one `se` to EFA_POOLED()/efa_mi(), so "re-fit with the same se" is
  # advice they already followed: the mixture comes from efa_fit() recording
  # se = "none" for a fit it could not compute the requested SEs for.
  local_reproducible_output()

  raw <- replicate(3L, matrix(stats::rnorm(40), 10L, 4L), simplify = FALSE)
  fits_boot <- list(.route_mock_fit("np-boot"), .route_mock_fit("none"),
                    .route_mock_fit("np-boot"))

  # A correlation matrix among raw datasets cannot be bootstrapped, which is the
  # reachable trigger; the message says so instead of blaming the `se` argument.
  cor_input <- raw
  cor_input[[2]] <- diag(4L)
  expect_error(
    EFAtools:::.efa_pooled_mixed_se_abort(fits_boot, cor_input),
    class = "efa_pooled_mixed_se"
  )
  expect_snapshot(error = TRUE,
                  EFAtools:::.efa_pooled_mixed_se_abort(fits_boot, cor_input))

  # Any other mixture keeps the generic remediation.
  fits_mixed <- list(.route_mock_fit("information"), .route_mock_fit("sandwich"),
                     .route_mock_fit("information"))
  expect_snapshot(error = TRUE,
                  EFAtools:::.efa_pooled_mixed_se_abort(fits_mixed, raw))
})

# ---- 7. Information route: pooling failure -> se-unavailable + fallback ------

test_that("an analytic-pool abort falls back with efa_pooled_se_unavailable", {
  imps <- replicate(5L, cormat, simplify = FALSE)

  # Simulate an unreliable analytic covariance (Heywood / singular information)
  # surfacing from the pool, regardless of the alignment branch.
  testthat::local_mocked_bindings(
    .efa_pooled_analytic_pool = function(...) {
      cli::cli_abort("unreliable", class = "efa_pooled_unreliable_vcov")
    }
  )

  expect_warning(
    pooled <- suppressMessages(EFA_POOLED(
      imps, n_factors = k, N = N_id, method = "ML", rotation = "none",
      se = "information"
    )),
    class = "efa_pooled_se_unavailable"
  )

  # The point-estimate pooling still completed; only the SEs were dropped.
  expect_false(is.null(pooled$unrot_loadings))
  expect_false(is.null(pooled$residuals))
  expect_false(is.null(pooled$fit_indices))
  expect_false("SE" %in% names(pooled))
  expect_identical(pooled$settings$se, "none")
  expect_identical(pooled$settings$component_se, "information")
})

test_that("the reliability gate aborts on an NA-filled vcov under any alignment", {
  base_fit <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                  rotation = "none", se = "information")
  fits <- replicate(3L, base_fit, simplify = FALSE)
  fits[[2]]$vcov_unrot_loadings[] <- NA_real_

  unrot_list <- lapply(fits, function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "signed_tucker_congruence"
  )

  # The default alignment now aborts too (previously only "procrustes" did), so
  # the dispatcher can fall back cleanly instead of emitting all-NA pooled SEs.
  expect_error(
    EFAtools:::.efa_pooled_analytic_pool(
      fits = fits,
      unrot_loadings_aligned = aligned$loadings,
      align_meta = aligned$meta,
      ci = 0.95,
      align_unrotated = "signed_tucker_congruence"
    ),
    class = "efa_pooled_unreliable_vcov"
  )
})

test_that("the reliability gate signals a classed abort (not a bare error) for >= 2 bad imputations", {
  # The classed abort's message uses cli pluralizers; with two or more flagged
  # imputations a {?...} pluralizer without an explicit cli::qty() count would
  # throw a bare formatting error that escapes the route's tryCatch. Assert the
  # abort is the documented classed condition for both one and several bad fits.
  base_fit <- EFA(cormat, n_factors = k, N = N_id, method = "ML",
                  rotation = "none", se = "information")
  unrot_list <- lapply(replicate(4L, base_fit, simplify = FALSE),
                       function(f) unclass(f$unrot_loadings))
  aligned <- EFAtools:::.efa_pooled_align_unrotated_list(
    unrot_list, align_unrotated = "signed_tucker_congruence"
  )

  for (bad in list(2L, c(2L, 3L), c(1L, 2L, 4L))) {
    fits <- replicate(4L, base_fit, simplify = FALSE)
    for (d in bad) fits[[d]]$vcov_unrot_loadings[] <- NA_real_
    expect_error(
      EFAtools:::.efa_pooled_analytic_pool(
        fits = fits,
        unrot_loadings_aligned = aligned$loadings,
        align_meta = aligned$meta,
        ci = 0.95,
        align_unrotated = "signed_tucker_congruence"
      ),
      class = "efa_pooled_unreliable_vcov"
    )
  }
})

# ---- 8. np-boot route: soft failure -> se-unavailable + fallback ------------

test_that("a NULL bootstrap pool falls back with efa_pooled_se_unavailable", {
  skip_on_cran()
  # Raw data (not a correlation matrix), so the component fits actually carry
  # se = "np-boot" and the np-boot route is taken; EFA() downgrades np-boot to
  # "none" for correlation-matrix input, which would otherwise route to "none".
  set.seed(7)
  imps <- lapply(1:3, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), 150L, replace = TRUE), ]
  })

  testthat::local_mocked_bindings(
    .efa_pooled_bootstrap_pool = function(...) NULL
  )

  expect_warning(
    pooled <- suppressMessages(EFA_POOLED(
      imps, n_factors = 1L, method = "ML", rotation = "none",
      se = "np-boot", b_boot = 6L
    )),
    class = "efa_pooled_se_unavailable"
  )

  expect_identical(pooled$settings$se, "none")
  expect_identical(pooled$settings$component_se, "np-boot")
  expect_false("SE" %in% names(pooled))
  expect_false(is.null(pooled$unrot_loadings))
})

# ---- 9. Sandwich route: structural failure aborts directly ------------------

test_that("a non-PD pooled ACOV aborts directly, with no se-unavailable fallback", {
  skip_on_cran()
  imps <- .route_ord_imps(m = 4L)

  testthat::local_mocked_bindings(
    .efa_pooled_mi2s_inputs = function(...) {
      cli::cli_abort("not psd", class = "efa_pooled_mi2s_acov_not_psd")
    }
  )

  # Propagates as an error (not converted to a warning + fallback).
  expect_error(
    suppressMessages(suppressWarnings(EFA_POOLED(
      imps, n_factors = 1L, method = "DWLS", rotation = "none",
      se = "sandwich", cor_method = "poly"
    ))),
    class = "efa_pooled_mi2s_acov_not_psd"
  )
})

# ---- 10. Print labels the SE source per route -------------------------------

test_that("print labels the SE source for the information and none routes", {
  info <- suppressMessages(EFA_POOLED(
    replicate(5L, cormat, simplify = FALSE), n_factors = k, N = N_id,
    method = "ML", rotation = "none", se = "information"
  ))
  note <- cli::ansi_strip(format(summary(info)))
  note <- note[grepl("Note:", note)]
  expect_true(any(grepl("expected information matrix", note)))

  none <- EFA_POOLED(replicate(3L, cormat, simplify = FALSE), n_factors = k,
                     N = N_id, method = "PAF", rotation = "none")
  body_none <- cli::ansi_strip(format(summary(none)))
  expect_false(any(grepl("information matrix|Bootstrap", body_none)))
})

#' Intelligence subtests from the Intelligence and Development Scales--2
#'
#' A matrix containing the bivariate correlations of the 14 intelligence subtests from the Intelligence and Development Scales--2 (IDS-2; Grob & Hagmann-von Arx, 2018), an intelligence and development test battery for children and adolescents aged 5 to 20 years, for the standardization and validation sample (N = 1,991). Details can be found in Grieder & Grob (2019).
#'
#'
#' @format A 14 x 14 matrix of bivariate correlations
#' \describe{
#'   \item{GS}{(numeric) - Geometric shapes.}
#'   \item{PL}{(numeric) - Plates.}
#'   \item{TC}{(numeric) - Two characteristics.}
#'   \item{CB}{(numeric) - Crossing out boxes.}
#'   \item{NL}{(numeric) - Numbers / letters.}
#'   \item{NLM}{(numeric) - Numbers / letter mixed.}
#'   \item{GF}{(numeric) - Geometric figures.}
#'   \item{RGF}{(numeric) - Rotated geometric figures.}
#'   \item{CM}{(numeric) - Completing matrices.}
#'   \item{EP}{(numeric) - Excluding pictures.}
#'   \item{CA}{(numeric) - Categories.}
#'   \item{OP}{(numeric) - Opposites.}
#'   \item{RS}{(numeric) - Retelling a story.}
#'   \item{DP}{(numeric) - Describing pictures.}
#'  }
#' @source Grieder, S., & Grob, A. (2019). Exploratory factor analyses of the intelligence and development scales--2: Implications for theory and practice. Assessment. Advance online publication. doi:10.1177/1073191119845051
#' @source Grob, A., & Hagmann-von Arx, P. (2018). Intelligence and Development Scales--2 (IDS-2). Intelligenz- und Entwicklungsskalen für Kinder und Jugendliche.
#' \[Intelligence and Development Scales for Children and Adolescents.\]. Bern, Switzerland: Hogrefe.
"IDS2_R"

#' Print and format an efa_kmo object
#'
#' `print()` shows the Kaiser-Meyer-Olkin (KMO) criterion computed by [efa_kmo()]: a
#' titled section with a verdict on the overall KMO value (and what it implies for
#' the suitability of the data for factor analysis), the overall value, and the
#' per-variable KMO values. `format()` assembles the same report and returns it as
#' a character vector; `print()` is `cat(format(x), sep = "\n")`. The lines follow
#' the active console theme, so they are plain when colours are disabled (for
#' example when captured into a file or stripped with [cli::ansi_strip()]).
#'
#' @param x An object of class `efa_kmo` (output from [efa_kmo()]).
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @export
#'
#' @method print efa_kmo
#'
#' @examples
#' KMO_base <- efa_kmo(test_models$baseline$cormat)
#' KMO_base
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(KMO_base))
#'
print.efa_kmo <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_kmo
#' @export
#' @method format efa_kmo
format.efa_kmo <- function(x, ...) {
  KMO <- x$KMO

  cli::cli_format_method({
    cli::cli_text("")
    cli::cli_rule(left = "Kaiser-Meyer-Olkin criterion (KMO)")
    cli::cli_text("")

    if (!is.null(KMO) && !is.na(KMO)) {

      # Kaiser's verbal bands for the KMO value (Kaiser, 1970; Kaiser & Rice, 1974). The band
      # carries its label, its verdict styling (green = suitable, yellow = mediocre, red =
      # unsuitable) and the suitability it implies; .kmo_band() holds the cut-offs, so this
      # report and efa_screen()'s cannot band the same value differently.
      band <- .kmo_band(KMO)

      label <- band$colour(cli::style_bold(band$label))
      band$alert("The overall KMO value for your data is {label}.", wrap = TRUE)
      cli::cli_text("These data are {band$suitability} suitable for factor analysis.")

      cli::cli_text("")
      overall <- cli::style_bold(as.character(round(KMO, 3)))
      cli::cli_text("Overall: {overall}")

      cli::cli_text("")
      cli::cli_text("For each variable:")
      # The per-variable values are kept as base R's named-vector layout, emitted verbatim so
      # cli does not reflow the aligned columns.
      cli::cli_verbatim(utils::capture.output(print(round(x$KMO_i, 3))))

    } else {
      cli::cli_alert_warning("Sorry, the KMO value for your data is not available.",
                             wrap = TRUE)
    }
  })
}

# Helpers for averaging many EFA solutions: build the per-rotation parameter grid, extract
# and align factors across runs by congruence, and aggregate loadings, communalities,
# variances, and fit indices.

### extract data from efa_list
.extract_data <- function(efa_list, R, n_factors, n_efa, rotation, salience_threshold) {


  L <- array(NA_real_, c(ncol(R), n_factors, n_efa))
  L_corres <- array(NA, c(ncol(R), n_factors, n_efa))
  h2 <- matrix(NA_real_, nrow = n_efa, ncol = ncol(R))
  if (n_factors > 1) {
    vars_accounted <- array(NA_real_, c(3, n_factors, n_efa))
  } else {
    vars_accounted <- array(NA_real_, c(2, n_factors, n_efa))
  }



  if (any(rotation %in% c(.oblq_rotations, "oblique"))) {
    extract_phi <- TRUE
    phi <- array(NA_real_, c(n_factors, n_factors, n_efa))
  } else {
    extract_phi <- FALSE
    phi <- NA
  }

  converged <- rep(NA, n_efa)
  errors <- rep(FALSE, n_efa)
  error_m <- rep(NA_character_, n_efa)
  heywood <- rep(NA, n_efa)
  admissible <- rep(NA, n_efa)
  aic <- rep(NA_real_, n_efa)
  bic <- rep(NA_real_, n_efa)
  chisq <- rep(NA_real_, n_efa)
  p_chi <- rep(NA_real_, n_efa)
  caf <- rep(NA_real_, n_efa)
  rmsea <- rep(NA_real_, n_efa)
  cfi <- rep(NA_real_, n_efa)
  srmr <- rep(NA_real_, n_efa)
  tli <- rep(NA_real_, n_efa)
  ecvi <- rep(NA_real_, n_efa)
  rmsr <- rep(NA_real_, n_efa)

  if (all(rotation == "none") || n_factors == 1) {
    load_ind <- "unrot_loadings"
    var_ind <- "vars_accounted"
  } else {
    load_ind <- "rot_loadings"
    var_ind <- "vars_accounted_rot"
  }

    for (row_i in seq_len(n_efa)) {

      efa_temp <- efa_list[[row_i]]

      if (inherits(efa_temp, "try-error")) {

        errors[row_i] <- TRUE
        error_m[row_i] <- efa_temp[[1]]

      } else {
        converged[row_i] <- efa_temp$convergence

        if (efa_temp$convergence == 0) {
          # Use the fit's Heywood flag so improper solutions are excluded
          # consistently across estimators (a communality >= 1, or an ML/ULS
          # uniqueness pinned at the estimation boundary).
          has_heywood <- length(efa_temp$heywood) > 0
          heywood[row_i] <- has_heywood

          if (!has_heywood) {

            aic[row_i] <- efa_temp$fit_indices$AIC
            bic[row_i] <- efa_temp$fit_indices$BIC
            chisq[row_i] <- efa_temp$fit_indices$chi
            p_chi[row_i] <- efa_temp$fit_indices$p_chi
            caf[row_i] <- efa_temp$fit_indices$CAF
            rmsea[row_i] <- efa_temp$fit_indices$RMSEA
            cfi[row_i] <- efa_temp$fit_indices$CFI
            srmr[row_i] <- efa_temp$fit_indices$SRMR
            tli[row_i] <- efa_temp$fit_indices$TLI
            ecvi[row_i] <- efa_temp$fit_indices$ECVI
            rmsr[row_i] <- efa_temp$fit_indices$RMSR

            h2[row_i, ] <- efa_temp$h2
            L[,, row_i] <- efa_temp[[load_ind]]
            if (n_factors > 1) {
              vars_accounted[,, row_i] <- efa_temp[[var_ind]][c(1, 2, 4),]
            } else {
              vars_accounted[,, row_i] <- efa_temp[[var_ind]]
            }

            temp_corres <- abs(efa_temp[[load_ind]]) >= salience_threshold
            L_corres[,, row_i] <-temp_corres

            # Admissible: no error, converged, no Heywood case, and at least two salient
            # loadings on every factor. The first three hold on this branch, so only the
            # salience count is left to check. The check lives here, where temp_corres was
            # just assigned, rather than relying on `||` short-circuiting to keep it from
            # reading the previous solution's correspondences.
            admissible[row_i] <- all(colSums(temp_corres) >= 2)

          } else {

            admissible[row_i] <- FALSE

          }


          if (isTRUE(extract_phi)) {
            phi[,, row_i] <- efa_temp$Phi
          }
        }
      }
    }

  # remove data from nonconverged EFAs
  excl <- which(converged != 0 | errors | heywood)
  if (length(excl) > 0) {
    L <- L[,, -excl, drop = FALSE]
    L_corres <- L_corres[,, -excl, drop = FALSE]
    vars_accounted <- vars_accounted[,, -excl, drop = FALSE]
    if (isTRUE(extract_phi)) {
      phi <- phi[,, -excl, drop = FALSE]
    }
  }

  out <- list(
    L = L,
    L_corres = L_corres,
    phi = phi,
    extract_phi = extract_phi,
    h2 = h2,
    vars_accounted = vars_accounted,
    for_grid = data.frame(
      errors = errors,
      error_m = error_m,
      converged = converged,
      heywood = heywood,
      admissible = admissible,
      chisq = chisq,
      p_chi = p_chi,
      caf = caf,
      cfi = cfi,
      rmsea = rmsea,
      aic = aic,
      bic= bic,
      srmr = srmr,
      tli = tli,
      ecvi = ecvi,
      rmsr = rmsr
    )
  )

  return(out)

}

# Range endpoints of a summary that may have no value at all. `min()`/`max()` with
# `na.rm = TRUE` warn and return -Inf/Inf for an entirely missing input, which happens
# routinely: a PAF-only grid carries no chi-square-based fit index, so those columns are
# missing for every solution. A summary with nothing to summarise is NA, not an infinite
# bound, so the endpoints are produced that way here rather than repaired afterwards.
.range_min <- function(x) if (all(is.na(x))) NA_real_ else min(x, na.rm = TRUE)
.range_max <- function(x) if (all(is.na(x))) NA_real_ else max(x, na.rm = TRUE)

### average arrays
.average_values <- function(vars_accounted, L, L_corres, h2, phi, extract_phi,
                              averaging, trim, for_grid, df, ind_names) {

  if (averaging == "mean") {

    if (trim == 0) {
      # faster, but only works without trimming
      L_av <- rowMeans(L, na.rm = TRUE, dims = 2)
      h2_av <- colMeans(h2, na.rm = TRUE)
      fit_av <- colMeans(for_grid, na.rm = TRUE)
      vars_accounted_av <- rowMeans(vars_accounted, na.rm = TRUE, dims = 2)


      if (isTRUE(extract_phi)) {
        phi_av <- rowMeans(phi, na.rm = TRUE, dims = 2)
      }
    } else {
      L_av <- apply(L, 1:2, mean, na.rm = TRUE, trim = trim)
      h2_av <- apply(h2, 2, mean, na.rm = TRUE, trim = trim)
      fit_av <- apply(for_grid, 2, mean, na.rm = TRUE, trim = trim)
      vars_accounted_av <- apply(vars_accounted, 1:2, mean, na.rm = TRUE,
                                  trim = trim)

      if (isTRUE(extract_phi)) {
        phi_av <- apply(phi, 1:2, mean, na.rm = TRUE, trim = trim)
      }
    }

  } else if (averaging == "median") {
    L_av <- apply(L, 1:2, stats::median, na.rm = TRUE)
    h2_av <- apply(h2, 2, stats::median, na.rm = TRUE)
    fit_av <- apply(for_grid, 2, stats::median, na.rm = TRUE)
    vars_accounted_av <- apply(vars_accounted, 1:2, stats::median, na.rm = TRUE)
    if (isTRUE(extract_phi)) {
      phi_av <- apply(phi, 1:2, stats::median, na.rm = TRUE)
    }
  }


  nf <- ncol(L_av)
  f_names <- paste0("F", 1:nf)

  L_corres_av <- rowMeans(L_corres, na.rm = TRUE, dims = 2)
  row.names(L_corres_av) <- ind_names
  colnames(L_corres_av) <- f_names

  L_min <- apply(L, 1:2, .range_min)
  L_max <- apply(L, 1:2, .range_max)
  L_range <- L_max - L_min
  L_sd <- apply(L, 1:2, stats::sd, na.rm = TRUE)
  rownames(L_av) <- ind_names
  colnames(L_av) <- f_names
  class(L_av) <- c("efa_loadings", "LOADINGS")
  rownames(L_min) <- ind_names
  colnames(L_min) <- f_names
  class(L_min) <- c("efa_loadings", "LOADINGS")
  rownames(L_max) <- ind_names
  colnames(L_max) <- f_names
  class(L_max) <- c("efa_loadings", "LOADINGS")
  rownames(L_range) <- ind_names
  colnames(L_range) <- f_names
  rownames(L_sd) <- ind_names
  colnames(L_sd) <- f_names



  vars_accounted_min <- apply(vars_accounted, 1:2, .range_min)
  vars_accounted_max <- apply(vars_accounted, 1:2, .range_max)
  vars_accounted_range <- vars_accounted_max - vars_accounted_min
  vars_accounted_sd <- apply(vars_accounted, 1:2, stats::sd, na.rm = TRUE)


  if (nrow(vars_accounted_av) == 2) {
    var_names <- c("SS loadings", "Prop Tot Var")
  } else {
    var_names <- c("SS loadings", "Prop Tot Var", "Prop Comm Var")
  }
  rownames(vars_accounted_av) <- var_names
  colnames(vars_accounted_av) <- f_names
  rownames(vars_accounted_min) <- var_names
  colnames(vars_accounted_min) <- f_names
  rownames(vars_accounted_max) <- var_names
  colnames(vars_accounted_max) <- f_names
  rownames(vars_accounted_range) <- var_names
  colnames(vars_accounted_range) <- f_names
  rownames(vars_accounted_sd) <- var_names
  colnames(vars_accounted_sd) <- f_names


  h2_min <- apply(h2, 2, .range_min)
  h2_max <- apply(h2, 2, .range_max)
  h2_range <- h2_max - h2_min
  h2_sd <- apply(h2, 2, stats::sd, na.rm = TRUE)
  names(h2_av) <- ind_names
  names(h2_min) <- ind_names
  names(h2_max) <- ind_names
  names(h2_range) <- ind_names
  names(h2_sd) <- ind_names


  fit_min <- apply(for_grid, 2, .range_min)
  fit_max <- apply(for_grid, 2, .range_max)
  fit_range <- fit_max - fit_min
  fit_sd <- apply(for_grid, 2, stats::sd, na.rm = TRUE)

  # colMeans()/mean() of an entirely missing column is NaN; the endpoints are already
  # NA, and the range inherits it.
  fit_av[is.nan(fit_av)] <- NA

  # df is the same for every averaged solution (fixed m and n_factors), so its
  # dispersion is zero: sd and range of a constant are 0, while average/min/max
  # equal df itself.
  fit_indices <- data.frame(
    index = c(names(fit_av), "df"),
    average = c(fit_av, df),
    sd = c(fit_sd, 0),
    range = c(fit_range, 0),
    min = c(fit_min, df),
    max = c(fit_max, df),
    stringsAsFactors = FALSE
  )

  if (isTRUE(extract_phi)) {
    phi_min <- apply(phi, 1:2, .range_min)
    phi_max <- apply(phi, 1:2, .range_max)
    phi_range <- phi_max - phi_min
    phi_sd <- apply(phi, 1:2, stats::sd, na.rm = TRUE)
    # All five summaries are factor x factor, so they take the same dimnames.
    phi_list <- lapply(
      list(average = phi_av, sd = phi_sd, min = phi_min, max = phi_max,
           range = phi_range),
      function(p) {
        dimnames(p) <- list(f_names, f_names)
        p
      })
  } else {
    phi_list <- NA
  }

  out <- list(
    h2 = list(
      average = h2_av,
      sd = h2_sd,
      min = h2_min,
      max = h2_max,
      range = h2_range
    ),
    loadings = list(
      average = L_av,
      sd = L_sd,
      min = L_min,
      max = L_max,
      range = L_range
    ),
    phi = phi_list,
    vars_accounted = list(
      average = vars_accounted_av,
      sd = vars_accounted_sd,
      min = vars_accounted_min,
      max = vars_accounted_max,
      range = vars_accounted_range
    ),
    ind_fac_corres = L_corres_av,
    fit_indices = fit_indices)

  return(out)
}


### reorder arrays according to factor congruence
.array_reorder <- function(vars_accounted, L, L_corres, phi, extract_phi, n_factors) {

  if (dim(L)[3] > 1) {
  	L1 <- L[,, 1]
    for (efa_i in 2:dim(L)[3]) {

      # Match each later solution's factors to the first via the optimal
      # (linear-sum-assignment) congruence alignment, which guarantees a true
      # permutation and applies sign flips consistently to loadings and Phi.
      aligned <- .align_solution(
        L_target = L1, L = L[,, efa_i],
        Phi = if (isTRUE(extract_phi)) phi[,, efa_i] else NULL)

      factor_order <- aligned$factor_order

      # The correspondence indicators and variances are sign-invariant, so they
      # only need the same column reordering.
      L[,, efa_i] <- aligned$loadings
      L_corres[,, efa_i] <- L_corres[,, efa_i][, factor_order]
      vars_accounted[,, efa_i] <- vars_accounted[,, efa_i][, factor_order]

      if (isTRUE(extract_phi)) {
        phi[,, efa_i] <- aligned$Phi
      }

    }
  }


  return(list(L=L, L_corres = L_corres, phi = phi, vars_accounted = vars_accounted))

}

### create grid for oblique rotations in efa_average()
.oblq_grid <- function(estimator, init_comm, criterion, criterion_type,
                       abs_eigen, max_iter, start_method, rotation, k_promax, normalize,
                       P_type, precision, varimax_type, k_simplimax){

  g_list <- list()

  if ("promax" %in% rotation) {

    g_list[["prmx"]] <- expand.grid(estimator = estimator, init_comm = init_comm,
                                    criterion = criterion, criterion_type = criterion_type,
                                    abs_eigen = abs_eigen, max_iter = max_iter,
                                    start_method = start_method,
                                    rotation = "promax",
                                    k_promax = k_promax, normalize = normalize, P_type = P_type,
                                    precision = precision, varimax_type = varimax_type,
                                    k_simplimax = NA, stringsAsFactors = FALSE)

  }

  if ("simplimax" %in% rotation) {

    g_list[["smplmx"]] <- expand.grid(estimator = estimator, init_comm = init_comm,
                                      criterion = criterion, criterion_type = criterion_type,
                                      abs_eigen = abs_eigen, max_iter = max_iter,
                                      start_method = start_method,
                                      rotation = "simplimax",
                                      k_promax = NA, normalize = normalize, P_type = NA,
                                      precision = precision, varimax_type = NA,
                                      k_simplimax = k_simplimax, stringsAsFactors = FALSE)

  }

  rotation_temp <- rotation[!(rotation %in% c("promax", "simplimax"))]

  if (length(rotation_temp) > 0) {
    g_list[["oblq"]] <- expand.grid(estimator = estimator, init_comm = init_comm,
                                    criterion = criterion, criterion_type = criterion_type,
                                    abs_eigen = abs_eigen, max_iter = max_iter,
                                    start_method = start_method,
                                    rotation = rotation_temp,
                                    k_promax = NA, normalize = normalize, P_type = NA,
                                    precision = precision, varimax_type = NA,
                                    k_simplimax = NA, stringsAsFactors = FALSE)
  }

  return(do.call(rbind, g_list))

}

### create grid for orthogonal rotations in efa_average()
.orth_grid <- function(estimator, init_comm, criterion, criterion_type,
                       abs_eigen, max_iter, start_method, rotation, normalize,
                       precision, varimax_type){

  g_list <- list()

  if ("varimax" %in% rotation) {

    g_list[["vrmx"]] <- expand.grid(estimator = estimator, init_comm = init_comm,
                                    criterion = criterion, criterion_type = criterion_type,
                                    abs_eigen = abs_eigen, max_iter = max_iter,
                                    start_method = start_method,
                                    rotation = "varimax",
                                    k_promax = NA, normalize = normalize, P_type = NA,
                                    precision = precision, varimax_type = varimax_type,
                                    k_simplimax = NA, stringsAsFactors = FALSE)

  }

  rotation_temp <- rotation[!(rotation %in% c("varimax"))]

  if (length(rotation_temp) > 0) {
    g_list[["orth"]] <- expand.grid(estimator = estimator, init_comm = init_comm,
                                    criterion = criterion, criterion_type = criterion_type,
                                    abs_eigen = abs_eigen, max_iter = max_iter,
                                    start_method = start_method,
                                    rotation = rotation_temp,
                                    k_promax = NA, normalize = normalize, P_type = NA,
                                    precision = precision, varimax_type = NA,
                                    k_simplimax = NA, stringsAsFactors = FALSE)
  }

  return(do.call(rbind, g_list))

}

.type_grid <- function(estimator, init_comm, criterion, criterion_type,
                       abs_eigen, max_iter, start_method, rotation, k_promax, normalize,
                       P_type, precision, varimax_type, k_simplimax) {

  t_grid_list <- list()
  if ("none" %in% rotation) {
    if (length(rotation) == 1) {

      t_grid_list[["nn"]] <- expand.grid(estimator = estimator, init_comm = init_comm,
                                         criterion = criterion, criterion_type = criterion_type,
                                         abs_eigen = abs_eigen, max_iter = max_iter,
                                         start_method = start_method,
                                         rotation = "none",
                                         k_promax = NA, normalize = NA, P_type = NA,
                                         precision = NA, varimax_type = NA,
                                         k_simplimax = NA, stringsAsFactors = FALSE)

    } else {

      cli::cli_abort(
        c("{.code rotation = \"none\"} was used but {.arg rotation} has length > 1.",
          "i" = "Can only average EFAs with rotations of the same type: {.val none}, {.val orthogonal}, or {.val oblique}."),
        class = "efa_rotation_length"
      )

    }
  } else if ("oblique" %in% rotation) {
    if (length(rotation) == 1) {

      t_grid_list[["blq"]] <- .oblq_grid(estimator = estimator, init_comm = init_comm,
                                         criterion = criterion, criterion_type = criterion_type,
                                         abs_eigen = abs_eigen, max_iter = max_iter,
                                         start_method = start_method,
                                         rotation = .oblq_rotations,
                                         k_promax = k_promax, normalize = normalize,
                                         P_type = P_type, precision = precision,
                                         varimax_type = varimax_type, k_simplimax = k_simplimax)

    } else {

      cli::cli_abort(
        c("{.code rotation = \"oblique\"} was used but {.arg rotation} has length > 1.",
          "i" = "Can only average EFAs with rotations of the same type: {.val none}, {.val orthogonal}, or {.val oblique}."),
        class = "efa_rotation_length"
      )

    }

  } else if ("orthogonal" %in% rotation) {

    if (length(rotation) == 1) {

      t_grid_list[["rth"]] <- .orth_grid(estimator = estimator, init_comm = init_comm,
                                         criterion = criterion, criterion_type = criterion_type,
                                         abs_eigen = abs_eigen, max_iter = max_iter,
                                         start_method = start_method,
                                         rotation = .orth_rotations,
                                         normalize = normalize, precision = precision,
                                         varimax_type = varimax_type)

    } else {

      cli::cli_abort(
        c("{.code rotation = \"orthogonal\"} was used but {.arg rotation} has length > 1.",
          "i" = "Can only average EFAs with rotations of the same type: {.val none}, {.val orthogonal}, or {.val oblique}."),
        class = "efa_rotation_length"
      )

    }

  } else if (all(rotation %in% .oblq_rotations)) {

    t_grid_list[["blq2"]] <- .oblq_grid(estimator = estimator, init_comm = init_comm,
                                        criterion = criterion, criterion_type = criterion_type,
                                        abs_eigen = abs_eigen, max_iter = max_iter,
                                        start_method = start_method,
                                        rotation = rotation,
                                        k_promax = k_promax, normalize = normalize,
                                        P_type = P_type, precision = precision,
                                        varimax_type = varimax_type, k_simplimax = k_simplimax)

  } else if (all(rotation %in% .orth_rotations)) {

    t_grid_list[["rth2"]] <- .orth_grid(estimator = estimator, init_comm = init_comm,
                                        criterion = criterion, criterion_type = criterion_type,
                                        abs_eigen = abs_eigen, max_iter = max_iter,
                                        start_method = start_method,
                                        rotation = rotation,
                                        normalize = normalize, precision = precision,
                                        varimax_type = varimax_type)

  } else if (any(rotation %in% .oblq_rotations) &&
             any(rotation %in% .orth_rotations)) {
    cli::cli_abort(
      c("{.arg rotation} mixes oblique and orthogonal rotations, but only rotations of the same kind can be averaged.",
        "*" = "Oblique rotations: {.val {(.oblq_rotations)}}.",
        "*" = "Orthogonal rotations: {.val {(.orth_rotations)}}."),
      class = "efa_rotation_mismatch"
    )
  }

  return(do.call(rbind, t_grid_list))
}

### assemble the full implementation grid for efa_average()
# Stack the per-(estimator, type) parameter grids into one grid of EFA() argument
# combinations. The three named types (EFAtools/psych/SPSS) take their tuning
# arguments from the shared `.efa_presets` table (the same source EFA() resolves),
# so an averaged sub-analysis matches EFA(type = <that type>); type "none" forwards
# the user-supplied arguments. Each estimator keeps only the arguments it uses
# (PAF: starting communalities and convergence criterion; ML: starting values),
# mirroring EFA() itself. Estimators are processed in the order PAF, ML, ULS and types
# in the order EFAtools, psych, SPSS, none, so the row order and names of the
# assembled grid are stable.
.build_avg_grid <- function(estimator, type, rotation, init_comm, criterion,
                            criterion_type, abs_eigen, max_iter, start_method,
                            k_promax, normalize, P_type, precision, varimax_type,
                            k_simplimax) {

  # Tags used to build the grid-list names (e.g. "ftls_pf", "nn_ml"), kept
  # identical to the names the grid is assembled under so row names stay stable.
  estimator_tag <- c(PAF = "pf", ML = "ml", ULS = "ls")
  type_tag   <- c(EFAtools = "ftls", psych = "psch", SPSS = "spss", none = "nn")

  # The communality/start arguments each estimator uses; the rest stay NA. The
  # iteration cap (`max_iter`) is one of the PAF arguments, since only the
  # principal-axis iterations consume it, so it is NA for ML and ULS just as the
  # other PAF settings are. On a PAF row it comes from the `.efa_presets` PAF
  # block for a named type (SPSS 25, psych 50, EFAtools 300) and from the user
  # for type "none": the cap is part of what distinguishes the implementations,
  # so two PAF rows differing only in it are genuinely different models. Leaving
  # it NA elsewhere keeps an ML or ULS row that a named type and type "none"
  # resolve to identically a single row, rather than the same fit repeated once
  # per type.
  estimator_args <- function(m, from_user, paf = NULL) {
    switch(m,
      PAF = if (from_user) {
        list(init_comm = init_comm, criterion = criterion,
             criterion_type = criterion_type, abs_eigen = abs_eigen,
             max_iter = max_iter, start_method = NA)
      } else {
        list(init_comm = paf$init_comm, criterion = paf$criterion,
             criterion_type = paf$criterion_type,
             abs_eigen = paf$abs_eigen, max_iter = paf$max_iter,
             start_method = NA)
      },
      ML  = list(init_comm = NA, criterion = NA, criterion_type = NA,
                 abs_eigen = NA, max_iter = NA,
                 start_method = if (from_user) start_method else "psych"),
      ULS = list(init_comm = NA, criterion = NA, criterion_type = NA,
                 abs_eigen = NA, max_iter = NA, start_method = NA))
  }

  grid_list <- list()

  for (m in c("PAF", "ML", "ULS")) {
    if (!(m %in% estimator)) next

    for (t in c("EFAtools", "psych", "SPSS", "none")) {
      if (!(t %in% type)) next

      if (t == "none") {
        ma   <- estimator_args(m, from_user = TRUE)
        rest <- list(k_promax = k_promax, normalize = normalize, P_type = P_type,
                     precision = precision, varimax_type = varimax_type)
      } else {
        # Resolve named-type tuning arguments through .resolve_settings() -- the
        # same resolver EFA() uses -- with EFA()'s default user arguments
        # (normalize on, everything else unset), so an averaged sub-analysis
        # matches EFA(type = t) in its estimation and rotation tuning: PAF's
        # convergence settings and iteration cap (max_iter) come from the PAF
        # block, the promax/varimax settings from the PROMAX block. `precision`
        # is not a preset, so it keeps EFA()'s default. The one deliberate
        # exception is order_type, hard-coded "eigen" when the grid is run: the
        # averaging re-aligns every solution to the first by congruence, so a
        # per-fit factor order (SPSS's "ss_factors") would never survive into the
        # averaged result anyway.
        paf <- .resolve_settings(
          type = t,
          user = list(init_comm = NA, criterion = NA, criterion_type = NA,
                      max_iter = NA, abs_eigen = NA),
          preset = .efa_presets$PAF)
        pro <- .resolve_settings(
          type = t,
          user = list(normalize = TRUE, P_type = NA, order_type = NA,
                      varimax_type = NA, k = NA),
          preset = .efa_presets$PROMAX)
        ma   <- estimator_args(m, from_user = FALSE, paf = paf)
        rest <- list(k_promax = pro$k, normalize = pro$normalize,
                     P_type = pro$P_type, precision = 1e-5,
                     varimax_type = pro$varimax_type)
      }

      grid_list[[paste0(type_tag[[t]], "_", estimator_tag[[m]])]] <- .type_grid(
        estimator = m, init_comm = ma$init_comm, criterion = ma$criterion,
        criterion_type = ma$criterion_type, abs_eigen = ma$abs_eigen,
        max_iter = ma$max_iter, start_method = ma$start_method, rotation = rotation,
        k_promax = rest$k_promax, normalize = rest$normalize,
        P_type = rest$P_type, precision = rest$precision,
        varimax_type = rest$varimax_type, k_simplimax = k_simplimax)
    }
  }

  unique(do.call(rbind, grid_list))
}

# Guards .build_avg_grid() against the previous explicit, per-(estimator, type)
# construction of the efa_average implementation grid. `ref_build_avg_grid()` is
# the earlier construction kept verbatim; the test asserts the current builder
# produces a byte-identical grid (values, column types, and row names) for every
# combination of estimators, types, and rotations, including vector-valued
# user arguments. Both implementations call the same .type_grid() helper, so any
# difference can only come from the arguments assembled per cell.

ref_build_avg_grid <- function(estimator, type, rotation, init_comm, criterion,
                               criterion_type, abs_eigen, max_iter, start_method,
                               k_promax, normalize, P_type, precision, varimax_type,
                               k_simplimax) {

  grid_list <- list()

  if ("PAF" %in% estimator) {

    if ("EFAtools" %in% type) {
      grid_list[["ftls_pf"]] <- .type_grid(estimator = "PAF", init_comm = "smc",
                                           criterion = 1e-3, criterion_type = "sum",
                                           abs_eigen = TRUE, max_iter = 300,
                                           start_method = NA,
                                           rotation = rotation, k_promax = 4,
                                           normalize = TRUE, P_type = "norm",
                                           precision = 1e-5,
                                           varimax_type = "kaiser",
                                           k_simplimax = k_simplimax)
    }

    if ("psych" %in% type) {
      grid_list[["psch_pf"]] <- .type_grid(estimator = "PAF", init_comm = "smc",
                                           criterion = 1e-3, criterion_type = "sum",
                                           abs_eigen = FALSE, max_iter = 50,
                                           start_method = NA,
                                           rotation = rotation, k_promax = 4,
                                           normalize = TRUE, P_type = "unnorm",
                                           precision = 1e-5,
                                           varimax_type = "svd",
                                           k_simplimax = k_simplimax)
    }

    if ("SPSS" %in% type) {
      grid_list[["spss_pf"]] <- .type_grid(estimator = "PAF", init_comm = "smc",
                                           criterion = 1e-3, criterion_type = "max_individual",
                                           abs_eigen = TRUE, max_iter = 25,
                                           start_method = NA,
                                           rotation = rotation, k_promax = 4,
                                           normalize = TRUE, P_type = "norm",
                                           precision = 1e-5,
                                           varimax_type = "kaiser",
                                           k_simplimax = k_simplimax)
    }

    if ("none" %in% type) {

        grid_list[["nn_pf"]] <- .type_grid(estimator = "PAF", init_comm = init_comm,
                                           criterion = criterion, criterion_type = criterion_type,
                                           abs_eigen = abs_eigen, max_iter = max_iter,
                                           start_method = NA,
                                           rotation = rotation, k_promax = k_promax,
                                           normalize = normalize, P_type = P_type,
                                           precision = precision,
                                           varimax_type = varimax_type,
                                           k_simplimax = k_simplimax)
    }
  }

    if ("ML" %in% estimator) {

      if ("EFAtools" %in% type) {
        grid_list[["ftls_ml"]] <- .type_grid(estimator = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, max_iter = NA,
                                             start_method = "psych",
                                             rotation = rotation, k_promax = 4,
                                             normalize = TRUE, P_type = "norm",
                                             precision = 1e-5,
                                             varimax_type = "kaiser",
                                             k_simplimax = k_simplimax)
      }

      if ("psych" %in% type) {
        grid_list[["psch_ml"]] <- .type_grid(estimator = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, max_iter = NA,
                                             start_method = "psych",
                                             rotation = rotation, k_promax = 4,
                                             normalize = TRUE, P_type = "unnorm",
                                             precision = 1e-5,
                                             varimax_type = "svd",
                                             k_simplimax = k_simplimax)
      }

      if ("SPSS" %in% type) {
        grid_list[["spss_ml"]] <- .type_grid(estimator = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, max_iter = NA,
                                             start_method = "psych",
                                             rotation = rotation, k_promax = 4,
                                             normalize = TRUE, P_type = "norm",
                                             precision = 1e-5,
                                             varimax_type = "kaiser",
                                             k_simplimax = k_simplimax)
      }

      if ("none" %in% type) {
          grid_list[["nn_ml"]] <- .type_grid(estimator = "ML", init_comm = NA,
                                             criterion = NA, criterion_type = NA,
                                             abs_eigen = NA, max_iter = NA,
                                             start_method = start_method,
                                             rotation = rotation, k_promax = k_promax,
                                             normalize = normalize, P_type = P_type,
                                             precision = precision,
                                             varimax_type = varimax_type,
                                             k_simplimax = k_simplimax)

      }
    }

      if ("ULS" %in% estimator) {

        if ("EFAtools" %in% type) {
          grid_list[["ftls_ls"]] <- .type_grid(estimator = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, max_iter = NA,
                                               start_method = NA,
                                               rotation = rotation, k_promax = 4,
                                               normalize = TRUE, P_type = "norm",
                                               precision = 1e-5,
                                               varimax_type = "kaiser",
                                               k_simplimax = k_simplimax)
        }

        if ("psych" %in% type) {
          grid_list[["psch_ls"]] <- .type_grid(estimator = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, max_iter = NA,
                                               start_method = NA,
                                               rotation = rotation, k_promax = 4,
                                               normalize = TRUE, P_type = "unnorm",
                                               precision = 1e-5,
                                               varimax_type = "svd",
                                               k_simplimax = k_simplimax)
        }

        if ("SPSS" %in% type) {
          grid_list[["spss_ls"]] <- .type_grid(estimator = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, max_iter = NA,
                                               start_method = NA,
                                               rotation = rotation, k_promax = 4,
                                               normalize = TRUE, P_type = "norm",
                                               precision = 1e-5,
                                               varimax_type = "kaiser",
                                               k_simplimax = k_simplimax)
        }

        if ("none" %in% type) {

            grid_list[["nn_ls"]] <- .type_grid(estimator = "ULS", init_comm = NA,
                                               criterion = NA, criterion_type = NA,
                                               abs_eigen = NA, max_iter = NA,
                                               start_method = NA,
                                               rotation = rotation, k_promax = k_promax,
                                               normalize = normalize, P_type = P_type,
                                               precision = precision,
                                               varimax_type = varimax_type,
                                               k_simplimax = k_simplimax)

        }
      }

  unique(do.call(rbind, grid_list))
}

test_that(".build_avg_grid reproduces the original grid construction", {

  estimator_sets <- list("PAF", "ML", "ULS", c("PAF", "ML", "ULS"))
  type_sets   <- list("EFAtools", "psych", "SPSS", "none",
                      c("EFAtools", "psych", "SPSS", "none"))
  # Each set is a single rotation kind (none / oblique / orthogonal), as required.
  rot_sets    <- list("none", "oblique", "orthogonal", "promax", "varimax",
                      "simplimax", c("oblimin", "geominQ"),
                      c("varimax", "quartimax"))

  # A non-preset max_iter for the type "none" rows, so the assembled grid carries a
  # genuinely varying max_iter column on the PAF rows (25/50/300 for the named types,
  # 500 here); it stays NA on the ML and ULS rows, which never use the cap.
  base_args <- list(init_comm = "smc", criterion = 1e-3, criterion_type = "sum",
                    abs_eigen = TRUE, max_iter = 500, start_method = "psych",
                    k_promax = 4, normalize = TRUE, P_type = "norm", precision = 1e-5,
                    varimax_type = "kaiser", k_simplimax = 1e-6)

  for (mm in estimator_sets) {
    for (tt in type_sets) {
      for (rr in rot_sets) {
        args <- c(list(estimator = mm, type = tt, rotation = rr), base_args)
        desc <- paste(paste(mm, collapse = "+"), "|", paste(tt, collapse = "+"),
                      "|", paste(rr, collapse = "+"))
        expect_equal(do.call(.build_avg_grid, args),
                     do.call(ref_build_avg_grid, args), info = desc)
      }
    }
  }

  # Vector-valued user arguments are only forwarded by type "none"; they must
  # expand to the same multi-row grid through both implementations.
  vec_args <- list(estimator = c("PAF", "ML", "ULS"), type = "none",
                   rotation = c("oblimin", "geominQ"),
                   init_comm = c("smc", "mac"), criterion = c(1e-3, 1e-4),
                   criterion_type = "sum", abs_eigen = TRUE, max_iter = 500,
                   start_method = c("psych", "factanal"), k_promax = c(3, 4),
                   normalize = TRUE, P_type = c("norm", "unnorm"),
                   precision = c(1e-4, 1e-5), varimax_type = "kaiser",
                   k_simplimax = c(1e-5, 1e-6))
  expect_equal(do.call(.build_avg_grid, vec_args),
               do.call(ref_build_avg_grid, vec_args))
})


### test .oblq_grid

obl_grid_1 <- .oblq_grid(c("PAF"), c("smc", "mac"), .001,
                         c("sum", "max_individual"), c(FALSE, TRUE),
                         500, NA, c("promax", "simplimax", "oblimin"),
                         c(3, 4), TRUE, c("norm", "unnorm"), 1e-5, c("kaiser", "svd"),
                         30)
obl_grid_2 <- .oblq_grid(c("PAF"), c("smc", "mac"), .001,
                         c("sum", "max_individual"), c(FALSE, TRUE),
                         500, NA, c("simplimax", "oblimin"),
                         c(3, 4), TRUE, c("norm", "unnorm"), 1e-5, c("kaiser", "svd"),
                         30)
obl_grid_3 <- .oblq_grid(c("PAF"), c("smc", "mac"), .001,
                         c("sum", "max_individual"), c(FALSE, TRUE),
                         500, NA, c("simplimax", "oblimin"),
                         NA, TRUE, NA, 1e-5, NA, 30)
obl_grid_4 <- .oblq_grid("ML", NA, NA, NA, NA, 500, c("psych", "factanal"),
                         "oblimin", NA, TRUE, NA, 1e-5, NA, NA)

test_that(".oblq_grid works", {
  ### tests for obl_grid_1 with vector-valued PAF arguments
  expect_s3_class(obl_grid_1, "data.frame")
  expect_named(obl_grid_1, c("estimator", "init_comm", "criterion", "criterion_type",
                            "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                            "normalize", "P_type", "precision", "varimax_type",
                            "k_simplimax"))
  expect_equal(nrow(obl_grid_1), 80)
  expect_equal(sum(is.na(obl_grid_1$k_simplimax)), 72)
  expect_equal(sum(is.na(obl_grid_1$k_promax)), 16)
  expect_equal(sum(is.na(obl_grid_1$varimax_type)), 16)
  expect_equal(unique(obl_grid_1$rotation), c("promax", "simplimax", "oblimin"))

  expect_s3_class(obl_grid_2, "data.frame")
  expect_named(obl_grid_2, c("estimator", "init_comm", "criterion", "criterion_type",
                             "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                             "normalize", "P_type", "precision", "varimax_type",
                             "k_simplimax"))
  expect_equal(nrow(obl_grid_2), 16)
  expect_equal(sum(is.na(obl_grid_2$k_promax)), 16)
  expect_equal(sum(is.na(obl_grid_2$varimax_type)), 16)
  expect_equal(unique(obl_grid_2$rotation), c("simplimax", "oblimin"))

  expect_equal(obl_grid_2, obl_grid_3)

  expect_s3_class(obl_grid_4, "data.frame")
  expect_named(obl_grid_4, c("estimator", "init_comm", "criterion", "criterion_type",
                             "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                             "normalize", "P_type", "precision", "varimax_type",
                             "k_simplimax"))
  expect_equal(nrow(obl_grid_4), 2)
  expect_equal(sum(is.na(obl_grid_4$k_simplimax)), 2)
  expect_equal(sum(is.na(obl_grid_4$k_promax)), 2)
  expect_equal(sum(is.na(obl_grid_4$varimax_type)), 2)
  expect_equal(unique(obl_grid_4$rotation), c("oblimin"))
  expect_equal(sum(is.na(obl_grid_4$init_comm)), 2)


})
### test .orth_grid

orth_grid_1 <- .orth_grid(c("PAF"), c("smc", "mac"), .001,
                         c("sum", "max_individual"), c(FALSE, TRUE),
                         500, NA, c("varimax", "quartimax"),
                         TRUE, 1e-5, c("kaiser", "svd"))
orth_grid_2 <- .orth_grid(c("PAF"), c("smc", "mac"), .001,
                          c("sum", "max_individual"), c(FALSE, TRUE),
                          500, NA, c("quartimax"), TRUE, 1e-5,
                          c("kaiser", "svd"))
orth_grid_3 <- .orth_grid(c("PAF"), c("smc", "mac"), .001,
                          c("sum", "max_individual"), c(FALSE, TRUE),
                          500, NA, c("quartimax"), TRUE, 1e-5, NA)
orth_grid_4 <- .orth_grid("ML", NA, NA, NA, NA, 500, c("psych", "factanal"),
                         "quartimax", TRUE, 1e-5, NA)

test_that(".orth_grid works", {
  ### tests for orth_grid_1 with vector-valued PAF arguments
  expect_s3_class(orth_grid_1, "data.frame")
  expect_named(orth_grid_1, c("estimator", "init_comm", "criterion", "criterion_type",
                             "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                             "normalize", "P_type", "precision", "varimax_type",
                             "k_simplimax"))
  expect_equal(nrow(orth_grid_1), 24)
  expect_equal(sum(is.na(orth_grid_1$varimax_type)), 8)
  expect_equal(sum(is.na(orth_grid_1$k_promax)), 24)
  expect_equal(unique(orth_grid_1$rotation), c("varimax", "quartimax"))

  expect_s3_class(orth_grid_2, "data.frame")
  expect_named(orth_grid_2, c("estimator", "init_comm", "criterion", "criterion_type",
                             "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                             "normalize", "P_type", "precision", "varimax_type",
                             "k_simplimax"))
  expect_equal(nrow(orth_grid_2), 8)
  expect_equal(sum(is.na(orth_grid_2$varimax_type)), 8)
  expect_equal(sum(is.na(orth_grid_2$k_promax)), 8)
  expect_equal(unique(orth_grid_2$rotation), c("quartimax"))

  expect_equal(orth_grid_2, orth_grid_3)

  expect_s3_class(orth_grid_4, "data.frame")
  expect_named(orth_grid_4, c("estimator", "init_comm", "criterion", "criterion_type",
                             "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                             "normalize", "P_type", "precision", "varimax_type",
                             "k_simplimax"))
  expect_equal(nrow(orth_grid_4), 2)
  expect_equal(sum(is.na(orth_grid_4$k_simplimax)), 2)
  expect_equal(sum(is.na(orth_grid_4$k_promax)), 2)
  expect_equal(sum(is.na(orth_grid_4$varimax_type)), 2)
  expect_equal(unique(orth_grid_4$rotation), c("quartimax"))
  expect_equal(sum(is.na(orth_grid_4$init_comm)), 2)

})


### test .type_grid

tg_ob <- .type_grid("PAF", c("smc", "mac"), .001,
                   c("sum", "max_individual"), c(FALSE, TRUE),
                   500, NA, "oblique", c(3, 4), TRUE, c("norm", "unnorm"),
                   1e-5, c("kaiser", "svd"), 30)
tg_ob2 <- .type_grid("PAF", c("smc", "mac"), .001,
                    c("sum", "max_individual"), c(FALSE, TRUE),
                    500, NA, c("oblimin", "promax"), c(3, 4), TRUE, c("norm", "unnorm"),
                    1e-5, c("kaiser", "svd"), 30)
tg_orth <- .type_grid("PAF", c("smc", "mac"), .001,
                    c("sum", "max_individual"), c(FALSE, TRUE),
                    500, NA, "orthogonal", c(3, 4), TRUE, c("norm", "unnorm"),
                    1e-5, c("kaiser", "svd"), 30)
tg_orth2 <- .type_grid("PAF", c("smc", "mac"), .001,
                      c("sum", "max_individual"), c(FALSE, TRUE),
                      500, NA, c("varimax", "quartimax"), c(3, 4), TRUE,
                      c("norm", "unnorm"), 1e-5, c("kaiser", "svd"), 30)
tg_nn <- .type_grid("PAF", c("smc", "mac"), .001,
                      c("sum", "max_individual"), c(FALSE, TRUE),
                      500, NA, "none", c(3, 4), TRUE, c("norm", "unnorm"),
                      1e-5, c("kaiser", "svd"), 30)

test_that(".type_grid works", {
  ### test errors
  expect_error(.type_grid("PAF", NA, NA, NA, NA, NA, NA, c("oblique", "none"), NA,
                          NA, NA, NA, NA, NA),
               class = "efa_rotation_length")
  expect_error(.type_grid("PAF", NA, NA, NA, NA, NA, NA, c("oblique", "varimax"), NA,
                          NA, NA, NA, NA, NA),
               class = "efa_rotation_length")
  expect_error(.type_grid("PAF", NA, NA, NA, NA, NA, NA, c("orthogonal", "varimax"),
                          NA, NA, NA, NA, NA, NA),
               class = "efa_rotation_length")
  expect_error(.type_grid("PAF", NA, NA, NA, NA, NA, NA, c("promax", "varimax"),
                          NA, NA, NA, NA, NA, NA),
               class = "efa_rotation_mismatch")

  expect_s3_class(tg_ob, "data.frame")
  expect_named(tg_ob, c("estimator", "init_comm", "criterion", "criterion_type",
                              "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                              "normalize", "P_type", "precision", "varimax_type",
                              "k_simplimax"))
  expect_equal(nrow(tg_ob), 112)
  expect_equal(sum(is.na(tg_ob$varimax_type)),
               nrow(tg_ob) - sum(tg_ob$rotation == "promax"))
  expect_equal(sum(is.na(tg_ob$k_promax)),
               nrow(tg_ob) - sum(tg_ob$rotation == "promax"))
  expect_equal(sum(is.na(tg_ob$P_type)),
               nrow(tg_ob) - sum(tg_ob$rotation == "promax"))
  expect_equal(sum(is.na(tg_ob$k_simplimax)),
               nrow(tg_ob) - sum(tg_ob$rotation == "simplimax"))
  expect_equal(sort(unique(tg_ob$rotation)),
               sort(c("promax", "oblimin", "quartimin", "simplimax",
                 "bentlerQ", "geominQ", "bifactorQ")))

  expect_s3_class(tg_ob2, "data.frame")
  expect_named(tg_ob2, c("estimator", "init_comm", "criterion", "criterion_type",
                        "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                        "normalize", "P_type", "precision", "varimax_type",
                        "k_simplimax"))
  expect_equal(nrow(tg_ob2), 72)
  expect_equal(sum(is.na(tg_ob2$varimax_type)),
               nrow(tg_ob2) - sum(tg_ob2$rotation == "promax"))
  expect_equal(sum(is.na(tg_ob2$k_promax)),
               nrow(tg_ob2) - sum(tg_ob2$rotation == "promax"))
  expect_equal(sum(is.na(tg_ob2$P_type)),
               nrow(tg_ob2) - sum(tg_ob2$rotation == "promax"))
  expect_equal(sum(is.na(tg_ob2$k_simplimax)),
               nrow(tg_ob2) - sum(tg_ob2$rotation == "simplimax"))
  expect_equal(unique(tg_ob2$rotation),
               c("promax", "oblimin"))

  expect_s3_class(tg_orth, "data.frame")
  expect_named(tg_orth, c("estimator", "init_comm", "criterion", "criterion_type",
                        "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                        "normalize", "P_type", "precision", "varimax_type",
                        "k_simplimax"))
  expect_equal(nrow(tg_orth), 56)
  expect_equal(sum(is.na(tg_orth$varimax_type)),
               nrow(tg_orth) - sum(tg_orth$rotation == "varimax"))
  expect_equal(sort(unique(tg_orth$rotation)),
               sort(c("varimax", "quartimax", "equamax",
                      "bentlerT", "geominT", "bifactorT")))

  expect_s3_class(tg_orth2, "data.frame")
  expect_named(tg_orth2, c("estimator", "init_comm", "criterion", "criterion_type",
                         "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                         "normalize", "P_type", "precision", "varimax_type",
                         "k_simplimax"))
  expect_equal(nrow(tg_orth2), 24)
  expect_equal(sum(is.na(tg_orth2$varimax_type)),
               nrow(tg_orth2) - sum(tg_orth2$rotation == "varimax"))
  expect_equal(unique(tg_orth2$rotation),
               c("varimax", "quartimax"))


  expect_s3_class(tg_nn, "data.frame")
  expect_named(tg_nn, c("estimator", "init_comm", "criterion", "criterion_type",
                           "abs_eigen", "max_iter", "start_method", "rotation", "k_promax",
                           "normalize", "P_type", "precision", "varimax_type",
                           "k_simplimax"))
  expect_equal(nrow(tg_nn), 8)
  expect_true(all(is.na(tg_nn$varimax_type)))
  expect_true(all(tg_nn$rotation == "none"))
  expect_true(all(is.na(tg_nn$k_promax)))
  expect_true(all(is.na(tg_nn$normalize)))
  expect_true(all(is.na(tg_nn$P_type)))
  expect_true(all(is.na(tg_nn$precision)))
  expect_true(all(is.na(tg_nn$k_simplimax)))

})


rm(obl_grid_1, obl_grid_2, obl_grid_3, obl_grid_4, orth_grid_1, orth_grid_2,
   orth_grid_3, orth_grid_4, tg_ob, tg_ob2, tg_orth, tg_orth2, tg_nn)

#' Screen data for exploratory factor analysis
#'
#' @description
#' Checks whether your data are suitable for exploratory factor analysis (EFA). From a
#' correlation matrix or raw data, it reports the Kaiser-Meyer-Olkin (KMO) measure of
#' sampling adequacy, Bartlett's test of sphericity, the determinant and condition number
#' of the correlation matrix, and each variable's squared multiple correlation (SMC). When
#' you supply raw data, it also reports each variable's variance and percentage of missing
#' values, category counts for categorical variables, tests of multivariate normality, and
#' multivariate outliers.
#'
#' @param x data.frame or matrix. Raw data, or a correlation matrix. Needs at least three
#'   variables, none of which is a perfect linear combination of the others.
#' @param N numeric. The number of observations. Set this only when you supply a
#'   correlation matrix; it is needed for Bartlett's test of sphericity and is taken from
#'   the data automatically when you supply raw data. Default is `NA`.
#' @param use character. How to handle missing values in raw data. For
#'   `cor_method = "pearson"`, `"spearman"`, or `"kendall"` this is passed to
#'   [stats::cor()]. For `"poly"` or `"tetra"` the same rule is applied to the raw data
#'   before the correlations are estimated; `"all.obs"` and `"everything"` then stop with
#'   an error on any missing value, instead of returning `NA` correlations. Default is
#'   `"pairwise.complete.obs"`.
#' @param cor_method character. How to compute correlations from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (via [stats::cor()]), or `"poly"` / `"tetra"` for
#'   polychoric / tetrachoric correlations of ordinal or binary data. A Spearman or Kendall
#'   correlation matrix is screened on its own scale, not converted to look like a Pearson
#'   correlation matrix; Kendall's tau in particular measures something different from a
#'   Pearson correlation, not just a rescaled version of it. Default is `"pearson"`.
#' @param mcd_alpha numeric. The proportion of cases used to build the robust outlier
#'   estimate, between 0.5 and 1. The default, `0.5`, is the most robust choice; a larger
#'   value uses more of the data but resists outliers less well. Used only with raw data.
#' @param outlier_cutoff numeric. The probability used to set the cutoff for flagging a
#'   multivariate outlier, between 0.5 and 0.9999. Default is `0.975`. Used only with raw
#'   data.
#' @param seed integer. A seed for the random subsets used by the outlier detection, so
#'   the result is reproducible. Does not affect your random-number generator elsewhere.
#'   Default is `NULL`. Used only with raw data.
#'
#' @details
#' The diagnostics are computed from the analysis correlation matrix \eqn{R}:
#' \describe{
#'   \item{KMO}{The Kaiser-Meyer-Olkin measure of sampling adequacy (Kaiser, 1970; Kaiser
#'     & Rice, 1974), overall and for each variable; see [efa_kmo()]. It shows how much
#'     common variance your variables share. Higher values are better; a common rule of
#'     thumb treats values below .50 as unacceptable.}
#'   \item{Bartlett}{Bartlett's (1951) test of sphericity: the likelihood-ratio test of
#'     whether the correlation matrix is an identity matrix, i.e., whether your variables
#'     correlate with each other at all; see [efa_bartlett()]. A significant result
#'     supports doing a factor analysis. The test needs the sample size `N`; without it,
#'     this diagnostic is skipped with a warning and `$bartlett` is `NULL`. If `N` is too
#'     small relative to the number of variables, the statistic is `NA`, also with a
#'     warning.}
#'   \item{Determinant}{The determinant of \eqn{R}, reported as a number only. It falls as
#'     you add variables even when the variables are not collinear, so a fixed cut-off on
#'     it (such as the 0.00001 often quoted from Field, 2018) says more about how many
#'     variables you have than about your data. Use the condition number instead.}
#'   \item{Condition number}{The ratio of the largest to the smallest eigenvalue of
#'     \eqn{R}. Its square root, the condition index, is the collinearity diagnostic of
#'     Belsley, Kuh & Welsch (1980); it drives the printed report and its recommendation.
#'     An index of 10 or less is rarely of interest. An index above 30 flags a near linear
#'     dependency: two or more variables that together carry much the same information.
#'     An index between the two is not negligible, but it stays below the value that
#'     flags a dependency. Belsley (1991) gives 30 as one example value and calls the
#'     choice of a cut-off "somewhat of an art form", so the report grades an index above
#'     30 by its position on the scale 1, 3, 10, 30, 100, 300, 1000: moderate (30 to 100),
#'     strong (100 to 300), or very strong (above 300). These values come from regression
#'     diagnostics on data that are not centred, but a correlation matrix is centred, so
#'     use them as a guide and not as a test.}
#'   \item{SMC}{The squared multiple correlation of each variable with all the others. A
#'     low value flags a variable that has little in common with the rest of your set.}
#'   \item{Variance and missing data}{For raw data: each variable's variance (over its
#'     available values) and percentage of missing values, computed from every row you
#'     supplied. These missing-value percentages explain why the correlation matrix's
#'     sample size (`N`) can be smaller than the number of rows in your data. Ordered-factor
#'     columns are recoded to integer levels first, so `variance` reflects those codes.}
#'   \item{Categories}{For raw data: for each variable with fewer than ten distinct values
#'     (treated as categorical), the count of responses in each category. A category with
#'     fewer than five responses is flagged as sparse, and an unused category between the
#'     smallest and largest response is flagged as empty. As a rule of thumb, items with
#'     fewer than five response categories are better analysed with `cor_method = "poly"`
#'     or `"tetra"` than with Pearson correlations (Rhemtulla et al., 2012).}
#'   \item{Multivariate normality}{For raw data, using only complete cases: two tests of
#'     multivariate normality, Mardia's (1970) test of skewness and kurtosis and the
#'     Henze-Zirkler (1990) test. A small p-value on either test suggests your data depart
#'     from a multivariate normal distribution, a reason to prefer a robust or ordinal
#'     method over normal-theory maximum likelihood. In a very small sample the kurtosis
#'     statistic is `NA`. The Henze-Zirkler p-value is not available with more than about
#'     50 to 60 variables; its test statistic is still reported.}
#'   \item{Outliers}{For raw data, using only complete cases: multivariate outliers, found
#'     from a robust estimate of each case's distance from the centre of your data (the
#'     minimum covariance determinant method; Rousseeuw & Van Driessen, 1999). A flagged
#'     case is unusually far from the rest of your sample. When there are too few complete
#'     cases, the variables are too collinear, or too many cases share identical answers,
#'     a plain (non-robust) distance is used instead, with a warning explaining why.}
#' }
#'
#' @returns An object of class `efa_screen`, a list containing:
#' \item{kmo}{A list with the overall KMO (`KMO`) and the per-variable KMO (`KMO_i`).}
#' \item{bartlett}{A list with Bartlett's chi-square statistic (`chisq`), its `p_value`,
#'   and its degrees of freedom (`df`); `chisq` and `p_value` are `NA` when `N` was too
#'   small for the correction. `NULL` when `N` is unavailable.}
#' \item{determinant}{The determinant of the correlation matrix.}
#' \item{condition}{The condition number of the correlation matrix (largest eigenvalue
#'   over smallest).}
#' \item{smc}{The per-variable squared multiple correlations.}
#' \item{per_item}{A data frame with one row per variable (row names are the variable
#'   names): `variance`, `missing` (percentage), `smc`, `kmo_i`, and `flags` (any
#'   sparse/empty-category issues). `NULL` when a correlation matrix is supplied instead
#'   of raw data.}
#' \item{normality}{A list with `mardia` (skewness `skewness`, `skewness_df`,
#'   `skewness_p`, kurtosis `kurtosis` and `kurtosis_p`, and the underlying `b1p`/`b2p`),
#'   `hz` (the Henze-Zirkler `statistic` and its `p_value`), and `n_complete` (the number
#'   of complete cases used). `NULL` without raw data, or a note explaining why when the
#'   complete-case data cannot support the tests.}
#' \item{outliers}{A list with `distances` (each complete case's robust distance, named by
#'   its row number), `cutoff` (the flagging threshold, on the same scale as `distances`),
#'   `flagged` (the row numbers exceeding `cutoff`), `center` and `cov` (the robust
#'   location and scatter), `method` (`"mcd"` or the `"classical"` fallback),
#'   `fallback_reason` (why the robust estimate was unavailable, when `method` is
#'   `"classical"`), and `n_complete`. `NULL` without raw data, or a note explaining why
#'   when no covariance can be formed.}
#' \item{categories}{A named list with the response-category counts for each categorical
#'   variable (in category order); `NA` for a variable treated as continuous. `NULL`
#'   without raw data.}
#' \item{note}{Explains why the raw-data diagnostics (`per_item`, `normality`,
#'   `outliers`, `categories`) are missing, when a correlation matrix is supplied instead
#'   of raw data. `NULL` when raw data are supplied.}
#' \item{settings}{The settings used: `N`, `n_obs` (rows in the raw data supplied, `NA`
#'   for a correlation-matrix input), `use`, `cor_method`, `mcd_alpha`, `outlier_cutoff`,
#'   and `seed`.}
#'
#' @source Bartlett, M. S. (1951). The effect of standardization on a Chi-square
#'   approximation in factor analysis. Biometrika, 38, 337-344.
#' @source Belsley, D. A. (1991). A guide to using the collinearity diagnostics. Computer
#'   Science in Economics and Management, 4, 33-50.
#' @source Belsley, D. A., Kuh, E. & Welsch, R. E. (1980). Regression diagnostics:
#'   Identifying influential data and sources of collinearity. Wiley.
#' @source Cochran, W. G. (1954). Some methods for strengthening the common
#'   \eqn{\chi^2} tests. Biometrics, 10, 417-451.
#' @source Croux, C. & Haesbroeck, G. (1999). Influence function and efficiency of the
#'   minimum covariance determinant scatter matrix estimator. Journal of Multivariate
#'   Analysis, 71, 161-190.
#' @source Field, A. (2018). Discovering statistics using IBM SPSS statistics (5th
#'   ed.). Sage.
#' @source Henze, N. & Zirkler, B. (1990). A class of invariant consistent tests for
#'   multivariate normality. Communications in Statistics - Theory and Methods, 19,
#'   3595-3617.
#' @source Kaiser, H. F. (1970). A second generation little jiffy. Psychometrika,
#'   35, 401-415.
#' @source Kaiser, H. F. & Rice, J. (1974). Little jiffy, mark IV. Educational and
#'   Psychological Measurement, 34, 111-117.
#' @source Mardia, K. V. (1970). Measures of multivariate skewness and kurtosis with
#'   applications. Biometrika, 57, 519-530.
#' @source Mardia, K. V. (1974). Applications of some measures of multivariate skewness
#'   and kurtosis in testing normality and robustness studies. Sankhya B, 36, 115-128.
#' @source Pison, G., Van Aelst, S. & Willems, G. (2002). Small sample corrections for LTS
#'   and MCD. Metrika, 55, 111-123.
#' @source Rhemtulla, M., Brosseau-Liard, P. E. & Savalei, V. (2012). When can
#'   categorical variables be treated as continuous? A comparison of robust continuous
#'   and categorical SEM estimation methods under suboptimal conditions. Psychological
#'   Methods, 17, 354-373.
#' @source Rousseeuw, P. J. & Van Driessen, K. (1999). A fast algorithm for the minimum
#'   covariance determinant estimator. Technometrics, 41, 212-223.
#'
#' @seealso [efa_kmo()] and [efa_bartlett()] for the individual suitability measures, and
#'   [efa_retain()] for factor retention criteria.
#'
#' @family factor analysis suitability
#'
#' @export
#'
#' @examples
#' # From a correlation matrix (supply N for Bartlett's test of sphericity)
#' efa_screen(test_models$baseline$cormat, N = 500)
#'
#' # From raw data (N is taken from the data; the seed makes the outlier
#' # diagnostics reproducible)
#' efa_screen(GRiPS_raw, seed = 1)
#'
efa_screen <- function(x, N = NA,
                       use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                               "everything", "na.or.complete"),
                       cor_method = c("pearson", "spearman", "kendall", "poly",
                                      "tetra"),
                       mcd_alpha = 0.5, outlier_cutoff = 0.975, seed = NULL) {

  # Perform argument checks
  .assert_cor_input(x)

  # Below three variables the report is degenerate rather than merely uninformative: the
  # KMO is 0/0 at p = 1 and identically .5 at p = 2 whatever the correlation, and
  # Bartlett's test has no or one degree of freedom, so the sections would contradict
  # each other. Refuse it instead of issuing a confident verdict on nothing.
  if (ncol(x) < 3L) {
    cli::cli_abort(
      c("Screening data for factor analysis needs at least three variables.",
        "x" = "{.arg x} has {ncol(x)} variable{?s}.",
        "i" = "The sampling adequacy of one or two variables is not defined: the KMO is
               undefined for a single variable and exactly {.val {0.5}} for any pair."),
      class = "efa_screen_too_few_vars")
  }

  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  checkmate::assert_count(N, na.ok = TRUE)
  checkmate::assert_number(mcd_alpha, lower = 0.5, upper = 1)
  # The endpoints of the natural [0, 1] range are degenerate rather than extreme: a
  # cutoff of 0 flags every observation and a cutoff of 1 puts the threshold at
  # infinity, so neither yields an outlier diagnostic. Bound the probability to the
  # usable range instead.
  checkmate::assert_number(outlier_cutoff, lower = 0.5, upper = 0.9999)
  checkmate::assert_int(seed, null.ok = TRUE)

  # Retain the raw data (when supplied) for the raw-data screening diagnostics;
  # NULL for a correlation-matrix input. data.matrix() codes factor / character
  # columns to their integer category codes, matching how .polychoric() reads
  # ordinal data (as.matrix() would coerce the whole frame to character as soon as
  # any one column is a factor, corrupting every column). data.matrix() only recodes
  # the columns of a data frame, so a non-numeric matrix (e.g. string-coded ordinal
  # responses) is passed through as.data.frame() first to receive the same coding.
  xr <- if (.is_cormat(x)) {
    NULL
  } else if (is.matrix(x) && !is.numeric(x)) {
    data.matrix(as.data.frame(x))
  } else {
    data.matrix(x)
  }

  # The level labels behind that coding, so the category tabulation can name its
  # counts by the responses the user supplied rather than by their integer codes.
  xl <- if (is.null(xr)) NULL else .screen_levels(x)

  # Detect or compute the correlation matrix, check it, and smooth it if needed.
  # N_policy = "optional" keeps N as NA for a correlation matrix without N (only
  # Bartlett's test needs it), and inform_from_data = FALSE suppresses the
  # "computing correlations from the raw data" note.
  # A singular matrix stays a refusal -- every suitability measure here is built from
  # R^-1, so there is nothing trustworthy to report -- but redundant items are the most
  # common reason to screen in the first place, so name them instead of leaving the user
  # to find them. The near-singular case is untouched and still yields a full report with
  # its multicollinearity flags.
  # The handler runs from tryCatch()'s own frame, so the re-raise is given this call
  # explicitly; otherwise the abort would name the internal helper rather than
  # efa_screen(), as every other condition here does.
  screen_call <- rlang::current_env()
  prep <- tryCatch(
    .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                       N_policy = "optional", inform_from_data = FALSE),
    efa_cor_singular = function(cnd) .screen_singular_abort(cnd, call = screen_call))
  R <- prep$R
  N <- prep$N
  p <- ncol(R)

  # Variable names for the per-variable diagnostics, with a V-fallback when the
  # input is unnamed (matching KMO()).
  nms <- colnames(R)
  if (is.null(nms)) nms <- rownames(R)
  if (is.null(nms)) nms <- paste0("V", seq_len(p))

  # Two factorisations of R serve every correlation-side measure below, each formed once
  # rather than once per measure: they cost O(p^3), which is the bulk of the work when a
  # large correlation matrix is screened.
  #
  # The eigenvalues give the condition number and the log-determinant, both taken in
  # absolute value: the singular values of a symmetric matrix are its absolute eigenvalues,
  # so max|lambda| / min|lambda| is exactly the 2-norm condition number, and
  # prod(sign(lambda)) with sum(log|lambda|) are exactly the sign and log-modulus of the
  # determinant. The raw ratio and the raw logarithm would be defined only for a positive
  # definite R.
  ev <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  # R^-1, needed by the KMO's anti-image matrix and by the squared multiple correlations,
  # from the Cholesky factor, which uses the symmetry and positive definiteness that
  # solve()'s LU ignores. .prepare_cor_input() above has refused a numerically singular R
  # and smoothed an indefinite one, so the factorisation exists. Should a Cholesky pivot
  # still turn non-positive, the inverse is left unset and each measure falls back to the
  # route it takes when it is given no inverse -- which keeps .compute_kmo()'s classed
  # singularity abort rather than replacing it with a bare solve() error here.
  R_inv <- tryCatch(chol2inv(chol(R)), error = function(e) NULL)

  # KMO measure of sampling adequacy, overall and per variable
  kmo <- .compute_kmo(R, R_inv = R_inv)
  names(kmo$KMO_i) <- nms

  # Log-determinant of R: passed to the Bartlett statistic (the log form keeps that
  # statistic finite even when |R| underflows to 0) and used here to report the
  # determinant itself. Shaped like determinant()'s return value, which is what
  # .null_chisq() reads.
  ld <- list(sign = prod(sign(ev)), modulus = sum(log(abs(ev))))
  det_R <- ld$sign * exp(ld$modulus)

  # Bartlett's test of sphericity (Bartlett, 1951): the null-model chi-square with
  # df = p(p - 1)/2. It needs N; without it the test is skipped with a warning and the
  # remaining diagnostics are still returned.
  df <- p * (p - 1) / 2
  bartlett <- if (is.na(N)) {
    cli::cli_warn(
      c("{.arg N} is {.val NA}; Bartlett's test of sphericity was skipped.",
        "i" = "Provide {.arg N} or raw data to include it."),
      class = "efa_suitability_no_n"
    )
    NULL
  } else {
    chisq <- .null_chisq(R, N, ld = ld)
    # Same reason, same condition and same wording as efa_bartlett(), so a user reads the
    # same explanation from either entry point.
    if (is.na(chisq)) .warn_bartlett_n_too_small(N, p)
    list(chisq = chisq,
         p_value = stats::pchisq(chisq, df, lower.tail = FALSE),
         df = df)
  }

  # Condition number of R: the ratio of its largest to its smallest eigenvalue in
  # absolute value, i.e. the exact 2-norm condition number. Large values flag
  # near-collinearity.
  condition <- max(abs(ev)) / min(abs(ev))

  # Squared multiple correlation of each variable with the others (1 - 1/diag(R^-1))
  smc <- .smc_start(R, R_inv = R_inv)
  names(smc) <- nms

  # Raw-data descriptive diagnostics: per-item variance, percentage missing, and a
  # category tabulation with sparse/empty-category flags, gathered into a per-item
  # data frame together with the correlation-side smc and per-variable KMO. These
  # need the raw data, so they are NULL for a correlation-matrix input, where a
  # classed note records why.
  if (is.null(xr)) {
    per_item <- NULL
    categories <- NULL
    normality <- NULL
    outliers <- NULL
    note <- structure(
      list(message = paste(
        "Per-item variance, missing-data, category, normality, and outlier",
        "diagnostics require raw data; only a correlation matrix was supplied.")),
      class = "efa_screen_no_raw"
    )
  } else {
    raw <- .screen_raw_diagnostics(xr, nms, xl)
    # The helper's vectors, smc, and kmo$KMO_i are all in column (nms) order, so the
    # columns are assembled positionally. make.unique() guards against duplicate
    # variable names (which base data.frame() row names would reject) and keys the
    # per-item table and the category list identically, so the two stay aligned.
    item_names <- make.unique(nms)
    per_item <- data.frame(
      variance = raw$variance,
      missing  = raw$missing,
      smc      = smc,
      kmo_i    = kmo$KMO_i,
      flags    = raw$flags,
      row.names = item_names,
      stringsAsFactors = FALSE
    )
    categories <- raw$categories
    names(categories) <- item_names
    # Multivariate-normality tests on the complete cases. These need an invertible
    # complete-case covariance; if the data have too few complete cases (heavy
    # missingness) or collinear variables the tests are skipped so the remaining
    # diagnostics are still returned.
    normality <- tryCatch(
      .screen_normality(xr),
      efa_screen_singular = function(cnd) {
        cli::cli_warn(
          c("Multivariate normality tests were skipped.",
            "i" = paste("The complete-case covariance is singular (too few complete",
                        "cases or collinear variables).")),
          class = "efa_screen_no_mvn"
        )
        structure(
          list(message = paste(
            "Multivariate normality tests require an invertible complete-case",
            "covariance; it was singular (too few complete cases or collinear",
            "variables).")),
          class = "efa_screen_no_mvn"
        )
      }
    )
    # Robust-Mahalanobis outlier diagnostics from the complete cases: a FAST-MCD robust
    # location and scatter, the resulting robust distances, and the flagged rows. The
    # helper degrades gracefully (a classed warning plus classical distances, or a classed
    # note) so the remaining diagnostics are still returned.
    outliers <- .screen_outliers(xr, mcd_alpha = mcd_alpha,
                                 outlier_cutoff = outlier_cutoff, seed = seed)
    note <- NULL
  }

  settings <- list(N = N,
                   n_obs = if (is.null(xr)) NA_integer_ else nrow(xr),
                   use = use,
                   cor_method = cor_method,
                   mcd_alpha = mcd_alpha,
                   outlier_cutoff = outlier_cutoff,
                   seed = seed)

  output <- list(
    kmo = kmo,
    bartlett = bartlett,
    determinant = det_R,
    condition = condition,
    smc = smc,
    # Raw-data screening diagnostics; NULL when a correlation matrix is analysed.
    per_item = per_item,
    normality = normality,
    outliers = outliers,
    categories = categories,
    note = note,
    settings = settings
  )

  class(output) <- "efa_screen"

  output

}

# Re-raise a singular-correlation-matrix condition with the redundancy that caused it.
# `cnd` carries the offending matrix, so the perfectly correlated pairs are read off it
# rather than recomputed. A singular matrix without such a pair is singular through a
# linear combination of several variables (a total score, say), which no pair names.
# The pairs are attached to the re-raised condition (`$pairs`) so a caller can act on
# them without parsing the message. Pairs are named as elsewhere in the package
# (.polychoric()'s pair labels), "first-second".
.screen_singular_abort <- function(cnd, call = rlang::caller_env()) {
  R0 <- cnd$R
  # Only .prepare_cor_input() raises this class here, and it attaches the matrix; keep
  # the shared message rather than failing inside the handler if that ever changes.
  if (!is.matrix(R0)) {
    cli::cli_abort(conditionMessage(cnd), class = "efa_cor_singular", call = call)
  }
  nms <- colnames(R0)
  if (is.null(nms)) nms <- rownames(R0)
  if (is.null(nms)) nms <- paste0("V", seq_len(ncol(R0)))
  hi <- which(abs(R0) > 1 - 1e-8 & upper.tri(R0), arr.ind = TRUE)
  # paste0() recycles its separator against an empty index, so the no-pair case has to
  # be taken before it, not after.
  culprits <- if (nrow(hi) > 0L) {
    paste0(nms[hi[, "row"]], "-", nms[hi[, "col"]])
  } else {
    character()
  }
  cli::cli_abort(
    c("The correlation matrix is singular; no further analyses are performed.",
      if (length(culprits)) {
        c("x" = "{cli::qty(culprits)}Perfectly correlated variable pair{?s}: {.val {culprits}}.",
          "i" = "Drop one variable of each pair and screen again.")
      } else {
        c("x" = "No pair is perfectly correlated, so one or more variables are an exact
                 combination of the others.",
          "i" = "Look for total or difference scores among the variables and drop them.")
      }),
    class = "efa_cor_singular", pairs = culprits, call = call)
}

# Level labels of each column of a raw-data input, in column order: the character
# labels behind the integer codes data.matrix() assigns to a factor or character
# column, and NULL for a column it leaves numeric. A character column is coded via
# factor(), so its labels are its sorted distinct values.
.screen_levels <- function(x) {
  if (is.matrix(x) && is.numeric(x)) return(vector("list", ncol(x)))
  d <- if (is.matrix(x)) as.data.frame(x) else x
  lapply(d, function(col) {
    if (is.factor(col)) levels(col)
    else if (is.character(col)) levels(factor(col))
    else NULL
  })
}

# Per-item raw-data descriptive diagnostics from the coded raw matrix `xr`
# (data.matrix() output, so factor columns are their integer level codes). Every
# statistic is computed column by column on the values as supplied - using each
# column's non-missing values - and named by `nms`, which shares the raw matrix's
# column order. `levels_list` (from .screen_levels()) supplies the labels behind
# those codes so the category counts can be named by the original responses.
# Returns variance, percentage missing, category counts, and the sparse/empty-category
# flags.
.screen_raw_diagnostics <- function(xr, nms, levels_list = NULL) {

  p <- ncol(xr)
  variance <- numeric(p)
  missing <- numeric(p)
  flags <- character(p)
  categories <- vector("list", p)
  if (is.null(levels_list)) levels_list <- vector("list", p)

  for (j in seq_len(p)) {
    col <- xr[, j]

    # Variance over the available values (NA when fewer than two are present) and
    # the percentage missing over all supplied rows.
    variance[j] <- stats::var(col, na.rm = TRUE)
    missing[j] <- 100 * mean(is.na(col))

    lv <- sort(unique(col[!is.na(col)]))
    n_cat <- length(lv)

    # Variables with many distinct values are treated as continuous (the >= 10 rule
    # .polychoric() uses to flag a probably-continuous variable): no tabulation and
    # no sparse/empty screening applies to them.
    if (n_cat >= 10L) {
      categories[[j]] <- NA_integer_
      flags[j] <- NA_character_
      next
    }

    # Marginal category counts in level order (tabulate over the recoded levels, as
    # in .polychoric(); table() would key by value-as-string and drop NA). A column
    # that came in as a factor or character is labelled by its original levels, so the
    # tabulation is readable without re-mapping the codes; otherwise the names use a
    # non-scientific format so a large code labels its category as the plain value
    # rather than, e.g., "1e+08".
    counts <- tabulate(match(col, lv), nbins = n_cat)
    labs <- levels_list[[j]]
    names(counts) <- if (!is.null(labs) && all(lv == floor(lv)) &&
                         min(lv) >= 1L && max(lv) <= length(labs)) {
      labs[lv]
    } else {
      format(lv, scientific = FALSE, trim = TRUE)
    }
    categories[[j]] <- counts

    # Sparse: an observed category with fewer than five responses (the >= 5
    # contingency-table rule; a low-frequency category destabilises polychoric /
    # chi-square estimation). Empty: for integer-coded variables only, an unused
    # integer strictly between the smallest and largest observed value (an ordinal
    # item with a skipped category). With `lv` the sorted distinct integer codes, an
    # interior gap exists exactly when the integers spanned (max - min + 1) outnumber
    # those observed - a constant-time test that handles negative and non-contiguous
    # codes without enumerating the (possibly enormous) span. The span is formed in
    # double precision so an integer-typed code range wider than the 32-bit integer
    # limit does not overflow to NA.
    sparse <- any(counts < 5L)
    empty <- n_cat >= 2L && all(lv == floor(lv)) &&
      (as.double(max(lv)) - min(lv) + 1) > n_cat
    flags[j] <- paste(c("sparse", "empty")[c(sparse, empty)], collapse = ", ")
  }

  names(variance) <- names(missing) <- names(flags) <- names(categories) <- nms
  list(variance = variance, missing = missing, flags = flags,
       categories = categories)
}

# Multivariate-normality diagnostics from the coded raw matrix `xr`. Computes the shared
# linear algebra once - the maximum-likelihood (divisor-n) covariance of the complete
# cases and its inverse - then dispatches to Mardia's and the Henze-Zirkler tests, which
# both reuse the same Mahalanobis geometry. Aborts on a correlation-matrix input or a
# singular complete-case covariance (efa_screen() catches the latter and skips the tests).
.screen_normality <- function(xr) {

  # Guard against a correlation matrix reaching the raw-data tests.
  if (.is_cormat(xr)) {
    cli::cli_abort(
      "Multivariate normality tests need raw data, not a correlation matrix.",
      class = "efa_screen_normality_cormat"
    )
  }

  X <- xr[stats::complete.cases(xr), , drop = FALSE]
  n <- nrow(X)
  p <- ncol(X)

  # Centred data and the biased (divisor-n) covariance S; S^-1 via a Cholesky.
  # Mardia's null moments (p(p + 2), 8p(p + 2)/n) and the Henze-Zirkler statistic are
  # both defined with this maximum-likelihood covariance. S is singular whenever there
  # are p or fewer complete cases: centring costs one degree of freedom, so crossprod(Xc)
  # has rank at most n - 1, and n <= p is checked explicitly because at the n == p
  # boundary a rounding-positive pivot can let chol() succeed on an effectively singular
  # S and return a nonsensical inverse (negative squared Mahalanobis distances). The
  # chol() itself then catches the remaining rank-deficient cases (e.g. collinear
  # variables). efa_screen() catches this abort and skips the tests.
  Xc <- scale(X, scale = FALSE)
  S <- crossprod(Xc) / n
  ch <- if (n <= p) NULL else tryCatch(chol(S), error = function(e) NULL)
  if (is.null(ch)) {
    cli::cli_abort(
      "The complete-case covariance is singular; multivariate normality tests need an invertible covariance.",
      class = "efa_screen_singular"
    )
  }
  Sinv <- chol2inv(ch)

  # Shared Mahalanobis geometry: XSinv %*% t(Xc) gives the Gram matrix D = Xc S^-1 Xc'
  # (reused block by block below), and its diagonal d2 holds the squared Mahalanobis
  # distances. Both tests need the full n x n D, so the accumulation is chunked into
  # row-blocks to cap memory (~ block * n doubles instead of n^2).
  XSinv <- Xc %*% Sinv
  d2 <- rowSums(XSinv * Xc)
  block <- max(1L, min(n, 4194304L %/% n))
  # Both tests form the Gram blocks as XSinv %*% t(Xc), so t(Xc) is transposed once here -
  # alongside d2 - rather than re-transposed inside each helper.
  tXc <- t(Xc)

  list(mardia = .mardia(XSinv, tXc, d2, n, p, block),
       hz = .hz(XSinv, tXc, d2, n, p, block),
       n_complete = n)
}

# Mardia's (1970) multivariate skewness and kurtosis. b1p = (1/n^2) sum_ij d_ij^3 with
# d_ij = x_i' S^-1 x_j (the Gram matrix D), accumulated in row-blocks; b2p = (1/n) sum_i
# d_ii^2. The skewness statistic n b1p / 6 is chi-square with p(p+1)(p+2)/6 df and always
# carries Mardia's (1974) correction; b2p is standardised to N(0, 1) with its exact null
# moments. Ref: Mardia (1970, 1974).
.mardia <- function(XSinv, tXc, d2, n, p, block) {

  s3 <- 0
  for (start in seq.int(1L, n, by = block)) {
    idx <- start:min(start + block - 1L, n)
    D_block <- XSinv[idx, , drop = FALSE] %*% tXc   # |idx| x n block of D
    s3 <- s3 + sum(D_block^3)
  }
  b1p <- s3 / n^2
  b2p <- sum(d2^2) / n

  chi_df <- p * (p + 1) * (p + 2) / 6
  # Mardia's (1974) correction k, eq. (5.5), which he gives "so that E(A') = f for all n"
  # - not only for small samples. From his eq. (5.1) the uncorrected n b1p / 6 has
  # expectation chi_df / k exactly, so without k the statistic is biased low, by a factor
  # that grows with the ratio of variables to observations. The bias is a small fraction
  # of chi_df, but the chi-square scale is sqrt(2 chi_df), so it reaches nearly 3 standard
  # deviations at p = 60 and n = 150 and leaves the test with no power at all there.
  # The correction fixes the mean exactly; the chi-square variance stays wrong by about
  # -20% to +30%, with no published exact variance of b1p to correct it with, so the test
  # keeps a rejection rate of roughly .04 to .08 against a nominal .05.
  # Ref: Mardia (1974), eq. (5.1) and (5.5), and section 6, recommendation (1).
  k <- (p + 1) * (n + 1) * (n + 3) / (n * ((n + 1) * (p + 1) - 6))
  skew <- k * n * b1p / 6

  # b2p is standardised with its EXACT null moments under normality,
  #   E[b2p]   = p(p + 2)(n - 1) / (n + 1)                  Mardia (1970), eq. (3.16)
  #   Var[b2p] = 8p(p + 2)(n - 3)(n - p - 1)(n - p + 1) /
  #              ((n + 1)^2 (n + 3)(n + 5))                 Mardia (1974), eq. (5.3)
  # giving Mardia's (1974) statistic B', eq. (5.6). Both moments become the asymptotic
  # pair p(p + 2) and 8p(p + 2)/n as n grows, which most other implementations use. That
  # pair overstates both, by an amount that grows with the ratio of variables to
  # observations. The overstated mean pushes the statistic down, so the two-sided test
  # rejects data from an exact multivariate normal far too often: more than half the time
  # at p = 40 and n = 200. The overstated variance works the opposite way, so the exact
  # mean with the asymptotic variance - Mardia (1970), eq. (3.20) - is conservative
  # instead, and rejects almost nothing at p = 50 and n = 100. At p = 1 the variance
  # reduces to Fisher's (1930) univariate result, 24n(n - 2)(n - 3) / ((n + 1)^2 (n + 3)
  # (n + 5)). b2p stays mildly right-skewed after standardisation, so the normal
  # approximation splits the two tails unevenly (about 1.7% and 3.0% against a nominal
  # 2.5% at p = 40 and n = 200, and closer with more observations); the two-sided level
  # that the p-value reports holds at .05.
  # The exact variance is 0 at n = p + 1, the smallest sample the caller admits, where
  # the centred data span the whole space orthogonal to the mean, so b2p equals (n - 1)^2
  # with probability 1 and the statistic does not exist.
  kurt_var <- 8 * p * (p + 2) * (n - 3) * (n - p - 1) * (n - p + 1) /
    ((n + 1)^2 * (n + 3) * (n + 5))
  kurt <- if (kurt_var > 0) {
    (b2p - p * (p + 2) * (n - 1) / (n + 1)) / sqrt(kurt_var)
  } else {
    NA_real_
  }

  list(skewness = skew,
       skewness_df = chi_df,
       skewness_p = stats::pchisq(skew, chi_df, lower.tail = FALSE),
       kurtosis = kurt,
       kurtosis_p = 2 * stats::pnorm(abs(kurt), lower.tail = FALSE),
       b1p = b1p, b2p = b2p)
}

# Henze-Zirkler (1990) omnibus test of multivariate normality: a weighted integral of the
# distance between the empirical and normal characteristic functions, evaluated with the
# smoothing parameter b. The three-term statistic reuses the pairwise Mahalanobis
# distances D_jk = d_jj + d_kk - 2 G_jk (G the Gram matrix, accumulated in row-blocks);
# under normality it is approximately lognormal with the parameters (mu, si2) below. That
# approximation degenerates at many variables, where the p-value is withheld (see below).
# Ref: Henze & Zirkler (1990).
.hz <- function(XSinv, tXc, d2, n, p, block) {

  b <- (1 / sqrt(2)) * ((2 * p + 1) * n / 4)^(1 / (p + 4))
  b2 <- b^2

  s1 <- 0
  for (start in seq.int(1L, n, by = block)) {
    idx <- start:min(start + block - 1L, n)
    D_block <- XSinv[idx, , drop = FALSE] %*% tXc
    Djk <- outer(d2[idx], d2, "+") - 2 * D_block   # pairwise Mahalanobis^2
    s1 <- s1 + sum(exp(-b2 / 2 * Djk))
  }
  term1 <- s1 / n^2
  term2 <- 2 * (1 + b2)^(-p / 2) * mean(exp(-(b2 / (2 * (1 + b2))) * d2))
  term3 <- (1 + 2 * b2)^(-p / 2)
  HZ <- n * (term1 - term2 + term3)

  # Lognormal null distribution of HZ (Henze & Zirkler, 1990): mean mu and variance si2,
  # reparameterised to the log scale (pmu, psi); large HZ signals non-normality, so the
  # p-value is the upper tail.
  a <- 1 + 2 * b2
  wb <- (1 + b2) * (1 + 3 * b2)
  mu <- 1 - a^(-p / 2) * (1 + p * b2 / a + p * (p + 2) * b2^2 / (2 * a^2))
  si2 <- 2 * (1 + 4 * b2)^(-p / 2) +
    2 * a^(-p) * (1 + 2 * p * b2^2 / a^2 + 3 * p * (p + 2) * b2^4 / (4 * a^4)) -
    4 * wb^(-p / 2) * (1 + 3 * p * b2^2 / (2 * wb) + p * (p + 2) * b2^4 / (2 * wb^2))
  pmu <- log(sqrt(mu^4 / (si2 + mu^2)))
  psi <- sqrt(log((si2 + mu^2) / mu^2))

  # The lognormal null approximation breaks down at many variables. As p grows si2 falls
  # geometrically while mu goes to 1, so si2 + mu^2 rounds to mu^2, psi rounds to 0, and
  # the z-score (log(HZ) - pmu) / psi becomes -Inf or +Inf: the p-value is then exactly 0
  # or exactly 1, decided by a rounding-level residual, even on data from an exact
  # multivariate normal. The p-value is thus withheld when si2 is no longer resolvable
  # against mu^2, the point at which the reparameterisation loses its input. This is a
  # property of the null moments alone, so where it applies follows n and p and is not a
  # fixed ratio of them. Writing the ratio in cancellation-free form moves the point but
  # does not remove it, because HZ saturates at mu as well. The statistic stays correct
  # and is still reported; only its p-value is unavailable.
  if (!is.finite(psi) || si2 <= mu^2 * .Machine$double.eps) {
    reason <- paste("The lognormal null distribution of the Henze-Zirkler statistic has",
                    "no resolvable spread at", p, "variables, so the statistic cannot be",
                    "turned into a p-value.")
    cli::cli_warn(
      c("The Henze-Zirkler p-value is not available.", "i" = reason),
      class = "efa_screen_no_hz"
    )
    return(structure(
      list(statistic = HZ, p_value = NA_real_, message = reason),
      class = "efa_screen_no_hz"
    ))
  }

  list(statistic = HZ,
       p_value = stats::pnorm((log(HZ) - pmu) / psi, lower.tail = FALSE))
}

# Robust-Mahalanobis outlier diagnostics from the coded raw matrix `xr`. Estimates a
# high-breakdown location and scatter from the robustly rescaled complete cases with
# FAST-MCD, scales it to consistency at the normal model with a small-sample correction,
# reweights it, returns it to the supplied units, and flags observations whose squared
# robust distance exceeds the chi-square cutoff. Degrades gracefully: too few complete
# cases, near-collinear variables, or an exact fit fall back to the classical Mahalanobis
# distance with a classed warning naming which of the three applies, and a fully singular
# covariance yields a classed note. The random subsets are reproducible via `seed`, and the
# caller's RNG state is preserved. Aborts on a correlation-matrix input.
.screen_outliers <- function(xr, mcd_alpha, outlier_cutoff, nsamp = 500L,
                             seed = NULL) {

  # Guard against a correlation matrix reaching the raw-data diagnostics.
  if (.is_cormat(xr)) {
    cli::cli_abort(
      "Outlier diagnostics need raw data, not a correlation matrix.",
      class = "efa_screen_outliers_cormat"
    )
  }

  # A Mahalanobis distance needs a complete row, so the diagnostics use the complete
  # cases; `rows` maps them back to the supplied data for reporting.
  cc <- stats::complete.cases(xr)
  X <- xr[cc, , drop = FALSE]
  rows <- which(cc)
  n <- nrow(X)
  p <- ncol(X)

  # The MCD estimator is affine equivariant -- det(A S A') = det(A)^2 det(S), so ranking
  # covering subsets by the determinant of their covariance, and the Mahalanobis distance
  # built from the winning one, are unchanged by any nonsingular A (Rousseeuw & Van Driessen,
  # 1999). The rcond() gates that decide whether a subset is usable are not: on a covariance
  # they measure how far apart the variables' units are as much as how close the data lie to
  # a lower-dimensional hyperplane, and on four well-conditioned variables a scale ratio of
  # about twenty already tips them. Running the search on columns divided by a robust scale
  # turns them into a test of the subset's *correlation* matrix instead, which is the rank
  # test they were always meant to be. The median absolute deviation is itself
  # high-breakdown, so the scale cannot be set by the very outliers it exists to find; a
  # column with no spread about its median falls back to the standard deviation, and one
  # with no spread at all to 1.
  sc <- apply(X, 2L, stats::mad)
  bad <- !is.finite(sc) | sc <= 0
  if (any(bad)) sc[bad] <- apply(X[, bad, drop = FALSE], 2L, stats::sd)
  bad <- !is.finite(sc) | sc <= 0
  sc[bad] <- 1
  Z <- sweep(X, 2L, sc, "/", check.margin = FALSE)

  # Reproducible random subsets without side effects on the caller's RNG: save the current
  # .Random.seed (or arrange to remove a freshly created one) and restore it on exit, then
  # seed the stream when a seed is supplied.
  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    old_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    on.exit(assign(".Random.seed", old_seed, envir = .GlobalEnv), add = TRUE)
  } else {
    on.exit(suppressWarnings(rm(".Random.seed", envir = .GlobalEnv)), add = TRUE)
  }
  if (!is.null(seed)) set.seed(seed)

  # Squared-distance critical value used for flagging; the reported cutoff (below) is on
  # the distance scale so that it is directly comparable to the reported `distances`.
  crit <- stats::qchisq(outlier_cutoff, p)

  # Robust location/scatter via FAST-MCD followed by a reweighting step. Too few complete
  # cases (n <= 2p) or a collinear MCD subset fall back to the classical mean/covariance
  # with a classed warning; if even the classical covariance is singular, no distances can
  # be formed and a classed note is returned.
  robust <- tryCatch(
    {
      if (n <= 2L * p) {
        cli::cli_abort("Too few complete cases for a robust covariance.",
                       class = "efa_screen_mcd_unusable")
      }
      h <- .mcd_hsize(n, p, mcd_alpha)
      fit <- .fast_mcd(Z, h, nsamp = nsamp)
      # Consistency (Croux & Haesbroeck, 1999) and small-sample (Pison et al., 2002)
      # scaling of the raw MCD scatter. The consistency factor uses the realised coverage
      # h/n; the small-sample correction uses the nominal `mcd_alpha` (as in its calibration).
      # Both are scalars in (p, n, mcd_alpha) alone, so they commute with the rescaling and
      # can be applied here, before the estimate is put back on the supplied scale.
      raw_center <- fit$center
      raw_cov <- .mcd_consistency(p, h / n) *
        .mcd_cnp2(p, n, mcd_alpha, reweighted = FALSE) * fit$cov
      # Reweighting: keep observations within the 0.975 cutoff of the raw fit, recompute
      # the mean/scatter, and rescale it (consistency at the 0.975 retention plus the
      # reweighted small-sample correction). Fall back to the raw fit if the reweighted
      # scatter would be singular.
      d2_raw <- stats::mahalanobis(Z, raw_center, raw_cov)
      w <- d2_raw <= stats::qchisq(0.975, p)
      rew_cov_sub <- if (sum(w) > p) stats::cov(Z[w, , drop = FALSE]) else NULL
      if (!is.null(rew_cov_sub) && rcond(rew_cov_sub) >= 1e-12) {
        list(center = colMeans(Z[w, , drop = FALSE]),
             cov = .mcd_consistency(p, 0.975) *
               .mcd_cnp2(p, n, mcd_alpha, reweighted = TRUE) * rew_cov_sub,
             method = "mcd")
      } else {
        list(center = raw_center, cov = raw_cov, method = "mcd")
      }
    },
    efa_screen_mcd_unusable = function(cnd) {
      # Fall back to the classical mean/covariance when it is invertible; otherwise signal
      # a total failure so a single, outcome-appropriate condition is raised below.
      # Gate on the reciprocal condition number, matching the reweighting gate above:
      # `mahalanobis()` inverts through `solve()`, which refuses anything whose rcond is
      # below the double-precision epsilon, so the gate has to test the same quantity.
      # Positive definiteness alone is not enough -- a collinear covariance can pass a
      # Cholesky on the sign of a rounding-level pivot (and does on some BLAS
      # implementations) yet still abort in `solve()`. A non-finite covariance is screened
      # off first: it has no condition number, and `rcond()` would abort on it.
      cov_cl <- stats::cov(Z)
      if (n > p && all(is.finite(cov_cl)) && rcond(cov_cl) >= 1e-12) {
        list(center = colMeans(Z), cov = cov_cl, method = "classical")
      } else {
        NULL
      }
    }
  )

  if (is.null(robust)) {
    # Name the condition that actually defeated the classical covariance in the handler
    # above, which gates on all three of n > p, a finite scatter, and its conditioning.
    reason <- if (n <= p) {
      "There are too few complete cases (n <= p) to form a covariance of full rank."
    } else if (!all(is.finite(stats::cov(Z)))) {
      # the same quantity the handler's finiteness gate rejected, recomputed here rather
      # than threaded out of it; this is a cold path, so the second cov() costs nothing
      paste("The complete-case covariance is not finite; the data contain non-finite",
            "or extreme values.")
    } else {
      paste("The complete-case covariance is singular: the variables are exactly or",
            "near-exactly linearly dependent.")
    }
    cli::cli_warn(
      c("Outlier diagnostics were skipped.", "i" = reason),
      class = "efa_screen_no_outliers"
    )
    return(structure(
      list(message = paste("Robust outlier diagnostics require an invertible",
                           "complete-case covariance.", reason),
           reason = reason),
      class = "efa_screen_no_outliers"
    ))
  }

  # A classed warning when the robust estimate was unavailable and classical distances were
  # used instead, naming the reason. Which of the three ways the search can fail applies is
  # read off the data rather than off whichever internal gate happened to fire first, since
  # the same gate is reachable from more than one cause: variables that are near-collinear
  # as a set starve the search of well-conditioned starting subsets, which is the same abort
  # a genuine exact fit produces. Too few complete cases is settled before the search runs;
  # a complete-case covariance that is itself ill-conditioned means the variables are
  # near-collinear; and a well-conditioned one that still admits no usable covering subset
  # means at least h of the cases lie on a lower-dimensional hyperplane. That distinction is
  # the whole of what a reader acts on -- a redundant item to drop, against tied answers
  # that are no fault of the variables -- so it is worth deciding on the evidence. The
  # threshold is the one the robust scatter itself has to clear in .fast_mcd(), on the same
  # rescaled columns.
  fallback_reason <- NULL
  if (identical(robust$method, "classical")) {
    fallback_reason <- if (n <= 2L * p) {
      "There are too few complete cases (n <= 2p) for a robust covariance."
    } else if (rcond(stats::cov(Z)) < 1e-3) {
      paste("The complete-case covariance is ill-conditioned: the variables are so nearly",
            "linearly dependent that no covering subset has a covariance stable enough to",
            "measure distances with. Look for redundant items rather than for outlying",
            "respondents.")
    } else {
      paste("At least half the complete cases lie exactly on a lower-dimensional",
            "hyperplane (an \"exact fit\"). This is common with coarse discrete items,",
            "where many respondents give identical answers on an item pair; it does not",
            "mean the data are collinear at the correlation level.")
    }
    cli::cli_warn(
      c("A robust (MCD) covariance could not be computed; classical Mahalanobis distances were used.",
        "i" = fallback_reason),
      class = "efa_screen_mcd_fallback"
    )
  }

  # Robust distances and outlier flags (squared distance beyond the chi-square cutoff).
  # `distances` and `flagged` share the supplied-data row numbers as their identifier: the
  # distances are named by row number (of the complete cases) and `flagged` holds the row
  # numbers whose distance exceeds the cutoff, so the two align even with incomplete rows.
  # The distances are taken on the rescaled columns, which is where the conditioning gates
  # above did their work: the Mahalanobis distance is the same either way, but putting the
  # scatter back on the supplied scale first can spoil its conditioning by as much as the
  # square of the ratio of the column scales, and `solve()` inside `mahalanobis()` would
  # then refuse a scatter the gates had already passed as usable.
  d2 <- stats::mahalanobis(Z, robust$center, robust$cov)
  distances <- sqrt(d2)
  names(distances) <- as.character(rows)
  flagged <- rows[d2 > crit]

  # The reported centre and scatter go back onto the supplied scale, so that they are in the
  # variables' own units; they describe the distances above, which the rescaling leaves
  # unchanged.
  robust$center <- robust$center * sc
  robust$cov <- outer(sc, sc) * robust$cov

  list(distances = distances,
       cutoff = sqrt(crit),
       flagged = flagged,
       center = robust$center,
       cov = robust$cov,
       method = robust$method,
       fallback_reason = fallback_reason,
       n_complete = n)
}

# Size of the MCD subset covering a proportion `alpha` of the observations (Rousseeuw &
# Van Driessen, 1999): at alpha = 0.5 the maximum-breakdown floor((n + p + 1)/2), rising
# to n at alpha = 1.
.mcd_hsize <- function(n, p, alpha) {
  n_half <- floor((n + p + 1) / 2)
  floor(2 * n_half - n + 2 * (n - n_half) * alpha)
}

# Consistency factor (Croux & Haesbroeck, 1999) that makes the raw MCD scatter consistent
# for the covariance at the multivariate normal model, for a subset covering a proportion
# `alpha`. Equivalent to alpha / F_{chi^2_{p+2}}(qchisq(alpha, p)).
.mcd_consistency <- function(p, alpha) {
  q <- stats::qchisq(alpha, p)
  alpha / stats::pgamma(q / 2, shape = p / 2 + 1)
}

# Small-sample correction factor (Pison, Van Aelst & Willems, 2002) that removes the finite-
# sample bias of the consistency-corrected MCD scatter. `reweighted` selects the factor for
# the raw (FALSE) or the reweighted (TRUE) estimator. For p > 2 the p = 2 and p = 3 fitted
# coefficients are interpolated in log p; for p <= 2 the published closed forms are used.
# The correction is evaluated at the alpha = 0.5 and alpha = 0.875 anchors and interpolated
# linearly in alpha. Ref: Pison et al. (2002), Metrika 55, 111-123.
.mcd_cnp2 <- function(p, n, alpha, reweighted = FALSE) {
  if (reweighted) {
    coef_875 <- matrix(c(-0.544482443573914, 1.25994483222292, 2,
                         -0.343791072183285, 1.25159004257133, 3), ncol = 2)
    coef_500 <- matrix(c(-1.02842572724793, 1.67659883081926, 2,
                         -0.26800273450853, 1.35968562893582, 3), ncol = 2)
    small_2 <- c(3.11101712909049, 1.91401056721863, 0.79473550581058, 1.10081930350091)
    small_1 <- c(1.11098143415027, 1.5182890270453, -0.66046776772861, 0.88939595831888)
  } else {
    coef_875 <- matrix(c(-0.455179464070565, 1.11192541278794, 2,
                         -0.294241208320834, 1.09649329149811, 3), ncol = 2)
    coef_500 <- matrix(c(-1.42764571687802, 1.26263336932151, 2,
                         -1.06141115981725, 1.28907991440387, 3), ncol = 2)
    small_2 <- c(0.673292623522027, 0.691365864961895, 0.446537815635445, 1.06690782995919)
    small_1 <- c(0.262024211897096, 0.604756680630497, -0.351584646688712, 1.01646567502486)
  }
  if (p > 2) {
    y500 <- log(-coef_500[1, ] / p^coef_500[2, ])
    y875 <- log(-coef_875[1, ] / p^coef_875[2, ])
    b500 <- solve(cbind(1, -log(coef_500[3, ] * p^2)), y500)
    b875 <- solve(cbind(1, -log(coef_875[3, ] * p^2)), y875)
    fp_500 <- 1 - exp(b500[1]) / n^b500[2]
    fp_875 <- 1 - exp(b875[1]) / n^b875[2]
  } else if (p == 2) {
    fp_500 <- 1 - exp(small_2[1]) / n^small_2[2]
    fp_875 <- 1 - exp(small_2[3]) / n^small_2[4]
  } else {
    fp_500 <- 1 - exp(small_1[1]) / n^small_1[2]
    fp_875 <- 1 - exp(small_1[3]) / n^small_1[4]
  }
  fp <- if (alpha <= 0.875) {
    fp_500 + (fp_875 - fp_500) / 0.375 * (alpha - 0.5)
  } else {
    fp_875 + (1 - fp_875) / 0.125 * (alpha - 0.875)
  }
  unname(1 / fp)
}

# One concentration step (Rousseeuw & Van Driessen, 1999): given a location/scatter, take
# the `h` observations with the smallest Mahalanobis distances and return their mean and
# covariance; each step cannot increase the covariance determinant. Returns NULL if the
# scatter is singular (so the calling candidate is abandoned).
.mcd_cstep <- function(X, mu, S, h) {
  ch <- tryCatch(chol(S), error = function(e) NULL)
  if (is.null(ch)) return(NULL)
  Xc <- sweep(X, 2L, mu, "-", check.margin = FALSE)
  d2 <- rowSums((Xc %*% chol2inv(ch)) * Xc)
  idx <- order(d2)[seq_len(h)]
  Xsub <- X[idx, , drop = FALSE]
  cov_new <- stats::cov(Xsub)
  # Reject a (near-)singular subset: its covariance cannot give stable distances, and the
  # minimum-determinant objective would otherwise collapse onto such a degenerate subset
  # (e.g. a tight cluster on a near-hyperplane) rather than the intended robust scatter.
  if (rcond(cov_new) < 1e-8) return(NULL)
  list(center = colMeans(Xsub), cov = cov_new)
}

# FAST-MCD (Rousseeuw & Van Driessen, 1999): find the h-subset of X whose covariance has the
# smallest determinant. From `nsamp` random (p + 1)-subsets (enlarged until non-singular),
# take two concentration steps each, keep the ten lowest-determinant candidates, and iterate
# those to convergence; return the best subset's mean and covariance. Aborts (classed,
# triggering the classical fallback) when no non-singular starting subset can be formed, when
# the minimum-determinant subset turns out to be singular (an exact fit: at least h
# observations lie on a lower-dimensional hyperplane, so a usable robust covariance does not
# exist), or when the winning subset's covariance is too ill-conditioned to give stable
# distances. Expects columns already on a common scale, so that its rcond() gates test rank
# rather than the variables' units.
.fast_mcd <- function(X, h, nsamp = 500L, maxit = 100L) {
  n <- nrow(X)
  p <- ncol(X)

  logdet <- function(S) as.numeric(determinant(S, logarithm = TRUE)$modulus)

  starts <- vector("list", nsamp)
  dets <- rep_len(NA_real_, nsamp)
  for (i in seq_len(nsamp)) {
    # Draw a (p + 1)-subset with a well-conditioned covariance, enlarging it if needed.
    sub <- sample.int(n, p + 1L)
    S <- stats::cov(X[sub, , drop = FALSE])
    tries <- 0L
    while (rcond(S) < 1e-8 && tries < n) {
      sub <- unique(c(sub, sample.int(n, 1L)))
      S <- stats::cov(X[sub, , drop = FALSE])
      tries <- tries + 1L
    }
    if (rcond(S) < 1e-8) next
    # Two concentration steps. A subset that reaches a degenerate scatter is dropped by
    # .mcd_cstep; only starts that completed at least one step (so `st` is a genuine
    # h-subset, not the raw (p + 1)-point start) are recorded, otherwise the tiny-sample
    # covariance of an unconcentrated start would win the minimum-determinant selection.
    st <- list(center = colMeans(X[sub, , drop = FALSE]), cov = S)
    concentrated <- FALSE
    for (k in 1:2) {
      nxt <- .mcd_cstep(X, st$center, st$cov, h)
      if (is.null(nxt)) break
      st <- nxt
      concentrated <- TRUE
    }
    if (concentrated) {
      starts[[i]] <- st
      dets[i] <- logdet(st$cov)
    }
  }

  ran <- which(!is.na(dets))
  if (!length(ran)) {
    cli::cli_abort("No well-conditioned MCD subset could be formed (exact fit).",
                   class = "efa_screen_mcd_unusable")
  }

  # Iterate the ten lowest-determinant candidates and keep the overall best that reaches a
  # stable fixed point (its determinant stops decreasing, i.e. the subset stabilises).
  keep <- ran[order(dets[ran])][seq_len(min(10L, length(ran)))]
  best <- NULL
  best_ld <- Inf
  for (i in keep) {
    st <- starts[[i]]
    prev <- logdet(st$cov)
    stabilised <- FALSE
    for (it in seq_len(maxit)) {
      nxt <- .mcd_cstep(X, st$center, st$cov, h)
      # A concentration step that lands on a singular h-subset has found a covering subset
      # of determinant zero, and no subset can have a smaller determinant than that: the
      # objective is minimised, and its minimiser is degenerate (an exact fit). Reaching
      # one therefore settles the question whichever candidate got there, and the search
      # stops. Dropping that candidate instead and returning the best non-degenerate fixed
      # point would label an estimate "MCD" that is demonstrably not the minimum-covariance-
      # determinant one, and would make the verdict depend on which random starts happened
      # to be drawn (Rousseeuw & Van Driessen, 1999, sec. 3).
      if (is.null(nxt)) {
        cli::cli_abort("The minimum-determinant subset is singular (exact fit).",
                       class = "efa_screen_mcd_unusable")
      }
      cur <- logdet(nxt$cov)
      st <- nxt
      if (cur >= prev - 1e-12) { prev <- cur; stabilised <- TRUE; break }
      prev <- cur
    }
    if (stabilised && prev < best_ld) {
      best_ld <- prev
      best <- st
    }
  }

  # No candidate settled within `maxit` concentration steps. Each step lowers the
  # determinant and there are finitely many h-subsets, so the iteration must reach a fixed
  # point eventually; exhausting the limit is a numerical stalemate rather than a statement
  # about the data. Signal the classical fallback.
  if (is.null(best)) {
    cli::cli_abort("No MCD subset settled within the concentration-step limit.",
                   class = "efa_screen_mcd_unusable")
  }
  # A near-singular best subset likewise marks a (near-)exact fit, so the robust distances
  # would be dominated by the near-null direction; fall back to the classical distances.
  # Note this is a property of the h-subset, not of the variables as a whole: coarse
  # discrete items routinely put more than h respondents on a hyperplane x_i = x_j while
  # the correlation matrix stays well conditioned, so the reported condition number is not
  # a proxy for it.
  if (rcond(best$cov) < 1e-3) {
    cli::cli_abort("The MCD scatter is near-singular (collinear variables / exact fit).",
                   class = "efa_screen_mcd_unusable")
  }
  best
}

scr_cor <- efa_screen(test_models$baseline$cormat, N = 500)
# GRiPS has an exact collinearity (enjoy == commonly for most respondents), so the robust
# MCD covariance is (near-)singular and the outlier diagnostic falls back to classical
# distances with a warning; suppress it here (it is asserted in its own test below). A seed
# makes the MCD random subsets - and hence the outlier diagnostics and the print snapshot -
# reproducible; it does not affect the seed-independent KMO/Bartlett/per-item/normality parts.
scr_raw <- suppressWarnings(efa_screen(GRiPS_raw, seed = 1))
scr_iris <- efa_screen(iris[, 1:4], seed = 1)

dat_nonames <- test_models$baseline$cormat
colnames(dat_nonames) <- NULL
rownames(dat_nonames) <- NULL
scr_nona <- efa_screen(dat_nonames, N = 500)

# correlation matrix without N: Bartlett's test skipped
scr_non <- suppressWarnings(efa_screen(test_models$baseline$cormat))

test_that("output class and structure are correct", {
  expect_s3_class(scr_cor, "efa_screen")
  expect_s3_class(scr_raw, "efa_screen")

  expect_named(scr_cor$kmo, c("KMO", "KMO_i"))
  expect_named(scr_cor$bartlett, c("chisq", "p_value", "df"))

  expect_length(scr_cor$kmo$KMO_i, ncol(test_models$baseline$cormat))
  expect_length(scr_cor$smc, ncol(test_models$baseline$cormat))

  # raw-only sections are NULL when a correlation matrix is analysed, with a note
  expect_null(scr_cor$per_item)
  expect_null(scr_cor$normality)
  expect_null(scr_cor$outliers)
  expect_null(scr_cor$categories)
  expect_s3_class(scr_cor$note, "efa_screen_no_raw")

  # the raw path fills the per-item table and clears the note
  expect_s3_class(scr_raw$per_item, "data.frame")
  expect_named(scr_raw$per_item, c("variance", "missing", "smc", "kmo_i", "flags"))
  expect_null(scr_raw$note)
})

test_that("KMO, Bartlett, determinant, condition, and SMC match the standalone tools", {
  R <- test_models$baseline$cormat

  # KMO consistent with the standalone KMO()
  k <- KMO(R)
  expect_equal(scr_cor$kmo$KMO, k$KMO, tolerance = 1e-10)
  expect_equal(unname(scr_cor$kmo$KMO_i), unname(k$KMO_i), tolerance = 1e-10)

  # Bartlett consistent with the standalone BARTLETT()
  b <- BARTLETT(R, N = 500)
  expect_equal(scr_cor$bartlett$chisq, b$chisq, tolerance = 1e-8)
  expect_equal(scr_cor$bartlett$df, b$df)
  expect_equal(scr_cor$bartlett$p_value, b$p_value, tolerance = 1e-8)

  # raw-data branch: N recovered from the data
  expect_equal(scr_raw$bartlett$chisq, suppressMessages(BARTLETT(GRiPS_raw))$chisq,
               tolerance = 1e-8)
  expect_equal(scr_raw$settings$N, nrow(GRiPS_raw))

  # determinant and condition number (eigenvalue-ratio convention)
  expect_equal(scr_cor$determinant, det(R), tolerance = 1e-8)
  expect_equal(scr_cor$condition, kappa(R, exact = TRUE), tolerance = 1e-8)
  ev <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
  expect_equal(scr_cor$condition, max(ev) / min(ev), tolerance = 1e-6)

  # SMC == the closed form 1 - 1/diag(R^-1)
  expect_equal(unname(scr_cor$smc), unname(1 - 1 / diag(solve(R))), tolerance = 1e-10)
  expect_length(scr_cor$smc, ncol(R))
})

test_that("KMO, Bartlett, and SMC match psych", {
  skip_on_cran()
  skip_if_not_installed("psych")

  R <- test_models$baseline$cormat

  ps <- psych::KMO(R)
  expect_equal(scr_cor$kmo$KMO, unname(ps$MSA), tolerance = 1e-4)
  expect_equal(unname(scr_cor$kmo$KMO_i), unname(ps$MSAi), tolerance = 1e-4)

  expect_equal(scr_cor$bartlett$chisq, psych::cortest.bartlett(R, n = 500)$chisq,
               tolerance = 1e-4)

  expect_equal(unname(scr_cor$smc), unname(psych::smc(R)), tolerance = 1e-6)
})

test_that("the raw-data path is silent and matches the standalone tools", {
  # inform_from_data = FALSE: the "computing correlations from the raw data" note
  # is suppressed here (unlike the standalone suitability tools)
  expect_no_message(suppressWarnings(efa_screen(GRiPS_raw)), class = "efa_cor_from_data")

  # the correlation-side diagnostics on the raw path go through the same pipeline
  # as the standalone KMO(), including the per-variable naming
  km <- suppressMessages(KMO(GRiPS_raw))
  expect_equal(scr_raw$kmo$KMO, km$KMO, tolerance = 1e-10)
  expect_equal(unname(scr_raw$kmo$KMO_i), unname(km$KMO_i), tolerance = 1e-10)
  expect_equal(names(scr_raw$kmo$KMO_i), colnames(GRiPS_raw))
  expect_equal(names(scr_raw$smc), colnames(GRiPS_raw))
})

test_that("per-variable diagnostics get a V-fallback when unnamed", {
  expect_equal(names(scr_nona$kmo$KMO_i), paste0("V", seq_len(ncol(dat_nonames))))
  expect_equal(names(scr_nona$smc), paste0("V", seq_len(ncol(dat_nonames))))
})

test_that("Bartlett's test is skipped with a note when N is unavailable", {
  expect_warning(efa_screen(test_models$baseline$cormat),
                 class = "efa_suitability_no_n")

  expect_null(scr_non$bartlett)
  expect_true(is.na(scr_non$settings$N))
  # the remaining diagnostics are still returned
  expect_false(is.null(scr_non$kmo))
  expect_false(is.null(scr_non$smc))
  expect_true(is.finite(scr_non$determinant))
  expect_true(is.finite(scr_non$condition))
})

test_that("a too-small N is reported rather than left as a bare NA", {
  # N is known here, so the test is attempted and comes back undefined: the multiplier
  # N - 1 - (2p + 5)/6 is non-positive at N = 5 for p = 18. Same condition and class as
  # efa_bartlett(), so a user reads the same reason from either entry point.
  expect_warning(scr_small <- efa_screen(test_models$baseline$cormat, N = 5),
                 class = "efa_bartlett_n_too_small")
  expect_true(is.na(scr_small$bartlett$chisq))
  expect_true(is.na(scr_small$bartlett$p_value))
  # the remaining suitability diagnostics are unaffected
  expect_true(is.finite(scr_small$kmo$KMO))
  expect_true(is.finite(scr_small$determinant))

  # the printed report carries the same reason, so it is not only in the warning. Matched on
  # the collapsed report rather than line by line, so the assertion does not depend on where
  # cli happens to break the sentence.
  expect_match(paste(cli::ansi_strip(format(scr_small)), collapse = " "),
               "Bartlett multiplier", fixed = TRUE)

  expect_no_warning(efa_screen(test_models$baseline$cormat, N = 500),
                    class = "efa_bartlett_n_too_small")
})

test_that("settings are returned correctly", {
  expect_named(scr_cor$settings, c("N", "n_obs", "use", "cor_method", "mcd_alpha",
                                   "outlier_cutoff", "seed"))
  expect_equal(scr_cor$settings$N, 500)
  expect_equal(scr_cor$settings$use, "pairwise.complete.obs")
  expect_equal(scr_cor$settings$cor_method, "pearson")
  expect_equal(scr_cor$settings$outlier_cutoff, 0.975)

  # n_obs is the number of supplied raw rows (NA for a correlation-matrix input)
  expect_true(is.na(scr_cor$settings$n_obs))
  expect_equal(scr_raw$settings$n_obs, nrow(GRiPS_raw))

  # the MCD arguments that govern the outlier diagnostics are recorded, so a saved
  # object says how its robust distances were obtained
  expect_equal(scr_cor$settings$mcd_alpha, 0.5)
  expect_null(scr_cor$settings$seed)
  # on the raw path both are the values the outlier diagnostics actually ran with
  expect_equal(scr_raw$settings$seed, 1)
  expect_equal(scr_raw$settings$mcd_alpha, 0.5)
  expect_equal(
    efa_screen(test_models$baseline$cormat, N = 500, mcd_alpha = 0.75)$settings$mcd_alpha,
    0.75)
})

test_that("raw-data per-item diagnostics: variance, missing, and category flags (GRiPS)", {
  pi <- scr_raw$per_item

  # per-item variance, and (zero) missingness for the complete GRiPS data
  expect_equal(pi$variance, apply(GRiPS_raw, 2, var),
               tolerance = 1e-10, ignore_attr = TRUE)
  expect_true(all(pi$missing == 0))

  # smc / kmo_i in the table are the same vectors reported at the top level
  expect_equal(pi$smc, unname(scr_raw$smc), tolerance = 1e-12, ignore_attr = TRUE)
  expect_equal(pi$kmo_i, unname(scr_raw$kmo$KMO_i), tolerance = 1e-12,
               ignore_attr = TRUE)

  # GRiPS items are 6-category (0-5) with no gaps; enjoy/hurt/attracted each have a
  # category with fewer than five responses (sparse), the other five are clean
  expect_equal(pi[c("enjoy", "hurt", "attracted"), "flags"], rep("sparse", 3L),
               ignore_attr = TRUE)
  expect_equal(pi[c("fun", "friends", "part", "commonly", "chances"), "flags"],
               rep("", 5L), ignore_attr = TRUE)

  # categories: one count table per item, summing to N, in category order
  expect_type(scr_raw$categories, "list")
  expect_length(scr_raw$categories, ncol(GRiPS_raw))
  expect_true(all(vapply(scr_raw$categories, sum, numeric(1)) == nrow(GRiPS_raw)))
  expect_equal(names(scr_raw$categories$fun), as.character(0:5))
})

test_that("raw-data flags separate sparse, empty, and continuous variables", {
  set.seed(123)
  n <- 60L
  cont1 <- rnorm(n)
  cont2 <- rnorm(n)
  miss <- rnorm(n); miss[sample.int(n, 6L)] <- NA           # 10% missing, continuous
  sparse <- sample(c(rep(1, 20), rep(2, 20), rep(3, 17), rep(4, 3)))  # 4 appears 3x
  empty <- sample(c(rep(1, 15), rep(2, 15), rep(4, 15), rep(5, 15)))  # 3 is skipped
  dat <- cbind(cont1, cont2, miss, sparse, empty)

  scr <- efa_screen(dat)
  pi <- scr$per_item

  # variance and percentage missing computed on all supplied rows, independent of use
  expect_equal(pi$variance, apply(dat, 2, var, na.rm = TRUE),
               tolerance = 1e-10, ignore_attr = TRUE)
  expect_equal(pi["miss", "missing"], 10)
  expect_equal(pi[c("cont1", "cont2", "sparse", "empty"), "missing"], rep(0, 4L),
               ignore_attr = TRUE)
  expect_equal(scr$settings$n_obs, n)

  # a low-frequency category is flagged sparse, a skipped interior category empty,
  # and variables with many distinct values are treated as continuous (flags NA)
  expect_equal(pi["sparse", "flags"], "sparse")
  expect_equal(pi["empty", "flags"], "empty")
  expect_true(all(is.na(pi[c("cont1", "cont2", "miss"), "flags"])))

  # the offending categories are visible in the tabulation; continuous columns are NA
  expect_equal(unname(scr$categories$sparse["4"]), 3L)
  expect_false("3" %in% names(scr$categories$empty))
  expect_true(is.na(scr$categories$cont1))
})

test_that("factor columns are coded (not corrupted) and an unused ordinal category is empty", {
  set.seed(1)
  n <- 120L
  # ordered factors carry the response order the polychoric path needs; f1 never uses its
  # third level ("c"), leaving an interior gap
  f1 <- ordered(sample(c("a", "b", "d"), n, replace = TRUE),
                levels = c("a", "b", "c", "d"))
  f2 <- ordered(sample(c("lo", "mid", "hi"), n, replace = TRUE),
                levels = c("lo", "mid", "hi"))
  f3 <- ordered(sample(c("lo", "mid", "hi"), n, replace = TRUE),
                levels = c("lo", "mid", "hi"))
  dat <- data.frame(f1 = f1, f2 = f2, f3 = f3)

  scr <- suppressWarnings(efa_screen(dat, cor_method = "poly"))

  # data.matrix() coding (not as.matrix()): variance is finite and matches the codes
  expect_true(all(is.finite(scr$per_item$variance)))
  expect_equal(scr$per_item$variance, apply(data.matrix(dat), 2, var),
               tolerance = 1e-10, ignore_attr = TRUE)
  # f1's unused middle level (c -> code 3, between b = 2 and d = 4) is an empty category
  expect_match(scr$per_item["f1", "flags"], "empty")

  # the category counts are labelled by the original levels, not by the integer codes,
  # and only the observed levels appear (f1 never uses "c")
  expect_equal(names(scr$categories$f1), c("a", "b", "d"))
  expect_equal(names(scr$categories$f2), c("lo", "mid", "hi"))
  expect_equal(scr$categories$f1, table(dat$f1)[c("a", "b", "d")], ignore_attr = TRUE)
})

test_that("raw data with duplicate column names is handled positionally, not by name", {
  set.seed(7)
  x <- data.frame(a = rnorm(150), a = rnorm(150, 5, 3), b = rnorm(150),
                  check.names = FALSE)
  scr <- efa_screen(x)

  # no crash; per-item rows carry make.unique()'d variable names
  expect_equal(rownames(scr$per_item), c("a", "a.1", "b"))
  # each column reports its OWN diagnostics (the duplicate name is not aliased to
  # the first matching column)
  expect_equal(scr$per_item$variance, apply(data.matrix(x), 2, var),
               tolerance = 1e-10, ignore_attr = TRUE)
  expect_length(scr$categories, ncol(x))
  # per_item and categories are keyed the same way, so named lookup agrees
  expect_equal(names(scr$categories), rownames(scr$per_item))
})

test_that("a wide integer-code span does not overflow the empty-category test", {
  set.seed(9)
  # an integer MATRIX (data.matrix keeps integer storage) whose ordinal column spans
  # more than the 32-bit integer range: max - min must be formed in double, or it
  # overflows to NA and mislabels the flag "NA" instead of "empty"
  mi <- matrix(0L, 200L, 3L)
  mi[, 1] <- sample(c(-2000000000L, 0L, 1L, 2000000000L), 200L, replace = TRUE)
  mi[, 2] <- sample(1:5, 200L, replace = TRUE)
  mi[, 3] <- sample(1:5, 200L, replace = TRUE)
  colnames(mi) <- c("v1", "v2", "v3")
  expect_identical(typeof(mi), "integer")

  scr <- suppressWarnings(efa_screen(mi))
  expect_equal(scr$per_item["v1", "flags"], "empty")
})

test_that("large category codes are labelled as plain values, not scientific notation", {
  set.seed(10)
  dat <- data.frame(big = sample(c(1, 2, 100000000), 80L, replace = TRUE),
                    v2 = rnorm(80), v3 = rnorm(80))
  scr <- efa_screen(dat)
  expect_true("100000000" %in% names(scr$categories$big))
  expect_false("1e+08" %in% names(scr$categories$big))
})

test_that("a non-numeric (character) matrix is rejected on the polychoric path", {
  set.seed(11)
  cm <- matrix(sample(c("lo", "mid", "hi"), 300L, replace = TRUE), nrow = 100L,
               dimnames = list(NULL, c("v1", "v2", "v3")))
  # A character matrix carries no response order: coding it would rank the labels
  # alphabetically (hi < lo < mid), silently attenuating the polychoric correlations, so the
  # polychoric path refuses it rather than returning a misleading screening report.
  expect_error(suppressMessages(efa_screen(cm, cor_method = "poly")),
               class = "efa_cor_unordered_factor")
})

test_that("a wide-span integer code is flagged empty without enumerating its range", {
  set.seed(8)
  # a sentinel code far from the ordinal range: the empty-category test must not
  # enumerate the full 1..2e9 span (which would exhaust memory)
  dat <- data.frame(v1 = sample(c(1L, 2L, 3L, 2000000000L), 200, replace = TRUE),
                    v2 = rnorm(200), v3 = rnorm(200))
  scr <- efa_screen(dat)
  expect_equal(scr$per_item["v1", "flags"], "empty")
})

test_that("multivariate normality diagnostics are present on the raw path", {
  nm <- scr_raw$normality
  expect_named(nm, c("mardia", "hz", "n_complete"))
  expect_named(nm$mardia, c("skewness", "skewness_df", "skewness_p",
                            "kurtosis", "kurtosis_p", "b1p", "b2p"))
  expect_named(nm$hz, c("statistic", "p_value"))

  # GRiPS is complete, so every row feeds the tests
  expect_equal(nm$n_complete, nrow(GRiPS_raw))

  # statistics finite and p-values in [0, 1]
  expect_true(all(is.finite(c(nm$mardia$skewness, nm$mardia$kurtosis, nm$hz$statistic))))
  expect_true(nm$mardia$skewness_p >= 0 && nm$mardia$skewness_p <= 1)
  expect_true(nm$mardia$kurtosis_p >= 0 && nm$mardia$kurtosis_p <= 1)
  expect_true(nm$hz$p_value >= 0 && nm$hz$p_value <= 1)

  # NULL for a correlation-matrix input
  expect_null(scr_cor$normality)
})

test_that("the Henze-Zirkler p-value is withheld where its null approximation degenerates", {
  # The lognormal null variance si2 falls geometrically with the number of variables
  # while the null mean mu goes to 1, so si2 + mu^2 rounds to mu^2 and the z-score
  # divides by a zero psi. Without the guard the p-value is then exactly 0 or exactly 1,
  # decided by a rounding-level residual, on data from an exact multivariate normal.
  set.seed(4)
  X <- matrix(stats::rnorm(150 * 60), 150, 60)

  expect_warning(nm <- EFAtools:::.screen_normality(X), class = "efa_screen_no_hz")
  expect_s3_class(nm$hz, "efa_screen_no_hz")
  expect_true(is.na(nm$hz$p_value))
  # only the null approximation fails: the statistic and Mardia's tests are unaffected
  expect_true(is.finite(nm$hz$statistic))
  expect_true(is.finite(nm$mardia$skewness_p))
  expect_true(is.finite(nm$mardia$kurtosis_p))

  # and the report neither passes nor rejects on it, and counts the tests that remain
  local_reproducible_output()
  colnames(X) <- paste0("v", seq_len(60))
  expect_warning(scr <- efa_screen(X, seed = 1), class = "efa_screen_no_hz")
  txt <- cli::ansi_strip(paste(format(scr), collapse = " "))
  expect_match(txt, "p not available at this number of variables", fixed = TRUE)
  expect_match(txt, "of the 2 tests", fixed = TRUE)
})

test_that("the multivariate normality verdict counts the tests that reject", {
  local_reproducible_output()
  # GRiPS departs from normality on all three tests, so the verdict names all three; the
  # count is what separates a unanimous rejection from a single marginal one.
  expect_match(cli::ansi_strip(paste(format(scr_raw), collapse = " ")),
               "3 of the 3 tests reject it", fixed = TRUE)
  # a clean multivariate normal sample takes the other branch, which names the
  # denominator too (all three p-values here are above .96, so the fixture is not tied)
  set.seed(381)
  clean <- matrix(stats::rnorm(200 * 5), 200, 5)
  colnames(clean) <- paste0("v", seq_len(5))
  expect_match(cli::ansi_strip(paste(format(efa_screen(clean, seed = 1)), collapse = " ")),
               "consistent with multivariate normality: none of the 3 tests rejects it",
               fixed = TRUE)
})

test_that("Mardia's skewness/kurtosis match psych after the divisor rescaling", {
  skip_on_cran()
  skip_if_not_installed("psych")

  X <- as.matrix(iris[, 1:4])
  n <- nrow(X)
  md <- scr_iris$normality$mardia
  m <- psych::mardia(X, plot = FALSE)

  # efa_screen uses the biased (divisor-n) covariance; psych::mardia uses cov()
  # (divisor n - 1). The two are related exactly by (n / (n - 1))^j, so the COEFFICIENTS
  # agree after rescaling. The two statistics built on them do not, by design: psych
  # standardises b2p with the asymptotic null moments where efa_screen uses the exact
  # ones, and psych drops Mardia's skewness correction above 20 observations where
  # efa_screen keeps it. Both divergences have their own tests below.
  expect_equal(md$b1p, m$b1p * (n / (n - 1))^3, tolerance = 1e-6)
  expect_equal(md$b2p, m$b2p * (n / (n - 1))^2, tolerance = 1e-6)
})

test_that("Mardia's kurtosis is standardised with Mardia's (1974) exact null moments", {
  md <- scr_iris$normality$mardia
  n <- nrow(iris)
  p <- 4L

  kurt_mean <- p * (p + 2) * (n - 1) / (n + 1)
  kurt_var <- 8 * p * (p + 2) * (n - 3) * (n - p - 1) * (n - p + 1) /
    ((n + 1)^2 * (n + 3) * (n + 5))
  expect_equal(md$kurtosis, (md$b2p - kurt_mean) / sqrt(kurt_var), tolerance = 1e-10)
})

test_that("Mardia's kurtosis holds its nominal level on exact multivariate normal data", {
  # The exact null moments matter most when the number of variables is large relative to
  # the sample size. The asymptotic pair p(p + 2) and 8p(p + 2)/n that most other
  # implementations use overstates both: at n = 50 and p = 20 it puts the statistic at
  # about -2.1 on average with a standard deviation near 0.54, and rejects data from an
  # exact multivariate normal about 59% of the time at the .05 level. Correcting the mean
  # alone leaves the standard deviation at 0.54 and turns that into a test that almost
  # never rejects, so all three checks below are needed to pin both corrections.
  set.seed(20240817)
  n <- 50L
  p <- 20L
  md <- lapply(seq_len(300L), function(i) {
    EFAtools:::.screen_normality(matrix(stats::rnorm(n * p), n, p))$mardia
  })
  z <- vapply(md, `[[`, numeric(1), "kurtosis")
  pv <- vapply(md, `[[`, numeric(1), "kurtosis_p")

  expect_lt(abs(mean(z)), 0.25)
  expect_gt(stats::sd(z), 0.85)
  expect_lt(stats::sd(z), 1.20)
  # the reported p-value drives the level, so assert it rather than rebuilding it
  expect_lt(mean(pv < .05), 0.12)
  # and it is the two-sided normal tail of the reported statistic
  expect_equal(pv, 2 * stats::pnorm(abs(z), lower.tail = FALSE), tolerance = 1e-12)
})

test_that("Mardia's kurtosis is NA when its exact null variance is zero", {
  # At n = p + 1 the centred data span the whole space orthogonal to the mean, so b2p
  # equals (n - 1)^2 with probability 1 and carries no information; its exact null
  # variance is 0 and the standardised statistic does not exist.
  set.seed(1)
  n <- 8L
  p <- 7L
  nrm <- EFAtools:::.screen_normality(matrix(stats::rnorm(n * p), n, p))

  expect_equal(nrm$mardia$b2p, (n - 1)^2, tolerance = 1e-8)
  expect_true(is.na(nrm$mardia$kurtosis))
  expect_true(is.na(nrm$mardia$kurtosis_p))
  # the skewness test is unaffected
  expect_true(is.finite(nrm$mardia$skewness_p))

  # the printed report calls it unavailable rather than showing it as a pass, and drops
  # it from the denominator of the verdict count
  set.seed(4)
  Xn <- matrix(stats::rnorm(12 * 11), 12, 11)
  colnames(Xn) <- paste0("v", seq_len(11))
  txt <- cli::ansi_strip(paste(suppressWarnings(format(efa_screen(Xn, seed = 1))),
                               collapse = " "))
  expect_match(txt, "Mardia's kurtosis: not available", fixed = TRUE)
  expect_match(txt, "of the 2 tests", fixed = TRUE)
})

test_that("the Henze-Zirkler statistic and p-value match the reference value on iris", {
  # Henze-Zirkler (1990) statistic for the four numeric iris columns, computed once from
  # this implementation and cross-checked against MVN::hz(iris[, 1:4]) = 2.336.
  expect_equal(scr_iris$normality$hz$statistic, 2.3363942003, tolerance = 1e-6)
  # p-value pinned relatively (an absolute tolerance is meaningless at ~1e-19)
  expect_equal(scr_iris$normality$hz$p_value / 4.1413116299e-19, 1, tolerance = 1e-6)
})

test_that("the corrected Mardia skewness statistic matches psych below 20 observations", {
  skip_on_cran()
  skip_if_not_installed("psych")

  set.seed(1)
  Xs <- matrix(rnorm(15 * 3), 15, 3)
  nrm <- EFAtools:::.screen_normality(Xs)
  ms <- psych::mardia(Xs, plot = FALSE)

  # below 20 observations psych also applies the correction, as `small.skew`, so the two
  # agree after the divisor rescaling
  expect_equal(nrm$mardia$skewness, ms$small.skew * (15 / 14)^3, tolerance = 1e-6)
})

test_that("Mardia's skewness correction applies at every sample size", {
  # Mardia's (1974) correction k, eq. (5.5), is what makes the statistic's expectation
  # equal its degrees of freedom; it is not a small-sample device. psych and most other
  # implementations apply it only below 20 observations, so above that the two diverge by
  # exactly k. Here n = 60, far above any small-sample threshold.
  set.seed(4)
  n <- 60L
  p <- 6L
  X <- matrix(stats::rnorm(n * p), n, p)
  nrm <- EFAtools:::.screen_normality(X)
  m <- psych::mardia(X, plot = FALSE)

  k <- (p + 1) * (n + 1) * (n + 3) / (n * ((n + 1) * (p + 1) - 6))
  expect_gt(k, 1)
  # psych's uncorrected `skew` is our statistic divided by k, after the divisor rescaling
  expect_equal(nrm$mardia$skewness, k * m$skew * (n / (n - 1))^3, tolerance = 1e-6)
  # and matches psych's own corrected form, which psych computes but does not use here
  expect_equal(nrm$mardia$skewness, m$small.skew * (n / (n - 1))^3, tolerance = 1e-6)
})

test_that("the corrected Mardia skewness statistic is centred on its degrees of freedom", {
  # The correction makes E(statistic) = df exactly at every n. Without it the statistic is
  # biased low by the factor k: at n = 50 and p = 20 its expectation is df / k = 1444.7
  # against df = 1540, which is 1.72 chi-square standard deviations below, and the test
  # then almost never rejects.
  set.seed(20240818)
  n <- 50L
  p <- 20L
  df <- p * (p + 1) * (p + 2) / 6
  s <- vapply(seq_len(200L), function(i) {
    EFAtools:::.screen_normality(matrix(stats::rnorm(n * p), n, p))$mardia$skewness
  }, numeric(1))

  # mean within 0.5 chi-square sd of the df; the uncorrected statistic sits 1.72 sd below
  expect_lt(abs(mean(s) - df) / sqrt(2 * df), 0.5)
})

test_that("normality tests abort on a correlation matrix and on a singular covariance", {
  expect_error(EFAtools:::.screen_normality(test_models$baseline$cormat),
               class = "efa_screen_normality_cormat")

  set.seed(1)
  # n_complete <= p: the complete-case covariance is singular (centring leaves rank at
  # most n - 1). Both n < p and the n == p boundary abort deterministically - at n == p
  # a rounding-positive pivot can let chol() alone succeed on the singular covariance.
  expect_error(EFAtools:::.screen_normality(matrix(rnorm(3 * 5), 3, 5)),
               class = "efa_screen_singular")
  expect_error(EFAtools:::.screen_normality(matrix(rnorm(4 * 4), 4, 4)),
               class = "efa_screen_singular")
})

test_that("efa_screen skips the normality tests when the covariance is singular", {
  set.seed(3)
  Sig <- matrix(0.35, 4, 4)
  diag(Sig) <- 1
  # Drawn through the Cholesky factor: for a positive definite Sigma that factor is unique
  # (the positive-diagonal convention fixes it), so the seeded sample is reproducible to
  # rounding on any LAPACK build, which an eigen-based draw is not -- eigenvector signs and
  # the basis of a repeated eigenspace are left free by the decomposition.
  dat <- matrix(stats::rnorm(200 * 4), 200) %*% chol(Sig)
  colnames(dat) <- paste0("v", seq_len(4))
  # a rotating block of missing values leaves no complete case (so the MVN tests cannot
  # run) while every pair of variables stays well observed (so the pairwise correlation
  # matrix, and the other diagnostics, are unaffected)
  blk <- rep(1:4, each = 50)
  for (i in seq_len(200)) dat[i, blk[i]] <- NA

  # with no complete case neither the normality tests nor the robust outlier diagnostics
  # can run, so each degrades with its own classed warning and note
  expect_warning(
    expect_warning(scr <- efa_screen(dat), class = "efa_screen_no_mvn"),
    class = "efa_screen_no_outliers"
  )
  expect_s3_class(scr$normality, "efa_screen_no_mvn")
  expect_s3_class(scr$outliers, "efa_screen_no_outliers")
  # the remaining diagnostics are still returned
  expect_false(is.null(scr$kmo))
  expect_s3_class(scr$per_item, "data.frame")
})

# ---- robust-Mahalanobis outlier diagnostics (FAST-MCD) ----

# injected-outlier fixture: a clean multivariate-normal core with a block of far
# outliers; a robust covariance should recover the clean centre/scatter and flag the
# injected rows, where the classical covariance (inflated by the outliers) would not
set.seed(2024)
n_out <- 200L
p_out <- 4L
Sig_out <- matrix(0.4, p_out, p_out)
diag(Sig_out) <- 1
X_out <- matrix(rnorm(n_out * p_out), n_out, p_out) %*% chol(Sig_out)
colnames(X_out) <- paste0("v", seq_len(p_out))
inj <- 1:12
# a shifted, jittered block of outliers (jittered so they do not collapse onto a single
# point, which would make their own covariance degenerate)
X_out[inj, ] <- matrix(rep(c(7, -7, 7, -7), each = length(inj)), ncol = p_out) +
  matrix(rnorm(length(inj) * p_out), ncol = p_out)
scr_out <- efa_screen(X_out, seed = 1)

test_that("robust outlier diagnostics have the expected structure", {
  o <- scr_out$outliers
  expect_named(o, c("distances", "cutoff", "flagged", "center", "cov", "method",
                    "fallback_reason", "n_complete"))
  expect_equal(o$method, "mcd")
  # nothing to explain when the robust estimate was available
  expect_null(o$fallback_reason)
  expect_equal(o$n_complete, n_out)
  expect_length(o$distances, n_out)
  # cutoff is on the distance scale (comparable to `distances`), and `flagged` is exactly
  # reproducible from the two reported fields
  expect_equal(o$cutoff, sqrt(qchisq(0.975, p_out)))
  expect_equal(sort(as.integer(names(o$distances)[o$distances > o$cutoff])), o$flagged)
  expect_true(all(is.finite(o$distances)))
  # the cutoff probability is exposed
  expect_equal(efa_screen(X_out, seed = 1, outlier_cutoff = 0.99)$outliers$cutoff,
               sqrt(qchisq(0.99, p_out)))
  # NULL on a correlation-matrix input
  expect_null(scr_cor$outliers)
})

test_that("the MCD flags the injected outliers and recovers the clean centre", {
  o <- scr_out$outliers
  # every injected row is flagged, and the injected rows are cleanly separated from the
  # rest (their robust distances exceed every non-injected distance)
  expect_true(all(inj %in% o$flagged))
  expect_gt(min(o$distances[inj]), max(o$distances[-inj]))
  # few false positives (near the 2.5% nominal rate), not a swamped solution
  expect_lte(length(setdiff(o$flagged, inj)), 15L)
  # the robust centre recovers the clean mean (~0) despite the outliers
  expect_equal(unname(o$center), rep(0, p_out), tolerance = 0.2)
  # the consistency + small-sample scaling makes the robust covariance unbiased for the
  # clean Sigma (unit variances); a scale-invariant correlation comparison cannot see this,
  # so a dropped consistency factor (which halves the scatter) would be caught here
  expect_equal(diag(o$cov), rep(1, p_out), tolerance = 0.35, ignore_attr = TRUE)
  # flagged indices refer to rows of the supplied data
  expect_true(all(o$flagged %in% seq_len(n_out)))
})

test_that("robust outlier diagnostics are reproducible via a seed and leave the RNG alone", {
  a <- efa_screen(X_out, seed = 7)
  b <- efa_screen(X_out, seed = 7)
  expect_identical(a$outliers$flagged, b$outliers$flagged)
  expect_equal(a$outliers$distances, b$outliers$distances)
  # a different seed can give a different (still-valid) solution but still flags injected
  expect_true(all(inj %in% efa_screen(X_out, seed = 99)$outliers$flagged))
  # the caller's random-number-generator stream is preserved
  set.seed(4321)
  before <- .Random.seed
  invisible(efa_screen(X_out, seed = 3))
  expect_identical(before, .Random.seed)
})

test_that("outlier row indices map back to the supplied rows when data are incomplete", {
  # incomplete rows are placed *before* the outliers, so a complete-case position differs
  # from the original row index; the reported flags and distance names must refer to the
  # supplied data's rows, not to compacted positions in the complete-case submatrix
  set.seed(77)
  n2 <- 220L
  S2 <- matrix(0.4, p_out, p_out)
  diag(S2) <- 1
  X2 <- matrix(rnorm(n2 * p_out), n2, p_out) %*% chol(S2)
  colnames(X2) <- paste0("v", seq_len(p_out))
  out2 <- 200:211
  X2[out2, ] <- matrix(rep(c(7, -7, 7, -7), each = length(out2)), ncol = p_out) +
    matrix(rnorm(length(out2) * p_out), ncol = p_out)
  X2[1:20, 1] <- NA                          # 20 incomplete rows before the outliers
  o <- efa_screen(X2, seed = 1)$outliers

  expect_equal(o$method, "mcd")
  expect_equal(o$n_complete, 200L)
  expect_true(all(out2 %in% o$flagged))
  # the outlier rows keep their supplied-data indices (>= 200); a missing remap would
  # report compacted positions <= 200
  expect_gte(max(o$flagged), 200L)
  expect_true("200" %in% names(o$distances))
  expect_false("1" %in% names(o$distances))  # row 1 is incomplete, so it has no distance
})

test_that("a non-default mcd_alpha threads through the whole outlier pipeline", {
  # a larger coverage (alpha = 0.75) enlarges the MCD subset but still yields a robust
  # estimate that flags the injected outliers
  a <- efa_screen(X_out, seed = 1, mcd_alpha = 0.75)$outliers
  expect_equal(a$method, "mcd")
  expect_true(all(inj %in% a$flagged))
  # the subset size follows the coverage: larger alpha -> larger h
  expect_gt(EFAtools:::.mcd_hsize(n_out, p_out, 0.75),
            EFAtools:::.mcd_hsize(n_out, p_out, 0.5))
})

test_that(".fast_mcd returns a genuine concentration fixed point, never a stopped-early subset", {
  # binary items make many h-subsets (near-)degenerate; the FAST-MCD search must return a
  # stable concentration fixed point (the h observations closest to its centre reproduce its
  # covariance) or abort to the classical fallback. It must never return an unconcentrated
  # (p + 1)-point or stopped-early over-tight scatter, which would silently over-flag.
  ld <- function(S) as.numeric(determinant(S, logarithm = TRUE)$modulus)
  set.seed(4)
  Sig <- 0.5 * matrix(1, 8L, 8L) + diag(0.5, 8L)
  ok <- replicate(20L, {
    B <- (matrix(rnorm(50L * 8L), 50L, 8L) %*% chol(Sig) > 0) + 0    # binary items
    B <- B[, apply(B, 2L, function(col) length(unique(col)) > 1L), drop = FALSE]
    h <- EFAtools:::.mcd_hsize(nrow(B), ncol(B), 0.5)
    fit <- tryCatch(EFAtools:::.fast_mcd(B, h), error = function(e) NULL)
    if (is.null(fit)) return(TRUE)                      # aborted -> classical fallback: fine
    d2 <- stats::mahalanobis(B, fit$center, fit$cov)
    idx <- order(d2)[seq_len(h)]
    abs(ld(stats::cov(B[idx, , drop = FALSE])) - ld(fit$cov)) < 0.1
  })
  expect_true(all(ok))
})

test_that("the robust covariance cross-checks against MASS::cov.rob", {
  skip_on_cran()
  skip_if_not_installed("MASS")

  h <- floor((n_out + p_out + 1) / 2)
  # MASS::cov.rob(method = "mcd") returns a *reweighted* MCD estimate (with its own
  # consistency scaling), so the comparable quantity is efa_screen's reweighted robust
  # location/scatter; they agree closely on this well-separated fixture
  set.seed(123)
  mm <- MASS::cov.rob(X_out, method = "mcd", quantile.used = h, nsamp = "best")
  o <- scr_out$outliers
  expect_equal(unname(o$center), unname(mm$center), tolerance = 0.1)
  expect_equal(cov2cor(o$cov), cov2cor(mm$cov), tolerance = 0.05, ignore_attr = TRUE)

  # the raw MCD subset location (before the finite-sample correction) is a robust centre
  # in the same region
  set.seed(1)
  fm <- EFAtools:::.fast_mcd(X_out, h, nsamp = 500L)
  expect_equal(unname(fm$center), unname(mm$center), tolerance = 0.25)
})

test_that("the MCD subset size, consistency, and small-sample factors match their forms", {
  # subset size: floor((n + p + 1) / 2) at alpha = 0.5 (max breakdown), rising to n at 1
  expect_equal(EFAtools:::.mcd_hsize(200, 4, 0.5), floor((200 + 4 + 1) / 2))
  expect_equal(EFAtools:::.mcd_hsize(200, 4, 1), 200)
  # consistency factor: alpha / F_{chi^2_{p + 2}}(qchisq(alpha, p))
  expect_equal(EFAtools:::.mcd_consistency(4, 0.5),
               0.5 / pchisq(qchisq(0.5, 4), 4 + 2), tolerance = 1e-12)
  # the small-sample correction inflates the scatter at small n and tends to 1 as n grows
  expect_gt(EFAtools:::.mcd_cnp2(4, 40, 0.5, reweighted = FALSE), 1)
  expect_equal(EFAtools:::.mcd_cnp2(4, 1e6, 0.5, reweighted = FALSE), 1, tolerance = 1e-3)
  # regression pins for both correction tables (raw and reweighted, all three p branches),
  # locking the Pison et al. (2002) coefficients used to scale the robust covariance
  expect_equal(EFAtools:::.mcd_cnp2(4, 40, 0.5, reweighted = FALSE), 1.26016971771, tolerance = 1e-8)
  expect_equal(EFAtools:::.mcd_cnp2(2, 50, 0.5, reweighted = FALSE), 1.15095829889, tolerance = 1e-8)
  expect_equal(EFAtools:::.mcd_cnp2(1, 50, 0.5, reweighted = FALSE), 1.13894209891, tolerance = 1e-8)
  expect_equal(EFAtools:::.mcd_cnp2(4, 40, 0.5, reweighted = TRUE),  1.06512515802, tolerance = 1e-8)
  expect_equal(EFAtools:::.mcd_cnp2(2, 50, 0.5, reweighted = TRUE),  1.01272751567, tolerance = 1e-8)
  expect_equal(EFAtools:::.mcd_cnp2(1, 50, 0.5, reweighted = TRUE),  1.00806218791, tolerance = 1e-8)
})

test_that("outlier diagnostics fall back to classical distances and degrade gracefully", {
  # n <= 2p: the robust covariance is undefined, so classical Mahalanobis is used
  set.seed(5)
  Xsmall <- matrix(rnorm(8 * 4), 8, 4)
  expect_warning(
    o <- EFAtools:::.screen_outliers(Xsmall, mcd_alpha = 0.5, outlier_cutoff = 0.975,
                                     seed = 1),
    class = "efa_screen_mcd_fallback"
  )
  expect_equal(o$method, "classical")
  expect_length(o$distances, 8L)
  # the recorded reason names the branch actually taken: n <= 2p is decided before
  # .fast_mcd() runs, so this case is "too few complete cases", never an exact fit
  expect_match(o$fallback_reason, "too few complete cases")

  # collinear variables with too few cases: even the classical covariance is singular,
  # so the diagnostic is skipped with a classed note
  set.seed(6)
  Xd <- matrix(rnorm(8 * 4), 8, 4)
  Xd[, 4] <- Xd[, 1] + Xd[, 2]
  expect_warning(
    o2 <- EFAtools:::.screen_outliers(Xd, mcd_alpha = 0.5, outlier_cutoff = 0.975,
                                      seed = 1),
    class = "efa_screen_no_outliers"
  )
  expect_s3_class(o2, "efa_screen_no_outliers")
  # each of the three ways the complete-case covariance can fail names itself, so the
  # report never sends a reader after collinearity when the cause is missingness or a
  # non-finite value
  expect_match(o2$reason, "linearly dependent")

  set.seed(6)
  Xnf <- matrix(rnorm(8 * 4), 8, 4)
  Xnf[3, 2] <- Inf
  expect_warning(
    o4 <- EFAtools:::.screen_outliers(Xnf, mcd_alpha = 0.5, outlier_cutoff = 0.975,
                                      seed = 1),
    class = "efa_screen_no_outliers"
  )
  expect_match(o4$reason, "not finite")

  X0 <- matrix(rnorm(5 * 4), 5, 4)
  X0[, 1] <- NA                              # no complete case at all
  expect_warning(
    o5 <- EFAtools:::.screen_outliers(X0, mcd_alpha = 0.5, outlier_cutoff = 0.975,
                                      seed = 1),
    class = "efa_screen_no_outliers"
  )
  expect_match(o5$reason, "too few complete cases")

  # near-collinear variables are skipped on the same terms as exactly collinear ones:
  # the covariance is numerically singular (rcond far below the double-precision epsilon)
  # even though it is positive definite, so it must not reach the distance step
  set.seed(6)
  Xn <- matrix(rnorm(8 * 4), 8, 4)
  Xn[, 4] <- Xn[, 1] + Xn[, 2] + 1e-9 * rnorm(8)
  expect_warning(
    o3 <- EFAtools:::.screen_outliers(Xn, mcd_alpha = 0.5, outlier_cutoff = 0.975,
                                      seed = 1),
    class = "efa_screen_no_outliers"
  )
  expect_s3_class(o3, "efa_screen_no_outliers")

  # a correlation matrix reaching the helper aborts
  expect_error(
    EFAtools:::.screen_outliers(test_models$baseline$cormat, mcd_alpha = 0.5,
                                outlier_cutoff = 0.975),
    class = "efa_screen_outliers_cormat"
  )
})

test_that("tied responses on coarse items fall back to classical distances (GRiPS)", {
  # GRiPS items are 6-point and heavily tied: more than h respondents give the same answer
  # on at least one item pair, so an h-subset sits exactly on the hyperplane x_i = x_j and
  # no robust MCD covariance exists. The outlier diagnostic then falls back to classical
  # Mahalanobis distances with a warning. A covering subset of determinant zero is the
  # smallest one there can be, so which random subsets the search happens to draw cannot
  # change that conclusion: the verdict, and the whole diagnostic behind it, must come out
  # the same at every seed rather than depending on one.
  seeds <- c(1L, 2L, 4L, 7L, 12L)
  scrs <- lapply(seeds, function(s) {
    expect_warning(scr <- efa_screen(GRiPS_raw, seed = s), class = "efa_screen_mcd_fallback")
    scr
  })
  outs <- lapply(scrs, `[[`, "outliers")
  expect_true(all(vapply(outs, function(o) o$method, character(1)) == "classical"))
  expect_true(all(vapply(outs, function(o) identical(o$flagged, outs[[1]]$flagged),
                         logical(1))))

  scr <- scrs[[1]]
  expect_length(scr$outliers$distances, nrow(GRiPS_raw))
  expect_equal(scr$outliers$cutoff, sqrt(qchisq(0.975, ncol(GRiPS_raw))))

  # with n = 810 well above 2p = 16 the cause cannot be too few complete cases, and the
  # correlation matrix is well conditioned, so the reported reason must be the exact fit
  expect_gt(nrow(GRiPS_raw), 2 * ncol(GRiPS_raw))
  expect_lt(sqrt(scr$condition), 10)
  expect_match(scr$outliers$fallback_reason, "exact fit")
})

test_that("the outlier diagnostic does not depend on the variables' units", {
  # The Mahalanobis distance, and the MCD location and scatter behind it, are affine
  # equivariant: recording a variable in cents rather than in euros cannot make anyone an
  # outlier who was not one before. Multiplying one column by 100 - a narrower spread of
  # units than an income variable beside a Likert item - must therefore leave the verdict,
  # the flagged rows and the distances where they were, and must carry the reported centre
  # and scatter into the new units exactly.
  set.seed(1809)
  X <- matrix(stats::rnorm(300 * 4), 300, 4)
  X[1:6, ] <- X[1:6, ] + 5                      # a block of genuine outliers
  colnames(X) <- paste0("V", seq_len(4))
  d <- c(100, 1, 1, 1)
  Xs <- sweep(X, 2L, d, "*", check.margin = FALSE)

  o <- efa_screen(X, seed = 1)$outliers
  os <- efa_screen(Xs, seed = 1)$outliers

  expect_equal(o$method, "mcd")
  expect_equal(os$method, "mcd")
  expect_identical(os$flagged, o$flagged)
  expect_equal(os$distances, o$distances, tolerance = 1e-8)
  expect_equal(os$center, o$center * d, tolerance = 1e-8)
  expect_equal(os$cov, o$cov * outer(d, d), tolerance = 1e-8)

  # a scale far enough apart to make the scatter singular once it is back in the supplied
  # units: the distances have to be taken where the conditioning was checked, or the
  # inversion inside mahalanobis() refuses a scatter that was passed as usable
  dw <- c(1e9, 1, 1, 1)
  ow <- efa_screen(sweep(X, 2L, dw, "*", check.margin = FALSE), seed = 1)$outliers
  expect_equal(ow$method, "mcd")
  expect_identical(ow$flagged, o$flagged)

  # the rescaling is taken from the complete cases, so an incomplete row cannot leave a
  # column unscaled - the median absolute deviation of a column holding a missing value is
  # itself missing - and quietly put the diagnostic back at the mercy of the units
  Xm <- Xs
  Xm[c(50, 120, 200), 1] <- NA
  expect_equal(efa_screen(Xm, seed = 1)$outliers$method, "mcd")
})

test_that("a near-collinear fallback is reported as collinearity, not as an exact fit", {
  # The robust scatter can be unusable for reasons that send a reader to quite different
  # places: near-collinear variables are dealt with by dropping a redundant item, while an
  # exact fit is a property of a covering subset of the respondents (tied answers on coarse
  # items) with nothing wrong with the variables at all. Here the fourth variable is a
  # near-exact sum of the first two, so the complete-case covariance is itself
  # ill-conditioned and the recorded reason has to say so.
  set.seed(12)
  Xc <- matrix(stats::rnorm(300 * 4), 300, 4)
  Xc[, 4] <- Xc[, 1] + Xc[, 2] + 0.01 * stats::rnorm(300)
  expect_warning(
    o <- EFAtools:::.screen_outliers(Xc, mcd_alpha = 0.5, outlier_cutoff = 0.975, seed = 1),
    class = "efa_screen_mcd_fallback"
  )
  expect_equal(o$method, "classical")
  expect_match(o$fallback_reason, "ill-conditioned")
  expect_false(grepl("exact fit", o$fallback_reason, fixed = TRUE))
})

test_that("outlier_cutoff is bounded to the range in which it defines a cutoff", {
  # 0 puts the threshold at zero (everything flagged) and 1 at infinity (nothing
  # flagged); neither is an outlier diagnostic. The argument check runs before anything
  # is computed, so the cheap correlation-matrix input reaches it - and pinning both
  # endpoints against the values just outside them fixes where the bound sits, which the
  # error class alone (a checkmate assertion carries no package class) cannot do
  R <- test_models$baseline$cormat
  expect_no_error(efa_screen(R, N = 500, outlier_cutoff = 0.5))
  expect_no_error(efa_screen(R, N = 500, outlier_cutoff = 0.9999))
  expect_error(efa_screen(R, N = 500, outlier_cutoff = 0.4999), class = "simpleError")
  expect_error(efa_screen(R, N = 500, outlier_cutoff = 1), class = "simpleError")
  expect_error(efa_screen(R, N = 500, outlier_cutoff = 0), class = "simpleError")
})

test_that("fewer than three variables is refused rather than reported on", {
  # At p = 1 the KMO is 0/0 and Bartlett's test has no degrees of freedom; at p = 2 the
  # KMO is algebraically .5 for every correlation. Both produced a confident report.
  expect_error(efa_screen(data.frame(a = rnorm(50))),
               class = "efa_screen_too_few_vars")
  expect_error(efa_screen(data.frame(a = rnorm(50), b = rnorm(50))),
               class = "efa_screen_too_few_vars")
  expect_error(efa_screen(matrix(1, 1, 1), N = 100),
               class = "efa_screen_too_few_vars")
  expect_error(efa_screen(diag(2), N = 100), class = "efa_screen_too_few_vars")
  # Three variables is the first usable width and still runs.
  expect_s3_class(efa_screen(test_models$baseline$cormat[1:3, 1:3], N = 500),
                  "efa_screen")
})

test_that("an exactly singular matrix names the redundancy that caused it", {
  # The abort is deliberate -- every measure here is built from R^-1 -- but the user
  # screening for redundant items is exactly the one who cannot get a report, so the
  # condition carries the culprit pairs as data (`$pairs`), which is what is asserted
  # here rather than the wording that reports them.
  set.seed(4)
  dup <- data.frame(a = rnorm(120), b = rnorm(120), c = rnorm(120))
  dup$copy_of_a <- dup$a
  expect_error(efa_screen(dup), class = "efa_cor_singular")
  cnd <- tryCatch(efa_screen(dup), efa_cor_singular = function(e) e)
  expect_identical(cnd$pairs, "a-copy_of_a")
  # The abort is attributed to efa_screen(), not to the internal handler that builds it.
  expect_identical(rlang::call_name(conditionCall(cnd)), "efa_screen")

  # A correlation-matrix input takes the same route, and an unnamed one falls back to
  # the V-names used elsewhere in the report.
  R_dup <- stats::cor(dup)
  expect_error(efa_screen(R_dup, N = 120), class = "efa_cor_singular")
  R_un <- unname(R_dup)
  cnd_un <- tryCatch(efa_screen(R_un, N = 120), efa_cor_singular = function(e) e)
  expect_identical(cnd_un$pairs, "V1-V4")

  # Rank deficiency without a perfect pair (a sum score) has no pair to name, so it
  # reports none rather than an empty pair label.
  cnd_sum <- tryCatch(efa_screen(sing_raw), efa_cor_singular = function(e) e)
  expect_identical(cnd_sum$pairs, character())
})

test_that("a near-singular matrix still produces the full report", {
  # The distinction is the point: perturbing a duplicate by 1e-7 is statistically the
  # same data, and it must come back as a report with multicollinearity flags rather
  # than as the singular abort.
  set.seed(5)
  near <- data.frame(a = rnorm(120), b = rnorm(120), c = rnorm(120))
  near$copy_of_a <- near$a + rnorm(120, sd = 1e-7)
  scr <- suppressWarnings(efa_screen(near, seed = 1))
  expect_s3_class(scr, "efa_screen")
  expect_lt(scr$determinant, 1e-10)
  expect_gt(scr$condition, 30)
})

test_that("the multicollinearity verdict follows the condition index, not the determinant", {
  local_reproducible_output()

  # A clean one-factor population with every loading 0.5: every off-diagonal correlation
  # is exactly 0.25 and the smallest eigenvalue exactly 0.75 at every p, so this pool is
  # no worse conditioned at 60 variables than at 10 (index 4.58). The determinant is the
  # product of the p eigenvalues, so it still falls below the 0.00001 that is commonly
  # quoted - a cut-off on it flags a pool with nothing wrong with it.
  p <- 60
  L <- matrix(0.5, p, 1)
  R60 <- L %*% t(L)
  diag(R60) <- 1
  dimnames(R60) <- list(paste0("v", seq_len(p)), paste0("v", seq_len(p)))
  scr_wide <- efa_screen(R60, N = 500)
  expect_lt(scr_wide$determinant, 1e-5)
  expect_lt(sqrt(scr_wide$condition), 10)

  txt <- cli::ansi_strip(paste(format(scr_wide), collapse = " "))
  # the determinant is still reported as a number, but carries no verdict of its own
  expect_match(txt, "Determinant: 0.0000006", fixed = TRUE)
  expect_match(txt, "so the condition index below carries the verdict", fixed = TRUE)
  expect_no_match(txt, "Multicollinearity likely", fixed = TRUE)
  # and no recommendation sends the reader after redundant items that do not exist
  expect_no_match(txt, "nearly linearly dependent", fixed = TRUE)

  # A genuinely ill-conditioned matrix still gets the verdict and the recommendation. A
  # duplicate perturbed by 1e-7 puts the index near 2e7, so far above the highest step of
  # the progression (300) that the graded label cannot move with the BLAS.
  set.seed(5)
  near2 <- data.frame(a = rnorm(120), b = rnorm(120), c = rnorm(120))
  near2$copy_of_a <- near2$a + rnorm(120, sd = 1e-7)
  scr_ill <- suppressWarnings(efa_screen(near2, seed = 1))
  expect_gt(sqrt(scr_ill$condition), 300)
  txt_ill <- cli::ansi_strip(paste(format(scr_ill), collapse = " "))
  expect_match(txt_ill, "which indicates a near linear dependency", fixed = TRUE)
  expect_match(txt_ill, "its relative strength is very strong", fixed = TRUE)
  expect_match(txt_ill, "A high condition index indicates multicollinearity", fixed = TRUE)
})

test_that("the condition-index verdict is graded on Belsley's progression", {
  local_reproducible_output()

  # An equicorrelated matrix of three variables has eigenvalues 1 + 2r and 1 - r (twice),
  # so its condition index is exactly sqrt((1 + 2r) / (1 - r)). Each r below puts the
  # index far from the band edges (10, 30, 100, 300), so the expected verdict cannot move
  # with the BLAS: the four indexes are 17.3, 44.7, 141.4, and 447.2.
  eqcor <- function(r) {
    R <- matrix(r, 3, 3)
    diag(R) <- 1
    dimnames(R) <- list(paste0("v", 1:3), paste0("v", 1:3))
    R
  }
  screened <- function(r) {
    scr <- efa_screen(eqcor(r), N = 500)
    expect_equal(sqrt(scr$condition), sqrt((1 + 2 * r) / (1 - r)), tolerance = 1e-6,
                 label = paste0("condition index at r = ", r))
    cli::ansi_strip(paste(format(scr), collapse = " "))
  }

  # Above 10 but not above the 30 that flags a near dependency: no flag, no recommendation.
  mid <- screened(0.99)
  expect_match(mid, "The index is above 10, but not above 30", fixed = TRUE)
  expect_no_match(mid, "A high condition index indicates", fixed = TRUE)

  # Flagged, and graded by the position on the progression the index sits at.
  for (case in list(list(r = 0.99850,  strength = "moderate"),
                    list(r = 0.99985,  strength = "strong"),
                    list(r = 0.999985, strength = "very strong"))) {
    txt <- screened(case$r)
    expect_match(txt, paste("its relative strength is", case$strength), fixed = TRUE)
    expect_match(txt, "A high condition index indicates", fixed = TRUE)
  }
})

# singular / non-positive-definite inputs
test_that("errors and warnings are thrown correctly", {
  # The rank-deficient raw data and its correlation matrix (`sing_raw` / `sing_cor`, with
  # `sing_N` rows) and the invertible-but-indefinite `cor_nposdef` come from
  # helper-singular.R, so both branches are exercised on the same inputs everywhere.
  expect_error(efa_screen(1:5), class = "efa_input_not_matrix")
  expect_error(efa_screen(sing_raw), class = "efa_cor_singular")
  expect_error(efa_screen(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_warning(efa_screen(cor_nposdef, N = 10), class = "efa_cor_smoothed")
})

test_that("a flag rate far above the nominal one is reported as a calibration issue", {
  # iris is three species clusters, so it is not elliptically distributed: a high-breakdown
  # estimate fitted to the most concentrated half legitimately calls a large share of the
  # rest distant. Far more flags than the 2.5% cutoff admits is evidence of that, not a list
  # of contaminated cases to work through.
  recs <- EFAtools:::.screen_recommendations(scr_iris, raw = TRUE,
                                             kmo = scr_iris$kmo$KMO, bart_sig = TRUE,
                                             mvn_nonnormal = TRUE, multicollinear = FALSE,
                                             digits = 3)
  nf <- length(scr_iris$outliers$flagged)
  expect_gt(nf / scr_iris$outliers$n_complete, 5 * (1 - scr_iris$settings$outlier_cutoff))
  expect_true(any(grepl("not elliptically distributed", recs$message, fixed = TRUE)))
  expect_false(any(grepl("before down-weighting or excluding", recs$message, fixed = TRUE)))

  # a rate in line with the cutoff keeps the ordinary inspect-them advice
  recs_ok <- EFAtools:::.screen_recommendations(scr_out, raw = TRUE,
                                                kmo = scr_out$kmo$KMO, bart_sig = TRUE,
                                                mvn_nonnormal = FALSE, multicollinear = FALSE,
                                                digits = 3)
  expect_true(any(grepl("before down-weighting or excluding", recs_ok$message, fixed = TRUE)))

  # the rate alone does not license the conclusion: a short flagged list is a list to
  # inspect however far above the nominal rate it sits (here 3 of 20 = 15% against 2.5%)
  few <- scr_iris
  few$outliers$flagged <- scr_iris$outliers$flagged[1:3]
  few$outliers$n_complete <- 20L
  recs_few <- EFAtools:::.screen_recommendations(few, raw = TRUE, kmo = few$kmo$KMO,
                                                 bart_sig = TRUE, mvn_nonnormal = TRUE,
                                                 multicollinear = FALSE, digits = 3)
  expect_gt(3 / 20, 5 * (1 - few$settings$outlier_cutoff))
  expect_true(any(grepl("before down-weighting or excluding", recs_few$message, fixed = TRUE)))
  expect_false(any(grepl("not elliptically distributed", recs_few$message, fixed = TRUE)))
})

test_that("the per-variable display marks the missing percentage and hides an empty flags column", {
  # `missing` is a percentage; the display says so
  expect_true("missing%" %in% names(EFAtools:::.screen_per_item_display(scr_raw, 3)))

  # all-continuous data: every flags entry is NA ("no category screening applies"), which
  # prints as a column of <NA>, so the column is dropped instead
  expect_true(all(is.na(scr_iris$per_item$flags)))
  expect_false("flags" %in% names(EFAtools:::.screen_per_item_display(scr_iris, 3)))

  # with some categorical variables the column stays and the NAs render as a dash
  set.seed(12)
  dat <- data.frame(cont = rnorm(80), cat = sample(1:5, 80, replace = TRUE),
                    v3 = rnorm(80))
  scr <- efa_screen(dat, seed = 1)
  disp <- EFAtools:::.screen_per_item_display(scr, 3)
  expect_true("flags" %in% names(disp))
  expect_equal(disp["cont", "flags"], "-")
})

test_that("the complete-case denominator is named when rows are incomplete", {
  # The normality tests and the outlier diagnostic are computed from the complete cases,
  # which under non-ignorable missingness are not a random subsample of the rows
  # supplied. Both bases therefore have to appear, so neither the normality verdict nor
  # the outlier rate can be read against the wrong one.
  local_reproducible_output(width = 200)
  set.seed(31)
  n3 <- 240L
  S3 <- matrix(0.4, p_out, p_out)
  diag(S3) <- 1
  X3 <- matrix(rnorm(n3 * p_out), n3, p_out) %*% chol(S3)
  colnames(X3) <- paste0("v", seq_len(p_out))
  X3[1:60, 1] <- NA
  scr_na <- efa_screen(X3, seed = 1)
  expect_equal(scr_na$settings$n_obs, 240L)
  expect_equal(scr_na$normality$n_complete, 180L)

  txt <- cli::ansi_strip(paste(format(scr_na), collapse = " "))
  expect_match(txt, "Computed from 180 complete cases of the 240 rows supplied",
               fixed = TRUE)
  expect_match(txt, "of 180 complete cases of the 240 rows supplied", fixed = TRUE)

  # With no missing values the two bases coincide, so the report stays as it was.
  scr_full <- efa_screen(X3[stats::complete.cases(X3), ], seed = 1)
  txt_full <- cli::ansi_strip(paste(format(scr_full), collapse = " "))
  expect_no_match(txt_full, "rows supplied", fixed = TRUE)
  expect_match(txt_full, "of 180 observations", fixed = TRUE)
})

test_that("print output is stable", {
  local_reproducible_output()

  # correlation-matrix input: KMO + Bartlett verdicts, multicollinearity, and the
  # per-variable MSA/SMC table, closing with the raw-data note (no raw-only sections)
  expect_snapshot(print(scr_cor), transform = scrub_num)

  # raw input: the full report incl. the per-item table, multivariate normality, the
  # classical-fallback outlier line, and the recommendations block (the non-normality,
  # over-powered-Bartlett caveat, sparse-category, and outlier recommendations)
  expect_snapshot(print(scr_raw), transform = scrub_num)

  # correlation-matrix input without N: Bartlett's test is reported as not computed
  expect_snapshot(print(scr_non), transform = scrub_num)
})

test_that("the report tracks a narrow console", {
  # Every other print snapshot in the package is recorded at the default width 80, so
  # nothing pins the behaviour of the report's advisory lines once the console is
  # narrower than they are. efa_screen() is the widest report in the package (banded
  # verdicts with a parenthetical rule attached to each), so it is the one that has to
  # reflow rather than overflow.
  local_reproducible_output(width = 60)

  expect_snapshot(print(scr_cor), transform = scrub_num)
})

rm(scr_cor, scr_raw, scr_iris, scr_nona, scr_non, dat_nonames, scr_out, X_out, Sig_out,
   n_out, p_out, inj)

#' Kaiser-Guttman criterion
#'
#' Probably the most popular factor retention criterion. Kaiser and Guttman suggested
#' to retain as many factors as there are sample eigenvalues greater than 1.
#' This is why the criterion is also known as eigenvalues-greater-than-one rule.
#'
#' @param x data.frame or matrix. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param eigen_type character. On what the eigenvalues should be found. Can be
#'  either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
#'  the diagonal values of the correlation matrices are left to be 1. If using
#'  "SMC", the diagonal of the
#'  correlation matrices is replaced by the squared multiple correlations (SMCs)
#'  of the indicators. If using "EFA", eigenvalues are found on the correlation
#'  matrices with the final communalities of an exploratory factor analysis
#'  solution (default is principal axis factoring extracting 1 factor) as
#'  diagonal. Default is `c("PCA", "SMC", "EFA")`, i.e. all three; `"EFA"` is the
#'  only one that fits a model.
#' @param use character. Passed to [stats::cor()] if raw
#'  data is given as input. Default is "pairwise.complete.obs".
#' @param cor_method character. Correlation computed from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (passed to [stats::cor()]), or `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations of ordinal / binary data
#'   (a two-step estimator).
#' Default is "pearson".
#' @param n_factors numeric. Number of factors to extract if "EFA" is included in
#' `eigen_type`. Default is 1.
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  [efa_fit()] fit that provides the communalities when `"EFA"` is included in `eigen_type`.
#'  `NULL` (default) uses the [efa_fit()] defaults. The fit is unrotated, so no rotation settings
#'  apply.
#' @param ... Additional arguments passed to [efa_fit()]. For example,
#' `estimator`, to change the estimator (PAF is default). The estimation tuning knobs are not
#' passed here; they live in `estimate_control`, and the standard-error arguments (`se`,
#' `b_boot`, `ci`, `seed`) are not accepted because the fit is an internal step that keeps
#' only its communalities.
#'
#' @details Originally, the Kaiser-Guttman criterion was intended for the use
#' with principal components, hence with eigenvalues derived from the original
#' correlation matrix. This can be done here by setting `eigen_type` to
#' "PCA". However, it is well-known that this criterion is often inaccurate and
#' that it tends to overestimate the number of factors, especially for unidimensional
#' or orthogonal factor structures (e.g., Zwick & Velicer, 1986).
#'
#' The criterion's inaccuracy in these cases is somewhat addressed if it is
#' applied on the correlation matrix with communalities in the diagonal, either
#' initial communalities estimated from SMCs (done setting `eigen_type` to
#' "SMC") or final communality estimates from an EFA (done setting `eigen_type`
#' to "EFA"; see Auerswald & Moshagen, 2019). However, although this variant
#' of the KGC is more accurate in some cases compared to the traditional KGC, it
#' is at the same time less accurate than the PCA-variant in other cases, and it
#' is still often less accurate than several of the other criteria available here,
#' such as parallel analysis ([efa_parallel()]), the Hull method ([efa_hull()]),
#' the empirical Kaiser criterion ([efa_ekc()]), or sequential \eqn{chi^2} model
#' tests ([efa_smt()]; see Auerswald & Moshagen, 2019). Which criteria are
#' informative depends on the data at hand, so rather than substituting one for
#' another, run several of them together and compare their suggestions.
#'
#' The `efa_kgc` function can also be called together with other factor
#' retention criteria in the [efa_retain()] function.
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). Its main fields are:
#' \item{n_factors}{A named numeric vector with the suggested number of factors
#'   for each requested eigenvalue type (`"PCA"`, `"SMC"`, and/or `"EFA"`).}
#' \item{results}{A list with one record per eigenvalue type, each holding the
#'   eigenvalues and the retained solution used for printing and plotting.}
#' \item{settings}{A list of the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Guttman, L. (1954). Some necessary conditions for common-factor analysis.
#' Psychometrika, 19, 149 –161. https://doi.org/10.1007/BF02289162
#'
#' @source Kaiser, H. F. (1960). The application of electronic computers to factor
#' analysis. Educational and Psychological Measurement, 20, 141–151.
#' https://doi.org/10.1177/001316446002000116
#'
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. Psychological Bulletin, 99,
#' 432–442. https://doi.org/10.1037/0033-2909.99.3.432
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#' @examples
#' efa_kgc(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
efa_kgc <- function(x, eigen_type = c("PCA", "SMC", "EFA"),
                use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                        "everything", "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"), n_factors = 1,
                estimate_control = NULL,
                ...){

  # Perform argument checks
  .reject_flat_knobs(...names(), fn = "efa_kgc")
  .reject_unknown_fit_dots(...names(), fn = "efa_kgc", unrotated = TRUE)
  .reject_rotation_dots(list(...), fn = "efa_kgc")
  .assert_cor_input(x)

  eigen_type <- .match_arg_ci(eigen_type, several.ok = TRUE)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  checkmate::assert_count(n_factors)
  .assert_estimate_control(estimate_control)

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, use = use, cor_method = cor_method,
                             N_policy = "none")
  R <- prep$R


  # Calculate the PCA / SMC / EFA eigenvalues for the requested types
  eigen_list <- .three_eigen(R, eigen_type, n_factors = n_factors,
                             estimate_control = estimate_control, ...)

  # Kaiser-Guttman: retain as many factors as there are eigenvalues greater than
  # 1 (NA for any eigenvalue type that was not requested)
  nfac_list <- lapply(eigen_list, function(eig) {
    if (length(eig) == 1L && is.na(eig)) NA else sum(eig > 1)
  })

  # one record per requested eigenvalue type (eigenvalues with the > 1 rule)
  results <- list()
  for (et in c("PCA", "SMC", "EFA")) {
    if (!(et %in% eigen_type)) next
    eig <- eigen_list[[et]]
    n_fac <- nfac_list[[et]]
    results[[et]] <- list(
      name = et,
      label = et,
      n_factors = n_fac,
      plot_type = "eigen",
      x = seq_along(eig),
      y = eig,
      reference = NULL,
      threshold = 1,
      highlight = if (!is.na(n_fac) && n_fac >= 1) n_fac else NULL
    )
  }

  output <- .new_efa_retention(
    "KGC",
    results = unname(results),
    settings = list(eigen_type = eigen_type, use = use,
                    cor_method = cor_method, n_factors = n_factors),
    subtitle = .eigen_subtitle(eigen_type)
  )

  return(output)

}

kgc_cor <- efa_kgc(test_models$baseline$cormat)
kgc_cor_smc <- efa_kgc(test_models$baseline$cormat, eigen_type = "SMC")
kgc_raw <- efa_kgc(GRiPS_raw)
# Check with an argument passed to "EFA"
kgc_efa_ml<- efa_kgc(test_models$baseline$cormat, eigen_type = "EFA", estimator = "ML")

test_that("output class and dimensions are correct", {
  expect_s3_class(kgc_cor, "efa_retention")
  expect_length(kgc_cor, 6)
  expect_s3_class(kgc_cor_smc, "efa_retention")
  expect_length(kgc_cor_smc, 6)
  expect_s3_class(kgc_raw, "efa_retention")
  expect_s3_class(kgc_efa_ml, "efa_retention")

  expect_named(kgc_cor$n_factors, c("PCA", "SMC", "EFA"))
  expect_named(kgc_cor_smc$n_factors, "SMC")
  expect_named(kgc_efa_ml$n_factors, "EFA")
  expect_equal(.retention_record(kgc_cor, "PCA")$plot_type, "eigen")
  expect_equal(.retention_record(kgc_cor, "PCA")$threshold, 1)
})

test_that("found eigenvalues are correct", {
  expect_equal(sum(.retention_record(kgc_cor, "PCA")$y),
               ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(kgc_cor, "SMC")$y),
            ncol(test_models$baseline$cormat))
  expect_lt(sum(.retention_record(kgc_cor, "EFA")$y),
            ncol(test_models$baseline$cormat))

  expect_equal(sum(.retention_record(kgc_raw, "PCA")$y), ncol(GRiPS_raw))
  expect_lt(sum(.retention_record(kgc_raw, "SMC")$y), ncol(GRiPS_raw))
  expect_lt(sum(.retention_record(kgc_raw, "EFA")$y), ncol(GRiPS_raw))

  # Only the requested eigenvalue type produces a record
  expect_null(.retention_record(kgc_cor_smc, "PCA"))
  expect_null(.retention_record(kgc_cor_smc, "EFA"))
  expect_null(.retention_record(kgc_efa_ml, "PCA"))
  expect_null(.retention_record(kgc_efa_ml, "SMC"))
})

test_that("identified number of factors is correct", {
  expect_equal(kgc_cor$n_factors[["PCA"]], 3)
  expect_equal(kgc_cor$n_factors[["SMC"]], 1)
  expect_equal(kgc_cor$n_factors[["EFA"]], 1)

  expect_equal(kgc_raw$n_factors[["PCA"]], 1)
  expect_equal(kgc_raw$n_factors[["SMC"]], 1)
  expect_equal(kgc_raw$n_factors[["EFA"]], 1)

  expect_equal(kgc_cor_smc$n_factors[["SMC"]], 1)
  expect_false(any(c("PCA", "EFA") %in% names(kgc_cor_smc$n_factors)))

  expect_equal(kgc_efa_ml$n_factors[["EFA"]], 1)
  expect_false(any(c("PCA", "SMC") %in% names(kgc_efa_ml$n_factors)))
})

test_that("the PCA suggestion matches nFactors::nScree", {
  skip_if_not_installed("nFactors")
  # nScree's nkaiser is the eigenvalues-greater-than-one count on the unaltered
  # correlation matrix, i.e. the PCA variant of the Kaiser-Guttman criterion.
  for (cmat in list(test_models$baseline$cormat, stats::cor(GRiPS_raw))) {
    expect_equal(
      efa_kgc(cmat, eigen_type = "PCA")$n_factors[["PCA"]],
      nFactors::nScree(cmat, model = "components")$Components$nkaiser
    )
  }
})

test_that("errors are thrown correctly", {
  expect_error(efa_kgc(1:5), class = "efa_input_not_matrix")
  expect_message(efa_kgc(GRiPS_raw, eigen_type = "PCA"), class = "efa_cor_from_data")
  expect_error(efa_kgc(sing_raw), class = "efa_cor_singular")
  expect_error(efa_kgc(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_warning(efa_kgc(cor_nposdef, N = 10), class = "efa_cor_smoothed")
})

test_that("settings are returned correctly", {
  expect_named(kgc_cor$settings, c("eigen_type", "use", "cor_method", "n_factors"))
  expect_named(kgc_raw$settings, c("eigen_type", "use", "cor_method", "n_factors"))
  expect_named(kgc_cor_smc$settings, c("eigen_type", "use", "cor_method", "n_factors"))
  expect_named(kgc_efa_ml$settings, c("eigen_type", "use", "cor_method", "n_factors"))

  expect_equal(kgc_cor$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(kgc_raw$settings$eigen_type, c("PCA", "SMC", "EFA"))
  expect_equal(kgc_cor_smc$settings$eigen_type, "SMC")
  expect_equal(kgc_efa_ml$settings$eigen_type, "EFA")

  expect_equal(kgc_cor$settings$use, "pairwise.complete.obs")
  expect_equal(kgc_raw$settings$use, "pairwise.complete.obs")
  expect_equal(kgc_cor_smc$settings$use, "pairwise.complete.obs")
  expect_equal(kgc_efa_ml$settings$use, "pairwise.complete.obs")

  expect_equal(kgc_cor$settings$cor_method, "pearson")
  expect_equal(kgc_raw$settings$cor_method, "pearson")
  expect_equal(kgc_cor_smc$settings$cor_method, "pearson")
  expect_equal(kgc_efa_ml$settings$cor_method, "pearson")

  expect_equal(kgc_cor$settings$n_factors, 1)
  expect_equal(kgc_raw$settings$n_factors, 1)
  expect_equal(kgc_cor_smc$settings$n_factors, 1)
  expect_equal(kgc_efa_ml$settings$n_factors, 1)

})

rm(kgc_cor, kgc_cor_smc, kgc_raw, kgc_efa_ml)

#' Plot efa_compare object
#'
#' Plot method for the [efa_compare()] function showing the distribution of the
#' absolute differences between the two compared objects as a violin plot with
#' jittered points. Differences above the threshold are highlighted.
#'
#' @param x list. An object of class `efa_compare` (output from the
#'  [efa_compare()] function).
#' @param plot_red numeric or `NULL`. Threshold above which to draw the absolute
#'  differences in red, documented in [efa_compare()]. `NULL` (default) uses the
#'  value recorded in `x$settings`; supplying one overrides it for this plot only,
#'  so the comparison need not be recomputed to redraw it at another threshold.
#' @param ... not used.
#'
#' @returns A ggplot object showing the absolute differences, with differences
#'  above `plot_red` highlighted in red.
#'
#' @importFrom rlang .data
#' @export
#' @method plot efa_compare
#'
#' @examples
#' # A type SPSS EFA to mimick the SPSS implementation
#' EFA_SPSS_5 <- efa_fit(IDS2_R, n_factors = 5,
#'                       estimate_control = estimate_control(type = "SPSS"),
#'                       rotate_control = rotate_control(type = "SPSS"))
#'
#' # A type psych EFA to mimick the psych::fa() implementation
#' EFA_psych_5 <- efa_fit(IDS2_R, n_factors = 5,
#'                        estimate_control = estimate_control(type = "psych"),
#'                        rotate_control = rotate_control(type = "psych"))
#'
#' # compare the two and plot the differences
#' comp <- efa_compare(EFA_SPSS_5$unrot_loadings, EFA_psych_5$unrot_loadings,
#'                     x_labels = c("SPSS", "psych"))
#' plot(comp)
plot.efa_compare <- function(x, plot_red = NULL, ...) {

  diff <- x$diff
  # As for the printed report's display controls, the recorded threshold is a drawing
  # setting: an argument supplied here overrides it for this plot, NULL falls back to it.
  if (is.null(plot_red)) {
    plot_red <- x$settings$plot_red
  } else {
    checkmate::assert_number(plot_red)
  }
  x_labels <- x$settings$x_labels

  if (length(c(diff)) <= 2) {
    cli::cli_abort(
      "{.fun plot.efa_compare} needs more than two differences to plot; this comparison has {length(c(diff))}.",
      class = "efa_compare_too_few_to_plot"
    )
  }

  # prepare variable for plot
  diffs <- as.vector(abs(diff))
  diff_dat <- data.frame(
    diffs = diffs,
    color = ifelse(diffs >= plot_red, "large difference", "acceptable difference"),
    comp = paste(x_labels, collapse = " vs. "),
    stringsAsFactors = FALSE
  )

  ggplot2::ggplot(diff_dat, ggplot2::aes(.data$comp, .data$diffs,
                                         col = .data$color)) +
    ggplot2::geom_violin(col = "grey20", width = .7, linewidth = .7) +
    ggplot2::geom_hline(yintercept = plot_red, lty = 2, alpha = .5,
                        linewidth = 1.25) +
    ggplot2::geom_jitter(alpha = .5, width = 0.05, height = 0, size = 2) +
    ggplot2::scale_color_manual(values = c("acceptable difference" = "black",
                                           "large difference" = "red")) +
    ggplot2::labs(
      subtitle = paste("Threshold for difference coloring:", plot_red),
      x = "Compared Variables",
      y = "Absolute Difference"
    ) +
    .gg_theme() +
    # The subtitle gives the threshold that decides the two point colours, so a legend
    # adds nothing.
    ggplot2::theme(legend.position = "none")

}

# Single estimation core: dispatch to a method fitter, then post-process once.
#
# `.PAF()`, `.ML()`, and `.ULS()` are thin fitters that return only their raw
# results (`L`, `h2`, objective `Fm`, `iter`, `convergence`, the original
# correlation matrix `orig_R`, and the matrix `R_final` whose eigenvalues are the
# final eigenvalues), plus any method-specific extras. `.finalize_fit()` performs
# the shared post-processing common to all methods, and `.estimate_model()`
# assembles the method-specific output object.

# Lower bound the ML and ULS optimisers impose on the uniquenesses. A solution
# whose uniqueness is pinned at this floor is an improper (boundary) solution:
# the unconstrained optimum would place it at or below zero. Shared by the ML/ULS
# fitters and the Heywood detector in .finalize_fit() so the value has one source.
.uniqueness_floor <- 0.005

# Squared multiple correlations (1 - 1/diag(R^-1)) used as starting communalities by the
# estimators. Shared by the PAF, ML, and ULS fitters so the start has one source.
#
# A squared multiple correlation is a squared correlation and so lies in [0, 1], but the
# expression only respects that bound for a positive definite R: on an indefinite or
# ill-conditioned matrix (an unsmoothed bootstrap resample, say) it escapes the range and
# would otherwise be written straight onto a correlation diagonal or handed to the
# optimiser as a start outside its box. Clamp it, as psych::smc() does.
#
# The inverse diagonal comes from a Cholesky factorisation, which uses the symmetry and
# positive definiteness that solve()'s LU ignores. Two cases fall back to solve(): a
# matrix the Cholesky rejects (not positive definite), and one whose implied uniqueness
# 1/diag(R^-1) has underflowed to p^2 * eps, where the Cholesky still succeeds but the
# matrix is singular to working precision. Both then reach solve()'s own singularity
# error, which callers rely on -- .parallel_sim_eig() rejects a simulated draw on it.
# The p^2 factor is deliberately more generous than the p * eps rank tolerance used on
# eigenvalue ratios elsewhere (.prepare_cor_input()): the two quantities are not on the
# same scale, and the asymmetry favours caution, because a needless fallback only repeats
# the calculation solve() would have done anyway while a missed one turns a rejected draw
# into an accepted degenerate one.
#
# The result is unnamed on both routes (chol2inv() drops dimnames, solve() keeps them), so
# the return shape does not depend on which one ran. Callers that surface the values name
# them from the correlation matrix themselves.
#
# `R_inv` lets a caller that has already formed R^-1 hand it over rather than pay for a
# second factorisation (efa_screen() builds one inverse for several measures at once). The
# underflow check and the solve() fallback below still run, so the singularity contract is
# the same on either route.
.smc_start <- function(R, R_inv = NULL) {
  d <- if (!is.null(R_inv)) {
    diag(R_inv)
  } else {
    tryCatch(diag(chol2inv(chol(R))), error = function(e) NULL)
  }
  if (is.null(d) || min(1 / d) <= ncol(R)^2 * .Machine$double.eps) {
    d <- diag(solve(R))
  }
  unname(pmin(pmax(1 - 1 / d, 0), 1))
}

# Sign convention shared by the unrotated (.finalize_fit) and rotated (.reflect_and_order)
# solutions: reflect each factor so its loadings sum to a non-negative value (as in the
# psych package and SPSS). Returns the +/-1 sign vector; apply it with
# `L %*% diag(signs, nrow = length(signs))` -- the `nrow` guards the single-factor case,
# where diag() would otherwise read a length-1 vector as a matrix dimension.
.reflect_signs <- function(L) {
  signs <- sign(colSums(L))
  signs[signs == 0] <- 1
  signs
}

# Shared post-processing for an unrotated solution. Reflects the loadings to a
# consistent sign, names them (with a V-fallback when the input is unnamed),
# computes the explained variances, fit indices, communalities, model-implied
# correlation matrix, residuals, and the original/final eigenvalues.
.finalize_fit <- function(fit, N, method, lean = FALSE, fiml = NULL) {

  L <- fit$L
  orig_R <- fit$orig_R
  h2 <- fit$h2
  n_factors <- ncol(L)

  # reverse the sign of loadings as done in the psych package and SPSS
  signs <- .reflect_signs(L)
  L <- L %*% diag(signs, nrow = length(signs))

  # Bootstrap replicate path: .boot_se_ci() aggregates only each replicate's
  # unrotated loadings, fit indices, and residuals, so only those are computed.
  # The two eigendecompositions, variable naming, explained variances, and
  # Heywood detection are skipped (none are aggregated per replicate), and the
  # analytic RMSEA confidence bounds are not solved (.gof(ci = FALSE)). The sign
  # reflection above is kept so the loadings are identical to the full path,
  # leaving the downstream target alignment unchanged.
  if (lean) {
    model_implied_R <- L %*% t(L) + diag(1 - h2)
    return(list(
      unrot_loadings = L,
      fit_indices = .gof(L, orig_R, N, method, fit$Fm, ci = FALSE, fiml = fiml),
      residuals = orig_R - model_implied_R,
      convergence = fit$convergence
    ))
  }

  if (!is.null(colnames(orig_R))) {
    # name the loading matrix so the variables can be identified
    rownames(L) <- colnames(orig_R)
  } else {
    varnames <- paste0("V", seq_len(ncol(orig_R)))
    colnames(orig_R) <- varnames
    rownames(orig_R) <- varnames
    rownames(L) <- varnames
  }

  colnames(L) <- paste0("F", seq_len(n_factors))

  vars_accounted <- .compute_vars(L_unrot = L, L_rot = L)
  colnames(vars_accounted) <- colnames(L)

  fit_ind <- .gof(L, orig_R, N, method, fit$Fm, fiml = fiml)

  # calculate model implied R
  model_implied_R <- L %*% t(L) + diag(1 - h2)

  # create the output object
  class(L) <- c("efa_loadings", "LOADINGS")

  # Name communalities
  names(h2) <- colnames(orig_R)

  # Detect Heywood (improper) cases. Named integer vector of the affected
  # variables (empty if none); surfaced to the user by EFA() and shown in
  # summary(). Under PAF the communality can reach or exceed 1 directly. Under
  # ML/ULS the optimiser constrains the uniquenesses to [floor, 1], so an improper
  # solution instead shows up as a uniqueness pinned at the lower floor (the
  # boundary case); flag those too so detection is consistent across estimators.
  # The ML/ULS fitters return psi as a p x 1 matrix; `|` propagates a dim
  # attribute in preference to names, which would strip the variable names off
  # the result, so drop the dim with as.vector() before combining.
  heywood_comm <- h2 >= 1
  heywood_boundary <- if (!is.null(fit$psi)) {
    as.vector(fit$psi) <= .uniqueness_floor + sqrt(.Machine$double.eps)
  } else {
    rep(FALSE, length(h2))
  }
  heywood <- which(heywood_comm | heywood_boundary)

  list(
    orig_R = orig_R,
    h2 = h2,
    # only the eigenvalues are reported, so skip the eigenvectors LAPACK would
    # otherwise be asked for
    orig_eigen = eigen(orig_R, symmetric = TRUE, only.values = TRUE)$values,
    final_eigen = eigen(fit$R_final, symmetric = TRUE, only.values = TRUE)$values,
    iter = fit$iter,
    convergence = fit$convergence,
    heywood = heywood,
    unrot_loadings = L,
    vars_accounted = vars_accounted,
    fit_indices = fit_ind,
    model_implied_R = model_implied_R,
    residuals = orig_R - model_implied_R
  )
}

# Dispatch to the requested fitter, run the shared post-processor, and assemble
# the method-specific output object (field set and order differ per method).
.estimate_model <- function(R, method, n_factors, N = NA,
                            type = "none", max_iter = NA, init_comm = NA,
                            criterion = NA, criterion_type = NA, abs_eigen = NA,
                            start_method = NA, weights = NULL, lean = FALSE,
                            fiml = NULL) {

  fit <- switch(
    method,
    PAF = .PAF(R, n_factors = n_factors, type = type, max_iter = max_iter,
               init_comm = init_comm, criterion = criterion,
               criterion_type = criterion_type, abs_eigen = abs_eigen),
    ML = .ML(R, n_factors = n_factors, start_method = start_method),
    ULS = .ULS(R, n_factors = n_factors),
    DWLS = .DWLS(R, n_factors = n_factors, weights = weights)
  )

  common <- .finalize_fit(fit, N = N, method = method, lean = lean, fiml = fiml)

  # The bootstrap replicate path needs only the post-processed common fields;
  # skip the method-specific output assembly, which the aggregation never reads.
  if (lean) {
    return(common)
  }

  if (method == "PAF") {

    h2_init <- fit$h2_init
    names(h2_init) <- colnames(common$orig_R)

    output <- list(
      orig_R = common$orig_R,
      h2_init = h2_init,
      h2 = common$h2,
      orig_eigen = common$orig_eigen,
      init_eigen = fit$init_eigen,
      final_eigen = common$final_eigen,
      iter = common$iter,
      convergence = common$convergence,
      heywood = common$heywood,
      unrot_loadings = common$unrot_loadings,
      vars_accounted = common$vars_accounted,
      fit_indices = common$fit_indices,
      model_implied_R = common$model_implied_R,
      residuals = common$residuals,
      settings = fit$settings
    )

  } else if (method == "ML") {

    output <- c(common, list(settings = fit$settings))

  } else {

    output <- common

  }

  output
}

ML_test <- .estimate_model(method = "ML",test_models$baseline$cormat, n_factors = 3, N = 500,
               start_method = "factanal")
ML_test_2 <- .estimate_model(method = "ML",test_models$baseline$cormat, n_factors = 3, N = 500,
               start_method = "psych")

test_that("output class and dimensions are correct", {
  expect_s3_class(ML_test$unrot_loadings, "LOADINGS")
  expect_output(str(ML_test), "List of 13")
  expect_s3_class(ML_test_2$unrot_loadings, "LOADINGS")
  expect_output(str(ML_test_2), "List of 13")

})

test_that("outputs are correct", {
  expect_equal(ML_test$orig_R, test_models$baseline$cormat)
  expect_equal(sum(ML_test$orig_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(ML_test$final_eigen), ncol(test_models$baseline$cormat))
  expect_equal(ML_test$convergence, 0)
  expect_equal(ML_test_2$orig_R, test_models$baseline$cormat)
  expect_equal(sum(ML_test_2$orig_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(ML_test_2$final_eigen), ncol(test_models$baseline$cormat))
  expect_equal(ML_test_2$convergence, 0)
})

test_that("fit indices are returned correctly", {
  expect_output(str(ML_test$fit_indices), "List of 18")

  expect_type(ML_test$fit_indices$chi, "double")
  expect_type(ML_test$fit_indices$df, "double")
  expect_type(ML_test$fit_indices$p_chi, "double")
  expect_type(ML_test$fit_indices$CAF, "double")
  expect_type(ML_test$fit_indices$CFI, "double")
  expect_type(ML_test$fit_indices$RMSEA, "double")
  expect_type(ML_test$fit_indices$RMSEA_LB, "double")
  expect_type(ML_test$fit_indices$RMSEA_UB, "double")
  expect_type(ML_test$fit_indices$AIC, "double")
  expect_type(ML_test$fit_indices$BIC, "double")
  expect_type(ML_test$fit_indices$Fm, "double")
  expect_type(ML_test$fit_indices$chi_null, "double")
  expect_type(ML_test$fit_indices$df_null, "double")
  expect_type(ML_test$fit_indices$p_null, "double")
  # ML_test_2 differs only in where the optimiser is started; both fits are assembled by
  # the same fit-index code on the same model, so the storage types are checked once.
})

test_that("settings are returned correctly", {
  expect_named(ML_test$settings, c("start_method"))
  expect_equal(ML_test$settings$start_method, "factanal")
  expect_named(ML_test_2$settings, c("start_method"))
  expect_equal(ML_test_2$settings$start_method, "psych")
})

rm(ML_test, ML_test_2)

paf_efatools <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                     max_iter = NA, type = "EFAtools", init_comm = NA,
                     criterion = NA, criterion_type = NA, abs_eigen = NA)
paf_psych <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                     max_iter = NA, type = "psych", init_comm = NA,
                     criterion = NA, criterion_type = NA, abs_eigen = NA)
paf_spss <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                  max_iter = NA, type = "SPSS", init_comm = NA,
                  criterion = NA, criterion_type = NA, abs_eigen = NA)
paf_none <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                 max_iter = 500, type = "none", init_comm = "unity",
                 criterion = 1e-4, criterion_type = "sum", abs_eigen = TRUE)
paf_mac_t <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                 max_iter = 500, type = "none", init_comm = "mac",
                 criterion = 1e-4, criterion_type = "max_individual", abs_eigen = TRUE)
paf_mac_f <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                  max_iter = 500, type = "none", init_comm = "mac",
                  criterion = 1e-4, criterion_type = "max_individual",
                  abs_eigen = FALSE)
paf_F1_t <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 1, N = 500,
                max_iter = 500, type = "none", init_comm = "mac",
                criterion = 1e-4, criterion_type = "max_individual", abs_eigen = TRUE)
paf_F1_f <- .estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 1, N = 500,
               max_iter = 500, type = "none", init_comm = "mac",
               criterion = 1e-4, criterion_type = "max_individual", abs_eigen = FALSE)


test_that("output class and dimensions are correct", {
  expect_s3_class(paf_efatools$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_psych$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_spss$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_none$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_mac_t$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_mac_f$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_F1_t$unrot_loadings, "LOADINGS")
  expect_s3_class(paf_F1_f$unrot_loadings, "LOADINGS")


  expect_output(str(paf_efatools), "List of 15")
  expect_output(str(paf_psych), "List of 15")
  expect_output(str(paf_spss), "List of 15")
  expect_output(str(paf_none), "List of 15")
  expect_output(str(paf_mac_t), "List of 15")
  expect_output(str(paf_mac_f), "List of 15")
  expect_output(str(paf_F1_t), "List of 15")
  expect_output(str(paf_F1_f), "List of 15")
})

test_that("original correlation matrix and eigenvalues are correct", {
  expect_equal(paf_efatools$orig_R, test_models$baseline$cormat)
  expect_equal(paf_psych$orig_R, test_models$baseline$cormat)
  expect_equal(paf_spss$orig_R, test_models$baseline$cormat)
  expect_equal(paf_none$orig_R, test_models$baseline$cormat)
  expect_equal(paf_mac_t$orig_R, test_models$baseline$cormat)
  expect_equal(paf_mac_f$orig_R, test_models$baseline$cormat)
  expect_equal(paf_F1_t$orig_R, test_models$baseline$cormat)
  expect_equal(paf_F1_f$orig_R, test_models$baseline$cormat)

  expect_equal(sum(paf_efatools$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_psych$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_spss$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_none$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_mac_t$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_mac_f$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_F1_t$orig_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_F1_f$orig_eigen), ncol(test_models$baseline$cormat))

  # The initial eigenvalues are taken from the correlation matrix after its diagonal has
  # been replaced by the initial communalities, so their sum is the trace of the reduced
  # matrix. Every init_comm rule that estimates communalities (smc, mac) puts something
  # below 1 on the diagonal and the sum therefore falls short of p; init_comm = "unity"
  # leaves the unit diagonal in place, so there the sum is exactly p.
  expect_lt(sum(paf_efatools$init_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_psych$init_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_spss$init_eigen), ncol(test_models$baseline$cormat))
  expect_equal(sum(paf_none$init_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_mac_t$init_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_mac_f$init_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_F1_t$init_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_F1_f$init_eigen), ncol(test_models$baseline$cormat))

  expect_lt(sum(paf_efatools$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_psych$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_spss$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_none$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_mac_t$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_mac_f$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_F1_t$final_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(paf_F1_f$final_eigen), ncol(test_models$baseline$cormat))
})

test_that("the PAF kernel extracts the leading eigenpairs", {
  # One pass of the kernel must reproduce the R-side extraction: the n_fac largest
  # eigenvalues in descending order with their eigenvectors. Eigenvector signs are free,
  # so the comparison uses the reproduced matrix and the communalities it implies.
  R <- test_models$baseline$cormat
  h2 <- .smc_start(R)
  R_red <- R
  diag(R_red) <- h2
  ev <- eigen(R_red, symmetric = TRUE)

  for (k in c(1L, 3L)) {
    out <- .paf_iter(h2, 1e-3, R_red, k, TRUE, 2L, 1L)
    lam <- abs(ev$values[seq_len(k)])
    L_ref <- ev$vectors[, seq_len(k), drop = FALSE] %*% diag(sqrt(lam), nrow = k)

    expect_equal(tcrossprod(out$L), tcrossprod(L_ref), info = paste("k =", k))
    expect_equal(as.vector(out$h2), rowSums(L_ref^2), info = paste("k =", k))
  }
})

test_that("fit indices are returned correctly", {
  expect_output(str(paf_efatools$fit_indices), "List of 18")
  expect_output(str(paf_psych$fit_indices), "List of 18")
  expect_output(str(paf_spss$fit_indices), "List of 18")
  expect_output(str(paf_none$fit_indices), "List of 18")
  expect_output(str(paf_mac_t$fit_indices), "List of 18")
  expect_output(str(paf_mac_f$fit_indices), "List of 18")
  expect_output(str(paf_F1_t$fit_indices), "List of 18")
  expect_output(str(paf_F1_f$fit_indices), "List of 18")

  # PAF has no discrepancy function, so the likelihood-based indices are undefined for
  # every one of these fits: they all reach the same branch of the fit-index assembly,
  # which sets the whole block to NA_real_. Neither the type preset, the initial
  # communalities, nor the number of factors enters any of these keys, so the eight
  # configurations are checked against one shared expectation, labelled so a failure still
  # names the configuration that produced it.
  paf_fits <- list(EFAtools = paf_efatools, psych = paf_psych, SPSS = paf_spss,
                   none = paf_none, mac_t = paf_mac_t, mac_f = paf_mac_f,
                   F1_t = paf_F1_t, F1_f = paf_F1_f)
  na_indices <- list(chi = NA_real_, p_chi = NA_real_, CFI = NA_real_,
                     RMSEA = NA_real_, RMSEA_LB = NA_real_, RMSEA_UB = NA_real_,
                     AIC = NA_real_, BIC = NA_real_, Fm = NA_real_,
                     chi_null = NA_real_, df_null = NA_real_, p_null = NA_real_)
  for (nm in names(paf_fits)) {
    expect_equal(paf_fits[[nm]]$fit_indices[names(na_indices)], na_indices, info = nm)
  }

  expect_type(paf_efatools$fit_indices$df, "double")
  expect_type(paf_psych$fit_indices$df, "double")
  expect_type(paf_spss$fit_indices$df, "double")
  expect_type(paf_none$fit_indices$df, "double")
  expect_type(paf_mac_t$fit_indices$df, "double")
  expect_type(paf_mac_f$fit_indices$df, "double")
  expect_type(paf_F1_t$fit_indices$df, "double")
  expect_type(paf_F1_f$fit_indices$df, "double")

  expect_type(paf_efatools$fit_indices$CAF, "double")
  expect_type(paf_psych$fit_indices$CAF, "double")
  expect_type(paf_spss$fit_indices$CAF, "double")
  expect_type(paf_none$fit_indices$CAF, "double")
  expect_type(paf_mac_t$fit_indices$CAF, "double")
  expect_type(paf_mac_f$fit_indices$CAF, "double")
  expect_type(paf_F1_t$fit_indices$CAF, "double")
  expect_type(paf_F1_f$fit_indices$CAF, "double")

})

test_that("settings are returned correctly", {
  expect_named(paf_efatools$settings, c("max_iter", "init_comm", "criterion",
                                        "criterion_type", "abs_eigen"))
  expect_named(paf_psych$settings, c("max_iter", "init_comm", "criterion",
                                        "criterion_type", "abs_eigen"))
  expect_named(paf_spss$settings, c("max_iter", "init_comm", "criterion",
                                        "criterion_type", "abs_eigen"))
  expect_named(paf_none$settings, c("max_iter", "init_comm", "criterion",
                                    "criterion_type", "abs_eigen"))
  expect_named(paf_mac_t$settings, c("max_iter", "init_comm", "criterion",
                                    "criterion_type", "abs_eigen"))
  expect_named(paf_mac_f$settings, c("max_iter", "init_comm", "criterion",
                                   "criterion_type", "abs_eigen"))
  expect_named(paf_F1_t$settings, c("max_iter", "init_comm", "criterion",
                                   "criterion_type", "abs_eigen"))
  expect_named(paf_F1_f$settings, c("max_iter", "init_comm", "criterion",
                                   "criterion_type", "abs_eigen"))

  expect_equal(paf_efatools$settings$max_iter, 300)
  expect_equal(paf_psych$settings$max_iter, 50)
  expect_equal(paf_spss$settings$max_iter, 25)
  expect_equal(paf_none$settings$max_iter, 500)
  expect_equal(paf_mac_t$settings$max_iter, 500)
  expect_equal(paf_mac_f$settings$max_iter, 500)
  expect_equal(paf_F1_t$settings$max_iter, 500)
  expect_equal(paf_F1_f$settings$max_iter, 500)

  expect_equal(paf_efatools$settings$init_comm, "smc")
  expect_equal(paf_psych$settings$init_comm, "smc")
  expect_equal(paf_spss$settings$init_comm, "smc")
  expect_equal(paf_none$settings$init_comm, "unity")
  expect_equal(paf_mac_t$settings$init_comm, "mac")
  expect_equal(paf_mac_f$settings$init_comm, "mac")
  expect_equal(paf_F1_t$settings$init_comm, "mac")
  expect_equal(paf_F1_f$settings$init_comm, "mac")

  expect_equal(paf_efatools$settings$criterion, 0.001)
  expect_equal(paf_psych$settings$criterion, 0.001)
  expect_equal(paf_spss$settings$criterion, 0.001)
  expect_equal(paf_none$settings$criterion, 1e-4)
  expect_equal(paf_mac_t$settings$criterion, 1e-4)
  expect_equal(paf_mac_f$settings$criterion, 1e-4)
  expect_equal(paf_F1_t$settings$criterion, 1e-4)
  expect_equal(paf_F1_f$settings$criterion, 1e-4)

  expect_equal(paf_efatools$settings$criterion_type, "sum")
  expect_equal(paf_psych$settings$criterion_type, "sum")
  expect_equal(paf_spss$settings$criterion_type, "max_individual")
  expect_equal(paf_none$settings$criterion_type, "sum")
  expect_equal(paf_mac_t$settings$criterion_type, "max_individual")
  expect_equal(paf_mac_f$settings$criterion_type, "max_individual")
  expect_equal(paf_F1_t$settings$criterion_type, "max_individual")
  expect_equal(paf_F1_f$settings$criterion_type, "max_individual")

  expect_equal(paf_efatools$settings$abs_eigen, TRUE)
  expect_equal(paf_psych$settings$abs_eigen, FALSE)
  expect_equal(paf_spss$settings$abs_eigen, TRUE)
  expect_equal(paf_none$settings$abs_eigen, TRUE)
  expect_equal(paf_mac_t$settings$abs_eigen, TRUE)
  expect_equal(paf_mac_f$settings$abs_eigen, FALSE)
  expect_equal(paf_F1_t$settings$abs_eigen, TRUE)
  expect_equal(paf_F1_f$settings$abs_eigen, FALSE)
})


test_that("warnings and errors are thrown correctly", {
  expect_error(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                    max_iter = 500, type = "none", init_comm = "unity",
                    criterion = 1e-4, criterion_type = "sum"), class = "efa_type_none")

  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                    type = "EFAtools", init_comm = "smc"), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                    type = "EFAtools", criterion = 0.001), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "EFAtools", criterion_type = "sum"), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "EFAtools", max_iter = 400), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "EFAtools", abs_eigen = TRUE), class = "efa_type_override")

  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "psych", init_comm = "smc"), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "psych", criterion = 0.001), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "psych", criterion_type = "sum"), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "psych", max_iter = 400), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "psych", abs_eigen = TRUE), class = "efa_type_override")

  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "SPSS", init_comm = "smc"), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "SPSS", criterion = 0.001), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "SPSS", criterion_type = "sum"), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "SPSS", max_iter = 400), class = "efa_type_override")
  expect_warning(.estimate_model(method = "PAF",test_models$baseline$cormat, n_factors = 3, N = 500,
                      type = "SPSS", abs_eigen = TRUE), class = "efa_type_override")
  expect_error(.estimate_model(method = "PAF",IDS2_R, n_factors = 7, N = 1991, max_iter = 500, type = "none",
                    init_comm = "smc", criterion = 1e-4, criterion_type = "sum",
                    abs_eigen = FALSE), class = "efa_paf_negative_eigen")
  expect_error(.estimate_model(method = "PAF",IDS2_R, n_factors = 7, N = 1991, max_iter = 500, type = "none",
                    init_comm = "smc", criterion = 1e-4,
                    criterion_type = "max_individual",
                    abs_eigen = FALSE), class = "efa_paf_negative_eigen")
})

rm(paf_efatools, paf_psych, paf_spss, paf_none, paf_mac_t, paf_mac_f, paf_F1_t,
   paf_F1_f)

#' RiskDimensions
#'
#' A list containing the bivariate correlations (cormat)
#' of the 9 dimensions on which participants in Fischhoff et al. (1978) rated
#' different activities and technologies as well as the sample size (N). This was
#' then analyzed together with ratings of the risks and benefits of these
#' activities and technologies.
#'
#' @format A list of 2 with elements "cormat" (9 x 9 matrix of bivariate
#' correlations) and "N" (scalar). The correlation matrix contains the following
#' risk dimensions:
#' \describe{
#'   \item{Voluntariness}{(numeric) - Voluntariness of exposure to the risk.}
#'   \item{Immediacy}{(numeric) - Immediacy of the risk's effect.}
#'   \item{Known to exposed}{(numeric) - How well the risk is known to those exposed to it.}
#'   \item{Known to science}{(numeric) - How well the risk is known to science.}
#'   \item{Controllability}{(numeric) - Controllability of the risk.}
#'   \item{Newness}{(numeric) - Newness of the risk.}
#'   \item{Chronic}{(numeric) - Whether the risk is chronic rather than catastrophic.}
#'   \item{Common}{(numeric) - Whether the risk is common rather than dreaded.}
#'   \item{Severity of consequences}{(numeric) - Severity of the consequences.}
#'  }
#' @source Fischhoff, B, Slovic, P, Lichtenstein, S, Read, S, and Combs, B. (1978). How safe is safe enough? A psychometric study of attitudes towards technological risks and benefits. Policy Sciences, 9, 127-152. doi: 10.1007/BF00143739
"RiskDimensions"

# Single rotation core: pick the engine by name, then post-process once.
#
# `.rotate_model()` resolves the `type` preset for the requested rotation, guards
# the single-factor case, runs the rotation engine that the rotation name maps to
# in a lookup table, and (for every engine except promax) hands the raw solution
# to `.reflect_and_order()` for the shared sign reflection, factor ordering,
# structure matrix, and explained variances. Selecting the engine from a name
# table keeps the rotation name and the criterion it runs in lockstep. Promax
# reflects and reorders its varimax base before the oblique fit, so it finalizes
# its own solution.

# Orthogonal rotation engines, keyed by rotation name. Each resolves its criterion-specific
# argument and hands the matching compiled entry to `.gpf_native()`, which returns the raw
# rotated solution ($loadings, $Th); `eps`, `normalize`, `randomStarts`, and any extra
# arguments are forwarded through `...`. The Crawford-Ferguson criteria (quartimax with
# kappa = 0, equamax with kappa = k / (2 p)), geomin, Bentler, and bifactor all use the native
# engine. The geomin offset `delta` defaults to GPArotation's 0.01 and may be overridden by an
# exact-named `delta` in `...`.
.orth_engines <- list(
  equamax   = function(L, ...) .gpf_native(.rotate_cf_orth, L,
                                           list(kappa = ncol(L) / (2 * nrow(L))), ...),
  quartimax = function(L, ...) .gpf_native(.rotate_cf_orth, L, list(kappa = 0), ...),
  bentlerT  = function(L, ...) .gpf_native(.rotate_bentler_orth, L, list(), ...),
  geominT   = function(L, ...) .gpf_native(.rotate_geomin_orth, L,
                                           list(delta = .gpf_crit(list(...), "delta", 0.01)),
                                           ...),
  bifactorT = function(L, ...) .gpf_native(.rotate_bifactor_orth, L, list(), ...)
)

# Warn when a native gradient-projection rotation did not reach the convergence tolerance
# within `maxit` iterations, mirroring the warning the GPArotation engines emit so that the
# native and GPArotation-backed rotations behave the same way on a non-convergent fit. The
# remedy names the knobs that move the outcome; `tunable_maxit` is FALSE for the SPSS varimax
# sweep, whose iteration budget is fixed in the algorithm, so there the tolerance is named alone.
.warn_rotation_no_convergence <- function(converged, maxit, tunable_maxit = TRUE) {
  if (!isTRUE(converged)) {
    remedy <- if (tunable_maxit) {
      "In {.fn rotate_control}, increase {.arg maxit} for more iterations, or {.arg precision}
       for a weaker tolerance."
    } else {
      "In {.fn rotate_control}, increase {.arg precision} for a weaker tolerance."
    }
    cli::cli_warn(c("The rotation did not converge within {maxit} iteration{?s}.",
                    "i" = remedy),
                  class = "efa_rotation_no_convergence")
  }
}

# Warn when an oblique rotation returns near-collinear factors. A factor correlation
# above `tol` in absolute value means two rotated factors are barely distinguishable,
# which is the usual signature of extracting more factors than the data support (and,
# for oblimin, of a large `gam`). The solution is still returned -- an extreme
# correlation is a property of the fit rather than a failure of the rotation -- but it
# should not be interpreted without inspecting the factor intercorrelations. The 0.9
# cut-off follows the extreme-correlation diagnostic GPArotation reports from its own
# oblique rotations (Bernaards & Jennrich, 2005).
#
# The reported value is TRUNCATED, not rounded: a correlation of 0.9999 must not print as
# "1.00", which is the perfectly collinear state a rejected singular transformation would
# have produced, and a correlation of 0.905 must not print as "0.90", which reads as
# sitting at the cut-off rather than above it.
.warn_extreme_phi <- function(Phi, tol = 0.9) {

  if (is.null(Phi) || !is.matrix(Phi) || nrow(Phi) < 2L) return(invisible(NULL))

  off <- abs(Phi[upper.tri(Phi)])
  off <- off[is.finite(off)]
  if (length(off) == 0L || max(off) <= tol) return(invisible(NULL))

  max_r <- formatC(trunc(max(off) * 1000) / 1000, format = "f", digits = 3)
  cli::cli_warn(
    c("Extreme factor correlations (max |r| = {max_r}).",
      "i" = "This usually indicates factor over-extraction; inspect {.field Phi}."),
    class = "efa_rotation_extreme_phi"
  )
  invisible(NULL)
}

# Warn when a bifactor criterion is requested for a two-factor solution. The
# Jennrich-Bentler bifactor criterion sums the products of squared loadings over pairs of
# distinct GROUP factors (Jennrich & Bentler, 2011, 2012), so with a general factor and a
# single group factor there is no such pair: the criterion is identically zero, every
# rotation attains it, and the identity start wins the tie. The returned "rotated" loadings
# are then the unrotated ones, while the multistart diagnostics -- one distinct optimum, zero
# criterion spread -- read exactly like a well-identified solution. The solution is still
# returned, so the situation is signalled rather than turned into an error.
.warn_bifactor_trivial <- function(rotation, n_factors) {
  if (!rotation %in% c("bifactorT", "bifactorQ") || n_factors != 2L) {
    return(invisible(NULL))
  }
  cli::cli_warn(
    c("The bifactor criterion is identically zero with only one group factor.",
      "i" = "With two factors every rotation is optimal, so the returned loadings are the
             unrotated ones.",
      "i" = "Extract at least three factors for a bifactor rotation."),
    class = "efa_bifactor_trivial"
  )
  invisible(NULL)
}

# Validate the arguments the compiled rotation entries take, in R, before the entry is
# called. The compiled entries check the same quantities and stay in place as backstops for
# direct internal callers, but they raise unclassed exceptions whose messages name neither
# the function nor the argument, and a non-numeric value never reaches them at all (the
# numeric coercion at the boundary fails first, with a message naming only the C type).
# Checking here gives every one of them the `efa_rotation_arg` class and the argument name
# the user typed. `crit` carries only the criterion parameter of the entry actually being
# run, so a criterion is never held to another criterion's contract; `maxit` is checked for
# every criterion because every entry takes it.
.assert_rotation_args <- function(crit, maxit, L) {

  bad_arg <- function(arg, requirement) {
    cli::cli_abort("{.arg {arg}} must be {requirement}.", class = "efa_rotation_arg")
  }

  # A fractional `maxit` is truncated at the C boundary, so the engine would run a different
  # budget from the one the non-convergence warning reports.
  if (!isTRUE(checkmate::test_int(maxit, lower = 0))) {
    bad_arg("maxit", "a single whole number of at least 0")
  }

  if ("gam" %in% names(crit) && !isTRUE(checkmate::test_number(crit$gam, finite = TRUE))) {
    bad_arg("gam", "a single finite number")
  }

  if ("delta" %in% names(crit) &&
      !(isTRUE(checkmate::test_number(crit$delta, finite = TRUE)) && crit$delta > 0)) {
    bad_arg("delta", "a single positive finite number")
  }

  # simplimax counts the loadings it drives toward zero, so its upper bound is the number of
  # loadings in the solution; the message quotes that bound rather than leaving the user to
  # work it out.
  if ("k" %in% names(crit)) {
    n_loadings <- nrow(L) * ncol(L)
    if (!isTRUE(checkmate::test_int(crit$k, lower = 1, upper = n_loadings))) {
      bad_arg("k", paste0("a single whole number between 1 and ", n_loadings,
                          ", the number of loadings"))
    }
  }

  invisible(NULL)
}

# Read an overridable criterion parameter (the geomin `delta`, the oblimin `gam`) from a
# rotation engine's `...` by EXACT name, falling back to `default`. Looking the value up by
# exact name -- rather than declaring the parameter as a named formal before `...` -- keeps
# R's partial matching from silently binding an abbreviated or misspelled dot argument (e.g.
# `EFA(rotation = "geominT", del = 1)`) to the criterion parameter. efa_fit() and
# rotate_control() reject a name their engines cannot consume up front (see
# `.rotation_dot_extras()`); the exact-name lookup remains the guard against partial matching
# for direct internal callers, where an unrecognised argument is simply ignored.
.gpf_crit <- function(dots, name, default) {
  if (name %in% names(dots)) dots[[name]] else default
}

# Shared driver for the native gradient-projection rotations. `entry` is the compiled rotation
# (e.g. `.rotate_cf_orth`); `crit` is a named list of the criterion-specific argument(s) for
# that entry (empty for the criteria that take none). The GPArotation-style controls (`eps`,
# `normalize`, `randomStarts` -> `random_starts`, `maxit`) are mapped onto the compiled entry, a
# non-convergent fit warns, and the raw rotated solution is returned in the convention
# `.reflect_and_order()` expects: the rotated `$loadings`, the rotation matrix `$Th`, and -- for
# oblique rotations -- the factor correlations `$Phi` (orthogonal entries return no `Phi`).
# Passing the criterion arguments explicitly via `crit` keeps them out of the partial-matching
# path; the trailing `...` simply absorbs the controls the engine-table closures forward
# through. The screen-and-triage multistart controls (`screen_keep`, `triage_maxit`,
# `triage_improve_tol`) are forwarded only when a criterion overrides the compiled entry's default
# (a `NULL` leaves that default in place); they are set per criterion by `.gpf_multistart_defaults`
# below. simplimax's entry takes no such controls (it always runs full multistart), and its closure
# passes none, so they are never forwarded to it.
.gpf_native <- function(entry, L, crit = list(), eps = 1e-5, normalize = TRUE,
                        randomStarts = 100, maxit = 1000L,
                        screen_keep = NULL, triage_maxit = NULL,
                        triage_improve_tol = NULL, ...) {
  .assert_rotation_args(crit, maxit, L)
  args <- list(eps = eps, normalize = normalize,
               random_starts = randomStarts, maxit = maxit)
  if (!is.null(screen_keep)) args$screen_keep <- screen_keep
  if (!is.null(triage_maxit)) args$triage_maxit <- triage_maxit
  if (!is.null(triage_improve_tol)) args$triage_improve_tol <- triage_improve_tol
  res <- do.call(entry, c(list(unclass(L)), crit, args))
  .warn_rotation_no_convergence(res$convergence, maxit)
  # `converged` is the flag of the start whose solution is RETURNED, which the aggregate
  # per-start counts below cannot reconstruct: the winner is chosen by lowest criterion
  # value, so a non-converged start with a lower value beats a converged one.
  out <- list(loadings = res$loadings, Th = res$Th, converged = isTRUE(res$convergence),
              all_values = res$all_values, all_converged = res$all_converged)
  if (!is.null(res$Phi)) out$Phi <- res$Phi
  out
}

# Oblique rotation engines, keyed by rotation name. Each resolves its criterion-specific
# argument and hands the matching compiled entry to `.gpf_native()`, which returns the raw
# rotated solution ($loadings, $Th, $Phi); `k` is the simplimax target count (consumed here and
# ignored by the others). The oblimin and quartimin criteria (quartimin is oblimin pinned at
# `gam = 0`, not overridable; oblimin's `gam` defaults to 0 and may be overridden by an
# exact-named `gam` in `...`), geomin, Bentler, bifactor, and simplimax all use the native
# engine.
.oblq_engines <- list(
  oblimin   = function(L, k, ...) .gpf_native(.rotate_oblimin, L,
                                              list(gam = .gpf_crit(list(...), "gam", 0)), ...),
  quartimin = function(L, k, ...) .gpf_native(.rotate_oblimin, L, list(gam = 0), ...),
  simplimax = function(L, k, ...) .gpf_native(.rotate_simplimax_oblq, L, list(k = k), ...),
  bentlerQ  = function(L, k, ...) .gpf_native(.rotate_bentler_oblq, L, list(), ...),
  geominQ   = function(L, k, ...) .gpf_native(.rotate_geomin_oblq, L,
                                              list(delta = .gpf_crit(list(...), "delta", 0.01)),
                                              ...),
  bifactorQ = function(L, k, ...) .gpf_native(.rotate_bifactor_oblq, L, list(), ...)
)

# Per-criterion screen-and-triage multistart overrides for the native gradient-projection
# rotations. The compiled entries triage the five best-screened random starts by default
# (screen_keep = 5, triage_maxit = 25); two already reach the global criterion optimum for every
# criterion except geominQ across a panel of real correlation matrices, and five is a cheap safety
# margin (a small, single-digit-millisecond cost
# per rotation) against harder, unseen data. The oblique geomin criterion (geominQ) is multimodal
# enough to need more: its best-screened starts can all miss the global basin, so it triages more
# starts for longer (screen_keep = 10, triage_maxit = 50), recovering the global optimum on every
# tested matrix while staying far cheaper than fully optimizing all starts. A criterion absent from
# this list uses the compiled default. (geominT, the orthogonal geomin, reaches the global optimum
# at the default; simplimax is omitted because it always runs full multistart, not
# screen-and-triage.)
.gpf_multistart_defaults <- list(
  geominQ = list(screen_keep = 10, triage_maxit = 50)
)

# Names a rotation's engine can consume from efa_fit()'s `...`, kept in lockstep with the
# engine tables above: `maxit` is the one `.gpf_native()` control `.rotate_model()` does not
# already pass explicitly, and the criterion extras are read by exact name via `.gpf_crit()`
# (`gam` for oblimin, `delta` for the geomin pair). varimax and promax run outside the native
# engines and take no extras, as does rotation = "none", where no engine runs at all.
# efa_fit() rejects any other dot argument up front; `.rotation_extra_union` is the
# across-rotation union rotate_control() validates its stored extras against.
.rotation_dot_extras <- function(rotation) {
  switch(rotation,
         none = ,
         varimax = ,
         promax = character(0),
         oblimin = c("maxit", "gam"),
         geominT = ,
         geominQ = c("maxit", "delta"),
         "maxit")
}

.rotation_extra_union <- c("maxit", "gam", "delta")

# Canonical rotation names by family. The engine tables above key the GPArotation
# engines and exclude the special-cased `varimax`/`promax`; these vectors are the
# *full* family memberships used to classify a rotation name. `.rotation_family()`
# maps a single rotation name to "none", "orthogonal", or "oblique".
.orth_rotations <- c("varimax", "quartimax", "equamax", "bentlerT", "geominT", "bifactorT")
.oblq_rotations <- c("promax", "oblimin", "quartimin", "simplimax",
                     "bentlerQ", "geominQ", "bifactorQ")

.rotation_family <- function(rotation) {
  if (identical(rotation, "none")) return("none")
  if (rotation %in% .orth_rotations) return("orthogonal")
  if (rotation %in% .oblq_rotations) return("oblique")
  cli::cli_abort("Unknown rotation: {.val {rotation}}.", class = "efa_unknown_rotation")
}

# Shared post-processing for a rotated solution. Reflects each factor to a
# consistent (positive) orientation, orders the factors as requested by
# `order_type`, and assembles the rotated loadings, factor correlations, structure
# matrix, and explained variances. The sign reflection and the factor ordering are
# applied to the loadings, the rotation matrix, and the factor intercorrelations
# alike, so the structure matrix and reported correlations stay consistent with the
# loadings. The factor columns are labelled "F1".."Fk" by their position in the
# ordered solution, so every rotation returns labelled loadings, factor
# intercorrelations, and structure coefficients. The labels deliberately do not
# follow the unrotated columns: a rotation mixes all k factors, so a rotated column
# corresponds to no single unrotated factor, and the engines return their columns in
# an arbitrary order (for the multistart engines it depends on which random start
# wins). Carrying the unrotated labels through the sort would therefore attach a
# run-dependent permutation of "F1".."Fk" to an otherwise stable solution.
# Used by every engine except promax.
.reflect_and_order <- function(loadings, Phi = NULL, rotmat, L_unrot, order_type) {

  oblique <- !is.null(Phi)
  var_names <- rownames(L_unrot)

  # reflect factors with negative sums
  signs <- .reflect_signs(loadings)
  loadings <- loadings %*% diag(signs, nrow = length(signs))

  if (oblique) {
    # reflect the factor intercorrelations the same way as the loadings so the
    # structure matrix and reported correlations stay in sync
    Phi <- diag(signs, nrow = length(signs)) %*% Phi %*% diag(signs, nrow = length(signs))
  }

  # order the factors by their explained variance (largest first). "eigen" orders
  # by the reported "SS loadings": colSums(L^2) for orthogonal solutions, the
  # Phi-weighted sum of squares diag(Phi L'L) for oblique ones, so the reported
  # variances decrease monotonically (as in psych). "ss_factors" orders oblique
  # factors by the raw pattern sum of squares colSums(L^2) instead. For orthogonal
  # solutions Phi = I, so the two keys coincide and order_type has no effect.
  ss <- if (oblique && order_type == "eigen") {
    diag(Phi %*% crossprod(loadings))
  } else {
    colSums(loadings^2)
  }
  ord <- order(ss, decreasing = TRUE)
  loadings <- loadings[, ord, drop = FALSE]

  fac_names <- colnames(L_unrot)
  dimnames(loadings) <- list(var_names, fac_names)

  # the rotation matrix follows the same sign reflection and factor ordering as
  # the loadings, so the documented reproduction identity holds (orthogonal:
  # L_unrot %*% rotmat; oblique: L_unrot %*% t(solve(rotmat)))
  rotmat <- (rotmat %*% diag(signs))[, ord, drop = FALSE]

  if (oblique) {
    Phi <- Phi[ord, ord, drop = FALSE]
    dimnames(Phi) <- list(fac_names, fac_names)
  }

  vars_accounted_rot <- .compute_vars(L_unrot = L_unrot, L_rot = loadings, Phi = Phi)
  colnames(vars_accounted_rot) <- fac_names

  if (!oblique) {
    class(loadings) <- c("efa_loadings", "LOADINGS")
    return(list(rot_loadings = loadings,
                rotmat = rotmat,
                vars_accounted_rot = vars_accounted_rot))
  }

  Structure <- loadings %*% Phi
  dimnames(Structure) <- list(var_names, fac_names)

  class(loadings) <- c("efa_loadings", "LOADINGS")
  class(Structure) <- c("efa_loadings", "LOADINGS")

  list(rot_loadings = loadings,
       Phi = Phi,
       Structure = Structure,
       rotmat = rotmat,
       vars_accounted_rot = vars_accounted_rot)
}

# Single-factor solutions cannot be rotated; warn and return the unrotated
# loadings in the family's output shape. The rotation-only outputs (factor
# intercorrelations, structure matrix, rotated variances) are returned as NULL so
# the print/summary sections that guard on `is.null` skip them.
.rotate_single_factor <- function(L, settings, oblique) {

  cli::cli_warn("A single factor cannot be rotated; returning the unrotated loadings.",
                class = "efa_single_factor")

  if (isTRUE(oblique)) {
    return(list(rot_loadings = L, Phi = NULL, Structure = NULL, rotmat = NA,
                vars_accounted_rot = NULL, settings = settings))
  }

  list(rot_loadings = L, rotmat = NA, vars_accounted_rot = NULL, settings = settings)
}

# Count distinct local minima among the criterion values reached by the rotation's
# random starts. Values within a small relative tolerance of each other are treated
# as the same minimum (different starts converging to one optimum differ only at the
# convergence tolerance), so the count reflects how many genuinely different optima
# the starts found.
.count_distinct_minima <- function(values, tol = 1e-6) {
  v <- sort(values[is.finite(values)])
  if (length(v) == 0L) return(0L)
  n <- 1L
  for (i in seq_along(v)[-1]) {
    if (v[i] - v[i - 1L] > tol * (1 + abs(v[i]))) n <- n + 1L
  }
  n
}

# Summarise a multistart rotation run for the output: how many starts were available in
# total, how many of them were actually optimized, how many converged to a genuine local
# optimum, how many distinct optima they found, and the spread and best value of the
# attained criterion. `all_values` and `all_converged` are the per-start criterion values
# and convergence flags returned by the native rotation engines, one entry per OPTIMIZED
# start, so `length(all_values)` is the number of optimizations that actually ran. Under the
# screen-and-triage strategy that is far fewer than the number of random starts: the starts
# that are not among the best-screened have their objective evaluated at the start point and
# are then discarded, so they can never yield a local optimum and are not counted here.
# `n_starts_total` includes the rational (identity) start alongside the `randomStarts` random
# ones, so it bounds `n_converged` from above. The distinct-optima count and spread are taken
# over the CONVERGED starts only: the screen-and-triage engine leaves the starts it does not
# promote at a short, unconverged triage iterate, which is not a local optimum and must not
# be counted as one (those starts have `all_converged == 0`). The best criterion value is the
# lowest attained over all finite starts (the selected solution).
# `converged` is the convergence flag of the start that WON, i.e. of the solution actually
# returned. It is reported separately from `n_converged` because the two can disagree: the
# winner is the lowest-criterion start, and a start that stopped short of the tolerance at a
# lower criterion value beats a converged start at a higher one. Collapsing the two would
# present such a solution as converged because its rivals were.
.rotation_diagnostics <- function(all_values, all_converged, randomStarts, converged) {
  finite <- is.finite(all_values)
  converged_values <- all_values[finite & all_converged == 1L]
  list(converged = isTRUE(converged),
       n_starts_total = as.integer(randomStarts) + 1L,
       n_optimized = length(all_values),
       n_converged = length(converged_values),
       n_distinct_minima = .count_distinct_minima(converged_values),
       criterion_spread = if (length(converged_values) > 1L) diff(range(converged_values)) else 0,
       criterion_best = if (any(finite)) min(all_values[finite]) else NA_real_)
}

# Resolve the preset, run the named engine, and post-process. The field set and
# order of the returned list differ per rotation family.
.rotate_model <- function(x, rotation, type = c("EFAtools", "psych", "SPSS", "none"),
                          normalize = TRUE, precision = 1e-5, order_type = NA,
                          varimax_type = NA, P_type = NA, k = NA,
                          randomStarts = 100, ...) {

  L <- x$unrot_loadings

  if (rotation == "varimax") {

    resolved <- .resolve_settings(
      type = type,
      user = list(normalize = normalize, order_type = order_type,
                  varimax_type = varimax_type),
      preset = .efa_presets$VARIMAX
    )
    settings <- list(normalize = resolved$normalize, precision = precision,
                     order_type = resolved$order_type,
                     varimax_type = resolved$varimax_type)

    if (ncol(L) < 2) return(.rotate_single_factor(L, settings, oblique = FALSE))

    AV <- if (resolved$varimax_type == "svd") {
      .varimax_svd(L, normalize = resolved$normalize, precision = precision)
    } else {
      .VARIMAX_SPSS(L, normalize = resolved$normalize, precision = precision)
    }
    out <- .reflect_and_order(AV$loadings, rotmat = AV$rotmat, L_unrot = L,
                              order_type = resolved$order_type)
    return(c(out, list(settings = settings)))

  }

  if (rotation == "promax") {

    resolved <- .resolve_settings(
      type = type,
      user = list(normalize = normalize, P_type = P_type, order_type = order_type,
                  varimax_type = varimax_type, k = k),
      preset = .efa_presets$PROMAX
    )
    settings <- list(normalize = resolved$normalize, P_type = resolved$P_type,
                     precision = precision, order_type = resolved$order_type,
                     varimax_type = resolved$varimax_type, k = resolved$k)

    if (ncol(L) < 2) return(.rotate_single_factor(L, settings, oblique = TRUE))

    out <- .rotate_promax(L, normalize = resolved$normalize, P_type = resolved$P_type,
                          precision = precision, order_type = resolved$order_type,
                          varimax_type = resolved$varimax_type, k = resolved$k)
    .warn_extreme_phi(out$Phi)
    return(c(out, list(settings = settings)))

  }

  if (rotation %in% names(.orth_engines)) {

    resolved <- .resolve_settings(
      type = type,
      user = list(normalize = normalize, order_type = order_type),
      preset = .efa_presets$ROTATE_ORTH
    )
    settings <- list(normalize = resolved$normalize, precision = precision,
                     order_type = resolved$order_type, randomStarts = randomStarts)

    if (ncol(L) < 2) return(.rotate_single_factor(L, settings, oblique = FALSE))

    .warn_bifactor_trivial(rotation, ncol(L))

    ms <- .gpf_multistart_defaults[[rotation]]
    AV <- .orth_engines[[rotation]](L, eps = precision,
                                    normalize = resolved$normalize,
                                    randomStarts = randomStarts,
                                    screen_keep = ms$screen_keep,
                                    triage_maxit = ms$triage_maxit,
                                    triage_improve_tol = ms$triage_improve_tol, ...)
    settings$rotation_diagnostics <- .rotation_diagnostics(AV$all_values,
                                                           AV$all_converged, randomStarts,
                                                           AV$converged)
    out <- .reflect_and_order(AV$loadings, rotmat = AV$Th, L_unrot = L,
                              order_type = resolved$order_type)
    return(c(out, list(settings = settings)))

  }

  if (rotation %in% names(.oblq_engines)) {

    resolved <- .resolve_settings(
      type = type,
      user = list(normalize = normalize, order_type = order_type),
      preset = .efa_presets$ROTATE_OBLQ
    )
    settings <- list(normalize = resolved$normalize, precision = precision,
                     order_type = resolved$order_type, k = k,
                     randomStarts = randomStarts)

    if (ncol(L) < 2) return(.rotate_single_factor(L, settings, oblique = TRUE))

    .warn_bifactor_trivial(rotation, ncol(L))

    if (rotation == "simplimax" && is.na(k)) {
      k <- nrow(L)
      settings$k <- k
    }

    ms <- .gpf_multistart_defaults[[rotation]]
    AV <- .oblq_engines[[rotation]](L, k = k, eps = precision,
                                    normalize = resolved$normalize,
                                    randomStarts = randomStarts,
                                    screen_keep = ms$screen_keep,
                                    triage_maxit = ms$triage_maxit,
                                    triage_improve_tol = ms$triage_improve_tol, ...)
    settings$rotation_diagnostics <- .rotation_diagnostics(AV$all_values,
                                                           AV$all_converged, randomStarts,
                                                           AV$converged)
    out <- .reflect_and_order(AV$loadings, Phi = AV$Phi, rotmat = AV$Th,
                              L_unrot = L, order_type = resolved$order_type)
    .warn_extreme_phi(out$Phi)
    return(c(out, list(settings = settings)))

  }

  cli::cli_abort("Unknown rotation: {.val {rotation}}.", class = "efa_unknown_rotation")
}

unrot <- EFA(test_models$baseline$cormat, 3, N = 500)
equa <- .rotate_model(unrot, rotation = "equamax", type = "EFAtools")

unrot_1 <- EFA(test_models$baseline$cormat, 1, N = 500)
equa_1 <- suppressWarnings(.rotate_model(unrot_1, rotation = "equamax",
                                        type = "EFAtools"))

quarti <- .rotate_model(unrot, rotation = "quartimax", type = "psych")
bentT <- .rotate_model(unrot, rotation = "bentlerT", type = "none",
                      order_type = "eigen")
geoT <- .rotate_model(unrot, rotation = "geominT", type = "SPSS")
bifacT <- .rotate_model(unrot, rotation = "bifactorT", type = "EFAtools")

test_that("output class and dimensions are correct", {
  expect_s3_class(equa$rot_loadings, "LOADINGS")
  expect_s3_class(equa_1$rot_loadings, "LOADINGS") # The unrotated loadings here!
  expect_s3_class(quarti$rot_loadings, "LOADINGS")
  expect_s3_class(bentT$rot_loadings, "LOADINGS")
  expect_s3_class(geoT$rot_loadings, "LOADINGS")
  expect_s3_class(bifacT$rot_loadings, "LOADINGS")

  expect_output(str(equa), "List of 4")
  expect_output(str(equa_1), "List of 4")
  expect_output(str(quarti), "List of 4")
  expect_output(str(bentT), "List of 4")
  expect_output(str(geoT), "List of 4")
  expect_output(str(bifacT), "List of 4")

  checkmate::expect_matrix(equa$rotmat)
  checkmate::expect_matrix(equa$vars_accounted_rot)

  checkmate::expect_matrix(quarti$rotmat)
  checkmate::expect_matrix(quarti$vars_accounted_rot)

  checkmate::expect_matrix(bentT$rotmat)
  checkmate::expect_matrix(bentT$vars_accounted_rot)

  checkmate::expect_matrix(geoT$rotmat)
  checkmate::expect_matrix(geoT$vars_accounted_rot)

  checkmate::expect_matrix(bifacT$rotmat)
  checkmate::expect_matrix(bifacT$vars_accounted_rot)

  expect_equal(equa_1$rotmat, NA)
  expect_null(equa_1$vars_accounted_rot)
})

test_that("every orthogonal rotation labels its factor columns", {
  # The oblique counterpart of this test lives in test-ROTATE_OBLQ.R; together they cover
  # every rotation efa_fit() offers. Downstream consumers match factors by column name, so
  # the labels must be present for the criterion rotations too, not just for varimax.
  for (rot in .orth_rotations) {
    set.seed(42)
    fit <- suppressWarnings(
      efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
              estimator = "PAF", rotation = rot)
    )
    expect_equal(sort(colnames(fit$rot_loadings)), paste0("F", 1:3),
                 label = paste0("sorted rot_loadings colnames for ", rot))
  }
})

test_that("settings are returned correctly", {
  expect_named(equa$settings, c("normalize", "precision", "order_type", "randomStarts", "rotation_diagnostics"))
  expect_named(equa_1$settings,c("normalize", "precision", "order_type", "randomStarts"))
  expect_named(quarti$settings, c("normalize", "precision", "order_type", "randomStarts", "rotation_diagnostics"))
  expect_named(bentT$settings, c("normalize", "precision", "order_type", "randomStarts", "rotation_diagnostics"))
  expect_named(geoT$settings, c("normalize", "precision", "order_type", "randomStarts", "rotation_diagnostics"))
  expect_named(bifacT$settings, c("normalize", "precision", "order_type", "randomStarts", "rotation_diagnostics"))

  # the diagnostic carries forwarded per-start criterion values for every native
  # orthogonal rotation (a dropped all_values would leave criterion_best at NA)
  for (obj in list(equa, quarti, bentT, geoT, bifacT)) {
    expect_true(is.finite(obj$settings$rotation_diagnostics$criterion_best))
  }

  expect_equal(equa$settings$normalize, TRUE)
  expect_equal(equa_1$settings$normalize, TRUE)
  expect_equal(quarti$settings$normalize, TRUE)
  expect_equal(bentT$settings$normalize, TRUE)
  expect_equal(geoT$settings$normalize, TRUE)
  expect_equal(bifacT$settings$normalize, TRUE)

  expect_equal(equa$settings$precision, 1e-5)
  expect_equal(equa_1$settings$precision, 1e-5)
  expect_equal(quarti$settings$precision, 1e-5)
  expect_equal(bentT$settings$precision, 1e-5)
  expect_equal(geoT$settings$precision, 1e-5)
  expect_equal(bifacT$settings$precision, 1e-5)

  expect_equal(equa$settings$order_type, "eigen")
  expect_equal(equa_1$settings$order_type, "eigen")
  expect_equal(quarti$settings$order_type, "eigen")
  expect_equal(bentT$settings$order_type, "eigen")
  expect_equal(geoT$settings$order_type, "ss_factors")
  expect_equal(bifacT$settings$order_type, "eigen")
})

test_that("errors etc. are thrown correctly", {
  expect_error(.rotate_model(unrot, rotation = "equamax", type = "none"), class = "efa_type_none")

  expect_warning(.rotate_model(unrot, rotation = "equamax", type = "EFAtools",
                              normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "equamax", type = "EFAtools",
                              order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "equamax", type = "psych",
                              normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "equamax", type = "psych",
                              order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "equamax", type = "SPSS",
                              normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "equamax", type = "SPSS",
                              order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot_1, rotation = "equamax", type = "EFAtools"), class = "efa_single_factor")
})

test_that("a two-factor bifactorT request warns that the criterion is degenerate", {
  # The Jennrich-Bentler criterion sums the products of squared loadings over pairs of
  # distinct group factors, so with a general factor and a single group factor there is no
  # such pair: the criterion is identically zero, every rotation attains it, and the
  # identity start wins the tie. The returned solution is then the unrotated one, while the
  # multistart diagnostics (one distinct optimum, zero spread) read like a perfectly
  # identified rotation -- so the degeneracy has to be signalled. The oblique counterpart of
  # this test lives in test-ROTATE_OBLQ.R.
  unrot_2 <- EFA(test_models$baseline$cormat, 2, N = 500)

  expect_warning(.rotate_model(unrot_2, rotation = "bifactorT", type = "EFAtools"),
                 class = "efa_bifactor_trivial")

  # the warning reaches the public surface too, not just the rotation core
  expect_warning(
    efa_fit(test_models$baseline$cormat, n_factors = 2, N = 500, estimator = "PAF",
            rotation = "bifactorT"),
    class = "efa_bifactor_trivial"
  )

  res <- suppressWarnings(.rotate_model(unrot_2, rotation = "bifactorT",
                                        type = "EFAtools"))
  expect_equal(res$settings$rotation_diagnostics$criterion_best, 0)
  expect_equal(res$settings$rotation_diagnostics$criterion_spread, 0)

  # no rotation was performed: the returned loadings are the unrotated ones put through the
  # shared sign/order post-processing, i.e. exactly what an identity rotation produces
  ident <- .reflect_and_order(unclass(unrot_2$unrot_loadings), rotmat = diag(2),
                              L_unrot = unrot_2$unrot_loadings, order_type = "eigen")
  expect_equal(unclass(res$rot_loadings), unclass(ident$rot_loadings), ignore_attr = TRUE)

  # with three factors there are two group factors, so the criterion is non-degenerate
  set.seed(42)
  expect_no_warning(.rotate_model(unrot, rotation = "bifactorT", type = "EFAtools"),
                    class = "efa_bifactor_trivial")
})

test_that("the orthogonal engines validate their arguments before the compiled entry", {
  # The compiled entries check the same quantities, but they raise unclassed exceptions
  # whose messages name neither the function nor the argument, and a non-numeric value never
  # reaches them at all (the numeric coercion at the C boundary fails first). Validating in
  # R gives all of them one class that user code can catch. The oblique criterion parameters
  # are covered in test-ROTATE_OBLQ.R.
  cm <- test_models$baseline$cormat
  fit <- function(...) efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", ...)

  expect_error(fit(rotation = "geominT", delta = 0), class = "efa_rotation_arg")
  expect_error(fit(rotation = "geominT", delta = -0.01), class = "efa_rotation_arg")
  expect_error(fit(rotation = "geominT", delta = "a"), class = "efa_rotation_arg")

  expect_error(fit(rotation = "quartimax", maxit = -1), class = "efa_rotation_arg")
  # a fractional maxit was truncated at the C boundary, so the engine ran a different
  # budget from the one the non-convergence warning quoted
  expect_error(fit(rotation = "bentlerT", maxit = 2.7), class = "efa_rotation_arg")
})

test_that("the native orthogonal rotation matrices are orthogonal and reproduce the loadings", {
  skip_on_cran()

  # bentlerT and bifactorT are computed by the native gradient-projection engine. These
  # invariants hold for any valid orthogonal solution and need no reference package (so
  # coverage survives a criterion moving off GPArotation): the rotation matrix is orthogonal
  # (t(Th) %*% Th == I), and it reproduces the rotated loadings via the documented identity
  # L_unrot %*% Th == rot_loadings. Both solutions are checked in one loop, labelled by
  # criterion so a failure still names the engine that produced it. The
  # sign-reflection/reordering of the rotation matrix is exercised deterministically by the
  # orthogonal reflect-and-order test below.
  L <- unrot$unrot_loadings
  k <- ncol(L)
  solutions <- list(bentlerT = bentT, bifactorT = bifacT)

  for (nm in names(solutions)) {
    res <- solutions[[nm]]
    expect_equal(crossprod(res$rotmat), diag(k), ignore_attr = TRUE, tolerance = 1e-6,
                 info = nm)
    expect_equal(unclass(L) %*% res$rotmat, unclass(res$rot_loadings),
                 ignore_attr = TRUE, tolerance = 1e-6, info = nm)
  }
})

test_that("orthogonal loadings and rotmat are reflected/reordered consistently", {
  skip_on_cran()

  # Feed .reflect_and_order a raw orthogonal solution whose first factor has a negative column
  # sum, and check that the documented sign reflection is applied consistently to the loadings
  # and the rotation matrix (the orthogonal counterpart of the oblique reflect-and-order test).
  # This exercises the package's orthogonal reflection branch directly and deterministically,
  # rather than relying on the bentlerT engine to return a negative-column-sum factor (an
  # orientation that varies with the random starts). The raw input is a real bentlerT solution
  # with its first factor negated -- still a valid orthogonal solution (loadings = L %*% Th) --
  # so reflecting it back must recover the canonical solution.
  L <- unrot$unrot_loadings
  k <- ncol(bentT$rot_loadings)

  neg <- replace(rep(1, k), 1L, -1)
  raw_loadings <- unclass(bentT$rot_loadings) %*% diag(neg)
  raw_rotmat   <- bentT$rotmat %*% diag(neg)

  res <- .reflect_and_order(raw_loadings, rotmat = raw_rotmat, L_unrot = L,
                            order_type = "eigen")

  # the negative factor is reflected back to a non-negative column sum
  expect_true(all(colSums(unclass(res$rot_loadings)) >= 0))
  # loadings and the rotation matrix are reflected consistently, recovering the canonical solution
  expect_equal(unclass(res$rot_loadings), unclass(bentT$rot_loadings), ignore_attr = TRUE)
  expect_equal(res$rotmat, bentT$rotmat, ignore_attr = TRUE)
  # the rotation matrix stays orthogonal and reproduces the rotated loadings via L_unrot %*% rotmat
  expect_equal(crossprod(res$rotmat), diag(k), ignore_attr = TRUE, tolerance = 1e-6)
  expect_equal(unclass(L) %*% res$rotmat, unclass(res$rot_loadings),
               ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("the multi-start screen retains the lowest-criterion starts", {
  # As on the oblique manifold: under a zero-iteration budget every optimization reports the
  # criterion at the start it was handed, so the starts kept by a smaller screen must be the
  # minima of the criterion values seen when every start is kept (same seed, same draws).
  L <- unclass(unrot$unrot_loadings)

  set.seed(31)
  all_kept <- .rotate_cf_orth(L, kappa = 0, random_starts = 20L, screen_keep = 20L,
                              triage_maxit = 0L, maxit = 0L)
  set.seed(31)
  three_kept <- .rotate_cf_orth(L, kappa = 0, random_starts = 20L, screen_keep = 3L,
                                triage_maxit = 0L, maxit = 0L)

  # entry 1 is the primary (identity) start in both runs; the rest are the random ones
  expect_length(all_kept$all_values, 21L)
  expect_length(three_kept$all_values, 4L)
  expect_equal(sort(three_kept$all_values[-1]),
               sort(all_kept$all_values[-1])[1:3])
})

rm(unrot, equa, unrot_1, equa_1, quarti, bentT, geoT, bifacT)

# Regression net: pin the current rotation output against the GPArotation / stats engine
# each rotation is designed to wrap, so an accidental change (or a re-implementation of the
# rotation engine) is caught. The rotation is isolated on a single fixed unrotated solution
# and compared against the oracle on that same loading matrix. Sign and column-permutation
# differences are removed with the package's own efa_compare() (Tucker-congruence reorder plus
# sign reflection); the oblique factor correlations are compared via a permutation/sign
# invariant fingerprint. For the GPArotation engines the oracle is given randomStarts under a
# shared seed (100 for the smooth criteria, matching EFAtools; 300 for the multimodal simplimax
# criterion, a matched budget at which both engines reach the global basin). Tolerance is generous
# relative to the observed agreement (exact for every rotation here) so the contract is
# portable.

# max absolute difference after aligning columns/signs of two loading matrices
aligned_max_diff <- function(x, y) {
  efa_compare(x, y, reorder = "congruence", corres = FALSE,
              plot = FALSE, print_diff = FALSE)$max_abs_diff
}

# permutation/sign-invariant fingerprint of the factor intercorrelations
phi_fingerprint <- function(Phi) sort(abs(Phi[upper.tri(Phi)]))

unrot <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500)
L <- unclass(unrot$unrot_loadings)
seed <- 42

# shared fixture set for the native-engine relaxed-parity tests below (the non-convex CF,
# oblimin, and geomin criteria are exercised across three real correlation matrices of
# different factor counts so the Q-dominance contract is not pinned to a single solution)
fixtures <- list(
  baseline = list(R = test_models$baseline$cormat, N = 500,                nf = 3),
  GRiPS    = list(R = stats::cor(GRiPS_raw),        N = nrow(GRiPS_raw),    nf = 2),
  DOSPERT  = list(R = stats::cor(DOSPERT_raw),      N = nrow(DOSPERT_raw),  nf = 4)
)

# Every block below rotates the same unrotated solution per fixture, so each fixture is
# extracted once and cached. The cache is filled on first request rather than eagerly:
# every consuming block is skip_on_cran(), so on CRAN no extraction is paid for at all.
# Unrotated PAF extraction is deterministic and draws no random numbers, and each site
# re-seeds immediately before the rotation, so caching cannot perturb the random-start
# stream the engines below share with their oracles.
unrot_fixture <- local({
  cache <- list()
  function(fx_name) {
    if (!fx_name %in% names(cache)) {
      fx <- fixtures[[fx_name]]
      cache[[fx_name]] <<- suppressWarnings(EFA(fx$R, n_factors = fx$nf, N = fx$N))
    }
    cache[[fx_name]]
  }
})

test_that("quartimax and equamax route through the native CF rotation engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # quartimax and equamax are computed by the native Crawford-Ferguson engine
  # (kappa = 0 and kappa = k / (2 p) respectively), as are the remaining orthogonal
  # criteria. CF is non-convex, so parity is relaxed rather than byte-exact:
  # the engine must reach an as-good-or-better minimum of the SAME criterion, and the
  # aligned loadings must agree to ~1e-4 rather than the ~1e-6 the convex engines reach.
  # The like-for-like reference is cfT() with the matching kappa -- GPArotation's
  # quartimax() optimises the same rotation but reports a differently scaled criterion
  # value, so cfT(kappa = 0) is the correct Q reference for quartimax.
  # A rotation-engine regression shifts loadings by >= 1e-2 regardless of fixture, so
  # a single baseline fixture is enough to catch one; cross-fixture coverage is kept
  # for the multimodal/flat criteria below.
  for (fx_name in "baseline") {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)
    p <- nrow(Lx)
    k <- ncol(Lx)

    for (rn in c("quartimax", "equamax")) {
      kappa <- if (rn == "quartimax") 0 else k / (2 * p)

      set.seed(seed)
      native <- .rotate_cf_orth(Lx, kappa = kappa, eps = 1e-5, normalize = TRUE,
                                random_starts = 100)
      set.seed(seed)
      ref <- GPArotation::cfT(Lx, kappa = kappa, normalize = TRUE, eps = 1e-5,
                              randomStarts = 100)
      ref_Q <- ref$Table[nrow(ref$Table), 2]

      # (a) as-good-or-better minimum of the same criterion (same Q definition)
      expect_lte(native$value, ref_Q + 1e-6)
      # (b) aligned loadings within relaxed parity
      expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)

      # the public rotation path routes through the native engine and reproduces it
      set.seed(seed)
      efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = rn, type = "EFAtools"))
      expect_lt(aligned_max_diff(efa$rot_loadings, ref$loadings), 1e-4)
    }
  }
})

test_that("geominT routes through the native geomin GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # geominT is computed by the native gradient-projection engine, as are the remaining orthogonal
  # criteria. The geomin criterion is prone to local minima, so parity is
  # relaxed: the PRIMARY contract is that the native engine reaches an as-good-or-better minimum
  # of the SAME criterion. geominT's Table column 2 is the geomin value, computed on the same
  # Kaiser-normalized loadings the native engine reports, so it is directly comparable to
  # native$value. On these fixtures the criterion is well identified at randomStarts = 100, so
  # the aligned loadings also agree to ~1e-4; should an environment ever settle on a different
  # but equal-Q minimum, the Q-dominance check still passes and only the aligned-loadings check
  # would legitimately diverge.

  # the native engine pins the geomin offset at delta = 0.01, chosen to match GPArotation's
  # default; guard that the reference below is taken at the same offset
  expect_equal(formals(GPArotation::geominT)$delta, 0.01)

  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_geomin_orth(Lx, delta = 0.01, eps = 1e-5, normalize = TRUE,
                                  random_starts = 100)
    set.seed(seed)
    ref <- GPArotation::geominT(Lx, delta = 0.01, normalize = TRUE, eps = 1e-5,
                                randomStarts = 100)
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) as-good-or-better minimum of the same geomin criterion
    expect_lte(native$value, ref_Q + 1e-6)
    # (b) aligned loadings within relaxed parity (well-identified on these fixtures)
    expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)

    # the public rotation path routes through the native engine and reproduces it exactly (same
    # seed, same compiled entry), so it inherits the Q-dominance guarantee above
    set.seed(seed)
    efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = "geominT", type = "EFAtools"))
    expect_lt(aligned_max_diff(efa$rot_loadings, native$loadings), 1e-6)
  }
})

test_that("bentlerT routes through the native GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # bentlerT is computed by the native gradient-projection engine, as are the remaining orthogonal
  # criteria. The Bentler criterion is non-convex, so parity is relaxed: the
  # PRIMARY contract is that the native engine reaches an as-good-or-better minimum of the SAME
  # criterion. bentlerT's Table column 2 is the Bentler value computed on the same
  # Kaiser-normalized loadings the native engine reports, so it is directly comparable to
  # native$value. On these fixtures the criterion is well identified at randomStarts = 100, so the
  # aligned loadings also agree to ~1e-4.
  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_bentler_orth(Lx, eps = 1e-5, normalize = TRUE, random_starts = 100)
    set.seed(seed)
    ref <- GPArotation::bentlerT(Lx, normalize = TRUE, eps = 1e-5, randomStarts = 100)
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) as-good-or-better minimum of the same Bentler criterion
    expect_lte(native$value, ref_Q + 1e-6)
    # (b) aligned loadings within relaxed parity (well-identified on these fixtures)
    expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)

    # the public rotation path routes through the native engine and reproduces it exactly (same
    # seed, same compiled entry), so it inherits the Q-dominance guarantee above
    set.seed(seed)
    efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = "bentlerT", type = "EFAtools"))
    expect_lt(aligned_max_diff(efa$rot_loadings, native$loadings), 1e-6)
  }
})

test_that("bifactorT routes through the native GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # bifactorT is computed by the native gradient-projection engine; with this every orthogonal
  # criterion is native. The Jennrich-Bentler bifactor criterion (the first factor is a general
  # factor, exempt from the penalty) is strongly prone to local minima, so parity is relaxed: the
  # PRIMARY contract is that the native engine reaches an as-good-or-better minimum of the SAME
  # criterion. bifactorT's Table column 2 is the bifactor value computed on the same
  # Kaiser-normalized loadings the native engine reports, so it is directly comparable to
  # native$value. The criterion needs at least two group factors to be non-trivial: with a single
  # group factor (two factors) it is identically zero, so any rotation is optimal and the aligned
  # loadings may legitimately differ. On the k >= 3 fixtures the criterion is well identified at
  # randomStarts = 100, so the aligned loadings also agree to ~1e-4; where an environment settles
  # on a different but equal-Q minimum, the Q-dominance check still holds.
  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_bifactor_orth(Lx, eps = 1e-5, normalize = TRUE, random_starts = 100)
    set.seed(seed)
    ref <- GPArotation::bifactorT(Lx, normalize = TRUE, eps = 1e-5, randomStarts = 100)
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) as-good-or-better minimum of the same bifactor criterion (every fixture)
    expect_lte(native$value, ref_Q + 1e-6)
    # (b) aligned loadings within relaxed parity, where the criterion is non-trivial (k >= 3)
    if (ncol(Lx) >= 3) {
      expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)
    }

    # the public rotation path routes through the native engine and reproduces it exactly (same
    # seed, same compiled entry), so it inherits the Q-dominance guarantee above
    set.seed(seed)
    efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = "bifactorT", type = "EFAtools"))
    expect_lt(aligned_max_diff(efa$rot_loadings, native$loadings), 1e-6)
  }
})

test_that("simplimax routes through the native oblique GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")
  skip_if_not_slow()

  # simplimax is computed by the native gradient-projection engine;
  # unlike the smooth criteria, the simplimax criterion reselects the
  # k smallest squared loadings at every evaluation (k = nrow(L) by default), so it is only
  # piecewise smooth and strongly multimodal: its gradient jumps as loadings cross the kth-smallest threshold (the
  # native engine uses a non-monotone line search to step across the kinks) and the global minimum
  # depends on the random restarts. The native engine therefore fully optimizes every restart and
  # keeps the lowest-criterion solution -- the standard remedy for the local minima of
  # complexity-based criteria (Kiers, 1994; Browne, 2001) -- rather than the screen-and-triage
  # heuristic the smooth criteria use.
  #
  # Because the criterion is multimodal, native and GPArotation draw independent restart sequences
  # and need not reach the SAME minimizer -- and neither engine's attained VALUE is a stable
  # target either. The minima are dense: on DOSPERT an exact variable relabelling (a permutation
  # under which every reported statistic is invariant by construction, with the seed fixed so the
  # same starts are drawn) moves GPArotation's attained value by ~45% (0.004373 to 0.006397 over
  # 20 relabellings) while the native engine's moves ~6% (0.005571 to 0.005881). The oracle's own
  # value is therefore not identified to the precision a direct engine-to-engine comparison would
  # need, so no dominance claim is asserted here.
  #
  # What is identified is asserted instead: (a) the reported value really is the simplimax
  # criterion of the reported rotation; (b) the rotation descends from its unrotated start;
  # (c) the two engines agree in order of magnitude, which is the scale on which a genuine
  # regression in the engine shows up. Aligned-loadings parity is NOT asserted: at comparable Q the
  # two engines may settle on different, equally valid simple-structure rotations.
  #
  # The criterion is evaluated here from its definition rather than taken from the engine or from
  # GPArotation: the sum of the k smallest squared loadings (Kiers, 1994, Psychometrika, 59(4),
  # 567-579, doi:10.1007/BF02294392).
  simplimax_Q <- function(L, k) sum(sort(as.vector(unclass(L))^2)[seq_len(k)])

  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_simplimax_oblq(Lx, k = nrow(Lx), eps = 1e-5, normalize = TRUE,
                                     random_starts = 300)
    set.seed(seed)
    ref <- suppressWarnings(GPArotation::simplimax(Lx, k = nrow(Lx), normalize = TRUE,
                                                   eps = 1e-5, randomStarts = 300))
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) the reported value is the criterion of the reported rotation. normalize = TRUE optimizes
    # on the Kaiser-normalized loadings and un-normalizes the returned ones, so the comparison is
    # made back on the scale the engine optimizes. Both sides evaluate one closed form on one
    # matrix, so this holds to machine precision on any platform.
    h <- sqrt(rowSums(Lx^2))
    expect_equal(native$value, simplimax_Q(native$loadings / h, nrow(Lx)), tolerance = 1e-10)

    # (b) the rotation descends from its unrotated start
    expect_lt(native$value, simplimax_Q(Lx / h, nrow(Lx)))

    # (c) both engines land on the same scale. The factor is set from measurement in both
    # directions: at the intended 300 starts the worst native/ref_Q ratio seen on any fixture,
    # including the 20 relabellings, is 1.27, while collapsing the multistart search to a single
    # start -- the regression this block guards -- drives the ratio to 1.95 (GRiPS), 4.04 (DOSPERT)
    # and 5.14 (baseline). 1.5 sits between the two, so it catches a restart regression on every
    # fixture without pinning the labelling-induced spread. GRiPS is the binding fixture; a factor
    # of 2 would let a single-start collapse pass there unnoticed.
    expect_lt(native$value, 1.5 * ref_Q)
    # the native solution is a valid oblique rotation (unit-diagonal factor correlations)
    expect_equal(diag(native$Phi), rep(1, ncol(Lx)))
  }
})

test_that("simplimax holds its contract at the released default random-start budget", {
  skip_on_cran()

  # The block above compares the engines at a raised, matched budget. What a user actually
  # gets is the shipped default, and simplimax is materially start-dependent there, so the
  # properties that ARE identified at that budget are pinned here.
  #
  # No particular optimum and no cross-engine value is asserted: the criterion is
  # multimodal and its minima are dense, so neither engine's attained value is a stable
  # target (see the reasoning in the block above). What is identified is that every start is
  # optimized, that the reported value really is the criterion of the returned rotation,
  # that the starts disagree at this budget, and that spending more starts cannot cost
  # anything -- the engine draws its random starts serially in a fixed order, so at one seed
  # a larger budget optimizes a superset of the same starts.
  simplimax_Q <- function(M, k) sum(sort(as.vector(unclass(M))^2)[seq_len(k)])
  h <- sqrt(rowSums(L^2))

  # the released budget itself, on both surfaces that carry it: raising either silently
  # would make everything below vacuous
  expect_equal(formals(rotate_control)$random_starts, 100)
  expect_equal(formals(.rotate_model)$randomStarts, 100)

  set.seed(seed)
  def <- suppressWarnings(.rotate_model(unrot, rotation = "simplimax", type = "EFAtools"))
  rd <- def$settings$rotation_diagnostics

  # (a) simplimax runs full multistart: every requested start is optimized, none screened
  expect_equal(rd$n_starts_total, 101L)
  expect_equal(rd$n_optimized, rd$n_starts_total)

  # (b) the reported value is the simplimax criterion of the returned rotation, evaluated
  # on the Kaiser-normalized scale the engine optimizes on (the criterion is invariant to
  # the column reordering and sign reflection the returned solution has been through)
  expect_equal(rd$criterion_best,
               simplimax_Q(unclass(def$rot_loadings) / h, nrow(L)), tolerance = 1e-10)

  # (c) the starts genuinely disagree at the default budget -- this is the documented start
  # dependence, and the reason a reported simplimax solution needs a seed
  expect_gt(rd$n_distinct_minima, 1L)
  expect_gt(rd$criterion_spread, 0)

  # (d) a larger budget at the same seed extends the same draw sequence, so the selected
  # criterion can only improve
  set.seed(seed)
  wide <- suppressWarnings(.rotate_model(unrot, rotation = "simplimax", type = "EFAtools",
                                         randomStarts = 300))
  expect_lte(wide$settings$rotation_diagnostics$criterion_best, rd$criterion_best)
})

test_that("the public simplimax rotation path reproduces the native engine", {
  skip_on_cran()

  # the public rotation path routes through the native engine and reproduces it exactly (same seed,
  # same compiled entry, k defaults to nrow(L)); checked at a small restart budget for speed
  set.seed(seed)
  efa <- suppressWarnings(.rotate_model(unrot, rotation = "simplimax", type = "EFAtools",
                                        randomStarts = 10))
  set.seed(seed)
  nat <- .rotate_simplimax_oblq(L, k = nrow(L), eps = 1e-5, normalize = TRUE, random_starts = 10)
  expect_lt(aligned_max_diff(efa$rot_loadings, nat$loadings), 1e-6)
  expect_equal(phi_fingerprint(efa$Phi), phi_fingerprint(nat$Phi), tolerance = 1e-6)
})

test_that("oblimin and quartimin route through the native oblique GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # oblimin and quartimin are computed by the native gradient-projection engine (oblimin
  # with the GPArotation default gam = 0; quartimin is the same criterion); the remaining
  # oblique criteria are computed by the native engine too. The criterion is non-convex, so
  # parity is relaxed rather than byte-exact: the engine must reach an as-good-or-better minimum
  # of the SAME criterion, the aligned loadings must agree to ~1e-4, and the factor correlations
  # must match under the permutation/sign-invariant fingerprint. The single GPArotation::oblimin()
  # reference serves both public names (quartimin is the same gam = 0 criterion).
  # A single baseline fixture is enough to catch any rotation-engine regression here;
  # cross-fixture coverage is kept for the multimodal/flat criteria below.
  for (fx_name in "baseline") {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_oblimin(Lx, gam = 0, eps = 1e-5, normalize = TRUE,
                              random_starts = 100)
    set.seed(seed)
    ref <- suppressWarnings(GPArotation::oblimin(Lx, normalize = TRUE, eps = 1e-5,
                                                 randomStarts = 100))
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # the native engine reaches an as-good-or-better minimum of the same criterion, with
    # aligned loadings and matching factor correlations
    expect_lte(native$value, ref_Q + 1e-6)
    expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)
    expect_equal(phi_fingerprint(native$Phi), phi_fingerprint(ref$Phi), tolerance = 1e-4)

    # both public rotation names dispatch to the native engine and reproduce the reference
    for (rn in c("oblimin", "quartimin")) {
      set.seed(seed)
      efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = rn, type = "EFAtools"))
      expect_lt(aligned_max_diff(efa$rot_loadings, ref$loadings), 1e-4)
      expect_equal(phi_fingerprint(efa$Phi), phi_fingerprint(ref$Phi), tolerance = 1e-4)
    }
  }
})

test_that("the native oblimin engine generalizes to non-zero gamma", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # Exercise the gamma-centering path of the native oblimin engine with gam = 0.5
  # (biquartimin) against GPArotation's oblimin at the same gamma, under
  # the same relaxed-parity scheme (Q-dominance, aligned loadings, Phi fingerprint).
  set.seed(seed)
  native <- .rotate_oblimin(L, gam = 0.5, eps = 1e-5, normalize = TRUE, random_starts = 100)
  set.seed(seed)
  ref <- suppressWarnings(GPArotation::oblimin(L, gam = 0.5, normalize = TRUE, eps = 1e-5,
                                               randomStarts = 100))

  expect_lte(native$value, ref$Table[nrow(ref$Table), 2] + 1e-6)
  expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)
  expect_equal(phi_fingerprint(native$Phi), phi_fingerprint(ref$Phi), tolerance = 1e-4)
})

test_that("geominQ routes through the native geomin GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # geominQ is computed by the native gradient-projection engine, as are the remaining oblique
  # criteria. As with geominT, the geomin criterion is prone to local minima, so the
  # PRIMARY contract is an as-good-or-better minimum of the SAME criterion (geominQ's Table
  # column 2, directly comparable to native$value) rather than byte-exact reproduction. On these
  # fixtures the criterion is well identified at randomStarts = 100, so the aligned loadings and
  # the factor correlations also agree; where they diverged at equal Q, the Q-dominance check
  # would still hold.

  # the native engine pins the geomin offset at delta = 0.01, chosen to match GPArotation's
  # default; guard that the reference below is taken at the same offset
  expect_equal(formals(GPArotation::geominQ)$delta, 0.01)

  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    # geominQ uses a wider multistart by default than the bare compiled entry (the oblique geomin
    # criterion is multimodal enough that the two best-screened starts can both miss the global
    # basin), so the direct call is given the same per-criterion settings the public path applies,
    # ensuring both optimize identically.
    set.seed(seed)
    native <- do.call(.rotate_geomin_oblq,
                      c(list(Lx, delta = 0.01, eps = 1e-5, normalize = TRUE, random_starts = 100),
                        .gpf_multistart_defaults$geominQ))
    set.seed(seed)
    ref <- suppressWarnings(GPArotation::geominQ(Lx, delta = 0.01, normalize = TRUE,
                                                 eps = 1e-5, randomStarts = 100))
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) as-good-or-better minimum of the same geomin criterion
    expect_lte(native$value, ref_Q + 1e-6)
    # (b) aligned loadings and factor correlations within relaxed parity (well-identified here)
    expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-4)
    expect_equal(phi_fingerprint(native$Phi), phi_fingerprint(ref$Phi), tolerance = 1e-4)

    # the public rotation path routes through the native engine and reproduces it exactly (same
    # seed, same compiled entry), so it inherits the Q-dominance guarantee above
    set.seed(seed)
    efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = "geominQ", type = "EFAtools"))
    expect_lt(aligned_max_diff(efa$rot_loadings, native$loadings), 1e-6)
    expect_equal(phi_fingerprint(efa$Phi), phi_fingerprint(native$Phi), tolerance = 1e-6)
  }
})

test_that("bentlerQ routes through the native GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # bentlerQ is computed by the native gradient-projection engine, as are the remaining oblique
  # criteria. The Bentler criterion is non-convex AND comparatively flat near its
  # oblique optimum, so parity is relaxed: the PRIMARY contract is that the native engine reaches
  # an as-good-or-better minimum of the SAME criterion (bentlerQ's Table column 2, directly
  # comparable to native$value). Because the oblique optimum is flat, the native engine and
  # GPArotation settle on the same equal-Q solution only to ~1e-4 here (rather than the ~1e-6 of
  # the convex engines), so the aligned loadings and factor correlations are checked at a
  # correspondingly looser ~1e-3 tolerance; a genuine rotation regression shifts loadings by
  # >= 1e-2.
  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_bentler_oblq(Lx, eps = 1e-5, normalize = TRUE, random_starts = 100)
    set.seed(seed)
    ref <- suppressWarnings(GPArotation::bentlerQ(Lx, normalize = TRUE, eps = 1e-5,
                                                  randomStarts = 100))
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) as-good-or-better minimum of the same Bentler criterion
    expect_lte(native$value, ref_Q + 1e-6)
    # (b) aligned loadings and factor correlations within relaxed parity (flat oblique optimum)
    expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-3)
    expect_lt(max(abs(phi_fingerprint(native$Phi) - phi_fingerprint(ref$Phi))), 1e-3)

    # the public rotation path routes through the native engine and reproduces it exactly (same
    # seed, same compiled entry), so it inherits the Q-dominance guarantee above
    set.seed(seed)
    efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = "bentlerQ", type = "EFAtools"))
    expect_lt(aligned_max_diff(efa$rot_loadings, native$loadings), 1e-6)
    expect_lt(max(abs(phi_fingerprint(efa$Phi) - phi_fingerprint(native$Phi))), 1e-6)
  }
})

test_that("bifactorQ routes through the native GPF engine", {
  skip_on_cran()
  skip_if_not_installed("GPArotation")

  # bifactorQ is computed by the native gradient-projection engine; with this every analytic
  # rotation criterion is native. The Jennrich-Bentler bifactor criterion (the first factor is a
  # general factor, exempt from the penalty) is strongly prone to local minima and comparatively
  # flat near its oblique optimum, so parity is relaxed: the PRIMARY contract is that the native
  # engine reaches an as-good-or-better minimum of the SAME criterion (bifactorQ's Table column 2,
  # directly comparable to native$value). The criterion needs at least two group factors to be
  # non-trivial: with a single group factor (two factors) it is identically zero, so any rotation is
  # optimal. On the k >= 3 fixtures the native engine and GPArotation settle on the same equal-Q
  # solution only to ~1e-4 (the flat oblique optimum), so the aligned loadings and factor
  # correlations are checked at a looser ~1e-3 tolerance; a genuine rotation regression shifts
  # loadings by >= 1e-2.
  for (fx_name in names(fixtures)) {
    unrot_fx <- unrot_fixture(fx_name)
    Lx <- unclass(unrot_fx$unrot_loadings)

    set.seed(seed)
    native <- .rotate_bifactor_oblq(Lx, eps = 1e-5, normalize = TRUE, random_starts = 100)
    set.seed(seed)
    ref <- suppressWarnings(GPArotation::bifactorQ(Lx, normalize = TRUE, eps = 1e-5,
                                                   randomStarts = 100))
    ref_Q <- ref$Table[nrow(ref$Table), 2]

    # (a) as-good-or-better minimum of the same bifactor criterion (every fixture)
    expect_lte(native$value, ref_Q + 1e-6)
    # (b) aligned loadings and factor correlations within relaxed parity, where non-trivial (k >= 3)
    if (ncol(Lx) >= 3) {
      expect_lt(aligned_max_diff(native$loadings, ref$loadings), 1e-3)
      expect_lt(max(abs(phi_fingerprint(native$Phi) - phi_fingerprint(ref$Phi))), 1e-3)
    }

    # the public rotation path routes through the native engine and reproduces it exactly (same
    # seed, same compiled entry), so it inherits the Q-dominance guarantee above
    set.seed(seed)
    efa <- suppressWarnings(.rotate_model(unrot_fx, rotation = "bifactorQ", type = "EFAtools"))
    expect_lt(aligned_max_diff(efa$rot_loadings, native$loadings), 1e-6)
    expect_lt(max(abs(phi_fingerprint(efa$Phi) - phi_fingerprint(native$Phi))), 1e-6)
  }
})

test_that("varimax reproduces stats::varimax", {
  skip_on_cran()

  efa <- suppressWarnings(.rotate_model(unrot, rotation = "varimax", type = "psych"))   # varimax_type 'svd'
  ref <- stats::varimax(L, normalize = TRUE, eps = 1e-5)
  expect_lt(aligned_max_diff(efa$rot_loadings, ref$loadings), 1e-4)
})

test_that("kaiser varimax reproduces stats::varimax", {
  skip_on_cran()

  # The svd path is pinned above; this targets the Kaiser successive-pairwise
  # algorithm (varimax_type "kaiser", used by type "EFAtools"/"SPSS"). It optimises
  # the same Kaiser-normalised varimax criterion as stats::varimax's SVD updates but
  # via a different engine, so the loadings agree to ~5e-4 rather than exactly; this
  # pins the rotated loadings only, not the .SV criterion-monitoring path. The bound
  # sits well below a genuine rotation regression (which shifts loadings by >=1e-2)
  # while leaving headroom over the cross-engine difference for BLAS portability.
  ref <- stats::varimax(L, normalize = TRUE, eps = 1e-5)
  for (ty in c("EFAtools", "SPSS")) {
    efa <- suppressWarnings(.rotate_model(unrot, rotation = "varimax", type = ty))
    expect_lt(aligned_max_diff(efa$rot_loadings, ref$loadings), 2e-3)
  }
})

test_that("promax reproduces stats::promax", {
  skip_on_cran()

  # type 'psych' uses the unnormalized Hendrickson-White target (k = 4) on an svd varimax
  # base, which is what stats::promax(m = 4) computes.
  efa <- suppressWarnings(.rotate_model(unrot, rotation = "promax", type = "psych"))
  ref <- stats::promax(L, m = 4)
  expect_lt(aligned_max_diff(efa$rot_loadings, ref$loadings), 1e-4)
})

test_that("promax with normalize = FALSE reproduces psych's Promax", {
  skip_on_cran()
  skip_if_not_installed("psych")

  # Pins the documented claim that switching off Kaiser normalization reproduces the
  # psych promax solution "to within the varimax convergence tolerance". psych::Promax is
  # what psych::fa(rotate = "Promax") calls. Each package converges its own varimax base
  # (psych::Promax forwards eps through ... to GPArotation::Varimax), so the residual is
  # bounded by both tolerances rather than being an algorithmic difference, and the
  # equivalence is asserted at 1e-4 rather than at machine precision.
  efa <- suppressWarnings(
    .rotate_model(unrot, rotation = "promax", type = "psych", normalize = FALSE)
  )
  ref <- psych::Promax(L, m = 4)
  expect_lt(aligned_max_diff(efa$rot_loadings, ref$loadings), 1e-4)
  expect_equal(phi_fingerprint(efa$Phi), phi_fingerprint(ref$Phi), tolerance = 1e-4)
})

rm(aligned_max_diff, phi_fingerprint, unrot, L, seed, fixtures, unrot_fixture)

unrot <- EFA(test_models$baseline$cormat, 3, N = 500)
obli <- .rotate_model(unrot, rotation = "oblimin", type = "EFAtools")

unrot_1 <- EFA(test_models$baseline$cormat, 1, N = 500)
obli_1 <- suppressWarnings(.rotate_model(unrot_1, rotation = "oblimin",
                                        type = "EFAtools"))

quarti <- suppressWarnings(.rotate_model(unrot, rotation = "quartimin", type = "psych"))
simpli <- suppressWarnings(.rotate_model(unrot, rotation = "simplimax", type = "SPSS",
                       maxit = 2000, randomStarts = 1))
bentQ <- suppressWarnings(.rotate_model(unrot, rotation = "bentlerQ", type = "none",
                       order_type = "eigen"))
geoQ <- suppressWarnings(.rotate_model(unrot, rotation = "geominQ", type = "EFAtools"))
bifacQ <- suppressWarnings(.rotate_model(unrot, rotation = "bifactorQ", type = "EFAtools"))

test_that("output class and dimensions are correct", {
  expect_s3_class(obli$rot_loadings, "LOADINGS")
  expect_s3_class(obli_1$rot_loadings, "LOADINGS") # The unrotated loadings here!
  expect_s3_class(quarti$rot_loadings, "LOADINGS")
  expect_s3_class(simpli$rot_loadings, "LOADINGS")
  expect_s3_class(bentQ$rot_loadings, "LOADINGS")
  expect_s3_class(geoQ$rot_loadings, "LOADINGS")
  expect_s3_class(bifacQ$rot_loadings, "LOADINGS")

  expect_output(str(obli), "List of 6")
  expect_output(str(obli_1), "List of 6")
  expect_output(str(quarti), "List of 6")
  expect_output(str(simpli), "List of 6")
  expect_output(str(bentQ), "List of 6")
  expect_output(str(geoQ), "List of 6")
  expect_output(str(bifacQ), "List of 6")

  checkmate::expect_matrix(obli$Phi)
  expect_s3_class(obli$Structure, "LOADINGS")
  checkmate::expect_matrix(obli$rotmat)
  checkmate::expect_matrix(obli$vars_accounted_rot)

  checkmate::expect_matrix(quarti$Phi)
  expect_s3_class(quarti$Structure, "LOADINGS")
  checkmate::expect_matrix(quarti$rotmat)
  checkmate::expect_matrix(quarti$vars_accounted_rot)

  checkmate::expect_matrix(simpli$Phi)
  expect_s3_class(simpli$Structure, "LOADINGS")
  checkmate::expect_matrix(simpli$rotmat)
  checkmate::expect_matrix(simpli$vars_accounted_rot)

  checkmate::expect_matrix(bentQ$Phi)
  expect_s3_class(bentQ$Structure, "LOADINGS")
  checkmate::expect_matrix(bentQ$rotmat)
  checkmate::expect_matrix(bentQ$vars_accounted_rot)

  checkmate::expect_matrix(geoQ$Phi)
  expect_s3_class(geoQ$Structure, "LOADINGS")
  checkmate::expect_matrix(geoQ$rotmat)
  checkmate::expect_matrix(geoQ$vars_accounted_rot)

  checkmate::expect_matrix(bifacQ$Phi)
  expect_s3_class(bifacQ$Structure, "LOADINGS")
  checkmate::expect_matrix(bifacQ$rotmat)
  checkmate::expect_matrix(bifacQ$vars_accounted_rot)

  expect_null(obli_1$Phi)
  expect_null(obli_1$Structure)
  expect_equal(obli_1$rotmat, NA)
  expect_null(obli_1$vars_accounted_rot)
})

test_that("every oblique rotation labels its factor columns", {
  # Downstream consumers match factors by column name (reliability coefficients and the
  # Schmid-Leiman transformation both reorder by the "F<j>" labels), so every rotation --
  # not only the varimax-based ones -- must label the rotated pattern, the factor
  # intercorrelations, and the structure matrix. The labels are positional: the j-th column
  # of the ordered solution is always Fj (see .reflect_and_order).
  for (rot in .oblq_rotations) {
    set.seed(42)
    fit <- suppressWarnings(
      efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
              estimator = "PAF", rotation = rot)
    )
    fac_names <- colnames(fit$rot_loadings)
    expect_equal(sort(fac_names), paste0("F", 1:3),
                 label = paste0("sorted rot_loadings colnames for ", rot))
    expect_equal(dimnames(fit$Phi), list(fac_names, fac_names),
                 label = paste0("Phi dimnames for ", rot))
    expect_equal(colnames(fit$Structure), fac_names,
                 label = paste0("Structure colnames for ", rot))
  }
})

test_that("maxit bounds every optimization the multistart solver runs", {
  skip_on_cran()

  # The screen-and-triage stage has its own short iteration budget, but it must still
  # respect a user-supplied `maxit`: with maxit = 1 no start can converge, so the fit warns
  # whether or not random starts are in play. Before the cap was applied to the triage
  # stage, the random-start fit silently returned a fully converged triage solution and the
  # two calls below disagreed about whether the same maxit had converged.
  expect_warning(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "geominQ", maxit = 1),
    class = "efa_rotation_no_convergence"
  )
  expect_warning(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "geominQ", maxit = 1,
            rotate_control = rotate_control(random_starts = 0)),
    class = "efa_rotation_no_convergence"
  )
})

test_that("the non-convergence warning names the knobs that move the outcome", {
  skip_on_cran()

  # Both knobs are reachable on this path: the iteration budget through `maxit` and the
  # convergence tolerance through `precision`. The warning used to report only the budget it
  # had exhausted, leaving the reader to find either of them.
  w <- tryCatch(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "geominQ", maxit = 1),
    efa_rotation_no_convergence = function(w) w)
  expect_s3_class(w, "efa_rotation_no_convergence")
  expect_snapshot(cat(conditionMessage(w)))
})

test_that("settings are returned correctly", {
  expect_named(obli$settings, c("normalize", "precision", "order_type", "k", "randomStarts", "rotation_diagnostics"))
  expect_named(obli_1$settings, c("normalize", "precision", "order_type", "k", "randomStarts"))
  expect_named(quarti$settings, c("normalize", "precision", "order_type", "k", "randomStarts", "rotation_diagnostics"))
  expect_named(simpli$settings, c("normalize", "precision", "order_type", "k", "randomStarts", "rotation_diagnostics"))
  expect_named(bentQ$settings, c("normalize", "precision", "order_type", "k", "randomStarts", "rotation_diagnostics"))
  expect_named(geoQ$settings, c("normalize", "precision", "order_type", "k", "randomStarts", "rotation_diagnostics"))
  expect_named(bifacQ$settings, c("normalize", "precision", "order_type", "k", "randomStarts", "rotation_diagnostics"))

  # the diagnostic carries forwarded per-start criterion values for every native oblique
  # rotation (a dropped all_values would leave criterion_best at NA)
  for (obj in list(obli, quarti, simpli, bentQ, geoQ, bifacQ)) {
    expect_true(is.finite(obj$settings$rotation_diagnostics$criterion_best))
  }

  expect_equal(obli$settings$normalize, TRUE)
  expect_equal(obli_1$settings$normalize, TRUE)
  expect_equal(quarti$settings$normalize, TRUE)
  expect_equal(simpli$settings$normalize, TRUE)
  expect_equal(bentQ$settings$normalize, TRUE)
  expect_equal(geoQ$settings$normalize, TRUE)
  expect_equal(bifacQ$settings$normalize, TRUE)

  expect_equal(obli$settings$precision, 1e-5)
  expect_equal(obli_1$settings$precision, 1e-5)
  expect_equal(quarti$settings$precision, 1e-5)
  expect_equal(simpli$settings$precision, 1e-5)
  expect_equal(bentQ$settings$precision, 1e-5)
  expect_equal(geoQ$settings$precision, 1e-5)
  expect_equal(bifacQ$settings$precision, 1e-5)

  expect_equal(obli$settings$order_type, "eigen")
  expect_equal(obli_1$settings$order_type, "eigen")
  expect_equal(quarti$settings$order_type, "eigen")
  expect_equal(simpli$settings$order_type, "ss_factors")
  expect_equal(bentQ$settings$order_type, "eigen")
  expect_equal(geoQ$settings$order_type, "eigen")
  expect_equal(bifacQ$settings$order_type, "eigen")

  expect_equal(obli$settings$k, NA)
  expect_equal(obli_1$settings$k, NA)
  expect_equal(quarti$settings$k, NA)
  expect_equal(simpli$settings$k, 18)
  expect_equal(bentQ$settings$k, NA)
  expect_equal(geoQ$settings$k, NA)
  expect_equal(bifacQ$settings$k, NA)
})

test_that("errors etc. are thrown correctly", {
  expect_error(.rotate_model(unrot, rotation = "oblimin", type = "none"), class = "efa_type_none")

  expect_warning(.rotate_model(unrot, rotation = "oblimin", type = "EFAtools",
                              normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "oblimin", type = "EFAtools",
                              order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "oblimin", type = "psych",
                              normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "oblimin", type = "psych",
                              order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot, rotation = "oblimin", type = "SPSS",
                              normalize = FALSE), class = "efa_type_override")
  expect_warning(.rotate_model(unrot, rotation = "oblimin", type = "SPSS",
                              order_type = "ss_factors"), class = "efa_type_override")

  expect_warning(.rotate_model(unrot_1, rotation = "oblimin", type = "EFAtools"), class = "efa_single_factor")

  # A rotation name no engine table knows aborts rather than falling off the end of the
  # dispatch and returning NULL, which would surface far downstream as a missing solution.
  expect_error(.rotate_model(unrot, rotation = "obimin", type = "EFAtools"),
               class = "efa_unknown_rotation")
})

test_that("a two-factor bifactorQ request warns that the criterion is degenerate", {
  # The oblique counterpart of the bifactorT test in test-ROTATE_ORTH.R, which carries the
  # reasoning: with a single group factor the Jennrich-Bentler criterion is identically
  # zero, so the identity start wins and no rotation is performed.
  unrot_2 <- EFA(test_models$baseline$cormat, 2, N = 500)

  expect_warning(.rotate_model(unrot_2, rotation = "bifactorQ", type = "EFAtools"),
                 class = "efa_bifactor_trivial")

  res <- suppressWarnings(.rotate_model(unrot_2, rotation = "bifactorQ",
                                        type = "EFAtools"))
  expect_equal(res$settings$rotation_diagnostics$criterion_best, 0)
  # the identity transformation leaves the factors uncorrelated, i.e. the oblique solution
  # is the unrotated one
  expect_equal(res$Phi, diag(2), ignore_attr = TRUE)

  set.seed(42)
  expect_no_warning(.rotate_model(unrot, rotation = "bifactorQ", type = "EFAtools"),
                    class = "efa_bifactor_trivial")
})

test_that("the oblique engines validate their criterion arguments before the compiled entry", {
  # The oblique counterpart of the argument-validation test in test-ROTATE_ORTH.R, which
  # carries the reasoning: the compiled entries raise unclassed exceptions, so the contract
  # is enforced in R where the class and the argument name are both available.
  cm <- test_models$baseline$cormat
  fit <- function(...) efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", ...)

  expect_error(fit(rotation = "oblimin", gam = NA), class = "efa_rotation_arg")
  expect_error(fit(rotation = "oblimin", gam = "a"), class = "efa_rotation_arg")
  expect_error(fit(rotation = "oblimin", gam = c(0, 1)), class = "efa_rotation_arg")
  expect_error(fit(rotation = "oblimin", gam = Inf), class = "efa_rotation_arg")

  expect_error(fit(rotation = "geominQ", delta = 0), class = "efa_rotation_arg")
  expect_error(fit(rotation = "geominQ", delta = "a"), class = "efa_rotation_arg")

  expect_error(fit(rotation = "oblimin", maxit = -1), class = "efa_rotation_arg")
  expect_error(fit(rotation = "quartimin", maxit = 2.7), class = "efa_rotation_arg")

  # simplimax counts the loadings it drives toward zero, so `k` is a whole number bounded by
  # the number of loadings (18 * 3 here). `rotate_control()` already refuses a non-positive
  # `k` with its own class, so the lower bound is exercised on the engine directly.
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", rotation = "simplimax",
            rotate_control = rotate_control(k = 55)),
    class = "efa_rotation_arg"
  )
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", rotation = "simplimax",
            rotate_control = rotate_control(k = 2.5)),
    class = "efa_rotation_arg"
  )
  expect_error(.rotate_model(unrot, rotation = "simplimax", type = "EFAtools", k = 0),
               class = "efa_rotation_arg")
})

test_that("a singular oblique transformation is rejected rather than pseudo-inverted", {
  # The oblique manifold evaluates a candidate transformation T through its inverse, so a
  # rank-deficient candidate has to be rejected: substituting an approximate
  # (pseudo-inverse) solution returns something that is not a rotation at all, with a
  # rank-deficient Phi and pattern coefficients many orders of magnitude too large. The
  # oblique Procrustes entry exposes the same checked inverse directly through its T_init
  # nonsingularity check, which is the deterministic way to exercise the guard. (The
  # rejection is raised by the compiled entry, so it carries no efa_* class.)
  set.seed(1)
  A <- matrix(stats::rnorm(12), 6, 2)
  B <- matrix(stats::rnorm(12), 6, 2)

  # rank 1: column normalization maps both columns to (1, 0)
  expect_error(.oblique_procrustes(A, B, T_init = cbind(c(1, 0), c(2, 0))),
               class = "Rcpp::exception")
  expect_type(.oblique_procrustes(A, B, T_init = diag(2)), "list")

  # Rejection has to be decided by conditioning, not by exact rank. A candidate that is
  # merely near-singular still inverts -- every LAPACK returns a finite inverse for it --
  # but the inverse amplifies by 1 / sigma_min, so evaluating it yields the same collapsed
  # pseudo-rotation that an exactly rank-deficient candidate would. Whether such a
  # candidate is caught must not be left to whether the LU happens to produce a bit-exact
  # zero pivot, which differs between LAPACK builds. Two unit columns separated by an
  # angle of d have sigma_min ~ d / sqrt(2), so d straddles the 1e-4 floor here.
  near_singular <- function(d) cbind(c(1, 0), c(sqrt(1 - d^2), d))
  expect_error(.oblique_procrustes(A, B, T_init = near_singular(1e-8)),
               class = "Rcpp::exception")
  expect_error(.oblique_procrustes(A, B, T_init = near_singular(1e-5)),
               class = "Rcpp::exception")
  expect_type(.oblique_procrustes(A, B, T_init = near_singular(1e-2)), "list")
})

test_that("oblimin at a degenerate gam returns a valid rotation and reports the collapse", {
  # gam > 0 increasingly rewards correlated factors, and by gam = 1 the oblimin solution
  # on this fixture collapses toward a single factor: the transformation degenerates.
  # Rejecting a candidate whose smallest singular value falls below the 1e-4 floor keeps
  # the returned solution a genuine rotation instead of the collapsed artefact (Phi
  # off-diagonals of 1 to working precision, a negative smallest eigenvalue, pattern
  # coefficients of order 1e16). The solution is still degenerate, so it must not be
  # returned silently.
  #
  # The bounds below are the ones that floor guarantees -- min eig(Phi) = sigma_min^2 and
  # |A| / sigma_min for the pattern -- asserted an order of magnitude loose, because how
  # far past the floor a run actually stops depends on the rounding of the platform's
  # LAPACK. Where exactly the collapse is caught is not the contract; that it is caught
  # before Phi stops being a correlation matrix is.
  set.seed(42)
  res <- suppressWarnings(.rotate_model(unrot, rotation = "oblimin", type = "EFAtools",
                                        gam = 1))
  Phi <- res$Phi

  expect_true(all(is.finite(unclass(res$rot_loadings))))
  expect_lt(max(abs(unclass(res$rot_loadings))), 1e6)
  expect_lt(max(abs(Phi[upper.tri(Phi)])), 1 - 1e-9)
  expect_gt(min(eigen(Phi, symmetric = TRUE, only.values = TRUE)$values), 1e-9)

  set.seed(42)
  expect_warning(
    withCallingHandlers(
      .rotate_model(unrot, rotation = "oblimin", type = "EFAtools", gam = 1),
      efa_rotation_no_convergence = function(w) invokeRestart("muffleWarning")),
    class = "efa_rotation_extreme_phi"
  )
})

test_that("an oblique rotation warns when the factors are near-collinear", {
  # A factor correlation above .9 in absolute value means two rotated factors are barely
  # distinguishable, which is the usual signature of extracting more factors than the data
  # support. The fit is still returned -- an extreme correlation is a property of the
  # solution, not a failure of the rotation -- so only the interpretation is flagged.
  expect_warning(.warn_extreme_phi(matrix(c(1, -.95, -.95, 1), 2, 2)),
                 class = "efa_rotation_extreme_phi")
  expect_silent(.warn_extreme_phi(matrix(c(1, .6, .6, 1), 2, 2)))

  # the unit diagonal never triggers the check, and a single-factor solution carries no
  # factor correlations at all
  expect_silent(.warn_extreme_phi(matrix(1, 1, 1)))
  expect_silent(.warn_extreme_phi(NULL))

  # The reported correlation is truncated, not rounded. A near-collapse must not print as
  # "1.000" -- that is the perfectly collinear Phi a rejected singular transformation would
  # have produced, and reporting it would undo the distinction. A value just above the
  # cut-off must likewise not print as "0.900", which reads as sitting at the cut-off.
  # (The condition is identified by class; only the number it carries is read back.)
  reported <- function(r) {
    w <- tryCatch(.warn_extreme_phi(matrix(c(1, r, r, 1), 2, 2)),
                  efa_rotation_extreme_phi = function(w) conditionMessage(w))
    sub(".*max \\|r\\| = ([0-9.]+).*", "\\1", w)
  }
  expect_equal(reported(0.9999846784), "0.999")
  expect_equal(reported(0.905), "0.905")

  # a well-conditioned oblique fit is not flagged. Asserted on this class alone rather
  # than with expect_silent(): the multistart rotation's own convergence flag is
  # BLAS-sensitive, and that warning is not what this test is about.
  set.seed(42)
  expect_no_warning(.rotate_model(unrot, rotation = "oblimin", type = "EFAtools"),
                    class = "efa_rotation_extreme_phi")

  # the check sits on both oblique return paths: promax finalizes its own solution and is
  # flagged the same way when its two factors are all but the same factor
  L <- cbind(c(.80, .75, .70, .65, .60, .55),
             c(.78, .74, .69, .66, .58, .56))
  expect_warning(.rotate_model(list(unrot_loadings = L), rotation = "promax",
                               type = "EFAtools"),
                 class = "efa_rotation_extreme_phi")
})

test_that("order_type orders oblique factors by the requested key", {
  # "eigen" orders by the reported SS loadings (the factor-intercorrelation-weighted
  # sum of squares diag(Phi L'L)); "ss_factors" orders by the unweighted pattern sum
  # of squares colSums(L^2). For an oblique solution these keys can rank the factors
  # differently, so the two order_type values can yield different factor orders.

  # On a real oblimin solution whose two keys rank the factors the same way, both
  # order_type values produce the identical, internally consistent solution sorted
  # (descending) by the relevant key.
  set.seed(42)
  eig <- suppressWarnings(.rotate_model(unrot, rotation = "oblimin", type = "EFAtools"))
  set.seed(42)
  ss <- suppressWarnings(.rotate_model(unrot, rotation = "oblimin", type = "SPSS"))
  expect_equal(unclass(ss$rot_loadings), unclass(eig$rot_loadings), ignore_attr = TRUE)
  expect_equal(unclass(eig$Structure), unclass(eig$rot_loadings) %*% eig$Phi,
               ignore_attr = TRUE)
  expect_equal(unclass(ss$Structure), unclass(ss$rot_loadings) %*% ss$Phi,
               ignore_attr = TRUE)
  expect_false(is.unsorted(rev(diag(eig$Phi %*% crossprod(unclass(eig$rot_loadings))))))
  expect_false(is.unsorted(rev(colSums(unclass(ss$rot_loadings)^2))))

  # Feed .reflect_and_order a fixed oblique solution whose two keys rank the factors
  # differently, and check that each order_type sorts by its own key, producing
  # different orders while keeping the solution consistent (structure = pattern %*% Phi).
  L <- rbind(
    c(0.85, 0.10, 0.05),
    c(0.80, 0.05, 0.00),
    c(0.05, 0.75, 0.55),
    c(0.10, 0.55, 0.75),
    c(0.00, 0.50, 0.70),
    c(0.05, 0.45, 0.65)
  )
  Phi <- matrix(c(1.00, 0.20, 0.20,
                  0.20, 1.00, 0.65,
                  0.20, 0.65, 1.00), 3, 3)
  expect_false(identical(order(colSums(L^2), decreasing = TRUE),
                         order(diag(Phi %*% crossprod(L)), decreasing = TRUE)))

  eig2 <- .reflect_and_order(L, Phi = Phi, rotmat = diag(3), L_unrot = L,
                             order_type = "eigen")
  ss2 <- .reflect_and_order(L, Phi = Phi, rotmat = diag(3), L_unrot = L,
                            order_type = "ss_factors")
  expect_false(is.unsorted(rev(diag(eig2$Phi %*% crossprod(unclass(eig2$rot_loadings))))))
  expect_false(is.unsorted(rev(colSums(unclass(ss2$rot_loadings)^2))))
  expect_false(isTRUE(all.equal(unclass(eig2$rot_loadings), unclass(ss2$rot_loadings),
                                check.attributes = TRUE)))
  expect_equal(unclass(eig2$Structure), unclass(eig2$rot_loadings) %*% eig2$Phi,
               ignore_attr = TRUE)
  expect_equal(unclass(ss2$Structure), unclass(ss2$rot_loadings) %*% ss2$Phi,
               ignore_attr = TRUE)
})

test_that("oblique Phi, structure, and rotmat are reflected/reordered with the loadings", {
  skip_on_cran()

  # Feed .reflect_and_order a raw oblique solution whose first factor has a negative column
  # sum, and check that the documented sign reflection is applied consistently to the
  # loadings, the factor intercorrelations, and the rotation matrix (a
  # structure == pattern %*% Phi check alone misses the Phi/rotmat reflection). This
  # exercises the package's own reflection branch directly and deterministically, rather
  # than relying on a rotation engine to return a negative-column-sum factor (an orientation
  # that varies across GPArotation versions). The raw input is a real oblimin solution with
  # its first factor negated -- still a valid oblique solution (loadings = L %*% t(solve(Th)),
  # Phi = t(Th) %*% Th) -- so reflecting it back must recover the canonical solution.
  L <- unrot$unrot_loadings
  o <- suppressWarnings(.rotate_model(unrot, rotation = "oblimin", type = "EFAtools"))
  k <- ncol(o$rot_loadings)

  neg <- replace(rep(1, k), 1L, -1)
  raw_loadings <- unclass(o$rot_loadings) %*% diag(neg)
  raw_Phi      <- diag(neg) %*% o$Phi %*% diag(neg)
  raw_rotmat   <- o$rotmat %*% diag(neg)

  res <- .reflect_and_order(raw_loadings, Phi = raw_Phi, rotmat = raw_rotmat,
                            L_unrot = L, order_type = "eigen")

  # the negative factor is reflected back to a non-negative column sum
  expect_true(all(colSums(unclass(res$rot_loadings)) >= 0))
  # loadings, Phi, and the rotation matrix are reflected consistently, recovering the
  # canonical solution
  expect_equal(unclass(res$rot_loadings), unclass(o$rot_loadings), ignore_attr = TRUE)
  expect_equal(res$Phi, o$Phi, ignore_attr = TRUE)
  expect_equal(res$rotmat, o$rotmat, ignore_attr = TRUE)
  # the structure matrix stays pattern %*% Phi, and the rotation matrix reproduces the
  # rotated pattern via L_unrot %*% t(solve(rotmat))
  expect_equal(unclass(res$Structure), unclass(res$rot_loadings) %*% res$Phi,
               ignore_attr = TRUE)
  expect_equal(unclass(L) %*% t(solve(res$rotmat)), unclass(res$rot_loadings),
               ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("the native oblique solutions satisfy the oblique structure invariants", {
  skip_on_cran()

  # bentlerQ, bifactorQ, and simplimax are computed by the native gradient-projection engine.
  # These invariants hold for any valid oblique solution and need no reference package (so
  # coverage survives a criterion moving off GPArotation): the structure matrix equals
  # pattern %*% Phi, the rotation matrix reproduces the rotated pattern via the documented
  # identity L_unrot %*% t(solve(Th)), and the factor correlation matrix has a unit diagonal.
  # The three solutions are checked in one loop, labelled by criterion so a failure still
  # names the engine that produced it.
  L <- unrot$unrot_loadings
  solutions <- list(bentlerQ = bentQ, bifactorQ = bifacQ, simplimax = simpli)

  for (nm in names(solutions)) {
    res <- solutions[[nm]]
    expect_equal(unclass(res$Structure), unclass(res$rot_loadings) %*% res$Phi,
                 ignore_attr = TRUE, info = nm)
    expect_equal(unclass(L) %*% t(solve(res$rotmat)), unclass(res$rot_loadings),
                 ignore_attr = TRUE, tolerance = 1e-6, info = nm)
    expect_equal(diag(res$Phi), rep(1, ncol(res$rot_loadings)), ignore_attr = TRUE,
                 info = nm)
  }
})

test_that("the multi-start screen retains the lowest-criterion starts", {
  # The screen must rank a random start by the criterion value the solver itself reports at
  # that start. Under a zero-iteration budget every optimization returns exactly that value,
  # so screening all starts lists the criterion at all of them, and screening the same starts
  # (same seed, same draws) with a smaller keep must retain their minima.
  L <- unclass(unrot$unrot_loadings)

  set.seed(31)
  all_kept <- .rotate_oblimin(L, gam = 0, random_starts = 20L, screen_keep = 20L,
                              triage_maxit = 0L, maxit = 0L)
  set.seed(31)
  three_kept <- .rotate_oblimin(L, gam = 0, random_starts = 20L, screen_keep = 3L,
                                triage_maxit = 0L, maxit = 0L)

  # entry 1 is the primary (identity) start in both runs; the rest are the random ones
  expect_length(all_kept$all_values, 21L)
  expect_length(three_kept$all_values, 4L)
  expect_equal(sort(three_kept$all_values[-1]),
               sort(all_kept$all_values[-1])[1:3])
})

rm(unrot, obli, unrot_1, obli_1, quarti, simpli, bentQ, geoQ, bifacQ)

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

.fit_ml_cpp <- function(R, n_fac, start, lower) {
    .Call(`_EFAtools_fit_ml_cpp`, R, n_fac, start, lower)
}

.error_ml <- function(psi, R, n_fac) {
    .Call(`_EFAtools_error_ml`, psi, R, n_fac)
}

.grad_ml <- function(psi, R, n_fac) {
    .Call(`_EFAtools_grad_ml`, psi, R, n_fac)
}

.fit_uls_cpp <- function(R, n_fac, start, lower) {
    .Call(`_EFAtools_fit_uls_cpp`, R, n_fac, start, lower)
}

.uls_residuals <- function(psi, R, n_fac) {
    .Call(`_EFAtools_uls_residuals`, psi, R, n_fac)
}

.grad_uls <- function(psi, R, n_fac) {
    .Call(`_EFAtools_grad_uls`, psi, R, n_fac)
}

.fit_dwls_cpp <- function(R, n_fac, W) {
    .Call(`_EFAtools_fit_dwls_cpp`, R, n_fac, W)
}

.dwls_residuals <- function(par, R, n_fac, W) {
    .Call(`_EFAtools_dwls_residuals`, par, R, n_fac, W)
}

.grad_dwls <- function(par, R, n_fac, W) {
    .Call(`_EFAtools_grad_dwls`, par, R, n_fac, W)
}

#' Compute number of non-matching indicator-to-factor correspondences
#'
#' @param x numeric matrix. A matrix of pattern coefficients.
#' @param y numeric matrix. A second matrix of coefficients.
#' @param thresh numeric. The threshold to classify a pattern coefficient as substantial.
#'
#' @keywords internal
.factor_corres <- function(x, y, thresh = 0.3) {
    .Call(`_EFAtools_factor_corres`, x, y, thresh)
}

#' Oblique Procrustes target rotation using a k x k inner objective
#'
#' Compute an oblique target rotation for a loading matrix using a
#' `targetQ`-compatible parameterization and a `k x k` objective.
#'
#' The rotated loading matrix is defined as
#' `L = A %*% solve(t(T))`, and the corresponding factor correlation matrix is
#' `Phi = t(T) %*% T`. The optimization is carried out over the transformation
#' matrix `T` under the oblique normalization constraint `diag(t(T) %*% T) = 1`.
#'
#' Non-invertible candidate transformations are rejected rather than evaluated
#' through a pseudo-inverse.
#'
#' Additional random starts may be requested. To reduce runtime, the solver uses
#' a two-stage strategy for extra starts: cheap objective screening, followed by
#' short triage optimization, followed by full optimization only for starts that
#' improve on the current incumbent by at least `triage_improve_tol`.
#'
#' @param A Numeric matrix. Loading matrix to be rotated.
#' @param B Numeric matrix. Target loading matrix with the same dimensions as
#'   `A`.
#' @param S_r Optional numeric `k x k` matrix containing `crossprod(A)`.
#'   Supplying this is useful when the same `A` is rotated repeatedly. Ignored
#'   when `normalize = TRUE` because the normalized cross-product is different.
#' @param T_init_r Optional numeric `k x k` starting transformation matrix.
#'   If `NULL`, the identity matrix is used for the primary start.
#' @param eps Numeric scalar. Convergence tolerance for the projected-gradient
#'   norm.
#' @param maxit Integer scalar. Maximum number of full projected-gradient
#'   updates.
#' @param max_line_search Integer scalar. Maximum number of step-halving
#'   attempts after the initial trial step in each line-search phase.
#' @param step0 Numeric scalar. Initial step size used in the projected-gradient
#'   update.
#' @param normalize Logical scalar. If `TRUE`, apply Kaiser normalization to the
#'   loadings (only) before rotation and reverse it afterwards; the target is left
#'   unnormalized, matching `GPArotation::targetQ(normalize = TRUE)`.
#' @param random_starts Integer scalar. Number of additional random starts.
#' @param screen_keep Integer scalar. Number of screened random starts retained
#'   for triage optimization.
#' @param triage_maxit Integer scalar. Number of short optimization iterations
#'   used in the triage stage.
#' @param triage_improve_tol Numeric scalar. Relative improvement required for a
#'   triaged start to be promoted to full optimization.
#'
#' @returns A named list containing the rotated loadings, transformation matrix,
#'   factor correlation matrix, target criterion value, convergence diagnostics,
#'   line-search diagnostics, and multi-start summaries.
#'
#' @details
#' The routine is intended for repeated oblique target rotations in workflows
#' such as bootstrap alignment or consensus alignment of exploratory factor
#' solutions across multiply imputed datasets. It follows the same oblique
#' transformation convention as `GPArotation::targetQ()`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms
#' and software for arbitrary rotation criteria in factor analysis.
#' *Educational and Psychological Measurement*, 65, 676-696.
#'
#' Browne, M. W. (2001). An overview of analytic rotation in exploratory factor
#' analysis. *Multivariate Behavioral Research*, 36, 111-150.
#'
#' @keywords internal
#'
.oblique_procrustes <- function(A, B, S_r = NULL, T_init_r = NULL, eps = 1e-5, maxit = 1000L, max_line_search = 10L, step0 = 1.0, normalize = FALSE, random_starts = 0L, screen_keep = 2L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_oblique_procrustes`, A, B, S_r, T_init_r, eps, maxit, max_line_search, step0, normalize, random_starts, screen_keep, triage_maxit, triage_improve_tol)
}

#' Batched oblique Procrustes target rotation over a cube of loading matrices
#'
#' Align each slice of a loading-matrix cube to a single shared target using the
#' same oblique target rotation as `.oblique_procrustes()`, in one call. This
#' removes the per-replicate marshalling overhead of looping `efa_procrustes()` in R
#' over bootstrap or multiple-imputation arrays.
#'
#' Each slice `A[, , i]` is aligned to `B`. For a single-factor cube the alignment
#' reduces to the closed-form sign match `T = sign(crossprod(A_i, B))` with factor
#' correlation `1`, matching the one-factor short-circuit in `efa_procrustes()`. For
#' two or more factors the slice is warm-started from the closed-form orthogonal
#' Procrustes solution (mirroring `efa_procrustes()`) and optimized with the same
#' multi-start oblique solver as `.oblique_procrustes()`. Random starts are drawn
#' serially with `R::rnorm` in the calling process.
#'
#' Slices are aligned independently. A slice that cannot be aligned (a non-finite
#' loading matrix, a failed warm-start decomposition, an invalid fit, or any
#' linear-algebra exception) is reported with `valid = FALSE` and `NA` for the
#' loadings, factor correlations, and all other per-slice diagnostics, rather than
#' aborting the whole call, so one degenerate replicate does not discard the rest.
#'
#' @param A Numeric array of dimension `n x m x b`: the `b` loading matrices to
#'   align.
#' @param B Numeric `n x m` target loading matrix shared across all slices.
#' @param eps Numeric scalar. Convergence tolerance for the projected-gradient
#'   norm.
#' @param maxit Integer scalar. Maximum number of full projected-gradient updates.
#' @param max_line_search Integer scalar. Maximum number of step-halving attempts
#'   after the initial trial step in each line-search phase.
#' @param step0 Numeric scalar. Initial step size used in the projected-gradient
#'   update.
#' @param normalize Logical scalar. If `TRUE`, apply Kaiser normalization to the
#'   loadings (only) before rotation and reverse it afterwards, leaving the target
#'   unnormalized (ignored for single-factor slices).
#' @param random_starts Integer scalar. Number of additional random starts per
#'   slice.
#' @param screen_keep Integer scalar. Number of screened random starts retained
#'   for triage optimization.
#' @param triage_maxit Integer scalar. Number of short optimization iterations
#'   used in the triage stage.
#' @param triage_improve_tol Numeric scalar. Relative improvement required for a
#'   triaged start to be promoted to full optimization.
#'
#' @returns A named list with the aligned-loadings array `loadings` (`n x m x b`),
#'   the factor-correlation array `Phi` (`m x m x b`), and the per-slice
#'   diagnostics `valid`, `convergence`, `value`, `iterations`, and
#'   `line_search_failed`.
#'
#' @keywords internal
#'
.oblique_procrustes_batch <- function(A, B, eps = 1e-5, maxit = 1000L, max_line_search = 10L, step0 = 1.0, normalize = FALSE, random_starts = 0L, screen_keep = 2L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_oblique_procrustes_batch`, A, B, eps, maxit, max_line_search, step0, normalize, random_starts, screen_keep, triage_maxit, triage_improve_tol)
}

#' Perform the iterative PAF procedure
#'
#' Function called from within PAF so usually no call to this is needed by the user.
#' Provides a C++ implementation of the PAF procedure
#'
#' @param h2 numeric. The initial communality estimates.
#' @param criterion double. The convergence criterion to use.
#' @param R matrix. The correlation matrix with the initial communality estimates in the diagonal.
#' @param n_fac numeric. The number of factors to extract.
#' @param abs_eig logical. Whether absolute eigenvalues should be used to compute the loadings.
#' @param crit_type numeric. Whether maximum absolute differences (crit_type = 1), or sum of differences (crit_type = 2) should be used
#' @param max_iter numeric. The number of iterations after which to end the procedure if no convergence has been reached by then.
#'
#' @keywords internal
.paf_iter <- function(h2, criterion, R, n_fac, abs_eig, crit_type, max_iter) {
    .Call(`_EFAtools_paf_iter`, h2, criterion, R, n_fac, abs_eig, crit_type, max_iter)
}

#' Parallel analysis on simulated data.
#'
#' Function called from within efa_parallel() so usually no call to this is needed by the user.
#' Provides a C++ implementation of the efa_parallel() simulation procedure
#'
#' @param n_datasets numeric. Number of datasets with dimensions (N, n_vars) to simulate.
#' @param n_vars numeric. Number of variables / indicators in dataset.
#' @param N numeric. Number of cases / observations in dataset.
#' @param eigen_type numeric. Whether PCA (eigen_type = 1; i.e., leaving diagonal of correlation matrix at 1), SMC (eigen_type = 2; i.e., setting diagonal of correlation matrix to SMCs), or both from the same simulated datasets (eigen_type = 3), in which case the returned matrix holds the PCA eigenvalues in the first n_vars columns and the SMC eigenvalues in the next n_vars.
#' @param maxit numeric. Maximum iterations to perform after which to abort.
#'
#' @keywords internal
.parallel_sim <- function(n_datasets, n_vars, N, eigen_type, maxit = 10000L) {
    .Call(`_EFAtools_parallel_sim`, n_datasets, n_vars, N, eigen_type, maxit)
}

.polychoric_cpp <- function(x, acov, nearest_pd) {
    .Call(`_EFAtools_polychoric_cpp`, x, acov, nearest_pd)
}

.bvn_rect_cpp <- function(a0, a1, b0, b1, rho) {
    .Call(`_EFAtools_bvn_rect_cpp`, a0, a1, b0, b1, rho)
}

#' Orthogonal Crawford-Ferguson factor rotation
#'
#' Rotate a loading matrix orthogonally under the Crawford-Ferguson criterion using a
#' gradient-projection optimizer along the orthogonal (Stiefel) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% T` define the search; the engine maps the gradient to the orthogonal
#' transformation `T`, projects it onto the tangent space, performs a
#' non-monotone line search, and retracts back onto the orthogonal group via a
#' polar (singular value) projection. `kappa = 0` is the quartimax criterion and
#' `kappa = ncol(A) / (2 * nrow(A))` is the equamax criterion.
#'
#' Additional random orthogonal starts may be requested. To bound runtime the solver
#' screens each random start by its objective, runs a short triage optimization on the
#' best-screened starts, and fully optimizes only those that improve on the current
#' incumbent by at least `triage_improve_tol`.
#'
#' @param L Numeric matrix. The unrotated loading matrix (variables by factors).
#' @param kappa Numeric scalar in `[0, 1]`. The Crawford-Ferguson parameter.
#' @param eps Numeric scalar. Convergence tolerance for the projected-gradient norm.
#' @param normalize Logical scalar. If `TRUE`, apply Kaiser normalization before
#'   rotation and reverse it afterwards.
#' @param random_starts Integer scalar. Number of additional random orthogonal starts.
#' @param maxit Integer scalar. Maximum number of projected-gradient updates.
#' @param max_line_search Integer scalar. Maximum number of step-halving attempts after
#'   the initial trial step in each line-search phase.
#' @param step0 Numeric scalar. Initial step size used in the projected-gradient update.
#' @param screen_keep Integer scalar. Number of screened random starts retained for
#'   triage optimization.
#' @param triage_maxit Integer scalar. Number of short optimization iterations used in
#'   the triage stage.
#' @param triage_improve_tol Numeric scalar. Relative improvement required for a triaged
#'   start to be promoted to full optimization.
#'
#' @returns A named list with the rotated loadings, the orthogonal rotation matrix `Th`
#'   (with `L %*% Th` reproducing the rotated loadings), the attained criterion value, and
#'   the convergence and validity flags. The list additionally reports the
#'   criterion value reached at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Crawford, C. B., & Ferguson, G. A. (1970). A general rotation criterion and its use
#' in orthogonal rotation. *Psychometrika*, 35, 321-332.
#'
#' @keywords internal
.rotate_cf_orth <- function(L, kappa, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_cf_orth`, L, kappa, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Oblique oblimin factor rotation
#'
#' Rotate a loading matrix obliquely under the oblimin criterion using a
#' gradient-projection optimizer along the oblique (column-normalized) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% solve(t(T))` define the search; the engine maps the gradient to the
#' transformation `T` on the manifold `diag(t(T) %*% T) = 1`, projects it onto the tangent
#' space, performs a non-monotone line search, and retracts back onto the manifold
#' by column normalization. `gam = 0` is the quartimin criterion.
#'
#' Additional random starts may be requested. To bound runtime the solver screens each
#' random start by its objective, runs a short triage optimization on the best-screened
#' starts, and fully optimizes only those that improve on the current incumbent by at
#' least `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#' @param gam Numeric scalar. The oblimin parameter; `gam = 0` is the quartimin criterion
#'   and the recommended setting. Larger values increasingly reward correlated factors and
#'   can drive the solution toward factor collapse, where the rotation may also fail to
#'   converge; inspect `Phi` before interpreting a fit with `gam > 0`.
#'
#' @returns A named list with the rotated loadings, the transformation matrix `Th`
#'   (with `L %*% t(solve(Th))` reproducing the rotated loadings), the factor correlation
#'   matrix `Phi` (`t(Th) %*% Th`), the attained criterion value, and the convergence and
#'   validity flags. The list additionally reports the criterion value reached
#'   at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Jennrich, R. I., & Sampson, P. F. (1966). Rotation for simple loadings.
#' *Psychometrika*, 31, 313-323.
#'
#' @keywords internal
.rotate_oblimin <- function(L, gam = 0.0, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_oblimin`, L, gam, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Orthogonal geomin factor rotation
#'
#' Rotate a loading matrix orthogonally under the geomin criterion using a
#' gradient-projection optimizer along the orthogonal (Stiefel) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% T` define the search; the engine maps the gradient to the orthogonal
#' transformation `T`, projects it onto the tangent space, performs a non-monotone
#' line search, and retracts back onto the orthogonal group via a polar (singular value)
#' projection. The geomin criterion sums the per-variable geometric mean of the squared
#' loadings offset by `delta`; it is prone to local minima, so additional random starts are
#' recommended.
#'
#' Additional random orthogonal starts may be requested. To bound runtime the solver screens
#' each random start by its objective, runs a short triage optimization on the best-screened
#' starts, and fully optimizes only those that improve on the current incumbent by at least
#' `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#' @param delta Numeric scalar. The geomin offset added to the squared loadings; must be a
#'   positive finite scalar. `delta = 0.01` is the usual default.
#'
#' @returns A named list with the rotated loadings, the orthogonal rotation matrix `Th`
#'   (with `L %*% Th` reproducing the rotated loadings), the attained criterion value, and the
#'   convergence and validity flags. The list additionally reports the criterion
#'   value reached at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Browne, M. W. (2001). An overview of analytic rotation in exploratory factor analysis.
#' *Multivariate Behavioral Research*, 36, 111-150.
#'
#' @keywords internal
.rotate_geomin_orth <- function(L, delta = 0.01, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_geomin_orth`, L, delta, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Oblique geomin factor rotation
#'
#' Rotate a loading matrix obliquely under the geomin criterion using a gradient-projection
#' optimizer along the oblique (column-normalized) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% solve(t(T))` define the search; the engine maps the gradient to the
#' transformation `T` on the manifold `diag(t(T) %*% T) = 1`, projects it onto the tangent
#' space, performs a non-monotone line search, and retracts back onto the manifold by
#' column normalization. The geomin criterion sums the per-variable geometric mean of the
#' squared loadings offset by `delta`; it is prone to local minima, so additional random
#' starts are recommended.
#'
#' Additional random starts may be requested. To bound runtime the solver screens each random
#' start by its objective, runs a short triage optimization on the best-screened starts, and
#' fully optimizes only those that improve on the current incumbent by at least
#' `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#' @param delta Numeric scalar. The geomin offset added to the squared loadings; must be a
#'   positive finite scalar. `delta = 0.01` is the usual default.
#'
#' @returns A named list with the rotated loadings, the transformation matrix `Th`
#'   (with `L %*% t(solve(Th))` reproducing the rotated loadings), the factor correlation
#'   matrix `Phi` (`t(Th) %*% Th`), the attained criterion value, and the convergence and
#'   validity flags. The list additionally reports the criterion value reached
#'   at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Browne, M. W. (2001). An overview of analytic rotation in exploratory factor analysis.
#' *Multivariate Behavioral Research*, 36, 111-150.
#'
#' @keywords internal
.rotate_geomin_oblq <- function(L, delta = 0.01, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_geomin_oblq`, L, delta, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Orthogonal Bentler factor rotation
#'
#' Rotate a loading matrix orthogonally under Bentler's invariant pattern simplicity criterion
#' using a gradient-projection optimizer along the orthogonal (Stiefel) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% T` define the search; the engine maps the gradient to the orthogonal
#' transformation `T`, projects it onto the tangent space, performs a non-monotone line
#' search, and retracts back onto the orthogonal group via a polar (singular value) projection.
#' The Bentler criterion measures the departure of the cross-products of squared loadings from a
#' diagonal pattern; it is prone to local minima, so additional random starts are recommended.
#'
#' Additional random orthogonal starts may be requested. To bound runtime the solver screens
#' each random start by its objective, runs a short triage optimization on the best-screened
#' starts, and fully optimizes only those that improve on the current incumbent by at least
#' `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#'
#' @returns A named list with the rotated loadings, the orthogonal rotation matrix `Th`
#'   (with `L %*% Th` reproducing the rotated loadings), the attained criterion value, and the
#'   convergence and validity flags. The list additionally reports the criterion
#'   value reached at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bentler, P. M. (1977). Factor simplicity index and transformations. *Psychometrika*, 42,
#' 277-295.
#'
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' @keywords internal
.rotate_bentler_orth <- function(L, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_bentler_orth`, L, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Oblique Bentler factor rotation
#'
#' Rotate a loading matrix obliquely under Bentler's invariant pattern simplicity criterion
#' using a gradient-projection optimizer along the oblique (column-normalized) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% solve(t(T))` define the search; the engine maps the gradient to the
#' transformation `T` on the manifold `diag(t(T) %*% T) = 1`, projects it onto the tangent
#' space, performs a non-monotone line search, and retracts back onto the manifold by
#' column normalization. The Bentler criterion measures the departure of the cross-products of
#' squared loadings from a diagonal pattern; it is prone to local minima, so additional random
#' starts are recommended.
#'
#' Additional random starts may be requested. To bound runtime the solver screens each random
#' start by its objective, runs a short triage optimization on the best-screened starts, and
#' fully optimizes only those that improve on the current incumbent by at least
#' `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#'
#' @returns A named list with the rotated loadings, the transformation matrix `Th`
#'   (with `L %*% t(solve(Th))` reproducing the rotated loadings), the factor correlation
#'   matrix `Phi` (`t(Th) %*% Th`), the attained criterion value, and the convergence and
#'   validity flags. The list additionally reports the criterion value reached
#'   at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bentler, P. M. (1977). Factor simplicity index and transformations. *Psychometrika*, 42,
#' 277-295.
#'
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' @keywords internal
.rotate_bentler_oblq <- function(L, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_bentler_oblq`, L, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Orthogonal bifactor factor rotation
#'
#' Rotate a loading matrix orthogonally under the Jennrich-Bentler bifactor criterion using a
#' gradient-projection optimizer along the orthogonal (Stiefel) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% T` define the search; the engine maps the gradient to the orthogonal
#' transformation `T`, projects it onto the tangent space, performs a non-monotone line
#' search, and retracts back onto the orthogonal group via a polar (singular value) projection.
#' The first factor is treated as a general factor and is exempt from the penalty; the criterion
#' measures the between-group-factor cross-products of the squared loadings, so it is minimized
#' when each variable loads on the general factor plus at most one group factor. The criterion is
#' prone to local minima, so additional random starts are recommended.
#'
#' Additional random orthogonal starts may be requested. To bound runtime the solver screens
#' each random start by its objective, runs a short triage optimization on the best-screened
#' starts, and fully optimizes only those that improve on the current incumbent by at least
#' `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#'
#' @returns A named list with the rotated loadings, the orthogonal rotation matrix `Th`
#'   (with `L %*% Th` reproducing the rotated loadings), the attained criterion value, and the
#'   convergence and validity flags. The list additionally reports the criterion
#'   value reached at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Jennrich, R. I., & Bentler, P. M. (2011). Exploratory bi-factor analysis. *Psychometrika*,
#' 76, 537-549.
#'
#' @keywords internal
.rotate_bifactor_orth <- function(L, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_bifactor_orth`, L, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Oblique bifactor factor rotation
#'
#' Rotate a loading matrix obliquely under the Jennrich-Bentler bifactor criterion using a
#' gradient-projection optimizer along the oblique (column-normalized) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% solve(t(T))` define the search; the engine maps the gradient to the
#' transformation `T` on the manifold `diag(t(T) %*% T) = 1`, projects it onto the tangent
#' space, performs a non-monotone line search, and retracts back onto the manifold by
#' column normalization. The first factor is treated as a general factor and is exempt from the
#' penalty; the criterion measures the between-group-factor cross-products of the squared
#' loadings, so it is minimized when each variable loads on the general factor plus at most one
#' group factor. The criterion is prone to local minima, so additional random starts are
#' recommended.
#'
#' Additional random starts may be requested. To bound runtime the solver screens each random
#' start by its objective, runs a short triage optimization on the best-screened starts, and
#' fully optimizes only those that improve on the current incumbent by at least
#' `triage_improve_tol`.
#'
#' @inheritParams .rotate_cf_orth
#'
#' @returns A named list with the rotated loadings, the transformation matrix `Th`
#'   (with `L %*% t(solve(Th))` reproducing the rotated loadings), the factor correlation
#'   matrix `Phi` (`t(Th) %*% Th`), the attained criterion value, and the convergence and
#'   validity flags. The list additionally reports the criterion value reached
#'   at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Jennrich, R. I., & Bentler, P. M. (2011). Exploratory bi-factor analysis. *Psychometrika*,
#' 76, 537-549.
#'
#' @keywords internal
.rotate_bifactor_oblq <- function(L, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0, screen_keep = 5L, triage_maxit = 25L, triage_improve_tol = 0.0) {
    .Call(`_EFAtools_rotate_bifactor_oblq`, L, eps, normalize, random_starts, maxit, max_line_search, step0, screen_keep, triage_maxit, triage_improve_tol)
}

#' Oblique simplimax factor rotation
#'
#' Rotate a loading matrix obliquely under the simplimax criterion using a gradient-projection
#' optimizer along the oblique (column-normalized) manifold.
#'
#' The criterion value `f` and its gradient `dQ/dL` at the rotated loadings
#' `L = A %*% solve(t(T))` define the search; the engine maps the gradient to the
#' transformation `T` on the manifold `diag(t(T) %*% T) = 1`, projects it onto the tangent
#' space, performs a non-monotone line search, and retracts back onto the manifold by column
#' normalization. The simplimax criterion sums the `k` smallest squared loadings, so it is
#' minimized when the `k` "close-to-zero" loadings are driven toward zero; the count `k` is a
#' tuning parameter. Because the set of `k` smallest loadings is reselected at every evaluation,
#' the criterion is only piecewise smooth: its gradient jumps as loadings cross the kth-smallest threshold, so the
#' line search accepts a step whenever it decreases the largest objective over a short window of
#' recent iterations (a non-monotone test; Grippo, Lampariello, & Lucidi, 1986), letting the
#' optimizer step across the kinks where a strictly monotone descent would stall.
#'
#' The criterion is strongly prone to local minima, so the solver fully optimizes the identity
#' start together with `random_starts` random orthogonal starts and keeps the solution with the
#' lowest criterion value. Fully optimizing every start -- rather than the screen-and-triage
#' strategy used for the smooth criteria, which assumes the rational start lies in the global
#' basin -- is the standard remedy for the local minima of complexity-based rotation criteria
#' (Kiers, 1994; Browne, 2001).
#'
#' @inheritParams .rotate_cf_orth
#' @param k Integer scalar. The number of "close-to-zero" loadings the criterion targets; must
#'   be in `[1, nrow(L) * ncol(L)]`. `k = nrow(L)` is the usual default.
#' @param eps Numeric scalar. Convergence tolerance for the projected-gradient norm. Because the
#'   simplimax criterion is only piecewise smooth, the projected gradient need not reach this
#'   tolerance at the optimum; convergence is then reported when the criterion value stalls (the
#'   non-monotone search described above), so `eps` mainly governs the smooth phases of the search.
#' @param random_starts Integer scalar. Number of random orthogonal starts fully optimized in
#'   addition to the identity start.
#' @param maxit Integer scalar. Maximum number of projected-gradient updates per start.
#'
#' @returns A named list with the rotated loadings, the transformation matrix `Th`
#'   (with `L %*% t(solve(Th))` reproducing the rotated loadings), the factor correlation
#'   matrix `Phi` (`t(Th) %*% Th`), the attained criterion value, and the convergence and
#'   validity flags. The list additionally reports the criterion value reached
#'   at each optimized start in `all_values`, with a per-start convergence flag in `all_converged`.
#'
#' @references
#' Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
#' software for arbitrary rotation criteria in factor analysis. *Educational and
#' Psychological Measurement*, 65, 676-696.
#'
#' Browne, M. W. (2001). An overview of analytic rotation in exploratory factor analysis.
#' *Multivariate Behavioral Research*, 36, 111-150.
#'
#' Grippo, L., Lampariello, F., & Lucidi, S. (1986). A nonmonotone line search technique for
#' Newton's method. *SIAM Journal on Numerical Analysis*, 23, 707-716.
#'
#' Kiers, H. A. L. (1994). Simplimax: Oblique rotation to an optimal target with simple
#' structure. *Psychometrika*, 59, 567-579.
#'
#' @keywords internal
.rotate_simplimax_oblq <- function(L, k, eps = 1e-5, normalize = TRUE, random_starts = 0L, maxit = 1000L, max_line_search = 10L, step0 = 1.0) {
    .Call(`_EFAtools_rotate_simplimax_oblq`, L, k, eps, normalize, random_starts, maxit, max_line_search, step0)
}

#' Rotation Jacobians for analytic rotation standard errors
#'
#' Forward-difference the warm-started re-rotation map `A -> (rotated loadings, Phi)` over the
#' unrotated loadings `A` to obtain the rotation Jacobians used by the analytic standard errors for
#' rotated loadings (`se = "information"` in [efa_fit()]). The full `nrow(A) * ncol(A)` finite-
#' difference loop runs in compiled code, re-solving the rotation from the converged transformation
#' `T_init` at each perturbation; the caller forms `J V J'` in R.
#'
#' @param A Numeric matrix. The unrotated loading matrix at the solution.
#' @param T_init Numeric matrix. The converged transformation that warm-starts each re-rotation.
#' @param method Character scalar. The criterion family: one of `"cf"`, `"oblimin"`, `"geomin"`,
#'   `"bentler"`, `"bifactor"`.
#' @param param Numeric scalar. The criterion's tuning argument (`kappa` for `"cf"`, `gam` for
#'   `"oblimin"`, `delta` for `"geomin"`); ignored for `"bentler"` and `"bifactor"`.
#' @param normalize Logical scalar. Apply Kaiser normalization before rotation and reverse it after.
#' @param oblique Logical scalar. Use the oblique (column-normalized) manifold; otherwise orthogonal.
#' @param eps Numeric scalar. The forward-difference step on the loadings.
#' @param general_col Integer scalar. For `method = "bifactor"`, the zero-based column holding the
#'   general factor in the (factor-reordered) reported solution; ignored by the other criteria.
#'
#' @returns A named list with the Jacobian `J_L` (`pk x pk`), the re-rotated `base_loadings`, a
#'   validity flag, and -- when oblique -- the Jacobian `J_Phi` (`k^2 x pk`) and `base_Phi`.
#'
#' @references
#' Jennrich, R. I. (1973). Standard errors for obliquely rotated factor loadings.
#' *Psychometrika*, 38, 593-604.
#'
#' @keywords internal
.rotation_se_jacobian <- function(A, T_init, method, param, normalize, oblique, eps, general_col = 0L) {
    .Call(`_EFAtools_rotation_se_jacobian`, A, T_init, method, param, normalize, oblique, eps, general_col)
}

#' Draw multivariate-normal data from a population correlation matrix.
#'
#' Internal helper called from efa_simulate(). Draws `N` cases from a
#' `p`-variate normal with correlation (or covariance) `R` by post-multiplying a
#' matrix of standard normal deviates by a matrix square root `M` of `R` (with
#' M' M = R, so the rows of Z * M are N(0, R)). This is the same Z * M rule used by
#' the NEST reference simulation (.simulate_cfm_eigen): there `M` is the transposed
#' factor-score matrix, here it is a Cholesky or eigen square root.
#' A positive-definite `R` is factored by Cholesky; a positive-semidefinite but
#' singular `R` (which makes the Cholesky fail although it is still a valid
#' covariance, e.g. a no-factor block or a smoothed factor intercorrelation
#' matrix) falls back to a symmetric eigen square root.
#'
#' @param R numeric matrix. Population correlation/covariance matrix.
#' @param N integer. Number of cases to draw.
#' @param tol numeric. Eigenvalues below `-tol` mark `R` as indefinite.
#'
#' @keywords internal
.simulate_cfm_mvn <- function(R, N, tol = 1e-8) {
    .Call(`_EFAtools_simulate_cfm_mvn`, R, N, tol)
}

#' Reference eigenvalues for the efa_nest() simulation via the shared kernel.
#'
#' Internal helper called from efa_nest(). Simulates `nreps` datasets from an
#' `(nf - 1)`-factor reference model, given that model's loadings `Lambda` and
#' uniquenesses `Psi`, and returns the `nf`-th largest eigenvalue of each simulated
#' correlation matrix. The data are drawn with the shared Z * M rule (see
#' .simulate_cfm_mvn()) using the factor-score square root
#' `M = t([Lambda | diag(sqrt(Psi))])`, so a row `randn(1, nf - 1 + p) * M` is
#' N(0, Lambda Lambda' + diag(Psi)). Drawing `nf - 1 + p` standard normals and
#' post-multiplying by the factor-score matrix is faster than forming the model-
#' implied correlation matrix and drawing from it, and matches the position at which
#' efa_nest() reads the reference eigenvalue.
#'
#' @param nf integer. Position of the empirical eigenvalue being tested (1-based);
#'   the `nf`-th largest simulated eigenvalue is returned per replicate.
#' @param N integer. Number of cases / observations per simulated dataset.
#' @param Lambda numeric matrix. Loadings of the `(nf - 1)`-factor reference model
#'   (`p x (nf - 1)`); pass a `p x 0` matrix for the `nf == 1` null (identity) model.
#' @param Psi numeric vector. Uniquenesses (`1 - h2`) of the reference model.
#' @param nreps integer. Number of datasets to simulate.
#'
#' @keywords internal
.simulate_cfm_eigen <- function(nf, N, Lambda, Psi, nreps = 1000L) {
    .Call(`_EFAtools_simulate_cfm_eigen`, nf, N, Lambda, Psi, nreps)
}

ULS_test <- .estimate_model(method = "ULS",test_models$baseline$cormat, n_factors = 3, N = 500)
ULS_test_1 <- .estimate_model(method = "ULS",test_models$baseline$cormat, n_factors = 1, N = 500)

test_that("output class and dimensions are correct", {
  expect_s3_class(ULS_test$unrot_loadings, "LOADINGS")
  expect_output(str(ULS_test), "List of 12")
  expect_s3_class(ULS_test_1$unrot_loadings, "LOADINGS")
  expect_output(str(ULS_test_1), "List of 12")

})

test_that("outputs are correct", {
  expect_equal(ULS_test$orig_R, test_models$baseline$cormat)
  expect_equal(sum(ULS_test$orig_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(ULS_test$final_eigen), ncol(test_models$baseline$cormat))
  expect_equal(ULS_test$convergence, 0)

  expect_equal(ULS_test_1$orig_R, test_models$baseline$cormat)
  expect_equal(sum(ULS_test_1$orig_eigen), ncol(test_models$baseline$cormat))
  expect_lt(sum(ULS_test_1$final_eigen), ncol(test_models$baseline$cormat))
  expect_equal(ULS_test_1$convergence, 0)

  # The reported eigenvalues are those of the original and of the reduced correlation
  # matrix, in decreasing order. Checked against the eigenvector-returning LAPACK
  # driver, which is a different routine from the values-only one used to compute them,
  # so the tolerance is set well above their disagreement at the noise floor (~6e-15)
  # and well below anything a real regression could hide in. Both factor counts are
  # checked: the single-factor path is where the extraction degenerates.
  for (fit in list(ULS_test, ULS_test_1)) {
    R_final <- test_models$baseline$cormat
    diag(R_final) <- fit$h2
    expect_equal(fit$orig_eigen,
                 eigen(test_models$baseline$cormat, symmetric = TRUE)$values,
                 tolerance = 1e-10)
    expect_equal(fit$final_eigen, eigen(R_final, symmetric = TRUE)$values,
                 tolerance = 1e-10)
    expect_equal(fit$orig_eigen, sort(fit$orig_eigen, decreasing = TRUE))
    expect_equal(fit$final_eigen, sort(fit$final_eigen, decreasing = TRUE))
  }
})

test_that("fit indices are returned correctly", {
  expect_output(str(ULS_test$fit_indices), "List of 18")
  expect_output(str(ULS_test_1$fit_indices), "List of 18")

  expect_type(ULS_test$fit_indices$chi, "double")
  expect_type(ULS_test$fit_indices$df, "double")
  expect_type(ULS_test$fit_indices$p_chi, "double")
  expect_type(ULS_test$fit_indices$CAF, "double")
  expect_type(ULS_test$fit_indices$CFI, "double")
  expect_type(ULS_test$fit_indices$RMSEA, "double")
  expect_type(ULS_test$fit_indices$RMSEA_LB, "double")
  expect_type(ULS_test$fit_indices$RMSEA_UB, "double")
  expect_type(ULS_test$fit_indices$AIC, "double")
  expect_type(ULS_test$fit_indices$BIC, "double")
  expect_type(ULS_test$fit_indices$Fm, "double")
  expect_type(ULS_test$fit_indices$chi_null, "double")
  expect_type(ULS_test$fit_indices$df_null, "double")
  expect_type(ULS_test$fit_indices$p_null, "double")

  expect_type(ULS_test_1$fit_indices$chi, "double")
  expect_type(ULS_test_1$fit_indices$df, "double")
  expect_type(ULS_test_1$fit_indices$p_chi, "double")
  expect_type(ULS_test_1$fit_indices$CAF, "double")
  expect_type(ULS_test_1$fit_indices$CFI, "double")
  expect_type(ULS_test_1$fit_indices$RMSEA, "double")
  expect_type(ULS_test_1$fit_indices$RMSEA_LB, "double")
  expect_type(ULS_test_1$fit_indices$RMSEA_UB, "double")
  expect_type(ULS_test_1$fit_indices$AIC, "double")
  expect_type(ULS_test_1$fit_indices$BIC, "double")
  expect_type(ULS_test_1$fit_indices$Fm, "double")
  expect_type(ULS_test_1$fit_indices$chi_null, "double")
  expect_type(ULS_test_1$fit_indices$df_null, "double")
  expect_type(ULS_test_1$fit_indices$p_null, "double")
})

test_that("the ULS objective is half the squared Frobenius norm of the reduced residual", {
  # The criterion is 0.5 * ||R - diag(psi) - LL'||_F^2 with L the top-n_fac eigen
  # extraction of the reduced matrix (negative eigenvalues clipped to zero), and its
  # gradient is the negated diagonal of that residual.
  set.seed(4)
  p <- 7L; m <- 2L
  A <- matrix(stats::rnorm(p * p), p, p)
  Rg <- stats::cov2cor(A %*% t(A))
  psi <- stats::runif(p, 0.25, 0.75)

  Rs <- Rg
  diag(Rs) <- diag(Rs) - psi
  ev <- eigen(Rs, symmetric = TRUE)
  lam <- pmax(ev$values[seq_len(m)], 0)
  L <- ev$vectors[, seq_len(m), drop = FALSE] %*% diag(sqrt(lam), nrow = m)
  E <- Rs - tcrossprod(L)

  expect_equal(.uls_residuals(psi, Rg, m), 0.5 * sum(E^2))
  expect_equal(as.vector(.grad_uls(psi, Rg, m)), -diag(E))
})

test_that("the ULS gradient matches finite differences", {
  # value and gradient must differentiate the same function: checked against central
  # differences at a non-trivial point away from the optimum and off the box boundary
  set.seed(3)
  p <- 9L; m <- 3L
  A <- matrix(stats::rnorm(p * p), p, p)
  Rg <- stats::cov2cor(A %*% t(A))
  psi <- stats::runif(p, 0.25, 0.75)

  ga <- as.vector(.grad_uls(psi, Rg, m))
  h <- 1e-6
  gn <- vapply(seq_along(psi), function(k) {
    e <- numeric(length(psi)); e[k] <- h
    (.uls_residuals(psi + e, Rg, m) - .uls_residuals(psi - e, Rg, m)) / (2 * h)
  }, numeric(1))

  expect_equal(ga, gn, tolerance = 1e-5)
})

test_that("MINRES is accepted as a synonym for ULS", {
  uls <- suppressWarnings(EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                              method = "ULS"))
  minres <- suppressWarnings(EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                                 method = "MINRES"))

  # minimum residual and unweighted least squares are the same estimator
  expect_equal(minres$unrot_loadings, uls$unrot_loadings)
  expect_equal(minres$h2, uls$h2)
  expect_equal(minres$fit_indices, uls$fit_indices)
  # the alias resolves to the canonical method name
  expect_identical(minres$settings$estimator, "ULS")
})

test_that("SL() and EFA_AVERAGE() accept method = 'MINRES'", {
  EFA_mod <- suppressWarnings(EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                                  method = "PAF", rotation = "promax"))
  expect_no_error(suppressWarnings(SL(EFA_mod, method = "MINRES")))
  expect_no_error(suppressWarnings(
    EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                method = "MINRES", show_progress = FALSE)))
})

rm(ULS_test, ULS_test_1)

# DWLS estimator: diagonally weighted least squares fit to a polychoric matrix. The
# matrix and its asymptotic covariance are covered in test-polychoric.R, the routing and
# bootstrap in test-cor-method-poly.R, and the lavaan parity in test-regression-estimators.R.
# These tests cover the fitter's output contract, the analytic gradient, and the
# fit-index handling.

x_dwls <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), ]
pc_dwls <- suppressWarnings(
  .polychoric(x_dwls, acov = "diag"),
  classes = "efa_cor_sparse_cells"
)
R_dwls <- pc_dwls$R
W_dwls <- .poly_weight_matrix(pc_dwls$acov, ncol(R_dwls))

DWLS_test   <- .estimate_model(R_dwls, method = "DWLS", n_factors = 3,
                               N = nrow(x_dwls), weights = W_dwls)
DWLS_test_1 <- .estimate_model(R_dwls, method = "DWLS", n_factors = 1,
                               N = nrow(x_dwls), weights = W_dwls)

test_that("output class and dimensions are correct", {
  expect_s3_class(DWLS_test$unrot_loadings, "LOADINGS")
  expect_output(str(DWLS_test), "List of 12")
  expect_s3_class(DWLS_test_1$unrot_loadings, "LOADINGS")
  expect_output(str(DWLS_test_1), "List of 12")
})

test_that("outputs are correct", {
  expect_equal(DWLS_test$orig_R, R_dwls)
  expect_equal(sum(DWLS_test$orig_eigen), ncol(R_dwls))
  expect_lt(sum(DWLS_test$final_eigen), ncol(R_dwls))
  expect_equal(DWLS_test$convergence, 0)

  expect_equal(DWLS_test_1$orig_R, R_dwls)
  expect_equal(sum(DWLS_test_1$orig_eigen), ncol(R_dwls))
  expect_lt(sum(DWLS_test_1$final_eigen), ncol(R_dwls))
  expect_equal(DWLS_test_1$convergence, 0)
})

test_that("the chi-square block is NA but the descriptive indices are reported", {
  # DWLS does not report the ML-discrepancy chi-square or the indices derived from it;
  # the weight-free residual indices are still computed.
  expect_true(is.na(DWLS_test$fit_indices$chi))
  expect_true(is.na(DWLS_test$fit_indices$p_chi))
  expect_true(is.na(DWLS_test$fit_indices$CFI))
  expect_true(is.na(DWLS_test$fit_indices$TLI))
  expect_true(is.na(DWLS_test$fit_indices$RMSEA))
  expect_true(is.na(DWLS_test$fit_indices$AIC))
  expect_true(is.na(DWLS_test$fit_indices$BIC))

  expect_false(is.na(DWLS_test$fit_indices$RMSR))
  expect_false(is.na(DWLS_test$fit_indices$SRMR))
  expect_false(is.na(DWLS_test$fit_indices$CAF))
  expect_type(DWLS_test$fit_indices$Fm, "double")
})

test_that("the reduced Model Fit block says why the chi-square block is missing", {
  # Without se = "sandwich" a DWLS report ends at CAF, SRMR and df, with nothing saying that
  # the statistic exists in scaled form and how to get it. Six items and one factor keep the
  # polychoric matrix small; df = 9 keeps the model overidentified, which is the case the note
  # is held to.
  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), 1:6][1:500, ]
  fit_dwls <- function(...) {
    suppressWarnings(suppressMessages(
      efa_fit(x, n_factors = 1, cor_method = "poly", estimator = "dwls",
              rotation = "none", ...)))
  }
  # Collapsed, because the note wraps to the console width and the fragment spans the break.
  shown <- function(obj, fun = print) {
    paste(cli::ansi_strip(utils::capture.output(fun(obj))), collapse = " ")
  }
  note <- "refit with `se = \"sandwich\"` for the scaled chi-square"

  plain <- fit_dwls()
  expect_true(is.na(plain$fit_indices$chi))
  expect_true(grepl(note, shown(plain), fixed = TRUE))
  # summary() renders the same section, so the note reaches both reports.
  expect_true(grepl(note, shown(summary(plain)), fixed = TRUE))

  # With the robust covariance the statistic is there, so the note must not appear.
  robust <- fit_dwls(se = "sandwich")
  expect_false(is.na(robust$fit_indices$chi))
  expect_false(grepl(note, shown(robust), fixed = TRUE))

  # PAF reaches the same reduced block, but no setting restores a chi-square for it.
  paf <- suppressWarnings(suppressMessages(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "none")))
  expect_false(grepl(note, shown(paf), fixed = TRUE))
})

test_that("the DWLS gradient matches finite differences", {
  # the analytic gradient -2 (W o (R - LL')) L of the weighted off-diagonal objective,
  # checked against central differences on a random fixture
  set.seed(1)
  p <- 6L; m <- 2L
  A <- matrix(stats::rnorm(p * p), p, p)
  Rg <- stats::cov2cor(A %*% t(A))
  Wg <- matrix(stats::runif(p * p, 0.5, 5), p, p)
  Wg <- (Wg + t(Wg)) / 2
  diag(Wg) <- 0
  par <- stats::rnorm(p * m, 0, 0.3)

  ga <- as.vector(.grad_dwls(par, Rg, m, Wg))
  h <- 1e-6
  gn <- vapply(seq_along(par), function(k) {
    e <- numeric(length(par)); e[k] <- h
    (.dwls_residuals(par + e, Rg, m, Wg) - .dwls_residuals(par - e, Rg, m, Wg)) / (2 * h)
  }, numeric(1))

  expect_equal(ga, gn, tolerance = 1e-5)

  expect_error(.grad_dwls(par[-1], Rg, m, Wg), "must have length")
  expect_error(.dwls_residuals(c(par, 0), Rg, m, Wg), "must have length")

  W_asym <- Wg
  W_asym[1, 2] <- W_asym[1, 2] + 1
  expect_error(.grad_dwls(par, Rg, m, W_asym), "must be symmetric")
})

test_that("the DWLS objective equals the weighted off-diagonal residual sum", {
  set.seed(2)
  p <- 5L; m <- 2L
  A <- matrix(stats::rnorm(p * p), p, p)
  Rg <- stats::cov2cor(A %*% t(A))
  Wg <- matrix(stats::runif(p * p, 0.5, 5), p, p)
  Wg <- (Wg + t(Wg)) / 2
  diag(Wg) <- 0
  par <- stats::rnorm(p * m, 0, 0.3)

  L <- matrix(par, p, m)
  E <- Rg - tcrossprod(L)
  expect_equal(.dwls_residuals(par, Rg, m, Wg),
               sum(Wg[upper.tri(Wg)] * E[upper.tri(E)]^2))
})

test_that(".DWLS requires a weight matrix", {
  expect_error(.DWLS(R_dwls, n_factors = 2, weights = NULL),
               class = "efa_dwls_no_weights")
})

test_that(".poly_weight_matrix rejects a non-positive asymptotic variance", {
  expect_error(.poly_weight_matrix(c(0.1, 0, 0.2), 3),
               class = "efa_dwls_degenerate_weight")
  expect_error(.poly_weight_matrix(c(0.1, -1, 0.2), 3),
               class = "efa_dwls_degenerate_weight")
})

test_that("DWLS leaves psi NULL so a near-1 communality is not a false boundary Heywood case", {
  # .DWLS optimises free loadings with no lower bound on the uniquenesses, so it must not
  # carry psi into the box-boundary Heywood heuristic that only applies to ML/ULS.
  raw <- .DWLS(R_dwls, n_factors = 3, weights = W_dwls)
  expect_null(raw$psi)

  # a proper solution (all communalities < 1) with one communality near 1 (psi ~ 0.004):
  # not a Heywood case for DWLS, but the box-constrained ML/ULS path treats a pinned
  # uniqueness as a boundary Heywood case.
  L <- matrix(c(0.998, 0.6, 0.5, 0.4), 4, 1)
  h2 <- as.vector(diag(tcrossprod(L)))
  orig_R <- matrix(c(1, .7, .6, .5, .7, 1, .5, .4, .6, .5, 1, .3, .5, .4, .3, 1), 4)
  R_final <- orig_R; diag(R_final) <- h2
  base <- list(L = L, h2 = h2, Fm = 0, iter = 1L, convergence = 0L,
               orig_R = orig_R, R_final = R_final)

  dwls_fin <- suppressWarnings(.finalize_fit(c(base, list(psi = NULL)), N = 500, method = "DWLS"))
  expect_length(dwls_fin$heywood, 0)

  uls_fin <- suppressWarnings(.finalize_fit(c(base, list(psi = 1 - h2)), N = 500, method = "ULS"))
  expect_true(1L %in% uls_fin$heywood)
})

test_that(".fit_dwls_cpp tolerates an indefinite (non-positive-definite) correlation matrix", {
  # bootstrap replicate matrices are fed in unsmoothed and can be indefinite; the warm-start
  # inverse must not throw (as inv_sympd would), so DWLS tolerates the same resamples ULS does.
  R_npd <- matrix(c(1, 0.6, 0.6, 0.6, 1, -0.6, 0.6, -0.6, 1), 3)
  expect_lt(min(eigen(R_npd, symmetric = TRUE, only.values = TRUE)$values), 0)
  W <- matrix(c(0, 2, 3, 2, 0, 1.5, 3, 1.5, 0), 3)
  fit <- .fit_dwls_cpp(R_npd, 1L, W)
  expect_true(all(is.finite(fit$loadings)))
})

test_that("DWLS ignores the weight matrix diagonal", {
  # objective and gradient weight only off-diagonal residuals, regardless of W's diagonal
  set.seed(3)
  p <- 5L; m <- 2L
  A <- matrix(stats::rnorm(p * p), p, p)
  Rg <- stats::cov2cor(A %*% t(A))
  W0 <- matrix(stats::runif(p * p, 0.5, 5), p, p)
  W0 <- (W0 + t(W0)) / 2
  diag(W0) <- 0
  W1 <- W0; diag(W1) <- c(-1, Inf, NA, NaN, -Inf)
  par <- stats::rnorm(p * m, 0, 0.3)

  expect_equal(.dwls_residuals(par, Rg, m, W1), .dwls_residuals(par, Rg, m, W0))
  expect_equal(as.vector(.grad_dwls(par, Rg, m, W1)), as.vector(.grad_dwls(par, Rg, m, W0)))
})

test_that("DWLS canonicalizes accepted round-off asymmetry", {
  R <- diag(3)
  W <- matrix(0, 3, 3)
  W[1, 2] <- 1e9
  W[2, 1] <- 0
  W[1, 3] <- W[3, 1] <- 1e20
  W_sym <- W / 2 + t(W) / 2
  par <- c(.2, .3, 0)

  expect_equal(.dwls_residuals(par, R, 1L, W),
               .dwls_residuals(par, R, 1L, W_sym))
  expect_equal(.grad_dwls(par, R, 1L, W),
               .grad_dwls(par, R, 1L, W_sym))
})

rm(x_dwls, pc_dwls, R_dwls, W_dwls, DWLS_test, DWLS_test_1)

# Tests for the non-parametric bootstrap standard error path of EFA()
# (se = "np-boot"): end-to-end coverage, output structure and validity,
# reproducibility, and graceful handling of degenerate bootstrap replicates.

# The reproducibility assertions below compare two separate invocations, so they use
# expect_equal() with an explicit tolerance rather than expect_identical(): a threaded BLAS
# (Apple's Accelerate, for one) is free to vary its GEMM reduction order between calls, so
# one function on one input can differ in the last ulp, and an iterative rotation carries
# that into the reported loadings. waldo still compares S3 classes, names, attributes and
# structure exactly, so only numeric values are given slack -- a seed that was ignored moves
# the compared values by orders of magnitude more than this and still fails loudly. (A set
# tolerance does relax integer against double, so where the storage mode is itself part of
# the contract it is asserted separately.)
fp_tol <- 1e-8

# A clean oblique (promax) bootstrap fit reused by several structure/validity
# tests. GRiPS_raw is well conditioned, so all replicates should succeed.
set.seed(42)
boot_promax <- suppressWarnings(suppressMessages(
  EFA(GRiPS_raw, n_factors = 2, method = "PAF", rotation = "promax",
      se = "np-boot", b_boot = 12)
))

test_that("np-boot runs end to end for all methods and rotation families", {
  skip_on_cran()

  # Three (method, rotation) representatives -- one per method and one per rotation
  # family (none / orthogonal / oblique) -- guard the method x np-boot dispatch
  # without redoing the full method x rotation cross product. The rotation-family
  # slot schema is independently pinned across se = information / sandwich / np-boot
  # by test-EFA-fields.R.
  combos <- data.frame(
    method   = c("PAF", "ML", "ULS"),
    rotation = c("promax", "varimax", "none"),
    stringsAsFactors = FALSE
  )

  for (i in seq_len(nrow(combos))) {
    method <- combos$method[i]
    rotation <- combos$rotation[i]
    label <- paste(method, rotation)

    set.seed(100 + i)
    res <- suppressWarnings(suppressMessages(
      EFA(GRiPS_raw, n_factors = 2, method = method, rotation = rotation,
          se = "np-boot", b_boot = 8)
    ))

    expect_s3_class(res, "EFA")
    expect_false(is.null(res$SE), info = label)
    expect_false(is.null(res$CI), info = label)
    expect_false(is.null(res$replicates), info = label)
    # the third array dimension is the number of bootstrap replicates
    expect_identical(dim(res$replicates$unrot_loadings)[3], 8L, info = label)
  }
})

test_that("np-boot resamples only the complete cases under listwise deletion", {
  skip_on_cran()

  # Build raw data whose complete cases sit at the tail: the first 35 rows each
  # carry a missing value and the 25 complete cases are rows 36-60. Under
  # use = "complete.obs" the correlation matrix - and hence N - rests on those
  # 25 complete cases, so the bootstrap must resample them. Resampling row
  # positions 1:N (= 1:25) would instead draw only all-missing rows, and
  # stats::cor(use = "complete.obs") then errors with "no complete element
  # pairs"; resampling the complete cases yields NA-free bootstrap correlations
  # and finite standard errors.
  set.seed(123)
  f <- rnorm(60)
  dat <- vapply(1:6, function(j) 0.6 * f + 0.8 * rnorm(60), numeric(60))
  colnames(dat) <- paste0("V", 1:6)
  dat[1:35, 1] <- NA

  expect_equal(sum(stats::complete.cases(dat)), 25)        # complete cases are rows 36-60
  expect_true(all(!stats::complete.cases(dat[1:25, ])))    # the old 1:N pool was all-missing

  set.seed(123)
  res <- suppressWarnings(suppressMessages(
    EFA(dat, n_factors = 1, method = "PAF", se = "np-boot", b_boot = 30,
        use = "complete.obs")
  ))

  expect_s3_class(res, "EFA")
  expect_true(all(is.finite(res$SE$unrot_loadings)))
})

test_that("oblique np-boot output has the expected structure", {
  se <- boot_promax$SE
  ci <- boot_promax$CI
  arr <- boot_promax$replicates

  expect_named(se, c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                     "fit_indices", "residuals", "valid_replicates",
                     "valid_target_rotations"))
  expect_named(ci, c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                     "fit_indices", "residuals"))
  expect_named(arr, c("unrot_loadings", "rot_loadings", "Phi", "Structure",
                      "fit_indices", "residuals"))

  L <- boot_promax$rot_loadings
  expect_equal(dim(se$rot_loadings), dim(L))
  expect_equal(dim(se$unrot_loadings), dim(L))
  expect_equal(dim(se$Phi), c(ncol(L), ncol(L)))
  expect_identical(dim(arr$rot_loadings)[3], 12L)

  # the effective B, and the number of usable target rotations, are reported and in bounds
  expect_true(se$valid_replicates >= 1 && se$valid_replicates <= 12)
  expect_true(se$valid_target_rotations >= 1 &&
                se$valid_target_rotations <= se$valid_replicates)
})

test_that("the printed bootstrap sample count reports the usable replicates when they differ", {
  # The effective B is what the unrotated-loading, residual and fit-index intervals rest on. Without
  # it the output asserts the requested b_boot whatever the survival rate, so an interval built from
  # 4 order statistics reads as one built from 20 -- and states a number that is wrong.
  degraded <- boot_promax
  degraded$SE$valid_replicates <- 4L

  body <- cli::ansi_strip(format(summary(degraded)))
  expect_true(any(grepl("Bootstrap samples: 12 (4 usable)", body, fixed = TRUE)))
  expect_true(any(grepl("12 bootstrap samples (4 usable)", body, fixed = TRUE)))

  # A run in which every replicate survived says nothing extra.
  clean <- cli::ansi_strip(format(summary(boot_promax)))
  expect_false(any(grepl("usable", clean, fixed = TRUE)))
})

test_that("np-boot standard errors and confidence intervals are valid", {
  se <- boot_promax$SE
  ci <- boot_promax$CI

  for (nm in c("unrot_loadings", "rot_loadings", "Phi", "Structure")) {
    expect_true(all(se[[nm]] >= 0), info = nm)            # SEs are non-negative
    expect_true(all(is.finite(se[[nm]])), info = nm)
    expect_true(all(ci[[nm]]$lower <= ci[[nm]]$upper), info = nm)  # ordered CIs
  }

  # standardized residuals are added from the bootstrap residual SEs; the
  # off-diagonal entries (the ones of interest) are finite
  sr <- boot_promax$standardized_residuals
  expect_equal(dim(sr), dim(boot_promax$residuals))
  expect_true(all(is.finite(sr[upper.tri(sr)])))
})

test_that("np-boot is reproducible with a fixed seed", {
  skip_on_cran()

  run <- function() suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", rotation = "promax",
        se = "np-boot", b_boot = 8)
  ))

  set.seed(7); a <- run()
  set.seed(7); b <- run()

  expect_equal(a$SE$unrot_loadings, b$SE$unrot_loadings)
  expect_equal(a$SE$rot_loadings, b$SE$rot_loadings)
  expect_equal(a$SE$Phi, b$SE$Phi)
})

test_that("np-boot with a fixed seed is reproducible at 1 vs 2 workers", {
  skip_on_cran()
  skip_if_not_slow()
  # The replicate fits are parallelised across workers with future.apply, and
  # future.seed = TRUE binds each replicate's RNG stream to its index. With a fixed
  # `seed` the bootstrap must therefore return the same result regardless of the
  # number of workers. The comparison uses a small tolerance rather than bit-for-bit
  # equality: the worker fits run in separate processes whose BLAS/LAPACK may sum in a
  # different order, so results can differ in the last bit or two while remaining
  # numerically equivalent. The multisession workers are fresh R processes that load
  # the installed package, so run this under devtools::check() / after
  # devtools::install() for the worker code to match the main process. (multicore is
  # unavailable on Windows.)
  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  run <- function() suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", rotation = "promax",
        se = "np-boot", b_boot = 12, seed = 2024)
  ))

  future::plan(future::sequential)
  one <- run()

  future::plan(future::multisession, workers = 2)
  two <- run()

  expect_equal(one$replicates, two$replicates, tolerance = 1e-10)
  expect_equal(one$SE, two$SE, tolerance = 1e-10)
  expect_equal(one$CI, two$CI, tolerance = 1e-10)
})

test_that("a supplied seed leaves the caller's RNG stream unchanged", {
  # Passing `seed` must not have a lasting side effect on the global RNG: the stream
  # is restored on exit, so a draw taken after a seeded bootstrap is identical to one
  # taken without the call having happened.
  set.seed(1)
  state_before <- get(".Random.seed", envir = globalenv(), inherits = FALSE)

  # An oblique rotation exercises every RNG consumer the restore must neutralize:
  # the case resampling, the point-estimate rotation random starts, and the oblique
  # Procrustes random starts that run after the replicate fits.
  invisible(suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", rotation = "promax",
        se = "np-boot", b_boot = 6, seed = 99)
  )))

  expect_identical(get(".Random.seed", envir = globalenv(), inherits = FALSE),
                   state_before)
})

test_that("seed makes a criterion rotation reproducible without a bootstrap", {
  # `seed` governs the whole fit, not just the bootstrap: the criterion-based rotations
  # draw random starts from the RNG, so two seeded calls at the default se = "none" must
  # agree. simplimax is used because its criterion is only piecewise smooth, making it
  # the most prone to landing in different optima from different starts -- unseeded, two
  # calls disagree by a wide margin, so this would fail if `seed` were ignored.
  fit <- function(s) suppressWarnings(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "simplimax", seed = s))

  a <- fit(42)
  b <- fit(42)
  expect_equal(unclass(a$rot_loadings), unclass(b$rot_loadings), tolerance = fp_tol)
  expect_equal(a$Phi, b$Phi, tolerance = fp_tol)

  # A different seed explores different starts, so the argument is doing something
  # rather than the fit being deterministic to begin with.
  set.seed(1)
  unseeded_1 <- suppressWarnings(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "simplimax"))
  set.seed(2)
  unseeded_2 <- suppressWarnings(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "simplimax"))
  expect_false(identical(unclass(unseeded_1$rot_loadings),
                         unclass(unseeded_2$rot_loadings)))
})

test_that("seed leaves the caller's RNG stream unchanged without a bootstrap", {
  # The restore-on-exit contract holds on the non-bootstrap path too.
  set.seed(1)
  state_before <- get(".Random.seed", envir = globalenv(), inherits = FALSE)

  invisible(suppressWarnings(
    efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500, estimator = "PAF",
            rotation = "simplimax", seed = 99)))

  expect_identical(get(".Random.seed", envir = globalenv(), inherits = FALSE),
                   state_before)
})

test_that("np-boot on a correlation matrix warns and disables the bootstrap", {
  expect_warning(
    res <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
               method = "PAF", rotation = "promax", se = "np-boot"),
    class = "efa_boot_cormat"
  )
  expect_s3_class(res, "EFA")
  expect_identical(res$settings$se, "none")
  expect_null(res$SE)
})

# A healthy PAF point-estimate fit on GRiPS_raw together with the matching list of `b`
# replicate fits. The failure-mode blocks below all start from this pair and damage it in
# different ways, so it is built by one helper rather than retyped. The seed lives inside the
# helper so every call draws the same resamples from the same stream whatever ran before it.
make_boot_pair <- function(b, n_factors) {
  set.seed(11)
  x <- GRiPS_raw
  R <- stats::cor(x)
  N <- nrow(x)
  m <- ncol(R)

  R_boot <- array(NA_real_, c(m, m, b))
  for (i in seq_len(b)) {
    ind <- sample(N, size = N, replace = TRUE)
    R_boot[, , i] <- stats::cor(x[ind, ])
  }

  list(
    fit_target = suppressWarnings(
      .estimate_model(R, method = "PAF", n_factors = n_factors, N = N, type = "EFAtools")),
    boot_fit = suppressWarnings(
      .boot_fun(R_boot, b, .estimate_model, method = "PAF", n_factors = n_factors,
                N = N, type = "EFAtools"))
  )
}

test_that("a failed bootstrap replicate is skipped with a warning", {
  b <- 6
  pair <- make_boot_pair(b, n_factors = 2)
  boot_fit <- pair$boot_fit

  # inject a failed replicate without dropping the list element
  boot_fit[2] <- list(NULL)

  expect_warning(
    res <- .boot_se_ci(pair$fit_target, L_rot = NULL, boot_fit,
                       boot_rot = "none", ci = 0.95, b = b),
    class = "efa_boot_replicate_failed"
  )

  # the failed replicate's slice stays NA; SEs from the rest are finite
  expect_true(all(is.na(res$replicates$unrot_loadings[, , 2])))
  expect_true(all(is.finite(res$SE$unrot_loadings)))
  # the effective B is recorded on the object, not only in the warning
  expect_equal(res$SE$valid_replicates, b - 1)
})

test_that(".boot_fun records a real failed replicate as NULL without dropping it", {
  # drive an actual fit failure (degenerate, non-finite correlation matrix)
  # through .boot_fun: a failed replicate must be recorded as a length-preserving
  # NULL, including when it is the LAST replicate, so .boot_se_ci can skip it.
  set.seed(13)
  x <- GRiPS_raw[1:200, ]
  R <- stats::cor(x)
  N <- nrow(x)
  m <- ncol(R)
  b <- 5

  bad <- R
  bad[1, ] <- NaN
  bad[, 1] <- NaN
  diag(bad) <- 1

  for (fail_at in c(2L, b)) {            # mid replicate and the last replicate
    R_boot <- array(NA_real_, c(m, m, b))
    for (i in seq_len(b)) {
      ind <- sample(N, size = N, replace = TRUE)
      R_boot[, , i] <- stats::cor(x[ind, ])
    }
    R_boot[, , fail_at] <- bad

    boot_fit <- suppressWarnings(
      .boot_fun(R_boot, b, .estimate_model, method = "PAF", n_factors = 2,
                N = N, type = "EFAtools"))

    # length preserved and exactly the degenerate replicate is NULL
    expect_length(boot_fit, b)
    expect_true(is.null(boot_fit[[fail_at]]))
    expect_equal(sum(vapply(boot_fit, is.null, logical(1))), 1L)

    # .boot_se_ci skips it gracefully (warns, does not error)
    fit_target <- suppressWarnings(
      .estimate_model(R, method = "PAF", n_factors = 2, N = N, type = "EFAtools"))
    expect_warning(
      res <- .boot_se_ci(fit_target, L_rot = NULL, boot_fit,
                         boot_rot = "none", ci = 0.95, b = b),
      class = "efa_boot_replicate_failed"
    )
    expect_true(all(is.finite(res$SE$unrot_loadings)))
  }
})

test_that("np-boot aborts when every replicate fails", {
  x <- GRiPS_raw[1:100, ]
  R <- stats::cor(x)
  b <- 4

  fit_target <- suppressWarnings(
    .estimate_model(R, method = "PAF", n_factors = 2, N = nrow(x),
                    type = "EFAtools"))
  boot_fit <- rep(list(NULL), b)

  expect_error(
    .boot_se_ci(fit_target, L_rot = NULL, boot_fit, boot_rot = "none",
                ci = 0.95, b = b),
    class = "efa_boot_all_failed"
  )
})

test_that("a replicate that cannot be target-rotated warns with its own class", {
  # The rotated block has a survival count of its own: a replicate can fit and align to the
  # unrotated point estimate and still fail the target rotation. A non-finite target forces that on
  # the orthogonal branch; the oblique branch signals the same class from its batched aligner.
  b <- 4
  pair <- make_boot_pair(b, n_factors = 2)

  bad_target <- unclass(pair$fit_target$unrot_loadings)
  bad_target[1, 1] <- NaN

  expect_warning(
    res <- .boot_se_ci(pair$fit_target, L_rot = bad_target, pair$boot_fit,
                       boot_rot = "orthogonal", ci = 0.95, b = b),
    class = "efa_boot_rotation_failed"
  )
  expect_equal(res$SE$valid_target_rotations, 0)
  expect_true(all(is.na(res$SE$rot_loadings)))
  # the unrotated block is unaffected: only the target rotation failed
  expect_equal(res$SE$valid_replicates, b)
  expect_true(all(is.finite(res$SE$unrot_loadings)))
})

test_that("b_boot below two is rejected", {
  # A bootstrap standard error is the dispersion across replicates: at b_boot = 1 every SE is the
  # sd() of a single value and comes back NA, and the interval collapses onto that replicate. That
  # used to happen silently -- the only route on which an SE returned NA with no condition at all.
  expect_error(
    efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF", rotation = "none",
            se = "np-boot", b_boot = 1),
    class = "efa_b_boot_too_small"
  )
  expect_error(
    efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF", rotation = "none",
            se = "np-boot", b_boot = 0),
    class = "efa_b_boot_too_small"
  )
  expect_error(
    efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF", rotation = "none",
            se = "np-boot", b_boot = -5),
    class = "efa_b_boot_too_small"
  )
})

test_that("a ci at either end of the unit interval is rejected", {
  # The bound is open at both ends, on every SE path. The analytic paths read the level as
  # z = qnorm(1 - (1 - ci) / 2): at ci = 1 that is infinite and each Wald bound came back
  # -Inf/Inf, at ci = 0 it is zero and each interval collapsed onto the point estimate. Both
  # passed silently, because the assertion took the closed interval the documentation described.
  cm <- test_models$baseline$cormat

  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "ML", se = "information", ci = 1),
    class = "efa_ci_out_of_bounds"
  )
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "ML", se = "information", ci = 0),
    class = "efa_ci_out_of_bounds"
  )
  # The guard sits in the argument checks, so it fires whatever the SE path -- including none.
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", se = "np-boot", ci = 1),
    class = "efa_ci_out_of_bounds"
  )
  # It owns the whole range, so a level outside the unit interval reports the same rule.
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "ML", se = "information", ci = 1.5),
    class = "efa_ci_out_of_bounds"
  )
  # A non-number stays with the shared argument assertion.
  expect_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "ML", se = "information", ci = NA),
    class = "efa_invalid_argument"
  )

  # A level inside the interval still passes.
  expect_no_error(
    efa_fit(cm, n_factors = 3, N = 500, estimator = "ML", rotation = "none",
            se = "information", ci = .99)
  )
})

test_that("fewer than two surviving replicates is flagged as unreliable", {
  # The input bound cannot see this one: b_boot is legal, but the replicates fail at run time and
  # leave a single usable draw behind. Without a condition the object returns all-NA SEs and a
  # collapsed interval that print as though the requested b_boot had stood.
  b <- 6
  pair <- make_boot_pair(b, n_factors = 1)
  fit_target <- pair$fit_target
  boot_fit <- pair$boot_fit

  one_left <- boot_fit
  one_left[2:b] <- rep(list(NULL), b - 1L)     # a single survivor

  expect_warning(
    expect_warning(
      res <- .boot_se_ci(fit_target, L_rot = NULL, one_left, boot_rot = "none",
                         ci = 0.95, b = b),
      class = "efa_boot_replicate_failed"
    ),
    class = "efa_se_unreliable"
  )
  expect_equal(res$SE$valid_replicates, 1L)
  expect_true(all(is.na(res$SE$unrot_loadings)))
  expect_equal(res$CI$unrot_loadings$lower, res$CI$unrot_loadings$upper)

  # Two survivors are enough for a defined (if very wide) standard error: the failed-replicate
  # warning still fires there, but the unreliability warning must not.
  two_left <- boot_fit
  two_left[3:b] <- rep(list(NULL), b - 2L)
  seen <- character(0)
  withCallingHandlers(
    res2 <- .boot_se_ci(fit_target, L_rot = NULL, two_left, boot_rot = "none",
                        ci = 0.95, b = b),
    warning = function(w) {
      seen <<- c(seen, class(w)[1])
      invokeRestart("muffleWarning")
    }
  )
  expect_false("efa_se_unreliable" %in% seen)
  expect_equal(res2$SE$valid_replicates, 2L)
  expect_true(all(is.finite(res2$SE$unrot_loadings)))
})

# Count how many signalled warnings carry a given condition class, muffling all
# warnings so the wrapped expression runs to completion.
count_warning_class <- function(expr, cls) {
  n <- 0L
  withCallingHandlers(
    force(expr),
    warning = function(w) {
      if (inherits(w, cls)) n <<- n + 1L
      invokeRestart("muffleWarning")
    }
  )
  n
}

test_that("a pinned argument re-warns once, not once per bootstrap replicate", {
  skip_on_cran()
  # `efa_type_override` reflects the (type, pinned-argument) combination, which is
  # identical for every replicate and already surfaced once by the point-estimate
  # fit. The bootstrap loop must not repeat it b_boot times.
  set.seed(42)
  n_override <- count_warning_class(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", type = "EFAtools",
        max_iter = 500, se = "np-boot", b_boot = 6),
    "efa_type_override"
  )
  expect_equal(n_override, 1L)
})

test_that("bootstrap non-convergence is summarized in a single classed warning", {
  skip_on_cran()
  # A tiny iteration cap forces every replicate to hit the maximum-iteration limit;
  # the per-replicate fitter warnings must be suppressed and replaced by a single
  # classed summary rather than one warning per replicate.
  set.seed(1)
  n_summary <- count_warning_class(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", type = "none",
        init_comm = "smc", criterion = 1e-3, criterion_type = "sum",
        abs_eigen = TRUE, max_iter = 1, se = "np-boot", b_boot = 5),
    "efa_boot_nonconvergence"
  )
  expect_equal(n_summary, 1L)

  # a cleanly converging bootstrap emits no non-convergence summary
  set.seed(7)
  n_clean <- count_warning_class(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", rotation = "promax",
        se = "np-boot", b_boot = 8),
    "efa_boot_nonconvergence"
  )
  expect_equal(n_clean, 0L)
})

test_that("eigendecomposition guards turn degenerate matrices into errors", {
  # a non-finite (constant-column-style) correlation matrix makes the symmetric
  # eigendecomposition fail; the guarded fitters must error, not crash R
  m <- 6
  R_bad <- diag(m)
  R_bad[1, ] <- NaN
  R_bad[, 1] <- NaN
  diag(R_bad) <- 1
  psi <- rep(0.5, m)

  expect_eigendecomp_error <- function(expr) {
    # Armadillo writes a diagnostic to stderr before returning failure; capture it
    # so the test asserts the classed R error without noisy logs.
    invisible(utils::capture.output(
      expect_error(force(expr), "Eigendecomposition failed"),
      type = "message"
    ))
  }

  expect_eigendecomp_error(.paf_iter(psi, 0.001, R_bad, 2L, TRUE, 2L, 100L))
  expect_eigendecomp_error(.grad_ml(psi, R_bad, 2L))
  expect_eigendecomp_error(.error_ml(psi, R_bad, 2L))
  expect_eigendecomp_error(.grad_uls(psi, R_bad, 2L))
  expect_eigendecomp_error(.uls_residuals(psi, R_bad, 2L))
})

test_that("over-extraction guards turn n_fac >= ncol into errors", {
  # the eigenvalue-based extraction reads the largest n_fac eigenpairs; with
  # n_fac >= ncol(R) it would index past the available eigenvalues (undefined
  # behaviour in an unchecked build). The guarded fitters must error, not crash R
  m <- 6L
  R <- diag(m)
  psi <- rep(0.5, m)

  expect_error(.paf_iter(rep(1, m), 0.001, R, m, TRUE, 2L, 10L),
               "smaller than the number of variables")
  expect_error(.grad_ml(psi, R, m), "smaller than the number of variables")
  expect_error(.error_ml(psi, R, m), "smaller than the number of variables")
  expect_error(.grad_uls(psi, R, m), "smaller than the number of variables")
  expect_error(.uls_residuals(psi, R, m), "smaller than the number of variables")
})

test_that("the PAF kernel rejects a non-positive iteration budget", {
  # With max_iter < 1 the iteration never runs and the kernel would return an empty
  # loading matrix, which .finalize_fit() would read as a zero-factor solution. The
  # R-side control validation is the normal user-facing path and keeps its condition
  # class; this asserts the backstop on the other side of the boundary.
  m <- 6L
  R <- matrix(0.3, m, m)
  diag(R) <- 1
  h2 <- rep(0.5, m)

  expect_error(.paf_iter(h2, 0.001, R, 2L, TRUE, 2L, 0L), "at least 1")
  expect_error(.paf_iter(h2, 0.001, R, 2L, TRUE, 2L, -5L), "at least 1")

  expect_error(estimate_control(max_iter = 0), class = "efa_control_input")
  expect_error(estimate_control(max_iter = -5), class = "efa_control_input")
})

test_that(".array_se_ci matches a per-probability sweep at every shape it is called with", {
  # The helper takes both percentile bounds from one pass over the replicate array. The bounds
  # must equal what a separate sweep per probability produced -- same values, same shapes, same
  # dimnames -- at every rank its call sites use: a replicate matrix (the fit indices, M = 2),
  # a 3-D cube (loadings, Phi, residuals, M = c(1, 2)), and the group bootstrap's 4-D cube.
  ref <- function(x, probs, M) {
    ci <- lapply(probs, function(p) apply(x, M, stats::quantile, probs = p, na.rm = TRUE))
    stats::setNames(ci, c("lower", "upper"))
  }
  probs <- c(0.025, 0.975)

  set.seed(20)
  cases <- list(
    matrix_M2 = list(
      x = matrix(stats::rnorm(40 * 5), nrow = 40,
                 dimnames = list(NULL, paste0("idx", 1:5))),
      M = 2),
    cube_M12 = list(
      x = array(stats::rnorm(6 * 3 * 40), c(6, 3, 40),
                dimnames = list(paste0("V", 1:6), paste0("F", 1:3), NULL)),
      M = c(1, 2)),
    cube4d_M123 = list(
      x = array(stats::rnorm(2 * 2 * 4 * 30), c(2, 2, 4, 30),
                dimnames = list(c("g1", "g2"), c("g1", "g2"), paste0("V", 1:4), NULL)),
      M = c(1L, 2L, 3L))
  )

  for (nm in names(cases)) {
    x <- cases[[nm]]$x
    M <- cases[[nm]]$M
    got <- .array_se_ci(x, probs, M = M)
    want <- ref(x, probs, M)
    expect_equal(got$ci$lower, want$lower, info = nm)
    expect_equal(got$ci$upper, want$upper, info = nm)
    expect_identical(dim(got$ci$lower), dim(got$se), info = nm)
    expect_identical(dimnames(got$ci$lower), dimnames(got$se), info = nm)
  }

  # NA handling is unchanged too: a cell whose replicates are all missing must come back NA
  # rather than error, and a partially missing cell must use its observed replicates only.
  y <- array(stats::rnorm(3 * 2 * 20), c(3, 2, 20))
  y[1, 1, ] <- NA_real_
  y[2, 1, 1:5] <- NA_real_
  got <- .array_se_ci(y, probs)
  want <- ref(y, probs, c(1, 2))
  expect_equal(got$ci$lower, want$lower)
  expect_equal(got$ci$upper, want$upper)
  expect_true(is.na(got$ci$lower[1, 1]))
  expect_false(is.na(got$ci$lower[2, 1]))
})

test_that(".estimate_model(lean = TRUE) returns only the bootstrap-aggregated quantities", {
  # The bootstrap replicate fitter computes just the loadings, fit indices, and
  # residuals that .boot_se_ci() aggregates. Those must match the full fit
  # exactly, except for the analytic RMSEA bounds, which the lean fit does not
  # solve (they are not meaningfully bootstrapped).
  R <- test_models$baseline$cormat
  N <- 500
  bounds <- c("RMSEA_LB", "RMSEA_UB")

  method_args <- list(
    PAF = list(type = "EFAtools"),
    ML  = list(start_method = "psych"),
    ULS = list()
  )

  for (method in names(method_args)) {
    common <- c(list(R, method = method, n_factors = 3, N = N),
                method_args[[method]])
    full <- suppressWarnings(do.call(.estimate_model, common))
    lean <- suppressWarnings(do.call(.estimate_model, c(common, list(lean = TRUE))))

    expect_named(lean, c("unrot_loadings", "fit_indices", "residuals", "convergence"))

    expect_equal(as.vector(lean$unrot_loadings), as.vector(full$unrot_loadings),
                 info = method)
    expect_equal(as.vector(lean$residuals), as.vector(full$residuals), info = method)
    expect_identical(lean$convergence, full$convergence, info = method)

    keep <- setdiff(names(full$fit_indices), bounds)
    expect_equal(lean$fit_indices[keep], full$fit_indices[keep], info = method)
    expect_true(all(is.na(unlist(lean$fit_indices[bounds]))), info = method)
  }
})

test_that("ML np-boot drops only the analytic RMSEA bounds from the fit-index SEs", {
  skip_on_cran()
  set.seed(202)
  res <- suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 2, method = "ML", rotation = "none",
        se = "np-boot", b_boot = 10)
  ))

  se_fit <- res$SE$fit_indices
  ci_fit <- res$CI$fit_indices
  # The replicates are fitted with ci = FALSE, so the analytic RMSEA bounds carry no
  # per-replicate value: they are dropped rather than shipped as guaranteed-NA entries.
  expect_false(any(c("RMSEA_LB", "RMSEA_UB") %in% names(se_fit)))
  expect_false(any(c("RMSEA_LB", "RMSEA_UB") %in% names(ci_fit$lower)))
  expect_false(anyNA(se_fit))
  # the replicate matrix labels its columns, and the replicate is its FIRST dimension
  expect_identical(colnames(res$replicates$fit_indices), names(se_fit))
  expect_identical(nrow(res$replicates$fit_indices), 10L)
  # the bootstrapped fit indices that are aggregated stay finite
  expect_true(all(is.finite(se_fit[c("CAF", "RMSR", "SRMR", "TLI", "ECVI")])))
  expect_true(all(is.finite(ci_fit$lower[c("SRMR", "TLI", "ECVI")])))
  expect_true(all(is.finite(ci_fit$upper[c("SRMR", "TLI", "ECVI")])))
  # the point estimate keeps its full analytic RMSEA confidence interval
  expect_true(is.finite(res$fit_indices$RMSEA_LB))
  expect_true(is.finite(res$fit_indices$RMSEA_UB))

  out <- cli::ansi_strip(format(res))
  expect_true("RMSR" %in% names(res$fit_indices))
  expect_false(any(grepl("^RMSR\\b", out)))
  expect_true(any(grepl("^SRMR \\[95% bootstrap-CI\\]:", out)))
  expect_true(any(grepl("^TLI \\[95% bootstrap-CI\\]:", out)))
  expect_true(any(grepl("^ECVI \\[95% bootstrap-CI\\]:", out)))
})

test_that("PAF np-boot fit output prints SRMR CIs but not RMSR", {
  out <- cli::ansi_strip(format(boot_promax))

  expect_true("RMSR" %in% names(boot_promax$fit_indices))
  expect_true(all(is.finite(boot_promax$CI$fit_indices$lower[c("CAF", "RMSR", "SRMR")])))
  expect_false(any(grepl("^RMSR\\b", out)))
  expect_true(any(grepl("^SRMR \\[95% bootstrap-CI\\]:", out)))
})

test_that(".oblique_procrustes_batch isolates an unalignable replicate", {
  # A single non-finite slice must not abort the whole batch: it is reported
  # valid = FALSE with NA loadings/Phi/diagnostics, while the OTHER (distinct)
  # slices align independently and correctly. This mirrors the per-replicate
  # failure isolation of the per-replicate efa_procrustes() loop the batch replaces.
  set.seed(311)
  efa <- suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 2, method = "PAF", rotation = "promax")))
  L_rot <- efa$rot_loadings
  p <- nrow(L_rot); m <- ncol(L_rot); N <- nrow(GRiPS_raw)

  # three DISTINCT replicate loadings, so a cross-slice contamination bug cannot
  # hide behind identical input slices
  slice_fit <- function(seed) {
    set.seed(seed)
    ind <- sample(N, N, replace = TRUE)
    suppressWarnings(.estimate_model(stats::cor(GRiPS_raw[ind, ]), method = "PAF",
        n_factors = m, N = N, type = "EFAtools", lean = TRUE))$unrot_loadings
  }
  cube <- array(NA_real_, c(p, m, 3))
  cube[, , 1] <- slice_fit(11)
  cube[, , 2] <- slice_fit(12)
  cube[, , 3] <- slice_fit(13)
  cube[1, 1, 2] <- Inf                       # make the middle replicate unalignable

  set.seed(1)
  res <- .oblique_procrustes_batch(cube, L_rot, random_starts = 5)

  expect_identical(as.logical(res$valid), c(TRUE, FALSE, TRUE))
  expect_true(all(is.na(res$loadings[, , 2])))           # invalid slice -> NA, not garbage
  expect_true(all(is.na(res$Phi[, , 2])))
  # invalid slice: every diagnostic is NA, only `valid` is a definite FALSE
  expect_true(all(is.na(c(res$value[2], res$iterations[2],
                          res$convergence[2], res$line_search_failed[2]))))
  expect_true(all(is.finite(res$loadings[, , c(1, 3)])))
  expect_true(all(is.finite(res$Phi[, , c(1, 3)])))

  # the surviving slices are aligned independently: distinct inputs give distinct
  # outputs (a contamination bug would tie them together)...
  expect_gt(max(abs(res$loadings[, , 1] - res$loadings[, , 3])), 1e-3)
  # ...and the first surviving slice reproduces a stand-alone single-slice
  # alignment of the same input under the same RNG offset (uncontaminated result)
  set.seed(1)
  solo <- .oblique_procrustes_batch(cube[, , 1, drop = FALSE], L_rot, random_starts = 5)
  expect_equal(res$loadings[, , 1], solo$loadings[, , 1], tolerance = 1e-10)
})

test_that(".oblique_procrustes_batch matches the per-replicate efa_procrustes alignment", {
  skip_on_cran()
  # The batched alignment must reproduce the single-matrix oblique efa_procrustes path
  # it replaces (same warm start, same R::rnorm random-start stream) to tolerance.
  set.seed(321)
  efa <- suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 3, method = "PAF", rotation = "promax")))
  L_rot <- efa$rot_loadings
  p <- nrow(L_rot); m <- ncol(L_rot)
  N <- nrow(GRiPS_raw)

  b <- 10L
  cube <- array(NA_real_, c(p, m, b))
  for (i in seq_len(b)) {
    ind <- sample(N, N, replace = TRUE)
    fit <- suppressWarnings(.estimate_model(stats::cor(GRiPS_raw[ind, ]),
              method = "PAF", n_factors = m, N = N, type = "EFAtools", lean = TRUE))
    cube[, , i] <- fit$unrot_loadings
  }

  set.seed(99)
  loop_L <- array(NA_real_, c(p, m, b))
  for (j in seq_len(b)) {
    loop_L[, , j] <- efa_procrustes(cube[, , j], Target = L_rot, rotation = "oblique",
                                    oblique_random_starts = 5)$loadings
  }
  set.seed(99)
  batch <- .oblique_procrustes_batch(cube, L_rot, random_starts = 5)

  expect_equal(batch$loadings, loop_L, tolerance = 1e-8)
})

test_that("oblique np-boot runs end to end for a single-factor model", {
  skip_on_cran()
  # A one-factor model with an oblique-family rotation reaches the batch's m == 1
  # closed-form path; the bootstrap must run end to end and return finite SEs.
  set.seed(404)
  res <- suppressWarnings(suppressMessages(
    EFA(GRiPS_raw, n_factors = 1, method = "PAF", rotation = "promax",
        se = "np-boot", b_boot = 8)))

  expect_s3_class(res, "EFA")
  expect_false(is.null(res$SE$rot_loadings))
  expect_true(all(is.finite(res$SE$rot_loadings)))
  expect_identical(dim(res$replicates$rot_loadings)[3], 8L)
})

# NEST simulates reference eigenvalues at every candidate factor count, which makes
# the file-top fixtures ~10s. Skipped by default; opt in with
# `Sys.setenv(EFATOOLS_TEST_SLOW = "true")`. See helper-slow.R.
if (is_slow_test()) {
# seed the reference-data simulation so the retained-factor counts are reproducible
set.seed(42)
nest_cor <- efa_nest(test_models$baseline$cormat, N = 500)
nest_raw <- efa_nest(GRiPS_raw)
}  # is_slow_test()

test_that("output class and dimensions are correct", {
  skip_if_not_slow()
  expect_s3_class(nest_cor, "efa_retention")
  expect_length(nest_cor, 6)
  expect_s3_class(nest_raw, "efa_retention")
  expect_length(nest_raw, 6)
  expect_equal(.retention_record(nest_cor, "NEST")$plot_type, "eigen")
})


test_that("found eigenvalues are correct", {
  skip_if_not_slow()
  expect_equal(sum(.retention_record(nest_cor, "NEST")$y),
               ncol(test_models$baseline$cormat))
  expect_equal(sum(.retention_record(nest_raw, "NEST")$y), ncol(GRiPS_raw))
  expect_length(.retention_record(nest_cor, "NEST")$y,
                ncol(test_models$baseline$cormat))
  expect_length(.retention_record(nest_raw, "NEST")$y, ncol(GRiPS_raw))
})

test_that("reference eigenvalues are correct", {
  skip_if_not_slow()
  nf_cor <- nest_cor$n_factors[["NEST"]]
  eig_cor <- .retention_record(nest_cor, "NEST")$y
  ref_cor <- .retention_record(nest_cor, "NEST")$reference
  nf_raw <- nest_raw$n_factors[["NEST"]]
  eig_raw <- .retention_record(nest_raw, "NEST")$y
  ref_raw <- .retention_record(nest_raw, "NEST")$reference

  # the reference series is padded with NA to the length of the empirical
  # eigenvalues; only the tested positions carry a value
  expect_length(ref_cor, length(eig_cor))
  expect_length(ref_raw, length(eig_raw))
  expect_equal(sum(!is.na(ref_cor)), nf_cor + 1)
  expect_equal(sum(!is.na(ref_raw)), nf_raw + 1)
  expect_lte(eig_cor[nf_cor + 1], ref_cor[nf_cor + 1])
  expect_true(all(eig_cor[1:nf_cor] > ref_cor[1:nf_cor]))
  expect_lte(eig_raw[nf_raw + 1], ref_raw[nf_raw + 1])
  expect_true(all(eig_raw[1:nf_raw] > ref_raw[1:nf_raw]))
})

test_that("identified number of factors is correct", {
  skip_if_not_slow()
  expect_equal(nest_cor$n_factors[["NEST"]], 3)
  expect_equal(nest_raw$n_factors[["NEST"]], 1)
})

test_that("errors are thrown correctly", {
  skip_if_not_slow()
  expect_error(efa_nest(1:5), class = "efa_input_not_matrix")
  expect_error(efa_nest(test_models$baseline$cormat), class = "efa_n_required")
  expect_message(efa_nest(GRiPS_raw), class = "efa_cor_from_data")
  expect_warning(efa_nest(GRiPS_raw, N = 20), class = "efa_n_from_data")
  expect_error(efa_nest(sing_raw), class = "efa_cor_singular")
  expect_error(efa_nest(sing_cor, N = sing_N), class = "efa_cor_singular")
})

test_that("NEST requires more observations than variables", {
  # the accepted call below simulates reference data, so pin the stream: the RNG
  # state reaching this test otherwise depends on which earlier blocks were skipped
  set.seed(321)
  R <- diag(5)
  expect_error(efa_nest(R, N = 0, n_datasets = 1), "Must be >= 1")
  expect_error(efa_nest(R, N = 1, n_datasets = 1), class = "efa_n_too_small")
  expect_error(efa_nest(R, N = 5, n_datasets = 1), class = "efa_n_too_small")
  expect_s3_class(efa_nest(R, N = 6, n_datasets = 1), "efa_retention")
})

test_that("a Heywood case in the reference model raises a classed error", {
  skip_if_not_slow()
  # a positive-definite matrix whose one-factor model has a communality above 1:
  # the reference data cannot be simulated (negative uniqueness), so NEST must
  # abort with a classed condition instead of an unclassed C++ error
  R_hey <- diag(4)
  R_hey[1, 2:4] <- R_hey[2:4, 1] <- .9
  R_hey[2, 3:4] <- R_hey[3:4, 2] <- .78
  R_hey[3, 4] <- R_hey[4, 3] <- .78
  expect_gt(min(eigen(R_hey, symmetric = TRUE, only.values = TRUE)$values), 0)
  # the reference EFA legitimately warns about the Heywood case; assert the abort
  expect_error(suppressWarnings(efa_nest(R_hey, N = 200)), class = "efa_nest_heywood")
})

test_that("the last accepted model is retained at the no-stop boundary", {
  skip_if_not_slow()
  # Four variables with two clear factors: NEST accepts the 1st empirical
  # eigenvalue (vs the identity reference) and the 2nd (vs the one-factor
  # reference). The search then runs out of testable factors -- floor(.8 * nvar)
  # is 3, but the (nf - 1)-factor reference model must stay over-identified
  # (df > 0), which caps it at 2. With no rejection, the retained count must be
  # the last *accepted* model (2), not the loop's just-past index (1).
  R_bound <- matrix(c(1, .7, .1, .1,
                      .7, 1, .1, .1,
                      .1, .1, 1, .7,
                      .1, .1, .7, 1), 4, 4)

  set.seed(123)
  nest_bound <- efa_nest(R_bound, N = 200, n_datasets = 300)
  rec_bound <- .retention_record(nest_bound, "NEST")

  expect_equal(nest_bound$n_factors[["NEST"]], 2)
  # every tested empirical eigenvalue exceeded its reference (no rejection), so
  # the reference series has one value per retained factor (not n_factors + 1),
  # padded with NA up to the number of empirical eigenvalues
  expect_length(rec_bound$reference, length(rec_bound$y))
  expect_equal(sum(!is.na(rec_bound$reference)),
               nest_bound$n_factors[["NEST"]])
  expect_true(all(rec_bound$y[seq_len(nest_bound$n_factors[["NEST"]])] >
                    rec_bound$reference[seq_len(nest_bound$n_factors[["NEST"]])]))

  # The reference fits stay over-identified: an uncapped floor(.8 * nvar) = 3
  # would fit a 2-factor (df < 0) reference for 4 variables; the df > 0 bound
  # prevents that, so no under-identified reference fit is triggered.
  expect_no_warning(
    {
      set.seed(123)
      efa_nest(R_bound, N = 200, n_datasets = 300)
    },
    class = "efa_underidentified"
  )
})

test_that("settings are returned correctly", {
  skip_if_not_slow()

  expect_equal(nest_cor$settings$N, 500)
  expect_equal(nest_raw$settings$N, 810)

  expect_equal(nest_cor$settings$use, "pairwise.complete.obs")
  expect_equal(nest_raw$settings$use, "pairwise.complete.obs")

  expect_equal(nest_cor$settings$cor_method, "pearson")
  expect_equal(nest_raw$settings$cor_method, "pearson")
})

# The reference-eigenvalue simulation is drawn by .simulate_cfm_eigen through the shared
# MVN kernel. These fast tests run unconditionally: the first two pin the two draw paths
# (the null model and a factor-score model), the third drives efa_nest() itself so the wiring
# from each reference model into the kernel is covered without the slow fixtures.
test_that("the null-model reference draw shares the MVN kernel's stream", {
  # The nf == 1 reference is the null (identity) model. Under a fixed seed the largest
  # reference eigenvalue must equal a direct recompute that draws the same null model
  # through .simulate_cfm_mvn(diag(p)) -- pinning that both kernel entries draw the same
  # stream. Tolerance covers arma::cor vs stats::cor (mirrors test-efa_parallel.R).
  N <- 60; p <- 5; nd <- 4

  set.seed(11)
  got <- .simulate_cfm_eigen(1L, N, matrix(numeric(0), p, 0), rep(1, p), nd)

  set.seed(11)
  want <- vapply(seq_len(nd), function(i) {
    eigen(stats::cor(.simulate_cfm_mvn(diag(p), N)), symmetric = TRUE,
          only.values = TRUE)$values[1]
  }, numeric(1))

  expect_equal(as.vector(got), want, tolerance = 1e-10)
})

test_that("the factor-score reference draw matches its pinned reference eigenvalues", {
  # A one-factor reference model (the nf = 2 draw path). The reference eigenvalues are
  # pinned; the draw must reproduce them under this seed. expect_equal (not identical)
  # because the linear algebra downstream of the platform-independent RNG draw varies by
  # ~1e-13 across BLAS implementations, whereas any change to the draw stream is O(1).
  N <- 50; nd <- 5
  L <- matrix(c(.7, .6, .5, .4), 4, 1)
  Psi <- 1 - L^2

  set.seed(123)
  got <- .simulate_cfm_eigen(2L, N, L, as.vector(Psi), nd)

  expect_equal(
    as.vector(got),
    c(0.851006411446006, 0.841528141373475, 0.876576983719939,
      0.856551711327252, 0.840572878485076)
  )
})

test_that("NEST recovers a clear two-factor structure end to end", {
  # Drives efa_nest() itself so the wiring from each reference model into .simulate_cfm_eigen
  # -- the loadings and uniquenesses efa_nest() passes it -- is exercised by a fast test. Two
  # well-separated factors keep the retained count stable at a small number of simulated
  # datasets; the nf = 3 step builds a two-factor reference, covering a multi-column
  # loading matrix in the draw.
  R2 <- matrix(.1, 6, 6)
  R2[1:3, 1:3] <- .6
  R2[4:6, 4:6] <- .6
  diag(R2) <- 1

  set.seed(2024)
  nest_2f <- efa_nest(R2, N = 300, n_datasets = 100)

  expect_s3_class(nest_2f, "efa_retention")
  expect_equal(nest_2f$n_factors[["NEST"]], 2)
})

if (is_slow_test()) rm(nest_cor, nest_raw)

x_mat <- matrix(c(.1, .1, .3,
                  .32, .4, .1,
                   0,  0,  0), ncol = 3, byrow = TRUE)
y_mat <- matrix(c(.1, .1, .3,
                  .4, .29, .1,
                   0,  0,  0), ncol = 3, byrow = TRUE)
test_that(".factor_corres works", {
  expect_equal(.factor_corres(x_mat, y_mat)$diff_corres, 1)
  expect_equal(.factor_corres(y_mat, x_mat)$diff_corres_cross, 1)

  expect_equal(.factor_corres(x_mat,
                              matrix(0, ncol = 3, nrow = 3))$diff_corres, 2)
  expect_equal(.factor_corres(x_mat,
                              matrix(0, ncol = 3, nrow = 3))$diff_corres_cross, 2)
  expect_equal(.factor_corres(matrix(0, ncol = 3, nrow = 3),
                              x_mat)$diff_corres, 2)
  expect_equal(.factor_corres(matrix(0, ncol = 3, nrow = 3),
                              x_mat)$diff_corres_cross, 2)

  expect_equal(.factor_corres(x_mat, x_mat)$diff_corres, 0)
  expect_equal(.factor_corres(x_mat, x_mat)$diff_corres_cross, 0)
})

test_that(".factor_corres cross-loading sets do not collide for >= 10 factors", {
  # One indicator loads on factors {1, 2}; the other on factor {12}. Without a
  # separator both collapse to "12" and the differing correspondence is missed.
  x12 <- matrix(0, nrow = 1, ncol = 12)
  x12[1, 1] <- x12[1, 2] <- 0.5
  y12 <- matrix(0, nrow = 1, ncol = 12)
  y12[1, 12] <- 0.5
  expect_equal(.factor_corres(x12, y12)$diff_corres_cross, 1)
})

test_that(".factor_corres ignores missing cells rather than the whole row", {
  x <- rbind(c(.8, NA), c(.1, .7))
  y <- rbind(c(.8, .1), c(.1, .7))
  out <- .factor_corres(x, y)
  expect_identical(out$x_corres, c(1L, 2L))
  expect_identical(out$diff_corres, 0L)
  expect_identical(out$diff_corres_cross, 0L)

  # A factor position missing on either side is excluded from both classifications.
  pairwise <- .factor_corres(matrix(c(.8, NA), nrow = 1),
                             matrix(c(.4, .9), nrow = 1))
  expect_identical(pairwise$x_corres, 1L)
  expect_identical(pairwise$y_corres, 1L)
  expect_identical(pairwise$diff_corres, 0L)

  # Rows without any jointly observed loading are unclassifiable and do not count
  # as a difference; if there are no comparable rows, the count is undefined.
  partly_empty <- .factor_corres(rbind(c(NA, NA), c(.7, .1)),
                                  rbind(c(.2, .8), c(.6, .2)))
  expect_true(is.na(partly_empty$x_corres[1]))
  expect_identical(partly_empty$diff_corres, 0L)
  fully_empty <- .factor_corres(matrix(c(NA, NA), nrow = 1),
                                 matrix(c(.2, .8), nrow = 1))
  expect_true(is.na(fully_empty$diff_corres))
})

test_that("native simulation and polychoric entry points reject malformed inputs", {
  expect_error(.parallel_sim(0, 3, 20, 1, 10), "n_datasets")
  expect_error(.parallel_sim(1, 3, 20, 9, 10), "eigen_type")
  expect_error(.parallel_sim(1, 3, 3, 2, 10), "larger than n_vars")

  bad_negative <- cbind(c(0L, 1L, -1L), c(0L, 1L, 0L))
  expect_error(.polychoric_cpp(bad_negative, "none", FALSE), "non-negative")
  bad_gap <- cbind(c(0L, 2L, 2L), c(0L, 1L, 0L))
  expect_error(.polychoric_cpp(bad_gap, "none", FALSE), "consecutive")
})

rm(x_mat, y_mat)

# EFAs of ten similar datasets (bootstrap samples of DOSPERT_raw). Only the first two are
# read by the blocks that run everywhere; samples 3..10 feed the warm-start comparison
# alone, which is skip_on_cran(). Each is a six-factor ML fit with promax on 3,123 x 30
# rows, so they are fitted on first request rather than up front.
#
# The resample row indices are drawn eagerly, in the original order and under the original
# seed, and only the fits are deferred. That keeps every fixture byte-identical to the
# eager version: an ML extraction with promax draws no random numbers, so interleaving the
# draws with the fits never mattered (verified against the eager construction).
set.seed(42)
boot_rows <- lapply(1:10, function(x) sample(1:nrow(DOSPERT_raw), replace = TRUE))

efa_at <- local({
  cache <- vector("list", 10)
  function(d) {
    if (is.null(cache[[d]])) {
      cache[[d]] <<- EFA(DOSPERT_raw[boot_rows[[d]], ], n_factors = 6, method = "ML",
                         rotation = "promax")
    }
    cache[[d]]
  }
})

unrot_loadings <- function(d) efa_at(d)$unrot_loadings
rot_loadings <- function(d) efa_at(d)$rot_loadings


test_that("efa_procrustes matches psych::Procrustes and GPArotation::targetQ outputs", {
  expect_equal(
    psych::Procrustes(unrot_loadings(2), rot_loadings(1))$loadings,
    efa_procrustes(unrot_loadings(2), rot_loadings(1))$loadings,
    ignore_attr = TRUE
    )
  # The default single start of GPArotation::targetQ() is itself trapped in a
  # local minimum on this matrix; random starts are needed for it to reach the
  # global oblique target-rotation optimum that efa_procrustes() now finds.
  skip_if_not_installed("GPArotation")
  set.seed(42)
  oracle <- suppressWarnings(
    GPArotation::targetQ(unrot_loadings(2), Target = rot_loadings(1),
                         randomStarts = 20)
  )
  expect_equal(
    oracle$loadings,
    efa_procrustes(unrot_loadings(2), rot_loadings(1), rotation = "oblique")$loadings,
    tolerance = 1e-3,
    ignore_attr = TRUE
  )
})


test_that("oblique efa_procrustes recovers a known oblique transform from its default start", {
  skip_on_cran()

  # Self-recovery: with A = L0 %*% t(T0) and target L0, the oblique optimum is T0
  # with objective 0. The default warm orthogonal-Procrustes start lands in the
  # global basin and recovers the generating loadings and factor correlations.
  L0 <- matrix(0, 9, 3)
  L0[1:3, 1] <- c(.75, .70, .65)
  L0[4:6, 2] <- c(.72, .68, .62)
  L0[7:9, 3] <- c(.80, .74, .66)
  L0[1, 2] <- .15; L0[5, 3] <- .12; L0[8, 1] <- .10
  Phi0 <- matrix(c(1, .3, .2, .3, 1, .25, .2, .25, 1), 3)
  T0 <- chol(Phi0)                       # unit-norm columns; t(T0) %*% T0 == Phi0
  A <- L0 %*% t(T0)

  fit <- efa_procrustes(A, Target = L0, rotation = "oblique")

  expect_lt(fit$value, 1e-6)
  expect_true(fit$convergence)
  expect_equal(unclass(fit$loadings), L0, tolerance = 1e-4, ignore_attr = TRUE)
  expect_equal(fit$Phi, Phi0, tolerance = 1e-4, ignore_attr = TRUE)

  # matches the GPArotation target-rotation oracle on the same clean problem
  skip_if_not_installed("GPArotation")
  oracle <- GPArotation::targetQ(A, Target = L0)
  expect_equal(unclass(fit$loadings), unclass(oracle$loadings),
               tolerance = 1e-3, ignore_attr = TRUE)
})

test_that("oblique efa_procrustes warm start escapes identity-start local minima", {
  skip_on_cran()

  # On the 6-factor bootstrap loadings the identity start is trapped in poor local
  # minima (still reporting convergence) for some matrices; the default warm
  # orthogonal-Procrustes start reaches a far lower target-criterion value.
  target <- rot_loadings(1)
  warm <- vapply(2:10, function(d)
    efa_procrustes(unrot_loadings(d), Target = target, rotation = "oblique")$value,
    numeric(1))
  identity_start <- vapply(2:10, function(d)
    efa_procrustes(unrot_loadings(d), Target = target, rotation = "oblique",
               T_init = diag(6))$value,
    numeric(1))

  # The default start is never worse than the identity start. Where the two starts land in
  # the same basin the objectives differ only by the solver's own stopping slack: the
  # projected-gradient norm is driven below oblique_eps = 1e-5, and at a flat oblique optimum
  # the corresponding objective gap is ||g||^2 / (2 * lambda), which for a small curvature
  # lambda can exceed 1e-16-scale rounding by several orders. The slack is therefore stated
  # at the solver's own resolution rather than at machine precision -- this is an all() over
  # nine bootstrap matrices, so a single such tie would otherwise fail it. The contract is
  # "never worse", which the >1-unit rescue asserted next shows is not vacuous.
  expect_true(all(warm <= identity_start + 1e-4))
  # ... and rescues at least one matrix the identity start traps badly
  expect_gt(max(identity_start - warm), 1)
})

test_that("oblique multi-start selection keeps the lowest objective, not the converged one", {
  # A converged suboptimal local minimum must not be preferred over a strictly
  # better but not-yet-converged fit. Use a reflected-factor stationary point
  # (a genuine converged local minimum) as the primary start, then let the random
  # starts reach the global optimum under a budget too small to converge there.
  set.seed(6)
  k <- 2
  p <- 8
  L0 <- matrix(runif(p * k, -.1, .1), p, k)
  for (j in seq_len(k)) L0[((j - 1) * 4 + 1):(j * 4), j] <- runif(4, .5, .85)
  T0 <- chol(matrix(c(1, .7, .7, 1), 2))
  A <- L0 %*% t(T0)

  # a converged suboptimal stationary point (objective ~2.13 vs 0 at the global)
  trapped <- efa_procrustes(A, Target = L0, rotation = "oblique",
                        T_init = T0 %*% diag(c(-1, 1)), oblique_maxit = 1000)$T

  set.seed(2)
  fit <- efa_procrustes(A, Target = L0, rotation = "oblique", T_init = trapped,
                    oblique_random_starts = 12, oblique_screen_keep = 12,
                    oblique_maxit = 3, oblique_triage_maxit = 2)

  # the returned fit is the minimum-objective start ...
  expect_equal(fit$value, min(fit$all_values))
  expect_lt(fit$value, 0.5)
  # ... even though the primary start (index 1) converged at a much higher
  # objective: a convergence-first rule would have returned that worse solution.
  expect_true(as.logical(fit$all_converged[1]))
  expect_gt(fit$all_values[1], fit$value + 0.5)
})


test_that("efa_procrustes rejects invalid matrices and oblique controls", {
  A <- rot_loadings(1)

  expect_error(efa_procrustes("a", A), class = "efa_not_matrix")
  expect_error(efa_procrustes(A, replace(A, 1, NA)), class = "efa_nonfinite")
  expect_error(efa_procrustes(matrix(numeric(0), 0, 0), A), class = "efa_empty_matrix")

  # the oblique control checks run only on the oblique branch
  ob <- function(...) efa_procrustes(A, A, rotation = "oblique", ...)
  expect_error(ob(oblique_maxit = 2.5), class = "efa_not_integer")
  expect_error(ob(oblique_eps = -1), class = "efa_out_of_range")
  expect_error(ob(oblique_eps = c(1, 2)), class = "efa_not_scalar")
  expect_error(ob(oblique_normalize = "y"), class = "efa_not_flag")

  k <- ncol(A)
  T_zero <- diag(k)
  T_zero[, 1] <- 0
  expect_error(ob(T_init = T_zero), class = "efa_zero_column")
  # nonzero columns throughout, but rank k - 1
  T_sing <- diag(k)
  T_sing[, k] <- T_sing[, 1]
  expect_error(ob(T_init = T_sing), class = "efa_singular")
})


test_that("S is rejected unless it is crossprod(A)", {
  A <- unrot_loadings(2)
  target <- rot_loadings(1)
  S_ok <- crossprod(A)
  # symmetric, correctly sized, and a perfectly plausible slip: the cross-product of a
  # different group's loadings. It enters the oblique criterion and its gradient, so
  # the solver would minimize a different problem and report that problem's value.
  S_wrong <- crossprod(unrot_loadings(1))
  S_asym <- S_ok
  S_asym[1, 2] <- S_asym[1, 2] + 1

  expect_error(efa_procrustes(A, target, rotation = "oblique", S = S_wrong),
               class = "efa_bad_crossprod")
  expect_error(efa_procrustes(A, target, rotation = "oblique", S = S_asym),
               class = "efa_bad_crossprod")
  expect_error(efa_procrustes(A, target, rotation = "oblique", S = S_ok[-1, -1]),
               class = "efa_dim_mismatch")

  # the correct cross-product is accepted, and is only the shortcut it claims to be
  fit_S <- efa_procrustes(A, target, rotation = "oblique", S = S_ok)
  fit_internal <- efa_procrustes(A, target, rotation = "oblique")
  expect_equal(fit_S$value, fit_internal$value, tolerance = 1e-8)
  expect_equal(fit_S$loadings, fit_internal$loadings, tolerance = 1e-6)
})

test_that("S is left unchecked, and unused, on the paths that never consume it", {
  A <- unrot_loadings(2)
  target <- rot_loadings(1)
  S_wrong <- crossprod(unrot_loadings(1))
  A_1 <- A[, 1, drop = FALSE]
  target_1 <- target[, 1, drop = FALSE]

  # Asserted as equality with the same call without S, not merely as the absence of
  # an error: a wrong S must not reach the result on these paths either.

  # orthogonal alignment is closed-form and never sees S ...
  expect_equal(efa_procrustes(A, target, rotation = "orthogonal", S = S_wrong),
               efa_procrustes(A, target, rotation = "orthogonal"))
  # ... a one-factor oblique request is answered with the orthogonal solution ...
  expect_equal(efa_procrustes(A_1, target_1, rotation = "oblique",
                              S = S_wrong[1, 1, drop = FALSE]),
               efa_procrustes(A_1, target_1, rotation = "oblique"))
  # ... and Kaiser normalization changes the loadings, so the solver recomputes the
  # cross-product on the normalized matrix regardless of what was supplied.
  expect_equal(efa_procrustes(A, target, rotation = "oblique", S = S_wrong,
                              oblique_normalize = TRUE, oblique_maxit = 5),
               efa_procrustes(A, target, rotation = "oblique",
                              oblique_normalize = TRUE, oblique_maxit = 5))
})

test_that("the native oblique solver uses only the symmetric part of S", {
  A <- unrot_loadings(2)
  S <- crossprod(A)
  S[1, 2] <- S[1, 2] + 1
  S_sym <- S / 2 + t(S) / 2

  # The quadratic form depends only on the symmetric part of S, and the analytic
  # gradient assumes that same matrix. The registered native entry point enforces
  # this contract when called directly, without the R validator above it.
  native <- .oblique_procrustes(A, rot_loadings(1), S_r = S)
  native_sym <- .oblique_procrustes(A, rot_loadings(1), S_r = S_sym)
  expect_equal(native$value, native_sym$value)
  expect_equal(native$loadings, native_sym$loadings)
})

test_that("smooth Procrustes does not use objective stalling as convergence", {
  set.seed(1)
  A <- matrix(stats::rnorm(24), 8, 3)
  B <- A + matrix(stats::rnorm(24), 8, 3) * 1e-7

  fit <- .oblique_procrustes(A, B, T_init_r = diag(3), eps = 1e-30,
                             maxit = 10L)

  expect_false(fit$convergence)
  expect_gt(10^tail(fit$Table[, "log10_s"], 1), 1e-30)
})

test_that("the oblique multi-start screen scores a start by the objective at that start", {
  # The screen ranks random starts by the objective alone. Under a zero-iteration budget
  # every optimization reports exactly the objective at the start it was handed, so the
  # screened value and the optimized value of each start must agree start for start -- a
  # screen scoring starts by any other quantity would diverge here.
  set.seed(15)
  A <- matrix(stats::rnorm(16), 8, 2)
  B <- A + matrix(stats::rnorm(16, sd = .1), 8, 2)

  set.seed(3)
  fit <- .oblique_procrustes(A, B, random_starts = 8L, screen_keep = 8L,
                             triage_maxit = 0L, maxit = 0L)

  expect_length(fit$screen_values, 8L)
  # screened ascending, so the retained starts are the best-scoring ones
  expect_false(is.unsorted(fit$screen_values))
  optimized <- fit$all_values[match(fit$screen_start_indices, fit$all_start_indices)]
  expect_equal(optimized, fit$screen_values, tolerance = 1e-8)
})

test_that("piecewise-smooth simplimax retains objective-stall convergence", {
  set.seed(1)
  L <- matrix(stats::rnorm(24), 8, 3)
  fit <- .rotate_simplimax_oblq(L, k = nrow(L), eps = 1e-30,
                                normalize = FALSE, random_starts = 0L,
                                maxit = 100L)
  expect_true(fit$convergence)
})


rm(boot_rows, efa_at, unrot_loadings, rot_loadings)

#' UPPS_raw
#'
#' A dataframe containing responses to the UPPS personality scale (Whiteside &
#' Lynam, 2005) of 645 participants of Study 2 of Steiner and Frey (2020). Each
#' column are the ratings to one of 45 items to assess urgency, premeditation,
#' perseverance, and sensation seeking. The original data can be accessed via
#' <https://osf.io/kxp8t/>.
#'
#' @format A data.frame with 645 rows (participants) and 45 columns, named by a
#' subscale prefix and item number, covering the four UPPS subscales:
#' \describe{
#'   \item{perseverance_1 to perseverance_10}{(numeric) - Perseverance-subscale items.}
#'   \item{premeditation_1 to premeditation_11}{(numeric) - Premeditation-subscale items.}
#'   \item{ss_1 to ss_12}{(numeric) - Sensation-seeking-subscale items.}
#'   \item{urgency_1 to urgency_12}{(numeric) - Urgency-subscale items.}
#'  }
#' @source Whiteside, S. P., Lynam, D. R., Miller, J. D., & Reynolds, S. K. (2005).
#'  Validation of the UPPS impulsive behaviour scale: A four-factor model of
#'  impulsivity. European Journal of Personality, 19 (7), 559–574.
#' @source Steiner, M., & Frey, R. (2020). Representative design in psychological assessment: A case study using the Balloon Analogue Risk Task (BART). PsyArXiv Preprint. doi:10.31234/osf.io/dg4ks
"UPPS_raw"

## Principal Axis Factoring
## Thin fitter: returns the raw PAF results; shared post-processing happens in
## .finalize_fit() / .estimate_model().
.PAF <- function(x, n_factors, max_iter = NA,
                type = c("EFAtools", "psych", "SPSS", "none"),
                init_comm = NA, criterion = NA,
                criterion_type = NA, abs_eigen = NA) {

  # Get correlation matrix entered or created in EFA
  R <- x

  # Fill the type preset defaults and warn about any pinned tuning arguments.
  resolved <- .resolve_settings(
    type = type,
    user = list(init_comm = init_comm, criterion = criterion,
                criterion_type = criterion_type, max_iter = max_iter,
                abs_eigen = abs_eigen),
    preset = .efa_presets$PAF
  )
  init_comm <- resolved$init_comm
  criterion <- resolved$criterion
  criterion_type <- resolved$criterion_type
  max_iter <- resolved$max_iter
  abs_eigen <- resolved$abs_eigen

  # set initial communality estimates. This can be done in three ways:
  #  init_comm == "smc": uses the Squared Multiple Correlation
  #  init_comm == "unity": uses unity as inital estimates
  #  init_comm == "mac": uses Maximum Absolute Correlations
  if (init_comm == "smc") {

    # save original correlation matrix
    orig_R <- R

    # compute smcs
    if (type == "psych") {

      h2_init <- psych::smc(R)

    } else {

      h2_init <- .smc_start(R)

    }

    # set diagonal of R to the initial communality estimates
    diag(R) <- h2_init


  } else if (init_comm == "unity") {
    # create h2_init object with a vector of 1s
    h2_init <- diag(R)

    # save original correlation matrix
    orig_R <- R

  } else if (init_comm == "mac") {

    # save original correlation matrix
    orig_R <- R

    # avoid using the diagonal as maximum correlations
    diag(R) <- 0

    # get maximum absolute correlations
    h2_init <- apply(R, 1, function(x){
      max(abs(x))
    })

    # set diagonal of R to the initial communality estimates
    diag(R) <- h2_init

  }

  # save initial eigenvalues
  init_eigen <- eigen(R, symmetric = TRUE, only.values = TRUE)$values

  # define the number of factors m
  m <- n_factors

  # set communalities to init_communalities for comparison in first iteration
  h2 <- h2_init

  crit_type <- ifelse(criterion_type == "max_individual", 1, 2)

  # run the iterative PAF procedure using Rcpp
  L_list <- .paf_iter(h2 = h2, criterion = criterion, R = R, n_fac = m,
                     abs_eig = abs_eigen, crit_type = crit_type,
                     max_iter = max_iter)

  # Negative eigenvalues make the communality estimates (square roots of the
  # eigenvalues) undefined. The kernel flags this rather than aborting so the
  # condition can be raised here as a classed R condition.
  if (isTRUE(L_list$neg_eigen)) {
    cli::cli_abort(
      c("Negative eigenvalues detected; cannot compute communality estimates.",
        "i" = 'Try {.code init_comm = "unity"} or {.code init_comm = "mac"}.'),
      class = "efa_paf_negative_eigen"
    )
  }

  h2 <- as.vector(L_list$h2)
  iter <- L_list$iter
  L <- L_list$L

  # save convergence status (0 = converged, 1 = not converged): the procedure has
  # not converged if the final change in the communality estimates still exceeds
  # the convergence criterion.
  if (isTRUE(L_list$delta > criterion)) {
    convergence <- 1
    cli::cli_warn(
      "Reached the maximum number of iterations without convergence; results may not be interpretable.",
      class = "efa_paf_nonconvergence"
    )
  } else {
    convergence <- 0
  }

  # store the settings used:
  settings <- list(
    max_iter = max_iter,
    init_comm = init_comm,
    criterion = criterion,
    criterion_type = criterion_type,
    abs_eigen = abs_eigen
  )

  # raw fit, finalized by .estimate_model()
  list(
    L = L,
    h2 = h2,
    Fm = NA_real_,
    iter = iter,
    convergence = convergence,
    orig_R = orig_R,
    R_final = L_list$R,
    h2_init = h2_init,
    init_eigen = init_eigen,
    settings = settings
  )

}

# Opt-in gate for slow tests. The heavy retention (HULL, PARALLEL, NEST, efa_retain),
# np-boot, polychoric, multimodal-rotation, lavaan-oracle, and grid-of-EFAs fixtures
# dominate the suite's runtime; they are skipped by default and run only when
# EFATOOLS_TEST_SLOW is set truthy, e.g. `Sys.setenv(EFATOOLS_TEST_SLOW = "true")` (or
# `EFATOOLS_TEST_SLOW=true` in the shell) before devtools::test() / devtools::check().
# `is_slow_test()` is the predicate the gate is built on, so file-top fixture blocks can
# guard with `if (is_slow_test()) { ... }` to avoid building heavy fixtures by default.
is_slow_test <- function() {
  isTRUE(as.logical(Sys.getenv("EFATOOLS_TEST_SLOW")))
}
skip_if_not_slow <- function() {
  testthat::skip_if_not(
    is_slow_test(),
    "slow test; set EFATOOLS_TEST_SLOW=true to run"
  )
}

# Counts the gated blocks in the test files themselves, so the number the notice below
# reports cannot drift away from the suite. Returns 0 when called from outside the test
# directory, where the files are not visible.
n_slow_gated <- function() {
  sum(vapply(
    list.files(pattern = "^test-.*\\.[Rr]$"),
    function(path) {
      sum(grepl("skip_if_not_slow()", readLines(path, warn = FALSE), fixed = TRUE))
    },
    integer(1)
  ))
}

# A skip is not a failure, so a default run can look complete when it is not. State once
# what the gate holds back -- a fact about the suite, not about the blocks this particular
# run happens to touch. local() keeps the count out of the environment the tests run in.
if (!is_slow_test()) local({
  n_gated <- n_slow_gated()
  if (n_gated > 0) {
    cli::cli_inform(c("i" = paste(
      "The suite has {n_gated} slow test block{?s} gated off; set",
      "{.envvar EFATOOLS_TEST_SLOW}=true (with {.envvar NOT_CRAN}=true) to run them."
    )))
  }
})

# HULL refits the model at every candidate factor count for every (method, gof)
# combination, which makes the file-top fixture block the single heaviest in the
# suite (~60s). Skip the fixtures + the tests below by default; opt in with
# `Sys.setenv(EFATOOLS_TEST_SLOW = "true")` to run the full HULL coverage. See
# helper-slow.R for the convention.
if (is_slow_test()) {
hull_cor_paf <- suppressMessages(efa_hull(test_models$baseline$cormat, N = 500))
hull_cor_ml <- efa_hull(test_models$baseline$cormat, N = 500, estimator = "ML")
hull_cor_uls <- efa_hull(test_models$baseline$cormat, N = 500, estimator = "ULS")
hull_cor_uls_CFI <- efa_hull(test_models$baseline$cormat, N = 500, estimator = "ULS",
                         gof = "CFI")
hull_cor_ml_nf <- suppressWarnings(efa_hull(test_models$baseline$cormat, N = 500,
                                        estimator = "ML", n_fac_theor = 12))

hull_PCA <- efa_hull(test_models$baseline$cormat, N = 500, estimator = "ML",
                 eigen_type = "PCA")
hull_EFA <- efa_hull(test_models$baseline$cormat, N = 500, estimator = "ML",
                 eigen_type = "EFA")

hull_raw_paf <- suppressMessages(suppressWarnings(efa_hull(GRiPS_raw)))
hull_raw_ml <- suppressMessages(suppressWarnings(efa_hull(GRiPS_raw, estimator = "ML")))
hull_raw_uls <- suppressMessages(suppressWarnings(efa_hull(GRiPS_raw, estimator = "ULS")))
hull_raw_uls_CFI <- suppressMessages(suppressWarnings(efa_hull(GRiPS_raw,
                                                           estimator = "ULS",
                                                           gof = "CFI")))
hull_raw_ml_nf <- suppressMessages(suppressWarnings(efa_hull(GRiPS_raw, N = 500,
                                                         estimator = "ML",
                                                         n_fac_theor = 7)))
}  # is_slow_test()


test_that("output class and dimensions are correct", {
  skip_if_not_slow()
  objs <- list(hull_cor_paf, hull_cor_ml, hull_cor_uls, hull_cor_uls_CFI,
               hull_PCA, hull_EFA, hull_raw_paf, hull_raw_ml, hull_raw_uls,
               hull_raw_uls_CFI)
  for (obj in objs) {
    expect_s3_class(obj, "efa_retention")
    expect_length(obj, 9)
    expect_length(obj$settings, 9)
  }
})

test_that("n_fac_max is correctly specified", {
  skip_if_not_slow()
  expect_lte(hull_cor_paf$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_cor_ml$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_cor_uls$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_cor_uls_CFI$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_PCA$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_lte(hull_EFA$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))

  expect_lte(hull_raw_paf$settings$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))
  expect_lte(hull_raw_ml$settings$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))
  expect_lte(hull_raw_uls$settings$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))
  expect_lte(hull_raw_uls_CFI$settings$n_fac_max,
             .det_max_factors(ncol(GRiPS_raw)))

  expect_equal(hull_cor_ml_nf$settings$n_fac_max,
             .det_max_factors(ncol(test_models$baseline$cormat)))
  expect_equal(hull_raw_ml_nf$settings$n_fac_max,
               .det_max_factors(ncol(GRiPS_raw)))
})

test_that("the factor-search bound stays over-identified with the minimum number of indicators", {
  skip_if_not_slow()
  # With 6 indicators the largest over-identified model has 2 factors (a 3-factor
  # solution is just-identified, df = 0). The search bound J must be capped at that
  # maximum and never forced up to 3, so the Hull method cannot select a df = 0 model.
  R6 <- test_models$baseline$cormat[1:6, 1:6]
  out <- suppressWarnings(suppressMessages(efa_hull(R6, N = 500, estimator = "ULS")))
  expect_equal(out$settings$n_fac_max, .det_max_factors(6))
  expect_lte(max(out$n_factors, na.rm = TRUE), .det_max_factors(6))
})

test_that("records are correctly returned", {
  skip_if_not_slow()
  expect_named(hull_cor_paf$n_factors, "CAF")
  expect_named(hull_cor_paf$results[[1]],
               c("name", "label", "n_factors", "plot_type", "x", "y", "reference",
                 "threshold", "highlight", "point_labels", "on_hull", "st"))
  expect_equal(hull_cor_paf$results[[1]]$plot_type, "hull")
  checkmate::expect_numeric(hull_cor_paf$results[[1]]$y)
  checkmate::expect_numeric(hull_cor_paf$results[[1]]$x)
  checkmate::expect_logical(hull_cor_paf$results[[1]]$on_hull)

  # the subtitle names the estimator under the name the argument has
  expect_equal(hull_cor_paf$subtitle, "Estimator: PAF")
  expect_equal(hull_cor_ml$subtitle, "Estimator: ML")

  expect_named(hull_cor_uls$n_factors, c("CAF", "CFI", "RMSEA"))
  expect_named(hull_cor_uls_CFI$n_factors, "CFI")

  expect_named(hull_raw_paf$n_factors, "CAF")
  expect_named(hull_raw_uls$n_factors, c("CAF", "CFI", "RMSEA"))
  expect_named(hull_raw_uls_CFI$n_factors, "CFI")
})

test_that("n_factors are correctly returned", {
  skip_if_not_slow()
  expect_equal(hull_cor_paf$n_factors[["CAF"]], 3)
  expect_false("CFI" %in% names(hull_cor_paf$n_factors))
  expect_false("RMSEA" %in% names(hull_cor_paf$n_factors))

  expect_equal(hull_cor_uls$n_factors[["CAF"]], 3)
  expect_equal(hull_cor_uls$n_factors[["CFI"]], 1)
  expect_equal(hull_cor_uls$n_factors[["RMSEA"]], 1)
  expect_equal(hull_cor_uls$n_fac_CAF, hull_cor_uls$n_factors[["CAF"]])
  expect_equal(hull_cor_uls$n_fac_CFI, hull_cor_uls$n_factors[["CFI"]])
  expect_equal(hull_cor_uls$n_fac_RMSEA, hull_cor_uls$n_factors[["RMSEA"]])

  expect_equal(hull_cor_uls_CFI$n_factors[["CFI"]], 1)
  expect_false("CAF" %in% names(hull_cor_uls_CFI$n_factors))
  expect_false("RMSEA" %in% names(hull_cor_uls_CFI$n_factors))
  expect_equal(hull_cor_uls_CFI$n_fac_CAF, NA_real_)
  expect_equal(hull_cor_uls_CFI$n_fac_CFI, hull_cor_uls_CFI$n_factors[["CFI"]])
  expect_equal(hull_cor_uls_CFI$n_fac_RMSEA, NA_real_)

  expect_equal(hull_raw_paf$n_factors[["CAF"]], 1)
  expect_false("CFI" %in% names(hull_raw_paf$n_factors))
  expect_false("RMSEA" %in% names(hull_raw_paf$n_factors))

  expect_equal(hull_raw_uls$n_factors[["CAF"]], 1)
  expect_equal(hull_raw_uls$n_factors[["CFI"]], 1)
  expect_equal(hull_raw_uls$n_factors[["RMSEA"]], 1)

  expect_equal(hull_raw_uls_CFI$n_factors[["CFI"]], 1)
  expect_false("CAF" %in% names(hull_raw_uls_CFI$n_factors))
  expect_false("RMSEA" %in% names(hull_raw_uls_CFI$n_factors))
})

test_that("every record carries the elbow sharpness of its solutions", {
  skip_if_not_slow()
  for (obj in list(hull_cor_paf, hull_cor_uls, hull_cor_uls_CFI, hull_raw_uls)) {
    for (rec in obj$results) {
      # one st per solution, aligned with the df / fit / hull-membership series
      expect_length(rec$st, length(rec$x))
      # st is defined only for the interior hull solutions: NA off the hull ...
      expect_true(all(is.na(rec$st[!rec$on_hull])))
      # ... and, asserted in the other direction so an all-NA regression cannot
      # pass, a value at every hull solution that has a neighbour on both sides
      # (with fewer than three hull solutions there are none and st is all NA)
      hull <- which(rec$on_hull)
      if (length(hull) >= 3) {
        expect_false(any(is.na(rec$st[hull[-c(1, length(hull))]])))
        # and the retained solution is the one with the largest st
        expect_equal(rec$point_labels[which.max(rec$st)], rec$n_factors)
      } else {
        expect_true(all(is.na(rec$st)))
      }
    }
  }
})

test_that("the convex-hull elimination tests every interior triplet", {
  # Hand-built fit table (columns: nfactors, goodness-of-fit, df, st). The fit
  # values increase monotonically, so the boundary step removes nothing; the
  # geometry is set so the last interior solution (two factors) lies below the
  # line connecting its neighbours and must be dropped from the hull.
  s <- cbind(0:3, c(0.00, 0.80, 0.85, 1.00), c(10, 6, 3, 1), 0)
  out <- .hull_calc(s, J = 3, gof_t = "CAF")

  # the below-chord last interior solution is off the hull (NA st) ...
  expect_true(is.na(out$s_complete[3, 4]))
  # ... while the supporting interior solution remains on it
  expect_false(is.na(out$s_complete[2, 4]))
  expect_equal(out$retain, 1)

  # the two hull endpoints have no neighbouring hull solution on one side, so their
  # st is undefined and reported as NA rather than as the initialised zero
  expect_true(is.na(out$s_complete[1, 4]))
  expect_true(is.na(out$s_complete[4, 4]))

  # a last interior solution lying above its neighbours' chord is kept
  s2 <- cbind(0:3, c(0.00, 0.85, 0.95, 1.00), c(10, 6, 3, 1), 0)
  out2 <- .hull_calc(s2, J = 3, gof_t = "CAF")
  expect_false(is.na(out2$s_complete[3, 4]))

  # a fully convex fit curve collapses the hull below three solutions; the
  # max-fit fallback is used instead of indexing past the remaining solutions
  s3 <- cbind(0:4, c(0.00, 0.10, 0.25, 0.55, 1.00), c(10, 8, 6, 3, 1), 0)
  expect_warning(out3 <- .hull_calc(s3, J = 4, gof_t = "CAF"),
                 class = "efa_hull_few_solutions")
  expect_equal(out3$retain, 4)
})

test_that("a non-finite goodness-of-fit value is dropped with a classed warning", {
  # A model with an undefined fit value (e.g. an NA CFI for a Heywood or
  # near-singular solution) cannot lie on the hull; it must be excluded with a
  # classed warning instead of crashing the elimination comparisons.
  s <- cbind(0:4, c(0.00, 0.80, NA, 0.90, 1.00), c(10, 8, 6, 3, 1), 0)
  expect_warning(out <- .hull_calc(s, J = 4, gof_t = "CFI"),
                 class = "efa_hull_na_fit")
  expect_false(is.na(out$retain))
  # the dropped (NA-fit) solution carries no hull membership
  expect_true(is.na(out$s_complete[3, 4]))
})


burt <- .burt_cormat()

test_that("errors etc are thrown correctly", {
  skip_if_not_slow()
  expect_error(efa_hull(1:5), class = "efa_input_not_matrix")
  expect_message(suppressWarnings(efa_hull(GRiPS_raw)), class = "efa_cor_from_data")
  expect_warning(suppressMessages(efa_hull(GRiPS_raw)), class = "efa_hull_min_factors")
  expect_warning(
    suppressWarnings(
      suppressMessages(efa_hull(IDS2_R, N = 20)),
      classes = "efa_hull_min_factors"
    ),
    class = "efa_hull_few_solutions"
  )
  expect_warning(
    suppressWarnings(
      suppressMessages(efa_hull(GRiPS_raw, N = 20)),
      classes = "efa_hull_min_factors"
    ),
    class = "efa_n_from_data"
  )
  expect_error(efa_hull(test_models$baseline$cormat), class = "efa_n_required")
  expect_error(efa_hull(test_models$baseline$cormat, estimator = "ML"), class = "efa_n_required")
  expect_error(efa_hull(test_models$baseline$cormat, estimator = "ULS"), class = "efa_n_required")

  expect_error(efa_hull(sing_raw, estimator = "ML"), class = "efa_cor_singular")
  expect_error(efa_hull(sing_cor, N = sing_N, estimator = "ML"),
               class = "efa_cor_singular")

  expect_error(efa_hull(matrix(rnorm(50), ncol = 5)), class = "efa_hull_min_indicators")

  expect_message(suppressWarnings(efa_hull(GRiPS_raw)), class = "efa_hull_gof_caf")

  expect_warning(efa_hull(test_models$baseline$cormat, n_fac_theor = 13, N = 500), class = "efa_hull_max_factors")
})

test_that("the Hull method needs at least one simulated dataset", {
  # refused at the argument boundary before anything is fitted, so this stays
  # outside the slow gate above
  expect_error(efa_hull(test_models$baseline$cormat, N = 500, n_datasets = 0),
               class = "efa_invalid_argument")
})

test_that("a non-positive-definite input is smoothed and its hull flagged", {
  # burt has to be smoothed, its hull collapses below three solutions, and the
  # selected solution is inadmissible. Several warnings fire, so collect their
  # classes rather than stopping at the first one (or muffling them wholesale).
  # An explicit n_fac_theor is not varied here: the internal parallel analysis
  # already suggests more factors than any small theoretical value, so the search
  # bound J -- and with it the whole result -- is the same either way.
  #
  # efa_hull() derives J from a simulated parallel analysis, so seed it rather than
  # inheriting whatever state the preceding tests left (which differs between the
  # default and the slow gate).
  #
  # The inadmissibility warning names the number of factors that was selected, so it
  # renders counts the way the report does. A negative options(scipen) is set here rather
  # than in a test of its own, so that the one fit covers both the conditions and that
  # text; the inner handler only reads the message and leaves the muffling to
  # `.warn_classes()`.
  withr::local_options(scipen = -5)
  set.seed(42)
  inadmissible <- NA_character_
  classes <- .warn_classes(withCallingHandlers(
    efa_hull(burt, N = 20, estimator = "ML"),
    efa_hull_inadmissible = function(w) inadmissible <<- conditionMessage(w)))

  expect_true("efa_cor_smoothed" %in% classes)
  expect_true("efa_hull_few_solutions" %in% classes)
  expect_true("efa_hull_inadmissible" %in% classes)

  # the hull stores its count as a double, so it would render as "1e+00" were the message
  # built with a bare paste0(); the pattern takes any exponent, because a count of ten or
  # more carries a different one
  expect_false(is.na(inadmissible))
  expect_false(grepl("e[+-][0-9]", inadmissible))
})

test_that("a hull with no finite goodness-of-fit value aborts", {
  # Every solution has an undefined fit value, so nothing can lie on the hull; the
  # elimination must abort with a classed condition instead of indexing an empty
  # matrix. The excluded-solutions warning fires first.
  s <- cbind(0:3, rep(NA_real_, 4), c(10, 6, 3, 1), 0)
  expect_error(
    suppressWarnings(.hull_calc(s, J = 3, gof_t = "CFI")),
    class = "efa_hull_no_fit"
  )
})

if (is_slow_test()) {
  rm(hull_cor_paf, hull_cor_ml, hull_cor_uls, hull_cor_uls_CFI, hull_raw_paf,
     hull_raw_ml, hull_raw_uls, hull_raw_uls_CFI, hull_raw_ml_nf, hull_cor_ml_nf,
     hull_PCA, hull_EFA)
}
rm(burt)

# efa_average runs its grid of EFAs through future.apply with future.seed = TRUE,
# so the random starts of the oblique rotation engines draw from the session RNG.
# Seed once here so the fixtures below - and the print snapshot built from them -
# are reproducible regardless of what ran earlier in the test run.
#
# The vector-valued default tuning settings expand to ~144 inner EFA fits per
# fixture, so the un-collapsed `efa_def` / `efa_all_none` / `efa_all_md`
# fixtures dominate the file (~25s combined). Skipped by default; opt in with
# `Sys.setenv(EFATOOLS_TEST_SLOW = "true")`. See helper-slow.R.
if (is_slow_test()) {
set.seed(42)

efa_def <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                       show_progress = FALSE)
efa_ml <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                      estimator = "ML", show_progress = FALSE)
efa_uls <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                      estimator = "ULS", show_progress = FALSE)

# Scalar overrides on the six default-vector tuning axes (init_comm,
# criterion_type, varimax_type, k_promax, P_type, start_method) for the
# fixtures whose downstream tests don't depend on the multi-value defaults.
# `efa_def` (above) still pins the vector defaults end-to-end; `efa_all_none`
# and `efa_all_md` (below) keep their full-grid expansion because they feed
# print-snapshot tests. Each of these scalar fixtures drops from ~1728 to ~12
# inner EFA fits.
.scalar_axes <- list(
  init_comm     = "smc",
  criterion_type = "sum",
  varimax_type  = "svd",
  k_promax      = 4,
  P_type        = "norm",
  start_method  = "psych"
)
efa_all <- do.call(efa_average, c(
  list(test_models$baseline$cormat, n_factors = 3, N = 500,
       estimator = c("PAF", "ML", "ULS"),
       type = c("none", "EFAtools", "psych", "SPSS"),
       salience_threshold = .2, show_progress = FALSE),
  .scalar_axes))
efa_all_oblq <- do.call(efa_average, c(
  list(test_models$baseline$cormat, n_factors = 3, N = 500,
       estimator = c("PAF", "ML", "ULS"),
       type = c("none", "EFAtools", "psych", "SPSS"),
       rotation = "oblique", show_progress = FALSE),
  .scalar_axes))
efa_all_orth <- do.call(efa_average, c(
  list(test_models$baseline$cormat, n_factors = 3, N = 500,
       estimator = c("PAF", "ML", "ULS"),
       type = c("none", "EFAtools", "psych", "SPSS"),
       rotation = "orthogonal", show_progress = FALSE),
  .scalar_axes))
efa_all_none <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                            estimator = c("PAF", "ML", "ULS"),
                            type = c("none", "EFAtools", "psych", "SPSS"),
                            rotation = "none", show_progress = FALSE)

efa_all_md <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                            estimator = c("PAF", "ML", "ULS"),
                            type = c("none", "EFAtools", "psych", "SPSS"),
                            rotation = "oblique", averaging = "median", show_progress = FALSE)
efa_all_tm <- do.call(efa_average, c(
  list(test_models$baseline$cormat, n_factors = 3, N = 500,
       estimator = c("PAF", "ML", "ULS"),
       type = c("none", "EFAtools", "psych", "SPSS"),
       rotation = "oblique", averaging = "mean",
       trim = .2, show_progress = FALSE),
  .scalar_axes))
efa_raw <- efa_average(GRiPS_raw, n_factors = 1, rotation = "none", show_progress = FALSE)
efa_raw_p <- efa_average(GRiPS_raw, n_factors = 2, rotation = "promax", show_progress = FALSE)
}  # is_slow_test()

test_that("efa_average runs end to end on a cheap grid (smoke test)", {
  # The rest of this file is gated behind EFATOOLS_TEST_SLOW, so without an
  # always-on test CRAN would never run efa_average() end to end. A two-row grid
  # (PAF and ML at the EFAtools preset, promax) fits in well under a second.
  local_reproducible_output()
  set.seed(42)
  res <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                     estimator = c("PAF", "ML"), type = "EFAtools",
                     rotation = "promax", start_method = "psych",
                     show_progress = FALSE)

  expect_s3_class(res, "efa_average")
  # PAF and ML at one preset and one rotation collapse to exactly two rows.
  expect_identical(nrow(res$implementations_grid), 2L)

  # Slot shapes: variables x factors loadings (carrying the LOADINGS class), a
  # square factor-correlation matrix from the oblique rotation, one communality
  # per variable, and the same-shaped indicator-to-factor correspondence matrix.
  m <- ncol(test_models$baseline$cormat)
  expect_s3_class(res$loadings$average, "LOADINGS")
  expect_identical(dim(res$loadings$average), c(m, 3L))
  expect_identical(dim(res$Phi$average), c(3L, 3L))
  expect_length(res$h2$average, m)
  expect_identical(dim(res$ind_fac_corres), c(m, 3L))

  # The averaged variance table carries three rows for more than one factor (the
  # single-factor form is pinned separately below).
  expect_identical(rownames(res$vars_accounted$average),
                   c("SS loadings", "Prop Tot Var", "Prop Comm Var"))

  # The averaging-rate sentence must never contain NaN (the empty-denominator guard).
  expect_no_match(paste(cli::ansi_strip(format(res)), collapse = " "), "NaN",
                  fixed = TRUE)
})

test_that("warnings raised while averaging reach the caller", {
  # The averaging step must not be run under a blanket suppression: an entirely missing
  # fit index is dealt with where the range is formed, so anything else the averaging
  # raises is a real signal and has to surface. Captured before mocking so the stand-in
  # can delegate to it instead of recursing.
  real_average_values <- .average_values
  testthat::local_mocked_bindings(
    .average_values = function(...) {
      cli::cli_warn("Something happened while averaging.",
                    class = "efa_test_average_warning")
      real_average_values(...)
    }
  )

  expect_warning(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = c("PAF", "ML"), type = "EFAtools", rotation = "promax",
                start_method = "psych", show_progress = FALSE),
    class = "efa_test_average_warning"
  )
})

test_that("an impossible N is rejected at the argument, not inside the grid", {
  # efa_average() runs every grid cell inside try(), so an N that efa_fit() refuses would
  # otherwise be recorded as a per-cell failure and surface as an unexplained
  # all-solutions-excluded result. It is rejected up front instead, with the same condition
  # class efa_fit() uses, while NA still means "sample size unknown".
  expect_error(efa_average(test_models$baseline$cormat, n_factors = 3, N = 0,
                           show_progress = FALSE),
               class = "efa_invalid_argument")
})

test_that("the averaged variance table drops 'Prop Comm Var' for a single factor", {
  # With one factor the proportion of common variance is identically 1, so the row is
  # omitted and these matrices have two rows rather than three. User code indexing a row by
  # name depends on it, so the documented return says so and this pins it.
  # Two grid rows, so the averaged (list) form is returned rather than the
  # single-combination shortcut's plain efa_fit() object.
  res_1 <- efa_average(test_models$baseline$cormat, n_factors = 1, N = 500,
                       estimator = c("PAF", "ML"), type = "EFAtools",
                       rotation = "none", start_method = "psych",
                       show_progress = FALSE)
  for (stat in c("average", "sd", "min", "max", "range")) {
    expect_identical(rownames(res_1$vars_accounted[[stat]]),
                     c("SS loadings", "Prop Tot Var"))
  }
})

test_that("output class and dimensions are correct", {
  skip_if_not_slow()
  expect_s3_class(efa_def, "efa_average")
  expect_s3_class(efa_ml, "efa_average")
  expect_s3_class(efa_uls, "efa_average")
  expect_s3_class(efa_all, "efa_average")
  expect_s3_class(efa_all_oblq, "efa_average")
  expect_s3_class(efa_all_orth, "efa_average")
  expect_s3_class(efa_all_none, "efa_average")
  expect_s3_class(efa_all_md, "efa_average")
  expect_s3_class(efa_all_tm, "efa_average")
  expect_s3_class(efa_raw, "efa_average")
  expect_s3_class(efa_raw_p, "efa_average")

  # the old class string is kept alongside the new one, so `inherits(x, "EFA_AVERAGE")`
  # in existing code still resolves
  expect_identical(class(efa_def), c("efa_average", "EFA_AVERAGE"))

  expect_s3_class(efa_def$loadings$average, "LOADINGS")
  expect_s3_class(efa_ml$loadings$average, "LOADINGS")
  expect_s3_class(efa_uls$loadings$average, "LOADINGS")
  expect_s3_class(efa_all$loadings$average, "LOADINGS")
  expect_s3_class(efa_all_oblq$loadings$average, "LOADINGS")
  expect_s3_class(efa_all_orth$loadings$average, "LOADINGS")
  expect_s3_class(efa_all_none$loadings$average, "LOADINGS")
  expect_s3_class(efa_all_md$loadings$average, "LOADINGS")
  expect_s3_class(efa_all_tm$loadings$average, "LOADINGS")
  expect_s3_class(efa_raw$loadings$average, "LOADINGS")
  expect_s3_class(efa_raw_p$loadings$average, "LOADINGS")

  expect_named(efa_def, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_ml, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_uls, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_oblq, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_orth, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_none, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_md, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_all_tm, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_raw, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
  expect_named(efa_raw_p, c("orig_R", "h2", "loadings", "Phi", "ind_fac_corres",
                          "vars_accounted", "fit_indices", "implementations_grid",
                          "efa_list", "settings"))
})

test_that("settings are returned correctly", {
  skip_if_not_slow()
  expect_named(efa_def$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold", "seed"))
  expect_named(efa_ml$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                  "init_comm", "criterion", "criterion_type",
                                  "abs_eigen", "varimax_type", "normalize",
                                  "k_promax", "k_simplimax", "P_type",
                                  "precision", "start_method", "use",
                                  "cor_method", "max_iter", "averaging",
                                  "trim", "salience_threshold", "seed"))
  expect_named(efa_uls$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold", "seed"))
  expect_named(efa_all$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold", "seed"))
  expect_named(efa_all_oblq$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                        "init_comm", "criterion", "criterion_type",
                                        "abs_eigen", "varimax_type", "normalize",
                                        "k_promax", "k_simplimax", "P_type",
                                        "precision", "start_method", "use",
                                        "cor_method", "max_iter", "averaging",
                                        "trim", "salience_threshold", "seed"))
  expect_named(efa_all_orth$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                        "init_comm", "criterion", "criterion_type",
                                        "abs_eigen", "varimax_type", "normalize",
                                        "k_promax", "k_simplimax", "P_type",
                                        "precision", "start_method", "use",
                                        "cor_method", "max_iter", "averaging",
                                        "trim", "salience_threshold", "seed"))
  expect_named(efa_all_none$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                        "init_comm", "criterion", "criterion_type",
                                        "abs_eigen", "varimax_type", "normalize",
                                        "k_promax", "k_simplimax", "P_type",
                                        "precision", "start_method", "use",
                                        "cor_method", "max_iter", "averaging",
                                        "trim", "salience_threshold", "seed"))
  expect_named(efa_all_md$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                      "init_comm", "criterion", "criterion_type",
                                      "abs_eigen", "varimax_type", "normalize",
                                      "k_promax", "k_simplimax", "P_type",
                                      "precision", "start_method", "use",
                                      "cor_method", "max_iter", "averaging",
                                      "trim", "salience_threshold", "seed"))
  expect_named(efa_all_tm$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                      "init_comm", "criterion", "criterion_type",
                                      "abs_eigen", "varimax_type", "normalize",
                                      "k_promax", "k_simplimax", "P_type",
                                      "precision", "start_method", "use",
                                      "cor_method", "max_iter", "averaging",
                                      "trim", "salience_threshold", "seed"))
  expect_named(efa_raw$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                   "init_comm", "criterion", "criterion_type",
                                   "abs_eigen", "varimax_type", "normalize",
                                   "k_promax", "k_simplimax", "P_type",
                                   "precision", "start_method", "use",
                                   "cor_method", "max_iter", "averaging",
                                   "trim", "salience_threshold", "seed"))
  expect_named(efa_raw_p$settings, c("estimator", "method", "rotation", "type", "n_factors", "N",
                                     "init_comm", "criterion", "criterion_type",
                                     "abs_eigen", "varimax_type", "normalize",
                                     "k_promax", "k_simplimax", "P_type",
                                     "precision", "start_method", "use",
                                     "cor_method", "max_iter", "averaging",
                                     "trim", "salience_threshold", "seed"))


  # the seed is part of the record, so a saved object says how its grid was seeded
  expect_null(efa_def$settings$seed)
  expect_equal(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = c("PAF", "ML"), rotation = "none", init_comm = "smc",
                criterion_type = "sum", start_method = "psych", show_progress = FALSE,
                seed = 11)$settings$seed,
    11)

  expect_equal(efa_def$settings$estimator, "PAF")
  expect_equal(efa_ml$settings$estimator, "ML")
  expect_equal(efa_uls$settings$estimator, "ULS")
  expect_equal(efa_all$settings$estimator, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_oblq$settings$estimator, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_orth$settings$estimator, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_none$settings$estimator, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_md$settings$estimator, c("PAF", "ML", "ULS"))
  expect_equal(efa_all_tm$settings$estimator, c("PAF", "ML", "ULS"))
  expect_equal(efa_raw$settings$estimator, "PAF")
  expect_equal(efa_raw_p$settings$estimator, "PAF")

  expect_equal(efa_def$settings$rotation, "promax")
  expect_equal(efa_ml$settings$rotation, "promax")
  expect_equal(efa_uls$settings$rotation, "promax")
  expect_equal(efa_all$settings$rotation, "promax")
  expect_equal(efa_all_oblq$settings$rotation, "oblique")
  expect_equal(efa_all_orth$settings$rotation, "orthogonal")
  expect_equal(efa_all_none$settings$rotation, "none")
  expect_equal(efa_all_md$settings$rotation, "oblique")
  expect_equal(efa_all_tm$settings$rotation, "oblique")
  expect_equal(efa_raw$settings$rotation, "none")
  expect_equal(efa_raw_p$settings$rotation, "promax")

  expect_equal(efa_def$settings$type, "none")
  expect_equal(efa_ml$settings$type, "none")
  expect_equal(efa_uls$settings$type, "none")
  expect_equal(efa_all$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_oblq$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_orth$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_none$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_md$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_all_tm$settings$type, c("none", "EFAtools", "psych", "SPSS"))
  expect_equal(efa_raw$settings$type, "none")
  expect_equal(efa_raw_p$settings$type, "none")

  expect_equal(efa_def$settings$n_factors, 3)
  expect_equal(efa_ml$settings$n_factors, 3)
  expect_equal(efa_uls$settings$n_factors, 3)
  expect_equal(efa_all$settings$n_factors, 3)
  expect_equal(efa_all_oblq$settings$n_factors, 3)
  expect_equal(efa_all_orth$settings$n_factors, 3)
  expect_equal(efa_all_none$settings$n_factors, 3)
  expect_equal(efa_all_md$settings$n_factors, 3)
  expect_equal(efa_all_tm$settings$n_factors, 3)
  expect_equal(efa_raw$settings$n_factors, 1)
  expect_equal(efa_raw_p$settings$n_factors, 2)

  expect_equal(efa_def$settings$N, 500)
  expect_equal(efa_ml$settings$N, 500)
  expect_equal(efa_uls$settings$N, 500)
  expect_equal(efa_all$settings$N, 500)
  expect_equal(efa_all_oblq$settings$N, 500)
  expect_equal(efa_all_orth$settings$N, 500)
  expect_equal(efa_all_none$settings$N, 500)
  expect_equal(efa_all_md$settings$N, 500)
  expect_equal(efa_all_tm$settings$N, 500)
  expect_equal(efa_raw$settings$N, 810)
  expect_equal(efa_raw_p$settings$N, 810)

  expect_equal(efa_def$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_ml$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_uls$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all$settings$init_comm, "smc")
  expect_equal(efa_all_oblq$settings$init_comm, "smc")
  expect_equal(efa_all_orth$settings$init_comm, "smc")
  expect_equal(efa_all_none$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_md$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_all_tm$settings$init_comm, "smc")
  expect_equal(efa_raw$settings$init_comm, c("smc", "mac", "unity"))
  expect_equal(efa_raw_p$settings$init_comm, c("smc", "mac", "unity"))

  expect_equal(efa_def$settings$criterion, 0.001)
  expect_equal(efa_ml$settings$criterion, 0.001)
  expect_equal(efa_uls$settings$criterion, 0.001)
  expect_equal(efa_all$settings$criterion, 0.001)
  expect_equal(efa_all_oblq$settings$criterion, 0.001)
  expect_equal(efa_all_orth$settings$criterion, 0.001)
  expect_equal(efa_all_none$settings$criterion, 0.001)
  expect_equal(efa_all_md$settings$criterion, 0.001)
  expect_equal(efa_all_tm$settings$criterion, 0.001)
  expect_equal(efa_raw$settings$criterion, 0.001)
  expect_equal(efa_raw_p$settings$criterion, 0.001)

  expect_equal(efa_def$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_ml$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_uls$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all$settings$criterion_type, "sum")
  expect_equal(efa_all_oblq$settings$criterion_type, "sum")
  expect_equal(efa_all_orth$settings$criterion_type, "sum")
  expect_equal(efa_all_none$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_md$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_all_tm$settings$criterion_type, "sum")
  expect_equal(efa_raw$settings$criterion_type, c("sum", "max_individual"))
  expect_equal(efa_raw_p$settings$criterion_type, c("sum", "max_individual"))

  expect_equal(efa_def$settings$abs_eigen, TRUE)
  expect_equal(efa_ml$settings$abs_eigen, TRUE)
  expect_equal(efa_uls$settings$abs_eigen, TRUE)
  expect_equal(efa_all$settings$abs_eigen, TRUE)
  expect_equal(efa_all_oblq$settings$abs_eigen, TRUE)
  expect_equal(efa_all_orth$settings$abs_eigen, TRUE)
  expect_equal(efa_all_none$settings$abs_eigen, TRUE)
  expect_equal(efa_all_md$settings$abs_eigen, TRUE)
  expect_equal(efa_all_tm$settings$abs_eigen, TRUE)
  expect_equal(efa_raw$settings$abs_eigen, TRUE)
  expect_equal(efa_raw_p$settings$abs_eigen, TRUE)

  expect_equal(efa_def$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_ml$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_uls$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all$settings$varimax_type, "svd")
  expect_equal(efa_all_oblq$settings$varimax_type, "svd")
  expect_equal(efa_all_orth$settings$varimax_type, "svd")
  expect_equal(efa_all_none$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_md$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_all_tm$settings$varimax_type, "svd")
  expect_equal(efa_raw$settings$varimax_type, c("svd", "kaiser"))
  expect_equal(efa_raw_p$settings$varimax_type, c("svd", "kaiser"))

  expect_equal(efa_def$settings$normalize, TRUE)
  expect_equal(efa_ml$settings$normalize, TRUE)
  expect_equal(efa_uls$settings$normalize, TRUE)
  expect_equal(efa_all$settings$normalize, TRUE)
  expect_equal(efa_all_oblq$settings$normalize, TRUE)
  expect_equal(efa_all_orth$settings$normalize, TRUE)
  expect_equal(efa_all_none$settings$normalize, TRUE)
  expect_equal(efa_all_md$settings$normalize, TRUE)
  expect_equal(efa_all_tm$settings$normalize, TRUE)
  expect_equal(efa_raw$settings$normalize, TRUE)
  expect_equal(efa_raw_p$settings$normalize, TRUE)

  expect_equal(efa_def$settings$k_promax, 2:4)
  expect_equal(efa_ml$settings$k_promax, 2:4)
  expect_equal(efa_uls$settings$k_promax, 2:4)
  expect_equal(efa_all$settings$k_promax, 4)
  expect_equal(efa_all_oblq$settings$k_promax, 4)
  expect_equal(efa_all_orth$settings$k_promax, 4)
  expect_equal(efa_all_none$settings$k_promax, 2:4)
  expect_equal(efa_all_md$settings$k_promax, 2:4)
  expect_equal(efa_all_tm$settings$k_promax, 4)
  expect_equal(efa_raw$settings$k_promax, 2:4)
  expect_equal(efa_raw_p$settings$k_promax, 2:4)

  expect_equal(efa_def$settings$k_simplimax, 18)
  expect_equal(efa_ml$settings$k_simplimax, 18)
  expect_equal(efa_uls$settings$k_simplimax, 18)
  expect_equal(efa_all$settings$k_simplimax, 18)
  expect_equal(efa_all_oblq$settings$k_simplimax, 18)
  expect_equal(efa_all_orth$settings$k_simplimax, 18)
  expect_equal(efa_all_none$settings$k_simplimax, 18)
  expect_equal(efa_all_md$settings$k_simplimax, 18)
  expect_equal(efa_all_tm$settings$k_simplimax, 18)
  expect_equal(efa_raw$settings$k_simplimax, 8)
  expect_equal(efa_raw_p$settings$k_simplimax, 8)

  expect_equal(efa_def$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_ml$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_uls$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all$settings$P_type, "norm")
  expect_equal(efa_all_oblq$settings$P_type, "norm")
  expect_equal(efa_all_orth$settings$P_type, "norm")
  expect_equal(efa_all_none$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_md$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_all_tm$settings$P_type, "norm")
  expect_equal(efa_raw$settings$P_type, c("norm", "unnorm"))
  expect_equal(efa_raw_p$settings$P_type, c("norm", "unnorm"))

  expect_equal(efa_def$settings$precision, 1e-5)
  expect_equal(efa_ml$settings$precision, 1e-5)
  expect_equal(efa_uls$settings$precision, 1e-5)
  expect_equal(efa_all$settings$precision, 1e-5)
  expect_equal(efa_all_oblq$settings$precision, 1e-5)
  expect_equal(efa_all_orth$settings$precision, 1e-5)
  expect_equal(efa_all_none$settings$precision, 1e-5)
  expect_equal(efa_all_md$settings$precision, 1e-5)
  expect_equal(efa_all_tm$settings$precision, 1e-5)
  expect_equal(efa_raw$settings$precision, 1e-5)
  expect_equal(efa_raw_p$settings$precision, 1e-5)

  expect_equal(efa_def$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_ml$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_uls$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all$settings$start_method, "psych")
  expect_equal(efa_all_oblq$settings$start_method, "psych")
  expect_equal(efa_all_orth$settings$start_method, "psych")
  expect_equal(efa_all_none$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_md$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_all_tm$settings$start_method, "psych")
  expect_equal(efa_raw$settings$start_method, c("psych", "factanal"))
  expect_equal(efa_raw_p$settings$start_method, c("psych", "factanal"))

  expect_equal(efa_def$settings$use, "pairwise.complete.obs")
  expect_equal(efa_ml$settings$use, "pairwise.complete.obs")
  expect_equal(efa_uls$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_oblq$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_orth$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_none$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_md$settings$use, "pairwise.complete.obs")
  expect_equal(efa_all_tm$settings$use, "pairwise.complete.obs")
  expect_equal(efa_raw$settings$use, "pairwise.complete.obs")
  expect_equal(efa_raw_p$settings$use, "pairwise.complete.obs")

  expect_equal(efa_def$settings$cor_method, "pearson")
  expect_equal(efa_ml$settings$cor_method, "pearson")
  expect_equal(efa_uls$settings$cor_method, "pearson")
  expect_equal(efa_all$settings$cor_method, "pearson")
  expect_equal(efa_all_oblq$settings$cor_method, "pearson")
  expect_equal(efa_all_orth$settings$cor_method, "pearson")
  expect_equal(efa_all_none$settings$cor_method, "pearson")
  expect_equal(efa_all_md$settings$cor_method, "pearson")
  expect_equal(efa_all_tm$settings$cor_method, "pearson")
  expect_equal(efa_raw$settings$cor_method, "pearson")
  expect_equal(efa_raw_p$settings$cor_method, "pearson")

  expect_equal(efa_def$settings$max_iter, 10000)
  expect_equal(efa_ml$settings$max_iter, 10000)
  expect_equal(efa_uls$settings$max_iter, 10000)
  expect_equal(efa_all$settings$max_iter, 10000)
  expect_equal(efa_all_oblq$settings$max_iter, 10000)
  expect_equal(efa_all_orth$settings$max_iter, 10000)
  expect_equal(efa_all_none$settings$max_iter, 10000)
  expect_equal(efa_all_md$settings$max_iter, 10000)
  expect_equal(efa_all_tm$settings$max_iter, 10000)
  expect_equal(efa_raw$settings$max_iter, 10000)
  expect_equal(efa_raw_p$settings$max_iter, 10000)

  expect_equal(efa_def$settings$averaging, "mean")
  expect_equal(efa_ml$settings$averaging, "mean")
  expect_equal(efa_uls$settings$averaging, "mean")
  expect_equal(efa_all$settings$averaging, "mean")
  expect_equal(efa_all_oblq$settings$averaging, "mean")
  expect_equal(efa_all_orth$settings$averaging, "mean")
  expect_equal(efa_all_none$settings$averaging, "mean")
  expect_equal(efa_all_md$settings$averaging, "median")
  expect_equal(efa_all_tm$settings$averaging, "mean")
  expect_equal(efa_raw$settings$averaging, "mean")
  expect_equal(efa_raw_p$settings$averaging, "mean")

  expect_equal(efa_def$settings$trim, 0)
  expect_equal(efa_ml$settings$trim, 0)
  expect_equal(efa_uls$settings$trim, 0)
  expect_equal(efa_all$settings$trim, 0)
  expect_equal(efa_all_oblq$settings$trim, 0)
  expect_equal(efa_all_orth$settings$trim, 0)
  expect_equal(efa_all_none$settings$trim, 0)
  expect_equal(efa_all_md$settings$trim, 0)
  expect_equal(efa_all_tm$settings$trim, 0.2)
  expect_equal(efa_raw$settings$trim, 0)
  expect_equal(efa_raw_p$settings$trim, 0)

  expect_equal(efa_def$settings$salience_threshold, 0.3)
  expect_equal(efa_ml$settings$salience_threshold, 0.3)
  expect_equal(efa_uls$settings$salience_threshold, 0.3)
  expect_equal(efa_all$settings$salience_threshold, 0.2)
  expect_equal(efa_all_oblq$settings$salience_threshold, 0.3)
  expect_equal(efa_all_orth$settings$salience_threshold, 0.3)
  expect_equal(efa_all_none$settings$salience_threshold, 0.3)
  expect_equal(efa_all_md$settings$salience_threshold, 0.3)
  expect_equal(efa_all_tm$settings$salience_threshold, 0.3)
  expect_equal(efa_raw$settings$salience_threshold, 0.3)
  expect_equal(efa_raw_p$settings$salience_threshold, 0.3)


})


test_that("errors are thrown correctly", {
  skip_if_not_slow()
  expect_error(efa_average(1:5, show_progress = FALSE), class = "efa_input_not_matrix")
  # Extracting two factors from the unidimensional GRiPS overshoots, and psych's
  # 50-iteration cap now stops that solution short of convergence, so it is
  # excluded with a warning; suppress that incidental warning to assert the target
  # condition (as the n_factors = 1 cases below already do).
  expect_message(suppressWarnings(
    efa_average(GRiPS_raw, n_factors = 2, estimator = "PAF", type = c("EFAtools", "psych"), show_progress = FALSE),
    classes = "efa_avg_excluded_solutions"),
                 class = "efa_cor_from_data")
  expect_warning(suppressWarnings(
    efa_average(GRiPS_raw, n_factors = 2, estimator = "PAF", type = c("EFAtools", "psych"),
                N = 20, show_progress = FALSE),
    classes = "efa_avg_excluded_solutions"),
                 class = "efa_n_from_data")
  expect_error(efa_average(sing_raw, n_factors = 1, show_progress = FALSE),
               class = "efa_cor_singular")
  expect_error(efa_average(sing_cor, N = sing_N, n_factors = 1, show_progress = FALSE),
               class = "efa_cor_singular")
  expect_error(efa_average(matrix(rnorm(30), ncol = 3), n_factors = 2, show_progress = FALSE),
               class = "efa_underidentified")
  expect_warning(suppressWarnings(efa_average(matrix(rnorm(30), ncol = 3), n_factors = 1,
                             estimator = "PAF", type = c("EFAtools", "psych"), show_progress = FALSE),
                             class = c("efa_avg_excluded_solutions")),
                 class = "efa_just_identified")
  expect_warning(suppressWarnings(efa_average(cor_nposdef, n_factors = 1, N = 10, estimator = "PAF",
                     type = c("EFAtools", "psych"), show_progress = FALSE), classes = c("efa_just_identified", "efa_avg_excluded_solutions")), class = "efa_cor_smoothed")
  expect_message(efa_average(GRiPS_raw, n_factors = 1, estimator = "PAF", type = c("EFAtools", "psych"), show_progress = FALSE),
                 class = "efa_avg_single_factor_rotation")
  expect_warning(efa_average(GRiPS_raw, n_factors = 1, estimator = "PAF", type = c("EFAtools"),
                             rotation = "none", show_progress = FALSE),
                 class = "efa_avg_single_combination")
})

test_that("a named type brings its tuning into the grid, but not its factor ordering", {
  # The solutions are realigned to a common target before averaging, so a per-fit factor
  # order would not survive into the averaged result; every row is therefore fitted
  # eigenvalue-ordered even under a named `type`. That is user-visible, because the
  # individual fits are returned, so it is documented and pinned here.
  avg <- suppressWarnings(efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                                      estimator = c("PAF", "ML"), rotation = "promax",
                                      type = "SPSS", show_progress = FALSE))
  order_types <- vapply(avg$efa_list, function(f) f$settings$order_type, character(1))
  expect_true(all(order_types == "eigen"))

  # ... and on the single-row path, where that fit *is* the return value, the recorded
  # settings are the concrete arguments it ran under rather than the type that supplied
  # them: the SPSS iteration cap, but type "none" and the grid's eigenvalue ordering.
  single <- suppressWarnings(efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                                         estimator = "PAF", rotation = "promax",
                                         type = "SPSS", show_progress = FALSE))
  expect_s3_class(single, "efa")
  expect_identical(single$settings$type, "none")
  expect_identical(single$settings$order_type, "eigen")
  expect_equal(single$settings$max_iter, 25)
})

test_that("the single-combination shortcut still records the seed", {
  single <- function(seed) suppressMessages(suppressWarnings(
    efa_average(GRiPS_raw, n_factors = 1, estimator = "PAF", type = "EFAtools",
                rotation = "none", show_progress = FALSE, seed = seed)))

  # the shortcut returns an efa_fit() object rather than an efa_average one, but the
  # run was still governed by `seed`, so the record must say so
  expect_equal(single(11)$settings$seed, 11)
  # and the field stays in the schema when no seed was supplied
  s <- single(NULL)$settings
  expect_true("seed" %in% names(s))
  expect_null(s$seed)
})

test_that("the grid progress bar declares exactly the steps it signals", {
  # progressr is inert in a non-interactive session and every other test in this file
  # runs with show_progress = FALSE, so this is the only place the progress accounting
  # is exercised. Sending more updates than the progressor declares makes
  # with_progress() warn that it is "no longer listening to this progressor"; the
  # warning then also styles the returned object's auto-printed output as warning
  # output. Small grids are the exposed case: they report on every EFA, so the per-EFA
  # updates alone complete the progressor.
  withr::local_options(progressr.enable = TRUE)

  # 3 EFAs (one per estimator at the EFAtools preset), the grid from @examples. Grids
  # of ten or fewer report on every EFA, so stepsize is 1.
  expect_no_warning(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = c("PAF", "ULS", "ML"), type = "EFAtools",
                start_method = "psych"))

  # 24 unrotated PAF EFAs, enough for stepsize to exceed 1 (round(24/10) = 2). That is
  # the branch in which the declared steps stop being a whole number of updates, so it
  # needs its own case; the cheap unrotated PAF grid keeps it fast.
  expect_no_warning(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = "PAF", type = "none", rotation = "none",
                init_comm = c("smc", "mac", "unity"),
                criterion = c(1e-3, 1e-4),
                criterion_type = c("sum", "max_individual"),
                abs_eigen = c(TRUE, FALSE)))
})

test_that("the Model Fit block reports the effective n per index class", {
  # The print snapshot masks these counts (they follow the per-solution convergence and
  # Heywood outcomes, which move across BLAS implementations), so the wiring - which grid
  # column feeds which printed line - is pinned here instead.
  local_reproducible_output()
  set.seed(42)

  # PAF leaves the chi-square-based indices NA but still contributes CAF, RMSR, and SRMR,
  # so the two index classes are averaged over different numbers of solutions.
  mixed <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                       estimator = c("PAF", "ML"), type = "EFAtools",
                       rotation = "promax", start_method = "psych",
                       show_progress = FALSE)
  grid <- mixed$implementations_grid
  n_chisq <- sum(!is.na(grid$chisq))
  n_resid <- sum(!is.na(grid$caf))
  expect_gt(n_resid, n_chisq)

  out <- paste(cli::ansi_strip(format(mixed)), collapse = " ")
  expect_match(out, paste0("Chi-square-based indices averaged over ", n_chisq,
                           " of ", nrow(grid), " solution"), fixed = TRUE)
  expect_match(out, paste0("CAF, RMSR, and SRMR averaged over ", n_resid,
                           " of ", nrow(grid), " solution"), fixed = TRUE)

  # With PAF alone no chi-square-based index is reported at all, so that line is absent.
  paf <- suppressWarnings(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = "PAF", type = c("EFAtools", "psych"),
                rotation = "promax", show_progress = FALSE))
  out_paf <- paste(cli::ansi_strip(format(paf)), collapse = " ")
  expect_no_match(out_paf, "Chi-square-based indices", fixed = TRUE)
  expect_match(out_paf, paste0("CAF, RMSR, and SRMR averaged over ",
                               sum(!is.na(paf$implementations_grid$caf)), " of "),
               fixed = TRUE)
})

test_that("the Model Fit block names the aggregation it reports", {
  # The sentence introducing the block must agree with the M / Md header below it.
  local_reproducible_output()
  set.seed(42)
  args <- list(test_models$baseline$cormat, n_factors = 3, N = 500,
               estimator = c("PAF", "ML"), type = "EFAtools", rotation = "promax",
               start_method = "psych", show_progress = FALSE)

  mn <- paste(cli::ansi_strip(format(do.call(efa_average, args))), collapse = " ")
  expect_match(mn, "the mean of the per-solution fit indices", fixed = TRUE)

  md <- paste(cli::ansi_strip(format(do.call(efa_average, c(args, averaging = "median")))),
              collapse = " ")
  expect_match(md, "the median of the per-solution fit indices", fixed = TRUE)
  expect_no_match(md, "the average of the per-solution fit indices", fixed = TRUE)
})

test_that("a trim that the median cannot use is flagged", {
  # trim only reaches mean(), so a non-zero trim under averaging = "median" changes
  # nothing; it must not be accepted silently. The message is raised during argument
  # checking, before any EFA is fitted, so this needs no grid.
  expect_message(
    suppressWarnings(efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                                 estimator = "PAF", type = "EFAtools", rotation = "none",
                                 averaging = "median", trim = 0.2,
                                 show_progress = FALSE)),
    class = "efa_avg_trim_ignored"
  )
  # ... and not raised when the trim is actually used.
  expect_no_message(
    suppressWarnings(efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                                 estimator = "PAF", type = "EFAtools", rotation = "none",
                                 averaging = "mean", trim = 0.2,
                                 show_progress = FALSE)),
    class = "efa_avg_trim_ignored"
  )
})

test_that("an all-failed averaging grid returns an empty (NA) result", {
  skip_if_not_slow()
  # When every solution fails (here all runs hit max_iter and do not converge),
  # the averaged result is NA rather than an error or an average over an empty set.
  res <- suppressWarnings(efa_average(test_models$baseline$cormat, n_factors = 3,
                                      N = 500, max_iter = 1, estimator = "PAF",
                                      type = "none", rotation = "none",
                                      show_progress = FALSE))
  expect_s3_class(res, "efa_average")
  expect_true(all((res$implementations_grid$converged != 0) %in% TRUE))
  expect_true(all(is.na(res$h2)))
  expect_true(all(is.na(res$loadings)))
})

test_that("an all-Heywood averaging grid returns an empty (NA) result", {
  skip_if_not_slow()
  # A 3-variable matrix whose single-factor solution implies a communality > 1, so
  # every fitted solution is a Heywood case and is excluded; with nothing left to
  # average, the communalities and loadings come back as NA rather than an error.
  m <- matrix(c(1, .8, .8,  .8, 1, .5,  .8, .5, 1), 3, 3)
  res <- suppressWarnings(efa_average(m, n_factors = 1, N = 200, estimator = "PAF",
                                      type = "none", rotation = "none",
                                      show_progress = FALSE))
  expect_s3_class(res, "efa_average")
  expect_true(all(res$implementations_grid$heywood))
  expect_true(all(is.na(res$h2)))
  expect_true(all(is.na(res$loadings)))
})

test_that("print output is stable", {
  skip_if_not_slow()
  skip_on_cran()
  local_reproducible_output()

  # default view (oblique -> Phi section, average + range)
  expect_snapshot(print(efa_def, plot = FALSE), transform = scrub_num_pct)

  # all statistics, no Phi (orthogonal-free rotation = "none")
  expect_snapshot(print(efa_all_none, stat = c("average", "sd", "min", "max"),
                        plot = FALSE), transform = scrub_num_pct)

  # median averaging (Md / Median labels)
  expect_snapshot(print(efa_all_md, plot = FALSE), transform = scrub_num_pct)
})

test_that("format.efa_average is the source of truth", {
  skip_on_cran()
  skip_if_not_slow()
  local_reproducible_output()

  # print() is exactly cat(format(x), sep = "\n") (with plot = FALSE, the default, so
  # nothing is drawn), so the two agree line for line.
  expect_identical(utils::capture.output(print(efa_def)), format(efa_def))
})

test_that("plot returns a ggplot", {
  skip_on_cran()
  skip_if_not_slow()
  expect_s3_class(plot(efa_def), "ggplot")
})

test_that("the average-loading plot is visually stable", {
  skip_if_not_installed("vdiffr")

  # A literal fixture: plot.efa_average() reads only the average/min/max loading
  # matrices and the two settings below, so hand-chosen values keep the baseline
  # deterministic (the fitted fixtures above are grids of many EFAs).
  L <- matrix(c( 0.82, -0.11,
                 0.45,  0.60,
                -0.30,  0.71,
                 0.04, -0.05),
              nrow = 4, byrow = TRUE,
              dimnames = list(paste0("V", 1:4), c("F1", "F2")))
  avg <- structure(
    list(loadings = list(average = L, min = L - 0.08, max = L + 0.12),
         settings = list(averaging = "mean", salience_threshold = 0.2)),
    class = "efa_average")

  # the axis title, the caption naming the four marks, and the marks themselves
  vdiffr::expect_doppelganger("efa_average loading plot", plot(avg))
})

test_that("a vector-valued precision is recycled across the grid", {
  skip_if_not_slow()
  # A vector precision is expanded into the grid; each EFA must receive its own
  # scalar value rather than the whole vector (which would fail every fit and
  # return an all-NA result).
  res <- suppressWarnings(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = "ML", rotation = "promax", type = "none",
                precision = c(1e-5, 1e-3), show_progress = FALSE))
  expect_s3_class(res, "efa_average")
  expect_setequal(unique(res$implementations_grid$precision), c(1e-5, 1e-3))
  expect_false(any(res$implementations_grid$errors))
  expect_false(all(is.na(res$loadings$average)))
  expect_true(all(is.finite(res$loadings$average)))
})

test_that("problematic solutions are summarised in a single warning", {
  skip_if_not_slow()
  # All runs hit max_iter and are excluded, so one summary warning is raised
  # rather than one per model.
  expect_warning(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                max_iter = 1, estimator = "PAF", type = "none", rotation = "none",
                show_progress = FALSE),
    class = "efa_avg_excluded_solutions"
  )
})

# The rates summary reads only the implementations_grid, so the denominator states below can
# be exercised without fitting anything.
rates_text <- function(grid) {
  paste(cli::ansi_strip(cli::cli_format_method(.efa_emit_average_rates(grid))),
        collapse = " ")
}

test_that("each rate names the denominator it is computed over", {
  local_reproducible_output()
  # Convergence is assessed for every solution that did not error, but Heywood cases and
  # admissibility only for those that then converged. Here: 2 errored, 2 did not converge,
  # and 6 converged (3 Heywood, 1 proper but not salient, 2 admissible). The Heywood and
  # admissibility denominators are therefore 6, not the 8 that did not error - over 8 the
  # Heywood rate would read 38% rather than 50%.
  grid <- data.frame(
    errors     = c(rep(TRUE, 2), rep(FALSE, 8)),
    converged  = c(rep(NA, 2), rep(1, 2), rep(0, 6)),
    heywood    = c(rep(NA, 4), rep(TRUE, 3), rep(FALSE, 3)),
    admissible = c(rep(NA, 4), rep(FALSE, 4), rep(TRUE, 2))
  )

  out <- rates_text(grid)
  expect_match(out, "The error rate is at 20%.", fixed = TRUE)
  expect_match(out, "Of the solutions that did not result in an error, 75% converged.",
               fixed = TRUE)
  expect_match(out,
               "Of the solutions that converged, 50% contained Heywood cases and 33% were admissible.",
               fixed = TRUE)
})

test_that("Heywood cases and admissibility are not reported when no solution converged", {
  local_reproducible_output()
  # Nothing reached the stage at which either is assessed, so both rates are undefined
  # rather than 0%.
  grid <- data.frame(errors = FALSE, converged = c(1, 1), heywood = NA, admissible = NA)

  out <- rates_text(grid)
  expect_no_match(out, "NaN", fixed = TRUE)
  expect_match(out, "Of the solutions that did not result in an error, 0% converged.",
               fixed = TRUE)
  expect_match(out,
               "No solution converged, so Heywood cases and admissibility could not be assessed.",
               fixed = TRUE)
})

test_that("only the error rate is reported when every EFA errored", {
  local_reproducible_output()
  # No model was fitted at all, so convergence is undefined too.
  grid <- data.frame(errors = TRUE, converged = c(NA, NA), heywood = NA, admissible = NA)

  out <- rates_text(grid)
  expect_no_match(out, "NaN", fixed = TRUE)
  expect_match(out, "The error rate is at 100%.", fixed = TRUE)
  expect_match(out,
               "No solution could be fitted, so convergence, Heywood cases, and admissibility could not be assessed.",
               fixed = TRUE)
})

test_that("printing a run in which no EFA converged reports no NaN rate", {
  skip_if_not_slow()
  local_reproducible_output()
  # Every run hits max_iter, so the Heywood and admissibility rates have an empty
  # denominator on the way through the public API.
  res <- suppressWarnings(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                max_iter = 1, estimator = "PAF", type = "none", rotation = "none",
                show_progress = FALSE))

  expect_no_match(paste(cli::ansi_strip(format(res)), collapse = " "), "NaN", fixed = TRUE)
})

test_that("printing does not plot by default", {
  expect_false(formals(getS3method("print", "efa_average"))$plot)
})

test_that("admissibility is reported as an outcome, not a varied setting", {
  skip_on_cran()
  skip_if_not_slow()
  local_reproducible_output()
  obj <- efa_def
  # Force mixed admissibility so the column would surface in the varied-settings
  # list if it were (incorrectly) treated as a setting.
  obj$implementations_grid$admissible[1] <- !obj$implementations_grid$admissible[1]
  out <- paste(cli::ansi_strip(utils::capture.output(print(obj, plot = FALSE))),
               collapse = " ")
  # "admissible" appears only in the convergence summary, never in the settings list.
  expect_equal(unname(lengths(gregexpr("admissible", out, fixed = TRUE))), 1L)
})

test_that("settings varied equally across the grid are still listed", {
  local_reproducible_output()
  # A grid whose settings columns each hold the same number of distinct values must
  # still report every varied setting. Every model is a Heywood case, so the report
  # stops after the summary and no averaged solution is needed.
  obj <- structure(list(
    settings = list(estimator = c("PAF", "ML"), rotation = c("promax", "oblimin"),
                    averaging = "mean", trim = 0),
    implementations_grid = data.frame(
      estimator  = c("PAF", "ML"),
      rotation   = c("promax", "oblimin"),
      errors     = FALSE,
      converged  = 0,
      heywood    = TRUE,
      admissible = FALSE
    )), class = "efa_average")

  out <- paste(cli::ansi_strip(format(obj)), collapse = " ")
  expect_match(out, "settings: estimator and rotation")
})

test_that("averaged fit indices match the per-model grid means", {
  skip_if_not_slow()
  # efa_def uses the untrimmed mean, so each averaged index equals the column
  # mean over the included models (and the new residual indices are averaged too).
  fi <- efa_def$fit_indices
  grid <- efa_def$implementations_grid
  expect_equal(fi$average[fi$index == "caf"], mean(grid$caf, na.rm = TRUE))
  expect_equal(fi$average[fi$index == "srmr"], mean(grid$srmr, na.rm = TRUE))

  # df is constant across the averaged solutions, so its dispersion is zero: the
  # sd and range cells of the df row are 0 (average/min/max hold df itself).
  expect_equal(fi$sd[fi$index == "df"], 0)
  expect_equal(fi$range[fi$index == "df"], 0)
  expect_equal(fi$average[fi$index == "df"], fi$min[fi$index == "df"])
})

if (is_slow_test()) {
  rm(efa_def, efa_ml, efa_uls, efa_all, efa_all_oblq, efa_all_orth, efa_all_none,
     efa_all_md, efa_all_tm, efa_raw, efa_raw_p)
}

# Corrected two-stage standard errors and the corrected (Satorra-Bentler) two-stage chi-square for
# cor_method = "fiml". The factor model is fitted to the EM-estimated saturated correlation R
# (Stage 1), so the naive Stage-2 standard errors -- which treat R as if it were complete raw data
# -- are inconsistent under missingness. The corrected covariance is the two-stage sandwich
#   Omega_tilde = (D' H D)^- D' H Omega_delta H D (D' H D)^-,
# with D = d sigma_offdiag / d vec(Lambda) the model derivative, H the Stage-2 weight at the
# solution (the normal-theory ML weight for ML, the identity for ULS), and Omega_delta the
# correlation-scale asymptotic covariance of the saturated FIML estimates
# (.fiml_saturated_acov()$cor). Likewise the plain two-stage
# likelihood-ratio chi-square is not asymptotically chi-square(df) under the two-stage estimator;
# it is rescaled by the Satorra-Bentler correction built from the same H and Omega_delta.
# Two-stage standard errors / rescaled statistic: Yuan & Bentler (2000); Savalei & Bentler (2009,
# SEM); Yuan, Marshall & Bentler (2002, Psychometrika 67:95-121). MAR/ignorability: Little & Rubin
# (2002). The construction shares its geometry with the ordinal/continuous robust sandwich -- both
# take D, H, the bordered bread and the gauge constraint from `.offdiag_sandwich_pieces()` -- and
# differs only in the meat, which here is the FIML saturated covariance.

# Robust two-stage loading covariance V_AA (p*k x p*k), the unrotated loading/uniqueness SEs, and
# the reliability flags, from the fitted loadings and the FIML moments. The model Jacobian, the
# Stage-2 weight (the normal-theory ML weight for ML, the identity for ULS), the bordered bread and
# the gauge constraint come from `.offdiag_sandwich_pieces()`, shared with the polychoric/ADF and
# expected-information sandwiches; only the meat differs. Returns NA SEs (reliable = FALSE) at a
# singular bordered system, a degenerate saturated covariance, or a non-PSD V_AA.
.se_fiml_core <- function(fit_out, fiml, N, method) {

  L <- unclass(fit_out$unrot_loadings)
  p <- nrow(L)
  k <- ncol(L)
  pk <- p * k

  # `gauge_reliable = TRUE` on the NA path, as on the other analytic paths: everything is already
  # withheld and the gauge is not what withheld it, so the caller's gauge-specific message must not
  # be attached to this branch.
  na_core <- list(
    V_AA = matrix(NA_real_, pk, pk),
    loadings_se = matrix(NA_real_, p, k),
    uniquenesses_se = rep(NA_real_, p),
    reliable = FALSE,
    gauge_reliable = TRUE
  )

  pieces <- .offdiag_sandwich_pieces(L, method)
  if (is.null(pieces)) return(na_core)

  # Stage-1 saturated correlation asymptotic covariance Omega_delta (the meat), on the variance
  # scale Var(r-hat) and in utils::combn() order. Reuse the cached `acov_cor` when EFA() already
  # built it for the scaled chi-square (the point-estimate analytic-SE path); else build it here.
  Omega <- fiml$acov_cor
  if (is.null(Omega)) {
    Omega <- tryCatch(.fiml_saturated_acov(fiml$data, fiml$mu, fiml$sigma)$cor,
                      error = function(e) NULL)
  }
  if (is.null(Omega) || anyNA(Omega) || nrow(Omega) != p * (p - 1L) / 2L) return(na_core)

  # Two-stage sandwich V = A^- (V D)' Omega_delta (V D) A^- / (N - 1). Omega_delta is the covariance
  # of the saturated estimates (the inverse observed information, so on the variance scale
  # Var(r-hat)); it is put on the unit asymptotic-variance scale by N and the covariance divided by
  # N - 1, exactly as the polychoric/ADF sandwich and the expected information do. That N - 1 is the
  # package-wide small-sample convention (it is also the scale `.gof()` and `.fiml_scaled_test()`
  # work on), and it makes the complete-data limit of this sandwich agree with the Pearson path
  # instead of differing from it by a systematic factor of N / (N - 1).
  VD <- pieces$VD
  Abread <- pieces$Abread
  meat <- crossprod(VD, (N * Omega) %*% VD)
  V_AA <- Abread %*% meat %*% Abread / (N - 1)
  V_AA <- (V_AA + t(V_AA)) / 2

  # The covariance must be positive semidefinite; a non-finite entry or negative eigenvalue would
  # otherwise corrupt the marginal, uniqueness, and rotated SEs that read the full V_AA.
  reliable <- .is_psd(V_AA)
  loadings_se <- if (reliable) sqrt(diag(V_AA)) else rep(NA_real_, pk)
  # Uniqueness SE = communality SE (psi_i = 1 - rowSums(Lambda^2)_i), via the shared gradient.
  uniq_se <- if (!reliable) rep(NA_real_, p) else .communality_se(L, V_AA)

  list(
    V_AA = V_AA,
    loadings_se = matrix(loadings_se, p, k),
    uniquenesses_se = uniq_se,
    reliable = reliable,
    # Conditioning of the rotational gauge the unrotated loadings are reported in, from the shared
    # geometry. Reported here so a degenerate orientation is caught on this path exactly as on the
    # Pearson/polychoric one, rather than shipping an inflated unrotated loading SE as informative.
    gauge_reliable = pieces$gauge_reliable
  )
}

# Corrected two-stage SE/CI for cor_method = "fiml" (se = "information" / "sandwich" both route
# here -- the naive Stage-2 information SE is never shipped). Fills the standard analytic SE/CI
# schema. The unrotated branch reuses the robust unrotated wrapper; the rotated branch reuses the
# information-SE rotation propagation by supplying the robust covariance through its `se0` contract
# (so promax/simplimax stay bootstrap-only, exactly as elsewhere). NA-fills with the classed
# efa_se_unreliable warning at a Heywood case / non-PSD covariance.
.se_fiml <- function(fit_out, rot_info, N, ci, fiml, method) {

  core <- .se_fiml_core(fit_out, fiml, N, method)

  # The two-stage sandwich withholds at a boundary solution on the same test, and through the same
  # shared policy, as every other analytic path. The corrected chi-square is untouched: it is built
  # in `.gof()` rather than here, and is a discrepancy-function quantity the boundary does not
  # invalidate.
  core <- .withhold_at_boundary(core, fit_out$unrot_loadings)

  if (is.null(rot_info)) {
    .se_sandwich_unrotated(fit_out, core, ci)
  } else {
    se0 <- list(vcov = core$V_AA,
                loadings_se = core$loadings_se,
                uniquenesses_se = core$uniquenesses_se,
                gauge_reliable = core$gauge_reliable)
    .se_information_rotated(fit_out, rot_info, N, ci, se0 = se0)
  }
}

# Satorra-Bentler-corrected two-stage chi-square block for cor_method = "fiml", in the schema
# .apply_scaled_test() consumes. The factor model is fitted to the saturated EM correlation R, so
# the corrected statistic is the residual-based two-stage statistic on the off-diagonal correlation
# structure: the normal-theory ML quadratic form rescaled by the saturated FIML covariance
# Omega_delta (the meat) -- exactly the .scaled_chisq() machinery the robust (sandwich) path uses,
# with the FIML covariance in place of the polychoric/ADF Gamma. Returns NULL for PAF (no
# discrepancy), a just/under-identified model, a missing N, or a degenerate/non-PD saturated
# covariance, where the plain likelihood-ratio / NA fallback is reported instead.
.fiml_scaled_test <- function(L, R, N, method, df, m, fiml) {

  if (!(method %in% c("ML", "ULS")) || is.na(N) ||
      is.null(df) || is.na(df) || df <= 0) {
    return(NULL)
  }

  L <- unclass(L)
  p <- nrow(L)

  # The rescaled statistic is gauge-invariant, so the gauge diagnostic is not read here; skip it.
  # This runs once per bootstrap replicate, where the discarded decomposition would be pure cost.
  pieces <- .offdiag_sandwich_pieces(L, method, gauge = FALSE)
  if (is.null(pieces)) return(NULL)

  # Reuse the cached saturated correlation covariance when EFA() pre-built it for the SE sandwich
  # (the point-estimate analytic-SE path); else build it here (every bootstrap replicate, and the
  # se = "none"/"np-boot" point estimate, omit the cache and land in this branch).
  Omega <- fiml$acov_cor
  if (is.null(Omega)) {
    Omega <- tryCatch(.fiml_saturated_acov(fiml$data, fiml$mu, fiml$sigma)$cor,
                      error = function(e) NULL)
  }
  if (is.null(Omega) || anyNA(Omega) || nrow(Omega) != p * (p - 1L) / 2L) return(NULL)

  # .scaled_chisq() takes Gamma on the unit asymptotic-variance scale (N * Var(rho-hat); = lavaan's
  # NACOV), so scale the variance-scale FIML covariance by N, exactly as the polychoric/ADF path
  # does. The minimal fit_out shape supplies the loadings, the analysis correlation, and df.
  fo <- list(unrot_loadings = L, orig_R = R, fit_indices = list(df = df))
  .scaled_chisq(fo, N * Omega, pieces$pairs, pieces$VD, pieces$vdiag, pieces$Vmat,
                pieces$Abread, N)
}

#' Print and format an efa_simulated object
#'
#' `print()` shows a compact summary of the data simulated by [efa_simulate()]: how many
#' datasets were drawn and their dimensions, the marginal distribution, whether the data were
#' discretized into ordered categories or given missing values, and -- when model error was
#' injected -- the method with the target and achieved RMSEA and CFI. The simulated data
#' themselves live in the `data` element and the population correlation matrix in `population`.
#' `format()` returns the same summary as a character vector; `print()` is
#' `cat(format(x), sep = "\n")`. The lines follow the active console theme, so they are plain
#' when colours are disabled.
#'
#' @param x An object of class `efa_simulated` (output from [efa_simulate()]).
#' @param digits Integer. The number of decimal places the reported fit values are rounded to.
#'   Default is 3.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a character vector
#'   with the summary lines.
#'
#' @family data simulation
#'
#' @export
#'
#' @method print efa_simulated
#'
#' @examples
#' Lambda <- population_models$loadings$baseline
#' Phi <- population_models$phis_3$moderate
#' efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, target_rmsea = 0.05, seed = 42)
#'
print.efa_simulated <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_simulated
#' @export
#' @method format efa_simulated
format.efa_simulated <- function(x, digits = 3, ...) {

  set <- x$settings
  p <- ncol(x$population)

  cli::cli_format_method({

    cli::cli_rule(left = "{.strong Simulated data}")
    cli::cli_text("")

    if (is.null(x$data)) {
      cli::cli_text("{cli::qty(p)}Population correlation matrix only ({p} variable{?s}); no data drawn.")
    } else {
      one <- if (is.list(x$data)) x$data[[1L]] else x$data
      nd <- set$n_datasets
      n <- nrow(one)
      kind <- if (is.integer(one)) "ordered-category" else "continuous"
      if (nd > 1L) {
        cli::cli_text("{cli::qty(nd)}{nd} dataset{?s} of {cli::qty(n)}{n} case{?s} by {cli::qty(p)}{p} variable{?s} ({kind}).")
      } else {
        cli::cli_text("{cli::qty(n)}{n} case{?s} by {cli::qty(p)}{p} variable{?s} ({kind}).")
      }
    }

    marg <- switch(set$marginals,
                   normal = "normal",
                   empirical = "empirical",
                   VM = "Vale-Maurelli non-normal",
                   IG = "independent-generator non-normal")
    cli::cli_text("Marginals: {marg}.")
    if (!is.null(set$categories)) {
      cli::cli_text("Discretized into ordered categories.")
    }
    if (!is.null(set$missing) && set$missing != "none") {
      cli::cli_text("Missing data: {set$missing}.")
    }

    cli::cli_text("")
    me <- x$model_error
    if (is.null(me)) {
      cli::cli_text("Model error: none (exact population).")
    } else {
      cli::cli_text("Model error: {me$method}")
      tr <- if (is.null(me$target_rmsea)) "--" else .efa_num(me$target_rmsea, digits, pad = FALSE)
      tc <- if (is.null(me$target_cfi)) "--" else .efa_num(me$target_cfi, digits, pad = FALSE)
      ar <- .efa_num(me$rmsea, digits, pad = FALSE)
      ac <- .efa_num(me$cfi, digits, pad = FALSE)
      cli::cli_text("  RMSEA: target {tr}, achieved {ar}")
      cli::cli_text("  CFI:   target {tc}, achieved {ac}")
    }

  })
}

# Percent of uncontaminated correlations (PUC): the proportion of item pairs that
# do not share a group factor (Reise et al., 2013; Bonifay et al., 2015). `grp_ind`
# is a logical items-by-group-factors membership matrix (the general factor is
# excluded, as it contaminates no correlation). Each item pair sharing at least one
# group factor is a single contaminated correlation, regardless of how many group
# factors it shares.
.puc <- function(grp_ind) {
  grp_ind <- as.matrix(grp_ind)
  shares_grp <- (grp_ind %*% t(grp_ind)) > 0
  n_items <- nrow(grp_ind)
  cont_corrs <- sum(shares_grp[upper.tri(shares_grp)])
  1 - cont_corrs / (n_items * (n_items - 1) / 2)
}

# H index over a construct's own loadings: the correlation between an optimally-weighted
# composite and the factor, 1 / (1 + 1 / sum(lambda^2 / (1 - lambda^2))) (Hancock & Mueller,
# 2001). A loading with |lambda| >= 1 makes the error term 1 - lambda^2 <= 0 and the summand
# undefined, so H is not interpretable for that construct; return NA -- as for a construct
# with no indicators -- rather than a negative or infinite value. For standardized loadings
# this only happens in an improper (Heywood) solution, but an oblique pattern coefficient can
# exceed 1 in a proper solution; callers flag genuine Heywood cases separately, via the unique
# variances. A missing loading (NA) is excluded from the >= 1 test and flows through the sum
# to yield NA.
.h_index <- function(lambda) {
  if (length(lambda) == 0L || any(abs(lambda) >= 1, na.rm = TRUE)) return(NA_real_)
  1 / (1 + 1 / sum(lambda^2 / (1 - lambda^2)))
}

# Reliability coefficients over a normalized model spec. `spec` carries the general-factor
# loadings (`g_load`), the group-factor loading matrix (`s_load`), the uniquenesses (`u2`),
# the item-to-factor correspondence map (`map`), the group-factor row labels (`fac_names`),
# the item names (`var_names`), optionally the model-implied correlation matrix of the
# variables (`implied`, see the total-variance denominators below), and -- for
# `variance = "correlation"` -- a correlation
# matrix (`cormat`) plus, for a correlated-factors solution, the factor
# intercorrelations (`Phi`, in the column order of `s_load`). `Phi` belongs to a spec
# with no general factor: it marks the group loadings as an oblique pattern whose
# factors covary, and enters the model-implied common variance in either variance mode.
# A spec that carries both a general factor and a `Phi` is not supported -- the callers that
# build a `Phi` set `g_load` to zero, and reject the combination where a user could supply
# one, because the general and group parts would no longer partition the composite.
# Given these, it computes
# McDonald's omega total, hierarchical, and subscale for the general factor and each group
# factor, and, when `add_ind = TRUE`, the H index (Hancock & Mueller, 2001) together with the
# ECV and PUC bifactor indices (Rodriguez, Reise & Haviland, 2016). Every omega numerator is
# the true score variance the model attributes to a composite, read off the model-implied
# common variance Lambda Psi Lambda'. Two total-variance denominators are supported:
# `"correlation"` takes a composite's variance from the correlation matrix, as McDonald's
# omega total does (1999, Test Theory, Eq. 6.2.1; Zinbarg, Yovel, Revelle & McDonald, 2006,
# Eqs. 4 and 6); `"sums_load"` uses the model-implied composite variance instead, and so needs
# no correlation matrix. That variance is the common variance plus the uniquenesses, unless the
# spec carries the model-implied correlation matrix of the variables in `implied`, in which case
# each composite's variance is the sum of its block of that matrix. An adapter supplies
# `implied` only where the uniquenesses do not hold the whole of the residual variance, which
# takes a fitted model with residual covariances; where they do, the two expressions are the
# same quantity and `implied` stays NULL. When
# `add_rel = TRUE`, three further columns are appended: standardized Cronbach's alpha
# (Cronbach, 1951) for the whole scale and each subscale, and -- per group factor, over its
# assigned (simple-structure) composite -- the composite reliability (congeneric omega;
# Joreskog, 1971; Raykov, 2001) and the average variance extracted (AVE, a convergent-validity
# index rather than a reliability; Fornell & Larcker, 1981). The function is purely
# computational: `spec` is assumed already normalized by the calling adapter, which owns all
# front-end input handling. `arg` names the user-facing map argument of the calling
# front-end, so the empty-factor warning points at the right one: `factor_map` for
# efa_reliability(), `factor_corres` on the frozen OMEGA() surface. Both front-ends pass
# it explicitly; the default is the frozen spelling so that a caller who forgets leaves
# OMEGA()'s wording as it has always been rather than silently changing it.
.reliability_core <- function(spec, variance = c("correlation", "sums_load"),
                              add_ind = TRUE, add_rel = FALSE,
                              arg = "factor_corres") {

  variance <- match.arg(variance)

  g_load <- spec$g_load
  s_load <- spec$s_load
  u2 <- spec$u2
  factor_corres <- spec$map
  cormat <- spec$cormat
  var_names <- spec$var_names

  # The map states membership, so only 0 and 1 (or the logicals they stand for) mean
  # anything. Any other value is read two ways below and cannot be read one way: the
  # composites are the entries equal to 1, while the map is also multiplied into the
  # loadings and the uniquenesses, where it weights them. A map of 2s therefore names the
  # same composites but empties every one of them -- the coefficients come back NA under a
  # whole-scale row that still looks right -- so it is refused rather than resolved.
  # Checked here rather than in the front ends because this is where every route meets: the
  # derived maps are 0/1 or logical by construction, but a user's arrives through several
  # callers, not all of which validate it.
  # %in% is match-based, so a missing entry answers FALSE rather than propagating.
  if (!(is.logical(factor_corres) || is.numeric(factor_corres)) ||
      !all(factor_corres %in% c(0, 1))) {
    vals <- unique(as.vector(as.matrix(factor_corres)))
    other <- utils::head(vals[!vals %in% c(0, 1)], 3L)
    cli::cli_abort(
      c("{.arg {arg}} must hold only 0 and 1.",
        "x" = "It also holds {.val {other}}.",
        "i" = "Mark each variable's group factors with 1 and every other entry with 0."),
      class = "efa_reliability_map_values"
    )
  }

  # Same general + group factor labels regardless of how s_load was obtained.
  factor_names <- c("g", seq_len(ncol(s_load)))

  # Create an input dataframe
  input <- data.frame(g_load, s_load)
  colnames(input) <- factor_names
  rownames(input) <- var_names
  input$u2 <- u2

  # seq_len(k) + 1 rather than 2:(k + 1): the latter counts down to c(2, 1) when a spec
  # arrives with no group factors, selecting the general column twice over.
  omega_mat <- input[, seq_len(ncol(s_load)) + 1L, drop = FALSE] * factor_corres

  # Sum of all g loadings
  sum_g <- sum(input$g)

  # Sum of all error variances
  sum_e <- sum(input$u2)

  # Sums of error variances and of loadings over each group factor's assigned items.
  sums_e_s <- colSums(input$u2 * factor_corres)
  sums_s_s <- colSums(omega_mat)

  # The variables a group factor's composite is made of. The omega numerators and
  # denominators are formed from this one index vector, so each coefficient's two halves
  # always describe the same composite. The map is 0/1 by the check above, so these are the
  # same items the masked sums above are taken over.
  members <- lapply(seq_len(ncol(s_load)), function(i) which(factor_corres[, i] == 1))

  # Model-implied common variance of the variables, Lambda Psi Lambda'. The general factor
  # is an extra column orthogonal to the group factors, so the quadratic form splits into a
  # general and a group part; Psi is the identity for a bifactor or Schmid-Leiman solution
  # and the factor intercorrelations Phi for a correlated-factors one, whose general column
  # is zero throughout (see above). Summing the block belonging to a set of variables gives
  # that composite's model-implied true score variance, 1' Lambda Psi Lambda' 1, which is the
  # numerator of every omega total below (Zinbarg, Revelle, Yovel & Li, 2005, Eq. 8; Zinbarg,
  # Yovel, Revelle & McDonald, 2006, Eq. 6). Its diagonal holds the communalities.
  s_mat <- as.matrix(input[, seq_len(ncol(s_load)) + 1L, drop = FALSE])
  Phi <- spec$Phi
  implied <- spec$implied
  group_common <- if (is.null(Phi)) tcrossprod(s_mat) else s_mat %*% Phi %*% t(s_mat)
  common <- tcrossprod(input$g) + group_common

  # The coefficients below describe the unit-weighted sum of the variables as supplied, so
  # flag a solution whose variables are not all keyed in the same direction before any of
  # them is reported.
  .rel_check_keying(common, var_names)

  if(variance == "correlation"){

    # Compute omega total, hierarchical, and subscale for g-factor
    omega_tot_g <- sum(common) / sum(cormat)
    omega_h_g <- sum_g^2 / sum(cormat)
    omega_sub_g <- sum(sums_s_s^2) / sum(cormat)

    # Compute omega total, hierarchical, and subscale for group factors
    omega_tot_sub <- NULL
    omega_h_sub <- NULL
    omega_sub_sub <- NULL

    # A composite's omega total is the true score variance the model attributes to it over
    # its observed variance. Every factor the composite's variables load on contributes:
    # with correlated factors through the cross-loadings and the factor correlations, and in
    # an estimated bifactor or Schmid-Leiman solution through the cross-loadings alone,
    # which such a solution always has -- only a confirmatory bifactor model carries
    # structural zeros off a composite's own column. Counting the general factor and the
    # composite's own column only would report that factor's congeneric omega, which the
    # omega subscale column already reports, under the name of a total. Reading the
    # composite's block of `common` counts every factor, whichever kind of solution the spec
    # holds.
    for (i in seq_len(ncol(s_load))) {
      subf <- members[[i]]
      Vgr <- sum(cormat[subf, subf])
      omega_tot_sub[i] <- sum(common[subf, subf]) / Vgr
      omega_h_sub[i] <- sum(input$g[subf])^2 / Vgr
      omega_sub_sub[i] <- sum(s_mat[subf, i])^2 / Vgr
    }

  } else if(variance == "sums_load") {

    # Sums of all group factor loadings for all group factors
    sums_s <- colSums(s_mat)

    # What the group factors contribute to the whole-scale composite, 1' S Psi S' 1. For
    # uncorrelated group factors that is the sum of the squared loading-column sums; for a
    # correlated-factors spec the factor correlations enter it as the quadratic form
    # (S'1)' Phi (S'1), which the squared column sums alone would drop. The subscale rows
    # below read the Phi-aware `common`, so scoring this row from the column sums would
    # describe an orthogonal model and a correlated one in the same table. The
    # uncorrelated branch keeps the literal expression, so nothing about the arithmetic of
    # a bifactor or Schmid-Leiman solution changes.
    group_var <- if (is.null(Phi)) sum(sums_s^2) else sum(sums_s * (Phi %*% sums_s))

    # Compute omega total, hierarchical, and subscale for the whole scale. The
    # composite here is all variables, so every loading on every group factor
    # contributes to its variance: the general and group terms use the full
    # loading-column sums, which add up to the whole of the model-implied common
    # variance. omega total is then 1 - sum(u^2) / V, with V
    # the model-implied composite variance, and the general and group variances
    # partition it exactly (tot = hier + sub for the g row).
    #
    # V comes from `implied` where the spec carries it. There the residual covariances hold
    # variance that `sum_e`, the sum of the residual variances alone, does not count
    # (Raykov, 2001). The numerators are the common variance either way, so the partition
    # is unaffected.
    var_g <- if (is.null(implied)) sum_g^2 + group_var + sum_e else sum(implied)
    omega_tot_g <- (sum_g^2 + group_var) / var_g
    omega_h_g <- sum_g^2 / var_g
    omega_sub_g <- group_var / var_g

    # Compute omega total, hierarchical, and subscale for group factors. A subscale
    # composite's model-implied variance is the common variance it receives from every
    # factor plus its unique variances, so the denominator reads that composite's block
    # of `common` exactly as the whole-scale row above reads the whole matrix. Counting
    # only the general and own-column terms would leave the variance the composite
    # receives through cross-loadings on the other group factors out of both the
    # numerator and the denominator, which understates omega total and inflates omega
    # hierarchical and omega subscale. The three add up on the whole-scale row but not
    # here, where omega subscale stays the factor's own contribution while omega total
    # counts them all.
    omega_tot_sub <- NULL
    omega_h_sub <- NULL
    omega_sub_sub <- NULL

    for (i in seq_len(ncol(s_load))) {
      subf <- members[[i]]
      common_sub <- sum(common[subf, subf])
      # As on the whole-scale row: that composite's block of `implied` where the spec
      # carries one, its own unique variances otherwise.
      Vsub <- if (is.null(implied)) common_sub + sum(u2[subf]) else
        sum(implied[subf, subf])
      omega_tot_sub[i] <- common_sub / Vsub
      omega_h_sub[i] <- sum(input$g[subf])^2 / Vsub
      omega_sub_sub[i] <- sum(s_mat[subf, i])^2 / Vsub
    }
  }

  # A group factor with no assigned items (an all-zero map column, e.g. a
  # type = "psych" factor that is never the largest loading for any item) has a
  # zero subscale-variance denominator and hence undefined coefficients. Report
  # them as NA -- consistent with the H index below -- instead of a silent NaN.
  empty_fac <- which(colSums(factor_corres == 1) == 0)
  if (length(empty_fac) > 0) {
    omega_tot_sub[empty_fac] <- NA
    omega_h_sub[empty_fac] <- NA
    omega_sub_sub[empty_fac] <- NA
    cli::cli_warn(
      c("Some group factors have no assigned items in {.arg {arg}}.",
        "i" = "Their omega coefficients and H index are returned as {.val NA}."),
      class = "efa_omega_empty_factor"
    )
  }

  # Combine and display results in a table
  omega_tot <- c(omega_tot_g, omega_tot_sub)
  omega_h <- c(omega_h_g, omega_h_sub)
  omega_sub <- c(omega_sub_g, omega_sub_sub)

  # A share of variance above 1, or a general-factor share above the share of all factors
  # together, is not an admissible reliability whichever total variance it was divided by.
  # The two conventions get there by different routes, so the remedy differs with them.
  #
  # In "correlation" mode the denominator (the correlation matrix) is independent of the
  # loading-based numerators, so a correlation matrix inconsistent with the supplied loadings
  # can push a coefficient above 1. It takes a composite whose loadings over-predict its
  # observed variance by more than the whole of its unique variance, which cannot happen when
  # the loadings were estimated from that same matrix, but can when the two come from
  # different sources.
  #
  # In "sums_load" mode the denominator is that composite's own common variance plus its own
  # unique variances, so an omega total above 1 takes those uniquenesses to sum to something
  # negative -- an improper solution, which the Heywood check below reports from the
  # uniquenesses themselves, except that that check is gated on `add_ind` while this one is
  # not. Omega subscale reaches it by a second route that needs no improper solution at all:
  # it divides a group factor's own squared loading sum by its composite's whole variance,
  # and once the group factors are correlated that sum is not a part of it, because the
  # composite's cross-loadings on the other factors can contribute negative common variance
  # and take the total below what its own factor puts in.
  #
  # Either way, a set of factor correlations that is not positive semi-definite makes the
  # group factors contribute negative common variance, which can put omega hierarchical above
  # omega total. Warn rather than return any of these silently.
  tol <- .Machine$double.eps^0.5
  if (any(c(omega_tot, omega_h, omega_sub) > 1 + tol, na.rm = TRUE) ||
      any(omega_h > omega_tot + tol, na.rm = TRUE)) {

    # Without a general factor omega hierarchical is zero throughout, so the second clause
    # there says an omega total came out negative -- and the front-end drops the hierarchical
    # column for such a solution, so naming it would point at a coefficient the result does
    # not carry.
    lead <- if (isTRUE(all(input$g == 0))) {
      "Some omega coefficients are out of range (above 1, or a negative omega total)."
    } else {
      "Some omega coefficients are out of range (above 1, or omega hierarchical above omega total)."
    }

    # The uniquenesses are named rather than pointed at as {.arg u2}: they are only an
    # argument on the manual and bare-matrix routes, and come from the model on the others.
    hints <- if (variance == "correlation") {
      "Check that {.arg cormat} is consistent with the loadings."
    } else {
      "Check the uniquenesses of the affected composite -- its own subtotal, not their sum over all variables -- for a negative value."
    }
    if (!is.null(Phi)) {
      hints <- c(hints, paste("Omega subscale can also exceed 1 in a proper solution, when a",
                              "composite's cross-loadings on the other group factors reduce its",
                              "variance below what its own factor contributes; inspect those and",
                              "{.field Phi}."))
    }
    names(hints) <- rep("i", length(hints))

    cli::cli_warn(c(lead, hints), class = "efa_omega_out_of_range")
  }

  # Optional reliability / convergent-validity coefficients, appended as columns when
  # requested. Kept behind add_rel so the OMEGA coefficient menu is unchanged otherwise.
  if (isTRUE(add_rel)) {

    # Standardized Cronbach's alpha (Cronbach, 1951), k / (k - 1) *
    # (1 - sum(diag(R_sub)) / sum(R_sub)), for the whole scale (all items) and each
    # subscale (the map's assigned items). It needs a correlation matrix: the supplied
    # cormat; else the model-implied one the spec carries in `implied`, which already has a
    # unit diagonal; else the model-implied common variance plus the uniquenesses (which
    # assumes the model holds and that the residual covariance matrix is diagonal),
    # standardized to a unit diagonal with cov2cor so the formula returns the standardized
    # coefficient even when the loadings and uniquenesses do not complete to unit item
    # variance. Standardized alpha uses the correlation matrix; raw, covariance-based alpha
    # would need item standard deviations the spec does not carry. Alpha is undefined for
    # fewer than two items (returned NA).
    R_rel <- if (!is.null(cormat)) {
      cormat
    } else if (!is.null(implied)) {
      implied
    } else {
      stats::cov2cor(common + diag(u2, nrow = length(u2)))
    }

    composites <- c(list(seq_len(nrow(s_mat))), members)
    alpha <- vapply(composites, function(idx) {
      k <- length(idx)
      if (k < 2L) return(NA_real_)
      Rk <- R_rel[idx, idx, drop = FALSE]
      k / (k - 1) * (1 - sum(diag(Rk)) / sum(Rk))
    }, numeric(1))

    # Composite reliability (congeneric omega; Joreskog, 1971; Raykov, 2001) and average
    # variance extracted (AVE; Fornell & Larcker, 1981) per group factor, over that
    # factor's assigned composite (simple structure): CR = (sum L)^2 / ((sum L)^2 +
    # sum u2), AVE = sum L^2 / (sum L^2 + sum u2), summed over the assigned items only --
    # a foreign item's cross-loading never enters the subscale score, so it must not enter
    # its reliability. AVE is a convergent-validity index, not a reliability. Both describe
    # a single group factor, so the general-factor (whole-scale) row is NA. sums_s_s and
    # sums_e_s are the assigned loading and uniqueness sums already formed for the omegas,
    # and omega_mat is the map-masked group loadings, so colSums(omega_mat^2) is the
    # assigned sum of squared loadings.
    sq_assigned <- colSums(omega_mat^2)
    CR  <- c(NA_real_, sums_s_s^2 / (sums_s_s^2 + sums_e_s))
    AVE <- c(NA_real_, sq_assigned / (sq_assigned + sums_e_s))

    # A group factor with no assigned items has no composite: NA, as for the omegas.
    # CR and AVE would otherwise be 0 / 0 = NaN; alpha is already NA for an empty
    # composite via its k < 2 guard, so only CR and AVE need the explicit override.
    if (length(empty_fac) > 0) {
      CR[empty_fac + 1] <- NA
      AVE[empty_fac + 1] <- NA
    }
  }

  if(isTRUE(add_ind)){

    # Compute H index, ECV, and PUC
    h_s <- vector("double", ncol(s_load))
    h_s_load <- vector("double", ncol(s_load))
    hiload_s <- logical(ncol(s_load))

    # The H index (Hancock & Mueller, 2001) of a group factor is defined over that factor's
    # own indicators, so its loadings are restricted to the items assigned to the factor by
    # the map; .h_index guards its denominator against a loading of absolute value at least 1.
    # `hiload_s` flags, per group factor, such a loading on an item with a positive unique
    # variance -- i.e. an oblique pattern coefficient >= 1 in a proper solution, not a Heywood
    # case -- so its undefined H is reported separately from (and independently of) any Heywood
    # case below. ECV (Rodriguez et al., 2016) is a ratio of common variances and uses the
    # full group-factor loading columns. The map defines item-to-factor membership, which also
    # enters the PUC.
    for (j in seq_len(ncol(s_load))){
      s_j <- input[[j + 1]]
      mem <- factor_corres[, j] == 1
      s_mem <- s_j[mem]
      h_s[j] <- .h_index(s_mem)
      hiload_s[j] <- length(s_mem) > 0 &&
        any(abs(s_mem) >= 1 & u2[mem] >= .Machine$double.eps, na.rm = TRUE)
      h_s_load[j] <- sum(s_j^2)
    }

    h_g <- .h_index(input$g)
    hiload_g <- any(abs(input$g) >= 1 & u2 >= .Machine$double.eps, na.rm = TRUE)

    # Heywood case: an improper solution, i.e. a unique variance at or below zero
    # (equivalently a communality at or above 1). This is the standard definition, as in
    # EFA() and the lavaan adapter, and is a property of the solution rather than of the H
    # index, so it is flagged from the uniquenesses (not the loadings). Coefficients
    # involving the affected variables may not be interpretable. which() drops any NA
    # uniqueness (a malformed component, not a Heywood case) so it neither triggers a
    # false Heywood warning nor suppresses the h-undefined note below.
    heywood_idx <- which(u2 < .Machine$double.eps)
    heywood_vars <- if (is.null(var_names)) as.character(heywood_idx) else
      var_names[heywood_idx]
    if (length(heywood_vars) > 0) {
      cli::cli_warn(
        c(paste("{cli::qty(heywood_vars)}Heywood case{?s} detected for {.val {heywood_vars}}:",
                "a communality at or above 1 leaves a non-positive unique variance."),
          "i" = "{cli::qty(heywood_vars)}Coefficients involving the affected variable{?s} may not be interpretable."),
        class = "efa_reliability_heywood"
      )
    }

    # Separately, the H index is undefined for any construct with a loading whose absolute
    # value is at least 1 (its 1 - lambda^2 error term is non-positive), so .h_index has
    # returned NA for it. For standardized (orthogonal) loadings a loading >= 1 only occurs
    # in a Heywood case, already flagged above from the uniquenesses; hiload_* count only the
    # constructs whose offending loading has a positive unique variance (an oblique pattern
    # coefficient in a proper solution), so this note is per-construct -- it fires for such a
    # construct even when an unrelated Heywood case exists elsewhere, and stays silent when
    # the loading >= 1 is itself the Heywood case.
    if (isTRUE(hiload_g) || any(hiload_s)) {
      cli::cli_warn(
        c("The H index could not be computed for some constructs (a loading with an absolute value of at least 1).",
          "i" = "The H index is returned as {.val NA} for those constructs."),
        class = "efa_reliability_h_undefined"
      )
    }

    ECV <- sum(input$g^2) / sum(sum(input$g^2), sum(h_s_load))

    # Proportion of uncontaminated correlations from the map membership (matching
    # the H index; the general factor is not a contaminant; see .puc).
    PUC <- .puc(factor_corres == 1)

    # Create output
    h <- c(h_g, h_s)

    omegas <- cbind(omega_tot, omega_h, omega_sub, h, NA, NA)
    omegas[1, 5] <- ECV
    omegas[1, 6] <- PUC
    colnames(omegas) <- c("tot", "hier", "sub", "H", "ECV", "PUC")

  } else {

    omegas <- cbind(omega_tot, omega_h, omega_sub)
    colnames(omegas) <- c("tot", "hier", "sub")

  }

  # Append the reliability / convergent-validity columns when computed.
  if (isTRUE(add_rel)) {
    omegas <- cbind(omegas, alpha = alpha, CR = CR, AVE = AVE)
  }

  rownames(omegas) <- c("g", spec$fac_names)

  class(omegas) <- "OMEGA"

  omegas

}

# Front-end adapters: normalize each reliability input source to the spec consumed
# by .reliability_core(). Every adapter returns the same list
# (g_load, s_load, u2, map, cormat, var_names, fac_names), so the core stays
# input-agnostic and the front-end only has to pick the right adapter.

# Resolve the item-to-factor correspondence map for adapters that take a separate
# `factor_corres`/`type` (SL, schmid, manual, efa): with a supplied `factor_corres`,
# validate its dimensions and use it; otherwise, for `type = "psych"`, assign each
# item to its highest-|loading| group factor (as psych::omega does); `type =
# "EFAtools"` requires an explicit `factor_corres`. `arg` names the user-facing
# argument the map came from, as in .rel_check_map().
.rel_map <- function(s_load, factor_corres = NULL, type = c("EFAtools", "psych"),
                     arg = "factor_map") {

  type <- match.arg(type)
  s_load <- as.matrix(s_load)

  if (!is.null(factor_corres)) {
    .rel_assert_map_dim(factor_corres, s_load, arg = arg)
    .rel_check_map(s_load, factor_corres, arg = arg)
    return(factor_corres)
  }

  if (type == "EFAtools") {
    cli::cli_abort(
      "Specify {.arg {arg}}, or set {.code type = \"psych\"} to derive variable-to-factor correspondences from the highest group-factor loading per variable.",
      class = "efa_reliability_need_corres"
    )
  }

  # type == "psych": each variable's correspondence is its largest |group loading|.
  map <- matrix(0, nrow = nrow(s_load), ncol = ncol(s_load))
  for (i in seq_len(nrow(s_load))) {
    map[i, which.max(abs(s_load[i, ]))] <- 1
  }
  map

}

# Abort when a user-supplied item-to-factor map does not conform to the group loadings
# it will be applied to: one row per variable and one column per group factor. Without
# this check a map with too few rows is recycled against the loadings and returns
# coefficients above 1 rather than an error. `arg` names the user-facing argument the map
# came from, as in .rel_check_map().
.rel_assert_map_dim <- function(map, s_load, arg = "factor_map") {

  s_load <- as.matrix(s_load)

  if (!is.matrix(map) || !identical(dim(map), dim(s_load))) {
    got <- if (is.matrix(map)) {
      "{.arg {arg}} has {nrow(map)} row{?s} and {ncol(map)} column{?s}."
    } else {
      "{.arg {arg}} is not a matrix."
    }
    cli::cli_abort(
      c("{.arg {arg}} must have one row per variable and one column per group factor.",
        "x" = got,
        "i" = "The solution has {nrow(s_load)} variable{?s} and {ncol(s_load)} group factor{?s}."),
      class = "efa_reliability_map_dim"
    )
  }

  invisible(map)

}

# Sanity-check a user-supplied item-to-factor map against the loadings it will be
# applied to. The map's columns are matched to the group factors by position, so a map
# written in a different factor order than the solution yields well-formed but
# meaningless subscale coefficients rather than an error. Flag each column whose
# assigned items barely load on it (mean |loading| below `min_load`) while the same
# items do load on some other column -- the signature of a permuted or mistranscribed
# map. Requiring a better home for the items keeps the check to that signature: a map
# whose items load on nothing is a weak solution, not a misaligned map, and is left
# alone. `min_load` is deliberately far below the salience conventions used elsewhere
# in the package (`salience_threshold`, .3 by default): a deliberate map may well
# disagree with the loadings, so only assignments that are essentially unsupported are
# flagged. Silent by construction for a map derived by .rel_map(): there each item is
# assigned to its own largest |loading|, so a column's assigned mean can never fall
# below another column's mean for those items. `arg` names the user-facing argument the
# map came from, so the message points at the right one for each front-end. A map that
# does not conform to the loadings is left to the caller, which validates the dimensions.
.rel_check_map <- function(s_load, map, min_load = 0.1, arg = "factor_map") {

  s_load <- abs(as.matrix(s_load))
  map <- as.matrix(map)
  if (!identical(dim(map), dim(s_load))) return(invisible(NULL))
  assigned <- map == 1

  labs <- colnames(s_load)
  if (is.null(labs)) labs <- as.character(seq_len(ncol(s_load)))

  implausible <- vapply(seq_len(ncol(assigned)), function(j) {
    idx <- which(assigned[, j])
    # An empty column carries no items to judge; the core reports it separately.
    if (length(idx) == 0L) return(FALSE)
    own <- mean(s_load[idx, j])
    others <- colMeans(s_load[idx, -j, drop = FALSE])
    isTRUE(own < min_load) && any(others > min_load, na.rm = TRUE)
  }, logical(1))

  if (any(implausible)) {
    bad <- labs[implausible]
    cli::cli_warn(
      c("{cli::qty(bad)}The items assigned to {.arg {arg}} column{?s} {.val {bad}} hardly load on {?that/those} group factor{?s}, but do load on another one.",
        "i" = "The columns of {.arg {arg}} are matched to the group factors by position; check that they are in the column order of the solution."),
      class = "efa_reliability_implausible_map"
    )
  }

  invisible(NULL)

}

# Warn when the variables of a solution are not all keyed in the same direction -- most
# often a scale whose reverse-worded items were never reverse-coded. Every coefficient
# computed here describes the raw unit-weighted sum of the variables the solution was fitted
# on, so a
# variable keyed against the rest subtracts from that sum's true score variance instead of
# adding to it and the coefficients collapse: correct for the sum that was scored, but
# reading as a poor scale rather than as a missing step. The variables are deliberately not
# reflected first, as psych::omega does, because reflection describes a different composite
# than the one the coefficients are defined on (Flora, 2020).
#
# "The variables the solution was fitted on" is meant literally: fitted to polychoric or
# tetrachoric correlations, or to an `ordered` lavaan fit, the composite is of the latent
# responses assumed to underlie the ordinal answers, not of the answers themselves, and the
# keying rule below reads that composite. The distinction does not change the rule -- a
# variable works against the sum on either metric -- but it does change what the reported
# coefficients are the reliability OF. See the Details of efa_reliability() and Green and
# Yang (2009) for the ordinal-score alternative, which is not computed here.
#
# `common` is the model-implied common variance Lambda Psi Lambda' of the variables, whose
# total sum(common) = 1' Lambda Psi Lambda' 1 is the composite's true score variance. Under
# any re-keying D = diag(+-1) of the variables that variance becomes sum_ij d_i d_j
# common_ij, which is at most sum(abs(common)); the ratio of the two is therefore 1 exactly
# when no variable works against the sum and falls towards 0 as the keying becomes balanced.
# The rule reads the common variance rather than the loading signs on purpose: `g_load` is
# zero throughout for a correlated-factors spec, so a sign test on it never fires there, and
# the group loadings carry no sign information either, because the factor columns are
# reflected -- a reverse-keyed block comes back with positive loadings and a negative factor
# correlation, which the common variance shows and the loadings do not.
#
# `cutoff` is deliberately far below any well-keyed solution, as `min_load` is in
# .rel_check_map(): the ratio is (P - N) / (P + N) over the positive and negative common
# variance, so firing takes negative common variance exceeding about a seventh of the
# positive. On the bundled data no correctly keyed solution reaches that: the lowest is .89
# (DOSPERT_raw at two factors, whose domains are only weakly related), whereas a scale whose
# reverse-worded items were never reverse-coded stays below .32 (UPPS_raw, two to six
# factors) and reverse-coding one factor's block of a well-keyed solution takes it to .23.
#
# The variables named are those whose own common variance with the composite, the row sum of
# `common`, is negative -- the ones subtracting from the sum, and so the ones to reverse-code.
# They are named for guidance and not as the trigger: a variable can drag the total down
# without its own row sum turning negative, so the ratio decides whether to warn and the row
# sums only say where to look. The list is therefore allowed to come out empty, in which case
# the message states the finding without it. `var_names` may be NULL, as for the Heywood
# report above, in which case the positions stand in for the names.
.rel_check_keying <- function(common, var_names = NULL, cutoff = 0.75) {

  aligned <- sum(abs(common))

  # Nothing to judge: no loadings at all, or a malformed component (NA) the callers
  # report separately.
  if (!is.finite(aligned) || aligned <= .Machine$double.eps^0.5) return(invisible(NULL))

  if (isTRUE(sum(common) < cutoff * aligned)) {

    against <- which(rowSums(common) < 0)
    n_bad <- length(against)

    lead <- if (n_bad > 0) {
      labs <- if (is.null(var_names)) as.character(against) else var_names[against]
      # Capped as elsewhere: a long scale can have dozens of reverse-worded items, and a
      # message that prints them all is unreadable.
      cap <- .cap_label_list(labs)
      paste("The variables are not all keyed in the same direction:",
            "{.val {cap$shown}}{cap$rest} {cli::qty(n_bad)}{?is/are} keyed against the rest.")
    } else {
      "The variables are not all keyed in the same direction."
    }

    cli::cli_warn(
      c(lead,
        "i" = "The coefficients describe the raw unit-weighted sum of the variables as supplied, in which the reversed variables cancel rather than add.",
        "i" = "Reverse-code them before fitting the solution (Flora, 2020)."),
      class = "efa_reliability_mixed_keying"
    )
  }

  invisible(NULL)

}

# Warn (classed) when supplied uniquenesses do not complete the supplied loadings to unit
# variance. `h2` is the diagonal of the model-implied common variance Lambda Psi Lambda',
# so a standardized solution has h2 + u2 = 1 for every variable. Only the routes whose
# uniquenesses come from the user need this: every fitted route derives them as 1 - h2, and
# so satisfies it by construction.
#
# What it catches is a slip on adjacent columns. The printed Schmid-Leiman table ends in
# `h2` and `u2`, and many published tables report only `h2`, leaving the reader to form
# `1 - h2` by hand -- so passing communalities where uniquenesses belong is a one-character
# mistake rather than a contrived one. It is silent otherwise: `u2` does not enter the
# coefficients at all under `variance = "correlation"` with a correlation matrix, and that
# is the mode a user without one is turned away from, so the mistake is inert exactly where
# it would be harmless and load-bearing exactly where it is not.
#
# A warning rather than an error, because unstandardized (covariance-metric) components are
# a defensible input that yields a legitimate raw-score coefficient; the user needs to know
# the components are not standardized, not to be stopped. `tol` sits well clear of both
# sides: a published table rounded to two decimals leaves a residual of about .015, while
# supplying communalities leaves at least .40.
.rel_check_u2 <- function(h2, u2, tol = 0.05) {

  # Nothing to compare: no variables at all, or a `u2` that does not line up with the
  # loadings, which is the callers' error to report and not a statement about
  # standardization. Either would otherwise reach `max()` over an empty vector or compare a
  # recycled one.
  if (length(h2) == 0L || length(h2) != length(u2)) return(invisible(NULL))

  resid <- max(abs(h2 + u2 - 1))

  # A non-finite residual is a malformed component, which the callers report separately.
  if (!isTRUE(resid > tol)) return(invisible(NULL))

  cli::cli_warn(
    c("The supplied {.arg u2} do not complete the loadings to unit variance.",
      "x" = "A communality plus its uniqueness differs from 1 by up to {round(resid, 3)}.",
      "i" = "A standardized solution has {.code u2 = 1 - h2}; check that {.arg u2} holds uniquenesses rather than communalities, and that they belong to these loadings.",
      "i" = "Unstandardized components are scored as given, and yield the coefficients of the raw composite."),
    class = "efa_reliability_u2_not_standardized"
  )

  invisible(NULL)

}

# Abort (classed) unless `Phi` can be the factor intercorrelation matrix of the `k` group
# factors it will be scored with. `NULL` means the identity -- uncorrelated group factors, as a
# Schmid-Leiman or bifactor solution has by construction -- and is always admissible, so this
# only runs where a matrix was actually supplied.
#
# The checks matter because `Phi` enters the model-implied common variance as the quadratic form
# S Phi S', where a matrix that is not a correlation matrix does not merely mislabel the solution:
# an asymmetric one is silently symmetrised by the quadratic form, a non-unit diagonal rescales
# the factors, and one that is not positive semi-definite attributes negative common variance to a
# composite and returns coefficients outside [0, 1] with nothing else to catch them.
#
# The shape tests are written out rather than delegated to .is_cormat(): that helper classifies
# raw data against a correlation matrix, so it rejects by design the 1 x 1 matrix a single group
# factor takes, and it aborts on a missing value instead of reporting one. Positive
# semi-definiteness is judged on the smallest eigenvalue against the package's usual tolerance --
# an estimated factor correlation matrix is often singular to rounding, which is admissible, while
# a genuinely indefinite one is not.
.rel_check_phi <- function(Phi, k, arg = "Phi") {

  if (is.null(Phi)) return(invisible(NULL))

  bad <- function(detail) {
    cli::cli_abort(
      c("{.arg {arg}} must be the correlation matrix of the {k} group factor{?s}.",
        "x" = detail,
        "i" = "Leave it {.val NULL} for uncorrelated group factors."),
      class = "efa_reliability_bad_phi"
    )
  }

  if (!is.matrix(Phi) || !is.numeric(Phi)) {
    bad("It is {.obj_type_friendly {Phi}}, not a numeric matrix.")
  }
  if (nrow(Phi) != k || ncol(Phi) != k) {
    bad("It is {nrow(Phi)} by {ncol(Phi)}, not {k} by {k}.")
  }
  if (anyNA(Phi)) {
    bad("It contains missing values.")
  }

  tol <- .Machine$double.eps^0.5
  if (any(abs(Phi - t(Phi)) > tol)) {
    bad("It is not symmetric.")
  }
  if (any(abs(diag(Phi) - 1) > tol)) {
    bad("Its diagonal is not all ones.")
  }
  # Symmetric by the test above, so the eigenvalues are real.
  if (min(eigen(Phi, symmetric = TRUE, only.values = TRUE)$values) < -tol) {
    bad("It is not positive semi-definite, so it is not a correlation matrix.")
  }

  invisible(NULL)

}

# Abort (classed) when a supplied correlation matrix fails the .is_cormat() check;
# shared guard for the adapters that accept a user cormat. A correlation matrix supplied as a
# data frame -- what read.csv() returns for a published correlation table -- is accepted and
# handed back as a matrix, as .prepare_cor_input() does on the other input route. The coercion
# is load-bearing for one operation only: the standardized-alpha branch of .reliability_core()
# takes diag() of a subset of it, which a data frame does not support. The omega quantities
# only sum and subset it, which a data frame does support -- so without the coercion they
# would return correct numbers and only add_rel = TRUE would fail. Keep the coercion here
# rather than at the one operation, so the spec carries a single type.
#
# The matrix is also checked against the solution it will be used with, given as the
# number of variables (`n_items`) and, where the solution carries them, their names
# (`var_names`). The subscale coefficients divide loading sums by the correlations of the
# same items, so a correlation matrix of other variables, or of the same variables in
# another order, yields well-formed but wrong subscale coefficients while leaving the
# whole-scale row -- which only sums the whole matrix -- untouched, and one of the wrong
# size fails much later as an out-of-bounds subscript. Named variables in a different
# order are an unambiguous permutation and are reordered to the solution; a different set
# of names cannot be resolved and aborts.
.rel_check_cormat <- function(cormat, var_names = NULL, n_items = NULL) {

  if (!.is_cormat(cormat)) {
    cli::cli_abort(
      c("{.arg cormat} is not a correlation matrix.",
        "i" = "Check the {.arg cormat} input, or leave it {.code NULL}."),
      class = "efa_reliability_not_cormat"
    )
  }

  cormat <- as.matrix(cormat)

  if (is.null(n_items) && !is.null(var_names)) n_items <- length(var_names)

  if (!is.null(n_items) && nrow(cormat) != n_items) {
    cli::cli_abort(
      c("{.arg cormat} must have one row and column per variable in the solution.",
        "x" = "{.arg cormat} has {nrow(cormat)} variable{?s}, but the solution has {n_items}.",
        "i" = "Supply the correlation matrix of the variables in the solution."),
      class = "efa_reliability_cormat_dim"
    )
  }

  # Only labelled variables can be matched. An unlabelled correlation matrix, or an
  # unlabelled solution, is matched by position as before.
  rn <- rownames(cormat)
  cn <- colnames(cormat)
  if (!is.null(rn) && !is.null(cn) && !identical(rn, cn)) {
    cli::cli_abort(
      c("The row and column names of {.arg cormat} differ.",
        "i" = "Use the same names on both axes, or remove them to match by position."),
      class = "efa_reliability_cormat_names"
    )
  }
  if (is.null(rn)) rn <- cn

  if (!is.null(var_names) && !is.null(rn) && !identical(rn, var_names)) {
    if (anyDuplicated(rn) == 0L && anyDuplicated(var_names) == 0L &&
        setequal(rn, var_names)) {
      idx <- match(var_names, rn)
      cormat <- cormat[idx, idx, drop = FALSE]
    } else {
      cli::cli_abort(
        c("The variables of {.arg cormat} are not the variables of the solution.",
          "i" = "Supply the correlation matrix of the solution's variables, in any order."),
        class = "efa_reliability_cormat_names"
      )
    }
  }

  cormat
}

# Adapter: normalize an SL() object to the reliability spec. Mirrors the SL branch of
# .OMEGA_FLEX. Honours a user cormat, else the SL object's stored correlation matrix
# (orig_R = NA for flexible-input SL objects, in which case cormat stays NULL and
# only variance = "sums_load" applies).
.rel_adapt_SL <- function(model, factor_corres = NULL, type = "EFAtools",
                          cormat = NULL, fac_names = NULL) {

  sl <- model$sl
  var_names <- rownames(sl)
  g_load <- sl[, 1]
  s_load <- sl[, 2:(ncol(sl) - 2), drop = FALSE]
  u2 <- sl[, "u2"]

  if (is.null(cormat)) {
    if (is.matrix(model$orig_R)) cormat <- model$orig_R
  } else {
    cormat <- .rel_check_cormat(cormat, var_names, nrow(sl))
  }

  if (is.null(fac_names)) fac_names <- colnames(s_load)

  list(g_load = g_load, s_load = s_load, u2 = u2,
       map = .rel_map(s_load, factor_corres, type),
       cormat = cormat, var_names = var_names, fac_names = fac_names)

}

# Adapter: normalize a psych::schmid() object to the reliability spec. Mirrors the
# schmid branch of .OMEGA_FLEX. When no cormat is supplied, it is reconstructed from
# the schmid pattern matrix and factor intercorrelations via psych::factor.model.
.rel_adapt_schmid <- function(model, factor_corres = NULL, type = "psych",
                              cormat = NULL, fac_names = NULL) {

  pattern <- model$oblique
  Phi <- model$phi
  sl <- model$sl

  s_load_names <- setdiff(colnames(sl[, -1]), c("h2", "u2", "p2", "com"))
  var_names <- rownames(sl)
  g_load <- sl[, 1]
  s_load <- sl[, s_load_names, drop = FALSE]
  u2 <- sl[, "u2"]

  if (is.null(cormat)) {
    cormat <- psych::factor.model(f = pattern, Phi = Phi, U2 = FALSE)
  } else {
    cormat <- .rel_check_cormat(cormat, var_names, nrow(sl))
  }

  if (is.null(fac_names)) fac_names <- s_load_names

  list(g_load = g_load, s_load = s_load, u2 = u2,
       map = .rel_map(s_load, factor_corres, type),
       cormat = cormat, var_names = var_names, fac_names = fac_names)

}

# Adapter: normalize manually supplied SL/bifactor components (g_load, s_load, u2) to
# the reliability spec. Mirrors the model = NULL branch of .OMEGA_FLEX. A cormat may be
# given directly or reconstructed from an oblique pattern matrix and factor
# intercorrelations -- one or the other, not both -- or left NULL, which only
# variance = "sums_load" can then score.
#
# `Phi` describes the loadings it is paired with. Supplied together with `pattern` it belongs
# to that separate oblique solution and is used only to rebuild the correlation matrix, as it
# always has been -- the pairing is what a Schmid-Leiman input needs, where `s_load` holds the
# orthogonalized group loadings and `pattern` the parent oblique ones, two different matrices
# whose model-implied correlations differ once more than three factors make the second-order
# model overidentified. Supplied without a `pattern` there is no other loading matrix for it
# to describe, so it is the factor intercorrelation matrix of `s_load` itself: the components
# then specify a correlated-factors solution, and it enters the model-implied common variance
# as the quadratic form S Phi S'. That is the one way this route can express correlated group
# factors; without it the coefficients would describe an orthogonal model instead, and fall
# short of the ones the same solution gets through `efa_fit()`.
.rel_adapt_manual <- function(g_load, s_load, u2, var_names, factor_corres = NULL,
                              type = "EFAtools", cormat = NULL, pattern = NULL,
                              Phi = NULL, fac_names = NULL) {

  # unclass() before as.matrix(): a Schmid-Leiman loading table is a matrix already, so
  # as.matrix() returns it with its class intact, and the data.frame() the core builds from
  # it then keeps it as a single column instead of expanding it -- which fails only later,
  # on a names/length mismatch that says nothing about the input.
  s_load <- as.matrix(unclass(s_load))

  # `cormat` and `pattern` are two ways to give one thing, so taking both is an ambiguous
  # request rather than a redundant one. It also cannot be resolved quietly: with a `cormat`
  # in hand the reconstruction never runs and `pattern` is read by nothing, yet its presence
  # still decides whether `Phi` describes this solution or that one -- an argument with no
  # effect of its own silently choosing between two different sets of coefficients. Refuse the
  # combination so the caller says which correlation matrix they mean.
  if (!is.null(cormat) && !is.null(pattern)) {
    cli::cli_abort(
      c("{.arg cormat} and {.arg pattern} are two ways to give the same correlation matrix, so supply one or the other.",
        "i" = "Drop {.arg pattern} to score the components against {.arg cormat}.",
        "i" = "Drop {.arg cormat} to reconstruct one instead, from {.arg pattern} and the {.arg Phi} of the oblique solution it came from."),
      class = "efa_reliability_cormat_and_pattern"
    )
  }

  # Whether `Phi` describes this solution's own group factors, rather than a `pattern`'s.
  phi_solution <- !is.null(Phi) && is.null(pattern)

  # Checked on either reading: paired with a `pattern` it is what the reconstructed
  # correlation matrix is built from, and so ends up in every coefficient's denominator,
  # which makes it exactly as load-bearing as it is on its own. It has to be the correlation
  # matrix of whichever loadings it describes, hence the two column counts.
  if (!is.null(Phi)) {
    .rel_check_phi(Phi, if (phi_solution) ncol(s_load) else NCOL(pattern))
  }

  if (phi_solution) {
    # A general factor and correlated group factors together do not give the variance
    # decomposition these coefficients report: omega hierarchical and omega subscale would
    # no longer partition the composite, and the PUC presupposes that two variables of
    # different group factors share only the general factor. The core makes the same
    # statement in its own terms; refuse the combination here, where it can be named.
    if (!isTRUE(all(g_load == 0))) {
      cli::cli_abort(
        c("{.arg Phi} describes correlated group factors, which cannot be combined with a general factor.",
          "i" = "Supply {.arg Phi} for a correlated-factors solution, whose {.arg g_load} is zero throughout.",
          "i" = "For a Schmid-Leiman or bifactor solution the group factors are orthogonal; pass {.arg pattern} as well if {.arg Phi} belongs to the oblique solution it came from."),
        class = "efa_reliability_phi_with_general"
      )
    }
  }

  # A `pattern` alongside `Phi` says the two belong together and `s_load` is a separate,
  # orthogonalized matrix -- which is exactly right for a Schmid-Leiman input, and exactly
  # wrong for components that carry no general factor, where `s_load` is itself the oblique
  # pattern and the factor correlations belong in its common variance. Here `pattern` is
  # load-bearing (it is reconstructing the correlation matrix, or the check above would have
  # aborted), so the reading cannot be settled from the arguments: a genuine parent oblique
  # solution and a `pattern` that merely repeats `s_load` look the same. Take the documented
  # one and say what was assumed, since the two give materially different coefficients.
  if (!is.null(Phi) && !is.null(pattern) && isTRUE(all(g_load == 0))) {
    cli::cli_warn(
      c("{.arg Phi} is read as the factor correlations of {.arg pattern}, not of {.arg s_load}.",
        "i" = "It is used to reconstruct the correlation matrix only, so the coefficients describe uncorrelated group factors.",
        "i" = "Omit {.arg pattern} to score {.arg s_load} as a correlated-factors solution instead."),
      class = "efa_reliability_phi_pattern"
    )
  }

  if (is.null(cormat)) {
    if (!is.null(Phi) && !is.null(pattern)) {
      cormat <- psych::factor.model(f = pattern, Phi = Phi, U2 = FALSE)
    }
  } else {
    # Matched against the loadings' own row names, not `var_names`: here `var_names` are
    # the labels the caller wants on the output rows, which need not be the names the
    # correlation matrix carries. Unlabelled components are matched by position, as the
    # manual contract -- every component in the row order of the loadings -- implies.
    cormat <- .rel_check_cormat(cormat, rownames(s_load), nrow(s_load))
  }

  # The communalities the supplied loadings imply -- the diagonal of Lambda Psi Lambda',
  # under the factor correlations where this spec carries them. The uniquenesses are the
  # caller's own here, so they are checked against these.
  group_h2 <- if (phi_solution) rowSums((s_load %*% Phi) * s_load) else rowSums(s_load^2)
  .rel_check_u2(g_load^2 + group_h2, u2)

  if (is.null(fac_names)) fac_names <- seq_len(ncol(s_load))

  list(g_load = g_load, s_load = s_load, u2 = u2,
       map = .rel_map(s_load, factor_corres, type),
       Phi = if (phi_solution) Phi else NULL,
       cormat = cormat, var_names = var_names, fac_names = fac_names)

}

# Adapter: normalize a lavaan single-factor, correlated-factors, bifactor, or
# second-order solution to the reliability spec, one entry per fitted group. Mirrors the
# structural detection and Schmid-Leiman transform of .OMEGA_LAVAAN (second-order group
# loadings via .sl_group_loadings), and drives the core with variance = "sums_load" (the
# composite variances are model-implied, and are read off the model-implied correlation
# matrix where the fit has residual covariances). Returns a list with, per group, a full spec
# (correlated factors / bifactor / second-order / single factor); a single-factor group is
# additionally flagged `single = TRUE`, since a solution with one factor defines fewer
# coefficients than the core computes for it. A correlated-factors spec is flagged by
# `correlated = TRUE` on the returned list: it carries no general factor, so a front-end
# that reports the general-factor coefficients has to drop them (efa_reliability) or
# refuse the input (OMEGA).
.rel_adapt_lavaan <- function(model, g_name = "g", group_names = NULL) {

  .require_lavaan()

  if (isFALSE(lavaan::lavInspect(model, what = "converged"))) {
    cli::cli_abort("The model did not converge; no reliability coefficients are computed.",
                   class = "efa_reliability_no_converge")
  }

  std_sol <- suppressWarnings(lavaan::lavInspect(model, what = "std",
                                drop.list.single.group = FALSE))

  # The model-implied correlation matrix of the variables, per group, fetched on first use
  # below and only where a group has residual covariances -- so a fit whose residual
  # covariance matrix is diagonal, which is every fit the coefficients were defined on until
  # now, asks lavaan for nothing more than it did.
  implied_cor <- NULL

  if (is.null(group_names)) {
    group_names <- names(std_sol)
  } else if (length(group_names) != length(std_sol)) {
    cli::cli_abort(
      c("{.arg group_names} does not match the number of groups in the {.cls lavaan} model.",
        "i" = "The model has {length(std_sol)} group{?s}, but {.arg group_names} has {length(group_names)}."),
      class = "efa_reliability_group_names"
    )
  }

  tol <- .Machine$double.eps * 100
  higherorder <- FALSE
  correlated <- FALSE
  few_loadings <- FALSE
  groups <- vector("list", length(std_sol))

  for (i in seq_along(std_sol)) {

    lambda <- std_sol[[i]][["lambda"]]
    theta <- std_sol[[i]][["theta"]]
    psi <- std_sol[[i]][["psi"]]

    if (any(is.na(lambda))) {
      cli::cli_abort("Some loadings are {.val NA} or {.val NaN}; no reliability coefficients are computed.",
                     class = "efa_reliability_na_loadings")
    }

    if (any(diag(theta) <= 0) || any(diag(psi) <= 0)) {
      cli::cli_abort("A Heywood case was detected (a variance of 0 or negative); no reliability coefficients are computed.",
                     class = "efa_reliability_heywood")
    }

    var_names_i <- rownames(lambda)

    # The uniquenesses below are the diagonal of Theta. A composite's model-implied variance
    # is 1' (Lambda Psi Lambda' + Theta) 1, so a fit with correlated residuals holds variance
    # that the diagonal does not carry (Raykov, 2001). Such a fit takes its composite
    # variances from the model-implied correlation matrix instead.
    #
    # That matrix is read from lavaan, not rebuilt from Theta. The standardized Theta holds
    # residual correlations off the diagonal and residual variances on it, so it is not one
    # matrix in one metric. For a second-order fit the rebuild would also have to add the
    # first-order factor covariances that `beta` carries.
    #
    # With a diagonal Theta the two expressions are the same quantity. This then stays NULL,
    # and the arithmetic of every such solution is unchanged.
    implied_i <- if (any(abs(theta[upper.tri(theta)]) > tol, na.rm = TRUE)) {
      if (is.null(implied_cor)) {
        implied_cor <- lavaan::lavInspect(model, what = "cor.ov",
                                          drop.list.single.group = FALSE)
      }
      implied_cor[[i]]
    }

    # A single factor has no group factors: normalize it to the spec such a solution takes on
    # every other input route, which the core scores like any of them. Marked `single = TRUE`
    # so the front-ends can report the coefficients it defines.
    if (ncol(lambda) == 1) {
      groups[[i]] <- c(
        # Carried beside the spec rather than in its `fac_names`, which the core reads to
        # build the group-factor rows this spec has none of.
        list(single = TRUE, fac_label = colnames(lambda)),
        .rel_single_factor_spec(list(g_load = lambda[, 1],
                                     s_load = lambda[, 0, drop = FALSE],
                                     u2 = diag(theta), cormat = NULL,
                                     implied = implied_i,
                                     var_names = var_names_i)))
      next
    }

    col_names <- colnames(lambda)

    # Detect the model type once (all groups share the fixed-zero structure).
    if (i == 1) {

      # A second-order model routes the covariances of its first-order factors through
      # `beta`, which none of the other supported structures has. Read from `beta` rather
      # than from an all-zero general-factor column, because that column cannot be found
      # at all when the general factor is misnamed -- and such a fit would then pass the
      # simple-structure test below and be scored as a set of correlated factors, whose
      # `psi` is here the residual covariance matrix of the first-order factors rather
      # than their correlation matrix.
      beta <- std_sol[[i]][["beta"]]
      higherorder <- !is.null(beta) && any(abs(beta) > tol, na.rm = TRUE)

      # A variable loading on two or more factors is what a general factor over and above
      # the group factors looks like in a loading matrix, and is the structure a bifactor
      # model is recognized by. Without one -- and without the second-order structure
      # above -- there is no general factor whose variance a composite could be
      # decomposed into, and the fit is an ordinary set of correlated factors. Decided
      # from the structure rather than from `g_name`, which says nothing on its own: a
      # factor may be named "g" without being a general factor, and a general factor need
      # not be named "g".
      bi_check <- lambda
      bi_check[abs(bi_check) > tol] <- 1
      correlated <- !higherorder && all(rowSums(bi_check) < 2)

      if (isTRUE(higherorder)) {
        if (sum(colSums(beta) > 0) > 1) {
          cli::cli_abort("The higher-order model has more than two latent strata or more than one second-order factor; only second-order models with one second-order factor are supported.",
                         class = "efa_reliability_higher_order")
        }
      } else if (!correlated) {
        # A genuine bifactor has each item loading on the general and a group
        # factor; flag the borderline case where some item loads on fewer than
        # two factors so the front-end can warn the user (it is still scored).
        few_loadings <- !all(rowSums(bi_check) > 1)
      }
    }

    # Both structures with a general factor locate it by name; a correlated-factors
    # solution has none to find, so it needs no name and reads none.
    if (!correlated && !any(col_names %in% g_name)) {
      cli::cli_abort(
        c("Could not find the specified general-factor name in the lavaan solution.",
          "i" = "Please check the spelling.",
          "i" = if (!higherorder) {
            "Some variables load on two or more factors, so the fit is read as a bifactor solution; a correlated-factors solution, which needs no general factor, has each variable loading on one factor only."
          }),
        class = "efa_reliability_g_name"
      )
    }

    if (isTRUE(correlated)) {
      # An ordinary set of correlated factors: no general factor, the loadings are the
      # oblique pattern, and the standardized `psi` is the factor correlation matrix
      # belonging to it. The core reads the pair as the model-implied common variance
      # Lambda Psi Lambda', exactly as it does for an oblique `efa_fit()` solution; the
      # front-end drops the coefficients such a solution does not define.
      # `psi` is indexed by name rather than taken whole, as the second-order branch below
      # does: the core reads it as the correlations of the columns of `s_load` in their
      # order, which the two matrices' shared latent ordering supplies but nothing checks.
      # Unlabelled columns, which a fitted model does not have, keep the positional
      # reading rather than subscripting `psi` down to nothing.
      Phi_i <- if (is.null(col_names)) psi else psi[col_names, col_names, drop = FALSE]
      groups[[i]] <- list(single = FALSE, g_load = rep(0, nrow(lambda)),
                          s_load = lambda, u2 = diag(theta), Phi = Phi_i,
                          map = abs(lambda) > tol, cormat = NULL,
                          implied = implied_i,
                          var_names = var_names_i, fac_names = col_names)
      next
    }

    # The coefficients decompose a composite's variance into a general part and one part
    # per group factor, which requires the latent variables to be uncorrelated: a
    # bifactor model is fitted with orthogonal factors, and a second-order model routes
    # the first-order covariances through `beta`, leaving `psi` diagonal either way.
    # `lavaan::cfa()` does not impose that by default, so a "bifactor" model fitted
    # without `orthogonal = TRUE` returns correlated factors whose covariances the
    # coefficients below would silently drop, as does a second-order model with a freed
    # residual covariance between first-order factors. Reject either rather than report
    # a decomposition of a model that does not admit one. An NA here is a malformed
    # solution rather than a correlation, and is left to the checks above.
    if (any(abs(psi[upper.tri(psi)]) > tol, na.rm = TRUE)) {
      cli::cli_abort(
        c("The factors of the {.cls lavaan} solution are correlated; the coefficients need uncorrelated factors.",
          "i" = "A bifactor model needs {.code orthogonal = TRUE}; a second-order model needs no freed covariances between the first-order factors."),
        class = "efa_reliability_correlated_factors"
      )
    }

    col_names <- col_names[!col_names %in% g_name]

    if (isTRUE(higherorder)) {
      # Schmid-Leiman the second-order solution: direct general-factor loadings from
      # the first-order loadings times the second-order (beta) loadings, and direct
      # group-factor loadings via .sl_group_loadings. The general column is computed
      # from the original first-order loadings before they are overwritten.
      lambda[, g_name] <- lambda[, col_names] %*% std_sol[[i]][["beta"]][col_names, g_name]
      lambda[, col_names] <- .sl_group_loadings(lambda[, col_names], psi, col_names)
    }

    s_load <- lambda[, col_names, drop = FALSE]

    groups[[i]] <- list(single = FALSE, g_load = lambda[, g_name], s_load = s_load,
                        u2 = diag(theta), map = abs(s_load) > tol, cormat = NULL,
                        implied = implied_i,
                        var_names = var_names_i, fac_names = col_names)

  }

  list(groups = groups, group_names = group_names, variance = "sums_load",
       higher_order = higherorder, few_loadings = few_loadings,
       correlated = correlated)

}

# Blank the cells of a computed coefficient matrix that a correlated-factors solution
# (one with no general factor) does not define, so the result builder omits them as it
# omits any other undefined coefficient.
#
# The general-factor decomposition is not identified without a general factor: omega
# hierarchical and ECV are structurally zero, and PUC ("percent of uncontaminated
# correlations") presupposes a general factor for the cross-factor correlations to be
# uncontaminated of. On the whole-scale row -- still labelled "g" here, and relabelled
# "total" by .reliability_result() -- the omega subscale and H index are
# further artifacts of the synthetic all-zero general-factor column rather than
# coefficients (the H of an all-zero loading vector is 0, and the whole-scale subscale
# omega does not partition the composite without a general factor). What remains is what
# such a solution does define: whole-scale omega total and alpha, and each group factor's
# congeneric omega, H, and alpha, which stay on their own rows.
.rel_drop_general <- function(x) {
  x[, c("hier", "ECV", "PUC")] <- NA
  x["g", c("sub", "H")] <- NA
  x
}

# The canonical spec of a solution with exactly one factor, or NULL when `spec` does not
# describe one.
#
# One factor is one factor whichever slot it arrived in. An input that names it the general
# factor (a single-factor `lavaan` fit, a one-column bifactor matrix) and one that names it
# the only group factor (a one-factor `efa_fit()` solution, manual components with a zero
# `g_load`) are the same model, so both are rewritten the same way -- the loadings as
# `g_load`, no group-factor columns -- and every route reaches the same coefficients rather
# than one reading of the same solution per input format. Any `Phi` goes with them: the
# correlation matrix of a single factor can only be the 1 x 1 identity, which .rel_check_phi()
# has already established and which says nothing about the solution. The correlation matrix
# `cormat` and the model-implied one `implied` are carried over untouched: both describe the
# variables, which the rewrite does not change.
#
# `fac_names` is emptied here only because the core builds the row labels as
# c("g", fac_names) and this spec has one row; the name the input gave the factor is applied
# to that row afterwards, by .rel_drop_single_factor(), from .rel_single_factor_label().
#
# The core scores the result as it scores every other spec -- its `seq_len(k) + 1` indexing
# and its empty-map guards already cover a spec with no group factors -- so a single factor
# needs no arithmetic of its own. What such a solution does not define is dropped afterwards,
# by .rel_drop_single_factor().
.rel_single_factor_spec <- function(spec) {

  s_load <- as.matrix(spec$s_load)
  n_g <- if (isTRUE(all(spec$g_load == 0))) 0L else 1L
  if (n_g + ncol(s_load) != 1L) return(NULL)

  g_load <- if (n_g == 1L) spec$g_load else s_load[, 1]
  # No group factors, in the p x 0 shape the core reads as none.
  none <- matrix(numeric(0), nrow = length(g_load), ncol = 0)

  list(g_load = g_load, s_load = none, u2 = spec$u2, map = none, Phi = NULL,
       cormat = spec$cormat, implied = spec$implied, var_names = spec$var_names,
       fac_names = character(0))

}

# The row label of a single factor: the name the input gives it, or "F1" where the input gives
# none. A factor supplied as the only group factor is named in `fac_names` -- from the
# solution's loading columns, or from the user's own argument -- and keeps that name, as does
# a `lavaan` factor, which the model syntax always names.
#
# The fallback is the default first-factor label rather than the general-factor "g" that the
# multi-factor solutions use on this row. A single factor is the whole model, not the general
# factor of a bifactor or hierarchical one, which is what "g" would state; a factor supplied
# as the general factor of a one-column matrix carries no name saying otherwise, so it takes
# the neutral label too. That also leaves every route agreeing on the label, and not only on
# the coefficients, wherever the input names the factor nothing.
#
# A name is a name: an empty or missing one is not, and neither is the bare column position
# the manual route falls back to when the caller names no factors (`seq_len(ncol(s_load))`,
# which is 1 for a single factor).
.rel_single_factor_label <- function(fac_names) {
  if (length(fac_names) != 1L || is.na(fac_names) || is.numeric(fac_names) ||
      !nzchar(as.character(fac_names))) {
    return("F1")
  }
  as.character(fac_names)
}

# Blank the cells of a computed coefficient matrix that a single-factor solution does not
# define, so the result builder omits them as it omits any other undefined coefficient, and
# label its one row with the name the factor came in under (see .rel_single_factor_label).
#
# What one factor leaves is omega total, standardized alpha, and the H index. Alpha assumes
# essentially tau-equivalent items, which is nested in a one-factor model, so this is the one
# solution for which reporting it is defensible rather than merely computable. The others are
# not withheld for want of an estimate but because each would state something the solution
# cannot: omega subscale is the variance due to the group factors, of which there are none;
# omega hierarchical is the same quantity as omega total here, the one factor accounting for
# all of the common variance, and two columns holding the same number invite a reader to
# compare a number with itself; and the ECV and the PUC are 1 by construction, which reads as
# evidence of unidimensionality rather than as the arithmetic of a model with one factor.
.rel_drop_single_factor <- function(x, fac_names = NULL) {
  x[, c("hier", "sub", "ECV", "PUC")] <- NA
  rownames(x) <- .rel_single_factor_label(fac_names)
  x
}

# The single-factor note, stated once for both efa_reliability() paths that reach it (the
# lavaan front-end and the shared spec route), so the two cannot drift apart.
.rel_inform_single_factor <- function() {
  cli::cli_inform(
    c("i" = "The solution has a single factor; omega total, alpha, and the H index are returned.",
      "i" = "Omega subscale needs group factors, omega hierarchical is the same quantity as omega total here, and the ECV and the PUC are 1 by construction, so they are omitted."),
    class = "efa_reliability_single_factor"
  )
}

# Adapter: normalize an oblique EFA() object to the reliability spec, treating it as
# the correlated-factors model it is. A correlated-factors solution identifies
# whole-scale omega total and each factor's congeneric omega/H, but not the
# hierarchical/bifactor indices (omega hierarchical, ECV, PUC), which require a
# general factor. Rather than manufacture one by a Schmid-Leiman transformation --
# which is underidentified for fewer than three factors, imposes proportionality
# constraints, and biases the general loadings under the cross-loadings EFA solutions
# almost always have (Flora, 2020, Adv. Methods Pract. Psychol. Sci.; Mansolf &
# Reise, 2016, Multivariate Behav. Res.) -- the spec carries a zero general factor,
# the oblique pattern as the group loadings, and Phi. The core then returns the
# whole-scale omega total as 1' L Phi L' 1 / 1'R1, each factor's omega total as the
# Phi-aware common variance of its composite over that composite's observed variance,
# each factor's congeneric omega subscale and H, and omega hierarchical / ECV / PUC 0.
#
# A one-factor solution is the exception to the oblique requirement below, having nothing to
# rotate: `efa_fit()` returns it unrotated, and where a rotation was asked for anyway leaves
# `rot_loadings` a copy of the unrotated loadings and no factor intercorrelations. It is read
# from `unrot_loadings`, which such an object always carries, under the 1 x 1 identity its
# single factor takes -- so the same expressions serve it -- and the front-end then rewrites
# the one-column spec into the single-factor one (.rel_single_factor_spec), as it does for
# every other route that can carry one factor. Refusing it for want of an oblique rotation
# would be the one piece of advice a one-factor solution cannot follow.
.rel_adapt_efa <- function(model, factor_corres = NULL, type = "psych",
                           cormat = NULL, fac_names = NULL) {

  L_unrot <- unclass(model$unrot_loadings)
  single <- NCOL(L_unrot) == 1L

  if (!single && !("Phi" %in% names(model))) {
    cli::cli_abort(
      c("{.arg model} is not an oblique EFA solution.",
        "i" = "Reliability from an EFA needs correlated factors; refit with an oblique rotation."),
      class = "efa_reliability_not_oblique"
    )
  }

  L1 <- if (single) L_unrot else unclass(model$rot_loadings)
  Phi <- if (single) diag(1) else model$Phi

  # Order the factor columns by number, as SL() does, for stable labels. Only some
  # rotations label their loading columns, so the order falls back to the columns'
  # own order when the labels are absent or carry no factor number.
  n_order <- .sl_factor_order(colnames(L1), ncol(L1))
  s_load <- L1[, n_order, drop = FALSE]
  Phi <- Phi[n_order, n_order, drop = FALSE]

  # Model-implied common variance L Phi L'; its diagonal gives the communalities.
  common <- s_load %*% Phi %*% t(s_load)
  u2 <- 1 - diag(common)

  if (is.null(cormat)) {
    cormat <- if (is.matrix(model$orig_R)) model$orig_R
              else common + diag(u2, nrow = length(u2))
  } else {
    cormat <- .rel_check_cormat(cormat, rownames(s_load), nrow(s_load))
  }

  if (is.null(fac_names)) {
    fac_names <- colnames(s_load)
    # Unlabelled loading columns still need one row label per group factor.
    if (is.null(fac_names)) fac_names <- paste0("F", seq_len(ncol(s_load)))
  }

  list(g_load = rep(0, nrow(s_load)), s_load = s_load, u2 = u2,
       map = .rel_map(s_load, factor_corres, type), Phi = Phi,
       cormat = cormat, var_names = rownames(s_load), fac_names = fac_names)

}

# Adapter: normalize a raw bifactor loading matrix (general factor in the first
# column, group factors in the rest) to the reliability spec. Uniquenesses default to
# the orthogonal-bifactor communalities (1 - rowSums(L^2)); the correspondence map
# defaults to the nonzero group-loading pattern (overridable via factor_corres); and,
# when no observed cormat is supplied, the model-implied L L' + diag(u2) is used so
# both variance conventions apply.
.rel_adapt_bifactor <- function(loadings, factor_corres = NULL, u2 = NULL,
                                cormat = NULL, fac_names = NULL) {

  loadings <- as.matrix(loadings)
  g_load <- loadings[, 1]
  s_load <- loadings[, -1, drop = FALSE]

  # Derived uniquenesses complete the loadings to unit variance by construction; only ones
  # the caller supplied are worth checking against them.
  if (is.null(u2)) {
    u2 <- 1 - rowSums(loadings^2)
  } else {
    .rel_check_u2(rowSums(loadings^2), u2)
  }

  if (is.null(factor_corres)) {
    factor_corres <- abs(s_load) > .Machine$double.eps * 100
  } else {
    # This adapter does not route through .rel_map(), so a supplied map gets the same
    # dimension check and plausibility check here. Only efa_reliability() reaches this
    # adapter, so the map is always its `factor_map`.
    .rel_assert_map_dim(factor_corres, s_load, arg = "factor_map")
    .rel_check_map(s_load, factor_corres, arg = "factor_map")
  }

  if (is.null(cormat)) {
    cormat <- loadings %*% t(loadings) + diag(u2, nrow = length(u2))
  } else {
    # The loadings' own row names, so an unlabelled matrix is matched by position
    # rather than against the V1, V2, ... fallback labels below.
    cormat <- .rel_check_cormat(cormat, rownames(loadings), nrow(loadings))
  }

  var_names <- rownames(loadings)
  if (is.null(var_names)) var_names <- paste0("V", seq_len(nrow(loadings)))

  if (is.null(fac_names)) {
    # A matrix with no group-factor columns holds a single factor, and the only place the
    # matrix names it is the general column. Take the name from there, so the label the
    # front-end puts on that solution's one row is the one the input gave it -- the same
    # label the same solution gets through the components.
    fac_names <- if (ncol(s_load) == 0L) colnames(loadings) else colnames(s_load)
    if (is.null(fac_names)) fac_names <- seq_len(ncol(s_load))
  }

  list(g_load = g_load, s_load = s_load, u2 = u2, map = factor_corres,
       cormat = cormat, var_names = var_names, fac_names = fac_names)

}

# Canonical map from a `.reliability_core()` output column to its public identity in an
# `efa_reliability` result: the long-format coefficient name, its kind (a reliability
# coefficient versus a common-variance / bifactor index), and the compact label used in
# the printed tables. Columns the core can also produce but that a reliability result does
# not surface (CR, AVE) are intentionally absent, so they never enter the output. The row
# order fixes the coefficient order in the tidy result and the column order in the print.
.reliability_registry <- function() {
  data.frame(
    core        = c("tot", "hier", "sub", "alpha", "H", "ECV", "PUC"),
    coefficient = c("omega_total", "omega_hierarchical", "omega_subscale",
                    "alpha", "H", "ECV", "PUC"),
    kind        = c("reliability", "reliability", "reliability", "reliability",
                    "reliability", "common_variance", "common_variance"),
    label       = c("tot", "hier", "sub", "alpha", "H", "ECV", "PUC"),
    stringsAsFactors = FALSE
  )
}

# Shape one or more computed coefficient matrices into the tidy `efa_reliability` result:
# a long data.frame `{coefficient, level, factor, group, value}` with a `settings`
# attribute, a `kind` attribute (each surfaced coefficient tagged reliability vs
# common-variance), and class `efa_reliability`. `x` is a single coefficient matrix (one
# unnamed group) or a named list of them (one entry per group); each matrix has factor
# rows (the general factor `"g"` first, then the group factors) and columns from the
# `.reliability_core()` menu. Only registry columns are carried, so the same builder serves
# a full multi-factor matrix and a single-factor `g`-only matrix alike; NA cells (a
# structurally undefined index such as ECV on a group row, or a Heywood / empty-factor
# coefficient the core has already warned about) are dropped, so the result holds only
# realized coefficients. Purely a reshaping helper -- it computes nothing.
.reliability_result <- function(x, settings = NULL) {

  if (is.list(x)) {
    # A list is always a multigroup result: every group gets a distinct label so the
    # print renders one block per group. Unnamed or blank entries fall back to their
    # position -- NA is reserved for the single, ungrouped matrix case below.
    groups <- x
    group_names <- names(x)
    if (is.null(group_names)) group_names <- rep("", length(x))
    blank <- !nzchar(group_names)
    group_names[blank] <- as.character(seq_along(x))[blank]
  } else {
    groups <- list(x)
    group_names <- NA_character_
  }

  reg <- .reliability_registry()

  parts <- lapply(seq_along(groups), function(g) {
    mat <- as.matrix(groups[[g]])
    factors <- rownames(mat)
    if (is.null(factors)) factors <- paste0("F", seq_len(nrow(mat)))
    keep <- reg[reg$core %in% colnames(mat), , drop = FALSE]

    # A regular (coefficient x factor) grid: each kept coefficient contributes one value
    # per factor. Built in one shot -- the columns of `mat[, keep$core]` unroll
    # column-major, matching the coefficient-major row order -- so a matrix with no
    # surfaced columns yields a 0-row frame rather than a NULL that would break the
    # assembly below.
    # The general factor is the first row of every matrix the core builds, whatever that row
    # is labelled: it is "g" for a solution with group factors, and the factor's own name for
    # a single-factor one, which reports that factor and no group factors at all. Read by
    # position rather than by that label, which also keeps a group factor a user happens to
    # name "g" at level "group". Vectorized over the row index so a matrix with no rows
    # yields no levels, matching the empty coefficient and factor columns below.
    row_level <- ifelse(seq_along(factors) == 1L, "general", "group")

    # A correlated-factors solution has no general factor to put in that row. The core
    # still builds it, and it holds the coefficients of the composite of every variable,
    # which is what a zero general loading leaves it describing. Label the row for what it
    # is: called "g" at level "general" it states a general factor the solution does not
    # define, and it would carry the label of a group factor a user names "g" as well.
    # Positional for the same reason the level is.
    if (isTRUE(settings$no_general) && length(factors) > 0L) {
      factors[1L] <- "total"
      row_level[1L] <- "total"
    }

    data.frame(
      coefficient = rep(keep$coefficient, each = length(factors)),
      level = rep(row_level, times = nrow(keep)),
      factor = rep(factors, times = nrow(keep)),
      group = rep(group_names[g], length(factors) * nrow(keep)),
      value = as.numeric(mat[, keep$core]),
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, parts)
  out <- out[!is.na(out$value), , drop = FALSE]
  rownames(out) <- NULL

  kind <- stats::setNames(reg$kind, reg$coefficient)[unique(out$coefficient)]

  attr(out, "settings") <- settings
  attr(out, "kind") <- kind
  class(out) <- c("efa_reliability", "data.frame")
  out

}

# Flexible omega function (e.g. to use with loadings obtained by MacOrtho)------
.OMEGA_FLEX <- function(model = NULL, type = c("EFAtools", "psych"),
                        factor_corres = NULL,
                        var_names = NULL, fac_names = NULL, g_load = NULL,
                        s_load = NULL, u2 = NULL, cormat = NULL, pattern = NULL,
                        Phi = NULL, variance = c("correlation", "sums_load"),
                        add_ind = TRUE){

  if(inherits(model, "schmid")){

    pattern <- model$oblique
    Phi <- model$phi

    model <-  model$sl
    s_load_names <- setdiff(colnames(model[, -1]),
                            c("h2", "u2", "p2", "com"))

    var_names <- rownames(model)
    g_load <- model[, 1]
    s_load <- model[, s_load_names]
    u2 <- model[, "u2"]

 } else if(inherits(model, "SL")){

    # Honour a user-supplied cormat; only fall back to the SL object's stored
    # correlation matrix when none was given. Flexible-input SL objects store
    # orig_R = NA, in which case cormat stays NULL and the checks below apply.
    if(is.null(cormat) && is.matrix(model$orig_R)){
      cormat <- model$orig_R
    }

    model <-  model$sl
    var_names <- rownames(model)
    g_load <- model[, 1]
    s_load <- model[, 2:(ncol(model) - 2)]
    u2 <- model[, "u2"]

  }

  # Same general + group factor labels regardless of how s_load was obtained.
  factor_names <- c("g", seq_len(ncol(s_load)))

    if(variance == "correlation"){

      if(is.null(cormat)){

        if(is.null(Phi) | is.null(pattern)) {
          cli::cli_abort(
            c("Specify either {.arg cormat}, or {.arg Phi} and {.arg pattern}.",
              "i" = "Alternatively, set {.code variance = \"sums_load\"}."),
            class = "efa_omega_need_cormat"
          )

        } else {

          # Create the correlation matrix from the pattern coefficients and factor
          # intercorrelations
          cormat <- psych::factor.model(f = pattern, Phi = Phi, U2 = FALSE)

        }

      } else {

        # Check if it is a correlation matrix
        if(!.is_cormat(cormat)) {

          cli::cli_abort(
            c("{.arg cormat} is not a correlation matrix.",
              "i" = "Check the {.arg cormat} input, supply {.arg Phi} and {.arg pattern} instead, or set {.code variance = \"sums_load\"}."),
            class = "efa_omega_not_cormat"
          )

        }

      }

    }

  # Check if input to factor_corres is correct (g_load is a vector here, so the
  # item count is its length, equivalently nrow(s_load)).
  checkmate::assert_matrix(factor_corres, null.ok = TRUE, nrows = length(g_load),
                           ncols = ncol(s_load))

  # Create an input dataframe
  input <- data.frame(g_load, s_load)
  colnames(input) <- factor_names
  rownames(input) <- var_names

  if(type == "EFAtools" & is.null(factor_corres)){

    cli::cli_abort("Specify {.arg factor_corres}, or set {.code type = \"psych\"} to derive variable-to-factor correspondences from the highest group-factor loading per variable.",
                   class = "efa_omega_need_corres")

  }

  if(type == "psych"){

    if(variance != "correlation"){

      cli::cli_warn(
        c("{.arg variance} is specified; the value {.val {variance}} is used.",
          "i" = "Results may differ from the specified {.arg type}."),
        class = "efa_omega_variance_override"
      )
      }

    if(is.null(factor_corres)){

      factor_corres <- matrix(0, nrow = nrow(s_load), ncol= ncol(s_load))

      for(i in seq_len(nrow(s_load))){

        # seq_len(k) + 1, not 2:(k + 1): the latter counts down to c(2, 1) with no group
        # factors and selects the general column, as .reliability_core() documents.
        factor_corres[i, which.max(abs(input[i, seq_len(ncol(s_load)) + 1L]))] <- 1

      }



    } else {

      cli::cli_warn(
        c("{.arg factor_corres} is specified; the supplied variable-to-factor correspondences are used.",
          "i" = "To compute correspondences as in psych, leave {.code factor_corres = NULL}."),
        class = "efa_omega_corres_override"
      )
    }
  }

  # Group-factor row labels: user-supplied names, else the factor names carried by a
  # model object, else integer positions for manually specified components.
  fac_names_out <- if (!is.null(fac_names)) {
    fac_names
  } else if (is.null(model)) {
    seq_len(ncol(s_load))
  } else {
    colnames(model)[seq_len(ncol(s_load)) + 1L]
  }

  # Hand the normalized components to the reliability engine.
  spec <- list(g_load = g_load, s_load = s_load, u2 = u2, map = factor_corres,
               cormat = cormat, var_names = var_names, fac_names = fac_names_out)

  .reliability_core(spec, variance = variance, add_ind = add_ind,
                    arg = "factor_corres")

}


# Reshape a lavaan reliability solution into OMEGA's legacy output ------------
#
# OMEGA's lavaan path is a thin front-end over the shared reliability machinery:
# the lavaan adapter normalizes the fitted model to a per-group spec (Schmid-
# Leiman transforming a second-order model) and the reliability core scores every
# group, single-factor ones included. This function only reassembles those results
# into OMEGA's historical shapes -- a coefficient matrix per group (unwrapped for a
# single group), a named list for several groups, and a named c(Omega, H) vector
# (or a bare omega) for a single factor -- and emits the user-facing notes about the
# model structure. All reliability math lives in .reliability_core.
#
# A single factor defines standardized alpha as well, which the core computes and
# efa_reliability() surfaces. OMEGA's single-factor output is a named c(Omega, H) vector
# rather than a coefficient matrix, so reporting alpha would change the shape of a
# superseded function's return value and not only its coefficient menu; it keeps the two
# coefficients it has always returned.
.OMEGA_LAVAAN <- function(model = NULL, g_name = "g", group_names = NULL,
                          add_ind = TRUE){

  adapt <- .rel_adapt_lavaan(model, g_name = g_name, group_names = group_names)

  # OMEGA's input is a Schmid-Leiman, bifactor, second-order, or single-factor solution,
  # each of which has a general factor, and its wide per-factor output reports that
  # factor's omega hierarchical, ECV, and PUC. A correlated-factors fit defines none of
  # them, so it is refused here rather than scored into a table whose general-factor
  # columns would be zeros. efa_reliability() scores it and omits those coefficients.
  if (isTRUE(adapt$correlated)) {
    cli::cli_abort(
      c("The lavaan input is invalid; no reliability coefficients are computed.",
        "i" = "Provide a bifactor model, a second-order model, or a single-factor model.",
        "i" = "To score a correlated-factors solution, use {.fn efa_reliability}."),
      class = "efa_reliability_invalid_lavaan"
    )
  }

  # A second-order general factor was Schmid-Leiman transformed by the adapter.
  if (isTRUE(adapt$higher_order)) {
    cli::cli_inform(
      c("i" = "The specified general factor is a second-order factor; omegas are computed on the Schmid-Leiman transformed second-order solution."),
      class = "efa_omega_g_second_order"
    )
  }

  # A supplied "bifactor" in which some item loads on fewer than two factors.
  if (isTRUE(adapt$few_loadings)) {
    cli::cli_inform(
      c("i" = "Some variables have fewer than two loadings; did you enter a bifactor model? Provide a bifactor model, a second-order model, or a single-factor model."),
      class = "efa_omega_few_loadings"
    )
  }

  omegas <- vector("list", length(adapt$groups))
  informed_single <- FALSE

  for (i in seq_along(adapt$groups)) {

    grp <- adapt$groups[[i]]

    if (isTRUE(grp$single)) {

      # A single factor: omega total (and, with add_ind, the H index) only.
      if (!informed_single) {
        msg <- if (isTRUE(add_ind)) {
          "The model contained a single factor; only omega total and the H index are returned."
        } else {
          "The model contained a single factor; only omega total is returned."
        }
        cli::cli_inform(c("i" = msg), class = "efa_omega_single_factor")
        informed_single <- TRUE
      }

      # Scored through the core on the spec the adapter normalized it to, as every other
      # group is; only the two coefficients OMEGA reports for a single factor are kept.
      sf <- unclass(.reliability_core(grp, "sums_load", add_ind = add_ind,
                                      arg = "factor_corres"))
      omegas[[i]] <- if (isTRUE(add_ind)) {
        stats::setNames(c(sf["g", "tot"], sf["g", "H"]), c("Omega", "H"))
      } else {
        sf["g", "tot"]
      }

    } else {

      # A multi-factor (bifactor / Schmid-Leiman) group: score through the core,
      # storing the unclassed matrix -- the OMEGA class is attached to the whole
      # output below, matching the historical list-of-matrices shape. The core
      # labels the general-factor row "g"; restore the user's general-factor name.
      mat <- unclass(.reliability_core(grp, "sums_load", add_ind = add_ind,
                                       arg = "factor_corres"))
      rownames(mat)[1] <- g_name
      omegas[[i]] <- mat

    }
  }

  if (length(omegas) > 1) {
    names(omegas) <- adapt$group_names
  } else {
    omegas <- omegas[[1]]
  }

  class(omegas) <- "OMEGA"
  omegas

}

## Tests for .OMEGA_LAVAAN --------

if (requireNamespace("lavaan", quietly = TRUE)) {
lav_mod_1 <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
              F2 =~ V7 + V8 + V9 + V10 + V11 + V12
              F3 =~ V13 + V14 + V15 + V16 + V17 + V18
              g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                   V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_1 <- suppressWarnings(lavaan::cfa(lav_mod_1,
                                          sample.cov = test_models$baseline$cormat,
                                          sample.nobs = 500, estimator = "ml",
                                          orthogonal = TRUE))
om_lav_bi_add <- .OMEGA_LAVAAN(lav_fit_1, g_name = "g")
om_lav_bi_noadd <- .OMEGA_LAVAAN(lav_fit_1, g_name = "g", add_ind = FALSE)

lav_mod_2 <- 'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                    V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_2 <- suppressWarnings(lavaan::cfa(lav_mod_2,
                                         sample.cov = test_models$baseline$cormat,
                                         sample.nobs = 500, estimator = "ml",
                                         orthogonal = TRUE))
om_lav_1_add <- suppressMessages(.OMEGA_LAVAAN(lav_fit_2))
om_lav_1_noadd <- suppressMessages(.OMEGA_LAVAAN(lav_fit_2, add_ind = FALSE))

lav_fit_3 <- suppressWarnings(lavaan::cfa(lav_mod_1, sample.cov =
                                            list(test_models$baseline$cormat,
                                                 test_models$baseline$cormat),
                                          sample.nobs = c(500, 500),
                                          estimator = "ml", orthogonal = TRUE))
om_lav_gr_add <- .OMEGA_LAVAAN(lav_fit_3, g_name = "g", group_names = c("Some",
                                                                    "Others"))
om_lav_gr_noadd <- .OMEGA_LAVAAN(lav_fit_3, g_name = "g", add_ind = FALSE,
                                 group_names = c("Some", "Others"))

lav_fit_4 <- suppressWarnings(lavaan::cfa(lav_mod_2, sample.cov =
                                            list(test_models$baseline$cormat,
                                                 test_models$baseline$cormat),
                                          sample.nobs = c(500, 500),
                                          estimator = "ml", orthogonal = TRUE))
om_lav_1_gr_add <- suppressMessages(.OMEGA_LAVAAN(lav_fit_4,
                                                  group_names = c("Some",
                                                                  "Others")))
om_lav_1_gr_noadd <- suppressMessages(.OMEGA_LAVAAN(lav_fit_4, add_ind = FALSE,
                                                    group_names = c("Some",
                                                                    "Others")))

lav_mod_ho_1 <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18
               g =~ F1 + F2 + F3'
lav_fit_ho_1 <- suppressWarnings(lavaan::cfa(lav_mod_ho_1,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))
om_lav_ho_1 <- suppressMessages(.OMEGA_LAVAAN(lav_fit_ho_1, g_name = "g"))
}

test_that("output class and dimensions are correct", {
  skip_if_not_installed("lavaan")
  expect_s3_class(om_lav_bi_add, "OMEGA")
  expect_s3_class(om_lav_1_add, "OMEGA")
  expect_s3_class(om_lav_ho_1, "OMEGA")
  expect_s3_class(om_lav_gr_add, "OMEGA")
  expect_s3_class(om_lav_1_gr_add, "OMEGA")
  expect_s3_class(om_lav_bi_noadd, "OMEGA")
  expect_s3_class(om_lav_1_noadd, "OMEGA")
  expect_s3_class(om_lav_gr_noadd, "OMEGA")
  expect_s3_class(om_lav_1_gr_noadd, "OMEGA")

  expect_output(str(om_lav_bi_add), "List of 2")
  expect_output(str(om_lav_1_add), "OMEGA")
  expect_output(str(om_lav_ho_1), "List of 2")
  expect_output(str(om_lav_gr_add), "List of 2")
  expect_output(str(om_lav_1_gr_add), "List of 2")
  expect_output(str(om_lav_bi_add), "List of 2")
  expect_output(str(om_lav_1_add), "OMEGA")
  expect_output(str(om_lav_ho_1), "List of 2")
  expect_output(str(om_lav_gr_add), "List of 2")
  expect_output(str(om_lav_1_gr_add), "List of 2")
})

test_that("output is correct (including group names for multiple groups)", {
  skip_if_not_installed("lavaan")
  expect_equal(rowSums(om_lav_bi_add[, 2:3]), om_lav_bi_add[, 1], tolerance = 1e-3)
  expect_equal(om_lav_bi_add[1, 4], 0.849, tolerance = 1e-3)
  expect_equal(om_lav_bi_add[1, 5], 0.672, tolerance = 1e-3)
  expect_equal(om_lav_bi_add[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(unname(om_lav_1_add[1]), 0.868, tolerance = 1e-3)
  expect_equal(unname(om_lav_1_add[2]), 0.869, tolerance = 1e-3)
  expect_equal(rowSums(om_lav_ho_1[, 2:3]), om_lav_ho_1[, 1], tolerance = 1e-3)
  expect_equal(om_lav_ho_1[1, 4], 0.848, tolerance = 1e-3)
  expect_equal(om_lav_ho_1[1, 5], 0.684, tolerance = 1e-3)
  expect_equal(om_lav_ho_1[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(rowSums(om_lav_gr_add$Some[, 2:3]), om_lav_gr_add$Some[, 1],
               tolerance = 1e-3)
  expect_equal(rowSums(om_lav_gr_add$Others[, 2:3]), om_lav_gr_add$Others[, 1],
               tolerance = 1e-3)
})

# Preparations for error tests
if (requireNamespace("lavaan", quietly = TRUE)) {
lav_mod_NA <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6 + V17
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12 + V2
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18 + V10
               g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                    V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_NA <- suppressWarnings(lavaan::cfa(lav_mod_NA,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

lav_mod_inv <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_inv <- suppressWarnings(lavaan::cfa(lav_mod_inv,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))

lav_mod_bi_red <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9 + V10 + V11 + V12
               F3 =~ V13 + V14 + V15 + V16 + V17 + V18
               g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 +
                    V13 + V14 + V15 + V16 + V17 + V18'
lav_fit_bi_red <- suppressWarnings(lavaan::cfa(lav_mod_bi_red,
                                              sample.cov = test_models$baseline$cormat,
                                              sample.nobs = 500, estimator = "ml",
                                              orthogonal = TRUE))

lav_mod_ho_2 <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
               F2 =~ V7 + V8 + V9
               F3 =~ V10 + V11 + V12
               F4 =~ V13 + V14 + V15 + V16 + V17 + V18
               g =~ F1 + F2
               h =~ F3 + F4'
lav_fit_ho_2 <- suppressWarnings(lavaan::cfa(lav_mod_ho_2,
                                           sample.cov = test_models$baseline$cormat,
                                           sample.nobs = 500, estimator = "ml"))
}

test_that("errors are thrown correctly", {
  skip_if_not_installed("lavaan")
  # OMEGA's lavaan path now delegates spec-building (and its validation) to the
  # shared reliability adapter, so the aborts carry the adapter's
  # efa_reliability_* classes; the informational notices remain OMEGA's own
  # efa_omega_* conditions.
  expect_error(.OMEGA_LAVAAN(lav_fit_NA, g_name = "g"), class = "efa_reliability_na_loadings")
  expect_error(.OMEGA_LAVAAN(lav_fit_1, g_name = "fu"), class = "efa_reliability_g_name")
  expect_message(.OMEGA_LAVAAN(lav_fit_2, add_ind = FALSE), class = "efa_omega_single_factor")
  expect_message(.OMEGA_LAVAAN(lav_fit_2), class = "efa_omega_single_factor")
  expect_message(.OMEGA_LAVAAN(lav_fit_ho_1, g_name = "g"), class = "efa_omega_g_second_order")
  expect_error(.OMEGA_LAVAAN(lav_fit_inv, g_name = "F3"), class = "efa_reliability_invalid_lavaan")
  expect_message(.OMEGA_LAVAAN(lav_fit_bi_red, g_name = "g"), class = "efa_omega_few_loadings")
  expect_error(.OMEGA_LAVAAN(lav_fit_ho_2, g_name = "g"), class = "efa_reliability_higher_order")
})

test_that("a bifactor model fitted with correlated factors is rejected", {
  skip_if_not_installed("lavaan")
  # The coefficients split a composite's variance into a general and a per-group-factor
  # part, which needs uncorrelated factors. lavaan::cfa() does not impose that, and the
  # same syntax fitted without orthogonal = TRUE returns factor correlations the
  # coefficients would silently drop -- here up to .24 on the covariance scale, worth
  # .016 on the whole-scale omega total.
  lav_fit_obl <- suppressWarnings(lavaan::cfa(lav_mod_1,
                                              sample.cov = test_models$baseline$cormat,
                                              sample.nobs = 500, estimator = "ml",
                                              orthogonal = FALSE))
  expect_error(.rel_adapt_lavaan(lav_fit_obl, g_name = "g"),
               class = "efa_reliability_correlated_factors")
  expect_error(.OMEGA_LAVAAN(lav_fit_obl, g_name = "g"),
               class = "efa_reliability_correlated_factors")
  expect_error(efa_reliability(lav_fit_obl, g_name = "g"),
               class = "efa_reliability_correlated_factors")
  # The supported shapes are unaffected: an orthogonal bifactor, a second-order model
  # and a single factor all leave psi diagonal.
  expect_no_error(.rel_adapt_lavaan(lav_fit_1, g_name = "g"))
  expect_no_error(suppressMessages(.rel_adapt_lavaan(lav_fit_ho_1, g_name = "g")))
  expect_no_error(.rel_adapt_lavaan(lav_fit_2))
})


## Tests for .OMEGA_FLEX -------

## Use with an output from the SL function, with type EFAtools
efa_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
sl_mod <- SL(efa_mod, type = "EFAtools", method = "PAF")
om_sl_add <- .OMEGA_FLEX(sl_mod, type = "EFAtools",
                     factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                     variance = "correlation")
om_sl_noadd <- .OMEGA_FLEX(sl_mod, type = "EFAtools",
                     factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                     variance = "correlation", add_ind = FALSE)

# Explicit factor names
om_sl_named_add <- .OMEGA_FLEX(sl_mod, type = "EFAtools",
                           fac_names = c("Fa1", "Fa2", "Fa3"),
                           factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                           variance = "correlation")
om_sl_named_noadd <- .OMEGA_FLEX(sl_mod, type = "EFAtools",
                           fac_names = c("Fa1", "Fa2", "Fa3"),
                           factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                           variance = "correlation", add_ind = FALSE)

## Use with an output from the psych::schmid function, with type psych
schmid_mod <- psych::schmid(test_models$baseline$cormat, nfactors = 3,
                            n.obs = 500, fm = "pa", rotate = "Promax")

# Omega also with type "psych"
# Find correlation matrix from phi and pattern matrix from psych::schmid output
om_schmid_1_add <- .OMEGA_FLEX(schmid_mod, type = "psych", variance = "correlation")
om_schmid_1_noadd <- .OMEGA_FLEX(schmid_mod, type = "psych", variance = "correlation",
                                add_ind = FALSE)
# Enter correlation matrix manually
om_schmid_2 <- .OMEGA_FLEX(schmid_mod, type = "psych",
                           cormat = test_models$baseline$cormat,
                           variance = "correlation")

## Manually specify components (here with type "EFAtools")
om_man_1_add <- .OMEGA_FLEX(model = NULL, type = "EFAtools",
                            var_names = rownames(sl_mod$sl),
                  g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                  u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
                  factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                  variance = "correlation")
om_man_1_noadd <- .OMEGA_FLEX(model = NULL, type = "EFAtools",
                              var_names = rownames(sl_mod$sl),
                              g_load = sl_mod$sl[, "g"],
                              s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                              u2 = sl_mod$sl[, "u2"],
                              cormat = test_models$baseline$cormat,
                              factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                              variance = "correlation", add_ind = FALSE)
# Now with other type of variance (model-based instead of based on corrmat)
om_man_2 <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                  g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                  u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
                  factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                  variance = "sums_load")

test_that("output class and dimensions are correct", {
  expect_s3_class(om_sl_add, "OMEGA")
  expect_s3_class(om_sl_named_add, "OMEGA")
  expect_s3_class(om_schmid_1_add, "OMEGA")
  expect_s3_class(om_schmid_2, "OMEGA")
  expect_s3_class(om_man_1_add, "OMEGA")
  expect_s3_class(om_man_2, "OMEGA")
  expect_s3_class(om_sl_noadd, "OMEGA")
  expect_s3_class(om_sl_named_noadd, "OMEGA")
  expect_s3_class(om_schmid_1_noadd, "OMEGA")
  expect_s3_class(om_man_1_noadd, "OMEGA")

  expect_output(str(om_sl_add), "List of 2")
  expect_output(str(om_sl_named_add), "List of 2")
  expect_output(str(om_schmid_1_add), "List of 2")
  expect_output(str(om_schmid_2), "List of 2")
  expect_output(str(om_man_1_add), "List of 2")
  expect_output(str(om_man_2), "List of 2")
  expect_output(str(om_sl_noadd), "List of 2")
  expect_output(str(om_sl_named_noadd), "List of 2")
  expect_output(str(om_schmid_1_noadd), "List of 2")
  expect_output(str(om_man_1_noadd), "List of 2")
})

test_that("output is correct", {
  # On a subscale row, omega total counts every factor the composite's variables load
  # on while omega subscale counts only that factor's own contribution, so the three
  # coefficients no longer add up: the gap is what the composite receives from the
  # other group factors through its cross-loadings. Every fixture here has an estimated
  # (dense) group-loading matrix, so that gap is strictly positive -- asserting only
  # `tot >= hier + sub` would pass just as well if the cross terms were dropped again,
  # which makes the gap exactly zero.
  tol <- sqrt(.Machine$double.eps)
  for (om in list(om_sl_add, om_sl_named_add, om_schmid_1_add, om_schmid_2,
                  om_man_1_add, om_man_2)) {
    expect_true(all(om[2:4, 1] - rowSums(om[2:4, 2:3]) > tol))
  }
  # And the gap is exactly the mass the composite draws from the other group factors,
  # 1' L_other L_other' 1 over its observed variance.
  s_all <- unclass(sl_mod$sl[, c("F1", "F2", "F3")])
  map_sl <- s_all >= .2
  for (j in 1:3) {
    mem <- which(map_sl[, j])
    expect_equal(unname(om_sl_add[j + 1, 1] - sum(om_sl_add[j + 1, 2:3])),
                 sum(colSums(s_all[mem, -j, drop = FALSE])^2) /
                   sum(test_models$baseline$cormat[mem, mem]))
  }
  expect_equal(om_sl_add[1, 4], 0.842, tolerance = 1e-3)
  expect_equal(om_sl_add[1, 5], 0.652, tolerance = 1e-3)
  expect_equal(om_sl_add[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(om_sl_named_add[1, 4], 0.842, tolerance = 1e-3)
  expect_equal(om_sl_named_add[1, 5], 0.652, tolerance = 1e-3)
  expect_equal(om_sl_named_add[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(om_schmid_1_add[1, 4], 0.845, tolerance = 1e-3)
  expect_equal(om_schmid_1_add[1, 5], 0.661, tolerance = 1e-3)
  expect_equal(om_schmid_1_add[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(om_schmid_2[1, 4], 0.845, tolerance = 1e-3)
  expect_equal(om_schmid_2[1, 5], 0.661, tolerance = 1e-3)
  expect_equal(om_schmid_2[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(om_man_1_add[1, 4], 0.842, tolerance = 1e-3)
  expect_equal(om_man_1_add[1, 5], 0.652, tolerance = 1e-3)
  expect_equal(om_man_1_add[1, 6], 0.706, tolerance = 1e-3)
  expect_equal(om_man_2[1, 4], 0.842, tolerance = 1e-3)
  expect_equal(om_man_2[1, 5], 0.652, tolerance = 1e-3)
  expect_equal(om_man_2[1, 6], 0.706, tolerance = 1e-3)

  expect_gte(om_sl_add[1, 1] + tol, sum(om_sl_add[1, 2:3]))
  expect_gte(om_sl_named_add[1, 1] + tol, sum(om_sl_named_add[1, 2:3]))
  expect_gte(om_schmid_1_add[1, 1] + tol, sum(om_schmid_1_add[1, 2:3]))
  expect_gte(om_schmid_2[1, 1] + tol, sum(om_schmid_2[1, 2:3]))
  expect_gte(om_man_1_add[1, 1] + tol, sum(om_man_1_add[1, 2:3]))
  expect_gte(om_man_2[1, 1] + tol, sum(om_man_2[1, 2:3]))

  # With variance = "sums_load", the general-factor omega total equals
  # hierarchical + subscale exactly (shared total-variance denominator).
  expect_equal(unname(om_man_2[1, 1]), unname(sum(om_man_2[1, 2:3])))
})

test_that("variance = 'sums_load' whole-scale omega total uses the full loading columns", {
  # A dense group-loading matrix: item 1 is assigned to F1 but also loads on F2.
  # The whole-scale composite contains all items, so its model-implied omega total
  # counts every group loading (full columns), giving 1 - sum(u2)/V, with the
  # general and group variances partitioning it exactly.
  g_load <- rep(0.5, 6)
  s_load <- matrix(0, 6, 2); s_load[1:3, 1] <- 0.5; s_load[4:6, 2] <- 0.5
  s_load[1, 2] <- 0.3
  u2 <- 1 - g_load^2 - rowSums(s_load^2)
  fc <- matrix(0, 6, 2); fc[1:3, 1] <- 1; fc[4:6, 2] <- 1   # item 1 assigned to F1 only

  om <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = paste0("V", 1:6),
                    g_load = g_load, s_load = s_load, u2 = u2,
                    factor_corres = fc, variance = "sums_load")

  full_s <- colSums(s_load)                       # full columns (all items)
  V <- sum(g_load)^2 + sum(full_s^2) + sum(u2)    # model-implied composite variance

  # omega total is the model-implied total, using the full loading columns, and the
  # whole-scale subscale omega is the full group-factor variance (not the
  # assigned-only mass), so tot = hier + sub exactly.
  expect_equal(unname(om[1, "tot"]), 1 - sum(u2) / V)
  expect_equal(unname(om[1, "sub"]), sum(full_s^2) / V)
  expect_equal(unname(om[1, "tot"]), unname(om[1, "hier"] + om[1, "sub"]))
  # the off-assignment loading makes the full mass strictly exceed the assigned one
  expect_gt(sum(full_s^2), sum(colSums(s_load * fc)^2))
})

test_that("type = 'psych' reproduces psych::omega (g omegas and ECV)", {
  skip_on_cran()
  skip_if_not_installed("psych")

  po <- psych::omega(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                     fm = "pa", rotate = "Promax", plot = FALSE)

  # The psych-type path derives the same variable-to-factor correspondences as
  # psych::omega and bases the ECV on the (unzeroed) group loadings, so the ECV and
  # omega hierarchical are the same quantity and match exactly. Omega total is not: it
  # is the model-implied common variance of all variables here and the observed total
  # variance less the unique variances in psych, which agree only to the extent that the
  # model reproduces the correlations. This fixture is a three-factor population matrix
  # fitted with three factors, so the residual mass is about 6e-5 and the tolerance below
  # holds with room to spare -- a worse-fitting fixture would separate the two on purpose.
  expect_equal(unname(om_schmid_1_add[1, "ECV"]), unname(po$ECV), tolerance = 1e-4)
  expect_equal(unname(om_schmid_1_add[1, "tot"]), po$omega.tot, tolerance = 1e-3)
  expect_equal(unname(om_schmid_1_add[1, "hier"]), po$omega_h, tolerance = 1e-3)
})

test_that("errors are thrown correctly", {
  expect_error(.OMEGA_FLEX(schmid_mod, type = "psych",
                           cormat = matrix(rnorm(50), ncol = 5),
                           variance = "correlation"), class = "efa_omega_not_cormat")
  expect_error(.OMEGA_FLEX(model = NULL, type = "EFAtools",
                           var_names = rownames(sl_mod$sl),
                           g_load = sl_mod$sl[, "g"],
                           s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                           u2 = sl_mod$sl[, "u2"],
                           factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                           variance = "correlation"), class = "efa_omega_need_cormat")
  expect_error(.OMEGA_FLEX(model = NULL, type = "EFAtools",
                           var_names = rownames(sl_mod$sl),
                           g_load = sl_mod$sl[, "g"],
                           s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                           u2 = sl_mod$sl[, "u2"], cormat = matrix(rnorm(50), ncol = 5),
                           factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                           variance = "correlation"), class = "efa_omega_not_cormat")
  expect_error(.OMEGA_FLEX(schmid_mod, type = "EFAtools",
                           variance = "correlation"), class = "efa_omega_need_corres")
  expect_warning(.OMEGA_FLEX(schmid_mod, type = "psych",
                             variance = "sums_load"), class = "efa_omega_variance_override")
  expect_warning(.OMEGA_FLEX(schmid_mod, type = "psych",
                             factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2,
                             variance = "correlation"), class = "efa_omega_corres_override")
  # A factor_corres with the wrong number of rows (one fewer than the number of
  # items) must be rejected, not silently used to corrupt the omega outputs.
  expect_error(.OMEGA_FLEX(sl_mod, type = "EFAtools",
                           factor_corres = (sl_mod$sl[, c("F1", "F2", "F3")] >= .2)[-1, ],
                           variance = "correlation"))
})

test_that("group-factor H index uses only the factor's assigned items", {
  # Item 6 has a sizeable loading on F2 but is assigned to F1; that cross-loading
  # must not enter F2's H index (Hancock & Mueller, 2001).
  g_load <- rep(0.5, 6)
  s_load <- matrix(0, 6, 2)
  s_load[1:3, 1] <- 0.6
  s_load[4:5, 2] <- 0.6
  s_load[6, 1] <- 0.5
  s_load[6, 2] <- 0.5
  fc <- matrix(0, 6, 2)
  fc[1:3, 1] <- 1
  fc[6, 1] <- 1
  fc[4:5, 2] <- 1
  u2 <- 1 - g_load^2 - rowSums(s_load^2)
  om <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = paste0("V", 1:6),
                    g_load = g_load, s_load = s_load, u2 = u2,
                    factor_corres = fc, variance = "sums_load")
  s2 <- s_load[fc[, 2] == 1, 2]
  expect_equal(unname(om["2", "H"]), 1 / (1 + 1 / sum(s2^2 / (1 - s2^2))))
})

test_that("PUC counts each contaminated pair once (no double counting)", {
  # Items 3 and 4 belong to BOTH group factors, so the pair {3, 4} is a single
  # contaminated correlation even though it sits within two group factors.
  fc <- matrix(c(1, 1, 1, 1, 0, 0,
                 0, 0, 1, 1, 1, 1), ncol = 2)
  s_load <- 0.4 * fc
  g_load <- rep(0.5, 6)
  u2 <- 1 - g_load^2 - rowSums(s_load^2)
  om <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = paste0("V", 1:6),
                    g_load = g_load, s_load = s_load, u2 = u2,
                    factor_corres = fc, variance = "sums_load")
  # C(6, 2) = 15 pairs; 11 unique contaminated -> PUC = 1 - 11/15.
  expect_equal(unname(om[1, "PUC"]), 1 - 11 / 15)
})

test_that("PUC for an incomplete-g bifactor depends only on the group factors", {
  skip_if_not_installed("lavaan")
  # Bifactor where the general factor does not load on V18; PUC must still be the
  # group-factor-only quantity (Reise et al., 2013): 1 - 3 * C(6,2) / C(18,2).
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g  =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                V13 + V14 + V15 + V16 + V17'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))
  om <- suppressMessages(suppressWarnings(.OMEGA_LAVAAN(fit, g_name = "g")))
  expect_equal(unname(om[1, "PUC"]), 1 - 3 * choose(6, 2) / choose(18, 2))
})

test_that("an inconsistent cormat triggers an out-of-range warning", {
  # Strong general loadings combined with a weakly correlated matrix push
  # omega hierarchical above 1 (and above omega total); this must not pass silently.
  g_load <- rep(0.7, 6)
  s_load <- matrix(0, 6, 2)
  s_load[1:3, 1] <- 0.1
  s_load[4:6, 2] <- 0.1
  u2 <- 1 - g_load^2 - rowSums(s_load^2)
  Rm <- matrix(0.3, 6, 6)
  diag(Rm) <- 1
  rownames(Rm) <- colnames(Rm) <- paste0("V", 1:6)
  expect_warning(
    .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = paste0("V", 1:6),
                g_load = g_load, s_load = s_load, u2 = u2, cormat = Rm,
                factor_corres = s_load > 0, variance = "correlation"),
    class = "efa_omega_out_of_range")

  # And the coefficients stay in range on solutions whose loadings came from the
  # correlation matrix they are scored against. Every omega total is now the composite's
  # model-implied common variance, which is weakly larger than the general-plus-own-column
  # term it replaced, so this is the direction in which the guard could start firing on
  # ordinary input. Asserted on the values rather than on the absence of the warning,
  # because an over-extracted solution legitimately raises the empty-factor one.
  tol_range <- sqrt(.Machine$double.eps)
  for (k in 2:5) {
    mod_k <- suppressWarnings(EFA(test_models$baseline$cormat, N = 500, n_factors = k,
                                  type = "EFAtools", method = "PAF",
                                  rotation = "promax"))
    sl_k <- suppressWarnings(SL(mod_k, type = "psych", method = "PAF"))
    om_k <- suppressWarnings(
      .OMEGA_FLEX(sl_k, type = "psych", cormat = test_models$baseline$cormat,
                  variance = "correlation"))
    expect_true(all(om_k[, c("tot", "hier", "sub")] <= 1 + tol_range, na.rm = TRUE))
    expect_true(all(om_k[, "hier"] <= om_k[, "tot"] + tol_range, na.rm = TRUE))
  }
})

test_that("the out-of-range guard runs in both variance conventions", {
  # A share of variance above 1 is inadmissible whichever total variance it was divided by.
  # One way there in the model-implied convention is an improper solution, whose unique
  # variances sum to something negative for a composite; the Heywood check reports that from
  # the uniquenesses, but only when the indices are computed, and never says that the
  # returned coefficients are themselves out of range.
  g <- rep(0.6, 6)
  s <- matrix(0, 6, 2)
  s[1:3, 1] <- 0.5
  s[4:6, 2] <- 0.5
  spec <- list(g_load = g, s_load = s, u2 = rep(-0.2, 6),
               map = cbind(rep(1:0, each = 3), rep(0:1, each = 3)), cormat = NULL,
               var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  # Each of these solutions raises both conditions, so the one an expectation is not about is
  # muffled to keep it from leaking out of the test.
  muffle <- function(expr, cls) {
    withCallingHandlers(expr,
      warning = function(w) if (inherits(w, cls)) invokeRestart("muffleWarning"))
  }
  expect_warning(om <- muffle(.reliability_core(spec, "sums_load"),
                              "efa_reliability_heywood"),
                 class = "efa_omega_out_of_range")
  expect_gt(unname(om["g", "tot"]), 1)

  # A proper solution with uncorrelated group factors stays silent, as it always has been.
  spec$u2 <- 1 - g^2 - rowSums(s^2)
  expect_no_warning(.reliability_core(spec, "sums_load"),
                    class = "efa_omega_out_of_range")

  # The second route needs no improper solution at all, which is what makes the guard worth
  # running in this convention: omega subscale divides a factor's own squared loading sum by
  # its composite's whole variance, and once the group factors are correlated that sum is not
  # a part of it -- the composite's cross-loadings on the other factor contribute negative
  # common variance and take its total below what its own factor puts in. Every uniqueness
  # here is positive and Phi is positive definite, so nothing else in the function signals it.
  S <- matrix(0, 8, 2)
  S[1:4, 1] <- .85; S[1:4, 2] <- -.30
  S[5:8, 2] <- .85; S[5:8, 1] <- -.30
  Phi <- matrix(c(1, .6, .6, 1), 2, 2)
  u2 <- 1 - diag(S %*% Phi %*% t(S))
  proper <- list(g_load = rep(0, 8), s_load = S, u2 = u2, Phi = Phi,
                 map = cbind(rep(1:0, each = 4), rep(0:1, each = 4)), cormat = NULL,
                 var_names = paste0("V", 1:8), fac_names = c("F1", "F2"))
  expect_true(all(u2 > 0))
  expect_gt(min(eigen(Phi, symmetric = TRUE, only.values = TRUE)$values), 0)

  expect_warning(om <- .reliability_core(proper, "sums_load"),
                 class = "efa_omega_out_of_range")
  expect_gt(unname(om["F1", "sub"]), 1)
  # and the improper-solution report stays silent, so this is the only signal there is
  expect_no_warning(muffle(.reliability_core(proper, "sums_load"),
                           "efa_omega_out_of_range"),
                    class = "efa_reliability_heywood")
})

test_that("a single group-factor column is handled without error", {
  g_load <- rep(0.7, 6)
  s_load <- matrix(0.2, 6, 1)
  u2 <- 1 - g_load^2 - rowSums(s_load^2)
  fc <- matrix(1, 6, 1)
  expect_s3_class(
    .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = paste0("V", 1:6),
                g_load = g_load, s_load = s_load, u2 = u2,
                factor_corres = fc, variance = "sums_load"), "OMEGA")
  Rm <- matrix(0.6, 6, 6)
  diag(Rm) <- 1
  rownames(Rm) <- colnames(Rm) <- paste0("V", 1:6)
  expect_s3_class(
    suppressWarnings(.OMEGA_FLEX(model = NULL, type = "EFAtools",
                var_names = paste0("V", 1:6), g_load = g_load, s_load = s_load,
                u2 = u2, cormat = Rm, factor_corres = fc,
                variance = "correlation")), "OMEGA")
})

test_that("a group factor with no assigned items reports NA coefficients and warns", {
  # type = 'psych' assigns each item to its highest-loading factor; F2 never wins,
  # so its correspondence column is all zero and its coefficients are undefined.
  g <- rep(0.5, 6)
  s <- matrix(c(rep(0.6, 6), rep(0.3, 6)), ncol = 2)
  u2 <- 1 - g^2 - rowSums(s^2)
  # Model-implied (consistent) correlation matrix so only the empty-factor warning
  # fires (no out-of-range, no variance override).
  L <- cbind(g, s)
  Rm <- L %*% t(L); diag(Rm) <- 1
  rownames(Rm) <- colnames(Rm) <- paste0("V", 1:6)
  expect_warning(
    om <- .OMEGA_FLEX(model = NULL, type = "psych", var_names = paste0("V", 1:6),
                      g_load = g, s_load = s, u2 = u2, cormat = Rm,
                      variance = "correlation"),
    class = "efa_omega_empty_factor")
  expect_true(all(is.na(om["2", c("tot", "hier", "sub", "H")])))
})

test_that("PUC membership follows factor_corres, not the loading magnitudes", {
  # Item 2 is assigned to F1 by factor_corres but has a zero loading on F1; it must
  # still count as an F1 member for PUC, consistent with the H index.
  g <- rep(0.5, 4)
  s <- matrix(0, 4, 2); s[1, 1] <- 0.4; s[3, 2] <- 0.4; s[4, 2] <- 0.4
  fc <- matrix(0, 4, 2); fc[1:2, 1] <- 1; fc[3:4, 2] <- 1
  u2 <- 1 - g^2 - rowSums(s^2)
  om <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = paste0("V", 1:4),
                    g_load = g, s_load = s, u2 = u2, factor_corres = fc,
                    variance = "sums_load")
  # F1 = {1, 2}, F2 = {3, 4}: 2 contaminated pairs of C(4, 2) = 6 -> PUC = 1 - 2/6.
  expect_equal(unname(om[1, "PUC"]), 1 - 2 / 6)
})

test_that("the sums_load whole-scale row honours the factor correlations", {
  # The whole-scale composite's model-implied common variance is 1' S Phi S' 1, the same
  # quantity the subscale rows of this branch read off `common`. Scoring the whole-scale
  # row from the squared loading-column sums alone drops the factor correlations, which
  # would put an orthogonal model and a correlated one in one table.
  S <- matrix(0, 6, 2)
  S[1:3, 1] <- c(.75, .70, .65)
  S[4:6, 2] <- c(.60, .65, .70)
  Phi <- matrix(c(1, .5, .5, 1), 2, 2)
  common <- S %*% Phi %*% t(S)
  u2 <- 1 - diag(common)
  spec <- list(g_load = rep(0, 6), s_load = S, u2 = u2, Phi = Phi,
               map = cbind(rep(1:0, each = 3), rep(0:1, each = 3)), cormat = NULL,
               var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))

  om <- suppressWarnings(.reliability_core(spec, "sums_load"))
  expect_equal(unname(om["g", "tot"]), sum(common) / (sum(common) + sum(u2)))

  # Not a vacuous identity: the factor correlations move it well away from the value the
  # column sums alone give.
  s <- colSums(S)
  expect_gt(unname(om["g", "tot"]), sum(s^2) / (sum(s^2) + sum(u2)))

  # The same spec without factor correlations keeps the loading-column-sum arithmetic
  # exactly, which is every solution the shipped adapters can produce.
  spec$Phi <- NULL
  om_orth <- suppressWarnings(.reliability_core(spec, "sums_load"))
  expect_identical(unname(om_orth["g", "tot"]), sum(s^2) / (sum(s^2) + sum(u2)))
})

test_that("variables keyed against each other are flagged, in both variance modes", {
  cls <- "efa_reliability_mixed_keying"
  make <- function(g, s) {
    u2 <- 1 - g^2 - rowSums(s^2)
    L <- cbind(g, s)
    Rm <- L %*% t(L)
    diag(Rm) <- 1
    rownames(Rm) <- colnames(Rm) <- paste0("V", seq_along(g))
    list(g_load = g, s_load = s, u2 = u2, map = abs(s) > 0, cormat = Rm,
         var_names = paste0("V", seq_along(g)), fac_names = c("F1", "F2"))
  }

  # Half the variables keyed the other way, so the true score variance of their
  # unit-weighted sum cancels instead of accumulating.
  s <- matrix(0, 6, 2)
  s[1:3, 1] <- 0.2
  s[4:6, 2] <- -0.2
  mixed <- make(c(rep(0.7, 3), rep(-0.7, 3)), s)
  for (v in c("correlation", "sums_load")) {
    expect_warning(.reliability_core(mixed, v), class = cls)
  }

  # The same components all keyed alike are silent.
  aligned <- make(rep(0.7, 6), abs(s))
  for (v in c("correlation", "sums_load")) {
    expect_no_warning(.reliability_core(aligned, v), class = cls)
  }

  # Nothing to judge: no common variance at all, or a malformed component the callers
  # report separately.
  expect_no_warning(.rel_check_keying(matrix(0, 4, 4)), class = cls)
  expect_no_warning(.rel_check_keying(matrix(NA_real_, 4, 4)), class = cls)

  # The variables the message names are those subtracting from the composite, which is
  # guidance rather than the trigger: a solution can fall below the cutoff with no single
  # variable's own common variance turning negative, and it must still warn.
  opposed <- matrix(c(1, -0.4, -0.4, 1), 2, 2)
  expect_true(all(rowSums(opposed) > 0))
  expect_warning(.rel_check_keying(opposed), class = cls)

  # A single factor reaches the core on a spec with no group factors, whose common
  # variance comes from the general column alone; the check must fire there too.
  single <- function(g) .rel_single_factor_spec(
    list(g_load = rep(0, length(g)), s_load = matrix(g, ncol = 1), u2 = rep(0.51, 6),
         cormat = tcrossprod(g) + diag(0.51, 6),
         var_names = paste0("V", seq_along(g))))
  for (v in c("correlation", "sums_load")) {
    expect_warning(.reliability_core(single(c(rep(0.7, 3), rep(-0.7, 3))), v),
                   class = cls)
    expect_no_warning(.reliability_core(single(rep(0.7, 6)), v), class = cls)
  }
})

test_that(".rel_single_factor_label takes a name where there is one, F1 where there is not", {
  # A name is used as given ...
  expect_identical(.rel_single_factor_label("Ability"), "Ability")
  expect_identical(.rel_single_factor_label("g"), "g")
  # ... including one that happens to read as a digit.
  expect_identical(.rel_single_factor_label("1"), "1")

  # ... and everything that is not a name takes the default first-factor label: no name at
  # all, a bare column position (what .rel_adapt_manual() falls back to), a missing one, and
  # an empty one. Not "g", which would claim a general factor over group factors a
  # single-factor solution does not have.
  expect_identical(.rel_single_factor_label(NULL), "F1")
  expect_identical(.rel_single_factor_label(character(0)), "F1")
  expect_identical(.rel_single_factor_label(1L), "F1")
  expect_identical(.rel_single_factor_label(seq_len(1)), "F1")
  expect_identical(.rel_single_factor_label(NA_character_), "F1")
  expect_identical(.rel_single_factor_label(""), "F1")

  # More than one name cannot be the label of one factor.
  expect_identical(.rel_single_factor_label(c("F1", "F2")), "F1")
})

rm(efa_mod, sl_mod,
   om_sl_add, om_sl_noadd, om_sl_named_add, om_sl_named_noadd, schmid_mod,
   om_schmid_1_add, om_schmid_1_noadd, om_schmid_2, om_man_1_add, om_man_1_noadd,
   om_man_2)
if (requireNamespace("lavaan", quietly = TRUE)) {
  rm(lav_mod_1, lav_fit_1, om_lav_bi_add, om_lav_bi_noadd, lav_mod_2, lav_fit_2,
     om_lav_1_add, om_lav_1_noadd, lav_fit_3, lav_fit_4, om_lav_gr_add, om_lav_gr_noadd,
     om_lav_1_gr_add, om_lav_1_gr_noadd, lav_mod_ho_1, lav_mod_ho_2, lav_fit_ho_1,
     lav_fit_ho_2, om_lav_ho_1, lav_mod_NA, lav_fit_NA,
     lav_mod_inv, lav_fit_inv, lav_mod_bi_red, lav_fit_bi_red)
}

## Tests for the exported efa_reliability() front-end: adapter dispatch ->
## .reliability_core -> tidy long result, the coefficient selector, and errors.

efa_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
sl_mod <- efa_schmid_leiman(efa_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF")
fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2

# The map efa_reliability derives when factor_map is omitted: each item on its highest
# |group loading|.
auto <- local({
  s <- sl_mod$sl[, c("F1", "F2", "F3")]
  m <- matrix(0, nrow(s), ncol(s))
  for (i in seq_len(nrow(s))) m[i, which.max(abs(s[i, ]))] <- 1
  m
})

# A population bifactor carrying two residual covariances, with the model syntaxes that
# reproduce it exactly. Shared by the residual-covariance blocks below: their cross-checks
# only compare like with like while they all describe one model. Pure algebra, so it costs
# nothing where lavaan is absent.
rc <- local({
  vn <- paste0("V", 1:18)
  g <- rep(.5, 18)
  S <- matrix(0, 18, 3)
  S[1:6, 1] <- .45; S[7:12, 2] <- .45; S[13:18, 3] <- .45
  common <- tcrossprod(g) + tcrossprod(S)
  Theta <- diag(1 - diag(common))
  Theta[1, 2] <- Theta[2, 1] <- .30
  Theta[3, 4] <- Theta[4, 3] <- .25
  Sigma <- common + Theta
  dimnames(Sigma) <- list(vn, vn)
  first_order <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
                  F2 =~ V7 + V8 + V9 + V10 + V11 + V12
                  F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
  all_items <- paste(vn, collapse = " + ")
  list(common = common, Sigma = Sigma,
       correlated = first_order,
       bifactor = paste(first_order, paste("g =~", all_items), sep = "\n"),
       second_order = paste(first_order, "g =~ F1 + F2 + F3", sep = "\n"),
       single = paste("g =~", all_items))
})

# The residual covariance matrix of a standardized lavaan solution. lavaan reports the
# off-diagonal of `theta` as a residual correlation and its diagonal as a residual variance,
# so the two are put back in one metric before they are summed. Written out here rather than
# taken from the package, so the expected values below are derived independently of the code
# under test.
rc_theta <- function(theta) {
  d <- diag(theta)
  out <- theta * sqrt(outer(d, d))
  diag(out) <- d
  out
}

# Fitted once and shared by the blocks that read it: the correction and its semTools
# cross-check have to be statements about one and the same fit.
rc_fit_bifactor <- if (requireNamespace("lavaan", quietly = TRUE)) {
  suppressWarnings(lavaan::cfa(paste(rc$bifactor, "V1 ~~ V2", "V3 ~~ V4", sep = "\n"),
                               sample.cov = rc$Sigma, sample.nobs = 1000,
                               estimator = "ml", orthogonal = TRUE,
                               sample.cov.rescale = FALSE))
}

test_that("efa_reliability wires the SL adapter, core, and result builder together", {
  spec <- .rel_adapt_SL(sl_mod, factor_corres = fc, type = "EFAtools")
  core <- .reliability_core(spec, "correlation", add_ind = TRUE, add_rel = TRUE)
  expected <- .reliability_result(core, settings = list(variance = "correlation",
                                                        no_general = FALSE))

  res <- efa_reliability(sl_mod, factor_map = fc)
  expect_s3_class(res, "efa_reliability")
  expect_equal(res, expected)
})

test_that("efa_reliability surfaces the full coefficient menu; CR/AVE never appear", {
  res <- efa_reliability(sl_mod, factor_map = fc)
  expect_setequal(res$coefficient,
                  c("omega_total", "omega_hierarchical", "omega_subscale",
                    "alpha", "H", "ECV", "PUC"))
  expect_false(any(c("CR", "AVE", "GLB", "beta") %in% res$coefficient))
})

test_that("the numbers tie to the validated OMEGA path", {
  res <- efa_reliability(sl_mod, factor_map = fc)
  om <- OMEGA(sl_mod, type = "EFAtools", factor_corres = fc)

  get <- function(coef, fac) res$value[res$coefficient == coef & res$factor == fac]
  expect_equal(get("omega_total", "g"), unname(om["g", "tot"]))
  expect_equal(get("omega_hierarchical", "F2"), unname(om["F2", "hier"]))
  expect_equal(get("H", "F3"), unname(om["F3", "H"]))
  expect_equal(get("ECV", "g"), unname(om["g", "ECV"]))
})

test_that("the coefficients selector returns only the requested coefficients", {
  res <- efa_reliability(sl_mod, factor_map = fc,
                         coefficients = c("omega_total", "alpha"))
  expect_s3_class(res, "efa_reliability")
  expect_setequal(res$coefficient, c("omega_total", "alpha"))
  # The kind attribute is subset to match, and the values are unchanged.
  expect_setequal(names(attr(res, "kind")), c("omega_total", "alpha"))
  full <- efa_reliability(sl_mod, factor_map = fc)
  expect_equal(res$value[res$coefficient == "alpha"],
               full$value[full$coefficient == "alpha"])
})

test_that("selecting only common-variance indices drops the reliability rows", {
  res <- efa_reliability(sl_mod, factor_map = fc,
                         coefficients = c("ECV", "PUC"))
  expect_setequal(res$coefficient, c("ECV", "PUC"))
  expect_true(all(res$level == "general"))
})

test_that("an unknown coefficient is a classed error", {
  expect_error(
    efa_reliability(sl_mod, factor_map = fc, coefficients = "omega"),
    class = "efa_reliability_bad_coefficients")
})

test_that("invalid model and missing/malformed manual components are classed errors", {
  expect_error(efa_reliability(model = 1:5), class = "efa_reliability_invalid_model")
  expect_error(
    efa_reliability(model = NULL, var_names = rownames(sl_mod$sl),
                    s_load = sl_mod$sl[, c("F1", "F2", "F3")], u2 = sl_mod$sl[, "u2"],
                    factor_map = fc),
    class = "efa_reliability_missing_args")
  # A vector s_load must abort, not be silently coerced to a one-column matrix.
  expect_error(
    efa_reliability(model = NULL, var_names = rownames(sl_mod$sl),
                    g_load = sl_mod$sl[, "g"], s_load = as.numeric(sl_mod$sl[, "F1"]),
                    u2 = sl_mod$sl[, "u2"]),
    class = "efa_reliability_bad_s_load")
  # A non-numeric matrix model is rejected rather than failing deep in the algebra.
  expect_error(efa_reliability(matrix("a", 4, 2)), class = "efa_reliability_bad_matrix")
})

test_that("correlation mode without an available correlation matrix aborts (not silent NA)", {
  s <- matrix(0, 6, 2); s[1:3, 1] <- 0.4; s[4:6, 2] <- 0.4
  g <- rep(0.5, 6)
  u2 <- 1 - g^2 - rowSums(s^2)
  # Manual components with no cormat (and no pattern/Phi): correlation mode has no
  # denominator, so it aborts with an actionable message instead of returning NA omegas.
  expect_error(
    efa_reliability(model = NULL, var_names = paste0("V", 1:6), g_load = g,
                    s_load = s, u2 = u2),
    class = "efa_reliability_need_cormat")
  # The same input is fine under variance = "sums_load" (no correlation matrix needed).
  expect_s3_class(
    efa_reliability(model = NULL, var_names = paste0("V", 1:6), g_load = g,
                    s_load = s, u2 = u2, variance = "sums_load"),
    "efa_reliability")
})

test_that("manual components of mismatched length abort with a classed error", {
  s <- matrix(0, 5, 2); s[1:3, 1] <- 0.4; s[4:5, 2] <- 0.4
  expect_error(
    efa_reliability(model = NULL, var_names = paste0("V", 1:6), g_load = rep(0.5, 6),
                    s_load = s, u2 = rep(0.5, 6)),
    class = "efa_reliability_length_mismatch")
})

test_that("only correlation mode needs a correlation matrix, general factor or not", {
  s <- matrix(0, 6, 2); s[1:3, 1] <- 0.4; s[4:6, 2] <- 0.4
  u2 <- 1 - rowSums(s^2)
  args <- list(model = NULL, var_names = paste0("V", 1:6), s_load = s, u2 = u2)

  # Correlation mode has no denominator without one, with or without a general factor.
  # The uniquenesses follow each general column so the components stay standardized.
  for (g in list(rep(0, 6), rep(0.3, 6))) {
    expect_error(
      do.call(efa_reliability,
              c(args[setdiff(names(args), "u2")],
                list(g_load = g, u2 = 1 - g^2 - rowSums(s^2)))),
      class = "efa_reliability_need_cormat")
  }

  # The model-implied convention needs none: its denominator is the composite's common
  # variance plus the uniquenesses. An all-zero general column says the solution has no
  # general factor, not that a correlation matrix is required.
  res <- efa_reliability(model = NULL, var_names = paste0("V", 1:6), g_load = rep(0, 6),
                         s_load = s, u2 = u2, variance = "sums_load")
  tot <- res$value[res$coefficient == "omega_total" & res$factor == "total"]
  expect_equal(tot, sum(tcrossprod(s)) / (sum(tcrossprod(s)) + sum(u2)))
  expect_identical(attr(res, "settings"),
                   list(variance = "sums_load", no_general = TRUE))
})

test_that("Phi without pattern makes manual components a correlated-factors solution", {
  # The defect this closes: the same solution scored through the components and through the
  # fitted object must return the same coefficients. Without Phi in the spec the numerator is
  # 1'LL'1, which describes an orthogonal model and falls short of 1'L Phi L'1.
  L <- matrix(0, 6, 2)
  L[1:3, 1] <- c(.75, .70, .65)
  L[4:6, 2] <- c(.60, .65, .70)
  rownames(L) <- paste0("V", 1:6)
  Phi <- matrix(c(1, .5, .5, 1), 2, 2)
  common <- L %*% Phi %*% t(L)
  u2 <- 1 - diag(common)
  R <- common + diag(u2)
  dimnames(R) <- list(rownames(L), rownames(L))
  manual <- function(...) efa_reliability(NULL, var_names = rownames(L),
                                          g_load = rep(0, 6), s_load = L, u2 = u2, ...)
  get <- function(x) x$value[x$coefficient == "omega_total" & x$factor == "total"]

  expect_equal(get(manual(cormat = R, Phi = Phi)), sum(common) / sum(R))
  # Without it, the orthogonal reading -- materially different, so the test is not vacuous.
  expect_equal(get(manual(cormat = R)), sum(tcrossprod(L)) / sum(R))
  expect_gt(get(manual(cormat = R, Phi = Phi)), get(manual(cormat = R)))

  # It needs no correlation matrix in the model-implied convention.
  expect_equal(get(manual(Phi = Phi, variance = "sums_load")),
               sum(common) / (sum(common) + sum(u2)))

  # A pattern says Phi belongs to that separate solution, which is stated rather than assumed.
  expect_warning(res <- manual(Phi = Phi, pattern = L),
                 class = "efa_reliability_phi_pattern")
  expect_equal(get(res), get(manual(cormat = R)))
})

test_that("missing and infinite components are rejected, not scored", {
  # Every check that might name them is blind to exactly these values -- the Heywood report
  # tests `u2 < eps` and skips NA, and the keying and standardization checks compare totals
  # that are themselves non-finite -- while the table that comes back looks ordinary: an
  # infinite uniqueness drives its composite's omega total to a flat 0, an infinite general
  # loading leaves a plausible coefficient, and a missing one silently drops rows. The
  # assertions come from checkmate, so unlike the other component checks they carry no
  # package class and can only be asserted as errors.
  L <- matrix(c(.7, .7, .7, 0, 0, 0, 0, 0, 0, .6, .6, .6), 6, 2)
  u2 <- 1 - rowSums(L^2)
  manual <- function(g = rep(0, 6), uniq = u2) {
    efa_reliability(NULL, var_names = paste0("V", 1:6), g_load = g, s_load = L,
                    u2 = uniq, variance = "sums_load")
  }

  s_bad <- function(bad) {
    efa_reliability(NULL, var_names = paste0("V", 1:6), g_load = rep(0, 6),
                    s_load = replace(L, 2, bad), u2 = u2, variance = "sums_load")
  }
  for (bad in list(NA_real_, Inf, -Inf)) {
    expect_error(manual(uniq = replace(u2, 2, bad)))
    expect_error(manual(g = replace(rep(0.3, 6), 2, bad)))
    expect_error(s_bad(bad))
  }
  expect_error(manual(uniq = rep(Inf, 6)))
  expect_error(manual(g = rep(NA_real_, 6)))

  # Checking the group loadings numerically also rejects a matrix that is not numeric,
  # which the class test alone admits.
  expect_error(efa_reliability(NULL, var_names = paste0("V", 1:6), g_load = rep(0, 6),
                               s_load = matrix(as.character(L), 6, 2), u2 = u2,
                               variance = "sums_load"))

  # The same components, complete and finite, are scored.
  expect_s3_class(manual(), "efa_reliability")
  expect_s3_class(manual(g = rep(0.3, 6), uniq = 1 - 0.09 - rowSums(L^2)),
                  "efa_reliability")

  # A bare bifactor loading matrix takes `u2` as an override of the uniquenesses it would
  # otherwise derive, so that route needs the same guard on its own account.
  Lb <- cbind(g = rep(.5, 6), L)
  rownames(Lb) <- paste0("V", 1:6)
  u2b <- 1 - rowSums(Lb^2)
  for (bad in list(NA_real_, Inf, -Inf)) {
    expect_error(efa_reliability(Lb, u2 = replace(u2b, 2, bad)))
  }
  # Omitting it, so the adapter derives them, is untouched.
  expect_s3_class(efa_reliability(Lb, u2 = u2b), "efa_reliability")
  expect_s3_class(efa_reliability(Lb), "efa_reliability")
})

test_that("supplied uniquenesses are checked against the loadings they accompany", {
  # The printed Schmid-Leiman table ends in h2 and u2, and published tables often report
  # only h2, so passing communalities where uniquenesses belong is a one-column slip. It is
  # silent otherwise: u2 does not enter the coefficients under variance = "correlation"
  # with a correlation matrix, only in the model-implied convention a user without one is
  # sent to.
  sl <- unclass(sl_mod$sl)
  g <- sl[, "g"]
  s <- sl[, c("F1", "F2", "F3")]
  h2 <- sl[, "h2"]
  manual <- function(u2, ...) {
    efa_reliability(NULL, var_names = rownames(sl), g_load = g, s_load = s, u2 = u2,
                    variance = "sums_load", ...)
  }

  expect_warning(manual(u2 = h2), class = "efa_reliability_u2_not_standardized")
  expect_no_warning(manual(u2 = sl[, "u2"]),
                    class = "efa_reliability_u2_not_standardized")

  # A published table rounded to two decimals is a legitimate transcription and must not
  # trip it, so the tolerance has to clear ordinary rounding by a wide margin.
  expect_no_warning(manual(u2 = round(sl[, "u2"], 2)),
                    class = "efa_reliability_u2_not_standardized")

  # The bare bifactor matrix takes its own uniquenesses, and derives them when it is not
  # given any -- derived ones complete by construction and must stay silent.
  L <- cbind(g = g, s)
  expect_warning(efa_reliability(L, u2 = h2, cormat = test_models$baseline$cormat),
                 class = "efa_reliability_u2_not_standardized")
  expect_no_warning(efa_reliability(L, cormat = test_models$baseline$cormat),
                    class = "efa_reliability_u2_not_standardized")

  # It reads the factor correlations where the solution has them. That only shows on a
  # solution with cross-loadings: under exact simple structure the off-diagonal terms of
  # L Phi L' never reach the diagonal, so the Phi-aware and Phi-blind uniquenesses coincide
  # and no fixture built that way could tell the two apart.
  Lo <- matrix(0, 8, 2)
  Lo[1:4, 1] <- .70; Lo[1:4, 2] <- .30
  Lo[5:8, 2] <- .70; Lo[5:8, 1] <- .30
  Phi <- matrix(c(1, .5, .5, 1), 2, 2)
  aware <- 1 - diag(Lo %*% Phi %*% t(Lo))
  blind <- 1 - rowSums(Lo^2)
  expect_gt(max(abs(aware - blind)), .05)

  obl <- function(u2) efa_reliability(NULL, var_names = paste0("V", 1:8),
                                      g_load = rep(0, 8), s_load = Lo, u2 = u2,
                                      Phi = Phi, variance = "sums_load")
  expect_no_warning(obl(aware), class = "efa_reliability_u2_not_standardized")
  expect_warning(obl(blind), class = "efa_reliability_u2_not_standardized")
})

test_that("a correlation matrix and a pattern to rebuild one are an either/or", {
  # Two ways to give one matrix. Taking both is ambiguous rather than redundant: the
  # reconstruction never runs when a cormat is in hand, so `pattern` is read by nothing --
  # yet its presence alone would decide whether `Phi` describes this solution or another.
  L <- matrix(0, 6, 2); L[1:3, 1] <- .7; L[4:6, 2] <- .6
  rownames(L) <- paste0("V", 1:6)
  Phi <- matrix(c(1, .5, .5, 1), 2, 2)
  u2 <- 1 - diag(L %*% Phi %*% t(L))
  R <- L %*% Phi %*% t(L) + diag(u2)
  dimnames(R) <- list(rownames(L), rownames(L))
  manual <- function(...) efa_reliability(NULL, var_names = rownames(L),
                                          g_load = rep(0, 6), s_load = L, u2 = u2, ...)

  expect_error(manual(cormat = R, pattern = L, Phi = Phi),
               class = "efa_reliability_cormat_and_pattern")
  # A pattern with no Phi cannot reconstruct anything either, so it is the same clash.
  expect_error(manual(cormat = R, pattern = L),
               class = "efa_reliability_cormat_and_pattern")

  # Each on its own is fine, and Phi with a cormat is the correlated-factors call.
  expect_no_error(manual(cormat = R, Phi = Phi))
  expect_no_error(suppressWarnings(manual(pattern = L, Phi = Phi)))
})

test_that("Phi is rejected or reported where it cannot describe the group factors", {
  L <- matrix(c(.7, .7, .7, 0, 0, 0, 0, 0, 0, .6, .6, .6), 6, 2)
  u2 <- 1 - rowSums(L^2)
  manual <- function(...) efa_reliability(NULL, var_names = paste0("V", 1:6),
                                          s_load = L, u2 = u2, variance = "sums_load", ...)
  ok <- matrix(c(1, .5, .5, 1), 2, 2)

  # Correlated group factors and a general factor are not the decomposition these
  # coefficients report.
  expect_error(manual(g_load = rep(.5, 6), Phi = ok),
               class = "efa_reliability_phi_with_general")

  # A Phi that is not a correlation matrix of the solution's group factors.
  for (bad in list(diag(3), diag(2, 2), matrix(c(1, .9, .1, 1), 2, 2),
                   matrix(c(1, NA, NA, 1), 2, 2), matrix(c(1, 1.4, 1.4, 1), 2, 2),
                   as.data.frame(ok), "nonsense")) {
    expect_error(manual(g_load = rep(0, 6), Phi = bad),
                 class = "efa_reliability_bad_phi")
  }

  # A Phi paired with a pattern builds the correlation matrix every coefficient is divided
  # by, so it is checked on that reading too, against the pattern's own factor count.
  expect_error(manual(g_load = rep(0, 6), pattern = L, Phi = diag(3)),
               class = "efa_reliability_bad_phi")
  expect_warning(manual(g_load = rep(0, 6), pattern = L, Phi = ok),
                 class = "efa_reliability_phi_pattern")

  # A single group factor takes a 1 x 1 Phi, which must stay reachable. It is one factor,
  # so the single-factor note is expected here and is muffled rather than left to print.
  expect_no_error(suppressMessages(
    efa_reliability(NULL, var_names = paste0("V", 1:6), g_load = rep(0, 6),
                    s_load = matrix(.6, 6, 1), u2 = rep(.64, 6),
                    Phi = matrix(1, 1, 1), variance = "sums_load")))

  # A fitted solution carries its own components, so every argument that describes manually
  # supplied ones is reported rather than dropped -- not just the two that changed meaning.
  cls <- "efa_reliability_args_ignored"
  expect_warning(efa_reliability(sl_mod, factor_map = fc, Phi = ok), class = cls)
  expect_warning(efa_reliability(sl_mod, factor_map = fc, pattern = L), class = cls)
  expect_warning(efa_reliability(sl_mod, factor_map = fc, u2 = rep(.5, 18)), class = cls)
  expect_warning(efa_reliability(sl_mod, factor_map = fc, g_load = rep(.5, 18)), class = cls)
  expect_warning(efa_reliability(sl_mod, factor_map = fc, var_names = paste0("z", 1:18)),
                 class = cls)
  # and the coefficients are untouched by any of them
  expect_equal(suppressWarnings(efa_reliability(sl_mod, factor_map = fc, Phi = ok)),
               efa_reliability(sl_mod, factor_map = fc))
  expect_equal(suppressWarnings(efa_reliability(sl_mod, factor_map = fc,
                                                u2 = rep(.5, 18))),
               efa_reliability(sl_mod, factor_map = fc))

  # A bare loading matrix is the one `model` that does read `u2`, so it must not be
  # reported there -- but it must line up with the loadings.
  Lb <- cbind(g = rep(.5, 6), matrix(c(rep(.4, 3), rep(0, 6), rep(.4, 3)), 6, 2))
  rownames(Lb) <- paste0("V", 1:6)
  expect_no_warning(efa_reliability(Lb, u2 = 1 - rowSums(Lb^2)), class = cls)
  expect_error(efa_reliability(Lb, u2 = rep(.5, 4)))

  # A lavaan fit reads neither the correspondences nor a correlation matrix, so those are
  # reported there too, and the remedy must not point the user back at them.
  skip_if_not_installed("lavaan")
  fit <- suppressWarnings(lavaan::cfa(
    paste('F1 =~ V1 + V2 + V3 + V4 + V5 + V6
           F2 =~ V7 + V8 + V9 + V10 + V11 + V12
           F3 =~ V13 + V14 + V15 + V16 + V17 + V18
           g =~', paste(paste0("V", 1:18), collapse = " + ")),
    sample.cov = test_models$baseline$cormat, sample.nobs = 500, estimator = "ml",
    orthogonal = TRUE))
  expect_warning(efa_reliability(fit, g_name = "g",
                                 cormat = test_models$baseline$cormat), class = cls)
  expect_warning(efa_reliability(fit, g_name = "g", factor_map = matrix(1, 18, 3)),
                 class = cls)
  # and they never moved a coefficient
  expect_equal(suppressWarnings(
    efa_reliability(fit, g_name = "g", cormat = test_models$baseline$cormat)),
    efa_reliability(fit, g_name = "g"))
})

test_that("a labelling argument is reported for an input that does not read it", {
  # `fac_names` and `group_names` label the result rather than describe components, so each
  # is reported where it does not apply rather than dropped: `fac_names` for a lavaan fit,
  # whose factor labels come from the model syntax, and `group_names` for anything but a
  # lavaan fit, every other input being a single ungrouped solution.
  cls <- "efa_reliability_arg_not_applicable"

  expect_warning(efa_reliability(sl_mod, factor_map = fc, group_names = c("a", "b")),
                 class = cls)
  Lb <- cbind(g = rep(.5, 6), F1 = c(rep(.4, 3), rep(0, 3)),
              F2 = c(rep(0, 3), rep(.4, 3)))
  rownames(Lb) <- paste0("V", 1:6)
  expect_warning(efa_reliability(Lb, group_names = c("a", "b"), variance = "sums_load"),
                 class = cls)
  # Manually supplied components read no `group_names` either, and are not covered by the
  # report for a fitted `model`, there being none.
  expect_warning(efa_reliability(NULL, var_names = rownames(Lb), g_load = Lb[, 1],
                                 s_load = Lb[, -1], u2 = 1 - rowSums(Lb^2),
                                 group_names = c("a", "b"), variance = "sums_load"),
                 class = cls)
  # and it never moved a coefficient
  expect_equal(suppressWarnings(
    efa_reliability(sl_mod, factor_map = fc, group_names = c("a", "b"))),
    efa_reliability(sl_mod, factor_map = fc))

  # `fac_names` is read on every non-lavaan route, so it stays silent there.
  expect_no_warning(efa_reliability(sl_mod, factor_map = fc,
                                    fac_names = c("A", "B", "C")), class = cls)

  skip_if_not_installed("lavaan")
  fit <- suppressWarnings(lavaan::cfa(
    paste('F1 =~ V1 + V2 + V3 + V4 + V5 + V6
           F2 =~ V7 + V8 + V9 + V10 + V11 + V12
           F3 =~ V13 + V14 + V15 + V16 + V17 + V18
           g =~', paste(paste0("V", 1:18), collapse = " + ")),
    sample.cov = test_models$baseline$cormat, sample.nobs = 500, estimator = "ml",
    orthogonal = TRUE))
  expect_warning(efa_reliability(fit, g_name = "g", fac_names = c("A", "B", "C")),
                 class = cls)
  # Including one of the wrong length, which the length check -- belonging to the routes
  # that read it -- does not refuse here.
  expect_warning(efa_reliability(fit, g_name = "g", fac_names = c("A", "B")), class = cls)
  expect_equal(suppressWarnings(
    efa_reliability(fit, g_name = "g", fac_names = c("A", "B", "C"))),
    efa_reliability(fit, g_name = "g"))

  # A single-group lavaan fit is scored as one ungrouped solution and drops the labels
  # unread, exactly as the inputs above do, so it is reported too -- the report keys on the
  # group count, not on the class.
  expect_warning(efa_reliability(fit, g_name = "g", group_names = "one"), class = cls)
  expect_true(all(is.na(suppressWarnings(
    efa_reliability(fit, g_name = "g", group_names = "one"))$group)))

  # The mirror, and the one input that does read them: a multiple-group fit, whose blocks
  # the labels key. Without it an inverted guard would warn here and pass the suite.
  fit_gr <- suppressWarnings(lavaan::cfa(
    paste('F1 =~ V1 + V2 + V3 + V4 + V5 + V6
           F2 =~ V7 + V8 + V9 + V10 + V11 + V12
           F3 =~ V13 + V14 + V15 + V16 + V17 + V18
           g =~', paste(paste0("V", 1:18), collapse = " + ")),
    sample.cov = list(test_models$baseline$cormat, test_models$baseline$cormat),
    sample.nobs = c(500, 500), estimator = "ml", orthogonal = TRUE))
  expect_no_warning(efa_reliability(fit_gr, g_name = "g",
                                    group_names = c("Young", "Old")), class = cls)
})

test_that("omitting factor_map auto-maps each item to its highest-loading factor", {
  # No factor_map: each item is assigned to its highest |group loading|, so the
  # result matches supplying that same 0/1 assignment explicitly.
  expect_equal(efa_reliability(sl_mod), efa_reliability(sl_mod, factor_map = auto))
})

test_that("a factor_map whose columns are in the wrong order is flagged", {
  # The columns of factor_map are matched to the group factors by position, so a
  # permuted map assigns each factor the items of another one -- well-formed, but
  # meaningless. The items then hardly load on the factor they are mapped to while
  # loading substantially on another, which is what the check picks up.
  expect_warning(efa_reliability(sl_mod, factor_map = auto[, c(2, 3, 1)]),
                 class = "efa_reliability_implausible_map")
  # A sound map -- explicit or auto-derived -- stays silent.
  expect_no_warning(efa_reliability(sl_mod, factor_map = fc))
  expect_no_warning(efa_reliability(sl_mod, factor_map = auto))
  expect_no_warning(efa_reliability(sl_mod))

  # The check belongs to the efa_reliability front-end; OMEGA()'s behaviour is
  # unchanged (it derives and validates its map in its own worker).
  expect_no_warning(OMEGA(sl_mod, type = "EFAtools",
                          factor_corres = auto[, c(2, 3, 1)]),
                    class = "efa_reliability_implausible_map")
})

test_that("a bifactor loading matrix validates and checks a supplied factor_map", {
  L <- cbind(g = rep(0.5, 6),
             F1 = c(rep(0.5, 3), rep(0, 3)),
             F2 = c(rep(0, 3), rep(0.5, 3)))
  rownames(L) <- paste0("V", 1:6)
  map <- cbind(c(rep(1, 3), rep(0, 3)), c(rep(0, 3), rep(1, 3)))

  expect_no_warning(efa_reliability(L, factor_map = map))
  # Swapping the columns points each factor at the other one's items.
  expect_warning(efa_reliability(L, factor_map = map[, c(2, 1)]),
                 class = "efa_reliability_implausible_map")
  # A map that does not conform to the group loadings is an error, not coefficients
  # above 1 from a recycled map.
  expect_error(efa_reliability(L, factor_map = map[1:3, , drop = FALSE]),
               class = "efa_reliability_map_dim")
  expect_error(efa_reliability(L, factor_map = cbind(map, 0)),
               class = "efa_reliability_map_dim")
  # a map of the right shape that is not a matrix is rejected by the same check
  expect_error(efa_reliability(L, factor_map = as.data.frame(map)),
               class = "efa_reliability_map_dim")
})

test_that("an oblique EFA is scored as a correlated-factors model (no general factor)", {
  res <- efa_reliability(efa_mod)
  # No general factor: the bifactor indices omega hierarchical, ECV, and PUC are not
  # meaningful and are omitted.
  expect_false(any(res$coefficient == "omega_hierarchical"))
  expect_false(any(res$coefficient == "ECV"))
  expect_false(any(res$coefficient == "PUC"))
  # The first row is the composite of every variable, not a general factor the solution
  # does not have, so it is labelled for that and takes its own level. Labelled "g" at
  # level "general" it would state a general factor, and would collide with a group factor
  # a user names "g".
  expect_true(any(res$factor == "total" & res$level == "total"))
  expect_false(any(res$factor == "g"))
  expect_false(any(res$level == "general"))
  # The synthetic all-zero general-factor column also leaves the whole-scale (total) row's
  # omega subscale and H index undefined; they are dropped (H of an all-zero loading
  # vector would otherwise print a misleading 0). Group factors keep their own omega
  # subscale and H.
  expect_false(any(res$coefficient == "omega_subscale" & res$factor == "total"))
  expect_false(any(res$coefficient == "H" & res$factor == "total"))
  expect_true(any(res$coefficient == "omega_subscale" & res$factor != "total"))
  expect_true(any(res$coefficient == "H" & res$factor != "total"))
  # The whole-scale omega total and alpha, and each group factor's omega/H, remain.
  expect_true(any(res$coefficient == "omega_total" & res$factor == "total"))
  expect_true(any(res$coefficient == "alpha" & res$factor == "total"))
  spec <- .rel_adapt_efa(efa_mod, type = "psych")
  expect_equal(res$value[res$coefficient == "omega_total" & res$factor == "total"],
               sum(spec$s_load %*% spec$Phi %*% t(spec$s_load)) / sum(spec$cormat))
})

test_that("the correlated-factors group omega totals match psych::omega", {
  skip_on_cran()
  skip_if_not_installed("psych")
  # An external check on the composite common variance. psych scores the same data
  # through its own Schmid-Leiman solution and adds the general and own-group terms of
  # that solution, so its group totals sit a hair below ours, which count the
  # cross-loading terms too. Compared as sorted values, because the two packages order
  # their factors independently.
  res <- efa_reliability(efa_mod)
  ours <- res$value[res$coefficient == "omega_total" & res$level == "group"]
  po <- suppressWarnings(suppressMessages(
    psych::omega(test_models$baseline$cormat, nfactors = 3, n.obs = 500, fm = "pa",
                 rotate = "Promax", plot = FALSE)))
  theirs <- sort(unname(as.matrix(po$omega.group)[-1, "total"]))
  expect_equal(sort(ours), theirs, tolerance = 5e-3)
  # Agreement to three decimals is not the whole claim. The values must belong to the
  # factors they are reported for: each factor's total is its own composite's, not a
  # sorted set. Checked against the composite common variance of that factor's assigned
  # items -- the assertion that fails if the cross terms are dropped again, whatever the
  # numerical agreement with psych. The cross terms are also shown to matter here, so
  # the identity is not vacuous; their sign is not fixed in general, because
  # cs' Phi cs is a quadratic form and can fall either side of cs_j^2 depending on how
  # the other columns' sums line up with the factor correlations.
  spec <- .rel_adapt_efa(efa_mod, type = "psych")
  common <- spec$s_load %*% spec$Phi %*% t(spec$s_load)
  for (j in seq_len(ncol(spec$s_load))) {
    mem <- which(spec$map[, j] == 1)
    Vgr <- sum(spec$cormat[mem, mem])
    expect_equal(res$value[res$coefficient == "omega_total" &
                             res$factor == spec$fac_names[j]],
                 sum(common[mem, mem]) / Vgr)
    expect_gt(sum(common[mem, mem]) / Vgr, sum(spec$s_load[mem, j])^2 / Vgr)
  }
})

test_that("sums_load on an oblique EFA is scored, and accounts for the factor correlations", {
  # Both conventions are defined for a correlated-factors solution: they share the
  # composite's model-implied common variance, which reads the factor correlations, and
  # differ only in the denominator -- the observed total variance against the
  # model-implied one.
  expect_no_warning(efa_reliability(efa_mod))
  res <- efa_reliability(efa_mod, variance = "sums_load")
  expect_identical(attr(res, "settings"), list(variance = "sums_load",
                                               no_general = TRUE))

  spec <- .rel_adapt_efa(efa_mod, type = "psych")
  common <- spec$s_load %*% spec$Phi %*% t(spec$s_load)
  tot <- res$value[res$coefficient == "omega_total" & res$factor == "total"]
  expect_equal(tot, sum(common) / (sum(common) + sum(spec$u2)))

  # The factor correlations are genuinely in there: ignoring them, as summing the squared
  # loading-column sums alone would, gives a materially different value.
  s <- colSums(spec$s_load)
  expect_false(isTRUE(all.equal(sum(common), sum(s^2))))

  # This solution reproduces its correlation matrix closely, so the two denominators --
  # and hence the two conventions -- nearly agree; they are not the same computation.
  expect_equal(tot, suppressWarnings(efa_reliability(efa_mod))$value[1],
               tolerance = 1e-4)
})

test_that("an explicitly requested but undefined coefficient is reported, not silent", {
  # omega hierarchical is not defined for a correlated-factors EFA.
  expect_message(efa_reliability(efa_mod, coefficients = "omega_hierarchical"),
                 class = "efa_reliability_coef_undefined")
  res <- suppressMessages(
    efa_reliability(efa_mod, coefficients = c("omega_total", "omega_hierarchical")))
  # The defined coefficient is still returned; the undefined one is dropped.
  expect_setequal(res$coefficient, "omega_total")
})

test_that("a correlation matrix is rejected, a square bifactor matrix is scored", {
  # Read as a bifactor matrix, a correlation matrix takes the first variable's
  # correlations for general-factor loadings and returns plausible but meaningless
  # coefficients. It carries no loading class, so nothing but its shape identifies it.
  expect_error(efa_reliability(test_models$baseline$cormat),
               class = "efa_reliability_cormat_as_model")
  expect_error(efa_reliability(stats::cov2cor(stats::cov(GRiPS_raw))),
               class = "efa_reliability_cormat_as_model")

  # The guard keys on the unit diagonal, not on squareness: a bifactor matrix of p
  # variables with a general factor and p - 1 group factors is square too, and is still
  # scored, because its diagonal holds loadings rather than ones.
  L <- matrix(c(0.60, 0.60, 0.60, 0.60,
                0.40, 0.40, 0.00, 0.00,
                0.00, 0.00, 0.40, 0.40,
                0.30, 0.00, 0.00, 0.30), nrow = 4)
  expect_s3_class(efa_reliability(L), "efa_reliability")

  # The manual route takes the group loadings separately, where a correlation matrix
  # passes the class and numeric checks and has one row per variable, so nothing else
  # would catch it.
  cm <- test_models$baseline$cormat
  p <- ncol(cm)
  expect_error(
    efa_reliability(var_names = colnames(cm), g_load = rep(0.5, p), s_load = cm,
                    u2 = rep(0.5, p), variance = "sums_load"),
    class = "efa_reliability_cormat_as_s_load")

  # The group loadings of a genuine bifactor solution are still accepted there.
  s <- L[, 2:3]
  expect_s3_class(
    efa_reliability(var_names = paste0("V", 1:4), g_load = L[, 1], s_load = s,
                    u2 = 1 - L[, 1]^2 - rowSums(s^2), variance = "sums_load"),
    "efa_reliability")

  # A missing value is reported as such: the correlation-matrix test aborts on one in
  # terms of its own argument, so it must not be the check that speaks first.
  s_na <- test_models$baseline$cormat
  s_na[2, 3] <- s_na[3, 2] <- NA
  expect_error(
    efa_reliability(var_names = colnames(s_na), g_load = rep(0.5, ncol(s_na)),
                    s_load = s_na, u2 = rep(0.5, ncol(s_na)), variance = "sums_load"),
    class = "efa_invalid_argument")
  # Likewise on the matrix route, where the guard sees an incomplete matrix at all.
  expect_error(efa_reliability(s_na), class = "efa_invalid_argument")
})

test_that("the whole Schmid-Leiman table is refused as s_load, its group columns are not", {
  # Passed whole, its h2 and u2 columns would be scored as two more group factors and its
  # general column as a third, on top of the g_load named separately.
  expect_error(
    efa_reliability(var_names = rownames(sl_mod$sl), g_load = sl_mod$sl[, "g"],
                    s_load = sl_mod$sl, u2 = sl_mod$sl[, "u2"], variance = "sums_load"),
    class = "efa_reliability_s_load_display_cols")

  # Either column alone is caught too, which also renders the singular message.
  for (nm in c("h2", "u2")) {
    s1 <- sl_mod$sl[, c(nm, "F2")]
    expect_error(
      efa_reliability(var_names = rownames(sl_mod$sl), g_load = sl_mod$sl[, "g"],
                      s_load = s1, u2 = sl_mod$sl[, "u2"], variance = "sums_load"),
      class = "efa_reliability_s_load_display_cols")
  }

  # The group-factor columns of that same table are the documented input and still score.
  expect_s3_class(
    efa_reliability(var_names = rownames(sl_mod$sl), g_load = sl_mod$sl[, "g"],
                    s_load = sl_mod$sl[, c("F1", "F2", "F3")], u2 = sl_mod$sl[, "u2"],
                    variance = "sums_load"),
    "efa_reliability")
})

test_that("a bare loading matrix with a missing or infinite loading is refused", {
  # With no u2 the adapter derives it as 1 - rowSums(loadings^2), so an unchecked loading
  # carries a missing or infinite uniqueness into the coefficients -- and the table it
  # produces looks ordinary. The manual route refuses the same components.
  L <- matrix(c(0.6, 0.6, 0.6, 0.6,
                0.5, 0.5, 0.0, 0.0,
                0.0, 0.0, 0.5, 0.5), nrow = 4)
  for (bad in list(NA_real_, Inf)) {
    L_bad <- L
    L_bad[2, 2] <- bad
    expect_error(efa_reliability(L_bad, variance = "sums_load"),
                 class = "efa_invalid_argument")
  }
  expect_s3_class(efa_reliability(L, variance = "sums_load"), "efa_reliability")
})

test_that("a factor_map holding values other than 0 and 1 is refused", {
  # The map states membership, and is read both as the set of assigned variables and as a
  # multiplier on the loadings and uniquenesses. A map of 2s names the same composites but
  # empties every one of them, returning NA coefficients under a plausible whole-scale row.
  L <- matrix(c(0.6, 0.6, 0.6, 0.6,
                0.5, 0.5, 0.0, 0.0,
                0.0, 0.0, 0.5, 0.5), nrow = 4)
  map <- matrix(c(1, 1, 0, 0, 0, 0, 1, 1), nrow = 4)
  expect_error(efa_reliability(L, factor_map = map * 2, variance = "sums_load"),
               class = "efa_reliability_map_values")
  # A missing entry is no more a membership statement than a 2 is.
  map_na <- map; map_na[1, 1] <- NA
  expect_error(efa_reliability(L, factor_map = map_na, variance = "sums_load"),
               class = "efa_reliability_map_values")
  # Several distinct offending values, so the message lists more than one of them.
  map_mixed <- matrix(c(1, 2, 3, 0, 0, 0, 1, 1), nrow = 4)
  expect_error(efa_reliability(L, factor_map = map_mixed, variance = "sums_load"),
               class = "efa_reliability_map_values")

  # The check sits where every route meets, so OMEGA()'s factor_corres gets it too --
  # that front end validates only the dimensions of its own argument.
  expect_error(OMEGA(sl_mod, factor_corres = matrix(2, nrow(sl_mod$sl), 3)),
               class = "efa_reliability_map_values")

  # 0/1 and the logical map that stands for it are both accepted, and agree.
  res <- efa_reliability(L, factor_map = map, variance = "sums_load")
  expect_s3_class(res, "efa_reliability")
  expect_equal(res, efa_reliability(L, factor_map = map == 1, variance = "sums_load"))
})

test_that("an ordinary loading matrix is rejected, a Schmid-Leiman table is scored", {
  # A pattern matrix read as a bifactor matrix takes its first factor for a general
  # factor and returns plausible but meaningless coefficients, so it is refused.
  expect_error(efa_reliability(efa_mod$rot_loadings),
               class = "efa_reliability_pattern_matrix")
  expect_error(efa_reliability(efa_mod$unrot_loadings),
               class = "efa_reliability_pattern_matrix")
  # A genuine bifactor matrix is still read as one once it carries no loading class.
  expect_s3_class(efa_reliability(unclass(efa_mod$rot_loadings)), "efa_reliability")

  # The Schmid-Leiman loading table is a bifactor matrix plus its h2/u2 display
  # columns: it is scored without them, and with its own uniquenesses, so it gives
  # exactly what the Schmid-Leiman object gives.
  res <- efa_reliability(sl_mod$sl, factor_map = fc,
                         cormat = test_models$baseline$cormat)
  expect_setequal(unique(res$factor), c("g", "F1", "F2", "F3"))
  expect_equal(res, efa_reliability(sl_mod, factor_map = fc))

  # With no factor_map the table gets the same auto-derived correspondences as the
  # object. The nonzero-pattern default of a raw bifactor matrix would instead assign
  # every variable to every group factor -- the table is estimated and has no structural
  # zeros -- which collapses PUC to zero and repeats the whole-scale alpha on every row.
  expect_equal(efa_reliability(sl_mod$sl, cormat = test_models$baseline$cormat),
               efa_reliability(sl_mod))

  # The h2/u2 columns are always the last two, so an unlabelled table is read off that
  # layout instead of scoring them as two more group factors (which puts omega total
  # above 1 and names a factor "5").
  unlabelled <- sl_mod$sl
  colnames(unlabelled) <- NULL
  res_bare <- efa_reliability(unlabelled, cormat = test_models$baseline$cormat)
  expect_length(unique(res_bare$factor), 4L)
  expect_true(all(res_bare$value[res_bare$coefficient == "omega_total"] <= 1))
  # Only the row labels differ from the labelled table, not the coefficients.
  expect_equal(res_bare$value, res$value)
})

test_that("an ordinary loading matrix with its Phi is scored as a correlated-factors solution", {
  # The same solution in three spellings -- the fitted object, its loading matrix with the
  # factor correlations, and the components one by one -- must give one result. The matrix
  # route is the object's own components read off the matrix, so this is an identity and
  # not an approximation.
  L <- efa_mod$rot_loadings
  Phi <- efa_mod$Phi
  cm <- test_models$baseline$cormat
  u2 <- 1 - diag(L %*% Phi %*% t(L))

  res <- efa_reliability(L, Phi = Phi, cormat = cm)
  expect_s3_class(res, "efa_reliability")
  expect_equal(res, efa_reliability(efa_mod, cormat = cm))
  expect_equal(res, efa_reliability(NULL, var_names = rownames(L), g_load = rep(0, nrow(L)),
                                    s_load = unclass(L), u2 = u2, Phi = Phi, cormat = cm,
                                    fac_names = colnames(L)))

  # It is the correlated-factors menu, not the bifactor one: the coefficients a solution
  # without a general factor does not define are omitted rather than returned as zeros.
  expect_setequal(res$coefficient, c("omega_total", "omega_subscale", "alpha", "H"))
  expect_identical(attr(res, "settings"), list(variance = "correlation",
                                               no_general = TRUE))
  # Labels come from the matrix itself.
  expect_setequal(unique(res$factor), c("total", "F1", "F2", "F3"))

  # Without Phi nothing says the columns are correlated factors, so the matrix is still
  # refused rather than read as a bifactor one.
  expect_error(efa_reliability(L, cormat = cm), class = "efa_reliability_pattern_matrix")
  # Phi is not reported as ignored on this route -- it is what the route reads.
  expect_no_warning(efa_reliability(L, Phi = Phi, cormat = cm),
                    class = "efa_reliability_args_ignored")
  # Stripped of its class the same matrix is still the pattern it was: the class does not
  # survive subsetting the matrix or reordering its rows, while Phi does, and Phi is what
  # says the columns are correlated factors. Read as one, and so giving the same result,
  # under a warning, since the bifactor reading reports coefficients this one does not.
  expect_warning(bare <- efa_reliability(unclass(L), Phi = Phi, cormat = cm),
                 class = "efa_reliability_phi_as_pattern")
  expect_equal(bare, res)

  # The uniquenesses are derived from the loadings under Phi, and a supplied set overrides
  # them, as on the bare bifactor matrix route. Shown in the model-implied convention,
  # which is the one they enter (with a correlation matrix in hand they do not).
  derived <- efa_reliability(L, Phi = Phi, variance = "sums_load")
  expect_equal(efa_reliability(L, Phi = Phi, variance = "sums_load", u2 = u2), derived)
  own <- suppressWarnings(efa_reliability(L, Phi = Phi, variance = "sums_load",
                                          u2 = u2 / 2))
  expect_false(isTRUE(all.equal(own$value, derived$value)))
  # A malformed one is refused there too, rather than scored into a plausible table.
  expect_error(efa_reliability(L, Phi = Phi, cormat = cm, u2 = replace(u2, 2, NA_real_)),
               class = "efa_invalid_argument")
  # A Phi that cannot be the correlation matrix of these columns is named as such.
  expect_error(efa_reliability(L, Phi = diag(2), cormat = cm),
               class = "efa_reliability_bad_phi")
})

test_that("Phi settles the reading of a matrix that states none, and is refused where it is stated", {
  L <- unclass(efa_mod$rot_loadings)
  cm <- test_models$baseline$cormat
  Phi <- efa_mod$Phi
  Phi2 <- matrix(c(1, .4, .4, 1), 2, 2)
  cls <- "efa_reliability_phi_with_general"

  # A matrix that carries no loading class states nothing about its columns, and the class
  # is not evidence of a hierarchy: subsetting a pattern matrix or reordering its rows drops
  # it while Phi survives. Phi is the evidence, since the group factors of a hierarchy are
  # orthogonal and have no factor correlations to supply, so the matrix is read as the
  # pattern it can only be -- and the reading is reported, because the bifactor one defines
  # coefficients the correlated-factors one does not.
  expect_warning(res <- efa_reliability(L, Phi = Phi, cormat = cm),
                 class = "efa_reliability_phi_as_pattern")
  expect_setequal(res$coefficient, c("omega_total", "omega_subscale", "alpha", "H"))

  # A Phi that cannot be the correlation matrix of those columns is named on its own, not
  # under a warning about a reading that is then abandoned.
  expect_error(efa_reliability(L, Phi = Phi2, cormat = cm),
               class = "efa_reliability_bad_phi")
  expect_no_warning(try(efa_reliability(L, Phi = Phi2, cormat = cm), silent = TRUE),
                    class = "efa_reliability_phi_as_pattern")

  # The components route names its general factor, so there a general factor together with
  # correlated group factors is a contradiction the arguments state outright, and it stays
  # refused rather than reinterpreted.
  expect_error(efa_reliability(NULL, var_names = rownames(L), g_load = L[, 1],
                               s_load = L[, -1, drop = FALSE], u2 = 1 - rowSums(L^2),
                               Phi = Phi2, variance = "sums_load"),
               class = cls)

  # The loading table of a Schmid-Leiman solution states that it is a hierarchy, so the
  # combination contradicts itself and takes the same refusal. Named for the reading rather
  # than for the shape of Phi: no Phi of any shape belongs with one.
  expect_error(efa_reliability(sl_mod$sl, Phi = Phi, cormat = cm), class = cls)
  expect_error(efa_reliability(sl_mod$sl, Phi = Phi2, cormat = cm), class = cls)

  # The refusal is the matrix routes only. A fitted solution carries its own factor
  # intercorrelations, so there Phi is still reported as ignored rather than refused.
  expect_warning(efa_reliability(sl_mod, factor_map = fc, Phi = Phi2, cormat = cm),
                 class = "efa_reliability_args_ignored")

  # One column is one factor, not a hierarchy. The only correlation matrix its factor can
  # have is the 1 by 1 identity, which says nothing about the solution, so it is checked and
  # the same coefficients come back with it or without it, classed or not.
  one <- efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF")$unrot_loadings
  res <- suppressMessages(efa_reliability(one, Phi = matrix(1, 1, 1), cormat = cm))
  expect_identical(res, suppressMessages(efa_reliability(unclass(one),
                                                         Phi = matrix(1, 1, 1),
                                                         cormat = cm)))
  expect_identical(res, suppressMessages(efa_reliability(unclass(one), cormat = cm)))
  # A Phi that cannot be that identity is still named on the unclassed route.
  expect_error(efa_reliability(unclass(one), Phi = matrix(2, 1, 1), cormat = cm),
               class = "efa_reliability_bad_phi")
})

test_that("the matrix route carries the factor correlations into the coefficients", {
  # Not a wiring test: the whole-scale omega total is the Phi-aware common variance over
  # the observed variance, and the Phi-blind reading is materially different.
  L <- efa_mod$rot_loadings
  Phi <- efa_mod$Phi
  cm <- test_models$baseline$cormat
  res <- efa_reliability(L, Phi = Phi, cormat = cm)
  tot <- res$value[res$coefficient == "omega_total" & res$factor == "total"]
  common <- unclass(L) %*% Phi %*% t(unclass(L))

  expect_equal(tot, sum(common) / sum(cm))
  expect_false(isTRUE(all.equal(tot, sum(tcrossprod(unclass(L))) / sum(cm))))

  # Model-implied variances need no correlation matrix; correlation mode does, and says so
  # rather than dividing by nothing.
  expect_s3_class(efa_reliability(L, Phi = Phi, variance = "sums_load"),
                  "efa_reliability")
  expect_error(efa_reliability(L, Phi = Phi), class = "efa_reliability_need_cormat")
})

test_that("a supplied cormat must hold the solution's variables", {
  cm <- test_models$baseline$cormat

  # The same variables in another order are reordered to the solution, so the
  # coefficients are the ones of the correctly ordered matrix rather than silently
  # wrong subscale values.
  expect_equal(efa_reliability(sl_mod, factor_map = fc, cormat = cm[18:1, 18:1]),
               efa_reliability(sl_mod, factor_map = fc, cormat = cm))
  # A different number of variables, or different ones, cannot be matched.
  expect_error(efa_reliability(sl_mod, factor_map = fc, cormat = cm[1:6, 1:6]),
               class = "efa_reliability_cormat_dim")
  other <- cm
  dimnames(other) <- list(paste0("X", 1:18), paste0("X", 1:18))
  expect_error(efa_reliability(sl_mod, factor_map = fc, cormat = other),
               class = "efa_reliability_cormat_names")
  # An unlabelled matrix of the right size is still matched by position.
  bare <- unname(cm)
  expect_equal(efa_reliability(sl_mod, factor_map = fc, cormat = bare),
               efa_reliability(sl_mod, factor_map = fc, cormat = cm))
})

test_that("a raw bifactor loading matrix is accepted", {
  L <- cbind(g = rep(0.5, 6), matrix(c(rep(0.4, 3), rep(0, 6), rep(0.4, 3)), 6, 2))
  colnames(L) <- c("g", "F1", "F2")
  rownames(L) <- paste0("V", 1:6)
  res <- efa_reliability(L)
  expect_s3_class(res, "efa_reliability")
  expect_setequal(unique(res$factor), c("g", "F1", "F2"))
})

## lavaan paths --------------------------------------------------------------

test_that("a lavaan bifactor fit matches OMEGA and yields one block", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))
  res <- efa_reliability(fit, g_name = "g")
  om <- OMEGA(fit, g_name = "g")

  expect_true(all(is.na(res$group)))       # single ungrouped fit
  expect_identical(attr(res, "settings"), list(variance = "sums_load",
                                               no_general = FALSE))
  get <- function(coef, fac) res$value[res$coefficient == coef & res$factor == fac]
  expect_equal(get("omega_total", "g"), unname(om["g", "tot"]))
  expect_equal(get("H", "F1"), unname(om["F1", "H"]))
})

test_that("a lavaan fit with correlated residuals counts them in the composite variance", {
  skip_if_not_installed("lavaan")
  # A composite's model-implied variance is 1' (Lambda Psi Lambda' + Theta) 1, so a freed
  # residual covariance belongs in it. The uniquenesses are only the diagonal of Theta, and
  # summing those alone leaves it out. Scored on a population bifactor built with two
  # residual covariances and fitted by the model that generated it, so the fit is exact and
  # the right answer is known rather than reproduced from the same code.
  fit_rc <- rc_fit_bifactor
  std <- suppressWarnings(lavaan::lavInspect(fit_rc, "std"))
  lam <- std$lambda
  common <- tcrossprod(lam[, "g"]) + tcrossprod(lam[, c("F1", "F2", "F3")])
  d <- diag(std$theta)
  Th <- rc_theta(std$theta)

  res <- efa_reliability(fit_rc, g_name = "g")
  get <- function(x, coef, fac) x$value[x$coefficient == coef & x$factor == fac]
  i1 <- seq_len(6)

  expect_equal(get(res, "omega_total", "g"), sum(common) / (sum(common) + sum(Th)))
  expect_equal(get(res, "omega_total", "F1"),
               sum(common[i1, i1]) / (sum(common[i1, i1]) + sum(Th[i1, i1])))
  # Standardized alpha follows the same composite variance, through the model-implied
  # correlation matrix it is computed from.
  expect_equal(get(res, "alpha", "g"), 18 / 17 * (1 - 18 / sum(common + Th)))
  expect_equal(get(res, "alpha", "F1"),
               6 / 5 * (1 - 6 / sum((common + Th)[i1, i1])))

  # Not an identity the uncorrected arithmetic also satisfies: dropping the off-diagonal
  # moves the subscale omega by more than .04 and the alpha by more than .01.
  expect_gt(abs(get(res, "omega_total", "F1") -
                sum(common[i1, i1]) / (sum(common[i1, i1]) + sum(diag(Th)[i1]))), 0.04)
  expect_gt(abs(get(res, "alpha", "F1") -
                6 / 5 * (1 - 6 / sum((common + diag(d))[i1, i1]))), 0.01)

  # The whole scale reproduces the population value the fit was built from.
  expect_equal(get(res, "omega_total", "g"), sum(rc$common) / sum(rc$Sigma),
               tolerance = 1e-6)

  # OMEGA() reaches the core through its own front end and is corrected with it.
  om <- OMEGA(fit_rc, g_name = "g")
  expect_equal(unname(om["g", "tot"]), get(res, "omega_total", "g"))
  expect_equal(unname(om["F1", "tot"]), get(res, "omega_total", "F1"))

  # A second-order fit reaches the same expression after the Schmid-Leiman transform.
  fit_ho <- suppressWarnings(lavaan::cfa(paste(rc$second_order, "V1 ~~ V2", sep = "\n"),
                                         sample.cov = rc$Sigma, sample.nobs = 1000,
                                         estimator = "ml", sample.cov.rescale = FALSE))
  res_ho <- suppressMessages(efa_reliability(fit_ho, g_name = "g"))
  sh <- suppressWarnings(lavaan::lavInspect(fit_ho, "std"))
  fo <- c("F1", "F2", "F3")
  L1 <- sh$lambda[, fo]
  # The Schmid-Leiman columns, written out rather than taken from the package's own helper,
  # so the expected value is derived independently of the code under test.
  sl <- tcrossprod(L1 %*% sh$beta[fo, "g"]) +
    tcrossprod(L1 %*% diag(sqrt(diag(sh$psi)[fo])))
  expect_equal(get(res_ho, "omega_total", "g"),
               sum(sl) / (sum(sl) + sum(rc_theta(sh$theta))))

  # Multiple groups are corrected per group, each from its own residual covariances.
  fit_gr <- suppressWarnings(lavaan::cfa(paste(rc$bifactor, "V1 ~~ V2", sep = "\n"),
                                         sample.cov = list(rc$Sigma, rc$Sigma),
                                         sample.nobs = c(1000, 1000), estimator = "ml",
                                         orthogonal = TRUE, sample.cov.rescale = FALSE))
  res_gr <- efa_reliability(fit_gr, g_name = "g")
  tot_gr <- res_gr$value[res_gr$coefficient == "omega_total" & res_gr$factor == "F1"]
  expect_length(tot_gr, 2L)
  # Each block against its own group's residual covariances, not against the other's: the
  # two groups are fitted separately, so their parameters agree only to optimizer tolerance.
  std_gr <- suppressWarnings(lavaan::lavInspect(fit_gr, "std",
                                                drop.list.single.group = FALSE))
  for (k in seq_along(std_gr)) {
    lam_k <- std_gr[[k]]$lambda
    common_k <- tcrossprod(lam_k[, "g"]) + tcrossprod(lam_k[, c("F1", "F2", "F3")])
    Th_k <- rc_theta(std_gr[[k]]$theta)
    expect_equal(tot_gr[k],
                 sum(common_k[i1, i1]) / (sum(common_k[i1, i1]) + sum(Th_k[i1, i1])))
    expect_gt(abs(tot_gr[k] - sum(common_k[i1, i1]) /
                    (sum(common_k[i1, i1]) + sum(diag(Th_k)[i1]))), 0.01)
  }

  # And the correction is confined to fits that have residual covariances: with a diagonal
  # Theta the composite variance stays the loading-and-uniqueness expression it always was.
  fit_plain <- suppressWarnings(lavaan::cfa(rc$bifactor, sample.cov = rc$Sigma,
                                            sample.nobs = 1000,
                                            estimator = "ml", orthogonal = TRUE,
                                            sample.cov.rescale = FALSE))
  sp <- suppressWarnings(lavaan::lavInspect(fit_plain, "std"))
  cp <- tcrossprod(sp$lambda[, "g"]) + tcrossprod(sp$lambda[, c("F1", "F2", "F3")])
  res_plain <- efa_reliability(fit_plain, g_name = "g")
  expect_equal(get(res_plain, "omega_total", "g"),
               sum(cp) / (sum(cp) + sum(diag(sp$theta))))
  expect_equal(get(res_plain, "omega_total", "F1"),
               sum(cp[i1, i1]) / (sum(cp[i1, i1]) + sum(diag(sp$theta)[i1])))
})

test_that("the residual-covariance correction reaches every lavaan spec shape", {
  skip_if_not_installed("lavaan")
  # The lavaan adapter builds three spec shapes, and each has to carry the model-implied
  # correlation matrix. The block above covers the bifactor and second-order one; these are
  # the other two, built at their own sites.
  i1 <- seq_len(6)
  get <- function(x, coef, fac) x$value[x$coefficient == coef & x$factor == fac]

  cf <- suppressWarnings(lavaan::cfa(paste(rc$correlated, "V1 ~~ V2", sep = "\n"),
                                     sample.cov = rc$Sigma, sample.nobs = 1000,
                                     estimator = "ml", sample.cov.rescale = FALSE))
  s_cf <- suppressWarnings(lavaan::lavInspect(cf, "std"))
  common_cf <- s_cf$lambda %*% s_cf$psi %*% t(s_cf$lambda)
  Th_cf <- rc_theta(s_cf$theta)
  res_cf <- suppressMessages(efa_reliability(cf))
  expect_equal(get(res_cf, "omega_total", "F1"),
               sum(common_cf[i1, i1]) / (sum(common_cf[i1, i1]) + sum(Th_cf[i1, i1])))
  expect_gt(abs(get(res_cf, "omega_total", "F1") - sum(common_cf[i1, i1]) /
                  (sum(common_cf[i1, i1]) + sum(diag(Th_cf)[i1]))), 0.02)

  # A single factor is normalized to its own spec, which rebuilds the list and so has to
  # carry the matrix over rather than drop it.
  sf <- suppressWarnings(lavaan::cfa(paste(rc$single, "V1 ~~ V2", sep = "\n"),
                                     sample.cov = rc$Sigma, sample.nobs = 1000,
                                     estimator = "ml", sample.cov.rescale = FALSE))
  s_sf <- suppressWarnings(lavaan::lavInspect(sf, "std"))
  common_sf <- tcrossprod(s_sf$lambda[, 1])
  Th_sf <- rc_theta(s_sf$theta)
  res_sf <- suppressMessages(efa_reliability(sf))
  tot_sf <- res_sf$value[res_sf$coefficient == "omega_total"]
  expect_equal(tot_sf, sum(common_sf) / (sum(common_sf) + sum(Th_sf)))
  expect_gt(abs(tot_sf - sum(common_sf) / (sum(common_sf) + sum(diag(Th_sf)))), 0.005)
  # OMEGA()'s single-factor vector is corrected with it.
  expect_equal(unname(suppressMessages(OMEGA(sf))["Omega"]), tot_sf)
})

test_that("the correlated-residual coefficients match semTools", {
  skip_on_cran()
  skip_if_not_installed("lavaan")
  skip_if_not_installed("semTools")
  # An independent implementation of the same model-based omega. semTools::compRelSEM()
  # counts the residual covariances in the composite variance, which is the quantity under
  # test here; obs.var = FALSE selects the model-implied total variance it is defined on.
  res <- efa_reliability(rc_fit_bifactor, g_name = "g")
  ref <- unlist(semTools::compRelSEM(rc_fit_bifactor, obs.var = FALSE,
                                     return.total = TRUE))

  get <- function(fac) res$value[res$coefficient == "omega_total" & res$factor == fac]
  expect_equal(get("g"), unname(ref[".TOTAL."]), tolerance = 1e-6)
  for (f in c("F1", "F2", "F3")) {
    expect_equal(get(f), unname(ref[f]), tolerance = 1e-6)
  }
})

test_that("a lavaan correlated-factors fit is scored as the oblique solution it is", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml"))

  # How the fit was read is stated rather than left to be inferred from a missing column.
  expect_message(efa_reliability(fit), class = "efa_reliability_no_general_factor")
  res <- suppressMessages(efa_reliability(fit))

  # The correlated-factors menu: the coefficients a solution with no general factor does
  # not define are omitted, not returned as the zeros the arithmetic would give.
  expect_setequal(res$coefficient, c("omega_total", "omega_subscale", "alpha", "H"))
  expect_false(any(res$coefficient == "omega_subscale" & res$factor == "total"))
  expect_false(any(res$coefficient == "H" & res$factor == "total"))
  expect_true(all(is.na(res$group)))
  expect_identical(attr(res, "settings"), list(variance = "sums_load",
                                               no_general = TRUE))

  # The values are the model-implied ones, as for every lavaan input: the composite's
  # common variance Lambda Psi Lambda' over that variance plus the residual variances.
  std <- suppressWarnings(lavaan::lavInspect(fit, "std",
                                             drop.list.single.group = FALSE))[[1]]
  common <- std$lambda %*% std$psi %*% t(std$lambda)
  expect_equal(res$value[res$coefficient == "omega_total" & res$factor == "total"],
               sum(common) / (sum(common) + sum(diag(std$theta))))
  # Ignoring the factor correlations would give a materially different value, so the
  # identity above is not one the orthogonal reading would also satisfy.
  expect_false(isTRUE(all.equal(sum(common), sum(tcrossprod(std$lambda)))))

  # `g_name` names a general factor, which this fit has none of: it is read from the
  # structure, so naming one of the ordinary factors changes nothing.
  expect_equal(suppressMessages(efa_reliability(fit, g_name = "F1")), res)

  # A multiple-group fit is scored per group, as the supported shapes are.
  fit_gr <- suppressWarnings(lavaan::cfa(mod, sample.cov =
    list(test_models$baseline$cormat, test_models$baseline$cormat),
    sample.nobs = c(500, 500), estimator = "ml"))
  res_gr <- suppressMessages(efa_reliability(fit_gr, group_names = c("Young", "Old")))
  expect_identical(unique(res_gr$group), c("Young", "Old"))
  expect_setequal(res_gr$coefficient, c("omega_total", "omega_subscale", "alpha", "H"))

  # The printed explanation of the two omega columns has to describe the table it stands
  # above for a grouped result as well. Every factor label repeats once per group here, so
  # a comparison keyed on the label alone matches only each factor's first group and can
  # never find the columns equal -- and under exact simple structure they are equal.
  tot <- res_gr$value[res_gr$coefficient == "omega_total" & res_gr$level == "group"]
  sub_om <- res_gr$value[res_gr$coefficient == "omega_subscale" & res_gr$level == "group"]
  expect_equal(tot, sub_om)
  testthat::local_reproducible_output()
  txt <- paste(cli::ansi_strip(format(res_gr)), collapse = " ")
  expect_match(txt, "subscale omega equals its total omega")

  # OMEGA() keeps its own surface: its wide per-factor output reports the general
  # factor's omega hierarchical, ECV, and PUC, which this fit does not define.
  expect_error(OMEGA(fit), class = "efa_reliability_invalid_lavaan")
  expect_error(OMEGA(fit, g_name = "F1"), class = "efa_reliability_invalid_lavaan")
})

test_that("a fit that is not a correlated-factors one still needs its general factor named", {
  skip_if_not_installed("lavaan")
  # A variable on two or more factors is the bifactor structure, and a fit with one is
  # read as a bifactor solution -- so its general factor still has to be named, and a
  # correlated-factors model with a cross-loading is not routed round that.
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6 + V7
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml"))
  expect_error(efa_reliability(fit), class = "efa_reliability_g_name")

  # And a genuine bifactor whose factors were left correlated stays refused: the general
  # and group parts of such a model do not partition a composite's variance, which is a
  # different problem from having no general factor and is named as one.
  mod_bi <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
             F2 =~ V7 + V8 + V9 + V10 + V11 + V12
             F3 =~ V13 + V14 + V15 + V16 + V17 + V18
             g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                  V13 + V14 + V15 + V16 + V17 + V18'
  fit_bi <- suppressWarnings(lavaan::cfa(mod_bi, sample.cov = test_models$baseline$cormat,
                                         sample.nobs = 500, estimator = "ml",
                                         orthogonal = FALSE))
  expect_error(efa_reliability(fit_bi, g_name = "g"),
               class = "efa_reliability_correlated_factors")
})

test_that("a lavaan single-factor fit surfaces omega total, alpha, and H, and informs", {
  skip_if_not_installed("lavaan")
  fit <- suppressWarnings(lavaan::cfa(
    'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
          V13 + V14 + V15 + V16 + V17 + V18',
    sample.cov = test_models$baseline$cormat, sample.nobs = 500,
    estimator = "ml", orthogonal = TRUE))

  expect_message(efa_reliability(fit, g_name = "g"),
                 class = "efa_reliability_single_factor")
  res <- suppressMessages(efa_reliability(fit, g_name = "g"))
  expect_setequal(res$coefficient, c("omega_total", "alpha", "H"))
  expect_true(all(res$factor == "g"))

  # The values are the model-implied ones every lavaan input takes: omega total from the
  # loading sum over that sum plus the residual variances, standardized alpha from the
  # model-implied correlation matrix, and the H index over the loadings.
  std <- suppressWarnings(lavaan::lavInspect(fit, "std",
                                             drop.list.single.group = FALSE))[[1]]
  lambda <- std$lambda[, 1]
  theta <- diag(std$theta)
  k <- length(lambda)
  R_imp <- stats::cov2cor(tcrossprod(lambda) + diag(theta))
  get <- function(coef) res$value[res$coefficient == coef]
  expect_equal(get("omega_total"), sum(lambda)^2 / (sum(lambda)^2 + sum(theta)))
  expect_equal(get("alpha"), k / (k - 1) * (1 - sum(diag(R_imp)) / sum(R_imp)))
  expect_equal(get("H"), 1 / (1 + 1 / sum(lambda^2 / (1 - lambda^2))))

  # Both expectations below are about the undefined-coefficient note alone, and every call
  # they make also raises the single-factor one, which would otherwise print outside any
  # expectation. Muffled by class rather than with suppressMessages(), which would hide the
  # note the expectations are about.
  quiet <- function(expr) {
    withCallingHandlers(
      expr,
      efa_reliability_single_factor = function(m) invokeRestart("muffleMessage"))
  }

  # Alpha assumes essential tau-equivalence, which is nested in a one-factor model, so it
  # is defined here and requesting it returns it instead of reporting it undefined.
  expect_no_message(quiet(efa_reliability(fit, g_name = "g", coefficients = "alpha")),
                    class = "efa_reliability_coef_undefined")
  sel <- suppressMessages(efa_reliability(fit, g_name = "g", coefficients = "alpha"))
  expect_identical(nrow(sel), 1L)
  expect_equal(sel$value, get("alpha"))

  # The coefficients a single factor genuinely does not define still report as undefined.
  expect_message(
    quiet(efa_reliability(fit, g_name = "g", coefficients = "omega_subscale")),
    class = "efa_reliability_coef_undefined")
})

test_that("every input route scores one and the same one-factor solution alike", {
  skip_if_not_installed("lavaan")

  # One solution, handed in four ways. It is scored against its own model-implied
  # correlation matrix, which a one-factor model reproduces exactly: lavaan then recovers
  # the loadings the EFA already has, and the observed and model-implied total variances
  # coincide, so the input route is the only difference left between the four results.
  cm <- test_models$baseline$cormat
  vn <- rownames(cm)
  fit_efa <- efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF")
  L <- as.numeric(unclass(fit_efa$unrot_loadings))
  u2 <- 1 - L^2
  L_col <- matrix(L, ncol = 1, dimnames = list(vn, NULL))
  R_imp <- tcrossprod(L) + diag(u2)
  dimnames(R_imp) <- dimnames(cm)

  fit_lav <- suppressWarnings(lavaan::cfa(
    paste("g =~", paste(vn, collapse = " + ")),
    sample.cov = R_imp, sample.nobs = 500, estimator = "ml"))

  routes <- suppressMessages(list(
    efa_fit = efa_reliability(fit_efa, cormat = R_imp),
    matrix  = efa_reliability(L_col, cormat = R_imp),
    manual  = efa_reliability(NULL, var_names = vn, g_load = rep(0, length(L)),
                              s_load = L_col, u2 = u2, Phi = matrix(1, 1, 1),
                              cormat = R_imp),
    lavaan  = efa_reliability(fit_lav, g_name = "g")))

  # Every route reaches the same menu -- what one factor defines, and nothing else -- on the
  # one row a single-factor solution has, which is its general-factor row whatever the input
  # calls the factor.
  for (route in routes) {
    expect_setequal(route$coefficient, c("omega_total", "alpha", "H"))
    expect_true(all(route$level == "general"))
    expect_identical(length(unique(route$factor)), 1L)
  }

  # The label is the name the input gives the factor: the EFA's own column label, the name
  # lavaan's model syntax gives it, and -- for the two routes that name it nothing -- the
  # default first-factor label, which the EFA's happens to be as well. Not "g", which would
  # claim a general factor over group factors this solution has none of.
  expect_identical(unique(routes$efa_fit$factor),
                   colnames(unclass(fit_efa$unrot_loadings)))
  expect_identical(unique(routes$efa_fit$factor), "F1")
  expect_identical(unique(routes$matrix$factor), "F1")
  expect_identical(unique(routes$manual$factor), "F1")
  expect_identical(unique(routes$lavaan$factor), "g")   # the model calls it "g"

  # ... and the same values, not only the same coefficient names.
  vals <- function(x) stats::setNames(x$value, x$coefficient)
  expect_equal(vals(routes$matrix), vals(routes$efa_fit))
  expect_equal(vals(routes$manual), vals(routes$efa_fit))
  # lavaan re-estimates the solution rather than being handed it, so it agrees to the
  # precision of its own ML fit of a model this matrix satisfies exactly.
  expect_equal(vals(routes$lavaan), vals(routes$efa_fit), tolerance = 1e-6)

  # The agreement is on the right numbers: each coefficient is what its definition gives
  # for this solution, so three routes matching a fourth is not three routes sharing a bug.
  k <- length(L)
  expect_equal(vals(routes$efa_fit)[["omega_total"]], sum(L)^2 / (sum(L)^2 + sum(u2)))
  expect_equal(vals(routes$efa_fit)[["alpha"]],
               k / (k - 1) * (1 - sum(diag(R_imp)) / sum(R_imp)))
  expect_equal(vals(routes$efa_fit)[["H"]], 1 / (1 + 1 / sum(L^2 / (1 - L^2))))
})

test_that("a one-column loading matrix is scored, not refused as a pattern matrix", {
  # The refusal of an ordinary loading matrix asks the caller either for the factor
  # intercorrelations or for a hierarchy, and a solution with one factor has neither to
  # give: one factor has no correlations with other factors, and one column is no
  # hierarchy for it to be mistaken for. So a one-column matrix is read as the single
  # factor it is, whether or not it carries the loading class of the solution it came
  # from -- which is what a one-factor `efa_fit()` object's loadings do carry.
  cm <- test_models$baseline$cormat
  fit <- efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF")
  L <- fit$unrot_loadings
  expect_s3_class(L, "efa_loadings")

  expect_message(efa_reliability(L, cormat = cm),
                 class = "efa_reliability_single_factor")
  res <- suppressMessages(efa_reliability(L, cormat = cm))
  expect_setequal(res$coefficient, c("omega_total", "alpha", "H"))
  expect_identical(unique(res$factor), "F1")

  # The class diverts nothing: the same matrix stripped of it takes the same route, and
  # the object the matrix came from reaches the same coefficients through its own adapter.
  expect_identical(res, suppressMessages(efa_reliability(unclass(L), cormat = cm)))
  expect_equal(res, suppressMessages(efa_reliability(fit)))
  # The superseded loading class is read the same way.
  L_old <- unclass(L)
  class(L_old) <- "LOADINGS"
  expect_identical(suppressMessages(efa_reliability(L_old, cormat = cm)), res)

  # Nor does it divert the model-implied correlation matrix the bare-matrix route falls
  # back to when `cormat` is absent, or either variance convention.
  expect_identical(suppressMessages(efa_reliability(L)),
                   suppressMessages(efa_reliability(unclass(L))))
  expect_identical(suppressMessages(efa_reliability(L, variance = "sums_load")),
                   suppressMessages(efa_reliability(unclass(L), variance = "sums_load")))

  # A single factor's Phi can only be the 1 x 1 identity, which says nothing about the
  # solution, so the route that reads the matrix as a correlated-factors one lands on the
  # same single factor and the same coefficients.
  expect_equal(suppressMessages(efa_reliability(L, Phi = matrix(1, 1, 1), cormat = cm)),
               res)

  # The exemption stops at one column: with two or more, nothing still says whether they
  # are a hierarchy or correlated factors, and the matrix is refused as before.
  two_col <- unclass(efa_mod$rot_loadings)[, 1:2, drop = FALSE]
  class(two_col) <- c("efa_loadings", "LOADINGS")
  expect_error(efa_reliability(two_col, cormat = cm),
               class = "efa_reliability_pattern_matrix")
})

test_that("a fac_names with the wrong number of names is refused", {
  # One label per factor. Too few or too many otherwise reach rownames() on the coefficient
  # matrix and fail there in base R's terms -- an error that names neither the argument nor
  # the count it needed, and that is localized on a non-English machine.
  expect_error(efa_reliability(sl_mod, factor_map = fc, fac_names = c("A", "B")),
               class = "efa_reliability_fac_names_length")
  expect_error(efa_reliability(sl_mod, factor_map = fc,
                               fac_names = c("A", "B", "C", "D")),
               class = "efa_reliability_fac_names_length")
  expect_error(efa_reliability(efa_mod, cormat = test_models$baseline$cormat,
                               fac_names = c("A", "B")),
               class = "efa_reliability_fac_names_length")
  # One name per group factor is what the solution has, so it is accepted and used.
  named <- efa_reliability(sl_mod, factor_map = fc, fac_names = c("A", "B", "C"))
  expect_setequal(unique(named$factor), c("g", "A", "B", "C"))

  # A single-factor solution has one factor to name, and the general-factor row is not one
  # of them -- so two names are as wrong here as they are above, and are refused rather
  # than dropped in favour of the default label.
  cm <- test_models$baseline$cormat
  fit1 <- efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF")
  expect_error(suppressMessages(efa_reliability(fit1, fac_names = c("A", "B"))),
               class = "efa_reliability_fac_names_length")
  expect_error(suppressMessages(efa_reliability(fit1$unrot_loadings, cormat = cm,
                                                fac_names = character(0))),
               class = "efa_reliability_fac_names_length")
  expect_identical(unique(suppressMessages(
    efa_reliability(fit1, fac_names = "Impulsivity"))$factor), "Impulsivity")
})

test_that("fac_names labels the one factor of a single-factor solution", {
  cm <- test_models$baseline$cormat
  vn <- rownames(cm)
  L <- as.numeric(unclass(efa_fit(cm, N = 500, n_factors = 1,
                                  estimator = "PAF")$unrot_loadings))
  L_col <- matrix(L, ncol = 1, dimnames = list(vn, NULL))
  manual <- function(...) suppressMessages(efa_reliability(
    NULL, var_names = vn, g_load = rep(0, length(L)), s_load = L_col,
    u2 = 1 - L^2, cormat = cm, ...))

  named <- manual(fac_names = "Impulsivity")
  expect_identical(unique(named$factor), "Impulsivity")
  # Still the whole-scale row: there are no group factors for it to be distinguished from,
  # which is why omega subscale is not among the coefficients.
  expect_true(all(named$level == "general"))
  expect_setequal(named$coefficient, c("omega_total", "alpha", "H"))
  # Only the label changes with it.
  expect_equal(named$value, manual()$value)

  # Named nothing, the factor takes the default first-factor label rather than "g", which
  # would claim a general factor over group factors a single-factor solution has none of.
  # The manual route's own default is the bare column position, which is not a name.
  expect_identical(unique(manual()$factor), "F1")
  expect_identical(unique(suppressMessages(efa_reliability(L_col, cormat = cm))$factor), "F1")
  # A name that happens to be a digit is still a name.
  expect_identical(unique(manual(fac_names = "1")$factor), "1")

  # A one-column matrix names its factor on the only column it has, which is read as the
  # general one; that name reaches the label, so the matrix and the components agree.
  named_col <- L_col
  colnames(named_col) <- "Impulsivity"
  expect_identical(
    unique(suppressMessages(efa_reliability(named_col, cormat = cm))$factor),
    unique(named$factor))
})

test_that("a general factor with one group factor is two factors, not one", {
  # The single-factor reading keys on how many factors a solution has, not on how many
  # columns any one slot holds: a bifactor solution with a single group factor still has
  # two, and keeps the coefficients that decompose a composite into them.
  L <- cbind(g = rep(0.5, 6), F1 = c(rep(0.4, 3), rep(0, 3)))
  rownames(L) <- paste0("V", 1:6)
  expect_no_message(efa_reliability(L), class = "efa_reliability_single_factor")
  res <- efa_reliability(L)
  expect_setequal(unique(res$factor), c("g", "F1"))
  expect_true(any(res$coefficient == "omega_hierarchical"))
  expect_true(any(res$coefficient == "omega_subscale"))
})

test_that("a one-factor efa_fit() is scored, whether or not a rotation was asked for", {
  cm <- test_models$baseline$cormat
  fit <- efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF")

  expect_message(efa_reliability(fit), class = "efa_reliability_single_factor")
  res <- suppressMessages(efa_reliability(fit))
  expect_setequal(res$coefficient, c("omega_total", "alpha", "H"))

  # A single factor cannot be rotated, so asking for a rotation changes neither the
  # solution nor its coefficients -- the unrotated loadings are read either way.
  rot <- suppressWarnings(suppressMessages(
    efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF", rotation = "promax")))
  expect_equal(suppressMessages(efa_reliability(rot)), res)

  # Scored against the object's own correlation matrix, as an oblique solution is.
  L <- as.numeric(unclass(fit$unrot_loadings))
  expect_equal(res$value[res$coefficient == "omega_total"],
               sum(tcrossprod(L)) / sum(cm))

  # A solution with more than one factor still needs an oblique rotation to be scored,
  # which is advice a one-factor solution could not follow.
  orth <- suppressWarnings(suppressMessages(
    efa_fit(cm, N = 500, n_factors = 3, estimator = "PAF", rotation = "varimax")))
  expect_error(efa_reliability(orth), class = "efa_reliability_not_oblique")
})

test_that("the one-factor menu does not depend on the variance convention", {
  # The two conventions are two total-variance denominators, not two coefficient menus:
  # they move omega total and leave the set of coefficients a single factor defines alone.
  fit <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 1, estimator = "PAF")
  corr <- suppressMessages(efa_reliability(fit, variance = "correlation"))
  sums <- suppressMessages(efa_reliability(fit, variance = "sums_load"))

  expect_setequal(corr$coefficient, sums$coefficient)
  expect_setequal(corr$coefficient, c("omega_total", "alpha", "H"))
  # Not vacuous: the denominators genuinely differ, the solution not reproducing the
  # observed correlations exactly.
  expect_false(isTRUE(all.equal(corr$value[corr$coefficient == "omega_total"],
                                sums$value[sums$coefficient == "omega_total"])))
})

test_that("the single-factor alpha is standardized alpha", {
  # An external check on the one coefficient the single-factor menu gained. Unlike the
  # psych::omega comparison above this needs no factor solution from psych -- alpha is a
  # closed-form function of the correlation matrix -- so it is cheap and stable enough to
  # run everywhere.
  skip_if_not_installed("psych")
  cm <- test_models$baseline$cormat
  fit <- efa_fit(cm, N = 500, n_factors = 1, estimator = "PAF")
  res <- suppressMessages(efa_reliability(fit))
  expect_equal(res$value[res$coefficient == "alpha"],
               suppressWarnings(psych::alpha(cm, n.obs = 500))$total$std.alpha,
               tolerance = 1e-8)
})

test_that("a lavaan second-order fit informs about the Schmid-Leiman transform", {
  skip_if_not_installed("lavaan")
  fit <- suppressWarnings(lavaan::cfa(
    'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
     F2 =~ V7 + V8 + V9 + V10 + V11 + V12
     F3 =~ V13 + V14 + V15 + V16 + V17 + V18
     g =~ F1 + F2 + F3',
    sample.cov = test_models$baseline$cormat, sample.nobs = 500, estimator = "ml"))
  expect_message(efa_reliability(fit, g_name = "g"),
                 class = "efa_reliability_g_second_order")
})

test_that("a lavaan multiple-group fit yields one labelled block per group", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod,
    sample.cov = list(test_models$baseline$cormat, test_models$baseline$cormat),
    sample.nobs = c(500, 500), estimator = "ml", orthogonal = TRUE))
  res <- efa_reliability(fit, g_name = "g", group_names = c("Young", "Old"))
  expect_identical(unique(res$group), c("Young", "Old"))

  # Duplicate labels would merge groups in the tidy output, so they abort.
  expect_error(efa_reliability(fit, g_name = "g", group_names = c("Dup", "Dup")),
               class = "efa_reliability_group_names")
})

test_that("every block of a multiple-group correlated-factors fit labels its own whole-scale row", {
  skip_if_not_installed("lavaan")
  # The whole-scale row is relabelled per block, so a fit with no general factor must
  # carry the correlated-factors labelling in every group and not only in the first.
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod,
    sample.cov = list(test_models$baseline$cormat, test_models$baseline$cormat),
    sample.nobs = c(500, 500), estimator = "ml"))
  res <- suppressMessages(efa_reliability(fit, group_names = c("Young", "Old")))

  for (grp in c("Young", "Old")) {
    block <- res[res$group == grp & res$coefficient == "omega_total", ]
    expect_identical(block$level, c("total", "group", "group", "group"))
    expect_identical(block$factor, c("total", "F1", "F2", "F3"))
  }
  # No block keeps the general-factor labelling of a solution that has one.
  expect_false(any(res$level == "general"))
  expect_false(any(res$factor == "g"))
})

test_that("a multiple-group single-factor fit yields one single-factor block per group", {
  skip_if_not_installed("lavaan")
  # The only path that assembles several single-factor matrices into one result: each block
  # has to be labelled, blanked, and scored on its own, and the note stated once for all.
  cm <- test_models$baseline$cormat
  fit <- suppressWarnings(lavaan::cfa(
    paste("ability =~", paste(rownames(cm), collapse = " + ")),
    sample.cov = list(cm, cm), sample.nobs = c(500, 500), estimator = "ml"))

  expect_message(efa_reliability(fit, group_names = c("Young", "Old")),
                 class = "efa_reliability_single_factor")
  res <- suppressMessages(efa_reliability(fit, group_names = c("Young", "Old")))

  expect_identical(unique(res$group), c("Young", "Old"))
  for (grp in c("Young", "Old")) {
    block <- res[res$group == grp, , drop = FALSE]
    expect_setequal(block$coefficient, c("omega_total", "alpha", "H"))
    # The factor's name in the model, on the whole-scale row.
    expect_true(all(block$factor == "ability"))
    expect_true(all(block$level == "general"))
  }
  # Identical groups, so the two blocks must agree coefficient by coefficient.
  expect_equal(res$value[res$group == "Young"], res$value[res$group == "Old"])
})

test_that("a bifactor with an under-loaded item informs about too few loadings", {
  skip_if_not_installed("lavaan")
  # g loads on all items except V11/V12, so those load on a single factor.
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))
  expect_message(efa_reliability(fit, g_name = "g"),
                 class = "efa_reliability_few_loadings")
})

test_that("efa_reliability is invariant across the oblique rotations", {
  skip_if_not_installed("GPArotation")

  # The factor columns are reordered by the number in their labels. A rotation whose
  # labels are absent, or carry no number, must therefore fall back to the columns'
  # own order: an unguarded reorder subscripts the loadings down to zero columns, and
  # the whole-scale omega total then silently becomes a function of the correlation
  # matrix alone while every per-factor row disappears. The whole-scale omega total is
  # rotation-invariant, so all oblique rotations must reach the same one.
  ref <- efa_reliability(efa_mod)
  g_tot <- function(x) x$value[x$coefficient == "omega_total" & x$factor == "total"]

  res_list <- lapply(.oblq_rotations, function(rot) {
    fit <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = rot)
    # bifactorQ leaves a group factor without salient items on this fixture, which is
    # reported by the classed efa_omega_empty_factor warning and is not under test here.
    suppressWarnings(efa_reliability(fit))
  })
  names(res_list) <- .oblq_rotations

  expect_true(all(vapply(res_list, nrow, integer(1)) > 2))
  expect_equal(vapply(res_list, g_tot, numeric(1)),
               stats::setNames(rep(g_tot(ref), length(res_list)), names(res_list)),
               tolerance = 1e-8)

  # An oblimin fit additionally keeps the full per-factor structure of a promax fit
  # of the same extraction.
  res_obl <- res_list[["oblimin"]]
  expect_setequal(res_obl$factor, c("total", "F1", "F2", "F3"))
  for (cf in c("omega_subscale", "H")) {
    expect_setequal(res_obl$factor[res_obl$coefficient == cf], c("F1", "F2", "F3"))
  }
  expect_setequal(res_obl$coefficient, unique(ref$coefficient))
})

## keying -------------------------------------------------------------------

# One factor's block of variables reverse-coded, which is what a scale whose
# reverse-worded items were never reverse-coded looks like. It exercises both shapes the
# check has to cover at once: in a Schmid-Leiman solution the block's general-factor
# loadings turn negative, whereas the oblique solution reflects the factor column instead
# and comes back with positive loadings throughout and a negative factor correlation.
flip <- rep(1, 18)
flip[7:12] <- -1
cormat_flipped <- diag(flip) %*% test_models$baseline$cormat %*% diag(flip)
dimnames(cormat_flipped) <- dimnames(test_models$baseline$cormat)
efa_flipped <- suppressWarnings(EFA(cormat_flipped, N = 500, n_factors = 3,
                                    type = "EFAtools", method = "PAF",
                                    rotation = "promax"))
sl_flipped <- suppressWarnings(efa_schmid_leiman(
  efa_flipped, estimate_control = estimate_control(type = "EFAtools"),
  estimator = "PAF"))

test_that("the fixture really is mis-keyed in both of the shapes the check must cover", {
  # Guards the tests below: were the fixture to lose either shape, they would still pass
  # while covering only one of the two routes.
  expect_true(any(sl_flipped$sl[, 1] < 0))
  expect_true(any(efa_flipped$Phi[upper.tri(efa_flipped$Phi)] < 0))
})

test_that("variables keyed against each other warn on every non-lavaan input route", {
  cls <- "efa_reliability_mixed_keying"
  sl <- unclass(sl_flipped$sl)
  g <- sl[, 1]
  s <- sl[, 2:4, drop = FALSE]
  u2 <- sl[, "u2"]

  expect_warning(efa_reliability(sl_flipped), class = cls)
  # The reverse-keyed block also leaves one of this fixture's subscale composites with a
  # smaller observed variance than the loadings imply, which trips the independent
  # out-of-range guard; muffle it so this expectation is about the keying alone.
  expect_warning(
    withCallingHandlers(efa_reliability(efa_flipped),
                        efa_omega_out_of_range = function(w) invokeRestart("muffleWarning")),
    class = cls)
  expect_warning(efa_reliability(cbind(g = g, s), u2 = u2, cormat = cormat_flipped),
                 class = cls)
  expect_warning(efa_reliability(NULL, var_names = rownames(sl), g_load = g,
                                 s_load = s, u2 = u2, cormat = cormat_flipped),
                 class = cls)

  skip_if_not_installed("psych")
  sm <- suppressWarnings(psych::schmid(cormat_flipped, nfactors = 3, fm = "minres",
                                       rotate = "promax"))
  expect_warning(efa_reliability(sm), class = cls)
})

test_that("a lavaan fit of mis-keyed variables warns, single-factor path included", {
  skip_if_not_installed("lavaan")
  cls <- "efa_reliability_mixed_keying"
  items <- paste0("V", 1:18)

  bi <- suppressWarnings(lavaan::cfa(
    paste('F1 =~ V1 + V2 + V3 + V4 + V5 + V6
           F2 =~ V7 + V8 + V9 + V10 + V11 + V12
           F3 =~ V13 + V14 + V15 + V16 + V17 + V18
           g =~', paste(items, collapse = " + ")),
    sample.cov = cormat_flipped, sample.nobs = 500, estimator = "ml",
    orthogonal = TRUE))
  expect_warning(efa_reliability(bi, g_name = "g"), class = cls)

  # The single-factor path reaches the core on a spec with no group factors, which no
  # other input to the check has, so it needs its own coverage.
  sf <- suppressWarnings(lavaan::cfa(paste("g =~", paste(items, collapse = " + ")),
                                     sample.cov = cormat_flipped, sample.nobs = 500,
                                     estimator = "ml"))
  expect_warning(suppressMessages(efa_reliability(sf, g_name = "g")), class = cls)
})

test_that("a correctly keyed solution raises no keying warning", {
  cls <- "efa_reliability_mixed_keying"
  expect_no_warning(efa_reliability(sl_mod, factor_map = fc), class = cls)
  expect_no_warning(efa_reliability(efa_mod), class = cls)
  expect_no_warning(efa_reliability(sl_mod$sl), class = cls)

  skip_if_not_installed("lavaan")
  items <- paste0("V", 1:18)
  sf <- suppressWarnings(lavaan::cfa(paste("g =~", paste(items, collapse = " + ")),
                                     sample.cov = test_models$baseline$cormat,
                                     sample.nobs = 500, estimator = "ml"))
  expect_no_warning(suppressMessages(efa_reliability(sf, g_name = "g")), class = cls)
})

test_that("the keying warning changes no coefficient", {
  # The check is a diagnostic only; the mis-keyed solution's coefficients are still the
  # ones the definitions give.
  res <- suppressWarnings(efa_reliability(sl_flipped))
  get <- function(x, coef, fac) x$value[x$coefficient == coef & x$factor == fac]

  # The whole-scale omega total is still the composite's model-implied common variance
  # over its observed variance -- collapsed by the keying, not by the warning.
  spec <- .rel_adapt_SL(sl_flipped, type = "psych")
  L <- cbind(spec$g_load, spec$s_load)
  expect_equal(get(res, "omega_total", "g"), sum(tcrossprod(L)) / sum(spec$cormat))

  # The same solution scored through the core directly, which is where the check sits, is
  # identical to what the front end returns.
  core <- .reliability_result(
    suppressWarnings(.reliability_core(spec, "correlation", add_ind = TRUE,
                                       add_rel = TRUE, arg = "factor_map")),
    settings = list(variance = "correlation", no_general = FALSE))
  expect_equal(res, core)
})

test_that("print output is stable", {
  local_reproducible_output()

  # full menu, single group (reliability and common-variance blocks)
  expect_snapshot(print(efa_reliability(sl_mod, factor_map = fc)),
                  transform = scrub_num)

  # coefficient selector (only the requested columns)
  expect_snapshot(print(efa_reliability(sl_mod, factor_map = fc,
                                        coefficients = c("omega_total", "alpha"))),
                  transform = scrub_num)

  # oblique EFA (no general factor): the total row omits omega subscale and H, which print
  # blank rather than as "NA", and the context line separates tot from sub.
  expect_snapshot(print(efa_reliability(efa_mod)), transform = scrub_num)
})

test_that("the correlated-factors context line follows the printed columns", {
  # It relates the tot and sub columns, so it is shown only when both are printed --
  # not for a selection that drops omega subscale, nor for a solution that has a
  # general factor.
  expect_match(format(efa_reliability(efa_mod)), "Correlated-factors", all = FALSE)
  sub_dropped <- suppressMessages(
    efa_reliability(efa_mod, coefficients = c("omega_total", "alpha")))
  expect_false(any(grepl("Correlated-factors", format(sub_dropped))))
  expect_false(any(grepl("Correlated-factors",
                         format(efa_reliability(sl_mod, factor_map = fc)))))

  # Which relation it states follows the printed values. An oblique fit has factor
  # correlations, so its totals exceed its subscale omegas.
  obl <- format(efa_reliability(efa_mod))
  expect_match(obl, "receives from every factor", all = FALSE)
  expect_false(any(grepl("equals its total omega", obl)))

  # Components supplied with a zero general factor, no factor correlations and simple
  # structure are a correlated-factors solution too, but nothing lets a composite draw
  # on another factor, so the two columns coincide and the line must say so instead.
  mk <- function(s) {
    u <- 1 - rowSums(s^2)
    efa_reliability(model = NULL, var_names = paste0("V", 1:6),
                    g_load = rep(0, 6), s_load = s, u2 = u,
                    cormat = s %*% t(s) + diag(u))
  }
  s <- matrix(0, 6, 2)
  s[1:3, 1] <- c(0.7, 0.6, 0.5)
  s[4:6, 2] <- c(0.7, 0.6, 0.5)
  no_phi <- mk(s)
  tot <- no_phi$value[no_phi$coefficient == "omega_total" & no_phi$level == "group"]
  sub_om <- no_phi$value[no_phi$coefficient == "omega_subscale"]
  expect_equal(tot, sub_om)
  txt <- format(no_phi)
  expect_match(txt, "equals its total omega", all = FALSE)
  expect_false(any(grepl("receives from every factor", txt)))

  # Cross-loadings alone are enough to separate the two columns, even with no factor
  # correlations at all: a composite still receives true score variance from the other
  # factor through them. The line must then take the general wording.
  s_cross <- s
  s_cross[3, 2] <- 0.3
  s_cross[6, 1] <- 0.25
  cross <- mk(s_cross)
  tot_c <- cross$value[cross$coefficient == "omega_total" & cross$level == "group"]
  sub_c <- cross$value[cross$coefficient == "omega_subscale"]
  expect_true(all(tot_c > sub_c))
  txt_c <- format(cross)
  expect_match(txt_c, "receives from every factor", all = FALSE)
  expect_false(any(grepl("equals its total omega", txt_c)))
})

test_that("the model error names the averaged matrix an efa_average object carries", {
  # An efa_average object is rejected while "a bifactor loading matrix" is listed as
  # acceptable, and its averaged loadings are a matrix. The message now says where they are.
  avg <- suppressWarnings(suppressMessages(
    efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                estimator = c("PAF", "ULS"), rotation = "promax", type = "EFAtools",
                start_method = "psych", show_progress = FALSE)))

  e <- tryCatch(efa_reliability(avg), efa_reliability_invalid_model = function(e) e)
  expect_s3_class(e, "efa_reliability_invalid_model")
  expect_snapshot(cat(conditionMessage(e)))

  # The component the message names clears this gate: what it meets next is the content
  # check for a correlated-factors pattern, which names `Phi` and `cormat` itself.
  e2 <- tryCatch(efa_reliability(avg$loadings$average), error = function(e) e)
  expect_false(inherits(e2, "efa_reliability_invalid_model"))
  expect_s3_class(
    suppressWarnings(efa_reliability(avg$loadings$average, Phi = avg$Phi$average,
                                     cormat = avg$orig_R)),
    "efa_reliability")
})

rm(efa_mod, sl_mod, fc, auto, flip, cormat_flipped, efa_flipped, sl_flipped)

## Tests for .reliability_result() -- the tidy long output built from computed
## coefficient matrices -- and its format/print method.

efa_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
sl_mod <- efa_schmid_leiman(efa_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF")
fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2
spec <- list(g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
             u2 = sl_mod$sl[, "u2"], map = fc, cormat = test_models$baseline$cormat,
             var_names = rownames(sl_mod$sl), fac_names = c("F1", "F2", "F3"))

# A full coefficient matrix (also carries the core's CR/AVE columns, which the result
# must drop).
mat <- .reliability_core(spec, "correlation", add_ind = TRUE, add_rel = TRUE)
res <- .reliability_result(mat, settings = list(variance = "correlation"))

# Multigroup: a named list of matrices, one entry per group.
res_mg <- .reliability_result(list(GroupA = mat, GroupB = mat),
                              settings = list(variance = "correlation"))

# Single factor: a one-row general-factor matrix carrying only omega total and H.
g_sf <- rep(0.7, 6)
u2_sf <- 1 - g_sf^2
mat_sf <- matrix(c(sum(g_sf)^2 / (sum(g_sf)^2 + sum(u2_sf)), .h_index(g_sf)),
                 nrow = 1, dimnames = list("g", c("tot", "H")))
res_sf <- .reliability_result(mat_sf, settings = list(variance = "sums_load"))

test_that("the result is a long data.frame with the documented schema", {
  expect_s3_class(res, "efa_reliability")
  expect_s3_class(res, "data.frame")
  expect_identical(names(res), c("coefficient", "level", "factor", "group", "value"))
  expect_type(res$coefficient, "character")
  expect_type(res$level, "character")
  expect_type(res$factor, "character")
  expect_type(res$group, "character")
  expect_type(res$value, "double")

  # level is general for the g row, group for the group factors, nothing else.
  expect_setequal(res$level, c("general", "group"))
  expect_true(all(res$level[res$factor == "g"] == "general"))
  expect_true(all(res$level[res$factor != "g"] == "group"))

  # A single unnamed group carries an NA group label.
  expect_true(all(is.na(res$group)))
})

test_that("only the surfaced menu is carried; CR/AVE/GLB/beta never appear", {
  # mat carries CR and AVE, but the result surfaces neither, nor any deferred surrogate.
  expect_true(all(c("CR", "AVE") %in% colnames(mat)))
  expect_false(any(c("CR", "AVE", "GLB", "beta") %in% res$coefficient))
  expect_setequal(res$coefficient,
                  c("omega_total", "omega_hierarchical", "omega_subscale",
                    "alpha", "H", "ECV", "PUC"))
})

test_that("the kind attribute tags each surfaced coefficient", {
  kind <- attr(res, "kind")
  expect_type(kind, "character")
  expect_identical(unname(kind[c("ECV", "PUC")]), c("common_variance", "common_variance"))
  expect_true(all(kind[c("omega_total", "omega_hierarchical", "omega_subscale",
                         "alpha", "H")] == "reliability"))
})

test_that("the settings attribute round-trips", {
  expect_identical(attr(res, "settings"), list(variance = "correlation"))
})

test_that("NA cells are dropped; ECV and PUC appear only for the general factor", {
  expect_false(anyNA(res$value))
  expect_true(all(res$level[res$coefficient %in% c("ECV", "PUC")] == "general"))
  # No group-factor row carries ECV or PUC.
  grp_rows <- res[res$level == "group", ]
  expect_false(any(grp_rows$coefficient %in% c("ECV", "PUC")))
})

test_that("a named list yields one labelled block per group", {
  expect_identical(unique(res_mg$group), c("GroupA", "GroupB"))
  # Each group holds the same coefficient set.
  expect_setequal(res_mg$coefficient[res_mg$group == "GroupA"],
                  res_mg$coefficient[res_mg$group == "GroupB"])
})

test_that("an unnamed list labels groups by position rather than collapsing them", {
  # A list is always multigroup; unnamed entries must get distinct positional labels so
  # each group renders as its own block (they must not all fall to a single NA label).
  res_un <- .reliability_result(list(mat, mat_sf))
  expect_identical(unique(res_un$group), c("1", "2"))
  expect_false(anyNA(res_un$group))
  # Both groups keep their own coefficients.
  expect_setequal(res_un$coefficient[res_un$group == "2"], c("omega_total", "H"))
})

test_that("a matrix with no surfaced columns yields an empty result, not an error", {
  # A CR/AVE-only matrix carries no registry column; the builder returns a 0-row
  # efa_reliability rather than aborting on a NULL assembly.
  cr_only <- matrix(c(0.9, 0.8), nrow = 1, dimnames = list("g", c("CR", "AVE")))
  expect_no_error(empty <- .reliability_result(cr_only))
  expect_s3_class(empty, "efa_reliability")
  expect_identical(nrow(empty), 0L)
})

test_that("a single-factor matrix carries only its available coefficients", {
  expect_setequal(res_sf$coefficient, c("omega_total", "H"))
  expect_true(all(res_sf$factor == "g"))
  expect_true(all(res_sf$level == "general"))
})

test_that("the general factor is the first row, whatever it is labelled", {
  # The general factor is row one of every matrix the core builds; the label on it varies
  # (a single-factor solution carries its own factor's name there), so the level cannot be
  # read off the label. Reading it off the label would also mislabel a group factor a user
  # names "g" as the whole-scale row.
  named_sf <- .reliability_result(
    matrix(c(0.86, 0.87), nrow = 1, dimnames = list("ability", c("tot", "H"))))
  expect_true(all(named_sf$factor == "ability"))
  expect_true(all(named_sf$level == "general"))

  collide <- .reliability_result(
    matrix(c(0.8, 0.7, 0.6, 0.5), nrow = 2,
           dimnames = list(c("F1", "g"), c("tot", "H"))))
  expect_identical(collide$level[collide$factor == "F1"], rep("general", 2L))
  expect_identical(collide$level[collide$factor == "g"], rep("group", 2L))
})

test_that("a solution with no general factor labels its whole-scale row for what it is", {
  # The core builds the first row for every solution, but a correlated-factors one has no
  # general factor to put there: the row holds the coefficients of the composite of every
  # variable. Labelled "g" at level "general" it would state a general factor the solution
  # does not define, so `no_general` relabels it. Read by position, as the level is, so a
  # group factor a user names "g" keeps its own label and level.
  mat_ng <- matrix(c(0.88, 0.77, 0.76, 0.75, 0.87, 0.76, 0.75, 0.74), nrow = 4,
                   dimnames = list(c("g", "F1", "F2", "g"), c("tot", "alpha")))
  ng <- .reliability_result(mat_ng, settings = list(variance = "correlation",
                                                    no_general = TRUE))
  expect_identical(ng$factor[ng$coefficient == "omega_total"],
                   c("total", "F1", "F2", "g"))
  expect_identical(ng$level[ng$coefficient == "omega_total"],
                   c("total", "group", "group", "group"))
  # The values are untouched: only the label of the first row moves.
  expect_identical(ng$value[ng$coefficient == "omega_total"],
                   c(0.88, 0.77, 0.76, 0.75))

  # The same matrix without the setting keeps the general-factor labelling, so the branch
  # is driven by the solution and not by the row's name.
  g_row <- .reliability_result(mat_ng, settings = list(variance = "correlation",
                                                       no_general = FALSE))
  expect_identical(g_row$factor[g_row$coefficient == "omega_total"],
                   c("g", "F1", "F2", "g"))
  expect_identical(g_row$level[g_row$coefficient == "omega_total"],
                   c("general", "group", "group", "group"))
})

test_that("a group factor sharing the whole-scale row's label keeps its own printed row", {
  local_reproducible_output()
  # The whole-scale row carries a label of its own -- "g" for a hierarchy, "total" for a
  # correlated-factors solution -- and a group factor may be named the same. The printed
  # table is keyed on the level and the factor together for that reason: keyed on the
  # label alone, both rows would be written into one and the group factor's value would be
  # printed as the whole scale's.
  mat <- matrix(c(0.88, 0.77, 0.76, 0.87, 0.76, 0.75), nrow = 3,
                dimnames = list(c("g", "total", "F2"), c("tot", "alpha")))
  res <- .reliability_result(mat, settings = list(variance = "correlation",
                                                  no_general = TRUE))
  # Both rows survive the reshape, at their own levels.
  tot <- res[res$coefficient == "omega_total", ]
  expect_identical(tot$level, c("total", "group", "group"))
  expect_identical(tot$factor, c("total", "total", "F2"))

  # And both reach the printed table, the whole-scale value among them.
  out <- format(res)
  expect_match(out, "\\.880", all = FALSE)
  expect_match(out, "\\.770", all = FALSE)
})

test_that("the correlated-factors note agrees with the table it introduces", {
  local_reproducible_output()
  # The note states whether the two omega columns coincide, so it has to be decided from
  # the values as printed: two coefficients less than one last decimal apart can still
  # straddle a rounding boundary and print as different numbers, which "equals" would
  # then contradict.
  build <- function(tot, sub) {
    m <- cbind(tot = c(0.9, tot), sub = c(NA_real_, sub))
    rownames(m) <- c("g", paste0("F", seq_along(tot)))
    .reliability_result(m, settings = list(variance = "correlation",
                                           no_general = TRUE))
  }
  txt <- function(x) gsub("\\s+", " ", paste(cli::ansi_strip(format(x)), collapse = " "))
  equal_note <- "subscale omega equals its total omega"

  expect_match(txt(build(c(0.769, 0.745), c(0.769, 0.745))), equal_note, fixed = TRUE)

  # Apart by 9e-4 -- inside the last printed decimal, but on either side of it.
  differ <- txt(build(c(0.7694, 0.745), c(0.7685, 0.745)))
  expect_false(grepl(equal_note, differ, fixed = TRUE))
  expect_match(differ, "F1 .769 .768", fixed = TRUE)
})

test_that("print output is stable", {
  local_reproducible_output()

  # single group, full menu (reliability and common-variance blocks)
  expect_snapshot(print(res), transform = scrub_num)

  # multiple groups (per-group headers)
  expect_snapshot(print(res_mg), transform = scrub_num)

  # single factor (reliability block only, omega total and H)
  expect_snapshot(print(res_sf), transform = scrub_num)
})

rm(efa_mod, sl_mod, fc, spec, mat, res, res_mg, g_sf, u2_sf, mat_sf, res_sf)

## Tests for the .rel_adapt_* front-end adapters. Each normalizes one input source
## to the spec consumed by .reliability_core(); the spec must drive the core to the
## same numbers the current OMEGA path produces for that input.

efa_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
sl_mod <- efa_schmid_leiman(efa_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF")
schmid_mod <- psych::schmid(test_models$baseline$cormat, nfactors = 3,
                            n.obs = 500, fm = "pa", rotate = "Promax")
fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2

test_that(".rel_adapt_SL reproduces the SL OMEGA path", {
  spec <- .rel_adapt_SL(sl_mod, factor_corres = fc, type = "EFAtools")

  expect_identical(
    .reliability_core(spec, "correlation"),
    .OMEGA_FLEX(sl_mod, type = "EFAtools", factor_corres = fc,
                variance = "correlation"))
  expect_identical(
    .reliability_core(spec, "sums_load"),
    .OMEGA_FLEX(sl_mod, type = "EFAtools", factor_corres = fc,
                variance = "sums_load"))
  expect_identical(
    .reliability_core(spec, "correlation", add_ind = FALSE),
    .OMEGA_FLEX(sl_mod, type = "EFAtools", factor_corres = fc,
                variance = "correlation", add_ind = FALSE))

  # The stored correlation matrix is picked up from the SL object.
  expect_identical(spec$cormat, sl_mod$orig_R)
})

test_that(".rel_adapt_schmid reproduces the psych::schmid OMEGA path", {
  # Cormat reconstructed from the schmid pattern and Phi.
  expect_equal(
    .reliability_core(.rel_adapt_schmid(schmid_mod, type = "psych"), "correlation"),
    .OMEGA_FLEX(schmid_mod, type = "psych", variance = "correlation"))

  # User-supplied cormat.
  expect_equal(
    .reliability_core(
      .rel_adapt_schmid(schmid_mod, type = "psych",
                        cormat = test_models$baseline$cormat), "correlation"),
    .OMEGA_FLEX(schmid_mod, type = "psych", cormat = test_models$baseline$cormat,
                variance = "correlation"))
})

test_that(".rel_adapt_manual reproduces the model = NULL OMEGA path", {
  spec <- .rel_adapt_manual(g_load = sl_mod$sl[, "g"],
                            s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                            u2 = sl_mod$sl[, "u2"],
                            var_names = rownames(sl_mod$sl),
                            factor_corres = fc, type = "EFAtools",
                            cormat = test_models$baseline$cormat)

  expect_identical(
    .reliability_core(spec, "correlation"),
    .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
                factor_corres = fc, variance = "correlation"))
  expect_identical(
    .reliability_core(spec, "sums_load"),
    .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
                g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
                u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
                factor_corres = fc, variance = "sums_load"))
})

test_that(".rel_adapt_manual checks cormat against the loadings, not the output labels", {
  # `var_names` names the result's rows; it is not a claim about what the correlation
  # matrix is labelled with. Relabelling the output must therefore not reject a correct
  # matrix -- the axis to match is the loadings' own row names.
  s_load <- unclass(sl_mod$sl)[, c("F1", "F2", "F3")]
  manual <- function(...) {
    .rel_adapt_manual(g_load = sl_mod$sl[, "g"], s_load = s_load,
                      u2 = sl_mod$sl[, "u2"], factor_corres = fc, type = "EFAtools",
                      ...)
  }
  cm <- test_models$baseline$cormat

  expect_identical(manual(var_names = paste0("Item", 1:18), cormat = cm)$cormat, cm)
  # The loadings' names still drive a permutation back into the solution's order.
  expect_identical(manual(var_names = paste0("Item", 1:18),
                          cormat = cm[18:1, 18:1])$cormat, cm)
  # Unlabelled components are matched by position: only the dimension is checked.
  bare <- s_load
  rownames(bare) <- NULL
  expect_identical(
    .rel_adapt_manual(g_load = sl_mod$sl[, "g"], s_load = bare, u2 = sl_mod$sl[, "u2"],
                      var_names = paste0("Item", 1:18), factor_corres = fc,
                      type = "EFAtools", cormat = cm[18:1, 18:1])$cormat,
    cm[18:1, 18:1])
  expect_error(manual(var_names = rownames(cm), cormat = cm[1:6, 1:6]),
               class = "efa_reliability_cormat_dim")
})

test_that(".rel_adapt_efa gives model-appropriate correlated-factors coefficients", {
  spec <- .rel_adapt_efa(efa_mod)

  # No general factor: g_load is zero, so omega hierarchical and ECV vanish.
  expect_true(all(spec$g_load == 0))
  expect_no_warning(om <- .reliability_core(spec, "correlation"))
  expect_true(all(om[, "hier"] == 0))
  expect_equal(unname(om["g", "ECV"]), 0)

  # Whole-scale omega total is the model-implied common variance of all variables over
  # their observed total variance.
  common <- spec$s_load %*% spec$Phi %*% t(spec$s_load)
  expect_equal(unname(om["g", "tot"]), sum(common) / sum(spec$cormat))

  # Each group row's omega subscale is that factor's congeneric omega from its own
  # indicators, while its omega total is the composite's full model-implied common
  # variance. The two differ whenever the composite draws on another factor, and on
  # this fit the total is the larger of the two (it need not be: the total is a
  # quadratic form in Phi and the cross terms can carry either sign).
  for (j in seq_len(ncol(spec$s_load))) {
    mem <- which(spec$map[, j] == 1)
    Vgr <- sum(spec$cormat[mem, mem])
    cong <- sum(spec$s_load[mem, j])^2 / Vgr
    expect_equal(unname(om[spec$fac_names[j], "sub"]), cong)
    expect_equal(unname(om[spec$fac_names[j], "hier"]), 0)
    expect_equal(unname(om[spec$fac_names[j], "tot"]), sum(common[mem, mem]) / Vgr)
    expect_gt(om[spec$fac_names[j], "tot"], cong)
  }

  # Works for a two-factor oblique EFA too (no Schmid-Leiman, so no identification
  # floor at three factors).
  efa2 <- EFA(test_models$baseline$cormat, N = 500, n_factors = 2,
              type = "EFAtools", method = "PAF", rotation = "promax")
  expect_no_error(.reliability_core(.rel_adapt_efa(efa2), "correlation"))
})

test_that(".rel_adapt_efa carries Phi in the column order of the loadings", {
  # The loading columns are ordered by the factor number in their labels, and the
  # group-row omega totals are a quadratic form in Phi, so Phi must be permuted the
  # same way or every correlated-factors total is scored against the wrong factors.
  efa_perm <- efa_mod
  colnames(efa_perm$rot_loadings) <- c("F2", "F3", "F1")
  spec <- .rel_adapt_efa(efa_perm)

  expect_equal(unname(spec$s_load), unname(unclass(efa_mod$rot_loadings)[, c(3, 1, 2)]))
  expect_equal(unname(spec$Phi), unname(efa_mod$Phi[c(3, 1, 2), c(3, 1, 2)]))
})

test_that(".rel_adapt_efa handles rotations that leave the loading columns unlabelled", {
  skip_if_not_installed("GPArotation")

  # efa_fit() labels the factor columns for every rotation, but the adapter also sees
  # loading matrices from other sources (a bare pattern matrix, an externally built fit),
  # where the labels can be missing. Unlabelled columns must not cost the adapter its
  # group factors, so construct that state explicitly.
  efa_obl <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                     estimator = "PAF", rotation = "oblimin")
  colnames(efa_obl$rot_loadings) <- NULL
  dimnames(efa_obl$Phi) <- NULL

  spec <- .rel_adapt_efa(efa_obl)

  # All three factors survive, and every group factor still gets a row label.
  expect_identical(dim(spec$s_load), c(18L, 3L))
  expect_identical(spec$fac_names, c("F1", "F2", "F3"))
  expect_identical(spec$s_load, unclass(efa_obl$rot_loadings))

  # Communalities come from the real pattern, not from an empty loading matrix.
  expect_true(all(spec$u2 > 0 & spec$u2 < 1))

  om <- .reliability_core(spec, "correlation")
  expect_identical(rownames(om), c("g", "F1", "F2", "F3"))
  # Each group factor scores a real congeneric omega. Tested on the group rows
  # only: the g row carries a "sub" value whether or not the group factors
  # survived, so it cannot discriminate a lost factor.
  expect_true(all(om[c("F1", "F2", "F3"), "sub"] > 0 &
                    om[c("F1", "F2", "F3"), "sub"] < 1))

  # Whole-scale omega total is 1' L Phi L' 1 / 1'R1, and L Phi L' is invariant to the
  # choice of oblique rotation, so the promax fit of the same extraction must agree on
  # it.
  om_pro <- .reliability_core(.rel_adapt_efa(efa_mod), "correlation")
  expect_equal(unname(om["g", "tot"]), unname(om_pro["g", "tot"]),
               tolerance = 1e-8)
})

test_that(".rel_adapt_bifactor reproduces the manual OMEGA path", {
  # A raw orthogonal-bifactor loading matrix: g in the first column.
  g_load <- rep(0.5, 6)
  s_load <- matrix(0, 6, 2)
  s_load[1:3, 1] <- 0.4
  s_load[4:6, 2] <- 0.4
  L <- cbind(g = g_load, s_load)
  colnames(L) <- c("g", "F1", "F2")
  rownames(L) <- paste0("V", 1:6)
  u2 <- 1 - rowSums(L^2)
  Rm <- L %*% t(L) + diag(u2)
  map <- abs(s_load) > 0

  spec <- .rel_adapt_bifactor(L, fac_names = c("F1", "F2"))
  expect_identical(spec$u2, u2)
  expect_identical(spec$cormat, Rm)

  ref_corr <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = rownames(L),
                          g_load = g_load, s_load = s_load, u2 = u2, cormat = Rm,
                          factor_corres = map, fac_names = c("F1", "F2"),
                          variance = "correlation")
  ref_sums <- .OMEGA_FLEX(model = NULL, type = "EFAtools", var_names = rownames(L),
                          g_load = g_load, s_load = s_load, u2 = u2,
                          factor_corres = map, fac_names = c("F1", "F2"),
                          variance = "sums_load")

  expect_identical(.reliability_core(spec, "correlation"), ref_corr)
  expect_identical(.reliability_core(spec, "sums_load"), ref_sums)
})

test_that("adapters reject a non-correlation cormat via .is_cormat", {
  bad <- matrix(rnorm(36), 6, 6)
  expect_error(.rel_adapt_SL(sl_mod, factor_corres = fc, cormat = bad),
               class = "efa_reliability_not_cormat")
  expect_error(.rel_adapt_bifactor(cbind(rep(0.5, 6), matrix(0.4, 6, 1)),
                                   cormat = bad),
               class = "efa_reliability_not_cormat")
  # the guard runs before the coercion, so a data frame that is not a correlation matrix is
  # still rejected rather than silently coerced
  expect_error(.rel_adapt_SL(sl_mod, factor_corres = fc, cormat = as.data.frame(bad)),
               class = "efa_reliability_not_cormat")
})

test_that(".rel_check_cormat matches the supplied matrix to the solution's variables", {
  cm <- test_models$baseline$cormat
  nms <- rownames(cm)

  # Same variables, other order: an unambiguous permutation, so it is reordered.
  expect_identical(.rel_check_cormat(cm[18:1, 18:1], nms, 18L), cm)
  # Already in the solution's order, or unlabelled: handed back untouched.
  expect_identical(.rel_check_cormat(cm, nms, 18L), cm)
  expect_identical(.rel_check_cormat(unname(cm), nms, 18L), unname(cm))

  # A different size, or a different set of variables, cannot be matched.
  expect_error(.rel_check_cormat(cm[1:6, 1:6], nms, 18L),
               class = "efa_reliability_cormat_dim")
  other <- cm
  dimnames(other) <- list(paste0("X", 1:18), paste0("X", 1:18))
  expect_error(.rel_check_cormat(other, nms, 18L),
               class = "efa_reliability_cormat_names")
  # Row and column names that disagree do not identify the variables either.
  mixed <- cm
  colnames(mixed) <- paste0("X", 1:18)
  expect_error(.rel_check_cormat(mixed, nms, 18L),
               class = "efa_reliability_cormat_names")

  # Without a solution to check against, only the correlation-matrix check applies.
  expect_identical(.rel_check_cormat(cm[18:1, 18:1]), cm[18:1, 18:1])
})

test_that("a cormat supplied as a data frame is accepted and handed on as a matrix", {
  # read.csv(file, row.names = 1) returns a data frame, and this guard is the only check a
  # user-supplied cormat passes on the reliability route -- it does not go through
  # .prepare_cor_input(). Without the coercion the omega columns still come out right (they
  # only sum and subset the matrix) and only the standardized-alpha branch fails, on diag();
  # efa_reliability() always requests that branch, so the whole call breaks.
  cm <- test_models$baseline$cormat
  cdf <- as.data.frame(cm)

  expect_true(is.matrix(.rel_check_cormat(cdf)))
  expect_identical(.rel_check_cormat(cdf), cm)

  # the adapters, and the exported entry point, give the matrix route's answer
  expect_equal(.rel_adapt_SL(sl_mod, factor_corres = fc, cormat = cdf),
               .rel_adapt_SL(sl_mod, factor_corres = fc, cormat = cm))
  expect_equal(efa_reliability(efa_mod, cormat = cdf),
               efa_reliability(efa_mod, cormat = cm))
})

test_that(".rel_adapt_efa requires an oblique solution", {
  efa_orth <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                  type = "EFAtools", method = "PAF", rotation = "varimax")
  expect_error(.rel_adapt_efa(efa_orth), class = "efa_reliability_not_oblique")
})

test_that(".rel_adapt_efa reads a one-factor solution from its unrotated loadings", {
  cm <- test_models$baseline$cormat
  efa_1 <- EFA(cm, N = 500, n_factors = 1, type = "EFAtools", method = "PAF")
  spec <- .rel_adapt_efa(efa_1)

  # One factor is nothing to rotate, so the unrotated loadings are the solution and the
  # oblique requirement does not apply. The spec is the ordinary one-factor one; the
  # front-end rewrites it into the single-factor spec the core scores.
  expect_identical(dim(spec$s_load), c(18L, 1L))
  expect_equal(unname(spec$s_load), unname(unclass(efa_1$unrot_loadings)))
  expect_equal(spec$u2, 1 - as.numeric(unclass(efa_1$unrot_loadings))^2,
               ignore_attr = TRUE)

  # Asking for a rotation leaves rot_loadings a copy of the unrotated ones and no Phi, so
  # the same solution arrives at the same spec.
  efa_1_rot <- suppressWarnings(suppressMessages(
    EFA(cm, N = 500, n_factors = 1, type = "EFAtools", method = "PAF",
        rotation = "promax")))
  expect_equal(.rel_adapt_efa(efa_1_rot), spec)

  # variance = "correlation" scores it against the object's own correlation matrix ...
  expect_equal(spec$cormat, efa_1$orig_R)
  # ... a supplied one takes precedence, and with neither the model-implied matrix stands
  # in, exactly as for an oblique solution.
  expect_equal(.rel_adapt_efa(efa_1, cormat = test_models$baseline$cormat)$cormat, cm)
  efa_no_R <- efa_1
  efa_no_R$orig_R <- NA
  expect_equal(.rel_adapt_efa(efa_no_R)$cormat,
               tcrossprod(as.numeric(unclass(efa_1$unrot_loadings))) + diag(spec$u2),
               ignore_attr = TRUE)
})

## lavaan adapters -----------------------------------------------------------

test_that(".rel_adapt_lavaan reproduces the bifactor OMEGA path", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))

  adapt <- .rel_adapt_lavaan(fit, g_name = "g")
  expect_length(adapt$groups, 1L)
  expect_false(adapt$groups[[1]]$single)
  expect_identical(adapt$variance, "sums_load")

  expect_equal(.reliability_core(adapt$groups[[1]], "sums_load"),
               .OMEGA_LAVAAN(fit, g_name = "g"))
})

test_that(".rel_adapt_lavaan Schmid-Leiman transforms a second-order model", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g =~ F1 + F2 + F3'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml"))

  spec <- .rel_adapt_lavaan(fit, g_name = "g")$groups[[1]]
  expect_equal(.reliability_core(spec, "sums_load"),
               suppressMessages(.OMEGA_LAVAAN(fit, g_name = "g")))
})

test_that(".rel_adapt_lavaan handles a single factor as a special path", {
  skip_if_not_installed("lavaan")
  mod <- 'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))

  sf <- .rel_adapt_lavaan(fit)$groups[[1]]
  expect_true(sf$single)

  ref <- suppressMessages(.OMEGA_LAVAAN(fit))  # c(Omega, H)
  omega <- sum(sf$g_load)^2 / (sum(sf$g_load)^2 + sum(sf$u2))
  expect_equal(unname(omega), unname(ref["Omega"]))
  expect_equal(unname(.h_index(sf$g_load)), unname(ref["H"]))
})

test_that(".rel_adapt_lavaan returns one spec per group", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18
          g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov =
                                        list(test_models$baseline$cormat,
                                             test_models$baseline$cormat),
                                      sample.nobs = c(500, 500),
                                      estimator = "ml", orthogonal = TRUE))

  adapt <- .rel_adapt_lavaan(fit, g_name = "g", group_names = c("Some", "Others"))
  expect_identical(adapt$group_names, c("Some", "Others"))
  expect_length(adapt$groups, 2L)

  ref <- .OMEGA_LAVAAN(fit, g_name = "g", group_names = c("Some", "Others"))
  expect_equal(.reliability_core(adapt$groups[[1]], "sums_load"), ref$Some)
  expect_equal(.reliability_core(adapt$groups[[2]], "sums_load"), ref$Others)
})

test_that(".rel_adapt_lavaan normalizes a correlated-factors fit to an oblique spec", {
  skip_if_not_installed("lavaan")
  mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
          F2 =~ V7 + V8 + V9 + V10 + V11 + V12
          F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml"))
  std <- suppressWarnings(lavaan::lavInspect(fit, "std",
                                             drop.list.single.group = FALSE))[[1]]

  adapt <- .rel_adapt_lavaan(fit)
  expect_true(adapt$correlated)
  expect_false(adapt$higher_order)
  # A simple structure would otherwise trip the "did you enter a bifactor model?" note.
  expect_false(adapt$few_loadings)

  spec <- adapt$groups[[1]]
  # The correlated-factors spec: no general factor, the oblique pattern as the group
  # loadings, and the standardized psi as their factor correlations.
  expect_true(all(spec$g_load == 0))
  expect_equal(spec$s_load, std$lambda)
  expect_equal(spec$Phi, std$psi)
  expect_equal(spec$u2, diag(std$theta))
  expect_identical(spec$fac_names, colnames(std$lambda))
  # The map is the model's own fixed-zero structure, as on the bifactor path: a
  # confirmatory model's unmodelled loadings are exact zeros, so the property holds
  # without repeating the adapter's tolerance here.
  expect_equal(spec$map, std$lambda != 0)
  # The factor correlations are substantial here, so a spec that dropped them would
  # score a materially different model.
  expect_gt(max(abs(std$psi[upper.tri(std$psi)])), .3)
})

test_that(".rel_adapt_lavaan reads the model structure, not the factor names", {
  skip_if_not_installed("lavaan")
  # An ordinary factor named "g" does not make a fit a bifactor one: no variable loads
  # on two or more factors, so it is a correlated-factors solution either way.
  mod_g <- 'g  =~ V1 + V2 + V3 + V4 + V5 + V6
            F2 =~ V7 + V8 + V9 + V10 + V11 + V12
            F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
  fit_g <- suppressWarnings(lavaan::cfa(mod_g, sample.cov = test_models$baseline$cormat,
                                        sample.nobs = 500, estimator = "ml"))
  expect_true(.rel_adapt_lavaan(fit_g, g_name = "g")$correlated)

  # A second-order model has simple structure too, so it must not be told apart by that
  # alone: it routes its factor covariances through beta, which is what marks it. Were it
  # read as a correlated-factors fit, its psi -- the residual covariances of the
  # first-order factors, not their correlations -- would be scored as factor correlations.
  mod_ho <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
             F2 =~ V7 + V8 + V9 + V10 + V11 + V12
             F3 =~ V13 + V14 + V15 + V16 + V17 + V18
             g =~ F1 + F2 + F3'
  fit_ho <- suppressWarnings(lavaan::cfa(mod_ho, sample.cov = test_models$baseline$cormat,
                                         sample.nobs = 500, estimator = "ml"))
  expect_false(.rel_adapt_lavaan(fit_ho, g_name = "g")$correlated)
  expect_true(.rel_adapt_lavaan(fit_ho, g_name = "g")$higher_order)
  # And it is still named as a misspelled general factor rather than quietly rescored.
  expect_error(.rel_adapt_lavaan(fit_ho, g_name = "G"), class = "efa_reliability_g_name")

  # A bifactor fit has variables on two factors, so it keeps needing its general factor
  # named, whatever that name is.
  mod_bi <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
             F2 =~ V7 + V8 + V9 + V10 + V11 + V12
             F3 =~ V13 + V14 + V15 + V16 + V17 + V18
             gen =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
                    V13 + V14 + V15 + V16 + V17 + V18'
  fit_bi <- suppressWarnings(lavaan::cfa(mod_bi, sample.cov = test_models$baseline$cormat,
                                         sample.nobs = 500, estimator = "ml",
                                         orthogonal = TRUE))
  expect_false(.rel_adapt_lavaan(fit_bi, g_name = "gen")$correlated)
  expect_error(.rel_adapt_lavaan(fit_bi, g_name = "g"), class = "efa_reliability_g_name")
})

rm(efa_mod, sl_mod, schmid_mod, fc)

## Tests for .reliability_core() -- the reliability engine behind OMEGA().

test_that(".reliability_core reproduces the OMEGA coefficient menu (regression)", {
  # Reference coefficients captured from OMEGA's flexible engine (both variance
  # modes), pinned to a tight tolerance. A regression in the lifted math would move
  # these numbers far beyond it; the comparison is not bit-exact because the last bits
  # differ across BLAS/platforms.
  ref_corr <- structure(c(0.882807712291674, 0.769347951077085,
    0.764969230283489, 0.74510159916687, 0.73987835131636559,
    0.49973769215022423, 0.4935927140365538, 0.51896424927096363,
    0.12494360637362421, 0.26942847755724941, 0.26981189190244209,
    0.22506712638466256, 0.8423884892261212, 0.4630872563777037,
    0.47192399221710424, 0.40799477228556474, 0.65197317077447814, NA, NA,
    NA, 0.70588235294117641, NA, NA, NA), dim = c(4L, 6L),
    dimnames = list(c("g", "1", "2", "3"),
                    c("tot", "hier", "sub", "H", "ECV", "PUC")), class = "OMEGA")
  ref_sums <- structure(c(0.88286138027838279, 0.769455598648182,
    0.765766612786389, 0.745673989947422, 0.73992333028627189,
    0.49980761571169, 0.49410722125874, 0.519362920206647,
    0.14293804999211088, 0.269466176131935, 0.270093136262557,
    0.225240024078469, 0.8423884892261212, 0.4630872563777037,
    0.47192399221710424, 0.40799477228556474, 0.65197317077447814, NA, NA,
    NA, 0.70588235294117641, NA, NA, NA), dim = c(4L, 6L),
    dimnames = list(c("g", "1", "2", "3"),
                    c("tot", "hier", "sub", "H", "ECV", "PUC")), class = "OMEGA")
  # The SL adapter labels the group rows with the factor names; same numbers.
  ref_sl <- ref_corr
  rownames(ref_sl) <- c("g", "F1", "F2", "F3")

  efa_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                 type = "EFAtools", method = "PAF", rotation = "promax")
  sl_mod <- efa_schmid_leiman(efa_mod, estimate_control = estimate_control(type = "EFAtools"), estimator = "PAF")
  fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2
  spec <- list(g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
               u2 = sl_mod$sl[, "u2"], map = fc, cormat = test_models$baseline$cormat,
               var_names = rownames(sl_mod$sl), fac_names = seq_len(3))

  # Both variance modes, driven directly through the core.
  expect_equal(.reliability_core(spec, "correlation", TRUE), ref_corr, tolerance = 1e-6)
  spec_sums <- spec
  spec_sums$cormat <- NULL
  expect_equal(.reliability_core(spec_sums, "sums_load", TRUE), ref_sums, tolerance = 1e-6)

  # add_ind = FALSE drops the H/ECV/PUC columns but leaves the omegas unchanged.
  noadd <- .reliability_core(spec, "correlation", FALSE)
  expect_equal(unclass(noadd), unclass(ref_corr)[, c("tot", "hier", "sub")],
               tolerance = 1e-6)
  expect_s3_class(noadd, "OMEGA")

  # The .OMEGA_FLEX adapter must normalise an SL object to a spec that yields the
  # same numbers through the core.
  expect_equal(.OMEGA_FLEX(sl_mod, type = "EFAtools", factor_corres = fc,
                           variance = "correlation"), ref_sl, tolerance = 1e-6)
})

test_that("a correlated-factors spec scores the group omega total from the full Phi-aware common variance", {
  # An exactly fitting correlated-factors model: two factors, two cross-loadings, and
  # a factor correlation of .5. Each subscale composite therefore also carries true
  # score variance from the other factor, both through the cross-loadings and through
  # Phi. R is the model-implied correlation matrix, so the coefficients are exact.
  L <- matrix(c(0.70, 0.60, 0.50, 0.00, 0.00, 0.25,
                0.00, 0.00, 0.30, 0.70, 0.60, 0.50), nrow = 6)
  Phi <- matrix(c(1, 0.5, 0.5, 1), 2)
  common <- L %*% Phi %*% t(L)
  u2 <- 1 - diag(common)
  R <- common + diag(u2)
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1

  spec <- list(g_load = rep(0, 6), s_load = L, u2 = u2, map = map, Phi = Phi,
               cormat = R, var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  om <- .reliability_core(spec, "correlation", add_ind = TRUE)

  # Independent reference: the common variance of the composite, 1' L Phi L' 1 read
  # off the model-implied common-variance matrix, over the composite's total variance.
  for (j in 1:2) {
    mem <- which(map[, j] == 1)
    expect_equal(unname(om[c("F1", "F2")[j], "tot"]),
                 sum(common[mem, mem]) / sum(R[mem, mem]))
    # omega subscale stays that factor's own contribution, which on this fixture is
    # less (it need not be: the total is the quadratic form cs' Phi cs and its cross
    # terms carry the sign of the other columns' sums times their correlations).
    expect_equal(unname(om[c("F1", "F2")[j], "sub"]),
                 sum(L[mem, j])^2 / sum(R[mem, mem]))
    expect_gt(om[c("F1", "F2")[j], "tot"], om[c("F1", "F2")[j], "sub"])
  }

  # Without Phi the spec is an orthogonal (bifactor / Schmid-Leiman) one. Its composites
  # still receive true score variance from the other factor through the cross-loadings,
  # just not the part carried by the factor correlation, so every omega total is read off
  # L L' instead of L Phi L' and stays above the factor's own congeneric omega. Only the
  # totals depend on Phi; the other coefficients do not.
  spec_orth <- spec
  spec_orth$Phi <- NULL
  om_orth <- .reliability_core(spec_orth, "correlation", add_ind = TRUE)
  common_orth <- L %*% t(L)
  for (j in 1:2) {
    mem <- which(map[, j] == 1)
    expect_equal(unname(om_orth[c("F1", "F2")[j], "tot"]),
                 sum(common_orth[mem, mem]) / sum(R[mem, mem]))
    expect_gt(om_orth[c("F1", "F2")[j], "tot"], om[c("F1", "F2")[j], "sub"])
  }
  expect_equal(unname(om_orth["g", "tot"]), sum(common_orth) / sum(R))
  expect_equal(om[, c("hier", "sub", "H", "ECV", "PUC")],
               om_orth[, c("hier", "sub", "H", "ECV", "PUC")])

  # Control: with exact simple structure a composite receives nothing from the other
  # factor, so its total reduces to the own-column congeneric omega whether or not the
  # spec carries Phi. The whole-scale composite spans both factors either way, so its
  # total does keep the factor correlation and is higher with Phi than without.
  L0 <- L; L0[3, 2] <- 0; L0[6, 1] <- 0
  common0 <- L0 %*% Phi %*% t(L0)
  spec0 <- spec
  spec0$s_load <- L0
  spec0$u2 <- 1 - diag(common0)
  spec0$cormat <- common0 + diag(spec0$u2)
  om0 <- .reliability_core(spec0, "correlation", add_ind = TRUE)
  spec0_orth <- spec0
  spec0_orth$Phi <- NULL
  om0_orth <- .reliability_core(spec0_orth, "correlation", add_ind = TRUE)
  expect_equal(om0[c("F1", "F2"), ], om0_orth[c("F1", "F2"), ])
  expect_equal(unname(om0[c("F1", "F2"), "tot"]), unname(om0[c("F1", "F2"), "sub"]))
  expect_gt(om0["g", "tot"], om0_orth["g", "tot"])
})

test_that("a bifactor spec scores the group omega total from every factor's contribution", {
  # An exactly fitting orthogonal bifactor solution with two cross-loadings. Each
  # subscale composite therefore carries true score variance from the general factor,
  # from its own group factor, and from the other group factor through the
  # cross-loading. An estimated Schmid-Leiman solution always has such cross-loadings;
  # only a confirmatory bifactor model has the structural zeros that make them vanish.
  g <- c(0.60, 0.55, 0.50, 0.60, 0.55, 0.50)
  S <- matrix(c(0.45, 0.40, 0.35, 0.00, 0.00, 0.20,
                0.00, 0.00, 0.25, 0.45, 0.40, 0.35), nrow = 6)
  common <- tcrossprod(g) + tcrossprod(S)
  u2 <- 1 - diag(common)
  R <- common + diag(u2)
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1
  facs <- c("F1", "F2")

  spec <- list(g_load = g, s_load = S, u2 = u2, map = map, cormat = R,
               var_names = paste0("V", 1:6), fac_names = facs)
  om <- .reliability_core(spec, "correlation", add_ind = TRUE)

  for (j in 1:2) {
    mem <- which(map[, j] == 1)
    Vgr <- sum(R[mem, mem])
    expect_equal(unname(om[facs[j], "tot"]), sum(common[mem, mem]) / Vgr)
    # The general and own-column terms alone -- what omega hierarchical and omega
    # subscale report -- leave out the other factor's contribution, so together they
    # fall short of the total by exactly that mass.
    own <- sum(g[mem])^2 + sum(S[mem, j])^2
    expect_equal(unname(om[facs[j], "hier"] + om[facs[j], "sub"]), own / Vgr)
    expect_gt(om[facs[j], "tot"], own / Vgr)
  }

  # Whole scale: every loading contributes, so the total is the whole of the
  # model-implied common variance over the observed total variance. Under exact fit
  # that is also the observed total variance less the unique variances.
  expect_equal(unname(om["g", "tot"]), sum(common) / sum(R))
  expect_equal(unname(om["g", "tot"]), 1 - sum(u2) / sum(R))

  # With exact simple structure nothing crosses, and the three coefficients add up on
  # the group rows again.
  S0 <- S; S0[3, 2] <- 0; S0[6, 1] <- 0
  common0 <- tcrossprod(g) + tcrossprod(S0)
  spec0 <- spec
  spec0$s_load <- S0
  spec0$u2 <- 1 - diag(common0)
  spec0$cormat <- common0 + diag(spec0$u2)
  om0 <- .reliability_core(spec0, "correlation", add_ind = TRUE)
  expect_equal(unname(om0[facs, "tot"]),
               unname(om0[facs, "hier"] + om0[facs, "sub"]))

  # The same contract holds under variance = "sums_load", where the composite's
  # model-implied variance replaces its observed one in the denominator.
  spec_s <- spec
  spec_s$cormat <- NULL
  om_s <- .reliability_core(spec_s, "sums_load", add_ind = TRUE)
  for (j in 1:2) {
    mem <- which(map[, j] == 1)
    V <- sum(common[mem, mem]) + sum(u2[mem])
    expect_equal(unname(om_s[facs[j], "tot"]), sum(common[mem, mem]) / V)
    expect_gt(om_s[facs[j], "tot"], om_s[facs[j], "hier"] + om_s[facs[j], "sub"])
  }
})

test_that("a loading of absolute value >= 1 returns that construct's H index as NA and notes it", {
  s <- matrix(0, 6, 2)
  s[1:3, 1] <- 0.4
  s[4:6, 2] <- 0.4
  u2 <- rep(0.5, 6)                          # proper unique variances: not a Heywood case
  map <- matrix(0, 6, 2)
  map[1:3, 1] <- 1
  map[4:6, 2] <- 1

  # A loading >= 1 on the general factor (item 6) NAs only the general-factor H. With
  # proper unique variances this is not a Heywood case (e.g. an oblique pattern
  # coefficient above 1), so it is reported as an undefined H, not a Heywood case.
  spec_g <- list(g_load = c(rep(0.5, 5), 1.05), s_load = s, u2 = u2, map = map,
                 cormat = NULL, var_names = paste0("V", 1:6),
                 fac_names = c("F1", "F2"))
  expect_warning(om_g <- .reliability_core(spec_g, "sums_load", TRUE),
                 class = "efa_reliability_h_undefined")
  expect_true(is.na(om_g["g", "H"]))
  expect_false(is.na(om_g["F1", "H"]))
  expect_false(is.na(om_g["F2", "H"]))

  # A loading >= 1 on a single group factor (item 4 on F2) NAs only that H.
  s2 <- s
  s2[4, 2] <- 1.1
  spec_s <- list(g_load = rep(0.5, 6), s_load = s2, u2 = u2, map = map, cormat = NULL,
                 var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  expect_warning(om_s <- .reliability_core(spec_s, "sums_load", TRUE),
                 class = "efa_reliability_h_undefined")
  expect_true(is.na(om_s["F2", "H"]))
  expect_false(is.na(om_s["F1", "H"]))
  expect_false(is.na(om_s["g", "H"]))
})

test_that("a Heywood case (a unique variance at or below zero) is flagged from the variances", {
  s <- matrix(0, 6, 2)
  s[1:3, 1] <- 0.4
  s[4:6, 2] <- 0.4
  map <- matrix(0, 6, 2)
  map[1:3, 1] <- 1
  map[4:6, 2] <- 1

  # Item 1 has a non-positive unique variance (a communality at or above 1): an improper
  # (Heywood) solution, detected from the uniquenesses rather than the loadings.
  u2_bad <- c(-0.05, rep(0.5, 5))
  spec_bad <- list(g_load = rep(0.5, 6), s_load = s, u2 = u2_bad, map = map,
                   cormat = NULL, var_names = paste0("V", 1:6),
                   fac_names = c("F1", "F2"))
  expect_warning(.reliability_core(spec_bad, "sums_load", TRUE),
                 class = "efa_reliability_heywood")

  # Proper loadings (|lambda| < 1) with proper variances warn about neither.
  spec_ok <- list(g_load = rep(0.5, 6), s_load = s, u2 = rep(0.5, 6), map = map,
                  cormat = NULL, var_names = paste0("V", 1:6),
                  fac_names = c("F1", "F2"))
  expect_no_warning(.reliability_core(spec_ok, "sums_load", TRUE))
})

test_that("an admissible spec yields a finite H index and no Heywood warning", {
  spec <- list(g_load = rep(0.6, 6), s_load = matrix(0.3, 6, 1), u2 = rep(0.4, 6),
               map = matrix(1, 6, 1), cormat = NULL,
               var_names = paste0("V", 1:6), fac_names = "F1")
  expect_no_warning(om <- .reliability_core(spec, "sums_load", TRUE))
  expect_true(all(is.finite(om[, "H"])))
})

test_that("the h-undefined note is per-construct, not globally suppressed by a Heywood case", {
  collect_classes <- function(spec) {
    cls <- character(0)
    withCallingHandlers(
      .reliability_core(spec, "sums_load", TRUE),
      warning = function(w) { cls <<- c(cls, class(w)[1]); invokeRestart("muffleWarning") })
    cls
  }

  # Item 1 is a Heywood case (u2 < 0). Group factor F2's item 4 has a proper oblique pattern
  # loading >= 1 (u2 > 0), so F2's H is NA for a non-Heywood reason. Both are reported: the
  # Heywood case (naming V1) AND the undefined H for F2.
  s <- matrix(0, 6, 2); s[1:3, 1] <- 0.4; s[4:6, 2] <- 0.4; s[4, 2] <- 1.2
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1
  spec <- list(g_load = rep(0.5, 6), s_load = s, u2 = c(-0.05, rep(0.5, 5)), map = map,
               cormat = NULL, var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  cls <- collect_classes(spec)
  expect_true("efa_reliability_heywood" %in% cls)
  expect_true("efa_reliability_h_undefined" %in% cls)

  # A loading >= 1 that is ITSELF the Heywood case (negative u2 on the same item) is reported
  # only as a Heywood case, not additionally as an undefined H (no redundant double warning).
  s2 <- matrix(0, 6, 2); s2[1:3, 1] <- 0.4; s2[4:6, 2] <- 0.4
  spec_hw <- list(g_load = c(rep(0.5, 5), 1.2), s_load = s2, u2 = c(rep(0.5, 5), -0.1),
                  map = map, cormat = NULL, var_names = paste0("V", 1:6),
                  fac_names = c("F1", "F2"))
  cls_hw <- collect_classes(spec_hw)
  expect_true("efa_reliability_heywood" %in% cls_hw)
  expect_false("efa_reliability_h_undefined" %in% cls_hw)
})

test_that("an NA uniqueness is not misread as a Heywood case", {
  # An NA in u2 (a malformed component) must not fire a false Heywood warning naming the
  # variable "NA", nor suppress the legitimate h-undefined note. which(u2 < eps) drops NA.
  s <- matrix(0, 6, 2); s[1:3, 1] <- 0.4; s[4:6, 2] <- 0.4
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1
  spec <- list(g_load = rep(0.5, 6), s_load = s, u2 = c(0.5, NA, rep(0.5, 4)), map = map,
               cormat = NULL, var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  expect_no_warning(.reliability_core(spec, "sums_load", TRUE))

  # A genuine |loading| >= 1 still triggers the h-undefined note even when an unrelated
  # uniqueness is NA (the NA does not count as a Heywood variable that would suppress it).
  spec_h <- spec
  spec_h$g_load <- c(rep(0.5, 5), 1.05)
  expect_warning(.reliability_core(spec_h, "sums_load", TRUE),
                 class = "efa_reliability_h_undefined")
})

test_that("a missing (NA) loading yields an NA H index rather than an error", {
  # A missing loading is not a Heywood case: it is excluded from the >= 1 test and
  # flows through the sum to NA, so H must be NA (not an aborted call). This guards
  # against `if (any(abs(NA) >= 1))` erroring with "missing value where TRUE/FALSE".
  s <- matrix(0, 6, 2)
  s[1:3, 1] <- 0.4
  s[4:6, 2] <- 0.4
  map <- matrix(0, 6, 2)
  map[1:3, 1] <- 1
  map[4:6, 2] <- 1
  spec <- list(g_load = c(0.5, NA, rep(0.5, 4)), s_load = s, u2 = rep(0.5, 6),
               map = map, var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  expect_no_warning(om <- .reliability_core(spec, "sums_load", TRUE))
  expect_true(is.na(om["g", "H"]))
  expect_true(is.finite(om["F1", "H"]))
  expect_true(is.finite(om["F2", "H"]))
})

## add_rel: Cronbach's alpha, composite reliability (CR), and AVE -----------------

test_that("add_rel = FALSE leaves the OMEGA coefficient columns unchanged", {
  # The OMEGA path never sets add_rel, so its coefficient menu must be untouched;
  # add_rel = TRUE appends exactly alpha, CR, and AVE.
  spec <- list(g_load = rep(0.4, 6), s_load = matrix(0.5, 6, 1), u2 = rep(0.59, 6),
               map = matrix(1, 6, 1), cormat = NULL, var_names = paste0("V", 1:6),
               fac_names = "F1")
  expect_identical(colnames(.reliability_core(spec, "sums_load", TRUE)),
                   c("tot", "hier", "sub", "H", "ECV", "PUC"))
  expect_identical(colnames(.reliability_core(spec, "sums_load", FALSE)),
                   c("tot", "hier", "sub"))
  expect_identical(colnames(.reliability_core(spec, "sums_load", TRUE, add_rel = TRUE)),
                   c("tot", "hier", "sub", "H", "ECV", "PUC", "alpha", "CR", "AVE"))
})

test_that("standardized alpha matches psych::alpha for the whole scale and subscales", {
  skip_if_not_installed("psych")
  R <- test_models$baseline$cormat            # 18 items, three 6-item blocks
  s <- matrix(0, 18, 3)
  s[1:6, 1] <- 0.5; s[7:12, 2] <- 0.5; s[13:18, 3] <- 0.5
  map <- matrix(0, 18, 3)
  map[1:6, 1] <- 1; map[7:12, 2] <- 1; map[13:18, 3] <- 1
  # variance = "sums_load" so the omega denominators ignore the (arbitrary) loadings;
  # only the appended alpha uses the supplied correlation matrix R.
  spec <- list(g_load = rep(0.4, 18), s_load = s, u2 = rep(0.2, 18), map = map,
               cormat = R, var_names = rownames(R), fac_names = c("F1", "F2", "F3"))
  om <- .reliability_core(spec, "sums_load", add_ind = FALSE, add_rel = TRUE)

  # Our alpha is standardized (computed from the correlation matrix), matching psych's
  # std.alpha; raw (covariance-based) alpha would need the item standard deviations.
  a_whole <- suppressWarnings(
    psych::alpha(R, check.keys = FALSE, warnings = FALSE))$total[["std.alpha"]]
  expect_equal(unname(om["g", "alpha"]), a_whole, tolerance = 1e-8)
  for (j in 1:3) {
    mem <- which(map[, j] == 1)
    a_sub <- suppressWarnings(
      psych::alpha(R[mem, mem], check.keys = FALSE, warnings = FALSE))$total[["std.alpha"]]
    expect_equal(unname(om[c("F1", "F2", "F3")[j], "alpha"]), a_sub, tolerance = 1e-8)
  }
})

test_that("alpha falls back to the model-implied correlation matrix when cormat is NULL", {
  g <- rep(0.4, 6)
  s <- matrix(0, 6, 2); s[1:3, 1] <- 0.5; s[4:6, 2] <- 0.5
  # Uniquenesses that do NOT complete to unit item variance (item 1), so the
  # model-implied Lambda Lambda' + diag(u2) has a non-unit, non-constant diagonal:
  # standardized alpha must standardize it (cov2cor) rather than read it as a
  # correlation matrix as-is.
  u2 <- c(0.30, rep(1 - 0.4^2 - 0.5^2, 5))    # item 1 diagonal = 0.16 + 0.25 + 0.30 != 1
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1
  spec <- list(g_load = g, s_load = s, u2 = u2, map = map, cormat = NULL,
               var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  om <- .reliability_core(spec, "sums_load", add_ind = FALSE, add_rel = TRUE)

  # Reference: standardize the model-implied matrix to a correlation matrix first.
  R <- stats::cov2cor(cbind(g, s) %*% t(cbind(g, s)) + diag(u2))
  expect_equal(unname(om["g", "alpha"]), 6 / 5 * (1 - sum(diag(R)) / sum(R)))
  Rk <- R[1:3, 1:3]
  expect_equal(unname(om["F1", "alpha"]), 3 / 2 * (1 - sum(diag(Rk)) / sum(Rk)))
})

test_that("CR and AVE match hand-computed values; the general-factor row is NA", {
  g <- rep(0.4, 6)
  s <- matrix(0, 6, 2); s[1:3, 1] <- 0.5; s[4:6, 2] <- 0.5
  u2 <- rep(1 - 0.4^2 - 0.5^2, 6)             # 0.59
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1
  spec <- list(g_load = g, s_load = s, u2 = u2, map = map, cormat = NULL,
               var_names = paste0("V", 1:6), fac_names = c("F1", "F2"))
  om <- .reliability_core(spec, "sums_load", add_ind = FALSE, add_rel = TRUE)

  eF1 <- 3 * 0.59
  expect_equal(unname(om["F1", "CR"]),  1.5^2 / (1.5^2 + eF1))
  expect_equal(unname(om["F1", "AVE"]), (3 * 0.25) / (3 * 0.25 + eF1))
  expect_equal(unname(om["F2", "CR"]),  unname(om["F1", "CR"]))   # symmetry
  expect_equal(unname(om["F2", "AVE"]), unname(om["F1", "AVE"]))

  # CR and AVE are per-subscale only; the general-factor row is NA. Alpha is not.
  expect_true(is.na(om["g", "CR"]))
  expect_true(is.na(om["g", "AVE"]))
  expect_false(is.na(om["g", "alpha"]))
})

test_that("CR uses the assigned composite: a foreign item's cross-loading does not change it", {
  s0 <- matrix(0, 6, 2); s0[1:3, 1] <- 0.5; s0[4:6, 2] <- 0.5
  map <- matrix(0, 6, 2); map[1:3, 1] <- 1; map[4:6, 2] <- 1
  base <- list(g_load = rep(0.3, 6), s_load = s0, u2 = 1 - 0.3^2 - rowSums(s0^2),
               map = map, cormat = NULL, var_names = paste0("V", 1:6),
               fac_names = c("F1", "F2"))
  om0 <- .reliability_core(base, "sums_load", add_ind = FALSE, add_rel = TRUE)

  # Item 4 (assigned to F2) gains a 0.3 cross-loading on F1. F1's assigned composite
  # (items 1-3) is unchanged, so CR_F1 and AVE_F1 must not move.
  sX <- s0; sX[4, 1] <- 0.3
  cross <- base; cross$s_load <- sX; cross$u2 <- 1 - 0.3^2 - rowSums(sX^2)
  omX <- .reliability_core(cross, "sums_load", add_ind = FALSE, add_rel = TRUE)

  expect_equal(unname(omX["F1", "CR"]),  unname(om0["F1", "CR"]))
  expect_equal(unname(omX["F1", "AVE"]), unname(om0["F1", "AVE"]))

  # It equals the assigned-only formula, not a full-column one that would fold in the 0.3.
  eF1 <- 3 * (1 - 0.3^2 - 0.25)
  expect_equal(unname(omX["F1", "CR"]), 1.5^2 / (1.5^2 + eF1))
  expect_false(isTRUE(all.equal(unname(omX["F1", "CR"]), 1.8^2 / (1.8^2 + eF1))))
})

test_that("CR equals the lavaan single-factor omega for the k = 1 case", {
  skip_if_not_installed("lavaan")
  mod <- 'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
               V13 + V14 + V15 + V16 + V17 + V18'
  fit <- suppressWarnings(lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                                      sample.nobs = 500, estimator = "ml",
                                      orthogonal = TRUE))
  ref <- suppressMessages(.OMEGA_LAVAAN(fit))          # c(Omega, H)

  # Represent the single factor as one group factor (no general factor) and read CR.
  sf <- .rel_adapt_lavaan(fit)$groups[[1]]
  p <- length(sf$g_load)
  spec <- list(g_load = rep(0, p), s_load = matrix(sf$g_load, ncol = 1), u2 = sf$u2,
               map = matrix(1, p, 1), cormat = NULL, var_names = sf$var_names,
               fac_names = "F1")
  om <- .reliability_core(spec, "sums_load", add_ind = FALSE, add_rel = TRUE)
  expect_equal(unname(om["F1", "CR"]), unname(ref[["Omega"]]), tolerance = 1e-8)
})

test_that("the coefficients a single factor does not define are degenerate, not merely absent", {
  # The grounds for reporting only omega total, alpha, and the H index for one factor. The
  # core computes the other four for such a spec, and each comes out saying something about
  # the number of factors rather than about the solution: omega hierarchical is omega total
  # again, the one factor accounting for all of the common variance; omega subscale is 0,
  # there being no group factors; and the ECV and the PUC are exactly 1, there being nothing
  # for the general factor to share the common variance with and no pair of variables
  # sharing a group factor.
  fit <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 1, estimator = "PAF")
  spec <- .rel_single_factor_spec(.rel_adapt_efa(fit))
  expect_identical(ncol(spec$s_load), 0L)

  for (v in c("correlation", "sums_load")) {
    om <- .reliability_core(spec, v, add_ind = TRUE, add_rel = TRUE)
    # The two omegas are one quantity, reached by two expressions that agree to rounding.
    expect_equal(unname(om["g", "hier"]), unname(om["g", "tot"]), tolerance = 1e-12)
    expect_identical(unname(om["g", "sub"]), 0)
    expect_identical(unname(om["g", "ECV"]), 1)
    expect_identical(unname(om["g", "PUC"]), 1)
    # The three that are reported carry information the solution actually supplies.
    expect_true(all(is.finite(om["g", c("tot", "alpha", "H")])))
  }
})

test_that("empty and single-item subscales get NA alpha/CR/AVE as appropriate", {
  # Empty F2: all items assigned to F1, so F2 has an all-zero map column.
  s <- matrix(0, 6, 2); s[, 1] <- 0.5
  map <- matrix(0, 6, 2); map[, 1] <- 1
  spec <- list(g_load = rep(0.3, 6), s_load = s, u2 = 1 - 0.3^2 - rowSums(s^2),
               map = map, cormat = NULL, var_names = paste0("V", 1:6),
               fac_names = c("F1", "F2"))
  expect_warning(
    om <- .reliability_core(spec, "sums_load", add_ind = FALSE, add_rel = TRUE),
    class = "efa_omega_empty_factor")
  expect_true(all(is.na(om["F2", c("alpha", "CR", "AVE")])))
  expect_false(any(is.na(om["F1", c("alpha", "CR", "AVE")])))

  # Single-item subscale F1: alpha is undefined (NA), CR and AVE are finite.
  s2 <- matrix(0, 4, 2); s2[1, 1] <- 0.6; s2[2:4, 2] <- 0.5
  map2 <- matrix(0, 4, 2); map2[1, 1] <- 1; map2[2:4, 2] <- 1
  spec2 <- list(g_load = rep(0.3, 4), s_load = s2, u2 = 1 - 0.3^2 - rowSums(s2^2),
                map = map2, cormat = NULL, var_names = paste0("V", 1:4),
                fac_names = c("F1", "F2"))
  om2 <- .reliability_core(spec2, "sums_load", add_ind = FALSE, add_rel = TRUE)
  expect_true(is.na(om2["F1", "alpha"]))
  expect_true(is.finite(om2["F1", "CR"]))
  expect_true(is.finite(om2["F1", "AVE"]))
})

#' Print and format a reliability object
#'
#' `print()` shows the reliability coefficients for the general factor and the
#' group factors, for a single group or for each group: the reliability
#' coefficients (omega total, omega hierarchical, and omega subscale, standardized
#' Cronbach's alpha, and the H index) and the common-variance indices (the explained
#' common variance, ECV, and the percent of uncontaminated correlations, PUC).
#' `format()` assembles the same report and returns it as a character vector;
#' `print()` is `cat(format(x), sep = "\n")`. The lines follow the active console
#' theme, so they are plain when colours are disabled (for example when captured
#' into a file or stripped with [cli::ansi_strip()]).
#'
#' @param x An object of class `efa_reliability`.
#' @param digits Integer. The number of decimal places the coefficients are
#'   rounded to. Default is 3.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @family reliability coefficients
#'
#' @export
#' @method print efa_reliability
#'
#' @examples
#' efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' rel <- efa_reliability(efa_mod)
#' rel
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(rel))
#'
print.efa_reliability <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_reliability
#' @export
#' @method format efa_reliability
format.efa_reliability <- function(x, digits = 3, ...) {

  reg <- .reliability_registry()
  settings <- attr(x, "settings")
  groups <- unique(x$group)
  # A single unnamed group (group = NA) prints without a group header; any named or
  # multigroup result labels each group.
  show_group <- !(length(groups) == 1L && is.na(groups))

  # Reconstruct the factor-by-coefficient matrix for one kind (reliability /
  # common-variance) from the group's long rows, so the shared matrix renderer can lay it
  # out as OMEGA does. Only factors and coefficients that carry a value for this kind
  # appear (e.g. ECV/PUC contribute a general-factor row only), so no all-NA rows are
  # rendered.
  build_mat <- function(sub, kind_sel) {
    keep <- reg[reg$kind == kind_sel & reg$coefficient %in% sub$coefficient, ,
                drop = FALSE]
    if (nrow(keep) == 0L) return(NULL)
    sk <- sub[sub$coefficient %in% keep$coefficient, , drop = FALSE]
    # Rows are keyed on the level and the factor together, and filled by index rather than
    # by label. The label alone does not identify a row: the whole-scale row carries one of
    # its own ("g" for a hierarchy, "total" for a correlated-factors solution) and a group
    # factor may be named the same, in which case indexing by label would write both into
    # one row and print the group factor's value as the whole scale's. The rendered labels
    # are still the factor names, so a collision shows as two rows of one name rather than
    # as a silently dropped value.
    key <- paste(sk$level, sk$factor, sep = "\r")
    rows <- unique(key)
    facs <- sk$factor[match(rows, key)]
    mat <- matrix(NA_real_, length(rows), nrow(keep),
                  dimnames = list(facs, keep$label))
    for (i in seq_len(nrow(keep))) {
      hit <- sk$coefficient == keep$coefficient[i]
      mat[match(key[hit], rows), keep$label[i]] <- sk$value[hit]
    }
    mat
  }

  cli::cli_format_method({

    if (!is.null(settings) && !is.null(settings$variance)) {
      vtext <- switch(settings$variance,
                      correlation = "the correlation matrix",
                      sums_load = "the model-implied composite variance",
                      settings$variance)
      cli::cli_text("")
      cli::cli_text("Total variance from {vtext}.")
    }

    # A correlated-factors solution has no general factor, so relate the two omega
    # columns rather than print them side by side unexplained. Which relation holds is a
    # property of the solution: a factor's composite also carries true score variance
    # from the other factors, through its cross-loadings and through any factor
    # correlations, so the two columns differ -- in either direction, since the total is
    # a quadratic form whose cross terms take the sign of the other columns' sums times
    # their correlations. Only under exact simple structure and no factor correlations
    # does the composite draw on nothing else and the two coincide. Decide from the
    # values as printed, so the sentence always describes the table below it. Only shown
    # when both columns are there (a coefficients = subset may drop one).
    if (isTRUE(settings$no_general) &&
        all(c("omega_total", "omega_subscale") %in% x$coefficient)) {
      gr <- x[x$level == "group", , drop = FALSE]
      tot <- gr[gr$coefficient == "omega_total", , drop = FALSE]
      sub_om <- gr[gr$coefficient == "omega_subscale", , drop = FALSE]
      # Paired on the group and the factor rather than on position: an undefined
      # coefficient is dropped row by row, so two equally long columns need not describe
      # the same factors, and a multigroup result repeats every factor label once per
      # group -- pairing on the label alone would match only each factor's first group and
      # leave the counts below unequal, so no multigroup table could ever be found to
      # coincide. Only the factors carrying both can be compared, and a factor carrying one
      # of them alone cannot establish equality, so it takes the general wording. Compared
      # through the printer's own formatter rather than against a tolerance, so "equals"
      # says the two columns show the same number at this `digits` -- a tolerance of one
      # last decimal calls values equal that straddle a rounding boundary and print apart.
      key <- function(d) paste(d$group, d$factor, sep = "\r")
      both <- intersect(key(tot), key(sub_om))
      shown <- function(d) .efa_num(d$value[match(both, key(d))], digits = digits,
                                    pad = FALSE)
      coincide <- length(both) > 0 &&
        length(both) == nrow(tot) && length(both) == nrow(sub_om) &&
        identical(shown(tot), shown(sub_om))
      cli::cli_text("")
      if (isTRUE(coincide)) {
        cli::cli_text("Correlated-factors solution: with no general factor, each factor's subscale omega equals its total omega.")
      } else {
        cli::cli_text("Correlated-factors solution: a factor's total omega counts the true score variance its composite receives from every factor, through its cross-loadings and any factor correlations; its subscale omega counts only that factor's own contribution.")
      }
    }

    for (grp in groups) {
      sub <- x[x$group %in% grp, , drop = FALSE]

      if (isTRUE(show_group)) {
        cli::cli_text("")
        cli::cli_text("Group {.emph {grp}}:")
      }

      # The tidy result carries no NA values (the builder drops them), so any NA hole in
      # the reconstructed factor-by-coefficient grid is a coefficient that is undefined for
      # that factor (e.g. omega subscale / H on the whole-scale "total" row of a
      # correlated-factors solution). Print those blank rather than as "NA", as OMEGA's
      # printer does.
      rel <- build_mat(sub, "reliability")
      if (!is.null(rel)) {
        .print_efa_rule("Reliability coefficients")
        .efa_emit_lines(.efa_corr_lines(rel, digits = digits, na = ""))
      }

      cv <- build_mat(sub, "common_variance")
      if (!is.null(cv)) {
        .print_efa_rule("Common-variance indices")
        .efa_emit_lines(.efa_corr_lines(cv, digits = digits, na = ""))
      }
    }

  })
}

#' Print and format an efa_screen object
#'
#' `print()` turns the factor-analysis screening diagnostics computed by
#' [efa_screen()] into a sectioned report with banded, colour-coded verdicts:
#' sampling adequacy and sphericity (the Kaiser-Meyer-Olkin measure and Bartlett's
#' test of sphericity), multicollinearity (the determinant and condition number of
#' the correlation matrix), the per-variable diagnostics, and, when raw data were
#' supplied, multivariate normality and multivariate outliers. It closes with a
#' consolidated list of actionable recommendations (for example, which items to
#' consider dropping, whether to prefer an ordinal or a robust estimator, and a
#' caveat that keeps an over-powered Bartlett's test from being over-trusted).
#' `format()` assembles the same report and returns it as a character vector;
#' `print()` is `cat(format(x), sep = "\n")`. The lines follow the active console
#' theme, so they are plain when colours are disabled (for example when captured
#' into a file or stripped with [cli::ansi_strip()]). `print()` does not draw a
#' plot.
#'
#' @param x An object of class `efa_screen` (output from [efa_screen()]).
#' @param digits Integer. The number of decimal places the reported values are
#'   rounded to. Default is 3.
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @family factor analysis suitability
#'
#' @export
#'
#' @method print efa_screen
#'
#' @examples
#' # From raw data
#' efa_screen(iris[, 1:4])
#'
#' # From a correlation matrix (supply N for Bartlett's test of sphericity)
#' efa_screen(test_models$baseline$cormat, N = 500)
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(efa_screen(test_models$baseline$cormat, N = 500)))
#'
print.efa_screen <- function(x, digits = 3, ...) {
  cat(format(x, digits = digits, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_screen
#' @export
#' @method format efa_screen
format.efa_screen <- function(x, digits = 3, ...) {

  raw <- !is.null(x$per_item)

  # Overall KMO band (reused Kaiser & Rice, 1974 cut-offs) and the significance flags
  # that several sections and the recommendations share.
  kmo <- x$kmo$KMO
  bart_sig <- .screen_is_sig(x$bartlett$p_value)
  # The multivariate-normality verdict is a count over the tests that are available. The
  # Henze-Zirkler p-value is withheld when its null approximation degenerates at many
  # variables, so the set is not always all three. `mvn_nonnormal` gates the
  # recommendations as before; the counts let the verdict sentence separate one marginal
  # rejection from a unanimous one.
  mvn_p <- if (raw && !inherits(x$normality, "efa_screen_no_mvn")) {
    c(x$normality$mardia$skewness_p, x$normality$mardia$kurtosis_p,
      x$normality$hz$p_value)
  }
  mvn_n_avail <- sum(!is.na(mvn_p))
  mvn_n_rej <- sum(vapply(mvn_p, .screen_is_sig, logical(1)))
  mvn_nonnormal <- mvn_n_rej > 0L

  cli::cli_format_method({

    # -- Sampling adequacy and sphericity --------------------------------------
    .print_efa_rule("Sampling adequacy and sphericity")

    if (!is.null(kmo) && !is.na(kmo)) {
      band <- .kmo_band(kmo)
      kval <- round(kmo, digits)
      band$alert("The overall KMO value for your data is {band$label} (Overall KMO = {kval}).",
                 wrap = TRUE)
      # The bands are named on screen for the same reason the condition-index rule is:
      # a verbal band is a convention, not a finding, and an unattributed cut-off
      # standing next to an attributed one reads as the latter.
      cli::cli_text(paste("These data are {band$suitability} suitable for factor",
                          "analysis (verbal bands: Kaiser & Rice, 1974)."))
    } else {
      cli::cli_alert_warning("The overall KMO value for your data is not available.",
                             wrap = TRUE)
    }

    cli::cli_text("")

    bart <- x$bartlett
    if (is.null(bart)) {
      cli::cli_alert_warning(
        "Bartlett's test of sphericity was not computed; no sample size (N) was supplied.",
        wrap = TRUE)
    } else if (!is.null(bart$p_value) && !is.na(bart$p_value)) {
      # Bartlett significance wording reused verbatim from format.efa_bartlett.
      if (bart$p_value < .05) {
        cli::cli_alert_success(
          "The Bartlett's test of sphericity was significant at an alpha level of .05.",
          wrap = TRUE)
        cli::cli_text("These data are probably suitable for factor analysis.")
      } else {
        cli::cli_alert_danger(
          "The Bartlett's test of sphericity was not significant at an alpha level of .05.",
          wrap = TRUE)
        cli::cli_text("These data are probably not suitable for factor analysis.")
      }
      cli::cli_verbatim(paste0("\U03C7\U00B2(", bart$df, ") = ", round(bart$chisq, 2),
                               ", ", cli::style_italic("p"), .screen_p_str(bart$p_value)))
    } else {
      .print_bartlett_no_result()
    }

    # -- Multicollinearity -----------------------------------------------------
    .print_efa_rule("Multicollinearity")

    # The determinant is the product of the p eigenvalues of R, so it shrinks
    # geometrically as variables are added even when every eigenvalue stays far from
    # zero. A fixed cut-off on it (below 0.00001; Field, 2018) is therefore a statement
    # about the number of variables as much as about the data: it fires on wide but
    # well-conditioned item pools and contradicts the condition index printed directly
    # under it. The determinant is reported as a number only, and the condition index
    # (above 30; Belsley, 1991) carries this section's verdict and the consolidated
    # recommendation, so the one cut-off lives in one place.
    det_R <- x$determinant
    dstr <- format(det_R, digits = digits, scientific = FALSE)
    cli::cli_alert_info(
      "Determinant: {dstr}. It falls as variables are added, so the condition index below carries the verdict.",
      wrap = TRUE)

    cond <- x$condition
    cstr <- formatC(cond, format = "f", digits = digits)
    # The verdict is on the condition *index*, and its bands live in .ci_band(). The
    # index is shown alongside the rule that governs it: quoting the rule without the
    # quantity invites the value to be read against the wrong scale. Above 30 the
    # strength is named as well, because a flat label would give an index of 35 and an
    # index of 3000 the same verdict.
    ci <- sqrt(cond)
    cistr <- formatC(ci, format = "f", digits = digits)
    ci_band <- .ci_band(ci)
    if (ci_band$flag) {
      ci_band$alert(
        "Condition number: {cstr} (condition index {cistr}). The index is above 30, which indicates a near linear dependency; its relative strength is {ci_band$strength} (Belsley, 1991).",
        wrap = TRUE)
    } else {
      ci_band$alert("Condition number: {cstr} (condition index {cistr}). {ci_band$note}",
                    wrap = TRUE)
    }

    # -- Per-variable diagnostics ----------------------------------------------
    .print_efa_rule("Per-variable diagnostics")
    # The standard display names are MSA (the per-variable measure of sampling adequacy,
    # stored as kmo_i) and SMC (the squared multiple correlation). The per-variable
    # values are laid out as base R's data-frame print, emitted verbatim so cli does not
    # reflow the aligned columns (as format.efa_kmo does for the KMO vector). This is the
    # one table not routed through .efa_num(), so it is also the one place a leading zero
    # is kept: the frame mixes bounded coefficients with variances and missing-data
    # percentages on the data's own scale, for which the leading zero is not redundant.
    disp <- .screen_per_item_display(x, digits)
    cli::cli_verbatim(utils::capture.output(print(disp)))

    # -- Multivariate normality (raw data only) --------------------------------
    if (raw) {
      .print_efa_rule("Multivariate normality")
      nm <- x$normality
      if (inherits(nm, "efa_screen_no_mvn")) {
        cli::cli_alert_warning(
          "Multivariate normality tests were skipped (singular complete-case covariance).",
          wrap = TRUE)
      } else {
        md <- nm$mardia
        .screen_mvn_line(.screen_is_sig(md$skewness_p),
                         paste0("Mardia's skewness: \U03C7\U00B2(", md$skewness_df, ") = ",
                                round(md$skewness, 2), ", p", .screen_p_str(md$skewness_p), "."))
        # An undefined kurtosis statistic is neither a pass nor a rejection, so it gets
        # an info alert, as a withheld Henze-Zirkler p-value does below.
        if (is.na(md$kurtosis)) {
          cli::cli_alert_info(
            "Mardia's kurtosis: not available at this number of complete cases.")
        } else {
          .screen_mvn_line(.screen_is_sig(md$kurtosis_p),
                           paste0("Mardia's kurtosis: z = ", round(md$kurtosis, 2),
                                  ", p", .screen_p_str(md$kurtosis_p), "."))
        }
        # A withheld Henze-Zirkler p-value is neither a pass nor a rejection, so it gets
        # an info alert rather than the success/danger pair the other two lines use.
        if (inherits(nm$hz, "efa_screen_no_hz")) {
          cli::cli_alert_info(
            paste0("Henze-Zirkler: HZ = ", round(nm$hz$statistic, 2),
                   ", p not available at this number of variables."))
        } else {
          .screen_mvn_line(.screen_is_sig(nm$hz$p_value),
                           paste0("Henze-Zirkler: HZ = ", round(nm$hz$statistic, 2),
                                  ", p", .screen_p_str(nm$hz$p_value), "."))
        }
        cli::cli_text(if (isTRUE(mvn_nonnormal)) {
          paste0("These data depart from multivariate normality: ", mvn_n_rej, " of the ",
                 mvn_n_avail, " tests ", .screen_plural(mvn_n_rej, "rejects", "reject"),
                 " it.")
        } else {
          paste0("These data are consistent with multivariate normality: none of the ",
                 mvn_n_avail, " tests rejects it.")
        })
        # The tests use the complete cases, which under non-ignorable missingness are not
        # a random subsample: with missing data the verdict above is a statement about a
        # subset, so say which one. Silent when every row is complete.
        if (.screen_incomplete(nm$n_complete, x$settings$n_obs)) {
          cli::cli_text("(Computed from {nm$n_complete} complete cases of the
                         {x$settings$n_obs} rows supplied.)")
        }
      }
    }

    # -- Outliers (raw data only) ----------------------------------------------
    if (raw) {
      .print_efa_rule("Outliers")
      o <- x$outliers
      if (inherits(o, "efa_screen_no_outliers")) {
        cli::cli_alert_warning("Outlier diagnostics were skipped.")
        # The recorded reason, as in the classical-fallback branch below: which of the
        # three ways the complete-case covariance failed decides what the user should do,
        # and "singular" alone points at collinearity even when the cause is missingness.
        if (!is.null(o$reason)) cli::cli_text(o$reason)
      } else {
        if (identical(o$method, "classical")) {
          cli::cli_alert_warning(paste(
            "A robust (MCD) covariance could not be computed; classical Mahalanobis",
            "distances were used."), wrap = TRUE)
          # The recorded reason (too few complete cases, near-collinear variables, or an
          # exact fit) and the consequence of the fallback: the classical covariance is
          # computed from every observation, outliers included, so the diagnostic no longer
          # has the high-breakdown property the MCD gives it.
          if (!is.null(o$fallback_reason)) cli::cli_text(o$fallback_reason)
          cli::cli_text(paste(
            "These distances come from a covariance the outliers themselves inflate, so",
            "the diagnostic is no longer high-breakdown and tends to under-flag."))
        }
        nf <- length(o$flagged)
        cstr_o <- format(o$cutoff, digits = digits, scientific = FALSE)
        # Name the distance for the estimate actually used (robust MCD, or classical
        # Mahalanobis when the robust covariance could not be formed).
        dist_label <- if (identical(o$method, "classical")) "Mahalanobis distance" else "robust distance"
        cli::cli_alert_info(paste0(
          nf, " of ", .screen_case_base(o$n_complete, x$settings$n_obs), " ",
          .screen_plural(nf, "was", "were"), " flagged as ",
          .screen_plural(nf, "a multivariate outlier", "multivariate outliers"),
          " (", dist_label, " > ", cstr_o, ")."), wrap = TRUE)
      }
    }

    # -- Recommendations -------------------------------------------------------
    .print_efa_rule("Recommendations")
    recs <- .screen_recommendations(x, raw, kmo, bart_sig, mvn_nonnormal,
                                    ci_band$flag, digits)
    # cli treats each bullet as a glue/cli template, so double any braces coming from
    # user variable names to render them literally (as .efa_emit_bullets does).
    msg <- gsub("}", "}}", gsub("{", "{{", recs$message, fixed = TRUE), fixed = TRUE)
    cli::cli_bullets(stats::setNames(msg, recs$symbol))

  })
}

# Kaiser & Rice (1974) verbal bands for the overall KMO value, and the single place they
# are written down: the highest band the value clears gives its label, the suitability it
# implies, and the severity (success >= .7, warning >= .6, danger below) as an alert
# function, the matching colour, and the equivalent cli_bullets() symbol, so that callers
# reporting the same value as an alert, a coloured label, or a bullet cannot drift apart.
.kmo_band <- function(kmo) {
  bands <- list(
    list(min = .9,   label = "marvellous",   colour = cli::col_green,  alert = cli::cli_alert_success, symbol = "v", suitability = "probably"),
    list(min = .8,   label = "meritorious",  colour = cli::col_green,  alert = cli::cli_alert_success, symbol = "v", suitability = "probably"),
    list(min = .7,   label = "middling",     colour = cli::col_green,  alert = cli::cli_alert_success, symbol = "v", suitability = "probably"),
    list(min = .6,   label = "mediocre",     colour = cli::col_yellow, alert = cli::cli_alert_warning, symbol = "!", suitability = "probably"),
    list(min = .5,   label = "miserable",    colour = cli::col_red,    alert = cli::cli_alert_danger,  symbol = "x", suitability = "hardly"),
    list(min = -Inf, label = "unacceptable", colour = cli::col_red,    alert = cli::cli_alert_danger,  symbol = "x", suitability = "not")
  )
  Find(function(b) kmo >= b$min, bands)
}

# Belsley's (1991) reading of the condition index, and the single place it is written
# down: the highest band the index clears gives the severity as an alert function, the
# verdict text, and `flag`, which gates both this section's verdict and the consolidated
# recommendation so the two cannot drift apart. The bands are steps on the progression
# 1, 3, 10, 30, 100, 300, 1000 that Belsley reads relative strength from (p. 38); they
# are not absolute cut-offs. An index of 10 or less is rarely of interest (pp. 43 and
# 46). Belsley gives 30 as one example value for flagging a near dependency, and a band
# must exceed it, not only reach it (p. 38). He calls 30 to 100 moderate (p. 42), and in
# his worked example he grades indexes of 35, 153, and 455 as moderately strong, strong,
# and very strong (pp. 44-45).
.ci_band <- function(ci) {
  bands <- list(
    list(above = 300,  alert = cli::cli_alert_danger,  flag = TRUE, strength = "very strong"),
    list(above = 100,  alert = cli::cli_alert_danger,  flag = TRUE, strength = "strong"),
    list(above = 30,   alert = cli::cli_alert_danger,  flag = TRUE, strength = "moderate"),
    list(above = 10,   alert = cli::cli_alert_warning, flag = FALSE,
         note = "The index is above 10, but not above 30, the value that flags a near linear dependency (Belsley, 1991)."),
    list(above = -Inf, alert = cli::cli_alert_success, flag = FALSE,
         note = "An index of 10 or less is rarely of interest (Belsley, 1991).")
  )
  # An index that is not available takes the lowest band, as the report has always done.
  if (is.na(ci)) return(bands[[length(bands)]])
  Find(function(b) ci > b$above, bands)
}

# TRUE when the complete cases the normality and outlier diagnostics use are fewer than
# the rows supplied, i.e. when the two possible denominators differ.
.screen_incomplete <- function(n_complete, n_obs) {
  !is.null(n_obs) && !is.na(n_obs) && !is.null(n_complete) && n_obs > n_complete
}

# The denominator of the complete-case diagnostics, named as such -- together with the
# number of rows supplied -- whenever missing values make the two differ, so a count or a
# rate cannot be read against the wrong base.
.screen_case_base <- function(n_complete, n_obs) {
  if (.screen_incomplete(n_complete, n_obs)) {
    paste0(n_complete, " complete cases of the ", n_obs, " rows supplied")
  } else {
    paste0(n_complete, " observations")
  }
}

# TRUE when a p-value is present and below .05 (guards NA / NULL p-values).
.screen_is_sig <- function(p) !is.null(p) && !is.na(p) && p < .05

# p-value tail formatting shared by the Bartlett and MVN lines and by format.efa_bartlett.
.screen_p_str <- function(p) {
  if (is.null(p) || is.na(p)) " = NA" else if (p < .001) " < .001" else paste0(" = ", round(p, 3))
}

# Singular/plural picker for a scalar count (avoids cli's numeric pluralizer).
.screen_plural <- function(n, singular, plural) if (n == 1L) singular else plural

# One multivariate-normality test line: a danger alert when significant, a success alert
# otherwise. `text` already carries the test statistic and p-value (no cli braces). Left
# unwrapped, unlike the prose alerts around it: it is a fixed-format statistic line, and
# wrapping would let a break fall inside the statistic.
.screen_mvn_line <- function(sig, text) {
  if (sig) cli::cli_alert_danger(text) else cli::cli_alert_success(text)
}

# Oxford-comma list of names ("a", "a and b", "a, b, and c").
.screen_and_list <- function(x) {
  if (length(x) == 1L) return(x)
  if (length(x) == 2L) return(paste(x, collapse = " and "))
  paste0(paste(x[-length(x)], collapse = ", "), ", and ", x[length(x)])
}

# Build the per-variable display data frame: the full per-item table for raw data
# (columns renamed to the standard MSA/SMC abbreviations), or MSA and SMC alone for a
# correlation-matrix input. Numeric columns are rounded to `digits`.
.screen_per_item_display <- function(x, digits) {
  if (!is.null(x$per_item)) {
    disp <- x$per_item
    num <- vapply(disp, is.numeric, logical(1))
    disp[num] <- lapply(disp[num], round, digits)
    names(disp)[names(disp) == "smc"] <- "SMC"
    names(disp)[names(disp) == "kmo_i"] <- "MSA"
    # `missing` is a percentage; the bare header reads as a count next to `variance`,
    # which is on the data's own scale.
    names(disp)[names(disp) == "missing"] <- "missing%"
    # A `flags` entry is NA for a variable treated as continuous, i.e. "no category
    # screening applies". Printed as <NA> that reads as a failed computation, so drop
    # the column when it says nothing at all and render the remaining NAs as a dash.
    if (all(is.na(disp$flags))) disp$flags <- NULL
    else disp$flags[is.na(disp$flags)] <- "-"
    disp
  } else {
    # make.unique() guards against duplicate variable names (a correlation matrix with
    # duplicate dimnames is a valid input), which base data.frame() row names reject; it
    # matches how the raw per-item table is keyed (efa_screen()).
    data.frame(MSA = round(x$kmo$KMO_i, digits),
               SMC = round(x$smc, digits),
               row.names = make.unique(names(x$kmo$KMO_i)))
  }
}

# Assemble the consolidated recommendations as parallel symbol/message vectors. Each
# check fires only when it applies; when no warning fires a single success line is
# returned, and a correlation-matrix input always closes with a raw-data note. Every
# recommendation is anchored in the literature cited in efa_screen()'s @source.
.screen_recommendations <- function(x, raw, kmo, bart_sig, mvn_nonnormal, multicollinear,
                                    digits) {
  sym <- character(0)
  msg <- character(0)
  push <- function(s, m) { sym[[length(sym) + 1L]] <<- s; msg[[length(msg) + 1L]] <<- m }

  # Few-category items (fewer than 5 response categories) are detected up front because
  # they steer the estimator recommendation away from the continuous-non-normal advice.
  few <- character(0)
  if (raw && !is.null(x$categories)) {
    n_cat <- vapply(x$categories, function(ct) {
      if (length(ct) == 1L && is.na(ct[1])) NA_integer_ else length(ct)
    }, integer(1))
    few <- names(n_cat)[!is.na(n_cat) & n_cat < 5L]
  }

  # Low overall sampling adequacy (KMO < .6): the set may not be factorable.
  if (!is.null(kmo) && !is.na(kmo) && kmo < .6) {
    push("!", paste0("Overall sampling adequacy is ", .kmo_band(kmo)$label,
                     " (KMO = ", round(kmo, digits),
                     "); the variables may not form a factorable set."))
  }

  # Variables with a low individual MSA (< .5) share little variance and are drop
  # candidates (Kaiser & Rice, 1974). make.unique() keys the names as the per-item table
  # and the sparse/empty recommendations do, so a variable is labelled consistently even
  # when the input has duplicate variable names.
  msa <- x$kmo$KMO_i
  low <- make.unique(names(msa))[!is.na(msa) & msa < .5]
  if (length(low)) {
    push("!", paste0(length(low), " ", .screen_plural(length(low), "variable has", "variables have"),
                     " a low individual MSA (< .5): ", .screen_and_list(low),
                     "; consider removing ", .screen_plural(length(low), "it", "them"),
                     " (little shared variance)."))
  }

  # Few-category ordinal data: prefer a categorical estimator (Rhemtulla et al., 2012).
  if (length(few)) {
    push("!", paste0(length(few), " ", .screen_plural(length(few), "item has", "items have"),
                     " fewer than 5 response categories; treating ",
                     .screen_plural(length(few), "it", "them"),
                     " as ordinal (polychoric correlations with a categorical estimator such as DWLS)",
                     " is less biased than normal-theory ML."))
  }

  # Continuous / 5+-category non-normal data: normal-theory standard errors are biased.
  # Suppressed when few-category items already routed the user to an ordinal estimator.
  if (isTRUE(mvn_nonnormal) && !length(few)) {
    push("!", paste("These data depart from multivariate normality; normal-theory standard errors",
                    "and fit statistics may be biased - prefer robust (sandwich) or bootstrapped",
                    "standard errors."))
  }

  # An over-powered Bartlett's test: significant but built on the normality it violates.
  if (bart_sig && isTRUE(mvn_nonnormal)) {
    push("!", paste("Bartlett's test is significant, but it assumes multivariate normality and grows",
                    "more sensitive as N increases; because these data are non-normal, treat it as",
                    "uninformative here and rely on the KMO."))
  }

  # Multicollinearity: a condition index above 30, the value Belsley (1991) gives for
  # flagging a near linear dependency, decided once by the caller from the same band
  # table as the section verdict. A high index does not have to come from one redundant
  # pair - a set of items that is together nearly linearly dependent gives the same
  # result - so the remedy names both.
  if (isTRUE(multicollinear)) {
    push("!", paste("A high condition index indicates multicollinearity; look for a redundant item",
                    "pair (r > .8) or a larger set of items that is nearly linearly dependent, and",
                    "consider removing some of them."))
  }

  if (raw) {
    flags <- x$per_item$flags
    items <- rownames(x$per_item)
    sparse <- items[!is.na(flags) & grepl("sparse", flags)]
    if (length(sparse)) {
      push("!", paste0(length(sparse), " ", .screen_plural(length(sparse), "variable has", "variables have"),
                       " a sparse response category (< 5 responses): ", .screen_and_list(sparse),
                       "; a low-frequency category can destabilise polychoric estimates",
                       " - consider collapsing it into an adjacent category."))
    }
    empty <- items[!is.na(flags) & grepl("empty", flags)]
    if (length(empty)) {
      push("!", paste0(length(empty), " ", .screen_plural(length(empty), "variable has", "variables have"),
                       " an unused interior response category (a gap in the scale): ",
                       .screen_and_list(empty), "; check the coding."))
    }
    if (!inherits(x$outliers, "efa_screen_no_outliers")) {
      nf <- length(x$outliers$flagged)
      nominal <- 1 - (x$settings$outlier_cutoff %||% 0.975)
      frac <- nf / x$outliers$n_complete
      # Far more flags than the cutoff nominally admits is not a contamination count to
      # work through: a high-breakdown estimate fitted to the most concentrated half of a
      # clustered or mixture sample legitimately calls the rest distant, so the excess is
      # evidence that the data are not elliptically distributed. The rate alone is not
      # enough to draw that conclusion - at a tight cutoff, or in a small sample, a couple
      # of genuine outliers clear it - so the flagged set must also be too long to work
      # through case by case, which is what makes the ordinary advice below unhelpful.
      if (nf >= 10L && frac > 5 * nominal) {
        # The rate is over the complete cases, so with missing data both bases are named:
        # read against the rows supplied it would look milder than it is.
        of_those <- if (.screen_incomplete(x$outliers$n_complete, x$settings$n_obs)) {
          "% of those cases"
        } else {
          "%"
        }
        push("!", paste0(nf, " of ",
                         .screen_case_base(x$outliers$n_complete, x$settings$n_obs),
                         " (", round(100 * frac), of_those,
                         ") exceed the outlier cutoff, far above the ",
                         # keep a significant digit: a tight cutoff has a nominal rate well
                         # below 0.1%, which rounds to a self-defeating "0%"
                         format(signif(100 * nominal, 2), trim = TRUE, scientific = FALSE),
                         "% expected under multivariate normality; this usually means the",
                         " data are not elliptically distributed (subgroups or a mixture)",
                         " rather than that this many cases are contaminated."))
      } else if (nf > 0L) {
        push("!", paste0(nf, " ", .screen_plural(nf, "observation was", "observations were"),
                         " flagged as ",
                         .screen_plural(nf, "a potential multivariate outlier", "potential multivariate outliers"),
                         "; inspect ", .screen_plural(nf, "it", "them"),
                         " (see `$outliers$flagged`) before down-weighting or excluding."))
      }
    }
  }

  # A clean bill of health when nothing above fired; the raw-data note always closes a
  # correlation-matrix report (its raw-only sections were absent above).
  if (!length(msg)) {
    push("v", "The data appear suitable for factor analysis.")
  }
  if (!raw) {
    push("i", paste("Per-item variance, missing-data, category, normality, and outlier diagnostics",
                    "require raw data; only a correlation matrix was supplied."))
  }

  list(symbol = sym, message = msg)
}

#' DOSPERT
#'
#' A list containing the bivariate correlations (cormat) of the 40 items of
#' the Domain Specific Risk Taking Scale (DOSPERT; Weber, Blais, & Betz, 2002)
#' and the sample size (N) based on the publicly available dataset at
#' (<https://osf.io/rce7g>) of the Basel-Berlin Risk Study (Frey et al., 2017).
#' The items measure risk-taking propensity on six different domains: social,
#' recreational, gambling, health/ safety, investment, and ethical.
#'
#' The underlying data deposit is licensed under CC BY 4.0
#' (<https://creativecommons.org/licenses/by/4.0/legalcode>). These correlations are a
#' derivative of it and are attributed to Frey et al. (2017), as the licence requires.
#'
#' @format A list of 2 with elements "cormat" (40 x 40 matrix of bivariate
#' correlations) and "N" (scalar).
#' \describe{
#'   \item{cormat}{(matrix) - Bivariate correlations of the 40 DOSPERT items,
#'   which span the six risk domains (social, recreational, gambling,
#'   health/safety, investment, and ethical).}
#'   \item{N}{(numeric) - The sample size the correlations are based on.}
#'  }
#' @source Weber, E. U., Blais, A.-R., & Betz, N. E. (2002). A domain specific risk-attitude scale: Measuring risk perceptions and risk behaviors. Journal of Behavioral Decision Making, 15(4), 263–290. doi: 10.1002/bdm.414
#' @source Frey, R., Pedroni, A., Mata, R., Rieskamp, J., & Hertwig, R. (2017). Risk preference shares the psychometric structure of major psychological traits. Science Advances, 3, e1701381.
#'
#' <https://osf.io/rce7g>
"DOSPERT"

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(EFAtools)

## -----------------------------------------------------------------------------
cor_mat <- test_models$baseline$cormat

# Recommended name -- exactly the same arguments as KMO():
efa_kmo(cor_mat)

# The old name still works and returns the same result:
identical(KMO(cor_mat)$KMO, efa_kmo(cor_mat)$KMO)

## -----------------------------------------------------------------------------
# Old flat interface
efa_old <- EFA(cor_mat, n_factors = 3, N = 500, type = "SPSS")

# New interface: the SPSS preset travels through the two control objects
efa_new <- efa_fit(cor_mat, n_factors = 3, N = 500,
                   estimate_control = estimate_control(type = "SPSS"),
                   rotate_control = rotate_control(type = "SPSS"))

# Identical numerical result
all.equal(efa_old$rot_loadings, efa_new$rot_loadings)

## ----eval = FALSE-------------------------------------------------------------
# efa_fit(cor_mat, n_factors = 3, N = 500, rotation = "promax",
#         estimate_control = estimate_control(type = "SPSS", max_iter = 500),
#         rotate_control   = rotate_control(type = "SPSS", k = 3))

## ----eval = FALSE-------------------------------------------------------------
# efa_fit(cor_mat, n_factors = 3, N = 500, max_iter = 500)
# #> Error: `max_iter` cannot be passed to `efa_fit()` directly.
# #> i The estimation and rotation tuning knobs live in `estimate_control()` and `rotate_control()`.
# #> i For example: `efa_fit(x, ..., estimate_control = estimate_control(max_iter = 500))`.

## -----------------------------------------------------------------------------
class(efa_new)
inherits(efa_new, "EFA")

#' Hull method for determining the number of factors to retain
#'
#' Implementation of the Hull method suggested by Lorenzo-Seva, Timmerman,
#' and Kiers (2011), with an extension to principal axis factoring. See details for
#' parallelization.
#'
#' @param x matrix or data.frame. Dataframe or matrix of raw data or matrix with
#' correlations.
#' @param N numeric. Number of cases in the data. This is passed to [efa_parallel].
#'  Only has to be specified if x is a correlation matrix, otherwise it is determined
#'  based on the dimensions of x.
#' @param n_fac_theor numeric. Theoretical number of factors to retain. One plus
#'   the larger of this number and the number of factors suggested by [efa_parallel]
#'   is used as the upper bound *J* of factors to extract in the Hull method.
#' @param estimator character. The estimator to use. One of  `"PAF"`,
#'    `"ULS"`, or  `"ML"`, for principal axis factoring, unweighted
#'    least squares, and maximum likelihood, respectively. Default is `"PAF"`.
#' @param gof character. The goodness of fit index to use. Either `"CAF"`,
#'   `"CFI"`, or `"RMSEA"`, or any combination of them.
#'   With the `"PAF"` estimator, only
#'   the CAF can be used as goodness of fit index. For details on the CAF, see
#'   Lorenzo-Seva, Timmerman, and Kiers (2011).
#' @param eigen_type character. On what the eigenvalues should be found in the
#'  parallel analysis. Can be one of `"SMC"`, `"PCA"`, or `"EFA"`.
#'   If using  `"SMC"` (default), the diagonal of the correlation matrices is
#'    replaced by the squared multiple correlations (SMCs) of the indicators. If
#'     using  `"PCA"`, the diagonal values of the correlation
#'  matrices are left to be 1. If using  `"EFA"`, eigenvalues are found on the
#'  correlation  matrices with the final communalities of an EFA solution as
#'  diagonal. This is passed to  [efa_parallel()].
#' @param use character. Passed to [stats::cor()] if raw data
#' is given as input. Default is `"pairwise.complete.obs"`.
#' @param cor_method character. One of `"pearson"`, `"spearman"`, or `"kendall"`,
#'   passed to [stats::cor()]. `"poly"` and `"tetra"` are not supported because
#'   `HULL` derives its factor-search bound from an internal parallel analysis
#'   against continuous reference data.
#'  Default is  `"pearson"`.
#' @param n_datasets numeric. The number of datasets to simulate. Must be at
#'   least 1. Default is 1000. This is passed to [efa_parallel()].
#' @param percent numeric. The percentile to take from the simulated eigenvalues.
#'  Default is 95. This is passed to [efa_parallel()].
#' @param decision_rule character. Which rule to use to determine the number of
#' factors to retain. Default is `"means"`, which will use the average
#' simulated eigenvalues. `"percentile"`, uses the percentiles specified
#' in percent. `"crawford"` uses the 95th percentile for the first factor
#' and the mean afterwards (based on Crawford et al, 2010). This is passed to [efa_parallel()].
#' @param n_factors numeric. Number of factors to extract if  `"EFA"` is
#' included in `eigen_type`. Default is 1. This is passed to
#' [efa_parallel()].
#' @param estimate_control an [estimate_control()] object with the estimation settings for the
#'  [efa_fit()] fits of the 0 to *J* factor solutions, and for the fit inside the internal
#'  [efa_parallel()] call. `NULL` (default) uses the [efa_fit()] defaults. This object carries
#'  estimation settings only; the fits are always unrotated, which the hull statistics (CFI,
#'  RMSEA, CAF) do not depend on.
#' @param ... Further arguments passed to [efa_fit()], also in
#' [efa_parallel()]. The estimation tuning knobs are not passed here; they live in
#' `estimate_control`, a rotation setting is not accepted because the fits are unrotated, and
#' neither are the standard-error arguments (`se`, `b_boot`, `ci`, `seed`), because the fits
#' are internal steps whose standard errors are not reported.
#'
#' @details The Hull method aims to find a model with an optimal balance between
#'  model fit and number of parameters, retaining only major factors
#'  (Lorenzo-Seva, Timmerman, & Kiers, 2011). It fits 0 to *J* factors -- where
#'  *J* is the number of factors suggested by parallel analysis (or `n_fac_theor`,
#'  if that is larger), plus one -- keeps the solutions on the upper boundary of
#'  the convex hull of goodness-of-fit against degrees of freedom, and selects the
#'  one at the sharpest elbow, i.e. with the highest *st* value.
#'
#'  Because it trades fit against parsimony instead of testing against a null model
#'  of uncorrelated variables, the Hull method does not lose accuracy for the
#'  correlated-factor structures where parallel analysis ([efa_parallel()]) tends to
#'  under-extract; the CAF variant in particular was among the more accurate criteria
#'  in Auerswald and Moshagen (2019). It needs at least six indicators and fits a
#'  model at every candidate factor count, so it is comparatively slow and is not an
#'  option for very short scales.
#'
#' The [efa_parallel] function and the principal axis factoring of the
#'   different number of factors can be parallelized using the future framework,
#'   by calling the [future::plan()] function. The examples
#'    provide example code on how to enable parallel processing.
#'
#'   The upper bound *J* comes from [efa_parallel()], which compares against simulated
#'   data, so the suggested number of factors varies slightly from run to run; a
#'   criterion-based rotation passed through `...` adds its own random starts. Call
#'   [base::set.seed()] beforehand to make a run reproducible; the result is then also
#'   independent of the parallel plan.
#'
#'   Note that if `gof = "RMSEA"` is used, 1 - RMSEA is actually used to
#'   compare the different solutions. This is necessary due to how the heuristic to
#'   locate the elbow of the hull works.
#'
#'   The solutions are fitted without inequality constraints, so a solution can be
#'   inadmissible (a Heywood case, or a fit that did not converge). Only the selected
#'   solution is checked for this; if it is inadmissible a warning is raised and the
#'   retained number of factors should be interpreted with caution.
#'
#'   The ML estimation method uses the [psych::fa()]
#'    starting values. See also the [efa_fit] documentation.
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). Its main fields are:
#' \item{n_factors}{A named numeric vector with the suggested number of factors
#'   for each requested goodness-of-fit index (`"CAF"`, `"CFI"`, and/or
#'   `"RMSEA"`).}
#' \item{results}{A list with one record per goodness-of-fit index, each holding
#'   the goodness-of-fit values, the degrees of freedom, the hull membership, and
#'   the retained solution used for printing and plotting. Each record also carries
#'   `st`, the elbow sharpness of every solution (see details): the retained
#'   solution has the largest value, and the runner-up shows how close the
#'   selection was. `st` is `NA` for the solutions where it is undefined, that is
#'   for those not on the hull and for the two hull endpoints, which have no
#'   neighbouring hull solution on one side. When fewer than three solutions
#'   remain on the hull, `st` is undefined throughout and the whole vector is `NA`;
#'   the retained solution is then the one with the highest goodness of fit, and a
#'   warning says so.}
#' \item{settings}{A list of the settings used, including `n_fac_max`, the upper
#'   bound *J* of the number of factors to extract (see details). For backwards
#'   compatibility the estimator is also repeated in `settings$method`.}
#'
#' For backwards compatibility the per-index suggestions are additionally available
#' as the top-level fields `n_fac_CAF`, `n_fac_CFI` and `n_fac_RMSEA`, each `NA` if
#' that index was not requested in `gof`. New code should read them from
#' `n_factors` instead.
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. (2011).
#' The Hull method for selecting the number of common factors. Multivariate
#' Behavioral Research, 46(2), 340-364.
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#'
#' @export
#'
#' @examples
#' \donttest{
#' # using PAF (this will print a message if gof is not specified manually
#' # and CAF will be used automatically)
#' efa_hull(test_models$baseline$cormat, N = 500, gof = "CAF", n_datasets = 100)
#'
#' # using ML with all available fit indices (CAF, CFI, and RMSEA)
#' efa_hull(test_models$baseline$cormat, N = 500, estimator = "ML", n_datasets = 100)
#'
#' # using ULS with only RMSEA
#' efa_hull(test_models$baseline$cormat, N = 500, estimator = "ULS", gof = "RMSEA",
#'          n_datasets = 100)
#'}
#'
#'\dontrun{
#' # using parallel processing (Note: plans can be adapted, see the future
#' # package for details). future::plan() returns the plan it replaces, so
#' # on.exit() puts the session back as it was -- also if the call fails.
#' local({
#'   old_plan <- future::plan(future::multisession, workers = 2)
#'   on.exit(future::plan(old_plan), add = TRUE)
#'   efa_hull(test_models$baseline$cormat, N = 500, gof = "CAF")
#' })
#' }
efa_hull <- function(x, N = NA, n_fac_theor = NA,
                 estimator = c("PAF", "ULS", "ML"), gof = c("CAF", "CFI", "RMSEA"),
                 eigen_type = c("SMC", "PCA", "EFA"),
                 use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                         "everything", "na.or.complete"),
                 cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                 n_datasets = 1000, percent = 95,
                 decision_rule = c("means", "percentile", "crawford"),
                 n_factors = 1, estimate_control = NULL, ...) {
  # Perform hull method following Lorenzo-Seva, Timmerman, and Kiers (2011)

  .reject_flat_knobs(...names(), fn = "efa_hull")
  .reject_unknown_fit_dots(...names(), fn = "efa_hull", unrotated = TRUE)
  .reject_rotation_dots(list(...), fn = "efa_hull")
  .assert_cor_input(x)

  estimator <- .match_arg_ci(estimator)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)
  # The Hull method derives its factor-search bound from an internal parallel
  # analysis, whose reference data are continuous; poly/tetra are therefore not
  # supported, consistent with PARALLEL/NEST/CD.
  .reject_poly_reference(cor_method, "efa_hull")
  gof <- .match_arg_ci(gof, several.ok = TRUE)
  eigen_type <- .match_arg_ci(eigen_type)
  .assert_args({
    checkmate::assert_count(n_fac_theor, na.ok = TRUE)
    checkmate::assert_count(N, na.ok = TRUE)
    decision_rule <- .match_arg_ci(decision_rule)
    .assert_estimate_control(estimate_control)
    checkmate::assert_count(n_factors)
    checkmate::assert_count(n_datasets, positive = TRUE)
    checkmate::assert_number(percent, lower = 0, upper = 100)
  })

  if (ncol(x) < 6) {
    cli::cli_abort(
      c("The data has fewer than 6 indicators.",
        "i" = "The Hull method needs at least 6."),
      class = "efa_hull_min_indicators"
    )
  }

  if (estimator == "PAF" && !all(gof == "CAF")) {
    cli::cli_inform(
      c("i" = 'Only CAF can be used as gof if estimator "PAF" is used. Setting gof to "CAF"'),
      class = "efa_hull_gof_caf"
    )
    gof <- "CAF"
  }

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(
    x, N = N, use = use, cor_method = cor_method, N_policy = "required",
    singular_tail = "the Hull method cannot be executed",
    N_required_msg = "{.arg N} is not specified but is needed to compute some fit indices.")
  R <- prep$R
  N <- prep$N

  m <- ncol(R)

  # 1) perform parallel analysis to find J as n_fac_theor + 1
  par_res <- efa_parallel(R, N = N, eigen_type = eigen_type, estimator = estimator,
                      n_datasets = n_datasets, percent = percent,
                      decision_rule = decision_rule, n_factors = n_factors,
                      estimate_control = estimate_control, ...)

  n_fac_PA <- unname(par_res$n_factors[eigen_type])

  if (is.na(n_fac_PA)) {

    if (!is.na(n_fac_theor)) {
      J <- n_fac_theor + 1
    } else {
      J <- .det_max_factors(ncol(R))
    }

  } else {

    J <- max(c(n_fac_PA, n_fac_theor), na.rm = TRUE) + 1

  }

  if (J > .det_max_factors(ncol(R))) {
    J <- .det_max_factors(ncol(R))
    cli::cli_warn("Setting the maximum number of factors to {J} to ensure overidentified models.",
                  class = "efa_hull_max_factors")
  }

  # The Hull method needs at least three solutions (for 0 to J factors) to form a
  # hull, i.e. J of at least 2. With few indicators the largest over-identified
  # model can itself be below the usual floor of three factors, so cap the floor at
  # that maximum to avoid forcing an under-identified (df = 0) top model.
  hull_floor <- min(3L, .det_max_factors(ncol(R)))
  if (J < hull_floor) {
    cli::cli_warn(
      c("The suggested maximum number of factors was {J}, but the Hull method needs at least {hull_floor}.",
        "i" = "Setting it to {hull_floor}."),
      class = "efa_hull_min_factors"
    )
    J <- hull_floor

  }

  # 2) perform factor analysis for the range of dimensions 1:J and compute f and
  #    df for every solution
  s <- matrix(0, ncol = 4, nrow = J + 1)
  s[, 1] <- 0:J

  # first for 0 factors
  if ("CAF" %in% gof) {
    s_CAF <- s
    colnames(s_CAF) <- c("nfactors", "CAF", "df", "st")
    s_CAF[1, 2] <- 1 - .compute_kmo(R)$KMO
    s_CAF[1, 3] <- (m**2 - m) / 2
  }

  if ("CFI" %in% gof) {
    s_CFI <- s
    colnames(s_CFI) <- c("nfactors", "CFI", "df", "st")
    s_CFI[1, 2] <- 0
    s_CFI[1, 3] <- (m**2 - m) / 2
  }

  if ("RMSEA" %in% gof) {
    s_RMSEA <- s
    colnames(s_RMSEA) <- c("nfactors", "RMSEA", "df", "st")
    # 0-factor (independence model) reference, on the same uncorrected (N - 1) discrepancy scale
    # the RMSEA of the 1:J solutions uses in .chi_fit_indices(). When N is too small for the
    # Bartlett correction (the corrected null chi-square is NA) the chi-square asymptotics break
    # down, so the reference is left undefined too -- matching how .gof()/SMT() drop the RMSEA at
    # such N (where every fitted 1:J solution is likewise NA, the factor-count term only lowering
    # the multiplier further). The log-determinant is computed once and reused for both calls.
    ld_R <- determinant(R, logarithm = TRUE)
    df <- (m**2 - m) / 2
    # compute 1 - RMSEA
    s_RMSEA[1, 2] <- if (is.na(.null_chisq(R, N, ld = ld_R))) {
      NA_real_
    } else {
      1 - .rmsea_point(.null_chisq(R, N, ld = ld_R, corrected = FALSE), df, N)
    }
    s_RMSEA[1, 3] <- (m**2 - m) / 2

  }

  # Calculate loadings with the EFA function. future.seed = TRUE because a
  # criterion-based rotation passed through `...` (e.g. oblimin) draws random starts, so
  # the workers need a managed random-number stream; without one an oblique HULL run
  # would not be reproducible from set.seed() under a parallel plan.
  loadings <- suppressWarnings(future.apply::future_lapply(seq_len(J), efa_fit,
                                                           x = R,
                                                           estimator = estimator,
                                                           N = N,
                                                           estimate_control = estimate_control,
                                                           ...,
                                                           future.seed = TRUE))

  # then for 1 to J factors. estimator == "PAF" forces gof to "CAF" above, so the
  # gof-keyed blocks already cover it; the df is the same for every index (Eq 4
  # gives the free parameters, and the difference in df equals the difference in
  # free parameters).
  for (i in seq_len(J)) {
    if ("CAF" %in% gof) {
      # compute goodness of fit "f" as CAF (common part accounted for; Eq 3)
      s_CAF[i + 1, 2] <- loadings[[i]]$fit_indices$CAF
      s_CAF[i + 1, 3] <- loadings[[i]]$fit_indices$df
    }

    if ("CFI" %in% gof) {
      s_CFI[i + 1, 2] <- loadings[[i]]$fit_indices$CFI
      s_CFI[i + 1, 3] <- loadings[[i]]$fit_indices$df
    }

    if ("RMSEA" %in% gof) {
      # compute 1 - RMSEA
      s_RMSEA[i + 1, 2] <- 1 - loadings[[i]]$fit_indices$RMSEA
      s_RMSEA[i + 1, 3] <- loadings[[i]]$fit_indices$df
    }

  }

  out_CAF <- list(s_complete = NA, retain = NA)
  out_CFI <- list(s_complete = NA, retain = NA)
  out_RMSEA <- list(s_complete = NA, retain = NA)

  if("CAF" %in% gof) {
    out_CAF <- .hull_calc(s = s_CAF, J = J, gof_t = "CAF")
  }
  if("CFI" %in% gof) {
    out_CFI <- .hull_calc(s = s_CFI, J = J, gof_t = "CFI")
  }
  if("RMSEA" %in% gof) {
    out_RMSEA <- .hull_calc(s = s_RMSEA, J = J, gof_t = "RMSEA")
  }

  gof_results <- list(CAF = out_CAF, CFI = out_CFI, RMSEA = out_RMSEA)

  # one record per requested goodness-of-fit index (df vs. fit, with the hull
  # membership and the retained solution used for printing and plotting)
  results <- list()
  inadmissible <- character(0)
  for (g in c("CAF", "CFI", "RMSEA")) {
    if (!(g %in% gof)) next
    sol <- gof_results[[g]]$s_complete
    retain <- gof_results[[g]]$retain

    # The right-most hull solutions over-extract and may be inadmissible, so flag
    # only the selected solution if it has a Heywood case or did not converge.
    if (!is.na(retain) && retain >= 1) {
      fit_sel <- loadings[[retain]]
      issues <- c(if (length(fit_sel$heywood) > 0) "Heywood case",
                  if (isTRUE(fit_sel$convergence != 0)) "non-convergence")
      if (length(issues) > 0) {
        inadmissible <- c(
          inadmissible,
          paste0(g, ": ", .retention_count(retain), " factor",
                 if (retain != 1) "s" else "",
                 " (", paste(issues, collapse = ", "), ")")
        )
      }
    }

    results[[g]] <- list(
      name = g,
      label = g,
      n_factors = retain,
      plot_type = "hull",
      x = unname(sol[, "df"]),
      y = unname(sol[, g]),
      reference = NULL,
      threshold = NULL,
      highlight = retain,
      point_labels = unname(sol[, "nfactors"]),
      on_hull = gof_results[[g]]$on_hull,
      # elbow sharpness of each solution (Eq 5); the selected one has the largest
      # value, so the runner-up shows how close the decision was
      st = unname(sol[, "st"])
    )
  }

  if (length(inadmissible) > 0) {
    cli::cli_warn(
      c("The Hull method selected an inadmissible solution: {inadmissible}.",
        "i" = "The selected solution has a Heywood case or did not converge, so the retained number of factors may be unreliable; interpret it with caution and cross-check with other criteria."),
      class = "efa_hull_inadmissible"
    )
  }

  out <- .new_efa_retention(
    "HULL",
    results = unname(results),
    settings = list(N = N,
                    estimator = estimator,
                    # back-compat alias, as for the frozen P_type key
                    method = estimator,
                    gof = gof,
                    n_fac_theor = n_fac_theor,
                    eigen_type = eigen_type,
                    use = use,
                    cor_method = cor_method,
                    n_fac_max = J),
    subtitle = paste0("Estimator: ", estimator)
  )

  # back-compat aliases of the per-index suggestions, as for the frozen method key
  out$n_fac_CAF <- unname(out$n_factors["CAF"])
  out$n_fac_CFI <- unname(out$n_factors["CFI"])
  out$n_fac_RMSEA <- unname(out$n_factors["RMSEA"])

  return(out)

}


.hull_calc <- function(s, J, gof_t){

  # 3) sort n solutions by their df values and denoted by s (already done)

  # 4) all solutions s are excluded for which a solution sj (j<i) exists such
  #    that fj > fi (eliminate solutions not on the boundary of the convex hull)

  s_complete <- s

  # A non-finite goodness-of-fit value (e.g. an undefined CFI/RMSEA for a Heywood
  # or near-singular model) cannot lie on the convex hull; drop those solutions
  # with a classed warning rather than failing in the comparisons below.
  na_rows <- !is.finite(s[, 2])
  if (any(na_rows)) {
    cli::cli_warn(
      c("{sum(na_rows)} solution{?s} had a non-finite {gof_t} value and {?was/were} excluded from the hull.",
        "i" = "Inspect the affected models, or try a different goodness-of-fit index or estimation method."),
      class = "efa_hull_na_fit"
    )
    s <- s[!na_rows, , drop = FALSE]
  }

  if (nrow(s) < 1) {
    cli::cli_abort(
      c("No solution had a finite {gof_t} value, so the Hull method cannot proceed.",
        "i" = "Try a different goodness-of-fit index or estimation method."),
      class = "efa_hull_no_fit"
    )
  }

  d_s <- diff(s[, 2])
  while (any(d_s < 0)) {
    s <- s[c(1, d_s) > 0, , drop = FALSE]
    if(nrow(s) == 1){
      break
    }
    d_s <- diff(s[, 2])
  }

  # 5) all triplets of adjacent solutions are considered consecutively.
  #    The middle solution is excluded if its point is below or on the line
  #    connecting its neighbors in GOF vs df.

  # 6) repeat 5) until no solution can be excluded

  # The `i <= nr_s - 1` bound ensures the final interior triplet is also tested;
  # the loop is a no-op while fewer than three boundary solutions remain.
  nr_s <- nrow(s)
  i <- 2

  while(i <= nr_s - 1) {

    f1 <- s[i - 1, 2]
    f2 <- s[i, 2]
    f3 <- s[i + 1, 2]
    df1 <- s[i - 1, 3]
    df2 <- s[i, 3]
    df3 <- s[i + 1, 3]

    # compute f2 if it were on the line between f1 and f3
    p_f2 <- f1 + (f3 - f1) / (df3 - df1) * (df2 - df1)

    # check if f2 is below or on the predicted line and if so, remove it
    if (f2 <= p_f2) {
      s <- s[-i, , drop = FALSE]
      nr_s <- nr_s -1
      i <- 1
    }
    i <- i + 1
  }

  if (nrow(s) < 3) {
    cli::cli_warn(
      c("Fewer than three solutions were located on the hull using {gof_t} as goodness-of-fit index.",
        "i" = "Proceeding with the maximum-{gof_t} value as a heuristic; consider additional indices or methods as a robustness check."),
      class = "efa_hull_few_solutions"
    )

    # the st values are undefined with fewer than three hull solutions
    s_complete[, 4] <- NA_real_

    # 8) select solution with highest gof value
    retain <- s[which.max(s[, 2]), 1]


  } else {

    # 7) the st values of the hull solutions are determined (Eq 5)
    for (i in 2:(nrow(s) - 1)) {

      f_i <- s[i, 2]
      f_p <- s[i - 1, 2]
      f_n <- s[i + 1, 2]
      df_i <- s[i, 3]
      df_p <- s[i - 1, 3]
      df_n <- s[i + 1, 3]

      s[i, 4] <- ((f_i - f_p) / (df_i - df_p)) / ((f_n - f_i) / (df_n - df_i))

    }

    # Combine values. Only the interior hull solutions have an st value: the ratio
    # needs both a predecessor and a successor on the hull, so the two hull
    # endpoints are as undefined as the solutions that are not on the hull at all,
    # and are reported as NA. The local `s` still carries the zero its endpoints
    # were initialised with, which is why the reported column is rebuilt here
    # rather than copied wholesale from it.
    interior <- s[-c(1, nrow(s)), , drop = FALSE]
    s_complete[, 4] <- NA_real_
    s_complete[match(interior[, 1], s_complete[, 1]), 4] <- interior[, 4]

    # 8) select solution with highest st value
    retain <- s[which.max(s[, 4]), 1]

  }


  # hull membership: the solutions that survived the boundary/triplet elimination.
  # Derived from the surviving solutions rather than from `st`, so the retained
  # solution is still flagged as on the hull in the < 3 fallback where `st` is
  # undefined.
  on_hull <- s_complete[, 1] %in% s[, 1]

  out <- list(s_complete = s_complete,
              retain = unname(retain),
              on_hull = on_hull)

  return(out)

}

#' GRiPS_raw
#'
#' A data.frame containing responses to the General Risk Propensity Scale (GRiPS, Zhang,
#' Highhouse & Nye, 2018) of 810 participants of Study 1 of Steiner and Frey (2020).
#' The original data can be accessed via <https://osf.io/kxp8t/>.
#'
#' @format A data.frame with 810 rows (participants) and 8 columns, one per GRiPS
#' item. Each item is a self-report indicator of general risk propensity, labelled
#' by a keyword from the item:
#' \describe{
#'   \item{fun}{(numeric) - Risk-taking makes life more fun.}
#'   \item{friends}{(numeric) - Friends would describe the respondent as a risk taker.}
#'   \item{enjoy}{(numeric) - Enjoyment of taking risks.}
#'   \item{hurt}{(numeric) - Willingness to take a risk even if it might hurt.}
#'   \item{part}{(numeric) - Risk-taking as an important part of life.}
#'   \item{commonly}{(numeric) - Commonly takes risks.}
#'   \item{chances}{(numeric) - Belief in taking chances.}
#'   \item{attracted}{(numeric) - Attracted, rather than scared, by risk.}
#'  }
#' @source Zhang, D. C., Highhouse, S., & Nye, C. D. (2019). Development and validation of the general risk propensity scale (GRiPS). Journal of Behavioral Decision Making, 32, 152–167. doi: 10.1002/bdm.2102
#' @source Steiner, M., & Frey, R. (2020). Representative design in psychological assessment: A case study using the Balloon Analogue Risk Task (BART). PsyArXiv Preprint. doi:10.31234/osf.io/dg4ks
"GRiPS_raw"

#' Print and summarise an efa object
#'
#' `print()` shows a concise overview of an [efa_fit()] or [efa_mi()] solution:
#' a model header, the loading matrix (with the factor intercorrelations for
#' oblique solutions), the variances accounted for, and the model fit.
#' `summary()` returns a `summary.efa` object whose print method adds the full
#' diagnostics: model and simple-structure diagnostics, confidence-interval
#' tables, the structure matrix, multiple-imputation uncertainty (for pooled
#' objects), and residual diagnostics. `format()` assembles the same report and
#' returns it as a character vector; `print()` is `cat(format(x), sep = "\n")`.
#' The lines follow the active console theme, so they are plain when colours are
#' disabled (for example when captured into a file or stripped with
#' [cli::ansi_strip()]).
#'
#' @details
#' The methods are shared by single-imputation `efa` objects and pooled
#' `efa_mi` objects. For `efa_mi` objects the header reports the number
#' of imputations and the alignment/pooling settings; confidence intervals and a
#' multiple-imputation uncertainty summary are shown by `summary()` when the
#' pooled object carries bootstrap/MI quantities.
#'
#' In `summary()`, `ci_filter` controls which loading intervals are shown:
#' `"salient"` reports intervals for loadings whose absolute point estimate is at
#' least `cutoff`, `"nonzero"` reports intervals excluding zero, and `"all"`
#' reports every finite interval.
#'
#' @param x,object An object of class `efa` (from [efa_fit()]) or `efa_mi` (from [efa_mi()]); for
#'   the `summary.efa` methods, the object returned by `summary()`.
#' @param cutoff numeric. The absolute value at or above which loadings are
#'   emphasised in the loading table. Default is .3.
#' @param digits numeric. Number of decimal places for the printed tables.
#'   Default is 3.
#' @param max_name_length numeric. Maximum length of the variable names to
#'   display; longer names are cut from the right, or abbreviated where cutting
#'   would give two variables the same label. Applies to every table that names
#'   variables. `name_style` (see [print.efa_loadings()]) can be passed through
#'   `...` to choose the shortening explicitly, but it reaches the loading table
#'   only; the confidence-interval and simple-structure tables always shorten by
#'   cutting.
#' @param sort_loadings character. Optional row sorting for the loading table.
#'   See [print.efa_loadings()].
#' @param show_loading_legend logical. Whether to print a short legend for the
#'   loading-table styling. Default is `TRUE`. The legend is only printed when
#'   that styling is actually rendered (a colour-capable console); in plain
#'   output it is omitted and this argument has no effect.
#' @param max_factors_per_block numeric or `NULL`. Maximum number of factor
#'   columns per loading-table block. If `NULL`, chosen from the console width.
#' @param ci character. Which confidence intervals `summary()` shows, if
#'   available. `"auto"` and `"separate"` print CI sections when CIs were
#'   computed; `"none"` suppresses them. Default is `"auto"`.
#' @param ci_filter character. Which loading CIs `summary()` prints: `"salient"`
#'   (default), `"all"`, or `"nonzero"`; see Details.
#' @param diagnostics_top_n numeric. Maximum number of item-level entries
#'   `summary()` prints per simple-structure diagnostic.
#' @param residual_cutoff numeric. Absolute residual cutoff for the residual
#'   diagnostics in `summary()`. Default is .1.
#' @param residual_top_n numeric. Maximum number of residuals `summary()` prints.
#'   Use `Inf` to print all residuals above `residual_cutoff`.
#' @param show_structure logical. Whether `summary()` prints the structure matrix
#'   for oblique solutions when available. Default is `TRUE`.
#' @param cross_loading_cutoff numeric. Cutoff for counting cross-loadings in the
#'   `summary()` diagnostics. Defaults to `cutoff`.
#' @param min_primary_gap numeric. Minimum desired absolute difference between the
#'   largest and second-largest absolute loading of an item, used in the
#'   `summary()` diagnostics.
#' @param min_salient_per_factor numeric. Minimum number of salient indicators per
#'   factor used in the `summary()` diagnostics. Default is 3.
#' @param show_mi_diagnostics logical or `NULL`. Whether `summary()` prints a
#'   multiple-imputation uncertainty summary for pooled EFAs. `NULL` shows it for
#'   pooled objects.
#' @param ... Further arguments passed to [print.efa_loadings()].
#'
#' @returns `print()` and the print method for `summary.efa` objects return their
#'   argument invisibly. `format()` returns a character vector with the report
#'   lines. `summary()` returns an object of class `summary.efa`.
#'
#' @export
#'
#' @method print efa
#'
#' @examples
#' mod <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
#'                estimator = "PAF", rotation = "promax")
#' mod
#'
#' # The full diagnostics, CI tables, and residual diagnostics:
#' summary(mod)
#'
#' # format() returns the same lines as a character vector, e.g. for a report file:
#' writeLines(format(mod))
#'
print.efa <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa
#' @export
#' @method print efa_mi
print.efa_mi <- function(x, ...) {
  print.efa(x, ...)
}

#' @rdname print.efa
#' @export
#' @method format efa
format.efa <- function(x, cutoff = .3, digits = 3, max_name_length = 10,
                       sort_loadings = c("none", "primary", "clustered"),
                       show_loading_legend = TRUE,
                       max_factors_per_block = NULL, ...) {
  sort_loadings <- .match_arg_ci(sort_loadings)

  .render_efa(x,
    view = "brief",
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    sort_loadings = sort_loadings,
    show_loading_legend = show_loading_legend,
    max_factors_per_block = max_factors_per_block,
    ...
  )
}

#' @rdname print.efa
#' @export
#' @method format efa_mi
format.efa_mi <- function(x, ...) {
  format.efa(x, ...)
}

#' @rdname print.efa
#' @export
#' @method summary efa
summary.efa <- function(object, cutoff = .3, digits = 3, max_name_length = 10,
                        ci = c("auto", "none", "separate"),
                        ci_filter = c("salient", "all", "nonzero"),
                        diagnostics_top_n = 10,
                        residual_cutoff = .1,
                        residual_top_n = 10,
                        show_structure = TRUE,
                        sort_loadings = c("none", "primary", "clustered"),
                        show_loading_legend = TRUE,
                        cross_loading_cutoff = cutoff,
                        min_primary_gap = .20,
                        min_salient_per_factor = 3,
                        max_factors_per_block = NULL,
                        show_mi_diagnostics = NULL, ...) {
  ci <- .match_arg_ci(ci)
  ci_filter <- .match_arg_ci(ci_filter)
  sort_loadings <- .match_arg_ci(sort_loadings)

  opts <- list(
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    diagnostics_top_n = diagnostics_top_n,
    residual_cutoff = residual_cutoff,
    residual_top_n = residual_top_n,
    show_structure = show_structure,
    sort_loadings = sort_loadings,
    show_loading_legend = show_loading_legend,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    max_factors_per_block = max_factors_per_block,
    show_mi_diagnostics = show_mi_diagnostics
  )

  structure(list(efa = object, opts = opts), class = "summary.efa")
}

#' @rdname print.efa
#' @export
#' @method summary efa_mi
summary.efa_mi <- function(object, ...) {
  summary.efa(object, ...)
}

#' @rdname print.efa
#' @export
#' @method print summary.efa
print.summary.efa <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa
#' @export
#' @method format summary.efa
format.summary.efa <- function(x, ...) {
  opts <- utils::modifyList(x$opts, list(...))
  do.call(.render_efa, c(list(x$efa, view = "full"), opts))
}

# Render the EFA report at the requested depth: "brief" (print) shows the header,
# loadings (with factor intercorrelations), variances accounted for, and model fit;
# "full" (summary) additionally shows model and simple-structure diagnostics, the
# CI tables, the structure matrix, MI uncertainty, and residual diagnostics.
.render_efa <- function(x, view = c("brief", "full"),
                        cutoff = .3, digits = 3, max_name_length = 10,
                        ci = c("auto", "none", "separate"),
                        ci_filter = c("salient", "all", "nonzero"),
                        diagnostics_top_n = 10,
                        residual_cutoff = .1,
                        residual_top_n = 10,
                        show_structure = TRUE,
                        sort_loadings = c("none", "primary", "clustered"),
                        show_loading_legend = TRUE,
                        cross_loading_cutoff = cutoff,
                        min_primary_gap = .20,
                        min_salient_per_factor = 3,
                        max_factors_per_block = NULL,
                        show_mi_diagnostics = NULL, ...) {
  view <- .match_arg_ci(view)
  ci <- .match_arg_ci(ci)
  ci_filter <- .match_arg_ci(ci_filter)
  sort_loadings <- .match_arg_ci(sort_loadings)
  full <- identical(view, "full")

  .efa_validate_print_options(
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    diagnostics_top_n = diagnostics_top_n,
    residual_cutoff = residual_cutoff,
    residual_top_n = residual_top_n,
    show_structure = show_structure,
    show_loading_legend = show_loading_legend,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    max_factors_per_block = max_factors_per_block,
    show_mi_diagnostics = show_mi_diagnostics
  )

  spec <- .efa_print_spec(x)
  show_mi <- if (!is.null(show_mi_diagnostics)) {
    isTRUE(show_mi_diagnostics)
  } else {
    isTRUE(spec$is_pooled)
  }

  # Assemble the whole report inside a single cli container so prose wraps to the console
  # width; `format()` returns these lines and `print()` is `cat(format(x), sep = "\n")`.
  cli::cli_format_method({
    .print_efa_header(spec)

    if (full) {
      .print_efa_diagnostics_section(x, spec,
        cutoff = cutoff,
        digits = digits,
        cross_loading_cutoff = cross_loading_cutoff,
        min_primary_gap = min_primary_gap,
        min_salient_per_factor = min_salient_per_factor
      )
    }

    .print_efa_loadings_section(x, spec,
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      ci = if (full) ci else "none",
      ci_filter = ci_filter,
      show_structure = isTRUE(show_structure) && full,
      sort_loadings = sort_loadings,
      show_loading_legend = show_loading_legend,
      max_factors_per_block = max_factors_per_block,
      ...
    )

    if (full) {
      .print_efa_simple_structure_section(x, spec,
        cutoff = cutoff,
        digits = digits,
        max_name_length = max_name_length,
        cross_loading_cutoff = cross_loading_cutoff,
        min_primary_gap = min_primary_gap,
        min_salient_per_factor = min_salient_per_factor,
        diagnostics_top_n = diagnostics_top_n,
        sort_loadings = sort_loadings
      )
    }

    .print_efa_variances_section(x, spec, digits = digits)

    # A negative df (underidentified) stops the report after the warning; df == 0 (just
    # identified) emits its caveat but the fit/bootstrap/residual sections still follow.
    if (.print_efa_identification_warning(spec)) {
      .print_efa_model_fit_section(x, spec)
      .print_efa_bootstrap_note(x, spec, ci = if (full) ci else "none")

      if (full && isTRUE(show_mi)) {
        .print_efa_mi_diagnostics_section(x, spec, digits = digits)
      }

      if (full) {
        .print_efa_residuals_section(x,
          residual_cutoff = residual_cutoff,
          residual_top_n = residual_top_n,
          digits = digits
        )
      }
    }
  })
}

.efa_print_spec <- function(x) {
  settings <- x$settings
  if (is.null(settings)) {
    settings <- list()
  }

  se <- settings$se
  is_pooled <- inherits(x, "EFA_POOLED") || isTRUE(settings$pooled)

  # The two-stage pooled solution IS a single fit on the pooled inputs, so that fit's
  # Heywood detector is the pooled solution's own. Without it the count falls back to
  # `h2 >= 1`, which cannot see the boundary uniqueness an ML/ULS solution expresses an
  # improper solution as. The Rubin routes average several fits and have no such
  # detector, so they keep the fallback.
  heywood <- x[["heywood"]]
  if (is.null(heywood) && !is.null(x[["mi_fit"]])) {
    heywood <- x[["mi_fit"]][["heywood"]]
  }

  list(
    estimator = settings$estimator,
    rotation = settings$rotation,
    rotation_diagnostics = settings$rotation_diagnostics,
    cor_method = settings$cor_method,
    # `[[` rather than `$`: the slot is absent on every non-FIML fit, and `$` partial-matches, so
    # any future slot whose name merely starts with "fiml" would be picked up here instead.
    fiml = x[["fiml"]],
    se = se,
    N = settings$N,
    fit = x$fit_indices,
    h2 = x$h2,
    heywood = heywood,
    np_boot = identical(.efa_setting_text(se), "np-boot"),
    ci = .efa_ci_level(x),
    rmsea_ci_level = .efa_rmsea_ci_level(x),
    b_boot = settings$b_boot,
    max_iter = settings$max_iter,
    iter = x$iter,
    convergence = x$convergence,
    is_pooled = is_pooled,
    n_imputations = .efa_n_imputations(x),
    target_method = settings$target_method,
    align_unrotated = settings$align_unrotated,
    fit_pool_method = settings$fit_pool_method,
    component_se = settings$component_se,
    has_ci = !is.null(x$CI)
  )
}

.print_efa_header <- function(spec) {
  cli::cli_text("")

  if (isTRUE(spec$is_pooled)) {
    .print_efa_pooled_header(spec)
  } else {
    cli::cli_verbatim(.efa_wrap_chunks(c("EFA", .efa_model_header_chunks(spec))))
  }

  # Flag two-stage FIML correlations (saturated mean/covariance EM-estimated from raw data
  # with missing values; Yuan, Marshall, & Bentler, 2002) so the analysed matrix is not read
  # as an ordinary complete-case correlation. Emitted verbatim (plain, no ANSI).
  if (identical(.efa_setting_text(spec$cor_method), "fiml")) {
    cli::cli_verbatim("Correlations: FIML (two-stage, missing data)")

    # A Stage-1 EM that stopped at its cap leaves the analysed matrix at the last iterate
    # rather than the FIML estimate, which can differ materially at a high fraction of
    # missing information. The warning fires only while the fit runs -- and is suppressed
    # wherever fits are run in bulk -- so a saved object would otherwise carry no visible
    # trace of it. Shown next to the line that names the correlation method.
    if (isFALSE(spec$fiml$converged)) {
      cli::cli_alert_warning(
        "The FIML EM stopped after {spec$fiml$iter} iteration{?s} without converging; the
         analysed correlation matrix is its last iterate. Raise {.arg fiml_max_iter}, or
         relax {.arg fiml_tol}, in {.fn estimate_control}."
      )
    }
  }

  if (.efa_iteration_nonconvergence(spec)) {
    cli::cli_text("")
    cli::cli_alert_danger(.efa_nonconvergence_banner(spec$estimator))
  }

  invisible(NULL)
}


# The estimator/rotation tail both model headers end in, as `.efa_wrap_chunks()` chunks.
# Emitted verbatim rather than via cli_text() so a "setting = 'value'" token is never split
# across a line break -- which is also why the values are style_bold() rather than {.strong}
# -- while the packer still breaks the line around those tokens.
.efa_model_header_chunks <- function(spec) {
  c("performed", "with",
    paste0("estimator = '", cli::style_bold(spec$estimator), "'"),
    "and",
    paste0("rotation = '", cli::style_bold(spec$rotation), "'."))
}

.print_efa_pooled_header <- function(spec) {
  n_imputations <- .efa_setting_text(spec$n_imputations)

  across <- if (nzchar(n_imputations)) {
    c("across", cli::style_bold(n_imputations), "imputations")
  } else {
    character(0)
  }
  cli::cli_verbatim(.efa_wrap_chunks(
    c("Pooled EFA", across, .efa_model_header_chunks(spec))
  ))

  pooling_settings <- .efa_pooled_settings_text(spec)
  if (length(pooling_settings) > 0L) {
    cli::cli_verbatim(.efa_wrap_chunks(
      c("Pooling settings:", .efa_setting_chunks(pooling_settings, end = "."))
    ))
  }

  invisible(NULL)
}

.efa_pooled_settings_text <- function(spec) {
  out <- character(0)

  # The MI2S sandwich route fits a single model to the pooled inputs, so the
  # per-imputation alignment and chi-square pooling settings (target_method,
  # align_unrotated, fit_pool_method) do not apply and are not reported.
  if (!identical(spec$component_se, "sandwich")) {
    target_method <- .efa_setting_text(spec$target_method)
    if (nzchar(target_method) && !identical(spec$rotation, "none")) {
      out <- c(out, paste0("target_method = '", target_method, "'"))
    }

    align_unrotated <- .efa_setting_text(spec$align_unrotated)
    if (nzchar(align_unrotated)) {
      out <- c(out, paste0("align_unrotated = '", align_unrotated, "'"))
    }

    fit_pool_method <- .efa_setting_text(spec$fit_pool_method)
    if (nzchar(fit_pool_method)) {
      out <- c(out, paste0("fit_pool_method = '", fit_pool_method, "'"))
    }
  }

  if (is.finite(spec$rmsea_ci_level) &&
      abs(spec$rmsea_ci_level - .90) > sqrt(.Machine$double.eps)) {
    out <- c(out, paste0("rmsea_ci_level = ", .efa_ci_level_text(spec$rmsea_ci_level)))
  }

  out
}

.print_efa_loadings_section <- function(x, spec, cutoff, digits, max_name_length,
                                        ci, ci_filter, show_structure,
                                        sort_loadings, show_loading_legend,
                                        max_factors_per_block, ...) {
  if (identical(spec$rotation, "none")) {
    .print_efa_loading_block(x, spec,
      component = "unrot_loadings",
      label = "unrotated",
      title = "Unrotated Loadings",
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      ci = ci,
      ci_filter = ci_filter,
      sort_loadings = sort_loadings,
      show_loading_legend = show_loading_legend,
      max_factors_per_block = max_factors_per_block,
      ...
    )
    return(invisible(NULL))
  }

  shown <- .print_efa_loading_block(x, spec,
    component = "rot_loadings",
    label = "rotated",
    title = "Rotated Loadings",
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    sort_loadings = sort_loadings,
    show_loading_legend = show_loading_legend,
    max_factors_per_block = max_factors_per_block,
    ...
  )
  if (!shown) {
    return(invisible(NULL))
  }

  if (!is.null(x$Phi)) {
    phi <- as.matrix(x$Phi)
    factor_names <- .efa_factor_names(phi)
    dimnames(phi) <- list(factor_names, factor_names)

    .print_efa_rule("Factor Intercorrelations")
    .efa_emit_lines(.efa_corr_lines(phi, digits = digits, lower_only = TRUE))
    .print_efa_phi_ci_section(x, spec, digits, ci)
  }

  if (!is.null(x$Structure) && isTRUE(show_structure)) {
    .print_efa_structure_section(x,
      cutoff = cutoff,
      digits = digits,
      max_name_length = max_name_length,
      sort_loadings = sort_loadings,
      max_factors_per_block = max_factors_per_block,
      ...
    )
  }

  invisible(NULL)
}

# Render one loading block -- its rule, the styled table, the Heywood warning, and the
# loading confidence intervals -- for either the unrotated or the rotated matrix.
# `component` names the slot on `x` and `label` the adjective used in the messages.
# Returns TRUE when a matrix was rendered, FALSE when the slot was empty, so the rotated
# caller knows whether to continue with the Phi and Structure blocks.
#
# Both call sites must pass every formal *by name*. The print methods forward the user's
# `...` down to here unfiltered, so an exactly-matched formal is what stops an abbreviated
# user argument (`ti =`, `lab =`) from partial-matching `title` or `label` instead of being
# passed on to format().
.print_efa_loading_block <- function(x, spec, component, label, title, cutoff, digits,
                                     max_name_length, ci, ci_filter, sort_loadings,
                                     show_loading_legend, max_factors_per_block, ...) {
  .print_efa_rule(title)

  loadings <- x[[component]]
  if (is.null(loadings)) {
    cli::cli_text("No {label} loading matrix available.")
    return(FALSE)
  }

  # format.efa_loadings() renders the styled table (h2/sorting/legend/column blocks) and
  # returns it as report lines, which .efa_emit_lines() embeds without reflowing. The
  # loading matrices are classed `efa_loadings`, so the generic dispatches there.
  .efa_emit_lines(format(loadings,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = x$h2,
    sort_loadings = sort_loadings,
    legend = show_loading_legend,
    max_factors_per_block = max_factors_per_block,
    ...
  ))
  .print_efa_heywood_warning(spec$heywood, spec$h2)
  .print_efa_loading_ci_section(x, spec,
    component = component,
    loadings = loadings,
    loading_label = label,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    ci = ci,
    ci_filter = ci_filter,
    sort_loadings = sort_loadings
  )

  TRUE
}

.print_efa_variances_section <- function(x, spec, digits = 3) {
  .print_efa_rule("Variances Accounted for")

  if (identical(spec$rotation, "none")) {
    vars_accounted <- x$vars_accounted
  } else {
    vars_accounted <- x$vars_accounted_rot
  }

  if (is.null(vars_accounted)) {
    cli::cli_text("No variance-accounted table available.")
    return(invisible(NULL))
  }

  .efa_emit_lines(.efa_corr_lines(vars_accounted, digits = digits))

  invisible(NULL)
}

# Format a "corr"-role matrix (factor intercorrelations or variances accounted for) into
# decimal-aligned lines for embedding in the cli report via `cli::cli_verbatim()`. Reuses
# the shared `.efa_format_matrix()` renderer, so these tables align consistently with the
# loading table; rows/columns are labelled from the matrix dimnames and `lower_only = TRUE`
# blanks the strictly-upper triangle (for symmetric matrices such as Phi).
.efa_corr_lines <- function(values, digits = 3, lower_only = FALSE, na = "NA") {
  values <- as.matrix(values)
  .efa_format_matrix(
    values = values,
    row_labels = .efa_variable_names(values),
    col_labels = .efa_factor_names(values),
    col_roles = rep("corr", ncol(values)),
    digits = digits,
    lower_only = lower_only,
    na = na
  )
}

# Emit pre-formatted table lines into the cli report. `cli::cli_verbatim()` silently drops
# interior empty-string elements, which would erase the blank line a loading table puts
# before its legend and the blank separators between column blocks of a wide table. So emit
# each run of non-empty lines verbatim (no reflow) and each blank line as `cli_text("")`,
# preserving the table's own vertical spacing.
.efa_emit_lines <- function(lines) {
  if (length(lines) < 1L) {
    return(invisible(NULL))
  }

  blank <- !nzchar(lines)
  run <- cumsum(c(TRUE, blank[-1L] != blank[-length(blank)]))
  for (r in unique(run)) {
    idx <- which(run == r)
    if (blank[idx[1L]]) {
      for (i in idx) cli::cli_text("")
    } else {
      cli::cli_verbatim(lines[idx])
    }
  }

  invisible(NULL)
}

# Emit a bullet list whose items embed user-supplied variable/factor names. `cli::cli_ul()`
# treats each item as a cli/glue template, so a name containing `{` or `}` would be evaluated
# as an expression (corrupting output or aborting). Double the braces so they render
# literally.
.efa_emit_bullets <- function(items) {
  items <- gsub("}", "}}", gsub("{", "{{", items, fixed = TRUE), fixed = TRUE)
  cli::cli_ul(items)
  invisible(NULL)
}

.print_efa_identification_warning <- function(spec) {
  df <- .efa_fit_scalar(spec$fit, "df")

  if (!is.finite(df)) {
    return(TRUE)
  }

  if (df == 0) {
    cli::cli_text("")
    cli::cli_alert_warning(
      "The model is just identified (df = 0). Goodness of fit indices may not be interpretable.",
      wrap = TRUE
    )
    return(TRUE)
  }

  if (df < 0) {
    # The residual summaries are populated in the object even here, so the note names what is
    # missing rather than claiming that nothing was computed; the wording is shared with the
    # warning efa_fit() raises for the same model.
    chisq_block <- .fit_unavailable_text("chisq_block")
    residuals_kept <- .fit_unavailable_text("residuals_kept")
    not_identification <- .fit_unavailable_text("residuals_not_identification")
    cli::cli_text("")
    cli::cli_alert_warning(
      "The model is underidentified (df < 0); {chisq_block}. {residuals_kept} {not_identification}",
      wrap = TRUE
    )
    return(FALSE)
  }

  TRUE
}

.print_efa_model_fit_section <- function(x, spec) {
  fit <- spec$fit
  fit_ci <- .efa_fit_ci_labels(x, spec)

  .print_efa_rule("Model Fit")

  if (is.null(fit) || length(fit) < 1L) {
    cli::cli_text("No model-fit indices available.")
    return(invisible(NULL))
  }

  # One "name[ CI-tag]: value[ CI]" data line. The CI tag is emitted only for
  # indices whose own CI suffix is available, so partially populated fit-index
  # CI lists do not label plain point estimates as interval estimates.
  fit_line <- function(name, key, ci = fit_ci[[key]], tag = TRUE, ...) {
    if (is.null(ci)) ci <- ""
    label <- if (isTRUE(tag) && nzchar(ci)) fit_ci$label else ""
    paste0(name, label, ": ", .efa_format_fit_value(fit, key, pad = FALSE, ...), ci)
  }
  is_finite_fit <- function(key) is.finite(.efa_fit_scalar(fit, key))

  estimator <- .efa_setting_text(spec$estimator)
  df <- .efa_fit_scalar(fit, "df")
  chi <- .efa_fit_scalar(fit, "chi")
  p_chi <- .efa_fit_scalar(fit, "p_chi")

  if (identical(estimator, "PAF") || .efa_is_missing_number(spec$N) ||
      !is.finite(chi) || !is.finite(df)) {
    lines <- fit_line("CAF", "CAF")
    if (is_finite_fit("SRMR")) {
      lines <- c(lines, fit_line("SRMR", "SRMR"))
    }
    lines <- c(lines, paste0("df: ",
      .efa_format_fit_value(fit, "df", digits = 0, print_zero = TRUE, pad = FALSE)))
    .efa_emit_lines(lines)
    # DWLS has no unscaled chi-square: the statistic exists only in the scaled-and-shifted
    # form the robust covariance supplies, so without se = "sandwich" the three lines above
    # are the whole of the fit output and nothing else says why. Name what is missing with
    # the wording shared with the other routes that lose the same block. The note is held to
    # a model that a scaled statistic would apply to: with df = 0 the caveat above already
    # covers it, and a non-finite df is not a DWLS matter. PAF gets no such line -- it has no
    # discrepancy function to test, so no setting restores the block for it.
    if (identical(estimator, "DWLS") && !identical(spec$se, "sandwich") &&
        is.finite(df) && df > 0) {
      chisq_block <- .fit_unavailable_text("chisq_block")
      cli::cli_text("")
      cli::cli_text("{.emph Note:} With {.code estimator = \"DWLS\"}, {chisq_block}; refit
                     with {.code se = \"sandwich\"} for the scaled chi-square.")
    }
    return(invisible(NULL))
  }

  p_text <- if (!is.finite(p_chi)) {
    " = NA"
  } else if (p_chi < .001) {
    " < .001"
  } else {
    paste0(" = ", .efa_format_number(p_chi, digits = 3, pad = FALSE))
  }

  df_text <- .efa_format_fit_value(fit, "df", digits = 0, print_zero = TRUE, pad = FALSE)
  chi_text <- .efa_format_fit_value(fit, "chi", digits = 2, print_zero = TRUE, pad = FALSE)

  rmsea_label <- paste0("RMSEA [", .efa_ci_level_text(spec$rmsea_ci_level), " CI]")
  rmsea_value <- paste0(
    .efa_format_fit_value(fit, "RMSEA", pad = FALSE), " [",
    .efa_format_fit_value(fit, "RMSEA_LB", pad = FALSE), "; ",
    .efa_format_fit_value(fit, "RMSEA_UB", pad = FALSE), "]"
  )

  # When the sandwich path supplies a scaled (Satorra-Bentler) chi-square, mark the line so
  # it is not read as an ordinary chi-square (the CFI/TLI/RMSEA below are derived from it).
  # A two-stage FIML fit whose correction could not be formed reports the plain
  # likelihood-ratio statistic under the "uncorrected.lrt" tag; it must NOT be labelled
  # scaled, because it carries no correction at all -- it is only approximately
  # chi-square(df) under the two-stage estimator, which is what the label says.
  # A D2-pooled fit is flagged the same way: its chi-square is the Li et al. (1991)
  # asymptotic statistic df * F_D2, whose p-value comes from the D2 reference F(df1, df2)
  # rather than the chi-square(df) tail (the note below spells this out when they diverge).
  is_d2 <- isTRUE(spec$is_pooled) && identical(fit[["pool_method"]], "D2")
  scaled_type <- fit[["chi_scaled_type"]]
  chi_prefix <- if (!is.null(scaled_type) && !is.na(scaled_type)) {
    if (identical(scaled_type, "uncorrected.lrt")) "uncorrected " else "scaled "
  } else if (is_d2) {
    "D2-pooled "
  } else {
    ""
  }

  # The chi-square line carries an italic "p"; emit verbatim (with style_italic) so the
  # styling survives and the line is not reflowed.
  lines <- paste0(chi_prefix, "\u03c7\u00b2(", df_text, ") = ", chi_text, ", ",
                  cli::style_italic("p"), p_text)
  # A D2-pooled fit reports CFI and TLI as the average of the per-imputation indices
  # rather than from the pooled statistic on the line above, so label them: they are the
  # quantities that will not reconcile against a reference which pools the model and
  # baseline noncentralities separately (see the note below). Only an index that is
  # actually reported is labelled, so an undefined one is not described as an average.
  inc_averaged <- if (is_d2) {
    c("CFI", "TLI")[c(is_finite_fit("CFI"), is_finite_fit("TLI"))]
  } else {
    character(0)
  }
  inc_tag <- function(key) {
    if (key %in% inc_averaged) " (avg. over imputations)" else ""
  }
  lines <- c(lines, fit_line(paste0("CFI", inc_tag("CFI")), "CFI"))
  if (is_finite_fit("TLI")) {
    lines <- c(lines, fit_line(paste0("TLI", inc_tag("TLI")), "TLI"))
  }
  lines <- c(lines, paste0(rmsea_label, fit_ci$label, ": ", rmsea_value, fit_ci$RMSEA))
  lines <- c(lines,
             fit_line("AIC", "AIC", fit_ci$AIC, print_zero = TRUE),
             fit_line("BIC", "BIC", fit_ci$BIC, print_zero = TRUE))
  if (is_finite_fit("ECVI")) {
    lines <- c(lines, fit_line("ECVI", "ECVI", print_zero = TRUE))
  }
  lines <- c(lines, fit_line("CAF", "CAF"))
  if (is_finite_fit("SRMR")) {
    lines <- c(lines, fit_line("SRMR", "SRMR"))
  }

  .efa_emit_lines(lines)

  # The D2-pooled chi-square (Li, Meng, Raghunathan & Rubin, 1991) is the asymptotic
  # statistic df * F_D2 and its p-value comes from the reference F(df1, df2); state that
  # so the reported p is not mistaken for the chi-square(df) tail probability. The two
  # references coincide only as df2 -> Inf (no between-imputation variance), so the note is
  # shown when df2 is finite.
  if (is_d2) {
    d2_df2 <- if (is.null(x$mi_diagnostics)) NA_real_ else x$mi_diagnostics$D2_df2
    if (is.finite(d2_df2)) {
      d2_df2_text <- .efa_format_plain_number(d2_df2, digits = 1)
      cli::cli_text(
        "{.emph Note:} the pooled \u03c7\u00b2 is the D2 statistic; its {.emph p} uses the D2 reference F({df_text}, {d2_df2_text}), not the \u03c7\u00b2({df_text}) tail."
      )
    }
    # The incremental indices are averaged over the imputations rather than formed from
    # separately pooled model and baseline statistics, which is what lavaan.mi / semTools
    # do; say so, because these are the indices a reader is most likely to cross-check.
    # Names only the indices that were reported above.
    if (length(inc_averaged) > 0L) {
      cli::cli_text(
        "{.emph Note:} {inc_averaged} {?is/are} averaged over the imputations, not formed from the separately pooled model and baseline statistics in {.code mi_diagnostics}."
      )
    }
  }

  invisible(NULL)
}

.efa_fit_ci_labels <- function(x, spec) {
  out <- list(
    label = "",
    CAF = "",
    RMSR = "",
    SRMR = "",
    CFI = "",
    TLI = "",
    RMSEA = "",
    AIC = "",
    BIC = "",
    ECVI = ""
  )

  if (!isTRUE(spec$has_ci)) {
    return(out)
  }

  fitind_ci <- .efa_get_fit_ci(x, spec)
  if (is.null(fitind_ci)) {
    return(out)
  }

  out$label <- paste0(" [", .efa_ci_header_label(spec), "-CI]")
  out$CAF <- .efa_paste_gof_ci(fitind_ci, "CAF")
  out$RMSR <- .efa_paste_gof_ci(fitind_ci, "RMSR")
  out$SRMR <- .efa_paste_gof_ci(fitind_ci, "SRMR")
  out$CFI <- .efa_paste_gof_ci(fitind_ci, "CFI")
  out$TLI <- .efa_paste_gof_ci(fitind_ci, "TLI")
  out$RMSEA <- .efa_paste_gof_ci(fitind_ci, "RMSEA")
  out$AIC <- .efa_paste_gof_ci(fitind_ci, "AIC")
  out$BIC <- .efa_paste_gof_ci(fitind_ci, "BIC")
  out$ECVI <- .efa_paste_gof_ci(fitind_ci, "ECVI")

  out
}

.print_efa_bootstrap_note <- function(x, spec, ci = NULL) {
  if (!isTRUE(spec$has_ci)) {
    return(invisible(NULL))
  }

  # Analytic SEs produce Wald intervals, not resampled ones; describe their source rather
  # than a bootstrap-sample count. For a NON-pooled analytic fit the Wald CIs are attached
  # only to the unrotated loadings (displayed only for an unrotated solution with CIs
  # enabled), so suppress the note when no such table is shown. Pooled objects, however,
  # attach and display rotated-loading CIs on both the information and MI2S routes, so the
  # note must still appear for a rotated pooled solution. The wording follows the SE method:
  # se = "information" draws Wald CIs from the expected information matrix (Rubin-pooled with
  # plain Rubin (1987) df when pooled), while se = "sandwich" draws them from the robust (Godambe)
  # sandwich covariance (built on the multiple-imputation-pooled asymptotic covariance when
  # pooled), so the pooled MI2S note must not claim information-matrix Rubin pooling.
  if (!isTRUE(spec$np_boot)) {
    if (identical(ci, "none") ||
        (!identical(spec$rotation, "none") && !isTRUE(spec$is_pooled))) {
      return(invisible(NULL))
    }
    cli::cli_text("")
    if (isTRUE(spec$is_pooled) && identical(spec$component_se, "sandwich")) {
      cli::cli_text("{.emph Note:} Wald CIs from the robust sandwich covariance built on the multiple-imputation-pooled asymptotic covariance.")
    } else if (isTRUE(spec$is_pooled)) {
      cli::cli_text("{.emph Note:} Wald CIs from the expected information matrix, Rubin-pooled across imputations (Rubin's 1987 degrees of freedom).")
    } else if (identical(.efa_setting_text(spec$cor_method), "fiml")) {
      # Under cor_method = "fiml" both analytic settings yield the corrected two-stage sandwich, so
      # the note describes that covariance rather than the expected information matrix.
      cli::cli_text("{.emph Note:} Wald CIs from the corrected two-stage (FIML) sandwich covariance.")
    } else if (identical(spec$se, "sandwich")) {
      cli::cli_text("{.emph Note:} Wald CIs from the robust (Godambe) sandwich covariance.")
    } else {
      cli::cli_text("{.emph Note:} Wald CIs from the expected information matrix.")
    }
    return(invisible(NULL))
  }

  note_label <- if (isTRUE(spec$is_pooled)) {
    "Bootstrap/MI CIs"
  } else {
    "Bootstrap CIs"
  }

  note <- note_label

  b_text <- .efa_bootstrap_sample_text(x, spec)
  if (nzchar(b_text)) {
    note <- paste0(note, " based on ", b_text)
  }

  target_rotation_note <- .efa_target_rotation_note(x, spec)
  if (nzchar(target_rotation_note)) {
    note <- paste0(note, "; ", target_rotation_note)
  }

  note <- paste0(note, ".")

  cli::cli_text("")
  cli::cli_text("{.emph Note:} {note}")

  invisible(NULL)
}

.print_efa_residuals_section <- function(x, residual_cutoff = .1,
                                     residual_top_n = 10, digits = 3) {
  .print_efa_rule("Residual Diagnostics")

  if (is.null(x$residuals)) {
    cli::cli_text("No residual matrix available.")
    return(invisible(NULL))
  }

  residuals <- as.matrix(x$residuals)
  idx <- which(upper.tri(residuals) & is.finite(residuals), arr.ind = TRUE)

  if (nrow(idx) < 1L) {
    cli::cli_text("No finite off-diagonal residuals available.")
    return(invisible(NULL))
  }

  values <- residuals[idx]
  keep <- abs(values) > residual_cutoff
  n_large <- sum(keep, na.rm = TRUE)
  largest_abs <- max(abs(values), na.rm = TRUE)

  cutoff_text <- .efa_format_plain_number(residual_cutoff, digits)
  largest_text <- .efa_format_plain_number(largest_abs, digits)
  cli::cli_text("Residual cutoff: |r| > {cutoff_text}")
  cli::cli_text("Number of large residuals: {n_large}")
  cli::cli_text("Largest absolute residual: {largest_text}")

  if (n_large < 1L) {
    cli::cli_text("")
    cli::cli_text("No absolute residuals > {cutoff_text} occurred.")
    cli::cli_text("")
    cli::cli_text("Inspect the residual matrix for details (e.g., with residuals()).")
    return(invisible(NULL))
  }

  large_idx <- idx[keep, , drop = FALSE]
  large_values <- values[keep]
  ord <- order(abs(large_values), decreasing = TRUE)
  large_idx <- large_idx[ord, , drop = FALSE]
  large_values <- large_values[ord]

  if (is.finite(residual_top_n)) {
    n_print <- min(length(large_values), as.integer(residual_top_n))
  } else {
    n_print <- length(large_values)
  }

  large_idx <- large_idx[seq_len(n_print), , drop = FALSE]
  large_values <- large_values[seq_len(n_print)]

  # The residual matrix is symmetric, so prefer column names but fall back to row
  # names (unlike `.efa_variable_names()`, which only consults row names).
  var_names <- colnames(residuals)
  if (is.null(var_names)) {
    var_names <- rownames(residuals)
  }
  if (is.null(var_names)) {
    var_names <- paste0("V", seq_len(ncol(residuals)))
  }

  heading <- if (n_print < n_large) {
    paste0("Largest residuals (top ", n_print, " of ", n_large, "):")
  } else {
    "Largest residuals:"
  }

  bullets <- vapply(seq_along(large_values), function(i) {
    paste0(var_names[large_idx[i, 1]], " ~~ ", var_names[large_idx[i, 2]],
           ": ", .efa_format_plain_number(large_values[i], digits))
  }, character(1L))

  cli::cli_text("")
  cli::cli_text("{heading}")
  .efa_emit_bullets(bullets)
  cli::cli_text("")
  cli::cli_text("Inspect the residual matrix for details (e.g., with residuals()).")

  invisible(NULL)
}

.print_efa_loading_ci_section <- function(x, spec, component, loadings, loading_label,
                                          cutoff, digits, max_name_length,
                                          ci, ci_filter,
                                          sort_loadings = c("none", "primary", "clustered")) {
  sort_loadings <- .match_arg_ci(sort_loadings)

  if (!.efa_should_print_ci(x, ci)) {
    return(invisible(NULL))
  }

  loading_ci <- .efa_get_component_ci(x, component)
  if (is.null(loading_ci)) {
    return(invisible(NULL))
  }

  loadings <- as.matrix(loadings)
  lower <- as.matrix(loading_ci$lower)
  upper <- as.matrix(loading_ci$upper)

  if (!all(dim(loadings) == dim(lower)) || !all(dim(loadings) == dim(upper))) {
    return(invisible(NULL))
  }

  row_order <- .efa_loading_row_order(loadings, sort_loadings)
  loadings <- loadings[row_order, , drop = FALSE]
  lower <- lower[row_order, , drop = FALSE]
  upper <- upper[row_order, , drop = FALSE]

  keep <- .efa_select_loading_cis(loadings, lower, upper, cutoff, ci_filter)
  title <- .efa_loading_ci_title(spec, loading_label, ci_filter)
  .print_efa_rule(title)

  if (!any(keep, na.rm = TRUE)) {
    cli::cli_text("No loading CIs matched ci_filter = '{ci_filter}'.")
    return(invisible(NULL))
  }

  idx <- which(keep, arr.ind = TRUE)
  row_names <- .efa_truncate_names(.efa_variable_names(loadings), max_name_length)
  factor_names <- .efa_factor_names(loadings)

  .print_efa_ci_table(
    key = row_names[idx[, 1]],
    second_key = factor_names[idx[, 2]],
    estimate = loadings[idx],
    lower = lower[idx],
    upper = upper[idx],
    digits = digits,
    key_header = "Variable",
    second_key_header = "Factor",
    highlight = abs(loadings[idx]) >= cutoff
  )

  invisible(NULL)
}

.print_efa_phi_ci_section <- function(x, spec, digits, ci) {
  if (!.efa_should_print_ci(x, ci)) {
    return(invisible(NULL))
  }

  phi_ci <- .efa_get_component_ci(x, "Phi")
  if (is.null(phi_ci) || is.null(x$Phi)) {
    return(invisible(NULL))
  }

  phi <- as.matrix(x$Phi)
  lower <- as.matrix(phi_ci$lower)
  upper <- as.matrix(phi_ci$upper)

  if (!all(dim(phi) == dim(lower)) || !all(dim(phi) == dim(upper))) {
    return(invisible(NULL))
  }

  keep <- lower.tri(phi) & is.finite(phi) & is.finite(lower) & is.finite(upper)
  if (!any(keep, na.rm = TRUE)) {
    return(invisible(NULL))
  }

  idx <- which(keep, arr.ind = TRUE)
  factor_names <- .efa_factor_names(phi)

  .print_efa_rule(.efa_phi_ci_title(spec))
  .print_efa_ci_table(
    key = paste0(factor_names[idx[, 2]], " ~~ ", factor_names[idx[, 1]]),
    second_key = NULL,
    estimate = phi[idx],
    lower = lower[idx],
    upper = upper[idx],
    digits = digits,
    key_header = "Factors",
    second_key_header = NULL,
    highlight = NULL
  )

  invisible(NULL)
}

# Count Heywood cases: prefer the detector's `heywood` field (covers communalities
# >= 1 and ML/ULS boundary uniquenesses); fall back to communalities >= 1 for objects
# that carry no such field (e.g. pooled solutions).
.efa_heywood_count <- function(heywood, h2 = NULL) {
  if (!is.null(heywood)) {
    length(heywood)
  } else if (!is.null(h2)) {
    sum(h2 >= 1 + .Machine$double.eps, na.rm = TRUE)
  } else {
    0L
  }
}

# The Heywood count as it is reported in the diagnostics section. A pooled solution
# counts the cases in the pooled matrix, and averaging aligned solutions pulls boundary
# communalities back inside the admissible range, so that count can be 0 for a pool whose
# component fits were improper. Where the components recorded Heywood cases, report them
# beside the pooled count rather than leaving the pooled number to read as an all-clear.
.efa_heywood_value_text <- function(x, count) {
  adm <- x[["mi_admissibility"]]
  if (is.null(adm) || is.null(adm$n_heywood_items)) {
    return(count)
  }

  n_flags <- sum(adm$n_heywood_items)
  if (n_flags < 1L) {
    return(count)
  }

  # A variable flagged in several imputations counts once per imputation, so the total is
  # a number of flags rather than of distinct variables.
  paste0(count, " pooled (", n_flags, if (n_flags == 1L) " flag" else " flags",
         " across ", length(adm$heywood_imputations), " of ", adm$m, " imputations)")
}

.print_efa_heywood_warning <- function(heywood, h2 = NULL) {
  n_heywood <- .efa_heywood_count(heywood, h2)

  if (n_heywood >= 1) {
    cli::cli_text("")
    cli::cli_alert_danger("{n_heywood} Heywood case{?s} {?was/were} detected!")
  }

  invisible(NULL)
}

.print_efa_rule <- function(title) {
  cli::cli_text("")
  cli::cli_rule(left = "{.strong {title}}")
  cli::cli_text("")

  invisible(NULL)
}

.efa_should_print_ci <- function(x, ci) {
  !identical(ci, "none") && !is.null(x$CI)
}

.efa_get_component_ci <- function(x, component) {
  if (is.null(x$CI)) {
    return(NULL)
  }

  ci <- x$CI[[component]]
  if (.efa_is_ci_pair(ci)) {
    return(ci)
  }

  NULL
}

.efa_get_fit_ci <- function(x, spec = .efa_print_spec(x)) {
  if (is.null(x$CI)) {
    return(NULL)
  }

  components <- if (isTRUE(spec$is_pooled)) {
    c("fit_indices_descriptive", "fit_indices")
  } else {
    c("fit_indices", "fit_indices_descriptive")
  }

  for (component in components) {
    ci <- x$CI[[component]]
    if (.efa_is_ci_pair(ci)) {
      return(ci)
    }
  }

  NULL
}

.efa_is_ci_pair <- function(x) {
  is.list(x) && !is.null(x$lower) && !is.null(x$upper)
}

.efa_select_loading_cis <- function(loadings, lower, upper, cutoff, ci_filter) {
  finite <- is.finite(loadings) & is.finite(lower) & is.finite(upper)

  if (identical(ci_filter, "all")) {
    return(finite)
  }

  if (identical(ci_filter, "nonzero")) {
    return(finite & (lower > 0 | upper < 0))
  }

  finite & abs(loadings) >= cutoff
}

.efa_n_imputations <- function(x) {
  n_imputations <- x$settings$n_imputations

  if ((is.null(n_imputations) || length(n_imputations) < 1L || is.na(n_imputations[1])) &&
      !is.null(x$fits)) {
    n_imputations <- length(x$fits)
  }

  if (is.null(n_imputations) || length(n_imputations) < 1L || is.na(n_imputations[1])) {
    return(NA_integer_)
  }

  as.integer(n_imputations[1])
}

.efa_setting_text <- function(x) {
  if (is.null(x) || length(x) < 1L || is.na(x[1])) {
    return("")
  }

  as.character(x[1])
}

.efa_ci_level <- function(x) {
  val <- x$settings$ci
  if (is.null(val) || length(val) < 1L || is.na(val[1])) {
    return(NA_real_)
  }

  out <- suppressWarnings(as.numeric(val[1]))
  if (length(out) < 1L || is.na(out)) {
    return(NA_real_)
  }

  out
}

.efa_rmsea_ci_level <- function(x) {
  val <- x$settings$rmsea_ci_level
  if (is.null(val) || length(val) < 1L || is.na(val[1])) {
    return(.90)
  }

  out <- suppressWarnings(as.numeric(val[1]))
  if (length(out) < 1L || !is.finite(out) || out <= 0 || out >= 1) {
    return(.90)
  }

  out
}

.efa_ci_level_text <- function(level) {
  if (!is.finite(level)) {
    return("")
  }

  paste0(formatC(level * 100, format = "fg", digits = 4, width = -1), "%")
}

.efa_ci_source_text <- function(spec) {
  if (isTRUE(spec$is_pooled)) {
    # Label the pooled CI source by route. The information route Rubin-pools the
    # per-imputation Wald intervals across imputations ("Wald/MI"). The sandwich
    # (MI2S) route fits a single model to the pooled inputs and carries the
    # imputation uncertainty in the pooled asymptotic covariance rather than by
    # per-parameter Rubin pooling, so its intervals are Wald intervals from that
    # single sandwich fit ("sandwich/MI") -- not Rubin-pooled. The bootstrap
    # route's within-imputation variance is bootstrap-based ("bootstrap/MI").
    if (identical(spec$component_se, "information")) {
      return("Wald/MI")
    }
    if (identical(spec$component_se, "sandwich")) {
      return("sandwich/MI")
    }
    return("bootstrap/MI")
  }

  # Analytic intervals are Wald (estimate +/- z * SE); only the bootstrap CIs are percentile.
  if (!isTRUE(spec$np_boot)) {
    return("Wald")
  }

  "bootstrap"
}

# Reportable valid target-rotation counts: the non-NA counts as integers, or NULL when
# there is nothing to report. Shared by the CI-source note and the diagnostics summary,
# which pass the same counts through different wording.
.efa_valid_target_rotation_counts <- function(x) {
  counts <- .efa_valid_target_rotations(x)
  if (is.null(counts) || length(counts) < 1L || all(is.na(counts))) {
    return(NULL)
  }

  counts <- as.integer(stats::na.omit(counts))
  if (length(counts) < 1L) {
    return(NULL)
  }

  counts
}

.efa_target_rotation_note <- function(x, spec) {
  counts <- .efa_valid_target_rotation_counts(x)
  if (is.null(counts)) {
    return("")
  }

  total <- .efa_total_bootstrap_samples(x, spec)
  if (!is.na(total) && !any(counts < total)) {
    return("")
  }

  component_text <- .efa_target_rotation_component_text(x)
  count_text <- .efa_valid_target_rotation_count_text(counts, spec)
  total_text <- if (!is.na(total)) {
    paste0(" out of ", total)
  } else {
    ""
  }
  scope_text <- if (isTRUE(spec$is_pooled)) {
    " per imputation"
  } else {
    ""
  }

  paste0(
    component_text,
    " based on ", count_text,
    " valid target-rotated bootstrap samples",
    total_text,
    scope_text
  )
}

.efa_valid_target_rotations <- function(x) {
  if (!is.null(x$SE$valid_target_rotations)) {
    return(x$SE$valid_target_rotations)
  }

  if (!is.null(x$MI$bootstrap_rotation_valid)) {
    return(x$MI$bootstrap_rotation_valid)
  }

  NULL
}

.efa_total_bootstrap_samples <- function(x, spec) {
  if (!is.null(spec$b_boot) && length(spec$b_boot) >= 1L &&
      is.finite(spec$b_boot[1])) {
    return(as.integer(spec$b_boot[1]))
  }

  if (!is.null(x$replicates$unrot_loadings)) {
    arr <- x$replicates$unrot_loadings
    if (length(dim(arr)) >= 3L) {
      return(dim(arr)[3L])
    }
  }

  NA_integer_
}

.efa_target_rotation_component_text <- function(x) {
  if (!is.null(x$Phi)) {
    return("rotated-loading, factor-intercorrelation, and structure-coefficient CIs")
  }

  "rotated-loading CIs"
}

.efa_valid_target_rotation_count_text <- function(counts, spec) {
  counts <- as.integer(counts)

  if (!isTRUE(spec$is_pooled) || length(unique(counts)) == 1L) {
    return(as.character(counts[1L]))
  }

  paste0("between ", min(counts), " and ", max(counts))
}

.efa_ci_header_label <- function(spec) {
  level <- .efa_ci_level_text(spec$ci)
  source <- .efa_ci_source_text(spec)

  if (nzchar(level)) {
    paste(level, source)
  } else {
    source
  }
}

.efa_loading_ci_title <- function(spec, loading_label, ci_filter) {
  filter_text <- switch(ci_filter,
    salient = "salient",
    all = "all",
    nonzero = "nonzero",
    ci_filter
  )

  paste0(
    .efa_ci_header_label(spec),
    " CIs for ", filter_text, " ", loading_label, " loadings"
  )
}

.efa_phi_ci_title <- function(spec) {
  paste0(.efa_ci_header_label(spec), " CIs for factor intercorrelations")
}

# Variable labels for the report's name columns (CI tables, simple-structure bullets).
# Shares the loading table's shortening, so a variable carries the same label wherever the
# report names it -- including the collision handling, without which the loading table and
# the sections below it could label the same item differently.
.efa_truncate_names <- function(x, max_name_length) {
  .shorten_loadings_names(x, max_length = max_name_length, name_style = "truncate")
}

# Format a vector of CI numbers for the CI tables. `.efa_num` is vectorized, so one
# call formats the whole vector; the result feeds data.frame columns that drop names.
.efa_format_ci_num <- function(x, digits) {
  .efa_num(round(x, digits = digits), digits = digits)
}

.print_efa_ci_table <- function(key, second_key = NULL, estimate, lower, upper,
                                digits, key_header, second_key_header = NULL,
                                highlight = NULL) {
  has_second_key <- !is.null(second_key)

  # Right-pad each cell to its column width (base equivalent of str_pad side = "right").
  pad <- function(s, w) formatC(s, width = w, flag = "-")

  estimate_chr <- .efa_format_ci_num(estimate, digits)
  lower_chr <- .efa_format_ci_num(lower, digits)
  upper_chr <- .efa_format_ci_num(upper, digits)

  if (has_second_key) {
    plain_rows <- data.frame(
      key = as.character(key),
      second_key = as.character(second_key),
      estimate = estimate_chr,
      lower = lower_chr,
      upper = upper_chr,
      stringsAsFactors = FALSE
    )
    headers <- c(key_header, second_key_header, "est", "lower", "upper")
  } else {
    plain_rows <- data.frame(
      key = as.character(key),
      estimate = estimate_chr,
      lower = lower_chr,
      upper = upper_chr,
      stringsAsFactors = FALSE
    )
    headers <- c(key_header, "est", "lower", "upper")
  }

  widths <- vapply(seq_along(headers), function(j) {
    max(nchar(c(headers[j], plain_rows[[j]])), na.rm = TRUE)
  }, numeric(1))

  header <- paste(
    mapply(pad, headers, widths, USE.NAMES = FALSE),
    collapse = "  "
  )

  rows <- vapply(seq_len(nrow(plain_rows)), function(i) {
    cells <- as.character(unlist(plain_rows[i, , drop = FALSE], use.names = FALSE))
    padded <- mapply(pad, cells, widths, USE.NAMES = FALSE)

    if (has_second_key && !is.null(highlight) && isTRUE(highlight[i])) {
      padded[3] <- cli::style_bold(padded[3])
    }

    paste(padded, collapse = "  ")
  }, character(1L))

  # A fixed-width table; emit verbatim so cli does not reflow the aligned columns.
  .efa_emit_lines(c(header, rows))

  invisible(NULL)
}


# Argument checks shared by the print/format methods (print.efa() and its summary, and
# print.efa_loadings()), which validate the same display options with the same wording.
# `arg` names the offender and `class` carries its condition class, both of which are part
# of each method's contract; `call` defaults to the calling validator's frame, so routing
# the abort through a helper reports the same call as an inline check would.
#
# The whole-number tests below compare against `round()` rather than `as.integer()`: a
# value beyond the integer range coerces to NA with a warning, and the resulting NA in the
# `||` chain aborts with "missing value where TRUE/FALSE needed" instead of the classed
# condition each caller documents. `round()` is exact for every finite double, and the
# separate integer-range bound then rejects the out-of-range value through the proper
# condition (no display option has a legitimate use above 2^31).

# A single finite number, zero or greater.
.efa_check_nonneg_number <- function(value, arg, class, call = rlang::caller_env()) {
  if (!is.numeric(value) || length(value) != 1L || !is.finite(value) || value < 0) {
    cli::cli_abort("{.arg {arg}} must be a single finite non-negative number.",
                   class = class, call = call)
  }
  invisible(TRUE)
}

# A single finite whole number, zero or greater.
.efa_check_nonneg_int <- function(value, arg, class, call = rlang::caller_env()) {
  if (!.efa_is_whole_number(value) || value < 0) {
    cli::cli_abort("{.arg {arg}} must be a single finite non-negative integer.",
                   class = class, call = call)
  }
  invisible(TRUE)
}

# A single finite whole number, one or greater.
.efa_check_pos_int <- function(value, arg, class, call = rlang::caller_env()) {
  if (!.efa_is_whole_number(value) || value < 1) {
    cli::cli_abort("{.arg {arg}} must be a single finite positive integer.",
                   class = class, call = call)
  }
  invisible(TRUE)
}

# As .efa_check_pos_int(), but `NULL` (the "choose it from the console width" default) passes.
.efa_check_opt_pos_int <- function(value, arg, class, call = rlang::caller_env()) {
  if (!is.null(value) && (!.efa_is_whole_number(value) || value < 1)) {
    cli::cli_abort("{.arg {arg}} must be {.val NULL} or a single finite positive integer.",
                   class = class, call = call)
  }
  invisible(TRUE)
}

# TRUE for a single finite whole number inside the integer range; never NA.
.efa_is_whole_number <- function(value) {
  is.numeric(value) && length(value) == 1L && is.finite(value) &&
    abs(value) <= .Machine$integer.max && value == round(value)
}

# A single non-missing TRUE/FALSE.
.efa_check_flag <- function(value, arg, class, call = rlang::caller_env()) {
  if (!is.logical(value) || length(value) != 1L || is.na(value)) {
    cli::cli_abort("{.arg {arg}} must be {.val TRUE} or {.val FALSE}.",
                   class = class, call = call)
  }
  invisible(TRUE)
}

.efa_validate_print_options <- function(cutoff, digits, max_name_length,
                                        diagnostics_top_n,
                                        residual_cutoff, residual_top_n,
                                        show_structure, show_loading_legend,
                                        cross_loading_cutoff, min_primary_gap,
                                        min_salient_per_factor,
                                        max_factors_per_block,
                                        show_mi_diagnostics) {
  .efa_check_nonneg_number(cutoff, "cutoff", "efa_print_invalid_cutoff")
  .efa_check_nonneg_int(digits, "digits", "efa_print_invalid_digits")
  .efa_check_pos_int(max_name_length, "max_name_length",
                     "efa_print_invalid_max_name_length")

  if (!.efa_is_top_n(diagnostics_top_n)) {
    cli::cli_abort("{.arg diagnostics_top_n} must be a positive integer or {.val Inf}.",
                   class = "efa_print_invalid_diagnostics_top_n")
  }

  .efa_check_nonneg_number(residual_cutoff, "residual_cutoff",
                           "efa_print_invalid_residual_cutoff")

  if (!.efa_is_top_n(residual_top_n)) {
    cli::cli_abort("{.arg residual_top_n} must be a positive integer or {.val Inf}.",
                   class = "efa_print_invalid_residual_top_n")
  }

  .efa_check_flag(show_structure, "show_structure",
                  "efa_print_invalid_show_structure")
  .efa_check_flag(show_loading_legend, "show_loading_legend",
                  "efa_print_invalid_show_loading_legend")
  .efa_check_nonneg_number(cross_loading_cutoff, "cross_loading_cutoff",
                           "efa_print_invalid_cross_loading_cutoff")
  .efa_check_nonneg_number(min_primary_gap, "min_primary_gap",
                           "efa_print_invalid_min_primary_gap")
  .efa_check_pos_int(min_salient_per_factor, "min_salient_per_factor",
                     "efa_print_invalid_min_salient_per_factor")
  .efa_check_opt_pos_int(max_factors_per_block, "max_factors_per_block",
                         "efa_print_invalid_max_factors_per_block")

  if (!is.null(show_mi_diagnostics) &&
      (!is.logical(show_mi_diagnostics) || length(show_mi_diagnostics) != 1L ||
       is.na(show_mi_diagnostics))) {
    cli::cli_abort("{.arg show_mi_diagnostics} must be {.val TRUE}, {.val FALSE}, or {.val NULL}.",
                   class = "efa_print_invalid_show_mi_diagnostics")
  }

  invisible(TRUE)
}

.efa_iteration_nonconvergence <- function(spec) {
  # All current fits report a convergence code (0 = converged); a non-zero code
  # means the estimator stopped before meeting its tolerance. The iteration-count
  # fallback below covers objects that carry no convergence code but do report iter
  # and max_iter (e.g. solutions saved by an older package version).
  convergence <- .efa_as_scalar_number(spec$convergence)
  if (is.finite(convergence) && convergence != 0) {
    return(TRUE)
  }

  iter <- .efa_as_scalar_number(spec$iter)
  max_iter <- .efa_as_scalar_number(spec$max_iter)

  is.finite(iter) && is.finite(max_iter) && iter >= max_iter
}

# Message for the non-convergence banner (shared by print.efa() and
# print.efa_schmid_leiman()).
# PAF non-convergence always means the iteration cap was hit; the optimiser-based
# estimators report a convergence code that need not be an iteration limit.
.efa_nonconvergence_banner <- function(estimator) {
  if (identical(estimator, "PAF")) {
    "Maximum number of iterations reached without convergence"
  } else {
    "The optimiser did not converge"
  }
}

.efa_as_scalar_number <- function(x) {
  if (is.null(x) || length(x) < 1L) {
    return(NA_real_)
  }

  out <- tryCatch(
    suppressWarnings(as.numeric(x[1])),
    error = function(e) NA_real_
  )
  if (length(out) < 1L) {
    return(NA_real_)
  }

  out
}

.efa_is_missing_number <- function(x) {
  val <- .efa_as_scalar_number(x)
  !is.finite(val)
}

.efa_fit_scalar <- function(fit, name) {
  if (is.null(fit) || is.null(fit[[name]]) || length(fit[[name]]) < 1L) {
    return(NA_real_)
  }

  out <- tryCatch(
    suppressWarnings(as.numeric(fit[[name]][1])),
    error = function(e) NA_real_
  )
  if (length(out) < 1L) {
    return(NA_real_)
  }

  out
}

.efa_format_number <- function(x, digits = 2, print_zero = FALSE, pad = TRUE) {
  if (length(x) != 1L || !is.finite(x)) {
    return("NA")
  }

  .efa_num(round(x, digits = digits), digits = digits,
           print_zero = print_zero, pad = pad)
}

# Fit indices are reported to two decimals, not the three the loading, correlation and
# coefficient tables use. That is deliberate: CFI/TLI/RMSEA/SRMR/CAF are conventionally
# reported to two decimals in the literature, and the third digit is well inside the
# sampling error of any of them. The leading-zero convention is shared (via .efa_num()).
.efa_format_fit_value <- function(fit, name, digits = 2,
                                  print_zero = FALSE, pad = TRUE) {
  .efa_format_number(
    .efa_fit_scalar(fit, name),
    digits = digits,
    print_zero = print_zero,
    pad = pad
  )
}

.efa_bootstrap_sample_text <- function(x, spec) {
  b_boot <- .efa_as_scalar_number(spec$b_boot)
  if (!is.finite(b_boot)) {
    return("")
  }

  label <- if (isTRUE(spec$is_pooled)) {
    "bootstrap samples per imputation"
  } else {
    "bootstrap samples"
  }

  paste0(as.integer(b_boot), " ", label, .efa_usable_replicate_text(x, spec))
}

# " (n usable)" whenever fewer replicates survived than were requested, else "". The count is the
# one the unrotated-loading, residual and fit-index statistics are actually based on, so without it
# the printed sample count claims a precision the intervals do not have. The rotated block carries
# its own count (`valid_target_rotations`) and is reported separately. The requested total is
# resolved here rather than by the caller, so both call sites mean the same number by it.
.efa_usable_replicate_text <- function(x, spec) {
  counts <- .efa_valid_replicate_counts(x)
  total <- .efa_total_bootstrap_samples(x, spec)
  if (is.null(counts) || !is.finite(total) || !any(counts < total)) {
    return("")
  }

  scope <- if (isTRUE(spec$is_pooled)) " per imputation" else ""
  paste0(" (", .efa_valid_target_rotation_count_text(counts, spec), " usable", scope, ")")
}

# Replicates that were fitted and aligned successfully, as an integer vector (one entry per
# imputation for a pooled fit, one entry otherwise), or NULL when the object records none. A single
# fit stores the count; the pooled bootstrap route stores per-imputation FAILURE counts instead, so
# they are subtracted from the requested total there. Mirrors `.efa_valid_target_rotations()`, which
# resolves the rotated block's count from the same two places.
.efa_valid_replicate_counts <- function(x) {
  if (!is.null(x$SE$valid_replicates)) {
    counts <- x$SE$valid_replicates
  } else if (!is.null(x$MI$bootstrap_source_failures) &&
             !is.null(x$settings$b_boot)) {
    counts <- x$settings$b_boot[1] - x$MI$bootstrap_source_failures
  } else {
    return(NULL)
  }

  counts <- suppressWarnings(as.integer(stats::na.omit(as.numeric(counts))))
  if (length(counts) < 1L) {
    return(NULL)
  }

  counts
}

.efa_alignment_summary <- function(x) {
  alignment <- x$alignment
  if (is.null(alignment)) {
    return("")
  }

  out <- character(0)

  method <- .efa_setting_text(alignment$method)
  if (!nzchar(method)) {
    method <- .efa_setting_text(x$settings$target_method)
  }
  if (nzchar(method)) {
    out <- c(out, paste0("method = '", method, "'"))
  }

  if (!is.null(alignment$converged)) {
    converged <- isTRUE(alignment$converged)
    out <- c(out, if (converged) "converged" else "not converged")
  }

  failures <- alignment$point_rotation_failures
  if (!is.null(failures) && length(failures) > 0L) {
    failures <- failures[!is.na(failures)]
    if (length(failures) > 0L) {
      out <- c(out, paste0("point-alignment failures = ", length(failures)))
    }
  }

  paste(out, collapse = ", ")
}

.efa_is_top_n <- function(x) {
  is.numeric(x) && length(x) == 1L && !is.na(x) &&
    (is.infinite(x) || (x >= 1 && x == as.integer(x)))
}

.efa_main_loadings <- function(x, spec) {
  if (identical(spec$rotation, "none")) {
    if (is.null(x$unrot_loadings)) {
      return(NULL)
    }
    return(as.matrix(x$unrot_loadings))
  }

  if (is.null(x$rot_loadings)) {
    return(NULL)
  }

  as.matrix(x$rot_loadings)
}

.efa_variable_names <- function(x) {
  rn <- rownames(x)
  if (is.null(rn)) {
    rn <- paste0("V", seq_len(nrow(x)))
  }
  rn
}

.efa_factor_names <- function(x) {
  cn <- colnames(x)
  if (is.null(cn)) {
    cn <- paste0("F", seq_len(ncol(x)))
  }
  cn
}

.print_efa_diagnostics_section <- function(x, spec, cutoff, digits,
                                           cross_loading_cutoff,
                                           min_primary_gap,
                                           min_salient_per_factor) {
  loadings <- .efa_main_loadings(x, spec)

  title <- if (isTRUE(spec$is_pooled)) {
    "Pooled Model Diagnostics"
  } else {
    "Model Diagnostics"
  }
  .print_efa_rule(title)

  if (is.null(loadings)) {
    cli::cli_text("No loading matrix available for diagnostics.")
    return(invisible(NULL))
  }

  .efa_print_key_value("Factors", ncol(loadings))
  .efa_print_key_value("Variables", nrow(loadings))

  n_text <- .efa_setting_text(spec$N)
  if (nzchar(n_text) && !identical(n_text, "NA")) {
    .efa_print_key_value("N", n_text)
  }

  if (isTRUE(spec$is_pooled)) {
    n_imp <- .efa_setting_text(spec$n_imputations)
    if (nzchar(n_imp)) {
      .efa_print_key_value("Imputations", n_imp)
    }
    pooling <- .efa_pooled_settings_text(spec)
    if (length(pooling) > 0L) {
      # The only diagnostics entry whose value is a list of settings rather than a single
      # number, so it is the only one long enough to need the console-width packing the
      # pooled header uses; the rest stay one-line key/value pairs.
      cli::cli_verbatim(.efa_wrap_chunks(c("Pooling:", .efa_setting_chunks(pooling))))
    }

    alignment_text <- .efa_alignment_summary(x)
    if (nzchar(alignment_text)) {
      .efa_print_key_value("Alignment", alignment_text)
    }
  }

  if (isTRUE(spec$np_boot) && !is.null(spec$b_boot)) {
    sample_label <- if (isTRUE(spec$is_pooled)) {
      "Bootstrap samples per imputation"
    } else {
      "Bootstrap samples"
    }
    # Take the requested count from the same resolver the "(n usable)" suffix compares against, so
    # the two can never disagree about what was asked for or render it differently.
    b_total <- .efa_total_bootstrap_samples(x, spec)
    b_text <- if (is.na(b_total)) {
      .efa_setting_text(spec$b_boot)
    } else {
      as.character(b_total)
    }
    .efa_print_key_value(
      sample_label,
      paste0(b_text, .efa_usable_replicate_text(x, spec))
    )

    valid_text <- .efa_valid_target_rotation_summary(x, spec)
    if (nzchar(valid_text)) {
      .efa_print_key_value("Valid target-rotated samples", valid_text)
    }
  }

  # Distinct local optima the rotation found across its random starts, counted over the
  # starts that converged (only present for the gradient-projection rotations that use
  # random starts; omitted when none converged). Both counts are reported because the
  # screen-and-triage strategy optimizes only a few of the available starts, and only an
  # optimized start can yield a local optimum.
  # The n_starts_total check keeps an object serialized before the two-count diagnostics
  # existed (it carried a single `n_starts`) from rendering a garbled line: paste0() would
  # silently drop the missing counts. Such an object omits the line instead.
  rot_diag <- spec$rotation_diagnostics
  if (!is.null(rot_diag) && isTRUE(rot_diag$n_converged >= 1L) &&
      !is.null(rot_diag$n_starts_total)) {
    .efa_print_key_value(
      "Rotation local optima",
      paste0(rot_diag$n_distinct_minima, " distinct from ",
             rot_diag$n_optimized, " of ", rot_diag$n_starts_total, " starts")
    )
  }

  # Count Heywood cases the same way as the printed Heywood note (see
  # `.efa_heywood_count()`), qualified by the component fits on a pooled solution.
  heywood <- .efa_heywood_count(spec$heywood, spec$h2)
  .efa_print_key_value("Heywood cases", .efa_heywood_value_text(x, heywood))

  cross_salient <- abs(loadings) >= cross_loading_cutoff
  cross_salient[is.na(cross_salient)] <- FALSE
  cross_loading_items <- sum(rowSums(cross_salient) > 1L)

  salient <- abs(loadings) >= cutoff
  salient[is.na(salient)] <- FALSE
  no_salient_items <- sum(rowSums(salient) < 1L)
  .efa_print_key_value(
    paste0("Cross-loading items (|loading| >= ",
           .efa_format_plain_number(cross_loading_cutoff, digits), ")"),
    cross_loading_items
  )
  .efa_print_key_value(
    paste0("Items without salient loading (|loading| >= ",
           .efa_format_plain_number(cutoff, digits), ")"),
    no_salient_items
  )

  weak_factors <- .efa_factors_below_indicator_count(
    loadings,
    cutoff = cutoff,
    min_salient_per_factor = min_salient_per_factor
  )
  .efa_print_key_value(
    paste0("Factors with fewer than ", min_salient_per_factor,
           " salient indicators"),
    length(weak_factors)
  )

  gap_count <- .efa_small_primary_gap_count(loadings,
    cutoff = cutoff,
    min_primary_gap = min_primary_gap
  )
  .efa_print_key_value(
    paste0("Items with primary-loading gap < ",
           .efa_format_plain_number(min_primary_gap, digits)),
    gap_count
  )

  if (!is.null(x$residuals)) {
    largest_resid <- .efa_largest_abs_residual(x$residuals)
    if (is.finite(largest_resid)) {
      .efa_print_key_value("Largest |residual|",
        .efa_format_plain_number(largest_resid, digits)
      )
    }
  }

  if (!is.null(x$Phi)) {
    high_phi <- .efa_high_factor_correlation_count(x$Phi, cutoff = .85)
    value <- if (high_phi > 0L) {
      as.character(high_phi)
    } else {
      "none"
    }
    .efa_print_key_value("Factor intercorrelations > .85", value)
  }

  invisible(NULL)
}

.efa_print_key_value <- function(key, value) {
  # A "label: value" data line; emit verbatim so a long key or a settings value (e.g. the
  # pooled "Pooling: ..." line) is never reflowed mid-token.
  cli::cli_verbatim(paste0(key, ": ", as.character(value)))
  invisible(NULL)
}

.efa_valid_target_rotation_summary <- function(x, spec) {
  counts <- .efa_valid_target_rotation_counts(x)
  if (is.null(counts)) {
    return("")
  }

  total <- .efa_total_bootstrap_samples(x, spec)
  count_text <- .efa_valid_target_rotation_count_text(counts, spec)

  if (!is.na(total)) {
    count_text <- paste0(count_text, " out of ", total)
  }

  if (isTRUE(spec$is_pooled)) {
    count_text <- paste0(count_text, " per imputation")
  }

  count_text
}

.efa_factors_below_indicator_count <- function(loadings, cutoff,
                                               min_salient_per_factor) {
  salient <- abs(loadings) >= cutoff
  salient[is.na(salient)] <- FALSE
  counts <- colSums(salient)
  which(counts < min_salient_per_factor)
}

.efa_small_primary_gap_count <- function(loadings, cutoff, min_primary_gap) {
  length(.efa_small_primary_gap_rows(loadings, cutoff = cutoff,
                                     min_primary_gap = min_primary_gap))
}

.efa_largest_abs_residual <- function(residuals) {
  residuals <- as.matrix(residuals)
  values <- residuals[upper.tri(residuals)]
  values <- values[is.finite(values)]
  if (length(values) < 1L) {
    return(NA_real_)
  }
  max(abs(values), na.rm = TRUE)
}

.efa_high_factor_correlation_count <- function(phi, cutoff = .85) {
  phi <- as.matrix(phi)
  values <- phi[lower.tri(phi)]
  values <- values[is.finite(values)]
  sum(abs(values) > cutoff, na.rm = TRUE)
}

.print_efa_simple_structure_section <- function(x, spec, cutoff, digits,
                                                max_name_length,
                                                cross_loading_cutoff,
                                                min_primary_gap,
                                                min_salient_per_factor,
                                                diagnostics_top_n,
                                                sort_loadings) {
  loadings <- .efa_main_loadings(x, spec)
  row_order <- .efa_loading_row_order(loadings, sort_loadings)
  loadings <- loadings[row_order, , drop = FALSE]

  summary <- .efa_simple_structure_summary(
    loadings = loadings,
    cutoff = cutoff,
    cross_loading_cutoff = cross_loading_cutoff,
    min_primary_gap = min_primary_gap,
    min_salient_per_factor = min_salient_per_factor,
    digits = digits,
    max_name_length = max_name_length
  )

  if (!summary$has_output) {
    return(invisible(NULL))
  }

  .print_efa_rule("Simple Structure Diagnostics")

  # The diagnostics print as bulleted lists in this fixed order, each under its own
  # heading and each omitted when it has no entries. Only the weak-factor list is never
  # truncated: it holds at most one entry per factor.
  blocks <- list(
    list(values = summary$no_salient,
         top_n = diagnostics_top_n,
         heading = paste0("Items with no salient loading, |loading| >= ",
                          .efa_format_plain_number(cutoff, digits), ":")),
    list(values = summary$cross_loadings,
         top_n = diagnostics_top_n,
         heading = paste0("Items with cross-loadings, |loading| >= ",
                          .efa_format_plain_number(cross_loading_cutoff, digits),
                          " on multiple factors:")),
    list(values = summary$small_gaps,
         top_n = diagnostics_top_n,
         heading = paste0("Items with primary-loading gap < ",
                          .efa_format_plain_number(min_primary_gap, digits), ":")),
    list(values = summary$weak_factors,
         top_n = Inf,
         heading = paste0("Factors with fewer than ", min_salient_per_factor,
                          " salient indicators:"))
  )

  for (block in blocks) {
    if (length(block$values) > 0L) {
      .efa_print_limited_bullets(
        heading = block$heading,
        values = block$values,
        top_n = block$top_n
      )
    }
  }

  invisible(NULL)
}

.efa_simple_structure_summary <- function(loadings, cutoff, cross_loading_cutoff,
                                          min_primary_gap,
                                          min_salient_per_factor,
                                          digits, max_name_length) {
  var_names <- .efa_truncate_names(.efa_variable_names(loadings), max_name_length)
  factor_names <- .efa_factor_names(loadings)

  abs_loadings <- abs(loadings)
  salient_for_items <- abs_loadings >= cutoff
  salient_for_items[is.na(salient_for_items)] <- FALSE

  cross_salient <- abs_loadings >= cross_loading_cutoff
  cross_salient[is.na(cross_salient)] <- FALSE

  no_salient <- var_names[rowSums(salient_for_items) < 1L]

  cross_rows <- which(rowSums(cross_salient) > 1L)
  cross_loadings <- vapply(cross_rows, function(i) {
    idx <- which(cross_salient[i, ])
    paste0(var_names[i], ": ",
      .efa_format_loading_pairs(loadings[i, idx], factor_names[idx], digits)
    )
  }, character(1L))

  small_gap_rows <- .efa_small_primary_gap_rows(loadings,
    cutoff = cutoff,
    min_primary_gap = min_primary_gap
  )
  small_gaps <- vapply(small_gap_rows, function(i) {
    ord <- order(abs(loadings[i, ]), decreasing = TRUE)
    idx <- ord[seq_len(min(2L, length(ord)))]
    paste0(var_names[i], ": ",
      .efa_format_loading_pairs(loadings[i, idx], factor_names[idx], digits)
    )
  }, character(1L))

  factor_counts <- colSums(salient_for_items)
  weak_factor_idx <- which(factor_counts < min_salient_per_factor)
  if (length(weak_factor_idx) > 0L) {
    weak_factors <- paste0(
      factor_names[weak_factor_idx],
      ": ",
      factor_counts[weak_factor_idx],
      " salient indicator",
      ifelse(factor_counts[weak_factor_idx] == 1L, "", "s")
    )
  } else {
    weak_factors <- NULL
  }


  list(
    no_salient = no_salient,
    cross_loadings = cross_loadings,
    small_gaps = small_gaps,
    weak_factors = weak_factors,
    has_output = length(no_salient) > 0L ||
      length(cross_loadings) > 0L ||
      length(small_gaps) > 0L ||
      length(weak_factors) > 0L
  )
}

.efa_small_primary_gap_rows <- function(loadings, cutoff, min_primary_gap) {
  if (ncol(loadings) < 2L) {
    return(integer(0))
  }

  abs_loadings <- abs(loadings)
  abs_loadings[is.na(abs_loadings)] <- -Inf
  sorted <- t(apply(abs_loadings, 1L, sort, decreasing = TRUE))
  primary <- sorted[, 1L]
  secondary <- sorted[, 2L]

  which(is.finite(primary) & primary >= cutoff &
          is.finite(secondary) & (primary - secondary) < min_primary_gap)
}

.efa_format_loading_pairs <- function(values, factor_names, digits) {
  ord <- order(abs(values), decreasing = TRUE)
  values <- values[ord]
  factor_names <- factor_names[ord]

  paste0(
    factor_names,
    " = ",
    vapply(values, .efa_format_plain_number, character(1L), digits = digits),
    collapse = ", "
  )
}

.efa_print_limited_bullets <- function(heading, values, top_n) {
  cli::cli_text("{heading}")

  if (length(values) < 1L) {
    return(invisible(NULL))
  }

  n_print <- if (is.finite(top_n)) {
    min(length(values), as.integer(top_n))
  } else {
    length(values)
  }

  bullets <- values[seq_len(n_print)]
  if (n_print < length(values)) {
    bullets <- c(bullets, paste0("... ", length(values) - n_print, " more not shown"))
  }

  .efa_emit_bullets(bullets)
  cli::cli_text("")
  invisible(NULL)
}

.print_efa_structure_section <- function(x, cutoff, digits, max_name_length,
                                         sort_loadings, max_factors_per_block,
                                         ...) {
  .print_efa_rule("Structure Matrix")
  .efa_emit_lines(format(x$Structure,
    cutoff = cutoff,
    digits = digits,
    max_name_length = max_name_length,
    h2 = NULL,
    sort_loadings = sort_loadings,
    legend = FALSE,
    max_factors_per_block = max_factors_per_block,
    ...
  ))
  invisible(NULL)
}

.print_efa_mi_diagnostics_section <- function(x, spec, digits = 3) {
  if (!isTRUE(spec$is_pooled) || is.null(x$MI)) {
    return(invisible(NULL))
  }

  fmi_values <- .efa_collect_mi_values(x$MI, pattern = "fmi")
  riv_values <- .efa_collect_mi_values(x$MI, pattern = "riv")

  if (length(fmi_values) < 1L && length(riv_values) < 1L) {
    return(invisible(NULL))
  }

  .print_efa_rule("MI Uncertainty Summary")

  if (length(fmi_values) > 0L) {
    .efa_print_key_value("Largest available FMI",
      .efa_format_plain_number(max(fmi_values, na.rm = TRUE), digits)
    )
    .efa_print_key_value("Median available FMI",
      .efa_format_plain_number(stats::median(fmi_values, na.rm = TRUE), digits)
    )
  }

  if (length(riv_values) > 0L) {
    .efa_print_key_value("Largest available RIV",
      .efa_format_plain_number(max(riv_values, na.rm = TRUE), digits)
    )
    .efa_print_key_value("Median available RIV",
      .efa_format_plain_number(stats::median(riv_values, na.rm = TRUE), digits)
    )
  }

  invisible(NULL)
}

.efa_collect_mi_values <- function(x, pattern, path = "") {
  pieces <- .efa_collect_mi_values_impl(x, pattern = pattern, path = path)
  if (length(pieces) < 1L) {
    return(numeric(0))
  }

  values <- unlist(pieces, use.names = FALSE)
  values[is.finite(values)]
}

.efa_collect_mi_values_impl <- function(x, pattern, path = "") {
  if (is.list(x)) {
    nm <- names(x)
    if (is.null(nm)) {
      nm <- rep("", length(x))
    }

    pieces <- vector("list", length(x))
    for (i in seq_along(x)) {
      child_path <- paste(c(path, nm[i]), collapse = "/")
      pieces[[i]] <- .efa_collect_mi_values_impl(x[[i]], pattern, child_path)
    }
    return(unlist(pieces, recursive = FALSE, use.names = FALSE))
  }

  if (!is.numeric(x) || !grepl(pattern, path, ignore.case = TRUE)) {
    return(list())
  }

  list(as.numeric(x))
}

.efa_loading_row_order <- function(x, sort_loadings = c("none", "primary", "clustered")) {
  sort_loadings <- .match_arg_ci(sort_loadings)

  if (identical(sort_loadings, "none") || nrow(x) < 2L || ncol(x) < 1L) {
    return(seq_len(nrow(x)))
  }

  abs_x <- abs(x)
  abs_x[is.na(abs_x)] <- -Inf
  primary_factor <- max.col(abs_x, ties.method = "first")
  primary_loading <- apply(abs_x, 1L, max, na.rm = TRUE)
  primary_loading[!is.finite(primary_loading)] <- -Inf

  if (identical(sort_loadings, "primary")) {
    return(order(primary_factor, seq_along(primary_factor)))
  }

  order(primary_factor, -primary_loading, seq_along(primary_factor))
}

.efa_format_plain_number <- function(x, digits = 3, pad = FALSE) {
  if (length(x) != 1L || is.na(x)) {
    return("NA")
  }

  .efa_num(round(x, digits = digits), digits = digits, pad = pad)
}


.efa_paste_gof_ci <- function(fitind_ci, name) {
  if (!.efa_is_ci_pair(fitind_ci)) {
    return("")
  }

  lower_names <- names(fitind_ci$lower)
  upper_names <- names(fitind_ci$upper)
  if (is.null(lower_names) || is.null(upper_names) ||
      !(name %in% lower_names) || !(name %in% upper_names)) {
    return("")
  }

  tryCatch(.paste_gof_ci(fitind_ci, name), error = function(e) "")
}

# Snapshot and structural tests for the shared loading-matrix renderer
# (.efa_format_matrix) behind format.efa_loadings / format.efa_sl_loadings. The inputs are literal
# matrices with hand-chosen values (including negatives, a Heywood loading > 1, h2 > 1 and
# u2 < 0), so the printed decimals are platform-stable and the snapshots pin down vertical
# alignment, styling structure, the legend, and the block-wise wrapping directly. Snapshots
# are recorded under local_reproducible_output() (plain text, no ANSI).

make_loadings <- function() {
  L <- matrix(c( 0.82, -0.11,  0.05,
                 0.45,  0.60, -0.02,
                -0.30,  0.71,  0.123,
                 0.04, -0.05,  1.08),
              nrow = 4, byrow = TRUE,
              dimnames = list(c("fun", "friends_long_name", "enjoy", "hurt"),
                              c("F1", "F2", "F3")))
  class(L) <- c("efa_loadings", "LOADINGS")
  L
}

make_wide_loadings <- function() {
  W <- matrix(round(seq(-0.9, 0.9, length.out = 4 * 8), 2), nrow = 4,
              dimnames = list(c("v1", "v2", "v3", "v4"), paste0("F", 1:8)))
  class(W) <- c("efa_loadings", "LOADINGS")
  W
}

# A non-loading "corr"-role table (variances accounted for) with enough factor columns to
# overflow an 80-column console; values are deterministic so the printed decimals are stable.
make_wide_vars <- function(n = 12) {
  prop <- seq(0.30, 0.02, length.out = n)
  va <- rbind("SS loadings" = seq(3.5, 0.4, length.out = n),
              "Prop Var"    = prop,
              "Cum Var"     = cumsum(prop))
  colnames(va) <- paste0("F", seq_len(n))
  va
}

# A symmetric factor-intercorrelation matrix (Phi) wide enough to overflow an 80-column
# console; only the lower triangle (plus the unit diagonal) is ever shown.
make_wide_phi <- function(n = 12) {
  phi <- matrix(0, n, n)
  phi[lower.tri(phi)] <- seq(0.10, 0.65, length.out = n * (n - 1) / 2)
  phi <- phi + t(phi)
  diag(phi) <- 1
  dimnames(phi) <- list(paste0("F", seq_len(n)), paste0("F", seq_len(n)))
  phi
}

make_sl <- function(heywood = TRUE) {
  f1 <- if (heywood) 1.05 else 0.62
  h2_2 <- if (heywood) 1.10 else 0.70
  u2_2 <- if (heywood) -0.10 else 0.30
  sl <- matrix(c(0.70,  0.20,  0.10, 0.53, 0.47,
                 0.65,  f1,   -0.05, h2_2, u2_2,
                 0.40,  0.10,  0.55, 0.46, 0.54),
               nrow = 3, byrow = TRUE,
               dimnames = list(c("i1", "i2", "i3"),
                               c("g", "F1", "F2", "h2", "u2")))
  class(sl) <- c("efa_sl_loadings", "SLLOADINGS")
  sl
}

test_that("format.efa_loadings aligns decimals and renders a plain table", {
  local_reproducible_output()
  expect_snapshot(print(make_loadings()))
})

test_that("format.efa_loadings prints communalities", {
  local_reproducible_output()
  # local_reproducible_output() turns colours off, so the requested legend is omitted:
  # it only describes styling, and none is rendered here (see the legend test below).
  h2 <- c(0.70, 0.58, 0.63, 1.18)
  expect_snapshot(print(make_loadings(), h2 = h2, legend = TRUE))
})

test_that("format.efa_loadings sorts rows when requested", {
  local_reproducible_output()
  expect_snapshot(print(make_loadings(), sort_loadings = "clustered"))
})

test_that("format.efa_loadings is the source of truth and honours the colour state", {
  L <- make_loadings()

  # print() is cat(format(x), sep = "\n") plus one blank line for console spacing
  expect_identical(utils::capture.output(print(L)), c(format(L), ""))

  withr::local_options(cli.num_colors = 256)
  # styling is controlled by `color`, in format() exactly as in print()
  expect_true(cli::ansi_has_any(paste(
    utils::capture.output(print(L, color = TRUE)), collapse = "")))
  expect_false(cli::ansi_has_any(paste(
    utils::capture.output(print(L, color = FALSE)), collapse = "")))
  expect_true(cli::ansi_has_any(paste(format(L, color = TRUE), collapse = "")))
  expect_false(cli::ansi_has_any(paste(format(L, color = FALSE), collapse = "")))

  # with colours off the table is plain whatever the caller asked for
  withr::local_options(cli.num_colors = 1)
  expect_false(cli::ansi_has_any(paste(format(L, color = TRUE), collapse = "")))

  # format() returns only the rendered table lines, with no trailing blank element
  fmt <- format(L)
  expect_true(nzchar(fmt[length(fmt)]))
})

test_that("the loading legend is printed only when its styling is rendered", {
  L <- make_loadings()
  legend_shown <- function(...) {
    any(grepl("Legend:", utils::capture.output(print(L, legend = TRUE, ...)),
              fixed = TRUE))
  }

  # The legend names bold/grey marks, so it is worth printing only when those marks are
  # actually rendered. That needs both the caller's `color` and a colour-capable target:
  # cli emits no escapes at one colour, where a legend would describe invisible styling.
  withr::with_options(list(cli.num_colors = 256), {
    expect_true(legend_shown(color = TRUE))
    expect_false(legend_shown(color = FALSE))
  })
  withr::with_options(list(cli.num_colors = 1), {
    expect_false(legend_shown(color = TRUE))
    expect_false(legend_shown(color = FALSE))
  })

  # Legend presence tracks ANSI presence exactly.
  withr::with_options(list(cli.num_colors = 256), {
    for (col in c(TRUE, FALSE)) {
      styled <- cli::ansi_has_any(paste(
        utils::capture.output(print(L, legend = TRUE, color = col)), collapse = ""))
      expect_equal(legend_shown(color = col), styled)
    }
  })

  # Suppressing the legend must not leave the spacer blank line behind: with no styling to
  # describe, `legend = TRUE` renders exactly as `legend = FALSE`.
  withr::with_options(list(cli.num_colors = 1), {
    expect_identical(utils::capture.output(print(L, legend = TRUE)),
                     utils::capture.output(print(L, legend = FALSE)))
  })

  # `format()` follows the same rule, since it is what `print()` renders: the legend
  # travels with the marks it names, and is dropped with them when colours are off.
  withr::with_options(list(cli.num_colors = 256), {
    out <- format(L, legend = TRUE)
    expect_true(any(grepl("Legend:", out, fixed = TRUE)))
    expect_true(cli::ansi_has_any(paste(out, collapse = "")))
  })
  withr::with_options(list(cli.num_colors = 1), {
    out <- format(L, legend = TRUE)
    expect_false(any(grepl("Legend:", out, fixed = TRUE)))
    expect_identical(out, format(L, legend = FALSE))
  })
})


test_that("the loading legend names the marks it describes", {
  local_reproducible_output()
  L <- make_loadings()
  withr::local_options(cli.num_colors = 8L)

  # The legend's wording is the contract; the exact escape sequences are cli's business,
  # so the assertion is on the stripped text. Nothing else pins these lines -- the printed
  # snapshot is taken with colours off, where the legend is correctly absent.
  out <- cli::ansi_strip(utils::capture.output(
    print(L, h2 = c(0.70, 0.58, 0.63, 1.18), legend = TRUE, cutoff = .3)))
  i <- which(out == "Legend:")
  expect_length(i, 1L)
  expect_false(nzchar(out[i - 1L]))                       # the spacer above it
  expect_identical(out[i + 1:3],
                   c("  bold = |loading| >= .300",
                     "  grey = below cutoff",
                     "  red h2/u2 = Heywood-relevant value"))
})


test_that("the efa report carries the loading legend through its nested render", {
  local_reproducible_output()

  # print.efa()/summary.efa() render the loading table through a format.efa_loadings()
  # call nested in their own cli::cli_format_method(), i.e. one cli_format_method inside
  # another. A sink makes cli report a single colour, so the legend would
  # be lost -- except that cli_format_method reads the colour count *before* installing its
  # sink and pins it for everything nested inside. Pinning `cli.num_colors` here would not
  # test that: it is the first thing cli consults, so it short-circuits the sink question
  # entirely and the assertion would hold with or without the outer cli_format_method.
  # `cli.default_num_colors` is read on the same side of the sink check as a real terminal's
  # isatty() detection, so it is the faithful stand-in. crayon.enabled and the environment
  # variables have to be cleared because each of them is consulted first.
  withr::local_envvar(c(NO_COLOR = NA, R_CLI_NUM_COLORS = NA))
  withr::local_options(cli.num_colors = NULL, crayon.enabled = NULL,
                       crayon.colors = NULL, knitr.in.progress = NULL,
                       cli.default_num_colors = 256L)
  skip_if_not(cli::num_ansi_colors() > 1L, "no colour-capable stand-in available")

  fit <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                 estimator = "PAF", rotation = "promax")
  has_legend <- function(lines) any(grepl("Legend:", lines, fixed = TRUE))

  expect_true(has_legend(format(fit)))
  expect_true(has_legend(format(summary(fit))))
  expect_false(has_legend(format(fit, show_loading_legend = FALSE)))

  # The same table rendered on its own reaches the same verdict as it does nested in the
  # report: the legend is a property of the console theme, not of the call depth.
  expect_true(has_legend(
    format(fit$rot_loadings, h2 = fit$h2, legend = TRUE)))

  # The guard that makes the assertions above meaningful: nothing renders a legend once
  # the target cannot show the marks it names, whether the table is nested or standalone.
  withr::local_options(cli.num_colors = 1)
  expect_false(has_legend(format(fit)))
  expect_false(has_legend(
    format(fit$rot_loadings, h2 = fit$h2, legend = TRUE)))
})


test_that(".efa_styling_visible tracks both the caller and the target", {
  withr::with_options(list(cli.num_colors = 256), {
    expect_true(.efa_styling_visible(TRUE))
    expect_false(.efa_styling_visible(FALSE))
  })
  withr::with_options(list(cli.num_colors = 1), {
    expect_false(.efa_styling_visible(TRUE))
    expect_false(.efa_styling_visible(FALSE))
  })
})

test_that("wide matrices wrap into stacked blocks via max_factors_per_block", {
  local_reproducible_output()
  expect_snapshot(print(make_wide_loadings(), max_factors_per_block = 3))
})

test_that("wide matrices wrap into stacked blocks on a narrow console", {
  local_reproducible_output()
  withr::local_options(cli.width = 40)
  expect_snapshot(print(make_wide_loadings()))
})

test_that("block wrapping stacks blocks and never folds a row", {
  local_reproducible_output()
  withr::local_options(cli.width = 40)
  out <- format(make_wide_loadings())
  # more than one block: each block carries its own "(block i/n)" label
  expect_gt(sum(grepl("(block ", out, fixed = TRUE)), 1L)
  # no emitted line exceeds the console width -> columns were split, lines not wrapped
  expect_true(all(cli::ansi_nchar(out) <= 40L))
})

test_that("stacked column blocks are separated by a blank line", {
  local_reproducible_output()

  # The renderer puts a blank line before every block after the first, and that separator has
  # to survive into format(): without it the blocks run together and the "(block i/n)" label
  # of one sits directly under the last data row of the previous. Both loading classes, and
  # both routes into a block split (an explicit cap and a narrow console).
  block_gaps <- function(lines) {
    label <- grep("(block ", lines, fixed = TRUE)
    sum(!nzchar(lines[label[-1L] - 1L]))
  }

  wide <- format(make_wide_loadings(), max_factors_per_block = 3)
  expect_equal(sum(grepl("(block ", wide, fixed = TRUE)), 3L)
  expect_equal(block_gaps(wide), 2L)

  narrow <- withr::with_options(list(cli.width = 40), format(make_wide_loadings()))
  expect_gt(sum(grepl("(block ", narrow, fixed = TRUE)), 1L)
  expect_equal(block_gaps(narrow),
               sum(grepl("(block ", narrow, fixed = TRUE)) - 1L)

  sl <- format(make_sl(heywood = FALSE), max_factors_per_block = 1)
  expect_equal(block_gaps(sl), sum(grepl("(block ", sl, fixed = TRUE)) - 1L)

  # the separator is a genuine blank line, not padding, and no table ends on one
  expect_true(all(wide[!nzchar(wide)] == ""))
  expect_true(nzchar(wide[length(wide)]))

  # a table that fits in one block gains no blank lines at all
  expect_false(any(!nzchar(format(make_loadings()))))
})

test_that("wide non-loading tables wrap into stacked blocks", {
  local_reproducible_output() # console width 80
  # a rectangular variances-accounted table and a symmetric Phi (lower triangle only),
  # through .efa_corr_lines(): the wrapper the reports actually use for these tables
  expect_snapshot(writeLines(.efa_corr_lines(make_wide_vars())))
  expect_snapshot(writeLines(.efa_corr_lines(make_wide_phi(), lower_only = TRUE)))
})

test_that("non-loading block wrapping splits columns without folding rows", {
  local_reproducible_output() # console width 80

  corr_lines <- function(values, lower_only = FALSE) {
    .efa_format_matrix(values,
                       row_labels = rownames(values),
                       col_labels = colnames(values),
                       col_roles = rep("corr", ncol(values)),
                       lower_only = lower_only)
  }

  vars <- corr_lines(make_wide_vars())
  # more than one block, and no emitted line exceeds the console width
  expect_gt(sum(grepl("(block ", vars, fixed = TRUE)), 1L)
  expect_true(all(cli::ansi_nchar(vars) <= 80L))

  phi <- corr_lines(make_wide_phi(), lower_only = TRUE)
  expect_gt(sum(grepl("(block ", phi, fixed = TRUE)), 1L)
  expect_true(all(cli::ansi_nchar(phi) <= 80L))
  # a later block's leading rows fall entirely in the blanked upper triangle and are dropped,
  # so no emitted line is a value-less row label (a phantom blank row would print as bare "Fk")
  expect_false(any(grepl("^F[0-9]+ *$", phi)))
})

test_that("truncated row labels never collide", {
  # Items sharing a long prefix (premeditation_1 ... premeditation_11) are the common
  # psychometric case: every one of them truncates to the same ten characters, and the
  # printed table then has rows the reader cannot tell apart.
  nms <- c(paste0("premeditation_", 1:11), paste0("perseverance_", 1:10))
  out <- .shorten_loadings_names(nms, max_length = 10, name_style = "truncate")
  expect_length(unique(out), length(nms))
  # the requested width is still respected
  expect_true(all(nchar(out) <= 10))

  # ... and the rendered table carries those labels, not just the helper
  L <- matrix(0.5, nrow = length(nms), ncol = 2,
              dimnames = list(nms, c("F1", "F2")))
  class(L) <- c("efa_loadings", "LOADINGS")
  labels <- sub("\\s.*$", "", format(L, color = FALSE)[-1L])
  expect_length(unique(labels), length(nms))

  # the same labels appear wherever else the report names a variable, so a row can be
  # matched between the loading table and the sections below it
  expect_identical(.efa_truncate_names(nms, 10), out)
})

test_that("row labels the caller supplied as duplicates are left as they are", {
  # Collisions that were already in the data are not the shortening's to repair; inventing
  # a distinction between two identically named variables would be worse than showing them
  # as what they are.
  nms <- c("premeditation_a", "premeditation_a", "perseverance_x")
  expect_identical(.shorten_loadings_names(nms, max_length = 10, name_style = "truncate"),
                   substr(nms, 1L, 10L))
})

test_that("disambiguation is budgeted into the width, and yields to it only when it must", {
  nms <- paste0("aaaaaaaaaa_", 1:3)

  # wide enough for abbreviation alone
  out <- .shorten_loadings_names(nms, max_length = 4, name_style = "truncate")
  expect_identical(out, c("aa_1", "aa_2", "aa_3"))

  # too narrow for that: the numbering suffix is paid for out of the label, not added to it
  expect_identical(.disambiguate_loadings_names(rep("premeditat", 3L), 10L),
                   c("premeditat", "premedit~1", "premedit~2"))

  # and where the width cannot hold a base character plus its suffix, uniqueness wins
  narrow <- .shorten_loadings_names(nms, max_length = 2, name_style = "truncate")
  expect_length(unique(narrow), 3L)
  expect_true(any(nchar(narrow) > 2))
})

test_that("a row label is a property of the variable, not of the row order", {
  # The numbering that resolves a collision is assigned in vector order, so deriving it from
  # the sorted rows would label the same item differently under different sort_loadings.
  nms <- paste0("Impulsivit\u00e4t_pers\u00f6nlich_", 1:4)
  L <- matrix(c(0.8, 0.1,
                0.2, 0.7,
                0.6, 0.3,
                0.1, 0.9),
              nrow = 4, byrow = TRUE, dimnames = list(nms, c("F1", "F2")))
  class(L) <- c("efa_loadings", "LOADINGS")

  labels_of <- function(...) sub("\\s.*$", "", format(L, color = FALSE, ...)[-1L])

  # all four names cut to the same ten characters, so the labels have to be disambiguated
  unsorted <- labels_of()
  expect_length(unique(unsorted), 4L)

  # "primary" groups the two F1 items ahead of the two F2 items, i.e. rows 1, 3, 2, 4.
  # Sorting must permute the labels, not re-derive them in the new order.
  expect_identical(labels_of(sort_loadings = "primary"), unsorted[c(1L, 3L, 2L, 4L)])
})

test_that(".efa_wrap_chunks packs to the width and never splits a chunk", {
  chunks <- c("Pooling settings:", "target_method = 'first_target',",
              "align_unrotated = 'signed_tucker_congruence',",
              "fit_pool_method = 'D2'.")

  wide <- .efa_wrap_chunks(chunks, width = 200L)
  expect_length(wide, 1L)
  expect_identical(wide, paste(chunks, collapse = " "))

  packed <- .efa_wrap_chunks(chunks, width = 60L)
  expect_gt(length(packed), 1L)
  expect_true(all(nchar(packed) <= 60L))
  # continuation lines carry the exdent, and no chunk is broken
  expect_true(all(startsWith(packed[-1L], "  ")))
  expect_identical(paste(trimws(packed), collapse = " "), paste(chunks, collapse = " "))

  # a chunk wider than the console has no split point that keeps it intact, so it is
  # emitted as it is rather than folded mid-token
  over <- .efa_wrap_chunks(chunks, width = 20L)
  expect_identical(trimws(over), chunks)

  # widths are display widths: a styled chunk packs by what it shows, not by its escapes
  withr::local_options(cli.num_colors = 256)
  styled <- .efa_wrap_chunks(c("a", cli::style_bold("b"), "c"), width = 5L)
  expect_length(styled, 1L)

  expect_identical(.efa_wrap_chunks(character(0)), character(0))
  expect_identical(.efa_wrap_chunks(c("", "only")), "only")
})

test_that("repairing a collision never warns about non-ASCII names", {
  # abbreviate() warns on non-ASCII input, so the automatic repair cannot route through it
  # there: a print method must not raise a warning of its own accord.
  nms <- paste0("Impulsivit\u00e4t_pers\u00f6nlich_", 1:3)
  expect_no_warning(
    out <- .shorten_loadings_names(nms, max_length = 10, name_style = "truncate"))
  expect_length(unique(out), 3L)
  expect_true(all(nchar(out) <= 10))
})

test_that("format.efa_sl_loadings honours the sibling display controls", {
  local_reproducible_output()
  sl <- make_sl(heywood = FALSE)
  rownames(sl) <- c("premeditation_1", "premeditation_2", "perseverance_1")

  # max_name_length + name_style reach the shared shortener instead of being absorbed by ...
  expect_true(any(grepl("^premeditation_1 ", format(sl, name_style = "full"))))
  short <- format(sl, max_name_length = 6)
  expect_true(any(grepl("^prmd_1 ", short)))
  # ... and still without a shared label
  labels <- sub("\\s.*$", "", short[-1L])
  expect_length(unique(labels), 3L)

  # sort_loadings reaches the shared row ordering, and compares the group factors only: on g
  # (the largest loading of all three items) "clustered" would leave the order untouched,
  # while on F1/F2 it lifts the stronger F1 item above the weaker one
  sorted <- sub("\\s.*$", "", format(sl, sort_loadings = "clustered", name_style = "full")[-1L])
  expect_identical(sorted, c("premeditation_2", "premeditation_1", "perseverance_1"))

  # max_factors_per_block splits the table the way it splits an ordinary loading table
  blocks <- format(sl, max_factors_per_block = 1)
  expect_gt(sum(grepl("(block ", blocks, fixed = TRUE)), 1L)

  # the numeric controls are rejected exactly as the sibling rejects them
  expect_error(format(sl, max_name_length = 0),
               class = "efa_print_invalid_max_name_length")
  expect_error(format(sl, max_factors_per_block = 0),
               class = "efa_print_invalid_max_factors_per_block")

  # and none of this disturbs the styling when colours are available
  withr::local_options(cli.num_colors = 256)
  expect_true(cli::ansi_has_any(paste(
    format(sl, sort_loadings = "clustered", max_name_length = 6), collapse = "")))
})

test_that("format.efa_sl_loadings flags a Heywood case", {
  local_reproducible_output()
  expect_snapshot(print(make_sl(heywood = TRUE)))
})

test_that("format.efa_sl_loadings prints cleanly without Heywood cases", {
  local_reproducible_output()
  expect_snapshot(print(make_sl(heywood = FALSE)))
})

test_that("format.efa_sl_loadings counts a communality-only Heywood item once", {
  # h2 > 1 (and u2 < 0) with no single loading > 1 must be flagged exactly once:
  # not missed (as a loading-only count would), nor multiplied across coupled cells.
  sl <- matrix(c(0.70, 0.20, 0.10, 0.53, 0.47,
                 0.80, 0.75, 0.10, 1.20, -0.20,
                 0.40, 0.10, 0.55, 0.46, 0.54),
               nrow = 3, byrow = TRUE,
               dimnames = list(c("i1", "i2", "i3"), c("g", "F1", "F2", "h2", "u2")))
  class(sl) <- c("efa_sl_loadings", "SLLOADINGS")
  out <- cli::ansi_strip(utils::capture.output(print(sl)))
  expect_true(any(grepl("Results contain a Heywood case!", out, fixed = TRUE)))
  expect_false(any(grepl("Heywood cases", out, fixed = TRUE)))
})

test_that("format.efa_sl_loadings is the source of truth and honours the colour state", {
  sl <- make_sl(heywood = TRUE)

  # print() is cat(format(x), sep = "\n") plus one blank line for console spacing
  expect_identical(utils::capture.output(print(sl)), c(format(sl), ""))

  old <- options(cli.num_colors = 256)
  on.exit(options(old), add = TRUE)
  expect_true(any(grepl("\033", format(sl), fixed = TRUE)))
  # `color = FALSE` governs the table cells, not the Heywood alert below them, which
  # cli styles from the active theme -- so it is not an ANSI switch for the whole block.
  # The table is the header row plus one row per item; the alert follows a blank line.
  table_lines <- seq_len(nrow(sl) + 1L)
  expect_false(any(grepl("\033", format(sl, color = FALSE)[table_lines], fixed = TRUE)))

  options(cli.num_colors = 1)
  plain <- format(sl)
  expect_false(any(grepl("\033", plain, fixed = TRUE)))
  # no trailing blank element
  expect_true(nzchar(plain[length(plain)]))
})

test_that("print.efa_loadings argument validators raise classed conditions", {
  L <- make_loadings()

  expect_error(print(L, cutoff = -1), class = "efa_print_invalid_cutoff")
  expect_error(print(L, digits = 1.5), class = "efa_print_invalid_digits")
  expect_error(print(L, max_name_length = 0),
               class = "efa_print_invalid_max_name_length")
  expect_error(print(L, max_factor_name_length = 0),
               class = "efa_print_invalid_max_factor_name_length")
  expect_error(print(L, max_factors_per_block = 0),
               class = "efa_print_invalid_max_factors_per_block")
  expect_error(print(L, color = NA), class = "efa_print_invalid_color")
  expect_error(print(L, legend = NA), class = "efa_print_invalid_legend")
  # h2 with the wrong length
  expect_error(print(L, h2 = c(0.5, 0.5)), class = "efa_print_invalid_h2")

  # x that is not a valid numeric matrix is rejected by the validator
  valid_args <- list(cutoff = .3, digits = 3, max_name_length = 10, h2 = NULL,
                     color = TRUE, max_factor_name_length = NULL,
                     max_factors_per_block = NULL, legend = FALSE)
  expect_error(do.call(.validate_loadings_print_args,
                       c(list(x = "not a matrix"), valid_args)),
               class = "efa_print_invalid_x")
  expect_error(do.call(.validate_loadings_print_args,
                       c(list(x = matrix(numeric(0), 0, 0)), valid_args)),
               class = "efa_print_invalid_x")
})

test_that("the shared argument checks keep their caller's frame and reject huge values", {
  L <- make_loadings()

  # The checks are shared helpers, so each abort has to report the validator that ran it
  # rather than the helper itself -- otherwise every message names the same internal
  # function whatever argument was wrong.
  cnd <- tryCatch(print(L, digits = -1), error = function(e) e)
  expect_s3_class(cnd, "efa_print_invalid_digits")
  expect_match(deparse(conditionCall(cnd))[1], ".validate_loadings_print_args",
               fixed = TRUE)

  # A whole number beyond the integer range must still reach the classed condition: an
  # as.integer() comparison would make it NA and abort with a bare "missing value where
  # TRUE/FALSE needed" instead.
  expect_error(print(L, digits = 1e10), class = "efa_print_invalid_digits")
  expect_error(print(L, max_name_length = 1e10),
               class = "efa_print_invalid_max_name_length")
  expect_error(print(L, max_factors_per_block = 3e9),
               class = "efa_print_invalid_max_factors_per_block")
  expect_error(format(efa_fit(test_models$baseline$cormat, n_factors = 2, N = 500,
                              estimator = "PAF"), digits = 1e10),
               class = "efa_print_invalid_digits")
})

test_that("a named h2 that omits a row name raises a classed condition", {
  # make_loadings() has row names; a named h2 missing one of them must abort in
  # .align_loadings_h2 rather than silently mismatching communalities to rows.
  L <- make_loadings()
  expect_error(
    print(L, h2 = c(fun = .7, friends_long_name = .6, enjoy = .6, wrong = .5)),
    class = "efa_print_invalid_h2"
  )
})

# Superseded functions
# =====================
#
# Home for the UPPERCASE names that have been superseded by lowercase `efa_*`
# equivalents. Each old name is a thin, silent bridge to its `efa_*` implementation:
# no message, no warning, no runtime signal. Most simply forward their arguments; the
# ones whose dots reach a fit repack the old flat tuning knobs into the control objects
# first (see the discipline below). Either way scripts that predate the rename keep
# working. The one deliberate exception is an ABBREVIATED tuning-knob name in a forwarded
# `...` (e.g. `max_it` for `max_iter`, `typ` for `type`): the old wrappers reached EFA()
# through do.call(), where R partial-matched such a name against the formals ahead of
# `...`, and it now aborts with `efa_unused_dots` instead. Silently running a different
# fit is worse than a loud error, so .drop_unknown_frozen_dots() lets it travel on to
# fail (see the comment there).
#
# SUPERSEDED DISCIPLINE (freeze the CONTRACT, not the implementation) -- every
# later wrapper added here MUST follow it, so the `superseded` badge stays honest:
#
#   * Never extend an old name's signature; keep each old argument list frozen
#     forever. New capabilities arrive only as new `efa_*` arguments (or
#     `estimate_control()` / `rotate_control()` knobs), which the frozen list
#     cannot name. Recorded exception: EFA() itself carries the rename of
#     `P_type`/`randomStarts` to `p_type`/`random_starts`, with silent
#     lifecycle::deprecated() catchers keeping every named legacy call working
#     unchanged. The wrappers that forward `...` (PARALLEL, KGC, HULL, SCREE,
#     NEST, N_FACTORS, SL, EFA_POOLED) are the exception: a new named `efa_*` formal IS reachable
#     through their dots. When adding one, decide deliberately whether the old
#     name should reach it, and intercept it here if it should not.
#   * A `...` that reaches a fit must be repacked with `.repack_flat_dots()` before it is
#     forwarded. The old flat interface let a caller tune the fit through these dots
#     (`type`, `max_iter`, `k`, ... -- "additional arguments passed to EFA()"), but those
#     knobs now live in the control objects and no fitting function has a formal for any of
#     them, so a bare knob forwarded on would be dropped and the fit would quietly run the
#     default preset. Repacking keeps the old call tuning the fit exactly as it used to.
#     PARALLEL, KGC, HULL, SCREE, NEST, SL, N_FACTORS and EFA_POOLED therefore translate rather
#     than forward verbatim. SL also repacks its frozen `type` argument, which the successor no
#     longer takes as a bare formal.
#   * A repacking wrapper must carry its OWN `@param ...`; never let `@inheritParams` supply
#     it. The successors tell their readers that the estimation knobs do not go in `...` --
#     true for them, and the exact opposite of what a repacking wrapper does with a legacy
#     flat call. Inheriting it documents the wrapper as breaking the code it exists to keep
#     working. Check this whenever a wrapper's dots reach a fit.
#   * Critical bug fixes in shared internals propagate through the wrapper -- this
#     is allowed for superseded functions and is why the wrapper forwards rather
#     than copies. Consciously admitted under this rule: the estimation control a
#     `N_FACTORS()` call repacks now reaches the NEST and SMT criteria, which silently dropped
#     it before, so a legacy `N_FACTORS(x, criteria = "NEST", type = "psych")` starts honouring
#     the knob it always claimed to pass on.
#   * For any DELIBERATE behaviour change to an existing pathway, consciously decide
#     whether to let it reach the old name or to pin the old behaviour inside the
#     wrapper, and document that decision. Consciously admitted under this rule: the
#     `EKC()` `type` and `N_FACTORS()` `ekc_type` arguments are deprecated on their
#     successors, so a legacy call that names either one now warns and gets a different
#     number of factors. The change is let through to the old name rather than pinned,
#     because the alternative -- returning the new number in silence -- is the one
#     outcome a caller cannot detect. Both formals stay, with the lifecycle sentinel as
#     their default so that forwarding it reads as "not supplied": a legacy call that
#     never named the argument stays silent, and the wrapper body is unchanged. The
#     returned object changes with them: `efa_ekc()` no longer records `type` in
#     `settings`, so `EKC(...)$settings$type` is now NULL rather than "BvA2017".
#     The successor raises the warning, so it names `efa_ekc(type)` even to a caller
#     who wrote `EKC(type)`. Letting the wrapper name itself would mean warning in the
#     wrapper and dropping the argument instead of forwarding it -- a second statement
#     in the body, which the frozen-signature test rejects (it requires the body to be
#     exactly one call that forwards every formal by name). The one-line forwarder is
#     worth more than the sharper label; do not "fix" this without moving that test.

# Reject the successor-only `estimator` name when it lands in a wrapper's `...`: the
# wrappers with a frozen `method` formal splice their own `estimator = method` translation
# into the do.call, so a caller's `estimator` dot would collide as an opaque base
# "matched by multiple actual arguments" error instead of naming the mistake.
.reject_estimator_dot <- function(nms, fn, successor) {
  if (!"estimator" %in% nms) return(invisible(NULL))
  cli::cli_abort(
    c("{.arg estimator} is not an argument of {.fn {fn}}.",
      "i" = "{.fn {fn}} selects the estimator with {.arg method}; {.arg estimator} is the
             {.fn {successor}} argument."),
    class = "efa_renamed_arg"
  )
}

#' Bartlett's test of sphericity
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `BARTLETT()` has been superseded by [efa_bartlett()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_bartlett
#'
#' @returns A list of class `c("efa_bartlett", "BARTLETT")`, identical to the value
#'   of [efa_bartlett()]; see there for the components.
#'
#' @seealso [efa_bartlett()]
#'
#' @export
BARTLETT <- function(x, N = NA, use = c("pairwise.complete.obs", "all.obs",
                                        "complete.obs", "everything",
                                        "na.or.complete"),
                     cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")) {
  efa_bartlett(x, N = N, use = use, cor_method = cor_method)
}

#' Kaiser-Meyer-Olkin criterion
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `KMO()` has been superseded by [efa_kmo()], which is the recommended interface
#' going forward. It remains available and unchanged so existing code keeps working.
#'
#' @inheritParams efa_kmo
#'
#' @returns A list of class `c("efa_kmo", "KMO")`, identical to the value of
#'   [efa_kmo()]; see there for the components.
#'
#' @seealso [efa_kmo()]
#'
#' @export
KMO <- function(x, use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                           "everything", "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")) {
  efa_kmo(x, use = use, cor_method = cor_method)
}

#' Parallel analysis
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `PARALLEL()` has been superseded by [efa_parallel()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_parallel
#' @param ... Further arguments passed on to the [efa_fit()] fits. For example,
#'  `estimator`, to change the estimator (default is "PAF"; PAF is more robust, but it
#'  will take longer compared to "ML" and "ULS"), or one of the estimation tuning knobs
#'  (`type`, `init_comm`, `criterion`, `criterion_type`, `max_iter`, `abs_eigen`,
#'  `start_method`), which are repacked into an [estimate_control()] object so that they
#'  tune the fits exactly as they always did.
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_parallel()]; see there for the components.
#'
#' @seealso [efa_parallel()]
#'
#' @export
PARALLEL <- function(x = NULL,
                     N = NA,
                     n_vars = NA,
                     n_datasets = 1000,
                     percent = 95,
                     eigen_type = c("PCA", "SMC", "EFA"),
                     use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                             "everything", "na.or.complete"),
                     cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                     decision_rule = c("means", "percentile", "crawford"),
                     n_factors = 1,
                     ...) {
  do.call(efa_parallel,
          c(list(x, N = N, n_vars = n_vars, n_datasets = n_datasets,
                 percent = percent, eigen_type = eigen_type, use = use,
                 cor_method = cor_method, decision_rule = decision_rule,
                 n_factors = n_factors),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)))))
}

#' Empirical Kaiser criterion
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `EKC()` has been superseded by [efa_ekc()], which is the recommended interface
#' going forward. It remains available and unchanged so existing code keeps working.
#'
#' @inheritParams efa_ekc
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_ekc()]; see there for the components.
#'
#' @seealso [efa_ekc()]
#'
#' @export
EKC <- function(x, N = NA,
                use = c("pairwise.complete.obs", "all.obs",
                           "complete.obs", "everything",
                           "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                type = lifecycle::deprecated()) {
  efa_ekc(x, N = N, use = use, cor_method = cor_method, type = type)
}

#' Kaiser-Guttman criterion
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `KGC()` has been superseded by [efa_kgc()], which is the recommended interface
#' going forward. It remains available and unchanged so existing code keeps working.
#'
#' @inheritParams efa_kgc
#' @param ... Further arguments passed on to the [efa_fit()] fit. For example,
#'  `estimator`, to change the estimator (PAF is default), or one of the estimation
#'  tuning knobs (`type`, `init_comm`, `criterion`, `criterion_type`, `max_iter`,
#'  `abs_eigen`, `start_method`), which are repacked into an [estimate_control()]
#'  object so that they tune the fit exactly as they always did.
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_kgc()]; see there for the components.
#'
#' @seealso [efa_kgc()]
#'
#' @export
KGC <- function(x, eigen_type = c("PCA", "SMC", "EFA"),
                use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                        "everything", "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                n_factors = 1, ...) {
  do.call(efa_kgc,
          c(list(x, eigen_type = eigen_type, use = use, cor_method = cor_method,
                 n_factors = n_factors),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)))))
}

#' Hull method
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `HULL()` has been superseded by [efa_hull()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_hull
#' @param method character. The estimator to use; passed to [efa_hull()] as its
#'   `estimator` argument. One of `"PAF"`, `"ULS"`, or `"ML"`.
#' @param ... Further arguments passed on to the [efa_fit()] fits, including the
#'  estimation tuning knobs (`type`, `init_comm`, `criterion`, `criterion_type`,
#'  `max_iter`, `abs_eigen`, `start_method`), which are repacked into an
#'  [estimate_control()] object so that they tune the fits exactly as they always did.
#'  The estimator is selected with `method`.
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_hull()]; see there for the components.
#'
#' @seealso [efa_hull()]
#'
#' @export
HULL <- function(x, N = NA, n_fac_theor = NA,
                 method = c("PAF", "ULS", "ML"), gof = c("CAF", "CFI", "RMSEA"),
                 eigen_type = c("SMC", "PCA", "EFA"),
                 use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                         "everything", "na.or.complete"),
                 cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                 n_datasets = 1000, percent = 95,
                 decision_rule = c("means", "percentile", "crawford"),
                 n_factors = 1, ...) {
  .reject_estimator_dot(...names(), "HULL", "efa_hull")
  do.call(efa_hull,
          c(list(x, N = N, n_fac_theor = n_fac_theor, estimator = method, gof = gof,
                 eigen_type = eigen_type, use = use, cor_method = cor_method,
                 n_datasets = n_datasets, percent = percent,
                 decision_rule = decision_rule, n_factors = n_factors),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)))))
}

#' Scree plot
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `SCREE()` has been superseded by [efa_scree()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_scree
#' @param ... Further arguments passed on to the [efa_fit()] fit. For example,
#'  `estimator`, to change the estimator (PAF is default), or one of the estimation
#'  tuning knobs (`type`, `init_comm`, `criterion`, `criterion_type`, `max_iter`,
#'  `abs_eigen`, `start_method`), which are repacked into an [estimate_control()]
#'  object so that they tune the fit exactly as they always did.
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_scree()]; see there for the components.
#'
#' @seealso [efa_scree()]
#'
#' @export
SCREE <- function(x, eigen_type = c("PCA", "SMC", "EFA"),
                  use = c("pairwise.complete.obs", "all.obs", "complete.obs",
                        "everything", "na.or.complete"),
                  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                  n_factors = 1, ...) {
  do.call(efa_scree,
          c(list(x, eigen_type = eigen_type, use = use, cor_method = cor_method,
                 n_factors = n_factors),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)))))
}

#' Minimum average partial
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `MAP()` has been superseded by [efa_map()], which is the recommended interface
#' going forward. It remains available and unchanged so existing code keeps working.
#'
#' @inheritParams efa_map
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_map()]; see there for the components.
#'
#' @seealso [efa_map()]
#'
#' @export
MAP <- function(x,
                use = c("pairwise.complete.obs", "all.obs",
                        "complete.obs", "everything",
                        "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")) {
  efa_map(x, use = use, cor_method = cor_method)
}

#' Next eigenvalue sufficiency test
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `NEST()` has been superseded by [efa_nest()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_nest
#' @param ... Further arguments passed on to the [efa_fit()] fits. For example,
#'  `estimator`, to change the estimator (PAF is default), or one of the estimation
#'  tuning knobs (`type`, `init_comm`, `criterion`, `criterion_type`, `max_iter`,
#'  `abs_eigen`, `start_method`), which are repacked into an [estimate_control()]
#'  object so that they tune the fits exactly as they always did.
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_nest()]; see there for the components.
#'
#' @seealso [efa_nest()]
#'
#' @export
NEST <- function(x, N = NA,
                 alpha = .05,
                 use = c("pairwise.complete.obs", "all.obs",
                         "complete.obs", "everything",
                         "na.or.complete"),
                 cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                 n_datasets = 1000,
                 ...) {
  do.call(efa_nest,
          c(list(x, N = N, alpha = alpha, use = use, cor_method = cor_method,
                 n_datasets = n_datasets),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)))))
}

#' Sequential model tests
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `SMT()` has been superseded by [efa_smt()], which is the recommended interface
#' going forward. It remains available and unchanged so existing code keeps working.
#'
#' @inheritParams efa_smt
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_smt()]; see there for the components.
#'
#' @seealso [efa_smt()]
#'
#' @export
SMT <- function(x, N = NA, use = c("pairwise.complete.obs", "all.obs",
                                   "complete.obs", "everything",
                                   "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")) {
  efa_smt(x, N = N, use = use, cor_method = cor_method)
}

#' Comparison data
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `CD()` has been superseded by [efa_cd()], which is the recommended interface
#' going forward. It remains available and unchanged so existing code keeps working.
#'
#' @inheritParams efa_cd
#'
#' @returns An object of class `efa_retention`, identical to the value of
#'   [efa_cd()]; see there for the components.
#'
#' @seealso [efa_cd()]
#'
#' @export
CD <- function(x, n_factors_max = NA, N_pop = 10000, N_samples = 500, alpha = .30,
               cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
               max_iter = 50) {
  efa_cd(x, n_factors_max = n_factors_max, N_pop = N_pop, N_samples = N_samples,
         alpha = alpha, cor_method = cor_method, max_iter = max_iter)
}

#' Various factor retention criteria
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `N_FACTORS()` has been superseded by [efa_retain()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_retain
#' @param criteria character. A vector with the factor retention methods to
#'   perform. Possible inputs are: `"CD"`, `"EKC"`, `"HULL"`, `"KGC"`, `"MAP"`,
#'   `"NEST"`, `"PARALLEL"`, `"SCREE"`, and `"SMT"` (see the details in
#'   [efa_retain()]). By default, a subset of often used, well-performing
#'   methods are performed.
#' @param method character. The estimator to use in the criteria that fit EFA models;
#'   passed to [efa_retain()] as its `estimator` argument. One of `"ML"`, `"PAF"`, or
#'   `"ULS"`.
#' @param ... Further arguments passed on to the [efa_fit()] fits, including the
#'  estimation tuning knobs (`type`, `init_comm`, `criterion`, `criterion_type`,
#'  `abs_eigen`, `start_method`), which are repacked into an [estimate_control()] object
#'  so that they tune the fits exactly as they always did. The estimator is selected with
#'  `method`; `max_iter` is taken by the `max_iter_CD` argument (R matches an abbreviated
#'  name against the arguments before `...`) and so does not reach the fits.
#'
#' @returns A list of class `c("efa_retain", "N_FACTORS")`, identical to the
#'   value of [efa_retain()]; see there for the components.
#'
#' @seealso [efa_retain()]
#'
#' @export
N_FACTORS <- function(x, criteria = c("CD", "EKC", "HULL", "MAP", "NEST", "PARALLEL"),
                      suitability = TRUE, N = NA,
                      use = c("pairwise.complete.obs", "all.obs",
                              "complete.obs", "everything", "na.or.complete"),
                      cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                      n_factors_max = NA, N_pop = 10000, N_samples = 500,
                      alpha = .30, max_iter_CD = 50, n_fac_theor = NA,
                      method = c("ML", "PAF", "ULS"),
                      gof = c("CAF", "CFI", "RMSEA"),
                      eigen_type_HULL = c("SMC", "PCA", "EFA"),
                      eigen_type_other = c("SMC"),
                      n_factors = 1, n_datasets = 1000,
                      percent = 95,
                      decision_rule = c("means", "percentile", "crawford"),
                      ekc_type = lifecycle::deprecated(),
                      n_datasets_nest = 1000, alpha_nest = .05,
                      show_progress = FALSE,
                      ...) {
  .reject_estimator_dot(...names(), "N_FACTORS", "efa_retain")
  do.call(efa_retain,
          c(list(x, criteria = criteria, suitability = suitability, N = N,
                 use = use, cor_method = cor_method, n_factors_max = n_factors_max,
                 N_pop = N_pop, N_samples = N_samples, alpha = alpha,
                 max_iter_CD = max_iter_CD, n_fac_theor = n_fac_theor,
                 estimator = method, gof = gof, eigen_type_HULL = eigen_type_HULL,
                 eigen_type_other = eigen_type_other, n_factors = n_factors,
                 n_datasets = n_datasets, percent = percent,
                 decision_rule = decision_rule, ekc_type = ekc_type,
                 n_datasets_nest = n_datasets_nest, alpha_nest = alpha_nest,
                 show_progress = show_progress),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)))))
}

#' Schmid-Leiman transformation
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `SL()` has been superseded by [efa_schmid_leiman()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_schmid_leiman
#' @param type character. One of "EFAtools" (default), "psych", "SPSS", or "none". This is
#'  used to control the procedure of the second-order factor analysis. In
#'  [efa_schmid_leiman()] it is set through the `type` of the [estimate_control()] object.
#' @param method character. The estimator for the second-order factor analysis; passed to
#'  [efa_schmid_leiman()] as its `estimator` argument. One of `"PAF"`, `"ML"`, `"ULS"`, or
#'  `"MINRES"`.
#' @param ... Further arguments passed on to the second-order [efa_fit()], including the
#'  estimation tuning knobs (`init_comm`, `criterion`, `criterion_type`, `max_iter`,
#'  `abs_eigen`, `start_method`), which are repacked, together with `type`, into an
#'  [estimate_control()] object so that they tune that fit exactly as they always did.
#'  The estimator is selected with `method`.
#'
#' @returns A list of class `c("efa_schmid_leiman", "SL")`, identical to the value
#'   of [efa_schmid_leiman()]; see there for the components.
#'
#' @seealso [efa_schmid_leiman()]
#'
#' @export
SL <- function(x, Phi = NULL, type = c("EFAtools", "psych", "SPSS", "none"),
               method = c("PAF", "ML", "ULS", "MINRES"), g_name = "g", ...) {
  .reject_estimator_dot(...names(), "SL", "efa_schmid_leiman")
  # The successor-only control objects must not ride through the frozen dots: with a control
  # present, .repack_flat_dots() translates nothing, so the wrapper's own `type` argument --
  # and any flat knob alongside it -- would be silently discarded and the second-order fit
  # would quietly run the control's preset instead of the requested one. No pre-rename code
  # could have passed these names meaningfully, so rejecting them breaks nothing frozen (the
  # same rule EFA() applies).
  new_only <- intersect(...names(), c("estimate_control", "rotate_control"))
  if (length(new_only) > 0L) {
    cli::cli_abort(
      c("{.arg {new_only}} {?is/are} not {?an argument/arguments} of {.fn SL}.",
        "i" = "{.fn SL} takes the tuning knobs flat and selects the preset with
               {.arg type}; the control objects belong to {.fn efa_schmid_leiman}."),
      class = "efa_renamed_arg"
    )
  }
  # Resolve `type` once so the frozen preset argument, and any flat knob the dots carry, reach
  # the second-order fit through the control objects it now takes. The second-order fit is
  # always unrotated, so the frozen filter's retention defaults apply.
  type <- match.arg(type)
  do.call(efa_schmid_leiman,
          c(list(x, Phi = Phi, estimator = method, g_name = g_name),
            .repack_flat_dots(.drop_unknown_frozen_dots(list(...)), type = type)))
}

#' Compare two vectors or matrices (communalities or loadings)
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `COMPARE()` has been superseded by [efa_compare()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_compare
#'
#' @returns A list of class `c("efa_compare", "COMPARE")`, identical to the value
#'   of [efa_compare()]; see there for the components.
#'
#' @seealso [efa_compare()]
#'
#' @export
COMPARE <- function(x,
                    y,
                    reorder = c("congruence", "names", "none"),
                    corres = TRUE,
                    thresh = .3,
                    digits = 4,
                    m_red = .001,
                    range_red = .001,
                    round_red = 3,
                    print_diff = TRUE,
                    na.rm = FALSE,
                    x_labels = c("x", "y"),
                    plot = TRUE,
                    plot_red = .01) {
  efa_compare(x, y = y, reorder = reorder, corres = corres, thresh = thresh,
              digits = digits, m_red = m_red, range_red = range_red,
              round_red = round_red, print_diff = print_diff, na.rm = na.rm,
              x_labels = x_labels, plot = plot, plot_red = plot_red)
}

#' Rotate a loading matrix to a target using Procrustes alignment
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `PROCRUSTES()` has been superseded by [efa_procrustes()], which is the
#' recommended interface going forward. It remains available and unchanged so
#' existing code keeps working.
#'
#' @inheritParams efa_procrustes
#'
#' @returns A list identical to the value of [efa_procrustes()]; see there for the
#'   components.
#'
#' @seealso [efa_procrustes()]
#'
#' @export
PROCRUSTES <- function(A,
                       Target,
                       rotation = c("orthogonal", "oblique"),
                       S = NULL,
                       T_init = NULL,
                       oblique_eps = 1e-5,
                       oblique_maxit = 1000,
                       oblique_max_line_search = 10,
                       oblique_step0 = 1,
                       oblique_normalize = FALSE,
                       oblique_random_starts = 0,
                       oblique_screen_keep = 2,
                       oblique_triage_maxit = 25,
                       oblique_triage_improve_tol = 0) {
  efa_procrustes(A, Target = Target, rotation = rotation, S = S,
                 T_init = T_init, oblique_eps = oblique_eps,
                 oblique_maxit = oblique_maxit,
                 oblique_max_line_search = oblique_max_line_search,
                 oblique_step0 = oblique_step0,
                 oblique_normalize = oblique_normalize,
                 oblique_random_starts = oblique_random_starts,
                 oblique_screen_keep = oblique_screen_keep,
                 oblique_triage_maxit = oblique_triage_maxit,
                 oblique_triage_improve_tol = oblique_triage_improve_tol)
}

# `EFA_AVERAGE` and `efa_average` differ only in case, so the two topics cannot both be
# written to `man/`: file names that differ only in case are not portable. `@rdname` moves
# this wrapper's topic aside; `?EFA_AVERAGE` still resolves through the alias.

#' Model averaging across different EFA methods and types
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `EFA_AVERAGE()` has been superseded by [efa_average()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_average
#'
#' @param method character vector. Any combination of `"PAF"`, `"ML"`, and `"ULS"`,
#'   the estimators to average across; passed to [efa_average()] as its `estimator`
#'   argument. Default is `"PAF"`.
#' @param P_type character vector. Any combination of `"norm"` and `"unnorm"`.
#'   How the promax target matrix P is computed if `"none"` is among the specified
#'   types and `"promax"` or `"oblique"` is among the specified rotations:
#'   `"unnorm"` uses the unnormalized target matrix of Hendrickson and White
#'   (1964), `"norm"` a normalized one. This frozen argument keeps its original
#'   name; [efa_average()] takes the same setting as `p_type`. Default is
#'   `c("norm", "unnorm")`.
#'
#' @returns The value of [efa_average()], normally a list of class
#'   `c("efa_average", "EFA_AVERAGE")`; see there for the components.
#'
#' @seealso [efa_average()]
#'
#' @rdname EFA_AVERAGE-superseded
#' @export
EFA_AVERAGE <- function(x, n_factors, N = NA, method = "PAF", rotation = "promax",
                        type = "none", averaging = c("mean", "median"), trim = 0,
                        salience_threshold = .3,
                        max_iter = 1e4,
                        init_comm = c("smc", "mac", "unity"),
                        criterion = c(1e-3),
                        criterion_type = c("sum", "max_individual"),
                        abs_eigen = c(TRUE),
                        varimax_type = c("svd", "kaiser"),
                        normalize = TRUE,
                        k_promax = 2:4, k_simplimax = ncol(x),
                        P_type = c("norm", "unnorm"), precision = 1e-5,
                        start_method = c("psych", "factanal"),
                        use = c("pairwise.complete.obs", "all.obs",
                                "complete.obs", "everything", "na.or.complete"),
                        cor_method = c("pearson", "spearman", "kendall", "poly",
                                       "tetra", "fiml"),
                        show_progress = TRUE) {
  efa_average(x, n_factors = n_factors, N = N, estimator = method,
              rotation = rotation, type = type, averaging = averaging,
              trim = trim, salience_threshold = salience_threshold,
              max_iter = max_iter, init_comm = init_comm, criterion = criterion,
              criterion_type = criterion_type, abs_eigen = abs_eigen,
              varimax_type = varimax_type, normalize = normalize,
              k_promax = k_promax, k_simplimax = k_simplimax, p_type = P_type,
              precision = precision, start_method = start_method, use = use,
              cor_method = cor_method, show_progress = show_progress)
}

# `EFA()` is a translating wrapper, not a pure forwarder: `efa_fit()` collects the
# estimation and rotation tuning knobs into two control objects, so the frozen flat
# argument list is repacked into `estimate_control()` / `rotate_control()` here. This is
# why `EFA()` is exempt from the static one-line-forwarder contract the plain forwarders meet;
# its behaviour is pinned by a byte-identical `EFA()` == `efa_fit()` regression test instead.

#' Exploratory factor analysis (EFA)
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `EFA()` has been superseded by [efa_fit()], which is the recommended interface going
#' forward. `efa_fit()` keeps the primary choices (data, factors, estimator, rotation,
#' standard errors) as top-level arguments and collects the estimation and rotation tuning
#' knobs into two control objects, [estimate_control()] and [rotate_control()]. `EFA()`
#' remains available and unchanged -- its full flat argument list still works exactly as
#' before -- so existing code keeps running.
#'
#' @inheritParams efa_fit
#' @param method character. The estimator used to fit the EFA; passed to [efa_fit()]
#'  as its `estimator` argument. One of "PAF", "ML", "ULS", "MINRES" (an accepted
#'  alias of "ULS"), or "DWLS"; see the [efa_fit()] documentation for their
#'  properties and data requirements.
#' @param type character. If one of "EFAtools" (default), "psych", or "SPSS" is
#'  used, and the following arguments with default NA are left with
#'  NA, these implementations are executed according to the respective program
#'  ("psych" and "SPSS") or according to the best solution found in Grieder &
#'  Steiner (2022; "EFAtools"). Individual properties can be adapted using one of
#'  the three types and specifying some of the following arguments. If set to
#'  "none" additional arguments must be specified depending on the `method`
#'  and `rotation` used (see details).
#' @param max_iter numeric. The maximum number of iterations to perform after which
#' the iterative PAF procedure is halted with a warning. If `type` is one of
#' "EFAtools", "SPSS", or "psych", this is automatically specified if `max_iter` is
#' left to be `NA`, but can be overridden by entering a number. Default is
#' `NA`.
#' @param init_comm character. The method to estimate the initial communalities
#' in `PAF`. "smc" will use squared multiple correlations, "mac" will use
#' maximum absolute correlations, "unity" will use 1s (see details).
#' Default is `NA`.
#' @param criterion numeric. The convergence criterion used for PAF.
#' If the change in communalities from one iteration to the next is smaller than
#' this criterion the solution is accepted and the procedure ends.
#' Default is `NA`.
#' @param criterion_type character. Type of convergence criterion used for
#' PAF. "max_individual" selects the maximum change in any of the
#' communalities from one iteration to the next and tests it against the
#' specified criterion. This is also used by SPSS. "sum" takes the difference of
#' the sum of all communalities in one iteration and the sum of all communalities
#' in the next iteration and tests this against the criterion. This procedure is
#' used by the [psych::fa()] function. Default is `NA`.
#' @param abs_eigen logical. Which algorithm to use in the PAF
#' iterations. If FALSE, the loadings are computed from the eigenvalues. This is
#' also used by the [psych::fa()] function. If TRUE the
#' loadings are computed with the absolute eigenvalues as done by SPSS.
#' Default is `NA`.
#' @param k numeric. Either the power used for computing the target matrix P in
#' the promax rotation or the number of 'close to zero loadings' for the simplimax
#' rotation. If left to `NA` (default), the value for promax depends on the specified type.
#' For simplimax, `nrow(L)`, where L is the matrix of unrotated loadings,
#' is used by default.
#' @param normalize logical. If `TRUE`, a kaiser normalization is
#' performed before the specified rotation. Default is `TRUE`.
#' @param p_type character. This specifies how the target
#' matrix P is computed in promax rotation. If "unnorm" it will use the
#' unnormalized target matrix as originally done in Hendrickson and White (1964).
#' This is also used in the psych and stats packages. If "norm" it will use the
#' normalized target matrix as used in SPSS. Default is `NA`.
#' @param precision numeric. The tolerance for stopping in the rotation
#' procedure. Default is 10^-5 for all rotation methods.
#' @param varimax_type character. The type of the varimax rotation performed.
#' If "svd", singular value decomposition is used, as [stats::varimax()] does. If
#' "kaiser", the varimax procedure performed in SPSS is used, following the original
#' procedure from Kaiser (1958) (see details). Default is `NA`.
#' @param order_type character. How to order the factors. "eigen" reorders the
#' factors by descending explained variance; "ss_factors" reorders the factors by
#' descending (unweighted) sum of squared factor loadings per factor. Default is `NA`.
#' @param start_method character. How to specify the starting values for the
#' optimization procedure for ML. Default is "psych" which takes the
#' starting values specified in [psych::fa()]. "factanal" takes the
#' starting values specified in the [stats::factanal()] function.
#' @param random_starts numeric. The number of random starts to use in the
#'  rotation to guard against local minima. Default is 100.
#' @param P_type,randomStarts `r lifecycle::badge("superseded")` Former names of
#'  `p_type` and `random_starts`. Still accepted (silently) for backwards
#'  compatibility; please use the new names.
#' @param ... Additional arguments passed to the rotation procedure (e.g., `maxit` for
#'  the maximum number of iterations).
#'
#' @returns The value of [efa_fit()], a list of class `c("efa", "EFA")`; see there for the
#'   components.
#'
#' @seealso [efa_fit()], [estimate_control()], [rotate_control()]
#'
#' @export
EFA <- function(x, n_factors, N = NA, method = c("PAF", "ML", "ULS", "MINRES", "DWLS"),
                rotation = c("none", "varimax", "equamax", "quartimax", "geominT",
                             "bentlerT", "bifactorT", "promax", "oblimin",
                             "quartimin", "simplimax", "bentlerQ", "geominQ",
                             "bifactorQ"),
                se = c("none", "information", "sandwich", "np-boot"),
                type = c("EFAtools", "psych", "SPSS", "none"), max_iter = NA,
                init_comm = NA, criterion = NA, criterion_type = NA,
                abs_eigen = NA, use = c("pairwise.complete.obs", "all.obs",
                                          "complete.obs", "everything",
                                          "na.or.complete"),
                varimax_type = NA,
                k = NA, normalize = TRUE, p_type = NA, precision = 1e-5,
                order_type = NA, start_method = "psych",
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra",
                               "fiml"),
                b_boot = 1000, ci = .95,
                random_starts = 100, seed = NULL,
                P_type = lifecycle::deprecated(),
                randomStarts = lifecycle::deprecated(),
                ...) {

  # Resolve the former argument spellings before translating to the control objects.
  if (lifecycle::is_present(P_type)) p_type <- P_type
  if (lifecycle::is_present(randomStarts)) random_starts <- randomStarts

  # Resolve `type` once so the estimation and rotation controls share the single preset the
  # old flat interface always applied to both. efa_fit() match.args the remaining choice
  # arguments and runs every input guard, so nothing else is resolved here.
  type <- match.arg(type)

  # The flat interface always ignored a `...` argument its rotation engine could not consume;
  # efa_fit() rejects such an argument. Keep the frozen contract by dropping the unconsumable
  # dots -- by the exact names the engines read -- before efa_fit() sees them. With
  # rotation = "none" no engine runs and the accepted set is empty, so the dots are dropped
  # in full and the call stays silent, as it always has been.
  #
  # `rotation` is resolved here so the lookup gets a canonical name, and the frozen
  # match.arg() is deliberate: efa_fit() matches its choices case-insensitively
  # (`.match_arg_ci()`), a successor-only feature, so the wrapper keeps the case-sensitive
  # matching its callers have always had. Forwarding the canonical value below also keeps
  # the successor's wider matching from reaching the frozen argument list.
  rotation <- match.arg(rotation)
  dots <- list(...)
  # The successor-only arguments must not ride through the frozen dots: none of the three is
  # a name a rotation engine reads, so the ignore-unknown-dots contract would silently drop
  # all of them and the fit would quietly run the flat defaults -- `estimator` losing out to
  # the wrapper's own `estimator = method` translation, the control objects to the ones the
  # wrapper builds below. No pre-rename code could have passed these names meaningfully, so
  # rejecting them breaks nothing frozen.
  new_only <- intersect(names(dots), c("estimator", "estimate_control", "rotate_control"))
  if (length(new_only) > 0L) {
    cli::cli_abort(
      c("{.arg {new_only}} {?is/are} not {?an argument/arguments} of {.fn EFA}.",
        "i" = "{.fn EFA} selects the estimator with {.arg method} and takes the tuning
               knobs flat; {.arg estimator} and the control objects belong to
               {.fn efa_fit}."),
      class = "efa_renamed_arg"
    )
  }
  dots <- dots[names(dots) %in% .rotation_dot_extras(rotation)]

  args <- c(list(x = x, N = N, estimator = method, rotation = rotation,
                 se = se, cor_method = cor_method, use = use, b_boot = b_boot, ci = ci,
                 seed = seed,
                 estimate_control = estimate_control(type = type, init_comm = init_comm,
                                                     criterion = criterion,
                                                     criterion_type = criterion_type,
                                                     max_iter = max_iter, abs_eigen = abs_eigen,
                                                     start_method = start_method),
                 rotate_control = rotate_control(type = type, normalize = normalize,
                                                 precision = precision, order_type = order_type,
                                                 varimax_type = varimax_type, p_type = p_type,
                                                 k = k, random_starts = random_starts)),
            dots)
  # `n_factors` has no default; do.call() would force it here and turn a call that omitted it
  # (e.g. `EFA(1:5)`) into an eager missing-argument error before efa_fit()'s input checks
  # run, so it is added only when supplied.
  if (!missing(n_factors)) args$n_factors <- n_factors
  do.call(efa_fit, args)
}

#' Exploratory factor analysis on multiple data imputations
#'
#' @description
#' `r lifecycle::badge("superseded")`
#'
#' `EFA_POOLED()` has been superseded by [efa_mi()], which is the recommended
#' interface going forward. It remains available and unchanged so existing code
#' keeps working.
#'
#' @inheritParams efa_mi
#'
#' @returns The value of [efa_mi()], normally a list of class
#'   `c("efa_mi", "EFA_POOLED", "efa", "EFA")`; see there for the components.
#'
#' @seealso [efa_mi()]
#'
#' @export
EFA_POOLED <- function(data_list,
                       p = 0.05,
                       target_method = c("first_target", "consensus"),
                       align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
                       fit_pool_method = c("D2"),
                       consensus_args = list(),
                       procrustes_args = list(),
                       rmsea_ci_level = .90,
                       rmsr_upper = TRUE,
                       ...) {
  # The per-imputation fits ARE rotated here, so the frozen filter keeps `rotation` and the
  # extras the selected rotation's engine reads -- the same per-rotation contract EFA()
  # froze. The rotation is read off the dots verbatim: pre-rename code always passed the
  # canonical spelling (the flat interface matched case-sensitively), and an invalid value
  # is rejected downstream by efa_mi() itself.
  dots <- list(...)
  rot <- if (is.character(dots$rotation) && length(dots$rotation) == 1L) {
    dots$rotation
  } else {
    "none"
  }
  # `rmsr_upper` stays a formal (the frozen signature) but is not forwarded: it is deprecated
  # on efa_mi(), and passing the frozen default on would fire that deprecation for every
  # legacy call. It never selected anything -- see the argument's documentation -- so dropping
  # it here leaves the result identical.
  do.call(efa_mi,
          c(list(data_list, p = p, target_method = target_method,
                 align_unrotated = align_unrotated, fit_pool_method = fit_pool_method,
                 consensus_args = consensus_args, procrustes_args = procrustes_args,
                 rmsea_ci_level = rmsea_ci_level),
            .repack_flat_dots(.drop_unknown_frozen_dots(
              dots, extras = .rotation_dot_extras(rot), unrotated = FALSE))))
}

vec_s <- c("a" = 1, "b" = 2, "c" = 4)
vec_L <- c("A" = 1, "B" = 2, "C" = 4)

mat_s <- matrix(c(0, 0, 0, 1), ncol = 2)
colnames(mat_s) <- c("a", "b")
mat_L <- matrix(c(0, 0, 0, 1), ncol = 2)
colnames(mat_L) <- c("A", "B")

int <- COMPARE(1:10, 1:10)
dec <- COMPARE(c(.1, .2), c(.1, .1))
matr <- COMPARE(matrix(c(1,1,1,2), ncol = 2), matrix(c(1,1,1,1), ncol = 2))

SPSS_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                type = "SPSS", method = "PAF", rotation = "none")
psych_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                 type = "psych", method = "PAF", rotation = "none")
load <- COMPARE(SPSS_PAF$unrot_loadings, SPSS_PAF$unrot_loadings)
load_ro1 <- COMPARE(SPSS_PAF$unrot_loadings, SPSS_PAF$unrot_loadings,
                    reorder = "names")
load_ro2 <- COMPARE(SPSS_PAF$unrot_loadings, SPSS_PAF$unrot_loadings,
                    reorder = "none")

SPSS_PAF_1 <- EFA(test_models$baseline$cormat, n_factors = 1, N = 500,
                type = "SPSS", method = "PAF", rotation = "none")
psych_PAF_1 <- EFA(test_models$baseline$cormat, n_factors = 1, N = 500,
                 type = "psych", method = "PAF", rotation = "none")
load_F1 <- COMPARE(SPSS_PAF_1$unrot_loadings, psych_PAF_1$unrot_loadings)

test_that("output class and dimensions are correct", {
  expect_is(int, "COMPARE")
  expect_is(dec, "COMPARE")
  expect_is(matr, "COMPARE")
  expect_is(load, "COMPARE")
  expect_is(load_ro1, "COMPARE")
  expect_is(load_ro2, "COMPARE")
  expect_is(load_F1, "COMPARE")

  expect_named(int, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                      "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                      "diff_corres_cross", "g", "settings"))
  expect_named(dec, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                      "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                      "diff_corres_cross", "g", "settings"))
  expect_named(matr, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                      "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                      "diff_corres_cross", "g", "settings"))
  expect_named(load, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                       "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                       "diff_corres_cross", "g", "settings"))
  expect_named(load_ro1, c("diff", "mean_abs_diff", "median_abs_diff",
                           "min_abs_diff", "max_abs_diff", "max_dec", "are_equal",
                           "diff_corres", "diff_corres_cross", "g", "settings"))
  expect_named(load_ro2, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                       "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                       "diff_corres_cross", "g", "settings"))
  expect_named(load_F1, c("diff", "mean_abs_diff", "median_abs_diff", "min_abs_diff",
                       "max_abs_diff", "max_dec", "are_equal", "diff_corres",
                       "diff_corres_cross", "g", "settings"))

  expect_is(int$diff, "integer")
  expect_is(int$mean_abs_diff, "numeric")
  expect_is(int$median_abs_diff, "numeric")
  expect_is(int$min_abs_diff, "integer")
  expect_is(int$max_abs_diff, "integer")
  expect_is(int$max_dec, "numeric")
  expect_is(int$are_equal, "numeric")
  expect_equal(int$diff_corres, NA)
  expect_equal(int$diff_corres_cross, NA)
  expect_is(int$g, "numeric")
  expect_is(int$settings, "list")

  expect_is(dec$diff, "numeric")
  expect_is(dec$mean_abs_diff, "numeric")
  expect_is(dec$median_abs_diff, "numeric")
  expect_is(dec$min_abs_diff, "numeric")
  expect_is(dec$max_abs_diff, "numeric")
  expect_is(dec$max_dec, "integer")
  expect_is(dec$are_equal, "numeric")
  expect_equal(dec$diff_corres, NA)
  expect_equal(dec$diff_corres_cross, NA)
  expect_is(dec$g, "numeric")
  expect_is(dec$settings, "list")

  expect_is(matr$diff, "matrix")
  expect_is(matr$mean_abs_diff, "numeric")
  expect_is(matr$median_abs_diff, "numeric")
  expect_is(matr$min_abs_diff, "numeric")
  expect_is(matr$max_abs_diff, "numeric")
  expect_is(matr$max_dec, "numeric")
  expect_is(matr$are_equal, "numeric")
  expect_is(matr$diff_corres, "integer")
  expect_is(matr$diff_corres_cross, "integer")
  expect_is(matr$g, "numeric")
  expect_is(matr$settings, "list")

  expect_is(load$diff, "matrix")
  expect_is(load$mean_abs_diff, "numeric")
  expect_is(load$median_abs_diff, "numeric")
  expect_is(load$min_abs_diff, "numeric")
  expect_is(load$max_abs_diff, "numeric")
  expect_is(load$max_dec, "integer")
  expect_is(load$are_equal, "numeric")
  expect_is(load$diff_corres, "integer")
  expect_is(load$diff_corres_cross, "integer")
  expect_is(load$g, "numeric")
  expect_is(load$settings, "list")

  expect_is(load_ro1$diff, "matrix")
  expect_is(load_ro1$mean_abs_diff, "numeric")
  expect_is(load_ro1$median_abs_diff, "numeric")
  expect_is(load_ro1$min_abs_diff, "numeric")
  expect_is(load_ro1$max_abs_diff, "numeric")
  expect_is(load_ro1$max_dec, "integer")
  expect_is(load_ro1$are_equal, "numeric")
  expect_is(load_ro1$diff_corres, "integer")
  expect_is(load_ro1$diff_corres_cross, "integer")
  expect_is(load_ro1$g, "numeric")
  expect_is(load_ro1$settings, "list")

  expect_is(load_ro2$diff, "matrix")
  expect_is(load_ro2$mean_abs_diff, "numeric")
  expect_is(load_ro2$median_abs_diff, "numeric")
  expect_is(load_ro2$min_abs_diff, "numeric")
  expect_is(load_ro2$max_abs_diff, "numeric")
  expect_is(load_ro2$max_dec, "integer")
  expect_is(load_ro2$are_equal, "numeric")
  expect_is(load_ro2$diff_corres, "integer")
  expect_is(load_ro2$diff_corres_cross, "integer")
  expect_is(load_ro2$g, "numeric")
  expect_is(load_ro2$settings, "list")

  expect_is(load_F1$diff, "matrix")
  expect_is(load_F1$mean_abs_diff, "numeric")
  expect_is(load_F1$median_abs_diff, "numeric")
  expect_is(load_F1$min_abs_diff, "numeric")
  expect_is(load_F1$max_abs_diff, "numeric")
  expect_is(load_F1$max_dec, "integer")
  expect_is(load_F1$are_equal, "numeric")
  expect_is(load_F1$diff_corres, "numeric")
  expect_is(load_F1$diff_corres_cross, "numeric")
  expect_is(load_F1$g, "numeric")
  expect_is(load_F1$settings, "list")

})

test_that("COMPARE returns the correct values", {
  expect_equal(int$diff, rep(0, 10))
  expect_equal(int$mean_abs_diff, 0)
  expect_equal(int$median_abs_diff, 0)
  expect_equal(int$min_abs_diff, 0)
  expect_equal(int$max_abs_diff, 0)
  expect_equal(int$max_dec, 0)
  expect_equal(int$are_equal, 0)
  expect_equal(int$g, 0)

  expect_equal(dec$diff, c(0, 0.1))
  expect_equal(dec$mean_abs_diff, 0.05)
  expect_equal(dec$median_abs_diff, 0.05)
  expect_equal(dec$min_abs_diff, 0)
  expect_equal(dec$max_abs_diff, .1)
  expect_equal(dec$max_dec, 1)
  expect_equal(dec$are_equal, 0)
  expect_equal(dec$g, 0.0707, tolerance = .01)

  expect_equal(matr$diff, matrix(c(0, 0, 0, 1), ncol = 2))
  expect_equal(matr$mean_abs_diff, 0.25)
  expect_equal(matr$median_abs_diff, 0)
  expect_equal(matr$min_abs_diff, 0)
  expect_equal(matr$max_abs_diff, 1)
  expect_equal(matr$max_dec, 0)
  expect_equal(matr$are_equal, 0)
  expect_equal(matr$g, 0.5, tolerance = .01)
  expect_equal(matr$diff_corres, 1)
  expect_equal(matr$diff_corres_cross, 0)

  expect_equal(load$diff, matrix(rep(0, 54), ncol = 3,
                                 dimnames = list(c(paste0("V", seq_len(18))),
                                                 c(paste0("F", seq_len(3))))))
  expect_equal(load$mean_abs_diff, 0)
  expect_equal(load$median_abs_diff, 0)
  expect_equal(load$min_abs_diff, 0)
  expect_equal(load$max_abs_diff, 0)
  expect_equal(load$max_dec, 17, tolerance = 1)
  expect_equal(load$are_equal, 17, tolerance = 1)
  expect_equal(load$g, 0)
  expect_equal(load$diff_corres, 0)
  expect_equal(load$diff_corres_cross, 0)

  expect_equal(load_ro1$mean_abs_diff, 0)
  expect_equal(load_ro1$median_abs_diff, 0)
  expect_equal(load_ro1$min_abs_diff, 0)
  expect_equal(load_ro1$max_abs_diff, 0)
  expect_equal(load_ro1$max_dec, 17, tolerance = 1)
  expect_equal(load_ro1$are_equal, 17, tolerance = 1)
  expect_equal(load_ro1$g, 0)
  expect_equal(load_ro1$diff_corres, 0)
  expect_equal(load_ro1$diff_corres_cross, 0)

  expect_equal(load_ro2$mean_abs_diff, 0)
  expect_equal(load_ro2$median_abs_diff, 0)
  expect_equal(load_ro2$min_abs_diff, 0)
  expect_equal(load_ro2$max_abs_diff, 0)
  expect_equal(load_ro2$max_dec, 17, tolerance = 1)
  expect_equal(load_ro2$are_equal, 17, tolerance = 1)
  expect_equal(load_ro2$g, 0)
  expect_equal(load_ro2$diff_corres, 0)
  expect_equal(load_ro2$diff_corres_cross, 0)

  expect_equal(round(load_F1$diff, 4), matrix(rep(0, 18), ncol = 1,
                                              dimnames = list(c(paste0("V",
                                                                       seq_len(18))),
                                                              "F1")))
  expect_equal(round(load_F1$mean_abs_diff, 4), 0)
  expect_equal(round(load_F1$median_abs_diff, 4), 0)
  expect_equal(round(load_F1$min_abs_diff, 4), 0)
  expect_equal(round(load_F1$max_abs_diff, 4), 0)
  expect_equal(load_F1$max_dec, 15)
  expect_equal(load_F1$are_equal, 2)
  expect_equal(round(load_F1$g, 4), 0)
  expect_equal(load_F1$diff_corres, 0)
  expect_equal(load_F1$diff_corres_cross, 0)

  expect_equal(COMPARE(vec_s, vec_s[c(3, 1, 2)],
                       reorder = "names")$mean_abs_diff, 0)
  expect_equal(COMPARE(mat_s, mat_s[, c(2, 1)],
                       reorder = "names")$mean_abs_diff, 0)
  expect_equal(COMPARE(psych_PAF$unrot_loadings,
                       psych_PAF$unrot_loadings[, c(3, 1, 2)])$mean_abs_diff, 0)
})


test_that("errors etc. are thrown correctly", {
  expect_error(COMPARE(c(1, 2), 1), " 'x' and 'y' have different lengths Compare only works with identical dimensions.\n")
  expect_error(COMPARE(c(1, 2), c("1", "2")), " 'x' is of class numeric and 'y' is of class character but must be numeric vectors or matrices\n")
  expect_error(COMPARE(c(1, 2), data.frame(x = "1", y = "2")),
               " 'x' is of class numeric and 'y' is of class data.frame but must be numeric vectors or matrices\n")

  expect_error(COMPARE(matrix(c(0, 0, 0, 1), ncol = 2),
                       matrix(c(0, 0, 0, 1), ncol = 1)), " 'x' and 'y' have different dimensions. Can only compare matrices with identical dimensions.\n")

  expect_warning(COMPARE(vec_s, vec_s), " reorder was set to 'congruence', but this only works for matrices. To reorder vectors, set reorder = 'names'. Proceeding without reordering.\n")
  expect_warning(COMPARE(vec_s, vec_L, reorder = "names"),
                 " reorder = 'names' was used but names of x and y were not identical. Results might be inaccurate.\n")
  expect_warning(COMPARE(vec_s, 1:3, reorder = "names"), " reorder was set to 'names' but at least one of 'x' and 'y' was not named. Proceeding without reordering.\n")
  expect_warning(COMPARE(1:3, 1:3, reorder = "names"), " reorder was set to 'names' but at least one of 'x' and 'y' was not named. Proceeding without reordering.\n")

  expect_warning(COMPARE(matrix(c(0, 0, 0, 1), ncol = 2),
                         matrix(c(0, 0, 0, 1), ncol = 2),
                         reorder = "names"), " reorder was set to 'names' but at least one of 'x' and 'y' was not named. Proceeding without reordering.\n")
  expect_warning(COMPARE(mat_s, mat_L, reorder = "names"),
                 " reorder = 'names' was used but colnames of x and y were not identical. Results might be inaccurate.\n")

  expect_error(COMPARE(mat_s, mat_s),
               " Tucker's congruence coefficients contained NAs, cannot reorder columns based on congruence. Try another reordering method.\n")

})

rm(int, dec, matr, SPSS_PAF, psych_PAF, load, load_ro1, load_ro2, SPSS_PAF_1,
   psych_PAF_1, load_F1, vec_s, vec_L, mat_s, mat_L)

# list of EFAs from similar datasets (10 bootstrap samples)
set.seed(42)
efa_list <- lapply(1:10,
                   function (x) {
                     EFA(DOSPERT_raw[sample(1:nrow(DOSPERT_raw), replace = TRUE),],
                         n_factors = 6, method = "ML", rotation = "promax")
                   })

# loadings lists
unrot_loadings <- lapply(efa_list, `[[`, "unrot_loadings")
rot_loadings <- lapply(efa_list, `[[`, "rot_loadings")


test_that("PROCRUSTES matches psych::Procrustes and GPArotation::targetQ outputs", {
  expect_equivalent(
    psych::Procrustes(unrot_loadings[[2]], rot_loadings[[1]])$loadings,
    PROCRUSTES(unrot_loadings[[2]], rot_loadings[[1]])$loadings
    )
  expect_equivalent(
    GPArotation::targetQ(unrot_loadings[[2]], Target = rot_loadings[[1]])$loadings,
    PROCRUSTES(unrot_loadings[[2]], rot_loadings[[1]], rotation = "oblique")$loadings,
    tolerance = 1e-3
  )
})


rm(efa_list, unrot_loadings, rot_loadings)

# A single-factor solution cannot be rotated and has no factor correlations, so
# an oblique rotation request on one factor must degenerate to the no-Phi case on
# every standard-error route (matching a single-fit EFA(), which returns
# Phi = NULL for one factor) rather than crashing on L %*% Phi with Phi = NULL.

cormat <- test_models$baseline$cormat
N_id   <- 500L
m_id   <- 3L
identical_list <- replicate(m_id, cormat, simplify = FALSE)

# A single-fit EFA() with one factor and an oblique rotation: the reference
# behaviour the pooled object must reproduce (no $Phi / $Structure).
single_k1 <- suppressWarnings(suppressMessages(
  EFA(cormat, n_factors = 1L, N = N_id, method = "ML", rotation = "oblimin",
      se = "none")
))


test_that("a single-fit EFA() with one factor and an oblique rotation carries no Phi/Structure", {
  expect_null(single_k1$Phi)
  expect_null(single_k1$Structure)
  expect_false(is.null(single_k1$rot_loadings))
})


test_that("k = 1 oblique pooling does not crash and returns no Phi/Structure (se = none, information)", {
  for (se in c("none", "information")) {
    pooled <- suppressWarnings(suppressMessages(
      EFA_POOLED(identical_list, n_factors = 1L, N = N_id, method = "ML",
                 rotation = "oblimin", se = se)
    ))
    expect_s3_class(pooled, "EFA_POOLED")
    expect_null(pooled$Phi)
    expect_null(pooled$Structure)
    expect_false(is.null(pooled$rot_loadings))
    # Communalities/loadings match the single-fit reference (identical imputations).
    expect_equal(unname(as.numeric(pooled$h2)), unname(as.numeric(single_k1$h2)),
                 tolerance = 1e-6, info = se)
  }
})


test_that("k = 1 oblique pooling fills rotated SE without Phi/Structure SE (se = information)", {
  pooled <- suppressWarnings(suppressMessages(
    EFA_POOLED(identical_list, n_factors = 1L, N = N_id, method = "ML",
               rotation = "oblimin", se = "information")
  ))
  # The rotated-loading SE family is populated (one column), but the factor
  # correlation / structure SE families do not exist for a single factor.
  expect_false(is.null(pooled$SE$rot_loadings))
  expect_identical(dim(as.matrix(pooled$SE$rot_loadings)), c(ncol(cormat), 1L))
  expect_null(pooled$SE$Phi)
  expect_null(pooled$SE$Structure)
  expect_null(pooled$MI$Phi)
})


test_that("k = 1 oblique pooling does not crash on the bootstrap route (se = np-boot)", {
  skip_on_cran()
  set.seed(11)
  raw <- lapply(1:3, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), 200L, replace = TRUE), ]
  })
  pooled <- suppressWarnings(suppressMessages(
    EFA_POOLED(raw, n_factors = 1L, method = "ML", rotation = "oblimin",
               se = "np-boot", b_boot = 20L)
  ))
  expect_s3_class(pooled, "EFA_POOLED")
  expect_null(pooled$Phi)
  expect_null(pooled$Structure)
  expect_false(is.null(pooled$rot_loadings))
  # The bootstrap MI families exist for the loadings, none for Phi/Structure.
  expect_false(is.null(pooled$MI$rot_loadings))
  expect_null(pooled$MI$Phi)
  expect_null(pooled$MI$Structure)
})


test_that("k = 1 oblique pooling does not crash on the sandwich route (MI2S)", {
  skip_on_cran()
  # A few ordinal imputations from a one-factor model.
  mk <- function(seed) {
    set.seed(seed)
    f <- stats::rnorm(300L)
    X <- outer(f, rep(0.6, 6L)) + matrix(stats::rnorm(300L * 6L, sd = 0.85), 300L, 6L)
    d <- as.data.frame(apply(X, 2L, function(col) {
      as.integer(cut(col, c(-Inf, stats::quantile(col, c(.25, .5, .75)), Inf)))
    }))
    names(d) <- paste0("x", seq_len(6L))
    d
  }
  imps <- lapply(1:4, function(i) mk(500L + i))

  pooled <- suppressWarnings(suppressMessages(
    EFA_POOLED(imps, n_factors = 1L, method = "ULS", rotation = "oblimin",
               se = "sandwich", cor_method = "poly")
  ))
  expect_s3_class(pooled, "EFA_POOLED")
  expect_s3_class(pooled$mi_fit, "EFA")
  expect_null(pooled$Phi)
  expect_null(pooled$Structure)
  expect_false(is.null(pooled$rot_loadings))
})

# The tuning arguments `P_type` and `randomStarts` were renamed to the snake_case
# `p_type` and `random_starts`, and the estimator argument `method` was renamed to
# `estimator`. The old tuning spellings must keep working, silently, and the returned
# object must keep its old settings field names (for the estimator, as a `method` entry
# duplicating `estimator`).

cormat <- test_models$baseline$cormat

test_that("EFA() maps the old randomStarts onto random_starts identically", {
  # The rotation random starts draw from the ambient RNG, so fix it identically before
  # each call (passing `seed` to both calls would do the same).
  set.seed(1)
  new <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "oblimin", type = "EFAtools",
        random_starts = 20))
  set.seed(1)
  old <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "oblimin", type = "EFAtools",
        randomStarts = 20))
  expect_equal(new$rot_loadings, old$rot_loadings)
  expect_equal(new$settings$randomStarts, old$settings$randomStarts)
  expect_equal(new$settings$randomStarts, 20)
})

test_that("EFA() maps the old P_type onto p_type identically", {
  new <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "promax", type = "none",
        init_comm = "smc", criterion = 1e-3, criterion_type = "sum",
        max_iter = 300, abs_eigen = TRUE, normalize = TRUE, order_type = "eigen",
        varimax_type = "kaiser", k = 4, p_type = "norm"))
  old <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "promax", type = "none",
        init_comm = "smc", criterion = 1e-3, criterion_type = "sum",
        max_iter = 300, abs_eigen = TRUE, normalize = TRUE, order_type = "eigen",
        varimax_type = "kaiser", k = 4, P_type = "norm"))
  expect_equal(new$rot_loadings, old$rot_loadings)
  expect_equal(new$settings$P_type, old$settings$P_type)
  expect_equal(new$settings$P_type, "norm")
})

test_that("the old spellings add no warning of their own", {
  # Comparing the emitted warnings isolates any deprecation warning from the
  # data-dependent warnings both calls share.
  set.seed(1)
  w_new <- capture_warnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "oblimin", type = "EFAtools",
        random_starts = 15))
  set.seed(1)
  w_old <- capture_warnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "oblimin", type = "EFAtools",
        randomStarts = 15))
  expect_equal(w_old, w_new)
})

test_that("EFA() keeps the old settings field names in its output", {
  m_promax <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "promax", type = "EFAtools"))
  expect_true("P_type" %in% names(m_promax$settings))
  # the estimator keeps its former field name alongside the current one
  expect_identical(m_promax$settings$method, m_promax$settings$estimator)
  expect_identical(m_promax$settings$method, "PAF")

  m_oblq <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, rotation = "oblimin", type = "EFAtools"))
  expect_true("randomStarts" %in% names(m_oblq$settings))
})

test_that("efa_average() and EFA_AVERAGE() accept p_type / P_type identically", {
  # A single-combination grid returns a plain EFA output, keeping the test fast. The
  # frozen wrapper still selects the estimator with `method`; the successor takes
  # `estimator`.
  args <- list(cormat, n_factors = 3, N = 500, rotation = "promax",
               type = "none", init_comm = "smc", criterion = 1e-3,
               criterion_type = "sum", abs_eigen = TRUE, varimax_type = "kaiser",
               normalize = TRUE, k_promax = 4, start_method = "psych")
  args_new <- c(args, list(estimator = "PAF"))
  args_old <- c(args, list(method = "PAF"))

  avg_new <- suppressWarnings(do.call(efa_average, c(args_new, list(p_type = "norm"))))
  avg_old <- suppressWarnings(do.call(efa_average, c(args_new, list(P_type = "norm"))))
  ea_old  <- suppressWarnings(do.call(EFA_AVERAGE, c(args_old, list(P_type = "norm"))))

  expect_equal(avg_new$rot_loadings, avg_old$rot_loadings)
  expect_equal(avg_new$rot_loadings, ea_old$rot_loadings)
  # the output settings keep the old P_type field name.
  expect_true("P_type" %in% names(avg_new$settings))
  expect_equal(avg_new$settings$P_type, "norm")
})

# The superseded UPPERCASE names must be thin, silent bridges to their efa_*
# implementations: identical result, no added condition, and the old class string kept so
# `inherits(x, "OLD")` and old S3 dispatch still resolve.

# The comparisons below use expect_equal() with an explicit tolerance rather than
# expect_identical(). The wrapper contract is that the old name runs the SAME fit, not that a
# fit is reproducible bit for bit: a threaded BLAS (Apple's Accelerate, for one) is free to
# vary its reduction order between two calls, so two invocations of one function on one input
# can differ in the last ulp, and an iterative oblique rotation carries that into the reported
# loadings. waldo still compares S3 classes, names, attributes and structure exactly, so only
# numeric values are given slack -- a wrapper that reached a different fit moves loadings in
# the second or third decimal and still fails loudly. The tolerance is deliberately not
# tightened further: `rotation_diagnostics$criterion_spread` is O(1e-12), and below the
# tolerance all.equal() compares such values absolutely rather than relatively.
fp_tol <- 1e-8

# `EFA()` is the *translating* wrapper of record: efa_fit() collects the estimation and
# rotation tuning knobs into estimate_control() / rotate_control(), so EFA() repacks its frozen
# flat argument list into those objects rather than forwarding it verbatim. It is therefore not
# part of the static `superseded_contract` below; its contract is that the old flat call and
# the new control-object call return the same fit, across several type / method /
# rotation combinations, and that EFA() adds no condition of its own. A shared seed pins the
# random-start rotations so the two calls draw from the same RNG state.
#
# The wrappers whose `...` reaches a fit (PARALLEL, KGC, HULL, SCREE, NEST, SL, N_FACTORS,
# EFA_POOLED) translate for the same reason -- the flat knobs a legacy call passes through
# their dots would otherwise match no efa_fit() formal and be dropped -- so the static contract
# pins only their frozen signature and the knob-forwarding tests at the end of this file pin
# their behaviour.

test_that("EFA() and efa_fit() return the same fit", {
  cormat <- test_models$baseline$cormat
  combos <- expand.grid(
    estimator = c("PAF", "ML", "ULS"),
    rotation = c("none", "varimax", "promax", "oblimin"),
    type = c("EFAtools", "psych", "SPSS"),
    stringsAsFactors = FALSE
  )
  for (i in seq_len(nrow(combos))) {
    m <- combos$estimator[i]
    r <- combos$rotation[i]
    ty <- combos$type[i]
    set.seed(42L)
    old <- suppressWarnings(EFA(cormat, n_factors = 3, N = 500, method = m,
                                rotation = r, type = ty))
    set.seed(42L)
    new <- suppressWarnings(efa_fit(cormat, n_factors = 3, N = 500, estimator = m,
                                    rotation = r,
                                    estimate_control = estimate_control(type = ty),
                                    rotate_control = rotate_control(type = ty)))
    expect_equal(old, new, tolerance = fp_tol,
                 info = paste("estimator =", m, "; rotation =", r, "; type =", ty))
  }
})

test_that("efa_fit() carries the new classes and its methods dispatch", {
  mod <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                 rotation = "promax")

  expect_identical(class(mod), c("efa", "EFA"))
  expect_identical(class(mod$unrot_loadings), c("efa_loadings", "LOADINGS"))
  expect_identical(class(mod$rot_loadings), c("efa_loadings", "LOADINGS"))
  expect_true(inherits(mod, "EFA"))

  # the renamed S3 methods resolve on the new class
  expect_s3_class(summary(mod), "summary.efa")
  expect_type(residuals(mod), "double")
  expect_no_error(invisible(format(mod)))
})

test_that("efa_fit() rejects a former flat knob forwarded through `...`, keeps rotation extras", {
  cm <- test_models$baseline$cormat

  # A knob that moved into the control objects (`k`, `type`, `max_iter`, ...) is no longer an
  # efa_fit() formal. Left to reach `.efa_core()` -- which still declares those names -- it
  # would supply the formal twice and abort with "matched by multiple actual arguments"; left
  # to be dropped, it would quietly run the default preset instead of the requested one.
  # Neither is acceptable, so it is rejected up front with a classed error that names the
  # control object owning the knob.
  expect_error(efa_fit(cm, n_factors = 3, N = 500, rotation = "promax", k = 4),
               class = "efa_flat_knob_in_dots")

  # a caller that forwards its user `...` (here efa_group) surfaces the same error rather
  # than silently ignoring the knob
  expect_error(
    suppressWarnings(efa_group(list(a = cm, b = cm), n_factors = 2, type = "SPSS")),
    class = "efa_flat_knob_in_dots"
  )
  # and the control objects ride through those same dots to reach the fit
  expect_s3_class(
    suppressWarnings(efa_group(list(a = cm, b = cm), n_factors = 2,
                               estimate_control = estimate_control(type = "SPSS"),
                               rotate_control = rotate_control(type = "SPSS"))),
    "efa_group"
  )

  # Genuine rotation extras still reach the rotation engine, via `...` or via rotate_control().
  # geominQ screens a hundred random starts drawn from the ambient stream, so the two routes
  # are seeded alike: unseeded they would search different candidate sets and could land in
  # different minima, which says nothing about whether the extra arrived.
  set.seed(1)
  via_dots <- efa_fit(cm, n_factors = 3, N = 500, rotation = "geominQ", maxit = 3000)
  set.seed(1)
  via_ctrl <- efa_fit(cm, n_factors = 3, N = 500, rotation = "geominQ",
                      rotate_control = rotate_control(maxit = 3000))
  expect_equal(via_dots$rot_loadings, via_ctrl$rot_loadings, tolerance = fp_tol)
})

test_that("EFA() keeps silently ignoring a rotation extra its engine cannot consume", {
  cm <- test_models$baseline$cormat

  # the flat interface froze the silent-ignore contract: promax consumes no engine extras,
  # so the call that is an error in efa_fit() still runs through EFA() -- and runs
  # identically to the call without the extra
  expect_error(efa_fit(cm, n_factors = 3, N = 500, rotation = "promax", maxit = 100),
               class = "efa_unused_dots")
  set.seed(1)
  with_extra <- EFA(cm, n_factors = 3, N = 500, rotation = "promax", maxit = 100)
  set.seed(1)
  without_extra <- EFA(cm, n_factors = 3, N = 500, rotation = "promax")
  expect_equal(with_extra$rot_loadings, without_extra$rot_loadings, tolerance = fp_tol)

  # a consumable extra still reaches the engine through EFA()'s dots
  set.seed(1)
  old <- EFA(cm, n_factors = 3, N = 500, rotation = "oblimin", maxit = 750)
  set.seed(1)
  new <- efa_fit(cm, n_factors = 3, N = 500, rotation = "oblimin", maxit = 750)
  expect_equal(old$rot_loadings, new$rot_loadings, tolerance = fp_tol)

  # rotation = "none" -- EFA()'s own default -- runs no engine, so the flat interface
  # consumed no extra there either and simply ignored it; the extra is still dropped
  # silently, and the fit is the one the call without it returns
  expect_no_condition(with_bogus <- EFA(cm, n_factors = 3, N = 500, bogus = 1))
  expect_equal(with_bogus, EFA(cm, n_factors = 3, N = 500), tolerance = fp_tol)
  expect_no_condition(EFA(cm, n_factors = 3, N = 500, rotation = "none", maxit = 500))
})

test_that("the retention wrappers keep silently ignoring an unknown dot", {
  cm <- test_models$baseline$cormat

  # the flat interface always dropped a `...` name its fits could never consume; the
  # successor functions reject such a name (see test-retention-controls.R), the frozen
  # names must not -- and the result is the one the call without the extra returns
  expect_no_condition(kgc_bogus <- KGC(cm, eigen_type = "PCA", bogus = 1))
  expect_equal(kgc_bogus, KGC(cm, eigen_type = "PCA"), tolerance = fp_tol)
  expect_no_condition(SCREE(cm, eigen_type = "PCA", cor_methd = "poly"))
  set.seed(1)
  par_bogus <- PARALLEL(N = 500, n_vars = 18, n_datasets = 2, eigen_type = "PCA",
                        bogus = 1)
  set.seed(1)
  par_plain <- PARALLEL(N = 500, n_vars = 18, n_datasets = 2, eigen_type = "PCA")
  expect_equal(par_bogus, par_plain, tolerance = fp_tol)
  expect_no_condition(
    nf_bogus <- N_FACTORS(cm, criteria = "EKC", N = 500, suitability = FALSE,
                          bogus = 1))
  expect_equal(nf_bogus,
               N_FACTORS(cm, criteria = "EKC", N = 500, suitability = FALSE),
               tolerance = fp_tol)

  # a rotation setting and the rotation-engine extras (maxit/gam/delta) were equally inert
  # on the flat interface -- the criterion fits are always unrotated, so EFA() consumed
  # neither -- and keep being dropped silently, where the successors refuse or reject them.
  # efa_fit()'s inference arguments (se/b_boot/ci/seed) are the one DELIBERATE exception to
  # this rule rather than another instance of it: the flat EFA() took all four, so they are
  # not junk, but they were inert on this path (it computed a standard error the criterion
  # discarded, and never reached its seeding branch at all for a correlation matrix) and are
  # now refused at the old names too -- see .drop_unknown_frozen_dots() and
  # .reject_inference_dots() in R/control.R
  expect_no_condition(N_FACTORS(cm, criteria = "EKC", N = 500, suitability = FALSE,
                                rotation = "promax"))
  expect_no_condition(N_FACTORS(cm, criteria = "EKC", N = 500, suitability = FALSE,
                                maxit = 500))
  # the drop also holds when a criterion fit actually runs (efa_fit() would hard-reject
  # `maxit` under its unrotated per-rotation whitelist)
  kgc_extra <- suppressWarnings(suppressMessages(
    KGC(cm, eigen_type = "EFA", maxit = 10)))
  kgc_noextra <- suppressWarnings(suppressMessages(KGC(cm, eigen_type = "EFA")))
  expect_equal(kgc_extra, kgc_noextra, tolerance = fp_tol)

  # a junk value under a successor-only name is NOT dropped: no pre-rename code could have
  # used the name, so the successor's loud classed validation is the right outcome (the
  # same rule EFA() applies to the successor-only names)
  expect_error(N_FACTORS(cm, criteria = "EKC", N = 500, suitability = FALSE,
                         rotate_control = "SPSS"),
               class = "efa_unused_dots")

  # An ABBREVIATED name is not junk: the flat interface reached its target through
  # do.call(), where R partial-matched it against the formals ahead of `...`, so dropping
  # it would silently run a different fit. It travels on and fails loudly instead.
  expect_error(KGC(cm, eigen_type = "EFA", meth = "ML"), class = "efa_unused_dots")
  expect_error(N_FACTORS(cm, criteria = "EKC", N = 500, suitability = FALSE,
                         estimat = "ML"),
               class = "efa_unused_dots")
})

test_that("HULL(), NEST(), and EFA_POOLED() keep silently ignoring an unknown dot", {
  cm <- test_models$baseline$cormat

  # both forward their dots into fits that reject unknown names (efa_hull by its own new
  # guard, efa_nest through efa_fit's), so the frozen filter has to drop the junk first
  set.seed(1)
  hull_bogus <- suppressWarnings(suppressMessages(
    HULL(cm, N = 500, n_datasets = 2, bogus = 1)))
  set.seed(1)
  hull_plain <- suppressWarnings(suppressMessages(
    HULL(cm, N = 500, n_datasets = 2)))
  expect_equal(hull_bogus, hull_plain, tolerance = fp_tol)

  # method = "ULS" rides through the repack alongside the dropped junk name (with the PAF
  # default and so few reference datasets the NEST walk runs deep enough to Heywood)
  set.seed(1)
  nest_bogus <- suppressWarnings(suppressMessages(
    NEST(cm, N = 500, n_datasets = 50, method = "ULS", bogus = 1)))
  set.seed(1)
  nest_plain <- suppressWarnings(suppressMessages(
    NEST(cm, N = 500, n_datasets = 50, method = "ULS")))
  expect_equal(nest_bogus, nest_plain, tolerance = fp_tol)

  # EFA_POOLED()'s fits are rotated, so its filter is rotation-aware like EFA()'s: a junk
  # name is dropped, and an extra its selected rotation consumes still reaches the engine
  cm_list <- list(cm, cm)
  pooled_bogus <- suppressWarnings(suppressMessages(
    EFA_POOLED(cm_list, n_factors = 3, N = 500, bogus = 1)))
  pooled_plain <- suppressWarnings(suppressMessages(
    EFA_POOLED(cm_list, n_factors = 3, N = 500)))
  expect_equal(pooled_bogus, pooled_plain, tolerance = fp_tol)
})

test_that("SL() keeps silently ignoring an unknown dot and rejects the control objects", {
  cm <- test_models$baseline$cormat
  efa_prom <- suppressWarnings(suppressMessages(
    efa_fit(cm, n_factors = 3, N = 500, rotation = "promax")))

  # the flat interface's second-order EFA silently ignored an unconsumable dot
  sl_bogus <- suppressWarnings(suppressMessages(SL(efa_prom, bogus = 1)))
  sl_plain <- suppressWarnings(suppressMessages(SL(efa_prom)))
  expect_equal(sl_bogus, sl_plain, tolerance = fp_tol)

  # a control object alongside the frozen `type` would silently win over it (the repack
  # translates nothing when a control is present), so the successor-only names are
  # rejected outright, as EFA() rejects them
  expect_error(SL(efa_prom, type = "psych", estimate_control = estimate_control()),
               class = "efa_renamed_arg")
  expect_error(SL(efa_prom, rotate_control = rotate_control()),
               class = "efa_renamed_arg")
})

test_that("EFA() stays transparent to efa_fit()'s conditions", {
  # no condition of its own on a clean correlation-matrix path
  expect_no_condition(EFA(test_models$baseline$cormat, n_factors = 3, N = 500))

  # it neither swallows nor adds to efa_fit()'s conditions: the raw-data path still
  # signals that the correlations were computed from the data
  expect_message(EFA(GRiPS_raw, n_factors = 1), class = "efa_cor_from_data")
})

test_that("BARTLETT() forwards to efa_bartlett() identically", {
  old <- BARTLETT(test_models$baseline$cormat, N = 500)
  new <- efa_bartlett(test_models$baseline$cormat, N = 500)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_bartlett", "BARTLETT"))
  expect_true(inherits(old, "BARTLETT"))
})

test_that("BARTLETT() adds no condition and stays transparent", {
  # no deprecation signal (or any other condition) on a clean path
  expect_no_condition(BARTLETT(test_models$baseline$cormat, N = 500))

  # but it neither swallows nor adds to the conditions efa_bartlett() itself
  # raises: the raw-data path still signals that the correlations were computed
  expect_message(BARTLETT(GRiPS_raw), class = "efa_cor_from_data")
})

test_that("KMO() forwards to efa_kmo() identically", {
  old <- KMO(test_models$baseline$cormat)
  new <- efa_kmo(test_models$baseline$cormat)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_kmo", "KMO"))
  expect_true(inherits(old, "KMO"))
})

test_that("KMO() adds no condition and stays transparent", {
  expect_no_condition(KMO(test_models$baseline$cormat))
  expect_message(KMO(GRiPS_raw), class = "efa_cor_from_data")
})

# The retention criteria share the lowercase `efa_retention` class and gain no
# extra class string on the old name, so the wrapper must return the same
# object with that single class and raise no condition of its own. EKC and MAP
# stand in for the family.

test_that("EKC() forwards to efa_ekc() identically", {
  old <- EKC(test_models$baseline$cormat, N = 500)
  new <- efa_ekc(test_models$baseline$cormat, N = 500)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), "efa_retention")
})

test_that("EKC() adds no condition and stays transparent", {
  expect_no_condition(EKC(test_models$baseline$cormat, N = 500))
  expect_message(EKC(GRiPS_raw), class = "efa_cor_from_data")
})

test_that("MAP() forwards to efa_map() identically", {
  old <- MAP(test_models$baseline$cormat)
  new <- efa_map(test_models$baseline$cormat)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), "efa_retention")
})

test_that("MAP() adds no condition and stays transparent", {
  expect_no_condition(MAP(test_models$baseline$cormat))
  expect_message(MAP(GRiPS_raw), class = "efa_cor_from_data")
})

# The retention orchestrator keeps the dual class its efa_retain() successor
# carries (so both `inherits(x, "N_FACTORS")` and the new S3 dispatch resolve),
# forwards the same result, and raises no condition of its own. EKC and MAP
# stand in for the criteria it runs.

test_that("N_FACTORS() forwards to efa_retain() identically", {
  old <- N_FACTORS(test_models$baseline$cormat, N = 500, suitability = FALSE,
                   criteria = c("EKC", "MAP"))
  new <- efa_retain(test_models$baseline$cormat, N = 500, suitability = FALSE,
                    criteria = c("EKC", "MAP"))

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_retain", "N_FACTORS"))
  expect_true(inherits(old, "N_FACTORS"))
})

test_that("N_FACTORS() adds no condition and stays transparent", {
  expect_no_condition(N_FACTORS(test_models$baseline$cormat, N = 500,
                                suitability = FALSE, criteria = c("EKC", "MAP")))
})

# The Schmid-Leiman transformation keeps the dual class on the returned list and
# on the nested loading matrix, so `inherits(x, "SL")`, `inherits(x$sl,
# "SLLOADINGS")`, and the OMEGA() / efa_reliability() input branches keyed on them
# all still resolve. An oblique EFA is the input every Schmid-Leiman path needs.

# Every path below is handed the same fit: promax rotates outside the random-start engines,
# so this is deterministic and no site that reads it can be told apart from one that fitted
# its own copy.
oblique_efa <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
                   type = "EFAtools", method = "PAF", rotation = "promax")

test_that("SL() forwards to efa_schmid_leiman() identically", {
  efa_mod <- oblique_efa
  old <- SL(efa_mod, type = "EFAtools", method = "PAF")
  new <- efa_schmid_leiman(efa_mod, estimator = "PAF",
                           estimate_control = estimate_control(type = "EFAtools"))

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_schmid_leiman", "SL"))
  expect_identical(class(old$sl), c("efa_sl_loadings", "SLLOADINGS"))
})

test_that("SL() adds no condition and stays transparent", {
  efa_mod <- oblique_efa

  expect_no_condition(SL(efa_mod, type = "EFAtools", method = "PAF"))

  # it neither swallows nor adds to the conditions efa_schmid_leiman() raises:
  # supplying Phi alongside an EFA object still warns
  expect_warning(SL(efa_mod, type = "EFAtools", method = "PAF", Phi = efa_mod$Phi),
                 class = "efa_sl_phi_specified")
})

# OMEGA() and FACTOR_SCORES() carry the superseded badge but keep their own
# implementation, output shape, and class. The supersession itself must add no runtime
# signal -- the statistical diagnostics they share with their successors through the
# reliability and scoring engines are pinned in test-OMEGA.R and test-FACTOR_SCORES.R.

test_that("OMEGA() adds no condition and keeps its class", {
  sl_mod <- efa_schmid_leiman(oblique_efa, estimator = "PAF",
                              estimate_control = estimate_control(type = "EFAtools"))
  fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2

  expect_no_condition(om <- OMEGA(sl_mod, type = "EFAtools", factor_corres = fc))
  expect_s3_class(om, "OMEGA")
})

test_that("FACTOR_SCORES() adds no condition and keeps its class", {
  efa_mod <- oblique_efa

  # the correlation-matrix path signals that only weights can be returned; the
  # wrapper neither swallows that message nor adds one of its own
  expect_message(FACTOR_SCORES(test_models$baseline$cormat, f = efa_mod),
                 class = "efa_scores_needs_raw")

  fs <- suppressMessages(FACTOR_SCORES(test_models$baseline$cormat, f = efa_mod))
  expect_s3_class(fs, "FACTOR_SCORES")
})

# The comparison helper keeps the dual class its efa_compare() successor carries, so
# `inherits(x, "COMPARE")` and the old print / format / plot dispatch all still resolve.

test_that("COMPARE() forwards to efa_compare() identically", {
  old <- COMPARE(matrix(c(1, 1, 1, 2), ncol = 2), matrix(c(1, 1, 1, 1), ncol = 2))
  new <- efa_compare(matrix(c(1, 1, 1, 2), ncol = 2), matrix(c(1, 1, 1, 1), ncol = 2))

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_compare", "COMPARE"))
  expect_true(inherits(old, "COMPARE"))
})

test_that("COMPARE() adds no condition and stays transparent", {
  expect_no_condition(COMPARE(matrix(c(1, 1, 1, 2), ncol = 2),
                              matrix(c(1, 1, 1, 1), ncol = 2)))

  # it neither swallows nor adds to the conditions efa_compare() itself raises: a
  # vector under the default congruence reordering still warns that it cannot be
  # reordered, named or not
  expect_warning(COMPARE(c(a = 1, b = 2, c = 4), c(a = 1, b = 2, c = 4)),
                 class = "efa_compare_reorder_vectors")
  expect_warning(COMPARE(1:10, 1:10), class = "efa_compare_reorder_vectors")
})

# The Procrustes alignment returns a plain, unclassed list, so the old name has no class
# string to preserve -- only the frozen signature and the silent pass-through.

# A fixed loading matrix and a quarter-turn rotation of it. The default orthogonal solver
# recovers that rotation exactly whatever the loadings are, so the two tests below need no
# random draw (and leave the RNG stream untouched for whatever runs next).
procrustes_A <- matrix(c(0.8, 0.7, 0.6, 0.1, 0.2, 0.1,
                         0.1, 0.2, 0.1, 0.8, 0.7, 0.6), ncol = 2)
procrustes_target <- procrustes_A %*% matrix(c(0, 1, -1, 0), ncol = 2)

test_that("PROCRUSTES() forwards to efa_procrustes() identically", {
  old <- PROCRUSTES(procrustes_A, Target = procrustes_target)
  new <- efa_procrustes(procrustes_A, Target = procrustes_target)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), "list")
})

test_that("PROCRUSTES() adds no condition and stays transparent", {
  expect_no_condition(PROCRUSTES(procrustes_A, Target = procrustes_target))

  # the dimension check efa_procrustes() performs still reaches the caller
  expect_error(PROCRUSTES(procrustes_A, Target = procrustes_target[, 1, drop = FALSE]),
               class = "efa_dim_mismatch")
})

# Model averaging keeps the dual class its efa_average() successor carries, so
# `inherits(x, "EFA_AVERAGE")` and the old print / format / plot dispatch all still
# resolve. Two grid cells (PAF and ML, both the EFAtools implementation) are enough to
# average over.

# The frozen wrapper still selects the estimators with `method`; the successor takes
# `estimator`, so the helper places the shared grid under the caller's argument name.
average_grid <- function(f, estimator_arg = "estimator") {
  args <- list(test_models$baseline$cormat, n_factors = 3, N = 500,
               type = "EFAtools", start_method = "psych", show_progress = FALSE)
  args[[estimator_arg]] <- c("PAF", "ML")
  do.call(f, args)
}

# Averaging a grid is the most expensive wrapper check in this file, so the silence
# assertion rides along with the identity one rather than fitting the grid a third time.
test_that("EFA_AVERAGE() forwards identically without user-facing conditions", {
  expect_no_warning(
    expect_no_message(
      old <- average_grid(EFA_AVERAGE, estimator_arg = "method")
    )
  )
  new <- average_grid(efa_average)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_average", "EFA_AVERAGE"))
  expect_true(inherits(old, "EFA_AVERAGE"))
})

test_that("EFA_AVERAGE() stays transparent to efa_average()'s conditions", {
  # it neither swallows nor adds to the conditions efa_average() itself raises: a grid
  # that collapses to a single EFA still warns and returns that EFA
  expect_warning(EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                             method = "PAF", type = "EFAtools", show_progress = FALSE),
                 class = "efa_avg_single_combination")
})

# EFA_POOLED() keeps the dual class its efa_mi() successor carries (so both
# `inherits(x, "EFA_POOLED")` and the new S3 dispatch resolve), forwards
# faithfully, and raises no condition of its own. A two-imputation list of
# identical correlation matrices needs no resampling here (se = "none").

test_that("EFA_POOLED() forwards to efa_mi() identically and adds no condition", {
  imps <- list(test_models$baseline$cormat, test_models$baseline$cormat)

  expect_no_condition(
    old <- EFA_POOLED(imps, n_factors = 3, N = 500, method = "PAF", rotation = "promax")
  )
  new <- efa_mi(imps, n_factors = 3, N = 500, estimator = "PAF", rotation = "promax")

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), c("efa_mi", "EFA_POOLED", "efa", "EFA"))
  expect_true(inherits(old, "EFA_POOLED"))
})

# Each old name is frozen: its argument list may never change. That half is checked
# statically for all seventeen names below. `OMEGA()` and `FACTOR_SCORES()` are superseded
# too, but they reshape their successor's output rather than forwarding to it, so -- like
# `EFA()` -- they are pinned behaviourally (`test-OMEGA.R`, `test-FACTOR_SCORES.R`) instead
# of statically, which is why twenty exported uppercase names yield seventeen table entries.
#
# A name marked `translating` also repacks the old flat tuning knobs its `...` may carry into
# the control objects the fit now takes (`.repack_flat_dots()`), so it is not a one-line
# forwarder and only its signature is pinned here; that its knobs still reach the fit is
# pinned behaviourally further down. Every other name forwards each of its arguments,
# unchanged, to the identically named parameter of its efa_* successor -- checked statically on
# the wrapper's own body, so a dropped, renamed, or mis-mapped argument is caught even for the
# simulation-heavy criteria that are too expensive to run twice in an identity test.
superseded_contract <- list(
  BARTLETT = list(
    callee = "efa_bartlett",
    formals = c("x", "N", "use", "cor_method")),
  KMO = list(
    callee = "efa_kmo",
    formals = c("x", "use", "cor_method")),
  PARALLEL = list(
    callee = "efa_parallel",
    translating = TRUE,
    formals = c("x", "N", "n_vars", "n_datasets", "percent", "eigen_type", "use",
                "cor_method", "decision_rule", "n_factors", "...")),
  EKC = list(
    callee = "efa_ekc",
    formals = c("x", "N", "use", "cor_method", "type")),
  KGC = list(
    callee = "efa_kgc",
    translating = TRUE,
    formals = c("x", "eigen_type", "use", "cor_method", "n_factors", "...")),
  HULL = list(
    callee = "efa_hull",
    translating = TRUE,
    formals = c("x", "N", "n_fac_theor", "method", "gof", "eigen_type", "use",
                "cor_method", "n_datasets", "percent", "decision_rule",
                "n_factors", "...")),
  SCREE = list(
    callee = "efa_scree",
    translating = TRUE,
    formals = c("x", "eigen_type", "use", "cor_method", "n_factors", "...")),
  MAP = list(
    callee = "efa_map",
    formals = c("x", "use", "cor_method")),
  NEST = list(
    callee = "efa_nest",
    translating = TRUE,
    formals = c("x", "N", "alpha", "use", "cor_method", "n_datasets", "...")),
  SMT = list(
    callee = "efa_smt",
    formals = c("x", "N", "use", "cor_method")),
  CD = list(
    callee = "efa_cd",
    formals = c("x", "n_factors_max", "N_pop", "N_samples", "alpha", "cor_method",
                "max_iter")),
  N_FACTORS = list(
    callee = "efa_retain",
    translating = TRUE,
    formals = c("x", "criteria", "suitability", "N", "use", "cor_method",
                "n_factors_max", "N_pop", "N_samples", "alpha", "max_iter_CD",
                "n_fac_theor", "method", "gof", "eigen_type_HULL",
                "eigen_type_other", "n_factors", "n_datasets", "percent",
                "decision_rule", "ekc_type", "n_datasets_nest", "alpha_nest",
                "show_progress", "...")),
  SL = list(
    callee = "efa_schmid_leiman",
    # `type` is frozen here but is no longer a bare formal of the successor: it is repacked
    # into the estimation control alongside any flat knob the dots carry.
    translating = TRUE,
    formals = c("x", "Phi", "type", "method", "g_name", "...")),
  COMPARE = list(
    callee = "efa_compare",
    formals = c("x", "y", "reorder", "corres", "thresh", "digits", "m_red",
                "range_red", "round_red", "print_diff", "na.rm", "x_labels",
                "plot", "plot_red")),
  EFA_AVERAGE = list(
    callee = "efa_average",
    # the frozen `method` and `P_type` formals forward to the successor's renamed
    # `estimator` and `p_type`.
    renamed = c(method = "estimator", P_type = "p_type"),
    formals = c("x", "n_factors", "N", "method", "rotation", "type", "averaging",
                "trim", "salience_threshold", "max_iter", "init_comm", "criterion",
                "criterion_type", "abs_eigen", "varimax_type", "normalize",
                "k_promax", "k_simplimax", "P_type", "precision", "start_method",
                "use", "cor_method", "show_progress")),
  PROCRUSTES = list(
    callee = "efa_procrustes",
    formals = c("A", "Target", "rotation", "S", "T_init", "oblique_eps",
                "oblique_maxit", "oblique_max_line_search", "oblique_step0",
                "oblique_normalize", "oblique_random_starts",
                "oblique_screen_keep", "oblique_triage_maxit",
                "oblique_triage_improve_tol")),
  EFA_POOLED = list(
    callee = "efa_mi",
    translating = TRUE,
    formals = c("data_list", "p", "target_method", "align_unrotated",
                "fit_pool_method", "consensus_args", "procrustes_args",
                "rmsea_ci_level", "rmsr_upper", "..."))
)

test_that("every superseded wrapper keeps a frozen signature and forwards it whole", {
  # covr instruments function bodies with counters, so their source structure
  # cannot be inspected meaningfully during a coverage run.
  skip_on_covr()

  for (old_name in names(superseded_contract)) {
    spec <- superseded_contract[[old_name]]
    wrapper <- get(old_name)

    expect_identical(names(formals(wrapper)), spec$formals)

    # a translating wrapper repacks the flat tuning knobs out of its `...` first, so it is not
    # a one-line forwarder; the frozen signature above is the static half of its contract, and
    # the knob-forwarding tests below are the behavioural half.
    if (isTRUE(spec$translating)) next

    # the body is `{ efa_*(<forwarded arguments>) }`: exactly one call, to the
    # successor and to nothing else
    b <- body(wrapper)
    expect_identical(as.character(b[[1]]), "{")
    expect_length(b, 2L)

    fwd <- b[[2]]
    expect_identical(as.character(fwd[[1]]), spec$callee)

    # the first formal (`x` for every name but PROCRUSTES, whose data argument is
    # `A`) passed positionally first, `...` positionally last, and every other
    # formal passed by name to the parameter of the same name
    args <- as.list(fwd)[-1]
    arg_names <- names(args)
    if (is.null(arg_names)) arg_names <- rep("", length(args))
    passed <- vapply(seq_along(args), function(i) {
      value <- paste(deparse(args[[i]]), collapse = "")
      if (nzchar(arg_names[i])) paste0(arg_names[i], " = ", value) else value
    }, character(1))

    positional <- spec$formals[[1]]
    forwarded <- setdiff(spec$formals, c(positional, "..."))
    # a renamed argument forwards `<successor param> = <frozen formal>`; every
    # other formal forwards unchanged as `<formal> = <formal>`.
    callee_param <- forwarded
    if (!is.null(spec$renamed)) {
      hit <- forwarded %in% names(spec$renamed)
      callee_param[hit] <- spec$renamed[forwarded[hit]]
    }
    expected <- c(positional, paste0(callee_param, " = ", forwarded))
    if ("..." %in% spec$formals) expected <- c(expected, "...")
    expect_identical(passed, expected)
  }
})

# The `...`-forwarding wrappers translate rather than forward verbatim: the old flat tuning
# knobs their dots may carry are repacked into the control objects the fit now takes. Forwarded
# bare, `type` (and every other moved knob) would match no `efa_fit()` formal and be dropped,
# and the fit would quietly run the default preset instead of the requested one.

test_that(".repack_flat_dots() splits the flat knobs into the two controls", {
  out <- .repack_flat_dots(list(type = "SPSS", max_iter = 42, k = 3, maxit = 900))

  expect_s3_class(out$estimate_control, "efa_estimate_control")
  expect_s3_class(out$rotate_control, "efa_rotate_control")
  expect_identical(out$estimate_control$type, "SPSS")
  expect_identical(out$rotate_control$type, "SPSS")
  expect_equal(out$estimate_control$max_iter, 42)
  expect_equal(out$rotate_control$k, 3)
  # a genuine rotation extra is not a control knob and rides on untouched
  expect_equal(out$maxit, 900)

  # an unnamed dot rides on too: the leftovers are kept by position, not by name
  pos <- .repack_flat_dots(list(42, max_iter = 5))
  expect_identical(pos[[3]], 42)
  expect_identical(names(pos), c("estimate_control", "rotate_control", ""))

  # the former argument spellings land on the current names
  ren <- .repack_flat_dots(list(P_type = "norm", randomStarts = 7))
  expect_identical(ren$rotate_control$p_type, "norm")
  expect_equal(ren$rotate_control$random_starts, 7)

  # the former `method` argument lands on `estimator`, even alongside an explicit
  # control object (it is a renamed argument, not a knob a control could carry)
  expect_identical(.repack_flat_dots(list(method = "ML")), list(estimator = "ML"))
  with_ctl <- .repack_flat_dots(list(method = "ML",
                                     estimate_control = estimate_control()))
  expect_identical(with_ctl$estimator, "ML")
  expect_false("method" %in% names(with_ctl))

  # a knob named without a `type` resolves against the preset the flat interface defaulted to
  expect_identical(.repack_flat_dots(list(k = 2))$rotate_control$type, "EFAtools")

  # nothing to translate: the dots are handed on exactly as they came in
  expect_identical(.repack_flat_dots(list(maxit = 5)), list(maxit = 5))

  # an explicit control object is the current interface and is never second-guessed
  ctl <- list(estimate_control = estimate_control(type = "psych"), type = "SPSS")
  expect_identical(.repack_flat_dots(ctl), ctl)
})

test_that("EFA_POOLED() still tunes the fit through its dots", {
  imps <- list(test_models$baseline$cormat, test_models$baseline$cormat)
  # the frozen interface still selects the estimator with `method`; the successor takes
  # `estimator`
  args_old <- list(n_factors = 3, N = 500, method = "PAF", rotation = "promax")
  args_new <- list(n_factors = 3, N = 500, estimator = "PAF", rotation = "promax")

  old <- do.call(EFA_POOLED, c(list(imps), args_old, list(type = "psych")))
  new <- do.call(efa_mi, c(list(imps), args_new,
                           list(estimate_control = estimate_control(type = "psych"),
                                rotate_control = rotate_control(type = "psych"))))
  expect_equal(old$rot_loadings, new$rot_loadings)

  # the knob is honoured, not merely ignored in both calls: psych is not the default preset
  default <- do.call(efa_mi, c(list(imps), args_new))
  expect_false(isTRUE(all.equal(old$rot_loadings, default$rot_loadings)))
})

test_that("PARALLEL() still tunes the EFA eigenvalues through its dots", {
  cormat <- test_models$baseline$cormat
  args <- list(N = 500, eigen_type = "EFA", n_datasets = 2)

  set.seed(42L)
  old <- suppressWarnings(do.call(PARALLEL, c(list(cormat), args, list(type = "psych"))))
  set.seed(42L)
  new <- suppressWarnings(do.call(
    efa_parallel, c(list(cormat), args,
                    list(estimate_control = estimate_control(type = "psych"),
                         rotate_control = rotate_control(type = "psych")))))
  expect_equal(old$eigenvalues, new$eigenvalues)
})

test_that("SL() still routes a non-default `type` into the second-order fit", {
  efa_mod <- oblique_efa

  # `efa_schmid_leiman()` has no `type` formal any more, so the frozen argument only reaches the
  # second-order fit if the wrapper repacks it into the estimation control. A default-valued
  # `type` would pass either way, so the preset pinned here has to be a non-default one.
  # The second-order fit is the only place the preset and the estimator take effect, so
  # its note that only PAF is validated against SPSS is what tells the user that this
  # combination is untested -- through the frozen wrapper and its successor alike.
  expect_warning(old <- SL(efa_mod, type = "SPSS", method = "ULS"),
                 class = "efa_spss_method_untested")
  expect_warning(
    new <- efa_schmid_leiman(efa_mod, estimator = "ULS",
                             estimate_control = estimate_control(type = "SPSS")),
    class = "efa_spss_method_untested")
  expect_equal(old, new, tolerance = fp_tol)

  # and it is honoured, not merely accepted: SPSS is not the preset the fit defaults to
  default <- efa_schmid_leiman(efa_mod, estimator = "ULS")
  expect_false(isTRUE(all.equal(old$settings, default$settings)))
})

test_that("SL() can still supply the PAF knobs that type = 'none' requires", {
  efa_mod <- oblique_efa

  # forwarded bare these knobs were dropped, leaving type = "none" with nothing to resolve
  expect_no_error(
    sl <- SL(efa_mod, type = "none", method = "PAF", init_comm = "smc",
             criterion = 1e-3, criterion_type = "sum", max_iter = 300, abs_eigen = TRUE)
  )
  expect_s3_class(sl, "efa_schmid_leiman")
})

test_that("HULL() still tunes the fit through its dots", {
  cormat <- test_models$baseline$cormat
  # the frozen wrapper still selects the estimator with `method`; the successor takes
  # `estimator`
  args_old <- list(N = 500, method = "PAF", gof = "CAF")
  args_new <- list(N = 500, estimator = "PAF", gof = "CAF")

  # HULL() fits every 1..J model through the estimation engine, so a flat knob passed
  # through its dots has to reach that fit. It is repacked into the control objects, and
  # efa_hull() must be able to take the controls it is handed.
  old <- do.call(HULL, c(list(cormat), args_old, list(type = "psych")))
  new <- do.call(efa_hull, c(list(cormat), args_new,
                             list(estimate_control = estimate_control(type = "psych"),
                                  rotate_control = rotate_control(type = "psych"))))
  expect_equal(old, new, tolerance = fp_tol)

  # the knob is honoured, not merely accepted: max_iter = 1 halts PAF after a single
  # iteration, so the goodness-of-fit values the hull is built from must change
  pinned <- suppressWarnings(
    do.call(efa_hull, c(list(cormat), args_new,
                        list(estimate_control = estimate_control(max_iter = 1)))))
  default <- do.call(efa_hull, c(list(cormat), args_new))
  expect_false(isTRUE(all.equal(pinned$results, default$results)))
})

# Silence: a superseded name must add no condition of its own. The criteria below take a
# correlation matrix, so nothing is emitted about deriving one from raw data.

test_that("the retention criteria wrappers add no condition", {
  cormat <- test_models$baseline$cormat

  expect_no_condition(PARALLEL(cormat, N = 500, n_datasets = 2))
  expect_no_condition(KGC(cormat))
  expect_no_condition(SCREE(cormat))
  expect_no_condition(SMT(cormat, N = 500))
  expect_no_condition(NEST(cormat, N = 500, n_datasets = 2))
  expect_no_condition(HULL(cormat, N = 500, method = "ML", gof = "CAF"))
})

test_that("CD() adds no condition and stays transparent", {
  # Comparison data is the most expensive call in this file, so the silence assertion is
  # made on the very run whose result is then compared rather than on a second one.
  set.seed(42L)
  expect_no_condition(old <- CD(GRiPS_raw, N_pop = 500, N_samples = 50))
  set.seed(42L)
  new <- efa_cd(GRiPS_raw, N_pop = 500, N_samples = 50)

  expect_equal(old, new, tolerance = fp_tol)
  expect_identical(class(old), "efa_retention")
})

# A flat tuning knob no longer has an efa_fit() formal, so a bare copy would land in the
# rotation extras and be dropped -- the fit would silently run the default preset. It is
# rejected instead, and the superseded names (which repack it into the controls) keep working.

test_that("efa_fit() rejects a flat tuning knob passed as a bare dot", {
  cormat <- test_models$baseline$cormat

  for (knob in list(list(type = "SPSS"), list(max_iter = 500), list(init_comm = "mac"),
                    list(criterion = 1e-4), list(criterion_type = "sum"),
                    list(abs_eigen = TRUE), list(start_method = "factanal"),
                    list(normalize = FALSE), list(precision = 1e-6),
                    list(order_type = "eigen"), list(varimax_type = "kaiser"),
                    list(p_type = "norm"), list(k = 3), list(random_starts = 2),
                    list(P_type = "norm"), list(randomStarts = 2))) {
    expect_error(
      do.call(efa_fit, c(list(cormat, n_factors = 3, N = 500, estimator = "PAF",
                              rotation = "promax"), knob)),
      class = "efa_flat_knob_in_dots")
  }

  # the functions that forward their dots to a fit surface the same classed error, rather
  # than an opaque "criterion could not be run"
  expect_error(efa_retain(cormat, N = 500, criteria = "HULL", type = "psych"),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_hull(cormat, N = 500, estimator = "PAF", gof = "CAF", max_iter = 500),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_kgc(cormat, eigen_type = "EFA", init_comm = "mac"),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_mi(list(cormat, cormat), n_factors = 2, type = "psych"),
               class = "efa_flat_knob_in_dots")

  # arguments that really are efa_fit() formals still pass through the same dots
  expect_no_error(suppressWarnings(
    efa_retain(cormat, N = 500, criteria = "HULL", estimator = "PAF")))

  # a genuine rotation extra is still forwarded, not rejected
  expect_no_error(efa_fit(cormat, n_factors = 3, N = 500, estimator = "PAF",
                          rotation = "oblimin", maxit = 750))
  # and the control objects remain the supported route (pinning a knob alongside a `type`
  # raises the usual preset-override notice, which is not what this test is about)
  expect_no_error(suppressWarnings(
    efa_fit(cormat, n_factors = 3, N = 500, estimator = "PAF", rotation = "promax",
            estimate_control = estimate_control(type = "SPSS", max_iter = 500))))
})

test_that("the former `method` argument is rejected with a pointer to `estimator`", {
  cormat <- test_models$baseline$cormat

  # Forwarded bare, `method` would either be rejected as an unconsumed rotation extra
  # (efa_fit) or travel into the per-criterion fits and surface as an opaque failure
  # (efa_retain); both name the rename instead.
  expect_error(efa_fit(cormat, n_factors = 3, N = 500, method = "ML"),
               class = "efa_renamed_arg")
  expect_error(efa_retain(cormat, N = 500, criteria = "EKC", method = "ML"),
               class = "efa_renamed_arg")
  expect_error(efa_schmid_leiman(oblique_efa, method = "PAF"),
               class = "efa_renamed_arg")
  # the functions that fit through their dots reject it at their own boundary, so the
  # message names the function the user called (not efa_fit, and in efa_group's case not
  # a spurious per-group fit failure)
  expect_error(efa_mi(list(cormat, cormat), n_factors = 2, method = "ML"),
               class = "efa_renamed_arg")
  expect_error(efa_group(list(a = cormat, b = cormat), n_factors = 2, N = 500,
                         method = "ML"),
               class = "efa_renamed_arg")

  # the superseded names keep selecting the estimator with `method`
  expect_no_error(suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, method = "ML", rotation = "none")))
})

test_that("the successor-only names are rejected in the superseded wrappers' dots", {
  cormat <- test_models$baseline$cormat

  # `estimator` (and the control objects, for EFA()) would either collide with the
  # wrapper's own translation as an opaque base duplicate-argument error, or -- for EFA(),
  # under its frozen ignore-unknown-dots contract -- be silently dropped so the fit quietly
  # runs the defaults. Rejected with a pointer to the frozen names.
  expect_error(EFA(cormat, n_factors = 3, N = 500, rotation = "promax",
                   estimator = "ML"),
               class = "efa_renamed_arg")
  expect_error(EFA(cormat, n_factors = 3, N = 500, rotation = "none",
                   estimate_control = estimate_control(type = "psych")),
               class = "efa_renamed_arg")
  expect_error(HULL(cormat, N = 500, estimator = "ML"), class = "efa_renamed_arg")
  expect_error(N_FACTORS(cormat, N = 500, criteria = "EKC", estimator = "ML"),
               class = "efa_renamed_arg")
  expect_error(SL(oblique_efa, estimator = "PAF"), class = "efa_renamed_arg")

  # both spellings at once through a repacking wrapper is a half-migrated call; neither
  # may silently win
  expect_error(.repack_flat_dots(list(method = "ML", estimator = "ULS")),
               class = "efa_renamed_arg")
})

test_that("the superseded names still tune the fit despite the flat-knob guard", {
  cormat <- test_models$baseline$cormat

  # EFA() repacks its frozen flat arguments into the controls, so the guard never sees them
  expect_no_error(old <- suppressWarnings(
    EFA(cormat, n_factors = 3, N = 500, method = "PAF", rotation = "promax",
        type = "SPSS", max_iter = 500)))
  new <- suppressWarnings(
    efa_fit(cormat, n_factors = 3, N = 500, estimator = "PAF", rotation = "promax",
            estimate_control = estimate_control(type = "SPSS", max_iter = 500),
            rotate_control = rotate_control(type = "SPSS")))
  expect_equal(old$rot_loadings, new$rot_loadings, tolerance = fp_tol)

  # the same holds for the wrappers whose dots reach a fit
  expect_no_error(HULL(cormat, N = 500, method = "PAF", gof = "CAF", max_iter = 500))
  expect_no_error(N_FACTORS(cormat, N = 500, criteria = "EKC", type = "psych"))
})

test_that("N_FACTORS() keeps partial-matching `max_iter` to its frozen `max_iter_CD`", {
  cormat <- test_models$baseline$cormat

  # `max_iter` is a unique prefix of the frozen `max_iter_CD` formal, so R has always matched it
  # there rather than passing it on through the dots. efa_retain() now rejects that spelling --
  # the tuning knob it looks like belongs in estimate_control() -- but the old name is frozen,
  # so the argument keeps the meaning every existing script gave it.
  nf <- suppressWarnings(suppressMessages(
    N_FACTORS(cormat, N = 500, suitability = FALSE, criteria = "EKC", max_iter = 5)))
  expect_identical(nf$settings$max_iter_CD, 5)

  expect_error(efa_retain(cormat, N = 500, criteria = "EKC", max_iter = 5),
               class = "efa_flat_knob_in_dots")
})

# cor_method = "fiml" propagated to the sibling entry points: efa_average() averages a
# two-stage FIML correlation (estimated once and reused across the grid), and the
# simulation-/eigenvalue-based retention and suitability functions leave "fiml" out of
# their choices. The EM engine and the EFA() routing themselves are covered elsewhere
# (test-fiml-moments.R, test-EFA-fiml.R).

# A missing-at-random fixture: a clean two-factor population, with column 1 fully observed
# and driving the missingness in the others (so the mechanism depends only on observed data).
# The deviates are coloured by the Cholesky factor rather than by an eigendecomposition: this
# population covariance has repeated eigenvalues, whose eigenvector basis is undetermined and
# hence settled by rounding, while chol() is unique for a positive definite matrix, so the same
# seed reproduces the same sample on every LAPACK build.
avg_fiml_mar_data <- function(n = 800, seed = 4242) {
  set.seed(seed)
  L <- matrix(0, 6, 2)
  L[1:3, 1] <- 0.7
  L[4:6, 2] <- 0.7
  S <- tcrossprod(L); diag(S) <- 1
  X <- matrix(stats::rnorm(n * 6), n) %*% chol(S)
  colnames(X) <- paste0("V", seq_len(6))
  X[X[, 1] >  0.8, 2] <- NA
  X[X[, 1] < -0.8, 3] <- NA
  X[X[, 1] >  1.2, 4] <- NA
  X[X[, 1] < -1.2, 5] <- NA
  X
}

test_that("efa_average with cor_method = 'fiml' analyses the two-stage FIML correlation", {
  skip_on_cran()

  X <- avg_fiml_mar_data()

  avg <- suppressMessages(suppressWarnings(
    efa_average(X, n_factors = 2, estimator = c("ML", "ULS"), rotation = "none",
                cor_method = "fiml", show_progress = FALSE)
  ))
  expect_s3_class(avg, "efa_average")

  # The analysed matrix is the two-stage FIML correlation: the EM saturated covariance,
  # standardized, estimated once up front from the raw data with missing values.
  em <- .fiml_em_moments(X)
  R_fiml <- stats::cov2cor(em$sigma)
  expect_equal(unclass(avg$orig_R), R_fiml, tolerance = 1e-8, ignore_attr = TRUE)
  expect_equal(avg$settings$N, em$n)

  # Routing raw data through cor_method = "fiml" is consistent with passing the precomputed
  # FIML correlation matrix through the same grid: both analyse the same matrix with the same
  # (deterministic, unrotated ML/ULS) solutions, so the averaged loadings coincide. (That the
  # EM is run only once for the whole grid is pinned separately below.)
  avg_ref <- suppressMessages(suppressWarnings(
    efa_average(R_fiml, N = em$n, n_factors = 2, estimator = c("ML", "ULS"),
                rotation = "none", show_progress = FALSE)
  ))
  expect_equal(unclass(avg$loadings$average), unclass(avg_ref$loadings$average),
               tolerance = 1e-8, ignore_attr = TRUE)
})

test_that("efa_average runs the FIML EM once for the whole grid, not per solution", {
  skip_on_cran()

  X <- avg_fiml_mar_data()

  # Count the EM calls during one efa_average run. The saturated moments are estimated once up
  # front in .prepare_cor_input() and the resulting correlation is reused by every solution in
  # the grid, so the engine is hit exactly once; a regression to per-cell EM would call it once
  # per grid cell. The per-cell EFA() calls receive that correlation matrix, so they never
  # re-enter the EM regardless of the future plan.
  real_em <- .fiml_em_moments
  n_em <- 0L
  testthat::local_mocked_bindings(
    .fiml_em_moments = function(...) {
      n_em <<- n_em + 1L
      real_em(...)
    }
  )

  invisible(suppressMessages(suppressWarnings(
    efa_average(X, n_factors = 2, estimator = c("ML", "ULS"), rotation = "none",
                cor_method = "fiml", show_progress = FALSE)
  )))

  expect_identical(n_em, 1L)
})

test_that("retention and suitability functions reject cor_method = 'fiml'", {
  # FIML is a single-fit missing-data correlation; the simulation-/eigenvalue-based
  # retention and suitability criteria do not support it, so it is left out of their
  # cor_method choices and rejected up front (before any computation) by the shared
  # choice-matching layer. Asserting that layer's condition class, rather than merely that
  # something failed, is what keeps the test tied to the rejection it is named for: any other
  # abort would otherwise satisfy it silently.
  expect_error(N_FACTORS(GRiPS_raw, cor_method = "fiml"), class = "efa_bad_choice")
  expect_error(KMO(GRiPS_raw, cor_method = "fiml"), class = "efa_bad_choice")
})

# EFA_POOLED() reports the incremental fit indices CFI (Bentler, 1990) and TLI
# (Tucker & Lewis, 1973) as the average of the per-imputation indices. Each
# per-imputation index is computed in range by the per-model path, so the average
# stays in range and consistent with the component fits, and reduces to EFA()'s
# value in the single-imputation limit. Pooling the model and baseline chi-squares
# separately (the lavaan.mi/semTools convention) shrinks the two noncentralities by
# their own between-imputation ARIV and can drive these non-linear indices outside
# [0, 1] and against every per-imputation value; those separately pooled
# noncentralities remain available, for reference, in mi_diagnostics. These tests
# check the averaging convention.
#
# Assertion groups:
#   1. Single-imputation (identical-imputation) limit reduces to EFA().
#   2. Pooled CFI/TLI equal the mean of the per-imputation indices (not the
#      separately pooled (N - 1)-scale formula).
#   3. Pooled CFI/TLI stay in range and within the per-imputation range under
#      genuine multi-imputation fits where separate pooling did not.

# ---- Group 1 ---------------------------------------------------------------

test_that("identical imputations reproduce EFA()'s CFI/TLI exactly", {
  cormat <- test_models$baseline$cormat
  N <- 500L
  k <- 3L
  single <- EFA(cormat, n_factors = k, N = N, method = "ML", rotation = "none")
  pooled <- suppressMessages(EFA_POOLED(
    replicate(4, cormat, simplify = FALSE),
    n_factors = k, N = N, method = "ML", rotation = "none"
  ))

  # Identical imputations have zero between-imputation spread, so every per-
  # imputation CFI/TLI equals the single-fit value and their average must match
  # EFA() to machine precision.
  expect_equal(pooled$fit_indices$CFI, single$fit_indices$CFI, tolerance = 1e-8)
  expect_equal(pooled$fit_indices$TLI, single$fit_indices$TLI, tolerance = 1e-8)
})

# ---- Group 2 ---------------------------------------------------------------

test_that("pooled CFI/TLI equal the mean of the per-imputation indices", {
  set.seed(42)
  grips_list <- lapply(1:3, function(i) {
    GRiPS_raw[sample(seq_len(nrow(GRiPS_raw)), replace = TRUE), ]
  })
  pooled <- suppressWarnings(suppressMessages(
    EFA_POOLED(grips_list, n_factors = 1, method = "ML")
  ))
  fi <- pooled$fit_indices

  cfis <- vapply(pooled$fits, function(f) f$fit_indices$CFI, numeric(1))
  tlis <- vapply(pooled$fits, function(f) f$fit_indices$TLI, numeric(1))

  expect_equal(fi$CFI, mean(cfis[is.finite(cfis)]), tolerance = 1e-12)
  expect_equal(fi$TLI, mean(tlis[is.finite(tlis)]), tolerance = 1e-12)

  # The reported indices are the average, not the discarded separately pooled
  # (N - 1)-scale formula. That formula is still recoverable from the diagnostics
  # mi_diagnostics exposes, and here it gives a different number, so a revert to it
  # would be caught.
  md <- pooled$mi_diagnostics
  d_m <- max(md$chi_cfi - fi$df, 0)
  d_null <- max(md$chi_null_cfi - fi$df_null, 0)
  CFI_sep <- if (max(d_m, d_null) == 0) 1 else 1 - d_m / max(d_m, d_null)
  ratio <- md$chi_null_cfi / fi$df_null
  TLI_sep <- (ratio - md$chi_cfi / fi$df) / (ratio - 1)
  expect_false(isTRUE(all.equal(fi$CFI, CFI_sep)))
  expect_false(isTRUE(all.equal(fi$TLI, TLI_sep)))
})

# ---- Group 3 ---------------------------------------------------------------

test_that("pooled CFI/TLI stay in range and consistent under multi-imputation fits", {
  skip_on_cran()

  # A 9-variable population with no clean two-factor structure: a k = 2 EFA
  # underfits, so each imputation lands at a moderate CFI/TLI well inside (0, 1).
  # Pooling the model and baseline chi-squares separately drove the pooled TLI
  # negative (and the pooled CFI far below every per-imputation value); averaging
  # the per-imputation indices keeps both in range and within the component range.
  set.seed(1)
  p <- 9L
  Sigma <- stats::cov2cor(tcrossprod(matrix(stats::rnorm(p * 4), p)) + diag(p))
  # Draw through the Cholesky factor: chol() is unique for a positive definite
  # matrix because the positive-diagonal convention pins it, so the same seed
  # yields the same sample on every LAPACK build. An eigen-based draw would not,
  # since eigenvector signs (and the basis of a repeated eigenspace) are left
  # undetermined by the decomposition.
  Rt <- chol(Sigma)

  check_in_range <- function(data_list, k) {
    pooled <- suppressWarnings(suppressMessages(
      EFA_POOLED(data_list, n_factors = k, method = "ML", rotation = "none")
    ))
    fi <- pooled$fit_indices
    cfis <- vapply(pooled$fits, function(f) f$fit_indices$CFI, numeric(1))
    tlis <- vapply(pooled$fits, function(f) f$fit_indices$TLI, numeric(1))

    # CFI is always in [0, 1] (mean of in-range per-imputation values).
    expect_gte(fi$CFI, 0)
    expect_lte(fi$CFI, 1)
    # Both indices lie within the per-imputation range -- the consistency the
    # separately pooled statistics violated (its TLI fell below every component).
    expect_gte(fi$CFI, min(cfis) - 1e-8)
    expect_lte(fi$CFI, max(cfis) + 1e-8)
    expect_gte(fi$TLI, min(tlis) - 1e-8)
    expect_lte(fi$TLI, max(tlis) + 1e-8)
    invisible(pooled)
  }

  # Homogeneous N.
  imps_hom <- lapply(1:5, function(i) {
    as.data.frame(matrix(stats::rnorm(300 * p), 300, p) %*% Rt)
  })
  check_in_range(imps_hom, k = 2)

  # Heterogeneous N: the degenerate case where each separately pooled chi-square
  # floored to zero, returning the self-contradictory CFI = 1 / TLI = 0.
  Ns <- c(200, 400, 600, 250, 550)
  imps_het <- lapply(Ns, function(n) {
    as.data.frame(matrix(stats::rnorm(n * p), n, p) %*% Rt)
  })
  check_in_range(imps_het, k = 2)
})

# Captured-output baselines for the print methods, recorded with
# local_reproducible_output() so the text is plain (no ANSI colour) and fixed
# width. Only methods whose printed output is integer factor counts or fixed
# text are snapshotted; methods that print computed decimals (e.g. KMO,
# Bartlett's chi-square) are omitted, since those digits can drift across BLAS
# and platforms and would make the baselines flaky.
#
# Only the criteria that simulate -- PARALLEL, NEST and HULL -- carry
# skip_if_not_slow(). They cost about an order of magnitude more than all the
# deterministic blocks together, which is what the gate is for. The rest run by
# default: gating a block that costs a tenth of a second only hides a baseline
# from the runs that would catch it drifting. Gate a new block here only if it
# simulates or otherwise joins the expensive group.

test_that("print.efa_retention output is stable for SMT", {
  local_reproducible_output()

  smt <- SMT(test_models$baseline$cormat, N = 500)
  expect_snapshot(print(smt))

  # Identity matrix: zero-factor solution.
  smt_id <- SMT(diag(5), N = 500)
  expect_snapshot(print(smt_id))
})

test_that("print.efa_retention output is stable for SCREE", {
  local_reproducible_output()

  scree <- SCREE(test_models$baseline$cormat)
  expect_snapshot(print(scree))

  scree_smc <- SCREE(test_models$baseline$cormat, eigen_type = "SMC")
  expect_snapshot(print(scree_smc))
})

test_that("print.efa_retention output is stable for CD", {
  local_reproducible_output()

  # CD is stochastic; seed and use a small simulation for a stable, fast snapshot
  set.seed(123)
  cd <- CD(GRiPS_raw, N_pop = 1000, N_samples = 100)
  expect_snapshot(print(cd))
})

test_that("print.efa_retention output is stable for PARALLEL", {
  skip_if_not_slow()
  local_reproducible_output()

  # seed the simulated eigenvalues so the printed suggestion cannot drift
  set.seed(123)
  pa <- PARALLEL(test_models$baseline$cormat, N = 500, eigen_type = "PCA")
  expect_snapshot(print(pa))

  # without real data there is no numeric suggestion (visual output only)
  pa_nodat <- PARALLEL(N = 20, n_vars = 5)
  expect_snapshot(print(pa_nodat))
})

test_that("print.efa_retention output is stable for KGC", {
  local_reproducible_output()

  kgc <- KGC(test_models$baseline$cormat)
  expect_snapshot(print(kgc))

  kgc_smc <- KGC(test_models$baseline$cormat, eigen_type = "SMC")
  expect_snapshot(print(kgc_smc))
})

test_that("print.efa_retention output is stable for NEST", {
  skip_if_not_slow()
  local_reproducible_output()

  nest <- NEST(test_models$baseline$cormat, N = 500)
  expect_snapshot(print(nest))
})

test_that("print.efa_retention output is stable for EKC", {
  local_reproducible_output()

  ekc <- EKC(test_models$baseline$cormat, N = 500)
  expect_snapshot(print(ekc))
})

test_that("print.efa_retention output is stable for HULL", {
  skip_if_not_slow()
  local_reproducible_output()

  hull <- HULL(test_models$baseline$cormat, N = 500, method = "ML")
  expect_snapshot(print(hull))

  hull_paf <- suppressMessages(HULL(test_models$baseline$cormat, N = 500))
  expect_snapshot(print(hull_paf))
})

test_that("print.efa_retention output is stable for MAP", {
  local_reproducible_output()

  map <- MAP(test_models$baseline$cormat)
  expect_snapshot(print(map))
})

test_that("print.efa_retain output is stable", {
  local_reproducible_output()

  # deterministic criteria only (no simulations); suitability = FALSE keeps
  # BLAS-sensitive decimals (chi-square, KMO) out of the snapshot
  nf <- efa_retain(test_models$baseline$cormat, N = 500, suitability = FALSE,
                   criteria = c("EKC", "KGC", "MAP", "SMT"),
                   eigen_type_other = c("PCA", "SMC"))
  expect_snapshot(print(nf))
})

test_that("every print method is exactly cat(format(x), sep = \"\\n\")", {
  skip_on_cran()
  local_reproducible_output()

  cm <- test_models$baseline$cormat
  fit <- efa_fit(cm, n_factors = 3, N = 500, estimator = "PAF", rotation = "promax")
  fit_ml <- efa_fit(cm, n_factors = 3, N = 500, estimator = "ML", rotation = "promax")
  sl <- efa_schmid_leiman(fit, estimator = "PAF")
  # correlation-matrix input, so efa_scores() notes that it returns weights only
  scores <- suppressMessages(efa_scores(cm, f = fit))

  # One object per printable class whose fixture is cheap to build. The three that are
  # not - efa_average, efa_group, and efa_mi - assert the same identity alongside their
  # own fixtures. The two loading tables print a trailing blank line for console spacing,
  # which is the only documented departure from the rule.
  fixtures <- list(
    efa                  = fit,
    summary.efa          = summary(fit),
    efa_loadings         = fit$rot_loadings,
    efa_sl_loadings      = sl$sl,
    efa_schmid_leiman    = sl,
    OMEGA                = OMEGA(sl, type = "EFAtools",
                                 factor_corres = sl$sl[, c("F1", "F2", "F3")] >= .2),
    efa_reliability      = efa_reliability(fit),
    efa_scores           = scores,
    summary.efa_scores   = summary(scores),
    efa_compare          = efa_compare(fit$unrot_loadings, fit_ml$unrot_loadings),
    efa_retention        = efa_ekc(cm, N = 500),
    efa_retain           = efa_retain(cm, N = 500, suitability = FALSE,
                                      criteria = c("EKC", "KGC")),
    efa_screen           = efa_screen(cm, N = 500),
    efa_kmo              = efa_kmo(cm),
    efa_bartlett         = efa_bartlett(cm, N = 500),
    efa_power            = efa_power(df = 100, N = 200),
    efa_simulated        = efa_simulate(N = 100, Lambda = population_models$loadings$baseline,
                                        seed = 42),
    efa_estimate_control = estimate_control(type = "SPSS"),
    efa_rotate_control   = rotate_control(type = "SPSS")
  )
  # trailing blank line, see above
  spacer <- c("efa_loadings", "efa_sl_loadings")

  for (nm in names(fixtures)) {
    obj <- fixtures[[nm]]
    expected <- format(obj)
    if (nm %in% spacer) expected <- c(expected, "")
    expect_identical(utils::capture.output(print(obj)), expected, info = nm)
  }
})

test_that("print.efa_retain summarises the criteria's suggestions", {
  local_reproducible_output()

  # A cheap deterministic set whose criteria disagree, pinning the summary line under
  # the section rule. The modal value is counted one vote per criterion: the empirical
  # Kaiser criterion and the Kaiser-Guttman criterion both suggest 3, while the
  # internally tied minimum average partial (1 and 3) is undecided and casts no vote
  # rather than casting two.
  nf <- efa_retain(test_models$baseline$cormat, N = 500, suitability = FALSE,
                   criteria = c("EKC", "KGC", "MAP"), eigen_type_other = "PCA")
  expect_snapshot(print(nf))
})

# The estimation settings must reach every factor-retention criterion that fits a model, and
# only through the control object: a criterion that accepts a control but never passes it on to
# its fit is indistinguishable, from the outside, from one that ignores it. Each criterion below
# is therefore shown to HONOUR the control, not merely to take it.
#
# Two ways to show it, one per estimator:
#   * PAF (what the criteria that fit use by default): `max_iter = 1` halts the iterations after
#     a single step, so the communalities -- and everything derived from them -- must change.
#   * ML: `start_method = NA` leaves the optimiser without starting values, which efa_fit()
#     rejects with a classed error. That the error surfaces proves the control reached the fit.
#     This is the only route for efa_smt(), whose models are maximum likelihood by definition,
#     so of the estimation knobs only `start_method` applies to it.
#
# CD, EKC and MAP fit no model (CD's `max_iter` caps its own comparison-data generation, not an
# EFA), so they take no control at all.

cormat <- test_models$baseline$cormat

pinned <- estimate_control(max_iter = 1)
no_start <- estimate_control(start_method = NA)

test_that("efa_hull() honours the estimation control", {
  args <- list(cormat, N = 500, estimator = "PAF", gof = "CAF", n_datasets = 10)

  set.seed(42L)
  default <- suppressWarnings(do.call(efa_hull, args))
  set.seed(42L)
  tuned <- suppressWarnings(do.call(efa_hull, c(args, list(estimate_control = pinned))))

  # the goodness-of-fit values the hull is built from come from the 0..J factor fits
  expect_false(isTRUE(all.equal(default$results, tuned$results)))

  # the control reaches those fits on the ML route too
  expect_error(
    efa_hull(cormat, N = 500, estimator = "ML", n_datasets = 10,
             estimate_control = no_start),
    class = "efa_ml_start_missing")
})

test_that("efa_kgc() honours the estimation control", {
  default <- efa_kgc(cormat, eigen_type = "EFA")
  tuned <- suppressWarnings(efa_kgc(cormat, eigen_type = "EFA",
                                    estimate_control = pinned))

  expect_false(isTRUE(all.equal(.retention_record(default, "EFA")$y,
                                .retention_record(tuned, "EFA")$y)))
})

test_that("efa_scree() honours the estimation control", {
  default <- efa_scree(cormat, eigen_type = "EFA")
  tuned <- suppressWarnings(efa_scree(cormat, eigen_type = "EFA",
                                      estimate_control = pinned))

  expect_false(isTRUE(all.equal(.retention_record(default, "EFA")$y,
                                .retention_record(tuned, "EFA")$y)))
})

test_that("efa_parallel() honours the estimation control", {
  args <- list(cormat, N = 500, eigen_type = "EFA", n_datasets = 2)

  set.seed(42L)
  default <- suppressWarnings(do.call(efa_parallel, args))
  set.seed(42L)
  tuned <- suppressWarnings(do.call(efa_parallel,
                                    c(args, list(estimate_control = pinned))))

  # the real-data EFA eigenvalues; the shared seed rules out a mere simulation difference
  expect_false(isTRUE(all.equal(.retention_record(default, "EFA")$y,
                                .retention_record(tuned, "EFA")$y)))
})

test_that("efa_nest() honours the estimation control", {
  args <- list(cormat, N = 500, n_datasets = 2)

  set.seed(42L)
  default <- suppressWarnings(do.call(efa_nest, args))
  set.seed(42L)
  tuned <- suppressWarnings(do.call(efa_nest, c(args, list(estimate_control = pinned))))

  # the reference eigenvalues are simulated from the (nf - 1)-factor PAF reference models
  expect_false(isTRUE(all.equal(.retention_record(default, "NEST")$reference,
                                .retention_record(tuned, "NEST")$reference)))

  expect_error(
    do.call(efa_nest, c(args, list(estimator = "ML", estimate_control = no_start))),
    class = "efa_ml_start_missing")
})

test_that("efa_smt() honours the estimation control", {
  # SMT could not be tuned at all before: it had no way to receive a control
  expect_error(efa_smt(cormat, N = 500, estimate_control = no_start),
               class = "efa_ml_start_missing")

  # the other ML starting values are accepted and still produce a valid result
  expect_s3_class(
    efa_smt(cormat, N = 500,
            estimate_control = estimate_control(start_method = "factanal")),
    "efa_retention")
})

test_that("efa_schmid_leiman() honours the estimation control", {
  efa_mod <- efa_fit(cormat, n_factors = 3, N = 500, estimator = "PAF",
                     rotation = "promax")

  # type = "none" resolves nothing from a preset, so the second-order fit only runs if the
  # knobs supplied in the control actually reach it
  expect_no_error(
    efa_schmid_leiman(efa_mod, estimator = "PAF",
                      estimate_control = estimate_control(
                        type = "none", init_comm = "smc", criterion = 1e-3,
                        criterion_type = "sum", max_iter = 300, abs_eigen = TRUE)))

  expect_error(
    efa_schmid_leiman(efa_mod, estimator = "ML", estimate_control = no_start),
    class = "efa_ml_start_missing")
})

test_that("efa_retain() threads the estimation control into every criterion that fits", {
  ids <- c("KGC", "NEST", "PARALLEL", "SCREE")
  args <- list(cormat, N = 500, suitability = FALSE, criteria = ids, estimator = "PAF",
               eigen_type_other = "EFA", n_datasets = 2, n_datasets_nest = 2)

  set.seed(42L)
  default <- suppressWarnings(suppressMessages(do.call(efa_retain, args)))
  set.seed(42L)
  tuned <- suppressWarnings(suppressMessages(
    do.call(efa_retain, c(args, list(estimate_control = pinned)))))

  for (id in ids) {
    expect_false(isTRUE(all.equal(default$outputs[[id]]$results,
                                  tuned$outputs[[id]]$results)),
                 info = id)
  }

  # the control reaches SMT as well: its models are fitted with ML, so the missing starting
  # values abort that criterion (efa_retain() excludes a failing criterion and carries on)
  smt <- suppressWarnings(suppressMessages(
    efa_retain(cormat, N = 500, suitability = FALSE, criteria = c("EKC", "SMT"),
               estimate_control = no_start)))
  expect_named(smt$not_run, "SMT")
  expect_named(smt$outputs, "EKC")

  expect_identical(tuned$settings$estimate_control, pinned)
})

test_that("efa_retain() rejects a flat knob that R would partial-match to a formal", {
  # `max_iter` is a unique prefix of the `max_iter_CD` formal, so R matched it there: the knob
  # silently capped the comparison-data iterations instead of tuning the fits, and never
  # reached the dots -- or the guard. `max_iter_CD` now sits behind `...`, where R matches it by
  # its full name only, so the knob lands in the dots and is named for what it is.
  expect_error(efa_retain(cormat, N = 500, criteria = "EKC", max_iter = 500),
               class = "efa_flat_knob_in_dots")

  # the same holds when the knob arrives through a caller's own `...`, where the partial match
  # would happen before efa_retain() ever ran and no inspection of the call could catch it
  forwarder <- function(...) efa_retain(cormat, N = 500, criteria = "EKC", ...)
  expect_error(forwarder(max_iter = 500), class = "efa_flat_knob_in_dots")

  # the comparison-data argument it was mistaken for still works, spelled out
  expect_no_error(suppressWarnings(suppressMessages(
    efa_retain(cormat, N = 500, criteria = "EKC", max_iter_CD = 5))))
  expect_identical(
    suppressWarnings(suppressMessages(
      efa_retain(cormat, N = 500, criteria = "EKC",
                 max_iter_CD = 5)))$settings$max_iter_CD,
    5)
})

test_that("a bare flat knob is rejected even when the call runs no fit", {
  # with eigen_type = "PCA" no model is fitted, so the knob never reached efa_fit()'s guard
  # and was dropped without a word
  expect_error(efa_kgc(cormat, eigen_type = "PCA", max_iter = 500),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_scree(cormat, eigen_type = "PCA", init_comm = "mac"),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_parallel(cormat, N = 500, eigen_type = "PCA", n_datasets = 2,
                            type = "SPSS"),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_nest(cormat, N = 500, n_datasets = 2, criterion = 1e-4),
               class = "efa_flat_knob_in_dots")
  expect_error(efa_hull(cormat, N = 500, gof = "CAF", abs_eigen = TRUE),
               class = "efa_flat_knob_in_dots")
  expect_error(
    efa_schmid_leiman(efa_fit(cormat, n_factors = 3, N = 500, estimator = "PAF",
                              rotation = "promax"),
                      max_iter = 300),
    class = "efa_flat_knob_in_dots")
})

test_that("an unknown dot is rejected even when the call runs no fit", {
  # the mirror of the flat-knob guard above: with eigen_type = "PCA" no model is fitted, so a
  # misspelled name never reached efa_fit()'s guard and the criterion silently ran the default
  expect_error(efa_kgc(cormat, eigen_type = "PCA", cor_methd = "poly"),
               class = "efa_unused_dots")
  expect_error(efa_scree(cormat, eigen_type = "PCA", percnt = 99),
               class = "efa_unused_dots")
  expect_error(efa_parallel(cormat, N = 500, eigen_type = "PCA", n_datasets = 2,
                            cor_methd = "poly"),
               class = "efa_unused_dots")
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "MAP",
                          nfactors = 3),
               class = "efa_unused_dots")

  # a genuine efa_fit() argument still travels through untouched
  expect_no_error(suppressWarnings(suppressMessages(
    efa_kgc(cormat, eigen_type = "PCA", estimator = "ML"))))
})

test_that("the retention criteria refuse a rotation setting and its engine extras", {
  # Every criterion fit is unrotated, so neither a rotation nor an engine extra can be
  # consumed: a `rotation` made the fit run an engine whose result no retention statistic
  # uses (and, for a criterion-based rotation, drew random starts from the caller's stream,
  # moving the seeded result), while an extra reached the fit and aborted there, naming
  # `rotation = "none"` instead of the function the user called. Both are refused at entry.
  expect_error(efa_kgc(cormat, eigen_type = "EFA", rotation = "promax"),
               class = "efa_unused_dots")
  expect_error(efa_scree(cormat, eigen_type = "EFA", rotation = "promax"),
               class = "efa_unused_dots")
  expect_error(efa_parallel(cormat, N = 500, eigen_type = "EFA", n_datasets = 2,
                            rotation = "promax"),
               class = "efa_unused_dots")
  expect_error(efa_nest(cormat, N = 500, n_datasets = 2, rotation = "promax"),
               class = "efa_unused_dots")
  expect_error(efa_hull(cormat, N = 500, rotation = "promax"),
               class = "efa_unused_dots")

  # the engine extras, refused at entry rather than deep inside the criterion's fit
  expect_error(efa_kgc(cormat, eigen_type = "EFA", maxit = 10),
               class = "efa_unused_dots")
  expect_error(efa_hull(cormat, N = 500, maxit = 500), class = "efa_unused_dots")
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "HULL",
                          maxit = 10),
               class = "efa_unused_dots")

  # a genuine efa_fit() argument still travels through untouched
  expect_no_error(suppressWarnings(suppressMessages(
    efa_kgc(cormat, eigen_type = "EFA", estimator = "ML"))))
})

test_that("the retention criteria refuse efa_fit()'s inference arguments", {
  # `se`, `b_boot`, `ci` and `seed` are efa_fit() formals, so they used to pass the dots
  # whitelist and reach the criterion fits, where nothing can come of them: the criteria keep
  # only eigenvalues and fit measures from those fits, so a standard error is computed and
  # discarded, and `seed` is spent on a fit that draws no random numbers while the simulated
  # reference data -- the only stochastic part -- are still drawn from the caller's stream.
  expect_error(efa_parallel(cormat, N = 500, eigen_type = "PCA", n_datasets = 2, seed = 42),
               class = "efa_unused_dots")
  expect_error(efa_nest(cormat, N = 500, n_datasets = 2, seed = 42),
               class = "efa_unused_dots")
  expect_error(efa_hull(cormat, N = 500, gof = "CAF", n_datasets = 2, seed = 42),
               class = "efa_unused_dots")
  expect_error(efa_kgc(cormat, eigen_type = "EFA", seed = 42), class = "efa_unused_dots")
  expect_error(efa_scree(cormat, eigen_type = "EFA", seed = 42), class = "efa_unused_dots")
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "PARALLEL",
                          n_datasets = 2, seed = 42),
               class = "efa_unused_dots")

  # refused even when the selected criteria run no fit at all -- that is the case the old
  # whitelist dropped in total silence, so it is the one a regression would restore first
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "MAP", seed = 42),
               class = "efa_unused_dots")
  expect_error(efa_kgc(cormat, eigen_type = "PCA", se = "information"),
               class = "efa_unused_dots")

  # the refused set is a hand-written mirror of four efa_fit() formals: if one is ever
  # renamed there, the guard stops matching and the name is silently forwarded again
  expect_true(all(.fit_inference_args %in% names(formals(efa_fit))))

  # An abbreviation must not slip past either: the dots are spliced into the fits with
  # do.call(), where R partial-matches them against efa_fit()'s formals, so `b_b` would
  # arrive as `b_boot` without matching the refused name exactly. The whitelist, which spells
  # every accepted name in full, is what refuses it -- with the misspelling message, since an
  # abbreviation is not one of this function's arguments either.
  expect_error(efa_kgc(cormat, eigen_type = "EFA", b_b = 10), class = "efa_unused_dots")
  expect_error(efa_parallel(cormat, N = 500, eigen_type = "PCA", n_datasets = 2, see = 42),
               class = "efa_unused_dots")

  # the standard-error group, refused at entry rather than doing discarded work in the fit
  expect_error(efa_kgc(cormat, eigen_type = "EFA", estimator = "ML", se = "information"),
               class = "efa_unused_dots")
  expect_error(efa_nest(cormat, N = 500, n_datasets = 2, se = "np-boot", b_boot = 10),
               class = "efa_unused_dots")
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "KGC",
                          eigen_type_other = "EFA", ci = .90),
               class = "efa_unused_dots")

  # the frozen wrappers forward their dots into these same functions, so the refusal reaches
  # the old names too. The flat interface did take all four, but they were inert on this path
  # (a correlation matrix never reached EFA()'s seeding branch, and a computed standard error
  # was discarded by the criterion), so refusing them changes no legacy result.
  expect_error(PARALLEL(cormat, N = 500, eigen_type = "PCA", n_datasets = 2, seed = 42),
               class = "efa_unused_dots")
  expect_error(N_FACTORS(cormat, N = 500, criteria = "PARALLEL", n_datasets = 2, seed = 42),
               class = "efa_unused_dots")

  # what replaces it: set.seed() beforehand, which pins the simulation the criterion draws
  set.seed(42L)
  a <- efa_parallel(cormat, N = 500, eigen_type = "PCA", n_datasets = 5)
  set.seed(42L)
  b <- efa_parallel(cormat, N = 500, eigen_type = "PCA", n_datasets = 5)
  expect_equal(a$results, b$results)
})

test_that("efa_retain() refuses a rotation setting but keeps the wrappers' repacked control", {
  # the criterion fits are always unrotated, so a rotation setting is meaningless here
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "MAP",
                          rotation = "promax"),
               class = "efa_unused_dots")
  expect_error(efa_retain(cormat, N = 500, suitability = FALSE, criteria = "MAP",
                          rotate_control = "SPSS"),
               class = "efa_unused_dots")

  # N_FACTORS() repacks a frozen `type` into a rotate_control() object that rides through
  # these same dots, so a real control object -- and a NULL, efa_fit()'s "not supplied"
  # default, which the repack passes through untouched -- must still be accepted
  expect_no_error(suppressWarnings(suppressMessages(
    efa_retain(cormat, N = 500, suitability = FALSE, criteria = "MAP",
               rotate_control = rotate_control()))))
  expect_no_error(suppressWarnings(suppressMessages(
    efa_retain(cormat, N = 500, suitability = FALSE, criteria = "MAP",
               rotate_control = NULL))))
})

test_that("an estimation control that is not a control object is rejected", {
  expect_error(efa_smt(cormat, N = 500, estimate_control = list(max_iter = 1)),
               class = "efa_control_input")
  expect_error(efa_kgc(cormat, eigen_type = "PCA", estimate_control = "SPSS"),
               class = "efa_control_input")
  expect_error(efa_retain(cormat, N = 500, criteria = "EKC",
                          estimate_control = rotate_control()),
               class = "efa_control_input")
})

test_that("N_FACTORS() now tunes the criteria its dots always claimed to reach", {
  # NEST and SMT never received the repacked control (the registry did not pass it on), so a
  # legacy call silently ran the default preset there. A coarse `criterion` ends the PAF
  # iterations almost at once; unlike `max_iter`, which R partial-matches to the frozen
  # `max_iter_CD` formal, it really does travel through the wrapper's dots.
  # the frozen wrapper still selects the estimator with `method`; the successor takes
  # `estimator`
  args <- list(cormat, N = 500, suitability = FALSE, criteria = "NEST",
               n_datasets_nest = 2)
  args_old <- c(args, list(method = "PAF"))
  args_new <- c(args, list(estimator = "PAF"))

  set.seed(42L)
  old <- suppressWarnings(suppressMessages(
    do.call(N_FACTORS, c(args_old, list(criterion = 0.5)))))
  set.seed(42L)
  new <- suppressWarnings(suppressMessages(
    do.call(efa_retain, c(args_new, list(estimate_control = estimate_control(criterion = 0.5))))))
  set.seed(42L)
  default <- suppressWarnings(suppressMessages(do.call(efa_retain, args_new)))

  expect_equal(old$outputs$NEST$results, new$outputs$NEST$results)
  expect_false(isTRUE(all.equal(old$outputs$NEST$results,
                                default$outputs$NEST$results)))
})

# cor_method = "poly" / "tetra": polychoric and tetrachoric correlations wired into
# the public API. The matrix estimator itself is covered in test-polychoric.R; these
# tests cover the routing through .prepare_cor_input() (EFA and the suitability /
# retention functions), the `use` handling, the EFA() bootstrap recompute, the
# tetrachoric binary assertion, and the clean rejection by the criteria whose
# reference data are continuous (CD, PARALLEL, NEST, HULL).

# The bootstrap reproducibility assertions below compare two separate invocations, so they
# use expect_equal() with an explicit tolerance rather than expect_identical(): a threaded
# BLAS (Apple's Accelerate, for one) is free to vary its GEMM reduction order between calls,
# so one function on one input can differ in the last ulp. waldo still compares S3 classes,
# names, attributes and structure exactly, so only numeric values are given slack -- a seed
# that was ignored resamples different cases entirely and still fails loudly. (A set tolerance
# does relax integer against double, which these floating-point SEs never are.)
fp_tol <- 1e-8

# Muffle only the named warning classes, letting any other (unexpected) warning
# surface so it is not silently hidden by a blanket suppressWarnings().
.muffle <- function(expr, ...) {
  classes <- c(...)
  withCallingHandlers(expr, warning = function(w) {
    if (inherits(w, classes)) invokeRestart("muffleWarning")
  })
}

# `.warn_classes()` -- which collects the class strings of every warning raised while
# evaluating an expression, so a single expected class can be asserted without
# expect_warning stopping at the first one -- lives in helper-efa_retention.R.

# A six-item ordinal set whose first pair is at a Frechet bound (its response table is the
# comonotone coupling of its own margins), so that pair is estimated at the boundary of the
# parameter space and has no asymptotic variance; the other four items are ordinary filler.
# Deterministic: a fixed contingency table expanded to raw data, plus seeded filler.
.degenerate_ordinal <- function() {
  tab <- rbind(c(41, 0), c(13, 0), c(28, 0), c(171, 47))
  pair <- do.call(rbind, lapply(which(tab > 0), function(k) {
    a <- ((k - 1L) %% nrow(tab)) + 1L
    b <- ((k - 1L) %/% nrow(tab)) + 1L
    cbind(rep(a, tab[a, b]), rep(b, tab[a, b]))
  }))
  set.seed(11)
  filler <- matrix(sample(1:4, nrow(pair) * 4L, TRUE), nrow(pair), 4L)
  out <- cbind(pair, filler)
  colnames(out) <- paste0("i", seq_len(ncol(out)))
  out
}

test_that("EFA() runs ordinal factor analysis and matches psych::fa on a polychoric matrix", {
  skip_on_cran()
  skip_if_not_installed("psych")

  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), ]
  N <- nrow(x)
  k <- 3L

  efa <- EFA(x, n_factors = k, method = "ULS", cor_method = "poly",
             rotation = "none")

  # Reference: psych's unweighted least squares factor analysis of psych's own
  # polychoric matrix, with correct = FALSE so psych applies no empty-cell continuity
  # correction, matching .polychoric(). fm = "uls" minimises the same criterion as this
  # package's ULS.
  Rp <- suppressWarnings(psych::polychoric(x, correct = FALSE, global = FALSE)$rho)
  fa_ref <- suppressWarnings(psych::fa(Rp, nfactors = k, fm = "uls",
                                       rotate = "none", n.obs = N))
  L_ref <- unclass(fa_ref$loadings)
  L <- efa$unrot_loadings

  # Compare via the reproduced common-factor correlations (L %*% t(L)), which are
  # invariant to factor sign and column order, so no fragile per-column alignment
  # is needed. The tolerance is dominated by the two polychoric matrices, which agree
  # to ~1e-4, not by the extraction, so it stays at 1e-3.
  expect_lt(max(abs(tcrossprod(L) - tcrossprod(L_ref))), 1e-3)
  expect_lt(max(abs(efa$h2 - fa_ref$communality)), 1e-3)
})

test_that("cor_method = 'tetra' equals 'poly' on binary data", {
  g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), ]
  gb <- apply(g, 2L, function(col) as.integer(col > stats::median(col)))

  # A Heywood case (and any non-PD smoothing) on this small binary example is
  # incidental to the equivalence being checked; muffle only those specific
  # warnings rather than blanket-suppressing every condition.
  e_poly  <- .muffle(EFA(gb, n_factors = 2, method = "ULS",
                         cor_method = "poly", rotation = "none"),
                     "efa_heywood", "efa_cor_smoothed")
  e_tetra <- .muffle(EFA(gb, n_factors = 2, method = "ULS",
                         cor_method = "tetra", rotation = "none"),
                     "efa_heywood", "efa_cor_smoothed")

  expect_equal(e_tetra$unrot_loadings, e_poly$unrot_loadings)
  expect_equal(e_tetra$orig_R, e_poly$orig_R)
})

test_that("cor_method = 'tetra' rejects variables with more than two categories", {
  expect_error(EFA(DOSPERT_raw, n_factors = 3, cor_method = "tetra"),
               class = "efa_cor_not_binary")
})

test_that("a Frechet-bound polychoric pair refuses DWLS and withholds the sandwich SEs", {
  x <- .degenerate_ordinal()

  # The first pair's response table is at a Frechet bound, so it has no asymptotic variance (see
  # test-polychoric.R). DWLS weights each element by the inverse of that variance, so it cannot be
  # formed and the fit is refused rather than run on a weight invented from rounding noise -- which
  # previously either dropped the pair from the solution or let it dominate, depending on the
  # platform.
  expect_error(
    suppressWarnings(
      efa_fit(x, n_factors = 2, estimator = "DWLS", cor_method = "poly", rotation = "none")),
    class = "efa_dwls_degenerate_weight")

  # The ULS sandwich consumes the same covariance. It still returns a fit -- the point estimates
  # need no asymptotic covariance -- but the standard errors are withheld rather than reported from
  # a contaminated meat matrix.
  cl <- character()
  sw <- withCallingHandlers(
    suppressMessages(efa_fit(x, n_factors = 2, estimator = "ULS", cor_method = "poly",
                             se = "sandwich", rotation = "none")),
    warning = function(w) { cl <<- c(cl, class(w)); invokeRestart("muffleWarning") })
  expect_s3_class(sw, "efa")
  expect_true(all(is.na(sw$SE$unrot_loadings)))
  expect_true("efa_se_unreliable" %in% cl)

  # The boundary pair is named, and reporting its correlation at the boundary makes this matrix
  # indefinite, so it is smoothed. Both are asserted because both are consequences of the boundary
  # estimate that a change to it would silently alter: at the old (lower) estimate this matrix was
  # positive definite and no smoothing happened.
  expect_true("efa_cor_boundary" %in% cl)
  expect_true("efa_cor_smoothed" %in% cl)

  # The degeneracy screen fires on this data, and does so on every platform: the variance reaches
  # it as NA by construction rather than as a number whose size decided the branch.
  expect_true("efa_acov_degenerate" %in% cl)

  # Clean ordinal data raises no degeneracy warning.
  cl_clean <- .warn_classes(suppressMessages(
    .prepare_cor_input(DOSPERT_raw[stats::complete.cases(DOSPERT_raw), 1:6],
                       cor_method = "poly", acov = "diag", dwls = TRUE,
                       inform_from_data = FALSE)))
  expect_false("efa_acov_degenerate" %in% cl_clean)
})

test_that("a continuity-corrected binary pair still refuses DWLS", {
  # The correction repairs the point estimate of a binary pair with an empty cell, but not its
  # asymptotic variance: the sandwich would treat the nudged counts as if they were the observed
  # data, and simulated coverage of an interval built that way is not trustworthy. So a corrected
  # pair is withheld exactly like a boundary pair, and DWLS refuses the data rather than weighting
  # by a number it cannot justify. This is the test that pins that decision.
  #
  # Six binary items that are monotone in one another, so every pair's 2x2 table has one empty
  # cell and every pair is corrected.
  x <- vapply(1:6, function(k) as.integer((0:239L) %% 7L >= k), integer(240L))
  colnames(x) <- paste0("i", 1:6)

  poly <- suppressWarnings(.polychoric(x, acov = "diag"))
  expect_true(all(poly$zero_corrected))
  expect_true(all(is.na(poly$acov)))

  expect_error(
    suppressWarnings(
      efa_fit(x, n_factors = 1, estimator = "DWLS", cor_method = "tetra", rotation = "none")),
    class = "efa_dwls_degenerate_weight")

  # ULS needs no asymptotic covariance, so the corrected matrix still fits -- which is the point
  # of the correction: these data are analysable now, where the boundary estimate made every
  # correlation 0.9999.
  cl <- character()
  fit <- withCallingHandlers(
    suppressMessages(efa_fit(x, n_factors = 1, estimator = "ULS", cor_method = "tetra",
                             rotation = "none")),
    warning = function(w) { cl <<- c(cl, class(w)); invokeRestart("muffleWarning") })
  expect_s3_class(fit, "efa")
  expect_true("efa_cor_zero_cell" %in% cl)
  expect_false("efa_cor_boundary" %in% cl)
  expect_true(all(abs(fit$orig_R[upper.tri(fit$orig_R)]) < 0.9999))
})

test_that(".warn_acov_degenerate flags an unusable asymptotic variance and passes clean ones", {
  # A variance above 1 already implies a +/- 1 SE interval as wide as the whole [-1, 1] range
  # of a correlation; non-finite and non-positive values are unusable a fortiori. Asserted on
  # supplied vectors rather than on an estimate, so the gate is exercised exactly rather than
  # through whichever side of it a fitted pair happens to land on.
  expect_warning(.warn_acov_degenerate(c(a = 0.01, b = 12, c = 0.02)),
                 class = "efa_acov_degenerate")
  expect_warning(.warn_acov_degenerate(c(a = 0.01, b = Inf, c = 0.02)),
                 class = "efa_acov_degenerate")
  expect_warning(.warn_acov_degenerate(c(a = 0.01, b = NaN, c = 0.02)),
                 class = "efa_acov_degenerate")
  expect_warning(.warn_acov_degenerate(c(a = 0.01, b = 0, c = 0.02)),
                 class = "efa_acov_degenerate")
  expect_warning(.warn_acov_degenerate(c(a = 0.01, b = -1, c = 0.02)),
                 class = "efa_acov_degenerate")

  # Ordinary variances pass, and an unnamed vector is reported positionally rather than failing
  # on the missing labels.
  expect_no_warning(.warn_acov_degenerate(c(a = 0.01, b = 0.5, c = 0.02)))
  expect_warning(.warn_acov_degenerate(c(0.01, 12, 0.02)), class = "efa_acov_degenerate")
})

test_that("the degenerate-variance warning reads correctly for one and for several pairs", {
  # Every bullet pluralises on the number of affected pairs, so both branches are recorded. The
  # first bullet opens with cli::qty(); a space after that call renders as a second space behind
  # the bullet marker, which is why it has none.
  degenerate_msg <- function(v) {
    w <- tryCatch(.warn_acov_degenerate(v), efa_acov_degenerate = function(w) w)
    expect_s3_class(w, "efa_acov_degenerate")
    conditionMessage(w)
  }
  expect_snapshot(cat(degenerate_msg(c(`v1-v2` = 0.01, `v1-v3` = -1))))
  expect_snapshot(cat(degenerate_msg(c(`v1-v2` = NA_real_, `v1-v3` = -1))))
})

test_that("both acov requests screen the same pair-labelled asymptotic variances", {
  # .prepare_cor_input() screens the variances once, before either consumer touches them, so a
  # degenerate pair is reported identically whether it reaches the DWLS weights (acov = "diag")
  # or the sandwich meat (acov = "full", whose diagonal is the same quantity). Captured from the
  # screen itself: the value that decides the warning is undetermined for a boundary pair, so
  # what is asserted is the routing -- that both paths screen, and screen the same labelled
  # diagonal -- on clean data where the screen is silent either way.
  seen <- list()
  local_mocked_bindings(
    .warn_acov_degenerate = function(acov_diag, labels = names(acov_diag)) {
      seen[[length(seen) + 1L]] <<- acov_diag
      invisible(NULL)
    }
  )
  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), 1:5]
  labels <- apply(utils::combn(colnames(x), 2L), 2L, paste, collapse = "-")

  suppressMessages(.prepare_cor_input(x, cor_method = "poly", acov = "diag", dwls = TRUE,
                                      inform_from_data = FALSE))
  suppressMessages(.prepare_cor_input(x, cor_method = "poly", acov = "full", dwls = TRUE,
                                      inform_from_data = FALSE))

  expect_length(seen, 2L)
  expect_named(seen[[1L]], labels)
  expect_named(seen[[2L]], labels)
  # The "full" path reaches the screen through diag(Gamma), which accumulates the same per-cell
  # influences by a different route (scatter then crossprod) than the "diag" path's direct sum.
  expect_equal(seen[[1L]], seen[[2L]], tolerance = 1e-6)
})

test_that("cor_method = 'poly' rejects unordered factor / character columns through efa_fit", {
  set.seed(123)
  N <- 400L
  z1 <- stats::rnorm(N)
  z2 <- 0.7 * z1 + sqrt(1 - 0.7^2) * stats::rnorm(N)
  br <- stats::qnorm(c(.33, .66))
  lab <- c("low", "mid", "high")
  df <- data.frame(a = factor(lab[findInterval(z1, br) + 1L]),   # unordered
                   b = factor(lab[findInterval(z2, br) + 1L]))
  expect_error(
    suppressMessages(efa_fit(df, n_factors = 1, cor_method = "poly")),
    class = "efa_cor_unordered_factor")
})

test_that("method = 'DWLS' requires a polychoric asymptotic covariance", {
  # a correlation matrix carries no raw data to estimate an asymptotic covariance from
  expect_error(EFA(test_models$baseline$cormat, n_factors = 3, method = "DWLS", N = 500),
               class = "efa_dwls_no_acov")
  # a continuous correlation method produces no polychoric asymptotic covariance
  expect_error(EFA(DOSPERT_raw, n_factors = 3, method = "DWLS", cor_method = "pearson"),
               class = "efa_dwls_no_acov")
  # the abort fires before fitting, regardless of bootstrap settings
  expect_error(EFA(test_models$baseline$cormat, n_factors = 3, method = "DWLS", N = 500,
                   se = "np-boot"),
               class = "efa_dwls_no_acov")
})

test_that("KMO() and BARTLETT() honour cor_method = 'poly'", {
  # Six items rather than all thirty: what is asserted is the routing, an exact identity
  # that holds for any item set, and six items make each polychoric matrix 15 pairs instead
  # of 435. The rows are still the complete cases of the full data, so N is unchanged.
  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), 1:6]
  N <- nrow(x)
  Rp <- .polychoric(x)$R

  # cor_method = "poly" must produce exactly the result of passing that polychoric
  # matrix in directly: this pins the routing through .prepare_cor_input(). The raw
  # call infers N = nrow(x), the same N supplied with the matrix.
  expect_equal(KMO(x, cor_method = "poly")$KMO, KMO(Rp)$KMO)
  expect_equal(BARTLETT(x, cor_method = "poly")$chisq,
               BARTLETT(Rp, N = N)$chisq)
})

test_that("KMO() and BARTLETT() with cor_method = 'poly' match an external polychoric reference", {
  skip_on_cran()
  skip_if_not_installed("psych")

  x <- DOSPERT_raw[stats::complete.cases(DOSPERT_raw), ]
  N <- nrow(x)
  Rp <- suppressWarnings(psych::polychoric(x, correct = FALSE, global = FALSE)$rho)

  # The two polychoric matrices agree to ~1e-4, so the suitability statistics agree
  # to well within 1e-2.
  expect_equal(KMO(x, cor_method = "poly")$KMO, psych::KMO(Rp)$MSA,
               tolerance = 1e-2)
  expect_equal(BARTLETT(x, cor_method = "poly")$chisq,
               psych::cortest.bartlett(Rp, n = N)$chisq,
               tolerance = 1e-2, ignore_attr = TRUE)
})

test_that("the simulation-based criteria reject poly / tetra with a classed condition", {
  expect_error(CD(DOSPERT_raw, cor_method = "poly"),
               class = "efa_cor_method_unsupported")
  expect_error(PARALLEL(DOSPERT_raw, cor_method = "tetra"),
               class = "efa_cor_method_unsupported")
  expect_error(NEST(DOSPERT_raw, N = 200, cor_method = "poly"),
               class = "efa_cor_method_unsupported")
  # HULL derives its factor-search bound from an internal parallel analysis, so it
  # rejects poly/tetra for the same reason.
  expect_error(HULL(DOSPERT_raw, N = 200, cor_method = "tetra"),
               class = "efa_cor_method_unsupported")
})

test_that("the poly rejection does not steer an ordinal analysis to Kendall's tau", {
  # The remedy is read by someone who reached for a polychoric correlation because the
  # items are ordinal. Kendall's tau is not on the Pearson scale the factor model
  # parameterises (for bivariate-normal data tau = (2/pi) asin(rho)), so it is the most
  # attenuated of the three and must not be offered here. The message is rendered from
  # every caller of the shared helper, so two of the five are checked: one continuous
  # reference criterion and efa_smt(), which reaches it with its own reason text.
  # `error =`, not `condition =`: these functions also emit an informational message about
  # computing correlations from raw data, which a condition handler would catch instead.
  remedy <- function(expr) {
    cnd <- tryCatch(expr, error = function(e) e)
    expect_s3_class(cnd, "efa_cor_method_unsupported")
    cli::ansi_strip(paste(conditionMessage(cnd), collapse = " "))
  }
  for (msg in c(remedy(efa_parallel(DOSPERT_raw, cor_method = "poly")),
                remedy(efa_smt(DOSPERT_raw, cor_method = "poly")))) {
    expect_match(msg, "pearson", fixed = TRUE)
    expect_match(msg, "spearman", fixed = TRUE)
    expect_no_match(msg, "kendall", fixed = TRUE)
  }
})

test_that("cor_method = 'poly' honours `use` for missing data", {
  g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), ]
  gm <- g
  gm[1:5, 1] <- NA

  # all.obs and everything do not delete missing data, so they error on it,
  # matching the outcome of stats::cor() for the Pearson path.
  expect_error(KMO(gm, cor_method = "poly", use = "all.obs"),
               class = "efa_cor_na")
  expect_error(KMO(gm, cor_method = "poly", use = "everything"),
               class = "efa_cor_na")

  # complete.obs listwise-deletes first, so the matrix equals the polychoric of
  # the complete-case data.
  R_listwise <- .polychoric(gm[stats::complete.cases(gm), ])$R
  expect_equal(KMO(gm, cor_method = "poly", use = "complete.obs")$KMO,
               KMO(R_listwise)$KMO)
})

test_that("cor_method = 'poly' uses pairwise-complete thresholds under missing data", {
  skip_on_cran()
  skip_if_not_installed("polycor")

  set.seed(202)
  N <- 2000
  L <- chol(matrix(c(1, .6, .6, 1), 2L))
  Z <- matrix(stats::rnorm(2 * N), N, 2L) %*% L
  a <- findInterval(Z[, 1], stats::qnorm(c(.3, .6)))
  b <- findInterval(Z[, 2], stats::qnorm(c(.25, .55, .8)))
  # Missing-at-random: a is dropped more often where b is high, so b's pairwise-complete
  # marginal differs from its full-column marginal. A pairwise polychoric must take its
  # thresholds from each pair's own complete cases (as polycor does); using the full-column
  # marginals instead would bias this estimate by ~0.06.
  am <- a; am[(b >= 2) & (stats::runif(N) < 0.7)] <- NA
  ours <- .polychoric(cbind(am, b))$R[1, 2]
  ref <- suppressWarnings(polycor::polychor(am, b, ML = FALSE))
  expect_equal(ours, ref, tolerance = 1e-3)
})

test_that("an asymptotic covariance over pairwise data reports the listwise override", {
  g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), ]
  gm <- g; gm[1:5, 1] <- NA

  # With missing data, an asymptotic covariance (here the DWLS weights) forces listwise
  # deletion even though `use` asks for pairwise-complete estimation; that override is reported.
  expect_message(
    suppressWarnings(
      .prepare_cor_input(gm, cor_method = "poly", acov = "diag",
                         use = "pairwise.complete.obs", inform_from_data = FALSE),
      classes = "efa_cor_sparse_cells"
    ),
    class = "efa_acov_listwise"
  )
  # Complete data needs no override, so nothing is reported.
  expect_no_message(
    suppressWarnings(
      .prepare_cor_input(g, cor_method = "poly", acov = "diag",
                         use = "pairwise.complete.obs", inform_from_data = FALSE),
      classes = "efa_cor_sparse_cells"
    ),
    class = "efa_acov_listwise"
  )
  # A diagonal (DWLS-weight) covariance is an ordinal construct; requesting it for a Pearson
  # correlation is rejected rather than silently producing no weights.
  expect_error(
    .prepare_cor_input(gm, cor_method = "pearson", acov = "diag",
                       use = "pairwise.complete.obs", inform_from_data = FALSE),
    class = "efa_acov_unsupported"
  )
})

test_that("an acov requested for a correlation matrix is ignored with a warning", {
  set.seed(1)
  R <- stats::cor(matrix(stats::rnorm(200L), ncol = 4L))
  expect_warning(
    .prepare_cor_input(R, N = 50, acov = "full", inform_from_data = FALSE),
    class = "efa_acov_ignored"
  )
})

test_that("an unsupported acov/cor_method is rejected before any listwise notice", {
  set.seed(1)
  dat <- matrix(stats::rnorm(150L), ncol = 3L)
  dat[1:3, 1L] <- NA
  # full + a rank correlation is unsupported; with missing data the abort must precede (and
  # replace) the listwise-override notice rather than follow it.
  expect_error(
    .prepare_cor_input(dat, cor_method = "spearman", acov = "full",
                       use = "pairwise.complete.obs", inform_from_data = FALSE),
    class = "efa_acov_unsupported"
  )
})

test_that("N_FACTORS skips reference-based criteria under poly with an informative note", {
  classes <- character()
  withCallingHandlers(
    N_FACTORS(GRiPS_raw, criteria = c("EKC", "HULL"), cor_method = "poly",
              method = "ULS", suitability = FALSE),
    warning = function(w) {
      classes <<- c(classes, class(w))
      invokeRestart("muffleWarning")
    }
  )
  # HULL is skipped with the dedicated 'skipped' class, not reported as a generic
  # failure; EKC still runs on the polychoric matrix.
  expect_true("efa_criterion_skipped" %in% classes)
  expect_false("efa_criterion_failed" %in% classes)
})

test_that("N_FACTORS skips SMT under poly because its normal-theory test is invalid", {
  classes <- character()
  out <- withCallingHandlers(
    N_FACTORS(GRiPS_raw, criteria = c("EKC", "SMT"), cor_method = "poly",
              method = "ULS", suitability = FALSE),
    warning = function(w) {
      classes <<- c(classes, class(w))
      invokeRestart("muffleWarning")
    }
  )
  # SMT's sequential model tests rest on the normal-theory ML chi-square, which is
  # not valid for polychoric/tetrachoric correlations, so it is skipped (not run on
  # an inappropriate matrix); EKC, a descriptive eigenvalue rule, still runs.
  expect_true("efa_criterion_skipped" %in% classes)
  expect_true("SMT" %in% names(out$not_run))
  expect_false("SMT" %in% names(out$outputs))
  expect_true("EKC" %in% names(out$outputs))
})

test_that("the polychoric bootstrap is reproducible and positive definite under resampling", {
  g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), ]

  b1 <- EFA(g, n_factors = 1, method = "ULS", cor_method = "poly",
            se = "np-boot", b_boot = 6, seed = 42)
  b2 <- EFA(g, n_factors = 1, method = "ULS", cor_method = "poly",
            se = "np-boot", b_boot = 6, seed = 42)

  # Same seed -> the same bootstrap SEs, independent of how many replicates were fit.
  expect_equal(b1$SE$unrot_loadings, b2$SE$unrot_loadings, tolerance = fp_tol)
  # Replicate matrices are recomputed per resample (degenerate ones are dropped at
  # the fit), so the surviving replicates yield finite SEs.
  expect_true(all(is.finite(b1$SE$unrot_loadings)))
})

test_that("the DWLS polychoric bootstrap is reproducible and positive definite under resampling", {
  g <- GRiPS_raw[stats::complete.cases(GRiPS_raw), ]

  b1 <- suppressWarnings(EFA(g, n_factors = 1, method = "DWLS", cor_method = "poly",
                             se = "np-boot", b_boot = 6, seed = 42))
  b2 <- suppressWarnings(EFA(g, n_factors = 1, method = "DWLS", cor_method = "poly",
                             se = "np-boot", b_boot = 6, seed = 42))

  # Same seed -> the same bootstrap SEs; each replicate recomputes its own polychoric
  # matrix and diagonal-ACOV weights, and the surviving replicates yield finite SEs.
  expect_equal(b1$SE$unrot_loadings, b2$SE$unrot_loadings, tolerance = fp_tol)
  expect_true(all(is.finite(b1$SE$unrot_loadings)))
})

test_that("the polychoric bootstrap on DOSPERT_raw completes (timed)", {
  skip_on_cran()
  skip_if_not_slow()

  b_boot <- 200L
  t <- system.time(
    fit <- EFA(DOSPERT_raw, n_factors = 3, method = "ULS", cor_method = "poly",
               se = "np-boot", b_boot = b_boot, seed = 1)
  )
  elapsed <- unname(t["elapsed"])
  cli::cli_inform(
    "Polychoric bootstrap: {b_boot} replicates on DOSPERT_raw in {round(elapsed, 1)}s ({round(elapsed / b_boot, 3)}s/replicate).")
  expect_true(all(is.finite(fit$SE$unrot_loadings)))
})

test_that("efa_retention plot methods return ggplot objects", {
  skip_if_not_slow()
  ekc <- EKC(test_models$baseline$cormat, N = 500)
  p_ekc <- plot(ekc)
  expect_s3_class(p_ekc, "ggplot")
  expect_equal(ekc$results[[1]]$plot_type, "eigen")

  hull <- HULL(test_models$baseline$cormat, N = 500, method = "ML")
  p_hull <- plot(hull)
  expect_s3_class(p_hull, "ggplot")
  expect_equal(hull$results[[1]]$plot_type, "hull")

  kgc <- KGC(test_models$baseline$cormat)
  p_kgc <- plot(kgc)
  expect_s3_class(p_kgc, "ggplot")
  expect_equal(.retention_record(kgc, "PCA")$plot_type, "eigen")

  # CD is stochastic, so only smoke-test that the plot builds with its custom
  # y-axis label (no vdiffr baseline)
  set.seed(123)
  cd <- CD(GRiPS_raw, N_pop = 1000, N_samples = 100)
  p_cd <- plot(cd)
  expect_s3_class(p_cd, "ggplot")
  expect_equal(.retention_record(cd, "CD")$y_label, "RMSE eigenvalues")

  scree <- SCREE(test_models$baseline$cormat)
  p_scree <- plot(scree)
  expect_s3_class(p_scree, "ggplot")

  # PARALLEL: smoke-test only (no vdiffr baseline) because the simulated
  # reference lines vary with the RNG state and the future plan's chunking,
  # so an SVG baseline would not be portable
  pa <- PARALLEL(test_models$baseline$cormat, N = 500, eigen_type = c("PCA", "SMC"))
  p_pa <- plot(pa)
  expect_s3_class(p_pa, "ggplot")

  # no real data: the plot shows only the dashed reference series
  pa_nodat <- PARALLEL(N = 20, n_vars = 5, eigen_type = "PCA")
  expect_s3_class(plot(pa_nodat), "ggplot")
})

test_that("plot-less criteria return NULL with a message", {
  for (obj in list(MAP(test_models$baseline$cormat),
                   SMT(test_models$baseline$cormat, N = 500))) {
    expect_message(p <- plot(obj), class = "efa_no_plot")
    expect_null(p)
  }
})

test_that("NEST plots its empirical eigenvalues against the reference series", {
  # few simulated datasets: the plot only needs a well-formed record
  set.seed(42)
  nest <- efa_nest(test_models$baseline$cormat, N = 500, n_datasets = 50)

  expect_s3_class(plot(nest), "ggplot")

  rec <- .retention_record(nest, "NEST")
  expect_equal(rec$plot_type, "eigen")
  # the shared eigenvalue plotter binds the series into one data frame, so the
  # reference has to be as long as the empirical eigenvalues
  expect_length(rec$reference, length(rec$y))
  # the retained solution is marked, as in the other eigenvalue plots (the record
  # carries no highlight when nothing is retained, so pin that there is one)
  expect_gte(nest$n_factors[["NEST"]], 1)
  expect_equal(rec$highlight, nest$n_factors[["NEST"]])
})

test_that("eigen plot of an empty record returns NULL with a message", {
  # e.g. CD on a tiny dataset that suggests 0 factors -> empty x/y record
  obj <- .new_efa_retention(
    "CD",
    results = list(list(name = "CD", label = "Suggested number of factors",
                        n_factors = 0, plot_type = "eigen",
                        x = integer(0), y = numeric(0))),
    settings = list()
  )
  expect_message(p <- plot(obj), class = "efa_no_plot")
  expect_null(p)
})

test_that("efa_retain reports when none of its criteria has a plot", {
  # plot.efa_retain() drops the per-criterion NULLs; with MAP and SMT the only
  # criteria run, nothing is left to return.
  nf <- efa_retain(test_models$baseline$cormat, N = 500, suitability = FALSE,
                   criteria = c("MAP", "SMT"))
  expect_message(p <- plot(nf), class = "efa_no_plot")
  expect_null(p)
})

test_that("EKC eigen plot is visually stable", {
  skip_if_not_installed("vdiffr")

  ekc <- EKC(test_models$baseline$cormat, N = 500)
  vdiffr::expect_doppelganger("EKC eigen plot", plot(ekc))
})

test_that("HULL hull plot is visually stable", {
  skip_if_not_installed("vdiffr")

  hull <- HULL(test_models$baseline$cormat, N = 500, method = "ML")
  vdiffr::expect_doppelganger("HULL hull plot", plot(hull))
})

test_that("KGC eigen plot is visually stable", {
  skip_if_not_installed("vdiffr")

  kgc <- KGC(test_models$baseline$cormat)
  vdiffr::expect_doppelganger("KGC eigen plot", plot(kgc))
})

test_that("SCREE eigen plot is visually stable", {
  skip_if_not_installed("vdiffr")

  scree <- SCREE(test_models$baseline$cormat)
  vdiffr::expect_doppelganger("SCREE eigen plot", plot(scree))
})

test_that("plot.efa_retain's deterministic members are visually stable", {
  skip_if_not_installed("vdiffr")

  # plot.efa_retain() dispatches one plot per plottable criterion that was run. Only the
  # deterministic ones carry a baseline: CD, PARALLEL, and NEST simulate, so their
  # reference series depend on the RNG state and no SVG baseline would be portable.
  nf <- efa_retain(test_models$baseline$cormat, N = 500, suitability = FALSE,
                   criteria = c("EKC", "KGC", "SCREE"), eigen_type_other = "PCA")
  p <- plot(nf)

  expect_named(p, c("EKC", "KGC", "SCREE"))
  for (id in names(p)) {
    vdiffr::expect_doppelganger(paste("efa_retain", id, "plot"), p[[id]])
  }
})

ekc_cor <- efa_ekc(test_models$baseline$cormat, N = 500)
ekc_raw <- efa_ekc(GRiPS_raw)

test_that("output class and dimensions are correct", {
  expect_s3_class(ekc_cor, "efa_retention")
  expect_length(ekc_cor, 6)
  expect_s3_class(ekc_raw, "efa_retention")
  expect_length(ekc_raw, 6)

  expect_equal(unname(ekc_cor$criterion["id"]), "EKC")
  expect_length(ekc_cor$results, 1)
  expect_named(ekc_cor$n_factors, "BvA2017")
  expect_equal(ekc_cor$results[[1]]$plot_type, "eigen")
})


test_that("found eigenvalues are correct", {
  expect_equal(sum(ekc_cor$results[[1]]$y),
               ncol(test_models$baseline$cormat))
  expect_equal(sum(ekc_raw$results[[1]]$y), ncol(GRiPS_raw))
  expect_length(ekc_cor$results[[1]]$y,
                ncol(test_models$baseline$cormat))
  expect_length(ekc_raw$results[[1]]$y, ncol(GRiPS_raw))
})

test_that("reference eigenvalues are correct", {
  expect_equal(ekc_cor$results[[1]]$reference[floor(ncol(test_models$baseline$cormat) / 2)], 1)
  expect_equal(ekc_raw$results[[1]]$reference[floor(ncol(GRiPS_raw) / 2)], 1)
  expect_length(ekc_cor$results[[1]]$reference,
                length(ekc_cor$results[[1]]$y))
  expect_length(ekc_raw$results[[1]]$reference, length(ekc_raw$results[[1]]$y))
})

test_that("identified number of factors is correct", {
  expect_equal(ekc_cor$n_factors[["BvA2017"]], 3)
  expect_equal(ekc_raw$n_factors[["BvA2017"]], 1)
})

test_that("the BvA2017 reference series reproduces the published equation", {
  # Braeken and van Assen (2017, p. 454) define the j-th reference eigenvalue as
  # max(((J - sum_{i < j} lambda_i) / (J - j + 1)) * (1 + sqrt(J / N))^2, 1),
  # transcribed here as an explicit loop -- an independent expression of the
  # equation the vectorised cumsum form in efa_ekc() implements. The equation holds
  # for any N, so one nominal value is used for both matrices.
  N <- 500
  for (cmat in list(test_models$baseline$cormat, stats::cor(GRiPS_raw))) {
    lambda <- eigen(cmat, symmetric = TRUE, only.values = TRUE)$values
    J <- ncol(cmat)

    ref <- numeric(J)
    for (j in seq_len(J)) {
      used <- if (j == 1) 0 else sum(lambda[seq_len(j - 1)])
      ref[j] <- max((J - used) / (J - j + 1) * (1 + sqrt(J / N))^2, 1)
    }

    out <- efa_ekc(cmat, N = N)
    expect_equal(.retention_record(out, "BvA2017")$reference, ref)
    # the paper's stopping rule: retain the leading run of eigenvalues above their
    # reference value
    expect_equal(out$n_factors[["BvA2017"]], sum(cumprod(lambda > ref)))
  }
})

test_that("the retained count is always an integer", {
  # The count is derived with cumprod(), which coerces the comparison to double, so pin
  # that it comes back in the integer storage mode a count belongs in. An empty solution
  # is checked alongside a retained one because a single expression has to cover both.
  # (How the count is rendered no longer depends on this: the report and the plot label
  # both go through .retention_count(); see test-efa_retention.R.)
  some <- efa_ekc(test_models$baseline$cormat, N = 500)
  expect_identical(some$results[[1]]$n_factors, 3L)
  expect_identical(some$results[[1]]$highlight, 3L)

  # a sample size just above the number of variables lifts the reference series above every
  # eigenvalue, so nothing is retained
  none <- efa_ekc(test_models$case_1a$cormat, N = 11)
  expect_identical(none$results[[1]]$n_factors, 0L)
  expect_null(none$results[[1]]$highlight)
})

test_that("the deprecated type argument warns and is ignored", {
  cmat <- test_models$baseline$cormat

  expect_warning(efa_ekc(cmat, N = 500, type = "AM2019"),
                 class = "lifecycle_warning_deprecated")
  expect_warning(efa_ekc(cmat, N = 500, type = c("BvA2017", "AM2019")),
                 class = "lifecycle_warning_deprecated")
  expect_warning(efa_retain(cmat, N = 500, criteria = "EKC", suitability = FALSE,
                            ekc_type = "AM2019"),
                 class = "lifecycle_warning_deprecated")

  # ignored: whatever is passed, the result is the one the argument-free call gives
  deprecated <- suppressWarnings(efa_ekc(cmat, N = 500, type = "AM2019"))
  expect_equal(deprecated, ekc_cor)

  # the criteria themselves never trip the deprecation, and the argument-free calls
  # stay silent
  expect_no_condition(efa_ekc(cmat, N = 500))
  expect_no_condition(efa_retain(cmat, N = 500, criteria = "EKC", suitability = FALSE))

  # the superseded wrappers keep both arguments, and a call that does not name one stays
  # silent
  expect_no_condition(EKC(cmat, N = 500))
  expect_warning(EKC(cmat, N = 500, type = "AM2019"),
                 class = "lifecycle_warning_deprecated")
  expect_no_condition(N_FACTORS(cmat, N = 500, criteria = "EKC", suitability = FALSE))
  expect_warning(N_FACTORS(cmat, N = 500, criteria = "EKC", suitability = FALSE,
                           ekc_type = "AM2019"),
                 class = "lifecycle_warning_deprecated")

  # a wrapper that warns but still forwards a value would pass every assertion above, so
  # pin what the wrappers return as well
  expect_equal(suppressWarnings(EKC(cmat, N = 500, type = "AM2019")),
               EKC(cmat, N = 500))
  expect_equal(
    suppressWarnings(N_FACTORS(cmat, N = 500, criteria = "EKC", suitability = FALSE,
                               ekc_type = "AM2019"))$n_factors,
    N_FACTORS(cmat, N = 500, criteria = "EKC", suitability = FALSE)$n_factors
  )
})

test_that("the deprecation warning repeats and is not attributed to the package", {
  skip_if_not_installed("withr")
  # lifecycle decides from the calling environment whether a user named the deprecated
  # argument, or whether a package named it in its own internals. Only the second case is
  # rate limited, and only it gets a footer that asks the reader to report a bug against
  # the package. Every route into the warning enters from this package, the direct call as
  # much as the one through a superseded wrapper, so without the caller EFAtools declares
  # all of them would be read as the second case.
  #
  # lifecycle exempts the package under test from that rule, which hides the difference
  # from the rest of the suite. Clearing TESTTHAT_PKG removes the exemption, so the calls
  # below take the path a user's session takes. The repeat is what separates the two: a
  # rate-limited warning is signalled at most once.
  withr::local_envvar(c(TESTTHAT_PKG = ""))
  cmat <- test_models$baseline$cormat

  count_two <- function(call) {
    fired <- 0L
    for (i in 1:2) {
      withCallingHandlers(call(), lifecycle_warning_deprecated = function(w) {
        fired <<- fired + 1L
        invokeRestart("muffleWarning")
      })
    }
    fired
  }

  # both routes, because both resolve to this package
  expect_identical(count_two(function() efa_ekc(cmat, N = 500, type = "AM2019")), 2L)
  expect_identical(count_two(function() EKC(cmat, N = 500, type = "AM2019")), 2L)
})

test_that("errors are thrown correctly", {
  expect_error(efa_ekc(1:5), class = "efa_input_not_matrix")
  expect_error(efa_ekc(test_models$baseline$cormat), class = "efa_n_required")
  expect_message(efa_ekc(GRiPS_raw), class = "efa_cor_from_data")
  expect_warning(efa_ekc(GRiPS_raw, N = 20), class = "efa_n_from_data")
  expect_error(efa_ekc(sing_raw), class = "efa_cor_singular")
  expect_error(efa_ekc(sing_cor, N = sing_N), class = "efa_cor_singular")
  expect_warning(efa_ekc(cor_nposdef, N = 20), class = "efa_cor_smoothed")
})

test_that("EKC requires more observations than variables", {
  # the EKC reference series (1 + sqrt(J / N))^2 is only defined above that boundary
  R8 <- stats::cor(GRiPS_raw)  # 8 variables
  expect_error(efa_ekc(R8, N = 8), class = "efa_n_too_small")
  expect_s3_class(efa_ekc(R8, N = 9), "efa_retention")
})

test_that("settings are returned correctly", {
  expect_named(ekc_cor$settings, c("use", "cor_method", "N"))
  expect_named(ekc_raw$settings, c("use", "cor_method", "N"))

  expect_equal(ekc_cor$settings$N, 500)
  expect_equal(ekc_raw$settings$N, 810)

  expect_equal(ekc_cor$settings$use, "pairwise.complete.obs")
  expect_equal(ekc_raw$settings$use, "pairwise.complete.obs")

  expect_equal(ekc_cor$settings$cor_method, "pearson")
  expect_equal(ekc_raw$settings$cor_method, "pearson")
})

rm(ekc_cor, ekc_raw)

#' Empirical Kaiser criterion
#'
#' The empirical Kaiser criterion incorporates random sampling variations of the
#' eigenvalues from the Kaiser-Guttman criterion ([efa_kgc()]; see Auerswald &
#' Moshagen, 2019; Braeken & van Assen, 2017). The implementation follows Braeken
#' and van Assen (2017).
#'
#' @inheritParams efa_kgc
#' @param N numeric. The number of observations. Only needed if x is a correlation
#'  matrix. Must be larger than the number of variables.
#' @param cor_method character. Correlation computed from raw data: `"pearson"`,
#'   `"spearman"`, or `"kendall"` (passed to [stats::cor()]), or `"poly"` /
#'   `"tetra"` for polychoric / tetrachoric correlations of ordinal / binary data
#'   (a two-step estimator). Default is `"pearson"`. Note that the EKC reference
#'   values rest on the Marchenko-Pastur law for the eigenvalues of a sample
#'   correlation matrix of independent variables, which assumes the sampling
#'   behaviour of product-moment correlations; with `"poly"` / `"tetra"` (and, to a
#'   lesser degree, the rank-based methods) the reference series is therefore an
#'   approximation.
#' @param type `r lifecycle::badge("deprecated")` Accepted and ignored. It selected
#'   between two ways to compute the reference values. The `"AM2019"` reference values
#'   do not depend on the observed eigenvalues, so they do not apply the empirical
#'   correction that defines the criterion, and they are no longer computed.
#'
#' @details The Kaiser-Guttman criterion was defined with the intend that a factor
#'  should only be extracted if it explains at least as much variance as a single
#'  factor (see [efa_kgc()]). However, this only applies to population-level
#'  correlation matrices. Due to sampling variation, the KGC strongly overestimates
#'  the number of factors to retrieve (e.g., Zwick & Velicer, 1986). To account
#'  for this and to introduce a factor retention method that performs well with
#'  small number of indicators and correlated factors (cases where the performance
#'  of parallel analysis, see [efa_parallel()], is known to deteriorate)
#'  Braeken and van Assen (2017) introduced the empirical Kaiser criterion in
#'  which a series of reference eigenvalues is created as a function of the
#'  variables-to-sample-size ratio and the observed eigenvalues.
#'
#'  Braeken and van Assen (2017) showed that "(a) EKC performs about as well as
#'  parallel analysis for data arising from the null, 1-factor, or orthogonal
#'  factors model; and (b) clearly outperforms parallel analysis for the specific
#'  case of oblique factors, particularly whenever factor intercorrelation is
#'  moderate to high and the number of variables per factor is small, which is
#'  characteristic of many applications these days" (p.463-464).
#'
#' @returns An object of class `efa_retention` (see [print.efa_retention()] and
#'   [plot.efa_retention()] for the print and plot methods). Its main fields are:
#' \item{n_factors}{A numeric vector of length one, named `"BvA2017"`, with the
#'   suggested number of factors. The factors up to the first observed eigenvalue
#'   that fails to exceed its reference value are retained. The "all-exceed"
#'   convention of parallel analysis ([efa_parallel()]), which retains all *J* factors
#'   when no such crossing is found, cannot be reached here: the reference values are
#'   never below 1, while the eigenvalues of a correlation matrix sum to *J* and are
#'   sorted downwards, so the last of them is never above 1.}
#' \item{results}{A list with one record, holding the eigenvalues, the reference
#'   eigenvalues, and the retained solution used for printing and plotting.}
#' \item{settings}{A list with the settings used.}
#'
#' @source Auerswald, M., & Moshagen, M. (2019). How to determine the number of
#' factors to retain in exploratory factor analysis: A comparison of extraction
#' methods under realistic conditions. Psychological Methods, 24(4), 468–491.
#' https://doi.org/10.1037/met0000200
#'
#' @source Braeken, J., & van Assen, M. A. (2017). An empirical Kaiser criterion.
#' Psychological Methods, 22, 450 – 466. https://doi.org/10.1037/met0000074
#'
#' @source Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
#' determining the number of components to retain. Psychological Bulletin, 99,
#' 432–442. https://doi.org/10.1037/0033-2909.99.3.432
#'
#' @family factor retention criteria
#'
#' @seealso [efa_retain()] as a wrapper function for this and the other factor
#'   retention criteria.
#' @export
#'
#' @examples
#' efa_ekc(test_models$baseline$cormat, N = 500)
efa_ekc <- function(x, N = NA,
                use = c("pairwise.complete.obs", "all.obs",
                           "complete.obs", "everything",
                           "na.or.complete"),
                cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
                type = lifecycle::deprecated()) {

  # Perform argument checks
  .assert_cor_input(x)

  checkmate::assert_count(N, na.ok = TRUE)
  use <- .match_arg_ci(use)
  cor_method <- .match_arg_ci(cor_method)

  if (lifecycle::is_present(type)) .deprecate_ekc_type("efa_ekc(type)")

  # Detect or compute the correlation matrix, check it, and smooth it if needed
  prep <- .prepare_cor_input(x, N = N, use = use, cor_method = cor_method,
                             N_policy = "required")
  R <- prep$R
  N <- prep$N
  .assert_n_gt_vars(N, ncol(R))

  # eigenvalues of the correlation matrix
  lambda <- eigen(R, symmetric = TRUE, only.values = TRUE)$values
  J <- ncol(R)

  ### implementation in Braeken & van Assen, 2017. An Empirical Kaiser
  ### Criterion. Psychological Methods, 22(3). pp. 450-466
  ### Calculation based on p. 454

  # lup: asymptotic max sample eigenvalue under the null model, used as the
  #      first reference eigenvalue
  lup <- (1 + sqrt(J / N))^2

  # correction factor: the variance the preceding eigenvalues do not explain, divided
  # over the factors that remain
  correction_factor <- c(J, (J - cumsum(lambda))[-J]) / (J:1)

  # Unrestricted EKC reference values
  l_REF <- lup * correction_factor

  # Restricted EKC reference values
  l_EKC <- l_REF
  l_EKC[l_REF < 1] <- 1

  # Number of factors to retain: the length of the leading run of eigenvalues above their
  # reference value. cumprod() is 1 across that run and 0 from the first failure onwards,
  # so its sum is the length of the run.
  #
  # cumprod() coerces the logical comparison to double, so the count is cast back to the
  # integer storage mode a count belongs in.
  n_factors <- as.integer(sum(cumprod(lambda > l_EKC)))

  out <- .new_efa_retention(
    "EKC",
    results = list(list(
      name = "BvA2017",
      label = "Braeken & van Assen (2017)",
      n_factors = n_factors,
      plot_type = "eigen",
      x = seq_along(lambda),
      y = lambda,
      reference = l_EKC,
      threshold = NULL,
      highlight = if (n_factors >= 1) n_factors else NULL
    )),
    settings = list(use = use, cor_method = cor_method, N = N)
  )

  return(out)

}

#' Woodcock Johnson IV: ages 3 to 5
#'
#' A list containing the bivariate correlations (N = 435) of the 29
#' cognitive and achievement subtests from the WJ IV for 3- to 5-year-olds from
#' the standardization sample obtained from
#' the WJ IV technical Manual (McGrew, LaForte, & Schrank, 2014).
#' Tables are reproduced with permission from the publisher.
#'
#'
#' @format A list of 2 with elements "cormat" (29 x 29 matrix of bivariate correlations)
#' and "N" (scalar). The correlation matrix contains the following variables:
#' \describe{
#'   \item{ORLVOC}{(numeric) - Oral Vocabulary.}
#'   \item{VRBATN}{(numeric) - Verbal Attention.}
#'   \item{LETPAT}{(numeric) - Phonological Processing.}
#'   \item{STYREC}{(numeric) - Story Recall.}
#'   \item{VISUAL}{(numeric) - Visualization.}
#'   \item{GENINF}{(numeric) - General Information.}
#'   \item{CONFRM}{(numeric) - Concept Formation.}
#'   \item{NUMREV}{(numeric) - Numbers Reversed.}
#'   \item{NUMPAT}{(numeric) - Number-Pattern Matching.}
#'   \item{NWDREP}{(numeric) - Nonword Repetition.}
#'   \item{VAL}{(numeric) - Visual-Auditory Learning.}
#'   \item{PICREC}{(numeric) - Picture Recognition.}
#'   \item{MEMWRD}{(numeric) - Memory for Words.}
#'   \item{PICVOC}{(numeric) - Picture Vocabulary.}
#'   \item{ORLCMP}{(numeric) - Oral Comprehension.}
#'   \item{SEGMNT}{(numeric) - Segmentation.}
#'   \item{RPCNAM}{(numeric) - Rapid Picture Naming.}
#'   \item{SENREP}{(numeric) - Sentence Repetition.}
#'   \item{UNDDIR}{(numeric) - Understanding Directions.}
#'   \item{SNDBLN}{(numeric) - Sound Blending.}
#'   \item{RETFLU}{(numeric) - Retrieval Fluency.}
#'   \item{SNDAWR}{(numeric) - Sound Awareness.}
#'   \item{LWIDNT}{(numeric) - Letter-Word Identification.}
#'   \item{APPROB}{(numeric) - Applied Problems.}
#'   \item{SPELL}{(numeric) - Spelling.}
#'   \item{PSGCMP}{(numeric) - Passage Comprehension.}
#'   \item{SCI}{(numeric) - Science.}
#'   \item{SOC}{(numeric) - Social Studies.}
#'   \item{HUM}{(numeric) - Humanities.}
#'  }
#' @source McGrew, K. S., LaForte, E. M., & Schrank, F. A. (2014). Technical
#'  Manual. Woodcock-Johnson IV. Rolling Meadows, IL: Riverside.
#'
#' @source Schrank, F. A., McGrew, K. S., & Mather, N. (2014). Woodcock-Johnson IV.
#' Rolling Meadows, IL: Riverside.
"WJIV_ages_3_5"

# Unified result class for the factor-retention criteria. Every criterion
# returns an `efa_retention` object built by `.new_efa_retention()`, printed by
# one `print.efa_retention()`, and plotted by one `plot.efa_retention()` that
# dispatches to the two ggplot helpers below. The registry is the single source
# of truth mapping each criterion id to its display label.

# Registry of the factor-retention criteria (id -> metadata). Each `fun` runs
# its criterion from the efa_retain() control list `ctl`; `x` is the raw data
# when `needs_raw`, the prepared correlation matrix otherwise. Every criterion
# that fits a model through efa_fit() (HULL, KGC, PARALLEL, SCREE, NEST, SMT)
# receives `ctl$estimate_control`, so the estimation settings reach all of them
# and not just some; those that also forward arguments to efa_fit() receive them
# via `ctl$dots`. CD, EKC and MAP run no efa_fit() model (CD's `max_iter` caps
# its own comparison-data generation, not an EFA), so neither applies to them.
# `poly_ok = FALSE` marks the criteria that do not support polychoric/tetrachoric
# correlations -- either because they compare the data against continuous
# reference data (CD, PARALLEL, NEST, HULL) or because their normal-theory
# chi-square test is not valid for such correlations (SMT); these are skipped
# with an informative note by efa_retain().
.retention_registry <- list(
  CD = list(
    label = "Comparison data", needs_raw = TRUE, poly_ok = FALSE,
    fun = function(x, ctl) {
      efa_cd(x, n_factors_max = ctl$n_factors_max, N_pop = ctl$N_pop,
         N_samples = ctl$N_samples, alpha = ctl$alpha,
         cor_method = ctl$cor_method, max_iter = ctl$max_iter_CD)
    }),
  EKC = list(
    label = "Empirical Kaiser Criterion", needs_raw = FALSE,
    fun = function(x, ctl) {
      efa_ekc(x, N = ctl$N, use = ctl$use, cor_method = ctl$cor_method)
    }),
  HULL = list(
    label = "Hull method", needs_raw = FALSE, poly_ok = FALSE,
    fun = function(x, ctl) {
      do.call(efa_hull, c(list(x, N = ctl$N, n_fac_theor = ctl$n_fac_theor,
                           estimator = ctl$estimator, gof = ctl$gof,
                           eigen_type = ctl$eigen_type_HULL, use = ctl$use,
                           cor_method = ctl$cor_method,
                           n_datasets = ctl$n_datasets, percent = ctl$percent,
                           decision_rule = ctl$decision_rule,
                           n_factors = ctl$n_factors,
                           estimate_control = ctl$estimate_control),
                      ctl$dots))
    }),
  KGC = list(
    label = "Kaiser-Guttman criterion", needs_raw = FALSE,
    fun = function(x, ctl) {
      do.call(efa_kgc, c(list(x, eigen_type = ctl$eigen_type_other, use = ctl$use,
                          cor_method = ctl$cor_method,
                          n_factors = ctl$n_factors, estimator = ctl$estimator,
                          estimate_control = ctl$estimate_control),
                     ctl$dots))
    }),
  MAP = list(
    label = "Minimum average partial", needs_raw = FALSE,
    fun = function(x, ctl) {
      efa_map(x, use = ctl$use, cor_method = ctl$cor_method)
    }),
  NEST = list(
    label = "Next Eigenvalue Sufficiency Test", needs_raw = FALSE, poly_ok = FALSE,
    fun = function(x, ctl) {
      do.call(efa_nest, c(list(x, N = ctl$N, use = ctl$use,
                           cor_method = ctl$cor_method, alpha = ctl$alpha_nest,
                           n_datasets = ctl$n_datasets_nest, estimator = ctl$estimator,
                           estimate_control = ctl$estimate_control),
                      ctl$dots))
    }),
  PARALLEL = list(
    label = "Parallel analysis", needs_raw = FALSE, poly_ok = FALSE,
    fun = function(x, ctl) {
      do.call(efa_parallel, c(list(x, N = ctl$N, n_datasets = ctl$n_datasets,
                               percent = ctl$percent,
                               eigen_type = ctl$eigen_type_other, use = ctl$use,
                               cor_method = ctl$cor_method,
                               decision_rule = ctl$decision_rule,
                               n_factors = ctl$n_factors, estimator = ctl$estimator,
                               estimate_control = ctl$estimate_control),
                          ctl$dots))
    }),
  SCREE = list(
    label = "Scree plot", needs_raw = FALSE, visual = TRUE,
    fun = function(x, ctl) {
      do.call(efa_scree, c(list(x, eigen_type = ctl$eigen_type_other, use = ctl$use,
                            cor_method = ctl$cor_method,
                            n_factors = ctl$n_factors, estimator = ctl$estimator,
                            estimate_control = ctl$estimate_control),
                       ctl$dots))
    }),
  SMT = list(
    label = "Sequential model tests", needs_raw = FALSE, poly_ok = FALSE,
    fun = function(x, ctl) {
      efa_smt(x, N = ctl$N, use = ctl$use, cor_method = ctl$cor_method,
          estimate_control = ctl$estimate_control)
    })
)

# Control list consumed by the factor-retention registry funs. The defaults
# mirror the efa_retain() argument defaults so any requested criterion resolves;
# callers pass only what they override. `gof` defaults to the Hull
# goodness-of-fit indices valid for `estimator` (PAF supports only the CAF).
.n_factors_ctl <- function(N = NA, use = "pairwise.complete.obs",
                           cor_method = "pearson", n_factors_max = NA,
                           N_pop = 10000, N_samples = 500, alpha = .30,
                           max_iter_CD = 50, n_fac_theor = NA, estimator = "ML",
                           gof = if (estimator == "PAF") "CAF" else c("CAF", "CFI", "RMSEA"),
                           eigen_type_HULL = "SMC", eigen_type_other = "SMC",
                           n_factors = 1, n_datasets = 1000, percent = 95,
                           decision_rule = "means",
                           n_datasets_nest = 1000, alpha_nest = .05,
                           estimate_control = NULL, dots = list()) {
  list(N = N, use = use, cor_method = cor_method,
       n_factors_max = n_factors_max, N_pop = N_pop,
       N_samples = N_samples, alpha = alpha, max_iter_CD = max_iter_CD,
       n_fac_theor = n_fac_theor, estimator = estimator, gof = gof,
       eigen_type_HULL = eigen_type_HULL,
       eigen_type_other = eigen_type_other, n_factors = n_factors,
       n_datasets = n_datasets, percent = percent,
       decision_rule = decision_rule,
       n_datasets_nest = n_datasets_nest, alpha_nest = alpha_nest,
       estimate_control = estimate_control, dots = dots)
}

# One sample-size rule for the criteria that need more observations than
# variables: parallel analysis and NEST draw reference data with N cases, and
# EKC and SMT rest on normal-theory quantities ((1 + sqrt(J / N))^2 and the
# Bartlett-corrected chi-square) that are not defined below that boundary. Kept
# in one place so the contract, the message, and the condition class stay
# identical across the four criteria. Both arguments must already be resolved:
# call it after the N-required guard (.prepare_cor_input(N_policy = "required"),
# or the caller's own check), because an NA would make the comparison itself fail
# instead of raising the classed condition.
.assert_n_gt_vars <- function(N, n_vars, error_call = rlang::caller_env()) {
  if (N <= n_vars) {
    cli::cli_abort(
      c("{.arg N} must be larger than the number of variables.",
        "x" = "You supplied {.arg N} = {N} for {n_vars} variable{?s}."),
      class = "efa_n_too_small",
      call = error_call
    )
  }
  invisible(N)
}

# Name a criterion's suggested factor counts the way efa_retain() aggregates
# them: a single-variant suggestion keeps the bare id, a multi-variant one
# becomes "<id>_<variant>".
.retention_key <- function(id, nf) {
  names(nf) <- ifelse(names(nf) == id, id, paste(id, names(nf), sep = "_"))
  nf
}

# Construct an efa_retention object from a list of per-sub-variant records. The
# top-level `n_factors` named vector is derived from the records. `subtitle` is an
# optional one-line context string (e.g. the estimation method); `note` is an
# optional vector of cli info lines.
.new_efa_retention <- function(id, results, settings,
                               subtitle = NULL, note = NULL) {

  if (!id %in% names(.retention_registry)) {
    cli::cli_abort("Unknown factor-retention criterion id {.val {id}}.",
                   class = "efa_unknown_criterion")
  }

  label <- .retention_registry[[id]]$label

  n_factors <- vapply(results, function(r) as.numeric(r$n_factors), numeric(1))
  names(n_factors) <- vapply(results, function(r) r$name, character(1))

  structure(
    list(
      criterion = c(id = id, label = label),
      n_factors = n_factors,
      results = results,
      subtitle = subtitle,
      note = note,
      settings = settings
    ),
    class = "efa_retention"
  )
}

# Subtitle naming the eigenvalue types a criterion computed its eigenvalues on,
# shared by the eigenvalue-based criteria so they all say it the same way. `detail`
# appends a clause (e.g. the number of simulated reference datasets) in place of the
# closing full stop.
.eigen_subtitle <- function(eigen_type, detail = NULL) {
  paste0("Eigenvalues found using ", cli::ansi_collapse(eigen_type),
         if (is.null(detail)) "." else paste0("; ", detail, "."))
}

# Render a whole-number count (of factors, or of simulated datasets) for the report, a
# condition message, or a plot label. The criteria differ in the storage mode they put in
# their record: a count derived with cumprod() or which.min() - 1 comes out double, one
# taken from an index integer. Under a negative options(scipen) a double is rendered in
# scientific notation ("3e+00" instead of "3") by as.character() and paste0() alike, and
# also by grid, which coerces a numeric plot label at draw time. A count is exempt from
# that option because it is whole by construction; a genuinely continuous quantity (an
# eigenvalue, a fit index) still honours it.
#
# scientific = FALSE pins the fixed notation whatever the option is set to. trim keeps a
# vector of counts unpadded, so that the point labels of a hull with ten or more solutions
# are not indented to the width of the widest of them.
#
# This covers the counts the criteria compute. A count that cli interpolates from a user
# argument -- `{N}` in .assert_n_gt_vars() below -- does not come through here.
.retention_count <- function(n) {
  format(n, scientific = FALSE, trim = TRUE)
}

# Bullet lines (one per record) shared by format.efa_retention and
# format.efa_retain.
.retention_bullets <- function(results) {
  vapply(results, function(r) {
    value <- if (is.na(r$n_factors)) "not applicable" else .retention_count(r$n_factors)
    paste0(r$label, ": ", value)
  }, character(1))
}

#' Print method for efa_retention objects
#'
#' @param x an object of class efa_retention, returned by a factor-retention
#'   criterion (e.g. [efa_ekc()] or [efa_hull()]).
#' @param ... not used.
#'
#' @returns `print()` returns its argument `x` invisibly; it is
#'   `cat(format(x), sep = "\n")`.
#'
#' @export
#' @method print efa_retention
#'
#' @examples
#' efa_ekc(test_models$baseline$cormat, N = 500)
print.efa_retention <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' Format method for efa_retention objects
#'
#' @param x an object of class efa_retention, returned by a factor-retention
#'   criterion (e.g. [efa_ekc()] or [efa_hull()]).
#' @param ... not used.
#'
#' @returns A character vector with the report lines (styled to the active
#'   console theme; plain when colours are disabled).
#'
#' @export
#' @method format efa_retention
#'
#' @examples
#' writeLines(format(efa_ekc(test_models$baseline$cormat, N = 500)))
format.efa_retention <- function(x, ...) {
  cli::cli_format_method({
    cli::cli_rule(left = x$criterion[["label"]])
    if (!is.null(x$subtitle)) {
      cli::cli_text("{x$subtitle}")
    }
    # criteria with no numeric suggestion (e.g. the visual scree plot) skip the
    # bullets and rely on their subtitle/note
    if (any(!is.na(x$n_factors))) {
      cli::cli_text("")
      cli::cli_ul(.retention_bullets(x$results))
    }
    if (!is.null(x$note)) {
      cli::cli_text("")
      for (msg in x$note) {
        cli::cli_alert_info("{msg}", wrap = TRUE)
      }
    }
  })
}

#' Plot method for efa_retention objects
#'
#' Plots the result of a factor-retention criterion. Eigenvalue-based criteria
#' (e.g. [efa_ekc()]) are shown as an eigenvalue plot, the Hull method ([efa_hull()]) as
#' a convex-hull plot. Criteria with more than one sub-variant are faceted.
#'
#' @param x an object of class efa_retention, returned by a factor-retention
#'   criterion (e.g. [efa_ekc()] or [efa_hull()]).
#' @param ... not used.
#'
#' @returns A [ggplot2::ggplot] object, or invisibly `NULL` if the criterion has
#'   no plottable result.
#'
#' @export
#' @method plot efa_retention
#'
#' @examples
#' plot(efa_ekc(test_models$baseline$cormat, N = 500))
plot.efa_retention <- function(x, ...) {

  plot_types <- unique(vapply(x$results, function(r) r$plot_type, character(1)))
  plot_types <- setdiff(plot_types, "none")

  if (length(plot_types) == 0) {
    cli::cli_inform("No plot is available for {x$criterion[['label']]}.",
                    class = "efa_no_plot")
    return(invisible(NULL))
  }

  if ("hull" %in% plot_types) {
    .gg_hull_plot(x)
  } else {
    .gg_eigen_plot(x)
  }

}

# Eigenvalue plot (factor index on x, eigenvalues on y) with an optional dashed
# reference series, an optional horizontal threshold, and an optional highlighted
# retained point. Covers the EKC/KGC/PARALLEL/SCREE/CD plots. Returns a ggplot.
#' @importFrom rlang .data
.gg_eigen_plot <- function(x) {

  # drop records with no plottable points (e.g. CD when it suggests 0 factors)
  records <- Filter(function(r) length(r$x) > 0, x$results)
  if (length(records) == 0) {
    cli::cli_inform("No plot is available for {x$criterion[['label']]}.",
                    class = "efa_no_plot")
    return(invisible(NULL))
  }

  variant_levels <- vapply(records, function(r) r$label, character(1))

  dat <- do.call(rbind, lapply(records, function(r) {
    data.frame(
      variant = r$label,
      factor = r$x,
      # a record may have no primary series (e.g. PARALLEL without real data
      # plots only its reference series)
      value = if (is.null(r$y)) NA_real_ else r$y,
      reference = if (is.null(r$reference)) NA_real_ else r$reference,
      stringsAsFactors = FALSE
    )
  }))
  dat$variant <- factor(dat$variant, levels = variant_levels)

  highlights <- do.call(rbind, lapply(records, function(r) {
    if (is.null(r$highlight) || is.na(r$highlight) || r$highlight < 1) return(NULL)
    # `factor` positions the point, `label` names it (see .retention_count())
    data.frame(variant = r$label, factor = r$highlight,
               value = r$y[r$highlight], label = .retention_count(r$highlight),
               stringsAsFactors = FALSE)
  }))

  thresholds <- do.call(rbind, lapply(records, function(r) {
    if (is.null(r$threshold)) return(NULL)
    data.frame(variant = r$label, yintercept = r$threshold,
               stringsAsFactors = FALSE)
  }))

  p <- ggplot2::ggplot(dat, ggplot2::aes(.data$factor, .data$value)) +
    ggplot2::geom_line(na.rm = TRUE) +
    ggplot2::geom_point(na.rm = TRUE)

  if (any(!is.na(dat$reference))) {
    p <- p + ggplot2::geom_line(ggplot2::aes(y = .data$reference),
                                linetype = 2, colour = "darkgray", na.rm = TRUE)
  }

  if (!is.null(thresholds)) {
    thresholds$variant <- factor(thresholds$variant, levels = variant_levels)
    p <- p + ggplot2::geom_hline(data = thresholds,
                                 ggplot2::aes(yintercept = .data$yintercept),
                                 linetype = 2, colour = "darkgray")
  }

  # records can carry several named reference series (e.g. PARALLEL's simulated
  # means and percentile eigenvalues), each drawn as a dashed coloured line with
  # a shared legend; criteria without `references` are unaffected
  ref_records <- Filter(function(r) !is.null(r$references), records)
  if (length(ref_records) > 0) {
    ref_dat <- do.call(rbind, lapply(ref_records, function(r) {
      data.frame(variant = r$label,
                 factor = rep(r$x, length(r$references)),
                 series = rep(names(r$references), each = length(r$x)),
                 value = unlist(r$references, use.names = FALSE),
                 stringsAsFactors = FALSE)
    }))
    ref_dat$variant <- factor(ref_dat$variant, levels = variant_levels)
    ref_dat$series <- factor(ref_dat$series, levels = unique(ref_dat$series))
    # label the primary series in the legend alongside the reference series
    # (drawn over the identical primary line, so only the legend key is added)
    ref_labels <- vapply(ref_records, function(r) r$label, character(1))
    real_dat <- dat[dat$variant %in% ref_labels & !is.na(dat$value), ,
                    drop = FALSE]
    if (nrow(real_dat) > 0) {
      p <- p +
        ggplot2::geom_line(data = real_dat,
                           ggplot2::aes(linetype = "Real Eigenvalues")) +
        ggplot2::scale_linetype_manual(values = c(`Real Eigenvalues` = 1))
    }
    p <- p +
      ggplot2::geom_line(data = ref_dat,
                         ggplot2::aes(.data$factor, .data$value,
                                      colour = .data$series),
                         linetype = 2, na.rm = TRUE) +
      ggplot2::scale_colour_viridis_d(end = 0.8) +
      ggplot2::labs(colour = NULL, linetype = NULL)
  }

  if (!is.null(highlights)) {
    highlights$variant <- factor(highlights$variant, levels = variant_levels)
    p <- p +
      ggplot2::geom_point(data = highlights, shape = 1, size = 4, colour = "red") +
      ggplot2::geom_text(data = highlights, ggplot2::aes(label = .data$label),
                         colour = "red", vjust = -1)
  }

  if (length(variant_levels) > 1) {
    p <- p + ggplot2::facet_wrap(ggplot2::vars(.data$variant), scales = "free_y")
  }

  # y-axis label defaults to "Eigenvalues" but a criterion can override it (e.g.
  # CD plots mean RMSE of the eigenvalues)
  y_label <- records[[1]]$y_label
  if (is.null(y_label)) y_label <- "Eigenvalues"

  p +
    ggplot2::scale_x_continuous(breaks = seq_len(max(dat$factor))) +
    ggplot2::labs(x = "Factor", y = y_label, title = x$criterion[["label"]]) +
    .gg_theme()

}

# Convex-hull plot (degrees of freedom on x, goodness-of-fit on y). Points on the
# hull are emphasised and connected; the retained solution is highlighted in red;
# each point is labelled with its number of factors. Records whose hull is empty
# (the degenerate fewer-than-three-solutions case) are dropped. Returns a ggplot.
#' @importFrom rlang .data
.gg_hull_plot <- function(x) {

  results <- Filter(function(r) any(r$on_hull), x$results)

  if (length(results) == 0) {
    cli::cli_inform("No hull plot is available for {x$criterion[['label']]}.")
    return(invisible(NULL))
  }

  # The RMSEA record carries 1 - RMSEA on the y-axis (see the HULL details), so its
  # facet is labelled accordingly; the record/bullet label stays "RMSEA".
  hull_facet <- function(label) ifelse(label == "RMSEA", "1 - RMSEA", label)

  variant_levels <- hull_facet(vapply(results, function(r) r$label, character(1)))

  dat <- do.call(rbind, lapply(results, function(r) {
    # `nfac` is only ever drawn as a text label (see .retention_count())
    data.frame(variant = hull_facet(r$label), df = r$x, fit = r$y,
               nfac = .retention_count(r$point_labels), on_hull = r$on_hull,
               stringsAsFactors = FALSE)
  }))
  dat$variant <- factor(dat$variant, levels = variant_levels)

  retained <- do.call(rbind, lapply(results, function(r) {
    idx <- which(r$point_labels == r$highlight)
    if (length(idx) == 0) return(NULL)
    data.frame(variant = hull_facet(r$label), df = r$x[idx], fit = r$y[idx],
               stringsAsFactors = FALSE)
  }))

  p <- ggplot2::ggplot(dat, ggplot2::aes(.data$df, .data$fit)) +
    ggplot2::geom_line(data = dat[dat$on_hull, , drop = FALSE]) +
    ggplot2::geom_point(ggplot2::aes(colour = .data$on_hull)) +
    ggplot2::geom_text(ggplot2::aes(label = .data$nfac), vjust = -1, size = 3) +
    ggplot2::scale_colour_manual(values = c(`TRUE` = "black", `FALSE` = "darkgray"),
                                 guide = "none")

  if (!is.null(retained)) {
    retained$variant <- factor(retained$variant, levels = variant_levels)
    p <- p + ggplot2::geom_point(data = retained, shape = 1, size = 4,
                                 colour = "red")
  }

  if (length(variant_levels) > 1) {
    p <- p + ggplot2::facet_wrap(ggplot2::vars(.data$variant), scales = "free_y")
  }

  p +
    ggplot2::labs(x = "Degrees of freedom", y = "Goodness of fit",
                  title = x$criterion[["label"]]) +
    .gg_theme()

}

test_that(".new_efa_retention rejects unknown criterion ids", {
  rec <- list(list(name = "X", label = "X", n_factors = 1, plot_type = "none"))
  expect_error(
    .new_efa_retention("NOPE", results = rec, settings = list()),
    class = "efa_unknown_criterion"
  )
})

test_that(".new_efa_retention builds the documented shape for a known id", {
  rec <- list(
    list(name = "TR2", label = "Original", n_factors = 3, plot_type = "eigen"),
    list(name = "TR4", label = "Revised", n_factors = 2, plot_type = "eigen")
  )
  out <- .new_efa_retention("MAP", results = rec, settings = list(a = 1))

  expect_s3_class(out, "efa_retention")
  expect_equal(unname(out$criterion["label"]), "Minimum average partial")
  expect_named(out$n_factors, c("TR2", "TR4"))
  expect_equal(unname(out$n_factors), c(3, 2))
})

test_that(".n_factors_ctl() defaults mirror efa_retain()'s", {
  # efa_power() calls .n_factors_ctl() relying on its defaults, so a default drifting from
  # efa_retain()'s would silently retain factors under undocumented settings.
  ctl <- .n_factors_ctl()
  shared <- intersect(names(ctl), names(formals(efa_retain)))
  ret <- lapply(formals(efa_retain)[shared], eval)
  # efa_retain() resolves these with a single-choice match.arg(), so its effective default
  # is the first element; every other default (including the several.ok ones, whose whole
  # vector is the default) has to match .n_factors_ctl()'s in full
  single <- c("use", "cor_method", "estimator", "eigen_type_HULL", "decision_rule")
  ret[single] <- lapply(ret[single], `[`, 1L)
  expect_equal(ctl[shared], ret[shared])
})

test_that(".eigen_subtitle names the eigenvalue types the same way for every criterion", {
  # The criteria that report which eigenvalues they used share this one sentence, so
  # pin both forms here: efa_parallel()'s `detail` variant is otherwise only covered
  # by a print snapshot behind the slow gate.
  expect_equal(.eigen_subtitle("SMC"), "Eigenvalues found using SMC.")
  expect_equal(.eigen_subtitle(c("PCA", "SMC")),
               "Eigenvalues found using PCA and SMC.")
  expect_equal(.eigen_subtitle(c("PCA", "SMC", "EFA")),
               "Eigenvalues found using PCA, SMC, and EFA.")

  # a detail clause replaces the closing full stop rather than following it
  expect_equal(.eigen_subtitle("SMC", "1000 simulated datasets"),
               "Eigenvalues found using SMC; 1000 simulated datasets.")

  # and the criteria really route through it
  expect_equal(efa_scree(test_models$baseline$cormat, eigen_type = "PCA")$subtitle,
               .eigen_subtitle("PCA"))
  expect_equal(efa_kgc(test_models$baseline$cormat, eigen_type = "PCA")$subtitle,
               .eigen_subtitle("PCA"))
})

test_that("printed bullets name what the count belongs to, not the eigenvalues", {
  # Every bullet reads "<variant>: <suggested count>": the eigenvalue-based criteria
  # key theirs by eigenvalue type, the Hull method by goodness-of-fit index, and the
  # single-variant criteria name the quantity itself. The eigenvalues the counts were
  # derived from belong in the subtitle, so that a count is never labelled as if it
  # were the eigenvalues.
  kgc <- efa_kgc(test_models$baseline$cormat)
  expect_equal(vapply(kgc$results, function(r) r$label, character(1)),
               c("PCA", "SMC", "EFA"))
  expect_equal(kgc$subtitle, "Eigenvalues found using PCA, SMC, and EFA.")

  scree <- efa_scree(test_models$baseline$cormat, eigen_type = "SMC")
  expect_equal(vapply(scree$results, function(r) r$label, character(1)), "SMC")
  expect_equal(scree$subtitle, "Eigenvalues found using SMC.")
  # the one purely visual criterion names the call that shows the plot
  expect_match(scree$note, "plot(x)", fixed = TRUE)
})

test_that(".retention_count() renders a count independently of the number options", {
  # Counts are whole by construction, so they are rendered the same way whatever
  # options(scipen) asks for. The two properties the call sites rely on:
  skip_if_not_installed("withr")
  withr::local_options(scipen = -5, digits = 2, OutDec = ",")

  # fixed notation, for a double as much as for an integer
  expect_identical(.retention_count(3), "3")
  expect_identical(.retention_count(3L), "3")
  expect_identical(.retention_count(1000), "1000")

  # and no padding to a common width: the hull draws a vector of point labels, and
  # .retention_summary() lists tied modal values, so a padded element would be visible
  expect_identical(.retention_count(c(0, 1, 2, 10)), c("0", "1", "2", "10"))
})

test_that("counts are rendered the same way whatever options(scipen) is", {
  # The criteria differ in the storage mode they put in their record: a count derived with
  # cumprod() or which.min() comes out double, one taken from an index integer. Under a
  # negative options(scipen) a double is rendered in scientific notation ("3e+00" instead
  # of "3") by as.character() in the report, and also by grid, which coerces a numeric plot
  # label at draw time. Both surfaces are therefore checked for every criterion, because
  # each renders its count on its own path.
  #
  # The option is set BEFORE the fixtures are built, because a criterion can render a
  # count into a string when it constructs its object rather than when the object is
  # printed (efa_parallel() does this for its subtitle). Fixtures built first would only
  # exercise the format()-time paths.
  skip_if_not_installed("withr")
  withr::local_options(scipen = -5)
  cmat <- test_models$baseline$cormat
  set.seed(42)
  # PARALLEL keeps a two-digit `n_datasets` on purpose: its subtitle names that number, so
  # this is the fixture whose exponent is not "e+00", and it holds the pattern below to a
  # general one. The other simulation counts are as low as still yields a record.
  fits <- list(
    EKC = efa_ekc(cmat, N = 500),
    KGC = efa_kgc(cmat, eigen_type = c("PCA", "SMC")),
    MAP = efa_map(cmat),
    PARALLEL = efa_parallel(cmat, N = 500, n_datasets = 20, eigen_type = "SMC"),
    SCREE = efa_scree(cmat, eigen_type = "SMC"),
    SMT = efa_smt(cmat, N = 500),
    HULL = efa_hull(cmat, N = 500, gof = "CAF"),
    NEST = efa_nest(cmat, N = 500, n_datasets = 5),
    # CD is the remaining criterion that stores its count as a double; it needs raw data
    CD = suppressMessages(efa_cd(GRiPS_raw[1:300, ], N_pop = 1000, N_samples = 20))
  )

  # every text a plot draws, coerced the way grid coerces it
  plot_labels <- function(p) {
    unlist(lapply(ggplot2::ggplot_build(p)$data, function(d) {
      if ("label" %in% names(d)) as.character(d$label) else NULL
    }))
  }

  # any exponent, not just "e+00": a count of ten or more carries a different one
  sci <- "e[+-][0-9]"
  for (id in names(fits)) {
    expect_false(any(grepl(sci, format(fits[[id]]))), info = id)
  }

  # the eigenvalue plots label the retained solution, the hull plot every point; the
  # criteria that draw neither (MAP, SMT, and SCREE, which retains nothing) have no label.
  # CD is checked on its report only: it is simulation-based, and a draw that suggested no
  # factors would leave its record with no point to label at all.
  for (id in c("EKC", "KGC", "PARALLEL", "NEST", "HULL")) {
    labels <- plot_labels(plot(fits[[id]]))
    expect_gt(length(labels), 0)
    expect_false(any(grepl(sci, labels)), info = id)
  }
})

test_that("the retention report fits a narrow console", {
  local_reproducible_output(width = 40)
  objs <- list(efa_kgc(test_models$baseline$cormat),
               efa_scree(test_models$baseline$cormat),
               efa_ekc(test_models$baseline$cormat, N = 500),
               efa_map(test_models$baseline$cormat))
  for (obj in objs) {
    expect_true(all(nchar(format(obj)) <= 40),
                info = obj$criterion[["id"]])
  }
})

test_that("format.efa_retention is the source of truth and honours the colour state", {
  # Two fixtures, because no one criterion covers both halves. The note is the only
  # styled element a report carries, and the scree plot is the criterion that always has
  # one -- but it makes no numeric suggestion, so its report has no bullets. EKC is the
  # other way round. Together they reach the rule, the note, and the bullet list.
  scree <- efa_scree(test_models$baseline$cormat, eigen_type = "SMC")
  ekc <- efa_ekc(test_models$baseline$cormat, N = 500)
  expect_true(all(is.na(scree$n_factors)))   # the bullets branch is skipped ...
  expect_true(any(!is.na(ekc$n_factors)))    # ... and taken

  # print() is exactly cat(format(x), sep = "\n"), so the two agree line for line.
  expect_identical(utils::capture.output(print(scree)), format(scree))
  expect_identical(utils::capture.output(print(ekc)), format(ekc))

  old <- options(cli.num_colors = 256)
  on.exit(options(old), add = TRUE)

  # With colours on the report embeds ANSI ...
  expect_true(any(grepl("\033", format(scree), fixed = TRUE)))

  # ... and with colours off it is plain.
  options(cli.num_colors = 1)
  expect_false(any(grepl("\033", format(scree), fixed = TRUE)))
  expect_false(any(grepl("\033", format(ekc), fixed = TRUE)))
})

## Unweighted Least Squares Estimation of Factor Loadings
## Thin fitter: returns the raw ULS results; shared post-processing happens in
## .finalize_fit() / .estimate_model().
.ULS <- function(x, n_factors) {

  # Get correlation matrix entered or created in EFA
  R <- x

  uls <- .fit_uls(R, n_factors)

  L <- uls$loadings
  orig_R <- R
  h2 <- rowSums(L^2)
  diag(R) <- h2

  # raw fit, finalized by .estimate_model(). `Fm` is the ULS objective (half the squared
  # Frobenius norm of the reduced residual) the C++ backend already minimised and returned.
  list(
    L = L,
    h2 = h2,
    psi = uls$res$par,
    Fm = uls$res$value,
    iter = uls$res$counts[1],
    convergence = uls$res$convergence,
    orig_R = orig_R,
    R_final = R
  )
}

# function to obtain the uls fit; adapted from the psych package
.fit_uls <- function(R, n_fac) {

  start <- diag(R) - .smc_start(R)

  # Bounded L-BFGS-B over the uniquenesses, run entirely in C++ (roptim drives
  # R's lbfgsb routine, so the box constraints and control match stats::optim).
  uls <- .fit_uls_cpp(R, n_fac, start, .uniqueness_floor)

  res <- list(par = uls$psi, value = uls$Fm, counts = c(uls$iter, NA_integer_),
              convergence = uls$convergence)

  list(loadings = uls$loadings, res = res, R = R)
}

#' Print and format an efa_schmid_leiman object
#'
#' `print()` shows a summarised output of the [efa_schmid_leiman()] function: a
#' model header (when the settings are available), the Schmid-Leiman loading
#' matrix, and the variances accounted for. `format()` assembles the same report
#' and returns it as a character vector; `print()` is `cat(format(x), sep = "\n")`.
#' The lines follow the active console theme, so they are plain when colours are
#' disabled (for example when captured into a file or stripped with
#' [cli::ansi_strip()]).
#'
#' @param x An object of class `efa_schmid_leiman` (output from
#'   [efa_schmid_leiman()]).
#' @param ... Not used; for consistency with the generic.
#'
#' @returns `print()` returns its argument `x` invisibly. `format()` returns a
#'   character vector with the report lines.
#'
#' @export
#' @method print efa_schmid_leiman
#'
#' @examples
#' EFA_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
#'                    estimator = "PAF", rotation = "promax")
#' sl_mod <- efa_schmid_leiman(EFA_mod, estimator = "PAF")
#' sl_mod
#'
#' # format() returns the same lines as a character vector:
#' writeLines(format(sl_mod))
#'
print.efa_schmid_leiman <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @rdname print.efa_schmid_leiman
#' @export
#' @method format efa_schmid_leiman
format.efa_schmid_leiman <- function(x, ...) {

  cli::cli_format_method({

    if (!all(is.na(x$settings))) {
      # extract the estimator of the second-order EFA
      estimator <- x$settings$estimator

      # Emitted verbatim (not via cli_text) so the "setting = 'value'" token is never split
      # across a line break; the bold value therefore uses style_bold() rather than {.strong}.
      # `.efa_wrap_chunks()` then packs the line to the console around that token.
      cli::cli_text("")
      cli::cli_verbatim(.efa_wrap_chunks(c(
        "EFA", "for", "second-order", "loadings", "performed", "with",
        paste0("estimator = '", cli::style_bold(estimator), "'")
      )))

      # Surface non-convergence of the second-order EFA, keyed on its convergence
      # code (with the iteration-count fallback), matching print.efa().
      if (.efa_iteration_nonconvergence(list(convergence = x$convergence,
                                             iter = x$iter,
                                             max_iter = x$settings$max_iter))) {
        cli::cli_text("")
        cli::cli_alert_danger(.efa_nonconvergence_banner(estimator))
      }
    }

    # print the loadings and the variances
    .print_efa_rule("Schmid-Leiman Solution")
    .efa_emit_lines(format(x$sl))

    .print_efa_rule("Variances Accounted for")
    .efa_emit_lines(.efa_corr_lines(x$vars_accounted, digits = 3))
  })
}

cd_grips <- CD(GRiPS_raw)

test_that("output class and dimensions are correct", {
  expect_is(cd_grips, "CD")
  expect_named(cd_grips, c("n_factors", "eigenvalues", "RMSE_eigenvalues",
                           "settings"))
  expect_is(cd_grips$RMSE_eigenvalues, "matrix")
})

test_that("CD returns the correct values", {
  expect_equal(cd_grips$n_factors, 1)
  expect_equal(sum(cd_grips$eigenvalues), 8)
})

grips_na <- GRiPS_raw
grips_na[c(1,3,5), c(2, 4)] <- NA
test_that("errors etc. are thrown correctly", {
  expect_error(CD(1:10), " 'x' is neither a matrix nor a dataframe. Provide a dataframe or matrix with raw data.\n")
  expect_error(CD(test_models$baseline$cormat), " 'x' is a correlation matrix, but CD only works with raw data.\n")

  expect_warning(CD(GRiPS_raw, n_factors_max = 5), " n_factors_max was set to 5 but maximum possible factors to extract is 4 . Setting n_factors_max to 4 .\n")
  expect_warning(CD(grips_na, n_factors_max = 3),
                 " The data contained missing values that were removed using stats::na.omit(). 3 row(s) were removed.\n",
                 fixed = TRUE)
})

rm(cd_grips, grips_na)

---
title: "EFAtools"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{EFAtools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.align = "center"
)

if (!requireNamespace("microbenchmark", quietly = TRUE)) {
      stop("Package \"microbenchmark\" needed for this vignette to work. Please install it.",
      call. = FALSE)
}

```



This vignette provides an overview for the functionalities of the EFAtools package. The general aim of the package is to provide flexible implementations of different algorithms for an exploratory factor analyses (EFA) procedure, including factor retention methods, factor extraction and rotation methods, as well as the computation of a Schmid-Leiman solution and McDonald's omega coefficients.

The package was first designed to enable a comparison of EFA (specifically, principal axis factoring with subsequent promax rotation) performed in R using the [**psych**](https://CRAN.R-project.org/package=psych) package and EFA performed in SPSS. That is why some functions allow the specification of a type, including `"psych"` and `"SPSS"`, such that the respective procedure will be executed to match the output of these implementations (which do not always lead to the same results; see separate vignette [**Replicate_SPSS_psych**](Replicate_SPSS_psych.html "Replicate SPSS and R psych results with EFAtools") for a demonstration of the replication of original results). This vignette will go through a complete example, that is, we will first show how to determine the number of factors to retain, then perform different factor extraction methods, run a Schmid-Leiman transformation and compute omegas.

The package can be installed from CRAN using `install.packages("EFAtools")`, or from GitHub using `devtools::install_github("mdsteiner/EFAtools")`, and then loaded using:

```{r}
library(EFAtools)
```

In this vignette, we will use the `DOSPERT_raw` dataset, which contains responses to the Domain Specific Risk Taking Scale (DOSPERT) of 3123 participants. The dataset is contained in the `EFAtools` package, for details, see `?DOSPERT_raw`. Note that this vignette is to provide a general overview and it is beyond its scope to explain all methods and functions in detail. If you want to learn more on the details and methods, please see the respective help functions for explanations and literature references. However, the dataset is rather large, so, just to save time when building the vignette, we will only use the first 500 observations. When you normally do your analyses, you use the full dataset.

```{r}
# only use a subset to make analyses faster
DOSPERT_sub <- DOSPERT_raw[1:500,]
```


## Test Suitability of Data

The first step in an EFA procedure is to test whether your data is suitable for factor analysis. To this end, the `EFAtools` package provides the `BARTLETT()` and the `KMO()` functions. The Bartlett's test of sphericity tests whether a correlation matrix is significantly different from an identity matrix (a correlation matrix with zero correlations between all variables). This test should thus be significant. The Kaiser-Meyer-Olkin criterion (KMO) represents the degree to which each observed variable is predicted by the other variables in the dataset and thus is another indicator for how correlated the different variables are.

We can test whether our `DOSPERT_sub` dataset is suitable for factor analysis as follows.

```{r}
# Bartlett's test of sphericity
BARTLETT(DOSPERT_sub)

# KMO criterion
KMO(DOSPERT_sub)
```

Note that these tests can also be run in the `N_FACTORS()` function.

## Factor Retention Methods

As the goal of EFA is to determine the underlying factors from a set of multiple variables, one of the most important decisions is how many factors can or should be extracted. There exists a plethora of factor retention methods to use for this decision. The problem is that there is no method that consistently outperforms all other methods. Rather, which factor retention method to use depends on the structure of the data: are there few or many indicators, are factors strong or weak, are the factor intercorrelations weak or strong. For rules on which methods to use, see, for example, [Auerswald and Moshagen, (2019)](https://doi.org/10.1037/met0000200).

There are multiple factor retention methods implemented in the `EFAtools` package. They can either be called with separate functions, or all (or a selection) of them using the `N_FACTORS()` function. 

### Calling Separate Functions

Let's first look at how to determine the number of factors to retain by calling separate functions. For example, if you would like to perform a parallel analysis based on squared multiple correlations (SMC; sometimes also called a parallel analysis with principal factors), you can do the following:

```{r}
# determine the number of factors to retain using parallel analysis
PARALLEL(DOSPERT_sub, eigen_type = "SMC")
```

Generating the plot can also be suppressed if the output is printed explicitly:

```{r}
# determine the number of factors to retain using parallel analysis
print(PARALLEL(DOSPERT_sub, eigen_type = "SMC"), plot = FALSE)
```


Other factor retention methods can be used accordingly. For example, to use the empirical Kaiser criterion, use the `EKC` function:

```{r}
# determine the number of factors to retain using parallel analysis
print(EKC(DOSPERT_sub), plot = FALSE)
```

The following factor retention methods are currently implemented: comparison data (`CD()`), empirical Kaiser criterion (`EKC()`), the hull method (`HULL()`), the Kaiser-Guttman criterion (`KGC()`), parallel analysis (`PARALLEL()`), scree test (`SCREE()`), and sequential model tests (`SMT()`). Many of these functions have multiple versions of the respective factor retention method implemented, for example, the parallel analysis can be done based on eigenvalues found using unity (principal components) or SMCs, or on an EFA procedure. Another example is the hull method, which can be used with different fitting methods (principal axis factoring [PAF], maximum likelihood [ML], or unweighted least squares [ULS]), and different goodness of fit indices. Please see the respective function documentations for details.

### Run Multiple Factor Retention Methods With `N_FACTORS()`

If you want to use multiple factor retention methods, for example, to compare whether different methods suggest the same number of factors, it is easier to use the `N_FACTORS()` function. This is a wrapper around all the implemented factor retention methods. Moreover, it also enables to run the Bartlett's test of sphericity and compute the KMO criterion.

For example, to test the suitability of the data for factor analysis and to determine the number of factors to retain based on parallel analysis (but only using eigen values based on SMCs and PCA), the EKC, and the sequential model test, we can run the following code:

```{r}
N_FACTORS(DOSPERT_sub, criteria = c("PARALLEL", "EKC", "SMT"),
          eigen_type_other = c("SMC", "PCA"))
```


If all possible factor retention methods should be used, it is sufficient to provide the data object (note that this takes a while, as the comparison data is computationally expensive and therefore relatively slow method, especially if larger datasets are used). We additionally specify the method argument to use unweighted least squares (ULS) estimation. This is a bit faster than using principle axis factoring (PAF) and it enables the computation of more goodness of fit indices:

```{r}
N_FACTORS(DOSPERT_sub, method = "ULS")
```

Now, this is not the scenario one is happy about, but it still does happen: There is no obvious convergence between the methods and thus the choice of the number of factors to retain becomes rather difficult (and to some extend arbitrary). We will proceed with 6 factors, as it is what is typically used with DOSPERT data, but this does not mean that other number of factors are not just as plausible.

Note that all factor retention methods, except comparison data (CD), can also be used with correlation matrices. We use `method = "ULS"` and `eigen_type_other = c("SMC", "PCA")` to skip the slower criteria. In this case, the sample size has to be specified:

```{r}
N_FACTORS(test_models$baseline$cormat, N = 500,
          method = "ULS", eigen_type_other = c("SMC", "PCA"))
```

## Exploratory Factor Analysis: Factor Extraction

Multiple algorithms to perform an EFA and to rotate the found solutions are implemented in the `EFAtools` package. All of them can be used using the `EFA()` function. To perform the EFA, you can use one of principal axis factoring (PAF), maximum likelihood estimation (ML), and unweighted least squares (ULS; also sometimes referred to as MINRES). To rotate the solutions, the `EFAtools` package offers varimax and promax rotations, as well as the orthogonal and oblique rotations provided by the `GPArotation` package (i.e., the `GPArotation` functions are called in the `EFA()` function in this case).

You can run an EFA with PAF and no rotation like this:

```{r}
EFA(DOSPERT_sub, n_factors = 6)
```

To rotate the loadings (e.g., using a promax rotation) adapt the `rotation` argument:

```{r}
EFA(DOSPERT_sub, n_factors = 6, rotation = "promax")
```

This now performed PAF with promax rotation with the specification, on average, we found to produce the most accurate results in a simulation analysis (see function documentation). If you want to replicate the implementation of the *psych* R package, you can set the `type` argument to `"psych"`:

```{r}
EFA(DOSPERT_sub, n_factors = 6, rotation = "promax", type = "psych")
```

If you want to use the *SPSS* implementation, you can set the `type` argument to `"SPSS"`: 

```{r}
EFA(DOSPERT_sub, n_factors = 6, rotation = "promax", type = "SPSS")
```

This enables comparisons of different implementations. The `COMPARE()` function provides an easy way to compare how similar two loading (pattern) matrices are:

```{r}
COMPARE(
  EFA(DOSPERT_sub, n_factors = 6, rotation = "promax", type = "psych")$rot_loadings,
  EFA(DOSPERT_sub, n_factors = 6, rotation = "promax", type = "SPSS")$rot_loadings
)

```

*Why would you want to do this?* One of us has had the experience that a reviewer asked whether the results can be reproduced in another statistical program than R. We therefore implemented this possibility in the package for an easy application of large scale, systematic comparisons.

Note that the `type` argument of the `EFA()` function only affects the implementations of principal axis factoring (PAF), varimax and promax rotations. The other procedures are not affected (except the order of the rotated factors for the other rotation methods).

As indicated previously, it is also possible to use different estimation and rotation methods. For example, to perform an EFA with ULS and an oblimin rotation, you can use the following code:

```{r}
EFA(DOSPERT_sub, n_factors = 6, rotation = "oblimin", method = "ULS")
```

Of course, `COMPARE()` can also be used to compare results from different estimation or rotation methods (in fact, to compare any two matrices), not just from different implementations:

```{r}
COMPARE(
  EFA(DOSPERT_sub, n_factors = 6, rotation = "promax")$rot_loadings,
  EFA(DOSPERT_sub, n_factors = 6, rotation = "oblimin", method = "ULS")$rot_loadings,
  x_labels = c("PAF and promax", "ULS and oblimin")
)
```

Finally, if you are interested in factor scores from the EFA solution, these can be obtained with `FACTOR_SCORES()`, a wrapper for `psych::factor.scores()` to be used directly with an output from `EFA()`:

```{r}
EFA_mod <- EFA(DOSPERT_sub, n_factors = 6, rotation = "promax")
fac_scores <- FACTOR_SCORES(DOSPERT_sub, f = EFA_mod)
```


### Performance

To improve performance of the iterative procedures (currently the parallel analysis, and the PAF, ML, and ULS methods) we implemented some of them in C++. For example, the following code compares the EFAtools parallel analysis with the corresponding one implemented in the psych package (the default of `PARALLEL()` is to use 1000 datasets, but 25 is enough to show the difference):

```{r message=FALSE}
microbenchmark::microbenchmark(
  PARALLEL(DOSPERT_sub, eigen_type = "SMC", n_datasets = 25),
  psych::fa.parallel(DOSPERT_sub, SMC = TRUE, plot = FALSE, n.iter = 25)
)
```


Moreover, the following code compares the PAF implementation (of type "psych") of the EFAtools package with the one from the psych package: 

```{r message=FALSE}
microbenchmark::microbenchmark(
  EFA(DOSPERT_raw, 6),
  psych::fa(DOSPERT_raw, 6, rotate = "none", fm = "pa")
)
```

While these differences are not large, they grow larger the more iterations the procedures need, which is usually the case if solutions are more tricky to find. Especially for simulations this might come in handy. For example, in one simulation analysis we ran over 10,000,000 EFAs, thus a difference of about 25 milliseconds per EFA leads to a difference in runtime of almost three days.

### Model Averaging

Instead of relying on one of the many possible implementations of, for example, PAF, and of using just one rotation (e.g., promax), it may be desirable to average different solutions to potentially arrive at a more robust, average solution. The `EFA_AVERAGE()` function provides this possibility. In addition to the average solution it provides the variation across solutions, a matrix indicating the robustness of indicator-to-factor correspondences, and a visualisation of the average solution and the variability across solutions. For example, to average across all available factor extraction methods and across all available oblique rotations, the following code can be run:

```{r}
# Average solution across many different EFAs with oblique rotations
EFA_AV <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                      method = c("PAF", "ML", "ULS"), rotation = "oblique",
                      show_progress = FALSE)

# look at solution
EFA_AV
```


The first matrix of the output tells us that the indicators are mostly allocated to the same factors. However, that some rowsums are larger than one also tells as that there likely are some cross loadings present in some solutions. Moreover, the relatively high percentages of salient pattern coefficients all loading on the first factor may indicate that some rotation methods failed to achieve simple structure and it might be desirable to exclude these from the model averaging procedure. The rest of the output is similar to the normal `EFA()` outputs shown above, only that in addition to the average coefficients their range is also shown. Finally, the plot shows the average pattern coefficients and their ranges. 

**Important disclaimer:** While it is possible that this approach provides more robust results, we are unaware of simulation studies that have investigated and shown this. Therefore, it might make sense to for now use this approach mainly to test the robustness of the results obtained with one single EFA implementation.

## Exploratory Factor Analysis: Schmid-Leiman transformation and McDonald's Omegas

For the Schmid-Leiman transformation and computation of omegas, we will use PAF and promax rotation:

```{r}
efa_dospert <- EFA(DOSPERT_sub, n_factors = 6, rotation = "promax")
efa_dospert
```

The indicator names in the output (i.e., the rownames of the rotated loadings section) tell us which domain (out of ethical, financial, health, recreational, and social risks) an indicator stems from. From the pattern coefficients it can be seen that these theoretical domains are recovered relatively well in the six factor solution, that is, usually, the indicators from the same domain load onto the same factor. When we take a look at the factor intercorrelations, we can see that there are some strong and some weak correlations. It might be worthwhile to explore whether a general factor can be obtained, and which factors load more strongly on it. To this end, we will use a Schmid-Leiman (SL) transformation.

## Schmid-Leiman Transformation

The SL transformation or orthogonalization transforms an oblique solution into a hierarchical, orthogonalized solution. To do this, the `EFAtools` package provides the `SL()` function.

```{r}
sl_dospert <- SL(efa_dospert)
sl_dospert
```

From the output, it can be seen that all, except the social domain indicators substantially load on the general factor. That is, the other domains covary substantially.

## McDonald's Omegas

Finally, we can compute omega estimates and additional indices of interpretive relevance based on the SL solution. To this end, we can either specify the variable-to-factor correspondences, or let them be determined automatically (in which case the highest factor loading will be taken, which might lead to a different solution than what is desired, in the presence of cross-loadings). Given that no cross-loadings are present here, it is easiest to let the function automatically determine the variable-to-factor correspondence. To this end, we will set the `type` argument to `"psych"`.

```{r}
OMEGA(sl_dospert, type = "psych")
```

If we wanted to specify the variable to factor correspondences explicitly (for example, according to theoretical expectations), we could do it in the following way:

```{r}
OMEGA(sl_dospert, factor_corres = matrix(c(rep(0, 18), rep(1, 6), rep(0, 30), 
                                         rep(1, 6), rep(0, 6), 1, 0, 1, 0, 1,
                                         rep(0, 19), rep(1, 6), rep(0, 31), 1, 0,
                                         1, 0, 1, rep(0, 30), rep(1, 6), 
                                         rep(0, 12)), ncol = 6, byrow = FALSE))
```

---
title: "Replicate SPSS and R psych results with EFAtools"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Replicate_SPSS_psych}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.align = "center"
)


if (!requireNamespace("psych", quietly = TRUE)) {
      stop("Package \"psych\" needed for this vignette to work. Please install it.",
      call. = FALSE)
}

```

This vignette demonstrates the replication of exploratory factor analysis (EFA) results (specifically, principal axis factoring [PAF] with subsequent promax rotation) from the SPSS `FACTOR` algorithm and from the `fa()` function from the `psych` R package.
For a general introduction to the `EFAtools` package, please see the [**EFAtools**](EFAtools.html "EFAtools") vignette. Same as in the EFAtools vignette, we will use the DOSPERT data set for this demonstration as well (see `?DOSPERT` for details).

First load the needed packages EFAtools and psych (original SPSS results for some data sets are available in the EFAtools package).

```{r setup}
library(psych)
library(EFAtools)
```

## Principal Axis Factoring

First, we will fit an EFA with PAF and without rotation using the `EFA` function from `EFAtools` using `type = "psych"` and `type = "SPSS"`. These types are intended to mimic the R psych and SPSS results, respectively.

```{r}
# EFAtools::EFA with type = "psych" without rotation
EFA_psych_paf <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych")
# EFAtools::EFA with type = "SPSS" without rotation
EFA_SPSS_paf <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS")
```

As a next step, we fit an EFA with the same configurations (PAF and no rotation) using the `fa` function from `psych` with the same data set.

```{r}
# psych::fa without rotation
psych_paf <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "none")
```

Now we can compare results from `EFA` with the respective types to the original R `psych` and `SPSS` results using the same data set. This is easily done using the `COMPARE` function available in the `EFAtools` package.

```{r}
# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_paf$unrot_loadings, psych_paf$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_paf$unrot_loadings, SPSS_27$DOSPERT$paf_load)

```

To see that this close match was not trivial, we can look at the match between the original R `psych` and `SPSS` solutions.

```{r}
# Compare loadings from psych::fa and SPSS
COMPARE(psych_paf$loadings, SPSS_27$DOSPERT$paf_load)
```

We can see that the solutions are slightly different, especially for the 9th and 10th factor. Although the differences are very small here, they can get quite large for other data sets, or get larger after rotation (see below).

## Varimax Rotation

Now we confirmed the replication of PAF results without rotation, we can continue to compare rotated factor solutions. We start by comparing varimax rotated PAF solutions.


```{r}
## Fit the models

# EFAtools::EFA with type = "psych" with varimax rotation
EFA_psych_var <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych", rotation = "varimax")
# EFAtools::EFA with type = "SPSS" with varimax rotation
EFA_SPSS_var <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS", rotation = "varimax")
# psych::fa with varimax rotation
psych_var <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "varimax")

## Check replication of results

# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_var$rot_loadings, psych_var$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_var$rot_loadings, SPSS_27$DOSPERT$var_load)

## Compare original results (just to see the difference)

# Compare loadings from psych::fa and SPSS
COMPARE(psych_var$loadings, SPSS_27$DOSPERT$var_load)

```

## Promax Rotation

Finally, we can do the same for promax rotated results as well.

```{r}
## Fit the models

# EFAtools::EFA with type = "psych" with promax rotation
EFA_psych_pro <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                     type = "psych", rotation = "promax")
# EFAtools::EFA with type = "SPSS" with promax rotation
EFA_SPSS_pro <- EFA(DOSPERT$cormat, n_factors = 10, N = DOSPERT$N,
                    type = "SPSS", rotation = "promax")
# psych::fa with promax rotation
psych_pro <- psych::fa(DOSPERT$cormat, nfactors = 10, n.obs = DOSPERT$N,
                       fm = "pa", rotate = "Promax")

## Check replication of results

# Compare loadings from psych::fa and EFAtools::EFA with type = "psych"
COMPARE(EFA_psych_pro$rot_loadings, psych_pro$loadings)

# Compare loadings from SPSS and EFAtools::EFA with type = "SPSS"
COMPARE(EFA_SPSS_pro$rot_loadings, SPSS_27$DOSPERT$pro_load)

## Compare original results (just to see the difference)

# Compare loadings from psych::fa and SPSS
COMPARE(psych_pro$loadings, SPSS_27$DOSPERT$pro_load)

```

---
title: "EFA with ordinal and missing data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{EFA with ordinal and missing data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.align = "center"
)

# The multiple-imputation section uses mice to create the imputations. It is only
# suggested by the package, so the imputation chunks are evaluated only when it is
# installed.
mice_ok <- requireNamespace("mice", quietly = TRUE)
```

Two features of real data complicate an exploratory factor analysis: items are often
**ordinal** (a handful of Likert categories rather than a continuous scale), and some
responses are usually **missing**. `EFAtools` handles both within the ordinary `efa_fit()`
workflow, without switching packages. This vignette shows how. It assumes familiarity with
the basic workflow covered in the [EFAtools](EFAtools.html) vignette and focuses on what
changes for ordinal and incomplete data.

```{r}
library(EFAtools)
```

So that the examples are self-contained and reproducible, we generate the data with
`efa_simulate()` from a known three-factor population (18 indicators, six per factor, with
moderately correlated factors), using fixed seeds throughout.

```{r}
Lambda <- population_models$loadings$baseline  # 18 x 3 loading pattern
Phi    <- population_models$phis_3$moderate    # moderate factor intercorrelations
```

## Ordinal Data

Rating-scale items are not continuous: they take a few ordered values, and a Pearson
correlation between two such items underestimates the association between the underlying
constructs. The polychoric correlation instead estimates the correlation of the continuous
latent variables assumed to underlie the observed categories, and pairing it with a
categorical estimator removes the bias that treating the items as continuous introduces.

We draw 400 responses on a four-category scale. Because the latent data are normal, cutting
them at the standard-normal category thresholds already leaves the population *polychoric*
correlation of the discretised data equal to the target correlation; `match = "polychoric"`
records that this is what we are after.

```{r}
d_ord <- efa_simulate(N = 400, Lambda = Lambda, Phi = Phi,
                      categories = 4, match = "polychoric", seed = 2024)$data
d_ord[1:5, 1:6]
```

### Screening Ordinal Data

`efa_screen()` reports, among its diagnostics, how many response categories each item has
and whether the data are multivariate normal — the two things that decide whether an
ordinal treatment is worthwhile.

```{r, warning = FALSE}
efa_screen(d_ord, seed = 42)
```

The sampling adequacy (KMO) and sphericity checks confirm the data are factorable. The
telling parts are the multivariate-normality section and the recommendations: Mardia's
kurtosis and the Henze-Zirkler test reject normality, and the recommendations flag that every
item has fewer than five response categories. Together, these point to the same conclusion.
With few categories and non-normal data, a polychoric correlation with a categorical
estimator such as DWLS is less biased than normal-theory maximum likelihood. The
normal-theory standard errors and fit indices are also better replaced by robust (sandwich)
versions.

### Polychoric Correlations, DWLS, and Robust Standard Errors

`efa_fit()` computes the polychoric correlation when `cor_method = "poly"` and fits it with
diagonally weighted least squares when `estimator = "DWLS"` — the estimator recommended for
ordinal data because it accounts for how precisely each polychoric correlation is estimated.
Requesting `se = "sandwich"` adds robust standard errors and a scaled (Satorra-Bentler)
chi-square that stay valid under the non-normality these data show.

```{r}
efa_poly <- efa_fit(d_ord, n_factors = 3, cor_method = "poly", estimator = "dwls",
                    rotation = "oblimin", se = "sandwich")
efa_poly
```

The pattern matrix recovers the three factors cleanly (six indicators each), and the model
fit reports a **scaled** chi-square with its CFI, TLI, and RMSEA. Because the chi-square is
a scaled statistic, the AIC and BIC (which are defined on the unscaled likelihood
discrepancy) are left `NA`.

For binary items, `cor_method = "tetra"` computes tetrachoric correlations and runs the same
DWLS and sandwich machinery.

The robust standard errors accompany each estimated quantity; for the rotated loadings, for
example:

```{r}
round(efa_poly$SE$rot_loadings, 3)
```

The matching confidence intervals live in `efa_poly$CI`, and `summary(efa_poly)` prints them
as a labelled table alongside the model diagnostics.

### Why Not Just Treat the Items as Continuous?

To see what the ordinal treatment buys, fit the same data as if they were continuous — a
Pearson correlation with maximum likelihood — and compare the rotated loadings with
`efa_compare()`.

```{r}
efa_cont <- efa_fit(d_ord, n_factors = 3, cor_method = "pearson", estimator = "ML",
                    rotation = "oblimin")

cmp <- efa_compare(efa_poly$rot_loadings, efa_cont$rot_loadings,
                   x_labels = c("Polychoric / DWLS", "Pearson / ML"))
cmp
plot(cmp)
```

The two solutions agree on the structure, but the polychoric loadings are systematically a
little larger: treating the items as continuous attenuates the loadings, because the Pearson
correlation understates the latent associations. With only four categories here the gap is
modest, but it widens as the number of categories drops (it is largest for binary items) and
as the category thresholds grow more asymmetric (skewed items). This is why a polychoric or
tetrachoric treatment is preferable for genuinely ordinal items with few categories.

The polychoric route does make its own demands, though: it assumes a normal latent variable
underlies each item, and it needs an adequate sample size and reasonably populated
response-category combinations. When categories are very sparse (rare responses, small
samples), the polychoric asymptotic covariance behind the DWLS weights and the robust
standard errors becomes unreliable, and collapsing rare categories can help.

## Missing Data

When some responses are missing, dropping every incomplete case (listwise deletion) wastes
data and can bias the results unless the values are missing completely at random. `EFAtools`
offers two principled alternatives that assume only that the data are missing at random
(MAR): a single-analysis route via full-information maximum likelihood, and a
multiple-imputation route via `efa_mi()`.

We simulate 250 continuous cases with about 15% of values missing at random, where each
item's missingness depends on another item's value. By default, `efa_simulate()` would make
each item's missingness depend on another item that is itself partly missing — a kind of
missingness the two routes below are not designed for. Rather than failing outright, they run
with a residual bias that grows as `missing_prop` and `missing_strength` increase.

`missing_vars` and `missing_predictor` avoid that. Here the first nine items carry the
missing values and each is driven by one of the last nine, which stay complete. Every
predictor is therefore fully observed, satisfying the *ignorable* MAR assumption the two
routes below rely on -- not just MAR, but MAR with no incomplete predictor left unaccounted
for.

```{r}
d_miss <- efa_simulate(N = 250, Lambda = Lambda, Phi = Phi,
                       missing = "MAR", missing_prop = 0.15,
                       missing_vars = 1:9, missing_predictor = 10:18,
                       seed = 2024)$data
round(mean(is.na(d_miss)), 3)          # overall proportion missing
round(colMeans(is.na(d_miss)), 3)      # holed items only
```

### Two-Stage Full-Information Maximum Likelihood

With `cor_method = "fiml"`, `efa_fit()` estimates the saturated mean and covariance from all
the observed data by an EM algorithm (assuming the data are MAR) and analyses the resulting
correlation — a single fit that uses every case rather than only the complete ones. The
model fit is reported as corrected two-stage (Satorra-Bentler) statistics.

```{r}
efa_fiml <- efa_fit(d_miss, n_factors = 3, cor_method = "fiml", estimator = "ml",
                    rotation = "oblimin")
efa_fiml
```

The solution again recovers the three factors, and the printout records that the correlation
was obtained by two-stage FIML. Standard errors are available here too: for `estimator = "ML"`
or `"ULS"`, `se = "information"` or `"sandwich"` return the corrected two-stage standard
errors, and `se = "np-boot"` works with any estimator.

FIML estimates the model from every case, but it does not fill in the missing values. Any step
that needs complete rows is therefore still complete-case. `efa_scores(d_miss, f = efa_fiml)`,
for example, scores only the complete cases and warns that the rest are `NA`:

```{r}
sum(complete.cases(d_miss))    # cases a score can be formed for
```

Use the `efa_mi()` route below if you want case-level scores for every respondent. Each imputed
dataset is complete, so scoring it with the pooled loadings covers all 250 cases.

The correction itself needs a well-behaved saturated covariance. When that cannot be formed —
typically in a small sample with a high proportion of missing values, or with near-collinear
items — `efa_fit()` warns and keeps the plain two-stage likelihood-ratio statistic instead of
discarding the test. Such a fit is never presented as corrected: the printed chi-square line
is labelled **uncorrected**, and the accompanying *p*-value, CFI, TLI, and RMSEA should be
read as indicative only.

### Multiple Imputation with `efa_mi()`

The alternative is to impute the missing values several times, fit each completed dataset,
and pool the results. `EFAtools` does not impute the data itself — use a dedicated tool such
as the [mice](https://CRAN.R-project.org/package=mice) package — but `efa_mi()` takes the
list of completed datasets and does the factor-analytic pooling.

Here we create five imputations with mice (a Bayesian linear-regression model, appropriate
for these continuous items) and collect them into a list.

```{r, eval = mice_ok}
imp <- mice::mice(as.data.frame(d_miss), m = 5, method = "norm",
                  printFlag = FALSE, seed = 123)
dat_list <- lapply(seq_len(imp$m), function(i) mice::complete(imp, i))
```

`efa_mi()` fits the same `efa_fit()` model to each imputed dataset — the extraction,
rotation, and standard-error options are passed through `...` — aligns the solutions to a
common factor space (rotation is only identified up to reflection and permutation, so the
imputations must be matched before averaging), and pools them.

```{r, eval = mice_ok}
efa_pooled <- efa_mi(dat_list, n_factors = 3, estimator = "ml", rotation = "oblimin")
efa_pooled
```

The pooled loadings recover the three factors. Point estimates are averaged across the
imputations after alignment. The model chi-square, and the AIC/BIC derived from it, are
pooled with the same rule as RMSEA, but applied to a different discrepancy scale — which is
why the printout labels the pooled chi-square **D2-pooled**, and why you should not expect it
to reconcile by hand with the pooled RMSEA. The incremental CFI
and TLI are instead averaged across the per-imputation fits. Requesting standard errors in
the call (for example `se = "information"` or `se = "np-boot"`) additionally pools them with
Rubin's rules, so the between-imputation variability inflates the pooled standard errors.
Because multiple imputation propagates the extra uncertainty from the missing data, its
pooled fit statistics are not directly comparable with the single FIML fit above; read them
together with the per-imputation fits stored in the returned object.

Which route to prefer is largely practical. FIML is a single, efficient fit and is the
simpler default when the analysis model is the whole story. Multiple imputation is more
flexible when the imputation model should draw on auxiliary variables not in the factor
model, or when the same imputations feed several downstream analyses.

## Ordinal *and* Missing Data

Questionnaire data are usually both: a few response categories *and* some unanswered items.
The two treatments above do not simply combine, because the ordinal machinery needs complete
cases exactly where the FIML route needs continuous ones. We draw the same four-category
items as before, now with 8% of the responses missing completely at random.

```{r}
d_ord_miss <- efa_simulate(N = 300, Lambda = Lambda, Phi = Phi,
                           categories = 4, match = "polychoric",
                           missing = "MCAR", missing_prop = 0.08, seed = 2024)$data

round(mean(is.na(d_ord_miss)), 3)     # overall proportion missing
sum(complete.cases(d_ord_miss))       # respondents who answered every item
```

Eight percent per item is mild, but spread over 18 items it leaves only a fifth of the
sample complete. That number, not the 300 rows supplied, is what the ordinal route has to
work with.

### The Effective Sample of a Polychoric DWLS Fit

Polychoric correlations can be estimated pair by pair, using every respondent who answered
a given pair of items. Their asymptotic covariance cannot: the DWLS weights and the
sandwich standard errors describe *one* set of estimates from *one* set of cases, so
whenever they are requested the correlations and the covariance are both computed on the
listwise-complete rows. `efa_fit()` announces that override rather than applying it
silently.

```{r}
efa_ord_miss <- efa_fit(d_ord_miss, n_factors = 3, cor_method = "poly",
                        estimator = "dwls", rotation = "oblimin")
efa_ord_miss$settings$N        # cases the fit is actually based on
```

Every loading, fit index, standard error, and confidence interval of that fit rests on the
complete cases. `efa_screen()` names both denominators for the same reason — its
multivariate-normality and outlier results state how many complete cases they used out of
how many rows were supplied — so the reduction is visible before a model is fitted.

Two adjustments avoid it, each with a cost:

- Ask for no asymptotic covariance. With `estimator = "ULS"` or `"ML"` and no sandwich
  standard errors, nothing needs the weight matrix, so the polychoric matrix stays
  pairwise-complete and every case contributes to the pairs it answered. In exchange you
  give up the DWLS weighting and the robust standard errors, and the reported `N` is the
  nominal row count while each correlation rests on its own smaller subset — so fit
  statistics and analytic standard errors treat the data as more complete than they are.
- Switch to FIML — but only as a continuous analysis. `cor_method = "fiml"` estimates a
  saturated *normal* covariance and analyses the resulting Pearson-type correlation; there
  is no polychoric asymptotic covariance in it, so it cannot supply DWLS with weights.
  Combining the two is refused rather than silently approximated:

```{r, error = TRUE}
efa_fit(d_ord_miss, n_factors = 3, cor_method = "fiml", estimator = "dwls")
```

### Keeping Both the Ordinal Treatment and the Cases

Multiple imputation is the route that keeps both. Each imputed dataset is complete, so the
polychoric correlations and their asymptotic covariance are estimated on all 300 cases, and
the imputation uncertainty is carried into the pooled results. Since `efa_mi()` forwards its
arguments to `efa_fit()`, this is the ordinal fit above applied to each imputation — here
with predictive mean matching, which resamples observed values and so respects the response
scale.

```{r, eval = mice_ok}
imp_ord <- mice::mice(as.data.frame(d_ord_miss), m = 5, method = "pmm",
                      printFlag = FALSE, seed = 123)
dat_ord <- lapply(seq_len(imp_ord$m), function(i) mice::complete(imp_ord, i))

efa_ord_pooled <- efa_mi(dat_ord, n_factors = 3, cor_method = "poly",
                         estimator = "dwls", rotation = "oblimin")
efa_ord_pooled
```

The pooled pattern matrix is estimated from all 300 respondents rather than the 62 complete
ones — compare it with `efa_ord_miss$rot_loadings` — and the between-imputation variability
enters any pooled standard errors requested in the call. As always with imputation, the
imputation model has to be defensible: for ordinal items that means a method returning
values on the observed scale (predictive mean matching, or an ordinal model such as
`method = "polr"`), and enough respondents per item to estimate it.

So, for ordinal items with missing values: use `efa_mi()` when the missingness is more than
incidental and the ordinal treatment matters; fit `cor_method = "poly"` with
`estimator = "DWLS"` directly when the complete-case sample is still comfortably large; and
use FIML when the items have enough categories to be treated as continuous in the first
place.

## Where to Next

This vignette covered the ordinal and missing-data extensions of the workflow. For the core
analysis — screening, factor retention, extraction and rotation, and the post-processing
tools — see the [EFAtools](EFAtools.html) vignette, and the individual help pages for the
statistical details and references. Run `browseVignettes("EFAtools")` for the vignettes
installed with the package, or visit the
[package website](https://mdsteiner.github.io/EFAtools/) for the full set of articles.

---
title: "Migrating to the efa_* interface"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Migrating to the efa_* interface}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`EFAtools` exposes its functionality through a set of lowercase `efa_*` functions:
`efa_screen()`, `efa_retain()`, `efa_fit()`, and so on. Earlier releases used uppercase
names instead (`EFA()`, `N_FACTORS()`, `OMEGA()`, ...). The uppercase names have **not** been
removed — they still work — but the `efa_*` functions are now the recommended interface. This
vignette explains the change, gives the full old-to-new mapping, and walks through the
migration with the largest change to the argument list: `EFA()` to `efa_fit()`.

```{r}
library(EFAtools)
```

## Why the efa_* Interface

The `efa_*` names are the interface we recommend for new code. They read consistently,
share a common prefix that groups them in tab-completion and the documentation index, and
they are where the package's development now happens.

The uppercase names are **superseded**, not deprecated. In practice that means:

* They keep working with their **original arguments**, and calling them emits **no warning
  or message**. Existing scripts and saved analyses continue to run byte-for-byte
  unchanged.
* Their argument lists are **frozen**. New capabilities are added only to the `efa_*`
  functions, so an uppercase name will never grow a new argument.

So there is no urgency to rewrite working code. Migrate a script when you want the new
functions' additional features (or simply for consistency); until then the old calls
remain valid.

## The Old-to-New Mapping

Every uppercase function has a lowercase counterpart. For most of them the migration is a
pure rename: the arguments are identical and only the name changes.

| Old name | Recommended name |
|---|---|
| `EFA()` | `efa_fit()` |
| `N_FACTORS()` | `efa_retain()` |
| `EFA_AVERAGE()` | `efa_average()` |
| `EFA_POOLED()` | `efa_mi()` |
| `COMPARE()` | `efa_compare()` |
| `SL()` | `efa_schmid_leiman()` |
| `OMEGA()` | `efa_reliability()` |
| `FACTOR_SCORES()` | `efa_scores()` |
| `PROCRUSTES()` | `efa_procrustes()` |
| `BARTLETT()` | `efa_bartlett()` |
| `KMO()` | `efa_kmo()` |
| `PARALLEL()` | `efa_parallel()` |
| `EKC()` | `efa_ekc()` |
| `KGC()` | `efa_kgc()` |
| `HULL()` | `efa_hull()` |
| `SCREE()` | `efa_scree()` |
| `MAP()` | `efa_map()` |
| `NEST()` | `efa_nest()` |
| `SMT()` | `efa_smt()` |
| `CD()` | `efa_cd()` |

`efa_reliability()` and `efa_scores()` are broader than the `OMEGA()` and `FACTOR_SCORES()`
they replace, but they cover the same use cases and are the recommended way to obtain those
quantities going forward.

Alongside the renamed functions, the package ships tools that have **no uppercase
predecessor** — they are new, and available only under the `efa_*` interface (and the two
control constructors):

| New function | Purpose |
|---|---|
| `efa_screen()` | Data screening and factorability diagnostics in one report |
| `efa_group()` | Multigroup EFA with factor congruence |
| `efa_simulate()` | Simulate data from a factor model |
| `efa_power()` | Power analysis for EFA |
| `estimate_control()` | Bundle the estimation tuning knobs (see below) |
| `rotate_control()` | Bundle the rotation tuning knobs (see below) |

For a plain rename, the call is unchanged apart from the name. For example, the
Kaiser-Meyer-Olkin criterion:

```{r}
cor_mat <- test_models$baseline$cormat

# Recommended name -- exactly the same arguments as KMO():
efa_kmo(cor_mat)

# The old name still works and returns the same result:
identical(KMO(cor_mat)$KMO, efa_kmo(cor_mat)$KMO)
```

## Migrating from `EFA()` to `efa_fit()`

`EFA()` exposed every estimation and rotation setting as a flat argument, which made for a
long and somewhat unwieldy signature. `efa_fit()` keeps the **primary** choices — the data,
factor count, sample size, estimator, rotation, and a few others — as top-level arguments
(see `?efa_fit` for the full list), and collects the **tuning** knobs into two small control
objects built by `estimate_control()` and `rotate_control()`.

Each flat `EFA()` tuning argument now lives in one of the two controls:

| Control object | Arguments it holds |
|---|---|
| `estimate_control()` | `type`, `init_comm`, `criterion`, `criterion_type`, `max_iter`, `abs_eigen`, `start_method` |
| `rotate_control()` | `type`, `normalize`, `precision`, `order_type`, `varimax_type`, `p_type`, `k`, `random_starts` |

A few points make the translation mechanical:

* **The `type` preset feeds both controls.** `EFA()` had a single `type`
  (`"EFAtools"`, `"psych"`, `"SPSS"`, or `"none"`) that governed both estimation and
  rotation. Pass the same `type` to each constructor to reproduce it. Because the two
  controls are independent, you *can* now give them different presets, but you do not have
  to.
* **Three arguments were renamed.** The estimator is selected with `estimator` instead of
  `method` — passing `method` to `efa_fit()` is an error that points to the new name, while
  `EFA()` keeps its `method` argument. `P_type` is now `p_type`, and `randomStarts` is now
  `random_starts`; `EFA()` still accepts those old spellings silently.
* **Rotation-engine extras** — the criterion-specific arguments a rotation may take, such
  as `maxit`, `gam` (oblimin), or `delta` (geomin) — are passed through `rotate_control()`'s
  `...` (or, equivalently, `efa_fit()`'s own `...`). Both validate the names: an extra the
  selected rotation's engine cannot consume (for example `gamma`, a misspelling of oblimin's
  `gam`) is rejected with an error, where `EFA()` silently ignores it.

The side-by-side below reproduces an SPSS-style analysis. The old flat call and the new
control-based call give identical results:

```{r}
# Old flat interface
efa_old <- EFA(cor_mat, n_factors = 3, N = 500, type = "SPSS")

# New interface: the SPSS preset travels through the two control objects
efa_new <- efa_fit(cor_mat, n_factors = 3, N = 500,
                   estimate_control = estimate_control(type = "SPSS"),
                   rotate_control = rotate_control(type = "SPSS"))

# Identical numerical result
all.equal(efa_old$rot_loadings, efa_new$rot_loadings)
```

An individual knob is set on the control it belongs to. A flat
`EFA(..., type = "SPSS", max_iter = 500, k = 3)` becomes:

```{r, eval = FALSE}
efa_fit(cor_mat, n_factors = 3, N = 500, rotation = "promax",
        estimate_control = estimate_control(type = "SPSS", max_iter = 500),
        rotate_control   = rotate_control(type = "SPSS", k = 3))
```

One behavioural difference is worth knowing about. `efa_fit()` **rejects** a tuning knob
passed directly, rather than silently ignoring it. This turns a common and previously
invisible mistake into an immediate, informative error that names the constructor the knob
belongs to:

```{r, eval = FALSE}
efa_fit(cor_mat, n_factors = 3, N = 500, max_iter = 500)
#> Error: `max_iter` cannot be passed to `efa_fit()` directly.
#> i The estimation and rotation tuning knobs live in `estimate_control()` and `rotate_control()`.
#> i For example: `efa_fit(x, ..., estimate_control = estimate_control(max_iter = 500))`.
```

The same move — collecting flat estimation knobs into `estimate_control()` — applies to the
other functions that fit a model internally: `efa_retain()`, `efa_schmid_leiman()`, and the
retention criteria that fit a model or use EFA-based eigenvalues — `efa_parallel()`,
`efa_kgc()`, `efa_scree()`, `efa_hull()`, `efa_nest()`, and `efa_smt()`. `efa_average()` is the
exception: it keeps its flat estimation and rotation arguments, and only renames `P_type` to
`p_type`.

## What the `type` Presets Replicate

The `type` presets are more than shorthand for a bundle of defaults. `"SPSS"` and `"psych"`
follow how SPSS's FACTOR procedure and `psych::fa()` implement principal axis factoring and
promax rotation, and `"EFAtools"`, the default, combines the settings that performed best in
a systematic comparison of the three
([Grieder and Steiner, 2022](https://doi.org/10.3758/s13428-021-01581-x)). There is one place a
preset alone is not enough: because every preset keeps EFAtools' Kaiser normalization, `"psych"`
also needs `normalize = FALSE` to reproduce `psych::fa()`'s promax — and even then a small
residual difference remains, from ordinary convergence slack between the two implementations.

## Returned Objects Keep Their Legacy Classes

Migrating a call does not break code that inspects or dispatches on the result. The renamed
functions attach the **legacy class alongside the new one**, so `inherits()` checks and S3
methods written against the old classes keep working — including on objects saved from
earlier sessions.

```{r}
class(efa_new)
inherits(efa_new, "EFA")
```

The same holds for the other direct renames: `efa_retain()` objects still inherit
`"N_FACTORS"`, `efa_bartlett()` still inherits `"BARTLETT"`, and likewise for
`efa_kmo()`, `efa_schmid_leiman()`, `efa_compare()`, `efa_average()`, and `efa_mi()`. The
retention criteria (`efa_parallel()`, `efa_ekc()`, and the rest) all share the single
`efa_retention` class they have returned since EFAtools 0.8.0. And the uppercase functions
themselves keep working unchanged: an `EFA()` call returns the same result as before, now
additionally classed `efa` so that it picks up the shared methods — `inherits()` checks
against `"EFA"` are unaffected.

## Where to Next

That is the whole migration: rename the function, and — for `EFA()` only — move the tuning
knobs into `estimate_control()` and `rotate_control()`. For an overview of the full analysis
workflow under the `efa_*` interface, see the [EFAtools](EFAtools.html) vignette; the
individual help pages document each function's arguments in full.

---
title: "EFAtools"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{EFAtools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  message = FALSE,
  fig.width = 7,
  fig.align = "center"
)
```

`EFAtools` provides a complete workflow for exploratory factor analysis (EFA). Starting
from raw data or a correlation matrix, it takes you through every stage of a typical
analysis:

1. **Screening** the data for factorability, multicollinearity, non-normality, and
   outliers (`efa_screen()`, with `efa_bartlett()` and `efa_kmo()` for the individual
   measures).
2. **Retention**: deciding how many factors to keep, with a comprehensive suite of factor
   retention criteria (`efa_retain()` and the individual criteria it wraps).
3. **Extraction and rotation** of the factor solution (`efa_fit()`), with a choice
   of estimators and orthogonal or oblique rotations.
4. **Post-processing**: comparing solutions (`efa_compare()`), averaging across analytic
   choices (`efa_average()`), estimating factor scores (`efa_scores()`), a Schmid-Leiman
   transformation (`efa_schmid_leiman()`), and reliability coefficients including
   McDonald's omegas (`efa_reliability()`).

The computationally intensive routines are implemented in C++ for speed. This vignette
walks through the whole workflow on a single data set; it is an overview rather than a
tutorial on the methods themselves, so please consult the individual help pages for the
statistical details and literature references.

The package can be installed from CRAN using `install.packages("EFAtools")`, or from GitHub
using `pak::pak("mdsteiner/EFAtools")`, and then loaded using:

```{r}
library(EFAtools)
```

We use the `DOSPERT_raw` data set throughout, which contains responses of 3123 participants
to the Domain Specific Risk Taking Scale (DOSPERT); see `?DOSPERT_raw` for details. The
data are rather large, so, purely to keep this vignette quick to build, we work with the
first 500 observations. In a real analysis you would of course use the full data set.

```{r}
# only use a subset to make analyses faster
DOSPERT_sub <- DOSPERT_raw[1:500, ]
```

## Screening the Data

Before extracting factors it is worth checking whether the data are suitable for factor
analysis, and whether they satisfy the assumptions of the estimator you plan to use.
`efa_screen()` collects the relevant diagnostics in one place. From raw data it reports the
Kaiser-Meyer-Olkin (KMO) measure of sampling adequacy (overall and per variable), Bartlett's
test of sphericity, the determinant and condition number of the correlation matrix, the
squared multiple correlation (SMC) and per-variable diagnostics for each item, tests of
multivariate normality (Mardia and Henze-Zirkler), and multivariate outliers, and it closes
with a set of recommendations.

```{r, warning = FALSE}
efa_screen(DOSPERT_sub, seed = 2)
```

KMO measures, on a 0-1 scale, how well each variable is predicted by the others, and
Bartlett's test checks that the correlations are not all zero. The overall KMO is
meritorious and Bartlett's test is clearly significant, so the data are factorable. The determinant looks small, but it is a product of 30 eigenvalues and falls as
variables are added; the condition index (6.8) is the measure that does not move with the
number of variables, and it is well below the value of 30 that flags a near linear
dependency.
More interesting are the last few
sections: three items have a sparse response category, and both the Mardia and the
Henze-Zirkler tests reject multivariate normality. A consequence is that normal-theory
standard errors and fit indices may be biased, so robust (sandwich) or bootstrapped standard
errors are preferable, and Bartlett's test should be read with caution. For a fully ordinal
treatment you would move to polychoric correlations with a diagonally weighted least-squares
(DWLS) estimator.

The outlier section reports something other than the usual case, and it says so. The
diagnostic normally measures each case against a high-breakdown location and scatter,
estimated from a minimum covariance determinant (MCD) subset found by a randomized search.
For these data the search keeps landing on a subset where one item is constant (several
DOSPERT items have a very common lowest response), so the covariance is singular and no valid
subset exists.

`efa_screen()` falls back to classical Mahalanobis distances. It names the reason and labels
the distances it reports. A classical covariance uses every observation, outliers included, so
the outliers inflate the scatter that measures them. The diagnostic is no longer
high-breakdown and it tends to under-flag. Read the 60 flagged cases as a lower bound, and as
a sign of the non-normality the tests above report, rather than as a list of cases to remove.
The MCD does not run, so its random subset search does not run either and `seed` changes
nothing here. `seed` matters only when the robust estimate is available, as in the ordinal
example in the [EFA with ordinal and missing data](Ordinal_and_missing_data.html) vignette.
Inspect the flagged cases (via `$outliers$flagged`) rather than deleting them automatically.

Bartlett's test and the KMO criterion, introduced above, are also available as standalone
functions if you only want the two classic factorability tests:

```{r}
# Bartlett's test of sphericity
efa_bartlett(DOSPERT_sub)

# KMO criterion
efa_kmo(DOSPERT_sub)
```

Both tests can also be requested directly inside `efa_retain()`, alongside the factor
retention criteria discussed next.

## Deciding How Many Factors to Retain

One of the most consequential decisions in an EFA is how many factors to extract. There is a
large number of factor retention criteria, and no single one dominates the others: which
criterion works best depends on the structure of the data — how many indicators there are,
how strong the factors are, and how strongly they intercorrelate. The current recommendation
is therefore to examine several criteria together and check whether they converge (see, for
example, [Auerswald and Moshagen, 2019](https://doi.org/10.1037/met0000200)).

`EFAtools` implements a broad set of these criteria. They can be called individually, or all
(or a selection) of them together through `efa_retain()`.

### Calling Individual Criteria

Each criterion has its own function. For a parallel analysis based on squared multiple
correlations (SMC; also called parallel analysis with principal factors), for example, you
would run:

```{r}
# parallel analysis based on SMC eigenvalues
pa <- efa_parallel(DOSPERT_sub, eigen_type = "smc")
pa
```

Printing the result reports the suggested number of factors. Each criterion also has a
`plot()` method that draws the figure behind the decision — for a parallel analysis, the
observed eigenvalues against the reference eigenvalues from the simulated data:

```{r}
plot(pa)
```

Other criteria work the same way. The empirical Kaiser criterion, for instance:

```{r}
# empirical Kaiser criterion
efa_ekc(DOSPERT_sub)
```

The following criteria are currently implemented: comparison data (`efa_cd()`), the empirical
Kaiser criterion (`efa_ekc()`), the hull method (`efa_hull()`), the Kaiser-Guttman criterion
(`efa_kgc()`), the minimum average partial test (`efa_map()`), the next eigenvalue sufficiency
test (`efa_nest()`), parallel analysis (`efa_parallel()`), the scree test (`efa_scree()`), and
sequential model tests (`efa_smt()`). Many of them offer several variants of the underlying
method. For instance, parallel analysis, can be based on eigenvalues from unity (principal
components), SMCs, or a full EFA. See the individual help pages for the options.

### Running Several Criteria at Once with `efa_retain()`

To compare criteria, it is easier to call `efa_retain()`, a wrapper around the implemented
retention criteria that also runs Bartlett's test and computes the KMO. You can pick a subset
with the `criteria` argument:

```{r}
ret <- efa_retain(DOSPERT_sub,
                  criteria = c("parallel", "ekc", "kgc", "smt", "map"))
ret
```

`plot()` on the result draws the figure for every criterion that has one:

```{r}
plot(ret)
```

Omitting `criteria` does not run every criterion listed above. The default is a subset of six:
comparison data, the empirical Kaiser criterion, the hull method, the minimum average partial
test, the next eigenvalue sufficiency test, and parallel analysis. The call above instead adds
the Kaiser-Guttman criterion and the sequential model tests, and drops comparison data (slow
on larger data sets), the hull method, and the next eigenvalue sufficiency test.

This is the situation one would rather avoid, but it does happen: the criteria disagree, with
suggestions ranging from four to well above ten factors. When there is no clear convergence,
the choice becomes partly a matter of judgement. We proceed with six factors, consistent with the minimum average partial test and the RMSEA-based sequential model test above (this version of the DOSPERT scale is usually
modeled with five factors; the sixth factor here splits the financial-risk items into two).

All criteria except comparison data (which needs raw data) can also be applied to a
correlation matrix, in which case the sample size must be supplied. On a cleaner data set the
criteria often agree:

```{r}
efa_retain(test_models$baseline$cormat, N = 500, estimator = "uls",
           criteria = c("parallel", "ekc", "smt"),
           eigen_type_other = c("smc", "pca"))
```

Here the criteria agree on three factors, with only the conservative RMSEA lower-bound
variant of the sequential model test suggesting two.

## Extracting Factors with `efa_fit()`

`efa_fit()` performs the factor extraction and rotation. For extraction you can choose
principal axis factoring (PAF), maximum likelihood (ML), or unweighted least squares (ULS,
also known as MINRES); diagonally weighted least squares (DWLS) is additionally available
for ordinal data analysed with polychoric or tetrachoric correlations (see the
[EFA with ordinal and missing data](Ordinal_and_missing_data.html) vignette). For rotation
the package offers varimax and promax as well as a range of further rotations (orthogonal
rotations keep the factors uncorrelated, oblique ones let them correlate): quartimax and
equamax are orthogonal; oblimin, quartimin, and simplimax are oblique; and geomin, bentler,
and bifactor come in both an orthogonal form (`geominT`, `bentlerT`, `bifactorT`) and an
oblique one (`geominQ`, `bentlerQ`, `bifactorQ`). The suffix is part of the name: `geomin` on
its own matches both forms and is rejected. All are computed by rotation engines built into
the package.

An EFA with PAF and no rotation is as simple as:

```{r}
efa_fit(DOSPERT_sub, n_factors = 6)
```

To rotate the loadings, set the `rotation` argument. Here we use a promax rotation. We assign the
result, as later sections reuse this solution:

```{r}
efa_dospert <- efa_fit(DOSPERT_sub, n_factors = 6, rotation = "promax")
efa_dospert
```

We can also use a different estimator and rotation, here for example ULS and an oblimin rotation:

```{r}
efa_uls <- efa_fit(DOSPERT_sub, n_factors = 6, rotation = "oblimin", estimator = "uls")
efa_uls
```

Note that `estimator = "uls"` was written in lowercase although the canonical spelling is
`"ULS"`. This holds throughout the `efa_*` interface: wherever an argument takes a fixed set
of values, capitalization does not matter, while the result stores and prints the canonical
spelling. For an argument that takes a single value an unambiguous abbreviation is enough as
well, and a value matching nothing raises an error naming the argument and listing the
choices. Arguments that take several values at once, such as the analytic choices
`efa_average()` averages over, need the full spelling of each value.

## Comparing Solutions with `efa_compare()`

`efa_compare()` provides a quick way to see how similar two loading (pattern) matrices are —
whether they come from different implementations, different estimators, or different rotations.
For instance, to compare the PAF/promax solution with the ULS/oblimin one:

```{r}
efa_compare(efa_dospert$rot_loadings, efa_uls$rot_loadings,
            x_labels = c("PAF and promax", "ULS and oblimin"))
```

## Averaging Across Analytic Choices with `efa_average()`

Comparing two solutions by hand answers the question for those two. `efa_average()` asks it
across a whole grid at once: it runs an EFA for every combination of the settings you allow it
to vary, averages the loadings across the solutions that converged, and reports the range each
loading spans. That range is the useful part — it shows how much of the result is the data and
how much is the analytic choice. Here we vary only the estimator, keeping one implementation
each of PAF, ULS, and ML with a promax rotation:

```{r}
avg_dospert <- efa_average(DOSPERT_sub, n_factors = 6,
                           estimator = c("paf", "uls", "ml"),
                           type = "EFAtools", start_method = "psych")
avg_dospert
```

Left unrestricted, the defaults sweep the different implementations of principal axis
factoring and promax rotation instead, which is considerably more work. See `?efa_average`
for the settings that can be varied, and for `plot()`, which shows the minimum, maximum, and
average loading of every indicator.

## Factor Scores with `efa_scores()`

If you need factor scores, `efa_scores()` estimates them directly from an `efa_fit()` solution
and reports score-quality diagnostics. Printing the object gives the determinacy of each score
and the Guttman indeterminacy index. `summary()` adds the factor weights, the score
intercorrelations, and a score validity/univocality matrix. The off-diagonals of that matrix
are the univocality — how strongly each score correlates with the *other* factors:

```{r}
fac_scores <- efa_scores(DOSPERT_sub, f = efa_dospert)
fac_scores

summary(fac_scores)
```

## Schmid-Leiman Transformation and McDonald's Omegas

For the Schmid-Leiman transformation and the omega coefficients we return to the PAF/promax
solution (`efa_dospert`) from above. Its indicator names encode the risk domain each item
belongs to (ethical, financial, health, recreational, and social), and its pattern
coefficients recovered these theoretical domains fairly well: items from the same domain
tend to load on the same factor. The factor intercorrelations also included some sizeable
values, which raises the question of whether a general factor underlies the domains. A
Schmid-Leiman transformation lets us examine that.

### Schmid-Leiman Transformation

The Schmid-Leiman (SL) transformation orthogonalises an oblique solution into a hierarchical
one with a general factor and orthogonalised group factors. `efa_schmid_leiman()` performs it:

```{r}
sl_dospert <- efa_schmid_leiman(efa_dospert)
sl_dospert
```

The general-factor column shows that the items load substantially on the general factor
across all domains except the social one, whose items load least strongly on it on average
(mean general-factor loading around .19, against roughly .38 to .55 in the other domains) —
that is, the non-social domains covary considerably, while the social domain stands somewhat
apart. Individual social items vary considerably around that average, so this is a statement
about the domain and not about every item in it.

### McDonald's Omegas

Finally, `efa_reliability()` computes omega estimates and related indices from the SL solution.
You can either specify the indicator-to-factor correspondences via the `factor_map` argument or
let the function determine them automatically, in which case each indicator is assigned to its
highest-loading factor. With a reasonably simple structure here, it is easiest to leave
`factor_map` unspecified — a few items do land on a neighbouring domain's factor, which is
exactly what motivates the explicit map shown afterwards:

```{r}
efa_reliability(sl_dospert)
```

To set the correspondences explicitly instead — for example to match a theoretical assignment —
pass a `factor_map` matrix:

```{r}
efa_reliability(sl_dospert,
  factor_map = matrix(c(rep(1, 6), rep(0, 24),                  # F1: ethical
                        rep(0, 18), rep(1, 6), rep(0, 6),       # F2: recreational
                        rep(0, 6), 1, 0, 1, 0, 1, rep(0, 19),   # F3: financial (1, 3, 5)
                        rep(0, 12), rep(1, 6), rep(0, 12),      # F4: health
                        rep(0, 24), rep(1, 6),                  # F5: social
                        rep(0, 7), 1, 0, 1, 0, 1, rep(0, 18)),  # F6: financial (2, 4, 6)
                      ncol = 6, byrow = FALSE))
```

The columns of `factor_map` are matched to the
group factors **by position**, so the assignment has to be read off the solution rather than
assumed: a column naming the wrong set of items yields a well-formed but meaningless subscale
omega instead of an error. Comparing the two tables, the health row (F4) drops more in omega
subscale than in omega total, because three of its six items actually load higher on the
ethical factor and still contribute their true-score variance to the total through that
factor.

Which coefficient to report depends on the question being asked. Omega hierarchical asks
whether a total score measures a single construct. Omega subscale asks whether a subscale
carries anything beyond the general factor. The H index asks whether a factor is well defined
by its indicators. ECV and PUC together ask whether a unidimensional model would be defensible
at all. Alpha is reported alongside them because it is still widely
expected, but it assumes essentially tau-equivalent items — equal true-score contributions —
whereas factor analysis yields congeneric solutions, for which alpha is only a lower bound. For
a multidimensional scale such as the DOSPERT it is therefore rarely the coefficient to report,
and the omegas answer the question alpha is usually asked to answer. See `?efa_reliability` for
the definition of each coefficient and the references behind it.

## Where to Next

This vignette covered the core workflow. `EFAtools` goes further in several directions, each
of which has its own vignette or article:

* [EFA with ordinal and missing data](Ordinal_and_missing_data.html): analysing ordinal
  items with polychoric correlations and handling missing data.
* [Migrating to the efa_* interface](Migrating_to_efa.html): a guide for users coming from
  the older uppercase function names.
* [Multigroup EFA](https://mdsteiner.github.io/EFAtools/articles/Multigroup_EFA.html)
  (on the package website): comparing factor structures across groups.
* [Simulation and power](https://mdsteiner.github.io/EFAtools/articles/Simulation_and_power.html)
  (on the package website): simulating data and running power analyses for EFA.

Run `browseVignettes("EFAtools")` to see the vignettes installed with the package, or visit
the [package website](https://mdsteiner.github.io/EFAtools/) for the full set of articles.

# print output is stable

    Code
      print(om_sl)
    Output
      Omega total, omega hierarchical, omega subscale, H index, explained common
      variance (ECV), and percent of uncontaminated correlations (PUC) for the
      general factor (top row) and omegas and H index for the group factors:
      
           tot  hier   sub     H   ECV   PUC
      g <num> <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num>            
      F2 <num> <num> <num> <num>            
      F3 <num> <num> <num> <num>            

---

    Code
      print(structure(0.85, class = "OMEGA"))
    Output
      Omega total for the single factor: <num>

---

    Code
      print(structure(c(0.85, 2.1), class = "OMEGA"))
    Output
      Omega total and H index for the single factor:
      
      Omega: <num>
      H index: <num>

---

    Code
      print(om_mg6)
    Output
      Omega total, omega hierarchical, omega subscale, H index, explained common
      variance (ECV), and percent of uncontaminated correlations (PUC) for the
      general factor (top row) and omegas and H index for the group factors for each
      group:
      
      Group GroupA:
           tot  hier   sub     H   ECV   PUC
      g <num> <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num>            
      F2 <num> <num> <num> <num>            
      F3 <num> <num> <num> <num>            
      
      Group GroupB:
           tot  hier   sub     H   ECV   PUC
      g <num> <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num>            
      F2 <num> <num> <num> <num>            
      F3 <num> <num> <num> <num>            

---

    Code
      print(om_mg3)
    Output
      Omega total, omega hierarchical, and omega subscale for the general factor (top
      row) and the group factors for each group:
      
      Group GroupA:
           tot  hier   sub
      g <num> <num> <num>
      F1 <num> <num> <num>
      F2 <num> <num> <num>
      F3 <num> <num> <num>
      
      Group GroupB:
           tot  hier   sub
      g <num> <num> <num>
      F1 <num> <num> <num>
      F2 <num> <num> <num>
      F3 <num> <num> <num>

---

    Code
      print(structure(list(GroupA = 0.85, GroupB = 0.8), class = "OMEGA"))
    Output
      Omega total for the single factor for each group:
      
      Group GroupA: <num>
      
      Group GroupB: <num>

---

    Code
      print(structure(list(GroupA = c(0.85, 2.1), GroupB = c(0.8, 1.9)), class = "OMEGA"))
    Output
      Omega total and H index for the single factor for each group:
      
      Group GroupA:
      Omega: <num>
      H index: <num>
      
      Group GroupB:
      Omega: <num>
      H index: <num>

# control objects print

    Code
      print(estimate_control(type = "SPSS"))
    Output
      Estimation control (type: "SPSS")
      
      init_comm: <from type preset>
      criterion: <from type preset>
      criterion_type: <from type preset>
      max_iter: <from type preset>
      abs_eigen: <from type preset>
      start_method: psych
      fiml_max_iter: 500
      fiml_tol: 1e-05

---

    Code
      print(estimate_control(type = "none", init_comm = "smc", criterion = 0.001,
        criterion_type = "sum", max_iter = 300, abs_eigen = TRUE))
    Output
      Estimation control (type: "none")
      
      init_comm: smc
      criterion: 0.001
      criterion_type: sum
      max_iter: 300
      abs_eigen: TRUE
      start_method: psych
      fiml_max_iter: 500
      fiml_tol: 1e-05

---

    Code
      print(rotate_control(type = "psych"))
    Output
      Rotation control (type: "psych")
      
      normalize: TRUE
      precision: 1e-05
      order_type: <from type preset>
      varimax_type: <from type preset>
      p_type: <from type preset>
      k: <from type preset>
      random_starts: 100

---

    Code
      print(rotate_control(type = "none", normalize = FALSE, order_type = "eigen",
        varimax_type = "kaiser", p_type = "norm", k = 4, random_starts = 50, gam = 0.5))
    Output
      Rotation control (type: "none")
      
      normalize: FALSE
      precision: 1e-05
      order_type: eigen
      varimax_type: kaiser
      p_type: norm
      k: 4
      random_starts: 50
      extra_args: gam = 0.5

# print output is stable

    Code
      print(kmo_cor)
    Output
      
      -- Kaiser-Meyer-Olkin criterion (KMO) ------------------------------------------
      
      v The overall KMO value for your data is marvellous.
      These data are probably suitable for factor analysis.
      
      Overall: <num>
      
      For each variable:
         V1    V2    V3    V4    V5    V6    V7    V8    V9   V10   V11   V12   V13 
       <num> <num> <num> <num> <num> <num> <num> <num> <num> <num> <num> <num> <num> 
        V14   V15   V16   V17   V18 
       <num> <num> <num> <num> <num> 

---

    Code
      print(kmo_low)
    Output
      
      -- Kaiser-Meyer-Olkin criterion (KMO) ------------------------------------------
      
      x The overall KMO value for your data is unacceptable.
      These data are not suitable for factor analysis.
      
      Overall: <num>
      
      For each variable:
        V1   V2   V3 
       <num> <num> <num> 

---

    Code
      print(kmo_na)
    Output
      
      -- Kaiser-Meyer-Olkin criterion (KMO) ------------------------------------------
      
      ! Sorry, the KMO value for your data is not available.

# printed output is stable

    Code
      print(efa_power(df = 100, N = 200))
    Output
      
      -- RMSEA power analysis --------------------------------------------------------
      
      Test of close fit: H0 RMSEA ≤ .050 vs. H1 RMSEA = .080.
      alpha = .050 · df = 100
      
      Power = .955 at N = 200.
      Critical value χ²(100) = 183.967 · noncentrality H0 = 49.750, H1 = 127.360.

---

    Code
      print(efa_power(df = 100, power = 0.8))
    Output
      
      -- RMSEA power analysis --------------------------------------------------------
      
      Test of close fit: H0 RMSEA ≤ .050 vs. H1 RMSEA = .080.
      alpha = .050 · df = 100
      
      Required N = 132 for a power of .800 (achieved .802).
      Critical value χ²(100) = 163.977 · noncentrality H0 = 32.750, H1 = 83.840.

---

    Code
      print(efa_power(df = 100, N = 200, type = "notclose", group = 2))
    Output
      
      -- RMSEA power analysis --------------------------------------------------------
      
      Test of not-close fit: H0 RMSEA ≥ .050 vs. H1 RMSEA = .010.
      alpha = .050 · df = 100 · groups = 2
      
      Power = .429 at N = 200 (total; 100 per group).
      Critical value χ²(100) = 97.795 · noncentrality H0 = 24.875, H1 = .995.

# a criterion that fails on every replicate is reported, not dropped

    Code
      print(sim)
    Output
      
      -- EFA power simulation --------------------------------------------------------
      
      18 variables · 3 factors · N = 6 · 3 datasets
      Estimation: PAF · rotation: varimax
      Model error: none. The population is exact, so the hit-rate and recovery are
      optimistic; set `target_rmsea` for realism.
      
      Retention hit-rate P(k-hat = 3)
      * MAP: NA (n = 0)
      
      Structure recovery (Tucker congruence ≥ .950)
      * recovery rate (min congruence): NA (n = 0)
      * recovery rate (mean congruence): NA (n = 0)
      * median min congruence: NA
      
      Convergence
      * fits completed: .000 (0/3)
      * converged (of completed): NA
      * Heywood cases (of completed): NA

# simulation-mode printed output is stable

    Code
      print(sim)
    Output
      
      -- EFA power simulation --------------------------------------------------------
      
      18 variables · 3 factors · N = 200 · 10 datasets
      Estimation: PAF · rotation: promax
      Model error: none. The population is exact, so the hit-rate and recovery are
      optimistic; set `target_rmsea` for realism.
      
      Retention hit-rate P(k-hat = 3)
      * EKC_BvA2017: <num> (n = 10)
      * MAP_TR2: <num> (n = 10)
      * MAP_TR4: <num> (n = 10)
      
      Structure recovery (Tucker congruence ≥ <num>)
      * recovery rate (min congruence): <num> (n = 10)
      * recovery rate (mean congruence): <num> (n = 10)
      * median min congruence: <num>
      
      Convergence
      * fits completed: <num> (10/10)
      * converged (of completed): <num>
      * Heywood cases (of completed): <num>

# print output is stable

    Code
      print(SL_EFAtools)
    Output
      
      EFA for second-order loadings performed with estimator = 'PAF'
      
      -- Schmid-Leiman Solution ------------------------------------------------------
      
             g     F1     F2     F3    h2    u2
      V1 <num> <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                           g     F1    F2     F3
      SS loadings <num> <num> <num> <num>
      Prop Tot Var <num> <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num> <num>
      Prop Comm Var <num> <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num> <num>

# the input error names the averaged matrix an efa_average object carries

    Code
      cat(conditionMessage(e))
    Output
      `x` must be an <EFA>, <fa>, or <lavaan> object, a matrix, or a <LOADINGS>/<loadings> object.
      i From an <efa_average> object, use the averaged loading matrix in `$loadings$average`, with the averaged factor correlations in `$Phi$average`.

# print output is stable

    Code
      print(bart_cor)
    Output
      
      v The Bartlett's test of sphericity was significant at an alpha level of <num>.
      These data are probably suitable for factor analysis.
      
      χ²(153) = <num>, p < <num>

---

    Code
      print(bart_rand)
    Output
      
      x The Bartlett's test of sphericity was not significant at an alpha level of
       <num>.
      These data are probably not suitable for factor analysis.
      
      χ²(6) = <num>, p = <num>

---

    Code
      print(bart_na)
    Output
      
      ! The Bartlett's test of sphericity did not render a result.
      The Bartlett multiplier N - 1 - (2p + 5)/6 is not positive for a sample this
      small relative to the number of variables.
      
      χ²(NA) = NA, p = NA

---

    Code
      print(bart_null)
    Output
      
      ! The Bartlett's test of sphericity did not render a result.
      The Bartlett multiplier N - 1 - (2p + 5)/6 is not positive for a sample this
      small relative to the number of variables.
      
      χ²() = NA, p = NA

# the non-convergence warning names the knobs that move the outcome

    Code
      cat(conditionMessage(w))
    Output
      The rotation did not converge within 1 iteration.
      i In `rotate_control()`, increase `maxit` for more iterations, or `precision` for a weaker tolerance.

# print.efa_simulated summarises the object

    Code
      print(sim)
    Output
      -- Simulated data --------------------------------------------------------------
      
      100 cases by 18 variables (continuous).
      Marginals: normal.
      
      Model error: CB
      RMSEA: target <num>, achieved <num>
      CFI: target --, achieved <num>

---

    Code
      print(rp)
    Output
      -- Simulated data --------------------------------------------------------------
      
      Population correlation matrix only (18 variables); no data drawn.
      Marginals: normal.
      
      Model error: none (exact population).

# print output is stable

    Code
      print(scr_cor)
    Output
      
      -- Sampling adequacy and sphericity --------------------------------------------
      
      v The overall KMO value for your data is marvellous (Overall KMO = <num>).
      These data are probably suitable for factor analysis (verbal bands: Kaiser &
      Rice, 1974).
      
      v The Bartlett's test of sphericity was significant at an alpha level of <num>.
      These data are probably suitable for factor analysis.
      χ²(153) = <num>, p < <num>
      
      -- Multicollinearity -----------------------------------------------------------
      
      i Determinant: <num>. It falls as variables are added, so the condition index
      below carries the verdict.
      v Condition number: <num> (condition index <num>). An index of 10 or less is
      rarely of interest (Belsley, 1991).
      
      -- Per-variable diagnostics ----------------------------------------------------
      
            MSA   SMC
      V1 <num> <num>
      V2 <num> <num>
      V3 <num> <num>
      V4 <num> <num>
      V5 <num> <num>
      V6 <num> <num>
      V7 <num> <num>
      V8 <num> <num>
      V9 <num> <num>
      V10 <num> <num>
      V11 <num> <num>
      V12 <num> <num>
      V13 <num> <num>
      V14 <num> <num>
      V15 <num> <num>
      V16 <num> <num>
      V17 <num> <num>
      V18 <num> <num>
      
      -- Recommendations -------------------------------------------------------------
      
      v The data appear suitable for factor analysis.
      i Per-item variance, missing-data, category, normality, and outlier diagnostics
        require raw data; only a correlation matrix was supplied.

---

    Code
      print(scr_raw)
    Output
      
      -- Sampling adequacy and sphericity --------------------------------------------
      
      v The overall KMO value for your data is marvellous (Overall KMO = <num>).
      These data are probably suitable for factor analysis (verbal bands: Kaiser &
      Rice, 1974).
      
      v The Bartlett's test of sphericity was significant at an alpha level of <num>.
      These data are probably suitable for factor analysis.
      χ²(28) = <num>, p < <num>
      
      -- Multicollinearity -----------------------------------------------------------
      
      i Determinant: <num>. It falls as variables are added, so the condition index
      below carries the verdict.
      v Condition number: <num> (condition index <num>). An index of 10 or less is
      rarely of interest (Belsley, 1991).
      
      -- Per-variable diagnostics ----------------------------------------------------
      
                variance missing%   SMC   MSA  flags
      fun <num>        0 <num> <num>       
      friends <num>        0 <num> <num>       
      enjoy <num>        0 <num> <num> sparse
      hurt <num>        0 <num> <num> sparse
      part <num>        0 <num> <num>       
      commonly <num>        0 <num> <num>       
      chances <num>        0 <num> <num>       
      attracted <num>        0 <num> <num> sparse
      
      -- Multivariate normality ------------------------------------------------------
      
      x Mardia's skewness: χ²(120) = <num>, p < <num>.
      x Mardia's kurtosis: z = <num>, p < <num>.
      x Henze-Zirkler: HZ = <num>, p < <num>.
      These data depart from multivariate normality: 3 of the 3 tests reject it.
      
      -- Outliers --------------------------------------------------------------------
      
      ! A robust (MCD) covariance could not be computed; classical Mahalanobis
      distances were used.
      At least half the complete cases lie exactly on a lower-dimensional hyperplane
      (an "exact fit"). This is common with coarse discrete items, where many
      respondents give identical answers on an item pair; it does not mean the data
      are collinear at the correlation level.
      These distances come from a covariance the outliers themselves inflate, so the
      diagnostic is no longer high-breakdown and tends to under-flag.
      i 71 of 810 observations were flagged as multivariate outliers (Mahalanobis
      distance > <num>).
      
      -- Recommendations -------------------------------------------------------------
      
      ! These data depart from multivariate normality; normal-theory standard errors
        and fit statistics may be biased - prefer robust (sandwich) or bootstrapped
        standard errors.
      ! Bartlett's test is significant, but it assumes multivariate normality and
        grows more sensitive as N increases; because these data are non-normal, treat
        it as uninformative here and rely on the KMO.
      ! 3 variables have a sparse response category (< 5 responses): enjoy, hurt, and
        attracted; a low-frequency category can destabilise polychoric estimates -
        consider collapsing it into an adjacent category.
      ! 71 observations were flagged as potential multivariate outliers; inspect them
        (see `$outliers$flagged`) before down-weighting or excluding.

---

    Code
      print(scr_non)
    Output
      
      -- Sampling adequacy and sphericity --------------------------------------------
      
      v The overall KMO value for your data is marvellous (Overall KMO = <num>).
      These data are probably suitable for factor analysis (verbal bands: Kaiser &
      Rice, 1974).
      
      ! Bartlett's test of sphericity was not computed; no sample size (N) was
      supplied.
      
      -- Multicollinearity -----------------------------------------------------------
      
      i Determinant: <num>. It falls as variables are added, so the condition index
      below carries the verdict.
      v Condition number: <num> (condition index <num>). An index of 10 or less is
      rarely of interest (Belsley, 1991).
      
      -- Per-variable diagnostics ----------------------------------------------------
      
            MSA   SMC
      V1 <num> <num>
      V2 <num> <num>
      V3 <num> <num>
      V4 <num> <num>
      V5 <num> <num>
      V6 <num> <num>
      V7 <num> <num>
      V8 <num> <num>
      V9 <num> <num>
      V10 <num> <num>
      V11 <num> <num>
      V12 <num> <num>
      V13 <num> <num>
      V14 <num> <num>
      V15 <num> <num>
      V16 <num> <num>
      V17 <num> <num>
      V18 <num> <num>
      
      -- Recommendations -------------------------------------------------------------
      
      v The data appear suitable for factor analysis.
      i Per-item variance, missing-data, category, normality, and outlier diagnostics
        require raw data; only a correlation matrix was supplied.

# the report tracks a narrow console

    Code
      print(scr_cor)
    Output
      
      -- Sampling adequacy and sphericity ------------------------
      
      v The overall KMO value for your data is marvellous
      (Overall KMO = <num>).
      These data are probably suitable for factor analysis
      (verbal bands: Kaiser & Rice, 1974).
      
      v The Bartlett's test of sphericity was significant at an
      alpha level of <num>.
      These data are probably suitable for factor analysis.
      χ²(153) = <num>, p < <num>
      
      -- Multicollinearity ---------------------------------------
      
      i Determinant: <num>. It falls as variables are added, so
      the condition index below carries the verdict.
      v Condition number: <num> (condition index <num>). An
      index of 10 or less is rarely of interest (Belsley, 1991).
      
      -- Per-variable diagnostics --------------------------------
      
            MSA   SMC
      V1 <num> <num>
      V2 <num> <num>
      V3 <num> <num>
      V4 <num> <num>
      V5 <num> <num>
      V6 <num> <num>
      V7 <num> <num>
      V8 <num> <num>
      V9 <num> <num>
      V10 <num> <num>
      V11 <num> <num>
      V12 <num> <num>
      V13 <num> <num>
      V14 <num> <num>
      V15 <num> <num>
      V16 <num> <num>
      V17 <num> <num>
      V18 <num> <num>
      
      -- Recommendations -----------------------------------------
      
      v The data appear suitable for factor analysis.
      i Per-item variance, missing-data, category, normality, and
        outlier diagnostics require raw data; only a correlation
        matrix was supplied.

# the degenerate-variance warning reads correctly for one and for several pairs

    Code
      cat(degenerate_msg(c(`v1-v2` = 0.01, `v1-v3` = -1)))
    Output
      The polychoric asymptotic covariance is unavailable for variable pair "v1-v3".
      x This pair was either estimated with a continuity correction for a response combination that never occurs -- which repairs the correlation but leaves it without a variance -- or is not identified well enough to have a usable one (e.g. a perfectly ordered or near-empty response table).
      i Any robust/sandwich standard error involving it is withheld. A DWLS fit refuses the data outright when the variance is missing or non-positive, and down-weights it out of the solution when it is merely far too large.
      i Fitting with `estimator = "ULS"` avoids the weights entirely; collapsing rare response categories in the variables involved resolves the under-identified cases.

---

    Code
      cat(degenerate_msg(c(`v1-v2` = NA_real_, `v1-v3` = -1)))
    Output
      The polychoric asymptotic covariance is unavailable for variable pairs "v1-v2" and "v1-v3".
      x These pairs were either estimated with a continuity correction for a response combination that never occurs -- which repairs the correlation but leaves it without a variance -- or are not identified well enough to have a usable one (e.g. a perfectly ordered or near-empty response table).
      i Any robust/sandwich standard error involving them is withheld. A DWLS fit refuses the data outright when the variance is missing or non-positive, and down-weights them out of the solution when it is merely far too large.
      i Fitting with `estimator = "ULS"` avoids the weights entirely; collapsing rare response categories in the variables involved resolves the under-identified cases.

# a failing component fit names its imputation and keeps the original condition

    Code
      efa_mi(raw, n_factors = 1, estimator = "PAF", rotation = "none")
    Message
      i `x` is not a correlation matrix; computing correlations from the raw data.
    Condition
      Error in `efa_mi()`:
      ! Imputation 2 could not be fitted.
      i Inspect `data_list[[2]]`, then re-fit or replace it.
      Caused by error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "V3" has zero variance.
      i A constant variable correlates with nothing; drop it before the analysis.

# bootstrap arrays are pooled into MI SEs and CIs

    Code
      print(summary(pooled_boot))
    Output
      
      Pooled EFA across 2 imputations performed with estimator = 'ML' and
        rotation = 'none'.
      Pooling settings: align_unrotated = 'signed_tucker_congruence',
        fit_pool_method = 'D2'.
      
      -- Pooled Model Diagnostics ----------------------------------------------------
      
      Factors: 1
      Variables: 8
      N: 250
      Imputations: 2
      Pooling: align_unrotated = 'signed_tucker_congruence', fit_pool_method = 'D2'
      Bootstrap samples per imputation: 6
      Valid target-rotated samples: 6 out of 6 per imputation
      Heywood cases: 0
      Cross-loading items (|loading| >= <num>): 0
      Items without salient loading (|loading| >= <num>): 0
      Factors with fewer than 3 salient indicators: 0
      Items with primary-loading gap < <num>: 0
      Largest |residual|: <num>
      
      -- Unrotated Loadings ----------------------------------------------------------
      
                  F1    h2    u2
      fun <num> <num> <num>
      friends <num> <num> <num>
      enjoy <num> <num> <num>
      hurt <num> <num> <num>
      part <num> <num> <num>
      commonly <num> <num> <num>
      chances <num> <num> <num>
      attracted <num> <num> <num>
      
      -- 95% bootstrap/MI CIs for salient unrotated loadings -------------------------
      
      Variable   Factor  est    lower  upper
      fun        F1 <num> <num> <num>
      friends    F1 <num> <num> <num>
      enjoy      F1 <num> <num> <num>
      hurt       F1 <num> <num> <num>
      part       F1 <num> <num> <num>
      commonly   F1 <num> <num> <num>
      chances    F1 <num> <num> <num>
      attracted  F1 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                      F1
      SS loadings <num>
      Prop Tot Var <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      D2-pooled χ²(20) = <num>, p = <num>
      CFI (avg. over imputations) [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      TLI (avg. over imputations) [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      RMSEA [90% CI] [95% bootstrap/MI-CI]: <num> [ <num>; <num>] [ <num>, <num>]
      AIC [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      BIC [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      ECVI [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      CAF [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      SRMR [95% bootstrap/MI-CI]: <num> [ <num>, <num>]
      Note: the pooled χ² is the D2 statistic; its p uses the D2 reference F(20, <num>),
      not the χ²(20) tail.
      Note: CFI and TLI are averaged over the imputations, not formed from the
      separately pooled model and baseline statistics in `mi_diagnostics`.
      
      Note: Bootstrap/MI CIs based on 6 bootstrap samples per imputation.
      
      -- MI Uncertainty Summary ------------------------------------------------------
      
      Largest available FMI: <num>
      Median available FMI: <num>
      Largest available RIV: <num>
      Median available RIV: <num>
      
      -- Residual Diagnostics --------------------------------------------------------
      
      Residual cutoff: |r| > <num>
      Number of large residuals: 0
      Largest absolute residual: <num>
      
      No absolute residuals > <num> occurred.
      
      Inspect the residual matrix for details (e.g., with residuals()).

# print.efa_mi output is stable (PAF, promax)

    Code
      print(pooled_obl)
    Output
      
      Pooled EFA across 3 imputations performed with estimator = 'PAF' and
        rotation = 'promax'.
      Pooling settings: target_method = 'first_target',
        align_unrotated = 'signed_tucker_congruence', fit_pool_method = 'D2'.
      
      -- Rotated Loadings ------------------------------------------------------------
      
             F1     F2     F3    h2    u2
      V1 <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num>
      
      -- Factor Intercorrelations ----------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                           F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      CAF: <num>
      SRMR: <num>
      df: 102

# print.efa_mi output is stable (ML, unrotated)

    Code
      print(pooled_none)
    Output
      
      Pooled EFA across 3 imputations performed with estimator = 'ML' and
        rotation = 'none'.
      Pooling settings: align_unrotated = 'signed_tucker_congruence',
        fit_pool_method = 'D2'.
      
      -- Unrotated Loadings ----------------------------------------------------------
      
                  F1    h2    u2
      fun <num> <num> <num>
      friends <num> <num> <num>
      enjoy <num> <num> <num>
      hurt <num> <num> <num>
      part <num> <num> <num>
      commonly <num> <num> <num>
      chances <num> <num> <num>
      attracted <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                      F1
      SS loadings <num>
      Prop Tot Var <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      D2-pooled χ²(20) = <num>, p = <num>
      CFI (avg. over imputations): <num>
      TLI (avg. over imputations): <num>
      RMSEA [90% CI]: <num> [ <num>; <num>]
      AIC: <num>
      BIC: <num>
      ECVI: <num>
      CAF: <num>
      SRMR: <num>
      Note: the pooled χ² is the D2 statistic; its p uses the D2 reference F(20, <num>),
      not the χ²(20) tail.
      Note: CFI and TLI are averaged over the imputations, not formed from the
      separately pooled model and baseline statistics in `mi_diagnostics`.

# summary.efa_mi output is stable (PAF, promax)

    Code
      print(summary(pooled_obl))
    Output
      
      Pooled EFA across 3 imputations performed with estimator = 'PAF' and
        rotation = 'promax'.
      Pooling settings: target_method = 'first_target',
        align_unrotated = 'signed_tucker_congruence', fit_pool_method = 'D2'.
      
      -- Pooled Model Diagnostics ----------------------------------------------------
      
      Factors: 3
      Variables: 18
      N: 500
      Imputations: 3
      Pooling: target_method = 'first_target',
        align_unrotated = 'signed_tucker_congruence', fit_pool_method = 'D2'
      Alignment: method = 'first_target', converged
      Heywood cases: 0
      Cross-loading items (|loading| >= <num>): 0
      Items without salient loading (|loading| >= <num>): 0
      Factors with fewer than 3 salient indicators: 0
      Items with primary-loading gap < <num>: 1
      Largest |residual|: <num>
      Factor intercorrelations > <num>: none
      
      -- Rotated Loadings ------------------------------------------------------------
      
             F1     F2     F3    h2    u2
      V1 <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num>
      
      -- Factor Intercorrelations ----------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Structure Matrix ------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Simple Structure Diagnostics ------------------------------------------------
      
      Items with primary-loading gap < <num>:
      * V11: F2 = <num>, F3 = <num>
      
      
      -- Variances Accounted for -----------------------------------------------------
      
                           F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      CAF: <num>
      SRMR: <num>
      df: 102
      
      -- Residual Diagnostics --------------------------------------------------------
      
      Residual cutoff: |r| > <num>
      Number of large residuals: 0
      Largest absolute residual: <num>
      
      No absolute residuals > <num> occurred.
      
      Inspect the residual matrix for details (e.g., with residuals()).

# summary.efa_mi output is stable (ML, unrotated)

    Code
      print(summary(pooled_none))
    Output
      
      Pooled EFA across 3 imputations performed with estimator = 'ML' and
        rotation = 'none'.
      Pooling settings: align_unrotated = 'signed_tucker_congruence',
        fit_pool_method = 'D2'.
      
      -- Pooled Model Diagnostics ----------------------------------------------------
      
      Factors: 1
      Variables: 8
      N: 810
      Imputations: 3
      Pooling: align_unrotated = 'signed_tucker_congruence', fit_pool_method = 'D2'
      Heywood cases: 0
      Cross-loading items (|loading| >= <num>): 0
      Items without salient loading (|loading| >= <num>): 0
      Factors with fewer than 3 salient indicators: 0
      Items with primary-loading gap < <num>: 0
      Largest |residual|: <num>
      
      -- Unrotated Loadings ----------------------------------------------------------
      
                  F1    h2    u2
      fun <num> <num> <num>
      friends <num> <num> <num>
      enjoy <num> <num> <num>
      hurt <num> <num> <num>
      part <num> <num> <num>
      commonly <num> <num> <num>
      chances <num> <num> <num>
      attracted <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                      F1
      SS loadings <num>
      Prop Tot Var <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      D2-pooled χ²(20) = <num>, p = <num>
      CFI (avg. over imputations): <num>
      TLI (avg. over imputations): <num>
      RMSEA [90% CI]: <num> [ <num>; <num>]
      AIC: <num>
      BIC: <num>
      ECVI: <num>
      CAF: <num>
      SRMR: <num>
      Note: the pooled χ² is the D2 statistic; its p uses the D2 reference F(20, <num>),
      not the χ²(20) tail.
      Note: CFI and TLI are averaged over the imputations, not formed from the
      separately pooled model and baseline statistics in `mi_diagnostics`.
      
      -- Residual Diagnostics --------------------------------------------------------
      
      Residual cutoff: |r| > <num>
      Number of large residuals: 0
      Largest absolute residual: <num>
      
      No absolute residuals > <num> occurred.
      
      Inspect the residual matrix for details (e.g., with residuals()).

# print and summary output are stable

    Code
      print(fs_grips)
    Output
      
      -- Factor scores (Bartlett) ----------------------------------------------------
      
      Scored 810 of 810 observations on 1 factor (see `$scores`).
      
      -- Score determinacy -----------------------------------------------------------
      
           rho  rho2  guttman
      F1 <num> <num> <num>

---

    Code
      print(fs_ob_cor)
    Output
      
      -- Factor scores (regression) --------------------------------------------------
      
      Weights and diagnostics only (correlation-matrix input; no scores).
      
      -- Score determinacy -----------------------------------------------------------
      
           rho  rho2  guttman
      F1 <num> <num> <num>
      F2 <num> <num> <num>
      F3 <num> <num> <num>

---

    Code
      print(summary(fs_ob_cor))
    Output
      
      -- Factor scores (regression) --------------------------------------------------
      
      Weights and diagnostics only (correlation-matrix input; no scores).
      
      -- Score determinacy -----------------------------------------------------------
      
           rho  rho2  guttman
      F1 <num> <num> <num>
      F2 <num> <num> <num>
      F3 <num> <num> <num>
      
      -- Factor weights --------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Score validity and univocality ----------------------------------------------
      
      Diagonal: validity (score-factor correlation). Off-diagonal: univocality.
      
           F1 F2 F3
      F1 <num> <num> <num>
      F2 <num> <num> <num>
      F3 <num> <num> <num>
      
      -- Score intercorrelations -----------------------------------------------------
      
            F1 F2 F3
      F1 <num> <num> <num>
      F2 <num> <num> <num>
      F3 <num> <num> <num>

# the f error names the averaged matrix an efa_average object carries

    Code
      cat(conditionMessage(e))
    Output
      `f` must be an <efa> object (from `efa_fit()`), a matrix, or an <efa_loadings> object.
      i From an <efa_average> object, use the averaged loading matrix in `$loadings$average`, with the averaged factor correlations in `$Phi$average` for an oblique solution.

# the uncomputable-correlation abort names the offending columns

    Code
      .prepare_cor_input(dat_fct, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "item_1" is not numeric.
      i Ordinal items stored as factors or character strings are correlated with `cor_method = "poly"` ("tetra" for binary items); anything else has to be converted or dropped.

---

    Code
      .prepare_cor_input(dat_const, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "item_2" has zero variance.
      i A constant variable correlates with nothing; drop it before the analysis.

---

    Code
      .prepare_cor_input(dat_both, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "item_1" is not numeric.
      i Ordinal items stored as factors or character strings are correlated with `cor_method = "poly"` ("tetra" for binary items); anything else has to be converted or dropped.
      x Column "item_2" has zero variance.
      i A constant variable correlates with nothing; drop it before the analysis.

---

    Code
      .prepare_cor_input(cbind(c(1, 2, 3, 4), rep(2, 4), c(4, 1, 3, 2)),
      inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "V2" has zero variance.
      i A constant variable correlates with nothing; drop it before the analysis.

---

    Code
      .prepare_cor_input(dat_lgl, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "b" has zero variance.
      i A constant variable correlates with nothing; drop it before the analysis.

---

    Code
      .prepare_cor_input(dat_inf, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "item_2" contains infinite values.
      i An infinite value has no correlation with anything; check for a division by zero or an out-of-range missing-value code.

---

    Code
      .prepare_cor_input(matrix(c(1, 2, 3), nrow = 1), inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      i Check that all columns are numeric and have non-zero variance. Correlations also need at least two observations.

---

    Code
      .prepare_cor_input(dat_lw, acov = "full", inform_from_data = FALSE)
    Message
      i An asymptotic covariance requires complete cases; incomplete rows were dropped (listwise), overriding `use = "pairwise.complete.obs"`.
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Column "c" has zero variance.
      i The variance was computed on the listwise-complete rows an `acov` needs, which can be far fewer than the data supplied; either supply more complete cases or drop the constant column.

---

    Code
      .prepare_cor_input(dat_many, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Columns "a" and "b" are not numeric.
      i Ordinal items stored as factors or character strings are correlated with `cor_method = "poly"` ("tetra" for binary items); anything else has to be converted or dropped.
      x Columns "c1" and "c2" have zero variance.
      i A constant variable correlates with nothing; drop them before the analysis.

---

    Code
      .prepare_cor_input(dat_wide, inform_from_data = FALSE)
    Condition
      Error:
      ! The correlation matrix could not be computed from the raw data.
      x Columns "V1", "V2", "V3", "V4", "V5", and 2 more have zero variance.
      i A constant variable correlates with nothing; drop them before the analysis.

# the non-symmetric abort names the triangle that is actually empty

    Code
      .assert_cor_input(R_lower)
    Condition
      Error:
      ! `x` looks like a correlation matrix but is not symmetric.
      x Its diagonal is all ones and every entry lies in [-1, 1], but `x[i, j]` and `x[j, i]` differ.
      i Only the lower triangle carries entries; mirror it: `x[upper.tri(x)] <- t(x)[upper.tri(x)]`.

---

    Code
      .assert_cor_input(R_upper)
    Condition
      Error:
      ! `x` looks like a correlation matrix but is not symmetric.
      x Its diagonal is all ones and every entry lies in [-1, 1], but `x[i, j]` and `x[j, i]` differ.
      i Only the upper triangle carries entries; mirror it: `x[lower.tri(x)] <- t(x)[lower.tri(x)]`.

---

    Code
      .assert_cor_input(R_na)
    Condition
      Error:
      ! `x` looks like a correlation matrix but is not symmetric.
      x Every entry that is present lies in [-1, 1] and no diagonal entry departs from one, but part of the matrix is missing and `x[i, j]` and `x[j, i]` differ.
      i Only the lower triangle carries entries; mirror it: `x[upper.tri(x)] <- t(x)[upper.tri(x)]`.

---

    Code
      .assert_cor_input(R_mismatch)
    Condition
      Error:
      ! `x` looks like a correlation matrix but is not symmetric.
      x Its diagonal is all ones and every entry lies in [-1, 1], but `x[i, j]` and `x[j, i]` differ.
      i Both triangles carry entries but they disagree, so neither can be mirrored onto the other; check the entries of the pairs that differ.

---

    Code
      .assert_cor_input(R_none)
    Condition
      Error:
      ! `x` looks like a correlation matrix but is not symmetric.
      x Every entry that is present lies in [-1, 1] and no diagonal entry departs from one, but part of the matrix is missing and `x[i, j]` and `x[j, i]` differ.
      i Neither triangle carries a correlation, so there is nothing to mirror; enter the off-diagonal correlations.

---

    Code
      .assert_cor_input(R_lower, raw_only = TRUE)
    Condition
      Error:
      ! `x` looks like a (non-symmetric) correlation matrix, not a data frame/matrix of raw data.
      x Its diagonal is all ones and every entry lies in [-1, 1], but `x[i, j]` and `x[j, i]` differ.
      i Supply the raw observations the correlation matrix was computed from.

# a singular matrix aborts unless the check is disabled

    Code
      .prepare_cor_input(sing_cor, N = sing_N)
    Condition
      Error:
      ! The correlation matrix is singular; no further analyses are performed.
      i Use `efa_screen()`, which reports the variables responsible.

---

    Code
      suppressMessages(.prepare_cor_input(GRiPS_raw[1:4, ]))
    Condition
      Error:
      ! The correlation matrix is singular; no further analyses are performed.
      i With N = 4 and 8 variables the correlation matrix cannot have full rank. N must be larger than the number of variables.

---

    Code
      suppressMessages(.prepare_cor_input(GRiPS_raw[1:8, ]))
    Condition
      Error:
      ! The correlation matrix is singular; no further analyses are performed.
      i With N = 8 and 8 variables the correlation matrix cannot have full rank. N must be larger than the number of variables.

# print output is stable

    Code
      print(efa_def, plot = FALSE)
    Output
      
      Averaging performed with averaging method mean (trim = 0) across 72 EFAs,
      varying the following settings: init_comm, criterion_type, k_promax, p_type,
      and varimax_type.
      
      The error rate is at <pct>. Of the solutions that did not result in an error, <pct>
      converged. Of the solutions that converged, <pct> contained Heywood cases and <pct>
      were admissible.
      
      == Indicator-to-Factor Correspondences =========================================
      
      For each cell, the proportion of solutions including the respective
      indicator-to-factor correspondence. A salience threshold of <num> was used to
      determine indicator-to-factor correspondences.
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      == Loadings ====================================================================
      
      -- Mean ------------------------------------------------------------------------
      
             F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Range (max − min) -----------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      == Factor Intercorrelations from Oblique Solutions =============================
      
      -- Mean ------------------------------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Range (max − min) -----------------------------------------------------------
      
           F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      == Variances Accounted for =====================================================
      
      -- Mean ------------------------------------------------------------------------
      
                       F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      -- Range (max − min) -----------------------------------------------------------
      
                      F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      == Model Fit ===================================================================
      
      The fit indices are the mean of the per-solution fit indices, not the fit of
      the averaged loadings printed above, which are a cell-wise summary rather than
      a fitted solution.
      
      CAF, RMSR, and SRMR averaged over <n> of 72 solutions.
      
             M (SD) [Min; Max]
      CAF: <num> ( <num>) [ <num>; <num>]
      RMSR: <num> ( <num>) [ <num>; <num>]
      SRMR: <num> ( <num>) [ <num>; <num>]
      df: 102

---

    Code
      print(efa_all_none, stat = c("average", "sd", "min", "max"), plot = FALSE)
    Output
      
      Averaging performed with averaging method mean (trim = 0) across 12 EFAs,
      varying the following settings: estimator, init_comm, criterion_type,
      abs_eigen, max_iter, and start_method.
      
      The error rate is at <pct>. Of the solutions that did not result in an error, <pct>
      converged. Of the solutions that converged, <pct> contained Heywood cases and <pct>
      were admissible.
      
      == Indicator-to-Factor Correspondences =========================================
      
      For each cell, the proportion of solutions including the respective
      indicator-to-factor correspondence. A salience threshold of <num> was used to
      determine indicator-to-factor correspondences.
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      == Loadings ====================================================================
      
      -- Mean ------------------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Standard Deviation ----------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Minimum ---------------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Maximum ---------------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      == Variances Accounted for =====================================================
      
      -- Mean ------------------------------------------------------------------------
      
                       F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      -- Standard Deviation ----------------------------------------------------------
      
                      F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      -- Minimum ---------------------------------------------------------------------
      
                       F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      -- Maximum ---------------------------------------------------------------------
      
                       F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      == Model Fit ===================================================================
      
      The fit indices are the mean of the per-solution fit indices, not the fit of
      the averaged loadings printed above, which are a cell-wise summary rather than
      a fitted solution.
      
      Chi-square-based indices averaged over <n> of 12 solutions.
      CAF, RMSR, and SRMR averaged over <n> of 12 solutions.
      
             M (SD) [Min; Max]
      χ²: <num> ( <num>) [ <num>; <num>]
      df: 102
      p: <num> ( <num>) [ <num>; <num>]
      CFI: <num> ( <num>) [ <num>; <num>]
      TLI: <num> ( <num>) [ <num>; <num>]
      RMSEA: <num> ( <num>) [ <num>; <num>]
      AIC: <num> ( <num>) [ <num>; <num>]
      BIC: <num> ( <num>) [ <num>; <num>]
      ECVI: <num> ( <num>) [ <num>; <num>]
      CAF: <num> ( <num>) [ <num>; <num>]
      RMSR: <num> ( <num>) [ <num>; <num>]
      SRMR: <num> ( <num>) [ <num>; <num>]

---

    Code
      print(efa_all_md, plot = FALSE)
    Output
      
      Averaging performed with averaging method median across 183 EFAs, varying the
      following settings: estimator, init_comm, criterion_type, abs_eigen, max_iter,
      start_method, rotation, k_promax, p_type, and varimax_type.
      
      The error rate is at <pct>. Of the solutions that did not result in an error, <pct>
      converged. Of the solutions that converged, <pct> contained Heywood cases and <pct>
      were admissible.
      
      == Indicator-to-Factor Correspondences =========================================
      
      For each cell, the proportion of solutions including the respective
      indicator-to-factor correspondence. A salience threshold of <num> was used to
      determine indicator-to-factor correspondences.
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      == Loadings ====================================================================
      
      -- Median ----------------------------------------------------------------------
      
             F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Range (max − min) -----------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      == Factor Intercorrelations from Oblique Solutions =============================
      
      -- Median ----------------------------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Range (max − min) -----------------------------------------------------------
      
           F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      == Variances Accounted for =====================================================
      
      -- Median ----------------------------------------------------------------------
      
                       F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      -- Range (max − min) -----------------------------------------------------------
      
                       F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      
      == Model Fit ===================================================================
      
      The fit indices are the median of the per-solution fit indices, not the fit of
      the averaged loadings printed above, which are a cell-wise summary rather than
      a fitted solution.
      
      Chi-square-based indices averaged over <n> of 183 solutions.
      CAF, RMSR, and SRMR averaged over <n> of 183 solutions.
      
             Md (SD) [Min; Max]
      χ²: <num> ( <num>) [ <num>; <num>]
      df: 102
      p: <num> ( <num>) [ <num>; <num>]
      CFI: <num> ( <num>) [ <num>; <num>]
      TLI: <num> ( <num>) [ <num>; <num>]
      RMSEA: <num> ( <num>) [ <num>; <num>]
      AIC: <num> ( <num>) [ <num>; <num>]
      BIC: <num> ( <num>) [ <num>; <num>]
      ECVI: <num> ( <num>) [ <num>; <num>]
      CAF: <num> ( <num>) [ <num>; <num>]
      RMSR: <num> ( <num>) [ <num>; <num>]
      SRMR: <num> ( <num>) [ <num>; <num>]

# the input error names where a fitted solution keeps its loadings

    Code
      cat(conditionMessage(e))
    Output
      `x` (<efa/EFA>) and `y` (<efa/EFA>) must be numeric vectors or matrices.
      i From an <efa> solution, use `$rot_loadings`, or `$unrot_loadings` if it was fitted without a rotation.

# print output is stable

    Code
      print(matr)
    Output
      
      -- Summary statistics ----------------------------------------------------------
      
      Mean [min, max] absolute difference:  .2500 [ .0000, 1.0000]
      Median absolute difference:  .0000
      Root mean squared distance (RMSE):  .5000
      Max decimals where all numbers agree in absolute value: none
      Minimum number of decimals provided: 0
      Differing indicator-to-factor correspondences: 1 (highest loading),
        0 (all |loadings| >= 0.3)
      
      -- Elementwise differences -----------------------------------------------------
      
      Differences: x - y.
      
            F1     F2
      V1  .0000   .0000
      V2  .0000  1.0000

---

    Code
      print(int)
    Output
      
      -- Summary statistics ----------------------------------------------------------
      
      Mean [min, max] absolute difference:  .0000 [ .0000,  .0000]
      Median absolute difference:  .0000
      Root mean squared distance (RMSE):  .0000
      Max decimals where all numbers agree in absolute value: 0
      Minimum number of decimals provided: 0
      
      -- Elementwise differences -----------------------------------------------------
      
      Differences: x - y.
      
      V1   .0000
      V2   .0000
      V3   .0000
      V4   .0000
      V5   .0000
      V6   .0000
      V7   .0000
      V8   .0000
      V9   .0000
      V10  .0000

# print.efa_retention output is stable for SMT

    Code
      print(smt)
    Output
      -- Sequential model tests ------------------------------------------------------
      
      * Sequential chi-square model tests: 3
      * Lower bound of RMSEA 90% CI: 2
      * Akaike Information Criterion: 3

---

    Code
      print(smt_id)
    Output
      -- Sequential model tests ------------------------------------------------------
      
      * Sequential chi-square model tests: 0
      * Lower bound of RMSEA 90% CI: 0
      * Akaike Information Criterion: 0

# print.efa_retention output is stable for SCREE

    Code
      print(scree)
    Output
      -- Scree plot ------------------------------------------------------------------
      Eigenvalues found using PCA, SMC, and EFA.
      
      i Scree plot is a visual criterion; call plot(x) to identify the elbow.

---

    Code
      print(scree_smc)
    Output
      -- Scree plot ------------------------------------------------------------------
      Eigenvalues found using SMC.
      
      i Scree plot is a visual criterion; call plot(x) to identify the elbow.

# print.efa_retention output is stable for CD

    Code
      print(cd)
    Output
      -- Comparison data -------------------------------------------------------------
      
      * Suggested number of factors: 1

# print.efa_retention output is stable for PARALLEL

    Code
      print(pa)
    Output
      -- Parallel analysis -----------------------------------------------------------
      Eigenvalues found using PCA; 1000 simulated datasets.
      
      * PCA: 3
      
      i Number of factors retained using the "means" decision rule.

---

    Code
      print(pa_nodat)
    Output
      -- Parallel analysis -----------------------------------------------------------
      Eigenvalues found using PCA, SMC, and EFA; 1000 simulated datasets.
      
      i No data were entered; showing the simulated eigenvalues only. No number of
      factors is suggested.

# print.efa_retention output is stable for KGC

    Code
      print(kgc)
    Output
      -- Kaiser-Guttman criterion ----------------------------------------------------
      Eigenvalues found using PCA, SMC, and EFA.
      
      * PCA: 3
      * SMC: 1
      * EFA: 1

---

    Code
      print(kgc_smc)
    Output
      -- Kaiser-Guttman criterion ----------------------------------------------------
      Eigenvalues found using SMC.
      
      * SMC: 1

# print.efa_retention output is stable for NEST

    Code
      print(nest)
    Output
      -- Next Eigenvalue Sufficiency Test --------------------------------------------
      
      * Suggested number of factors: 3

# print.efa_retention output is stable for EKC

    Code
      print(ekc)
    Output
      -- Empirical Kaiser Criterion --------------------------------------------------
      
      * Braeken & van Assen (2017): 3

# print.efa_retention output is stable for HULL

    Code
      print(hull)
    Output
      -- Hull method -----------------------------------------------------------------
      Estimator: ML
      
      * CAF: 3
      * CFI: 1
      * RMSEA: 1

---

    Code
      print(hull_paf)
    Output
      -- Hull method -----------------------------------------------------------------
      Estimator: PAF
      
      * CAF: 3

# print.efa_retention output is stable for MAP

    Code
      print(map)
    Output
      -- Minimum average partial -----------------------------------------------------
      
      * Original implementation (TR2): 1
      * Revised implementation (TR4): 3

# print.efa_retain output is stable

    Code
      print(nf)
    Output
      -- Suggested number of factors -------------------------------------------------
      
      8 suggestions from 4 criteria, ranging from 1 to 3 factors (most common: 3).
      
      Empirical Kaiser Criterion
      * Braeken & van Assen (2017): 3
      
      Kaiser-Guttman criterion
      Eigenvalues found using PCA and SMC.
      * PCA: 3
      * SMC: 1
      
      Minimum average partial
      * Original implementation (TR2): 1
      * Revised implementation (TR4): 3
      
      Sequential model tests
      * Sequential chi-square model tests: 3
      * Lower bound of RMSEA 90% CI: 2
      * Akaike Information Criterion: 3

# print.efa_retain summarises the criteria's suggestions

    Code
      print(nf)
    Output
      -- Suggested number of factors -------------------------------------------------
      
      4 suggestions from 3 criteria, ranging from 1 to 3 factors (most common: 3).
      
      Empirical Kaiser Criterion
      * Braeken & van Assen (2017): 3
      
      Kaiser-Guttman criterion
      Eigenvalues found using PCA.
      * PCA: 3
      
      Minimum average partial
      * Original implementation (TR2): 1
      * Revised implementation (TR4): 3

# the SPSS varimax non-convergence warning names the tolerance alone

    Code
      cat(conditionMessage(w))
    Output
      The rotation did not converge within 1000 iterations.
      i In `rotate_control()`, increase `precision` for a weaker tolerance.

# print output is stable

    Code
      print(efa_reliability(sl_mod, factor_map = fc))
    Output
      
      Total variance from the correlation matrix.
      
      -- Reliability coefficients ----------------------------------------------------
      
           tot  hier   sub  alpha    H
      g <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num> <num>
      F2 <num> <num> <num> <num> <num>
      F3 <num> <num> <num> <num> <num>
      
      -- Common-variance indices -----------------------------------------------------
      
          ECV   PUC
      g <num> <num>

---

    Code
      print(efa_reliability(sl_mod, factor_map = fc, coefficients = c("omega_total",
        "alpha")))
    Output
      
      Total variance from the correlation matrix.
      
      -- Reliability coefficients ----------------------------------------------------
      
           tot  alpha
      g <num> <num>
      F1 <num> <num>
      F2 <num> <num>
      F3 <num> <num>

---

    Code
      print(efa_reliability(efa_mod))
    Output
      
      Total variance from the correlation matrix.
      
      Correlated-factors solution: a factor's total omega counts the true score
      variance its composite receives from every factor, through its cross-loadings
      and any factor correlations; its subscale omega counts only that factor's own
      contribution.
      
      -- Reliability coefficients ----------------------------------------------------
      
              tot   sub  alpha    H
      total <num> <num>
      F1 <num> <num> <num> <num>
      F2 <num> <num> <num> <num>
      F3 <num> <num> <num> <num>

# the model error names the averaged matrix an efa_average object carries

    Code
      cat(conditionMessage(e))
    Output
      Invalid `model`.
      i Enter a <lavaan>, `psych::schmid()`, <SL>, or <EFA> object, a bifactor loading matrix, or specify `var_names`, `g_load`, `s_load`, and `u2`.
      i From an <efa_average> object, use the averaged loading matrix in `$loadings$average`, with `$Phi$average` as `Phi` and `$orig_R` as `cormat`.

# a failing group names the group and hands the cause on unaltered

    Code
      cat(conditionMessage(e))
    Output
      The `efa_fit()` fit failed for group "a".
      Caused by error:
      ! `estimator = "DWLS"` requires a polychoric asymptotic covariance.
      x You supplied a correlation matrix, so no asymptotic covariance can be estimated.
      i Supply raw ordinal data with `cor_method = "poly"` or `"tetra"`.

# print and format render the efa_group report

    Code
      print(mg)
    Output
      -- Multigroup exploratory factor analysis --------------------------------------
      
      2 groups (age_6_8, age_14_19) · 3 factors · PAF extraction · varimax rotation
      Aligned to a symmetric consensus target.
      N: age_6_8 = 825, age_14_19 = 1685
      
      -- Factor congruence -----------------------------------------------------------
      
      Tucker's congruence between the aligned group loadings (matched factors):
      
                             F1 F2 F3
      age_6_8 vs age_14_19 <num> <num> <num>
      
      -- Loading differences ---------------------------------------------------------
      
      Absolute loading differences by group pair (salience threshold |Δ| ≥ <num>):
      
                            mean  median   min   max  rmse  flagged
      age_6_8 vs age_14_19 <num> <num> <num> <num> <num> <flagged>
      
      -- Approximate invariance ------------------------------------------------------
      
      Congruence bands (Lorenzo-Seva & ten Berge, 2006: ≥ <num> equal, ≥ <num> fair):
      
                              F1 F2 F3
      age_6_8 vs age_14_19  equal  equal  equal

---

    Code
      print(mg_ident)
    Output
      -- Multigroup exploratory factor analysis --------------------------------------
      
      3 groups (a, b, c) · 2 factors · PAF extraction · varimax rotation
      Aligned to a symmetric consensus target.
      N: a = 500, b = 500, c = 500
      
      -- Factor congruence -----------------------------------------------------------
      
      Tucker's congruence between the aligned group loadings (matched factors):
      
                F1 F2
      a vs b <num> <num>
      a vs c <num> <num>
      b vs c <num> <num>
      
      -- Loading differences ---------------------------------------------------------
      
      Absolute loading differences by group pair (salience threshold |Δ| ≥ <num>):
      
              mean  median   min   max  rmse  flagged
      a vs b <num> <num> <num> <num> <num> <flagged>
      a vs c <num> <num> <num> <num> <num> <flagged>
      b vs c <num> <num> <num> <num> <num> <flagged>
      
      -- Approximate invariance ------------------------------------------------------
      
      Congruence bands (Lorenzo-Seva & ten Berge, 2006: ≥ <num> equal, ≥ <num> fair):
      
                F1 F2
      a vs b  equal  equal
      a vs c  equal  equal
      b vs c  equal  equal

---

    Code
      print(mg_ob)
    Output
      -- Multigroup exploratory factor analysis --------------------------------------
      
      2 groups (age_6_8, age_14_19) · 3 factors · PAF extraction · promax rotation
      Aligned to reference group age_6_8.
      N: age_6_8 = 825, age_14_19 = 1685
      
      -- Factor congruence -----------------------------------------------------------
      
      Tucker's congruence between the aligned group loadings (matched factors):
      
                             F1 F2 F3
      age_6_8 vs age_14_19 <num> <num> <num>
      
      -- Loading differences ---------------------------------------------------------
      
      Absolute loading differences by group pair (salience threshold |Δ| ≥ <num>):
      
                            mean  median   min   max  rmse  flagged
      age_6_8 vs age_14_19 <num> <num> <num> <num> <num> <flagged>

# print reports the bootstrap congruence intervals and verdicts

    Code
      print(mg)
    Output
      -- Multigroup exploratory factor analysis --------------------------------------
      
      2 groups (g1, g2) · 1 factor · PAF extraction · unrotated
      Aligned to a symmetric consensus target.
      N: g1 = 405, g2 = 405
      
      -- Factor congruence -----------------------------------------------------------
      
      Tucker's congruence between the aligned group loadings (matched factors):
      
                 F1
      g1 vs g2 <num>
      
      i 95% percentile bootstrap CIs (50 replicates) are in `$congruence$matched_ci`.
      
      -- Loading differences ---------------------------------------------------------
      
      Absolute loading differences by group pair (salience threshold |Δ| ≥ <num>):
      
                mean  median   min   max  rmse  flagged
      g1 vs g2 <num> <num> <num> <num> <num> <flagged>
      
      i 1 loading has a bootstrap CI excluding 0 (see `$flags`).
      
      -- Approximate invariance ------------------------------------------------------
      
      Congruence bands (Lorenzo-Seva & ten Berge, 2006), read off the CI lower bound:
      
                  F1
      g1 vs g2  equal

# print points out an 'equal' verdict that rests on scale-invariance

    Code
      print(mg)
    Output
      -- Multigroup exploratory factor analysis --------------------------------------
      
      2 groups (g1, g2) · 2 factors · PAF extraction · varimax rotation
      Aligned to a symmetric consensus target.
      N: g1 = 500, g2 = 500
      
      -- Factor congruence -----------------------------------------------------------
      
      Tucker's congruence between the aligned group loadings (matched factors):
      
                  F1 F2
      g1 vs g2 <num> <num>
      
      -- Loading differences ---------------------------------------------------------
      
      Absolute loading differences by group pair (salience threshold |Δ| ≥ <num>):
      
                mean  median   min   max  rmse  flagged
      g1 vs g2 <num> <num> <num> <num> <num> <flagged>
      
      -- Approximate invariance ------------------------------------------------------
      
      Congruence bands (Lorenzo-Seva & ten Berge, 2006: ≥ <num> equal, ≥ <num> fair):
      
                  F1 F2
      g1 vs g2  equal  equal
      
      i Every factor is graded "equal" while the aligned loadings differ by up to
       <num> on average: Tucker's congruence is invariant to a proportional rescaling
      of a factor's loadings, so read the verdicts alongside `$diffs`.

# the report wraps its header and splits its tables to the console width

    Code
      print(mg)
    Output
      -- Multigroup exploratory factor analysis ------------------
      
      4 groups (age_6_8, age_9_13, age_14_19, age_20_39) · 3
        factors · PAF extraction · varimax rotation
      Aligned to a symmetric consensus target.
      N: age_6_8 = 825, age_9_13 = 1572, age_14_19 = 1685,
        age_20_39 = 1251
      
      -- Factor congruence ---------------------------------------
      
      Tucker's congruence between the aligned group loadings
      (matched factors):
      
                               F1 F2 F3
      age_6_8 vs age_9_13 <num> <num> <num>
      age_6_8 vs age_14_19 <num> <num> <num>
      age_6_8 vs age_20_39 <num> <num> <num>
      age_9_13 vs age_14_19 <num> <num> <num>
      age_9_13 vs age_20_39 <num> <num> <num>
      age_14_19 vs age_20_39 <num> <num> <num>
      
      -- Loading differences -------------------------------------
      
      Absolute loading differences by group pair (salience
      threshold |Δ| ≥ <num>):
      
                              mean  median   min   max  rmse
      age_6_8 vs age_9_13 <num> <num> <num> <num> <num>
      age_6_8 vs age_14_19 <num> <num> <num> <num> <num>
      age_6_8 vs age_20_39 <num> <num> <num> <num> <num>
      age_9_13 vs age_14_19 <num> <num> <num> <num> <num>
      age_9_13 vs age_20_39 <num> <num> <num> <num> <num>
      age_14_19 vs age_20_39 <num> <num> <num> <num> <num>
      
                              flagged
      age_6_8 vs age_9_13 <flagged>
      age_6_8 vs age_14_19 <flagged>
      age_6_8 vs age_20_39 <flagged>
      age_9_13 vs age_14_19 <flagged>
      age_9_13 vs age_20_39 <flagged>
      age_14_19 vs age_20_39 <flagged>
      
      -- Approximate invariance ----------------------------------
      
      Congruence bands (Lorenzo-Seva & ten Berge, 2006: ≥ <num>
      equal, ≥ <num> fair):
      
                                F1 F2 F3
      age_6_8 vs age_9_13     equal  equal  equal
      age_6_8 vs age_14_19    equal  equal  equal
      age_6_8 vs age_20_39    equal  equal  equal
      age_9_13 vs age_14_19   equal  equal  equal
      age_9_13 vs age_20_39   equal  equal  equal
      age_14_19 vs age_20_39  equal  equal  equal

# a whole retention object as n_factors names the component that holds the counts

    Code
      cat(bad_n_factors(ret))
    Output
      `n_factors` must be a single count, not an <efa_retain> object.
      i Its `$n_factors` holds 3 suggested counts; supply one of them.

---

    Code
      cat(bad_n_factors(suppressMessages(efa_ekc(cm, N = 500))))
    Output
      `n_factors` must be a single count, not an <efa_retention> object.
      i Its `$n_factors` holds 1 suggested count; supply it.

# print.efa output is stable (PAF, promax)

    Code
      print(efa_psych)
    Output
      
      EFA performed with estimator = 'PAF' and rotation = 'promax'.
      
      -- Rotated Loadings ------------------------------------------------------------
      
             F1     F2     F3    h2    u2
      V1 <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num>
      
      -- Factor Intercorrelations ----------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                           F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      CAF: <num>
      SRMR: <num>
      df: 102

# print.efa output is stable (ML, promax)

    Code
      print(efa_ml_moderate)
    Output
      
      EFA performed with estimator = 'ML' and rotation = 'promax'.
      
      -- Rotated Loadings ------------------------------------------------------------
      
            F1    F2    F3    h2    u2
      V1 <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num>
      
      -- Factor Intercorrelations ----------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                           F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      χ²(102) = <num>, p = <num>
      CFI: <num>
      TLI: <num>
      RMSEA [90% CI]: <num> [ <num>; <num>]
      AIC: <num>
      BIC: <num>
      ECVI: <num>
      CAF: <num>
      SRMR: <num>

# summary.efa output is stable (PAF, promax)

    Code
      print(summary(efa_psych))
    Output
      
      EFA performed with estimator = 'PAF' and rotation = 'promax'.
      
      -- Model Diagnostics -----------------------------------------------------------
      
      Factors: 3
      Variables: 18
      N: 500
      Heywood cases: 0
      Cross-loading items (|loading| >= <num>): 0
      Items without salient loading (|loading| >= <num>): 0
      Factors with fewer than 3 salient indicators: 0
      Items with primary-loading gap < <num>: 1
      Largest |residual|: <num>
      Factor intercorrelations > <num>: none
      
      -- Rotated Loadings ------------------------------------------------------------
      
             F1     F2     F3    h2    u2
      V1 <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num>
      
      -- Factor Intercorrelations ----------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Structure Matrix ------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Simple Structure Diagnostics ------------------------------------------------
      
      Items with primary-loading gap < <num>:
      * V11: F2 = <num>, F3 = <num>
      
      
      -- Variances Accounted for -----------------------------------------------------
      
                           F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      CAF: <num>
      SRMR: <num>
      df: 102
      
      -- Residual Diagnostics --------------------------------------------------------
      
      Residual cutoff: |r| > <num>
      Number of large residuals: 0
      Largest absolute residual: <num>
      
      No absolute residuals > <num> occurred.
      
      Inspect the residual matrix for details (e.g., with residuals()).

# summary.efa output is stable (ML, promax)

    Code
      print(summary(efa_ml_moderate))
    Output
      
      EFA performed with estimator = 'ML' and rotation = 'promax'.
      
      -- Model Diagnostics -----------------------------------------------------------
      
      Factors: 3
      Variables: 18
      N: 500
      Heywood cases: 0
      Cross-loading items (|loading| >= <num>): 0
      Items without salient loading (|loading| >= <num>): 0
      Factors with fewer than 3 salient indicators: 0
      Items with primary-loading gap < <num>: 0
      Largest |residual|: <num>
      Factor intercorrelations > <num>: none
      
      -- Rotated Loadings ------------------------------------------------------------
      
            F1    F2    F3    h2    u2
      V1 <num> <num> <num> <num> <num>
      V2 <num> <num> <num> <num> <num>
      V3 <num> <num> <num> <num> <num>
      V4 <num> <num> <num> <num> <num>
      V5 <num> <num> <num> <num> <num>
      V6 <num> <num> <num> <num> <num>
      V7 <num> <num> <num> <num> <num>
      V8 <num> <num> <num> <num> <num>
      V9 <num> <num> <num> <num> <num>
      V10 <num> <num> <num> <num> <num>
      V11 <num> <num> <num> <num> <num>
      V12 <num> <num> <num> <num> <num>
      V13 <num> <num> <num> <num> <num>
      V14 <num> <num> <num> <num> <num>
      V15 <num> <num> <num> <num> <num>
      V16 <num> <num> <num> <num> <num>
      V17 <num> <num> <num> <num> <num>
      V18 <num> <num> <num> <num> <num>
      
      -- Factor Intercorrelations ----------------------------------------------------
      
            F1 F2 F3
      F1 <num>
      F2 <num> <num>
      F3 <num> <num> <num>
      
      -- Structure Matrix ------------------------------------------------------------
      
            F1 F2 F3
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
                           F1 F2 F3
      SS loadings <num> <num> <num>
      Prop Tot Var <num> <num> <num>
      Cum Prop Tot Var <num> <num> <num>
      Prop Comm Var <num> <num> <num>
      Cum Prop Comm Var <num> <num> <num>
      
      -- Model Fit -------------------------------------------------------------------
      
      χ²(102) = <num>, p = <num>
      CFI: <num>
      TLI: <num>
      RMSEA [90% CI]: <num> [ <num>; <num>]
      AIC: <num>
      BIC: <num>
      ECVI: <num>
      CAF: <num>
      SRMR: <num>
      
      -- Residual Diagnostics --------------------------------------------------------
      
      Residual cutoff: |r| > <num>
      Number of large residuals: 0
      Largest absolute residual: <num>
      
      No absolute residuals > <num> occurred.
      
      Inspect the residual matrix for details (e.g., with residuals()).

# print/summary.efa omit the inapplicable tables for a rotated single factor

    Code
      print(efa_1fac)
    Output
      
      EFA performed with estimator = 'PAF' and rotation = 'promax'.
      
      -- Rotated Loadings ------------------------------------------------------------
      
            F1    h2    u2
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
      No variance-accounted table available.
      
      -- Model Fit -------------------------------------------------------------------
      
      CAF: <num>
      SRMR: <num>
      df: 135

---

    Code
      print(summary(efa_1fac))
    Output
      
      EFA performed with estimator = 'PAF' and rotation = 'promax'.
      
      -- Model Diagnostics -----------------------------------------------------------
      
      Factors: 1
      Variables: 18
      N: 500
      Heywood cases: 0
      Cross-loading items (|loading| >= <num>): 0
      Items without salient loading (|loading| >= <num>): 0
      Factors with fewer than 3 salient indicators: 0
      Items with primary-loading gap < <num>: 0
      Largest |residual|: <num>
      
      -- Rotated Loadings ------------------------------------------------------------
      
            F1    h2    u2
      V1 <num> <num> <num>
      V2 <num> <num> <num>
      V3 <num> <num> <num>
      V4 <num> <num> <num>
      V5 <num> <num> <num>
      V6 <num> <num> <num>
      V7 <num> <num> <num>
      V8 <num> <num> <num>
      V9 <num> <num> <num>
      V10 <num> <num> <num>
      V11 <num> <num> <num>
      V12 <num> <num> <num>
      V13 <num> <num> <num>
      V14 <num> <num> <num>
      V15 <num> <num> <num>
      V16 <num> <num> <num>
      V17 <num> <num> <num>
      V18 <num> <num> <num>
      
      -- Variances Accounted for -----------------------------------------------------
      
      No variance-accounted table available.
      
      -- Model Fit -------------------------------------------------------------------
      
      CAF: <num>
      SRMR: <num>
      df: 135
      
      -- Residual Diagnostics --------------------------------------------------------
      
      Residual cutoff: |r| > <num>
      Number of large residuals: 9
      Largest absolute residual: <num>
      
      Largest residuals:
      * V10 ~~ V12: <num>
      * V1 ~~ V6: <num>
      * V9 ~~ V10: <num>
      * V14 ~~ V17: <num>
      * V8 ~~ V10: <num>
      * V1 ~~ V4: <num>
      * V13 ~~ V17: <num>
      * V17 ~~ V18: <num>
      * V4 ~~ V5: <num>
      
      Inspect the residual matrix for details (e.g., with residuals()).

# the Heywood warning names the affected variables

    Code
      for (est in c("ML", "ULS")) {
        cat("--", est, "--\n")
        withCallingHandlers(efa_fit(R_bnd, 1, N = 200, estimator = est), efa_heywood = function(
          w) cat(conditionMessage(w), "\n"), warning = function(w) invokeRestart(
          "muffleWarning"))
      }
    Output
      -- ML --
      Heywood case detected for "item1": the solution is improper (a communality at or above 1, or a uniqueness fixed at the estimation boundary).
      i Interpret the affected loadings and uniquenesses with caution. 
      -- ULS --
      Heywood case detected for "item1": the solution is improper (a communality at or above 1, or a uniqueness fixed at the estimation boundary).
      i Interpret the affected loadings and uniquenesses with caution. 

# the category guards read correctly for one and for several variables

    Code
      cat(constant_msg(one_const))
    Output
      Polychoric correlations need at least two response categories per variable.
      x Variable "a" is constant.

---

    Code
      cat(constant_msg(two_const))
    Output
      Polychoric correlations need at least two response categories per variable.
      x Variables "a" and "b" are constant.

---

    Code
      cat(binary_msg(one_wide))
    Output
      Tetrachoric correlations require binary variables (at most two categories).
      x Variable "a" has more than two categories.

---

    Code
      cat(binary_msg(base))
    Output
      Tetrachoric correlations require binary variables (at most two categories).
      x Variables "a" and "c" have more than two categories.

# the uncomputable-pair error reads correctly for one and for several pairs

    Code
      cat(na_msg(data.frame(a = a, b = b, c = half)))
    Output
      The polychoric correlation could not be computed for variable pair "a-b".
      i This pair has no overlapping complete observations.

---

    Code
      cat(na_msg(data.frame(a = a, b = b, d = d, c = half)))
    Output
      The polychoric correlation could not be computed for variable pairs "a-b" and "b-d".
      i These pairs have no overlapping complete observations.

# the sparse-cell warning names the offending pairs and caps the list

    Code
      cat(sparse_msg(one))
    Output
      1 variable pair has an empty response-category combination despite a non-negligible expected count.
      x Affected pair: "risk-gamble".
      i The polychoric asymptotic covariance (and any DWLS weights or robust standard errors derived from it) can be unreliable for such structurally sparse cells; interpret them with caution and consider collapsing rare response categories in these variables.

---

    Code
      cat(sparse_msg(nudged))
    Output
      15 variable pairs have an empty response-category combination despite a non-negligible expected count.
      x Affected pairs: "v1-v2", "v1-v3", "v1-v4", "v1-v5", "v1-v6", and 10 more.
      i The polychoric asymptotic covariance (and any DWLS weights or robust standard errors derived from it) can be unreliable for such structurally sparse cells; interpret them with caution and consider collapsing rare response categories in these variables.

# the boundary warning names the offending pairs and caps the list

    Code
      cat(boundary_msg(one))
    Output
      1 variable pair has a perfectly monotone response table, so its correlation is estimated at the boundary value 0.9999.
      x Affected pair: "risk-gamble".
      i The two variables are never observed out of order -- either they always vary together, or (for a negative estimate) they always vary in opposite directions -- which the model can reproduce only with a correlation of 1 or -1, so the data put no limit on how close to that boundary the correlation lies and no asymptotic variance or standard error is available.
      i Consider collapsing rare response categories in the variables involved, or dropping one item of a redundant pair.

---

    Code
      cat(boundary_msg(rev))
    Output
      1 variable pair has a perfectly monotone response table, so its correlation is estimated at the boundary value -0.9999.
      x Affected pair: "risk-gamble".
      i The two variables are never observed out of order -- either they always vary together, or (for a negative estimate) they always vary in opposite directions -- which the model can reproduce only with a correlation of 1 or -1, so the data put no limit on how close to that boundary the correlation lies and no asymptotic variance or standard error is available.
      i Consider collapsing rare response categories in the variables involved, or dropping one item of a redundant pair.

---

    Code
      cat(boundary_msg(both))
    Output
      3 variable pairs have a perfectly monotone response table, so their correlation is estimated at the boundary values 0.9999 and -0.9999.
      x Affected pairs: "risk-gamble", "risk-mirror", and "gamble-mirror".
      i The two variables are never observed out of order -- either they always vary together, or (for a negative estimate) they always vary in opposite directions -- which the model can reproduce only with a correlation of 1 or -1, so the data put no limit on how close to that boundary the correlation lies and no asymptotic variance or standard error is available.
      i Consider collapsing rare response categories in the variables involved, or dropping one item of a redundant pair.

# the boundary warning quotes the estimate, not the nearest-PD projection of it

    Code
      cat(conditionMessage(w))
    Output
      1 variable pair has a perfectly monotone response table, so its correlation is estimated at the boundary value 0.9999.
      x Affected pair: "v1-v2".
      i The two variables are never observed out of order -- either they always vary together, or (for a negative estimate) they always vary in opposite directions -- which the model can reproduce only with a correlation of 1 or -1, so the data put no limit on how close to that boundary the correlation lies and no asymptotic variance or standard error is available.
      i Consider collapsing rare response categories in the variables involved, or dropping one item of a redundant pair.

# the continuity-correction warning names the offending pairs and caps the list

    Code
      cat(zero_cell_msg(guttman))
    Output
      21 binary variable pairs have a response combination that never occurs, so a continuity correction of 0.5 was applied before estimating their correlation.
      x Affected pairs: "i1-i2", "i1-i3", "i1-i4", "i1-i5", "i1-i6", and 16 more.
      i Every two-by-two table with an empty cell is reproduced exactly by a correlation of 1 (or -1), so without the correction the estimate would be the boundary value whatever the underlying correlation. The correction adds 0.5 to the empty cell while preserving the table margins, as lavaan and psych do by default.
      i The corrected correlation is a point estimate only: no asymptotic variance or standard error is available for these pairs.

---

    Code
      cat(zero_cell_msg(one))
    Output
      1 binary variable pair has a response combination that never occurs, so a continuity correction of 0.5 was applied before estimating its correlation.
      x Affected pair: "a-b".
      i Every two-by-two table with an empty cell is reproduced exactly by a correlation of 1 (or -1), so without the correction the estimate would be the boundary value whatever the underlying correlation. The correction adds 0.5 to the empty cell while preserving the table margins, as lavaan and psych do by default.
      i The corrected correlation is a point estimate only: no asymptotic variance or standard error is available for this pair.

# the mixed-se abort names the imputations that broke the shared se

    Code
      EFAtools:::.efa_pooled_mixed_se_abort(fits_boot, cor_input)
    Condition
      Error in `EFAtools:::.efa_pooled_mixed_se_abort()`:
      ! The component `efa_fit()` fits use different `se` methods, so their standard errors cannot be pooled.
      x Imputation 2 recorded `se` "none"; the remaining 2 recorded "np-boot".
      i Imputation 2 is a correlation matrix, which cannot be bootstrapped, so `efa_fit()` fitted it with `se = "none"`.

---

    Code
      EFAtools:::.efa_pooled_mixed_se_abort(fits_mixed, raw)
    Condition
      Error in `EFAtools:::.efa_pooled_mixed_se_abort()`:
      ! The component `efa_fit()` fits use different `se` methods, so their standard errors cannot be pooled.
      x Imputation 2 recorded `se` "sandwich"; the remaining 2 recorded "information".
      i Re-fit every imputation with the same `se` (all "none", "information", "sandwich", or "np-boot").

# print output is stable

    Code
      print(res)
    Output
      
      Total variance from the correlation matrix.
      
      -- Reliability coefficients ----------------------------------------------------
      
           tot  hier   sub  alpha    H
      g <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num> <num>
      F2 <num> <num> <num> <num> <num>
      F3 <num> <num> <num> <num> <num>
      
      -- Common-variance indices -----------------------------------------------------
      
          ECV   PUC
      g <num> <num>

---

    Code
      print(res_mg)
    Output
      
      Total variance from the correlation matrix.
      
      Group GroupA:
      
      -- Reliability coefficients ----------------------------------------------------
      
           tot  hier   sub  alpha    H
      g <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num> <num>
      F2 <num> <num> <num> <num> <num>
      F3 <num> <num> <num> <num> <num>
      
      -- Common-variance indices -----------------------------------------------------
      
          ECV   PUC
      g <num> <num>
      
      Group GroupB:
      
      -- Reliability coefficients ----------------------------------------------------
      
           tot  hier   sub  alpha    H
      g <num> <num> <num> <num> <num>
      F1 <num> <num> <num> <num> <num>
      F2 <num> <num> <num> <num> <num>
      F3 <num> <num> <num> <num> <num>
      
      -- Common-variance indices -----------------------------------------------------
      
          ECV   PUC
      g <num> <num>

---

    Code
      print(res_sf)
    Output
      
      Total variance from the model-implied composite variance.
      
      -- Reliability coefficients ----------------------------------------------------
      
          tot    H
      g <num> <num>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{CD}
\alias{CD}
\title{Comparison data}
\usage{
CD(
  x,
  n_factors_max = NA,
  N_pop = 10000,
  N_samples = 500,
  alpha = 0.3,
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  max_iter = 50
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data.}

\item{n_factors_max}{numeric. The maximum number of factors to test against.
Larger numbers will increase the duration the procedure takes, but test more
possible solutions. If left NA (default) the maximum number of factors for
which the model is still over-identified (df > 0) is used.}

\item{N_pop}{numeric. Size of finite populations of comparison data. Default
is 10000.}

\item{N_samples}{numeric. Number of samples drawn from each population.
Default is 500.}

\item{alpha}{numeric. The alpha level used to test the significance of the
improvement added by an additional factor. Default is .30.}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{CD} compares the data against simulated continuous reference data.
Default is "pearson".}

\item{max_iter}{numeric. The maximum number of iterations after which the
iterative PAF procedure inside the comparison-data generation is halted; it
does not cap an EFA of \code{x}. Default is 50.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_cd]{efa_cd()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{CD()} has been superseded by \code{\link[=efa_cd]{efa_cd()}}, which is the recommended interface
going forward. It remains available and unchanged so existing code keeps working.
}
\seealso{
\code{\link[=efa_cd]{efa_cd()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_retain.R
\name{format.efa_retain}
\alias{format.efa_retain}
\title{Format method for efa_retain objects}
\usage{
\method{format}{efa_retain}(x, ...)
}
\arguments{
\item{x}{an object of class efa_retain, returned by \code{\link[=efa_retain]{efa_retain()}}.}

\item{...}{not used.}
}
\value{
A character vector with the report lines (styled to the active
console theme; plain when colours are disabled).
}
\description{
Format method for efa_retain objects
}
\examples{
\donttest{
nf <- efa_retain(test_models$baseline$cormat, criteria = c("EKC", "SMT"),
                 N = 500)
writeLines(format(nf))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.BARTLETT.R
\name{print.BARTLETT}
\alias{print.BARTLETT}
\title{Print BARTLETT object}
\usage{
\method{print}{BARTLETT}(x, ...)
}
\arguments{
\item{x}{list of class BARTLETT (output from the \link{BARTLETT} function)}

\item{...}{additional arguments passed to print}
}
\description{
Print BARTLETT object
}
\examples{
BARTLETT(test_models$baseline$cormat, N = 500)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_simulated.R
\name{print.efa_simulated}
\alias{print.efa_simulated}
\alias{format.efa_simulated}
\title{Print and format an efa_simulated object}
\usage{
\method{print}{efa_simulated}(x, digits = 3, ...)

\method{format}{efa_simulated}(x, digits = 3, ...)
}
\arguments{
\item{x}{An object of class \code{efa_simulated} (output from \code{\link[=efa_simulate]{efa_simulate()}}).}

\item{digits}{Integer. The number of decimal places the reported fit values are rounded to.
Default is 3.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a character vector
with the summary lines.
}
\description{
\code{print()} shows a compact summary of the data simulated by \code{\link[=efa_simulate]{efa_simulate()}}: how many
datasets were drawn and their dimensions, the marginal distribution, whether the data were
discretized into ordered categories or given missing values, and -- when model error was
injected -- the method with the target and achieved RMSEA and CFI. The simulated data
themselves live in the \code{data} element and the population correlation matrix in \code{population}.
\code{format()} returns the same summary as a character vector; \code{print()} is
\code{cat(format(x), sep = "\\n")}. The lines follow the active console theme, so they are plain
when colours are disabled.
}
\examples{
Lambda <- population_models$loadings$baseline
Phi <- population_models$phis_3$moderate
efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, target_rmsea = 0.05, seed = 42)

}
\seealso{
Other data simulation:
\code{\link[=efa_simulate]{efa_simulate()}}
}
\concept{data simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.rotate_bifactor_orth}
\alias{.rotate_bifactor_orth}
\title{Orthogonal bifactor factor rotation}
\usage{
.rotate_bifactor_orth(
  L,
  eps = 1e-05,
  normalize = TRUE,
  random_starts = 0L,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1,
  screen_keep = 5L,
  triage_maxit = 25L,
  triage_improve_tol = 0
)
}
\arguments{
\item{L}{Numeric matrix. The unrotated loading matrix (variables by factors).}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient norm.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization before
rotation and reverse it afterwards.}

\item{random_starts}{Integer scalar. Number of additional random orthogonal starts.}

\item{maxit}{Integer scalar. Maximum number of projected-gradient updates.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving attempts after
the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient update.}

\item{screen_keep}{Integer scalar. Number of screened random starts retained for
triage optimization.}

\item{triage_maxit}{Integer scalar. Number of short optimization iterations used in
the triage stage.}

\item{triage_improve_tol}{Numeric scalar. Relative improvement required for a triaged
start to be promoted to full optimization.}
}
\value{
A named list with the rotated loadings, the orthogonal rotation matrix \code{Th}
(with \code{L \%*\% Th} reproducing the rotated loadings), the attained criterion value, and the
convergence and validity flags. The list additionally reports the criterion
value reached at each optimized start in \code{all_values}, with a per-start convergence flag in \code{all_converged}.
}
\description{
Rotate a loading matrix orthogonally under the Jennrich-Bentler bifactor criterion using a
gradient-projection optimizer along the orthogonal (Stiefel) manifold.
}
\details{
The criterion value \code{f} and its gradient \code{dQ/dL} at the rotated loadings
\code{L = A \%*\% T} define the search; the engine maps the gradient to the orthogonal
transformation \code{T}, projects it onto the tangent space, performs a non-monotone line
search, and retracts back onto the orthogonal group via a polar (singular value) projection.
The first factor is treated as a general factor and is exempt from the penalty; the criterion
measures the between-group-factor cross-products of the squared loadings, so it is minimized
when each variable loads on the general factor plus at most one group factor. The criterion is
prone to local minima, so additional random starts are recommended.

Additional random orthogonal starts may be requested. To bound runtime the solver screens
each random start by its objective, runs a short triage optimization on the best-screened
starts, and fully optimizes only those that improve on the current incumbent by at least
\code{triage_improve_tol}.
}
\references{
Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
software for arbitrary rotation criteria in factor analysis. \emph{Educational and
Psychological Measurement}, 65, 676-696.

Jennrich, R. I., & Bentler, P. M. (2011). Exploratory bi-factor analysis. \emph{Psychometrika},
76, 537-549.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.PARALLEL.R
\name{plot.PARALLEL}
\alias{plot.PARALLEL}
\title{Plot PARALLEL object}
\usage{
\method{plot}{PARALLEL}(x, ...)
}
\arguments{
\item{x}{list of class PARALLEL. An output from the \link{PARALLEL} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \link{PARALLEL} function
}
\examples{
\donttest{
# example with correlation matrix and "ML" estimation
x <- PARALLEL(test_models$case_11b$cormat, N = 500, method = "ML")
plot(x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_schmid_leiman.R
\name{efa_schmid_leiman}
\alias{efa_schmid_leiman}
\title{Schmid-Leiman transformation}
\source{
Schmid, J. & Leiman, J. M. (1957). The development of hierarchical
factor solutions. Psychometrika, 22(1), 53–61. doi:10.1007/BF02289209

Wolff, H.-G., & Preising, K. (2005). Exploring item and higher order
factor structure with the Schmid-Leiman solution: Syntax codes for SPSS and
SAS. Behavior Research Methods, 37 , 48–58. doi:10.3758/BF03206397
}
\usage{
efa_schmid_leiman(
  x,
  Phi = NULL,
  estimator = c("PAF", "ML", "ULS", "MINRES"),
  g_name = "g",
  estimate_control = NULL,
  ...
)
}
\arguments{
\item{x}{object of class \code{\link[=efa_fit]{efa_fit()}}, class \code{\link[psych:fa]{psych::fa()}},
class \code{lavaan::lavaan()}, a matrix, or an \code{efa_loadings}/\code{loadings} object. If class \code{\link[=efa_fit]{efa_fit()}} or
class \code{\link[psych:fa]{psych::fa()}}, pattern coefficients and factor
intercorrelations are taken from this object. If class \code{lavaan::lavaan()},
it must be a second-order CFA solution. In this case first-order and second-order
factor loadings are taken from this object and the \code{g_name} argument has
to be specified.
x can also be a pattern matrix from an oblique factor solution (see \code{Phi}).}

\item{Phi}{matrix. A matrix of factor intercorrelations from an oblique factor
solution. Only needs to be specified if a pattern matrix is entered directly
into \code{x}.}

\item{estimator}{character. One of "PAF", "ML", or "ULS" to use
principal axis factoring, maximum likelihood, or unweighted least squares,
respectively, used in \code{\link[=efa_fit]{efa_fit()}} to find the second-order loadings. "MINRES" is
accepted as a synonym for "ULS" (the same estimator).}

\item{g_name}{character. The name of the general factor. This needs only be
specified if \code{x} is a \code{lavaan} second-order solution. Default is "g".}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
second-order \code{\link[=efa_fit]{efa_fit()}} fit, including the \code{type} preset. \code{NULL} (default) uses the
\code{\link[=efa_fit]{efa_fit()}} defaults. The second-order fit is unrotated, so no rotation settings apply.}

\item{...}{Arguments to be passed to \code{\link[=efa_fit]{efa_fit()}}. The estimation tuning knobs are not passed
here; they live in \code{estimate_control}, and the standard-error arguments (\code{se}, \code{b_boot},
\code{ci}, \code{seed}) are not accepted because the second-order fit is an internal step run against
a placeholder sample size and only its loadings are kept.}
}
\value{
A list of class \code{c("efa_schmid_leiman", "SL")} containing the following
\item{orig_R}{Original correlation matrix.}
\item{sl}{A matrix with general factor loadings, group factor loadings, communalities,
and uniquenesses.}
\item{L2}{Second-order factor loadings.}
\item{vars_accounted}{A matrix of explained variances and sums of squared loadings.}
\item{iter}{The number of iterations needed for convergence in EFA.}
\item{convergence}{Integer convergence code of the second-order EFA (0 =
converged); \code{NA} for a lavaan input. See \code{\link[=efa_fit]{efa_fit()}}.}
\item{settings}{list. The settings (arguments) used in EFA to get the
second-order loadings.}
}
\description{
This function implements the Schmid-Leiman (SL) transformation
(Schmid & Leiman, 1957). It takes the pattern coefficients and factor
intercorrelations from an oblique factor solution as
input and can reproduce the results from \code{\link[psych:schmid]{psych::schmid()}}
and from the SPSS implementation from Wolff & Preising (2005). Other arguments
from \code{\link[=efa_fit]{efa_fit()}} can be used to control the procedure to find the
second-order loadings more flexibly. The function can also be used on a
second-order confirmatory factor analysis (CFA) solution from lavaan.
The group factors of the returned solution are sorted and relabelled, so their
column order can differ from the input solution's (see Details).
}
\details{
The SL transformation (also called SL orthogonalization) is a procedure with
which an oblique factor solution is transformed into a hierarchical,
orthogonalized solution. As a first step, the factor intercorrelations are
factor analyzed to extract a single second-order (general) factor, yielding a
two-level hierarchical structure. The first-order factor and the second-order
factor are then orthogonalized, resulting in an orthogonalized factor solution
with proportionality constraints. The procedure thus makes a suggested
hierarchical data structure based on factor intercorrelations explicit. One
major advantage of SL transformation is that it enables variance
partitioning between higher-order and first-order factors, including the
calculation of McDonald's omegas (see \code{\link[=efa_reliability]{efa_reliability()}}).

Where the first-order factors come from a loading matrix -- an \code{\link[=efa_fit]{efa_fit()}} or a
\code{\link[psych:fa]{psych::fa()}} solution, or a pattern matrix supplied with \code{Phi} -- they are sorted
by the number in their column labels, so that \code{"F10"} follows \code{"F2"} rather than
\code{"F1"}. The sort needs a number in every column label; columns that carry no
labels, or a label without a number, keep the order they arrive in. A second-order
\code{lavaan} solution is not sorted at all: its first-order factors keep the order the
model declares them in.

The columns are then labelled \code{"F1"} to \code{"Fk"} by position, on every route, and the
input solution's own factor names are not carried over. A factor a \code{lavaan} model
calls \code{"F3"} can therefore come back as \code{"F1"}. Where the sort applies it is
independent of how the input orders its factors, so the group
factors of the returned \code{sl} matrix can also be in a different order from the columns
they came from. A \code{\link[psych:fa]{psych::fa()}} solution shows this most readily: it orders its
columns by their sums of squared loadings, but keeps each factor's own number in its
label, so those numbers arrive out of order. A
solution whose columns are \code{"PA2"}, \code{"PA3"}, \code{"PA1"} comes back with those same
three factors sorted as \code{PA1}, \code{PA2}, \code{PA3} and labelled \code{"F1"}, \code{"F2"}, \code{"F3"}.
The first group factor of the result is then the third column of the input. The
same holds against \code{\link[psych:schmid]{psych::schmid()}}, whose columns keep the input order: the two
solutions agree column for column only after one of them is permuted to the
other's order. An \code{\link[=efa_fit]{efa_fit()}} solution already labels its factors \code{"F1"} to \code{"Fk"}
in that order, so nothing moves for one; the reordering shows itself for a
\code{\link[psych:fa]{psych::fa()}} solution, and for a pattern matrix supplied with labels of its own.

Read the group factors from the returned matrix, therefore, rather than from the
input. An indicator-to-factor map is matched to the group factors by position, so
one built in the input solution's column order lines up only where the columns did
not move; where they did, \code{\link[=efa_reliability]{efa_reliability()}} or \code{\link[=OMEGA]{OMEGA()}} scores each composite
against the wrong factor. Build such a map from the \code{"F1"} to \code{"Fk"} columns of the
returned \code{sl} matrix instead, which is right on every route. The \code{fac_names} of
\code{\link[=efa_reliability]{efa_reliability()}} are matched by position in the same way, so names given in the
input solution's order label the wrong subscales, and do so without any sign.
}
\examples{
## Use with an output from the EFAtools::efa_fit function, both with type EFAtools
EFA_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
SL_EFAtools <- efa_schmid_leiman(EFA_mod, estimator = "PAF",
                                 estimate_control = estimate_control(type = "EFAtools"))

\donttest{
## Use with an output from the psych::fa function with type psych
fa_mod <- psych::fa(test_models$baseline$cormat, nfactors = 3, n.obs = 500,
                    fm = "pa", rotate = "Promax")
SL_psych <- efa_schmid_leiman(fa_mod, estimator = "PAF",
                              estimate_control = estimate_control(type = "psych"))
}

## Use more flexibly by entering a pattern matrix and phi directly (useful if
## a factor solution found with another program should be subjected to SL
## transformation)

## For demonstration, take pattern matrix and phi from an EFA output
## This gives the same solution as the first example
SL_flex <- efa_schmid_leiman(EFA_mod$rot_loadings, Phi = EFA_mod$Phi, estimator = "PAF",
                             estimate_control = estimate_control(type = "EFAtools"))

\donttest{
## Use with a lavaan second-order CFA output
if (requireNamespace("lavaan", quietly = TRUE)) {

# Create and fit model in lavaan (assume all variables have SDs of 1)
mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
        F2 =~ V7 + V8 + V9 + V10 + V11 + V12
        F3 =~ V13 + V14 + V15 + V16 + V17 + V18
        g =~ F1 + F2 + F3'
fit <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                   sample.nobs = 500, estimator = "ml")

SL_lav <- efa_schmid_leiman(fit, g_name = "g")

}
}
}
\seealso{
Other factor rotation:
\code{\link[=efa_procrustes]{efa_procrustes()}}

Other reliability coefficients:
\code{\link[=efa_reliability]{efa_reliability()}},
\code{\link[=print.efa_reliability]{print.efa_reliability()}}
}
\concept{factor rotation}
\concept{reliability coefficients}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_cd.R
\name{efa_cd}
\alias{efa_cd}
\title{Comparison data}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200

Ruscio, J., & Roche, B. (2012). Determining the number of factors to
retain in an exploratory factor analysis using comparison data of known
factorial structure. Psychological Assessment, 24, 282–292.
doi: 10.1037/a0025697
}
\usage{
efa_cd(
  x,
  n_factors_max = NA,
  N_pop = 10000,
  N_samples = 500,
  alpha = 0.3,
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  max_iter = 50
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data.}

\item{n_factors_max}{numeric. The maximum number of factors to test against.
Larger numbers will increase the duration the procedure takes, but test more
possible solutions. If left NA (default) the maximum number of factors for
which the model is still over-identified (df > 0) is used.}

\item{N_pop}{numeric. Size of finite populations of comparison data. Default
is 10000.}

\item{N_samples}{numeric. Number of samples drawn from each population.
Default is 500.}

\item{alpha}{numeric. The alpha level used to test the significance of the
improvement added by an additional factor. Default is .30.}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{CD} compares the data against simulated continuous reference data.
Default is "pearson".}

\item{max_iter}{numeric. The maximum number of iterations after which the
iterative PAF procedure inside the comparison-data generation is halted; it
does not cap an EFA of \code{x}. Default is 50.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). Its main fields are:
\item{n_factors}{A named numeric vector (\code{"CD"}) with the suggested number of
factors according to comparison data results.}
\item{results}{A list with a single record holding the mean RMSE between the
eigenvalues of the generated and the entered data per number of factors
(used for the plot) and, in \code{rmse_eigenvalues}, the per-sample RMSE matrix
(rows are samples, columns are factor counts; columns beyond the last tested
factor count are left as zero).}
\item{settings}{A list of the settings used.}
}
\description{
Factor retention method introduced by Ruscio and Roche (2012). The code was
adapted from the CD code published by Auerswald and Moshagen (2019), available
at \url{https://osf.io/x5cz2/}.
}
\details{
Comparison data (CD) extends parallel analysis by reproducing the
observed correlation matrix rather than generating random data: datasets with a
known factor structure are generated with an increasing number of factors, and
the smallest number for which adding a further factor no longer significantly
improves the reproduction of the observed eigenvalues is retained (Ruscio &
Roche, 2012).

Because it reproduces the observed correlation structure instead of a null model,
CD was among the more accurate criteria across a broad range of conditions in
Ruscio and Roche (2012). It is, however, the only criterion in this family that
requires raw data, and by some margin the most computationally intensive one: a
finite population of \code{N_pop} cases is generated and \code{N_samples} samples are drawn
from it at every candidate factor count. It is therefore a good choice when the
raw data are at hand and the runtime is acceptable, and a poor one for a quick
look at a correlation matrix.

The comparison data are obtained by simulation, so the suggested number of factors
varies slightly from run to run. Call \code{\link[base:set.seed]{base::set.seed()}} beforehand to make a run
reproducible.

Note that if the data contains missing values, these will be removed for the
comparison data procedure using \code{\link[stats:na.fail]{stats::na.omit()}}. If
missing data should be treated differently, e.g., by imputation, do this outside
\code{efa_cd()} and then pass the complete data.
}
\examples{
\donttest{
# determine n factors of the GRiPS
efa_cd(GRiPS_raw, N_pop = 500, N_samples = 20)

# determine n factors of the DOSPERT risk subscale
efa_cd(DOSPERT_raw, N_pop = 500, N_samples = 20)
}
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.KGC.R
\name{plot.KGC}
\alias{plot.KGC}
\title{Plot KGC object}
\usage{
\method{plot}{KGC}(x, ...)
}
\arguments{
\item{x}{a list of class KGC. An output from the \link{KGC} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \link{KGC} function
}
\examples{
KGC_base <- KGC(test_models$baseline$cormat)
plot(KGC_base)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{EFA_AVERAGE}
\alias{EFA_AVERAGE}
\title{Model averaging across different EFA methods and types}
\usage{
EFA_AVERAGE(
  x,
  n_factors,
  N = NA,
  method = "PAF",
  rotation = "promax",
  type = "none",
  averaging = c("mean", "median"),
  trim = 0,
  salience_threshold = 0.3,
  max_iter = 10000,
  init_comm = c("smc", "mac", "unity"),
  criterion = c(0.001),
  criterion_type = c("sum", "max_individual"),
  abs_eigen = c(TRUE),
  varimax_type = c("svd", "kaiser"),
  normalize = TRUE,
  k_promax = 2:4,
  k_simplimax = ncol(x),
  P_type = c("norm", "unnorm"),
  precision = 1e-05,
  start_method = c("psych", "factanal"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra", "fiml"),
  show_progress = TRUE
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations. If raw data is entered, the correlation matrix is found from the
data.}

\item{n_factors}{numeric. Number of factors to extract.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used. If input is a correlation matrix and \code{N} = NA
(default), not all fit indices can be computed.}

\item{method}{character vector. Any combination of \code{"PAF"}, \code{"ML"}, and \code{"ULS"},
the estimators to average across; passed to \code{\link[=efa_average]{efa_average()}} as its \code{estimator}
argument. Default is \code{"PAF"}.}

\item{rotation}{character vector. Either perform no rotation ("none"),
any combination of orthogonal rotations ("varimax", "equamax", "quartimax", "geominT",
"bentlerT", and "bifactorT"; using "orthogonal" runs all of these), or of
oblique rotations ("promax", "oblimin", "quartimin", "simplimax", "bentlerQ",
"geominQ", and "bifactorQ"; using "oblique" runs all of these). Rotation types
(no rotation, orthogonal rotations, and oblique rotations) cannot be mixed.
Default is "promax".}

\item{type}{character vector. Any combination of "none" (default), "EFAtools",
"psych", and "SPSS" can be entered. "none" allows the specification of various
combinations of the arguments controlling both factor extraction methods and
the rotations. The others ("EFAtools", "psych", and "SPSS") take the extraction
and rotation tuning of the respective implementation: this package's default
procedure, the psych package's, and SPSS's. A specific psych implementation
exists for PAF, ML, varimax, and promax. The SPSS implementation exists for
PAF, varimax, and promax. For details, see \code{\link[=efa_fit]{efa_fit()}}. The factor ordering is
the one setting a named \code{type} does not bring here: every solution in the grid
is fitted with the eigenvalue-based ordering, so that the solutions can be
realigned to a common target before averaging.}

\item{averaging}{character. One of "mean" (default), and "median". Controls
whether the different results should be averaged using the (trimmed) mean,
or the median.}

\item{trim}{numeric. If averaging is set to "mean", this argument controls
the trimming of extremes (for details see \code{\link[base:mean]{base::mean()}}).
By default no trimming is done (i.e., trim = 0).}

\item{salience_threshold}{numeric. The threshold to use to classify a pattern
coefficient or loading as salient (i.e., substantial enough to assign it to
a factor). Default is 0.3. Indicator-to-factor correspondences will be inferred
based on this threshold. Note that this may not be meaningful if rotation = "none"
and n_factors > 1 are used, as no simple structure is present there.}

\item{max_iter}{numeric. The maximum number of iterations to perform after which
the iterative PAF procedure is halted with a warning. Default is 10,000. It is
only evaluated for the "PAF" solutions run under \code{type} "none": a named \code{type}
brings the iteration cap that defines it ("SPSS" 25, "psych" 50, and
"EFAtools" 300), and "ML" and "ULS" do not iterate this way. Note
that non-converged procedures are excluded from the averaging procedure.}

\item{init_comm}{character vector. Any combination of "smc", "mac", and "unity".
Controls the methods to estimate the initial communalities in \code{PAF} if
"none" is among the specified types. "smc" will use squared multiple
correlations, "mac" will use maximum absolute correlations, "unity" will use
1s (for details see \code{\link[=efa_fit]{efa_fit()}}). Default is \code{c("smc", "mac", "unity")}.}

\item{criterion}{numeric vector. The convergence criterion used for PAF if
"none" is among the specified types.
If the change in communalities from one iteration to the next is smaller than
this criterion the solution is accepted and the procedure ends.
Default is \code{0.001}.}

\item{criterion_type}{character vector. Any combination of "max_individual" and
"sum". Type of convergence criterion used for PAF if "none" is among the
specified types. "max_individual" selects the maximum change in any of the
communalities from one iteration to the next and tests it against the
specified criterion. "sum" takes the difference of
the sum of all communalities in one iteration and the sum of all communalities
in the next iteration and tests this against the criterion
(for details see \code{\link[=efa_fit]{efa_fit()}}). Default is \code{c("sum", "max_individual")}.}

\item{abs_eigen}{logical vector. Any combination of TRUE and FALSE.
Which algorithm to use in the PAF iterations if "none" is among the specified
types. If FALSE, the loadings are computed from the eigenvalues. This is also
used by the \code{\link[psych:fa]{psych::fa()}} function. If TRUE the
loadings are computed with the absolute eigenvalues as done by SPSS
(for details see \code{\link[=efa_fit]{efa_fit()}}). Default is \code{TRUE}.}

\item{varimax_type}{character vector. Any combination of "svd" and "kaiser".
The type of the varimax rotation performed if "none" is among the specified
types and "varimax", "promax", "orthogonal", or "oblique" is among the specified
rotations. "svd" uses singular value decomposition, as
\code{\link[stats:varimax]{stats::varimax()}} does, and "kaiser" uses the varimax
procedure performed in SPSS. This is the original procedure from Kaiser (1958),
but with slight alterations in the varimax criterion (for details, see
\code{\link[=efa_fit]{efa_fit()}} and Grieder & Steiner, 2022).
Default is \code{c("svd", "kaiser")}.}

\item{normalize}{logical vector. Any combination of TRUE and FALSE.
\code{TRUE} performs a kaiser normalization before the specified rotation(s).
Default is \code{TRUE}.}

\item{k_promax}{numeric vector. The power used for computing the target matrix
P in the promax rotation if "none" is among the specified types and "promax"
or "oblique" is among the specified rotations. Default is \code{2:4}.}

\item{k_simplimax}{numeric. The number of 'close to zero loadings' for the
simplimax rotation if "simplimax" or "oblique" is among the specified rotations. Default
is \code{ncol(x)}, where x is the entered data. It counts loadings, so each value must be a
whole number no larger than the number of loadings in the solution; a simplimax fit
given anything else fails and is reported as an errored solution in the grid.}

\item{P_type}{character vector. Any combination of \code{"norm"} and \code{"unnorm"}.
How the promax target matrix P is computed if \code{"none"} is among the specified
types and \code{"promax"} or \code{"oblique"} is among the specified rotations:
\code{"unnorm"} uses the unnormalized target matrix of Hendrickson and White
(1964), \code{"norm"} a normalized one. This frozen argument keeps its original
name; \code{\link[=efa_average]{efa_average()}} takes the same setting as \code{p_type}. Default is
\code{c("norm", "unnorm")}.}

\item{precision}{numeric vector. The tolerance for stopping in the rotation
procedure(s). Default is 10^-5.}

\item{start_method}{character vector. Any combination of "psych" and "factanal".
How to specify the starting values for the optimization procedure for ML.
"psych" takes the starting values specified in \code{\link[psych:fa]{psych::fa()}}.
"factanal" takes the starting values specified in the
\code{\link[stats:factanal]{stats::factanal()}} function. Default is
\code{c("psych", "factanal")}.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is "pairwise.complete.obs". It is ignored when
\code{cor_method = "fiml"}, which handles the missingness itself, so every case
contributes.}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator), or \code{"fiml"}
for a two-stage full-information maximum-likelihood correlation from raw data
with missing values. With \code{"fiml"} the saturated multivariate-normal mean and
covariance are estimated by an EM algorithm assuming the data are missing at
random and the standardized covariance is analysed, reproducing
\code{psych::corFiml()} followed by \code{psych::fa()} and
\code{lavaan(missing = "two.stage")}, \emph{not} \code{lavaan::efa(missing = "ml")} (see
\code{\link[=efa_fit]{efa_fit()}} and the details). Default is "pearson".}

\item{show_progress}{logical. Whether a progress bar should be shown in the
console. Default is TRUE.}
}
\value{
The value of \code{\link[=efa_average]{efa_average()}}, normally a list of class
\code{c("efa_average", "EFA_AVERAGE")}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{EFA_AVERAGE()} has been superseded by \code{\link[=efa_average]{efa_average()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_average]{efa_average()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SPSS_23_doc.R
\docType{data}
\name{SPSS_23}
\alias{SPSS_23}
\title{Various outputs from SPSS (version 23) FACTOR}
\format{
A list of 9 containing EFA results for each of the data sets mentioned above. Each of these nine entries is a list of 4 or 8 (see details), of the following structure:
\describe{
  \item{paf_comm}{(vector) - The final communalities obtained with the FACTOR algorithm with PAF and no rotation. For details, see Grieder and Grob (2019).}
  \item{paf_load}{(matrix) - F1 to FN = unrotated factor loadings obtained with the FACTOR algorithm with PAF. Rownames are the abbreviated subtest names.}
  \item{paf_iter}{(numeric) - Number of iterations needed for the principal axis factoring to converge.}
  \item{var_load}{(matrix) - F1 to FN = varimax rotated factor loadings obtained with the FACTOR algorithm with PAF. Rownames are the abbreviated subtest names.}
  \item{pro_load}{(matrix) - F1 to FN = promax rotated factor loadings obtained with the FACTOR algorithm with PAF. Rownames are the abbreviated subtest names.}
  \item{pro_phi}{(matrix) - F1 to FN = intercorrelations of the promax rotated loadings.}
  \item{sl}{(matrix) - g = General / second order factor of the Schmid-Leiman solution. F1 to FN  = First order factors of the Schmid-Leiman solution. h2 = Communalities of the Schmid-Leiman solution. This Schmid-Leiman solution was found using the SPSS Syntax provided by Wolff and Preising (2005).}
  \item{L2}{(matrix) - Second order loadings used for the Schmid-Leiman transformation. This Schmid-Leiman solution was found using the SPSS Syntax provided by Wolff and Preising (2005).}
 }
}
\source{
Grieder, S., & Steiner, M.D. (2020). Algorithmic Jingle Jungle:
A Comparison of Implementations of Principal Axis Factoring and Promax Rotation
 in R and SPSS. Manuscript in Preparation.

Wolff, H.G., & Preising, K. (2005). Exploring item and higher order factor structure with the Schmid-Leiman solution: Syntax codes for SPSS and SAS. Behavior Research Methods, 37, 48–58. doi: 10.3758/BF03206397

Grieder, S., & Grob, A. (2019). Exploratory factor analyses of the intelligence and development scales–2: Implications for theory and practice. Assessment. Advance online publication. doi:10.1177/10731911198450

Grob, A., & Hagmann-von Arx, P. (2018). Intelligence and Development Scales--2 (IDS-2). Intelligenz- und Entwicklungsskalen für Kinder und Jugendliche.
[Intelligence and Development Scales for Children and Adolescents.]. Bern, Switzerland: Hogrefe.

Frey, R., Pedroni, A., Mata, R., Rieskamp, J., & Hertwig, R. (2017). Risk preference shares the psychometric structure of major psychological traits. Science Advances, 3, e1701381.

McGrew, K. S., LaForte, E. M., & Schrank, F. A. (2014). Technical
 Manual. Woodcock-Johnson IV. Rolling Meadows, IL: Riverside.

Schrank, F. A., McGrew, K. S., & Mather, N. (2014). Woodcock-Johnson IV.
Rolling Meadows, IL: Riverside.

Costa, P. T., & McCrae, R. R. (1992). NEO PI-R professional manual. Odessa, FL: Psychological Assessment Resources, Inc.
}
\usage{
SPSS_23
}
\description{
Various outputs from SPSS (version 23) FACTOR for the IDS-2 (Grob & Hagmann-von Arx, 2018), the WJIV (3 to 5 and 20 to 39 years; McGrew, LaForte, & Schrank, 2014), the DOSPERT (Frey et al., 2017; Weber,
Blais, & Betz, 2002), the NEO-PI-R (Costa, & McCrae, 1992), and four simulated datasets (baseline, case_1a, case_6b, and case_11b, see \link{test_models} and \link{population_models}) used in Grieder and Steiner (2020).
}
\details{
The IDS-2, the two WJIV, the DOSPERT, and the NEO-PI-R contain all the above entries, while the four simulated datasets contain only paf_load, var_load, pro_load, and pro_phi.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_schmid_leiman.R
\name{print.efa_schmid_leiman}
\alias{print.efa_schmid_leiman}
\alias{format.efa_schmid_leiman}
\title{Print and format an efa_schmid_leiman object}
\usage{
\method{print}{efa_schmid_leiman}(x, ...)

\method{format}{efa_schmid_leiman}(x, ...)
}
\arguments{
\item{x}{An object of class \code{efa_schmid_leiman} (output from
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}).}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows a summarised output of the \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} function: a
model header (when the settings are available), the Schmid-Leiman loading
matrix, and the variances accounted for. \code{format()} assembles the same report
and returns it as a character vector; \code{print()} is \code{cat(format(x), sep = "\\n")}.
The lines follow the active console theme, so they are plain when colours are
disabled (for example when captured into a file or stripped with
\code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
EFA_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
sl_mod <- efa_schmid_leiman(EFA_mod, estimator = "PAF")
sl_mod

# format() returns the same lines as a character vector:
writeLines(format(sl_mod))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WJIV_ages_3_5_doc.R
\docType{data}
\name{WJIV_ages_3_5}
\alias{WJIV_ages_3_5}
\title{Woodcock Johnson IV: ages 3 to 5}
\format{
A list of 2 with elements "cormat" (29 x 29 matrix of bivariate correlations)
and "N" (scalar). The correlation matrix contains the following variables:
\describe{
\item{ORLVOC}{(numeric) - Oral Vocabulary.}
\item{VRBATN}{(numeric) - Verbal Attention.}
\item{LETPAT}{(numeric) - Phonological Processing.}
\item{STYREC}{(numeric) - Story Recall.}
\item{VISUAL}{(numeric) - Visualization.}
\item{GENINF}{(numeric) - General Information.}
\item{CONFRM}{(numeric) - Concept Formation.}
\item{NUMREV}{(numeric) - Numbers Reversed.}
\item{NUMPAT}{(numeric) - Number-Pattern Matching.}
\item{NWDREP}{(numeric) - Nonword Repetition.}
\item{VAL}{(numeric) - Visual-Auditory Learning.}
\item{PICREC}{(numeric) - Picture Recognition.}
\item{MEMWRD}{(numeric) - Memory for Words.}
\item{PICVOC}{(numeric) - Picture Vocabulary.}
\item{ORLCMP}{(numeric) - Oral Comprehension.}
\item{SEGMNT}{(numeric) - Segmentation.}
\item{RPCNAM}{(numeric) - Rapid Picture Naming.}
\item{SENREP}{(numeric) - Sentence Repetition.}
\item{UNDDIR}{(numeric) - Understanding Directions.}
\item{SNDBLN}{(numeric) - Sound Blending.}
\item{RETFLU}{(numeric) - Retrieval Fluency.}
\item{SNDAWR}{(numeric) - Sound Awareness.}
\item{LWIDNT}{(numeric) - Letter-Word Identification.}
\item{APPROB}{(numeric) - Applied Problems.}
\item{SPELL}{(numeric) - Spelling.}
\item{PSGCMP}{(numeric) - Passage Comprehension.}
\item{SCI}{(numeric) - Science.}
\item{SOC}{(numeric) - Social Studies.}
\item{HUM}{(numeric) - Humanities.}
}
}
\source{
McGrew, K. S., LaForte, E. M., & Schrank, F. A. (2014). Technical
Manual. Woodcock-Johnson IV. Rolling Meadows, IL: Riverside.

Schrank, F. A., McGrew, K. S., & Mather, N. (2014). Woodcock-Johnson IV.
Rolling Meadows, IL: Riverside.
}
\usage{
WJIV_ages_3_5
}
\description{
A list containing the bivariate correlations (N = 435) of the 29
cognitive and achievement subtests from the WJ IV for 3- to 5-year-olds from
the standardization sample obtained from
the WJ IV technical Manual (McGrew, LaForte, & Schrank, 2014).
Tables are reproduced with permission from the publisher.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.EFA_AVERAGE.R
\name{plot.EFA_AVERAGE}
\alias{plot.EFA_AVERAGE}
\title{Plot EFA_AVERAGE object}
\usage{
\method{plot}{EFA_AVERAGE}(x, ...)
}
\arguments{
\item{x}{list. An output from the \link{EFA_AVERAGE} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \link{EFA_AVERAGE} function
}
\examples{
\dontrun{
EFA_aver <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)
EFA_aver
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_kmo.R
\name{efa_kmo}
\alias{efa_kmo}
\title{Kaiser-Meyer-Olkin criterion}
\source{
Kaiser, H. F. (1970). A second generation little jiffy. Psychometrika,
35, 401-415.

Kaiser, H. F. & Rice, J. (1974). Little jiffy, mark IV. Educational
and Psychological Measurement, 34, 111-117.

Cureton, E. E. & D'Agostino, R. B. (1983). Factor analysis: An
applied approach. Hillsdale, N.J.: Lawrence Erlbaum Associates, Inc.
}
\usage{
efa_kmo(
  x,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{use}{character. The missing-data policy for raw data. Passed to
\code{\link[stats:cor]{stats::cor()}} for \code{"pearson"}, \code{"spearman"}, and \code{"kendall"}; for \code{"poly"} /
\code{"tetra"} the same policies are applied to the raw data before the polychoric
estimation, where \code{"all.obs"} and \code{"everything"} abort on a missing value instead
of returning \code{NA} correlations. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator).
Default is "pearson".}
}
\value{
A list containing
\item{KMO}{Overall KMO.}
\item{KMO_i}{KMO for each variable.}
\item{settings}{A list of the settings used.}
}
\description{
This function computes the Kaiser-Meyer-Olkin (KMO) criterion overall and for
each variable in a correlation matrix. The KMO represents the degree to
which each observed variable is predicted by the other variables in the
dataset and with this indicates the suitability for factor analysis.
}
\details{
Kaiser (1970) proposed this index, originally called measure of
sampling adequacy (MSA), that indicates how near the inverted correlation
matrix \eqn{R^{-1}} is to a diagonal matrix to determine a given
correlation matrix's (\eqn{R}) suitability for factor analysis.
The index is
\deqn{KMO = \frac{\sum_{i \neq j} r_{ij}^2}{\sum_{i \neq j} r_{ij}^2 + \sum_{i \neq j} q_{ij}^2}}
with \eqn{Q = SR^{-1}S} and S = \eqn{(diag R^{-1})^{-1/2}} where
\eqn{\sum_{i \neq j} r_{ij}^2} is the sum of squares of the off-diagonal
elements of \eqn{R} and \eqn{\sum_{i \neq j} q_{ij}^2} is the sum of squares of
the off-diagonal elements of \eqn{Q} (see also Cureton & D'Agostino, 1983).

So KMO varies between 0 and 1, with larger values indicating higher suitability
for factor analysis. Kaiser and Rice (1974) suggest that KMO should at least
exceed .50 for a correlation matrix to be suitable for factor analysis.

This function was heavily influenced by the \code{\link[psych:KMO]{psych::KMO()}}
function.

See also \code{\link[=efa_bartlett]{efa_bartlett()}} for another test of suitability for factor
analysis.

The \code{efa_kmo} function can also be called together with the
\code{\link[=efa_bartlett]{efa_bartlett()}} function and with factor retention criteria in the
\code{\link[=efa_retain]{efa_retain()}} function.
}
\examples{
efa_kmo(test_models$baseline$cormat)
}
\seealso{
\code{\link[=efa_bartlett]{efa_bartlett()}} for another measure to determine
suitability for factor analysis.

\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this function,
\code{\link[=efa_bartlett]{efa_bartlett()}} and several factor retention criteria.

Other factor analysis suitability:
\code{\link[=efa_bartlett]{efa_bartlett()}},
\code{\link[=efa_screen]{efa_screen()}},
\code{\link[=print.efa_screen]{print.efa_screen()}}
}
\concept{factor analysis suitability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.efa_power.R
\name{plot.efa_power}
\alias{plot.efa_power}
\title{Plot the RMSEA power curve}
\usage{
\method{plot}{efa_power}(x, n = NULL, df = NULL, eps1 = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{efa_power} (output from \code{\link[=efa_power]{efa_power()}}).}

\item{n}{numeric. The total sample sizes to evaluate. If \code{NULL} (the default) a
sequence bracketing the object's sample size is chosen automatically.}

\item{df}{numeric. The model degrees of freedom (must be positive). Defaults to the
object's \code{df}; a vector of length greater than one draws one curve per value.}

\item{eps1}{numeric. The alternative-hypothesis RMSEA (must differ from the null
\code{eps0}). Defaults to the object's \code{eps1}; a vector of length greater than one draws
one curve per value. At most one of \code{df} and \code{eps1} may be a vector.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2::ggplot} object.
}
\description{
Draws the analytic RMSEA power (MacCallum, Browne, & Sugawara, 1996) of an
\code{\link[=efa_power]{efa_power()}} result as a function of the total sample size, mirroring
\code{semTools::plotRMSEApower()} but returning a \link[ggplot2:ggplot]{ggplot2::ggplot} object rather than
drawing to the active device. The test, its null and alternative RMSEA, the
significance level, and the number of groups are taken from the object; only the
sample-size axis is swept, with an optional sweep of the degrees of freedom or the
alternative RMSEA to overlay several curves.
}
\details{
When the plotted curve is the object's own -- a single curve with neither \code{df} nor
\code{eps1} overridden -- it is annotated with the object's result: a dashed vertical line
at its sample size \code{x$N}, a dashed horizontal line at the reference power (the target
power when a sample size was solved for, otherwise the power achieved at \code{x$N}), and a
point at \code{x$N} and the achieved power. Overriding \code{df} or \code{eps1}, sweeping either as a
vector, or supplying an \code{n} that does not span \code{x$N} moves that point off the drawn
curve, so the marks are then omitted.
}
\examples{
pw <- efa_power(df = 100, N = 200)

# Power curve for the test of close fit, marking the object's own N
plot(pw)

# Overlay several models by sweeping the degrees of freedom
plot(pw, df = c(50, 100, 200))

# Sweep the alternative RMSEA instead
plot(pw, eps1 = c(0.06, 0.08, 0.10))
}
\references{
MacCallum, R. C., Browne, M. W., & Sugawara, H. M. (1996). Power analysis and
determination of sample size for covariance structure modeling. \emph{Psychological
Methods, 1}(2), 130-149. \doi{10.1037/1082-989X.1.2.130}
}
\seealso{
Other power analysis:
\code{\link[=efa_power]{efa_power()}},
\code{\link[=print.efa_power]{print.efa_power()}}
}
\concept{power analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_procrustes.R
\name{efa_procrustes}
\alias{efa_procrustes}
\title{Rotate a loading matrix to a target using Procrustes alignment}
\usage{
efa_procrustes(
  A,
  Target,
  rotation = c("orthogonal", "oblique"),
  S = NULL,
  T_init = NULL,
  oblique_eps = 1e-05,
  oblique_maxit = 1000,
  oblique_max_line_search = 10,
  oblique_step0 = 1,
  oblique_normalize = FALSE,
  oblique_random_starts = 0,
  oblique_screen_keep = 2,
  oblique_triage_maxit = 25,
  oblique_triage_improve_tol = 0
)
}
\arguments{
\item{A}{Numeric loading matrix to be aligned.}

\item{Target}{Numeric target matrix with the same dimensions as \code{A}.}

\item{rotation}{Character string, either \code{"orthogonal"} or \code{"oblique"}.}

\item{S}{Optional \verb{k x k} cross-product matrix \code{crossprod(A)}, kept for
compatibility. It enters both the oblique criterion and its gradient, so any
other matrix would minimize a different criterion: where \code{S} is used it is
checked against \code{crossprod(A)} and must agree with it up to a relative
tolerance of \code{1e-8}. That check forms \code{crossprod(A)} itself, so passing \code{S}
no longer avoids any work: omitting it gives the same result for slightly
less. \code{S} is used, and therefore checked, only on the oblique path with more
than one factor and \code{oblique_normalize = FALSE}; if Kaiser normalization is
requested, the cross-product must be recomputed on the normalized matrix and
\code{S} is ignored.}

\item{T_init}{Optional \verb{k x k} starting transformation matrix for the oblique
solver. Its columns are normalized internally, and the normalized matrix must
be well enough conditioned to define a proper factor correlation matrix: its
smallest singular value must be at least \code{1e-4}, the same floor the solver
applies to every candidate it evaluates. If \code{NULL} (the default), the oblique
solver is warm-started from the closed-form orthogonal Procrustes solution.}

\item{oblique_eps}{Positive convergence tolerance for the projected-gradient
norm in the oblique solver.}

\item{oblique_maxit}{Non-negative integer. Maximum number of projected-gradient
updates in the full oblique solver.}

\item{oblique_max_line_search}{Non-negative integer. Maximum number of
step-halving attempts after the initial line-search step.}

\item{oblique_step0}{Positive initial step size for the oblique solver.}

\item{oblique_normalize}{Logical; if \code{TRUE}, apply Kaiser row normalization to
the loadings (only) in the oblique solver and back-transform the aligned
loadings afterwards, leaving \code{Target} unnormalized (as in
\code{GPArotation::targetQ(normalize = TRUE)}).}

\item{oblique_random_starts}{Non-negative integer. Number of additional random
starts used by the oblique solver.}

\item{oblique_screen_keep}{Non-negative integer. Number of random starts
retained after cheap objective screening and sent to triage optimization.}

\item{oblique_triage_maxit}{Non-negative integer. Number of short optimization
iterations used in the triage stage.}

\item{oblique_triage_improve_tol}{Non-negative scalar. Relative improvement
required for a triaged start to be promoted to full optimization.}
}
\value{
A list. Every path returns the following components:

\item{loadings}{Aligned loading matrix.}
\item{T}{Transformation matrix.}
\item{Phi}{Factor intercorrelation matrix; the identity for orthogonal and
one-factor alignment.}
\item{value}{Target criterion at the returned solution.}
\item{convergence}{Logical; \code{TRUE} for the closed-form orthogonal solution.}
\item{valid}{Logical; whether the transformation defines an admissible \code{Phi}.}
\item{iterations}{Number of solver iterations; \code{0} for the closed-form
orthogonal solution.}
\item{kappa_T}{Condition number of \code{T}; a constant \code{1} on the orthogonal path.}
\item{Table}{Iteration history with columns \code{iter}, \code{f}, \code{log10_s}, and \code{step};
a single placeholder row on the orthogonal path.}
\item{method}{\code{"orthogonal_procrustes"}, \code{"oblique_procrustes"}, or
\code{"single_factor_procrustes"} for a one-factor oblique request.}
\item{line_search_failed}{Logical line-search diagnostic.}
\item{best_start_index, all_start_indices, all_values, all_converged,
all_iterations}{Multi-start summary of the starts that were fully
optimized; each has a single entry when no random starts were used.}

The oblique solver additionally returns \code{screen_start_indices} and
\code{screen_values} (the starts kept by cheap objective screening and their
criterion values) together with the counts \code{n_random_starts}, \code{n_screened},
\code{n_triaged}, and \code{n_fully_optimized}. These six components are absent for
\code{rotation = "orthogonal"} and for one-factor models, which are aligned with the
orthogonal solution.

Row and column names are preserved where possible. When
\code{oblique_normalize = TRUE} the returned \code{loadings} are back-transformed to the
original scale, but \code{value} is the criterion on the Kaiser-normalized loadings,
so it is not \code{0.5 * sum((loadings - Target)^2)}.
}
\description{
\code{efa_procrustes()} aligns one loading matrix to a target loading matrix with the
same dimensions. It is used internally by \code{efa_mi()}, but can also be used
directly when factor columns must be brought into a common orientation before
averaging or comparing solutions.
}
\details{
For \code{rotation = "orthogonal"}, the function solves the closed-form orthogonal
Procrustes problem

\deqn{\min_T \frac{1}{2}\|A T - B\|_F^2 \quad \textrm{subject to}\quad T'T = I,}

where \code{A} is the loading matrix and \code{B} is \code{Target}.

For \code{rotation = "oblique"}, the function calls the compiled
\code{.oblique_procrustes()} optimizer. The oblique convention is the same as in
\code{GPArotation::targetQ()}:

\deqn{L = A T^{-T}, \qquad \Phi = T'T, \qquad diag(\Phi) = 1.}

By default the oblique solver is warm-started from the closed-form orthogonal
Procrustes solution, which resolves the factor permutation and sign
indeterminacy and avoids the poor local minima an identity start can fall
into. Supply \code{T_init} to override this start. Random starts are only used for
oblique alignment. For one-factor models, oblique and orthogonal alignment are
equivalent, so the function uses the stable one-factor orthogonal solution
instead of calling the oblique optimizer.
}
\examples{
## Align an estimated loading matrix to a known target pattern: fit an
## unrotated three-factor model, then rotate its loadings toward the true
## population pattern.
efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "none")
target <- population_models$loadings$baseline

## Orthogonal target rotation (rigid rotation/reflection):
efa_procrustes(efa_mod$unrot_loadings, target, rotation = "orthogonal")

## Oblique target rotation (lets the aligned factors correlate):
efa_procrustes(efa_mod$unrot_loadings, target, rotation = "oblique")

}
\seealso{
Other factor rotation:
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}
}
\concept{factor rotation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_retention.R
\name{plot.efa_retention}
\alias{plot.efa_retention}
\title{Plot method for efa_retention objects}
\usage{
\method{plot}{efa_retention}(x, ...)
}
\arguments{
\item{x}{an object of class efa_retention, returned by a factor-retention
criterion (e.g. \code{\link[=efa_ekc]{efa_ekc()}} or \code{\link[=efa_hull]{efa_hull()}}).}

\item{...}{not used.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2::ggplot} object, or invisibly \code{NULL} if the criterion has
no plottable result.
}
\description{
Plots the result of a factor-retention criterion. Eigenvalue-based criteria
(e.g. \code{\link[=efa_ekc]{efa_ekc()}}) are shown as an eigenvalue plot, the Hull method (\code{\link[=efa_hull]{efa_hull()}}) as
a convex-hull plot. Criteria with more than one sub-variant are faceted.
}
\examples{
plot(efa_ekc(test_models$baseline$cormat, N = 500))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.SL.R
\name{print.SL}
\alias{print.SL}
\title{Print SL object}
\usage{
\method{print}{SL}(x, ...)
}
\arguments{
\item{x}{list. An object of class SL to be printed}

\item{...}{Further arguments for print.}
}
\description{
Print Method showing a summarized output of the \link{SL} function.
}
\examples{
EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
SL(EFA_mod, type = "EFAtools", method = "PAF")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.HULL.R
\name{print.HULL}
\alias{print.HULL}
\title{Print function for HULL objects}
\usage{
\method{print}{HULL}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class HULL. Output from the \code{\link{HULL}} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for HULL objects
}
\examples{
\donttest{
HULL(test_models$baseline$cormat, N = 500, method = "ML")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.change_class}
\alias{.change_class}
\title{Convert an \code{"efa_loadings"} table to matrix or a matrix to \code{"efa_loadings"}}
\usage{
.change_class(x, cl = "matrix")
}
\arguments{
\item{x}{A table of class \code{"matrix"} or \code{"efa_loadings"}.}

\item{cl}{A character vector with the class to change the table to. Should be
\code{c("efa_loadings", "LOADINGS")} or \code{"matrix"}.}
}
\value{
A table with the loadings, of class either \code{"efa_loadings"} or
\code{"matrix"}.
}
\description{
The loadings tables returned by \code{\link[=efa_fit]{efa_fit()}} are of class
\code{c("efa_loadings", "LOADINGS")}, which prevents applying functions on them.
This function changes their class to \code{"matrix"}, and changes it back to
\code{"efa_loadings"} when done.
}
\author{
Andreas Soteriades
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{EFA}
\alias{EFA}
\title{Exploratory factor analysis (EFA)}
\usage{
EFA(
  x,
  n_factors,
  N = NA,
  method = c("PAF", "ML", "ULS", "MINRES", "DWLS"),
  rotation = c("none", "varimax", "equamax", "quartimax", "geominT", "bentlerT",
    "bifactorT", "promax", "oblimin", "quartimin", "simplimax", "bentlerQ", "geominQ",
    "bifactorQ"),
  se = c("none", "information", "sandwich", "np-boot"),
  type = c("EFAtools", "psych", "SPSS", "none"),
  max_iter = NA,
  init_comm = NA,
  criterion = NA,
  criterion_type = NA,
  abs_eigen = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  varimax_type = NA,
  k = NA,
  normalize = TRUE,
  p_type = NA,
  precision = 1e-05,
  order_type = NA,
  start_method = "psych",
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra", "fiml"),
  b_boot = 1000,
  ci = 0.95,
  random_starts = 100,
  seed = NULL,
  P_type = lifecycle::deprecated(),
  randomStarts = lifecycle::deprecated(),
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations. If raw data is entered, the correlation matrix is found from the
data.}

\item{n_factors}{numeric. Number of factors to extract. Must be at least 1 and
smaller than the number of variables (the common factor model is not identified
otherwise). Use \code{\link[=efa_retain]{efa_retain()}} to decide on a value.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used; with raw data, \code{N} is found from the data instead.
\itemize{
\item With \code{N = NA}, not all fit indices can be computed; a positive \code{N} that is very
small relative to the number of variables leaves the chi-square-derived indices
unavailable as well, with a warning.
\item With raw data, \code{N} is the number of cases the correlation matrix was actually
computed from -- see \code{use} for the general rule and how missing values change it.
Under \code{cor_method = "fiml"}, \code{use} is ignored and \code{N} is instead the number of
cases carrying at least one observed value.
}}

\item{method}{character. The estimator used to fit the EFA; passed to \code{\link[=efa_fit]{efa_fit()}}
as its \code{estimator} argument. One of "PAF", "ML", "ULS", "MINRES" (an accepted
alias of "ULS"), or "DWLS"; see the \code{\link[=efa_fit]{efa_fit()}} documentation for their
properties and data requirements.}

\item{rotation}{character. Either perform no rotation ("none"; default),
an orthogonal rotation ("varimax", "equamax", "quartimax", "geominT",
"bentlerT", or "bifactorT"), or an oblique rotation ("promax", "oblimin",
"quartimin", "simplimax", "bentlerQ", "geominQ", or "bifactorQ"). See the
\emph{Rotations} section in Details for their properties and known issues.}

\item{se}{character. Whether and how to compute standard errors (and matching
confidence intervals): "none" (default), "information" (analytic
standard errors from the expected Fisher information of the ML solution), "sandwich"
(robust "sandwich" standard errors from raw data, which stay reliable under
non-normality or a misspecified estimator weight), or "np-boot" (non-parametric
bootstrap). The
methods differ in their assumptions, their data requirements, and which estimator,
rotation, and \code{cor_method} combinations they support; see the \emph{Standard errors}
section in Details.}

\item{type}{character. If one of "EFAtools" (default), "psych", or "SPSS" is
used, and the following arguments with default NA are left with
NA, these implementations are executed according to the respective program
("psych" and "SPSS") or according to the best solution found in Grieder &
Steiner (2022; "EFAtools"). Individual properties can be adapted using one of
the three types and specifying some of the following arguments. If set to
"none" additional arguments must be specified depending on the \code{method}
and \code{rotation} used (see details).}

\item{max_iter}{numeric. The maximum number of iterations to perform after which
the iterative PAF procedure is halted with a warning. If \code{type} is one of
"EFAtools", "SPSS", or "psych", this is automatically specified if \code{max_iter} is
left to be \code{NA}, but can be overridden by entering a number. Default is
\code{NA}.}

\item{init_comm}{character. The method to estimate the initial communalities
in \code{PAF}. "smc" will use squared multiple correlations, "mac" will use
maximum absolute correlations, "unity" will use 1s (see details).
Default is \code{NA}.}

\item{criterion}{numeric. The convergence criterion used for PAF.
If the change in communalities from one iteration to the next is smaller than
this criterion the solution is accepted and the procedure ends.
Default is \code{NA}.}

\item{criterion_type}{character. Type of convergence criterion used for
PAF. "max_individual" selects the maximum change in any of the
communalities from one iteration to the next and tests it against the
specified criterion. This is also used by SPSS. "sum" takes the difference of
the sum of all communalities in one iteration and the sum of all communalities
in the next iteration and tests this against the criterion. This procedure is
used by the \code{\link[psych:fa]{psych::fa()}} function. Default is \code{NA}.}

\item{abs_eigen}{logical. Which algorithm to use in the PAF
iterations. If FALSE, the loadings are computed from the eigenvalues. This is
also used by the \code{\link[psych:fa]{psych::fa()}} function. If TRUE the
loadings are computed with the absolute eigenvalues as done by SPSS.
Default is \code{NA}.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is "pairwise.complete.obs". It is ignored when
\code{cor_method = "fiml"} (which handles the missingness itself, so every case
contributes), and it is overridden to listwise deletion whenever an asymptotic
covariance is required (the "DWLS" estimator, or \code{se = "sandwich"}), because the
covariance must describe the same cases as the correlation matrix.}

\item{varimax_type}{character. The type of the varimax rotation performed.
If "svd", singular value decomposition is used, as \code{\link[stats:varimax]{stats::varimax()}} does. If
"kaiser", the varimax procedure performed in SPSS is used, following the original
procedure from Kaiser (1958) (see details). Default is \code{NA}.}

\item{k}{numeric. Either the power used for computing the target matrix P in
the promax rotation or the number of 'close to zero loadings' for the simplimax
rotation. If left to \code{NA} (default), the value for promax depends on the specified type.
For simplimax, \code{nrow(L)}, where L is the matrix of unrotated loadings,
is used by default.}

\item{normalize}{logical. If \code{TRUE}, a kaiser normalization is
performed before the specified rotation. Default is \code{TRUE}.}

\item{p_type}{character. This specifies how the target
matrix P is computed in promax rotation. If "unnorm" it will use the
unnormalized target matrix as originally done in Hendrickson and White (1964).
This is also used in the psych and stats packages. If "norm" it will use the
normalized target matrix as used in SPSS. Default is \code{NA}.}

\item{precision}{numeric. The tolerance for stopping in the rotation
procedure. Default is 10^-5 for all rotation methods.}

\item{order_type}{character. How to order the factors. "eigen" reorders the
factors by descending explained variance; "ss_factors" reorders the factors by
descending (unweighted) sum of squared factor loadings per factor. Default is \code{NA}.}

\item{start_method}{character. How to specify the starting values for the
optimization procedure for ML. Default is "psych" which takes the
starting values specified in \code{\link[psych:fa]{psych::fa()}}. "factanal" takes the
starting values specified in the \code{\link[stats:factanal]{stats::factanal()}} function.}

\item{cor_method}{character. How the correlation is computed from raw data:
\code{"pearson"}, \code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}); \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data; or
\code{"fiml"} for a two-stage full-information maximum-likelihood correlation from raw data
with missing values. See the \emph{Correlation methods} section in Details for their
properties and the combinations they support. Default is "pearson".}

\item{b_boot}{numeric. The number of bootstrap samples to draw. Default is 1000.
Must be at least 2, the smallest number from which a standard error is defined.
Under \code{cor_method = "fiml"} each bootstrap sample re-runs the EM moment
estimation, so a smaller value may be advisable.}

\item{ci}{numeric. The level of the confidence intervals: the percentile intervals from
the bootstrap samples under \code{se = "np-boot"}, and the analytic Wald intervals under
\code{se = "information"} and \code{se = "sandwich"}, the corrected two-stage intervals of
\code{cor_method = "fiml"} included. Must be greater than 0 and smaller than 1. Default is
.95 for 95\% CIs.}

\item{random_starts}{numeric. The number of random starts to use in the
rotation to guard against local minima. Default is 100.}

\item{seed}{numeric. An optional seed for the random-number generator.}

\item{P_type, randomStarts}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}} Former names of
\code{p_type} and \code{random_starts}. Still accepted (silently) for backwards
compatibility; please use the new names.}

\item{...}{Additional arguments passed to the rotation procedure (e.g., \code{maxit} for
the maximum number of iterations).}
}
\value{
The value of \code{\link[=efa_fit]{efa_fit()}}, a list of class \code{c("efa", "EFA")}; see there for the
components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{EFA()} has been superseded by \code{\link[=efa_fit]{efa_fit()}}, which is the recommended interface going
forward. \code{efa_fit()} keeps the primary choices (data, factors, estimator, rotation,
standard errors) as top-level arguments and collects the estimation and rotation tuning
knobs into two control objects, \code{\link[=estimate_control]{estimate_control()}} and \code{\link[=rotate_control]{rotate_control()}}. \code{EFA()}
remains available and unchanged -- its full flat argument list still works exactly as
before -- so existing code keeps running.
}
\seealso{
\code{\link[=efa_fit]{efa_fit()}}, \code{\link[=estimate_control]{estimate_control()}}, \code{\link[=rotate_control]{rotate_control()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.efa_average.R
\name{plot.efa_average}
\alias{plot.efa_average}
\title{Plot efa_average object}
\usage{
\method{plot}{efa_average}(x, ...)
}
\arguments{
\item{x}{list. An output from the \code{\link[=efa_average]{efa_average()}} function.}

\item{...}{not used.}
}
\value{
A ggplot object showing, for each indicator and factor, the minimum,
maximum, and average (mean or median) loading across the averaged solutions.
Each panel carries a point at the average, a bar spanning the minimum to the
maximum with a tick at each endpoint, and a grey band marking the loadings
that fall below the salience threshold; the caption names the four marks.
}
\description{
Plot method showing a summarized output of the \code{\link[=efa_average]{efa_average()}} function
}
\examples{
\donttest{
EFA_aver <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500)
plot(EFA_aver)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/population_models_doc.R
\docType{data}
\name{population_models}
\alias{population_models}
\title{population_models}
\format{
A list of 3 lists "loadings", "phis_3", and "phis_6".
\describe{
\code{loadings} contains the following matrices of pattern coefficients:
\item{baseline}{(matrix) - The pattern coefficients of the baseline model. Three factors with six indicators each, all with pattern coefficients of .6. Same baseline model as used in de Winter and Dodou (2012).}
\item{case_1a}{(matrix) - Three factors with 2 indicators per factor.}
\item{case_1b}{(matrix) - Three factors with 3 indicators per factor. Case 5 in de Winter and Dodou (2012).}
\item{case_1c}{(matrix) - Three factors with 4 indicators per factor.}
\item{case_1d}{(matrix) - Three factors with 5 indicators per factor.}
\item{case_2}{(matrix) - Same as baseline model but with low pattern coefficients of .3.}
\item{case_3}{(matrix) - Same as baseline model but with high pattern coefficients of .9.}
\item{case_4}{(matrix) - Three factors with different pattern coefficients \emph{between} factors (one factor with .9, one with .6, and one with .3, respectively). Case 7 in de Winter and Dodou (2012).}
\item{case_5}{(matrix) - Three factors with different pattern coefficients \emph{within} factors (each factor has two pattern coefficients of each .9, .6, and .3). Similar to cases 8/ 9 in de Winter and Dodou (2012).}
\item{case_6a}{(matrix) - Same as baseline model but with one cross loading of .4. Similar to case 10 in de Winter and Dodou (2012).}
\item{case_6b}{(matrix) - Same as baseline model but with three cross loading of .4 (One factor with 2 and one with 1 crossloading). Similar to case 10 in de Winter and Dodou (2012).}
\item{case_7}{(matrix) - Three factors with different number of indicators per factor (2, 4, and 6 respectively). Similar to cases 11/ 12 in de Winter and Dodou (2012).}
\item{case_8}{(matrix) - Three factors with random variation in pattern coefficients added, drawn from a uniform distribution between [-.2, .2]. Case 13 in de Winter and Dodou (2012).}
\item{case_9a}{(matrix) - Three factors with 2 indicators per factor, with different pattern coefficients within one of the factors.}
\item{case_9b}{(matrix) - Three factors with 3 indicators per factor, with different pattern coefficients.}
\item{case_9c}{(matrix) - Three factors with 4 indicators per factor, with different pattern coefficients.}
\item{case_9d}{(matrix) - Three factors with 5 indicators per factor, with different pattern coefficients.}
\item{case_10a}{(matrix) - Six factors with 2 indicators per factor, all with pattern coefficients of .6.}
\item{case_10b}{(matrix) - Six factors with 3 indicators per factor, all with pattern coefficients of .6.}
\item{case_10c}{(matrix) - Six factors with 4 indicators per factor, all with pattern coefficients of .6.}
\item{case_10d}{(matrix) - Six factors with 5 indicators per factor, all with pattern coefficients of .6.}
\item{case_10e}{(matrix) - Six factors with 6 indicators per factor, all with pattern coefficients of .6.}
\item{case_11a}{(matrix) - Six factors with 2 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
\item{case_11b}{(matrix) - Six factors with 3 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
\item{case_11c}{(matrix) - Six factors with 4 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
\item{case_11d}{(matrix) - Six factors with 5 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
\item{case_11e}{(matrix) - Six factors with 6 indicators per factor, with different pattern coefficients within and between factors (.3, .6, and .9).}
\item{case_12a}{(matrix) - One factor, with 2 equal pattern coefficients (.6).}
\item{case_12b}{(matrix) - One factor, with 3 equal pattern coefficients (.6).}
\item{case_12c}{(matrix) - One factor, with 6 equal pattern coefficients (.6).}
\item{case_12d}{(matrix) - One factor, with 10 equal pattern coefficients (.6).}
\item{case_12e}{(matrix) - One factor, with 15 equal pattern coefficients (.6).}
\item{case_13a}{(matrix) - One factor, with 2 different pattern coefficients (.3, and .6).}
\item{case_13b}{(matrix) - One factor, with 3 different pattern coefficients (.3, .6, and .9).}
\item{case_13c}{(matrix) - One factor, with 6 different pattern coefficients (.3, .6, and .9).}
\item{case_13d}{(matrix) - One factor, with 10 different pattern coefficients (.3, .6, and .9).}
\item{case_13e}{(matrix) - One factor, with 15 different pattern coefficients (.3, .6, and .9).}
\item{case_14a}{(matrix) - No factor, 2 variables (0).}
\item{case_14b}{(matrix) - No factor, 3 variables (0).}
\item{case_14c}{(matrix) - No factor, 6 variables (0).}
\item{case_14d}{(matrix) - No factor, 10 variables (0).}
\item{case_14e}{(matrix) - No factor, 15 variables (0).}
\code{phis_3} contains the following 3x3 matrices:
\item{zero}{(matrix) - Matrix of factor intercorrelations of 0. Same intercorrelations as used in de Winter and Dodou (2012).}
\item{moderate}{(matrix) - Matrix of moderate factor intercorrelations of .3.}
\item{mixed}{(matrix) - Matrix of mixed (.3, .5, and .7) factor intercorrelations.}
\item{strong}{(matrix) - Matrix of strong factor intercorrelations of .7. Same intercorrelations as used in de Winter and Dodou (2012).}
\code{phis_6} contains the following 6x6 matrices:
\item{zero}{(matrix) - Matrix of factor intercorrelations of 0. Same intercorrelations as used in de Winter and Dodou (2012).}
\item{moderate}{(matrix) - Matrix of moderate factor intercorrelations of .3.}
\item{mixed}{(matrix) - Matrix of mixed (around .3, .5, and .7; smoothing was necessary for the matrix to be positive definite) factor intercorrelations.}
\item{strong}{(matrix) - Matrix of strong factor intercorrelations of .7. Same intercorrelations as used in de Winter and Dodou (2012).}
}
}
\source{
Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
of implementations of principal axis factoring and promax rotation in R and SPSS.
Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x

de Winter, J.C.F., & Dodou, D. (2012). Factor recovery by principal axis factoring and maximum likelihood factor analysis as a function of factor pattern and sample size. Journal of Applied Statistics. 39.
}
\usage{
population_models
}
\description{
Population factor models, some of which (baseline to case_11e) used for the
simulation analyses reported in Grieder and Steiner (2022). All combinations
of the pattern matrices and the factor
intercorrelations were used in the simulations. Many models are based on cases
used in de Winter and Dodou (2012).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.KGC.R
\name{print.KGC}
\alias{print.KGC}
\title{Print function for KGC objects}
\usage{
\method{print}{KGC}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class KGC. Output from \link{KGC} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for KGC objects
}
\examples{
KGC_base <- KGC(test_models$baseline$cormat)
KGC_base

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KGC.R
\name{KGC}
\alias{KGC}
\title{Kaiser-Guttman Criterion}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200

Guttman, L. (1954). Some necessary conditions for common-factor analysis.
Psychometrika, 19, 149 –161. http://dx.doi.org/10.1007/BF02289162

Kaiser, H. F. (1960). The application of electronic computers to factor
analysis. Educational and Psychological Measurement, 20, 141–151.
http://dx.doi.org/10.1177/001316446002000116

Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
determining the number of components to retain. Psychological Bulletin, 99,
432–442. http://dx.doi.org/10.1037/0033-2909.99.3.432
}
\usage{
KGC(
  x,
  eigen_type = c("PCA", "SMC", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall"),
  n_factors = 1,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{eigen_type}{character. On what the eigenvalues should be found. Can be
either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
the diagonal values of the correlation matrices are left to be 1. If using
"SMC", the diagonal of the
correlation matrices is replaced by the squared multiple correlations (SMCs)
of the indicators. If using "EFA", eigenvalues are found on the correlation
matrices with the final communalities of an exploratory factor analysis
solution (default is principal axis factoring extracting 1 factor) as
diagonal.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Passed to \code{\link[stats:cor]{stats::cor}}.
Default is "pearson".}

\item{n_factors}{numeric. Number of factors to extract if "EFA" is included in
\code{eigen_type}. Default is 1.}

\item{...}{Additional arguments passed to \code{\link{EFA}}. For example,
to change the extraction method (PAF is default).}
}
\value{
A list of class KGC containing

\item{eigen_PCA}{ A vector containing the eigenvalues found with PCA.}
\item{eigen_SMC}{ A vector containing the eigenvalues found with SMCs.}
\item{eigen_EFA}{ A vector containing the eigenvalues found with EFA.}
\item{n_fac_PCA}{ The number of factors to retain according to the Kaiser-
Guttmann criterion with PCA eigenvalues type.}
\item{n_fac_SMC}{ The number of factors to retain according to the Kaiser-
Guttmann criterion with SMC eigenvalues type.}
\item{n_fac_EFA}{ The number of factors to retain according to the Kaiser-
Guttmann criterion with EFA eigenvalues type.}
\item{settings}{A list of the settings used.}
}
\description{
Probably the most popular factor retention criterion. Kaiser and Guttman suggested
to retain as many factors as there are sample eigenvalues greater than 1.
This is why the criterion is also known as eigenvalues-greater-than-one rule.
}
\details{
Originally, the Kaiser-Guttman criterion was intended for the use
with prinicpal components, hence with eigenvalues derived from the original
correlation matrix. This can be done here by setting \code{eigen_type} to
"PCA". However, it is well-known that this criterion is often inaccurate and
that it tends to overestimate the number of factors, especially for unidimensional
or orthogonal factor structures (e.g., Zwick & Velicer, 1986).

The criterion's inaccuracy in these cases is somewhat addressed if it is
applied on the correlation matrix with communalities in the diagonal, either
initial communalities estimated from SMCs (done setting \code{eigen_type} to
"SMC") or final communality estimates from an EFA (done setting \code{eigen_type}
to "EFA"; see Auerswald & Moshagen, 2019). However, although this variant
of the KGC is more accurate in some cases compared to the traditional KGC, it
is at the same time less accurate than the PCA-variant in other cases, and it
is still often less accurate than other factor retention methods, for
example parallel analysis (\code{\link{PARALLEL}}), the Hull method
\code{\link{HULL}}, or sequential \eqn{chi^2} model tests (\code{\link{SMT}};
see Auerswald & Moshagen, 2019).

The \code{KGC} function can also be called together with other factor
retention criteria in the \code{\link{N_FACTORS}} function.
}
\examples{
KGC(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
}
\seealso{
Other factor retention criteria: \code{\link{CD}}, \code{\link{EKC}},
\code{\link{HULL}}, \code{\link{PARALLEL}}, \code{\link{SMT}}

\code{\link{N_FACTORS}} as a wrapper function for this and all the
above-mentioned factor retention criteria.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OMEGA.R
\name{OMEGA}
\alias{OMEGA}
\title{McDonald's omega}
\source{
McDonald, R. P. (1978). Generalizability in factorable domains: ‘‘Domain
validity and generalizability’’. Educational and Psychological Measurement,
38, 75–79.

McDonald, R. P. (1985). Factor analysis and related methods. Hillsdale,
NJ: Erlbaum.

McDonald, R. P. (1999). Test theory: A unified treatment. Mahwah,
NJ: Erlbaum.

Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016a). Applying bifactor
statistical indices in the evaluation of psychological measures. Journal of
Personality Assessment, 98, 223-237.

Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016b). Evaluating
bifactor models: Calculating and interpreting statistical indices.
Psychological Methods, 21, 137-150.

Hancock, G. R., & Mueller, R. O. (2001). Rethinking construct reliability
within latent variable systems. In R. Cudeck, S. du Toit, & D. Sörbom (Eds.),
Structural equation modeling: Present and future—A Festschrift in honor of Karl
Jöreskog (pp. 195–216). Lincolnwood, IL: Scientific Software International.

Reise, S. P., Scheines, R., Widaman, K. F., & Haviland, M. G. (2013).
Multidimensionality and structural coefficient bias in structural equation
modeling: A bifactor perspective. Educational and Psychological Measurement,
73, 5–26.

Bonifay, W. E., Reise, S. P., Scheines, R., & Meijer, R. R. (2015).
When are multidimensional data unidimensional enough for structural equation
modeling?: An evaluation of the DETECT multidimensionality index. Structural
Equation Modeling, 22, 504—516.

Gignac, G. E. (2014). On the Inappropriateness of Using Items to
Calculate Total Scale Score Reliability via Coefficient Alpha for Multidimensional
Scales. European Journal of Psychological Assessment, 30, 130-139.
}
\usage{
OMEGA(
  model = NULL,
  type = c("EFAtools", "psych"),
  g_name = "g",
  group_names = NULL,
  add_ind = TRUE,
  factor_corres = NULL,
  var_names = NULL,
  fac_names = NULL,
  g_load = NULL,
  s_load = NULL,
  u2 = NULL,
  cormat = NULL,
  pattern = NULL,
  Phi = NULL,
  variance = c("correlation", "sums_load")
)
}
\arguments{
\item{model}{class \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}, class \code{schmid}, or class
\code{lavaan} object. That is, an output object from \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} or
\code{\link[psych:schmid]{psych::schmid()}}, or a \code{lavaan} fit object with a
single factor, second-order, or bifactor solution. If of class \code{lavaan},
only \code{g_name} needs to be specified additionally. If of class
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} or \code{schmid}, only the arguments \code{factor_corres}
and \code{cormat} need to be specified additionally.}

\item{type}{character. Either \code{"EFAtools"} (default) or \code{"psych"}
(see details)}

\item{g_name}{character. The name of the general factor from the lavaan solution.
This needs only be specified if \code{model} is a \code{lavaan} second-order
or bifactor solution. Default is "g".}

\item{group_names}{character. An optional vector of group names. The length
must correspond to the number of groups for which the \code{lavaan} model
was fitted.}

\item{add_ind}{logical. Whether additional indices (H index, ECV, PUC) should
be calculated or not (see details for these indices). If FALSE, only omegas
are returned. Default is \code{TRUE}.}

\item{factor_corres}{matrix. A logical matrix or a numeric matrix containing
0's and 1's that indicates which variable corresponds to which group factor.
Must have the same dimensions as the matrix of group factor loadings from the
SL solution. Cross-loadings are allowed here. See examples for use.}

\item{var_names}{character. A vector with subtest names in the order
of the rows from the SL solution. This needs only be specified if \code{model}
is left \code{NULL}.}

\item{fac_names}{character. An optional vector of group factor names in the
order of the columns of the SL solution. If left \code{NULL}, names of the
group factors from the entered solution are taken.}

\item{g_load}{numeric. A vector of general factor loadings from an SL solution.
This needs only be specified if \code{model} is left \code{NULL}.}

\item{s_load}{matrix. A matrix of group factor loadings from an SL solution.
This needs only be specified if \code{model} is left \code{NULL}.}

\item{u2}{numeric. A vector of uniquenesses from an SL solution. This needs
only be specified if \code{model} is left \code{NULL}.}

\item{cormat}{matrix. A correlation matrix to be used when
\code{variance = "correlation"}. If left \code{NULL} and an \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}
output is entered in \code{model}, the correlation matrix is taken from the
output. If left \code{NULL} and a \code{\link[psych:schmid]{psych::schmid()}}
output is entered, the correlation matrix will be found based on the pattern
matrix and Phi from the \code{\link[psych:schmid]{psych::schmid()}} output
using \code{\link[psych:factor.model]{psych::factor.model()}}.
If left \code{NULL} and model is also left \code{NULL}, the correlation matrix
is found based on the pattern matrix and Phi entered. However, if the
correlation matrix is available, \code{cormat} should be specified instead
of \code{Phi} and \code{pattern}.}

\item{pattern}{matrix. Pattern coefficients from an oblique factor solution.
This needs only be specified if \code{model} is left \code{NULL},
\code{variance = "correlation"} and \code{cormat} is also left \code{NULL}.}

\item{Phi}{matrix. Factor intercorrelations from an oblique factor solution.
This needs only be specified if \code{model} is left \code{NULL},
\code{variance = "correlation"} and \code{cormat} is also left \code{NULL}.}

\item{variance}{character. If \code{"correlation"} (default), then total
variances for the whole scale as well as for the subscale composites are
calculated based on the correlation
matrix. If \code{"sums_load"}, then total variances are calculated using the
squared sums of general factor loadings and group factor loadings and
the sum of uniquenesses (see details).}
}
\value{
If found for an SL or \code{lavaan} second-order or bifactor solution
without multiple groups:
A matrix with omegas for the whole scale and for the subscales and (only if
\code{add_ind = TRUE}) with the H index, ECV, and PUC.
\item{tot}{Omega total.}
\item{hier}{Omega hierarchical.}
\item{sub}{Omega subscale.}
\item{H}{H index.}
\item{ECV}{Explained common variance.}
\item{PUC}{Percent of uncontaminated correlations.}

If found for a \code{lavaan} single factor solution without multiple groups:
A (named) vector with omega total and (if \code{add_ind = TRUE}) the H index
for the single factor.

If found for a \code{lavaan} output from a multiple group analysis: A list
containing the output described above for each group.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{OMEGA()} has been superseded by \code{\link[=efa_reliability]{efa_reliability()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.

This function finds omega total, hierarchical, and subscale, as well as additional
model-based indices of interpretive relevance (H index, ECV, PUC)
from a Schmid-Leiman (SL) solution or lavaan single factor, second-order (see below),
or bifactor solution. The SL-based omegas can either be found from a
\code{\link[psych:schmid]{psych::schmid()}}, \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}, or,
in a more flexible way, by leaving
\code{model = NULL} and specifying additional arguments. The \code{type} argument selects
how variables are assigned to group factors, and can reproduce the assignment
\code{\link[psych:omega]{psych::omega()}} makes.
}
\details{
\subsection{What this function does}{

All types of McDonald's omegas (total, hierarchical, and subscale; McDonald,
1978, 1985, 1999) are calculated for
the general factor as well as for the subscales / group factors (see, e.g.,
Gignac, 2014; Rodriguez et al., 2016a, 2016b). Omegas refer to the correlation
between a factor and a unit-weighted composite score and thus the
true score variance in a unit-weighted composite based on the respective
indicators. Omega total is the total true score variance in a composite.
Omega hierarchical is the true score variance in a composite that is attributable
to the general factor, and omega subscale is the true score variance in a
composite attributable to all subscales / group factors (for the whole scale)
or to the specific subscale / group factor (for subscale composites).

Accordingly, on a subscale row the \code{hier} column reports the share of that
subscale's composite variance due to the general factor and the \code{sub} column
the share due to the subscale-specific factor; the latter corresponds to the
omega hierarchical subscale of Rodriguez et al. (2016a, 2016b).

The H index (also construct reliability or replicability index) is the
correlation between an optimally-weighted composite score
and a factor (Hancock & Mueller, 2001; Rodriguez et al., 2016a, 2016b). It, too,
can be calculated for the whole scale / general factor as well as for the
subscales / group factors. Low values indicate that a latent variable is not well
defined by its indicators.

The ECV (Rodriguez et al., 2016a, 2016b) is the ratio of the
variance explained by the general factor and the variance explained by the
general factor and the group factors.

The PUC (Bonifay et al., 2015; Reise et al., 2013, Rodriguez et al., 2016a,
2016b) refers to the proportion
of correlations in the underlying correlation matrix that is not contaminated
by variance of both the general factor and the group factors (i.e., correlations
between indicators from different group factors, which reflect only general
factor variance). The higher the PUC, the more similar a general factor from
a multidimensional model will be to the single factor from a unidimensional
model.
}

\subsection{How to use this function}{

If \code{model} is a \code{lavaan} second-order or bifactor solution,
only the name of the general factor from the lavaan model needs to be specified
additionally with the \code{g_name} argument. It is then determined whether this
general factor is a second-order factor (second-order model with one second-order
factor assumed) or a breadth factor (bifactor model assumed). Please note that
this function only works for second-order models if they contain no more than
one second-order factor. In case of a second-order solution, a
Schmid-Leiman transformation is performed on the first- and second-order loadings
and omega coefficients are obtained from the transformed (orthogonalized) solution
(see \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} for more information on Schmid-Leiman transformation).
There is also the possibility to enter a \code{lavaan} single factor solution.
In this case, \code{g_name} is not needed. Finally, if a solution from a
\code{lavaan} multiple group analysis is entered, the indices are computed for
each group. For \code{lavaan} input the composite variances entering the omegas are
model-implied: they are computed from the fitted loadings and the fitted residual
covariance matrix, and count any freed residual covariance as well as the residual
variances. The coefficients thus coincide with the observed-score versions when the
model fits perfectly. The omegas split a composite's variance into a general part and one part
per group factor, which needs uncorrelated latent variables: fit a bifactor model
with \code{orthogonal = TRUE} (not \code{lavaan}'s default) and leave the covariances between
a second-order model's first-order factors at zero. A fit whose factors correlate is
rejected rather than scored as though they did not.
The type argument is not evaluated if \code{model} is of class
\code{lavaan}.

If \code{model} is of class \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} or
\code{\link[psych:schmid]{psych::schmid()}} only the
\code{type} and, depending on the type (see below), the \code{factor_corres}
arguments need to be specified additionally. If model is of class
\code{\link[psych:schmid]{psych::schmid()}} and \code{variance = "correlation"}
(default), it is
recommended to also provide the original correlation matrix in \code{cormat}
to get more accurate results. Otherwise, the correlation matrix will be found
based on the pattern matrix and Phi from the
\code{\link[psych:schmid]{psych::schmid()}} output
using the \code{\link[psych:factor.model]{psych::factor.model()}} function.

If \code{model = NULL}, the arguments \code{type}, \code{factor_corres}
(depending on the type, see below), \code{var_names}, \code{g_load}, \code{s_load},
and \code{u2} and either \code{cormat} (recommended) or \code{Phi} and
\code{pattern} need to be specified. If \code{Phi} and \code{pattern} are
specified instead of \code{cormat}, the correlation matrix is found using
the \code{\link[psych:factor.model]{psych::factor.model()}} function.

The only difference between \code{type = "EFAtools"} and \code{type = "psych"}
is the determination of variable-to-factor correspondences. \code{type = "psych"}
derives them as \code{\link[psych:omega]{psych::omega()}} does, by
taking the highest
group factor loading for each variable as the relevant group factor loading.
To do this, \code{factor_corres} must be left \code{NULL}.

Both settings score a composite by the true score variance the model attributes
to it, counting every factor its variables load on; they differ only in the
variance that is divided into. \code{variance = "correlation"} uses the composite's
observed variance, giving the observed-score form of omega; \code{"sums_load"} uses its
model-implied variance, which partitions exactly into omega hierarchical plus omega
subscale on the whole-scale row. The two settings agree up to model misfit, and
differ mainly in the whole-scale omega subscale, which counts all group-factor
variance under \code{"sums_load"} but only the assigned subscale composites under
\code{"correlation"}.
}
}
\examples{
\donttest{
## Use with lavaan outputs
if (requireNamespace("lavaan", quietly = TRUE)) {

# Create and fit bifactor model in lavaan (assume all variables have SDs of 1)
mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
        F2 =~ V7 + V8 + V9 + V10 + V11 + V12
        F3 =~ V13 + V14 + V15 + V16 + V17 + V18
        g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 + V11 + V12 +
             V13 + V14 + V15 + V16 + V17 + V18'
fit_bi <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                      sample.nobs = 500, estimator = "ml", orthogonal = TRUE)

# Compute omegas and additional indices for bifactor solution
OMEGA(fit_bi, g_name = "g")

# Compute only omegas
OMEGA(fit_bi, g_name = "g", add_ind = FALSE)

# Create and fit second-order model in lavaan (assume all variables have SDs of 1)
mod <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
        F2 =~ V7 + V8 + V9 + V10 + V11 + V12
        F3 =~ V13 + V14 + V15 + V16 + V17 + V18
        g =~ F1 + F2 + F3'
fit_ho <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                      sample.nobs = 500, estimator = "ml")

# Compute omegas and additional indices for second-order solution
OMEGA(fit_ho, g_name = "g")
}
}

## Use with an output from the SL function, with type EFAtools
efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
sl_mod <- efa_schmid_leiman(efa_mod, estimator = "PAF")

# Indicator-to-factor correspondences from a salience threshold (here: .20):
factor_corres_1 <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2

OMEGA(sl_mod, type = "EFAtools", factor_corres = factor_corres_1)

## Use with an output from the psych::schmid function, with type psych for
## OMEGA
schmid_mod <- psych::schmid(test_models$baseline$cormat, nfactors = 3,
                            n.obs = 500, fm = "pa", rotate = "Promax")
# Find correlation matrix from phi and pattern matrix from psych::schmid output
OMEGA(schmid_mod, type = "psych")
# Use specified correlation matrix
OMEGA(schmid_mod, type = "psych", cormat = test_models$baseline$cormat)

## Manually specify components (useful if omegas should be computed for a SL
## or bifactor solution found with another program)
## As an example, we extract the elements from an SL output here. This gives
## the same results as in the second example above.

factor_corres <- matrix(c(rep(0, 12), rep(1, 6), rep(0, 6), rep(1, 6),
                        rep(0, 6), rep(1, 6), rep(0, 12)), ncol = 3,
                        byrow = FALSE)

OMEGA(model = NULL, type = "EFAtools", var_names = rownames(sl_mod$sl),
      g_load = sl_mod$sl[, "g"], s_load = sl_mod$sl[, c("F1", "F2", "F3")],
      u2 = sl_mod$sl[, "u2"], cormat = test_models$baseline$cormat,
      factor_corres = factor_corres)

}
\seealso{
\code{\link[=efa_reliability]{efa_reliability()}} for the same coefficients in a tidy, long-format
result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_retention.R
\name{format.efa_retention}
\alias{format.efa_retention}
\title{Format method for efa_retention objects}
\usage{
\method{format}{efa_retention}(x, ...)
}
\arguments{
\item{x}{an object of class efa_retention, returned by a factor-retention
criterion (e.g. \code{\link[=efa_ekc]{efa_ekc()}} or \code{\link[=efa_hull]{efa_hull()}}).}

\item{...}{not used.}
}
\value{
A character vector with the report lines (styled to the active
console theme; plain when colours are disabled).
}
\description{
Format method for efa_retention objects
}
\examples{
writeLines(format(efa_ekc(test_models$baseline$cormat, N = 500)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.PARALLEL.R
\name{print.PARALLEL}
\alias{print.PARALLEL}
\title{Print function for PARALLEL objects}
\usage{
\method{print}{PARALLEL}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class PARALLEL. Output from \link{PARALLEL} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for PARALLEL objects
}
\examples{
\donttest{
# example without real data
PARALLEL(N = 500, n_vars = 10)

# example with correlation matrix and "ML" estimation
PARALLEL(test_models$case_11b$cormat, N = 500, method = "ML")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.extract_list_object}
\alias{.extract_list_object}
\title{Extract a list object by its name}
\usage{
.extract_list_object(alist, object)
}
\arguments{
\item{alist}{A list of sub-lists, typically a list of \eqn{m} objects of class
\code{"efa"}, where \eqn{m} is the number of imputations passed to
\code{\link[=efa_mi]{efa_mi()}}.}

\item{object}{String of length 1. The name of the object to extract e.g.
\code{"h2"} or \code{"vars_accounted"}.}
}
\value{
A list of length \eqn{m}, with each element containing the extracted
\code{object} for the \eqn{k}th element (\eqn{k = 1,..., m}).
}
\description{
Consider a list of named sub-lists. This function extracts, for each sub-list,
the sub-list element that is specified by the user. This function is useful
for extracting results from \code{\link[=efa_fit]{efa_fit()}} for each imputation in
\code{\link[=efa_mi]{efa_mi()}}.
}
\author{
Andreas Soteriades
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_reliability.R
\name{efa_reliability}
\alias{efa_reliability}
\title{Reliability and common-variance coefficients for a factor solution}
\source{
McDonald, R. P. (1978). Generalizability in factorable domains: Domain
validity and generalizability. Educational and Psychological Measurement, 38,
75-79.

McDonald, R. P. (1985). Factor analysis and related methods. Hillsdale,
NJ: Erlbaum.

McDonald, R. P. (1999). Test theory: A unified treatment. Mahwah,
NJ: Erlbaum.

Cronbach, L. J. (1951). Coefficient alpha and the internal structure of
tests. Psychometrika, 16, 297-334.

Gignac, G. E. (2014). On the inappropriateness of using items to
calculate total scale score reliability via coefficient alpha for
multidimensional scales. European Journal of Psychological Assessment, 30,
130-139.

Flora, D. B. (2020). Your coefficient alpha is probably wrong, but which
coefficient omega is right? A tutorial on using R to obtain better reliability
estimates. Advances in Methods and Practices in Psychological Science, 3,
484-501.

Green, S. B., & Yang, Y. (2009). Reliability of summed item scores using
structural equation modeling: An alternative to coefficient alpha. Psychometrika,
74, 155-167.

Zinbarg, R. E., Revelle, W., Yovel, I., & Li, W. (2005). Cronbach's
alpha, Revelle's beta, and McDonald's omega H: Their relations with each
other and two alternative conceptualizations of reliability. Psychometrika,
70, 123-133.

Zinbarg, R. E., Yovel, I., Revelle, W., & McDonald, R. P. (2006).
Estimating generalizability to a latent variable common to all of a scale's
indicators: A comparison of estimators for omega H. Applied Psychological
Measurement, 30, 121-144.

Hancock, G. R., & Mueller, R. O. (2001). Rethinking construct
reliability within latent variable systems. In R. Cudeck, S. du Toit, & D.
Sörbom (Eds.), Structural equation modeling: Present and future - A
Festschrift in honor of Karl Jöreskog (pp. 195-216). Lincolnwood, IL:
Scientific Software International.

Bonifay, W. E., Reise, S. P., Scheines, R., & Meijer, R. R. (2015). When
are multidimensional data unidimensional enough for structural equation
modeling? An evaluation of the DETECT multidimensionality index. Structural
Equation Modeling, 22, 504-516.

Reise, S. P., Scheines, R., Widaman, K. F., & Haviland, M. G. (2013).
Multidimensionality and structural coefficient bias in structural equation
modeling: A bifactor perspective. Educational and Psychological Measurement,
73, 5-26.

Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016a). Applying bifactor
statistical indices in the evaluation of psychological measures. Journal of
Personality Assessment, 98, 223-237.

Rodriguez, A., Reise, S. P., & Haviland, M. G. (2016b). Evaluating
bifactor models: Calculating and interpreting statistical indices.
Psychological Methods, 21, 137-150.
}
\usage{
efa_reliability(
  model = NULL,
  coefficients = NULL,
  g_name = "g",
  group_names = NULL,
  factor_map = NULL,
  variance = c("correlation", "sums_load"),
  var_names = NULL,
  fac_names = NULL,
  g_load = NULL,
  s_load = NULL,
  u2 = NULL,
  cormat = NULL,
  pattern = NULL,
  Phi = NULL
)
}
\arguments{
\item{model}{an \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}, \code{schmid} (\code{\link[psych:schmid]{psych::schmid()}}), \code{\link[=efa_fit]{efa_fit()}}
(oblique), or \code{lavaan} object; a raw bifactor loading matrix (general factor
first), the loading table of an \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} solution, or the pattern
matrix of an oblique solution together with its \code{Phi}; or \code{NULL} to
supply the components manually via \code{g_load}, \code{s_load}, \code{u2}, and \code{var_names}.}

\item{coefficients}{character. An optional subset of the coefficients to
return, any of \code{"omega_total"}, \code{"omega_hierarchical"}, \code{"omega_subscale"},
\code{"alpha"}, \code{"H"}, \code{"ECV"}, and \code{"PUC"}. Default \code{NULL} returns all of them.}

\item{g_name}{character. The name of the general factor in the \code{lavaan}
solution. Only needed for a \code{lavaan} second-order or bifactor fit. A fit in
which every variable loads on a single factor has no general factor, and
reads none. No other input is affected by it. Default is \code{"g"}.}

\item{group_names}{character. An optional vector of group names for a \code{lavaan}
multiple-group fit. Its length must match the number of groups. Not used for
any other input -- including a single-group \code{lavaan} fit, since every one of
those is scored as a single ungrouped solution with no group label.}

\item{factor_map}{matrix. A logical or 0/1 matrix indicating which variable
belongs to which group factor, with the same dimensions as the group
loading matrix (cross-loadings are allowed). Match its columns to the
group factors by position -- a map given in a different factor order than
the solution still runs, but produces meaningless subscale coefficients.
The function warns if a mapped item loads weakly on its assigned factor
and more strongly on another one. If \code{NULL} (default), each variable is
assigned to the group factor on which it loads most strongly. Not used
for \code{lavaan} input.}

\item{variance}{character. The total-variance denominator for the
coefficients. \code{"correlation"} (default) takes each composite's variance
from the correlation matrix, giving the observed-score omega. \code{"sums_load"}
uses the model-implied composite variance from the loadings and the
uniquenesses instead; it needs no correlation matrix, so it is the way to
score a bare loading matrix or manual components given without one.
\code{lavaan} input fixes the convention: its composite variances are always
model-implied, and include any freed residual covariance as well as the
residual variances. Neither convention changes the metric the coefficients
are on -- with polychoric or tetrachoric correlations, or an \code{ordered}
\code{lavaan} fit, both describe the latent-response composite rather than the
ordinal sum score (see Details).}

\item{var_names}{character. Subtest names in the row order of the loadings.
Only needed when \code{model} is \code{NULL}.}

\item{fac_names}{character. An optional vector of group-factor names in the
column order of the loadings. Taken from the input if \code{NULL}. A
single-factor solution has no group factors; \code{fac_names} then labels its
one factor instead. If neither \code{fac_names} nor the solution names that
factor, it is labelled \code{"F1"}. Not used for \code{lavaan} input, whose factor
labels come from the model syntax.}

\item{g_load}{numeric. General-factor loadings. Only needed when \code{model} is
\code{NULL}.}

\item{s_load}{matrix. Group-factor loadings. Only needed when \code{model} is
\code{NULL}.}

\item{u2}{numeric. Uniquenesses. Only needed when \code{model} is \code{NULL}, or to
override the communality-based default for a loading matrix. Under
\code{variance = "correlation"}, the coefficients follow from the loadings and
the correlation matrix alone, so \code{u2} only enters the check for an improper
solution. Under \code{"sums_load"}, it is part of every composite's variance.}

\item{cormat}{matrix or data.frame. A correlation matrix used when
\code{variance = "correlation"}. It must hold the same variables as the
solution: named variables in a different order are reordered to match; a
different set of variables, or a different number of them, is an error.
The matrix must use the solution's own variable names -- for manually
supplied components, that means the row names of \code{s_load}, not
\code{var_names}, which only labels the output. Without row names, the
variables are matched by position, so supply the matrix in the row order
of the loadings. If \code{NULL}, it is taken from the input object, or
reconstructed from \code{pattern} and \code{Phi} where possible. Supply the matrix
on the same metric as the loadings: a polychoric or tetrachoric matrix
keeps the coefficients on the latent-response metric, while a Pearson
matrix given with loadings fitted to a polychoric one mixes the two (see
Details).}

\item{pattern}{matrix. Pattern coefficients from a separate oblique solution, used
with \code{Phi} to reconstruct a correlation matrix when \code{model} is \code{NULL}. Supply it for
a Schmid-Leiman input, whose \code{s_load} holds the orthogonalized group loadings rather
than the oblique ones. It is an alternative to \code{cormat}, not a supplement: giving
both is an error.}

\item{Phi}{matrix. Factor intercorrelations. \code{NULL} (default) means
uncorrelated group factors, as a Schmid-Leiman or bifactor solution has.
It cannot be combined with a general factor: supplied together with a
non-zero \code{g_load}, or with the loading table of an
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} solution, it is an error.

Supply it together with \code{s_load} (manually supplied components, with
\code{g_load} zero throughout), or with a loading matrix of two or more
factors in \code{model}, to score the input as a correlated-factors solution
instead.

Without \code{Phi}, a loading matrix in \code{model} is read as a bifactor
solution (general factor first). The one exception is a matrix that
still carries the loading class \code{\link[=efa_fit]{efa_fit()}} gives its output -- a class
that subsetting a matrix or reordering its rows drops. Such a matrix is
rejected instead; supply \code{Phi} to score it as a correlated-factors
solution.

With \code{Phi}, a matrix in \code{model} that carries no loading class is read as
a correlated-factors solution, and a warning names this reading (drop
\code{Phi} to read it as a bifactor matrix instead).

Given together with \code{pattern} instead, see \code{pattern} above. With a
fitted solution already in \code{model}, \code{Phi} is ignored, with a warning,
since that solution carries its own factor intercorrelations.}
}
\value{
An object of class \code{efa_reliability}: a long-format data frame with
one row per computed coefficient, with columns
\item{coefficient}{the coefficient name (e.g. \code{"omega_total"}).}
\item{level}{\code{"general"} for the general-factor row, \code{"total"} for the
whole-scale row of a correlated-factors solution, and \code{"group"} for
every other row. A single-factor solution has only the general-factor
row.}
\item{factor}{the factor label: \code{"g"} for the general factor of a
solution with group factors. A single-factor solution's one factor
takes its own name, or \code{"F1"} if neither the input nor \code{fac_names}
names it. A correlated-factors solution has no general factor, so its
first row is labelled \code{"total"} instead -- it describes the composite
of every variable, not a factor of the model.}
\item{group}{the group label, or \code{NA} for a single ungrouped solution.}
\item{value}{the coefficient value.}
Structurally undefined cells (for example, ECV and PUC on a group factor)
are omitted. The object also carries a \code{settings} attribute (the
total-variance convention used, and whether the solution has a general
factor) and a \code{kind} attribute tagging each coefficient as a reliability
coefficient or a common-variance index. It has a
\code{\link[=print.efa_reliability]{print.efa_reliability()}} method.
}
\description{
Computes model-based reliability coefficients for a factor solution:
McDonald's omega (total, hierarchical, and subscale), standardized Cronbach's
alpha, and the H index. For a bifactor solution, it also computes two
common-variance indices, ECV and PUC, for the general factor. The result is a
tidy, long-format table with one row per coefficient.
}
\details{
The function reads many kinds of input: a Schmid-Leiman solution
(\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} or \code{\link[psych:schmid]{psych::schmid()}}), an oblique \code{\link[=efa_fit]{efa_fit()}}
(correlated-factors) solution, a \code{lavaan} fit (single-factor,
correlated-factors, second-order, or bifactor), a raw bifactor loading matrix,
an oblique pattern matrix given with its factor intercorrelations, or manually
supplied components.
\subsection{Coefficients}{

The reliability coefficients are McDonald's omegas (McDonald, 1978, 1985,
1999; Zinbarg et al., 2005, 2006, for omega hierarchical specifically),
standardized Cronbach's alpha (Cronbach, 1951), and the H index (construct
replicability; Hancock & Mueller, 2001).

The omegas give the share of true score variance in a unit-weighted
composite. Omega total is the share due to all factors together. Omega
hierarchical is the share due to the general factor only. Omega subscale is
the share due to the group factors: for the whole scale, or for one specific
group factor in a subscale composite.

Alpha is the standardized coefficient, computed from the correlation matrix
of the items. Where no such matrix is available -- for \code{lavaan} input, and
for components supplied without one -- alpha is computed from the
model-implied correlation matrix instead, so it then reflects the fitted
model rather than the raw data.

The H index is the reliability of an optimally weighted composite. A low
value means the factor is not well defined by its indicators.

All of these coefficients describe the raw sum of the variables as fitted,
without reverse-coding. If some items are keyed in the opposite direction --
for example, a reverse-worded item that was not reverse-scored -- they lower
that sum's true-score variance. The coefficients then look poor even though
the model fits well. The function warns when it detects this. Reverse-code
such items before fitting the solution (Flora, 2020).

The sum these coefficients describe is not always the sum of the raw answers.
Polychoric and tetrachoric correlations describe the continuous latent
responses assumed to underlie ordinal answers; a \code{lavaan} fit that declares
its indicators \code{ordered} does the same. In that case the loadings, the
uniquenesses, the correlation matrix, and the coefficients are all on that
latent-response metric. They give the reliability of the unit-weighted sum of
the latent responses, which is not observed -- not the reliability of the
ordinal sum score the user actually computes from the answers. The two can
differ substantially, especially where the answers use few categories or are
strongly skewed. Green and Yang (2009) give an omega for the ordinal sum
score itself, computed from the fitted model and its thresholds; this
package does not compute it. Pearson correlations raise no such distinction,
because their metric is the answers as scored.

Omega total is lower when a solution reproduces the observed correlations
poorly, because residual covariance does not count as true score.
\code{\link[psych:omega]{psych::omega()}} computes the whole-scale omega total differently:
residually, from the observed total-score variance. The two agree when the
model reproduces the correlations exactly, and diverge otherwise.

The three coefficients are not generally additive. Omega total need not
equal omega hierarchical plus omega subscale, except on the whole-scale row
under \code{variance = "sums_load"}.

A single-factor solution is scored as such on every input route, and
returns omega total, alpha, and the H index. Alpha assumes essentially
tau-equivalent items, an assumption nested within a one-factor model -- so a
single factor is the case where reporting alpha is defensible, not merely
possible.

The other coefficients are omitted because a single factor does not define
them: omega subscale is the variance due to the group factors, and there
are none; omega hierarchical would equal omega total, since the one factor
accounts for all common variance; and ECV and PUC would both be 1 by
construction, which reflects the number of factors in the model rather than
evidence of unidimensionality.

Each coefficient answers a different question:
\itemize{
\item Omega hierarchical: can the total score be read as a measure of one
construct?
\item Omega subscale: does a subscale score add anything beyond the general
factor?
\item The H index: is a factor well defined by its indicators?
\item ECV together with PUC: is a unidimensional model defensible?
}

Alpha assumes essentially tau-equivalent items. Factor analysis instead
yields congeneric solutions, for which alpha is only a lower bound. For a
multidimensional scale, alpha is rarely the coefficient to report (Gignac,
2014).

Composite reliability and average variance extracted (AVE) are not among
the coefficients this function computes. Use \code{semTools::compRelSEM()} and
\code{semTools::AVE()} to compute them from a \code{lavaan} fit.

The common-variance indices ECV and PUC (Bonifay et al., 2015; Reise et al.,
2013; Rodriguez et al., 2016a, 2016b) describe the general factor, so they
are reported for the general factor only. ECV is the share of the common
variance explained by the general factor. PUC is the proportion of
correlations that reflect general-factor variance alone -- correlations
between indicators of different group factors. The higher the PUC, the more
the general factor resembles the single factor of a unidimensional model.
}

\subsection{Input}{

\code{efa_reliability()} reads several kinds of input, illustrated in the
examples below.
\itemize{
\item \strong{An oblique \code{\link[=efa_fit]{efa_fit()}} solution} is scored as the correlated-factors
model it is. It has no general factor, so it omits the bifactor indices
(omega hierarchical, ECV, and PUC). Its whole-scale row is labelled
\code{"total"} rather than \code{"g"}.
\item \strong{An \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} solution} -- or a \code{schmid} object from
\code{\link[psych:schmid]{psych::schmid()}}, or a raw bifactor loading matrix with the general
factor in its first column -- is scored as a bifactor solution, with the
whole-scale row labelled \code{"g"}. Indicator-to-factor correspondences come
from \code{factor_map} if supplied (see \code{factor_map} below). Without one, an
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} or \code{\link[psych:schmid]{psych::schmid()}} solution is mapped by each
variable's strongest group loading. A raw bifactor matrix instead
defaults to its exact zero pattern -- supply \code{factor_map} explicitly if
that matrix has no exact zeros (e.g. an estimated rather than a target
matrix).
\item \strong{A \code{lavaan} fit} -- single-factor, correlated-factors, second-order, or
bifactor -- is scored per its structure. The structure is detected
automatically, not taken from \code{g_name}. The general-factor coefficients
need the latent factors to be uncorrelated: fit a bifactor model with
\code{orthogonal = TRUE} (\code{lavaan}'s default leaves the factors correlated),
and leave a second-order model's first-order factor covariances at zero.
A fit whose factors correlate is rejected rather than scored as though
they did not. \code{variance} is not used for \code{lavaan} input: its composite
variances are always model-implied, computed separately per group for a
multiple-group fit.
\item \strong{A pattern matrix with its \code{Phi}} -- for example, the loadings and
\code{Phi} of an oblique \code{\link[=efa_fit]{efa_fit()}} solution -- is scored as the
correlated-factors solution it represents. Uniquenesses are derived from
the loadings under \code{Phi}, unless \code{u2} is given. Supply the pattern
matrix, not the structure matrix -- the structure matrix would be read
as a pattern too, giving the wrong result. Unlike a fitted solution, this
pair carries no correlation matrix. \code{variance = "correlation"} then
needs one, supplied in \code{cormat}, and errors without it. \code{variance = "sums_load"} needs none.
}

A one-factor solution -- a one-column loading matrix, a single-factor
\code{lavaan} fit, a single-factor \code{\link[=efa_fit]{efa_fit()}} solution, or single-factor
components -- is scored the same way regardless of route (see Coefficients
above for what it returns). Its row is labelled with the factor's own name,
with \code{fac_names} if supplied, or with \code{"F1"} if neither names it; it is
never labelled \code{"g"}.

Manually supplied components (\code{g_load}, \code{s_load}, \code{u2}, \code{var_names}, \code{Phi})
follow the same reading rules as the matrix routes above. Do not pass a
correlation matrix as \code{s_load} (or as \code{model}, for the matrix routes) --
supply it as \code{cormat} instead.
}
}
\examples{
## From an oblique EFA (correlated-factors) solution. With no factor_map, each
## item is auto-assigned to its highest-loading factor.
efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
efa_reliability(efa_mod)

## From a Schmid-Leiman solution, with an explicit indicator-to-factor map.
sl_mod <- efa_schmid_leiman(efa_mod, estimator = "PAF")
fc <- sl_mod$sl[, c("F1", "F2", "F3")] >= .2
efa_reliability(sl_mod, factor_map = fc)

## Request a subset of the coefficients only.
efa_reliability(sl_mod, factor_map = fc,
                coefficients = c("omega_total", "alpha"))

## From an oblique pattern matrix and its factor intercorrelations. This is
## the same correlated-factors solution, and gives the same coefficients.
efa_reliability(efa_mod$rot_loadings, Phi = efa_mod$Phi,
                cormat = test_models$baseline$cormat)

## From lavaan fits: a bifactor solution, and a correlated-factors one.
\donttest{
if (requireNamespace("lavaan", quietly = TRUE)) {
mod_cf <- 'F1 =~ V1 + V2 + V3 + V4 + V5 + V6
           F2 =~ V7 + V8 + V9 + V10 + V11 + V12
           F3 =~ V13 + V14 + V15 + V16 + V17 + V18'
mod <- paste(mod_cf, 'g =~ V1 + V2 + V3 + V4 + V5 + V6 + V7 + V8 + V9 + V10 +
                           V11 + V12 + V13 + V14 + V15 + V16 + V17 + V18',
             sep = "\n")
fit <- lavaan::cfa(mod, sample.cov = test_models$baseline$cormat,
                   sample.nobs = 500, estimator = "ml", orthogonal = TRUE)
print(efa_reliability(fit, g_name = "g"))

## No general factor: omega hierarchical, ECV, and PUC are omitted.
fit_cf <- lavaan::cfa(mod_cf, sample.cov = test_models$baseline$cormat,
                      sample.nobs = 500, estimator = "ml")
efa_reliability(fit_cf)
}
}

}
\seealso{
\code{\link[=efa_fit]{efa_fit()}} for the solution these are computed from, and \code{\link[=OMEGA]{OMEGA()}}, the
superseded function that returns these same coefficients in a wide, per-factor layout.

Other reliability coefficients:
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}},
\code{\link[=print.efa_reliability]{print.efa_reliability()}}
}
\concept{reliability coefficients}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.EKC.R
\name{plot.EKC}
\alias{plot.EKC}
\title{Plot EKC object}
\usage{
\method{plot}{EKC}(x, ...)
}
\arguments{
\item{x}{a list of class EKC. An output from the \link{EKC} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \link{EKC} function
}
\examples{
EKC_base <- EKC(test_models$baseline$cormat, N = 500)
plot(EKC_base)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{print.efa_control}
\alias{print.efa_control}
\alias{print.efa_estimate_control}
\alias{format.efa_estimate_control}
\alias{print.efa_rotate_control}
\alias{format.efa_rotate_control}
\title{Print and format a control object}
\usage{
\method{print}{efa_estimate_control}(x, ...)

\method{format}{efa_estimate_control}(x, ...)

\method{print}{efa_rotate_control}(x, ...)

\method{format}{efa_rotate_control}(x, ...)
}
\arguments{
\item{x}{A control object from \code{\link[=estimate_control]{estimate_control()}} or \code{\link[=rotate_control]{rotate_control()}}.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows the chosen \code{type} and each tuning knob, with an unset (\code{NA})
preset-driven knob marked as resolved from the \code{type} preset. \code{format()}
assembles the same report and returns it as a character vector; \code{print()} is
\code{cat(format(x), sep = "\\n")}. The lines follow the active console theme, so they
are plain when colours are disabled.
}
\examples{
est <- estimate_control(type = "SPSS")
est
writeLines(format(est))

}
\seealso{
\code{\link[=estimate_control]{estimate_control()}}, \code{\link[=rotate_control]{rotate_control()}}

Other Control functions:
\code{\link[=estimate_control]{estimate_control()}}
}
\concept{Control functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.OMEGA.R
\name{print.OMEGA}
\alias{print.OMEGA}
\alias{format.OMEGA}
\title{Print and format an OMEGA object}
\usage{
\method{print}{OMEGA}(x, digits = 3, ...)

\method{format}{OMEGA}(x, digits = 3, ...)
}
\arguments{
\item{x}{An object of class \code{OMEGA} (output from \code{\link[=OMEGA]{OMEGA()}}).}

\item{digits}{Integer. The number of decimal places the coefficients are
rounded to (passed to \code{\link[base:round]{base::round()}}). Default is 3.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows the omega coefficients computed by \code{\link[=OMEGA]{OMEGA()}}: omega total (and,
for multi-factor solutions, omega hierarchical, omega subscale, the H index,
the explained common variance, and the percent of uncontaminated correlations)
for the general factor and the group factors, for a single group or for each
group. \code{format()} assembles the same report and returns it as a character
vector; \code{print()} is \code{cat(format(x), sep = "\\n")}. The lines follow the active
console theme, so they are plain when colours are disabled (for example when
captured into a file or stripped with \code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
sl_mod <- efa_schmid_leiman(efa_mod, estimator = "PAF")

om <- OMEGA(sl_mod, type = "EFAtools",
            factor_corres = sl_mod$sl[, c("F1", "F2", "F3")] >= .2)
om

# format() returns the same lines as a character vector:
writeLines(format(om))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_ekc.R
\name{efa_ekc}
\alias{efa_ekc}
\title{Empirical Kaiser criterion}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200

Braeken, J., & van Assen, M. A. (2017). An empirical Kaiser criterion.
Psychological Methods, 22, 450 – 466. https://doi.org/10.1037/met0000074

Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
determining the number of components to retain. Psychological Bulletin, 99,
432–442. https://doi.org/10.1037/0033-2909.99.3.432
}
\usage{
efa_ekc(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  type = lifecycle::deprecated()
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Only needed if x is a correlation
matrix. Must be larger than the number of variables.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator). Default is \code{"pearson"}. Note that the EKC reference
values rest on the Marchenko-Pastur law for the eigenvalues of a sample
correlation matrix of independent variables, which assumes the sampling
behaviour of product-moment correlations; with \code{"poly"} / \code{"tetra"} (and, to a
lesser degree, the rank-based methods) the reference series is therefore an
approximation.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Accepted and ignored. It selected
between two ways to compute the reference values. The \code{"AM2019"} reference values
do not depend on the observed eigenvalues, so they do not apply the empirical
correction that defines the criterion, and they are no longer computed.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). Its main fields are:
\item{n_factors}{A numeric vector of length one, named \code{"BvA2017"}, with the
suggested number of factors. The factors up to the first observed eigenvalue
that fails to exceed its reference value are retained. The "all-exceed"
convention of parallel analysis (\code{\link[=efa_parallel]{efa_parallel()}}), which retains all \emph{J} factors
when no such crossing is found, cannot be reached here: the reference values are
never below 1, while the eigenvalues of a correlation matrix sum to \emph{J} and are
sorted downwards, so the last of them is never above 1.}
\item{results}{A list with one record, holding the eigenvalues, the reference
eigenvalues, and the retained solution used for printing and plotting.}
\item{settings}{A list with the settings used.}
}
\description{
The empirical Kaiser criterion incorporates random sampling variations of the
eigenvalues from the Kaiser-Guttman criterion (\code{\link[=efa_kgc]{efa_kgc()}}; see Auerswald &
Moshagen, 2019; Braeken & van Assen, 2017). The implementation follows Braeken
and van Assen (2017).
}
\details{
The Kaiser-Guttman criterion was defined with the intend that a factor
should only be extracted if it explains at least as much variance as a single
factor (see \code{\link[=efa_kgc]{efa_kgc()}}). However, this only applies to population-level
correlation matrices. Due to sampling variation, the KGC strongly overestimates
the number of factors to retrieve (e.g., Zwick & Velicer, 1986). To account
for this and to introduce a factor retention method that performs well with
small number of indicators and correlated factors (cases where the performance
of parallel analysis, see \code{\link[=efa_parallel]{efa_parallel()}}, is known to deteriorate)
Braeken and van Assen (2017) introduced the empirical Kaiser criterion in
which a series of reference eigenvalues is created as a function of the
variables-to-sample-size ratio and the observed eigenvalues.

Braeken and van Assen (2017) showed that "(a) EKC performs about as well as
parallel analysis for data arising from the null, 1-factor, or orthogonal
factors model; and (b) clearly outperforms parallel analysis for the specific
case of oblique factors, particularly whenever factor intercorrelation is
moderate to high and the number of variables per factor is small, which is
characteristic of many applications these days" (p.463-464).
}
\examples{
efa_ekc(test_models$baseline$cormat, N = 500)
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_mi.R
\name{efa_mi}
\alias{efa_mi}
\title{Exploratory factor analysis on multiple data imputations}
\usage{
efa_mi(
  data_list,
  p = 0.05,
  target_method = c("first_target", "consensus"),
  align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
  fit_pool_method = c("D2"),
  consensus_args = list(),
  procrustes_args = list(),
  rmsea_ci_level = 0.9,
  rmsr_upper = lifecycle::deprecated(),
  ...
)
}
\arguments{
\item{data_list}{A list of length \eqn{m}, where \eqn{m} is the number of
imputations. Each list element is a data frame or matrix of raw data, or a
correlation matrix. See argument \code{x} in \code{\link[=efa_fit]{efa_fit()}}. A \code{mids} object from
\pkg{mice} must be converted first, with \code{mice::complete(x, "all")}.}

\item{p}{Numeric in \eqn{(0, 1)}. One minus the confidence level for the
pooled confidence intervals, whichever \code{se} method produced them
(\code{"information"}, \code{"np-boot"}, or \code{"sandwich"}). For example, \code{p = .05}
gives 95\% intervals.}

\item{target_method}{Character. How rotated solutions are aligned across imputations
before pooling: \code{"first_target"} (the default) aligns every imputation to the first
imputation's rotated solution, while \code{"consensus"} refines a centroid target by
Generalized Procrustes Analysis, started from the medoid imputation so that the
pooled rotated solution does not depend on the order of \code{data_list} (orthogonal
rotations only). See \emph{Aligning solutions across imputations} in Details.}

\item{align_unrotated}{Character. How unrotated loadings are aligned before pooling:
\code{"signed_tucker_congruence"} (the default; sign/permutation via Tucker congruence,
anchored on the medoid imputation and returned in the extraction's canonical
gauge), \code{"procrustes"} (orthogonal Procrustes to the first imputation), or
\code{"none"}. See \emph{Aligning solutions across imputations} in Details.}

\item{fit_pool_method}{Character. Only \code{"D2"} is implemented for pooling
chi-square-type fit. If no chi-square is available, only residual-based
fit and descriptive quantities are returned. See \emph{Pooling the model chi-square and
fit indices} in Details.}

\item{consensus_args}{List of additional arguments controlling the
GPA-consensus iteration when \code{target_method = "consensus"}. Recognised tuning
parameters include the convergence tolerances \code{tol} and \code{loss_tol}, the
iteration bounds \code{min_iter} and \code{max_iter}, the target-update damping \code{alpha},
the multi-start controls \code{multi_start} and \code{starts}, and \code{start}, which
overrides the medoid imputation the iteration is otherwise started from.}

\item{procrustes_args}{List of \code{\link[=efa_procrustes]{efa_procrustes()}} algorithm controls for
fixed-target alignment, for example \code{oblique_maxit} or \code{oblique_random_starts}.
The loadings \code{A}, the alignment \code{Target}, the \code{rotation} family, and the
cross-product \code{S} are derived from the imputations and cannot be set here.}

\item{rmsea_ci_level}{Numeric. Confidence level for the RMSEA CI.}

\item{rmsr_upper}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Deprecated and ignored.
\code{efa_mi()} now always computes RMSR the same way, from the unique
off-diagonal residuals; SRMR is reported alongside it. Supplying it to
\code{\link[=efa_mi]{efa_mi()}} signals a deprecation warning; the superseded \code{\link[=EFA_POOLED]{EFA_POOLED()}}
accepts it silently.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}} (e.g. \code{estimator}, \code{rotation}, \code{se},
\code{n_factors}, \code{N}). These select the estimator, rotation, standard-error method, and
fit indices used for every imputation; see \code{\link[=efa_fit]{efa_fit()}} for the available options, their
properties, and which combinations are valid. Two of them shape the pooled object
rather than a single fit: \code{seed} sets the random state once for the whole
\code{efa_mi()} call -- every component bootstrap and every random-start rotation draws
from it, so a seeded call is reproducible as a whole, and the caller's random stream
is restored afterwards -- and \code{b_boot} sets the number of bootstrap replicates drawn
per imputation under \code{se = "np-boot"}, which is what the pooled within-imputation
variances are estimated from and is recorded in \code{settings$b_boot}. The
\code{\link[=estimate_control]{estimate_control()}} and
\code{\link[=rotate_control]{rotate_control()}} objects are accepted through \code{...} as well, although they are not
declared formals: pass them as \verb{estimate_control =} / \verb{rotate_control =} exactly as you
would to \code{\link[=efa_fit]{efa_fit()}}.}
}
\value{
A list of class \code{c("efa_mi", "EFA_POOLED", "efa", "EFA")} containing
pooled estimates, residuals, fit indices, the individual fits, and MI
diagnostics. The trailing legacy classes keep \code{inherits(x, "EFA_POOLED")} and
the single-fit EFA accessors and S3 dispatch working. In
addition to the slots inherited from \code{\link[=efa_fit]{efa_fit()}} (including \code{SE}, \code{CI}, and,
on the bootstrap path, \code{replicates}), the object carries:
\describe{
\item{SE, CI}{Pooled standard errors and confidence intervals, named as
\code{\link[=efa_fit]{efa_fit()}} names them: where a pooled communality standard error and interval
are produced, they are \code{SE$communalities} and \code{CI$communalities} on every route.
The Rubin routes (\code{se = "information"}, \code{se = "np-boot"}) additionally return
them under the compatibility alias \code{h2}, which holds the same values. The
analytic route builds the communality family only when a rotation was requested;
an unrotated analytic pool reports \code{uniquenesses} instead.}
\item{fit_indices}{The pooled fit indices. Every route reports \code{chi}, \code{df},
\code{p_chi}, \code{CAF}, \code{CFI}, \code{TLI}, \code{RMSEA}, \code{RMSEA_LB}, \code{RMSEA_UB}, \code{AIC}, \code{BIC},
\code{ECVI}, \code{RMSR}, \code{SRMR}, \code{chi_null}, \code{df_null}, \code{p_null}, and \code{pool_method}
under those names and in that order. \code{pool_method} records the rule the model
chi-square was pooled with (\code{"D2"}); it is \code{NA} on the \code{se = "sandwich"} (MI2S)
path, which fits once on the pooled inputs and reports that fit's own scaled
statistic rather than pooling several. On that path a few extra scaled-statistic
fields are appended after the common block, for advanced diagnostics.}
\item{standardized_residuals}{The pooled residuals divided by their pooled
bootstrap standard errors, with a zero diagonal. Returned on the
\code{se = "np-boot"} path only, the one route that pools a residual standard error.}
\item{MI}{Multiple-imputation diagnostics for each pooled parameter family.
On the bootstrap path: \code{unrot_loadings}, \code{communalities}, \code{residuals}, optionally
\code{rot_loadings}, \code{Phi}, \code{Structure}, and \code{fit_indices_descriptive}, plus
integer vectors \code{bootstrap_source_failures} (replicates the component \code{\link[=efa_fit]{efa_fit()}}
could not fit), \code{bootstrap_rotation_failures} (replicates whose Procrustes
alignment to the target was invalid), and \code{bootstrap_rotation_valid} (those
that entered the pool, \code{B - source - rotation} failures). Both paths use the
plain Rubin (1987) df. On the analytic path (\code{se = "information"}):
\code{unrot_loadings} and \code{uniquenesses}, plus, when a rotation was requested,
\code{rot_loadings}, \code{communalities}, and (oblique) \code{Phi} and \code{Structure}. The
communality family is keyed by its canonical name here, without the \code{SE}/\code{CI}
alias, so each family is counted once in the printed FMI/RIV summary. Each per-family
entry is a list with \code{RIV} (relative increase in variance), \code{FMI} (the
fraction of missing information, reported as Rubin's asymptotic
\eqn{\lambda = RIV / (1 + RIV)}), and \code{df}; the
rotated families on the analytic path additionally carry a \code{method} string
recording the gauge alignment used (\code{"gauge_invariant"} for communalities and
\code{"signed_permutation_approx"} for rotated loadings and, for oblique rotations,
factor correlations and structure coefficients). \code{fit_indices_descriptive}, on
the bootstrap path, pools every fit index the bootstrap replicates carry, so the
structural constants among them (\code{df}, \code{df_null}) appear with a standard error
of 0. The RMSEA confidence bounds are not among them: the replicate fits run
without confidence intervals, so no per-replicate value exists to pool.}
\item{mi_fit}{On the \code{se = "sandwich"} (MI2S) path only: the single \code{\link[=efa_fit]{efa_fit()}}
fit on the pooled correlation matrix \eqn{\bar r} and pooled asymptotic
covariance \eqn{\tilde\Gamma}. Its \code{orig_R} is \eqn{\bar r} and its \code{Gamma}
is \eqn{\tilde\Gamma}; the pooled \code{SE} and \code{CI} are taken from it, as are the
pooled \code{fit_indices} (put into the common order above and extended with
\code{pool_method}, while \code{mi_fit} keeps \code{\link[=efa_fit]{efa_fit()}}'s own layout). \code{MI} is \code{NULL}
on this path because the imputation uncertainty is carried by
\eqn{\tilde\Gamma} rather than by per-parameter Rubin pooling.}
\item{mi_diagnostics}{Diagnostics for the pooled model fit, \code{NULL} on the
\code{se = "sandwich"} (MI2S) path, where there is one fit and no D2 pool. \code{m} is the
number of imputations that entered the pool. \code{D2_F}, \code{D2_df1}, \code{D2_df2},
\code{D2_chi_asymptotic}, \code{ARIV} and \code{FMI} describe the D2 pool of the model
chi-square (the average relative increase in variance and the fraction of
missing information it implies), and \code{chi_bar_naive} is the plain mean of the
per-imputation statistics for comparison; the \verb{*_null} entries are the same
quantities for the independence baseline. \code{D2_F} is the rule's raw statistic and
is reported unfloored, so it is negative whenever the between-imputation
variability of the component statistics exceeds the pooled discrepancy -- a
diagnostic of the pool rather than a fit statistic. The reported fit is not
affected: the pooled chi-square is floored at zero and its p-value is 1 in that
case. \code{chi_cfi} and \code{chi_null_cfi} are the
pooled model and baseline \strong{chi-squares} on the common \eqn{N - 1}
noncentrality scale. \code{chi_cfi} is the statistic the reported RMSEA is formed
from; the pair also gives a reference CFI formed the conventional way,
\code{1 - (chi_cfi - df) / (chi_null_cfi - df_null)} (and analogously for TLI) --
a different quantity from the reported CFI/TLI, which average the
per-imputation indices.}
\item{mi_admissibility}{Admissibility and convergence of the component fits, kept on
the pooled object so a saved solution carries the record independently of \code{fits}: \code{m}
(the number of imputations that entered the pool), \code{heywood_imputations} (the indices
of the fits with at least one Heywood case -- an improper solution where a variable's
communality is at or above 1, or its uniqueness is fixed at the estimation
boundary), \code{n_heywood_items} (the number of flagged variables per imputation),
\code{nonconverged} (the indices whose extraction
reported a non-zero convergence code), and \code{iter} (the iterations each extraction
used). Averaging aligned solutions pulls boundary communalities back inside the
admissible range, so a pooled matrix with no Heywood case can still rest on component
fits that had them; \code{summary()} reports the pooled count together with these.}
\item{fits}{The list of \eqn{m} component \code{\link[=efa_fit]{efa_fit()}} fits, in the order of
\code{data_list}, kept for per-imputation diagnostics. On the MI2S path these are the
per-imputation fits whose inputs were pooled, not the pooled fit itself (which
is \code{mi_fit}).}
\item{alignment}{Metadata from aligning the rotated solutions, \code{NULL} when no
rotation was requested or on the MI2S path (one fit, one gauge). Under
\code{target_method = "first_target"}: the \code{method} used, the \code{target} it aligned to,
the per-imputation \code{target_rotations}, the indices of any
\code{point_rotation_failures}, and whether every inner alignment \code{converged}. Under
\code{target_method = "consensus"} it is the full \code{\link[=efa_procrustes]{efa_procrustes()}}-based GPA record:
the converged \code{target}, the \code{aligned_loadings} and \code{aligned_phi}, the iteration
\code{history}, convergence flags, and the multi-start summary.}
\item{settings}{The component fits' \code{\link[=efa_fit]{efa_fit()}} settings with the pooling
settings added: \code{pooled} (always \code{TRUE}), \code{pooled_N} and \code{N} (the mean N across
imputations), \code{n_imputations}, \code{component_se} (the \code{se} the component fits
used), \code{target_method}, \code{align_unrotated}, \code{fit_pool_method}, \code{p}, \code{ci} and
\code{rmsea_ci_level}. \code{se} records what was actually pooled, so it
is \code{"none"} when pooled standard errors could not be produced although the
component fits computed them (\code{component_se} keeps the request).}
}
}
\description{
Fits \code{\link[=efa_fit]{efa_fit()}} to each of several imputed datasets, aligns the
factor solutions to a common factor space, and pools the resulting estimates
and selected fit quantities across imputations.
}
\details{
\code{efa_mi()} is the multiple-imputation route to handling missing data:
several imputed datasets are each fitted with \code{\link[=efa_fit]{efa_fit()}} and the solutions pooled.
A single-fit alternative is full-information maximum likelihood, available
directly in \code{\link[=efa_fit]{efa_fit()}} as \code{cor_method = "fiml"}, which EM-estimates a two-stage
correlation from one raw dataset with missing values. Both feed the same
correlation-scale EFA core and differ only in how the missingness is handled;
FIML is intentionally not routed through \code{efa_mi()}, which is a multi-fit
pooler by construction.

Both routes assume the values are missing at random (MAR). Which one to prefer
is largely practical: FIML is a single, efficient fit and is the simpler
default when the analysis model is the whole story, whereas multiple imputation
is more flexible when the imputation model should draw on auxiliary variables
not in the factor model, or when the same imputations feed several downstream
analyses.
\subsection{Standard-error pooling routes}{

The pooling pathway is selected automatically from the \code{se} method recorded on
the component \code{\link[=efa_fit]{efa_fit()}} fits, which must be identical across imputations:
\itemize{
\item \code{se = "none"}: no standard errors are pooled.
\item \code{se = "information"}: the per-imputation expected-information standard errors
are pooled with Rubin's (1987) rules (Wald intervals).
\item \code{se = "sandwich"}: the two-stage pooled-inputs (MI2S) approach fits a single
model on the Rubin-pooled correlation matrix and asymptotic covariance.
\item \code{se = "np-boot"}: the non-parametric bootstrap replicates are re-aligned to
the multiple-imputation target and Rubin-pooled.
}

On the information and np-boot routes, if pooled standard errors cannot be
produced (for example an unreliable analytic covariance or too few bootstrap
replicates) the pool falls back to point-estimate-only pooling and downgrades
\code{settings$se} to \code{"none"}. The MI2S route is the exception: its single fit
fuses the point estimates and standard errors through the pooled asymptotic
covariance, so a structural failure aborts directly rather than falling back.
}

\subsection{Aligning solutions across imputations}{

The same \code{\link[=efa_fit]{efa_fit()}} model is fitted to each imputed dataset and the solutions are
put into a common factor space before averaging. For oblique solutions the
factor intercorrelations are aligned together with the loadings so the model
stays internally consistent.

\code{target_method} controls how rotated solutions are aligned. \code{"first_target"}
(the default) aligns every imputation to the first imputation's rotated
solution by one Procrustes rotation each. \code{"consensus"} instead refines a
centroid target by Generalized Procrustes Analysis (Gower 1975; van Ginkel &
Kroonenberg 2014; Lorenzo-Seva & Van Ginkel 2016), starting from the \emph{medoid}
imputation's rotated solution -- the one closest in aligned squared distance to
all the others. \code{"consensus"} is supported for orthogonal rotations only.

Factor loadings are only unique up to rotation; "gauge" here means which
particular rotation, or orientation, a solution is expressed in. The two
methods differ in the rotational gauge the pooled solution ends up in, and so
in how it responds to the order of \code{data_list}. The GPA iteration moves its
target toward the centroid but keeps the gauge of the solution it started from,
so starting it at the medoid -- a property of the set, not of the list order --
makes the pooled rotated solution invariant to that order, as the pooled
unrotated solution already is. \code{"first_target"} anchors on the first imputation
by construction, so an atypical first imputation fixes the orientation for
every other one. Permuting \code{data_list} therefore moves the pooled pattern --
by a few hundredths of a loading unit when imputations are similar, more when
they disagree. For oblique rotations, the factor correlations move with it.
Where the two anchors coincide -- and, more generally,
where the imputations agree -- the two methods give effectively the same pooled
estimate and \code{"consensus"} is simply the more expensive; where they do not, the
pooled patterns differ by the rotation between the two gauges. Passing \code{start}
through \code{consensus_args} overrides the medoid anchor and makes the consensus
order dependent again.

\code{align_unrotated} controls how unrotated loadings are aligned before pooling:
\code{"signed_tucker_congruence"} (the default) matches them up to factor reordering
and sign changes, \code{"procrustes"} aligns them to the first imputation by
orthogonal Procrustes rotation, and \code{"none"} averages them as returned by
\code{\link[=efa_fit]{efa_fit()}}.

The default anchors the matching on the \emph{medoid} imputation (defined above)
rather than on whichever imputation happens to come first, so the pooled
\emph{unrotated} solution does not depend on the order of \code{data_list}. The rotated
solution is aligned separately, against a reference chosen by \code{target_method},
and still depends on that reference.

The pooled unrotated matrix is then returned in the same gauge as a single
\code{\link[=efa_fit]{efa_fit()}} fit. Its identifying constraint differs by extraction method -- a
principal-axis extraction and a maximum-likelihood extraction fix the
rotation differently (Anderson & Rubin 1956; Lawley & Maxwell 1971) -- and is
read off each component fit automatically, so the pooled matrix can be
compared element-by-element with an \code{\link[=efa_fit]{efa_fit()}} solution. A solution that
meets neither constraint -- an improper one, say -- is left as aligned. The
correction is a common orthogonal rotation, so communalities, the total
variance accounted for, the model-implied correlation matrix, the residuals,
and RMSR are unchanged; only the split of variance across factors moves.
\code{"procrustes"} and \code{"none"} keep their first-imputation anchor and are
returned as aligned.
}

\subsection{Pooling point estimates}{

Point estimates are pooled by arithmetic averaging after alignment. For oblique
rotations the structure matrix is recomputed from the pooled pattern matrix and
pooled factor correlations, \eqn{Structure = \Lambda \Phi}, and communalities
are the diagonal of the reproduced correlation matrix,
\eqn{diag(\Lambda \Phi \Lambda')} for oblique rotations and
\eqn{diag(\Lambda \Lambda')} otherwise. Residuals are not averaged across
imputations; they are the pooled observed correlation matrix minus the
model-implied correlation of the pooled solution, so RMSR/SRMR are based on
these pooled residuals. Both are returned, though the print and summary methods
show SRMR only.
}

\subsection{Pooling the model chi-square and fit indices}{

The model chi-square and the indices derived from it (ECVI and the descriptive
AIC/BIC) are pooled with the D2 rule (Li, Meng, Raghunathan &
Rubin, 1991), not arithmetically averaged. RMSEA is pooled by the same rule but
from a second D2 pool of the per-imputation discrepancies taken on the
uncorrected \eqn{N - 1} scale. The printed RMSEA therefore does not reconcile
by hand with the printed chi-square; the statistic it is formed from is
\code{chi_cfi} in \code{mi_diagnostics}. Because D2 shrinks the pooled chi-square in
proportion to the between-imputation variability, the pooled RMSEA can fall
below the mean of the per-imputation RMSEAs; read it together with the
per-imputation fit. The incremental indices CFI (Bentler, 1990) and TLI
(Tucker & Lewis, 1973) are instead the average of the per-imputation indices,
which keeps them consistent with the component fits and avoids out-of-range
values; the separately pooled model and baseline chi-squares those indices
would be formed from remain available in \code{mi_diagnostics}.
AIC and BIC, if returned, are chi-square-derived descriptive quantities and are
not likelihood-based MI information criteria. They are reported only where the
component fits report them: whenever a component withholds them -- any
\code{cor_method = "fiml"} fit, and any fit whose chi-square is a scaled statistic,
such as \code{se = "sandwich"} -- the pooled AIC, BIC, and ECVI are \code{NA} too, matching
what \code{\link[=efa_fit]{efa_fit()}} returns for a single such fit. On the sandwich/MI2S route the
chi-square is the single fit's scaled statistic rather than a D2 pool.
}

\subsection{Bootstrap pooling (np-boot)}{

If each component \code{\link[=efa_fit]{efa_fit()}} call was run with \code{se = "np-boot"}, pooled bootstrap SEs
and Wald-type MI confidence intervals are computed for loadings, communalities,
residuals, and, when applicable, factor correlations and structure
coefficients. The unrotated bootstrap replicates are re-aligned to the final MI
target before the within-imputation covariance is estimated, and Rubin pooling
is applied with \eqn{T = Ubar + (1 + 1/m) B}. The confidence level of the pooled
intervals is set by \code{p}, not by the component \code{\link[=efa_fit]{efa_fit()}} calls' \code{ci}.
}

\subsection{Analytic pooling (information)}{

With \code{se = "information"}, the analytic unrotated-loading and uniqueness SEs
returned by each fit are pooled element-wise with Rubin's rules
(\eqn{T = Ubar + (1 + 1/m) B}), with Wald intervals on the plain Rubin (1987)
degrees of freedom (the analytic loadings are asymptotically normal, so the
Barnard-Rubin (1999) adjustment reduces to this form). NA
propagation is fail-closed: if any imputation is NA at an element, all pooled
outputs for that element are NA. When a rotation was requested, the rotated
loadings, communalities, and (for oblique rotations) factor correlations and
structure coefficients are pooled as well; residual SE pooling is available only
on the bootstrap path. Under \code{align_unrotated = "procrustes"} the full unrotated
covariance \code{vcov_unrot_loadings} (populated by \code{se = "information"}) is
propagated through the alignment, so it must be present and reliable on every
fit. The default alignment also mixes loading columns, through the common
canonical-gauge rotation, and so propagates the same covariance; where a fit
does not carry it, the unrotated standard errors are returned as \code{NA} rather
than aborting, and the remaining families still pool.

A rotated-loading standard error is conditional on the rotation criterion
used (see References). For both orthogonal and oblique rotations the
within-imputation variance is therefore each fit's own
criterion-aware delta-method rotated SE (the quantity \code{efa_fit()} returns), reused
after a signed-permutation alignment to the MI target, and the
between-imputation variance is the sample variance of the aligned rotated
loadings. This is a deliberate approximation -- each SE is conditional on its
own fit's rotation optimum rather than on a common gauge -- and is flagged by
\verb{MI$<param>$method = "signed_permutation_approx"}. Communalities are
rotation-invariant and pool element-wise. For a fully gauge-consistent rotated
uncertainty, cross-check with \code{se = "np-boot"}.
}

\subsection{Two-stage pooling (sandwich / MI2S)}{

With \code{se = "sandwich"} (robust SEs from a polychoric/tetrachoric or
continuous-Pearson asymptotic covariance), pooling follows the two-stage,
pooled-inputs approach (Chung & Cai 2019; Sriutaisuk, Liu, Chung, Kim & Gu
2025): the correlation matrix and the asymptotic covariance of its off-diagonal
entries are Rubin-pooled across imputations, and a single \code{EFA} model is
fitted to the pooled correlation with the pooled covariance, \eqn{\tilde\Gamma},
as the robust meat (its diagonal as the weights for
\code{estimator = "DWLS"}). Because there is only one fit and one rotational gauge, this
route bypasses the per-imputation alignment: \code{target_method} and
\code{align_unrotated} do not apply. The fitted object carries native scaled-shifted
chi-square statistics and sandwich SEs that already reflect the
multiple-imputation uncertainty, so the chi-square is not D2-pooled and the
likelihood-ratio-based AIC/BIC/ECVI are \code{NA}; it is returned in the \code{mi_fit}
slot, with the per-imputation \code{fits} retained for diagnostics. The pooled fit
uses the same \code{\link[=estimate_control]{estimate_control()}} and \code{\link[=rotate_control]{rotate_control()}} tuning (including
any rotation-engine extras) as the per-imputation fits. At least 20
imputations are recommended for the scaled-shifted statistic, and more (around
100) at higher rates of missingness (Sriutaisuk et al. 2025). The
polychoric/tetrachoric (ordinal) case is the primary, best-evaluated target; the
continuous-Pearson case uses the same recipe but is less benchmarked.
}
}
\section{Conditions}{

Errors and warnings raised by \code{efa_mi()} are classed, with an \code{efa_pooled_}
prefix (\code{efa_consensus_} for the consensus target) -- except the dots validation
shared with \code{\link[=efa_fit]{efa_fit()}}, which signals \code{efa_flat_knob_in_dots} or \code{efa_renamed_arg}
-- so they can be caught programmatically. The message shown explains what went
wrong and, where relevant, how to fix it.
}

\examples{

# create a list of three datasets, mimicking a list you would obtain from
# e.g. mice.
dat_list <- lapply(1:3, function(x) GRiPS_raw[sample(1:nrow(GRiPS_raw), replace = TRUE),])
mod <- efa_mi(dat_list, n_factors = 1, estimator = "ML")
mod

\donttest{
# add computation of standard errors and CIs
mod <- efa_mi(dat_list, n_factors = 1, estimator = "ML", se = "np-boot")
mod
}
}
\references{
Anderson, T. W., & Rubin, H. (1956). Statistical inference in factor analysis.
In \emph{Proceedings of the Third Berkeley Symposium on Mathematical Statistics and
Probability} (Vol. 5, pp. 111-150). University of California Press.

Archer, C. O., & Jennrich, R. I. (1973). Standard errors for rotated factor
loadings. \emph{Psychometrika}, 38(4), 581-592.

Barnard, J., & Rubin, D. B. (1999). Small-sample degrees of freedom with
multiple imputation. \emph{Biometrika}, 86(4), 948-955.

Bentler, P. M. (1990). Comparative fit indexes in structural models.
\emph{Psychological Bulletin}, 107(2), 238-246.

Chung, S., & Cai, L. (2019). Alternative multiple imputation inference for
categorical structural equation modeling. \emph{Multivariate Behavioral
Research}, 54(3), 323-337.

Gower, J. C. (1975). Generalized Procrustes analysis. \emph{Psychometrika},
40(1), 33-51.

Jennrich, R. I. (1973). Standard errors for obliquely rotated factor
loadings. \emph{Psychometrika}, 38(4), 593-604.

Jennrich, R. I. (1974). Simplified formulae for standard errors in
maximum-likelihood factor analysis. \emph{British Journal of Mathematical and
Statistical Psychology}, 27(1), 122-131.

Lawley, D. N., & Maxwell, A. E. (1971). \emph{Factor analysis as a statistical
method} (2nd ed.). Butterworths.

Li, K. H., Meng, X.-L., Raghunathan, T. E., & Rubin, D. B. (1991).
Significance levels from repeated p-values with multiply-imputed data.
\emph{Statistica Sinica}, 1(1), 65-92.

Lorenzo-Seva, U., & Van Ginkel, J. R. (2016). Multiple imputation of missing
values in exploratory factor analysis of multidimensional scales. \emph{Anales de
Psicologia}, 32(2), 596-608.

Rubin, D. B. (1987). \emph{Multiple imputation for nonresponse in surveys}. Wiley.

Schoenemann, P. H. (1966). A generalized solution of the orthogonal
Procrustes problem. \emph{Psychometrika}, 31(1), 1-10.

Sriutaisuk, S., Liu, Y., Chung, S., Kim, H., & Gu, F. (2025). Evaluating
imputation-based fit statistics in structural equation modeling with ordinal
data: The MI2S approach. \emph{Educational and Psychological Measurement}, 85(1),
82-113.

Tucker, L. R., & Lewis, C. (1973). A reliability coefficient for maximum
likelihood factor analysis. \emph{Psychometrika}, 38(1), 1-10.

van Ginkel, J. R., & Kroonenberg, P. M. (2014). Using generalized Procrustes
analysis for multiple imputation in principal component analysis. \emph{Journal of
Classification}, 31(2), 242-269.

Zhang, G., & Preacher, K. J. (2015). Factor rotation and standard errors in
exploratory factor analysis. \emph{Journal of Educational and Behavioral
Statistics}, 40(6), 579-603.

Zhang, G., Preacher, K. J., & Jennrich, R. I. (2012). The infinitesimal
jackknife with exploratory factor analysis. \emph{Psychometrika}, 77(4), 634-648.
}
\seealso{
Other factor analysis:
\code{\link[=efa_average]{efa_average()}},
\code{\link[=efa_fit]{efa_fit()}},
\code{\link[=efa_group]{efa_group()}},
\code{\link[=plot.efa_group]{plot.efa_group()}},
\code{\link[=print.efa_group]{print.efa_group()}}
}
\author{
Andreas Soteriades, Markus Steiner
}
\concept{factor analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.SLLOADINGS.R
\name{print.SLLOADINGS}
\alias{print.SLLOADINGS}
\title{Print SLLOADINGS object}
\usage{
\method{print}{SLLOADINGS}(x, cutoff = 0.2, digits = 3, ...)
}
\arguments{
\item{x}{class SLLOADINGS matrix.}

\item{cutoff}{numeric. The number above which to print loadings in bold
(default is .2).}

\item{digits}{numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
to round the loadings to (default is 3).}

\item{...}{additional arguments passed to print}
}
\description{
Print SLLOADINGS object
}
\examples{
EFA_mod <- EFA(test_models$baseline$cormat, N = 500, n_factors = 3,
               type = "EFAtools", method = "PAF", rotation = "promax")
SL(EFA_mod, type = "EFAtools", method = "PAF")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.parallel_sim}
\alias{.parallel_sim}
\title{Parallel analysis on simulated data.}
\usage{
.parallel_sim(n_datasets, n_vars, N, eigen_type, maxit = 10000L)
}
\arguments{
\item{n_datasets}{numeric. Number of datasets with dimensions (N, n_vars) to simulate.}

\item{n_vars}{numeric. Number of variables / indicators in dataset.}

\item{N}{numeric. Number of cases / observations in dataset.}

\item{eigen_type}{numeric. Whether PCA (eigen_type = 1; i.e., leaving diagonal of correlation matrix at 1), SMC (eigen_type = 2; i.e., setting diagonal of correlation matrix to SMCs), or both from the same simulated datasets (eigen_type = 3), in which case the returned matrix holds the PCA eigenvalues in the first n_vars columns and the SMC eigenvalues in the next n_vars.}

\item{maxit}{numeric. Maximum iterations to perform after which to abort.}
}
\description{
Function called from within efa_parallel() so usually no call to this is needed by the user.
Provides a C++ implementation of the efa_parallel() simulation procedure
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{estimate_control}
\alias{estimate_control}
\alias{rotate_control}
\title{Control objects for estimation and rotation settings}
\usage{
estimate_control(
  type = c("EFAtools", "psych", "SPSS", "none"),
  init_comm = NA,
  criterion = NA,
  criterion_type = NA,
  max_iter = NA,
  abs_eigen = NA,
  start_method = "psych",
  fiml_max_iter = 500,
  fiml_tol = 1e-05
)

rotate_control(
  type = c("EFAtools", "psych", "SPSS", "none"),
  normalize = TRUE,
  precision = 1e-05,
  order_type = NA,
  varimax_type = NA,
  p_type = NA,
  k = NA,
  random_starts = 100,
  ...
)
}
\arguments{
\item{type}{character. One of \code{"EFAtools"} (default), \code{"psych"}, \code{"SPSS"}, or
\code{"none"}. Selects the preset that fills the \code{NA}-defaulted knobs below when the
control is used to fit a model.}

\item{init_comm}{character. Method for the initial communalities in principal
axis factoring: \code{"smc"} (squared multiple correlations), \code{"mac"} (maximum
absolute correlations), or \code{"unity"}. \code{NA} (default) resolves from \code{type}.}

\item{criterion}{numeric. The convergence criterion for principal axis
factoring: iteration stops once the change in communalities falls below it. A
single number greater than 0 and smaller than 1; \code{NA} (default) resolves from \code{type}.}

\item{criterion_type}{character. The convergence criterion type for principal
axis factoring: \code{"max_individual"} (the largest change in any communality, as
in SPSS) or \code{"sum"} (the change in the summed communalities, as in
\code{\link[psych:fa]{psych::fa()}}). \code{NA} (default) resolves from \code{type}.}

\item{max_iter}{numeric. The maximum number of principal-axis-factoring
iterations before the procedure is halted with a warning. A single whole
number of at least 1; \code{NA} (default) resolves from \code{type}.}

\item{abs_eigen}{logical. Which algorithm the principal-axis-factoring
iterations use: \code{FALSE} computes the loadings from the eigenvalues (as in
\code{\link[psych:fa]{psych::fa()}}); \code{TRUE} uses the absolute eigenvalues (as in SPSS). \code{NA}
(default) resolves from \code{type}.}

\item{start_method}{character. Starting values for the maximum-likelihood
optimiser: \code{"psych"} (default, the \code{\link[psych:fa]{psych::fa()}} starts) or \code{"factanal"} (the
\code{\link[stats:factanal]{stats::factanal()}} starts); abbreviations are matched. Not governed by \code{type}.
Only maximum likelihood uses it, so \code{NA} leaves it unset and is rejected only by
a fit that is actually run with \code{estimator = "ML"}.}

\item{fiml_max_iter}{numeric. The maximum number of EM iterations used to estimate the
two-stage full-information maximum-likelihood moments from raw data with missing
values (\code{cor_method = "fiml"}); the last iterate is returned, with a warning, if the
cap is reached. A single whole number of at least 1; default \code{500}. Not governed by
\code{type}, and unused by every other correlation method. The EM converges linearly and
needs more iterations the larger the fraction of missing information, so raise it when
a fit reports that the moments did not converge.}

\item{fiml_tol}{numeric. The convergence tolerance of that EM: iteration stops once the
largest change in the standardized moments (the standardized means, log-variances, and
correlations) falls below it, so it does not depend on the variables' measurement
scale. A single number greater than 0 and smaller than 1 (at or above 1 the criterion is
met immediately and the starting moments would be returned as converged); default \code{1e-5}.
Not governed by \code{type}.}

\item{normalize}{logical. If \code{TRUE} (default), a Kaiser normalization is
performed before the rotation. The one knob that is always on unless you turn
it off with \code{FALSE}.}

\item{precision}{numeric. The convergence tolerance of the rotation procedure. A
single number greater than 0 and at most 1; default \code{1e-5}. Each rotation stage
monitors its own quantity, so the same number is not the same tolerance everywhere:
\code{varimax_type = "kaiser"} stops on the \emph{absolute} change in the varimax simplicity
criterion, which is an average over variables (and so does not scale with how many
there are) but rises with the number of factors, roughly toward
\code{1 - 1 / n_factors}, so a fixed value is a relatively weaker tolerance the more
factors are extracted; \code{varimax_type = "svd"} stops on the \emph{relative} change in the
singular values (as in \code{\link[stats:varimax]{stats::varimax()}}); and the criterion rotations fitted by
gradient projection stop when the \emph{projected-gradient norm} falls below it. Promax
inherits whichever of the two varimax tests its \code{varimax_type} selects, because it
rotates a varimax base.}

\item{order_type}{character. How the factors are ordered: \code{"eigen"} (by
descending sum of squared loadings, as in \code{\link[psych:fa]{psych::fa()}}) or \code{"ss_factors"} (by
descending unweighted sum of squared loadings). \code{NA} (default) resolves from
\code{type}.}

\item{varimax_type}{character. The varimax variant used (for the varimax and
promax rotations): \code{"svd"} (as in \code{\link[stats:varimax]{stats::varimax()}}) or \code{"kaiser"} (the SPSS
/ Kaiser (1958) procedure). \code{NA} (default) resolves from \code{type}.}

\item{p_type}{character. How the promax target matrix is computed: \code{"unnorm"}
(the unnormalized target of Hendrickson & White (1964), also used by psych and
stats) or \code{"norm"} (the normalized target used by SPSS). \code{NA} (default)
resolves from \code{type}.}

\item{k}{numeric. The promax power (for the target matrix) or the number of
near-zero loadings for simplimax. A single number greater than 0; \code{NA}
(default) leaves it to the fit (the \code{type}-dependent promax value, or
\code{nrow(loadings)} for simplimax). Simplimax counts loadings, so a fit using it
additionally requires a whole number no larger than the number of loadings in
the solution; promax's power has no such restriction.}

\item{random_starts}{numeric. The number of random starts used by the
criterion-based rotations to guard against local minima. A single whole number
of at least 0, where \code{0} runs the rotation from its warm start only; default
\code{100}. The default suffices for the smooth criteria; \code{simplimax} remains
materially start-dependent at it, so raise it there (see the \emph{Rotations} section
of \code{\link[=efa_fit]{efa_fit()}}).}

\item{...}{Additional arguments forwarded to the rotation engine. Only the names
a rotation engine can consume are accepted: \code{maxit} (a whole number of at least 0
bounding a \emph{single} gradient-projection optimization -- the multi-start search runs
several of them and each is bounded separately, so it is not a budget for the run as a
whole; varimax and promax have no such stage and take only \code{precision}), and the
criterion parameters \code{gam} (oblimin; \code{gam = 0} is the
recommended default, and larger values increasingly reward correlated factors and
can drive the solution toward factor collapse, so inspect \code{Phi} before interpreting
a fit with \code{gam > 0}) and \code{delta} (geomin; a positive number, default \code{0.01});
anything else is rejected as a
misspelling. They are stored in \code{extra_args} and passed on to the rotation engine
when the control is used to fit a model; an extra a given fit's rotation does not
consume is ignored by that fit, so one control can serve fits with different
rotations. An estimation knob (which belongs in \code{\link[=estimate_control]{estimate_control()}}) or one of the
former spellings \code{P_type} and \code{randomStarts} is likewise rejected here, because the
fit would silently drop it.}
}
\value{
\code{estimate_control()} returns a list of class \code{efa_estimate_control} with
the components \code{type}, \code{init_comm}, \code{criterion}, \code{criterion_type}, \code{max_iter},
\code{abs_eigen}, \code{start_method}, \code{fiml_max_iter}, and \code{fiml_tol}.
\code{rotate_control()} returns a list of class
\code{efa_rotate_control} with the components \code{type}, \code{normalize}, \code{precision},
\code{order_type}, \code{varimax_type}, \code{p_type}, \code{k}, \code{random_starts}, and \code{extra_args}
(a named list of any additional arguments forwarded to the rotation engine).
}
\description{
\code{estimate_control()} and \code{rotate_control()} collect the estimation and rotation
\emph{tuning} arguments of a factor analysis into two small, validated objects. They
are a declarative surface over the same settings resolved internally by the
package's estimation and rotation engines, so that a fit's many tuning knobs can
be prepared, inspected, and reused as a single value instead of being passed one
by one.
}
\details{
Each argument that governs a \code{type} preset defaults to \code{NA}, meaning "leave this
knob to the preset". Setting \code{type} to one of \code{"EFAtools"}, \code{"psych"}, or
\code{"SPSS"} fills those knobs from the corresponding preset when the fit is run;
setting \code{type = "none"} requires the relevant knobs to be supplied explicitly.
The control object only records the chosen \code{type} and the knobs you set: the
preset is resolved (and any "argument set alongside \code{type}" warning issued) when
the object is used to fit a model, exactly as it is today, because which preset
applies depends on the estimator and rotation.
}
\examples{
# Estimation knobs taken entirely from a preset:
estimate_control(type = "SPSS")

# A preset with one knob pinned to a non-preset value:
estimate_control(type = "EFAtools", max_iter = 500)

# Every knob supplied explicitly (type = "none"):
estimate_control(type = "none", init_comm = "smc", criterion = 1e-3,
                 criterion_type = "sum", max_iter = 300, abs_eigen = TRUE)


# Rotation knobs taken from a preset:
rotate_control(type = "psych")

# A criterion-specific extra argument, forwarded to the rotation engine:
rotate_control(type = "EFAtools", k = 3, gam = 0.5)

}
\seealso{
\code{\link[=efa_fit]{efa_fit()}}, which takes both controls; \code{\link[=efa_retain]{efa_retain()}}, the retention
criteria, and \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}, which take an \code{estimate_control} for the
fits they run.

Other Control functions:
\code{\link{print.efa_control}}
}
\concept{Control functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_reliability.R
\name{print.efa_reliability}
\alias{print.efa_reliability}
\alias{format.efa_reliability}
\title{Print and format a reliability object}
\usage{
\method{print}{efa_reliability}(x, digits = 3, ...)

\method{format}{efa_reliability}(x, digits = 3, ...)
}
\arguments{
\item{x}{An object of class \code{efa_reliability}.}

\item{digits}{Integer. The number of decimal places the coefficients are
rounded to. Default is 3.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows the reliability coefficients for the general factor and the
group factors, for a single group or for each group: the reliability
coefficients (omega total, omega hierarchical, and omega subscale, standardized
Cronbach's alpha, and the H index) and the common-variance indices (the explained
common variance, ECV, and the percent of uncontaminated correlations, PUC).
\code{format()} assembles the same report and returns it as a character vector;
\code{print()} is \code{cat(format(x), sep = "\\n")}. The lines follow the active console
theme, so they are plain when colours are disabled (for example when captured
into a file or stripped with \code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
efa_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
rel <- efa_reliability(efa_mod)
rel

# format() returns the same lines as a character vector:
writeLines(format(rel))

}
\seealso{
Other reliability coefficients:
\code{\link[=efa_reliability]{efa_reliability()}},
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}
}
\concept{reliability coefficients}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.KMO.R
\name{print.KMO}
\alias{print.KMO}
\title{Print KMO object}
\usage{
\method{print}{KMO}(x, ...)
}
\arguments{
\item{x}{list of class KMO (output from the \link{KMO} function)}

\item{...}{additional arguments passed to print}
}
\description{
Print KMO object
}
\examples{
KMO_base <- KMO(test_models$baseline$cormat)
KMO_base

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.average_matrices}
\alias{.average_matrices}
\title{Average a list of matrices elementwise}
\usage{
.average_matrices(x)
}
\arguments{
\item{x}{List of conformable matrices.}
}
\value{
A matrix with the same dimensions as the inputs.
}
\description{
Average a list of matrices elementwise
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.EKC.R
\name{print.EKC}
\alias{print.EKC}
\title{Print function for EKC objects}
\usage{
\method{print}{EKC}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class EKC. Output from \code{\link{EKC}} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for EKC objects
}
\examples{
EKC_base <- EKC(test_models$baseline$cormat, N = 500)
EKC_base

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.MAP.R
\name{print.MAP}
\alias{print.MAP}
\title{Print function for MAP objects}
\usage{
\method{print}{MAP}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class MAP. Output from \code{\link{MAP}} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for MAP objects
}
\examples{
MAP_base <- MAP(test_models$baseline$cormat)
MAP_base

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RiskDimensions_doc.R
\docType{data}
\name{RiskDimensions}
\alias{RiskDimensions}
\title{RiskDimensions}
\format{
A list of 2 with elements "cormat" (9 x 9 matrix of bivariate
correlations) and "N" (scalar). The correlation matrix contains the following
risk dimensions:
\describe{
\item{Voluntariness}{(numeric) - Voluntariness of exposure to the risk.}
\item{Immediacy}{(numeric) - Immediacy of the risk's effect.}
\item{Known to exposed}{(numeric) - How well the risk is known to those exposed to it.}
\item{Known to science}{(numeric) - How well the risk is known to science.}
\item{Controllability}{(numeric) - Controllability of the risk.}
\item{Newness}{(numeric) - Newness of the risk.}
\item{Chronic}{(numeric) - Whether the risk is chronic rather than catastrophic.}
\item{Common}{(numeric) - Whether the risk is common rather than dreaded.}
\item{Severity of consequences}{(numeric) - Severity of the consequences.}
}
}
\source{
Fischhoff, B, Slovic, P, Lichtenstein, S, Read, S, and Combs, B. (1978). How safe is safe enough? A psychometric study of attitudes towards technological risks and benefits. Policy Sciences, 9, 127-152. doi: 10.1007/BF00143739
}
\usage{
RiskDimensions
}
\description{
A list containing the bivariate correlations (cormat)
of the 9 dimensions on which participants in Fischhoff et al. (1978) rated
different activities and technologies as well as the sample size (N). This was
then analyzed together with ratings of the risks and benefits of these
activities and technologies.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_power.R
\name{print.efa_power}
\alias{print.efa_power}
\alias{format.efa_power}
\title{Print and format an efa_power object}
\usage{
\method{print}{efa_power}(x, digits = 3, ...)

\method{format}{efa_power}(x, digits = 3, ...)
}
\arguments{
\item{x}{An object of class \code{efa_power} (output from \code{\link[=efa_power]{efa_power()}}).}

\item{digits}{Integer. The number of decimal places the reported values are
rounded to. Default is 3.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} turns an \code{\link[=efa_power]{efa_power()}} result into a short report, and \code{format()}
builds the same report as a character vector (\code{print()} is
\code{cat(format(x), sep = "\\n")}).
}
\details{
For an RMSEA-mode object, the report has a header naming the test, the null and
alternative hypotheses with the significance level and degrees of freedom, the
headline result (the power at the sample size, or the required sample size for
the target power), and the critical value and noncentrality parameters.

For a simulation-mode object, the report instead has the population and design,
the retention hit-rate per criterion, the structure-recovery rate, and the
convergence and Heywood-case rate.

The lines follow the active console theme, so they print as plain text when
colours are disabled -- for example when captured into a file, or stripped with
\code{\link[cli:ansi_strip]{cli::ansi_strip()}}.
}
\examples{
pw <- efa_power(df = 100, N = 200)
pw

# format() returns the same lines as a character vector:
writeLines(format(pw))

}
\seealso{
Other power analysis:
\code{\link[=efa_power]{efa_power()}},
\code{\link[=plot.efa_power]{plot.efa_power()}}
}
\concept{power analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.EFA_AVERAGE.R
\name{print.EFA_AVERAGE}
\alias{print.EFA_AVERAGE}
\title{Print EFA_AVERAGE object}
\usage{
\method{print}{EFA_AVERAGE}(x, stat = c("average", "range"), plot = TRUE, ...)
}
\arguments{
\item{x}{list. An object of class EFA_AVERAGE to be printed}

\item{stat}{character. A vector with the statistics to print. Possible inputs
are "average", "sd", "range", "min", and "max". Default is "average" and
"range".}

\item{plot}{logical. Whether a plot of the average and min- max loadings should
be created. Default is TRUE. If more than 10 factors are extracted, no plot is
created.}

\item{...}{Further arguments for print.}
}
\description{
Print Method showing a summarized output of the \link{EFA_AVERAGE} function
}
\examples{
\dontrun{
EFA_aver <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)
EFA_aver
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UPPS_raw_doc.R
\docType{data}
\name{UPPS_raw}
\alias{UPPS_raw}
\title{UPPS_raw}
\format{
A data.frame with 645 rows (participants) and 45 columns, named by a
subscale prefix and item number, covering the four UPPS subscales:
\describe{
\item{perseverance_1 to perseverance_10}{(numeric) - Perseverance-subscale items.}
\item{premeditation_1 to premeditation_11}{(numeric) - Premeditation-subscale items.}
\item{ss_1 to ss_12}{(numeric) - Sensation-seeking-subscale items.}
\item{urgency_1 to urgency_12}{(numeric) - Urgency-subscale items.}
}
}
\source{
Whiteside, S. P., Lynam, D. R., Miller, J. D., & Reynolds, S. K. (2005).
Validation of the UPPS impulsive behaviour scale: A four-factor model of
impulsivity. European Journal of Personality, 19 (7), 559–574.

Steiner, M., & Frey, R. (2020). Representative design in psychological assessment: A case study using the Balloon Analogue Risk Task (BART). PsyArXiv Preprint. doi:10.31234/osf.io/dg4ks
}
\usage{
UPPS_raw
}
\description{
A dataframe containing responses to the UPPS personality scale (Whiteside &
Lynam, 2005) of 645 participants of Study 2 of Steiner and Frey (2020). Each
column are the ratings to one of 45 items to assess urgency, premeditation,
perseverance, and sensation seeking. The original data can be accessed via
\url{https://osf.io/kxp8t/}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{SMT}
\alias{SMT}
\title{Sequential model tests}
\usage{
SMT(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used. Must be larger than the number of variables.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{SMT} rests on a normal-theory chi-square test that is not valid for
polychoric / tetrachoric correlations.
Default is \code{"pearson"}.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_smt]{efa_smt()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{SMT()} has been superseded by \code{\link[=efa_smt]{efa_smt()}}, which is the recommended interface
going forward. It remains available and unchanged so existing code keeps working.
}
\seealso{
\code{\link[=efa_smt]{efa_smt()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.rotate_geomin_oblq}
\alias{.rotate_geomin_oblq}
\title{Oblique geomin factor rotation}
\usage{
.rotate_geomin_oblq(
  L,
  delta = 0.01,
  eps = 1e-05,
  normalize = TRUE,
  random_starts = 0L,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1,
  screen_keep = 5L,
  triage_maxit = 25L,
  triage_improve_tol = 0
)
}
\arguments{
\item{L}{Numeric matrix. The unrotated loading matrix (variables by factors).}

\item{delta}{Numeric scalar. The geomin offset added to the squared loadings; must be a
positive finite scalar. \code{delta = 0.01} is the usual default.}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient norm.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization before
rotation and reverse it afterwards.}

\item{random_starts}{Integer scalar. Number of additional random orthogonal starts.}

\item{maxit}{Integer scalar. Maximum number of projected-gradient updates.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving attempts after
the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient update.}

\item{screen_keep}{Integer scalar. Number of screened random starts retained for
triage optimization.}

\item{triage_maxit}{Integer scalar. Number of short optimization iterations used in
the triage stage.}

\item{triage_improve_tol}{Numeric scalar. Relative improvement required for a triaged
start to be promoted to full optimization.}
}
\value{
A named list with the rotated loadings, the transformation matrix \code{Th}
(with \code{L \%*\% t(solve(Th))} reproducing the rotated loadings), the factor correlation
matrix \code{Phi} (\code{t(Th) \%*\% Th}), the attained criterion value, and the convergence and
validity flags. The list additionally reports the criterion value reached
at each optimized start in \code{all_values}, with a per-start convergence flag in \code{all_converged}.
}
\description{
Rotate a loading matrix obliquely under the geomin criterion using a gradient-projection
optimizer along the oblique (column-normalized) manifold.
}
\details{
The criterion value \code{f} and its gradient \code{dQ/dL} at the rotated loadings
\code{L = A \%*\% solve(t(T))} define the search; the engine maps the gradient to the
transformation \code{T} on the manifold \code{diag(t(T) \%*\% T) = 1}, projects it onto the tangent
space, performs a non-monotone line search, and retracts back onto the manifold by
column normalization. The geomin criterion sums the per-variable geometric mean of the
squared loadings offset by \code{delta}; it is prone to local minima, so additional random
starts are recommended.

Additional random starts may be requested. To bound runtime the solver screens each random
start by its objective, runs a short triage optimization on the best-screened starts, and
fully optimizes only those that improve on the current incumbent by at least
\code{triage_improve_tol}.
}
\references{
Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
software for arbitrary rotation criteria in factor analysis. \emph{Educational and
Psychological Measurement}, 65, 676-696.

Browne, M. W. (2001). An overview of analytic rotation in exploratory factor analysis.
\emph{Multivariate Behavioral Research}, 36, 111-150.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.oblique_procrustes}
\alias{.oblique_procrustes}
\title{Oblique Procrustes target rotation using a k x k inner objective}
\usage{
.oblique_procrustes(
  A,
  B,
  S_r = NULL,
  T_init_r = NULL,
  eps = 1e-05,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1,
  normalize = FALSE,
  random_starts = 0L,
  screen_keep = 2L,
  triage_maxit = 25L,
  triage_improve_tol = 0
)
}
\arguments{
\item{A}{Numeric matrix. Loading matrix to be rotated.}

\item{B}{Numeric matrix. Target loading matrix with the same dimensions as
\code{A}.}

\item{S_r}{Optional numeric \verb{k x k} matrix containing \code{crossprod(A)}.
Supplying this is useful when the same \code{A} is rotated repeatedly. Ignored
when \code{normalize = TRUE} because the normalized cross-product is different.}

\item{T_init_r}{Optional numeric \verb{k x k} starting transformation matrix.
If \code{NULL}, the identity matrix is used for the primary start.}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient
norm.}

\item{maxit}{Integer scalar. Maximum number of full projected-gradient
updates.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving
attempts after the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient
update.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization to the
loadings (only) before rotation and reverse it afterwards; the target is left
unnormalized, matching \code{GPArotation::targetQ(normalize = TRUE)}.}

\item{random_starts}{Integer scalar. Number of additional random starts.}

\item{screen_keep}{Integer scalar. Number of screened random starts retained
for triage optimization.}

\item{triage_maxit}{Integer scalar. Number of short optimization iterations
used in the triage stage.}

\item{triage_improve_tol}{Numeric scalar. Relative improvement required for a
triaged start to be promoted to full optimization.}
}
\value{
A named list containing the rotated loadings, transformation matrix,
factor correlation matrix, target criterion value, convergence diagnostics,
line-search diagnostics, and multi-start summaries.
}
\description{
Compute an oblique target rotation for a loading matrix using a
\code{targetQ}-compatible parameterization and a \verb{k x k} objective.
}
\details{
The rotated loading matrix is defined as
\code{L = A \%*\% solve(t(T))}, and the corresponding factor correlation matrix is
\code{Phi = t(T) \%*\% T}. The optimization is carried out over the transformation
matrix \code{T} under the oblique normalization constraint \code{diag(t(T) \%*\% T) = 1}.

Non-invertible candidate transformations are rejected rather than evaluated
through a pseudo-inverse.

Additional random starts may be requested. To reduce runtime, the solver uses
a two-stage strategy for extra starts: cheap objective screening, followed by
short triage optimization, followed by full optimization only for starts that
improve on the current incumbent by at least \code{triage_improve_tol}.

The routine is intended for repeated oblique target rotations in workflows
such as bootstrap alignment or consensus alignment of exploratory factor
solutions across multiply imputed datasets. It follows the same oblique
transformation convention as \code{GPArotation::targetQ()}.
}
\references{
Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms
and software for arbitrary rotation criteria in factor analysis.
\emph{Educational and Psychological Measurement}, 65, 676-696.

Browne, M. W. (2001). An overview of analytic rotation in exploratory factor
analysis. \emph{Multivariate Behavioral Research}, 36, 111-150.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_scree.R
\name{efa_scree}
\alias{efa_scree}
\title{Scree plot}
\source{
Cattell, R. B. (1966). The scree test for the number of factors.
Multivariate Behavioral Research, 1(2), 245–276.
https://doi.org/10.1207/s15327906mbr0102_10

Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
determining the number of components to retain. Psychological Bulletin, 99,
432–442. https://doi.org/10.1037/0033-2909.99.3.432
}
\usage{
efa_scree(
  x,
  eigen_type = c("PCA", "SMC", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_factors = 1,
  estimate_control = NULL,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{eigen_type}{character. On what the eigenvalues should be found. Can be
either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
the diagonal values of the correlation matrices are left to be 1. If using
"SMC", the diagonal of the
correlation matrices is replaced by the squared multiple correlations (SMCs)
of the indicators. If using "EFA", eigenvalues are found on the correlation
matrices with the final communalities of an exploratory factor analysis
solution (default is principal axis factoring extracting 1 factor) as
diagonal. Default is \code{c("PCA", "SMC", "EFA")}, i.e. all three; \code{"EFA"} is the
only one that fits a model.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator).
Default is "pearson".}

\item{n_factors}{numeric. Number of factors to extract if "EFA" is included in
\code{eigen_type}. Default is 1.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
\code{\link[=efa_fit]{efa_fit()}} fit that provides the communalities when \code{"EFA"} is included in \code{eigen_type}.
\code{NULL} (default) uses the \code{\link[=efa_fit]{efa_fit()}} defaults. The fit is unrotated, so no rotation settings
apply.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}}. For example,
\code{estimator}, to change the estimator (PAF is default). The estimation tuning knobs are not
passed here; they live in \code{estimate_control}, and the standard-error arguments (\code{se},
\code{b_boot}, \code{ci}, \code{seed}) are not accepted because the fit is an internal step that keeps
only its communalities.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). The scree plot is a
visual criterion, so it returns no numeric suggestion. Its main fields are:
\item{results}{A list with one record per requested eigenvalue type, each
holding the eigenvalues used for the scree plot.}
\item{settings}{A list of the settings used.}
}
\description{
The scree plot was originally introduced by Cattell (1966) to perform the
scree test. In a scree plot, the eigenvalues of the factors / components are
plotted against the index of the factors / components, ordered from 1 to N
factors components, hence from largest to smallest eigenvalue. According to
the scree test, the number of factors / components to retain is the number of
factors / components to the left of the "elbow" (where the curve starts to
level off) in the scree plot.
}
\details{
As the scree test requires visual examination, the test has been
especially criticized for its subjectivity and with this low inter-rater
reliability. Moreover, a scree plot can be ambiguous if there are either no
clear "elbow" or multiple "elbows", making it difficult to judge just where
the eigenvalues do level off. Finally, the scree test has also been found to
be less accurate than other factor retention criteria. For all these reasons,
the scree test has been recommended against, at least for exclusive use as a
factor retention criterion (Zwick & Velicer, 1986)

The \code{efa_scree} function can also be called together with other factor
retention criteria in the \code{\link[=efa_retain]{efa_retain()}} function.
}
\examples{
efa_scree(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.calc_cis}
\alias{.calc_cis}
\title{Confidence intervals around mean}
\usage{
.calc_cis(means, sds, p = 0.05, n)
}
\arguments{
\item{means}{A matrix with the pooled loadings or pooled interfactor
correlations}

\item{sds}{A matrix with the standard deviations for the loadings or
interfactor correlations}

\item{p}{Numeric. One minus the confidence level of the CIs. Defaults to 0.05
for 95\% CIs.}

\item{n}{Numeric. Typically, the number of permutations \eqn{m}. See
\code{\link{EFA_POOLED}}.}
}
\value{
A list with the lower and upper CIs.
}
\description{
Confidence intervals around mean
}
\details{
The standard error (SE) to use in the creation of the CIs is calculated
according to Rubin's (1987) formula (eq. 11 in Hayes & Enders, 2023):

\eqn{\sqrt{mean(SE^2) + var(\hat{\theta_{m}}) + var(\hat{\theta_{m}}) / M}}

According to Hayes & Enders (2023) p. 42:

\emph{[T]he first term under the radical represents the average squared
standard error (the within imputation sampling variance [...]), the second
term depends on the variance of the M parameter estimates around their
average (the between imputation variance [...]), and the final term
represents the squared standard error of the pooled estimate [...].
Conceptually, the first term estimates the sampling error of a complete-data
analysis, and the next two terms are essentially correction factors that
inflate the standard error to compensate for uncertainty due to the
imputations–that is, additional uncertainty (sampling variability) in the
parameter estimates caused by missing data.}

Currently, it is not possible to calculate the first term, because
\code{\link{EFA}} does not calculate SEs for the loadings. Only the second
and third terms are used in the calculation of SE for the CIs.

The CI is generally calculated as:

CI = Point estimate ± Margin of error,

where

Margin of error = Critical value × SE of point estimate.

To account for situations where the sample size is small, instead of using
z-values, the critical value is derived from the \emph{t} distribution with
\code{n - 1} degrees of freedom (Hazra, 2017).
}
\references{
Hayes, T. & Enders, C. K. (2023). Maximum likelihood and multiple imputation
missing data handling: how they work, and how to make them work in practice.
In \emph{APA Handbook of Research Methods in Psychology, Second Edition} Vol.
3. Data Analysis and Research Publication, H. Cooper (Editor-in-Chief).

Hazra, A. (2017). Using the confidence interval confidently. \emph{Journal of
Thoracic Disease} 9(10), 4125--4130.

Rubin, D. B. (1987). Multiple imputation for nonresponse in surveys. Wiley.
https://doi.org/10.1002/9780470316696
}
\author{
Andreas Soteriades

This function is used internally by \code{\link{EFA_POOLED}} to calculate
confidence intervals (CIs) around the pooled loadings and pooled interfactor
correlations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_group.R
\name{efa_group}
\alias{efa_group}
\title{Multigroup exploratory factor analysis}
\usage{
efa_group(
  x,
  groups = NULL,
  n_factors,
  N = NA,
  reference_group = NULL,
  b_boot = 0L,
  ci = 0.95,
  seed = NULL,
  delta = 0.1,
  invariance = FALSE,
  se = NULL,
  ...
)
}
\arguments{
\item{x}{A data frame or matrix of raw data (with \code{groups}), or a named list
of per-group data sets -- either raw data frames/matrices or correlation
matrices (all of one kind).}

\item{groups}{A vector with one value per row of \code{x}, giving each row's group.
Only used when \code{x} is a single raw data set; leave \code{NULL} when \code{x} is a list.
Rows with a missing group value are dropped with a warning.}

\item{n_factors}{numeric. The common number of factors extracted in every
group.}

\item{N}{numeric. The number of observations per group, used only for
correlation-matrix input: either a single value applied to all groups or one
value per group. Ignored for raw data, where \code{N} is taken from each group's
data. Default is \code{NA}.}

\item{reference_group}{The group to align the others to (a group name or an
integer index). If \code{NULL} (default), orthogonal and unrotated solutions use
the symmetric consensus target; oblique solutions with more than one factor
fall back to the first group as reference. Supplying a value forces the
reference alignment.}

\item{b_boot}{numeric. The number of non-parametric bootstrap replicates used to
form percentile confidence intervals for the between-group Tucker congruences.
\code{0} (the default) skips the bootstrap and returns the congruence point estimates
only. Bootstrapping requires raw data; it is skipped with a warning for
correlation-matrix input.}

\item{ci}{numeric. The confidence level for the bootstrap congruence intervals, a
single value in \verb{(0, 1)}. Default is \code{0.95}.}

\item{seed}{numeric or \code{NULL}. An optional seed making the analysis reproducible. It
covers the per-group fits, some of whose rotations draw random starts, whether or not a
bootstrap is run. With a bootstrap, it also makes the result independent of how many
parallel workers are used (bootstrap replicates run with
\link[future.apply:future_lapply]{future_lapply()}, configurable via \code{\link[future:plan]{future::plan()}}). The
caller's random-number stream is restored afterwards, leaving no side effect. Default is
\code{NULL}.}

\item{delta}{numeric. The salience threshold for the per-item loading-difference flag
table: an item's loading on a factor is flagged for a group pair when the groups'
aligned loadings differ by at least \code{delta} in absolute value. This is a descriptive
salience heuristic, not a significance test; common alternatives are \code{0.15} and \code{0.20}.
The threshold applies to whatever loading metric the chosen rotation produces (pattern
coefficients for an oblique rotation). \code{0} flags every cell. Default is \code{0.1}.

The geomin rotations take a criterion parameter of the same name. A \code{delta} given
directly is always this salience threshold and never reaches the rotation; give the
geomin parameter as \code{rotate_control(delta = ...)}. With \code{rotation = "geominT"} or
\code{"geominQ"}, a supplied \code{delta} gives a warning that says which of the two applies.}

\item{invariance}{logical. Whether to add an approximate-invariance verdict per factor and
group pair from the Lorenzo-Seva and ten Berge (2006) congruence bands (see \emph{Value}).
Default is \code{FALSE}.}

\item{se}{Not used. \code{efa_group()} itself sets the standard-error method of the
per-group \code{\link[=efa_fit]{efa_fit()}} calls, so a supplied value is dropped with a warning. Ask for
bootstrap confidence intervals of the between-group congruences with \code{b_boot}.
Default is \code{NULL}.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}} for every group (for
example \code{estimator}, \code{rotation}, or \code{cor_method}). The \code{\link[=estimate_control]{estimate_control()}} and
\code{\link[=rotate_control]{rotate_control()}} objects are accepted through \code{...} as well, although they are not
declared formals: pass them as \verb{estimate_control =} / \verb{rotate_control =} exactly
as you would to \code{\link[=efa_fit]{efa_fit()}}. A name that is neither an \code{\link[=efa_fit]{efa_fit()}} argument nor a
rotation-engine extra is rejected.

A rotation-engine extra that shares a name with an \code{efa_group()} argument cannot reach
the rotation through \code{...}, because the argument takes the name first (for example
geomin's \code{delta}; see \code{delta} above).}
}
\value{
An object of class \code{efa_group}, a list containing:
\item{loadings}{A named list of the aligned per-group loading matrices. Their
columns match the columns of \code{target} in order and sign.}
\item{target}{The alignment target: the symmetric consensus target, or the
reference group's own loadings.}
\item{Phi}{A named list of the aligned per-group factor intercorrelations for
an oblique rotation; \code{NULL} otherwise.}
\item{congruence}{Tucker congruence between the aligned group loadings, a list
with: \code{matrices}, a nested list whose \verb{[[g]][[h]]} element is the
factor-by-factor congruence matrix between the aligned loadings of groups
\code{g} and \code{h}; \code{matched}, a groups-by-groups-by-factors array of the
matched-factor congruences (the diagonal of each pairwise matrix); and
\code{degenerate}, a groups-by-groups logical matrix flagging pairs whose
congruence is undefined (for example, a near-zero factor), for which the
corresponding entries are \code{NA}. When \code{b_boot > 0} (raw data), three further
elements are added: \code{matched_se}, the bootstrap standard error of each
matched congruence; \code{matched_ci}, a list of \code{lower} and \code{upper} percentile
confidence limits (each a groups-by-groups-by-factors array); and \code{n_boot},
the number of bootstrap replicates that aligned in every group and so
contributed to the intervals (a replicate whose fit did not converge is
retained, as in \code{\link[=efa_fit]{efa_fit()}}).}
\item{diffs}{A data frame with one row per group pair summarising the differences
between their aligned loadings: the mean, median, minimum, and maximum absolute
difference, the root-mean-square difference (\code{rmse}), and \code{n_flagged}, the number of
loading cells whose absolute difference reaches \code{delta}.}
\item{flags}{A data frame with one row per group pair, item, and factor giving the signed
loading difference (\code{diff}), its absolute value (\code{abs_diff}), and \code{flagged}, whether it
reaches \code{delta}. When a bootstrap was run (\code{b_boot > 0}, raw data), \code{ci_lower},
\code{ci_upper}, and \code{ci_excludes_0} add the percentile confidence interval for the
difference and whether it excludes zero; otherwise these are \code{NA}.}
\item{invariance}{When \code{invariance = TRUE}, a data frame with one row per group pair and
factor giving the matched Tucker congruence (\code{phi}), its bootstrap CI lower bound
(\code{phi_lower}, \code{NA} without a bootstrap), and an approximate-invariance \code{verdict} based on
the Lorenzo-Seva and ten Berge (2006) similarity bands: \code{phi >= 0.95} is "equal" and
\verb{[0.85, 0.95)} is "fair"; congruences \verb{< 0.85}, below their bands, are labelled
"incongruent". The verdict is read from \code{phi_lower} when a bootstrap is available
(conservative) and from \code{phi} otherwise. A wide interval therefore lowers the verdict:
\code{phi} = 0.989 with \code{phi_lower} = 0.726 is labelled "incongruent", because the band is
applied to the lower bound. Tucker's congruence is invariant to a
proportional rescaling of a factor's loadings, so a factor can be graded "equal"
even when one group's loadings on it are uniformly stronger; read the verdict
alongside \code{diffs}. \code{NULL} when \code{invariance = FALSE}.}
\item{efa}{The named list of per-group \code{\link[=efa_fit]{efa_fit()}} objects (each retains its own
diagnostics, e.g. \code{heywood}).}
\item{alignment}{The alignment result: the consensus object (see
\code{\link[=efa_procrustes]{efa_procrustes()}}), or a list with the reference group, the target, and the
per-group Procrustes results. On the consensus path, this is the raw Procrustes
iteration output: its \code{target}/\code{aligned_loadings} are in a different (pre-gauge)
orientation than the gauged \code{target}/\code{loadings} returned above. Use \code{target}/\code{loadings}
above for comparisons.}
\item{settings}{A list of the settings used, including the per-group \code{N}, the
alignment method, the group that seeded the consensus frame (\code{alignment_start},
\code{NULL} on the reference path), the orientation the shared frame was put in
(\code{gauge}: the rotation's own name, \code{"principal_axes"}, or \code{"identity"} for a
single factor; \code{NULL} on the reference path), the rotation, the estimator, the
input type, whether a bootstrap is available (\code{can_bootstrap}, \code{FALSE} for
correlation-matrix input), and \code{seed} (\code{NULL} when none was supplied).}
}
\description{
Fit an exploratory factor analysis in each of several groups at a common
number of factors and bring the per-group solutions into one shared
orientation so their loadings can be compared. Each group is fitted with
\code{\link[=efa_fit]{efa_fit()}}; the solutions are then aligned either to a symmetric consensus target
or to a chosen reference group (see \emph{Alignment}).
}
\details{
\subsection{Input}{

Groups can be supplied in two ways: raw data together with a grouping vector
(\code{x} a data frame or matrix, \code{groups} one value per row), or a named list of
per-group data sets in \code{x} (with \code{groups} left \code{NULL}). The list may hold raw
data frames or correlation matrices (supply \code{N}), but not a mix of the two.
All groups must contain the same items in the same order; a different item set
or order is an error rather than being silently reordered.

Every group is fitted at the same \code{n_factors}. Extra arguments in \code{...} (for example
\code{estimator}, \code{rotation}, \code{cor_method}, or an \code{\link[=estimate_control]{estimate_control()}} / \code{\link[=rotate_control]{rotate_control()}}
carrying the tuning knobs) are forwarded unchanged to each \code{\link[=efa_fit]{efa_fit()}} call, so the
estimator and rotation are common to all groups.

The requested number of factors must be small enough, relative to the number of items,
for the \code{n_factors}-factor model to be identified for the shared item set. Unlike a single
\code{\link[=efa_fit]{efa_fit()}} fit -- which only warns on an under-identified model -- a multigroup fit aborts
when this fails, because a shared alignment target across an under-identified group is not
interpretable.
}

\subsection{Alignment}{

A factor solution is identified only up to a rotation of its factors, so the
per-group solutions must be brought into a common orientation before their
loadings can be compared. Two strategies are available and are chosen
automatically:
\itemize{
\item \strong{Consensus} (the default for orthogonal rotations and for unrotated
solutions): a symmetric target is built across all groups using Generalized Procrustes
Analysis (Gower, 1975), and every group's loadings are rotated to it. Because this
target's own orientation is arbitrary, it is then rotated once more into a fixed
convention (called the gauge), and the same transform is applied to every group. The
gauge uses the same simple-structure criterion as the requested rotation, applied to the
target itself, so the shared loadings are in the same kind of frame as the per-group
solutions they summarise. Where no rotation criterion identifies a unique frame (an
unrotated solution, or a two-factor \code{bifactorT} request), the target's principal-axes
orientation is used instead. Either way the columns are ordered by decreasing sum of
squares and signed by their column sums, and the shared orientation -- and hence every
reported congruence, difference, and flag -- does not depend on the order the groups are
supplied, to well beyond the precision loadings are reported at.
\item \strong{Reference}: every group's loadings are aligned by Procrustes rotation to
one reference group's loadings, which are kept fixed. This path is used when
\code{reference_group} is given, and is used automatically for oblique rotations
because the consensus iteration is not defined for oblique transforms with
more than one factor. When an oblique rotation triggers the reference path
without an explicit \code{reference_group}, the first group is used and a message
reports this; the requested rotation is never silently changed.
}

In both cases the returned per-group loadings share the column order and sign
of the returned \code{target}.
}

\subsection{Comparing the aligned loadings}{

Because the per-group loadings share one orientation, they can be compared cell by cell.
\code{efa_group()} reports a per-pair summary of their differences (\code{diffs}) and a per-item,
per-factor flag table (\code{flags}) marking cells whose absolute difference reaches \code{delta}; a
bootstrap (\code{b_boot > 0}) additionally reports, for every cell, whether its difference's
confidence interval excludes zero. With \code{invariance = TRUE}, each factor and group pair also gets an
approximate-invariance verdict based on the matched Tucker congruence (see \emph{Value} for the
similarity bands and how a bootstrap is used).
}
}
\examples{
# Raw data split by a grouping vector (unrotated, consensus alignment)
g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
mg <- efa_group(GRiPS_raw, groups = g, n_factors = 1)
mg$loadings

# Per-pair difference summary and the per-item salience-flag table
mg$diffs
mg$flags

\donttest{
# Percentile bootstrap confidence intervals for the between-group congruences, with an
# approximate-invariance verdict read conservatively off the congruence CI lower bound
mg_ci <- efa_group(GRiPS_raw, groups = g, n_factors = 1, b_boot = 100, seed = 42,
                   invariance = TRUE)
mg_ci$congruence$matched_ci
mg_ci$invariance

# A named list of correlation matrices sharing the same items, common
# three-factor model, orthogonal rotation -> symmetric consensus target
bands <- list(age_6_8 = WJIV_ages_6_8$cormat, age_14_19 = WJIV_ages_14_19$cormat)
Ns <- c(WJIV_ages_6_8$N, WJIV_ages_14_19$N)
efa_group(bands, n_factors = 3, N = Ns, rotation = "varimax")

# An oblique rotation aligns to a reference group (reported via a message)
efa_group(bands, n_factors = 3, N = Ns, rotation = "promax")
}

}
\references{
Efron, B., & Tibshirani, R. J. (1993). \emph{An Introduction to the Bootstrap}.
Chapman & Hall.

Gower, J. C. (1975). Generalized Procrustes analysis. \emph{Psychometrika}, 40,
33-51. doi: 10.1007/BF02291478

Lorenzo-Seva, U., and ten Berge, J. M. F. (2006). Tucker's congruence coefficient as a
meaningful index of factor similarity. \emph{Methodology}, 2, 57-64.
doi: 10.1027/1614-2241.2.2.57
}
\seealso{
Other factor analysis:
\code{\link[=efa_average]{efa_average()}},
\code{\link[=efa_fit]{efa_fit()}},
\code{\link[=efa_mi]{efa_mi()}},
\code{\link[=plot.efa_group]{plot.efa_group()}},
\code{\link[=print.efa_group]{print.efa_group()}}
}
\concept{factor analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.SMT.R
\name{print.SMT}
\alias{print.SMT}
\title{Print SMT object}
\usage{
\method{print}{SMT}(x, ...)
}
\arguments{
\item{x}{list of class SMT (output from the \link{SMT} function)}

\item{...}{additional arguments passed to print}
}
\description{
Print SMT object
}
\examples{
SMT_base <- SMT(test_models$baseline$cormat, N = 500)
SMT_base

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{EFA_POOLED}
\alias{EFA_POOLED}
\title{Exploratory factor analysis on multiple data imputations}
\usage{
EFA_POOLED(
  data_list,
  p = 0.05,
  target_method = c("first_target", "consensus"),
  align_unrotated = c("signed_tucker_congruence", "none", "procrustes"),
  fit_pool_method = c("D2"),
  consensus_args = list(),
  procrustes_args = list(),
  rmsea_ci_level = 0.9,
  rmsr_upper = TRUE,
  ...
)
}
\arguments{
\item{data_list}{A list of length \eqn{m}, where \eqn{m} is the number of
imputations. Each list element is a data frame or matrix of raw data, or a
correlation matrix. See argument \code{x} in \code{\link[=efa_fit]{efa_fit()}}. A \code{mids} object from
\pkg{mice} must be converted first, with \code{mice::complete(x, "all")}.}

\item{p}{Numeric in \eqn{(0, 1)}. One minus the confidence level for the
pooled confidence intervals, whichever \code{se} method produced them
(\code{"information"}, \code{"np-boot"}, or \code{"sandwich"}). For example, \code{p = .05}
gives 95\% intervals.}

\item{target_method}{Character. How rotated solutions are aligned across imputations
before pooling: \code{"first_target"} (the default) aligns every imputation to the first
imputation's rotated solution, while \code{"consensus"} refines a centroid target by
Generalized Procrustes Analysis, started from the medoid imputation so that the
pooled rotated solution does not depend on the order of \code{data_list} (orthogonal
rotations only). See \emph{Aligning solutions across imputations} in Details.}

\item{align_unrotated}{Character. How unrotated loadings are aligned before pooling:
\code{"signed_tucker_congruence"} (the default; sign/permutation via Tucker congruence,
anchored on the medoid imputation and returned in the extraction's canonical
gauge), \code{"procrustes"} (orthogonal Procrustes to the first imputation), or
\code{"none"}. See \emph{Aligning solutions across imputations} in Details.}

\item{fit_pool_method}{Character. Only \code{"D2"} is implemented for pooling
chi-square-type fit. If no chi-square is available, only residual-based
fit and descriptive quantities are returned. See \emph{Pooling the model chi-square and
fit indices} in Details.}

\item{consensus_args}{List of additional arguments controlling the
GPA-consensus iteration when \code{target_method = "consensus"}. Recognised tuning
parameters include the convergence tolerances \code{tol} and \code{loss_tol}, the
iteration bounds \code{min_iter} and \code{max_iter}, the target-update damping \code{alpha},
the multi-start controls \code{multi_start} and \code{starts}, and \code{start}, which
overrides the medoid imputation the iteration is otherwise started from.}

\item{procrustes_args}{List of \code{\link[=efa_procrustes]{efa_procrustes()}} algorithm controls for
fixed-target alignment, for example \code{oblique_maxit} or \code{oblique_random_starts}.
The loadings \code{A}, the alignment \code{Target}, the \code{rotation} family, and the
cross-product \code{S} are derived from the imputations and cannot be set here.}

\item{rmsea_ci_level}{Numeric. Confidence level for the RMSEA CI.}

\item{rmsr_upper}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Deprecated and ignored.
\code{efa_mi()} now always computes RMSR the same way, from the unique
off-diagonal residuals; SRMR is reported alongside it. Supplying it to
\code{\link[=efa_mi]{efa_mi()}} signals a deprecation warning; the superseded \code{\link[=EFA_POOLED]{EFA_POOLED()}}
accepts it silently.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}} (e.g. \code{estimator}, \code{rotation}, \code{se},
\code{n_factors}, \code{N}). These select the estimator, rotation, standard-error method, and
fit indices used for every imputation; see \code{\link[=efa_fit]{efa_fit()}} for the available options, their
properties, and which combinations are valid. Two of them shape the pooled object
rather than a single fit: \code{seed} sets the random state once for the whole
\code{efa_mi()} call -- every component bootstrap and every random-start rotation draws
from it, so a seeded call is reproducible as a whole, and the caller's random stream
is restored afterwards -- and \code{b_boot} sets the number of bootstrap replicates drawn
per imputation under \code{se = "np-boot"}, which is what the pooled within-imputation
variances are estimated from and is recorded in \code{settings$b_boot}. The
\code{\link[=estimate_control]{estimate_control()}} and
\code{\link[=rotate_control]{rotate_control()}} objects are accepted through \code{...} as well, although they are not
declared formals: pass them as \verb{estimate_control =} / \verb{rotate_control =} exactly as you
would to \code{\link[=efa_fit]{efa_fit()}}.}
}
\value{
The value of \code{\link[=efa_mi]{efa_mi()}}, normally a list of class
\code{c("efa_mi", "EFA_POOLED", "efa", "EFA")}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{EFA_POOLED()} has been superseded by \code{\link[=efa_mi]{efa_mi()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_mi]{efa_mi()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_screen.R
\name{print.efa_screen}
\alias{print.efa_screen}
\alias{format.efa_screen}
\title{Print and format an efa_screen object}
\usage{
\method{print}{efa_screen}(x, digits = 3, ...)

\method{format}{efa_screen}(x, digits = 3, ...)
}
\arguments{
\item{x}{An object of class \code{efa_screen} (output from \code{\link[=efa_screen]{efa_screen()}}).}

\item{digits}{Integer. The number of decimal places the reported values are
rounded to. Default is 3.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} turns the factor-analysis screening diagnostics computed by
\code{\link[=efa_screen]{efa_screen()}} into a sectioned report with banded, colour-coded verdicts:
sampling adequacy and sphericity (the Kaiser-Meyer-Olkin measure and Bartlett's
test of sphericity), multicollinearity (the determinant and condition number of
the correlation matrix), the per-variable diagnostics, and, when raw data were
supplied, multivariate normality and multivariate outliers. It closes with a
consolidated list of actionable recommendations (for example, which items to
consider dropping, whether to prefer an ordinal or a robust estimator, and a
caveat that keeps an over-powered Bartlett's test from being over-trusted).
\code{format()} assembles the same report and returns it as a character vector;
\code{print()} is \code{cat(format(x), sep = "\\n")}. The lines follow the active console
theme, so they are plain when colours are disabled (for example when captured
into a file or stripped with \code{\link[cli:ansi_strip]{cli::ansi_strip()}}). \code{print()} does not draw a
plot.
}
\examples{
# From raw data
efa_screen(iris[, 1:4])

# From a correlation matrix (supply N for Bartlett's test of sphericity)
efa_screen(test_models$baseline$cormat, N = 500)

# format() returns the same lines as a character vector:
writeLines(format(efa_screen(test_models$baseline$cormat, N = 500)))

}
\seealso{
Other factor analysis suitability:
\code{\link[=efa_bartlett]{efa_bartlett()}},
\code{\link[=efa_kmo]{efa_kmo()}},
\code{\link[=efa_screen]{efa_screen()}}
}
\concept{factor analysis suitability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.numformat}
\alias{.numformat}
\title{Format numbers for print method}
\usage{
.numformat(x, digits = 2, print_zero = FALSE)
}
\arguments{
\item{x}{numeric. Number to be formatted.}

\item{digits}{numeric. Number of digits after the comma to keep.}

\item{print_zero}{logical. Whether, if a number is between [-1, 1], the
zero should be omitted or printed (default is FALSE, i.e. omit zeros).}
}
\value{
A formated number
}
\description{
Helper function used in the print method for class LOADINGS and SLLOADINGS.
Strips the 0 in front of the decimal point of a number if number < 1, only
keeps the first \code{digits} number of digits, and adds an empty space in
front of the number if the number is positive. This way all returned strings
(except for those > 1, which are exceptions in LOADINGS) have the same number
of characters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{PARALLEL}
\alias{PARALLEL}
\title{Parallel analysis}
\usage{
PARALLEL(
  x = NULL,
  N = NA,
  n_vars = NA,
  n_datasets = 1000,
  percent = 95,
  eigen_type = c("PCA", "SMC", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  decision_rule = c("means", "percentile", "crawford"),
  n_factors = 1,
  ...
)
}
\arguments{
\item{x}{matrix or data.frame. The real data to compare the simulated eigenvalues
against. Must not contain variables of classes other than numeric. Can be a
correlation matrix or raw data.}

\item{N}{numeric. The number of cases / observations to simulate. Only has to
be specified if \code{x} is either a correlation matrix or \code{NULL}. If
x contains raw data, \code{N} is found from the dimensions of \code{x}. Must be larger
than the number of variables.}

\item{n_vars}{numeric. The number of variables / indicators to simulate.
Only has to be specified if \code{x} is left as \code{NULL} as otherwise the
dimensions are taken from \code{x}.}

\item{n_datasets}{numeric. The number of datasets to simulate. Must be at
least 1. Default is 1000.}

\item{percent}{numeric. The percentile to take from the simulated eigenvalues.
Default is 95.}

\item{eigen_type}{character. On what the eigenvalues should be found. Can be
either "SMC", "PCA", or "EFA". If using "SMC", the diagonal of the correlation
matrix is replaced by the squared multiple correlations (SMCs) of the
indicators. If using "PCA", the diagonal values of the correlation matrices
are left to be 1. If using "EFA", eigenvalues are found on the correlation
matrices with the final communalities of an EFA solution as diagonal. Default
is \code{c("PCA", "SMC", "EFA")}, i.e. all three, which costs roughly six times a
single non-EFA type: \code{"EFA"} fits an EFA to every simulated dataset and
dominates that total. Pass a single type if the run is time-critical.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{PARALLEL} compares the data against simulated continuous reference data.
Default is "pearson".}

\item{decision_rule}{character. Which rule to use to determine the number of
factors to retain. Default is \code{"means"}, which will use the average
simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
in percent. \code{"crawford"} uses the 95th percentile for the first factor
and the mean afterwards (based on Crawford et al, 2010). All three rules retain
the factors up to the first observed eigenvalue that fails to exceed its
reference value; an eigenvalue further down the series that rises above its own
reference again therefore adds no factor. Because the average simulated
eigenvalue is a lower reference than the percentile, \code{"means"} tends to retain
more factors than the more conservative \code{"percentile"} rule (Glorfeld, 1995).}

\item{n_factors}{numeric. Number of factors to extract if "EFA" is included in
\code{eigen_type}. Default is 1.}

\item{...}{Further arguments passed on to the \code{\link[=efa_fit]{efa_fit()}} fits. For example,
\code{estimator}, to change the estimator (default is "PAF"; PAF is more robust, but it
will take longer compared to "ML" and "ULS"), or one of the estimation tuning knobs
(\code{type}, \code{init_comm}, \code{criterion}, \code{criterion_type}, \code{max_iter}, \code{abs_eigen},
\code{start_method}), which are repacked into an \code{\link[=estimate_control]{estimate_control()}} object so that they
tune the fits exactly as they always did.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_parallel]{efa_parallel()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{PARALLEL()} has been superseded by \code{\link[=efa_parallel]{efa_parallel()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_parallel]{efa_parallel()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.NEST.R
\name{print.NEST}
\alias{print.NEST}
\title{Print function for NEST objects}
\usage{
\method{print}{NEST}(x, ...)
}
\arguments{
\item{x}{a list of class NEST Output from \code{\link{NEST}} function.}

\item{...}{Further arguments for print.}
}
\description{
Print function for NEST objects
}
\examples{
NEST(test_models$baseline$cormat, N = 500)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.N_FACTORS.R
\name{print.N_FACTORS}
\alias{print.N_FACTORS}
\title{Print function for N_FACTORS objects}
\usage{
\method{print}{N_FACTORS}(x, ...)
}
\arguments{
\item{x}{a list of class N_FACTORS. Output from \link{N_FACTORS} function.}

\item{...}{Further arguments for print.}
}
\description{
Print function for N_FACTORS objects
}
\examples{
\donttest{
# All criteria except "CD", with correlation matrix and fit method "ML"
# (where needed)
N_FACTORS(test_models$baseline$cormat, criteria = c("EKC", "HULL", "KGC",
          "PARALLEL", "SCREE", "SMT"), N = 500, method = "ML")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.EFA.R
\name{print.EFA}
\alias{print.EFA}
\alias{print.EFA_POOLED}
\alias{format.EFA}
\alias{format.EFA_POOLED}
\title{Print EFA object}
\usage{
\method{print}{EFA}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  ci = c("auto", "none", "separate"),
  ci_filter = c("salient", "all", "nonzero"),
  details = c("standard", "compact", "full"),
  diagnostics = TRUE,
  diagnostics_top_n = 10,
  residual_cutoff = 0.1,
  residual_top_n = 10,
  show_structure = FALSE,
  sort_loadings = c("none", "primary", "clustered"),
  show_loading_legend = TRUE,
  cross_loading_cutoff = cutoff,
  min_primary_gap = 0.2,
  min_salient_per_factor = 3,
  max_factors_per_block = NULL,
  show_mi_diagnostics = NULL,
  ...
)

\method{print}{EFA_POOLED}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  ci = c("auto", "none", "separate"),
  ci_filter = c("salient", "all", "nonzero"),
  details = c("standard", "compact", "full"),
  diagnostics = TRUE,
  diagnostics_top_n = 10,
  residual_cutoff = 0.1,
  residual_top_n = 10,
  show_structure = FALSE,
  sort_loadings = c("none", "primary", "clustered"),
  show_loading_legend = TRUE,
  cross_loading_cutoff = cutoff,
  min_primary_gap = 0.2,
  min_salient_per_factor = 3,
  max_factors_per_block = NULL,
  show_mi_diagnostics = NULL,
  ...
)

\method{format}{EFA}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  ci = c("auto", "none", "separate"),
  ci_filter = c("salient", "all", "nonzero"),
  details = c("standard", "compact", "full"),
  diagnostics = TRUE,
  diagnostics_top_n = 10,
  residual_cutoff = 0.1,
  residual_top_n = 10,
  show_structure = FALSE,
  sort_loadings = c("none", "primary", "clustered"),
  show_loading_legend = TRUE,
  cross_loading_cutoff = cutoff,
  min_primary_gap = 0.2,
  min_salient_per_factor = 3,
  max_factors_per_block = NULL,
  show_mi_diagnostics = NULL,
  ...
)

\method{format}{EFA_POOLED}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  ci = c("auto", "none", "separate"),
  ci_filter = c("salient", "all", "nonzero"),
  details = c("standard", "compact", "full"),
  diagnostics = TRUE,
  diagnostics_top_n = 10,
  residual_cutoff = 0.1,
  residual_top_n = 10,
  show_structure = FALSE,
  sort_loadings = c("none", "primary", "clustered"),
  show_loading_legend = TRUE,
  cross_loading_cutoff = cutoff,
  min_primary_gap = 0.2,
  min_salient_per_factor = 3,
  max_factors_per_block = NULL,
  show_mi_diagnostics = NULL,
  ...
)
}
\arguments{
\item{x}{list. An object of class EFA to be printed}

\item{cutoff}{numeric. Passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.
The number above which to print loadings in bold. Default is .3.}

\item{digits}{numeric. Passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}
Number of digits to round the loadings to (default is 3).}

\item{max_name_length}{numeric. Passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.
The maximum length of the variable names to display. Everything beyond this
will be cut from the right.}

\item{ci}{character. Whether to print confidence intervals for loadings and
factor intercorrelations, if available. \code{"auto"} and \code{"separate"}
print separate CI sections when CIs were computed; \code{"none"} suppresses
these sections. Default is \code{"auto"}.}

\item{ci_filter}{character. Which loading CIs to print. \code{"salient"}
prints CIs for loadings with absolute values greater than or equal to
\code{cutoff}; \code{"all"} prints all loading CIs; \code{"nonzero"} prints
loading CIs that exclude zero. Default is \code{"salient"}.}

\item{details}{character. Amount of information to print. \code{"standard"}
prints the regular output, \code{"compact"} suppresses longer diagnostic
sections, and \code{"full"} additionally prints available optional sections.}

\item{diagnostics}{logical. Whether to print model and simple-structure
diagnostics. Default is \code{TRUE}.}

\item{diagnostics_top_n}{numeric. Maximum number of item-level diagnostic
entries to print per diagnostic type.}

\item{residual_cutoff}{numeric. Absolute residual cutoff used in the residual
diagnostics section. Default is .1.}

\item{residual_top_n}{numeric. Maximum number of residuals to print. Use
\code{Inf} to print all residuals above \code{residual_cutoff}.}

\item{show_structure}{logical. Whether to print the structure matrix for
oblique solutions when available. Default is \code{FALSE}.}

\item{sort_loadings}{character. Optional row sorting for loading tables.
See \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.}

\item{show_loading_legend}{logical. Whether to print a compact legend for
loading-table styling. Default is \code{TRUE}.}

\item{cross_loading_cutoff}{numeric. Cutoff for counting cross-loadings in
the diagnostics. Defaults to \code{cutoff}.}

\item{min_primary_gap}{numeric. Minimum desired absolute difference between
the largest and second-largest absolute loading of an item. Used only for
descriptive diagnostics.}

\item{min_salient_per_factor}{numeric. Minimum number of salient indicators
per factor used in the diagnostics. Default is 3.}

\item{max_factors_per_block}{numeric or \code{NULL}. Maximum number of factor
columns per loading-table block. If \code{NULL}, this is chosen from the
console width.}

\item{show_mi_diagnostics}{logical or \code{NULL}. Whether to print a compact
MI uncertainty summary for pooled EFAs when available. \code{NULL} prints it
only for \code{details = "full"}.}

\item{...}{Further arguments passed to \code{\link[EFAtools:print.LOADINGS]{print.LOADINGS}}.}
}
\description{
Print Method showing a summarized output of the \link{EFA} function.
}
\details{
The method is shared by single-imputation `EFA` objects and pooled
`EFA_POOLED` objects. It prints, in order, a compact model header, optional
diagnostics, the loading table, optional confidence-interval tables, variance
accounted for, model fit, bootstrap notes, optional multiple-imputation
uncertainty diagnostics, and residual diagnostics.

For `EFA_POOLED` objects, the header and diagnostics report the number of
imputations and the alignment/pooling settings stored in the object. Loading
and factor-correlation confidence intervals are printed only when `boot.CI`
is present. Pooled intervals created by `EFA_POOLED()` are labelled as
bootstrap/MI intervals.

`ci_filter` controls which loading intervals are shown. `salient` reports
intervals for loadings whose absolute point estimate is at least `cutoff`;
`nonzero` reports intervals excluding zero; and `all` reports every finite
interval.
}
\examples{
mod <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
           method = "PAF", rotation = "promax")
mod

# With SEs and CIs, needs raw-data
mod <- EFA(GRiPS_raw, n_factors = 1, method = "ML", se = "np-boot")
mod
# with additional infos
print(mod, details = "full")
# omit some diagnostic parts
print(mod, details = "compact")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_screen.R
\name{efa_screen}
\alias{efa_screen}
\title{Screen data for exploratory factor analysis}
\source{
Bartlett, M. S. (1951). The effect of standardization on a Chi-square
approximation in factor analysis. Biometrika, 38, 337-344.

Belsley, D. A. (1991). A guide to using the collinearity diagnostics. Computer
Science in Economics and Management, 4, 33-50.

Belsley, D. A., Kuh, E. & Welsch, R. E. (1980). Regression diagnostics:
Identifying influential data and sources of collinearity. Wiley.

Cochran, W. G. (1954). Some methods for strengthening the common
\eqn{\chi^2} tests. Biometrics, 10, 417-451.

Croux, C. & Haesbroeck, G. (1999). Influence function and efficiency of the
minimum covariance determinant scatter matrix estimator. Journal of Multivariate
Analysis, 71, 161-190.

Field, A. (2018). Discovering statistics using IBM SPSS statistics (5th
ed.). Sage.

Henze, N. & Zirkler, B. (1990). A class of invariant consistent tests for
multivariate normality. Communications in Statistics - Theory and Methods, 19,
3595-3617.

Kaiser, H. F. (1970). A second generation little jiffy. Psychometrika,
35, 401-415.

Kaiser, H. F. & Rice, J. (1974). Little jiffy, mark IV. Educational and
Psychological Measurement, 34, 111-117.

Mardia, K. V. (1970). Measures of multivariate skewness and kurtosis with
applications. Biometrika, 57, 519-530.

Mardia, K. V. (1974). Applications of some measures of multivariate skewness
and kurtosis in testing normality and robustness studies. Sankhya B, 36, 115-128.

Pison, G., Van Aelst, S. & Willems, G. (2002). Small sample corrections for LTS
and MCD. Metrika, 55, 111-123.

Rhemtulla, M., Brosseau-Liard, P. E. & Savalei, V. (2012). When can
categorical variables be treated as continuous? A comparison of robust continuous
and categorical SEM estimation methods under suboptimal conditions. Psychological
Methods, 17, 354-373.

Rousseeuw, P. J. & Van Driessen, K. (1999). A fast algorithm for the minimum
covariance determinant estimator. Technometrics, 41, 212-223.
}
\usage{
efa_screen(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  mcd_alpha = 0.5,
  outlier_cutoff = 0.975,
  seed = NULL
)
}
\arguments{
\item{x}{data.frame or matrix. Raw data, or a correlation matrix. Needs at least three
variables, none of which is a perfect linear combination of the others.}

\item{N}{numeric. The number of observations. Set this only when you supply a
correlation matrix; it is needed for Bartlett's test of sphericity and is taken from
the data automatically when you supply raw data. Default is \code{NA}.}

\item{use}{character. How to handle missing values in raw data. For
\code{cor_method = "pearson"}, \code{"spearman"}, or \code{"kendall"} this is passed to
\code{\link[stats:cor]{stats::cor()}}. For \code{"poly"} or \code{"tetra"} the same rule is applied to the raw data
before the correlations are estimated; \code{"all.obs"} and \code{"everything"} then stop with
an error on any missing value, instead of returning \code{NA} correlations. Default is
\code{"pairwise.complete.obs"}.}

\item{cor_method}{character. How to compute correlations from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (via \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} / \code{"tetra"} for
polychoric / tetrachoric correlations of ordinal or binary data. A Spearman or Kendall
correlation matrix is screened on its own scale, not converted to look like a Pearson
correlation matrix; Kendall's tau in particular measures something different from a
Pearson correlation, not just a rescaled version of it. Default is \code{"pearson"}.}

\item{mcd_alpha}{numeric. The proportion of cases used to build the robust outlier
estimate, between 0.5 and 1. The default, \code{0.5}, is the most robust choice; a larger
value uses more of the data but resists outliers less well. Used only with raw data.}

\item{outlier_cutoff}{numeric. The probability used to set the cutoff for flagging a
multivariate outlier, between 0.5 and 0.9999. Default is \code{0.975}. Used only with raw
data.}

\item{seed}{integer. A seed for the random subsets used by the outlier detection, so
the result is reproducible. Does not affect your random-number generator elsewhere.
Default is \code{NULL}. Used only with raw data.}
}
\value{
An object of class \code{efa_screen}, a list containing:
\item{kmo}{A list with the overall KMO (\code{KMO}) and the per-variable KMO (\code{KMO_i}).}
\item{bartlett}{A list with Bartlett's chi-square statistic (\code{chisq}), its \code{p_value},
and its degrees of freedom (\code{df}); \code{chisq} and \code{p_value} are \code{NA} when \code{N} was too
small for the correction. \code{NULL} when \code{N} is unavailable.}
\item{determinant}{The determinant of the correlation matrix.}
\item{condition}{The condition number of the correlation matrix (largest eigenvalue
over smallest).}
\item{smc}{The per-variable squared multiple correlations.}
\item{per_item}{A data frame with one row per variable (row names are the variable
names): \code{variance}, \code{missing} (percentage), \code{smc}, \code{kmo_i}, and \code{flags} (any
sparse/empty-category issues). \code{NULL} when a correlation matrix is supplied instead
of raw data.}
\item{normality}{A list with \code{mardia} (skewness \code{skewness}, \code{skewness_df},
\code{skewness_p}, kurtosis \code{kurtosis} and \code{kurtosis_p}, and the underlying \code{b1p}/\code{b2p}),
\code{hz} (the Henze-Zirkler \code{statistic} and its \code{p_value}), and \code{n_complete} (the number
of complete cases used). \code{NULL} without raw data, or a note explaining why when the
complete-case data cannot support the tests.}
\item{outliers}{A list with \code{distances} (each complete case's robust distance, named by
its row number), \code{cutoff} (the flagging threshold, on the same scale as \code{distances}),
\code{flagged} (the row numbers exceeding \code{cutoff}), \code{center} and \code{cov} (the robust
location and scatter), \code{method} (\code{"mcd"} or the \code{"classical"} fallback),
\code{fallback_reason} (why the robust estimate was unavailable, when \code{method} is
\code{"classical"}), and \code{n_complete}. \code{NULL} without raw data, or a note explaining why
when no covariance can be formed.}
\item{categories}{A named list with the response-category counts for each categorical
variable (in category order); \code{NA} for a variable treated as continuous. \code{NULL}
without raw data.}
\item{note}{Explains why the raw-data diagnostics (\code{per_item}, \code{normality},
\code{outliers}, \code{categories}) are missing, when a correlation matrix is supplied instead
of raw data. \code{NULL} when raw data are supplied.}
\item{settings}{The settings used: \code{N}, \code{n_obs} (rows in the raw data supplied, \code{NA}
for a correlation-matrix input), \code{use}, \code{cor_method}, \code{mcd_alpha}, \code{outlier_cutoff},
and \code{seed}.}
}
\description{
Checks whether your data are suitable for exploratory factor analysis (EFA). From a
correlation matrix or raw data, it reports the Kaiser-Meyer-Olkin (KMO) measure of
sampling adequacy, Bartlett's test of sphericity, the determinant and condition number
of the correlation matrix, and each variable's squared multiple correlation (SMC). When
you supply raw data, it also reports each variable's variance and percentage of missing
values, category counts for categorical variables, tests of multivariate normality, and
multivariate outliers.
}
\details{
The diagnostics are computed from the analysis correlation matrix \eqn{R}:
\describe{
\item{KMO}{The Kaiser-Meyer-Olkin measure of sampling adequacy (Kaiser, 1970; Kaiser
& Rice, 1974), overall and for each variable; see \code{\link[=efa_kmo]{efa_kmo()}}. It shows how much
common variance your variables share. Higher values are better; a common rule of
thumb treats values below .50 as unacceptable.}
\item{Bartlett}{Bartlett's (1951) test of sphericity: the likelihood-ratio test of
whether the correlation matrix is an identity matrix, i.e., whether your variables
correlate with each other at all; see \code{\link[=efa_bartlett]{efa_bartlett()}}. A significant result
supports doing a factor analysis. The test needs the sample size \code{N}; without it,
this diagnostic is skipped with a warning and \verb{$bartlett} is \code{NULL}. If \code{N} is too
small relative to the number of variables, the statistic is \code{NA}, also with a
warning.}
\item{Determinant}{The determinant of \eqn{R}, reported as a number only. It falls as
you add variables even when the variables are not collinear, so a fixed cut-off on
it (such as the 0.00001 often quoted from Field, 2018) says more about how many
variables you have than about your data. Use the condition number instead.}
\item{Condition number}{The ratio of the largest to the smallest eigenvalue of
\eqn{R}. Its square root, the condition index, is the collinearity diagnostic of
Belsley, Kuh & Welsch (1980); it drives the printed report and its recommendation.
An index of 10 or less is rarely of interest. An index above 30 flags a near linear
dependency: two or more variables that together carry much the same information.
An index between the two is not negligible, but it stays below the value that
flags a dependency. Belsley (1991) gives 30 as one example value and calls the
choice of a cut-off "somewhat of an art form", so the report grades an index above
30 by its position on the scale 1, 3, 10, 30, 100, 300, 1000: moderate (30 to 100),
strong (100 to 300), or very strong (above 300). These values come from regression
diagnostics on data that are not centred, but a correlation matrix is centred, so
use them as a guide and not as a test.}
\item{SMC}{The squared multiple correlation of each variable with all the others. A
low value flags a variable that has little in common with the rest of your set.}
\item{Variance and missing data}{For raw data: each variable's variance (over its
available values) and percentage of missing values, computed from every row you
supplied. These missing-value percentages explain why the correlation matrix's
sample size (\code{N}) can be smaller than the number of rows in your data. Ordered-factor
columns are recoded to integer levels first, so \code{variance} reflects those codes.}
\item{Categories}{For raw data: for each variable with fewer than ten distinct values
(treated as categorical), the count of responses in each category. A category with
fewer than five responses is flagged as sparse, and an unused category between the
smallest and largest response is flagged as empty. As a rule of thumb, items with
fewer than five response categories are better analysed with \code{cor_method = "poly"}
or \code{"tetra"} than with Pearson correlations (Rhemtulla et al., 2012).}
\item{Multivariate normality}{For raw data, using only complete cases: two tests of
multivariate normality, Mardia's (1970) test of skewness and kurtosis and the
Henze-Zirkler (1990) test. A small p-value on either test suggests your data depart
from a multivariate normal distribution, a reason to prefer a robust or ordinal
method over normal-theory maximum likelihood. In a very small sample the kurtosis
statistic is \code{NA}. The Henze-Zirkler p-value is not available with more than about
50 to 60 variables; its test statistic is still reported.}
\item{Outliers}{For raw data, using only complete cases: multivariate outliers, found
from a robust estimate of each case's distance from the centre of your data (the
minimum covariance determinant method; Rousseeuw & Van Driessen, 1999). A flagged
case is unusually far from the rest of your sample. When there are too few complete
cases, the variables are too collinear, or too many cases share identical answers,
a plain (non-robust) distance is used instead, with a warning explaining why.}
}
}
\examples{
# From a correlation matrix (supply N for Bartlett's test of sphericity)
efa_screen(test_models$baseline$cormat, N = 500)

# From raw data (N is taken from the data; the seed makes the outlier
# diagnostics reproducible)
efa_screen(GRiPS_raw, seed = 1)

}
\seealso{
\code{\link[=efa_kmo]{efa_kmo()}} and \code{\link[=efa_bartlett]{efa_bartlett()}} for the individual suitability measures, and
\code{\link[=efa_retain]{efa_retain()}} for factor retention criteria.

Other factor analysis suitability:
\code{\link[=efa_bartlett]{efa_bartlett()}},
\code{\link[=efa_kmo]{efa_kmo()}},
\code{\link[=print.efa_screen]{print.efa_screen()}}
}
\concept{factor analysis suitability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{HULL}
\alias{HULL}
\title{Hull method}
\usage{
HULL(
  x,
  N = NA,
  n_fac_theor = NA,
  method = c("PAF", "ULS", "ML"),
  gof = c("CAF", "CFI", "RMSEA"),
  eigen_type = c("SMC", "PCA", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_datasets = 1000,
  percent = 95,
  decision_rule = c("means", "percentile", "crawford"),
  n_factors = 1,
  ...
)
}
\arguments{
\item{x}{matrix or data.frame. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. Number of cases in the data. This is passed to \link{efa_parallel}.
Only has to be specified if x is a correlation matrix, otherwise it is determined
based on the dimensions of x.}

\item{n_fac_theor}{numeric. Theoretical number of factors to retain. One plus
the larger of this number and the number of factors suggested by \link{efa_parallel}
is used as the upper bound \emph{J} of factors to extract in the Hull method.}

\item{method}{character. The estimator to use; passed to \code{\link[=efa_hull]{efa_hull()}} as its
\code{estimator} argument. One of \code{"PAF"}, \code{"ULS"}, or \code{"ML"}.}

\item{gof}{character. The goodness of fit index to use. Either \code{"CAF"},
\code{"CFI"}, or \code{"RMSEA"}, or any combination of them.
With the \code{"PAF"} estimator, only
the CAF can be used as goodness of fit index. For details on the CAF, see
Lorenzo-Seva, Timmerman, and Kiers (2011).}

\item{eigen_type}{character. On what the eigenvalues should be found in the
parallel analysis. Can be one of \code{"SMC"}, \code{"PCA"}, or \code{"EFA"}.
If using  \code{"SMC"} (default), the diagonal of the correlation matrices is
replaced by the squared multiple correlations (SMCs) of the indicators. If
using  \code{"PCA"}, the diagonal values of the correlation
matrices are left to be 1. If using  \code{"EFA"}, eigenvalues are found on the
correlation  matrices with the final communalities of an EFA solution as
diagonal. This is passed to  \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is \code{"pairwise.complete.obs"}.}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{HULL} derives its factor-search bound from an internal parallel analysis
against continuous reference data.
Default is  \code{"pearson"}.}

\item{n_datasets}{numeric. The number of datasets to simulate. Must be at
least 1. Default is 1000. This is passed to \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{percent}{numeric. The percentile to take from the simulated eigenvalues.
Default is 95. This is passed to \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{decision_rule}{character. Which rule to use to determine the number of
factors to retain. Default is \code{"means"}, which will use the average
simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
in percent. \code{"crawford"} uses the 95th percentile for the first factor
and the mean afterwards (based on Crawford et al, 2010). This is passed to \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{n_factors}{numeric. Number of factors to extract if  \code{"EFA"} is
included in \code{eigen_type}. Default is 1. This is passed to
\code{\link[=efa_parallel]{efa_parallel()}}.}

\item{...}{Further arguments passed on to the \code{\link[=efa_fit]{efa_fit()}} fits, including the
estimation tuning knobs (\code{type}, \code{init_comm}, \code{criterion}, \code{criterion_type},
\code{max_iter}, \code{abs_eigen}, \code{start_method}), which are repacked into an
\code{\link[=estimate_control]{estimate_control()}} object so that they tune the fits exactly as they always did.
The estimator is selected with \code{method}.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_hull]{efa_hull()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{HULL()} has been superseded by \code{\link[=efa_hull]{efa_hull()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_hull]{efa_hull()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_loadings.R
\name{print.efa_loadings}
\alias{print.efa_loadings}
\alias{format.efa_loadings}
\title{Print a loading matrix}
\usage{
\method{print}{efa_loadings}(x, ...)

\method{format}{efa_loadings}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  h2 = NULL,
  color = TRUE,
  name_style = c("truncate", "abbreviate", "full"),
  max_factor_name_length = NULL,
  max_factors_per_block = NULL,
  sort_loadings = c("none", "primary", "clustered"),
  legend = FALSE,
  ...
)
}
\arguments{
\item{x}{a loading matrix of class \code{efa_loadings} (or the legacy \code{LOADINGS}).}

\item{...}{additional arguments passed to print or format}

\item{cutoff}{numeric. The value at or above which loadings are emphasized;
default is .3.}

\item{digits}{numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
to round the loadings to (default is 3).}

\item{max_name_length}{numeric. The maximum length of the variable names to
display. Everything beyond this will be cut from the right unless
\code{name_style = "abbreviate"} or \code{name_style = "full"} is used. Cutting never
leaves two variables sharing a row label: if it would (as for items with a
long common prefix), the names are abbreviated instead, and numbered if
needed.}

\item{h2}{numeric. Vector of communalities to print. If named and \code{x}
has row names, names are used to align communalities to rows.}

\item{color}{logical. Whether to apply console styling using \pkg{cli}.
Default is \code{TRUE}.}

\item{name_style}{character. How to shorten variable names longer than
\code{max_name_length}. \code{"truncate"} cuts names from the right,
\code{"abbreviate"} uses \code{\link[base:abbreviate]{base::abbreviate()}}, and
\code{"full"} prints full names.}

\item{max_factor_name_length}{numeric or \code{NULL}. Optional maximum length
of factor names. If \code{NULL}, factor names are not shortened.}

\item{max_factors_per_block}{numeric or \code{NULL}. Maximum number of factor
columns to print per block. If \code{NULL}, the number is chosen from the
console width.}

\item{sort_loadings}{character. Optional row sorting. \code{"none"} preserves
the input order, \code{"primary"} groups rows by the factor with the largest
absolute loading, and \code{"clustered"} additionally sorts within each
factor by the size of the primary loading.}

\item{legend}{logical. Whether to append a short explanation of the styling.
Default is \code{FALSE} for standalone loading matrices. The legend is only printed
when the styling it describes is actually rendered (\code{color = TRUE} and a
colour-capable console); in plain output it is omitted and this argument has
no effect.}
}
\value{
\code{print()} returns its argument \code{x} invisibly; it is
\code{cat(format(x, ...), sep = "\\n")} followed by a blank line for console
spacing. \code{format()} returns a character vector with the table lines (styled
to the active console theme; plain when colours are disabled).
}
\description{
Print a loading matrix
}
\details{
The method prints a loading matrix in a compact, console-oriented table.
Loadings with absolute value greater than or equal to \code{cutoff} are emphasized,
smaller loadings are de-emphasized, and Heywood-relevant communality/
uniqueness values are marked when \code{h2} is supplied. Long variable names can
be truncated, abbreviated, or printed in full. If the matrix has many factor
columns, the table is split into column blocks so that the output remains
readable in narrower consoles.

If \code{h2} is named and \code{x} has row names, \code{h2} is matched to the row names of
\code{x} before any optional row sorting is applied. If \code{x} has no row names, a
named \code{h2} vector is used in the supplied order.
}
\examples{
EFAtools_PAF <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                        estimator = "PAF", rotation = "promax")
EFAtools_PAF

# format() returns the same lines as a character vector:
writeLines(format(EFAtools_PAF$rot_loadings))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KMO.R
\name{KMO}
\alias{KMO}
\title{Kaiser-Meyer-Olkin criterion}
\source{
Kaiser, H. F. (1970). A second generation little jiffy. Psychometrika,
35, 401-415.

Kaiser, H. F. & Rice, J. (1974). Little jiffy, mark IV. Educational
and Psychological Measurement, 34, 111-117.

Cureton, E. E. & D'Augustino, R. B. (1983). Factor analysis: An
 applied approach. Hillsdale, N.J.: Lawrence Erlbaum Associates, Inc.
}
\usage{
KMO(
  x,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall")
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Passed to \code{\link[stats:cor]{stats::cor}}.
Default is "pearson".}
}
\value{
A list containing
\item{KMO}{Overall KMO.}
\item{KMO_i}{KMO for each variable.}
\item{settings}{A list of the settings used.}
}
\description{
This function computes the Kaiser-Meyer-Olkin (KMO) criterion overall and for
each variable in a correlation matrix. The KMO represents the degree to
which each observed variable is predicted by the other variables in the
dataset and with this indicates the suitability for factor analysis.
}
\details{
Kaiser (1970) proposed this index, originally called measure of
sampling adequacy (MSA), that indicates how near the inverted correlation
matrix \eqn{R^{-1}} is to a diagonal matrix \eqn{S} to determine a given
correlation matrix's (\eqn{R}) suitability for factor analysis.
The index is
\deqn{KMO = \frac{\sum\limits_{i<j}\sum r_{ij}^2}{\sum\limits_{i<j}\sum r_{ij}^2 + \sum\limits_{i<j}\sum q_{ij}^2}}
with \eqn{Q = SR^{-1}S} and S = \eqn{(diag R^{-1})^{-1/2}} where
\eqn{\sum\limits_{i<j}\sum r_{ij}^2} is the sum of squares of the upper
off-diagonal elements of \eqn{R} and \eqn{\sum\limits_{i<j}\sum q_{ij}^2} is the
sum of squares of the upper off-diagonal elements of \eqn{Q} (see also Cureton & D'Augustino, 1983).

So KMO varies between 0 and 1, with larger values indicating higher suitability
for factor analysis. Kaiser and Rice (1974) suggest that KMO should at least
exceed .50 for a correlation matrix to be suitable for factor analysis.

This function was heavily influenced by the \code{\link[psych:KMO]{psych::KMO}}
function.

See also \code{\link{BARTLETT}} for another test of suitability for factor
analysis.

The \code{KMO} function can also be called together with the
\code{\link{BARTLETT}} function and with factor retention criteria in the
 \code{\link{N_FACTORS}} function.
}
\examples{
KMO(test_models$baseline$cormat)
}
\seealso{
\code{\link{BARTLETT}} for another measure to determine
suitability for factor analysis.

\code{\link{N_FACTORS}} as a wrapper function for this function,
\code{\link{BARTLETT}} and several factor retention criteria.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.rotate_bifactor_oblq}
\alias{.rotate_bifactor_oblq}
\title{Oblique bifactor factor rotation}
\usage{
.rotate_bifactor_oblq(
  L,
  eps = 1e-05,
  normalize = TRUE,
  random_starts = 0L,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1,
  screen_keep = 5L,
  triage_maxit = 25L,
  triage_improve_tol = 0
)
}
\arguments{
\item{L}{Numeric matrix. The unrotated loading matrix (variables by factors).}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient norm.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization before
rotation and reverse it afterwards.}

\item{random_starts}{Integer scalar. Number of additional random orthogonal starts.}

\item{maxit}{Integer scalar. Maximum number of projected-gradient updates.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving attempts after
the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient update.}

\item{screen_keep}{Integer scalar. Number of screened random starts retained for
triage optimization.}

\item{triage_maxit}{Integer scalar. Number of short optimization iterations used in
the triage stage.}

\item{triage_improve_tol}{Numeric scalar. Relative improvement required for a triaged
start to be promoted to full optimization.}
}
\value{
A named list with the rotated loadings, the transformation matrix \code{Th}
(with \code{L \%*\% t(solve(Th))} reproducing the rotated loadings), the factor correlation
matrix \code{Phi} (\code{t(Th) \%*\% Th}), the attained criterion value, and the convergence and
validity flags. The list additionally reports the criterion value reached
at each optimized start in \code{all_values}, with a per-start convergence flag in \code{all_converged}.
}
\description{
Rotate a loading matrix obliquely under the Jennrich-Bentler bifactor criterion using a
gradient-projection optimizer along the oblique (column-normalized) manifold.
}
\details{
The criterion value \code{f} and its gradient \code{dQ/dL} at the rotated loadings
\code{L = A \%*\% solve(t(T))} define the search; the engine maps the gradient to the
transformation \code{T} on the manifold \code{diag(t(T) \%*\% T) = 1}, projects it onto the tangent
space, performs a non-monotone line search, and retracts back onto the manifold by
column normalization. The first factor is treated as a general factor and is exempt from the
penalty; the criterion measures the between-group-factor cross-products of the squared
loadings, so it is minimized when each variable loads on the general factor plus at most one
group factor. The criterion is prone to local minima, so additional random starts are
recommended.

Additional random starts may be requested. To bound runtime the solver screens each random
start by its objective, runs a short triage optimization on the best-screened starts, and
fully optimizes only those that improve on the current incumbent by at least
\code{triage_improve_tol}.
}
\references{
Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
software for arbitrary rotation criteria in factor analysis. \emph{Educational and
Psychological Measurement}, 65, 676-696.

Jennrich, R. I., & Bentler, P. M. (2011). Exploratory bi-factor analysis. \emph{Psychometrika},
76, 537-549.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.stat_over_list}
\alias{.stat_over_list}
\title{Calculate statistics for a list of matrices}
\usage{
.stat_over_list(alist, stat)
}
\arguments{
\item{alist}{A list of sub-lists, typically a list of \eqn{m} matrices from
\code{"EFA"}, where \eqn{m} is the number of imputations passed to
\code{\link{EFA_POOLED}}.}

\item{stat}{A function, e.g. \code{mean} or \code{sd}.}
}
\value{
A matrix with the aggregated results
}
\description{
Calculate statistics for a list of matrices
}
\author{
Andreas Soteriades

Given a list of matrices, this function calculates user-supplied statistics
(e.g. mean, median) over the matrices. This function is useful for averaging
results from \code{\link{EFA}} (e.g. loadings) over all permutation runs in
\code{\link{EFA_POOLED}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.nest_sym}
\alias{.nest_sym}
\title{Get reference values for nest.}
\usage{
.nest_sym(nf, N, M, nreps = 1000L)
}
\arguments{
\item{nf}{numeric. Current number of factors.}

\item{N}{numeric. Number of cases / observations in dataset.}

\item{M}{numeric matrix. A transposed matrix of cbind(loadings, diag(sqrt(uniquenesses))).}

\item{nreps}{numeric. Number of datasets to create.}
}
\description{
Function called from within NEST so usually no call to this is needed by the user.
Provides a C++ implementation of the NEST simulation procedure
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_compare.R
\name{print.efa_compare}
\alias{print.efa_compare}
\alias{format.efa_compare}
\title{Print and format an efa_compare object}
\usage{
\method{print}{efa_compare}(x, ...)

\method{format}{efa_compare}(
  x,
  digits = NULL,
  m_red = NULL,
  range_red = NULL,
  round_red = NULL,
  print_diff = NULL,
  ...
)
}
\arguments{
\item{x}{An object of class \code{efa_compare} (output from \code{\link[=efa_compare]{efa_compare()}}).}

\item{...}{Passed from \code{print()} to \code{format()}; not otherwise used.}

\item{digits, m_red, range_red, round_red, print_diff}{Display controls, documented
in \code{\link[=efa_compare]{efa_compare()}}. Each defaults to \code{NULL}, meaning the value \code{efa_compare()}
recorded in \code{x$settings} is used; supplying one overrides it for this call
only, so the comparison need not be recomputed to change the printed report.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows a summarised output of the \code{\link[=efa_compare]{efa_compare()}} function: the mean
(with its range), median, and root mean squared distance (RMSE) of the
differences, the number of decimals to which all numbers agree, the minimum
number of decimals provided, and (for matrices) the number of differing
indicator-to-factor correspondences, followed (optionally) by the table of
elementwise differences. \code{format()} assembles the same report and returns it
as a character vector; \code{print()} is \code{cat(format(x), sep = "\\n")}.
The lines follow the active console theme, so they are plain when colours are
disabled (for example when captured into a file or stripped with
\code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\details{
The line reporting the minimum number of decimals provided is shown only when it
carries information: two ordinary double matrices carry the full double precision,
for which the count is uninformative and the line is omitted.

The summary statistics are absolute differences, so they carry no direction. The
elementwise differences are signed, and the table is headed by the direction of the
subtraction, named with the \code{x_labels} recorded by \code{\link[=efa_compare]{efa_compare()}} (\code{"x"} and \code{"y"} by
default): a negative cell means the first solution is the lower of the two there.
}
\examples{
# A type SPSS EFA to mimick the SPSS implementation
EFA_SPSS_5 <- efa_fit(IDS2_R, n_factors = 5,
                      estimate_control = estimate_control(type = "SPSS"),
                      rotate_control = rotate_control(type = "SPSS"))

# A type psych EFA to mimick the psych::fa() implementation
EFA_psych_5 <- efa_fit(IDS2_R, n_factors = 5,
                       estimate_control = estimate_control(type = "psych"),
                       rotate_control = rotate_control(type = "psych"))

# compare the two
comp <- efa_compare(EFA_SPSS_5$unrot_loadings, EFA_psych_5$unrot_loadings,
                    x_labels = c("SPSS", "psych"))
comp

# format() returns the same lines as a character vector:
writeLines(format(comp))

# the display settings can be changed without recomputing the comparison:
print(comp, digits = 2, print_diff = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GRiPS_raw_doc.R
\docType{data}
\name{GRiPS_raw}
\alias{GRiPS_raw}
\title{GRiPS_raw}
\format{
A data.frame with 810 rows (participants) and 8 columns, one per GRiPS
item. Each item is a self-report indicator of general risk propensity, labelled
by a keyword from the item:
\describe{
\item{fun}{(numeric) - Risk-taking makes life more fun.}
\item{friends}{(numeric) - Friends would describe the respondent as a risk taker.}
\item{enjoy}{(numeric) - Enjoyment of taking risks.}
\item{hurt}{(numeric) - Willingness to take a risk even if it might hurt.}
\item{part}{(numeric) - Risk-taking as an important part of life.}
\item{commonly}{(numeric) - Commonly takes risks.}
\item{chances}{(numeric) - Belief in taking chances.}
\item{attracted}{(numeric) - Attracted, rather than scared, by risk.}
}
}
\source{
Zhang, D. C., Highhouse, S., & Nye, C. D. (2019). Development and validation of the general risk propensity scale (GRiPS). Journal of Behavioral Decision Making, 32, 152–167. doi: 10.1002/bdm.2102

Steiner, M., & Frey, R. (2020). Representative design in psychological assessment: A case study using the Balloon Analogue Risk Task (BART). PsyArXiv Preprint. doi:10.31234/osf.io/dg4ks
}
\usage{
GRiPS_raw
}
\description{
A data.frame containing responses to the General Risk Propensity Scale (GRiPS, Zhang,
Highhouse & Nye, 2018) of 810 participants of Study 1 of Steiner and Frey (2020).
The original data can be accessed via \url{https://osf.io/kxp8t/}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.HULL.R
\name{plot.HULL}
\alias{plot.HULL}
\title{Plot HULL object}
\usage{
\method{plot}{HULL}(x, ...)
}
\arguments{
\item{x}{list of class HULL. An output from the \code{\link{HULL}} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \code{\link{HULL}} function
}
\examples{
\donttest{
x <- HULL(test_models$baseline$cormat, N = 500, method = "ML")
plot(x)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_simulate.R
\name{efa_simulate}
\alias{efa_simulate}
\title{Simulate data from a common-factor population model}
\usage{
efa_simulate(
  N = NULL,
  Lambda = NULL,
  Phi = NULL,
  Psi = NULL,
  R = NULL,
  model_error = c("CB", "TKL", "WB", "none"),
  target_rmsea = NULL,
  target_cfi = NULL,
  marginals = c("normal", "empirical", "VM", "IG"),
  marginal_data = NULL,
  n_factors = NULL,
  skewness = NULL,
  kurtosis = NULL,
  force_pd = FALSE,
  categories = NULL,
  match = NULL,
  missing = c("none", "MCAR", "MAR", "MNAR"),
  missing_prop = NULL,
  missing_strength = NULL,
  missing_predictor = NULL,
  missing_vars = NULL,
  n_datasets = 1L,
  seed = NULL,
  return_pop = FALSE
)
}
\arguments{
\item{N}{numeric. Number of cases (rows) to draw per dataset. Required unless
\code{return_pop = TRUE}.}

\item{Lambda}{matrix. A \code{p} by \code{m} matrix of factor loadings. Supply this
(optionally with \code{Phi} and \code{Psi}) instead of \code{R} to build the population
from a factor model.}

\item{Phi}{matrix. The \code{m} by \code{m} factor intercorrelation matrix. Only used
with \code{Lambda}. Default is \code{NULL}, in which case the factors are orthogonal
(an identity matrix).}

\item{Psi}{numeric vector or matrix. The unique variances: either a length-\code{p}
vector or a \code{p} by \code{p} matrix (added as the residual covariance). Only used
with \code{Lambda}. Default is \code{NULL}, in which case the unique variances that
standardize the population to a correlation matrix are used.}

\item{R}{matrix. A \code{p} by \code{p} population correlation matrix to draw from
directly. Supply this instead of \code{Lambda}/\code{Phi}/\code{Psi}. It must have a unit
diagonal; standardize a covariance matrix with \code{\link[stats:cov2cor]{stats::cov2cor()}} first.}

\item{model_error}{character. The method used to perturb the population so the factor model
fits it imperfectly ("model error"): one of \code{"CB"} (Cudeck-Browne, the default), \code{"TKL"}
(Tucker-Koopman-Linn), \code{"WB"} (Wu-Browne), or \code{"none"}. Model error is only applied when a
target is supplied in \code{target_rmsea} or \code{target_cfi}; without one the population is exact,
whatever \code{model_error}. Only used with a factor-model population (\code{Lambda}). Note that
\code{\link[=efa_power]{efa_power()}} defaults to \code{"TKL"} instead, because its minor common factors degrade
factor recovery as well as the fit, so passing the same target to the two functions
perturbs the population differently unless \code{model_error} is set explicitly.}

\item{target_rmsea}{numeric. The population RMSEA the factor model should have relative to
the perturbed population, a single number strictly in \verb{(0, 1)}. Supplying it activates model
error. Simulating from an exact population overstates recovery, so a realistic value (around
\code{0.05}) is recommended for simulation studies (MacCallum, 2003). Default is \code{NULL} (an
exact population; do not pass \code{0}). Required for \code{"CB"} and \code{"WB"}; optional for \code{"TKL"}.}

\item{target_cfi}{numeric. Only used with \code{model_error = "TKL"}: the population CFI to
target, a single number strictly in \verb{(0, 1)}, on its own or together with \code{target_rmsea}
(TKL then trades the two off). Default is \code{NULL} (do not pass \code{1}). \code{"CB"} and \code{"WB"} target
the RMSEA only.}

\item{marginals}{character. The marginal distribution of the drawn data: one
of \code{"normal"} (the default), which draws normal marginals; \code{"empirical"},
which reproduces the population correlation while preserving the empirical
marginals supplied in \code{marginal_data}; or \code{"VM"} (Vale-Maurelli) and \code{"IG"}
(independent generator), which draw non-normal marginals with the target
\code{skewness} and \code{kurtosis}.}

\item{marginal_data}{matrix or data frame. Only used with
\code{marginals = "empirical"}, where it is required: a data set with one numeric
column per variable (\code{p} columns), each with at least two distinct values,
whose per-column distributions are resampled as the marginals of the drawn
data. Its correlations are ignored. Default is \code{NULL}.}

\item{n_factors}{numeric. Only used with \code{marginals = "empirical"}: the number
of factors the rank-matching reproduction fits. Default is \code{NULL}, in which
case it is the number of columns of \code{Lambda} when the population is built
from a factor model; it must be given when the population is supplied via \code{R}.}

\item{skewness}{numeric. Only used with \code{marginals = "VM"} or \code{"IG"}: the
target marginal skewness, as a single value applied to every variable or a
length-\code{p} vector. Default is \code{NULL} (0, a symmetric marginal). At least one
of \code{skewness} or \code{kurtosis} must be given for these marginals.}

\item{kurtosis}{numeric. Only used with \code{marginals = "VM"} or \code{"IG"}: the
target marginal excess kurtosis (0 for a normal marginal), as a single value
applied to every variable or a length-\code{p} vector. Default is \code{NULL} (0).}

\item{force_pd}{logical. Used with \code{marginals = "VM"} and with Cudeck-Browne model error
(\code{model_error = "CB"}), where \code{FALSE} (the default) rejects a population or intermediate
matrix that is not positive definite. What \code{TRUE} accepts instead differs by path, in
both cases with a warning: for \code{marginals = "VM"} the intermediate correlation matrix is
projected to the nearest correlation matrix (via \code{\link[psych:cor.smooth]{psych::cor.smooth()}}), and the returned
\code{population} becomes the one the projected draw attains; for \code{model_error = "CB"} there is
no such projection -- the closest positive-definite perturbation is kept instead, whose
achieved RMSEA is \strong{below} the target. Has no effect for the \code{"TKL"} or \code{"WB"} methods.}

\item{categories}{numeric or list. Requests ordinal output by discretizing each
variable into ordered categories. Either a count of equally probable categories
(a single value applied to every variable or a length-\code{p} vector), or a
length-\code{p} list of numeric vectors giving the marginal category proportions per
variable (each strictly positive and summing to 1). Default is \code{NULL}, which
returns the continuous data.}

\item{match}{character. Only used with \code{categories}: an assertion about how the
categorization relates to the population correlation. With a normal latent, cutting
at the normal-scale thresholds already leaves the population polychoric correlation
of the categorized data equal to the target correlation, so both values compute the
same thresholds and produce identical data whenever both are legal. \code{"thresholds"}
(the default) also cuts the \code{"VM"} and \code{"IG"} draws, whose ordinal Pearson and
polychoric correlations then both depart from the population; \code{"polychoric"} states
that the polychoric match is required and therefore rejects non-normal marginals.
Not available with \code{marginals = "empirical"}. Default is \code{NULL}
(\code{"thresholds"} when \code{categories} is set).}

\item{missing}{character. An optional missing-data mechanism to impose on the drawn
data: one of \code{"none"} (the default, complete data), \code{"MCAR"} (missing completely at
random), \code{"MAR"} (missing at random, depending on another variable), or \code{"MNAR"}
(missing not at random, depending on the variable's own value). Introduced values
become \code{NA}. By default every variable is holed, so under \code{"MAR"} each variable's
predictor is itself subject to missingness and the mechanism is MAR given the
\emph{complete} data rather than ignorably MAR; \code{missing_vars} and \code{missing_predictor}
together give an ignorable design (see Details).}

\item{missing_prop}{numeric. Only used when \code{missing} is not \code{"none"}, where it is
required: the target marginal proportion of missing values per variable, a single
number strictly between 0 and 1. This is the expected rate; the realized rate of a
given draw varies around it.}

\item{missing_strength}{numeric. Only used with \code{missing = "MAR"} or \code{"MNAR"}: the
slope of the logistic missingness model, setting how strongly the missing
probability depends on the predictor. Default is \code{NULL} (1, a moderate dependence);
\code{0} removes the dependence (equivalent to MCAR at the same rate) and large
magnitudes make missingness nearly deterministic.}

\item{missing_predictor}{integer or character. Only used with \code{missing = "MAR"}:
which variable drives each holed variable's missingness, as one column index or
variable name per variable in \code{missing_vars} (so one per variable by default), in
that order. A variable cannot be its own predictor; if the same predictor is reused
for several \code{missing_vars}, it must not itself be one of them. Default is
\code{NULL}, in which case each variable's missingness is driven by the next variable
cyclically (so variable order matters; supply this explicitly when the order is
arbitrary). Predictors outside \code{missing_vars} are fully observed (see Details for
why this matters).}

\item{missing_vars}{integer or character. Only used when \code{missing} is not \code{"none"}:
which variables carry missing values, as column indices or variable names. Default is
\code{NULL}, in which case every variable is holed. Restricting it, together with
pointing \code{missing_predictor} at variables outside that set, is how an ignorably
MAR design is obtained (see Details).}

\item{n_datasets}{numeric. The number of datasets to draw. Default is 1. With
more than one, a list of datasets is returned.}

\item{seed}{numeric. Optional seed for reproducible draws. When supplied, the
caller's random-number stream is saved and restored, so the call leaves the
global RNG state unchanged. Default is \code{NULL} (no seeding).}

\item{return_pop}{logical. If \code{TRUE}, return only the population correlation
matrix and draw no data. Default is \code{FALSE}. Because no data are drawn, the
Vale-Maurelli intermediate correlation matrix is never formed, so its
positive-definiteness (and any \code{force_pd} projection of it) is neither checked nor
reflected in the returned population.}
}
\value{
An object of class \code{efa_simulated}, a list containing:
\item{data}{The simulated data: an \code{N} by \code{p} numeric matrix, an integer matrix of
category codes when \code{categories} is set, or a length-\code{n_datasets} list of these when
\code{n_datasets > 1}. \code{NULL} when \code{return_pop = TRUE}.}
\item{population}{The \code{p} by \code{p} population correlation matrix drawn from,
model-error-perturbed when requested. When \code{force_pd = TRUE} causes \code{marginals = "VM"}'s intermediate matrix to be projected, \code{population} is the population the
projected draw actually reaches, differing from the target by the drift the warning
reports.}
\item{model_error}{\code{NULL}, or, when model error was applied, a list of the method, the
target and achieved RMSEA/CFI, the model degrees of freedom \code{df}, and the method's
tuning parameter: \code{kappa} for \code{"CB"}, \code{v} and \code{eps} for \code{"TKL"}, \code{m} for \code{"WB"}.}
\item{settings}{The call's key arguments (\code{N}, \code{n_datasets}, \code{marginals}, \code{categories},
\code{match}, \code{missing}, \code{seed}), kept for reference and used when printing.}

Printing the object shows a compact summary.
}
\description{
Draws data from a population correlation matrix, given either directly or
built from a factor model. The population correlation is either supplied in
\code{R}, or assembled from a loading matrix \code{Lambda}, the factor intercorrelations
\code{Phi}, and the unique variances \code{Psi} as
\eqn{R = Lambda\, Phi\, Lambda' + Psi}, standardized to a correlation matrix.
\code{marginals} chooses the marginal distribution of the drawn cases, \code{categories}
discretizes them into ordered categories, \code{missing} imposes a missing-data mechanism,
and a misfit target perturbs the population with model error; see \emph{Details}.
}
\details{
Provide the population either as a ready correlation matrix in \code{R}, or through
the model components \code{Lambda}, \code{Phi}, and \code{Psi}; the two ways are mutually
exclusive. When the model components are used, \code{Phi} defaults to the identity
matrix (orthogonal factors) and \code{Psi} defaults to the unique variances that
make the population a correlation matrix (\eqn{1 - \mathrm{diag}(Lambda\, Phi\,
Lambda')}); the assembled covariance is standardized with
\code{\link[stats:cov2cor]{stats::cov2cor()}} so a non-standardized \code{Psi} still yields a correlation
matrix. With the default \code{Psi}, a factor model whose implied communalities
exceed 1 (a Heywood case) leaves no unique variance and is rejected; a \code{Psi}
you supply is instead only required to give positive variances and a
positive-semidefinite (a mathematically valid, internally consistent
correlation/covariance structure) population. Cases with normal marginals (the
default) are drawn through a matrix square root of the population correlation --
a Cholesky factor, or a symmetric eigen square root when it is singular (e.g. a
communality of exactly 1).

With \code{marginals = "empirical"}, the iterative rank-matching algorithm of Ruscio
and Kaczetow (2008) reproduces the population correlation while each variable
takes the empirical marginal distribution of the matching column of
\code{marginal_data} (resampled with replacement). Only the marginals of
\code{marginal_data} are used; its own correlations are ignored, and the drawn columns
follow the population's variables, not those of \code{marginal_data}.

With \code{marginals = "VM"} (Vale-Maurelli, 1983) or \code{"IG"} (the independent-generator
method; Foldnes & Olsson, 2016), the cases reproduce the population correlation
while carrying non-normal marginals with the target \code{skewness} and (excess)
\code{kurtosis}. The Vale-Maurelli family does not span every valid non-normal
distribution (Foldnes & Grønneberg, 2015); \code{"IG"} covers distributions \code{"VM"}
cannot. Not every (\code{skewness}, \code{kurtosis}) pair is attainable -- every distribution
needs excess kurtosis of at least \code{skewness^2 - 2}, and either method covers a smaller
region still -- so an unreachable request is rejected, as is a \code{"VM"} intermediate
correlation matrix that is not positive definite unless \code{force_pd} allows it.

With \code{categories}, the drawn data are discretized into ordered categories (an integer
code \code{1} to \code{K}) at the thresholds that reproduce the requested category proportions
(Olsson, 1979): the standard-normal quantiles for \code{marginals = "normal"}, and for
\code{marginals = "VM"} those quantiles mapped through the same Fleishman (1978) cubic the
draw uses, so the requested proportions are reproduced on the non-normal scale too. Under
\code{marginals = "IG"} the thresholds stay on the standard-normal scale while the data do
not, so the achieved proportions depart from the request systematically rather than by
sampling noise, and only the \emph{number} of categories is guaranteed; the same holds for a
\code{"VM"} variable whose Fleishman cubic is not increasing over its own thresholds and the
tails beyond them, which keeps the normal-scale thresholds and is reported with a
warning. This is more likely with strong skewness/kurtosis or very unequal category
proportions; the warning names the affected variable.
Because categorization attenuates product-moment correlations, the categorized
data's Pearson correlation is smaller in magnitude than the population correlation;
under non-normal marginals its polychoric correlation departs from the population as
well. Ordinal output is not available with \code{marginals = "empirical"}. Empty categories
left by a draw are reported with a warning, as they destabilize the polychoric
correlation and the factor analysis.

With \code{missing}, missing values are introduced into the drawn data under a chosen
mechanism (Rubin, 1976), each variable holed at a target expected rate
\code{missing_prop}. \code{"MCAR"} draws an independent mask, so missingness is unrelated to
the data. \code{"MAR"} and \code{"MNAR"} set each case's missing probability by a logistic
model of a standardized predictor: another variable for \code{"MAR"} (chosen by
\code{missing_predictor}) or the variable's own value for \code{"MNAR"}, with slope
\code{missing_strength}. The mechanism acts on the drawn (latent) values, so when
\code{categories} also discretizes the data the missingness is keyed on the underlying
value, not the category code. For \code{"MAR"} the predictor is evaluated on the complete
drawn values, so whether the mechanism is \emph{ignorably} MAR depends on which variables
carry missing values. By default every variable is holed, which leaves a variable's MAR
predictor itself missing for roughly a \code{missing_prop} fraction of the cases whose
missingness it drove. This breaks ignorability: the mechanism is then MAR conditional
on the \emph{complete} data but \strong{not} ignorable for an analyst who sees only the observed
data. As a result, estimators that are consistent under ignorable MAR -- \code{cor_method = "fiml"} in \code{\link[=efa_fit]{efa_fit()}}, or the multiple imputation behind \code{\link[=efa_mi]{efa_mi()}} -- keep a residual
bias that grows with \code{missing_prop} and \code{missing_strength}. Restricting the holed
variables with \code{missing_vars} and pointing
\code{missing_predictor} at variables outside that set makes every predictor fully observed,
which is ignorably MAR and recovers the unbiasedness those estimators are advertised
with. The returned matrix carries the \code{NA}s, which the correlation estimators handle
downstream.

With \code{model_error}, the population is perturbed away from the exact factor
structure so the \code{q}-factor model (\code{q = ncol(Lambda)}) fits it only approximately,
at a prescribed misfit; exact factor structures are unrealistic (see
\code{target_rmsea}). The perturbation is applied once to
the population, and the achieved misfit of the specified generating model is computed
with the same fit-index formulas \code{\link[=efa_fit]{efa_fit()}} uses and returned in the \code{model_error}
element. It needs a factor-model population with residual degrees of freedom and an
exact factor structure (a diagonal \code{Psi}), and is orthogonal to the marginal, ordinal,
and missing-data options. Three methods are available. \code{"CB"} (Cudeck & Browne,
1992) matches the target RMSEA to numerical precision and keeps the \code{q}-factor
model the exact minimizer (the CFI follows as a derived quantity). \code{"TKL"} (Tucker,
Koopman & Linn, 1969) adds minor common factors tuned so the achieved RMSEA -- and,
optionally, CFI -- match the target(s); with a single target the match is close, with
both it is a compromise, reported with a warning when the two cannot be reconciled.
\code{"WB"} (Wu & Browne, 2015) draws the population from
an inverse-Wishart distribution around the model-implied correlation; its calibration
applies to the \emph{best-fitting} model, so the reported misfit of the \emph{generating} model
is systematically larger than the target -- about 1.4 times for a typical 12-variable,
3-factor model. Use \code{"CB"} when the reported RMSEA must equal the target.

Replicated draws (\code{n_datasets > 1}) are generated in parallel across
replicates with \pkg{future.apply}; a parallel plan can be selected with
\code{\link[future:plan]{future::plan()}} (the default plan runs sequentially). Each replicate is
assigned its own reproducible random-number stream, so with a fixed \code{seed} the
output is identical regardless of the number of workers.
}
\examples{
# Build a population from a shipped loading pattern and factor correlations
Lambda <- population_models$loadings$baseline
Phi <- population_models$phis_3$moderate

# Draw one normal dataset of 500 cases (the data live in $data)
sim <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, seed = 42)
dim(sim$data)

# Return only the population correlation matrix
R_pop <- efa_simulate(Lambda = Lambda, Phi = Phi, return_pop = TRUE)$population

# Draw several datasets at once from a supplied correlation matrix
sims <- efa_simulate(N = 500, R = R_pop, n_datasets = 3, seed = 42)
length(sims$data)

# Reproduce the population correlation but with skewed, empirical marginals
# (here from a chi-squared source with one column per variable)
src <- matrix(rchisq(200 * nrow(Lambda), df = 3), ncol = nrow(Lambda))
dat_emp <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi,
                        marginals = "empirical", marginal_data = src, seed = 42)

# Draw skewed, leptokurtic data with the Vale-Maurelli method
dat_vm <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, marginals = "VM",
                       skewness = 1.5, kurtosis = 4, seed = 42)

# Draw five-category ordinal data whose polychoric correlation matches R
dat_ord <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi,
                        categories = 5, match = "polychoric", seed = 42)

# Draw data with 15\% missing at random, driven by a neighbouring item
dat_mar <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, missing = "MAR",
                        missing_prop = 0.15, seed = 42)
colMeans(is.na(dat_mar$data))

# An ignorably MAR design: only the first nine items are holed, each driven by one of
# the last nine, which stay complete
dat_ign <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi, missing = "MAR",
                        missing_prop = 0.15, missing_vars = 1:9,
                        missing_predictor = 10:18, seed = 42)
colMeans(is.na(dat_ign$data))

# Add realistic model error: a population the model fits with RMSEA of about .05
# (Cudeck-Browne, the default method; the achieved fit is reported)
sim_me <- efa_simulate(N = 500, Lambda = Lambda, Phi = Phi,
                       target_rmsea = 0.05, seed = 42)
sim_me$model_error$rmsea
}
\references{
Cudeck, R., & Browne, M. W. (1992). Constructing a covariance matrix
that yields a specified minimizer and a specified minimum discrepancy function
value. Psychometrika, 57(3), 357-369. \doi{10.1007/BF02295424}

Fleishman, A. I. (1978). A method for simulating non-normal
distributions. Psychometrika, 43(4), 521-532. \doi{10.1007/BF02293811}

Foldnes, N., & Grønneberg, S. (2015). How general is the Vale-Maurelli
simulation approach? Psychometrika, 80(4), 1066-1083.
\doi{10.1007/s11336-014-9414-0}

Foldnes, N., & Olsson, U. H. (2016). A simple simulation technique for
nonnormal data with prespecified skewness, kurtosis, and covariance matrix.
Multivariate Behavioral Research, 51(2-3), 207-219.
\doi{10.1080/00273171.2015.1133274}

MacCallum, R. C. (2003). 2001 Presidential Address: Working with imperfect
models. Multivariate Behavioral Research, 38(1), 113-139.
\doi{10.1207/S15327906MBR3801_5}

Olsson, U. (1979). Maximum likelihood estimation of the polychoric correlation
coefficient. Psychometrika, 44(4), 443-460. \doi{10.1007/BF02296207}

Rubin, D. B. (1976). Inference and missing data. Biometrika, 63(3), 581-592.
\doi{10.1093/biomet/63.3.581}

Ruscio, J., & Kaczetow, W. (2008). Simulating multivariate nonnormal data
using an iterative algorithm. Multivariate Behavioral Research, 43(3),
355-381. \doi{10.1080/00273170802285693}

Tucker, L. R., Koopman, R. F., & Linn, R. L. (1969). Evaluation of factor
analytic research procedures by means of simulated correlation matrices.
Psychometrika, 34(4), 421-459. \doi{10.1007/BF02290601}

Vale, C. D., & Maurelli, V. A. (1983). Simulating multivariate nonnormal
distributions. Psychometrika, 48(3), 465-471. \doi{10.1007/BF02293687}

Wu, H., & Browne, M. W. (2015). Quantifying adventitious error in a covariance
structure as a random effect. Psychometrika, 80(3), 571-600.
\doi{10.1007/s11336-015-9451-3}
}
\seealso{
\code{\link[=efa_power]{efa_power()}}, whose simulation mode draws its replicate data sets with this
function, and \code{\link[=efa_fit]{efa_fit()}} for analysing the simulated data.

Other data simulation:
\code{\link[=print.efa_simulated]{print.efa_simulated()}}
}
\concept{data simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_kgc.R
\name{efa_kgc}
\alias{efa_kgc}
\title{Kaiser-Guttman criterion}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200

Guttman, L. (1954). Some necessary conditions for common-factor analysis.
Psychometrika, 19, 149 –161. https://doi.org/10.1007/BF02289162

Kaiser, H. F. (1960). The application of electronic computers to factor
analysis. Educational and Psychological Measurement, 20, 141–151.
https://doi.org/10.1177/001316446002000116

Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
determining the number of components to retain. Psychological Bulletin, 99,
432–442. https://doi.org/10.1037/0033-2909.99.3.432
}
\usage{
efa_kgc(
  x,
  eigen_type = c("PCA", "SMC", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_factors = 1,
  estimate_control = NULL,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{eigen_type}{character. On what the eigenvalues should be found. Can be
either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
the diagonal values of the correlation matrices are left to be 1. If using
"SMC", the diagonal of the
correlation matrices is replaced by the squared multiple correlations (SMCs)
of the indicators. If using "EFA", eigenvalues are found on the correlation
matrices with the final communalities of an exploratory factor analysis
solution (default is principal axis factoring extracting 1 factor) as
diagonal. Default is \code{c("PCA", "SMC", "EFA")}, i.e. all three; \code{"EFA"} is the
only one that fits a model.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator).
Default is "pearson".}

\item{n_factors}{numeric. Number of factors to extract if "EFA" is included in
\code{eigen_type}. Default is 1.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
\code{\link[=efa_fit]{efa_fit()}} fit that provides the communalities when \code{"EFA"} is included in \code{eigen_type}.
\code{NULL} (default) uses the \code{\link[=efa_fit]{efa_fit()}} defaults. The fit is unrotated, so no rotation settings
apply.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}}. For example,
\code{estimator}, to change the estimator (PAF is default). The estimation tuning knobs are not
passed here; they live in \code{estimate_control}, and the standard-error arguments (\code{se},
\code{b_boot}, \code{ci}, \code{seed}) are not accepted because the fit is an internal step that keeps
only its communalities.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). Its main fields are:
\item{n_factors}{A named numeric vector with the suggested number of factors
for each requested eigenvalue type (\code{"PCA"}, \code{"SMC"}, and/or \code{"EFA"}).}
\item{results}{A list with one record per eigenvalue type, each holding the
eigenvalues and the retained solution used for printing and plotting.}
\item{settings}{A list of the settings used.}
}
\description{
Probably the most popular factor retention criterion. Kaiser and Guttman suggested
to retain as many factors as there are sample eigenvalues greater than 1.
This is why the criterion is also known as eigenvalues-greater-than-one rule.
}
\details{
Originally, the Kaiser-Guttman criterion was intended for the use
with principal components, hence with eigenvalues derived from the original
correlation matrix. This can be done here by setting \code{eigen_type} to
"PCA". However, it is well-known that this criterion is often inaccurate and
that it tends to overestimate the number of factors, especially for unidimensional
or orthogonal factor structures (e.g., Zwick & Velicer, 1986).

The criterion's inaccuracy in these cases is somewhat addressed if it is
applied on the correlation matrix with communalities in the diagonal, either
initial communalities estimated from SMCs (done setting \code{eigen_type} to
"SMC") or final communality estimates from an EFA (done setting \code{eigen_type}
to "EFA"; see Auerswald & Moshagen, 2019). However, although this variant
of the KGC is more accurate in some cases compared to the traditional KGC, it
is at the same time less accurate than the PCA-variant in other cases, and it
is still often less accurate than several of the other criteria available here,
such as parallel analysis (\code{\link[=efa_parallel]{efa_parallel()}}), the Hull method (\code{\link[=efa_hull]{efa_hull()}}),
the empirical Kaiser criterion (\code{\link[=efa_ekc]{efa_ekc()}}), or sequential \eqn{chi^2} model
tests (\code{\link[=efa_smt]{efa_smt()}}; see Auerswald & Moshagen, 2019). Which criteria are
informative depends on the data at hand, so rather than substituting one for
another, run several of them together and compare their suggestions.

The \code{efa_kgc} function can also be called together with other factor
retention criteria in the \code{\link[=efa_retain]{efa_retain()}} function.
}
\examples{
efa_kgc(test_models$baseline$cormat, eigen_type = c("PCA", "SMC"))
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.SCREE.R
\name{print.SCREE}
\alias{print.SCREE}
\title{Print function for SCREE objects}
\usage{
\method{print}{SCREE}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class SCREE Output from \link{SCREE} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for SCREE objects
}
\examples{
SCREE_base <- SCREE(test_models$baseline$cormat)
SCREE_base

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_hull.R
\name{efa_hull}
\alias{efa_hull}
\title{Hull method for determining the number of factors to retain}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200

Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. (2011).
The Hull method for selecting the number of common factors. Multivariate
Behavioral Research, 46(2), 340-364.
}
\usage{
efa_hull(
  x,
  N = NA,
  n_fac_theor = NA,
  estimator = c("PAF", "ULS", "ML"),
  gof = c("CAF", "CFI", "RMSEA"),
  eigen_type = c("SMC", "PCA", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_datasets = 1000,
  percent = 95,
  decision_rule = c("means", "percentile", "crawford"),
  n_factors = 1,
  estimate_control = NULL,
  ...
)
}
\arguments{
\item{x}{matrix or data.frame. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. Number of cases in the data. This is passed to \link{efa_parallel}.
Only has to be specified if x is a correlation matrix, otherwise it is determined
based on the dimensions of x.}

\item{n_fac_theor}{numeric. Theoretical number of factors to retain. One plus
the larger of this number and the number of factors suggested by \link{efa_parallel}
is used as the upper bound \emph{J} of factors to extract in the Hull method.}

\item{estimator}{character. The estimator to use. One of  \code{"PAF"},
\code{"ULS"}, or  \code{"ML"}, for principal axis factoring, unweighted
least squares, and maximum likelihood, respectively. Default is \code{"PAF"}.}

\item{gof}{character. The goodness of fit index to use. Either \code{"CAF"},
\code{"CFI"}, or \code{"RMSEA"}, or any combination of them.
With the \code{"PAF"} estimator, only
the CAF can be used as goodness of fit index. For details on the CAF, see
Lorenzo-Seva, Timmerman, and Kiers (2011).}

\item{eigen_type}{character. On what the eigenvalues should be found in the
parallel analysis. Can be one of \code{"SMC"}, \code{"PCA"}, or \code{"EFA"}.
If using  \code{"SMC"} (default), the diagonal of the correlation matrices is
replaced by the squared multiple correlations (SMCs) of the indicators. If
using  \code{"PCA"}, the diagonal values of the correlation
matrices are left to be 1. If using  \code{"EFA"}, eigenvalues are found on the
correlation  matrices with the final communalities of an EFA solution as
diagonal. This is passed to  \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is \code{"pairwise.complete.obs"}.}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{HULL} derives its factor-search bound from an internal parallel analysis
against continuous reference data.
Default is  \code{"pearson"}.}

\item{n_datasets}{numeric. The number of datasets to simulate. Must be at
least 1. Default is 1000. This is passed to \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{percent}{numeric. The percentile to take from the simulated eigenvalues.
Default is 95. This is passed to \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{decision_rule}{character. Which rule to use to determine the number of
factors to retain. Default is \code{"means"}, which will use the average
simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
in percent. \code{"crawford"} uses the 95th percentile for the first factor
and the mean afterwards (based on Crawford et al, 2010). This is passed to \code{\link[=efa_parallel]{efa_parallel()}}.}

\item{n_factors}{numeric. Number of factors to extract if  \code{"EFA"} is
included in \code{eigen_type}. Default is 1. This is passed to
\code{\link[=efa_parallel]{efa_parallel()}}.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
\code{\link[=efa_fit]{efa_fit()}} fits of the 0 to \emph{J} factor solutions, and for the fit inside the internal
\code{\link[=efa_parallel]{efa_parallel()}} call. \code{NULL} (default) uses the \code{\link[=efa_fit]{efa_fit()}} defaults. This object carries
estimation settings only; the fits are always unrotated, which the hull statistics (CFI,
RMSEA, CAF) do not depend on.}

\item{...}{Further arguments passed to \code{\link[=efa_fit]{efa_fit()}}, also in
\code{\link[=efa_parallel]{efa_parallel()}}. The estimation tuning knobs are not passed here; they live in
\code{estimate_control}, a rotation setting is not accepted because the fits are unrotated, and
neither are the standard-error arguments (\code{se}, \code{b_boot}, \code{ci}, \code{seed}), because the fits
are internal steps whose standard errors are not reported.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). Its main fields are:
\item{n_factors}{A named numeric vector with the suggested number of factors
for each requested goodness-of-fit index (\code{"CAF"}, \code{"CFI"}, and/or
\code{"RMSEA"}).}
\item{results}{A list with one record per goodness-of-fit index, each holding
the goodness-of-fit values, the degrees of freedom, the hull membership, and
the retained solution used for printing and plotting. Each record also carries
\code{st}, the elbow sharpness of every solution (see details): the retained
solution has the largest value, and the runner-up shows how close the
selection was. \code{st} is \code{NA} for the solutions where it is undefined, that is
for those not on the hull and for the two hull endpoints, which have no
neighbouring hull solution on one side. When fewer than three solutions
remain on the hull, \code{st} is undefined throughout and the whole vector is \code{NA};
the retained solution is then the one with the highest goodness of fit, and a
warning says so.}
\item{settings}{A list of the settings used, including \code{n_fac_max}, the upper
bound \emph{J} of the number of factors to extract (see details). For backwards
compatibility the estimator is also repeated in \code{settings$method}.}

For backwards compatibility the per-index suggestions are additionally available
as the top-level fields \code{n_fac_CAF}, \code{n_fac_CFI} and \code{n_fac_RMSEA}, each \code{NA} if
that index was not requested in \code{gof}. New code should read them from
\code{n_factors} instead.
}
\description{
Implementation of the Hull method suggested by Lorenzo-Seva, Timmerman,
and Kiers (2011), with an extension to principal axis factoring. See details for
parallelization.
}
\details{
The Hull method aims to find a model with an optimal balance between
model fit and number of parameters, retaining only major factors
(Lorenzo-Seva, Timmerman, & Kiers, 2011). It fits 0 to \emph{J} factors -- where
\emph{J} is the number of factors suggested by parallel analysis (or \code{n_fac_theor},
if that is larger), plus one -- keeps the solutions on the upper boundary of
the convex hull of goodness-of-fit against degrees of freedom, and selects the
one at the sharpest elbow, i.e. with the highest \emph{st} value.

Because it trades fit against parsimony instead of testing against a null model
of uncorrelated variables, the Hull method does not lose accuracy for the
correlated-factor structures where parallel analysis (\code{\link[=efa_parallel]{efa_parallel()}}) tends to
under-extract; the CAF variant in particular was among the more accurate criteria
in Auerswald and Moshagen (2019). It needs at least six indicators and fits a
model at every candidate factor count, so it is comparatively slow and is not an
option for very short scales.

The \link{efa_parallel} function and the principal axis factoring of the
different number of factors can be parallelized using the future framework,
by calling the \code{\link[future:plan]{future::plan()}} function. The examples
provide example code on how to enable parallel processing.

The upper bound \emph{J} comes from \code{\link[=efa_parallel]{efa_parallel()}}, which compares against simulated
data, so the suggested number of factors varies slightly from run to run; a
criterion-based rotation passed through \code{...} adds its own random starts. Call
\code{\link[base:set.seed]{base::set.seed()}} beforehand to make a run reproducible; the result is then also
independent of the parallel plan.

Note that if \code{gof = "RMSEA"} is used, 1 - RMSEA is actually used to
compare the different solutions. This is necessary due to how the heuristic to
locate the elbow of the hull works.

The solutions are fitted without inequality constraints, so a solution can be
inadmissible (a Heywood case, or a fit that did not converge). Only the selected
solution is checked for this; if it is inadmissible a warning is raised and the
retained number of factors should be interpreted with caution.

The ML estimation method uses the \code{\link[psych:fa]{psych::fa()}}
starting values. See also the \link{efa_fit} documentation.
}
\examples{
\donttest{
# using PAF (this will print a message if gof is not specified manually
# and CAF will be used automatically)
efa_hull(test_models$baseline$cormat, N = 500, gof = "CAF", n_datasets = 100)

# using ML with all available fit indices (CAF, CFI, and RMSEA)
efa_hull(test_models$baseline$cormat, N = 500, estimator = "ML", n_datasets = 100)

# using ULS with only RMSEA
efa_hull(test_models$baseline$cormat, N = 500, estimator = "ULS", gof = "RMSEA",
         n_datasets = 100)
}

\dontrun{
# using parallel processing (Note: plans can be adapted, see the future
# package for details). future::plan() returns the plan it replaces, so
# on.exit() puts the session back as it was -- also if the call fails.
local({
  old_plan <- future::plan(future::multisession, workers = 2)
  on.exit(future::plan(old_plan), add = TRUE)
  efa_hull(test_models$baseline$cormat, N = 500, gof = "CAF")
})
}
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.oblique_procrustes_batch}
\alias{.oblique_procrustes_batch}
\title{Batched oblique Procrustes target rotation over a cube of loading matrices}
\usage{
.oblique_procrustes_batch(
  A,
  B,
  eps = 1e-05,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1,
  normalize = FALSE,
  random_starts = 0L,
  screen_keep = 2L,
  triage_maxit = 25L,
  triage_improve_tol = 0
)
}
\arguments{
\item{A}{Numeric array of dimension \verb{n x m x b}: the \code{b} loading matrices to
align.}

\item{B}{Numeric \verb{n x m} target loading matrix shared across all slices.}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient
norm.}

\item{maxit}{Integer scalar. Maximum number of full projected-gradient updates.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving attempts
after the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient
update.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization to the
loadings (only) before rotation and reverse it afterwards, leaving the target
unnormalized (ignored for single-factor slices).}

\item{random_starts}{Integer scalar. Number of additional random starts per
slice.}

\item{screen_keep}{Integer scalar. Number of screened random starts retained
for triage optimization.}

\item{triage_maxit}{Integer scalar. Number of short optimization iterations
used in the triage stage.}

\item{triage_improve_tol}{Numeric scalar. Relative improvement required for a
triaged start to be promoted to full optimization.}
}
\value{
A named list with the aligned-loadings array \code{loadings} (\verb{n x m x b}),
the factor-correlation array \code{Phi} (\verb{m x m x b}), and the per-slice
diagnostics \code{valid}, \code{convergence}, \code{value}, \code{iterations}, and
\code{line_search_failed}.
}
\description{
Align each slice of a loading-matrix cube to a single shared target using the
same oblique target rotation as \code{.oblique_procrustes()}, in one call. This
removes the per-replicate marshalling overhead of looping \code{efa_procrustes()} in R
over bootstrap or multiple-imputation arrays.
}
\details{
Each slice \code{A[, , i]} is aligned to \code{B}. For a single-factor cube the alignment
reduces to the closed-form sign match \code{T = sign(crossprod(A_i, B))} with factor
correlation \code{1}, matching the one-factor short-circuit in \code{efa_procrustes()}. For
two or more factors the slice is warm-started from the closed-form orthogonal
Procrustes solution (mirroring \code{efa_procrustes()}) and optimized with the same
multi-start oblique solver as \code{.oblique_procrustes()}. Random starts are drawn
serially with \code{R::rnorm} in the calling process.

Slices are aligned independently. A slice that cannot be aligned (a non-finite
loading matrix, a failed warm-start decomposition, an invalid fit, or any
linear-algebra exception) is reported with \code{valid = FALSE} and \code{NA} for the
loadings, factor correlations, and all other per-slice diagnostics, rather than
aborting the whole call, so one degenerate replicate does not discard the rest.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.CD.R
\name{plot.CD}
\alias{plot.CD}
\title{Plot CD object}
\usage{
\method{plot}{CD}(x, ...)
}
\arguments{
\item{x}{a list of class CD. An output from the \link{CD} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \link{CD} function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_average.R
\name{print.efa_average}
\alias{print.efa_average}
\alias{format.efa_average}
\title{Print and format an efa_average object}
\usage{
\method{print}{efa_average}(x, stat = c("average", "range"), plot = FALSE, ...)

\method{format}{efa_average}(x, stat = c("average", "range"), ...)
}
\arguments{
\item{x}{An object of class \code{efa_average} (output from \code{\link[=efa_average]{efa_average()}}).}

\item{stat}{character. A vector with the statistics to print. Possible inputs
are "average", "sd", "range", "min", and "max". Default is "average" and
"range".}

\item{plot}{logical. Whether a plot of the average and min- max loadings should
be created. Default is FALSE. If more than 10 factors are extracted, no plot is
created. Only used by \code{print()}; \code{\link[=plot.efa_average]{plot.efa_average()}} draws the same plot and
returns it, so it is the route to take when the plot object itself is wanted.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows a summarised output of the \code{\link[=efa_average]{efa_average()}} function: the
averaging settings, the error/convergence/Heywood/admissibility rates, the
indicator-to-factor correspondences, the averaged loadings (and, for oblique
solutions, the factor intercorrelations), the variances accounted for, and the
model fit. \code{format()} assembles the same report and returns it as a character
vector; by default (\code{plot = FALSE}) \code{print()} is \code{cat(format(x), sep = "\\n")}.
With \code{plot = TRUE} it additionally draws the loading plot, which is the one
thing \code{format()} cannot return: the printed lines are the same, but the call
has a side effect beyond them. The lines follow the active console theme, so
they are plain when colours are disabled (for example when captured into a file
or stripped with \code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
\donttest{
EFA_aver <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500)
EFA_aver

# format() returns the same lines as a character vector:
writeLines(format(EFA_aver))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DOSPERT_doc.R
\docType{data}
\name{DOSPERT}
\alias{DOSPERT}
\title{DOSPERT}
\format{
A list of 2 with elements "cormat" (40 x 40 matrix of bivariate
correlations) and "N" (scalar).
\describe{
\item{cormat}{(matrix) - Bivariate correlations of the 40 DOSPERT items,
which span the six risk domains (social, recreational, gambling,
health/safety, investment, and ethical).}
\item{N}{(numeric) - The sample size the correlations are based on.}
}
}
\source{
Weber, E. U., Blais, A.-R., & Betz, N. E. (2002). A domain specific risk-attitude scale: Measuring risk perceptions and risk behaviors. Journal of Behavioral Decision Making, 15(4), 263–290. doi: 10.1002/bdm.414

Frey, R., Pedroni, A., Mata, R., Rieskamp, J., & Hertwig, R. (2017). Risk preference shares the psychometric structure of major psychological traits. Science Advances, 3, e1701381.

\url{https://osf.io/rce7g}
}
\usage{
DOSPERT
}
\description{
A list containing the bivariate correlations (cormat) of the 40 items of
the Domain Specific Risk Taking Scale (DOSPERT; Weber, Blais, & Betz, 2002)
and the sample size (N) based on the publicly available dataset at
(\url{https://osf.io/rce7g}) of the Basel-Berlin Risk Study (Frey et al., 2017).
The items measure risk-taking propensity on six different domains: social,
recreational, gambling, health/ safety, investment, and ethical.
}
\details{
The underlying data deposit is licensed under CC BY 4.0
(\url{https://creativecommons.org/licenses/by/4.0/legalcode}). These correlations are a
derivative of it and are attributed to Frey et al. (2017), as the licence requires.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_nest.R
\name{efa_nest}
\alias{efa_nest}
\title{Next eigenvalue sufficiency test (NEST)}
\source{
Achim, A. (2017). Testing the number of required dimensions in exploratory factor analysis. The Quantitative Methods for Psychology, 13(1), 64–74. https://doi.org/10.20982/tqmp.13.1.p064

Brandenburg, N., & Papenberg, M. (2024). Reassessment of innovative methods to determine the number of factors: A simulation-based comparison of exploratory graph analysis and Next Eigenvalue Sufficiency Test. Psychological Methods, 29(1), 21–47. https://doi.org/10.1037/met0000527

Caron, P.-O. (2025). A Comparison of the Next Eigenvalue Sufficiency Test to Other Stopping Rules for the Number of Factors in Factor Analysis.
Educational and Psychological Measurement, Online-first publication. https://doi.org/10.1177/00131644241308528
}
\usage{
efa_nest(
  x,
  N = NA,
  alpha = 0.05,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_datasets = 1000,
  estimate_control = NULL,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Only needed if x is a correlation
matrix. Must be larger than the number of variables.}

\item{alpha}{numeric. The alpha level to use (i.e., 1-alpha percentile of eigenvalues is used for reference values).}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{NEST} compares the data against simulated continuous reference data.
Default is  \code{"pearson"}.}

\item{n_datasets}{numeric. The number of datasets to simulate. Default is 1000.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
\code{\link[=efa_fit]{efa_fit()}} reference-model fits. \code{NULL} (default) uses the \code{\link[=efa_fit]{efa_fit()}} defaults. The
reference models are unrotated, so no rotation settings apply.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}}. For example,
\code{estimator}, to change the estimator (default is "PAF"). PAF is more
robust, but it will take longer compared to the other estimators
available ("ML" and "ULS"). The estimation tuning knobs are not passed here; they
live in \code{estimate_control}, and the standard-error arguments (\code{se}, \code{b_boot}, \code{ci},
\code{seed}) are not accepted because the reference-model fits are internal steps.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). Its main fields are:
\item{n_factors}{A named numeric vector (\code{"NEST"}) with the suggested number of
factors according to the NEST procedure.}
\item{results}{A list with a single record holding the empirical eigenvalues
and the reference eigenvalues. Only the positions the search actually tested
carry a reference value; beyond the position at which it stopped the series is
\code{NA}.}
\item{settings}{A list of control settings used.}
}
\description{
NEST uses many synthetic datasets to generate reference eigenvalues against
which to compare the empirical eigenvalues. This is similar to parallel
analysis, but other than parallel analysis, NEST does not just rely on
synthetic eigenvalues based on an identity matrix as null model.
It was introduced by Achim (2017), see also Brandenburg and Papenberg (2024) and
Caron (2025) for further simulation studies including NEST.
}
\details{
NEST compares the first empirical eigenvalue against the first eigenvalues
of \code{n_dataset} synthetic datasets based on a null model  (i.e.,
with uncorrelated variables; same as in parallel analysis, see \code{\link[=efa_parallel]{efa_parallel()}}).
The following eigenvalues are compared against synthetic datasets based on an EFA-model with one fewer factors
than the position of the respective empirical eigenvalue. E.g, the second
empirical eigenvalue is compared against synthetic data based on a one-factor
model. In each comparison the \eqn{k}-th empirical eigenvalue is tested against
the \eqn{k}-th largest eigenvalue of the synthetic datasets. The \code{alpha}-level
defines against which percentile of the synthetic
eigenvalue distribution to compare the empirical eigenvalues against, i.e., an
alpha of .05 (the default) uses the 95th percentile as reference value.

The number of factors tested is capped at \eqn{\lfloor 0.8 \times p \rfloor}
(with \eqn{p} the number of variables; Achim, 2017) and additionally limited so
that the \eqn{(k - 1)}-factor reference model used at each step stays
over-identified. If no empirical eigenvalue falls at or below its reference
within this range, every tested factor is accepted and this capped number is
returned.

Because each reference model carries the factors already retained, NEST does not
lose accuracy for the strongly correlated factor structures where parallel
analysis tends to under-extract, and it was among the more accurate criteria in
the simulation studies of Brandenburg and Papenberg (2024) and Caron (2025). The
price is runtime: a fresh set of \code{n_datasets} reference datasets is drawn and
eigen-decomposed at every candidate factor count, which makes NEST one of the
slowest criteria available here.

The reference models are fitted without inequality constraints. A Heywood case in
one of them leaves no unique variance to simulate the reference data from, so
NEST aborts rather than continuing from an inadmissible reference.

The reference eigenvalues are obtained from simulated data, so the suggested number
of factors varies slightly from run to run. Call \code{\link[base:set.seed]{base::set.seed()}} beforehand to make a
run reproducible.

For details on the method, including simulation studies, see Achim (2017),
Brandenburg and Papenberg (2024), and Caron (2025).

The \code{efa_nest} function can also be called together with other factor
retention criteria in the \code{\link[=efa_retain]{efa_retain()}} function.
}
\examples{
\donttest{
# with correlation matrix
efa_nest(test_models$baseline$cormat, N = 500)

# with raw data
efa_nest(GRiPS_raw)
}
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.CD.R
\name{print.CD}
\alias{print.CD}
\title{Print function for CD objects}
\usage{
\method{print}{CD}(x, plot = TRUE, ...)
}
\arguments{
\item{x}{a list of class CD. Output from \link{CD} function.}

\item{plot}{logical. Whether to plot the results.}

\item{...}{Further arguments for print.}
}
\description{
Print function for CD objects
}
\examples{
\donttest{
# determine n factors of the GRiPS
CD(GRiPS_raw)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_fit.R
\name{efa_fit}
\alias{efa_fit}
\title{Exploratory factor analysis (EFA)}
\source{
Bollen, K. A., & Stine, R. A. (1992). Bootstrapping goodness-of-fit measures
in structural equation models. Sociological Methods & Research, 21, 205–229.
doi: 10.1177/0049124192021002004

Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
of implementations of principal axis factoring and promax rotation in R and SPSS.
Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x

Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
rotation to oblique simple structure. British Journal of Statistical Psychology,
17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x

Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
Research, 46, 340-364, doi: 10.1080/00273171.2011.564527

Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233

Lawley, D. N., & Maxwell, A. E. (1971). Factor analysis as a statistical
method (2nd ed.). Butterworths.

Cudeck, R. (1989). Analysis of correlation matrices using covariance
structure models. Psychological Bulletin, 105, 317–327.
doi: 10.1037/0033-2909.105.2.317

Olkin, I., & Siotani, M. (1976). Asymptotic distribution of functions of a
correlation matrix. In S. Ikeda (Ed.), Essays in probability and statistics
(pp. 235–251). Shinko Tsusho.

Jennrich, R. I. (1973). Standard errors for obliquely rotated factor
loadings. Psychometrika, 38, 593–604. doi: 10.1007/BF02291497

Zhang, G., & Preacher, K. J. (2015). Factor rotation and standard errors
in exploratory factor analysis. Journal of Educational and Behavioral Statistics,
40, 579–603. doi: 10.3102/1076998615606098

Browne, M. W. (1984). Asymptotically distribution-free methods for the
analysis of covariance structures. British Journal of Mathematical and Statistical
Psychology, 37, 62–83. doi: 10.1111/j.2044-8317.1984.tb00789.x

Satorra, A., & Bentler, P. M. (1994). Corrections to test statistics and
standard errors in covariance structure analysis. In A. von Eye & C. C. Clogg (Eds.),
Latent variables analysis: Applications for developmental research (pp. 399–419). Sage.

Asparouhov, T., & Muthén, B. (2010). Simple second order chi-square
correction. Mplus Technical Appendix.

Muthén, B., du Toit, S. H. C., & Spisic, D. (1997). Robust inference using
weighted least squares and quadratic estimating equations in latent variable modeling
with categorical and continuous outcomes. Unpublished manuscript.

Yuan, K.-H., & Bentler, P. M. (2000). Three likelihood-based methods for mean
and covariance structure analysis with nonnormal missing data. Sociological Methodology,
30, 165–200. doi: 10.1111/0081-1750.00078

Yuan, K.-H., Marshall, L. L., & Bentler, P. M. (2002). A unified approach to
exploratory factor analysis with missing data, nonnormal data, and in the presence of
outliers. Psychometrika, 67, 95–121. doi: 10.1007/BF02294711

Savalei, V., & Bentler, P. M. (2009). A two-stage approach to missing data:
Theory and application to auxiliary variables. Structural Equation Modeling, 16, 477–497.
doi: 10.1080/10705510903008238

Little, R. J. A., & Rubin, D. B. (2002). Statistical analysis with missing data
(2nd ed.). Wiley.

Bartlett, M. S. (1951). The effect of standardization on a Chi-square
approximation in factor analysis. Biometrika, 38, 337–344.

Bentler, P. M. (1990). Comparative fit indexes in structural models.
Psychological Bulletin, 107, 238–246. doi: 10.1037/0033-2909.107.2.238

Tucker, L. R., & Lewis, C. (1973). A reliability coefficient for maximum
likelihood factor analysis. Psychometrika, 38, 1–10. doi: 10.1007/BF02291170

Browne, M. W., & Cudeck, R. (1989). Single sample cross-validation indices
for covariance structures. Multivariate Behavioral Research, 24, 445–455.
doi: 10.1207/s15327906mbr2404_4

Browne, M. W., & Cudeck, R. (1992). Alternative ways of assessing model fit.
Sociological Methods & Research, 21, 230–258. doi: 10.1177/0049124192021002005

Bentler, P. M. (1995). EQS structural equations program manual. Multivariate
Software.
}
\usage{
efa_fit(
  x,
  n_factors,
  N = NA,
  estimator = c("PAF", "ML", "ULS", "MINRES", "DWLS"),
  rotation = c("none", "varimax", "equamax", "quartimax", "geominT", "bentlerT",
    "bifactorT", "promax", "oblimin", "quartimin", "simplimax", "bentlerQ", "geominQ",
    "bifactorQ"),
  se = c("none", "information", "sandwich", "np-boot"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra", "fiml"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  estimate_control = NULL,
  rotate_control = NULL,
  b_boot = 1000,
  ci = 0.95,
  seed = NULL,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations. If raw data is entered, the correlation matrix is found from the
data.}

\item{n_factors}{numeric. Number of factors to extract. Must be at least 1 and
smaller than the number of variables (the common factor model is not identified
otherwise). Use \code{\link[=efa_retain]{efa_retain()}} to decide on a value.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used; with raw data, \code{N} is found from the data instead.
\itemize{
\item With \code{N = NA}, not all fit indices can be computed; a positive \code{N} that is very
small relative to the number of variables leaves the chi-square-derived indices
unavailable as well, with a warning.
\item With raw data, \code{N} is the number of cases the correlation matrix was actually
computed from -- see \code{use} for the general rule and how missing values change it.
Under \code{cor_method = "fiml"}, \code{use} is ignored and \code{N} is instead the number of
cases carrying at least one observed value.
}}

\item{estimator}{character. The estimator used to fit the EFA: "PAF" (principal axis
factoring), "ML" (maximum likelihood), "ULS" (unweighted least squares; "MINRES" is an
accepted alias returning identical results), or "DWLS" (diagonally weighted least
squares, for ordinal data). See the \emph{Estimators} section in Details for their
properties and data requirements. Lower-case versions (e.g., "paf") are also accepted.}

\item{rotation}{character. Either perform no rotation ("none"; default),
an orthogonal rotation ("varimax", "equamax", "quartimax", "geominT",
"bentlerT", or "bifactorT"), or an oblique rotation ("promax", "oblimin",
"quartimin", "simplimax", "bentlerQ", "geominQ", or "bifactorQ"). See the
\emph{Rotations} section in Details for their properties and known issues.}

\item{se}{character. Whether and how to compute standard errors (and matching
confidence intervals): "none" (default), "information" (analytic
standard errors from the expected Fisher information of the ML solution), "sandwich"
(robust "sandwich" standard errors from raw data, which stay reliable under
non-normality or a misspecified estimator weight), or "np-boot" (non-parametric
bootstrap). The
methods differ in their assumptions, their data requirements, and which estimator,
rotation, and \code{cor_method} combinations they support; see the \emph{Standard errors}
section in Details.}

\item{cor_method}{character. How the correlation is computed from raw data:
\code{"pearson"}, \code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}); \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data; or
\code{"fiml"} for a two-stage full-information maximum-likelihood correlation from raw data
with missing values. See the \emph{Correlation methods} section in Details for their
properties and the combinations they support. Default is "pearson".}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is "pairwise.complete.obs". It is ignored when
\code{cor_method = "fiml"} (which handles the missingness itself, so every case
contributes), and it is overridden to listwise deletion whenever an asymptotic
covariance is required (the "DWLS" estimator, or \code{se = "sandwich"}), because the
covariance must describe the same cases as the correlation matrix.}

\item{estimate_control}{a control object from \code{\link[=estimate_control]{estimate_control()}} bundling the
estimation control arguments: the \code{type} preset; the principal-axis-factoring iteration
settings \code{init_comm}, \code{criterion}, \code{criterion_type}, \code{max_iter}, and \code{abs_eigen}; and
the maximum-likelihood \code{start_method}. Defaults to \code{estimate_control()}, which uses
the \code{"EFAtools"} type. See \code{\link[=estimate_control]{estimate_control()}}.}

\item{rotate_control}{a control object from \code{\link[=rotate_control]{rotate_control()}} bundling the rotation
control arguments: the \code{type} preset; Kaiser \code{normalize}; the convergence \code{precision}; the
factor \code{order_type}; the varimax/promax settings \code{varimax_type} and \code{p_type}; the
simplimax/promax \code{k}; and \code{random_starts}. Defaults to \code{rotate_control()}, which uses
the \code{"EFAtools"} type. The estimation and rotation presets
are independent. See \code{\link[=rotate_control]{rotate_control()}} for details.}

\item{b_boot}{numeric. The number of bootstrap samples to draw. Default is 1000.
Must be at least 2, the smallest number from which a standard error is defined.
Under \code{cor_method = "fiml"} each bootstrap sample re-runs the EM moment
estimation, so a smaller value may be advisable.}

\item{ci}{numeric. The level of the confidence intervals: the percentile intervals from
the bootstrap samples under \code{se = "np-boot"}, and the analytic Wald intervals under
\code{se = "information"} and \code{se = "sandwich"}, the corrected two-stage intervals of
\code{cor_method = "fiml"} included. Must be greater than 0 and smaller than 1. Default is
.95 for 95\% CIs.}

\item{seed}{numeric. An optional seed for the random-number generator.}

\item{...}{Additional arguments forwarded to the rotation engine (usually not needed);
an unrecognized one is an error, not a setting that is silently ignored.}
}
\value{
A list of class \code{c("efa", "EFA")} containing (a subset of) the following:

\item{orig_R}{Original correlation matrix.}
\item{h2_init}{Initial communality estimates from PAF.}
\item{h2}{Final communality estimates from the unrotated solution.}
\item{orig_eigen}{Eigen values of the original correlation matrix.}
\item{init_eigen}{Initial eigenvalues, obtained from the correlation matrix
with the initial communality estimates as diagonal in PAF.}
\item{final_eigen}{Eigenvalues obtained from the correlation matrix
with the final communality estimates as diagonal.}
\item{iter}{For PAF, the number of iterations until convergence. For ML, ULS, and
DWLS, the number of objective-function evaluations used by the optimiser (not the
number of optimiser iterations).}
\item{convergence}{Integer convergence code (0 = converged), using the codes of
\code{\link[stats:optim]{stats::optim()}}. For ML and ULS it is the code from the bounded
optimiser; for DWLS the fit runs a bounded warm start followed by an
unconstrained polish, and the reported code is from the final polish. For PAF
it is 1 if the maximum number of iterations was reached
without meeting the convergence criterion and 0 otherwise. A non-zero code is
also reported with a warning.}
\item{heywood}{A named integer vector indicating which variables have a
Heywood (improper) case in the unrotated solution; empty if there are none.}
\item{unrot_loadings}{Loading matrix containing the final unrotated loadings.}
\item{vars_accounted}{Matrix of explained variances and sums of squared loadings. Based
on the unrotated loadings. Its rows are \code{"SS loadings"} and \code{"Prop Tot Var"}, followed by
\code{"Cum Prop Tot Var"}, \code{"Prop Comm Var"}, and \code{"Cum Prop Comm Var"}; those last three are
omitted for a single-factor solution, where they would only repeat the two above them. So
code that indexes a row by name (for example \verb{["Prop Comm Var", ]}) must allow for the
two-row form at \code{n_factors = 1}.}
\item{fit_indices}{A named list of fit indices computed from the unrotated
loadings. ML and ULS report the full set: the model Chi Square (with its
p-value and df), CFI, TLI, RMSEA with its 90\% confidence interval, AIC, BIC,
ECVI, RMSR, SRMR, and CAF. PAF and DWLS report only RMSR, SRMR, CAF, and df;
the Chi-Square-based indices are \code{NA} there, unless DWLS uses
\code{se = "sandwich"}, which fills the full block from a scaled Chi Square
instead. RMSR and SRMR are both included, but the print and summary methods
display only SRMR (see \emph{Fit indices} in Details). Whenever the Chi Square is
scaled (\code{se = "sandwich"}, or a \code{cor_method = "fiml"} fit whose correction
could be formed), AIC, BIC, and ECVI are \code{NA}, and a few extra
scaled-statistic fields are appended for advanced diagnostics (see
\emph{Fit indices} in Details). The list also carries \code{Fm}, the estimator's own
objective value at the solution -- on its own scale, not the discrepancy
the Chi Square is built from -- and the independence-model baseline
\code{chi_null}, \code{df_null}, and \code{p_null} that CFI and TLI are computed from.}
\item{model_implied_R}{The model implied correlation
matrix.}
\item{residuals}{Residual correlations, i.e., orig_R - model_implied_R}
\item{standardized_residuals}{Residual correlations standardized by their
bootstrap standard errors. Only returned, if \code{se = "np-boot"}.}
\item{rot_loadings}{Loading matrix containing the final rotated loadings. For an oblique
rotation this is the pattern matrix -- each variable's unique contribution from each
factor, with the other factors partialled out -- and is the matrix normally interpreted
(see \emph{Rotations}).}
\item{Phi}{The factor intercorrelations (only for oblique rotations).}
\item{Structure}{The structure matrix \code{rot_loadings \%*\% Phi}, holding the plain
variable-factor correlations, which are inflated by the factor intercorrelations (only
for oblique rotations).}
\item{rotmat}{The rotation matrix. The rotated loadings are recovered from the
unrotated loadings as \code{unrot_loadings \%*\% rotmat} for orthogonal rotations and
for promax, and as \code{unrot_loadings \%*\% t(solve(rotmat))} for the other oblique
rotations.}
\item{vars_accounted_rot}{Matrix of explained variances and sums of squared
loadings. Based on rotated loadings and, for oblique rotations, the factor
intercorrelations. Same rows as \code{vars_accounted}; it is returned only when a rotation
was actually applied, which never happens for a single-factor solution, so it always
has the five-row form.}
\item{settings}{A list of the settings used, including \code{seed} (\code{NULL} when none was
supplied), \code{input_type} (\code{"raw"} or \code{"correlation"}, what \code{x} was), and
\code{cor_method_used} (the correlation method that actually ran, \code{NA_character_} for a
correlation-matrix input, which consumes none). \code{cor_method} keeps the requested value
whether or not it was used. For the criterion rotations fitted by gradient projection
it additionally carries \code{rotation_diagnostics}, a list summarising the multi-start run:
\itemize{
\item \code{converged}: whether the start whose solution is returned reached the convergence
tolerance (reported separately from the counts below because the returned solution
is the one with the lowest criterion value, which need not be a converged one).
\item \code{n_starts_total}: the \code{random_starts} random starts plus the rational start.
\item \code{n_optimized}: how many of those starts were actually optimized -- fewer than
\code{n_starts_total} whenever the solver screens the random starts and optimizes only
the most promising ones.
\item \code{n_converged}: how many optimized starts reached the convergence tolerance.
\item \code{n_distinct_minima}: how many distinct local optima those converged starts found;
more than one means the criterion is multimodal on these data.
\item \code{criterion_spread}: the range of the criterion values they attained.
\item \code{criterion_best}: the criterion value of the returned solution.
}

When \code{normalize = TRUE}, \code{criterion_best} and \code{criterion_spread} are evaluated on
the Kaiser-normalized loadings the criterion is optimized on, not on the returned
\code{rot_loadings}, so they are not directly comparable to a criterion recomputed from
the returned loadings.}
\item{fiml}{Diagnostics of the FIML correlation's EM estimation, present only for
\code{cor_method = "fiml"}: whether it \code{converged} before \code{fiml_max_iter}, how many
iterations it took (\code{iter}), the number of distinct missingness patterns
(\code{n_patterns}), and the number of cases used (\code{n}, the reported \code{N}). A \code{converged}
of \code{FALSE} means the analysed correlation is the EM's last iterate, not the true
FIML estimate; this is flagged in the printed output. Set
\code{fiml_max_iter} and \code{fiml_tol} through \code{\link[=estimate_control]{estimate_control()}}.}
\item{SE}{A named list of standard error matrices, returned only when \code{se} is not
\code{"none"} (see \emph{Standard errors} in Details for what each method assumes and when a
component comes back \code{NA}). For \code{se = "np-boot"}: bootstrap SDs of the loadings
(rotated too, if a rotation was applied), the residuals, and the fit indices, plus
-- for oblique rotations -- \code{Phi} and the structure coefficients; \code{valid_replicates}
(and, when rotated, \code{valid_target_rotations}) records how many bootstrap replicates
each of those is based on. For \code{se = "information"} and \code{se = "sandwich"}: Wald-type
SEs for the loadings, the uniquenesses, and the communalities -- identical to each
other, since communality is just \code{1 - uniqueness} -- plus \code{Phi} and the structure
coefficients for oblique rotations. \code{"sandwich"} stays valid under non-normality;
\code{"information"} assumes the model is correctly specified.}
\item{CI}{A named list of confidence intervals of width \code{ci}, matching the
components of \code{SE}: percentile intervals for \code{se = "np-boot"}, Wald intervals for
\code{se = "information"} and \code{se = "sandwich"}. Only returned if \code{se} is not \code{"none"}.}
\item{replicates}{A named list of raw bootstrap replicate arrays -- the aligned
loadings, \code{Phi}, structure coefficients, residuals, and fit indices behind \code{SE} and
\code{CI} -- with one replicate per array's last dimension (first dimension for
\code{fit_indices}). Failed replicates are left \code{NA}. Populated only for
\code{se = "np-boot"}; \code{NULL} otherwise.}
\item{vcov_unrot_loadings}{The full unrotated-loading covariance matrix behind
\code{SE$unrot_loadings}: a \code{p * n_factors} by \code{p * n_factors} matrix in column-major
\code{vec(Lambda)} order, with rows and columns labelled \code{"<variable>_<factor>"}.
Populated for \code{se = "information"} and \code{se = "sandwich"} -- always the unrotated
block, even when a rotation is applied -- and \code{NA}-filled if the analytic covariance
is unreliable (a Heywood case or a singular information matrix); \code{NULL} for
\code{se = "np-boot"} and \code{se = "none"}. It can be populated even when
\code{SE$unrot_loadings} is \code{NA}: a weakly determined rotational orientation invalidates
only the marginal SEs, not the underlying covariance.}
\item{Gamma}{The asymptotic covariance of the off-diagonal sample correlations --
the meat of the robust sandwich SEs. A \code{p (p - 1) / 2} by \code{p (p - 1) / 2} matrix,
rows and columns ordered by \code{\link[utils:combn]{utils::combn()}} over the column pairs and labelled
\code{"<var_i>-<var_j>"}. Populated for \code{se = "sandwich"} on the polychoric/tetrachoric
and Pearson paths; \code{NULL} otherwise (including under \code{cor_method = "fiml"}, whose
meat is not returned). It typically dominates a sandwich fit's size (about 11 MB at
\code{p = 50}) and is kept because \code{\link[=efa_mi]{efa_mi()}} needs it from each per-imputation fit to
build the pooled covariance.}
}
\description{
This function does an EFA with either \code{PAF}, \code{ML}, \code{ULS}/\code{MINRES},
or \code{DWLS} with or without subsequent rotation.
Estimation and rotation are controlled through the control objects built by
\code{\link[=estimate_control]{estimate_control()}} and \code{\link[=rotate_control]{rotate_control()}}; each accepts a \code{type}
("EFAtools", "SPSS", "psych", or "none") that fills in its remaining settings.
}
\details{
\subsection{Estimators}{

The estimator is chosen with \code{estimator}.
\itemize{
\item \strong{PAF} (principal axis factoring) iteratively estimates the communalities and makes
no distributional assumptions, which makes it robust and a good general-purpose
default. Because it minimises no likelihood or weighted discrepancy it provides no
model chi-square, and hence no chi-square-based fit indices (see \emph{Fit indices}).
\item \strong{ML} (maximum likelihood) maximises the normal-theory likelihood. It yields the
full set of fit indices and is the only estimator with analytic expected-information
standard errors (\code{se = "information"}), but it assumes multivariate normality and is
the most prone to Heywood (improper) cases. Its starting values are set by
\code{start_method}.
\item \strong{ULS} (unweighted least squares) minimises the sum of squared correlation
residuals. "MINRES" (minimum residual) is the same estimator under a different name
and returns identical results. It makes no normality assumption, is robust to mild
non-normality, and yields the full set of fit indices.
\item \strong{DWLS} (diagonally weighted least squares) is the recommended estimator for ordinal
data. It weights each off-diagonal correlation residual by the inverse asymptotic
variance of the corresponding polychoric correlation (Muthén, du Toit, & Spisic, 1997),
reproducing the loadings of a diagonally weighted least squares fit. It therefore
requires raw ordinal data with \code{cor_method = "poly"} or \code{"tetra"}. Because the
weighting follows the polychoric
asymptotic covariance, the matrix and the weights are estimated on the
listwise-complete cases. Its fit-index behaviour is described under \emph{Fit indices}.
}
}

\subsection{Correlation methods}{

When raw data are supplied, \code{cor_method} selects how the correlation matrix is computed
(it is ignored when a correlation matrix is entered directly).
\itemize{
\item \strong{"pearson"} (default), \strong{"spearman"}, and \strong{"kendall"} are passed to
\code{\link[stats:cor]{stats::cor()}}. The factor model assumes a Pearson correlation, but a rank
correlation is analysed on its own scale, not converted to a Pearson-equivalent
value; Kendall's tau in particular gives the most attenuated loadings of the three.
For ordinal items prefer \code{"poly"} / \code{"tetra"} below, which estimate the correlation
of the underlying continuous variables.
\item \strong{"poly"} / \strong{"tetra"} compute polychoric / tetrachoric correlations for ordinal /
binary data, assuming an underlying bivariate-normal latent variable. They use a
two-step estimator. The polychoric asymptotic covariance that
underlies both the DWLS weights and the scaled (sandwich) statistic relies on
large-sample theory that degrades for empty or near-empty response-category
combinations; with very sparse cells the resulting weights and standard errors can be
unreliable (a warning is issued when empty cells are present), so interpret them with
caution and consider collapsing rare categories.
Each of the \code{p(p - 1)/2} variable pairs is a separate numerical optimisation, so a
polychoric matrix takes much longer to compute than a Pearson one, and the difference grows
quadratically in the number of variables; with \code{se = "np-boot"} the whole matrix is
re-estimated for every bootstrap replicate.
\item \strong{"fiml"} estimates a two-stage full-information maximum-likelihood correlation. The
saturated multivariate-normal mean and covariance are estimated from raw data with
missing values by an EM algorithm assuming the data are missing at random (Yuan,
Marshall, & Bentler, 2002; Little & Rubin, 2002), and the standardized covariance is
then analysed. The model fit indices are corrected
two-stage statistics wherever the correction can be formed (see \emph{Fit indices}).
\code{"fiml"} uses every case and handles the
missingness itself, so \code{use} is ignored; it supplies a continuous (Pearson-type)
correlation only and is therefore not compatible with \code{estimator = "DWLS"}. Standard
errors are available analytically for \code{estimator = "ML"} or \code{"ULS"} and, for any estimator,
by the non-parametric bootstrap (see \emph{Standard errors}). For multiply imputed data,
\code{\link[=efa_mi]{efa_mi()}} is the alternative route to handling missingness. Both routes assume
the values are missing at random (MAR), and which one to prefer is largely
practical: FIML is a single, efficient fit and is the simpler default when the
analysis model is the whole story, whereas multiple imputation is more flexible when
the imputation model should draw on auxiliary variables not in the factor model, or
when the same imputations feed several downstream analyses.
}
}

\subsection{Rotations}{

A rotation transforms the unrotated loadings toward a simpler, more interpretable
pattern; all rotations are performed by rotation engines built into the package.
Orthogonal rotations keep the factors uncorrelated, whereas oblique rotations let them
correlate (returning a pattern matrix, a structure matrix, and the factor
intercorrelations \code{Phi}) and are usually more realistic for psychological constructs.

For an oblique solution the \strong{pattern} matrix (\code{rot_loadings}) holds each variable's
unique contribution from each factor, with the other factors partialled out; it is what
is normally interpreted and reported. The \strong{structure} matrix
(\code{Structure = rot_loadings \%*\% Phi}) holds the plain variable-factor correlations, which
are inflated by the factor intercorrelations. The two coincide only when \code{Phi} is the
identity, which is why an orthogonal rotation returns \code{rot_loadings} alone.

Orthogonal rotations:
\itemize{
\item \strong{varimax} maximises the variance of the squared loadings within each factor (column
simplicity). It is the most widely used orthogonal rotation and spreads variance
across factors rather than concentrating it in a general factor.
\item \strong{quartimax} simplifies the variables (rows) so that each loads mainly on one factor;
it tends to produce a strong general factor.
\item \strong{equamax} is a Crawford-Ferguson compromise between varimax (column) and quartimax
(row) simplicity.
\item \strong{geominT} uses a geometric-mean criterion that rewards a sparse pattern and
tolerates variables with cross-loadings; a smaller offset \code{delta} gives a sparser
solution but sharper local minima.
\item \strong{bentlerT} uses Bentler's invariant pattern simplicity criterion.
\item \strong{bifactorT} is the Jennrich-Bentler orthogonal bifactor criterion: a general factor
plus group factors (bifactor simple structure). It needs at least two group factors
(\code{n_factors >= 3}): with two factors the criterion is identically zero, so no rotation
is performed and the unrotated loadings are returned with a warning.
}

Oblique rotations:
\itemize{
\item \strong{promax} is a fast two-step rotation: a varimax solution is raised to a power
(controlled by \code{k} and \code{p_type}) to form a target that is then fitted obliquely.
\item \strong{quartimin} simplifies the variables (rows), like quartimax, so each loads mainly
on one factor, but factors are allowed to correlate.
\item \strong{oblimin} is quartimin with a tunable argument, \code{gam} (default 0, i.e. quartimin
itself); turning it up trades some of that row simplicity for more strongly
correlated factors, and can drive the solution toward factor collapse, so inspect
\code{Phi} before interpreting a fit with \code{gam > 0}.
\item \strong{simplimax} drives the \code{k} smallest loadings toward zero. Its criterion is only
piecewise smooth and strongly multimodal, which makes it by far the most
start-dependent rotation offered here: different random seeds can reach noticeably
different solutions. Because every start is fully optimised (there is no screening
stage), raising \code{random_starts} to several hundred costs proportionally more time but
buys a better optimum.
\item \strong{bentlerQ} is the oblique Bentler invariant pattern simplicity criterion.
\item \strong{geominQ} is the oblique geomin criterion; it handles complex (cross-loading)
structure well but is multimodal, so it benefits from more \code{random_starts} (and uses a
more thorough multi-start search internally).
\item \strong{bifactorQ} is the oblique (correlated) Jennrich-Bentler bifactor criterion, with the
same two-group-factor requirement as \strong{bifactorT}.
}

Prefer an oblique rotation unless there is a substantive reason to force the factors to
be uncorrelated: if \code{Phi} comes back near zero the oblique solution is essentially the
orthogonal one anyway, whereas imposing orthogonality on genuinely correlated factors
distorts the pattern.

The criterion-based rotations (all except varimax and promax) are fitted by gradient
projection with \code{random_starts} random starts to guard against local minima; the
complexity criteria (simplimax and geominQ in particular) are the most multimodal. The
starts are drawn from the random-number generator, so different starts can reach
genuinely different optima and such a fit is reproducible only when the generator is
controlled: pass \code{seed}, or call \code{\link[base:set.seed]{base::set.seed()}} beforehand.
}

\subsection{Standard errors}{

\code{se} selects whether and how standard errors (and matching confidence intervals) are
computed. Which quantities they cover depends on the method. The analytic methods
(\code{"information"} and \code{"sandwich"}) cover the unrotated loadings, the uniquenesses and
the communalities and, when a rotation is applied, the rotated loadings and -- for
oblique rotations -- the factor correlations and the structure coefficients. The
bootstrap (\code{"np-boot"}) covers the unrotated loadings, the residuals, and the fit
indices and, when a rotation is applied, the rotated loadings and -- for oblique
rotations -- the factor correlations and the structure coefficients; it reports no
uniqueness or communality standard errors.
\itemize{
\item \strong{"none"} (default) computes no standard errors.
\item \strong{"information"} returns analytic standard errors from the expected (Fisher)
information matrix of the maximum-likelihood solution, and therefore requires
\code{estimator = "ML"} and \code{cor_method = "pearson"} (or \code{"fiml"}, see below). The rotated
standard errors are obtained by propagating the unrotated-loading covariance through
the rotation (Jennrich, 1973); because rotated quantities do not depend on how the
unrotated solution happens to be oriented, they are directly comparable across
programs. Unlike the bootstrap it also works from a correlation matrix as long as \code{N}
is supplied. These standard errors assume multivariate normality and a correctly
specified model; under heavy-tailed data or model misfit they can understate the
sampling variability, where \code{"sandwich"} or \code{"np-boot"} are more robust.

The rotated loadings, \code{Phi}, the structure coefficients, the uniquenesses, and the
communalities do not depend on how the unrotated solution happens to be oriented, and
so are comparable across programs. The \strong{unrotated} loading standard errors are not:
a program using a different orientation will report different unrotated loading
standard errors for the same fit. That orientation can also fail to be well defined --
for example when two factors are of near-equal strength -- in which case the
unrotated loadings have no well-defined sampling distribution and their standard
errors diverge. \code{efa_fit()} detects this and returns \code{NA} for the unrotated loading
standard errors with a warning, rather than reporting a number that looks like a
standard error but is an artefact of the orientation; the rotated loadings, \code{Phi}, the
structure coefficients, the uniquenesses, and the communalities are unaffected and
still reported. Use those, or \code{se = "np-boot"}, when the unrotated loadings
themselves are the quantity of interest. This detection applies to every analytic
route (the Pearson and polychoric paths and the two-stage \code{cor_method = "fiml"}
sandwich alike). A Heywood case (a uniqueness at its lower boundary) is
separate and more severe: there
the whole \code{SE}/\code{CI} block comes back \code{NA} with a warning, because the standard-error
approximation fails for every parameter. The scaled chi-square (from \code{"sandwich"}, or
from \code{cor_method = "fiml"}) does not rely on that approximation and is still
reported, so the fit indices are unaffected.
\item \strong{"sandwich"} returns robust ("sandwich") standard errors estimated from raw data,
combining the estimator weight with a distribution-free covariance of the
correlations, so it stays valid under non-normality and weight misspecification
(Browne, 1984; Satorra & Bentler, 1994). It is available either for ordinal data with
\code{cor_method = "poly"} or \code{"tetra"} and \code{estimator} one of \code{"ML"}, \code{"ULS"}, or
\code{"DWLS"}, or for continuous data with \code{cor_method = "pearson"} and
\code{estimator = "ML"} or \code{"ULS"}. It reports the same coverage as \code{"information"} and
additionally fills the model fit's chi-square block with a scaled chi-square (see
\emph{Fit indices}); the statistic reported as \code{chi} is always the scaled-and-shifted one
(flagged by \code{chi_scaled_type}), and a mean-adjusted alternative is returned alongside
it as \code{chi_mean_adjusted}.
Because the covariance must describe the same cases as the correlation matrix, the
sandwich (like \code{estimator = "DWLS"}) is computed on the listwise-complete cases; on
data with missing values the reported \code{N}, the correlation matrix, and the point
estimate therefore reflect the complete cases regardless of \code{use}.
\item \strong{"np-boot"} draws a non-parametric (case-resampling) bootstrap and needs raw data.
A correlation matrix carries no cases to resample; alone among the unsupported
combinations this one does not error but warns and downgrades \code{se} to \code{"none"}, so
the fit is returned without an \code{SE} slot.
It is the most general method -- available for any \code{estimator}, \code{rotation}, and
\code{cor_method} -- and the most robust to non-normality and misfit, at the cost of speed;
its intervals are bootstrap percentile intervals. The replicate fits run across
replicates with the \code{future} framework; by default they run sequentially, but
registering a plan with \code{\link[future:plan]{future::plan()}} runs them in parallel instead, as the
examples show. With a fixed \code{seed} the bootstrap is reproducible and yields the same
result regardless of the number of workers. Under \code{cor_method = "fiml"} each resample
also re-runs the EM moment estimation and is therefore slow, so a smaller \code{b_boot}
may be advisable.

The percentile intervals are centred on the point estimate for the loadings, the
factor correlations, the structure coefficients, and the residuals, but \strong{not} for
the indices derived from the chi-square (\code{RMSEA}, \code{AIC}, \code{BIC}, and \code{ECVI}): a
resample already carries the model's own misfit, so those intervals ride upward and
can even place the point estimate below their own lower bound. Read them as a spread
rather than a range for the point estimate; correcting that shift needs resampling
from a population transformed to fit the model (Bollen & Stine, 1992), which is not
what this bootstrap does. \code{CFI} and \code{TLI} are not affected, being ratios in which the
baseline chi-square shifts along with the model one.
}

The analytic methods (\code{"information"} and \code{"sandwich"}) are not available with the
\code{"promax"} or \code{"simplimax"} rotations, which have no supported analytic route for the
rotated standard errors; use \code{"np-boot"} there. Under \code{cor_method = "fiml"},
\code{"information"} and \code{"sandwich"} instead return, for \code{estimator = "ML"} or \code{"ULS"},
corrected two-stage sandwich standard errors (Yuan & Bentler, 2000; Savalei & Bentler,
2009). \code{estimator = "PAF"} carries no Stage-2 weight to build the sandwich from, so use
\code{se = "np-boot"} there instead.
}

\subsection{Fit indices}{

For ML and ULS, \code{\link[=efa_fit]{efa_fit()}} returns the model chi-square (with its p-value and degrees of
freedom), the Comparative Fit Index (CFI; Bentler, 1990), the Tucker-Lewis Index (TLI,
also called the non-normed fit index; Tucker & Lewis, 1973), the Root Mean Square Error
of Approximation (RMSEA) with its 90\% confidence interval (Browne & Cudeck, 1992), the
Akaike and Bayesian Information Criteria (AIC, BIC), the Expected Cross-Validation Index
(ECVI; Browne & Cudeck, 1989), the Root Mean Squared Residual (RMSR), the Standardized
Root Mean Squared Residual (SRMR; Bentler, 1995), and the common-part-accounted-for
(CAF) index (Lorenzo-Seva, Timmerman, & Kiers, 2011). They come with the
independence-baseline statistics \code{chi_null}, \code{df_null}, and \code{p_null}. On the unscaled ML
and ULS paths \code{chi_null} is Bartlett's test of sphericity; a scaled chi-square, and the
two-stage statistic of a \code{cor_method = "fiml"} fit, each carry their own baseline
instead.

The degrees of freedom depend on the number of variables and factors; the baseline
degrees of freedom \code{df_null} are \eqn{p(p - 1)/2} for \eqn{p} variables.

RMSR is the root mean square of the off-diagonal residuals; SRMR rescales it by a
fixed factor that depends only on the number of variables. The print and summary
methods show SRMR, not RMSR; RMSR remains in the returned object for backward
compatibility. Both are computed over the same residuals, so an unavailable residual
leaves both \code{NA}. (\pkg{psych}'s \code{rms} uses a different divisor and equals
RMSR\eqn{/\sqrt{2}}; the two are not directly comparable.)

The model chi-square is the Bartlett-corrected discrepancy (matching
\code{\link[stats:factanal]{stats::factanal()}} for ML). For ULS it is the same maximum-likelihood discrepancy,
evaluated at the ULS-fitted solution, as in \code{\link[psych:fa]{psych::fa()}} -- not the least-squares
criterion \pkg{lavaan} reports as its standard ULS test statistic, so the two are not
comparable. AIC and BIC are built on this chi-square and can therefore be negative;
ECVI (built on the same chi-square plus a non-negative penalty) cannot. Because of the
Bartlett correction, ECVI differs slightly from the uncorrected Browne-Cudeck form
reported by \pkg{lavaan} and Mplus (their AIC/BIC use an unrelated log-likelihood-based
formula, so they are not comparable). On the unscaled ML/ULS path, CFI and
TLI are computed on a slightly different discrepancy scale than the reported
chi-square, so you cannot recompute one from the other by hand there; on the scaled
(sandwich) and FIML paths, the reported chi and chi_null are exactly the pair the
indices use.

Which indices are reported depends on the estimator:
\itemize{
\item \strong{ML and ULS} compute the full set above.
\item \strong{PAF} returns only the descriptive residual indices (RMSR, SRMR, CAF) and df; the
printed model-fit block shows CAF and SRMR. The chi-square-based indices are \code{NA},
because PAF minimises no discrepancy.
\item \strong{DWLS} by default returns only RMSR, SRMR, CAF, and df, because the ordinary
maximum-likelihood discrepancy is not its fit function. When \code{se = "sandwich"}, a
scaled chi-square and the CFI, TLI, and RMSEA derived from it are reported (AIC and
BIC remain \code{NA}); that statistic is a two-stage correction applied to the polychoric
correlation residuals (Browne, 1984), not identical to the full WLSMV test of
\pkg{lavaan} or Mplus, which also corrects for the response thresholds.
\item \strong{\code{cor_method = "fiml"}} (with ML or ULS) reports two-stage-corrected statistics
(Yuan, Marshall, & Bentler, 2002); AIC, BIC, and ECVI are left \code{NA}, as for any
scaled chi-square. The correction itself can be degenerate -- typically with a small
sample, a high proportion of missing values, or near-collinear variables -- in which
case an uncorrected likelihood-ratio statistic is reported in its place with a
warning, and the print methods label that line \code{uncorrected} rather than \code{scaled};
read its p-value and the CFI, TLI, and RMSEA derived from it as indicative only.
}

Beyond the estimator, the chi-square and everything derived from it are \code{NA} whenever
the statistic itself is undefined: when \code{N} is not supplied, when the model is
underidentified (a negative df), and when a positive \code{N} is too small relative to the
number of variables and factors for the small-sample correction to remain valid. Each
case raises its own warning. The residual summaries (RMSR, SRMR, CAF) and the degrees
of freedom are still returned there, but residual size does not establish that a model
is identified: below zero degrees of freedom a near-zero residual is an artefact of
over-parameterisation, not close fit.

Whenever the chi-square is a scaled one (\code{se = "sandwich"}, or a \code{cor_method = "fiml"}
fit whose correction could be formed), AIC, BIC, and ECVI are \code{NA}; see the
\code{fit_indices} entry in Value for the additional components then returned. AIC, BIC,
and ECVI are \code{NA} on every \code{cor_method = "fiml"} fit, including the uncorrected
fallback above. Lorenzo-Seva, Timmerman, and Kiers (2011) describe CAF as ranging from
0 to 1, with values near 1 indicating close fit; that does not hold here, where a
well-fitting model produces a CAF near 0.5, not near 1. Read it as a relative rather
than an absolute measure.
}

\subsection{Available combinations}{

Not every estimator, rotation, standard-error, and correlation method can be combined:
\itemize{
\item \strong{Estimator and correlation method.} \code{estimator = "DWLS"} requires ordinal data with
\code{cor_method = "poly"} or \code{"tetra"}. \code{cor_method = "fiml"} works with PAF, ML, and ULS
(not DWLS) and needs raw data with missing values.
\item \strong{Standard errors.} \code{se = "information"} requires \code{estimator = "ML"} and
\code{cor_method = "pearson"} or \code{"fiml"}, and can be computed from a correlation matrix
when \code{N} is supplied. \code{se = "sandwich"} requires raw data,
with either a polychoric/tetrachoric \code{cor_method} (ML, ULS, or DWLS) or a Pearson
\code{cor_method} (ML or ULS); it is not available for PAF. Under \code{cor_method = "fiml"},
\code{"information"} and \code{"sandwich"} are available for ML and ULS only and both return the
corrected two-stage sandwich. \code{se = "np-boot"} requires raw data and works with any
estimator, rotation, and correlation method. Neither \code{"information"} nor \code{"sandwich"}
is available with the \code{"promax"} or \code{"simplimax"} rotations.
\item \strong{Fit indices.} The chi-square-based indices are available for ML and ULS (and, as
scaled statistics, for \code{cor_method = "fiml"} and for DWLS with \code{se = "sandwich"}); PAF
and DWLS otherwise report only the descriptive residual indices.
}
}
}
\examples{

# Principal axis factoring with oblimin rotation
mod_oblimin <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                       rotation = "oblimin")
mod_oblimin
summary(mod_oblimin)

# ML estimation with oblimin rotation
mod_oblimin <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                       estimator = "ML", rotation = "oblimin")
mod_oblimin
summary(mod_oblimin)

# Tuning knobs are supplied through the control objects. Here the SPSS preset is
# used for the estimation and rotation, with the maximum PAF iterations raised.
mod_spss <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                    rotation = "promax",
                    estimate_control = estimate_control(type = "SPSS", max_iter = 500),
                    rotate_control = rotate_control(type = "SPSS"))
mod_spss

# Analytic (expected-information) standard errors for the above
ML_info <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                   estimator = "ML", rotation = "oblimin", se = "information")
ML_info
summary(ML_info)

\donttest{
# Robust (sandwich) standard errors and a scaled chi-square for ordinal raw data.
# These need a polychoric/tetrachoric correlation method and estimator ML, ULS, or DWLS.
DWLS_rob <- efa_fit(DOSPERT_raw, n_factors = 6, cor_method = "poly",
                    estimator = "DWLS", rotation = "oblimin", se = "sandwich")
DWLS_rob
summary(DWLS_rob)

# The same robust SEs and scaled chi-square for continuous data: a Pearson
# correlation with estimator ML or ULS (the fourth-moment ADF covariance).
ML_rob <- efa_fit(GRiPS_raw, n_factors = 1, cor_method = "pearson",
                  estimator = "ML", rotation = "none", se = "sandwich")
ML_rob
summary(ML_rob)
}

\donttest{
# Two-stage FIML correlations from raw data with missing values: the saturated
# multivariate-normal moments are EM-estimated (assuming the data are missing at
# random) and the standardized covariance is analysed.
x_miss <- GRiPS_raw
x_miss[cbind(1:20, 1)] <- NA
efa_fiml <- efa_fit(x_miss, n_factors = 1, estimator = "ML", cor_method = "fiml")
efa_fiml
}

\dontrun{
# Bootstrap standard errors from raw data, reproducible via a fixed seed and run
# in parallel across replicates. future::plan() returns the plan it replaces, so
# on.exit() puts the session back as it was -- also if the fit fails.
efa_boot <- local({
  old_plan <- future::plan(future::multisession, workers = 2)
  on.exit(future::plan(old_plan), add = TRUE)
  efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF", rotation = "none",
          se = "np-boot", b_boot = 1000, seed = 42)
})
}

}
\seealso{
\code{\link[=estimate_control]{estimate_control()}} and \code{\link[=rotate_control]{rotate_control()}} for the estimation and rotation
tuning knobs. \code{\link[=efa_retain]{efa_retain()}} for choosing \code{n_factors}, and \code{\link[=efa_scores]{efa_scores()}},
\code{\link[=efa_reliability]{efa_reliability()}}, \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}, and \code{\link[=efa_compare]{efa_compare()}} for working with the
fitted solution.

Other factor analysis:
\code{\link[=efa_average]{efa_average()}},
\code{\link[=efa_group]{efa_group()}},
\code{\link[=efa_mi]{efa_mi()}},
\code{\link[=plot.efa_group]{plot.efa_group()}},
\code{\link[=print.efa_group]{print.efa_group()}}
}
\concept{factor analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit-indices.R
\name{.compute_vars}
\alias{.compute_vars}
\title{Compute explained variances from loadings}
\usage{
.compute_vars(L_unrot, L_rot, Phi = NULL)
}
\arguments{
\item{L_unrot}{matrix. Unrotated factor loadings.}

\item{L_rot}{matrix. Rotated factor loadings.}

\item{Phi}{matrix. Factor intercorrelations. Provide only if oblique rotation
is used.}
}
\value{
A matrix with sum of squared loadings, proportion explained variance
from total variance per factor, same as previous but cumulative, Proportion
of explained variance from total explained variance, and same as previous but
cumulative. The three cumulative and common-variance rows are omitted when
\code{L_rot} has a single column, where they would only repeat the two above them,
so the result has two rows there and five otherwise.
}
\description{
From unrotated loadings compute the communalities and uniquenesses for total
variance. Compute explained variances per factor from rotated loadings (and
factor intercorrelations Phi if oblique rotation was used).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa.R
\name{print.efa}
\alias{print.efa}
\alias{print.efa_mi}
\alias{format.efa}
\alias{format.efa_mi}
\alias{summary.efa}
\alias{summary.efa_mi}
\alias{print.summary.efa}
\alias{format.summary.efa}
\title{Print and summarise an efa object}
\usage{
\method{print}{efa}(x, ...)

\method{print}{efa_mi}(x, ...)

\method{format}{efa}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  sort_loadings = c("none", "primary", "clustered"),
  show_loading_legend = TRUE,
  max_factors_per_block = NULL,
  ...
)

\method{format}{efa_mi}(x, ...)

\method{summary}{efa}(
  object,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  ci = c("auto", "none", "separate"),
  ci_filter = c("salient", "all", "nonzero"),
  diagnostics_top_n = 10,
  residual_cutoff = 0.1,
  residual_top_n = 10,
  show_structure = TRUE,
  sort_loadings = c("none", "primary", "clustered"),
  show_loading_legend = TRUE,
  cross_loading_cutoff = cutoff,
  min_primary_gap = 0.2,
  min_salient_per_factor = 3,
  max_factors_per_block = NULL,
  show_mi_diagnostics = NULL,
  ...
)

\method{summary}{efa_mi}(object, ...)

\method{print}{summary.efa}(x, ...)

\method{format}{summary.efa}(x, ...)
}
\arguments{
\item{x, object}{An object of class \code{efa} (from \code{\link[=efa_fit]{efa_fit()}}) or \code{efa_mi} (from \code{\link[=efa_mi]{efa_mi()}}); for
the \code{summary.efa} methods, the object returned by \code{summary()}.}

\item{...}{Further arguments passed to \code{\link[=print.efa_loadings]{print.efa_loadings()}}.}

\item{cutoff}{numeric. The absolute value at or above which loadings are
emphasised in the loading table. Default is .3.}

\item{digits}{numeric. Number of decimal places for the printed tables.
Default is 3.}

\item{max_name_length}{numeric. Maximum length of the variable names to
display; longer names are cut from the right, or abbreviated where cutting
would give two variables the same label. Applies to every table that names
variables. \code{name_style} (see \code{\link[=print.efa_loadings]{print.efa_loadings()}}) can be passed through
\code{...} to choose the shortening explicitly, but it reaches the loading table
only; the confidence-interval and simple-structure tables always shorten by
cutting.}

\item{sort_loadings}{character. Optional row sorting for the loading table.
See \code{\link[=print.efa_loadings]{print.efa_loadings()}}.}

\item{show_loading_legend}{logical. Whether to print a short legend for the
loading-table styling. Default is \code{TRUE}. The legend is only printed when
that styling is actually rendered (a colour-capable console); in plain
output it is omitted and this argument has no effect.}

\item{max_factors_per_block}{numeric or \code{NULL}. Maximum number of factor
columns per loading-table block. If \code{NULL}, chosen from the console width.}

\item{ci}{character. Which confidence intervals \code{summary()} shows, if
available. \code{"auto"} and \code{"separate"} print CI sections when CIs were
computed; \code{"none"} suppresses them. Default is \code{"auto"}.}

\item{ci_filter}{character. Which loading CIs \code{summary()} prints: \code{"salient"}
(default), \code{"all"}, or \code{"nonzero"}; see Details.}

\item{diagnostics_top_n}{numeric. Maximum number of item-level entries
\code{summary()} prints per simple-structure diagnostic.}

\item{residual_cutoff}{numeric. Absolute residual cutoff for the residual
diagnostics in \code{summary()}. Default is .1.}

\item{residual_top_n}{numeric. Maximum number of residuals \code{summary()} prints.
Use \code{Inf} to print all residuals above \code{residual_cutoff}.}

\item{show_structure}{logical. Whether \code{summary()} prints the structure matrix
for oblique solutions when available. Default is \code{TRUE}.}

\item{cross_loading_cutoff}{numeric. Cutoff for counting cross-loadings in the
\code{summary()} diagnostics. Defaults to \code{cutoff}.}

\item{min_primary_gap}{numeric. Minimum desired absolute difference between the
largest and second-largest absolute loading of an item, used in the
\code{summary()} diagnostics.}

\item{min_salient_per_factor}{numeric. Minimum number of salient indicators per
factor used in the \code{summary()} diagnostics. Default is 3.}

\item{show_mi_diagnostics}{logical or \code{NULL}. Whether \code{summary()} prints a
multiple-imputation uncertainty summary for pooled EFAs. \code{NULL} shows it for
pooled objects.}
}
\value{
\code{print()} and the print method for \code{summary.efa} objects return their
argument invisibly. \code{format()} returns a character vector with the report
lines. \code{summary()} returns an object of class \code{summary.efa}.
}
\description{
\code{print()} shows a concise overview of an \code{\link[=efa_fit]{efa_fit()}} or \code{\link[=efa_mi]{efa_mi()}} solution:
a model header, the loading matrix (with the factor intercorrelations for
oblique solutions), the variances accounted for, and the model fit.
\code{summary()} returns a \code{summary.efa} object whose print method adds the full
diagnostics: model and simple-structure diagnostics, confidence-interval
tables, the structure matrix, multiple-imputation uncertainty (for pooled
objects), and residual diagnostics. \code{format()} assembles the same report and
returns it as a character vector; \code{print()} is \code{cat(format(x), sep = "\\n")}.
The lines follow the active console theme, so they are plain when colours are
disabled (for example when captured into a file or stripped with
\code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\details{
The methods are shared by single-imputation \code{efa} objects and pooled
\code{efa_mi} objects. For \code{efa_mi} objects the header reports the number
of imputations and the alignment/pooling settings; confidence intervals and a
multiple-imputation uncertainty summary are shown by \code{summary()} when the
pooled object carries bootstrap/MI quantities.

In \code{summary()}, \code{ci_filter} controls which loading intervals are shown:
\code{"salient"} reports intervals for loadings whose absolute point estimate is at
least \code{cutoff}, \code{"nonzero"} reports intervals excluding zero, and \code{"all"}
reports every finite interval.
}
\examples{
mod <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
               estimator = "PAF", rotation = "promax")
mod

# The full diagnostics, CI tables, and residual diagnostics:
summary(mod)

# format() returns the same lines as a character vector, e.g. for a report file:
writeLines(format(mod))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFA_AVERAGE.R
\name{EFA_AVERAGE}
\alias{EFA_AVERAGE}
\title{Model averaging across different EFA methods and types}
\source{
Grieder, S., & Steiner, M.D. (2020). Algorithmic Jingle Jungle:
A Comparison of Implementations of Principal Axis Factoring and Promax Rotation
 in R and SPSS. Manuscript in Preparation.

Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
rotation to oblique simple structure. British Journal of Statistical Psychology,
17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x

Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
Research, 46, 340-364, doi: 10.1080/00273171.2011.564527

Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233
}
\usage{
EFA_AVERAGE(
  x,
  n_factors,
  N = NA,
  method = "PAF",
  rotation = "promax",
  type = "none",
  averaging = c("mean", "median"),
  trim = 0,
  salience_threshold = 0.3,
  max_iter = 10000,
  init_comm = c("smc", "mac", "unity"),
  criterion = c(0.001),
  criterion_type = c("sum", "max_individual"),
  abs_eigen = c(TRUE),
  varimax_type = c("svd", "kaiser"),
  normalize = TRUE,
  k_promax = 2:4,
  k_simplimax = ncol(x),
  P_type = c("norm", "unnorm"),
  precision = 1e-05,
  start_method = c("psych", "factanal"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall"),
  show_progress = TRUE
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations. If raw data is entered, the correlation matrix is found from the
data.}

\item{n_factors}{numeric. Number of factors to extract.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used. If input is a correlation matrix and \code{N} = NA
(default), not all fit indices can be computed.}

\item{method}{character vector. Any combination of  "PAF", "ML", and "ULS",
to use principal axis factoring, maximum likelihood, or unweighted least
squares (also called minres), respectively, to fit the EFAs. Default is "PAF".}

\item{rotation}{character vector. Either perform no rotation ("none"),
any combination of orthogonal rotations ("varimax", "equamax", "quartimax", "geominT",
"bentlerT", and "bifactorT"; using "orthogonal" runs all of these), or of
oblique rotations ("promax", "oblimin", "quartimin", "simplimax", "bentlerQ",
"geominQ", and "bifactorQ"; using "oblique" runs all of these). Rotation types
(no rotation, orthogonal rotations, and oblique rotations) cannot be mixed.
Default is "promax".}

\item{type}{character vector. Any combination of "none" (default), "EFAtools",
"psych", and "SPSS" can be entered. "none" allows the specification of various
combinations of the arguments controlling both factor extraction methods and
the rotations. The others ("EFAtools", "psych", and "SPSS"), control the execution
of the respective factor extraction method and rotation to be in line with how
it is executed in this package (i.e., the respective default procedure), in the
psych package, and in SPSS. A specific psych implementation exists for PAF, ML, varimax,
and promax. The SPSS implementation exists for PAF, varimax, and promax. For
details, see \code{\link{EFA}}.}

\item{averaging}{character. One of "mean" (default), and "median". Controls
whether the different results should be averaged using the (trimmed) mean,
or the median.}

\item{trim}{numeric. If averaging is set to "mean", this argument controls
the trimming of extremes (for details see \code{\link[base:mean]{base::mean}}).
By default no trimming is done (i.e., trim = 0).}

\item{salience_threshold}{numeric. The threshold to use to classify a pattern
coefficient or loading as salient (i.e., substantial enough to assign it to
a factor). Default is 0.3. Indicator-to-factor correspondences will be inferred
based on this threshold. Note that this may not be meaningful if rotation = "none"
and n_factors > 1 are used, as no simple structure is present there.}

\item{max_iter}{numeric. The maximum number of iterations to perform after which
the iterative PAF procedure is halted with a warning. Default is 10,000. Note
that non-converged procedures are excluded from the averaging procedure.}

\item{init_comm}{character vector. Any combination of "smc", "mac", and "unity".
Controls the methods to estimate the initial communalities in \code{PAF} if
"none" is among the specified types. "smc" will use squared multiple
correlations, "mac" will use maximum absolute correlations, "unity" will use
1s (for details see \code{\link{EFA}}). Default is \code{c("smc", "mac", "unity")}.}

\item{criterion}{numeric vector. The convergence criterion used for PAF if
"none" is among the specified types.
If the change in communalities from one iteration to the next is smaller than
this criterion the solution is accepted and the procedure ends.
Default is \code{0.001}.}

\item{criterion_type}{character vector. Any combination of "max_individual" and
"sum". Type of convergence criterion used for PAF if "none" is among the
specified types. "max_individual" selects the maximum change in any of the
communalities from one iteration to the next and tests it against the
specified criterion. "sum" takes the difference of
the sum of all communalities in one iteration and the sum of all communalities
in the next iteration and tests this against the criterion
(for details see \code{\link{EFA}}). Default is \code{c("sum", "max_individual")}.}

\item{abs_eigen}{logical vector. Any combination of TRUE and FALSE.
Which algorithm to use in the PAF iterations if "none" is among the specified
types. If FALSE, the loadings are computed from the eigenvalues. This is also
used by the \code{\link[psych:fa]{psych::fa}} function. If TRUE the
loadings are computed with the absolute eigenvalues as done by SPSS
(for details see \code{\link{EFA}}). Default is \code{TRUE}.}

\item{varimax_type}{character vector. Any combination of "svd" and "kaiser".
The type of the varimax rotation performed if "none" is among the specified
types and "varimax", "promax", "orthogonal", or "oblique" is among the specified
rotations. "svd" uses singular value decomposition, as
\link[stats:varimax]{stats::varimax} does, and "kaiser" uses the varimax
procedure performed in SPSS. This is the original procedure from Kaiser (1958),
but with slight alterations in the varimax criterion (for details, see
\code{\link{EFA}} and Grieder & Steiner, 2020).
Default is \code{c("svd", "kaiser")}.}

\item{normalize}{logical vector. Any combination of TRUE and FALSE.
\code{TRUE} performs a kaiser normalization before the specified rotation(s).
Default is \code{TRUE}.}

\item{k_promax}{numeric vector. The power used for computing the target matrix
P in the promax rotation if "none" is among the specified types and "promax"
or "oblique" is among the specified rotations. Default is \code{2:4}.}

\item{k_simplimax}{numeric. The number of 'close to zero loadings' for the
simplimax rotation (see \code{\link[GPArotation:GPA]{GPArotation::GPFoblq}})
if "simplimax" or "oblique" is among the specified rotations. Default
is \code{ncol(x)}, where x is the entered data.}

\item{P_type}{character vector. Any combination of "norm" and "unnorm".
This specifies how the target matrix P is computed in promax rotation if
"none" is among the specified types and "promax" or "oblique" is among the
specified rotations. "unnorm" will use the unnormalized target matrix as
originally done in Hendrickson and White (1964). "norm" will use a
normalized target matrix (for details see \code{\link{EFA}}).
Default is \code{c("norm", "unnorm")}.}

\item{precision}{numeric vector. The tolerance for stopping in the rotation
procedure(s). Default is 10^-5.}

\item{start_method}{character vector. Any combination of "psych" and "factanal".
How to specify the starting values for the optimization procedure for ML.
"psych" takes the starting values specified in \link[psych:fa]{psych::fa}.
"factanal" takes the starting values specified in the
\link[stats:factanal]{stats::factanal} function. Default is
\code{c("psych", "factanal")}.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor}} if raw data
is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Passed to \code{\link[stats:cor]{stats::cor}}.
Default is "pearson".}

\item{show_progress}{logical. Whether a progress bar should be shown in the
console. Default is TRUE.}
}
\value{
A list of class EFA_AVERAGE containing
\item{orig_R}{Original correlation matrix.}
\item{h2}{A list with the average, standard deviation, minimum, maximum, and
range of the final communality estimates across the factor solutions.}
\item{loadings}{A list with the average, standard deviation, minimum, maximum,
and range of the final loadings across the factor solutions. If rotation was
"none", the unrotated loadings, otherwise the rotated loadings (pattern
coefficients).}
\item{Phi}{A list with the average, standard deviation, minimum, maximum, and
range of the factor intercorrelations across factor solutions obtained with
oblique rotations.}
\item{ind_fac_corres}{A matrix with each cell containing the proportion of
the factor solutions in which the respective indicator-to-factor correspondence
occurred, i.e., in which the loading exceeded the specified salience threshold.
Note: Rowsums can exceed 1 due to cross-loadings.}
\item{vars_accounted}{A list with the average, standard deviation, minimum,
maximum, and range of explained variances and sums of squared loadings across
the factor solutions. Based on the unrotated loadings.}
\item{fit_indices}{A matrix containing the average, standard deviation,
minimum, maximum, and range for all applicable fit indices across the respective
factor solutions, and the degrees of freedom (df). If the method argument
contains ML or ULS: Fit indices derived
from the unrotated factor loadings: Chi Square (chisq), including significance
level, Comparative Fit Index (CFI), Root Mean Square Error of Approximation
(RMSEA), Akaike Information Criterion (AIC), Bayesian Information Criterion
(BIC)and the common part accounted for (CAF) index as proposed by
Lorenzo-Seva, Timmerman, & Kiers (2011). For PAF, only the CAF can be
calculated (see details).}
\item{implementations_grid}{A matrix containing, for each performed EFA,
the setting combination, if an error occurred (logical), the error message
(character), an integer code for convergence as returned by
\code{\link[stats:optim]{stats:optim}} (0 indicates successful completion.),
if heywood cases occurred (logical, see details for definition), if the
solution was admissible (logical, see details for definition), and the fit
indices.}
\item{efa_list}{A list containing the outputs of all performed EFAs. The names
correspond to the rownames from the implementations_grid.}
\item{settings}{A list of the settings used.}
}
\description{
Not all EFA procedures always arrive at the same solution. This function allows
you perform a number of EFAs from different methods (e.g., Maximum Likelihood
and Principal Axis Factoring), with different implementations (e.g., the SPSS
and psych implementations of Principal Axis Factoring), and across different
rotations of the same type (e.g., multiple oblique rotations, like promax and
oblimin). EFA_AVERAGE will then run all these EFAs (using the \code{\link{EFA}}
function) and provide a summary across the different solutions.
}
\details{
As a first step in this function, a grid is produced containing the setting
combinations for the to-be-performed EFAs. These settings are then entered as
arguments to the \code{\link{EFA}} function and the EFAs are run in a second
step. After all EFAs are run, the factor solutions are averaged and their
variability determined in a third step.

The grid containing the setting combinations is produced based on the entries
to the respective arguments. To this end, all possible combinations resulting
in unique EFA models are considered. That is, if, for example, the \code{type}
argument was set to \code{c("none", "SPSS")} and one combination of the specific
settings entered was identical to the SPSS combination, this combination
would be included in the grid and run only once. We include here a list
of arguments that are only evaluated under specific conditions:

The arguments \code{init_comm}, \code{criterion}, \code{criterion_type},
\code{abs_eigen} are only evaluated if "PAF" is included in \code{method}
and "none" is included in \code{type}.

The argument \code{varimax_type} is only evaluated if "varimax", "promax",
"oblique", or "orthogonal" is included in \code{rotation} and "none" is
included in \code{type}.

The argument \code{normalize} is only evaluated if \code{rotation} is not
set to "none" and "none" is included in \code{type}.

The argument \code{k_simplimax} is only evaluated if "simplimax" or "oblique"
is included in \code{rotation}.

The arguments \code{k_promax} and \code{P_type} are only evaluated if
"promax" or "oblique" is included in \code{rotation} and "none" is included
in \code{type}.

The argument \code{start_method} is only evaluated if "ML" is included in
\code{method}.

To avoid a bias in the averaged factor solutions from problematic solutions,
these are excluded prior to averaging. A solution is deemed problematic if
at least one of the following is true: an error occurred, the model did not
converge, or there is at least one Heywood case (defined as a loading or communality of >= .998).
Information on errors, convergence, and Heywood cases are returned in the
implementations_grid and a summary of these is given when printing the output.
In addition to these, information on the admissibility of the factor solutions
is also included. A solution was deemed admissible if (1) no error occurred,
(2) the model converged, (3) no Heywood cases are present, and (4) there are
at least two salient loadings (i.e., loadings exceeding the specified
\code{salience_threshold}) for each factor. So, solutions failing one of the
first three of these criteria of admissibility are also deemed problematic and
therefore excluded from averaging. However, solutions failing only
the fourth criterion of admissibility are still included for averaging.
Finally, if all solutions are problematic (e.g., all solutions contain
Heywood cases), no averaging is performed and the respective outputs are NA.
In this case, the implementations_grid should be inspected to see if there
are any error messages, and the separate EFA solutions that are also included
in the output can be inspected as well, for example, to see where Heywood
cases occurred.

A core output of this function includes the average, minimum, and maximum
loadings derived from all non-problematic (see above) factor solutions. Please
note that these are not entire solutions, but the matrices include the average,
minimum, or maximum value for each cell (i.e., each loading separately). This
means that, for example, the matrix with the minimum loadings will contain
the minimum value in any of the factor solutions for each specific loading,
and therefore most likely contains loadings from different factor solutions.
The matrices containing the minimum and maximum factor solutions can
therefore not be interpreted as whole factor solutions.

The output also includes information on the average, minimum, maximum, and
variability of the fit indices across the non-problematic factor solutions.
It is important to note that not all fit indices are computed for all fit
methods: For ML and ULS, all fit indices can be computed, while for PAF, only
the common part accounted for (CAF) index (Lorenzo-Seva, Timmerman, & Kiers, 2011)
can be computed. As a consequence, if only "PAF" is included in the
\code{method} argument, averaging can only be performed for the CAF, and the
other fit indices are NA. If a combination of "PAF" and "ML" and/or "ULS" are
included in the \code{method} argument, the CAF is averaged across all non-
problematic factor solutions, while all other fit indices are only averaged
across the ML and ULS solutions. The user should therefore keep in mind that
the number of EFAs across which the fit indices are averaged can diverge for
the CAF compared to all other fit indices.
}
\examples{
\dontrun{
# Averaging across different implementations of PAF and promax rotation (72 EFAs)
Aver_PAF <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500)

# Use median instead of mean for averaging (72 EFAs)
Aver_PAF_md <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                           averaging = "median")

# Averaging across different implementations of PAF and promax rotation,
# and across ULS and different versions of ML (108 EFAs)
Aver_meth_ext <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                             method = c("PAF", "ULS", "ML"))

# Averaging across one implementation each of PAF (EFAtools type), ULS, and
# ML with one implementation of promax (EFAtools type) (3 EFAs)
Aver_meth <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                         method = c("PAF", "ULS", "ML"), type = "EFAtools",
                         start_method = "psych")

# Averaging across different oblique rotation methods, using one implementation
# of ML and one implementation of promax (EFAtools type) (7 EFAs)
Aver_rot <- EFA_AVERAGE(test_models$baseline$cormat, n_factors = 3, N = 500,
                         method = "ML", rotation = "oblique", type = "EFAtools",
                         start_method = "psych")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_procrustes.R
\name{.gpa_consensus_target}
\alias{.gpa_consensus_target}
\title{Generalized Procrustes Analysis consensus target across loading matrices}
\usage{
.gpa_consensus_target(
  unrotated_list,
  init_targets = NULL,
  rotation = c("orthogonal", "oblique"),
  start = 1,
  multi_start = FALSE,
  starts = NULL,
  tol = 0.001,
  loss_tol = 1e-06,
  loss_patience = 5,
  convergence = c("either", "target", "loss", "both"),
  min_iter = 2,
  max_iter = 200,
  alpha = 1,
  match_target = TRUE,
  hyper_cutoff = 0.15,
  verbose = FALSE
)
}
\arguments{
\item{unrotated_list}{List of unrotated loading matrices to be aligned. All
matrices must be numeric, finite, and have identical dimensions.}

\item{init_targets}{Optional list of starting target matrices. These are
typically rotated loading matrices from the corresponding analyses. If
\code{NULL}, \code{unrotated_list} is used.}

\item{rotation}{Character string, either \code{"orthogonal"} or \code{"oblique"}.}

\item{start}{Either a single integer selecting an element of \code{init_targets},
or an explicit target matrix. Used when \code{multi_start = FALSE}.}

\item{multi_start}{Logical. If \code{FALSE}, perform one consensus-target run. If
\code{TRUE}, repeat the single-start algorithm for each element of \code{starts}.}

\item{starts}{Integer vector selecting elements of \code{init_targets} used as
starting targets when \code{multi_start = TRUE}. If \code{NULL}, all elements of
\code{init_targets} are used. Duplicate entries are removed.}

\item{tol}{Positive relative Frobenius-norm convergence tolerance for the
outer target update.}

\item{loss_tol}{Positive tolerance for the relative change in the outer
consensus loss. If \code{NULL}, loss-based convergence is disabled. It cannot be
\code{NULL} when \code{convergence} is \code{"loss"} or \code{"both"}.}

\item{loss_patience}{Positive integer. Number of consecutive iterations with
relative loss change below \code{loss_tol} required for loss-based convergence.}

\item{convergence}{Character string controlling the stopping rule. \code{"either"}
stops when either target or loss convergence is satisfied; \code{"target"} uses
only target change; \code{"loss"} uses only loss change; \code{"both"} requires both.}

\item{min_iter}{Non-negative integer. Minimum number of outer iterations
before convergence can be declared.}

\item{max_iter}{Positive integer. Maximum number of outer consensus
iterations.}

\item{alpha}{Damping factor for the target update. \code{alpha = 1} uses the full
centroid update. Smaller values, such as \code{0.5}, can reduce oscillation.}

\item{match_target}{Logical. If \code{TRUE}, the updated centroid is signed and
column-matched to the previous target before convergence is evaluated.}

\item{hyper_cutoff}{Non-negative cutoff used by \code{.hyperplane_count()} for
summary output.}

\item{verbose}{Logical; if \code{TRUE}, print convergence messages for the outer
loop.}
}
\value{
A list with the converged target, aligned matrices, pooled loadings,
pooled \code{Phi}, convergence history, inner-alignment diagnostics, and
hyperplane summaries. If \code{multi_start = TRUE}, the \code{multi_start} element also
contains the per-start losses, convergence indicators, run summaries, all
run objects, and between-run Tucker congruence matrices.
}
\description{
Internal helper that constructs a Generalized Procrustes Analysis (GPA)
consensus target across a list of loading matrices and returns the aligned
loadings, the centroid target, and convergence diagnostics. Used by
\code{\link[=efa_mi]{efa_mi()}} under \code{target_method = "consensus"} to build a common
rotation target across imputations. Oblique rotations are not supported
here: the iteration is degenerate for oblique transforms with more than
one factor (cf. Lorenzo-Seva & Van Ginkel 2016, who use a Promin step on
top of the centroid rather than iterated oblique Procrustes); callers
should pass the unrotated solutions of an orthogonal rotation, or use
\code{target_method = "first_target"}.
}
\details{
The iteration alternates two steps:
\enumerate{
\item each loading matrix is aligned to the current target with \code{efa_procrustes()};
\item the target is updated to the elementwise centroid of the aligned matrices.
}

The outer loop stops when the target stabilises, when the consensus loss
stabilises, or when both criteria are satisfied.

If \code{multi_start = FALSE}, one consensus run is performed. If
\code{multi_start = TRUE}, the same engine is repeated for the selected starting
targets and the run with the smallest final mean loss is returned as the
main result; all runs and a between-run congruence summary are retained in
the \code{multi_start} component.
}
\references{
Gower, J. C. (1975). Generalized Procrustes analysis. \emph{Psychometrika}, 40,
33-51.

Van Ginkel, J. R., & Kroonenberg, P. M. (2014). Using Generalized
Procrustes Analysis for Multiple Imputation in Principal Component
Analysis. \emph{Journal of Classification}, 31, 242-269.

Lorenzo-Seva, U., & Van Ginkel, J. R. (2016). Multiple Imputation of
missing values in exploratory factor analysis of multidimensional scales:
estimating latent trait scores. \emph{Anales de Psicologia}, 32, 596-608.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_bartlett.R
\name{efa_bartlett}
\alias{efa_bartlett}
\title{Bartlett's test of sphericity}
\source{
Bartlett, M. S. (1951). The effect of standardization on a Chi-square
approximation in factor analysis. Biometrika, 38, 337-344.
}
\usage{
efa_bartlett(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used.}

\item{use}{character. The missing-data policy for raw data. Passed to
\code{\link[stats:cor]{stats::cor()}} for \code{"pearson"}, \code{"spearman"}, and \code{"kendall"}; for \code{"poly"} /
\code{"tetra"} the same policies are applied to the raw data before the polychoric
estimation, where \code{"all.obs"} and \code{"everything"} abort on a missing value instead
of returning \code{NA} correlations. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator).
Default is "pearson".}
}
\value{
A list containing
\item{chisq}{The chi square statistic, or \code{NA}, with a warning, if \code{N} is too
small for the Bartlett correction (i.e. \eqn{N - 1 - (2p + 5)/6 \le 0}).}
\item{p_value}{The p value of the chi square statistic, or \code{NA} when \code{chisq}
is \code{NA}.}
\item{df}{The degrees of freedom for the chi square statistic.}
\item{settings}{A list of the settings used.}
}
\description{
This function tests whether a correlation matrix is significantly different
from an identity matrix (Bartlett, 1951). If the Bartlett's test is not
significant, the correlation matrix is not suitable for factor analysis
because the variables show too little covariance.
}
\details{
Bartlett (1951) proposed this statistic to determine a correlation
matrix' suitability for factor analysis. The statistic is approximately
chi square distributed with \eqn{df = \frac{p(p - 1)}{2}} and is given by

\deqn{chi^2 = -log(det(R)) (N - 1 - (2 * p + 5)/6)}

where \eqn{det(R)} is the determinant of the correlation matrix, \eqn{N} is
the sample size, and \eqn{p} is the number of variables.

This test requires multivariate normality. If this condition is not met,
the Kaiser-Meyer-Olkin criterion (\code{\link[=efa_kmo]{efa_kmo()}})
can still be used.

This function was heavily influenced by the \code{\link[psych:cortest.bartlett]{psych::cortest.bartlett()}} function from the psych package.

The \code{efa_bartlett} function can also be called together with the
(\code{\link[=efa_kmo]{efa_kmo()}}) function and with factor retention criteria
in the \code{\link[=efa_retain]{efa_retain()}} function.
}
\examples{
efa_bartlett(test_models$baseline$cormat, N = 500)

}
\seealso{
\code{\link[=efa_kmo]{efa_kmo()}} for another measure to determine
suitability for factor analysis.

\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this function,
\code{\link[=efa_kmo]{efa_kmo()}} and several factor retention criteria.

Other factor analysis suitability:
\code{\link[=efa_kmo]{efa_kmo()}},
\code{\link[=efa_screen]{efa_screen()}},
\code{\link[=print.efa_screen]{print.efa_screen()}}
}
\concept{factor analysis suitability}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{NEST}
\alias{NEST}
\title{Next eigenvalue sufficiency test}
\usage{
NEST(
  x,
  N = NA,
  alpha = 0.05,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_datasets = 1000,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Only needed if x is a correlation
matrix. Must be larger than the number of variables.}

\item{alpha}{numeric. The alpha level to use (i.e., 1-alpha percentile of eigenvalues is used for reference values).}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{NEST} compares the data against simulated continuous reference data.
Default is  \code{"pearson"}.}

\item{n_datasets}{numeric. The number of datasets to simulate. Default is 1000.}

\item{...}{Further arguments passed on to the \code{\link[=efa_fit]{efa_fit()}} fits. For example,
\code{estimator}, to change the estimator (PAF is default), or one of the estimation
tuning knobs (\code{type}, \code{init_comm}, \code{criterion}, \code{criterion_type}, \code{max_iter},
\code{abs_eigen}, \code{start_method}), which are repacked into an \code{\link[=estimate_control]{estimate_control()}}
object so that they tune the fits exactly as they always did.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_nest]{efa_nest()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{NEST()} has been superseded by \code{\link[=efa_nest]{efa_nest()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_nest]{efa_nest()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.COMPARE.R
\name{print.COMPARE}
\alias{print.COMPARE}
\title{Print COMPARE object}
\usage{
\method{print}{COMPARE}(x, ...)
}
\arguments{
\item{x}{list. An object of class COMPARE to be printed}

\item{...}{Further arguments for print.}
}
\description{
Print Method showing a summarized output of the \code{\link{COMPARE}} function.
}
\examples{
# A type SPSS EFA to mimick the SPSS implementation
EFA_SPSS_5 <- EFA(IDS2_R, n_factors = 5, type = "SPSS")

# A type psych EFA to mimick the psych::fa() implementation
EFA_psych_5 <- EFA(IDS2_R, n_factors = 5, type = "psych")

# compare the two
COMPARE(EFA_SPSS_5$unrot_loadings, EFA_psych_5$unrot_loadings,
        x_labels = c("SPSS", "psych"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_group.R
\name{print.efa_group}
\alias{print.efa_group}
\alias{format.efa_group}
\title{Print and format a multigroup factor analysis}
\usage{
\method{print}{efa_group}(x, digits = 3, ...)

\method{format}{efa_group}(x, digits = 3, ...)
}
\arguments{
\item{x}{An object of class \code{efa_group} (output from \code{\link[=efa_group]{efa_group()}}).}

\item{digits}{Integer. The number of decimal places the reported values are rounded to.
Default is 3.}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a character
vector with the report lines.
}
\description{
\code{print()} turns an \code{\link[=efa_group]{efa_group()}} result into a sectioned report: a header recapping
the groups, the common number of factors, the estimator, the rotation, and the
alignment; a group-pair table of the matched Tucker congruences between the aligned
loadings; a per-pair summary of the cross-group loading differences (with the salient
and, when a bootstrap was run, the confidence-interval flags); and, when
\code{invariance = TRUE}, a group-pair by factor grid of the approximate-invariance verdicts.
\code{format()} assembles the same report and returns it as a character vector; \code{print()} is
\code{cat(format(x), sep = "\\n")}. The lines follow the active console theme, so they are
plain when colours are disabled (for example when captured into a file or stripped with
\code{\link[cli:ansi_strip]{cli::ansi_strip()}}). \code{print()} does not draw a plot; use \code{\link[=plot.efa_group]{plot.efa_group()}}.
}
\examples{
g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
mg <- efa_group(GRiPS_raw, groups = g, n_factors = 1)
mg

# format() returns the same lines as a character vector:
writeLines(format(mg))

}
\seealso{
Other factor analysis:
\code{\link[=efa_average]{efa_average()}},
\code{\link[=efa_fit]{efa_fit()}},
\code{\link[=efa_group]{efa_group()}},
\code{\link[=efa_mi]{efa_mi()}},
\code{\link[=plot.efa_group]{plot.efa_group()}}
}
\concept{factor analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{SCREE}
\alias{SCREE}
\title{Scree plot}
\usage{
SCREE(
  x,
  eigen_type = c("PCA", "SMC", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_factors = 1,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{eigen_type}{character. On what the eigenvalues should be found. Can be
either "PCA", "SMC", or "EFA", or some combination of them. If using "PCA",
the diagonal values of the correlation matrices are left to be 1. If using
"SMC", the diagonal of the
correlation matrices is replaced by the squared multiple correlations (SMCs)
of the indicators. If using "EFA", eigenvalues are found on the correlation
matrices with the final communalities of an exploratory factor analysis
solution (default is principal axis factoring extracting 1 factor) as
diagonal. Default is \code{c("PCA", "SMC", "EFA")}, i.e. all three; \code{"EFA"} is the
only one that fits a model.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator).
Default is "pearson".}

\item{n_factors}{numeric. Number of factors to extract if "EFA" is included in
\code{eigen_type}. Default is 1.}

\item{...}{Further arguments passed on to the \code{\link[=efa_fit]{efa_fit()}} fit. For example,
\code{estimator}, to change the estimator (PAF is default), or one of the estimation
tuning knobs (\code{type}, \code{init_comm}, \code{criterion}, \code{criterion_type}, \code{max_iter},
\code{abs_eigen}, \code{start_method}), which are repacked into an \code{\link[=estimate_control]{estimate_control()}}
object so that they tune the fit exactly as they always did.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_scree]{efa_scree()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{SCREE()} has been superseded by \code{\link[=efa_scree]{efa_scree()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_scree]{efa_scree()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.SCREE.R
\name{plot.SCREE}
\alias{plot.SCREE}
\title{Plot SCREE object}
\usage{
\method{plot}{SCREE}(x, ...)
}
\arguments{
\item{x}{a list of class SCREE An output from the \link{SCREE} function.}

\item{...}{not used.}
}
\description{
Plot method showing a summarized output of the \link{SCREE} function
}
\examples{
SCREE_base <- SCREE(test_models$baseline$cormat)
plot(SCREE_base)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{COMPARE}
\alias{COMPARE}
\title{Compare two vectors or matrices (communalities or loadings)}
\usage{
COMPARE(
  x,
  y,
  reorder = c("congruence", "names", "none"),
  corres = TRUE,
  thresh = 0.3,
  digits = 4,
  m_red = 0.001,
  range_red = 0.001,
  round_red = 3,
  print_diff = TRUE,
  na.rm = FALSE,
  x_labels = c("x", "y"),
  plot = TRUE,
  plot_red = 0.01
)
}
\arguments{
\item{x}{matrix, or vector. Loadings or communalities of a factor
analysis output.}

\item{y}{matrix, or vector. Loadings or communalities of another
factor analysis output to compare to x.}

\item{reorder}{character. Whether and how elements / columns should be
reordered. If "congruence" (default), the columns of \code{y} are matched to those of
\code{x} by a joint one-to-one assignment that maximizes the total Tucker's congruence
coefficient (a standard measure of similarity between two loading vectors) across
all columns at once, and each matched column's sign is flipped if needed. This way,
mismatched factor order or sign between two solutions does not distort the
comparison. It applies to matrices only, and warns when \code{x}
and \code{y} are vectors. If "names", the columns of a matrix -- or the elements of
a vector -- are put in alphabetical order of their names; the rows of a matrix
are assumed to be aligned already and are left untouched. If "none", no
reordering is done.}

\item{corres}{logical. Whether factor correspondences should be compared if a
matrix is entered. Default is TRUE.}

\item{thresh}{numeric. The threshold at or above which a loading is classified as substantial. Default is .3.}

\item{digits}{numeric. Number of decimals to print in the output. Default is 4.}

\item{m_red}{numeric. Number above which the mean and median should be printed
in red (i.e., if .001 is used, the mean will be in red if it is larger than
.001, otherwise it will be displayed in green.) Default is .001.}

\item{range_red}{numeric. Number above which the min and max should be printed
in red (i.e., if .001 is used, min and max will be in red if the max is larger
than .001, otherwise it will be displayed in green). Default is .001. Note that
the color of min also depends on max, that is min will be displayed in the
same color as max.}

\item{round_red}{numeric. The number of agreeing decimals below which the
report highlights the agreement in red (i.e., if 3 is used, the value is
shown in red when the compared numbers agree to fewer than 3 decimals,
otherwise in green). Default is 3.}

\item{print_diff}{logical. Whether the difference vector or matrix should be
printed or not. Default is TRUE.}

\item{na.rm}{logical. Whether NAs should be removed from the difference
summaries and factor-correspondence classifications. With \code{FALSE}, a missing
loading makes the correspondence counts undefined (\code{NA}). Default is FALSE.}

\item{x_labels}{character. A vector of length two containing identifying
labels for the two objects x and y that will be compared. These will be used
as labels on the x-axis of the plot, and to name the direction of the signed
elementwise differences in the printed report (see \code{\link[=print.efa_compare]{print.efa_compare()}}).
Default is "x" and "y".}

\item{plot}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}} Accepted and validated, but
without effect; retained for backwards compatibility. The difference plot is
drawn by \code{\link[=plot.efa_compare]{plot.efa_compare()}}. Default is TRUE.}

\item{plot_red}{numeric. Threshold above which to plot the absolute differences
in red. Default is .01.}
}
\value{
A list of class \code{c("efa_compare", "COMPARE")}, identical to the value
of \code{\link[=efa_compare]{efa_compare()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{COMPARE()} has been superseded by \code{\link[=efa_compare]{efa_compare()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_compare]{efa_compare()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FACTOR_SCORES.R
\name{FACTOR_SCORES}
\alias{FACTOR_SCORES}
\title{Estimate factor scores for an EFA model}
\usage{
FACTOR_SCORES(
  x,
  f,
  Phi = NULL,
  rho = NULL,
  method = c("Thurstone", "tenBerge", "Anderson", "Bartlett", "Harman", "components")
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data (needed to get
factor scores) or matrix with correlations.}

\item{f}{object of class \code{\link[=efa_fit]{efa_fit()}} or matrix.}

\item{Phi}{matrix. A matrix of factor intercorrelations. Only needs to be
specified if a factor loadings matrix is entered directly into \code{f}; for an
\code{\link[=efa_fit]{efa_fit()}} object the intercorrelations are taken from the object, and a
supplied \code{Phi} is ignored with a warning. Default is \code{NULL}, in which case the
intercorrelations of a directly supplied loading matrix are assumed to be zero.}

\item{rho}{matrix. Correlation matrix used to derive the scoring weights.
Defaults to \code{NULL}, in which case the matrix the EFA in \code{f} was fit on
(\code{f$orig_R}) is used, so the weights stay consistent with the loadings even for
a non-Pearson correlation (e.g. polychoric); for a directly supplied loading
matrix, \code{x} itself is used when it is a correlation matrix, otherwise the
Pearson correlation of \code{x}. Pass a matrix here to score against a different
correlation.}

\item{method}{character. The method used to calculate factor scores. One of
"Thurstone" (regression-based; default), "tenBerge", "Anderson", "Bartlett",
"Harman", or "components".}
}
\value{
A list of class FACTOR_SCORES containing the following:

\item{scores}{The factor scores (only if raw data are provided.)}
\item{weights}{The factor weights.}
\item{r.scores}{The correlations of the factor score estimates.}
\item{missing}{Whether the raw data contained missing values (only if raw data
are provided).}
\item{R2}{The squared factor-score determinacy for each factor: the squared
correlation between a factor and the score the requested \code{method} produces.
For \code{method = "Thurstone"} this equals the squared multiple correlation
between the factor and the observed variables; for every other method it is
specific to those scores and smaller. See \code{\link[=efa_scores]{efa_scores()}} for the underlying
score-quality diagnostics.}
\item{settings}{A list of the settings used.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{FACTOR_SCORES()} has been superseded by \code{\link[=efa_scores]{efa_scores()}}, which is the
recommended interface going forward. It remains available so existing code
keeps working. Note that \code{R2} is now the squared factor-score determinacy of
the \emph{requested} \code{method}: the squared correlation between a factor and the
scores that method produces. For \code{method = "Thurstone"} this is each factor's
squared multiple correlation with the observed variables (the value
\code{\link[psych:factor.scores]{psych::factor.scores()}} returns with \code{Grice = TRUE}); for every other method
it is smaller, because no estimator correlates more highly with the factor
than the regression estimator does. Earlier versions returned psych's default
\code{Grice = FALSE} validity coefficient, so the slot is not comparable across
versions.

A convenience wrapper around \code{\link[=efa_scores]{efa_scores()}} that returns factor scores and
weights in a compact list. Factor scores are calculated according to the
specified method if raw data are provided, and only factor weights if a
correlation matrix is provided.
}
\examples{
# Example with raw data with method "Bartlett"
EFA_raw <- efa_fit(DOSPERT_raw, n_factors = 10, estimator = "PAF",
                   rotation = "oblimin",
                   rotate_control = rotate_control(random_starts = 1))
fac_scores_raw <- FACTOR_SCORES(DOSPERT_raw, f = EFA_raw, method = "Bartlett")

# Same as above, but with raw data AND a correlation matrix
cor_pearson <- cor(DOSPERT_raw)
EFA_cor_pearson <- efa_fit(cor_pearson, n_factors = 10, N = nrow(DOSPERT_raw),
                           estimator = "PAF", rotation = "oblimin",
                           rotate_control = rotate_control(random_starts = 1))
fac_scores_cor_pearson <- FACTOR_SCORES(DOSPERT_raw, f = EFA_cor_pearson,
                                        rho = cor_pearson,
                                        method = "Bartlett")

# Scores between two alternatives above are identical
isTRUE(all.equal(fac_scores_raw$scores, fac_scores_cor_pearson$scores,
                 check.attributes = FALSE))

# Example with a correlation matrix only (does not return factor scores)
EFA_cor <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
                   estimator = "PAF", rotation = "oblimin")
fac_scores_cor <- FACTOR_SCORES(test_models$baseline$cormat, f = EFA_cor)

}
\seealso{
\code{\link[=efa_scores]{efa_scores()}} for the factor-score weights together with the full
set of score-quality diagnostics (determinacy, univocality, and Guttman
indeterminacy index) and a print/summary method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alignment.R
\name{.tucker_congruence}
\alias{.tucker_congruence}
\title{Tucker congruence between factors}
\usage{
.tucker_congruence(L1, L2)
}
\arguments{
\item{L1}{Numeric matrix.}

\item{L2}{Numeric matrix with the same dimensions as \code{L1}.}
}
\value{
A square matrix whose \verb{(i, j)} entry is the Tucker congruence
between column \code{i} of \code{L1} and column \code{j} of \code{L2}.
}
\description{
Compute the Tucker congruence matrix between the columns of two loading
matrices.
}
\references{
Lorenzo-Seva, U., and ten Berge, J. M. F. (2006). Tucker's congruence
coefficient as a meaningful index of factor similarity. \emph{Methodology}, 2,
57-64.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.hyperplane_count}
\alias{.hyperplane_count}
\title{Count near-zero loadings}
\usage{
.hyperplane_count(L, cutoff = 0.15)
}
\arguments{
\item{L}{Numeric loading matrix.}

\item{cutoff}{Numeric scalar. Loadings with `abs(L) < cutoff` are counted as
being in the hyperplane.}
}
\value{
A list with the total hyperplane count and counts by factor and item.
}
\description{
Hyperplane count is the number of loadings with absolute value smaller than a
user-specified cutoff.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{SL}
\alias{SL}
\title{Schmid-Leiman transformation}
\usage{
SL(
  x,
  Phi = NULL,
  type = c("EFAtools", "psych", "SPSS", "none"),
  method = c("PAF", "ML", "ULS", "MINRES"),
  g_name = "g",
  ...
)
}
\arguments{
\item{x}{object of class \code{\link[=efa_fit]{efa_fit()}}, class \code{\link[psych:fa]{psych::fa()}},
class \code{lavaan::lavaan()}, a matrix, or an \code{efa_loadings}/\code{loadings} object. If class \code{\link[=efa_fit]{efa_fit()}} or
class \code{\link[psych:fa]{psych::fa()}}, pattern coefficients and factor
intercorrelations are taken from this object. If class \code{lavaan::lavaan()},
it must be a second-order CFA solution. In this case first-order and second-order
factor loadings are taken from this object and the \code{g_name} argument has
to be specified.
x can also be a pattern matrix from an oblique factor solution (see \code{Phi}).}

\item{Phi}{matrix. A matrix of factor intercorrelations from an oblique factor
solution. Only needs to be specified if a pattern matrix is entered directly
into \code{x}.}

\item{type}{character. One of "EFAtools" (default), "psych", "SPSS", or "none". This is
used to control the procedure of the second-order factor analysis. In
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} it is set through the \code{type} of the \code{\link[=estimate_control]{estimate_control()}} object.}

\item{method}{character. The estimator for the second-order factor analysis; passed to
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}} as its \code{estimator} argument. One of \code{"PAF"}, \code{"ML"}, \code{"ULS"}, or
\code{"MINRES"}.}

\item{g_name}{character. The name of the general factor. This needs only be
specified if \code{x} is a \code{lavaan} second-order solution. Default is "g".}

\item{...}{Further arguments passed on to the second-order \code{\link[=efa_fit]{efa_fit()}}, including the
estimation tuning knobs (\code{init_comm}, \code{criterion}, \code{criterion_type}, \code{max_iter},
\code{abs_eigen}, \code{start_method}), which are repacked, together with \code{type}, into an
\code{\link[=estimate_control]{estimate_control()}} object so that they tune that fit exactly as they always did.
The estimator is selected with \code{method}.}
}
\value{
A list of class \code{c("efa_schmid_leiman", "SL")}, identical to the value
of \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{SL()} has been superseded by \code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_schmid_leiman]{efa_schmid_leiman()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residuals.EFA.R
\name{residuals.EFA}
\alias{residuals.EFA}
\title{Residuals function for EFA objects}
\usage{
\method{residuals}{EFA}(object, print = TRUE, digits = 3, ...)
}
\arguments{
\item{object}{a list of class EFA. Output from \code{\link{EFA}} or
\code{\link{EFA_POOLED}}.}

\item{print}{logical. Whether to print residuals. If FALSE, they are just returned.}

\item{digits}{numeric. The number of digits to round printed residuals.}

\item{...}{Further arguments.}
}
\description{
Residuals function for EFA objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{EKC}
\alias{EKC}
\title{Empirical Kaiser criterion}
\usage{
EKC(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  type = lifecycle::deprecated()
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Only needed if x is a correlation
matrix. Must be larger than the number of variables.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator). Default is \code{"pearson"}. Note that the EKC reference
values rest on the Marchenko-Pastur law for the eigenvalues of a sample
correlation matrix of independent variables, which assumes the sampling
behaviour of product-moment correlations; with \code{"poly"} / \code{"tetra"} (and, to a
lesser degree, the rank-based methods) the reference series is therefore an
approximation.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Accepted and ignored. It selected
between two ways to compute the reference values. The \code{"AM2019"} reference values
do not depend on the observed eigenvalues, so they do not apply the empirical
correction that defines the criterion, and they are no longer computed.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_ekc]{efa_ekc()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{EKC()} has been superseded by \code{\link[=efa_ekc]{efa_ekc()}}, which is the recommended interface
going forward. It remains available and unchanged so existing code keeps working.
}
\seealso{
\code{\link[=efa_ekc]{efa_ekc()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_retain.R
\name{efa_retain}
\alias{efa_retain}
\title{Various factor retention criteria}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200
}
\usage{
efa_retain(
  x,
  criteria = c("CD", "EKC", "HULL", "MAP", "NEST", "PARALLEL"),
  suitability = TRUE,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_factors_max = NA,
  N_pop = 10000,
  N_samples = 500,
  alpha = 0.3,
  ...,
  max_iter_CD = 50,
  n_fac_theor = NA,
  estimator = c("ML", "PAF", "ULS"),
  gof = c("CAF", "CFI", "RMSEA"),
  eigen_type_HULL = c("SMC", "PCA", "EFA"),
  eigen_type_other = c("SMC"),
  n_factors = 1,
  n_datasets = 1000,
  percent = 95,
  decision_rule = c("means", "percentile", "crawford"),
  ekc_type = lifecycle::deprecated(),
  n_datasets_nest = 1000,
  alpha_nest = 0.05,
  show_progress = FALSE,
  estimate_control = NULL
)
}
\arguments{
\item{x}{data.frame or matrix. Raw data, or a correlation matrix. If \code{"CD"} is
included as a criterion, x must be raw data.}

\item{criteria}{character. Which factor retention methods to run: one or
more of \code{"CD"}, \code{"EKC"}, \code{"HULL"}, \code{"KGC"}, \code{"MAP"}, \code{"NEST"}, \code{"PARALLEL"},
\code{"SCREE"}, and \code{"SMT"} (see details). The default runs a subset of commonly
used, well-performing methods, listed in the details.}

\item{suitability}{logical. Whether the data should be checked for suitability
for factor analysis using Bartlett's test of sphericity and the
Kaiser-Meyer-Olkin criterion (see details). Default is \code{TRUE}.}

\item{N}{numeric. The number of observations. Only needed if x is a
correlation matrix.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is \code{"pairwise.complete.obs"}.}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations (a two-step estimator).
\code{CD}, \code{PARALLEL}, \code{NEST}, \code{HULL}, and \code{SMT} do not support \code{"poly"} / \code{"tetra"}
and are skipped automatically if you request them together.
Default is \code{"pearson"}.}

\item{n_factors_max}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. The maximum number
of factors to test against.
Larger numbers will increase the duration the procedure takes, but test more
possible solutions. If left NA (default), the maximum number of factors for
which the model is still over-identified (df > 0) is used.}

\item{N_pop}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. Size of finite populations
of comparison data. Default is 10000.}

\item{N_samples}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. Number of samples drawn
from each population. Default is 500.}

\item{alpha}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. The alpha level used to test
the significance of the improvement added by an additional factor.
Default is .30.}

\item{...}{Further arguments passed to \code{\link[=efa_fit]{efa_fit()}} in
\code{\link[=efa_parallel]{efa_parallel()}}, \code{\link[=efa_kgc]{efa_kgc()}}, \code{\link[=efa_scree]{efa_scree()}}, \code{\link[=efa_nest]{efa_nest()}}, and (through its parallel
analysis and its own candidate fits) \code{\link[=efa_hull]{efa_hull()}}. An argument that \code{\link[=efa_fit]{efa_fit()}} does not
recognize causes an error. The estimation tuning knobs are not passed here; they live in
\code{estimate_control}. The standard-error arguments (\code{se}, \code{b_boot}, \code{ci}, \code{seed}) are not
accepted, because the criterion fits are internal steps whose standard errors are not
reported. Arguments listed after \code{...} must be given by their full name (R matches an
abbreviated name only against the arguments before \code{...}), so a tuning knob such as
\code{max_iter} cannot be mistaken for \code{max_iter_CD}.}

\item{max_iter_CD}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. The maximum number of
iterations to perform after which the iterative PAF procedure is halted.
Default is 50.}

\item{n_fac_theor}{numeric. Passed to \code{\link[=efa_hull]{efa_hull()}}. Theoretical number
of factors to retain. The Hull method uses one plus the larger of this
number and the number of factors suggested by \code{\link[=efa_parallel]{efa_parallel()}} as its
upper bound.}

\item{estimator}{character. Passed to \code{\link[=efa_fit]{efa_fit()}} in \code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}}, \code{\link[=efa_scree]{efa_scree()}}, \code{\link[=efa_parallel]{efa_parallel()}}, and \code{\link[=efa_nest]{efa_nest()}}. The
estimator to use. One of  \code{"PAF"}, \code{"ULS"}, or  \code{"ML"},
for principal axis factoring, unweighted least squares, and maximum
likelihood, respectively. The default here is \code{"ML"}. Some criteria default to
something else when called on their own (for example, \code{\link[=efa_hull]{efa_hull()}} defaults to
\code{"PAF"}), so results from \code{efa_retain()} can differ from calling that criterion
directly unless you set \code{estimator} to match. In \code{\link[=efa_kgc]{efa_kgc()}}, \code{\link[=efa_scree]{efa_scree()}}, and
\code{\link[=efa_parallel]{efa_parallel()}} it only takes effect when the respective \code{eigen_type} includes \code{"EFA"}.}

\item{gof}{character. Passed to \code{\link[=efa_hull]{efa_hull()}}. The goodness of fit index
to use. Either \code{"CAF"}, \code{"CFI"}, or \code{"RMSEA"}, or any
combination of them. With the \code{"PAF"} estimator, only
the CAF can be used as goodness of fit index. For details on the CAF, see
Lorenzo-Seva, Timmerman, and Kiers (2011).}

\item{eigen_type_HULL}{character. Passed to \code{\link[=efa_parallel]{efa_parallel()}} in
\code{\link[=efa_hull]{efa_hull()}}. What the eigenvalues in the parallel analysis are based on.
One of \code{"SMC"}, \code{"PCA"}, or \code{"EFA"} -- different ways of estimating how much
variance each indicator shares with the others before the eigenvalues are
computed. \code{"SMC"} (default) uses each indicator's squared multiple
correlation with the others (its diagonal value in the correlation matrix).
\code{"PCA"} leaves the diagonal at 1, so each indicator's total variance -- not
just the shared part -- feeds into the eigenvalues. \code{"EFA"} uses the
communalities from a fitted EFA solution instead.}

\item{eigen_type_other}{character. Passed to \code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_scree]{efa_scree()}}, and \code{\link[=efa_parallel]{efa_parallel()}}. The same as eigen_type_HULL,
but multiple inputs are possible here (any combination of \code{"PCA"}, \code{"SMC"},
and \code{"EFA"}). Default is \code{"SMC"}.}

\item{n_factors}{numeric. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}), \code{\link[=efa_kgc]{efa_kgc()}}, and \code{\link[=efa_scree]{efa_scree()}}. Number of
factors to extract if \code{"EFA"} is included in \code{eigen_type_HULL} or
\code{eigen_type_other}. Default is 1.}

\item{n_datasets}{numeric. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}). The number of datasets to simulate. Default is 1000.}

\item{percent}{numeric. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}). The percentile to take from the simulated eigenvalues.
Default is 95.}

\item{decision_rule}{character. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}). Which rule to use to determine the number of
factors to retain. Default is \code{"means"}, which uses the average simulated
eigenvalues. \code{"percentile"} uses the percentiles specified in \code{percent}.
\code{"crawford"} uses the 95th percentile for the first factor and the mean
afterwards (based on Crawford et al., 2010).}

\item{ekc_type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Accepted and ignored. It
used to select between two ways to compute the \code{\link[=efa_ekc]{efa_ekc()}} reference values.
The \code{"AM2019"} reference values do not depend on the observed eigenvalues.
They therefore skip the empirical correction that defines the criterion, so
they are no longer computed.}

\item{n_datasets_nest}{numeric. Passed to \code{\link[=efa_nest]{efa_nest()}}. The number of
datasets to simulate. Default is 1000.}

\item{alpha_nest}{numeric. Passed to \code{\link[=efa_nest]{efa_nest()}}. The alpha level to use.
The reference values are the eigenvalues at the (1 - alpha_nest)
percentile. Default is .05.}

\item{show_progress}{logical. Whether a progress bar should be shown in the
console. Default is FALSE.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
\code{\link[=efa_fit]{efa_fit()}} fits run by the criteria that fit a model (\code{\link[=efa_hull]{efa_hull()}}, \code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_scree]{efa_scree()}}, \code{\link[=efa_parallel]{efa_parallel()}}, \code{\link[=efa_nest]{efa_nest()}}, and \code{\link[=efa_smt]{efa_smt()}}). \code{NULL} (default) uses the
\code{\link[=efa_fit]{efa_fit()}} defaults. It only applies to criteria that fit a model, and only to the parts of
that fit each criterion actually runs. \code{\link[=efa_cd]{efa_cd()}}, \code{\link[=efa_ekc]{efa_ekc()}}, and \code{\link[=efa_map]{efa_map()}} fit no model.
\code{\link[=efa_kgc]{efa_kgc()}}, \code{\link[=efa_scree]{efa_scree()}}, and \code{\link[=efa_parallel]{efa_parallel()}} only fit one when their \code{eigen_type}
includes \code{"EFA"}. \code{\link[=efa_smt]{efa_smt()}} fits with maximum likelihood by definition, so only
\code{start_method} takes effect there. All fits are unrotated, so no rotation settings apply.}
}
\value{
A list of class \code{c("efa_retain", "N_FACTORS")}, the trailing class
keeping \code{inherits(x, "N_FACTORS")} working for code written against the
superseded name. It contains
\item{suitability}{A list with the results from \code{\link[=efa_bartlett]{efa_bartlett()}} and
\code{\link[=efa_kmo]{efa_kmo()}} (\code{bartlett} and \code{kmo}), or \code{NULL} if
\code{suitability = FALSE}.}
\item{outputs}{A named list with one \code{efa_retention} object per factor
retention criterion that was run (see, e.g., \code{\link[=efa_ekc]{efa_ekc()}}).}
\item{n_factors}{A named numeric vector with the suggested number of factors
per criterion and, where a criterion has several variants, per variant
(e.g. \code{EKC_BvA2017} or \code{PARALLEL_SMC}). Criteria without a numeric
suggestion (the scree plot) are not included. The printed summary's "most common"
value is based on each criterion's own most frequent (modal) suggestion among its
variants, not a plain tally of this vector, so counting entries here by hand can
give a different answer.}
\item{not_run}{A named character vector with the criteria that were skipped
or failed and the reason, or \code{NULL} if all requested criteria ran.}
\item{settings}{A list of the settings used. Its \code{criteria} element records the
requested criteria, in the order they were given, while \code{outputs} and \code{n_factors}
are in the order in which the criteria were run. \code{gof} records the requested Hull
goodness-of-fit indices, and \code{gof_used} records the ones the Hull method actually
computed (it reduces them to \code{"CAF"} for the PAF estimator). \code{gof_used} is \code{NA}
when HULL was not requested, was skipped, or failed.}
}
\description{
Choosing the number of factors to retain is one of the most important
decisions in an exploratory factor analysis (EFA). Many criteria exist to
help with this choice. This function runs several of them together, and can
also check whether the data are suitable for factor analysis.
}
\details{
By default, the entered data are checked for suitability for factor analysis
using the following methods (see the respective documentation for details):
\itemize{
\item Bartlett's test of sphericity (see \code{\link[=efa_bartlett]{efa_bartlett()}})
\item Kaiser-Meyer-Olkin criterion (see \code{\link[=efa_kmo]{efa_kmo()}})
}

The available factor retention criteria are the following (see the
respective documentation for details):
\itemize{
\item Comparison data (see \code{\link[=efa_cd]{efa_cd()}})
\item Empirical Kaiser criterion (see \code{\link[=efa_ekc]{efa_ekc()}})
\item Hull method (see \code{\link[=efa_hull]{efa_hull()}})
\item Kaiser-Guttman criterion (see \code{\link[=efa_kgc]{efa_kgc()}})
\item Velicer's minimum average partial, MAP (see \code{\link[=efa_map]{efa_map()}})
\item Next Eigenvalue Sufficiency Test, NEST (see \code{\link[=efa_nest]{efa_nest()}})
\item Parallel analysis (see \code{\link[=efa_parallel]{efa_parallel()}})
\item Scree plot (see \code{\link[=efa_scree]{efa_scree()}})
\item Sequential chi-square model tests, RMSEA lower bound, and AIC
(see \code{\link[=efa_smt]{efa_smt()}})
}

The default \code{criteria} are comparison data, the empirical Kaiser criterion, the
Hull method, MAP, NEST, and parallel analysis. No single criterion is the most
accurate in all conditions. \code{efa_retain()} therefore runs several criteria
together, and the printed summary gives the range of their suggestions and the
most common one. Auerswald and Moshagen (2019) compare the criteria and give
guidance on the selection.

The comparison data, parallel analysis, and NEST criteria compare the data
against simulated reference data, so their suggested numbers of factors vary
slightly from run to run. The Hull method also varies, because it calls
\code{\link[=efa_parallel]{efa_parallel()}} to set its upper bound. Call \code{\link[base:set.seed]{base::set.seed()}} before
\code{efa_retain()} to make the results reproducible.
}
\examples{
\donttest{
# Default criteria, with correlation matrix and estimator "ML" (where needed)
# This will throw a warning for CD, as no raw data were specified
# The simulation-based criteria are seeded to make the run reproducible
set.seed(42)
nfac_all <- efa_retain(test_models$baseline$cormat, N = 500, estimator = "ML",
                       n_datasets = 100, n_datasets_nest = 100)

# The same as above, but without "CD"
nfac_wo_CD <- efa_retain(test_models$baseline$cormat, criteria = c("EKC",
                         "HULL", "PARALLEL", "NEST"), N = 500,
                         estimator = "ML", n_datasets = 100,
                         n_datasets_nest = 100)

# Use PAF instead of ML (this will take longer). PAF only supports "CAF" as
# gof for the Hull method, so set it explicitly to avoid the automatic message.
nfac_PAF <- efa_retain(test_models$baseline$cormat, criteria = c("EKC",
                       "HULL", "PARALLEL", "NEST"), N = 500,
                       estimator = "PAF", gof = "CAF", n_datasets = 100,
                       n_datasets_nest = 100)

# Back to the default ML estimator (unlike above), with only "PCA" type eigenvalues
nfac_PCA <- efa_retain(test_models$baseline$cormat, criteria = c("EKC",
                       "HULL", "PARALLEL", "NEST"), N = 500,
                       estimator = "ML", eigen_type_other = "PCA",
                       n_datasets = 100, n_datasets_nest = 100)

# Use raw data, such that CD can also be performed
nfac_raw <- efa_retain(GRiPS_raw, estimator = "ML", N_pop = 500,
                        N_samples = 20, n_datasets = 100,
                        n_datasets_nest = 100)
}
}
\seealso{
\code{\link[=efa_screen]{efa_screen()}} for data screening before retention, and \code{\link[=efa_fit]{efa_fit()}} to extract
the chosen number of factors.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_bartlett.R
\name{print.efa_bartlett}
\alias{print.efa_bartlett}
\alias{format.efa_bartlett}
\title{Print and format an efa_bartlett object}
\usage{
\method{print}{efa_bartlett}(x, ...)

\method{format}{efa_bartlett}(x, ...)
}
\arguments{
\item{x}{An object of class \code{efa_bartlett} (output from \code{\link[=efa_bartlett]{efa_bartlett()}}).}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} reports the outcome of \code{\link[=efa_bartlett]{efa_bartlett()}}'s test of sphericity: a verdict
on whether the test was significant (and what that implies for the suitability
of the data for factor analysis), followed by the chi-square statistic, its
degrees of freedom, and the p-value. \code{format()} assembles the same report and
returns it as a character vector; \code{print()} is \code{cat(format(x), sep = "\\n")}.
The lines follow the active console theme, so they are plain when colours are
disabled (for example when captured into a file or stripped with
\code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
bart <- efa_bartlett(test_models$baseline$cormat, N = 500)
bart

# format() returns the same lines as a character vector:
writeLines(format(bart))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.factor_corres}
\alias{.factor_corres}
\title{Compute number of non-matching indicator-to-factor correspondences}
\usage{
.factor_corres(x, y, thresh = 0.3)
}
\arguments{
\item{x}{numeric matrix. A matrix of pattern coefficients.}

\item{y}{numeric matrix. A second matrix of coefficients.}

\item{thresh}{numeric. The threshold to classify a pattern coefficient as substantial.}
}
\description{
Compute number of non-matching indicator-to-factor correspondences
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_map.R
\name{efa_map}
\alias{efa_map}
\title{Velicer's minimum average partial (MAP) criterion}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. \emph{Psychological Methods, 24}(4), 468--491.
https://doi.org/10.1037/met0000200

Velicer, W. F. (1976). Determining the number of components from the matrix of partial correlations.
\emph{Psychometrika, 41}, 321--327.

Velicer, W. F., Eaton, C. A., & Fava, J. L. (2000). Construct explication through factor or component analysis: A review and evaluation of alternative procedures for determining the number of factors or components. In Goffin, R. D. & Helmes, E. (Eds.), \emph{Problems and Solutions in
Human Assessment: Honoring Douglas N. Jackson at Seventy} (pp. 41--71). Boston: Kluwer.

Zwick, W. R., & Velicer, W. F. (1986). Comparison of five rules for
determining the number of components to retain. \emph{Psychological Bulletin, 99},
432--442. https://doi.org/10.1037/0033-2909.99.3.432
}
\usage{
efa_map(
  x,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")
)
}
\arguments{
\item{x}{A numeric \code{matrix} or \code{data.frame}. Can be either (a) a correlation matrix, or
(b) raw data (rows = observations, columns = variables) from which correlations are computed.}

\item{use}{Character string specifying the treatment of missing values when computing correlations.
Passed to \code{\link[stats:cor]{stats::cor()}}. Defaults to \code{"pairwise.complete.obs"}.}

\item{cor_method}{Character string specifying the correlation coefficient to be computed if raw
data are supplied. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"} (passed to
\code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} / \code{"tetra"} for polychoric / tetrachoric correlations
of ordinal / binary data (a two-step estimator). Defaults to \code{"pearson"}.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} for the
print method). MAP has no plot; \code{\link[=plot.efa_retention]{plot.efa_retention()}} returns \code{NULL} with a
message for it. Its main elements are:
\itemize{
\item \code{n_factors}: A named numeric vector (\code{"TR2"}, \code{"TR4"}) with the index
\eqn{m} that minimizes the original (TR2) and revised (TR4) MAP criterion.
\item \code{results}: A list with one record per criterion, each holding the
criterion values over \eqn{m} and, in \code{m_last}, the largest \eqn{m} at which
the criterion could be evaluated (see details).
\item \code{settings}: A list containing \code{use} and \code{cor_method}.
}
}
\description{
Computes Velicer's Minimum Average Partial (MAP) criterion for determining the number of
factors/components to retain. The function implements the original MAP criterion
(Velicer, 1976), expressed via the \eqn{\mathrm{TR2}} representation, and the revised
\eqn{\mathrm{TR4}} variant proposed by Velicer, Eaton, and Fava (2000).
}
\details{
MAP partials successive principal components out of the correlation matrix and,
after removing \eqn{m} components, summarizes the off-diagonal partial
correlations \eqn{r^*_{ij}} that remain in the \eqn{m}-th partial correlation
matrix \eqn{M} (which has a unit diagonal); the suggested number of factors is
the \eqn{m} that minimizes the criterion. Two criteria are returned, each
rescaling the trace of a matrix power of \eqn{M} by the number of off-diagonal
cells \eqn{p(p-1)}:
\itemize{
\item \strong{TR2 (original MAP; Velicer, 1976):} the average squared off-diagonal
partial correlation,
\deqn{\mathrm{TR2}_m = \frac{\mathrm{tr}(M^2) - p}{p(p-1)} = \frac{\sum_{i \neq j} (r^*_{ij})^2}{p(p-1)},}
where subtracting \eqn{p} removes the \eqn{p} unit diagonal entries.
\item \strong{TR4 (revised MAP; Velicer, Eaton, & Fava, 2000):} the analogous
fourth-power summary, formed from the trace of the fourth matrix power,
\deqn{\mathrm{TR4}_m = \frac{\mathrm{tr}(M^4) - p}{p(p-1)}.}
Moving from the squared to the fourth power downweights the small partial
correlations relative to the large ones, which can sharpen the minimum.
Unlike TR2, \eqn{\mathrm{tr}(M^4)} is \emph{not} the sum of the fourth powers of the
individual partial correlations; the matrix power is intended and is what
Velicer, Eaton, and Fava (2000) describe.
}

Both criteria are returned for every call and they can suggest different numbers
of factors on the same correlation matrix. Both are in use in the literature and
neither is treated as the default here, so be sure to state which of the two you
report, as you would for any other analysis choice.

MAP is most dependable when the components are well determined, that is with many
indicators per factor and substantial loadings. It has a well-documented tendency
to under-extract, particularly with few indicators per factor or weak loadings
(Zwick & Velicer, 1986; Auerswald & Moshagen, 2019), so it is best read as a lower
bound and paired with a criterion that errs in the other direction, such as the
Kaiser-Guttman criterion (\code{\link[=efa_kgc]{efa_kgc()}}).

The criterion is evaluated over \eqn{m = 0, \ldots, p - 1}. Each step standardizes
the partial covariance matrix by its residual standard deviations, which requires
every residual variance to stay positive. Partialling out all but one component
leaves a rank-one residual, so the final point \eqn{m = p - 1} is undefined for
most correlation matrices and is routinely returned as \code{NA}. A residual variance
can also reach zero earlier, most often on a near-singular matrix; the search then
stops there, the criterion values that could be computed are kept, the remaining
values stay \code{NA}, and a warning (class \code{efa_map_truncated}) reports how far the
grid was searched. In that case the suggested \eqn{m} is the minimum over the
evaluated range only, so it should be read together with the returned series.

A non-positive-definite input correlation matrix (e.g. from sampling error) is
smoothed with \code{\link[psych:cor.smooth]{psych::cor.smooth()}}.
}
\examples{
## Example with raw data
res <- efa_map(GRiPS_raw)
res

## Example with a correlation matrix
res2 <- efa_map(test_models$baseline$cormat)
res2

}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.rotation_se_jacobian}
\alias{.rotation_se_jacobian}
\title{Rotation Jacobians for analytic rotation standard errors}
\usage{
.rotation_se_jacobian(
  A,
  T_init,
  method,
  param,
  normalize,
  oblique,
  eps,
  general_col = 0L
)
}
\arguments{
\item{A}{Numeric matrix. The unrotated loading matrix at the solution.}

\item{T_init}{Numeric matrix. The converged transformation that warm-starts each re-rotation.}

\item{method}{Character scalar. The criterion family: one of \code{"cf"}, \code{"oblimin"}, \code{"geomin"},
\code{"bentler"}, \code{"bifactor"}.}

\item{param}{Numeric scalar. The criterion's tuning argument (\code{kappa} for \code{"cf"}, \code{gam} for
\code{"oblimin"}, \code{delta} for \code{"geomin"}); ignored for \code{"bentler"} and \code{"bifactor"}.}

\item{normalize}{Logical scalar. Apply Kaiser normalization before rotation and reverse it after.}

\item{oblique}{Logical scalar. Use the oblique (column-normalized) manifold; otherwise orthogonal.}

\item{eps}{Numeric scalar. The forward-difference step on the loadings.}

\item{general_col}{Integer scalar. For \code{method = "bifactor"}, the zero-based column holding the
general factor in the (factor-reordered) reported solution; ignored by the other criteria.}
}
\value{
A named list with the Jacobian \code{J_L} (\verb{pk x pk}), the re-rotated \code{base_loadings}, a
validity flag, and -- when oblique -- the Jacobian \code{J_Phi} (\verb{k^2 x pk}) and \code{base_Phi}.
}
\description{
Forward-difference the warm-started re-rotation map \verb{A -> (rotated loadings, Phi)} over the
unrotated loadings \code{A} to obtain the rotation Jacobians used by the analytic standard errors for
rotated loadings (\code{se = "information"} in \code{\link[=efa_fit]{efa_fit()}}). The full \code{nrow(A) * ncol(A)} finite-
difference loop runs in compiled code, re-solving the rotation from the converged transformation
\code{T_init} at each perturbation; the caller forms \verb{J V J'} in R.
}
\references{
Jennrich, R. I. (1973). Standard errors for obliquely rotated factor loadings.
\emph{Psychometrika}, 38, 593-604.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper.R
\name{.consensus_loss}
\alias{.consensus_loss}
\title{Mean squared discrepancy to a consensus target}
\usage{
.consensus_loss(aligned_loadings, target)
}
\arguments{
\item{aligned_loadings}{List of aligned loading matrices.}

\item{target}{Consensus target matrix.}
}
\value{
Mean sum of squared deviations from the target across matrices.
}
\description{
Mean squared discrepancy to a consensus target
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_parallel.R
\name{efa_parallel}
\alias{efa_parallel}
\title{Parallel analysis}
\source{
Braeken, J., & van Assen, M. A. (2017). An empirical Kaiser criterion.
Psychological Methods, 22, 450--466. https://doi.org/10.1037/met0000074

Crawford, A. V., Green, S. B., Levy, R., Lo, W. J., Scott, L.,
Svetina, D., & Thompson, M. S. (2010). Evaluation of parallel analysis methods
for determining the number of factors. Educational and Psychological
Measurement, 70(6), 885-901.

Glorfeld, L. W. (1995). An improvement on Horn's parallel analysis
methodology for selecting the correct number of factors to retain. Educational
and Psychological Measurement, 55(3), 377-393.

Horn, J. L. (1965). A rationale and test for the number of factors in
factor analysis. Psychometrika, 30(2), 179--185. https://doi.org/10.1007/BF02289447
}
\usage{
efa_parallel(
  x = NULL,
  N = NA,
  n_vars = NA,
  n_datasets = 1000,
  percent = 95,
  eigen_type = c("PCA", "SMC", "EFA"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  decision_rule = c("means", "percentile", "crawford"),
  n_factors = 1,
  estimate_control = NULL,
  ...
)
}
\arguments{
\item{x}{matrix or data.frame. The real data to compare the simulated eigenvalues
against. Must not contain variables of classes other than numeric. Can be a
correlation matrix or raw data.}

\item{N}{numeric. The number of cases / observations to simulate. Only has to
be specified if \code{x} is either a correlation matrix or \code{NULL}. If
x contains raw data, \code{N} is found from the dimensions of \code{x}. Must be larger
than the number of variables.}

\item{n_vars}{numeric. The number of variables / indicators to simulate.
Only has to be specified if \code{x} is left as \code{NULL} as otherwise the
dimensions are taken from \code{x}.}

\item{n_datasets}{numeric. The number of datasets to simulate. Must be at
least 1. Default is 1000.}

\item{percent}{numeric. The percentile to take from the simulated eigenvalues.
Default is 95.}

\item{eigen_type}{character. On what the eigenvalues should be found. Can be
either "SMC", "PCA", or "EFA". If using "SMC", the diagonal of the correlation
matrix is replaced by the squared multiple correlations (SMCs) of the
indicators. If using "PCA", the diagonal values of the correlation matrices
are left to be 1. If using "EFA", eigenvalues are found on the correlation
matrices with the final communalities of an EFA solution as diagonal. Default
is \code{c("PCA", "SMC", "EFA")}, i.e. all three, which costs roughly six times a
single non-EFA type: \code{"EFA"} fits an EFA to every simulated dataset and
dominates that total. Pass a single type if the run is time-critical.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{PARALLEL} compares the data against simulated continuous reference data.
Default is "pearson".}

\item{decision_rule}{character. Which rule to use to determine the number of
factors to retain. Default is \code{"means"}, which will use the average
simulated eigenvalues. \code{"percentile"}, uses the percentiles specified
in percent. \code{"crawford"} uses the 95th percentile for the first factor
and the mean afterwards (based on Crawford et al, 2010). All three rules retain
the factors up to the first observed eigenvalue that fails to exceed its
reference value; an eigenvalue further down the series that rises above its own
reference again therefore adds no factor. Because the average simulated
eigenvalue is a lower reference than the percentile, \code{"means"} tends to retain
more factors than the more conservative \code{"percentile"} rule (Glorfeld, 1995).}

\item{n_factors}{numeric. Number of factors to extract if "EFA" is included in
\code{eigen_type}. Default is 1.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
\code{\link[=efa_fit]{efa_fit()}} fits (of both the real and the simulated data) when \code{"EFA"} is included in
\code{eigen_type}. \code{NULL} (default) uses the \code{\link[=efa_fit]{efa_fit()}} defaults. The fits are unrotated, so no
rotation settings apply.}

\item{...}{Additional arguments passed to \code{\link[=efa_fit]{efa_fit()}}. For example,
\code{estimator}, to change the estimator (default is "PAF"). PAF is more
robust, but it will take longer compared to the other estimators
available ("ML" and "ULS"). The estimation tuning knobs are not passed here; they live in
\code{estimate_control}, and the standard-error arguments (\code{se}, \code{b_boot}, \code{ci}, \code{seed}) are
not accepted because the fits are internal steps that keep only their eigenvalues.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} and
\code{\link[=plot.efa_retention]{plot.efa_retention()}} for the print and plot methods). Its main fields are:
\item{n_factors}{A named numeric vector with the suggested number of factors for
each requested eigenvalue type (\code{"PCA"}, \code{"SMC"}, and/or \code{"EFA"}). These are
\code{NA} when no real data are supplied (i.e. only \code{N} and \code{n_vars} are given). When
every observed eigenvalue exceeds its reference value (no crossing is found), all
\code{n_vars} components are retained and a warning is issued.}
\item{results}{A list with one record per eigenvalue type, each holding the
observed eigenvalues (when real data were supplied) and the simulated reference
values (means and percentiles) used for printing and plotting.}
\item{settings}{A list of the settings used.}
}
\description{
Various methods for performing parallel analysis. This function uses
\link[future.apply:future_lapply]{future_lapply()} for which a parallel processing plan can
be selected. To do so, register a plan with \code{\link[future:plan]{future::plan()}}, for example
\code{future::plan(future::multisession, workers = 2)}; see examples.
}
\details{
Parallel analysis (Horn, 1965) compares the eigenvalues obtained from
the sample
correlation matrix against those of null model correlation matrices (i.e.,
with uncorrelated variables) of the same sample size. This way, it accounts
for the variation in eigenvalues introduced by sampling error and thus
eliminates the main problem inherent in the Kaiser-Guttman criterion
(\code{\link[=efa_kgc]{efa_kgc()}}).

Parallel analysis is often argued to be one of the most accurate factor
retention criteria. However, for highly correlated
factor structures it has been shown to underestimate the correct number of
factors. The reason for this is that a null model (uncorrelated variables)
is used as reference. However, when factors are highly correlated, the first
eigenvalue will be much larger compared to the following ones, as
later eigenvalues are conditional on the earlier ones in the sequence and thus
the shared variance is already accounted in the first eigenvalue (e.g.,
Braeken & van Assen, 2017).

The reference eigenvalues are obtained from simulated data, so the suggested number
of factors varies slightly from run to run. Call \code{\link[base:set.seed]{base::set.seed()}} beforehand to make a
run reproducible; the result is then also independent of the parallel plan set via
\code{\link[future:plan]{future::plan()}}, so it can be reproduced on a machine with a different number of
cores. For \code{"PCA"} and \code{"SMC"} the simulation is drawn in independently seeded blocks;
a block that fails -- which happens when a simulated correlation matrix is singular, so
that no eigenvalues can be taken from it -- is redrawn on its own, leaving the blocks
that succeeded with the draws they already made. The \code{"EFA"} series instead redraws the
single dataset that could not be fitted; if that dataset still cannot be fitted, the
call stops with an error.

When both \code{"PCA"} and \code{"SMC"} are requested, the two are read off the \emph{same} simulated
datasets rather than from two independent simulations: they differ only in the diagonal
substituted into the simulated correlation matrix, so one set of draws serves both and
the two reference series are paired dataset by dataset. A draw that cannot be used for
the SMC series -- a simulated matrix with no inverse, and hence no squared multiple
correlations -- is discarded for the \code{"PCA"} series as well, so that the pairing stays
exact. \code{"EFA"} fits a model to each simulated dataset and draws its own.

The \code{efa_parallel} function can also be called together with other factor
retention criteria in the \code{\link[=efa_retain]{efa_retain()}} function.
}
\examples{
\donttest{
# example without real data
pa_unreal <- efa_parallel(N = 500, n_vars = 10, n_datasets = 100)

# example with correlation matrix with all eigen_types and PAF estimation
pa_paf <- efa_parallel(test_models$case_11b$cormat, N = 500, n_datasets = 100)

# example with correlation matrix with all eigen_types and ML estimation
# this will be faster than the above with PAF)
pa_ml <- efa_parallel(test_models$case_11b$cormat, N = 500, estimator = "ML",
                      n_datasets = 100)
}

\dontrun{
# for parallel computation. future::plan() returns the plan it replaces, so
# on.exit() puts the session back as it was -- also if the call fails.
pa_faster <- local({
  old_plan <- future::plan(future::multisession, workers = 2)
  on.exit(future::plan(old_plan), add = TRUE)
  efa_parallel(test_models$case_11b$cormat, N = 500)
})
}
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}},
\code{\link[=efa_smt]{efa_smt()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.LOADINGS.R
\name{print.LOADINGS}
\alias{print.LOADINGS}
\alias{format.LOADINGS}
\title{Print LOADINGS object}
\usage{
\method{print}{LOADINGS}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  h2 = NULL,
  color = TRUE,
  name_style = c("truncate", "abbreviate", "full"),
  max_factor_name_length = NULL,
  max_factors_per_block = NULL,
  sort_loadings = c("none", "primary", "clustered"),
  legend = FALSE,
  ...
)

\method{format}{LOADINGS}(
  x,
  cutoff = 0.3,
  digits = 3,
  max_name_length = 10,
  h2 = NULL,
  color = TRUE,
  name_style = c("truncate", "abbreviate", "full"),
  max_factor_name_length = NULL,
  max_factors_per_block = NULL,
  sort_loadings = c("none", "primary", "clustered"),
  legend = FALSE,
  ...
)
}
\arguments{
\item{x}{class LOADINGS matrix.}

\item{cutoff}{numeric. The number above which to print loadings in bold
default is .3.}

\item{digits}{numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
to round the loadings to (default is 3).}

\item{max_name_length}{numeric. The maximum length of the variable names to
display. Everything beyond this will be cut from the right unless
\code{name_style = "abbreviate"} or \code{name_style = "full"} is used.}

\item{h2}{numeric. Vector of communalities to print. If named and \code{x}
has row names, names are used to align communalities to rows.}

\item{color}{logical. Whether to apply console styling using \pkg{crayon}.
Default is \code{TRUE}.}

\item{name_style}{character. How to shorten variable names longer than
\code{max_name_length}. \code{"truncate"} cuts names from the right,
\code{"abbreviate"} uses \code{\link[base:abbreviate]{abbreviate}}, and
\code{"full"} prints full names.}

\item{max_factor_name_length}{numeric or \code{NULL}. Optional maximum length
of factor names. If \code{NULL}, factor names are not shortened.}

\item{max_factors_per_block}{numeric or \code{NULL}. Maximum number of factor
columns to print per block. If \code{NULL}, the number is chosen from the
console width.}

\item{sort_loadings}{character. Optional row sorting. \code{"none"} preserves
the input order, \code{"primary"} groups rows by the factor with the largest
absolute loading, and \code{"clustered"} additionally sorts within each
factor by the size of the primary loading.}

\item{legend}{logical. Whether to append a short explanation of the styling.
Default is \code{FALSE} for standalone loading matrices.}

\item{...}{additional arguments passed to print or format}
}
\description{
Print LOADINGS object
}
\details{
The method prints a loading matrix in a compact, console-oriented table.
Loadings with absolute value greater than or equal to `cutoff` are emphasized,
smaller loadings are de-emphasized, and Heywood-relevant communality/
uniqueness values are marked when `h2` is supplied. Long variable names can
be truncated, abbreviated, or printed in full. If the matrix has many factor
columns, the table is split into column blocks so that the output remains
readable in narrower consoles.

If `h2` is named and `x` has row names, `h2` is matched to the row names of
`x` before any optional row sorting is applied. If `x` has no row names, a
named `h2` vector is used in the supplied order.
}
\examples{
EFAtools_PAF <- EFA(test_models$baseline$cormat, n_factors = 3, N = 500,
                    type = "EFAtools", method = "PAF", rotation = "promax")
EFAtools_PAF

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-package.R
\docType{package}
\name{EFAtools-package}
\alias{EFAtools}
\alias{EFAtools-package}
\title{EFAtools: Fast and Flexible Implementations of Exploratory Factor Analysis Tools}
\description{
Provides a complete workflow for exploratory factor analysis (EFA). It covers data screening and factorability checks, a suite of factor retention criteria for choosing the number of factors, and factor extraction by principal axis factoring, maximum likelihood, unweighted least squares, or diagonally weighted least squares from Pearson, Spearman, Kendall, polychoric, tetrachoric, or two-stage full-information maximum likelihood correlations. A built-in rotation engine offers a range of orthogonal and oblique rotations, and standard errors for loadings and related quantities can be obtained by analytic, robust, or bootstrap methods. Further tools support model averaging across analytic choices, multigroup EFA with factor congruence, EFA on multiply imputed data, Schmid-Leiman transformation, reliability coefficients (including McDonald's omegas), factor score estimation, data simulation, and power analysis. Computationally intensive procedures are implemented in 'C++' for speed.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/mdsteiner/EFAtools}
  \item \url{https://mdsteiner.github.io/EFAtools/}
  \item Report bugs at \url{https://github.com/mdsteiner/EFAtools/issues}
}

}
\author{
\strong{Maintainer}: Markus Steiner \email{markus.d.steiner@gmail.com} (\href{https://orcid.org/0000-0002-8126-0757}{ORCID})

Authors:
\itemize{
  \item Markus Steiner \email{markus.d.steiner@gmail.com} (\href{https://orcid.org/0000-0002-8126-0757}{ORCID})
  \item Silvia Steiner \email{silvia.steiner.grieder@gmail.com} (\href{https://orcid.org/0000-0002-0118-7722}{ORCID})
}

Other contributors:
\itemize{
  \item William Revelle [contributor]
  \item Max Auerswald [contributor]
  \item Morten Moshagen [contributor]
  \item John Ruscio [contributor]
  \item Brendan Roche [contributor]
  \item Urbano Lorenzo-Seva [contributor]
  \item David Navarro-Gonzalez [contributor]
  \item Johan Braeken [contributor]
  \item Andreas Soteriades [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.paf_iter}
\alias{.paf_iter}
\title{Perform the iterative PAF procedure}
\usage{
.paf_iter(h2, criterion, R, n_fac, abs_eig, crit_type, max_iter)
}
\arguments{
\item{h2}{numeric. The initial communality estimates.}

\item{criterion}{double. The convergence criterion to use.}

\item{R}{matrix. The correlation matrix with the initial communality estimates in the diagonal.}

\item{n_fac}{numeric. The number of factors to extract.}

\item{abs_eig}{logical. Whether absolute eigenvalues should be used to compute the loadings.}

\item{crit_type}{numeric. Whether maximum absolute differences (crit_type = 1), or sum of differences (crit_type = 2) should be used}

\item{max_iter}{numeric. The number of iterations after which to end the procedure if no convergence has been reached by then.}
}
\description{
Function called from within PAF so usually no call to this is needed by the user.
Provides a C++ implementation of the PAF procedure
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_retain.R
\name{plot.efa_retain}
\alias{plot.efa_retain}
\title{Plot method for efa_retain objects}
\usage{
\method{plot}{efa_retain}(x, ...)
}
\arguments{
\item{x}{an object of class efa_retain, returned by \code{\link[=efa_retain]{efa_retain()}}.}

\item{...}{not used.}
}
\value{
A named list of \link[ggplot2:ggplot]{ggplot2::ggplot} objects, one per criterion with a
plottable result, or invisibly \code{NULL} if there is none.
}
\description{
Plots every factor-retention criterion in the \code{\link[=efa_retain]{efa_retain()}} result that has
a plottable outcome (see \code{\link[=plot.efa_retention]{plot.efa_retention()}}). Criteria without a plot
(e.g. \code{\link[=efa_map]{efa_map()}} or \code{\link[=efa_smt]{efa_smt()}}) are skipped.
}
\examples{
\donttest{
nf <- efa_retain(test_models$baseline$cormat, criteria = c("EKC", "SMT"),
                 N = 500)
plot(nf)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IDS2_R_doc.R
\docType{data}
\name{IDS2_R}
\alias{IDS2_R}
\title{Intelligence subtests from the Intelligence and Development Scales--2}
\format{
A 14 x 14 matrix of bivariate correlations
\describe{
\item{GS}{(numeric) - Geometric shapes.}
\item{PL}{(numeric) - Plates.}
\item{TC}{(numeric) - Two characteristics.}
\item{CB}{(numeric) - Crossing out boxes.}
\item{NL}{(numeric) - Numbers / letters.}
\item{NLM}{(numeric) - Numbers / letter mixed.}
\item{GF}{(numeric) - Geometric figures.}
\item{RGF}{(numeric) - Rotated geometric figures.}
\item{CM}{(numeric) - Completing matrices.}
\item{EP}{(numeric) - Excluding pictures.}
\item{CA}{(numeric) - Categories.}
\item{OP}{(numeric) - Opposites.}
\item{RS}{(numeric) - Retelling a story.}
\item{DP}{(numeric) - Describing pictures.}
}
}
\source{
Grieder, S., & Grob, A. (2019). Exploratory factor analyses of the intelligence and development scales--2: Implications for theory and practice. Assessment. Advance online publication. doi:10.1177/1073191119845051

Grob, A., & Hagmann-von Arx, P. (2018). Intelligence and Development Scales--2 (IDS-2). Intelligenz- und Entwicklungsskalen für Kinder und Jugendliche.
[Intelligence and Development Scales for Children and Adolescents.]. Bern, Switzerland: Hogrefe.
}
\usage{
IDS2_R
}
\description{
A matrix containing the bivariate correlations of the 14 intelligence subtests from the Intelligence and Development Scales--2 (IDS-2; Grob & Hagmann-von Arx, 2018), an intelligence and development test battery for children and adolescents aged 5 to 20 years, for the standardization and validation sample (N = 1,991). Details can be found in Grieder & Grob (2019).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DOSPERT_raw_doc.R
\docType{data}
\name{DOSPERT_raw}
\alias{DOSPERT_raw}
\title{DOSPERT_raw}
\format{
A data.frame with 3,123 rows (participants) and 30 columns, named by a
domain prefix and item number, with six items in each of five risk domains:
\describe{
\item{ethR_1 to ethR_6}{(numeric) - Ethical-domain risk-taking items.}
\item{finR_1 to finR_6}{(numeric) - Financial-domain risk-taking items.}
\item{heaR_1 to heaR_6}{(numeric) - Health/safety-domain risk-taking items.}
\item{recR_1 to recR_6}{(numeric) - Recreational-domain risk-taking items.}
\item{socR_1 to socR_6}{(numeric) - Social-domain risk-taking items.}
}
}
\source{
Blais, A.-R., & Weber, E. U. (2006). A domain-specific risk-taking (DOSPERT) scale for adult populations. Judgment and Decision Making, 1(1), 33–47. doi: 10.1017/S1930297500000334

Frey, R., Duncan, S. M., & Weber, E. U. (2023). Towards a typology of risk preference: Four risk profiles describe two-thirds of individuals in a large sample of the U.S. population. Journal of Risk and Uncertainty, 66(1), 1–17. doi:10.1007/s11166-022-09398-5
}
\usage{
DOSPERT_raw
}
\description{
A data.frame containing responses to the risk subscale of the Domain Specific
Risk Taking Scale (DOSPERT; Weber, Blais, & Betz, 2002) based on the publicly
available dataset (at \url{https://osf.io/pjt57/}) by Frey, Duncan, and Weber (2023).
The items measure risk-taking propensity on five different domains: social,
recreational, financial, health/ safety, and ethical.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_scores.R
\name{efa_scores}
\alias{efa_scores}
\title{Estimate factor scores and score-quality diagnostics for an EFA model}
\source{
Thurstone, L. L. (1935). \emph{The vectors of mind}. University of Chicago
Press.

Bartlett, M. S. (1937). The statistical conception of mental factors.
\emph{British Journal of Psychology, 28}, 97-104.

Anderson, T. W., & Rubin, H. (1956). Statistical inference in factor
analysis. In \emph{Proceedings of the Third Berkeley Symposium on Mathematical
Statistics and Probability} (Vol. 5, pp. 111-150). University of California
Press.

Guttman, L. (1955). The determinacy of factor score matrices with
implications for five other basic problems of common-factor theory.
\emph{British Journal of Statistical Psychology, 8}, 65-81.

ten Berge, J. M. F., Krijnen, W. P., Wansbeek, T., & Shapiro, A. (1999).
Some new results on correlation-preserving factor scores prediction methods.
\emph{Linear Algebra and its Applications, 289}, 311-318.

Grice, J. W. (2001). Computing and evaluating factor scores. \emph{Psychological
Methods, 6}, 430-450.
}
\usage{
efa_scores(
  x,
  f,
  Phi = NULL,
  rho = NULL,
  method = c("regression", "Bartlett", "Anderson", "tenBerge", "Harman", "components")
)
}
\arguments{
\item{x}{data.frame or matrix. Raw data (needed to obtain factor scores) or a
correlation matrix (yields weights and diagnostics only). When \code{f} is a
directly supplied loading matrix, a correlation-matrix \code{x} also supplies the
correlations the weights are derived from; when \code{f} is an \code{\link[=efa_fit]{efa_fit()}} object,
its own fitted correlations are used instead (supply \code{rho} to derive the
weights from another matrix). \code{x} describes the model variables either way.
When raw data carry column names, they are matched to the model variables by
name (any extra columns are ignored, and a model variable missing from \code{x} is
an error). A named correlation matrix is likewise matched to the loading rows
by name; its row and column names must use the same order, and it must carry
one row and column for each model variable. Unnamed input is matched by
position.

Raw data are scored as supplied: no imputation is performed, so a case with a
missing value on any model variable receives \code{NA} scores (and is reported as
not scored). A model variable that carries no usable spread in \code{x} -- constant,
infinite, or observed fewer than twice -- is an error.}

\item{f}{object of class \code{\link[=efa_fit]{efa_fit()}}, an \code{efa_loadings} object, or a matrix of factor
loadings.}

\item{Phi}{matrix. Factor intercorrelations. Only used when a loading matrix is
supplied directly in \code{f}; taken from the \code{efa} object otherwise, in which
case a supplied \code{Phi} is ignored with a warning. Named rows and columns are
matched to the loading columns and must use the same order.
Default is \code{NULL}, in which case the factors are assumed uncorrelated.}

\item{rho}{matrix. Correlation matrix used to derive the scoring weights.
Defaults to \code{NULL}, in which case \code{f$orig_R} is used for an \code{efa} object; for
a directly supplied loading matrix, \code{x} itself when it is a correlation
matrix, and \code{cor(x, use = "pairwise")} otherwise. Pass a matrix here to score
against a correlation other than the one implied by \code{f}/\code{x}. Named rows and
columns are matched to the loading rows; row and column names must use the
same order.}

\item{method}{character. The factor-score method: one of \code{"regression"}
(default), \code{"Bartlett"}, \code{"Anderson"}, \code{"tenBerge"}, \code{"Harman"}, or
\code{"components"}.}
}
\value{
An object of class \code{efa_scores}, a list containing:

\item{weights}{The \code{p} by \code{m} factor-score weight matrix.}
\item{scores}{The factor scores (\code{n} by \code{m}), or \code{NULL} when a correlation
matrix was supplied. A case with a missing value on any model variable is not
scored and keeps \code{NA} in every column.}
\item{r.scores}{The \code{m} by \code{m} correlations of the factor-score estimates (see
Details for the \code{"components"}-method scale caveat).}
\item{score_cor}{The \code{m} by \code{m} score-factor correlation matrix; its diagonal
is the determinacy (validity) of each score and its off-diagonals the
univocality.}
\item{determinacy}{A data frame with, per factor, the determinacy \code{rho}, the
squared determinacy \code{rho2}, and Guttman's indeterminacy index \code{guttman}.}
\item{settings}{A list of the settings used, including the number of supplied
observations \code{n_obs} and the number of them that could be scored \code{n_scored}.}
}
\description{
Computes factor-score weights, and (from raw data) the factor scores
themselves, for an \code{\link[=efa_fit]{efa_fit()}} solution or a directly supplied loading matrix.
It also returns score-quality diagnostics: the score intercorrelations, the
determinacy (validity) and univocality of each score, and Guttman's
indeterminacy index. Factor scores are returned only when raw data are
supplied; a correlation matrix yields the weights and diagnostics alone.
}
\details{
Each method combines the loadings with some or all of the factor correlations and the
scoring correlation matrix into weights in a different way:
\describe{
\item{\code{"regression"}}{Thurstone's (1935) regression scores.}
\item{\code{"Bartlett"}}{Bartlett's (1937) scores.}
\item{\code{"Anderson"}}{Anderson & Rubin's (1956) scores.}
\item{\code{"tenBerge"}}{ten Berge, Krijnen, Wansbeek & Shapiro's (1999) scores.}
\item{\code{"Harman"}}{Harman's (1976) scores, based on an idealized variable (a
hypothetical variable that would correlate perfectly with the factor).}
\item{\code{"components"}}{Component scores. These are formed from the raw,
uncentered data (\code{X \%*\% W}) rather than the standardized data, so unlike
the other methods they are on the scale of the input variables. The
diagnostics below describe the standardized combination \code{scale(X) \%*\% W},
and therefore differ from the realized correlations of the returned scores
whenever the variables have unequal variances.}
}
The determinacy (validity) of a score is its correlation with the factor it
estimates, computed from the returned weights; for regression scores it is the
multiple correlation between the factor and the observed variables (Guttman,
1955; Grice, 2001). The off-diagonal score-factor correlations give the
univocality (the correlation of a score with the \emph{other} factors). Guttman's
(1955) indeterminacy index, \verb{2 rho^2 - 1}, is the minimum correlation between
two equally valid sets of scores. For a \code{method} other than \code{"regression"}
both quantities are specific to those scores: the determinacy is the method's
own score-factor correlation (never larger than the regression value), and
the reported \code{guttman} follows from it.

Determinacies close to 1 mean the scores stand in for the factor with little
loss; Grice (2001) regards values of about .90 and above as the level required
before scores are interpreted for individual cases, and treats lower values as
usable only for group-level research. The Guttman index makes the same point
more sharply, because a factor score is never the factor: at \code{rho = .90} two
equally valid sets of scores can still correlate as low as .62, and at
\code{rho = .80} as low as .28, so the rank order of cases is not unique.

Which \code{method} to prefer follows from what the scores are for. Regression
scores correlate most highly with the factor, but they are biased towards it
and correlate across factors even when the model is orthogonal. Bartlett
scores are conditionally unbiased, which makes them the choice when the scores
stand in for the factor in a later model. \code{"tenBerge"} reproduces the factor
intercorrelations \code{Phi}, so it is the choice when the scores will be correlated
with each other or with external variables. \code{"Anderson"} forces the scores to
be uncorrelated with unit variance and is appropriate only when the factors
themselves are orthogonal. \code{"components"} is a weighted sum of the observed
variables rather than an estimate of a common factor.
}
\examples{
# Weights and score diagnostics from an EFA on a correlation matrix
efa <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
               estimator = "PAF", rotation = "oblimin")
fs <- efa_scores(test_models$baseline$cormat, f = efa)
fs
summary(fs)

# Factor scores from raw data (Bartlett method)
\donttest{
efa_raw <- efa_fit(GRiPS_raw, n_factors = 1, estimator = "PAF")
efa_scores(GRiPS_raw, f = efa_raw, method = "Bartlett")
}

# Loadings supplied directly, with the factor intercorrelations
efa_scores(test_models$baseline$cormat, f = efa$rot_loadings, Phi = efa$Phi)

}
\seealso{
\code{\link[=efa_fit]{efa_fit()}} for the solution these are computed from.

Other factor scoring:
\code{\link[=print.efa_scores]{print.efa_scores()}}
}
\concept{factor scoring}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_scores.R
\name{print.efa_scores}
\alias{print.efa_scores}
\alias{format.efa_scores}
\alias{summary.efa_scores}
\alias{print.summary.efa_scores}
\alias{format.summary.efa_scores}
\title{Print and format an efa_scores object}
\usage{
\method{print}{efa_scores}(x, digits = 3, ...)

\method{format}{efa_scores}(x, digits = 3, ...)

\method{summary}{efa_scores}(object, digits = 3, ...)

\method{print}{summary.efa_scores}(x, ...)

\method{format}{summary.efa_scores}(x, digits = x$opts$digits, ...)
}
\arguments{
\item{x, object}{An object of class \code{efa_scores}; for the \code{summary.efa_scores}
methods, the object returned by \code{summary()}.}

\item{digits}{numeric. Number of decimal places for the printed tables.
Default is 3.}

\item{...}{Not used; for consistency with the generics.}
}
\value{
\code{print()} and the print method for \code{summary.efa_scores} objects return
their argument invisibly. \code{format()} returns a character vector with the
report lines. \code{summary()} returns an object of class \code{summary.efa_scores}.
}
\description{
\code{print()} shows a concise overview of an \code{\link[=efa_scores]{efa_scores()}} result: a header naming
the method and whether factor scores were computed, and the per-factor
determinacy table (determinacy, squared determinacy, and Guttman index).
\code{summary()} returns a \code{summary.efa_scores} object whose print method adds the
full factor-weight matrix, the score validity/univocality matrix, and the score
intercorrelations. \code{format()} assembles the same report and returns it as a
character vector; \code{print()} is \code{cat(format(x), sep = "\\n")}. The lines follow
the active console theme, so they are plain when colours are disabled (for
example when captured into a file or stripped with \code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
efa <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500,
               estimator = "PAF", rotation = "oblimin")
fs <- efa_scores(test_models$baseline$cormat, f = efa)
fs
summary(fs)

# format() returns the same lines as a character vector:
writeLines(format(fs))

}
\seealso{
Other factor scoring:
\code{\link[=efa_scores]{efa_scores()}}
}
\concept{factor scoring}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{BARTLETT}
\alias{BARTLETT}
\title{Bartlett's test of sphericity}
\usage{
BARTLETT(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used.}

\item{use}{character. The missing-data policy for raw data. Passed to
\code{\link[stats:cor]{stats::cor()}} for \code{"pearson"}, \code{"spearman"}, and \code{"kendall"}; for \code{"poly"} /
\code{"tetra"} the same policies are applied to the raw data before the polychoric
estimation, where \code{"all.obs"} and \code{"everything"} abort on a missing value instead
of returning \code{NA} correlations. Default is "pairwise.complete.obs".}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator).
Default is "pearson".}
}
\value{
A list of class \code{c("efa_bartlett", "BARTLETT")}, identical to the value
of \code{\link[=efa_bartlett]{efa_bartlett()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{BARTLETT()} has been superseded by \code{\link[=efa_bartlett]{efa_bartlett()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_bartlett]{efa_bartlett()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/procrustes-consensus.R
\name{.consensus_target_procrustes_single}
\alias{.consensus_target_procrustes_single}
\title{Internal single-start GPA-consensus engine}
\usage{
.consensus_target_procrustes_single(
  unrotated_list,
  init_targets = NULL,
  rotation = c("orthogonal", "oblique"),
  start = 1,
  tol = 0.001,
  loss_tol = 1e-06,
  loss_patience = 5,
  convergence = c("either", "target", "loss", "both"),
  min_iter = 2,
  max_iter = 200,
  alpha = 1,
  match_target = TRUE,
  hyper_cutoff = 0.15,
  verbose = FALSE
)
}
\arguments{
\item{unrotated_list}{List of unrotated loading matrices to be aligned. All
matrices must be numeric, finite, and have identical dimensions.}

\item{init_targets}{Optional list of starting target matrices. These are
typically rotated loading matrices from the corresponding analyses. If
\code{NULL}, \code{unrotated_list} is used.}

\item{rotation}{Character string, either \code{"orthogonal"} or \code{"oblique"}.}

\item{start}{Either a single integer selecting an element of \code{init_targets},
or an explicit target matrix. Used when \code{multi_start = FALSE}.}

\item{tol}{Positive relative Frobenius-norm convergence tolerance for the
outer target update.}

\item{loss_tol}{Positive tolerance for the relative change in the outer
consensus loss. If \code{NULL}, loss-based convergence is disabled. It cannot be
\code{NULL} when \code{convergence} is \code{"loss"} or \code{"both"}.}

\item{loss_patience}{Positive integer. Number of consecutive iterations with
relative loss change below \code{loss_tol} required for loss-based convergence.}

\item{convergence}{Character string controlling the stopping rule. \code{"either"}
stops when either target or loss convergence is satisfied; \code{"target"} uses
only target change; \code{"loss"} uses only loss change; \code{"both"} requires both.}

\item{min_iter}{Non-negative integer. Minimum number of outer iterations
before convergence can be declared.}

\item{max_iter}{Positive integer. Maximum number of outer consensus
iterations.}

\item{alpha}{Damping factor for the target update. \code{alpha = 1} uses the full
centroid update. Smaller values, such as \code{0.5}, can reduce oscillation.}

\item{match_target}{Logical. If \code{TRUE}, the updated centroid is signed and
column-matched to the previous target before convergence is evaluated.}

\item{hyper_cutoff}{Non-negative cutoff used by \code{.hyperplane_count()} for
summary output.}

\item{verbose}{Logical; if \code{TRUE}, print convergence messages for the outer
loop.}
}
\description{
Performs a single GPA-consensus run from one starting target. The
multi-start wrapper \code{\link[=.gpa_consensus_target]{.gpa_consensus_target()}} dispatches here.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_smt.R
\name{efa_smt}
\alias{efa_smt}
\title{Sequential chi square model tests, RMSEA lower bound, and AIC}
\source{
Auerswald, M., & Moshagen, M. (2019). How to determine the number of
factors to retain in exploratory factor analysis: A comparison of extraction
methods under realistic conditions. Psychological Methods, 24(4), 468–491.
https://doi.org/10.1037/met0000200

Browne, M.W., & Cudeck, R. (1992). Alternative ways of assessing model
fit. Sociological Methods and Research, 21, 230–258.

Preacher, K. J., Zhang G., Kim, C., & Mels, G. (2013). Choosing the
Optimal Number of Factors in Exploratory Factor Analysis: A Model Selection
Perspective, Multivariate Behavioral Research, 48(1), 28-56,
doi:10.1080/00273171.2012.710386

Steiger, J. H., & Lind, J. C. (1980, May). Statistically based tests
for the number of common factors. Paper presented at the annual meeting of
the Psychometric Society, Iowa City, IA.
}
\usage{
efa_smt(
  x,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  estimate_control = NULL
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used. Must be larger than the number of variables.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is "pairwise.complete.obs".}

\item{cor_method}{character. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"},
passed to \code{\link[stats:cor]{stats::cor()}}. \code{"poly"} and \code{"tetra"} are not supported because
\code{SMT} rests on a normal-theory chi-square test that is not valid for
polychoric / tetrachoric correlations.
Default is \code{"pearson"}.}

\item{estimate_control}{an \code{\link[=estimate_control]{estimate_control()}} object with the estimation settings for the
sequential \code{\link[=efa_fit]{efa_fit()}} fits. \code{NULL} (default) uses the \code{\link[=efa_fit]{efa_fit()}} defaults. The sequential
models are fitted with maximum likelihood (the chi-square, RMSEA, and AIC the SMT is built on
are defined for it), so of the estimation knobs only \code{start_method} takes effect; the ones
governing principal axis factoring do not apply. The models are unrotated, so no rotation
settings apply either.}
}
\value{
An object of class \code{efa_retention} (see \code{\link[=print.efa_retention]{print.efa_retention()}} for
the print method). SMT has no plot; \code{\link[=plot.efa_retention]{plot.efa_retention()}} returns \code{NULL}
with a message for it. Its main fields are:
\item{n_factors}{A named numeric vector (\code{"chi"}, \code{"RMSEA"}, \code{"AIC"}) with the
suggested number of factors from the sequential chi-square model tests, the
RMSEA lower bound, and the AIC.}
\item{results}{A list with one record per criterion, each holding the criterion
values for the null model (zero factors) through the maximum number of
factors.}
\item{settings}{A list of the settings used.}
}
\description{
Sequential chi square model tests (SMT) are a factor retention method where
multiple
EFAs with increasing numbers of factors are fitted and the number of factors
for which the Chi Square value first becomes non-significant is taken as the
suggested number of factors.
Preacher, Zhang, Kim, & Mels (2013) suggested a similar approach with the
lower bound of the 90\% confidence interval of the Root Mean Square Error of
Approximation (RMSEA; Browne & Cudeck, 1992; Steiger & Lind, 1980), and with
the Akaike Information Criterion (AIC). For the RMSEA, the
number of factors for which this lower bound first falls below .05 is the
suggested number of factors to retain. For the AIC, it is the number of factors
where the AIC is lowest.
}
\details{
As a first step in the procedure, a maximum number of factors to extract is
determined for which the model is still over-identified (df > 0).

Then, EFAs with increasing numbers of factors from 1 to the maximum number are
fitted with maximum likelihood estimation.

For the SMT, first the significance of the chi
square value for a model with 0 factors is determined. If this value is
not significant, 0 factors are suggested to retain. If it is significant,
a model with 1 factor is estimated and the significance of its chi square value
is determined, and so on, until a non-significant result is obtained. The
suggested number of factors is the number of factors for the model where the
chi square value first becomes non-significant.

Regarding the RMSEA, the suggested number of factors is the number of factors
for the model where the lower bound of the 90\% confidence interval of the
RMSEA first falls below the .05 threshold.

Regarding the AIC, the suggested number of factors is the number of factors
for the model with the lowest AIC.

The sequential models are fitted without inequality constraints, so a solution
can be inadmissible (a Heywood case, or a fit that did not converge). Only the
models the three rules actually select are checked for this; if one of them is
inadmissible a warning is raised and the corresponding suggestion should be
interpreted with caution.

In comparison with other prominent factor retention criteria, SMT performed
well at determining the number of factors to extract in EFA (Auerswald &
Moshagen, 2019). The RMSEA lower bound also performed well at determining the true
number of factors, while the AIC performed well at determining the
most generalizable model (Preacher, Zhang, Kim, & Mels, 2013).
}
\examples{
SMT_base <- efa_smt(test_models$baseline$cormat, N = 500)
SMT_base

}
\seealso{
\code{\link[=efa_retain]{efa_retain()}} as a wrapper function for this and the other factor
retention criteria.

Other factor retention criteria:
\code{\link[=efa_cd]{efa_cd()}},
\code{\link[=efa_ekc]{efa_ekc()}},
\code{\link[=efa_hull]{efa_hull()}},
\code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_map]{efa_map()}},
\code{\link[=efa_nest]{efa_nest()}},
\code{\link[=efa_parallel]{efa_parallel()}},
\code{\link[=efa_retain]{efa_retain()}},
\code{\link[=efa_scree]{efa_scree()}}
}
\concept{factor retention criteria}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_sl_loadings.R
\name{print.efa_sl_loadings}
\alias{print.efa_sl_loadings}
\alias{format.efa_sl_loadings}
\title{Print an efa_sl_loadings object}
\usage{
\method{print}{efa_sl_loadings}(x, ...)

\method{format}{efa_sl_loadings}(
  x,
  cutoff = 0.2,
  digits = 3,
  max_name_length = 10,
  color = TRUE,
  name_style = c("truncate", "abbreviate", "full"),
  max_factors_per_block = NULL,
  sort_loadings = c("none", "primary", "clustered"),
  ...
)
}
\arguments{
\item{x}{class efa_sl_loadings matrix.}

\item{...}{additional arguments passed to print or format.}

\item{cutoff}{numeric. The value at or above which loadings are emphasized
(default is .2). The default is lower than the .3 of an ordinary loading
table (\code{\link[=print.efa_loadings]{print.efa_loadings()}}): the group-factor loadings are residualized,
that is, they carry only the variance left once the general factor has been
partialled out, and are therefore smaller than the corresponding first-order
loadings.}

\item{digits}{numeric. Passed to \code{\link[base:Round]{round}}. Number of digits
to round the loadings to (default is 3).}

\item{max_name_length}{numeric. The maximum length of the variable names to
display; see \code{\link[=print.efa_loadings]{print.efa_loadings()}}.}

\item{color}{logical. Whether to apply console styling using \pkg{cli}.
Default is \code{TRUE}.}

\item{name_style}{character. How to shorten variable names longer than
\code{max_name_length}; see \code{\link[=print.efa_loadings]{print.efa_loadings()}}.}

\item{max_factors_per_block}{numeric or \code{NULL}. Maximum number of factor
columns to print per block. If \code{NULL}, the number is chosen from the
console width.}

\item{sort_loadings}{character. Optional row sorting; see
\code{\link[=print.efa_loadings]{print.efa_loadings()}}. The default \code{"none"} keeps the input order. When
sorting is requested, rows are grouped by their largest \emph{group}-factor
loading: the general factor is left out of the comparison, since it is the
largest loading of almost every item and sorting on it would leave the order
untouched.}
}
\value{
\code{print()} returns its argument \code{x} invisibly; it is
\code{cat(format(x, ...), sep = "\\n")} followed by a blank line for console
spacing. \code{format()} returns a character vector with the table lines (styled
to the active console theme; plain when colours are disabled).
}
\description{
Print an efa_sl_loadings object
}
\details{
Prints a Schmid-Leiman loading matrix (general factor, group factors, and the
communality/uniqueness columns) as a styled, decimal-aligned table. Loadings with
absolute value greater than or equal to \code{cutoff} are emphasised, smaller loadings are
de-emphasised, and Heywood-relevant cells (a loading or communality above 1, or a
negative uniqueness) are highlighted. If the matrix has many columns or the console is
narrow, the table is split into stacked column blocks so the output stays readable.
}
\examples{
EFA_mod <- efa_fit(test_models$baseline$cormat, N = 500, n_factors = 3,
                   estimator = "PAF", rotation = "promax")
efa_schmid_leiman(EFA_mod, estimator = "PAF")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_retain.R
\name{print.efa_retain}
\alias{print.efa_retain}
\title{Print method for efa_retain objects}
\usage{
\method{print}{efa_retain}(x, ...)
}
\arguments{
\item{x}{an object of class efa_retain, returned by \code{\link[=efa_retain]{efa_retain()}}.}

\item{...}{not used.}
}
\value{
\code{print()} returns its argument \code{x} invisibly; it is
\code{cat(format(x), sep = "\\n")}.
}
\description{
Print method for efa_retain objects
}
\examples{
\donttest{
efa_retain(test_models$baseline$cormat, criteria = c("EKC", "SMT"), N = 500)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.efa_compare.R
\name{plot.efa_compare}
\alias{plot.efa_compare}
\title{Plot efa_compare object}
\usage{
\method{plot}{efa_compare}(x, plot_red = NULL, ...)
}
\arguments{
\item{x}{list. An object of class \code{efa_compare} (output from the
\code{\link[=efa_compare]{efa_compare()}} function).}

\item{plot_red}{numeric or \code{NULL}. Threshold above which to draw the absolute
differences in red, documented in \code{\link[=efa_compare]{efa_compare()}}. \code{NULL} (default) uses the
value recorded in \code{x$settings}; supplying one overrides it for this plot only,
so the comparison need not be recomputed to redraw it at another threshold.}

\item{...}{not used.}
}
\value{
A ggplot object showing the absolute differences, with differences
above \code{plot_red} highlighted in red.
}
\description{
Plot method for the \code{\link[=efa_compare]{efa_compare()}} function showing the distribution of the
absolute differences between the two compared objects as a violin plot with
jittered points. Differences above the threshold are highlighted.
}
\examples{
# A type SPSS EFA to mimick the SPSS implementation
EFA_SPSS_5 <- efa_fit(IDS2_R, n_factors = 5,
                      estimate_control = estimate_control(type = "SPSS"),
                      rotate_control = rotate_control(type = "SPSS"))

# A type psych EFA to mimick the psych::fa() implementation
EFA_psych_5 <- efa_fit(IDS2_R, n_factors = 5,
                       estimate_control = estimate_control(type = "psych"),
                       rotate_control = rotate_control(type = "psych"))

# compare the two and plot the differences
comp <- efa_compare(EFA_SPSS_5$unrot_loadings, EFA_psych_5$unrot_loadings,
                    x_labels = c("SPSS", "psych"))
plot(comp)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.rotate_simplimax_oblq}
\alias{.rotate_simplimax_oblq}
\title{Oblique simplimax factor rotation}
\usage{
.rotate_simplimax_oblq(
  L,
  k,
  eps = 1e-05,
  normalize = TRUE,
  random_starts = 0L,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1
)
}
\arguments{
\item{L}{Numeric matrix. The unrotated loading matrix (variables by factors).}

\item{k}{Integer scalar. The number of "close-to-zero" loadings the criterion targets; must
be in \verb{[1, nrow(L) * ncol(L)]}. \code{k = nrow(L)} is the usual default.}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient norm. Because the
simplimax criterion is only piecewise smooth, the projected gradient need not reach this
tolerance at the optimum; convergence is then reported when the criterion value stalls (the
non-monotone search described above), so \code{eps} mainly governs the smooth phases of the search.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization before
rotation and reverse it afterwards.}

\item{random_starts}{Integer scalar. Number of random orthogonal starts fully optimized in
addition to the identity start.}

\item{maxit}{Integer scalar. Maximum number of projected-gradient updates per start.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving attempts after
the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient update.}
}
\value{
A named list with the rotated loadings, the transformation matrix \code{Th}
(with \code{L \%*\% t(solve(Th))} reproducing the rotated loadings), the factor correlation
matrix \code{Phi} (\code{t(Th) \%*\% Th}), the attained criterion value, and the convergence and
validity flags. The list additionally reports the criterion value reached
at each optimized start in \code{all_values}, with a per-start convergence flag in \code{all_converged}.
}
\description{
Rotate a loading matrix obliquely under the simplimax criterion using a gradient-projection
optimizer along the oblique (column-normalized) manifold.
}
\details{
The criterion value \code{f} and its gradient \code{dQ/dL} at the rotated loadings
\code{L = A \%*\% solve(t(T))} define the search; the engine maps the gradient to the
transformation \code{T} on the manifold \code{diag(t(T) \%*\% T) = 1}, projects it onto the tangent
space, performs a non-monotone line search, and retracts back onto the manifold by column
normalization. The simplimax criterion sums the \code{k} smallest squared loadings, so it is
minimized when the \code{k} "close-to-zero" loadings are driven toward zero; the count \code{k} is a
tuning parameter. Because the set of \code{k} smallest loadings is reselected at every evaluation,
the criterion is only piecewise smooth: its gradient jumps as loadings cross the kth-smallest threshold, so the
line search accepts a step whenever it decreases the largest objective over a short window of
recent iterations (a non-monotone test; Grippo, Lampariello, & Lucidi, 1986), letting the
optimizer step across the kinks where a strictly monotone descent would stall.

The criterion is strongly prone to local minima, so the solver fully optimizes the identity
start together with \code{random_starts} random orthogonal starts and keeps the solution with the
lowest criterion value. Fully optimizing every start -- rather than the screen-and-triage
strategy used for the smooth criteria, which assumes the rational start lies in the global
basin -- is the standard remedy for the local minima of complexity-based rotation criteria
(Kiers, 1994; Browne, 2001).
}
\references{
Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
software for arbitrary rotation criteria in factor analysis. \emph{Educational and
Psychological Measurement}, 65, 676-696.

Browne, M. W. (2001). An overview of analytic rotation in exploratory factor analysis.
\emph{Multivariate Behavioral Research}, 36, 111-150.

Grippo, L., Lampariello, F., & Lucidi, S. (1986). A nonmonotone line search technique for
Newton's method. \emph{SIAM Journal on Numerical Analysis}, 23, 707-716.

Kiers, H. A. L. (1994). Simplimax: Oblique rotation to an optimal target with simple
structure. \emph{Psychometrika}, 59, 567-579.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.simulate_cfm_eigen}
\alias{.simulate_cfm_eigen}
\title{Reference eigenvalues for the efa_nest() simulation via the shared kernel.}
\usage{
.simulate_cfm_eigen(nf, N, Lambda, Psi, nreps = 1000L)
}
\arguments{
\item{nf}{integer. Position of the empirical eigenvalue being tested (1-based);
the \code{nf}-th largest simulated eigenvalue is returned per replicate.}

\item{N}{integer. Number of cases / observations per simulated dataset.}

\item{Lambda}{numeric matrix. Loadings of the \code{(nf - 1)}-factor reference model
(\verb{p x (nf - 1)}); pass a \verb{p x 0} matrix for the \code{nf == 1} null (identity) model.}

\item{Psi}{numeric vector. Uniquenesses (\code{1 - h2}) of the reference model.}

\item{nreps}{integer. Number of datasets to simulate.}
}
\description{
Internal helper called from efa_nest(). Simulates \code{nreps} datasets from an
\code{(nf - 1)}-factor reference model, given that model's loadings \code{Lambda} and
uniquenesses \code{Psi}, and returns the \code{nf}-th largest eigenvalue of each simulated
correlation matrix. The data are drawn with the shared Z * M rule (see
.simulate_cfm_mvn()) using the factor-score square root
\verb{M = t([Lambda | diag(sqrt(Psi))])}, so a row \code{randn(1, nf - 1 + p) * M} is
N(0, Lambda Lambda' + diag(Psi)). Drawing \code{nf - 1 + p} standard normals and
post-multiplying by the factor-score matrix is faster than forming the model-
implied correlation matrix and drawing from it, and matches the position at which
efa_nest() reads the reference eigenvalue.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.rotate_cf_orth}
\alias{.rotate_cf_orth}
\title{Orthogonal Crawford-Ferguson factor rotation}
\usage{
.rotate_cf_orth(
  L,
  kappa,
  eps = 1e-05,
  normalize = TRUE,
  random_starts = 0L,
  maxit = 1000L,
  max_line_search = 10L,
  step0 = 1,
  screen_keep = 5L,
  triage_maxit = 25L,
  triage_improve_tol = 0
)
}
\arguments{
\item{L}{Numeric matrix. The unrotated loading matrix (variables by factors).}

\item{kappa}{Numeric scalar in \verb{[0, 1]}. The Crawford-Ferguson parameter.}

\item{eps}{Numeric scalar. Convergence tolerance for the projected-gradient norm.}

\item{normalize}{Logical scalar. If \code{TRUE}, apply Kaiser normalization before
rotation and reverse it afterwards.}

\item{random_starts}{Integer scalar. Number of additional random orthogonal starts.}

\item{maxit}{Integer scalar. Maximum number of projected-gradient updates.}

\item{max_line_search}{Integer scalar. Maximum number of step-halving attempts after
the initial trial step in each line-search phase.}

\item{step0}{Numeric scalar. Initial step size used in the projected-gradient update.}

\item{screen_keep}{Integer scalar. Number of screened random starts retained for
triage optimization.}

\item{triage_maxit}{Integer scalar. Number of short optimization iterations used in
the triage stage.}

\item{triage_improve_tol}{Numeric scalar. Relative improvement required for a triaged
start to be promoted to full optimization.}
}
\value{
A named list with the rotated loadings, the orthogonal rotation matrix \code{Th}
(with \code{L \%*\% Th} reproducing the rotated loadings), the attained criterion value, and
the convergence and validity flags. The list additionally reports the
criterion value reached at each optimized start in \code{all_values}, with a per-start convergence flag in \code{all_converged}.
}
\description{
Rotate a loading matrix orthogonally under the Crawford-Ferguson criterion using a
gradient-projection optimizer along the orthogonal (Stiefel) manifold.
}
\details{
The criterion value \code{f} and its gradient \code{dQ/dL} at the rotated loadings
\code{L = A \%*\% T} define the search; the engine maps the gradient to the orthogonal
transformation \code{T}, projects it onto the tangent space, performs a
non-monotone line search, and retracts back onto the orthogonal group via a
polar (singular value) projection. \code{kappa = 0} is the quartimax criterion and
\code{kappa = ncol(A) / (2 * nrow(A))} is the equamax criterion.

Additional random orthogonal starts may be requested. To bound runtime the solver
screens each random start by its objective, runs a short triage optimization on the
best-screened starts, and fully optimizes only those that improve on the current
incumbent by at least \code{triage_improve_tol}.
}
\references{
Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms and
software for arbitrary rotation criteria in factor analysis. \emph{Educational and
Psychological Measurement}, 65, 676-696.

Crawford, C. B., & Ferguson, G. A. (1970). A general rotation criterion and its use
in orthogonal rotation. \emph{Psychometrika}, 35, 321-332.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/procrustes-consensus.R
\name{.orthogonal_procrustes}
\alias{.orthogonal_procrustes}
\title{Closed-form orthogonal Procrustes rotation}
\usage{
.orthogonal_procrustes(A, B)
}
\arguments{
\item{A}{Numeric matrix to be rotated.}

\item{B}{Numeric target matrix with the same dimensions as \code{A}.}
}
\value{
A list with the rotated loadings, orthogonal transformation matrix,
target criterion value, and basic diagnostics.
}
\description{
Rotate \code{A} to the orthogonal target \code{B} by minimizing
\verb{||A \%*\% T - B||_F^2} subject to \code{t(T) \%*\% T = I}.
}
\references{
Schoenemann, P. H. (1966). A generalized solution of the orthogonal
Procrustes problem. \emph{Psychometrika}, 31, 1-10.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_average.R
\name{efa_average}
\alias{efa_average}
\title{Model averaging across different EFA estimators and types}
\source{
Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
of implementations of principal axis factoring and promax rotation in R and SPSS.
Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x

Hendrickson, A. E., & White, P. O. (1964). Promax: A quick method for
rotation to oblique simple structure. British Journal of Statistical Psychology,
17 , 65–70. doi: 10.1111/j.2044-8317.1964.tb00244.x

Lorenzo-Seva, U., Timmerman, M. E., & Kiers, H. A. L. (2011). The
Hull Method for Selecting the Number of Common Factors, Multivariate Behavioral
Research, 46, 340-364, doi: 10.1080/00273171.2011.564527

Kaiser, H. F. (1958). The varimax criterion for analytic rotation in
factor analysis. Psychometrika, 23, 187–200. doi: 10.1007/BF02289233
}
\usage{
efa_average(
  x,
  n_factors,
  N = NA,
  estimator = "PAF",
  rotation = "promax",
  type = "none",
  averaging = c("mean", "median"),
  trim = 0,
  salience_threshold = 0.3,
  max_iter = 10000,
  init_comm = c("smc", "mac", "unity"),
  criterion = c(0.001),
  criterion_type = c("sum", "max_individual"),
  abs_eigen = c(TRUE),
  varimax_type = c("svd", "kaiser"),
  normalize = TRUE,
  k_promax = 2:4,
  k_simplimax = ncol(x),
  p_type = c("norm", "unnorm"),
  precision = 1e-05,
  start_method = c("psych", "factanal"),
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra", "fiml"),
  show_progress = TRUE,
  seed = NULL,
  P_type = lifecycle::deprecated()
)
}
\arguments{
\item{x}{data.frame or matrix. Dataframe or matrix of raw data or matrix with
correlations. If raw data is entered, the correlation matrix is found from the
data.}

\item{n_factors}{numeric. Number of factors to extract.}

\item{N}{numeric. The number of observations. Needs only be specified if a
correlation matrix is used. If input is a correlation matrix and \code{N} = NA
(default), not all fit indices can be computed.}

\item{estimator}{character vector. Any combination of  "PAF", "ML", and "ULS",
to use principal axis factoring, maximum likelihood, or unweighted least
squares, respectively, to fit the EFAs. "MINRES" is accepted as a synonym for
"ULS" (the same estimator). Default is "PAF".
"DWLS", which \code{\link[=efa_fit]{efa_fit()}} does accept, is deliberately not
offered here: it weights each residual correlation by the inverse of its
asymptotic variance, which is only available from raw ordinal data analysed
with \code{cor_method = "poly"} or \code{"tetra"}, whereas every EFA in the grid is fitted
to the single correlation matrix computed once from \code{x}. Fit a DWLS solution
with \code{\link[=efa_fit]{efa_fit()}} directly.}

\item{rotation}{character vector. Either perform no rotation ("none"),
any combination of orthogonal rotations ("varimax", "equamax", "quartimax", "geominT",
"bentlerT", and "bifactorT"; using "orthogonal" runs all of these), or of
oblique rotations ("promax", "oblimin", "quartimin", "simplimax", "bentlerQ",
"geominQ", and "bifactorQ"; using "oblique" runs all of these). Rotation types
(no rotation, orthogonal rotations, and oblique rotations) cannot be mixed.
Default is "promax".}

\item{type}{character vector. Any combination of "none" (default), "EFAtools",
"psych", and "SPSS" can be entered. "none" allows the specification of various
combinations of the arguments controlling both factor extraction methods and
the rotations. The others ("EFAtools", "psych", and "SPSS") take the extraction
and rotation tuning of the respective implementation: this package's default
procedure, the psych package's, and SPSS's. A specific psych implementation
exists for PAF, ML, varimax, and promax. The SPSS implementation exists for
PAF, varimax, and promax. For details, see \code{\link[=efa_fit]{efa_fit()}}. The factor ordering is
the one setting a named \code{type} does not bring here: every solution in the grid
is fitted with the eigenvalue-based ordering, so that the solutions can be
realigned to a common target before averaging.}

\item{averaging}{character. One of "mean" (default), and "median". Controls
whether the different results should be averaged using the (trimmed) mean,
or the median.}

\item{trim}{numeric. If averaging is set to "mean", this argument controls
the trimming of extremes (for details see \code{\link[base:mean]{base::mean()}}).
By default no trimming is done (i.e., trim = 0).}

\item{salience_threshold}{numeric. The threshold to use to classify a pattern
coefficient or loading as salient (i.e., substantial enough to assign it to
a factor). Default is 0.3. Indicator-to-factor correspondences will be inferred
based on this threshold. Note that this may not be meaningful if rotation = "none"
and n_factors > 1 are used, as no simple structure is present there.}

\item{max_iter}{numeric. The maximum number of iterations to perform after which
the iterative PAF procedure is halted with a warning. Default is 10,000. It is
only evaluated for the "PAF" solutions run under \code{type} "none": a named \code{type}
brings the iteration cap that defines it ("SPSS" 25, "psych" 50, and
"EFAtools" 300), and "ML" and "ULS" do not iterate this way. Note
that non-converged procedures are excluded from the averaging procedure.}

\item{init_comm}{character vector. Any combination of "smc", "mac", and "unity".
Controls the methods to estimate the initial communalities in \code{PAF} if
"none" is among the specified types. "smc" will use squared multiple
correlations, "mac" will use maximum absolute correlations, "unity" will use
1s (for details see \code{\link[=efa_fit]{efa_fit()}}). Default is \code{c("smc", "mac", "unity")}.}

\item{criterion}{numeric vector. The convergence criterion used for PAF if
"none" is among the specified types.
If the change in communalities from one iteration to the next is smaller than
this criterion the solution is accepted and the procedure ends.
Default is \code{0.001}.}

\item{criterion_type}{character vector. Any combination of "max_individual" and
"sum". Type of convergence criterion used for PAF if "none" is among the
specified types. "max_individual" selects the maximum change in any of the
communalities from one iteration to the next and tests it against the
specified criterion. "sum" takes the difference of
the sum of all communalities in one iteration and the sum of all communalities
in the next iteration and tests this against the criterion
(for details see \code{\link[=efa_fit]{efa_fit()}}). Default is \code{c("sum", "max_individual")}.}

\item{abs_eigen}{logical vector. Any combination of TRUE and FALSE.
Which algorithm to use in the PAF iterations if "none" is among the specified
types. If FALSE, the loadings are computed from the eigenvalues. This is also
used by the \code{\link[psych:fa]{psych::fa()}} function. If TRUE the
loadings are computed with the absolute eigenvalues as done by SPSS
(for details see \code{\link[=efa_fit]{efa_fit()}}). Default is \code{TRUE}.}

\item{varimax_type}{character vector. Any combination of "svd" and "kaiser".
The type of the varimax rotation performed if "none" is among the specified
types and "varimax", "promax", "orthogonal", or "oblique" is among the specified
rotations. "svd" uses singular value decomposition, as
\code{\link[stats:varimax]{stats::varimax()}} does, and "kaiser" uses the varimax
procedure performed in SPSS. This is the original procedure from Kaiser (1958),
but with slight alterations in the varimax criterion (for details, see
\code{\link[=efa_fit]{efa_fit()}} and Grieder & Steiner, 2022).
Default is \code{c("svd", "kaiser")}.}

\item{normalize}{logical vector. Any combination of TRUE and FALSE.
\code{TRUE} performs a kaiser normalization before the specified rotation(s).
Default is \code{TRUE}.}

\item{k_promax}{numeric vector. The power used for computing the target matrix
P in the promax rotation if "none" is among the specified types and "promax"
or "oblique" is among the specified rotations. Default is \code{2:4}.}

\item{k_simplimax}{numeric. The number of 'close to zero loadings' for the
simplimax rotation if "simplimax" or "oblique" is among the specified rotations. Default
is \code{ncol(x)}, where x is the entered data. It counts loadings, so each value must be a
whole number no larger than the number of loadings in the solution; a simplimax fit
given anything else fails and is reported as an errored solution in the grid.}

\item{p_type}{character vector. Any combination of "norm" and "unnorm".
This specifies how the target matrix P is computed in promax rotation if
"none" is among the specified types and "promax" or "oblique" is among the
specified rotations. "unnorm" will use the unnormalized target matrix as
originally done in Hendrickson and White (1964). "norm" will use a
normalized target matrix (for details see \code{\link[=efa_fit]{efa_fit()}}).
Default is \code{c("norm", "unnorm")}.}

\item{precision}{numeric vector. The tolerance for stopping in the rotation
procedure(s). Default is 10^-5.}

\item{start_method}{character vector. Any combination of "psych" and "factanal".
How to specify the starting values for the optimization procedure for ML.
"psych" takes the starting values specified in \code{\link[psych:fa]{psych::fa()}}.
"factanal" takes the starting values specified in the
\code{\link[stats:factanal]{stats::factanal()}} function. Default is
\code{c("psych", "factanal")}.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw data
is given as input. Default is "pairwise.complete.obs". It is ignored when
\code{cor_method = "fiml"}, which handles the missingness itself, so every case
contributes.}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations of ordinal / binary data
(a two-step estimator), or \code{"fiml"}
for a two-stage full-information maximum-likelihood correlation from raw data
with missing values. With \code{"fiml"} the saturated multivariate-normal mean and
covariance are estimated by an EM algorithm assuming the data are missing at
random and the standardized covariance is analysed, reproducing
\code{psych::corFiml()} followed by \code{psych::fa()} and
\code{lavaan(missing = "two.stage")}, \emph{not} \code{lavaan::efa(missing = "ml")} (see
\code{\link[=efa_fit]{efa_fit()}} and the details). Default is "pearson".}

\item{show_progress}{logical. Whether a progress bar should be shown in the
console. Default is TRUE.}

\item{seed}{numeric or \code{NULL}. An optional seed making the averaging run
reproducible and independent of the number of parallel workers (the grid runs with
\link[future.apply:future_lapply]{future_lapply()}, for which a parallel plan can be set
via \code{\link[future:plan]{future::plan()}}). It matters whenever a criterion-based rotation is included,
since those draw random starts. When supplied, the caller's random-number stream is
restored afterwards, leaving no side effect. Default is \code{NULL}.}

\item{P_type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}} Former name of \code{p_type}. Still
accepted (silently) for backwards compatibility; please use \code{p_type}.}
}
\value{
A list of class \code{c("efa_average", "EFA_AVERAGE")} containing the
components below. Throughout, \code{range} is the width \code{maximum - minimum} of each
cell across the factor solutions, not the interval \verb{[minimum, maximum]}, and
\code{average} is the (trimmed) mean or the median, following \code{averaging}.
\item{orig_R}{Original correlation matrix.}
\item{h2}{A list with the average, standard deviation, minimum, maximum, and
range of the final communality estimates across the factor solutions.}
\item{loadings}{A list with the average, standard deviation, minimum, maximum,
and range of the final loadings across the factor solutions. If rotation was
"none", the unrotated loadings, otherwise the rotated loadings (pattern
coefficients). \code{average} is a cell-wise summary of the solutions and not itself
a fitted solution (see Details).}
\item{Phi}{A list with the average, standard deviation, minimum, maximum, and
range of the factor intercorrelations across factor solutions obtained with
oblique rotations.}
\item{ind_fac_corres}{A matrix with each cell containing the proportion of
the factor solutions in which the respective indicator-to-factor correspondence
occurred, i.e., in which the loading exceeded the specified salience threshold.
Note: Rowsums can exceed 1 due to cross-loadings.}
\item{vars_accounted}{A list with the average, standard deviation, minimum,
maximum, and range of explained variances and sums of squared loadings across
the factor solutions. Based on the unrotated loadings if rotation was "none"
or only one factor was extracted, otherwise on the rotated loadings. Each entry
is a matrix with rows \code{"SS loadings"}, \code{"Prop Tot Var"}, and \code{"Prop Comm Var"};
the last is omitted for a single-factor solution, where it is identically 1, so
those matrices have two rows there and three otherwise.}
\item{fit_indices}{A matrix containing the average, standard deviation,
minimum, maximum, and range for all applicable fit indices across the respective
factor solutions, and the degrees of freedom (df). If the estimator argument
contains ML or ULS: Fit indices derived
from the unrotated factor loadings: Chi Square (chisq), including significance
level, Comparative Fit Index (CFI), Tucker-Lewis Index (TLI), Root Mean Square
Error of Approximation (RMSEA), Akaike Information Criterion (AIC), Bayesian
Information Criterion (BIC), Expected Cross-Validation Index (ECVI), and the
common part accounted for (CAF) index as proposed by Lorenzo-Seva, Timmerman,
& Kiers (2011). The residual-based Standardized Root Mean Square Residual
(SRMR) and Root Mean Square Residual (RMSR) and the CAF are also computed for
PAF; for PAF the remaining (chi-square-based) indices are not available (see
details).}
\item{implementations_grid}{A matrix containing, for each performed EFA,
the setting combination, if an error occurred (logical), the error message
(character), an integer convergence code (0 = converged; for ML and ULS the
same codes as \code{\link[stats:optim]{stats::optim()}}'s "L-BFGS-B", for PAF 1 if the
maximum number of iterations was reached without meeting the convergence
criterion and 0 otherwise),
if heywood cases occurred (logical, see details for definition), if the
solution was admissible (logical, see details for definition), and the fit
indices.}
\item{efa_list}{A list containing the outputs of all performed EFAs. The names
correspond to the rownames from the implementations_grid.}
\item{settings}{A list of the settings used, including \code{seed} (\code{NULL} when none was
supplied).}

If the supplied arguments admit only a single EFA, there is nothing to average
across: that one \code{\link[=efa_fit]{efa_fit()}} object is returned instead, with a warning. Its
\code{settings} are that fit's, with \code{seed} recording the seed the run was governed by.
They therefore describe the concrete arguments the fit ran under rather than the
\code{type} that supplied them: a row taken from a named preset records \code{type = "none"}
together with the preset's resolved values (for example \code{max_iter = 25} for
"SPSS"), and \code{order_type = "eigen"} as for every other row in the grid.
}
\description{
Not all EFA procedures always arrive at the same solution. This function allows
you perform a number of EFAs from different estimators (e.g., Maximum Likelihood
and Principal Axis Factoring), with different implementations (e.g., the SPSS
and psych implementations of Principal Axis Factoring), and across different
rotations of the same type (e.g., multiple oblique rotations, like promax and
oblimin). \code{efa_average()} will then run all these EFAs (using the \code{\link[=efa_fit]{efa_fit()}}
function) and provide a summary across the different solutions.
}
\details{
As a first step in this function, a grid is produced containing the setting
combinations for the to-be-performed EFAs. These settings are then entered as
arguments to the \code{\link[=efa_fit]{efa_fit()}} function and the EFAs are run in a second
step. After all EFAs are run, the factor solutions are averaged and their
variability determined in a third step.

When raw data are supplied, the correlation matrix is computed once before the
grid is run and reused for every EFA in it. Under \code{cor_method = "fiml"} this
means the saturated multivariate-normal moments are EM-estimated a single time
(from the raw data with missing values, assuming the data are missing at
random) and the resulting two-stage correlation is analysed by every solution
in the grid; the EM is not re-run per solution. Under \code{cor_method = "fiml"},
\code{use} does not select cases (every case contributes to the EM). The averaged
loadings and communalities are then the two-stage FIML estimates, but the
averaged Chi-Square and the indices derived from it (CFI, TLI, RMSEA, AIC, BIC,
ECVI) are the ordinary ML/ULS discrepancy statistics on the EM correlation, not
the corrected two-stage (Satorra-Bentler) statistics that a standalone
\code{\link[=efa_fit]{efa_fit()}} with \code{cor_method = "fiml"} reports; in particular the averaged AIC and
BIC are finite here rather than \code{NA}.

The grid containing the setting combinations is produced based on the entries
to the respective arguments. To this end, all possible combinations resulting
in unique EFA models are considered: combinations that resolve to the same
model are run only once. Two combinations are the same model only if every
setting the fit consumes agrees, and for "PAF" that includes the iteration cap
\code{max_iter}. Since a named \code{type} brings its own cap, a \code{type} of
\code{c("none", "SPSS")} whose specific settings match the SPSS combination in
every other respect still gives two "PAF" models, unless \code{max_iter} is also
set to the cap of the SPSS implementation. We include here a list
of arguments that are only evaluated under specific conditions:

The arguments \code{init_comm}, \code{criterion}, \code{criterion_type},
\code{abs_eigen}, and \code{max_iter} are only evaluated if "PAF" is included in
\code{estimator} and "none" is included in \code{type}.

The argument \code{varimax_type} is only evaluated if "varimax", "promax",
"oblique", or "orthogonal" is included in \code{rotation} and "none" is
included in \code{type}.

The argument \code{normalize} is only evaluated if \code{rotation} is not
set to "none" and "none" is included in \code{type}.

The argument \code{k_simplimax} is only evaluated if "simplimax" or "oblique"
is included in \code{rotation}.

The arguments \code{k_promax} and \code{p_type} are only evaluated if
"promax" or "oblique" is included in \code{rotation} and "none" is included
in \code{type}.

The argument \code{start_method} is only evaluated if "ML" is included in
\code{estimator}.

Every solution in the grid is fitted with the eigenvalue-based factor ordering,
including under a named \code{type}: the solutions are realigned to a common target
before averaging, so a per-fit ordering (SPSS orders by the sum of squared
loadings) would not survive into the averaged result. That target is the first
solution the grid retains, in grid order; solutions dropped for an error,
non-convergence, or a Heywood case cannot become it. The averaged loadings are
therefore in the factor order and sign of that solution. This is the one setting a
named \code{type} does not carry, and it is visible in the two places the individual
fits are: the solutions returned in \code{efa_list} are eigenvalue-ordered, and so is
the single \code{\link[=efa_fit]{efa_fit()}} object returned when the grid collapses to one row. Their
loadings can therefore appear in a different factor order than the same fit run
through \code{\link[=efa_fit]{efa_fit()}} under that \code{type}, even though the solutions are the same.

To avoid a bias in the averaged factor solutions from problematic solutions,
these are excluded prior to averaging. A solution is deemed problematic if
at least one of the following is true: an error occurred, the model did not
converge, or there is at least one Heywood (improper) case (a communality at
or above 1, or, for \code{ML}/\code{ULS}, a uniqueness pinned at the estimator's lower
bound).
Information on errors, convergence, and Heywood cases are returned in the
implementations_grid and a summary of these is given when printing the output.
In addition to these, information on the admissibility of the factor solutions
is also included. A solution was deemed admissible if (1) no error occurred,
(2) the model converged, (3) no Heywood cases are present, and (4) there are
at least two salient loadings (i.e., loadings exceeding the specified
\code{salience_threshold}) for each factor. So, solutions failing one of the
first three of these criteria of admissibility are also deemed problematic and
therefore excluded from averaging. However, solutions failing only
the fourth criterion of admissibility are still included for averaging.
Finally, if all solutions are problematic (e.g., all solutions contain
Heywood cases), no averaging is performed and the respective outputs are NA.
In this case, the implementations_grid should be inspected to see if there
are any error messages, and the separate EFA solutions that are also included
in the output can be inspected as well, for example, to see where Heywood
cases occurred.

A core output of this function includes the average, minimum, and maximum
loadings derived from all non-problematic (see above) factor solutions. Please
note that these are not entire solutions, but the matrices include the average,
minimum, or maximum value for each cell (i.e., each loading separately). This
means that, for example, the matrix with the minimum loadings will contain
the minimum value in any of the factor solutions for each specific loading,
and therefore most likely contains loadings from different factor solutions.
The matrices containing the minimum and maximum factor solutions can
therefore not be interpreted as whole factor solutions.

The averaged loading matrix is likewise a cell-wise summary rather than a fitted
solution: it is not itself the solution of any EFA, does not in general reproduce
the correlation matrix, and need not reproduce the averaged communalities. The
fit indices described below are correspondingly the mean (or, under
\code{averaging = "median"}, the median) of the per-solution fit indices, not the fit
of the averaged loadings, so the averaged loadings and the reported fit do not
describe one and the same model.

The output also includes information on the average, minimum, maximum, and
variability of the fit indices across the non-problematic factor solutions.
It is important to note that not all fit indices are computed for all fit
methods: For ML and ULS, all fit indices can be computed, while for PAF the
chi-square-based indices (the chi-square statistic and its significance, CFI,
TLI, RMSEA, AIC, BIC, and ECVI) are NA. The common part accounted for (CAF)
index (Lorenzo-Seva, Timmerman, & Kiers, 2011) and the residual-based SRMR and
RMSR are still computed for PAF. As a consequence, if only "PAF" is included in
the \code{estimator} argument, averaging is performed for the CAF, SRMR, and RMSR, while
the chi-square-based indices are NA. If a combination of "PAF" and "ML" and/or
"ULS" are included in the \code{estimator} argument, the CAF, SRMR, and RMSR are
averaged across all non-problematic factor solutions, while the chi-square-based
indices are only averaged across the ML and ULS solutions. The user should
therefore keep in mind that the number of EFAs across which the fit indices are
averaged can diverge for the CAF, SRMR, and RMSR compared to the chi-square-based
indices.

Each reported fit index is summarised across the (non-problematic) solutions in
the same descriptive way: the average, standard deviation, minimum, and maximum
of the per-solution values. This includes the chi-square significance level
(\code{p_chi}), which is therefore the mean (or median) of the per-solution p-values
and is purely descriptive; it is not the p-value of any pooled chi-square test.
}
\examples{
# Averaging across one implementation each of PAF (EFAtools type), ULS, and
# ML with one implementation of promax (EFAtools type) (3 EFAs)
Aver_meth <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                         estimator = c("PAF", "ULS", "ML"), type = "EFAtools",
                         start_method = "psych")

\donttest{
# Averaging across different implementations of PAF and promax rotation (72 EFAs)
Aver_PAF <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500)

# Use median instead of mean for averaging (72 EFAs)
Aver_PAF_md <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                           averaging = "median")

# Averaging across different implementations of PAF and promax rotation,
# and across ULS and different versions of ML (108 EFAs)
Aver_meth_ext <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                             estimator = c("PAF", "ULS", "ML"))

# Averaging across different oblique rotation methods, using one implementation
# of ML and one implementation of promax (EFAtools type) (7 EFAs)
Aver_rot <- efa_average(test_models$baseline$cormat, n_factors = 3, N = 500,
                         estimator = "ML", rotation = "oblique", type = "EFAtools",
                         start_method = "psych")
}

\donttest{
# Two-stage FIML correlations from raw data with missing values: the EM
# saturated moments are estimated once and the resulting correlation is
# averaged across the grid of EFAs.
x_miss <- GRiPS_raw
x_miss[cbind(1:20, 1)] <- NA
Aver_fiml <- efa_average(x_miss, n_factors = 1, estimator = c("PAF", "ML"),
                         cor_method = "fiml")
}

}
\seealso{
Other factor analysis:
\code{\link[=efa_fit]{efa_fit()}},
\code{\link[=efa_group]{efa_group()}},
\code{\link[=efa_mi]{efa_mi()}},
\code{\link[=plot.efa_group]{plot.efa_group()}},
\code{\link[=print.efa_group]{print.efa_group()}}
}
\concept{factor analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.efa_group.R
\name{plot.efa_group}
\alias{plot.efa_group}
\title{Plot a multigroup factor analysis}
\usage{
\method{plot}{efa_group}(x, type = c("congruence", "differences"), ...)
}
\arguments{
\item{x}{An object of class \code{efa_group} (output from \code{\link[=efa_group]{efa_group()}}).}

\item{type}{character. Which plot to draw: \code{"congruence"} (per-factor congruence with
confidence intervals) or \code{"differences"} (a per-item loading-difference heatmap).}

\item{...}{Not used; for consistency with the generic.}
}
\value{
A \link[ggplot2:ggplot]{ggplot2::ggplot} object.
}
\description{
Two views of an \code{\link[=efa_group]{efa_group()}} result, selected by \code{type}:
}
\details{
\itemize{
\item \code{"congruence"} (the default) plots the matched Tucker congruence of each factor
between every group pair, with a percentile bootstrap confidence interval when one
was computed (\code{b_boot > 0}). The Lorenzo-Seva and ten Berge (2006) reference bands
(\code{.95} "equal", \code{.85} "fair") are drawn so a factor's cross-group similarity can be
read against them at a glance.
\item \code{"differences"} draws a heatmap of the signed cross-group loading differences (item
by factor, one panel per group pair). Cells whose absolute difference reaches the
salience threshold \code{delta} are outlined.
}
}
\examples{
g <- rep(c("g1", "g2"), length.out = nrow(GRiPS_raw))
mg <- efa_group(GRiPS_raw, groups = g, n_factors = 1)

# Per-factor congruence against the Lorenzo-Seva & ten Berge bands
plot(mg)

# Per-item cross-group loading-difference heatmap
plot(mg, type = "differences")
}
\references{
Lorenzo-Seva, U., and ten Berge, J. M. F. (2006). Tucker's congruence coefficient as a
meaningful index of factor similarity. \emph{Methodology}, 2, 57-64.
doi: 10.1027/1614-2241.2.2.57
}
\seealso{
Other factor analysis:
\code{\link[=efa_average]{efa_average()}},
\code{\link[=efa_fit]{efa_fit()}},
\code{\link[=efa_group]{efa_group()}},
\code{\link[=efa_mi]{efa_mi()}},
\code{\link[=print.efa_group]{print.efa_group()}}
}
\concept{factor analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residuals.efa.R
\name{residuals.efa}
\alias{residuals.efa}
\title{Extract residuals from an efa object}
\usage{
\method{residuals}{efa}(object, type = c("raw", "standardized"), ...)
}
\arguments{
\item{object}{a list of class \code{efa}. Output from \code{\link[=efa_fit]{efa_fit()}} or \code{\link[=efa_mi]{efa_mi()}}.}

\item{type}{character. Which residuals to return. \code{"raw"} (default) returns
\code{orig_R - model_implied_R}; \code{"standardized"} returns the standardized
residuals (residuals divided by their standard errors), available only when
the object was fitted with bootstrap standard errors.}

\item{...}{Further arguments (currently unused).}
}
\value{
A numeric matrix of residual correlations.
}
\description{
Returns the residual correlation matrix of an \code{\link[=efa_fit]{efa_fit()}} or \code{\link[=efa_mi]{efa_mi()}}
solution. Residuals are a pure extractor here; their diagnostics and a
formatted display are part of \code{\link[=summary.efa]{summary.efa()}}.
}
\examples{
efa <- efa_fit(test_models$baseline$cormat, n_factors = 3, N = 500)
residuals(efa)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test_models_doc.R
\docType{data}
\name{test_models}
\alias{test_models}
\title{Four test models used in Grieder and Steiner (2022)}
\format{
A list of 4 lists "baseline", "case_1a", "case_6b", and "case_11b", each with the following elements.
\describe{
\item{cormat}{(matrix) - The correlation matrix of the simulated data.}
\item{n_factors}{(numeric) - The true number of factors.}
\item{N}{(numeric) - The sample size of the generated data.}
}
}
\source{
Grieder, S., & Steiner, M. D. (2022). Algorithmic jingle jungle: A comparison
of implementations of principal axis factoring and promax rotation in R and SPSS.
Behavior Research Methods, 54, 54–74. doi: 10.3758/s13428-021-01581-x
}
\usage{
test_models
}
\description{
Correlation matrices created from simulated data from four of the
\code{\link{population_models}} cases, each with strong factor intercorrelations.
These are used in Grieder & Steiner (2022) to compare the psych and SPSS
implementations in this package with the actual implementations of the programs.
For details on the cases, see \code{\link{population_models}}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.efa_kmo.R
\name{print.efa_kmo}
\alias{print.efa_kmo}
\alias{format.efa_kmo}
\title{Print and format an efa_kmo object}
\usage{
\method{print}{efa_kmo}(x, ...)

\method{format}{efa_kmo}(x, ...)
}
\arguments{
\item{x}{An object of class \code{efa_kmo} (output from \code{\link[=efa_kmo]{efa_kmo()}}).}

\item{...}{Not used; for consistency with the generic.}
}
\value{
\code{print()} returns its argument \code{x} invisibly. \code{format()} returns a
character vector with the report lines.
}
\description{
\code{print()} shows the Kaiser-Meyer-Olkin (KMO) criterion computed by \code{\link[=efa_kmo]{efa_kmo()}}: a
titled section with a verdict on the overall KMO value (and what it implies for
the suitability of the data for factor analysis), the overall value, and the
per-variable KMO values. \code{format()} assembles the same report and returns it as
a character vector; \code{print()} is \code{cat(format(x), sep = "\\n")}. The lines follow
the active console theme, so they are plain when colours are disabled (for
example when captured into a file or stripped with \code{\link[cli:ansi_strip]{cli::ansi_strip()}}).
}
\examples{
KMO_base <- efa_kmo(test_models$baseline$cormat)
KMO_base

# format() returns the same lines as a character vector:
writeLines(format(KMO_base))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{PROCRUSTES}
\alias{PROCRUSTES}
\title{Rotate a loading matrix to a target using Procrustes alignment}
\usage{
PROCRUSTES(
  A,
  Target,
  rotation = c("orthogonal", "oblique"),
  S = NULL,
  T_init = NULL,
  oblique_eps = 1e-05,
  oblique_maxit = 1000,
  oblique_max_line_search = 10,
  oblique_step0 = 1,
  oblique_normalize = FALSE,
  oblique_random_starts = 0,
  oblique_screen_keep = 2,
  oblique_triage_maxit = 25,
  oblique_triage_improve_tol = 0
)
}
\arguments{
\item{A}{Numeric loading matrix to be aligned.}

\item{Target}{Numeric target matrix with the same dimensions as \code{A}.}

\item{rotation}{Character string, either \code{"orthogonal"} or \code{"oblique"}.}

\item{S}{Optional \verb{k x k} cross-product matrix \code{crossprod(A)}, kept for
compatibility. It enters both the oblique criterion and its gradient, so any
other matrix would minimize a different criterion: where \code{S} is used it is
checked against \code{crossprod(A)} and must agree with it up to a relative
tolerance of \code{1e-8}. That check forms \code{crossprod(A)} itself, so passing \code{S}
no longer avoids any work: omitting it gives the same result for slightly
less. \code{S} is used, and therefore checked, only on the oblique path with more
than one factor and \code{oblique_normalize = FALSE}; if Kaiser normalization is
requested, the cross-product must be recomputed on the normalized matrix and
\code{S} is ignored.}

\item{T_init}{Optional \verb{k x k} starting transformation matrix for the oblique
solver. Its columns are normalized internally, and the normalized matrix must
be well enough conditioned to define a proper factor correlation matrix: its
smallest singular value must be at least \code{1e-4}, the same floor the solver
applies to every candidate it evaluates. If \code{NULL} (the default), the oblique
solver is warm-started from the closed-form orthogonal Procrustes solution.}

\item{oblique_eps}{Positive convergence tolerance for the projected-gradient
norm in the oblique solver.}

\item{oblique_maxit}{Non-negative integer. Maximum number of projected-gradient
updates in the full oblique solver.}

\item{oblique_max_line_search}{Non-negative integer. Maximum number of
step-halving attempts after the initial line-search step.}

\item{oblique_step0}{Positive initial step size for the oblique solver.}

\item{oblique_normalize}{Logical; if \code{TRUE}, apply Kaiser row normalization to
the loadings (only) in the oblique solver and back-transform the aligned
loadings afterwards, leaving \code{Target} unnormalized (as in
\code{GPArotation::targetQ(normalize = TRUE)}).}

\item{oblique_random_starts}{Non-negative integer. Number of additional random
starts used by the oblique solver.}

\item{oblique_screen_keep}{Non-negative integer. Number of random starts
retained after cheap objective screening and sent to triage optimization.}

\item{oblique_triage_maxit}{Non-negative integer. Number of short optimization
iterations used in the triage stage.}

\item{oblique_triage_improve_tol}{Non-negative scalar. Relative improvement
required for a triaged start to be promoted to full optimization.}
}
\value{
A list identical to the value of \code{\link[=efa_procrustes]{efa_procrustes()}}; see there for the
components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{PROCRUSTES()} has been superseded by \code{\link[=efa_procrustes]{efa_procrustes()}}, which is the
recommended interface going forward. It remains available and unchanged so
existing code keeps working.
}
\seealso{
\code{\link[=efa_procrustes]{efa_procrustes()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{N_FACTORS}
\alias{N_FACTORS}
\title{Various factor retention criteria}
\usage{
N_FACTORS(
  x,
  criteria = c("CD", "EKC", "HULL", "MAP", "NEST", "PARALLEL"),
  suitability = TRUE,
  N = NA,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra"),
  n_factors_max = NA,
  N_pop = 10000,
  N_samples = 500,
  alpha = 0.3,
  max_iter_CD = 50,
  n_fac_theor = NA,
  method = c("ML", "PAF", "ULS"),
  gof = c("CAF", "CFI", "RMSEA"),
  eigen_type_HULL = c("SMC", "PCA", "EFA"),
  eigen_type_other = c("SMC"),
  n_factors = 1,
  n_datasets = 1000,
  percent = 95,
  decision_rule = c("means", "percentile", "crawford"),
  ekc_type = lifecycle::deprecated(),
  n_datasets_nest = 1000,
  alpha_nest = 0.05,
  show_progress = FALSE,
  ...
)
}
\arguments{
\item{x}{data.frame or matrix. Raw data, or a correlation matrix. If \code{"CD"} is
included as a criterion, x must be raw data.}

\item{criteria}{character. A vector with the factor retention methods to
perform. Possible inputs are: \code{"CD"}, \code{"EKC"}, \code{"HULL"}, \code{"KGC"}, \code{"MAP"},
\code{"NEST"}, \code{"PARALLEL"}, \code{"SCREE"}, and \code{"SMT"} (see the details in
\code{\link[=efa_retain]{efa_retain()}}). By default, a subset of often used, well-performing
methods are performed.}

\item{suitability}{logical. Whether the data should be checked for suitability
for factor analysis using Bartlett's test of sphericity and the
Kaiser-Meyer-Olkin criterion (see details). Default is \code{TRUE}.}

\item{N}{numeric. The number of observations. Only needed if x is a
correlation matrix.}

\item{use}{character. Passed to \code{\link[stats:cor]{stats::cor()}} if raw
data is given as input. Default is \code{"pairwise.complete.obs"}.}

\item{cor_method}{character. Correlation computed from raw data: \code{"pearson"},
\code{"spearman"}, or \code{"kendall"} (passed to \code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} /
\code{"tetra"} for polychoric / tetrachoric correlations (a two-step estimator).
\code{CD}, \code{PARALLEL}, \code{NEST}, \code{HULL}, and \code{SMT} do not support \code{"poly"} / \code{"tetra"}
and are skipped automatically if you request them together.
Default is \code{"pearson"}.}

\item{n_factors_max}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. The maximum number
of factors to test against.
Larger numbers will increase the duration the procedure takes, but test more
possible solutions. If left NA (default), the maximum number of factors for
which the model is still over-identified (df > 0) is used.}

\item{N_pop}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. Size of finite populations
of comparison data. Default is 10000.}

\item{N_samples}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. Number of samples drawn
from each population. Default is 500.}

\item{alpha}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. The alpha level used to test
the significance of the improvement added by an additional factor.
Default is .30.}

\item{max_iter_CD}{numeric. Passed to \code{\link[=efa_cd]{efa_cd()}}. The maximum number of
iterations to perform after which the iterative PAF procedure is halted.
Default is 50.}

\item{n_fac_theor}{numeric. Passed to \code{\link[=efa_hull]{efa_hull()}}. Theoretical number
of factors to retain. The Hull method uses one plus the larger of this
number and the number of factors suggested by \code{\link[=efa_parallel]{efa_parallel()}} as its
upper bound.}

\item{method}{character. The estimator to use in the criteria that fit EFA models;
passed to \code{\link[=efa_retain]{efa_retain()}} as its \code{estimator} argument. One of \code{"ML"}, \code{"PAF"}, or
\code{"ULS"}.}

\item{gof}{character. Passed to \code{\link[=efa_hull]{efa_hull()}}. The goodness of fit index
to use. Either \code{"CAF"}, \code{"CFI"}, or \code{"RMSEA"}, or any
combination of them. With the \code{"PAF"} estimator, only
the CAF can be used as goodness of fit index. For details on the CAF, see
Lorenzo-Seva, Timmerman, and Kiers (2011).}

\item{eigen_type_HULL}{character. Passed to \code{\link[=efa_parallel]{efa_parallel()}} in
\code{\link[=efa_hull]{efa_hull()}}. What the eigenvalues in the parallel analysis are based on.
One of \code{"SMC"}, \code{"PCA"}, or \code{"EFA"} -- different ways of estimating how much
variance each indicator shares with the others before the eigenvalues are
computed. \code{"SMC"} (default) uses each indicator's squared multiple
correlation with the others (its diagonal value in the correlation matrix).
\code{"PCA"} leaves the diagonal at 1, so each indicator's total variance -- not
just the shared part -- feeds into the eigenvalues. \code{"EFA"} uses the
communalities from a fitted EFA solution instead.}

\item{eigen_type_other}{character. Passed to \code{\link[=efa_kgc]{efa_kgc()}},
\code{\link[=efa_scree]{efa_scree()}}, and \code{\link[=efa_parallel]{efa_parallel()}}. The same as eigen_type_HULL,
but multiple inputs are possible here (any combination of \code{"PCA"}, \code{"SMC"},
and \code{"EFA"}). Default is \code{"SMC"}.}

\item{n_factors}{numeric. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}), \code{\link[=efa_kgc]{efa_kgc()}}, and \code{\link[=efa_scree]{efa_scree()}}. Number of
factors to extract if \code{"EFA"} is included in \code{eigen_type_HULL} or
\code{eigen_type_other}. Default is 1.}

\item{n_datasets}{numeric. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}). The number of datasets to simulate. Default is 1000.}

\item{percent}{numeric. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}). The percentile to take from the simulated eigenvalues.
Default is 95.}

\item{decision_rule}{character. Passed to \code{\link[=efa_parallel]{efa_parallel()}} (also within
\code{\link[=efa_hull]{efa_hull()}}). Which rule to use to determine the number of
factors to retain. Default is \code{"means"}, which uses the average simulated
eigenvalues. \code{"percentile"} uses the percentiles specified in \code{percent}.
\code{"crawford"} uses the 95th percentile for the first factor and the mean
afterwards (based on Crawford et al., 2010).}

\item{ekc_type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Accepted and ignored. It
used to select between two ways to compute the \code{\link[=efa_ekc]{efa_ekc()}} reference values.
The \code{"AM2019"} reference values do not depend on the observed eigenvalues.
They therefore skip the empirical correction that defines the criterion, so
they are no longer computed.}

\item{n_datasets_nest}{numeric. Passed to \code{\link[=efa_nest]{efa_nest()}}. The number of
datasets to simulate. Default is 1000.}

\item{alpha_nest}{numeric. Passed to \code{\link[=efa_nest]{efa_nest()}}. The alpha level to use.
The reference values are the eigenvalues at the (1 - alpha_nest)
percentile. Default is .05.}

\item{show_progress}{logical. Whether a progress bar should be shown in the
console. Default is FALSE.}

\item{...}{Further arguments passed on to the \code{\link[=efa_fit]{efa_fit()}} fits, including the
estimation tuning knobs (\code{type}, \code{init_comm}, \code{criterion}, \code{criterion_type},
\code{abs_eigen}, \code{start_method}), which are repacked into an \code{\link[=estimate_control]{estimate_control()}} object
so that they tune the fits exactly as they always did. The estimator is selected with
\code{method}; \code{max_iter} is taken by the \code{max_iter_CD} argument (R matches an abbreviated
name against the arguments before \code{...}) and so does not reach the fits.}
}
\value{
A list of class \code{c("efa_retain", "N_FACTORS")}, identical to the
value of \code{\link[=efa_retain]{efa_retain()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{N_FACTORS()} has been superseded by \code{\link[=efa_retain]{efa_retain()}}, which is the recommended
interface going forward. It remains available and unchanged so existing code
keeps working.
}
\seealso{
\code{\link[=efa_retain]{efa_retain()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.simulate_cfm_mvn}
\alias{.simulate_cfm_mvn}
\title{Draw multivariate-normal data from a population correlation matrix.}
\usage{
.simulate_cfm_mvn(R, N, tol = 1e-08)
}
\arguments{
\item{R}{numeric matrix. Population correlation/covariance matrix.}

\item{N}{integer. Number of cases to draw.}

\item{tol}{numeric. Eigenvalues below \code{-tol} mark \code{R} as indefinite.}
}
\description{
Internal helper called from efa_simulate(). Draws \code{N} cases from a
\code{p}-variate normal with correlation (or covariance) \code{R} by post-multiplying a
matrix of standard normal deviates by a matrix square root \code{M} of \code{R} (with
M' M = R, so the rows of Z * M are N(0, R)). This is the same Z * M rule used by
the NEST reference simulation (.simulate_cfm_eigen): there \code{M} is the transposed
factor-score matrix, here it is a Cholesky or eigen square root.
A positive-definite \code{R} is factored by Cholesky; a positive-semidefinite but
singular \code{R} (which makes the Cholesky fail although it is still a valid
covariance, e.g. a no-factor block or a smoothed factor intercorrelation
matrix) falls back to a symmetric eigen square root.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EFAtools-superseded.R
\name{MAP}
\alias{MAP}
\title{Minimum average partial}
\usage{
MAP(
  x,
  use = c("pairwise.complete.obs", "all.obs", "complete.obs", "everything",
    "na.or.complete"),
  cor_method = c("pearson", "spearman", "kendall", "poly", "tetra")
)
}
\arguments{
\item{x}{A numeric \code{matrix} or \code{data.frame}. Can be either (a) a correlation matrix, or
(b) raw data (rows = observations, columns = variables) from which correlations are computed.}

\item{use}{Character string specifying the treatment of missing values when computing correlations.
Passed to \code{\link[stats:cor]{stats::cor()}}. Defaults to \code{"pairwise.complete.obs"}.}

\item{cor_method}{Character string specifying the correlation coefficient to be computed if raw
data are supplied. One of \code{"pearson"}, \code{"spearman"}, or \code{"kendall"} (passed to
\code{\link[stats:cor]{stats::cor()}}), or \code{"poly"} / \code{"tetra"} for polychoric / tetrachoric correlations
of ordinal / binary data (a two-step estimator). Defaults to \code{"pearson"}.}
}
\value{
An object of class \code{efa_retention}, identical to the value of
\code{\link[=efa_map]{efa_map()}}; see there for the components.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

\code{MAP()} has been superseded by \code{\link[=efa_map]{efa_map()}}, which is the recommended interface
going forward. It remains available and unchanged so existing code keeps working.
}
\seealso{
\code{\link[=efa_map]{efa_map()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/efa_power.R
\name{efa_power}
\alias{efa_power}
\title{Power analysis for exploratory factor analysis}
\usage{
efa_power(
  mode = c("rmsea", "simulation"),
  type = c("close", "notclose"),
  eps0 = NULL,
  eps1 = NULL,
  N = NULL,
  p = NULL,
  k = NULL,
  df = NULL,
  alpha = 0.05,
  power = NULL,
  group = 1,
  Lambda = NULL,
  Phi = NULL,
  Psi = NULL,
  R = NULL,
  n_datasets = 500,
  criteria = c("EKC", "MAP"),
  estimator = "PAF",
  rotation = NULL,
  recovery_threshold = 0.95,
  model_error = c("TKL", "CB", "WB", "none"),
  target_rmsea = NULL,
  target_cfi = NULL,
  seed = NULL
)
}
\arguments{
\item{mode}{character. The kind of power analysis: \code{"rmsea"} (the default; analytic
RMSEA power) or \code{"simulation"} (Monte-Carlo hit-rate and structure recovery).
\code{type}, \code{eps0}, \code{eps1}, \code{df}, \code{alpha}, \code{power}, and \code{group} apply to RMSEA mode only;
the arguments marked \emph{Simulation mode} below apply to the other; \code{N}, \code{p}, and \code{k} are
used in both.}

\item{type}{character. The RMSEA test: \code{"close"} (test of close fit) or
\code{"notclose"} (test of not-close fit). See \emph{Details}.}

\item{eps0}{numeric. The null-hypothesis RMSEA. Default is \code{0.05}.}

\item{eps1}{numeric. The alternative-hypothesis RMSEA (the true RMSEA power is
evaluated at). Default is \code{0.08} for \code{type = "close"} and \code{0.01} for
\code{type = "notclose"}.}

\item{N}{numeric. In \code{"rmsea"} mode, the total sample size across groups (the
plain sample size when \code{group} is \code{1}): give \code{N} to compute power, or leave it
\code{NULL} to solve for the required \code{N} at a target \code{power}. In \code{"simulation"}
mode \code{N} is required and is the size of each drawn sample, with no sample size
solved for and no \code{group} division.}

\item{p}{numeric. The number of observed variables. In \code{"rmsea"} mode, used with
\code{k} to derive \code{df} when \code{df} is not given directly. In \code{"simulation"} mode it is
read off the population, so leave it unset (or matching \code{nrow(Lambda)} / \code{nrow(R)}).}

\item{k}{numeric. The number of factors. In \code{"rmsea"} mode, used with \code{p} to
derive \code{df} when \code{df} is not given directly. In \code{"simulation"} mode it is the
true number of factors: it is required with an \code{R} population and must be left
unset (or match \code{ncol(Lambda)}) with a factor-model population.}

\item{df}{numeric. The model degrees of freedom. Either supply \code{df} directly or
supply both \code{p} and \code{k}, from which \code{df = ((p - k)^2 - (p + k)) / 2}. Must be
positive.}

\item{alpha}{numeric. The significance level. Default is \code{0.05}.}

\item{power}{numeric. The target power. Give \code{power} (or leave both \code{power} and
\code{N} \code{NULL}, defaulting to \code{0.80}) to solve for the required \code{N}; leave it
\code{NULL} while giving \code{N} to compute power. Exactly one of \code{N} and \code{power} is
solved for.}

\item{group}{numeric. The number of groups. Default is \code{1}. \code{N} is the total
across all \code{group} groups, not the size of each one, and a solved \code{N} is a
multiple of \code{group}. See \emph{Details}.}

\item{Lambda}{matrix. Simulation mode. A \code{p} by \code{k_true} population loading matrix.
Supply this (optionally with \code{Phi}/\code{Psi}) to build a factor-model population;
structure recovery is available only with this form. Passed to \code{\link[=efa_simulate]{efa_simulate()}}.}

\item{Phi}{matrix. Simulation mode. The \code{k_true} by \code{k_true} population factor
intercorrelations. Only used with \code{Lambda}; defaults to orthogonal factors. When
\code{rotation} is unset, an oblique \code{Phi} selects a \code{"promax"} recovery fit and an
orthogonal one a \code{"varimax"} fit.}

\item{Psi}{numeric or matrix. Simulation mode. The population unique variances (a
length-\code{p} vector or a \code{p} by \code{p} matrix). Only used with \code{Lambda}. Passed to
\code{\link[=efa_simulate]{efa_simulate()}}.}

\item{R}{matrix. Simulation mode. A \code{p} by \code{p} population correlation matrix to draw
from directly, instead of a factor model. Structure recovery is not available for
this form (there are no population loadings to recover), and \code{k} is required.}

\item{n_datasets}{numeric. Simulation mode. The number of samples to draw and
analyse. Default is \code{500}.}

\item{criteria}{character. Simulation mode. The factor-retention criteria to
evaluate the hit-rate for, any of \code{"CD"}, \code{"EKC"}, \code{"HULL"}, \code{"KGC"}, \code{"MAP"},
\code{"NEST"}, \code{"PARALLEL"}, and \code{"SMT"} (see \code{\link[=efa_retain]{efa_retain()}}). Default is
\code{c("EKC", "MAP")}. Criteria that simulate internally (\code{"CD"}, \code{"HULL"},
\code{"NEST"}, \code{"PARALLEL"}) make each run substantially slower.}

\item{estimator}{character. Simulation mode. The estimator (\code{"PAF"}, \code{"ML"},
or \code{"ULS"}) used for the recovery fit and the retention criteria. Default
is \code{"PAF"}.}

\item{rotation}{character. Simulation mode. The rotation for the recovery fit,
passed to \code{\link[=efa_fit]{efa_fit()}}. Default is \code{NULL}, which matches the population: \code{"varimax"}
for orthogonal factors and \code{"promax"} for oblique ones (a single factor is left
unrotated). Recovery aligns the fitted loadings to the population pattern by
permutation and sign only. A rotation that does not seek that structure -- for
example \code{"none"} with more than one factor -- will understate recovery, so keep
the default (or another structure-seeking rotation) for a meaningful recovery
rate.}

\item{recovery_threshold}{numeric. Simulation mode. The matched-factor Tucker
congruence a replicate must reach to count as recovered. Default is \code{0.95}:
Lorenzo-Seva and ten Berge (2006) treat congruence at or above this level as
indicating the same factor.}

\item{model_error}{character. Simulation mode. The \code{\link[=efa_simulate]{efa_simulate()}} method that
perturbs the population with model error: \code{"TKL"} (Tucker-Koopman-Linn, the
default here), \code{"CB"} (Cudeck-Browne), \code{"WB"} (Wu-Browne), or \code{"none"} for an
exact population. It only takes effect when a target is supplied (\code{target_rmsea}
and/or \code{target_cfi}), and only for a factor-model population; without a target
the population stays exact whatever the method. \code{"TKL"} adds minor common
factors, giving a realistically imperfect population but lowering both the
hit-rate and structure recovery; \code{"CB"} and \code{"WB"} target the RMSEA only. \code{"CB"}
keeps the population loadings as the exact minimizer of the perturbed
population, so recovery stays close to perfect; \code{"WB"}'s loadings are not the
minimizer, so they carry no such guarantee. Note that \code{\link[=efa_simulate]{efa_simulate()}} itself
defaults to \code{"CB"}: the same \code{target_rmsea} passed to both functions gives an
easier population there, unless \code{model_error} is also set explicitly here.}

\item{target_rmsea}{numeric. Simulation mode. The population RMSEA the model should
have relative to the perturbed population, activating model error. Default is
\code{NULL}. Passed to \code{\link[=efa_simulate]{efa_simulate()}}.}

\item{target_cfi}{numeric. Simulation mode. The population CFI target (only with
\code{model_error = "TKL"}). Default is \code{NULL}. Passed to \code{\link[=efa_simulate]{efa_simulate()}}.}

\item{seed}{numeric. Simulation mode. Optional seed making the draws and analysis
reproducible and worker-count independent; the caller's random-number stream is
restored afterwards. Default is \code{NULL}.}
}
\value{
An object of class \code{efa_power}. For \code{mode = "rmsea"}, a list containing:
\item{power}{The power of the test at \code{N} (the achieved power, which for a
solved sample size is at least the target).}
\item{N}{The total sample size across groups: the supplied \code{N}, or the solved
required sample size (a multiple of \code{group}).}
\item{N_per_group}{The per-group sample size \code{N / group}, equal to \code{N} when
\code{group} is \code{1}. A whole number for a solved \code{N}; for a supplied \code{N} that is not
a multiple of \code{group} it is the fraction that the noncentrality uses.}
\item{crit}{The critical chi-square value the fit statistic is compared against.}
\item{ncp}{The noncentrality parameters under the null (\code{H0}, from \code{eps0}) and
the alternative (\code{H1}, from \code{eps1}).}
\item{solve_for}{\code{"power"} or \code{"N"}, recording which quantity was solved for.}
\item{settings}{A list of the inputs: \code{mode}, \code{type}, \code{eps0}, \code{eps1}, \code{df},
\code{p}, \code{k}, \code{alpha}, \code{group}, and the target \code{power} (the value solved to when
\code{solve_for} is \code{"N"}, otherwise \code{NULL}).}

For \code{mode = "simulation"}, a list containing:
\item{hit_rate}{A named numeric vector of the retention hit-rate per criterion (and,
where a criterion has several variants, per variant); \code{NA} for a criterion that
returned no suggestion on any replicate.}
\item{hits}{A data frame with one row per criterion (\code{criterion}) giving the
number of replicates it returned a definite suggestion on (\code{n_valid}), the
number of those that matched \code{k_true} (\code{hits}), and the \code{hit_rate}
(\code{hits / n_valid}).}
\item{recovery}{For a factor-model population, a list with the structure-recovery
rates (\code{min_rate}, \code{mean_rate}), the \code{threshold}, and the number of usable fits
(\code{n_valid}); \code{NULL} for an \code{R} population. Rates are over every replicate whose fit
returned loadings, including non-converged or Heywood solutions (their rates are
reported separately in \code{convergence}).}
\item{convergence}{A list with the number of datasets (\code{n_datasets}), the number
of fits that completed (\code{n_fit_ok}), how many of those converged
(\code{n_converged}) and how many produced a Heywood case (\code{n_heywood}), and the
corresponding rates: \code{fit_rate} (fits completed, over all datasets) and
\code{convergence_rate} / \code{heywood_rate} (converged / Heywood, over the completed
fits).}
\item{replicates}{The raw per-replicate values: the suggested factor counts
(\code{n_hat}), the matched congruences (\code{rec_min}, \code{rec_mean}), the \code{converged},
\code{heywood}, and \code{fit_ok} flags, and \code{fit_error}, the message of the fit that did
not complete (\code{NA} where it did).}
\item{k_true}{The true number of factors.}
\item{model_error}{The \code{\link[=efa_simulate]{efa_simulate()}} model-error record, or \code{NULL}.}
\item{settings}{A list of the simulation inputs.}
}
\description{
Analyses power for exploratory factor analysis, in one of two modes chosen with
\code{mode}.

\code{mode = "rmsea"} (the default) gives the analytic power of the root mean square
error of approximation (RMSEA) tests of close and not-close fit (MacCallum,
Browne, & Sugawara, 1996). Give a sample size to get the power of the test, or
give a target power to get the sample size needed to reach it.

\code{mode = "simulation"} runs a Monte-Carlo study: it draws \code{n_datasets} samples from
a known population (via \code{\link[=efa_simulate]{efa_simulate()}}), analyses each one, and reports how well
the analysis recovers that population. See \emph{Details} for what is reported.

Here the number of variables is \code{p} and the number of factors is \code{k} (elsewhere in
the package: \code{n_vars} and \code{n_factors}).
}
\section{RMSEA mode}{
Power rises with a larger sample, a larger model (more degrees of freedom), and a
bigger gap between the null and alternative RMSEA (MacCallum, Browne, & Sugawara,
1996).

Two tests are supported, chosen with \code{type} (never by the order of \code{eps0} and
\code{eps1}):
\describe{
\item{\code{"close"}}{Tests \emph{close fit} (MacCallum et al., 1996). The null
hypothesis is that the fit is close (RMSEA \eqn{\le} \code{eps0}; conventionally
0.05). Power is the chance of detecting a worse alternative (\code{eps1};
conventionally 0.08, so \code{eps0 < eps1}), in the upper tail.}
\item{\code{"notclose"}}{Tests \emph{not-close fit}. The null hypothesis is that the fit
is not close (RMSEA \eqn{\ge} \code{eps0}). Power is the chance of detecting a
better alternative (\code{eps1}; conventionally 0.01, so \code{eps0 > eps1}), in the
lower tail.}
}
When \code{eps0} and \code{eps1} are in the wrong order for the chosen \code{type}, a message is
shown but the requested test still runs. Equal \code{eps0} and \code{eps1} leave nothing to
detect and are an error.

Power always increases with \code{N}, so the required sample size (the smallest \code{N}
reaching \code{power}) is found by bisection. \code{N} is the \strong{total} sample size across
groups: with \code{group > 1} the power calculation divides by \code{group} (the
\code{1 / group} factor), so spreading a fixed total over more groups gives less
power. The matching per-group sample size, \code{N / group}, is returned as
\code{N_per_group}.

The \code{1 / group} factor makes all \code{group} groups the same size, so a required
total is rounded up to the next multiple of \code{group}. A solved \code{N_per_group} is thus
a whole number of persons, and the reported \code{power} is the power at a total that a
study can collect. With \code{group = 2} and \code{df = 102}, for example, the required total
is 260, or 130 per group. Bisection on the total alone gives 259, which asks for
129.5 persons in each group.
}

\section{Simulation mode}{
The population is passed to \code{\link[=efa_simulate]{efa_simulate()}}, which draws \code{n_datasets} samples of
size \code{N} from it. The population's true number of factors, \code{k_true}, is
\code{ncol(Lambda)} for a factor-model population, or \code{k} for a bare \code{R}. By default
the population fits the factor model exactly, which overstates how well the
criteria and the fit recover its structure; setting a misfit target makes the
population more realistic (MacCallum, 2003).

Each replicate is analysed three ways:
\describe{
\item{\strong{Hit-rate}}{The share of replicates where a criterion's suggested factor
count (from \code{criteria}) matches \code{k_true}. A replicate where the criterion
errored or gave no answer is left out of this count -- it does not count as a
miss.}
\item{\strong{Structure recovery} (factor-model populations only)}{The \code{k_true}-factor
model is fitted with \code{\link[=efa_fit]{efa_fit()}}, its loadings are matched to the population
loadings, and the matched-factor Tucker congruences (Lorenzo-Seva & ten Berge,
2006) are compared with \code{recovery_threshold}. A replicate succeeds when its
smallest (\code{min}) or average (\code{mean}) matched congruence reaches the
threshold.}
\item{\strong{Convergence}}{Among the replicates whose fit completed, the share that
converged and the share that produced a Heywood case.}
}
A replicate whose fit fails completely is not counted in any of the three
measures above. If any fit fails, a warning reports how many failed and the cause
of the first failure.

Replicates are analysed in parallel with \pkg{future.apply}; choose a parallel
plan with \code{\link[future:plan]{future::plan()}}. Each replicate uses its own reproducible
random-number stream, so with a fixed \code{seed} the result does not depend on the
number of workers, and the caller's random-number state is left unchanged.
}

\examples{
# Power of the test of close fit at N = 200 for a 100-df model
efa_power(df = 100, N = 200)

# Deriving df from the model dimensions instead of giving it directly
efa_power(p = 20, k = 3, N = 200)

# Required total sample size for 80\% power
efa_power(df = 100, power = 0.80)

# Test of not-close fit
efa_power(df = 100, N = 200, type = "notclose")

\donttest{
# Simulation mode: retention hit-rate and structure recovery for a known
# three-factor population at N = 300
efa_power("simulation", Lambda = population_models$loadings$baseline,
          Phi = population_models$phis_3$moderate, N = 300,
          n_datasets = 50, criteria = c("EKC", "MAP"), seed = 42)
}

}
\references{
MacCallum, R. C., Browne, M. W., & Sugawara, H. M. (1996). Power analysis and
determination of sample size for covariance structure modeling. \emph{Psychological
Methods, 1}(2), 130-149. \doi{10.1037/1082-989X.1.2.130}

MacCallum, R. C. (2003). 2001 Presidential Address: Working with imperfect models.
\emph{Multivariate Behavioral Research, 38}(1), 113-139. \doi{10.1207/S15327906MBR3801_5}

Lorenzo-Seva, U., & ten Berge, J. M. F. (2006). Tucker's congruence coefficient as a
meaningful index of factor similarity. \emph{Methodology, 2}(2), 57-64.
\doi{10.1027/1614-2241.2.2.57}
}
\seealso{
\code{\link[=efa_simulate]{efa_simulate()}} draws the replicate datasets used in simulation mode.
\code{\link[=efa_retain]{efa_retain()}} implements the retention criteria whose hit-rates simulation
mode reports.

Other power analysis:
\code{\link[=plot.efa_power]{plot.efa_power()}},
\code{\link[=print.efa_power]{print.efa_power()}}
}
\concept{power analysis}
